# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   05-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import math
import tkinter as tk
from tkinter import ttk, messagebox
try:
	import PIL.Image
except ImportError:
	PIL = None

import bmath
import icons
import Ribbon
import tkExtra
import tkFlair
try:
	import ZoomImage
except ImportError:
	pass

import Plot
import PlotEngine
from palette import paletteStore

_GEO_NO    = "-No-"
_GEO_AUTO  = "-Auto-"

_AXES = ["x","y","z","cb","x2","y2","img"]

_NORM_BALLOON = "Normalization factor (eg. 1e11) or a constant expression "\
		"(eg. 7e12*1.6e-10) or an expression using variable 'x' "\
		"for the value of the bin (eg. 7e12*1.6e-10*x/4.2)"

_NEW         = 0
_TOPLEFT     = 1
_TOP         = 2
_TOPRIGHT    = 3
_RIGHT       = 4
_BOTTOMRIGHT = 5
_BOTTOM      = 6
_BOTTOMLEFT  = 7
_LEFT        = 9

#-------------------------------------------------------------------------------
def setFloat(entry, value):
	value = float(value)
	if abs(value) < 1e-15: value=0.0
	s = bmath.format(value,10)

	# remember state
	state = entry["state"]
	entry["state"] = tk.NORMAL

	# set value
	entry.delete(0,tk.END)
	entry.insert(0,s)

	# restore state
	entry["state"] = state

#===============================================================================
# Base of Plot Frames
#===============================================================================
class BasePlotFrame(tk.Frame):
	# ----------------------------------------------------------------------
	def __init__(self, master, flair):
		super().__init__(master)
		self.flair    = flair
		self.project  = flair.project
		self.widget   = {}
		self.variable = {}
		self._update  = False
		self.plot     = None

	# ----------------------------------------------------------------------
	def type(self):
		raise Exception("BasePlotFrame.type not overridden")

	# ----------------------------------------------------------------------
	# set fields
	# ----------------------------------------------------------------------
	def set(self, plot):
		if plot is None: return
		self._update = True
		self.plot = plot
		for name, widget in self.widget.items():
			if name[0]=="@": continue
			if isinstance(widget,tuple): widget = widget[0]
			var = self.variable.get(name)
			if var is not None:
				value = self.plot(name)
				if value in ("True","False"): value = (value=="True")
				if isinstance(var,(tk.IntVar, tk.BooleanVar)):
					try:
						var.set(int(value))
					except ValueError:
						var.set(0)
				else:
					var.set(value)
			elif isinstance(widget, tk.Entry):
				widget.delete(0, tk.END)
				widget.insert(0, str(self.plot(name)))
			elif isinstance(widget, tk.Text):
				widget.delete("1.0", tk.END)
				widget.insert("1.0", str(self.plot(name)))
			elif isinstance(widget, tkExtra.Combobox):
				value = self.plot(name)
				if value:
					widget.set(value)
				elif widget.size()>0:
					widget.select(0)
				else:
					widget.set("")
		self._update = False

	# ----------------------------------------------------------------------
	# get fields
	# ----------------------------------------------------------------------
	def get(self, plot):
		if plot is None: return
		# Get values from widgets
		for name,widget in self.widget.items():
			if name[0]=="@": continue
			if isinstance(widget,tuple):
				for w in widget:
					if w.cget("state") != tk.DISABLED:
						break
				else:
					continue
				widget = widget[0]
			elif widget.cget("state") == tk.DISABLED:
				continue

			var = self.variable.get(name)
			if var is not None:
				try:
					value = var.get()
				except:
					value = ""
			elif isinstance(widget, tk.Entry):
				value = widget.get()
			elif isinstance(widget, tk.Text):
				value = widget.get(1.0, tk.END).strip()
			elif isinstance(widget, tkExtra.Combobox):
				value = widget.get()
			else:
				continue

			plot[name] = value

	# ----------------------------------------------------------------------
	# update fields after a plot
	# ----------------------------------------------------------------------
	def afterPlot(self):
		pass

	# ----------------------------------------------------------------------
	# setState in all widgets
	# ----------------------------------------------------------------------
	def setState(self, state):
		#for name,widget in self.widget.items():
		#	say(">>>",name)
		for widget in self.widget.values():
			if isinstance(widget,tuple):
				for w in widget: w.config(state=state)
			else:
				widget.config(state=state)

	# ----------------------------------------------------------------------
	def toggleField(self, event):
		if event.widget["state"] != tk.DISABLED:
			state = tk.DISABLED
		else:
			state = tk.NORMAL

		if isinstance(event.widget.master, tkExtra.Combobox):
			event.widget.master.config(state=state)
		else:
			event.widget.config(state=state)

		# Search linked list of widgets
		for widget in self.widget.values():
			if isinstance(widget,tuple) and event.widget in widget:
				for w in widget:
					w.config(state=state)

	#-----------------------------------------------------------------------
	def bindToggle(self):
		for widget in self.widget.values():
			if isinstance(widget,tuple):
				for w in widget: w.bind("<Button-3>", self.toggleField)
			elif isinstance(widget,tkExtra.Combobox):
				widget.bindWidgets("<Button-3>", self.toggleField)
			else:
				widget.bind("<Button-3>", self.toggleField)

	# ----------------------------------------------------------------------
	def styleFrame(self, frame):
		row,col = 0,0
		tk.Label(frame, text="Type:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.plot_with = tkExtra.Combobox(frame, width=10)
		self.plot_with.fill(Plot.STYLE)
		self.plot_with.set(Plot.STYLE[Plot.STYLE_DEF])
		self.plot_with.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(self.plot_with, "Plotting style")
		self.widget["with.0"] = self.plot_with

		# ===
		row += 1
		col = 0
		tk.Label(frame, text="Color:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.plot_lc = tkExtra.Combobox(frame, width=15)
		self.plot_lc.fill(PlotEngine._COLOR_NAMES)
		self.plot_lc.grid(row=row,column=col,sticky=tk.EW)
		tkExtra.Balloon.set(self.plot_lc, "Line color if gnuplot>=4.2")
		self.widget["lc.0"] = self.plot_lc

		# ---
		col += 1
		tk.Label(frame, text=" Line width:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.plot_lw = tk.IntVar()
		self.plot_lw.set(1)
		self.plot_lw_spin = tk.Spinbox(frame, text=self.plot_lw, from_=1, to=6,
				background="White",width=3)
		self.plot_lw_spin.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(self.plot_lw_spin, "Line width")
		self.widget["lw.0"] = self.plot_lw_spin
		self.variable["lw.0"] = self.plot_lw

		# ===
		row, col = row+1, 0
		tk.Label(frame, text=" Point type:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.plot_pt = tkExtra.Combobox(frame, width=15)
		self.plot_pt.fill(PlotEngine.POINTS)
		self.plot_pt.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(self.plot_pt, "Point type to use")
		self.widget["@pt.0"] = self.plot_pt

		# ---
		col += 1
		tk.Label(frame, text="Point size:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.plot_ps = tk.IntVar()
		self.plot_ps.set(1)
		self.plot_ps_spin = tk.Spinbox(frame, text=self.plot_ps, from_=1, to=6,
				background="White",width=3)
		self.plot_ps_spin.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(self.plot_ps_spin, "Point size")
		self.widget["ps.0"] = self.plot_ps_spin
		self.variable["ps.0"] = self.plot_ps

		return frame

#===============================================================================
# Common of Plot Frames
#===============================================================================
class CommonPlotFrame(BasePlotFrame):
	# ----------------------------------------------------------------------
	def __init__(self, master, flair):
		super().__init__(master, flair)
		#splitter = tkExtra.VSplitter(self,0.7)
		#splitter.pack(expand=tk.YES, fill=tk.BOTH)

		# === Top frame with common info ===
		frame = tk.Frame(self) #splitter.topFrame())
		frame.pack(side=tk.TOP, fill=tk.X)

		f = tk.Frame(frame)
		f.pack(side=tk.TOP, fill=tk.X)

		#tk.Label(f, text="Title:").pack(side=tk.LEFT)

		self.titleOptions = tk.IntVar()
		b = tk.Checkbutton(f, text="Title: ", #+"\N{BLACK DOWN-POINTING TRIANGLE}",
				variable=self.titleOptions,
				selectcolor="LightYellow",
				indicatoron=tk.FALSE,
				command=self.showTitleOptions)
		tkExtra.Balloon.set(b, "Show plot options")
		b.pack(side=tk.LEFT)

		b = tk.Entry(f, background="White")
		tkExtra.Balloon.set(b, "Plot Title")
		b.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X)
		self.widget["title"] = b

		var = tk.IntVar()
		var.set(0)
		tk.Label(f, text="Display:").pack(side=tk.LEFT)
		b = tk.Spinbox(f, text=var, from_=0,
			to=100, background="White", width=3)
		tkExtra.Balloon.set(b, "Gnuplot display window index. "
			"Change index to display many windows")
		b.pack(side=tk.LEFT)
		self.widget["index"] = b
		self.variable["index"] = var

		# ==== Title Options Frame ====
		self.titleOptionsFrame = tk.LabelFrame(frame, text="Options ", foreground="DarkBlue")

		# --- 1st row ---
		f = tk.Frame(self.titleOptionsFrame)
		f.pack(side=tk.TOP, fill=tk.X)
		self.optionsFrame(f, "title")

		# --- 2nd row ---
		f = tk.Frame(self.titleOptionsFrame)
		f.pack(side=tk.TOP, fill=tk.X)

		# - grid -
		var = tk.IntVar()
		var.set(0)
		b = tk.Checkbutton(f, text="grid",
				foreground="DarkBlue",
				variable=var,
				anchor=tk.W)
		b.pack(side=tk.LEFT, padx=5)
		tkExtra.Balloon.set(b,"Display plot grid")
		self.widget["grid"] = b
		self.variable["grid"] = var

		# - aspect -
		tk.Label(f, text=" aspect:", foreground="DarkBlue").pack(side=tk.LEFT)
		b = tkExtra.FloatEntry(f, background="white", width=7)
		tkExtra.Balloon.set(b, "Plotting aspect ratio")
		b.pack(side=tk.LEFT)
		self.widget["ratio"] = b

		# - auto -
		b = tk.Button(f, text="Auto", foreground="DarkBlue",
				command=self.calcAspectRatio,
				pady=1)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b,
			"Calculate aspect ratio, using the plot range information. "
			"Make a plot before clicking")

		# line type
		tk.Label(f, text="  lines:", foreground="DarkBlue").pack(side=tk.LEFT)
		b = tkExtra.Combobox(f, width=6)
		b.fill(Plot.LINE_TYPES)
		b.set(Plot.LINE_TYPES[0])
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b,"Line type style dashed or solid")
		self.widget["linetype"] = b

		# === Top frame with common info ===
		topframe = tk.Frame(self) #splitter.topFrame())
		topframe.pack(fill=tk.X)

		groupframe = tk.LabelFrame(topframe, text="Axes")
		groupframe.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES, padx=3, pady=3)

		row = 0
		col = 0

		self.axesButton = tk.Button(groupframe,
				text="\N{BLACK DOWN-POINTING TRIANGLE}",
				command=self.axesMenu,
				width=3,
				padx=0,
				pady=0)
		self.axesButton.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self.axesButton, "Select axes to display")
		col += 1
		for label,anchor,cmd,hlp in [
				("Label", tk.W     , None, None),
				("Log"  , tk.W     , None, None),
				("Min"  , tk.CENTER, self.getLimits, "Get limits from plot window" ),
				(None   , None     , None, None),
				("Max"  , tk.CENTER, self.getLimits, "Get limits from plot window" )]:
			if label:
				if cmd:
					b = tk.Button(groupframe,
						text=label,
						padx=0, pady=0,
						command=cmd,
						anchor=anchor)
					b.grid(row=row, column=col)
					tkExtra.Balloon.set(self.axesButton, hlp)
				else:
					tk.Label(groupframe,
						text=label,
						anchor=anchor).grid(row=row, column=col, sticky=tk.NSEW)
			col += 1

		# Axes frame
		self.axesVar = [("z", tk.IntVar()),
				("cb",tk.IntVar()),
				("x2",tk.IntVar()),
				("y2",tk.IntVar()),
				("img",tk.IntVar())]

		# Options
		self.axesOptions = []

		for axis in _AXES:
			row += 1
			col  = 0

			#b = tk.Label(groupframe, text=f"{axis}:", anchor=tk.E)
			#b.grid(row=row, column=col, sticky=tk.EW)
			#self.widget[f"{axis}name"] = b

			# Options button
			optvar = tk.IntVar()
			b = tk.Checkbutton(groupframe, text=f"{axis}:",
					#text="\N{BLACK DOWN-POINTING TRIANGLE}",
					anchor=tk.E,
					variable=optvar,
					selectcolor="LightYellow",
					indicatoron=tk.FALSE,
					width=4,
					command=self.axesShowOptions)
			tkExtra.Balloon.set(b, "Show axis options")
			self.widget[f"{axis}name"] = b
			b.grid(row=row, column=col, sticky=tk.EW)

			if axis=="img":
				# File instead of label
				col += 1
				self._backgroundImage = tkExtra.FileEntry(groupframe,
						title="Load background image",
						initialdir=self.project.dir(),
						filetypes=[
							("All images", ("*.png", "*.gif", "*.jpg", "*.jpeg")),
							("PNG - Portable Network Graphics", "*.png"),
							("GIF - Graphics Interchange Format", "*.gif"),
							("JPG - Joint Photographic Experts Group", ("*.jpg","*.jpeg")),
							("All","*")],
						background="White")
				self._backgroundImage.grid(row=row, column=col, sticky=tk.EW)
				tkExtra.Balloon.set(self._backgroundImage, "Select background image")
				self.widget[f"{axis}file"] = self._backgroundImage
				self.axesOptions.append((axis,None,row,None))

				# Calibrate image
				col += 2
				self._calibrate = tk.Button(groupframe,
							text="Calibrate",
							command=self._calibrateImage)
				self._calibrate.grid(row=row, column=col, sticky=tk.EW)
				continue
			else:
				# Label
				col += 1
				b = tk.Entry(groupframe, background="White")
				b.grid(row=row, column=col, sticky=tk.EW)
				tkExtra.Balloon.set(b, f"{axis} axis label")
				self.widget[f"{axis}label"] = b

			# Log
			col += 1
			var = tk.IntVar()
			var.set(False)
			b = tk.Checkbutton(groupframe,
					text="",
					variable=var,
					anchor=tk.W)
			b.grid(row=row, column=col, sticky=tk.EW)
			tkExtra.Balloon.set(b, "Log/Linear")
			self.widget[f"{axis}log"] = b
			self.variable[f"{axis}log"] = var

			if axis == "cb":	# Special handling for color band
				self.cblogVar = var
				var.trace("w", self.cpdChange)

			# Limits:  min - max
			col += 1
			if axis == "cb":	# Special handling for color band
				self.cb_min_var = var = tk.StringVar()
				self.cb_min_var.trace("w", self.cpdChange)
			else:
				var = None

			b = tkExtra.FloatEntry(groupframe,
					textvariable=var,
					width=15,
					background="White")
			b.grid(row=row, column=col, sticky=tk.EW)
			tkExtra.Balloon.set(b, f"{axis} axis minimum range")
			self.widget[f"{axis}min"] = b

			col += 1
			b = tk.Label(groupframe, text="-")
			b.grid(row=row, column=col, sticky=tk.EW)
			self.widget[f"{axis}dash"] = b

			col += 1
			b = tkExtra.FloatEntry(groupframe, width=15, background="White")
			b.grid(row=row, column=col, sticky=tk.EW)
			b.bind("<Key>", self.cbmaxChange)
			if axis == "cb":
				tkExtra.Balloon.set(b,
					"cb axis maximum range, leave empty "
					"to be dynamically calculated")
			else:
				tkExtra.Balloon.set(b, f"{axis} axis maximum range")
			self.widget[f"{axis}max"] = b

			# Special handling for color band
			row += 1
			frame = tk.Frame(groupframe)
			# grid now now
			self.axesOptions.append((axis,optvar,row,frame))

			if axis == "cb":
				row += 1
				f = tk.Frame(frame)
				f.pack(side=tk.TOP, fill=tk.X)

				# ---
				var = tk.IntVar()
				var.set(1)
				b = tk.Checkbutton(f, text="Round", variable=var)
				b.pack(side=tk.RIGHT)
				tkExtra.Balloon.set(b, "Round lower limit to closest decade boundary")
				self.widget["cbround"]   = b
				self.variable["cbround"] = var
				var.trace("w", self.cpdChange)

				# ---
				var = tk.IntVar()
				var.set(3)
				b = tk.Spinbox(f, text=var,
						from_=0, to=50,
						background="White",
						width=3)
				tkExtra.Balloon.set(b, "Colors per decade")
				b.pack(side=tk.RIGHT)
				tk.Label(f, text="  CPD:", foreground="DarkBlue").pack(side=tk.RIGHT)
				self.widget["cbcpd"]   = b
				self.variable["cbcpd"] = var
				var.trace("w", self.cpdChange)

				# ---
				var = tk.IntVar()
				var.set(30)
				b = tk.Spinbox(f, text=var,
						from_=4, to=256,
						background="White",
						width=4)
				tkExtra.Balloon.set(b, "Total number of colors")
				b.pack(side=tk.RIGHT)
				tk.Label(f, text="  Colors:", foreground="DarkBlue").pack(side=tk.RIGHT)
				self.widget["cbcolors"]   = b
				self.variable["cbcolors"] = var
				var.trace("w", self.cpdChange)

				# ---
				b = tkExtra.Combobox(f, width=10)
				b.fill(sorted(paletteStore.keys()))
				b.set("fluka")
				b.pack(side=tk.RIGHT)
				tk.Label(f, text="Palette:", foreground="DarkBlue").pack(side=tk.RIGHT)
				self.widget["cbpalette"] = b

				# Replace frame with the axes options
				frame = tk.Frame(frame)
				frame.pack(side=tk.BOTTOM, fill=tk.X)

			# option field
			self.optionsFrame(frame, axis, 0, "label")
			self.optionsFrame(frame, axis+"tics", 1, "tics")

		# -----
		groupframe.grid_columnconfigure(1, weight=1)
		self.axesShow()

		# =====
		#self.plot_keys = IntVar()
		#self.plot_keys.set(1)
		#b = tk.Checkbutton(frame, text="legend", variable=self.plot_keys,
		#		anchor=tk.W)
		#b.pack(fill=tk.X)
		#tkExtra.Balloon.set(b, "Display label keys")

		## =====
		#tk.Label(frame, text="Width:").grid(row=row, column=col, sticky=tk.E)
		#col += 1
		#self.plot_sizex = tkExtra.FloatEntry(frame, background="white", width=4)
		#tkExtra.Balloon.set(self.plot_sizex, "Plotting width")
		#self.plot_sizex.grid(row=row, column=col, sticky=tk.EW)

		#tk.Label(frame, text="Height:").grid(row=row, column=col, sticky=tk.E)
		#self.plot_sizey = tkExtra.FloatEntry(frame, background="white", width=4)
		#tkExtra.Balloon.set(self.plot_sizey, "Plotting height")
		#self.plot_sizey.grid(row=row, column=col, sticky=tk.EW)

		# === Common Main Frame ===
		self.common = tk.Frame(self) #splitter.topFrame())
		self.common.pack(fill=tk.BOTH)

		# === Commands frame with common info ===
		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM,fill=tk.BOTH, expand=tk.YES, pady=3)

		b = tk.Text(frame, #splitter.bottomFrame(),
				background="white", height=5, undo=True)
		b.pack(side=tk.LEFT,fill=tk.BOTH, expand=tk.YES)
		self.widget["commands"] = b
		b.bind("<Control-Return>",   self.execute)
		b.bind("<Control-KP_Enter>", self.execute)

		sb = ttk.Scrollbar(frame, #splitter.bottomFrame(),
				takefocus=False, command=b.yview)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		b.config(yscrollcommand=sb.set)

		self.bindToggle()

	# ----------------------------------------------------------------------
	def execute(self, event=None):
		self.flair.event_generate("<<Execute>>")
		return "break"

	# ----------------------------------------------------------------------
	def optionsFrame(self, frame, prefix, row=0, label=""):
		col = 0
		if label:
			tk.Label(frame, text=label, fg="DarkBlue").grid(row=row, column=col, sticky=tk.E)

		# font
		col += 1
		tk.Label(frame, text="font:", fg="DarkBlue").grid(row=row, column=col, sticky=tk.E)
		col += 1
		b = tkExtra.Combobox(frame, width=12)
		b.grid(row=row, column=col, sticky=tk.EW)
		b.fill(Plot.FONTS)
		b.set(Plot.FONTS[0])
		tkExtra.Balloon.set(b,"Font family")
		self.widget[f"{prefix}font"] = b

		# size
		col += 1
		b = tkExtra.Combobox(frame, width=4)
		b.grid(row=row, column=col, sticky=tk.EW)
		b.fill(Plot.SIZE)
		b.set(Plot.SIZE[0])
		tkExtra.Balloon.set(b,"Font size")
		self.widget[f"{prefix}size"] = b

		# color
		col += 1
		tk.Label(frame, text="color:", fg="DarkBlue").grid(row=row, column=col, sticky=tk.EW)
		col += 1
		b = tkExtra.Combobox(frame, width=12)
		b.grid(row=row, column=col, sticky=tk.EW)
		b.fill(PlotEngine._COLOR_NAMES)
		b.set(PlotEngine._COLOR_NAMES[0])
		tkExtra.Balloon.set(b,"Font color")
		self.widget[f"{prefix}color"] = b

		# --- 3rd row ---
		# options
		col += 1
		tk.Label(frame, text="options:", fg="DarkBlue").grid(row=row, column=col, sticky=tk.EW)
		col += 1
		b = tk.Entry(frame, background="White")
		b.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Additional options")
		self.widget[f"{prefix}options"] = b

		frame.grid_columnconfigure(col, weight=1)

	# ----------------------------------------------------------------------
	def showTitleOptions(self):
		if self.titleOptions.get():
			self.titleOptionsFrame.pack(side=tk.BOTTOM, fill=tk.X)
		else:
			self.titleOptionsFrame.pack_forget()

	# ----------------------------------------------------------------------
	# Display axes menu to select from
	# ----------------------------------------------------------------------
	def axesMenu(self):
		menu = tk.Menu(self, tearoff=0)
		for axis,var in self.axesVar:
			menu.add_checkbutton(label=axis,  variable=var,  command=self.axesShow)
		menu.tk_popup(	self.axesButton.winfo_rootx(),
				self.axesButton.winfo_rooty() + self.axesButton.winfo_height())

	# ----------------------------------------------------------------------
	# Hide/Show axes
	# ----------------------------------------------------------------------
	def axesShow(self):
		for axis,var in self.axesVar:
			for label in ["name", "file", "label", "log", "min", "dash", "max"]: #, "show"]):
				name = axis+label
				try:
					if var.get(): # Show
						self.widget[name].grid()
					else:	# Hide
						self.widget[name].grid_remove()
				except KeyError:
					pass
				# Special treatment for img
				if axis=="img":
					if var.get():
						self._calibrate.grid()
					else:
						self._calibrate.grid_remove()
			if not var.get():
				for a,v,_,_ in self.axesOptions:
					if axis==a and v is not None:
						v.set(False)
		self.axesShowOptions()

	# ----------------------------------------------------------------------
	def axesShowOptions(self):
		for axis,var,row,frame in self.axesOptions:
			if frame is None:
				pass
			elif var.get():
				frame.grid(row=row, column=1, columnspan=6,sticky=tk.NSEW)
			else:
				frame.grid_forget()

	# ----------------------------------------------------------------------
	# set fields
	# ----------------------------------------------------------------------
	def set(self, plot):
		super().set(plot)
		# Special values
		for i,(axis, var) in enumerate(self.axesVar):
			axis = plot(f"{axis}tics")
			try:
				val = int(axis)
			except:
				val = int(axis=="True")
			if val == 0:
				# find option frame to remove
				a,v,r,f = self.axesOptions[i+2]
				if v is not None: v.set(0)
				if f is not None: f.grid_forget()
			var.set(val)
		self.axesShow()

	# ----------------------------------------------------------------------
	# get fields
	# ----------------------------------------------------------------------
	def get(self, plot):
		super().get(plot)
		# Special values
		for axis, var in self.axesVar:
			plot[f"{axis}tics"] = int(var.get())

	# ----------------------------------------------------------------------
	def getLimits(self):
		if not self.flair.plotEngine:
			self.flair.notify("Get limits",
				"You need to plot first before getting the extends from plot window",
				tkFlair.NOTIFY_ERROR)
			return
		xmin, xmax = self.flair.plotEngine.getRange("x")
		ymin, ymax = self.flair.plotEngine.getRange("y")

		if xmax is None: return

		if self.widget["xmin"]!=tk.DISABLED:
			self.widget["xmin"].set(xmin)
		if self.widget["xmax"]!=tk.DISABLED:
			self.widget["xmax"].set(xmax)
		if self.widget["ymin"]!=tk.DISABLED:
			self.widget["ymin"].set(ymin)
		if self.widget["ymax"]!=tk.DISABLED:
			self.widget["ymax"].set(ymax)

	# ----------------------------------------------------------------------
	# Override method to calculate aspect ratio from plot extends
	# ----------------------------------------------------------------------
	def calcAspectRatio(self):
		if not self.flair.plotEngine: return
		xlow, xhigh = self.flair.plotEngine.getRange("x")
		if xhigh is None: return
		ylow, yhigh = self.flair.plotEngine.getRange("y")
		if yhigh is None: return
		try: ratio = (yhigh-ylow) / (xhigh-xlow)
		except: return
		self.widget["ratio"].set(ratio)

	# ----------------------------------------------------------------------
	# CPD Change
	# ----------------------------------------------------------------------
	def cpdChange(self, *abc):
		if self._update: return
		state = tk.NORMAL if int(self.cblogVar.get()) else tk.DISABLED

		self.widget["cbcpd"].config(state=state)
		self.widget["cbround"].config(state=state)
		if state == tk.DISABLED: return

		try:    cpd = float(self.variable["cbcpd"].get())
		except: return
		if cpd < 1.0: return

		try:    cbmin  = float(self.widget["cbmin"].get())
		except: return
		try:    ncolors = float(self.widget["cbcolors"].get())
		except: return

		# round minimum to lowest round color value
		if self.variable["cbround"].get():
			try:
				cbmin = 10.0**(math.floor(math.log10(cbmin)*cpd)/cpd)
			except (ValueError, OverflowError):
				pass

		# Calculate max
		if cbmin>0 and ncolors>0: # and self.widget["cbmax"].get().strip():
			try:
				self.widget["cbmax"].set("%g"%(cbmin * 10.0**(ncolors/cpd)))
			except (ValueError, OverflowError):
				pass

	#-----------------------------------------------------------------------
	# When the cb Max is manually edited, set the CPD to 0 to avoid
	# any future recalculation
	#-----------------------------------------------------------------------
	def cbmaxChange(self, event=None):
		if self._update: return
		try:
			float(self.widget["cbmax"].get())
			self.variable["cbcpd"].set(0)
		except ValueError:
			pass

	#-----------------------------------------------------------------------
	def _calibrateImage(self):
		if PIL is None: return
		calibrate = BackgroundImageCalibration(self)
		filename = self._backgroundImage.filename()
		if not filename: return
		calibrate.loadImage(filename)
		xlow  = self.widget["x2min"].get()
		xhigh = self.widget["x2max"].get()
		ylow  = self.widget["y2min"].get()
		yhigh = self.widget["y2max"].get()
		calibrate.rectangle(xlow, xhigh, ylow, yhigh)
		if calibrate.show():
			self.widget["x2min"].set(calibrate.xmin())
			self.widget["x2max"].set(calibrate.xmax())
			self.widget["y2min"].set(calibrate.ymin())
			self.widget["y2max"].set(calibrate.ymax())

#===============================================================================
class BackgroundImageCalibration(tk.Toplevel):
	width  = -1
	height = -1

	#-----------------------------------------------------------------------
	def __init__(self, master, *options):
		super().__init__(master, *options)
		self.title("Calibrate Background Image")

		# --- working canvas ---
		frame = tk.Frame(self)
		frame.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

		self.canvas = ZoomImage.ZoomImage(frame,
				#width=640, height=480,
				takefocus=True)
		self.canvas.grid(row=0, column=0, sticky=tk.NSEW)

		sbv = ttk.Scrollbar(frame, orient=tk.VERTICAL)
		sbv.grid(row=0, column=1, sticky=tk.NSEW)
		sbv.config(command=self.canvas.yview)

		sbh = ttk.Scrollbar(frame, orient=tk.HORIZONTAL)
		sbh.grid(row=1, column=0, sticky=tk.NSEW)
		sbh.config(command=self.canvas.xview)

		self.canvas.config( yscrollcommand=sbv.set,
				    xscrollcommand=sbh.set)
		frame.grid_columnconfigure(0, weight=1)
		frame.grid_rowconfigure(0, weight=1)

		# --- canvas ---
		ribbon = tk.Frame(self)
		ribbon.place(x=0, y=0)

		# ===
		group = Ribbon.LabelGroup(ribbon, "View")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid2rows()

		b = Ribbon.LabelButton(group.frame,
				text="Zoom In",
				image=icons.get("zoom_in"),
				compound=tk.LEFT,
				command=self.canvas.zoomIn,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.grid(row=0, column=0, stick=tk.W)

		b = Ribbon.LabelButton(group.frame,
				text="Zoom Out",
				image=icons.get("zoom_out"),
				compound=tk.LEFT,
				command=self.canvas.zoomOut,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.grid(row=1, column=0, stick=tk.W)

		# ===
		group = Ribbon.LabelGroup(ribbon, "Confirm")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		b = Ribbon.LabelButton(group.frame,
				text="Ok",
				image=icons.get("ok32"),
				compound=tk.TOP,
				command=self.ok,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.pack(side=tk.LEFT, fill=tk.Y)

		b = Ribbon.LabelButton(group.frame,
				text="Cancel",
				image=icons.get("close32"),
				compound=tk.TOP,
				command=self.close,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.pack(side=tk.LEFT, fill=tk.Y)

		# Override zooming
		self.canvas.bindMotion()
		self.canvas.bindArrows()
		self.canvas.bindZoom()
		self.canvas.bind("<<Zoom>>",       self.drawRect)
		#self.canvas.bind("<<View>>",      self.move)
		#self.canvas.bind("<Motion>",      self.motion)
		self.canvas.bind("<Button-1>",     self.click)
		self.canvas.bind("<B1-Motion>",    self.b1motion)
		self.canvas.bind("<Key-Return>",   self.ok)
		self.canvas.bind("<Key-KP_Enter>", self.ok)
		self.canvas.bind("<Key-Escape>",   self.close)
		self.canvas.focus_set()

		self._rect = self.canvas.create_rectangle(10,10,100,100,
					outline="Magenta",
					width=2,
					dash=(4,2))

		# Global variables
		self.imagefile = ""
		self.filename  = None
		self.item      = None
		self.rc        = False

	#-----------------------------------------------------------------------
	def show(self):
		self.deiconify()
		#self.transient(self.master)
		if not self.winfo_ismapped():
			self.wait_visibility()
		if BackgroundImageCalibration.width > 0:
			self.geometry(f"{BackgroundImageCalibration.width}x{BackgroundImageCalibration.height}")
		else:
			self.geometry(f"{self.winfo_width()}x{self.winfo_height()}")
		self.wait_window()
		return self.rc

	#-----------------------------------------------------------------------
	def ok(self, event=None):
		self.rc = True
		self.close()

	#-----------------------------------------------------------------------
	def close(self, event=None):
		BackgroundImageCalibration.width  = self.winfo_width()
		BackgroundImageCalibration.height = self.winfo_height()
		self.destroy()

	#-----------------------------------------------------------------------
	def loadImage(self, fn):
		self.config(cursor="watch")
		try:
			image = PIL.Image.open(fn)
			image.load()
		except:
			self.config(cursor="")
			messagebox.showerror("Error opening file",
				f"Error opening image file {fn}",
				parent=self)
			return
		self.imagefile = fn
		self.image     = image
		self.canvas.setImage(image)
		self.config(cursor="")

	# ----------------------------------------------------------------------
	def setImage(self, img):
		self.canvas.setImage(img)

	#-----------------------------------------------------------------------
	def rectangle(self, xmin, xmax, ymin, ymax):
		try:    self._xmin = int(float(xmin))
		except: self._xmin = 0
		try:    self._xmax = int(float(xmax))
		except: self._xmax = self.image.width
		try:    self._ymin = self.image.height-int(float(ymax))
		except: self._ymin = 0
		try:    self._ymax = self.image.height-int(float(ymin))
		except: self._ymax = self.image.height
		self.drawRect()

	#-----------------------------------------------------------------------
	def xmin(self):	return self._xmin
	def xmax(self):	return self._xmax
	def ymin(self):	return self.image.height - self._ymax
	def ymax(self):	return self.image.height - self._ymin

	# ----------------------------------------------------------------------
	def drawRect(self, event=None):
		xmin = self._xmin * self.canvas.zoom
		xmax = self._xmax * self.canvas.zoom
		ymin = self._ymin * self.canvas.zoom+1
		ymax = self._ymax * self.canvas.zoom+1
		self.canvas.coords(self._rect, (xmin,ymin, xmax, ymax))

	# --------------------------------------------------------------------
	# Button1 clicked find item to change in clipping rectangle
	# --------------------------------------------------------------------
	def click(self, event):
		self.canvas.focus_set()
		x = self.canvas.canvasx(event.x)
		y = self.canvas.canvasy(event.y)
		M = 20
		if self._rect in self.canvas.find_overlapping(x-M, y-M, x+M, y+M):
			# Find item to drag
			x /= self.canvas.zoom
			y /= self.canvas.zoom
			if abs(y - self._ymin) <= M:	# On top line
				if  abs(x - self._xmin) <= M:
					self._item = _TOPLEFT
				elif abs(x  - self._xmax) <= M:
					self._item = _TOPRIGHT
				else:
					self._item = _TOP
			elif abs(y - self._ymax) <= M:	# On bottom line
				if  abs(x - self._xmin) <= M:
					self._item = _BOTTOMLEFT
				elif abs(x - self._xmax) <= M:
					self._item = _BOTTOMRIGHT
				else:
					self._item = _BOTTOM
			elif abs(x - self._xmin) <= M:
				self._item = _LEFT
			else:
				self._item = _RIGHT
		else:
			self._item = _NEW
		self._startx = x
		self._starty = y

	# --------------------------------------------------------------------
	# Draw clipping rectangle
	# --------------------------------------------------------------------
	def b1motion(self, event):
		x = self.canvas.canvasx(event.x)//self.canvas.zoom
		y = self.canvas.canvasy(event.y)//self.canvas.zoom
		if self._item in (_TOPRIGHT, _RIGHT, _BOTTOMRIGHT):
			x -= 1
		if self._item in (_BOTTOMLEFT, _BOTTOM, _BOTTOMRIGHT):
			y -= 1

		if self._item == _NEW:
			if abs(self._startx - self.canvas.canvasx(event.x))>3 or \
			   abs(self._starty - self.canvas.canvasy(event.y))>3:
				self._xmin = x
				self._ymin = y
				self._item = _BOTTOMRIGHT

		elif self._item == _TOPLEFT:
			self._xmin = x
			self._ymin = y

		elif self._item == _TOP:
			self._ymin = y

		elif self._item == _TOPRIGHT:
			self._xmax = x
			self._ymin = y

		elif self._item == _RIGHT:
			self._xmax = x

		elif self._item == _BOTTOMRIGHT:
			self._xmax = x
			self._ymax = y

		elif self._item == _BOTTOM:
			self._ymax = y

		elif self._item == _BOTTOMLEFT:
			self._xmin = x
			self._ymax = y

		elif self._item == _LEFT:
			self._xmin = x

		if self._item != _NEW: self.drawRect()
