#!/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   04-Oct-2012

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re
import sys
import tkinter as tk
from tkinter import ttk
from itertools import count

_LIB=os.path.join(os.path.abspath(os.path.dirname(__file__)), "lib")
if _LIB not in sys.path: sys.path.insert(0,_LIB)

import rexx
import bmath
import icons
import tkFlair
import tkExtra

import Ribbon
import Evaluation
import FlairRibbon

ERROR_COLOR    = "Red"
RESULT_COLOR   = "DarkGreen"
COMMENT_COLOR  = "Blue"

_ASSIGN = re.compile(r"^([A-Za-z_][A-Za-z0-9_]*)\s*=\s*(.*)$")
_ASSRES = re.compile(r"^([A-Za-z_][A-Za-z0-9_]*) = ([+\-]?\d*\.?\d*[eE]?[+\-]?\d*)$",re.MULTILINE)
_PATANS = re.compile(r"^> *(ans\d+) = .*$",re.MULTILINE)
_PATCOM = re.compile(r"^ *#.*$")
_PATERR = re.compile(r"^> Error: .*$",re.MULTILINE)

#===============================================================================
class Calculator(tk.Frame):
	def __init__(self, master, *args):
		super().__init__(master, *args)
		self.text = tk.Text(self, background="White")
		self.text.pack(expand=tk.YES, fill=tk.BOTH, side=tk.LEFT)
		sb = ttk.Scrollbar(self, orient=tk.VERTICAL, command=self.text.yview)
		self.text.config(yscrollcommand=sb.set, undo=True)
		sb.pack(fill=tk.Y, side=tk.RIGHT)

		self.bind("<<Evaluate>>",		self.evaluate)
		self.text.bind("<Return>",		self.evaluate)
		self.text.bind("<KP_Enter>",		self.evaluate)
		self.text.bind("<<Paste>>",		self.paste)
		self.text.bind("<Control-Key-a>",	self.selectAll)
		self.text.bind("<Shift-Key-Return>",	self.shiftReturn)
		self.localDict  = {}
		self.globalDict = Evaluation._globalDict
		self._insert    = None	# remember paste position

	# ----------------------------------------------------------------------
	def setInput(self, inp):
		self.localDict  = {} #inp.localDict[:]
		self.globalDict = inp.globalDict

	# ----------------------------------------------------------------------
	# Evaluate all lines
	# ----------------------------------------------------------------------
	def evaluate(self, event=None):
		if self._insert is not None:
			self.text.mark_set(tk.INSERT, self._insert)
			self._insert = None

		row, col  = list(map(int,self.text.index(tk.INSERT).split(".")))
		rowe,cole = list(map(int,self.text.index(tk.END).split(".")))

		if self.text.tag_names(tk.INSERT) in ("comment", "result", "error"):
			row += 1

		remain = self.text.get(f"{row+1}.0",tk.END)
		remain = remain.replace("\n","").strip()
		last = (remain=="")	# last line?
		if last:
			try:
				self.text.delete(f"{row+1}.0",tk.END)
				rowe=row+1
			except tk.TclError:
				pass

		# Cleanup old tags
		#for tag in ("comment", "result", "error"):
		#	self.text.tag_remove(tag, f"{row}.0", tk.END)

		# Re-evaluate all lines from current to the end
		#first = True
		while row < rowe:
			expr = self.text.get(f"{row}.0", f"{row}.end")
			#print(row,expr)
			comment = expr.find("#")
			if comment>=0:
				self.text.tag_add("comment", f"{row}.{comment}", f"{row}.end")
				expr = expr[:comment].strip()
				if not expr:
					row += 1
					continue

			if not expr:
				# skip empty lines
				row += 1
				continue

			if expr[0]==">":
				#if first and col<cole:
				#	# delete old result
				##	self.text.delete(f"{row}.0",f"{row}.end")
				#	rowe -= 1
				#else:
				row += 1
				continue
			#else:
			#	first = False

			# Get continuation lines
			r = row
			while r>0 and expr[:1]==" ":
				r -= 1
				line = self.text.get(f"{r}.0",f"{r}.end")
				expr = line[:line.find("#")]+expr

			# Possibly check previous line for continuation
			# Maybe first char could be space
			m = _ASSIGN.match(expr)
			if m:
				variable = m.group(1)
				expr     = m.group(2)

				if rexx.abbrev("PRECISION",variable,4):
					variable = "PRECISION"
				elif rexx.abbrev("USEEXP",variable,4):
					variable = "USEEXP"
			else:
				# check next line
				line = self.text.get(f"{row+1}.0",f"{row+1}.end")
				m = _PATANS.match(line)
				if m:
					variable = m.group(1)
				else:
					variable = self.findUniqueAns()

			# Check for possible vector definitions
			# replace any [ to Vector( and ] to )
			expr = expr.replace("{","Vector(")
			expr = expr.replace("}",")")

			# Check if it is a simple assignment
			try:
				result = float(expr)
				self.localDict[variable] = result
				self.localDict["ans"] = result
				row += 1
				# Do not append any result
				continue
			except:
				pass

			# Evaluate the expression
			try:
				result = eval(expr, self.globalDict, self.localDict)
				self.localDict[variable] = result
				self.localDict["ans"] = result
				tag = "result"
				if isinstance(result,float):
					prec = int(self.localDict.get("PRECISION",17))
					uexp = int(self.localDict.get("USEEXP",False))
					result = bmath.format(result, prec, useExp=uexp)
				result = f"{variable} = {result}"
			except Exception as err:
				result = f"Error: {err}"
				tag    = "error"

			# Append result in next line
			row += 1
			line = self.text.get(f"{row}.0", f"{row}.end").strip()
			if line and line[0]==">":
				self.text.delete(f"{row}.0", f"{row}.end")
				self.text.insert(f"{row}.0", f"> {result}", tag)
			else:
				prefix = "\n" if row==rowe else ""
				if last:
					self.text.insert(f"{row}.0",f"{prefix}> {result}", tag)
				else:
					self.text.insert(f"{row}.0",f"{prefix}> {result}\n", tag)
				rowe += 1
			row += 1

		self.text.tag_config("comment", foreground=COMMENT_COLOR)
		self.text.tag_config("result",  foreground=RESULT_COLOR)
		self.text.tag_config("error",   foreground=ERROR_COLOR)
		if last:
			self.text.insert(tk.END,"\n")
		else:
			self.text.mark_set(tk.INSERT, f"{row}.end")
		self.text.see(tk.INSERT)
		return "break"

	# ----------------------------------------------------------------------
	def shiftReturn(self, event=None):
		self.text.insert(tk.INSERT,"\n")
		return "break"

	# ----------------------------------------------------------------------
	def paste(self, event=None):
		self._insert = self.text.index(tk.INSERT)
		self.event_generate("<<Evaluate>>")

	# ----------------------------------------------------------------------
	def findUniqueAns(self):
		for i in count(1):
			ans=f"ans{i}"
			if ans not in self.localDict: return ans

	# ----------------------------------------------------------------------
	# Set text and highlight answers and errors
	# ----------------------------------------------------------------------
	def set(self, txt):
		if txt.strip() != self.text.get("1.0", tk.END).strip():
			self.localDict.clear()
			self.text.delete("1.0", tk.END)
			self.text.insert("1.0", txt)
			self.text.mark_set(tk.INSERT,"1.0")
			self.evaluate()

	# ----------------------------------------------------------------------
	def get(self):
		return self.text.get("1.0",tk.END).strip()

	# ----------------------------------------------------------------------
	def clear(self):
		self.text.delete("1.0",tk.END)

	# ----------------------------------------------------------------------
	def selectAll(self, event=None):
		self.text.tag_add(tk.SEL, "1.0", tk.END)
		return "break"

#===============================================================================
# Button with Label that popups a menu
#===============================================================================
class MenuButton(Ribbon.MenuButton):
	#----------------------------------------------------------------------
	def __init__(self, master, insert, menulist, **kw):
		self.insert = insert
		super().__init__(master, menulist, **kw)

	#----------------------------------------------------------------------
	def createMenuFromList(self, menulist):
		mainmenu = menu = tk.Menu(self, tearoff=0, activebackground=Ribbon._ACTIVE_COLOR)
		for item in menulist:
			if item is None:
				menu.add_separator()
			elif isinstance(item,str):
				menu = tk.Menu(mainmenu)
				mainmenu.add_cascade(label=item, menu=menu)
			else:
				name, acc = item
				cmd = lambda t=name, ins=self.insert: ins(t)
				menu.add_command(label=name,
						accelerator=acc,
						command=cmd)
		return mainmenu

#===============================================================================
# Calculator frame class
#===============================================================================
class CalculatorPage(FlairRibbon.FlairPage):
	"""Scientific calculator for FLUKA"""

	_name_ = "Calculator"
	_icon_ = "calculator"

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Plot List ===========
		group = Ribbon.LabelGroup(self.ribbon, "Functions")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		menulist = [
			("a",	"Fine constant"),
			("amu",	"Atomic Mass Unit"),
			("amuC12","Atomic Mass Unit C12"),
			("amugr","Atomic Mass Unit in gr"),
			("c",	"Speed of light"),
			("fwhm","Full Width at Half Max"),
			("Na",	"Avogadro number"),
			("pi",	"3.14159..."),
			("qe",	"Electron charge"),
			("re",	"Electron radius") ]

		col,row = 1,0
		b = MenuButton(group.frame, self.insert, menulist,
				text="Constants \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("constant"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Constants")

		# ---
		menulist = [
			"Angle",
			(	"deg"	, "pi/180"),
			(	"rad"	, "1"),
			(	"mrad"	, "0.001"),

			"Distance",
			(	"nm",	"0.1E-6"),
			(	"um",	"0.1E-3"),
			(	"mm",	"0.1"),
			(	"cm",	"1"),
			(	"m",	"100"),
			(	"km",	"100E3"),
			None,
			(	"inch"	, "2.54"),
			(	"feet"	, "30.48"),
			(	"ft"	, "30.48"),
			(	"mile"	, "160934.4"),
			(	"mi"	, "160934.4"),

			"Energy",
			(	"eV"	, "1E-9"),
			(	"keV"	, "1E-6"),
			(	"MeV"	, "1e-3"),
			(	"GeV"	, "1"),
			(	"TeV"	, "1000"),
			(	"PeV"	, "1E6"),
			(	"J"     , "eV/qe"),

			"Time",
			(	"ns"	, "1.0E-9"),
			(	"us"	, "1.0E-6"),
			(	"ms"	, "0.001"),
			(	"s"	, "1"),
			(	"min"	, "60"),
			(	"hour"	, "3600"),
			(	"day"	, "24*hour"),
			(	"week"	, "7*day"),
			(	"month"	, "365.25/12*day"),
			(	"year"	, "365.25*day") ]

		col,row = 1,1
		b = MenuButton(group.frame, self.insert, menulist,
				text="Units \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("units"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Conversion Units")

		# ---
		menulist = [
			"Vector",
			("{x,y,z}",	"vector"),
			(".x",		"x component"),
			(".y",		"y component"),
			(".z",		"z component"),
			("*",		"dot product"),
			(".dot(v)",	"dot product"),
			("^",		"cross product"),
			(".cross(v)",	"cross product"),
			(".length()",	"length"),
			("abs(v)",	"length of vector"),
			(".unit()",	"unit vector"),
			(".phi()",	"azimuthal angle"),
			(".theta()",	"polar angle"),
			#"Matrix",
			"List",
			("[x,y,z]",	"list or array")]

		col,row = 1,2
		b = MenuButton(group.frame, self.insert, menulist,
				text="Vector \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("matrix"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Vector functions")

		# ---
		col,row = 2,0
		menulist = [
			("Mp",	"Proton mass"),
			("Mn",	"Neutron mass"),
			("Me",	"Electron mass") ]

		b = MenuButton(group.frame, self.insert, menulist,
				text="Particles \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("particle"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Particle masses")

		# ---
		col,row = 2,1
		menulist = [
			"Physics",
			("T2p(E,m)"	, "Energy to Momentum"),
			("p2T(p,m)"	, "Momentum to Energy"),
			("dT2dp(E,dE,m)", "Energy spread in momentum"),
			("T2g(E,m)"	, "Energy to gamma"),
			("g2b(g)"	, "gamma to beta"),
			("X0(Z,A)"      , "Radiation length in g/cm2"),
			("lhad(A)"      , "Hadronic interaction length (g/cm2)"),

			"Logarithmic",
			("exp(x)",	"e**x"),
			("expm1(x)",	"e**x-1"),
			("log(x)",	"Natural logarithm of x (base e)"),
			("log(x,b)",	"logarithm of x to the given base"),
			("log1p(x)",	"Natural logarithm of 1+x (base e)"),
			("log10(x)",	"Base-10 logarithm of x"),
			("pow(x,y)",	"x raised to the power y"),
			("sqrt(x)",	"Return the square root of x"),

			"Trigonometric",
			("acos(x)",	"Arc cosine to radians"),
			("asin(x)",	"Arc sine to radians"),
			("atan(x)",	"Arc tangent to radians"),
			("atan2(y,x)",	"Atan(y/x) to radians"),
			("cos(x)",	"Cosine from radians"),
			("hypot(x,y)",	"Euclidean norm, sqrt(x*x + y*y)"),
			("sin(x)",	"Sine from radians"),
			("tan(x)",	"Tangent from radians"),
			("omega(x)"	, "Solid angle from radians"),
			None,
			("acosd(x)",	"acos to degrees"),
			("asind(x)",	"asin to degrees"),
			("atand(x)",	"atan to degrees"),
			("cosd(x)",	"cos from degrees"),
			("sind(x)",	"sin from degrees"),
			("tand(x)",	"tan from degrees"),
			("omegad(x)",	"Solid angle from degrees"),

			"Hyperbolic",
			("acosh(x)",	"Inverse hyperbolic cosine"),
			("asinh(x)",	"Inverse hyperbolic sine"),
			("atanh(x)",	"Inverse hyperbolic tangent"),
			("cosh(x)",	"Hyperbolic cosine"),
			("sinh(x)",	"Hyperbolic sine"),
			("tanh(x)",	"Hyperbolic tangent"),

			"Special",
			("erf(x)",	"Error function"),
			("erfc(x)",	"Complementary error function"),
			("gamma(x)",	"Gamma function"),
			("lgamma(x)",	"Natural logarithm of the absolute Gamma function"),
			]
		b = MenuButton(group.frame, self.insert, menulist,
				text="Functions \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("function"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Physics and mathematical functions")

		# ---
		col,row = 2,2
		menulist = [
			("w(n)",	"n'th what of current card"),
			("what(n)",	"n'th what of current card"),
			("b(n,w)",	"w'th what of body named n"),
			("bbody(n,w)",	"w'th what of body named n"),
			("C(t,n,w)",	"n'th what of card tag t with name n"),
			("card(t,n,w)",	"n'th what of card tag t with name n")
			]
		b = MenuButton(group.frame, self.insert, menulist,
				text="Card \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("input"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Card cross reference")

		# ---
		# ========== Plot List ===========
		group = Ribbon.LabelGroup(self.ribbon, "Action")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		group.frame.grid_rowconfigure(0, weight=1)
		col,row = 0,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("clean32"),
				text="Clear",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.clear)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Clear calculator")

	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage()

		# --- Frame with listbox ---
		self.calculator = Calculator(self.frame)
		self.calculator.text.config(height=5)
		self.calculator.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

	# ----------------------------------------------------------------------
	# Refresh calculator
	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		super().refresh(what)
		self.calculator.set(self.project.calculator)

	# ----------------------------------------------------------------------
	# Update fields
	# ----------------------------------------------------------------------
	def get(self):
		if self.page is None or self.project is None: return

		txt= self.calculator.get()
		if txt != self.project.calculator:
			self.project.calculator = txt
			self.modified()

	#----------------------------------------------------------------------
	def clear(self):
		self.calculator.clear()

	# ----------------------------------------------------------------------
	def insert(self, txt):
		self.calculator.text.insert(tk.INSERT, txt)

#-------------------------------------------------------------------------------
if __name__ == "__main__":
	sys.path.insert(0,os.path.join(os.path.abspath(os.path.dirname(__file__)), "lib"))

	root = tk.Tk(className="Flair_calculator")
	tkFlair.openIni()
	root.title("Calculator")
	icons.loadIcon(os.path.join(tkFlair.prgDir,"icons","calculator.gif"))
	try:
		root.tk.call('wm','iconphoto',root._w, icons.get("calculator"))
	except tk.TclError:
		root.iconbitmap(f"@{tkFlair.prgDir}/icons/calculator.xbm")
	tkExtra.changeBehaviourTextClasses(root)
	calc = Calculator(root)
	calc.pack(expand=tk.YES, fill=tk.BOTH)

	def quit(event):  root.destroy()
	def cut(event):   root.event_generate("<<Cut>>")
	def copy(event):  root.event_generate("<<Copy>>")
	def paste(event): root.event_generate("<<Paste>>")
	def undo(event):  root.event_generate("<<Undo>>")
	def redo(event):  root.event_generate("<<Redo>>")

	root.bind("<Escape>", quit)
	root.bind("<Control-Key-x>", cut)
	root.bind("<Control-Key-c>", copy)
	root.bind("<Control-Key-v>", paste)
	root.bind("<Control-Key-q>", quit)
	root.bind("<Control-Key-z>", undo)
	root.bind("<Control-Key-Z>", redo)
	root.bind("<Control-Key-y>", redo)
	calc.text.focus_set()
	root.mainloop()
	icons.delete()
