# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import math

import tkinter as tk
from tkinter import ttk
from tkinter import messagebox

try:
	from PIL import Image
	import ZoomImage
except ImportError:
	Image = None

import bmath
import icons
import Ribbon
import tkExtra

CURSOR = 3

MXi = 0
MYi = 1
MX  = 2		# Position in markers list
MY  = 3
MZ  = 4
L1  = 5
L2  = 6
TXT = 7

ACTION_NONE   = 0
ACTION_ADD    = 1
ACTION_EDIT   = 2
ACTION_DELETE = 3

_FORMATXYZ = "[%g, %g, %g]"

root = None

#===============================================================================
class Marker:
	idx   = 0
	color = "Magenta"
	font  = "Helvetica,14"
	width = 5.0

	#-----------------------------------------------------------------------
	def __init__(self, xi, yi, x, y, z):
		self.xi = float(xi)
		self.yi = float(yi)
		self.x  = x
		self.y  = y
		self.z  = z
		self.tag = f"M{Marker.idx}"
		Marker.idx += 1

	#-----------------------------------------------------------------------
	def create(self, canvas):
		self.canvas = canvas
		zoom = canvas.zoom
		xi = round(self.xi*zoom)
		yi = round(self.yi*zoom)
		self.line1 = canvas.create_line(
				xi - Marker.width,
				yi - Marker.width,
				xi + Marker.width + 1,
				yi + Marker.width + 1,
				fill=Marker.color,
				tags=self.tag)

		self.line2 = canvas.create_line(
				xi - Marker.width,
				yi + Marker.width,
				xi + Marker.width + 1,
				yi - Marker.width - 1,
				fill=Marker.color,
				tags=self.tag)

		#canvas.create_rectangle(( xi, yi, xi+2, yi+2), fill="#FFFF00")

		self.txt = canvas.create_text(
				xi, yi + Marker.width,
				text=str(self),
				fill=Marker.color,
				font=Marker.font,
				tags=self.tag,
				justify=tk.CENTER, anchor=tk.N)

	#-----------------------------------------------------------------------
	def zoom(self):
		zoom = self.canvas.zoom
		self.canvas.coords(self.line1,
				self.xi*zoom - Marker.width,
				self.yi*zoom - Marker.width,
				self.xi*zoom + Marker.width + 1,
				self.yi*zoom + Marker.width + 1)
		self.canvas.coords(self.line2,
				self.xi*zoom - Marker.width,
				self.yi*zoom + Marker.width + 1,
				self.xi*zoom + Marker.width + 1,
				self.yi*zoom - Marker.width)
		self.canvas.coords(self.txt,
				self.xi*zoom,
				self.yi*zoom + Marker.width)

	#-----------------------------------------------------------------------
	def delete(self):
		self.canvas.delete(self.tag)

	#-----------------------------------------------------------------------
	def __str__(self):
		return _FORMATXYZ%(self.x, self.y, self.z)

	#-----------------------------------------------------------------------
	def edit(self, master):
		self.toplevel = tk.Toplevel(master)
		self.toplevel.title("Coordinates")
		self.toplevel.transient(master)

		tk.Label(self.toplevel, text="X:").grid(row=0,column=0)
		tk.Label(self.toplevel, text="Y:").grid(row=1,column=0)
		tk.Label(self.toplevel, text="Z:").grid(row=2,column=0)

		self.xentry = tkExtra.FloatEntry(self.toplevel, background="White")
		self.yentry = tkExtra.FloatEntry(self.toplevel, background="White")
		self.zentry = tkExtra.FloatEntry(self.toplevel, background="White")

		self.xentry.grid(row=0, column=1, columnspan=3, sticky=tk.EW)
		self.yentry.grid(row=1, column=1, columnspan=3, sticky=tk.EW)
		self.zentry.grid(row=2, column=1, columnspan=3, sticky=tk.EW)

		b = tk.Button(self.toplevel, text="Ok",
				image=icons.get("ok"),
				compound=tk.LEFT,
				command=self.ok,
				padx=2, pady=2)
		b.grid(row=3, column=2)
		b = tk.Button(self.toplevel, text="Cancel",
				image=icons.get("x"),
				compound=tk.LEFT,
				command=self.close,
				padx=2, pady=2)
		b.grid(row=3, column=3)
		self.toplevel.grid_columnconfigure(1, weight=1)

		self.xentry.insert(0, str(self.x))
		self.yentry.insert(0, str(self.y))
		self.zentry.insert(0, str(self.z))
		self.xentry.focus_set()

		self.toplevel.bind("<Return>",   self.ok)
		self.toplevel.bind("<KP_Enter>", self.ok)
		self.toplevel.bind("<Escape>",   self.close)
		self.toplevel.protocol("WM_DELETE_WINDOW", self.close)
		tkExtra.placeWindow(self.toplevel, master)
		try:
			if not self.toplevel.winfo_ismapped():
				self.toplevel.wait_visibility()
			self.toplevel.grab_set()
		except tk.TclError:
			pass
		self.toplevel.wait_window()

	#-----------------------------------------------------------------------
	def ok(self, event=None):
		try: self.x = float(self.xentry.get())
		except: pass
		try: self.y = float(self.yentry.get())
		except: pass
		try: self.z = float(self.zentry.get())
		except: pass

		self.canvas.itemconfig(self.txt, text=str(self))
		self.close()

	#-----------------------------------------------------------------------
	def close(self, event=None):
		self.toplevel.destroy()
		del self.toplevel

#===============================================================================
# Image Calibrate class
#===============================================================================
class Calibrate(tk.Toplevel):
	width  = -1
	height = -1

	#-----------------------------------------------------------------------
	def __init__(self, master, *options):
		super().__init__(master, *options)
		self.title("Calibrate Image")

		# ---------- create layout -------------
		#self.splitter = tkExtra.HSplitter(self, 250, True)
		#self.splitter.pack(expand=tk.YES, fill=tk.BOTH)

		# --- working canvas ---
		frame = tk.Frame(self)
		frame.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

		self.canvas = ZoomImage.ZoomImage(frame,
				#width=640, height=480,
				takefocus=True)
		self.canvas.grid(row=0, column=0, sticky=tk.NSEW)

		sbv = ttk.Scrollbar(frame, orient=tk.VERTICAL)
		sbv.grid(row=0, column=1, sticky=tk.NSEW)
		sbv.config(command=self.canvas.yview)

		sbh = ttk.Scrollbar(frame, orient=tk.HORIZONTAL)
		sbh.grid(row=1, column=0, sticky=tk.NSEW)
		sbh.config(command=self.canvas.xview)

		self.canvas.config( yscrollcommand=sbv.set,
				    xscrollcommand=sbh.set)
		frame.grid_columnconfigure(0, weight=1)
		frame.grid_rowconfigure(0, weight=1)

		# Override zooming
		self.canvas.bindMotion()
		self.canvas.bindArrows()
		self.canvas.bind('<Button-4>',  self.zoomIn)
		self.canvas.bind('<Button-5>',  self.zoomOut)
		self.canvas.bind('<Key-minus>', self.zoomOut)
		self.canvas.bind('<Key-equal>', self.zoomIn)
		self.canvas.bind('<Key-plus>',  self.zoomIn)
		self.canvas.bind('<Key-1>',     self.zoomOne)

		self.canvas.bind('<Button-1>',  self.click)
		self.canvas.bind('<Control-Button-1>', self.addMarkerAction)
		self.canvas.bind('<Double-1>',  self.editMarkerAction)
		self.canvas.bind('<Button-3>',  self.popup)
		self.canvas.bind('<Key-space>', self.popup)

		self.canvas.bind('<B1-Motion>', self.motion)
		self.canvas.bind('<ButtonRelease-1>', self.release)
		self.canvas.bind('<Enter>',	self.enter)
		self.canvas.bind('<Insert>',	self.addMarkerAction)
		self.canvas.focus_set()

		# --- canvas ---
		ribbon = tk.Frame(self)
		ribbon.place(x=0, y=0)

		# ===
		group = Ribbon.LabelGroup(ribbon, "Edit")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		b = Ribbon.LabelButton(group.frame,
				text="Add",
				image=icons.get("add32"),
				compound=tk.TOP,
				command=self.addMarker,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.pack(side=tk.LEFT, fill=tk.Y)

		b = Ribbon.LabelButton(group.frame,
				text="Edit",
				image=icons.get("edit32"),
				compound=tk.TOP,
				command=self.editMarker,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.pack(side=tk.LEFT, fill=tk.Y)

		b = Ribbon.LabelButton(group.frame,
				text="Delete",
				image=icons.get("x32"),
				compound=tk.TOP,
				command=self.delMarker,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.pack(side=tk.LEFT, fill=tk.Y)

		# ===
		group = Ribbon.LabelGroup(ribbon, "View")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid2rows()

		b = Ribbon.LabelButton(group.frame,
				text="Zoom In",
				image=icons.get("zoom_in"),
				compound=tk.LEFT,
				command=self.zoomIn,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.grid(row=0, column=0, stick=tk.W)

		b = Ribbon.LabelButton(group.frame,
				text="Zoom Out",
				image=icons.get("zoom_out"),
				compound=tk.LEFT,
				command=self.zoomOut,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.grid(row=1, column=0, stick=tk.W)

		# ---
		b = Ribbon.LabelButton(group.frame,
				text="Info",
				image=icons.get("info32"),
				compound=tk.TOP,
				command=self.showInfo,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.grid(row=0, rowspan=2, column=1, stick=tk.NSEW)

		# ===
		group = Ribbon.LabelGroup(ribbon, "Confirm")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		b = Ribbon.LabelButton(group.frame,
				text="Ok",
				image=icons.get("ok32"),
				compound=tk.TOP,
				command=self.ok,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.pack(side=tk.LEFT, fill=tk.Y)

		b = Ribbon.LabelButton(group.frame,
				text="Cancel",
				image=icons.get("close32"),
				compound=tk.TOP,
				command=self.close,
				padx=3,
				pady=0,
				background=Ribbon._BACKGROUND)
		b.pack(side=tk.LEFT, fill=tk.Y)

		# Global variables
		self.imagefile = ""
		self.filename  = None

		#self.info = tk.Frame(self.splitter.right())
		#self.info.pack(fill=tk.BOTH)

		self.action    = ACTION_NONE
		self.w         = 5
		self.marker    = []
		self.point     = []
		self.closest   = None
		self.closest_x = 0.0
		self.closest_y = 0.0
		self.oldZoom   = 1.0
		self.rc        = False

		#self.protocol("WM_DELETE_WINDOW", self.close)
		#self.cursor = (self.canvas.create_line(-2, 0, 2, 0, fill="Pink"),
		#	       self.canvas.create_line( 0,-2, 0, 2, fill="Pink"))

		# ---
		#viewmenu = tk.Menu(self)
		#viewmenu.add_command(label="Zoom 1:1", underline=0,
		#			accelerator="1", command=self.zoomOne)
		#viewmenu.add_command(label="Zoom In", underline=0,
		#			accelerator="=", command=self.zoomIn)
		#viewmenu.add_command(label="Zoom Out", underline=0,
		#			accelerator="-", command=self.zoomOut)

	#-----------------------------------------------------------------------
	def show(self):
		self.transient(self.master)
		self.deiconify()
		if not self.winfo_ismapped():
			self.wait_visibility()
		if Calibrate.width > 0:
			self.geometry(f"{Calibrate.width}x{Calibrate.height}")
		else:
			self.geometry(f"{self.winfo_width()}x{self.winfo_height()}")
		self.wait_window()
		return self.rc

	#-----------------------------------------------------------------------
	def ok(self):
		self.rc = True
		self.close()

	#-----------------------------------------------------------------------
	def close(self):
		Calibrate.width  = self.winfo_width()
		Calibrate.height = self.winfo_height()
		self.destroy()
		if root: root.destroy()

	#-----------------------------------------------------------------------
	def showInfo(self):
		InfoDialog(self)

	#-----------------------------------------------------------------------
	#def printInfo(self):
	#	markers = self.getMarkers()
	#	self.matrix = compute(markers)
	#	i = 1
	#	say("##     x          y          z         xi       yi      err    dx     dy")
	#	say("-- ---------- ---------- ---------- -------- -------- ------ ------ ------")
	#	for xi,yi,x,y,z in markers:
	#		pixel = self.matrix * bmath.Vector(x,y,z)
	#		e = math.sqrt((pixel[0]-xi)**2 + (pixel[1]-yi)**2)
	#		say("%2d %10g %10g %10g %8.4g %8.4g %6.3g %6.3g %6.3g" % \)
	#			(i, x,y,z, xi,yi, e,(pixel[0]-xi),(pixel[1]-yi))
	#		i += 1
	#	say("matrix=\n",self.matrix)
	#	say("Error=",error(markers, self.matrix))

	#-----------------------------------------------------------------------
	def loadImage(self, fn):
		self.config(cursor="watch")
		try:
			image = Image.open(fn)
			image.load()
		except:
			self.config(cursor="")
			messagebox.showerror("Error opening file",
				f"Error opening image file {fn}",
				parent=self)
			return
		self.imagefile = fn
		self.image     = image
		self.canvas.setImage(image)
		self.config(cursor="")

	# ----------------------------------------------------------------------
	def setImage(self, img):
		self.canvas.setImage(img)

	#-----------------------------------------------------------------------
	def addMarker(self):
		self.action = ACTION_ADD
		self.config(cursor="tcross")

	#-----------------------------------------------------------------------
	def editMarker(self):
		self.action = ACTION_EDIT
		self.config(cursor="question_arrow")

	#-----------------------------------------------------------------------
	def delMarker(self):
		self.action = ACTION_DELETE
		self.config(cursor="pirate")

	# ----------------------------------------------------------------------
	def setMarkers(self, markers):
		self.deleteAll()
		for xi,yi,x,y,z in markers:
			mark = Marker(xi,yi,x,y,z)
			self.marker.append(mark)
			mark.create(self.canvas)

	# ----------------------------------------------------------------------
	def getMarkers(self):
		markers = []
		for mark in self.marker:
			markers.append([mark.xi, mark.yi, mark.x, mark.y, mark.z])
		return markers

	# ----------------------------------------------------------------------
	def zoomMarkers(self):
		for mark in self.marker:
			mark.zoom()

	# ----------------------------------------------------------------------
	# find closest marker
	# ----------------------------------------------------------------------
	def find(self, x, y):
		#items = self.canvas.find_closest(x, y,
		#		self.w*self.canvas.zoom,
		#		self.canvas.image)
		items = self.canvas.find_overlapping(x-1, y-1, x+2, y+2)

		for item in items:
			if item == self.canvas.image: continue

			for mark in self.marker:
				if item in (mark.line1, mark.line2,  mark.txt):
					tag = self.canvas.gettags(mark.line1)[0]
					return tag

		return None

	# ----------------------------------------------------------------------
	def _click(self, event):
		self.cx = self.canvas.canvasx(event.x)
		self.cy = self.canvas.canvasy(event.y)
		self.x = float(self.cx)/self.canvas.zoom
		self.y = float(self.cy)/self.canvas.zoom

		self.closest   = self.find(self.cx, self.cy)
		self.start_x   = event.x
		self.start_y   = event.y
		self.closest_x = event.x
		self.closest_y = event.y

		pos = f"{self.x:6.1f} {self.y:6.1f}"
		self.clipboard_clear()
		self.clipboard_append(pos)

	# ----------------------------------------------------------------------
	def addMarkerAction(self, event):
		self.action = ACTION_NONE
		self._click(event)
		if self.marker:
			mark = self.marker[-1]
			x = mark.x
			y = mark.y
			z = mark.z
		else:
			x = y = z = 0.0
		mark = Marker(self.x,self.y,x,y,z)
		self.marker.append(mark)
		mark.create(self.canvas)
		mark.edit(self.canvas)

	# ----------------------------------------------------------------------
	def delMarkerAction(self, event):
		self.action = ACTION_NONE
		self._click(event)
		self.deleteMarker()

	# ----------------------------------------------------------------------
	def click(self, event):
		if self.canvas.tkimage is None: return
		self._click(event)

		if self.action == ACTION_ADD:
			self.addMarkerAction(event)
		elif self.action == ACTION_EDIT:
			self.editMarkerAction(event)
		elif self.action == ACTION_DELETE:
			self.delMarkerAction(event)
		#if self.closest is None:
		self.config(cursor="")

	# ----------------------------------------------------------------------
	def motion(self, event):
		if self.canvas.tkimage is None: return
		if self.closest is None: return
		if event.x == self.closest_x and event.y == self.closest_y:
			return

		# Moving a marker or an object
		self.canvas.move(self.closest,
			event.x-self.closest_x,
			event.y-self.closest_y)

		self.closest_x = event.x
		self.closest_y = event.y

	# ----------------------------------------------------------------------
	def release(self, event):
		if self.canvas.tkimage is None: return
		if self.closest is None: return

		# Modify position of marker
		#x = float(self.canvas.canvasx(event.x))/self.canvas.zoom
		#y = float(self.canvas.canvasy(event.y))/self.canvas.zoom

		for mark in self.marker:
			if mark.tag == self.closest:
				mark.xi += (event.x-self.start_x)/self.canvas.zoom
				mark.yi += (event.y-self.start_y)/self.canvas.zoom
				return

	# ----------------------------------------------------------------------
	#def setCursor(self, x, y):
	#	self.canvas.coords(self.cursor[0], x-CURSOR, y, x+CURSOR+1, y)
	#	self.canvas.coords(self.cursor[1], x, y-CURSOR, x, y+CURSOR+1)

	# ----------------------------------------------------------------------
	def editMarkerAction(self, event=None):
		if self.closest is None: return
		# Edit marker
		for mark in self.marker:
			if mark.tag == self.closest:
				mark.edit(self.canvas)
				return

	# ----------------------------------------------------------------------
	def deleteMarker(self, event=None):
		if self.closest is None: return

		# Delete marker
		for mark in self.marker:
			if mark.tag == self.closest:
				mark.delete()
				self.marker.remove(mark)
				return

	# ----------------------------------------------------------------------
	def deleteAll(self):
		for mark in self.marker:
			mark.delete()
		del self.marker[:]

	# ----------------------------------------------------------------------
	def popup(self, event=None):
		if self.canvas.tkimage is None: return
		self._click(event)
		#self.setCursor(self.cx, self.cy)

		menu = tk.Menu(self, tearoff=0)
		menu.add_command(label="Add",    command=self.addMarker)
		menu.add_command(label="Edit",   command=self.editMarker)
		menu.add_command(label="Delete", command=self.delMarker)
		#menu.add_separator()
		menu.tk_popup(event.x_root+1, event.y_root+1)

	# ----------------------------------------------------------------------
	def zoomIn(self, event=None):
		self.oldZoom = self.canvas.zoom
		self.canvas.zoomIn(event)
		self.zoomMarkers()

	# ----------------------------------------------------------------------
	def zoomOut(self, event=None):
		self.oldZoom = self.canvas.zoom
		self.canvas.zoomOut(event)
		self.zoomMarkers()

	# ----------------------------------------------------------------------
	def zoomOne(self, event=None):
		self.oldZoom = self.canvas.zoom
		self.canvas.zoomImage(1)
		self.zoomMarkers()

	# ----------------------------------------------------------------------
	def enter(self, event):
		self.canvas.focus_set()

#===============================================================================
# InfoDialog
#===============================================================================
class InfoDialog(tk.Toplevel):
	def __init__(self, master, **kw):
		super().__init__(master, class_="InfoDialog", **kw)
		self.transient(master)
		self.title("Calibration Information")

		# calibration data
		self.markers = master.getMarkers()

		self.listbox = tkExtra.MultiListbox(self,
					(("#",   2, None),
					 ("xi",  7, None),
					 ("yi",  7, None),
					 ("x",  10, None),
					 ("y",  10, None),
					 ("z",  10, None),
					 ("Δx",  7, None),
					 ("Δy",  7, None),
					 ("err", 7, None)),
					 background='White')
		self.listbox.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
		self.listbox.bindList("<ButtonRelease-1>", self.edit)
		self.listbox.bindList("<Return>",          self.edit)
		self.listbox.bindList("<KP_Enter>",        self.edit)
		self.listbox.bindLeftRight()
		self.listbox.setPopupMenu([
				("Edit",   0, self.edit,   icons.get("edit")),
				("Add",	   0, self.add,    icons.get("add")),
				("Delete", 0, self.delete, icons.get("x")),
				("Clone",  0, self.clone,  icons.get("clone"))])

		# Total error
		frame = tk.Frame(self)
		frame.pack(fill=tk.X)
		tk.Button(frame, image=icons.get("add"), command=self.add).pack(side=tk.LEFT)
		tk.Button(frame, image=icons.get("del"), command=self.delete).pack(side=tk.LEFT)
		tk.Button(frame, image=icons.get("clone"), command=self.clone).pack(side=tk.LEFT)

		self.pixelerror = tk.Label(frame, width=10, background="White")
		self.pixelerror.pack(side=tk.RIGHT)
		tk.Label(frame, text="Error Pixel:").pack(side=tk.RIGHT)

		# M matrix
		frame = tk.LabelFrame(self, text="Transformation: Real to Image plane")
		frame.pack(fill=tk.X)
		self.mLabel = []
		for j in range(3):
			tk.Label(frame, text=f"#{j+1}").grid(
				row=j+1, column=0, sticky=tk.EW)
			tk.Label(frame, text=f"#{j+1}").grid(
				row=0, column=j+1, sticky=tk.EW)
			frame.grid_columnconfigure(j+1, weight=1)
			row = []
			for i in range(3):
				l = tk.Label(frame, width=10, background="White", anchor=tk.E)
				l.grid(row=j+1, column=i+1, padx=2, pady=2, sticky=tk.EW)
				row.append(l)
			self.mLabel.append(row)

		# R matrix
		frame = tk.LabelFrame(self, text="Image Plane to Pixel")
		frame.pack(fill=tk.X)
		self.rLabel = []
		for j in range(2):
			tk.Label(frame, text=f"#{j+1}").grid(row=j+1, column=0, sticky=tk.EW)
			tk.Label(frame, text=f"#{j+1}").grid(row=0, column=j+1, sticky=tk.EW)
			frame.grid_columnconfigure(j+1, weight=1)
			row = []
			for i in range(3):
				l = tk.Label(frame, width=10, background="White", anchor=tk.E)
				l.grid(row=j+1, column=i+1, padx=2, pady=2, sticky=tk.EW)
				row.append(l)
			self.rLabel.append(row)
		tk.Label(frame, text="#3").grid(row=0, column=3, sticky=tk.EW)
		frame.grid_columnconfigure(3, weight=1)

		# Buttons
		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, fill=tk.X)
		tk.Button(frame, text="Close", command=self.quit).pack(side=tk.RIGHT)

		self.fill()

		# Window
		self.protocol("WM_DELETE_WINDOW", self.quit)
		tkExtra.placeWindow(self, master)
		try:
			if not self.winfo_ismapped():
				self.wait_visibility()
			self.grab_set()
		except tk.TclError:
			pass
		self.wait_window()

	#-------------------------------------------------------------------------------
	def fill(self):
		# Matrix
		R,M = compute(self.markers)
		for j in range(3):
			for i in range(3):
				value = "" if M is None else f"{M[j][i]:g}"
				self.mLabel[j][i].config(text=value)
		for j in range(2):
			for i in range(3):
				value = "" if R is None else f"{R[j][i]:g}"
				self.rLabel[j][i].config(text=value)

		# Error
		if M is not None:
			value = f"{error(self.markers,R,M):g}"
			self.pixelerror.config(text=value)
		else:
			self.pixelerror.config(text="")

		# Markers
		sel = self.listbox.curselection()
		act = self.listbox.index(tk.ACTIVE)
		self.listbox.delete(0,tk.END)
		for i,(xi,yi,x,y,z) in enumerate(self.markers):
			if R is not None:
				uvw = M * bmath.Vector(x,y,z)
				uvw[2] = 1.0
				pixel = R * uvw
				dx = (pixel[0]-xi)
				dy = (pixel[1]-yi)
				e  = math.sqrt(dx**2 + dx**2)
				if abs(dx<1e-04): dx = 0.0
				if abs(dy<1e-04): dy = 0.0
				if abs(e <1e-04): e  = 0.0
				dx = "%-.4g"%(dx)
				dy = "%-.4g"%(dy)
				e  = "%-.4g"%(e)
			else:
				dx = ""
				dy = ""
				e  = ""
			try:   x  = f"{x:g}"
			except: pass
			try: y  = f"{y:g}"
			except: pass
			try: z  = f"{z:g}"
			except: pass
			try: xi = f"{xi:g}"
			except: pass
			try: yi = f"{yi:g}"
			except: pass
			self.listbox.insert(tk.END, (i+1, xi,yi, x,y,z, dx, dy, e))

		for s in sel: self.listbox.selection_set(s)
		self.listbox.activate(act)

	#-------------------------------------------------------------------------------
	def edit(self, event=None):
		listbox = self.listbox.focus_get()
		nl = self.listbox.listboxIndex(listbox)
		if nl<1 or nl>5: return
		sel = listbox.curselection()
		if len(sel) != 1: return
		self.listbox.activate(sel[0])
		edit = tkExtra.InPlaceEdit(listbox)
		if edit:
			try: value = float(edit.value)
			except: return
			self.markers[int(sel[0])][nl-1] = value
			self.fill()
			self.master.setMarkers(self.markers)

	# ----------------------------------------------------------------------
	def add(self, event=None):
		self.markers.append([0,0,0,0,0])
		self.listbox.insert(tk.END, (len(self.markers),
						"0", "0", "0", "0", "0", "", "", ""))
		self.listbox.selection_clear(0,tk.END)
		self.listbox.activate(tk.END)
		self.listbox.selection_set(tk.END)

	# ----------------------------------------------------------------------
	def delete(self, event=None):
		lst = list(map(int,self.listbox.curselection()))
		lst.reverse()
		for i in lst:
			self.listbox.delete(i)
			del self.markers[i]
		self.fill()
		self.master.setMarkers(self.markers)

	# ----------------------------------------------------------------------
	def clone(self, event=None):
		sel = list(map(int, self.listbox.curselection()))
		self.listbox.selection_clear(0,tk.END)
		for i in sel:
			row = self.listbox.get(i)
			row[0] = len(self.markers)
			self.markers.append(self.markers[i][:])
			self.listbox.insert(tk.END, row)
			self.listbox.selection_set(tk.END)
		self.fill()
		self.master.setMarkers(self.markers)

	# ----------------------------------------------------------------------
	def quit(self):
		self.destroy()

#-------------------------------------------------------------------------------
# Calculate transformation matrix from markers
#-------------------------------------------------------------------------------
def compute(markers):
	if markers is None or len(markers)<3: return None,None

	# Find perpendicular (plot or projection) plane
	abcd = bmath.fitPlane([x[2:] for x in markers])
	if abcd:
		w = bmath.Vector(abcd[0], abcd[1], abcd[2])
		w.normalize()
	else:
		return None,None

	# Create a 3D system u,v,w   with w=plane
	u = w.orthogonal()
	u.normalize()
	v = w ^ u

	M = bmath.Matrix(3)
	M.make(u,v,w)

	# Transformation from real(x,y,z) to plane(u,v,w) coordinates
	# [u,v,w] = M * [x,y,z]
	A = []
	B = []
	C = []
	# Find transformation from plane(u,v) coord to image(i,j) coordinates
	# ignoring u
	# Solve first for Xi,Yi = k*u + l*v + m
	for i,j,x,y,z in markers:
		uvw = M*bmath.Vector(x,y,z)
		uvw[2]=1.0	# ignore w set it to 1.0
		A.append(uvw)
		B.append([i])
		C.append([j])
	A = bmath.Matrix(A)
	B = bmath.Matrix(B)
	C = bmath.Matrix(C)
	try:
		RTi = bmath.solveOverDetermined(A, B)
	except:
		return None,None
	try:
		RTj = bmath.solveOverDetermined(A, C)
	except:
		return None,None

	# plane to image. Normally should be 2x3
	R = bmath.Matrix(3,type=1)
	for i in range(3):
		R[0][i] = RTi[i]
		R[1][i] = RTj[i]
	return R,M

#-------------------------------------------------------------------------------
# Find pixel error
#-------------------------------------------------------------------------------
def error(markers, R,M):
	err = 0.0
	for i,j,x,y,z in markers:
		uvw = M * bmath.Vector(x,y,z)
		uvw[2] = 1.0
		ijk = R * uvw
		err += (ijk[0]-i)**2 + (ijk[1]-j)**2
	return math.sqrt(err/len(markers))

#-------------------------------------------------------------------------------
# Find real error
#-------------------------------------------------------------------------------
def realerror(markers, matrix):
	err = 0.0
	invMatrix = matrix.clone()
	invMatrix.inv()
	for xi,yi,x,y,z in markers:
		r = invMatrix * bmath.Vector(xi,yi,0.0)
		err += (r[0]-x)**2 + (r[1]-y)**2 + (r[2]-z)**2
	return math.sqrt(err/len(markers))
