# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   06-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import time
import shutil
import tkinter as tk

import tool
import icons
import Ribbon
import Global
import tkFlair
import tkExtra
import Project
import bFileDialog
import FlairRibbon
import CompileProcess

from log import say
from fluka import FlukaConfig

#===============================================================================
# Database scan
#===============================================================================
class _DatabaseMenuButton(FlairRibbon.FlairMenu):
	#----------------------------------------------------------------------
	def createMenu(self):
		user = list(UserRoutinesDialog.scanInput(self.page.project.input))
		user.sort()

		if not user:
			self.page.userRoutines()
			return None

		menu = tk.Menu(self, tearoff=0)
		for u in user:
			menu.add_command(label=u,
				compound=tk.LEFT, image=icons.get("new"),
				command=lambda f=f"{u}.f",p=self.page : p.addUsermvaxFile(f))
		return menu

#===============================================================================
# Compile Page
#===============================================================================
class CompilePage(FlairRibbon.FlairPage):
	"""Compile a FLUKA executable from user routines."""

	_name_ = "Compile"
	_icon_ = "compile"

	#----------------------------------------------------------------------
	def init(self):
		self.refreshOn = ("exe",)

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Executable ===========
		group = FlairRibbon.FlairMenuGroup(self.ribbon, self, "Executable",
			[("Export Makefile","compile", self.flair.exportMakefile)])
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("save32"),
				text="Save As",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.saveAs)
		b.pack(side=tk.LEFT, expand=tk.YES, fill=tk.Y)
		tkExtra.Balloon.set(b, "Save executable as")

		# ========== Files ===========
		group = Ribbon.LabelGroup(self.ribbon, "Files")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		# ===
		col,row = 0,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("add32"),
				anchor=tk.CENTER,
				command=self.add)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add fortran file or library")

		col,row = 0,2
		menulist = [	("Fortran", "new",   self.addFortran),
				("C/C++",   "new",   self.addC),
				("Object",  "new",   self.addObject),
				("Library", "new",   self.addLibrary) ]
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Add",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add fortran file or library")

		# ===
		col,row = 1,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("database32"),
				anchor=tk.CENTER,
				command=self.userRoutines)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Copy from FLUKA user routines to the current directory/project")

		col,row = 1,2
		b = _DatabaseMenuButton(group.frame, self,
				text="Database",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Copy from FLUKA user routines to the current directory/project")

		# ---
		col,row = 2,0
		b = Ribbon.LabelButton(group.frame,
				text="Remove",
				image=icons.get("x"),
				command=self.delete)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Remove file from the compile list")

		# ---
		col,row = 2,1
		b = Ribbon.LabelButton(group.frame,
				text="Move Up",
				image=icons.get("up"),
				command=self.moveUp)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Move up card")

		# ---
		col,row = 2,2
		b = Ribbon.LabelButton(group.frame,
				text="Move Down",
				image=icons.get("down"),
				command=self.moveDown)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Move down card")

		# ========== View ===========
		group = Ribbon.LabelGroup(self.ribbon, "View")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("viewer32"),
				text="Viewer",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.viewer)
		b.pack(side=tk.LEFT, expand=tk.YES, fill=tk.Y)
		tkExtra.Balloon.set(b, "Open in internal viewer")

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("edit32"),
				text="Editor",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.editor)
		b.pack(side=tk.LEFT, expand=tk.YES, fill=tk.Y)
		tkExtra.Balloon.set(b, "Open in external editor")

		# ========== Build ===========
		group = Ribbon.LabelGroup(self.ribbon, "Action")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		group.grid2rows()

		# ===
		col,row = 0,0
		b = tk.Label(group.frame,
					text="Link",
					image=icons.get("link"),
					font=Ribbon._FONT,
					anchor=tk.E,
					background=Ribbon._BACKGROUND)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.SW)
		tkExtra.Balloon.set(b, "Linking program")

		# ===
		col,row = 0,1
		self.link = tkExtra.Combobox(group.frame,
					width=10,
					font=Ribbon._FONT,
					background=Ribbon._BACKGROUND)
		tk.Frame.config(self.link, background=Ribbon._BACKGROUND, padx=0, pady=0)
		self.link.grid(row=row, column=col, padx=0, pady=0, sticky=tk.N+tk.EW)
		tkExtra.Balloon.set(self.link, "Linking program")

		self.link.fill(Project.CompileInfo.LINK_PRG)
		self.link.set(Project.CompileInfo.LINK_PRG[0])

		# ---
		col,row = 1,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("clean32"),
				text="Clean",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.clean)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Remove all generated files")

		# ---
		col,row = 2,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("build32"),
				text="Build",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.execute)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Build executable file")

	#----------------------------------------------------------------------
	# Create Project page
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage()

		# -----
		frame = tk.Frame(self.frame)
		frame.pack(side=tk.TOP, fill=tk.X)

		tk.Label(frame, text="Executable:").pack(side=tk.LEFT)

		self.exe = tk.Label(frame,
				width=32,
				foreground="DarkBlue", background="White",
				relief=tk.SUNKEN, anchor=tk.W)
		tkExtra.Balloon.set(self.exe, "Executable filename")
		self.exe.pack(side=tk.LEFT)
		self.exe.bind('<Button-1>', self.saveAs)

		b = tk.Button(frame, image=icons.get("x"),
				command=self.defaultExecutable, pady=1, padx=1)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b, "Remove executable")

		tk.Label(frame, text=" Options:").pack(side=tk.LEFT)
		self.opt = tk.Entry(frame, background="White")
		self.opt.pack(side=tk.LEFT, fill=tk.X, expand=tk.TRUE)
		tkExtra.Balloon.set(self.opt, "Options to be passed to the compiler")

		# --- fileList
		self.fileList = tkExtra.ColorMultiListbox(self.frame,
				(('File', 10, None),
				 ('Type',  8, None),
				 ('Size',  6, None),
				 ('Date', 15, None)),
				 height=7)
		self.fileList.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=tk.YES)
		self.fileList.setPopupMenu([
				("Edit",   0, self.editor, icons.get("edit")),
				("View",   0, self.viewer, icons.get("view")),
				None,
				("Add",	   0, self.add,    icons.get("add")),
				("Remove", 0, self.delete, icons.get("x")),
				("Rename", 0, self.rename, icons.get("edit")) ])
		self.fileList.bindList('<Double-1>',	self.editor)
		self.fileList.bindList('<Return>',	self.editor)
		self.fileList.bindList('<KP_Enter>',	self.editor)
		self.fileList.bindList('<<Rename>>',	self.rename)
		self.fileList.bindList('<<MoveUp>>',	self.moveUp)
		self.fileList.bindList('<<MoveDown>>',	self.moveDown)
		self.fileList.focus_set()

		self.process = CompileProcess.CompileProcess()

	# ----------------------------------------------------------------------
	# Rename in place filename
	# ----------------------------------------------------------------------
	def rename(self, event = None):
		listbox  = self.fileList.listbox(0)
		selected = list(map(int,listbox.curselection()))
		if len(selected) != 1: return
		pos = selected[0]
		edit = tkExtra.InPlaceEdit(listbox)
		if edit.value and os.path.isfile(edit.value):
			# Replace file
			fn = self.project.relpath(edit.value)
			self.project.compile["source"][pos] = fn
			ext,color = bFileDialog.fileTypeColor(fn)
			s = os.stat(edit.value)
			self.fileList.set(pos,[fn, ext, s.st_size,
				time.strftime(bFileDialog._TIME_FORMAT, time.localtime(s.st_mtime))])
			if color is not None:
				self.fileList.itemconfig(pos,foreground=color)
			self.flair.notify("Success: File Changed",
				f"Success: file {edit.value} successfully added",
				tkFlair.NOTIFY_NORMAL)
			self.modified()
		else:
			self.refresh()
			self.flair.notify("Error: File not found",
					f"Error: file {edit.value} NOT found",
					tkFlair.NOTIFY_ERROR)
		self.fileList.selection_set(pos)
		self.fileList.activate(pos)

	# ----------------------------------------------------------------------
	# Show Frame
	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		super().refresh(what)
		if Global.developer:
			self.opt.delete(0, tk.END)
			self.opt.insert(0, self.project.compile("fopts"))

		self.link.set(self.project.compile("link"))
		self.exe["text"] = tool.abbrevPath(self.project.compile.name)

		self.fileList.delete(0,tk.END)
		for f in self.project.compile("source"):
			try:
				s = os.stat(f)
				ext,color = bFileDialog.fileTypeColor(f)
				self.fileList.insert(tk.END, (f, ext, s.st_size,
					time.strftime(bFileDialog._TIME_FORMAT,
						time.localtime(s.st_mtime))))
				if color is not None:
					self.fileList.itemconfig(tk.END,foreground=color)
			except OSError:
				self.fileList.insert(tk.END, (f, '-', '-', '-'))

	# --------------------------------------------------------------------
	def refreshStatus(self):
		if self.process.isAlive():
			self.setStatus("%2d%%  %s: %s" %
				(self.process.percent,
				 self.process.message[0],
				 self.process.message[1]))
			self.setStatusPercent(self.process.percent)
		else:
			self.setStatus(f"Files: {self.fileList.size()}")
			self.resetPercent()

	# ----------------------------------------------------------------------
	def addFortran(self):
		self.add(filetypes=[("Fortran files", ("*.f","*.for","*.F","*.FOR")),
				   ("All","*")])

	def addC(self):
		self.add(filetypes=[("C/C++ files", ("*.c", "*.cc", "*.cpp","*.C","*.CC","*.CPP")),
				   ("All","*")])

	def addObject(self):
		self.add(filetypes=[("Object Files","*.o"),
				   ("All","*")])

	def addLibrary(self):
		self.add(filetypes=[("Libraries",("*.a","*.so")),
				   ("All","*")])

	# ----------------------------------------------------------------------
	# Add a new file
	# ----------------------------------------------------------------------
	def add(self, event=None, filetypes=None):
		if filetypes is None:
			filetypes=[("Fortran files", ("*.f","*.for","*.F","*.FOR")),
				   ("C/C++ files", ("*.c", "*.cc", "*.cpp","*.C","*.CC","*.CPP")),
				   ("Object Files","*.o"),
				   ("Libraries",("*.a","*.so")),
				   ("All","*")]

		files = bFileDialog.askopenfilenames(master=self.page,
					initialdir=self.project.dir(),
					filetypes=filetypes)
		if files is None: return
		for f in files: self.addFile(f)

	# ----------------------------------------------------------------------
	def addFile(self, filename):
		fn = self.project.relpath(filename)
		ext,color = bFileDialog.fileTypeColor(fn)
		s = os.stat(filename)
		self.fileList.insert(tk.END,[fn, ext, s.st_size,
			time.strftime(bFileDialog._TIME_FORMAT, time.localtime(s.st_mtime))])
		if color is not None:
			self.fileList.itemconfig(tk.END,foreground=color)
		self.project.compile.appendSource(fn)
		self.modified()

	# ----------------------------------------------------------------------
	def addUsermvaxFile(self, filename):
		# Copy the file
		src = os.path.join(FlukaConfig.directory, "src/user", filename)
		dst = os.path.join(self.project.dir(), filename)
		try:
			os.stat(dst)
			if not tkFlair.askyesno(
					f"Override existing file '{filename}'",
					f"File '{filename}' already exists in project directory. "
					"Do you want to copy and override the existing file?",
					parent=self.page):
				return

			# FIXME XXX
			try:
				self.project.compile["source"].remove(filename)
			except KeyError:
				pass
			for i in range(self.fileList.size()):
				fn = self.fileList.listbox(0).get(i)
				if fn==filename:
					self.fileList.delete(i)
					break
		except OSError:
			pass

		try:
			shutil.copyfile(src, dst)
			# Add file to the compile list of files
			self.addFile(dst)
		except IOError:
			self.flair.notify(f"Error in copying file '{filename}'",
				f"Error during copying file '{filename}'",
				tkFlair.NOTIFY_ERROR)

	# ----------------------------------------------------------------------
	# Delete source
	# ----------------------------------------------------------------------
	def delete(self, event=None):
		# protect against accidental delete
		src = self.project.compile("source")
		for i in reversed(list(map(int, self.fileList.curselection()))):
			self.fileList.delete(i)
			del src[i]

	# ----------------------------------------------------------------------
	# Move up select items by one
	# ----------------------------------------------------------------------
	def moveUp(self, event=None):
		for i in map(int,self.fileList.curselection()):
			if i==0: continue
			prev = i-1
			if not self.fileList.selection_includes(prev):
				act = self.fileList.index(tk.ACTIVE)
				self.swap(prev,i)
				self.fileList.selection_set(prev)
				if act == i: self.fileList.activate(prev)
		return "break"

	# ----------------------------------------------------------------------
	# Move down select items by one
	# ----------------------------------------------------------------------
	def moveDown(self, event=None):
		sz  = self.fileList.size()-1
		for i in reversed(list(map(int, self.fileList.curselection()))):
			if i >= sz: continue
			inext = i+1
			if not self.fileList.selection_includes(inext):
				act = self.fileList.index(tk.ACTIVE)
				self.swap(i,inext)
				self.fileList.selection_set(inext)
				if act == i: self.fileList.activate(inext)
		return "break"

	# ----------------------------------------------------------------------
	# Swap two items in the list
	# ----------------------------------------------------------------------
	def swap(self, a, b):
		lb = self.fileList
		if a>b: a, b = b, a

		at = lb.get(a);		al = self.project.compile["source"][a]
		bt = lb.get(b);		bl = self.project.compile["source"][b]

		lb.delete(b)
		lb.delete(a)

		lb.insert(a, bt);	self.project.compile["source"][a] = bl
		lb.insert(b, at);	self.project.compile["source"][b] = al

	# ----------------------------------------------------------------------
	# Edit file
	# ----------------------------------------------------------------------
	def editor(self, event=None):
		tkFlair.editor(*[self.fileList.get(int(x))[0]
				for x in self.fileList.curselection() ])

	# ----------------------------------------------------------------------
	# View file
	# ----------------------------------------------------------------------
	def viewer(self, event=None):
		self.flair.viewer(*[self.fileList.get(int(x))[0]
				for x in self.fileList.curselection() ])

	# ----------------------------------------------------------------------
	def userRoutines(self, event=None):
		UserRoutinesDialog(self.page, self, self.project)

	# ----------------------------------------------------------------------
	# Save executable as
	# ----------------------------------------------------------------------
	def saveAs(self, event=None):
		self.get()
		self.activate()
		fn = bFileDialog.asksaveasfilename(master=self.page,
				title="Save executable as",
				initialdir=self.project.dir(),
				filetypes=[ ("All","*"), ("Exe","*.exe") ])
		if fn:
			self.project.compile.name = fn
			self.exe["text"] = self.project.compile.name
			self.project.runs.default()["exe"] = self.project.compile.name
			self.flair.refresh("exe")
			self.modified()

	# ----------------------------------------------------------------------
	# Default executable
	# ----------------------------------------------------------------------
	def defaultExecutable(self):
		self.project.compile.name = ""
		self.project.defaultRun["exe"] = ""
		self.exe["text"] = ""
		self.flair.refresh("exe")
		self.modified()

	# ----------------------------------------------------------------------
	# Select All
	# ----------------------------------------------------------------------
	def selectAll(self):
		self.fileList.selection_set(0,tk.END)

	# ----------------------------------------------------------------------
	# Update fields
	# ----------------------------------------------------------------------
	def get(self):
		if self.page is None or self.project is None: return

		o = self.opt.get()
		if o!=self.project.compile("fopts"):
			self.project.compile["fopts"] = o
			self.modified()

		l = self.link.get()
		if l!=self.project.compile("link"):
			self.project.compile["link"] = l
			self.modified()

	# ----------------------------------------------------------------------
	# build executable
	# ----------------------------------------------------------------------
	def execute(self):
		self.get()

		if not self.project.compile.name:
			self.saveAs()
			if not self.project.compile.name: return

		if self.process.isAlive():
			if not tkFlair.askyesno("Compile process active",
					"A compile process is still running.\n"
					"Do you want to kill it?",
					parent=self.page):
				return
			self.process.kill()
			self.process.wait()

		log = self.flair.newLog(self.name, self.project.compile.name)

		# Check if libfluka is newer than the executable
		# then rebuild everything
		lib = os.path.join(FlukaConfig.directory, FlukaConfig.DEFAULT_LIB)
		if tool.isnewer(lib, self.project.compile.name):
			say(f"Executable is older than {FlukaConfig.DEFAULT_LIB} rebuilding\n")
			self.process.clean(self.project)

		self.setStatus("Compiling executable")
		self.process.init()
		self.flair.addProcess(self.process, log, self.project)

	# ----------------------------------------------------------------------
	# clean files
	# ----------------------------------------------------------------------
	def clean(self):
		if self.process.isAlive():
			if tkFlair.askyesno("Compile process active",
					"A compile process is still running.\n"
					"Do you want to kill it?",
					parent=self.page):
				return
			self.process.kill()
			self.process.wait()

		log = self.flair.newLog(self.name, self.project.compile.name)
		self.setStatus("Cleaning files")
		self.process.clean(self.project)
		log.write(self.process.output())
		if self.process.message:
			self.flair.notify(self.process.message[0],
				self.process.message[1])

#===============================================================================
# User Routines Dialog
#===============================================================================
class UserRoutinesDialog(tk.Toplevel):
	def __init__(self, master, page, project):
		super().__init__(master)
		self.transient(master)
		self.title("FLUKA User routines")

		self.page    = page
		self.project = project
		self.path    = os.path.join(FlukaConfig.directory, "src/user")

		self.fileList = tkExtra.MultiListbox(self,
				(('File', 10, None),
				 ('Size',  6, None),
				 ('Date', 15, None),
				 ('Desc', 50, None)))
		self.fileList.config(height=20)
		self.fileList.bindList('<Double-1>', self.view)
		self.fileList.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, fill=tk.X)

		btn = tk.Button(frame, text="Copy to Project", underline=0,
				command=self.copy)
		btn.pack(side=tk.LEFT)

		btn = tk.Button(frame, text="Scan Input", underline=0,
				command=self.scan)
		btn.pack(side=tk.LEFT)

		btn = tk.Button(frame, text="View", underline=0,
				command=self.view)
		btn.pack(side=tk.LEFT)

		btn = tk.Button(frame, text="Close", underline=0,
			command=self.close)
		btn.pack(side=tk.LEFT)

		self.bind("<Escape>", self.close)
		self.protocol("WM_DELETE_WINDOW", self.close)

		self.fill()
		self.scan()

		tkExtra.placeWindow(self, master)
		self.grab_set()
		self.focus_set()
		self.fileList.focus_set()
		self.wait_window()

	# ----------------------------------------------------------------------
	def close(self, event=None):
		self.destroy()

	# ----------------------------------------------------------------------
	# Fill with files generated by fluka
	# ----------------------------------------------------------------------
	def fill(self):
		try:
			for filename in os.listdir(self.path):
				fn,ext = os.path.splitext(filename)
				if ext != ".f": continue
				try:
					s = os.lstat(os.path.join(self.path,filename))
					desc = FlukaConfig.user.get(fn, "")
					self.fileList.insert(tk.END,
						(filename, s.st_size,
						 time.strftime(bFileDialog._TIME_FORMAT,
							time.localtime(s.st_mtime)),
						 desc))
				except OSError:
					say("ERROR accessing file", filename)
		except OSError:
			say("ERROR accessing folder", self.path)
		self.fileList.sort(0)

	# ----------------------------------------------------------------------
	# View document
	# ----------------------------------------------------------------------
	def view(self, event=None):
		self.page.flair.viewer(*[
			os.path.join(self.path,self.fileList.get(int(x))[0])
				for x in self.fileList.curselection()])

	# ----------------------------------------------------------------------
	# Copy selected files to project
	# ----------------------------------------------------------------------
	def copy(self, event=None):
		for sel in self.fileList.curselection():
			filename, size, date, desc = self.fileList.get(sel)
			self.page.addUsermvaxFile(filename)

	# ----------------------------------------------------------------------
	# Scan input and highlight necessary files
	# ----------------------------------------------------------------------
	def scan(self):
		user = UserRoutinesDialog.scanInput(self.project.input)

		# Return a set of routines
		active = self.fileList.index(tk.ACTIVE)
		self.fileList.selection_clear(0,tk.END)
		activate = -1
		for i in range(self.fileList.size()):
			fn,_ = os.path.splitext(self.fileList.get(i)[0])
			if fn in user:
				self.fileList.selection_set(i)
				if active == i: activate = i
				if activate < 0:
					activate = i
		if activate>=0:
			self.fileList.activate(activate)
			self.fileList.see(activate)

	# ----------------------------------------------------------------------
	@staticmethod
	def scanInput(inp):
		user = set()

		# Check for abscff, dffcff, frghns
		for card in inp["OPT-PROP"]:
			sdum = card.sdum().strip()
			user.add("frghns")
			user.add("rflctv")
			if sdum=="":
				if card.intWhat(1)<-99: user.add("rfrndx")
				if card.intWhat(2)<-99: user.add("abscff")
				if card.intWhat(3)<-99: user.add("dffcff")
			elif sdum=="SPEC-BDX":
				user.add("ophbdx")
			elif sdum=="SENSITIV":
				if card.intWhat(1)<-99: user.add("queffc")

		if "OPT-PROD" in inp.cards:
			user.add("frghns")

		# Check for comscw, fluscw, endscp, fldscp, usrrnc
		for card in inp["USERWEIG"]:
			w3 = card.intWhat(3)
			if w3 > 0: user.add("fluscw")
			if w3 in (2,4): user.add("fldscp")
			if card.intWhat(5)>0: user.add("usrrnc")
			w6 = card.intWhat(6)
			if w6 > 0: user.add("comscw")
			if w6 in (2,4): user.add("endscp")

		# Check for formfu
		if "MULSOPT" in inp.cards:
			user.add("formfu")

		# FUSRBV, LUSRBL, MUSRBR
		for card in inp["USRBIN"]:
			if card.intWhat(1) in (8,18):
				user.add("fusrbv")
				user.add("lusrbl")
				user.add("musrbr")
				break

		# LATTIC
		for card in inp["LATTICE"]:
			if card.sdum().strip()=="":
				user.add("lattic")
				break

		# MAGFLD
		for card in inp["ASSIGNMA"]:
			if card.intWhat(5) == 1:
				user.add("magfld")
				break
		if "MGNFIELD" in inp.cards:
			user.add("magfld")

		# MGDRAW
		if "USERDUMP" in inp.cards:
			user.add("mgdraw")

		# Source
		if "SOURCE" in inp.cards:
			user.add("source")

		# Biasing
		cards = inp.cards.get("BIASING")
		if cards is not None:
			user.add("ubsset")
			for card in cards:
				if card.sdum()=="USER":
					user.add("usimbs")
					break
		for tag in ["EMF-BIAS", "EMFCUT", "EXPTRANS",
				"EXPTRANS", "LOW-BIAS", "LOW-DOWN", "WW-FACTO"]:
			if tag in inp.cards:
				user.add("ubsset")
				break

		# UDCDRL
		for card in inp["LAM-BIAS"]:
			if card.sdum() == "DCY-DIRE" and card.intWhat(4)==0:
				user.add("udcdrl")
				break
			if card.sdum() == "N1HSCBS":
				user.add("uesdrl")
				break

		# USRINI
		if "USRICALL" in inp.cards:
			user.add("usrini")

		# USROUT
		if "USROCALL" in inp.cards:
			user.add("usrout")

		# USRGLO
		if "USRGCALL" in inp.cards:
			user.add("usrglo")

		# USRMED
		for card in inp["MAT-PROP"]:
			if card.sdum()[:8] == "USERDIRE":
				if card.intWhat(1)>0:
					user.add("usrmed")

		return user
