#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

__author__  = "Vasilis Vlachoudis"
__email__   = "Vasilis.Vlachoudis@cern.ch"

import os
import time
import subprocess

import tool
import tkFlair
from process import Process
from bFileDialog import _TIME_FORMAT
from fluka import FlukaConfig

#===============================================================================
# Compile FLUKA project process
#===============================================================================
class CompileProcess(Process):
	#----------------------------------------------------------------------
	# Cleaning files in synchronous way
	# @return rc (and output filled)
	#----------------------------------------------------------------------
	def clean(self, project):
		super().clean()
		# Remove files
		rc = 0
		count = 0
		try:
			os.remove(project.compile.name)
			self.output(f"Removing: {project.compile.name}\n")
			count += 1
		except OSError as err:
			if err.errno!=2:		# Ignore No such file or directory
				self.output(f"{err}\n")
		except Exception as err:
			self.output(f"{err}\n")
			self.message = ("Compile clean", "Cannot delete executable")
			rc = 1

		try:
			os.remove(project.compile.name+".map")
			self.output(f"Removing: {project.compile.name}.map\n")
			count += 1
		except OSError as err:
			if err.errno!=2:		# Ignore No such file or directory
				self.output(f"{err}\n")
		except Exception as err:
			self.output(f"{err}\n")
			self.message = ("Compile clean", "Cannot delete map file")
			rc = 2

		for f in project.compile("source"):
			(fn, ext) = os.path.splitext(f)
			# Skip libraries
			if ext in (".a",".so",".o"): continue
			obj = fn + ".o"
			try:
				os.remove(obj)
				count += 1
				self.output(f"Removing: {obj}\n")
			except OSError as err:
				if err.errno!=2:	# Ignore No such file or directory
					self.output(f"{err}\n")
			except Exception as err:
				self.output(f"{err}\n")
				self.message = ("Compile clean", f"Cannot delete map {obj}")
				rc = 3
		if rc == 0:
			if count:
				self.message = ("Compile clean", f"{count} files deleted")
			else:
				self.message = ("Compile clean", "No files deleted")
		return rc

	#----------------------------------------------------------------------
	# Compile executable
	#----------------------------------------------------------------------
	@tkFlair.exception
	def execute(self, project):
		super().execute()
		# compile all sources that need compilation
		exe_time = tool.getmtime(project.compile.name)
		libfluka = os.path.join(FlukaConfig.directory, FlukaConfig.DEFAULT_LIB)
		lib_time = tool.getmtime(libfluka)

		maxobj_time = 0
		tot = len(project.compile("source")) + 1
		for i,f in enumerate(project.compile("source")):
			self.message = ("Compiling", f)
			if self.status == Process.STATUS_KILLED:
				self.output("*** Killed ***\n")
				return 0
			(fn, ext) = os.path.splitext(f)
			if ext in (".a",".so",".o"):
				# Check date of libraries
				try:	o_time = os.stat(f).st_mtime
				except:	o_time = 0
			else:
				# or object
				try:	o_time = os.stat(fn+".o").st_mtime
				except:	o_time = 0

				try:	f_time = os.stat(f).st_mtime
				except:
					self.message = ("File do not exist",
							f"Cannot find file {f}")
					return 1

				if o_time <= f_time or o_time <= lib_time:
					# Compile Fortran
					cmd = project.compile.compileCmd(f)
					self.output(f"\n*-* Compiling: {' '.join(list(cmd))}\n")
					try:
						p = subprocess.Popen(cmd,
								stdout=subprocess.PIPE,
								stderr=subprocess.STDOUT)
					except Exception as err:
						self.message = ("Error", str(err))
						return 2
					out,err = p.communicate()
					self.output(out)
					try:    o_time = os.stat(fn+".o").st_mtime
					except: o_time = 0
					if o_time < f_time:
						if o_time == 0:
							self.message = ("Error compiling",
									f"Error: compiling file: {f}\n"
									"No object generated")
						else:
							self.message = ("Error compiling",
									"Error: file %s time: %s\n"
									"is newer than the object %s"
									%(f,
									 time.strftime(_TIME_FORMAT,
										time.localtime(f_time)),
									 time.strftime(_TIME_FORMAT,
										time.localtime(o_time))))
						return 2

				if self.isKilled():
					self.message = ("Compilation Stopped",
							"User killed the compilation")
					return 3
			if o_time > maxobj_time:
				maxobj_time = o_time
			# Update percentage
			self.percent = (i*100) // tot

		# Link executable
		if exe_time <= maxobj_time:
			cmd = project.compile.linkCmd()
			self.output(f"\n*-* Linking: {' '.join(list(cmd))}\n")
			try:
				p = subprocess.Popen(cmd,
						stdout=subprocess.PIPE,
						stderr=subprocess.STDOUT)
			except Exception as err:
				self.message = ("Error", str(err))
				return 4
			out,err = p.communicate()
			self.output(out)
			try:
				exe_time = os.stat(project.compile.name).st_mtime
				if exe_time < maxobj_time:
					self.message = ("Error linking",
						"Error: executable %s time: %s\n"
						"is older than objects %s"
							%(project.compile.name,
							 time.strftime(_TIME_FORMAT,
								 time.localtime(exe_time)),
							 time.strftime(_TIME_FORMAT,
								time.localtime(maxobj_time))))
					return 4
			except:
				self.message = ("Error linking", "Error: executable is not generated")
				return 4
			try:
				exe_time = os.stat(project.compile.name).st_mtime
				self.message = ("Successful Build",
					f"Executable {project.compile.name} is successfully built")
				return 0
			except:
				self.message = ("Error linking",
					f"Error linking executable {project.compile.name}")
				return 5
		else:
			self.message = ("Up to date", "All files are up to date")
			return 0
