# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Dec-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import configparser
import tkinter as tk
from tkinter import ttk
import tkinter.font as tkFont
try:
	import multiprocessing
except ImportError:
	multiprocessing = None

import icons
import tkFlair
import tkExtra
import bFileDialog

import Manual
import Project

from fluka    import FlukaConfig
from mcnp     import McnpConfig
from moira    import MoiraConfig
from penelope import PenelopeConfig
from phits    import PhitsConfig

# synonyms of pages
_PAGES = {
	"Runs"  : "Run",
	"Input" : "Interface",
	"Plot"  : "Gnuplot"
}

_COLOR_LIST = [ "input.active",
		"input.comment",
		"input.define",
		"input.diff",
		"input.disable",
		"input.error",
		"input.hidden",
		"input.hiddenbg",
		"input.label",
		"input.mode",
		"input.normal",
		"input.name",
		"input.tag",
		"input.value",

		"geometry.background",
		"geometry.bodybboxincolor",
		"geometry.bodybboxoutcolor",
		"geometry.zonebboxcolor",
		"geometry.regionbboxcolor",
		"geometry.error",
		"geometry.gridtext",
		"geometry.label",
		"geometry.lattice",
		"geometry.lock",
		"geometry.palettetext",
		"geometry.region",
		"geometry.select",
		"geometry.title",
		"geometry.vertex",
		"geometry.visible",
		"geometry.wireframe",
		"geometry.zone",
		"geometry.zonebody",

		"balloon.foreground",
		"balloon.background",
		"title.foreground",
		"title.background",
		"title.backinactive",
		"entrylabel.foreground",
		"entrylabel.background",
		"infolabel.foreground",
		"infolabel.background",
		"calclabel.foreground",
		"calclabel.background",
		#"*entry*foreground",
		#"*entry*background",
		#"*label*foreground",
		#"*label*background",
		#"*text*foreground",
		#"*text*background",
		]

#===============================================================================
# Configuration Dialog
#===============================================================================
class ConfigDialog(tk.Toplevel):
	activepage = None
	width  = -1
	height = -1
	_PAGES = None

	def __init__(self, master, config):
		super().__init__(master)
		self.transient(master)
		self.config = config
		self.title("Preferences")

		# Action buttons
		frameButtons = tk.Frame(self)
		frameButtons.pack(side=tk.BOTTOM)
		self.buttonHelp = tk.Button(frameButtons,
				text='Help',
				image=icons.get("info"),
				compound=tk.LEFT,
				command=self.help, takefocus=tk.FALSE)
		self.buttonOk = tk.Button(frameButtons,
				text='Ok',
				image=icons.get("ok"),
				compound=tk.LEFT,
				command=self.ok,
				takefocus=tk.FALSE)
		self.buttonCancel = tk.Button(frameButtons,
				text='Cancel',
				image=icons.get("x"),
				compound=tk.LEFT,
				command=self.cancel, takefocus=tk.FALSE)
		self.buttonHelp.pack(side=tk.RIGHT, padx=5, pady=5)
		self.buttonOk.pack(side=tk.LEFT, padx=5, pady=5)
		self.buttonCancel.pack(side=tk.LEFT, padx=5, pady=5)

		# Selection listbox
		self.listbox = tk.Listbox(self,
				width=15,
				height=20,
				selectmode=tk.BROWSE,
				selectborderwidth=0,
				background="White",
				takefocus=True,
				exportselection=tk.FALSE,
				borderwidth=2,
				relief=tk.SUNKEN)
		self.listbox.pack(side=tk.LEFT, fill=tk.Y, padx=2, pady=5)
		self.listbox.bind("<<ListboxSelect>>", self.listboxChanged)

		# main frame
		#self.labelframe = tk.LabelFrame(self, text="")
		#self.labelframe.pack(expand=tk.YES, fill=tk.BOTH, padx=2, pady=5)
		self.mainframe = tk.Frame(self, borderwidth=2, relief=tk.SUNKEN)
		self.mainframe.pack(expand=tk.YES, fill=tk.BOTH, padx=2, pady=5)

		# initialize font dictionary...
		self.fontDict = tkFlair.fontDictionary()
		# and colors
		self.initColorDictionary()

		# Initialize pages, but only after the class is defined!
		if ConfigDialog._PAGES is None:
			ConfigDialog._PAGES = (
				("Colors",	ConfigDialog.colorsFrame),
				("Data",	ConfigDialog.dataProcessingFrame),
				("Fluka",	ConfigDialog.flukaFrame),
				("Fonts",	ConfigDialog.fontsFrame),
				("Geometry",	ConfigDialog.geometryFrame),
				("Gnuplot",	ConfigDialog.gnuplotFrame),
				("Interface",	ConfigDialog.interfaceFrame),
				("Mcnp",	ConfigDialog.mcnpFrame),
				("Moira",	ConfigDialog.moiraFrame),
				("Penelope",	ConfigDialog.penelopeFrame),
				("Phits",	ConfigDialog.phitsFrame),
				("Programs",	ConfigDialog.programsFrame),
				("Run",		ConfigDialog.batchSystemFrame),
			)

		# Populate pages
		self.pages = {}
		for page, method in ConfigDialog._PAGES:
			self.listbox.insert(tk.END, page)
			self.pages[page] = method(self)

		self.bind('<<Help>>',	self.help)
		self.bind('<Escape>',	self.cancel)

		self.loadFonts()
		self.rc = False

	# --------------------------------------------------------------------
	def show(self, page=None):
		if page:
			try:
				page = _PAGES[page]
			except KeyError:
				pass
		# Select last page
		self.listbox.selection_clear(0,tk.END)

		if page is not None:
			# First check if pagename exists
			if page not in self.pages:
				page = None

		if page is not None:
			self.showPage(page)
		elif ConfigDialog.activepage is None:
			self.showPage(ConfigDialog._PAGES[0][0])
		else:
			self.showPage(ConfigDialog.activepage)

		self.init()

		if ConfigDialog.width > 0:
			self.geometry("%dx%d" %
				(ConfigDialog.width, ConfigDialog.height))

		self.deiconify()
		# Get the focus
		self.focus_set()
		self.grab_set()
		self.wait_window()
		return self.rc

	# --------------------------------------------------------------------
	# Display new page
	# --------------------------------------------------------------------
	def listboxChanged(self, event):
		sel = self.listbox.get(self.listbox.curselection())
		if sel: self.showPage(sel)

	# --------------------------------------------------------------------
	def showPage(self, page):
		for i,p in enumerate(ConfigDialog._PAGES):
			if page == p[0]:
				if not self.listbox.selection_includes(i):
					self.listbox.selection_set(i)
				break

		if ConfigDialog.activepage:
			self.pages[ConfigDialog.activepage].pack_forget()
		ConfigDialog.activepage = page
		self.pages[page].pack(expand=tk.YES, fill=tk.BOTH)

	# --------------------------------------------------------------------
	def initColorDictionary(self):
		self.colorDict = {}

		# and populate color list with extra colors
		for n,v in self.config.items(tkFlair._COLOR_SECTION):
			if n not in _COLOR_LIST:
				if not n.startswith("material.") and \
				   not n.startswith("alpha."):
					_COLOR_LIST.append(n)

			self.colorDict[n] = v
		_COLOR_LIST.sort()

	# --------------------------------------------------------------------
	# Create programs frame
	# --------------------------------------------------------------------
	def programsFrame(self):
		pageframe = tk.Frame(self.mainframe)

		# ---
		row = 0
		tk.Label(pageframe, text="Kill Command").grid(
				row=row, column=0, sticky=tk.W)
		self.killCmd = tk.Entry(pageframe, background="white")
		self.killCmd.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.killCmd,
			"""Kill command. No default.
If empty uses the kill system command.
Override with the appropriate program for batch systems""")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.killCmd:
					self.chooseFile("kill", e))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Editor").grid(
				row=row, column=0, sticky=tk.W)
		self.editor = tk.Entry(pageframe, background="white")
		self.editor.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.editor,
			"Editor program. No default")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.editor:
					self.chooseFile("editor", e))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Terminal").grid(
				row=row, column=0, sticky=tk.W)
		self.terminal = tk.Entry(pageframe, background="white")
		self.terminal.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.terminal,
			"Terminal program. No default")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.terminal:
					self.chooseFile("terminal", e))
		b.grid(row=row, column=2)

		row += 1
		tk.Label(pageframe, text="Debugger").grid(
				row=row, column=0, sticky=tk.W)
		self.debugger = tk.Entry(pageframe, background="white")
		self.debugger.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.debugger, "Debugging program.")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.debugger:
					self.chooseFile("debugger", e))
		b.grid(row=row, column=2)

		pageframe.grid_columnconfigure(1, weight=1)

		return pageframe

	# --------------------------------------------------------------------
	# Create fluka programs frame
	# --------------------------------------------------------------------
	def flukaFrame(self):
		pageframe = tk.Frame(self.mainframe)

		self.flukafiles  = tk.BooleanVar(self)
		self.warnmultmat = tk.BooleanVar(self)

		# ---
		row = 0
		tk.Label(pageframe, text="Fluka Directory").grid(
				row=row, column=0, sticky=tk.W)
		self.flukadir = tk.Entry(pageframe, background="white")
		self.flukadir.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.flukadir,
			"Override or set FLUKA directory. Default: <find fluke from path>")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
			command=lambda f=self.flukadir : self.chooseDirectory(f))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="rfluka").grid(
				row=row, column=0, sticky=tk.W)
		self.rfluka = tk.Entry(pageframe, background="white")
		self.rfluka.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.rfluka,
			"rfluka executable. Default: {FlukaDir}/bin/rfluka")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.rfluka, d=self.flukadir:
					self.chooseFile("rfluka", e, d))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Executable").grid(
				row=row, column=0, sticky=tk.W)
		self.flukaexe = tk.Entry(pageframe, background="white")
		self.flukaexe.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.flukaexe,
			"fluka default executable. "
			"Default: {FlukaDir}/fluka")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.flukaexe, d=self.flukadir:
					self.chooseFile("fluka", e, d))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="fff").grid(
				row=row, column=0, sticky=tk.W)
		self.fff = tk.Entry(pageframe, background="white")
		self.fff.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.fff,
			"fff compiling program. "
			"Default: {FlukaDir}/bin/fff")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.fff, d=self.flukadir:
					self.chooseFile("fff", e, d))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="USBMAX trace scan").grid(
				row=row, column=0, sticky=tk.W)
		self.usbmax = tk.Entry(pageframe, background="white")
		self.usbmax.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.usbmax,
			"USBMAX max trace scanning program. "
			"Default: {FlukaDir}/bin/usbmax")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.usbmax, d=self.flukadir:
					self.chooseFile("usbmax", e, d))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Gplevbin projection").grid(
				row=row, column=0, sticky=tk.W)
		self.gplevbin = tk.Entry(pageframe, background="white")
		self.gplevbin.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.gplevbin,
			"USRBIN processing program for gnuplot. "
			"Default: {FlukaDir}/bin/gplevbin")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.gplevbin, d=self.flukadir:
					self.chooseFile("gplevbin", e, d))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="USBREA usrbin to ascii").grid(
				row=row, column=0, sticky=tk.W)
		self.usbrea = tk.Entry(pageframe, background="white")
		self.usbrea.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.usbrea,
			"USBREA convert USRBIN files to ascii. "
			"Default: {FlukaDir}/bin/usbrea")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.usbrea, d=self.flukadir:
					self.chooseFile("usbrea", e, d))
		b.grid(row=row, column=2)

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text="Warn on multiple material assignment to region",
				variable=self.warnmultmat,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Show warnings if a region is assigned multiple materials")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text="Show fluka files in fluka_XXX dir",
				variable=self.flukafiles,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Show the fluka files (usually links) in the running fluka_XXX directory listing")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		pageframe.grid_columnconfigure(1, weight=1)
		return pageframe

	# --------------------------------------------------------------------
	# Create moira programs frame
	# --------------------------------------------------------------------
	def moiraFrame(self):
		pageframe = tk.Frame(self.mainframe)

		# ---
		row = 0
		tk.Label(pageframe, text="Moira Directory").grid(
				row=row, column=0, sticky=tk.W)
		self.moiradir = tk.Entry(pageframe, background="white")
		self.moiradir.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.moiradir,
			"Set Moira directory. Default: no-default")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
			command=lambda f=self.moiradir : self.chooseDirectory(f))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Program").grid(
				row=row, column=0, sticky=tk.W)
		self.moiraprg = tk.Entry(pageframe, background="white")
		self.moiraprg.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.moiraprg,
			"moira submitting program. "
			"Default: {MoiraDir}/moira")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.moiraprg, d=self.moiradir:
					self.chooseFile("moira.sh", e, d))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Executable").grid(
				row=row, column=0, sticky=tk.W)
		self.moiraexe = tk.Entry(pageframe, background="white")
		self.moiraexe.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.moiraexe,
			"moira executable. "
			"Default: {MoiraDir}/moira")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.moiraexe, d=self.moiradir:
					self.chooseFile("moira", e, d))
		b.grid(row=row, column=2)

		pageframe.grid_columnconfigure(1, weight=1)
		return pageframe

	# --------------------------------------------------------------------
	# Create mcnp programs frame
	# --------------------------------------------------------------------
	def mcnpFrame(self):
		pageframe = tk.Frame(self.mainframe)

		# ---
		row = 0
		tk.Label(pageframe, text="Mcnp Directory").grid(
				row=row, column=0, sticky=tk.W)
		self.mcnpdir = tk.Entry(pageframe, background="white")
		self.mcnpdir.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.mcnpdir,
			"Set Mcnp directory. Default: no-default")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
			command=lambda f=self.mcnpdir : self.chooseDirectory(f))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Executable").grid(
				row=row, column=0, sticky=tk.W)
		self.mcnpexe = tk.Entry(pageframe, background="white")
		self.mcnpexe.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.mcnpexe,
			"mcnp executable. "
			"Default: {McnpDir}/mcnp")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.mcnpexe, d=self.mcnpdir:
					self.chooseFile("mcnp", e, d))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Datapath").grid(
				row=row, column=0, sticky=tk.W)
		self.mcnpdata = tk.Entry(pageframe, background="white")
		self.mcnpdata.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.mcnpdata,
			"mcnp datapath directory. "
			"Default: $DATAPATH")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
			command=lambda f=self.mcnpdata :
					self.chooseDirectory(f,"DATAPATH"))
		b.grid(row=row, column=2)

		pageframe.grid_columnconfigure(1, weight=1)
		return pageframe

	# --------------------------------------------------------------------
	# Create penelope programs frame
	# --------------------------------------------------------------------
	def penelopeFrame(self):
		pageframe = tk.Frame(self.mainframe)

		# ---
		row = 0
		tk.Label(pageframe, text="Penelope Directory").grid(
				row=row, column=0, sticky=tk.W)
		self.penelopedir = tk.Entry(pageframe, background="white")
		self.penelopedir.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.penelopedir,
			"Set Penelope directory. Default: no-default")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
			command=lambda f=self.penelopedir : self.chooseDirectory(f))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Executable").grid(
				row=row, column=0, sticky=tk.W)
		self.penelopeprg = tk.Entry(pageframe, background="white")
		self.penelopeprg.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.penelopeprg,
			"Penelope executable. "
			"Default: {PenelopeDir}/bin/penmain")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.penelopeprg, d=self.penelopedir:
					self.chooseFile("penmain", e, d))
		b.grid(row=row, column=2)

		pageframe.grid_columnconfigure(1, weight=1)
		return pageframe

	# --------------------------------------------------------------------
	# Create phits programs frame
	# --------------------------------------------------------------------
	def phitsFrame(self):
		pageframe = tk.Frame(self.mainframe)

		# ---
		row = 0
		tk.Label(pageframe, text="Phits Directory").grid(
				row=row, column=0, sticky=tk.W)
		self.phitsdir = tk.Entry(pageframe, background="white")
		self.phitsdir.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.phitsdir,
			"Set Phits directory. Default: no-default")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
			command=lambda f=self.phitsdir : self.chooseDirectory(f))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Executable").grid(
				row=row, column=0, sticky=tk.W)
		self.phitsprg = tk.Entry(pageframe, background="white")
		self.phitsprg.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.phitsprg,
			"Phits executable. "
			"Default: {PhitsDir}/phits")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.phitsprg, d=self.phitsdir:
					self.chooseFile("phits", e, d))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="Datapath").grid(
				row=row, column=0, sticky=tk.W)
		self.phitsdata = tk.Entry(pageframe, background="white")
		self.phitsdata.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.phitsdata,
			"phits datapath directory. "
			"Default: $DATAPATH")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
			command=lambda f=self.phitsdata :
					self.chooseDirectory(f,"DATAPATH"))
		b.grid(row=row, column=2)

		pageframe.grid_columnconfigure(1, weight=1)
		return pageframe

	# --------------------------------------------------------------------
	# Choose Directory
	# --------------------------------------------------------------------
	def chooseDirectory(self, widget, envvar=None):
		directory = widget.get()
		if not directory and envvar is not None:
			directory = os.getenv(envvar)
		if directory is None: directory = ""

		directory = bFileDialog.askdirectory(
					master=self,
					title="Select directory",
					initialdir=directory.strip())
		directory = directory.rstrip(os.sep)
		if directory:
			widget.delete(0,tk.END)
			#if envvar is not None and \
			#   os.getenv(envvar).rstrip(os.sep)==directory:
			#	return	# do not insert anything
			widget.insert(0, directory)

	# --------------------------------------------------------------------
	def chooseFile(self, default, entry, directory=None): #, envvar=None):
		if directory is None:
			directory = tkFlair.prgDir
		elif isinstance(directory,tk.Entry):
			directory = directory.get()

		#if not directory and envvar is not None:
		#	directory = os.getenv(envvar)

		fn = entry.get().strip()
		if not fn:
			fn = default
		if fn[0]!="/":
			fn = os.path.join(directory, "bin", fn)

		ft=[(f"{default} files",f"*{default}*"), ("All","*")]

		fn = bFileDialog.askopenfilename(master=self,
				title=f"Select {default}",
				initialdir=directory.strip(),
				initialfile=fn,
				filetypes=ft)
		if fn:
			entry.delete(0, tk.END)
			entry.insert(0, fn)

	# --------------------------------------------------------------------
	# Create batch system frame
	# --------------------------------------------------------------------
	def batchSystemFrame(self):
		pageframe = tk.Frame(self.mainframe, padx=5, pady=5)
		self.enableBatchSystem = tk.BooleanVar(self)

		# row = 0
		# b = tk.Checkbutton(pageframe, text="Enable Batch System",
		#		variable=self.enableBatchSystem,
		#		onvalue=1, offvalue=0,
		#		anchor=tk.W, command=self.batchEdit)
		# if self.enableBatchSystem : b.select()

		#tkExtra.Balloon.set(b,
		#	"Enable Batch System")
		#b.grid(row=row, column=0, columnspan=2, sticky=tk.W)
		#b.pack(side=tk.TOP, fill=tk.X)

		# ---
		row = 0
		tk.Label(pageframe, text="Spawn Run Name").grid(
				row=row, column=0, sticky=tk.E)
		self.spawnRunName = tk.Entry(pageframe, background="white")
		self.spawnRunName.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.spawnRunName,
			"Prefix to be used for all temporary files created by flair")

		# ---
		row += 1
		tk.Label(pageframe, text="Attach timeout").grid(
				row=row, column=0, sticky=tk.E)
		self.timethreshold = tkExtra.IntegerEntry(pageframe, background="white")
		self.timethreshold.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.timethreshold,
			"Time out threshold for attaching to a running job")
		tk.Label(pageframe, text="[s]").grid(row=row, column=2, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text="Refresh Interval (s)").grid(
				row=row, column=0, sticky=tk.W)
		self.refreshinterval = tkExtra.IntegerEntry(pageframe, background="white")
		self.refreshinterval.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.refreshinterval,
			"Refresh interval for updating information on running programs")
		tk.Label(pageframe, text="[s]").grid(row=row, column=2, sticky=tk.W)

		# --- Queue Types ---
		row += 1
		frame = tk.LabelFrame(pageframe, text="Queues / Submit command",
				relief=tk.GROOVE)
		frame.grid(row=row, column=0, columnspan=3, sticky=tk.NSEW, padx=5, pady=5)

		pageframe.grid_rowconfigure(row, weight=1)
		pageframe.grid_columnconfigure(1, weight=1)

		# ----
		self.queueTypes = tkExtra.MultiListbox(frame,
			(("Queue ", 16, None),
			 ("Execution command", 30, None)))
		self.queueTypes.pack(fill=tk.BOTH, expand=tk.YES)

		self.queueTypes.bindList('<ButtonRelease-1>',
				self.queueRelease)
		self.queueTypes.bindList('<Return>',
				self.queueEdit)
		self.queueTypes.bindLeftRight()
		self.queueTypes.setPopupMenu(
				[('Add',    0, self.queueAdd,    icons.get("add")),
				 ('Delete', 0, self.queueDelete, icons.get("x")),
				 ('Clone',  0, self.queueClone,  icons.get("clone"))] )
		return pageframe

	# ----------------------------------------------------------------------
	def queueEdit(self, event):
		sel = self.queueTypes.curselection()
		if len(sel) != 1: return
		sel = int(sel[0])
		try: active = event.widget.index(tk.ACTIVE)
		except: return
		if active != sel: return
		tkExtra.InPlaceEdit(event.widget)

	# ----------------------------------------------------------------------
	def queueRelease(self, event=None):
		act = event.widget.nearest(event.y)
		if act: self.queueTypes.activate(act)
		self.queueEdit(event)

	# ----------------------------------------------------------------------
	def queueAdd(self):
		self.queueTypes.insert(tk.END, ("myQueue", ""))
		self.queueTypes.selection_clear(0,tk.END)
		self.queueTypes.selection_set(tk.END)
		self.queueTypes.activate(tk.END)
		self.queueTypes.see(tk.END)
		self.queueTypes.focus_set()

	# ----------------------------------------------------------------------
	def queueDelete(self):
		lst = list(map(int, self.queueTypes.curselection()))
		if not lst: return
		lst.reverse()
		for i in lst:
			self.queueTypes.delete(i)

	# ----------------------------------------------------------------------
	def queueClone(self):
		sel = list(map(int,self.queueTypes.curselection()))
		if not sel: return
		oldSize = self.queueTypes.size()
		for i in sel:
			item = self.queueTypes.get(i)
			self.queueTypes.insert(tk.END, item)
		self.queueTypes.selection_clear(0,tk.END)
		self.queueTypes.selection_set(oldSize,tk.END)
		self.queueTypes.activate(tk.END)

	# ----------------------------------------------------------------------
	def batchEdit(self):
		if self.enableBatchSystem.get():
			self.queueTypes["state"] = tk.NORMAL
		else:
			self.queueTypes["state"] = tk.DISABLED

	# --------------------------------------------------------------------
	# Graphical Interface frame
	# --------------------------------------------------------------------
	def interfaceFrame(self):
		pageframe = tk.Frame(self.mainframe)

		self.skipintro  = tk.BooleanVar(self)
		self.keepbackup = tk.BooleanVar(self)
		self.cleanup    = tk.BooleanVar(self)
		self.hrlines    = tk.BooleanVar(self)

		row = 0
		b = tk.Checkbutton(pageframe, text="Skip About dialog",
				variable=self.skipintro,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Skip the about dialog during startup")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text="Keep backups",
				variable=self.keepbackup,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"Flag whether you want to keep backups of input and project files")

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text="Cleanup temporary files",
				variable=self.cleanup,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"Cleanup flair temporary files (Recommended: ON)")

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text="Editor: Show separation lines",
				variable=self.hrlines,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"Show horizontal separation lines between cards in input editor")#

		# ---
		row += 1
		tk.Label(pageframe, text="Manual viewer").grid(
				row=row, column=0, sticky=tk.W)

		self.manual_viewer = tkExtra.Combobox(pageframe, width=10)
		self.manual_viewer.grid(row=row, column=1, sticky=tk.EW)
		self.manual_viewer.fill(Manual.VIEWERS)
		tkExtra.Balloon.set(self.manual_viewer,
			"Select manual viewer or browser")

		# ---
		row += 1
		tk.Label(pageframe, text="Window scaling").grid(
				row=row, column=0, sticky=tk.W)
		self.dpiscale = tk.Scale(pageframe, from_=1, to_=5,
				orient=tk.HORIZONTAL, showvalue=1)
		self.dpiscale.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.dpiscale,
			"DPI scaling of text and icons by an integer factor (for HiDPI)")

		# ---
		row += 1
		tk.Label(pageframe, text="Temporary prefix").grid(
				row=row, column=0, sticky=tk.E)
		self.tmpprefix = tk.Entry(pageframe, background="white")
		self.tmpprefix.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.tmpprefix,
			"Prefix to be used for all temporary files created by flair")

		# ---
		row += 1
		tk.Label(pageframe, text="Auto save time").grid(
				row=row, column=0, sticky=tk.E)
		self.autosave = tkExtra.IntegerEntry(pageframe, background="white")
		self.autosave.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.autosave,
			"Input file auto save period. File is written in pickle format")
		tk.Label(pageframe, text="[s]").grid(row=row, column=2, sticky=tk.W)

		# ---
		#row += 1
		#tk.Label(pageframe, text="Key time threshold").grid(
		#	row=row, column=0, sticky=tk.E)
		#self.keytime = tkExtra.IntegerEntry(pageframe, width=7, background="white")
		#self.keytime.grid(row=row, column=1, sticky=tk.EW)
		#tkExtra.Balloon.set(self.keytime,
		#	"Set the keyboard time out threshold during listbox searches")
		#tk.Label(pageframe, text="[ms]").grid(row=row, column=2, sticky=tk.W)

		# ---
		#row += 1
		#tk.Label(pageframe, text="Balloon delay").grid(
		#	row=row, column=0, sticky=tk.E)
		#self.balloondelay = tkExtra.IntegerEntry(pageframe, width=7, background="white")
		#self.balloondelay.grid(row=row, column=1, sticky=tk.EW)
		#tkExtra.Balloon.set(self.balloondelay,
		#	"Delay time to display the help-balloon")
		#tk.Label(pageframe, text="[ms]").grid(row=row, column=2, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text="Card Display timeout").grid(
				row=row, column=0, sticky=tk.E)
		self.cardstimeout = tkExtra.IntegerEntry(pageframe, background="white")
		self.cardstimeout.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.cardstimeout,
			"Time out for displaying long input files")
		tk.Label(pageframe, text="[s]").grid(row=row, column=2, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text="Time format").grid(
			row=row, column=0, sticky=tk.E)
		self.timeformat = tk.Entry(pageframe, width=20, background="white")
		self.timeformat.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.timeformat,
			"Set the format to display time in file dialogs. Check strftime python syntax")

		pageframe.grid_columnconfigure(1, weight=1)

		return pageframe

	# --------------------------------------------------------------------
	def dataProcessingFrame(self):
		pageframe = tk.Frame(self.mainframe)

		row = 0
		self.usrTypes = tkExtra.MultiListbox(pageframe,
			(("Type", 12, None),
			 ("Default Output File Name", 20, None)),
			 height=6)
		self.usrTypes.grid(row=row, column=0, columnspan=3, sticky=tk.NSEW)
		self.usrTypes.bindList('<ButtonRelease-1>',
				self.usrTypesRelease)
		self.usrTypes.bindLeftRight()

		# ---
		row += 1
		self.usrRules = tkExtra.ExListbox(pageframe,
					height=4,
					takefocus=tk.TRUE,
					exportselection=tk.TRUE)
		self.usrRules.grid(row=row, column=0, columnspan=3, sticky=tk.NSEW)
		self.usrRules.bind('<ButtonRelease-1>', self.usrRulesRelease)
		self.usrRules.setPopupMenu(
				[('Add',    0, self.usrRulesAdd,    icons.get("add")),
				 ('Delete', 0, self.usrRulesDelete, icons.get("x")),
				 ('Clone',  0, self.usrRulesClone,  icons.get("clone"))] )

		# ---
		row += 1
		tk.Label(pageframe, text="USRBIN processing").grid(
				row=row, column=0, sticky=tk.W)
		self.usbsuw = tk.Entry(pageframe, background="white")
		self.usbsuw.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.usbsuw,
			"USRBIN processing program. Default: {FlukaDir}/bin/usbsuw")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.usbsuw:
					self.chooseFile("usbsuw", e))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="USRTRACK processing").grid(
				row=row, column=0, sticky=tk.W)
		self.ustsuw = tk.Entry(pageframe, background="white")
		self.ustsuw.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.ustsuw,
			"USRTRACK/USRCOLL processing program. Default: {FlukaDir}/bin/ustsuw")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.ustsuw:
					self.chooseFile("ustsuw", e))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="USRBDX processing").grid(
				row=row, column=0, sticky=tk.W)
		self.usxsuw = tk.Entry(pageframe, background="white")
		self.usxsuw.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.usxsuw,
			"USRBDX processing program. Default: {FlukaDir}/bin/usxsuw")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.usxsuw:
					self.chooseFile("usxsuw", e))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="USRYIELD processing").grid(
				row=row, column=0, sticky=tk.W)
		self.usysuw = tk.Entry(pageframe, background="white")
		self.usysuw.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.usysuw,
			"USRYIELD processing program. Default: {FlukaDir}/bin/usysuw")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.usysuw:
					self.chooseFile("usysuw", e))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="RESNUCLEI processing").grid(
				row=row, column=0, sticky=tk.W)
		self.usrsuw = tk.Entry(pageframe, background="white")
		self.usrsuw.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.usrsuw,
			"RESNUCLEI processing program. Default: {FlukaDir}/bin/usrsuw")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.usrsuw:
					self.chooseFile("usrsuw", e))
		b.grid(row=row, column=2)

		# ---
		row += 1
		tk.Label(pageframe, text="DETECT processing").grid(
				row=row, column=0, sticky=tk.W)
		self.detsuw = tk.Entry(pageframe, background="white")
		self.detsuw.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.detsuw,
			"DETECT processing program. Default: {FlukaDir}/bin/detsuw")

		b = tk.Button(pageframe, image=icons.get("load"), pady=0,
				command=lambda e=self.detsuw:
					self.chooseFile("detsuw", e))
		b.grid(row=row, column=2)

		# ---
		pageframe.grid_rowconfigure(0, weight=1)
		pageframe.grid_columnconfigure(1, weight=1)

		return pageframe

	# ----------------------------------------------------------------------
	def usrTypesRelease(self, event=None):
		listbox = event.widget
		act = listbox.nearest(event.y)
		if act: self.usrTypes.activate(act)
		sel = self.usrTypes.curselection()
		if len(sel) != 1: return
		sel = int(sel[0])
		try: active = listbox.index(tk.ACTIVE)
		except: return
		if active != sel: return
		# Find list box
		lid = self.usrTypes.listboxIndex(listbox)
		if lid==0: return
		tkExtra.InPlaceEdit(listbox)

	# ----------------------------------------------------------------------
	def usrRulesRelease(self, event=None):
		act = self.usrRules.nearest(event.y)
		if act: self.usrRules.activate(act)
		sel = self.usrRules.curselection()
		if len(sel) != 1: return
		sel = int(sel[0])
		try: active = self.usrRules.index(tk.ACTIVE)
		except: return
		if active != sel: return
		tkExtra.InPlaceEdit(self.usrRules)
		self.usrRules.selection_set(sel)

	# ----------------------------------------------------------------------
	def usrRulesAdd(self):
		self.usrRules.insert(tk.END, r"+\I\d\d\d_fort.\U")
		self.usrRules.selection_clear(0,tk.END)
		self.usrRules.selection_set(tk.END)
		self.usrRules.activate(tk.END)
		self.usrRules.see(tk.END)
		self.usrRules.focus_set()

	# ----------------------------------------------------------------------
	def usrRulesDelete(self):
		lst = list(map(int, self.usrRules.curselection()))
		if not lst: return
		lst.reverse()
		for i in lst:
			self.usrRules.delete(i)

	# ----------------------------------------------------------------------
	def usrRulesClone(self):
		sel = list(map(int,self.usrRules.curselection()))
		if not sel: return
		oldSize = self.usrRules.size()
		for i in sel:
			item = self.usrRules.get(i)
			self.usrRules.insert(tk.END, item)
		self.usrRules.selection_clear(0,tk.END)
		self.usrRules.selection_set(oldSize,tk.END)
		self.usrRules.activate(tk.END)

	# --------------------------------------------------------------------
	# Gnuplot Frame
	# --------------------------------------------------------------------
	def gnuplotFrame(self):
		pageframe = tk.Frame(self.mainframe)

		frame = tk.Frame(pageframe)
		frame.pack(side=tk.TOP, padx=5, fill=tk.X)

		# -- gnuplot --
		row = 0
		tk.Label(frame, text="Gnuplot").grid(
				row=row, column=0, sticky=tk.W)
		self.gnuplot = tk.Entry(frame, background="white")
		self.gnuplot.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.gnuplot, "Gnuplot program.")

		b = tk.Button(frame, image=icons.get("load"), pady=0,
				command=lambda e=self.gnuplot:
					self.chooseFile("gnuplot", e))
		b.grid(row=row, column=2)

		# -- terminal --
		row += 1
		tk.Label(frame, text="Terminal:").grid(row=row, column=0, sticky=tk.W)
		self.gnuplotTerminal = tk.Entry(frame, background="White")
		self.gnuplotTerminal.grid(row=row, column=1, columnspan=2, sticky=tk.EW)
		tkExtra.Balloon.set(self.gnuplotTerminal,
			"Gnuplot terminal. Default=empty, let gnuplot decide")

		frame.grid_columnconfigure(1, weight=1)

		# -- commands --
		frame = tk.LabelFrame(pageframe, text="Global Commands",
				relief=tk.GROOVE)
		frame.pack(side=tk.TOP, padx=5, pady=5, fill=tk.BOTH, expand=tk.YES)

		self.gnuplotCommands = tk.Text(frame, background="white", width=40, height=7)
		self.gnuplotCommands.pack(fill=tk.BOTH, expand=tk.YES)

		# -- File Types --
		frame = tk.LabelFrame(pageframe, text="File Types",
				relief=tk.GROOVE)
		frame.pack(padx=5, pady=5, fill=tk.BOTH, expand=tk.YES)

		self.gnuplotFileTypes = tkExtra.MultiListbox(frame,
			(("Type", 5, None),
			 ("Settings", 30, None)))
		self.gnuplotFileTypes.pack(fill=tk.BOTH, expand=tk.YES)

		self.gnuplotFileTypes.bindList('<ButtonRelease-1>',
				self.gnuplotRelease)
		self.gnuplotFileTypes.bindLeftRight()
		self.gnuplotFileTypes.setPopupMenu(
				[('Add',    0, self.gnuplotAdd,    icons.get("add")),
				 ('Delete', 0, self.gnuplotDelete, icons.get("x")),
				 ('Clone',  0, self.gnuplotClone,  icons.get("clone"))] )
		return pageframe

	# ----------------------------------------------------------------------
	def gnuplotRelease(self, event=None):
		listbox = event.widget
		act = listbox.nearest(event.y)
		if act: self.gnuplotFileTypes.activate(act)
		sel = self.gnuplotFileTypes.curselection()
		if len(sel) != 1: return
		sel = int(sel[0])
		try: active = listbox.index(tk.ACTIVE)
		except: return
		if active != sel: return
		tkExtra.InPlaceEdit(listbox)

	# ----------------------------------------------------------------------
	def gnuplotAdd(self):
		self.gnuplotFileTypes.insert(tk.END, (".ext", ""))
		self.gnuplotFileTypes.selection_clear(0,tk.END)
		self.gnuplotFileTypes.selection_set(tk.END)
		self.gnuplotFileTypes.activate(tk.END)
		self.gnuplotFileTypes.see(tk.END)
		self.gnuplotFileTypes.focus_set()

	# ----------------------------------------------------------------------
	def gnuplotDelete(self):
		lst = list(map(int, self.gnuplotFileTypes.curselection()))
		if not lst: return
		lst.reverse()
		for i in lst:
			self.gnuplotFileTypes.delete(i)

	# ----------------------------------------------------------------------
	def gnuplotClone(self):
		sel = list(map(int,self.gnuplotFileTypes.curselection()))
		if not sel: return
		oldSize = self.gnuplotFileTypes.size()
		for i in sel:
			item = self.gnuplotFileTypes.get(i)
			self.gnuplotFileTypes.insert(tk.END, item)
		self.gnuplotFileTypes.selection_clear(0,tk.END)
		self.gnuplotFileTypes.selection_set(oldSize,tk.END)
		self.gnuplotFileTypes.activate(tk.END)

	# --------------------------------------------------------------------
	# Geometry Frame
	# --------------------------------------------------------------------
	def geometryFrame(self):
		pageframe = tk.Frame(self.mainframe)

		self.laptopMode  = tk.BooleanVar(self)
		self.invertWheel = tk.BooleanVar(self)
		self.pinZoom     = tk.BooleanVar(self)
		self.useQUA      = tk.BooleanVar(self)
		self.useBOX      = tk.BooleanVar(self)

		# ---
		row = 0
		b = tk.Checkbutton(pageframe, text="Laptop Mode",
				variable=self.laptopMode,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"Enable laptop mode for geometry editor: "
			"Swap middle-mouse button with right-mouse button."
			"Requires restart of the program" )

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text="Invert Mouse Wheel",
				variable=self.invertWheel,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"Invert the zooming with the mouse wheel."
			"Requires restart of the program" )

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text="Zoom on Mouse location",
				variable=self.pinZoom,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"Check to zoom in on mouse location. "
			"Uncheck to zoom on viewport center. "
			"[Alt] inverts the action" )

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text="Prefer QUA on rotation",
				variable=self.useQUA,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"Create infinite QUA rotated instead of RCC "
			"when rotating infinite cylinders [XYZ]CC or [XYZ]EC")

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text="Use of deprecated bodies (BOX,RAW,ARB)",
				variable=self.useBOX,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"Allow the use of deprecated bodies BOX/RAW/WED/ARB for back compatibility")

		# ---
		row += 1
		tk.Label(pageframe, text="# CPU/Cores").grid(
				row=row, column=0, sticky=tk.W)
		if multiprocessing:
			cores = multiprocessing.cpu_count()
		else:
			cores = 8
		self.cores = tk.Scale(pageframe, from_=0, to_=cores,
				orient=tk.HORIZONTAL, showvalue=1)
		self.cores.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.cores, "Number of cores per viewport to use.\n"
						"0=all available (default); 1=synchronous, "
						"nn=user defined.\n"
						"Restart of flair is required.")

		# ---
		row += 1
		tk.Label(pageframe, text="Snapping distance").grid(
				row=row, column=0, sticky=tk.W)
		self.snap = tk.Scale(pageframe, from_=1, to_=30,
				orient=tk.HORIZONTAL, showvalue=1)
		self.snap.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.snap, "Distance in pixels for snapping.")

		# ---
		row += 1
		tk.Label(pageframe, text="Expansion Limit:").grid(row=row, column=0, sticky=tk.W)
		self.explimit = tkExtra.FloatEntry(pageframe, background="white")
		self.explimit.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.explimit,
			"Region parentheses expansion limit in number of terms")

		# ---
		row += 1
		tk.Label(pageframe, text="Zero").grid(
				row=row, column=0, sticky=tk.W)
		self.zero = tkExtra.FloatEntry(pageframe, background="white")
		self.zero.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.zero,
			"Threshold value to be considered zero or equal in comparison.")

		# ---
		row += 1
		tk.Label(pageframe, text="Infinite:").grid(row=row, column=0, sticky=tk.W)
		self.infinite = tkExtra.FloatEntry(pageframe, background="white")
		self.infinite.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.infinite,
			"Length of cylinders when from infinite [XYZ]CC are converted to finite RCC")

		# ---
		row += 1
		tk.Label(pageframe, text="Accuracy:").grid(row=row, column=0, sticky=tk.W)
		self.accuracy = tk.IntVar()
		self.accuracy.set(15)
		s = tk.Spinbox(pageframe, text=self.accuracy, from_=3, to=17, background="white")
		s.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(s,
			"Number of digits to keep during transformation")

		pageframe.grid_columnconfigure(1, weight=1)

		return pageframe

	# --------------------------------------------------------------------
	# Fonts Frame
	# --------------------------------------------------------------------
	def fontsFrame(self):
		fontframe = tk.Frame(self.mainframe)

		self.fontChanged = False
		self.fontSize   = tk.StringVar(self)
		self.fontAbs    = tk.BooleanVar(self)
		self.fontBold   = tk.BooleanVar(self)
		self.fontItalic = tk.BooleanVar(self)
		self.fontFamily = tk.StringVar(self)
		self.font       = tkFont.Font(self,
					('helvetica', 12, 'normal'))
		self.fontSize.set('12')
		self.fontFamily.set('helvetica')

		# --- Which Font ---
		self.fontList = tkExtra.ExListbox(fontframe,
					takefocus=tk.TRUE,
					selectmode=tk.EXTENDED,
					exportselection=tk.TRUE)
		self.fontList.bind("<<ListboxSelect>>", self.fontListChange)
		self.fontList.pack(side=tk.LEFT, padx=2, pady=2,
					fill=tk.BOTH, expand=tk.YES)
		for font in tkFlair._FONT_LIST:
			self.fontList.insert(tk.END, font)

		# --- Font ---
		rightframe = tk.Frame(fontframe)
		rightframe.pack(side=tk.RIGHT, padx=2, pady=2, fill=tk.BOTH, expand=tk.YES)

		frame = tk.Frame(rightframe)
		frame.pack(side=tk.TOP, padx=2, pady=2, fill=tk.BOTH, expand=tk.YES)

		#tk.Label(frame, justify=tk.LEFT, text='Font:').pack(side=tk.TOP, anchor=tk.W)
		self.listFontName = tkExtra.ExListbox(frame,
					height=5,
					takefocus=tk.FALSE,
					exportselection=tk.FALSE)
		self.listFontName.pack(side=tk.LEFT, expand=tk.TRUE, fill=tk.BOTH)
		self.listFontName.bind('<ButtonRelease-1>',
					self.listFontButtonRelease)
		#self.listFontName.bind("<<ListboxSelect>>",
		#			self.listFontButtonRelease)
		self.listFontName.bind("<KeyRelease>",
					self.listFontKeyRelease)
		sb = ttk.Scrollbar(frame)
		sb.pack(side=tk.LEFT, fill=tk.Y)
		sb.config(command=self.listFontName.yview)
		self.listFontName.config(yscrollcommand=sb.set)

		# --- Param ---
		frame = tk.Frame(rightframe)
		frame.pack(padx=5, pady=2, fill=tk.X)
		tk.Label(frame, text='Size:').pack(side=tk.LEFT, anchor=tk.W)
		self.menuFontSize = tkExtra.ExOptionMenu(frame,
					self.fontSize,
					command=self.setFontSample,
					*tkFlair._FONT_SIZE)
		self.menuFontSize.pack(side=tk.LEFT, anchor=tk.W)

		frame = tk.Frame(rightframe)
		frame.pack(padx=5, pady=2, fill=tk.X)
		self.fontAbs.set(1)
		self.checkFontAbs = tk.Checkbutton(frame, text='DPI',
					variable=self.fontAbs,
					onvalue=1, offvalue=0,
					command=self.setFontSample)
		self.checkFontAbs.pack(side=tk.LEFT, anchor=tk.W)

		self.checkFontBold = tk.Checkbutton(frame, text='Bold',
					variable=self.fontBold,
					onvalue=1, offvalue=0,
					command=self.setFontSample)
		self.checkFontBold.pack(side=tk.LEFT, anchor=tk.W)

		self.checkFontItalic = tk.Checkbutton(frame, text='Italic',
					variable=self.fontItalic,
					onvalue=1, offvalue=0,
					command=self.setFontSample)
		self.checkFontItalic.pack(side=tk.LEFT, anchor=tk.W)

		# --- Sample ---
		frame = tk.Frame(rightframe, relief=tk.SOLID, borderwidth=1)
		frame.pack(side=tk.TOP, padx=5, pady=5, expand=tk.TRUE, fill=tk.BOTH)

		self.labelFontSample = tk.Label(frame,
			text =  "AaBbCcDdEe\nFfGgHhIiJjK\n"
				"AαBβΓγΔδEε\n"
				"ZζHηΘθIιKκ\n"
				"ΛλMμNνΞξOo\n"
				"ΠπPρΣσTτYυΦ\n"
				"ϕXχΨψΩω\n"
				"1234567890\n#:+=(){}[]",
			justify=tk.LEFT, font=self.font)
		self.labelFontSample.pack(expand=tk.TRUE, fill=tk.BOTH)
		self.fontFamily.trace("w", self.fontFamilyChange)
		return fontframe

	# --------------------------------------------------------------------
	def loadFonts(self):
		fonts = list(tkFont.families(self))
		fonts.sort()
		# remove duplicates
		self.listFontName.insert(tk.END,'')
		for font in fonts:
			self.listFontName.insert(tk.END, font)
		self.listFontName.see(0)
		self.listFontName.activate(0)
		self.listFontName.selection_set(0)

	# --------------------------------------------------------------------
	def listFontButtonRelease(self, event=None):
		font = self.listFontName.get(tk.ANCHOR)
		self.fontFamily.set(font)
		self.setFontSample()

	# --------------------------------------------------------------------
	def listFontKeyRelease(self, event=None):
		self.after_idle(self.listFontButtonRelease)

	# --------------------------------------------------------------------
	def setFontSample(self, event=None):
		self.fontChanged = True
		fontFamily = self.fontFamily.get()

		if self.fontAbs.get():
			fontSign = 1
		else:
			fontSign = -1

		if self.fontBold.get():
			fontWeight = tkFont.BOLD
		else:
			fontWeight = tkFont.NORMAL

		if self.fontItalic.get():
			fontSlant = tkFont.ITALIC
		else:
			fontSlant = tkFont.ROMAN

		self.font.config(size=fontSign*int(self.fontSize.get()),
			weight=fontWeight, slant=fontSlant,
			family=fontFamily)

	# --------------------------------------------------------------------
	def fontListChange(self, event=None):
		if self.fontCurrent and self.fontChanged:
			for font in self.fontCurrent:
				self.fontDict[font] = \
					(self.font.cget("family"), self.font.cget("size"),
					 self.font.cget("weight"), self.font.cget("slant"))

		sel = self.fontList.curselection()
		self.fontCurrent = list(map(self.fontList.get, sel))
		if self.fontCurrent:
			info = self.fontDict.get(self.fontCurrent[0])
			family = info[0]
			self.fontFamily.set(family)
			sz = int(info[1])
			if sz>0:
				self.fontAbs.set(1)
			else:
				self.fontAbs.set(0)
			self.fontSize.set(abs(sz))
			self.fontBold.set(info[2] == tkFont.BOLD)
			self.fontItalic.set(info[3] == tkFont.ITALIC)
		else:
			family = ""
			self.fontFamily.set("")
			self.fontAbs.set(1)
			self.fontSize.set(12)
			self.fontBold.set(0)
			self.fontItalic.set(0)

		self.listFontName.selection_clear(0, tk.END)
		for i in range(self.listFontName.size()):
			if self.listFontName.get(i) == family:
				self.listFontName.selection_set(i)
				self.listFontName.selection_anchor(i)
				self.listFontName.activate(i)
				self.listFontName.see(i)
				break
		self.setFontSample()
		self.fontChanged = False

	# --------------------------------------------------------------------
	def fontFamilyChange(self, *abc):
		if self.fontFamily.get()=="":
			state = tk.DISABLED
		else:
			state = tk.NORMAL

		self.menuFontSize["state"]    = state
		self.checkFontAbs["state"]    = state
		self.checkFontBold["state"]   = state
		self.checkFontItalic["state"] = state

	# --------------------------------------------------------------------
	def colorsFrame(self):
		pageframe = tk.Frame(self.mainframe)

		self.colorsList = tkExtra.MultiListbox(pageframe,
					(("Item", 20, None),
					 ("Color", 15, None)))
		self.colorsList.pack(expand=tk.YES, fill=tk.BOTH)
		self.colorsList.setPopupMenu(
				[("Edit",   0, self.editColor,icons.get("edit"))])
				#("Add",    0, self.addColor, icons.get("add")),
				#("Delete", 0, self.delColor, icons.get("x"))])

		self.colorsList.bindList('<ButtonRelease-1>', self.editColor)
		self.colorsList.sortAssist = None

		for name in _COLOR_LIST:
			color = str(self.colorDict.get(name,""))
			self.colorsList.insert(tk.END, (name,color))
			try:
				self.colorsList.listbox(1).itemconfig(tk.END, background=color)
			except:
				self.colorDict[name] = ""

		return pageframe

	# ----------------------------------------------------------------------
	# Edit color
	# ----------------------------------------------------------------------
	def editColor(self, event=None):
		listbox = self.focus_get()
		if listbox is None: return

		if event is not None:
			active = self.colorsList.nearest(event.y)
			if active is not None:
				self.colorsList.activate(active)
			else:
				return
		else:
			active = self.colorsList.index(tk.ACTIVE)

		if len(self.colorsList.curselection())>1: return

		#values = self.colorsList.get(active)
		lid    = self.colorsList.listboxIndex(listbox)

		# Edit and change value
		if lid == 1:
			name = self.colorsList.listbox(0).get(active)
			edit = tkExtra.InPlaceColor(listbox)

			if edit.value is None or edit.value=="":
				self.colorsList.listbox(lid).set(active, "")
			else:
				self.colorsList.listbox(lid).set(active, edit.value)
				self.colorsList.listbox(lid).itemconfig(active,
					background=edit.value)
				self.colorDict[name] = str(edit.value)

	# --------------------------------------------------------------------
	def _setBool(self, var, section, name):
		try: var.set(self.config.getboolean(section, name))
		except configparser.NoOptionError: pass

	# --------------------------------------------------------------------
	def _setStr(self, var, section, name):
		var.delete(0,tk.END)
		try: var.insert(0, self.config.get(section, name))
		except: pass

	# --------------------------------------------------------------------
	def _setInt(self, var, section, name):
		try: var.set(0, self.config.get(section, name))
		except: pass

	# --------------------------------------------------------------------
	# Initialize configuration dialog
	# --------------------------------------------------------------------
	def init(self):
		# -------------------------------------------------------------
		section = tkFlair._FLAIR_SECTION
		self._setBool(self.skipintro,	section,  "skipintro")
		self._setStr(self.autosave,	section,  "autosave")
		#self._setStr(self.keytime,	section,  "keytimethreshold")
		#self._setStr(self.balloondelay,section,  "balloondelay")
		self._setStr(self.timeformat,	section,  "timeformat")
		self.manual_viewer.set(self.config.get(section,"manualviewer"))
		self.dpiscale.set(self.config.get(section,"dpiscale"))

		# -------------------------------------------------------------
		section = tkFlair._PROJECT_SECTION
		self._setBool(self.keepbackup, section,   "keepbackup")
		self._setBool(self.cleanup,    section,   "cleanup")
		self._setStr(self.killCmd,     section,   "kill")
		self._setStr(self.editor,      section,   "editor")
		self._setStr(self.terminal,    section,   "terminal")
		self._setStr(self.debugger,    section,   "debugger")
		self._setStr(self.tmpprefix,   section,   "tmpprefix")
		self._setStr(self.timethreshold,section,  "timethreshold")
		self._setStr(self.refreshinterval,section,"refreshinterval")

		# -------------------------------------------------------------
		section = FlukaConfig.SECTION
		self._setStr(self.flukadir,    section, "directory")
		self._setStr(self.rfluka,      section, "program")
		self._setStr(self.flukaexe,    section, "executable")
		self._setStr(self.fff,         section, "fff")
		self._setStr(self.usbsuw,      section, "usbsuw")
		self._setStr(self.ustsuw,      section, "ustsuw")
		self._setStr(self.usxsuw,      section, "usxsuw")
		self._setStr(self.usrsuw,      section, "usrsuw")
		self._setStr(self.usysuw,      section, "usysuw")
		self._setStr(self.detsuw,      section, "detsuw")
		self._setStr(self.usbmax,      section, "usbmax")
		self._setStr(self.gplevbin,    section, "gplevbin")
		self._setStr(self.usbrea,      section, "usbrea")
		self._setBool(self.warnmultmat,section, "warnmultmat")
		self._setBool(self.flukafiles, section, "showflukafiles")

		# -------------------------------------------------------------
		section = MoiraConfig.SECTION
		self._setStr(self.moiradir,    section, "directory")
		self._setStr(self.moiraprg,    section, "program")
		self._setStr(self.moiraexe,    section, "executable")

		# -------------------------------------------------------------
		section = McnpConfig.SECTION
		self._setStr(self.mcnpdir,     section, "directory")
		#self._setStr(self.mcnpprg,     section, "program")
		self._setStr(self.mcnpexe,     section, "executable")
		self._setStr(self.mcnpdata,    section, "datapath")

		# -------------------------------------------------------------
		section = PenelopeConfig.SECTION
		self._setStr(self.penelopedir, section, "directory")
		self._setStr(self.penelopeprg, section, "program")

		# -------------------------------------------------------------
		section = PhitsConfig.SECTION
		self._setStr(self.phitsdir,    section, "directory")
		self._setStr(self.phitsprg,    section, "program")
		self._setStr(self.phitsdata,   section, "datapath")

		# -------------------------------------------------------------
		section = tkFlair._INPUT_SECTION
		self._setStr(self.cardstimeout,section, "cardstimeout")
		self._setBool(self.hrlines,    section, "hrlines")

		# Populate gnuplot commands
		section = tkFlair._PLOT_SECTION
		for k,v in self.config.items(section):
			if k[0] == ".": self.gnuplotFileTypes.insert(tk.END, (k,v))
		self.gnuplotFileTypes.sort(0)
		self.gnuplotCommands.insert("0.0",
			self.config.get(section, "commands"))
		self._setStr(self.gnuplotTerminal,    section, "terminal")
		self._setStr(self.gnuplot,    section, "program")

		# Populate Data
		section = tkFlair._DATA_SECTION
		for k,v in self.config.items(section):
			if "." in k:
				if k.startswith("arg.") or \
				   k.startswith("inp.") or \
				   k.startswith("cmd.") or \
				   k.startswith("icon.") or \
				   k.startswith("rule."): continue
			self.usrTypes.insert(tk.END, (k, v))
		self.usrTypes.sort(0)

		for op,r in Project.UsrInfo.RULES:
			self.usrRules.insert(tk.END,f"{op}{r}")

		#Add custom batch system script
		section = tkFlair._BATCH_SECTION
		self._setStr(self.spawnRunName,    section, "spawnname")
		for k,v in self.config.items(section):
			if k=="spawnname": continue
			self.queueTypes.insert(tk.END, (k,v))
		self.queueTypes.sort(0)

		# Geometry Section
		section = tkFlair._GEOMETRY_SECTION
		self._setBool(self.laptopMode,  section, "laptop")
		self._setBool(self.invertWheel, section, "invertwheel")
		self._setBool(self.pinZoom,     section, "pinzoom")
		self._setBool(self.useQUA,      section, "usequa")
		self._setBool(self.useBOX,      section, "usebox")
		self.cores.set(self.config.get( section, "cores"))
		self.snap.set(self.config.get(  section, "snap"))
		self._setStr(self.zero,		section, "zero")
		self._setStr(self.infinite,	section, "infinite")
		self._setStr(self.explimit,	section, "explimit")
		self._setInt(self.accuracy,	section, "accuracy")

		# Fonts
		self.fontCurrent = []
		self.fontList.selection_clear(0,tk.END)
		self.fontList.selection_set(0)
		self.fontListChange()

	# --------------------------------------------------------------------
	# Ok button pressed
	# --------------------------------------------------------------------
	def ok(self, event=None):
		self.fontListChange()
		# -------------------------------------------------------------
		# Copy all selected fonts to config
		tkFlair.updateConfigFonts(self.fontDict)

		# -------------------------------------------------------------
		# Copy all selected colors
		section = tkFlair._COLOR_SECTION
		for n,v in self.colorDict.items():
			if v is not None:
				tkFlair.config.setColor(section, n, str(v))
			else:
				self.config.remove_option(section, n)

		# -------------------------------------------------------------
		# Get General options
		section = tkFlair._FLAIR_SECTION
		self.config.set(section, "skipintro",       str(bool(self.skipintro.get())))
		self.config.set(section, "autosave",        self.autosave.get())
		#self.config.set(section, "keytimethreshold",self.keytime.get())
		#self.config.set(section, "balloondelay",    self.balloondelay.get())
		self.config.set(section, "timeformat",      self.timeformat.get())
		self.config.set(section, "manualviewer",    self.manual_viewer.get())
		self.config.set(section, "dpiscale",        self.dpiscale.get())

		# -------------------------------------------------------------
		section = tkFlair._PROJECT_SECTION
		self.config.set(section, "keepbackup",      str(bool(self.keepbackup.get())))
		self.config.set(section, "cleanup",         str(bool(self.cleanup.get())))
		self.config.set(section, "kill",            self.killCmd.get())
		self.config.set(section, "editor",          self.editor.get())
		self.config.set(section, "terminal",        self.terminal.get())
		self.config.set(section, "debugger",        self.debugger.get())
		self.config.set(section, "tmpprefix",       self.tmpprefix.get())
		self.config.set(section, "timethreshold",   self.timethreshold.get())
		self.config.set(section, "refreshinterval", self.refreshinterval.get())

		# -------------------------------------------------------------
		section = FlukaConfig.SECTION
		self.config.set(section, "directory",       self.flukadir.get())
		self.config.set(section, "program",         self.rfluka.get())
		self.config.set(section, "executable",      self.flukaexe.get())
		self.config.set(section, "fff",             self.fff.get())
		self.config.set(section, "usbsuw",          self.usbsuw.get())
		self.config.set(section, "ustsuw",          self.ustsuw.get())
		self.config.set(section, "usxsuw",          self.usxsuw.get())
		self.config.set(section, "usrsuw",          self.usrsuw.get())
		self.config.set(section, "usysuw",          self.usysuw.get())
		self.config.set(section, "detsuw",          self.detsuw.get())
		self.config.set(section, "usbmax",          self.usbmax.get())
		self.config.set(section, "gplevbin",        self.gplevbin.get())
		self.config.set(section, "usbrea",          self.usbrea.get())
		self.config.set(section, "warnmultmat",     str(bool(self.warnmultmat.get())))
		self.config.set(section, "showflukafiles",  str(bool(self.flukafiles.get())))

		# -------------------------------------------------------------
		section = MoiraConfig.SECTION
		self.config.set(section, "directory",       self.moiradir.get())
		self.config.set(section, "program",         self.moiraprg.get())
		self.config.set(section, "executable",      self.moiraexe.get())

		# -------------------------------------------------------------
		section = McnpConfig.SECTION
		self.config.set(section, "directory",       self.mcnpdir.get())
		#self.config.set(section, "program",         self.mcnpprg.get())
		self.config.set(section, "executable",      self.mcnpexe.get())
		self.config.set(section, "datapath",        self.mcnpdata.get())

		# -------------------------------------------------------------
		section = PenelopeConfig.SECTION
		self.config.set(section, "directory",       self.penelopedir.get())
		self.config.set(section, "program",         self.penelopeprg.get())

		# -------------------------------------------------------------
		section = PhitsConfig.SECTION
		self.config.set(section, "directory",       self.phitsdir.get())
		self.config.set(section, "program",         self.phitsprg.get())
		self.config.set(section, "datapath",        self.phitsdata.get())

		# -------------------------------------------------------------
		section = tkFlair._INPUT_SECTION
		self.config.set(section, "cardstimeout",   self.cardstimeout.get())
		self.config.set(section, "hrlines",        str(bool(self.hrlines.get())))

		# Gnuplot
		section = tkFlair._PLOT_SECTION
		self.config.remove_section(section)
		self.config.add_section(section)
		for name,value in self.gnuplotFileTypes.get(0,tk.END):
			self.config.set(section, name, value)

		self.config.set(section, "commands",
			self.gnuplotCommands.get("1.0",tk.END))
		self.config.set(section, "terminal", self.gnuplotTerminal.get())
		self.config.set(section, "program",  self.gnuplot.get())

		# Batch
		section = tkFlair._BATCH_SECTION
		self.config.remove_section(section)
		self.config.add_section(section)
		for name,value in self.queueTypes.get(0,tk.END):
			self.config.set(section, name, value)
		self.config.set(section, "spawnname", self.spawnRunName.get())

		# Data processing
		section = tkFlair._DATA_SECTION
		self.config.remove_section(section)
		self.config.add_section(section)
		for name,value in self.usrTypes.get(0,tk.END):
			self.config.set(section, name, value)

		for i,rule in enumerate(self.usrRules.get(0,tk.END),1):
			self.config.set(section, f"rule.{i}", rule)

		# Geometry Section
		section = tkFlair._GEOMETRY_SECTION
		self.config.set(section, "laptop",	str(bool(self.laptopMode.get())))
		self.config.set(section, "invertwheel",	str(bool(self.invertWheel.get())))
		self.config.set(section, "pinzoom",	str(bool(self.pinZoom.get())))
		self.config.set(section, "usequa",	str(bool(self.useQUA.get())))
		self.config.set(section, "usebox",	str(bool(self.useBOX.get())))
		self.config.set(section, "cores",	self.cores.get())
		self.config.set(section, "snap",	self.snap.get())
		self.config.set(section, "zero",	self.zero.get())
		self.config.set(section, "infinite",	self.infinite.get())
		self.config.set(section, "explimit",	self.explimit.get())
		self.config.set(section, "accuracy",	self.accuracy.get())

		# Destroy the dialog and exit
		self.rc = True
		self.close()

	# --------------------------------------------------------------------
	def cancel(self, event=None):
		self.rc = False
		self.close()

	# --------------------------------------------------------------------
	def close(self):
		ConfigDialog.width  = self.winfo_width()
		ConfigDialog.height = self.winfo_height()
		self.destroy()

	# --------------------------------------------------------------------
	def help(self, event=None):
		Manual.show("","Preferences")
