#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

__author__  = "Vasilis Vlachoudis"
__email__   = "Vasilis.Vlachoudis@cern.ch"

import os
import time
import select
import subprocess

import log
import tkFlair
from process import Process

#===============================================================================
# Data process
#===============================================================================
class DataProcess(Process):
	#----------------------------------------------------------------------
	# Cleaning files in synchronous way
	# @return rc (and output filled)
	#----------------------------------------------------------------------
	def clean(self, usr):
		super().clean()
		rc    = 0
		count = 0
		for u in usr:
			fn,ext = os.path.splitext(u.name)
			files = [fn+ext, fn+"_tab.lis", fn+"_sum.lis"]
			for f in files:
				# Remove files
				try:
					os.remove(f)
					count += 1
					self.output(f"Removing: {u.project.relpath(f)}\n")
				except OSError as err:
					if err.errno!=2:	# Ignore No such file or directory
						self.output(f"{err}\n")
				except Exception as err:
					self.output(f"{err}\n")
					rc = 1
		self.message = ("Data clean", f"{count} files removed")
		return rc

	#----------------------------------------------------------------------
	# Process all USRxxx data files
	#----------------------------------------------------------------------
	@tkFlair.exception
	def execute(self, usr):
		super().execute()
		files    = []
		errfiles = []
		rc       = 0
		for i,u in enumerate(usr):
			if u is None:
				self.message = ("Error", "Invalid detector. usr=None")
				return 1
			self.message = ("Processing", u.name)
			if self.status == Process.STATUS_KILLED:
				self.output("*** Killed ***\n")
				return rc
			if u is None:
				self.message = ("Warning", "No USRxxx rules found")
				return 1
			if self._processUsrInfo(u):
				rc = 2
				self.output(f"Error processing: {u.name}\n")
				errfiles.append(u.project.relpath(u.name))
			else:
				files.append(u.project.relpath(u.name))
			# Update percentage
			self.percent = (i*100) // len(usr)

		if rc:
			self.message = ("Warning",
				f"Possible errors: {' '.join(errfiles)}\n"
				f"Correctly processed: {' '.join(files)}")
		elif files:
			self.message = ("Data Merging",
				f"Files created: {' '.join(files)}")
		else:
			self.message = ("Data Merging", "No files were created")
		return rc

	# --------------------------------------------------------------------
	def _processUsrInfo(self, usr):
		startTime = time.time()-2.0
		# command, input
		cmd,inp = usr.command()
		if cmd is None:
			self.output(f"\nWARNING: No input data found for {usr.name!r}\n")
			return True

		scmd = " ".join(cmd)
		self.output(f"\n*-* Processing: {usr.name}\n")
		self.output(f"Cmd: {scmd}\n")

		# start execution
		try:
			pipe = subprocess.Popen(cmd,
					stdin =subprocess.PIPE,
					stdout=subprocess.PIPE,
					stderr=subprocess.STDOUT,
					close_fds=True)
			for x in inp:
				ready = select.select([pipe.stdout], [], [], 0)
				while ready[0]:
					# Use the os.read instead of the pout.readline
					# to avoid blocking situations
					# readline waits for the \n or EOF
					data = os.read(pipe.stdout.fileno(), 1024)
					if not data: break
					self.output(data)
					ready = select.select([pipe.stdout], [], [], 0)

				self.output(f"<<< {x}\n")
				pipe.stdin.write((f"{x}\n").encode())
				pipe.stdin.flush()
			pipe.stdin.close()

			self.output("\n")
			while True:
				line = pipe.stdout.readline()
				if not line: break
				self.output(line)

		except (IOError, OSError):
			self.output(f"{log.ERROR} executing command {scmd}\n")
			self.message = ("Error", f"{log.ERROR} executing command {scmd}")
			return True

		# Assert that generated file is newer than inp
		err = usr.verify(startTime)
		if err:
			self.output(f"{log.ERROR} {err}\n\tMaybe the run is still going on?\n")
		return bool(err)
