#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import time
import tkinter as tk
from tkinter import ttk

import log
import icons
import Ribbon
import tkFlair
import tkExtra
import bFileDialog

import Manual
import Project
import DataProcess
import FlairRibbon

#===============================================================================
# Commands Menu
#===============================================================================
class _CommandsMenu(FlairRibbon.FlairMenu):
	#----------------------------------------------------------------------
	def createMenu(self):
		menu = tk.Menu(self, tearoff=0, activebackground=Ribbon._ACTIVE_COLOR)
		for cmd in Project.UsrInfo.TYPES:
			try:
				icon = icons.get(cmd.upper()[:8])
			except:
				icon = icons.get("empty")
			menu.add_command(label=cmd,
				compound=tk.LEFT, image=icon,
				command=lambda t=cmd,p=self.page:p.addUsr(t))

		if not Project.Command.commands: return menu

		menu.add_separator()
		for cmd in sorted([x for x in Project.Command.commands
						if isinstance(x,str)]):
			try:
				icon = icons.get(Project.Command.commands[cmd].icon)
			except:
				icon = icons.get("empty")
			menu.add_command(label=cmd,
				compound=tk.LEFT, image=icon,
				command=lambda t=cmd,p=self.page:p.addUsr(t))

		return menu

#===============================================================================
# Data Tab in Page Frame
#===============================================================================
class DataTab(FlairRibbon.FlairTab):
	#----------------------------------------------------------------------
	def __init__(self, master, page, **kw):
		super().__init__(master, page, "Data", **kw)

		# --- Run-Detectors splitter ---
		splitter = tk.PanedWindow(self, orient=tk.VERTICAL,
			showhandle=0,
			sashrelief=tk.RAISED, sashwidth=4, opaqueresize=1)
		splitter.pack(expand=tk.YES, fill=tk.BOTH)

		# --- Detectors ---
		frame = tk.LabelFrame(splitter, text="Detectors", foreground="DarkBlue")
		splitter.add(frame, minsize=50)

		self.usrList = tkExtra.MultiListbox(frame,
				(("Run",      32, None),
				 ("Type",      8, None),
				 ("Output",   32, None),
				 ("Name/Unit",16, None)),
				 background="White",
				 height=10)
		self.usrList.pack(expand=tk.YES, fill=tk.BOTH)
		#self.usrList.sortAssist = None

		self.usrList.bind('<<ListboxSelect>>',		self.fillListboxes)
		self.usrList.bindList('<<Rename>>',		self.usrInPlaceEdit)
		self.usrList.bindList('<Control-Key-e>',	self.usrInPlaceEdit)
		self.usrList.bindList('<Return>',		self.usrInPlaceEdit)
		self.usrList.bindList('<KP_Enter>',		self.usrInPlaceEdit)
		self.usrList.bindList('<ButtonRelease-1>',	self.usrInPlaceEdit)
		self.usrList.bindList('<Delete>',		self.deleteUsr)
		self.usrList.listbox(3).bind('<Button-1>',	self.button1Ignore)
		self.usrList.bindList('<Control-Key-d>',	self.cloneUsr)
		self.usrList.bindList('<Control-Key-D>',	self.cloneUsr)

		self.usrList.setPopupMenu(
				[('Add',     0, self.addUsr,    icons.get("add")),
				 ('Remove', -1, self.deleteUsr, icons.get("x")),
				 ('Scan',    0, self.filter,    icons.get("find")) ])

		# --- Frame with Multilistbox ---
		notebook = ttk.Notebook(splitter)
		splitter.add(notebook)

		# --- Files ---
		self.fileList = tkExtra.ColorMultiListbox(notebook,
				(("File", 30, None),
				 ("Type", 10, None),
				 ("Size",  8, None),
				 ("Date", 15, None)),
				 background="White")
		self.fileList.setPopupMenu(
				[('Add',     0, self.add,     icons.get("add")),
				 ('Remove', -1, self.delete,  icons.get("x")),
				 ('Filter',  0, self.filter,  icons.get("filter")),
				 ('Refresh', 0, self.refresh, icons.get("refresh"))])
		self.fileList.bindList('<Delete>',        self.delete)
		notebook.add(self.fileList,
				text     = "Files",
				image    = icons.get("file"),
				compound = tk.LEFT)

		# --- Parameters ---
		self.inputList = tkExtra.MultiListbox(notebook,
				(("Name",   24, None),
				 ("Tag",    10, None),
				 ("Value",  32, None),
				 ("Type",    8, None),
				 ("Default",10, None)),
				 background="White")
		self.inputList.sortAssist = None
		self.inputList.listbox(2).bind('<<Rename>>',		self.inputInPlaceEdit)
		self.inputList.listbox(2).bind('<Control-Key-e>',	self.inputInPlaceEdit)
		self.inputList.listbox(2).bind('<Return>',		self.inputInPlaceEdit)
		self.inputList.listbox(2).bind('<KP_Enter>',		self.inputInPlaceEdit)
		self.inputList.listbox(2).bind('<ButtonRelease-1>',	self.inputInPlaceEdit)
		notebook.add(self.inputList,
				text     = "Parameters",
				image    = icons.get("gear"),
				compound = tk.LEFT)

		self.process = DataProcess.DataProcess()

	# ----------------------------------------------------------------------
	# create the ribbon buttons for the Run tab
	# ----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		group = Ribbon.LabelGroup(self.ribbon, "Command")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.label["background"] = Ribbon._BACKGROUND_GROUP4

		group.grid3rows()

		# ---
		col,row = 0,0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("scan32"),
				text="Scan",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.scan)
		b.grid(row=row, rowspan=3, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Scan input for detectors")

		# ---
		col,row = 1,0
		b = _CommandsMenu(group.frame, self,
				text="New \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("add"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add new command")

		# ---
		col,row = 1,1
		b = Ribbon.LabelButton(group.frame,
				text="Clone",
				image=icons.get("clone"),
				compound=tk.LEFT,
				command=self.cloneUsr)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Clone selected command")

		# ---
		col,row = 1,2
		b = Ribbon.LabelButton(group.frame,
				text="Remove",
				image=icons.get("x"),
				command=self.deleteUsr)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Remove command")

		# ========== Data/Files ===========
		group = Ribbon.LabelGroup(self.ribbon, "Files")
		group.label["background"] = Ribbon._BACKGROUND_GROUP4
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		group.grid3rows()

		# ---
		col,row = 0,0
		b = Ribbon.LabelButton(group.frame,
				text="Refresh",
				image=icons.get("refresh32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.refresh)
		b.grid(row=row, rowspan=3, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Refresh file")

		# ---
		col,row = 1,0
		b = Ribbon.LabelButton(group.frame,
				text="Add",
				image=icons.get("add"),
				command=self.add)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add detector")

		# FIXME SHOULD  we ask for the UNIT?

		# ---
		col,row = 1,1
		b = Ribbon.LabelButton(group.frame,
				text="Remove",
				image=icons.get("x"),
				command=self.delete)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Remove detector")

		# ---
		col,row = 1,2
		b = Ribbon.LabelButton(group.frame,
				text="Filter",
				image=icons.get("filter"),
				command=self.filter)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Open the file search filter/rules dialog")

		# ======= Data =========
		group = Ribbon.LabelGroup(self.ribbon, "Action")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.label["background"] = Ribbon._BACKGROUND_GROUP4

		group.frame.grid_rowconfigure(0, weight=1)

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("clean32"),
				text="Clean",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.clean)
		b.pack(side=tk.LEFT, expand=tk.YES, fill=tk.Y)
		tkExtra.Balloon.set(b, "Remove all generated files")

		# ---
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("compress32"),
				text="Process",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.execute)
		b.pack(side=tk.LEFT, expand=tk.YES, fill=tk.Y)
		tkExtra.Balloon.set(b, "Merged data files [Ctrl-Enter]")

		return self.ribbon

	# ----------------------------------------------------------------------
	def activate(self):
		self.fillUsrList()
		self.selectAllUsr()
		self.fillListboxes()
		self.refreshStatus()

	# --------------------------------------------------------------------
	# Handle global refresh message
	# --------------------------------------------------------------------
	def refresh(self):
		# Restore selections
		self.page.runList.refreshRuns()
		self.fillUsrList()
		self.fillListboxes()

	# --------------------------------------------------------------------
	def scan(self, all_=False):
		# XXX FIXME XXX when finding ASCII files issue a warning message
		for run in self.page.runList.selectedRuns():
			self.project.findRunUsrInfo(run)
		self.fillUsrList()
		self.selectAllUsr()
		self.fillListboxes()

	# --------------------------------------------------------------------
	# Fill USRxxx list
	# --------------------------------------------------------------------
	def fillUsrList(self, event=None):
		self.log = self.flair.newLog("Data", "Process")
		oldSize = self.log.size()	# remember size before loading
		self.page.busy()
		savesort = self.usrList.saveSort()
		self.usrList.delete(0,tk.END)
		for run in self.page.runList.selectedRuns():
			if not run.usrinfo:
				self.project.findRunUsrInfo(run)
			runname = run.inputName2()
			parent  = run("parent")
			if parent:
				runname = f"{parent}:{runname}"
			for usr in run.usrinfo:
				usrname = self.project.relpath(usr.name)
				self.usrList.insert(tk.END,
					(runname, usr.type, usrname, usr.unit))
		self.usrList.restoreSort(savesort)
		self.page.notBusy()
		if self.log.size() != oldSize:
			self.flair.notify("Errors during file scanning",
				"Error or warning messages appeared while USRxxx file scanning. "
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR,
				"Output", self.flair.showOutput,
				showOnOutput=False)

	# --------------------------------------------------------------------
	def selectAllUsr(self):
		self.usrList.selection_set(0, tk.END)
		self.usrList.activate(0)
		self.usrList.see(0)

	# --------------------------------------------------------------------
	# Find UsrInfo for a specific usrname from the selected ones
	# --------------------------------------------------------------------
	def findUsrInfo(self, usrname, runname=None):
		if runname is not None:
			if ":" in runname:
				parent, runname = runname.split(":")
			else:
				parent = ""
			run = self.project.runs.getRunByName(runname, parent)
			for usr in run.usrinfo:
				if usrname == self.project.relpath(usr.name):
					return usr
		else:
			for run in self.page.runList.selectedRuns():
				for usr in run.usrinfo:
					if usrname == self.project.relpath(usr.name):
						return usr
		return None

	# --------------------------------------------------------------------
	# Return selected UsrInfo list
	# --------------------------------------------------------------------
	def getUsr(self):
		usr = []
		for u in self.usrList.curselection():
			runname = self.usrList.listbox(0).get(u)
			usrname = self.usrList.listbox(2).get(u)
			u = self.findUsrInfo(usrname, runname)
			if u is not None: usr.append(u)
		return usr

	# --------------------------------------------------------------------
	# Return selected UsrInfo
	# --------------------------------------------------------------------
	def selectedUsr(self):
		sel = self.usrList.curselection()
		if len(sel) != 1: return None
		return self.findUsrInfo(self.usrList.listbox(2).get(sel[0]),
				self.usrList.listbox(0).get(sel[0]))

	# --------------------------------------------------------------------
	# Fill files in listbox
	# --------------------------------------------------------------------
	def fillListboxes(self, event=None):
		self.page.busy()
		savesort = self.fileList.saveSort()

		# Empty listboxes
		self.inputList.delete(0,tk.END)
		self.fileList.delete(0,tk.END)

		# Check if all usr selected are the same
		# FIXME NEEDS CHECKING FOR MULTIPLE SELECTION
		t = None
		for usr in self.getUsr():
			if t is None: t = usr.type
			if t != usr.type:
				t = None
				break

		# Fill filelist
		self.files = []
		repeat = None
		for usr in self.getUsr():
			self.files.extend(usr.depends())
			if t is not None and usr.cmd is not None:
				for i,(ctag,ctype,cdef,cname) in enumerate(usr.cmd.arg):
					try:
						u = usr.arg[i]
					except:
						u = ""
					if not u:
						if ctype=="output":
							u = usr.name
						elif cdef:
							u = usr.expand(cdef)
					self.inputList.insert(tk.END,(cname,ctag,u,ctype,cdef))
				i = 0	# rule index
				j = 0	# input index
				while i < len(usr.cmd.inp):
					ctype,cdef,cname = usr.cmd.inp[i]
					i += 1
					try:
						u = usr.inp[j]
						j += 1
					except:
						u = ""
					if not u:
						if ctype=="output":
							u = usr.name
						elif cdef:
							u = usr.expand(cdef)
					if ctype=="repeat":
						repeat = int(cname)
						if u != cdef:
							# correct pointers and repeat sequence
							i  = repeat-1
							j -= 1
							continue
						ctype,cdef,cname = usr.cmd.inp[int(cname)-1]
						cname += " [repeat?]"
						ctype = f"{ctype}:{repeat}"
						repeat = None
					if repeat is not None:
						ctype = f"{ctype}:{i}"
					self.inputList.insert(tk.END,(cname,"",u,ctype,cdef))
				t = None

		self.files.sort()
		self.updateFiles()
		self.fileList.restoreSort(savesort)
		self.page.notBusy()

	# --------------------------------------------------------------------
	# Update file list
	# --------------------------------------------------------------------
	def updateFiles(self):
		# Populate list
		self.fileList.delete(0,tk.END)
		totalSize = 0
		for fn in self.files:
			try:
				s  = os.stat(self.project.abspath(fn))
				sz = s.st_size
				ti = s.st_mtime
			except OSError:
				self.log(f"{log.WARNING} File not found: {fn}")
				continue
			totalSize += sz
			ext,color = bFileDialog.fileTypeColor(fn)
			self.fileList.insert(tk.END,
				(fn, ext, sz, time.strftime(bFileDialog._TIME_FORMAT,
					time.localtime(ti))))
			if color is not None:
				self.fileList.itemconfig(tk.END,foreground=color)
		if self.flair is None: return

	# --------------------------------------------------------------------
	def refreshStatus(self):
		if self.process.isAlive():
			self.page.setStatus("%2d%%  %s: %s" %
				(self.process.percent,
				 self.process.message[0],
				 self.process.message[1]), "DarkGreen")
			self.page.setStatusPercent(self.process.percent)
		else:
			self.page.resetPercent()
			self.page.setStatus(f"Files: {len(self.files)}")

	# --------------------------------------------------------------------
	# Add a new file
	# --------------------------------------------------------------------
	def add(self):
		usr = self.selectedUsr()
		if usr is None:
			self.flair.notify("Please select one usrxxx",
				"Select only one detector to add a file(s) to it",
				tkFlair.NOTIFY_WARNING)
			return

		inpname = usr.run.inputBaseName()
		files = bFileDialog.askopenfilenames(master=self,
				 title='Add Files',
				 initialdir=usr.run.getRunDir(),
				 filetypes=[
					("Fluka Run fort files",
						(f"{inpname}*_fort.{usr.unit}",
						 f"{inpname}*_ftn.{usr.unit}")),
					(f"Fluka unit {usr.unit} fort files",
						(f"*_fort.{usr.unit}",
						 f"*_ftn.{usr.unit}")),
					("Fluka fort files","*_fort.*"),
					("Fluka ftn files","*_ftn.*"),
					("All","*")])
		if files:
			for f in files:
				fn = self.project.relpath(f)
				usr._doRule(self.files, '+', "", fn)
				usr.rules.append(('+',fn))
				usr.auto = False
				self.page.modified()
			self.updateFiles()

	# --------------------------------------------------------------------
	# Delete a file from list
	# --------------------------------------------------------------------
	def delete(self, event=None):
		usr = self.selectedUsr()
		if usr is None:
			self.flair.notify("Please select one usrxxx",
				"Select only one detector to delete a file(s) from it",
				tkFlair.NOTIFY_WARNING)
			return
		cursel = list(map(int, self.fileList.curselection()))

		# Delete all?
		if len(cursel) == self.fileList.size():
			usr.rules  = []
			self.files = []
		else:
			# Delete selected files
			for i in cursel:
				fn = self.fileList.get(i)[0]
				# search if + rule already exists then remove it
				for j,(o,f) in enumerate(usr.rules):
					if f==fn:
						if o=="+": del usr.rules[j]
						break
				else:
					usr.rules.append(("-",fn))
				usr._doRule(self.files, '-', "", fn)
				usr.auto = False
		self.updateFiles()

	# --------------------------------------------------------------------
	def filter(self):
		usr = self.getUsr()
		dlg = FilterDialog(self.master)
		rules = dlg.show(usr)
		if rules is not None:
			# Set rules for all runs
			for u in usr:
				if u is not None:
					u.auto  = False
					u.rules = rules[:]
		self.fillListboxes()
		self.page.modified()

	# --------------------------------------------------------------------
	# Set USRxxx
	# --------------------------------------------------------------------
	def addUsr(self, cmd="usrbin"):
		# do for all runs
		added = []
		for run in self.page.runList.selectedRuns():
			# do for all selected usr
			#for u in self.getUsr():
			usr = Project.UsrInfo(run)
			usr.auto = False
			usr.set(type_=cmd)
			if usr.cmd:
				usr.name = usr.cmd.defaultName()
			run.usrinfo.append(usr)
			added.append((run.inputName2(), usr.type, usr.name))

		# Update display
		self.fillUsrList()
		self.usrList.selection_clear(0,tk.END)

		see = tk.END
		for r,t,n in added:
			for i,row in enumerate(self.usrList.get(0,tk.END)):
				if row[0]==r and row[1]==t and row[2]==n:
					if see == tk.END: see = i
					self.usrList.selection_set(i)
					self.usrList.activate(i)
		self.fillListboxes()
		self.usrList.see(see)
		self.page.modified()

	# --------------------------------------------------------------------
	def cloneUsr(self, event=None):
		sel = list(map(int, self.usrList.curselection()))
		if not sel: return
		for u in sel:
			usrname = self.usrList.listbox(2).get(u)
			usr = self.findUsrInfo(usrname)
			new = usr.clone()
			usr.run.usrinfo.append(new)
			new.run.findUniqueUsrName(new)
		self.fillUsrList()
		self.page.modified()
		self.fillListboxes()
		return "break"

	# --------------------------------------------------------------------
	def deleteUsr(self, event=None):
		sel = list(map(int, self.usrList.curselection()))
		if not sel: return
		if not tkFlair.askyesno("Delete USRxxx files",
				f"You are about to delete {len(sel)} "
				f"USRxxx summary data files. "
				"Do you want to continue?",
				parent=self.ribbon):
			return

		sel.reverse()
		for u in sel:
			usrname = self.usrList.listbox(2).get(u)
			usr = self.findUsrInfo(usrname)
			usr.run.usrinfo.remove(usr)
		self.fillUsrList()
		self.page.modified()
		self.fillListboxes()

	# --------------------------------------------------------------------
	# Button1 hit on selected item, ignore then action so not to unselect
	# --------------------------------------------------------------------
	def button1Ignore(self, event):
		act = event.widget.nearest(event.y)
		if event.widget.selection_includes(act):
			event.widget.focus_set()
			event.widget.activate(act)
			return "break"

	# --------------------------------------------------------------------
	# USR list edit in Place
	# --------------------------------------------------------------------
	def usrInPlaceEdit(self, event=None):
		listbox = event.widget
		if event.type == "2":		# keyboard event
			active = listbox.index(tk.ACTIVE)
		else:
			active = listbox.nearest(event.y)
			self.usrList.activate(active)
		lid = self.usrList.listboxIndex(listbox)
		sel = self.usrList.curselection()

		cmd = self.usrList.listbox(1).get(active)
		ypos = listbox.yview()[0]	# remember y position

		changed = 0
		if lid == 1:
			lst = Project.UsrInfo.TYPES[:]
			for cmd in sorted([x
					for x in Project.Command.commands
						if isinstance(x,str)]):
				lst.append(cmd)
			# Change the usrxxx type
			edit = tkExtra.InPlaceList(listbox, values=lst, height=8)
			if edit.value is None: return
			for usr in self.getUsr():
				if usr.type != edit.value:
					usr.auto = False
					try:
						usr.type = edit.value
						changed += 1
						self.page.modified()
					except:
						pass

		elif lid == 2:
			# Edit the filename
			usrs = self.getUsr()	# before the name change
						# after the name change the getUsr will fail
						# to find the usr
			edit = tkExtra.InPlaceFile(listbox, title="Rename usrxxx")
			if edit.value is None: return
			for usr in usrs:
				if usr.type != cmd: continue
				fn = self.project.relpath(edit.value, usr.run)
				usr.auto = False
				usr.name = fn
				changed += 1
				self.page.modified()

		elif lid == 3:
			edit = tkExtra.InPlaceSpinbox(listbox, from_=21, to_=99)
			if edit.value is None: return
			try: edit.value = int(edit.value)
			except: pass
			changed = 2	# Force filling of UsrList below
			for usr in self.getUsr():
				if usr.type != cmd: continue
				if usr.unit != edit.value:
					usr.auto = False
					usr.unit = edit.value
					row = self.usrList.get(active)
					row[1] = usr.type
					row[2] = self.project.relpath(usr.name)
					changed += 1
					self.page.modified()

		if changed>1: self.fillUsrList()
		for s in sel: self.usrList.selection_set(s)
		self.usrList.selection_set(active)
		self.usrList.activate(active)
		self.usrList.yview_moveto(ypos)
		self.fillListboxes()

	# --------------------------------------------------------------------
	# Input list edit in Place
	# --------------------------------------------------------------------
	def inputInPlaceEdit(self, event=None):
		listbox = event.widget
		if event.type == "2":		# keyboard event
			active = listbox.index(tk.ACTIVE)
		else:
			active = listbox.nearest(event.y)
			self.inputList.activate(active)
		#lid = self.inputList.listboxIndex(listbox)

		usr = self.selectedUsr()
		if usr is None or len(usr.type)==1: return
		ypos = listbox.yview()[0]	# remember y position
		oldname = usr.name

		# depends on type
		name,tag,value,vartype,default = self.inputList.get(tk.ACTIVE)
		if ":" in vartype:
			vartype,repeat = vartype.split(":")
			repeat = int(repeat)-1
		else:
			repeat = None

		if vartype == "int":
			edit = tkExtra.InPlaceInteger(listbox)
		elif vartype == "float":
			edit = tkExtra.InPlaceFloat(listbox)
		elif vartype == "output":
			edit = tkExtra.InPlaceFile(listbox, title=name, save=True,
				filetypes=[("All","*"),
					("Histogram",  "*.hist"),
					("Fluka fort", "*_fort.*"),
					("Tab lis",    "*_tab.lis"),
					("Voxel",      "*.vxl"),
					("Data",       "*.dat"),
					("Usrbin",     "*.bnn")])
		elif vartype == "file":
			edit = tkExtra.InPlaceFile(listbox, title=name, save=False,
				filetypes=[("All","*"),
					("Histogram",  "*.hist"),
					("Fluka fort", "*_fort.*"),
					("Tab lis",    "*_tab.lis"),
					("Voxel",      "*.vxl"),
					("Data",       "*.dat"),
					("Usrbin",     "*.bnn")])
		elif vartype == "filelist":
			edit = tkExtra.InPlaceEdit(listbox)
		elif "," in vartype:
			choices = [""]
			choices.extend(vartype.split(","))
			edit = tkExtra.InPlaceList(listbox, values=choices)
		elif vartype == "fixed":
			# Do nothing
			return
		else:
			edit = tkExtra.InPlaceEdit(listbox)

		if edit.value is None: return

		if vartype in ("output","file"): #,"filelist")
			edit.value = self.project.relpath(edit.value)

		# Argument?
		if active < len(usr.cmd.arg):
			idx = active
			if not edit.value:
				ctag,ctype,cdef,cname = usr.cmd.arg[idx]
				edit.value = usr.expand(cdef)
			usr.setArg(idx, edit.value)
			usr.auto = False

		# or Input?
		else:
			idx = active-len(usr.cmd.arg)
			modified = False
			# Empty
			if not edit.value:
				# if before was not empty and repeat is active
				if value and repeat is not None:
					# delete all repeated terms
					usr.deleteRepeat(active, repeat)
					modified = True

				if repeat is not None:
					ctype,cdef,cname = usr.cmd.inp[repeat]
				else:
					ctype,cdef,cname = usr.cmd.inp[idx]
				edit.value = usr.expand(cdef)

			else:   # Or has a new value
				# if before was empty and repeat is active
				if not value and repeat is not None:
					# insert repeated terms
					usr.insertRepeat(active,repeat)
					modified = True
			usr.setInp(idx, edit.value)
			usr.auto = False

			if modified: self.fillListboxes()

		if vartype == "output":
			usr.name = edit.value
			usr.auto = False

		listbox.set(active, edit.value)

		if usr.name != oldname:
			r = usr.run.inputName2()
			t = usr.type
			n = usr.name

			# Name has changed re-fill UsrList and select the usr with the new name
			self.fillUsrList()
			for i,row in enumerate(self.usrList.get(0,tk.END)):
				if row[0]==r and row[1]==t and row[2]==n:
					self.usrList.selection_set(i)
					self.usrList.activate(i)
					self.usrList.see(i)
					break

		self.inputList.selection_set(active)
		self.inputList.activate(active)
		self.inputList.yview_moveto(ypos)

	# --------------------------------------------------------------------
	# Process
	# --------------------------------------------------------------------
	def execute(self):
		if not self.usrList.curselection():
			self.flair.notify("Nothing is selected",
				"Please select the detectors to process and execute again",
				tkFlair.NOTIFY_WARNING)
			return
		if self.process.isAlive():
			if not tkFlair.askyesno("Data process is active",
					"A data merging process is still running.\n"
					"Do you want to kill it?",
					parent=self.ribbon):
				return
			self.process.kill()
			self.process.wait()

		self.log = self.flair.newLog("Data", "Process")
		self.process.init()
		self.flair.addProcess(self.process, self.log, self.getUsr())
		self.page.setStatus("Data merging..")

	# ----------------------------------------------------------------------
	# clean files
	# ----------------------------------------------------------------------
	def clean(self):
		if not self.usrList.curselection():
			self.flair.notify("Nothing is selected",
				"Please select the detectors to clean and execute again",
				tkFlair.NOTIFY_WARNING)
			return
		if not tkFlair.askyesno("Warning",
				"Do you want to remove all generated files?",
				parent=self.ribbon):
			return

		if self.process.isAlive():
			if not tkFlair.askyesno("Data process is active",
					"A data merging process is still running.\n"
					"Do you want to kill it?",
					parent=self.ribbon):
				return
			self.process.kill()
			self.process.wait()

		self.log = self.flair.newLog("Data", "Process")
		self.process.clean(self.getUsr())
		self.log.write(self.process.output())
		if self.process.message:
			self.flair.notify(self.process.message[0],
				self.process.message[1])
			self.page.setStatus(self.process.message[1])

	# ----------------------------------------------------------------------
	def help(self):
		Manual.show("","Run:Data")

#=============================================================================
# Rules dialog
#=============================================================================
class FilterDialog(tk.Toplevel):
	def __init__(self, master):
		super().__init__(master)
		self.title("File Selection Rules")
		self.transient(master)

		frame = tk.LabelFrame(self, text="Rules")
		frame.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=5, pady=3)

		# ---
		self.rulesList = tkExtra.ExListbox(frame,
				borderwidth=0,
				selectmode=tk.EXTENDED,
				selectborderwidth=0,
				background="White",
				takefocus=True,
				width=50,
				exportselection=tk.FALSE)
		self.rulesList.grid(row=0, column=0, rowspan=7, sticky=tk.NSEW)
		sb = ttk.Scrollbar(frame, orient=tk.VERTICAL, takefocus=False,
				command=self.rulesList.yview)
		sb.grid(row=0, column=1, rowspan=7, sticky=tk.NSEW)
		self.rulesList.config(yscrollcommand=sb.set)

		# --- Buttons ---
		b = tk.Button(frame, image=icons.get("edit"),
				command=self.edit)
		tkExtra.Balloon.set(b, "Edit Rule")
		b.grid(row=0, column=2, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("add"),
				command=self.add)
		tkExtra.Balloon.set(b, "Add new rule")
		b.grid(row=1, column=2, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("del"),
				command=self.delete)
		tkExtra.Balloon.set(b, "Delete rule(s)")
		b.grid(row=2, column=2, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("clone"),
				command=self.clone)
		tkExtra.Balloon.set(b, "Duplicate rule(s)")
		b.grid(row=3, column=2, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("up"),
				command=self.rulesList.moveUp)
		tkExtra.Balloon.set(b, "Move up")
		b.grid(row=4, column=2, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("down"),
				command=self.rulesList.moveDown)
		tkExtra.Balloon.set(b, "Move down")
		b.grid(row=5, column=2, sticky=tk.EW)

		m = tk.Label(frame,
			text="Syntax: [+-][filename | ^regexpr$]\n"
				"Special Characters:\n"
				" \\I\tInput name\t"
				" \\U\tUnit number\n"
				" \\T\tType (usrtrack...)\t"
				" \\t\tShort type (t,x,...)\n"
				" \\e\tDefault extension\n"
				"  .\tAny character\t"
				"  *\t0 or more char\n"
				"  +\t1 or more char\t"
				"  ?\t0 or 1 match of char\n"
				"  \\d\tDigit\t\t"
				"  \\D\tNon Digit\n"
				"http://docs.python.org/library/re.html",
				foreground="DarkBlue", justify=tk.LEFT,
				relief=tk.GROOVE, anchor=tk.W)
		m.grid(row=7, column=0, columnspan=3, sticky=tk.NSEW)

		frame.grid_rowconfigure(6, weight=1)
		frame.grid_columnconfigure(0, weight=1)

		# --------------
		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, fill=tk.X)

		tk.Button(frame, text="Cancel",
				image=icons.get("x"),
				compound=tk.LEFT,
				command=self.cancel).pack(side=tk.RIGHT)
		tk.Button(frame, text="Ok",
				image=icons.get("ok"),
				compound=tk.LEFT,
				command=self.ok).pack(side=tk.RIGHT)

		# --------------
		self.rulesList.setPopupMenu(
				[('Edit',   0, self.edit,   icons.get("edit")),
				 ('Add',    0, self.add,    icons.get("add")),
				 ('Clone',  0, self.clone,  icons.get("clone")),
				 ('Remove',-1, self.delete, icons.get("x")) ])
		# --------------
		self.rulesList.bind("<Return>",		self.edit)
		self.rulesList.bind("<KP_Enter>",	self.edit)
		self.rulesList.bind("<Double-Button-1>",self.edit)
		self.rulesList.bind("<<Rename>>",	self.edit)
		self.rulesList.bind("<Insert>",		self.add)
		self.rulesList.bind("<Delete>",		self.delete)
		self.rulesList.bind("<<MoveUp>>",	lambda e,s=self: s.rulesList.moveUp())
		self.rulesList.bind("<<MoveDown>>",	lambda e,s=self: s.rulesList.moveDown())
		self.bind("<Escape>", self.cancel)
		self.protocol("WM_DELETE_WINDOW", self.cancel)

	# ----------------------------------------------------------------------
	def show(self, usr):
		self.rules = None
		self.rulesList.delete(0,tk.END)
		rulesstr = {}
		for u in usr:
			if u is None: continue
			rules = u.rules
			for pat,filename in rules:
				rule = pat+filename
				if rule not in rulesstr:
					rulesstr[rule] = True
					self.rulesList.insert(tk.END, rule)
		self.wait_window()
		return self.rules

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		self.rules = []
		for rule in self.rulesList.get(0,tk.END):
			pat = "+"
			filename = rule
			if len(rule)>1:
				pat = rule[0]
				if pat in ("+", "-"):
					filename = rule[1:]
				else:
					pat = "+"
			self.rules.append([pat, filename])

		self.destroy()

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		self.destroy()

	# ----------------------------------------------------------------------
	# Edit rule
	# ----------------------------------------------------------------------
	def edit(self, event=None):
		tkExtra.InPlaceEdit(self.rulesList)

	# ----------------------------------------------------------------------
	# Add new rules
	# ----------------------------------------------------------------------
	def add(self, event=None):
		self.rulesList.insert(tk.END,"")
		self.rulesList.selection_clear(0,tk.END)
		self.rulesList.selection_set(tk.END)
		self.rulesList.activate(tk.END)
		self.edit()
		if self.rulesList.get(tk.ACTIVE).strip() == "":
			self.rulesList.delete(tk.ACTIVE)
			self.rulesList.selection_set(tk.END)
			self.rulesList.activate(tk.END)

	# ----------------------------------------------------------------------
	# Delete rules
	# ----------------------------------------------------------------------
	def delete(self, event=None):
		sel = list(self.rulesList.curselection())
		if not sel: return
		sel.reverse()
		for i in sel:
			self.rulesList.delete(i)

	# ----------------------------------------------------------------------
	# Clone
	# ----------------------------------------------------------------------
	def clone(self, event=None):
		sel = list(map(int, self.rulesList.curselection()))
		self.rulesList.selection_clear(0, tk.END)
		for i in sel:
			rule = self.rulesList.get(i)
			self.rulesList.insert(tk.END, rule)
			self.rulesList.selection_set(tk.END)
		self.rulesList.activate(tk.END)
