# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   14-Aug-2012

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import sys
import math
import time
import struct
import fnmatch

try:
	import numpy as np
except ImportError:
	np = None
from operator import itemgetter
from collections import defaultdict, namedtuple

try:
	from PIL import Image
	from PIL import ImageDraw
except ImportError:
	Image = None

try:
	import pydicom
except ImportError:
	try:
		import dicom as pydicom
	except ImportError:
		pydicom = None

import log
import tool
import bmath
import Global
import fortran
import tkFlair
import tkDialogs
from Project import DicomInfo
from material import Material
from DicomDraw import drawLine
from process import PopenProcess
from fluka import Fluka, FlukaCard, FlukaVoxel

DicomHeader = namedtuple("DicomHeader",
		["uid", "modality", "datetime", "name", "age", "sex", "institution", "study"])
DataSets    = namedtuple("DataSets", ["frames","path","files","modality","location"])

HOUNSFIELD_MIN = -2000
HOUNSFIELD_MAX =  4096

MODALITY = {
	"AS"       : "Angioscopy",
	"AU"       : "Audio",
	"BI"       : "Biomagnetic imaging",
	"CD"       : "Color flow Doppler",
	"CP"       : "Colposcopy",
	"CR"       : "Computed Radiography",
	"CS"       : "Cystoscopy",
	"CT"       : "Computed Tomography",
	"DD"       : "Duplex Doppler",
	"DG"       : "Diaphanography",
	"DM"       : "Digital microscopy",
	"DX"       : "Digital Radiography",
	"EC"       : "Echocardiography",
	"ECG"      : "Electrocardiography",
	"EPS"      : "Cardiac Electrophysiology",
	"ES"       : "Endoscopy",
	"FA"       : "Fluorescein angiography",
	"FS"       : "Fundoscopy",
	"GM"       : "General Microscopy",
	"HC"       : "Hard Copy",
	"HD"       : "Hemodynamic Waveform",
	"IO"       : "Intra-oral Radiography",
	"IVUS"     : "Intravascular Ultrasound",
	"KO"       : "Key Object Selection",
	"LP"       : "Laparoscopy",
	"LS"       : "Laser surface scan",
	"MA"       : "Magnetic resonance angiography",
	"MG"       : "Mammography",
	"MR"       : "Magnetic Resonance",
	"MS"       : "Magnetic resonance spectroscopy",
	"NM"       : "Nuclear Medicine",
	"OCT"      : "Optical Coherence Tomography",
	"OP"       : "Ophthalmic Photography",
	"OPM"      : "Ophthalmic Mapping",
	"OPR"      : "Ophthalmic Refraction",
	"OPV"      : "Ophthalmic Visual Field",
	"OT"       : "Other",
	"PR"       : "Presentation State",
	"PT"       : "Positron emission tomography (PET)",
	"PX"       : "Panoramic X-Ray",
	"REG"      : "Registration",
	"RF"       : "Radio Fluoroscopy",
	"RG"       : "Radiographic imaging (conventional film/screen)",
	"RTDOSE"   : "Radiotherapy Dose",
	"RTIMAGE"  : "Radiotherapy Image",
	"RTPLAN"   : "Radiotherapy Plan",
	"RTRECORD" : "RT Treatment Record",
	"RTSTRUCT" : "Radiotherapy Structure Set",
	"SEG"      : "Segmentation",
	"SM"       : "Slide Microscopy",
	"SMR"      : "Stereometric Relationship",
	"SR"       : "SR Document",
	"ST"       : "Single-photon emission computed tomography (SPECT)",
	"TG"       : "Thermography",
	"US"       : "Ultrasound",
	"XA"       : "X-Ray Angiography",
	"XC"       : "External-camera Photography",
}

# ----------------------------------------------------------------------
# Return the header of a dicom for filling the list
# ----------------------------------------------------------------------
def dicomHeader(uid, filename):
	try:
		dataset = pydicom.read_file(filename,
					force=True,
					stop_before_pixels=True)
	except pydicom.filereader.InvalidDicomError:
		return DicomHeader(uid,"?","?","?","?","?","?","?")

	try:
		modality = dataset.Modality
	except:
		modality = 'dicom'

	try:
		acqdate = dataset.StudyDate
		acqtime = dataset.StudyTime
		datetime = "%s.%s.%s %s:%s"% \
			(acqdate[: 4], acqdate[ 4: 6], acqdate[ 6:],
			 acqtime[:-4], acqtime[-4:-2])
	except AttributeError:
		datetime = "?"

	if "PatientIdentityRemoved" in dataset:
		name = "-patient identity removed-"
	elif "PatientName" in dataset:
		pat_name = dataset.PatientName
		try:
			name = f"{pat_name.family_name}, {pat_name.given_name}"
		except:
			name = "none"
	else:
		name = "none"

	try:
		institution_name = dataset.InstitutionName
	except AttributeError:
		institution_name = "none"

	try:
		study_description = dataset.StudyDescription
	except AttributeError:
		study_description = "none"

	try:
		age = dataset.PatientAge
	except AttributeError:
		age = "?"

	try:
		birth = dataset.PatientBirthDate
	except:
		birth = "?"

	try:
		sex = dataset.PatientSex
	except AttributeError:
		sex = "?"

	return DicomHeader(uid,
			modality,
			datetime,
			name,
			f"{age} [{birth}]",
			sex,
			institution_name,
			study_description)

#===============================================================================
# Scan a directory for dicom sets and cache the results
#===============================================================================
class Scan:
	directory = ""
	depth     = 3
	datasets  = {}

	SKIP = ["*~",			# patterns to skip
		"*.bak",
		"*.flair",
		"*.inp",
		"*.out",
		"*.log",
		"*.err",
		"*.bnn",
		"*.vxl",
		"*.bdx",
		"*_fort.[0-9][0-9]",
		]

	# ----------------------------------------------------------------------
	# Scan directory for unique dicom files
	# ----------------------------------------------------------------------
	@staticmethod
	def scan(path, depth=None, master=None, force=False):
		#Global.dump(f"*-* Scan.scan({path},{depth})")
		apath = os.path.abspath(path)

		if depth is None: depth = Scan.depth
		if not force and apath==Scan.directory and depth==Scan.depth: return
		Scan.depth = depth

		Scan.directory = apath
		Scan.datasets.clear()

		try:
			listdir = os.listdir(apath)
		except OSError:
			log.error(f"Cannot access directory {apath!r}")
			return False

		dlg = None
		if master:
			# BUG python or tk/ttk?
			# sometimes it doesn't pass the ProgressDialog line
			# maybe it generates a refresh event and the selection is set
			# on the treeview screwing up everything?
			try:
				dlg = tkDialogs.ProgressDialog(master,
						f"Scanning directory: {path}")
				dlg.setLimits(0.0, len(listdir))
			except:
				pass
		i = -1
		for root,dirs,files in tool.walk(apath, level=depth):
			for fn in files:
				# skip known no dicom files
				# Files larger than 10Mb sould have a known extension (dcm) or none
				#if os.path.getsize(fn) > 10*1024*1024:
				if fn[-1] == "~": continue	# skip backups
				if any(fnmatch.fnmatch(fn,x) for x in Scan.SKIP): continue
				path = os.path.join(apath, root)
				filename = os.path.join(root, fn)
				if not os.path.isfile(filename): continue
				try:
					dataset = pydicom.read_file(filename,
								force=True,
								stop_before_pixels=True)
				except Exception as ex:
					log.error(f"Cannot open file {filename!r}")
					if Global.developer:
						log.errorCont(ex)
					continue
				try:
					uid = dataset.SeriesInstanceUID
				except AttributeError:
					continue
				try:
					modality = dataset.Modality
				except:
					modality = 'dicom'
				try:
					iop = list(map(float,dataset.ImageOrientationPatient))
				except:
					iop = [1., 0., 0.,  0., 1., 0.]
				rowV    = bmath.Vector(iop[:3])	# row Vector
				columnV = bmath.Vector(iop[3:])	# column Vector
				sliceV  = rowV ^ columnV

				# Sort by instance
				try:
					instance = int(dataset.InstanceNumber)
				except:
					instance = 1

				location = instance
				if modality in ("RTDOSE","RTPLAN"):
					pass
				# Except when the location can be deduced
				elif "ImagePositionPatient" in dataset:
					imagepos = bmath.Vector(map(float,
							dataset.ImagePositionPatient))
					location = imagepos * sliceV
				else:
					try:
						imagepos = bmath.Vector(map(float,
							dataset.PerFrameFunctionalGroupsSequence[0]
								.PlanePositionSequence[0]
								.ImagePositionPatient))
						location = imagepos * sliceV
					except AttributeError:
						pass
				try:
					sop = dataset.SOPInstanceUID
				except:
					sop = None
				try:
					# append file
					ds = Scan.datasets[uid]
					if ds.path != path:
						fn = os.path.join(os.path.relpath(path, ds.path), fn)
					ds.files.append(DicomInfo.FileTuple(fn,instance,sop))
					ds.location.append(location)
				except KeyError:
					try:
						frames = int(dataset.NumberOfFrames)
					except AttributeError:
						frames = 0
					Scan.datasets[uid] = DataSets(
								frames,
								path,
								[DicomInfo.FileTuple(fn,instance,sop)],
								modality,
								[location])

				# Show progress
				if dlg and dlg.show(i,"Scanning: "+filename): break
				del dataset

		# Sort lists by location
		for ds in Scan.datasets.values():
			loc,files = zip(*sorted(zip(ds.location, ds.files)))
			# namedtuple is immutable so we need to replace the lists content
			del ds.files[:]
			del ds.location[:]
			ds.files.extend(files)
			ds.location.extend(loc)

		if dlg: dlg.stop()
		return True

	# ----------------------------------------------------------------------
	@staticmethod
	def rescan(path, master=None):
		Scan.directory = ""
		Scan.scan(path, master)

	# ----------------------------------------------------------------------
	@staticmethod
	def __getitem__(uid):
		return Scan.datasets[uid]

	# ----------------------------------------------------------------------
	@staticmethod
	def headerUID(uid):
		return dicomHeader(uid,
				os.path.join(Scan.path(uid), Scan.files(uid)[0][0]))

	# ----------------------------------------------------------------------
	@staticmethod
	def frames(uid):
		return Scan.datasets[uid].frames

	# ----------------------------------------------------------------------
	@staticmethod
	def path(uid):
		return os.path.join(Scan.directory, Scan.datasets[uid].path)

	# ----------------------------------------------------------------------
	@staticmethod
	def files(uid):
		return Scan.datasets[uid].files

	# ----------------------------------------------------------------------
	@staticmethod
	def modality(uid):
		return Scan.datasets[uid].modality

#===============================================================================
# Hounsfield units class
#===============================================================================
class Hounsfield:
	prev   = -9999

	#-----------------------------------------------------------------------
	def __init__(self, humax, mat, crho_min, crho_max, cdedx_min, cdedx_max):
		self.humin  = Hounsfield.prev
		self.humax  = humax
		Hounsfield.prev = humax
		self.mat    = mat
		self.reg    = 0
		self.crho_min  = crho_min
		self.crho_max  = crho_max
		self.cdedx_min = cdedx_min
		self.cdedx_max = cdedx_max

		if self.humax > self.humin+1:
			self._rho_scale  = (self.crho_max  - self.crho_min)  \
						/ float(self.humax - self.humin - 1)
			self._dedx_scale = (self.cdedx_max - self.cdedx_min) \
						/ float(self.humax - self.humin - 1)
		else:
			self._rho_scale  = 0.0
			self._dedx_scale = 0.0
#		if abs(self.crho(self.humin) - self.crho_min) > 0.001:
#			print("error1", self)
#		if abs(self.crho(self.humax-1) - self.crho_max) > 0.001:
#			print()
#			print("error2", self)
#			print("interpolated:",self.crho(self.humax-1))
#			print("max:",self.crho_max)

	#-----------------------------------------------------------------------
	def __str__(self):
		return "[%5d .. %5d)	%s  rho:[%g .. %g]  dedx:[%g .. %g]" % \
			(self.humin, self.humax, self.mat,
			 self.crho_min, self.crho_max,
			 self.cdedx_min, self.cdedx_max)

	#-----------------------------------------------------------------------
	def str(self):
		return "%d %s  %.10g %.10g  %.10g  %.10g" % \
			(self.humax, self.mat,
			 self.crho_min, self.crho_max,
			 self.cdedx_min, self.cdedx_max)

	#-----------------------------------------------------------------------
	def crho(self, h):
		return (h - self.humin) * self._rho_scale  + self.crho_min

	#-----------------------------------------------------------------------
	def cdedx(self, h):
		return (h - self.humin) * self._dedx_scale + self.cdedx_min

	#-----------------------------------------------------------------------
	@staticmethod
	def reset():
		Hounsfield.prev = -9999

#===============================================================================
# Dicom helper class
#===============================================================================
class Dicom:
	#-----------------------------------------------------------------------
	def __init__(self, filename=None, log_=None):
		self.setLog(log_)
		self.dataset  = None
		self.clear()
		self.clearHounsfield()

		if filename is not None:
			self.open(filename)
			self.init()

	#-----------------------------------------------------------------------
	# Clear information comming from scanning
	#-----------------------------------------------------------------------
	def clear(self):
		self.uid      = None
		self.name     = ""
		self.modality = ""
		self.m        = 1.0
		self.b        = 0.0
		self.x        = 0.0
		self.y        = 0.0
		self.z        = 0.0
		self.dx       = 0.0
		self.dy       = 0.0
		self.dz       = 0.0
		self.nx       = 0
		self.ny       = 0
		self.nz       = 0
		self.iop      = [1., 0., 0.,  0., 1., 0.]	# Image patient orientation
		self.instance = 0
		self.hasSlices= False
		self.hasPixel = False
		self.dtype    = None
		self.xfrom    = None
		self.yfrom    = None
		self.zfrom    = None
		self.xto      = None
		self.yto      = None
		self.zto      = None
		self.sliceFrom = 0
		self.sliceTo  = sys.maxsize
		self.rowFrom  = 0
		self.rowTo    = sys.maxsize
		self.colFrom  = 0
		self.colTo    = sys.maxsize
		self.rowV     = None
		self.colV     = None
		self.sliceV   = None
		self.patientPosition = None

		self.array    = None		# 3D array from all slices used in RTViewer
		self._field   = 0

		self.scanned(False)

	# ----------------------------------------------------------------------
	def scanned(self, s=None):
		if s is None:
			return bool(self.slices)
		else:
			self.slices  = []
			self._values = set()
			self.minHU   =  99999999999
			self.maxHU   = -99999999999

	# ----------------------------------------------------------------------
	def clearHounsfield(self):
		Hounsfield.reset()
		self.hounsfield = []
		self.corrrho    = []
		self.corrdedx   = []
		self.correction = {}

	# ----------------------------------------------------------------------
	def setLog(self, log_):
		self._log = log_

	# ----------------------------------------------------------------------
	def log(self, s):
		if self._log is None:
			print(s)		# log
		else:
			self._log(s)

	# ----------------------------------------------------------------------
	# Open dicom with filename
	# ----------------------------------------------------------------------
	def open(self, filename):
		if isinstance(filename,str):
			try:
				self.dataset = pydicom.read_file(filename, force=True)
				return True
			except (FileNotFoundError, pydicom.filereader.InvalidDicomError):
				self.log(f"{log.ERROR}Cannot open file {filename}")
				self.dataset = None
				return False
		else:
			self.dataset = filename

	# ----------------------------------------------------------------------
	def close(self):
		del self.dataset
		self.dataset = None

	# ----------------------------------------------------------------------
	def checkTransferSyntaxUID(self):
		try:
			if "TransferSyntaxUID" not in self.dataset.file_meta:
				self.dataset.file_meta.TransferSyntaxUID = \
					pydicom.uid.ImplicitVRLittleEndian
		except:
			pass

	#-----------------------------------------------------------------------
	# Initialize dicom information by reading the first slice
	#-----------------------------------------------------------------------
	def init(self, filename=None, nz=0):
		#Global.dump("*-* Dicom.init")
		self.clear()
		self.nz = nz

		# Load one slice to get dimensions
		if isinstance(filename,str):
			self.open(filename)
		elif filename is not None:
			self.dataset = filename

		self.uid = self.dataset.SeriesInstanceUID
		try:
			self.instance = int(self.dataset.InstanceNumber)
		except:
			self.instance = 1

		if "PatientIdentityRemoved" in self.dataset:
			self.name = "-patient identity removed-"
		elif "PatientName" in self.dataset:
			pat_name = self.dataset.PatientName
			try:
				self.name = f"{pat_name.family_name}, {pat_name.given_name}"
			except:
				self.name = str(pat_name)
		else:
			self.name = "none"

		try:
			self.modality = self.dataset.Modality
		except AttributeError:
			self.modality = 'dicom'

		if "PixelData" not in self.dataset:
			self.hasPixel = False
			self.dtype = None
			return True
		else:
			self.checkTransferSyntaxUID()
			try:
				self.dtype = self.dataset.pixel_array.dtype
				self.hasPixel = True
			except (OSError, NotImplementedError, TypeError) as err:
				self.log(f"{log.ERROR} Error reading pixel_array of '{filename}'")
				self.log(f"{log.ERRORCONT}{err}")

		# Image patient orientation
		try:
			self.iop = [float(x) for x in self.dataset.ImageOrientationPatient]
		except:
			self.iop = [1., 0., 0.,  0., 1., 0.]
		self._makeVectors()

		try:
			self.patientPosition = self.dataset.PatientPosition
		except:
			self.patientPosition = "HFS"

		frame = None
		if "PixelSpacing" not in self.dataset:
			# Most probably new 3D format
			self.dx = self.dy = 0.0
			self.dz = 0.0
			found = False

			# Try first in SharedFunctionalGroupsSequence
			if "SharedFunctionalGroupsSequence" in self.dataset:
				found = True
				frame = self.dataset.SharedFunctionalGroupsSequence[0]
				try:
					self.dx, self.dy = map(float,
							frame.PixelMeasuresSequence[0].PixelSpacing)
					self.dz = float(frame.PixelMeasuresSequence[0].SliceThickness)
				except:
					pass

			# Then in PerFrameFunctionalGroupsSequence
			if "PerFrameFunctionalGroupsSequence" in self.dataset:
				found = True
				frame = self.dataset.PerFrameFunctionalGroupsSequence[0]
				if "PixelMeasuresSequence" in frame:
					try:
						self.dx, self.dy = map(float,
							frame.PixelMeasuresSequence[0].PixelSpacing)
					except:
						pass
					if "SliceThickness" in frame.PixelMeasuresSequence[0]:
						self.dz = float(frame.PixelMeasuresSequence[0]
									.SliceThickness)
					elif "SliceThickness" in frame[0x2005140f][0]:	# Private tag?
						self.dz = float(frame[0x2005140f][0].SliceThickness)

			if not found:
				self.log(f"{log.ERROR}Unknown format: No PixelSpacing or"
					" PerFrameFunctionalGroupsSequence or"
					" SharedFunctionalGroupsSequence found in DICOM")

			if self.dx*self.dy==0.0:
				self.log(f"{log.ERROR}Unknown dicom format, PixelSpacing is not found.")

		else:
			# in mm
			self.dx, self.dy = map(float,self.dataset.PixelSpacing)
			try:
				self.dz = float(self.dataset.SliceThickness)
			except:
				self.dz = 0.0
				if self.modality == "RTDOSE":
					try :
						self.dz = float(self.dataset.GridFrameOffsetVector[1])
					except AttributeError:
						pass

		#if self.dataset.RescaleType != "HU":
		#	self.log(f"{log.ERROR} cannot handle DICOM with rescale type different than HU")
		#	return False

		# FIXME self.m,b should be per slice/frame not global!

		self.m = 1.0	# rescale slope
		self.b = 0.0	# rescale intercept

		if frame is not None:
			try:
				self.m = float(frame.PixelValueTransformationSequence[0]
							.RescaleSlope)
				self.b = float(frame.PixelValueTransformationSequence[0]
							.RescaleIntercept)
			except AttributeError:
				pass
		elif self.modality == "RTDOSE":
			try :
				self.m = float(self.dataset.DoseGridScaling)
			except AttributeError:
				pass
		else:
			try:
				self.m = float(self.dataset.RescaleSlope)
				self.b = float(self.dataset.RescaleIntercept)
			except AttributeError:
				pass

		if frame is not None:
			try:
				self.x, self.y, self.z = map(float,
						frame.PlanePositionSequence[0]
							.ImagePositionPatient)
			except:
				self.x = self.y = self.z = 0.0
		else:
			try:
				self.x, self.y, self.z = map(float,
						self.dataset.ImagePositionPatient)
			except:
				self.x = self.y = self.z = 0.0

		# get limits
		self.nx = int(self.dataset.Columns)
		self.ny = int(self.dataset.Rows)
		if 'NumberOfFrames' in self.dataset:
			# 3D data
			self.nz        = int(self.dataset.NumberOfFrames)
			self.hasSlices = False
		else:
			# 2D slice data
			self.hasSlices = True
			if 'SliceLocation' not in self.dataset and \
			   'NumberOfSlices' not in self.dataset:
				self.log(f"{log.ERROR}Cannot handle DICOM without both "
					f"SliceLocation and NumberOfFrames {filename}")
				return False

		# Calculate the limits for the first time
		self.calcLimits()
		if isinstance(filename,str): self.close()
		return True

	# ----------------------------------------------------------------------
	# Construct vectors and rotation matrices based on IOP
	# ----------------------------------------------------------------------
	def _makeVectors(self):
		#Global.dump("*-* Dicom._makeVectors")
		self.rowV    = bmath.Vector(self.iop[:3])
		self.columnV = bmath.Vector(self.iop[3:])
		self.sliceV  = self.rowV ^ self.columnV
		self.X = bmath.Vector(self.rowV[0], self.columnV[0], self.sliceV[0])
		self.Y = bmath.Vector(self.rowV[1], self.columnV[1], self.sliceV[1])
		self.Z = bmath.Vector(self.rowV[2], self.columnV[2], self.sliceV[2])
		# transformation matrix from Image to RCS (ijk to xyz)
		self.image2RCS = bmath.Matrix([self.rowV, self.columnV, self.sliceV])
		self.RCS2image = self.image2RCS.T()
		self.Rx, self.Ry, self.Rz = self.RCS2image.getEulerRotation()
		#Global.dump("\tIOP   =",self.iop)
		#Global.dump("\tRow   =",self.rowV)
		#Global.dump("\tColumn=",self.columnV)
		#Global.dump("\tSlice =",self.sliceV)
		#Global.dump("\tImage2RCS")
		#Global.dump(self.image2RCS)
		#Global.dump("\tRCS2image")
		#Global.dump(self.RCS2image)
		#Global.dump("\tEuler Rx,Ry,Rz=",self.Rx,self.Ry,self.Rz)

	# ----------------------------------------------------------------------
	# @return true if they have changed
	# ----------------------------------------------------------------------
	def setLimits(self, xfrom=None, xto=None, yfrom=None, yto=None, zfrom=None, zto=None):
		self.xfrom = xfrom
		self.yfrom = yfrom
		self.zfrom = zfrom
		self.xto   = xto
		self.yto   = yto
		self.zto   = zto
		return self.calcLimits()

	# ----------------------------------------------------------------------
	def calcLimits(self):
		sf = self.sliceFrom
		st = self.sliceTo
		rf = self.rowFrom
		rt = self.rowTo
		cf = self.colFrom
		ct = self.colTo
		try:    self.sliceFrom = min(max(0,
					int(math.floor((float(self.zfrom)-self.z)/self.dz))),self.nz-1)
		except: self.sliceFrom = 0
		try:    self.sliceTo   = min(max(self.sliceFrom,
					int(math.floor((float(self.zto)-self.z)/self.dz))),self.nz-1)
		except: self.sliceTo   = self.nz-1

		try:    self.rowFrom   = min(max(0,
					int(math.floor((float(self.yfrom)-self.y)/self.dy))),self.ny-1)
		except: self.rowFrom   = 0
		try:    self.rowTo     = min(max(self.rowFrom,
					int(math.floor((float(self.yto)-self.y)/self.dy))),self.ny-1)
		except: self.rowTo     = self.ny-1

		try:    self.colFrom   = min(max(0,
					int(math.floor((float(self.xfrom)-self.x)/self.dx))),self.nx-1)
		except: self.colFrom   = 0
		try:    self.colTo     = min(max(self.colFrom,
					int(math.floor((float(self.xto)-self.x)/self.dx))),self.nx-1)
		except: self.colTo     = self.nx-1

		return sf!=self.sliceFrom or st!=self.sliceTo or \
		       rf!=self.rowFrom   or rt!=self.rowTo   or \
		       cf!=self.colFrom   or ct!=self.colTo

	#-----------------------------------------------------------------------
	def ncolumns(self): return min(self.nx, self.colTo  -self.colFrom+1)
	def nrows(self):    return min(self.ny, self.rowTo  -self.rowFrom+1)
	def nslices(self):  return min(self.nz, self.sliceTo-self.sliceFrom+1)

	#-----------------------------------------------------------------------
	# WARNING, returns the coordinates in the Image system
	#-----------------------------------------------------------------------
	def xmin(self): return self.x + (self.colFrom-0.5)  *self.dx
	def ymin(self): return self.y + (self.rowFrom-0.5)  *self.dy
	def zmin(self): return self.z + (self.sliceFrom-0.5)*self.dz

	def xmax(self): return self.x + (self.ncolumns()+ self.colFrom-0.5)  *self.dx
	def ymax(self): return self.y + (self.nrows()   + self.rowFrom-0.5)  *self.dy
	def zmax(self): return self.z + (self.nslices() + self.sliceFrom-0.5)*self.dz

	#def dimx(self): return self.ncolumns()*self.dx
	#def dimy(self): return self.nrows()   *self.dy
	#def dimz(self): return self.nslices() *self.dz

	#-----------------------------------------------------------------------
	# return true if the iop is the default [1,0,0, 0,1,0] XYZ aligned
	# Cartesian system
	#-----------------------------------------------------------------------
	def iopIsXYZ(self):
		return  self.iop[0]==1 and self.iop[1]==0 and self.iop[2]==0 and \
			self.iop[3]==0 and self.iop[4]==1 and self.iop[5]==0

	#-----------------------------------------------------------------------
	# @return origin vector
	#-----------------------------------------------------------------------
	def origin(self):
		return bmath.Vector(self.x, self.y, self.z) \
				+ self.dx*(self.colFrom-0.5)*self.X \
				+ self.dy*(self.rowFrom-0.5)*self.Y \
				+ self.dz*(self.sliceFrom-0.5)*self.Z

	#-----------------------------------------------------------------------
	# Read Hounsfield material and ranges from a file
	#-----------------------------------------------------------------------
	def readHounsfield(self, filename):
		Hounsfield.reset()
		del self.hounsfield[:]

		try: f = open(filename,"r",encoding="UTF-8",errors="replace")
		except: return False
		with f:
			lines = []
			for line in f:
				line = line.strip()
				if not line or line[0] in "#*": continue
				elem     = line.split()
				humax    = int(elem[0])
				lines.append((humax, elem))
			lines.sort(key=itemgetter(0))

			for humax,elem in lines:
				material = elem[1]
				try:	crho_min = float(elem[2])
				except: crho_min = 1.0
				try:	crho_max = float(elem[3])
				except: crho_max = 1.0
				try:	cdedx_min = float(elem[4])
				except: cdedx_min = 1.0
				try:	cdedx_max = float(elem[5])
				except: cdedx_max = 1.0
				hu = Hounsfield(humax, material, crho_min, crho_max, cdedx_min, cdedx_max)
				self.hounsfield.append(hu)
		self.interpolateCorrections()

	#-----------------------------------------------------------------------
	def interpolateCorrections(self):
		self.correction.clear()
		for hu in self.hounsfield:
			# Interpolate correction
			if hu.humax > hu.humin:
				for h in range(hu.humin, hu.humax):
					self.correction[h] = (-hu.cdedx(h), hu.crho(h))
			else:
				self.correction[hu.humin] = (-hu.cdedx_min, hu.crho_min)

	#-----------------------------------------------------------------------
	def readCorrection(self, filename):
		self.correction = {}
		try: f = open(filename,"r",encoding="UTF-8")
		except: return False
		for line in f:
			line = line.strip()
			if not line or line[0]=="*": continue
			elem = list(map(float,line.split()))
			self.correction[int(elem[0])] = elem[1:]
		f.close()

	#-----------------------------------------------------------------------
	# Make a zero array
	#-----------------------------------------------------------------------
	def makeArray(self):
		self.array = np.zeros((self.nx, self.ny, self.nz), dtype=self.dtype)

	#-----------------------------------------------------------------------
	# Scan slice for unique Hounsfield values
	#-----------------------------------------------------------------------
	def scanSlice(self, filename=None, s=None, update=None):
		if not self.open(filename):
			return False

		if self.uid != self.dataset.SeriesInstanceUID:
			self.log(f"{log.ERROR}Different Series Instance UID found")
			return False

		self.checkTransferSyntaxUID()

		# Scan for unique Hounsfield regions
		if self.hasSlices:
			assert s is not None

			if self.sliceFrom <= s <= self.sliceTo:
				self.__scanSliceData(self.dataset.pixel_array)
			try:
				k = bmath.Vector(map(float, self.dataset.ImagePositionPatient)) \
						* self.sliceV
				self.slices.append((k, filename, self.dataset.SOPInstanceUID))
			except AttributeError:
				try:
					self.slices.append((float(self.dataset.SliceLocation),
							filename, self.dataset.SOPInstanceUID))
				except AttributeError:
					self.slices.append((0, filename, self.dataset.SOPInstanceUID))
		else:
			try:	# FIXME Only for DOSE maps
				zoff = self.dataset.GridFrameOffsetVector
			except:
				zoff = [x*self.dz for x in range(len(self.dataset.pixel_array))]
			for s,data_slice in enumerate(self.dataset.pixel_array):
				self.slices.append((float(zoff[s]),
							filename,
							self.dataset.SOPInstanceUID))
				if s is None or self.sliceFrom <= s <= self.sliceTo:
					self.__scanSliceData(data_slice)
				if update and update(s): break
		self.close()
		return True

	#-----------------------------------------------------------------------
	# Scan data to create a histogram of unique Hounsfield values
	#-----------------------------------------------------------------------
	def __scanSliceData(self, data):
		if self.rowFrom>0 or self.rowTo<self.ny-1 or \
		   self.colFrom>0 or self.colTo<self.nx-1:
			array = data[self.rowFrom:self.rowTo+1, self.colFrom:self.colTo+1]
		else:
			array = data
		self._values.update(np.unique(np.int32(self.m*array + self.b)))

	#-----------------------------------------------------------------------
	def scanSliceEnd(self):
		# Sorted list of existing Hounsfield units in all slices
		# Each Hounsfield is one region
		# XXX WARNING the conversion should be the same with the _writeVoxelSlice ###
		# convert to int from numpy.int16, otherwise we might end
		# in overflows in operations e.g. maxHU-minHU
		self.reg2hu = [int(x) for x in sorted(self._values)]
		self.minHU  = self.reg2hu[0]
		self.maxHU  = self.reg2hu[-1]
		# inverse table (kreg in the notion of fluka)
		self.vxl2reg = [0]*(self.maxHU - self.minHU)
		# Keep first Unit (normally -1024) as the master voxel/organ no.0
		# the vxl2reg(kreg) should start from voxel/organ no.1
		if self.vxl2reg:
			for i,h in enumerate(self.reg2hu[1:],1):
				self.vxl2reg[h-self.minHU-1] = i

		# DEBUG
		if Global.developer and Global.verbose>1:
			Global.dump("Limits:",self.minHU, self.maxHU)
			Global.dump("\nreg2hu")
			for i,r in enumerate(self.reg2hu):
				Global.dump("\t",i,r)

			#print(f"{self._values=}")
			#print(f"{len(self.vxl2reg)=}")
			#print(f"{self.vxl2reg=}")
			Global.dump("\nvxl2reg")
			for i,r in enumerate(self.vxl2reg):
				Global.dump("\t",i,r)

		# Sort slices
		self.slices = sorted(self.slices, key=itemgetter(0))
		if self.nz==0: self.nz = len(self.slices)

		# Correct slice thickness, use the average from the last SliceLocation to the first
		# FIXME will it work with the slices that are in the image space?
		if len(self.slices)>1:
			dz = (self.slices[-1][0] - self.slices[0][0]) / (len(self.slices)-1)
			if abs(self.dz-dz) > 1e-5:
				self.log(f"{log.WARNING} Slice thickness vs calculated one "
					"from SliceLocation do not match")
				self.log(f"   Slice thickness: {self.dz}")
				self.log(f"   Slice Location:  {dz}")
				self.dz = dz

		prev = self.slices[0][0] - self.dz
		err = ""
		dzs = []
		for l,fn,sop in self.slices:
			dzs.append(l-prev)
			if abs((l-prev) - self.dz) > 1E-5:
				self.log(f"\nSlice: {fn}")
				self.log(f"Difference from previous: {l-prev}")
				self.log(f"Report slice thickness:   {self.dz}")
				if not err:
					err = f"{log.WARNING}Slice is missing or wrong thickness."
				err +=  f"\nLocation z= {l}\n" \
					f"Difference from previous: {l-prev}\n" \
					f"Slice thickness: {self.dz}"
			prev = l
		return err

	#-----------------------------------------------------------------------
	def assignMaterials2Regions(self, field=0):
		self._field = field
		self.assignmat = []
		Global.dump("Assign Materials")
		for h in self.reg2hu:
			for hounsfield in self.hounsfield:
				if hounsfield.humin <= h < hounsfield.humax:
					self.assignmat.append(hounsfield.mat)
					Global.dump("\t",len(self.assignmat),self.assignmat[-1])
					break
			else:
				self.log(f"{log.WARNING}Out of range HU={h} assigned last material")
				self.assignmat.append(self.hounsfield[-1].mat)
				Global.dump("\t",len(self.assignmat)-1,self.assignmat[-1])
		return True

	#-----------------------------------------------------------------------
	def _write(self, data):
		fortran.write(self._f, data)

	#-----------------------------------------------------------------------
	def _writeStr(self, data):
		self._write(data.encode())

	#-----------------------------------------------------------------------
	def _writeStruct(self, fmt, *args):
		fortran.write(self._f, struct.pack(fmt, *args))

	#-----------------------------------------------------------------------
	# Prepare all the voxel data from the slices
	#-----------------------------------------------------------------------
	def makeVoxel(self, update=None):
		# voxel data
		self._data = []

		# initialize the rtstuct arrays
		self._roiname   = {}	# index=ROINumber, value=name,volume
		self._roicomb   = {}	# index=:RTSTRUCT[:RTSTRUCT] (Unique combination)
					# value=sequential index
		self._roidata   = []	# matrix of seq-index
		self._roicolor  = {}	# color of each structure
		self._roivolume = {}	# volume of each structure
		#self._roiplanar = {}	# planar information for CLOSED_PLANAR

		if self.hasSlices:
			# Multiple slices
			for i,(l,fn,sop) in enumerate(self.slices):
				if not self.sliceFrom <= i <= self.sliceTo:
					continue
				# Show progress
				if update and update(i, f"Processing: {os.path.basename(fn)}"):
					return False
				self.open(fn)
				self.checkTransferSyntaxUID()
				self.__processSliceData(self.dataset.pixel_array)
				self.close()
		else:
			self.open(self.slices[0][1])
			self.checkTransferSyntaxUID()
			for i,data_slice in enumerate(self.dataset.pixel_array):
				if self.sliceFrom <= i <= self.sliceTo:
					self.__processSliceData(data_slice)
				if update and \
				   update(i, f"Processing: {os.path.basename(self.slices[0][1])}"):
					self.close()
					return False
			self.close()
		return True

	#-----------------------------------------------------------------------
	def __processSliceData(self, data):
		if self.rowFrom>0 or self.rowTo<self.ny-1 or \
		   self.colFrom>0 or self.colTo<self.nx-1:
			array = data[self.rowFrom:self.rowTo+1, self.colFrom:self.colTo+1]
		else:
			array = data

		# XXX in the past it was uint16(m*array+b-minHU)
		# and it was introducing numerical errors +/-1 value
		self._data.append(np.uint16(np.int32(self.m*array + self.b) - self.minHU))

		# Verify that all unique values are referred in the vxl2reg
		if Global.developer:
			for u in np.unique(self._data[-1]):
				if u==0: continue	# default value
				assert self.vxl2reg[u-1] != 0

	#-----------------------------------------------------------------------
	# Open an rtstruct file and process the ROI present
	#-----------------------------------------------------------------------
	def processRTStruct(self, rtstruct_file, update=None):
		try:
			rtstruct = pydicom.read_file(rtstruct_file, force=True)
		except pydicom.filereader.InvalidDicomError:
			return False
		# For all(cropped) SOP of DICOM scan the RTSTRUCT
		for i,(z,fn,sop) in enumerate(self.slices):
			self.__processRTStructSlice(rtstruct,sop,i)
			if update and update(i, f"Processing: {sop}"):
				return False
		del rtstruct
		Global.dump(f"Drawing Time={self._drawingTime}")
		Global.dump(f"Scanning Time={self._scanningTime}")
		return True

	#-----------------------------------------------------------------------
	# Find the ROI (RTSTRUCT) and create the 3D matrix as well the
	# correspondence table
	#-----------------------------------------------------------------------
	def __processRTStructSlice(self, rtstruct, sop, s=None):
		if Image is None: return
		if s is not None and not self.sliceFrom <= s <= self.sliceTo: return
		rois   = []
		images = []

		# Prepare table with index, names, volume
		for roi in rtstruct.StructureSetROISequence:
			try:
				# Associate ROI-number/index with the ROI-name
				self._roiname[int(roi.ROINumber)] = str(roi.ROIName)
			except:
				pass

		Xx = self.iop[0]; Xy = self.iop[1]
		Yx = self.iop[3]; Yy = self.iop[4]
		x0 = self.xmin()
		y0 = self.ymin()

		t0 = time.time()
		# Loop over all ROIs
		for sequence in rtstruct.ROIContourSequence:
			# Loop over all contours to find SOP
			draw = None
			roi = int(sequence.ReferencedROINumber)
			#planar = []

			if roi not in self._roicolor:
				try:
					color = sequence.ROIDisplayColor
					# Swap color order to be used in geoviewer
					self._roicolor[roi] =   int(color[0])<<16 | \
								int(color[1])<< 8 | \
								int(color[2])
				except AttributeError:
					self._roicolor[roi] = 128

				try:
					self._roivolume[roi] = float(sequence.ROIVolume)
				except:
					self._roivolume[roi] = 0.0

			if "ContourSequence" not in sequence: continue
			for contour in sequence.ContourSequence:
				# Ignore non-planar (for the moment)
				if "ContourGeometricType" not in contour: continue
				if contour.ContourGeometricType != "CLOSED_PLANAR" or \
				   contour.NumberOfContourPoints == 1:
					continue
				if contour.ContourImageSequence[0].ReferencedSOPInstanceUID != sop:
					continue
				# We might have more than one contour per slice
				if draw is None:
					# First time create the image/draw
					rois.append(int(sequence.ReferencedROINumber))
					image = Image.new("L", (self.ncolumns(),self.nrows()))
					images.append(image)
					draw = ImageDraw.Draw(image)

				coords = list(map(float, contour.ContourData))
				XY = list(zip(coords[::3], coords[1::3]))
				X  = [Xx*(x-x0) + Xy*(y-y0) for x,y in XY]
				Y  = [Yx*(x-x0) + Yy*(y-y0) for x,y in XY]

				# FIXME me correct the origin!
				#planar.append((zi,zip(X,Y)))
				iX = [int(round(x/self.dx)) for x in X]
				iY = [int(round(y/self.dy)) for y in Y]
				draw.polygon(list(zip(iX,iY)), fill=0x80)	# fill

				# draw outline with our routine to correctly mark all
				# possible crossing voxels for the outline, based on the
				# true floating point position of the structure
				x1 = X[-1]/self.dx
				y1 = Y[-1]/self.dy
				for x,y in zip(X,Y):
					x2 = x/self.dx
					y2 = y/self.dy
					drawLine(image, x1, y1, x2, y2, 0xff)
					x1,y1 = x2,y2
				#image.save("ROI%02d_z%03d.png"%(roi, zi))
			#if planar: self._roiplanar[roi] = planar
			del draw
		try: self._drawingTime += time.time()-t0
		except: self._drawingTime = time.time()-t0

		roimg = [(str(roi),img.tobytes()) for roi,img in zip(rois,images)]

		t0 = time.time()
		# Process images
		ptr  = 0
		data = b""
		# Warning ZERO based
		self._roicomb[""] = 0	# first element
		for j in range(self.nrows()):
			col = []
			for i in range(self.ncolumns()):
				# rois appearing in pixel as unique string :roi[:roi]
				pixel = ""
				for roi,img in roimg:
					cimg = img[ptr]
					if   cimg==0x80: pixel += f":{roi}"
					elif cimg==0xff: pixel += f":-{roi}"
				ptr += 1
				try:
					col.append(self._roicomb[pixel])
				except KeyError: # add new id
					self._roicomb[pixel] = idx = len(self._roicomb)
					col.append(idx)
			data += struct.pack("=%dH"%(len(col)), *col)
		self._roidata.append(data)
		try: self._scanningTime += time.time()-t0
		except: self._scanningTime = time.time()-t0
		del data
		del images

	#-----------------------------------------------------------------------
	# Apply ROI filters to define specific materials
	# filters should be a list in the form
	#	filters: [ ("material", [expr])... ]
	#	expr:    [ [+/-]roiid... ] even empty
	#
	# WARNING: filter will after the vxl2reg structure, therefore
	#	analyze slices will be needed afterwards
	#-----------------------------------------------------------------------
	def filter(self, filters, update=None):
		if not filters: return

		# Prepare the vxl2reg, assigmat association vectors
		filterId = []
		for mat,expr in filters:
			self.assignmat.append(mat)
			self.vxl2reg.append(len(self.assignmat)-1)
			filterId.append(len(self.vxl2reg))
			Global.dump(f"*-* Filter {len(self.assignmat)}={mat}")
			Global.dump(f"\t vxl2reg {len(self.vxl2reg)}={self.vxl2reg[-1]}")
			Global.dump(f"\t filter {len(filterId)}={filterId[-1]}")

		# Remove unused entries from assign mat and vxl2reg array
		# Filter the CT data
		for i,(data,roi) in enumerate(zip(self._data, self._roidata)):
			R = np.frombuffer(roi, dtype=np.uint16).reshape(*data.shape)
			for matid,(name,expr) in zip(filterId, filters):
				for roicomb,idx in self._roicomb.items():
					if expr:
						if roicomb:
							rois = {abs(int(x))
								for x in roicomb[1:].split(":")}
						else:
							rois = set()
						for item in expr:
							if item>0 and  item not in rois: break
							if item<0 and -item     in rois: break
						else:
							data[R==idx] = matid
					elif not expr and not roicomb:
						data[R==idx] = matid
			if update and update(i, f"Material override filtering slice {i}"):
				return False

		if Global.verbose>1:
			Global.dump("\nvxl2reg before")
			for i,k in enumerate(self.vxl2reg):
				if k>0:
					Global.dump("\t",i,k,self.assignmat[k])

		# Remove unused entries from assign mat and vxl2reg array
		unique = set()
		for data in self._data:
			unique.update(np.unique(data))

		t0 = time.time()
		for i,k in reversed(list(enumerate(self.vxl2reg))):
			if k!=0 and i+1 not in unique:
				Global.dump(f">>> Removing organ={i} "
					f"region={k+1} {self.assignmat[k]}")
				self.vxl2reg[i] = 0
				del self.assignmat[k]
				try:
					del self.reg2hu[k]
				except IndexError:
					pass
				for j in range(i+1,len(self.vxl2reg)):
					if self.vxl2reg[j] > 0:
						self.vxl2reg[j] -= 1
			if update and update(len(self.vxl2reg)-i,
					f"Material override cleanup slice {i}"):
				return False

		# Special case for the cage organ
		if 0 not in unique:
			Global.dump(f">>> Removing organ={0} region={1} {self.assignmat[0]}")
			# shift everything by the first n
			humin = min(unique)
			for i,data in enumerate(self._data):
				self._data[i] = data - humin
			del self.assignmat[0]
			del self.reg2hu[0]
			del self.vxl2reg[0:humin]
			self.minHU += humin
			for j in range(len(self.vxl2reg)):
				if self.vxl2reg[j] > 0:
					self.vxl2reg[j] -= 1
		# Trim trailing zeros if any
		while self.vxl2reg and self.vxl2reg[-1] == 0:
			del self.vxl2reg[-1]
		assert len(self.vxl2reg) == max(unique)-min(unique)

		Global.dump(f"Time to remove organs= {time.time()-t0}")

		if Global.verbose>1:
			Global.dump("\nvxl2reg after")
			for i,k in enumerate(self.vxl2reg):
				if k>0:
					Global.dump("\t",i,k,self.assignmat[k])

	#-----------------------------------------------------------------------
	# Finalize the file by writing the data, materials, etc..
	#-----------------------------------------------------------------------
	def writeVoxel(self, filename):
		try: self._f = open(filename,"wb")
		except: return False

		# write title
		self._writeStr("%-80s"%(self.name))

		# memory size
		header = struct.pack("=5i", self.ncolumns(), self.nrows(), self.nslices(),
						len(self.assignmat)-1, len(self.vxl2reg))
		if len(self._roicomb)>1:
			# Number of structures (consecutive)
			header += struct.pack("=i", max(self._roiname))
			# Number of combinations
			header += struct.pack("=i", len(self._roicomb)-1)	# Minus the 0 index
			# Maximum length of combinations
			maxcomb = max([x.count(":") for x in self._roicomb])
			header += struct.pack("=i", maxcomb)	# Minus the 0 index
			# Number of CLOSED_PLANAR structures
			if Global.experimental:
				header += struct.pack("=i", 0)		# ignore for the moment
				#header += struct.pack("=i", len(self._roiplanar))
		self._write(header)

		# voxel dimension
		self._writeStruct("=3d", 0.1*self.dx, 0.1*self.dy, 0.1*self.dz)

		# data
		self._write(b"".join([x.data for x in self._data]))
		#print(f"{self._data=}")

		# hu -> region matrix
		self._writeStruct("=%dH"%(len(self.vxl2reg)), *self.vxl2reg)
		#print(f"{self.vxl2reg=}")

		# rt struct
		self.__writeVoxelRTstruct()

		# embed material definition
		self.writeMaterials()

		# assignmat matrix
		self.writeAssignmats()

		# correction factors
		self.writeCorrFactors()

		self._f.close()

		self.printReport()
		del self._data
		return True

	#-----------------------------------------------------------------------
	# Print report of the Dicom to voxel conversion
	#-----------------------------------------------------------------------
	def printReport(self):
		log.say("Dicom to voxel conversion report")
		log.say(f"{'Dicom':5} {'Counts':>8} {'HU':>5} {'Organ':5} {'Region':6} "
			f"{'Material':16} {'Density':>8} {'C-rho':>7} {'C-dE/dx':>7}")

		# make statistics
		count = defaultdict(int)
		for data in self._data:
			for u,c in zip(*np.unique(data, return_counts=True)):
				count[u] += c
		# 0=cage cell, 1=first organ vxl2reg[0] ...
		for vxl in range(len(self.vxl2reg)+1):
			if vxl==0:
				reg = 0
			else:
				reg = self.vxl2reg[vxl-1]
				if reg==0: continue	# skip empty vxl
			mat = self.assignmat[reg]
			hu  = vxl + self.minHU
			dicom = int((hu-self.b) / self.m)
			if Global.developer:
				assert count[vxl] != 0
			try:
				density = Material.get(mat).density
			except:
				density = 0.0
			try:
				cdedx,crho = self.correction[hu]
				rho = density*crho
			except KeyError:
				cdedx = -1.0
				crho  =  1.0
				rho = density
			log.say(f"{dicom:5} {count[vxl]:8} {hu:5} {vxl:5} {reg+1:6} "
				f"{mat:16} {rho:8.5f} {crho:7.5f} {cdedx:7.5f}")

		if len(self._roicomb)>1:
			log.say()
			log.say("RT structure dictionary")
			log.say(f"{'Id':4} {'Length':6} Rois")
			for i in range(1,len(self._roicomb)):
				for key,value in self._roicomb.items():
					if value == i:
						# Found break the key into the ROI sequence
						rois = sorted(map(int,key[1:].split(":")))
						lenrois = len(rois)
						log.say(f"{i:4} {lenrois:6} {rois}")
						break

	#-----------------------------------------------------------------------
	# Write the RTSTRUCT if they are there
	#-----------------------------------------------------------------------
	def __writeVoxelRTstruct(self):
		if len(self._roicomb)<=1: return

		# Write structure information
		for i in range(1,max(self._roiname)+1):
			name   = self._roiname.get(i,"")
			color  = self._roicolor.get(i,0)
			volume = self._roivolume.get(i,0.0)
			self._writeStruct("=id64s", color, volume, f"{name:64}".encode())

		# Write indexes (WARNING skip 0 which is empty)
		maxcomb = max([x.count(":") for x in self._roicomb]) # maxHU combination
		for i in range(1,len(self._roicomb)):
			# Find the key that has this index
			for key,value in self._roicomb.items():
				if value == i:
					# Found break the key into the ROI sequence
					rois = sorted(map(int,key[1:].split(":")))
					lenrois = len(rois)
					Global.dump("*-*", i,lenrois,rois)
					rois.extend([0]*(maxcomb-lenrois))
					self._writeStruct(f"=i{len(rois)}h",lenrois,*rois)
					break

		# Write the data
		self._write(b"".join(self._roidata))

		Global.dump("*-* RTSTRUCT len=",len(b"".join(self._roidata)))
		Global.dump("*-* Dictionary len=",len(self._roicomb),
				"max=",max([len(x.split(":")) for x in self._roicomb]))

		###################
		# Write planar information
		#if self._roiplanar:
		#	for roi in sorted(self._roiplanar):
		#		planar = self._roiplanar[roi]
		#		self._writeStruct("=ii", roi, len(planar))
		#		for zi,xy in planar:
		#			rec = struct.pack("=ii",zi,len(xy))
		#			for x,y in xy:
		#				rec += struct.pack("=ff",x,y)
		#			self._write(rec)
		###################

		# Cleanup memory
		#del self._roiplanar
		#del self._roicomb
		del self._roidata

	#-----------------------------------------------------------------------
	# Write USRBIN header
	#-----------------------------------------------------------------------
	def writeUsrbinInit(self, filename):
		try: self._f = open(filename,"wb")
		except: return False

		# write title
		self._writeStruct("=80s32sfiii",
					("%-80s"%(self.name)).encode(),
					("%-32s"%(time.asctime())).encode(),
					0.0,0,0,0)

		# write dicom as first usrbin file
		self._writeStruct("=i10siiffifffifffififff",
				1, (self.modality+" "*10)[:10].encode(), 0, 208,
				self.xmin()/10.0, self.xmax()/10.0, self.ncolumns(), self.dx/10.0,
				self.ymin()/10.0, self.ymax()/10.0, self.nrows(),    self.dy/10.0,
				self.zmin()/10.0, self.zmax()/10.0, self.nslices(),  self.dz/10.0,
				0, 0.0, 0.0, 0.0)

		self._data = []
		self._min =  1e10
		self._max = -1e10
		return True

	#-----------------------------------------------------------------------
	# Process a single slice of USRBIN
	#-----------------------------------------------------------------------
	def writeUsrbinSlice(self, filename, s=None, update=None):
		if s is not None and not self.sliceFrom <= s <= self.sliceTo: return
		if not self.open(filename): return
		self.checkTransferSyntaxUID()

		m = self.m
		b = self.b

		if self.modality == "RTDOSE":
			try :
				m = float(self.dataset.DoseGridScaling)
			except AttributeError:
				pass
		else:
			try:
				m = float(self.dataset.RescaleSlope)
				b = float(self.dataset.RescaleIntercept)
			except AttributeError:
				pass

		if self.hasSlices:
			self.__writeUsrbinSliceData(self.dataset.pixel_array, m, b)
		else:
			for s,data_slice in enumerate(self.dataset.pixel_array):
				if s is None or self.sliceFrom <= s <= self.sliceTo:
					self.__writeUsrbinSliceData(data_slice, m, b)
				if update and update(s): break
		self.close()

	#-----------------------------------------------------------------------
	# append data
	#-----------------------------------------------------------------------
	def __writeUsrbinSliceData(self, data, m, b):
		if self.rowFrom>0 or self.rowTo<self.ny-1 or \
		   self.colFrom>0 or self.colTo<self.nx-1:
			array = data[self.rowFrom:self.rowTo+1, self.colFrom:self.colTo+1]
		else:
			array = data
		array2 = np.float32(m*array + b)
		self._min = min(self._min, array2.min())
		self._max = max(self._max, array2.max())
		self._data.append(bytes(array2.data))

	#-----------------------------------------------------------------------
	# Finalize USRBIN by writing the data
	#-----------------------------------------------------------------------
	def writeUsrbinEnd(self):
		# data
		self._write(b"".join(self._data))
		del self._data
		self._f.close()
		Global.dump("Limits:", self._min, self._max)
		return True

	#-----------------------------------------------------------------------
	def _writeCard(self, card):
		scard = card.toStr(fmt=FlukaCard.FORMAT_SINGLE)
		for line in scard.splitlines():
			self._writeStr("%-80s"%(line[:80]))

	#-----------------------------------------------------------------------
	# Embed materials inside voxel file
	#-----------------------------------------------------------------------
	def writeMaterials(self):
		materials = set(list(Fluka.defaultMaterials.keys()) +
				list(Fluka.icruMaterials.keys()))
		for name in sorted(list(set(self.assignmat))):
			if name in materials: continue
			mat = Material.get(name)
			if mat is None:
				raise ValueError(f"Material {name} not found")
			for card in mat.materialCard(name, materials, Fluka.MODE):
				materials.add(card.name())
				self._writeCard(card)
			card = mat.matPropCard(name, Fluka.MODE)
			if card:
				self._writeCard(card)
		return True

	#-----------------------------------------------------------------------
	# Write assignmats
	#-----------------------------------------------------------------------
	def writeAssignmats(self):
		if self._field:
			extra="%20s%8d.0"%("",self._field)
		else:
			extra=""
		self._writeStr("ASSIGNMA  %10s%10s%-50s" % ("VACUUM","VOXEL",extra))
		for i,asmat in enumerate(self.assignmat,1):
			self._writeStr("ASSIGNMA  %10s%10s%-50s" %
				(asmat[:8],FlukaVoxel.regionName(i),extra))
			Global.dump("ASSIGNMA  %10s%10s%-50s" %
				(asmat[:8],FlukaVoxel.regionName(i),extra))

	#-----------------------------------------------------------------------
	# Write correction factors
	#-----------------------------------------------------------------------
	def writeCorrFactors(self):
		if not self.correction: return
		card = FlukaCard("CORRFACT",[""])
		for i,hu in enumerate(self.reg2hu,1):
			try:
				for j,c in enumerate(self.correction[hu],1):
					card.setWhat(j, c)
				card.setWhat(4, FlukaVoxel.regionName(i))
				scard = card.toStr()
				self._writeStr("%-80s"%(scard))
			except:
				pass

	#-----------------------------------------------------------------------
	# initialize dicom from a Project.DicomInfo file
	# @return None if everything ok, else error message
	#-----------------------------------------------------------------------
	def initDicomInfo(self, di):
		# Find slices and zmin
		self.clear()
		fn = di.filename()
		if di["frames"] == 0:
			# dicom with multiple slices, 3D array
			if not self.init(fn, len(di.files)):
				return f"loading slice {fn}"
		else:
			# Single slice dicom
			try:
				if not self.init(fn):
					return f"loading slice {fn}"
			except FileNotFoundError:
				return None
		self.calcLimits()	# calculate correctly the limits now we know the dz
		return None

	#-----------------------------------------------------------------------
	# Associate a dicom to dicominfo???
	# Not very nice place here!
	#-----------------------------------------------------------------------
	@staticmethod
	def add2DicomInfo(di):
		di.dicom = Dicom()
		di.dicom.initDicomInfo(di)

#===============================================================================
# RT structure processing
#===============================================================================
class RTStruct:
	#-----------------------------------------------------------------------
	def __init__(self, dataset, ctinfo=None):
		self.idx     = {}	# index to name
		self.names   = {}	# name to index
		self.volumes = defaultdict(lambda : (0.,0.,0.))	# volumes per index

		if isinstance(dataset,str):
			try:
				self.dataset = pydicom.read_file(dataset, force=True)
			except pydicom.filereader.InvalidDicomError:
				return
		else:
			self.dataset = dataset
		self.nstruct = len(self.dataset.ROIContourSequence)

		if ctinfo is not None:
			dicom = Dicom(ctinfo)
			self.x   = dicom.x
			self.y   = dicom.y
			self.dx  = dicom.dx
			self.dy  = dicom.dy
			self.dz  = dicom.dz
			self.nx  = dicom.nx
			self.ny  = dicom.ny
			self.iop = dicom.iop[:]
		else:
			self.x  = self.y  = None
			self.dx = self.dy = self.dz = None
			self.nx = self.ny = None
			self.iop = None

		self.findNames()
		#self.calcVolumes(self.dataset)

	#-----------------------------------------------------------------------
	def findNames(self):
		for roi in self.dataset.StructureSetROISequence:
			self.names[int(roi.ROINumber)] = str(roi.ROIName)
			self.idx[str(roi.ROIName)]     = int(roi.ROINumber)
			try:
				volume = float(roi.ROIVolume)
			except:
				volume = 0.0
			self.volumes[int(roi.ROINumber)] = (volume,0.,0.)

	#-----------------------------------------------------------------------
	# Calculate real volumes based on coordinates
	#-----------------------------------------------------------------------
	def calcVolumes(self, update=None):
		if self.iop is None: return

		Xx = self.iop[0]; Xy = self.iop[1]
		Yx = self.iop[3]; Yy = self.iop[4]
		x0 = self.x-self.dx/2.0
		y0 = self.y-self.dy/2.0
		dz = 1e99

		# Loop over all sequences
		for s,sequence in enumerate(self.dataset.ROIContourSequence):
			# Loop over all contours
			zmin =  1e99
			zmax = -1e99
			zp   = None
			Sv   = 0
			Sp   = 0.0
			if self.nx is not None:
				image = Image.new("L",(self.nx, self.ny))
				draw = ImageDraw.Draw(image)
			else:
				image = None

			if "ContourSequence" not in sequence: continue
			for contour in sequence.ContourSequence:
				# Ignore non-planar (for the moment)
				if "ContourGeometricType" not in contour: continue
				if contour.ContourGeometricType != "CLOSED_PLANAR" or \
				   contour.NumberOfContourPoints == 1:
					continue

				# Coordinates
				coords = list(map(float, contour.ContourData))
				# convert zip->list otherwise on
				# the first use it will be consumed
				# and on the second it will be empty!
				XY  = list(zip(coords[::3], coords[1::3]))
				X   = [Xx*(x-x0)+Xy*(y-y0) for x,y in XY]
				Y   = [Yx*(x-x0)+Yy*(y-y0) for x,y in XY]
				# WARNING center of voxel
				#X  = [x-self.x+self.dx/2.0 for x in X]
				#Y  = [y-self.y+self.dy/2.0 for y in Y]
				z    = coords[2]
				zmin = min(zmin,z)
				zmax = max(zmax,z)

				# Calculate volume from polygon
				S = 0.0
				xp = X[-1]
				yp = Y[-1]
				for x,y in zip(X,Y):
					S += (xp+x) * (yp-y)
					xp = x
					yp = y

				# Voxel-based volume
				if image:
					iX = [round(x/self.dx) for x in X]
					iY = [round(y/self.dy) for y in Y]
					# clear image
					draw.rectangle(((0,0),(self.nx,self.ny)), fill=0, outline=0)
					draw.polygon(list(zip(iX,iY)), fill=1)

					# draw outline with our routine to correctly mark all
					# possible crossing voxels for the outline, based on the
					# true floating point position of the structure
					x1 = X[-1]/self.dx
					y1 = Y[-1]/self.dy
					for x,y in zip(X,Y):
						x2 = x/self.dx
						y2 = y/self.dy
						drawLine(image, x1, y1, x2, y2, 1)
						x1,y1 = x2,y2
					# Some all entries with weight=1
					Sv += image.histogram()[1]
				Sp += abs(S)
				# Find next z
				if zp is not None and z != zp:
					dz = min(dz, z-zp)
				zp = z

			if image:
				Sv *= self.dx*self.dy*self.dz
			Sp = Sp * self.dz / 2.0
			Rv = self.volumes[int(sequence.ReferencedROINumber)][0]
			self.volumes[int(sequence.ReferencedROINumber)] = (Rv,Sv,Sp)
			if update is not None: update(s)

#===============================================================================
# DVH process
#===============================================================================
class DVHProcess(PopenProcess):
	def __init__(self):
		super().__init__()
		self.remove = False

	#-----------------------------------------------------------------------
	@tkFlair.exception
	def execute(self, cmd):
		log.say(f"Executing: {cmd}")
		rc = super().execute(cmd)
		outfile = cmd[cmd.index("-o")+1]
		bnnfile = cmd[cmd.index("-u")+1]
		if self.remove:
			try: os.remove(bnnfile)
			except: pass
		self.output(f"\nRemoved usrbin file {bnnfile}\n")
		self.message = "DVH created", f"DVH file {outfile}"
		return rc
