#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   21-Aug-2014

import tkinter as tk

try:
	import pydicom
except ImportError:
	try:
		import dicom as pydicom
	except ImportError:
		pydicom = None

import tool
import icons
import tkExtra
from stream import Stream

#===============================================================================
# Display the DICOM information in a browsable
#===============================================================================
class DicomBrowser(tkExtra.MultiListbox):
	_NONE   = "\N{DOTTED CIRCLE}"
	_OPEN   = "\N{BLACK DOWN-POINTING TRIANGLE}"
	_CLOSE  = "\N{BLACK RIGHT-POINTING TRIANGLE}"
	_MULTI  = "\N{DIE FACE-6}"
	_INDENT = "   "

	# ----------------------------------------------------------------------
	def __init__(self, master, dataset=None, filenames=None, **options):
		super().__init__(master,
				(('Name',  32, None),
				 ('Value', 48, None),
				 ('Tag',   12, None),
				 ('VR',     4, None),
				 ('VM',     4, None)), **options)
		self.sortAssist = None
		self.listbox(0).bind("<ButtonRelease-1>", self.buttonRelease)
		self.bindLeftRight()

		self.bindList("<Delete>",		self.deleteItem)
		self.bindList("<Key-space>",		self.toggle)

		self.listbox(0).bind("<Right>",		self.expand)
		self.listbox(0).bind("<Left>",		self.collapse)

		self.listbox(1).bind('<<Rename>>',	 self.inPlaceEdit)
		self.listbox(1).bind('<Control-Key-e>',	 self.inPlaceEdit)
		self.listbox(1).bind('<Return>',	 self.inPlaceEdit)
		self.listbox(1).bind('<KP_Enter>',	 self.inPlaceEdit)
		self.listbox(1).bind('<ButtonRelease-1>',self.inPlaceEdit)
		self.setPopupMenu(
				[('Copy',    0, self.copy,    icons.get("copy"))])

		self._tags = []		# tags per line
		self._open = {}		# open tags
		self._dataset     = None
		self._filenames   = None
		self._changed     = False
		self._filter      = None
		self._filterUpper = False
		self.dataset(dataset)
		self.filenames(filenames)

	# ----------------------------------------------------------------------
	def dataset(self, ds=None):
		if ds is not None:
			if isinstance(ds, list):
				self._dataset = ds
			else:
				self._dataset = [ds]
			self._changed = False
		return self._dataset

	# ----------------------------------------------------------------------
	def filenames(self, fn=None):
		if fn is not None:
			if isinstance(fn, list):
				self._filenames = fn
			else:
				self._filenames = [fn]
			assert len(self._filenames) == len(self._dataset)
		return self._filenames

	# ----------------------------------------------------------------------
	def clear(self):
		del self._tags[:]
		self._dataset   = None
		self._filenames = None
		self._changed   = False
		self.delete(0,tk.END)

	# ----------------------------------------------------------------------
	def openClear(self):
		self._open.clear()

	# ----------------------------------------------------------------------
	def changed(self): return self._changed

	# ----------------------------------------------------------------------
	def save(self, backup=True):
		if not self._changed or \
		   self._filenames is None: return False
		for ds,fn in zip(self._dataset, self._filenames):
			if backup: tool.backup(fn)
			ds.save_as(fn)
		self._changed = False
		return True

	# ----------------------------------------------------------------------
	def copy(self):
		sel = self.curselection()
		if not sel: return
		txt = []
		listbox = self.focus_get()
		first = listbox is self.listbox(0)
		for i in sel:
			s = listbox.get(i)
			if first:
				# strip the Unicode prefix
				s = s.lstrip()[2:]
			txt.append(s)
		self.clipboard_clear()
		self.clipboard_append("\n".join(txt))

	# ----------------------------------------------------------------------
	# Recursively fill a DICOM file
	# ----------------------------------------------------------------------
	def _fill(self, rootlist, nesttag, depth=0, prefix=""):
		maxlen = 16

		# --------------------------------------------------------------
		def byteData(value):
			s = [str(value[x]) for x in range(min(len(value),maxlen))]
			if len(value)>maxlen: s.append("...")
			return f"{len(value)} [{', '.join(s)}]"

		# --------------------------------------------------------------
		def wordData(value):
			if value is None: return ""
			s = []
			for i in range(min(len(value)//2,maxlen)):
				s.append(str(value[i*2] + value[i*2]<<8))
			if len(value)//2>maxlen: s.append("...")
			return f"{len(value)//2} [{', '.join(s)}]"

		for i,item in enumerate(rootlist[0]):
			if isinstance(item, pydicom.dataelem.RawDataElement):
				name  = "RAW"
				value = item.value
				tag   = item.tag
				VM    = "-"
				VR    = item.VR
			elif isinstance(item, pydicom.dataelem.DataElement):
				name  = item.description()
				value = item.value
				tag   = item.tag
				VM    = str(item.VM)
				VR    = item.VR
			elif isinstance(item, pydicom.dataset.Dataset):
				items = [x[i] for x in rootlist if x is not None]
				self._fill(items, f"{nesttag}:{i}", depth, f"{i+1}. ")
				continue
			else:
				raise Exception(f"Unknown item type: {type(item)}")

			# Check other values if they are the same
			ntag = f"{nesttag}:{tag.numerator}"
			if isinstance(value, pydicom.sequence.Sequence):
				if self._open.get(ntag,0):
					handle = DicomBrowser._OPEN
				else:
					handle = DicomBrowser._CLOSE
				svalue = ""
			else:
				same = True
				for root in rootlist[1:]:
					try:
						if root[tag.real].value != value:
							same = False
							break
					except KeyError:
						same = False
						break
				handle = DicomBrowser._NONE
				if not same:
					handle = DicomBrowser._MULTI
					svalue = ""
				elif VR == "OB":
					svalue = byteData(value)
				elif VR == "OW":
					svalue = wordData(value)
				else:
					try:
						svalue = str(value)
					except:
						svalue = byteData(value)

			sname = f"{DicomBrowser._INDENT*depth}{handle} {prefix}{name}"
			stag = str(tag)
			if self._filter:
				if self._filterUpper:
					sn = sname.upper()
					sv = svalue.upper()
				else:
					sn = sname
					sv = svalue
				if self._filter in sn or \
				   self._filter in sv or \
				   self._filter in stag:
					self.insert(tk.END, (sname, svalue, stag, VR, VM))
					self._tags.append(ntag)
			else:
				self.insert(tk.END, (sname, svalue, stag, VR, VM))
				self._tags.append(ntag)

			# If open sequence
			if handle == DicomBrowser._OPEN:
				values = [value]
				for root in rootlist[1:]:
					try:
						values.append(root[tag.real].value)
					except KeyError:
						values.append(None)
				self._fill(values, ntag, depth+1)

	# ----------------------------------------------------------------------
	# Fill a dicom
	# ----------------------------------------------------------------------
	def fill(self, dataset=None):
		yview = self.yview()
		del self._tags[:]
		self.delete(0,tk.END)
		if dataset is not None: self.dataset(dataset)
		self._fill(self._dataset,'')
		self.yview_moveto(yview[0])

	# ----------------------------------------------------------------------
	# Open close a sub-element when the button-1 is released
	# ----------------------------------------------------------------------
	def buttonRelease(self, event):
		if len(event.widget.curselection()) != 1: return
		if event.x > event.widget.winfo_width()/5: return
		self._toggle(event.widget.nearest(event.y))

	# ----------------------------------------------------------------------
	# Toggle state expand/collapse of current node
	# ----------------------------------------------------------------------
	def toggle(self, event):
		self._toggle(self.index(tk.ACTIVE))

	# ----------------------------------------------------------------------
	def _toggle(self, i):
		item = self.listbox(0).get(i)
		handle = item.strip()[0]
		if handle == DicomBrowser._NONE: return
		tag = self._tags[i]
		self._open[tag] = bool(handle==DicomBrowser._CLOSE)
		self._changeTag([tag])

	# ----------------------------------------------------------------------
	# Expand current node
	# ----------------------------------------------------------------------
	def expand(self, event=None):
		tkExtra.busy(self)
		tags = []
		changed = False
		for i in map(int, self.curselection()):
			item = self.listbox(0).get(i)
			handle = item.strip()[0]
			tags.append(self._tags[i])
			if handle == DicomBrowser._CLOSE:
				self._open[tags[-1]] = True
				changed = True
		if tags and changed: self._changeTag(tags)
		tkExtra.notBusy(self)
		return "break"

	# ----------------------------------------------------------------------
	# Collapse current or master node
	# ----------------------------------------------------------------------
	def collapse(self, event=None):
		tkExtra.busy(self)
		tags = []
		for i in map(int, self.curselection()):
			item = self.listbox(0).get(i)
			handle = item.strip()[0]
			tag = self._tags[i]
			if handle==DicomBrowser._OPEN:
				self._open[tag] = False
				tags.append(tag)
			else:
				if item[0] != ' ': continue
				# We are in sub nesting find the head node and close it
				last = tag.rfind(":")
				if last==-1: continue
				prev = tag.rfind(":",0,last)
				if prev==-1: continue
				tag = tag[:prev]
				self._open[tag] = False
				tags.append(tag)
		if tags: self._changeTag(tags)
		tkExtra.notBusy(self)
		return "break"

	# ----------------------------------------------------------------------
	def _changeTag(self, tags):
		#active = self.index(tk.ACTIVE)
		#activetag = self._tags[active]
		active = None
		self.fill()
		for tag in tags:
			for i,t in enumerate(self._tags):
				if t.startswith(tag):
					if active is None:
						active = i
					self.selection_set(i)
		if active is not None:
			self.activate(active)
			self.see(active)

	# ----------------------------------------------------------------------
	# Return a list of items for the tag addressed by a path/list of tags
	# for all datasets loaded
	# ----------------------------------------------------------------------
	def findItems(self, tag_path):
		items = []
		for item in self._dataset:
			iterator = iter(tag_path)
			try:
				for tag in iterator:
					parent = item
					item = parent[tag]
					try:
						idx  = next(iterator)
						parent = item
						item = parent[idx]
					except StopIteration:
						break
				items.append((parent,item))
			except KeyError:
				items.append((None,None))
		if all(x[0] is None for x in items): return None
		return items

	# ----------------------------------------------------------------------
	def deleteItem(self, event=None):
		for idx in reversed(self.curselection()):
			#name, value, stag, vr, vm = self.get(active)
			# Find item pointer
			ntag = list(map(int,self._tags[idx].split(":")[1:]))
			items = self.findItems(ntag)
			if items is None: return
			for parent,item in items:
				del parent[item.tag]
				self._changed = True
		self.fill()

	# ----------------------------------------------------------------------
	def inPlaceEdit(self, event=None):
		listbox = event.widget
		if event.type == "2":		# keyboard event
			active = listbox.index(tk.ACTIVE)
		else:
			active = listbox.nearest(event.y)
			self.activate(active)
		name, value, stag, vr, vm = self.get(active)
		try:
			vm = int(vm)
		except ValueError:
			vm = 1

		# Find item pointer
		ntag = list(map(int,self._tags[active].split(":")[1:]))
		items = self.findItems(ntag)
		if items is None: return

		lid = self.listboxIndex(listbox)
		if lid == 1:
			if vr in ("OW","OB","SQ"): return
			edit = tkExtra.InPlaceEdit(listbox)
			if edit.value is None: return
			if vm == 1:
				value = edit.value
			else:
				# split to comma items
				try:
					value = Stream.str2list(edit.value)
				except:
					return
			for parent,item in items:
				if item is None: continue
				item.value = value
				self._changed = True
			listbox.set(active,str(item.value))

	# ----------------------------------------------------------------------
	# Search for a tag
	# @param txt	txt to search
	# @param upper	case insensitive search
	# @return None or tag nesting
	# ----------------------------------------------------------------------
	def filter(self, txt=None, upper=False):
		try:
			active = self._tags[self.index(tk.ACTIVE)]
		except:
			active = None
		if txt is None:
			self._filter = None
		else:
			if upper:
				self._filter = txt.upper()
			else:
				self._filter = txt
			self._filterUpper = upper
		self.fill()
		if active:
			self.selectTag(active)

	# ----------------------------------------------------------------------
	# Search for a tag
	# @param txt	txt to search
	# @param upper	case insensitive search
	# @return None or tag nesting
	# ----------------------------------------------------------------------
	def search(self, txt, upper=False):
		#dump(f"*-* search({txt!r},{upper})")
		sel = self.curselection()
		if len(sel)==1:
			self._skip = self._tags[int(sel[0])]
		else:
			self._skip = None
		if upper: txt = txt.upper()
		return self._search(txt, self._dataset[0], '', upper)

	# ----------------------------------------------------------------------
	# Recursive function to perform the search
	# ----------------------------------------------------------------------
	def _search(self, needle, root, nesttag, upper):
		for i,item in enumerate(root):
			if isinstance(item, pydicom.dataelem.RawDataElement):
				name  = "RAW"
				value = item.value
				tag   = item.tag
			elif isinstance(item, pydicom.dataelem.DataElement):
				name  = item.description()
				value = item.value
				tag   = item.tag
			elif isinstance(item, pydicom.dataset.Dataset):
				pos = self._search(needle, item, f"{nesttag}:{i}", upper)
				if pos is not None: return pos
				continue
			else:
				raise Exception("Unknown item type: "+str(type(item)))

			ntag = f"{nesttag}:{tag.numerator}"
			if isinstance(value, pydicom.sequence.Sequence) or \
			   item.VR=="OB":
				svalue = ""
			else:
				svalue = str(value)

			if self._skip is None:
				if upper:
					name   = name.upper()
					svalue = svalue.upper()
				if needle in name or \
				   needle in svalue or \
				   needle in str(tag):
					return ntag

			elif self._skip == ntag:
				self._skip = None

			# If open sequence
			if isinstance(value, pydicom.sequence.Sequence):
				pos = self._search(needle, item, ntag, upper)
				if pos is not None: return pos
		return None

	# ----------------------------------------------------------------------
	# Open all tags and select the current one
	# ----------------------------------------------------------------------
	def selectTag(self, tag):
		# select to open all parent tags
		parent = ""
		refill = False
		for t in tag.split(":")[1:-1]:
			parent = f"{parent}:{t}"
			if not refill and not self._open.get(parent,False):
				refill = True
			self._open[parent] = True

		if refill: self.fill()

		self.selection_clear(0,tk.END)
		try:
			idx = self._tags.index(tag)
			self.selection_set(idx)
			self.activate(idx)
			self.see(idx)
		except ValueError:
			pass

#===============================================================================
if __name__ == "__main__":
	import os
	import sys
	sys.path.insert(0,os.path.join(os.path.abspath(os.path.dirname(__file__)), "lib"))

	if len(sys.argv) < 2:
		fn = os.path.basename(sys.argv[0])
		print(f"syntax: {fn} <dicom_file>...")		# main
		sys.exit()

	if len(sys.argv)==2:
		fn = sys.argv[1]
		_fn = [fn]
		try:
			_ds = pydicom.read_file(fn, force=True)
		except pydicom.filereader.InvalidDicomError:
			print(f"Error opening file: {fn}")	# main
			sys.exit()
	else:
		_ds = []
		_fn = sys.argv[1:]
		for fn in _fn:
			try:
				_ds.append(pydicom.read_file(fn, force=True))
			except pydicom.filereader.InvalidDicomError:
				print(f"Error opening file: {fn}")	# main
				sys.exit()

	rootwin = tk.Tk(className="flair_dicombrowser")
	icons.load(f"{os.path.dirname(os.path.abspath(__file__))}/icons")
	rootwin.title(f"DICOM Viewer: {' '.join([os.path.basename(x) for x in _fn])}")

	tree = DicomBrowser(rootwin, _ds, _fn, height=30)
	tree.pack(expand=tk.TRUE, fill=tk.BOTH)
	tree.focus_set()
	entry = tkExtra.LabelEntry(rootwin,
			label="Search (Ctrl-f) or Filter (Ctrl-F)",
			labelcolor="Gray")
	entry.pack(side=tk.BOTTOM, fill=tk.X)

	#-----------------------------------------------------------------------
	def searchTree(event=None):
		needle = entry.get()
		if not needle:
			tree.filter(needle)
			entry.focus_set()
			return
		upper = True
		tag = tree.search(needle, upper)
		if tag:
			print("Found:", tag)	# main
			tree.selectTag(tag)

		else:
			print("Not found")	# main

	#-----------------------------------------------------------------------
	def filterTree(event=None):
		needle = entry.get()
		upper = True
		tree.filter(needle, upper)

	entry.bind("<Return>",            searchTree)
	entry.bind("<<LabelEntryClear>>", filterTree)
	entry.bind("<Escape>",            lambda e : tree.focus_set())
	rootwin.bind("<Control-Key-f>",   searchTree)
	rootwin.bind("<Control-Key-F>",   filterTree)
	rootwin.bind("<Control-Key-s>",   lambda e : tree.save())
	rootwin.bind("<Control-Key-q>",   lambda e : rootwin.destroy())

	tree.fill()
	rootwin.mainloop()
	icons.delete()
