#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   23-Jun-2017

import math
import tkinter as tk

try:
	from PIL import Image
	from PIL import ImageDraw
except:
	Image = None

try:
	import numpy as np
except ImportError:
	np = None

import bmath

#-------------------------------------------------------------------------------
# Rescale using the Window, Center from Look Up Table
# Apply the Palette Look-Up Table for the given data and min/max value.
#       <--       width       ->
#       <-----------|---------->
#             center/level
#-------------------------------------------------------------------------------
def rescaleWindow(data, width, level):
	width = float(width)
	level  = float(level)

	# Fix for numpy 1.9.2 error: all arrays must have the same dimension
	below = (data <= (level-0.5-(width-1.0)/2.0))
	above = (data >  (level-0.5+(width-1.0)/2.0))
	totlist = np.logical_or.reduce([below,above], axis=0)

	return np.piecewise(data,
		[below, above, ~totlist],
		[0, 255, lambda x: ((x-(level-0.5))/(width-1.0)+0.5)*255])

#------------------------------------------------------------------------------
# Apply the Palette Look-Up Table for the given data and min/max value.
#-----------------------------------------------------------------------------
def rescale(data, minimum, maximum):
	# Fix for np 1.9.2 error: all arrays must have the same dimension
	below = (data <= minimum)
	above = (data >  maximum)
	totlist = np.logical_or.reduce([below,above], axis=0)

	width = maximum - minimum
	return np.piecewise(data,
		[below, above, ~totlist],
		[0,255, lambda data: ((data-minimum)/width)*255])

#-------------------------------------------------------------------------------
# Convert a DICOM slice to Tk photoimage
#-------------------------------------------------------------------------------
def dicom2image(dataset, array, width=None, center=None, z=0):
	#dump("*-* DicomDraw.dicom2image")

	if 'PixelData' not in dataset:
		raise TypeError("Cannot show image -- DICOM dataset does not have pixel data")

	# Multi frame image
	if 'NumberOfFrames' in dataset:
		try:
			frame = dataset.PerFrameFunctionalGroupsSequence[z]
		except:
			frame = None

		try:
			if width is None or width==0: width = dataset.WindowWidth
		except: pass
		try:
			if center is None or center==-4096: center = dataset.WindowCenter
		except: pass

		if width:
			width = float(width)
		else:
			try: width = float(frame.FrameVOILUTSequence[0].WindowWidth)
			except: width = 0
		if center:
			center = float(center)
		else:
			try: center = float(frame.FrameVOILUTSequence[0].WindowCenter)
			except: center = 0
		try:
			if frame:
				m = float(frame.PixelValueTransformationSequence[0].RescaleSlope)
				b = float(frame.PixelValueTransformationSequence[0].RescaleIntercept)
			else:
				m = dataset.RescaleSlope
				b = dataset.RescaleIntercept
		except AttributeError:
			m = 1	# rescale slope
			b = 0	# rescale intercept

		width /= m
		center = (center-b) / m

		try:
			image = rescaleWindow(array[z], width, center)
		except IndexError:
			return

		# Convert mode to L since LUT has only 256 values
		# http://www.pythonware.com/library/pil/handbook/image.htm
		# when image is not in int16 (e.g. uint16) format the PIL convert doesn't work
		image = Image.fromarray(image.astype(np.int16)).convert('L')

	# can only apply LUT if these values exist
	elif 'WindowWidth' not in dataset or 'WindowCenter' not in dataset:
		bits = dataset.BitsAllocated
		samples = dataset.SamplesPerPixel
		if bits == 8 and samples == 1:
			mode = "L"
		elif bits == 8 and samples == 3:
			mode = "RGB"
		elif bits == 16:
			# not sure about this -- PIL source says is 'experimental'
			# and no documentation. Also, should bytes swap depending
			# on endian of file and system??
			mode = "I;16"
		else:
			raise TypeError(f"Don't know PIL mode for {bits} BitsAllocated "
					f"and {samples} SamplesPerPixel")

		# PIL size = (width, height)
		size = (dataset.Columns, dataset.Rows)

		# Recommended to specify all details by
		# http://www.pythonware.com/library/pil/handbook/image.htm
		image = Image.frombuffer(mode, size,
					dataset.PixelData,
					"raw",
					mode,
					0, 1).convert('L')

	else:
		if width is None or width==0:     width = dataset.WindowWidth
		if center is None or center in (-4096,0): center = dataset.WindowCenter
		try:    width = float(width)
		except: width = float(width[-1])
		try:    center = float(center)
		except: center = float(center[-1])
		try:
			m = dataset.RescaleSlope
			b = dataset.RescaleIntercept
		except AttributeError:
			m = 1	# rescale slope
			b = 0	# rescale intercept
		width /= m
		center = (center-b) / m
		image  = rescaleWindow(array, width, center)
		# Convert mode to L since LUT has only 256 values
		# http://www.pythonware.com/library/pil/handbook/image.htm
		# when image is not in int16 (e.g. uint16)
		# format the PIL convert doesn't work
		image = Image.fromarray(image.astype(np.int16)).convert('L')

	return image

#-------------------------------------------------------------------------------
# Draw an RT structure on image
#-------------------------------------------------------------------------------
def rtstruct2image(dataset, ctds, z, sop, rois, width, center):
	if ctds:
		if "TransferSyntaxUID" not in ctds.file_meta:
			ctds.file_meta.TransferSyntaxUID = \
				ctds.uid.ImplicitVRLittleEndian
		image = dicom2image(ctds, ctds.pixel_array, width, center, z).convert('RGB')
		x0, y0, z0 = map(float,ctds.ImagePositionPatient)
		dx, dy     = map(float,ctds.PixelSpacing)
		iop        = list(map(float,ctds.ImageOrientationPatient))
		x0 -= dx/2	# correct for half a voxel
		y0 -= dy/2
	else:
		image = Image.new("RGB", (512,512))
		x0 = y0 = z0 = 0
		dx = dy = dz = 1.
		iop = [1., 0., 0., 0., 1., 0.]
	try:
		dz = ctds.SliceThickness
	except:
		dz = 1.

	Xx = iop[0]; Xy = iop[1]
	Yx = iop[3]; Yy = iop[4]

	# Find coordinates for transformation of ROI
	draw = ImageDraw.Draw(image)
	for roi in rois:
		# XXX damn it ROI sequence is not incremental can be anything!!
		# Find the sequence idx from the roi
		for sequence in dataset.ROIContourSequence:
			if int(sequence.ReferencedROINumber) == roi:
				break
		else:
			continue

		# Color
		try:
			color = sequence.ROIDisplayColor	# RGB
			color = int(color[2])<<16 | int(color[1])<<8 | int(color[0])	# BGR
		except AttributeError:
			color = 128

		# Find contour if any based on SOP
		if "ContourSequence" not in sequence: continue
		for contour in sequence.ContourSequence:
			try:
				if contour.ContourImageSequence[0].ReferencedSOPInstanceUID == sop \
				   and contour.ContourGeometricType == "CLOSED_PLANAR" \
				   and contour.NumberOfContourPoints > 1:
					npoints = contour.NumberOfContourPoints
					coords = list(map(float, contour.ContourData))
					XY = zip(coords[::3], coords[1::3])
					iXY = [(round((Xx*(x-x0)+Xy*(y-y0))/dx),
						round((Yx*(x-x0)+Yy*(y-y0))/dy))
							for x,y in XY]
					draw.polygon(iXY, outline=color)
			except AttributeError:
				if contour.ContourGeometricType == "POINT" \
				   and abs(float(contour.ContourData[2]) - z) <= dz/2:
					npoints = contour.NumberOfContourPoints
					coords = list(map(float, contour.ContourData))
					if npoints == 1:
						x = coords[0]
						y = coords[1]
						iX = round((Xx*(x-x0)+Xy*(y-y0))/dx)
						iY = round((Yx*(x-x0)+Yy*(y-y0))/dy)
						draw.line((iX,iY-5,iX,iY+5), fill=color)
						draw.line((iX-5,iY,iX+5,iY), fill=color)
	del draw
	return image

# ----------------------------------------------------------------------
# Draw line for every pixel that it is intercepted, accepting
# floating points as pixel position
# FIXME move to some external PIL functions module
# ----------------------------------------------------------------------
def drawLine(image, x1, y1, x2, y2, c):
	dx = x2-x1
	dy = y2-y1
	length = math.sqrt(dx**2 + dy**2)
	if length < 1e-10: return
	dirx = dx / length
	diry = dy / length

	i  = int(x1)
	j  = int(y1)

	if dirx > 1e-10:
		tx    = (float(i+1) - x1)/ dirx	# distance to next cell
		tdx   = 1.0 / dirx
		stepx = 1
	elif dirx < -1e-10:
		tx    = (float(i) - x1)/ dirx	# distance to next cell
		tdx   = -1.0 / dirx
		stepx = -1
	else:
		tx    = 1e10
		tdx   = 0.0
		stepx = 0

	if diry > 1e-10:
		ty    = (float(j+1) - y1)/ diry	# distance to next cell
		tdy   = 1.0 / diry
		stepy = 1
	elif diry < -1e-10:
		ty    = (float(j) - y1)/ diry	# distance to next cell
		tdy   = -1.0 / diry
		stepy = -1
	else:
		ty    = 1e10
		tdy   = 0.0
		stepy = 0

	try:
		image.putpixel((i,j),c)
	except IndexError:
		pass
	while tx<=length or ty<=length:
		if tx<ty:
			i += stepx
			tx += tdx
		else:
			j += stepy
			ty += tdy
		if i<0 or i>=image.size[0] or j<0 or j>=image.size[1]: break
		try:
			image.putpixel((i,j),c)
		except IndexError:
			pass

	i2 = int(x2)
	j2 = int(y2)
	try:
		image.putpixel((i2,j2),c)
	except IndexError:
		pass

# ----------------------------------------------------------------------
# Draw orientation axis for dicom
# FIXME todo add option for Biped, Quadruped
# ----------------------------------------------------------------------
def drawOrientation(canvas, iop, cx, cy, size=20, font=("Helvetica","-12")):
	#          0   1   2
	# iop = [ cxx cxy cxz;
	#         cyx cyy cyz ]
	#          3   4   5
	orient = []

	r = bmath.Vector(iop[:3])
	c = bmath.Vector(iop[3:])
	s = r ^ c
	# /dx*i\   / a b c \ / x \
	# |dy*j| = | d e f | | y |
	# \dz*k/   \ s s s / \ z /

	# X:R->L   L=(size,0,0), r=(-size,0,0)
	x,y,z = size,0,0
	L = cx+r[0]*x+r[1]*y+r[2]*z, cy-(c[0]*x+c[1]*y+c[2]*z)
	Lz = s[0]*x + s[1]*y + s[2]*z
	x = -x
	R = cx+r[0]*x+r[1]*y+r[2]*z, cy-(c[0]*x+c[1]*y+c[2]*z)
	Rz = s[0]*x + s[1]*y + s[2]*z

	# Y:A->P   A=(0,size,0), P=(0,-size,0)
	x,y,z = 0,size,0
	A = cx+r[0]*x+r[1]*y+r[2]*z, cy-(c[0]*x+c[1]*y+c[2]*z)
	Az = s[0]*x + s[1]*y + s[2]*z
	y = -y
	P = cx+r[0]*x+r[1]*y+r[2]*z, cy-(c[0]*x+c[1]*y+c[2]*z)
	Pz = s[0]*x + s[1]*y + s[2]*z

	# Z:F->H   H=(0,0,size), F=(0,0,-size)
	x,y,z = 0,0,size
	H = cx+r[0]*x+r[1]*y+r[2]*z, cy-(c[0]*x+c[1]*y+c[2]*z)
	Hz = s[0]*x + s[1]*y + s[2]*z

	z = -z
	F = cx+r[0]*x+r[1]*y+r[2]*z, cy-(c[0]*x+c[1]*y+c[2]*z)
	Fz = s[0]*x + s[1]*y + s[2]*z

	#-----------------------------------------------------------------------
	def isCenter(A,Az,At,B,Bz,Bt):
		nonlocal C, Ct
		if (A[0]-B[0])**2 + (A[1]-B[1])**2 < 5**2:
			if Hz > Fz:
				C = A[0]+2, A[1]+2
				Ct = At
			else:
				C = B[0]+2, B[0]+2
				Ct = Bt

	C = Ct = None
	isCenter(H,Hz,"H",F,Fz,"F")
	isCenter(P,Pz,"P",A,Az,"A")
	isCenter(L,Lz,"L",R,Rz,"R")

	def anchor(M):
		if   M[0] > cx+size/4: return tk.W
		elif M[0] < cx-size/4: return tk.E
		elif M[1] > cy+size/4: return tk.N
		elif M[1] < cy-size/4: return tk.S
		else:                  return tk.SW

	orient.append(canvas.create_line(L+R,
			tag="orientation",
			fill="#FF0000",
			width=3))
	orient.append(canvas.create_line(A+P,
			tag="orientation",
			fill="#00FF00",
			width=3))
	orient.append(canvas.create_line(H+F,
			tag="orientation",
			fill="#0000FF",
			width=3))

	if Ct is None or Ct not in "RL":
		orient.append(canvas.create_text(R,
				text="R",
				fill="Yellow",
				tag="orientation",
				anchor=anchor(R),
				font=font))
		orient.append(canvas.create_text(L,
				text="L",
				fill="Yellow",
				tag="orientation",
				anchor=anchor(L),
				font=font))

	if Ct is None or Ct not in "PA":
		orient.append(canvas.create_text(A,
				text="A",
				fill="Yellow",
				tag="orientation",
				anchor=anchor(A),
				font=font))
		orient.append(canvas.create_text(P,
				text="P",
				fill="Yellow",
				tag="orientation",
				anchor=anchor(P),
				font=font))

	if Ct is None or Ct not in "HF":
		orient.append(canvas.create_text(H,
				text="H",
				fill="Yellow",
				tag="orientation",
				anchor=anchor(H),
				font=font))
		orient.append(canvas.create_text(F,
				text="F",
				fill="Yellow",
				tag="orientation",
				anchor=anchor(F),
				font=font))

	if C is not None:
		orient.append(canvas.create_text(C,
				text=Ct,
				fill="Yellow",
				tag="orientation",
				anchor=tk.SW,
				font=font))
	return orient
