#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Aug-2012

import os
import tkinter as tk

try:
	import pydicom
except ImportError:
	try:
		import dicom as pydicom
	except ImportError:
		pydicom = None

import icons
import Ribbon
import tkFlair
import tkExtra
import Project
import bFileDialog
import FlairRibbon
import DicomBrowser

#===============================================================================
# Dicom Browser Tab
#===============================================================================
class DicomEditorTab(FlairRibbon.FlairTab):
	#----------------------------------------------------------------------
	def __init__(self, master, page, **kw):
		super().__init__(master, page, "Editor", **kw)
		# -----
		self.browser = DicomBrowser.DicomBrowser(self)
		self.browser.pack(fill=tk.BOTH, expand=tk.YES)
		self.browser.bindList("<Control-Key-f>", self.findFocus)
		self.browser.bindList("<Control-Key-g>", self.find)

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		self.contextGroup = {}

		# ========== Browser ===========
		group = Ribbon.LabelGroup(self.ribbon, "Browser")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.grid3rows()

		# ---
		col,row = 0,0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("save32"),
				anchor=tk.CENTER,
				command=self.save)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Save dicom file")

		row = 2
		b = Ribbon.LabelButton(group.frame,
				text="Save",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.CENTER,
				command=self.saveAs)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Save dicom file with a different name")

		# ---
		col += 1
		row = 1
		self.findString = tkExtra.LabelEntry(group.frame,
					"Search",
					"DarkGray",
					background="White",
					font=Ribbon._FONT,
					width=20)
		self.findString.grid(row=row, column=col, padx=0, pady=0) #, sticky=tk.NSEW)
		tkExtra.Balloon.set(self.findString , "Search string [Ctrl-F, Ctrl-G]")
		self.findString.bind("<Return>",        self.find)
		self.findString.bind("<KP_Enter>",      self.find)
		self.findString.bind("<Control-Key-g>", self.find)
		self.findString.bind("<Escape>",        lambda e,s=self: s.browser.focus_set())

		# ---
		col += 1
		self.findCase = tk.BooleanVar()
		self.findCase.set(False)
		b = tk.Checkbutton(group.frame,
				variable=self.findCase,
				padx=0,
				pady=0,
				borderwidth=0,
				highlightthickness=0,
				font=Ribbon._FONT,
				activebackground="LightYellow",
				background=Ribbon._BACKGROUND)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Case sensitive")

		# ---
		col += 1
		row = 0
		b = Ribbon.LabelButton(group.frame,
				text = "Search",
				image=icons.get("find"),
				command=self.find)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.W)
		tkExtra.Balloon.set(b, "Find cards with string")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text = "Filter",
				image=icons.get("filter"),
				command=self.filterBrowser)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.W)
		tkExtra.Balloon.set(b, "Filter: display only cards matching the search string")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text = "Clear",
				image=icons.get("clean"),
				command=self.filterClear)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.W)
		tkExtra.Balloon.set(b, "Filter clear")

		# ---
		col += 1
		row = 0
		b = Ribbon.LabelButton(group.frame,
				text="Viewer",
				image=icons.get("viewer32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.viewer)
		b.grid(row=row, rowspan=3, column=col, padx=0, pady=0, sticky=tk.W)
		tkExtra.Balloon.set(b, "Open in internal viewer")

		self.dataset     = None			# current loaded dataset
		self._loadedFile = None			# currently loaded file
		self._selFiles   = 0			# number of selected files

	# ----------------------------------------------------------------------
	def dicom(self):	return self.page.dicom

	# ----------------------------------------------------------------------
	def refresh(self):
		di = self.page.dicominfo
		if di is None:
			self.browser.clear()
			return

	# ----------------------------------------------------------------------
	# Select a new dataset and populate the listbox
	# ----------------------------------------------------------------------
	def activate(self):
		self.refresh()
		#if self.dataset is not None and self.dataset.SeriesInstanceUID == self.dicom().uid:
		#	return
		if self.page.dicomfile:
			self.loadSlice(self.page.dicomfile)

	# ----------------------------------------------------------------------
	# Load slice
	# ----------------------------------------------------------------------
	def loadSlice(self, filename):
		if isinstance(filename,list):
			fn = filename[0]
			n  = len(filename)
		else:
			fn = filename
			n  = 1
		if fn == self._loadedFile and self._selFiles == n: return

		if self.browser.changed():
			if len(self.browser.filenames())==1:
				msg = "Slice is modified. Do you want to save it?"
			else:
				msg = "Slices are modified. Do you want to save them?"
			if tkFlair.askyesno("Warning", msg, parent=self):
				self.save()

		# remember position
		yview = self.browser.yview()
		self.browser.clear()
		try:
			self.dataset = pydicom.read_file(fn, force=True)
			self._loadedFile = fn
			self._selFiles = n
		except pydicom.filereader.InvalidDicomError:
			self._loadedFile = None
			self._selFiles = 0
			return

		# update browser
		if n>1:
			datasets  = [self.dataset]
			filenames = [fn]
			for fn in filename[1:]:
				try:
					datasets.append(pydicom.read_file(fn, force=True))
					filenames.append(fn)
				except pydicom.filereader.InvalidDicomError:
					continue
		else:
			datasets  = [self.dataset]
			filenames = [fn]
		self.browser.dataset(datasets)
		self.browser.filenames(filenames)
		self.browser.fill()

		self.browser.yview_moveto(yview[0])
		txt = self.findString.get()
		if txt: self.search(txt)

	# ----------------------------------------------------------------------
	def save(self):
		if self.browser.save(Project.keepBackup):
			self.flair.notify("Dicom saved",
				f"{len(self.browser.dataset())} dicom files saved")
		else:
			self.flair.notify("Dicom save error",
				"Error saving dicom files",
				tkFlair.NOTIFY_ERROR)

	# ----------------------------------------------------------------------
	def saveAs(self):
		newfilenames = []
		for fn in self.browser.filenames():
			filename = bFileDialog.asksaveasfilename(master=self,
						title=f"Save slice {fn} as",
						initialdir=os.path.dirname(fn),
						initialfile=os.path.basename(fn),
						filetypes=[
							("Dicom files",("*.dcm")),
							("All","*")])
			if filename:
				newfilenames.append(filename)
			else:
				return
		self.browser.filenames(newfilenames)
		self.save()

	# ----------------------------------------------------------------------
	def findFocus(self, event=None):
		self.findString.focus_set()

	# --------------------------------------------------------------------
	def find(self, event=None):
		txt = self.findString.get()
		self.search(txt, not self.findCase.get())

	# --------------------------------------------------------------------
	# Search inside dataset information header
	# --------------------------------------------------------------------
	def search(self, needle, upper=False):
		tag = self.browser.search(needle, upper)
		if tag:
			self.browser.selectTag(tag)
		else:
			self.flair.notify("Not found",
				f"'{needle}' not found",
				tkFlair.NOTIFY_WARNING)

	# ----------------------------------------------------------------------
	def filterBrowser(self, event=None):
		txt = self.findString.get()
		self.browser.filter(txt, not self.findCase.get())

	# ----------------------------------------------------------------------
	def filterClear(self, event=None):
		self.findString.set("")
		self.browser.filter()

	# --------------------------------------------------------------------
	# Open file in viewer
	# --------------------------------------------------------------------
	def viewer(self, event=None):
		lst = []
		if self.dicom().modality == "RTSTRUCT":
			lst.append(("RTSTRUCT", str(self.dataset)))
		else:
			try:
				dataset = pydicom.read_file(
						self._loadedFile,
						force=True)
				lst.append((self._loadedFile, str(dataset)))
			except pydicom.filereader.InvalidDicomError:
				pass

		if lst:
			self.flair.viewer(*lst)
