#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Aug-2012

import os
import math
import time
import tempfile
import tkinter as tk
from collections import defaultdict

import icons
import tkFlair
import tkExtra
import tkDialogs
import bFileDialog

import deck
import Dicom
import Global
import Ribbon
import Project
import FlairRibbon
from fluka import FlukaCard
from usrbin2rtdose import usrbin2rtdose

try:
	import pydicom
except ImportError:
	try:
		import dicom as pydicom
	except ImportError as err:
		pydicom = None
		pydicom_err = str(err)
try:
	import DicomInfoTab
	import DicomSliceTab
	import DicomVoxelTab
	import DicomEditorTab
	import RTPlanTab
	import RTViewerTab
except ImportError as err:
	pydicom = None
	pydicom_err = str(err)

#===============================================================================
# Dicom Frame class
#===============================================================================
class DicomPage(FlairRibbon.FlairPage):
	"""Generate Voxel and USRBINs from DICOM files CT, PET, RT..."""
	_name_ = "Dicom"
	_icon_ = "dicom"

	#-----------------------------------------------------------------------
	# Page variables
	#-----------------------------------------------------------------------
	def init(self):
		#Global.dump("*-* DicomPage.init")
		# XXX Warning all dicominfo/file/instances
		# should be of the same modality
		self.dicominfo = None		# selected dicominfo
		self.dicomfile = None		# selected dicomfile(s)
		self.dicominst = None		# selected dicom instances
		self._lastHash = 0

	# ----------------------------------------------------------------------
	@property
	def dicom(self):	return self.dicominfo.dicom

	#-----------------------------------------------------------------------
	def reset(self):
		Global.dump("*-* DicomPage.reset")
		self._lastHash = 0
		self.dicominfo = None
		self.dicomfile = None
		self.dicominst = None
		self.changeTab(0)

	#-----------------------------------------------------------------------
	def setLimits(self):
		if self.dicominfo and self.dicom:
			# update dicom limits
			self.dicom.setLimits(
					self.dicominfo("xfrom"),
					self.dicominfo("xto"),
					self.dicominfo("yfrom"),
					self.dicominfo("yto"),
					self.dicominfo("zfrom"),
					self.dicominfo("zto"))

	#-----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		if pydicom is None: return

		# ========== Tools ===========
		group = Ribbon.LabelGroup(self.ribbon, "Datasets")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		kw = {
			"compound" : tk.TOP,
			"variable" : self.tabVar,
			"command"  : self.changeTab,
			"padx"     : 6
		}

		# dynamic buttons based on modality selected [*=all]
		self._dynamicButtons = defaultdict(list)

		# ---
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Dicom",
				image=icons.get("dicom32"),
				#image=icons.get("total32"),
				value=0,
				**kw)
		tkExtra.Balloon.set(b, "Information on selected dataset")
		self._dynamicButtons["*"].append(b)

		# ---
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Editor",
				image=icons.get("edit32"),
				value=1,
				**kw)
		tkExtra.Balloon.set(b, "Dicom text editor and browser")
		self._dynamicButtons["*"].append(b)

		# ---
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Viewer",
				image=icons.get("slice32"),
				value=2,
				**kw)
		tkExtra.Balloon.set(b, "Dicom slices viewer")
		self._dynamicButtons["CT"].append(b)
		self._dynamicButtons["PT"].append(b)
		self._dynamicButtons["NM"].append(b)
		self._dynamicButtons["MR"].append(b)
		#self._dynamicButtons["RTDOSE"].append(b)
		self._dynamicButtons["RTIMAGE"].append(b)
		self._dynamicButtons["RTSTRUCT"].append(b)

		# ---
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Voxel",
				image=icons.get("voxel32"),
				value=3,
				**kw)
		tkExtra.Balloon.set(b, "Create VOXEL file from DICOM")
		self._dynamicButtons["CT"].append(b)
		self._dynamicButtons["PT"].append(b)
		self._dynamicButtons["NM"].append(b)
		self._dynamicButtons["MR"].append(b)
		self._dynamicButtons["RTSTRUCT"].append(b)

		# ---
		b = Ribbon.LabelRadiobutton(group.frame,
				text="RTPlan",
				image=icons.get("plan32"),
				value=4,
				**kw)
		tkExtra.Balloon.set(b, "Process RTPlan Data")
		self._dynamicButtons["RTPLAN"].append(b)

		# ---
		b = Ribbon.LabelRadiobutton(group.frame,
				text="RTViewer",
				image=icons.get("xray32"),
				value=5,
				**kw)
		tkExtra.Balloon.set(b, "View RTPlan Dose")
		self._dynamicButtons["RTDOSE"].append(b)

		# ====== basic actions =======
		actionsFrame = tk.Frame(group.frame, background=Ribbon._BACKGROUND)
		actionsFrame.pack(side=tk.RIGHT, fill=tk.BOTH)
		for i in range(3):
			actionsFrame.grid_rowconfigure(i, weight=1)

		col,row=0,0
		b = Ribbon.LabelButton(actionsFrame,
					#image=icons.get("add32"),
					image=icons.get("add"),
					text="Add",
					command=self.add,
					#compound=tk.TOP,
					compound=tk.LEFT,
					anchor=tk.W,
					background=Ribbon._BACKGROUND)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add a dicom set to project")

		# ---
		row += 1
		b = Ribbon.LabelButton(actionsFrame, image=icons.get("x"),
				text="Remove",
				compound=tk.LEFT,
				command=self.remove,
				anchor=tk.W,
				background=Ribbon._BACKGROUND)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Remove dicom set")

		# ----
		row += 1
		b = Ribbon.LabelButton(actionsFrame,
					image=icons.get("import"),
					text="Import",
					compound=tk.LEFT,
					anchor=tk.W,
					command=self.importUSRBIN)
		b.grid(row=row, column=col, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Import USRBIN file as RTDOSE")
		# ---
		#col,row = 1,1
		#b = Ribbon.LabelButton(group.frame, image=icons.get("change"),
		#		text="Change",
		#		compound=tk.LEFT,
		#		command=self.change,
		#		anchor=tk.W,
		#		background=Ribbon._BACKGROUND)
		#b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		#tkExtra.Balloon.set(b, "Change dicom data set")

		# ---
		#col,row = 1,2
		#b = Ribbon.LabelButton(group.frame,
		#		image=icons.get("clone"),
		#		text="Clone",
		#		compound=tk.LEFT,
		#		anchor=tk.W,
		#		command=self.clone,
		#		background=Ribbon._BACKGROUND)
		#b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		#tkExtra.Balloon.set(b, "Clone dicom set")

		self.refreshOn = ("save","project")

	#----------------------------------------------------------------------
	def createContextGroups(self):
		super().createContextGroups()

		# ========== USRBIN ===========
		group = Ribbon.LabelGroup(self.ribbon, "Options")
		self.addContextGroup("USRBIN", group)
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		row,col = 0,0
		menulist = [	#("Voxel
				("Usrbin",      "USRBIN",   self.insertUSRBIN),
				("Usrbin-Bio",  "USRBIN",   self.insertUSRBINbio),
				("Remove",      "x",        self.removeCards)]
		b = Ribbon.MenuButton(group.frame, menulist,
				image=icons.get("input"),
				text="Card \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Insert/Update USRBIN card with transformation "
					"corresponding to the Dataset")

		row += 1
		menulist = [	("Usrbin",   "USRBIN",   self.exportUSRBIN),
				("DVH",      "USR-1D",   self.exportDVH)]
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Export \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("export"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Export PixelData to USRBIN,DVH... file")

		#row += 1
		#b = Ribbon.LabelButton(group.frame,
		#			image=icons.get("import"),
		#			text="Import",
		#			compound=tk.LEFT,
		#			anchor=tk.W,
		#			command=self.importUSRBIN)
		#b.grid(row=row, column=col, sticky=tk.NSEW)
		#tkExtra.Balloon.set(b, "Import USRBIN file")

		self.showContextGroup()

	#-----------------------------------------------------------------------
	# Create Project page
	#-----------------------------------------------------------------------
	def createPage(self):
		super().createPage()

		if pydicom is None:
			t = tk.Message(self.frame,
				text = "Error: " + str(pydicom_err) + """\n\n
pydicom library cannot be loaded
either pydicom is missing or some dependency is missing\n
Website: http://code.google.com/p/pydicom/\n
Install:\n
\tFedora\t# dnf install python3-pydicom
\tUbuntu\t# apt-get install python3-pydicom python3-scipy python3-matplotlib
\tall\t# python -m pip install pydicom
\t\t(Requires the python-setuptools),\n
Extra requirements (name depends on distribution):
\tpython3-numeric, python3-imaging or python3-pillow,
\tpython3-imaging-tk or python3-pillow-tk or python3-pil.imagetk\n
""",
				aspect = 600,
				relief=tk.RAISED)
			t.pack(expand=tk.YES, fill=tk.BOTH)
			b = tk.Button(self.frame,
				text = "Help",
				command=self.help)
			b.pack()
			return

		split = tk.PanedWindow(self.frame,
					orient=tk.HORIZONTAL,
					showhandle=0,
					sashrelief=tk.RAISED,
					sashwidth=4,
					opaqueresize=1)
		split.pack(expand=tk.YES, fill=tk.BOTH)

		# Tree frame
		frame = tk.Frame(split)
		split.add(frame, minsize=50)

		# Tree
		self.tree = tkExtra.Treeview(frame)
		self.tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
		self.tree["columns"] = ("instance",)
		#self.tree.column("#0" width=, minwidth=, stretch=tk.NO)
		self.tree.column("instance", width=50*tkFlair.dpiscale,
						stretch=tk.NO, anchor=tk.E)

		self.tree.heading("#0",       text="Dataset",  anchor=tk.W)
		self.tree.heading("instance", text="Instance", anchor=tk.CENTER)

		self.tree.bind("<<TreeviewSelect>>",	self.changeDataset)

		sb = tkExtra.AutoScrollbar(frame, orient=tk.VERTICAL,
					takefocus=False,
					command=self.tree.yview)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		self.tree["yscrollcommand"] = sb.set

		# --------- Dicom tab ---------
		tabFrame = tk.Frame(split, padx=0, pady=0)
		split.add(tabFrame)

		# --------- Info tab ---------
		self.infoTab = DicomInfoTab.InfoTab(tabFrame, self)
		self.addTab(self.infoTab)	# 0

		# --------- Editor tab ---------
		self.editTab = DicomEditorTab.DicomEditorTab(tabFrame, self)
		self.addTab(self.editTab)	# 1

		# --------- Slice tab ---------
		self.sliceTab = DicomSliceTab.SliceTab(tabFrame, self)
		self.addTab(self.sliceTab)	# 2

		# --------- Voxel tab ---------
		self.voxelTab = DicomVoxelTab.VoxelTab(tabFrame, self)
		self.addTab(self.voxelTab)	# 3

		# -------- RTPlan tab ---------
		self.RTPlanTab = RTPlanTab.RTPlanTab(tabFrame, self)
		self.addTab(self.RTPlanTab)	# 4

		# -------- RTViewer tab ---------
		self.RTViewerTab = RTViewerTab.RTViewerTab(tabFrame, self)
		self.addTab(self.RTViewerTab)	# 5

	# ----------------------------------------------------------------------
	# Update fields
	# ----------------------------------------------------------------------
	def get(self):
		Global.dump("*-* DicomPage.get",level=2)
		if self.page is None or pydicom is None: return
		if self.project is None: return
		if self.dicominfo is None: return
		oldhash = hash(self.dicominfo)
		self.sliceTab.get()
		#self.infoTab.get()
		self.voxelTab.get()
		self.RTPlanTab.get()
		self.RTViewerTab.get()
		#self.RTOptTab.get()
		if hash(self.dicominfo) != oldhash:
			self.modified()

	# ----------------------------------------------------------------------
	# Refresh directory listing
	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		Global.dump("*-* DicomPage.refresh")
		if self.page is None or pydicom is None: return
		if what == "project" or self._lastHash != self._dicomsHash():
			self._lastHash = 0
			self.fill()

		super().refresh(what)	# << refresh tabs

		self.activeTab.refresh()
		self._lastHash = self._dicomsHash()

		if not self.tree.selection() and self.project.dicoms:
			try:
				self.tree.selection_add(self.tree.get_children()[0])
			except IndexError:
				pass

	# ----------------------------------------------------------------------
	def _execute(self):
		self.activeTab.execute()

	# ----------------------------------------------------------------------
	def afterDock(self):
		self._lastHash = 0
		self.refresh()

	# ----------------------------------------------------------------------
	def _dicomsHash(self):
		h = 0
		for di in self.project.dicoms:
			h += hash(di.name)
		return h

	# ----------------------------------------------------------------------
	def fill(self):
		self.tree.delete(*self.tree.get_children())
		i = 0
		while i < len(self.project.dicoms):
			di = self.project.dicoms[i]
			if self.insertDicom(di):
				i += 1
			else:
				del self.project.dicoms[i]

		if not self.project.dicoms:
			self.removeAllTabs()

	# ----------------------------------------------------------------------
	def insertDicom(self, di):
		uid = di.name
		if uid is None: return False
		if di.files is None:
			Dicom.Scan.scan(di("directory")) #, self.page)
			try:
				di["frames"] = Dicom.Scan.frames(uid)
			except KeyError:
				return False
			di.files    = Dicom.Scan.files(uid)
			di.modality = Dicom.Scan.modality(uid)
			directory   = self.project.relpath(Dicom.Scan.path(uid))
			if di("directory") != directory:
				di["directory"] = directory

		# Insert group
		if not self.tree.exists(di.modality):
			modname = Dicom.MODALITY.get(di.modality,"")
			if modname != di.modality:
				modname = f"{di.modality}: {modname}"
			self.tree.insert("", tk.END, di.modality,
					text=modname,
					tag="section",
					open=True)

		name = uid
		if di.modality == "CT":
			try:
				ds = pydicom.read_file(di.filename(), force=True)
				name = ds.StudyDescription
			except:
				pass

		elif len(di.files)==1:
			name = os.path.basename(di.filename())

		item = self.tree.insert(di.modality, tk.END, uid,
					text=name,
					values=di.files[0].instance)
		if len(di.files)>1:
			for fn in di.files:
				self.tree.insert(uid, tk.END,
						text=fn.name,
						values=fn.instance)

		self.tree.tag_configure("section", foreground="DarkRed")
		return item

	# ----------------------------------------------------------------------
	# Add a new dicom set
	# ----------------------------------------------------------------------
	def add(self, event=None):
		result = DicomSelectDialog(self.page, self.project).show()
		if result is None: return

		path = result.pop(0)
		first = None
		for uid in result:
			item = self.addDicomInfo(path, uid)
			if first is None: first = item
		if first:
			self.tree.selection_remove(self.tree.selection())
			self.tree.selection_add(first)
			self.changeDataset()
		self.refresh()

	# ----------------------------------------------------------------------
	# Remove selected data sets
	# ----------------------------------------------------------------------
	def remove(self, event=None):
		sel = self.tree.selection()
		remove = set()
		for item in sel:
			di = self.findDicomInfo(item)[0]
			if di: remove.add(di)
		self.tree.selection_remove(sel)
		for di in list(remove):
			self.project.dicoms.remove(di)
		self.refresh()

	# ----------------------------------------------------------------------
	# Clone selected data sets
	# ----------------------------------------------------------------------
	#def clone(self, event=None):
	#	sel = self.dicomList.curselection()
	#	self.dicomList.selection_clear(0,tk.END)
	#	for i in sel:
	#		header = self.dicomList.get(i)
	#		di = self.project.dicomInfo(header[0])
	#		self.project.dicoms.append(di.clone())
	#		self.dicomList.insert(tk.END, header)
	#		self.dicomList.selection_set(tk.END)
	#		self._lastHash += hash(header[0])
	#	self.changeDataset()

	# ----------------------------------------------------------------------
	# Add a new dicom dataset
	# ----------------------------------------------------------------------
	def addDicomInfo(self, path, uid):
		di = self.project.dicomInfo(uid)
		if di is not None:
			self.flair.notify("Dicom already exists",
				f"Dicom {uid} already loaded",
				tkFlair.NOTIFY_WARNING)
			# FIXME select the existing uid
			return	# already exists selected it

		di = Project.DicomInfo(self.project, uid)
		self.project.dicoms.append(di)
		di["directory"] = self.project.relpath(Dicom.Scan.path(uid))
		di["frames"]    = Dicom.Scan.frames(uid)
		di.files        = Dicom.Scan.files(uid)
		di.modality     = Dicom.Scan.modality(uid)
		item = self.insertDicom(di)
		self._lastHash += hash(uid)
		return item

	#-----------------------------------------------------------------------
	# Remove all tabs
	#-----------------------------------------------------------------------
	def removeAllTabs(self):
		btns = {b for x in self._dynamicButtons.values()
				for b in x}
		for b in btns:
			try:
				b.pack_forget()
			except tk.TclError:
				pass
		self.showContextGroup()

	# ----------------------------------------------------------------------
	# Change selection
	# FIXME it is called twice (selection_remove, selection_add) pairs
	# ----------------------------------------------------------------------
	def changeDataset(self, event=None):
		Global.dump("*-* DicomPage.changeDataset",level=2)
		self.log = self.flair.newLog("Dicom", "Dicom")
		diprev = self.dicominfo
		fnprev = self.dicomfile
		inprev = self.dicominst
		self.get()
		self.selectedDicom()		# find new selected tab
		if self.dicominfo is None or \
		   (diprev == self.dicominfo and
		    fnprev == self.dicomfile and
		    inprev == self.dicominst):
			return

		# Force cleaning of dataset in case of RTDOSE
		if self.dicom and not self.dicom.hasSlices and \
		   (inprev and inprev[0]!=self.dicominst[0]):
			self.dicom.scanned(False)
			self.dicominfo.cache.clear()
			self.dicom.init(self.dicomfile[0])
		else:
			self.loadDicom()
		self.setLimits()

		# Display dynamic buttons depending on the modality
		modality = self.dicominfo.modality if self.dicominfo else None

		if diprev is not None:
			if diprev.modality == modality:
				self.changeTab()	# refresh tab
				return
			try:
				for b in self._dynamicButtons[diprev.modality]:
					b.pack_forget()
			except KeyError:
				pass

		if modality is None:
			for b in self._dynamicButtons["*"]:
				b.pack_forget()
			self.changeTab(0)
		else:
			shown = False
			try:
				for b in self._dynamicButtons["*"]:
					if b["text"] == self.activeTab.name: shown = True
					b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
					#print(">>>",b["text"])
				for b in self._dynamicButtons[modality]:
					if b["text"] == self.activeTab.name: shown = True
					b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
					#print(">>>",b["text"])
			except KeyError:
				pass
			if shown:
				self.changeTab()	# force refreshing of tab
			else:
				self.changeTab(0)	# go to default

	# ----------------------------------------------------------------------
	def selectedDicom(self):
		sel = self.tree.selection()
		if not sel:
			Global.dump("*-* DicomPage.selectedDicom1",level=2)
			self.dicominfo = None
			self.dicomfile = None
			self.dicominst = None
		else:
			Global.dump("*-* DicomPage.selectedDicom2",level=2)
			self.dicominfo, fn, inst = self.findDicomInfo(sel[0])
			self.dicomfile = [fn]
			self.dicominst = [inst]
			if len(sel)>1:
				for item in sel[1:]:
					di, df, ins = self.findDicomInfo(item)
					if di != self.dicominfo:
						Global.dump("*-* DicomPage.selectedDicom3",level=2)
						self.dicominfo = None
						self.dicomfile = None
						self.dicominst = None
						break
					self.dicomfile.append(df)
					self.dicominst.append(ins)
		return self.dicominfo

	# ----------------------------------------------------------------------
	# Select file by id or name for the current dicominfo
	# ----------------------------------------------------------------------
	def selectFile(self, fileid):
		if isinstance(fileid,int):
			# find item
			child = self.tree.get_children(self.dicominfo.name)[fileid]
			self.tree.selection_remove(self.tree.selection())
			self.tree.selection_add(child)
			if self.tree.item(self.dicominfo.name,"open"):
				self.tree.see(child)
		else:
			raise Exception("Not implemented yet")

	# ----------------------------------------------------------------------
	# @return number of selected file from current dicominfo
	# ----------------------------------------------------------------------
	def selectedFile(self):
		sel = self.tree.selection()
		if not sel: return 0
		return self.tree.index(sel[0])

	# ----------------------------------------------------------------------
	# Find project dicominfo of tree item
	# ----------------------------------------------------------------------
	def findDicomInfo(self, item):
		if not self.tree.parent(item):
			item = self.tree.get_children(item)[0]
		di = self.project.dicomInfo(item)
		if di is not None:
			return di,di.filename(),di.files[0].instance

		# remember initial filename
		fn   = self.tree.item(item,"text")
		inst = int(self.tree.item(item,"values")[0])

		# search all parents
		while item:
			item = self.tree.parent(item)
			di = self.project.dicomInfo(item)
			if di is not None:
				return di,di.filename(fn),inst
		return None,None,None

	# ----------------------------------------------------------------------
	# Load dicom structure
	# @return True on success (or already existing)
	#	  False otherwise
	# ----------------------------------------------------------------------
	def loadDicom(self):
		if self.dicominfo is None:
			if self.project.dicoms:
				self.flair.notify("DICOM select",
					"Please select only ONE DICOM dataset")
				return "No dicom selected"
			return
		if self.dicominfo.dicom is not None: return
		self.dicominfo.dicom = dicom = Dicom.Dicom()
		dicom.setLog(self.log)
		oldSize = self.log.size()	# remember size before loading
		error = dicom.initDicomInfo(self.dicominfo)
		if isinstance(error,str) or self.log.size() != oldSize:
			if isinstance(error,str): self.log(error)
			self.flair.notify("Errors during loading",
				"Error or warning messages appeared while loading dicom slices. "
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR,
				"Output", self.flair.showOutput,
				showOnOutput=False)
		return error

	# ----------------------------------------------------------------------
	# Analyze slices to find unique Hounsfield units
	# first call loadDicom()
	# FIXME could be made as staticmethod, self we need only for the window
	# ----------------------------------------------------------------------
	def analyzeSlices(self, fn=None, dicominfo=None):
		Global.dump("*-* DicomPage.analyzeSlices")
		if dicominfo is None: dicominfo = self.dicominfo
		dicom = dicominfo.dicom
		if dicom is None or not dicom.hasPixel or dicom.scanned(): return

		self.setStatus("Analyzing slices", "DarkBlue")
		t = time.time()
		if dicominfo("frames") == 0:
			self.resetPercent(len(dicominfo.files))
			for i,(fn,z,sop) in enumerate(dicominfo.files):
				if not dicom.scanSlice(dicominfo.filename(fn),i):
					dicom.scanned(False)	# clear scan data
					self.resetPercent()
					return f"loading slice {fn}"
				self.setStatus(f"Analyzing slice: {fn}", "DarkBlue")
				self.setStatusPercent(i, update=True)
		else:
			# Unique slice dicom
			if fn is None:
				fn = dicominfo.filename()
			self.resetPercent(dicom.nz, True)
			self.setStatus(f"Slice: {fn}", "DarkBlue")
			def update(n): return self.setStatusPercent(n, True)

			if not dicom.scanSlice(fn, update=update):
				dicom.scanned(False)	# clear scan data
				self.resetPercent()
				return f"Error scanning slice {fn}"

		# End scanning
		dicom.scanSliceEnd()
		self.resetPercent()

		t = time.time() - t
		self.flair.log(f"Analyzing DICOM slices finished in {t:.1f}s")
		return None

	# ----------------------------------------------------------------------
	# For RTSTRUCT return a dictionary with sop and CT filename
	# ----------------------------------------------------------------------
	def rtstructReferencedCT(self, dataset=None):
		if dataset is None:
			try:
				dataset = pydicom.read_file(self.dicominfo.filename(), force=True)
			except pydicom.filereader.InvalidDicomError:
				return None
		for cnt in dataset\
				.ReferencedFrameOfReferenceSequence[0]\
				.RTReferencedStudySequence[0]\
				.RTReferencedSeriesSequence[0]\
				.ContourImageSequence:
			ct = self.project.dicomInfoFromSOPInstanceUID(cnt.ReferencedSOPInstanceUID)
			if ct is not None: return ct
		return None

	# ----------------------------------------------------------------------
	# Create a USRBIN card compatible with the dicom
	# ----------------------------------------------------------------------
	def insertUSRBIN(self, event=None):
		self._insertUSRBIN(["DOSE"])

	#-----------------------------------------------------------------------
	def insertUSRBINbio(self, event=None):
		self._insertUSRBIN(["DOSE-H2O","ALPHA-D","SQBETA-D"])
	#	pos = self._addCard(
	#		FlukaCard("RAD-BIOL", ["V79_MCTPS"]),pos,undoinfo)
	#	pos = self._addCard(
	#		FlukaCard("TPSSCORE", ["BIOTOBIN", 1,"","","Dose-H2O",
	#			"Sqbeta-D"]),pos,undoinfo)
	#	return pos

	#-----------------------------------------------------------------------
	def _insertUSRBIN(self, score):
		if self.dicom is None: return
		dicom = self.dicom
		#self.get()
		undoinfo = self.flair.undolistRefresh()

		inp = self.flair.project.input
		if not inp.cardlist: inp.minimumInput()

		# dicom flag
		name = f"{dicom.modality}-{dicom.instance}"

		# delete previously existing cards referring to that name
		# USRBIN, ROT-DEFI, ROTPRBIN
		posdict = {}
		for card in reversed(self.project.input.cardlist):
			if card.tag in ("USRBIN","ROT-DEFI","ROTPRBIN") and card["dicom"]==name:
				posdict[card.tag] = pos = card.pos()
				for t,p in posdict.items():
					if p>pos: posdict[t] -= 1
				undoinfo.append(self.flair.delCardUndo(card.pos(), False))

		#---------------------------------------------------------------
		def bestPos(tag):
			nonlocal posdict, inp
			try:
				return posdict[tag]
			except KeyError:
				return inp.bestPosition(tag)

		# --- USRBIN ---
		for s in score:
			score_name = f"{dicom.instance}-{s}"

			whats = [score_name,
				10.0,	# Cartesian mesh
				s,
				-99.0,	# File
				# XXX embed translation in the USRBIN
				# correct in the ROT-DEFI if needed
				dicom.xmax()/10, #dicom.nx*dicom.dx/10.0,
				dicom.ymax()/10, #dicom.ny*dicom.dy/10.0,
				dicom.zmax()/10, #dicom.nz*dicom.dz/10.0,

				dicom.xmin()/10, #0,
				dicom.ymin()/10, #0,
				dicom.zmin()/10, #0,

				dicom.nx,
				dicom.ny,
				dicom.nz]

			# add card to input
			card = FlukaCard("USRBIN",whats)
			card["dicom"] = name
			card["pos"]   = dicom.origin()/10
			card["iop"]   = dicom.iop
			pos = bestPos(card.tag)
			undoinfo.append(self.flair.addCardUndo(card, pos, False))

			if not dicom.iopIsXYZ():
				# --- ROTPRBIN ---
				what = ["", "", f"-{name}", "", score_name]
				card = FlukaCard("ROTPRBIN", what)
				card["dicom"] = name
				pos = bestPos(card.tag)
				undoinfo.append(self.flair.addCardUndo(card,pos,False))

		if not dicom.iopIsXYZ():
			# --- ROT-DEFI ---
			# contains only the rotation IOP
			pos = bestPos("ROT-DEFI")
			# Add the new transformations
			R = round(math.degrees(dicom.Rz),6)
			if R!=0.0:	# rotation around Z
				what = [name, "", "", R]
				card = FlukaCard("ROT-DEFI", what)
				card["dicom"] = name
				undoinfo.append(self.flair.addCardUndo(card,pos,False))
				pos += 1
			R = round(math.degrees(dicom.Ry),6)
			if R!=0.0:	# rotation around Y
				what = [name, 200.0, "", R]
				card = FlukaCard("ROT-DEFI", what)
				card["dicom"] = name
				undoinfo.append(self.flair.addCardUndo(card,pos,False))
				pos += 1
			R = round(math.degrees(dicom.Rx),6)
			if R!=0.0:	# rotation around X
				what = [name, 100.0, "", R]
				card = FlukaCard("ROT-DEFI", what)
				card["dicom"] = name
				undoinfo.append(self.flair.addCardUndo(card,pos,False))
				pos += 1

			# Add final translation
			#xo,yo,zo = dicom.origin()/10.
			#what = [name, 100.0, "", "", xo, yo, zo]
			#card = FlukaCard("ROT-DEFI", what)
			#card["dicom"] = name
			#undoinfo.append(self.flair.addCardUndo(card,pos,False))
			#pos += 1

		# Finalize
		self.flair.addUndolistRefresh(undoinfo,f"Added USRBIN cards {name!r}")
		self.flair.refresh("card")
		self.flair.notify(f"Add USRBIN: {name!r}",
			f"A USRBIN,ROT-DEFI,ROTPRBIN named {name!r} cards "
			"corresponding to the dicom file has been added to your input")

	# ----------------------------------------------------------------------
	def removeCards(self, event=None):
		name = f"{self.dicom.modality}-{self.dicom.instance}"
		undoinfo = self.flair.undolistRefresh()
		for card in reversed(self.project.input.cardlist):
			if card["dicom"]==name:
				undoinfo.append(self.flair.delCardUndo(card.pos(), False))
		self.flair.addUndolistRefresh(undoinfo, "Delete dicom inserted cards")
		return True

	# ----------------------------------------------------------------------
	# Create usrbin from pixel data
	# ----------------------------------------------------------------------
	def exportUSRBIN(self, filename=None):
		if self.dicom is None: return False
		if not self.dicom.slices:
			self.flair.notify("No DICOM processed",
				"Please select a DICOM dataset",
				tkFlair.NOTIFY_WARNING)
			return False

		if filename is None and len(self.dicomfile)==1:
			filename = self.dicominfo("usrbin")
			if not filename:
				fn,ext = os.path.splitext(self.dicomfile[0])
				filename=f"{fn}.bnn"
			newfilename = bFileDialog.asksaveasfilename(
					master=self.page,
					title="Save as Usrbin file",
					filetypes=[("Usrbin files",("*.bnn","*usrbin*")),
						("All","*")],
					initialfile=filename)
			if not newfilename: return False
			if newfilename != filename:
				filename = newfilename
				self.dicominfo["usrbin"] = self.project.relpath(filename)

			t0 = time.time()
			self._exportUSRBIN(self.dicom, filename)
			dt = time.time()-t0
			self.flair.notify("DICOM to USRBIN",
					f"DICOM to Usrbin file {os.path.basename(filename)!r}"
					f"conversion finish in {dt:g} s")
		else:
			t0 = time.time()
			filenames = []
			for fn in self.dicomfile:
				self.dicom.scanned(False)
				self.dicom.init(fn)
				self.dicom.calcLimits()
				self.analyzeSlices(fn)
				filename=f"{os.path.splitext(fn)}.bnn"
				self._exportUSRBIN(self.dicom, filename)
				filenames.append(os.path.basename(filename))
			dt = time.time()-t0
			self.dicom.scanned(False)
			self.dicominfo.cache.clear()
			self.flair.notify("DICOM to USRBIN",
					f"DICOM to Usrbin file {' '.join(filenames)!r}"
					f"conversion finish in {dt:g} s")
		self.flair.log(f"DICOM to Usrbin file conversion finish in {dt:g} s")

	#-----------------------------------------------------------------------
	# Export a single file
	#-----------------------------------------------------------------------
	def _exportUSRBIN(self, dicom, filename):
		# Do the real processing
		self.log(f"Creating USRBINfile {filename}")
		dlg = tkDialogs.ProgressDialog(self.page, "Creating USRBIN file")

		if not dicom.writeUsrbinInit(filename):
			self.flair.notify("DICOM to USRBIN",
				"Cannot open output Usrbin file",
				tkFlair.NOTIFY_ERROR)
			return False

		if dicom.hasSlices:
			# Multiple slices
			dlg.setLimits(0, len(dicom.slices))
			for i,(l,fn,sop) in enumerate(dicom.slices):
				# Show progress
				if dlg.show(i, "Processing: "+os.path.basename(fn)):
					return False
				dicom.writeUsrbinSlice(fn,i)
		else:
			# Unique 3D data
			fn = dicom.slices[0][1]
			# Show progress
			dlg.setLimits(0.0, dicom.nz)
			dlg.show(0, "Processing: "+os.path.basename(fn))
			def update(s): return dlg.show(s)
			dicom.writeUsrbinSlice(fn, update=update)
		dlg.stop()

		if not dicom.writeUsrbinEnd():
			self.flair.notify("DICOM to USRBIN",
				"Error writing USRBIN file",
				tkFlair.NOTIFY_ERROR)
			return False
		return True

	# ----------------------------------------------------------------------
	# Create DVH from pixel data
	# ----------------------------------------------------------------------
	def exportDVH(self, filename=None):
		if self.dicom is None: return False
		if not self.dicom.slices:
			self.flair.notify("No DICOM processed",
					"Please select a DICOM dataset",
					tkFlair.NOTIFY_WARNING)
			return False

		if filename is None and len(self.dicomfile)==1:
			try:
				filename = self.dicominfo.filename(self.dicominfo["dvh"])
			except KeyError:
				fn,ext = os.path.splitext(self.dicomfile[0])
				filename=f"{fn}_dvh.lis"
			newfilename = bFileDialog.asksaveasfilename(
					master=self.page,
					title="Save as Usrbin file",
					filetypes=[("DVH data list files","*_dvh.lis"),
						("Data list files",("*_tab.lis","*.dat","*.hist")),
						("All","*")],
					initialfile=filename)
			if not newfilename: return False
			if newfilename != filename:
				filename = newfilename
				self.dicominfo["dvh"] = self.project.relpath(filename)
			self._exportDVH(self.dicom, filename)
		else:
			filenames = []
			for fn in self.dicomfile:
				self.dicom.scanned(False)
				self.dicom.init(fn)
				self.dicom.calcLimits()
				self.analyzeSlices(fn)
				filename=f"{os.path.splitext(fn)}_dvh.lis"
				self._exportDVH(self.dicom, filename)
				filenames.append(os.path.basename(filename))
			self.dicom.scanned(False)
			self.dicominfo.cache.clear()

	# ----------------------------------------------------------------------
	# Export dicom as DVH, creates a temporary usrbin file
	# ----------------------------------------------------------------------
	def _exportDVH(self, dicom, filename, notify=None):
		# use a temporary name
		usrbinfile = tempfile.mktemp(".bnn")
		self._exportUSRBIN(dicom, usrbinfile)

		# Find ct and voxel file
		ct = [x for x in self.project.dicoms
				if x.modality=="CT"]
		ct = ct[0] if ct else None
		if ct is None or not os.path.isfile(ct("voxel")):
			self.flair.notify("Voxel file not found",
				"Please load a CT dataset and create the VOXEL file first",
				tkFlair.NOTIFY_WARNING)
			return False
		voxelfile = ct("voxel")

		self.flair.log("DVH creation started")
		self.setStatus("Creating DVH file", "DarkBlue")
		prg = deck.DeckConfig.command("usrbin2dvh")
		xo,yo,zo = ct.dicom.origin() / 10
		cmd = [prg,
			"-o", filename,
			"-u", usrbinfile,
			"-d", "1",
			"-v", voxelfile,
			"-x", str(xo),
			"-y", str(yo),
			"-z", str(zo),
			#"--ux",
			#"--uy",
			#"--uz",
			#"--vx",
			#"--vy",
			#"--vz"
			]
		process = Dicom.DVHProcess()
		process.notify = notify
		process.remove = True
		self.flair.addProcess(process, self.log, cmd)
		return True

	# ----------------------------------------------------------------------
	# Import USRBIN file as RTDOSE
	# ----------------------------------------------------------------------
	def importUSRBIN(self):
		name = self.project.name()
		usrbin_file = bFileDialog.askopenfilename(master=self.page,
				title="Load USRBIN file",
				initialdir=self.project.dir(),
				filetypes=[
					("Fluka USRBIN files",("*usrbin*","*.bnn")),
					(f"{name} USRBIN files", f"{name}*usrbin*"),
					("Fluka fort files",("*_ftn.*","*_fort.*")),
					("All","*")])
		if not usrbin_file: return

		if self.dicomfile is not None and len(self.dicomfile)==1:
			dicom_path = self.dicominfo("directory")
			rtdose_file = self.dicomfile[0]
			rtdose = pydicom.read_file(rtdose_file, force=True)
			rtplan_file = None
			if rtdose.Modality != "RTDOSE":
				rtdose_file = None
			else:
				try:
					refplan_uid = rtdose.ReferencedRTPlanSequence[0].\
							ReferencedSOPInstanceUID
					rtplan = self.project.dicomInfoFromSOPInstanceUID(refplan_uid)
					if rtplan is not None:
						rtplan_file = rtplan.filename()
				except AttributeError:
					pass
			del rtdose
		else:
			dicom_path  = self.project.dir()
			rtdose_file = None
			rtplan_file = None

		fn = os.path.splitext(os.path.basename(usrbin_file))[0]
		if rtdose_file is None:
			output_file = os.path.join(dicom_path, f"{fn}.dcm")
		else:
			output_file = os.path.join(os.path.dirname(rtdose_file), f"{fn}.dcm")

		try:
			uid = usrbin2rtdose(usrbin_file, 0,
					rtdose_file,
					rtplan_file,
					output_file)
		except usrbin2rtdose.RTException as ex:
			self.flair.notify("Error converting to RTDOSE",
				f"There was a problem converting to RTDOSE\n{ex}",
				tkFlair.NOTIFY_ERROR,
				"Output", self.flair.showOutput)
		else:
			self.flair.notify("Converted to RTDOSE",
				f"Usrbin file {self.project.relpath(usrbin_file)!r}\n"
				f"Converted to {self.project.relpath(output_file)!r}\n",
				tkFlair.NOTIFY_NORMAL,
				"Output", self.flair.showOutput)

			# check if the uid is already present in the project else add it
			Dicom.Scan.scan(dicom_path,force=True)
			di = self.project.dicomInfo(uid)
			if di is None:
				self.addDicomInfo(dicom_path, uid)


#===============================================================================
# Dicom Select Dialog
#===============================================================================
class DicomSelectDialog(tk.Toplevel):
	width  = -1
	height = -1
	depth  =  1

	#-----------------------------------------------------------------------
	def __init__(self, master, project, path="", uid=None):
		super().__init__(master)
		self.title("Select DICOM data set")
		self.transient(master)

		# -------------- Directory ---------------
		tk.Label(self, text="Directory:").grid(
				row=0, column=0, pady=1, sticky=tk.E)
		self.path = tk.Label(self,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				#background=tkFlair._ILABEL_BACKGROUND_COLOR,
				background="LightYellow",
				relief=tk.SUNKEN, anchor=tk.W)
		self.path.grid(row=0, column=1, sticky=tk.EW)
		self.path.bind('<Button-1>', self.chdir)
		tkExtra.Balloon.set(self.path, "Dicom directory")

		b = tk.Button(self, image=icons.get("load"), pady=0,
				command=self.chdir)
		b.grid(row=0, column=2)
		tkExtra.Balloon.set(b, "Change Dicom directory")

		tk.Label(self, text="Depth:").grid(
				row=0, column=3, pady=1, sticky=tk.E)
		self.depth = tk.Spinbox(self, from_=0, to_=5, width=5,
				command=self.scan)
		self.depth.delete(0,tk.END)
		self.depth.insert(0, str(DicomSelectDialog.depth))
		self.depth.grid(row=0, column=4)
		tkExtra.Balloon.set(self.depth, "How many subdirectory levels to scan")

		# -------------- Datasets ---------------
		self.dicomList = tkExtra.MultiListbox(self,
				(('Series Instance UID', 30, None),
				 ('Modality',             8, None),
				 ('Date',                12, None),
				 ('Patient',             20, None),
				 ('Age',                  4, None),
				 ('Sex',                  3, None),
				 ('Institution',         10, None),
				 ('Study',               15, None)),
				height=10)
		self.dicomList.grid(row=1, column=0, columnspan=5, sticky=tk.NSEW)
		self.dicomList.bindList("<Double-1>", self.ok)
		self.dicomList.bindList("<Return>",   self.ok)

		frame = tk.Frame(self)
		frame.grid(row=2, column=0, columnspan=5, sticky=tk.NSEW)
		b = tk.Button(frame, text="Cancel",
				image=icons.get("x"),
				compound=tk.LEFT,
				command=self.close)
		b.pack(side=tk.RIGHT)
		b = tk.Button(frame, text="Ok",
				image=icons.get("ok"),
				compound=tk.LEFT,
				command=self.ok)
		b.pack(side=tk.RIGHT)

		self.grid_columnconfigure(1, weight=1)
		self.grid_rowconfigure(1, weight=1)

		self.bind("<Escape>", self.close)
		self.protocol("WM_DELETE_WINDOW", self.close)

		self.project  = project

		if os.path.abspath(path) == Dicom.Scan.directory:
			self.path["text"] = path
			self.fill(uid)

		elif path:
			self.path["text"] = path
			self.scan()

		elif Dicom.Scan.directory:
			self.path.config(text=self.project.relpath(Dicom.Scan.directory))
			self.fill(uid)

		self._result = None

	# ----------------------------------------------------------------------
	def show(self):
		self.deiconify()
		self.grab_set()
		if self.dicomList.size()==0:
			self.chdir()
		if DicomSelectDialog.width > 0:
			self.geometry("%dx%d" %
				(DicomSelectDialog.width, DicomSelectDialog.height))
		self.wait_window()
		return self._result

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		sel = self.dicomList.curselection()
		self._result = [self.path["text"]]
		for i in sel:
			self._result.append(self.dicomList.get(i)[0])
		self.close()

	# ----------------------------------------------------------------------
	def close(self, event=None):
		DicomSelectDialog.width  = self.winfo_width()
		DicomSelectDialog.height = self.winfo_height()
		self.grab_release()
		self.destroy()

	# ----------------------------------------------------------------------
	# Change dicom directory
	# ----------------------------------------------------------------------
	def chdir(self, event=None):
		self.grab_release()
		d = bFileDialog.askdirectory(master=self,
				title="Choose DICOM directory",
				initialdir=self.path["text"],
				filetypes=[("Dicom files","*.dcm"),
					("All","*")])
		if d:
			self.path["text"] = self.project.relpath(d)
			self.config(cursor="watch")
			self.scan()
			self.config(cursor="")
			self.dicomList.focus_set()
			self.dicomList.selection_set(0)
		self.grab_set()

	# ----------------------------------------------------------------------
	# Scan directory for unique dicom files
	# ----------------------------------------------------------------------
	def scan(self):
		try:
			depth = int(self.depth.get())
			DicomSelectDialog.depth = depth
		except:
			depth = DicomSelectDialog.depth
		Dicom.Scan.scan(self.path["text"], depth=depth) #, self)
		self.fill()

	# ----------------------------------------------------------------------
	# Fill listbox
	# ----------------------------------------------------------------------
	def fill(self, select=None):
		self.dicomList.delete(0,tk.END)
		for uid in Dicom.Scan.datasets:
			self.dicomList.insert(tk.END, Dicom.Scan.headerUID(uid))
			if uid == select:
				self.dicomList.selection_set(tk.END)
				self.dicomList.see(tk.END)
				self.dicomList.activate(tk.END)
		self.dicomList.sort(1)
