#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Aug-2012

import math
import tkinter as tk
from tkinter import ttk
try:
	import pydicom
except ImportError:
	try:
		import dicom as pydicom
	except ImportError:
		pydicom = None
from legacy import cmp

import Dicom
import Global
import Ribbon
import tkFlair
import tkExtra
import DicomDraw
import ZoomImage
import FlairRibbon
from Project import DicomInfo

_NEW         = 0
_TOPLEFT     = 1
_TOP         = 2
_TOPRIGHT    = 3
_RIGHT       = 4
_BOTTOMRIGHT = 5
_BOTTOM      = 6
_BOTTOMLEFT  = 7
_LEFT        = 9

#-------------------------------------------------------------------------------
def _cmp(column, x, y):
	if column==0:
		return cmp(x[column], y[column])
	else:
		return cmp(float(x[column]), float(y[column]))

#===============================================================================
# Slice Tab
#===============================================================================
class SliceTab(FlairRibbon.FlairTab):
	#----------------------------------------------------------------------
	def __init__(self, master, page, **kw):
		super().__init__(master, page, "Viewer", **kw)

		# -------------- Slices ---------------
		self.roiList = tkExtra.MultiListbox(self,
					(("ROI", 16, None),
					 ("Id",   4, None)))
		self.roiList.pack(side=tk.LEFT, fill=tk.BOTH)
		self.roiList.bind("<<ListboxSelect>>", self.roiSelect)
		self.roiList.sortAssist = None

		# --------------- Canvas -----------
		frame = tk.Frame(self)
		frame.pack(side=tk.RIGHT, fill=tk.BOTH, expand=tk.YES)
		row,col = 0,0
		self.canvas = ZoomImage.ZoomImage(frame, background="Black", takefocus=True)
		self.canvas.grid(row=row, column=col, columnspan=2, sticky=tk.NSEW)

		self.canvas.bindMotion(wheel=False)
		self.canvas.bindArrows()
		self.canvas.bindZoom()
		self.canvas.bind("<Motion>",          self.motion)
		self.canvas.bind("<Button-1>",        self.click)
		self.canvas.bind("<B1-Motion>",       self.b1motion)
		self.canvas.bind("<Button-3>",        self.button3)
		self.canvas.bind("<B3-Motion>",       self.b3motion)
		self.canvas.bind("<Next>",            self.downZ)
		self.canvas.bind("<Prior>",           self.upZ)
		self.canvas.bind("<Control-Button-4>",self.downZ)
		self.canvas.bind("<Control-Button-5>",self.upZ)
		self.canvas.bind("<<Zoom>>",          self.zoom)
		self.canvas.bind("<<View>>",          self.move)

		col += 2
		sbv = ttk.Scrollbar(frame, orient=tk.VERTICAL, takefocus=False,
				command=self.canvas.yview)
		sbv.grid(row=row, column=col, sticky=tk.NS)

		row,col = row+1,0
		sbh = ttk.Scrollbar(frame, orient=tk.HORIZONTAL, takefocus=False,
				command=self.canvas.xview)
		sbh.grid(row=row, column=col, columnspan=2, sticky=tk.EW)

		self.canvas.config(xscrollcommand=sbh.set, yscrollcommand=sbv.set)

		# -------------
		row,col = row+1,0
		f = tk.Frame(frame)
		f.grid(row=row, column=col, columnspan=3, sticky=tk.NSEW)

		tk.Label(f, text="x:").pack(side=tk.LEFT)
		self.xpixel = tk.Label(f, width=10, foreground="DarkBlue", anchor=tk.W)
		self.xpixel.pack(side=tk.LEFT)

		tk.Label(f, text=" y:").pack(side=tk.LEFT)
		self.ypixel = tk.Label(f, width=10, foreground="DarkBlue", anchor=tk.W)
		self.ypixel.pack(side=tk.LEFT)

		tk.Label(f, text=" z:").pack(side=tk.LEFT)
		self.zpixel = tk.Label(f, width=10, foreground="DarkBlue", anchor=tk.W)
		self.zpixel.pack(side=tk.LEFT)

		tk.Label(f, text=" Dicom:").pack(side=tk.LEFT)
		self.dicompixel = tk.Label(f, width=10, foreground="DarkBlue", anchor=tk.W)
		self.dicompixel.pack(side=tk.LEFT)

		tk.Label(f, text=" Value:").pack(side=tk.LEFT)
		self.valuepixel = tk.Label(f, width=10, foreground="DarkBlue", anchor=tk.W)
		self.valuepixel.pack(side=tk.LEFT)

		# -------------
		row,col = row+1,0
		lf = tk.Frame(frame)
		lf.grid(row=row, column=col, columnspan=3, sticky=tk.NSEW)

		r,c = 0,0
		tk.Label(lf, text="Window Center:").grid(row=r, column=c, sticky=tk.SE)

		c += 1
		self.windowCenter = tk.Scale(lf, from_=-4096, to_=4096,
					showvalue=1,
					orient=tk.HORIZONTAL,
					takefocus=True,
					command=self.levelChanged)
		self.windowCenter.grid(row=r, column=c, sticky=tk.EW)
		self.windowCenter.bind("<Button-3>", lambda e,s=self: s.windowCenter.set(0))
		tkExtra.Balloon.set(self.windowCenter,
				"Center or Level of values to display. Right-click to reset")

		# ---
		c += 1
		tk.Label(lf, text="Width:").grid(row=r, column=c, sticky=tk.SE)

		c += 1
		self.windowWidth = tk.Scale(lf, from_=0, to_=16000,
					showvalue=1,
					orient=tk.HORIZONTAL,
					takefocus=True,
					command=self.levelChanged)
		self.windowWidth.grid(row=r, column=c, sticky=tk.EW)
		self.windowWidth.bind("<Button-3>", lambda e,s=self: s.windowWidth.set(0))
		tkExtra.Balloon.set(self.windowWidth,
				"Window around level/center values to display. Right-click to reset")
		lf.grid_columnconfigure(c, weight=1)

		# ---
		c += 1
		tk.Label(lf, text="Preset:").grid(row=r, column=c, sticky=tk.SE)

		c += 1
		windows = list(DicomInfo.windows)
		windows.insert(0,"")
		self.windowPreset = tkExtra.Combobox(lf, True,
						command=self.presetChanged,
						width=10)
		self.windowPreset.fill(sorted(windows))
		self.windowPreset.set("",False)
		self.windowPreset.grid(row=r, column=c, sticky=tk.EW+tk.S)
		tkExtra.Balloon.set(self.windowPreset, "Window presets")

		# -------------
		r,c = r+1,0
		tk.Label(lf, text="Slice:").grid(row=r, column=c, sticky=tk.SE)

		c += 1
		self.sliceScaleVar = tk.IntVar()
		self.sliceScale = tk.Scale(lf, from_=1, to_=100,
					orient=tk.HORIZONTAL,
					variable=self.sliceScaleVar,
					command=self.sliceChange)
		self.sliceScale.grid(row=r, column=c, columnspan=5, sticky=tk.EW)
		tkExtra.Balloon.set(self.sliceScale, "Slice/Frame to display")

		lf.grid_columnconfigure(1, weight=1)
		lf.grid_columnconfigure(3, weight=1)

		# -------------
		frame.grid_rowconfigure(0, weight=1)
		frame.grid_columnconfigure(1, weight=1)
		#frame.grid_columnconfigure(3, weight=1)

		self.grid_columnconfigure(1, weight=1)
		self._photo  = None			# photo
		self._rect   = None			# cropping rectangle
		self._item   = None
		self.iop     = [1.,0.,0.,0.,1.,0.]	# Transformation matrix
		self.origin  = (0.,0.,0.)		# Dicom origin
		self.dxyz    = (1.,1.,1.)		# voxel size (dx,dy,dz)
		self.size    = (0, 0)			# Slice size nx,ny
		self.dataset = None			# current loaded dataset
		self._loadedFile = None			# currently loaded file
		self.frame   = 0			# Number of frames for scale
		self.ctinfo  = None			# CT project DicomInfo structure related to this dataset
		self._after  = None
		self._fromPreset = False

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Viewer ===========
		group = Ribbon.LabelGroup(self.ribbon, "Crop")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.grid3rows()

		# --- Xcrop ---
		row,column = 0,0
		tk.Label(group.frame, text="Xmin:",
				font=Ribbon._FONT,
				background=Ribbon._BACKGROUND).grid(
					row=row, column=column, sticky=tk.E)
		column += 1
		self.xfrom = tkExtra.FloatEntry(group.frame, background="white", width=8)
		self.xfrom.grid(row=row, column=column, sticky=tk.EW)
		self.xfrom.bind("<Return>",   self.drawRectangle)
		self.xfrom.bind("<KP_Enter>", self.drawRectangle)
		self.xfrom.bind("<FocusOut>", self.drawRectangle)
		group.frame.grid_columnconfigure(column, weight=1)

		column += 1
		tk.Label(group.frame, text="Xmax:",
				font=Ribbon._FONT,
				background=Ribbon._BACKGROUND).grid(
					row=row, column=column, sticky=tk.E)
		column += 1
		self.xto = tkExtra.FloatEntry(group.frame, background="white", width=8)
		self.xto.grid(row=row, column=column, sticky=tk.EW)
		self.xto.bind("<Return>",   self.drawRectangle)
		self.xto.bind("<KP_Enter>", self.drawRectangle)
		self.xto.bind("<FocusOut>", self.drawRectangle)
		group.frame.grid_columnconfigure(column, weight=1)

		# --- Ycrop ---
		row,column = row+1,0
		tk.Label(group.frame, text="Ymin:",
				font=Ribbon._FONT,
				background=Ribbon._BACKGROUND).grid(
					row=row, column=column, sticky=tk.E)
		column += 1
		self.yfrom = tkExtra.FloatEntry(group.frame, background="white", width=8)
		self.yfrom.grid(row=row, column=column, sticky=tk.EW)
		self.yfrom.bind("<Return>",   self.drawRectangle)
		self.yfrom.bind("<KP_Enter>", self.drawRectangle)
		self.yfrom.bind("<FocusOut>", self.drawRectangle)

		column += 1
		tk.Label(group.frame, text="Ymax:",
				font=Ribbon._FONT,
				background=Ribbon._BACKGROUND).grid(
					row=row, column=column, sticky=tk.E)
		column += 1
		self.yto = tkExtra.FloatEntry(group.frame, background="white", width=8)
		self.yto.grid(row=row, column=column, sticky=tk.EW)
		self.yto.bind("<Return>",   self.drawRectangle)
		self.yto.bind("<KP_Enter>", self.drawRectangle)
		self.yto.bind("<FocusOut>", self.drawRectangle)

		# --- Zcrop ---
		row,column = row+1,0
		tk.Label(group.frame, text="Zmin:",
				font=Ribbon._FONT,
				background=Ribbon._BACKGROUND).grid(
					row=row, column=column, sticky=tk.E)
		column += 1
		self.zfrom = tkExtra.FloatEntry(group.frame, background="white", width=8)
		self.zfrom.grid(row=row, column=column, sticky=tk.EW)
		column += 1
		tk.Label(group.frame, text="Zmax:",
				font=Ribbon._FONT,
				background=Ribbon._BACKGROUND).grid(
					row=row, column=column, sticky=tk.E)
		column += 1
		self.zto = tkExtra.FloatEntry(group.frame, background="white", width=8)
		self.zto.grid(row=row, column=column, sticky=tk.EW)

		# --- Buttons
		row,column = 0, column+1
		Ribbon.LabelButton(group.frame,
				text="Reset",
				anchor=tk.W,
				command=self.reset).grid(
					row=row, column=column, sticky=tk.EW)
		row += 1
		Ribbon.LabelButton(group.frame,
				text="Slice \N{RIGHTWARDS ARROW} Zmin",
				anchor=tk.W,
				command=self.setZFrom).grid(
					row=row, column=column, sticky=tk.EW)
		row += 1
		Ribbon.LabelButton(group.frame,
				text="Slice \N{RIGHTWARDS ARROW} Zmax",
				anchor=tk.W,
				command=self.setZTo).grid(
					row=row, column=column, sticky=tk.EW)

	# ----------------------------------------------------------------------
	@property
	def dicom(self):	return self.page.dicominfo.dicom

	# ----------------------------------------------------------------------
	# Select a new dataset and populate the listbox
	# ----------------------------------------------------------------------
	def activate(self):
		self.refresh()

		if self.dataset is not None and self.dataset.SeriesInstanceUID == self.dicom.uid:
			return
		#self.dicom.pixel_array = None

		self.clearDatasetSelection()
		di = self.page.selectedDicom()
		if di is None:
			self.flair.notify("No DICOM selected",
				"Please select a DICOM set first")
			return

		# Init dicom with the first slice
		try:
			self.dataset = pydicom.read_file(
						self.page.dicominfo.filename(),
						force=True)
		except pydicom.filereader.InvalidDicomError:
			self.dataset = None
			return

		self.ctinfo = None
		# Fill list according to modality
		if self.dicom.modality == "RTSTRUCT":
			self.roiList.pack(side=tk.LEFT, fill=tk.BOTH)
			self.roiList.delete(0,tk.END)
			for roi in self.dataset.StructureSetROISequence:
				self.roiList.insert(tk.END,(str(roi.ROIName), str(roi.ROINumber)))
			self.loadSlice(self.page.dicomfile[0])

			# Find the corresponding CT data set
			self.ctinfo = self.page.rtstructReferencedCT(self.dataset)
			if self.ctinfo is None:
				self.flair.notify("Referenced CT not found",
					"RT structure is referring to an DICOM "
					"image which is not loaded",
					tkFlair.NOTIFY_ERROR)
				return

			# Set scale to the number of SOP UIDs
			self.sliceScale.config(to_=len(self.ctinfo.files))
		else:
			self.roiList.pack_forget()
			if self.page.dicomfile:
				self.loadSlice(self.page.dicomfile[0])

			if self.dicom.hasSlices:
				# 2D slice data
				self.sliceScale.config(to_=len(self.page.dicominfo.files))
			else:
				# 3D multi frame data
				self.sliceScale.config(to_=self.dicom.nz)
			self.frame = self.page.selectedFile()
		self.updateOnIdle()	#<<<<< instead of sliceScale set
		# through the var not to trigger the callback
		self.sliceScaleVar.set(self.frame+1)

	# ----------------------------------------------------------------------
	def refresh(self):
		di = self.page.dicominfo
		if di is None:
			self.canvas.setImage()
			self.draw()
			self._photo = None
			return

		self.dataset = None

		self.xfrom.set(di("xfrom"))
		self.xto.set(  di("xto"))

		self.yfrom.set(di("yfrom"))
		self.yto.set(  di("yto"))

		self.zfrom.set(di("zfrom"))
		self.zto.set(  di("zto"))

		# Changing the center/window will update the image/crop
		self.windowCenter.set(float(di("center")))
		self.windowWidth.set(float(di("window")))
		self.windowPreset.set(di("preset"))

	# ----------------------------------------------------------------------
	def get(self):
		Global.dump("*-* Slice.get", self.xfrom.get(),level=2)
		if not self.winfo_ismapped(): return
		di = self.page.dicominfo
		if di is None: return

		oldHash = hash(di)
		di["xfrom"] = self.xfrom.get()
		di["xto"]   = self.xto.get()

		di["yfrom"] = self.yfrom.get()
		di["yto"]   = self.yto.get()

		di["zfrom"] = self.zfrom.get()
		di["zto"]   = self.zto.get()

		di["window"] = self.windowWidth.get()
		di["center"] = self.windowCenter.get()
		di["preset"] = self.windowPreset.get()

		if self.dicom: self.dicom.setLimits(*self.limits())

		if hash(di) != oldHash:
			di.cache.clear()
			self.dicom.scanned(False)

	# ----------------------------------------------------------------------
	def deactivate(self):
		self.get()

	# ----------------------------------------------------------------------
	# Return crop image in pixel coordinates
	# ----------------------------------------------------------------------
	def crop(self):
		if self.dataset is None or self.page.dicominfo is None: return

		try:
			xfrom = max(0,min(self.dicom.nx-1,
				int(math.floor((float(self.xfrom.get())-self.dicom.x)/self.dicom.dx))))
		except (ValueError, ZeroDivisionError):
			xfrom = 0

		try:
			xto = max(0,min(self.dicom.nx-1,
				int(math.floor((float(self.xto.get())-self.dicom.x)/self.dicom.dx))))
		except (ValueError, ZeroDivisionError):
			xto = self.dicom.nx-1

		try:
			yfrom = max(0,min(self.dicom.ny-1,
				int(math.floor((float(self.yfrom.get())-self.dicom.y)/self.dicom.dy))))
		except (ValueError, ZeroDivisionError):
			yfrom = 0

		try:
			yto = max(0,min(self.dicom.ny-1,
				int(math.floor((float(self.yto.get())-self.dicom.y)/self.dicom.dy))))
		except (ValueError, ZeroDivisionError):
			yto = self.dicom.ny-1

		if xto < xfrom:
			xfrom, xto = xto, xfrom
			tmp = self.xfrom.get()
			self.xfrom.set(self.xto.get())
			self.xto.set(tmp)

		if yto < yfrom:
			yfrom, yto = yto, yfrom
			tmp = self.yfrom.get()
			self.yfrom.set(self.yto.get())
			self.yto.set(tmp)

		return xfrom, yfrom, xto, yto

	# ----------------------------------------------------------------------
	# @return limits as entered by the user
	# ----------------------------------------------------------------------
	def limits(self):
		try:    xfrom = float(self.xfrom.get())
		except: xfrom = None

		try:    xto   = float(self.xto.get())
		except: xto   = None

		try:    yfrom = float(self.yfrom.get())
		except: yfrom = None

		try:    yto   = float(self.yto.get())
		except: yto   = None

		try:    zfrom = float(self.zfrom.get())
		except: zfrom = None

		try:    zto   = float(self.zto.get())
		except: zto   = None

		return (xfrom, xto, yfrom, yto, zfrom, zto)

	# ----------------------------------------------------------------------
	# Select a roi
	# ----------------------------------------------------------------------
	def roiSelect(self, event=None):
		sel = self.roiList.curselection()
		if not sel: return
		if self.dicom.modality == "RTSTRUCT":
			self.updateOnIdle()

	# ----------------------------------------------------------------------
	def sliceChange(self, event=None):
		self.frame = self.sliceScale.get()-1
		if self.dicom.hasSlices:
			self.page.selectFile(self.frame)
		if not self.dicom.hasSlices:
			# update canvas
			self.updateOnIdle()
		self.zpixel["text"] = "%.4g"%(float(self.frame)*self.dicom.dz + self.dicom.z)

	# ----------------------------------------------------------------------
	# Load slice
	# ----------------------------------------------------------------------
	def loadSlice(self, filename):
		if filename == self._loadedFile: return

		self.clearSlice()
		try:
			self.dataset = pydicom.read_file(filename, force=True)
			self._loadedFile = filename
		except pydicom.filereader.InvalidDicomError:
			self._loadedFile = None
			return
		#self.dicom.init(self.dataset,
		#		z=self.page.selectedDicom().files[0][1])

	# ----------------------------------------------------------------------
	def updateOnIdle(self):
		if self._after is None:
			self._after = self.after_idle(self.updateImage)

	# ----------------------------------------------------------------------
	# Convert dicom to image and show it in canvas
	# ----------------------------------------------------------------------
	def updateImage(self, event=None):
		self._after = None
		if self.dataset is None or self.page.dicominfo is None:
			self.draw()
			self.canvas.setImage()
			self._photo = None
			return

		try:    window = int(self.windowWidth.get())
		except: window = None
		try:    center = int(self.windowCenter.get())
		except: center = None

		if self.dicom.modality == "RTSTRUCT":
			# XXX WARNING ROI order is not the same with the sequence index!!!
			rois = []
			for i in self.roiList.curselection():
				rois.append(int(self.roiList.listbox(1).get(int(i))))
			if self.ctinfo is None:
				self.flair.notify("Referenced SOP not found",
					"RT structure is referring to an DICOM image "
					"which is not loaded",
					tkFlair.NOTIFY_ERROR)
			else:
				fn,z,sop  = self.ctinfo.files[self.frame]
				dd = Dicom.Dicom(self.ctinfo.filename(fn))
				self.origin = (dd.x,  dd.y,  dd.z)
				self.dxyz   = (dd.dx, dd.dy, dd.dz)
				self.size   = (dd.nx, dd.ny)
				self.iop    = dd.iop
				self._photo = DicomDraw.rtstruct2image(self.dataset,
							dd.dataset, z, sop, rois, window, center)
		else:
			if "TransferSyntaxUID" not in self.dataset.file_meta:
				self.dataset.file_meta.TransferSyntaxUID = \
					pydicom.uid.ImplicitVRLittleEndian
			try:
				#if self.dicom.pixel_array is None:
				#	self.dicom.pixel_array = self.dicom.m * self.dataset.pixel_array \
				#				+ self.dicom.b
				self._photo = DicomDraw.dicom2image(self.dataset,
							self.dataset.pixel_array,
							#self.dicom.pixel_array,
							window, center, self.frame)
			except Exception as ex:
				if not isinstance(ex, (AttributeError,TypeError,
							OSError,NotImplementedError)):
					tkFlair.addException()
				self.flair.notify("Error displaying dicom",
						str(ex),
						tkFlair.NOTIFY_ERROR)
				self.canvas.setImage()
				self._photo = None
			try:
				self.iop = list(map(float,self.dataset.ImageOrientationPatient))
			except:
				self.iop = [1., 0., 0., 0., 1., 0.]

			try:
				self.origin = (self.dicom.x,  self.dicom.y,  self.dicom.z)
				self.dxyz   = (self.dicom.dx, self.dicom.dy, self.dicom.dz)
				self.size   = (self.dicom.nx, self.dicom.ny)
			except:
				self.origin = (0., 0., 0.)
				self.dxyz   = (1., 1., 1.)
				self.size   = (0, 0)

			if self._photo is None: return

		self.canvas.setImage(self._photo)
		self.draw()

		bb = self.canvas.bbox('all')
		if bb is None: return
		self.canvas.configure(scrollregion=bb)

	# ----------------------------------------------------------------------
	def draw(self, event=None):
		self.drawRectangle()
		self.drawOrientation()

	# ----------------------------------------------------------------------
	def drawRectangle(self, event=None):
		if self.dataset is None or self.page.dicominfo is None: return
		xfrom, yfrom, xto, yto = self.crop()
		xfrom *= self.canvas.zoom
		yfrom *= self.canvas.zoom
		xto += 1
		yto += 1
		xto   *= self.canvas.zoom
		yto   *= self.canvas.zoom
		if self._rect is None:
			self._rect = self.canvas.create_rectangle((xfrom,yfrom, xto,yto),
							outline="Orange")
		else:
			self.canvas.coords(self._rect, (xfrom,yfrom, xto, yto))

	# ----------------------------------------------------------------------
	def drawOrientation(self):
		self.canvas.delete("orientation")
		if self.dataset is None: return
		self.__x = 0			# orientation canvas offset
		self.__y = 0
		DicomDraw.drawOrientation(self.canvas, self.iop, 50, 50, 20)
		self.move()

	# ----------------------------------------------------------------------
	def reset(self):
		self.xfrom.set("")
		self.xto.set("")
		self.yfrom.set("")
		self.yto.set("")
		self.zfrom.set("")
		self.zto.set("")
		self.draw()

	# ----------------------------------------------------------------------
	def setZFrom(self):
		if self.dataset:
			try:
				self.zfrom.set(self.dataset.SliceLocation)
			except AttributeError:
				pass
		#else:
		#	self.setZRange()

	# ----------------------------------------------------------------------
	def setZTo(self):
		if self.dataset:
			try:
				self.zto.set(self.dataset.SliceLocation)
			except AttributeError:
				pass
		#else:
		#	self.setZRange()

	# ----------------------------------------------------------------------
	#def setZRange(self):
	#	sel = self.sliceList.curselection()
	#	if len(sel)<=1: return
	#	zmin = 9999999.0
	#	zmax = -zmin
	#	for i in sel:
	#		z = float(self.sliceList.get(i)[1])
	#		zmin = min(z,zmin)
	#		zmax = max(z,zmax)
	#	self.zfrom.set(zmin)
	#	self.zto.set(zmax)

	# --------------------------------------------------------------------
	def motion(self, event):
		if self._photo    is None or \
		   self.dataset   is None or \
		   self.page.dicominfo is None or \
		   self.dicom     is None:
			return

		i = min(self.size[0]-1, max(0, self.canvas.canvasx(event.x)//self.canvas.zoom))
		j = min(self.size[1]-1, max(0, self.canvas.canvasy(event.y)//self.canvas.zoom))

		di = self.dxyz[0]*i
		dj = self.dxyz[0]*j

		# FIXME add cross product.... on Z
		x = self.iop[0]*di + self.iop[3]*dj + self.origin[0]
		y = self.iop[1]*di + self.iop[4]*dj + self.origin[1]
		z = self.iop[2]*di + self.iop[5]*dj + self.origin[2] + float(self.frame)*self.dxyz[2]

		self.xpixel["text"] = "%.6g"%(x)
		self.ypixel["text"] = "%.6g"%(y)
		self.zpixel["text"] = "%.6g"%(z)

		if self.dicom.modality == "RTSTRUCT":
			self.dicompixel["text"] = ""
			self.valuepixel["text"] = ""
			return

		i = int(i)
		j = int(j)
		try:
			if self.dicom.hasSlices:
				d = self.dataset.pixel_array[j][i]
			else:
				d = self.dataset.pixel_array[self.frame][j][i]
			self.dicompixel["text"] = "%.5g"%(d)
			self.valuepixel["text"] = "%.5g"%(self.dicom.m*d + self.dicom.b)
		except IndexError:
			self.dicompixel["text"] = ""
			self.valuepixel["text"] = ""

	# --------------------------------------------------------------------
	# Button1 clicked find item to change in clipping rectangle
	# --------------------------------------------------------------------
	def click(self, event):
		if self._photo is None: return
		self.canvas.focus_set()
		x = self.canvas.canvasx(event.x)
		y = self.canvas.canvasy(event.y)
		enclosed = self.canvas.find_overlapping(x-5, y-5, x+5, y+5)

		if self._rect in enclosed:
			# Find item to drag
			x /= self.canvas.zoom
			y /= self.canvas.zoom
			xfrom, yfrom, xto, yto = self.crop()
			MARGIN = 0.2
			xto += 1
			yto += 1
			if abs(y - yfrom) <= MARGIN:	# On top line
				if  abs(x - xfrom) <= MARGIN:
					self._item = _TOPLEFT
				elif abs(x  - xto) <= MARGIN:
					self._item = _TOPRIGHT
				else:
					self._item = _TOP
			elif abs(y - yto) <= MARGIN:	# On bottom line
				if  abs(x - xfrom) <= MARGIN:
					self._item = _BOTTOMLEFT
				elif abs(x - xto) <= MARGIN:
					self._item = _BOTTOMRIGHT
				else:
					self._item = _BOTTOM
			elif abs(x - xfrom) <= MARGIN:
				self._item = _LEFT
			else:
				self._item = _RIGHT
		else:
			self._item = _NEW

		self._startx = x
		self._starty = y

	# --------------------------------------------------------------------
	# Draw clipping rectangle
	# --------------------------------------------------------------------
	def b1motion(self, event):
		if self._photo is None: return
		i = min(self.dicom.nx-1, max(0, self.canvas.canvasx(event.x)//self.canvas.zoom))
		j = min(self.dicom.ny-1, max(0, self.canvas.canvasy(event.y)//self.canvas.zoom))
		if self._item in (_TOPRIGHT, _RIGHT, _BOTTOMRIGHT):
			i -= 1
		if self._item in (_BOTTOMLEFT, _BOTTOM, _BOTTOMRIGHT):
			j -= 1
		x = str(i*self.dicom.dx + self.dicom.x)
		y = str(j*self.dicom.dy + self.dicom.y)

		if   self._item == _NEW:
			if abs(self._startx - self.canvas.canvasx(event.x))>3 or \
			   abs(self._starty - self.canvas.canvasy(event.y))>3:
				self.xfrom.set(x)
				self.yfrom.set(y)
				self._item = _BOTTOMRIGHT

		elif self._item == _TOPLEFT:
			self.xfrom.set(x)
			self.yfrom.set(y)

		elif self._item == _TOP:
			self.yfrom.set(y)

		elif self._item == _TOPRIGHT:
			self.xto.set(x)
			self.yfrom.set(y)

		elif self._item == _RIGHT:
			self.xto.set(x)

		elif self._item == _BOTTOMRIGHT:
			self.xto.set(x)
			self.yto.set(y)

		elif self._item == _BOTTOM:
			self.yto.set(y)

		elif self._item == _BOTTOMLEFT:
			self.xfrom.set(x)
			self.yto.set(y)

		elif self._item == _LEFT:
			self.xfrom.set(x)

		if self._item != _NEW: self.drawRectangle()

	# ----------------------------------------------------------------------
	def button3(self, event):
		self._startx = event.x
		self._starty = event.y
		self._window = self.windowWidth.get()
		self._center = self.windowCenter.get()

	# ----------------------------------------------------------------------
	def b3motion(self, event):
		self.windowWidth.set( self._center + 5*(self._startx - event.x))
		self.windowCenter.set(self._window + 5*(self._starty - event.y))

	# --------------------------------------------------------------------
	# Compensate moving of orientation axis, to appear immobile on screen
	# --------------------------------------------------------------------
	def move(self, event=None):
		x = self.canvas.canvasx(0)
		y = self.canvas.canvasy(0)
		dx = x - self.__x
		dy = y - self.__y
		self.canvas.move("orientation", dx, dy)
		self.__x = x
		self.__y = y

	# ----------------------------------------------------------------------
	def zoom(self, event=None):
		self.canvas.delete("orientation")
		self.draw()

	# ----------------------------------------------------------------------
	def downZ(self, event=None):
		self.sliceScale.set(max(1,self.frame))

	# ----------------------------------------------------------------------
	def upZ(self, event=None):
		self.sliceScale.set(min(self.sliceScale["to"],self.frame+2))

	# ----------------------------------------------------------------------
	def levelChanged(self, s=None):
		#self.windowPreset.set("", False)
		self.updateOnIdle()

	# ----------------------------------------------------------------------
	def presetChanged(self):
		preset = self.windowPreset.get()
		try:
			center, width = DicomInfo.windows[preset]
		except KeyError:
			return
		if center != self.windowCenter.get():
			self.windowCenter.set(center)
		if width != self.windowWidth.get():
			self.windowWidth.set(width)

	# ----------------------------------------------------------------------
	# Clear current selection
	# ----------------------------------------------------------------------
	def clearDatasetSelection(self):
		#self.sliceList.delete(0,tk.END)
		self.clearSlice()

	# ----------------------------------------------------------------------
	def clearSlice(self):
		self.dataset = None
		if self._photo:
			self.draw()
			self.canvas.setImage()
			self._photo = None
