#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Aug-2012

import os
import math
import time
import tkinter as tk
from tkinter import ttk

try:
	import pydicom
except ImportError:
	try:
		import dicom as pydicom
	except ImportError:
		pydicom = None

import tool
import icons
import Dicom
import Global
import Ribbon
import tkFlair
import tkExtra
import tkDialogs
import bFileDialog
import FlairRibbon

from material import Material
from fluka import Fluka, FlukaCard, FlukaLayout

#-------------------------------------------------------------------------------
class UnitListAssist(tkExtra.SortAssist):
	def __call__(self, x):
		try:
			return float(x[self.column])
		except:
			return x[self.column]

#===============================================================================
# Voxel Tab
#===============================================================================
class VoxelTab(FlairRibbon.FlairTab):
	NEW = "<new>"
	ADD = "<add>"
	BACKGROUNDS = ("White", "#DDEEFF")

	#-----------------------------------------------------------------------
	def __init__(self, master, page, **kw):
		super().__init__(master, page, "Voxel", **kw)

		self.ctinfo    = None
		self.roi       = {}
		self.roiList   = []
		self.lastPath  = ""
		self.materials = []
		self.unit2matModified    = False
		self.roiMaterialModified = False

		self.tabPage = ttk.Notebook(self)
		self.tabPage.pack(expand=tk.YES, fill=tk.BOTH)

		# -------------------------------------------------------------
		frame = tk.Frame(self.tabPage)
		self.tabPage.add(frame,
				text     = "1. Hounsfield Units",
				image    = icons.get("dicom"),
				compound = tk.LEFT,
				sticky   = tk.NSEW)

		# -------------- Units to Materials ---------------
		row,col = 0,0
		tk.Label(frame, text="HU to Material File:").grid(
				row=row, column=col, pady=1, sticky=tk.E)
		col += 1
		self.unit2mat = tkExtra.Entry(frame,
				foreground=tkFlair._ELABEL_FOREGROUND_COLOR,
				background=tkFlair._ELABEL_BACKGROUND_COLOR)
		self.unit2mat.grid(row=row, column=col, sticky=tk.EW)
		self.unit2mat.bind('<Button-3>',         self.loadUnit2Material)
		self.unit2mat.bind('<Control-Button-1>', self.loadUnit2Material)
		self.unit2mat.bind('<Return>',           self.readUnit2Material)
		self.unit2mat.bind('<KP_Enter>',         self.readUnit2Material)
		self.unit2mat.bind('<FocusOut>',         self.readUnit2Material)
		tkExtra.Balloon.set(self.unit2mat, "Hounsfield Unit to material conversion data file")

		col += 1
		b = tk.Button(frame, image=icons.get("load"), pady=0,
				command=self.loadUnit2Material)
		b.grid(row=row, column=col)
		tkExtra.Balloon.set(b, "Load unit to material conversion")

		col += 1
		b = tk.Button(frame, image=icons.get("edit"), pady=0,
				command=self.editUnit2Material)
		b.grid(row=row, column=col)
		tkExtra.Balloon.set(b, "Edit Hounsfield units in an external editor")

		col += 1
		b = tk.Button(frame, image=icons.get("save"), pady=0,
				command=self.saveUnit2Material)
		b.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(b, "Save unit to material conversion")

		colspan = col
		# -------------- Unit to Material ---------------
		# --- Table editing ---
		row += 1
		col  = 0
		self.unit2matList = tkExtra.MultiListbox(frame,
				(#('< Unit',        5, None),
				 ('HUmin',          5, None),
				 ('HUmax',          5, None),
				 ('Material',      10, None),
				 ('Crho_min',      10, None),
				 ('Crho_max',      10, None),
				 ('CdEdx_rel_min', 10, None),
				 ('CdEdx_rel_max', 10, None)),
				 background="White")
		self.unit2matList.sortAssist = UnitListAssist
		self.unit2matList.grid(row=row, column=0,
				columnspan=colspan, rowspan=4, sticky=tk.NSEW)
		self.unit2matList.bindList("<ButtonRelease-1>",	self.unit2matRelease)
		self.unit2matList.bindList("<Return>",		self.unit2matEdit)
		self.unit2matList.bindList("<KP_Enter>",	self.unit2matEdit)
		self.unit2matList.bindList("<<Rename>>",	self.unit2matEdit)
		self.unit2matList.bindList("<Insert>",		self.add)
		self.unit2matList.bindList("<Control-Return>",	self.add)
		self.unit2matList.bindList("<Delete>",		self.delete)
		self.unit2matList.bindList("<<Clone>>",		self.clone)
		self.unit2matList.bindLeftRight()
		self.unit2matList.sortAssist = None

		# ---
		col += colspan
		b = tk.Button(frame, image=icons.get("add"),
				takefocus=False, command=self.add)
		tkExtra.Balloon.set(b, "Add unit range")
		b.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		b = tk.Button(frame, image=icons.get("del"),
				takefocus=False, command=self.delete)
		tkExtra.Balloon.set(b, "Delete unit range")
		b.grid(row=row, column=col, sticky=tk.EW)

		# ---
		row += 1
		b = tk.Button(frame, image=icons.get("clone"),
				takefocus=False, command=self.clone)
		tkExtra.Balloon.set(b, "Clone unit range")
		b.grid(row=row, column=col, sticky=tk.EW)

		# Stretching
		frame.grid_rowconfigure(row+1, weight=1)
		frame.grid_columnconfigure(1, weight=1)

		# -------------------------------------------------------------
		frame = tk.Frame(self.tabPage)
		self.tabPage.add(frame,
				text     = "2. RTSTRUCT and Material",
				image    = icons.get("MATERIAL"),
				compound = tk.LEFT)

		row,col = 0,0
		tk.Label(frame, text="RTSTRUCT:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.rtstruct = tkExtra.Combobox(frame,
					command=self.rtstructChanged,
					width=30,
					foreground="DarkBlue")
		self.rtstruct.grid(row=row, column=col, columnspan=4, sticky=tk.EW)

		# ---
		col += 1
		tk.Label(frame, text="Field:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.field = tkExtra.Combobox(frame, width=10,
					foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
					background="White")
		self.field.grid(row=row, column=col, columnspan=2, sticky=tk.EW)
		self.field.fill([x[0] for x in FlukaLayout.ASSIGNMA_field])

		frame.grid_rowconfigure(row+1, weight=1)
		frame.grid_columnconfigure(1, weight=1)

		# -------------- Material Override ---------------
		row += 1
		col  = 0
		f = tk.LabelFrame(frame, text="Material Override", foreground="DarkRed")
		f.grid(row=row, column=col, columnspan=4, sticky=tk.NSEW)
		# --- Table editing ---
		row = 0
		col = 0
		self.roiMaterial = tkExtra.MultiListbox(f,
				 (('Material',      16, 'MATERIAL'),
				  ('Operation',      6, None),
				  ('ROI',           32, None)),
				 background="White")
		self.roiMaterial.sortAssist = UnitListAssist
		self.roiMaterial.grid(row=row, column=0, rowspan=6, sticky=tk.NSEW)
		self.roiMaterial.bindList("<ButtonRelease-1>",	self.roiRelease)
		self.roiMaterial.bindList("<Return>",		self.roiEdit)
		self.roiMaterial.bindList("<KP_Enter>",		self.roiEdit)
		self.roiMaterial.bindList("<<Rename>>",		self.roiEdit)
		self.roiMaterial.bindList("<Insert>",		self.roiAdd)
		self.roiMaterial.bindList("<Control-Return>",	self.roiAdd)
		self.roiMaterial.bindList("<Delete>",		self.roiDelete)
		self.roiMaterial.bindList("<<Clone>>",		self.roiClone)
		self.roiMaterial.bindList("<<MoveUp>>",		self.roiMoveUp)
		self.roiMaterial.bindList("<<MoveDown>>",	self.roiMoveDown)
		self.roiMaterial.bindLeftRight()
		#self.roiMaterial.listbox(1).config(justify=tk.CENTER)
		self.roiMaterial.sortAssist = None
		self.roiMaterial.setPopupMenu([
				('Add',    0, self.roiAdd,    icons.get("add")),
				('Clone',  0, self.roiClone,  icons.get("clone")),
				('Remove', 0, self.roiDelete, icons.get("x")),
				None,
				('Move Up',   0, self.roiMoveUp,   icons.get("up")),
				('Move Down', 0, self.roiMoveDown, icons.get("down")),
				#('Cut',    0, self.cut,           icons.get("cut")),
				#('Copy',   0, self.copy,          icons.get("copy")),
				#('Paste',  0, self.paste,         icons.get("paste")),
				])

		# ---
		col += colspan
		b = tk.Button(f, image=icons.get("add"),
				takefocus=False,
				command=self.roiAdd)
		tkExtra.Balloon.set(b, "Add material filter")
		b.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		b = tk.Button(f, image=icons.get("clone"),
				takefocus=False,
				state = tk.DISABLED,
				command=self.roiClone)
		tkExtra.Balloon.set(b, "Clone material filter")
		b.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		b = tk.Button(f, image=icons.get("del"),
				takefocus=False,
				command=self.roiDelete)
		tkExtra.Balloon.set(b, "Delete material filter")
		b.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		b = tk.Button(f, image=icons.get("up"),
				takefocus=False,
				state = tk.DISABLED,
				command=self.roiMoveUp)
		tkExtra.Balloon.set(b, "Move up material")
		b.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		b = tk.Button(f, image=icons.get("down"),
				state = tk.DISABLED,
				takefocus=False,
				command=self.roiMoveDown)
		tkExtra.Balloon.set(b, "Move up material")
		b.grid(row=row, column=col, sticky=tk.EW)

		# ---
		f.grid_rowconfigure(row+1, weight=1)
		f.grid_columnconfigure(0, weight=1)

	#-----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Files ===========
		group = Ribbon.LabelGroup(self.ribbon, "Execute")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid2rows()

		# -------------- Output voxel file ---------------
		col, row = 0,0
		self.voxelfile = tkExtra.LabelEntry(group.frame,
					"Voxel file",
					tkFlair._LABEL_ENTRY_COLOR,
					width=30,
					#font=Ribbon._FONT,
					relief=tk.SUNKEN)
		self.voxelfile.grid(row=row, column=col, rowspan=2, sticky=tk.EW)
		self.voxelfile.bind('<Button-3>',         self.saveVoxelFile)
		self.voxelfile.bind('<Control-Button-1>', self.saveVoxelFile)
		tkExtra.Balloon.set(self.voxelfile, "Output voxel file")

		col += 1
		b = Ribbon.LabelButton(group.frame, image=icons.get("VOXELS"),
				text="VOXEL",
				command=self.makeVoxel,
				foreground="Darkred")
		b.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(b, "Generate a VOXEL files from the dicom files")

		row += 1
		b = Ribbon.LabelButton(group.frame, image=icons.get("input"),
				text="Card",
				command=self.voxelCard)

		b.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(b, "Add or modify VOXEL card")

	# ----------------------------------------------------------------------
	@property
	def dicom(self):	 return self.ctinfo.dicom

	# ----------------------------------------------------------------------
	def activate(self):
		self.page.analyzeSlices()
		mats = set(Material.materials.keys())
		mats.update(Material.names.keys())
		mats.update(Fluka.defaultMaterials.keys())
		mats.update(Fluka.icruMaterials.keys())
		self.materials = list(mats)
		self.materials.sort()
		self.refresh()

	# ----------------------------------------------------------------------
	def deactivate(self):
		if self.page.dicominfo is not None:
			self.get()

	# ----------------------------------------------------------------------
	# Refresh values
	# ----------------------------------------------------------------------
	def refresh(self):
		Global.dump("*-* DicomVoxelTab refresh", level=2)
		self.ctinfo = self.page.dicominfo
		if self.ctinfo is None: return
		if self.ctinfo.modality == "RTSTRUCT":
			self.ctinfo = self.page.rtstructReferencedCT()
		if self.ctinfo is None or self.dicom is None: return

		try:
			v = self.ctinfo["unit2mat"]
		except KeyError:
			v = tkFlair.config.getStr("Dicom","unit2mat","")
		if v != self.unit2mat.get():
			self.unit2mat.set(v)
			self.readUnit2Material()

		self.voxelfile.set(self.ctinfo("voxel"))
		self.field.set(FlukaLayout.ASSIGNMA_field[int(self.ctinfo("field"))][0])

		# Fill a list of RTSTRUCT
		self.rtstruct.config(state=tk.NORMAL)
		self.rtstruct.delete(0,tk.END)
		if self.page.dicominfo.modality == "RTSTRUCT":
			self.rtstruct.set(self.page.dicominfo.name)
			self.rtstruct.config(state=tk.DISABLED)
		else:
			rtfirst = None
			self.rtstruct.insert(tk.END,"")
			for d in self.project.dicoms:
				if d.modality=="RTSTRUCT":
					self.rtstruct.insert(tk.END,d.name)
					if rtfirst is None:
						rtfirst = d.name
			rtstruct = self.ctinfo("rtstruct").strip()
			if rtstruct:
				self.rtstruct.set(rtstruct)
			elif rtfirst:
				self.rtstruct.set(rtfirst)

		self.fillRoiMaterial()
		self.materials.sort()

	# ----------------------------------------------------------------------
	def get(self):
		Global.dump("*-* DicomVoxelTab.get", level=2)
		if not self.winfo_ismapped(): return
		if self.page.dicominfo is None or \
		   self.ctinfo is None or \
		   self.dicom is None: return
		oldHash = hash(self.ctinfo)
		self.ctinfo["unit2mat"]  = self.unit2mat.get()
		self.ctinfo["voxel"]     = self.voxelfile.get()
		self.ctinfo["rtstruct"]  = self.rtstruct.get()
		self.ctinfo["field"]     = self.fieldIndex()
		self.getRoiMaterial()

		if hash(self.ctinfo) != oldHash:
			self.page.modified()
		tkFlair.config.set("Dicom", "unit2mat",  self.ctinfo["unit2mat"])

	# ----------------------------------------------------------------------
	def fieldIndex(self):
		mag = self.field.get()
		for n,i in FlukaLayout.ASSIGNMA_field:
			if n==mag:
				if i=="": i=0
				return i
		return 0

	# ----------------------------------------------------------------------
	def loadUnit2Material(self, event=None):
		if self.dicom is None: return
		if self.unit2matModified:
			if not tkFlair.askyesno("Load Unit 2 Material",
					"Current Unit to material list is modified\nDiscard changes",
					parent=self):
				return
		fn = bFileDialog.askopenfilename(master=self,
			title="Unit to material conversion",
			filetypes=[("Unit to material data files","*.mat"),
				("Data files","*.dat"),
				("Text files","*.txt"),
				("All","*")],
			initialdir=self.lastPath,
			initialfile=self.unit2mat.get())

		if not fn: return False

		relfn = self.project.relpath(fn)
		self.lastPath = os.path.abspath(os.path.dirname(relfn))
		self.unit2mat.set(relfn)
		self.readUnit2Material()
		return True

	# ----------------------------------------------------------------------
	def saveUnit2Material(self, event=None):
		if self.dicom is None: return
		#if not self.unit2matModified: return
		self.getHounsfieldFromList()
		fn = self.unit2mat.get()
		tool.backup(fn)
		try:
			with open(fn,"w") as f:
				for hu in self.dicom.hounsfield:
					f.write(f"{hu.str()}\n")
		except:
			self.flair.notify("Cannot open file",
				f"Cannot open file '{fn}' for writing hounsfield units",
				tkFlair.NOTIFY_ERROR)
			return

		self.unit2matModified = False
		self.flair.notify("Unit to Materials file written",
			f"File '{fn} with units to materials updated",
			tkFlair.NOTIFY_NORMAL)

	# ----------------------------------------------------------------------
	def readUnit2Material(self, event=None):
		self.dicom.readHounsfield(self.unit2mat.get())
		self.updateUnit2Material()
		self.unit2matModified = False

	# ----------------------------------------------------------------------
	def editUnit2Material(self, event=None):
		tkFlair.editor(self.unit2mat.get())
		self.updateUnit2Material()

	# ----------------------------------------------------------------------
	def updateUnit2Material(self):
		if self.dicom is None: return
		self.unit2matList.delete(0,tk.END)
		humin = -3072
		for hu in self.dicom.hounsfield:
			self.unit2matList.insert(tk.END,
				(humin, hu.humax-1, hu.mat,
				 hu.crho_min, hu.crho_max,
				 hu.cdedx_min, hu.cdedx_max))
			humin = hu.humax
		self.unit2matList.sort(0,False)

	#-----------------------------------------------------------------------
	# Find all ROIs and fill the menu
	# Handling the visualization of the ROI structures
	# ----------------------------------------------------------------------
	def rtstructChanged(self, event=None):
		self.roi.clear()
		del self.roiList[:]
		rtuid = self.rtstruct.get()
		if not rtuid: return
		roidi = self.project.dicomInfo(rtuid)
		if roidi:
			try:
				dataset = pydicom.read_file(roidi.filename(), force=True)
				for roi in dataset.StructureSetROISequence:
					name = f"{roi.ROIName} [{roi.ROINumber}]"
					self.roi[name]          = roi.ROINumber
					self.roi[roi.ROINumber] = name
					self.roiList.append(name)
			except:
				self.flair.notify("RTSTRUCT Error",
						  "Problem with opening RTSTRUCT File")

	# ----------------------------------------------------------------------
	def getRoiMaterial(self):
		# delete previous material. info from self.ctinfo
		for key in list(self.ctinfo.keys()):
			if key.startswith("material."):
				del self.ctinfo[key]
		idx      = 1
		material = None
		expr     = []
		for i in range(self.roiMaterial.size()):
			mat,op,roi = self.roiMaterial.get(i)
			if mat != "":
				if material is not None:
					self.ctinfo[f"material.{idx}"] = f"{material!r} {' '.join(expr)}"
					idx += 1
				if mat == VoxelTab.ADD:
					material = None
					break
				material = mat
				expr = []

			if op == VoxelTab.ADD:
				continue
			if op != "":
				try:
					roid = self.roi[roi]
					expr.append(f"{op}{roid}")
				except KeyError:
					pass
		if material:
			self.ctinfo[f"material.{idx}"] = f"{material!r} {' '.join(expr)}"

	# ----------------------------------------------------------------------
	# Fill roi material listbox
	# ----------------------------------------------------------------------
	def fillRoiMaterial(self):
		sel = self.roiMaterial.curselection()
		active = self.roiMaterial.index(tk.ACTIVE)
		self.roiMaterial.delete(0,tk.END)
		first   = True
		itemadd = True
		for mat,expr in self.ctinfo.materialFilters():
			if not first and mat!="" and itemadd:
				self.roiInsertOpAdd(tk.END)
			if not expr:
				self.roiMaterial.insert(tk.END, (mat,"",""))
				itemadd = False
			else:
				for j,roid in enumerate(expr):
					roid = int(roid)
					op  = "+" if roid>0 else "-"
					roi = self.roi.get(abs(roid),"")
					self.roiMaterial.insert(tk.END, (mat,op,roi))
					mat = ""
				itemadd = True
			first = False

		if itemadd and not first: self.roiInsertOpAdd(tk.END)
		self.roiInsertAdd()
		self.roiColors()

		if not sel:
			sel = (0,)
			active = 0
		for i in sel:
			self.roiMaterial.selection_set(i)
		self.roiMaterial.activate(active)
		self.roiMaterial.see(sel[0])

	# ----------------------------------------------------------------------
	def roiInsertAdd(self):
		self.roiMaterial.insert(tk.END, (VoxelTab.ADD,"",""))
		self.roiMaterial.itemconfig(tk.END, foreground=tkFlair._CLABEL_FOREGROUND_COLOR)

	# ----------------------------------------------------------------------
	def roiInsertOpAdd(self, pos):
		self.roiMaterial.insert(pos, ("",VoxelTab.ADD,""))
		self.roiMaterial.listbox(1).itemconfig(pos,
				foreground=tkFlair._CLABEL_FOREGROUND_COLOR)

	# ----------------------------------------------------------------------
	# Set the background as alternating color
	# ----------------------------------------------------------------------
	def roiColors(self):
		background = 1
		bgcolor = VoxelTab.BACKGROUNDS[background]
		for i in range(self.roiMaterial.size()):
			mat = self.roiMaterial.listbox(0).get(i)
			if mat:
				background ^= 1
				bgcolor = VoxelTab.BACKGROUNDS[background]
			self.roiMaterial.itemconfig(i, background=bgcolor)

	# ----------------------------------------------------------------------
	def roiRelease(self, event):
		act = event.widget.nearest(event.y)
		if act != self.roiMaterial.index(tk.ACTIVE):
			self.roiMaterial.activate(act)
			if self.roiMaterial.get(act)[0] != VoxelTab.ADD:
				return
		else:
			return self.roiEdit(event)

	# ----------------------------------------------------------------------
	def roiEdit(self, event):
		sel = self.roiMaterial.curselection()
		if len(sel) != 1: return

		listbox = event.widget
		try: active = listbox.index(tk.ACTIVE)
		except: return
		if active != sel[0]: return

		# Find list box
		lid = self.roiMaterial.listboxIndex(listbox)

		# remember current record
		save = self.roiMaterial.get(active)
		savebg = self.roiMaterial.itemcget(active, "background")

		if lid == 0:	# Material
			if save[0]=="": return
			edit = tkExtra.InPlaceSearchList(listbox, values=self.materials, height=10)
			if edit.value is not None:
				self.roiMaterial.focusRight()
				if save[0] == VoxelTab.ADD:
					self.roiInsertAdd()
					self.roiColors()

		# Skip other two listbox if we click on the last line
		elif save[0] == VoxelTab.ADD:
			return

		elif lid == 1:	# Operation
			values = ["+","-"]
			if save[0] != "" and self.roiMaterial.get(active+1)[0] != "":
				values.insert(0,"")
			edit = tkExtra.InPlaceList(listbox, values=values)
			if edit.value is not None:
				self.roiMaterial.focusRight()

		elif lid == 2:	# ROI
			edit = tkExtra.InPlaceSearchList(listbox, values=self.roiList, height=10)
			if edit.value is not None:
				if save[1] in ("",VoxelTab.ADD):
					self.roiMaterial.listbox(1).set(active, "+")
					self.roiMaterial.listbox(1).itemconfig(active,
							background=savebg)
				self.roiMaterial.focusLeft()

		if edit.value is None: return
		listbox.itemconfig(active, background=savebg)
		if save[1] in ("",VoxelTab.ADD):
			following = self.roiMaterial.get(active+1)
			if following[0]==VoxelTab.ADD or following[1] not in ("", VoxelTab.ADD):
				self.roiInsertOpAdd(active+1)
				self.roiMaterial.itemconfig(active+1, background=savebg)

		self.roiMaterial.selection_set(active)
		self.roiMaterial.activate(active)
		self.roiMaterial.see(active)
		self.roiMaterialModified = True
		self.page.modified()
		return "break"

	# ----------------------------------------------------------------------
	def roiAdd(self, event=None):
		active = self.roiMaterial.index(tk.ACTIVE)
		if self.roiMaterial.get(active)[0]!=VoxelTab.ADD:
			active += 1
			while self.roiMaterial.get(active)[0]=="" and active<self.roiMaterial.size():
				active += 1
		self.roiMaterial.insert(active,("VACUUM","",""))
		self.roiColors()
		self.roiMaterial.selection_clear(0,tk.END)
		self.roiMaterial.selection_set(active)
		self.roiMaterial.activate(active)
		self.roiMaterial.see(active)
		#if event is None:
		#	event = tk.Event()
		#event.widget = self.roiMaterial.listbox(0)
		#self.roiEdit(event)

	# ----------------------------------------------------------------------
	def roiDelete(self, event=None):
		sel = list(map(int,self.roiMaterial.curselection()))
		if not sel: return
		for i in reversed(sel):
			item = self.roiMaterial.get(i)
			if VoxelTab.ADD in item[:2]: continue
			if item[0] == "":
				self.roiMaterial.delete(i)
			else:
				for j in range(i+1,self.roiMaterial.size()):
					if self.roiMaterial.get(j)[0]!="":
						break
				self.roiMaterial.delete(i,j-1)
		active = sel[0]
		self.roiMaterial.selection_set(active)
		self.roiMaterial.activate(active)
		self.roiMaterial.see(active)
		self.roiColors()

	# ----------------------------------------------------------------------
	def roiClone(self, event=None):
		pass

	# ----------------------------------------------------------------------
	def roiMoveUp(self, event=None):
		pass

	# ----------------------------------------------------------------------
	def roiMoveDown(self, event=None):
		pass

	# ----------------------------------------------------------------------
	def saveVoxelFile(self, event=None):
		fn = bFileDialog.asksaveasfilename(master=self,
			title="Voxel file",
			filetypes=[("Voxel files","*.vxl"),
				("All","*")],
			initialdir=self.lastPath,
			initialfile=self.voxelfile.get())

		if not fn: return False

		self.lastPath = os.path.abspath(os.path.dirname(fn))
		relfn = self.project.relpath(fn)
		self.voxelfile.set(relfn)
		return True

	# ----------------------------------------------------------------------
	# Create the voxel file
	# ----------------------------------------------------------------------
	def makeVoxel(self):
		self.dicom.scanned(False)	# << Force to re-analyze slices, filter alters the structures
		self.page.analyzeSlices()
		if self.dicom is None or not self.dicom.slices:
			self.flair.notify("No DICOM processed",
				"Please select a DICOM dataset",
				tkFlair.NOTIFY_WARNING)
			return
		self.get()

		if not self.dicom.hounsfield:
			self.flair.notify("No Unit to Material file",
				"Please select or create a 'Unit to Material' conversion file",
				tkFlair.NOTIFY_WARNING)
			return

		if not self.voxelfile.get():
			if not self.saveVoxelFile():
				return

		fn,ext = os.path.splitext(self.voxelfile.get())
		if ext != ".vxl":
			self.voxelfile.set(f"{fn}.vxl")

		# time keeping
		t0 = t00 = time.time()

		# Interpolate correction factors
		if self.unit2matModified:
			self.getHounsfieldFromList()
			self.dicom.interpolateCorrections()

		# Setup the assignmats
		self.dicom.assignMaterials2Regions(self.fieldIndex())

		# Prepare the voxel arrays
		dlg = tkDialogs.ProgressDialog(self, "Creating VOXEL file")
		dlg.setLimits(0.0, self.dicom.nz)
		def update(s,msg=None): return dlg.show(s,msg)

		if not self.dicom.makeVoxel(update):
			self.flair.notify("DICOM to Voxel",
				"Processing stopped from user",
				tkFlair.NOTIFY_ERROR)
			return
		dlg.stop()
		self.flair.log(f"CT processed in {time.time()-t0:g} s")

		# Processing RTSTRUCT
		t0 = time.time()
		if self._processRTStruct():
			self.flair.log(f"RTSTRUCT processed in {time.time()-t0:g} s")

		# Process any material filters
		filters = self.ctinfo.materialFilters()
		if filters:
			self.flair.log(f"Material Filters: {filters}")
			t0 = time.time()
			dlg = tkDialogs.ProgressDialog(self, "Material override")
			dlg.setLimits(0.0, self.dicom.nz)
			def update(s,msg=None): return dlg.show(s,msg)
			self.dicom.filter(filters,update)
			dlg.stop()
			self.flair.log(f"Material override processed in {time.time()-t0:g} s")

		# Finalize voxel by writing data
		if not self.dicom.writeVoxel(self.voxelfile.get()):
			self.flair.notify("DICOM to Voxel",
				"Error writing Voxel file",
				tkFlair.NOTIFY_ERROR)
			return

		# time keeping
		t3 = time.time()
		self.flair.log(f"Voxel generated in {t3-t00:g} s")

		self.flair.notify("DICOM to Voxel",
			f"Voxel file generated successfully from DICOM in {t3-t00:g} s")
		self.flair.refresh("card")

	# ----------------------------------------------------------------------
	def _processRTStruct(self):
		# Processing RTSTRUCT
		rtuid = self.rtstruct.get()
		if rtuid == "": return False

		# Find and open RTSTRUCT dataset
		di = self.project.dicomInfo(rtuid)
		if di is None: return False
		dlg = tkDialogs.ProgressDialog(self, "Processing RTSTRUCTs")
		dlg.setLimits(0, len(self.ctinfo.files))
		def update(s,msg=None): return dlg.show(s,msg)

		rc = self.dicom.processRTStruct(di.filename(), update)

		dlg.stop()
		return rc

	# ----------------------------------------------------------------------
	# Create or modify the input VOXEL card
	# ----------------------------------------------------------------------
	def voxelCard(self, event=None):
		if self.dicom is None: return
		self.page.get()
		undoinfo = self.flair.undolistRefresh()

		inp = self.flair.project.input
		if not inp.cardlist: inp.minimumInput()
		rotpos = inp.bestPosition("ROT-DEFI")

		xo,yo,zo = self.dicom.origin()/10.

		# Add the VOXELS card if not there
		rotdefi_name = "VOXEL"
		fn,ext = os.path.splitext(os.path.basename(self.voxelfile.get()))
		whats  = [fn, xo, yo, zo, ""]
		if not self.dicom.iopIsXYZ():
			whats[4] = rotdefi_name
		prop = { "pos"  : [xo,yo,zo], "iop"  : self.dicom.iop }
		card = inp["VOXELS"]
		if card:
			card = card[0]
			msg  = "Modify VOXELS card"
			undoinfo.append(self.flair.saveCardUndoInfo(card.pos()))
			for i,w in enumerate(whats):
				card.setWhat(i,w)
			for k,v in prop.items():
				card[k] = v
		else:
			try:
				geobegin = inp["GEOBEGIN"][0]
			except IndexError:
				self.flair.notify("Failed",
					"Input file does not have a GEOBEGIN card\n"
					"Please ensure to have a minimal input file before\n"
					"You could load the 'voxel' template from Project->New",
					tkFlair.NOTIFY_ERROR)
				return
			pos = geobegin.pos() + 1
			# add card to input
			card = FlukaCard("VOXELS",whats)
			for k,v in prop.items():
				card[k] = v
			msg = "Add VOXELS card"
			undoinfo.append(self.flair.addCardUndo(card, pos, False))
			if rotpos>pos: rotpos += 1

		# Load the new voxel file
		card.loadVoxel()

		# Delete existing ROT-DEFI with rotdefi_name
		for card in reversed(inp.cardsSorted("ROT-DEFI",-1).values()):
			if card.name()==rotdefi_name:
				rotpos = card.pos()
				undoinfo.append(self.flair.delCardUndo(card.pos(),False))

		# Add the new transformations
		if not self.dicom.iopIsXYZ():
			R = round(math.degrees(self.dicom.Rz),6)
			if R!=0.0:	# rotation around Z
				what = [rotdefi_name, "", "", R]
				card = FlukaCard("ROT-DEFI", what)
				undoinfo.append(self.flair.addCardUndo(card,rotpos,False))
				rotpos += 1
			R = round(math.degrees(self.dicom.Ry),6)
			if R!=0.0:	# rotation around Y
				what = [rotdefi_name, 200.0, "", R]
				card = FlukaCard("ROT-DEFI", what)
				undoinfo.append(self.flair.addCardUndo(card,rotpos,False))
				rotpos += 1
			R = round(math.degrees(self.dicom.Rx),6)
			if R!=0.0:	# rotation around X
				what = [rotdefi_name, 100.0, "", R]
				card = FlukaCard("ROT-DEFI", what)
				undoinfo.append(self.flair.addCardUndo(card,rotpos,False))
				rotpos += 1
			# Add final translation (now embedded in the VOXELS cards)
			# FIXME WARNING on the order
			#what = [rotdefi_name, 100.0, "", "", xo, yo, zo]
			#card = FlukaCard("ROT-DEFI", what)
			#undoinfo.append(self.flair.addCardUndo(card,rotpos,False))
			#rotpos += 1

		# Finalize
		self.flair.addUndolistRefresh(undoinfo,msg)
		self.flair.refresh("card")

		self.flair.notify(msg,
			"A VOXELS card corresponding to the dicom file "
			" has been added/modified to your input")

	# ----------------------------------------------------------------------
	def unit2matRelease(self, event):
		self.unit2matList.activate(event.widget.nearest(event.y))
		return self.unit2matEdit(event)

	# ----------------------------------------------------------------------
	def unit2matEdit(self, event):
		if self.dicom is None: return
		sel = self.unit2matList.curselection()
		if len(sel) != 1: return
		sel = int(sel[0])

		listbox = event.widget
		try: active = listbox.index(tk.ACTIVE)
		except: return
		if active != sel: return

		# Find list box
		lid = self.unit2matList.listboxIndex(listbox)

		# remember current record
		save = self.unit2matList.get(sel)

		if lid == 0:	# HUmin changed
			edit = tkExtra.InPlaceInteger(listbox)
			# Value cannot be:
			#  1. smaller than the previous minimum
			#  2. bigger than the maximum
			try:
				value = int(edit.value)
				if save[0]==VoxelTab.NEW:
					save[0] = save[1] = edit.value
					for i,hu in enumerate(self.unit2matList.get(0,tk.END)):
						try:
							humax = int(hu[1])
						except ValueError:
							self.unit2matList.set(sel,save)
							break
						if value<=humax:
							hu = list(hu)
							hu[1] = str(value-1)
							self.unit2matList.delete(sel)
							self.unit2matList.set(i,hu)
							self.unit2matList.insert(i+1,save)
							self.correctValues()
							active = i+1
							break
				elif value > save[1]:
					self.unit2matList.set(sel, save)	# restore record
				elif sel>0:
					prev = self.unit2matList.get(sel-1)
					if value <= int(prev[0]):
						self.unit2matList.set(sel, save) # restore record
					else:
						# Correct also previous record
						prev[1] = str(value-1)
						self.unit2matList.set(sel-1, prev)
			except (TypeError, ValueError):
				self.unit2matList.set(sel, save)	# restore record

		elif save[0]==VoxelTab.NEW:
			return

		elif lid == 1:	# HUmax changed
			edit = tkExtra.InPlaceInteger(listbox)
			# Value cannot be:
			#  1. bigger than the maximum of the next one
			#  2. smaller than the minimum
			try:
				value = int(edit.value)
				if value < save[0]:
					self.unit2matList.set(sel, save)	# restore record
				elif sel<self.unit2matList.size()-1:
					nxt = self.unit2matList.get(sel+1)
					if value >= int(nxt[1]):
						self.unit2matList.set(sel, save) # restore record
					else:
						# Correct also previous record
						nxt[0] = str(value+1)
						self.unit2matList.set(sel+1, nxt)
			except (TypeError, ValueError):
				self.unit2matList.set(sel, save) # restore record

		elif lid == 2:
			edit = tkExtra.InPlaceSearchList(listbox, values=self.materials, height=10)
		else:
			edit = tkExtra.InPlaceFloat(listbox)
		if edit.value is None: return

		#if self.unit2matList._sortColumn == lid:
		#	self.unit2matList.sort(lid, self.unit2matList._sortReverse)

		self.unit2matList.activate(active)
		self.unit2matList.selection_set(active)
		self.unit2matModified = True
		return "break"

	# ----------------------------------------------------------------------
	# Scan the list to correct the Hounsfield unit values if needed
	# ----------------------------------------------------------------------
	def correctValues(self):
		active = self.unit2matList.index(tk.ACTIVE)
		sel = self.unit2matList.curselection()
		for i,hu in enumerate(self.unit2matList.get(0,tk.END)):
			hu = list(hu)
			try:    humin = int(hu[0])
			except: pass
			try:    humax = int(hu[1])
			except: humax = humin

			if i==0:
				huprev = humax
				continue

			changed = False
			if humin != huprev+1:
				humin = huprev+1
				hu[0] = str(humin)
				changed = True
			if humax < humin:
				humax = humin
				hu[1] = str(humin)
				changed = True
			if changed: self.unit2matList.set(i, hu)
			huprev = humax
		self.unit2matList.activate(active)
		for s in sel: self.unit2matList.selection_set(s)

	# ----------------------------------------------------------------------
	def getHounsfieldFromList(self):
		if self.dicom is None: return
		Dicom.Hounsfield.reset()
		del self.dicom.hounsfield[:]

		self.unit2matList.sort(0,False)

		prev = -9999
		dups = []

		for elem in self.unit2matList.get(0,tk.END):
			#try: humin = int(elem[0])
			#except: continue
			try: humax = int(elem[1])+1
			except: continue
			if humax == prev: dups.append(humax)
			prev = humax
			material = elem[2]
			try:	crho_min = float(elem[3])
			except: crho_min = 1.0
			try:	crho_max = float(elem[4])
			except: crho_max = 1.0
			try:	cdedx_min = float(elem[5])
			except: cdedx_min = 1.0
			try:	cdedx_max = float(elem[6])
			except: cdedx_max = 1.0
			hu = Dicom.Hounsfield(humax, material,
						crho_min, crho_max,
						cdedx_min, cdedx_max)
			self.dicom.hounsfield.append(hu)

		if dups:
			self.flair.notify("Duplicate Hounsfield units",
					f"Duplicate units found {dups}",
					tkFlair.NOTIFY_WARNING)
			self.unit2matList.selection_clear(0,tk.END)
			for i,hu in enumerate(self.unit2matList.listbox(0).get(0,tk.END)):
				if hu in dups:
					self.unit2matList.selection_set(i)

	# ----------------------------------------------------------------------
	def add(self, event=None):
		# add an empty record (only once)
		for hu in self.unit2matList.listbox(0).get(0,tk.END):
			if hu == VoxelTab.NEW:
				break
		else:
			self.unit2matList.insert(tk.END,(VoxelTab.NEW,"","","","","",""))
		self.unit2matList.selection_clear(0,tk.END)
		self.unit2matList.activate(tk.END)
		self.unit2matList.selection_set(tk.END)
		self.unit2matModified = True
		self.page.modified()

	# ----------------------------------------------------------------------
	def delete(self, event=None):
		sel = list(map(int,self.unit2matList.curselection()))
		if not sel: return
		sel.reverse()
		for item in sel:
			self.unit2matList.delete(item)
		self.unit2matModified = True
		self.page.modified()
		self.correctValues()

	# ----------------------------------------------------------------------
	def clone(self, event=None):
		sel = list(map(int,self.unit2matList.curselection()))
		if not sel: return
		for item in sel:
			self.unit2matList.insert(tk.END, self.unit2matList.get(item))
		self.unit2matModified = True
		self.page.modified()
		self.correctValues()
