# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

import math
import bmath
from spline import CardinalSpline

Na     = 6.0221367e23
Qe     = 1.60217646e-19
a      = 1.0 / 137.035989561
h      = 4.135667516e-15		# eV s
c      = 299792458e2
fwhm   = 2.0*math.sqrt(2.0*math.log(2.0))
ProtonAMU = 1.007276470			# Proton mass in amu

#-------------------------------------------------------------------------------
# flair vector returning a string with {} representation
class Vector(bmath.Vector):
	def __str__(self):
		return f"{{{super().__str__()[1:-1]}}}"

#-------------------------------------------------------------------------------
# @return int value also from strings-floats as '35.3'
# and 0 if case of failure
#-------------------------------------------------------------------------------
def __int(x):
	try:
		return int(float(x))
	except:
		return 0

#-------------------------------------------------------------------------------
# @return float on success and 0 on failure
#-------------------------------------------------------------------------------
def __float(x):
	try:
		return float(x)
	except:
		return 0.0

#-------------------------------------------------------------------------------
# Global dict defined in cm, GeV, s
#-------------------------------------------------------------------------------
_globalDict = {
	# Mathematical constants
	"fwhm"	: fwhm,

	# Physics constants
	"a"	: a,
	"h"	: h,	# eV s
	"c"	: c,
	"Na"	: Na,
	"qe"	: Qe,
	"re"	: 2.8179409183694872e-13,

	# Distances
	"nm"	: 0.1E-6,
	"um"	: 0.1E-3,
	"mm"	: 0.1,
	"cm"	: 1.0,
	"dm"	: 10.0,
	"m"	: 100.0,
	"km"	: 100.0e3,

	# Imperial system distances
	"inch"	: 2.54,
	"feet"	: 30.48,
	"ft"	: 30.48,
	"mile"	: 160934.4,
	"mi"	: 160934.4,

	# Masses
	"amu"	: 0.93149432,
	"amuC12": 0.93123890,
	"amugr"	: 1.6605402E-24,
	"Mp"	: 0.93827231,
	"Mn"	: 0.93956563,
	"Me"	: 0.510998910e-3,
	"g"     : 1,
	"kg"    : 1e3,
	"pound" : 0.453592*1e3,

	# Density
	"g/cm3"  : 1.0,
	"mg/cm3" : 1e-3,
	"g/mol"  : 1.0,
	"g/mole" : 1.0,

	# Energy
	"meV"	: 1e-12,
	"eV"	: 1e-9,
	"keV"	: 1e-6,
	"MeV"	: 1e-3,
	"GeV"	: 1.0,
	"TeV"	: 1e3,
	"PeV"	: 1e6,
	"J"     : 1.0/(Qe*1e9),
	"Gy"    : 1.0/(Qe*1e9*1000),

	# Time
	"ns"	: 1.0e-9,
	"us"	: 1.0e-6,
	"ms"	: 0.001,
	"s"	: 1.0,
	"minute": 60.0,
	"hour"	: 3600.0,
	"day"	: 86400.0,
	"week"	: 7.0*86400.0,
	"month"	: 365.24219647*86400.0/12.0,
	"year"	: 365.24219647*86400.0,

	# Temperature
	"K"     : 1.0,

	# Pressure
	"pascal"     : 1./101325.,
	"bar"        : 1./1.01325,
	"atmosphere" : 1.0,

	# Prefixes
	"yotta"	: 1e24,
	"zetta"	: 1e21,
	"exa"	: 1e18,
	"peta"	: 1e15,
	"tera"	: 1e12,
	"giga"	: 1e9,
	"mega"	: 1e6,
	"kilo"	: 1e3,
	"hecto"	: 1e2,
	"deca"	: 1e1,

	"deci"	: 1e-1,
	"centi"	: 1e-2,
	"milli"	: 1e-3,
	"micro"	: 1e-6,
	"nano"	: 1e-9,
	"pico"	: 1e-12,
	"femto"	: 1e-15,
	"atto"	: 1e-18,
	"zepto"	: 1e-21,
	"yocto"	: 1e-24,

	# Angle
	"deg"	: math.pi/180.0,
	"degree": math.pi/180.0,
	"rad"	: 1.0,
	"radian": 1.0,
	"mrad"	: 0.001,

	# Trig functions in degrees
	"sind"	: lambda x : math.sin(math.radians(x)),
	"cosd"	: lambda x : math.cos(math.radians(x)),
	"tand"	: lambda x : math.tan(math.radians(x)),
	"asind"	: lambda x : 180.*math.asin(x)/math.pi,
	"acosd"	: lambda x : 180.*math.acos(x)/math.pi,
	"atand"	: lambda x : 180.*math.atan(x)/math.pi,

	# Generic functions
	"abs"	: abs,
	"float"	: __float,
	"int"	: __int,
	"len"	: len,
	"str"	: str,
	#"type"  : type,

	# Physics functions
	"T2g"	: lambda T,m    : 1. + T/m,
	"g2b"	: lambda g      : math.sqrt(1.-1./g**2),
	"b2g"	: lambda b      : 1./math.sqrt(1.-b**2),
	"p2T"	: lambda p,m    : math.sqrt(p**2 + m**2) - m,
	"T2p"	: lambda T,m    : math.sqrt(T**2 + 2.*m*T),
	"dT2dp"	: lambda T,dT,m : (T+m)/math.sqrt(T**2 + 2.*m*T)*dT,
	"omega"	: lambda x      : 2.*math.pi*(1.-math.cos(x)),
	"omegad": lambda x      : 2.*math.pi*(1.-math.cos(math.radians(x))),
	"X0"	: lambda Z,A    : 716.4*float(A)/(float(Z)*(Z+1.)*math.log(287./math.sqrt(float(Z)))),
	"lhad"	: lambda A      : 35.0*math.pow(A,1./3.),

	# Vector and matrices
	"Vector": Vector,
	"Matrix": bmath.Matrix,
}

# Populate math functions on the local dict
for _n in dir(math):
	if _n[0]=="_": continue
	_globalDict[_n] = getattr(math, _n)

#-------------------------------------------------------------------------------
# return a float or int otherwise as is
#-------------------------------------------------------------------------------
def float_or_int(val):
	try:
		f = float(val)
		if float(int(f)) == f:
			return int(f)
		return f
	except:
		return val

#===============================================================================
# Global dictionary with units which should be sub classed for each input mode
#===============================================================================
class GlobalDict(dict):
	def __init__(self):
		super().__init__()
		self.update(_globalDict)

	#-----------------------------------------------------------------------
	# Redefine length based on cm, e.g for mm 1cm=10
	#-----------------------------------------------------------------------
	def redefineDistance_cm(self, cm):
		# Distances
		self["nm"] *= cm
		self["um"] *= cm
		self["mm"] *= cm
		self["cm"] *= cm
		self["dm"] *= cm
		self["m"]  *= cm
		self["km"] *= cm

		# Imperial system distances
		self["inch"] *= cm
		self["feet"] *= cm
		self["ft"]   *= cm
		self["mile"] *= cm
		self["mi"]   *= cm

	#-----------------------------------------------------------------------
	# Redefine energy based on GeV
	#-----------------------------------------------------------------------
	def redefineEnergy_GeV(self, GeV):
		# Masses
		self["amu"]    *= GeV
		self["amuC12"] *= GeV
		self["amugr"]  *= GeV
		self["Mp"]     *= GeV
		self["Mn"]     *= GeV
		self["Me"]     *= GeV

		# Energy
		self["meV"] *= GeV
		self["eV"]  *= GeV
		self["keV"] *= GeV
		self["MeV"] *= GeV
		self["GeV"] *= GeV
		self["TeV"] *= GeV
		self["PeV"] *= GeV
		self["J"]   *= GeV

	#-----------------------------------------------------------------------
	# Redefine time based on s
	#-----------------------------------------------------------------------
	def redefineTime_s(self, s):
		# Time
		self["ns"]     *= s
		self["us"]     *= s
		self["ms"]     *= s
		self["s"]      *= s
		self["minute"] *= s
		self["hour"]   *= s
		self["day"]    *= s
		self["week"]   *= s
		self["month"]  *= s
		self["year"]   *= s

#===============================================================================
# Input Local Dictionary
#===============================================================================
class LocalDict(dict):
	"""Implement the local dictionary"""
	def __init__(self, input_):
		super().__init__()
		self.input = input_
		self.card  = None		# used for the what()
		self.clear()
		from deck import Card
		LocalDict.BODY = Card.BODY

	# ----------------------------------------------------------------------
	# Clear dictionary and set default variables
	# ----------------------------------------------------------------------
	def clear(self):
		super().clear()
		self["what"]   = lambda w,s=self : s.card.numWhat(w)
		self["body"]   = self.bodyWhat
		self["card"]   = self.cardWhat
		self["spline"] = self.spline
		self["frame"]  = 0

	# ----------------------------------------------------------------------
	# Default getitem
	# ----------------------------------------------------------------------
	def __getitem__(self, item):
		try:
			return super().__getitem__(item)
		except:
			# check defines
			defines = self.input.cardsCache("#define")
			try:
				return float_or_int(defines[item].evalWhat(1))
			except KeyError:
				pass
			if item not in self.input.globalDict:
				return str(item)
			# FIXME check for a card like BEAM(0,1)
			#if cards: return lambda n,w,t=item,s=self: s.cardWhat(t,n,w)
			#return str(item)
			raise

	# ----------------------------------------------------------------------
	# return what of body
	# ----------------------------------------------------------------------
	def bodyWhat(self, name, what):
		body = self.input.cardsCache(LocalDict.BODY)[name] # 2) #Card.BODY)
		return body.numWhat(what)

	# ----------------------------------------------------------------------
	# return a card's what
	# ----------------------------------------------------------------------
	def cardWhat(self, tag, name, what):
		cards = self.input.cardsCache(tag)
		if cards is None:
			raise Exception(f"Card {tag!r} do not exist")
		if isinstance(name,str):
			try:
				val = cards[name].evalWhat(what)
			except KeyError:
				raise Exception(f"No card {tag!r} with name={name!r} found")
		else:
			val = list(cards.values())[name].evalWhat(what)
		# return a float or int otherwise as is
		return float_or_int(val)

	# ----------------------------------------------------------------------
	# evaluate a spline and return its value
	# ----------------------------------------------------------------------
	def spline(self, name, t):
		cards = self.input.cardsCache("spline")
		if cards is None:
			raise Exception("Card 'spline' do not exist")
		try:
			card = cards[name]
		except KeyError:
			raise Exception(f"No card 'spline' with name={name!r} found")

		N = card.intProperty("node:0")
		xo = card.numWhat(1)
		yo = card.numWhat(2)
		zo = card.numWhat(3)

		option  = card.intWhat(4)
		closed  = bool(option & 1)	# open=0,  closed=1
		lines   = bool(option & 2)	# spline=0, lines=2
		tension = card.numWhat(7, 0.5)
		Nnodes = N-1 if closed else N	# ignore last node for position, but not for time

		#---------------------------------------------------------------
		# @return node vector
		#---------------------------------------------------------------
		def nodeVector(node):
			nonlocal card, xo, yo, zo, Nnodes, closed
			# last node is only used for time
			if closed and node==Nnodes: Nnodes = 1
			x = card.numProperty(f"node:{node}.x")
			y = card.numProperty(f"node:{node}.y")
			z = card.numProperty(f"node:{node}.z")
			return bmath.Vector(x+xo, y+yo, z+zo)

		#---------------------------------------------------------------
		# @return true if we have to ignore the node equal with the previous
		#---------------------------------------------------------------
		def nodeIgnore(node):
			nonlocal Nnodes
			if node<2: return False
			x = card.numProperty(f"node:{node}.x")
			y = card.numProperty(f"node:{node}.y")
			z = card.numProperty(f"node:{node}.z")
			node -= 1
			xp = card.numProperty(f"node:{node}.x")
			yp = card.numProperty(f"node:{node}.y")
			zp = card.numProperty(f"node:{node}.z")
			if (x-xp)**2 + (y-yp)**2 + (z-zp)**2 < 1e-10: return True
			return False

		#---------------------------------------------------------------
		# @return previous non-ignored node
		#---------------------------------------------------------------
		def nodePrev(node):
			nonlocal Nnodes, closed
			while True:
				node -= 1
				if node<1: return Nnodes if closed else 1
				if not nodeIgnore(node): return node

		#---------------------------------------------------------------
		# @return next non-ignored node
		#---------------------------------------------------------------
		def nodeNext(node):
			nonlocal Nnodes, closed
			while True:
				node += 1
				if node>Nnodes: return 1 if closed else Nnodes
				if not nodeIgnore(node): return node

		#---------------------------------------------------------------
		# WARNING node.t must be ascending
		if t <= card.numProperty("node:1.t"):
			return nodeVector(1)

		elif t >= card.numProperty(f"node:{N}.t"):	# for time use all nodes
			return nodeVector(1 if closed else Nnodes)

		# find interval of nodes with bisection
		nlow  = 1
		nhigh = N					# for time use all nodes
		while nlow+1 < nhigh:
			nmid = (nlow+nhigh)//2
			tmid = card.numProperty(f"node:{nmid}.t")
			if t < tmid:
				nhigh = nmid
			elif t > tmid:
				nlow = nmid
			else:
				return nodeVector(nmid)

		# it lies in between nlow..nlow+1 skip same nodes
		if nlow < Nnodes and nodeIgnore(nlow+1):
			return nodeVector(nlow)			# are standing still here

		t1 = card.numProperty(f"node:{nlow}.t")
		if nodeIgnore(nlow):
			nlow = nodePrev(nlow)			# start from a non-ignored node

		nidx = [nodePrev(nlow), nlow, nodeNext(nlow)]
		nidx.append(nodeNext(nidx[-1]))

		if closed and nidx[2]==1:
			t2 = card.numProperty(f"node:{N}.t")
		else:
			t2 = card.numProperty(f"node:{nidx[2]}.t")

		try:
			tfraction = (t-t1) / (t2-t1)
		except ZeroDivisionError:
			return nodeVector(nidx[1])

		# Convert to lines if needed
		if abs(tension)<1e-10 or \
		   (not closed and Nnodes<4) or (closed and Nnodes<3):
			lines = True

		pos = bmath.Vector(xo,yo,zo)
		if lines:
			# Linear interpolation between nidx[1] .. nidx[2]
			for i,s in enumerate("xyz"):
				v1 = card.numProperty(f"node:{nidx[1]}.{s}")
				v2 = card.numProperty(f"node:{nidx[2]}.{s}")
				pos[i] += tfraction*(v2-v1) + v1
		else:
			# Cardinal spline interpolation
			P = [0.0]*4
			spline = CardinalSpline(tension)
			for i,s in enumerate("xyz"):
				for j in range(4):
					P[j] = card.numProperty(f"node:{nidx[j]}.{s}")
				pos[i] += spline(P,tfraction)
		return pos
