#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   06-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re
import stat
import time
import fnmatch
import tkinter as tk
from tkinter import ttk

import pipe
import tool
import icons
import inputs
import Ribbon
import tkExtra
import bFileDialog
from log import say

import Global
import Manual
import tkFlair
import FlairRibbon

from fluka import FlukaConfig

_INPUT_CYCLE = "input"
_DATA_CYCLE  = "data"
_PLOT_CYCLE  = "plot"
_COMP_CYCLE  = "compile"
_TEMP_CYCLE  = "temporary"
SPECIAL_CYCLES = [_INPUT_CYCLE, _DATA_CYCLE, _PLOT_CYCLE, _COMP_CYCLE, _TEMP_CYCLE]

#===============================================================================
# Files Tab in Page Frame
#===============================================================================
class FilesTab(FlairRibbon.FlairTab):
	#----------------------------------------------------------------------
	def __init__(self, master, page, **kw):
		super().__init__(master, page, "Files", **kw)

		# --- Run-Cycle splitter ---
		#splitter = tk.PanedWindow(self,
		#			orient=tk.HORIZONTAL,
		#			showhandle=0,
		#			sashrelief=tk.RAISED,
		#			sashwidth=4,
		#			opaqueresize=1)
		#splitter.pack(expand=tk.YES, fill=tk.BOTH)

		# --- Cycles ---
		#frame = tk.Frame(splitter)
		#splitter.add(frame, minsize=40)
		frame = tk.Frame(self)
		frame.pack(side=tk.LEFT, fill=tk.Y)

		tk.Label(frame,
				text="Cycles",
				borderwidth=1,
				relief=tk.RAISED).pack(side=tk.TOP,fill=tk.X)

		self.cycleList = tkExtra.ExListbox(frame, borderwidth=0,
				selectmode=tk.EXTENDED,
				selectborderwidth=0,
				takefocus=True,
				width=15,
				height=7,
				exportselection=tk.FALSE)
		self.cycleList.pack(side=tk.LEFT,expand=tk.YES,fill=tk.BOTH)

		sb = ttk.Scrollbar(frame, orient=tk.VERTICAL, takefocus=False,
			command=self.cycleList.yview)
		sb.pack(side=tk.RIGHT,fill=tk.Y)
		self.cycleList.config(yscrollcommand=sb.set)

		# Bind events
		self.cycleList.bind('<<ListboxSelect>>', self.refresh)

		# --- Files ---
		frame = tk.Frame(self)
		frame.pack(side=tk.RIGHT, fill=tk.BOTH, expand=tk.YES)

		self.fileList = tkExtra.ColorMultiListbox(frame, #self, #frame,
				(('File', 30, None),
				 ('Type', 10, None),
				 ('Size',  8, None),
				 ('Date', 15, None)),
				 background="White")
		self.fileList.pack(side=tk.RIGHT, expand=tk.YES, fill=tk.BOTH)

		# Bind events
		self.fileList.bindList('<Double-1>',     self.viewer)
		self.fileList.bindList('<Return>',       self.viewer)
		self.fileList.bindList('<KP_Enter>',     self.viewer)
		self.fileList.bindList('<Delete>',       self.delete)

		self.fileList.setPopupMenu(
				[('View',     0, self.viewer,  icons.get("view")),
				 ('Edit',     0, self.editor,  icons.get("edit")),
				 ('To ascii', 0, self.toAscii, icons.get("ascii")),
				 ('Refresh',  0, self.refresh, icons.get("refresh")),
				 None,
				 ('Delete',   0, self.delete,  icons.get("x"))] )
		self.ascending = True

		self.pattern   = None
		self.totalSize = 0
		self._files    = {}	# to avoid duplicates

	# ----------------------------------------------------------------------
	# create the ribbon buttons for the Run tab
	# ----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		group = Ribbon.LabelGroup(self.ribbon, "Files")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.label["background"] = Ribbon._BACKGROUND_GROUP3

		group.grid2rows()

		# ---
		col,row = 0,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("ascii"),
				text="\N{RIGHTWARDS ARROW} Ascii",
				command=self.toAscii)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Convert file to ASCII")

		# ----
		col,row = 0,1
		b = Ribbon.LabelButton(group.frame, image=icons.get("x"),
				text="Delete",
				command=self.delete)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Delete files [Del]")

		# ---
		col,row = 0,2
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("refresh"),
				text="Refresh",
				command=self.refresh)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Refresh files")

		# ---
		col,row = 1,1
		self.patternEntry = tkExtra.LabelEntry(group.frame,
					"Filter",
					"DarkGray",
					background="White", #Ribbon._BACKGROUND,
					width=20,
					font=Ribbon._FONT)
		self.patternEntry.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self.patternEntry,
				"Filter files with a pattern (wildchars: * and ?)")
		self.patternEntry.bind("<Return>",   self.refresh)
		self.patternEntry.bind("<KP_Enter>", self.refresh)
		self.patternEntry.bind("<Button-3>", self.patternClear)

		b = Ribbon.LabelButton(group.frame, image=icons.get("filter"),
				command=self.refresh)
		b.grid(row=row, column=col+1, padx=0, pady=0, sticky=tk.E)
		tkExtra.Balloon.set(b, "Filter files with pattern")

		# ======================
		group = Ribbon.LabelGroup(self.ribbon, "Action")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.label["background"] = Ribbon._BACKGROUND_GROUP3

		group.grid3rows()

		# ----
		col,row = 0,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("viewer32"),
				text="Viewer",
				compound=tk.TOP,
				command=self.viewer)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Open in viewer")

		# ----
		col,row = 1,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("edit32"),
				text="Editor",
				compound=tk.TOP,
				command=self.editor)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Open in editor")

		return self.ribbon

	# ----------------------------------------------------------------------
	def activate(self):
		self.fillCycles()
		self.refreshStatus()

	# ----------------------------------------------------------------------
	# Lose focus
	# ----------------------------------------------------------------------
	def _focusOut(self, event=None):
		if self.project is None: return
		self.page.runList.updateSelection()

	# ----------------------------------------------------------------------
	def patternClear(self, event=None):
		self.patternEntry.delete(0,tk.END)
		self.pattern = ""
		self.refresh()

	# ----------------------------------------------------------------------
	def selectAll(self):
		self.page.runList.selection_set(0,tk.END)
		self.fillCycles()

	# ----------------------------------------------------------------------
	def selectClear(self):
		self.page.runList.selection_clear(0,tk.END)
		self.fillCycles()

	# ----------------------------------------------------------------------
	# Select Run
	# ----------------------------------------------------------------------
	def runSelect(self, event):
		i = self.page.runList.nearest(event.y)
		self.page.runList.focus_set()
		self.page.runList.activate(i)
		self.fillCycles()

	# ----------------------------------------------------------------------
	# Fill Run Cycles
	# ----------------------------------------------------------------------
	def fillCycles(self, event=None, select=True):
		cycles = SPECIAL_CYCLES[:]	#copy
		selected_cycles = set(map(self.cycleList.get, self.cycleList.curselection()))
		for run in self.page.runList.selectedRuns(True):
			inpname = run.inputBaseName()
			inpfile = f"{inpname}.inp"
			inpname = re.escape(inpname)
			rundir  = run.dir()

			cpat1  = re.compile(rf"^{inpname}(\d\d\d)[._].+$")
			cpat2  = re.compile(rf"^ran{inpname}(\d\d\d)$")
			patout = re.compile(rf"^{inpname}(\d\d\d).out$")
			patlog = re.compile(rf"^{inpname}(\d\d\d).log$")

			config = run.config()

			try:
				os.stat(rundir)
			except OSError:
				continue

			for f in tool.listdir(rundir):
				m = cpat1.match(f)
				if m is None: m = cpat2.match(f)
				if m:
					c = m.group(1)
					if c not in cycles:
						cycles.append(c)

				# Check sub directory if ...out file exist
				if not config.CYCLEDIR_PAT.match(f): continue

				fn = os.path.join(rundir,f)
				try: s = os.lstat(fn)
				except OSError: continue

				if f.startswith(config.CYCLE) and stat.S_ISDIR(s.st_mode):
					for fsub in tool.listdir(fn):
						if fsub==inpfile or \
						   patout.match(fsub) or \
						   patlog.match(fsub):
							cycles.append(f)
							break
		cycles.sort()

		# Populate list
		self.cycleList.delete(0,tk.END)
		selected = False
		for cycle in cycles:
			self.cycleList.insert(tk.END,cycle)
			if cycle in selected_cycles:
				self.cycleList.selection_set(tk.END)
				selected = True

		if not selected:
			self.cycleList.selection_set(0)

		if select: self.refresh()

	# ----------------------------------------------------------------------
	# Select Cycle
	# ----------------------------------------------------------------------
	def cycleSelect(self, event):
		i = self.page.runList.nearest(event.y)
		self.cycleList.focus_set()
		self.cycleList.activate(i)
		self.refresh()

	# ----------------------------------------------------------------------
	def _compileFiles(self):
		totalSize = 0

		# Source files
		d = self.project.dir()
		for f in self.project.compile("source"):
			fn,ext = os.path.splitext(f)
			if ext in (".f", ".for", ".fpp", ".c", ".cpp", ".cc"):
				fo = f"{fn}.o"
				try:
					os.stat(fo)
					totalSize += self.addFile(d, fo)
				except:
					pass

		# ... executable
		try:
			os.stat(self.project.compile.name)
			totalSize += self.addFile(d, self.project.compile.name)
		except:
			pass

		# ... and map
		fn,ext = os.path.splitext(self.project.compile.name)
		fm = f"{fn}.map"
		try:
			os.stat(fm)
			totalSize += self.addFile(d, fm)
		except:
			pass

		return totalSize

	# ----------------------------------------------------------------------
	def _plotFiles(self):
		totalSize = 0
		d = self.project.dir()
		for f in tool.listdir(d):
			fn,_ = os.path.splitext(f)
			if fn in ("gplevh","doslev"):
				totalSize += self.addFile(d, f)
			else:
				for plot in self.project.plots():
					if fn in (plot.name, plot.name+".geo"):
						totalSize += self.addFile(d, f)
		return totalSize

	# ----------------------------------------------------------------------
	def _tmpFiles(self):
		totalSize = 0
		tmp = Global.tmpPrefix
		rantmp = f"ran{tmp}"
		d = self.project.dir()
		for f in tool.listdir(d):
			fn,_ = os.path.splitext(f)
			if fn.startswith(tmp) or fn.startswith(rantmp):
				totalSize += self.addFile(d, f)
		return totalSize

	# ----------------------------------------------------------------------
	# Refresh file list
	# ----------------------------------------------------------------------
	def refresh(self, event=None):
		self.page.busy()
		self.page.runList.refreshRuns()
		savesort = self.fileList.saveSort()
		#runs   = self.page.runList.curselection()
		cycles = [self.cycleList.get(int(x))
				for x in self.cycleList.curselection()]
		self.pattern = self.patternEntry.get()

		self._files.clear()
		self.fileList.delete(0,tk.END)
		self.totalSize = 0
		first = True

		tstart = time.time()
		#self.listdir.clear()

		for run in self.page.runList.selectedRuns(True):
			inpname = re.escape(run.inputBaseName())
			rundir  = run.dir()
			mainout = False

			# Populate usrinfo if needed
			if not run.usrinfo:
				self.project.findRunUsrInfo(run)

			nohup = os.path.join(rundir, "nohup.out")
			if os.path.isfile(nohup):
				self.totalSize += self.addFile(rundir, "nohup.out")

			for c in cycles:
				if c == _INPUT_CYCLE:
					for ext in (".inp", ".fluka", ".moira", ".macro", ".mac", ".out"):
						filename = f"{rundir}/{run.inputBaseName()}{ext}"
						if os.path.isfile(filename):
							self.totalSize += self.addFile(rundir, filename)

				elif c == _DATA_CYCLE:
					# Populate with data files
					for usr in run.usrinfo:
						for f in usr.provides(True):
							self.totalSize += self.addFile("", f)

				elif c == _COMP_CYCLE and first:
					# Show compile only for the first cycle
					self.totalSize += self._compileFiles()

				elif c == _PLOT_CYCLE and first:
					self.totalSize += self._plotFiles()

				elif c == _TEMP_CYCLE and first:
					self.totalSize += self._tmpFiles()

				elif c.startswith(run.config().CYCLE):
					cdir = os.path.join(rundir,c)
					try:
						os.stat(cdir)
						self.addFile(rundir, c)
						for f in os.listdir(cdir):
							self.totalSize += self.addFile(cdir, f,
								not tkFlair._showFlukaFiles)
					except OSError:
						pass
				else:
					# Populate with cycle files
					cpat1 = re.compile(rf"^{inpname}{c}[._].+$")
					cpat2 = re.compile(rf"^ran{inpname}{c}$")
					try:
						for f in tool.listdir(rundir):
							if cpat1.match(f) or cpat2.match(f):
								self.totalSize += self.addFile(rundir, f)
					except OSError as err:
						self.flair.notify("OSError", str(err),
								tkFlair.NOTIFY_WARNING)

					if not mainout:
						mainout = True
						self.totalSize += self.addFile(rundir,
									f"{run.inputBaseName()}.out")

			first = False

		# Display total
		Global.dump("+++ FilesTab::refresh time=",time.time() - tstart)
		self.fileList.restoreSort(savesort)
		self.page.notBusy()

	# ----------------------------------------------------------------------
	def refreshStatus(self):
		self.page.setStatus(f"Files: {self.fileList.size()} Total Size: {self.totalSize}")

	# ----------------------------------------------------------------------
	# Add a file to the list
	# ----------------------------------------------------------------------
	def addFile(self, path, filename, checklink=False):
		if self.pattern:
			if not fnmatch.fnmatch(filename, self.pattern):
				return 0
		f  = os.path.join(path, filename)
		if f in self._files: return 0
		self._files[f] = True
		fn = self.project.relpath(f)
		try:
			s = os.lstat(f)
			if checklink and stat.S_ISLNK(s.st_mode): return 0
			ext,color = bFileDialog.fileTypeColor(f)
			self.fileList.insert(tk.END, (fn, ext, s.st_size,
				 time.strftime(bFileDialog._TIME_FORMAT,
					time.localtime(s.st_mtime))))
			if color is not None:
				self.fileList.itemconfig(tk.END,foreground=color)
			return s.st_size
		except OSError:
			#self.fileList.insert(tk.END, (fn, "", 0, "-"))
			return 0

	# ----------------------------------------------------------------------
	# Delete selected files
	# ----------------------------------------------------------------------
	def delete(self, event=None):
		#log = self.flair.newLog("Files", "Files")
		files = [ self.fileList.get(int(x))[0]
				for x in self.fileList.curselection() ]
		if not files:
			self.flair.notify("No files selected",
				"Please select the files you want to delete",
				tkFlair.NOTIFY_WARNING)
			return

		elif len(files)>20:
			msg = f"Are you sure you want to delete {len(files)} files?"
		else:
			msg = "Are you sure you want to delete the files:\n" + \
				' '.join(files)

		if not tkFlair.askyesno("Delete Files", msg, parent=self.ribbon): return

		# First delete files
		refreshCycles = False

		fn = self.project.name()
		skip = [fn+x for x in inputs.ALL_EXTS]
		for f in files:
			try: s = os.lstat(f)
			except OSError: continue

			# Skip files
			if f in skip: continue

			if stat.S_ISDIR(s.st_mode): continue
			if stat.S_ISLNK(s.st_mode):
				say(f"Removing link: {f}")
			else:
				say(f"Deleting file: {f}")
			try:
				os.remove(f)
			except OSError:
				say(f"Error deleting file: {f}")

		# Try to delete subdirectories if empty
		for f in files:
			try: s = os.stat(f)
			except OSError: continue
			if not stat.S_ISDIR(s.st_mode): continue

			# Try to delete all links in directory
			if not tkFlair._showFlukaFiles:
				for fn in os.listdir(f):
					ff = os.path.join(f,fn)
					s = os.lstat(ff)
					if stat.S_ISLNK(s.st_mode):
						say(f"Removing link: {ff}")
						try:
							os.remove(ff)
						except OSError:
							say(f"Error removing link: {ff}")

			say(f"Removing directory: {f}")
			try:
				os.rmdir(f)
				refreshCycles = True
			except OSError:
				say(f"Error removing directory: {f}")

		self.refresh()
		if refreshCycles: self.fillCycles()

	# ----------------------------------------------------------------------
	# Edit files
	# ----------------------------------------------------------------------
	def editor(self):
		tkFlair.editor(*[self.project.abspath(self.fileList.get(int(x))[0])
				for x in self.fileList.curselection()])

	# ----------------------------------------------------------------------
	# Edit files
	# ----------------------------------------------------------------------
	def viewer(self, event=None):
		self.flair.viewer(*[self.project.abspath(self.fileList.get(int(x))[0])
				for x in self.fileList.curselection()])

	# ----------------------------------------------------------------------
	def toAscii(self, event=None):
		cmd = FlukaConfig.command("usbrea")

		files = []
		#log = self.flair.newLog("Files", "Files")
		for x in self.fileList.curselection():
			f = self.fileList.get(int(x))[0]
			files.append(f)

			say(f"Processing: {f} cmd={cmd}")
			self.page.setStatus(f"Converting {f} to ascii")

			inp = [f, f"{f}.lis"]
			try:
				rc, out = pipe.system(cmd, inp)
			except (KeyboardInterrupt, IOError):
				say("*** Interrupt ***")
				return True
			for line in out:
				say(line)
		self.refresh()
		self.flair.notify("To Ascii",
			f"Files converted to ascii: {' '.join(files)}\n")

	# ----------------------------------------------------------------------
	def help(self):
		Manual.show("","Run:Files")
