# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   05-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re
import math
import time
from tkinter import messagebox

import log
import tool
import bmath

import Data
import Plot
import tkFlair
import Gnuplot
import Evaluation
from palette import paletteStore

#===============================================================================
# ----------------------------------------------------------------------
# Base class to start plotting (Launches gnuplot)
# ----------------------------------------------------------------------
def start():
	engine = Gnuplot.Gnuplot()
	while 20:	# wait a while to get version
		try:
			ver = engine.version()
		except:
			return None
		if ver is not None: break
	while 2:
		Gnuplot.defaultTerminal = engine.getTerminal()
		if Gnuplot.defaultTerminal is not None: break
	else:
		Gnuplot.defaultTerminal = "x11"

	return engine

#===============================================================================
# Gnuplot Plot class
#===============================================================================
class GPPlot(Plot.Plot):
	def __init__(self, engine, project, plot, logger=None):
		super().__init__(engine, project, plot, logger)
		self.engine.setLogger(logger)

	# ----------------------------------------------------------------------
	# Send a string command to gnuplot and copy it on output
	# ----------------------------------------------------------------------
	def __call__(self, s):
		self.engine.write(s)

	# ----------------------------------------------------------------------
	# Save plot to file
	# ----------------------------------------------------------------------
	def save(self, ext):
		filename = super().save(ext)
		if ext in (".gnu", ".gp"):
			self(f"save \'{filename}\'")
		else:
			opt = tkFlair._PLOT_FORMAT.get(ext,None)
			if opt:
				self(f"set terminal {opt} {self.plot('linetype')}")
			else:
				self(f"set terminal {ext[1:]}")
			self.engine.output(filename)
			self.engine.replot()
			self("set terminal pop")
			self.engine.output()
		self.checkOutput()
		return filename

	# ----------------------------------------------------------------------
	# Send plot on printer
	# ----------------------------------------------------------------------
	def hardcopy(self, printer):
		super().hardcopy(printer)
		if printer.landscape:
			orient = "landscape"
		else:
			orient = "portrait"

		self(f"set terminal postscript {orient} {self.plot('linetype')} enhanced color")

		if printer.printTo:
			self.engine.output("|"+printer.command())
		else:
			self.engine.output(printer.filename)
		self.engine.replot()
		self("set terminal pop")
		self.engine.output()

	# ----------------------------------------------------------------------
	def _startPlot(self):
		lt = str(self.plot("linetype"))

		term = tkFlair._gnuplotTerminal
		if term == "":
			term = Gnuplot.defaultTerminal
		if term == "wxt":
			lt = ""	# Don't use line type with wxt

		if term:
			idx = self.plot('index')
			title = self.engine.escape(self.windowtitle())
			self(f"set terminal {term} {idx} enhanced {lt} title '{title}'")
		self(tkFlair._PLOT_COMMANDS)
		super()._startPlot()

		#if self.plot("size_x") != "" and self.plot("size_y") != "":
		#	if setsize == "": setsize = "set size"
		#	setsize = "%s %s,%s" % (setsize, self.plot("size_x"), self.plot("size_y"))
		#	self(setsize)
		self("set key default")
		self(f"plotname='{self.plot.name}'")

	# ----------------------------------------------------------------------
	# @return plot options for prefix: title, x, y, ...
	# ----------------------------------------------------------------------
	def _plotOptions(self, prefix):
		opts = self.plot(f"{prefix}options")
		if opts: return opts
		font  = self.plot(f"{prefix}font")
		size  = self.plot(f"{prefix}size")
		color = self.plot(f"{prefix}color")
		if font or size:
			opts = f"font '{font},{size}'"
		if color:
			opts += f" tc rgb '{color}'"
		return opts

	# ----------------------------------------------------------------------
	# Set color band
	# ----------------------------------------------------------------------
	def colorBand(self, pm3dopts=None):
		cbmin,cbmax = self.cbLimits("*","*")
		self(f"set cbrange [{cbmin}:{cbmax}]")
		if int(self.plot("cblog")):
			self("set logscale cb")
			#if 4.4 < self.engine.version() < 5.0:
			#	self("set logscale z")	# Necessary for 4.4.4+
							# to treat log(0) as 0 and not as undefined
		try:
			ncolors = int(self.plot("cbcolors"))
		except ValueError:
			ncolors = 30
		self.plotPm3D(ncolors, self.plot("cbpalette"), pm3dopts)

	# ----------------------------------------------------------------------
	# Get axes range from engine
	# ----------------------------------------------------------------------
#	def getAxesRangeFromGnuplot(self):
#		if not self.engine: return
#		xlow, xhigh = self.engine.getRange("x")
#		ylow, yhigh = self.engine.getRange("y")
#		try:
#			zlow, zhigh = self.engine.getRange("z")
#		except:
#			pass
#
#		self.resetAxesRange(False)
#		if xlow is not None:  self.x_min.insert(0, xlow)
#		if ylow is not None:  self.y_min.insert(0, ylow)
#		if xhigh is not None: self.x_max.insert(0, xhigh)
#		if yhigh is not None: self.y_max.insert(0, yhigh)
#
#		try:
#			if zlow is not None:  self.z_min.insert(0, zlow)
#			if zhigh is not None: self.z_max.insert(0, zhigh)
#		except:
#			pass
#
#		try:
#			if self.x2_tics.get():
#				self.x2_min.delete(0,END)
#				self.x2_max.delete(0,END)
#				xlow, xhigh = self.engine.getRange("x2")
#
#				if xlow is not None:  self.x2_min.insert(0, xlow)
#				if xhigh is not None: self.x2_max.insert(0, xhigh)
#
#			if self.y2_tics.get():
#				self.y2_min.delete(0,END)
#				self.y2_max.delete(0,END)
#				ylow, yhigh = self.engine.getRange("y2")
#
#				if ylow is not None:  self.y2_min.insert(0, ylow)
#				if yhigh is not None: self.y2_max.insert(0, yhigh)
#		except:
#			pass

	# ----------------------------------------------------------------------
	# Plot Using
	# return using string for line plots depending on parameters requested
	# ----------------------------------------------------------------------
	def _plotUsing(self, style, x, y, ex, ey):
		try:
			specs = Gnuplot.SPEC[Gnuplot.STYLE.index(style)]
		except ValueError:
			specs = 2
		if specs == 2:
			return f"{x}:{y}"
		elif specs == 3:
			if style in ["xerrorbars", "xerrorlines"]:
				return f"{x}:{y}:{ex}"
			else:
				return f"{x}:{y}:{ey}"
		elif specs == 4:
			return f"{x}:{y}:{ex}:{ey}"

	# ----------------------------------------------------------------------
	# Plot Style
	# @return string with plot options
	# ----------------------------------------------------------------------
	def _plotStyle(self, det, style=None):
		cmd = []
		# axes
		axes = self.plot(f"axes.{det}")
		if axes not in ("","x1y1"):
			cmd.append(f"axes {axes}")

		# with
		if style is None:
			style = self.plot(f"with.{det}")
			if style in ("histogram","histerror"): style = "steps"
		cmd.append(f"w {style}")

		# smooth
		smooth = self.plot(f"smooth.{det}")
		if smooth != "":
			cmd.append(f"smooth {smooth}")

		# line style
		#if style not in ["points", "dots"]:
		x = int(self.plot(f"ls.{det}"))
		if x:
			cmd.append(f"ls {x}")
		else:
			x = self.plot(f"lt.{det}")
			if x: cmd.append(f"lt {x}")

			x = self.plot(f"lw.{det}")
			if x: cmd.append(f"lw {x}")

			try:
				x = int(self.plot(f"dt.{det}"))
				if x: cmd.append(f"dt {x}")
			except:
				pass

			if self.engine.version() >= 4.2:
				x = self.plot(f"lc.{det}")
				if x: cmd.append(f"lc rgb '{x}'")

			# point style
			if style not in ["lines", "steps", "impulses", "fsteps", "dots",
					"filledcurves", "histeps", "boxes",
					"boxerrorbars", "boxxyerrorbars"]:
				cmd.append("pt %s ps %s"
					% (self.plot(f"pt.{det}"),
					   self.plot(f"ps.{det}")))

		#if style in ["filledcurves", "boxes", \
		#		"boxerrorbars", "boxxyerrorbars"]:
		#	cmd.append("fs %s " % (self.plot(f"fs.{det}","1")))
		return " ".join(cmd)

	# ----------------------------------------------------------------------
	# Plot commands
	#
	# execute plot commands in two passes
	#   before = True - execute all commands before a "[s]plot"
	#                   target is found
	#   before = False - execute all commands after a possible "[s]plot"
	#
	# triggers:
	# - plot | splot  : substitute with the program plot command
	# - replot        : substitute with replot + cmd
	# - else          : anything else is send to gnuplot as is
	# ----------------------------------------------------------------------
	def _plotCommands(self, plot, cmd):
		pline = ""		# previous line
		plotuser = False	# a user plot command found "[s]plot cmd"
		plotcmd  = False	# a system plot command was issued
		for line in self.plot("commands").splitlines():
			line = line.strip()
			# check possible line continuation with "\"
			if pline != "":
				line = pline + line
				pline = ""
			if line!="" and line[-1] == "\\":
				pline = line[:-1]
				continue
			if not line: continue

			# check for a line like "plot" or "splot" only
			if line in ("plot", "splot"):
				self(f"{plot} {cmd}\n")
				plotcmd = True
			elif line == "replot":
				self(f"replot {cmd}\n")
				plotcmd = True
			else:
				if line.startswith("plot ") or line.startswith("splot "):
					plotuser = True
				elif line.startswith("replot ") and not plotcmd and not plotuser:
					self(f"{plot} {cmd}\n")
					plotcmd = True
				self(line+"\n")		# execute command

		if not plotcmd and not plotuser:
			self(f"{plot} {cmd}\n")

	# ----------------------------------------------------------------------
	# Perform plot
	# ----------------------------------------------------------------------
	def _plot(self, cmd):
		if self.engine.version() == 0: return None
		background = self._backgroundPlot()
		if background is not None: cmd = f"{background}, {cmd}"
		self._plotCommands("plot", cmd)
		return self.checkOutput()

	# ----------------------------------------------------------------------
	# Perform splot
	# ----------------------------------------------------------------------
	def _splot(self, cmd):
		if self.engine.version() == 0: return None
		self._plotCommands("splot", cmd)
		return self.checkOutput()

	# ----------------------------------------------------------------------
	def checkOutput(self):
		self.engine.writeline('print "end"')
		# wait until we get the line end
		line  = "\n"
		END   = "end\n"
		err   = None
		tstart = time.time()
		while line!=END and time.time()-tstart < 10.0:	# 10s
			while self.engine.check(1.0):
				line = self.engine.readline()
				if line not in ("\n", END):
					if "Warning" in line:
						self.log(f"{log.WARNING_PREFIX} {line}")
					else:
						self.log(f"{log.ERROR_PREFIX} {line}")
					if err is None:
						err = "Gnuplot errors or warnings found"
					if self.engine.poll() is not None:
						return "Fatal error Gnuplot terminated with "\
							f"exit code={self.engine.poll()}"
		return err

	# ----------------------------------------------------------------------
	# Enable pm3d mode for gnuplot
	# ----------------------------------------------------------------------
	def plotPm3D(self, ncolors=None, palette="fluka", pm3dopts=None):
		self("set colorbox vertical")
		if pm3dopts is None:
			self("set pm3d map explicit corners2color c1")
		else:
			self(f"set pm3d {pm3dopts}")

		#if not isinstance(palette, str):
		#	palette.gnuplot_setPalette(self.engine)
		#	return
		#else:
		try:
			self(paletteStore[palette].gnuplot())
		except KeyError:
			self(f"set palette {palette}")

		if ncolors:
			# FIXME Stupid bug fix of gnuplot
			# doesn't work when printing!!!!
			if self.engine.version() < 4.2 or ncolors<=1: ncolors += 1
			self(f"set palette maxcolors {ncolors}")

	# ----------------------------------------------------------------------
	# Set normalization factor operated on a column and/or expression
	#	norm=""           no normalization is assumed return column
	#	norm=(norm)       if it contains $column terms
	#	norm=expr->float  if it evaluates to float return (norm)*(column)
	#	norm=expr-of-$xxx expression with columns $xxx return norm as is
	#	norm=func(x)      set a gnuplot function func(x)=norm and use the
	#                         function(column)
	# ----------------------------------------------------------------------
	def plotNorm(self, norm, column, mult="", axis="y"):
		# Normalization and default using
		if norm=="": return str(column)
		if isinstance(column,int):
			column = f"(${column})"
		if "$" in norm:
			return f"({norm})"

		try:
			# Try for a fixed value
			norm = float(eval(norm, Evaluation._globalDict))
			return f"({norm:g}*{column})"
		except:
			pass

		if str(norm).find("$")>=0:
			# Expression with $col inside
			if mult:
				return f"({norm}*{mult})"
			return f"({norm})"

		func=f"{axis}norm"
		if re.search(r"\bx\b", norm):
			self(f"{func}(x)={norm}")
		else:
			self(f"{func}(x)=({norm})*x")
		if mult:
			return f"({func}({column})*{mult})"
		else:
			return f"({func}({column}))"

	# ----------------------------------------------------------------------
	# Generate the plot
	# ----------------------------------------------------------------------
	def _geometryPlot(self):
		err = super()._geometryPlot()
		if err is not None: return err
		# Important variables
		kind = self.plot("kind")
		cd = self.plot("coord")
		O  = bmath.Vector(map(float,self.plot("origin").split()))
		U  = bmath.Vector(map(float,self.plot("basisu").split()))
		V  = bmath.Vector(map(float,self.plot("basisv").split()))
		extU,extV = map(float, self.plot("extends").split())
		loP  = -extU*U -extV*V + O
		hiP  =  extU*U +extV*V + O
		Us = Vs = ""		# Sign of field operation
		if cd == Plot.COORD[0]:
			#center = "0,0"
			using  = "us 1:2"
			self("set xrange [%g:%g]"%(-extU, extU))
			self("set yrange [%g:%g]"%(-extV, extV))
		elif cd == Plot.COORD[1]:
			#center   = "0,0"
			using  = "us 2:1"
			self("set xrange [%g:%g]"%(-extV, extV))
			self("set yrange [%g:%g]"%(-extU, extU))
		else:
			u = ord(cd[ 0])-ord('X')
			v = ord(cd[-1])-ord('X')
			#center = "%g,%g"%(O[u],O[v])
			self("set xrange [%g:%g]"%(loP[u],hiP[u]))
			self("set yrange [%g:%g]"%(loP[v],hiP[v]))
			u += 3
			v += 3
			using  = f"us {u}:{v}"
			if U.direction()[0]=="-": Us = "-"
			if V.direction()[0]=="-": Vs = "-"

		# Styles
		self("set style line 1 lt -1 lw 1")
		self("set style line 2 lt 6 lw 1")

		# Vector field scaling
		try:
			vscale = float(self.plot('vectscale'))
		except ValueError:
			vscale = 1
		self(f"vscale={abs(vscale)}")

		# Do the plot
		if kind == Plot.KIND[0]:
			return self._plot(f"'{self.plot.name}.dat' ind 0 {using} w l ls 1 notitle")

		elif kind == Plot.KIND[-2]:		# Field Vector
			every = 5
			return self._plot(f"'{self.plot.name}.dat' ind 0 {using} w l ls 1 notitle,"
				f"'' ind 1 every {every}:{every} "
				f"{using}:({Us}$9*vscale):({Vs}$10*vscale) w vectors ls 2 notitle")

		else:
			# FIXME if one day I convert to with image (not working until 4.2.3)
			#if self.engine.version() > 4.2:
			if self.engine.version() >= 5.1:
				plot      = self._plot
				withImage = "with image"
				with0     = ""
			else:
				plot      = self._splot
				withImage = ""
				with0     = ":(0)"

			extra = ""
			colors = 32
			if kind == Plot.KIND[2]:		# Materials
				try:
					colors = self.project.materialPalette(self.plot._materialList)
					palette = ",".join([f"{x[0]} \"{x[1]}\"" for x in colors])
					self.plotPm3D(len(colors), f"defined ({palette})")
					# FIXME correct bug that might set cblog by mistake to 1!
					if int(self.plot("cblog")):
						self("unset logscale cb")
					if isinstance(self.plot._materialList, list):
						self(f"set cbrange [1:{len(colors)}]")
					else:
						cbmin = min(self.plot._materialList.keys())
						cbmax = max(self.plot._materialList.keys())
						if cbmin==cbmax:
							cbmin -= 0.1
							cbmax += 0.1
						self(f"set cbrange [{cbmin}:{cbmax}]")
				except AttributeError:
					# Due to some error the _materialList is not defined
					self.log(f"{log.ERROR} materialList is not defined")

			elif kind == Plot.KIND[-1]: # Field Intensity + Vectors
				every = 5
				if vscale>0:
					# uniform length
					us_vect = f"({Us}vscale*$9/sqrt($9**2+$10**2)):"\
						  f"({Vs}vscale*$10/sqrt($9**2+$10**2))"
				else:
					# proportional length to intensity
					us_vect = f"({Us}$9*vscale):({Vs}$10*vscale)"

				extra=f",'' ind 1 every {every}:{every} " \
					f"{using}{with0}:{us_vect}{with0} " \
					f"w vectors ls 2 notitle"
				self.plotPm3D(colors, "field")

			elif kind in (Plot.KIND[-3], Plot.KIND[-1]):
				self.plotPm3D(colors, "field")
			else:
				self.plotPm3D(colors, "geometry")

			if kind in (Plot.KIND[-3], Plot.KIND[-1]):
				col=11
				u3 ="11"
			else:
				col = Plot.KIND.index(kind) + 5
				if kind == Plot.KIND[2]:
					u3 = str(col)
				else:
					# Use random colors for region/lattice plots
					self("rnd(x)=abs(int((x*1103515245+12345)/65536)%32768)")
					u3 ="(rnd($%d)%%%d)"%(col,colors)
				self("unset colorbox")

			if int(self.plot("labels")) and col<8:
				# Check version of engine
				if self.engine.version() < 4.2:
					# Convert to notify
					messagebox.showwarning(
						"No Material labels",
						"Gnuplot version 4.2 or higher is required "
						"for plotting material labels. "
						"Option will be disabled.")
					self.plot["labels"] = 0
				else:
					font = self.plot("labelfont")
					font = f"font '{font}' " if font is not None else ""
					extra = f",'' ind 2 {using}{with0}:{col} w labels " \
						"offset 0,.3 point ps 0.7 pt 11 " \
						f"tc lt -1 {font}notitle"

			# FIXME if one day I convert to with image (not working until 4.2.3)
			if self.engine.version() >= 5.0:
				return plot(f"'{self.plot.name}.dat' ind 1 {using}:{u3} "
					f"{withImage} notitle,"
					f"'' ind 0 {using}{with0} w l ls 1 notitle{extra}")

	# ----------------------------------------------------------------------
	# @return the plot string for the background plot
	# ----------------------------------------------------------------------
	def _backgroundPlot(self):
		if self.plot("imgtics"):
			filename = self.plot('imgfile').strip()
			if filename:
				_,ext = os.path.splitext(filename)
				ext = ext[1:].lower()
				return f"'{filename}' binary filetype={ext} with rgbimage axis x2y2"
		return None

	# ----------------------------------------------------------------------
	# Generate the plot
	# ----------------------------------------------------------------------
	def _usr1DPlot(self):
		# --- Loop over detectors to build plot string ---
		cmd = ""
		prevfile = ""
		for n in range(int(self.plot("ndetectors"))):
			if int(self.plot(f"show.{n}")) != 1:
				continue		# Skip graph
			filename = self.plot(f"file.{n}")
			if not filename: continue
			filefmt = Data.detectFormat(filename)
			det = int(self.plot(f"det.{n}"))

			if filefmt == 'phits':
				# convert to gnuplot
				fileout = f"{filename}.data"
				if not tool.isnewer(fileout, filename):
					data = Data.PhitsData(filename, readdata=True)
					with open(fileout, "w") as fout:
						for i,d in enumerate(data.detector):
							if i>0:
								fout.write("\n\n")
							fout.write(f"# {d.num} {d.name}\n")
							for row in d.data:
								print(" ".join([str(x) for x in row]),
									file=fout)
					filename = fileout

			# Check if file is same as previous
			if prevfile == filename:
				filename = ""
			else:
				prevfile = filename

			if cmd!="": cmd += ", \\\n\t"

			# Detector index
			det = f"ind {det}"

			# block to display
			block = int(self.plot(f"block.{n}"))
			if block > 0:
				block -= 1
				det += f" every :::{block}::{block}"

			cmd += f"'{filename}' {det} "

			xlog  = int(self.plot("xlog"))
			style = self.plot(f"with.{n}")
			addErrors = style=="histerror"	# Additional errors

			# Default using (middle of bin)
			if xlog:
				usingx = usingxe = "(sqrt($1*$2))"
			else:
				usingx = usingxe = "(($1+$2)/2.0)"
			usingex = "(($2-$1)/2.0)"

			if style in ("histogram", "histerror"):
				style  = "steps"
				usingx  = 1
				usingex = ""

			yval = int(self.plot(f"y.{n}"))
			errory  = "($4/100.)"
			if yval == 1:	# <X>*Y
				if xlog:
					usingy  = "($3*sqrt($1*$2))"
					usingey = f"({usingy}*{errory})"
				else:
					usingy  = "($3*($1+$2)/2)"
					usingey = f"({usingy}*{errory})"
			elif yval == 2:	# DX*Y
				usingy  = "($3*($2-$1))"
				usingey = f"({usingy}*{errory})"

			else:	# Leave default Y
				usingy  = 3
				usingey = f"($3*{errory})"

			# Normalization of x-axis
			xnorm = self.plot(f"xnorm.{n}")
			if xnorm:
				usingx = self.plotNorm(xnorm, usingx,  axis=f"x{n}")
				if usingex:
					usingex = self.plotNorm(xnorm, usingex, axis=f"x{n}")
				if addErrors:
					usingxe = self.plotNorm(xnorm, usingxe, axis=f"x{n}")

			# Normalization and default using
			norm = str(self.plot(f"norm.{n}"))
			if norm:
				if ":" in norm:
					norm,enorm = norm.split(":")
				else:
					enorm = norm
				usingy  = self.plotNorm(norm, usingy,  axis=f"y{n}")
				usingey = self.plotNorm(enorm, usingey, errory, axis=f"y{n}")

			cmd += "us %s "% \
				(self._plotUsing(style, usingx,
					usingy, usingex, usingey))

			# Plotting style
			cmd += self._plotStyle(n, style)

			# Label
			title = str(self.plot(f"name.{n}"))
			#if title.startswith("#Detector "):
			#	title = self.plot(f"detname.{n}")
			key = int(self.plot(f"key.{n}"))
			if title.strip()=="" or key==0:
				cmd += " notitle"
			else:
				cmd += f" t '{title}'"

			if addErrors:
				cmd += ", \\\n\t'' %s us %s " % \
					(det, self._plotUsing("errorbars", usingxe,
						usingy, "", usingey))
				cmd += self._plotStyle(n, "errorbars")
				cmd += " notitle"

		return self._plot(cmd)

	# ----------------------------------------------------------------------
	# USR-2D plot
	# ----------------------------------------------------------------------
	def _usr2DPlot(self):
		# Normalization and default using
		using = "us 1:5:%s"%(self.plotNorm(self.plot("norm"), 3))

		# Color band
		self("set hidden3d")
		self.colorBand("at s")

		return self._splot("'%s' ind %s %s with lines lt -1 notitle"
			% (self.plot("datafile"), self.plot("det"), using))

	# ----------------------------------------------------------------------
	# Usrbin plot
	# ----------------------------------------------------------------------
	def _usrbinPlot(self):
		err = super()._usrbinPlot()
		if err is not None or self._det is None:
			return err

		regionPlot = self._det.type in (2, 8, 12, 18)
		b10  = self._det.type % 10
		# Unfortunately with image doesn't require
		# the last additional row/col that pm3d needs
		# so it duplicates the last row/col
		withimage = (self.engine.version() >= 5.0)
		withimage = False
		dx = self._det.dx
		dy = self._det.dy
		dz = self._det.dz
		proj = self.plot("proj")
		swap = int(self.plot("swap"))
		norm = self.plot("norm")
		font = ""
		try:
			h = Plot.USRBIN_HIST.index(self.plot("hist"))
		except ValueError:
			h = 0

		# Styles
		self("set style line 1 lt -1 lw 1")

		# --- 2D Histogram ---
		if h == 0:
			# Geometry information
			geo  = self.plot("geo")
			axes = self.plot("axes")
			if geo == Plot.GEO_NO or err:
				plotgeo = ""
			else:
				if geo == Plot.GEO_AUTO:
					name = f"{self.plot.name}.geo"
				else:
					name = geo

				if axes == Plot.COORDAUTO[0]:
					if proj == "X":
						u1, u2 = 5, 4
					elif proj == "Y":
						u1, u2 = 5, 3
					else:
						u1, u2 = 3, 4
					if swap:
						u1, u2 = u2, u1

					if regionPlot:
						# Find plot information
						for p in self.project.plots():
							if p.name == geo:
								axes = p("coord")
								u1 = ord(axes[ 0])-ord('U')+1
								u2 = ord(axes[-1])-ord('U')+1
								if u1>3: u1 -= 1
								if u2>3: u2 -= 1
								#font = p("labelfont")
								#if font is not None:
								#	font = f"font '{font}' "
								break
				else:
					u1 = ord(axes[ 0])-ord('U')+1
					u2 = ord(axes[-1])-ord('U')+1
					if u1>3: u1 -= 1
					if u2>3: u2 -= 1
				plotgeo = f",'{name}.dat' ind 0 us {u1}:{u2}:(0.) w l ls 1 notit"

			# Plot region or normal usrbin plot
			extra = ""
			if regionPlot:
				if geo==Plot.GEO_NO or geo==Plot.GEO_AUTO or err:
					# Convert to notify
					return "Plot of Region or Special binning has been requested,\n" \
						"however no geometry to plot-onto has been selected\n" \
						"-Auto- or -No- plot cannot work with this type of binning.\n" \
						"Please create first a geometry plot"
				ind   = "ind 1"
				using = f"{u1}:{u2}:"

				# label plot
				extra = f",'' ind 2 us {using}(0):6 w labels " \
					"offset 0,.3 point ps 0.7 pt 11 " \
					f"tc lt -1 {font} notitle"
				# Color band
				if int(self.plot("errors")) == 0:
					using += self.plotNorm(norm, 7)
				else:
					using += self.plotNorm(norm, 8)

			else:
				ind   = ""
				if b10 in (1,7) and proj=="Z":
					if swap:
						using = "($2*sin($1)):($2*cos($1)):"
					else:
						using = "($2*cos($1)):($2*sin($1)):"
				#elif b10==2:
				#	pass
				#elif b10==8:
				#	pass
				else:
					if proj=="X":
						du = dz/2
						dv = dx/2
					elif proj=="Y":
						du = dz/2
						dv = dx/2
					else:
						du = dx/2
						dv = dy/2
					if (swap and proj!="Z") or (not swap and proj=="Z"):
						if withimage:
							using = f"($2+{dv}):($1+{du}):"
						else:
							using = "2:1:"
					else:
						if withimage:
							using = f"($1+{du}):($2+{dv}):"
						else:
							using = "1:2:"
				# Color band
				if int(self.plot("errors")) == 0:
					using += self.plotNorm(norm, 3)
				else:
					using += self.plotNorm(norm, 4)

			# -- Common plotting --
			self.colorBand()
			if withimage:
				err2 = self._plot(f"'{self.plot.name}.dat' {ind} "
						f"us {using} with image notitle {plotgeo}{extra}")
			else:
				err2 = self._splot(f"'{self.plot.name}.dat' {ind} "
						f"us {using} notitle {plotgeo}{extra}")
			return err2 or err

		# --- 1D Histogram ---
		elif h == 1:
			style = self.plot("with.0")
			if regionPlot:
				#self("set xtics rotate by 90")
				x  = 0
				ex = "(0)"
				y  = self.plotNorm(norm, 4)
				ey = self.plotNorm(norm, "($4*$5/100)")
				if style in ("histogram", "histerror"):
					self("set style fill solid 1.0")
					self("set boxwidth 0.75 absolute")
					if style == "histerror":
						extra = f",'' us 0:{y}:{ey} with errorbars " \
							f"lt 1 lw 1 notitle"
					else:
						extra = ""
					self._plot(f"'{self.plot.name}.dat' ind 0 "
							f"us 0:{y}:xticlabels(3) with boxes notitle{extra}")
				else:
					self._plot(f"'{self.plot.name}.dat' ind 0 "
							f"us {self._plotUsing(style,x,y,ex,ey)}:xticlabels(3) "
							f"{self._plotStyle(0)} notitle")
			else:
				if style in ("histogram", "histerror"):
					x  = 1
				else:
					x  = "(($1+$2)/2.0)"
				ex = "(0)"
				y  = self.plotNorm(norm, 3)
				ey = self.plotNorm(norm, "($3*$4/100)")
				if style=="histerror":
					return self._plot("'%s.dat' us %s %s notitle,'' us %s %s notitle"
						% (self.plot.name,
						   self._plotUsing(style,x,y,ex,ey),
						   self._plotStyle(0),
						   self._plotUsing("errorbars","(($1+$2)/2.0)",y,ex,ey),
						   self._plotStyle(0,"errorbars")))
				else:
					return self._plot("'%s.dat' us %s %s notitle"
						% (self.plot.name,
						   self._plotUsing(style,x,y,ex,ey),
						   self._plotStyle(0)))

		# --- 1D Max + ...
		else:
			style = self.plot("with.0")
			if style in ("histogram", "histerror"):
				x  = 1
			else:
				x  = "(($1+$2)/2.0)"
			ex = "(0)"
			ind = max(0,h-3)
			if ind==0:	# Max trace
				y  = self.plotNorm(norm, 3)
				ey = self.plotNorm(norm, "($3*$4/100)")
				if style=="histerror":
					return self._plot("'%s.dat' ind %d us %s %s notitle,'' ind %d us %s %s notitle"
						% (self.plot.name,
						   ind,
						   self._plotUsing(style,x,y,ex,ey),
						   self._plotStyle(0),
						   ind,
						   self._plotUsing("errorbars","(($1+$2)/2.0)",y,ex,ey),
						   self._plotStyle(0,"errorbars")))
				else:
					return self._plot("'%s.dat' ind %d us %s %s notitle"
						% (self.plot.name,
						   ind,
						   self._plotUsing(style,x,y,ex,ey),
						   self._plotStyle(0)))

			else:	# horizontal/vertical trace
				# make plotstyle
				axes = self.plot("axes.0")
				if axes not in ("","x1y1"):
					plotstyle = f"axes {axes} "
				else:
					plotstyle = ""

				plotstyle += "w yerrorbars lt %s" % (str(self.plot("lt.0")))
				if self.engine.version() >= 4.2:
					lc = self.plot("lc.0")
					if lc != "": plotstyle += f"lc rgb '{lc}' "

				lw = "lw %s"%(str(self.plot("lw.0")))

				return self._plot("'%s.dat' ind %d us %s:3:4:5 %s %s notitle"
						% (self.plot.name, ind, x, plotstyle, lw))

	# ----------------------------------------------------------------------
	# RESNUCLEi plot
	# ----------------------------------------------------------------------
	def _resnucleiPlot(self):
		err = super()._resnucleiPlot()
		if err is not None: return err

		plotType = self.plot("plot")

		# Styles
		self("set style line 1 lt -1 lw 1")
		self("set style line 2 lt -1 lw 0.5")

		# --- histogram plotting ---
		norm = self.plot("norm")
		if plotType in (Plot.RESNUCLEI_A, Plot.RESNUCLEI_Z):
			# --- 1D plotting ---
			index = "ind %d" % (plotType == Plot.RESNUCLEI_A)

			x  = "($1-0.5)"
			ex = "(0)"
			y  = self.plotNorm(norm, 3)
			ey = self.plotNorm(norm, "($3*$4/100)")
			style = self.plot("with.0")

			if style == "histerror":
				return self._plot("'%s.dat' %s us %s %s notitle,'' %s us %s %s notitle" %
					(self.plot.name,
					 index,
					 self._plotUsing(style,x,y,ex,ey),
					 self._plotStyle(0),
					 index,
					 self._plotUsing("errorbars","1",y,ex,ey),
					 self._plotStyle(0,"errorbars")))
			else:
				return self._plot("'%s.dat' %s us %s %s notitle" %
					(self.plot.name,
					 index,
					 self._plotUsing(style,x,y,ex,ey),
					 self._plotStyle(0)))

		else:
			# --- 2D plotting ---
			#self("set hidden3d")
			self.colorBand() #"at s")

			# Assume format x/y
			using = "us "
			for p in plotType.split("/"):
				if p=="Z":
					using += "($1-0.5):"
				elif p=="A":
					using += "($2-0.5):"
				elif p=="N":
					using += "($2-$1)>0?($2-$1):1/0:"

			# Normalization and default using
			using += self.plotNorm(norm, 3)

			return self._splot("'%s.dat' ind 2 %s notitle"% (self.plot.name, using))

	# ----------------------------------------------------------------------
	def _userdumpPlot(self):
		err = super()._userdumpPlot()
		if err: return err

		# --- Geometry ---
		geo  = self.plot("geo")
		axes = self.plot("axes")
		if geo in ("",Plot.GEO_NO):
			plotgeo = ""
		else:
			name = geo
			u1 = ord(axes[ 0])-ord('U')+1
			u2 = ord(axes[-1])-ord('U')+1
			if u1>3: u1 -= 1
			if u2>3: u2 -= 1
			plotgeo = f",'{name}.dat' ind 0 us {u1}:{u2} w l lt -1 lw 1 notit"

		case = self.plot("case")
		if case == "Continuous":
			return self._splot(f"'{self.plot.name}.dat' us 4:5:6 w l")
		elif case == "Point":
			pass
		elif case == "Source":
			column = Plot.SOURCE_HIST.index(self.plot("x"))

			# Limits
			xmin  = self.plot("xmin")
			if xmin == "":
				xmin = self.dmin[column]
			else:
				xmin = float(xmin)

			xmax  = self.plot("xmax")
			if xmax == "":
				xmax = self.dmax[column]
			else:
				xmax = float(xmax)

			# --- Histogram or Scatter
			if self.plot("type") == Plot.HISTOGRAM:
				try: bins = float(self.plot("bins"))
				except: bins = 50.0
				if int(self.plot("xlog")):
					if xmin <= 0.0: xmin = 1e-10
					if xmax <= 0.0: xmax = 1.
					step  = (math.log10(xmax) - math.log10(xmin)) / bins
					self("logbin(x,m,s)=10**(floor((log10(x)-m)/s)*s+m)")
					func = "logbin"
				else:
					step  = (xmax - xmin) / bins
					self("bin(x,m,s)=floor((x-m)/s)*s+m")
					func = "bin"

				if int(self.plot("weight")):
					weight = "3"
				else:
					weight = "(1)"

				# Using...
				using = "(%s($%d,%g,%g)):%s" % \
					(func, column+1, self.dmin[column],
					 step, weight)
				self("set style fill solid 0.5 noborder")
				return self._plot(f"'{self.plot.name}.dat' us {using} "
						"smooth freq with boxes lt 7 notitle,"
						f"'' us {using} smooth freq with histeps lt -1 notitle")

			else:	# Scatter plot
				column2 = Plot.SOURCE_HIST.index(self.plot("y"))
				return self._plot(f"'{self.plot.name}.dat' "
						f"us {column+1}:{column2+1} with points {plotgeo}")
