#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   14-Oct-2011

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import math
import tkinter as tk
from tkinter import ttk

import bmath
import icons
import Manual
import tkExtra
import tkFlair
import Project

from quad import Quad
from deck  import Input, CardInfo, Card
from fluka import FlukaCard

_NONE   = "-none-"
_GROUPS = ["all", "plane", "box", "sphere", "cylinder", "quadric"]

# FIXME move variables in another module and import them from Viewer and Body
_VIEW   = ["all", "3D", "Red", "Green", "Blue", "Magenta"]

# FIXME should be defined in tkFlair
SMALL = 1E-10
BIG   = 1E10

_NAMEPAT = re.compile(r"^(.+?)\d+$")

#-------------------------------------------------------------------------------
# Round number to closest integer if needed
#-------------------------------------------------------------------------------
def fmt(x):
	if abs(x)<SMALL:
		return 0.0
	else:
		return bmath.format(x,15)

#-------------------------------------------------------------------------------
# Create a new body aligned to a certain viewport
# @param P	body position
# @param size	grid size as a hint for initial size of body
# @param matrix	view port matrix
#-------------------------------------------------------------------------------
def newBody(tag, P, size, matrix):
	# Find direction if needed
	u = bmath.Vector(matrix[0][0], matrix[1][0], matrix[2][0])
	v = bmath.Vector(matrix[0][1], matrix[1][1], matrix[2][1])
	w = bmath.Vector(matrix[0][2], matrix[1][2], matrix[2][2])

	# Create basic whats
	what = ["body"]

	mode,tag = tag.split(":")

	if   tag in ("XYP","PZ"):
		what.append(P.z)
	elif tag in ("XZP","PY"):
		what.append(P.y)
	elif tag in ("YZP","PX"):
		what.append(P.x)
	elif tag == "PLA":
		what.extend(size*u)
		what.extend(P)
	elif tag == "P":
		what.extend(size*u)
		what.append(P*u)
	elif tag == "P3":
		what.extend(P)
		what.extend(P+u*size)
		what.extend(P+w*size)

	elif tag == "RPP":
		if abs(abs(w[0])-1.0) < 1.0E-5:
			what.append(P.x-size/2.0)
			what.append(P.x+size/2.0)
		else:
			what.append(P.x)
			what.append(P.x+size)

		if abs(abs(w[1])-1.0) < 1.0E-5:
			what.append(P.y-size/2.0)
			what.append(P.y+size/2.0)
		else:
			what.append(P.y)
			what.append(P.y+size)

		if abs(abs(w[2])-1.0) < 1.0E-5:
			what.append(P.z-size/2.0)
			what.append(P.z+size/2.0)
		else:
			what.append(P.z)
			what.append(P.z+size)

	elif tag in ("BOX", "REC"):
		what.extend(P)
		what.extend( size*u)
		what.extend( size*v)
		what.extend(-size*w)

	elif tag in ("WED", "RAW"):
		what.extend(P)
		what.extend( size*u)
		what.extend( size*v)
		what.extend(-size*w)

	elif tag in ("SPH","S"):
		what.extend(P)
		what.append(size/2.0)
	elif tag == "SO":
		what.append(P.length())
	elif tag == "SX":
		what.append(P.x)
		what.append(size/2.0)
	elif tag == "SY":
		what.append(P.y)
		what.append(size/2.0)
	elif tag == "SZ":
		what.append(P.z)
		what.append(size/2.0)

	elif tag == "ELL":
		what.append(P.x - size*u[0])
		what.append(P.y - size*u[1])
		what.append(P.z - size*u[2])
		what.append(P.x + size*u[0])
		what.append(P.y + size*u[1])
		what.append(P.z + size*u[2])
		what.append(3.0*size)

	elif tag in ("RCC", "TRC"):
		what.extend(P)
		what.extend(size*u)
		what.append(size)
		if tag == "TRC":
			what.append(size/2.0)

	elif tag in ("XCC", "XEC", "C/X"):
		what.append(P.y)
		what.append(P.z)
		what.append(size)
		if tag=="XEC": what.append(size*2.0)

	elif tag in ("YCC", "YEC", "C/Y"):
		what.append(P.z)
		what.append(P.x)
		what.append(size)
		if tag=="YEC": what.append(size*2.0)

	elif tag in ("ZCC", "ZEC", "C/Z"):
		what.append(P.x)
		what.append(P.y)
		what.append(size)
		if tag=="ZEC": what.append(size*2.0)

	elif tag == "CX":
		what.append(math.sqrt(P.y**2 + P.z**2))
	elif tag == "CY":
		what.append(math.sqrt(P.x**2 + P.z**2))
	elif tag == "CZ":
		what.append(math.sqrt(P.x**2 + P.y**2))

	elif tag in ("K/X", "K/Y", "K/Z"):
		what.extend(P)
		what.append(1)			# same initial nonsense t2
		what.append(0)			# double sided cone

	elif tag == "KX":
		what.append(P.x)
		what.append(1)			# same initial nonsense t2
		what.append(0)			# double sided cone
	elif tag == "KY":
		what.append(P.y)
		what.append(1)			# same initial nonsense t2
		what.append(0)			# double sided cone
	elif tag == "KZ":
		what.append(P.z)
		what.append(1)			# same initial nonsense t2
		what.append(0)			# double sided cone

	elif tag in ("QUA", "GQ"):
		what.append( 1.0/size**2)	# sphere at P location
		what.append( 1.0/size**2)
		what.append( 1.0/size**2)
		what.append( 0.0)
		what.append( 0.0)
		what.append( 0.0)
		what.append(-2.0*P.x/size**2)
		what.append(-2.0*P.y/size**2)
		what.append(-2.0*P.z/size**2)
		what.append(P.length2()/size**2-1.0)

	elif tag == "SQ":			# sphere at P location
		what.append( 1.0/size**2)	# A
		what.append( 1.0/size**2)	# B
		what.append( 1.0/size**2)	# C
		what.append( 0.0)		# D
		what.append( 0.0)		# E
		what.append( 0.0)		# F
		what.append(-1.0)		# G
		what.extend(P)

	card = Card.new(tag,mode)	# FIXME
	card.setWhats(what)
	return card

# ----------------------------------------------------------------------
# Convert BOX to RPP
# ----------------------------------------------------------------------
#def box2rpp(card):
#	assert card.tag == "BOX"
#	P = card.bodyP()
#	X = card.bodyX()
#	Y = card.bodyY()
#	Z = card.bodyZ()
#	xlen = X.norm()
#	ylen = Y.norm()
#	zlen = Z.norm()
#	dump()
#	dump("position=",P)
#	dump("Dim=",xlen,ylen,zlen)
#	M = bmath.Matrix([X, Y, Z])
#	dump(M)
#	dump(M.T())
#	dump([round(math.degrees(x)) for x in M.getEulerRotation()])
#	dump([round(math.degrees(x)) for x in M.T().getEulerRotation()])

#-------------------------------------------------------------------------------
# return a PLA on the viewing plane, assuming w pointing inside
# @param pts	a list of points (u,v) on the viewing plane
# @return None or (tag, [whats], err)
#-------------------------------------------------------------------------------
def findPlaneUV(pts, matrix=None):
	N = len(pts)

	err = 0.0
	if N==1:
		# with one point we return
		a = b = d = 0.0
		c = 1.0
	else:
		A = bmath.Matrix(pts)
		B = bmath.Matrix([-1.0]*N)

		# default plane
		a = b = c = 0.0; d = 1.0
		try:
			# Check for line equation a*u + b*v + 1 = 0
			# and the plane equation a*u + b*v + 0*w + 1 = 0
			a,b = bmath.solveOverDetermined(A,B)

			if abs(a)>BIG or abs(b)>BIG:      raise Exception("Cannot fit plane UV")
			if abs(a)<SMALL and abs(b)<SMALL: raise Exception("Cannot fit plane UV")
		except:
			# search for a solution a*u + b*v = 0 <=> a*u = -v
			ATA = ATB = 0.0
			for u,v in pts:
				ATA += u*u
				ATB -= u*v
			if ATA>SMALL:
				a = ATB/ATA
				b = 1.0
				d = 0.0
			else:
				# check if u=0 or v=0
				su = sv = 0.0
				for u,v in pts:
					su += abs(u)
					sv += abs(v)

				if abs(su)<SMALL and abs(sv)<SMALL: return None

				d = 0.0
				if abs(su)<SMALL:
					a = 1.0
				elif abs(sv)<SMALL:
					b = 1.0
				else:
					return None

		# find error
		for u,v in pts:
			err += (a*u+b*v+d)**2
		err = math.sqrt(err)/float(N)

	# Rotate if matrix is present
	if matrix is not None:
		q = Quad(Cx=a,Cy=b,Cz=c,C=d)
		q.transform(matrix)
		a = q.Cx
		b = q.Cy
		c = q.Cz
		d = q.C

	# Avoid rounding
	if abs(a) < SMALL: a = 0.0
	if abs(b) < SMALL: b = 0.0
	if abs(c) < SMALL: c = 0.0
	if abs(d) < SMALL: d = 0.0

	return "PLA", (a,b,c,d), err

#-------------------------------------------------------------------------------
# @param pts		a list of points that form the plane
#-------------------------------------------------------------------------------
def findPlane(pts):
	N = len(pts)

	if N < 3: return None

	A = bmath.Matrix(pts)
	B = bmath.Matrix([-1.0]*N)

	# default plane
	a = b = c = 0.0; d = 1.0
	try:	# try generic plane
		a, b, c = bmath.solveOverDetermined(A,B)
		if abs(a)>BIG or abs(b)>BIG or abs(c)>BIG:         raise Exception("Cannot fit plane")
		if abs(a)<SMALL and abs(b)<SMALL and abs(c)<SMALL: raise Exception("Cannot fit plane")
	except:
		# search for a solution a*x + b*y + c*z = 0 <=> a*x + b*y = -z
		A = []
		B = []
		for x,y,z in pts:
			A.append([x,y])
			B.append(-z)

		A = bmath.Matrix(A)
		B = bmath.Matrix(B)

		d = 0.0
		try:
			a, b = bmath.solveOverDetermined(A,B)
			if abs(a)>BIG or abs(b)>BIG:      raise Exception("Cannot fit plane")
			if abs(a)<SMALL and abs(b)<SMALL: raise Exception("Cannot fit plane")
			c = 1.0
		except:
			# Check for a solution of x=0 or y=0 or z=0
			sx = sy = sz = 0.0
			for x,y,z in pts:
				sx += abs(x)
				sy += abs(y)
				sz += abs(z)

			if abs(sx)<SMALL and abs(sy)<SMALL and abs(sz)<SMALL:
				return None

			# check if all points fall on a line
			#X = [p[0] for p in pts]
			#Y = [p[1] for p in pts]
			#Z = [p[2] for p in pts]
			#r1 = bmath.linear(X,Y)
			#r2 = bmath.linear(X,Z)
			#if r1 is not None and r2 is not None and \
			#   abs(r1[2])>0.999999 and abs(r2[2])>0.999999: return None
			#say("r1=",r1)
			#say("r2=",r2)

			if abs(sx)<SMALL:
				a = 1.0
			elif abs(sy)<SMALL:
				b = 1.0
			elif abs(sz)<SMALL:
				c = 1.0
			else:
				# Check for 1) a*x + b*y = 0 <=> a*x = -y
				#        or 2) a*x + c*z = 0 <=> a*x = -z
				#        or 3) b*y + c*z = 0 <=> b*y = -z
				# search for a solution a*u + b*v = 0 <=> a*u = -v
				ATA1 = ATB1 = 0.0
				#ATA2 = ATB2 = 0.0
				ATB2 = 0.0
				ATA3 = ATB3 = 0.0

				for x,y,z in pts:
					ATA1 += x*x
					ATB1 -= x*y

					ATB2 -= x*z

					ATA3 += y*y
					ATB3 -= y*z
				#ATA2 = ATA1

				if ATA1 > SMALL:
					a = ATB1/ATA1
					b = 1.0
				elif ATA3 > SMALL:
					b = ATB3/ATA3
					c = 1.0
				else:
					return None

			# FIXME Check if all points fall on a line

	# find error
	err = 0.0
	for x,y,z in pts:
		err += (a*x+b*y+c*z+d)**2
	err = math.sqrt(err)/float(N)

	return "PLA", (a,b,c,d), err

#-------------------------------------------------------------------------------
# Convert a generic plane to a more appropriate form
# @param tagwhat	("PLA", [a,b,c,d], err)
# @return a ("PLA", "XYP", "XZP", "YZP", [x|y|z|(nx,ny,nz,px,py,pz)], err)
#-------------------------------------------------------------------------------
def findBestPlane(tagwhat):
	if tagwhat is None: return None
	tag,what,err = tagwhat
	a,b,c,d = what

	# prefer the positive solution
	if a<=SMALL and b<=SMALL and c<=SMALL:
		a = -a
		b = -b
		c = -c
		d = -d

	if abs(a)<SMALL and abs(b)<SMALL:
		w = -d/c
		return "XYP", [fmt(w)], err
	elif abs(a)<SMALL and abs(c)<SMALL:
		w = -d/b
		return "XZP", [fmt(w)], err
	elif abs(b)<SMALL and abs(c)<SMALL:
		w = -d/a
		return "YZP", [fmt(w)], err

	# searh an intersection of the plane along  various intersection
	# for the most simpler
	solutions = []
	try: solutions.append((-d/(a+b+c), -d/(a+b+c), -d/(a+b+c)))
	except: pass
	try: solutions.append((-d/a, 0.0, 0.0))
	except: pass
	try: solutions.append((0.0, -d/b, 0.0))
	except: pass
	try: solutions.append((0.0, 0.0, -d/c))
	except: pass

	m = -1
	l = 10000
	for i,(x,y,z) in enumerate(solutions):
		s = "%s %s %s"%(bmath.format(x,22), bmath.format(y,22), bmath.format(z,22))
		if len(s)<l:
			m = i
			l = len(s)

	# normalize normal
	s = math.sqrt(a*a + b*b + c*c)
	a /= s
	b /= s
	c /= s

	return "PLA", map(fmt,[a, b, c, solutions[m][0], solutions[m][1], solutions[m][2]]), err

#-------------------------------------------------------------------------------
# Return a SPHere UV
# @param pts	a list of points (u,v) on the viewing plane
# @return None or (tag, [whats], err)
#-------------------------------------------------------------------------------
def findSphereUV(pts, matrix=None):
	N = len(pts)
	if N < 3: return None

	A = []
	B = []
	for x,y in pts:
		A.append([2.0*x, 2.0*y, -1.0])
		B.append([x*x+y*y])
	A = bmath.Matrix(A)
	B = bmath.Matrix(B)

	try:
		x,y,C = bmath.solveOverDetermined(A,B)
	except:
		return None

	R2 = C + x*x + y*y
	if R2 < SMALL: return None

	# find error
	#err = 0.0
	#for x,y,z in pts:
	#	err += ()**2
	#err = math.sqrt(err)/float(N)

	# Rotate if matrix is present
	p = bmath.Vector(x,y,0.0)
	if matrix is not None:
		p = matrix * p

	return "SPH", map(fmt,[ p[0], p[1], p[2], math.sqrt(R2)]), 0.0

#-------------------------------------------------------------------------------
# Return a SPHere
#-------------------------------------------------------------------------------
def findSphere(pts):
	N = len(pts)
	if N < 4: return None

	A = []
	B = []
	for x,y,z in pts:
		A.append([2.0*x, 2.0*y, 2.0*z, -1.0])
		B.append([x*x+y*y+z*z])
	A = bmath.Matrix(A)
	B = bmath.Matrix(B)

	try:
		x,y,z,C = bmath.solveOverDetermined(A,B)
	except:
		return None

	R2 = C + x*x + y*y + z*z
	if R2 < SMALL: return None

	return "SPH", map(fmt, [x, y, z, math.sqrt(R2)]), 0.0

#-------------------------------------------------------------------------------
# Return a Cylinder [XYZ]CC UV
# @param pts	a list of points (u,v) on the viewing plane
# @return None or (tag, [whats], err)
#-------------------------------------------------------------------------------
def findCylinderUV(pts, matrix=None):
	N = len(pts)
	if N < 3: return None

	A = []
	B = []
	for x,y in pts:
		A.append([2.0*x, 2.0*y, -1.0])
		B.append([x*x+y*y])
	A = bmath.Matrix(A)
	B = bmath.Matrix(B)

	try:
		x,y,C = bmath.solveOverDetermined(A,B)
	except:
		return None

	R2 = C + x*x + y*y
	if R2 < SMALL: return None

	# find error
	#err = 0.0
	#for x,y,z in pts:
	#	err += ()**2
	#err = math.sqrt(err)/float(N)

	# Rotate if matrix is present
	p = bmath.Vector(x,y,0.0)
	z = bmath.Vector(0.0,0.0,1.0)
	if matrix is not None:
		p = matrix * p
		z = matrix * z

	R = math.sqrt(R2)

	if   abs(z[0])<=SMALL and abs(z[1])<=SMALL and abs(abs(z[2])-1.0)<=SMALL:
		return "ZCC", map(fmt, [p[0], p[1], R]), 0.0
	elif abs(z[0])<=SMALL and abs(abs(z[1])-1.0)<=SMALL and abs(z[2])<=SMALL:
		return "YCC", map(fmt, [p[2], p[0], R]), 0.0
	elif abs(abs(z[0])-1.0)<=SMALL and abs(z[1])<=SMALL and abs(z[2])<=SMALL:
		return "XCC", map(fmt, [p[1], p[2], R]), 0.0
	else:
		q = Quad(1.0, 1.0, 0.0, 0.0, 0.0, 0.0, -2.0*x, -2.0*y, 0.0, -C)
		q.transform(matrix)
		return "QUA", map(fmt, [q.Cxx, q.Cyy, q.Czz, q.Cxy, q.Cxz, q.Cyz, q.Cx, q.Cy, q.Cz, q.C]), 0.0

#-------------------------------------------------------------------------------
# Return a Cylinder [XYZ]CC
#-------------------------------------------------------------------------------
def findCylinder(pts):
	N = len(pts)
	if N < 3: return None

	AX = []
	AY = []
	AZ = []
	B = bmath.Matrix([1.0]*N)
	for x,y,z in pts:
		AX.append([y*y+z*z, -2.0*y, -2.0*z])
		AY.append([x*x+z*z, -2.0*x, -2.0*z])
		AZ.append([x*x+y*y, -2.0*x, -2.0*y])
	AX = bmath.Matrix(AX)
	AY = bmath.Matrix(AY)
	AZ = bmath.Matrix(AZ)

	try:
		C,y,z = bmath.solveOverDetermined(AX,B)
		y /= C
		z /= C
		R2 = 1.0/C+(y*y+z*z)
		if R2<0.0:
			return "XCC", [-y,-z,math.sqrt(-R2)]
		else:
			return "XCC", [ y, z,math.sqrt( R2)]
	except:
		pass

	try:
		C,x,z = bmath.solveOverDetermined(AY,B)
		x /= C
		z /= C
		R2 = 1.0/C+(x*x+z*z)
		if R2<0.0:
			return "YCC", [-z,-x,math.sqrt(-R2)]
		else:
			return "YCC", [ z, x,math.sqrt( R2)]
	except:
		pass

	try:
		C,x,y = bmath.solveOverDetermined(AZ,B)
		x /= C
		y /= C
		R2 = 1.0/C+(x*x+y*y)
		if R2<0.0:
			return "ZCC", [-x,-y,math.sqrt(-R2)]
		else:
			return "ZCC", [ x, y,math.sqrt( R2)]
	except:
		pass

	return None

#-------------------------------------------------------------------------------
# Return an Elliptical Cylinder [XYZ]EC
#-------------------------------------------------------------------------------
def findEllipticalCylinder(pts):
	N = len(pts)
	if N < 4: return None

	AX = []
	AY = []
	AZ = []
	B = bmath.Matrix([1.0]*N)
	for x,y,z in pts:
		AX.append([y*y, z*z, -2.0*y, -2.0*z])
		AY.append([x*x, z*z, -2.0*x, -2.0*z])
		AZ.append([x*x, y*y, -2.0*x, -2.0*y])
	AX = bmath.Matrix(AX)
	AY = bmath.Matrix(AY)
	AZ = bmath.Matrix(AZ)

	try:
		a,b,y,z = bmath.solveOverDetermined(AX,B)
		y /= a
		z /= b
		Ry = math.sqrt(y*y + b/a * z*z + 1.0/a)
		Rz = math.sqrt(a/b)*Ry
		return "YEC", [y,z,Ry,Rz]
	except:
		pass

	try:
		a,b,x,z = bmath.solveOverDetermined(AY,B)
		x /= a
		z /= b
		Rx = math.sqrt(x*x + b/a * z*z + 1.0/a)
		Rz = math.sqrt(a/b)*Rx
		return "YEC", [x,z,Rx,Rz]
	except:
		pass

	try:
		a,b,x,y = bmath.solveOverDetermined(AZ,B)
		x /= a
		y /= b
		Rx = math.sqrt(x*x + b/a * y*y + 1.0/a)
		Ry = math.sqrt(a/b)*Rx
		return "ZEC", [x,y,Rx,Ry]
	except:
		pass

	return None

#-------------------------------------------------------------------------------
# Return a QUAdratic
#-------------------------------------------------------------------------------
def findQUA(pts):
	N = len(pts)
	if N < 5: return None

	B = bmath.Matrix([1.0]*N)

	A  = []
	AX = []		# ignore x
	AY = []		# ignore y
	AZ = []		# ignore z
	for x,y,z in pts:
		A.append([x*x, y*y, z*z, x*y, x*z, y*z, x, y, z])
		AX.append([y*y, z*z, y*z, y, z])
		AY.append([x*x, z*z, x*z, x, z])
		AZ.append([x*x, y*y, x*y, x, y])
	A = bmath.Matrix(A)

	try:
		X = bmath.solveOverDetermined(A,B)
		return "QUA", X
	except:
		pass

	try:
		X = bmath.solveOverDetermined(AX,B)
		return "QUX", X
	except:
		pass

	try:
		X = bmath.solveOverDetermined(AY,B)
		return "QUY", X
	except:
		pass

	try:
		X = bmath.solveOverDetermined(AZ,B)
		return "QUZ", X
	except:
		pass

	return None

#===============================================================================
# Geometry Auto Body
# Detect all possible body types and propose to the user
#===============================================================================
class GeometryAutoBody(tk.Toplevel):
	activepage = None
	width  = -1
	height = -1
	_PAGES = None

	def __init__(self, master, editor, project):
		super().__init__(master)
		self.transient(master)
		self.editor  = editor
		self.project = project
		self.title("Geometry Auto Body Creation")
		self.layer = None

		# --- Layer ---
		row = 0
		tk.Label(self, text="Name:").grid(row=row, column=0, sticky=tk.E)
		self.name = tk.Entry(self, background="White", width=9)
		self.name.insert(0, self.editor.findUniqueName(FlukaCard.BODY,"body"))
		self.name.grid(row=row, column=1,sticky=tk.EW)
		tkExtra.Balloon.set(self.name,
			"Proposed name for the newly created body")
		b = tk.Button(self, text="A", command=self.autoname,
				padx=2,pady=1)
		b.grid(row=row, column=2,sticky=tk.EW)
		tkExtra.Balloon.set(b, "Assign automatically a name to the body")

		# ---
		tk.Label(self, text="# Points:").grid(row=row, column=3, sticky=tk.E)
		self.npoints = tk.Label(self, text=_NONE, background="White", anchor=tk.W)
		self.npoints.grid(row=row, column=4, sticky=tk.EW)
		tkExtra.Balloon.set(self.npoints, "Number of selected points")

		# ---
		row += 1
		tk.Label(self, text="Comment:").grid(row=row, column=0, sticky=tk.NE)
		self.comment = tk.Text(self, background="White", width=30, height=3)
		self.comment.grid(row=row, column=1,columnspan=4,sticky=tk.NSEW)
		tkExtra.Balloon.set(self.comment,
			"Comment describing the new body")

		# ---
		row += 1
		tk.Label(self, text="View:").grid(row=row, column=0, sticky=tk.E)
		self.view = tkExtra.Combobox(self, width=10)
		self.view.grid(row=row, column=1, columnspan=2, sticky=tk.EW)
		self.view.fill(_VIEW)
		self.view.set(_VIEW[0])
		# Hack
		for i in range(2,len(_VIEW)):
			self.view._listbox.itemconfig(i,
				foreground=f"Dark{_VIEW[i]}")
		tkExtra.Balloon.set(self.view,
			"Show bodies that are parallel or perpendicular to a viewport")

		tk.Label(self, text="Group:").grid(row=row, column=3, sticky=tk.E)
		self.group = tkExtra.Combobox(self, width=10)
		self.group.grid(row=row, column=4, sticky=tk.EW)
		self.group.fill(_GROUPS)
		self.group.set(_GROUPS[0])
		tkExtra.Balloon.set(self.group,
			"Show only bodies belonging to a certain group")

		# ---
		row += 1
		b = tk.Button(self, text="Bodies:",
				command = self.forceRefresh,
				image=icons.get("refresh"),
				compound=tk.LEFT, anchor=tk.W)
		b.grid(row=row, column=0, columnspan=5, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Click to refresh")
		row += 1
		self.text = tk.Text(self, background="White", takefocus=1, width=40)
		self.text.grid(row=row, column=0, columnspan=5, sticky=tk.NSEW)
		sbv = tkExtra.AutoScrollbar(self, takefocus=0)
		sbv.grid(row=row, column=2, sticky='ns')
		sbh = tkExtra.AutoScrollbar(self, takefocus=0,
				orient=tk.HORIZONTAL)
		sbh.grid(row=row+1, column=0, columnspan=2, sticky='ew')

		self.text.config(yscrollcommand=sbv.set, xscrollcommand=sbh.set)
		sbv.config(command=self.text.yview)
		sbh.config(command=self.text.xview)
		self.text["wrap"]  = tk.NONE
		self.text["state"] = tk.DISABLED

		self.grid_rowconfigure(row, weight=1)
		self.grid_columnconfigure(1, weight=1)
		self.grid_columnconfigure(4, weight=1)

		# ---
		row += 1
		frame = tk.Frame(self)
		frame.grid(row=row, column=0, columnspan=5, sticky=tk.EW)

		# buttons
		tk.Button(frame, text='Help', image=icons.get("info"),
			compound=tk.LEFT, takefocus=tk.FALSE,
			command=self.help).pack(side=tk.LEFT, padx=5, pady=5)
		tk.Button(frame, text='Close', image=icons.get("x"),
			compound=tk.LEFT, takefocus=tk.FALSE,
			command=self.close).pack(side=tk.RIGHT, padx=5, pady=5)
		tk.Button(frame, text='Insert', image=icons.get("add"),
			compound=tk.LEFT, takefocus=tk.FALSE,
			command=self.insert).pack(side=tk.RIGHT, padx=5, pady=5)

		self.deiconify()
		if not self.winfo_ismapped():
			self.wait_visibility()
		if GeometryAutoBody.width > 0:
			self.geometry("%dx%d" %
				(GeometryAutoBody.width, GeometryAutoBody.height))
		else:
			self.geometry("%dx%d" %
				(self.winfo_width(), self.winfo_height()))

		self.protocol("WM_DELETE_WINDOW",self.close)
		self.bind("<Escape>",		self.close)
		self.bind("<Control-Key-q>",	self.close)
		self.bind("<Control-Key-w>",	self.close)
		self.bind("<<Help>>",		self.help)
		self.bind("<FocusIn>",		self.refresh)
		self.bind("<Control-Key-r>",	self.refresh)

		self.text.bind("<Button-1>",	self.textFocus)
		self.text.bind("<Home>",	self.selectHome)
		self.text.bind("<End>",		self.selectEnd)
		self.text.bind("<Prior>",	self.selectPageUp)
		self.text.bind("<Next>",	self.selectPageDown)
		self.text.bind("<Up>",		self.selectUp)
		self.text.bind("<Down>",	self.selectDown)

		# Set commands only after initializing everything
		self.view.command  = self.viewChanged
		self.group.command = self.refresh
		self._bodies = []
		self._select = -1
		self._inrefresh = False
		self._hash   = -1
		self.refresh()

	# --------------------------------------------------------------------
	# Close button pressed
	# --------------------------------------------------------------------
	def close(self, event=None):
		self.update()
		GeometryAutoBody.width  = self.winfo_width()
		GeometryAutoBody.height = self.winfo_height()
		self.destroy()

	# --------------------------------------------------------------------
	# Help me please
	# --------------------------------------------------------------------
	def help(self, event=None):
		Manual.show("F4.11")

	# --------------------------------------------------------------------
	# Insert body to list
	# --------------------------------------------------------------------
	def insert(self, event=None):
		if self._select>=0:
			color, (tag, what, err) = self._bodies[self._select]
			what.insert(0, self.name.get())
			self.editor.addBody(FlukaCard(tag, what))
			self.autoname()

	# --------------------------------------------------------------------
	def autoname(self):
		name = self.name.get()
		match = _NAMEPAT.match(name)
		if match:
			prefix = match.group(1)
		elif name:
			prefix = name
		else:
			prefix = "body"

		self.name.delete(0,tk.END)
		self.name.insert(0, self.editor.findUniqueName(FlukaCard.BODY,prefix))

	# --------------------------------------------------------------------
	# Projection Changed
	# --------------------------------------------------------------------
	def viewChanged(self, event=None):
		# Correct color
		view = self.view.get()
		if _VIEW.index(view) < 2:
			color = "Black"
		else:
			color = f"Dark{view}"
		self.view._text["foreground"] = color
		#self.refresh()

	# --------------------------------------------------------------------
	# @return a tuple with the uv view information
	# --------------------------------------------------------------------
	def _createUV(self, idx, xyz):
		# Convert to U,V (ignore W)
		viewer = self.master.viewers[idx]
		#matrix = bmath.Matrix(viewer.matrix(-1))
		matrix = bmath.Matrix(viewer.matrix())
		uv = [list(viewer.xyz2uv(*r)) for r in xyz]
		return (_VIEW[idx+2], uv, matrix)

	# --------------------------------------------------------------------
	# Refresh body list
	# --------------------------------------------------------------------
	def refresh(self, event=None):
		if self._inrefresh: return
		self._inrefresh = True
		self._refresh()
		self._inrefresh = False

	# --------------------------------------------------------------------
	# Force recalculation of bodies
	# --------------------------------------------------------------------
	def forceRefresh(self, event=None):
		self._hash = -1
		self.refresh()

	# --------------------------------------------------------------------
	def _refresh(self):
		selection = self.editor.getSelection(FlukaCard.OBJECT)
		if not selection:
			self.npoints["text"] = _NONE
			return

		self.npoints["text"] = str(len(selection))
		group = self.group.get()

		#say("*-* Selection:",selection)
		#say("*-* Projection:",self.view.get())
		#say("*-* Group:",self.group.get())

		# List of points
		# Get X,Y,Z coordinates
		#cards = self.project.input.cardsSorted(CardInfo.FLAIR_TAGS)

		xyz = []
		# Calculate hash to see if something has changed
		h = hash(self.view.get()) + hash(self.group.get())
		for card in selection:
			p = tuple(card.bodyP())
			xyz.append(p)
			h += hash(p[0]) + hash(p[1]) + hash(p[2])

		# -------------------------------
		# Check if something has changed
		if self._hash == h: return
		self._hash = h

		view = []	# a list of view to check
				# should be in the format (Name, xyz|uv, matrix|None)
		del self._bodies[:]	# Bodies found and categories
				# either: str: Category
				#	or (view-name, body-information)

		# Projection group index requested 0=all, 1=3d, 2,3,4,5 red,green,blue,magenta
		proj = _VIEW.index(self.view.get())

		# 3d view requested
		if proj in (0,1):
			view.append(("3D",xyz,None))
		# UV view requested:
		if proj==0:
			for i in range(4):
				view.append(self._createUV(i,xyz))
		elif proj>=2:
			view.append(self._createUV(proj-2, xyz))

		# ----------------
		# Check for planes
		# ----------------
		if group in ("all", "plane"):
			header = False
			for name, pts, matrix in view:
				if matrix is None:
					body = findBestPlane(findPlane(pts))
				else:
					body = findBestPlane(findPlaneUV(pts, matrix))

				if body is not None:
					if not header:
						self._bodies.append("Plane")
						header = True
					self._bodies.append((name,body))

		# ----------------
		# Check for sphere
		# ----------------
		if group in ("all", "sphere"):
			header = False
			for name, pts, matrix in view:
				if matrix is None:
					body = findSphere(pts)
				else:
					body = findSphereUV(pts, matrix)

				if body is not None:
					if not header:
						self._bodies.append("Sphere")
						header = True
					self._bodies.append((name,body))

		# ------------------
		# Check for cylinder
		# ------------------
		if group in ("all", "cylinder"):
			header = False
			for name, pts, matrix in view:
				if matrix is None:
					body = None
					#body = findCylinder(pts)
				else:
					body = findCylinderUV(pts, matrix)

				if body is not None:
					if not header:
						self._bodies.append("Cylinder")
						header = True
					self._bodies.append((name,body))

		self.fillBodies()

	# --------------------------------------------------------------------
	# Populate bodies textbox
	# --------------------------------------------------------------------
	def fillBodies(self):
		text = self.text
		text["state"] = tk.NORMAL
		text.delete("0.0",tk.END)
		text.tag_delete("*")

		for idx,body in enumerate(self._bodies):
			if isinstance(body,str):
				text.insert(tk.END, f"{body}\n", "title")
			else:
				name, (card, whats, err) = body
				if name == "3D":
					text.insert(tk.END, name, name)
				else:
					text.insert(tk.END, "2D", name)
				text.insert(tk.END, " ")
				text.image_create(tk.END, image=icons.get(card))
				text.insert(tk.END, " ")

				tag = f"B{idx}"
				text.insert(tk.END, "%s %s" %
					(card, "\t".join([bmath.format(x) for x in whats])),
					tag)
				text.tag_bind(tag, "<Button-1>", lambda e,s=self,i=idx : s.select(i))
				text.insert(tk.END, f"\t[{err}]\n")

		if not self._bodies:
			self.select(-1)
			text.insert(tk.END,"None found")
		else:
			text.tag_configure("title",   foreground="DarkBlue", font="Helvetica 14")
			text.tag_configure("3D",      background="#FFFF70")
			text.tag_configure("Red",     background="#FF7070")
			text.tag_configure("Green",   background="#70FF70")
			text.tag_configure("Blue",    background="#7070FF")
			text.tag_configure("Magenta", background="#FF70FF")
			self.select()

		text["state"] = tk.DISABLED
		text.focus_set()

	# --------------------------------------------------------------------
	# Highlight and change select
	# --------------------------------------------------------------------
	def select(self, idx=None):
		if idx is not None:
			if self._select>=0:
				self.text.tag_configure(f"B{self._select}",
					background="White")
			self._select = idx
		if self._select>=0:
			tag = f"B{self._select}"
			self.text.tag_configure(tag, background="LightBlue")
			#index = self.text.tag_nextrange(tag,"1.0")
			#if index: self.text.see(index[0])
			try: self.text.see(f"{tag}.first")
			except tk.TclError: pass

	# --------------------------------------------------------------------
	def selectHome(self, event=None):
		self.select(0)

	# --------------------------------------------------------------------
	def selectEnd(self, event=None):
		self.select(len(self._bodies)-1)

	# --------------------------------------------------------------------
	def selectUp(self, event=None):
		self.select(max(0,self._select-1))

	# --------------------------------------------------------------------
	def selectDown(self, event=None):
		self.select(min(len(self._bodies)-1,self._select+1))

	# --------------------------------------------------------------------
	selectPageUp   = selectUp
	selectPageDown = selectDown

	# ----------------------------------------------------------------------
	def textFocus(self, event):
		self.text.focus_set()

#===============================================================================
# General geometry transformation dialog
#===============================================================================
class TransformGeometryDialog(tk.Toplevel):
	lastPage =  0
	width    = -1
	height   = -1

	def __init__(self, master, flair, cardlist):
		super().__init__(master)
		self.transient(master)
		self.title("Transform geometry")

		self.flair    = flair
		self.input    = flair.project.input
		self.cardlist = cardlist

		# Tab page set
		self.tabPage = ttk.Notebook(self)
		self.tabPage.pack(expand=tk.YES, fill=tk.BOTH)

		# --------------------------------------------------------------
		# Bodies
		# --------------------------------------------------------------
		self.bodies = TransformBodies(self.tabPage, flair, cardlist)
		self.tabPage.add(self.bodies,
				text     = "Bodies",
				image    = icons.get("TRC"),
				compound = tk.LEFT)

		# --------------------------------------------------------------
		# Repeat
		# --------------------------------------------------------------
		self.repeat = RepeatBodies(self.tabPage, flair, cardlist)
		self.tabPage.add(self.repeat,
				text     = "Repeat",
				image    = icons.get("repeat"),
				compound = tk.LEFT)

		# --------------------------------------------------------------
		# Buttons
		# --------------------------------------------------------------
		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, fill=tk.X)

		tk.Button(frame, text="Close", command=self.close,
				image=icons.get("x"), compound=tk.LEFT).pack(side=tk.RIGHT)
		tk.Button(frame, text="Apply", command=self.apply,
				image=icons.get("ok"), compound=tk.LEFT,
				foreground="DarkRed", background="LightYellow",
				activeforeground="Red",
				activebackground="LightYellow").pack(side=tk.RIGHT)

		# --------------------------------------------------------------
		# Bindings
		# --------------------------------------------------------------
		self.bind("<Escape>",		  self.close)
		self.protocol("WM_DELETE_WINDOW", self.close)

		# --- Show ---
		self.deiconify()
		if TransformGeometryDialog.width > 0:
			self.geometry("%dx%d"%
				(TransformGeometryDialog.width, TransformGeometryDialog.height))
		self.lift()
		self.grab_set()
		self.wait_window()

	# ----------------------------------------------------------------------
	def apply(self, event=None):
		page = self.tabPage.index(tk.CURRENT)
		if page == 0:
			self.bodies.transform()
		elif page == 1:
			self.repeat.transform()
		else:
			pass

	# ----------------------------------------------------------------------
	def close(self, event=None):
		# remember last position and size
		TransformGeometryDialog.width    = self.winfo_width()
		TransformGeometryDialog.height   = self.winfo_height()
		TransformGeometryDialog.lastPage = self.tabPage.index(tk.CURRENT)

		if TransformGeometryDialog.lastPage == 0:
			self.bodies.getTransformations()

		elif TransformGeometryDialog.lastPage == 1:
			self.repeat.get()

		self.destroy()

#===============================================================================
# Transform body dialog
#===============================================================================
_TRANSFORM_TYPES = [ "O", "T", "TX", "TY", "TZ", "RX" ,"RY", "RZ", "MX", "MY", "MZ", "S", "ignore" ]
class TransformBodies(tk.Frame):
	transformList = [("RZ","90.0")]

	def __init__(self, master, flair, cardlist):
		super().__init__(master)
		self.flair    = flair
		self.input    = flair.project.input
		self.cardlist = [x for x in cardlist if x.type()==FlukaCard.BODY]

		# --- ROT-DEFI ----
		frame = tk.Frame(self)
		frame.pack(side=tk.TOP, fill=tk.X)
		l = tk.Label(frame, text="ROT-DEFini:")
		l.pack(side=tk.LEFT)
		self.combo = tkExtra.Combobox(frame,False,width=15)
		self.combo.pack(side=tk.LEFT)
		tkExtra.Balloon.set(self.combo,
			"Import or create ROT-DEFini cards in input")
		self.combo._text.config(background="White")
		b = tk.Button(frame, text="Add to Input", command=self.add2Input,
				pady=0, padx=3)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b,
			"Convert transformation to ROT-DEFI cards in input")
		b = tk.Button(frame, text="Get from Input", command=self.getFromInput,
				pady=0, padx=3)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b,
			"Get transformation from input ROT-DEFI cards")

		# --- Transformations ----
		frame = tk.Frame(self)
		frame.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

		self.listbox = tkExtra.MultiListbox(frame,
				(("Type",   10, None),
				 ("Value",  40, None)),
				 height=5)
		self.listbox.grid(row=0, rowspan=7, column=0, sticky=tk.NSEW)
		for lst in self.listbox.listboxes(): lst.config(takefocus=1)
		self.listbox.sortAssist = None

		b = tk.Button(frame, image=icons.get("add"), takefocus=False,
				command=self.add)
		tkExtra.Balloon.set(b, "Add transformation")
		b.grid(row=0, column=1, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("del"), takefocus=False,
				command=self.delete)
		tkExtra.Balloon.set(b, "Delete transformation")
		b.grid(row=1, column=1, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("clone"), takefocus=False,
				command=self.clone)
		tkExtra.Balloon.set(b, "Clone Transformation")
		b.grid(row=2, column=1, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("up"), takefocus=False,
				command=self.moveUp)
		tkExtra.Balloon.set(b, "Move Up")
		b.grid(row=3, column=1, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("down"), takefocus=False,
				command=self.moveDown)
		tkExtra.Balloon.set(b, "Move Down")
		b.grid(row=4, column=1, sticky=tk.EW)

		b = tk.Button(frame, image=icons.get("toggle"), padx=0, takefocus=False,
				command=self.invert)
		tkExtra.Balloon.set(b, "Invert Transformation")
		b.grid(row=5, column=1, sticky=tk.EW)

		frame.grid_columnconfigure(0, weight=1)
		frame.grid_rowconfigure(6, weight=1)

		# --- Global Variables ----
		frame = tk.Frame(self)
		frame.pack(fill=tk.X)

		tk.Label(frame, text="Zero:").grid(row=0, column=0, sticky=tk.W)
		self.zero = tkExtra.FloatEntry(frame, background="white")
		self.zero.grid(row=0, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.zero,
			"Threshold value to be considered zero or equal in comparison. ")

		tk.Label(frame, text="Infinite:").grid(row=1, column=0, sticky=tk.W)
		self.infinite = tkExtra.FloatEntry(frame, background="white")
		self.infinite.grid(row=1, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.infinite,
			"Length of cylinders when from infinite [XYZ]CC are converted to finite RCC")

		self.useQua = tk.BooleanVar(self)
		self.useQua.set(Input._useQUA)

		b = tk.Checkbutton(frame, text="Use QUA (instead of RCC/REC)",
				variable=self.useQua,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		b.grid(row=1, column=2, columnspan=2, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"If set infinite cylinders will be converted to QUA when rotated otherwise to RCC/REC")

		#frame.grid_columnconfigure(1, weight=1)

		# --- Bindings ---
		self.listbox.bindList('<ButtonRelease-1>',	self.release)
		self.listbox.bindList('<<MoveUp>>',		self.moveUp)
		self.listbox.bindList('<<MoveDown>>',		self.moveDown)
		#self.listbox.bindList('<Double-1>',		self.edit)
		self.listbox.bindList('<Return>',		self.release)
		self.listbox.bindList('<KP_Enter>',		self.release)

		self.listbox.setPopupMenu([
				 ('Add',    0, self.add,    icons.get("add")),
				 ('Delete', 0, self.delete, icons.get("x")),
				 ('Clone',  0, self.clone,  icons.get("clone")),
				 ('Invert', 0, self.invert, icons.get("invert"))] )
		self.listbox.bindLeftRight()
		self.listbox.focus_set()

		self.fill()
		self.zero.delete(0,tk.END)
		self.infinite.delete(0,tk.END)
		self.zero.insert(0, str(Input.zero))
		self.infinite.insert(0, str(Input.infinite))

	# ----------------------------------------------------------------------
	def fill(self):
		# Find all ROT-DEFI cards in input and fill combobox
		self.combo.fill(self.input.rotationList())

		# Fill transformations
		for entry in TransformBodies.transformList:
			self.listbox.insert(tk.END, entry)

	# ----------------------------------------------------------------------
	def release(self, event=None):
		listbox = event.widget
		sel = self.listbox.curselection()
		if len(sel) != 1: return
		sel = int(sel[0])

		try: active = listbox.index(tk.ACTIVE)
		except: return

		if active != sel: return

		lid = self.listbox.listboxIndex(listbox)

		if lid == 0:
			# Edit the transform type
			tkExtra.InPlaceList(listbox,
					values=_TRANSFORM_TYPES,
					height=10)
		else:
			# Edit the value
			tkExtra.InPlaceEdit(listbox)

	# ----------------------------------------------------------------------
	def add(self, event=None):
		self.listbox.insert(tk.END, (_TRANSFORM_TYPES[0], "0.0 0.0 0.0"))
		self.listbox.selection_clear(0,tk.END)
		self.listbox.selection_set(tk.END)
		self.listbox.activate(tk.END)
		self.listbox.see(tk.END)
		self.listbox.focus_set()

	# ----------------------------------------------------------------------
	def delete(self, event=None):
		self.listbox.focus_set()
		lst = list(map(int, self.listbox.curselection()))
		if not lst: return
		lst.reverse()
		act = lst[0]
		for i in lst:
			self.listbox.delete(i)
		if act >= self.listbox.size(): act = self.listbox.size()-1
		if act >= 0:
			self.listbox.activate(act)
			self.listbox.selection_set(act)

	# ----------------------------------------------------------------------
	def clone(self, event=None):
		self.listbox.focus_set()
		sel = list(map(int,self.listbox.curselection()))
		if not sel: return
		oldSize = self.listbox.size()
		for i in sel:
			entry = self.listbox.get(i)
			self.listbox.insert(tk.END, entry)
		self.listbox.selection_clear(0,tk.END)
		self.listbox.selection_set(oldSize,tk.END)
		self.listbox.activate(tk.END)

	# ----------------------------------------------------------------------
	def invert(self, event=None):
		self.getTransformations()
		self.listbox.delete(0, tk.END)
		for type_, value in TransformBodies.transformList:
			value = value.split()
			for i,v in enumerate(value):
				try:
					value[i] = str(-float(v))
				except:
					value[i] = "0.0"

			self.listbox.insert(0, (type_, " ".join(value)))

	# ----------------------------------------------------------------------
	def moveUp(self, event=None):
		self.listbox.moveUp()
		return "break"

	# ----------------------------------------------------------------------
	def moveDown(self, event=None):
		self.listbox.moveDown()
		return "break"

	# ----------------------------------------------------------------------
	def transform(self):
		# Threshold
		try:
			zero = abs(float(self.zero.get()))
			if 1e-16 < zero < 0.01:
				Input.zero = zero
		except ValueError:
			pass

		# Infinite
		try:
			infinite = float(self.infinite.get())
			if 1.0 < infinite < 1E10:
				Input.infinite = infinite
		except ValueError:
			pass

		# QUA or RCC/REC
		Input._useQUA = self.useQua.get()

		# Transformation
		self.getTransformations()
		matrix = bmath.Matrix(4, type=1)

		origin = None

		for trans, value in TransformBodies.transformList:
			if trans == "O":
				value = value.split()
				origin = bmath.Vector()
				for i in range(3):
					try:
						origin[i] = float(value[i])
					except:
						self.flair.notify("WARNING",
							"invalid origin vector",
							tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix.translate(-origin)
			elif trans == "T":
				value = value.split()
				t = bmath.Vector()
				for i in range(3):
					try:
						t[i] = float(value[i])
					except:
						self.flair.notify("WARNING",
							"invalid translate vector",
							tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix.translate(t)
			elif trans == "TX":
				t = bmath.Vector()
				try:
					t[0] = float(value)
				except:
					self.flair.notify("WARNING",
						"invalid translate-X value",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix.translate(t)
			elif trans == "TY":
				t = bmath.Vector()
				try:
					t[1] = float(value)
				except:
					self.flair.notify("WARNING",
						"invalid translate-Y value",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix.translate(t)
			elif trans == "TZ":
				t = bmath.Vector()
				try:
					t[2] = float(value)
				except:
					self.flair.notify("WARNING",
						"invalid translate-Z value",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix.translate(t)
			elif trans == "RX":
				try:
					a = math.radians(float(value))
				except:
					a = 0.0
					self.flair.notify("WARNING",
						"invalid rotate-X value",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix.rotX(-a)
			elif trans == "RY":
				try:
					a = math.radians(float(value))
				except:
					a = 0.0
					self.flair.notify("WARNING",
						"invalid rotate-Y value",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix.rotY(-a)
			elif trans == "RZ":
				try:
					a = math.radians(float(value))
				except:
					a = 0.0
					self.flair.notify("WARNING",
						"invalid rotate-Z value",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix.rotZ(-a)
			elif trans == "MX":
				if value!="":
					self.flair.notify("WARNING",
						"no value required for mirror-X",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix(4,type=1)
				m[0][0] = -1.0
			elif trans == "MY":
				if value!="":
					self.flair.notify("WARNING",
						"no value required for mirror-Y",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix(4,type=1)
				m[1][1] = -1.0
			elif trans == "MZ":
				if value!="":
					self.flair.notify("WARNING",
						"no value required for mirror-Z",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix(4,type=1)
				m[2][2] = -1.0
			elif trans == "S":
				try:
					s = float(value)
				except:
					s = 1.0
					self.flair.notify("WARNING",
						"invalid Scale value",
						tkFlair.NOTIFY_WARNING)
				m = bmath.Matrix.scale(s)
			else:
				continue

			matrix = m * matrix

		if origin:
			matrix = bmath.Matrix.translate(origin) * matrix

		self.flair.log("\nTransformation matrix=")
		self.flair.log(str(matrix))
		undoinfo = [self.flair.refreshUndo()]
		for card in self.cardlist:
			undoinfo.append(self.flair.saveCardUndoInfo(card.pos()))
			self.flair.log("\n>>> Card=")
			self.flair.log(str(card))
			if self.input.transformBody(card, matrix):
				self.flair.log("*** Invalid card for transformation...")
			else:
				self.flair.log("<<< Card=")
				self.flair.log(str(card))
		undoinfo.append(self.flair.refreshUndo())
		self.flair.addUndo(undoinfo,"Transform cards")

		self.flair.inputModified()
		self.flair.refresh("card")

	# ----------------------------------------------------------------------
	def getTransformations(self):
		# Remember transformations
		del TransformBodies.transformList[:]
		for i in range(self.listbox.size()):
			TransformBodies.transformList.append(self.listbox.get(i))

	# ----------------------------------------------------------------------
	def add2Input(self, event=None):
		# Get the name from the combo box
		name = self.combo.get().strip()
		hashtag = name.find('#')
		idx   = 0
		if hashtag >= 0:
			try:
				idx = int(name[hashtag+1:])
				name = ""
			except:
				pass
		else:
			try:
				idx = int(name)
			except:
				idx = 0
		if idx==0 and name=="": return

		# Check if ROT-DEFI already exists
		undoinfo = [(self.flair.refreshUndo,True), self.flair.renumberUndo()]
		found = False
		for card in self.input["ROT-DEFI"]:
			w1 = card.intWhat(1)
			if w1 < 1000:
				(j, i) = divmod(w1, 100)
			else:
				(i, j) = divmod(w1, 1000)
			if (idx!=0 and i==idx) or (name!="" and name==card.name()):
				found = True
				break
		if found:
			ans = tkFlair.askyesno(f"ROT-DEFI {name} already exists",
				f"ROT-DEFI cards with id='{idx}' name='{name}' already"
				"exists in input. Do you want to override?",
				parent=self)
			if not ans: return
			# Delete existing ROT-DEFI cards...
			for card in self.input["ROT-DEFI"]:
				w1 = card.intWhat(1)
				if w1 < 1000:
					(j, i) = divmod(w1, 100)
				else:
					(i, j) = divmod(w1, 1000)
				if (idx!=0 and i==idx) or (name!="" and name==card.name()):
					undoinfo.append(
						self.flair.delCardUndo(card.pos(), False))

		# Create the necessary ROT-DEFI cards
		pos = self.input.bestPosition("ROT-DEFI")
		self.getTransformations()
		for trans, value in TransformBodies.transformList:
			what = [""]*7
			what[0] = name
			if trans == "T":
				axis  = 0
				value = value.split()
				for i in range(3):
					try:
						what[4+i] = float(value[i])
					except:
						self.flair.notify("WARNING"
							"Invalid T vector",
							tkFlair.NOTIFY_WARNING)
						break
			elif trans in ("TX", "TY", "TZ"):
				axis = ord(trans[1])-ord('X')+1
				try:
					what[3+axis] = float(value)
				except:
					self.flair.notify("WARNING",
						f"Invalid {trans} value",
						tkFlair.NOTIFY_WARNING)
					break
			elif trans in ("RX", "RY", "RZ"):
				axis = ord(trans[1])-ord('X')+1
				try:
					what[3] = float(value)
				except:
					self.flair.notify("WARNING",
						f"Invalid {trans} value",
						tkFlair.NOTIFY_WARNING)
					break
			else:
				self.flair.notify("WARNING",
					"No Scaling is allowed in ROT-DEFI",
					tkFlair.NOTIFY_WARNING)
				break
			if idx < 100:
				what[1] = axis*100 + idx
			else:
				what[1] = idx*1000 + axis

			card = FlukaCard("ROT-DEFI", what,
				"Added by flair Transformation Dialog")
			undoinfo.append(self.flair.addCardUndo(card, pos, False))
			pos += 1

		undoinfo.append(self.flair.renumberUndo())
		undoinfo.append(undoinfo[0])
		self.flair.addUndo(undoinfo,"Add ROT-DEFI cards")
		self.combo.fill(self.input.rotationList())
		self.combo._text.delete(0,tk.END)
		self.combo._text.insert(0,name)
		self.flair.redraw()

	# ----------------------------------------------------------------------
	def getFromInput(self, event=None):
		name = self.combo.get().strip()
		hashtag = name.find('#')
		idx   = 0
		if hashtag >= 0:
			try:
				idx = int(name[hashtag+1:])
				name = ""
			except:
				pass
		else:
			try:
				idx = int(name)
			except:
				idx = 0
		if idx==0 and name=="": return

		self.listbox.delete(0, tk.END)
		for card in self.input["ROT-DEFI"]:
			if card.ignore(): continue
			w1 = card.intWhat(1)
			if w1 < 1000:
				(j, i) = divmod(w1, 100)
			else:
				(i, j) = divmod(w1, 1000)
			if i==idx and card.name()==name:
				if j==0: j=3
				theta = card.numWhat(2)
				phi   = card.numWhat(3)
				x = card.numWhat(4)
				y = card.numWhat(5)
				z = card.numWhat(6)
				if x!=0.0 or y!=0.0 or z!=0.0:
					self.listbox.insert(tk.END,
						('T',"%g %g %g"%(x,y,z)))
				if phi!=0.0:
					self.listbox.insert(tk.END,
						('R%s'%("XYZ"[j-1]), str(phi)))
				if theta!=0.0:
					self.listbox.insert(tk.END,
						('R%s'%("ZXY"[j-1]), str(theta)))
		self.listbox.selection_set(0)
		self.listbox.activate(0)

#===============================================================================
# Repeat bodies
#===============================================================================
class RepeatBodies(tk.Frame):
	Nu = 5
	Nv = 1
	Nw = 1
	U  = bmath.Vector.X
	V  = bmath.Vector.Y
	W  = bmath.Vector.Z

	def __init__(self, master, flair, cardlist):
		super().__init__(master)
		self.flair    = flair
		self.input    = flair.project.input
		self.cardlist = [x for x in cardlist if x.type()==FlukaCard.BODY]

		self.nu = tk.IntVar()
		self.nv = tk.IntVar()
		self.nw = tk.IntVar()

		self.nu.set(RepeatBodies.Nu)
		self.nv.set(RepeatBodies.Nv)
		self.nw.set(RepeatBodies.Nw)

		# ---- Labels ----
		row = 0
		tk.Label(self, text="Axis").grid(row=row, column=0)
		tk.Label(self, text="N (#)").grid(row=row, column=1)
		tk.Label(self, text="x").grid(row=row, column=2)
		tk.Label(self, text="y").grid(row=row, column=3)
		tk.Label(self, text="z").grid(row=row, column=4)

		# ---------- U ---------
		row += 1
		tk.Label(self, text="U:", foreground="DarkRed").grid(row=row, column=0, sticky=tk.E)

		nu = tk.Spinbox(self, text=self.nu, background="White", from_=1, to=10000, width=4)
		nu.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(nu, "Number of replicas along U axis")

		self.ux = tkExtra.FloatEntry(self, background="White", width=8)
		self.ux.grid(row=row, column=2, sticky=tk.EW)
		tkExtra.Balloon.set(self.ux, "Ux direction")

		self.uy = tkExtra.FloatEntry(self, background="White", width=8)
		self.uy.grid(row=row, column=3, sticky=tk.EW)
		tkExtra.Balloon.set(self.uy, "Uy direction")

		self.uz = tkExtra.FloatEntry(self, background="White", width=8)
		self.uz.grid(row=row, column=4, sticky=tk.EW)
		tkExtra.Balloon.set(self.uz, "Uz direction")

		tk.Button(self, text="x", command=self.setUx, takefocus=False).grid(row=row, column=5)
		tk.Button(self, text="y", command=self.setUy, takefocus=False).grid(row=row, column=6)
		tk.Button(self, text="z", command=self.setUz, takefocus=False).grid(row=row, column=7)

		# ---------- V ---------
		row += 1
		tk.Label(self, text="V:", foreground="DarkGreen").grid(row=row, column=0, sticky=tk.E)

		nv = tk.Spinbox(self, text=self.nv, background="White", from_=1, to=10000, width=4)
		nv.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(nv, "Number of replicas along V axis")

		self.vx = tkExtra.FloatEntry(self, background="White", width=8)
		self.vx.grid(row=row, column=2, sticky=tk.EW)
		tkExtra.Balloon.set(self.vx, "Vx direction")

		self.vy = tkExtra.FloatEntry(self, background="White", width=8)
		self.vy.grid(row=row, column=3, sticky=tk.EW)
		tkExtra.Balloon.set(self.vy, "Vy direction")

		self.vz = tkExtra.FloatEntry(self, background="White", width=8)
		self.vz.grid(row=row, column=4, sticky=tk.EW)
		tkExtra.Balloon.set(self.vz, "Vz direction")

		tk.Button(self, text="x", command=self.setVx, takefocus=False).grid(row=row, column=5)
		tk.Button(self, text="y", command=self.setVy, takefocus=False).grid(row=row, column=6)
		tk.Button(self, text="z", command=self.setVz, takefocus=False).grid(row=row, column=7)

		# ---------- W ---------
		row += 1
		tk.Label(self, text="W:", foreground="DarkBlue").grid(row=row, column=0, sticky=tk.E)

		nw = tk.Spinbox(self, text=self.nw, background="White", from_=1, to=10000, width=4)
		nw.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(nw, "Number of replicas along W axis")

		self.wx = tkExtra.FloatEntry(self, background="White", width=8)
		self.wx.grid(row=row, column=2, sticky=tk.EW)
		tkExtra.Balloon.set(self.wx, "Wx direction")

		self.wy = tkExtra.FloatEntry(self, background="White", width=8)
		self.wy.grid(row=row, column=3, sticky=tk.EW)
		tkExtra.Balloon.set(self.wy, "Wy direction")

		self.wz = tkExtra.FloatEntry(self, background="White", width=8)
		self.wz.grid(row=row, column=4, sticky=tk.EW)
		tkExtra.Balloon.set(self.wz, "Wz direction")

		tk.Button(self, text="x", command=self.setWx, takefocus=False).grid(row=row, column=5)
		tk.Button(self, text="y", command=self.setWy, takefocus=False).grid(row=row, column=6)
		tk.Button(self, text="z", command=self.setWz, takefocus=False).grid(row=row, column=7)

		self.grid_columnconfigure(1, weight=1)
		self.grid_columnconfigure(2, weight=2)
		self.grid_columnconfigure(3, weight=2)
		self.grid_columnconfigure(4, weight=2)

		# ---- Fill values ----
		self.ux.insert(0,str(RepeatBodies.U[0]))
		self.uy.insert(0,str(RepeatBodies.U[1]))
		self.uz.insert(0,str(RepeatBodies.U[2]))

		self.vx.insert(0,str(RepeatBodies.V[0]))
		self.vy.insert(0,str(RepeatBodies.V[1]))
		self.vz.insert(0,str(RepeatBodies.V[2]))

		self.wx.insert(0,str(RepeatBodies.W[0]))
		self.wy.insert(0,str(RepeatBodies.W[1]))
		self.wz.insert(0,str(RepeatBodies.W[2]))

	# ----------------------------------------------------------------------
	def setU(self, x,y,z):
		self.ux.delete(0,tk.END)
		self.uy.delete(0,tk.END)
		self.uz.delete(0,tk.END)

		self.ux.insert(0, str(x))
		self.uy.insert(0, str(y))
		self.uz.insert(0, str(z))

	# ----------------------------------------------------------------------
	def setV(self, x,y,z):
		self.vx.delete(0,tk.END)
		self.vy.delete(0,tk.END)
		self.vz.delete(0,tk.END)

		self.vx.insert(0, str(x))
		self.vy.insert(0, str(y))
		self.vz.insert(0, str(z))

	# ----------------------------------------------------------------------
	def setW(self, x,y,z):
		self.wx.delete(0,tk.END)
		self.wy.delete(0,tk.END)
		self.wz.delete(0,tk.END)

		self.wx.insert(0, str(x))
		self.wy.insert(0, str(y))
		self.wz.insert(0, str(z))

	# ----------------------------------------------------------------------
	def setUx(self):	self.setU(1.0,0.0,0.0)
	def setUy(self):	self.setU(0.0,1.0,0.0)
	def setUz(self):	self.setU(0.0,0.0,1.0)

	def setVx(self):	self.setV(1.0,0.0,0.0)
	def setVy(self):	self.setV(0.0,1.0,0.0)
	def setVz(self):	self.setV(0.0,0.0,1.0)

	def setWx(self):	self.setW(1.0,0.0,0.0)
	def setWy(self):	self.setW(0.0,1.0,0.0)
	def setWz(self):	self.setW(0.0,0.0,1.0)

	# ----------------------------------------------------------------------
	def get(self):
		RepeatBodies.Nu = int(self.nu.get())
		RepeatBodies.Nv = int(self.nv.get())
		RepeatBodies.Nw = int(self.nw.get())

		RepeatBodies.U[0] = float(self.ux.get())
		RepeatBodies.U[1] = float(self.uy.get())
		RepeatBodies.U[2] = float(self.uz.get())

		RepeatBodies.V[0] = float(self.vx.get())
		RepeatBodies.V[1] = float(self.vy.get())
		RepeatBodies.V[2] = float(self.vz.get())

		RepeatBodies.W[0] = float(self.wx.get())
		RepeatBodies.W[1] = float(self.wy.get())
		RepeatBodies.W[2] = float(self.wz.get())

	# ----------------------------------------------------------------------
	def transform(self):
		if not self.cardlist:
			self.flair.notify("WARNING",
				"No body is selected",
				tkFlair.NOTIFY_WARNING)
		self.get()
		bodynames = set(list(self.input.cardsCache(Card.BODY).keys()))
		pos = self.cardlist[-1].pos()+1
		undoinfo = [self.flair.refreshUndo()]
		for w in range(RepeatBodies.Nw):
			v1 = w*bmath.Vector(RepeatBodies.W)
			for v in range(RepeatBodies.Nv):
				v2 = v1 + v*bmath.Vector(RepeatBodies.V)
				for u in range(RepeatBodies.Nu):
					if u==0 and v==0 and w==0: continue
					vector = v2 + u*bmath.Vector(RepeatBodies.U)
					matrix = bmath.Matrix.translate(vector)
					for card in self.cardlist:
						clone = card.clone()
						# Find unique name
						# find number in the end of the name
						pat = Project._LAST_NUM.match(clone.name())
						if pat:
							name = pat.group(1)
							n = int(pat.group(2))+1
						else:
							name = clone.name()
							n = 1
						while True:
							sn = str(n)
							guess = "%s%s"%(name[:8-len(sn)],sn)
							if guess not in bodynames:
								clone.setSdum(guess)
								break
							n += 1
						bodynames.add(clone.name())
						undoinfo.append(
							self.flair.addCardUndo(clone, pos, True))
						self.input.transformBody(clone, matrix)
						pos += 1

		undoinfo.append(self.flair.refreshUndo())
		self.flair.addUndo(undoinfo,"Repeat bodies cards")

		self.flair.inputModified()
		self.flair.refresh("card")
