#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   25-Nov-2019

import tkinter as tk
from tkinter import ttk

import bmath
import icons
import Manual

from Global import VIEWER

MAXERRORS  = 50

#===============================================================================
# Errors Frame
#===============================================================================
class ErrorsFrame(tk.Frame):
	#-----------------------------------------------------------------------
	def __init__(self, master, editor):
		super().__init__(master)
		self.editor = editor

		# --------------------------------------------------------------
		tk.Button(self, text="Refresh", image=icons.get("refresh"),
			compound=tk.LEFT, takefocus=tk.FALSE,
			command=self.refresh).pack(side=tk.TOP, fill=tk.X)

		# --------------------------------------------------------------
		sbh = ttk.Scrollbar(self, orient=tk.HORIZONTAL)
		sbh.pack(side=tk.BOTTOM, fill=tk.X)
		sbv = ttk.Scrollbar(self, orient=tk.VERTICAL)
		sbv.pack(side=tk.RIGHT, fill=tk.Y)

		self.text = tk.Text(self, width=80, cursor="arrow", wrap="none")
		self.text.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		self.text.bind("<Button-1>",	self.button1)
		self.text.bind("<Escape>",	self.escape)
		self.text.bind("<B1-Motion>",	self.donothing)
		self.text.bind("<Double-1>",	self.donothing)
		self.text.bind("<<Help>>",	self.help)
		self.text.bind("<F1>",		self.help)

		self.text.config(yscrollcommand=sbv.set, xscrollcommand=sbh.set, state=tk.DISABLED)
		sbh.config(command=self.text.xview)
		sbv.config(command=self.text.yview)

		# Remember error-list
		self._hash   = [0, 0, 0, 0]
		self._errors = [0, 0, 0, 0]
		self._cache  = [[], [], [], []]
		self._show   = [True, True, True, True, True]
		self._select = ""
		self._expand = {}

	# ----------------------------------------------------------------------
	# Refresh all errors
	# ----------------------------------------------------------------------
	def refresh(self, event=None):
		for i,viewer in enumerate(self.editor.viewers):
			self.readErrors(i,viewer)
		self.fill()

	# ----------------------------------------------------------------------
	# Read errors from viewport
	# ----------------------------------------------------------------------
	def readErrors(self, idx, viewer):
		self._hash[idx]   = viewer.error("hash")
		self._errors[idx] = viewer.error("n")
		lst = self._cache[idx]
		del lst[:]
		for i in range(1,min(MAXERRORS, self._errors[idx]+1)):
			err = viewer.error("get",i)
			if err is None: break
			lst.append(err)

	# ----------------------------------------------------------------------
	# Fill text with error messages
	# ----------------------------------------------------------------------
	def fill(self, event=None):
		self.text.config(state=tk.NORMAL)
		self.text.delete("0.0",tk.END)
		self.text.tag_delete("*")

		for i,name in enumerate(VIEWER.NAME):
			viewer = self.editor.viewers[i]
			if viewer.error("hash") != self._hash[i]:
				self.readErrors(i,viewer)
			self.insertErrors(i, name)

		ne = len(self.editor.viewers.errors)
		nw = len(self.editor.viewers.warnings)

		if self._show[4]:
			header = f"\N{BLACK DOWN-POINTING TRIANGLE} Input [Errors:{ne}, Warnings:{nw}]\n"
		else:
			header = f"\N{BLACK RIGHT-POINTING TRIANGLE} Input [Errors:{ne}, Warnings:{nw}]\n"

		self.text.insert(tk.END, header)
		self.text.tag_add("Input", "end - 2 lines", "end -1 lines")
		self.text.tag_configure("Input",   background="#FFFF70")

		if self._show[4]:
			if self.editor.viewers.errors:
				tag = "Errors"
				self.text.insert(tk.END, "Errors:\n",tag)
				self.text.tag_configure(tag, foreground="DarkRed")
				for i,err in enumerate(self.editor.viewers.errors):
					tag = f"Error.{i}"
					self.text.insert(tk.END, "%2d: %s\n"%(i+1,err), tag)
					#if not i%2: self.text.tag_config(tag, background="DarkGray")
			if self.editor.viewers.warnings:
				tag = "Warnings"
				if self.editor.viewers.errors: self.text.insert(tk.END, "\n")
				self.text.insert(tk.END, "Warnings:\n",tag)
				self.text.tag_configure(tag, foreground="Magenta")
				for i,err in enumerate(self.editor.viewers.warnings):
					tag = f"Warning.{i}"
					self.text.insert(tk.END, "%2d: %s\n"%(i+1,err), tag)
					#if not i%2: self.text.tag_config(tag, background="DarkGray")
		self.text.config(state=tk.DISABLED)

	# ----------------------------------------------------------------------
	# Insert errors in Text from a specific viewport
	# ----------------------------------------------------------------------
	def insertErrors(self, idx, name):
		if self._show[idx]:
			header = "\N{BLACK DOWN-POINTING TRIANGLE} %s [%d]\n"%\
				(name, abs(self._errors[idx]))
		else:
			header = "\N{BLACK RIGHT-POINTING TRIANGLE} %s [%d]\n"%\
				(name, abs(self._errors[idx]))

		self.text.insert(tk.END, header)
		self.text.tag_add(name,"end - 2 lines", "end -1 lines")
		self.text.tag_configure(name, background="#%06X"%(VIEWER.COLOR[idx]))

		if self._errors[idx] and self._show[idx]:
			viewer = self.editor.viewers[idx]
			viewer.error("show", MAXERRORS)
			viewer.expose()
			for i,err in enumerate(self._cache[idx]):
				x,y,z,xmin,xmax,ymin,ymax,body,regIn,regOut = err
				tag = f"{name}.{i+1}"
				if tag in self._expand:
					self.text.insert(tk.END, "-%2d:  %8s %8s %8s\n"
								 "   body: %s\n"
								 "   Reg-: %s\n"
								 "   Reg+: %s\n"
							% (i+1,
							   bmath.format(x,8),
							   bmath.format(y,8),
							   bmath.format(z,8),
							   body,
							   ",".join(regOut),
							   ",".join(regIn)),
							tag)
				else:
					self.text.insert(tk.END, "+%2d:  %8s %8s %8s\n"
							% (i+1,
							   bmath.format(x,8),
							   bmath.format(y,8),
							   bmath.format(z,8)),
							tag)
				#if not i%2: self.text.tag_config(tag, background="DarkGray")
		self.text.insert(tk.END, "\n")

	# ----------------------------------------------------------------------
	def button1(self, event=None):
		index = self.text.index(f"@{event.x},{event.y}")
		row,col = map(int,index.split("."))
		tags = self.text.tag_names(index)
		if not tags: return

		for tag in tags:
			if tag == tk.SEL: continue

			try:
				name,err = tag.split(".")
				err = int(err)
			except ValueError:
				name = tag
				err = -1

			try:
				idx = VIEWER.NAME.index(name)
			except ValueError:
				idx = 4 if name == "Input" else -1

			if err>=0:
				if col<3:
					if tag in self._expand:
						del self._expand[tag]
					else:
						self._expand[tag] = True
					self.fill()
				self.select(tag)
				if col>=3:
					self.showError(idx, err, row, col)
			else:
				self._show[idx] = not self._show[idx]
				self.fill()
			break

		self.text.see(index)
		return "break"

	# ----------------------------------------------------------------------
	def escape(self, event=None):
		self.select("")

	# ----------------------------------------------------------------------
	def select(self, tag):
		self.text.tag_config(self._select,
					background="",
					foreground="")
		self._select = tag
		if self._select:
			self.text.tag_config(self._select,
					background=self.text["selectbackground"],
					foreground=self.text["selectforeground"])

	# ----------------------------------------------------------------------
	def donothing(self, event=None):
		return "break"

	# ----------------------------------------------------------------------
	def showError(self, idx, err, row, col):
		if idx>=0:
			# Zoom on selected error
			self.editor.viewers.frames[idx].viewer.zoomOnError(err)
			return

		# display the error message
		self.editor.flair.showCard(self.text.get(f"{row}.0",f"{row}.end"))
		return

	# ----------------------------------------------------------------------
	def help(self, event=None):
		Manual.show("","Debugging")
