# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Dec-2006

__author__ = 'Vasilis Vlachoudis'
__email__  = 'Vasilis.Vlachoudis@cern.ch'

import io
import os
import re
import pickle
import tkinter as tk
from itertools import count
from tkinter.colorchooser import askcolor

import Data
import icons
import Global
import Manual
import tkFlair
import tkExtra
import Project
import CalibrateImage

from log import say
from deck import Card
from palette import paletteStore
from fluka import FlukaCard, FlukaParticle

_ADD = "<add>"

_PROJECTIONS = ["Orthographic", "Perspective", "Combo"]
_SKIPBLACK   = ["Stop", "Ignore first", "Ignore"]

# ------------------------- Colors ---------------------------------------------
COLORVALUES = [ 'None',			# 0
		'Region',		# 1
		'Material',		# 2
		'Material-Decay',	# 3
		'Corrfactor',
		'Corrfactor-dE/dx',
		'Corrfactor-rho',
		'Deltaray',
		'Density',
		'Field',
		'e-Production',
		'e-Transport',
		'g-Production',
		'g-Transport',
		'Importance',
		'Importance-E',
		'Importance-H',
		'Importance-N',
		'Splitting',
		'Splitting-E',
		'Splitting-H',
		'Splitting-N' ]

# Labels
LABELS = [ 'None', 'Region', 'Material', 'Value' ]
LABEL_NONE     = 0
LABEL_REGION   = 1
LABEL_MATERIAL = 2
LABEL_VALUE    = 3
_usrbinPat = re.compile(r'(\d+) (.+) \[(\d+)\]')

# ------------------------------------------------------------------------------
def getMarkers(layer):
	markers = []
	n = int(layer('Image.marker'))
	for i in range(n):
		key = f"Image.marker.{i}"
		mark = layer(key)
		if mark is None: return None
		markers.append(list(map(float,mark.split())))
	return markers

#===============================================================================
class Combobox(tkExtra.Combobox):
	def setGeometryLayersDialog(self, dlg):
		self._geoDlg = dlg

	# --------------------------------------------------------------------
	def beforeShow(self):
		self._geoDlg.get()
		self._geoDlg.fillLayers()

#===============================================================================
# Geometry Layers Frame
#===============================================================================
class GeometryLayersFrame(tk.Frame):
	_PAGES = None

	#-----------------------------------------------------------------------
	def __init__(self, master, flair):
		super().__init__(master)
		self.flair        = flair
		self.layer        = None
		self.activepage   = None
		self._apply_after = None
		self._inInit      = True	# Avoid updating

		# --- Layer ---
		frame = tk.Frame(self)
		frame.pack(side=tk.TOP, fill=tk.X)
		self.layerBox = Combobox(frame, False, width=8,
					command=self.changeLayer)
		self.layerBox.setGeometryLayersDialog(self)
		self.layerBox._text.config(background='White')
		self.layerBox._text.bind('<Return>',   self.nothing)
		self.layerBox._text.bind('<KP_Enter>', self.nothing)
		self.layerBox.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X)
		tkExtra.Balloon.set(self.layerBox, 'Layer name currently being edited')

		# buttons
		b = tk.Button(frame, image=icons.get('clone'), command=self.cloneLayer,
					pady=0, padx=3)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b, 'Clone active layer')

		b = tk.Button(frame, image=icons.get('add'), command=self.addLayer,
					pady=0, padx=3)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b, 'Add layer to project')

		b = tk.Button(frame, image=icons.get('del'), command=self.delLayer,
					pady=0, padx=3)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b, 'Delete layer from project')

		self.globalLayer = tk.BooleanVar(self)
		self.globalButton = tk.Checkbutton(frame, text='Global',
					variable=self.globalLayer,
					onvalue=1, offvalue=0,
					anchor=tk.W)
		tkExtra.Balloon.set(self.globalButton,
			'If selected the layer will be saved in the flair.ini, '
			"instead of the project and will be available to all projects")
		self.globalButton.pack(side=tk.RIGHT)

		# --- Action buttons ---
		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, fill=tk.X)
		tk.Button(frame, text='Help',
				image=icons.get('info'),
				compound=tk.LEFT,
				takefocus=tk.FALSE,
				command=self.help).pack(side=tk.LEFT, padx=1, pady=5)
		tk.Button(frame, text='Reset',
				image=icons.get('reset'),
				compound=tk.LEFT,
				takefocus=tk.FALSE,
				command=self.reset).pack(side=tk.LEFT, pady=5)
		self.applyButton = tk.Button(frame, text='Apply',
				image=icons.get('ok'),
				compound=tk.LEFT,
				takefocus=tk.FALSE,
				command=self.apply)
		self.applyButton.pack(side=tk.RIGHT, padx=5, pady=5)

		# --- Center frame ---
		splitter = tkExtra.VSplitter(self, 0.35)
		splitter.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

		# --- Selection listbox ---
		self.listbox = tkExtra.MultiListbox(splitter.topFrame(),
					(("Value", 16, None),
					 (" ",      3, None)),
					header=False,
					selectmode=tk.BROWSE,
					height=5,
					stretch="first",
					background="White")
		self.listbox.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=2, pady=2)
		self.listbox.bind('<<ListboxSelect>>',		self.listboxChanged)
		self.listbox.bindList('<space>',		self.selectToggle)
		self.listbox.bindList('<Return>',		self.add2Listbox)
		self.listbox.bindList('<KP_Enter>',		self.add2Listbox)
		self.listbox.bindList('<Insert>',		self.add2Listbox)
		self.listbox.bindList('<Delete>',		self.removeListbox)
		self.listbox.listbox(0).bind("<ButtonRelease-1>",self.listbox0Release)
		self.listbox.listbox(1).bind("<ButtonRelease-1>",self.listbox1Release)
		self.listbox.bindList('<<Copy>>',		self.copy)
		self.listbox.bindList('<<Cut>>',		self.cut)
		self.listbox.bindList('<<Paste>>',		self.paste)
		self.listbox.setPopupMenu([
				('Add',    0, self.add2Listbox,   icons.get("add")),
				('Clone',  0, self.cloneListbox,  icons.get("clone")),
				('Remove', 0, self.removeListbox, icons.get("x")),
				None,
				('Cut',    0, self.cut,           icons.get("cut")),
				('Copy',   0, self.copy,          icons.get("copy")),
				('Paste',  0, self.paste,         icons.get("paste")),
				])

		# ---- Bottom frame ----
		# Frame title
		self.title = tk.Label(splitter.bottomFrame(),
					foreground=tkFlair._TITLE_FOREGROUND_COLOR,
					background=tkFlair._TITLE_BACKACTIVE_COLOR,
					relief=tk.RAISED)
		#self.title.pack(side=tk.TOP, fill=tk.X, padx=2, pady=2)
		self.title.grid(row=0, column=0, columnspan=2, sticky=tk.EW)

		# main frame
		self._scrollFrame = tkExtra.ScrollFrame(splitter.bottomFrame(),
						borderwidth=2,
						relief=tk.SUNKEN)
		self._scrollFrame.grid(row=1, column=0, sticky=tk.NSEW, padx=2, pady=2)
		#self._scrollFrame.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH, padx=2, pady=2)
		self.mainframe = self._scrollFrame()

		#sbx=tkExtra.AutoScrollbar(splitter.bottomFrame(), orient=tk.HORIZONTAL,
		#		command=self._scrollFrame.xview)
		#sbx.grid(row=2,column=0,sticky=tk.NSEW)
		sby=tkExtra.AutoScrollbar(splitter.bottomFrame(), orient=tk.VERTICAL,
					command=self._scrollFrame.yview)
		sby.grid(row=1, column=1, sticky=tk.NS)
		#self._scrollFrame["xscrollcommand"]=sbx.set
		self._scrollFrame["yscrollcommand"]=sby.set
		splitter.bottomFrame().grid_columnconfigure(0, weight=1)
		splitter.bottomFrame().grid_rowconfigure(1, weight=1)

		# Bindings
		self.bind("<<Help>>",	self.help)
		self.bind("<Return>",	self.apply)
		self.bind("<KP_Enter>",	self.apply)

		# Initialize pages, but only after the class is defined!
		if GeometryLayersFrame._PAGES is None:
			GeometryLayersFrame._PAGES = {
				'Options'   : GeometryLayersFrame.optionsFrame,
				'Show'      : GeometryLayersFrame.showFrame,
				'Image'     : GeometryLayersFrame.imageFrame,
				'Beam'      : GeometryLayersFrame.emptyFrame,
				'Userdump'  : GeometryLayersFrame.userdumpFrame,
				'Usrbin'    : GeometryLayersFrame.usrbinFrame,
				'3D'        : GeometryLayersFrame.d3Frame,
				'Voxel'     : GeometryLayersFrame.voxelFrame,
				'Palette'   : GeometryLayersFrame.paletteFrame,
			}
		# Populate fonts
		self.populateFonts()

		# Populate pages
		self.pages = {}

		# default values
		for page, method in GeometryLayersFrame._PAGES.items():
			self.pages[page] = method(self)
		self._scrollFrame.defaultBinds()

		# copy pages for Usrbin# and Colorband#
		page = self.pages["Usrbin"]
		for i in range(2, Project.GeometryLayerInfo._MAXUSRBIN+1):
			self.pages[f"Usrbin{i}"] = page
		page = self.pages['Palette']
		for i in range(2, Project.GeometryLayerInfo._MAXPALETTE+1):
			self.pages[f"Palette{i}"] = page
		self._inInit = False	# Avoid updating

	# --------------------------------------------------------------------
	@property
	def project(self):
		return self.flair.project

	# --------------------------------------------------------------------
	def show(self, layer=None):
		# set the list of layers
		layers = self.project.geometry.layersList()
		if layer is not None or not self.layerBox.get():
			if layer not in layers: layer = layers[0]
			self.layerBox.fill(layers)
			self.layerBox.set(layer)

		# Select last page
		self.listbox.selection_clear(0,tk.END)
		oldPage = self.activepage
		self.activepage = None
		self.showPage(Project.GeometryLayerInfo._ORDER[0])	# Show options by default
		# do not show immediately Usrbin/Colorband otherwise it
		# will save empty options
		if oldPage is not None:
			self.listbox.selection_clear(0,tk.END)
			self.showPage(oldPage)

	# --------------------------------------------------------------------
	def refresh(self):
		self.d3Refresh()
		self.voxelRefresh()

	# --------------------------------------------------------------------
	# Apply button
	# --------------------------------------------------------------------
	def apply(self, event=None, force=False):
		self.applyButton.focus_set()
		self._apply(event, force)

	#-----------------------------------------------------------------------
	def _apply(self, event=None, force=False):
		if self._inInit: return
		if self._apply_after is not None:
			self.after_cancel(self._apply_after)
			self._apply_after = None
		oldHash = hash(self.project.geometry)
		self.get()
		self.project.geometry.correctGlobalLayers()
		tkFlair.updateLayers()
		if hash(self.project.geometry) != oldHash or force:
			# FIXME unfortunately data= is one of the missing parameters in tkinter!!
			# and it is not copied in the event class
			self.event_generate('<<LayersUpdate>>') #,data=self.layer.name)

			if self.projection.get() == _PROJECTIONS[0]:
				self.fov["state"] = tk.DISABLED
				self._fovLabel["state"] = tk.DISABLED
			else:
				self.fov["state"] = tk.NORMAL
				self._fovLabel["state"] = tk.NORMAL

		# For debugging multiple applies
		#say('\n====================================================')
		#import traceback; traceback.print_stack()

	# --------------------------------------------------------------------
	def applyLate(self, event=None):
		if self._apply_after is not None: self.after_cancel(self._apply_after)
		self._apply_after = self.after_idle(self._apply)

	# --------------------------------------------------------------------
	def nothing(self, event=None):
		return 'break'

	# --------------------------------------------------------------------
	def entry(self, key):
		return f"{self.activepage}.{key}"

	# ----------------------------------------------------------------------
	def copy(self, event=None):
		self.get()
		selection = {}
		item = None
		for i in self.listbox.curselection():
			item = self.listbox.get(i)[0]
			for name,value in self.layer.items():
				if name.startswith(item):
					selection[name] = value
		if item is None: return

		# Write to clipboard
		bio = io.BytesIO()
		pickler = pickle.Pickler(bio)
		pickler.dump(item)
		pickler.dump(selection)
		self.clipboard_clear()
		self.flair.clipboardAppend(bio, tkFlair._LAYER_MIME)
		return "break"

	# ----------------------------------------------------------------------
	def cut(self, event=None):
		return "break"

	# ----------------------------------------------------------------------
	def paste(self, event=None):
		self.get()
		if tkFlair._LAYER_MIME in self.flair.clipboardTargets():
			# Pickle format
			unpickler = self.flair.clipboardUnpickler(tkFlair._LAYER_MIME)
			try:
				item      = unpickler.load()
				selection = unpickler.load()
			except:
				self.flair.notify("Cannot paste",
						"Invalid content in clipboard",
						tkFlair.NOTIFY_WARNING)
				return "break"
		else:
			self.flair.notify("Cannot paste",
					"Invalid content in clipboard",
					tkFlair.NOTIFY_WARNING)
			return "break"

		self.layer.update(selection)
		self.showPage(item)
		return "break"

	# --------------------------------------------------------------------
	# Display new page
	# --------------------------------------------------------------------
	def listboxChanged(self, event):
		sel = self.listbox.get(self.listbox.curselection())
		if sel and sel[0] != _ADD:
			self.showPage(sel[0])

	# --------------------------------------------------------------------
	# Populate listbox with visible and active layers
	# --------------------------------------------------------------------
	def populateListbox(self):
		self.listbox.delete(0,tk.END)
		self.listbox.insert(tk.END, (Project.GeometryLayerInfo._ORDER[0], tkFlair._ON))
		for page in Project.GeometryLayerInfo._ORDER[1:]:
			on = self.layer(page)
			if on or self.layer(page+'._visible'):
				self.layer[page+'._visible'] = True
				self.listbox.insert(tk.END, (page,
						tkFlair._ON if on else tkFlair._OFF))
		self.listbox.insert(tk.END, (_ADD,''))
		self.listbox.itemconfig(tk.END, foreground=tkFlair._CLABEL_FOREGROUND_COLOR)

		if int(self.layer('_system')):
			self.globalButton.config(state=tk.DISABLED)
		else:
			self.globalButton.config(state=tk.NORMAL)
		self.listbox.itemconfigure(0, fg='DarkRed')

	# --------------------------------------------------------------------
	# Add layer to listbox
	# --------------------------------------------------------------------
	def add2Listbox(self, event=None):
		self.flair.modified()
		# Add only on the non-system layers
		self.get()

		# Layers that could be added
		layers = []
		# Add only one of these layers
		usrbin  = False
		palette = False
		for page in Project.GeometryLayerInfo._ORDER[1:]:
			if not self.layer(page) and \
			   not self.layer(page+"._visible"):
				if page.startswith("Usrbin"):
					if usrbin: continue
					usrbin = True
				elif page.startswith("Palette"):
					if palette: continue
					palette = True
				layers.append(page)
		edit = tkExtra.InPlaceList(self.listbox.listbox(0),
				values=layers)
		if edit and edit.value:
			self.layer[edit.value] = True
			self.layer[edit.value+"._visible"] = True
			self.populateListbox()
			self.showPage(edit.value)
		#else:
		#	print("Restore <add>")
		return 'break'

	# --------------------------------------------------------------------
	def cloneListbox(self, event=None):
		if not self.activepage.startswith("Usrbin") and \
		   not self.activepage.startswith("Palette"):
			return
		self.flair.modified()

		# Search what to add
		if self.activepage.startswith("Usrbin"):
			for page in Project.GeometryLayerInfo.USRBINNAMES:
				if not self.layer(page) and \
				   not self.layer(page+"._visible"):
					break
			else:
				return

		elif self.activepage.startswith("Palette"):
			for page in Project.GeometryLayerInfo.PALETTENAMES:
				if not self.layer(page) and \
				   not self.layer(page+"._visible"):
					break
			else:
				return

		self.layer[page] = True
		self.layer[page+"._visible"] = True
		self.populateListbox()
		# Don't call showPage() but simply activate listbox
		for i,p in enumerate(self.listbox.listbox(0).get(0,tk.END)):
			if p==page:
				self.listbox.selection_set(i)
				self.listbox.activate(i)
				break
		self.activepage = page
		self.applyLate()

	# --------------------------------------------------------------------
	def removeListbox(self, event=None):
		if self.activepage == Project.GeometryLayerInfo._ORDER[0]: return
		self.flair.modified()

		# find selected item
		active = self.listbox.index(tk.ACTIVE)

		self.layer[self.activepage] = False
		self.layer[self.activepage + "._visible"] = False
		self.populateListbox()

		try:
			page = self.listbox.listbox(0).get(active)
			self.showPage(page)
		except:
			pass
		self.applyLate()

	# --------------------------------------------------------------------
	def showPage(self, page):
		if not page or \
		   not self.layer(page+"._visible"):
			page = Project.GeometryLayerInfo._ORDER[0]

		for i,p in enumerate(self.listbox.listbox(0).get(0,tk.END)):
			if p==page:
				self.listbox.selection_set(i)
				self.listbox.activate(i)
				break

		if self.activepage is not None and \
		   self.activepage != page:
			self.usrbinUpdate()
			self.paletteUpdate()
			if self.activepage == "Voxel": self.voxelUpdate()
			try: self.pages[self.activepage].pack_forget()
			except KeyError: pass

		self.activepage = page
		self.title.config(text=page)
		if   page == '3D':    self.d3Refresh()
		elif page == "Voxel": self.voxelRefresh()
		self.usrbinInit()
		self.paletteInit()

		try: self.pages[page].pack(expand=tk.YES, fill=tk.BOTH)
		except KeyError: pass
		self._scrollFrame.updateScrollRegion()

	# --------------------------------------------------------------------
	def listbox0Release(self, event):
		if len(self.listbox.curselection())>1: return

		active = self.listbox.nearest(event.y)
		self.listbox.activate(active)
		if active and self.listbox.listbox(0).get(active) == _ADD:
			self.add2Listbox()

	# --------------------------------------------------------------------
	def listbox1Release(self, event):
		if len(self.listbox.curselection())>1: return

		active = self.listbox.nearest(event.y)
		if active is not None:
			self.listbox.activate(active)
		else:
			return

		self.selectToggle()

	# --------------------------------------------------------------------
	def selectToggle(self, event=None):
		lst = self.listbox.listbox(1)
		for i in map(int,lst.curselection()):
			if i==0: continue
			sel = lst.get(i)
			if not sel: continue
			lst.set(i,tkFlair._ON if sel==tkFlair._OFF else tkFlair._OFF)
			lst.selection_set(i)
		self.applyLate()

	# --------------------------------------------------------------------
	# Populate the font list
	# Scan the .tga in the system/font directory then ini/font
	# --------------------------------------------------------------------
	def populateFonts(self):
		self.fonts = []

		fonts = {}
		# User defined tga fonts
		try:
			for f in os.listdir(os.path.join(tkFlair.prgDir,'fonts')):
				fn,ext = os.path.splitext(f)
				if ext=='.tga': fonts[fn] = True
		except OSError:
			self.flair.notify('Fonts directory not found',
				'Please check your flair installation. the "fonts" '
				'directory is not found!',
				tkFlair.NOTIFY_WARNING)
			return

		# flair default fonts
		try:
			for f in os.listdir(os.path.join(tkFlair.iniDir,'fonts')):
				fn,ext = os.path.splitext(f)
				if ext=='.tga': fonts[fn] = True
		except OSError:
			pass

		# Try system bitmap fonts
		#for f in subprocess.check_output('xlsfonts').splitlines():
		#	if f[0]=='-': continue
		#	fonts[f] = True
		self.fonts = sorted(fonts)

	# --------------------------------------------------------------------
	# Create options frame
	# --------------------------------------------------------------------
	def optionsFrame(self):
		pageframe = tk.Frame(self.mainframe)

		self.axes       = tk.BooleanVar(self)
		self.ocube      = tk.BooleanVar(self)
		self.showErrors = tk.BooleanVar(self)
		self.gridVar    = tk.BooleanVar(self)
		self.viewport   = tk.BooleanVar(self)

		# ---
		row = 0
		b = tk.Checkbutton(pageframe, text='Coordinate system',
					variable=self.axes,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Draw coordinate system on bottom left corner")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text='Viewport lines',
					variable=self.viewport,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Display other viewport cross sections")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text='Show Errors',
					variable=self.showErrors,
					onvalue=1, offvalue=0,
					foreground="Red",
					activeforeground="Red",
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Show Errors. Disabling this you will not be notified for geometry errors")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text='Grid',
					variable=self.gridVar,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Draw grid")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		# ---
		row += 1
		b = tk.Checkbutton(pageframe, text='Orientation Cube',
					variable=self.ocube,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		b.grid(row=row, column=0, sticky=tk.W)
		tkExtra.Balloon.set(b,
			"Display on/off orientation cube")
		self.ocubelength = tk.Scale(pageframe,
					from_=8, to_=32,
					orient=tk.HORIZONTAL,
					showvalue=0,
					command=self.applyLate)
		tkExtra.Balloon.set(self.ocubelength,
			"Set size of orientation cube")
		self.ocubelength.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Grid Level').grid(
			row=row, column=0, sticky=tk.W)
		self.gridlevel = tk.Scale(pageframe,
					from_=32, to_=0,
					orient=tk.HORIZONTAL,
					showvalue=0,
					command=self.applyLate)
		tkExtra.Balloon.set(self.gridlevel,
			"Set the grid lines intensity level")
		self.gridlevel.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Lattice Level').grid(
			row=row, column=0, sticky=tk.W)
		self.latticeLevel = tk.Scale(pageframe,
					from_=0, to_=32,
					orient=tk.HORIZONTAL,
					showvalue=0,
					command=self.applyLate)
		tkExtra.Balloon.set(self.latticeLevel,
			"Set the lattice&voxel hash lines intensity level")
		self.latticeLevel.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Crosshair:').grid(
			row=row, column=0, sticky=tk.W)
		self.crosshair = tk.Scale(pageframe,
					from_=0, to_=32,
					orient=tk.HORIZONTAL,
					showvalue=0,
					command=self.applyLate)
		tkExtra.Balloon.set(self.crosshair,
			"Draw crosshair at the center of the plot")
		self.crosshair.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Text Background').grid(
			row=row, column=0, sticky=tk.W)
		self.background = tk.Scale(pageframe,
					from_=0, to_=255,
					orient=tk.HORIZONTAL,
					showvalue=0,
					command=self.applyLate)
		tkExtra.Balloon.set(self.background,
			"Text background level")
		self.background.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='General Font:').grid(
			row=row, column=0, sticky=tk.W)
		self.generalFont = tkExtra.Combobox(pageframe, width=10,
					command=self.applyLate)
		tkExtra.Balloon.set(self.generalFont,
			"Bitmap font used in Geometry (located in [flair]/fonts folder")
		self.generalFont.grid(row=row, column=1, sticky=tk.EW)
		self.generalFont.fill(self.fonts)

		# ---
		row += 1
		tk.Label(pageframe, text='Grid Font:').grid(
			row=row, column=0, sticky=tk.W)
		self.gridFont = tkExtra.Combobox(pageframe, width=10,
					command=self.applyLate)
		tkExtra.Balloon.set(self.gridFont,
			"Bitmap font used in Geometry (located in [flair]/fonts folder")
		self.gridFont.grid(row=row, column=1, sticky=tk.EW)
		self.gridFont.fill(self.fonts)

		# ---
		row += 1
		tk.Label(pageframe, text='Palette Font:').grid(
			row=row, column=0, sticky=tk.W)
		self.paletteFont = tkExtra.Combobox(pageframe, width=10,
					command=self.applyLate)
		tkExtra.Balloon.set(self.paletteFont,
			"Bitmap font used in Geometry (located in [flair]/fonts folder")
		self.paletteFont.grid(row=row, column=1, sticky=tk.EW)
		self.paletteFont.fill(self.fonts)

		# ---
		pageframe.grid_columnconfigure(1, weight=1)

		return pageframe

	# --------------------------------------------------------------------
	# Show frame
	# --------------------------------------------------------------------
	def showFrame(self):
		pageframe = tk.Frame(self.mainframe)

		self.showlattice = tk.IntVar(self)
		self.showvoxel   = tk.IntVar(self)
		self.showvertex  = tk.BooleanVar(self)

		# ---
		row = 0
		b = tk.Checkbutton(pageframe, text='Vertices',
					variable=self.showvertex,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Display vertices on the intersections of selected bodies")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		# ---
		row += 1
		l = tk.Label(pageframe, text='Labels:')
		l.grid(row=row, column=0, sticky=tk.W)

		self.labels = tkExtra.Combobox(pageframe, width=10,
					command=self.applyLate)
		self.labels.grid(row=row, column=1, sticky=tk.EW)
		self.labels.fill(LABELS)

		# ---
		row += 1
		l = tk.Label(pageframe, text="Color:")
		l.grid(row=row, column=0, sticky=tk.W)

		self.colorvalue = tkExtra.Combobox(pageframe, width=10,
					command=self.check4ShowPalette)
		self.colorvalue.grid(row=row, column=1, sticky=tk.EW)
		self.colorvalue.fill(COLORVALUES)

		# ---
		row += 1

		l = tk.Label(pageframe, text="Lattices:")
		l.grid(row=row, column=0, sticky=tk.W)

		f = tk.Frame(pageframe)
		f.grid(row=row, column=1, sticky=tk.EW)

		b = tk.Radiobutton(f, text="Off",
					variable=self.showlattice,
					value=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b, "Don't show lattice regions")
		b.pack(side=tk.LEFT)

		b = tk.Radiobutton(f, text="Show",
					variable=self.showlattice,
					value=1,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b, "Show lattice regions")
		b.pack(side=tk.LEFT)

		b = tk.Radiobutton(f, text="Prompt",
					foreground="DarkRed",
					variable=self.showlattice,
					value=2,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b, "Show promptly lattices. Can slow down refreshing")
		b.pack(side=tk.LEFT)

		# ---
		row += 1
		l = tk.Label(pageframe, text="Voxel:")
		l.grid(row=row, column=0, sticky=tk.W)

		f = tk.Frame(pageframe)
		f.grid(row=row, column=1, sticky=tk.EW)

		b = tk.Radiobutton(f, text="Off",
					variable=self.showvoxel,
					value=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b, "Don't show any voxel")
		b.pack(side=tk.LEFT)

		b = tk.Radiobutton(f, text="Show",
					variable=self.showvoxel,
					value=1,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b, "Show voxels")
		b.pack(side=tk.LEFT)

		b = tk.Radiobutton(f, text="Prompt",
					foreground="DarkRed",
					variable=self.showvoxel,
					value=2,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b, "Show promptly voxels. Can slow down refreshing")
		b.pack(side=tk.LEFT)

		# ---
		row += 1
		l = tk.Label(pageframe, text="Palette:")
		l.grid(row=row, column=0, sticky=tk.W)

		self.showpalette = tkExtra.Combobox(pageframe, width=10,
					command=self.check4ShowPalette)
		self.showpalette.grid(row=row, column=1, sticky=tk.EW)
		self.showpalette.fill(Project.GeometryLayerInfo.PALETTENAMES)

		# ---
		pageframe.grid_columnconfigure(1, weight=1)

		return pageframe

	# --------------------------------------------------------------------
	def check4ShowPalette(self, event=None):
		if self._inInit: return

		try: idx = COLORVALUES.index(self.colorvalue.get())
		except: return
		if idx < 4:
			self._apply()
			return

		# Check if palette is visible otherwise added
		palette = self.showpalette.get()

		if not self.layer(palette+'._visible'):
			self.layer[palette] = True
			self.layer[palette+'._visible'] = True
			self.populateListbox()
			self.showPage(self.activepage)
		self._apply()

	# --------------------------------------------------------------------
	# Background Image Frame
	# --------------------------------------------------------------------
	def imageFrame(self):
		pageframe = tk.Frame(self.mainframe)

		self.imagePrompt = tk.BooleanVar(self)

		# ---
		row = 0
		tk.Label(pageframe, text='Image:').grid(
					row=row, column=0, sticky=tk.W)

		self.imageFilename = ''
		self.imageFile = tkExtra.FileEntry(pageframe,
					title="Load image file",
					filetypes=[
						("All images", ("*.png", "*.gif", "*.jpg", "*.jpeg")),
						("PNG - Portable Network Graphics", "*.png"),
						("GIF - Graphics Interchange Format", "*.gif"),
						("JPG - Joint Photographic Experts Group", ("*.jpg","*.jpeg")),
						("All","*")],
					width=5,
					foreground="DarkBlue", background="White")
		self.imageFile.bind("<<FileChanged>>", self.loadImage)
		tkExtra.Balloon.set(self.imageFile, "Select background image")
		self.imageFile.grid(row=row, column=1, sticky=tk.EW)

		pageframe.grid_columnconfigure(1, weight=1)

		# ---
		if CalibrateImage.Image is None:
		# XXX FIXME give a message
			state = tk.DISABLED
		else:
			state = tk.NORMAL

		# ---
		row += 1
		b = tk.Button(pageframe, text='Calibrate',
					image=icons.get('edit'),
					compound=tk.LEFT,
					command=self.calibrateImage,
					state=state)
		b.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Calibrate image. Requires PIL package")

		# ---
		row += 1
		tk.Label(pageframe, text='Transparency:').grid(
			row=row, column=0, sticky=tk.W)
		self.imageAlpha = tk.Scale(pageframe, from_=0, to_=255,
					orient=tk.HORIZONTAL, showvalue=0,
					command=self.applyLate)
		tkExtra.Balloon.set(self.imageAlpha,
			"Set the transparency level of the superimposed plot")
		self.imageAlpha.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Color Adjust:').grid(
					row=row, column=0, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text='Black:').grid(
					row=row, column=0, sticky=tk.W)

		self.blackImageColor = tk.Button(pageframe,text='  ',
					background='#000000', activebackground='#000000',
					command=self.chooseBlackColor)
		tkExtra.Balloon.set(self.blackImageColor,
			"Choose color for black pixels")
		self.blackImageColor.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='White:').grid(
					row=row, column=0, sticky=tk.W)

		self.whiteImageColor = tk.Button(pageframe,text='  ',
					background='#FFFFFF', activebackground='#FFFFFF',
					command=self.chooseWhiteColor)
		tkExtra.Balloon.set(self.whiteImageColor,
			"Choose color for black pixels")
		self.whiteImageColor.grid(row=row, column=1, sticky=tk.EW)

		pageframe.grid_columnconfigure(1, weight=1)

		# ---
		row += 1
		tk.Label(pageframe, text='Prompt draw', foreground='DarkRed').grid(
					row=row, column=0, sticky=tk.W)

		b = tk.Checkbutton(pageframe, text='',
					variable=self.imagePrompt,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Default (Off) is to draw image only when the display is inactive. "
			"Enable to draw the image immediately, "
			"BUT slowing down the refreshing")
		b.grid(row=row, column=1, sticky=tk.W)
		return pageframe

	# --------------------------------------------------------------------
	def loadImage(self, event=None):
		self.setImagefile(self.imageFile.filename())

	# --------------------------------------------------------------------
	def setImagefile(self, filename):
		self.imageFilename = filename
		self.imageFile.set(self.project.relpath(filename))

	# --------------------------------------------------------------------
	def calibrateImage(self):
		if self.imageFilename:
			dlg = CalibrateImage.Calibrate(self)
			dlg.loadImage(self.imageFilename)

			# prepare markers
			markers = getMarkers(self.layer)

			dlg.setMarkers(markers)

			if dlg.show():
				# Calibration changed save to input and flair
				markers = dlg.getMarkers()

				# Save to layers and create points out of the markers
				points = self.project.input['point'][:]
				if points: pos = points[-1].pos()+1
				else: pos = self.project.input.bestPosition('point')

				self.layer["Image.marker"] = len(markers)

				undoinfo = [self.flair.refreshUndo(), self.flair.renumberUndo()]
				for i,mark in enumerate(markers):
					self.layer[f"Image.marker.{i}"] = " ".join(map(str,mark))
					xi,yi,x,y,z = mark
					for card in points:
						if abs(card.numWhat(1)-x)<1e-10 and \
						   abs(card.numWhat(2)-y)<1e-10 and \
						   abs(card.numWhat(3)-z)<1e-10:
							break
					else:
						# Add a new point
						j=0
						while True:	# Find unique name
							j += 1
							name = f"marker{j}"
							for c in points:
								if c.name()==name: break
							else: break
						card = Card('point',[name,x,y,z])
						card[Card.SELECT] = Global.BIT.VISIBLE
						undoinfo.append(self.flair.addCardUndo(card, pos, False))
						#self.project.input.add(card, pos, True)
						points.append(card)
						#self.project.inputModified()
						pos += 1
				undoinfo.append(self.flair.renumberUndo())
				undoinfo.append(self.flair.refreshUndo())
				self.flair.addUndo(undoinfo,"Add Image markers")
				self.flair.inputModified()
				self._apply()

	# --------------------------------------------------------------------
	def setImageBlack(self, col):
		if isinstance(col,int):
			color = f"#{col:06X}"
		else:
			color = col

		if self.blackImageColor['background'] != color:
			self.blackImageColor['background'] = color
			self.blackImageColor['activebackground'] = color
			self._apply()

	# --------------------------------------------------------------------
	def setImageWhite(self, col):
		if isinstance(col,int):
			color = f"#{col:06X}"
		else:
			color = col
			self.whiteImageColor['background'] = color
			self.whiteImageColor['activebackground'] = color
			self._apply()

	# --------------------------------------------------------------------
	def chooseBlackColor(self):
		try:
			rgb, color = askcolor(title='Select Black color',
				parent=self,
				initialcolor=self.blackImageColor['background'])
			self.setImageBlack(color)
		except tk.TclError:
			pass

	# --------------------------------------------------------------------
	def chooseWhiteColor(self):
		try:
			rgb, color = askcolor(title='Select White color',
				parent=self,
				initialcolor=self.whiteImageColor['background'])
			self.setImageWhite(color)
		except tk.TclError:
			pass

	# --------------------------------------------------------------------
	# Usrdump Frame
	# --------------------------------------------------------------------
	def userdumpFrame(self):
		pageframe = tk.Frame(self.mainframe)

		# ---
		row = 0
		tk.Label(pageframe, text='Userdump file:').grid(
				row=row, column=0, sticky=tk.W)

		self.userdumpFilename = ''
		self.userdumpFile = tkExtra.FileEntry(pageframe,
					title="Load user dump file",
					filetypes=[
						("Fluka USERDUMP files",("*dump*","*draw*")),
						("Fluka fort files",("*_ftn.*","*_fort.*")),
						("All","*")],
					width=5,
					foreground="DarkBlue", background="White")
		self.userdumpFile.bind("<<FileChanged>>", self.loadUserdump)
		tkExtra.Balloon.set(self.userdumpFile, "Select userdump file")
		self.userdumpFile.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Start:').grid(
				row=row, column=0, sticky=tk.W)

		self.userdumpStart = tk.IntVar()
		b = tk.Spinbox(pageframe,
					text = self.userdumpStart,
					from_=1,
					to=10000000,
					command=self.applyLate,
					background='white')
		tkExtra.Balloon.set(b, "Starting event for userdump")
		b.bind("<Return>",   self.apply)
		b.bind("<KP_Enter>", self.apply)
		b.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='# Events:').grid(
				row=row, column=0, sticky=tk.W)

		self.userdumpEvents = tk.IntVar()
		b = tk.Spinbox(pageframe,
					text = self.userdumpEvents,
					from_=0,
					to=10000,
					command=self.applyLate,
					background='white')
		tkExtra.Balloon.set(b, "Userdump events to load")
		b.bind("<Return>",   self.apply)
		b.bind("<KP_Enter>", self.apply)
		b.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Transparency:').grid(
			row=row, column=0, sticky=tk.W)
		self.userdumpAlpha = tk.Scale(pageframe, from_=0, to_=255,
					orient=tk.HORIZONTAL, showvalue=0,
					command=self.applyLate)
		tkExtra.Balloon.set(self.userdumpAlpha,
				"Set the transparency level of the superimposed tracks "
				"in 3D when hidden by geometry")
		self.userdumpAlpha.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		self.userdumpList = tkExtra.MultiListbox(pageframe,
					(("  ",        3, None),
					 ("Particle", 10, None),
					 ("Emin",      5, None),
					 ("Emax",      5, None),
					 ("Color",     8, None)),
					 height=5,
					 background="White")
		self.userdumpList.bindList("<Button-1>",        self.userdumpClick)
		self.userdumpList.bindList("<ButtonRelease-1>", self.userdumpRelease)
		self.userdumpList.bindList("<Key-space>",       self.userdumpToggle)
		self.userdumpList.grid(row=row, column=0, columnspan=2, sticky=tk.NSEW)
		self.userdumpList.setPopupMenu([
				("Show", 0, self.userdumpShow, icons.get("light_on")),
				("Hide", 0, self.userdumpHide, icons.get("light_off"))])
		pageframe.columnconfigure(1,weight=1)
		pageframe.rowconfigure(row,weight=1)

		for particle in FlukaParticle.list:
			if not particle or particle[0]=='@': continue
			self.userdumpList.insert(tk.END, (tkFlair._OFF, particle, "", "", ""))

		return pageframe

	# --------------------------------------------------------------------
	def loadUserdump(self, event=None):
		self.setUserdumpFile(self.userdumpFile.filename())
		self.apply()

	# --------------------------------------------------------------------
	def setUserdumpFile(self, filename):
		self.userdumpFilename = filename
		self.userdumpFile.set(self.project.relpath(filename))

	# --------------------------------------------------------------------
	def userdumpClick(self, event):
		y = event.widget.nearest(event.y)
		if event.widget is self.userdumpList.listbox(0):
			if event.widget.selection_includes(y):
				sel = event.widget.curselection()
				if len(sel)>1:
					status = tkFlair._ON \
							if event.widget.get(y)==tkFlair._OFF \
							else tkFlair._OFF
					for i in sel:
						event.widget.set(i, status)
				self.applyLate()
				return "break"

	# --------------------------------------------------------------------
	def userdumpRelease(self, event):
		y = event.widget.nearest(event.y)
		listid = self.userdumpList.listboxIndex(event.widget)
		if listid == 0:
			if len(event.widget.curselection())==1:
				status = tkFlair._ON \
						if event.widget.get(y)==tkFlair._OFF \
						else tkFlair._OFF
				event.widget.set(y, status)
				self.applyLate()
			return "break"
		elif listid in (2,3):
			tkExtra.InPlaceFloat(event.widget, y)
			self.applyLate()
		elif listid == 4:
			edit = tkExtra.InPlaceColor(event.widget, y)
			if edit:
				event.widget.itemconfig(y, background=edit.value)
				self.applyLate()

	# --------------------------------------------------------------------
	def userdumpToggle(self, event):
		status = tkFlair._ON if self.userdumpList.listbox(0).get(tk.ACTIVE)==tkFlair._OFF \
				else tkFlair._OFF
		for i in self.userdumpList.curselection():
			self.userdumpList.listbox(0).set(i, status)
		self.applyLate()

	# --------------------------------------------------------------------
	def userdumpShow(self, event=None):
		sel = self.userdumpList.curselection()
		if not sel:
			for i in range(self.userdumpList.size()):
				self.userdumpList.listbox(0).set(i,tkFlair._ON)
		else:
			for i in sel:
				self.userdumpList.listbox(0).set(i,tkFlair._ON)
		self.applyLate()

	# --------------------------------------------------------------------
	def userdumpHide(self, event=None):
		sel = self.userdumpList.curselection()
		if not sel:
			for i in range(self.userdumpList.size()):
				self.userdumpList.listbox(0).set(i,tkFlair._OFF)
		else:
			for i in sel:
				self.userdumpList.listbox(0).set(i,tkFlair._OFF)
		self.applyLate()

	# --------------------------------------------------------------------
	# Usrbin Frame
	# --------------------------------------------------------------------
	def usrbinFrame(self):
		pageframe = tk.Frame(self.mainframe)

		self.usrbinInput  = tk.BooleanVar(self)
		self.usrbinErr    = tk.BooleanVar(self)
		self.usrbinPrompt = tk.BooleanVar(self)

		# ---
		row = 0
		tk.Label(pageframe, text="From Input:").grid(
					row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe,
				variable=self.usrbinInput,
				command=self.usrbinInputChanged,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Unselected load a USRBIN file with data to display. "
			"Select to display a USRBINs grid from Input as checker.")
		b.grid(row=row, column=1, columnspan=2, sticky=tk.W)

		# ---
		row += 1

		self.usrbinLabel = tk.Label(pageframe, text='File:')
		self.usrbinLabel.grid(row=row, column=0, sticky=tk.W)

		self.usrbinFile = tkExtra.FileEntry(pageframe,
					title="Load USRBIN file",
					initialdir=self.project.dir(),
					filetypes=[
						("Fluka USRBIN files",("*usrbin*","*.bnn")),
						("Fluka fort files",("*_ftn.*","*_fort.*")),
						("All","*")],
					width=5,
					foreground="DarkBlue", background="White")
		self.usrbinFile.bind("<<FileChanged>>", self.loadUsrbin)
		tkExtra.Balloon.set(self.usrbinFile, 'Load usrbin file')
		self.usrbinFile.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Detector:').grid(
				row=row, column=0, sticky=tk.W)

		self.usrbinDet = tkExtra.Combobox(pageframe,
				width=10,
				command=self.usrbinDetChanged)
		tkExtra.Balloon.set(self.usrbinDet,
			"Usrbin detector to load")
		self.usrbinDet.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		self.usrbinErrLabel = tk.Label(pageframe, text='Show Error:')
		self.usrbinErrLabel.grid(row=row, column=0, sticky=tk.W)
		self.usrbinErrCheck = tk.Checkbutton(pageframe, text='',
					variable=self.usrbinErr,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(self.usrbinErrLabel, "Show Usrbin errors instead of data")
		tkExtra.Balloon.set(self.usrbinErrCheck, "Show Usrbin errors instead of data")
		self.usrbinErrCheck.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		self.usrbinNormLabel = tk.Label(pageframe, text='Norm:')
		self.usrbinNormLabel.grid(row=row, column=0, sticky=tk.W)
		self.usrbinNorm = tk.Entry(pageframe, background='white')
		self.usrbinNorm.bind('<Return>',   self.apply)
		self.usrbinNorm.bind('<KP_Enter>', self.apply)
		tkExtra.Balloon.set(self.usrbinNorm,
			"Usrbin normalisation. Default=1.0")
		self.usrbinNorm.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		self.usrbinXofs = tk.Label(pageframe, text='X-offset:')
		self.usrbinXofs.grid(row=row, column=0, sticky=tk.W)
		self.usrbinXofs = tkExtra.FloatEntry(pageframe, background='white')
		self.usrbinXofs.bind('<Return>',   self.apply)
		self.usrbinXofs.bind('<KP_Enter>', self.apply)
		tkExtra.Balloon.set(self.usrbinXofs,
			"Usrbin X offset. Default=0.0")
		self.usrbinXofs.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		self.usrbinYofs = tk.Label(pageframe, text='Y-offset:')
		self.usrbinYofs.grid(row=row, column=0, sticky=tk.W)
		self.usrbinYofs = tkExtra.FloatEntry(pageframe, background='white')
		self.usrbinYofs.bind('<Return>',   self.apply)
		self.usrbinYofs.bind('<KP_Enter>', self.apply)
		tkExtra.Balloon.set(self.usrbinYofs,
			"Usrbin Y offset. Default=0.0")
		self.usrbinYofs.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		self.usrbinZofs = tk.Label(pageframe, text='Z-offset:')
		self.usrbinZofs.grid(row=row, column=0, sticky=tk.W)
		self.usrbinZofs = tkExtra.FloatEntry(pageframe, background='white')
		self.usrbinZofs.bind('<Return>',   self.apply)
		self.usrbinZofs.bind('<KP_Enter>', self.apply)
		tkExtra.Balloon.set(self.usrbinZofs,
			"Usrbin Z offset. Default=0.0")
		self.usrbinZofs.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Rotdefi:').grid(
				row=row, column=0, sticky=tk.W)

		self.usrbinRotdefi = tkExtra.Combobox(pageframe, width=10,
					command=self.applyLate)
		tkExtra.Balloon.set(self.usrbinRotdefi,
			"Rotdefi to be applied to USRBIN")
		self.usrbinRotdefi.grid(row=row, column=1, columnspan=2, sticky=tk.EW)
		# ---
		row += 1
		tk.Label(pageframe, text='Rotdefi2:').grid(
				row=row, column=0, sticky=tk.W)

		self.usrbinRotdefi2 = tkExtra.Combobox(pageframe, width=10,
					command=self.applyLate)
		tkExtra.Balloon.set(self.usrbinRotdefi2,
			"Second Rotdefi to be applied to USRBIN")
		self.usrbinRotdefi2.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Transparency:').grid(
			row=row, column=0, sticky=tk.W)
		self.usrbinAlpha = tk.Scale(pageframe, from_=0, to_=255,
					orient=tk.HORIZONTAL,
					command=self.applyLate,
					showvalue=0)
		tkExtra.Balloon.set(self.usrbinAlpha,
			"Set the transparency level of the superimposed usrbin plot")
		self.usrbinAlpha.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Prompt draw', foreground='DarkRed').grid(
					row=row, column=0, sticky=tk.W)

		b = tk.Checkbutton(pageframe, text='',
					variable=self.usrbinPrompt,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b, "Show promptly usrbins. Can slow down refreshing")
		b.grid(row=row, column=1, columnspan=2, sticky=tk.W)

		pageframe.grid_columnconfigure(1, weight=1)
		pageframe.columnconfigure(1,weight=1)

		# ---
		row += 1
		l = tk.Label(pageframe, text='Palette:')
		l.grid(row=row, column=0, sticky=tk.W)

		self.usrbinPalette = tkExtra.Combobox(pageframe, width=10,
				command=self.usrbinPaletteChange)
		self.usrbinPalette.grid(row=row, column=1, columnspan=2, sticky=tk.EW)
		self.usrbinPalette.fill(Project.GeometryLayerInfo.PALETTENAMES)

		return pageframe

	# --------------------------------------------------------------------
	def usrbinPaletteChange(self, event=None):
		# Check if palette is visible otherwise added
		palette = self.usrbinPalette.get()
		self.layer[self.entry('palette')] = palette

		if not self.layer(palette+'._visible'):
			self.layer[palette] = True
			self.layer[palette+'._visible'] = True
			self.populateListbox()
			self.showPage(self.activepage)
		self._apply()

	# --------------------------------------------------------------------
	# Initialize usrbin values set to fields
	# --------------------------------------------------------------------
	def usrbinInit(self):
		if self.activepage is None or \
		   not self.activepage.startswith('Usrbin'): return
		oldInit = self._inInit
		self._inInit = True
		self.usrbin  = Data.Usrbin()
		self.usrbinFile.initialdir = self.project.dir(),
		self._usrbinFile = ''
		self.usrbinInput.set(     self.layer(self.entry('input')))
		self.usrbinInputChanged()
		self.setUsrbinFile(       self.layer(self.entry('file')))
		self.setUsrbinDet(        self.layer(self.entry('det')))
		self.usrbinNorm.delete(0, tk.END)
		self.usrbinNorm.insert(0, self.layer(str(self.entry('norm'))))
		self.usrbinXofs.set(      self.layer(str(self.entry('xofs'))))
		self.usrbinYofs.set(      self.layer(str(self.entry('yofs'))))
		self.usrbinZofs.set(      self.layer(str(self.entry('zofs'))))
		self.usrbinRotdefi.clear()
		self.usrbinRotdefi.fill(self.project.input.rotationList(True))
		self.usrbinRotdefi.set(   self.layer(self.entry('rotdefi')))
		self.usrbinRotdefi2.clear()
		self.usrbinRotdefi2.fill(self.project.input.rotationList(True))
		self.usrbinRotdefi2.set(  self.layer(self.entry('rotdefi2')))
		self.usrbinAlpha.set(     self.layer(self.entry('alpha')))
		self.usrbinPalette.set(   self.layer(self.entry('palette')))
		self.usrbinErr.set(       self.layer(self.entry('showerror')))
		self.usrbinPrompt.set(    self.layer("Usrbin.prompt"))
		self._inInit = oldInit

	# --------------------------------------------------------------------
	# usrbinUpdate: Copies the current values on the interface to the
	# project data (only updates current usrbin)
	# --------------------------------------------------------------------
	def usrbinUpdate(self):
		if not self.activepage.startswith('Usrbin'): return

		self.layer[self.entry('input')] = self.usrbinInput.get()
		try:
			self.layer[self.entry('file')] = self._usrbinFile
		except OSError:
			pass

		try:    det = int(self.usrbinDet.get().split()[0])
		except: det = 1
		self.layer[self.entry('det')] = det

		try:    self.layer[self.entry('norm')] = self.usrbinNorm.get()
		except: self.layer[self.entry('norm')] = 1.0
		try:    self.layer[self.entry('xofs')] = self.usrbinXofs.get()
		except: self.layer[self.entry('xofs')] = ""
		try:    self.layer[self.entry('yofs')] = self.usrbinYofs.get()
		except: self.layer[self.entry('yofs')] = ""
		try:    self.layer[self.entry('zofs')] = self.usrbinZofs.get()
		except: self.layer[self.entry('zofs')] = ""
		self.layer[self.entry('rotdefi')]      = self.usrbinRotdefi.get()
		self.layer[self.entry('rotdefi2')]     = self.usrbinRotdefi2.get()
		self.layer[self.entry('alpha')]        = self.usrbinAlpha.get()
		self.layer[self.entry('palette')]      = self.usrbinPalette.get()
		self.layer[self.entry('showerror')]    = self.usrbinErr.get()
		self.layer["Usrbin.prompt"]    = self.usrbinPrompt.get()

	# --------------------------------------------------------------------
	def usrbinInputChanged(self, event=None):
		#self.setUsrbinFile('')
		self.populateUsrbinDet()
		state = tk.DISABLED if self.usrbinInput.get() else tk.NORMAL
		self.usrbinLabel.config(state=state)
		self.usrbinFile.config(state=state)
		self.usrbinErrLabel.config(state=state)
		self.usrbinErrCheck.config(state=state)
		self.usrbinNormLabel.config(state=state)
		self.usrbinNorm.config(state=state)
		self.usrbinXofs.config(state=state)
		self.usrbinYofs.config(state=state)
		self.usrbinZofs.config(state=state)

	# --------------------------------------------------------------------
	def populateUsrbinDet(self):
		oldInit = self._inInit
		self._inInit = True
		self.usrbinDet.clear()

		if self.usrbinInput.get():
			self.usrbinDet.delete(0,tk.END)
			# Populate with input usrbins + EVENTBIN
			for i, card in enumerate(self.project.input.cardsSorted("USRBIN",-1).values(),1):
				self.usrbinDet.insert(tk.END,
					f"{i} {card.name()} [{abs(card.intWhat(3)):02d}]")
		else:
			# --- Load USRBIN ---
			if self._usrbinFile:
				say("Loading:", self._usrbinFile)
				try:
					self.usrbin.readHeader(self._usrbinFile)
				except Exception as err:
					# FIXME should display a nice message
					# Displaying an error before the parent (from the __init__)
					# is displayed makes funny things!!!
					self.flair.notify("Error loading USRBIN",
						f"Error loading USRBIN file: {err}",
						tkFlair.NOTIFY_ERROR)
					say(f"ERROR loading USRBIN file: {err}",)
					self._inInit = oldInit
					return

				# Populate the detector menu
				self.usrbinDet.delete(0,tk.END)
				for i,det in enumerate(self.usrbin.detector):
					self.usrbinDet.insert(tk.END, f"{i+1} {det.name}")
		self._inInit = oldInit

	# ----------------------------------------------------------------------
	def usrbinFileFocusOut(self, event=None):
		if self._usrbinFile == self.usrbinFile.get(): return
		self.setUsrbinFile(self.usrbinFile.get())
		self._apply()

	# --------------------------------------------------------------------
	def loadUsrbin(self, event=None):
		self.setUsrbinFile(self.usrbinFile.filename())
		self.apply()

	# --------------------------------------------------------------------
	def setUsrbinFile(self, filename):
		self.usrbinFile.initialdir = self.project.dir()
		self._usrbinFile = self.project.relpath(filename)
		self.usrbinFile.set(self._usrbinFile)
		self.populateUsrbinDet()

	# --------------------------------------------------------------------
	def setUsrbinDet(self, det):
		try: det = int(det)
		except: det = 1

		oldInit = self._inInit
		self._inInit = True
		if self.usrbinInput.get():
			try:
				card = list(self.project.input.cardsSorted('USRBIN',-1).values())[det-1]
				self.usrbinDet.set(f"{det} {card.name()} [{abs(card.intWhat(3)):02d}]")
			except IndexError:
				self.usrbinDet.set(f"{det} ? [?]")
		else:
			if 0 < det <= len(self.usrbin.detector):
				detector = self.usrbin.detector[det-1]
				newdet = f"{det} {detector.name}"
				if self.usrbinDet.get() != newdet:
					# FIXME self._apply()
					self.usrbinDet.set(newdet)
			else:
				if self.usrbinDet.get():
					# FIXME self._apply()
					self.usrbinDet.clear()
		self._inInit = oldInit

	# --------------------------------------------------------------------
	# usrbin detector changed, check if there is rotation associated with it
	# --------------------------------------------------------------------
	def usrbinDetChanged(self):
		if self.usrbinInput.get():
			match = re.match(r'(\d+) (.+) \[(\d+)\]', self.usrbinDet.get())
		else:
			match = re.match(r'(\d+) (.+)', self.usrbinDet.get())

		oldInit = self._inInit
		#print("\nSET usrbinDetChanged", self._inInit)
		self._inInit = True
		if match:
			idx  = int(match.group(1))
			name = match.group(2)

			# Check ROTPRBIN
			usrbins = list(self.project.input.cardsSorted('USRBIN',-1).values())
			names   = [x.name() for x in usrbins]
			self.usrbinRotdefi.set("")
			self.usrbinRotdefi2.set("")
			for card in self.project.input['ROTPRBIN']:
				if card.ignore(): continue
				from_ = card.what(4)
				to_   = card.what(5)
				#step  = card.intWhat(6)

				try:    ifrom = names.index(from_)
				except: ifrom = -1
				try:    ito = names.index(to_)
				except: ito = ifrom

				try:
					if ifrom <= names.index(name) <= ito:
						self.usrbinRotdefi.set(card.what(2))
						rot2 = card.name()
						if rot2 != "RHO-FACT":
							self.usrbinRotdefi2.set(rot2)
						break
				except:
					pass

			# Check USRBIN in input (maybe)
			if self.usrbinInput.get():
				self.usrbinXofs.config(state=tk.NORMAL)
				self.usrbinYofs.config(state=tk.NORMAL)
				self.usrbinZofs.config(state=tk.NORMAL)
				self.usrbinXofs.set("")
				self.usrbinYofs.set("")
				try:
					card = usrbins[idx-1]
					if card.intWhat(1) in (1,7,11,17):
						self.usrbinXofs.set(card.numWhat(8))
						self.usrbinYofs.set(card.numWhat(5))
				except:
					pass
				self.usrbinXofs.config(state=tk.DISABLED)
				self.usrbinYofs.config(state=tk.DISABLED)
				self.usrbinZofs.config(state=tk.DISABLED)
		self._inInit = oldInit
		#print("RESET usrbinDetChanged", self._inInit)
		self._apply()

	# --------------------------------------------------------------------
	# 3D ray tracing frame
	# --------------------------------------------------------------------
	def d3Frame(self):
		pageframe = tk.Frame(self.mainframe)

		self.deflights       = tk.BooleanVar(self)
		self.shadows         = tk.BooleanVar(self)
		self.edgedetect      = tk.BooleanVar(self)
		self.skipblack       = tk.StringVar(self)
		self.light           = tk.IntVar(self)
		self.clipnegative    = tk.BooleanVar(self)
		self.clipnegative1   = tk.BooleanVar(self)
		self.clipnegative2   = tk.BooleanVar(self)
		self.usrbinastexture = tk.BooleanVar(self)
		self.usrbinastexture.set(True)

		# ---
		row = 0
		tk.Label(pageframe, text='Projection:').grid(
				row=row, column=0, sticky=tk.W)
		self.projection = tkExtra.Combobox(pageframe,
					width=12,
					command=self.applyLate)
		self.projection.fill(_PROJECTIONS)
		tkExtra.Balloon.set(self.projection,
			"Orthographic, Perspective or Combo=Perspective starting "
			"from the Orthographic/Viewing plane")
		self.projection.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		self._fovLabel = tk.Label(pageframe, text='FOV')
		self._fovLabel.grid(row=row, column=0, sticky=tk.W)
		self.fov = tk.Scale(pageframe, from_=0.1, to_=160.0,  # FIXME For fishEye 360.0,
					resolution=0.1,
					orient=tk.HORIZONTAL, showvalue=0,
					command=self.applyLate)
		self.fov.grid(row=row, column=1, columnspan=2, sticky=tk.EW)
		tkExtra.Balloon.set(self.fov,
			"Camera Field of View (Angle in U). "
			"Enabled only when perspective or combo mode is enabled")

		# ---
		row += 1
		tk.Label(pageframe, text='Def. Lights:').grid(
					row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe, text='',
					variable=self.deflights,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b, "Use default (predefined) lights")
		b.grid(row=row, column=1, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text="Ambient Light").grid(
			row=row, column=0, sticky=tk.W+tk.S)
		self.ambient = tk.Scale(pageframe,
					from_=0, to_=255,
					orient=tk.HORIZONTAL,
					command=self.applyLate,
					showvalue=1)
		tkExtra.Balloon.set(self.ambient,
			"Set the level of ambient background intensity")
		self.ambient.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Antialias:", foreground="DarkRed").grid(
			row=row, column=0, sticky=tk.W)
		self.antialias = tk.Scale(pageframe,
					from_=1, to_=4,
					orient=tk.HORIZONTAL,
					command=self.applyLate,
					showvalue=0)
		self.antialias.grid(row=row, column=1, columnspan=2, sticky=tk.EW)
		tkExtra.Balloon.set(self.antialias,
			"Enable anti-aliasing by super sampling for \n"
			"each pixel to enhance image quality.\n"
			"WARNING Increases exponentially the rendering time.")

		# ---
		row += 1
		tk.Label(pageframe, text="Quality:", foreground="DarkRed").grid(
			row=row, column=0, sticky=tk.W)
		self.reflections = tk.Scale(pageframe,
					from_=0, to_=4,
					orient=tk.HORIZONTAL,
					command=self.applyLate,
					showvalue=0)
		self.reflections.grid(row=row, column=1, columnspan=2, sticky=tk.EW)
		tkExtra.Balloon.set(self.reflections,
			"Enable higher quality effects, like reflections, "
			"refractions etc... from objects\n"
			"WARNING Increases the rendering time.")

		# ---
		row += 1
		tk.Label(pageframe, text='Shadows:', foreground='DarkRed').grid(
			row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe, text='',
					variable=self.shadows,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Enable casting of shadows from lights\n"
			"WARNING Increases the rendering time.")
		b.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		# FIXME disable when shadows are off
		tk.Label(pageframe, text='Ambient Occlusion:', foreground='DarkRed').grid(
			row=row, column=0, sticky=tk.W)
		self.aoc = tk.Scale(pageframe,
					from_=0, to_=128, resolution=8,
					orient=tk.HORIZONTAL,
					command=self.applyLate,
					showvalue=0)
		tkExtra.Balloon.set(self.aoc,
			"Enable calculation of ambient occlusion for softer shadows\n"
			"The higher the value more samples are used\n"
			"WARNING Increases the rendering time.")
		self.aoc.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Edge Detection:', foreground='DarkRed').grid(
				row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe, text='',
					variable=self.edgedetect,
					onvalue=1, offvalue=0,
					command=self.applyLate,
					anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Draw edges as black lines.\n"
			"WARNING Increases the rendering time")
		b.grid(row=row, column=1, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text='Black Holes:').grid(
				row=row, column=0, sticky=tk.W)
		b = tk.OptionMenu(pageframe, self.skipblack,
				*_SKIPBLACK,
				command=self.applyLate)
		b.configure(pady=0, anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Select how to treat the rays when hitting a black hole material"
			"Stop the ray and treated as transparent."
			"Ignore only the first (external) black hole region."
			"Ignore and treat black holes as vacuum."
			"WARNING Can lead to errors in display.")
		b.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Xray Level').grid(
			row=row, column=0, sticky=tk.W+tk.S)
		self.xray = tk.Scale(pageframe,
					from_=0, to_=255,
					orient=tk.HORIZONTAL,
					command=self.applyLate,
					showvalue=1)
		tkExtra.Balloon.set(self.xray,
			"Automatic transparency for each region")
		self.xray.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Clipped by:').grid(
				row=row, column=0, sticky=tk.W)

		b = tk.Checkbutton(pageframe, text='',
				variable=self.clipnegative,
				onvalue=1, offvalue=0,
				command=self.applyLate,
				anchor=tk.W)
		tkExtra.Balloon.set(b, "Use the inverse of the clipping body")
		b.grid(row=row, column=1, sticky=tk.W)

		self.clipbody = tkExtra.Combobox(pageframe,
					width=9,
					command=self.applyLate)
		tkExtra.Balloon.set(self.clipbody, "Clipping body")
		self.clipbody.grid(row=row, column=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text='Clipped by #2:').grid(
				row=row, column=0, sticky=tk.W)

		b = tk.Checkbutton(pageframe, text='',
				variable=self.clipnegative1,
				onvalue=1, offvalue=0,
				command=self.applyLate,
				anchor=tk.W)
		tkExtra.Balloon.set(b, "Use the inverse of the clipping body")
		b.grid(row=row, column=1, sticky=tk.W)

		self.clipbody1 = tkExtra.Combobox(pageframe,
					width=9,
					command=self.applyLate)
		tkExtra.Balloon.set(self.clipbody1,
			"Second Clipping body")
		self.clipbody1.grid(row=row, column=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Clipped by #3:").grid(
				row=row, column=0, sticky=tk.W)

		b = tk.Checkbutton(pageframe, text='',
				variable=self.clipnegative2,
				onvalue=1, offvalue=0,
				command=self.applyLate,
				anchor=tk.W)
		tkExtra.Balloon.set(b, "Use the inverse of the clipping body")
		b.grid(row=row, column=1, sticky=tk.W)

		self.clipbody2 = tkExtra.Combobox(pageframe,
					width=9,
					command=self.applyLate)
		tkExtra.Balloon.set(self.clipbody2,
			"Third Clipping body ")
		self.clipbody2.grid(row=row, column=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Project body:").grid(
				row=row, column=0, sticky=tk.W)

		self.projectbody = tkExtra.Combobox(pageframe,
					width=9,
					command=self.applyLate)
		tkExtra.Balloon.set(self.projectbody,
			"USRBIN projection body")
		self.projectbody.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Project body #2:").grid(
				row=row, column=0, sticky=tk.W)

		self.projectbody1 = tkExtra.Combobox(pageframe,
					width=9,
					command=self.applyLate)
		tkExtra.Balloon.set(self.projectbody1,
			"USRBIN second projection body")
		self.projectbody1.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Project body #3:").grid(
				row=row, column=0, sticky=tk.W)

		self.projectbody2 = tkExtra.Combobox(pageframe,
					width=9,
					command=self.applyLate)
		tkExtra.Balloon.set(self.projectbody2,
			"USRBIN third projection body")
		self.projectbody2.grid(row=row, column=1, columnspan=2, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Usrbin as texture:").grid(
				row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe, text='',
				variable=self.usrbinastexture,
				onvalue=1, offvalue=0,
				command=self.applyLate,
				anchor=tk.W)
		tkExtra.Balloon.set(b, "Usrbin color mapping as region texture")
		b.grid(row=row, column=1, sticky=tk.W)

		# ---
		row += 1
		tk.LabelFrame(pageframe, text="Light Information").grid(
			row=row, column=0, columnspan=2, sticky=tk.NSEW)

		pageframe.grid_columnconfigure(2, weight=1)

		return pageframe

	# --------------------------------------------------------------------
	def d3Refresh(self):
		# Populate with clipping bodies
		bodies = sorted(self.project.input.cardsSorted(Card.BODY).keys())
		bodies.insert(0,"")
		self.clipbody.fill(bodies)
		self.clipbody1.fill(bodies)
		self.clipbody2.fill(bodies)
		self.projectbody.fill(bodies)
		self.projectbody1.fill(bodies)
		self.projectbody2.fill(bodies)

	# --------------------------------------------------------------------
	# Voxel structures frame
	# --------------------------------------------------------------------
	def voxelFrame(self):
		pageframe = tk.Frame(self.mainframe)

		# ---
		row = 0
		tk.Label(pageframe, text="ROI Colors:").grid(
			row=row, column=0, sticky=tk.W)
		self.roiAlpha = tk.Scale(pageframe, from_=0, to_=255,
					orient=tk.HORIZONTAL,
					command=self.applyLate,
					showvalue=0)
		tkExtra.Balloon.set(self.roiAlpha,
			"Set blending level of roi colors")
		self.roiAlpha.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		self.roiList = tkExtra.MultiListbox(pageframe,
					(("Id",    3, None),
					 ("Name", 16, None)),
					#selectmode=tk.BROWSE,
					height=10,
					#stretch="first",
					background="White")
		tkExtra.Balloon.set(self.roiList, "Select ROIS to be displayed")
		self.roiList.grid(row=row, column=0, columnspan=2, sticky=tk.NSEW)
		self.roiList.bind("<<ListboxSelect>>", self._apply)
		pageframe.grid_rowconfigure(row, weight=1)

		# ---
		row += 1
		b = tk.Button(pageframe, text="Clear", command=self.roiClear)
		b.grid(row=row, column=0, columnspan=2)
		tkExtra.Balloon.set(b, "Clear roi selection")

		pageframe.grid_columnconfigure(1, weight=1)

		return pageframe

	# --------------------------------------------------------------------
	def voxelRefresh(self):
		self.roiList.delete(0,tk.END)
		for card in self.project.input["VOXELS"]:
			if card.ignore() or card[FlukaCard.VOXEL] is None: continue
			voxel = card[FlukaCard.VOXEL]
			break
		else:
			return

		rois = list(map(int,str(self.layer("Voxel.roi")).split()))
		for roi,name in voxel.roiName.items():
			self.roiList.insert(tk.END,(roi, name))
			if roi in rois:
				self.roiList.selection_set(tk.END)

	# --------------------------------------------------------------------
	def voxelUpdate(self):
		rois = []
		for sel in self.roiList.curselection():
			rois.append(str(self.roiList.get(sel)[0]))
		self.layer["Voxel.roi"] = " ".join(rois)

	# --------------------------------------------------------------------
	def roiClear(self, event=None):
		self.roiList.selection_clear(0,tk.END)
		self._apply()

	# --------------------------------------------------------------------
	# Color band frame
	# --------------------------------------------------------------------
	def paletteFrame(self):
		pageframe = tk.Frame(self.mainframe)

		self.cblog      = tk.BooleanVar(self)
		self.cbinv      = tk.BooleanVar(self)
		self.cbmin      = tk.StringVar()
		self.cbmax      = tk.StringVar()
		self.cbsmooth   = tk.BooleanVar(self)
		self.cbcolors   = tk.IntVar()
		self.cpd        = tk.IntVar()
		self.cbalphamin = tk.BooleanVar(self)
		self.cbalphamax = tk.BooleanVar(self)
		self.cb_min_variable = tk.StringVar()

		# ---
		row = 0
		tk.Label(pageframe, text="Label:").grid(
				row=row, column=0, sticky=tk.W)
		self.cblabel = tk.Entry(pageframe, background="White")
		self.cblabel.bind("<Return>",   self.apply)
		self.cblabel.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.cblabel, "Palette label")
		self.cblabel.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Palette:").grid(
				row=row, column=0, sticky=tk.W)

		self.cbpalette = tkExtra.Combobox(pageframe, width=10,
					command=self.applyLate)
		self.cbpalette.fill(sorted(paletteStore.keys()))
		tkExtra.Balloon.set(self.cbpalette,
			"Color palette to use")
		self.cbpalette.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Log:").grid(
				row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe, text='',
				variable=self.cblog,
				onvalue=1, offvalue=0,
				command=self.applyLate,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Logarithmic or linear")
		b.grid(row=row, column=1, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text="Invert:").grid(
				row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe, text='',
				variable=self.cbinv,
				onvalue=1, offvalue=0,
				command=self.applyLate,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Reverse order of colors")
		b.grid(row=row, column=1, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text="Transparent (<Min):").grid(
				row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe, text='',
				variable=self.cbalphamin,
				onvalue=1, offvalue=0,
				command=self.applyLate,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Transparent below the minimum limit")
		b.grid(row=row, column=1, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text="Transparent (>Max):").grid(
				row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe, text='',
				variable=self.cbalphamax,
				onvalue=1, offvalue=0,
				command=self.applyLate,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Transparent above the maximum limit")
		b.grid(row=row, column=1, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text='Minimum:').grid(
				row=row, column=0, sticky=tk.W)
		e = tkExtra.FloatEntry(pageframe,
				textvariable=self.cbmin,
				background="white")
		e.bind("<Return>",   self.apply)
		e.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(e, "Minimum value for color band")
		e.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Maximum:").grid(
				row=row, column=0, sticky=tk.W)
		b = tkExtra.FloatEntry(pageframe,
				textvariable=self.cbmax,
				disabledforeground="Black")
		b.bind("<Key>",      self.cbmaxChange)
		b.bind("<Return>",   self.apply)
		b.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(b,
				"Maximum value for color band")
		b.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(pageframe, text="Smooth:").grid(
				row=row, column=0, sticky=tk.W)
		b = tk.Checkbutton(pageframe, text='',
				variable=self.cbsmooth,
				onvalue=1, offvalue=0,
				command=self.applyLate,
				anchor=tk.W)
		tkExtra.Balloon.set(b, 'Smooth palette')
		b.grid(row=row, column=1, sticky=tk.W)

		# ---
		row += 1
		tk.Label(pageframe, text="Colors:").grid(
				row=row, column=0, sticky=tk.W)
		self.cbcolors.set(30)
		self.cbspin = tk.Spinbox(pageframe, text=self.cbcolors,
					from_=2,
					to=256,
					command=self.applyLate,
					background="LightYellow",width=4)
		self.cbspin.grid(row=row, column=1, sticky=tk.W)
		tkExtra.Balloon.set(self.cbspin, "Total number of colors")

		# ---
		row += 1
		tk.Label(pageframe, text="CPD:").grid(
				row=row, column=0, sticky=tk.W)
		self.cpd_spin = tk.Spinbox(pageframe,
					text=self.cpd,
					from_=0,
					to=50,
					command=self.applyLate,
					background="LightYellow",width=4)
		self.cpd_spin.grid(row=row, column=1, sticky=tk.W)
		tkExtra.Balloon.set(self.cpd_spin,
			"Colors per decay for log scales")

		# set trace sentinels
		self.cbmin.trace(   'w', self.cpdChange)
		self.cblog.trace(   'w', self.cpdChange)
		self.cpd.trace(     'w', self.cpdChange)
		self.cbsmooth.trace('w', self.cpdChange)
		self.cbcolors.trace('w', self.cpdChange)

		pageframe.grid_columnconfigure(1, weight=1)

		return pageframe

	# --------------------------------------------------------------------
	# When the cb Max is manually edited, set the CPD to 0 to avoid
	# any future recalculation
	#-----------------------------------------------------------------------
	def cbmaxChange(self, event=None):
		try:
			float(self.cbmax.get())
			self.cpd.set(0)
		except ValueError:
			pass

	# --------------------------------------------------------------------
	def cpdChange(self, *abc):
		cblog = int(self.cblog.get())

		if self.cbsmooth.get():
			self.cbspin.config(state=tk.DISABLED)
			self.cpd_spin.config(state=tk.DISABLED)
			return
		else:
			self.cbspin.config(state=tk.NORMAL)

		if cblog:
			self.cpd_spin.config(state=tk.NORMAL)
		else:
			self.cpd_spin.config(state=tk.DISABLED)
			return

		try:
			cpd = float(self.cpd.get())
		except:
			cpd = 0.0

		if cpd < 1.0: return

		# Calculate max
		try: cbmin = float(self.cbmin.get())
		except: return
		try: colors = float(self.cbcolors.get())
		except: return

		try:
			self.cbmax.set("%g"%(cbmin * 10.0**(colors/cpd)))
		except (ValueError, OverflowError):
			pass

	# --------------------------------------------------------------------
	def paletteInit(self):
		if self.activepage is None or \
		   not self.activepage.startswith("Palette"): return

		oldInit = self._inInit
		#Global.dump("*-* paletteInit", self._inInit)
		self._inInit = True
		self.cbpalette.set( self.layer(self.entry('palette')))
		self.cblog.set(     self.layer(self.entry('log')))
		self.cbsmooth.set(  self.layer(self.entry('smooth')))
		self.cbcolors.set(  self.layer(self.entry('n')))
		self.cpd.set(       self.layer(self.entry('cpd')))
		self.cbmin.set(     self.layer(self.entry('min')))
		# cbmax must be set as last to be sure that cpd is not changing it
		self.cbmax.set(     self.layer(self.entry('max')))
		self.cbinv.set(     self.layer(self.entry('inv')))
		self.cbalphamin.set(self.layer(self.entry('alphamin')))
		self.cbalphamax.set(self.layer(self.entry('alphamax')))

		self.cblabel.delete(0,tk.END)
		self.cblabel.insert(0,self.layer(self.entry('label')))

		self._inInit = oldInit
		#Global.dump(">>> paletteInit RESET paletteInit", self._inInit)

	# --------------------------------------------------------------------
	def paletteUpdate(self):
		if not self.activepage.startswith('Palette'): return

		self.layer[self.entry('palette')]     = self.cbpalette.get()
		self.layer[self.entry('log')]         = self.cblog.get()
		self.layer[self.entry('inv')]         = self.cbinv.get()
		self.layer[self.entry('alphamin')]    = self.cbalphamin.get()
		self.layer[self.entry('alphamax')]    = self.cbalphamax.get()
		self.layer[self.entry('smooth')]      = self.cbsmooth.get()
		try:    self.layer[self.entry('min')] = float(self.cbmin.get())
		except: self.layer[self.entry('min')] = 1E-5
		try:    self.layer[self.entry('max')] = float(self.cbmax.get())
		except: self.layer[self.entry('max')] = 1E5
		try:    self.layer[self.entry('n')]   = int(self.cbcolors.get())
		except: self.layer[self.entry('n')]   = 30
		try:    self.layer[self.entry('cpd')] = int(self.cpd.get())
		except: self.layer[self.entry('cpd')] = 3
		self.layer[self.entry('label')]       = self.cblabel.get()

	# --------------------------------------------------------------------
	# Empty frame
	# --------------------------------------------------------------------
	def emptyFrame(self):
		pageframe = tk.Frame(self.mainframe)
		tk.Label(pageframe, text='Not implemented yet').pack(expand=tk.YES,fill=tk.BOTH)
		return pageframe

	# --------------------------------------------------------------------
	# Initialize and set defaults options
	# --------------------------------------------------------------------
	def init(self):
		Global.dump("*-* GeometryLayersFrame.init")
		self._inInit = True	# Avoid updating

		# Layers
		for i in range(1, self.listbox.size()):
			page, sel = self.listbox.get(i)
			if sel=='': continue
			s = self.listbox.selection_includes(i)
			if self.layer(page):
				sel = tkFlair._ON
			else:
				sel = tkFlair._OFF
			self.listbox.set(i, (page,sel))
			if s: self.listbox.selection_set(i)

		# Global layer
		self.globalLayer.set(self.layer._global)

		# Options
		self.axes.set(        self.layer("Options.axes"))
		self.showErrors.set(  self.layer("Options.errors"))
		self.gridVar.set(     self.layer("Options.grid"))
		ocubelength = int(self.layer("Options.ocube"))
		self.ocubelength.set(abs(ocubelength))
		self.ocube.set(ocubelength>0)
		self.gridlevel.set(   self.layer("Options.gridlevel"))
		self.latticeLevel.set(self.layer("Options.latticelevel"))
		self.crosshair.set(   self.layer("Options.crosshair"))
		self.background.set(  self.layer("Options.textbackground"))
		self.viewport.set(    self.layer("Options.viewport"))
		self.generalFont.set( self.layer("Options.font"))
		self.gridFont.set(    self.layer("Options.gridfont"))
		self.paletteFont.set( self.layer("Options.palettefont"))

		# Show
		self.showvertex.set( self.layer("Show.vertex"))
		self.labels.set(     self.layer("Show.label"))
		self.colorvalue.set( self.layer("Show.color"))
		self.showlattice.set(self.layer("Show.lattice"))
		self.showvoxel.set(  self.layer("Show.voxel"))
		self.showpalette.set(self.layer("Show.palette"))

		# Background image
		self.imageFile.initialdir = self.project.dir()
		self.setImagefile(     self.layer("Image.file"))
		self.imageAlpha.set(   self.layer("Image.alpha"))
		self.setImageBlack(int(self.layer("Image.black")))
		self.setImageWhite(int(self.layer("Image.white")))
		self.imagePrompt.set(  self.layer("Image.prompt"))

		# Userdump
		self.userdumpFile.initialdir = self.project.dir()
		self.setUserdumpFile(   self.layer("Userdump.file"))
		self.userdumpStart.set( self.layer("Userdump.start"))
		self.userdumpEvents.set(self.layer("Userdump.n"))
		self.userdumpAlpha.set( self.layer("Userdump.alpha"))
		for i,name in enumerate(self.userdumpList.listbox(1).get(0,tk.END)):
			on    = tkFlair._ON if self.layer(f"Userdump.{name}") else tkFlair._OFF
			emin  = self.layer(f"Userdump.emin.{name}")
			emax  = self.layer(f"Userdump.emax.{name}")
			color = self.layer(f"Userdump.color.{name}")
			self.userdumpList.set(i, (on,name,emin,emax,color))
			if color!="":
				self.userdumpList.listbox(-1).itemconfig(i, background=color)

		# 3D
		self.projection.set(_PROJECTIONS[self.layer('3D.projection')])
		self.ambient.set(        self.layer('3D.ambient'))
		self.antialias.set(      self.layer('3D.antialias'))
		self.clipbody.set(       self.layer('3D.clip.0'))
		self.clipbody1.set(      self.layer('3D.clip.1'))
		self.clipbody2.set(      self.layer('3D.clip.2'))
		self.clipnegative.set(   self.layer('3D.clipnegative.0'))
		self.clipnegative1.set(  self.layer('3D.clipnegative.1'))
		self.clipnegative2.set(  self.layer('3D.clipnegative.2'))
		self.deflights.set(      self.layer('3D.deflights'))
		self.edgedetect.set(     self.layer('3D.edgedetect'))
		self.fov.set(            self.layer('3D.fov'))
		self.projectbody.set(    self.layer('3D.project.0'))
		self.projectbody1.set(   self.layer('3D.project.1'))
		self.projectbody2.set(   self.layer('3D.project.2'))
		self.reflections.set(    self.layer('3D.reflections'))
		self.shadows.set(        self.layer('3D.shadows'))
		self.aoc.set(            self.layer('3D.aoc'))
		self.skipblack.set(_SKIPBLACK[int(self.layer('3D.skipblack'))])
		self.usrbinastexture.set(self.layer('3D.usrbinastexture'))
		self.xray.set(           self.layer('3D.xray'))

		# Voxel
		self.roiAlpha.set(self.layer('Voxel.roialpha'))
		self.voxelRefresh()

		# Multi init Usrbin/Palette
		self._usrbinFile = ''
		self.usrbinInit()
		self.paletteInit()

		self._inInit = False

	# --------------------------------------------------------------------
	# Get layers information and save it in project
	# --------------------------------------------------------------------
	def get(self):
		Global.dump("*-* GeometryLayersFrame.get", level=2)
		if self.layer is None: return

		# Get the active layers
		for page, sel in self.listbox.get(0,tk.END):
			if sel!='': self.layer[page] = (sel==tkFlair._ON)

		# Update layer
		oldname = self.layer.name
		self.layer.name = self.layerBox.get()
		if oldname != self.layer.name:
			if self.layer._global:
				try: del Project._globalLayers[oldname]
				except KeyError: pass
			try: del self.project.geometry.layers[oldname]
			except KeyError: pass
			self.project.geometry.layers[self.layer.name] = self.layer

		# Global layer or not
		self.layer._global = self.globalLayer.get()

		# Options
		self.layer["Options.axes"]          = self.axes.get()
		self.layer["Options.errors"]        = self.showErrors.get()
		self.layer["Options.grid"]          = self.gridVar.get()
		if self.ocube.get():
			ocubelength = abs(self.ocubelength.get())
		else:
			ocubelength = -abs(self.ocubelength.get())
		self.layer["Options.ocube"]         = ocubelength
		self.layer["Options.gridlevel"]     = self.gridlevel.get()
		self.layer["Options.latticelevel"]  = self.latticeLevel.get()
		self.layer["Options.crosshair"]     = self.crosshair.get()
		self.layer["Options.textbackground"]= self.background.get()
		self.layer["Options.viewport"]      = self.viewport.get()
		self.layer["Options.font"]          = self.generalFont.get()
		self.layer["Options.gridfont"]      = self.gridFont.get()
		self.layer["Options.palettefont"]   = self.paletteFont.get()

		# Show
		self.layer["Show.vertex"]  = self.showvertex.get()
		self.layer["Show.label"]   = self.labels.get()
		self.layer["Show.color"]   = self.colorvalue.get()
		self.layer["Show.lattice"] = self.showlattice.get()
		self.layer["Show.voxel"]   = self.showvoxel.get()
		self.layer["Show.palette"] = self.showpalette.get()

		# Background image
		try:
			self.layer['Image.file']   = self.project.relpath(self.imageFilename)
		except OSError:
			pass
		self.layer["Image.alpha"]  = self.imageAlpha.get()
		self.layer["Image.black"]  = int(self.blackImageColor["background"][1:],16)
		self.layer["Image.white"]  = int(self.whiteImageColor["background"][1:],16)
		self.layer["Image.prompt"] = self.imagePrompt.get()

		# Userdump
		try:
			self.layer["Userdump.file"] = self.project.relpath(self.userdumpFilename)
		except OSError:
			pass
		try:    self.layer["Userdump.start"] = int(self.userdumpStart.get())
		except: self.layer["Userdump.start"] = 1
		try:    self.layer["Userdump.n"] = int(self.userdumpEvents.get())
		except: self.layer["Userdump.n"] = 100
		try:    self.layer["Userdump.alpha"] = int(self.userdumpAlpha.get())
		except: self.layer["Userdump.alpha"] = 0
		for on, name, emin, emax, color in self.userdumpList.get(0,tk.END):
			self.layer[f"Userdump.{name}"] = int(on==tkFlair._ON)
			if emin != "":
				self.layer[f"Userdump.emin.{name}"] = emin
			else:
				try: del self.layer[f"Userdump.emin.{name}"]
				except KeyError: pass

			if emax != "":
				self.layer[f"Userdump.emax.{name}"] = emax
			else:
				try: del self.layer[f"Userdump.emax.{name}"]
				except KeyError: pass

			if color != "":
				self.layer[f"Userdump.color.{name}"] = color
			else:
				try: del self.layer[f"Userdump.color.{name}"]
				except KeyError: pass

		# 3D
		try: self.layer["3D.projection"] = _PROJECTIONS.index(self.projection.get())
		except: self.layer["3D.projection"] = 0
		self.layer["3D.deflights"]       = self.deflights.get()
		try:    self.layer["3D.fov"]     = float(self.fov.get())
		except: self.layer["3D.fov"]     = 60.0
		self.layer["3D.ambient"]         = self.ambient.get()
		self.layer["3D.antialias"]       = self.antialias.get()
		self.layer["3D.clip.0"]          = self.clipbody.get()
		self.layer["3D.clip.1"]          = self.clipbody1.get()
		self.layer["3D.clip.2"]          = self.clipbody2.get()
		self.layer["3D.clipnegative.0"]  = self.clipnegative.get()
		self.layer["3D.clipnegative.1"]  = self.clipnegative1.get()
		self.layer["3D.clipnegative.2"]  = self.clipnegative2.get()
		self.layer["3D.edgedetect"]      = self.edgedetect.get()
		self.layer["3D.project.0"]       = self.projectbody.get()
		self.layer["3D.project.1"]       = self.projectbody1.get()
		self.layer["3D.project.2"]       = self.projectbody2.get()
		self.layer["3D.reflections"]     = self.reflections.get()
		self.layer["3D.shadows"]         = self.shadows.get()
		self.layer["3D.aoc"]             = self.aoc.get()
		self.layer["3D.skipblack"]       = _SKIPBLACK.index(self.skipblack.get())
		self.layer["3D.usrbinastexture"] = self.usrbinastexture.get()
		self.layer["3D.xray"]            = self.xray.get()

		# Voxel
		self.layer["Voxel.roialpha"]     = self.roiAlpha.get()
		self.voxelUpdate()

		# Multi updates Usrbin/Palette
		self.usrbinUpdate()
		self.paletteUpdate()

		# Correct layer
		self.removeSameWithDefaults()
		self.checkSystem()

	# --------------------------------------------------------------------
	# Reset current/active layer
	# --------------------------------------------------------------------
	def reset(self):
		if self.layer is None: return
		self.flair.modified()
		prefix = self.activepage + "."
		# delete all entries starting with the activepage
		todel = [x for x in self.layer.keys()
				if x.startswith(prefix) ]
		for key in todel:
			del self.layer[key]
		#try:	# remove the system flag
		#	del layer["_system"]
		#	del layer["_order"]
		#except KeyError:
		#	pass
		#layer._global = False
		self.init()
		self._apply(force=True)

	# --------------------------------------------------------------------
	# Remove layer options that are the same with the default ones
	# --------------------------------------------------------------------
	def removeSameWithDefaults(self):
		remove = []
		for name,value in self.layer.items():
			if name == '_system': continue
			if "._visible" in name: continue
			if value == Project.GeometryLayerInfo.DEFAULT.get(name):
				remove.append(name)
		for name in remove:
			del self.layer[name]

	# --------------------------------------------------------------------
	# Check if system was changed and save it to local
	# --------------------------------------------------------------------
	def checkSystem(self):
		if not self.layer("_system"): return

		# Find in the global
		default = Project._systemLayers.get(self.layer.name)
		if default is None:
			self.makeLocal()
			return
		else:
			# Compare with the default
			for name,value in self.layer.items():
				# skip options
				if name.startswith("Options"): continue
				defvalue = default(name)
				if defvalue is not None and value != defvalue:
					self.makeLocal()
					return

	# --------------------------------------------------------------------
	# Make a system layer local
	# --------------------------------------------------------------------
	def makeLocal(self):
		# clone layer to project
		self.layer = self.layer.clone()
		try:	# remove the system flag
			del self.layer["_system"]
			del self.layer["_order"]
		except KeyError:
			pass
		self.layer._global = False

		self.project.geometry.layers[self.layer.name] = self.layer
		self.globalLayer.set(False)
		self.globalButton.config(state=tk.NORMAL)

		# restore default layer
		try:
			Project._globalLayers[self.layer.name] = \
				Project._systemLayers[self.layer.name].clone()
		except KeyError:
			pass

	# --------------------------------------------------------------------
	def addLayer(self, event=None):
		self.flair.modified()
		self.get()
		self.layer = None

		# find unique name
		for i in count(1):
			name = 'Layer %02d'%(i)
			if name not in self.project.geometry.layers: break
		self.project.geometry.layers[name] = Project.GeometryLayerInfo(self.project, name)
		self.fillLayers(name)
		self.layerBox._text.select_range(0,tk.END)
		self.layerBox._text.focus_set()

	# --------------------------------------------------------------------
	def cloneLayer(self, event=None):
		self.flair.modified()
		self.get()

		layer = self.layer.clone()
		try:	# remove the system flag
			del layer["_system"]
			del layer["_order"]
		except KeyError:
			pass
		layer._global = False

		# Find unique name
		pat = Project._LAST_NUM.match(layer.name)
		if pat:
			name = pat.group(1)
			n    = int(pat.group(2))+1
		else:
			name = layer.name
			n    = 1
		while True:
			guess = f"{name} {n:02d}"
			if guess not in self.project.geometry.layers:
				layer.name = guess
				break
			n += 1

		self.project.geometry.layers[guess] = layer
		self.fillLayers(layer.name)
		self.layerBox._text.select_range(0,tk.END)
		self.layerBox._text.focus_set()

	# --------------------------------------------------------------------
	def delLayer(self, event=None):
		if self.layer('_system'): return

		if not tkFlair.askyesno("Delete Layer",
				f"Do you want to delete the active layer \'{self.layer.name}\'",
				parent=self.listbox):
			return

		lst = self.project.geometry.layersList()

		if self.layer._global:
			try: del Project._globalLayers[self.layer.name]
			except KeyError: pass

		try: del self.project.geometry.layers[self.layer.name]
		except KeyError: pass
		tkFlair.updateLayers()

		try:
			idx = lst.index(self.layer.name)
			del lst[idx]
			while idx>0 and idx >= len(lst):
				idx -= 1
			name = lst[idx]
			self.layer = self.project.geometry.layers[name]
			self.fillLayers(name)
		except:
			self.layer = None
			self.addLayer()

	# --------------------------------------------------------------------
	def fillLayers(self, layername=None):
		layers = self.project.geometry.layersList()
		self.layerBox.fill(layers)
		if layername is not None and layername in self.project.geometry.layers:
			self.layerBox.set(layername)

	# --------------------------------------------------------------------
	def changeLayer(self, event=None):
		self.setLayer(self.layerBox.get())

	# --------------------------------------------------------------------
	def setLayer(self, layer):
		self.layerBox._text.delete(0, tk.END)
		self.layerBox._text.insert(0, layer)
		try:
			self.layer = self.project.geometry.layers[layer]
		except KeyError:
			pass
		self.populateListbox()
		self.showPage(self.activepage)
		self.init()

	# --------------------------------------------------------------------
	def help(self, event=None):
		Manual.show("","Layers")
