#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

import os
import math
import tkinter as tk

import icons
import tkFlair
import tkExtra
import Project
import bFileDialog

from log import error
from deck import Card
from Global import BIT, ACTION
from postscript import Postscript

#===============================================================================
# ToolBase base class for a floating dialog inside the geometry editor
#===============================================================================
class ToolBase(tk.Frame):
	TITLE_BG = "DarkBlue"
	TITLE_FG = "White"

	TEXT_NO  = "Black"
	TEXT_OK  = "DarkGreen"
	TEXT_ERR = "Red"

	EXPAND   = "\N{BLACK RIGHT-POINTING TRIANGLE}"
	COLLAPSE = "\N{BLACK DOWN-POINTING TRIANGLE}"

	# ----------------------------------------------------------------------
	def __init__(self, master, title="", ok="", *args, **kwargs):
		super().__init__(master, *args, **kwargs)

		self.bind("<Configure>",	self._configure)
		self.master.bind("<Configure>",	self._configure)

		self.action = tk.IntVar()
		self.action.set(ACTION.SELECT)
		self.action.trace("w", self.actionChange)

		frame = tk.Frame(self, bg=ToolBase.TITLE_BG)
		frame.pack(side=tk.TOP, fill=tk.X)
		frame.bind("<B1-Motion>",	self._motion)
		frame.bind("<ButtonRelease-1>",	self._motionEnd)

		self._minimize = tk.Label(frame,
					text=ToolBase.COLLAPSE,
					bg=ToolBase.TITLE_BG,
					fg=ToolBase.TITLE_FG)
		self._minimize.pack(side=tk.LEFT)
		self._minimize.bind("<Button-1>", self._toggle)

		self._title = tk.Label(frame, text=title,
					bg=ToolBase.TITLE_BG,
					fg=ToolBase.TITLE_FG)
		self._title.pack(side=tk.LEFT, fill=tk.X) #, expand=tk.YES)
		self._title.bind("<B1-Motion>",		self._motion)
		self._title.bind("<ButtonRelease-1>",	self._motionEnd)

		self.frame = tk.Frame(self, relief=tk.GROOVE, padx=2, pady=2)
		self.frame.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=2, pady=2)
		self.frame.grid_columnconfigure(1, weight=1)

		self._buttonFrame = tk.Frame(self)
		self._buttonFrame.pack(side=tk.TOP, fill=tk.X)

		if ok:
			self.okButton = self.addButton(ok, "ok", "Perform action", self.ok, padx=10)
			self.okButton.bind("<Return>",   self.ok)
			self.okButton.bind("<KP_Enter>", self.ok)
			self.frame.bind("<Return>",      self.ok)
			self.frame.bind("<KP_Enter>",    self.ok)
		self.cancelButton = self.addButton("Cancel", "x", "Close tool dialog", self.cancel)
		self.cancelButton.bind("<Escape>", self.cancel)
		self.frame.bind("<Escape>", self.cancel)

		b = tk.Checkbutton(self._buttonFrame, image=icons.get("info"),
						command=self.toggleInfo,
						indicatoron=0,
						padx=1, pady=1)
		b.pack(side=tk.LEFT)

		self._info = tk.Label(self, justify=tk.LEFT, relief=tk.SUNKEN)
		self.fillInfo(title)
		self.title = title

		self._sideX = tk.E	# docking side
		self._sideY = None	#    -//-
		self._x = 0		# remember position
		self._y = 50		#    -//-
		self.__motion = None	# initial coordinates during motion

		# selections buttons
		self._viewportBtn = None
		self._regionsBtn  = None
		self._zonesBtn    = None
		self._bodiesBtn   = None
		self._objectsBtn  = None

		# number of items selected (None if button is disabled else number of items)
		self.nregions = None
		self.nzones   = None
		self.nobjects = None
		self.exclude  = []		# exclude from selection counting

	#-----------------------------------------------------------------------
	def project(self):	return self.master.project
	def input(self):	return self.master.project.input
	def geometry(self):	return self.master.geometry()
	def editor(self):	return self.master.editor
	def viewers(self):	return self.master
	def flair(self):	return self.master.flair()
	def hasViewportSelection(self):	return self._viewportBtn is not None

	#-----------------------------------------------------------------------
	def show(self, side=None):
		self.place(relx=0.0, rely=0.0, x=self._x, y=self._y,
				anchor=tk.NW, bordermode="outside")
		if not self.winfo_ismapped():
			self.wait_visibility()

		if side is None:
			pass		# do nothing
		elif side == tk.N:
			self._sideY = tk.N
			self._sideX = None
		elif side == tk.NE:
			self._sideY = tk.N
			self._sideX = tk.E
		elif side == tk.E:
			self._sideY = None
			self._sideX = tk.E
		elif side == tk.SE:
			self._sideY = tk.S
			self._sideX = tk.E
		elif side == tk.S:
			self._sideY = tk.S
			self._sideX = None
		elif side == tk.SW:
			self._sideY = tk.S
			self._sideX = tk.W
		elif side == tk.W:
			self._sideY = None
			self._sideX = tk.W
		elif side == tk.NW:
			self._sideY = tk.N
			self._sideX = tk.W
		else:
			self._sideX = self._sideY = None

		self.placeAt()
		self.updateSelection()

	#-----------------------------------------------------------------------
	def hide(self):
		if not self.winfo_ismapped(): return
		self._x = self.winfo_x()
		self._y = self.winfo_y()
		self.place_forget()

	#-----------------------------------------------------------------------
	def _configure(self, event=None):
		if not self.winfo_ismapped(): return
		if self.__motion is not None: return	# during motion
		self.placeAt()

	#-----------------------------------------------------------------------
	def placeAt(self, x=None, y=None):
		W = self.master.winfo_width()
		H = self.master.winfo_height()
		w = self.winfo_width()
		h = self.winfo_height()

		margin = 5

		# horizontal
		if self._sideX == tk.E:
			x = W-w
		elif self._sideX == tk.W:
			x = 0
		elif x is None:
			x = self.winfo_x()

		if x<margin:
			x = 0
			self._sideX = tk.W
		elif x>=W-w-margin:
			x = W-w
			self._sideX = tk.E
		else:
			self._sideX = None

		# vertical
		if self._sideY == tk.S:
			y = H-h
		elif self._sideY == tk.N:
			y = 0
		elif y is None:
			y = self.winfo_y()

		if y<margin:
			y = 0
			self._sideY = tk.N
		elif y>H-h-margin:
			y = H-h
			self._sideY = tk.S
		else:
			self._sideY = None

		self._x = x
		self._y = y
		self.place(x=x, y=y)

	#-----------------------------------------------------------------------
	def _motion(self, event):
		if self.__motion is None:
			self.__motion = (event.x, event.y)
		x = event.x_root - self.master.winfo_rootx() - self.__motion[0]
		y = event.y_root - self.master.winfo_rooty() - self.__motion[1]
		self._sideX = self._sideY = None
		self.placeAt(x,y)

	#-----------------------------------------------------------------------
	def _motionEnd(self, event):
		self.__motion = None

	#-----------------------------------------------------------------------
	def _toggle(self, event=None):
		if self._minimize["text"] == ToolBase.COLLAPSE:
			self.frame.pack_forget()
			self._buttonFrame.pack_forget()
			self._minimize["text"] = ToolBase.EXPAND
		else:
			self.frame.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
			self._buttonFrame.pack(side=tk.BOTTOM, fill=tk.X)
			self._minimize["text"] = ToolBase.COLLAPSE

	#-----------------------------------------------------------------------
	def cancel(self):
		self.action.set(ACTION.SELECT)
		self.hide()

	#-----------------------------------------------------------------------
	def ok(self, event=None):
		self.hide()

	#-----------------------------------------------------------------------
	def toggleInfo(self):
		if self._info.winfo_ismapped():
			self._info.pack_forget()
		else:
			self._info.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=tk.YES)

	#-----------------------------------------------------------------------
	def fillInfo(self, title=""):
		self._info["text"] = f"{title}:\n\n{self.__doc__}"

	#-----------------------------------------------------------------------
	# Next action triggered by GeometryViewer. It should be based on the
	# current selection
	#-----------------------------------------------------------------------
	def nextAction(self):
		pass

	#-----------------------------------------------------------------------
	def setAction(self, act):
		#print("\n>>> set action",act)
		self.action.set(act)

	#-----------------------------------------------------------------------
	def actionChange(self, *abc):
		self.viewers().setMouseAction(self.action.get())

	#-----------------------------------------------------------------------
	def addButton(self, label, icon, hlp=None, command=None, **kwargs):
		if "padx" not in kwargs: kwargs["padx"] = 1
		if "pady" not in kwargs: kwargs["pady"] = 1
		btn = tk.Button(self._buttonFrame,
				text=label,
				image=icons.get(icon),
				compound=tk.LEFT,
				command=command,
				**kwargs)
		btn.pack(side=tk.RIGHT)
		if hlp is not None: tkExtra.Balloon.set(btn, hlp)
		return btn

	#-----------------------------------------------------------------------
	# Add a label at column 0
	#-----------------------------------------------------------------------
	def addLabel(self, row, label, hlp):
		l = tk.Label(self.frame, text=label)
		l.grid(row=row, column=0, sticky=tk.W)
		tkExtra.Balloon.set(l, hlp)

	#-----------------------------------------------------------------------
	# Add a label with info
	#-----------------------------------------------------------------------
	def addLabelInfo(self, row, label, hlp, **kwargs):
		self.addLabel(row, label, hlp)
		if "fg" not in kwargs or "foreground" not in kwargs:
			kwargs["fg"] = "DarkBlue"
		if "bg" not in kwargs or "background" not in kwargs:
			kwargs["bg"] = "White"
		e = tk.Label(self.frame, **kwargs)
		e.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(e, hlp)
		return e

	#-----------------------------------------------------------------------
	# Add a label with an entry
	#-----------------------------------------------------------------------
	def addLabelEntry(self, row, label, hlp, class_=tk.Entry, **kwargs):
		self.addLabel(row, label, hlp)
		if "fg" not in kwargs or "foreground" not in kwargs:
			kwargs["fg"] = "DarkBlue"
		if "bg" not in kwargs or "background" not in kwargs:
			kwargs["bg"] = "White"
		e = class_(self.frame, **kwargs)
		e.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(e, hlp)
		e.bind("<Return>",   self.ok)
		e.bind("<KP_Enter>", self.ok)
		return e

	#-----------------------------------------------------------------------
	# Add a label with a floating point entry
	#-----------------------------------------------------------------------
	def addFloatEntry(self, row, label, hlp, **kwargs):
		return self.addLabelEntry(row, label, hlp, tkExtra.FloatEntry, **kwargs)

	#-----------------------------------------------------------------------
	# Add a label with an integer point entry
	#-----------------------------------------------------------------------
	def addIntegerEntry(self, row, label, hlp, **kwargs):
		return self.addLabelEntry(row, label, hlp, tkExtra.IntegerEntry, **kwargs)

	#-----------------------------------------------------------------------
	# Add a label with a spinbox
	#-----------------------------------------------------------------------
	def addSpinbox(self, row, label, hlp, from_, to_):
		self.addLabel(row, label, hlp)
		e = tk.Spinbox(self.frame,
				fg="DarkBlue", bg="White",
				from_=from_, to_=to_)
		e.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(e, hlp)
		return e

	#-----------------------------------------------------------------------
	# Add a label with a radiobutton
	#-----------------------------------------------------------------------
	def addRadiobutton(self, row, label, hlp, image, value, clear=None):
		self.addLabel(row, label, hlp)
		btn = tk.Radiobutton(self.frame,
					image=image, compound=tk.LEFT,
					variable=self.action,
					value=value,
					selectcolor="LightGreen",
					indicatoron=0)
		btn.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(btn, hlp)
		if clear:
			b = tk.Button(self.frame,
					text="\N{HEAVY MULTIPLICATION X}",
					padx=1, pady=1,
					command=clear,
					relief=tk.FLAT)
			b.grid(row=row, column=2, sticky=tk.EW)
			tkExtra.Balloon.set(b, f"Clear {label}")
		return btn

	#-----------------------------------------------------------------------
	# Regions selection
	#-----------------------------------------------------------------------
	def addRegionsBtn(self, row, hlp):
		self._regionsBtn = self.addRadiobutton(row,
					"Region",
					hlp,
					icons.get("select"),
					ACTION.SELECT_REGION,
					self.clearRegion)

	#-----------------------------------------------------------------------
	def clearRegion(self):
		self.editor().selectClear(True, Card.REGION)
		self.setAction(ACTION.SELECT_REGION)

	#-----------------------------------------------------------------------
	# Zones selection
	#-----------------------------------------------------------------------
	def addZonesBtn(self, row, hlp):
		self._zonesBtn = self.addRadiobutton(row,
					"Zone",
					hlp,
					icons.get("select"),
					ACTION.SELECT_ZONE,
					self.clearZone)

	#-----------------------------------------------------------------------
	def clearZone(self):
		self.editor().propListClear()
		if self.action.get() != ACTION.SELECT_BODY:
			self.setAction(ACTION.SELECT_ZONE)

	#-----------------------------------------------------------------------
	# Bodies selection
	#-----------------------------------------------------------------------
	def addBodiesBtn(self, row, hlp):
		self._bodiesBtn = self.addRadiobutton(row,
					"Bodies",
					hlp,
					icons.get("select"),
					ACTION.SELECT_BODY,
					self.clearBodies)

	#-----------------------------------------------------------------------
	def clearBodies(self):
		self.editor().selectClear(True, Card.BODY)
		self.setAction(ACTION.SELECT_BODY)

	#-----------------------------------------------------------------------
	# Objects selection
	#-----------------------------------------------------------------------
	def addObjectsBtn(self, row, hlp):
		self._objectsBtn = self.addRadiobutton(row,
					"Objects",
					hlp,
					icons.get("select"),
					ACTION.SELECT_OBJECT,
					self.clearObjects)

	#-----------------------------------------------------------------------
	def clearObjects(self):
		self.editor().selectClear(True, Card.OBJECT)
		self.setAction(ACTION.SELECT_OBJECT)

	#-----------------------------------------------------------------------
	# return spline card
	#-----------------------------------------------------------------------
	def splineCard(self):
		for card in self.input().cardsCache("spline").values():
			if card.ignore(): continue
			if card.get(Card.SELECT,0) & BIT.SELECT:
				return card
		return None

	#-----------------------------------------------------------------------
	# Viewport selection
	#-----------------------------------------------------------------------
	def addViewportBtn(self, row, hlp):
		self._viewportBtn = self.addRadiobutton(row,
					"Viewport",
					hlp,
					icons.get("select"),
					ACTION.SELECT_VIEWPORT,
					self.clearViewport)

	#-----------------------------------------------------------------------
	def clearViewport(self):
		self._viewportBtn.config(text="select", foreground="Black")
		self.setAction(ACTION.SELECT_VIEWPORT)

	#-----------------------------------------------------------------------
	# Update selection
	#-----------------------------------------------------------------------
	def updateSelection(self):
		if self._objectsBtn is not None:
			self.nobjects = self.geometry().selection("o")
			if self.exclude and self.nobjects:
				self.nobjects -= len([x for x in self.viewers().selectedCards()
								if x.tag in self.exclude])
			if self.nobjects==0:
				self._objectsBtn.config(text="select",
						fg=ToolBase.TEXT_NO,
						state=tk.NORMAL)
			else:
				self._objectsBtn.config(text=f"{self.nobjects} selected",
						fg=ToolBase.TEXT_OK,
						state=tk.NORMAL)
		else:
			self.nobjects = None

		if self._bodiesBtn is not None:
			self.nbodies = self.geometry().selection("b")
			if self.nbodies==0:
				self._bodiesBtn.config(text="select",
						fg=ToolBase.TEXT_NO,
						state=tk.NORMAL)
			else:
				self._bodiesBtn.config(text=f"{self.nbodies} selected",
						fg=ToolBase.TEXT_OK,
						state=tk.NORMAL)
		else:
			self.nbodies = None

		if self._zonesBtn is not None:
			self.nzones = self.editor().propListZoneSelected()
			if self.nzones==0:
				self._zonesBtn.config(text="select",
						fg=ToolBase.TEXT_NO,
						state=tk.NORMAL)
			else:
				self._zonesBtn.config(text=f"{self.nzones} selected",
						fg=ToolBase.TEXT_OK,
						state=tk.NORMAL)
		else:
			self.nzones = None

		if self._regionsBtn is not None:
			self.nregions = self.geometry().selection("r")
			if self.nregions==0:
				self._regionsBtn.config(text="select",
						fg=ToolBase.TEXT_NO,
						state=tk.NORMAL)
			else:
				self._regionsBtn.config(text=f"{self.nregions} selected",
						fg=ToolBase.TEXT_OK,
						state=tk.NORMAL)
		else:
			self.nregions = None

		if self._viewportBtn is not None:
			viewport = self.viewers().selectedViewer
			if viewport:
				self._viewportBtn.config(text=viewport.name(),
						foreground=viewport.name())
			else:
				self._viewportBtn.config(text="select", foreground="Black")

		return self.nbodies, self.nzones, self.nregions

#===============================================================================
# ZoneTool
#===============================================================================
class ZoneTool(ToolBase):
	"""\
Create NEW or EDIT a Zone of a Region

Add NEW Zone to a NEW Region:
 - Select only the bodies surrounding the first zone
 - Do not select any region
 - Do not select any zone

Add NEW Zone to an EXITING Region:
 - Select ONE region
 - Select only the bodies surrounding the zone
 - Do not select any zone

Edit EXISTING Zone:
 - Select ONE region
 - Select ONE zone of the region
   Automatically all bodies of that zone will be selected
 - Change the selection of bodies

Selecting the paint operation multiple zones can be
added at once.
"""
	#-----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, "Zone Edit", *args, **kwargs)

		# ===
		row = 0
		self.addRegionsBtn(row, "Region selection:\n"
					"None - to add a new one\n"
					"One - To append a new zone")
		row += 1
		self.addZonesBtn(row,	"Zone selection:\n"
					"None - to add a new zone\n"
					"One - to modify the selected one")
		row += 1
		self.addBodiesBtn(row, "Select the bodies enclosing the zone you want to describe")
		row += 1
		self._operationBtn = self.addRadiobutton(row,
					"Operation",
					"Create a new region, new zone or modify an existing one [d]",
					icons.get("add"),
					ACTION.ADD_ZONE)
		self._paintBtn = tk.Radiobutton(self.frame,
					image=icons.get("paint"),
					variable=self.action,
					value=ACTION.ADD_ZONEPAINT,
					selectcolor="LightGreen",
					indicatoron=0)
		self._paintBtn.grid(row=row, column=2, stick=tk.EW)
		tkExtra.Balloon.set(self._paintBtn,"Enter PAINT mode for adding multiple zones")

		self.hide()

	# ----------------------------------------------------------------------
	def updateSelection(self):
		super().updateSelection()

		err = False
		if self.nbodies==0:
			err = True
		else:
			self._operationBtn.config(text="Add Zone",
					fg=ToolBase.TEXT_NO,
					state=tk.NORMAL)
			self._paintBtn.config(state=tk.NORMAL)

		if self.nzones==0:
			self._operationBtn.config(text="Add Zone",
					fg=ToolBase.TEXT_NO,
					state=tk.NORMAL)
			self._paintBtn.config(state=tk.NORMAL)
		elif self.nzones==1:
			self._operationBtn.config(text="Modify Zone",
					fg=ToolBase.TEXT_NO,
					state=tk.NORMAL)
			self._paintBtn.config(state=tk.NORMAL)
			if self.action.get() == ACTION.SELECT_ZONE:
				self.setAction(ACTION.SELECT_BODY)
		else:
			self._zonesBtn["fg"] = ToolBase.TEXT_ERR
			err = True

		if self.nregions==0:
			self._operationBtn.config(text="Add Region",
					fg=ToolBase.TEXT_NO,
					state=tk.NORMAL)
			self._paintBtn.config(state=tk.NORMAL)
		elif self.nregions==1:
			if self.action.get() == ACTION.SELECT_REGION:
				self.setAction(ACTION.SELECT_ZONE)
		else:
			self._regionsBtn["fg"] = ToolBase.TEXT_ERR
			err = True

		if err:
			self._operationBtn.config(text="Error",
						fg=ToolBase.TEXT_ERR,
						state=tk.DISABLED)
			self._paintBtn.config(state=tk.DISABLED)

	# ----------------------------------------------------------------------
	def nextAction(self):
		#print("Next Action", "B:",self.nbodies, "Z:", self.nzones, "R:", self.nregions)
		if self.nregions<=1 and self.nzones<=1 and self.nbodies>0:
			self.setAction(ACTION.ADD_ZONE)
		elif self.nregions>1:
			self.setAction(ACTION.SELECT_REGION)
		elif self.nregions==1:
			self.setAction(ACTION.SELECT_BODY)
		elif self.nzones>1:
			self.setAction(ACTION.SELECT_ZONE)
		elif self.nbodies==0:
			self.setAction(ACTION.SELECT_BODY)
		else:
			self.setAction(ACTION.ADD_ZONE)

	#-----------------------------------------------------------------------
	def setAction(self, act):
		if act in (ACTION.ADD_ZONE, ACTION.ADD_ZONEPAINT):
			# Special case if the default action is selected then
			# toggle with the other probably which is the body
			if self.action.get()==act:
				act = ACTION.SELECT_BODY
			elif self._operationBtn["state"] == tk.DISABLED:
				return
		super().setAction(act)

	#-----------------------------------------------------------------------
	def show(self, side=None):
		super().show(side)
		# Special case to reset action to ZONE, just in case
		# multiple times the action is called and we are swapping
		# the action from the @see ZoneTool.setAction method
		super().setAction(ACTION.ZONE)

#===============================================================================
# Move Tool
#===============================================================================
class MoveTool(ToolBase):
	"""\
Move selected items:
    objects, bodies, zones or regions
Either:
- dragging with the mouse
- adding a translation vector
- relative to screen
"""
	# ----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, "Move items", *args, **kwargs)

		# ===
		row = 0
		self.addRegionsBtn(row,	"Region selection")
		row += 1
		self.addZonesBtn(row,	"Zone selections")
		row += 1
		self.addBodiesBtn(row,	"Bodies selection")
		row += 1
		self.addObjectsBtn(row,	"Objects selection")
		row += 1
		self.addViewportBtn(row, "Viewport to operate on")

#		self._operationBtn     = self.addRadiobutton(row,
#					"Operation",
#					"Create a new region, new zone or modify an existing one [d]",
#					icons.get("add"),
#					ACTION.ADD_ZONE)
#		self.hide()

#===============================================================================
# Rotate Tool
#===============================================================================
class RotateTool(ToolBase):
	"""\
Rotate selected items:
    objects, bodies, zones or regions
Either:
- dragging with the mouse
- adding a translation vector
- relative to screen
"""
	# ----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, "Rotate items", *args, **kwargs)

		# ===
		row = 0
		self.addRegionsBtn(row,	"Region selection")
		row += 1
		self.addZonesBtn(row,	"Zone selections")
		row += 1
		self.addBodiesBtn(row,	"Bodies selection")
		row += 1
		self.addObjectsBtn(row,	"Objects selection")
		row += 1

#===============================================================================
# Repeat Tool
#===============================================================================
class RepeatTool(ToolBase):
	"""\
Repeat selected items:
    objects, bodies, zones or regions
Either:
- dragging with the mouse
- adding a translation vector
- relative to screen
"""
	# ----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, "Repeat items", *args, **kwargs)

		# ===
		row = 0
		self.addRegionsBtn(row,	"Region selection")
		row += 1
		self.addZonesBtn(row,	"Zone selections")
		row += 1
		self.addBodiesBtn(row,	"Bodies selection")
		row += 1
		self.addObjectsBtn(row,	"Objects selection")
		row += 1

#===============================================================================
# Link Tool
#===============================================================================
class LinkTool(ToolBase):
	"""\
Link object node coordinates with a spline
Select a target object and a handler node
Select also a spline to be linked on this node coordinates
"""
	POSITION = "position"

	# ----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, "Link to spline", "Link", *args, **kwargs)
		# ===
		row = 0
		self.addObjectsBtn(row,	"Select objects to link with spline")

		# ----
		row += 1
		self.addBodiesBtn(row, "Select bodies to link with the spline")
		self.exclude = ["spline"]

		# ----
		row += 1
		hlp = "Select node of object"
		self.node = self.addRadiobutton(row, "Node", hlp,
						icons.get("select"),
						ACTION.SELECT_NODE)

		# ----
		row += 1
		hlp = "Spline to be linked to object"
		self.spline = self.addRadiobutton(row, "Spline", hlp,
						icons.get("select"),
						ACTION.SELECT_SPLINE)

	#-----------------------------------------------------------------------
	def show(self, side=None):
		self.node.config(text=LinkTool.POSITION)
		self.spline.config(text="select")
		super().show(side)
		self.setAction(ACTION.SELECT_OBJECT)

	#-----------------------------------------------------------------------
	def updateSelection(self):
		super().updateSelection()
		# Find selected spline
		#if self.geometry().selection("o"):
		#print("action=",self.action.get())

		if self.action.get() in (ACTION.SELECT_OBJECT, ACTION.SELECT_NODE):
			opt = LinkTool.POSITION
			if self.viewers().closest and self.viewers().closest[2]>0:
				opt = str(self.viewers().closest[2])
			self.node.config(text=opt)
			if self.nobjects: self.nextAction()

		card = self.splineCard()
		self.spline.config(text=card.name() if card else "select",
				image=icons.get("spline"))

	# ----------------------------------------------------------------------
	def nextAction(self):
		#print("nextAction=",self.action.get())
		if self.action.get() == ACTION.SELECT_OBJECT:
			self.setAction(ACTION.SELECT_NODE)
		elif self.action.get() == ACTION.SELECT_NODE:
			self.setAction(ACTION.SELECT_SPLINE)
		else:	# default action
			self.setAction(ACTION.SELECT_OBJECT)

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		super().ok(event)
		# find objects to assign spline
		opt = self.node["text"]
		spline = self.splineCard()
		if spline is None:
			self.flair().notify("Link error",
				"No spline was selected\n"
				"Nothing was done",
				tkFlair.NOTIFY_WARNING)
			return

		s  = f"=spline('{spline.name()}',frame)"
		xs = f"{s}.x"
		ys = f"{s}.y"
		zs = f"{s}.z"
		errors = 0

		#---------------------------------------------------------------
		def setwhat(what,value):
			nonlocal card, undoinfo
			undoinfo.append(self.flair().setWhatUndo(card, what, value))

		#---------------------------------------------------------------
		undoinfo = self.flair().undolistRefreshOnly()
		for card in self.viewers().selectedCards():
			if card.tag == "spline": continue	# ignore splines
			if opt in (LinkTool.POSITION, "0", "1"):
				if card.type() == Card.OBJECT or \
				   card.tag in ("BOX", "SPH", "RCC", "TRC", "WED", "REC", "ELL", "ARB"):
					setwhat(1, xs)
					setwhat(2, ys)
					setwhat(3, zs)
				elif card.tag == "RPP":
					dx = card.numWhat(2)-card.numWhat(1)
					dy = card.numWhat(4)-card.numWhat(3)
					dz = card.numWhat(6)-card.numWhat(5)
					setwhat(1, xs)
					setwhat(2, f"{xs}+{dx}")
					setwhat(3, ys)
					setwhat(4, f"{ys}+{dy}")
					setwhat(5, zs)
					setwhat(6, f"{zs}+{dz}")
				elif card.tag == "PLA":
					setwhat(4, xs)
					setwhat(5, ys)
					setwhat(6, zs)
				elif card.tag == "XYP":
					setwhat(1, zs)
				elif card.tag == "XZP":
					setwhat(1, ys)
				elif card.tag == "YZP":
					setwhat(1, xs)
				elif card.tag in ("XCC", "XEC"):
					setwhat(1,ys)
					setwhat(2,zs)
				elif card.tag in ("YCC", "YEC"):
					setwhat(1, zs)
					setwhat(2, xs)
				elif card.tag in ("ZCC", "ZEC"):
					setwhat(1, xs)
					setwhat(2, ys)
				else:
					errors += 1
			elif opt == "2":
				if card.type() == Card.OBJECT:
					setwhat(7, xs)
					setwhat(8, ys)
					setwhat(9, zs)
				else:
					errors += 1
			elif opt == "3":
				if card.type() == Card.OBJECT:
					setwhat(10, xs)
					setwhat(11, ys)
					setwhat(12, zs)
				else:
					errors += 1

		self.flair().addUndolistRefreshOnly(undoinfo, "Link with spline")

		if errors>0:
			self.flair().notify("Link errors",
				"Could not perform the spline-link on selected objects",
				tkFlair.NOTIFY_ERROR)
		else:
			self.flair().notify("Link",
				f"All selected objects where linked with spline {spline.name()}")

#===============================================================================
# VolumeTool
#===============================================================================
class VolumeTool(ToolBase):
	"""\
Select regions to calculate the volume in cm³.
 - The error is given in absolute values
 - The number of evaluations is 2^iterations.

Notes:
 - Warning: Every iteration doubles the CPU time
 - Every new region is selected a new job will be submitted
   sequentially, using all the user defined CPU cores
 - There is an internal caching of the calculated volumes.
   Therefore, if the regions or bodies are modified
   please restart the calculation
 - Copying [Ctrl-C] has different behaviour:
   - 0 lines are selected on the table. It will copy
             the full table as tab separated list
   - 1 line
"""
	ITERATIONS = 24
	DIGITS     =  5

	# ----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, "Volume Calculation", *args, **kwargs)

		# ===
		row = 0
		self.addRegionsBtn(row,
				"Region selection to calculate volume")
		row += 1
		self.iterations = tk.IntVar()
		self.iterations.set(VolumeTool.ITERATIONS)

		e = self.addSpinbox(row, "Iterations",
				"Number of maximum iterations 2^n", 5, 31)
		e.config(textvariable=self.iterations)

		row += 1
		self.digits = tk.IntVar()
		self.digits.set(VolumeTool.DIGITS)
		e = self.addSpinbox(row, "Digits",
				"Number of significant digits to calculate", 3, 12)
		e.config(textvariable=self.digits)

		self.iterations.trace("w",self.updateVariables)
		self.digits.trace("w",    self.updateVariables)

		row += 1
		self.listbox = tkExtra.MultiListbox(self.frame,
					(("Region", 10, None),
					 ("V(cm³)", 12, None),
					 ("S(cm²)", 12, None),
					 ("M(g)",   12, None),
					 ("err(%)",  7, None),
					 ("n",       9, None)),
					 header=True,
					 stretch="last",
					 background="White")
		self.listbox.bindList("<Control-Key-c>",	self.copy)
		self.listbox.setPopupMenu([
				("Copy",         0, self.copy,         icons.get("copy")),
				("Recalculate",  2, self.recalculate,  icons.get("redo")),
				("Restart",      0, self.restart,      icons.get("reset"))])
		self.listbox.grid(row=row, column=0, columnspan=3, sticky=tk.NSEW)

		self.addButton("Copy", "copy", "Copy information to clipboard",   self.copy)
		self.addButton("Restart", "reset", "Restart volume calculations", self.restart)

		self.hide()
		self.regions = {}		# dictionary of regions to calculate
		self._runid  = None		# running region id
		self._after  = None

	#-----------------------------------------------------------------------
	def show(self, side=None):
		super().show(side)
		self.updateSelection()

	#-----------------------------------------------------------------------
	def cancel(self):
		self.reset()
		super().cancel()

	#-----------------------------------------------------------------------
	def nextAction(self):
		self.setAction(ACTION.SELECT_REGION)

	# ----------------------------------------------------------------------
	def record(self, name, vol, area, mass, err, samples):
		try:
			n = int(math.floor(math.log(samples)/math.log(2.0)))
		except ValueError:
			n = 0
		return (name,
			"%.*g"%(VolumeTool.DIGITS,vol),
			"%.*g"%(VolumeTool.DIGITS,area),
			"%.*g"%(VolumeTool.DIGITS,mass),
			"%.1g"%(err*100),
			str(n))

	# ----------------------------------------------------------------------
	def updateVariables(self, *abc):
		i = VolumeTool.ITERATIONS
		d = VolumeTool.DIGITS
		try:
			VolumeTool.ITERATIONS = max(5,min(30,int(self.iterations.get())))
		except:
			VolumeTool.ITERATIONS = 24
		try:
			VolumeTool.DIGITS = int(self.digits.get())
		except:
			VolumeTool.DIGITS = 5
		if i!=VolumeTool.ITERATIONS or d!=VolumeTool.DIGITS: self.restart()

	# ----------------------------------------------------------------------
	def reset(self):
		if self._after is not None:
			self.after_cancel(self._after)
			self._after = None
		if self._runid is not None:
			self.geometry().volume("stop")
			self._runid = None
		self.regions.clear()

	# ----------------------------------------------------------------------
	# Force restarting of the calculation
	# ----------------------------------------------------------------------
	def restart(self):
		self.reset()
		self.updateSelection()

	# ----------------------------------------------------------------------
	# Force recalculation of everything
	# ----------------------------------------------------------------------
	def recalculate(self):
		for i in self.listbox.curselection():
			name = self.listbox.listbox(0).get(i)
			rid  = self.regions[name][0]
			if self._runid == rid:
				self.geometry().volume("stop")
				self._runid = None
			self.regions[name] = rid, 0.0, 0.0, 0
			self.listbox.insert(i, (name,"?","?","?","?","?"))
		self.updateSelection()

	# ----------------------------------------------------------------------
	# Update selection of regions and start calculation if needed
	# ----------------------------------------------------------------------
	def updateSelection(self):
		super().updateSelection()

		# remove all non calculated regions
		todel = [name for name, (rid,vol,area,mass,err,samples) in self.regions.items()
					if samples==0 ]
		for i in todel:
			del self.regions[i]

		# Update listbox
		self.listbox.delete(0,tk.END)
		for card in self.input().cardsCache(Card.REGION).values():
			if card.ignore(): continue
			if card.get(Card.SELECT,0) & BIT.SELECT:
				try:
					rid,vol,area,mass,err,samples = self.regions[card.name()]
					if samples==0: raise KeyError()
					self.listbox.insert(tk.END,
						self.record(card.name(),vol,area,mass,err,samples))
				except KeyError:
					self.regions[card.name()] = card[Card.ID], 0.0,0.0,0.0,0.0,0
					self.listbox.insert(tk.END, (card.name(),"?","?","?","?","?"))

		if self._after is None and self.regions: self.refresh()

	# ----------------------------------------------------------------------
	# Periodic refresh of the volume progress. Submit new jobs if needed
	# ----------------------------------------------------------------------
	def refresh(self):
		self._after = None
		# Check calculation status
		status,vol,area,mass,err,samples = self.geometry().volume("status")
		if status in ("finished","running"):
			# Store information on previous calculation
			if self._runid is not None:
				name = self.geometry().region(self._runid,"name")
				self.regions[name] = self._runid,vol,area,mass,err,samples
				for i,n in enumerate(self.listbox.listbox(0).get(0,tk.END)):
					if name==n:
						self.listbox.set(i,
							self.record(name,vol,area,mass,err, samples))
						if status == "finished":
							self._runid = None
						break
			# submit a new job
			if status=="finished":
				for name, (rid,vol,area,mass,err,samples) in self.regions.items():
					if samples!=0: continue	# ignore already calculated
					try:
						vol = self.geometry().volume("run",
								rid, 1<<VolumeTool.ITERATIONS)
						self._runid = rid
						break		# submit only one
					except ValueError as ex:
						error(str(ex))
		if self._runid is not None:
			self._after = self.after(500, self.refresh)

	# ----------------------------------------------------------------------
	# Copy one value, or as a tab-separated all list or only selection
	# ----------------------------------------------------------------------
	def copy(self):
		f = self.listbox.focus_get()
		sel = self.listbox.curselection()
		if len(sel)==1:
			sel_text = f.get(sel[0])
		elif not sel:
			sel_text = "\n".join(["\t".join(x) for x in self.listbox.get(0,tk.END)])
		else:
			sel_text = "\n".join(["\t".join(self.listbox.get(x)) for x in sel])
		self.clipboard_clear()
		self.clipboard_append(sel_text)

#===============================================================================
# Export/Print selection
#===============================================================================
class ExportTool(ToolBase):
	"""Select viewport to export to an image format"""

	TYPES   = ["Notes", "Print", "Image", "Vector", "Postscript", "Nox"]
	UNIT = { "px" : 1,
		 "mm" : Postscript.mm,
		 "cm" : Postscript.cm,
		 "in" : Postscript.inch }

	TEMPLATES  = ["", "320x240", "640x480", "800x600", "1024x768", "1440x900", "1600x1200"]

	lastType   = "Image"
	lastFile   = ""
	lastWidth  = "0","px"
	lastHeight = "0","px"

	# ----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, "Export Viewport", "Export", *args, **kwargs)

		# ===
		row = 0
		self.addViewportBtn(row, "Select viewport to export")

		# --- Format type ---
		#  - Notes
		#  - Bitmap Image
		#  - Vector drawing
		#  - Postscript
		#  - Nox
		row += 1
		hlp = "Select image exporting type"
		self.addLabel(row, "Type", hlp)
		self.expType = tkExtra.Combobox(self.frame,
					width=10,
					command=self.expTypeCommand)
		self.expType.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.expType, hlp)
		self.expType.fill(ExportTool.TYPES)

		# --- filename --
		row += 1

		hlp = "Select file name of exported image"
		self.file = self.addLabelEntry(row, "File", hlp)
		self.file.bind('<Return>',   self.ok)
		self.file.bind('<KP_Enter>', self.ok)
		self.fileBtn = tk.Button(self.frame,
					image=icons.get("save"),
					command=self.saveFile)
		tkExtra.Balloon.set(self.fileBtn, hlp)
		self.fileBtn.grid(row=row, column=2, sticky=tk.EW)

		# --- template ---
		row += 1
		hlp = "Select custom image format"
		self.addLabel(row, "Template", hlp)
		self.expTemplate = tkExtra.Combobox(self.frame,
					width=10,
					command=self.expTemplateCommand)
		self.expTemplate.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.expTemplate, hlp)
		self.expTemplate.fill(ExportTool.TEMPLATES)

		# --- width ---
		row += 1
		self.width = self.addFloatEntry(row, "Width",
					"Width of image in pixels, mm, cm or inches")
		self.widthUnit = tkExtra.Combobox(self.frame, width=3)
		self.widthUnit.grid(row=row, column=2, sticky=tk.EW)
		tkExtra.Balloon.set(self.expType, "Select width unit")
		self.widthUnit.fill(["px", "mm", "cm", "in"])
		self.widthUnit.set("px")

		# --- height ---
		row += 1
		self.height = self.addFloatEntry(row, "Height",
					"Height of image in pixels, mm, cm or inches")
		self.heightUnit = tkExtra.Combobox(self.frame, width=3)
		self.heightUnit.grid(row=row, column=2, sticky=tk.EW)
		tkExtra.Balloon.set(self.expType, "Select height unit")
		self.heightUnit.fill(["px", "mm", "cm", "in"])
		self.heightUnit.set("px")

		# --- options ---
		row += 1
		self.xwin = tk.BooleanVar()
		hlp = "Include all drawn objects (object, viewport lines, orientation cube...)"
		self.addLabel(row, "Objects", hlp)
		self.xwinBtn = tk.Checkbutton(self.frame, variable=self.xwin)
		tkExtra.Balloon.set(self.xwinBtn, hlp)
		self.xwinBtn.grid(row=row, column=1, sticky=tk.W)

		# --- resolution? ---
		# Image only or including x11 decorations (e.g. ocube, viewport lines etc...)

	#-----------------------------------------------------------------------
	def show(self, side=None):
		super().show(side)

		self.setFilename(ExportTool.lastFile)
		self.expType.set(ExportTool.lastType)

		self.width.delete(0, tk.END)
		self.width.insert(0, ExportTool.lastWidth[0])
		self.widthUnit.set(ExportTool.lastWidth[1])

		self.height.delete(0, tk.END)
		self.height.insert(0, ExportTool.lastHeight[0])
		self.heightUnit.set(ExportTool.lastHeight[1])

	#-----------------------------------------------------------------------
	def nextAction(self):
		self.setAction(ACTION.SELECT_VIEWPORT)

	#-----------------------------------------------------------------------
	def expTypeCommand(self):
		ExportTool.lastType = etype = self.expType.get()
		state = tk.DISABLED if etype in ("Notes","Print") else tk.NORMAL
		self.file["state"]    = state
		self.fileBtn["state"] = state

		state = tk.DISABLED if etype in ("Vector","Nox") else tk.NORMAL
		self.width["state"]       = state
		self.height["state"]      = state
		self.widthUnit.config(state=state)
		self.heightUnit.config(state=state)
		self.expTemplate.config(state=state)

		state = tk.NORMAL if etype in ("Notes","Image") else tk.DISABLED
		self.xwinBtn["state"] = state

		fn,ext = os.path.splitext(self.file.get())
		if not fn: fn = self.project().name()
		if not fn: fn = "untitled"
		if etype == "Postscript" and ext not in (".ps", ".eps"):
			self.setFilename(f"{fn}.ps")
		elif etype == "Vector" and ext not in (".dxf", ".svg"):
			self.setFilename(f"{fn}.dxf")
		elif etype == "Image" and ext not in (".png", ".jpg", ".jpeg", ".gif"):
			self.setFilename(f"{fn}.png")
		elif etype == "Nox" and ext != ".nox":
			self.setFilename(f"{fn}.nox")

	#-----------------------------------------------------------------------
	def expTemplateCommand(self):
		try:
			width, height = self.expTemplate.get().split("x")
		except (ValueError,IndexError):
			return
		self.width.set(width)
		self.height.set(height)

	#-----------------------------------------------------------------------
	def detectType(self):
		ext = os.path.splitext(self.file.get())[1]
		if ext in (".dxf", ".svg"):
			self.expType.set("Vector")
		elif ext == ".nox":
			self.expType.set("Nox")
		elif ext in (".ps", ".eps"):
			self.expType.set("Postscript")
		else:
			self.expType.set("Image")

	#-----------------------------------------------------------------------
	def setFilename(self, filename):
		if filename=="":
			viewport = self.viewers().selectedViewer
			filename = self.project().name()
			if not filename: filename = "untitled"
			vname    = "-"+viewport.name() if viewport else ""
			filename = f"{filename}{vname}.png"
		state = self.file.cget("state")
		self.file.config(state=tk.NORMAL)
		self.file.delete(0, tk.END)
		self.file.insert(0, filename)
		self.file.config(state=state)

	#-----------------------------------------------------------------------
	def saveFile(self):
		filetypes=[
			("PNG - Portable Network Graphics", "*.png"),			# 0
			("GIF - Graphics Interchange Format", "*.gif"),			# 1
			("JPG - Joint Photographic Experts Group", ("*.jpg","*.jpeg")),	# 2
			("DXF - Data Exchange Format","*.dxf"),				# 3
			("SVG - Scalable Vector Graphics","*.svg"),			# 4
			("PS  - Postscript",("*.ps","*.eps")),				# 5
			("NOX - No-X11 Format","*.nox"),				# 6
			("All","*")]

		promote = 0
		etype = self.expType.get()
		if   etype == "Nox":        promote = 6
		elif etype == "Postscript": promote = 5
		elif etype == "Vector":     promote = 3

		filename = self.file.get()
		if promote==0:
			ext = os.path.splitext(filename)[1]
			if ext:
				for i,ft in enumerate(filetypes):
					if ext in ft[1]:	# promote type
						promote = i
						break
		if promote>0:
			filetypes.insert(0, filetypes.pop(promote))

		filename = bFileDialog.asksaveasfilename(master=self,
				title="Export viewport as",
				initialdir=self.project().dir(),
				initialfile=filename,
				filetypes=filetypes)
		if not filename: return
		self.setFilename(filename)
		self.detectType()

	#-----------------------------------------------------------------------
	def ok(self, event=None):
		viewport = self.viewers().selectedViewer
		self.setAction(ACTION.SELECT)

		if viewport is None:
			self.flair().notify(self.title,
				"No viewport selected to export",
				tkFlair.NOTIFY_ERROR)
			return

		# get width and height in pixels assuming 72 dpi
		try:
			width = int(float(self.width.get())
					* ExportTool.UNIT[self.widthUnit.get()])
			if width==0: width=None
		except ValueError:
			width = None
		try:
			height = int(float(self.height.get())
					* ExportTool.UNIT[self.heightUnit.get()])
			if height==0: height=None
		except ValueError:
			height = None

		option = "win" if self.xwin.get() else None

		ok = None

		# Export based on type and filename
		etype    = self.expType.get()
		filename = self.file.get()
		ext      = os.path.splitext(filename)[1]

		msg = f"Viewport {viewport.name()} exported to {filename!r}"

		if etype == "Notes":
			viewport.toNotes(width, height)
		elif etype == "Print":
			viewport.hardcopy(width=width, height=height)
			msg = f"Viewport {viewport.name()} send to printer"
		elif ext in (".dxf", ".svg"):
			viewport._viewer.export(filename)
		elif ext in (".ps", ".eps"):
			paper = "EPS" if ext == ".eps" else "A4"
			viewport.exportPostscript(filename,
					paper, False,
					width=width,
					height=height)
		elif ext == ".nox":
			ok = viewport.exportNox(filename)
		elif etype == "Image":
			viewport.exportImage(filename, width, height, option)
		else:
			self.flair().notify(self.title,
				f"Unrecognized format/filetype specified '{ext}'",
				tkFlair.NOTIFY_ERROR)

		if ok is None:
			pass	# late exporting
		elif ok:
			self.flair().notify(self.title, msg)
		else:
			self.flair().notify(self.title,
				f"Error exporting to format/filetype '{ext}'",
				tkFlair.NOTIFY_ERROR)

		ExportTool.lastFile = filename

		def zero(x): return x if x else "0"
		ExportTool.lastWidth  = zero(self.width.get()),  self.widthUnit.get()
		ExportTool.lastHeight = zero(self.height.get()), self.heightUnit.get()

		super().ok(event)

#===============================================================================
# Movie tool
#
# FIXME save to Project.MovieInfo as multiple movies
#===============================================================================
class MovieTool(ToolBase):
	"""Create a Movie from a camera object and a viewport"""

	lastFile       = "untitled.mp4"
	lastWidth      = 320
	lastHeight     = 240
	lastFFopts     = ""
	lastStartFrame = None
	lastStopFrame  = None
	lastFPS        = 24
	#lastWin       = False

	# ----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, "Movie Tool", "Render", *args, **kwargs)

		# ===
		row = 0
		self.addViewportBtn(row, "Select viewport to render")

		# --- Format type ---
		#  - Notes
		#  - Bitmap Image
		#  - Vector drawing
		#  - Postscript
		#  - Nox
		#row += 1
		#hlp = "Select image exporting type"
		#self.addLabel(row, "Type", hlp)
		#self.expType = tkExtra.Combobox(self.frame,
		#			width=10,
		#			command=self.expTypeCommand)
		#self.expType.grid(row=row, column=1, sticky=tk.EW)
		#tkExtra.Balloon.set(self.expType, hlp)
		#self.expType.fill(MovieTool.TYPES)

		# --- filename --
		row += 1

		hlp = "Select file name of exported movie"
		self.file = self.addLabelEntry(row, "File", hlp)
		self.fileBtn = tk.Button(self.frame,
					image=icons.get("save"),
					command=self.saveFile)
		tkExtra.Balloon.set(self.fileBtn, hlp)
		self.fileBtn.grid(row=row, column=2, sticky=tk.EW)

		# Variables
		self.startFrame = tk.IntVar()
		self.stopFrame  = tk.IntVar()

		self.startFrame.set(self.timeframe.start)
		self.stopFrame.set(self.timeframe.stop)

		self.startFrame.trace("w", self.timeframeChanged)
		self.stopFrame.trace("w",  self.timeframeChanged)

		# --- frames ---
		row += 1
		e = self.addIntegerEntry(row, "Start Frame", "Start frame index")
		e.config(textvariable=self.startFrame)
		row += 1
		e = self.addIntegerEntry(row, "End Frame", "Last frame index")
		e.config(textvariable=self.stopFrame)
		row += 1
		self.fps = self.addFloatEntry(row, "Frame rate", "Frames per second")
		row += 1
		self.ffopts = self.addLabelEntry(row, "ffmpeg Options",
						"Additional options to ffmpeg")

		# --- template ---
		row += 1
		hlp = "Select custom image format"
		self.addLabel(row, "Template", hlp)
		self.expTemplate = tkExtra.Combobox(self.frame,
					width=10,
					command=self.expTemplateCommand)
		self.expTemplate.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.expTemplate, hlp)
		self.expTemplate.fill(ExportTool.TEMPLATES)

		# --- width ---
		row += 1
		self.width = self.addFloatEntry(row, "Width", "Width of image in pixels")

		# --- height ---
		row += 1
		self.height = self.addFloatEntry(row, "Height",
				"Height of image in pixels.\n"
				"Set to '0' to be calculated from aspect ratio.")

		# --- options ---
		#row += 1
		#self.xwin = tk.BooleanVar()
		#hlp = "Include all drawn objects (object, viewport lines, orientation cube...)"
		#self.addLabel(row, "Objects", hlp)
		#self.xwinBtn = tk.Checkbutton(self.frame, variable=self.xwin)
		#tkExtra.Balloon.set(self.xwinBtn, hlp)
		#self.xwinBtn.grid(row=row, column=1, sticky=tk.W)

		# --- resolution? ---
		# Image only or including x11 decorations (e.g. ocube, viewport lines etc...)

	#-----------------------------------------------------------------------
	@property
	def timeframe(self):
		return self.viewers().timeframe

	#-----------------------------------------------------------------------
	def show(self, side=None):
		super().show(side)

		self.file.delete(0, tk.END)
		self.file.insert(0, MovieTool.lastFile)

		self.startFrame.set(self.timeframe.start
				if MovieTool.lastStartFrame is None
					else MovieTool.lastStartFrame)
		self.stopFrame.set(self.timeframe.stop
				if MovieTool.lastStopFrame is None
					else MovieTool.lastStopFrame)
		self.fps.set(str(MovieTool.lastFPS))
		self.ffopts.delete(0,tk.END)
		self.ffopts.insert(0,MovieTool.lastFFopts)

		self.width.set(str(MovieTool.lastWidth))
		self.height.set(str(MovieTool.lastHeight))

	#-----------------------------------------------------------------------
	# Time frame limits changed
	#-----------------------------------------------------------------------
	def timeframeChanged(self, name, index, mode):
		try:
			start = int(self.startFrame.get())
			stop  = int(self.stopFrame.get())
		except:
			return

		if not (self.timeframe.start <= start <= self.timeframe.stop) or \
		   not (self.timeframe.start <= stop  <= self.timeframe.stop):
			# increase limits if needed
			self.timeframe.setLimits(start, stop)

	#-----------------------------------------------------------------------
	def nextAction(self):
		self.setAction(ACTION.SELECT_VIEWPORT)

	#-----------------------------------------------------------------------
	def updateSelection(self):
		super().updateSelection()
		self.setAction(ACTION.MOVIE)

	#-----------------------------------------------------------------------
	def expTemplateCommand(self):
		try:
			width, height = self.expTemplate.get().split("x")
		except:
			return
		self.width.set(width)
		self.height.set(height)

	#-----------------------------------------------------------------------
	def saveFile(self):
		pass
		#filetypes=[
		#	("PNG - Portable Network Graphics", "*.png"),			# 0
		#	("GIF - Graphics Interchange Format", "*.gif"),			# 1
		#	("JPG - Joint Photographic Experts Group", ("*.jpg","*.jpeg")),	# 2
		#	("DXF - Data Exchange Format","*.dxf"),				# 3
		#	("SVG - Scalable Vector Graphics","*.svg"),			# 4
		#	("PS  - Postscript",("*.ps","*.eps")),				# 5
		#	("NOX - No-X11 Format","*.nox"),				# 6
		#	("All","*")]

		#promote = 0
		#etype = self.expType.get()
		#if   etype == "Nox":        promote = 6
		#elif etype == "Postscript": promote = 5
		#elif etype == "Vector":     promote = 3

		#filename = self.file.get()
		#if promote==0:
		#	ext = os.path.splitext(filename)[1]
		#	if ext:
		#		for i,ft in enumerate(filetypes):
		#			if ext in ft[1]:	# promote type
		#				promote = i
		#				break
		#if promote>0:
		#	filetypes.insert(0, filetypes.pop(promote))

		#filename = bFileDialog.asksaveasfilename(master=self,
		#		title="Export viewport as",
		#		initialdir=self.project().dir,
		#		initialfile=filename,
		#		filetypes=filetypes)
		#if not filename: return
		#self.setFilename(filename)
		#self.detectType()

	#-----------------------------------------------------------------------
	def ok(self, event=None):
		args = {}

		# Get the variables
		viewport = self.viewers().selectedViewer
		self.setAction(ACTION.SELECT)
		filename = self.file.get()
		# get width and height in pixels assuming 72 dpi
		try:
			args["width"] = int(self.width.get())
		except ValueError:
			args["width"] = 0
		try:
			args["height"] = int(self.height.get())
		except ValueError:
			args["height"] = 0
		try:
			args["fps"] = int(self.fps.get())
		except:
			args["fps"] = 0

		args["start"]  = self.startFrame.get()
		args["stop"]   = self.stopFrame.get()
		args["ffopts"] = self.ffopts.get()

		# save them in the class
		MovieTool.lastFile       = filename
		MovieTool.lastWidth      = args["width"]
		MovieTool.lastHeight     = args["height"]
		MovieTool.lastStartFrame = args["start"]
		MovieTool.lastStopFrame  = args["stop"]
		MovieTool.lastFPS        = args["fps"]
		MovieTool.lastFFopts     = args["ffopts"]

		# Error checking
		if not filename:
			self.flair().notify(self.title,
				"Invalid filename specified",
				tkFlair.NOTIFY_ERROR)
			return
		if viewport is None:
			self.flair().notify(self.title,
				"No viewport selected to export",
				tkFlair.NOTIFY_ERROR)
			return
		if args["width"]<=5:
			self.flair().notify(self.title,
				"Invalid image width",
				tkFlair.NOTIFY_ERROR)
			return
		if args["height"]<=5:
			self.flair().notify(self.title,
				"Invalid image height",
				tkFlair.NOTIFY_ERROR)
			return
		if args["fps"]<=0:
			self.flair().notify(self.title,
				"Invalid FPS (frames per second)",
				tkFlair.NOTIFY_ERROR)
			return

		# execute
		viewport.exportMovie(filename, **args)
		super().ok(event)

#===============================================================================
# Planner
#
# calculate integrated dose along a path
#===============================================================================
class PlannerTool(ToolBase):
	"""Calculate integrated dose along a spline path.
	Select a viewport containing a layer with a USRBIN,
	Select a spline path"""

	lastFile   = "planner.dat"
	lastTScale = 1.0

	# ----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, "Planner", "Calculate", *args, **kwargs)

		# ===
		row = 0
		self.addViewportBtn(row, "Select viewport to use information from")

		# ----
		row += 1
		hlp = "Usrbin associated with viewport/layer"
		self.usrbin = self.addLabelInfo(row, "Usrbin", hlp, fg="DarkBlue")

		# ----
		row += 1
		hlp = "Normalization of the usrbin"
		self.norm = self.addLabelInfo(row, "Norm", hlp, fg="DarkBlue")

		# ----
		row += 1
		hlp = "Spline path to integrate usrbin"
		self.spline = self.addRadiobutton(row, "Path", hlp,
						icons.get("select"),
						ACTION.SELECT_SPLINE)

		# ----
		row += 1
		self.tscale = self.addFloatEntry(row, "Time scale",
					"Time scaling factor for spline. Default 1s")

		# --- filename --
		row += 1
		hlp = "Select file name of planner report and data"
		self.file = self.addLabelEntry(row, "File", hlp)
		self.fileBtn = tk.Button(self.frame,
					image=icons.get("save"),
					command=self.saveFile)
		tkExtra.Balloon.set(self.fileBtn, hlp)
		self.fileBtn.grid(row=row, column=2, sticky=tk.EW)

	#-----------------------------------------------------------------------
	@property
	def timeframe(self):
		return self.viewers().timeframe

	#-----------------------------------------------------------------------
	def show(self, side=None):
		super().show(side)
		self.file.delete(0, tk.END)
		self.file.insert(0, PlannerTool.lastFile)
		self.tscale.set(self.lastTScale)

	#-----------------------------------------------------------------------
	def nextAction(self):
		self.setAction(ACTION.SELECT_VIEWPORT)

	#-----------------------------------------------------------------------
	def updateSelection(self):
		super().updateSelection()
		viewport = self.viewers().selectedViewer
		if not viewport: return

		# Find Usrbin and normalization
		tag = Project.GeometryLayerInfo.USRBINNAMES[0]
		viewframe = viewport.frame
		if viewframe.getInt(tag):
			filename = viewframe.getVar(tag+".file")
			try: det = viewframe.getInt(tag+".det")
			except: det = 1
			self.usrbin["text"] = f"{filename}:{det}"
			self.norm["text"]   = viewframe.getVar(tag+".norm")

		# Find selected spline
		if self.geometry().selection("o") != 1:
			self.setAction(ACTION.SELECT_SPLINE)
			self.spline.config(text="select",
					image=icons.get("select"))
		else:
			card = self.splineCard()
			if card:
				self.spline.config(text=card.name(),
						image=icons.get("spline"))
				self.file.focus_set()
			else:
				self.spline.config(text="select",
						image=icons.get("select"))

		self.setAction(ACTION.PLANNER)

	#-----------------------------------------------------------------------
	def saveFile(self):
		pass

	#-----------------------------------------------------------------------
	def ok(self, event=None):
		viewport = self.viewers().selectedViewer
		self.setAction(ACTION.SELECT)

		if viewport is None:
			self.flair().notify(self.title,
				"No viewport selected to export",
				tkFlair.NOTIFY_ERROR)
			return

		filename = self.file.get()
		if not filename:
			self.flair().notify(self.title,
				"Invalid filename specified",
				tkFlair.NOTIFY_ERROR)
			return

		if not self.usrbin["text"]:
			self.flair().notify(self.title,
				"No USRBIN file specified",
				tkFlair.NOTIFY_ERROR)
			return

		spline = self.splineCard()
		if not spline:
			self.flair().notify(self.title,
				"No path selected",
				tkFlair.NOTIFY_ERROR)
			return

		try:
			tscale = float(self.tscale.get())
		except:
			tscale = 1.0

		try:
			viewport._viewer.planner(filename, spline[Card.ID], tscale)
		except Exception as ex:
			self.flair().notify(self.title,
				f"Error in planner {ex}",
				tkFlair.NOTIFY_ERROR)
			return

		self.flair().notify(self.title,
			f"Planner calculated in file {filename}")
		self.flair().viewer(filename)
		self.flair().startPlotEngine("Gnuplot")
		self.flair().plotEngine.write(f"load '{filename}'")

		PlannerTool.lastFile   = filename
		PlannerTool.lastTScale = tscale
		super().ok(event)

#-------------------------------------------------------------------------------
def loadConfig(config):
	section = tkFlair._GEOMETRY_SECTION
	VolumeTool.ITERATIONS  = config.getInt(section,   "volume.iterations", VolumeTool.ITERATIONS)
	VolumeTool.DIGITS      = config.getInt(section,   "volume.digits",     VolumeTool.DIGITS)
	ExportTool.lastWidth   = config.getStr(section,   "export.width",  " ".join(ExportTool.lastWidth)).split()
	ExportTool.lastHeight  = config.getStr(section,   "export.height", " ".join(ExportTool.lastHeight)).split()
	PlannerTool.lastTScale = config.getFloat(section, "planner.tscale",    PlannerTool.lastTScale)

#-------------------------------------------------------------------------------
def saveConfig(config):
	section = tkFlair._GEOMETRY_SECTION
	config.setInt(section,   "volume.iterations",  VolumeTool.ITERATIONS)
	config.setInt(section,   "volume.digits",      VolumeTool.DIGITS)
	config.setStr(section,   "export.width",       " ".join(ExportTool.lastWidth))
	config.setStr(section,   "export.height",      " ".join(ExportTool.lastHeight))
	config.setFloat(section, "planner.tscale",     PlannerTool.lastTScale)
