# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   27-Jan-2012

import math
import tkinter as tk

import bmath
import icons
import tkExtra

from fluka import FlukaCard

FMT = "%.10g"
CAMERA_PREFIX = "\N{PLACE OF INTEREST SIGN} "

#===============================================================================
# Base frame for origin manipulation
#===============================================================================
class _BaseFrame(tk.Frame):
	def __init__(self, master, geometryFrame, *args):
		super().__init__(master, *args)
		self.geometryFrame = geometryFrame
		self.viewer   = geometryFrame.viewer
		self._viewer  = geometryFrame.viewer._viewer

	# ----------------------------------------------------------------------
	# Return the apply button
	# ----------------------------------------------------------------------
	def applyButton(self, msg, compound=tk.LEFT):
		b = tk.Button(self, text="Apply",
				image=icons.get("ok"),
				compound=compound,
				command=self.apply,
				padx=1,
				pady=0)
		tkExtra.Balloon.set(b, msg)
		return b

	# ----------------------------------------------------------------------
	# fill information from viewer on every draw
	# ----------------------------------------------------------------------
	def fill(self):  pass

	# ----------------------------------------------------------------------
	# load information from project/input
	# ----------------------------------------------------------------------
	def load(self):	 pass

	# ----------------------------------------------------------------------
	# apply changes to viewer
	# ----------------------------------------------------------------------
	def apply(self, event=None):
		self.viewer.focus_set()

	# ----------------------------------------------------------------------
	# update viewer
	# ----------------------------------------------------------------------
	def draw(self, invalid, expose):
		self.viewer.draw(invalid, expose)
		self.viewer.synchronizeViewports()

#===============================================================================
class _NavigationFrame(_BaseFrame):
	name = "Navigation"
	icon = "navigation"

	#-----------------------------------------------------------------------
	def __init__(self, master, geometryFrame, *args):
		super().__init__(master, geometryFrame, *args)

		# back
		self.backwardButton = tk.Button(self, image=icons.get("left"),
				takefocus=0,
				command=self.viewer.historyBackward,
				state=tk.DISABLED,
				padx=0, pady=0)
		self.backwardButton.pack(side=tk.LEFT)
		tkExtra.Balloon.set(self.backwardButton, "Move back one view [<]")

		# forward
		self.forwardButton = tk.Button(self, image=icons.get("right"),
				takefocus=0,
				command=self.viewer.historyForward,
				state=tk.DISABLED,
				padx=0, pady=0)
		self.forwardButton.pack(side=tk.LEFT)
		tkExtra.Balloon.set(self.forwardButton, "Go forward one view [>]")

		# reload
		b = tk.Button(self, image=icons.get("refresh"),
				takefocus=0,
				command=geometryFrame.redraw,
				padx=0, pady=0)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b, "Refresh display")

		# sep
		tk.Label(self, image=icons.get("sep")).pack(side=tk.LEFT, padx=3)

		# layers
		b = tk.Button(self, image=icons.get("layers"),
				takefocus=0,
				command=self.geometryFrame.layersDialog,
				padx=0, pady=0)
		tkExtra.Balloon.set(b, "Configure the layers [Ctrl-L]")
		b.pack(side=tk.LEFT)

		# layer Combo
		self.layerCombo = tkExtra.Combobox(self, label=True, width=10,
					takefocus=0,
					command=self.geometryFrame.changeLayer)
		self.layerCombo.pack(side=tk.LEFT, padx=1, pady=0, expand=tk.YES, fill=tk.BOTH)
		self.layerCombo.select(1)

		# Bookmarks
		b = tkExtra.AnyButton(self, image=icons.get("star"),
					command=self.addBookmark,
					command3=self.updateBookmark)
		b.pack(side=tk.LEFT)

		tkExtra.Balloon.set(b, "Save/Bookmark current view as a new geometry plot.\n"
					"Right click to update existing plot")

		self.bookmarkCombo = tkExtra.Combobox(self, False, width=9,
					command=self.changeBookmark)
		self.bookmarkCombo.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
		self.bookmarkCombo._text.config(background="White")
		self.bookmarkCombo._text.bind("<FocusOut>", self.renameBookmark)
		self.bookmarkCombo._text.bind("<Return>",   self.returnHit)
		self.bookmarkCombo._text.bind("<KP_Enter>", self.returnHit)

		b = tk.Button(self,
				text="\N{VECTOR OR CROSS PRODUCT}",
				padx=1, pady=0,
				command=self.delBookmark)
		b.pack(side=tk.LEFT, fill=tk.Y)
		tkExtra.Balloon.set(b, "Delete selected bookmark")

		self.lastBookmark = self.geometryFrame.name

	# ----------------------------------------------------------------------
	def historyButton(self, left, right):
		if left:
			self.backwardButton["state"]=tk.NORMAL
		else:
			self.backwardButton["state"]=tk.DISABLED

		if right:
			self.forwardButton["state"]=tk.NORMAL
		else:
			self.forwardButton["state"]=tk.DISABLED

	# ----------------------------------------------------------------------
	def _loadBookmarks(self):
		lst = []
		for plot in self.geometryFrame.project().plots():
			if plot.type == "Geometry":
				lst.append(plot.name)
		for camera in self.geometryFrame.project().input["camera"]:
			if camera.ignore(): continue
			lst.append(f"{CAMERA_PREFIX}{camera.name()}")
		self.bookmarkCombo.fill(lst)

	# ----------------------------------------------------------------------
	def load(self):
		self._loadBookmarks()
		if self.bookmarkCombo.get() != self.lastBookmark:
			self.bookmarkCombo.set(self.geometryFrame.name)

	# ----------------------------------------------------------------------
	def apply(self, event=None):
		super().apply()
		self.geometryFrame.loadPlot(self.bookmarkCombo.get())

	# ----------------------------------------------------------------------
	def changeBookmark(self):
		self.lastBookmark = self.bookmarkCombo.get()
		self.apply()

	# ----------------------------------------------------------------------
	def addBookmark(self, event=None):
		for i in range(1,1000):
			name = f"Geometry{i:02d}"
			for plot in self.geometryFrame.project().plots():
				if plot.type == "Geometry" and plot.name == name:
					break
			else:
				break
		self.geometryFrame.savePlot(name)
		self.bookmarkCombo.insert(tk.END,name)
		self.lastBookmark = name
		self.bookmarkCombo.set(name)
		self.geometryFrame.parent.loadOrigin()

	# ----------------------------------------------------------------------
	def updateBookmark(self, event=None):
		name = self.bookmarkCombo.get()
		self.geometryFrame.savePlot(name)

	# ----------------------------------------------------------------------
	def delBookmark(self, event=None):
		name = self.bookmarkCombo.get()
		for i, plot in enumerate(self.geometryFrame.project().plots()):
			if plot.name == name and plot.type == "Geometry":
				del self.geometryFrame.project().plots[i]
				self.bookmarkCombo.delete(i)
				break
		self._loadBookmarks()
		self.lastBookmark = self.geometryFrame.name
		self.bookmarkCombo.set(self.lastBookmark)
		self.geometryFrame.parent.loadOrigin()

	# ----------------------------------------------------------------------
	def renameBookmark(self, event=None):
		if self.lastBookmark == self.bookmarkCombo.get(): return
		if self.focus_get() == self.bookmarkCombo._text: return
		for plot in self.geometryFrame.project().plots():
			if plot.name == self.lastBookmark and plot.type == "Geometry":
				plot.name = self.bookmarkCombo.get()
				self._loadBookmarks()
				self.lastBookmark = plot.name
				self.bookmarkCombo.set(self.lastBookmark)
				self.geometryFrame.parent.loadOrigin()
				return

	# ----------------------------------------------------------------------
	def returnHit(self, event=None):
		self.viewer.focus_set()

#===============================================================================
class _OriginFrame(_BaseFrame):
	name = "Origin"
	icon = "center"

	#-----------------------------------------------------------------------
	def __init__(self, master, geometryFrame, *args):
		super().__init__(master, geometryFrame, *args)

		tk.Label(self,text="x:").pack(side=tk.LEFT)
		self.xentry = tkExtra.VectorEntry(self, width=6, background="White")
		self.xentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.xentry.bind("<Return>", self.apply)
		self.xentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.xentry, "Origin X-coordinate or all coordinates (x y z) as a vector")

		tk.Label(self,text="y:").pack(side=tk.LEFT, )
		self.yentry = tkExtra.FloatEntry(self, width=6, background="White")
		self.yentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.yentry.bind("<Return>", self.apply)
		self.yentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.yentry, "Origin Y-coordinate")

		tk.Label(self,text="z:").pack(side=tk.LEFT, )
		self.zentry = tkExtra.FloatEntry(self, width=6, background="White")
		self.zentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.zentry.bind("<Return>", self.apply)
		self.zentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.zentry, "Origin Z-coordinate")

		self.applyButton("Set Origin").pack(side=tk.RIGHT)

	# ----------------------------------------------------------------------
	def fill(self):
		xyz = self._viewer.origin()
		self.xentry.set(FMT%(xyz[0]))
		self.yentry.set(FMT%(xyz[1]))
		self.zentry.set(FMT%(xyz[2]))

	# ----------------------------------------------------------------------
	def apply(self, event=None):
		super().apply()
		x,y,z = self._viewer.origin()
		self.xentry.split((self.yentry, self.zentry))
		x = self.xentry.getfloat(x)
		y = self.yentry.getfloat(y)
		z = self.zentry.getfloat(z)
		self._viewer.viewOto0()
		self._viewer.origin(x,y,z)
		self.draw(True, True)

#===============================================================================
class _ExtendsFrame(_BaseFrame):
	name = "Extends"
	icon = "expand"

	#-----------------------------------------------------------------------
	def __init__(self, master, geometryFrame, *args):
		super().__init__(master, geometryFrame, *args)

		tk.Label(self,text="Δu:").pack(side=tk.LEFT)
		self.duentry = tkExtra.FloatEntry(self, width=6, background="White")
		self.duentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.duentry.bind("<Return>", self.apply)
		self.duentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.duentry, "Extends on U")

		tk.Label(self,text="Δv:").pack(side=tk.LEFT)
		self.dventry = tk.Label(self, background="White", anchor=tk.W, relief=tk.SUNKEN)
		self.dventry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		tkExtra.Balloon.set(self.dventry, "Extends on V")

		tk.Label(self,text="Aspect(X/Y):").pack(side=tk.LEFT)
		self.aspectentry = tkExtra.FloatEntry(self, width=6, background="White")
		self.aspectentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.aspectentry.bind("<Return>", self.apply)
		self.aspectentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.aspectentry, "Aspect Ratio")

		self.applyButton("Set extends and aspect ratio").pack(side=tk.RIGHT)

	# ----------------------------------------------------------------------
	def fill(self):
		x1,y1,x2,y2 = self._viewer.extends()
		z = self._viewer.zoom()
		du = (x2-x1)/2.0/z
		w,h = self._viewer.size()
		dv = float(h)/float(w) * du
		self.duentry.set(FMT%(du))
		self.dventry["text"] = FMT%(dv)
		self.aspectentry.set(FMT%(self._viewer.aspect()))

	# ----------------------------------------------------------------------
	def apply(self, event=None):
		super().apply()
		try: du = float(self.duentry.get())
		except: return
		try:
			aspect = float(self.aspectentry.get())
		except:
			aspect = 1.0
		du = 2.0*du
		self._viewer.extends(-du, -du, du, du)
		self._viewer.zoom(2.0)
		self._viewer.aspect(aspect)
		self.draw(True, True)

#===============================================================================
class _MoveFrame(_BaseFrame):
	name = "Move"
	icon = "pan"

	#-----------------------------------------------------------------------
	def __init__(self, master, geometryFrame, *args):
		super().__init__(master, geometryFrame, *args)

		tk.Label(self,text="+u:").pack(side=tk.LEFT)
		self.uentry = tkExtra.VectorEntry(self, width=6, background="White")
		self.uentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.uentry.bind("<Return>", self.apply)
		self.uentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.uentry, "Shift Origin on U or by vector (ux, uy, uz)")

		tk.Label(self,text="+v:").pack(side=tk.LEFT)
		self.ventry = tkExtra.FloatEntry(self, width=6, background="White")
		self.ventry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.ventry.bind("<Return>", self.apply)
		self.ventry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.ventry, "Shift Origin on V")

		tk.Label(self,text="+w:").pack(side=tk.LEFT)
		self.wentry = tkExtra.FloatEntry(self, width=6, background="White")
		self.wentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.wentry.bind("<Return>", self.apply)
		self.wentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.wentry, "Shift Origin on W")

		self.applyButton("Move origin by").pack(side=tk.RIGHT)

	# ----------------------------------------------------------------------
	def apply(self, event=None):
		super().apply()
		u,v = self._viewer.offset()
		self.uentry.split((self.ventry,self.wentry))
		u += self.uentry.getfloat()
		v += self.ventry.getfloat()
		self._viewer.offset(u,v)

		w = self.wentry.getfloat()
		if abs(w)>1e-10:
			matrix = bmath.Matrix(self._viewer.matrix())
			self._viewer.matrix(matrix*bmath.Matrix.translate(0.0, 0.0, w))
			self.draw(True, True)
		else:
			self.draw(False, True)

#===============================================================================
class _BasisFrame(_BaseFrame):
	name = "Basis"
	icon = "axes"

	#-----------------------------------------------------------------------
	def __init__(self, master, geometryFrame, *args):
		super().__init__(master, geometryFrame, *args)

		row = 0
		tk.Label(self,text="ux:").grid(row=row, column=0)
		self.uxentry = tkExtra.VectorEntry(self, width=5, background="White")
		self.uxentry.grid(row=row, column=1, sticky=tk.EW)
		self.uxentry.bind("<Return>", self.apply)
		self.uxentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.uxentry, "Enter U-axis x-direction cosine or all (ux uy uz) as a vector")

		tk.Label(self,text="uy:").grid(row=row, column=2)
		self.uyentry = tkExtra.FloatEntry(self, width=5, background="White")
		self.uyentry.grid(row=row, column=3, sticky=tk.EW)
		self.uyentry.bind("<Return>", self.apply)
		self.uyentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.uyentry, "Enter U-axis y-direction cosine")

		tk.Label(self,text="uz:").grid(row=row, column=4)
		self.uzentry = tkExtra.FloatEntry(self, width=5, background="White")
		self.uzentry.grid(row=row, column=5, sticky=tk.EW)
		self.uzentry.bind("<Return>", self.apply)
		self.uzentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.uzentry, "Enter U-axis z-direction cosine")

		f = tk.Frame(self)
		f.grid(row=row, column=6, sticky=tk.EW)
		tk.Button(f, text="x-y", padx=3, pady=1, command=self.orientXY).pack(side=tk.LEFT)
		tk.Button(f, text="x-z", padx=2, pady=1, command=self.orientXZ).pack(side=tk.LEFT)
		tk.Button(f, text="-v",  padx=2, pady=1, command=self.orient_V).pack(side=tk.LEFT)
		tk.Button(f, text="-u",  padx=2, pady=1, command=self.orient_U).pack(side=tk.LEFT)

		f = tk.Frame(self)
		f.grid(row=row, column=7, sticky=tk.EW)
		tk.Button(f, image=icons.get("rotate_left"), command=self.rotWp, pady=1).pack(side=tk.LEFT)
		tk.Button(f, image=icons.get("rotate_vleft"),command=self.rotVp, pady=1).pack(side=tk.LEFT)
		tk.Button(f, image=icons.get("rotate_up"),   command=self.rotUp, pady=1).pack(side=tk.LEFT)

		row = 1
		tk.Label(self,text="vx:").grid(row=row, column=0)
		self.vxentry = tkExtra.VectorEntry(self, width=5, background="White")
		self.vxentry.grid(row=row, column=1, sticky=tk.EW)
		self.vxentry.bind("<Return>", self.apply)
		self.vxentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.vxentry, "Enter V-axis x-direction cosine or all (vx vy vz) as a vector")

		tk.Label(self,text="vy:").grid(row=row, column=2)
		self.vyentry = tkExtra.FloatEntry(self, width=5, background="White")
		self.vyentry.grid(row=row, column=3, sticky=tk.EW)
		self.vyentry.bind("<Return>", self.apply)
		self.vyentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.vyentry, "Enter V-axis y-direction cosine")

		tk.Label(self,text="vz:").grid(row=row, column=4)
		self.vzentry = tkExtra.FloatEntry(self, width=5, background="White")
		self.vzentry.grid(row=row, column=5, sticky=tk.EW)
		self.vzentry.bind("<Return>", self.apply)
		self.vzentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.vyentry, "Enter V-axis z-direction cosine")

		f = tk.Frame(self)
		f.grid(row=row, column=6, sticky=tk.EW)
		tk.Button(f, text="y-z", padx=2, pady=1, command=self.orientYZ).pack(side=tk.LEFT)
		tk.Button(f, text="iso",  padx=2, pady=1, command=self.orientISO).pack(side=tk.LEFT)
		tk.Button(f, text="swap", padx=2, pady=1, command=self.orientSWAP).pack(side=tk.LEFT)

		f = tk.Frame(self)
		f.grid(row=row, column=7, sticky=tk.EW)
		tk.Button(f, image=icons.get("rotate_right"), command=self.rotWm, pady=1).pack(side=tk.LEFT)
		tk.Button(f, image=icons.get("rotate_vright"),command=self.rotVm, pady=1).pack(side=tk.LEFT)
		tk.Button(f, image=icons.get("rotate_down"),  command=self.rotUm, pady=1).pack(side=tk.LEFT)

		self.applyButton("Set basis orientation", compound=tk.TOP).grid(row=0, column=13, rowspan=2)

		self.grid_columnconfigure(1,weight=1)
		self.grid_columnconfigure(3,weight=1)
		self.grid_columnconfigure(5,weight=1)

	# ----------------------------------------------------------------------
	def fill(self):
		matrix = bmath.Matrix(self._viewer.matrix())
		self.orientSet(	matrix[0][0], matrix[1][0], matrix[2][0],
				matrix[0][1], matrix[1][1], matrix[2][1])

	# ----------------------------------------------------------------------
	def rotUp(self): self.rotate( 90.0,  0.0,  0.0)
	def rotUm(self): self.rotate(-90.0,  0.0,  0.0)
	def rotVp(self): self.rotate(  0.0, 90.0,  0.0)
	def rotVm(self): self.rotate(  0.0,-90.0,  0.0)
	def rotWp(self): self.rotate(  0.0,  0.0,-90.0)
	def rotWm(self): self.rotate(  0.0,  0.0, 90.0)

	# ----------------------------------------------------------------------
	def rotate(self, ru, rv, rw):
		super().apply()
		self._viewer.viewOto0()
		rotated = False
		ang = math.radians(ru)
		if abs(ang)>1e-10:
			self._viewer.matrix(bmath.Matrix(self._viewer.matrix())*bmath.Matrix.rotX(ang))
			rotated = True

		ang = math.radians(rv)
		if abs(ang)>1e-10:
			self._viewer.matrix(bmath.Matrix(self._viewer.matrix())*bmath.Matrix.rotY(ang))
			rotated = True

		ang = math.radians(rw)
		if abs(ang)>1e-10:
			self._viewer.matrix(bmath.Matrix(self._viewer.matrix())*bmath.Matrix.rotZ(ang))
			rotated = True
		if rotated: self.draw(True, True)

	# ----------------------------------------------------------------------
	def apply(self, event=None):
		super().apply()
		try: u,v,w = self.orientNorm()
		except: return

		xyz    = self._viewer.origin()
		matrix = self._viewer.matrix()
		for i in range(3):
			matrix[i][0] = u[i]
			matrix[i][1] = v[i]
			matrix[i][2] = w[i]
		self._viewer.matrix(matrix)
		self._viewer.origin(xyz)
		self.draw(True, True)

	# ----------------------------------------------------------------------
	def orientSet(self, ux=None, uy=None, uz=None, vx=None, vy=None, vz=None):
		if ux is not None: self.uxentry.set(str(ux))
		if uy is not None: self.uyentry.set(str(uy))
		if uz is not None: self.uzentry.set(str(uz))
		if vx is not None: self.vxentry.set(str(vx))
		if vy is not None: self.vyentry.set(str(vy))
		if vz is not None: self.vzentry.set(str(vz))

	# ----------------------------------------------------------------------
	def orientXY(self, event=None):
		self.orientSet(1.0,0.0,0.0, 0.0,1.0,0.0)
		self.apply()

	# ----------------------------------------------------------------------
	def orientXZ(self, event=None):
		self.orientSet(1.0,0.0,0.0, 0.0,0.0,1.0)
		self.apply()

	# ----------------------------------------------------------------------
	def orientYZ(self, event=None):
		self.orientSet(0.0,1.0,0.0, 0.0,0.0,1.0)
		self.apply()

	# ----------------------------------------------------------------------
	def orientISO(self, event=None):
		# As in the 'Green' viewport from GeometryViewers ORIENTATIONS
		Rx,Ry,Rz = map(math.radians, (30,30,15))
		matrix = bmath.Matrix.eulerRotation(Rx,Ry,Rz)
		self.orientSet(	matrix[0][0], matrix[1][0], matrix[2][0],
				matrix[0][1], matrix[1][1], matrix[2][1])
		self.apply()

	# ----------------------------------------------------------------------
	def orientSWAP(self, event=None):
		self.orientSet(	self.vxentry.get(), self.vyentry.get(), self.vzentry.get(),
				self.uxentry.get(), self.uyentry.get(), self.uzentry.get())
		self.apply()

	# ----------------------------------------------------------------------
	# Negate U vector
	# ----------------------------------------------------------------------
	def orient_U(self, event=None):
		try: ux = -float(self.uxentry.get())
		except: ux = None
		try: uy = -float(self.uyentry.get())
		except: uy = None
		try: uz = -float(self.uzentry.get())
		except: uz = None
		self.orientSet(	ux, uy, uz )
		self.apply()

	# ----------------------------------------------------------------------
	# Negate V vector
	# ----------------------------------------------------------------------
	def orient_V(self, event=None):
		try: vx = -float(self.vxentry.get())
		except: vx = None
		try: vy = -float(self.vyentry.get())
		except: vy = None
		try: vz = -float(self.vzentry.get())
		except: vz = None
		self.orientSet(	None, None, None, vx, vy, vz )
		self.apply()

	# ----------------------------------------------------------------------
	def orientNorm(self, event=None):
		# try it as a vector
		self.uxentry.split((self.uyentry, self.uzentry))
		self.vxentry.split((self.vyentry, self.vzentry))

		# Get vectors
		u = bmath.Vector(self.uxentry.getfloat(),
				 self.uyentry.getfloat(),
				 self.uzentry.getfloat())

		v = bmath.Vector(self.vxentry.getfloat(0.0),
				 self.vyentry.getfloat(1.0),
				 self.vzentry.getfloat(0.0))

		# normalize vectors
		u.norm()
		v.norm()

		# find w
		w = u.cross(v)
		if w.length() < 0.01:
			# try another vector
			w = u.cross(bmath.Vector.Z)
			if w.length() < 0.01:
				w = u.cross(bmath.Vector.X)
		w.norm()
		v = w.cross(u)
		v.norm()

		self.orientSet(u[0], u[1], u[2], v[0], v[1], v[2])
		return u,v,w

#===============================================================================
class _EulerFrame(_BaseFrame):
	name = "Euler"
	icon = "orbit"

	#-----------------------------------------------------------------------
	def __init__(self, master, geometryFrame, *args):
		super().__init__(master, geometryFrame, *args)

		tk.Label(self,text="Rx:").pack(side=tk.LEFT)
		self.rxentry = tkExtra.VectorEntry(self, width=6, background="White")
		self.rxentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.rxentry.bind("<Return>", self.apply)
		self.rxentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.rxentry, "Enter Rx Euler angle or all (Rx Ry Rz)")

		tk.Label(self,text="Ry:").pack(side=tk.LEFT)
		self.ryentry = tkExtra.FloatEntry(self, width=6, background="White")
		self.ryentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.ryentry.bind("<Return>", self.apply)
		self.ryentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.ryentry, "Enter Ry Euler angle")

		tk.Label(self,text="Rz:").pack(side=tk.LEFT)
		self.rzentry = tkExtra.FloatEntry(self, width=6, background="White")
		self.rzentry.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
		self.rzentry.bind("<Return>", self.apply)
		self.rzentry.bind("<KP_Enter>", self.apply)
		tkExtra.Balloon.set(self.rzentry, "Enter Rz Euler angle")

		self.applyButton("Set orientation using Euler angles").pack(side=tk.RIGHT)

	# ----------------------------------------------------------------------
	def fill(self):
		matrix = bmath.Matrix(self._viewer.matrix())
		rxyz = matrix.getEulerRotation()
		self.rxentry.set(FMT%(math.degrees(rxyz[0])))
		self.ryentry.set(FMT%(math.degrees(rxyz[1])))
		self.rzentry.set(FMT%(math.degrees(rxyz[2])))

	# ----------------------------------------------------------------------
	def apply(self, event=None):
		super().apply()
		self._viewer.viewOto0()

		self.rxentry.split((self.ryentry, self.rzentry))
		rx = math.radians(self.rxentry.getfloat())
		ry = math.radians(self.ryentry.getfloat())
		rz = math.radians(self.rzentry.getfloat())

		xyz = self._viewer.origin()
		self._viewer.matrix(bmath.Matrix.eulerRotation(rx,ry,rz))
		self._viewer.origin(xyz)
		self.draw(True, True)

#===============================================================================
class _TransformFrame(_BaseFrame):
	name = "Transform"
	icon = "ROT-DEFI"

	#-----------------------------------------------------------------------
	def __init__(self, master, geometryFrame, *args):
		super().__init__(master, geometryFrame, *args)
		tk.Label(self, text="Rotdefi:").pack(side=tk.LEFT)
		self.transform = tkExtra.Combobox(self, width=10)
		self.transform.pack(side=tk.LEFT,fill=tk.X, expand=tk.YES)
		self.applyButton("Transform using ROT-DEFI").pack(side=tk.RIGHT)

	# ----------------------------------------------------------------------
	def fill(self):
		if self.geometryFrame.parent.editor.editType() == FlukaCard.REGION:
			rotdefi = self.geometryFrame.parent.editor.editObject.cards[0]["@rotdefi"]
			if rotdefi: self.transform.set(rotdefi)

	# ----------------------------------------------------------------------
	def load(self):
		trans = self.geometryFrame.input().rotationList(True)
		self.transform.fill(trans)

	# ----------------------------------------------------------------------
	def apply(self, event=None):
		super().apply()
		trans = self.geometryFrame.input().getTransformation(str(self.transform.get()))
		self._viewer.viewOto0()
		matrix = bmath.Matrix(self._viewer.matrix())
		self._viewer.matrix(trans * matrix)
		self.draw(True, True)

#===============================================================================
# Projection Frame
#===============================================================================
class ProjectionFrame(tk.Frame):
	def __init__(self, master, geometryFrame, *args):
		super().__init__(master, *args)
		self.comboButton = tk.Button(self,image=icons.get("axes"),
				text=" \N{BLACK DOWN-POINTING TRIANGLE}",
				compound=tk.LEFT,
				command=self.change,
				padx=0, pady=0)
		self.comboButton.grid(row=0, column=0, sticky=tk.N+tk.EW)
		tkExtra.Balloon.set(self.comboButton, "Origin and projection [o]")

		#b = tk.Button(self, text="Apply",
		#		image=icons.get("ok"),
		#		compound=tk.LEFT,
		#		command=self.apply)
		#b.grid(row=0, column=2, padx=2, sticky=tk.EW)
		#tkExtra.Balloon.set(b, "Apply operation")

		self._frames = []
		self._frames.append(_NavigationFrame(self,geometryFrame))
		self._frames.append(_OriginFrame(self,    geometryFrame))
		self._frames.append(_BasisFrame(self,     geometryFrame))
		self._frames.append(_ExtendsFrame(self,   geometryFrame))
		self._frames.append(_MoveFrame(self,      geometryFrame))
		self._frames.append(_EulerFrame(self,     geometryFrame))
		self._frames.append(_TransformFrame(self, geometryFrame))
		self.lastFrame = None
		self.show(self._frames[0])

		self.grid_columnconfigure(1, weight=1)

	# ----------------------------------------------------------------------
	def change(self):
		menu = tk.Menu(self, tearoff=0)
		for frame in self._frames:
			menu.add_command(label=frame.name,
					image=icons.get(frame.icon),
					compound=tk.LEFT,
					command=lambda s=self,f=frame:s.show(f))
		menu.tk_popup(self.comboButton.winfo_rootx(),
			      self.comboButton.winfo_rooty() + self.comboButton.winfo_height())

	# ----------------------------------------------------------------------
	def show(self, frame):
		if self.lastFrame: self.lastFrame.grid_forget()
		self.comboButton["image"] = icons.get(frame.icon)
		frame.grid(row=0, column=1, sticky=tk.EW)
		frame.fill()
		self.lastFrame = frame

	# ----------------------------------------------------------------------
	def load(self):
		for frame in self._frames:
			frame.load()

	# ----------------------------------------------------------------------
	def apply(self, event=None):
		self.lastFrame.apply()
