#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Feb-2007

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import tkinter as tk
from tkinter import ttk

import icons
import Manual
#import Ribbon
import tkFlair
import tkExtra
import InputPage
import bFileDialog
import tkTextEditor

from deck  import CardInfo
from fluka import FlukaCard, FlukaCardInfo, FlukaParticle
from material import Material

#===============================================================================
# Open filename dialog
#===============================================================================
filename = ""

# ------------------------------------------------------------------------------
def openOldFile(master, fn="", ft=None):
	global filename
	if ft is None: ft=("All","*")
	filename = bFileDialog.askopenfilename(master=master,
				title="Open file ReadOnly",
				initialfile=fn,
				filetypes=ft)
	return filename

# ------------------------------------------------------------------------------
def openNewFile(master, fn="", ft=None):
	global filename
	if ft is None: ft=("All","*")
	filename = bFileDialog.asksaveasfilename(master=master,
				title="Open file Read/Write",
				initialfile=fn,
				filetypes=ft)
	return filename

# ------------------------------------------------------------------------------
def openFile(master, fn="", ft=None):
	global filename
	if ft is None: ft=("All","*")
	filename = bFileDialog.askfilename(master=master,
				title="Open file Read/Write",
				initialfile=fn,
				filetypes=ft)
	return filename

#===============================================================================
# Error Dialog
#===============================================================================
class ErrorDialog(tk.Toplevel):
	def __init__(self, master, card, errmsg):
		super().__init__(master)
		self.transient(master)
		self.title("Card Validation Errors")

		from InputPage import _FIXED_FONT

		lbl = tk.Label(self, text=FlukaCard.SCALE,
				font=_FIXED_FONT,
				justify=tk.LEFT,
				anchor=tk.NW,
				relief=tk.SUNKEN)
		lbl.pack(side=tk.TOP, fill=tk.X, pady=3)

		msg = tk.Label(self, text=errmsg,
				justify=tk.LEFT,
				anchor=tk.NW,
				relief=tk.SUNKEN)
		msg.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, pady=3)

		b = tk.Button(self, text="Close", command=self.close)
		b.pack(side=tk.BOTTOM)
		self.bind("<Escape>",   self.close)
		self.bind("<Return>",   self.close)
		self.bind("<KP_Enter>", self.close)

		self.deiconify()
		self.grab_set()
		self.wait_window()

	# ----------------------------------------------------------------------
	def close(self, event=None):
		self.destroy()

#===============================================================================
# Select List Frame
#===============================================================================
class SelectionListFrame(tk.Frame):
	def __init__(self, master, *kw, **kwargs):
		super().__init__(master,  *kw, **kwargs)

		# Create the listbox inside the dropdown window
		sb = ttk.Scrollbar(self, takefocus=0)
		self._listbox = tkExtra.SearchListbox(self,
					width=10,
					yscrollcommand=sb.set,
					selectmode=tk.BROWSE,
					foreground=InputPage._VALUE_COLOR,
					font=InputPage._VALUE_FONT)
		sb.config(command=self._listbox.yview)

		self._listbox.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		sb.pack(side=tk.RIGHT, fill=tk.Y)

		self._listbox.bind("<ButtonRelease-1>",	self._buttonRelease)

	# ----------------------------------------------------------------------
	def resetSearch(self):
		tkExtra.SearchBox.reset()

	# ----------------------------------------------------------------------
	def filter(self, search=None, refill=False):
		self._listbox.filter(search, refill)

	# ----------------------------------------------------------------------
	def isPosted(self):
		return self.winfo_ismapped()

	# ----------------------------------------------------------------------
	def up(self):
		self._listbox.event_generate("<<PrevLine>>")
		self.selectActive()

	# ----------------------------------------------------------------------
	def down(self):
		self._listbox.event_generate("<<NextLine>>")
		self.selectActive()

	# ----------------------------------------------------------------------
	def prior(self):
		self._listbox.yview(tk.SCROLL, -1, tk.PAGES)
		self._listbox.activate("@0,0")
		self.selectActive()

	# ----------------------------------------------------------------------
	def next(self):
		self._listbox.yview(tk.SCROLL, 1, tk.PAGES)
		self._listbox.activate(f"@0,{self._listbox.winfo_height()}")
		self.selectActive()

	# ----------------------------------------------------------------------
	def fill(self, lst, sel):
		# Populate listbox
		self._list = lst if lst else []
		self._listbox.delete(0, tk.END)
		tkExtra.SearchBox.reset()

		# Find best match
		selected = False

		# Convert selection to int if possible
		try:
			sf = float(sel)
			si = int(sf)
			if sf == float(si): sel = si
		except ValueError:
			pass

		# FIXME only for FLUKA mode
		try:
			sel8 = sel[:8]
		except TypeError:
			sel8 = None

		# Fill list
		for i,item in enumerate(self._list):
			if isinstance(item, tuple):
				n, v = item
			else:
				n = v = item
			self._listbox.insert(tk.END, str(n))
			if v in (sel,sel8):
				self.selectionSet(i)
				selected = True

		if not selected:
			self._listbox.see(0)

	# ----------------------------------------------------------------------
	# listbox handling methods
	# ----------------------------------------------------------------------
	def post(self, x, y, w, h):
		if self.winfo_ismapped(): return
		self._listbox.config(	font=InputPage._VALUE_FONT,
					foreground=InputPage._VALUE_COLOR)
		self.place(x=x, y=y, width=w, height=h)

	# ----------------------------------------------------------------------
	def unpost(self, event=None):
		if not self.winfo_ismapped(): return
		self.place_forget()

	# ----------------------------------------------------------------------
	def _selectUnpost(self, event=None):
		cur = self._listbox.index(tk.ACTIVE)
		self.selectionClear()
		self.selectionSet(cur)
		self.unpost()

	# ----------------------------------------------------------------------
	def _buttonRelease(self, event):
		self.selectionClear()
		i = self._listbox.nearest(event.y)
		self._listbox.activate(i)
		self._listbox.see(i)
		self._selectUnpost(event)
		self.event_generate("<<Selected>>")

	# ----------------------------------------------------------------------
	# return selected item
	# ----------------------------------------------------------------------
	def selection(self):
		sel = self._listbox.curselection()
		if not sel: return (None, None)
		item = self._list[int(sel[0])]
		if isinstance(item, tuple):
			return item
		else:
			return (item, item)

	# ----------------------------------------------------------------------
	def unselect(self):
		self.selectionClear()

	# ----------------------------------------------------------------------
	def selectionClear(self):
		self._listbox.selection_clear(0,tk.END)
		for i in range(0,self._listbox.size()):
			self._listbox.itemconfig(i, background="")

	# ----------------------------------------------------------------------
	def selectionSet(self, i):
		self._listbox.selection_set(i)
		self._listbox.activate(i)
		try:
			self._listbox.itemconfig(i,
				background=self._listbox.cget("selectbackground"))
			self._listbox.see(i)
		except tk.TclError:
			pass

	# ----------------------------------------------------------------------
	def select(self, txt):
		self.selectionClear()
		for i,item in enumerate(self._list):
			if isinstance(item, tuple):
				n,_ = item
			else:
				n = item
			if txt==n:
				self.selectionSet(i)
				break

	# ----------------------------------------------------------------------
	def selectActive(self):
		self.selectionClear()
		self.selectionSet(tk.ACTIVE)

	# ----------------------------------------------------------------------
	# return width & height of the frame
	# ----------------------------------------------------------------------
	def width(self):  return self.winfo_reqwidth()
	def height(self): return self.winfo_reqheight()

#===============================================================================
# Select groups dialog
#===============================================================================
class SelectGroupsDialog(tk.Toplevel):
	# ----------------------------------------------------------------------
	def __init__(self, master, inp):
		super().__init__(master)
		self.transient(master)
		self.title("Select cards to append")

		# --- Categories ---
		frame = tk.LabelFrame(self, text="Categories to Append")
		frame.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

		self.categoryList = tkExtra.ExListbox(frame, borderwidth=0,
			selectmode=tk.EXTENDED,
			selectborderwidth=0,
			background="White",
			takefocus=True,
			exportselection=tk.FALSE)
		self.categoryList.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

		sb = ttk.Scrollbar(frame, orient=tk.VERTICAL,
				command=self.categoryList.yview)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		self.categoryList.config(yscrollcommand=sb.set)

		# --- Buttons ---
		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM)
		b = tk.Button(frame, text="Ok",
				image=icons.get("ok"),
				compound=tk.LEFT,
				command=self.ok)
		b.pack(side=tk.LEFT)
		b = tk.Button(frame, text="Cancel",
				image=icons.get("x"),
				compound=tk.LEFT,
				command=self.cancel)
		b.pack(side=tk.RIGHT)

		# Fill categories
		self.categoryList.insert(tk.END, tkFlair._ALL_GROUP)
		for group in CardInfo.groups:
			# Check if category is not empty
			cards = [ x for x in FlukaCardInfo._db.values()
					if group in x.group ]
			empty = True
			for c in cards:
				if c.tag in inp.cards:
					empty = False
					break
			if not empty:
				self.categoryList.insert(tk.END, group)

		self.categoryList.selection_set(0)
		self.categoryList.activate(0)

		self.groups = None	# Empty list is all, None is none

	# ----------------------------------------------------------------------
	def show(self):
		self.wait_window()
		return self.groups

	# ----------------------------------------------------------------------
	def ok(self):
		items = list(map(int, self.categoryList.curselection()))
		if not items or items[0] == tkFlair._ALL_GROUP:
			self.groups = []
		else:
			self.groups = [self.categoryList.get(i) for i in items]
		self.destroy()

	# ----------------------------------------------------------------------
	def cancel(self):
		self.groups = None
		self.destroy()

#===============================================================================
# Manual entry of card
#===============================================================================
class EditCardDialog(tk.Toplevel):
	# ----------------------------------------------------------------------
	def __init__(self, editor, flair, card):
		super().__init__(editor)
		self.transient(editor)
		self.card   = card
		self.editor = editor
		self.flair  = flair

		self.title("Edit Card")

		# --- Card Frame ---
		frame = tk.Frame(self, relief=tk.GROOVE, padx=3, pady=3)
		frame.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=3, pady=3)

		# ---
		row,col = 0, 0
		l = tk.Label(frame, text="Comment:")
		l.grid(row=row, column=col, sticky=tk.NE)

		col += 1
		self.comment = tkTextEditor.TextEditor(frame, background="white")
		self.comment.txt.config(height=5, foreground=InputPage._COMMENT_COLOR)
		self.comment.grid(row=row, column=col, columnspan=8, sticky=tk.NSEW)

		# ---
		row,col = row+1, 0
		l = tk.Label(frame, text="Card:")
		l.grid(row=row, column=col, sticky=tk.E)

		col += 1
		self.tag = tk.Label(frame, text=self.card.tag, anchor=tk.W,
				relief=tk.SUNKEN,
				foreground=InputPage._TAG_COLOR,
				background=tkFlair._ELABEL_BACKGROUND_COLOR)
		self.tag.bind('<Button-1>', self.tagButton)
		self.tag.bind('<Button-3>', self.tagButton)
		self.tag.grid(row=row, column=col, sticky=tk.EW)

		col += 1
		b = tk.Button(frame, text="\N{BLACK DOWN-POINTING TRIANGLE}", padx=1, pady=0,
				command=self.tagButton)
		b.grid(row=row, column=col)

		# -
		col += 1
		l = tk.Label(frame, text="Lines:")
		l.grid(row=row, column=col, sticky=tk.E)

		col += 1
		self.linesVar = tk.IntVar()
		lines = list(range(1,11))
		self.linesVar.trace("w", self.linesChange)
		l = tk.OptionMenu(frame, self.linesVar, *lines)
		l.grid(row=row, column=col, columnspan=2, sticky=tk.W)
		col += 1

		# -
		self.what = []

		col += 1
		l = tk.Label(frame, text="name:")
		l.grid(row=row, column=col, sticky=tk.E)

		col += 1
		cmd = lambda e=None,s=self,w=0:s.whatButton(w)
		e = tk.Entry(frame,
				foreground=InputPage._VALUE_COLOR,
				background="White")
		e.bind('<Button-3>', cmd)
		e.grid(row=row, column=col, sticky=tk.EW)

		col += 1
		b = tk.Button(frame, text="\N{BLACK DOWN-POINTING TRIANGLE}", padx=1, pady=0,
				command=cmd)
		b.grid(row=row, column=col)

		self.what.append((l, e, b, row, col-2))

		# --- Whats ---
		w = 0
		ln = 0
		for i in range(10):
			ln += 1
			for j in range(6):
				if j % 3 == 0:
					row,col = row+1, 0
				else:
					col += 1
				w += 1
				if w < 7:
					lbl = f"{j+1}:"
				else:
					lbl = f"{ln}-{j+1}:"
				l = tk.Label(frame, text=lbl)
				l.grid(row=row, column=col, sticky=tk.E)

				col += 1
				cmd = lambda e=None,s=self,w=w:s.whatButton(w)

				e = tk.Entry(frame,
						foreground=InputPage._VALUE_COLOR,
						background="White")
				e.bind('<Button-3>', cmd)
				e.grid(row=row, column=col, sticky=tk.EW)

				col += 1
				b = tk.Button(frame, text="\N{BLACK DOWN-POINTING TRIANGLE}",
						padx=1, pady=0,
						command=cmd)
				b.grid(row=row, column=col)

				self.what.append((l, e, b, row, col-2))

		# ---
		row, col = row+1, 0
		l = tk.Label(frame, text="Text:")
		l.grid(row=row, column=col, sticky=tk.NE)

		col += 1
		self.text = tkTextEditor.TextEditor(frame, background="white")
		self.text.txt.config(height=6, foreground=InputPage._VALUE_COLOR)
		self.text.grid(row=row, column=col, columnspan=8, sticky=tk.NSEW)
		tkExtra.Balloon.set(self.text,
			"Extra text information like title or region description of card")

		# ---
		row, col = row+1, 0
		l = tk.Label(frame, text="Properties:")
		l.grid(row=row, column=col, sticky=tk.NE)

		col += 1
		self.properties = tkTextEditor.TextEditor(frame, background="white")
		self.properties.txt.config(height=6, foreground=InputPage._VALUE_COLOR)
		self.properties.grid(row=row, column=col, columnspan=8, sticky=tk.NSEW)
		tkExtra.Balloon.set(self.properties,
			"Additional variables/properties associated with card")

		frame.grid_rowconfigure(0, weight=1)
		frame.grid_rowconfigure(row, weight=1)
		frame.grid_columnconfigure(1, weight=1)
		frame.grid_columnconfigure(4, weight=1)
		frame.grid_columnconfigure(7, weight=1)

		# --- Buttons Frame ---
		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, fill=tk.X)

		b = tk.Button(frame, text="Help",
				image=icons.get("info"),
				compound=tk.LEFT,
				command=self.help)
		b.pack(side=tk.RIGHT)
		b = tk.Button(frame, text="Cancel",
				image=icons.get("x"),
				compound=tk.LEFT,
				command=self.cancel)
		b.pack(side=tk.RIGHT)
		b = tk.Button(frame, text="Ok",
				image=icons.get("ok"),
				compound=tk.LEFT,
				command=self.ok)
		b.pack(side=tk.RIGHT)
		self.rc = False

		self.bind('<Escape>',	self.cancel)
		self.bind('<<Help>>',	self.help)

	# ----------------------------------------------------------------------
	def fill(self):
		lines, rem = divmod(self.card.nwhats()-1, 6)
		if rem>0: lines += 1
		lines = max(1, lines)

		# fill comment
		self.comment.txt.delete(0.0, tk.END)
		self.comment.txt.insert(0.0, self.card.comment())

		# set tag
		self.tag.config(text=self.card.tag)

		# line number
		self.linesVar.set(lines)

		# set whats
		w = 0
		for l,e,b,r,c in self.what[:lines*6+1]:
			e.delete(0, tk.END)
			e.insert(0, self.card.what(w))
			w += 1

		# fill the extra text
		self.text.txt.delete(0.0, tk.END)
		self.text.txt.insert(0.0, self.card.text())

		# fill the properties
		self.properties.txt.delete(0.0, tk.END)
		if self.card.property:
			props = [f"{n}={v}"
					for n,v in self.card.property.items()
						if n[0]!="@"]
			self.properties.txt.insert(0.0, "\n".join(props))

	# ----------------------------------------------------------------------
	def linesChange(self, *abc):
		lines = self.linesVar.get()
		# display whats
		for l,e,b,r,c in self.what[:lines*6+1]:
			l.grid(row=r, column=c, sticky=tk.E)
			e.grid(row=r, column=c+1, sticky=tk.NSEW)
			b.grid(row=r, column=c+2)

		# remove the rest
		for l,e,b,r,c in self.what[lines*6+1:]:
			l.grid_forget()
			e.grid_forget()
			b.grid_forget()

	# ----------------------------------------------------------------------
	def show(self):
		self.fill()
		self.grab_set()
		self.wait_window()
		return self.rc

	# ----------------------------------------------------------------------
	# Add the information to the card and exit
	# ----------------------------------------------------------------------
	def ok(self, event=None):
		undoinfo = [(self.flair.refreshUndo,),
				self.flair.saveCardUndoInfo(self.card.pos()),
				(self.flair.refreshUndo,)]
		# Collect the data

		# Change the tag if needed!
		if self.card.tag != self.tag["text"]:
			self.editor.input.changeTag(self.card, self.tag["text"])

		# set the comment
		self.card.setComment(self.comment.txt.get(1.0, tk.END).rstrip())

		# Set number of whats
		lines  = self.linesVar.get()

		# Find number of whats
		wend = lines*6
		while wend >= 0:
			l,e,b,r,c = self.what[wend]
			if e.get().strip() != "":
				break
			wend -= 1
		wend += 1

		self.card.setNWhats(wend)

		w = 0
		for l,e,b,r,c in self.what[:wend]:
			self.card.setWhat(w, e.get().strip())
			w += 1

		# Set text field
		self.card.setText(self.text.txt.get("1.0", tk.END).rstrip())

		# Set properties
		prop = self.properties.txt.get("1.0", tk.END).splitlines()
		if prop:
			self.card.clearProperties(False)
			var = value = None
			for line in prop:
				line = line.strip()
				if line:
					try:
						equal = line.index("=")
						if var is not None: self.card[var] = value
						var   = line[:equal].strip()
						value = line[equal+1:].strip()
					except:
						value = line
			if var is not None: self.card[var] = value

		self.flair.addUndo(undoinfo,"Edit card")
		self.rc = True
		self.destroy()
		self.flair.inputModified()

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		self.rc = False
		self.destroy()

	# ----------------------------------------------------------------------
	def help(self, event=None):
		Manual.show(self.card.MODE, self.card.tag)

	# ----------------------------------------------------------------------
	def tagButton(self, event=None):
		l,e,b,c,r = self.what[1]
		x = self.tag.winfo_rootx()
		y = self.tag.winfo_rooty() + self.tag.winfo_height()
		width = self.tag.winfo_width() + b.winfo_width()

		cards = sorted([x for x in FlukaCardInfo._db if x])
		tag = self.tag["text"]
		sel = SelectionListDialog(self).show(x, y, width, cards, tag)
		if sel:
			self.tag.config(text=sel)

	# ----------------------------------------------------------------------
	def whatButton(self, w):
		l,e,b,c,r = self.what[w]
		self.x = b.winfo_rootx()
		self.y = b.winfo_rooty() + b.winfo_height()
		self.w = w
		self.width = e.winfo_width() + b.winfo_width()

		# Show menu first
		menu = tk.Menu(self, tearoff=0)
		menu.add_command(label='#define',  underline=0, command=self.defines)
		menu.add_command(label='Binning',  underline=1, command=self.binnings)
		menu.add_command(label='Body',     underline=0, command=self.bodies)
		menu.add_command(label='Card',     underline=0, command=self.cards)
		menu.add_command(label='Material', underline=0, command=self.materials)
		menu.add_command(label='Particle', underline=0, command=self.particles)
		menu.add_command(label='Region',   underline=0, command=self.regions)
		menu.add_command(label='Rot-defi', underline=1, command=self.rotdefis)
		menu.tk_popup(self.x, self.y)

		self.x = e.winfo_rootx()
		self.y = b.winfo_rooty() + b.winfo_height()

	# ----------------------------------------------------------------------
	def defines(self):
		lst = self.editor.layout[self.card.MODE].defineList()
		self.showList(lst)

	# ----------------------------------------------------------------------
	def binnings(self):
		lst = self.editor.layout[self.card.MODE].binningList()
		self.showList(lst)

	# ----------------------------------------------------------------------
	def cards(self):
		self.showList(sorted([x for x in FlukaCardInfo._db if x]))

	# ----------------------------------------------------------------------
	def bodies(self):
		lst = self.editor.layout[self.card.MODE].bodyList()
		self.showList(lst)

	# ----------------------------------------------------------------------
	def materials(self):
		lst = self.editor.layout[self.card.MODE].materialList()
		self.showList(lst)

	# ----------------------------------------------------------------------
	def particles(self):
		lst = FlukaParticle.list
		self.showList(lst)

	# ----------------------------------------------------------------------
	def regions(self):
		lst = self.editor.layout[self.card.MODE].regionList()
		self.showList(lst)

	# ----------------------------------------------------------------------
	def rotdefis(self):
		lst = self.editor.layout[self.card.MODE].rotationList2()
		self.showList(lst)

	# ----------------------------------------------------------------------
	def showList(self, lst):
		sel = SelectionListDialog(self).show(
				self.x, self.y, self.width, lst,
				self.what[self.w][1].get())
		if sel:
			l,e,b,c,r = self.what[self.w]
			e.delete(0, tk.END)
			e.insert(0, sel)

#===============================================================================
# Selection List Dialog
#===============================================================================
class SelectionListDialog(tk.Toplevel):
	# ----------------------------------------------------------------------
	def __init__(self, master):
		super().__init__(master)
		self.transient(master)
		self.overrideredirect(1)
		self.title("")

		sb = ttk.Scrollbar(self, takefocus=0)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		self._listbox = tkExtra.ExListbox(
					self,
					width=10,
					yscrollcommand=sb.set,
					selectmode=tk.BROWSE,
					takefocus=1,
					#background="White",
					font=InputPage._VALUE_FONT)
		self._listbox.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		sb.config(command=self._listbox.yview)

		self.bind('<Escape>',	self.cancel)
		self.bind('<FocusOut>', self.focusOut)
		self.bind('<Return>',	self.ok)
		self.bind('<KP_Enter>',	self.ok)
		self._listbox.bind('<ButtonRelease-1>',	self.ok)
		self._listbox.bind('<Double-1>',	self.ok)
		self._listbox.bind('<Button-3>',	self.cancel)

	# ----------------------------------------------------------------------
	def show(self, x, y, w, lst, val):
		self.ret = None
		self._grab_window = self.grab_current()
		if self._grab_window is not None:
			self._grab_window.grab_release()

		self._listbox.delete(0, tk.END)
		for item in lst:
			self._listbox.insert(tk.END, item)
			if val == item:
				self._listbox.selection_set(tk.END)
				self._listbox.activate(tk.END)
				self._listbox.see(tk.END)

		h = 200
		self.wm_geometry(f"{w}x{h}+{x}+{y}")
		self.deiconify()
		self._listbox.focus_set()
		self.wait_window()
		try:
			if self._grab_window is not None:
				self._grab_window.grab_set()
			self.master.focus_set()
		except tk.TclError:
			pass
		return self.ret

	# ----------------------------------------------------------------------
	# Add the information to the card and exit
	# ----------------------------------------------------------------------
	def ok(self, event=None):
		try:
			sel = int(self._listbox.curselection()[0])
			self._listbox.activate(sel)
		except:
			pass
		self.ret = self._listbox.get(tk.ACTIVE)
		self.destroy()

	# ----------------------------------------------------------------------
	def focusOut(self, event=None):
		focus = self.focus_get()
		if focus is None: return
		self.cancel()

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		self.destroy()

# ------------------------------------------------------------------------------
# Filter menu
# ------------------------------------------------------------------------------
def filterMenu(page, menu, card, search=None, keepcase=False):
	if search:
		target = search if keepcase else search.upper()

		rc = True
		cmd = lambda f=('*',target),p=page:p.filterOn(f)
		menu.add_command(label="Any: "+search,
				image=icons.get("GLOBAL"), compound=tk.LEFT,
				command=cmd)
		cmd = lambda f=('body',target),p=page:p.filterOn(f)
		menu.add_command(label="Body: "+search,
				image=icons.get("TRC"), compound=tk.LEFT,
				command=cmd)
		cmd = lambda f=('region',target),p=page:p.filterOn(f)
		menu.add_command(label="Region: "+search,
				image=icons.get("REGION"), compound=tk.LEFT,
				command=cmd)
		cmd = lambda f=('solid',target),p=page:p.filterOn(f)
		menu.add_command(label="Solid: "+search,
				image=icons.get("TRC"), compound=tk.LEFT,
				command=cmd)
		cmd = lambda f=('volume',target),p=page:p.filterOn(f)
		menu.add_command(label="Volume: "+search,
				image=icons.get("REGION"), compound=tk.LEFT,
				command=cmd)
		cmd = lambda f=('material',target),p=page:p.filterOn(f)
		menu.add_command(label="Material: "+search,
				image=icons.get("MATERIAL"), compound=tk.LEFT,
				command=cmd)
		cmd = lambda f=('transform',target),p=page:p.filterOn(f)
		menu.add_command(label="Transformation: "+search,
				image=icons.get("ROT-DEFI"), compound=tk.LEFT,
				command=cmd)
		cmd = lambda f=('detector',target),p=page:p.filterOn(f)
		menu.add_command(label="Detector: "+search,
				image=icons.get("USRBIN"), compound=tk.LEFT,
			command=cmd)

		if target in FlukaParticle.particles:
			cmd = lambda f=('particle',target),p=page:p.filterOn(f)
			menu.add_command(
				label="FlukaParticle: "+search,
				image=icons.get("particle"), compound=tk.LEFT,
				command=cmd)

		try:
			target = str(int(target))
			cmd = lambda f=('lunit',target),p=page:p.filterOn(f)
			menu.add_command(
				label="Unit: "+search,
				image=icons.get("new"), compound=tk.LEFT,
				command=cmd)
		except ValueError:
			pass
	else:
		rc = False

	if card is None:
		pass

	elif card.tag[0] == "#":
		for w in card.whats():
			if w != "":
				cmd = lambda f=('*',w),p=page:p.filterOn(f)
				menu.add_command(label="Any: "+str(w),
						image=icons.get("GLOBAL"), compound=tk.LEFT,
						command=cmd)
	else:
		case = card.case()
		if case is None: return
		range_ = card.info.range[case]
		if card.tag in FlukaCardInfo.DETECTOR_TAGS:
			w = card.name()
			if w!="":
				cmd = lambda f=('detector',w),p=page:p.filterOn(f)
				if rc: menu.add_separator()
				menu.add_command(label="Detector: "+str(w),
					image=icons.get("USRBIN"), compound=tk.LEFT,
					command=cmd)
				rc = False

		prev = ""
		for i,r in enumerate(range_):
			w = card.what(i)
			if w=="": continue
			if isinstance(w,str) and w[0]=="-": w=w[1:]

			# strip leading minus if any
			if isinstance(w,str) and w[0]=='-': w = w[1:]
			m = CardInfo._PATDEF.match(r)
			if m: r = m.group(3)		# remove definitions
			if r[-1]=="-":  r=r[:-1]	# remove the - at the end

			def addSeparator(r):
				nonlocal rc, prev, menu
				if rc and prev!=r:
					menu.add_separator()
				prev = r
				rc = False

			search = {"body"      : "TRC",
				  "solid"     : "TRC",
				  "region"    : "REGION",
				  "volume"    : "REGION",
				  "material"  : "MATERIAL",
				  "detector"  : "DETECT",
				  "binning"   : "USRBIN",
				  "particle"  : "particle",
				  "transform" : "ROT-DEFI",
				  "lunit"     : "new"}

			try:
				ws = w if keepcase else w.upper()
			except:
				ws = str(w)

			if r in search:
				addSeparator(r)
				menu.add_command(
					label=f"{r.capitalize()}: {w}",
					image=icons.get(search[r]),
					compound=tk.LEFT,
					command=lambda f=(r,ws),p=page:p.filterOn(f))
				rc = True

	# Filter invalid cards
	if rc: menu.add_separator()
	menu.add_command(label="Show Invalid Cards",
			image=icons.get("debug"),
			compound=tk.LEFT,
			command=page.showInvalid)

	if InputPage._SHOW_PREPROCESSOR:
		icon = icons.get("undef")
		txt = "Hide preprocessor"
	else:
		icon = icons.get("define")
		txt = "Show preprocessor"

	menu.add_command(label=txt,
			image=icon,
			compound=tk.LEFT,
			command=page.togglePreprocessor)
	return rc

# ==============================================================================
# List popup toplevel
# ==============================================================================
class ListPopup(tk.Toplevel):
	selected = None
	shown    = None

	# ----------------------------------------------------------------------
	def __init__(self, master, **kw):
		if ListPopup.shown: return
		super().__init__(master, **kw)
		self.overrideredirect(1)
		self.transient(master)

		sb = ttk.Scrollbar(self, takefocus=0)
		self._listbox = tkExtra.SearchListbox(self,
					selectmode=tk.BROWSE,
					yscrollcommand=sb.set,
					#font=Ribbon._FONT,
					width=30,
					height=20,
					exportselection=tk.FALSE,
					takefocus=tk.FALSE)
		self._listbox.pack(side=tk.LEFT, fill=tk.BOTH)
		sb.config(command=self._listbox.yview)
		sb.pack(side=tk.RIGHT, fill=tk.Y)

		# Bind events to the dropdown window.
		self._listbox.bind('<Escape>',		self.cancel)
		self._listbox.bind('<KeyRelease-Up>',	self.key)
		self._listbox.bind('<KeyRelease-Down>',	self.key)
		self._listbox.bind('<KeyRelease-Return>',self.ok)
		self._listbox.bind('<KeyRelease-KP_Enter>',self.ok)
		self._listbox.bind('<ButtonRelease-1>',	self.ok)
		self._listbox.bind('<FocusOut>',		self.focusOut)
		self._listbox.bind('<Motion>',		self.motion)
		self._lasty = -1
		tkExtra.Balloon.set(self._listbox, "Select item")

	# ----------------------------------------------------------------------
	def fill(self, pattern):
		pass

	# ----------------------------------------------------------------------
	def show(self, x, y):
		if ListPopup.shown:
			ListPopup.shown.destroy()
			ListPopup.shown = None
			return
		try:
			self._focus = self.focus_get()
		except:
			self._focus = None
		ListPopup.shown = self
		self.deiconify()
		self.geometry(f"+{x}+{y}")
		self.lift()
		self.focus_set()
		self._listbox.focus_set()
		self.wait_window()
		ListPopup.shown = None
		return ListPopup.selected

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		sel = self._listbox.curselection()
		if sel:
			ListPopup.selected = self._listbox.get(sel[0])
		tkExtra.SearchBox.reset()
		if self._focus: self._focus.focus_set()
		self.destroy()
		return "break"

	# ----------------------------------------------------------------------
	def focusOut(self, event=None):
		self._focus = self.focus_get()
		self.cancel()

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		ListPopup.selected = None
		tkExtra.SearchBox.reset()
		if self._focus: self._focus.focus_set()
		self.destroy()

	# ----------------------------------------------------------------------
	def motion(self, event=None):
		self.showtooltip(self._listbox.nearest(event.y))

	# ----------------------------------------------------------------------
	def key(self, event=None):
		self.showtooltip(self._listbox.index(tk.ACTIVE))

	# ----------------------------------------------------------------------
	def select(self, item):
		ListPopup.selected = None
		self._listbox.selection_clear(0,tk.END)
		self._listbox.selection_set(item)
		self._listbox.activate(item)

	# ----------------------------------------------------------------------
	def showtooltip(self, y):
		if y == self._lasty: return
		self._lasty = y
		tkExtra.Balloon.sethelp(self._listbox, self.tooltip(y))
		tkExtra.Balloon.show()

	# ----------------------------------------------------------------------
	def tooltip(self, y):
		return f"Item={y} {self._listbox.get(y)}"

# ==============================================================================
# Material Selection Popup
# ==============================================================================
class MaterialList(ListPopup):
	# ----------------------------------------------------------------------
	def fill(self):
		sel = 0
		for m in sorted(Material.materials):
			self._listbox.insert(tk.END, m)
			if ListPopup.selected == m:
				sel = self._listbox.size()-1
		self.select(sel)

	# ----------------------------------------------------------------------
	def tooltip(self, y):
		item = self._listbox.get(y)
		try:
			return f"{item}:\n{Material.materials[item].notes}"
		except KeyError:
			return None

# ==============================================================================
# Add search card popup
# ==============================================================================
class AddSearchList(ListPopup):
	# ----------------------------------------------------------------------
	def fill(self, mode, pattern):
		self.mode = mode
		patterns  = [x.upper() for x in pattern.split()]
		cards     = [x for x in CardInfo.cards(mode) if x.find(patterns)]
		CardInfo.sort(cards)
		sel = 0
		for card in cards:
			self._listbox.insert(tk.END, card.tag)
			if ListPopup.selected == card.tag:
				sel = self._listbox.size()-1
		self.select(sel)

	# ----------------------------------------------------------------------
	def tooltip(self, y):
		item = self._listbox.get(y)
		return f"{item}:\n{CardInfo.get(self.mode,item).meaning}"
