#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
#	  Andre.Donadon.Servelle@cern.ch
# Date:   05-Sep-2020

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import csv
import math
import itertools

try:
	import matplotlib
except ImportError:
	matplotlib = None

try:
	import numpy as np
except ImportError:
	np = None

#import log
#import Data
import Plot
import bmath
import Matplotlib
from palette import rndColor, RGB2str, paletteStore

#===============================================================================
# ----------------------------------------------------------------------
# Base class to start plotting (start matplotlib engine)
# ----------------------------------------------------------------------
def start():
	engine = Matplotlib.Matplotlib()
	while 20:	# wait a while to get version
		try:
			ver = engine.version()
		except:
			return None
		if ver is not None: break
	return engine

#===============================================================================
# MatPlotlib Plot class
#===============================================================================
class MPPlot(Plot.Plot):
	#-----------------------------------------------------------------------
	#def __init__(self, engine, project, plot, log=None):
	#	super().__init__(engine, project, plot, log)
	#	#self.engine.setLogger(log)

	# ----------------------------------------------------------------------
	def __call__(self, s):
		self.write(s)
		self.plot.write(s)

	# ----------------------------------------------------------------------
	# Save plot to file
	# ----------------------------------------------------------------------
	def save(self, ext):
		pass

	# ----------------------------------------------------------------------
	# Send plot on printer
	# ----------------------------------------------------------------------
	def hardcopy(self, printer):
		pass

	# ----------------------------------------------------------------------
	def _startPlot(self):
		self.canvas = self.engine.tkdisplay(self.plot('index'),
						self.windowtitle())
		#self(tkFlair._PLOT_COMMANDS)
		super()._startPlot()

		#self("set key default")
		#self(f"plotname='{self.plot.name}'")

	# ----------------------------------------------------------------------
	def show(self):
		super().show()
		#self.plotCommands()
		self.canvas.draw()
		self.engine.show(False)
		#self._plotEnd()

	# ----------------------------------------------------------------------
	# @return plot options for prefix: title, x, y, ...
	# ----------------------------------------------------------------------
	def _plotOptions(self, prefix):
		opts = {"color" : "black"}

		def setopt(optname, name, default=None):
			nonlocal prefix, opts
			value = self.plot(f"{prefix}{name}")
			if value:
				opts[optname] = value
			elif default is not None:
				opts[optname] = default

		setopt("fontname", "font")
		setopt("fontsize", "size")
		setopt("color",    "color")
		return opts

		#font  = self.plot(f"{prefix}font")
		#size  = self.plot(f"{prefix}size")
		#color = self.plot(f"{prefix}color")
		#if font or size:
		#	opts = f"font '{font},{size}'"
		#if color:
		#	opts += f" tc rgb '{color}'"
		#return opts

	# ----------------------------------------------------------------------
	# Base class to start plotting (Launches engine)
	# ----------------------------------------------------------------------
	#def _plotEnd(self):
	#	self.plotCommands(False)
	#	return self.checkOutput()

	# ----------------------------------------------------------------------
	#def checkOutput(self):
	#	self.engine.writeline('print "end"')
	#	# wait until we get the line end
	#	line  = "\n"
	#	END   = "end\n"
	#	err   = None
	#	tstart = time.time()
	#	while line!=END and time.time()-tstart < 10.0:	# 10s
	#		while self.engine.check(1.0):
	#			line = self.engine.readline()
	#			if line not in ("\n", END):
	#				if "Warning" in line:
	#					self.log(f"{log.WARNING_PREFIX} {line}")
	#				else:
	#					self.log(f"{log.ERROR_PREFIX} {line}")
	#				if err is None:
	#					err = "Matplot error or warning"
	#	return err

	# ----------------------------------------------------------------------
	# @return color band and limits
	# ----------------------------------------------------------------------
	def colorBand(self):
		try:
			ncolors = int(self.plot("cbcolors"))
		except ValueError:
			ncolors = 30
		palette = self.plot("cbpalette")
		try:
			clist = paletteStore[palette].resize(ncolors)
		except KeyError:
			clist = paletteStore["fluka"].resize(ncolors)
		return matplotlib.colors.ListedColormap(clist)

	# ----------------------------------------------------------------------
	# Set normalization factor operated on a column and/or expression
	#	norm=""           no normalization is assumed return column
	#	norm=expr->float  if it evaluates to float return (norm)*(column)
	#	norm=expr-of-$xxx expression with columns $xxx return norm as is
	#	norm=func(x)      set a gnuplot function func(x)=norm and use the
	#                         function(column)
	# ----------------------------------------------------------------------
	def plotNorm(self, norm, column, mult="", axis="y"):
		pass

	# ----------------------------------------------------------------------
	# Plot commands
	# ----------------------------------------------------------------------
	def plotCommands(self, prefix=True):
		first = True
		pline = ""
		self.plotappend = ""
		for line in self.plot("commands").splitlines():
			line = line.strip()
			if pline != "":
				line = pline + line
				pline = ""
			if line!="" and line[-1] == "\\":
				pline = line[:-1]
				continue

			words = line.split()
			if not words: continue
			# stop on the first plot command
			if words[0] in ["plot", "replot", "splot"]:
				if words[0] == "plot":
					self.plotappend = line[5:]
				if prefix: break
				if words[0] != "plot":
					self(line+"\n")
				first = False
			elif prefix == first:
				self(line+"\n")

	# ----------------------------------------------------------------------
	# Generate the plot
	# ----------------------------------------------------------------------
	def _geometryPlot(self):
		err = super()._geometryPlot()
		if err is not None: return err
		# Important variables
		kind = self.plot("kind")
		cd = self.plot("coord")
		O  = bmath.Vector(map(float,self.plot("origin").split()))
		U  = bmath.Vector(map(float,self.plot("basisu").split()))
		V  = bmath.Vector(map(float,self.plot("basisv").split()))
		extU,extV = map(float, self.plot("extends").split())
		loP  = -extU*U -extV*V + O
		hiP  =  extU*U +extV*V + O
		#Us = Vs = ""		# Sign of field operation
		ui = vi = 0		# coordinate index
		if cd == Plot.COORD[0]:
			ui = 0
			vi = 1
			minx  = -extU
			maxx  =  extU
			miny  = -extV
			maxy  =  extV
		elif cd == Plot.COORD[1]:
			ui = 1
			vi = 0
			minx  = -extV
			maxx  =  extV
			miny  = -extU
			maxy  =  extU
		else:
			ui = ord(cd[ 0])-ord('X')
			vi = ord(cd[-1])-ord('X')
			minx  = loP[ui]
			maxx  = hiP[ui]
			miny  = loP[vi]
			maxy  = hiP[vi]
			ui += 2
			vi += 2
			#if U.direction()[0]=="-": Us = "-"
			#if V.direction()[0]=="-": Vs = "-"

		for axis in ['x','y']:
			amin = self.plot(f"{axis}min")
			amax = self.plot(f"{axis}max")
			if (amin and amax):
				self.engine.range(axis, amin, amax)
			elif amin and axis == 'x':
				self.engine.range(axis, amin, maxx)
			elif amin and axis == 'y':
				self.engine.range(axis, amin, maxy)
			elif amax and axis == 'x':
				self.engine.range(axis, minx, amax)
			elif amax and axis == 'y':
				self.engine.range(axis, miny, amax)
			elif axis == 'x':
				self.engine.range(axis, minx, maxx)
			else:
				self.engine.range(axis, miny, maxy)

		# load data
		nu = int(self.plot('nu'))
		nv = int(self.plot('nv'))
		worms  = []
		uimg   = np.ndarray((nu,nv), dtype=np.float)
		vimg   = np.ndarray((nu,nv), dtype=np.float)
		ximg   = np.ndarray((nu,nv), dtype=np.float)
		yimg   = np.ndarray((nu,nv), dtype=np.float)
		zimg   = np.ndarray((nu,nv), dtype=np.float)
		regimg = np.ndarray((nu,nv), dtype=np.int)
		matimg = np.ndarray((nu,nv), dtype=np.int)
		latimg = np.ndarray((nu,nv), dtype=np.int)
		buimg  = np.ndarray((nu,nv), dtype=np.float)
		bvimg  = np.ndarray((nu,nv), dtype=np.float)
		bimg   = np.ndarray((nu,nv), dtype=np.float)
		#           0     1    2     3     4      5       6        7      8      9     10
		image  = [uimg, vimg, ximg, yimg, zimg, regimg, matimg, latimg, buimg, bvimg, bimg]
		labels  = []
		regions = {}
		with open(f"{self.plot.name}.dat","r") as f:
			u   = []
			v   = []
			i = j = 0
			empty = 0		# empty lines counter
			index = 0		# index counter every after 2 empty lines
			for line in f:
				line = line.strip()
				if not line:
					# empty line
					empty += 1
					if u:
						if index==0:
							worms.append((u,v))
							u = []
							v = []
					if empty == 2:
						index += 1
						empty = 0
					continue
				empty = 0
				if line[0] == "#":
					continue
				if index == 0:
					# Parse worms
					coord = list(map(float, line.split()))
					u.append(coord[ui])
					v.append(coord[vi])
				elif index == 1:
					# parse image 2D array
					words  = line.split()
					uimg[i,j]   = float(words[0])		# u
					vimg[i,j]   = float(words[1])		# v
					ximg[i,j]   = float(words[2])		# x
					yimg[i,j]   = float(words[3])		# y
					zimg[i,j]   = float(words[4])		# z
					regimg[i,j] =   int(words[5])		# reg
					matimg[i,j] =   int(words[6])		# mat
					latimg[i,j] =   int(words[7])		# lat
					buimg[i,j]  = float(words[8])		# bu
					bvimg[i,j]  = float(words[9])		# bv
					bimg[i,j]   = float(words[10])		# b
					i += 1
					if i==nu:
						i = 0
						j += 1
				elif index == 2:
					# parse label position
					words = line.split()
					labels.append(list(map(float, words[:5])) + words[5:])
					regions[int(words[7])] = words[5]

		labelid = None
		plotid  = None
		clist   = None
		vmin    = None
		vmax    = None
		vscale  = None
		if kind == Plot.KIND[1]:	# Region
			labelid = plotid = 5
			clist = ["#ffffff"] * max(regions.keys())
			vmin  = 1
			vmax  = len(clist)+1
			for k,v in regions.items():
				clist[k-1] = RGB2str(rndColor(hash(v)))

		elif kind == Plot.KIND[2]:	# Material
			# FIXME if index x[0] is not sequential e.g. from Moira
			labelid = plotid = 6
			clist = [x[1] for x in self.project.materialPalette(self.plot._materialList)]
			vmin  = 1
			vmax  = len(clist)+1

		elif kind == Plot.KIND[3]:	# Lattice
			#labelid = ???
			plotid = 7
			#vmin = 1
			#vmax = len(clist)+1

		elif kind in (Plot.KIND[4], Plot.KIND[6]):	# Field Intensity & Field
			labelid = 5
			plotid  = 10
			clist   = paletteStore["field"].palette
			vmax    = np.ndarray.max(image[plotid])
			vmax   *= (11./10)			# 10% extra for colorbar
			vscale  = float(self.plot("vectscale"))
		elif kind in Plot.KIND[5]:
			vscale = float(self.plot("vectscale"))
			# plotid = 8

		if plotid is not None:
			if clist is not None:
				cmap = matplotlib.colors.ListedColormap(clist)
				colormesh = self.engine.ax.pcolormesh(
						image[ui], image[vi], image[plotid],
						shading="auto", cmap=cmap,
						vmin=vmin, vmax=vmax)
			else:
				colormesh = self.engine.ax.pcolormesh(
						image[ui], image[vi], image[plotid])
			if vmin is None:
				self.engine.fig.colorbar(colormesh)

		if kind in (Plot.KIND[5], Plot.KIND[6]):	# Field Vector & Field
			vscale = float(self.plot("vectscale"))
			every  = 5
			self.engine.ax.quiver(
					uimg[::every,::every],  vimg[::every,::every],
					buimg[::every,::every], bvimg[::every,::every],
					headwidth=5, headaxislength=5, width=0.002,
					headlength=4, scale = 50./vscale, color='#0072B2')

		# Plot the worms
		for u,v in worms:
			self.engine.ax.plot(u, v, c="black", linewidth=1, alpha=.5)

		# Plot the labels
		if int(self.plot("labels")) and labelid is not None:
			for label in labels:
				self.engine.ax.text(label[ui], label[vi],
						label[labelid],
						horizontalalignment='center')
				self.engine.ax.text(label[ui], label[vi],
						'$\\odot$',
						horizontalalignment='center',
						verticalalignment='top')

	# ----------------------------------------------------------------------
	# Find Index
	# ----------------------------------------------------------------------
	@staticmethod
	def findIndex(x, v, m):
		for i in range(len(x)-1):
			if x[i] < v < x[i+1]:
				# print( "< >" )
				return i if m else i+1
			elif v == x[i]:
				# print( "==" )
				return i
			elif v < x[0]:
				# print( "<" )
				return 0
			elif v >= x[len(x)-1]:
				# print( ">" )
				return len(x)-1

	# ----------------------------------------------------------------------
	# Generate the plot
	# ----------------------------------------------------------------------
	def _usr1DPlot(self):
		# --- Loop over detectors to build plot string ---
		for i in range(int(self.plot("ndetectors"))):
			if not int(self.plot(f"show.{i}")) != 1:
				continue		# Skip graph
			filename = self.plot(f"file.{i}")
			if not filename: continue
			det = int(self.plot(f"det.{i}"))

			detector = np.array([])
			lastline = 0
			with open(filename,'r') as f:
				for num, line in enumerate(f,1):
					lastline += 1
					if 'Detector' in line:
						detector = np.append(detector, num)
			detector = np.append(detector, lastline)

			# load data
			xlow  = np.array([])
			xhigh = np.array([])
			y     = np.array([])
			yerr  = np.array([])

			f = open(filename, 'r')
			for line in itertools.islice(f, int(detector[det]), int(detector[det+1])):
				linesp = line.split()
				if not linesp: continue
				if linesp[0][0]== '#': continue
				xlow  = np.append(xlow,  float(linesp[0]))
				xhigh = np.append(xhigh, float(linesp[1]))
				y     = np.append(y,     float(linesp[2]))
				yerr  = np.append(yerr,  float(linesp[3]))
			f.close()

			xlog = int(self.plot("xlog"))
			style = self.plot(f"with.{i}")
			#addErrors = style == "histerror"	# Additional errors

			# Default using (middle of bin)
			if xlog:
				X = np.sqrt(xlow * xhigh)
			else:
				X = (xlow + xhigh) / 2.0
			EX = (xhigh - xlow ) / 2.0

			if style in ("histogram", "histerror"):
				kwargs = {'drawstyle': 'steps-mid'}
				X = xlow
				usingex = ""
			else:
				kwargs = {}

			yval = int(self.plot(f"y.{i}"))
			if yval == 1:			# <X>*Y
				Y  = X * y
				EY = X * y * yerr / 100.0
			elif yval == 2:			# DX*Y
				Y  = (xhigh - xlow) * y
				EY = (xhigh - xlow) * y * yerr / 100.0

			else:  # Leave default Y
				Y  = y
				EY = y * yerr / 100.0

			pt = self.plot(f"pt.{i}")
			if self.plot(f"fit_type.{i}") != "":
				#pyplot.errorbar(X, Y, yerr=EY, marker=".", xerr=0, **kwargs)
				#if (self.plot(f"pt.{i}")).find("filled"):
				#if pt == 4 or pt == 6 or pt == 8 or pt == 10:
				#	pyplot.errorbar(X, Y, yerr=EY, xerr=0,
				#		color=self.engine.color(self.plot(f"lc.{i}")),
				#		marker=self.engine.marker(self.plot(f"pt.%d"%(i))),
				#		label="gato{i}",
				#		markersize=int(self.plot(f"ps.{i}")),
				#		linewidth=int(self.plot(f"lw.{i}")))
				#pyplot.legend()
				#else:
				#	 pyplot.errorbar(X, Y, yerr=EY, xerr=0,
				#		color=self.engine.color(self.plot(f"lc.{i}")),
				#		mfc='white',
				#		marker=self.engine.marker(self.plot(f"pt.%d"%(i))),
				#		markersize=self.plot("ps.{i}"),
				#		linewidth=self.plot(f"lw.{i}"))
				minx = MPPlot.findIndex(X, float(self.plot(f"fit_min.{i}")), True)
				maxx = MPPlot.findIndex(X, float(self.plot(f"fit_max.{i}")), False)
			style  = self.plot("with.0")
			color  = self.engine.color(self.plot(f"lc.{i}"))
			marker = self.engine.marker(self.plot(f"pt.{i}"))
			markersize = int(self.plot(f"ps.{i}"))*5
			linewidth  = int(self.plot(f"lw.{i}"))
			label = self.plot(f"name.{i}")
			if style=="histerror":
				self.engine.ax.step(X, Y, where='post',
						color = color,
						label = label,
						linewidth=linewidth)
				uu = (X[1:] + X[:-1])/2.0
				self.engine.ax.errorbar(uu, Y[0:-1], yerr=EY[0:-1],
						linewidth       = linewidth,
						marker          = marker,
						markersize      = markersize,
						ecolor          = color,
						markerfacecolor = color,
						markeredgecolor = color,
						fmt             = '.k')
			elif style=="errorbars":
				uu = (X[1:] + X[:-1])/2.0
				self.engine.ax.errorbar(uu, Y[0:-1], yerr=EY[0:-1],
						linewidth=linewidth,
						marker=marker,
						markersize=markersize,
						ecolor = color,
						markerfacecolor = color,
						markeredgecolor = color,
						label = label,
						fmt='.k')
			elif style=="errorlines":
				uu = (X[1:] + X[:-1])/2.0
				self.engine.ax.plot(uu, Y[0:-1], color = color,
						linewidth = linewidth)
				self.engine.ax.errorbar(uu, Y[0:-1], yerr=EY[0:-1],
						linewidth       = linewidth,
						marker          = marker,
						markersize      = markersize,
						ecolor          = color,
						markerfacecolor = color,
						markeredgecolor = color,
						label           = label,
						fmt             = '.k')
			elif style=="histogram":
				self.engine.ax.step(X, Y, where='post', color = color,
						linewidth = linewidth)
			elif style=="lines":
				uu = (X[1:] + X[:-1])/2.0
				self.engine.ax.plot(uu, Y[0:-1], color = color,
						label = label,
						linewidth = linewidth)
			elif style=="linespoints":
				uu = (X[1:] + X[:-1])/2.0
				self.engine.ax.plot(uu, Y[0:-1], color = color,
						linewidth = linewidth)
				self.engine.ax.scatter(uu, Y[0:-1],
						color  = color,
						label  = label,
						marker = marker,
						s      = markersize)
			elif style=="points":
				uu = (X[1:] + X[:-1])/2.0
				self.engine.ax.scatter(uu, Y[0:-1],
						label  = label,
						color  = color,
						marker = marker,
						s      = markersize)
			else:
				uu = (X[1:] + X[:-1])/2.0
				self.engine.ax.errorbar(uu, Y[0:-1],
						xerr   = EX[0:-1],
						yerr   = EY[0:-1],
						ecolor = color,
						label  = label,
						fmt    = '.k')
		self.engine.ax.legend()

	# ----------------------------------------------------------------------
	# USR-2D plot
	# FIXME
	# ----------------------------------------------------------------------
	def _usr2DPlot(self):
		filename = self.plot("datafile")
		det = int(self.plot("det"))

		# load data
		u = np.array([])
		v = np.array([])
		w = np.array([])
		PLOTON = False
		num_det = 0
		with open(filename,'r') as csvfile:
			plots = csv.reader(csvfile)
			for row in plots:
				#print(row)
				if not row: continue
				if "Detector" in row[0]:
					num_det = int(row[0].split()[3])
					if num_det == det: PLOTON = True
				# if row[0].find('Histogram') != -1:  num_lines = int(row[0].split()[2])
				if row[0][0]== '#': continue
				if row[0][0]== '#' and PLOTON: break
				u = np.append(xlow,  float(((row[0]).split())[0]))
				v = np.append(xhigh, float(((row[0]).split())[4]))
				w = np.append(y,     float(((row[0]).split())[2]))
		u = np.unique(u)
		u = np.sort(u)
		v = np.unique(v)
		v = np.sort(v)
		uu, vv = np.meshgrid(u,v)
		B2D = np.reshape(w, (-1, len(u)))
		#pyplot.pcolormesh(uu, vv, B2D, cmap='Greens')
		#pyplot.contour3D(uu, vv, B2D, cmap='hot')
		#pyplot.plot_surface(uu, vv, B2D, cmap='hot')
		#pyplot.colormaps()

	# ----------------------------------------------------------------------
	# Usrbin plot
	# ----------------------------------------------------------------------
	def _usrbinPlot(self):
		err = super()._usrbinPlot()
		if err is not None or self._det is None:
			return err

		regionPlot = self._det.type in (2, 8, 12, 18)
		b10  = self._det.type % 10
		# Unfortunately with image doesn't require
		# the last additional row/col that pm3d needs
		# so it duplicates the last row/col
		withimage = (self.engine.version() >= 5.0)
		withimage = False
		dx = self._det.dx
		dy = self._det.dy
		dz = self._det.dz
		proj = self.plot("proj")
		swap = int(self.plot("swap"))
		norm = self.plot("norm")
		font = ""
		try:
			h = Plot.USRBIN_HIST.index(self.plot("hist"))
		except ValueError:
			h = 0

		u        = np.array([])
		v        = np.array([])
		w        = np.array([])
		uu       = np.array([])
		errv     = np.array([])
		detector = np.array([])
		filename = f"{self.plot.name}.dat"
		lastline = 0
		i = 0
		with open(filename,'r') as f:
			for num, line in enumerate(f,1):
				lastline += 1
				if 'Detector' in line:
					detector = np.append(detector, num)
		detector = np.append(detector, lastline)

		if h == 0: # 2D Projection
			# Geometry information
			geo  = self.plot("geo")
			axes = self.plot("axes")
			plotgeo = ""

			worm  = []
			index = []
			if geo == Plot.GEO_NO or err:
				plotgeo = ""
			else:
				if geo == Plot.GEO_AUTO:
					name = f"{self.plot.name}.geo"
				else:
					name = geo

				if axes == Plot.COORDAUTO[0]:
					if proj == "X":
						u1, u2 = 4, 3
					elif proj == "Y":
						u1, u2 = 4, 2
					else:
						u1, u2 = 2, 3
					if swap:
						u1, u2 = u2, u1

					if regionPlot:
						# Find plot information
						for p in self.project.plots():
							if p.name == geo:
								axes = p("coord")
								u1 = ord(axes[ 0])-ord('U')
								u2 = ord(axes[-1])-ord('U')
								if u1>3: u1 -= 1
								if u2>3: u2 -= 1
								#font = p("labelfont")
								#if font is not None:
								#	font = "font '%s' "%(font)
								break
				else:
					u1 = ord(axes[ 0])-ord('U')
					u2 = ord(axes[-1])-ord('U')
					if u1>3: u1 -= 1
					if u2>3: u2 -= 1

				idx   = 1		# gnuplot index
				empty = 0		# continuous empty line count 2*empty lines = 1-line :)
				blockid = 0
				with open(f"{name}.dat",'r') as f:
					u = []
					v = []
					for line in f:
						line = line.strip()
						if line=="":
							if u:
								self.engine.ax.plot(u, v, c="black")
								u = []
								v = []
							empty += 1
							if empty == 2:
								empty = 0
								idx  += 1
								break
							continue
						empty = 0
						if line[0]=="#": continue
						words = line.split()
						u.append(float(words[u1]))
						v.append(float(words[u2]))
					if u:
						self.engine.ax.plot(u, v, c="black")

			filename = f"{self.plot.name}.dat"
			nu = self._det.nx + 1 # Probably FIXME
			with open(filename, 'r') as f:
				for line in f:
					linesp = line.split()
					if not linesp: continue
					#if s in line:
					#	rowsp = row[0].split()
					#	pyplot.text(float(rowsp[u0]),float(rowsp[u1]) +2,
					#			rowsp[6].replace('\\',''),
					#			horizontalalignment='center')
					#	pyplot.text(float(rowsp[u0]),float(rowsp[u1]),
					#			'$\\odot$', horizontalalignment='center')
					if b10 in (1,7) and proj=="Z":
						if swap:
							u = np.append(u, float((linesp[1]))*math.sin(float(linesp[0])))
							v = np.append(v, float((linesp[1]))*math.cos(float(linesp[0])))
						else:
							u = np.append(u, float((linesp[1]))*math.cos(float(linesp[0])))
							v = np.append(v, float((linesp[1]))*math.sin(float(linesp[0])))
					else:
						if   proj=="X":
							du = dz/2
							dv = dx/2
							nu = self._det.ny + 1
						elif proj=="Y":
							du = dz/2
							dv = dx/2
							nu = self._det.nz + 1
						else:
							du = dx/2
							dv = dy/2
						if (swap and proj!="Z") or (not swap and proj=="Z"):
							if withimage:
								u = np.append(u, float((linesp[1+dv])))
								v = np.append(v, float((linesp[0+du])))
							else:
								u = np.append(u, float((linesp[1])))
								v = np.append(v, float((linesp[0])))
						else:
							if withimage:
								u = np.append(u, float((linesp[0+du])))
								v = np.append(v, float((linesp[1+dv])))
							else:
								u = np.append(u, float((linesp[0])))
								v = np.append(v, float((linesp[1])))
					w = np.append(w, float((linesp[2])))
			uu = np.reshape(u, (-1, nu))
			vv = np.reshape(v, (-1, nu))
			ww = np.reshape(w, (-1, nu))
			# pyplot.contourf(uu, vv, ww, cmap='hot')

			vmin,vmax = self.cbLimits(
						np.min(w[np.nonzero(w)]),
						np.max(w[np.nonzero(w)]))
			cmap = self.colorBand()

			colormesh = self.engine.ax.pcolormesh(uu, vv, ww,
						norm=matplotlib.colors.LogNorm(vmin=vmin, vmax=vmax),
						cmap=cmap)
			# pyplot.pcolormesh(uu, vv, ww, vmin=0, vmax=0.005, cmap='cool')
			# self.engine.fig.colormaps()
			self.engine.fig.colorbar(colormesh)
			# -- Common plotting --

		# --- 1D Histogram ---
		elif h == 1: # 1D Projection
			filename = f"{self.plot.name}.dat"
			#grid = 0
			#ROWDONE = False
			#PLOTON  = False
			with open(filename, 'r') as f:
				for line in f:
					linesp = line.split()
					if not linesp: continue
					if linesp[0][0] == '#': continue
					style = self.plot("with.0")
					if style in ("histogram", "histerror"):
						u = np.append(u, float((linesp[0])))
					else:
						u = np.append(u, (float((linesp[0]))+float((linesp[0])))/2.0)
					v = np.append(v, float(linesp[2]))
					errv = np.append(errv, (float(linesp[2])*float(linesp[3]))/100)
			style  = self.plot("with.0")
			color  = self.engine.color(self.plot(f"lc.{i}"))
			marker = self.engine.marker(self.plot(f"pt.{i}"))
			markersize = int(self.plot(f"ps.{i}"))*5
			linewidth  = int(self.plot(f"lw.{i}"))
			label = self.plot(f"name.{i}")
			if style=="histerror":
				self.engine.ax.step(u, v, where='post',
						color = color,
						label = label,
						linewidth=linewidth)
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.errorbar(uu, v[0:-1], yerr=errv[0:-1],
							linewidth       = linewidth,
							marker          = marker,
							markersize      = markersize,
							ecolor          = color,
							markerfacecolor = color,
							markeredgecolor = color,
							fmt             = '.k')
			elif style=="errorbars":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.errorbar(uu, v[0:-1], yerr=errv[0:-1],
							linewidth       = linewidth,
							marker          = marker,
							markersize      = markersize,
							ecolor          = color,
							markerfacecolor = color,
							markeredgecolor = color,
							label           = label,
							fmt             = '.k')
			elif style=="errorlines":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.plot(uu, v[0:-1] , color = color,
						linewidth = linewidth)
				self.engine.ax.errorbar(uu, v[0:-1], yerr=errv[0:-1],
							linewidth       = linewidth,
							marker          = marker,
							markersize      = markersize,
							ecolor          = color,
							markerfacecolor = color,
							markeredgecolor = color,
							label           = label,
							fmt             = '.k')
			elif style=="histogram":
				self.engine.ax.step(u, v, where='post', color = color,
						linewidth = linewidth)
			elif style=="lines":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.plot(uu, v[0:-1], color = color,
						label = label,
						linewidth = linewidth)
			elif style=="linespoints":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.plot(uu, v[0:-1], color = color,
						linewidth = linewidth)
				self.engine.ax.scatter(uu, v[0:-1],
						color = color,
						label = label,
						marker = marker,
						s = markersize)
			elif style=="points":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.scatter(uu, v[0:-1],
						label = label,
						color = color,
						marker = marker,
						s = markersize)
			else:
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.errorbar(uu, v[0:-1], yerr=errv[0:-1], xerr = 0,
						ecolor = color,
						label  = label,
						fmt    = '.k')
		else: # 1D Max, Min, Trace-H, Trace-V
			filename = f"{self.plot.name}.dat"
			ind = max(0,h-3)
			style  = self.plot("with.0")
			color  = self.engine.color(self.plot(f"lc.{i}"))
			marker = self.engine.marker(self.plot(f"pt.{i}"))
			markersize = int(self.plot(f"ps.{i}"))*5
			linewidth  = int(self.plot(f"lw.{i}"))
			label = self.plot(f"name.{i}")
			if os.path.getsize(filename) == 0:
				self.show()
				return
			with open(filename, 'r') as f:
				for line in itertools.islice(f, int(detector[ind]), int(detector[ind+1])-1):
					linesp = line.split()
					if not linesp: continue
					if linesp[0][0] == '#': continue
					if style in ("histogram", "histerror"):
						u = np.append(u, float(linesp[0]))
					else:
						u = np.append(u, (float(linesp[0])+float(linesp[1]))/2.0)
					if ind == 0: # Max trace
						v = np.append(v, float(linesp[2]))
						errv = np.append(errv, (float(linesp[2])*float(linesp[3]))/100)
					else:
						v = np.append(v, float(linesp[2]))
						errv = np.append(errv, (float(linesp[2])*float(linesp[3]))/100)
			if style=="histerror":
				self.engine.ax.step(u, v, where='post',
						color = color,
						label = label,
						linewidth=linewidth)
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.errorbar(uu, v[0:-1], yerr=errv[0:-1],
							linewidth       = linewidth,
							marker          = marker,
							markersize      = markersize,
							ecolor          = color,
							markerfacecolor = color,
							markeredgecolor = color,
							fmt             = '.k')
			elif style=="errorbars":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.errorbar(uu, v[0:-1], yerr=errv[0:-1],
							linewidth       = linewidth,
							marker          = marker,
							markersize      = markersize,
							ecolor          = color,
							markerfacecolor = color,
							markeredgecolor = color,
							label           = label,
							fmt             = '.k')
			elif style=="errorlines":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.plot(uu, v[0:-1] , color = color,
						linewidth = linewidth)
				self.engine.ax.errorbar(uu, v[0:-1], yerr=errv[0:-1],
							linewidth       = linewidth,
							marker          = marker,
							markersize      = markersize,
							ecolor          = color,
							markerfacecolor = color,
							markeredgecolor = color,
							label           = label,
							fmt             = '.k')
			elif style=="histogram":
				self.engine.ax.step(u, v, where='post', color = color,
						linewidth = linewidth)
			elif style=="lines":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.plot(uu, v[0:-1], color = color,
						label = label,
						linewidth = linewidth)
			elif style=="linespoints":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.plot(uu, v[0:-1], color = color,
						linewidth = linewidth)
				self.engine.ax.scatter(uu, v[0:-1],
						color = color,
						label = label,
						marker = marker,
						s = markersize)
			elif style=="points":
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.scatter(uu, v[0:-1],
							label = label,
							color = color,
							marker = marker,
							s = markersize)
			else:
				uu = (u[1:] + u[:-1])/2.0
				self.engine.ax.errorbar(uu, v[0:-1], yerr=errv[0:-1], xerr = 0,  # FIXME xerr
							ecolor = color,
							label  = label,
							fmt    = '.k')

	# ----------------------------------------------------------------------
	# RESNUCLEi plot
	# ----------------------------------------------------------------------
	def _resnucleiPlot(self):
		pass

	# ----------------------------------------------------------------------
	def _userdumpPlot(self):
		pass
