#!/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   21-Nov-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re
import sys
import glob
import getopt
import tkinter as tk
import multiprocessing
from tkinter import ttk,messagebox

try:
	import webview
except ImportError:
	webview = None

import webbrowser

if __name__ == "__main__":
	sys.path.append(os.path.join(os.path.abspath(os.path.dirname(__file__)), "lib"))

import log
import rexx
import Global
import tkExtra
import tkFlair
import tkDialogs
from markdown import Markdown
from tkmarkdown import MarkdownFormatter

VIEWERS = [
		"",		# automatically select
		#"internal",
		"webview",
		"webbrowser" ]

WEBVIEW_TYPE = [
	"gtk",
	"qt",
	"edgechromium",
	"edgehtml",
	"mshtml",
	"cef"]

WEBBROWSER_TYPE = [
	"chrome",
	"chromium",
	"chromium-browser",
	"elinks",
	"epiphany",
	"firefox",
	"galeon",
	"google-chrome",
	"grail",
	"kfm",
	"kfmclient",
	"konqueror",
	"links",
	"lynx",
	"macosx",
	"mosaic",
	"mozilla",
	"netscape",
	"opera",
	"safari",
	"skipstone",
	"w3m",
	"windows-default"]

VIEWER = ""
TYPE   = ""

#===============================================================================
# Build manual index from
#===============================================================================
class ManualIndex:
	Hnn_PAT     = re.compile(r"<h[0-9]>(.*)</h[0-9]>")
	ID_PAT      = re.compile(r"<span id=\"(.*?)\"></span>")
	TITLE_PATH  = re.compile(r"<span class=\"section-number\">(.*?)</span>(.*?)<a class=.*")
	TITLE_PATH2 = re.compile(r"(.*?)<a class=.*")

	_index = {}

	#-----------------------------------------------------------------------
	@staticmethod
	def addref(mode, filename, title=None, span_id=None):
		index = ManualIndex._index.setdefault(mode, {})
		if title:
			index[title.strip().lower()] = filename
		if span_id:
			index[span_id.strip().lower()] = f"{filename}#{span_id}"

	#-----------------------------------------------------------------------
	@staticmethod
	def scan(mode, path):
		for root, dirs, files in os.walk(path):
			for file in files:
				fn,ext = os.path.splitext(file)
				if ext != ".html": continue
				filename = os.path.join(root,file)
				ManualIndex.addref(mode, filename, file)
				ManualIndex._scanHtml(mode, filename)

	#-----------------------------------------------------------------------
	@staticmethod
	def _scanHtml(mode, filename):
		with open(filename, "r", encoding="UTF-8", errors="replace") as f:
			for line in f:
				# search for heading and span id
				m = ManualIndex.Hnn_PAT.search(line)
				if m:
					m2 = ManualIndex.TITLE_PATH.search(m.group(1))
					if m2:
						ManualIndex.addref(mode, filename, m2.group(1))
						ManualIndex.addref(mode, filename, m2.group(2))
					else:
						m2 = ManualIndex.TITLE_PATH2.search(m.group(1))
						if m2:
							ManualIndex.addref(mode, filename, m2.group(1))
						else:
							ManualIndex.addref(mode, filename, m.group(1))
							Global.dump("H:", filename, m.group(1))
							Global.dump(line)
							#raise Exception("Not found")

				m = ManualIndex.ID_PAT.search(line)
				if m:
					ManualIndex.addref(mode, filename, span_id=m.group(1))

	#-----------------------------------------------------------------------
	@staticmethod
	def find(mode, target):
		index = ManualIndex._index.get(mode, ManualIndex._index.get("",None))
		if index is None: return None
		target = target.lower()
		#if target in index: Global.dump(f"<<< return '{target}'")
		try:
			return index[target]
		except KeyError:
			pass

		# find best match
		for key,value in index.items():
			if key.startswith(target):
				#Global.dump(f"<<< found as prefix '{key}'")
				return value

		for key,value in index.items():
			if target in key:
				#Global.dump(f"<<< found inside '{key}'")
				return value

		#---------------------------------------------------------------
		# search all loaded manuals for a possible match
		#---------------------------------------------------------------
		for anyindex in ManualIndex._index.values():
			try:
				return anyindex[target]
			except KeyError:
				pass
			for key,value in anyindex.items():
				if key.startswith(target):
					#Global.dump(f"<<< found as prefix '{key}'")
					return value

			for key,value in anyindex.items():
				if target in key:
					#Global.dump(f"<<< found inside '{key}'")
					return value

		#Global.dump(f"<<< not found '{target}'")
		return index["index.html"]

#===============================================================================
# webview manual browser
#===============================================================================
class ManualBrowser:
	_queue   = None
	_process = None

	#-------------------------------------------------------------------------------
	# On close event, send "quit" to window and kill process
	#-------------------------------------------------------------------------------
	@staticmethod
	def _on_closed():
		ManualBrowser._queue.put("quit")

	#-------------------------------------------------------------------------------
	# Running in a thread started in a multiprocess
	#-------------------------------------------------------------------------------
	@staticmethod
	def _control(window, queue):
		while True:
			url = queue.get()
			#Global.dump(f">>> ManualBrowser._control: {url}")
			if url == "quit": break
			window.load_url(url)
		window.destroy()

	#-------------------------------------------------------------------------------
	# Create webview window
	# Running in a multiprocess
	#-------------------------------------------------------------------------------
	@staticmethod
	def _help_process(queue, url, title="flair manual"):
		#Global.dump("*-* ManualBrowser._help_process()")
		#def configure():
		#	window.resize(432, 630)
		window = webview.create_window(title, url)
		if hasattr(window,"events"):
			window.events.closed += ManualBrowser._on_closed
		else:
			window.closed    += ManualBrowser._on_closed
		#window.events.closing   += on_closing
		#window.events.shown     += on_shown
		#window.events.loaded    += on_loaded
		#window.events.minimized += on_minimized
		#window.events.maximized += on_maximized
		#window.events.restored  += on_restored
		#window.events.resized   += on_resized
		webview.start(ManualBrowser._control, (window, queue)) #, gui="gtk")

	#-------------------------------------------------------------------------------
	# Open help webview if not open else show url
	#-------------------------------------------------------------------------------
	@staticmethod
	def show(url):
		Global.dump(f"*-* help('{url}')")
		if '#' in url: url,link = url.split('#', maxsplit=1)
		if ManualBrowser._process is not None and ManualBrowser._process.is_alive():
			ManualBrowser._queue.put(url)
		else:
			ManualBrowser._queue   = multiprocessing.Queue()
			ManualBrowser._process = multiprocessing.Process(
						target=ManualBrowser._help_process,
						args=(ManualBrowser._queue, url))
			ManualBrowser._process.start()

	#-------------------------------------------------------------------------------
	@staticmethod
	def quit():
		if ManualBrowser._queue is not None:
			ManualBrowser._queue.put("quit")
		if ManualBrowser._process is not None:
			ManualBrowser._process.kill()

#===============================================================================
# Tk based Manual browser
#===============================================================================
class Manual:
	# --- patterns ---
	SECTION_PAT = re.compile(r"^([0-9.]+\.?) (.*)$")
	SECTION_CHARS = '*=-~"+`'
	#_change = [	("\\`e",  "è"),
	#		("\\'e",  "é"),
	#		("\\`o",  "ò"),
	#		("\\\"u", "ü"),
	#		("\\o",   "ö"),
	#		("\\ie\\", "i.e.") ]
	#_patterns = [	re.compile(r"^(\d+)} +[A-Z].*$"),
	#		re.compile(r"^([A-Za-z\-]\d+)} +[A-Za-z\[\]].*$"),
	#		re.compile(r"^{([A-Za-z0-9\-:\[\]]+)}[a-z]*$"),
	#		re.compile(r"^{([A-Za-z0-9\-:\[\]]+)} +[A-Za-z\[\]].*$"),
	#		re.compile(r"^({[\w/]+})")] #Moira

	#_subpatterns = [ re.compile(r"^ *(\d+\.\d+)} +[A-Za-z0-9].*$"),
	#		 re.compile(r"^ *(\d+\.\d+.\d+)} +[A-Za-z].*$"),
	#		 re.compile(r"^ *(\d+\.\d+.\d+\.\d+)} +[A-Za-z].*$"),
	#		 re.compile(r"^ *([A-Za-z]\d+\.\d+)} +[A-Za-z].*$"),
	#		 re.compile(r"^ *([A-Za-z]\d+\.\d+\.\d+)} +[A-Za-z].*$"),
	#		 re.compile(r"^ *([A-Za-z]\d+\.\d+\.\d+\.\d+)} +[A-Za-z].*$") ]
	#_patbody1   = re.compile(r"^ +\d+\. [A-Z].*\. Code: ([A-Z].*)$")
	#_patbody2   = re.compile(r"^ +\d+\.\d+\.\d+\.\d+} [A-Z].*\. Code: ([A-Z].*)$")
	_patwhat    = re.compile(r"\bWHAT\([1-9.]+\)")
	_patsdum    = re.compile(r"\bSDUM\b")
	#_patcomment = re.compile(r"^ \*.*$", re.MULTILINE)
	#_pathtmltag = re.compile(r"<([a-z]+)[ >].*?</\1>")

	# global variables
	_MAX_VIEWS  = 20
	imgfolders = [	os.path.join(tkFlair.prgDir, "manual","html","_images"),
			os.path.join(tkFlair.prgDir, "icons")]

	# Manual is stored as a dictionary of sections
	manual     = {}
	section    = []
	title      = []
	pattern    = []
	images     = {}
	topSection = None
	current    = None	# current Manual browser to open only one instance

	# ----------------------------------------------------------------------
	def __init__(self, path):
		if not Manual.manual:
			if isinstance(path,str):
				Manual.load(Manual.findManual(path))
			else:
				for d in path:
					filename = Manual.findManual(d)
					if filename:
						Manual.load(filename)
		self.views   = []
		self.curview = 0
		self._move2End = False

	# ----------------------------------------------------------------------
	# before end delete all images
	# ----------------------------------------------------------------------
	@staticmethod
	def done():
		for i in Manual.images.values():
			del i
		Manual.images = {}

	# ----------------------------------------------------------------------
	# show manual window
	# ----------------------------------------------------------------------
	def show(self, master=None):
		if master is not None:
			self.toplevel = master
		else:
			# Start a new interpreter (not Toplevel),
			# to have the manual page responsive at all times
			self.toplevel = tk.Tk(className="flair_manual")

		self.toplevel.title("flair manual")
		try:
			image = Manual.images.get("info.gif")
			if image is None:
				image = self.loadImage("info.gif")
			self.toplevel.call('wm','iconphoto',self.toplevel._w, image)
		except tk.TclError:
			self.toplevel.iconbitmap("info")	# standard bitmap

		section = tkFlair._FONT_SECTION
		self.defaultFont = tkFlair.config.getFont(section,
					"manual.default",
					("Sans",11), self.toplevel)
		self.fixedFont   = tkFlair.config.getFont(section,
					"manual.fixed",
					("Mono",11), self.toplevel)
		self.titleFont   = tkFlair.config.getFont(section,
					"manual.title",
					("Sans",16), self.toplevel)

		# Paned Window
		self.splitter = tkExtra.HSplitter(self.toplevel, 250*tkFlair.dpiscale, True)
		self.splitter.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
		tocFrame  = self.splitter.leftFrame()
		textFrame = self.splitter.rightFrame()

		# Search
		self.backwardBtn = tk.Button(tocFrame, text="\N{LEFTWARDS ARROW}",
					font=self.defaultFont,
					padx=2, pady=0,
					command=self.backward)
		self.backwardBtn.grid(row=0, column=0, sticky=tk.NSEW)
		self.forwardBtn = tk.Button(tocFrame, text="\N{RIGHTWARDS ARROW}",
					font=self.defaultFont,
					padx=2, pady=0,
					command=self.forward)
		self.forwardBtn.grid(row=0, column=1, sticky=tk.NSEW)
		self.entry = tkExtra.LabelEntry(tocFrame,
					label="Search",
					labelcolor="DarkGray",
					background="White",
					font=self.defaultFont,
					width=7)
		self.entry.grid(row=0, column=2, columnspan=2, sticky=tk.NSEW)
		self.entry.bind("<<LabelEntryClear>>", self.searchClear)

		self.tocList = tkExtra.ExListbox(tocFrame, borderwidth=0,
				selectmode=tk.BROWSE,
				selectborderwidth=0,
				background="White",
				font=self.defaultFont,
				takefocus=True,
				exportselection=tk.FALSE)
		self.tocList.grid(row=1, column=0, columnspan=3, sticky=tk.NSEW)

		sb = ttk.Scrollbar(tocFrame, orient=tk.VERTICAL, takefocus=0,
				command=self.tocList.yview)
		sb.grid(row=1, column=3, sticky=tk.NSEW)
		self.tocList.config(yscrollcommand=sb.set)
		tocFrame.grid_rowconfigure(1, weight=1)
		tocFrame.grid_columnconfigure(2, weight=1)

		# Text Viewer
		self.txt = tk.Text(textFrame, background="White",
				font=self.fixedFont, takefocus=1)
		self.txt.config(wrap=tk.NONE, cursor="top_left_arrow")
		self.txt.grid(row=0, column=0, sticky='nsew')
		sbv = tkExtra.AutoScrollbar(textFrame, takefocus=0)
		sbv.grid(row=0, column=1, sticky='ns')
		sbh = tkExtra.AutoScrollbar(textFrame, takefocus=0,
				orient=tk.HORIZONTAL)
		sbh.grid(row=1, column=0, sticky='ew')

		# Markdown formatter
		self.mdf = MarkdownFormatter(self.txt)

		self.txt.config(yscrollcommand=sbv.set, xscrollcommand=sbh.set)
		sbv.config(command=self.txt.yview)
		sbh.config(command=self.txt.xview)

		textFrame.grid_rowconfigure(0, weight=1)
		textFrame.grid_columnconfigure(0, weight=1)

		# Bindings
		self.txt.bind('<Button-1>',		self.textFocus)
		self.tocList.bind('<ButtonRelease-1>',	self.searchClick)
		self.tocList.bind('<Return>',		self.searchReturn)
		self.tocList.bind('<KP_Enter>',		self.searchReturn)
		self.toplevel.bind('<Control-Key-f>',	self.findDialog)
		self.toplevel.bind('<Control-Key-g>',	self.findNext)
		self.entry.bind('<Return>',		self.entryReturn)
		self.entry.bind('<KP_Enter>',		self.entryReturn)
		self.txt.bind('<Up>',	lambda e, s=self.txt:
						s.yview(tk.SCROLL, -1, tk.UNITS))
		self.txt.bind('<Down>',	lambda e, s=self.txt:
						s.yview(tk.SCROLL,  1, tk.UNITS))
		self.toplevel.bind("<F11>",		self.toggleTree)
		self.toplevel.bind("<F12>",		self.toggleHeight)
		self.toplevel.bind("<Escape>",		self.close)
		self.toplevel.bind('<Control-Key-w>',	self.close)
		self.toplevel.bind('<Control-Key-q>',	self.close)
		self.toplevel.protocol("WM_DELETE_WINDOW", self.close)

		self.tocList.bind('<Button-3>',		self.popupMenu)
		self.tocList.bind('<Control-Key-space>',self.popupMenu)
		self.tocList.bind('<<ListboxSelect>>',	self.searchAfter)
		self.txt.bind('<Key-slash>',		self.findDialog)
		self.txt.bind('<Button-3>',		self.popupMenu)
		self.txt.bind('<Control-Key-space>',	self.popupMenu)
		self.txt.bind('<Prior>',		self.prior)
		self.txt.bind('<Next>',			self.next)
		self.txt.bind('<Control-Prior>',	self.pageUp)
		self.txt.bind('<Control-Next>',		self.pageDown)

		# Fill TOC
		first = True
		for title in Manual.title:
			if not first and title[0:3] == "---":
				self.tocList.insert(tk.END,"")
			self.tocList.insert(tk.END, title)
			if title[0]!=" ":
				if title[0]=="-":
					self.tocList.itemconfig(tk.END,foreground="DarkRed")
				else:
					self.tocList.itemconfig(tk.END,foreground="DarkBlue")
			first = False

		# Initial values
		self.current_section = None
		self.searchStr       = None
		Manual.current       = self
		self.showSection(Manual.section[0])

		# Show window
		self.toplevel.deiconify()
		self.toplevel.focus_set()
		self.tocList.focus_set()
		self._oldHeight = 700
		if not self.toplevel.winfo_ismapped():
			self.toplevel.wait_visibility()
		self.toplevel.geometry("1024x800")
		self.splitter_pos = self.splitter.split
		self.inafter = False

	# ----------------------------------------------------------------------
	def close(self, event=None):
		if isinstance(self.toplevel, tk.Tk):
			# Destroy local icons...
			Manual.done()
		self.toplevel.destroy()
		Manual.current = None

	# ----------------------------------------------------------------------
	def textFocus(self, event):
		self.txt.focus_set()

	# ----------------------------------------------------------------------
	def toggleTree(self, event):
		if self.splitter.split > self.splitter.min:
			self.splitter.toggle()
			self.splitter.rightFrame().focus_set()
			self.toplevel.focus_set()
		else:
			self.splitter.toggle()

	# ----------------------------------------------------------------------
	def toggleHeight(self, event):
		self._oldHeight = tkExtra.toggleHeight(self.toplevel,
					self._oldHeight)

	# ----------------------------------------------------------------------
	def searchClear(self, event=None):
		self.search("")

	# ----------------------------------------------------------------------
	def backward(self, event=None):
		if self.curview>=len(self.views):
			self.curview -= 1
		if self.curview>0:
			self.curview -= 1
			section, searchstr = self.views[self.curview]
			if searchstr != self.entry.get():
				self.entry.delete(0,tk.END)
				self.entry.insert(0, searchstr)
				self.search(searchstr, False)
			self.showSection(section)

	# ----------------------------------------------------------------------
	def forward(self, event=None):
		if self.curview < len(self.views)-1:
			self.curview += 1
			section, searchstr = self.views[self.curview]
			if searchstr != self.entry.get():
				self.entry.delete(0,tk.END)
				self.entry.insert(0, searchstr)
				self.search(searchstr, False)
			self.showSection(section)

	# ----------------------------------------------------------------------
	# Create the popup menu
	# ----------------------------------------------------------------------
	def popupMenu(self, event):
		menu=tk.Menu(self.toplevel, tearoff=0)
		menu.add_command(label='Backward', underline=0,
				command=self.backward)
		menu.add_command(label='Forward', underline=0,
				command=self.forward)
		menu.add_command(label='Clear', underline=0,
				command=self.searchClear)
		menu.add_separator()
		menu.add_command(label='Find', underline=0,
				command=self.findDialog)
		menu.add_command(label='Find Next', underline=5,
				command=self.findNext)
		menu.add_separator()
		menu.add_command(label='Close',
				command=self.close)
		menu.tk_popup(event.x_root, event.y_root)
		return "break"

	# ----------------------------------------------------------------------
	# Return hit on entry box
	# ----------------------------------------------------------------------
	def entryReturn(self, event=None):
		self.search(self.entry.get())

	# ----------------------------------------------------------------------
	# Search for an expression in help pages
	# ----------------------------------------------------------------------
	def search(self, expr, select=True):
		self.searchStr = expr.upper().split()
		self.current_section = None
		self.tocList.delete(0,tk.END)
		for section,title in zip(Manual.section,Manual.title):
			txt = Manual.manual[section].upper()
			found = True
			for w in self.searchStr:
				if txt.find(w) == -1:
					found = False
					break
			if found:
				self.tocList.insert(tk.END,title)
				if title[0]!=" ":
					if title[0]=="-":
						self.tocList.itemconfig(tk.END,foreground="DarkRed")
					else:
						self.tocList.itemconfig(tk.END,foreground="DarkBlue")
		if select and self.tocList.size() > 0:
			self.tocList.selection_clear(0,tk.END)
			self.tocList.selection_set(0)
			self.tocList.activate(0)
			self.tocList.focus_set()
			self.searchReturn()

	# ----------------------------------------------------------------------
	# button click
	# ----------------------------------------------------------------------
	def searchClick(self, event):
		self.tocList.focus_set()
		idx = self.tocList.nearest(event.y)
		if idx<0: return
		title = self.tocList.get(idx)
		if not title:
			idx  += 1
			title = self.tocList.get(idx)
		item    = Manual.title.index(title)
		section = Manual.section[item]
		self.showSection(section)

	# ----------------------------------------------------------------------
	def searchReturn(self, event=None):
		if self.tocList.size()==0: return
		sel = self.tocList.index(tk.ACTIVE)
		if sel<0: return
		self.tocList.selection_clear(0,tk.END)
		self.tocList.selection_set(sel)
		title = self.tocList.get(sel)
		if not title:
			self.tocList.selection_clear(sel)
			sel = int(sel) + 1
			self.tocList.selection_set(sel)
			title = self.tocList.get(sel)
		item    = Manual.title.index(title)
		section = Manual.section[item]
		self.showSection(section)
		if self._move2End:
			self.txt.see(tk.END)
			self._move2End = False

	# ----------------------------------------------------------------------
	def searchAfter(self, event=None):
		if self.inafter: return
		self.inafter = True
		self.toplevel.after(250, self.searchReturn)

	# ----------------------------------------------------------------------
	def prior(self, event=None):
		top,_ = self.txt.yview()
		if top < 0.001:
			self.pageUp()
			return "break"

	# ----------------------------------------------------------------------
	def next(self, event=None):
		top, bot = self.txt.yview()
		if bot > 1.0 - (bot-top)/5.0:
			self.pageDown()
			return "break"

	# ----------------------------------------------------------------------
	def pageUp(self, event=None):
		active = self.tocList.index(tk.ACTIVE)
		if active==0: return
		try: active = int(active)-1
		except: pass
		self.tocList.activate(active)
		self._move2End = True
		self.searchReturn()

	# ----------------------------------------------------------------------
	def pageDown(self, event=None):
		active = self.tocList.index(tk.ACTIVE)
		try: active = int(active)+1
		except: pass
		self.tocList.activate(active)
		self.searchReturn()

	# ----------------------------------------------------------------------
	def loadImage(self, imageName):
		# Search the following directories
		# 1. current manual location
		# 2. image folder of flair
		# 3. icons folder of flair
		# Try to load from icons directory
		if imageName.find('/') != -1:
			filename = imageName
		else:
			for folder in Manual.imgfolders:
				filename = os.path.join(folder, imageName)
				if os.path.isfile(filename):
					break
			else:
				log.warning(f"Image '{filename}' not found")
				return None
		try:
			image = tk.PhotoImage(master=self.toplevel, file=filename)
			if tkFlair.dpiscale>1:
				image = image.zoom(tkFlair.dpiscale)
		except tk.TclError:
			log.warning(f"Cannot open '{filename}'")
			return None
		Manual.images[imageName] = image
		return image

	# ----------------------------------------------------------------------
	# show section
	# ----------------------------------------------------------------------
	def showSection(self, section=None, err=True):
		self.inafter = False
		if section is None:
			section = Manual.topSection

		try:
			idx = Manual.section.index(section)
		except:
			# Try to find the best match
			idx = 0
			su = section[:8].upper()
			for s in Manual.section:
				if su == s[:8].upper():
					section = s
					break
				idx += 1
			# For Moira
			idx = 0
			su = section[:]
			for s in Manual.section:
				if su in s[:]:
					section = s
					break
				idx += 1

		# Search item in list to highlight
		try: title=Manual.title[idx]
		except IndexError:
			if err:
				try:
					messagebox.showwarning("Section not found",
							f"Section {section} not found",
							parent=self.toplevel)
				except tk.TclError:
					pass
			return False
		for i in range(self.tocList.size()):
			item = self.tocList.get(i)
			if item == title or item.strip() == section:
				self.tocList.selection_clear(0,tk.END)
				self.tocList.selection_set(i)
				self.tocList.activate(i)
				self.tocList.see(i)
				break

		if self.current_section == section: return True
		self.current_section = section

		# Add to view history
		if self.curview == 0:
			self.backwardBtn.config(state=tk.DISABLED)
		else:
			self.backwardBtn.config(state=tk.NORMAL)
		if self.curview >= len(self.views):
			self.curview += 1
			self.views.append((section, self.entry.get()))
			if len(self.views) > Manual._MAX_VIEWS:
				self.views.pop(0)
				self.curview -= 1
		else:
			if self.views[self.curview][0] != section:
				# delete the rest and append the new view
				self.curview += 1
				del self.views[self.curview:]
				self.views.append((section, self.entry.get()))
		if self.curview >= len(self.views)-1:
			self.forwardBtn.config(state=tk.DISABLED)
		else:
			self.forwardBtn.config(state=tk.NORMAL)

		txt = self.txt
		txt.config(state=tk.NORMAL)
		txt.delete(0.0, tk.END)
		text = Manual.manual.get(section,"")
		#txt.insert(0.0, text)

		#heading = '#'*(section_no.count('.'))
		#md = Markdown(f"{heading} {firstline}\n{txt}")
		md = Markdown(text)

		self.mdf.insert(0.0, md)

		# Highlight items
		#txt.tag_delete("title")
		#txt.tag_add("title", "1.0", "1.end")
		#txt.tag_configure("title", foreground="DarkRed", font=self.titleFont)
		#txt.update_idletasks()

		## WHAT's & SDUM
		#txt.tag_delete("what")
		#for m in Manual._patwhat.finditer(text):
		#	txt.tag_add("what",
		#			f"1.0 + {m.start()} chars",
		#			f"1.0 + {m.end()} chars")
		#for m in Manual._patsdum.finditer(text):
		#	txt.tag_add("what",
		#			f"1.0 + {m.start()} chars",
		#			f"1.0 + {m.end()} chars")
		#txt.tag_configure("what", foreground="DarkMagenta")

		## Hyper links
		#txt.tag_delete("card")
		#tag = "card"
		#for i in range(len(Manual.section)-1,-1,-1):
		#	s = Manual.section[i]
		#	if s == section: continue
		#	pat1, pat2 = Manual.pattern[i]
		#	for m in pat1.finditer(text):
		#		txt.tag_add(tag,
		#				f"1.0 + {m.start()} chars",
		#				f"1.0 + {m.end()} chars")

		#	if pat2 is not None:
		#		for m in pat2.finditer(text):
		#			txt.tag_add(tag,
		#					f"1.0 + {m.start()} chars",
		#					f"1.0 + {m.end()} chars")
		#txt.tag_remove(tag, "1.0", "2.0")

		## Highlight current section
		#try:
		#	i = Manual.section.index(section)
		#	tag = "me"
		#	pat1, pat2 = Manual.pattern[i]
		#	for m in pat1.finditer(text):
		#		txt.tag_add(tag,
		#				f"1.0 + {m.start()} chars",
		#				f"1.0 + {m.end()} chars")

		#	if pat2 is not None:
		#		for m in pat2.finditer(text):
		#			txt.tag_add(tag,
		#				f"1.0 + {m.start()} chars",
		#				f"1.0 + {m.end()} chars")
		#	txt.tag_configure("me", foreground="DarkBlue")
		#except ValueError:
		#	pass
		#	#if section == Manual.section[i]:
		#	#	tag = "me"
		#	#else:
		#txt.tag_configure("card",
		#		foreground = "Blue",
		#		underline  = 1)
		#txt.tag_bind("card", "<Button-1>", self.hyperlink)

		## Highlight search string
		#if self.searchStr is not None:
		#	txt.tag_delete("find")
		#	firstIndex = None
		#	for target in self.searchStr:
		#		index = "1.0"
		#		while True:
		#			index = txt.search(target, index,
		#					nocase=1, stopindex=tk.END)
		#			if not index: break
		#			endex = f"{index} + {len(target)} chars"
		#			txt.tag_add("find", index, endex)
		#			txt.update_idletasks()
		#			if firstIndex is None:
		#				firstIndex = index
		#			index = endex
		#	if firstIndex is not None:
		#		txt.see(firstIndex)
		#	txt.tag_configure("find",
		#			background = '#800000',
		#			foreground = '#FFFFFF')

		## Find&insert all image tags
		#index = "0.0"
		#while True:
		#	index = txt.search("{img:", index)
		#	if not index: break
		#	indexEnd = txt.search("}", index)
		#	imageName = txt.get(index, indexEnd)[5:]
		#	txt.delete(index, f"{indexEnd} + 1 char")
		#	image = Manual.images.get(imageName)
		#	if image is None:
		#		image = self.loadImage(imageName)

		#	if image is not None:
		#		try:
		#			txt.image_create(index, image=image)
		#		except tk.TclError:
		#			log.error(f"displaying image '{imageName}'")

		## html hyper links
		#index = "0.0"
		#while True:
		#	idx = index
		#	index  = txt.search("http://", idx, stopindex=tk.END)
		#	index2 = txt.search("https://", idx, stopindex=tk.END)
		#	if not index: index = index2
		#	if not index: break
		#	if index2 and txt.compare(index2, '<', index): index = index2
		#	line = index.split(".")[0]
		#	indexEnd = txt.search(" ", index, stopindex=line+".end")
		#	if not indexEnd: indexEnd = line+".end"
		#	txt.tag_add("url", index, indexEnd)
		#	index = indexEnd
		#txt.tag_configure("url",
		#		foreground = "Blue",
		#		underline  = 1)
		#txt.tag_bind("url", "<Button-1>", self.urllink)

		# No more editing is allowed
		txt.config(state=tk.DISABLED)
		return True

	# --------------------------------------------------------------------
	# FindDialog
	# --------------------------------------------------------------------
	def findDialog(self, event=None):
		txt = self.entry.get()
		fd = tkDialogs.FindReplaceDialog(self.toplevel, replace=False)
		fd.show(self.find, None, None, txt)
		try:
			self.txt.tag_delete("find")
		except tk.TclError:
			pass

	# ----------------------------------------------------------------------
	# Find target(s)
	# ----------------------------------------------------------------------
	def find(self, findStr=None, matchCase=None):
		if findStr is not None:
			self.searchStr = [findStr]
		else:
			if self.searchStr is None: return

		minrow = 100000
		mincol = 100000
		minlen = 0

		self.txt.tag_delete("find")
		for target in self.searchStr:
			index = str(self.txt.search(target,
					"insert + 1 chars",
					nocase = matchCase))
			if not index: continue
			row,col = map(int,index.split('.'))
			if row<minrow or (row==minrow and col<mincol):
				minrow = row
				mincol = col
				minlen = len(target)

		if minrow < 100000:
			index = f"{minrow}.{mincol}"
			self.txt.mark_set('insert', index)
			self.txt.tag_add("find", 'insert', f"insert + {minlen} chars")
			self.txt.see(index)
			self.txt.tag_configure("find",
					background = '#800000',
					foreground = '#FFFFFF')

	# ----------------------------------------------------------------------
	def findNext(self, event=None):
		self.find()

	# ----------------------------------------------------------------------
	# hyper link
	# ----------------------------------------------------------------------
	def hyperlink(self, event):
		row,col = map(int, self.txt.index(f"@{event.x},{event.y}").split("."))

		# Find line being clicked
		lines = Manual.manual[self.current_section].splitlines()
		line = lines[row-1]
		del lines

		# Find word
		if line[col]=='}': col -= 1
		begin = col
		while begin>=0:
			ch = line[begin]
			if ch.isalnum() or ch in "-.:[]":
				begin -= 1
			else:
				break
		begin += 1

		end = col
		while end<len(line):
			ch = line[end]
			if ch.isalnum() or ch in "-.:[]":
				end += 1
			else:
				break

		section = line[begin:end]
		if section[-1]==".": section = section[:-1]

		try:
			item = Manual.section.index(section)
		except:
			# Try to find the best match
			item = 0
			su = section[:8].upper()
			for s in Manual.section:
				if rexx.abbrev(su, s[:8].upper(), 5):
					section = s
					break
				item += 1

		self.tocList.selection_clear(0,tk.END)
		self.tocList.selection_set(item)
		self.tocList.activate(item)
		self.searchStr = None
		self.showSection(section)

	# ----------------------------------------------------------------------
	def urllink(self, event=None):
		idx = self.txt.index(f"@{event.x},{event.y}")
		line = idx.split(".")[0]
		index  = self.txt.search("http://", idx, backwards=True, stopindex=line+".0")
		index2 = self.txt.search("https://", idx, backwards=True, stopindex=line+".0")
		if not index: index = index2
		if not index: return
		if index2 and self.txt.compare(index2, ">", index):
			index = index2
		indexEnd = self.txt.search(" ", index, stopindex=line+".end")
		if not indexEnd: indexEnd = line+".end"
		tkFlair.openurl( self.txt.get(index, indexEnd).strip() )

	# ----------------------------------------------------------------------
	# load manual and split it into sections
	# ----------------------------------------------------------------------
	#@staticmethod
	#def load(filename):
	#	try:
	#		self._load(filename)
	#	except IOError as err:
	#		messagebox.showerror("Error", err)
	#		return

	# ----------------------------------------------------------------------
	@staticmethod
	def load(filename):
		with open(filename, mode="r", encoding="UTF-8") as fmanual:
			Manual.imgfolders.insert(0, os.path.dirname(os.path.abspath(filename)))
			fn,_ = os.path.splitext(os.path.basename(filename))
			#index   = True
			#txt     = ""
			#section = fn
			#if Manual.topSection is None:
			#	Manual.topSection = section
			#title    = f"----- {fn} -----"
			#inGeo   = False
			#depth   = 1

			prevline      = ""
			section_no    = ""
			section_title = ""
			text          = []
			zero_section  = [0, 0, 0, 0]

			for line in fmanual:
				line = line.rstrip()
				if line and line[0] in Manual.SECTION_CHARS:
					if re.fullmatch(rf"\{line[0]}*", line):
						section_level = Manual.SECTION_CHARS.index(line[0])
						# Delete last line which should be the section header
						if text and text[-1] == prevline:
							text.pop()
						if section_title:
							Manual._addSection(fn, section_no,
								section_title, "\n".join(text))
						pat = Manual.SECTION_PAT.match(prevline)
						if pat:
							section_no = pat.group(1)
							section_title = pat.group(2)
						else:
							zero_section[section_level] += 1
							section_no = f"0.{'.'.join(map(str, zero_section[:section_level+1]))}."
							section_title = prevline
						text = []
						continue
				line = line.replace("[image]","")
				text.append(line)
				prevline = line
			Manual._addSection(fn, section_no, section_title, "\n".join(text))

	# ----------------------------------------------------------------------
	@staticmethod
	def _addSection(filename, section_no, section_title, txt):
		section = f"{filename} {section_no}"
		if section_no[0]!='0':
			firstline = f"{section_no} {section_title}"
		else:
			firstline = section_title

		heading = '#'*(section_no.count('.'))
		#md = Markdown(f"{heading} {firstline}\n{txt}")
		Manual.manual[section] = f"{heading} {firstline}\n{txt}"

		Manual.section.append(section)
		Manual.title.append(firstline)
		ch = section[0]
		target = section.replace(".",r"\.")
		target = target.replace("[",r"\[")
		target = target.replace("]",r"\]")
		if 'A'<=ch<='Z':
			pat1 = re.compile(rf"\b({section[:8]})\b")
			f = firstline.split()[0]
			if f != section[:8]:
				pat2 = re.compile(r"\b({f[:8]}){f[8:]}\b")
			else:
				pat2 = None
		elif ch in ":[":
			pat1 = re.compile(f"({target})")
			pat2 = None
		else:
			pat1 = re.compile(rf"\b({target})}}")
			pat2 = None
		Manual.pattern.append((pat1, pat2))

	# ----------------------------------------------------------------------
	# Find latest manual file in the FLUKA directory
	# ----------------------------------------------------------------------
	@staticmethod
	def findManual(path):
		if not path:
			return None
		elif os.path.isdir(path):
			path = os.path.abspath(path)
			maxt   = 0
			manual_file = None
			for f in glob.glob(f"{path}/*.manual"):
				try:
					t = os.path.getmtime(os.path.join(path,f))
					if t > maxt: # get the newer only
						manual_file = f
						maxt = t
				except:
					pass
			return manual_file
		elif os.path.isfile(path):
			return os.path.abspath(path)

#-------------------------------------------------------------------------------
def viewer():
	if not VIEWER:
		return "webview" if webview else "webbrowser"
	return VIEWER

#-------------------------------------------------------------------------------
# Show keyword
#-------------------------------------------------------------------------------
def show(mode, keyword=None):
	if viewer() in ("webview", "webbrowser"):
		def showPage(mode, target):
			url = ManualIndex.find(mode, target)
			if url is None:
				# FIXME maybe show an error message on the tkmessagebox
				return False
			elif viewer() == "webview":
				ManualBrowser.show(url)
			elif viewer() == "webbrowser":
				webbrowser.open_new_tab(url)
			return True

		if not keyword or not isinstance(keyword,str):
			showPage(mode, "index.html")
		else:
			showPage(mode, keyword)
	#else:
	#	if Manual.current is not None:
	#		Manual.current.searchClear()
	#		#Manual.current.showSection(keyword)
	#		Manual.current.toplevel.deiconify()
	#		Manual.current.toplevel.lift()
	#		Manual.current.toplevel.focus_set()
	#		manual = Manual.current
	#	else:
	#		manual = Manual(_directories)
	#		try:
	#			manual.show()
	#		except IndexError:
	#			messagebox.showwarning(
	#					"Manual error",
	#					"Error opening manual")

	#	if isinstance(keyword,str):
	#		manual.showSection(keyword)
	#	else:
	#		for k in keyword:
	#			if manual.showSection(k,False):
	#				break

#-------------------------------------------------------------------------------
def destroy():
	if viewer() == "webview":
		ManualBrowser.quit()
	elif viewer() == "webbrowser":
		pass
	elif Manual.current is not None:
		Manual.current.close()

#-------------------------------------------------------------------------------
# Initialize Manual by settings the default search path
#-------------------------------------------------------------------------------
def init(dirs):
	for mode,path in dirs:
		Global.dump("*-* Manual.init",mode,path)
		ManualIndex.scan(mode, os.path.join(path,"html"))
		try:
			Global.dump(">>>     loaded", len(ManualIndex._index[mode]))
		except KeyError:
			Global.dump(">>>     nothing loaded")

#-------------------------------------------------------------------------------
def usage():
	print()
	print("Usage: Manual.py [options] [page]")
	print()
	print("Options:")
	print("\t-?\t--help\t\tShow this help page")
	print("\t-m\t--mode\t\tUse this simulation 'mode' (default: fluka)")
	print("\t-v\t--viewer\tPreferred viewer for the manual")
	print("\tpage\t\t\tpage to show (default index)")
	sys.exit(0)

#-------------------------------------------------------------------------------
if __name__ == "__main__":
	from fluka import Fluka,FlukaConfig
	#from moira import Moira,MoiraConfig
	manualpath = None
	page = None
	mode = Fluka.MODE

	if len(sys.argv)>1:
		try:
			optlist, args = getopt.getopt(sys.argv[1:],
					"?hv:m:",
					["help",
					 "viewer=",
					 "mode="])
		except getopt.GetoptError:
			usage()

		for opt, val in optlist:
			if opt in ("-?", "--help"):
				usage()
			elif opt in ("-v", "--viewer"):
				VIEWER = val
			elif opt in ("-m", "--mode"):
				mode = val
			else:
				print(f"Unknown option '{opt}'\n")
				usage()

		if args: page = args[0]

	tkFlair.openIni()
	FlukaConfig.loadConfig(tkFlair.config)
	init([("", os.path.join(tkFlair.prgDir,"manual")),
		(Fluka.MODE, FlukaConfig.documentDir())])
		#(Moira.MODE, MoiraConfig.documentDir())])
	show(mode,page)

	#while True:
	#	mode,target = input().strip().split(maxsplit=1)
	#	if mode == "quit": break
	#	print(f">>> Found: {ManualIndex.find(mode, target)}")
	#	print()
	#exit()
