# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   07-Mar-2007

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import tkinter as tk
from tkinter import ttk, messagebox

import rexx
import icons
#import Global
import Ribbon
import tkFlair
import tkExtra
import FlairRibbon
import tkTextEditor
import PeriodicTable

from element import Element
from material import Material

from deck  import Card
from fluka import Fluka
from phits import Phits

_clipboard = []

#-------------------------------------------------------------------------------
# Create material and insert into a fluka input
#-------------------------------------------------------------------------------
def insertMaterialFluka(flair, mat):
	# Create the basic materials
	ignore = set(flair.project.input.materialList())

	undoinfo = [flair.refreshUndo(), flair.renumberUndo()]

	# Find a name that doesn't exist or replace material?
	for n in mat.name.split():
		matname = n[:8]
		if matname=="PHOSPHOR": matname = "PHOSPHO"
		if matname not in ignore:
			break
	else:
		if not tkFlair.askyesno("Material exists",
				f"Material '{mat.name}' already exists!\n"
				"Do you want to insert it anyway?",
				parent=flair):
			return None, None

	# Find position
	# Add MATERIAL/COMPOUND cards after the last MATERIAL
	# or COMPOUND card
	pos = 0
	for tag in ["MATERIAL", "COMPOUND"]:
		try:
			card = flair.project.input.cards[tag][-1]
			pos = max(pos, card.pos())
		except:
			pass
	# If everything fails try GEOEND
	if pos == 0:
		try:
			card = flair.project.input.cards["GEOEND"][-1]
			pos = card.pos()
		except:
			pass
	pos += 1

	matelem = []
	mode = flair.project.input.mode
	for card in mat.materialCard(matname, ignore, mode):
		undoinfo.append(flair.addCardUndo(card, pos, False))
		pos += 1
		if card.tag == "MATERIAL":
			matelem.append(card.name())

	# Check for special material properties
	prop = []

	# Check for MAT-PROP
	if mat.pressure or mat.rhor or mat.ionization:
		prop.append("MAT-PROP")
	if mat.dpa_threshold:
		prop.append("DPA")
	if mat.s_cbar or mat.s_x0 or \
	   mat.s_x1   or mat.s_a  or \
	   mat.s_m    or mat.s_d0:
		prop.append("STERNHEI")

	if prop:
		dlg = SelectPropertiesDlg(flair, matname, prop)
		if dlg.add:
			if "MAT-PROP" in dlg.add:
				undoinfo.append(flair.addCardUndo(
						mat.matPropCard(matname,mode),
						pos, False))
				pos += 1
			if "DPA" in dlg.add:
				undoinfo.append(flair.addCardUndo(
						mat.dpaCard(matname,mode),
						pos, False))
				pos += 1
			if "STERNHEI" in dlg.add:
				undoinfo.append(flair.addCardUndo(
						mat.sternheimerCard(matname,mode),
						pos, False))
				pos += 1
	undoinfo.append(flair.renumberUndo())
	undoinfo.append(flair.refreshUndo())
	flair.addUndo(undoinfo, "Materials added")

	return matname, matelem

#-------------------------------------------------------------------------------
# Create material and insert into a phits input
#-------------------------------------------------------------------------------
def insertMaterialPhits(flair, mat):
	# Create the basic materials
	ignore = set(flair.project.input.materialList())

	undoinfo = [flair.refreshUndo(), flair.renumberUndo()]

	# Find a name that doesn't exist or replace material?
	for matname in mat.name.split():
		if matname not in ignore:
			break
	else:
		if not tkFlair.askyesno("Material exists",
				f"Material {matname} already exists!\n"
				"Do you want to insert it anyway?",
				parent=flair):
			return None, None

	# Find position
	pos = 0
	card = list(flair.project.input.cardsSorted(Card.MATERIALS,-1).values())
	if card: pos = card[-1].pos()+1
	if pos == 0:
		card = flair.project.input["[Material]"]
		if card: pos = card[0].pos()+1

	matelem = []
	mode = flair.project.input.mode
	for card in mat.materialCard(matname, ignore, mode):
		undoinfo.append(flair.addCardUndo(card, pos, False))
		matelem.append(card.name())
		pos += 1

	# Check for special material properties
	#prop = []

	# Check for MAT-PROP
	#if mat.pressure or mat.rhor or mat.ionization:
	#	prop.append("MAT-PROP")
	#if mat.dpa_threshold:
	#	prop.append("DPA")
	#if mat.s_cbar or mat.s_x0 or \
	#   mat.s_x1   or mat.s_a  or \
	#   mat.s_m    or mat.s_d0:
	#	prop.append("STERNHEI")

	#if prop:
	#	dlg = SelectPropertiesDlg(flair, matname, prop)
	#	if dlg.add:
	#		if "MAT-PROP" in dlg.add:
	#			undoinfo.append(flair.addCardUndo(
	#					mat.matPropCard(matname,mode),
	#					pos, False))
	#			pos += 1
	#		if "DPA" in dlg.add:
	#			undoinfo.append(flair.addCardUndo(
	#					mat.dpaCard(matname,mode),
	#					pos, False))
	#			pos += 1
	#		if "STERNHEI" in dlg.add:
	#			undoinfo.append(flair.addCardUndo(
	#					mat.sternheimerCard(matname,mode),
	#					pos, False))
	#			pos += 1
	undoinfo.append(flair.renumberUndo())
	undoinfo.append(flair.refreshUndo())
	flair.addUndo(undoinfo, "Materials added")

	return matname, matelem

#-------------------------------------------------------------------------------
# Insert the material list to input
#-------------------------------------------------------------------------------
def insert2Input(flair, mats):
	assert flair.project.ismode(Fluka, Phits)
	if not flair.project.input.cardlist:
		flair.project.input.minimumInput()
	if isinstance(mats, list):
		matnames = []
		elements = set()
		for m in mats:
			if flair.project.ismode(Fluka):
				n, elems = insertMaterialFluka(flair, Material.materials[m])
			#elif flair.project.ismode(Moira):
			elif flair.project.ismode(Phits):
				n, elems = insertMaterialPhits(flair, Material.materials[m])
			#elif flair.project.ismode(Mcnp):
			if n:
				matnames.append(n)
				for e in elems:
					elements.add(e)
		matnames = " ".join(matnames)
		elements = " ".join(sorted(list(elements)))
	else:
		if flair.project.ismode(Fluka):
			matnames,elements = insertMaterialFluka(flair, Material.materials[mats])
		#elif flair.project.ismode(Moira):
		elif flair.project.ismode(Phits):
			matnames,elements = insertMaterialPhits(flair, Material.materials[mats])
		#elif flair.project.ismode(Mcnp):
		if elements: elements = " ".join(sorted(elements))

	if matnames:
		flair.notify(f"Material: {matnames}",
			f"Added materials:\n{matnames}\n\n"
			f"Based on:\n{elements}",
			tkFlair.NOTIFY_NORMAL)
	else:
		flair.notify("No material added",
			"No material added to input.",
			tkFlair.NOTIFY_WARNING)
	return matnames, elements

#===============================================================================
# Material Frame
#===============================================================================
class MaterialPage(FlairRibbon.FlairPage):
	"""Database of predefined and user materials."""

	_name_ = "Materials"
	_icon_ = "materialdb"

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Input ===========
		group = Ribbon.LabelGroup(self.ribbon, "Materials")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col,row = 0,0
		b = Ribbon.LabelButton(group.frame,
				text="To Input",
				image=icons.get("input32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.insert2Input)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Insert selected materials to input")
		self.list_widgets.append(b)

		# ---
		col,row = 1,0
		b = Ribbon.LabelButton(group.frame,
				text="Import",
				image=icons.get("import32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.importFromInput)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Import Materials FROM Input")
		self.list_widgets.append(b)

		# ---
		col,row = 2,0
		b = Ribbon.LabelButton(group.frame,
				text="Add",
				image=icons.get("add"),
				command=self.add)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Create new material")
		self.list_widgets.append(b)

		# ---
		col,row = 2,1
		b = Ribbon.LabelButton(group.frame,
				text="Delete",
				image=icons.get("x"),
				command=self.delete)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Clone run [Ctrl-D]")
		self.list_widgets.append(b)

		# ---
		col,row = 2,2
		b = Ribbon.LabelButton(group.frame,
				text="Edit",
				image=icons.get("edit"),
				command=self.edit)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Edit material")
		self.list_widgets.append(b)

		# -------------- Edit Ribbon -------------
		self.editribbon = group = Ribbon.LabelGroup(self.ribbon, "Edit")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		b = Ribbon.LabelButton(group,
				text="Ok",
				image=icons.get("ok32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.ok)
		tkExtra.Balloon.set(b, "Update Material")
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
		self.properties_widgets.append(b)

		b = Ribbon.LabelButton(group,
				text="Cancel",
				image=icons.get("x32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.showMaterial)
		tkExtra.Balloon.set(b, "Cancel changes")
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
		self.properties_widgets.append(b)

	#----------------------------------------------------------------------
	# Create Project page
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage()

		self.list_widgets = []
		self.properties_widgets = []

		# ============ Search ============
		frame = tk.Frame(self.frame)
		frame.pack(side=tk.TOP, fill=tk.BOTH, padx=3, pady=2)

		tk.Label(frame, text="Search:").pack(side=tk.LEFT)
		self.searchText = tk.Entry(frame, background="White")
		self.searchText.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X)
		tkExtra.Balloon.set(self.searchText, "Filter materials with search string")
		self.list_widgets.append(self.searchText)

		tk.Button(frame, text="X", padx=1, pady=1,
				command=self.clearSearch).pack(side=tk.RIGHT)

		# ----
		frame = tk.Frame(self.frame)
		frame.pack(side=tk.TOP, fill=tk.BOTH, padx=3, pady=2)

		# -- Group --
		lblframe = tk.LabelFrame(frame, text="Group",
				foreground="DarkBlue", padx=2, pady=2)
		lblframe.pack(side=tk.LEFT, fill=tk.Y)

		self.groupList = tkExtra.ExListbox(lblframe,
					borderwidth=0,
					width=20,
					selectmode=tk.EXTENDED,
					selectborderwidth=0,
					background="White",
					takefocus=True,
					exportselection=tk.FALSE)
		self.groupList.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		self.list_widgets.append(self.groupList)
		sb = ttk.Scrollbar(lblframe, orient=tk.VERTICAL, takefocus=False,
				command=self.groupList.yview)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		self.groupList.config(yscrollcommand=sb.set)

		# -- Material --
		lblframe = tk.LabelFrame(frame, text="Material List",
				foreground="DarkBlue", padx=2, pady=2)
		lblframe.pack(side=tk.RIGHT, expand=tk.YES, fill=tk.BOTH)

		self.materialList = tkExtra.MultiListbox(lblframe,
					(('Material',      30, None),
					 ('Density',        9, None),
					 ('Stoichiometry', 10, None)))
		self.materialList.pack(expand=tk.YES, fill=tk.BOTH)
		self.list_widgets.append(self.materialList)

		# ========== Properties ==========
		lblframe = tk.LabelFrame(self.frame, text="Material Properties",
				foreground="DarkBlue", padx=2, pady=2)
		lblframe.pack(side=tk.BOTTOM, expand=tk.YES, fill=tk.BOTH)

		# ---
		frame = tk.Frame(lblframe)
		frame.pack(side=tk.TOP, fill=tk.X)
		tk.Label(frame, text="Title:").pack(side=tk.LEFT)
		self.title = tk.Entry(frame, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.title, "Title of the material")
		self.title.pack(side=tk.RIGHT, expand=tk.YES, fill=tk.X)
		self.properties_widgets.append(self.title)

		# === Specific Info ===
		notebook = ttk.Notebook(lblframe)
		notebook.pack(side=tk.BOTTOM, expand=tk.YES, fill=tk.BOTH)

		# ========== Notes ==========
		pageframe = tk.Frame(notebook)
		notebook.add(pageframe,
				text     = "Notes",
				image    = icons.get("list"),
				compound = tk.LEFT)
		# ---
		row, col = 0, 0
		tk.Label(pageframe, text="Notes:").grid(row=row, column=col, sticky=tk.W)

		# ---
		col += 1
		tk.Label(pageframe, text="Names:").grid(row=row, column=col, sticky=tk.SW)

		row, col = row+1, 0
		rowspan = 6
		self.notes = tkTextEditor.TextEditor(pageframe)
		self.notes.txt.config(width=40, height=5, background="White")
		self.notes.txt.bind("<Button-1>", lambda e,s=self: s.notes.txt.focus_set())
		self.notes.grid(row=row, rowspan=rowspan,
				column=col, #columnspan=colspan,
				padx=3, sticky=tk.NSEW)
		self.properties_widgets.append(self.notes.txt)

		pageframe.grid_columnconfigure(0, weight=1)
		pageframe.grid_rowconfigure(row+rowspan-1, weight=1)

		# --- Names ---
		col += 1
		self.matNameList = tkExtra.ExListbox(pageframe, borderwidth=0,
					selectmode=tk.EXTENDED,
					selectborderwidth=0,
					background="White",
					takefocus=True,
					relief=tk.SUNKEN,
					disabledforeground="Black",
					height=5,
					width=15,
					exportselection=tk.FALSE)
		self.matNameList.grid(row=row, rowspan=rowspan, column=col,
				padx=3, sticky=tk.NSEW)
		self.properties_widgets.append(self.matNameList)

		col += 1
		sb = ttk.Scrollbar(pageframe, orient=tk.VERTICAL, takefocus=False,
				command=self.matNameList.yview)
		sb.grid(row=row, rowspan=rowspan, column=col, sticky=tk.NSEW)
		#self.properties_widgets.append(sb)
		self.matNameList.config(yscrollcommand=sb.set)

		# --
		col += 1
		b = tk.Button(pageframe, image=icons.get("rename"),
				command=self.renameName)
		tkExtra.Balloon.set(b, "Rename")
		b.grid(row=row, column=col, sticky=tk.EW)
		self.properties_widgets.append(b)

		b = tk.Button(pageframe, image=icons.get("add"),
				command=self.addName)
		tkExtra.Balloon.set(b, "Add name")
		b.grid(row=row+1, column=col, sticky=tk.EW)
		self.properties_widgets.append(b)

		b = tk.Button(pageframe, image=icons.get("del"),
				command=self.deleteName)
		tkExtra.Balloon.set(b, "Delete name")
		b.grid(row=row+2, column=col, sticky=tk.EW)
		self.properties_widgets.append(b)

		b = tk.Button(pageframe, image=icons.get("up"),
				command=self.matNameList.moveUp)
		tkExtra.Balloon.set(b, "Move up")
		b.grid(row=row+3, column=col, sticky=tk.EW)
		self.properties_widgets.append(b)

		b = tk.Button(pageframe, image=icons.get("down"),
				command=self.matNameList.moveDown)
		tkExtra.Balloon.set(b, "Move down")
		b.grid(row=row+4, column=col, sticky=tk.EW)
		self.properties_widgets.append(b)

		# ========== Stoichiometry ==========
		pageframe = tk.Frame(notebook)
		notebook.add(pageframe,
				text     = "Stoichiometry",
				image    = icons.get("attributes"),
				compound = tk.LEFT)
		# ---
		row, col = 0, 0
		tk.Label(pageframe, text="Composition:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.compositionType = tkExtra.Combobox(pageframe, width=7,
				command=self.compositionChanged)
		self.compositionType.fill(Material._COMPOSITION)
		self.compositionType.config(disabledforeground='Black')
		tkExtra.Balloon.set(self.compositionType, "Type of mixture")
		self.compositionType.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.compositionType)

		# ---
		col += 1
		tk.Label(pageframe, text="State:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.stateType = tkExtra.Combobox(pageframe, width=7)
		self.stateType.fill(Material._STATE)
		self.stateType.set(Material._STATE[0])
		self.stateType.config(disabledforeground='Black')
		tkExtra.Balloon.set(self.stateType, "Material state")
		self.stateType.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.stateType)

		# ---
		col += 1
		tk.Label(pageframe, text="Density:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.density = tkExtra.FloatEntry(pageframe, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.density, "Density of material in g/cm^3")
		self.density.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.density)

		# ---
		col += 2
		tk.Label(pageframe, text="Group:").grid(row=row, column=col, sticky=tk.SW)

		# --- Stoichiometry ---
		row, col = row+1, 0
		colspan = 6
		self.stoichiometryList = tkExtra.MultiListbox(pageframe,
					(('Z', 3, None),
					 ('A', 4, None),
					 ('El', 4, None),
					 ('Name', 12, None),
					 ('Frac', 10, None)),
					 height=5)
		self.stoichiometryList.config(disabledforeground='Black')
		self.stoichiometryList.grid(row=row, rowspan=4,
				column=col, columnspan=colspan, sticky=tk.NSEW)
		self.properties_widgets.append(self.stoichiometryList)

		# --
		col += colspan
		b = tk.Button(pageframe, image=icons.get("edit"),
				command=self.editElement)
		tkExtra.Balloon.set(b, "Edit element")
		b.grid(row=row, column=col, sticky=tk.EW)
		self.properties_widgets.append(b)

		b = tk.Button(pageframe, image=icons.get("add"),
				command=self.addElement)
		tkExtra.Balloon.set(b, "Add element")
		b.grid(row=row+1, column=col, sticky=tk.EW)
		self.properties_widgets.append(b)

		b = tk.Button(pageframe, image=icons.get("del"),
				command=self.deleteElement)
		tkExtra.Balloon.set(b, "Delete element")
		b.grid(row=row+2, column=col, sticky=tk.EW)
		self.properties_widgets.append(b)

		# --- Group ---
		col += 1
		rowspan=6
		self.matGroupList = tkExtra.ExListbox(pageframe, borderwidth=0,
					selectmode=tk.MULTIPLE,
					selectborderwidth=0,
					background="White",
					takefocus=True,
					relief=tk.SUNKEN,
					width=15,
					disabledforeground="Black",
					exportselection=tk.FALSE)
		self.matGroupList.grid(row=row, rowspan=rowspan, column=col,
				padx=3, sticky=tk.NSEW)
		self.properties_widgets.append(self.matGroupList)
		col += 1
		sb = ttk.Scrollbar(pageframe, orient=tk.VERTICAL, takefocus=False,
				command=self.matGroupList.yview)
		sb.grid(row=row, rowspan=rowspan, column=col, sticky=tk.NSEW)
		#self.properties_widgets.append(sb)
		self.matGroupList.config(yscrollcommand=sb.set)

		# ---
		pageframe.grid_rowconfigure(row+3, weight=1)
		pageframe.grid_columnconfigure(2, weight=1)
		pageframe.grid_columnconfigure(4, weight=1)

		# --- Frame grid weights ---
#		mainframe.grid_columnconfigure(0, weight=1)
#		mainframe.grid_columnconfigure(1, weight=5)
#		mainframe.grid_rowconfigure(0, weight=1)
#		mainframe.grid_rowconfigure(1, weight=3)

		# --- Bindings ---
		self.searchText.bind('<Return>',         self.fillMaterials)
		self.searchText.bind('<KP_Enter>',       self.fillMaterials)
		self.groupList.bind('<<ListboxSelect>>', self.fillMaterials)
		#self.materialList.bindList('<ButtonRelease-1>', self.showMaterial)
		self.materialList.bind('<<ListboxSelect>>', self.showMaterial)

		# Doesn't work!
		self.materialList.bindList('<Double-1>', self.insert2Input)
		self.materialList.bindList('<Return>',   self.insert2Input)
		self.materialList.bindList('<KP_Enter>', self.insert2Input)
		self.matNameList.bind('<Double-1>',      self.renameName)
		self.matNameList.bind('<Control-Key-E>', self.renameName)

		# --- Menus ---
		self.materialList.setPopupMenu([
				("Insert to Input", 0, self.insert2Input,   icons.get("input")),
				None,
				("Edit",   0, self.edit,   icons.get("edit")),
				("Add",    0, self.add,    icons.get("add")),
				("Delete", 0, self.delete, icons.get("x"))])

		self.matNameList.setPopupMenu(
				[("Rename", 0, self.renameName,  icons.get("rename")),
				 ("Add",    0, self.addName,     icons.get("add")),
				 ("Delete", 0, self.deleteName,  icons.get("x"))])

		self.stoichiometryList.setPopupMenu(
				[("Edit",   0, self.editElement,   icons.get("edit")),
				 ("Add",    0, self.addElement,    icons.get("add")),
				 ("Delete", 0, self.deleteElement, icons.get("x"))])

		self.stoichiometryList.bindList('<Double-1>', self.editElement)
		self.stoichiometryList.bindLeftRight()
		self.stoichiometryList.bindList('<Return>',   self.editElement)
		self.stoichiometryList.bindList('<KP_Enter>', self.editElement)

		# ========== Properties ==========
		pageframe = tk.Frame(notebook)
		notebook.add(pageframe,
				text     = "Properties",
				image    = icons.get("properties"),
				compound = tk.LEFT)

		# ---
		row, col = 0, 0
		#tk.Label(pageframe, text="State:").grid(
		#		row=row, column=col, sticky=tk.E)
		#col += 1
		#self.stateType = tk.StringVar()
		#self.stateType.set(Material._STATE[0])
		#o = tkExtra.ExOptionMenu(pageframe, self.stateType,
		#			*Material._STATE)
		#o.config(disabledforeground='Black')
		#tkExtra.Balloon.set(o, "Material state")
		#o.grid(row=row, column=col, sticky=tk.W)
		#self.properties_widgets.append(o)
#
#		# ---
#		row, col = row+1, 0
		tk.Label(pageframe, text="Pressure (atm):").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.pressure = tkExtra.FloatEntry(pageframe, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.pressure, "Gas pressure in atmospheres")
		self.pressure.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.pressure)

		# ---
		row, col = row+1, 0
		tk.Label(pageframe, text="RHOR Factor:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.rhor = tkExtra.FloatEntry(pageframe, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.rhor,
			"Factor to multiply the density of the material, "
			"when calculating density effect parameters")
		self.rhor.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.rhor)

		# ---
		row, col = row+1, 0
		tk.Label(pageframe, text="Ionization Potential (eV):").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.ionization = tkExtra.FloatEntry(pageframe, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.ionization,
			"Average Ionization potential to be used for dE/dx")
		self.ionization.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.ionization)

		# ---
		row, col = row+1, 0
		tk.Label(pageframe, text="DPA Atomic Displacement Energy (eV):").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.dpa_threshold = tkExtra.FloatEntry(pageframe, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.dpa_threshold,
			"DPA Displacement energy threshold averaged "
			"over all crystallographic directions")
		self.dpa_threshold.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.dpa_threshold)

		# ---
		row, col = row+1, 0
		tk.Label(pageframe, text="b_arcdpa:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.barcdpa = tkExtra.FloatEntry(pageframe, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.barcdpa,
			"Nordlund b_arcdpa parameter")
		self.barcdpa.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.barcdpa)

		# ---
		row, col = row+1, 0
		tk.Label(pageframe, text="c_arcdpa:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.carcdpa = tkExtra.FloatEntry(pageframe, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.carcdpa,
			"Nordlund c_arcdpa parameter")
		self.carcdpa.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.carcdpa)

		# ---
		row = 0
		col += 1
		rowspan = 6
		frame = tk.LabelFrame(pageframe, text="Sternheimer parameters",
				foreground="DarkBlue", padx=2, pady=2)
		frame.grid(row=row, rowspan=rowspan, column=col, sticky=tk.NSEW)

		# ---
		row, col = 0,0
		tk.Label(frame, text="Cbar:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.sternheimer_cbar = tkExtra.FloatEntry(frame, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.sternheimer_cbar, "Sternheimer Cbar parameter")
		self.sternheimer_cbar.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.sternheimer_cbar)

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="X0:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.sternheimer_x0 = tkExtra.FloatEntry(frame, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.sternheimer_x0, "Sternheimer X0 parameter")
		self.sternheimer_x0.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.sternheimer_x0)

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="X1:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.sternheimer_x1 = tkExtra.FloatEntry(frame, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.sternheimer_x1, "Sternheimer X1 parameter")
		self.sternheimer_x1.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.sternheimer_x1)

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="a:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.sternheimer_a = tkExtra.FloatEntry(frame, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.sternheimer_a, "Sternheimer a parameter")
		self.sternheimer_a.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.sternheimer_a)

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="m:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.sternheimer_m = tkExtra.FloatEntry(frame, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.sternheimer_m, "Sternheimer m (SK) parameter")
		self.sternheimer_m.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.sternheimer_m)

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="δ0:").grid(
				row=row, column=col, sticky=tk.E)
		col += 1
		self.sternheimer_d0 = tkExtra.FloatEntry(frame, background="White",
				disabledforeground="Black")
		tkExtra.Balloon.set(self.sternheimer_d0, "Sternheimer Delta0 parameter")
		self.sternheimer_d0.grid(row=row, column=col, sticky=tk.W)
		self.properties_widgets.append(self.sternheimer_d0)

		# ================================
		tkExtra.multiConfig(self.properties_widgets, state=tk.DISABLED)
		self.editing = False
		self.shown   = -1
		self.compositionValue = ""
		self.refresh()

	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		super().refresh(what)
		self.fillGroups()
		self.groupList.selection_set(0, tk.END)
		self.fillMaterials()

	# ----------------------------------------------------------------------
	def fillGroups(self, event=None):
		# Fill groups
		self.groupList.delete(0, tk.END)

		glist = Material.groups[:]
		glist.sort()
		for g in glist:
			self.groupList.insert(tk.END, g)

	# ----------------------------------------------------------------------
	def clearSearch(self):
		self.searchText.delete(0,tk.END)
		self.fillMaterials()

	# ----------------------------------------------------------------------
	def fillMaterials(self, event=None):
		# Selected groups
		selgroup = []
		for sel in map(int, self.groupList.curselection()):
			group = self.groupList.get(sel)
			gid = Material.groups.index(group) + 1
			selgroup.append(gid)

		# Search string
		search_string = self.searchText.get().upper().split()
		if not search_string: search_string = None

		# Fill list
		self.materialList.delete(0, tk.END)
		savesort = self.materialList.saveSort()

		for mat in Material.materials.values():
			# Find intersection of groups
			for mg in mat.group:
				if mg in selgroup: break
			else:
				continue

			# Check also search pattern
			if search_string:
				if rexx.search(mat.title.upper(), search_string):
					pass
				elif rexx.search(mat.stoichiometry.upper(), search_string):
					pass
				elif rexx.search(mat.notes.upper(), search_string):
					pass
				elif rexx.search(mat.name.upper(), search_string):
					pass
				else:
					continue

			# Add material
			self.materialList.insert(tk.END,
					(mat.title, mat.density, mat.stoichiometry))

		self.materialList.restoreSort(savesort)
		self.materialList.selection_clear(0, tk.END)
		self.materialList.selection_set(0)
		self.showMaterial()

	# ----------------------------------------------------------------------
	def get(self):
		if self.page is None: return
		if self.editing and self.materialChanged():
			# -----------------------------------
			# Ask user to update the old material
			# -----------------------------------
			if tkFlair.askyesno(f"Update material: {self.material.title}",
			   f"Material \"{self.material.title}\" has changed.\n"
			   "Do you want to update the material?",
			   parent=self.page):
				self.updateMaterial()

		self.editing = False
		tkExtra.multiConfig(self.list_widgets, state=tk.NORMAL)
		tkExtra.multiConfig(self.properties_widgets, state=tk.DISABLED)

	# ----------------------------------------------------------------------
	def showMaterial(self, event=None):
		self.get()

		# Empty material info
		tkExtra.multiConfig(self.properties_widgets, state=tk.NORMAL)
		self.title.delete(0, tk.END)
		self.density.delete(0, tk.END)
		self.notes.txt.delete(1.0, tk.END)
		self.stoichiometryList.delete(0, tk.END)
		self.matGroupList.delete(0, tk.END)
		self.matNameList.delete(0, tk.END)
		self.pressure.delete(0, tk.END)
		self.rhor.delete(0, tk.END)
		self.ionization.delete(0, tk.END)
		self.dpa_threshold.delete(0, tk.END)
		self.barcdpa.delete(0, tk.END)
		self.carcdpa.delete(0, tk.END)
		self.sternheimer_cbar.delete(0, tk.END)
		self.sternheimer_x0.delete(0, tk.END)
		self.sternheimer_x1.delete(0, tk.END)
		self.sternheimer_a.delete(0, tk.END)
		self.sternheimer_m.delete(0, tk.END)
		self.sternheimer_d0.delete(0, tk.END)
		self.compositionType.set("")
		self.stateType.set("")

		# Show selected one
		self.shown = -1
		sel = self.materialList.curselection()
		if len(sel) != 1:
			tkExtra.multiConfig(self.properties_widgets, state=tk.DISABLED)
			return

		# Search for material in database
		self.shown = int(sel[0])
		(title, density, stoichiometry) = self.materialList.get(self.shown)
		mat = Material.materials[title]
		self.material = mat

		# Fill values
		self.title.insert(0, mat.title)
		self.density.insert(0, mat.density)
		self.notes.txt.insert(0.0, mat.notes)
		self.compositionType.set(Material._COMPOSITION[mat.composition])
		self.compositionValue = ""	# Empty
		self.stateType.set(Material._STATE[mat.state])

		self.pressure.insert(         0, mat.pressure)
		self.rhor.insert(             0, mat.rhor)
		self.ionization.insert(       0, mat.ionization)
		self.dpa_threshold.insert(    0, mat.dpa_threshold)
		self.barcdpa.insert(          0, mat.barcdpa)
		self.carcdpa.insert(          0, mat.carcdpa)

		self.sternheimer_cbar.insert( 0, mat.s_cbar)
		self.sternheimer_x0.insert(   0, mat.s_x0)
		self.sternheimer_x1.insert(   0, mat.s_x1)
		self.sternheimer_a.insert(    0, mat.s_a)
		self.sternheimer_m.insert(    0, mat.s_m)
		self.sternheimer_d0.insert(   0, mat.s_d0)

		# Fill Stoichiometry
		for Z, A, f in mat.elements:
			self.insertElement(tk.END, Z, A, f)

		# Fill groups
		for g in mat.group:
			self.matGroupList.insert(tk.END, Material.groups[g-1])

		# Fill names
		for name in mat.name.split():
			self.matNameList.insert(tk.END, name)

		tkExtra.multiConfig(self.properties_widgets, state=tk.DISABLED)

	# ----------------------------------------------------------------------
	def edit(self, event=None):
		if self.shown<0: return

		self.editing = True

		tkExtra.multiConfig(self.list_widgets, state=tk.DISABLED)
		tkExtra.multiConfig(self.properties_widgets, state=tk.NORMAL)

		# Remember composition type
		self.compositionValue = Material._COMPOSITION[self.material.composition]

		# Change group list to tk.BROWSE
		self.title.focus_set()
		self.matGroupList.delete(0, tk.END)
		glist = Material.groups[:]
		glist.sort()
		row = 0
		for g in glist:
			self.matGroupList.insert(tk.END, g)
			gi = Material.groups.index(g)+1
			if gi in self.material.group:
				self.matGroupList.selection_set(row)
			row += 1

	# ----------------------------------------------------------------------
	def compositionChanged(self):
		if self.compositionValue == "": return
		new = self.compositionType.get()
		if new in ("",self.compositionValue): return

		old = self.compositionValue
		self.compositionValue = new

		# XXX Ask the user before?
		if not tkFlair.askyesno("Composition type changed",
			"Do you want to update the elements fractions\n"
			f"from \"{old}\" to \"{new}\"?",
			parent=self.page):
			return

		# Convert everything to mass
		# FIXME I need to total weight of the compound on the transformation
		# look at the Input.materialZAID()
		flist = []
		for i in range(self.stoichiometryList.size()):
			Z, A, S, N, f = self.stoichiometryList.get(i)
			element = Element.byZ(Z)
			if A == "": A = element.Amass()
			mf = 0.0
			if   old == "atom":
				try:
					mf = f*float(A)
				except:
					pass
			elif old == "volume":
				try:
					mf = f*element.density()
				except:
					pass
			else:
				mf = f

			# And then to the new type
			if   new == "atom":
				try:
					f = mf/float(A)
				except:
					pass
			elif new == "volume":
				try:
					f = mf/element.density()
				except:
					pass
			else:
				f = mf
			flist.append(f)

		# Insert items
		sel = list(map(int, self.stoichiometryList.curselection()))
		listbox = self.stoichiometryList.listbox(-1)
		listbox.delete(0, tk.END)
		for i,f in enumerate(flist):
			listbox.insert(tk.END, f)
			if i in sel: listbox.selection_set(i)

	# ----------------------------------------------------------------------
	def ok(self):
		self.updateMaterial()
		self.showMaterial()

	# ----------------------------------------------------------------------
	# Check if material description has changed
	# ----------------------------------------------------------------------
	def materialChanged(self, event=None):
		mat = self.material

		if mat.title != self.title.get().strip(): return True

		if mat.notes != self.notes.txt.get(1.0, tk.END).strip(): return True

		name = " ".join(list(self.matNameList.get(0,tk.END)))
		if mat.name  != name: return True

		composition = Material._COMPOSITION.index(self.compositionType.get())
		if mat.composition != composition: return True

		# find groups
		group = []
		for g in self.matGroupList.curselection():
			gname = self.matGroupList.get(g)
			group.append(Material.groups.index(gname)+1)
		group.sort()

		if " ".join(map(str,mat.group)) != " ".join(map(str,group)):
			return True

		# Do the checksum
		chksum = 0.0
		try: chksum += float(self.density.get())
		except: pass

		try: chksum += float(self.pressure.get())
		except: pass
		try: chksum += float(self.rhor.get())
		except: pass
		try: chksum += float(self.ionization.get())
		except: pass
		try: chksum += float(self.dpa_threshold.get())
		except: pass
		try: chksum += float(self.barcdpa.get())
		except: pass
		try: chksum += float(self.carcdpa.get())
		except: pass

		try: chksum += float(self.sternheimer_cbar.get())
		except: pass
		try: chksum += float(self.sternheimer_x0.get())
		except: pass
		try: chksum += float(self.sternheimer_x1.get())
		except: pass
		try: chksum += float(self.sternheimer_a.get())
		except: pass
		try: chksum += float(self.sternheimer_m.get())
		except: pass
		try: chksum += float(self.sternheimer_d0.get())
		except: pass

		for row in range(self.stoichiometryList.size()):
			Z,A,symbol,name,f = self.stoichiometryList.get(row)
			if A=="": A=0
			try: chksum += float(Z)
			except: pass
			try: chksum += float(A)
			except: pass
			try: chksum += float(f)
			except: pass

		if abs(chksum - mat.chksum) > 1e-10: return True

		return False

	# ----------------------------------------------------------------------
	def updateMaterial(self):
		mat = self.material

		for name in mat.name.split():
			try:
				Material.names[name]
			except KeyError:
				pass
		newtitle = self.title.get()
		if newtitle != mat.title:
			del Material.materials[mat.title]
			Material.materials[newtitle] = mat
			mat.title = newtitle

		mat.name  = " ".join(list(self.matNameList.get(0,tk.END)))
		for name in mat.name.split():
			Material.names[name] = mat
		mat.notes = self.notes.txt.get(1.0, tk.END).strip()

		try: mat.density     = float(self.density.get())
		except: mat.density  = 0.0

		try: mat.pressure    = float(self.pressure.get())
		except: mat.pressure = ""
		try: mat.rhor        = float(self.rhor.get())
		except: mat.rhor     = ""
		try: mat.ionization  = float(self.ionization.get())
		except: mat.ionization = ""
		try: mat.dpa_threshold = float(self.dpa_threshold.get())
		except: mat.dpa_threshold = ""
		try: mat.barcdpa    = float(self.barcdpa.get())
		except: mat.barcdpa = ""
		try: mat.carcdpa    = float(self.carcdpa.get())
		except: mat.carcdpa = ""

		try: mat.s_cbar  = float(self.sternheimer_cbar.get())
		except: mat.s_cbar = ""
		try: mat.s_x0    = float(self.sternheimer_x0.get())
		except: mat.s_x0 = ""
		try: mat.s_x1    = float(self.sternheimer_x1.get())
		except: mat.s_x1 = ""
		try: mat.s_a     = float(self.sternheimer_a.get())
		except: mat.s_a  = ""
		try: mat.s_m     = float(self.sternheimer_m.get())
		except: mat.s_m  = ""
		try: mat.s_d0    = float(self.sternheimer_d0.get())
		except: mat.s_d0 = ""

		mat.group = []
		for g in self.matGroupList.curselection():
			gname = self.matGroupList.get(g)
			mat.group.append(Material.groups.index(gname)+1)
		mat.group.sort()
		mat.composition = Material._COMPOSITION.index(self.compositionType.get())
		mat.state       = Material._STATE.index(self.stateType.get())

		mat.elements = []
		for row in range(self.stoichiometryList.size()):
			Z,A,symbol,name,f = self.stoichiometryList.get(row)
			if A=="": A=0
			try: Z = int(Z)
			except ValueError: Z = 0
			try: A = int(A)
			except ValueError: A = 0
			try: f = float(f)
			except ValueError: f = 0
			mat.elements.append((Z,A,f))

		mat.make()
		mat.modified()
		Material.changed = True

		# Update listbox
		self.materialList.config(state=tk.NORMAL)
		active = self.materialList.index(tk.ACTIVE)
		self.materialList.delete(self.shown)
		self.materialList.insert(self.shown,
				(mat.title, mat.density, mat.stoichiometry))
		if active == self.shown:
			self.materialList.selection_set(self.shown)
			self.materialList.activate(self.shown)

	# ----------------------------------------------------------------------
	def add(self, event=None):
		if self.editing: return
		Material.changed = True
		mat = Material("New Material")
		mat.name = "NewMat"
		mat.group = [1]		# User
		mat.make()
		Material.materials[mat.title] = mat
		self.materialList.insert(tk.END,
			(mat.title, mat.density, mat.stoichiometry))
		self.materialList.selection_clear(0, tk.END)
		self.materialList.selection_set(tk.END)
		self.materialList.activate(tk.END)
		self.materialList.see(tk.END)
		self.showMaterial()
		self.edit()

	# ----------------------------------------------------------------------
	def delete(self, event=None):
		if self.editing: return
		if not tkFlair.askyesno("Delete materials",
				"Do you want to delete the selected materials?", parent=self.page):
			return
		Material.changed = True
		sel = list(self.materialList.curselection())
		sel.reverse()
		for i in sel:
			mat = self.materialList.get(i)
			del Material.materials[mat[0]]
			self.materialList.delete(i)

	# ----------------------------------------------------------------------
	def editElement(self, event=None):
		if not self.editing: return
		listbox = self.page.focus_get()
		if listbox is None: return

		try: active = listbox.index(tk.ACTIVE)
		except: return
		values = self.stoichiometryList.get(tk.ACTIVE)
		lid    = self.stoichiometryList.listboxIndex(listbox)

		# Edit and change value
		if lid < 2:
			# Edit Z or A
			edit = tkExtra.InPlaceInteger(listbox)
			if not edit.value or lid==1:
				return
			try:
				Z = int(edit.value)
			except ValueError:
				return
			element = Element.byZ(Z)
			if element is None: return

		elif lid < 4:
			# Edit Symbol or Name
			edit = InPlaceElement(listbox)
			if edit.value is None:
				return
			element = Element.bySymbol(edit.value)
			if element is None:
				element = Element.byName(edit.value)
			if element:
				Z = element.Z
			else:
				return

		else:
			# Edit fraction
			edit = tkExtra.InPlaceFloat(listbox)
			return

		# Update active row
		self.stoichiometryList.set(active,
		(Z, values[1], element.symbol, element.name, values[4]))
		self.stoichiometryList.selection_set(active)
		self.stoichiometryList.activate(active)

	# ----------------------------------------------------------------------
	def addElement(self, event=None):
		self.insertElement(tk.END, 1, 0, 0.0)
		self.stoichiometryList.selection_clear(0,tk.END)
		self.stoichiometryList.selection_set(tk.END)
		self.stoichiometryList.activate(tk.END)
		self.stoichiometryList.focus_set()
		self.editElement()

	# ----------------------------------------------------------------------
	def insertElement(self, pos, Z, A, f):
		if A==0: A=""
		element = Element.byZ(Z)
		if element:
			self.stoichiometryList.insert(pos,
				(Z, A, element.symbol, element.name, f))

	# ----------------------------------------------------------------------
	def deleteElement(self, event=None):
		sel = list(self.stoichiometryList.curselection())
		if not sel: return
		sel.reverse()
		for i in sel:
			self.stoichiometryList.delete(i)

	# ----------------------------------------------------------------------
	def renameName(self, event=None):
		if self.matNameList["state"] != tk.DISABLED:
			tkExtra.InPlaceEdit(self.matNameList)

	# ----------------------------------------------------------------------
	def addName(self, event=None):
		self.matNameList.insert(tk.END,"")
		self.matNameList.selection_clear(0,tk.END)
		self.matNameList.selection_set(tk.END)
		self.matNameList.activate(tk.END)
		self.renameName()
		if self.matNameList.get(tk.ACTIVE).strip() == "":
			self.matNameList.delete(tk.ACTIVE)
			self.matNameList.selection_set(tk.END)
			self.matNameList.activate(tk.END)

	# ----------------------------------------------------------------------
	def deleteName(self, event=None):
		sel = list(self.matNameList.curselection())
		if not sel: return
		sel.reverse()
		for i in sel:
			self.matNameList.delete(i)

	# ----------------------------------------------------------------------
	# Insert selected materials to input
	# ----------------------------------------------------------------------
	def insert2Input(self, event=None):
		# Insert selected materials
		mats = []
		for sel in self.materialList.curselection():
			mats.append(self.materialList.get(sel)[0])
		matnames, elements = insert2Input(self.flair, mats)
		if matnames: self.modified()

	# ----------------------------------------------------------------------
	# Import materials from input
	# ----------------------------------------------------------------------
	def importFromInput(self, event=None):
		materialDict = self.project.input.materialDict()
		material = self.project.input["MATERIAL"]
		compound = self.project.input["COMPOUND"]

		errorstr = ""
		imported = []

		for name in SelectInputMaterialsDialog(self.page, self.project.input).show():
			# Find material card
			for matcard in material:
				if matcard.name() == name: break
			else:
				errorstr += f"Material {name!r} card not found\n"
				continue

			mat = Material(name)
			mat.name    = name
			mat.group   = [1]		# User
			mat.notes   = matcard.comment()
			mat.density = matcard.numWhat(3)

			for card in compound:
				if card.name() != name: continue
				for i in range(1,card.nwhats(),2):
					f = card.numWhat(i)
					iso  = card.what(i+1)
					if not iso: continue
					if f>0.0:
						mat.composition = 0
					elif iso[0]=="-":
						iso = iso[1:]
						mat.composition = 2
					else:
						f = -f
						mat.composition = 1

					# Find isotope
					ciso = materialDict.get(iso)
					if not ciso:
						errorstr += f"Material {name!r} Isotope {iso!r} not found\n"
						continue
					Z = ciso.intWhat(1)
					if Z==0:
						errorstr += f"Material {name!r} "\
							f"cannot handle compound with compounds {iso!r}\n"
					A = ciso.intWhat(6)
					mat.elements.append((Z,A,f))
			else:
				# material without a compound card
				mat.composition = 0
				mat.elements.append((matcard.intWhat(1),0,1))
			imported.append(name)

			mat.modified()
			mat.make()
			Material.materials[mat.title] = mat
			Material.changed = True

		if imported:
			if errorstr:
				self.flair.notify("Imported Materials",
					f"{len(imported)} materials imported with errors\n{errorstr}",
					tkFlair.NOTIFY_WARNING)
			else:
				self.flair.notify("Imported Materials",
					f"{len(imported)} materials imported")
		elif errorstr:
			self.flair.notify("Import Errors",
					errorstr,
					tkFlair.NOTIFY_ERROR)

#		self.materialList.insert(tk.END,
#			(mat.title, mat.density, mat.stoichiometry))
#		self.materialList.selection_clear(0, tk.END)
#		self.materialList.selection_set(tk.END)
#		self.materialList.activate(tk.END)
#		self.materialList.see(tk.END)

	# ----------------------------------------------------------------------
	def copy(self, event=None):
		global _clipboard
		if self.page.focus_get() in self.stoichiometryList.listboxes():
			del _clipboard[:]
			for i in map(int, self.stoichiometryList.curselection()):
				(Z,A,S,N,f) = self.stoichiometryList.get(i)
				_clipboard.append((Z,A,f))
		else:
			super().copy(event)

	# ----------------------------------------------------------------------
	def cut(self, event=None):
		# XXX cut material, stoichiometryList etc...
		if self.page.focus_get() in self.stoichiometryList.listboxes():
			self.copy()
			self.deleteElement()
		else:
			super().cut(event)

	# ----------------------------------------------------------------------
	def paste(self, event=None):
		# XXX paste material, stoichiometryList etc...
		if self.page.focus_get() in self.stoichiometryList.listboxes():
			pos = tk.END
			sel = self.stoichiometryList.curselection()
			if sel: pos = int(sel[-1])+1
			for Z, A, f in _clipboard:
				self.insertElement(pos, Z, A, f)
				if pos != tk.END: pos += 1
			self.stoichiometryList.selection_clear(0,tk.END)
			self.stoichiometryList.selection_set(pos)
			self.stoichiometryList.activate(pos)
			self.stoichiometryList.see(pos)
		else:
			super().paste(event)

	# ----------------------------------------------------------------------
	def find(self):
		self.notes.findDialog()

	# ----------------------------------------------------------------------
	def replace(self):
		self.notes.replaceDialog()

	# ----------------------------------------------------------------------
	def findNext(self):
		self.notes.findNext()

#===============================================================================
class SelectPropertiesDlg(tk.Toplevel):
	def __init__(self, master, material, prop):
		super().__init__(master)
		self.transient(master)
		self.title(f"{material} properties")

		l = tk.Label(self, text=
			"Select the additional properties of "
			f"material '{material}' to be added in the input",
			foreground="DarkBlue", background="White",
			justify=tk.CENTER)
		l.pack(side=tk.TOP, fill=tk.X)

		self.prop = prop
		self.vars = {}

		v = self._addButton(self,
			"MAT-PROP", "pressure, ionization potential")
		v.pack(side=tk.TOP, fill=tk.X)

		v = self._addButton(self,
			"DPA", "DPA damage threshold")
		v.pack(side=tk.TOP, fill=tk.X)

		v = self._addButton(self,
			"STERNHEI", "Sternheimer parameters")
		v.pack(side=tk.TOP, fill=tk.X)

		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, expand=tk.YES, fill=tk.X)
		self.bind("<Escape>", self.close)
		b = tk.Button(frame, text="Only MATERIAL card", command=self.close)
		b.pack(side=tk.RIGHT)
		b = tk.Button(frame, text="All the above", command=self.insertAll)
		b.pack(side=tk.RIGHT)
		b = tk.Button(frame, text="Selected Only", command=self.insert)
		b.pack(side=tk.RIGHT)

		self.add = []
		x = master.winfo_rootx() + master.winfo_width()//2 -100
		y = master.winfo_rooty() + master.winfo_height()//2 -50
		self.geometry(f"+{x}+{y}")
		self.focus_set()
		try:
			self.wait_visibility()
			self.grab_set()
		except tk.TclError:
			pass
		self.wait_window()

	# ----------------------------------------------------------------------
	def _addButton(self, frame, var, text):
		v = tk.IntVar()
		v.set(0)
		c = tk.Checkbutton(frame, text=f"{var}: {text}",
			variable=v, anchor=tk.W, justify=tk.LEFT,
			background="White")
		if var not in self.prop:
			c["state"] = tk.DISABLED
		self.vars[var] = v
		return c

	# ----------------------------------------------------------------------
	def insert(self):
		for k,v in self.vars.items():
			if v.get():
				self.add.append(k)
		if not self.add:
			messagebox.showinfo("None selected",
				"No property is selected for insertion in the input",
				parent=self)
			return
		self.destroy()

	# ----------------------------------------------------------------------
	def insertAll(self):
		self.add = self.prop
		self.destroy()

	# ----------------------------------------------------------------------
	def close(self, event=None):
		self.destroy()

#===============================================================================
class SelectInputMaterialsDialog(tk.Toplevel):
	def __init__(self, master, input_):
		super().__init__(master)
		self.transient(master)

		# Listbox
		frame = tk.Frame(self)
		frame.pack(expand=tk.YES,fill=tk.BOTH)

		self.listbox = tkExtra.ExListbox(frame, borderwidth=0,
			selectmode=tk.MULTIPLE,
			selectborderwidth=0,
			background="White",
			takefocus=True,
			exportselection=tk.FALSE)
		self.listbox.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

		sb = ttk.Scrollbar(frame, orient=tk.VERTICAL, command=self.listbox.yview)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		self.listbox.config(yscrollcommand=sb.set)

		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM)

		b = tk.Button(frame,text="Cancel",
				image=icons.get("x"),
				compound=tk.LEFT,
				takefocus=tk.FALSE,
				command=self.cancel)
		b.pack(side=tk.RIGHT)
		b = tk.Button(frame,text="All",
				image=icons.get("select_all"),
				compound=tk.LEFT,
				takefocus=tk.FALSE,
				command=self.all)
		b.pack(side=tk.RIGHT)
		b = tk.Button(frame,text="Ok",
				image=icons.get("ok"),
				compound=tk.LEFT,
				takefocus=tk.FALSE,
				command=self.ok)
		b.pack(side=tk.RIGHT)
		self.bind("<Escape>", self.cancel)
		self.bind("<Control-Key-a>", self.all)
		self.protocol("WM_DELETE_WINDOW", self.cancel)
		self.title("Import Materials")

		# Fill template box
		mats = set()
		for card in input_["MATERIAL"]:
			mats.add(card.name())
		for m in sorted(mats):
			self.listbox.insert(tk.END, m)
		self.selection = []

	# ----------------------------------------------------------------------
	def show(self):
		self.focus_set()
		self.listbox.focus_set()
		try:
			self.wait_visibility()
			self.grab_set()
		except tk.TclError:
			pass
		self.wait_window()
		return self.selection

	# ----------------------------------------------------------------------
	def all(self, event=None):
		self.listbox.selection_set(0,tk.END)
		self.ok()

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		self.selection = [self.listbox.get(x)
					for x in self.listbox.curselection()]
		self.destroy()

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		self.destroy()

#===============================================================================
class InPlaceElement(tkExtra.InPlaceEdit):
	# ----------------------------------------------------------------------
	def createWidget(self):
		self.edit = tk.Entry(self.frame, width=10, background="White")
		self.edit.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		b = tk.Button(self.frame, image=icons.get("toi"),
			padx=0, pady=0, command=self.showTable)
		b.pack(side=tk.RIGHT)
		self.edit.focus_set()

	# ----------------------------------------------------------------------
	def showTable(self):
		self.frame.unbind("<FocusOut>")
		element = PeriodicTable.show(None, False, True)
		self.frame.bind("<FocusOut>", self.cancel)
		self.edit.focus_set()
		if element is not None:
			self.edit.delete(0,tk.END)
			self.edit.insert(0, element.symbol)
