#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   07-Dec-2007

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import log
import csg
import bmath
from deck import Card

_DENSITY_LIMIT = 0.1	# Ligther than that is ignored
#_FAR_AWAY      = 1E10

#-------------------------------------------------------------------------------
# Create rotation matrix using the X,Y,Z vectors
#-------------------------------------------------------------------------------
def write_matrix(fout, X, Y, Z, T):
	M = bmath.Matrix()
	M.make(X,Y,Z)
	M.transpose()
	for i in range(3): M[i][3] = T[i]
	fout.write("\tmultmatrix(m=[\n")
	for r in range(4):
		if r==3: comma=""
		else: comma=","
		fout.write(f"\t\t\t{M[r]}{comma}\n")
	fout.write("\t\t])\n")

#-------------------------------------------------------------------------------
# Create rotation matrix using the Z vector
#-------------------------------------------------------------------------------
def write_matrixZ(fout, Z, T):
	X = Z.orthogonal()
	Y = Z ^ X
	write_matrix(fout,X,Y,Z,T)

##-------------------------------------------------------------------------------
def writePlane0(fout):
	fout.write("\t\tpolyhedron( [[0,INF,0], [INF,-INF/2,0], [-INF,-INF/2,0], [0,0,-INF]],\n")
	fout.write("\t\t\t[[0,1,2], [0,3,1], [1,3,2], [2,3,0]] );")

#-------------------------------------------------------------------------------
# Return a valid povray name
#-------------------------------------------------------------------------------
#def name(s):
#	# First convert all invalid characters to underscore (_)
#	# Minimize the occurrences of _
#	s = s.encode("ascii","replace")
#	new = s.translate(_TRANS)
#	while True:
#		old = new
#		new = new.replace("__","_")
#		if old == new: break
#	return old

#-------------------------------------------------------------------------------
def header(fout, inp):
	fout.write("// flair converted to openscad\n")
	try:
		title = inp.cards["TITLE"][0]
		fout.write(f"// Title: {title.text()}\n\n")
	except:
		pass
	fout.write("INF=1000000;\n")
	fout.write("SPH=32;\n")
	fout.write("CYL=32;\n")
	fout.write("\n")

#-------------------------------------------------------------------------------
def exportBodies(fout, inp):
	fout.write("\n// ---- Bodies ----\n")

	bodies = []

	# Loop all cards
	for card in inp.cardlist:
		if not card.isGeo(): continue
		if len(card.tag) != 3: continue
		if card.ignore(): continue
		if card.tag == "END": continue

		name = card.name()
		if name == "": continue
		name = "b_"+name
		bodies.append(name)

		fout.write("\n")
		for line in str(card).splitlines():
			if line.startswith("*"): line = line[1:]
			fout.write(f"//{line.rstrip()}\n")
		fout.write(f"module {name}()\n{{\n")

		if   card.tag == "RPP":
			fout.write(f"\ttranslate({card.bodyP()})\n")
			fout.write("\t\tcube([%g,%g,%g], center=false);\n"%
				( card.numWhat(2)-card.numWhat(1),
				  card.numWhat(4)-card.numWhat(3),
				  card.numWhat(6)-card.numWhat(5)))

		elif card.tag == "BOX":
			P = card.bodyP()
			X = card.bodyX()
			Y = card.bodyY()
			Z = card.bodyZ()
			lx = X.norm()
			ly = Y.norm()
			lz = Z.norm()
			write_matrix(fout, X,Y,Z,P)
			fout.write(f"\t\tcube([{lx},{ly},{lz}], center=false);\n")

		elif card.tag == "SPH":
			fout.write(f"\ttranslate({card.bodyP()})\n")
			fout.write(f"\t\tsphere({card.bodyR()},$fn=SPH);\n")

		elif card.tag == "RCC":
			P = card.bodyP()
			Z = card.bodyZ()
			h = Z.norm()
			write_matrixZ(fout,Z,P)
			fout.write(f"\tcylinder(h={h}, r={card.bodyR()}, center=false, $fn=CYL);\n")

		elif card.tag == "TRC":
			P = card.bodyP()
			Z = card.bodyZ()
			h = Z.norm()
			write_matrixZ(fout,Z,P)
			fout.write(f"\tcylinder(h={h}, r1={card.numWhat(7)}, "
				f"r2={card.numWhat(8)}, center=false, $fn=CYL);\n")

		#elif card.tag == "ELL":
		#	WriteEllipsoid

		#elif card.tag == "WED":
		#	WriteAngleWedge

		#elif card.tag == "ARB"
		#	WritePolyhedron

		elif card.tag == "XYP":
			fout.write(f"\ttranslate({card.bodyP()})\n")
			writePlane0(fout)

		elif card.tag == "XZP":
			fout.write(f"\ttranslate({card.bodyP()})\n")
			fout.write("\trotate(90,[1,0,0])\n")
			writePlane0(fout)

		elif card.tag == "YZP":
			fout.write(f"\ttranslate({card.bodyP()})\n")
			fout.write("\trotate(90,[0,1,0])\n")
			writePlane0(fout)

		elif card.tag == "PLA":
			Z = card.bodyN()
			Z.norm()
			write_matrixZ(fout, Z, card.bodyP())
			writePlane0(fout)

		elif card.tag in ("XCC", "YCC", "ZCC"):
			fout.write(f"\ttranslate({card.bodyP()})\n")
			if card.tag == "XCC":
				fout.write("\t\trotate(90,[0,1,0])\n")
			elif card.tag == "YCC":
				fout.write("\t\trotate(90,[1,0,0])\n")

			fout.write(f"\t\t\tcylinder(h=INF,r={card.bodyR()},center=true,$fn=CYL);")

		elif card.tag == "REC":
			P = card.bodyP()
			X = card.bodyX()
			Y = card.bodyY()
			Z = card.bodyZ()
			lx = X.norm()
			ly = Y.norm()
			lz = Z.norm()
			write_matrix(fout,X,Y,Z,P)
			fout.write(f"\tscale([{lx},{ly},1.0])\n")
			fout.write(f"\t\tcylinder(h={lz}, r=1, center=false, $fn=CYL);\n")

		elif card.tag in ("XEC", "YEC", "ZEC"):
			fout.write(f"\ttranslate({card.bodyP()})\n")
			if card.tag == "XCC":
				fout.write("\t\trotate(90,[0,1,0])\n")
			elif card.tag == "YCC":
				fout.write("\t\trotate(90,[1,0,0])\n")

			fout.write(f"\t\tscale([{card.numWhat(3)},{card.numWhat(4)},1.0])\n")
			fout.write("\t\t\tcylinder(h=INF, r=1, center=true, $fn=CYL);\n")

		else:
			raise Exception(f"Unknown body {card.tag}")

		fout.write("\n}\n")
	return bodies

#-------------------------------------------------------------------------------
def _exportRegion(fout, region, expr):
	# Find material
	try:
		material = region[Card.MATERIAL]
	except:
		log.error(f"\nError: region {region.name()!r} is not assigned material\n")
		return True

	try: density = float(material.what(3))
	except: density = 0.0
	if density <= _DENSITY_LIMIT: return False

	# Write card
	fout.write(f"\n// Region: {region.name()}\tMaterial: {material.name()}\n\n")
	for line in str(region).splitlines():
		if line.startswith("*"): line = line[1:]
		fout.write(f"//{line}\n")

	fout.write(f"module {region.name()}()\n{{\n")

	# Check expression
	exp = csg.tokenize(expr)
	try:
		csg.exp2rpn(exp)
		if exp and not csg.check(exp):
			log.error(f"REGION {region.name()!r} has invalid expression")
		else:
			stack = []
			for token in exp:
				if token in ("-", "+", "|"): #, "@"):
					if   token == "+": op = "intersection() {"
					elif token == "-": op = "difference() {"
					elif token == "|": op = "union() {"
					b = stack.pop()
					a = stack.pop()

					if isinstance(a,list) and a[0]==op:
						if isinstance(b,str):
							a.insert(len(a)-1, "\t"+b)
						else:
							a[len(a)-1:len(a)-1] = ["\t"+x for x in b]
						lst = a

					elif isinstance(b,list) and b[0]==op and token in ("+","|"):
						if isinstance(a,str):
							b.insert(1, "\t"+a)
						else:
							b[1:1] = ["\t"+x for x in a]
						lst = b

					else:
						lst = [op]
						if isinstance(a,str):
							lst.append("\t"+a)
						else:
							for i in a: lst.append("\t"+i)
						if isinstance(b,str):
							lst.append("\t"+b)
						else:
							for i in b: lst.append("\t"+i)
						lst.append("}")

					stack.append(lst)

				else:
					stack.append(f"b_{token}();")

	except csg.CSGException:
		log.error("Unbalanced parenthesis")

	lst = stack.pop()
	if isinstance(lst,str):
		fout.write(f"\t{lst}\n")
	else:
		for i in lst:
			fout.write(f"\t{i}\n")
	fout.write("}\n")

	return True

#-------------------------------------------------------------------------------
def exportRegions(fout, fluka):
	fout.write("\n\n// ------ Regions ------\n")
	fluka.regionProperties()

	region = None
	expr   = ""
	regions = []
	for card in fluka.cards["REGION"]:
		if card.ignore(): continue
		if card.name()=="&":
			# append to previous expression
			expr += card.expr()
			continue

		if region is None:
			# Remember starting region
			region = card
			expr   = region.expr()
			continue

		if _exportRegion(fout, region, expr):
			regions.append(region.name())

		# Remember starting region
		region = card
		expr   = region.expr()

	# Last region
	if region:
		if _exportRegion(fout, region, expr):
			regions.append(region.name())

	return regions

#-------------------------------------------------------------------------------
def export(fluka, filename):
	with open(filename, "w") as fout:
		fluka.preprocess()
		header(fout, fluka)
		bodies = exportBodies(fout, fluka)

		for body in bodies:
			fout.write(f"//{body}();\n")

		regions = exportRegions(fout, fluka)

		fout.write("\n")
		for region in regions:
			fout.write(f"{region}();\n")
