#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   03-Oct-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import sys
import time
import tkinter as tk
from tkinter import ttk

import log
import ansi
import icons
import Global
import Ribbon
import tkFlair
import tkExtra
import bFileDialog
import FlairRibbon
from process import Process

stdout_orig = sys.stdout
stderr_orig = sys.stderr

_ORDER = {"flair"   : 0,
	  "Compile" : 1,
	  "Run"     : 2,
	  "Data"    : 3,
	  "Plot"    : 4,
	  "Dicom"   : 5
	 }

# Colors to match with process status
#           0        1           2          3          4         5
#STATUS = ["Idle", "Waiting", "Running", "Finished", "Error", "Killed"]
_COLOR  = ["Black", "Orange", "Blue",    "DarkGreen", "Red", "DarkRed"]

_TITLE_COLOR   = "DarkRed"
_COMMAND_COLOR = "DarkBlue"
_INPUT_COLOR   = "DarkGreen"
_WARNING_COLOR = "Purple"
_ERROR_COLOR   = "Red"

_UNICODE_RANGE = re.compile('[^\u0000-\uFFFF]', re.UNICODE)

#===============================================================================
# NullLog
# Empty wrapper to display messages on screen
#===============================================================================
class NullLog:
	def __init__(self, process):
		print(f"{ansi.BOLD}{time.ctime()} {process}{ansi.NORMAL}")

	# ----------------------------------------------------------------------
	def write(self, text):
		print(text,end="")

	# ----------------------------------------------------------------------
	def writeln(self, text):
		print(text)

	# ----------------------------------------------------------------------
	def __call__(self, text):
		self.writeln(text)

	# ----------------------------------------------------------------------
	def warning(self, text):
		self.writeln(f"{log.WARNING}{text}")

	# ----------------------------------------------------------------------
	def error(self, text):
		self.writeln(f"{log.ERROR}{text}")

#===============================================================================
# Log
#===============================================================================
class Log:
	def __init__(self, page, typ, name):
		self.page = page
		self.type = typ
		self.idx  = 0
		self.name = name
		self._status  = 0
		self._percent = 0
		self.reset()

	# ----------------------------------------------------------------------
	def reset(self):
		self.messages  = ""
		self._status   = 0
	clear = reset

	# ----------------------------------------------------------------------
	def status(self, st, p):
		if self._status!=st or self._percent!=p:
			self._status  = st
			self._percent = p
			self.page.updateStatus(self)

	# ----------------------------------------------------------------------
	def statusText(self):
		if self._status == Process.STATUS_RUNNING:
			return f"{Process.STATUS[self._status]} {self._percent}%"
		else:
			return Process.STATUS[self._status]

	# ----------------------------------------------------------------------
	def statusColor(self):
		return _COLOR[self._status]

	# ----------------------------------------------------------------------
	def timestamp(self, msg):
		self.messages += "%s %s\n" % \
			(msg, time.strftime(bFileDialog._TIME_FORMAT, time.localtime(time.time())))

	# ----------------------------------------------------------------------
	def write(self, text):
		if Global.verbose or Global.developer:
			print(text,end="")
		#if "Unbalanced parenthesis" in text: import traceback; traceback.print_stack()
		if not self.messages: self.timestamp("Started")
		self.messages += str(text)
		self.page.refreshLate()		# FIXME not very brilliant
						# it needs to insert just the last string
						# to insert(tk.END, text\n)
						# if the proper item is selected
						# otherwise highlight the item
						# that new info is printed

	# ----------------------------------------------------------------------
	def writeln(self, text):
		self.write(f"{text}\n")

	# ----------------------------------------------------------------------
	def warning(self, text):
		prefix = log.WARNING
		if Global.verbose and Global._ansi:
			print(ansi.BOLD,end="")
		for line in text.splitlines():
			self.write(f"{prefix}{line}\n")
			if prefix==log.WARNING:
				prefix = log.WARNINGCONT
		if Global.verbose and Global._ansi:
			print(ansi.NORMAL,end="")

	# ----------------------------------------------------------------------
	def error(self, text):
		prefix = log.ERROR
		if Global.verbose and Global._ansi:
			print(ansi.BOLD,end="")
		for line in text.splitlines():
			self.write(f"{prefix}{line}\n")
			if prefix==log.ERROR:
				prefix = log.ERRORCONT
		if Global.verbose and Global._ansi:
			print(ansi.NORMAL,end="")

	# ----------------------------------------------------------------------
	def __call__(self, text):
		self.writeln(text)

	# ----------------------------------------------------------------------
	def size(self):
		return len(self.messages)

#===============================================================================
# Output Text Stream
#===============================================================================
class OutputBaseStream:
	txt = None

	#-----------------------------------------------------------------------
	def __init__(self, master):
		if OutputBaseStream.txt is None:
			OutputBaseStream.txt = tk.Text(master)

	# ----------------------------------------------------------------------
	def write(self, text):
		pass

#===============================================================================
class OutputStream(OutputBaseStream):
	#def __init__(self, master):
	#	super().__init__(master)

	# ----------------------------------------------------------------------
	def write(self, text):
		OutputBaseStream.txt.insert(tk.END, text)

#===============================================================================
class ErrorStream(OutputBaseStream):
	#def __init__(self, master):
	#	super().__init__(master)

	# ----------------------------------------------------------------------
	def write(self, text):
		OutputBaseStream.txt.insert(tk.END, text)

#===============================================================================
# Output frame class
#===============================================================================
class OutputPage(FlairRibbon.FlairPage):
	"""Monitor the progress of a process and its output"""

	_name_ = "Output"
	_icon_ = "output"

	# ----------------------------------------------------------------------
	def init(self):
		self._log = []	# list of log structures
		self._refreshAfter = None

	# ----------------------------------------------------------------------
	# create the ribbon buttons
	# ----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		group = Ribbon.LabelGroup(self.ribbon, "Process")
		group.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.TRUE, padx=0, pady=0)

		# ---
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("play32"),
				text="Run",
				compound=tk.TOP,
				anchor=tk.CENTER,
				state=tk.DISABLED)
#				command=self.execute)
		b.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		tkExtra.Balloon.set(b, "Execute selected processes [Ctrl-Enter]")

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("stop32"),
				text="Stop",
				compound=tk.TOP,
				anchor=tk.CENTER,
				state=tk.DISABLED)
#				command=self.stopRun)
		b.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		tkExtra.Balloon.set(b, "Stop selected processes")

		# ---
#		b = Ribbon.LabelButton(group.frame, image=icons.get("x32"),
#				text="Files",
#				compound=tk.TOP,
#				anchor=tk.CENTER,
#				state=tk.DISABLED,
##				command=self.cleanFiles,
#				background=Ribbon._BACKGROUND)
#		b.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
#		tkExtra.Balloon.set(b, "Delete all generated files from this process.")

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("clean32"),
				text="Clean",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.deleteOutput)
		b.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		tkExtra.Balloon.set(b, "Clean output of selected processes")

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("x32"),
				text="Close",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.close)
		b.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		tkExtra.Balloon.set(b, "Close output")

		# ===== Action ===
		group = Ribbon.LabelGroup(self.ribbon, "Action")
		group.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.TRUE, padx=0, pady=0)

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("save32"),
				text="Save",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.save)
		b.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		tkExtra.Balloon.set(b, "Save selected process output to file")

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("input32"),
				text="Go to",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.goto)
		b.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		tkExtra.Balloon.set(b, "Show selected item on Input Editor or in Geometry Editor")

		# ========== Find ===========
		group = Ribbon.LabelGroup(self.ribbon, "Search")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		self.findString = tkExtra.LabelEntry(group.frame,
					"Search",
					"DarkGray",
					width=20,
					font=Ribbon._FONT,
					background="White") #Ribbon._BACKGROUND,
		self.findString.pack(side=tk.LEFT)
		tkExtra.Balloon.set(self.findString , "Find cards with string")
		self.findString.bind("<Return>",   self.find)
		self.findString.bind("<KP_Enter>", self.find)
		self.findString.bind("<Button-3>", self.findClear)
		self.findString.bind("<Control-Key-g>", self.find)
		self.findString.bind("<Control-Key-f>", lambda e : self.findString.set_focus())

		self.findCase = tk.BooleanVar()
		self.findCase.set(False)
		b = tk.Checkbutton(group.frame,
				variable=self.findCase,
				padx=0,
				pady=0,
				borderwidth=0,
				highlightthickness=0,
				font=Ribbon._FONT,
				activebackground="LightYellow",
				background=Ribbon._BACKGROUND)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b, "Case sensitive")

		b = Ribbon.LabelButton(group.frame, image=icons.get("find"),
				command=self.find)
		b.pack(side=tk.LEFT)
		tkExtra.Balloon.set(b, "Find cards with string")

		return self.ribbon

	#----------------------------------------------------------------------
	# Create Project page
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage()

		# --- Run-Usrxxx splitter ---
		splitter = tk.PanedWindow(self.frame, orient=tk.VERTICAL,
			showhandle=0,
			sashrelief=tk.RAISED, sashwidth=4, opaqueresize=1)
		splitter.pack(expand=tk.YES, fill=tk.BOTH)

		# List of commands
		self.listbox = tkExtra.MultiListbox(splitter,
				(("Type",     8, None),
				 ("Process", 20, None),
				 ("Status",  10, None)),
				 height=6)
		self.listbox.sortAssist = None
		splitter.add(self.listbox)

		# Text with output messages
		frame = tk.Frame(splitter)
		splitter.add(frame)

		self.text = tk.Text(frame,
					state=tk.DISABLED,
					takefocus=tk.TRUE,
					background="White",
					height=6)
		self.text.grid(row=0, column=0, sticky=tk.NSEW)
		sbv = ttk.Scrollbar(frame, orient=tk.VERTICAL, takefocus=False,
			command=self.text.yview)
		sbv.grid(row=0, column=1, sticky=tk.NS)
		sbh = ttk.Scrollbar(frame, orient=tk.HORIZONTAL, takefocus=False,
			command=self.text.xview)
		sbh.grid(row=1, column=0, sticky=tk.EW)
		self.text.config(yscrollcommand=sbv.set,
				xscrollcommand=sbh.set)

		self.text.bind("<ButtonRelease-1>", self.release)
		self.text.bind("<Double-1>",        self.double)
		self.text.bind("<Up>",    lambda e,s=self : s.text.yview(tk.SCROLL,-1,tk.UNITS))
		self.text.bind("<Down>",  lambda e,s=self : s.text.yview(tk.SCROLL, 1,tk.UNITS))

		frame.grid_columnconfigure(0, weight=1)
		frame.grid_rowconfigure(0, weight=1)

		self.populateLog()

		self.text.tag_configure("title",   foreground=_TITLE_COLOR)
		self.text.tag_configure("command", foreground=_COMMAND_COLOR)
		self.text.tag_configure("input",   foreground=_INPUT_COLOR)
		self.text.tag_configure("warning", foreground=_WARNING_COLOR)
		self.text.tag_configure("error",   foreground=_ERROR_COLOR)

		self.listbox.bind("<<ListboxSelect>>", self.refresh)

	# ----------------------------------------------------------------------
	def populateLog(self):
		self.listbox.delete(0,tk.END)
		for l in self._log:
			self.listbox.insert(tk.END, (l.type, l.name, l.statusText()))
		self.listbox.selection_set(tk.END)
		self.listbox.see(tk.END)

	# ----------------------------------------------------------------------
	# Return log with name
	# ----------------------------------------------------------------------
	def newLog(self, typ, name, append=False):
		self.listbox.selection_clear(0,tk.END)
		# Check for an existing log with the same name/type
		for i, l in enumerate(self._log):
			if l.name == name and l.type==typ:
				if not append: l.reset()
				self.listbox.selection_set(i)
				self.refresh()
				return l

		newlog = Log(self,typ,name)
		self.insert(newlog)
		self.refresh()
		return newlog

	# ----------------------------------------------------------------------
	# Find position to insert log
	# ----------------------------------------------------------------------
	def insert(self, newlog):
		order = _ORDER.get(newlog.type, len(_ORDER))

		# Find the last
		for i,l in enumerate(self._log):
			if _ORDER.get(l.type, len(_ORDER)) > order:
				self._log.insert(i,newlog)
				self.listbox.insert(i, (newlog.type, newlog.name, newlog.statusText()))
				self.listbox.selection_set(i)
				self.listbox.see(i)
				break
		else:
			self._log.append(newlog)
			self.listbox.insert(tk.END, (newlog.type, newlog.name, newlog.statusText()))
			self.listbox.selection_set(tk.END)
			self.listbox.see(tk.END)

		for i,l in enumerate(self._log):
			l.idx = i

	# ----------------------------------------------------------------------
	def refreshLate(self):
		# FIXME Normally it should be protected with semaphore!!!!
		if self._refreshAfter is not None:
			if self.page:
				self.page.after_cancel(self._refreshAfter)
		if self.page:
			try:
				self._refreshAfter = self.page.after(1000, self.refresh)
			except tk.TclError:
				# catch possible infinite loop error when multiple writes happen
				pass

	# ----------------------------------------------------------------------
	# Update the display
	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		self._refreshAfter = None
		if self.page is None: return
		super().refresh(what)
		try:
			sel = self.listbox.curselection()
		except tk.TclError:
			self.page = None
			return

		self.text.config(state=tk.NORMAL)
		self.text.delete("0.0",tk.END)
		if len(sel)==1:
			idx = int(sel[0])
			try:
				logger = self._log[idx]
				try:
					self.text.insert("0.0", logger.messages)
				except tk.TclError:
					self.text.insert("0.0", _UNICODE_RANGE.sub("?",logger.messages))
				self.text.see(tk.END)
			except IndexError:
				pass

		self.text.config(state=tk.DISABLED)
		self.colorize()

	# ----------------------------------------------------------------------
	def deleteOutput(self):
		for sel in self.listbox.curselection():
			idx = int(sel)
			try:
				logger = self._log[idx]
				logger.reset()
			except IndexError:
				pass
		self.refresh()

	# ----------------------------------------------------------------------
	# Find warning and error patterns and mark them
	# ----------------------------------------------------------------------
	def colorize(self, begin="1.0", end=tk.END):
		def mark(tag, target):
			start = begin
			while True:
				start = pos = self.text.search(target, start, end)
				if not pos: break
				# find end of error
				while True:
					start = self.text.index(start + "+1l")	# next line
					if self.text.get(start, f"{start}+3c")!="   ": break
				self.text.tag_add(tag, pos, start)
		mark("title",   "*-*")
		mark("command", "Cmd:")
		mark("input",   "<<<")
		mark("warning", log.WARNING_PREFIX)
		mark("error",   log.ERROR_PREFIX)

	# ----------------------------------------------------------------------
	def updateStatus(self, logger):
		sel = self.listbox.selection_includes(logger.idx)
		self.listbox.set(logger.idx, (logger.type, logger.name, logger.statusText()))
		if logger._status != 0:
			self.listbox.itemconfig(logger.idx, foreground=logger.statusColor())

		if sel: self.listbox.selection_set(logger.idx)

	# ----------------------------------------------------------------------
	def copy(self, event=None):
		if event.widget is self.text: return
		self.text.event_generate("<<Copy>>")
	cut = copy

	# ----------------------------------------------------------------------
	def close(self):
		self.hide(True)

	# ----------------------------------------------------------------------
	def mouseLocation(self, event):
		index = self.text.index(f"@{event.x},{event.y}")
		return map(int,index.split("."))

	# ----------------------------------------------------------------------
	def release(self, event):
		self._focusIn()
		self.text.focus_set()
		if not self.text.tag_nextrange(tk.SEL,"1.0"):
			row,col = self.mouseLocation(event)
			self.text.tag_remove(tk.SEL, "1.0", tk.END)
			self.text.tag_add(tk.SEL, f"{row}.0", f"{row}.end")

	# ----------------------------------------------------------------------
	def double(self, event):
		#index = self.text.index("@%d,%d"%(event.x,event.y))
		row,col = self.mouseLocation(event)
		self.flair.showCard(self.text.get(f"{row}.0",f"{row}.end"))
		return "break"

	# ----------------------------------------------------------------------
	def save(self):
		filename = bFileDialog.asksaveasfilename(master=self.page,
					title="Save output",
					initialdir=self.project.dir(),
					filetypes=[
						("Output file",("*.txt","*.out")),
						("Text file","*.txt"),
						("All","*")])
		if not filename: return
		with open(filename, "w", encoding="UTF-8", errors="replace") as f:
			f.write(self.text.get("1.0",tk.END))
			self.flair.notify("Output Saved",
					f"Output saved to file {filename!r}")

	# ----------------------------------------------------------------------
	def goto(self):
		tag = self.text.tag_nextrange(tk.SEL,"1.0")
		if not tag:
			self.flair.notify("Nothing selected",
				"No line selected to display on Input or Geometry Editor",
				tkFlair.NOTIFY_WARNING,
				showOnOutput=False)
			return
		self.flair.showCard(self.text.get(*tag))

	#----------------------------------------------------------------------
	def find(self, event=None):
		findStr = self.findString.get()
		if not findStr: return
		self.text.tag_remove(tk.SEL, "0.0", tk.END)
		try:
			index = self.text.search(findStr,
				"insert + 1 chars",
				nocase = self.findCase.get())
			self.text.mark_set('insert', index)
			endex = f"insert + {len(findStr)} chars"
			self.text.tag_add(tk.SEL, index, endex)
			self.text.see(index)
			return True
		except tk.TclError:
			self.flair.notify("Not found",
				f"Target '{findStr}' not found",
				tkFlair.NOTIFY_WARNING,
				showOnOutput=False)
			return False

	# --------------------------------------------------------------------
	def findNext(self, event=None):
		self.find()

	#----------------------------------------------------------------------
	def findClear(self, event=None):
		self.findString.delete(0, tk.END)
