#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Sep-2006

# This file implements the Element and PeriodicTable
# classes.
# The first displays a dialog with the element properties.
# The creates the Medeleyv table

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import tkinter as tk
from tkinter import ttk

import icons
import Ribbon
import tkExtra
import FlairRibbon
from element import Element

#===============================================================================
# Static-local Variables
#-------------------------------------------------------------------------------
_PTBL_ROWS = 12
_PTBL_COLS = 20

#-------------------------------------------------------------------------------
# Labels of periodic table
#-------------------------------------------------------------------------------
_ptableLbl = [	"1",			# -1
		"2",			# -2
		"3",			# -3
		"4",			# -4
		"5",			# -5
		"6",			# -6
		"7",			# -7
		"8",			# -8
		"9",			# -9
		"10",			# -10
		"11",			# -11
		"12",			# -12
		"13",			# -13
		"14",			# -14
		"15",			# -15
		"16",			# -16
		"17",			# -17
		"18",			# -18
		"Group",		# -19
		"Period",		# -20
		"*",			# -21
		"**",			# -22
		"*\nLanthanides",	# -23
		"**\nActinides"]	# -24

#-------------------------------------------------------------------------------
# Positive is given the Z of the elements
# Zero is empty space
# Negative is a string indexed from the above array ptable[abs(num)]
#-------------------------------------------------------------------------------
_ptable = [
	[-19, -1, -2,  0, -3, -4, -5, -6, -7, -8, -9,-10,-11,-12,-13,-14,-15,-16,-17,-18],
	[-20,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
	[ -1,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  2],
	[ -2,  3,  4,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  5,  6,  7,  8,  9, 10],
	[ -3, 11, 12,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 13, 14, 15, 16, 17, 18],
	[ -4, 19, 20,  0, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36],
	[ -5, 37, 38,  0, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54],
	[ -6, 55, 56,-21, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86],
	[ -7, 87, 88,-22,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118],
	[  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
	[-23,  0,  0,-21, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70,  0,  0],
	[-24,  0,  0,-22, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99,100,101,102,  0,  0] ]

#-------------------------------------------------------------------------------
#            A:0      B:1        C:2        D:3        E:4         F:5       G:6
#-------------------------------------------------------------------------------
#               w         A         B           K          O
_bcolor  = [ "White", "#CCCCCC", "#CCCCFF", "#FFFFCC", "#FFCCCC",
#               G         M         C          P           T
	    "#CCFFCC", "#FFCC00", "#CCFFFF", "#FFCCFF", "#AADDFF" ]
#0 1 2 3 4 5 6 7 8 9
(w,A,B,K,O,G,M,C,P,T) = range(10)
_ecolor = [	[ A,A,A,w,A, A,A,A,A,A, A,A,A,A,A, A,A,A,A,A ],
		[ A,w,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,w,w ],
		[ A,G,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,w,O ],
		[ A,M,C,w,w, w,w,w,w,w, w,w,w,w,G, G,G,G,K,O ],
		[ A,M,C,w,w, w,w,w,w,w, w,w,w,w,P, G,G,G,K,O ],
		[ A,M,C,w,B, B,B,B,B,B, B,B,B,B,P, P,G,G,K,O ],
		[ A,M,C,w,B, B,B,B,B,B, B,B,B,B,P, P,P,G,K,O ],
		[ A,M,C,w,B, B,B,B,B,B, B,B,B,B,P, P,P,P,K,O ],
		[ A,M,C,w,B, B,B,B,B,B, B,B,B,B,A, A,A,A,A,A ],
		[ w,w,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,w,w ],
		[ A,A,A,w,T, T,T,T,T,T, T,T,T,T,T, T,T,T,w,w ],
		[ A,A,A,w,T, T,T,T,T,T, T,T,T,T,T, T,T,T,w,w ] ]

#             w         B      R        G
_fcolor = [ "Black", "Blue", "Red", "#707070" ]
(w,B,R,G) = range(4)
_tcolor = [	[ w,w,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,w,w ],
		[ w,w,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,w,w ],
		[ w,B,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,w,B ],
		[ w,w,w,w,w, w,w,w,w,w, w,w,w,w,w, w,B,B,B,B ],
		[ w,w,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,B,B ],
		[ w,w,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,R,B ],
		[ w,w,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,w,B ],
		[ w,R,w,w,w, w,w,w,w,w, w,w,w,R,w, w,w,w,w,B ],
		[ w,R,w,w,w, G,G,G,G,G, G,G,G,G,G, G,G,G,G,G ],
		[ w,w,w,w,w, w,w,w,w,w, w,w,w,w,w, w,w,w,w,w ],
		[ w,w,w,w,w, w,w,w,G,w, w,w,w,w,w, w,w,w,w,w ],
		[ w,w,w,w,w, w,w,w,G,G, G,G,G,G,G, G,G,G,w,w ] ]

#===============================================================================
# Element Dialog
#===============================================================================
class ElementDialog:
	def __init__(self, master, element):
		if master is None:
			self.toplevel = tk.Toplevel(master)
		else:
			self.toplevel = master

		frame = self.toplevel
		self.element = element

		self.symbol = tk.Label(frame, font="Helvetica -72 bold",
				relief=tk.GROOVE, padx=30)
		self.symbol.grid(row=0, rowspan=5, column=0, columnspan=3, sticky=tk.NSEW)

		b = tk.Button(frame, text="\N{WHITE LEFT-POINTING TRIANGLE}", padx=2, pady=0,
				command=self.prevZ)
		b.grid(row=5, column=0, sticky=tk.W)
		self.name = tk.Label(frame, foreground="DarkBlue",
				width=13, relief=tk.GROOVE)
		self.name.grid(row=5, column=1, sticky=tk.EW)
		b = tk.Button(frame, text="\N{WHITE RIGHT-POINTING TRIANGLE}", padx=2, pady=0,
				command=self.nextZ)
		b.grid(row=5, column=2, sticky=tk.E)

		# ------
		row = 0
		col = 3
		l = tk.Label(frame, text="Z:", justify=tk.RIGHT, anchor=tk.E)
		l.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		l = tk.Label(frame, text="Atomic Weight:", justify=tk.RIGHT, anchor=tk.E)
		l.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		l = tk.Label(frame, text="Density:", justify=tk.RIGHT, anchor=tk.E)
		l.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		l = tk.Label(frame, text="Melting:", justify=tk.RIGHT, anchor=tk.E)
		l.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		l = tk.Label(frame, text="Boiling:", justify=tk.RIGHT, anchor=tk.E)
		l.grid(row=row, column=col, sticky=tk.EW)

		row += 1
		l = tk.Label(frame, text="Oxidation:", justify=tk.RIGHT, anchor=tk.E)
		l.grid(row=row, column=col, sticky=tk.EW)

		# ------
		row = 0
		col += 1
		self.Z = tk.Label(frame, justify=tk.LEFT,
			background="White", foreground="DarkBlue",
			anchor=tk.W)
		self.Z.grid(row=row, column=col, sticky=tk.EW, pady=1)

		row += 1
		self.Amass = tk.Label(frame, justify=tk.LEFT,
			background="White", foreground="DarkBlue",
			anchor=tk.W)
		self.Amass.grid(row=row, column=col, sticky=tk.EW, pady=1)

		row += 1
		self.density = tk.Label(frame, justify=tk.LEFT,
			background="White", foreground="DarkBlue",
			anchor=tk.W)
		self.density.grid(row=row, column=col, sticky=tk.EW, pady=1)

		row += 1
		self.melting = tk.Label(frame, justify=tk.LEFT,
			background="White", foreground="DarkBlue",
			anchor=tk.W)
		self.melting.grid(row=row, column=col, sticky=tk.EW, pady=1)

		row += 1
		self.boiling = tk.Label(frame, justify=tk.LEFT,
			background="White", foreground="DarkBlue",
			anchor=tk.W)
		self.boiling.grid(row=row, column=col, sticky=tk.EW, pady=1)

		row += 1
		self.oxidation = tk.Label(frame, justify=tk.LEFT,
			background="White", foreground="DarkBlue",
			anchor=tk.W)
		self.oxidation.grid(row=row, column=col, sticky=tk.EW, pady=1)

		# ------
		row += 1
		self.isotopeList = tkExtra.MultiListbox(frame,
				(("A",                   5, None),
				 ("Jπ",                  5, None),
				 ("Δ (MeV)",             8, None),
				 ("T1/2, Γ, Abundance", 15, None),
				 ("Decay Mode",          8, None)))
		self.isotopeList.grid(row=row, column=0, columnspan=5, sticky=tk.NSEW)
		self.isotopeList.bindLeftRight()

		row += 1
		self.footnote = tk.Label(frame, justify=tk.LEFT, anchor=tk.W,
				relief=tk.GROOVE, background="White")
		self.footnote.grid(row=row, column=0, columnspan=5, sticky=tk.NSEW)

		frame.grid_rowconfigure(row-1, weight=3)
		frame.grid_columnconfigure(3, weight=1)
		frame.grid_columnconfigure(4, weight=3)

		self.fill()

		if master is None:
			self.toplevel.bind("<Escape>", self.close)
			self.toplevel.bind("<Control-Key-q>", self.close)
			self.toplevel.protocol("WM_DELETE_WINDOW", self.close)
			#self.toplevel.transient(master)
			self.toplevel.focus_set()
			#self.toplevel.grab_set()
			self.toplevel.wait_window()

	# ----------------------------------------------------------------------
	def fill(self):
		element = self.element

		try:
			self.toplevel.title(f"Element: {element.Z}-{element.symbol} {element.name}")
		except:
			pass

		# Find color
		bcol = "White"
		for j in range(_PTBL_ROWS):
			try:
				i = _ptable[j].index(element.Z)
				fcol = _fcolor[_tcolor[j][i]]
				bcol = _bcolor[_ecolor[j][i]]
				break
			except:
				pass

		self.symbol.config(text=element.symbol,
				background=bcol, foreground=fcol)
		self.name.config(text=element.name)

		self.Z.config(text=element.Z)
		self.Amass.config(text=element.AmassStr)
		self.density.config(text=element.densityStr)
		self.melting.config(text=element.melting)
		self.boiling.config(text=element.boiling)
		self.oxidation.config(text=element.oxidation)

		# Fill isotope List
		self.isotopeList.delete(0,tk.END)
		for iso in element.isotope:
			jpi = element.isotope[iso].spin
			d   = element.isotope[iso].massex
			t12 = element.isotope[iso].decayt
			dec = element.isotope[iso].decayp
			self.isotopeList.insert(tk.END, (iso,jpi,d,t12,dec))

		# Show footnote if needed
		f = element.densityNote()
		if f:
			try:
				note = Element.footnote[f]
				self.footnote.config(text=f"{f}) {note}")
			except:
				pass

	# ----------------------------------------------------------------------
	def prevZ(self, event=None):
		if self.element.Z>1:
			self.element = Element.byZ(self.element.Z-1)
			self.fill()

	# ----------------------------------------------------------------------
	def nextZ(self, event=None):
		if self.element.Z<118:
			self.element = Element.byZ(self.element.Z+1)
			self.fill()

	# ----------------------------------------------------------------------
	# Close window
	# ----------------------------------------------------------------------
	def close(self, event=None):
		self.toplevel.destroy()

#===============================================================================
# Periodic Table class
#===============================================================================
class PeriodicTable(tk.Frame):
	def __init__(self, master, isotopeFrame=None):
		self.selected = None
		self.element  = None
		self.isotopeFrame = isotopeFrame

		super().__init__(master, class_="PeriodicTable", background="White")

		for j in range(_PTBL_ROWS):
			for i in range(_PTBL_COLS):
				val = _ptable[j][i]
				fcol = _fcolor[_tcolor[j][i]]
				bcol = _bcolor[_ecolor[j][i]]

				if val>0:
					elem = Element.elements[val-1]
					l = tk.Label(self,
						text=f"{val}\n{elem.symbol}",
						foreground=fcol,
						background=bcol,
						justify=tk.CENTER,
						borderwidth=2,
						width = 3,
						padx = 1,
						pady = 0,
						relief=tk.FLAT)
					l.bind('<Button-1>', self.click)
				elif val<0:
					l = tk.Label(self,
						pady=0,
						text=_ptableLbl[-val-1])
				else:
					continue
				l.grid(row=j, column=i, sticky=tk.NSEW, padx=1, pady=1)

		# ---
		row, col, span = _PTBL_ROWS, 2, 4
		tk.Label(self, background="White").grid(row=row, column=col)

		row += 1
		tk.Label(self, text="H - gas",
			foreground=_fcolor[1], background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		col += span
		tk.Label(self, text="Li - solid",
			foreground=_fcolor[0], background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		col += span
		tk.Label(self, text="Br - liquid",
			foreground=_fcolor[2], background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		col += span
		tk.Label(self, text="Tc - synthetic",
			foreground=_fcolor[3], background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		# ---
		row += 1
		col  = 1
		tk.Message(self, text=" ", background=_bcolor[5],
			relief=tk.GROOVE).grid(row=row, column=col, pady=3)
		col += 1
		span = 3
		tk.Label(self, text="Non-Metals", background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		col += span
		tk.Message(self, text=" ", background=_bcolor[2],
			relief=tk.GROOVE).grid(row=row, column=col)
		col += 1
		tk.Label(self, text="Transition Metals", background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		col += span
		tk.Message(self, text=" ", background=_bcolor[9],
			relief=tk.GROOVE).grid(row=row, column=col)
		col += 1
		tk.Label(self, text="Rare Earth Metals", background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		col += span
		tk.Message(self, text=" ", background=_bcolor[3],
			relief=tk.GROOVE).grid(row=row, column=col)
		col += 1
		tk.Label(self, text="Halogens", background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		# ---
		row += 1
		col  = 1
		tk.Message(self, text=" ", background=_bcolor[6],
			relief=tk.GROOVE).grid(row=row, column=col)
		col += 1
		span = 3
		tk.Label(self, text="Alkali Metals", background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		col += span
		tk.Message(self, text=" ", background=_bcolor[7],
			relief=tk.GROOVE).grid(row=row, column=col)
		col += 1
		tk.Label(self, text="Alkali Earth Metals", background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		col += span
		tk.Message(self, text=" ", background=_bcolor[8],
			relief=tk.GROOVE).grid(row=row, column=col)
		col += 1
		tk.Label(self, text="Other Metals", background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

		col += span
		tk.Message(self, text=" ", background=_bcolor[4],
			relief=tk.GROOVE).grid(row=row, column=col)
		col += 1
		tk.Label(self, text="Inert Elements", background="White").grid(
			row=row, column=col, columnspan=span, sticky=tk.W)

	# ----------------------------------------------------------------------
	def click(self, event):
		if self.selected:
			#self.selected.config(borderwidth=1)
			self.selected.config(relief=tk.FLAT)
		z = self.buttonZ(event.widget)
		self.selected = event.widget
		self.selected.config(relief=tk.RAISED) #borderwidth=3)
		dlg = ElementDialog(self.isotopeFrame, Element.byZ(z))
		self.element = dlg.element

	# ----------------------------------------------------------------------
	def buttonZ(self, widget):
		return int(widget["text"].split()[0])

#===============================================================================
#class PeriodicTableSelect(PeriodicTable):
#	def __init__(self, master):
#		PeriodicTable.__init__(self, master)
#
#	# ----------------------------------------------------------------------
#	def click(self, event):
#		self.selected = self.buttonZ(event.widget)
#		self.destroy()

#===============================================================================
# Periodic Elements List
#===============================================================================
class ElementsList(tk.Frame):
	def __init__(self, master, isotopeFrame=None):

		self.element = None
		self.isotopeFrame = isotopeFrame
		super().__init__(master, class_="ElementsList")

		self.listbox = tkExtra.MultiListbox(self,
				(("Z"        ,  3, None),
				 ("El"       ,  3, None),
				 ("Name"     , 12, None),
				 ("Weight"   , 14, None),
				 ("Density"  , 10, None),
				 ("Melting"  ,  8, None),
				 ("Boiling"  ,  8, None),
				 ("Oxidation",  8, None)),
				 background="White",
				 height=20)
		self.listbox.pack(expand=tk.YES, fill=tk.BOTH)
		self.listbox.bindLeftRight()

		for elem in Element.elements:
			if elem is None: continue
			self.listbox.insert(tk.END,
				(elem.Z, elem.symbol, elem.name, elem.AmassStr,
				elem.densityStr, elem.melting, elem.boiling, elem.oxidation))
		self.listbox.bind('<<ListboxSelect>>', self.show)

	# --------------------------------------------------------------------
	def show(self, event=None):
		sel = [ int(x) for x in self.listbox.curselection() ]
		if len(sel) == 1:
			isotope = self.listbox.get(sel[0])
			elem = Element.elements[isotope[0]-1]
			if elem is None: return
			dlg = ElementDialog(self.isotopeFrame, elem)
			self.element = dlg.element

#===============================================================================
# Periodic Table Frame class
#===============================================================================
class PeriodicTablePage(FlairRibbon.FlairPage):

	_name_ = "Elements"
	_icon_ = "toi"

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Input ===========
		group = Ribbon.LabelGroup(self.ribbon, "View")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		kw = {
			"compound" : tk.TOP,
			"variable" : self.tabVar,
			"command"  : self.change,
			"padx"     : 5
		}

		# ---
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Table",
				image=icons.get("toi32"),
				anchor=tk.CENTER,
				value=0,
				**kw)
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
		tkExtra.Balloon.set(b, "Show as periodic table")

		b = Ribbon.LabelRadiobutton(group.frame,
				text="List",
				image=icons.get("list32"),
				anchor=tk.CENTER,
				value=1,
				**kw)
		tkExtra.Balloon.set(b, "Show as list")
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)

	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage(False)

		self.frame.config(background="White")
		self.isotope = tk.Frame(self.frame)

		self.periodic_frame = tk.Frame(self.frame, background="White")

		self.periodic_table = PeriodicTable(self.periodic_frame, self.isotope)
		self.periodic_table.pack(expand=tk.YES, fill=tk.X)

		self.periodic_list = ElementsList(self.frame, self.isotope)

		self.tabVar = tk.IntVar()
		self.tabVar.set(0)

		self.change()
		self.isotope.pack(side=tk.RIGHT, expand=tk.YES, fill=tk.BOTH)

	#----------------------------------------------------------------------
	def change(self):
		if self.tabVar.get() == 0:
			self.periodic_frame.pack(side=tk.LEFT, anchor=tk.NW)
			self.periodic_list.pack_forget()
		else:
			self.periodic_frame.pack_forget()
			self.periodic_list.pack(side=tk.LEFT, anchor=tk.NW, fill=tk.Y)

#===============================================================================
# TOIFrame
#===============================================================================
class TOIFrame(tk.Frame):
	def __init__(self, master, isotopeFrame=None):
		super().__init__(master, class_="TOIFrame")

		self.tabPage = ttk.Notebook(self)
		self.tabPage.pack(expand=tk.YES, fill=tk.BOTH)

		self.periodic_table = PeriodicTable(self.tabPage, isotopeFrame)
		self.tabPage.add(self.periodic_table,
				text     = "Table")

		self.periodic_list = ElementsList(self.tabPage, isotopeFrame)
		self.tabPage.add(self.periodic_list,
				text     = "List")

	# ----------------------------------------------------------------------
	def getLast(self):
		try:
			actPage = self.tabPage.index(tk.CURRENT)
		except tk.TclError:
			return self.periodic_table.element

		if actPage == 0:
			return self.periodic_table.element
		else:
			return self.periodic_list.element

#===============================================================================
def show(master, swallow=False, select=False):
	if swallow:
		toplevel = master
	else:
		toplevel = tk.Toplevel(master)
		toplevel.bind("<Escape>", lambda e:toplevel.destroy())
		toplevel.bind("<Control-Key-q>", lambda e:toplevel.destroy())

	toplevel.title("Table of Elements")	#Periodic Table")
	toi = TOIFrame(toplevel)
	toi.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
	toplevel.deiconify()
	toplevel.wm_resizable(False,False)
	toplevel.focus_set()
	toplevel.lift()
	if select:
		toplevel.grab_set()
		toplevel.wait_window()
		return toi.getLast()
	return None

#===============================================================================
if __name__ == "__main__":
	import os
	import sys
	sys.path.insert(0,os.path.join(os.path.abspath(os.path.dirname(__file__)), "lib"))
	Element.init(os.path.join(os.path.dirname(sys.argv[0]),"db/isotopes.ini"))
	root = tk.Tk(className="flair_toi")
	root.deiconify()
	show(root,True)
	root.mainloop()
