# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   05-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import sys
import time
import math
import random
import subprocess

import tool
import pipe
import bmath

import Data
import convert
import Project
from deck    import Card
from moira   import Moira
from fluka   import Fluka, FlukaConfig, FlukaParticle

DEFAULT_FORMAT = ".eps"

GEO_NO    = "-No-"
GEO_AUTO  = "-Auto-"

PLOT_TYPES = [ "Geometry",
		"USR-1D",
		"USR-2D",
		"USRBIN",
		"USERDUMP",
		#"Output",	<- Thresholds, WW, Energy balance, etc...
		"RESNUCLE" ]

#AXES = ["x","y","z","cb","x2","y2"]

COORD     = ["U-V", "V-U", "X-Y", "X-Z", "Y-Z", "Y-X", "Z-X", "Z-Y" ]
COORDAUTO = ["Auto"] + COORD

STYLE = [
		"errorbars",		# X
		"errorlines",		# X
		"histerror",		# histogram with error
		"histogram",		# Xlow
		"lines",		# X
		"linespoints",		# X
		"points",
		"xyerrorbars"]
STYLE_DEF = 3

Y_VALUE = [	"Y",			# 0
		"<X>*Y",		# 1
		"DX*Y" ]		# 2

X_SCALE =  [	"",
		"1/eV",
		"1/keV",
		"1/MeV",
		"1/GeV",
		"1/J",
		"1/Gy",
		"1/um",
		"1/mm",
		"1/cm",
		"1/dm",
		"1/m",
		"1/km",
		]

LINE_TYPES = ["", "solid", "dashed"]
FONTS      = ["",
		"serif",		# Matplotlib families
		"sans-serif",
		"monospace",
		"cursive",
		"fantasy",

		"Arial",		# Gnuplot families
		"Courier",
		"Helvetica",
		"Symbol",
		"Times",
		]
SIZE       = ["","8","9","10","11","12","14","16","18","20","22","24","26","28","36","48","72"]

#                 0                   1            2           3            4           5
USRBIN_HIST = ['2D Projection', '1D Projection', '1D Max', '1D Min', '1D Trace-H', '1D Trace-V']

KIND = [	"Geometry",		# 0
		"Region",		# 1	-6
		"Material",		# 2	-5
		"Lattice",		# 3	-4
		"Field Intensity",	# 4	-3
		"Field Vector",		# 5	-2
		"Field" ]		# 6	-1

RESNUCLEI_Z   = "Z"
RESNUCLEI_A   = "A"
RESNUCLEI_Z_A = "Z/A"
RESNUCLEI_Z_N = "Z/N"
RESNUCLEI_A_Z = "A/Z"
RESNUCLEI_N_Z = "N/Z"

RESNUCLEI_TYPES  = [	RESNUCLEI_Z,
			RESNUCLEI_A,
			RESNUCLEI_Z_A,
			RESNUCLEI_Z_N,
			RESNUCLEI_A_Z,
			RESNUCLEI_N_Z ]
RESNUCLEI_TITLE  = ["Atomic Number",
		"Mass Number",
		"Mass vs Atomic Number",
		"Neutron vs Atomic Number",
		"Atomic vs Mass Number",
		"Atomic vs Neutron Number" ]
#RESNUCLEI_STABLE = ["No", "Lines", "Dots"]

SOURCE_HIST = [  'Particle',
		 'Total-Energy',
		 'Weight',
		 'x',
		 'y',
		 'z',
		 'tx',
		 'ty',
		 'tz',
		 'Kinetic-Energy',	# Are not in the file
		 'Momentum' ]

HISTOGRAM = "Histogram"
SCATTER   = "Scatter"

USERDUMP_CASE = ["Continuous", "Point", "Source"]

#===============================================================================
# Plot class
#===============================================================================
class Plot:
	def __init__(self, engine, project, plot, log=None):
		self.project = project
		self.engine  = engine
		self.plot    = plot
		self._log    = log

	# ----------------------------------------------------------------------
	def log(self, *args):
		s = " ".join(map(str,args))
		if self._log is None:
			print(s,end="")
		else:
			self._log.write(s)

	# ----------------------------------------------------------------------
	def error(self, *args):
		s = " ".join(map(str,args))
		if self._log is None:
			print(s, end="", file=sys.stderr)
		else:
			self._log.error(s)

	# ----------------------------------------------------------------------
	def write(self, s):
		self.log(s)

	# ----------------------------------------------------------------------
	def reset(self):
		self.engine.reset()

	# ----------------------------------------------------------------------
	# @return window title
	# ----------------------------------------------------------------------
	def windowtitle(self):
		return f"{self.project.name()}-{self.plot('index')}: {self.plot('title')}"

	# ----------------------------------------------------------------------
	# Show plot on screen
	# ----------------------------------------------------------------------
	def show(self):
		self.reset()
		self._startPlot()
		if self.plot.type == "Geometry":
			err = self._geometryPlot()
		elif self.plot.type == "USR-1D":
			err = self._usr1DPlot()
		elif self.plot.type == "USR-2D":
			err = self._usr2DPlot()
		elif self.plot.type == "USRBIN":
			err = self._usrbinPlot()
		elif self.plot.type == "RESNUCLE":
			err = self._resnucleiPlot()
		elif self.plot.type == "USERDUMP":
			err = self._userdumpPlot()
		else:
			err = None

		if err:
			if isinstance(err,tuple):
				self.error(err[0])
			else:
				self.error(err)
		# Keep plot information
		self.plot.time = time.time()
		return err

	# ----------------------------------------------------------------------
	# Save plot to file
	# ----------------------------------------------------------------------
	def save(self, ext):
		return self.plot.name+ext

	# ----------------------------------------------------------------------
	# Send plot on printer
	# ----------------------------------------------------------------------
	def hardcopy(self, printer):
		pass

	# ----------------------------------------------------------------------
	# Clean up files and variables
	# ----------------------------------------------------------------------
	def cleanup(self):
		if self.plot.type in ("Geometry", "USRBIN"):
			self.plot._plotgeom = {}
			if self.plot.type == "USRBIN":
				self.plot._inp = None

		elif self.plot.type == "USR-1D":
			for i in range(int(self.plot("ndetectors"))):
				filename = self.plot(f"file.{i}")
				if not filename: continue
				fileout = f"{filename}.data"
				if os.path.isfile(fileout):
					os.remove(fileout)

		elif self.plot.type == "USR-2D":
			pass

		elif self.plot.type == "RESNUCLE":
			self.plot._previous = None

		elif self.plot.type == "USERDUMP":
			self.plot._case    = -1
			self.plot._mgfile  = ""
			self.plot._limfile = ""

		for f in [f"{self.plot.name}.dat", f"{self.plot.name}.lim",
			  "gplevh.lim", "gplevh.npo", "gplevh.poi"]:
			try: os.remove(f)
			except OSError: pass

	# ----------------------------------------------------------------------
	# Perform plot
	# ----------------------------------------------------------------------
	def _plot(self, cmd):
		return None

	# ----------------------------------------------------------------------
	# Perform splot
	# ----------------------------------------------------------------------
	def _splot(self, cmd):
		return None

	# ======================================================================
	# Generate the plot
	# ----------------------------------------------------------------------
	def _geometryPlot(self):
		if self._getGeometryPlotgeom():
			return self._runProgram(self.plot.name)
		return None

	# ----------------------------------------------------------------------
	# Overridden method to return the plotting options for prefix
	# ----------------------------------------------------------------------
	def _plotOptions(self, prefix):
		pass

	# ----------------------------------------------------------------------
	# return colorband limits
	# ----------------------------------------------------------------------
	def cbLimits(self, vmin="*", vmax="*"):
		try:    cbmin = float(self.plot("cbmin"))
		except: cbmin = None

		try:    cbmax = float(self.plot("cbmax"))
		except: cbmax = None

		try:    cpd = float(self.plot("cbcpd"))
		except: cpd = 3

		try:    ncolors = float(self.plot("cbcolors"))
		except: ncolors = 30

		if int(self.plot("cblog")) and cpd>=1.0:
			if cbmin is None: cbmin = vmin
			if int(self.plot("cbround")):
				try:	# round limits that colors boundaries falls on decades
					cbmin = 10**(math.floor(math.log10(cbmin)*cpd)/cpd)
				except:
					pass
			if isinstance(cbmin,float) and not isinstance(cbmax,float) and ncolors>0:
				cbmax = cbmin * 10.0**(ncolors/cpd)

		if cbmin is None: cbmin = vmin
		if cbmax is None: cbmax = vmax
		return cbmin, cbmax

	# ----------------------------------------------------------------------
	# Initialize plot
	# ----------------------------------------------------------------------
	def _startPlot(self):
		self.engine.title(self.plot("title"), self._plotOptions("title"))
		self.engine.grid(self.plot("grid"))
		if self.plot("ratio"): self.engine.aspect(self.plot("ratio"))
		self._setupAxes()

	# ----------------------------------------------------------------------
	# Setup plotting axes
	# ----------------------------------------------------------------------
	def _setupAxes(self):
		self.engine.label('x', self.plot("xlabel"), self._plotOptions("x"))
		self.engine.tics( 'x', self._plotOptions("xtics"))

		self.engine.label('y', self.plot("ylabel"), self._plotOptions("y"))
		self.engine.tics( 'y', self._plotOptions("ytics"))

		#def isenable(p):
		#	try:
		#		return int(p)
		#	except:
		#		return False
		#if isenable(self.plot["ztics"]):
		if self.plot("ztics"):
			self.engine.label('z',  self.plot("zlabel"),  self._plotOptions("z"))
			self.engine.tics( 'z', self._plotOptions("ztics"))

		#if isenable(self.plot["cbtics"]):
		if self.plot("cbtics"):
			self.engine.label('cb', self.plot("cblabel"), self._plotOptions("cb"))
			self.engine.tics( 'cb', self._plotOptions("cbtics"))

		#if isenable(self.plot["x2tics"]):
		if self.plot("x2tics"):
			self.engine.label('x2', self.plot("x2label"), self._plotOptions("x2"))
			self.engine.tics( 'x2', self._plotOptions("x2tics"))
			self.engine("set x2tics")	# FIXME through engine

		#if isenable(self.plot["y2tics"]):
		if self.plot("y2tics"):
			self.engine.label('y2', self.plot("y2label"), self._plotOptions("y2"))
			self.engine.tics( 'y2', self._plotOptions("y2tics"))
			self.engine("set y2tics")	# FIXME through engine

		self._plotRange("x")
		self._plotRange("y")
		self._plotRange("z")
		self._plotRange("cb")
		self._plotRange("x2")
		self._plotRange("y2")

	# ----------------------------------------------------------------------
	# Set Plot range and log/linear
	# ----------------------------------------------------------------------
	def _plotRange(self, axis):
		self.engine.log(axis, int(self.plot(f"{axis}log")))
		amin = self.plot(f"{axis}min")
		amax = self.plot(f"{axis}max")
		self.engine.range(axis, amin, amax)

	# ----------------------------------------------------------------------
	def _backgroundPlot(self):
		pass

	# ----------------------------------------------------------------------
	# Generate the plot
	# ----------------------------------------------------------------------
	def _usr1DPlot(self):
		return None

	# ----------------------------------------------------------------------
	def _usr2DPlot(self):
		return None

	# ----------------------------------------------------------------------
	# Usrbin plot
	# ----------------------------------------------------------------------
	def _usrbinPlot(self):
		# Read detector
		usr = Data.Usrbin()
		self._det = None	# usrbin detector

		# Read file
		try:
			usr.readHeader(self.project.abspath(self.plot("datafile")))
		except (FileNotFoundError, IOError, MemoryError) as ex:
			return f"Error loading USRBIN file\n{ex}"

		# Using...
		d = int(self.plot("det"))
		try:
			self._det = usr.detector[d-1]
		except IndexError:
			return f"Detector {d} not found"

		# Start plotting
		regionPlot = self._det.type in (2, 8, 12, 18)
		try:
			h = USRBIN_HIST.index(self.plot("hist"))
		except ValueError:
			h = 0

		if h<=1:
			if regionPlot:
				# Region plot
				err = self._mergeGeoData(usr, d, self._det)
				if err is not None: return err

			else:	# Normal plot with GPLEVBIN
				err = self._runLevbin(usr,d)
				if err is not None: return err

			# --- 2D Histogram ---
			if h==0 and self.plot("geo")==GEO_AUTO:
				name = self.plot.name + ".geo"
				if self._getUsrbinPlotgeom(name, self._det):
					err = self._runProgram(name)
					if err is not None:
						self.plot._plotgeom = {}
						return err

		else:
			if regionPlot: return
			return self._runUsbmax(d)

	# ======================================================================
	# Run Fluka and prepare the output file
	# @return the error message to display
	# ----------------------------------------------------------------------
	def _runProgram(self, name):
		# Write the inputfile
		title = self.plot("title")
		if title == "": title = "-no title-"

		run = self.project.runs.getRunByName(self.plot("run"))
		if run is not None:
			mode = run("mode")
			inp = convert.input(self.project.input, mode, run.defines())
		else:
			inp = self.project.input
		inpname = inp.writePlotgeomInput(self.plot._plotgeom, self.project.dir(), title)

		# Execute application to create a geometry plot
		stime = time.time()
		cmd = self.project.plotgeomCmd(run, inpname)
		self.log(f"\nCmd: {' '.join(list(cmd))}\n")

		process = subprocess.Popen(cmd,
					stdout=subprocess.PIPE,
					stderr=subprocess.STDOUT)
		out,err = process.communicate()
		self.log(out.decode(errors="replace"))
		if err: self.log(err.decode(errors="replace"))
		pid = process.pid
		rc  = process.returncode
		if rc is None: rc=0
		self.log(f"... {self.project.mode()} run: {time.time()-stime:.3f} s\n")

		# Error?
		if rc != 0:
			return f"Return code {rc}",inpname,pid

		# Process output files
		# if everything went well we should have a file called
		if isinstance(inp, Fluka):
			ret = self._processFlukaPlotgeom(name, inpname, pid)
		elif isinstance(inp, Moira):
			ret = self._processMoiraPlotter(name, inpname, pid)
		else:
			ret = None

		# Clean up
		if Project.cleanup and ret is None:
			self.project.deleteFiles(inpname, pid)
		return ret

	#-----------------------------------------------------------------------
	# Process the fluka plotgeom file to be converted something that
	# plot is using
	#-----------------------------------------------------------------------
	def _processFlukaPlotgeom(self, name, inpname, pid):
		fn = f"{inpname}001_PLOTGEOM.STORE"
		try:
			fpg = open(fn, "r", errors="replace")
		except:
			return f"No {fn} file",inpname,pid

		try:
			fgeo = open(f"{name}.dat","w")
		except:
			return f"Opening file {self.plot.name}.dat for writing",inpname,pid

		# Work out the origin information from the what's
		O = self.plot._plotgeom.get("O", bmath.Vector.O)
		U = self.plot._plotgeom.get("U", bmath.Vector.Z)
		V = self.plot._plotgeom.get("V", bmath.Vector.X)
		extu = self.plot._plotgeom.get("extu", 50.0)
		extv = self.plot._plotgeom.get("extv", 50.0)

		# Skip first 7 lines
		for i in range(7):
			fgeo.write("#"+fpg.readline())

		nworm      = 0
		firstblock = 0	# Check that at least one worm have different size
		blocksize  = 0	# ... otherwise gnuplot thinks we are dealing with isolines

		self.log("Processing geometry plot file")
		stime = time.time()
		while True:
			line = fpg.readline()
			if line=="" or line.startswith(" ***MATERIAL***FIELD***"): break
			if line.startswith(" Cold storage") or line.startswith(" Worm no"):
				if nworm>0:
					# Write the 2 last points
					lx -= extu
					ly -= extv
					r = O + lx*U + ly*V
					fgeo.write("%.8g %.8g %.8g %.8g %.8g\n"
						% (lx,ly,r[0],r[1],r[2]))
					mx -= extu
					my -= extv
					r = O + mx*U + my*V
					fgeo.write("%.8g %.8g %.8g %.8g %.8g\n\n"
						% (mx,my,r[0],r[1],r[2]))
					blocksize += 2
					if firstblock == 0:
						firstblock = blocksize
					elif firstblock != blocksize:
						firstblock = -1
				nworm = 0
				blocksize = 0
				continue
			try:
				x, y = map(float, line.split())
			except:
				return "Processing geometry plot file",inpname,pid

			# Check if we have to forget mid point
			nworm += 1
			if nworm >= 3:
				dx1 = x  - lx
				dy1 = y  - ly
				dx2 = mx - lx
				dy2 = my - ly
				diff = (dx1*dx2 + dy1*dy2) - \
					(math.sqrt(dx1**2 + dy1**2) * math.sqrt(dx2**2 + dy2**2))
				if diff != 0.0:
					lx -= extu
					ly -= extv
					r = O + lx*U + ly*V
					fgeo.write("%.8g %.8g %.8g %.8g %.8g\n"
						% (lx,ly,r[0],r[1],r[2]))
					blocksize += 1
					lx, ly = mx, my
				else:
					nworm -= 1
				mx, my = x, y
			elif nworm == 2:
				mx, my = x, y
			else:
				lx, ly = x, y

		# Write the 2 last points
		if nworm >= 1:
			lx -= extu
			ly -= extv
			r = O + lx*U + ly*V
			fgeo.write("%.8g %.8g %.8g %.8g %.8g\n"
				% (lx,ly,r[0],r[1],r[2]))
			blocksize += 1
			if nworm >= 2:
				mx -= extu
				my -= extv
				r = O + mx*U + my*V
				fgeo.write("%.8g %.8g %.8g %.8g %.8g\n"
					% (mx,my,r[0],r[1],r[2]))
				blocksize += 1
			else:
				mx, my = lx, ly

		if firstblock == blocksize:
			if blocksize == 0:
				return "Geometry Problem or empty plot at the specified location.\n"\
					"Try to change the location or enlarge the plot.\n"\
					"For USRBIN change either the 'Use:' to '-No-' or the 'Pos:'", \
					inpname, pid
			else:
				fgeo.write("%.8g %.8g %.8g %.8g %.8g\n"
					% (mx,my,r[0],r[1],r[2]))
		self.log("... Worms:     %g s"%(time.time()-stime))
		stime = time.time()

		# Material section
		if line:
			fgeo.write("\n\n")

			# Split #Grid nx ny line
			grid = fpg.readline().split()
			nu = int(grid[1])
			nv = int(grid[2])

			# New 2010 release
			region2material = [0]*(len(self.project.input["REGION"])+20)

			# Histogram
			histogram = [[0]*nu for i in range(nv)]
			us = (2.0*extu)/nu
			vs = (2.0*extv)/nv
			v  = -extv
			#v  = -extv + vs/2.0	# with image
			for i in range(nv):
				u  = -extu
				#u  = -extu + us/2.0.0	# with image
				histrow = histogram[i]
				for j in range(nu):
					line = fpg.readline()
					try:
						reg = int(line[ 0: 5])
						mat = int(line[ 5:10])
						lat = int(line[10:15])
						bu  = float(line[15:27])
						bv  = float(line[27:39])
						b   = float(line[39:51])
					except ValueError:
						return "PLOTGEOM Incorrect region/field table "\
							"returned, maybe a problem with the input",\
							inpname, pid
					r   = O + u*U + v*V
					fgeo.write("%.8g %.8g %.8g %.8g %.8g %d %d %d %g %g %g\n"
						% (u,v,r[0],r[1],r[2],reg,mat,lat,bu,bv,b))
					histrow[j] = reg
					try:
						region2material[reg] = mat
					except IndexError:	# It should come here for VOXELS
						region2material.extend([0] * (reg-len(region2material)+20))
						region2material[reg] = mat

					u += us
				fgeo.write("\n")
				v += vs
			self.log("... Histogram: %g s"%(time.time()-stime))
			stime = time.time()

			line = fpg.readline()
			self.plot._materialList = []

			if line.startswith(" ***MATERIAL***NAMES***"):
				# New 2010 release
				# materials
				while True:
					line = fpg.readline()
					if not line:
						break	# XXX should never arrive here ###
					if line.startswith(" ***REGION***NAMES***"):
						break
					# should be in serial order!!!!
					try:
						self.plot._materialList.append(line.split()[1])
					except IndexError:
						self.plot._materialList.append("?")

				# regions
				regionList = []
				while True:
					line = fpg.readline()
					if not line:
						break
					# should be in serial order!!!!
					rid,regname = line.split()
					regionList.append(regname)
			else:
				assert False

			# Scan histograms for labels
			first = True
			for reg,x,y in Plot._scan4Labels(histogram):
				if first:
					fgeo.write("\n")
					first = False
				u = x*us-extu
				v = y*vs-extv
				r = O + u*U + v*V
				try:
					mat = str(self.plot._materialList[
							region2material[reg]-1]).replace("_",r"\\_")
					fgeo.write(f"{u} {v} {r[0]} {r[1]} {r[2]} "
						f"{regionList[reg-1]} {mat} {reg}\n")
				except IndexError:
					if reg >= len(regionList):
						sreg = str(reg)
					else:
						sreg = regionList[reg-1]
					mat = region2material[reg]
					if mat >= len(self.plot._materialList):
						smat = str(mat)
					else:
						smat = self.plot._materialList[mat-1]
					smat = smat.replace("_",r"\\_")
					fgeo.write(f"{u} {v} {r[0]} {r[1]} {r[2]} {sreg} {smat} {reg}\n")
			self.log(f"... Labels:    {time.time()-stime} s")

		# Close files and do the plot
		fgeo.close()
		fpg.close()

		return None

	# ----------------------------------------------------------------------
	# Process the moira plotgeom file to be converted something that
	# plot is using
	# ----------------------------------------------------------------------
	def _processMoiraPlotter(self, name, inpname, pid):
		fn   = f"{inpname}001_plotter.dat"
		fgeo = f"{name}.dat"

		try:	fpg = open(fn,"r")
		except: return f"No {fn} file",inpname,pid

		try:	fgeo = open(f"{name}.dat","w")
		except: return f"Opening file {self.plot.name}.dat for writing",inpname,pid

		# Work out the origin information from the what's
		O = self.plot._plotgeom.get("O", bmath.Vector.O)
		U = self.plot._plotgeom.get("U", bmath.Vector.Z)
		V = self.plot._plotgeom.get("V", bmath.Vector.X)
		extu = self.plot._plotgeom.get("extu", 50.0)
		extv = self.plot._plotgeom.get("extv", 50.0)

		# Copy everything until index 1
		for line in fpg:
			fgeo.write(line)
			if "# grid: " in line:
				words = line.split()
				nu = int(words[2])
				nv = int(words[4])
			elif "# Index: 1" in line:
				break

		# Copy next line
		fgeo.write(fpg.readline())

		# Material section
		stime = time.time()
		histogram = [[0]*nu for i in range(nv)]
		us = (2.0*extu)/nu
		vs = (2.0*extv)/nv
		v  = -extv
		#v  = -extv + vs/2.0	# with image
		region2material = {}
		region2name = {}
		self.plot._materialList = {}
		for i in range(nv):
			u  = -extu
			#u  = -extu + us/2.0.0	# with image
			histrow = histogram[i]
			for j in range(nu):
				line = fpg.readline()
				words = line.split()
				regid = int(words[0])
				matid = int(words[1])
				#lat   = int(words[2])
				#bu    = float(words[3])
				#bv    = float(words[4])
				#b     = float(words[5])
				reg   = words[6]
				mat   = words[7]
				r   = O + u*U + v*V
				fgeo.write(f"{u:12g} {v:12g} {r[0]:12g} {r[1]:12g} {r[2]:12g} ")
				fgeo.write(line)
				histrow[j] = regid
				region2material[regid] = mat
				region2name[regid] = reg
				self.plot._materialList[matid] = mat

				u += us
			fgeo.write("\n")
			v += vs
		self.log(f"... Histogram: {time.time()-stime} s")
		stime = time.time()

		# Scan histograms for labels
		first = True
		for regid,x,y in Plot._scan4Labels(histogram):
			if first:
				fgeo.write("\n")
				first = False
			u = x*us-extu
			v = y*vs-extv
			r = O + u*U + v*V
			mat = region2material[regid].replace("_",r"\\_")
			reg = region2name[regid].replace("_",r"\\_")
			fgeo.write(f"{u:12g} {v:12g} {r[0]:12g} {r[1]:12g} {r[2]:12g} {reg} {mat} {regid}\n")
		self.log(f"... Labels:    {time.time()-stime} s")

		# Close files and do the plot
		fgeo.close()
		fpg.close()
		return None

	# -----------------------------------------------------------------------------
	# perform a flood fill algorithm to determine regions
	# -----------------------------------------------------------------------------
	@staticmethod
	def _scan4Labels(hist):
		regions = []	# List of lists
				# Each row contains [reg/0, n/1, sumx/2, sumy/3]
		rows = len(hist)
		cols = len(hist[0])

		rnd = random.Random()
		rnd.seed(314159)
		pos = []
		pos_append = pos.append

		for row in range(rows):
			for col in range(cols):
				reg = hist[row][col]
				if reg < 0: continue

				# Create a new entry
				index = -len(regions)-1
				n, sx, sy  = 1, col, row
				stack = [(col,row)]
				stack_append = stack.append
				del pos[:]

				while stack:
					xo, y = stack.pop()
					# Scan to the left
					x = xo+1
					yup = y-1
					ydn = y+1
					while x>0:
						x -= 1
						if hist[y][x] != reg: break
						hist[y][x] = index
						n  += 1
						sx += x
						sy += y
						pos_append((x,y))

						# Check up/down
						if yup>=0 and hist[yup][x]==reg:
							# Push up value
							# XXX Check if neighbor is pushed!
							stack_append((x, yup))
						if ydn<rows and hist[ydn][x]==reg:
							# Push down value
							# XXX Check if neighbor is pushed!
							stack_append((x, ydn))

					# Scan to the right
					x = xo
					while True:
						x += 1
						if x==cols or hist[y][x] != reg: break
						hist[y][x] = index
						n  += 1
						sx += x
						sy += y
						pos_append((x,y))

						# Check up/down
						if yup>=0 and hist[yup][x]==reg:
							# Push up value
							# XXX Check if neighbor is pushed!
							stack_append((x, yup))
						if ydn<rows and hist[ydn][x]==reg:
							# Push down value
							# XXX Check if neighbor is pushed!
							stack_append((x, ydn))

				# Ignore small regions
				if n<5: continue
				ax = int(float(sx)/float(n))
				ay = int(float(sy)/float(n))

				# Check neighboring materials
				SCAN=5
				ok = True
				for y in range(max(0, ay-SCAN), min(ay+SCAN,rows)):
					for x in range(max(0, ax-SCAN), min(ax+SCAN,cols)):
						if hist[y][x] != index:
							ok = False
							break
					if not ok: break
				if ok:
					regions.append((reg, ax+0.5, ay+0.5))
				else:
					# Try few times to find a point not close
					# to anything else
					max_dist2 = 0
					max_x = -1
					max_y = -1
					if not regions:
						# Select randomly
						r = rnd.randint(0, len(pos)-1)
						max_x,max_y = pos[r]
					else:
						for _ in range(SCAN):
							# Select randomly
							r = rnd.randint(0, len(pos)-1)
							rx,ry = pos[r]
							min_dist2 = 10000000
							min_x = -1
							min_y = -1
							for rg,x,y in regions:
								d2 = (x-rx)**2 + (y-ry)**2
								if d2 <= min_dist2:
									min_dist2 = d2
									min_x = rx
									min_y = ry

							if min_dist2 > max_dist2:
								max_dist2 = min_dist2
								max_x = min_x
								max_y = min_y

					regions.append((reg, max_x+0.5, max_y+0.5))
		return regions

	# ----------------------------------------------------------------------
	# Get plotgeom info and return True is it has changed
	# ----------------------------------------------------------------------
	def _getGeometryPlotgeom(self):
		try:
			plotgeom = self.plot._plotgeom
		except AttributeError:
			self.plot._plotgeom = {}
			plotgeom = self.plot._plotgeom

		# Get variables
		O = bmath.Vector(self.plot("origin").split())
		U = bmath.Vector(self.plot("basisu").split())
		V = bmath.Vector(self.plot("basisv").split())
		extu,extv = map(float, self.plot("extends").split())
		#loP  = -extu*U -extv*V + O
		#hiP  =  extu*U +extv*V + O

		try:
			nu = float(self.plot("nu"))
		except ValueError:
			nu = 200
		try:
			nv = float(self.plot("nv"))
		except ValueError:
			nv = 200

		# Run fluka only if something has changed
		datfile = self.plot.name+".dat"
		try:
			# Check the output data file
			os.stat(datfile)
			changed = self.project.input.modifiedTime() > self.plot.time
		except (OSError, AttributeError):
			changed = True

		run = self.plot("run")

		if not changed: changed = plotgeom.get("boundaries",0) != 1-int(self.plot("boundaries"))
		if not changed: changed = abs(plotgeom.get("O",bmath.Vector.O)-O)>1e-6
		if not changed: changed = abs(plotgeom.get("U",bmath.Vector.Z)-U)>1e-6
		if not changed: changed = abs(plotgeom.get("V",bmath.Vector.X)-V)>1e-6
		if not changed: changed = abs(plotgeom.get("extu",50.0)-extu)>1e-6
		if not changed: changed = abs(plotgeom.get("extv",50.0)-extv)>1e-6
		if not changed: changed = plotgeom.get("nu",200) != nu
		if not changed: changed = plotgeom.get("nv",200) != nv
		if not changed: changed = tool.isnewer(self.project.input,
						self.plot.name+".dat")
		if not changed: changed = plotgeom.get("run","") != run
		if not changed: return False

		plotgeom["boundaries"] = 1-int(self.plot("boundaries"))	# Plot regions
		plotgeom["O"]    = O
		plotgeom["U"]    = U
		plotgeom["V"]    = V
		plotgeom["extu"] = extu
		plotgeom["extv"] = extv
		plotgeom["nu"]   = nu
		plotgeom["nv"]   = nv
		plotgeom["run"]  = run

		return True

	# ----------------------------------------------------------------------
	# Get plotgeom info and return True is it has changed
	# ----------------------------------------------------------------------
	def _getUsrbinPlotgeom(self, name, det):
		b10  = det.type % 10

		proj = ord(self.plot("proj")) - ord("X")
		pos  = self.plot("geopos")
		try:
			plotgeom = self.plot._plotgeom
		except AttributeError:
			self.plot._plotgeom = {}
			plotgeom = self.plot._plotgeom

		# Get variables
		if b10 in (1,7):
			# Cylindrical x=R, y=Phi, z=z
			# ymin=X, yhigh=Y
			# XXX bug in fluka does not keep the X,Y in the binning
			X = 0
			Y = 0
			if   proj==0: # X
				return False	# No geometry plot when Z-Phi
			elif proj==1: # Y
				# More tricky pos should be the rotation
				loP = bmath.Vector(X, Y, det.zlow)
				hiP = bmath.Vector(X+det.xhigh,  Y, det.zhigh)
				U = bmath.Vector.X
				V = bmath.Vector.Z
			elif proj==2: # Z
				if pos!="":
					try:
						z = float(pos)
					except ValueError:
						z = (det.zlow+det.zhigh)/2
				else:
					z = (det.zlow+det.zhigh)/2
				loP = bmath.Vector(-det.xhigh + X,
						   -det.xhigh + Y,
						    z)
				hiP = bmath.Vector( det.xhigh + X,
						    det.xhigh + Y,
						    z)
				U = bmath.Vector.X
				V = bmath.Vector.Y
		elif b10==2:
			return False
		elif b10==8:
			return False
		else:
			loP = bmath.Vector(det.xlow,  det.ylow,  det.zlow)
			hiP = bmath.Vector(det.xhigh, det.yhigh, det.zhigh)

			# Override limits if defined by user
			try: loP[0] = float(self.plot("bxmin"))
			except: pass
			try: loP[1] = float(self.plot("bymin"))
			except: pass
			try: loP[2] = float(self.plot("bzmin"))
			except: pass

			try: hiP[0] = float(self.plot("bxmax"))
			except: pass
			try: hiP[1] = float(self.plot("bymax"))
			except: pass
			try: hiP[2] = float(self.plot("bzmax"))
			except: pass

			# Add a small margin to avoid numerical precision
			# errors from the PLOTGEOM
			for i in range(3):
				loP[i] -= (abs(loP[i])+1.0)*1e-6
				hiP[i] += (abs(hiP[i])+1.0)*1e-6

			# Unless there are limits set...
			if pos != "":
				loP[proj] = hiP[proj] = float(pos)
			else:
				loP[proj] = hiP[proj] = (loP[proj]+hiP[proj])/2

			if   proj == 0:	# X
				U = bmath.Vector.Y
				V = bmath.Vector.Z
			elif proj == 1: # Y
				U = bmath.Vector.X
				V = bmath.Vector.Z
			elif proj == 2: # Z
				U = bmath.Vector.X
				V = bmath.Vector.Y

		# calculate origin and extends
		O = 0.5*(loP+hiP)
		D = hiP - loP
		extu = (D * U) / 2 * 1.00001
		extv = (D * V) / 2 * 1.00001

		try:
			changed = self.project.input.modifiedTime() > self.plot.time
		except AttributeError:
			changed = True
		if not changed: changed = abs(plotgeom.get("O",bmath.Vector.O)-O)>1e-6
		if not changed: changed = abs(plotgeom.get("U",bmath.Vector.Z)-U)>1e-6
		if not changed: changed = abs(plotgeom.get("V",bmath.Vector.X)-V)>1e-6
		if not changed: changed = abs(plotgeom.get("extu",50.0)-extu)>1e-6
		if not changed: changed = abs(plotgeom.get("extv",50.0)-extv)>1e-6
		if not changed: changed = tool.isnewer(self.project.input, name+".dat")
		if not changed: return False

		plotgeom["boundaries"] = 1	# Plot materials
		plotgeom["O"] = 0.5*(loP + hiP)
		plotgeom["U"] = U
		plotgeom["V"] = V
		plotgeom["extu"] = extu
		plotgeom["extv"] = extv
		return True

	# ----------------------------------------------------------------------
	# Run the gplevbin program to process the data
	# ----------------------------------------------------------------------
	def _runLevbin(self, usr, d):
		if d<=0 or d>len(usr.detector):
			return "Requested detector out of range"
		det = usr.detector[d-1]

		# Find limits and bins
		# --- X ---
		xmin = self.plot("bxmin")
		xmax = self.plot("bxmax")
		if xmin=="" and xmax=="":
			limx = ""
			xmin = det.xlow
			xmax = det.xhigh
		else:
			try: xmin = float(xmin)
			except ValueError: xmin = det.xlow
			try: xmax = float(xmax)
			except ValueError: xmax = det.xhigh
			if xmin<det.xlow:  xmin = det.xlow
			if xmax>det.xhigh: xmax = det.xhigh
			limx = "%g %g"%(xmin,xmax)
		try:
			xrebin = int(self.plot("xrebin"))
			if xrebin<=0: xrebin=1
		except:
			xrebin = 1
		dx = (det.xhigh - det.xlow) / float(det.nx) * float(xrebin)
		if abs(dx)<1e-19:
			self.log("Too small bin in x",dx)
			return False
		nx = str(int(round((xmax-xmin) / dx)))

		# --- Y ---
		ymin = self.plot("bymin")
		ymax = self.plot("bymax")
		if ymin=="" and ymax=="":
			limy = ""
			ymin = det.ylow
			ymax = det.yhigh
		else:
			try: ymin = float(ymin)
			except ValueError: ymin = det.ylow
			try: ymax = float(ymax)
			except ValueError: ymax = det.yhigh
			if ymin<det.ylow:  ymin = det.ylow
			if ymax>det.yhigh: ymax = det.yhigh
			limy = "%g %g"%(ymin,ymax)
		try:
			yrebin = int(self.plot("yrebin"))
			if yrebin<=0: yrebin=1
		except:
			yrebin = 1
		dy = (det.yhigh - det.ylow) / float(det.ny) * float(yrebin)
		if abs(dy)<1e-19:
			self.log("Too small bin in y",dy)
			return False
		ny = str(int(round((ymax-ymin) / dy)))

		# --- Z ---
		zmin = self.plot("bzmin")
		zmax = self.plot("bzmax")
		if zmin=="" and zmax=="":
			limz = ""
			zmin = det.zlow
			zmax = det.zhigh
		else:
			try: zmin = float(zmin)
			except ValueError: zmin = det.zlow
			try: zmax = float(zmax)
			except ValueError: zmax = det.zhigh
			if zmin<det.zlow:  zmin = det.zlow
			if zmax>det.zhigh: zmax = det.zhigh
			limz = "%g %g"%(zmin,zmax)
		try:
			zrebin = int(self.plot("zrebin"))
			if zrebin<=0: zrebin=1
		except:
			zrebin = 1
		dz = (det.zhigh - det.zlow) / float(det.nz) * float(zrebin)
		if abs(dz)<1e-19:
			self.log("Too small bin in z",dz)
			return False
		nz = str(int(round((zmax-zmin) / dz)))

		p = self.plot("proj")
		try:
			h = USRBIN_HIST.index(self.plot("hist"))
		except ValueError:
			h = 0
		if h == 0:
			if   p == "X": nx = "1"
			elif p == "Y": ny = "1"
			else:          nz = "1"
		elif h == 1:
			if   p == "X": ny,nz = "1","1"
			elif p == "Y": nx,nz = "1","1"
			else:          nx,ny = "1","1"
		else:
			# Special type... not handled now
			nz = "1"

		datafile = self.project.abspath(self.plot("datafile"))
		inp = [ "-",		# PLOTGEOM
			"",		# Swap plotgeom
			datafile,	# Bin file
			"",		# Normalization
			"",		# Threshold density
			d,		# Binning
			limx, nx,	# Limits and bins
			limy, ny,
			limz, nz,
			"" ]		# End

		# Compare self.inp and inp
		try:
			if self.plot._inp == inp: return None
		except AttributeError:
			pass

		cmd = FlukaConfig.command("gplevbin")
		self.log(f"\nCmd: {cmd}\n")
		for line in inp:
			self.log(f"<<< {line}")
		rc, out = pipe.system(cmd, inp)
		for line in out: self.log(line)
		if rc: self.log(f"\nReturn code {rc}")

		try:
			# To be propagated into the viewport
			minmax = out[-2].split()
			self.plot["min"] = minmax[0]
			self.plot["max"] = minmax[1]
			n = -3
			while True:
				sp = out[n].split()
				if sp[0] == "Binning":
					self.plot["int"] = sp[-1]
					break
				n -= 1
		except:
			self.plot["min"] = ""
			self.plot["max"] = ""
			self.plot["int"] = ""

		try:
			os.stat("gplevh.dat")
		except:
			return "File gplevh.dat is not created."

		if not tool.isnewer("gplevh.dat",datafile):
			return f"gplevh.dat is older than {self.plot('datafile')}"

		fndat = self.project.abspath(self.plot.name+".dat")
		try:
			os.stat(fndat)
			try:
				os.remove(fndat)
			except OSError:
				return f"Cannot remove old {fndat} file"
		except OSError:
			pass

		try:
			os.rename("gplevh.dat", fndat)
		except:
			return f"Cannot rename gplevh.dat to {fndat}"

		if Project.cleanup:
			for fn in ("gplevh.lim", "gplevh.npo", "gplevh.poi", "doslev.dat",
				  "fort.11", "fort.15"):
				try: os.remove(fn)
				except: pass

		self.plot._inp = inp

	# ----------------------------------------------------------------------
	# Run the usbmax program to process the data
	# ----------------------------------------------------------------------
	def _runUsbmax(self, det):
		# Find limits
		xmin = self.plot("bxmin")
		xmax = self.plot("bxmax")
		ymin = self.plot("bymin")
		ymax = self.plot("bymax")
		zmin = self.plot("bzmin")
		zmax = self.plot("bzmax")
		axis = self.plot("proj")

		cmd = [ FlukaConfig.command("usbmax"),
				"-d", str(det),
				"-a", axis,
				"-o", f"{self.plot.name}.dat"]
		if self.plot("hist") == "1D Min":
			cmd.append("-m")
		cmd.append(self.plot("datafile")) # Bin file

		if xmin!="": cmd.extend(["-x", str(xmin)])
		if xmax!="": cmd.extend(["-X", str(xmax)])
		if ymin!="": cmd.extend(["-y", str(ymin)])
		if ymax!="": cmd.extend(["-Y", str(ymax)])
		if zmin!="": cmd.extend(["-z", str(zmin)])
		if zmax!="": cmd.extend(["-Z", str(zmax)])

		try:
			if self.plot._inp == cmd: return None
		except AttributeError:
			pass

		self.log(f"\nCmd: {' '.join(cmd)}\n")

		try:
			#out = subprocess.check_output(cmd)
			process = subprocess.Popen(cmd,
						stdout=subprocess.PIPE,
						stderr=subprocess.STDOUT)
			out,err = process.communicate()
			self.log(out.decode(errors="replace"))
			if err: self.log(err.decode(errors="replace"))
			if process.returncode != 0: return out
		except subprocess.CalledProcessError as ex:
			self.log(str(ex))
			return str(ex)
		self.plot._inp = cmd

	# ----------------------------------------------------------------------
	def _mergeGeoData(self, usr, d, det):
		datfile = self.plot("datafile")
		outfile = f"{self.plot.name}.dat"
		geofile = f"{self.plot('geo')}.dat"

		# check if conditions have changed
		try:
			plotgeom = self.plot._plotgeom
		except AttributeError:
			self.plot._plotgeom = {}
			plotgeom = self.plot._plotgeom

		changed = datfile != plotgeom.get("dat")
		if not changed: changed = geofile != plotgeom.get("geo")
		if not changed: changed = tool.isnewer(self.plot("datafile"), outfile)
		if not changed: changed = tool.isnewer(geofile, outfile)
		if not changed: return None
		plotgeom["dat"] = datfile
		plotgeom["geo"] = geofile

		# If no geometry is provided then make histogram plot
		# of the data "Region" / "Value"
		# if a geometry is provided then superimpose the data on
		# the geometry
		try:
			data = usr.readData(d-1)
			stat = usr.readStat(d-1)
		except (FileNotFoundError, IOError) as ex:
			return f"Error reading data {d}\n{ex}"

		fdata  = Data.unpackArray(data)
		edata  = Data.unpackArray(stat)
		del data
		del stat

		# Find regions
		cardlist = self.project.input["REGION"]
		Card.sort(cardlist)
		regionList = [x.name() for x in cardlist
				if x.notIgnore() and x.name()!="&"]

		# open output file
		try:
			f = open(outfile,"w")
		except (FileNotFoundError, IOError) as ex:
			return f"Error opening output\n{ex}"

		# Dump energy/region
		rstart  = int(det.xlow)-1	# zero based
		#rend    = int(det.xhigh)-1
		rstep   = int(det.dx)
		rlen    = int(det.nx)

		lstart  = int(det.ylow)
		lstep   = int(det.dy)
		llen    = int(det.ny)

		special = det.type in (8,18)
		reglen  = len(regionList)
		if special:
			# Special binning
			# accept only if the last field has 1 bin!
			if det.nz != 1:
				return "Error cannot process special binning " \
					"with more than one binning in the third variable"

			idx = 0
			for j in range(llen):
				l  = j*lstep + lstart
				#jl = j*rlen
				for i in range(rlen):
					r = int(i*rstep) + rstart
					if r<reglen:
						f.write("%d %d %s:%d %g %g\n" %
							(r+1, l, regionList[r],l,
							 fdata[idx], edata[idx]))
					idx += 1

		else:	# Region binning
			for i in range(rlen):
				r = int(i*rstep) + rstart
				if r<reglen:
					f.write("%d 0 %s:0 %g %g\n" %
						(r+1, regionList[r],
						 fdata[i], edata[i]))

		# If custom geometry plot is available, get the second part
		# and add the energy/region
		if self.plot("geo") not in (GEO_NO, GEO_AUTO):
			try:
				fgeo = open(geofile,"r")

				# Make a new separation
				f.write("\n\n")

				# Scan fgeo for the scan part (double empty line)
				count = 0
				for line in fgeo:
					if line=="\n":
						count += 1
						if count == 2: break
					else:
						count = 0

				# Read remaining lines and append energy deposition if any
				count = 0
				for line in fgeo:
					if line=="\n":
						f.write(line)
						count += 1
						if count == 2: break
						continue
					count = 0
					words = line.split()

					# Check if it is a round integer
					r,m = divmod(int(words[5])-1-rstart, rstep)
					l = int(words[7])	# before we substitute
					# set the default 0.0
					words[6] = '0.0'
					words[7] = '0.0'
					if m==0 and 0 <= r < rlen:
						if special:
							l,m = divmod(l-lstart, lstep)
							if m==0 and 0 <= l < llen:
								idx = l*rlen+r
								words[6] = str(fdata[idx])
								words[7] = str(edata[idx])
						else:
							words[6] = str(fdata[r])
							words[7] = str(edata[r])
					f.write("%s\n"%(" ".join(words[:8])))

				# Copy the remaining lines
				for line in fgeo: f.write(line)

				fgeo.close()
			except (FileNotFoundError, IOError) as ex:
				return str(ex)

			f.close()

		del fdata
		del edata

	# ----------------------------------------------------------------------
	# Prepare the data for the resnuclei plot
	# ----------------------------------------------------------------------
	def _resnucleiPlot(self):
		# Check if we have to re-write the file
		d = int(self.plot("det"))-1
		if d<0: return "RESNUCLEi invalid detector specified"

		usr = Data.Resnuclei()
		try:
			usr.readHeader(self.plot("datafile"))
		except (FileNotFoundError, IOError) as ex:
			return f"reading resnuclei file\n{ex}"

		if not usr.detector: return "empty RESNUCLEi file"
		current = [self.plot.name, self.plot("datafile"), d]
		try:
			if current == self.plot._previous:
				return None
		except AttributeError:
			pass

		det = usr.detector[d]
		try:
			data = usr.readData(d)
			stat = usr.readStat(d)
		except (FileNotFoundError, IOError):
			return f"Error reading data file {d}"

		fdata = Data.unpackArray(data)
		del data

		if stat:
			#total = Data.unpackArray(stat[0])
			A     = Data.unpackArray(stat[1])
			eA    = Data.unpackArray(stat[2])
			Z     = Data.unpackArray(stat[3])
			eZ    = Data.unpackArray(stat[4])
			edata = Data.unpackArray(stat[5])
			del stat
			stat  = True
		else:
			stat  = False

		# Fill min/max values
		try:
			fmin = min([x for x in fdata if x>0.0])
		except:
			return "Empty data set"

		#fmax = max(fdata)
		#self.usr_min.config(text="%-10g"%(fmin))
		#self.usr_max.config(text="%-10g"%(fmax))

		# open output file
		try:
			f = open(f"{self.plot.name}.dat","w")
		except (FileNotFoundError, IOError) as ex:
			return f"Error opening output\n{ex}"

		zh  = det.zhigh
		mh  = det.mhigh
		nmz = det.nmzmin

		length = len(fdata)

		# Z distribution
		f.write("# Detector n: 1 Z\n")
		if stat:
			for z in range(zh):
				f.write("%d %d %15g %g\n"%(z+1,z+2,Z[z],eZ[z]*100.0))
		else:
			for z in range(zh):
				s  = 0.0
				for m in range(det.mhigh):
					pos = z + m * zh
					if 0 <= pos < length:
						s += fdata[pos]
				f.write("%d %d %15g 0.\n"%(z+1,z+2,s))

		# A distribution
		f.write("\n\n# Detector n: 2 A\n")
		amax = 2*zh + mh + nmz
		if stat:
			for a in range(1,min(amax,len(A))+1):
				f.write("%d %d %15g %g\n"%(a,a+1, A[a-1],eA[a-1]))
		else:
			for a in range(1,amax+1):
				s = 0.0
				for z in range(zh):
					m = a - 2*z - nmz - 3
					pos = z + m * zh
					if 0 <= pos < length:
						s += fdata[pos]
				f.write("%d %d %15g 0.\n"%(a,a+1, s))

		# Z,M distribution
		#f.write("\n\n#Isotope Yield as a function of Atomic " \
		#	"Number / N-Z-NMZmin (nuclei / cmc / pr)\n")
		#pos = 0
		#for m in range(mh):
		#	for z in range(zh):
		#		f.write("%d %d %15g\n"%(z+1,m+1,fdata[pos]))
		#		pos += 1
		#	f.write("\n")

		# Z,A distribution
		f.write("\n\n# Detector n: 3 Z/A\n")
		for a in range(1,amax+1):
			f.write("# Block n: 3 A:%d\n"%(a))
			for z in range(zh):
				m = a - 2*z - nmz - 3
				if m<0 or m>=mh:
					f.write("%d %d 0. 0.\n"%(z+1,a))
				else:
					pos = z + m * zh
					if stat:
						err = edata[pos]*100
					else:
						err = 0.0
					f.write("%d %d %15g %g\n"%(z+1,a,
						fdata[pos], err))
			f.write("\n")

		f.close()
		del fdata
		if stat: del edata
		self.plot._previous = current

	# ----------------------------------------------------------------------
	# Userdump file creation
	# ----------------------------------------------------------------------
	def _userdumpPlot(self):
		try:
			case = USERDUMP_CASE.index(self.plot("case"))
		except:
			case = 0

		mgfile = self.plot.name + ".dat"
		#if self.case == case and \
		#   mgfile == self.mgfile and \
		#   tool.isnewer(self.mgfile, self.plot("datafile")):
		#	return True
		try:
			fout = open(os.path.join(self.project.dir(),mgfile),"w")
		except (FileNotFoundError, IOError):
			return f"cannot open file {mgfile} for writing"

		mg = Data.Mgdraw(self.plot("datafile"))
		if mg.hnd is None:
			return f"opening userdump file {self.plot('datafile')}"

		try: from_ = int(self.plot("from"))
		except: from_ = 0
		try: n = int(self.plot("n"))
		except: n = 0
		try: emin = float(self.plot("emin"))
		except: emin = None
		try: emax = float(self.plot("emax"))
		except: emax = None
		particles = []
		for p in self.plot("particles").split():
			particles.append(FlukaParticle._db[p].id)
		if not particles:
			particles = None

		if n==0:
			to_ = from_ + 10000
		else:
			to_ = from_ + n

		if case == 2:
			# Find limits
			dmin = [ 1e30]*11
			dmax = [-1e30]*11

		i = 0
		startTime = time.time()
		#last = startTime
		try:
			while True:
				try:
					t = mg.readEvent(case)
				except:
					return "Not a valid MGREAD file"

				if t is None: break
				if t != case: continue

				if case == 0:		# Tracking
					n      = mg.ntrack
					ptype  = mg.jtrack
					weight = mg.wtrack
					try:
						mass = FlukaParticle._db[int(ptype)].mass
					except:
						# FIXME XXX XXX
						# for ptype < -6!!!!!
						# or ptype == -2
						# only kinetic energy is possible
						mass = 0.0

					ekin   = mg.etrack - mass
					if emin and ekin < emin: continue
					if emax and ekin > emax: continue
					if particles and ptype not in particles:
						continue

					i += 1
					if i < from_: continue
					if i > to_: break

					j = 0
					while j < n:
						x = mg.data[j];	j += 1
						y = mg.data[j];	j += 1
						z = mg.data[j];	j += 1
						fout.write("%d %g %g %g %g %g\n" %
							(ptype, ekin, weight, x, y, z))
					fout.write("\n")

				elif case == 2:		# Source particles
					(ptype, etot, weight, x, y, z, tx, ty, tz) \
							= mg.data
					try:
						mass = FlukaParticle._db[int(ptype)].mass
					except KeyError:
						return f"Unknown particle id={int(ptype)}"
					ekin = max(etot - mass, 0.0)
					if emin and ekin < emin: continue
					if emax and ekin > emax: continue
					if particles and ptype not in particles:
						continue

					i += 1
					if i < from_: continue
					if i > to_:   break

					pmom = math.sqrt(ekin*(etot+mass))
					fout.write("%d %g %g %g %g %g %g %g %g %g %g\n" %
						(ptype, etot, weight,
							x, y, z, tx, ty, tz, ekin, pmom))

					for d in range(9):
						dmin[d] = min(dmin[d], mg.data[d])
						dmax[d] = max(dmax[d], mg.data[d])
					dmin[ 9] = min(dmin[ 9], ekin)
					dmin[10] = min(dmin[10], pmom)
					dmax[ 9] = max(dmin[ 9], ekin)
					dmax[10] = max(dmin[10], pmom)

				#if time.time() - last > 2.0:
				#	last = time.time()
					#self.processed.config(text="%d"%(i))
					#self.processed.update_idletasks()

		except KeyboardInterrupt:
			self.log(f"*** Interrupted after {time.time()-startTime} s ***")
		else:
			self.log(f"Processing finished in {time.time()-startTime} s")
		#self.processed.config(text="%d"%(i))
		mg.close()
		fout.close()

		if case == 2:
			#self.limfile = "%s-%d.lim" % (self.plot.name, case)
			self.limfile = f"{self.plot.name}.lim"
			try:
				fout = open(os.path.join(self.project.dir(), self.limfile),"w")
			except (FileNotFoundError, IOError):
				return f"cannot open file {self.limfile} for writing"
			for dmi,dma in zip(dmin, dmax):
				fout.write("%g %g\n"%(dmi, dma))
			fout.close()
			self.dmin = dmin
			self.dmax = dmax
		else:
			self.limfile   = ""
			self.dmin = None
			self.dmax = None
