# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   05-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import io
import os
import pickle
import tkinter as tk
from tkinter import ttk

import undo
#import Global
import Ribbon
import Project
import FlairRibbon
from Project import GroupInfo, PlotInfo

import icons
import tkFlair
import tkExtra
import tkDialogs
import bFileDialog

import Plot
import GPPlot
import MPPlot
import UsrPlot
import BasePlot
import Usr2Plot
import UsrbinPlot
import GeometryPlot
import UserdumpPlot
import ResnucleiPlot

#from flair import PLOT_ENGINES
PLOT_ENGINES = ["Gnuplot", "Matplotlib"]

#===============================================================================
# Action Group
#===============================================================================
class _ActionGroup(FlairRibbon.FlairMenuGroup):
	#----------------------------------------------------------------------
	def createMenu(self):
		menu = tk.Menu(self, tearoff=0)

		for n in PLOT_ENGINES:
			menu.add_radiobutton(
					label=n,
					image=icons.get("plot"), compound=tk.LEFT,
					variable=self.page._plotEngine,
					value=n)
		return menu

#===============================================================================
# Override Tree view with selection and dnd mechanism
#===============================================================================
class PlotTreeview(tkFlair.InfoTreeview):
	#----------------------------------------------------------------------
	def __init__(self, master, *kw, **args):
		super().__init__(master, *kw, **args)
		self.bindSearch()
		#self.bindVi()

	# ----------------------------------------------------------------------
	# Insert a plot or issue a standard Treeview.insert() call
	# ----------------------------------------------------------------------
	def insert(self, parent, index, plot=None, **kw):
		#print(f">>> TREE INSERT {parent=} {index=} {name=} {type(plot)}")
		if isinstance(plot, GroupInfo):
			if plot.name:
				item = super().insert(parent, index,
							f"{parent}:{plot.name} [{id(plot)}]",
							text=plot.name,
							image=icons.get("group"),
							**kw)
							#open=
				self.item2info[item] = plot
			else:
				item = parent
			open_ = None
			for p in plot.childs:
				if open_ is None and isinstance(p,PlotInfo):
					open_ = p("open")
				self.insert(item, tk.END, p)
			if open_ is not None:
				self.item(item, open=open_)
		elif isinstance(plot, PlotInfo):
			item = super().insert(parent, index,
						f"{parent}:{plot.name} [{id(plot)}]",
						text=plot.name,
						image=icons.get(plot.type.upper()),
						**kw)
			self.item2info[item] = plot
		else:
			item = super().insert(parent, index, **kw)
			self.item2info[item] = plot
		return item

	# ----------------------------------------------------------------------
	# Delete items and update dictionaries
	# ----------------------------------------------------------------------
	def delete(self, *items):
		super().delete(*items)
		for item in items:
			# FIXME if item has children we need to remove them as well
			try:
				del self.item2info[item]
			except KeyError:
				pass

	# ----------------------------------------------------------------------
	def editInPlace(self, event=None):
		self.event_generate("<<Rename>>")

	# ----------------------------------------------------------------------
	# @return the path of a item
	# ----------------------------------------------------------------------
	def path(self, item):
		return "/".join(super().path(item))

	# ----------------------------------------------------------------------
	def selectedPlots(self):
		return [self.item2info[x] for x in self.selection()
				if isinstance(self.item2info[x], PlotInfo)]

#===============================================================================
# Plot List Frame
#===============================================================================
class PlotListPage(FlairRibbon.FlairPage):
	"""Create plots of geometry or from the detectors defined in the input."""
	_name_ = "Plot"
	_icon_ = "plot"

	#----------------------------------------------------------------------
	def init(self):
		self._plotEngine = tk.StringVar()
		self._plotEngine.set(PLOT_ENGINES[0])
		self.undoredo  = undo.UndoRedo(100)
		self.refreshOn  = ("plot",)
		self.selected  = None		# Selected plots to refresh when changed
		self.displayed = None		# Displayed plot type

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Plot List ===========
		group = Ribbon.LabelGroup(self.ribbon, "Plot List")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col,row = 0,0
		menulist = [	("group",    "group",     self.addGroup),
				("Geometry", "GEOMETRY",  self.addGeometry),
				("USR-1D",   "USR-1D",    self.addUsr1D),
				("USR-2D",   "USR-2D",    self.addUsr2D),
				("USRBIN",   "USRBIN",    self.addUsrbin),
				("USERDUMP", "USERDUMP",  self.addUserdump),
				("RESNUCLE", "RESNUCLE",  self.addResnuclei)]

		b = Ribbon.MenuButton(group.frame, menulist,
				image=icons.get("add32"),
				text="Add \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				compound=tk.TOP,
				command=self.add)
		b.grid(row=row, rowspan=3, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add new plot [Ins]")

		# ---
		col += 1
		b = Ribbon.LabelButton(group.frame,
				text="Oz",
				image=icons.get("wizard32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.wizard)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Automatically create plots")

		# ---
		col += 1
		row  = 0
		b = Ribbon.LabelButton(group.frame,
				text="Delete",
				image=icons.get("x"),
				command=self.delete)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Delete selected plots [Del]")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text="Clone",
				image=icons.get("clone"),
				command=self.clone)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Clone selected plots [Ctrl-D]")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame, self.page, "<<Refresh>>",
				text="Refresh",
				image=icons.get("refresh"))
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Refresh plots [Ctrl-R]")

		# ========== Move ===========
		group = Ribbon.LabelGroup(self.ribbon, "Move")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col  = 0
		row  = 1
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("left"),
				command=self.moveLeft)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Move selected plots up-one parent [Ctrl-Left]")

		# ---
		col += 1
		row  = 0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("up"),
				command=self.moveUp)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Move selected plots up [Ctrl-Up]")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame,
				#text="Rename",
				image=icons.get("rename"),
				command=self.rename)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Rename plot [F2]")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("down"),
				command=self.moveDown)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Move selected plots down [Ctrl-Down]")

		# ---
		col += 1
		row  = 1
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("right"),
				command=self.moveRight)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Move selected plots down-one child [Ctrl-Right]")

		# ========== Plot ===========
		group = _ActionGroup(self.ribbon, self, "Action")
		#group = Ribbon.LabelGroup(self.ribbon, "Action")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()
		group.frame.grid_columnconfigure(2, weight=1)

		# ---
		col,row = 0,0
		menulist = []
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("save"),
				command=self.save)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Save plot")

		col,row = 1,0
		menulist = []
		for fmt in sorted(tkFlair._PLOT_FORMAT):
			cmd = lambda s=self,f=fmt: s.save(f)
			menulist.append((fmt, "save", cmd))
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Save",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Save plot as")

		# ---
		col,row = 0,1
		b = Ribbon.LabelButton(group.frame,
				text="Print",
				image=icons.get("print"),
				command=self.hardcopy)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Print selected plots")

		# ---
		col,row = 0,2
		b = Ribbon.LabelButton(group.frame,
				text="Notes",
				image=icons.get("document"),
				command=self.toNotes)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Insert to notes")

		# ---
		col,row = 2,0
		b = Ribbon.LabelButton(group.frame,
				text="Clean",
				image=icons.get("clean32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.clean)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Remove all generated files")

		# ---
		col,row = 3,0
		b = Ribbon.LabelButton(group.frame,
				text="Plot",
				image=icons.get("plot32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.execute)
		b.grid(row=row, rowspan=3, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Do the plot. Generate files when needed")

	#----------------------------------------------------------------------
	# Create Project page
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage(scrolly=True)

		self.splitter = tk.PanedWindow(self.frame,
					orient=tk.HORIZONTAL,
					showhandle=0,
					sashrelief=tk.RAISED,
					sashwidth=4,
					opaqueresize=1)
		self.splitter.pack(expand=tk.YES, fill=tk.BOTH)

		# -----------
		frame = tk.Frame(self.splitter)
		self.splitter.add(frame, minsize=50,
				width=tkFlair.config.getInt(tkFlair._PAGE_SECTION,
						f"{self.name}.split", 240))
		self.tree = PlotTreeview(frame, show="tree")
		self.tree.pack(fill=tk.BOTH, expand=tk.YES, side=tk.LEFT)

		sb = ttk.Scrollbar(frame, orient=tk.VERTICAL, command=self.tree.yview)
		sb.pack(fill=tk.Y, side=tk.RIGHT)
		self.tree.config(yscrollcommand=sb.set)
		#self.tree["yscrollcommand"] = sb.set

		self.tree.bind('<Button-3>',		self.popup)
		self.tree.bind('<<Rename>>',		self.rename)
		self.tree.bind('<Return>',		self.rename)
		self.tree.bind('<KP_Enter>',		self.rename)
		self.tree.bind('<Control-Return>',	self.execute)
		#self.tree.bind('<Escape>',		self.selectionClean)
		self.tree.bind('<Delete>',		self.delete)
		self.tree.bind('<<MoveLeft>>',		self.moveLeft)
		self.tree.bind('<<MoveRight>>',		self.moveRight)
		self.tree.bind('<<MoveUp>>',		self.moveUp)
		self.tree.bind('<<MoveDown>>',		self.moveDown)
		self.tree.bind('<<TreeviewSelect>>',	self.select)
		self.tree.bind('<<Copy>>',		self.copy)
		self.tree.bind('<<Cut>>',		self.cut)
		self.tree.bind('<<Paste>>',		self.paste)
		self.tree.bind('<<PasteSpecial>>',	self.pasteAbove)

		self.tree.focus_set()

		# -----------
		self.base = BasePlot.CommonPlotFrame(self.splitter, self.flair)
		self.splitter.add(self.base, minsize=320)

		# ----- Create Plot Frames ----
		self.frames = {}
		for C in [GeometryPlot.GeometryPlotFrame,
			  UsrPlot.Usr1DPlotFrame,
			  Usr2Plot.Usr2DPlotFrame,
			  UsrbinPlot.UsrbinPlotFrame,
			  UserdumpPlot.UserdumpPlotFrame,
			  ResnucleiPlot.ResnucleiPlotFrame]:
			frame = C(self.base.common, self.flair)
			self.frames[frame.type()] = frame

	# ----------------------------------------------------------------------
	def help(self):
		self.displayed.help()

	# ----------------------------------------------------------------------
	def canUndo(self):	return self.undoredo.canUndo()
	def canRedo(self):	return self.undoredo.canRedo()
	def resetUndo(self):	self.undoredo.reset()

	# ----------------------------------------------------------------------
	def undo(self, event=None):
		if self.canUndo():
			self.undoredo.undo()
			self.selected = None
			self.refresh()

	# ----------------------------------------------------------------------
	def redo(self, event=None):
		if self.canRedo():
			self.undoredo.redo()
			self.selected = None
			self.refresh()

	# ----------------------------------------------------------------------
	# Update tree and add undo structure
	# ----------------------------------------------------------------------
	def addUndo(self):
		self.undoredo.add(self.project.plots.setUndo(self.tree.items()))
		self.modified()

	# ----------------------------------------------------------------------
	def copy(self, event=None):
		# Operate only on the tree
		if self.page.focus_get() is not self.tree:
			return super().copy(event)
		self.get()
		self.tree.selectBranches()

		# Copy to clipboard
		bio = io.BytesIO()
		pickler = pickle.Pickler(bio)
		pickler.dump(self.tree.selectedItems())
		self.flair.clipboard_clear()
		self.flair.clipboardAppend(bio, tkFlair._PLOT_MIME)
		return "break"

	# ----------------------------------------------------------------------
	def cut(self, event=None):
		# Operate only on the tree
		if self.page.focus_get() is not self.tree:
			return super().cut(event)
		self.copy()
		self.delete()
		return "break"

	# ----------------------------------------------------------------------
	# Paste below current item
	# ----------------------------------------------------------------------
	def paste(self, event=None, below=True):
		if self.page.focus_get() is not self.tree:
			return super().paste(event)

		# Operate only on the tree
		self.get()
		if tkFlair._PLOT_MIME in self.flair.clipboardTargets():
			unpickler = self.flair.clipboardUnpickler(tkFlair._PLOT_MIME)
			plots = unpickler.load()
		else:
			self.flair.notify("Cannot paste",
					"Invalid content in clipboard",
					tkFlair.NOTIFY_WARNING)
			return "break"

		active = self.tree.focus()
		plot   = self.tree.item2info.get(active)
		if active is None:
			parent = ""
			index  = 0
		elif self.tree.item(active, "open") and \
		     isinstance(plot, GroupInfo):
			parent = active
			index  = 0
		else:
			parent = self.tree.parent(active)
			index  = self.tree.index(active)+int(below)

		selection = []
		for p in plots.childs:
			item = self.tree.insert(parent, index, p)
			selection.append(item)
			index += 1
		if selection:
			self.tree.selection_set(selection)
			first = selection[0]
		else:
			first = ""
		self.tree.focus(first)
		self.tree.see(first)
		self.addUndo()
		self.select()
		self.flair.refreshGeometry()
		return "break"

	# ----------------------------------------------------------------------
	# Paste plots above current plot
	# ----------------------------------------------------------------------
	def pasteAbove(self, event=None):
		return self.paste(event, False)

	# ----------------------------------------------------------------------
	# Delete plot
	# ----------------------------------------------------------------------
	def delete(self, event=None):
		if self.page.focus_get() is not self.tree:
			return super().delete(event)
		self.get()	# get all information before pushing into undo

		selection = self.tree.selection()
		active    = self.tree.focus()
		path      = self.tree.path(active)
		idx       = [self.tree.index(x) for x in self.tree.dir(active)]
		self.tree.delete(*selection)
		self.tree.focus_set()

		# find item to select
		parent = ""
		child = None
		for i,p in zip(idx,path):
			try:
				child = self.tree.get_children(parent)[i]
				if self.tree.item(child, "text") != p:
					break
				parent = child
			except IndexError:
				while True:
					try:
						child = self.tree.get_children(parent)[-1]
						break
					except IndexError:
						if parent=="": break
						parent = self.tree.parent(parent)
		if child is not None:
			self.tree.selection_set(child)
			self.tree.focus(child)
			self.tree.see(child)

		self.selected = None
		self.addUndo()
		self.select()
		self.flair.refreshGeometry()
		return "break"

	# ----------------------------------------------------------------------
	# Clone plots
	# ----------------------------------------------------------------------
	def clone(self, event=None):
		if self.page.focus_get() is not self.tree:
			return super().clone(event)
		if not self.tree.selection(): return
		self.copy()
		active = self.tree.focus()
		if self.tree.item(active, "open"):
			self.tree.focus(self.tree.next(active))
		self.paste()
		return "break"

	# ----------------------------------------------------------------------
	# Activate and set focus
	# ----------------------------------------------------------------------
	#def activate(self):
	#	#self.refresh()
	#	super().activate()

	# ----------------------------------------------------------------------
	# Refresh display and reload all plots
	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		super().refresh(what)
		engine = self.project["plotengine"]
		if engine: self._plotEngine.set(engine)
		if self.flair.geometryPage:
			self.flair.geometryPage.saveState()

		# remember open and selection
		# WARNING: record the selection in a set AFTER
		#          the "plot" are modified for the "open"
		#          Otherwise the hash(plot) will be different
		#          from what was used to construct the set
		#          and it will not find the items
		self.tree.scanOpen()
		focused  = self.tree.item2info.get(self.tree.focus())
		selected = {self.tree.item2info.get(x) for x in self.tree.selection()}

		ttk.Treeview.delete(self.tree, *self.tree.get_children())
		self.tree.clear()
		self.tree.insert("", tk.END, self.project.plots._tree)

		if selected:
			for plot in selected:
				item = self.tree.info2item(plot)
				if item:
					self.tree.selection_add(item)
			item = self.tree.info2item(focused)
			if item:
				self.tree.focus(item)
				self.tree.see(item)
		else:
			try:
				first = self.tree.get_children()[0]
				self.tree.focus(first)
				self.tree.selection_set(first)
			except IndexError:
				pass
		self.select()

	# ----------------------------------------------------------------------
	# Reset before loading new projects
	# ----------------------------------------------------------------------
	def reset(self):
		if self.page is None: return
		self.selected = None
		self.tree.delete(*self.tree.get_children())

	# ----------------------------------------------------------------------
	# Get information from the plots
	# ----------------------------------------------------------------------
	def get(self):
		#Global.dump("*** PlotPage.get")
		if not self.selected: return
		self.tree.scanOpen()
		self.project["plotengine"] = self._plotEngine.get()
		for plot in self.selected:
			# Remember for undo
			oldplot = plot.clone()

			# Get information
			try:
				self.base.get(plot)
				if self.displayed is not None and \
				   plot.type == self.displayed.type():
					self.displayed.get(plot)
			except tk.TclError:
				continue

			if hash(plot) != hash(oldplot):
				undoinfo = (plot.setUndo, oldplot)
				self.undoredo.add(undoinfo)
				self.flair.modified()

	# ----------------------------------------------------------------------
	# Selection changed in the tree
	# ----------------------------------------------------------------------
	def select(self, event=None):
		#Global.dump("*** PlotPage.select")
		self.get()

		displayed = self.displayed
		oldplot   = displayed.plot if displayed else None

		# As a precaution remove everything
		#for child in self.base.common.winfo_children():
		#	child.pack_forget()
		self.selected = self.tree.selectedPlots()
		if not self.selected: return

		# Find common type
		plot = self.tree.item2info.get(self.tree.focus())
		if plot not in self.selected: plot = None
		if isinstance(plot, PlotInfo):
			t = plot.type
		else:
			plot = t = None

		for p in self.selected:
			if plot is None:
				plot = p
				t = plot.type
			elif t != p.type:
				t = None
				break

		# Display the new frame
		if t:
			self.displayed = self.frames[plot.type]
			if self.displayed is not displayed:
				for child in self.base.common.winfo_children():
					child.pack_forget()
				self.displayed.pack(expand=tk.YES, fill=tk.BOTH)
		else:
			for child in self.base.common.winfo_children():
				child.pack_forget()
			self.displayed = None

		# Fill values with plot type
		if self.displayed: self.displayed.setState(tk.NORMAL)
		self.base.setState(tk.NORMAL)

		# Fill in reverse order so maybe the set to modify the plot like in usrbin
		if plot is not None and \
		   (len(self.selected)==1 or self.displayed is not displayed):
			self.base.set(plot)
			if self.displayed:
				self.displayed.set(plot)

		if len(self.selected)>1:
			self.base.setState(tk.DISABLED)
			if self.displayed: self.displayed.setState(tk.DISABLED)
		else:
			self.base.setState(tk.NORMAL)
			if self.displayed: self.displayed.setState(tk.NORMAL)

	# ----------------------------------------------------------------------
	def selectAll(self):
		if self.page.focus_get() is not self.tree:
			return super().selectAll()
		else:
			self.tree.selectAll()
		return "break"

	# ----------------------------------------------------------------------
	def selectClear(self):
		if self.page.focus_get() is not self.tree:
			return super().selectClear()
		else:
			self.tree.selectClear()
		return "break"

	# ----------------------------------------------------------------------
	def selectInvert(self):
		if self.page.focus_get() is not self.tree:
			return super().selectInvert()
		else:
			self.tree.selectInvert()
		return "break"

	# ----------------------------------------------------------------------
	# Add a new plot
	# ----------------------------------------------------------------------
	def add(self, type_=None):
		plot = PlotInfo(self.project, "")
		plot.type = type_ if type_ else Plot.PLOT_TYPES[0]
		plot["title"] = f"Plot {plot.type}"
		if self.project.name() == "":
			name   = "plot"
		else:
			fn = os.path.splitext(self.project.name())[0]
			name = f"{fn}_plot"
		plot.name = self.project.plots.uniqueName(name)
		plot["format"] = Plot.DEFAULT_FORMAT

		# Detector specific
		if plot.type == "USRBIN":
			plot["cbtics"]   = 1
			plot["cblog"]    = 1
			plot["cbcolors"] = 30

		elif plot.type == "USR-1D":
			plot["ndetectors"] = 1
			plot["name.0"] = "#Detector 1"
			plot["det.0"]  = 0

		elif plot.type == "USR-2D":
			plot["ztics"]  = 1
			plot["cbtics"] = 1

		elif plot.type == "RESNUCLE":
			plot["cbtics"] = 1

		elif plot.type == "USERDUMP":
			plot["type"] = "Histogram"

		item = self.tree.add(plot)
		self.tree.item2info[item] = plot
		#item = self.tree.focus()
		#itemplot = self.tree.item2info.get(item)
		#if self.tree.item(item, "open") and \
		#   not isinstance(itemplot, PlotInfo):
		#	plot.dirname(self.tree.path(item))
		#	item = self.tree.insert(item, 0, plot)
		#else:
		#	plot.dirname(self.tree.path(self.tree.parent(item)))
		#	item = self.tree.insert(self.tree.parent(item),
		#			self.tree.index(item)+1,
		#			plot)
		self.addUndo()
		self.tree.selection_set(item)
		self.tree.focus(item)
		self.tree.see(item)
		self.tree.focus_set()
		self.lastrelease = (item, 0)
		self.select()
		self.flair.refreshGeometry()

	# ----------------------------------------------------------------------
	def addGeometry(self):	self.add("Geometry")
	def addUsr1D(self):	self.add("USR-1D")
	def addUsr2D(self):	self.add("USR-2D")
	def addUsrbin(self):	self.add("USRBIN")
	def addUserdump(self):	self.add("USERDUMP")
	def addResnuclei(self):	self.add("RESNUCLE")

	# ----------------------------------------------------------------------
	def addGroup(self, event=None):
		if self.tree.addGroup():
			self.addUndo()

	# ----------------------------------------------------------------------
	# Move up select items by one
	# ----------------------------------------------------------------------
	def moveUp(self, event=None):
		if self.page.focus_get() is not self.tree: return
		self.get()
		self.tree.moveUp()
		self.addUndo()
		self.selected = self.tree.selectedPlots()
		return "break"

	# ----------------------------------------------------------------------
	# Move down select items by one
	# ----------------------------------------------------------------------
	def moveDown(self, event=None):
		if self.page.focus_get() is not self.tree: return
		self.get()
		self.tree.moveDown()
		self.addUndo()
		self.selected = self.tree.selectedPlots()
		return "break"

	# ----------------------------------------------------------------------
	# Move left (up one parent)
	# ----------------------------------------------------------------------
	def moveLeft(self, event=None):
		if self.page.focus_get() is not self.tree: return
		self.get()
		self.tree.moveLeft()
		self.addUndo()
		self.selected = self.tree.selectedPlots()
		return "break"

	# ----------------------------------------------------------------------
	# Move right (down one parent if next is group)
	# ----------------------------------------------------------------------
	def moveRight(self, event=None):
		if self.page.focus_get() is not self.tree: return
		self.get()
		self.tree.moveRight()
		self.addUndo()
		self.selected = self.tree.selectedPlots()
		return "break"

	# ----------------------------------------------------------------------
	def wizard(self):
		"""scan input file for possible plots"""
		if self.project.name() == "":
			pattern = "%s"
		else:
			pattern = "%s_%%s" % (os.path.splitext(self.project.name())[0])

		# Add one geometry plot if doesn't exist
		for p in self.project.plots():
			if p.type == Plot.PLOT_TYPES[0]: break
		else:
			plot = PlotInfo(self.project, pattern % ("geometry"))
			plot["title"]  = f"Geometry {self.project.title}"
			plot.type      = Plot.PLOT_TYPES[0]
			plot["format"] = ".eps"
			self.project.plots.add(plot)

		# For default Run find all usrinfos
		run = self.project.runs.default()
		for usr in run.usrinfo:
			name = os.path.splitext(usr.name)[0]

			plot = PlotInfo(self.project, f"{name}_plot")
			self.project.plots.add(plot)
			if usr.type=="usrbin":
				plot.type = "USRBIN"
				plot["datafile"] = usr.name
				plot["geo"]      = BasePlot._GEO_AUTO
				plot["cbtics"]   = 1
				plot["cblog"]    = 1
				plot["cbcolors"] = 30

			elif usr.type=="resnucle":
				plot.type = "RESNUCLE"
				plot["datafile"] = usr.name

			else:
				plot.type = "USR-1D"
				plot["xlog"] = 1
				plot["ylog"] = 1
				plot["ndetectors"] = 1
				tablis = f"{name}_tab.lis"
				plot["name.0"] = "Detector 1"
				plot["file.0"] = tablis
				plot["det.0"]  = "0"
				plot["x.0"]    = "2"
				plot["y.0"]    = "0"
				plot["lw.0"]   = "2"
				plot["lc.0"]   = "red"
				plot["with.0"] = "histerror"
			plot["title"]  = f"{plot.type} {name.replace('_',' ')}"
			plot["format"] = ".eps"

		self.refresh()
		self.modified()

	# ----------------------------------------------------------------------
	# Rename plot
	# ----------------------------------------------------------------------
	def rename(self, event=None):
		active = self.tree.focus()
		if not active: return
		plot = self.tree.item2info.get(active)
		if isinstance(plot, PlotInfo):
			# Edit the filename
			filetypes = []
			for ext in sorted(tkFlair._PLOT_FORMAT):
				desc = bFileDialog.DESCRIPTION.get(ext[1:],f"File type {ext}")
				filetypes.append((desc, "*"+ext))
			filetypes.append(("All","*"))
			edit = tkExtra.InPlaceFile(self.tree,
							item=active,
							value=plot.name+plot("format"),
							title="Save plot file as",
							showext=False,
							filetypes=filetypes)
			if edit.value:
				fn = self.project.relpath(edit.value)
				plot.name, ext = os.path.splitext(fn)
				if ext: plot["format"] = ext
				self.tree.item(active, text=plot.name)
				self.modified()
		else:
			# Edit the group name
			value = self.tree.item(active, "text")
			edit  = tkExtra.InPlaceEdit(self.tree, item=active, value=value)
			if edit.value and edit.value != value:
				self.addUndo()
		# Break not to insert the <Return> in the Text and modify the text
		return "break"

	# ----------------------------------------------------------------------
	def popup(self, event):
		menu=tk.Menu(self.tree, tearoff=0)

		submenu = tk.Menu(menu)
		menu.add_command(label="Rename", image=icons.get("rename"), compound=tk.LEFT,
				command=self.rename)

		menu.add_cascade(label="Add", image=icons.get("add"), compound=tk.LEFT,
				underline=0, menu=submenu)
		submenu.add_command(label="group",
				compound=tk.LEFT, image=icons.get("group"),
				command=self.addGroup)
		for t in Plot.PLOT_TYPES:
			icon = icons.get(t.upper())
			submenu.add_command(label=t,
				compound=tk.LEFT, image=icon,
				command=lambda s=self,t=t:s.add(t))

		submenu = tk.Menu(menu)
		menu.add_cascade(label="Change to", image=icons.get("change"), compound=tk.LEFT,
				underline=7, menu=submenu)
		self._changeTypeMenu(submenu)

		# --- Std cut & copy & paste menus ---
		menu.add_separator()
		menu.add_command(label="Cut", underline=2, command=self.cut,
				accelerator="Ctrl-X",
				compound=tk.LEFT, image=icons.get("cut"))
		menu.add_command(label="Copy", underline=0, command=self.copy,
				accelerator="Ctrl-C",
				compound=tk.LEFT, image=icons.get("copy"))
		menu.add_command(label="Paste", underline=0, command=self.paste,
				accelerator="Ctrl-V",
				compound=tk.LEFT, image=icons.get("paste"))
		menu.add_command(label="Clone", underline=1, command=self.clone,
				accelerator="Ctrl-D",
				compound=tk.LEFT, image=icons.get("clone"))
		menu.add_command(label="Delete", image=icons.get("x"), compound=tk.LEFT,
				underline=0, command=self.delete)

		menu.tk_popup(event.x_root, event.y_root)

	# ----------------------------------------------------------------------
	def changeTypePopup(self):
		menu=tk.Menu(self.tree, tearoff=0)
		self._changeTypeMenu(menu)
		bbox = self.tree.bbox(self.tree.focus())
		menu.tk_popup(self.tree.winfo_rootx(),
			      self.tree.winfo_rooty() + bbox[1] + bbox[3])

	# ----------------------------------------------------------------------
	def _changeTypeMenu(self, menu):
		for t in Plot.PLOT_TYPES:
			icon = icons.get(t.upper())
			menu.add_command(label=t,
				compound=tk.LEFT, image=icon,
				command=lambda s=self,t=t:s.changePlotType(t))

	# ----------------------------------------------------------------------
	# Change the Plot type
	# ----------------------------------------------------------------------
	def changePlotType(self, newtype):
		self.tree.focus_set()
		selection = self.tree.selection()
		if not selection: return
		for pid in selection:
			plot = self.tree.item2info.get(pid)
			if isinstance(plot, PlotInfo) and plot.type != newtype:
				self.modified()
				plot.type = newtype
				self.tree.item(pid, image = icons.get(plot.type.upper()))
		self.flair.refreshGeometry()

	# ----------------------------------------------------------------------
	# generate the plot
	# ----------------------------------------------------------------------
	def execute(self, show=True, save=False, export=False, clean=False,
			printer=None, fmt=None):
		if not self.selected: return

		self.get()
		if self.flair.startPlotEngine(self._plotEngine.get()) is None: return

		doall    = False
		saved    = []
		printed  = []
		indexes  = {}		# display indexes updated
		contonerror = None	# continue on errors
		for plot in self.selected:
			if not isinstance(plot, PlotInfo): continue
			# Ask to wait before next plot
			if show and not save  and not export and not clean and \
			   plot is not self.selected[0] and not doall and \
			   int(plot("index")) in indexes:
				try:
					ans = tkDialogs.Dialog(self.page.winfo_toplevel(),
							{"title":"Multiple Plot",
							"text":"Continue with the next\n"
								"Or do all plots",
							"bitmap": "questhead",
							"default": 0,
							"strings": ("Next","All","Cancel")})
				except tk.TclError:
					ans = 1
				indexes.clear()
				if ans.num==1:
					doall = True
				elif ans.num in (2,-1):
					break

			self.base.get(plot)
			if self.displayed and plot.type == self.displayed.type():
				self.displayed.get(plot)

			self.busy()
			#try
			log = self.flair.newLog(f"Plot {plot.type}", plot.name)

			# FIXME to be changed with one class!!!
			if self._plotEngine.get() == PLOT_ENGINES[0]:
				plotter = GPPlot.GPPlot(
						self.flair.plotEngine,
						self.project,
						plot,
						log)
			elif self._plotEngine.get() == PLOT_ENGINES[1]:
				plotter = MPPlot.MPPlot(
						self.flair.plotEngine,
						self.project,
						plot,
						log)
			if clean or plot.verify() is not None:
				plotter.cleanup()

			err = None
			if show:
				self.setStatus(f"Processing plot: {plot.name}",
						"Tomato", update=True)
				try:
					err = plotter.show()
				except (OSError, IOError) as error:
					self.flair.notify("OS Error",
						str(error),
						tkFlair.NOTIFY_ERROR)
					err = error
				if self.displayed: self.displayed.afterPlot()
				self.setStatus("Plot completed")
			self.notBusy()
			#except
			#...
			if err:
				if isinstance(err,tuple):
					err,inp,pid = err
				else:
					inp = pid = None
				self.flair.notify("Plot Error",
					f"{err}\nCheck Output window for additional messages.",
					tkFlair.NOTIFY_ERROR,
					"Output", self.flair.showOutput, showOnOutput=False)
				if inp is not None:
					tkFlair.OutputFilesDialog(self.flair,
							os.path.basename(inp),
							pid, 1, self.project)
					if Project.cleanup:
						self.project.deleteFiles(inp, pid)
				if contonerror is None and plot is not self.selected[-1]:
					contonerror = tkFlair.askyesno("Errors found",
							"Errors found during plotting. "
							"Do you want to continue with the next plots?",
							parent=self.page.winfo_toplevel())
				if not contonerror: break

			# Save file if needed
			if save:
				if fmt is None:
					saved.append(plotter.save(plot("format")))
				else:
					plot["format"] = fmt
					saved.append(plotter.save(fmt))

			# Print if needed
			if printer:
				plotter.hardcopy(printer)
				printed.append(plot.name)

			# Update which display indexes have been used up to now
			indexes[int(plot("index"))] = True

		if clean:
			self.setStatus("Plot files cleaned", "DarkGreen")

		if save:
			self.setStatus(f"Saved: {' '.join(saved)}", "DarkBlue")

		if printer:
			self.setStatus(f"Printed: {' '.join(printed)}", "DarkBlue")
		return "break"

	# ----------------------------------------------------------------------
	def export(self):		self.execute(export=True)
	def save(self,fmt=None):	self.execute(save=True, fmt=fmt)
	def clean(self):		self.execute(show=False, clean=True)

	# ----------------------------------------------------------------------
	# Select plot by name
	# ----------------------------------------------------------------------
	def showPlot(self, name):
		for pid, plot in self.tree.item2info.items():
			if plot.name == name:
				self.tree.selectClear()
				self.tree.selection_set(pid)
				self.tree.focus(pid)
				self.tree.see(pid)
				self.tree.focus_set()
				self.select()
				break

	# ----------------------------------------------------------------------
	def hardcopy(self):
		printer = tkFlair.printer(self.page)
		if printer is None: return
		self.execute(printer=printer)
		printer.close()

	# ----------------------------------------------------------------------
	def showNotesPage(self):
		self.flair.notifyHide()
		self.flair.tabs.changePage("Flair")

	# ----------------------------------------------------------------------
	# Insert to project notes a link to the image
	# ----------------------------------------------------------------------
	def toNotes(self):
		added = 0
		self.save(".png")
		for item in self.tree.selection():
			plot = self.tree.item2info[item]
			link = f"{{img:{plot.name}.png}}"
			if link not in self.project.notes:
				self.project.notes += f"\n\n{link}"
				added += 1

		if added:
			self.flair.modified()
			self.flair.refresh("image")
			self.setStatus(f"{added} plots added to notes")
			self.flair.notify("Plots to Notes",
				f"{added} plots added to notes")
		else:
			self.setStatus("Replaced existing plots in Notes")
			self.flair.notify("Plots to Notes",
				"Replaced existing plots to notes",
				tkFlair.NOTIFY_NORMAL,
				"Notes",
				self.showNotesPage)
