# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

__author__  = "Vasilis Vlachoudis"
__email__   = "Vasilis.Vlachoudis@cern.ch"
__version__ = 300

import io
import os
import re
import stat
import time
import shlex
import random
import signal
import shutil
import hashlib
import itertools
import threading
import subprocess
from fnmatch     import fnmatch
from dataclasses import dataclass, field
from operator    import attrgetter
from collections import namedtuple, defaultdict, OrderedDict

import log
import tool
import rexx
import bmath
import inputs				# Used to initialize the various mc
import Global
import convert

from stream   import Stream
from palette  import rndColor, RGB2str

from deck     import Card
from fluka    import Fluka, FlukaCard, FlukaConfig
from mcnp     import Mcnp, McnpConfig
from moira    import Moira, MoiraCard, MoiraConfig
from penelope import Penelope, PenelopeCard, PenelopeConfig
from phits    import Phits, PhitsCard, PhitsConfig

_TIME_FORMAT  = "%Y.%m.%d %H:%M:%S"	# instead from bFileDialog import _TIME_FORMAT

section = "Project"

class InputMD5Exception(Exception): pass

#===============================================================================
# Global variables
refreshInterval = 5 * 1000		# ms
cleanup         = True			# Cleanup temporary files
editor          = "/usr/bin/emacs"	# Default editor
kill            = ""			# kill command
terminal        = "xterm"		# terminal command
debugger        = "gdb"			# debugger
#alphaSortOrder = True			# For tree view menus
keepBackup      = True			# Keep backups
ccprogram       = "/usr/bin/cc"		# C compiler
cppprogram      = "/usr/bin/c++"	# C++ compiler
EXT             = ".flair"		# Project extension
default_mode    = Fluka.MODE		# default mode

# default clone input name
_DEFAULT_RUNNAME = r'\I_\a'

# submit commands
_submit_command = { }
_SUBMIT_DEFAULT = "*Default"

# patterns
_LAST_NUM    = re.compile(r"^(.*?) *(\d+)$")
FLAIRVER_PAT = re.compile(r".*^flair-(\d+\.\d+.\d+\.?\d*[a-z]?)\.tgz (.*?)$",
			re.MULTILINE | re.DOTALL)

# list of global layers GeometryLayerInfo
_systemLayers = {}
_globalLayers = {}

__gnucolors = ("black","red", "green","blue", "magenta","cyan", "yellow")

#-------------------------------------------------------------------------------
# Initialize global layers to the 4 basic ones
#-------------------------------------------------------------------------------
def initSystemLayers():
	"""Initialize global layers to the 4 basic ones"""
	_systemLayers.clear()

	# WARNING it has to be in sync with the flair.ini
	layer = GeometryLayerInfo(None, "Borders", True)
	_systemLayers[layer.name] = layer
	layer["_system"]    = 1
	layer["_order"]     = 0
	layer["Show"]       = 1
	layer["Show.color"] = "None"

	layer = GeometryLayerInfo(None, "Media", True)
	_systemLayers[layer.name] = layer
	layer["_system"]    = 1
	layer["_order"]     = 1
	layer["Show"]       = 1
	layer["Show.color"] = "Material"

	layer = GeometryLayerInfo(None, "3D", True)
	_systemLayers[layer.name] = layer
	layer["_system"]       = 1
	layer["_order"]        = 2
	layer["Show"]          = 0
	layer["Show.lattice"]  = 1
	layer["Show.voxel"]    = 1
	layer["3D"]            = 1
	layer["3D.deflights"]  = 1
	layer["3D.edgedetect"] = 1

	layer = GeometryLayerInfo(None, "Lattice", True)
	_systemLayers[layer.name] = layer
	layer["_system"]      = 1
	layer["_order"]       = 3
	layer["Show"]         = 1
	layer["Show.color"]   = "Material"
	layer["Show.lattice"] = 2
	layer["Show.voxel"]   = 2

	initGlobalLayers()

#-------------------------------------------------------------------------------
def initGlobalLayers():
	_globalLayers.clear()

	for layer in _systemLayers.values():
		_globalLayers[layer.name] = layer.clone()

#-------------------------------------------------------------------------------
# Load configuration
#-------------------------------------------------------------------------------
def loadConfig(config):
	global cleanup, refreshInterval, keepBackup
	global editor, terminal, debugger, kill
	global default_mode

	Global.tmpPrefix= config.getStr(section, "tmpprefix",       Global.tmpPrefix)
	cleanup         = config.getBool(section,"cleanup",         cleanup)
	refreshInterval = config.getInt(section, "refreshinterval", refreshInterval/1000)*1000
	keepBackup      = config.getBool(section,"keepbackup",      keepBackup)
	editor          = config.getStr(section, "editor",          editor)
	terminal        = config.getStr(section, "terminal",        terminal)
	debugger        = config.getStr(section, "debugger",        debugger)
	kill            = config.getStr(section, "kill",            kill)
	default_mode    = config.getStr(section, "defaultmode",     default_mode)
	RunInfo.timeThreshold = config.getInt(section, "timethreshold", RunInfo.timeThreshold)

#-------------------------------------------------------------------------------
def saveConfig(config):
	global default_mode
	config.set(section, "defaultmode", default_mode)

#===============================================================================
# Rules for commands
#===============================================================================
class Command:
	section  = "Data"
	commands = {}	# available commands

	# ----------------------------------------------------------------------
	def __init__(self, idx, name, cmd):
		self.name = name
		if cmd:
			self.cmd = cmd
		else:
			self.cmd = name
		self.idx  = idx
		self.arg  = []
		self.inp  = []
		self.icon = None

	# ----------------------------------------------------------------------
	# Append an argument string separated with ":"
	#
	#  tag, type, default, name
	#
	#  tag=     is the argument tag as "-x"
	#  type=    int, float, output, file, filelist, fixed, or list of values
	#           look the DataTab.inputInPlaceEdit for data types
	#  default= default value, can be any expression on the input
	#           e.g. =C(VOXELS,0,0)+".vxl"
	#  name=    to appear in listbox
	# ----------------------------------------------------------------------
	def addArgument(self, arg):
		self.arg.append(arg.split(":"))

	# ----------------------------------------------------------------------
	# Append an input string separated with ":"
	#
	#  type, default, name
	#
	#  type= like in the arguments
	#        + conditional
	#        repeat : value : label	(repeat from label until value is given)
	# ----------------------------------------------------------------------
	def addInput(self, inp):
		self.inp.append(inp.split(":"))

	# ----------------------------------------------------------------------
	# Return default name
	# ----------------------------------------------------------------------
	def defaultName(self):
		for tag,ty,default,name in self.arg:
			if ty=="output": return default
		for ty,default,name in self.inp:
			if ty=="output": return default
		return UsrInfo.DEFAULT_USRNAME

	# ----------------------------------------------------------------------
	@staticmethod
	def addCommand(idx, name, cmd):
		cmd = Command(idx, name, cmd)
		Command.commands[name] = cmd
		Command.commands[idx]  = cmd
		return cmd

	# ----------------------------------------------------------------------
	@staticmethod
	def clearCommands():
		Command.commands.clear()

	# ----------------------------------------------------------------------
	@staticmethod
	def get(name):
		return Command.commands.get(name)

	# ----------------------------------------------------------------------
	# Load command configuration
	# From 1..99 system configuration and from 100.. user configurations
	# ----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		# Load all commands
		Command.clearCommands()
		i = 1
		while True:
			try:
				value = config.get(Command.section, f"cmd.{i}")
			except:
				if i<100:
					i=100
					continue
				break

			name,prg = value.split(":")
			cmd = Command.addCommand(i,name,prg)

			try:
				cmd.icon = config.get(Command.section, f"icon.{i}")
			except:
				pass
			for a in range(1,100):
				try:
					cmd.addArgument(config.get(Command.section, f"arg.{i}.{a}"))
				except: break
			for j in range(1,100):
				try:
					cmd.addInput(config.get(Command.section, f"inp.{i}.{j}"))
				except: break
			i += 1

	# ----------------------------------------------------------------------
	@staticmethod
	def saveConfig(config):
		pass

#===============================================================================
# Base class of holding Info in the project
#===============================================================================
class Info(dict):
	CLASS   = "Info"	# Type name of class
	PREFIX  = ""		# Indentation prefix when writing
	DEFAULT = {}		# Default values of parameters
	FORMAT  = {}		# Special format handling
				# b - block
				# l - list
	_idcnt  = 0		# unique id used for the undo/redo to identify
				# the cloned objects in the undolist

	# ----------------------------------------------------------------------
	def __init__(self, parent, name=""):
		super().__init__()
		self._parent = parent		# parent of info, typically the project
		self._name   = name		# name of block
		self._id     = Info.newid()
		self.type    = self.CLASS	# sub-type of information

	# ----------------------------------------------------------------------
	@staticmethod
	def newid():
		Info._idcnt += 1
		return Info._idcnt

	# ----------------------------------------------------------------------
	def id(self): return self._id

	# ----------------------------------------------------------------------
	# Pickle information to copy
	# ----------------------------------------------------------------------
	def __getstate__(self):
		state = {
			"_name" : self._name,
			"_id"   : self._id,
			"type"  : self.type,
		}
		return state

	# ----------------------------------------------------------------------
	# UnPickle information to restore
	# ----------------------------------------------------------------------
	def __setstate__(self, state):
		self.__dict__.update(state)
		self._parent = None

	# ----------------------------------------------------------------------
	@property
	def name(self):
		return self._name

	# ----------------------------------------------------------------------
	@name.setter
	def name(self, x):
		self._name = x

	# ----------------------------------------------------------------------
	def dirname(self, x=None):
		if x is None:
			return os.path.dirname(self._name)
		else:
			self.name = os.path.join(x, self.basename())

	# ----------------------------------------------------------------------
	def basename(self, x=None):
		if x is None:
			return os.path.basename(self._name)
		else:
			self.name = os.path.join(self.dirname(), x)

	# ----------------------------------------------------------------------
	# set/get parent
	# ----------------------------------------------------------------------
	def parent(self, val=None):
		if val is None:
			return self._parent
		else:
			self._parent = val

	# ----------------------------------------------------------------------
	# Write header only once if any information exists
	# ----------------------------------------------------------------------
	def _writeHeader(self, stream):
		if not self._headerWritten:
			if self._name:
				stream.write(f"{self.PREFIX}{self.CLASS}: {self._name}\n")
			else:
				stream.write(f"\n{self.PREFIX}# {self.CLASS}\n")
			self._headerWritten = True

	# ----------------------------------------------------------------------
	# Write information
	# ----------------------------------------------------------------------
	def write(self, stream):
		self._headerWritten = False
		if self.type != self.CLASS:
			self._writeHeader(stream)
			stream.write(f"{self.PREFIX}\tType: {self.type}\n")
		for tag in sorted(self):
			val = self[tag]
			if isinstance(val,bool): val=int(val)
			d = self.default(tag)
			if val == d: continue
			val = str(val)
			if val == str(d): continue
			if val != "":
				self._writeHeader(stream)
				if "\n" in val:
					stream.write(f"{self.PREFIX}\t{tag}:\n")
					stream.writeBlock(val.rstrip(), f"{self.PREFIX}\t\t|")
				else:
					stream.write(f"{self.PREFIX}\t{tag}: {val}\n")
		self._customWrite(stream)
		if self._headerWritten: stream.write(f"{self.PREFIX}End\n")

	# ----------------------------------------------------------------------
	# Custom write function to override
	# ----------------------------------------------------------------------
	def _customWrite(self, stream):
		pass

	# ----------------------------------------------------------------------
	# Read information from file.
	# ----------------------------------------------------------------------
	def read(self, stream):
		while True:
			tag, val = stream.parseVarValue()
			fmt = self.FORMAT.get(tag)
			if tag is None: break
			if tag in ("End", "End"+self.CLASS): break
			if tag=="Type":
				self.type = val	# @FIXME special treatment. Maybe add to dictionary
						# and use a @property
			elif self._customRead(stream, tag, val):
				pass
			elif val=="" and fmt == "b":
				self[tag] = stream.readBlock()
			else:
				if fmt == "l":
					self[tag] = Stream.str2list(val)
				elif val in ("True","False"):
					# Convert True/False to int
					self[tag] = int(val=="True")
				else:
					try:
						val = int(val)
					except ValueError:
						pass
					# rest keep it as string
					self[tag] = val

		# FIXME call _upgrade() but we need to check the version of the project!
		# or potentially pass a boolean value to upgrade or not
		if self.version() < __version__: self._upgrade()

	# ----------------------------------------------------------------------
	# Custom read function to override
	# @return True if the tag is handled
	# ----------------------------------------------------------------------
	def _customRead(self, stream, tag, val):
		return False

	# ----------------------------------------------------------------------
	def clone(self):
		c = self.__class__(self._parent, self._name)
		c.copy(self)
		return c

	# ----------------------------------------------------------------------
	def copy(self, src):
		self._parent = src._parent
		self._name   = src._name
		self.type    = src.type
		self.update(src)

	# ----------------------------------------------------------------------
	def __hash__(self):
		dicthash = 0
		for key,value in self.items():
			if str(value) != str(self.default(key)):
				try:
					#Global.dump(">V> key='%s' value='%s'
					#	type=%s def=%s"
					#	%(key,value,type(value),type(self.default(key))),level=2)
					dicthash += hash(key)*61 + hash(str(value))*167
				except TypeError:
					pass
		return hash(self._name) + hash(self.type)*3 + dicthash

	# ----------------------------------------------------------------------
	# Return always a variable if exist else the default
	# ----------------------------------------------------------------------
	def get(self, key):
		return super().get(key, self.default(key))
	__call__ = get

	# ----------------------------------------------------------------------
	# Return default value of key.
	# Keys with a '.' will be matched only for the prefix until the dot
	# ----------------------------------------------------------------------
	def default(self, key):
		try:	# try key as is
			return self.DEFAULT[key]
		except KeyError:
			# check if it is an array
			p = key.rfind(".")
			if p>=0:
				try:
					return self.DEFAULT[key[:p+1]]
				except KeyError:
					if Global.developer:
						Global.dump(f"{log.WARNING}Project.Info "
							f"{self.name!r} missing field "
							f"{self.CLASS}.{key} prefix={key[:p+1]}")
			elif Global.developer:
				Global.dump(f"{log.WARNING}Project.Info "
					f"{self.name!r} missing field "
					f"{self.CLASS}.{key} = {dict.get(self,key,'')}")
			return ""

	# ----------------------------------------------------------------------
	# set default value if doesn't exist
	# ----------------------------------------------------------------------
	def _default(self, key, default):
		if key not in self:
			self[key] = default

	# ----------------------------------------------------------------------
	def __str__(self):
		return f"{self.__class__.__name__}(name={self._name!r})={dict(self)}"

	# ----------------------------------------------------------------------
	def __eq__(self, b):
		return self is b

	# ----------------------------------------------------------------------
	def setUndo(self, undo):
		"""set all parameters and return undo as info"""
		Global.dump("*-* Info.setUndo")
		undoinfo = (self.setUndo, self.clone())
		self.copy(undo)
		return undoinfo

	# ----------------------------------------------------------------------
	# @return a list of dependencies
	# ----------------------------------------------------------------------
	def depends(self):
		return []

	# ----------------------------------------------------------------------
	# @return a list of what files it provides
	# ----------------------------------------------------------------------
	def provides(self, derivatives=False):
		return [self.name]

	# ----------------------------------------------------------------------
	# Verify that all provided files are newer than the dependencies
	# @return the error message of the faulty file, None on success
	# ----------------------------------------------------------------------
	def verify(self, t0=0):
		# Find oldest provides() time
		pt = 1e99
		for p in self.provides():
			try:
				pt = min(pt, os.path.getmtime(p))
			except FileNotFoundError:
				return f"cannot access file '{p}'"

		for d in self.depends():
			try:
				dt = os.path.getmtime(d)
			except FileNotFoundError:
				return f"cannot access file '{d}'"
			if pt < dt:
				return f"'{p}':{pt} was created before than '{d}':{dt}"
			if pt < t0:
				return f"'{p}':{pt} was created before the command" \
					f" was started {t0}"
		return None

	# ----------------------------------------------------------------------
	# create the 'provided' files
	# ----------------------------------------------------------------------
	#def make(self):
	#	pass

	# ----------------------------------------------------------------------
	# clean all temporary and provided files
	# ----------------------------------------------------------------------
	#def clean(self):
	#	pass

	# ----------------------------------------------------------------------
	def delPrefix(self, prefix):
		"""Delete all keys starting with prefix"""
		for name in self:
			if name.startswith(prefix):
				del self[name]

	# ----------------------------------------------------------------------
	def _changeTag(self, old, new):
		"""change name tag or set a default value"""
		if old in self:
			self[new] = self[old]
			del self[old]

	# ----------------------------------------------------------------------
	# Possible upgrade of the data from an old version to the current one
	# ----------------------------------------------------------------------
	def _upgrade(self):
		pass

	# ----------------------------------------------------------------------
	# Return version of the project file
	# ----------------------------------------------------------------------
	def version(self):
		parent = self._parent
		while not isinstance(parent,Project):
			parent = parent._parent
		return parent.version

#===============================================================================
# UsrInfo
#===============================================================================
class UsrException(Exception): pass
class UsrInfo(Info):
	CLASS = "Data"

	# FIXME Fluka extensions move to config
	EXTENSIONS = {	"detect"    : "dtc",
			"resnucle"  : "rnc",
			"usrbin"    : "bnn",
			"usrbdx"    : "bnx",
			"usrtrack"  : "trk",
			"usryield"  : "yie",
			"usrcoll"   : "cll",
		}
	TYPES = [ "detect",
		  "resnucle",
		  "usrbdx",
		  "usrbin",
		  "usrcoll",
		  "usrtrack",
		  "usryield",
		  "merge.py" ]

	# default names
	DEFAULT_USRNAME = r'\I_\U\e'
	USR_NAME        = {}	# custom user names

	# default rules
	RULES = [('+',r"\I\d\d\d_fort\.\U"),
		 ('+',r"\I\d\d\d_\U\e") ]

	# ----------------------------------------------------------------------
	def __init__(self, run, name=DEFAULT_USRNAME):
		super().__init__(run, name)
		self.auto  = True	# automatically generated or manually modified
		self.type  = 'usrbin'
		self.unit  = 0
		self.rules = []
		self.arg   = []		# Arguments for generic commands
		self.inp   = []		# Input parameters for generic commands
		self.cmd   = None

	# ----------------------------------------------------------------------
	# Pickle information to copy
	# ----------------------------------------------------------------------
	def __getstate__(self):
		state = super().__getstate__()
		state.update({
			"auto"  : self.unit,
			"unit"  : self.unit,
			"rules" : self.rules,
			"arg"   : self.arg,
			"inp"   : self.inp,
			"cmd"   : self.cmd,
		})
		return state

	# ----------------------------------------------------------------------
	@property
	def run(self):
		return self._parent

	# ----------------------------------------------------------------------
	@run.setter
	def run(self, x):
		#raise Exception("Cannot set project in UsrInfo")
		self._parent = x

	# ----------------------------------------------------------------------
	@property
	def project(self):
		return self.run.project

	# ----------------------------------------------------------------------
	# Write usr data information
	# ----------------------------------------------------------------------
	def write(self, stream):
		stream.write(f"\n\t# USRxxx data file: {self.name}\n")
		stream.write(f"\tData: {self._name}\n")
		stream.write(f"\t\tUnit: {self.unit}\n")
		stream.write(f"\t\tType: {self.type}\n")
		if self.rules != UsrInfo.RULES:
			for i in self.rules:
				stream.write(f"\t\tRule: {i[0]},{i[1]}\n")
		for a in self.arg:
			stream.write(f"\t\tArg: {a}\n")
		for i in self.inp:
			stream.write(f"\t\tInp: {i}\n")
		stream.write("\tEnd\n")

	# ----------------------------------------------------------------------
	# Read usr data information
	# ----------------------------------------------------------------------
	def read(self, stream):
		self.auto  = False
		self.rules = []
		while True:
			tag, val = stream.parseVarValue()
			if tag is None:
				break
			if tag=="Unit":
				self.unit = val
			elif tag=="Type":
				self.type = val
				self.cmd = Command.get(self.type)
			elif tag=="Recursive": pass	# from V1
			elif tag=="Rule":
				self.rules.append(tuple(val.split(",")))
			elif tag=="Arg":
				self.arg.append(val)
			elif tag=="Inp":
				self.inp.append(val)
			elif tag in ("End","EndData"):
				break
			else:	# Add in dictionary
				raise UsrException("Unknown Data keyword "+tag)
		if self.version() < __version__: self._upgrade()

	# ----------------------------------------------------------------------
	def set(self, unit=None, type_=None):
		if unit is not None: self.unit = unit
		if type_ is not None:
			# Only if name is the default one change it
			if self._name == UsrInfo.USR_NAME.get(self.type, UsrInfo.DEFAULT_USRNAME):
				self._name = UsrInfo.USR_NAME.get(type_, UsrInfo.DEFAULT_USRNAME)
			self.type = type_
			self.cmd  = Command.get(self.type)
		self.rules = UsrInfo.RULES[:]

	# ----------------------------------------------------------------------
	# return/set name
	# ----------------------------------------------------------------------
	@property
	def name(self):
		filename = self.expand(self._name)
		if os.path.sep in filename: return filename
		d = self.run.dir()
		if d: filename = os.path.join(d,filename)
		#if self.run("parent"):
		#	parent = self.run.parentRun()
		#	d = self.project.relpath(parent.dir())
		#	if d: filename = os.path.join(d,filename)
		return self.project.relpath(filename)

	# ----------------------------------------------------------------------
	@name.setter
	def name(self, name):
		if name:
			self._name = name
		else:
			self._name = self.cmd.defaultName()

	# ----------------------------------------------------------------------
	# Set argument, expand array if needed
	# ----------------------------------------------------------------------
	def setArg(self, idx, value):
		if idx >= len(self.arg):
			self.arg.extend([""]*(idx-len(self.arg)+1))
		self.arg[idx] = value

	# ----------------------------------------------------------------------
	def setInp(self, idx, value):
		if idx >= len(self.inp):
			self.inp.extend([""]*(idx-len(self.inp)+1))
		self.inp[idx] = value

	# ----------------------------------------------------------------------
	# Insert empty values for repeated statements starting from repeat
	# ----------------------------------------------------------------------
	def insertRepeat(self, pos, repeat):
		while repeat < len(self.cmd.inp):
			ctype,cdef,cname = self.cmd.inp[repeat]
			repeat += 1
			if ctype == "repeat": break
			self.inp.insert(pos, "")
			pos += 1

	# ----------------------------------------------------------------------
	# Delete all terms until "repeat"
	# ----------------------------------------------------------------------
	def deleteRepeat(self, pos, repeat):
		while repeat < len(self.cmd.inp):
			ctype,cdef,cname = self.cmd.inp[repeat]
			repeat += 1
			if ctype == "repeat": break
			del self.inp[pos]

	# ----------------------------------------------------------------------
	def __str__(self):
		return f"UsrInfo(u={self.unit}, t={self.type}, n={self._name}, files={self.rules})"
	__repr__ = __str__

	# ----------------------------------------------------------------------
	def clone(self):
		c = UsrInfo(self.run)
		c.copy(self)
		return c

	# ----------------------------------------------------------------------
	def copy(self, src):
		self._name = src._name
		# will override the one from clone! and will break the RunInfo.clone()
		self.auto  = src.auto
		self.run   = src.run
		self.type  = src.type
		self.unit  = src.unit
		self.rules = src.rules[:]
		self.arg   = src.arg[:]
		self.inp   = src.inp[:]
		self.cmd   = src.cmd

	# ----------------------------------------------------------------------
	def __hash__(self):
		h = hash(self._name) ^ hash(self.type) ^ hash(self.unit)
		for _,r in self.rules:
			h ^= r
		h ^= hash(self.arg)
		h ^= hash(self.inp)
		return h

	# ----------------------------------------------------------------------
	# Return argument value for name or index
	# ----------------------------------------------------------------------
	def getArg(self, name):
		if isinstance(name,int):
			try:
				value = self.arg[name]
			except:
				value = ""
			cdef = self.cmd.arg[name][2]
			if not value and cdef: value = self.expand(cdef)
			return value
		else:
			return None

	# ----------------------------------------------------------------------
	# expand escape characters in string
	# ----------------------------------------------------------------------
	def expand(self, s, esc=False, run=None):
		if run is None: run = self.run
		if esc:
			s = s.replace(r"\F", re.escape(run.project.name()))
			s = s.replace(r"\I", re.escape(run.inputBaseName()))
		else:
			s = s.replace(r"\F", run.project.name())
			s = s.replace(r"\I", run.inputBaseName())
		s = s.replace(r"\U", str(self.unit))
		s = s.replace(r"\N", str(self.unit))
		s = s.replace(r"\t", self.type)
		s = s.replace(r"\T", self.type)
		if "." in str(self.unit):
			s = s.replace(r"\e", "")
		if r"\e" in s:
			ext = "." + UsrInfo.EXTENSIONS.get(self.type,"dat")
			s = s.replace(r".\e", ext)	# from V2
			s = s.replace(r"\e", ext)	# and the rest

		if self.cmd is not None:
			for i in range(1,10):
				target = rf"\{i}"
				if target in s:
					s = s.replace(target, self.getArg(i-1))

		if s and s[0]=="=":
			# evaluate expression and return result
			try:
				s = eval(s[1:], run.project.input.globalDict,
						run.project.input.localDict)
			except:
				pass
		return s

	# ----------------------------------------------------------------------
	def compile(self, rule, run=None):
		rule = self.expand(rule, True, run)
		if rule[0]  != "^": rule = f"^{rule}"
		if rule[-1] != "$": rule = f"{rule}$"
		return re.compile(rule)

	# --------------------------------------------------------------------
	# Return processing program command
	# --------------------------------------------------------------------
	def command(self, cwd=None):
		if self.cmd is None or (not self.cmd.arg and not self.cmd.inp):
			inp = self.depends(cwd)
			if not inp: return None,None
			inp.append("")
			if cwd:
				inp.append(os.path.join(cwd,self.name))
			else:
				inp.append(self.name)
			config = self.run.config()
			if config is FlukaConfig:
				prg = config.command(self.type)
			else:
				prg = self.type
			return [config.command(prg)], inp
		else:
			cmd = [self.run.config().command(self.cmd.cmd)]
			for i,(ctag,ctype,cdef,cname) in enumerate(self.cmd.arg):
				try:
					value = self.arg[i]
				except:
					value = ""
				if not value and cdef: value = self.expand(cdef)
				if value:
					if ctag: cmd.append(ctag)
					cmd.append(str(value))

			inp = []
			i = 0		# rule index
			j = 0		# input index
			while i < len(self.cmd.inp):
				ctype,cdef,cname = self.cmd.inp[i]
				i += 1
				try:
					u = self.inp[j]
					j += 1
				except:
					u = ""
				if ctype == "repeat":
					if u != cdef:
						i  = int(cname)-1
						j -= 1
						continue
				if not u and cdef: inp.append(self.expand(cdef))
				inp.append(u)
			return cmd,inp

	# --------------------------------------------------------------------
	# Override default provides() list with addition _sum.lis
	# and _tab.lis files
	# ----------------------------------------------------------------------
	def provides(self, derivatives=False):
		p = super().provides(derivatives)
		fn = os.path.splitext(self.name)[0]
		if self.type in ('usrcoll','detect','usrtrack','usrbdx','usryield','resnucle'):
			if self.type != 'detect':
				p.append(f"{fn}_sum.lis")
			p.append(f"{fn}_tab.lis")
		elif self.type == 'usrbin' and derivatives:
			p.append(f"{p[0]}.lis")
		return p

	# --------------------------------------------------------------------
	# Find files by executing the rules
	# --------------------------------------------------------------------
	def depends(self, cwd=None):
		files = []
		if cwd is None:
			rundir = self.run.dir()
		else:
			rundir = cwd

		if self.cmd:
			for i,(ctag,ctype,cdef,cname) in enumerate(self.cmd.arg):
				if ctype not in ("file","filelist"): continue
				fn = self.getArg(i)
				if fn: files.append(fn)
			i = 0	# rule index
			j = 0	# input index
			while i < len(self.cmd.inp):
				ctype,cdef,cname = self.cmd.inp[i]
				i += 1
				try:
					u = self.inp[j]
					j += 1
				except:
					u = ""
				if ctype == "repeat":
					if u != cdef:
						i  = int(cname)-1
						j -= 1
						continue
				if ctype not in ("file","filelist"): continue
				try:
					files.append(u)
				except:
					if cdef:
						files.append(self.expand(cdef))
		else:
			# add files in an ordered dict to avoid duplications
			filelist = OrderedDict()
			for op,rule in self.rules:
				if "*" in rule or "\\" in rule:
					try:
						pattern = [(rundir, self.compile(rule))]
						for child in self.run.family:
							if cwd is None:
								childdir = child.dir()
							else:
								childdir = cwd
							pattern.append((childdir,
								self.compile(rule, child)))
					except Exception as ex:
						log.error(ex)
						continue
					self._scanDir(filelist, op, pattern)
				else:
					self._doRule(filelist, op, rundir, rule)
			files.extend(filelist.keys())
		return files

	# --------------------------------------------------------------------
	# Scan Directory for Files
	# --------------------------------------------------------------------
	def _scanDir(self, files, op, patterns):
		for path, pattern in patterns:
			try: os.stat(path)
			except OSError: continue
			# In the current path
			for f in tool.listdir(path):
				if pattern.match(f):
					self._doRule(files, op, path, f)

	# --------------------------------------------------------------------
	# Add/Del a single file to the list
	# WARNING files can be an ordereddict
	# (from UsrInfo.depends) and/or list from DataTab
	# --------------------------------------------------------------------
	def _doRule(self, files, op, path, filename):
		if op == "-":	# Delete file
			# delete filename from list/dict
			while True:
				if isinstance(files, list):
					try:
						files.remove(filename)
					except ValueError:
						return
				else:
					try:
						del files[filename]
					except KeyError:
						return
		# Add filename to list/dict
		elif isinstance(files, list):
			files.append(self.run.project.relpath(
					os.path.join(path, filename)))
		else:
			files[self.run.project.relpath(os.path.join(path, filename))] = None

	# ----------------------------------------------------------------------
	def _upgrade(self):
		if self.version() <= 2:
			if len(self.type)==1:
				types = {"d": "detect",
					 "r": "resnucle",
					 "b": "usrbin",
					 "x": "usrbdx",
					 "t": "usrtrack",
					 "y": "usryield",
					 "c": "usrcoll",
					}
				self.type = types.get(self.type,self.type)

	# ----------------------------------------------------------------------
	# Load configuration
	# ----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		del UsrInfo.RULES[:]	# clear the rules
		#items = config.items(Command.section)
		#if not items:
		#	for k,v in Project.UsrInfo.USR_NAME.items():
		#		config.set(Command.section, k, v)
		#	i = 1
		#	for t,r in Project.UsrInfo.RULES:
		#		config.set(Command.section, f"rule.{i}", "%s%s"%(t,r))
		#		i += 1
		for k,v in config.items(Command.section):
			# --- XXX  cleanup of rule: +_\I_\d\d\d\d\d_fort\.\U  XXX ---
			# To be cleaned latter
			# Added: 21.03.2013
			if v==r"+_\I_\d\d\d\d\d_fort\.\U": continue
			# --- XXX XXX XXX ---
			if k.startswith("rule."):
				try:
					idx = int(k[5:])-1
				except:
					idx = None
				UsrInfo.addRule(v, idx)
			elif "." not in k:
				UsrInfo.USR_NAME[k] = v

	#-------------------------------------------------------------------------------
	@staticmethod
	def addRule(rule, pos=None):
		"""Add rule from configuration file"""

		pat  = "+"
		name = rule
		if len(rule)>1:
			pat  = rule[0]
			if pat in ("+","-"):
				name = rule[1:]
			else:
				pat = "+"
		rule = (pat, name)

		if pos is None:
			UsrInfo.RULES.append(rule)
		else:
			if pos >= len(UsrInfo.RULES):
				UsrInfo.RULES.extend([("","")]*(pos-len(UsrInfo.RULES)+1))
			UsrInfo.RULES[pos] = rule

	# ----------------------------------------------------------------------
	@staticmethod
	def saveConfig(config):
		pass

#===============================================================================
# Run Information Class
#===============================================================================
class RunException(Exception): pass
class RunInfo(Info):
	CLASS   = "Run"
	section = "Batch"

	DEFAULT_INPUT = "<default>"

	STATUS_NOT_RUNNING  = 0
	STATUS_WAIT2ATTACH  = 1
	STATUS_RUNNING      = 2
	STATUS_FINISHED     = 3
	STATUS_FINISHED_ERR = 4
	STATUS_FINISHED_INT = 5
	STATUS_TIMEOUT      = 6

	STATUS_STR = [					# should match the statuses before
		"Not Running",
		"Waiting to attach",
		"Running",
		"Finished OK",
		"Finished with ERRORS",
		"Finished interrupted",
		"*** TIMED-OUT ***"
	]

	statusColor = {
		STATUS_NOT_RUNNING  : "Black",
		STATUS_WAIT2ATTACH  : "Orange",
		STATUS_RUNNING      : "DarkGreen",
		STATUS_FINISHED     : "DarkBlue",
		STATUS_FINISHED_ERR : "Red",
		STATUS_FINISHED_INT : "DarkCyan",
		STATUS_TIMEOUT      : "Purple"
	}

	statusIcon = {
		STATUS_WAIT2ATTACH  : "attach",
		STATUS_RUNNING      : "run",
		STATUS_TIMEOUT      : "TIME-CUT"
	}

	DEFAULT = {
		"exe"       : "",			# override executable
		"last"      : 5,			# Last cycle
		"mode"      : "",			# default simulation mode
		"parent"    : "",			# Parent name of a family
		"pid"       : 0,			# Process id from running dir is fluka_{pid}
		"prev"      : 0,			# Previous cycle
		"primaries" : 0,			# override No of primaries
		"queue"     : _SUBMIT_DEFAULT,		# submitting queue
		"rnd"       : 0,			# override random seed
		"spawnname" : "",			# Spawn pattern name
		"startrun"  : 0,			# time when run button was pressed
		"status"    : STATUS_NOT_RUNNING,	# Last known running status
		"time"      : 0,			# override time limit
		"open"      : 0,			# open/close in treeview
		"title"     : "",			# override title of the run
		}
	timeThreshold   = 120				# more that XXs the run is considered
							# ... to be finished
	MAX_TIME     = 1.0E100				# maximum time allocated for run
	G4WTPAT      = re.compile(r"^(G4WT\d+ > )*(.*)$")
	UNTITLED     = "<untitled>"
	refreshQueue = None				# queue holding runs to be refreshed

	#-----------------------------------------------------------------------
	# XXX WARNINGS XXX
	# 1. The .parent of the Info is not the parent of the family
	#    but rather the .project (as it is set below with the setter/getter)
	#    The parent of the family is a field in the dictionary
	#-----------------------------------------------------------------------
	def __init__(self, project, name):
		super().__init__(project, name)
		self.defineList = []	# override defines
		self.family     = []	# other members of the family
		self.usrinfo    = []	# UsrInfo list
		self.init()

	# ----------------------------------------------------------------------
	# Init temporary local variables
	# ----------------------------------------------------------------------
	def init(self):
		# variables showing the running status
		self.attachTime = 0	# when the last attached was requested
		self.cycle      = 0	# current running cycle
		self.handled    = 0	# number of primaries handled up to now
		self.iniTime    = 0	# initialization time
		self.maxTime    = RunInfo.MAX_TIME	# max time allocated for the run
		self.remaining  = 0	# number of remaining particles
		self.startTime  = ""	# start run time
		self.timeperprim= 0	# time per primary during run

		self._exclude   = None	# Exclude running directories that are present
		self._fout      = None	# Output file handler
		self._lastOut   = None	# last output file opened
		self._lastPos   = 0	# last output file position scanned
		self._log       = log.say
		self._popen     = None	# process id (in case we have it: e.g. local run)
		self._prevStatus= RunInfo.STATUS_NOT_RUNNING
		self._runDir    = None	# running directory

	# ----------------------------------------------------------------------
	# Pickle information to copy
	# ----------------------------------------------------------------------
	def __getstate__(self):
		state = super().__getstate__()
		state.update({
			"defineList" : self.defineList,
			"family"     : self.family,
			"usrinfo"    : self.usrinfo,
		})
		return state

	# ----------------------------------------------------------------------
	# UnPickle information to restore
	# ----------------------------------------------------------------------
	def __setstate__(self, state):
		# Unfortunately we cannot call super().__setstate__(state)
		self.__dict__.update(state)
		self._parent = None
		for u in self.usrinfo: u.parent(self)
		self.init()

	# ----------------------------------------------------------------------
	@property
	def project(self):
		return self._parent

	# ----------------------------------------------------------------------
	@project.setter
	def project(self, x):
		#raise Exception("Cannot set project in RunInfo")
		self._parent = x

	# ----------------------------------------------------------------------
	def __str__(self):
		return f"{self.__class__.__name__}(name={self._name!r})={dict(self)}"
	__repr__ = __str__

	# ----------------------------------------------------------------------
	# @return true if run is the master (default) run
	# ----------------------------------------------------------------------
	def isdefault(self):
		return self.name == RunInfo.DEFAULT_INPUT

	# ----------------------------------------------------------------------
	# FIXME very buggy in case of similar names!
	# ----------------------------------------------------------------------
	def parentRun(self):
		"""Return parent RunInfo"""
		p = self("parent")
		if not p: return None
		for parent in self.project.runs():
			if p == parent.name:
				return parent
		return None

	# ----------------------------------------------------------------------
	def getChildOrParent(self, parent, key):
		"""Return a value from child or parent for key"""
		if self(key): return self(key)
		if parent and parent(key): return parent(key)
		return RunInfo.DEFAULT[key]

	# ----------------------------------------------------------------------
	# Write run information
	# ----------------------------------------------------------------------
	def write(self, stream):
		stream.write(f"Run: {self.name}\n")
		p = not self("parent")
		if p and self("title"): stream.write(f"\tTitle:    {self('title')}\n")
		if p:		# Write defines only when it is not a child
			for n,v in self.defineList:
				v = v.strip()
				if v!="":
					stream.write(f"\tDefine:   {n}={v}\n")
				else:
					stream.write(f"\tDefine:   {n}\n")

		# FIXME to be cleaned!!!!
		# it should go through the default write()
		if self("rnd"):         stream.write(f"\tRnd:      {self['rnd']}\n")
		if p and self("primaries")>0: stream.write(f"\tStart:    {self('primaries')}\n")
		if self("time") > 0:    stream.write(f"\tTime:     {self('time')}\n")
		if p and self("exe"):   stream.write(f"\tExe:      {self('exe')}\n")
		if self("prev"):        stream.write(f"\tPrev:     {self['prev']}\n")
		if self("last")!=5:     stream.write(f"\tLast:     {self('last')}\n")
		if self("status"):      stream.write(f"\tStatus:   {self['status']}\n")
		if self("pid"):         stream.write(f"\tPid:      {self['pid']}\n")
		if p and self("mode"):  stream.write(f"\tmode:     {self['mode']}\n")
		if self("startrun"):    stream.write(f"\tStartRun: {self('startrun')}\n")
		if self("parent"):      stream.write(f"\tParent:   {self['parent']}\n")
		if self("queue") not in ("",_SUBMIT_DEFAULT):
					stream.write(f"\tqueue:    {self['queue']}\n")
		if self("open"):        stream.write(f"\topen:     {int(self['open'])}\n")

		# if all are auto generated then don't write anything
		if not all([x.auto for x in self.usrinfo]):
			for u in self.usrinfo:
				u.write(stream)

		stream.write("End\n")

	# ----------------------------------------------------------------------
	# Read run from file. The file pointer should be exactly
	# after the Run: tag
	# ----------------------------------------------------------------------
	def read(self, stream):
		child = False
		while True:
			tag, val = stream.parseVarValue()
			if tag is None: break
			if   tag=="Title":    self["title"]     = val
			elif tag=="Rnd":      self["rnd"]       = int(val)
			elif tag=="Start":    self["primaries"] = float(val)
			elif tag=="Time":     self["time"]      = float(val)
			elif tag=="Exe":      self["exe"]       = val
			elif tag=="Prev":     self["prev"]      = int(val)
			elif tag=="Last":     self["last"]      = int(val)
			elif tag=="Status":   self["status"]    = int(val)
			elif tag=="Pid":      self["pid"]       = int(val)
			elif tag=="mode":     self["mode"]      = val
			elif tag=="StartRun": self["startrun"]  = float(val)
			elif tag=="queue":    self["queue"]     = val
			elif tag=="open":     self["open"]      = bool(int(val))
			elif tag=="Parent":   self["parent"]    = val
			elif tag=="Defines":
				self.defineList  = [(n,"") for n in val.split()]	# deprecated
			elif tag=="Define":
				e = val.find("=")
				if e>0:
					n = val[:e]
					v = val[e+1:].strip()
				else:
					n = val
					v = ""
				self.defineList.append((n,v))
			elif tag==UsrInfo.CLASS:
				usr = UsrInfo(self, val)
				usr.read(stream)
				self.usrinfo.append(usr)
			elif tag in ("End", "End"+self.CLASS):
				break
			else:
				pass
				#raise RunException("Unknown Run keyword "+tag)
		if child and not self("parent"):
			# Try to find parent run for old projects
			for parent in self.project.runs():
				if self.name in parent.family:
					self["parent"] = parent.name
					break

	# ----------------------------------------------------------------------
	def clone(self):
		c = RunInfo(self.project, self.name)
		c.copy(self)
		return c

	# ----------------------------------------------------------------------
	def copy(self, src):
		super().copy(src)
		self["title"]    = src("title")
		self["rnd"]      = src("rnd")
		self["primaries"]= src("primaries")
		self["time"]     = src("time")
		if src.isdefault():
			self["exe"] = src.project.compile.name
		else:
			self["exe"] = src("exe")
		self["prev"]     = src("prev")
		self["last"]     = src("last")
		self["pid"]      = src("pid")
		self["status"]   = RunInfo.STATUS_NOT_RUNNING
		self["startrun"] = src("startrun")

		self.init()
		self.defineList  = src.defineList[:]
		self.family      = []			# ?
		self.usrinfo     = [x.clone() for x in src.usrinfo]
		for u in self.usrinfo: u.parent(self)

	# ----------------------------------------------------------------------
	# @return full input name with extension
	# ----------------------------------------------------------------------
	def inputFullName(self):
		if self.isdefault():
			return self.project.inputName()
		else:
			return os.path.join(self.dir(), os.path.basename(self.name))

	# ----------------------------------------------------------------------
	# @return run input name without extension
	# ----------------------------------------------------------------------
	def inputName(self):
		if self.isdefault():
			return os.path.splitext(self.project.inputName())[0]
		else:
			return os.path.join(self.dir(), os.path.basename(self.name))

	# ----------------------------------------------------------------------
	# @return run input name for listboxes
	# ----------------------------------------------------------------------
	def inputName2(self):
		if not self.isdefault():			# sub run name
			return self.name
		elif self.project.filename:
			return f"<{self.project.name()}>"	# project name
		else:
			return RunInfo.UNTITLED			# no name

	# ----------------------------------------------------------------------
	def inputBaseName(self):
		return os.path.basename(self.inputName())

	# ----------------------------------------------------------------------
	def inputBaseName2(self):
		return os.path.basename(self.inputName2())

	# ----------------------------------------------------------------------
	def inputBaseFullName(self):
		return os.path.basename(self.inputFullName())

	# ----------------------------------------------------------------------
	def inputDirName(self):
		return os.path.dirname(self.inputName())

	# ----------------------------------------------------------------------
	# Get default directory
	# ----------------------------------------------------------------------
	def dir(self):
		subdir = os.path.dirname(self.name)
		if subdir:
			if subdir[0] == os.path.sep:
				return subdir
			elif self("parent"):
				parent = self.parentRun()
				if parent is not None:
					return os.path.join(parent.dir(), subdir)
			return os.path.join(self.project.dir(), subdir)

		elif self("parent"):
			parent = self.parentRun()
			if parent is not None:
				return parent.dir()

		return self.project.dir()

	# ----------------------------------------------------------------------
	# Get temporary running directory if any otherwise default
	# ----------------------------------------------------------------------
	def getRunDir(self):
		if self._runDir is not None:
			return self._runDir
		else:
			return self.dir()

	# ----------------------------------------------------------------------
	# setExecutable: always as absolute
	# ----------------------------------------------------------------------
	def setExecutable(self, fn):
		if fn:
			path = os.path.dirname(fn)
			if path != self.project.dir():
				self["exe"] = fn		# Absolute path
			else:
				self["exe"] = os.path.basename(fn)
		else:
			self["exe"] = fn

	# ----------------------------------------------------------------------
	# Reset defines list to default
	# WARNING empty list means no defines are selected apart for the
	#         default run
	# ----------------------------------------------------------------------
	def defaultDefines(self):
		self.defineList = []
		for card in self.project.input.cardsSorted("#define").values():
			if card._indent != 0: continue
			name = card.name()
			for n,w in self.defineList:
				if n==name: break
			else:
				self.defineList.append((name, card.what(1)))

	# ----------------------------------------------------------------------
	# @return defines list for preprocess(), None for the default run
	# ----------------------------------------------------------------------
	def defines(self):
		if self.isdefault():
			return None
		elif self("parent"):
			parent = self.parentRun()
			if parent is not None and parent.isdefault():
				return None
			return parent.defineList
		else:
			return self.defineList

	# ----------------------------------------------------------------------
	# Find possible directories where the run might take place
	# @return [ (dir, filename), ... ]
	# ----------------------------------------------------------------------
	def findRunningDirs(self, check_core=True):
		#Global.dump("*-* RunInfo.findRunningDirs", level=2)
		# Scan all subdirs with pattern "fluka_\d+"
		# for a file named "input\d\d\d.out"
		# check for a core.\d+ file
		found  = []
		inpname= re.escape(self.inputBaseName())
		rundir = self.getRunDir()
		if not os.path.isdir(rundir): return found
		patout = re.compile(rf"^{inpname}(\d\d\d).out$")
		patlog = re.compile(rf"^{inpname}(\d\d\d).log$")
		config = self.config()
		for d in tool.listdir(rundir):
			if config.CYCLEDIR_PAT.match(d):	# Directory found
				fullpath = os.path.join(rundir,d)
				relpath  = self.project.relpath(fullpath)
				if not os.path.isdir(fullpath): continue
				if self._exclude and d in self._exclude: continue
				# Check files in directory
				outf = None	# output file
				logf = None	# log file
				core = None	# core file
				for f in tool.listdir(fullpath):
					# ignore links
					try: s = os.lstat(os.path.join(fullpath,f))
					except: continue
					if stat.S_ISLNK(s.st_mode):
						continue
					if s.st_mtime < self("startrun"):
						continue
					if check_core and (f=="core" or config.CORE_PAT.match(f)):
						core = (relpath,None)
					elif patout.match(f):
						outf = (relpath,f)
					elif patlog.match(f):
						logf = (relpath,f)
				if outf is not None:		# Prefer out file
					found.append(outf)
				elif logf is not None:		# Otherwise log file
					found.append(logf)
				elif core and (outf or logf):	# FIXME nonsense condition .inp must be present
					found.append(core)	# core.xxx file
		return found

	# ----------------------------------------------------------------------
	# return current run mode
	# ----------------------------------------------------------------------
	def mode(self):
		try:
			mode = self["mode"]
			if mode: return mode
		except KeyError:
			pass
		if self("parent"): # Child
			parent = self.parentRun()
			if parent("mode"):
				try:
					return parent["mode"]
				except KeyError:
					pass
		return self.project.input.mode

	# ----------------------------------------------------------------------
	# return config for the current mode
	# ----------------------------------------------------------------------
	def config(self):
		return inputs.MODE2CONFIG[self.mode()]

	# ----------------------------------------------------------------------
	# Return exe information from the child parent or the master
	# ----------------------------------------------------------------------
	def exe(self):
		Global.dump("*-* RunInfo.exe", level=2)
		exe = self("exe")
		if exe:
			if exe[0] == "<":
				exe = exe[1:-1]
				if exe == self.config().DEFAULT_EXE:
					return self.config().executable()
				return self.config().command(exe)
			else:
				return exe
		elif self("parent"): # Child
			parent = self.parentRun()
			if parent is not None:
				return parent.exe()
		elif self.isdefault():
			if self.project.compile.name:
				return self.project.compile.name
			elif self.mode()==Fluka.MODE and self.config().isDefaultExecutable():
				return ""
			else:
				return self.config().executable()
		else:
			return self.config().executable()

	# ----------------------------------------------------------------------
	# Start the run and try to attach
	# ----------------------------------------------------------------------
	def start(self, cwd=None, defines=None, output=None):
		Global.dump("*-* RunInfo.start", level=2)
		if cwd is not None:
			self._runDir = cwd
		else:
			self._runDir = None
			cwd = self.dir()
		if output is None:
			self._log = log.say
		else:
			self._log = output

		# Check if directory exist
		if not os.path.isdir(cwd):
			os.makedirs(cwd)

		# Check last random number if exist
		rc = 0
		self["status"] = RunInfo.STATUS_NOT_RUNNING
		#parent = self.parentRun()
		last = self("prev") + 1
		mode = self.mode()
		if last > 1:
			inpname = self.inputBaseName()
			try:
				if mode == Fluka.MODE:
					rndfile = f"{cwd}/ran{inpname}{last:03d}"
				elif mode == Moira.MODE:
					rndfile = f"{cwd}/{inpname}{last:03d}_run.rndm"
				os.stat(rndfile)
			except OSError:
				raise RunException(
					f"Error running input '{inpname}'.\n"
					f"Random number not found.\n"
					f"'{rndfile}'")
		else:
			# Prepare input file
			errors = self.project.writeInput(self, cwd, defines)
			if errors:
				for card,err in errors:
					if card:
						self._log(f"{log.ERROR}{card}")
						prefix = log.ERRORCONT
					else:
						prefix = log.ERROR
					self._log(f"{prefix}{err}")
				rc = len(errors)
				return rc

		# Create run command
		if self.project.submit == "null": return 0
		cmd = self.runCommand(cwd)
		if not os.path.isfile(cmd[0]):
			raise RunException(f"Cannot find executable {cmd[0]}\n"
				"Please verify that the FLUKA directory\n"
				f" {FlukaConfig.directory!r}\n"
				"or the submit command are correct!")

		filename = os.path.splitext(self.inputName())[0] + ".out"
		fout = open(filename, "w", encoding="UTF-8")		# For the pipe
		fout.write(f"*-* Running: {self.inputBaseName()}\n")
		fout.write(f"Dir: {cwd}\n")
		fout.write(f"Cmd: {' '.join(list(cmd))}\n")
		self.updateOut()

		# Exclude existing directories
		self._exclude = [d for d,f in self.findRunningDirs(check_core=False)]

		# spawn in the background
		env = os.environ.copy()
		# force unbuffered stdio
		env["GFORTRAN_UNBUFFERED_PRECONNECTED"] = "y"
		if mode == Fluka.MODE:
			env["FLUKAPATH"] = self.project.dir()

		self._popen = subprocess.Popen(cmd,
					cwd=cwd,
					env=env,
					stdout=fout,
					stderr=subprocess.STDOUT,
					close_fds=True,
					preexec_fn=os.setpgrp)
					# start a new session, detached from the father

		# Change status to reflect changes
		self["status"]   = RunInfo.STATUS_WAIT2ATTACH	# Wait for the second update
		self["startrun"] = self.attachTime = time.time()
		self.iniTime     = 0
		self.maxTime     = RunInfo.MAX_TIME
		self._lastOut    = None			# last output file opened
		self._lastPos    = 0			# last output file position scanned
		return rc

	# ----------------------------------------------------------------------
	# return a tuple to spawn a fluka command
	# ----------------------------------------------------------------------
	def runCommand(self, cwd=None):
		Global.dump("*-* RunInfo.runCommand", level=2)
		queue  = self("queue")
		mode   = self.mode()
		config = self.config()
		if not queue: queue = self.project.submit
		submit = _submit_command.get(queue,"")

		# Prepare command
		cmd = []
		if submit: cmd.extend(shlex.split(submit))
		cmd.append(config.program())
		if mode in (Fluka.MODE, Moira.MODE):
			if self("exe"):
				if self("exe") == f"<{FlukaConfig.DEFAULT_EXE}>":
					pass
				elif self("exe") == f"<{FlukaConfig.DEFAULT_EXE2}>":
					cmd.append("-d")
				else:
					exe = self.exe()
					if cwd is not None and exe.find("/") < 0:
						exe = os.path.join(self.project.dir(), exe)
					cmd.append("-e")
					cmd.append(exe)
			if self("prev") > 0:
				cmd.append("-N")
				cmd.append(str(self("prev")))
			cmd.append("-M")
			cmd.append(str(self("last")))

		#elif mode == Penelope.MODE:
		#	cmd.append(config.exe())
		#elif mode == Phits.MODE:
		#	cmd.append(self.exe())

		cmd.append(self.inputBaseFullName())
		return tuple(cmd)

	# ----------------------------------------------------------------------
	# run finished, open last .out and check if everything is Ok
	#-----------------------------------------------------------------------
	def _checkFinished(self):
		#Global.dump("*-* RunInfo._checkFinished", level=2)
		self["pid"] = 0
		for cycle in range(self("last"), self("prev"),-1):
			fn = "%s/%s%03d.out" % (self.getRunDir(),
						self.inputBaseName(),
						cycle)
			try:
				if os.path.getmtime(fn) < self("startrun"):
					# Ignore this output
					#self["status"] = RunInfo.STATUS_TIMEOUT
					continue
				break
			except FileNotFoundError:
				pass
		else:
			if self("status") == RunInfo.STATUS_RUNNING:
				self["status"] = RunInfo.STATUS_FINISHED
			self.closeOut()
			return
		try:
			with open(fn, mode="rb") as fout:
				fout.seek(-500, io.SEEK_END)
				mode = self.mode()
				if mode == Fluka.MODE:
					if b"End of FLUKA" in fout.read():
						self["status"] = RunInfo.STATUS_FINISHED \
							if cycle == self("last") \
							else RunInfo.STATUS_FINISHED_INT
					else:
						self["status"] = RunInfo.STATUS_FINISHED_ERR
				elif mode == Moira.MODE:
					if b"End of Moira Run" in fout.read():
						self["status"] = RunInfo.STATUS_FINISHED \
							if cycle == self("last") \
							else RunInfo.STATUS_FINISHED_INT
					else:
						self["status"] = RunInfo.STATUS_FINISHED_ERR
		except IOError:
			self["status"] = RunInfo.STATUS_NOT_RUNNING
		self.closeOut()

	# ----------------------------------------------------------------------
	# Check if submitting process returned any error
	# Return true on returncode not zero
	# ----------------------------------------------------------------------
	def checkSubProcess(self):
		#Global.dump("*-* RunInfo.checkSubProcess", level=2)
		if self._popen is None: return False
		try:
			# It can happen that _popen terminates between the
			# instructions and _popen becomes None from the
			# checking thread
			self._popen.poll()
			if self._popen.returncode:
				self.updateOut()
				self._log(f"Error code: {self._popen.returncode}")
				self["status"] = RunInfo.STATUS_FINISHED_ERR
				self.closeOut()
				return True
		except AttributeError:
			pass
		return False

	# ----------------------------------------------------------------------
	# Read any possible output buffer and append to log
	# ----------------------------------------------------------------------
	def updateOut(self):
		if self._fout is None:
			try:
				self._fout = open(f"{self.inputBaseName()}.out", "r",
							encoding="UTF-8",
							errors="replace")
			except: pass
		if self._fout:
			s = self._fout.read()
			if s: self._log(s)

	#-----------------------------------------------------------------------
	# Close reading of output
	#-----------------------------------------------------------------------
	def closeOut(self):
		if self._fout is not None:
			self._fout.close()
		self._fout  = None
		self._popen = None

	# ----------------------------------------------------------------------
	# Set status to wait to attach to a process
	# ----------------------------------------------------------------------
	def wait2Attach(self):
		self._exclude = None
		if self.family:
			for child in self.family:
				child["status"]  = RunInfo.STATUS_WAIT2ATTACH
				child.attachTime = time.time()
				child._exclude   = None
		else:
			self["status"]  = RunInfo.STATUS_WAIT2ATTACH
			self.attachTime = time.time()

	# ----------------------------------------------------------------------
	# Attach to a process
	# ----------------------------------------------------------------------
	def attach(self):
		#Global.dump("*-* RunInfo.attach", level=2)
		if self.family: return
		if self.checkSubProcess(): return

		# Set default info
		if self.attachTime>0 and \
		   time.time() - self.attachTime > RunInfo.timeThreshold:
			self["status"] = RunInfo.STATUS_TIMEOUT
			return

		# reset information
		self["pid"]  = 0
		self.cycle   = 0
		self.handled = 0

		# Scan all subdirs with pattern "fluka_\d+"
		# for a file named "input\d\d\d.out"
		# check for a core.\d+ file
		inpname = self.inputBaseName()
		try:
			found = self.findRunningDirs()
		except OSError:
			found = []

		# Check if run is already finished
		if not found:
			try:
				fout = f"{self.getRunDir()}/{inpname}{self('last'):03d}.out"
				if os.path.getmtime(fout) >= self("startrun"):
					self._checkFinished()
			except FileNotFoundError:
				pass
			return		# Nothing found

		# Filter out all core files
		found = [ (d,f) for d,f in found if f ]
		if not found:
			self.closeOut()
			self.refresh()
			return
		elif len(found) > 1:
			return found	# Multiple choices ask the user

		# Else take the first one
		self["status"] = RunInfo.STATUS_RUNNING
		self["pid"]    = int(self.config().CYCLEDIR2_PAT.match(found[0][0]).group(2))
		self.updateOut()
		pat = re.compile(rf"^{re.escape(inpname)}(\d\d\d).(out|log)$")
		m = pat.match(found[0][1])
		if m is None:
			self["status"] = RunInfo.STATUS_FINISHED_ERR
			m = pat.match(found[0][1])
		if m is not None:
			self.cycle = int(m.group(1))
		self.attachTime = time.time()
		return found[0]

	# ----------------------------------------------------------------------
	# refresh status if running
	# @param output	if True check the running output for # events handled
	# ----------------------------------------------------------------------
	def refresh(self, output=False):
		#Global.dump("*-* RunInfo.refresh", self.inputFullName(), self('status'))
		#import traceback; traceback.print_stack()
		if self.checkSubProcess(): return
		inpname = self.inputBaseName()

		# If it is waiting to attach check only in case of timeout
		if self("status") == RunInfo.STATUS_WAIT2ATTACH:
			if self.attachTime>0 and \
			   time.time()-self.attachTime > RunInfo.timeThreshold:
				self._checkFinished()
			return
		if self("status") not in (RunInfo.STATUS_RUNNING, RunInfo.STATUS_TIMEOUT):
			return
		if self("pid")==0: return

		rundir = "%s/%s%d" % (self.getRunDir(), self.config().CYCLE, self("pid"))
		if not os.path.isdir(rundir):
			self._checkFinished()
			return

		# find if cycle still exist
		found = False
		for self.cycle in range(self.cycle, self("last")+1):
			try:
				fnout = os.path.join(rundir,f"{inpname}{self.cycle:03d}.out")
				os.stat(fnout)
				found = True
				break
			except OSError:
				pass
			try:
				os.stat(os.path.join(rundir,f"{inpname}{self.cycle:03d}.log"))
				found = True
				break
			except OSError:
				pass

		# Check if a core.### file exists
		maxtime = 0
		try:
			for fn in tool.listdir(rundir):
				try: s = os.lstat(os.path.join(rundir,fn))
				except: continue
				if stat.S_ISLNK(s.st_mode): continue
				maxtime = max(maxtime, s.st_mtime)
				if fn=="core" or FlukaConfig.CORE_PAT.match(fn):
					self["status"] = RunInfo.STATUS_FINISHED_ERR
					self["pid"]    = 0
					self.closeOut()
					return
		except OSError:
			self["status"] = RunInfo.STATUS_FINISHED
			self["pid"]    = 0
			self.closeOut()
			return

		# No output, log file and no core try to re-attach
		if not found:
			# FIXME Should have a counter, how many times to attach
			# Might end-up in looping and flashing
			#self["status"] = RunInfo.STATUS_WAIT2ATTACH
			self["pid"]    = 0
			self.cycle     = 0
			self.handled   = 0
			self.startTime = ""
			self.closeOut()
			return

		# Running
		self["status"] = RunInfo.STATUS_RUNNING
		self.updateOut()

		# Find progress
		if fnout != self._lastOut:
			# New file reset information
			self._lastOut    = fnout
			self._lastPos    = 0
			self.handled     = 0
			self.remaining   = 0
			self.timeperprim = 0.0

		mode = self.mode()
		try:
			with open(self._lastOut, mode="rb") as fout:
				if self._lastPos: fout.seek(self._lastPos)
				infoline = ""
				if mode == Fluka.MODE:
					prev = ""
					for line in fout:
						if line.startswith(b" NEXT SEEDS:"):
							infoline = prev
							self._lastPos = fout.tell()
						elif self.iniTime == 0:
							if line.startswith(b" Total time used for initialization:"):
								try: self.iniTime = float(line.split()[-2])
								except: pass
							elif line.startswith(b" Maximum cpu-time allocated for this run:"):
								try: self.maxTime = float(line.split()[-2])
								except: pass
						else:
							prev = line
				elif mode == Moira.MODE:
					for line in fout:
						if line.startswith(b"G4WT"):
							# strip multithreading headers
							pat = RunInfo.G4WTPAT.match(line.decode("ascii","replace"))
							if pat:
								line = pat.group(2).encode()

						if line.startswith(b"*-* event: "):
							infoline = line
							self._lastPos = fout.tell()
						elif not self.startTime and line.startswith(b"Moira started at: "):
							self.startTime = line[17:].strip().decode()
						elif self.iniTime==0 and line.startswith(b'Initialization Time:'):
							try: self.iniTime = float(line.split()[-2])
							except: pass
						elif line == b"***** Batch is interrupted!! *****\n" or \
						     b"*** G4Exception: Aborting execution ***" in line:
							self["status"] = RunInfo.STATUS_FINISHED_ERR
							self.closeOut()
							return
		except IOError:
			# Rest information
			self._lastOut    = None
			self._lastPos    = 0
			self.handled     = 0
			self.remaining   = 0
			self.timeperprim = 0.0
			return

		# check maximum time of all files
		if time.time() - maxtime > RunInfo.timeThreshold:
			if time.time() - maxtime > 10*RunInfo.timeThreshold:
				# Hmm it appears dead!
				self["status"] = RunInfo.STATUS_FINISHED_ERR
			else:
				self["status"] = RunInfo.STATUS_TIMEOUT
			self.closeOut()
			return

		# Parse info line
		word = infoline.split()
		if word:
			if mode == Fluka.MODE:
				try:
					self.handled     = int(word[0])
					self.remaining   = int(word[1])
					self.timeperprim = float(word[3])
				except:
					pass
				# Get the starting time from when the fort.2 file was modified
				try:
					st = os.lstat(os.path.join(rundir,"fort.1")).st_mtime
					self.startTime = time.strftime(
								_TIME_FORMAT,
								time.localtime(st))
				except:
					self.startTime = ""
			elif mode == Moira.MODE:
				try:
					handled = int(word[2])
					if handled > self.handled:
						self.handled   = handled
						total          = int(word[5].replace(")",""))
						self.remaining = total - self.handled
						# 10 real, 12 user
						self.timeperprim = float(word[10])
				except:
					pass

	# ----------------------------------------------------------------------
	# Run refresh command as a thread
	# ----------------------------------------------------------------------
	def refreshAsThread(self):
		#Global.dump("*-* RunInfo.refreshAsThread", self.inputFullName(), self("status"), level=2)
		if self("status") not in (RunInfo.STATUS_RUNNING, RunInfo.STATUS_TIMEOUT): return
		#---------------------------------------------------------------
		# Running thread refreshing runs
		#---------------------------------------------------------------
		def _refreshThread():
			#print("*+* Starting refresh thread")
			cnt = 0
			while True:
				while RunInfo.refreshQueue:
					run = RunInfo.refreshQueue.pop()
					run.refresh()
					cnt += 1
				# before exiting wait for a while if something was
				# pushed from the main thread
				time.sleep(0.1)
				if not RunInfo.refreshQueue: break
			RunInfo.refreshQueue = None
			#print(f"<<< Ending refresh thread, runs processed {cnt}")
		#---------------------------------------------------------------
		if RunInfo.refreshQueue is None:
			RunInfo.refreshQueue = set()
			start_thread = True
		else:
			start_thread = False
		RunInfo.refreshQueue.add(self)
		#print(f"+++ refreshQueue size={len(RunInfo.refreshQueue)} {threading.active_count()}")
		if start_thread:
			#print("### Start new thread", self.inputFullName(), self("status"))
			threading.Thread(target=_refreshThread).start()

	# ----------------------------------------------------------------------
	# Remove all generated files and
	# return True on error, False otherwise, and number of files removed
	# ----------------------------------------------------------------------
	def clean(self, exclude=None, output=None):
		Global.dump("*-* RunInfo.clean", level=2)
		if output is None: output = log.say
		inpname = self.inputBaseName()
		p  = re.compile(rf"^{re.escape(inpname)}\d\d\d[._].+$")
		rp = re.compile(rf"^ran{re.escape(inpname)}\d\d\d$")
		rundir = self.getRunDir()
		try:
			os.stat(rundir)
		except:
			return False, 0
		errors = False
		nfiles = 0
		cycle  = self.config().CYCLE
		try:
			fn = os.path.join(rundir, inpname+".out")
			os.remove(fn)
			output(f"Remove: {fn}")
			nfiles += 1
		except:
			pass
		for fn in tool.listdir(rundir):
			if exclude:
				if isinstance(exclude,list):
					skip = False
					for pattern in exclude:
						if fnmatch(fn,pattern):
							skip = True
							break
					if skip: continue
				elif fnmatch(fn,exclude): continue
			if p.match(fn) or rp.match(fn):
				output(f"Remove: {rundir}/{fn}")
				try:
					os.remove(os.path.join(rundir,fn))
					nfiles += 1
				except OSError as ex:
					log.error(ex)
					errors = True
			elif fn.startswith(cycle):
				dfn = os.path.normpath(os.path.join(rundir,fn))
				try: s = os.lstat(dfn)
				except OSError: continue
				if not stat.S_ISDIR(s.st_mode): continue
				found = False
				inpfile = f"{inpname}.inp"
				for fsub in tool.listdir(dfn):
					if  p.match(fsub) or \
					   rp.match(fsub) or \
					   fsub==inpfile:
						found = True
						break
				if found:
					output(f"Remove folder: {fn}")
					try:
						shutil.rmtree(dfn)
						nfiles += 1
					except OSError as ex:
						log.error(ex)
						errors = True
		self["status"] = RunInfo.STATUS_NOT_RUNNING
		self.closeOut()
		return errors, nfiles

	# ----------------------------------------------------------------------
	def stop(self, cycle=True):
		Global.dump("*-* RunInfo.stop", level=2)
		#---------------------------------------------------------------
		def _stop(run):
			# Check status, if not running start the run
			if run("status") != RunInfo.STATUS_RUNNING: return
			# Create fluka.stop or rfluka.stop files
			mode = self.mode()
			if mode == Fluka.MODE:
				stopfile = "fluka.stop" if cycle else "rfluka.stop"
			elif mode == Moira.MODE:
				stopfile = "moira.stop" if cycle else "rmoira.stop"

			try:
				f = open("%s/%s_%d/%s"
					% (run.getRunDir(), mode, run["pid"], stopfile),"w")
				f.close()
			except:
				pass

		_stop(self)
		for child in self.family:
			_stop(child)

	# ----------------------------------------------------------------------
	# Kill a run
	# ----------------------------------------------------------------------
	def kill(self):
		Global.dump("*-* RunInfo.kill", level=2)
		# Check status, if not running start the run
		self.stop()
		pid = self("pid")
		if self("status") not in (RunInfo.STATUS_RUNNING, RunInfo.STATUS_WAIT2ATTACH) or \
		   self("pid")==0:
			self["pid"] = 0
			if self("status") != RunInfo.STATUS_FINISHED_ERR:
				self["status"] = RunInfo.STATUS_FINISHED
			return

		self["pid"]    = 0
		self["status"] = RunInfo.STATUS_FINISHED_ERR

		if kill == "":
			# For Linux
			os.spawnv(os.P_WAIT, "/usr/bin/pkill",
				[ "/usr/bin/pkill", "-P", str(pid) ])
			os.kill(pid, signal.SIGKILL)
		else:
			os.spawnv(os.P_NOWAIT, kill, [ kill, str(pid) ] )

	# ----------------------------------------------------------------------
	# Scan directory to find last random number seed for a possible cont.
	# ----------------------------------------------------------------------
	def lastRun(self):
		Global.dump("*-* RunInfo.lastRun", level=2)
		mode = self.mode()
		if mode == Fluka.MODE:
			ranpat = re.compile(r"^ran%s(\d\d\d)$" % (re.escape(self.inputBaseName())))
		elif mode == Moira.MODE:
			ranpat = re.compile(r"^%s(\d\d\d)_run\.rndm$" % (re.escape(self.inputBaseName())))
		else:
			return 0
		# Scan run directory...
		last = 1
		for fn in tool.listdir(self.getRunDir()):
			m = ranpat.match(fn)
			if m:
				no = int(m.group(1))
				last = max(last, no)
		return last-1

	# ----------------------------------------------------------------------
	# expand escape characters in string
	# ----------------------------------------------------------------------
	def spawnName(self, idx, esc=False):
		name = self("spawnname") or _DEFAULT_RUNNAME
		if esc:
			name = name.replace(r"\F", re.escape(self.project.name()))
			name = name.replace(r"\I", re.escape(self.inputBaseName()))
		else:
			name = name.replace(r"\F", self.project.name())
			name = name.replace(r"\I", self.inputBaseName())

		name = name.replace(r"\0n", str(idx))
		name = name.replace(r"\02", f"{idx:02d}")
		name = name.replace(r"\03", f"{idx:03d}")
		name = name.replace(r"\04", f"{idx:04d}")
		name = name.replace(r"\n",  str(idx))
		name = name.replace(r"\2",  f"{idx:02d}")
		name = name.replace(r"\3",  f"{idx:03d}")
		name = name.replace(r"\4",  f"{idx:04d}")

		if r"\a" in name or r"\A" in name:
			hi,lo = divmod(idx-1,26)
			if hi>=26:
				# FIXME we need to add a new letter
				return name
			name = name.replace(r"\a", chr(hi+ord('a')) + chr(lo+ord('a')))
			name = name.replace(r"\A", chr(hi+ord('A')) + chr(lo+ord('A')))
		name = name.replace(r"\x", f"{idx:02X}")
		if r"\X" in name:
			name = name.replace(r"\X", bmath.int2roman(idx))
		return name

	# ----------------------------------------------------------------------
	def setRandomSeed(self, parent=None, index=None):
		if index is not None:
			# Get father and add child index
			if parent is None or parent("rnd") == 0:
				rnd = 0
				# Scan in input to find the randomize card
				for card in self.project.input["RANDOMIZ"]:
					if card.ignore(): continue
					rnd = card.intWhat(2)+index
					break
			else:
				rnd = parent("rnd")+index
			if rnd == 0:
				rnd = index
		else:
			# Random random seed
			for _ in range(1000):
				rnd = int(random.randint(3, 0x7fffffff))
				found = False
				for r in self.project.runs():
					if isinstance(r, RunInfo) and rnd == r("rnd"):
						found = True
						break
				if not found: break
		self["rnd"] = rnd

	# ----------------------------------------------------------------------
	def findUniqueUsrName(self, usr):
		Global.dump("*-* RunInfo.findUniqueUsrName", level=2)
		# Find a unique file name for usrinfo
		fn,ext = os.path.splitext(self._parent.relpath(usr.name))
		if not fn[-1].isdigit():
			name  = fn
			start = 1
		else:
			m = _LAST_NUM.match(fn)
			name = m.group(1)
			start = int(m.group(2))+1

		for i in itertools.count(start):
			namei = f"{name}{i:02}"
			for u in self.usrinfo:
				if u.name == f"{namei}{ext}":
					break
			else:
				usr.name = f"{namei}{ext}"
				break

	# ----------------------------------------------------------------------
	def __hash__(self):
		h = super().__hash__()
		for i in self.family: h ^= hash(i)
		for n,v in self.defineList: h ^= hash(n)^hash(v)
		return hash(self("title")) \
			  ^ self("rnd") ^ hash(self("primaries")) ^ hash(self("time")) \
			  ^ hash(self("exe")) ^ self("prev") \
			  ^ self("status") ^ self("pid") ^ int(self("startrun")) \
			  ^ hash(self("parent")) ^ h

	# ----------------------------------------------------------------------
	def _upgrade(self):
		# FIXME for the moment disable, but we have to enable in the future
		pass
#		self._changeTag("Title",    "title")
#		self._changeTag("Rnd",      "rnd")
#		self._changeTag("Start",    "primaries")
#		self._changeTag("Time",     "time")
#		self._changeTag("Exe",      "exe")
#		self._changeTag("Prev",     "prev")
#		self._changeTag("Last",     "last")
#		self._changeTag("Status",   "status")
#		self._changeTag("Pid",      "pid")
#		self._changeTag("StartRun", "startrun")
#		self._changeTag("Parent",   "parent")

	# ----------------------------------------------------------------------
	# Load commands
	# ----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		global _DEFAULT_RUNNAME
		_DEFAULT_RUNNAME = config.getStr(RunInfo.section, "spawnname",	_DEFAULT_RUNNAME)
		_submit_command.clear()
		for tag, value in config.items(RunInfo.section):
			_submit_command[tag] = os.path.expandvars(value)

#===============================================================================
# Plot Information Class
#===============================================================================
class PlotInfo(Info):
	CLASS = "Plot"
	DEFAULT = {
		"axes"          : "Auto",
		"axes."         : "",
		"aspect"        : 1.0,
		"basisu"        : "1 0 0",
		"basisv"        : "0 1 0",
		"bins"          : 50,
		"block."        : 0,
		"boundaries"    : 1,
		"bxmax"         : "",
		"bxmin"         : "",
		"bymax"         : "",
		"bymin"         : "",
		"bzmax"         : "",
		"bzmin"         : "",
		"case"          : "Source",
		"cbcolor"       : "",
		"cbcolors"      : 30,
		"cbcpd"         : 3,
		"cbdash"        : "",
		"cbfont"        : "",
		"cblabel"       : "",
		"cblog"         : 0,
		"cbmax"         : "",
		"cbmin"         : "",
		"cbname"        : "",
		"cboptions"     : "",
		"cbpalette"     : "fluka",
		"cbround"       : 1,
		"cbsize"        : "",
		"cbtics"        : 0,
		"cbticscolor"   : "",
		"cbticsfont"    : "",
		"cbticsoptions" : "",
		"cbticssize"    : "",
		"commands"      : "",
		"coord"         : "U-V",
		"datafile"      : "",
		"det"           : 1,
		"det."          : 0,
		"dt."           : 0,
		"emax"          : "",
		"emin"          : "",
		"errors"        : 0,
		"extends"       : "100 100",
		"file."         : "",
		"fill."         : 0,
		"format"        : ".eps",
		"from"          : 0,
		"fs."           : 1,
		"geo"           : "-Auto-",
		"geopos"        : "",
		"grid"          : "",
		"hist"          : "",
		"imgtics"       : 0,
		"imgfile"       : "",
		"index"         : 0,
		"int"           : "",
		"key."          : 1,
		"keys"          : 1,
		"kind"          : "Material",
		"labelfont"     : "",
		"labels"        : 1,
		"layer"         : "",
		"lc."           : "",
		"linetype"      : "solid",
		"ls."           : 0,
		"lt."           : 0,
		"lw."           : 1,
		"max"           : "",
		"min"           : "",
		"n"             : 0,
		"name."         : "",
		"ndetectors"    : 0,
		"norm"          : "",
		"norm."         : "",
		"nu"            : 200,
		"nv"            : 200,
		"open"          : 0,		# open/close in treeview
		"origin"        : "0 0 0",
		"particles"     : "",
		"group"         : "",		# clustering group path of current plot
		"plot"          : "",
		"proj"          : "Z",
		"ps."           : 1,
		"pt."           : 0,
		"ratio"         : "",
		"run"           : RunInfo.DEFAULT_INPUT,
		"show."         : 1,
		"smooth."       : "",
		"swap"          : 0,
		"title"         : "",
		"titlecolor"    : "",
		"titlefont"     : "",
		"titleoptions"  : "",
		"titlesize"     : "",
		"type"          : "Histogram",
		"vectscale"     : 1,
		"weight"        : 1,
		"weighted"      : 0,
		"with."         : "histerror",
		"x"             : "x",
		"x."            : "",
		"x2color"       : "",
		"x2dash"        : "",
		"x2font"        : "",
		"x2label"       : "",
		"x2log"         : 0,
		"x2max"         : "",
		"x2min"         : "",
		"x2name"        : "",
		"x2options"     : "",
		"x2size"        : "",
		"x2tics"        : 0,
		"x2ticscolor"   : "",
		"x2ticsfont"    : "",
		"x2ticsoptions" : "",
		"x2ticssize"    : "",
		"xcolor"        : "",
		"xdash"         : "",
		"xfont"         : "",
		"xlabel"        : "",
		"xlog"          : 0,
		"xmax"          : "",
		"xmin"          : "",
		"xname"         : "",
		"xnorm."        : "",
		"xoptions"      : "",
		"xrebin"        : 1,
		"xsize"         : "",
		"xtics"         : 0,
		"xticscolor"    : "",
		"xticsfont"     : "",
		"xticsoptions"  : "",
		"xticssize"     : "",
		"y"             : "y",
		"y."            : "",
		"y2color"       : "",
		"y2dash"        : "",
		"y2font"        : "",
		"y2label"       : "",
		"y2log"         : 0,
		"y2max"         : "",
		"y2min"         : "",
		"y2name"        : "",
		"y2options"     : "",
		"y2size"        : "",
		"y2tics"        : 0,
		"y2ticscolor"   : "",
		"y2ticsfont"    : "",
		"y2ticsoptions" : "",
		"y2ticssize"    : "",
		"ycolor"        : "",
		"ydash"         : "",
		"yfont"         : "",
		"ylabel"        : "",
		"ylog"          : 0,
		"ymax"          : "",
		"ymin"          : "",
		"yname"         : "",
		"yoptions"      : "",
		"yrebin"        : 1,
		"ysize"         : "",
		"ytics"         : 0,
		"yticscolor"    : "",
		"yticsfont"     : "",
		"yticsoptions"  : "",
		"yticssize"     : "",
		"zcolor"        : "",
		"zdash"         : "",
		"zfont"         : "",
		"zlabel"        : "",
		"zlog"          : 0,
		"zmax"          : "",
		"zmin"          : "",
		"zname"         : "",
		"zoptions"      : "",
		"zrebin"        : 1,
		"zsize"         : "",
		"ztics"         : 0,
		"zticscolor"    : "",
		"zticsfont"     : "",
		"zticsoptions"  : "",
		"zticssize"     : "",
	}
	# Special format handling
	FORMAT = { "commands" : "b" }

	CAMERA_KIND = "Camera:"		# kind for camera prefix

	# ----------------------------------------------------------------------
	def __init__(self, parent, name):
		super().__init__(parent, name)
		self.time = 0	# last plotting time

	# ----------------------------------------------------------------------
	def dirname(self, x=None):
		if x is None:
			return self("group")
		else:
			self["group"] = x

	# ----------------------------------------------------------------------
	def basename(self, x=None):
		if x is None:
			return self.name
		else:
			self.name = x

	# ----------------------------------------------------------------------
	def depends(self):
		lst = []
		#---------------------------------------------------------------
		def appendIf(filename):
			if filename:
				lst.append(filename)
		appendIf(self("datafile"))
		for i in range(int(self("ndetectors"))):
			try:
				appendIf(self[f"file.{i}"])
			except KeyError:
				pass
		return lst

	# ----------------------------------------------------------------------
	def provides(self, derivatives=False):
		return [self.name+".dat"]

	# ----------------------------------------------------------------------
	def _upgrade(self):
		# Change from 1->2
		if self.version() < 2:
			self._changeTag("cpd", "cbcpd")
			if self.type in ("USRBIN","Geometry"):
				# move [xyz][min|max] -> b\0
				self._changeTag("xmin","bxmin")
				self._changeTag("xmax","bxmax")
				self._changeTag("ymin","bymin")
				self._changeTag("ymax","bymax")
				self._changeTag("zmin","bzmin")
				self._changeTag("zmax","bzmax")
				self._changeTag("log", "cblog")
				self._default("cbtics", 1)
				lc = self("lc.0")
				if lc=="":
					try:
						lt = int(self("lt.0"))
						if lt<=0: lc = __gnucolors[0]
						else: lc = __gnucolors[lt%7]
						self["lc.0"] = lc
						del self["lt.0"]
					except:
						pass

			elif self.type == "USR-1D":
				for i in range(int(self("ndetectors"))):
					lc = self(f"lc.{i}")
					if lc == "":
						try:
							lt = int(self(f"lt.{i}"))
							if lt<=0: lc = __gnucolors[0]
							else: lc = __gnucolors[lt%7]
							self[f"lc.{i}"] = lc
							del self[f"lt.{i}"]
						except:
							pass

					# It should be consistent with the Plot.VALUE
					try:
						y = int(self(f"y.{i}"))
					except ValueError:
						y = 0
					if 1<=y<=4:	# Y*<X>, Y*<Xgeo>, Y*Xl, Y*Xh
						y=1	# convert to isoolethargic
					elif y==5:
						y=2	# convert to Y*DX
					else:
						y=0
					self[f"y.{i}"] = y

			elif self.type == "USR-2D":
				self._changeTag("log", "cblog")

		self._changeTag("Format", "format")

		#  old tags to delete:
		todel = ["autoaspect",
			"log",
			"ndet",
			"projtype",
			"showcenter",
			"subtype"
			]
		for key in todel:
			try:
				del self[key]
			except KeyError:
				pass

		#pw.0 prefix=pw.
		#using.0 prefix=using.

#===============================================================================
# Geometry Layers Info
#===============================================================================
class GeometryLayerInfo(Info):
	CLASS  = "Layer"
	PREFIX = "\t"
	DEFAULT = {
			"_order"            : 0,
			"_system"           : 0,

			# Options
			"Options"           : 1,
			"Options."          : "",
			"Options._visible"  : 1,
			"Options.axes"      : 0,
			"Options.errors"    : 1,
			"Options.grid"      : 1,
			"Options.ocube"     : 16,
			"Options.gridlevel" : 25,
			"Options.latticelevel": 25,
			"Options.crosshair" : 0,
			"Options.viewport"  : 1,
			"Options.font"      : "fixed8x13",
			"Options.gridfont"  : "fixed8x13",
			"Options.palettefont":"fixed8x13",
			"Options.textbackground": 0,

			# Show
			"Show"              : 1,
			"Show._visible"     : 1,
			"Show.vertex"       : 1,
			"Show.label"        : "Region",
			"Show.color"        : "Material",
			"Show.lattice"      : 0,
			"Show.voxel"        : 0,
			"Show.palette"      : "Palette",

			# Background image
			"Image"             : 0,
			"Image._visible"    : 0,
			"Image.file"        : "",
			"Image.alpha"       : 127,
			"Image.black"       : 0,
			"Image.white"       : 0xffffff,
			"Image.prompt"      : 0,
			"Image.marker"      : 0,
			"Image.marker."     : "",

			# Transformations
			"Transformations"   : 0,
			"Transformations._visible" : 0,

			# Beam
			"Beam"              : 0,
			"Beam._visible"     : 0,

			# Userdump
			"Userdump"          : 0,
			"Userdump._visible" : 0,
			"Userdump.start"    : 0,
			"Userdump.n"        : 100,
			"Userdump.file"     : "",
			"Userdump."         : 0,
			"Userdump.emin."    : "",
			"Userdump.emax."    : "",
			"Userdump.alpha."   : 0x30,

			"Userdump.color."   : "",
			"Userdump.color.4-HELIUM" : "#FF8C00",
			"Userdump.color.3-HELIUM" : "#FF1493",
			"Userdump.color.TRITON"   : "#B03060",
			"Userdump.color.DEUTERON" : "#9932CC",
			"Userdump.color.HEAVYION" : "#0000FF",
			"Userdump.color.PROTON"   : "#FF0000",
			"Userdump.color.ELECTRON" : "#00FFFF",
			"Userdump.color.POSITRON" : "#FF00FF",
			"Userdump.color.NEUTRIE"  : "#708080",
			"Userdump.color.ANEUTRIE" : "#708080",
			"Userdump.color.PHOTON"   : "#FFFF00",
			"Userdump.color.NEUTRON"  : "#00C000",
			"Userdump.color.MUON+"    : "#00FFFF",
			"Userdump.color.MUON-"    : "#00FFFF",
			"Userdump.color.PION+"    : "#8B4513",
			"Userdump.color.PION-"    : "#8B4513",
			"Userdump.color.KAON+"    : "#006400",
			"Userdump.color.KAON-"    : "#006400",
			"Userdump.color.NEUTRIM"  : "#FFC0CB",
			"Userdump.color.ANEUTRIM" : "#FFC0CB",

			# Usrbin
			"Usrbin"            : 0,
			"Usrbin._visible"   : 0,
			"Usrbin.input"      : 0,
			"Usrbin.file"       : "",
			"Usrbin.det"        : 1,
			"Usrbin.norm"       : 1.0,
			"Usrbin.xofs"       : "",
			"Usrbin.yofs"       : "",
			"Usrbin.zofs"       : "",
			"Usrbin.rotdefi"    : "",
			"Usrbin.rotdefi2"   : "",
			"Usrbin.palette"    : "Palette",
			"Usrbin.alpha"      : 0,
			"Usrbin.prompt"     : 0,
			"Usrbin.showerror"  : 0,

			# 3D
			"3D"                : 0,
			"3D._visible"       : 0,
			"3D.ambient"        : 64,
			"3D.aoc"            : 0,
			"3D.antialias"      : 1,
			#"3D.focal"          : 100.0,
			"3D.fov"            : 60,
			"3D.clip."          : "",
			"3D.clipnegative."  : 0,
			"3D.deflights"      : 1,
			"3D.edgedetect"     : 0,
			"3D.project."       : "",
			"3D.projection"     : 0,
			"3D.reflections"    : 0,
			"3D.shadows"        : 1,
			"3D.skipblack"      : 2,
			"3D.usrbinastexture": 1,
			"3D.xray"           : 0,

			# Color band
			"Palette"           : 0,
			"Palette._visible"  : 0,
			"Palette.palette"   : "fluka",
			"Palette.log"       : 1,
			"Palette.min"       : 1.0E-5,
			"Palette.max"       : 1.0E5,
			"Palette.smooth"    : 0,
			"Palette.n"         : 30,
			"Palette.cpd"       : 3,
			"Palette.inv"       : 0,
			"Palette.alphamin"  : 0,	# Transparent below minimum limit
			"Palette.alphamax"  : 0,	# Transparent above maximum limit
			"Palette.label"     : "",

			# Voxel, ROI - regions of interest in voxel
			"Voxel"             : 0,
			"Voxel._visible"    : 0,
			"Voxel.roialpha"    : 200,
			"Voxel.roi"         : "",
	}
	_ORDER = [	"Options",
			 "Show",
			 "3D",
			 "Beam",
			 "Image",
			 "Palette",
			 "Userdump",
			 "Usrbin",
			 "Voxel",
		 ]
	_MAXUSRBIN   =  20
	_MAXPALETTE  =   5
	USRBINNAMES  = [ "Usrbin" ]
	PALETTENAMES = [ "Palette" ]

	#-----------------------------------------------------------------------
	def __init__(self, parent, name, isglobal=False):
		super().__init__(parent, name)
		self._global = isglobal

	# ----------------------------------------------------------------------
	def copy(self, src):
		super().copy(src)
		self._global = src._global

	# ----------------------------------------------------------------------
	def _upgrade(self):
		# there are some garbage from the past
		todel = ["Options.title", "Show.borders", "3D.density",
			"layer", "._visible", "deflights", ""]
		obsolete = [	re.compile(r"^Usrbin[0-9]*\.inv$"),
				re.compile(r"^Usrbin[0-9]*\.log$"),
				re.compile(r"^Usrbin[0-9]*\.max$"),
				re.compile(r"^Usrbin[0-9]*\.min$"),
				re.compile(r"^Usrbin[0-9]*\.n$"),
				re.compile(r"^Palette[0-9]*\.alpha$"),
				re.compile(r"^Palette[0-9]*\.input$"),
				re.compile(r"^Palette[0-9]*\.det$"),
			]
		for key in list(self.keys()):
			value = self[key]
			if key.startswith("Colorband"):
				newkey = "Palette" + key[9:]
				self[newkey] = value
				todel.append(key)
			elif key == "Show.label" and not value:
				todel.append("Region")
			else:
				for pat in obsolete:
					if pat.match(key):
						todel.append(key)
						break
		for key in todel:
			try:
				del self[key]
			except KeyError:
				pass

	# ----------------------------------------------------------------------
	# Init defaults with dynamic variables
	# ----------------------------------------------------------------------
	@staticmethod
	def initDefaults():
		# Execute for first time
		# Insert Additional USRBINS
		pos  = GeometryLayerInfo._ORDER.index("Usrbin")+1
		keys = [ x for x in GeometryLayerInfo.DEFAULT if x.startswith('Usrbin') ]
		for i in range(GeometryLayerInfo._MAXUSRBIN,1,-1):
			GeometryLayerInfo._ORDER.insert(pos, f"Usrbin{i}")
			for k in keys:
				newkey = f"Usrbin{i}{k[6:]}"
				GeometryLayerInfo.DEFAULT[newkey] = GeometryLayerInfo.DEFAULT[k]
		for i in range(2,GeometryLayerInfo._MAXUSRBIN+1):
			GeometryLayerInfo.USRBINNAMES.append(f"Usrbin{i}")

		# Insert Additional Palette
		pos  = GeometryLayerInfo._ORDER.index("Palette")+1
		keys = [ x for x in GeometryLayerInfo.DEFAULT if x.startswith('Palette') ]
		for i in range(GeometryLayerInfo._MAXPALETTE,1,-1):
			GeometryLayerInfo._ORDER.insert(pos, f"Palette{i}")
			for k in keys:
				newkey = f"Palette{i}{k[7:]}"
				GeometryLayerInfo.DEFAULT[newkey] = GeometryLayerInfo.DEFAULT[k]
		for i in range(2,GeometryLayerInfo._MAXPALETTE+1):
			GeometryLayerInfo.PALETTENAMES.append(f"Palette{i}")
		del keys, i, pos

# Initialize additional defaults
GeometryLayerInfo.initDefaults()

#===============================================================================
# Geometry Info
# FIXME make it based on the Info class
#===============================================================================
class GeometryInfo(Info):
	CLASS = "Geometry"
	ORIENTATION = "flair"
	DEFAULT = {
			"Lock"                   : 0,
			"Synchronize"            : 0,
			"StartFrame"             : 0,
			"StopFrame"              : 250,
			"Orientation"            : ORIENTATION,

			"Frame.bsplit"           : 0.5,
			"Frame.tsplit"           : 0.5,
			"Frame.vsplit"           : 0.5,

			# Unify with the GeometryLayers.py
			"Options."               : "",
			"Options.axes"           : 1,
			"Options.crosshair"      : 0,
			"Options.grid"           : 1,
			"Options.gridlevel"      : 25,
			"Options.latticelevel"   : 25,
			"Options.title"          : 1,
			"Options.vertex"         : 1,
			"Options.errors"         : 1,
			"Options.viewport"       : 1,
			"Options.font"           : 'fixed8x13',
			"Options.gridfont"       : 'fixed8x13',
			"Options.palettefont"    : 'fixed8x13',
			"Options.textbackground" : 200,
		}

	#-----------------------------------------------------------------------
	def __init__(self, parent, name):
		super().__init__(parent, name)
		self.layers = {}
		self.layers = _globalLayers.copy()	# Make a copy of global layers

	# ----------------------------------------------------------------------
	# Pickle information to copy
	# ----------------------------------------------------------------------
	def __getstate__(self):
		state = super().__getstate__()
		state.update({
			"layers" : self.layers,
		})
		return state

	# ----------------------------------------------------------------------
	#def _writeHeader(self, stream):
	#	stream.write("\n# Geometry information\n")
	#	super()._writeHeader(stream)

	# ----------------------------------------------------------------------
	# Write geometry information
	# ----------------------------------------------------------------------
	def _customWrite(self, stream):
		for layer in self.layers.values():
			# from pickle it might not have the _global set
			if hasattr(layer,"_global") and not layer._global:
				self._writeHeader(stream)
				layer.write(stream)

	# ----------------------------------------------------------------------
	def _customRead(self, stream, tag, val):
		if tag == "Layer":
			# Parse layer
			layer = GeometryLayerInfo(self, val)
			layer.read(stream)
			self.layers[layer.name] = layer
			return True
		return False

	# ----------------------------------------------------------------------
	def __hash__(self):
		h = super().__hash__()
		for layer in self.layers.values():
			h += hash(layer)
		return h

	# ----------------------------------------------------------------------
	# Scan all layers to see if something has moved out from global
	# or the opposite
	# ----------------------------------------------------------------------
	def correctGlobalLayers(self):
		for layer in self.layers.values():
			# Layer was in global before
			if layer in _globalLayers.values():
				# if it was moved out from globals
				if not layer._global:
					# FIXME Warning if other flair are open
					# they will keep a duplicate!
					del _globalLayers[layer.name]

			# else if wasn't in global but now it has to move there
			elif layer._global:
				_globalLayers[layer.name] = layer

	# ----------------------------------------------------------------------
	def layersList(self):
		lst = [x.name for x in self.layers.values()
				if x.get("_system") and int(x.get("_order"))<2]
		lst.sort()
		l = [x.name for x in self.layers.values()
				if x.get("_system") and int(x.get("_order"))>=2]
		l.sort()
		lst.extend(l)
		l = [x.name for x in self.layers.values() if not x.get("_system")]
		l.sort()
		lst.extend(l)
		return lst

#===============================================================================
# DicomInfo
#===============================================================================
class DicomInfo(Info):
	CLASS = "Dicom"
	DEFAULT = {
		"beammodel"   : "",
		"calcdose"    : "",
		"calcdosedet" : "",
		"calcnorm"    : 1.0,
		"center"      : 0,
		"ct"          : "",
		"directory"   : "",	# different than '.' to force writing of entry
		"dvh"         : "",
		"field"       : 0,
		"frames"      : 0,
		"material."   : "",
		"palmax"      : 0,
		"palmin"      : 0,
		"palmult"     : 1.0,
		"plandose"    : "",
		"plandosedet" : "",
		"plannorm"    : 1.0,
		"preset"      : "",
		"rtdose"      : "",
		"rtstruct"    : "",
		"slice"       : 0,
		"unit2mat"    : "",
		"usrbin"      : "",
		"view"        : "Axial",
		"voxel"       : "",
		"window"      : 0,
		"xfrom"       : "",
		"xto"         : "",
		"yfrom"       : "",
		"yto"         : "",
		"zfrom"       : "",
		"zto"         : "",
		}
	windows = {}

	FileTuple = namedtuple("FileTuple", ["name","instance","sop"])

	#-----------------------------------------------------------------------
	def __init__(self, parent, name):
		super().__init__(parent, name)
		self.modality = None
		self.dicom    = None
		self.files    = None
		self.cache    = {}

	# ----------------------------------------------------------------------
	def _upgrade(self):
		# correct uid -> name
		s = self.get("uid")
		if s is not None and s != "":
			self.name = s
		if "directory" not in self:
			self["directory"] = "."

	# ----------------------------------------------------------------------
	def clone(self):
		c = super().clone()
		c.files = self.files[:]
		return c

	# ----------------------------------------------------------------------
	def clearCache(self):
		self.cache.clear()

	# ----------------------------------------------------------------------
	# return abs filename of dicominfo file. default for fist file
	# else for any file provided as string (fileidx)
	# ----------------------------------------------------------------------
	def filename(self, fileidx=0):
		if isinstance(fileidx,int):
			return os.path.abspath(os.path.join(self("directory"),
					self.files[fileidx].name))
		else:
			return os.path.abspath(os.path.join(self("directory"),fileidx))

	# ----------------------------------------------------------------------
	# Return a list of ROI material override filters
	# ----------------------------------------------------------------------
	def materialFilters(self):
		filters = []
		for i in itertools.count(1):
			try:
				matexpr = self[f"material.{i}"]
			except KeyError:
				return filters
			mat,*expr = shlex.split(matexpr)
			filters.append((mat,list(map(int,expr))))

	# ----------------------------------------------------------------------
	def __hash__(self):
		h = super().__hash__()
		if self.files:
			for t in self.files:
				for i in t:
					h += hash(i)
		return h

	# ----------------------------------------------------------------------
	# Load configuration
	# ----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		DicomInfo.windows.clear()
		for i in itertools.count(1):
			key = f"window.{i}"
			try:
				value = config.get(DicomInfo.CLASS, key)
			except:
				break
			value = value.split(":")
			try:
				name = value[0]
			except IndexError:
				continue
			try:
				level = int(value[1])
			except:
				continue
			try:
				width = int(value[2])
			except:
				continue
			DicomInfo.windows[name] = level,width

	# ----------------------------------------------------------------------
	@staticmethod
	def saveConfig(config):
		pass

#===============================================================================
# PetInfo
#===============================================================================
class PetInfo(Info):
	CLASS = "Pet"
	DEFAULT = {
		"angle"    : "",
		"axisux"   : "",
		"axisuy"   : "",
		"axisuz"   : "",
		"centrx"   : "",
		"centry"   : "",
		"centrz"   : "",
		"da"       : "",
		"dr"       : "",
		"dz"       : "",
		"ma"       : 1,
		"material" : "",
		"modules"  : 0,
		"mr"       : 1,
		"mz"       : 1,
		"na"       : 1,
		"nr"       : 1,
		"nz"       : 1,
		"radius"   : "",
		"region"   : "",
		"sa"       : "",
		"sr"       : "",
		"sz"       : "",
	}

#===============================================================================
# CompileInfo
#===============================================================================
class CompileInfo(Info):
	CLASS    = "Compile"
	LINK_PRG = [ "lfluka", "ldpmqmd" ]
	DEFAULT  = {
		"link"     : LINK_PRG[0],
		"fopts"    : "",
		"source"   : [],
	}
	FORMAT = {"source" : "l"}

	# ----------------------------------------------------------------------
	# Override name
	# ----------------------------------------------------------------------
	@property
	def name(self):
		return self._name

	# ----------------------------------------------------------------------
	@name.setter
	def name(self, filename):
		oldexe = self._name
		if filename != "":
			path = os.path.dirname(filename)
			if path != self._parent.dir():
				self._name = filename		# Absolute path
			else:
				self._name = os.path.basename(filename)
		else:
			self._name = filename

		# Modify all runs with the same exe
		for run in self._parent.runs():
			if run.isdefault(): continue
			if run("exe") == oldexe:
				run["exe"] = self._name

	# ----------------------------------------------------------------------
	def appendSource(self, filename):
		if not filename: return
		try:
			self["source"].append(filename)
		except KeyError:
			self["source"] = [filename]

	# ----------------------------------------------------------------------
	# return a tuple with the compile command to be spawned
	# ----------------------------------------------------------------------
	def compileCmd(self, filename):
		"""Return a tuple containing the compile command to be spawned"""
		fn, ext = os.path.splitext(filename)

		# FIXME should find the MODE and use the RunInfo.config() instead of
		# FlukaConfig
		if ext == ".c":
			return ( ccprogram, "-g",
				"-c",
				"-I"+FlukaConfig.includeDir(),
				f"-o{fn}.o",
				filename )

		elif ext in (".cc",".cpp"):
			return ( cppprogram, "-g",
				"-c",
				"-I"+FlukaConfig.includeDir(),
				f"-o{fn}.o",
				filename )

		else:
			# FIXME if mode is FLUKA
			cmd = [ FlukaConfig.command("fff") ]
			if self("fopts"):
				for i in self("fopts").split():
					cmd.append("-o")
					cmd.append(i)
			cmd.append(filename)
			return tuple(cmd)

	# ----------------------------------------------------------------------
	# Link command
	# ----------------------------------------------------------------------
	def linkCmd(self):
		"""Return a tuple containing the link command to be spawned"""
		cmd = [ os.path.join(FlukaConfig.binDir(), self("link")), "-o", self.name]

		dirs = []
		libs = []
		for filename in self("source"):
			fn, ext = os.path.splitext(filename)
			if ext in (".a", ".so"):
				folder = os.path.dirname(fn)
				if folder=="": folder = "."
				if folder not in dirs:
					dirs.append(folder)
				lib = os.path.basename(fn)
				if lib[0:3] == "lib":
					libs.append(lib[3:])
				else:
					libs.append(filename)

		if dirs:
			cmd.append("-L")
			cmd.append(FlukaConfig.directory)
			for d in dirs:
				cmd.append("-L")
				cmd.append(d)
			for l in libs:
				cmd.append("-b")	# Before
				cmd.append(l)
				cmd.append("-a")	# and after
				cmd.append(l)

		for filename in self("source"):
			fn, ext = os.path.splitext(filename)
			if ext not in (".a", ".so"):
				cmd.append(fn+".o")

		return tuple(cmd)

	# ----------------------------------------------------------------------
	# write a makefile to build the executable
	# ----------------------------------------------------------------------
	def writeMakefile(self, filename="Makefile"):
		"""Write a makefile to build the executable"""
		f = open(filename, "w", encoding="UTF-8")

		f.write("# Exported Makefile from project: %s\n" % (self._parent.name))
		f.write("# Title: %s\n" % (self._parent.title))

		f.write("\nF77  = %s\n" % (FlukaConfig.command("fff")))
		f.write("LINK = $(%s)/bin/%s\n\n" % (FlukaConfig.directory, self("link")))
		f.write("EXE  = %s\n" % (self.name))

		srcs = []
		dirs = []
		libs = []
		for filename in self("source"):
			fn, ext = os.path.splitext(filename)
			if ext == ".a":
				folder = os.path.dirname(fn)
				if folder: dirs.append(folder)
				lib = os.path.basename(fn)
				if lib[0:3] == "lib":
					libs.append(lib[3:])
				else:
					libs.append(filename)
			else:
				srcs.append(filename)

		f.write("SRCS = %s\n" % (" ".join(srcs)))
		f.write("OBJS = $(SRCS:.f=.o)\n\n")
		comp = self.compileCmd("")
		f.write("%%.o: %%.f\n\t$(F77) %s $<\n\n" % (" ".join(comp[1:-1])))
		f.write("$(EXE): $(OBJS)\n")

		dirstr = " ".join(["-L "+x for x in dirs])
		libstr = " ".join(["-l %s -a %s" % (x,x) for x in libs])
		f.write("\t$(LINK) -o $@ %s %s $^\n" % (dirstr, libstr))
		f.write("\nclean:\n")
		f.write("\trm -f $(EXE) $(EXE).map $(OBJS)\n")
		f.close()

#===============================================================================
# Deprecated class for assigning properties
# FIXME should be removed!!!
#===============================================================================
class Properties(Info):
	CLASS = "Properties"

	#-----------------------------------------------------------------------
	# Get properties from all cards of input
	#-----------------------------------------------------------------------
	def getFromInput(self, inp):
		self.clear()
		for card in inp.cardlist:
			for key,value in card.property.items():
				# ignore internals starting with '@', except select if not 0
				#if key[0] == '@':
				#	if not (key=="@select" and value!=0): continue
				# FIXME for the moment it only saves the @select
				#       all others are saved inside the .inp
				if not (key=="@select" and value!=0): continue
				name = f"{card.tag}/{card.name()}/{key}"
				self[name] = value

	#-----------------------------------------------------------------------
	# Set properties to all cards of input
	#-----------------------------------------------------------------------
	def setToInput(self, inp):
		for key,value in self.items():
			try:
				tag,sdum,name = key.split("/")
			except ValueError:
				continue
			for card in inp[tag]:
				if card.name()==sdum:
					card[name] = int(value)

#===============================================================================
# Material information
#===============================================================================
class MaterialInfo(Info):
	CLASS   = "Material"
	globalMaterials = {}
	# XXX order is important, is how it will appear in
	# GeometryEditor MATERIAL properties
	DEFAULT = {
		"color"     : "#707070",
		"alpha"     :  0,
		"bxdf"      :  "",
		"fuzz"      :  0.0,
		"ior"       :  1.0,
		"roughness" :  1.0,
		"shine"     : 10.0,
		"specular"  :  0.0,
	}

	#-----------------------------------------------------------------------
	def __init__(self, parent, name):
		super().__init__(parent, name)
		self["color"] = self.random()

	# ----------------------------------------------------------------------
	def random(self):
		"""Return a pseudo-random always the same color, based on the name"""
		return RGB2str(rndColor(rexx.hash(self.name)))

	# ----------------------------------------------------------------------
	def color32(self):
		try:
			return int(self["color"][1:],16)
		except:
			return 0x707070

	# ----------------------------------------------------------------------
	def default(self, key):
		"""Override the default, to return a dynamic color"""
		if key=="color": return self.random()
		return super().default(key)

	# ----------------------------------------------------------------------
	def isGlobal(self):
		return self in MaterialInfo.globalMaterials.values()

	# ----------------------------------------------------------------------
	# Check if a color is valid or not
	# ----------------------------------------------------------------------
	@staticmethod
	def valid(color):
		if len(color)!=7: return False
		if color[0]!='#': return False
		try: int(color[1:],16)
		except: return False
		return True

	# ----------------------------------------------------------------------
	# Return global material.
	# If it doesn't exist create it
	# ----------------------------------------------------------------------
	@staticmethod
	def getGlobal(name):
		mat = MaterialInfo.globalMaterials.get(name)
		if mat is None:
			MaterialInfo.globalMaterials[name] = mat = MaterialInfo(None, name)
		return mat

	# ----------------------------------------------------------------------
	# Load materials
	# ----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		for tag, value in config.items(MaterialInfo.CLASS):
			var, name = tag.split(".")
			material  = MaterialInfo.getGlobal(name)
			material[var] = value

	# ----------------------------------------------------------------------
	@staticmethod
	def saveConfig(config):
		for name,material in MaterialInfo.globalMaterials.items():
			for var in MaterialInfo.DEFAULT.keys():
				if str(material(var)) == str(material.default(var)): continue
				config.set(MaterialInfo.CLASS, f"{var}.{name}", str(material(var)))

#===============================================================================
# GroupInfo data class
#===============================================================================
@dataclass
class GroupInfo:
	name   : str  = ""
	childs : list = field(default_factory=list)

	#-----------------------------------------------------------------------
	def append(self, item):
		self.childs.append(item)

	#-----------------------------------------------------------------------
	def __hash__(self):
		return hash(self.name)

	#-----------------------------------------------------------------------
	def clone(self):
		c = self.__class__(self.name)
		return c

#===============================================================================
# Base class for Info Trees, containing a certain type of info and groups
#===============================================================================
#===============================================================================
# Isotope class
#===============================================================================
class InfoTree:
	def __init__(self, project):
		self.project = project
		self.init()

	#-----------------------------------------------------------------------
	def init(self):
		self._tree   = GroupInfo()
		self._groups = {}

	#-----------------------------------------------------------------------
	def empty(self):
		return not bool(self._tree.childs)

	# ----------------------------------------------------------------------
	# @return list of info items from tree
	# ----------------------------------------------------------------------
	def __call__(self, branch=None):
		if branch is None: branch = self._tree
		for item in branch.childs:
			if isinstance(item,GroupInfo):
				yield from self(item)
			else:
				yield item

	# ----------------------------------------------------------------------
	# @param pattern	pattern following fnmatch syntax
	#			if int is assumed as direct index
	# @return a list of parent-runs based on pattern
	# ----------------------------------------------------------------------
	def pattern(self, pattern):
		for item in self():
			if item("parent"): continue
			if fnmatch(item.name, pattern):
				yield item

	# ----------------------------------------------------------------------
	# Add info to the list and construct the tree
	# ----------------------------------------------------------------------
	def add(self, info):
		#Global.dump("*-* InfoTree.add:",(info.name, info.dirname()),level=2)
		group  = info.dirname()
		parent = self._tree
		if group:
			try:
				parent = self._groups[group]
			except KeyError:
				prevpath = ""
				group = group.split(os.sep)
				if group[0]=="":	# for absolute paths use the / infront of first dir
					group.pop(0)
					group[0] = os.sep + group[0]
				for item in group:
					path = f"{prevpath}/{item}" if prevpath else item
					try:
						parent = self._groups[path]
					except KeyError:
						self._groups[path] = branch = GroupInfo(item)
						parent.append(branch)
						parent = branch
					prevpath = path
		parent.append(info)

	# ----------------------------------------------------------------------
	# @return a unique name for a info
	# ----------------------------------------------------------------------
	def uniqueName(self, name):
		# Find a unique file name
		if not name[-1].isdigit():
			start   = 1
			newname = name
		else:
			newname = name
			m       = _LAST_NUM.match(name)
			name    = m.group(1)
			start   = int(m.group(2))+1

		for num in itertools.count(start):
			for i in self():
				if i.name == newname:
					break
			else:
				return newname
			newname = f"{name}{num:02}"

	# ----------------------------------------------------------------------
	# dump infos
	# ----------------------------------------------------------------------
	def dump(self, tree=None):
		def _dumpBranch(prefix, branch):
			if isinstance(branch,GroupInfo):
				for item in branch.childs:
					if isinstance(item,GroupInfo):
						print(f"{prefix}+ {item.name} ")
					_dumpBranch(prefix+"   ",item)
			else:
				print(f"{prefix}- {branch.name} [{branch.type}] {branch._parent}")
				if isinstance(branch,RunInfo) and branch.family:
					for child in branch.family:
						print(f"   {prefix}- {child.name} [{child.type}] {child._parent}")
		_dumpBranch("", self._tree if tree is None else tree)

	# ----------------------------------------------------------------------
	# @return a tree with ids as childs instead of info-items
	# ----------------------------------------------------------------------
	def toId(self):
		def _toid(branch):
			if isinstance(branch,GroupInfo):
				group = GroupInfo(branch.name)
				for item in branch.childs:
					group.append(_toid(item))
				return group
			else:
				return branch.id()
		return _toid(self._tree)

	# ----------------------------------------------------------------------
	# @return a tree with info-items from tree with ids
	# ----------------------------------------------------------------------
	@staticmethod
	def fromId(tree, infodict):
		def _fromid(branch):
			nonlocal infodict
			if isinstance(branch,GroupInfo):
				group = GroupInfo(branch.name)
				for item in branch.childs:
					group.append(_fromid(item))
				return group
			else:
				return infodict[branch]
		return _fromid(tree)

	# ----------------------------------------------------------------------
	# Set a new tree
	# ----------------------------------------------------------------------
	def set(self, tree):
		Global.dump("*-* InfoTree.set")
		#---------------------------------------------------------------
		def _scanBranches(path, branch):
			for item in branch.childs:
				if isinstance(item,GroupInfo):
					newpath = f"{path}/{item.name}" if path else item.name
					self._groups[newpath] = item
					_scanBranches(newpath, item)
				else:
					item.parent(self.project)
					if isinstance(item,RunInfo):
						for child in item.family:
							child.parent(self.project)
		self._tree = tree
		self._groups.clear()
		_scanBranches("",self._tree)

	# ----------------------------------------------------------------------
	# re-parent all runs
	# ----------------------------------------------------------------------
	#def reparent(self, project):
	#	def _reparent(item):
	#		if isinstance(item,GroupInfo):
	#			for item in item.childs:
	#				_reparent(item)
	#		else:
	#			item.parent(project)
	#			if isinstance(item,RunInfo):
	#				for child in item.family:
	#					child.parent(project)
	#	_reparent(runs)

	# ----------------------------------------------------------------------
	# Set a new plot tree
	# find the items that are added/deleted to return the undo information
	# and update the tree structure to reflect the one from Treeview
	# ----------------------------------------------------------------------
	def setUndo(self, tree):
		Global.dump("*-* InfoTree.setUndo")
		oldtree  = self.toId()
		olditems = list(self())
		newitems = list(self(tree))
		for item in newitems[:]:
			if item in olditems[:]:
				newitems.remove(item)
				olditems.remove(item)
		added   = {x.id():x for x in olditems}
		deleted = [x.id()   for x in newitems]
		self.set(tree)
		return (self.undoTree, oldtree, added, deleted)

	# ----------------------------------------------------------------------
	# called from undo/redo mechanism to reconstruct the previous state
	# of the tree
	# @param tree     tree with id's
	# @param added    dictionary of id:plot added new plots from the
	#                 previous tree state
	# @param deleted  list of deleted id's to become added in the undo
	# @return redo tuple
	# ----------------------------------------------------------------------
	def undoTree(self, tree, added, deleted):
		infodict = {x.id():x for x in self()}
		infodict.update(added)
		newtree = InfoTree.fromId(tree, infodict)
		newadd  = {xid:infodict[xid] for xid in deleted}
		newdel  = [xid               for xid in added.keys()]
		oldtree = self.toId()
		self.set(newtree)
		return (self.undoTree, oldtree, newadd, newdel)

#===============================================================================
# Run Tree - create a tree with all runs and groups
#===============================================================================
class RunTree(InfoTree):
	# ----------------------------------------------------------------------
	def init(self):
		super().init()
		run = RunInfo(self.project, RunInfo.DEFAULT_INPUT)
		self._tree.childs.append(run)
		# contain pointers (parent,child) -> RunInfo
		self._dict = {(RunInfo.DEFAULT_INPUT,"") : run}
		self._groups = {}

	# ----------------------------------------------------------------------
	def default(self):
		return self._tree.childs[0]

	#-----------------------------------------------------------------------
	def __setstate__(self, state):
		# add to _dict and restore parent for usrinfo
		for run in self():
			for usr in run.usrinfo:
				usr._parent = run
			if run.name[0]=="<":
				self._dict[(RunInfo.DEFAULT_INPUT,"")] = run
			else:
				self._dict[(run.name, run("parent"))] = run
		self._check()

	# ----------------------------------------------------------------------
	def _check(self):
		for run in self.runs():
			r = self.getRunByName(run.name, run("parent"))
			assert r==run
		for run in self._dict.values():
			assert run in self._runs

	# ----------------------------------------------------------------------
	# @return a generator of only runs (exclude groups)
	# ----------------------------------------------------------------------
	def __call__(self, branch=None):
		if branch is None: branch = self._tree
		for child in branch.childs:
			if isinstance(child,RunInfo):
				yield child
				for child in child.family:
					yield child
			else:
				yield from self(child)

	# ----------------------------------------------------------------------
	# Add a run and update the run-dictionary
	# @param run	run to add
	# @param pos	if None then append, else insert
	# ----------------------------------------------------------------------
	def add(self, run, pos=None):
		#Global.dump("*-* RunTree.add:",(run.name, run.dirname()),level=2)
		if pos==0 or run.name and run.name[0]=="<": return	# DEFAULT_INPUT
		assert pos is None or pos==0
		parent = run("parent")
		if parent:
			parentrun = self.getRunByName(run("parent"))
			#if parentrun is not None and run.name not in parentrun.family:
			#if parentrun is None: breakpoint()
			parentrun.family.append(run)
			return
		self._dict[(run.name, run("parent"))] = run
		return super().add(run)

	# ----------------------------------------------------------------------
	# Delete run
	# FIXME needs to be optimized to use the _dict
	# ----------------------------------------------------------------------
	def delete(self, run):
		# Check for child
		assert False
		if run("parent"):
			# find parent
			for parent in self._runs:
				found = False
				if run("parent") == parent.name:
					parent.family.remove(run.name)
					found = True
					break
				if found: break
		# Check for family
		elif run.family:
			family = run.family	# save family
			run.family = []		# set it to none
			for child in family:
				# to make it faster cut the link with the parent
				try:	# but first remove it from the dict
					del self._dict[(child.name, run.name)]
					#Global.dump("*-* RunTree.delete:",(child.name, run.name),level=2)
				except KeyError:
					pass
				child["parent"] = ""
				self.delete(child)

		# remove the run
		try:
			del self._dict[(run.name, run("parent"))]
			#Global.dump("*-* RunTree.delete:",(run.name, run("parent")),level=2)
		except KeyError:
			pass
		self._runs.remove(run)

	# ----------------------------------------------------------------------
	# Rename a run and its children
	# ----------------------------------------------------------------------
	def rename(self, run, name):
		# rename all children with the same way
		oldname = run.name
		try: del self._dict[(run.name, run("parent"))]
		except KeyError: pass

		run.name = name
		self._dict[(run.name, run("parent"))] = run

		if run.family:
			for i,child in enumerate(run.family,1):
				newname = run.spawnName(i)
				try: del self._dict[(child.name, oldname)]
				except KeyError: pass
				child.name = newname
				child["parent"] = name
				self._dict[(newname, name)] = child

	# ----------------------------------------------------------------------
	# return run by name under a certain parent if not None
	# ----------------------------------------------------------------------
	def getRunByName(self, name, parent=""):
		if name and name[0]=="<": return self.default()		# default run/input
		if parent is None:
			parent = ""
		try:
			return self._dict[(name, parent)]
		except KeyError:
			# try to correct dictionary
			for run in self():
				if run.name == name:
					if not parent:
						self._dict[(name,parent)] = run
						Global.dump(f"{log.WARNING} RunTree.getRunByName "
							f"adding non existing run '{name}'")
						return run
					elif parent == run("parent"):
						self._dict[(name,parent)] = run
						Global.dump(f"{log.WARNING} RunTree.getRunByName "
							f"adding non existing run '{parent}:{name}'")
						return run
			#Global.dump(f"{log.WARNING} getRunByName: "
			#		f"Run {parent}:{name} not Found")
			return None

	# ----------------------------------------------------------------------
	# Scan status of all runs and return True if something has changed
	# @param spawn	if spawn then spawn refresh() for all runs and check state
	#		else change change of state only
	# ----------------------------------------------------------------------
	def refresh(self, spawn=True):
		#print("\n*-* Refresh Tree")
		changed = False
		if spawn:
			for run in self():
				if run("status") == RunInfo.STATUS_WAIT2ATTACH or \
				  (run("status") == RunInfo.STATUS_RUNNING and run("pid")==0):
					if run.family:
						run["status"] = RunInfo.STATUS_NOT_RUNNING
						run.closeOut()
						changed = True
						for child in run.family:
							if child("status") in (RunInfo.STATUS_NOT_RUNNING,
									       RunInfo.STATUS_WAIT2ATTACH):
								child.attach()
							else:
								child.refreshAsThread()
					else:
						run.attach()
				elif run("status") in (RunInfo.STATUS_RUNNING, RunInfo.STATUS_TIMEOUT):
					if run.family:
						run["status"] = RunInfo.STATUS_NOT_RUNNING
						run.closeOut()
						changed = True
						for child in run.family:
							if child("status") == RunInfo.STATUS_WAIT2ATTACH or \
							  (child("status") == RunInfo.STATUS_RUNNING and child("pid")==0):
								child.attach()
							else:
								child.refreshAsThread()
					else:
						run.refreshAsThread()
				if run("status") != run._prevStatus:
					run._prevStatus = run("status")
					changed = True

		#---------------------------------------------------------------
		# set run status
		#---------------------------------------------------------------
		def setRunStatus(run, status):
			if run("status") != status:
				run["status"] = status
				return True
			return False

		#---------------------------------------------------------------
		# Update families colors
		#---------------------------------------------------------------
		for run in self():
			if not run.family: continue
			# Operate only on families
			count = defaultdict(int)
			for child in run.family:
				count[child("status")] += 1
			# Check if all child are the same
			#print(run.name)
			#if run.name == "jeff/fe56_14MeV": breakpoint()
			if len(count)==1:
				# check against last child status
				changed |= setRunStatus(run, child("status"))
			elif RunInfo.STATUS_FINISHED_ERR in count:
				changed |= setRunStatus(run, RunInfo.STATUS_FINISHED_ERR)
			elif RunInfo.STATUS_TIMEOUT in count:
				changed |= setRunStatus(run, RunInfo.STATUS_TIMEOUT)
			elif RunInfo.STATUS_RUNNING in count:
				changed |= setRunStatus(run, RunInfo.STATUS_RUNNING)
			elif RunInfo.STATUS_WAIT2ATTACH in count:
				changed |= setRunStatus(run, RunInfo.STATUS_WAIT2ATTACH)
			elif RunInfo.STATUS_FINISHED_INT in count:
				changed |= setRunStatus(run, RunInfo.STATUS_FINISHED_INT)
		return changed

	# ----------------------------------------------------------------------
	# Stop any refresh thread
	# ----------------------------------------------------------------------
	def stopRefresh(self):
		if RunInfo.refreshQueue is not None:
			RunInfo.refreshQueue.clear()

#===============================================================================
# Plot Tree - create a tree with all plots and groups
#===============================================================================
class PlotTree(InfoTree):
	...

#===============================================================================
# Project class holder
#===============================================================================
class Project:
	#-----------------------------------------------------------------------
	def __init__(self, mode=None):
		"""Initialize a fluka project"""
		self.filename = ""		# filename without the path
		self.input    = None		# input
		self.runs     = RunTree(self)
		self.plots    = PlotTree(self)
		self.init(mode)

	# ----------------------------------------------------------------------
	# initialize/clear all parameters except directory & filename
	# ----------------------------------------------------------------------
	def init(self, mode=None):
		"""Set to default all parameters of fluka project"""
		self.version    = __version__
		self.include    = {}		# input projects when inputs/includes are present
		self.master     = self		# Master flair file -//-
		self.external   = None		# input is contained in an external project
		self.newInput(mode)		# Create a new input

		# Project information
		self.title      = ""
		self.notes      = ""
		self.calculator = ""
		self.submit     = _SUBMIT_DEFAULT
		self.page       = "Flair"	# Last/active page visited
		self.modified(False)

		# Compile list FIXME maybe a list
		self.compile    = CompileInfo(self, "")
		self.runs.init()
		self.plots.init()
		self.dicoms     = []
		self.materials  = {}
		self.geometry   = GeometryInfo(self, "Geometry")
		self.pet        = PetInfo(self, "pet")
		self.var        = {}		# other variables

	# ----------------------------------------------------------------------
	# Pickle information to copy
	# ----------------------------------------------------------------------
	def __getstate__(self):
		# return the basic information without the unpicklable entries
		state = {
			"version"    : self.version,
			"filename"   : self.filename,
			"input"      : self.input,
			"title"      : self.title,
			"notes"      : self.notes,
			"calculator" : self.calculator,
			"submit"     : self.submit,
			"page"       : self.page,
			"compile"    : self.compile,
			"runs"       : self.runs,	# FIXME establish UsrInfo parent-ship!!!
			"plots"      : self.plots,
			"dicoms"     : self.dicoms,
			"materials"  : self.materials,
			"geometry"   : self.geometry,
			"pet"        : self.pet,
			"var"        : self.var,
		}
		return state

	# ----------------------------------------------------------------------
	# UnPickle information to restore
	# ----------------------------------------------------------------------
	def __setstate__(self, state):
		self.__init__()
		self.__dict__.update(state)
		# Restore parents
		for info in itertools.chain([self.geometry, self.compile, self.pet], self.runs,
					self.plots, self.dicoms, self.materials.values()):
			info._parent = self
		# include correction?
		#self.include = {}

	# ----------------------------------------------------------------------
	def __getitem__(self, key):
		return self.var.get(key)

	# ----------------------------------------------------------------------
	def __setitem__(self, key, value):
		self.var[key] = value

	# ----------------------------------------------------------------------
	def __hash__(self):
		h = hash(self.filename) \
		  ^ hash(self.title) \
		  ^ hash(self.notes) \
		  ^ hash(self.calculator)
		return h

	# ----------------------------------------------------------------------
	# Set a new input to the project
	# ----------------------------------------------------------------------
	def newInput(self, mode=None):
		if mode is None:
			mode = self.mode()
			if mode not in inputs.MODE2INPUT:
				mode = Fluka.MODE
		self.input = inputs.MODE2INPUT[mode]()
		return self.input

	# ----------------------------------------------------------------------
	def mode(self):
		global default_mode
		if self.input:
			return self.input.MODE
		else:
			return default_mode

	# ----------------------------------------------------------------------
	# Check if current input is of mode class type e.g. Fluka, Mcnp...
	# ----------------------------------------------------------------------
	def ismode(self, *modes):
		return isinstance(self.input, modes)

	#-----------------------------------------------------------------------
	# @return directory of project
	#-----------------------------------------------------------------------
	def dir(self):
		if self.filename == "":
			try:
				return os.getcwd()
			except FileNotFoundError:
				return ""
		else:
			return os.path.dirname(self.filename)

	# ----------------------------------------------------------------------
	# @return name of project (without extension)
	# ----------------------------------------------------------------------
	def name(self):
		return os.path.splitext(os.path.basename(self.filename))[0]

	# ----------------------------------------------------------------------
	# return input name from flair project with extension
	#	geometry.flair -> filename.inp  or filename.moira
	#	geometry.bodies.flair -> geometry.bodies
	# ----------------------------------------------------------------------
	def inputName(self):
		if self.external:
			name = self.external.filename
		else:
			name = self.filename
		if not name: return ""

		name = os.path.splitext(name)[0]
		if "." in name and self.master is not self:
			# return file as is
			return name
		else:
			# FIXME dirty hack for master input file always + .inp
			return name + self.input.EXT()

	# ----------------------------------------------------------------------
	# @return complete path-filename
	# ----------------------------------------------------------------------
	def abspath(self, filename=None):
		if not filename:
			if self.filename:
				return self.filename
			#elif self.external: FIXME
			else:
				return ""
		elif filename[0] == os.path.sep:
			return filename
		else:
			return os.path.abspath(os.path.join(self.dir(), filename))

	# ----------------------------------------------------------------------
	# Change the directory of the project to the new "path"
	# Warning: on the self.filename before the call
	#          the current directory self.dir() depends on the self.filename
	# ----------------------------------------------------------------------
	def chdir(self, path):
		Global.dump("*-* chdir",path)
		curdir = self.dir()		# holds the current/old path
		if path == curdir: return
		curdir = os.path.abspath(curdir)
		try:
			path = os.path.abspath(path)
		except OSError:
			log.error("No such directory",path)
			return

		#---------------------------------------------------------------
		# Make relative filename path
		#---------------------------------------------------------------
		def relative(filename):
			nonlocal path, curdir
			if filename and filename[0]!=os.sep:
				return tool.relativePath(os.path.join(curdir, filename), path)
				#return os.path.relpath(os.path.join(curdir, filename), path)
			else:
				return filename

		# FIXME input.filename
		# Fix all paths to become relative to the new path
		Global.dump("+++ FIXME Project.chdir() includes when changing directory",level=2)
		if self.include:
			include = {}
			for name, proj in self.include.items():
				include[relative(name)] = proj
			self.include = include

		# Update the dicom directory paths
		for dicom in self.dicoms:
			for name in ("directory", "materials", "unit2mat", "voxel", "usrbin"):
				if dicom(name):
					dicom[name] = relative(dicom(name))

		# FIXME other paths to be converted...
		Global.dump("*-* Project.chdir:", path,level=2)
		os.chdir(path)

	# ----------------------------------------------------------------------
	# Make relative or smallest representation of a filename path
	# ----------------------------------------------------------------------
	def relpath(self, filename, run=None):
		"""Convert path to relative to the project's directory"""
		if filename=="":
			return filename
		elif run is None:
			directory = self.dir()
		else:
			directory = os.path.abspath(run.dir())
			if directory == "" and run("parent"):
				parent = run.parentRun()
				directory = os.path.abspath(parent.dir())
				if directory == "":
					directory = self.dir()

		return tool.relativePath(filename, directory)
		#return os.path.relpath(os.path.abspath(filename), directory)

	# ----------------------------------------------------------------------
	# load project
	# ----------------------------------------------------------------------
	def load(self, filename=None, master=None):
		"""Load a project filename and change directory to it.
		All files are referred relative to this directory"""
		Global.dump("*-* Project.load", filename, "master=",master, bold=True)

		self.version  = 0	# set to unknown

		# check if external and remember the name
		if not self.filename and self.external:
			filename = self.external.filename

		self.init()
		if filename is not None:
			self.filename = os.path.abspath(filename)

		if not os.path.isfile(self.abspath()):
			if self.inputName():
				# in case that the user opened only an inputfile
				# refresh the input
				self.loadInput(self.inputName())
				if master is None: self._fillIncludes()
			return

		if master is None:
			# FIXME Change to projects directory
			Global.dump("***    Project.load.chdir:", os.path.dirname(self.filename),level=2)
			os.chdir(os.path.dirname(self.filename))
		else:
			self.master = master

		mode = Fluka.MODE	# FIXME for compatibility

		# Open project filename
		stream = Stream(self.abspath(), "r")
		md5 = None
		while True:
			tag, val = stream.parseVarValue()
			if tag is None: break
			if   tag=="Version":
				try:    self.version = int(float(val))
				except: self.version = 2
				if self.version<3:
					mode = "fluka"
				else:
					mode = default_mode
			elif tag=="Title":	self.title = val
			elif tag=="md5":	md5  = val
			elif tag=="Mode":
				if val in Card.classes:
					mode = val
				else:
					log.error(f"Project.load: invalid simulation mode '{val}'")
			elif tag=="Input":
				if val == self.name():
					# Infinite loop
					log.error("Project.load: infinite loop, "
						"input is the same as the project file")
				elif val:
					self.loadInput(val, self.master, mode)
				else:
					try:
						self._loadInputFromProject(stream, mode, md5)
					except InputMD5Exception:
						if self.master is not self: raise
						log.warning(f"Loading input from: {self.inputName()!r}")
						self.newInput()
						self.input.load(self.inputName())
					self.input.refresh()
				if master is None:
					self.input.renumber()
					self.input.preprocess()
					self.input.validate()

			# FIXME move executable to the first RunInfo
			elif tag=="Exec":	self.compile.name = val
			elif tag=="Submit":	self.submit = val
			elif tag=="Notes":	self.notes  = stream.readBlock()
			elif tag=="Page":	self.page   = val
			elif tag=="Calculator":	self.calculator = stream.readBlock()
			# Deprecated compile information
			elif tag=="Main":	pass
			elif tag=="DefMain":	pass
			elif tag=="Source":	self.compile.appendSource(val)
			elif tag=="LinkPrg":	self.compile["link"]   = val
			elif tag in ("fopts","F77opts"): self.compile["fopts"]  = val
			# ---
			elif tag==CompileInfo.CLASS:
				self.compile.name = val
				self.compile.read(stream)
			elif tag==RunInfo.CLASS:
				run = self.runs.default() if val == RunInfo.DEFAULT_INPUT else RunInfo(self, val)
				run.read(stream)
				self.runs.add(run)
			elif tag==MaterialInfo.CLASS:
				material = self.materials.get(val)
				if material is None:
					self.materials[val] = material = MaterialInfo(self, val)
				material.read(stream)
			elif tag==GeometryInfo.CLASS:
				self.geometry.read(stream)
			elif tag==Properties.CLASS:
				# Obsolete feature
				properties = Properties(self, "Input")
				properties.read(stream)
				properties.setToInput(self.input)
			elif tag==DicomInfo.CLASS:
				dicom = DicomInfo(self, val)
				dicom.read(stream)
				self.dicoms.append(dicom)
			elif tag==PetInfo.CLASS:
				self.pet.read(stream)
			elif tag==PlotInfo.CLASS:
				plot = PlotInfo(self, val)
				plot.read(stream)
				self.plots.add(plot)
			elif tag.startswith("material."):
				# Add in dictionary v.0 correction
				name = tag[9:]
				self.materials[name] = material = MaterialInfo(self, name)
			elif val != "":
				self.var[tag] = val
			elif tag != "End":
				self.var[tag] = stream.readBlock("\t")

		stream.close()
		if self.master is self:
			self._fillIncludes()
			Global.dump("+++    Included projects:", self.include,level=2)
			Global.dump("+++    cwd:", os.getcwd(),level=2)

	#-----------------------------------------------------------------------
	# Load input from the project file
	#-----------------------------------------------------------------------
	def _loadInputFromProject(self, stream, mode, md5):
		Global.dump("*-* _loadInputFromProject, mode=",mode,"md5=",md5)
		# Load input from project
		_saveComment = stream.commentChar
		self.newInput(mode)
		inputname = self.inputName()
		Global.dump(">V>    input file: ",inputname,level=2)
		stream.commentChar = "*"
		while True:
			card = Card.read(stream, mode)
			if card is None: break
			self.input.add(card, renumber=False)
			if card.tag == "#include":
				if card.enable and card.name():
					# FIXME relative to master or relative to self?
					# @see Input.open()
					# Use self.abspath() or self.master.abspath()
					# FIXME check if include is already loaded and skip!!!
					if Global.experimental:
						# Relative to self
						filename = self.abspath(card.name())
					else:
						# Relative to master
						filename = self.master.abspath(card.name())
					try:
						project = self._loadInputOrProject(filename,
									self.master, mode)
					except InputMD5Exception as ex:
						# skip remaining cards until EndOfInput
						while Card.read(stream,mode): pass
						stream.commentChar = _saveComment
						raise ex	# chain the exception
					self.master.include[self.abspath(project.name())] = project
					self.input.extend(project.input)
					project.input.delAll()
				enable = card.enable
				card = Card("#endinclude")
				card.setEnable(enable)
				self.input.add(card, renumber=False)
		stream.commentChar = _saveComment

		# Check external input and load it if needed
		if md5 and tool.isnewer(inputname, self.abspath()):
			Global.dump(f"+++    Input {inputname} is newer")
			inpmd5 = hashlib.md5()
			with open(inputname, "r", encoding="UTF-8", errors="replace") as f:
				inpmd5.update(f.read().encode())
			#Global.dump(">V>    Flair  md5=", md5,level=2)
			#Global.dump(">V>    Input  md5=", inpmd5.hexdigest(),level=2)
			if inpmd5.hexdigest() != md5:
				log.warning(f"Input file: {inputname!r}")
				log.warningCont("was modified outside flair")
				log.warningCont(f"Loading from: '{inputname}'")
				log.warningCont(f"Instead from: '{self.filename}'")
				log.warningCont("Risk of losing functionality")
				log.warningCont("The recommended way is to edit the '.flair' "
						"instead of the '.inp' files")
				raise InputMD5Exception(inputname)
		self.modified(False)

	# ----------------------------------------------------------------------
	# Load input file
	# ----------------------------------------------------------------------
	def loadInput(self, filename, master=None, mode=None, onlyinput=False):
		"""Load input file"""
		Global.dump("*-* Project.loadInput", filename, bold=True)
		# Create a new input class
		self.modified(False)
		if master is None: master = self
		# FIXME input.filename
		#if filename is not None and master is self:
		#	os.chdir(os.path.dirname(os.path.abspath(filename)))
		filename = self.abspath(filename)
		fn, ext = os.path.splitext(filename)
		# no ext set the default one
		if not ext and not os.path.isfile(filename):
			fn = filename
			filename += EXT

		# Maybe .flair or .inp...
		try:
			project = self._loadInputOrProject(filename, master, mode, onlyinput)
		except InputMD5Exception:
			if master is not self: raise	# chain the exception
			if filename == self.filename:
				self.newInput()
				self.input.load(fn+self.input.EXT())
			else:
				# Find project if any
				try:
					project = self.master.include[fn]
				except KeyError:
					project = Project(self.mode())
					project.filename = fn+EXT
					project.master   = self.master
				project.input.load(fn+project.input.EXT())

		# FIXME input.filename check the condition
		if master is self and project is not self:
			Global.dump(">V>    self=",self,level=2)
			Global.dump(">V>    self.filename=",self.filename,level=2)
			Global.dump(">V>    project.filename=",project.filename,level=2)
			Global.dump(f">+>    master:{master.filename!r}"
				f".external=project:{project.filename!r}", bold=True,level=2)
			self.include[os.path.splitext(project.filename)[0]] = project
			self.external = project
			self.input    = project.input
			Global.dump(">V>    includes=",self.include,level=2)

		# Post loading input specific action
		if master is self:
			self._postLoadInput()

	#-----------------------------------------------------------------------
	# After loading the master input do some processing
	#-----------------------------------------------------------------------
	def _postLoadInput(self):
		self._fillIncludes()
		if isinstance(self.input, Fluka):
			if not self.title:	# Update title if needed
				try: self.title = self.input.cards["TITLE"][0].text().strip()
				except KeyError: pass
			# Check expressions if something has changed
			if not self.input.checkExpr():
				log.error("Some expressions are not consistent with the input.")
				log.errorCont("Please save the input file from flair.")
				self.modified()
		elif isinstance(self.input, Moira):
			if not self.title:
				try: self.title = self.input.cards["/moira/title"][0].text().strip()
				except KeyError: pass
		elif isinstance(self.input, Phits):
			if not self.title:
				try: self.title = self.input.cards["[Title]"][0].text().strip()
				except KeyError: pass
		elif isinstance(self.input, Mcnp):
			if not self.title:
				try: self.title = self.input.cards["TITLE"][0].what(0).strip()
				except KeyError: pass

	#-----------------------------------------------------------------------
	# Make one project for each include
	# FIXME input.filename, move after _loadInputOrProject
	#-----------------------------------------------------------------------
	def _fillIncludes(self):
		Global.dump("*-* _fillIncludes", bold=True)
		#for card in self.input["#include"]:
		#	filename = card.name()
		#	fn,ext = os.path.splitext(filename)
		#	if ext not in inputs.ALL_EXTS:
		#		fn += ext
		#		ext = ""
		#	if fn not in self.include:
		#		project = Project(self.input.mode)
		#		project.filename = fn + EXT
		#		self.include[self.abspath(fn)] = project
		Global.dump(">>>    project.filename=", self.filename,level=2)
		if self.external:
			Global.dump(">>>    external.filename=", self.external.filename,level=2)
		else:
			Global.dump(">>>    external=          None",level=2)
		Global.dump(">>>    includes:", self.include.keys(),level=2)

	# ----------------------------------------------------------------------
	# Load input file or project that it is associated with
	# Warning: expects an absolute filename
	# ----------------------------------------------------------------------
	def _loadInputOrProject(self, filename, master, mode, onlyinput=False):
		Global.dump("*-* Project._loadInputOrProject", filename, bold=True)
		fn,ext = os.path.splitext(filename)
		if mode is None:
			try:
				mode = inputs.EXT2MODE[ext]
			except KeyError:
				fn  = filename		# Non standard extension
				ext = ""
		if ext and ext != EXT and ext not in inputs.ALL_EXTS:	# Unknown extension
			fn  = filename
			ext = ""
		try:
			# Recursive loading or same flair?
			if fn == self.abspath(self.name()):
				Global.dump(">V>    fn=",fn,"self.name()=", self.name(),
						self.filename,level=2)
				self.newInput(mode)
				filename = fn + self.input.EXT()	# load the input only
				Global.dump("+++    loading input1",filename,
						mode, bold=True,level=2)
				self.input.load(filename)
				return self

			# load a project file
			elif os.path.isfile(fn+EXT) and not onlyinput:
				Global.dump(f"+++    loading project {fn}{EXT}",
						bold=True,level=2)
				project = Project(mode)
				project.load(fn+EXT, master)
				return project

			# load an input as a project file
			else:
				project = Project(mode)
				project.master = master
				project.filename = fn + EXT
				if not ext:
					if os.path.isfile(fn):
						filename = fn
					elif ext == EXT:
						filename = fn + project.input.EXT()
				if os.path.isfile(filename):
					Global.dump("+++    loading input2",filename,
							mode, bold=True,level=2)
					project.input.load(filename)
				else:
					ext = inputs.MODE2INPUT[mode].EXT()
					if os.path.isfile(fn + ext):
						project.input.load(fn+ext)
						log.warning(f"Opening file '{fn}{ext}'\n"
							f"instead of '{filename}'")
					else:
						log.error(f"Cannot open file '{filename}'")
				return project
		except (IOError, OSError) as ex:
			log.error(ex)
			return None

	# ----------------------------------------------------------------------
	# save project to file
	# @param filename	if None then save on the current filename
	#			if specified save project under this new filename
	# @param cardlist	if None then save the whole input file
	#			if specified save those cards to file.
	#			Used to save #include portions, since currently
	#			we are handling the whole input file as one list (flat)
	#			FIXME in the future when it will split into many
	#			sub-projects I can remove this option
	# ----------------------------------------------------------------------
	def save(self, filename=None, cardlist=None):
		"""Save a fluka project file. If a filename is specified
		it will be saved under this name.
		WARNING it will also change the current directory"""
		Global.dump("\n*-* Project.save",filename, bold=True)
		if filename is not None:
			# Warning also change of directory!!!
			filename = os.path.abspath(filename)
			if self.filename != filename:
				if self.external and self.external.filename==filename:
					# he asked to override the external with self
					self.external = None
				self.chdir(os.path.realpath(os.path.dirname(filename)))
				self.filename = filename

		Global.dump(">>>    Project.save",self.filename,level=2)
		if keepBackup: tool.backup(self.filename)
		try:
			stream = Stream(self.filename, "w")
		except Exception as ex:
			log.error(f"Cannot open file '{self.filename}' for writing")
			log.errorCont(ex)
			stream = Stream("/dev/null", "w")
		stream.write("# flair project file\n")
		stream.write(f"Version: {__version__}\n")
		stream.write(f"Mode: {self.input.MODE}\n")
		if self.title:
			stream.write(f"Title: {self.title}\n")

		# Save input
		if self.external is None:
			# Save to same file
			Global.dump(">>>    Internally to:", self.inputName(),level=2)
			md5 = ""
			if cardlist is None:
				try:
					fn = self.saveInput()
					try:
						md5 = self.input.md5(fn)
					except KeyError as ex:
						if Global.developer:
							log.error(ex)
							log.errorCont("fn=",fn)
							log.errorCont("name=",self.name())
				except Exception as ex:
					log.error(ex)
					log.exception()
					fn = None

				# Make a copy of the cardlist
				cardlist = self.input.cardlist[:]
			else:
				try:
					md5 = self.master.input.md5(self.inputName())
				except KeyError:
					md5 = ""
			# Internally into project
			if md5: stream.write(f"md5: {md5}\n")
			self._saveInputToProject(stream, cardlist)
		else:
			# Externally to another file
			Global.dump(">>>    Externally to:", self.external.filename,level=2)
			stream.write(f"Input: {self.relpath(self.external.filename)}\n")

			fn,ext = os.path.splitext(self.inputName())
			if ext not in inputs.ALL_EXTS:
				fn += ext

			# Find project if any
			try:
				project = self.include[fn]
			except KeyError:
				project = Project(self.mode())
				self.include[fn] = project
				project.filename = fn+EXT
				project.master   = self.master
			# ... and save it
			assert project is not self
			if project is not self:
				project.save(cardlist=cardlist)

		if self.notes:
			stream.write("Notes:\n")
			stream.writeBlock(self.notes.rstrip(), "\t|")
		if self.submit != _SUBMIT_DEFAULT:
			stream.write(f"Submit: {self.submit}\n")
		if self.page:	stream.write(f"Page: {self.page}\n")
		if self.calculator:
			stream.write("Calculator:\n")
			stream.writeBlock(self.calculator.rstrip(), "\t|")
		for k in sorted(self.var):
			v = str(self.var[k])
			if v: stream.write(f"{k}: {v}\n")

		if self.compile.name:
			self.compile.write(stream)

		stream.write("\n# Run information\n")
		for run in self.runs():
			run.write(stream)

		if self.materials:
			for material in self.materials.values():
				material.write(stream)

		self.geometry.write(stream)

		if self.dicoms:
			stream.write("\n# Dicom information\n")
			for dicom in self.dicoms:
				dicom.write(stream)

		self.pet.write(stream)

		if not self.plots.empty():
			stream.write("\n# Plots information\n")
			for plot in self.plots():
				plot.write(stream)
		stream.close()
		self.modified(False)

	# ----------------------------------------------------------------------
	# Save input file
	# ----------------------------------------------------------------------
	def saveInput(self, filename=None):
		"""Save input file"""
		Global.dump("*-* Project.saveInput", filename, bold=True)
		# Check if modified or it doesn't exist then save the file
		#if self.isModified() or not os.path.exists(inputname+self.input.EXT()):
		#if  not os.path.exists(filename+self.input.EXT()):
		if self.mode() == Fluka.MODE:
			kwargs = {"expr" : FlukaCard.EXPR_OLD}
		else:
			kwargs = {}
		if filename is None: filename = self.inputName()
		Global.dump("+++    Saving input", filename,level=2)
		self.input.save(filename, backup=keepBackup, **kwargs)
		return filename

	#-----------------------------------------------------------------------
	# Save input from cardlist inside the project file
	# Warning: cardlist should be a clone since it is consumable
	#-----------------------------------------------------------------------
	def _saveInputToProject(self, stream, cardlist):
		Global.dump("*-* Project._saveInputToProject", stream.name())
		_saveComment = stream.commentChar
		stream.commentChar = "*"
		stream.write("Input:\n")
		if cardlist is None: cardlist = self.input.cardlist
		mode = self.mode()
		while cardlist:
			card = cardlist.pop(0)
			if card.tag == "#endinclude":
				if card.enable: break	# close file and exit
				continue		# skip endinclude
			# write card to stream
			card.write(stream, mode)
			if card.tag == "#include" and card.enable:
				# FIXME relative to master or relative to self?
				# @see Input.open()
				# self.abspath() or self.master.abspath()
				filename = self.master.abspath(card.name())
				fn,ext = os.path.splitext(filename)
				if not ext or ext in (EXT, self.input.EXT()):
					filename = fn + EXT
				else:
					fn = filename
					filename += EXT
				Global.dump(">V>    new-filename=",filename,level=2)
				try:
					project = self.master.include[fn]
				except KeyError:
					self.include[fn] = project = Project(self.mode())
					project.filename = filename
					project.master   = self.master
				Global.dump(">V>    project-mode:", project.mode(),
						"self.mode=",self.mode(),level=2)
				# the #endinclude will stop the writing
				project.save(cardlist=cardlist)

		stream.write("EndInput\n\n")
		stream.commentChar = _saveComment
		self.inputModified()
		self.modified(False)

	# ----------------------------------------------------------------------
	# Set project modified flag
	# ----------------------------------------------------------------------
	def modified(self, m=True):
		"""set last time modified, else None"""
		self._modified = m
		self._modifiedTime = time.time()
		for project in self.include.values():
			if project is not self:
				project.modified()

	# ----------------------------------------------------------------------
	# FIXME to be revised, since now there is an input as master
	# that contains everything and sub-projects.... :(
	# ----------------------------------------------------------------------
	def inputModified(self):
		if self.external is None:
			self.input.modified()
		for project in self.include.values():
			if project is not self:
				project.inputModified()

	# ----------------------------------------------------------------------
	# Return True if project is modified
	# ----------------------------------------------------------------------
	def isModified(self):
		"""Return TRUE if project is modified"""
		return self._modified

	# ----------------------------------------------------------------------
	# Check if a new project file exists
	# or any of the include is modified
	# ----------------------------------------------------------------------
	def checkFileTimestamp(self, modtime=None):
		#Global.dump("*-* checkFileTimestamp", self.name())
		if modtime is None: modtime = self._modifiedTime
		if self.filename:
			try:
				if os.path.getmtime(self.abspath()) > modtime:
					#Global.dump(">f>    Checking:",self.abspath(), "CHANGED")
					return True
			except FileNotFoundError:
				pass
			#Global.dump(">f>    Checking:",self.abspath(),level=2)

		#Global.dump(">>> includes=", [x.filename for x in self.include.values()])
		for project in self.include.values():
			if project is not self:
				if project.checkFileTimestamp(modtime):
					return True

		if self.external is None:
			rc = self.input.checkFileTimestamp(self.inputName())
			#Global.dump(">i>    Checking:", self.inputName(),
			#		"CHANGED" if rc else "",level=2)
			if rc: return rc
		return False

	# ----------------------------------------------------------------------
	# Run Input file
	# Write the input file with the info in Run
	# FIXME move to RunInfo
	# ----------------------------------------------------------------------
	def writeInput(self, run=None, cwd=None, defines=None):
		Global.dump("*-* Project.writeInput",
				"<default>" if run is None else run.name)
		# If default input
		parent = None
		if run is None or run.isdefault():
			# if we are writing the default input in the default directory
			# ---> use the standard writing routine
			if (cwd is None or cwd==self.dir()) and defines is None:
				# Default input save&exit
				try:
					self.saveInput()
				except Exception as ex:
					log.error(ex)
					return str(ex)
				return
			else:
				# else create a clone of the input
				run = self.runs.default()
		else:
			# Create a clone and override parameters if necessary
			if run("parent"):
				parent = run.parentRun()
				if parent is not None and parent.isdefault():
					defines = None
				elif defines is None:
					defines = parent.defineList
			elif defines is None:
				defines = run.defineList

		# Special run override is needed
		runinput = convert.input(self.input, run("mode"), defines)

		# ------------------------
		# Create a new FLUKA Input
		# ------------------------
		if isinstance(runinput, Fluka):
			# --- title ---
			title = run.getChildOrParent(parent,"title")
			if title:
				# Modify title cards
				for card in runinput["TITLE"]:
					card.setText(title)

			# --- random number seed ---
			rnd = run.getChildOrParent(parent,"rnd")
			if rnd:
				# Modify title randomize cards
				for card in runinput["RANDOMIZ"]:
					card.setWhat(2, rnd)

			# --- primaries/time ---
			primaries = run.getChildOrParent(parent,"primaries")
			timerun   = run.getChildOrParent(parent,"time")
			if primaries or timerun:
				# Modify title start cards
				for card in runinput["START"]:
					if primaries:
						card.setWhat(1, primaries)
					if timerun:
						card.setWhat(6, timerun)

		# ------------------------
		# Create a new Moira Input
		# ------------------------
		elif isinstance(runinput, Moira):
			## --- title ---
			#title = run.getChildOrParent(parent,"title")
			#if title:
			#	# Modify title cards
			#	for card in runinput["TITLE"]:
			#		card.setText(title)

			# --- random number seed ---
			rnd = run.getChildOrParent(parent,"rnd")
			if rnd:
				try:	# Modify random seeds cards
					for card in runinput.cards["/random/setSeeds"]:
						# FIXME dirty hack!!!!!
						# setSeeds needs more than 1 value!!!
						card.setWhat(0,rnd)
						card.setWhat(1,rnd)
						#card.setWhat(0, "%d %d"%(rnd, rnd))
				except KeyError:
					# No /random/setSeets present insert one
					#card = MoiraCard("/random/setSeeds",["%d %d"%(rnd,rnd)])
					card = MoiraCard("/random/setSeeds",[rnd,rnd])
					# Find position of /run/beamOn and insert above
					# FIXME what if beamOn is not there as well
					try:
						pos = runinput["/run/beamOn"][0].pos()
					except:
						pos = None
					runinput.add(card, pos, True)

			# --- primaries/time ---
			primaries = run.getChildOrParent(parent,"primaries")
			timerun   = run.getChildOrParent(parent,"time")
			if primaries or timerun:
				# Modify title start cards
				for card in runinput["/run/beamOn"]:
					if primaries:
						card.setWhat(0, int(primaries))
					#if timerun:
					#	card.setWhat(6, timerun)

		# ------------------------
		# Create a new Mcnp Input
		# ------------------------
		elif isinstance(runinput, Mcnp):
			## --- title ---
			#title = run.getChildOrParent(parent,"title")
			#if title:
			#	# Modify title cards
			#	for card in runinput["TITLE"]:
			#		card.setText(title)

			# --- random number seed ---
			rnd = run.getChildOrParent(parent,"rnd")
			if rnd:
				try:	# Modify random seeds cards
					for card in runinput.cards["/random/setSeeds"]:
						# FIXME dirty hack!!!!!
						# setSeeds needs more than 1 value!!!
						card.setWhat(0,rnd)
						card.setWhat(1,rnd)
						#card.setWhat(0, "%d %d"%(rnd, rnd))
				except KeyError:
					# No /random/setSeets present insert one
					#card = MoiraCard("/random/setSeeds",["%d %d"%(rnd,rnd)])
					card = MoiraCard("/random/setSeeds",[rnd,rnd])
					# Find position of /run/beamOn and insert above
					# FIXME what if beamOn is not there as well
					try:
						pos = runinput["/run/beamOn"][0].pos()
					except:
						pos = None
					runinput.add(card, pos, True)

			# --- primaries/time ---
			primaries = run.getChildOrParent(parent,"primaries")
			timerun   = run.getChildOrParent(parent,"time")
			if primaries or timerun:
				# Modify title start cards
				for card in runinput["/run/beamOn"]:
					if primaries:
						card.setWhat(0, int(primaries))
					#if timerun:
					#	card.setWhat(6, timerun)

		if cwd is not None:
			filename = os.path.join(cwd, run.inputBaseName()+runinput.EXT())
		else:
			filename = run.name + runinput.EXT()

		# Pre-process and save input
		errors = runinput.preprocess()
		# FIXME Do I need this one???
		errors.extend(runinput.disableDuplicateDefines())
		if isinstance(runinput, Fluka):
			kwargs = {"expr" : FlukaCard.EXPR_EXPAND}
		else:
			kwargs = {}

		try:
			runinput.save(filename, backup=keepBackup, **kwargs)
		except Exception as ex:
			errors.append((None,str(ex)))

		del runinput
		return errors

	# ----------------------------------------------------------------------
	# return a tuple to spawn a fluka command
	# FIXME only used by Plot, should be canceled
	# ----------------------------------------------------------------------
	def plotgeomCmd(self, run, inpname):
		"""Return a tuple containing the command to spawn a fluka run"""
		if run is None:
			run  = self.runs.default()
			mode = self.mode()
		else:
			mode = run.mode()

		cmd = [run.config().program()]
		if mode == Fluka.MODE:
			exe = run.exe()
			if exe:
				cmd.append("-e")
				cmd.append(exe)
		elif mode == Moira.MODE:
			cmd.append("-p")
		cmd.append("-M")
		cmd.append("1")
		if inpname:
			cmd.append(inpname)
		else:
			cmd.append(self.inputName())
		return tuple(cmd)

	# ----------------------------------------------------------------------
	# Delete files generated by running fluka/moira
	# [ran]name[0-9][0-9][0-9]*
	# ----------------------------------------------------------------------
	def deleteFiles(self, name, pid=None):
		"""Delete temporary files and directory"""
		path = os.path.dirname(name)
		name = os.path.basename(name)
		if path and path[0]!="/": path = os.path.join(self.dir(),path)
		p  = re.compile(rf"^{name}\d\d\d[._].+$")
		rp = re.compile(rf"^ran{name}\d\d\d$")
		for fn in tool.listdir(path):
			if p.match(fn) or rp.match(fn):
				os.remove(os.path.join(path,fn))
		if pid is not None:
			for s in (FlukaConfig.CYCLE, MoiraConfig.CYCLE):
				rundir = os.path.join(path, s + str(pid))
				try:
					os.stat(rundir)
					for f in tool.listdir(rundir):
						os.remove(os.path.join(rundir,f))
					os.rmdir(rundir)
				except OSError:
					pass

		for ext in inputs.ALL_EXTS:
			try: os.remove(os.path.join(path,name+ext))
			except: pass

		# try to remove the fluka echo file
		try: os.remove(os.path.join(path,name+"-echo.inp"))
		except: pass

	# ----------------------------------------------------------------------
	def _scanUSR(self, usr, score, tag):
		for card in self.input[tag]:
			if card.ignore(): continue

			if isinstance(card, MoiraCard):
				if tag == "/score/dumpAllQuantitiesToFile":
					unit = card.what(1)
				else:
					unit = card.name()

			elif isinstance(card, FlukaCard):
				if tag == "DETECT":
					unit = '17'
				else:
					try: unit = int(card.fileunits(False)[0])
					except IndexError: continue
					# Use only binary files
					if unit>0: continue
					unit = str(abs(unit))

			if unit not in usr:
				usr[unit] = score

	# ----------------------------------------------------------------------
	# Scan Input for available binary scoring cards
	# ----------------------------------------------------------------------
	def scanInput4Usr(self):
		"""Scan Input for available binary scoring cards"""
		usr = {}

		# Fluka
		for tag in UsrInfo.TYPES:
			self._scanUSR(usr, tag, tag.upper())

		# Moira
		# FIXME should be made like fluka
		self._scanUSR(usr, "merge.py", "/tally/file")
		self._scanUSR(usr, "usrbin",   "MESH.BOX")
		self._scanUSR(usr, "usrbin",   "MESH.CYLINDER")
		self._scanUSR(usr, "usrbin",   "/score/dumpAllQuantitiesToFile")
		return usr

	# ----------------------------------------------------------------------
	# For each run fill default UsrInfo
	# ----------------------------------------------------------------------
	def findRunUsrInfo(self, run):
		self.input.preprocess(run.defines())
		# add all non existing units in run
		for u,t in self.scanInput4Usr().items():
			for i in run.usrinfo:
				if i.unit == u:
					break
			else:
				ui = UsrInfo(run)
				ui.set(u, t)
				run.usrinfo.append(ui)
		run.usrinfo.sort(key=attrgetter("unit"))

	# ----------------------------------------------------------------------
	# @return local or global material
	# If it doesn't exist create it in local
	# ----------------------------------------------------------------------
	def getMaterial(self, name):
		mat = self.materials.get(name)
		if mat is None:
			mat = MaterialInfo.globalMaterials.get(name)
			if mat is None:
				self.materials[name] = mat = MaterialInfo(self, name)
		return mat

	# ----------------------------------------------------------------------
	# Move a local material to global
	# ----------------------------------------------------------------------
	def material2Global(self, name):
		mat = self.materials.get(name)
		if mat is None:
			MaterialInfo.globalMaterials[name] = name
			del self.materials[name]

	# ----------------------------------------------------------------------
	# Move a local material to global
	# ----------------------------------------------------------------------
	def material2Local(self, name):
		mat = MaterialInfo.globalMaterials.get(name)
		if mat:
			self.materials[name] = mat
			del MaterialInfo.globalMaterials[name]

	# ----------------------------------------------------------------------
	# Prepare the material palette
	# ----------------------------------------------------------------------
	def materialPalette(self, materialList):
		palette = []
		if isinstance(materialList,list):
			for i,name in enumerate(materialList,1):
				palette.append((i, self.getMaterial(name)['color']))
		else:
			for i,name in materialList.items():
				palette.append((i, self.getMaterial(name)['color']))
		return palette

	# ----------------------------------------------------------------------
	# return dicom info structure from UID
	# FIXME could become a dictionary
	# ----------------------------------------------------------------------
	def dicomInfo(self, uid):
		for di in self.dicoms:
			if uid == di.name:
				return di
		return None

	# ----------------------------------------------------------------------
	# Find specific slice from SOPInstanceUID
	# ----------------------------------------------------------------------
	def dicomInfoFromSOPInstanceUID(self, sop):
		# First time scan all dicom info sets
		for di in self.dicoms:
			for fn,inst,s in di.files:
				if sop == s:
					return di
		return None

	# ----------------------------------------------------------------------
	# Find specific slice from instance
	# ----------------------------------------------------------------------
	def dicomFileFromInstance(self, instance):
		# First time scan all dicom info sets
		for di in self.dicoms:
			for f in di.files:
				if f.instance == instance:
					return f
		return None

	#----------------------------------------------------------------------
	# Return the project mode of a file
	#----------------------------------------------------------------------
	@staticmethod
	def projectMode(filename):
		try:
			with open(filename,"r") as f:
				for line in f:
					word = line.split()
					if word and word[0] == "Mode:":
						return word[1]
		except IOError:
			pass
		return "project"

#===============================================================================
def init():
	initSystemLayers()
	FlukaConfig.setDir()
	McnpConfig.setDir()
	MoiraConfig.setDir()
	PenelopeConfig.setDir()
	PhitsConfig.setDir()

#-----------------------------------------------------------------------
# Execute only the first time the module is loaded
#-----------------------------------------------------------------------
init()
