#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import tkinter as tk

import icons
import Ribbon
import tkFlair
import tkExtra
import Project
import Updates
import RichText
import tkDialogs
import bFileDialog
import FlairRibbon
from log import say
from fluka import Fluka
from mcnp  import Mcnp
from moira import Moira
from phits import Phits
from pstext import PSTextExporter

#===============================================================================
# Recent Menu button
#===============================================================================
class _RecentMenuButton(FlairRibbon.FlairMenu):
	#def __init__(self, master, menulist, **kw):
	#	super().__init__(master, menulist, **kw)

	#-----------------------------------------------------------------------
	def showInNewWindow(self, event=None):
		self.page.inNewWindow = True

	#----------------------------------------------------------------------
	def createMenu(self):
		menu = tk.Menu(self, tearoff=0, activebackground=Ribbon._ACTIVE_COLOR)
		empty = True
		for i in range(tkFlair._maxRecent):
			filename = tkFlair.getRecent(i)
			if not filename: break
			path  = os.path.dirname(filename)
			fn    = os.path.basename(filename)
			icon  = icons.get(Project.Project.projectMode(filename), "flair")
			state = tk.NORMAL if os.path.isfile(filename) else tk.DISABLED
			menu.add_command(label=fn,
				image=icon,
				compound=tk.LEFT,
				accelerator=path, # Show as accelerator in order to be aligned
				state=state,
				command=lambda i=i,p=self.page: p.loadRecent(i))
			empty = False
		if empty:
			self.page.flair.openProject()
			return None
		else:
			menu.bind("<Control-Button-1>", self.showInNewWindow)
			menu.bind("<Shift-Button-1>",   self.showInNewWindow)
			return menu

#===============================================================================
# Input template menu button
#===============================================================================
class _TemplateMenuButton(FlairRibbon._TemplateMenuButton):
	#-----------------------------------------------------------------------
	def showInNewWindow(self, event=None):
		self.page.inNewWindow = True

	#----------------------------------------------------------------------
	def createMenu(self):
		menu = super().createMenu()
		menu.bind("<Control-Button-1>", self.showInNewWindow)
		menu.bind("<Shift-Button-1>",   self.showInNewWindow)
		return menu

#===============================================================================
# Project Tab Page
#===============================================================================
class ProjectPage(FlairRibbon.FlairPage):
	"""Create and/or edit a flair project."""

	_name_ = "Flair"
	_icon_ = "project"

	#----------------------------------------------------------------------
	def init(self):
		self.refreshOn = ("exe","image")
		self.inNewWindow = False

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Project ===========
		group = Ribbon.LabelGroup(self.ribbon, "Project")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col,row=0,0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("newflair32"),
				#command=self.flair.newProject,
				anchor=tk.CENTER,
				command=self.openTemplate)
		b.bind('<Control-Button-1>', self.openTemplateInNewFlair)
		b.bind('<Shift-Button-1>',   self.openTemplateInNewFlair)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "New project with the basic template as input")

		col,row=0,2
		b = _TemplateMenuButton(group.frame, self,
				text="New",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.CENTER)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "New project + input from template")

		# ---
		col,row=1,0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("openflair32"),
				anchor=tk.CENTER,
				command=self.flair.openProject)
		b.bind('<Control-Button-1>', self.openProjectInNewFlair)
		b.bind('<Shift-Button-1>',   self.openProjectInNewFlair)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Open project. Control-click to open in new flair window.")

		col,row=1,2
		b = _RecentMenuButton(group.frame, self,
				text="Open",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.CENTER)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Open recent project")

		# ---
		col,row=2,0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("saveflair32"),
				anchor=tk.CENTER,
				command=self.flair.saveProject)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Save project")

		col,row=2,2
		b = Ribbon.LabelButton(group.frame,
				text="Save",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.CENTER,
				command=self.flair.saveProjectAs)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Save project as")

		# ========== Edit ==============
		group = FlairRibbon.FlairMenuGroup(self.ribbon, self, "Edit",
			[("Edit style", "edit", self.editStyle)])
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		col,row = 0,0
		self.styleCombo = Ribbon.LabelCombobox(group.frame,
					width=12,
					command=self.setStyleFromCombo)
		self.styleCombo.grid(row=row, column=col, columnspan=5, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self.styleCombo, "Set editing style")
		self.styleCombo.fill(self.notes.styleList())

		# ---
		col,row = 0,1
		b = Ribbon.LabelRadiobutton(group.frame,
				image=icons.get("alignleft"),
				variable=self._style,
				value="Left")
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Align left text")

		# ---
		col,row = 1,1
		b = Ribbon.LabelRadiobutton(group.frame,
				image=icons.get("aligncenter"),
				variable=self._style,
				value="Center")
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Align center text")

		# ---
		col,row = 2,1
		b = Ribbon.LabelRadiobutton(group.frame,
				image=icons.get("alignright"),
				variable=self._style,
				value="Right")
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Align right text")

		# ---
		col,row = 3,1
		b = Ribbon.LabelRadiobutton(group.frame,
				image=icons.get("hyperlink"),
				variable=self._style,
				value="Link")
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Hyperlink")

		# ---
		col,row = 4,1
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("image"),
				command=self.insertImage)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Insert image")

		# ---
		col,row = 0,2
		b = Ribbon.LabelRadiobutton(group.frame,
				image=icons.get("bold"),
				variable=self._style,
				value="Bold")
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Bold font [Ctrl-B]")

		# ---
		col,row = 1,2
		b = Ribbon.LabelRadiobutton(group.frame,
				image=icons.get("italic"),
				variable=self._style,
				value="Italic")
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Italics font [Ctrl-I]")

		# ---
		col,row = 2,2
		b = Ribbon.LabelRadiobutton(group.frame,
				image=icons.get("underline"),
				variable=self._style,
				value="Underline")
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Underline font [Ctrl-U]")

		# ---
		col,row = 3,2
		b = Ribbon.LabelRadiobutton(group.frame,
				image=icons.get("overstrike"),
				variable=self._style,
				value="Overstrike")
		b.grid(row=row,  column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Overstrike font [Ctrl-O]")

		# ========== Publish ==============
		group = Ribbon.LabelGroup(self.ribbon, "Publish")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col,row=0,0
		b = Ribbon.LabelButton(group.frame, self.page, "<<Document>>",
				text="Document",
				image=icons.get("new32"),
				compound=tk.TOP,
				#anchor=tk.S,
				state=tk.DISABLED)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NS)
		tkExtra.Balloon.set(b, "Create document from project")

		# ---
		col,row=1,0
		b = Ribbon.LabelButton(group.frame,
				text="Print",
				image=icons.get("print32"),
				compound=tk.TOP,
				#anchor=tk.S,
				command=self.hardcopy)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NS)
		tkExtra.Balloon.set(b, "Print input")

		# ---
		col,row=2,0
		b = Ribbon.LabelButton(group.frame,
				text="Refresh",
				image=icons.get("refresh32"),
				compound=tk.TOP,
				#anchor=tk.S,
				command=self.refreshButton)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NS)
		tkExtra.Balloon.set(b, "Refresh document")

		# ========== Tools ==============
		group = FlairRibbon.FlairMenuGroup(self.ribbon, self, "Tools",
			[("License", "document", self.showLicense)])
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col,row=0,0
		b = Ribbon.LabelButton(group.frame, #self.page, "<<Config>>",
				text="Config",
				image=icons.get("config32"),
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.flair.preferences)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NS)
		tkExtra.Balloon.set(b, "Open configuration dialog")

		# ===
		col,row=1,0
		b = Ribbon.LabelButton(group.frame,
				text="Report",
				image=icons.get("debug"),
				command=self.reportDialog)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Send Error Report")

		# ---
		col,row=1,1
		b = Ribbon.LabelButton(group.frame,
				text="Updates",
				image=icons.get("GLOBAL"),
				command=self.flair.checkUpdates)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Check Updates")

		col,row=1,2
		b = Ribbon.LabelButton(group.frame,
				text="About",
				image=icons.get("about"),
				command=lambda s=self: tkFlair.aboutDialog(s.page))
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "About flair")

		# ========== Tools ==============
		group = Ribbon.LabelGroup(self.ribbon, "Close")
		group.pack(side=tk.RIGHT, fill=tk.Y, padx=0, pady=0)

		# ---
		b = Ribbon.LabelButton(group.frame,
				text="Exit",
				image=icons.get("exit32"),
				compound=tk.TOP,
				command=self.flair.quit)
		b.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		tkExtra.Balloon.set(b, "Close program")

	#----------------------------------------------------------------------
	# Create Project page
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage(False)

		lf = tk.LabelFrame(self.frame, text="Notes", padx=2, pady=2)
		lf.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=tk.YES)

		tk.Label(self.frame, text="Title:").pack(side=tk.LEFT)
		self.title = tk.Entry(self.frame, background="White")
		self.title.pack(side=tk.RIGHT, fill=tk.X, expand=tk.YES)
		tkExtra.Balloon.set(self.title, "Project title")

		self._style = tk.StringVar()
		self._setVar = False
		self._style.set("Normal")
		self._style.trace("w",self.styleChanged)

		self.notes = RichText.RichText(lf,
		#self.notes = tkTextEditor.TextEditor(lf,
			height=4, wrap=tk.WORD, background="white", undo=True)
		self.notes.grid(row=0, column=0, sticky=tk.NSEW)
		self.notes.bind("<ButtonRelease-1>",	self.checkStyle)
		self.notes.bind("<Key>",		self.checkStyle)
		self.notes.bind('<Control-Key-a>',	self.selectAll)
		self.notes.bind('<Control-Key-A>',	self.selectClear)
		self.notes.bind('<Control-Key-f>',	self.findDialog)
		self.notes.bind('<Control-Key-r>',	self.replaceDialog)
		self.notes.bind('<Control-Key-b>',	lambda e,s=self,t="Bold":       s.setStyle(t))
		self.notes.bind('<Control-Key-i>',	lambda e,s=self,t="Italic":     s.setStyle(t))
		self.notes.bind('<Control-Key-u>',	lambda e,s=self,t="Underline":  s.setStyle(t))
		self.notes.bind('<Control-Key-o>',	lambda e,s=self,t="Overstrike": s.setStyle(t))
		self.notes.bind('<Control-Key-l>',	lambda e,s=self,t="Left":       s.setStyle(t))
		self.notes.bind('<Control-Key-r>',	lambda e,s=self,t="Right":      s.setStyle(t))
		self.notes.bind('<Control-Key-1>',	lambda e,s=self,t="Heading 1":  s.setStyle(t))
		self.notes.bind('<Control-Key-2>',	lambda e,s=self,t="Heading 2":  s.setStyle(t))
		self.notes.bind('<Control-Key-3>',	lambda e,s=self,t="Heading 3":  s.setStyle(t))
		self.notes.bind("<<Hardcopy>>",		self.hardcopy)

		self.notes.bind("<<Image>>",     self.openImage)
		self.notes.bind("<<Link>>",      self.openLink)
		self.notes.bind("<<Selection>>", self.selection)
		self.notes_First = self.notes_Last  = None

		sbv = tkExtra.AutoScrollbar(lf, takefocus=False)
		sbv.grid(row=0, column=1, sticky='ns')
		sbh = tkExtra.AutoScrollbar(lf,
				takefocus=False,
				orient=tk.HORIZONTAL)
		sbh.grid(row=1, column=0, sticky='ew')

		self.notes.config(  yscrollcommand=sbv.set,
				    xscrollcommand=sbh.set )
		sbv.config(command=self.notes.yview)
		sbh.config(command=self.notes.xview)

		lf.grid_rowconfigure(0, weight=1)
		lf.grid_columnconfigure(0, weight=1)

		self.title.bind("<Return>",   self.updateTitle)
		self.title.bind("<KP_Enter>", self.updateTitle)
		self.title.bind("<FocusOut>", self.updateTitle)

	# ----------------------------------------------------------------------
	def reportDialog(self):
		tkFlair.ReportDialog.show(self.flair)

	# ----------------------------------------------------------------------
	def openTemplateInNewFlair(self, event=None):
		new = self.flair.new()
		new.openTemplate()
		return "break"

	# ----------------------------------------------------------------------
	def openProjectInNewFlair(self, event=None):
		new = self.flair.new()
		new.openProject()
		return "break"

	#----------------------------------------------------------------------
	def selection(self, event=None):
		if self.notes.focus_get() is self.notes:
			try:
				self.notes_First = self.notes.index(tk.SEL_FIRST)
				self.notes_Last  = self.notes.index(tk.SEL_LAST)
			except tk.TclError:
				self.notes_First = self.notes_Last = None

	#----------------------------------------------------------------------
	def createStatus(self):
		super().createStatus()
		self.dirLabel = tk.Label(self.status,
					text="Dir:",
					justify=tk.LEFT,
					anchor=tk.W,
					width=32,
					relief=tk.SUNKEN)
		self.dirLabel.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X)
		self.exeLabel = tk.Label(self.status,
					text="Exe:",
					foreground="DarkBlue",
					justify=tk.LEFT,
					anchor=tk.W,
					width=10,
					relief=tk.SUNKEN)
		self.exeLabel.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X)
		self.exeLabel.bind("<Button-1>", self.loadExecutable)
		return self.status

	# ----------------------------------------------------------------------
	# refresh contents of files
	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		#say("ProjectPage.refresh")
		super().refresh(what)
		self.title.delete(0, tk.END)
		self.title.insert(0, self.project.title)
		# Find styles remove the ones from the buttons
		styles = self.notes.styleList()
		for s in ("Bold", "Italic", "Underline", "Overstrike",
				"Left", "Center", "Right", "Link"):
			styles.remove(s)
		self.styleCombo.fill(styles)
		self.notes.setText(self.project.notes)
		self.checkStyle()

	# ----------------------------------------------------------------------
	def refreshButton(self):
		self.get()
		self.refresh()

	# ----------------------------------------------------------------------
	def refreshStatus(self):
		super().refreshStatus()
		self.dirLabel["text"] = f"Dir: {self.project.dir()}"
		self.exeLabel["text"] = f"Exe: {os.path.basename(self.project.compile.name)}"

	# ----------------------------------------------------------------------
	def showLicense(self):
		Updates.License(self.page, True)

	# ----------------------------------------------------------------------
	# Load executable
	# ----------------------------------------------------------------------
	def loadExecutable(self, event=None):
		self.get()
		fn = bFileDialog.askopenfilename(master=self.page,
					title="Load Executable",
					filetypes=[ ("All","*"), ("Exe","*.exe") ],
					initialdir=self.project.dir(),
					initialfile=self.project.compile.name)
		if fn:
			self.project.compile.name = fn
			self.project.defaultRun.setExecutable(fn)
			self.modified()
			self.flair.refresh("exe")
			self.refreshStatus()

	# ----------------------------------------------------------------------
	def selectAll(self, event=None):
		if self.title.focus_get() is not self.notes:
			super().selectAll()
		else:
			self.selectClear()
			self.notes.tag_add(tk.SEL, "1.0", tk.END)
		return "break"

	# ----------------------------------------------------------------------
	def selectClear(self, event=None):
		if self.title.focus_get() is not self.notes:
			super().selectClear()
		else:
			self.notes.tag_remove(tk.SEL, "1.0", tk.END)
		return "break"

	# ----------------------------------------------------------------------
	def setStyle(self, style):
		self._style.set(style)

	# ----------------------------------------------------------------------
	def setStyleFromCombo(self):
		if self.notes_First is not None and self.notes.focus_get() is not self.notes:
			self.notes.tag_add(tk.SEL, self.notes_First, self.notes_Last)
		self._style.set(self.styleCombo.get())
		self.notes.focus_set()

	# ----------------------------------------------------------------------
	def styleChanged(self, *abc):
		if self._setVar: return
		self.notes.setStyle(self._style.get())
		self.styleCombo.set(self._style.get())

	# ----------------------------------------------------------------------
	def checkStyle(self, event=None):
		self._setVar = True
		style = self.notes.styleAtCursor()
		self._style.set(style)
		self.styleCombo.set(style)
		self._setVar = False

	# ----------------------------------------------------------------------
	def editStyle(self, event=None):
		RichText.StyleConfigurationDialog(self.page, self.notes, self._style.get())
		self.notes.configureTags()

	# ----------------------------------------------------------------------
	def insertImage(self, event=None):
		filename = bFileDialog.askopenfilename(master=self.page,
				title="Insert image",
				initialdir=self.project.dir(),
				filetypes=[
					("All images", ("*.png","*.gif","*.jpg","*.jpeg")),
					("PNG - Portable Network Graphics", "*.png"),
					("GIF - Graphics Interchange Format", "*.gif"),
					("JPG - Joint Photographic Experts Group", ("*.jpg","*.jpeg")),
					("All","*")])

		if not filename: return
		filename = self.project.relpath(filename)
		self.notes.insertImage(filename)

	# ----------------------------------------------------------------------
	def openImage(self, event=None):
		try:
			image = self.notes.image_cget(tk.INSERT,"image")
		except tk.TclError:
			try:
				image = self.notes.image_cget("insert - 1 char","image")
			except tk.TclError:
				return "break"

		imageName = self.notes.imageName(image)
		if imageName is not None:
			# Switch to Plots page and activate the plot
			name,ext = os.path.splitext(imageName)
			self.flair.showPlot(name)
		return "break"

	# ----------------------------------------------------------------------
	def openLink(self, event=None):
		prev = self.notes.tag_prevrange("link", tk.INSERT)
		if prev:
			from_,to_ = prev
			url = tk.Text.get(self.notes, from_,to_)
			if "://" not in url: url = f"http://{url}"
			tkFlair.openurl(url)

	# --------------------------------------------------------------------
	def findDialog(self, event=None):
		try: txt = self.notes.selection_get()
		except:
			txt = ""
		fd = tkDialogs.FindReplaceDialog(self.page, replace=False)
		fd.show(self.find, None, None, txt)
		try:
			self.notes.tag_delete("find")
		except tk.TclError:
			pass

	# --------------------------------------------------------------------
	def replaceDialog(self, event=None):
		try: txt = self.notes.selection_get()
		except: txt = ""
		fd = tkDialogs.FindReplaceDialog(self.page,replace=True)
		fd.show(self.find, self.replace, self.replaceAll, txt)
		self.notes.tag_delete("find")

	# --------------------------------------------------------------------
	def find(self, findStr=None, matchCase=None):
		if findStr:
			ProjectPage.save_findStr   = findStr
			ProjectPage.save_matchCase = matchCase
		else:
			try:
				findStr   = ProjectPage.save_findStr
				matchCase = ProjectPage.save_matchCase
			except: return

		self.notes.tag_remove(tk.SEL,"0.0",tk.END)
		self.notes.tag_delete("find")
		try:
			index = self.notes.search(findStr,
				"insert + 1 chars",
				nocase = matchCase)
			self.notes.mark_set('insert', index)
			endex = f"insert + {len(findStr)} chars"
			self.notes.tag_add(tk.SEL, index, endex)
			self.notes.tag_add("find", index, endex)
			self.notes.tag_configure("find",
					background = '#800000',
					foreground = '#FFFFFF')
			self.notes.see(index)
			self.found = True
			return True
		except:
			self.flair.notify("Not Found",
				f"Target \"{findStr}\" not found")
			self.found = False
			return False

	# --------------------------------------------------------------------
	def findNext(self, event=None):
		self.find()

	# --------------------------------------------------------------------
	def replace(self, findStr, replaceStr, matchCase):
		if not self.found:
			if not self.find(findStr, matchCase): return
		index = self.notes.index(tk.SEL_FIRST)
		self.notes.delete(tk.SEL_FIRST, tk.SEL_LAST)
		self.notes.insert(index, replaceStr)
		endex = f"{index} + {len(replaceStr)} chars"
		self.notes.tag_add(tk.SEL, index, endex)
		self.notes.tag_add("find", index, endex)
		self.found = False

	# --------------------------------------------------------------------
	def replaceAll(self, findStr, replaceStr, matchCase):
		say("replaceAll", findStr, replaceStr, matchCase)

	# ----------------------------------------------------------------------
	# Update (Get) fields
	# ----------------------------------------------------------------------
	def get(self):
		if self.project is None: return
		self.updateTitle()

		#n = self.notes.txt.get(1.0, tk.END).strip()
		n = self.notes.getText().strip()
		if n != self.project.notes:
			self.project.notes = n
			self.modified()

	# ----------------------------------------------------------------------
	def updateTitle(self, event=None):
		t = self.title.get().strip()
		if t != self.project.title:
			self.project.title = t
			self._updateInputTitle()
			self.flair.inputModified(True)

	# ----------------------------------------------------------------------
	# update input title's
	# ----------------------------------------------------------------------
	def _updateInputTitle(self):
		if self.project.title == "": return

		if isinstance(self.project.input, Moira):
			tag = "/moira/title"
		elif isinstance(self.project.input, (Fluka, Mcnp)):
			tag = "TITLE"
		elif isinstance(self.project.input, Phits):
			tag = "[Title]"

		try: titlecard = self.project.input.cards[tag][0]
		except: return

		# Project title has been modified
		# ask user if he wants to update the titles
		if not tkFlair.askyesno("Update title?",
				"Project title is modified.\n"
				"Update TITLE card?",
				parent=self.page):
			return

		titlecard.setTitle(self.project.title)
		self.flair.refresh("card")

	# ----------------------------------------------------------------------
	# Load Recent
	# ----------------------------------------------------------------------
	def loadRecent(self, r):
		if self.inNewWindow:
			flair = self.flair.new()
			self.inNewWindow = False
		else:
			if self.project.isModified() and \
			   not tkFlair.askyesno("Load recent project",
					"Current Project is modified\nDiscard changes",
					parent=self.page):
				return
			flair = self.flair

		filename = tkFlair.config.get(tkFlair.__program__, f"recent.{r}")
		flair.loadProject(filename)

	# ----------------------------------------------------------------------
	def openTemplate(self, template=None):
		if self.inNewWindow:
			flair = self.flair.new()
			self.inNewWindow = False
		else:
			flair = self.flair

		if template is None:
			flair.openTemplate(tkFlair.inputTemplate(self.project.mode()))
		else:
			flair.openTemplate(template)

	# ----------------------------------------------------------------------
	def hardcopy(self, event=None):
		self.get()
		printer = tkFlair.printer(self.page)
		if not printer: return

		printer.open()
		ps = PSTextExporter(self.notes, printer, printer.paper, printer.landscape)
		ps.header(self.project.title if self.project.title else self.project.name())
		ps.comment()
		ps.prolog()
		ps.export()
		ps.close()
		printer.close()

		self.setStatus("Notes printed",  "DarkBlue")
