#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Wioletta.Kozlowska@cern.ch
#         Vasilis.Vlachoudis@cern.ch
# Date:   16-Aug-2015

from __future__ import print_function

import os
import math
import numpy as np
import tkinter as tk

try:
	import pydicom
except ImportError:
	try:
		import dicom as pydicom
	except ImportError:
		pydicom = None
from scipy import interpolate

import log
import tool
import bmath
import icons
import Global
import Ribbon
import tkFlair
import tkExtra
import FlairRibbon
import bFileDialog

from deck   import Card
from fluka  import FlukaCard
from DicomVoxelTab import UnitListAssist

NEW = "<new>"
AMPRMU = 1.007276470

#===============================================================================
# RT plan processing
#===============================================================================
class RTPlan:
	#-----------------------------------------------------------------------
	def __init__(self, dataset):
		# Patient Info
		self.patientName = self.patientAge = self.patientSex = "No information"

		# RTPlan info
		self.planName = self.planDate = self.doseType = self.planIntent = \
			self.appStatus = self.reviewDate = self.reviewerName = \
			self.studyDate = "No information"
		self.patientSetup = "HFS"

		self.maxEnergy    = 0		# Maximum energy for beam tabulation
		self.ionplan      = True	# ion or normal plan

		# Beam Info
		self.bidx         = {}	# index to beam name#
		self.bName        = {}	# Beam Name to idx
		self.bWeight      = {}	# Beam Weight
		self.bType        = {}	# Planned Beam Type
		self.radType      = {}	# Radiation Type. Description of ions not supported
		self.scanMode     = {}	# Scanning Mode
		self.numCP        = {}	# Number of Control Points
		self.sourceDistance={}	# Virtual Source-Axis Distance  or  Source Axis Distance
		self.totWeight    = {}	# Final Cumulative Meterset Weight
		self.numPart      = {}	# Real number of particles MU-> NP conversion applied
		self.unit         = {}	# Unit Number of Particles
		self.isocenter    = {}	# ISO center position
		self.gantryAngle  = {}	# Gantry Angle		- No Gantry Rotation Supported
		self.patientAngle = {}	# Patient Angle		- No Patient Rotation Supported
		self.gantryPAngle = {}	# Gantry Pitch Angle	- No Gantry Rotation Supported
		self.tableTPAngle = {}	# Table Top Pitch Angle	- No Table Rotation Supported
		self.tableTRAngle = {}	# Table Top Roll Angle	- No Table Rotation Supported
		self.beamModel    = {}	# Beam Model dictionary
		self.Emin         = {}	# minimum energy
		self.Emax         = {}	# maximum energy

		# no beam limiting devices
		if isinstance(dataset,str):
			try:
				dataset = pydicom.read_file(dataset, force=True)
				self.dataset = dataset
			except:
				return
		self.fillRTPlanData(dataset)

	#-----------------------------------------------------------------------
	# Filling information from RTPlan
	#-----------------------------------------------------------------------
	def fillRTPlanData(self, dataset):
		try:
			self.patientName = str(dataset.PatientsName.family_name)+ \
					", " +str(dataset.PatientsName.given_name)
		except AttributeError:
			self.patientName = "No information"

		try:
			self.patientAge = str(dataset.PatientsAge)
		except AttributeError:
			self.patientAge = "No information"

		try:
			self.patientSex = str(dataset.PatientsSex)
		except AttributeError:
			self.patientSex = "No information"

		try:
			self.planName = str(dataset.RTPlanName)
		except AttributeError:
			self.planName = "No information"

		try:
			RTdate = dataset.RTPlanDate
			RTtime = dataset.RTPlanTime
			self.planDate = "%s.%s.%s %s:%s" % \
				(RTdate[: 4], RTdate[ 4: 6], RTdate[ 6:],
				 RTtime[:-4], RTtime[-4:-2])
		except AttributeError:
			self.planDate = "No information"

		try:
			self.doseType = str(dataset.DoseType)
		except AttributeError:
			self.doseType = "No information"

		try:
			self.planIntent = str(dataset.PlanIntent)
		except AttributeError:
			self.planIntent = "No information"

		try:
			self.appStatus = str(dataset.ApprovalStatus)
		except AttributeError:
			self.appStatus = "No information"

		try:
			Revdate = dataset.ReviewDate
			Revtime = dataset.ReviewTime
			self.reviewDate = "%s.%s.%s %s:%s" % \
				(Revdate[: 4], Revdate[ 4: 6], Revdate[ 6:],
				 Revtime[:-4], Revtime[-4:-2])
		except AttributeError:
			self.reviewDate = "No information"

		try:
			self.reviewerName = str(dataset.ReviewerName.family_name)+\
					", " +str(dataset.ReviewerName.given_name)
		except AttributeError:
			self.reviewerName = "No information"

		try:
			Studydate = dataset.StudyDate
			Studytime = dataset.StudyTime
			self.studyDate = "%s.%s.%s %s:%s" % \
				(Studydate[: 4], Studydate[ 4: 6], Studydate[ 6:],
				 Studytime[:-4], Studytime[-4:-2])
		except AttributeError:
			self.studyDate = "No information"

		self.ionplan = "IonBeamSequence" in dataset
		beamSequence = dataset.IonBeamSequence if self.ionplan \
					else dataset.BeamSequence
		try:
			self.beamNum = len(beamSequence)
		except AttributeError:
			self.beamNum = 0

		# TODO: Only one patient position
		#try:
		#	self.patientSetup = str(dataset.PatientSetupSequence[0].PatientPosition)
		#except AttributeError:
		#	self.patientSetup = "HFS"

		for beamId in range(self.beamNum):
			try:
				beam = beamSequence[beamId]
			except AttributeError:
				return
			try:
				self.bidx[beamId] = int(beam.BeamNumber)
			except AttributeError:
				self.bidx[beamId] = "No information"
			try:
				self.numCP[beamId] = int(beam.NumberOfControlPoints)
			except AttributeError:
				self.numCP[beamId] = 0
			try:
				self.bName[beamId] = str(beam.BeamName)
			except AttributeError:
				self.bName[beamId] = "No information"
			try:
				self.bWeight[beamId] = float(beam.FinalCumulativeMetersetWeight)
			except AttributeError:
				self.bWeight[beamId] = 0
			try:
				self.bType[beamId] = str(beam.BeamType)
			except AttributeError:
				self.bType[beamId] = "No information"
			try:
				self.radType[beamId] = str(beam.RadiationType)
			except AttributeError:
				self.radType[beamId] = "No information"
			try:
				self.scanMode[beamId] = str(beam.ScanMode)
			except AttributeError:
				self.scanMode[beamId] = "No information"
			try:
				cp = beam.IonControlPointSequence[0] if self.ionplan \
					else beam.ControlPointSequence[0]
			except:
				return
			try:
				self.gantryAngle[beamId] = float(self.correctAng(cp.GantryAngle))
			except:
				self.gantryAngle[beamId] = 0
			try:
				self.gantryPAngle[beamId] = float(self.correctAng(cp.GantryPitchAngle))
			except:
				self.gantryPAngle[beamId] = 0
			try:
				self.patientAngle[beamId] = float(self.correctAng(cp.PatientSupportAngle))
			except:
				self.patientAngle[beamId] = 0
			try:
				self.tableTPAngle[beamId] = float(self.correctAng(cp.TableTopPitchAngle))
			except:
				self.tableTPAngle[beamId] = 0
			try:
				self.tableTRAngle[beamId] = float(self.correctAng(cp.TableTopRollAngle))
			except:
				self.tableTRAngle[beamId] = 0
			try:
				self.isocenter[beamId] = cp.IsocenterPosition
			except AttributeError:
				self.isocenter[beamId] = [0,0,0]
			#try:
			#	self.refPatientSetup[int(beam.BeamNumber)] = beam.ReferencedPatientSetupNumber
			#except AttributeError:
			#	self.refPatientSetup[int(beam.BeamNumber)] = 0
			if self.ionplan:
				try:
					self.sourceDistance[beamId] = beam.VirtualSourceAxisDistances
				except AttributeError:
					self.sourceDistance[beamId] = "No information"
			else:
				try:
					self.sourceDistance[beamId] = beam.SourceAxisDistance
				except AttributeError:
					self.sourceDistance[beamId] = "No information"
			try:
				self.totWeight[beamId] = beam.FinalCumulativeMetersetWeight
			except AttributeError:
				self.totWeight[beamId] = "No information"
			try:
				self.unit[beamId] = str(beam.PrimaryDosimeterUnit)
			except AttributeError:
				self.unit[beamId] = "None"

			self.Emin[beamId] = 1e99
			self.Emax[beamId] = 0
			for cp in beam.IonControlPointSequence if self.ionplan \
					else beam.ControlPointSequence:
				try:
					E = float(cp.NominalBeamEnergy)
					self.Emin[beamId] = min(self.Emin[beamId], E)
					self.Emax[beamId] = max(self.Emax[beamId], E)
				except AttributeError:
					pass

		# FIXME no beam limiting devices
		#...

	#-----------------------------------------------------------------------
	# Correcting angle values, if angle do not exist return 0
	#-----------------------------------------------------------------------
	def correctAng(self, ang):
		if ang in ([],0.0): return 0
		return ang

	#----------------------------------------------------------------------
	# Read beam model file
	#----------------------------------------------------------------------
	def readBModel(self, bmFile, srcpos=None):
		self.beamModel.clear()
		if not bmFile: return
		prefixes=["#","/","*"," ","\t","\n"]
		en,norm,mom_spread,ang_divX,ang_divY,fwhmx,fwhmy=([] for i in range(7))
		try:
			with open(bmFile,'r') as mfile:
				for line in mfile:
					if line.startswith(tuple(prefixes)):
						pass
					else:
						try:
							data=line.split()
							en.append(        float(data[0]))
							norm.append(      float(data[1]))
							mom_spread.append(float(data[2]))
							ang_divX.append(  float(data[3]))
							ang_divY.append(  float(data[4]))
							fwhmx.append(     float(data[5]))
							fwhmy.append(     float(data[6]))
						except (TypeError, IndexError):
							log.error("Wrong beam model file formatting")
							return False
			self.beamModel = {
					"en"         : np.array(en),
					"norm"       : np.array(norm),	# np per mu
					"mom_spread" : np.array(mom_spread),
					"ang_divX"   : np.array(ang_divX),
					"ang_divY"   : np.array(ang_divY),
					"fwhmx"      : np.array(fwhmx),
					"fwhmy"      : np.array(fwhmy)
				}
			if srcpos: self.beamModel["srcpos"] = float(srcpos)
			return True
		except IOError:
			log.error(f"Cannot open file {bmFile!r}")
			return False

	#-----------------------------------------------------------------------
	def interpolate(self, val, what, kval=1, sval=0.002):
		tcx=interpolate.splrep(self.beamModel["en"],
					self.beamModel[what],
					k=kval,
					s=sval)
		return float(interpolate.splev(val,tcx))

	#-----------------------------------------------------------------------
	# Convert beam parameters to the consistent table
	#-----------------------------------------------------------------------
	def beamData(self, dataset, beamId):
		try:
			if self.ionplan:
				beamSequence = dataset.IonBeamSequence[beamId]
			else:
				beamSequence = dataset.BeamSequence[beamId]
		except Exception as ex:
			log.error("Beam File Error")
			log.errorCont(f"No beam data file was created\n{ex}")
			return None

		if self.unit[beamId] == 'MU' and not self.beamModel:
			log.error("Primary Dosimeter Unit is MU\n")
			log.errorCont("Beam Model needs to be specified for convert to NP")
			return None

		# Radiation type
		radtype = self.radType[beamId]
		if radtype == "PROTON":
			particle = "PROTON"
		elif radtype == "ION":
			try:
				AtNb = int(beamSequence.RadiationAtomicNumber)
				MsNb = int(beamSequence.RadiationMassNumber)
				particle = int(AtNb*1000+MsNb)
			except AttributeError:
				log.error("Beam File Error")
				log.errorCont("No information on Ion Type")
				return None
		elif radtype == "PHOTON":
			#import traceback; traceback.print_stack()
			particle = "PHOTON"
		else:
			log.error("Beam File Error")
			log.errorCont("No information on Ion Type")
			if self.ionplan:
				particle = "PROTON"
			else:
				particle = "PHOTON"

		# particle weight
		totweight = self.totWeight[beamId]	# for MU needs scaling
		#try:
		#	beamMet = dataset.FractionGroupSequence[0].\
		#			ReferencedBeamSequence[beamId].\
		#			BeamMeterset
		#except AttributeError:
		#	beamMet = 1

		#partWeight = beamMet/totweight
		sumweight = 0		# Control sum weight

		# Source position/distance
		srcDist = self.sourceDistance[beamId]
		try:
			zPos = self.beamModel["srcpos"]
		except KeyError:
			if self.ionplan:
				zPos = np.mean(srcDist)/10
			else:
				zPos = srcDist/10

		BEnergy  = 0
		GAngle   = 0
		GPAngle  = 0
		PAngle   = 0
		TTPAngle = 0
		TTRAngle = 0
		cptab    = []
		for i in range(self.numCP[beamId]):
			try:
				if self.ionplan:
					cp = beamSequence.IonControlPointSequence[i]
				else:
					cp = beamSequence.ControlPointSequence[i]
			except AttributeError:
				log.error("Beam File Error")
				log.errorCont("No Control Points Exist")
				return None

			try:
				numSSP = cp.NumberOfScanSpotPositions
			except AttributeError:
				numSSP = 1

			if self.ionplan:
				if "ScanSpotMetersetWeights" in cp:
					if numSSP == 1:
						weights = [cp.ScanSpotMetersetWeights]
					else:
						weights = cp.ScanSpotMetersetWeights
				else:
					thisCum = cp.CumulativeMetersetWeight
					try:
						nextCum = beamSequence.ControlPointSequence[i+1]. \
							CumulativeMetersetWeight
					except:
						nextCum = thisCum
					weights = [nextCum-thisCum]
				# Scan Pos in mm
				try:
					xPos = [x/10 for x in cp.ScanSpotPositionMap[0::2]]
					yPos = [x/10 for x in cp.ScanSpotPositionMap[1::2]]
				except:
					xPos = [0]
					yPos = [0]
			else:
				# if weight are not given find it as the difference
				# from the cumulative weight of the next cp
				thisCum = cp.CumulativeMetersetWeight
				try:
					nextCum = beamSequence.ControlPointSequence[i+1]. \
						CumulativeMetersetWeight
				except:
					nextCum = thisCum
				weights = [nextCum-thisCum]
				xPos = [0]
				yPos = [0]

			try:
				BEnergy = float(cp.NominalBeamEnergy) # Nominal Beam Energy in MeV/n
			except:
				pass	# use previous
			self.maxEnergy = max(self.maxEnergy, BEnergy)

			try:	# np per mu
				if self.unit[0]=="MU":
					norm = self.interpolate(BEnergy, "norm")
					SSweight = [norm*x for x in weights ]
				else:
					SSweight = weights
			except:
				#log.error("Wrong format of the normalization file. "
				#	"No normalization made")
				SSweight = weights

			# check is this scanning spot is correct
			if abs(sum(SSweight))<0.001: continue

			try:
				fwhmx = self.interpolate(BEnergy, "fwhmx")
			except:
				#log.error("Wrong format of the beam model file. "
				#	"Missing fwhmX info")
				try:
					fwhmx = cp.ScanningSpotSize[0]/10
				except:
					fwhmx = 0
			try:
				fwhmy = self.interpolate(BEnergy, "fwhmy")
			except:
				#log.error("Wrong format of the beam model file. "
				#	"Missing fwhmY info")
				try:
					fwhmy = cp.ScanningSpotSize[0]/10
				except:
					fwhmy = 0
			try:
				dMomentum = self.interpolate(BEnergy, "mom_spread")
			except:
				#log.error("Wrong format of the beam model file. "
				#	"Missing momentum spread info")
				dMomentum = 0
			try:
				angularX = self.interpolate(BEnergy, "ang_divX")
				angularY = self.interpolate(BEnergy, "ang_divY")
			except:
				#log.error("Wrong format of the beam model file. "
				#	"Missing angular divergence info")
				angularX = 0
				angularY = 0

			sumweight += abs(sum(weights)) # Control sum weight

			if radtype == "ION":
				BEnergy *= AMPRMU
			#elif radtype == "PHOTON":
			#	import traceback; traceback.print_stack()

			try:
				GAngle = float(self.correctAng(cp.GantryAngle))
			except:
				pass	# use previous
			try:
				GPAngle = float(self.correctAng(cp.GantryPitchAngle))
			except:
				pass	# use previous
			try:
				PAngle = float(self.correctAng(cp.PatientSupportAngle))
			except:
				pass	# use previous
			try:
				TTPAngle = float(self.correctAng(cp.TableTopPitchAngle))
			except:
				pass	# use previous
			try:
				TTRAngle = float(self.correctAng(cp.TableTopRollAngle))
			except:
				pass	# use previous
			#try:
			#	isocenter = cp.IsocenterPosition
			#except AttributeError:
			#	pass	# use previous

			# Solve all angles
			def f1(ang, a, b):
				return a*math.cos(math.radians(ang)) - b*math.sin(math.radians(ang))

			def f2(ang, a, b):
				return a*math.cos(math.radians(ang)) + b*math.sin(math.radians(ang))

			for spot in range(numSSP):
				posx, posy = 0,0
				posz = abs(zPos)

				# Gantry Angle (X,Z)
				ang  = GAngle
				cosx = f2(ang, xPos[spot], -abs(zPos))
				cosy = yPos[spot]
				cosz = f1(ang, -abs(zPos), xPos[spot])
				Xx = f2(ang,0,-yPos[spot])
				Xy = -abs(zPos)
				Xz = f1(ang,-yPos[spot],0)
				posx,posz = f2(ang,posx,posz),f1(ang,posz,posx)

				# Gantry Pitch Angle (Y,Z)
				ang  = GPAngle
				cosy,cosz = f1(ang,cosy,cosz),f2(ang,cosz,cosy)
				Xy,Xz     = f2(ang,Xy,Xz),    f1(ang,Xz,Xy)
				posy,posz = f1(ang,posy,posz),f2(ang,posz,posy)

				# Patient Angle (X,Y)
				ang  = PAngle
				cosx,cosy = f2(ang,cosx,cosy),f1(ang,cosy,cosx)
				Xy,Xz     = f2(ang,Xy,Xz),    f1(ang,Xz,Xy)
				posx,posy = f2(ang,posx,posy),f1(ang,posy,posx)

				# Table Top Pitch Angle (Y,Z)
				ang  = TTPAngle
				cosy,cosz = f2(ang,cosy,cosz),f1(ang,cosz,cosy)
				Xx,Xy     = f2(ang,Xx,Xy),    f1(ang,Xy,Xx)
				posy,posz = f2(ang,posy,posz),f1(ang,posz,posy)

				# Table Top Roll Angle (X,Z)
				ang  = TTRAngle
				cosx,cosz = f1(ang,cosx,cosz),f2(ang,cosz,cosx)
				Xy,Xz     = f2(ang,Xy,Xz),    f1(ang,Xz,Xy)
				posx,posz = f1(ang,posx,posz),f2(ang,posz,posx)

				r = math.sqrt(math.pow(cosx,2)+math.pow(cosy,2)+math.pow(cosz,2))
				cosx /= r
				cosy /= r
				cosz /= r

				r = math.sqrt(math.pow(Xx,2)+math.pow(Xy,2)+math.pow(Xz,2))
				Xx /= r
				Xy /= r
				Xz /= r

				# fill in the Control Point Table
				cptab.append([	particle,
						-BEnergy/1000,
						SSweight[spot], #*partWeight,
						-dMomentum/1000,
						-angularX, -angularY,
						posx, posy, posz,
						-fwhmx, -fwhmy,
						cosx, cosy, cosz,
						Xx, Xy, Xz,
						self.bidx[beamId]])

		if (sumweight-totweight) > 0.0001*sumweight:
			log.error("Beam File Error")
			log.errorCont("Beam Meterset Weight not correct")

		self.numPart[beamId] = sum(row[2] for row in cptab)
		return cptab

	#-----------------------------------------------------------------------
	# Exporting data to the external file
	#-----------------------------------------------------------------------
	@staticmethod
	def exportData(cptab, filename, bidx):
		# Output file
		fname = f"RTPlan_{filename}_beam{bidx}"
		fname_inp = f"{fname}.inp"
		try:	# try removing any previous flair to avoid timestamp problems
			os.remove(f"{fname}.flair")
			tool.backup(fname_inp)
		except:
			pass

		with open(fname_inp, "w") as fout:
			# Print to output file
			fout.write(f"{'*'*20} RTPLAN Data for Beam Spot Cards {'*'*20}\n")
			fout.write("* Card Spotbeam\n")
			fout.write("* 1) particle id\n")
			fout.write("* 2) beam laboratory kinetic energy/n [GeV/n]\n")
			fout.write("* 3) beam spot weight\n")
			fout.write("* 4) momentum FWHM spread (GeV/c)\n")
			fout.write("* 5) beam angular divergence X (mrad)\n")
			fout.write("* 6) beam angular divergence Y\n")
			fout.write("* 1-5) not used\n")
			fout.write("* 6) print beam spot info\n")
			fout.write("*\n")
			fout.write("* Card Spotpos\n")
			fout.write("* 1) x-position of the beam spot\n")
			fout.write("* 2) y-position of the beam spot\n")
			fout.write("* 3) z-position of the beam spot\n")
			fout.write("* 4) beam FWHM with in x-direction (cm)\n")
			fout.write("* 5) beam FWHM width in y-direction (cm)\n")
			fout.write("*\n")
			fout.write("* Card Spotdir\n")
			fout.write("* 1) x-cosine of the beam spot\n")
			fout.write("* 2) y-cosine of the beam spot\n")
			fout.write("* 3) z-cosine of the beam spot\n")
			fout.write("* 4) x-cosine of the beam x-axis\n")
			fout.write("* 5) y-cosine of the beam x-axis\n")
			fout.write("* 6) z-cosine of the beam x-axis\n")
			fout.write("*\n")
			fout.write("* The !@dicom=rtplan variable marks the lines created by RTPlan\n")
			fout.write("!@dicom=rtplan\n")
			fout.write("FREE\n")
			for i,spot in enumerate(cptab):
				fout.write(f"* #{i+1}\n")
				fout.write("!@dicom=rtplan\n")
				print("SPOTBEAM",", ".join(map(bmath.format,spot[0:6]  )),
						sep=", ", file=fout)
				fout.write("!@dicom=rtplan\n")
				print("SPOTPOS", ", ".join(map(bmath.format,spot[6:11] )),"",
						sep=", ", file=fout)
				fout.write("!@dicom=rtplan\n")
				print("SPOTDIR", ", ".join(map(bmath.format,spot[11:17])),
						sep=", ", file=fout)
			fout.write("!@dicom=rtplan\n")
			fout.write("FIXED\n")
		return fname_inp

	#-----------------------------------------------------------------------
	# Arranging the beam to correct files
	#-----------------------------------------------------------------------
	def beams2Files(self, dataset, filename):
		data  = []
		files = []
		spots = []

		# create beam TP files
		for idx in self.bidx:
			data.append(self.beamData(dataset,idx))

		#data[0] = [item for sublist in data for item in sublist]
		for i,d in enumerate(data):
			if d is not None:
				files.append(RTPlan.exportData(d, filename, i+1))
				spots.append(len(d))
			else:
				files.append(None)
				spots.append(None)
		return files,spots

#===============================================================================
# RTPlan Tab class
#===============================================================================
class RTPlanTab(FlairRibbon.FlairTab):
	#-----------------------------------------------------------------------
	def __init__(self, master, page, **kw):
		super().__init__(master, page, "RTPlan", **kw)

		self.filesel = None

		pane = tk.PanedWindow(self, orient=tk.HORIZONTAL, opaqueresize=1)
		pane.pack(fill=tk.BOTH, expand=tk.YES)

		# ------------------ Files ---------------
		vert_pane = tk.PanedWindow(pane, orient=tk.VERTICAL, opaqueresize=1)
		pane.add(vert_pane,stretch='always')

		# ------------------ RTPlan --------------
		beamframe = tk.LabelFrame(vert_pane, text = "Beam Number")
		vert_pane.add(beamframe, stretch="always")

		self.beamList = tkExtra.MultiListbox(beamframe,
						(("Beam",	10, None),
						 ("# points",	10, None),
						 ("Gantry Ang", 10, None),
						 ("Patient Ang",10, None)),
						selectmode=tk.BROWSE)
		self.beamList.pack(anchor=tk.NW, side=tk.TOP, fill=tk.BOTH, expand=tk.Y)
		self.beamList.bind("<<ListboxSelect>>", self.updateInfo)
		self.beamList.sortAssist = None

		# --------------- Beam Model --------------
		modelframe = tk.LabelFrame(vert_pane, text = "Edit beam model file")
		vert_pane.add(modelframe, stretch="always")
		row,col = 0,0

		tk.Label(modelframe, text="Beam Model File:").grid(
				row=row, column=col, sticky=tk.W)
		col+=1
		self.bmFile= tk.Label(modelframe,
					width=24,
					foreground=tkFlair._ELABEL_FOREGROUND_COLOR,
					background=tkFlair._ELABEL_BACKGROUND_COLOR,
					relief=tk.SUNKEN,
					anchor=tk.W)
		self.bmFile.grid(row=row, column=col, sticky=tk.EW)
		self.bmFile.bind('<Button-1>', self.loadBMFile)
		tkExtra.Balloon.set(self.bmFile, "File with beam model")

		col+=1
		b=tk.Button(modelframe,image=icons.get("load"), pady=0,
			 command=self.loadBMFile)
		b.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(b, "Load file with beam model")

		col += 1
		b = tk.Button(modelframe, image=icons.get("edit"), pady=0,
				command=self.editBMFile)
		b.grid(row=row, column=col,sticky=tk.W)
		tkExtra.Balloon.set(b, "Edit beam model file")

		col += 1
		b = tk.Button(modelframe, image=icons.get("save"), pady=0,
				command=self.saveBMFile)
		b.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(b, "Save beam model file")

		col += 1
		b = tk.Button(modelframe, image=icons.get("x"), pady=0,
				command=self.cancelBMFile)
		b.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(b, "Remove beam model file")

		#---------------------------------------------------------------
		row += 1
		col = 0
		self.modelList = tkExtra.MultiListbox(modelframe,
						(("Energy [MeV]", 12, None),
						 ("NP per MU",    12, None),
						 ("Δp [MeV/c]",   12, None),
						 ("Δφx [mrad]",   12, None),
						 ("Δφy [mrad]",   12, None),
						 ("x(FWHM) [cm]", 12, None),
						 ("y(FWHM) [cm]", 12, None)))
		self.modelList.grid(row=row, column=col, columnspan=5, rowspan=4, sticky=tk.NSEW)
		self.modelList.sortAssist=UnitListAssist
		self.modelList.bindList("<ButtonRelease-1>",	self.modelRelease)
		self.modelList.bindList("<Return>",		self.modelEdit)
		self.modelList.bindList("<KP_Enter>",		self.modelEdit)
		self.modelList.bindList("<<Rename>>",		self.modelEdit)
		self.modelList.bindList("<Insert>",		self.modelAdd)
		self.modelList.bindList("<Control-Return>",	self.modelAdd)
		self.modelList.bindList("<Delete>",		self.modelDelete)
		self.modelList.bindList("<<Clone>>",		self.modelClone)
		self.modelList.bindLeftRight()
		#self.modelList.sortAssist = None
		self.bmFileModified = False

		# ---
		col += 5
		b = tk.Button(modelframe, image=icons.get("add"),
				takefocus=False, command=self.modelAdd)
		tkExtra.Balloon.set(b, "Add energy characteristic")
		b.grid(row=row, column=col, sticky=tk.NW)

		row += 1
		b = tk.Button(modelframe, image=icons.get("del"),
				takefocus=False, command=self.modelDelete)
		tkExtra.Balloon.set(b, "Delete energy characteristic")
		b.grid(row=row, column=col, sticky=tk.NW)

		# ---
		row += 1
		b = tk.Button(modelframe, image=icons.get("clone"),
				takefocus=False, command=self.modelClone)
		tkExtra.Balloon.set(b, "Clone energy characteristic")
		b.grid(row=row, column=col, sticky=tk.NW)

		# stretching
		modelframe.grid_rowconfigure(row+1, weight=1)
		modelframe.grid_columnconfigure(1, weight=1)

		#-------------------------Information frame ---------------------
		frame = tk.Frame(pane)
		pane.add(frame, stretch="always")

		infoframe= tk.LabelFrame(frame, text="RT Plan Information")
		infoframe.pack(side=tk.TOP, fill=tk.BOTH)

		# ---
		row = 0
		tk.Label(infoframe, text="Patient Info:",
				foreground="DarkRed").grid(
				row=row, column=0, sticky=tk.W)

		row += 1
		tk.Label(infoframe, text="Patient's Name:").grid(
				row=row, column=0, sticky=tk.E)
		self.patname = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				width=32,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.patname.grid(row=row, column=1, sticky=tk.EW)

		row += 1
		tk.Label(infoframe, text="Patient's Age:").grid(row=row, column=0, sticky=tk.E)
		self.patage = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.patage.grid(row=row, column=1, sticky=tk.EW)

		row += 1
		tk.Label(infoframe, text="Patient's Sex:").grid(row=row, column=0, sticky=tk.E)
		self.patsex = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.patsex.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 5
		tk.Label(infoframe,text="RT Plan Info:", foreground="DarkRed").grid(
				row=row, column=0, sticky=tk.W)

		row +=1
		tk.Label(infoframe, text="RT Plan Name:").grid(row=row, column=0, sticky=tk.E)
		self.planname = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.planname.grid(row=row, column=1, sticky=tk.EW)

		row += 1
		tk.Label(infoframe, text="Dose Type:").grid(row=row, column=0, sticky=tk.E)
		self.dosetype = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.dosetype.grid(row=row, column=1, sticky=tk.EW)

		row += 1
		tk.Label(infoframe, text="Plan Intent:").grid(row=row, column=0, sticky=tk.E)
		self.planintent = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.planintent.grid(row=row, column=1, sticky=tk.EW)

		row += 1
		tk.Label(infoframe, text="Approval Status:").grid(row=row, column=0, sticky=tk.E)
		self.appstatus = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.appstatus.grid(row=row, column=1, sticky=tk.EW)

		row += 1
		tk.Label(infoframe, text="Review Date").grid(row=row, column=0, sticky=tk.E)
		self.reviewdate = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.reviewdate.grid(row=row, column=1, sticky=tk.EW)

		row += 1
		tk.Label(infoframe, text="Reviewer Name").grid(row=row, column=0, sticky=tk.E)
		self.reviewername = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.reviewername.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 5
		tk.Label(infoframe,text="General Beam Info:", foreground="DarkRed").grid(
				row=row, column=0, sticky=tk.W)

		row +=1
		tk.Label(infoframe, text="Beam Name:").grid(row=row, column=0, sticky=tk.E)
		self.beamname = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.beamname.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Beam Weight:").grid(row=row, column=0, sticky=tk.E)
		self.beamdescr = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.beamdescr.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Beam Type:").grid(row=row, column=0, sticky=tk.E)
		self.beamtype = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.beamtype.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Radiation Type:").grid(row=row, column=0, sticky=tk.E)
		self.radtype = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.radtype.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Scan Mode:").grid(row=row, column=0, sticky=tk.E)
		self.scanmode = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.scanmode.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="# Control Points:").grid(row=row, column=0, sticky=tk.E)
		self.numcp = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.numcp.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Energy Range:").grid(row=row, column=0, sticky=tk.E)
		self.energyRange = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.energyRange.grid(row=row, column=1, sticky=tk.EW)

		# ---
		row += 1
		tk.Label(infoframe,text="Positioning:", foreground="DarkRed").grid(
				row=row, column=0, sticky=tk.W)

		row +=1
		tk.Label(infoframe, text="ISO Center:").grid(row=row, column=0, sticky=tk.E)
		self.isolabel= tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
					foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
					background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.isolabel.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Gantry Angle:").grid(row=row, column=0, sticky=tk.E)
		self.gangle = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.gangle.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Gantry Pitch Angle:").grid(row=row, column=0, sticky=tk.E)
		self.gpangle = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.gpangle.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Patient Support Angle:").grid(row=row, column=0, sticky=tk.E)
		self.pangle = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.pangle.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Table Top Pitch Angle:").grid(row=row, column=0, sticky=tk.E)
		self.ttpangle = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.ttpangle.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Table Roll Angle:").grid(row=row, column=0, sticky=tk.E)
		self.ttrangle = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.ttrangle.grid(row=row, column=1, sticky=tk.EW)

		row +=1
		tk.Label(infoframe, text="Source Distance:").grid(row=row, column=0, sticky=tk.E)
		self.srcDist = tk.Label(infoframe, justify=tk.LEFT, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.srcDist.grid(row=row, column=1, sticky=tk.EW)

		infoframe.grid_columnconfigure(1, weight=1)

		self.lastPath = ""

	#-----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# -------------------Process RTPlan -------------------
		group = Ribbon.LabelGroup(self.ribbon, "Process RTPlan")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		b = Ribbon.LabelButton(group.frame, image=icons.get("clean32"),
				text="Clean",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.cleanInput)
		b.pack(side=tk.LEFT, fill=tk.Y)
		tkExtra.Balloon.set(b, "Remove all automatically generated cards for RTPLAN")

		# ---
		b = Ribbon.LabelButton(group.frame, image=icons.get("gear32"),
				text="Create",
				compound=tk.TOP,
				anchor=tk.CENTER,
				command=self.writeInput)
		b.pack(side=tk.LEFT, fill=tk.Y)
		tkExtra.Balloon.set(b, "Create input file with RTPLAN information")

		self.rtplan   = None
		self._updated = False

	# ----------------------------------------------------------------------
	@property
	def dicom(self):	return self.page.dicominfo.dicom

	# ----------------------------------------------------------------------
	def activate(self):
		Global.dump("*-* RTPlanTab.activate", level=2)
		di = self.page.selectedDicom()

		if di is None:
			self.clear()
			return
		try:
			self.dataset = pydicom.read_file(
						self.page.dicominfo.filename(),
						force=True)
		except pydicom.filereader.InvalidDicomError:
			self.dataset = None
			return

		if self.dataset:
			self.dicom.init(self.dataset)

		if self.dicom.modality == "RTPLAN":
			self.refresh()
		else:
			self.flair.notify("No DICOM file selected",
				"Please select RTPlan file first")
			self.clear()

	# ----------------------------------------------------------------------
	def deactivate(self):
		Global.dump("*-* RTPlanTab.deactivate", level=2)
		self.get()

	# ----------------------------------------------------------------------
	def refresh(self):
		Global.dump("*-* RTPlanTab.refresh", level=2)
		self.clear()

		di = self.page.dicominfo
		if di is None: return
		self.dataset = None

		self.filesel = self.page.dicominfo.filename(self.page.dicomfile[0])
		self.dataset = pydicom.read_file(self.filesel, force=True)
		self.rtplan  = RTPlan(self.dataset)
		self.fillBeams()
		self.bmFile["text"] = di("beammodel")
		if self.rtplan.readBModel(self.bmFile["text"]):
			self.updateBMFile()
			self.bmFileModified = False

	# ----------------------------------------------------------------------
	# Get the variables from the sliders
	# ----------------------------------------------------------------------
	def get(self):
		Global.dump("*-* RTPlanTab.get",level=2)
		if not self.winfo_ismapped(): return
		if self.dicom is None: return
		di = self.page.dicominfo
		if di is None: return
		di["beammodel"] = self.bmFile["text"]

	# ----------------------------------------------------------------------
	# Clear table info
	# ----------------------------------------------------------------------
	def clear(self):
		self.dataset=None
		self.beamList.delete(0,tk.END)
		self.patname["text"]      = ""
		self.patage["text"]       = ""
		self.patsex["text"]       = ""
		self.planname["text"]     = ""
		self.dosetype["text"]     = ""
		self.planintent["text"]   = ""
		self.appstatus["text"]    = ""
		self.reviewdate["text"]   = ""
		self.reviewername["text"] = ""
		self.beamname["text"]     = ""
		self.beamdescr["text"]    = ""
		self.beamtype["text"]     = ""
		self.radtype["text"]      = ""
		self.numcp["text"]        = ""
		self.gangle["text"]       = ""
		self.gpangle["text"]      = ""
		self.pangle["text"]       = ""
		self.ttpangle["text"]     = ""
		self.ttrangle["text"]     = ""
		self.srcDist["text"]      = ""
		self.scanmode["text"]     = ""
		self.energyRange["text"]  = ""

	#---------------------------Load Beam Model files---------------------------------
	def loadBMFile(self, event=None):
		if self.rtplan is None: return
		if self.bmFileModified:
			if not tkFlair.askyesno("Load beam model file",
					"Current beam model file is modified\nDiscard changes",
					parent=self):
				return

		fn=bFileDialog.askopenfilename(parent=self,
					title="Import beam model file",
					filetypes=[("Data files","*.dat"),
						   ("Text files","*.txt"),
						   ("All","*")],
					initialdir=self.lastPath,
					initialfile=self.bmFile["text"])

		if not fn: return False

		self.lastPath = os.path.abspath(os.path.dirname(fn))
		relfn = self.project.relpath(fn)
		self.bmFile["text"] = relfn

		if self.rtplan.readBModel(relfn):
			self.updateBMFile()
			self.bmFileModified = False
		self.page.modified()
		return True

	# ----------------------------------------------------------------------
	def saveBMFile(self, event=None):
		if self.rtplan is None: return
		self.getModelFromList()
		fn = self.bmFile["text"]
		if not fn:
			fn = bFileDialog.asksaveasfilename(master=self,
					title="Save beam model file",
					filetypes=[("Data files","*.dat"),
						   ("Text files","*.txt"),
						   ("All","*")],
					initialdir=self.lastPath,
					initialfile=fn)
			self.lastPath = os.path.abspath(os.path.dirname(fn))
			fn=self.project.relpath(fn)
			self.bmFile["text"] = fn
		try:
			f = open(fn,"w")
		except:
			self.flair.notify("Cannot open file",
				f"Cannot open file {fn!r} for writing beam model",
				tkFlair.NOTIFY_ERROR)
			return

		#model = sorted(self.rtplan.beamModel.items(), key=operator.itemgetter(0))
		#  print(model)
		model=self.rtplan.beamModel
		for idx in range(len(model['en'])):
			print(	str(model['en'][idx]),
				str(model['norm'][idx]),
				str(model['mom_spread'][idx]),
				str(model['ang_divX'][idx]),
				str(model['ang_divY'][idx]),
				str(model['fwhmx'][idx]),
				str(model['fwhmy'][idx]),
				sep='\t', end='\n',file=f)
		f.close()

		self.bmFileModified = False
		self.flair.notify("beam model file written",
			f"File {fn!r} with beam model updated",
			tkFlair.NOTIFY_NORMAL)

	# ----------------------------------------------------------------------
	def getModelFromList(self):
		if self.rtplan is None: return
		self.rtplan.beamModel.clear()

		self.modelList.sort(0,False)
		en,norm,mom_spread,ang_divX,ang_divY,fwhmx,fwhmy=([] for i in range(7))

		for elem in self.modelList.get(0,tk.END):
			try:    en.append(float(elem[0]))
			except: continue
			try:    norm.append(float(elem[1]))
			except: norm.append(1.0)
			try:    mom_spread.append(float(elem[2]))
			except: mom_spread.append(0.0)
			try:    ang_divX.append(float(elem[3]))
			except: ang_divX.append(0.0)
			try:    ang_divY.append(float(elem[4]))
			except: ang_divY.append(0.0)
			try:    fwhmx.append(float(elem[5]))
			except: fwhmx.append(1.0)
			try:    fwhmy.append(float(elem[6]))
			except: fwhmy.append(1.0)

		self.rtplan.beamModel = {"en"         : np.array(en),
					 "norm"       : np.array(norm),	# np per mu
					 "mom_spread" : np.array(mom_spread),
					 "ang_divX"   : np.array(ang_divX),
					 "ang_divY"   : np.array(ang_divY),
					 "fwhmx"      : np.array(fwhmx),
					 "fwhmy"      : np.array(fwhmy)
					}

	# ----------------------------------------------------------------------
	def cancelBMFile(self, event=None):
		self.bmFile["text"] = ""
		self.rtplan.beamModel.clear()
		self.bmFileModified = False
		self.modelList.delete(0, tk.END)
		self.page.dicominfo["beammodel"] = ""
		self.page.modified()

	# ----------------------------------------------------------------------
	def editBMFile(self, event=None):
		if self.bmFile["text"]:
			tkFlair.editor(self.bmFile["text"])
		else:
			return
		if self.rtplan is None: return
		if self.rtplan.readBModel(self.bmFile["text"]):
			self.updateBMFile()

	# ----------------------------------------------------------------------
	def updateBMFile(self):
		self.modelList.delete(0,tk.END)
		model=self.rtplan.beamModel
		for idx in range(len(model['en'])):
			self.modelList.insert(tk.END,
				(model['en'][idx],
				 model['norm'][idx],
				 model['mom_spread'][idx],
				 model['ang_divX'][idx],
				 model['ang_divY'][idx],
				 model['fwhmx'][idx],
				 model['fwhmy'][idx]))
		self.modelList.sort(0)

	# ----------------------------------------------------------------------
	def modelAdd(self):
		# add an empty record (only once)
		for en in self.modelList.listbox(0).get(0,tk.END):
			if en == NEW: return
		self.modelList.insert(0,(NEW,"","","","","",""))
		self.modelList.selection_clear(0,tk.END)
		self.modelList.activate(tk.END)
		self.modelList.selection_set(tk.END)
		self.modelList.sort(0)
		self.bmFileModified = True

	# ----------------------------------------------------------------------
	def modelDelete(self, event=None):
		sel = list(map(int,self.modelList.curselection()))
		if not sel: return
		sel.reverse()
		for item in sel:
			self.modelList.delete(item)
		self.bmFileModified = True

	# ----------------------------------------------------------------------
	def modelClone(self, event=None):
		sel = list(map(int,self.modelList.curselection()))
		if not sel: return
		for item in sel:
			self.modelList.insert(tk.END, self.modelList.get(item))
		self.bmFileModified = True
		self.modelList.sort(0)

	# ----------------------------------------------------------------------
	def modelEdit(self, event=None):
		if self.rtplan is None: return
		sel = self.modelList.curselection()
		if len(sel) != 1: return
		sel = int(sel[0])

		listbox = event.widget
		try: active = listbox.index(tk.ACTIVE)
		except: return
		if active != sel: return

		edit = tkExtra.InPlaceFloat(listbox)
		if edit.value is None: return

		self.modelList.activate(active)
		self.modelList.selection_set(active)
		self.bmFileModified = True
		return "break"

	# ----------------------------------------------------------------------
	def modelRelease(self, event):
		act = event.widget.nearest(event.y)
		if act: self.modelList.activate(act)
		return self.modelEdit(event)

	#-----------------------------------------------------------------------
	# Update RTPlan info in the table
	#-----------------------------------------------------------------------
	def fillBeams(self):
		try:
			sel = int(self.beamList.curselection()[0])
		except:
			sel = 0
		self.beamList.delete(0,tk.END)
		self.patname["text"]      = self.rtplan.patientName
		self.patage["text"]       = self.rtplan.patientAge
		self.patsex["text"]       = self.rtplan.patientSex
		self.planname["text"]     = self.rtplan.planName
		self.dosetype["text"]     = self.rtplan.doseType
		self.planintent["text"]   = self.rtplan.planIntent
		self.appstatus["text"]    = self.rtplan.appStatus
		self.reviewdate["text"]   = self.rtplan.reviewDate
		self.reviewername["text"] = self.rtplan.reviewerName

		for beam in list(self.rtplan.bidx):
			self.beamList.insert(tk.END,
					(f"{self.rtplan.bName[beam]} [{self.rtplan.bidx[beam]}]",
					 self.rtplan.numCP[beam],
					 self.rtplan.gantryAngle[beam],
					 self.rtplan.patientAngle[beam]))
		self.beamList.selection_set(sel)
		self.updateInfo()

	#-----------------------------------------------------------------------
	# Update Beam info in the table
	#-----------------------------------------------------------------------
	def updateInfo(self, event=None):
		try:
			sel = int(self.beamList.curselection()[0])
		except:
			sel = 0
		self.beamname["text"]  = self.rtplan.bName[sel]
		self.beamdescr["text"] = self.rtplan.bWeight[sel]
		self.beamtype["text"]  = self.rtplan.bType[sel]
		self.radtype["text"]   = self.rtplan.radType[sel]
		self.numcp["text"]     = self.rtplan.numCP[sel]
		#self.energy["text"]    = self.rtplan.energy[sel]
		self.isolabel["text"]  = ", ".join(map(str,self.rtplan.isocenter[sel]))
		self.gangle["text"]    = self.rtplan.gantryAngle[sel]
		self.gpangle["text"]   = self.rtplan.gantryPAngle[sel]
		self.pangle["text"]    = self.rtplan.patientAngle[sel]
		self.ttpangle["text"]  = self.rtplan.tableTPAngle[sel]
		self.ttrangle["text"]  = self.rtplan.tableTRAngle[sel]
		self.srcDist["text"]   = self.rtplan.sourceDistance[sel]
		self.scanmode["text"]  = self.rtplan.scanMode[sel]
		self.energyRange["text"] = f"{self.rtplan.Emin[sel]} "\
					f".. {self.rtplan.Emax[sel]} MeV"

	#-----------------------------------------------------------------------
	# Clean RTPLAN information from the input file
	#-----------------------------------------------------------------------
	def cleanInput(self):
		self.page.setStatus("Cleanup previously created RTPlan cards", "DarkBlue")
		self.page.setStatusPercent(0, update=True)
		undoinfo = self.flair.undolistRefresh()
		n = len(self.project.input.cardlist)
		for i,card in enumerate(reversed(self.project.input.cardlist)):
			if card["dicom"]=="rtplan":
				pos = card.pos()	# remember pos since del will set it to -1
				undoinfo.append(self.flair.delCardUndo(card.pos(), False))
				if card.tag == "#include":
					if self.project.input.cardlist[pos].tag == "#endinclude":
						# Delete as well the #endinclude (during loading
						# of previous file) it might not have the rtplan set
						undoinfo.append(self.flair.delCardUndo(pos, False))
				self.page.setStatusPercent((i*100) // n, True)
		self.page.setStatusPercent(100, update=True)
		self.flair.addUndolistRefresh(undoinfo, "Delete all RTPlan cards")
		self.page.resetPercent(update=True)
		return True

	#-----------------------------------------------------------------------
	# Add all required information into the input file
	#-----------------------------------------------------------------------
	def writeInput(self):
		# First clean input it will record undo information
		self.get()
		self.cleanInput()
		self.page.setStatus("Creating RTPlan cards", "DarkBlue")
		self.page.setStatusPercent(0, update=True)
		inp=self.project.input
		undoinfo = self.flair.undolistRefresh()
		self.page.get()
		#self.findRTDose()
		self._inccards = []

		self.makePatRD(undoinfo, inp)
		self.page.setStatusPercent(10, update=True)		# fake to show progress
		if not self.makeSPOTFile(undoinfo, inp): return False
		self.makeDefines(undoinfo, inp)
		self.expandIncludeFiles(self._inccards)
		self.page.setStatusPercent(100, update=True)

		self.flair.addUndolistRefresh(undoinfo,"Update input file from RTPLAN")

		self.flair.notify("RTPLAN",
				"Input File was modified:\n"
				"VOXEL card modified\n"
				"USRBIN scoring added\n"
				"BEAMSPOT added with RTPlan information\n"
				"BEAM card modified")
		self.page.resetPercent(update=True)
		self.flair.inputModified()

		if not self.rtplan.ionplan:
			self.flair.notify("PHOTON Plan",
				"PHOTON plan doesn't not use collimator data for the moment.\n"
				"Only pencil beams are generated",
				tkFlair.NOTIFY_WARNING)

		return True

	#-----------------------------------------------------------------------
	# add a card and flag it as rtplan owned
	#-----------------------------------------------------------------------
	def _addCard(self, card, pos, undoinfo):
		undoinfo.append(self.flair.addCardUndo(card, pos, False))
		card["dicom"] = "rtplan"
		return pos+1

	#-----------------------------------------------------------------------
	# Update the beam numbers define the preprocessor cards
	#-----------------------------------------------------------------------
	def makeDefines(self, undoinfo, inp):
		pos = 0
		for i in range(self.rtplan.beamNum):
			card = Card("#define", [f"rtplan_beam{i+1}"])
			pos = self._addCard(card,pos,undoinfo)

		for i in range(self.rtplan.beamNum):
			card = Card("#if", [f"rtplan_beam{i+1}"])
			pos = self._addCard(card,pos,undoinfo)
			card = Card("#define", [f"rtplan_spots{i+1}",self.nspots[i]])
			pos = self._addCard(card,pos,undoinfo)
			card = Card("#define", [f"rtplan_weight{i+1}",
						self.rtplan.numPart[i]])
			pos = self._addCard(card,pos,undoinfo)
			card = Card("#define", [f"rtplan_norm{i+1}",f"=rtplan_weight{i+1}/Gy"])
			pos = self._addCard(card,pos,undoinfo)
			pos = self._addCard(Card("#endif"),pos,undoinfo)

		spots = "="+"+".join([f"int(rtplan_spots{x})"
					for x in range(1,len(self.nspots)+1)])
		card = Card("#define", ["rtplan_spots", spots])
		pos = self._addCard(card,pos,undoinfo)

	#-----------------------------------------------------------------------
	# Update the info of the patient position
	#-----------------------------------------------------------------------
	def makePatRD(self, undoinfo, inp):
		pos = inp.bestPosition("ROT-DEFI")
		rotname = "RTPLAN"

		# defining the isocenter
		self.isocenter = []

		for i in range(self.rtplan.beamNum):
			self.isocenter.append(self.rtplan.isocenter[i])

		if self.isocenter[1:] == self.isocenter[:-1]:
			pos = self._isocenter(self.isocenter[0], undoinfo, pos, rotname)
		else:
			for i in range(len(self.isocenter)):
				pos = self._addCard(
					Card("#if", [f"rtplan_beam{i+1}"],
						f"Defines isocenter setup for beam{i+1}"),
						pos,undoinfo)
				pos = self._isocenter(self.isocenter[i], undoinfo, pos, rotname)
				pos = self._addCard(Card("#endif"),pos,undoinfo)

		# defining patient setup - only one available!
		pos  = self._patposition(0, self.rtplan.patientSetup, undoinfo, pos, rotname)

	#-----------------------------------------------------------------------
	# Create the beam data files
	#-----------------------------------------------------------------------
	def makeSPOTFile(self, undoinfo, inp):
		fn = os.path.splitext(os.path.basename(self.filesel))[0]
		try:
			filenames, self.nspots = self.rtplan.beams2Files(self.dataset, fn)
		except Exception as ex:
			self.flair.notify("Beam file error",
				f"Error creating include beam data file {ex}",
				tkFlair.NOTIFY_ERROR)
			raise ex
			return False
		if all([x is None for x in filenames]):
			self.flair.notify("RTPlan error",
				"No Beam Profile Files created, check your DICOM and beam model",
				tkFlair.NOTIFY_ERROR)
			return False

		card = inp["DEFAULTS"][0]
		card.setSdum("HADROTHE")

		# set the maximum energy for tabulation
		card = inp["BEAM"][0]
		card.setWhat(1,(-float(self.rtplan.maxEnergy)/1000.))

		if str(self.rtplan.radType)=="ION":
			card.setWhat(7,("HEAVYION"))
			pos = inp.bestPosition("IONTRANS")
			pos = self._addCard(FlukaCard("IONTRANS"),pos,undoinfo)
			pos = inp.bestPosition("PHYSICS")
			pos = self._addCard(FlukaCard("PHYSICS",["EVAPORAT",'',2]),pos,undoinfo)
			pos = self._addCard(FlukaCard("PHYSICS",["COALESCE",'',1]),pos+1,undoinfo)

		pos = inp.bestPosition("BEAMPOS")
		card = FlukaCard("SPECSOUR", ["BEAMSPOT", "=rtplan_spots"],
				"Card to initialize special beam - SPOTBEAM")
		pos = self._addCard(card, pos, undoinfo)
		for i,fn in enumerate(filenames):
			if self.nspots[i]:
				pos = self._includeCards(i+1, fn, self.nspots[i], undoinfo, pos)
			self.page.setStatusPercent(10+((i+1)*40)//len(filenames), update=True)

		# Add a global spottran card
		card = FlukaCard("SPOTTRAN",["INVERSE","RTPLAN","RTPLAN","",1,"=rtplan_spots"])
		pos = self._addCard(card, pos, undoinfo)
		return True

	#-----------------------------------------------------------------------
	# Adding ROT-DEFI card based on VOXEL to GANTRY coordinate system
	#-----------------------------------------------------------------------
	def _patposition(self, iTrasf, setup, undoinfo, pos, rotname):
		if setup == "HFS":
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 100, 0, 90],
					"HFS Voxel Rotation"),pos,undoinfo)

		elif setup == "FFS":
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 100, 0, -90],
					"FFS Voxel Rotation"),pos,undoinfo)
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 200, 0, 180],
					"FFS Voxel Rotation"),pos,undoinfo)

		elif setup == "HFP":
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 100, 0, 90],
					"HFP Voxel Rotation"),pos,undoinfo)
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 200, 0, 180],
					"HFP Voxel Rotation"),pos,undoinfo)

		elif setup == "FFP":
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 100, 0, -90],
					"FFP Voxel Rotation"),pos,undoinfo)

		elif setup == "HFDL":
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 100, 0, 90],
					"HFDL Voxel Rotation"),pos,undoinfo)
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 200, 0, 90],
					"HFDL Voxel Rotation"),pos,undoinfo)

		elif setup == "HFDR":
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 100, 0, 90],
					"HFDR Voxel Rotation"),pos,undoinfo)
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 200, 0, -90],
					"HFDR Voxel Rotation"),pos,undoinfo)

		elif setup == "FFDL":
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 100, 0, 90],
					"FFDL Voxel Rotation"),pos,undoinfo)
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 200, 0, -90],
					"FFDL Voxel Rotation"),pos,undoinfo)

		elif setup == "FFDR":
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 100, 0, -90],
					"FFDR Voxel Rotation"),pos,undoinfo)
			pos = self._addCard(
				FlukaCard("ROT-DEFI", [rotname, 200, 0, 90],
					"FFDR Voxel Rotation"),pos,undoinfo)

		return pos

	#-----------------------------------------------------------------------
	# Updating Voxel files - beam isocenter
	#-----------------------------------------------------------------------
	def _isocenter(self, isocenter, undoinfo, pos, rotname):
		pos = self._addCard(FlukaCard("ROT-DEFI", [rotname, 100, "", "",
				-(isocenter[0]/10), -(isocenter[1]/10), -(isocenter[2]/10)],
				"Isocenter translation"),
				pos,undoinfo)
		return pos

	#-------------------------------------------------------------------------
	# Adding source file for beam description
	#-------------------------------------------------------------------------
	def _includeCards(self, bidx, filename, nspots, undoinfo, pos):
		card = Card("#if", [f"rtplan_beam{bidx}"],
				f"Defines beam source file for beam{bidx}")
		pos = self._addCard(card, pos,undoinfo)

		card = Card("#include", [str(filename)])
		self._inccards.append(card)
		pos = self._addCard(card,pos,undoinfo)
		pos = self._addCard(Card("#endinclude"),pos,undoinfo)
		pos = self._addCard(Card("#endif",[]),pos,undoinfo)
		return pos

	#-----------------------------------------------------------------------
	def expandIncludeFiles(self, cards):
		self.project.input.renumber()
		for i,card in enumerate(cards):
			if card.enable:
				self.flair.include(card)
			self.page.setStatusPercent(40+((i+1)*60)//len(cards), update=True)
