# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Wioletta.Kozlowska@cern.ch
#         Vasilis.Vlachoudis@cern.ch
# Date:   16-Aug-2015

import os
import math
import numpy as np
import tkinter as tk

try:
	import pydicom
except ImportError:
	try:
		import dicom as pydicom
	except ImportError:
		pydicom = None

from collections import namedtuple
from matplotlib.figure import Figure
from matplotlib.ticker import FormatStrFormatter
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg #, NavigationToolbar2Tk
#from matplotlib.backend_bases import key_press_handler

import log
import deck
import tool
import Data
import bmath
import icons
import Dicom
import Global
import Ribbon
import tkFlair
import tkExtra
import tkPalette
import DicomDraw
import bFileDialog
import FlairRibbon

from Project import DicomInfo

try:
	import ZoomImage
	from PIL import Image, ImageDraw, ImageOps, ImageFilter
except ImportError:
	pass

PALETTE_WIDTH = 50

VoxelsTuple = namedtuple('VoxelsTuple', ["array","size","low","high","dim","zfirst"])

#===============================================================================
# RTViewer Tab
#===============================================================================
class RTViewerTab(FlairRibbon.FlairTab):
	DASHES = [5,2]

	#----------------------------------------------------------------------
	def __init__(self, master, page, **kw):
		super().__init__(master, page, "RTViewer", **kw)

		#pane = tk.PanedWindow(self, orient=tk.HORIZONTAL, opaqueresize=1)
		#pane.pack(fill=tk.BOTH, expand=tk.YES)

		# -------------- Slices ---------------
		self.roiList = tkExtra.MultiListbox(self,
					(("ROI", 16, None),
					 ("Id",   4, None)))
		#pane.add(self.roiList)
		self.roiList.pack(side=tk.LEFT, fill=tk.BOTH)
		self.roiList.bind("<<ListboxSelect>>", self.roiShow)
		self.roiList.sortAssist = None

		#---------------------------------------------------------------
		# Firstly draw the controls, so when we resize the window not to disappear
		cframe = tk.Frame(self)
		#pane.add(cframe, stretch="always")
		cframe.pack(side=tk.RIGHT, fill=tk.BOTH, expand=tk.YES)

		# controls
		frame = tk.Frame(cframe)
		frame.pack(side=tk.BOTTOM, fill=tk.BOTH)

		row,col = 0,0
		tk.Label(frame, text="Min:").grid(row=row, column=col, sticky=tk.SE)
		col += 1
		self.palmin = tk.Scale(frame, from_=0, to_=100,
					orient=tk.HORIZONTAL,
					command=self.updateImageOnIdle)
		self.palmin.grid(row=row, column=col, sticky=tk.EW)
		self.palmin.bind("<Button-3>", lambda e,s=self: s.palmin.set(0))
		tkExtra.Balloon.set(self.palmin,
				"Minimum palette values to display. Right-click to reset")

		row= 0
		col +=1
		tk.Label(frame, text="Max:").grid(row=row, column=col, sticky=tk.SE)

		col += 1
		self.palmax = tk.Scale(frame, from_=0, to_=100,
					orient=tk.HORIZONTAL,
					command=self.updateImageOnIdle)
		self.palmax.grid(row=row, column=col, sticky=tk.EW)
		self.palmax.bind("<Button-3>", lambda e,s=self : s.palmax.set(s.palmax["to"]))
		tkExtra.Balloon.set(self.palmax,
				"Maximum palette values to display. Right-click to reset")
		col += 1
		tk.Label(frame, text="Δ x :").grid(row=row, column=col, sticky=tk.SE)
		col += 1
		self.palmult = tk.Scale(frame,
					from_=0.1, to_=10, resolution=0.1,
					orient=tk.HORIZONTAL,
					command=self.updateImageOnIdle)
		self.palmult.grid(row=row, column=col, sticky=tk.EW)
		self.palmult.bind("<Button-3>", lambda e,s=self : s.palmult.set(1.0))
		tkExtra.Balloon.set(self.palmult,
				"Difference palette zooming factor. Right-click to reset")

		row,col = row+1,0
		tk.Label(frame, text="Window Center:").grid(row=row, column=col, sticky=tk.SE)

		col += 1
		self.windowCenter = tk.Scale(frame, from_=-4096, to_=4096,
					showvalue=0,
					orient=tk.HORIZONTAL,
					command=self.levelChanged)
		self.windowCenter.grid(row=row, column=col, sticky=tk.EW)
		self.windowCenter.bind("<Button-3>", lambda e,s=self: s.windowCenter.set(0))
		tkExtra.Balloon.set(self.windowCenter,
				"Center or Level of values to display. Right-click to reset")

		# ---
		col += 1
		tk.Label(frame, text="Width:").grid(row=row, column=col, sticky=tk.SE)

		col += 1
		self.windowWidth = tk.Scale(frame, from_=0, to_=16000,
					showvalue=0,
					orient=tk.HORIZONTAL,
					command=self.levelChanged)
		self.windowWidth.grid(row=row, column=col, sticky=tk.EW)
		self.windowWidth.bind("<Button-3>", lambda e,s=self: s.windowWidth.set(0))
		tkExtra.Balloon.set(self.windowWidth,
				"Window around level/center values to display. Right-click to reset")

		# ---
		col += 1
		tk.Label(frame, text="Preset:").grid(row=row, column=col, sticky=tk.SE)

		col += 1
		windows = list(DicomInfo.windows)
		windows.insert(0,"")
		self.windowPreset = tkExtra.Combobox(frame, True,
						command=self.presetChanged,
						width=10)
		self.windowPreset.fill(sorted(windows))
		self.windowPreset.set("")
		self.windowPreset.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(self.windowPreset, "Window presets")

		# -------------
		row,col = row+1,0
		tk.Label(frame, text="Slice:").grid(row=row, column=col, sticky=tk.SE)

		col += 1
		self.sliceScale = tk.Scale(frame, from_=0, to_=100,
					orient=tk.HORIZONTAL,
					command=self.sliceChange)
		self.sliceScale.grid(row=row, column=col, columnspan=5, sticky=tk.EW)
		tkExtra.Balloon.set(self.sliceScale, "Slice/Frame to display")

		frame.grid_columnconfigure(1, weight=1)
		frame.grid_columnconfigure(3, weight=1)

		# -------------- Slices ---------------
		self._gridframe = tk.Frame(cframe, width=320, height=240)
		self._gridframe.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)

		# --------------- Canvas -----------
		self._grid = [[None,None,None],[None,None,None]]	# grid widgets to maximizing
		self.canvas     = []
		self._maxbutton = []
		self._orientIOP = None
		self._orientXY  = [(0,0)]*3
		for i,(label,color) in enumerate((("Planned","Blue"),
					("Calculated","Red"),
					("Calc-Plan","Black"))):
			row = i//2
			col = (i%2)*2
			canvas = ZoomImage.ZoomImage(self._gridframe,
						width=160, heigh=120,
						background="Black")
			# Possible resampling methods
			#Image.NEAREST (0)
			#Image.LANCZOS (1)
			#Image.BILINEAR (2)
			#Image.BICUBIC (3)
			#Image.BOX (4)
			#Image.HAMMING (5)
			canvas.resample = Image.NEAREST
			canvas.grid(row=row, column=col, sticky=tk.NSEW)
			canvas.bindMotion(wheel=False)
			#canvas.bindArrows()
			canvas.bind('<Button-4>',       self.imageZoomIn)
			canvas.bind('<Button-5>',       self.imageZoomOut)
			canvas.bind('<Key-1>',          self.imageZoomOne)
			canvas.bind('<B2-Motion>',      self.imageScanDrag)
			canvas.bind('<ButtonRelease-2>',self.imageScanRelease)
			canvas.bind('<Button-3>',       self.button3)
			canvas.bind('<B3-Motion>',      self.b3motion)
			canvas.bind('<Control-Button-4>',self.downZ)
			canvas.bind('<Control-Button-5>',self.upZ)
			canvas.bind('<<Zoom>>',         self.imageZoom)
			canvas.bind("<<View>>",         self.imageMove)

			self._grid[row][col] = canvas
			self.canvas.append(canvas)

			l = tk.Label(canvas, text=label, foreground=color,
					anchor=tk.N, justify=tk.CENTER)
			l.place(relx=0.5, y=5, anchor=tk.N)
			b = tk.Button(canvas, text="\N{BLACK UP-POINTING TRIANGLE}",
					command=lambda i=i: self.maximize(i),
					relief=tk.FLAT,
					padx=0, pady=0)
			b.place(relx=1.0, y=0, anchor=tk.N+tk.E)
			self._maxbutton.append(b)

		# Binning palette
		self.usrbin_palette = tkPalette.PaletteCanvas(self._gridframe,
						width=PALETTE_WIDTH,
						background="White")
		self.usrbin_palette.grid(row=0, column=1, sticky=tk.NS)
		self._grid[0][1] = self.usrbin_palette

		# Difference palette
		self.diff_palette = tkPalette.PaletteCanvas(self._gridframe,
						width=PALETTE_WIDTH,
						background="White")
		self.diff_palette.grid(row=1, column=1, sticky=tk.NS)
		self._grid[1][1] = self.diff_palette

		# matplotlib canvas for DVH
		self.plotframe = tk.Frame(self._gridframe, width=160, heigh=120)
		self.plotframe.grid(row=1, column=2, sticky=tk.NSEW)
		self._grid[1][2] = self.plotframe

		#self.fig = Figure(figsize=(5,4), dpi=100)
		self.fig = Figure()
		self.plot = self.fig.add_subplot(111)
		#t = np.arange(0, 3, .01)
		#self.fig.add_subplot(111).plot(t, 2 * np.sin(2 * np.pi*t))
		# A tk.DrawingArea.
		self.dvhCanvas = FigureCanvasTkAgg(self.fig, master=self.plotframe)
		self.dvhCanvas.draw()
		self.dvhCanvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)

		b = tk.Button(self.plotframe, text="\N{BLACK UP-POINTING TRIANGLE}",
				command=lambda i=3: self.maximize(i),
				padx=0, pady=0)
		b.place(relx=1.0, y=0, anchor=tk.N+tk.E)
		self._maxbutton.append(b)
		#toolbar = NavigationToolbar2Tk(self.dvhCanvas, self.plotframe)
		#toolbar.update()

		self.bind("<Configure>", self.configureEvent)

		self.view       = tk.StringVar()
		self.view.set("Axial")
		self.dataset    = None		# current loaded dataset
		self.frame      = 0		# Number of frames for scale
		self.planDose   = None		# planned dose		(VoxelsTuple)
		self.planDVH    = Data.Tablis()
		self.planFile   = ""		# planned file RTDOSE/USRBIN
		self.calcDose   = None		# calculated dose	(VoxelsTuple)
		self.calcDVH    = Data.Tablis()
		self.calcFile   = ""		# calculated data file RTDOSE/USRBIN
		self.ct         = None		# ct voxel information	(VoxelsTuple)
		self.rtinfo     = None		# current rtdose dicominfo
		self.ctinfo     = None		# CT DicomInfo structure related to this dataset
		self.maxCanvas  = None

		self.roi_selected = []
		self.roi_file     = None
		self.roi_colors   = {}
		self.roi_array    = {}

		self.usrbin_palette.setPalette("fluka",  30)
		self.diff_palette.setPalette(  "diff",   30)
		self._fluka_palette = self.usrbin_palette.palette.pil()
		self._diff_palette  = self.diff_palette.palette.pil()
		self.maximize()
		self._after = None
		self._force = False		# forced refreshing

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== View ===========
		group = FlairRibbon.FlairMenuGroup(self.ribbon, self, "View",
			[("Refresh", "refresh", self.forceRefresh)])
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		#group.label["background"] = Ribbon._BACKGROUND_GROUP2

		b = Ribbon.LabelRadiobutton(group.frame,
					image=icons.get("axial32"),
					text="Axial",
					compound=tk.TOP,
					variable=self.view,
					command=self.updateImageOnIdle,
					value="Axial",
					padx=5,
					pady=10)
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
		tkExtra.Balloon.set(b, "Axial view")
		b.select()

		b = Ribbon.LabelRadiobutton(group.frame,
					image=icons.get("coronal32"),
					text="Coronal",
					compound=tk.TOP,
					variable=self.view,
					command=self.updateImageOnIdle,
					value="Coronal",
					padx=5)
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
		tkExtra.Balloon.set(b, "Coronal View")

		b = Ribbon.LabelRadiobutton(group.frame,
					image=icons.get("sagittal32"),
					text="Sagittal",
					compound=tk.TOP,
					variable=self.view,
					command=self.updateImageOnIdle,
					value="Sagittal",
					padx=5)
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
		tkExtra.Balloon.set(b, "Sagittal View")

		# ========== Binnings ===========
		group = Ribbon.LabelGroup(self.ribbon, "Dose")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---- Planned Dose ----
		row,col = 0,0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("load"),
				text="Planned:",
				compound=tk.RIGHT,
				font=Ribbon._FONT,
				command=self.loadPUsrbin,
				background=Ribbon._BACKGROUND)
		b.grid(row=row, column=col, sticky=tk.E)
		tkExtra.Balloon.set(b, "Load USRBIN file for planned dose")

		col += 1
		self.pdose = tkExtra.Combobox(group.frame, True,
						command=self.loadPDose,
						width=30,
						foreground=tkFlair._ELABEL_FOREGROUND_COLOR,
						background=Ribbon._BACKGROUND)
		self.pdose.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(self.pdose, "Planned dose USRBIN or RTDOSE")

		col += 1
		self.pdosedet = tkExtra.Combobox(group.frame, True,
						command=self.loadPDoseDet,
						width=10,
						foreground=tkFlair._ELABEL_FOREGROUND_COLOR,
						background=Ribbon._BACKGROUND)
		self.pdosedet.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(self.pdosedet, "Detector id of planned dose")

		# ---
		col += 1
		Ribbon.LabelButton(group.frame,
				text="Norm:",
				font=Ribbon._FONT,
				command=self.updateNorm,
				background=Ribbon._BACKGROUND).grid(
					row=row, column=col, pady=1, sticky=tk.E)

		col +=1
		self.planNorm = tkExtra.Combobox(group.frame, False,
						command=self.updateNorm,
						width=16,
						background="White")
		self.planNorm.set(1.0,False)
		self.planNorm.grid(row=row, column=col, sticky=tk.EW)
		self.planNorm.bind("<Return>",   self.updateNorm)
		self.planNorm.bind("<FocusOut>", self.updateNorm)
		tkExtra.Balloon.set(self.planNorm, "Planned dose normalization factor")

		# ---- Calculated Dose ----
		row,col = 1,0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("load"),
				text="Calculated:",
				compound=tk.RIGHT,
				font=Ribbon._FONT,
				command=self.loadCUsrbin,
				background=Ribbon._BACKGROUND)
		b.grid(row=row, column=col, sticky=tk.E)
		tkExtra.Balloon.set(b, "Load USRBIN file for calculated dose")

		col += 1
		self.cdose = tkExtra.Combobox(group.frame, True,
						command=self.loadCDose,
						width=30,
						foreground=tkFlair._ELABEL_FOREGROUND_COLOR,
						background=Ribbon._BACKGROUND)
		self.cdose.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(self.cdose, "Calculated Dose USRBIN or RTDOSE")

		col += 1
		self.cdosedet = tkExtra.Combobox(group.frame, True,
						command=self.loadCDoseDet,
						width=10,
						foreground=tkFlair._ELABEL_FOREGROUND_COLOR,
						background=Ribbon._BACKGROUND)
		self.cdosedet.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(self.cdosedet, "Detector id of calculated dose")

		# ---
		col += 1
		Ribbon.LabelButton(group.frame,
				text="Norm:",
				font=Ribbon._FONT,
				command=self.updateNorm,
				background=Ribbon._BACKGROUND).grid(
					row=row, column=col, pady=1, sticky=tk.E)

		col +=1
		self.calcNorm = tkExtra.Combobox(group.frame, False,
						command=self.updateNorm,
						width=16,
						background="White")
		self.calcNorm.set(1.0,False)
		self.calcNorm.grid(row=row, column=col, sticky=tk.EW)
		self.calcNorm.bind("<Return>",   self.updateNorm)
		self.calcNorm.bind("<FocusOut>", self.updateNorm)
		tkExtra.Balloon.set(self.calcNorm, "Calculated dose normalization factor")

		# ---- RT Structures ----
		row,col = 2,0
		tk.Label(group.frame,
				text="RTSTRUCT:",
				font=Ribbon._FONT,
				background=Ribbon._BACKGROUND).grid(
					row=row, column=col, sticky=tk.E)
		col += 1
		self.rtstruct = tkExtra.Combobox(group.frame,
					command=self.rtstructChanged,
					width=30,
					foreground=tkFlair._ELABEL_FOREGROUND_COLOR,
					background=Ribbon._BACKGROUND)
		self.rtstruct.grid(row=row, column=col, columnspan=4, sticky=tk.EW)
		tkExtra.Balloon.set(self.rtstruct, "Select RTStruct file")

		group.frame.grid_columnconfigure(1,weight=1)

	# ----------------------------------------------------------------------
	# Fill the dicom array
	# ----------------------------------------------------------------------
	def activate(self):
		Global.dump("*-* RTViewerTab.activate")
		#if self.page.dicominfo == self.rtinfo: return
		self.rtinfo = self.page.dicominfo

		if self.rtinfo is None or self.rtinfo.modality != "RTDOSE":
			# It should never come here but lets protect our self
			self.flair.notify("No RTDOSE selected",
					"Please select a RTDOSE set first",
					tkFlair.NOTIFY_ERROR)
			self.rtinfo = None
			return
		self.refresh()

		# FIXME find referenced rtstruct instead of first one
		# load rtstruct
		if not self.rtinfo("rtstruct"):
			di = [x for x in self.project.dicoms
					if x.modality=="RTSTRUCT"]
			if di:
				self.rtinfo["rtstruct"] = di[0].name
		self.rtstruct.set(self.rtinfo("rtstruct").strip())
		self.fillRTCombos()
		self.fillNorm()

	# ----------------------------------------------------------------------
	def deactivate(self):
		self.get()

	# ----------------------------------------------------------------------
	# Refresh the view in the beginning and set the variables
	# ----------------------------------------------------------------------
	def refresh(self):
		Global.dump("*-* RTViewerTab.refresh")
		di = self.rtinfo
		if di is None:
			self._photo   = None
			self.view.set("Axial")
			self.planFile = ""
			self.calcFile = ""
			self.frame    = 0	# Number of frames for scale
			return

		self.page.analyzeSlices()

		# Find "CT"
		ct = self.project.dicomInfo(di("ct"))
		# scan to find a ct
		if ct is None:
			cts = [x for x in self.project.dicoms
					if x.modality=="CT"]
			ct = cts[0] if cts else None
		if ct is not self.ctinfo:
			self.ctinfo = ct
			if ct is not None:
				di["ct"] = ct.name
			self.loadVoxel()

		if self.ct is None:
			self.flair.notify("CT file not found",
				"CT dicom file is required",
				tkFlair.NOTIFY_ERROR)
			return

		self.dataset = None
		self.view.set(di("view"))

		# Changing the center/window will update the image/crop
		self.windowWidth.set(float(di("window")))
		self.windowCenter.set(float(di("center")))

		self.palmin.set(float(di("palmin")))
		self.palmax.set(float(di("palmax")))
		self.palmult.set(float(di("palmult")))

		# Load planned dose
		self.planNorm.set(di("plannorm"),False)
		self.planFile = di("plandose")
		if self.planFile=="":
			self.pdose.set("", False)
			self.pdosedet.set("", False)
		else:
			self.pdose.set(self.planFile, False)
			self.pdosedet.set(di("plandosedet"), False)
		if os.path.isfile(self.planFile):
			self.planDose = self._readUsrbin(self.planFile, False)
		else:
			self.planFile = self.page.dicomfile[0]
			self.planDose = self._readRTDose(self.planFile)

		# Load calculated dose
		self.calcNorm.set(di("calcnorm"),False)
		self.calcFile = di("calcdose")
		self.cdose.set(self.calcFile, False)
		self.cdosedet.set(di("calcdosedet"), False)
		if os.path.isfile(self.calcFile):
			self.calcDose = self._readUsrbin(self.calcFile, True)
		else:
			self.loadRTDose(True)

		self.setnMax()
		self.loadDVH()
		self.roiShow()

		self.sliceScale.set(float(di("slice")))
		self.updateImageOnIdle()

	# ----------------------------------------------------------------------
	# Force refreshing of all quantities
	# ----------------------------------------------------------------------
	def forceRefresh(self):
		self._force   = True
		self.calcDose = None
		self.planDose = None
		self.calcFile = ""
		self.planFile = ""
		self.calcDVH.reset()
		self.planDVH.reset()
		self.refresh()
		self._force = False

	# ----------------------------------------------------------------------
	# Get the variables from the sliders
	# ----------------------------------------------------------------------
	def get(self):
		if not self.winfo_ismapped(): return
		if self.ctinfo is None: return
		di = self.rtinfo
		if di is None: return
		di["center"]      = self.windowCenter.get()
		di["palmax"]      = self.palmax.get()
		di["palmin"]      = self.palmin.get()
		di["palmult"]     = self.palmult.get()

		di["plandose"]    = self.pdose.get()
		di["plandosedet"] = self.pdosedet.get()
		di["plannorm"]    = self.planNorm.get()

		di["calcdose"]    = self.cdose.get()
		di["calcdosedet"] = self.cdosedet.get()
		di["calcnorm"]    = self.calcNorm.get()

		di["rtstruct"]    = self.rtstruct.get()
		di["slice"]       = self.frame
		di["view"]        = self.view.get()
		di["window"]      = self.windowWidth.get()

	# ----------------------------------------------------------------------
	# Fill combo boxes with all RT DOSE available in project
	# ----------------------------------------------------------------------
	def fillRTCombos(self):
		self.rtstruct.delete(0,tk.END)
		self.rtstruct.insert(tk.END,"")

		self.pdose.delete(0,tk.END)
		self.pdose.insert(tk.END,"")
		self.cdose.delete(0,tk.END)
		self.cdose.insert(tk.END,"")

		for di in self.project.dicoms:
			if di.modality=="RTSTRUCT":
				self.rtstruct.insert(tk.END, di.name)
			elif di.modality == "RTDOSE":
				self.pdose.insert(tk.END, di.name)
				self.cdose.insert(tk.END, di.name)

	# ----------------------------------------------------------------------
	# Fill normalization from input
	# ----------------------------------------------------------------------
	def fillNorm(self):
		self.planNorm.delete(0,tk.END)
		self.planNorm.insert(tk.END,"1.0")
		self.calcNorm.delete(0,tk.END)
		self.calcNorm.insert(tk.END,"1.0")
		for card in self.project.input["#define"]:
			if card.name().startswith("rtplan_norm"):
				self.planNorm.insert(tk.END, card.what(1))
				self.calcNorm.insert(tk.END, card.what(1))

	#-----------------------------------------------------------------------
	# Load voxel file
	#-----------------------------------------------------------------------
	def loadVoxel(self):
		if self.ctinfo.dicom is None:
			Dicom.Dicom.add2DicomInfo(self.ctinfo)
		self.loadArray()
		# FIXME it the user has cropped the image then xmin,ymin,zmin
		# will be different than the array....
		d = self.ctinfo.dicom
		self.ct = VoxelsTuple(d.array,
					(d.nrows(), d.ncolumns(), d.nslices()),		# size
					bmath.Vector(d.xmin(), d.ymin(), d.zmin()),	# low
					bmath.Vector(d.xmax(), d.ymax(), d.zmax()),	# high
					bmath.Vector(d.dx, d.dy, d.dz),			# bin-size
					False)						# shape = nx,ny,nz
		self.ct_iop  = d.iop[:]
		self.ct_iop.extend([	d.iop[1]*d.iop[5]-d.iop[2]*d.iop[4],
					d.iop[2]*d.iop[3]-d.iop[0]*d.iop[5],
					d.iop[0]*d.iop[4]-d.iop[1]*d.iop[3]])

		Global.dump("Voxel:")
		Global.dump(f"\tLow:  {self.ct.low}")
		Global.dump(f"\tHigh: {self.ct.high}")
		Global.dump(f"\tSize: {self.ct.size}")
		Global.dump(f"\tDim:  {self.ct.dim}")
		Global.dump(f"\tShape: {self.ct.array.shape}")

	#-----------------------------------------------------------------------
	# Create the 3D numpy matrix.
	# if master is provided then a dialog will be displayed to show
	# the progress
	# FIXME: cropped arrays
	#-----------------------------------------------------------------------
	def loadArray(self):
		dicom = self.ctinfo.dicom
		if dicom.array is not None: return
		dicom.makeArray()
		if not dicom.scanned():
			self.page.analyzeSlices(dicominfo=self.ctinfo)
		try:
			if dicom.hasSlices:
				self.page.resetPercent(len(dicom.slices))
				for i,(z,fn,sop) in enumerate(dicom.slices):
					# read the file
					ds = pydicom.read_file(self.ctinfo.filename(fn),
								force=True)
					if "TransferSyntaxUID" not in ds.file_meta:
						ds.file_meta.TransferSyntaxUID = \
							pydicom.uid.ImplicitVRLittleEndian
					dicom.array[:,:,i] = ds.pixel_array
					self.page.setStatus(f"Analyzing slices2: {fn[0]}",
							"DarkBlue")
					self.page.setStatusPercent(i, update=True)
			else:
				self.flair.log("makeArray with 3D frames not implemented yet")
		except AttributeError:
			self.flair.notify("Dicom Array",
				"Error making dicom array",
				tkFlair.NOTIFY_ERROR)
		self.page.resetPercent()

	#-----------------------------------------------------------------------
	# Decide if the dose is from usrbin or from dicom and load it accordingly
	#-----------------------------------------------------------------------
	def loadDose(self, calc, detOnly):
		if os.path.isfile(self.cdose.get() if calc else self.pdose.get()):
			if detOnly:
				self.loadUsrbinDet(calc)
			else:
				self.loadUsrbin(calc)
		else:
			self.loadRTDose(calc)
		self.setnMax()
		self.loadDVH()
		self.roiShow()
		self.updateImageOnIdle()

	#-----------------------------------------------------------------------
	def loadPDose(self): return self.loadDose(False, False)
	def loadPDoseDet(self): return self.loadDose(False, True)
	def loadCDose(self): return self.loadDose(True, False)
	def loadCDoseDet(self): return self.loadDose(True, True)

	#-----------------------------------------------------------------------
	# Reading the TPS and FLUKA dose files in usrbin format
	# load Prescribed Dose from bnn File
	# TODO: catch an errors
	#------------------------------------------------------------------------
	def loadPUsrbin(self): self.loadUsrbin(False)
	def loadCUsrbin(self): self.loadUsrbin(True)

	#-----------------------------------------------------------------------
	def loadUsrbin(self, calc):
		fn = bFileDialog.askopenfilename(master=self,
				title="Import Dose from USRBIN file",
				filetypes=[("Fluka USRBIN files",("*.bnn")),
					("All","*")],
				initialdir=os.path.dirname(self.calcFile),
				initialfile=self.cdose.get())
		if not fn:
			if calc:
				self.calcDose = None
				self.calcFile = ""
				self.calcDVH.reset()
			else:
				#self.planDose = None
				self.planFile = ""
				self.planDVH.reset()
			return
		if calc:
			self.calcFile = os.path.abspath(fn)
			self.cdose.set(self.project.relpath(self.calcFile), False)
		else:
			self.planFile = os.path.abspath(fn)
			self.pdose.set(self.project.relpath(self.planFile), False)
		self.loadUsrbinDet(calc)

	#-----------------------------------------------------------------------
	def loadUsrbinDet(self, calc):
		if calc:
			self.calcDose = self._readUsrbin(self.calcFile, calc)
		else:
			self.planDose = self._readUsrbin(self.planFile, calc)
		self.setnMax()
		self.loadDVH()
		self.roiShow()
		self.updateImageOnIdle()

	#-----------------------------------------------------------------------
	# Reading the TPS and FLUKA dose files variables
	#-----------------------------------------------------------------------
	def _readUsrbin(self, filename, calc):
		dosedet = self.cdosedet if calc else self.pdosedet
		if not filename:
			dosedet.set("",False)
			dosedet.delete(0,tk.END)
			return None

		usrbin = Data.Usrbin(filename)
		n = None
		detname = dosedet.get()
		dosedet.delete(0,tk.END)
		for i,d in enumerate(usrbin.detector):
			value = f"{i+1} {d.name}"
			if n is None and detname == value:
				n = i
			dosedet.insert(tk.END, value)
		if n is None: n=0
		det = usrbin.detector[n]
		dosedet.set(f"{n+1} {det.name}", False)

		array = usrbin.readArray(n)
		Global.dump(f"Usrbin: {filename}:{n}")
		Global.dump(f"\tLow:    {det.xlow*10}, {det.ylow*10}, {det.zlow*10}")
		Global.dump(f"\tHigh:   {det.xhigh*10}, {det.yhigh*10}, {det.zhigh*10}")
		Global.dump(f"\tSize:   {det.nx} x {det.ny} x {det.nz}")
		Global.dump(f"\tDim:    {det.dx*10} x {det.dy*10} x {det.dz*10}")
		Global.dump(f"\tShape:  {array.shape}")
		Global.dump(f"\tValues: {array.min()} .. {array.max()}")

		return VoxelsTuple(usrbin.readArray(n),
				(det.nx, det.ny, det.nz),			# size
				# low position wrt to ct.low
				bmath.Vector(det.xlow*10, det.ylow*10, det.zlow*10)-self.ct.low,
				# high position wrt to ct.low
				bmath.Vector(det.xhigh*10, det.yhigh*10, det.zhigh*10)-self.ct.low,
				bmath.Vector(det.dx*10, det.dy*10, det.dz*10),	# bin-size
				False)						# shape = nx,ny,nz

	#-----------------------------------------------------------------------
	# Load RTDose from dicominfo
	#-----------------------------------------------------------------------
	def loadRTDose(self, calc):
		dosedet = self.cdosedet if calc else self.pdosedet
		det = dosedet.get()
		dosedet.delete(0, tk.END)
		if calc:
			di = self.project.dicomInfo(self.cdose.get())
		elif self.pdose.get()=="":
			# Switch back to the default RTDOSE
			di = self.rtinfo
			self.pdosedet.set("",False)
			self.planDose = self._readRTDose(self.page.dicomfile[0])
			return
		else:
			di = self.project.dicomInfo(self.pdose.get())
		if di is None:
			dosedet.set("",False)
			if calc:
				self.calcDose = None
				self.calcFile = ""
				self.calcDVH.reset()
			else:
				self.planDose = None
				self.planFile = ""
				self.planDVH.reset()
			return
		assert di.modality == "RTDOSE"

		# Populate detector combo with all RTDOSEs with the same name
		selected = None
		for fn,inst,_ in di.files:
			inst = str(inst)
			dosedet.insert(tk.END, inst)
			if det == inst:
				selected = fn
		if selected is None:
			selected = di.files[0].name
			dosedet.set(di.files[0].instance, False)
		if calc:
			self.calcFile = di.filename(selected)
			self.calcDose = self._readRTDose(self.calcFile)
		else:
			self.planFile = di.filename(selected)
			self.planDose = self._readRTDose(self.planFile)

	#-----------------------------------------------------------------------
	# @return dose from RTDOSE dicom file
	#-----------------------------------------------------------------------
	def _readRTDose(self, filename):
		dicom = Dicom.Dicom(filename)
		m        = float(dicom.dataset.DoseGridScaling)
		array    = dicom.dataset.pixel_array * m
		nz,ny,nx = array.shape
		low      = dicom.origin()
		high     = low + bmath.Vector(dicom.dx*nx, dicom.dy*ny, dicom.dz*nz)
		ctlow    = self.ct.low if self.ct is not None else bmath.Vector.O

		Global.dump(f"RTDOSE {filename}")
		Global.dump(f"\tLow:    {low}    Voxel-diff={low-ctlow}")
		Global.dump(f"\tHigh:   {high}")
		Global.dump(f"\tSize:   {nx} x {ny} x {nz}")
		Global.dump(f"\tDim:    {dicom.dx} x {dicom.dy} x {dicom.dz}")
		Global.dump(f"\tShape:  {array.shape}")
		Global.dump(f"\tValues: {array.min()} .. {array.max()}")
		return VoxelsTuple(array,
				(nx,ny,nz),			# size
				low - ctlow,			# low wrt to ct
				high - ctlow,			# high wrt to ct
				(dicom.dx,dicom.dy,dicom.dz),	# bin-size
				True)				# shape = nx,ny,nz

	# ----------------------------------------------------------------------
	# For RTSTRUCT return a dictionary with sop and CT filename
	# ----------------------------------------------------------------------
	#def scan4ReferencedCT(self):
	#	self.ctinfo = None
	#	for cnt in self.dataset\
	#			.ReferencedFrameOfReferenceSequence[0]\
	#			.RTReferencedStudySequence[0]\
	#			.RTReferencedSeriesSequence[0]\
	#			.ContourImageSequence:
	#		self.ctinfo = self.project.dicomInfoFromSOPInstanceUID(
	#					cnt.ReferencedSOPInstanceUID)
	#		if self.ctinfo is not None:
	#			return

	# ----------------------------------------------------------------------
	# Handle configure event to redraw palettes
	# ----------------------------------------------------------------------
	def configureEvent(self, event=None):
		self.usrbin_palette.draw()
		self.diff_palette.draw()

	# ----------------------------------------------------------------------
	# Maximize canvas window
	# ----------------------------------------------------------------------
	def maximize(self, idx=None):
		if idx is None or self.maxCanvas is not None:
			if self.maxCanvas is not None:
				self._maxbutton[self.maxCanvas]["text"] = \
						"\N{BLACK UP-POINTING TRIANGLE}"
			self.maxCanvas = None
			if idx is not None and idx<3:
				z = self.canvas[idx].zoom
				#x = self.canvas[idx].canvasx(0)
				#y = self.canvas[idx].canvasy(0)
			else:
				z = None
			# remap everything
			for row,rowwidgets in enumerate(self._grid):
				for col,w in enumerate(rowwidgets):
					w.grid()

			self._gridframe.grid_columnconfigure(0, uniform=1, weight=1)
			self._gridframe.grid_columnconfigure(2, uniform=1, weight=1)
			self._gridframe.grid_rowconfigure(0, uniform=1, weight=1)
			self._gridframe.grid_rowconfigure(1, uniform=1, weight=1)

			if z is not None:
				for c in self.canvas:
					c.zoomImage(z)
					c.xview_moveto(0)
					c.yview_moveto(0)
		else:
			self.maxCanvas = idx
			self._maxbutton[self.maxCanvas]["text"] = \
					"\N{BLACK DOWN-POINTING TRIANGLE}"
			row,col = divmod(idx,2)
			for w in self._grid[1-row]:
				w.grid_remove()
			self._grid[row][(1-col)*2].grid_remove()
			if idx==3: self._grid[row][1].grid_remove()

			self._gridframe.grid_columnconfigure(2*col, uniform=1, weight=1)
			self._gridframe.grid_columnconfigure(2*(1-col), uniform=0,
							weight=0, minsize=0)
			self._gridframe.grid_rowconfigure(row, uniform=1, weight=1)
			self._gridframe.grid_rowconfigure(1-row, uniform=0, weight=0, minsize=0)

		# updatee_idletasks needed to re-configure the palette size
		self._gridframe.update_idletasks()
		self.configureEvent()

	# ----------------------------------------------------------------------
	# Zoom all canvases
	# ----------------------------------------------------------------------
	def imageZoomIn(self, event):
		for c in self.canvas:
			c.zoomIn(event)
		return "break"

	# ----------------------------------------------------------------------
	def imageZoomOut(self, event):
		for c in self.canvas:
			c.zoomOut(event)
		return "break"

	# ----------------------------------------------------------------------
	def imageZoomOne(self, event):
		for c in self.canvas:
			c.zoomOne(event)
		return "break"

	# ----------------------------------------------------------------------
	def imageScanDrag(self, event):
		for c in self.canvas:
			c.scanDrag(event)
		return "break"

	# ----------------------------------------------------------------------
	def imageScanRelease(self, event):
		for c in self.canvas:
			c.scanRelease(event)
		return "break"

	# ----------------------------------------------------------------------
	def button3(self, event):
		self._startx = event.x
		self._starty = event.y
		self._center = self.windowCenter.get()
		self._window = self.windowWidth.get()

	# ----------------------------------------------------------------------
	def b3motion(self, event):
		self.windowWidth.set(self._center + 5*(self._startx - event.x))
		self.windowCenter.set(self._window + 5*(self._starty - event.y))
		self.levelChanged()

	#-----------------------------------------------------------------------
	# Set maximum value to scale of both calculated and planned dose
	# nearest power of 10
	#-----------------------------------------------------------------------
	def setnMax(self, event=None):
		nmax = max(self.planDose.array.max()*self.getPlanNorm(),
			   self.calcDose.array.max()*self.getCalcNorm()
				if self.calcDose else 0)
		if abs(nmax)<1e-10:
			nmax = 10

		nmax *= 2
		#nmax = 10**math.ceil(math.log10(nmax))
		if nmax<=1e-10: nmax = 1.0
		self.palmin["to"] = nmax
		self.palmax["to"] = nmax

		res = 10**(round(math.log10(nmax))-3)
		self.palmin["resolution"] = res
		self.palmax["resolution"] = res
		if self.palmax.get() == 0:
			self.palmax.set(nmax)

	# ----------------------------------------------------------------------
	def updateNorm(self, event=None):
		self.setnMax()
		self.roiShow()
		self.updateImageOnIdle()

	# ----------------------------------------------------------------------
	def getPlanNorm(self):
		norm = self.project.input.eval(self.planNorm.get())
		if norm<=1e-10: norm = 1.0
		return norm

	# ----------------------------------------------------------------------
	def getCalcNorm(self):
		norm = self.project.input.eval(self.calcNorm.get())
		if norm<=1e-10: norm = 1.0
		return norm

	# ----------------------------------------------------------------------
	# Select a slice
	# ----------------------------------------------------------------------
	def sliceChange(self, event=None):
		self.frame = self.sliceScale.get()
		self.updateImageOnIdle()

	#-----------------------------------------------------------------------
	def downZ(self, event=None):
		self.sliceScale.set(max(0,self.frame-1))

	# ----------------------------------------------------------------------
	def upZ(self, event=None):
		self.sliceScale.set(min(self.sliceScale["to"]-1,self.frame+1))

	# ----------------------------------------------------------------------
	def updateImageOnIdle(self, event=None):
		if self._after is not None: return
		self._after = self.after_idle(self.updateImage)

	# ----------------------------------------------------------------------
	# Convert dicom to image and show it on canvas
	# ----------------------------------------------------------------------
	def updateImage(self, event=None):
		self._after = None
		if self.ct is None: return
		di = self.ctinfo

		if di is None:
			self.flair.notify("No DICOM selected",
					"Please select a DICOM set first",
					tkFlair.NOTIFY_ERROR)
			return
		if self.dataset is None:
			try:
				self.dataset = pydicom.read_file(
							di.filename(),
							force=True)
			except pydicom.filereader.InvalidDicomError:
				self.canvas[0].setImage()
				return
		try:	window = int(self.windowWidth.get())
		except: window = None
		try:	center = int(self.windowCenter.get())
		except: center = None
		try:	palmin = float(self.palmin.get())
		except: palmin = 0
		try:
			palmax = float(self.palmax.get())
		except:
			palmax = 100
		if palmax<=palmin:
			palmax = palmin+float(self.palmin["resolution"])
			self.palmax.set(palmax)
		try:	palmult = float(self.palmult.get())
		except: palmult = 1

		#-----------------------------------------------------------------------------
		# Setting the DICOM slice on 3 canvas
		#-----------------------------------------------------------------------------
		images = None
		idx = ("Coronal", "Sagittal", "Axial").index(self.view.get())
		IOP = [[1, 0,0, 0,0,1],	# XXX to be verified
		       [0,-1,0, 0,0,1],	# ok
		       [1, 0,0, 0,1,0],	# ok
		       ][idx]

		sliceLen = self.ct.array.shape[idx]-1
		self.frame = min(self.frame, sliceLen)

		self.sliceScale.config(to_=sliceLen)

		try:
			images = self.views(self.view.get(), window, center, palmin, palmax,palmult)
		except Exception as err:
			tkFlair.addException()
			self.flair.notify("Cannot display dicom",
					str(err),
					tkFlair.NOTIFY_WARNING)
			self.canvas[0].setImage()
			images = None
			return

		for idx,canvas in enumerate(self.canvas):
			canvas.setImage(images[idx])
			self.drawOrientation(idx, IOP)
		self._orientIOP = IOP

		self.usrbin_palette.setLimits(palmin, palmax, False)
		self.usrbin_palette.draw()

		self.diff_palette.setLimits(-palmax/palmult, palmax/palmult, False)
		self.diff_palette.draw()

	# ----------------------------------------------------------------------
	def drawOrientation(self, idx, IOP):
		if self._orientIOP == IOP: return
		self.canvas[idx].delete("orientation")
		DicomDraw.drawOrientation(self.canvas[idx], IOP, 50, 50, 20)
		self._orientXY[idx] = (0,0)
		self._imageMove(idx)

	# ----------------------------------------------------------------------
	def imageZoom(self, event=None):
		self._imageMove(self.canvas.index(event.widget))

	# --------------------------------------------------------------------
	# Compensate moving of orientation axis, to appear immobile on screen
	# --------------------------------------------------------------------
	def imageMove(self, event=None):
		self._imageMove(self.canvas.index(event.widget))

	# --------------------------------------------------------------------
	def _imageMove(self, idx):
		ox,oy = self._orientXY[idx]
		x = self.canvas[idx].canvasx(0)
		y = self.canvas[idx].canvasy(0)
		dx = x - ox
		dy = y - oy
		self.canvas[idx].move("orientation", dx, dy)
		self._orientXY[idx] = x,y

	# ----------------------------------------------------------------------
	def levelChanged(self, s=None):
		self.updateImageOnIdle()

	# ----------------------------------------------------------------------
	def presetChanged(self):
		preset = self.windowPreset.get()
		try:
			center, width = DicomInfo.windows[preset]
		except KeyError:
			return
		self.windowCenter.set(center)
		self.windowWidth.set(width)
		self.updateImageOnIdle()

	#-----------------------------------------------------------------------
	# Set the views on each canvas
	#-----------------------------------------------------------------------
	def views(self, view, window, center, palmin, palmax, palmult):
		#---------------------------------------------------------------
		# Calculate the best resolution to show the image
		# keeping the correct aspect ratio
		#---------------------------------------------------------------
		def calcImageSize(voxels, ix, iy, d=None):
			dx = voxels.dim[ix]
			dy = voxels.dim[iy]
			nx = voxels.size[ix]
			ny = voxels.size[iy]
			if d is not None:
				nx = int(round(nx*dx/d))
				ny = int(round(ny*dy/d))
			elif dx<dy:
				d = dx
				ny = int(round(ny*dy/d))
			else:
				d = dy
				nx = int(round(nx*dx/d))
			return (nx,ny),d

		#---------------------------------------------------------------
		# Return slice image position and size
		#---------------------------------------------------------------
		def bnnView(bnn, norm=1.0):
			nonlocal dpixel

			if view == "Axial":
				idx = int(round((self.frame*self.ct.dim[2]-bnn.low[2]) / bnn.dim[2]))
				if idx<0:
					idx = 0
				elif idx>=bnn.size[2]:
					idx = bnn.size[2]-1
				size,d = calcImageSize(bnn,0,1,dpixel)
				if bnn.zfirst:
					img = bnn.array[idx,:,:]
				else:
					img = np.swapaxes(bnn.array[:,:,idx], 0, 1)
				pos  = (int(round(bnn.low[0]/d)),  int(round(bnn.low[1]/d)))

			elif view == "Coronal":
				idx = int(round((self.frame*self.ct.dim[1]-bnn.low[1]) / bnn.dim[1]))
				if idx < 0:
					idx = 0
				elif idx >= bnn.size[1]:
					idx = bnn.size[1]-1
				size,d = calcImageSize(bnn,0,2,dpixel)
				if bnn.zfirst:
					img = np.flip(bnn.array[:,idx,:], 0)
				else:
					img = np.rot90(bnn.array[:,idx,:], 1)
				pos = (int(round(bnn.low[0]/d)), int(round(bnn.low[2]/d)))

			elif view == "Sagittal":
				idx = int(round((self.frame*self.ct.dim[0]-bnn.low[0]) / bnn.dim[0]))
				if idx < 0:
					idx = 0
				elif idx >= bnn.size[0]:
					idx = bnn.size[0]-1
				size,d = calcImageSize(bnn,1,2,dpixel)
				if bnn.zfirst:
					img = np.flip(bnn.array[:,:,idx],0)
				else:
					img = np.rot90(bnn.array[idx,:,:], 1)
				pos  = (int(round(bnn.low[1]/d)), int(round(bnn.low[2]/d)))
			else:
				assert False
				return None

			return img*norm, pos, size

		#---------------------------------------------------------------
		# bnn to image
		#---------------------------------------------------------------
		def bnn2image(dicomImage, size, bnnArray, bnnSize, bnnPos, palette):
			# arrange dicomIm and bnn results based on the palette
			#Global.dump(f"bnn2image: size={size} bnnSize={bnnSize} bnnPos={bnnPos}")
			bnnImage = Image.fromarray(bnnArray.astype(np.int16)).convert('L')
			bnnImage.putpalette(palette)
			bnnImage = bnnImage.resize(bnnSize, resample=Image.BILINEAR)
			image = Image.new("RGB",size)
			image.paste(bnnImage, bnnPos)
			mask  = image.convert("L")
			image.paste(dicomImage, (0,0), mask)
			return image

		#---------------------------------------------------------------
		# Find edges and merge rois
		#---------------------------------------------------------------
		def roiMerge(array, color):
			nonlocal roiimage
			roiL = Image.fromarray(array, mode="L")
			roiL = roiL.filter(ImageFilter.FIND_EDGES)
			roi = ImageOps.colorize(roiL,
						black="#000000",
						white=f"#{color:06X}")
			if roiimage is None:
				roiimage = roi
			else:
				roiimage.paste(roi, (0,0), roiL)

		#---------------------------------------------------------------
		# Set the image and the size
		#---------------------------------------------------------------
		roiimage = None
		if view == "Axial":
			image = self.ct.array[:,:,self.frame]
			size,dpixel = calcImageSize(self.ct,1,0)
			for idx in self.roi_selected:
				try:
					roiMerge(self.roi_array[idx][:,:,self.frame],
						self.roi_colors[idx])
				except KeyError:
					pass
		elif view == "Coronal":
			image = np.rot90(self.ct.array[self.frame,:,:], 1)
			size,dpixel = calcImageSize(self.ct,0,2)
			for idx in self.roi_selected:
				try:
					roiMerge(np.rot90(self.roi_array[idx][self.frame,:,:], 1),
						self.roi_colors[idx])
				except KeyError:
					pass
		elif view == "Sagittal":
			image = np.rot90(self.ct.array[:,self.frame,:],1)
			size,dpixel = calcImageSize(self.ct,1,2)
			for idx in self.roi_selected:
				try:
					roiMerge(np.rot90(self.roi_array[idx][:,self.frame,:], 1),
						self.roi_colors[idx])
				except KeyError:
					pass
		else:
			return None

		image = DicomDraw.dicom2image(self.dataset, image, window, center)
		image = image.resize(size, resample=Image.BILINEAR)
		image = image.convert("RGB")
		imageP = imageC = imageD = image

		# TODO better interpolation between slices
		# http://stackoverflow.com/questions/10854903/
		# what-is-causing-dimension-dependent-attributeerror-in-pil-fromarray-function
		imgP, bnnPos, bnnSize = bnnView(self.planDose, self.getPlanNorm())
		imageP = bnn2image(image, size,
				DicomDraw.rescale(imgP, palmin, palmax),
				bnnSize, bnnPos, self._fluka_palette)

		if self.calcDose:
			imgC, bnnPos, bnnSize = bnnView(self.calcDose, self.getCalcNorm())
			imageC = bnn2image(image, size,
					DicomDraw.rescale(imgC, palmin, palmax),
					bnnSize, bnnPos, self._fluka_palette)

		if self.calcDose:
			try:
				lev    = palmax/palmult
				img    = DicomDraw.rescale(imgC-imgP, -lev, lev)
				imageD = bnn2image(image, size, img, bnnSize, bnnPos,
							self._diff_palette)
			except ValueError:
				log.error("Planned dose and calculated dose have different dimensions")

		if roiimage:
			roiimage = roiimage.resize(size, resample=Image.NEAREST)
			mask = roiimage.convert("L").point(lambda p: 255*(p>0))
			for image in (imageP, imageC, imageD):
				image.paste(roiimage, (0,0), mask)

		return imageP,imageC,imageD

	#-----------------------------------------------------------------------
	# Find all ROIs and fill the menu
	# Handling the visualization of the ROI structures
	# ----------------------------------------------------------------------
	def rtstructChanged(self, event=None):
		self.roi_selected = []
		self.roi_array    = {}
		self.roi_colors   = {}

		# remember selection if possible
		selected = set()
		for i in self.roiList.curselection():
			selected.add(self.roiList.get(i)[0])
		self.roiList.delete(0,tk.END)
		rtuid = self.rtstruct.get()
		self.roi_file = None
		if not rtuid:
			#self.roiList.pack_forget()
			self.roiShow()
			return
		roidi = self.project.dicomInfo(rtuid)
		if roidi:
			try:
				self.roi_file = pydicom.read_file(roidi.filename(), force=True)
				for roi in self.roi_file.StructureSetROISequence:
					self.roiList.insert(tk.END, (roi.ROIName, roi.ROINumber))
					if roi.ROIName in selected:
						self.roiList.selection_set(tk.END)
			except:
				self.flair.notify("RTSTRUCT Error",
						  "Problem with opening RTSTRUCT File")

		self.loadDVH()
		self.roiShow()
		#if self.roi_file:
			## FIXME for some reason it doens't work it requires a <Configure> event?
			#print("PACK")
			#self.roiList.pack(side=tk.LEFT, fill=tk.BOTH)
			#self.event_generate("<Configure>")
		#else:
		#	self.roiList.pack_forget()

	# ----------------------------------------------------------------------
	# load/create DVH files if needed
	# ----------------------------------------------------------------------
	def loadDVH(self):
		Global.dump("*-* RTViewer.loadDVH")
		self.planDVH.reset()
		self.calcDVH.reset()
		if self.roi_file is None or self.ct is None: return

		# -------------------
		# Load planned DVH
		# -------------------
		try:
			filename = self.rtinfo.filename(self.rtinfo["dvh"])
		except KeyError:
			fn = os.path.splitext(self.planFile)[0]
			filename = f"{fn}_dvh.lis"

		if tool.isnewer(filename, self.planFile) and not self._force:
			self._loadDVH(self.planDVH, filename)
		else:
			self.page._exportDVH(self.rtinfo.dicom, filename,
				lambda s=self, f=filename: s._loadDVHshow(s.planDVH,f))
			self.flair.notify("Creating DVH",
				"DVH creation spawned in background.\n"
				"They will be displayed when ready")
			filename = None

		# -------------------
		# Load calculated DVH
		# -------------------
		if not self.calcFile: return
		fn = os.path.splitext(self.calcFile)[0]
		filename = f"{fn}_dvh.lis"
		if tool.isnewer(filename, self.calcFile) and not self._force:
			self._loadDVH(self.calcDVH, filename)
		elif os.path.isfile(self.cdose.get()):
			# From USRBIN
			prg = deck.DeckConfig.command("usrbin2dvh")
			xo,yo,zo = self.ctinfo.dicom.origin() / 10
			cmd = [prg,
				"-o", filename,
				"-u", self.calcFile,
				"-d", self.cdosedet.get().split()[0],
				"-v", self.ctinfo("voxel"),
				"-x", str(xo),
				"-y", str(yo),
				"-z", str(zo),
				#"--ux",
				#"--uy",
				#"--uz",
				#"--vx",
				#"--vy",
				#"--vz"
				]
			process = Dicom.DVHProcess()
			process.notify = lambda s=self, f=filename: s._loadDVHshow(s.calcDVH,f)
			process.remove = False
			self.flair.addProcess(process, self.page.log, cmd)
			self.flair.log("USRBIN DVH creation started")
			self.page.setStatus("Creating DVH file", "DarkBlue")
			self.flair.notify("Creating DVH",
				"DVH creation spawned in background.\n"
				"They will be displayed when ready")
		else:
			# From RTDOSE
			di = self.project.dicomInfo(self.cdose.get())
			if di.dicom is None:
				Dicom.Dicom.add2DicomInfo(di)
			di.dicom.scanned(False)
			di.dicom.init(self.calcFile)
			di.dicom.calcLimits()
			di.analyzeSlices(fn)
			self.page._exportDVH(di.dicom, filename,
				lambda s=self, f=filename: s._loadDVHshow(s.calcDVH,f))
			self.flair.notify("Creating DVH",
				"DVH creation spawned in background.\n"
				"They will be displayed when ready")

	# ----------------------------------------------------------------------
	# Load DVH from filename
	# ----------------------------------------------------------------------
	def _loadDVH(self, dvh, filename):
		if filename and os.path.isfile(filename):
			Global.dump(f">>>\tloading planned {filename}")
			dvh.read(filename)
			dvh.toNumpy()
			self.flair.log(f"Loaded DVH from {filename}")
			self.flair.notify("Loaded DVH",
				f"Loaded DVH from {filename}")
			return True
		return False

	# ----------------------------------------------------------------------
	# load DVH and show, hook for background process
	# ----------------------------------------------------------------------
	def _loadDVHshow(self, dvh, filename):
		if self._loadDVH(dvh, filename):
			self.roiShow()

	# ----------------------------------------------------------------------
	# ROI listbox changed
	# ----------------------------------------------------------------------
	def roiShow(self, event=None):
		if self.ct is None: return
		self.roi_selected = []
		sel = self.roiList.curselection()
		self.plot.clear()
		if not sel:
			self.updateImageOnIdle()
			return
		self.page.resetPercent(len(sel))
		shown = 0
		for i,item in enumerate(map(int,sel)):
			name,roi = self.roiList.get(item)
			roi = int(roi)
			self.roi_selected.append(roi)
			if roi not in self.roi_array:
				self.page.setStatus(f"Creating ROIs: {name}", "DarkBlue")
				self._addROI(roi)
			shown |= self.showDVH(name,roi)
			self.page.setStatusPercent(i, update=True)

		for roi in list(self.roi_array.keys()):
			if roi not in self.roi_selected:
				self._removeROI(roi)

		if shown:
			if shown&1:
				self.plot.plot([],[],linewidth=0.5,
						color='Black',
						dashes=RTViewerTab.DASHES,
						label="dashes:planned")
			if shown&2:
				self.plot.plot([],[],linewidth=0.5,
						color='Black',
						label="solid:calculated")

			self.plot.set_title('Dose Volume Histogram')
			self.plot.set_xlabel('Dose [Gy]')
			self.plot.set_ylabel('Volume [%]')
			self.plot.xaxis.set_major_formatter(FormatStrFormatter("%g"))
			self.plot.legend()
			self.plot.grid(which='both')

		self.dvhCanvas.draw()
		self.updateImageOnIdle()
		self.page.resetPercent()
		self.page.setStatus()

	# ----------------------------------------------------------------------
	# create ROI matrix
	# ----------------------------------------------------------------------
	def _addROI(self, roi):
		for sequence in self.roi_file.ROIContourSequence:
			if int(sequence.ReferencedROINumber) == roi:
				break
		else:
			return

		array = np.zeros([self.ct.size[0], self.ct.size[1], self.ct.size[2]], dtype=np.byte)
		x0,y0,z0 = self.ct.low
		dx,dy,dz = self.ct.dim
		Xx = self.ct_iop[0]
		Xy = self.ct_iop[1]
		Yx = self.ct_iop[3]
		Yy = self.ct_iop[4]
		if "ContourSequence" not in sequence: return
		for contour in sequence.ContourSequence:
			try:
				if contour.ContourGeometricType != "CLOSED_PLANAR" and \
				   contour.NumberOfContourPoints == 1:
					continue
				coords = list(map(float, contour.ContourData))
				XY = list(zip(coords[::3], coords[1::3]))
				X  = [Xx*(x-x0) + Xy*(y-y0) for x,y in XY]
				Y  = [Yx*(x-x0) + Yy*(y-y0) for x,y in XY]
				if len(X)>1:
					iX = [int(round(x/dx)) for x in X]
					iY = [int(round(y/dy)) for y in Y]
					iZ = int((coords[2]-z0) / dz)
					img = Image.new("L",(self.ct.size[0],self.ct.size[1]))
					ImageDraw.Draw(img).polygon(list(zip(iX,iY)),
								outline=0xff,
								fill=0xff)
					array[:,:,iZ] = np.array(img)
			except AttributeError:
				if contour.ContourGeometricType == "POINT" and \
				   abs(float(contour.ContourData[2]) - z0) <= dz/2:
					npoints = contour.NumberOfContourPoints
					coords = list(map(float, contour.ContourData))
					if npoints == 1:
						x = coords[0]
						y = coords[1]
						iX = int((Xx*(x-x0)+Xy*(y-y0))/dx)
						iY = int((Yx*(x-x0)+Yy*(y-y0))/dy)
						img = Image.new("L",(self.ct.size[0],self.ct.size[1]))
						ImageDraw.Draw(img).line((iX,iY-5,iX,iY+5), fill=1)
						ImageDraw.Draw(img).line((iX-5,iY,iX+5,iY), fill=1)
						array[:,:,iZ] = np.array(img)
		try:
			color = sequence.ROIDisplayColor
			self.roi_colors[roi] = int(color[0])<<16 | int(color[1])<<8 | int(color[2])
		except AttributeError:
			self.roi_colors[roi] = 0x7f7f7f
		self.roi_array[roi] = array

	# ----------------------------------------------------------------------
	# Show DVH plot
	# ----------------------------------------------------------------------
	def showDVH(self, name, roi):
		if roi not in self.roi_array: return False
		title = f"{name} [{roi}]"
		color = f"#{self.roi_colors[roi]:06X}"
		shown = 0
		error = ""

		if self.calcDVH.detector:
			try:
				data = self.calcDVH[title].data
				x=np.add(data[:,0],np.subtract(data[:,1],
						data[:,0])/2)*self.getCalcNorm()
				options = {"color":color, "linewidth":2 }
				self.plot.plot(x,data[:,2],
						color=color,
						linewidth=2,
						label=title)
				shown |= 2
			except KeyError:
				error += f"\nCalculated data {title!r} not found"
				Global.dump(f"{log.WARNING} {error}")

		if self.planDVH.detector:
			try:
				data = self.planDVH[title].data
				x=np.add(data[:,0],np.subtract(data[:,1],
						data[:,0])/2*self.getPlanNorm())
				options = {"color":color,
					"linewidth":1,
					"dashes": RTViewerTab.DASHES }
				if not shown: options["label"] = title
				self.plot.plot(x,data[:,2], **options)
				shown |= 1
			except KeyError:
				error = f"Planned data {title!r} not found"
				Global.dump(f"{log.WARNING} {error}")

		if error:
			self.flair.notify("DVH", error, tkFlair.NOTIFY_WARNING)
		return shown

	# ----------------------------------------------------------------------
	# remove ROI to recover memory
	# ----------------------------------------------------------------------
	def _removeROI(self, idx):
		del self.roi_array[idx]
		del self.roi_colors[idx]
