# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   06-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import io
import os
import pickle
import tkinter as tk

import icons
import inputs
import tkExtra
import tkFlair
from Project import GroupInfo, RunInfo
from tkconst import STATE

#===============================================================================
# Run Listbox
#===============================================================================
class RunListbox(tkFlair.InfoTreeview):
	def __init__(self, master, page, **kw):
		super().__init__(master, **kw)
		self["columns"] = ("spawn",)
		self.column("#0",    width=100, stretch=tk.YES)
		self.column("spawn", width= 50*tkFlair.dpiscale, stretch=tk.NO, anchor=tk.E)
		self.heading("#0",    text="Run",   anchor=tk.W)
		self.heading("spawn", text="Spawn", anchor=tk.E)

		# Bind events
		self.bind('<<PasteSpecial>>',	self.pasteAbove)
		self.bindSearch()
		#self.bindVi()
		#self.bindLeftRight()

		self.bind("<<Rename>>",		self.rename)
		self.bind("<Return>",		self.editInPlace)
		self.bind("<KP_Enter>",		self.editInPlace)
		self.bind('<Key-F3>',		self.editSpawn)
		self.bind('<Control-Key-e>',	self.editSpawn)
		self.bind("<Control-Key-t>",	self.attach)

		self._groupicon = icons.get("folder")
		self.setPopupMenu([
			("Rename",     0, self.rename,        icons.get("rename")),
			("Add",        0, self.add,           icons.get("add")),
			("Add Folder", 0, self.addGroup,      self._groupicon),
			("Remove",     0, self.remove,        icons.get("x")),
			("Clone",      0, self.clone,         icons.get("clone")),
			None,
			("Run",        0, self.executeRun,    icons.get("run")),
			("Refresh",    0, self.refreshRun,    icons.get("refresh")),
			])

		# Remember selected runs
		self.flair    = None
		self.project  = None
		self.page     = page

	# ---------------------------------------------------------------------
	def set(self, flair):
		self.flair = flair
		self.project = flair.project
		self.fill()

	# ----------------------------------------------------------------------
	# Insert a run or issue a standard Treeview.insert() call
	# ----------------------------------------------------------------------
	def insert(self, parent, index, run=None, **kw):
		#print(f">>> TREE INSERT {parent=} {index=} {run.name=} {type(run)}")
		if isinstance(run, GroupInfo):
			if run.name:
				item = super().insert(parent, index,
							#f"{parent}:{run.name} [{id(run)}]",
							text=run.name,
							image=self._groupicon,
							**kw)
				self.item2info[item] = run
			else:
				item = parent
			open_ = None
			for p in run.childs:
				if open_ is None and isinstance(p,RunInfo):
					open_ = p("open")
				self.insert(item, tk.END, p)
			if open_ is not None:
				self.item(item, open=open_)
		elif isinstance(run, RunInfo):
			item = super().insert(parent, index,
						#f"{parent}:{run.name} [{id(run)}]",
						text=run.inputBaseName2(),
						values=len(run.family) if run.family else "",
						image=icons.get(run.mode()),
						tag=run("status"),
						**kw)
			self.item2info[item] = run
			if run.family:
				for child in run.family:
					chitem = super().insert(item, tk.END,
								#f"{item}:{child.name} [{id(child)}]",
								text=child.inputBaseName2(),
								image=icons.get(run.mode()),
								tag=child("status"),
								**kw)
					self.item2info[chitem] = child
				self.item(item, open=run.family[0]("open"))
		else:
			item = super().insert(parent, index, **kw)
			self.item2info[item] = run
		return item

	# ----------------------------------------------------------------------
	# fill run list
	# ----------------------------------------------------------------------
	def fill(self):
		#print("*-* FILL")
		#import traceback; traceback.print_stack()
		self.scanOpen()
		#focused  = self.item2info.get(self.focus())
		selected = [self.item2info.get(x) for x in self.selection()]

		super().delete(*self.get_children())
		self.item2info.clear()
		self.insert("", tk.END, self.project.runs._tree)

		selected = [self.info2item(x) for x in selected if self.info2item(x)]

		if selected:
			self.selection_set(selected)
			self.focus(selected[0])
			self.see(selected[0])
		else:
			try:
				first = self.get_children()[0]
				self.selection_set(first)
				self.focus(first)
				self.see(first)
			except IndexError:
				pass

		for tag,color in RunInfo.statusColor.items():
			self.tag_configure(str(tag), foreground=RunInfo.statusColor[tag])
			try:
				self.tag_configure(str(tag),
					image=icons.get(RunInfo.statusIcon[tag]))
			except KeyError:
				pass

	# ----------------------------------------------------------------------
	# @return selected visible only runs
	# ----------------------------------------------------------------------
	def selectedRuns(self, include_family=False):
		if not include_family:
			return [self.item2info[x] for x in self.selection()
					if self.bbox(x) and isinstance(self.item2info[x], RunInfo)]

		runs = []
		for item in self.selection():
			run = self.item2info[item]
			if not self.bbox(item): continue
			if not isinstance(run, RunInfo): continue
			runs.append(run)
			if run.family and not self.item(item,"open"):
				for child in run.family:
					runs.append(child)
		return runs

	# ----------------------------------------------------------------------
	# @return focused run info
	# ----------------------------------------------------------------------
	def focusedRun(self):
		return self.item2info.get(self.focus())

	# ----------------------------------------------------------------------
	# @return directory for focused element
	# ----------------------------------------------------------------------
	def focusedDir(self):
		run = self.focusedRun()
		if not run:
			return self.project.dir()
		elif isinstance(run,GroupInfo):
			return "/".join(self.path(self.focus()))
		else:
			return run.inputDirName()

	# ----------------------------------------------------------------------
	def add(self, event=None):
		self.page.add()

	# ----------------------------------------------------------------------
	# Delete items and update dictionaries
	# ----------------------------------------------------------------------
	def delete(self, *items):
		for item in reversed(items):
			run = self.item2info[item]
			if isinstance(run,RunInfo) and run.isdefault(): continue
			if isinstance(run,GroupInfo) or run("parent")=="":
				super().delete(item)
				del self.item2info[item]

	# ----------------------------------------------------------------------
	# Update tree and add undo structure
	# ----------------------------------------------------------------------
	def addUndo(self):
		self.page.undoredo.add(self.project.runs.setUndo(self.items()))
		self.project.modified()

	# ----------------------------------------------------------------------
	# Add a run
	# ----------------------------------------------------------------------
	def addRun(self):
		focused = self.focus()
		run  = RunInfo(self.project,"")
		item = super().add(run)
		edit = tkExtra.InPlaceFile(self,
					item=item,
					#initialdir=run.dirname(),
					title=f"Rename run '{run.name}' to",
					filetypes=inputs.filetypes(run.mode()))
		#---------------------------------------------------------------
		def _delete():
			self.delete(item)
			self.focus(focused)
			self.selection_set(focused)
			self.see(focused)
		#---------------------------------------------------------------
		if edit.value is None:
			_delete()
			return False

		edit.value = edit.value.strip()
		if not edit.value:
			_delete()
			return False

		n,e  = os.path.splitext(edit.value)
		name = self.project.relpath(os.path.join(run.dirname(), n))
		self.item2info[item] = run
		dirname = os.path.dirname(n)
		self.item(item, text=run.basename())
		run.basename(name)
		if dirname:
			run["open"] = True
			parent = self.parent(item)
			index  = self.index(item)
			self.detach(item)
			dirname = dirname.split(os.sep)
			if dirname[0]=="":	# for absolute paths use the / infront of first dir
				dirname.pop(0)
				dirname[0] = os.sep + dirname[0]
			for d in dirname:
				group  = GroupInfo(d)
				parent = self.insert(parent, index, group, open=True)
				index  = 0
			self.reattach(item, parent, index)
		self.addUndo()
		return True

	# ----------------------------------------------------------------------
	# Add a group entry
	# ----------------------------------------------------------------------
	def addGroup(self, event=None):
		if super().addGroup():
			self.addUndo()

	# ----------------------------------------------------------------------
	def refreshRun(self, event=None):
		self.page.refreshRun()

	# ----------------------------------------------------------------------
	def remove(self, event=None):
		self.page.delete()

	# ----------------------------------------------------------------------
	def clone(self, event=None):
		self.page.clone()

	# ----------------------------------------------------------------------
	def executeRun(self, event=None):
		self.page.executeRun()

	# ----------------------------------------------------------------------
	# Update run status color and image
	# ----------------------------------------------------------------------
	def updateStatus(self, item=""):
		if item != "":
			run = self.item2info[item]
			if isinstance(run, RunInfo):
				found = False
				for tag in self.item(item, "tags"):
					if str(run("status")) != tag:
						self.tag_remove(tag, item)
					else:
						found = True
				if not found:
					self.tag_add(run("status"), item)

		for child in self.get_children(item):
			run = self.item2info[child]
			if isinstance(run, RunInfo):
				found = False
				for tag in self.item(item, "tags"):
					if str(run("status")) != tag:
						self.tag_remove(tag, child)
					else:
						found = True
				if not found:
					self.tag_add(run("status"), child)
			self.updateStatus(child)

	# ----------------------------------------------------------------------
	# Refresh runs status and update colors if needed
	# ----------------------------------------------------------------------
	def refreshRuns(self):
		changed = False
		#---------------------------------------------------------------
		def setRunStatus(run, status):
			nonlocal changed
			if run("status") != status:
				run["status"] = status
				changed = True

		# update colors on listbox
		if self.project.runs.refresh(): self.updateStatus()
		return changed

	# ----------------------------------------------------------------------
	# Attach to selected runs
	# ----------------------------------------------------------------------
	def attach(self, event=None):
		self.page.busy()
		for run in self.selectedRuns():
			if not isinstance(run, RunInfo): continue
			run.wait2Attach()
		self.page.notBusy()

	# ----------------------------------------------------------------------
	# Hitting F2 rename in place
	# ----------------------------------------------------------------------
	def editInPlace(self, event=None):
		#print("rename",event)
		if event and event.state & STATE.CONTROL: return

		if event is not None:
			column = self.identify_column(event.x)
		else:
			column = "#0"

		sel = self.selection()
		if not sel: return "break"

		if column == "#0":
			if len(sel) != 1: return "break"
			self.focus(sel[0])
			return self.rename()

		elif column == "#1":
			return self.editSpawn()

	# ----------------------------------------------------------------------
	# Rename run
	# ----------------------------------------------------------------------
	def rename(self, event=None):
		active = self.focus()
		run    = self.item2info.get(active)
		if isinstance(run, RunInfo):
			if run.isdefault(): return "break"
			if run("parent"): return "break"
			if run("status") in (RunInfo.STATUS_WAIT2ATTACH, RunInfo.STATUS_RUNNING):
				return "break"
			#path = run.dirname()
			edit = tkExtra.InPlaceFile(self,
						item=active,
						#initialdir=path,
						title=f"Rename run '{run.name}' to",
						filetypes=inputs.filetypes(run.mode()))

			if edit.value is None: return "break"
			edit.value = edit.value.strip()
			if not edit.value:
				self.item(active, text=run.basename())
			elif edit.value != run.basename():
				n,e  = os.path.splitext(edit.value)
				name = self.project.relpath(os.path.join(run.dirname(), n))
				dirname = os.path.dirname(n)
				if dirname:
					run["open"] = True
					parent = self.parent(active)
					index  = self.index(active)
					self.detach(active)
					dirname = dirname.split(os.sep)
					if dirname[0]=="":	# for absolute paths use the / infront of first dir
						dirname.pop(0)
						dirname[0] = os.sep + dirname[0]
					for d in dirname:
						group  = GroupInfo(d)
						parent = self.insert(parent, index, group, open=True)
						index  = 0
					self.reattach(active, parent, index)
					self.item(active, text=run.basename())
					self.addUndo()
				else:
					self.renameRun(run, name)
					self.flair.refresh("run")
					self.flair.modified()
		else:
			oldpath  = "/".join(self.path(active))
			edit = tkExtra.InPlaceEdit(self, item=active)
			if edit.value is not None and edit.value != run.name:
				run.name = edit.value
				newpath  = "/".join(self.path(active))
				for i, run in enumerate(self.project.runs()):
					if isinstance(run,RunInfo):
						group = run.dirname()
						if not group:
							continue
						elif group==oldpath:
							run.dirname(newpath)
						elif group.startswith(oldpath+"/"):
							run.dirname(newpath + group[len(oldpath):])
				self.flair.modified()
				self.flair.refresh("run")
		return "break"

	# ----------------------------------------------------------------------
	def editSpawn(self, event=None):
		active = self.focus()
		# Number of spawned runs has been edited
		try:
			oldSpawn = int(self.item(active,"values")[0])
			if oldSpawn == 1: oldSpawn = 0
		except:
			oldSpawn = 0

		edit = tkExtra.InPlaceEdit(self, item=active, column=0)
		if edit.value is None: return "break"
		try:
			spawn = int(edit.value)
			if spawn == 1: spawn = 0
		except:
			spawn = 0
		if spawn == oldSpawn: return "break"

		for item in reversed(self.selection()):
			run = self.item2info.get(item)
			if not isinstance(run, RunInfo): continue
			if run("parent") or \
			   run("status") in (RunInfo.STATUS_WAIT2ATTACH, RunInfo.STATUS_RUNNING):
				continue

			# Remove additional runs
			del run.family[spawn:]

			# Fill runs
			for i in range(len(run.family),spawn):
				name = run.spawnName(i+1)
				#if self.project.runs.getRunByName(name, run.name) is not None:
				#	continue
				child = RunInfo(self.project, name)
				child["parent"] = run.name
				child.setRandomSeed(run, i+1)
				run.family.append(child)
			if run.family:
				run.family[0]["open"] = True

		#self.addUndo()
		self.fill()
		self.flair.modified()
		return "break"

	# ----------------------------------------------------------------------
	# Rename a run
	# @param run	to rename
	# @param name	new name
	# ----------------------------------------------------------------------
	def renameRun(self, run, name):
		# Check if run is there
		newname = self.project.runs.uniqueName(name)
		if newname != name:
			self.flair.notify("Run already exists",
				f"Run '{name}' already exists. Renamed to '{newname}'",
				tkFlair.NOTIFY_WARNING)
			name = newname
		self.project.runs.rename(run,name)
		self.flair.refresh("run")
		self.flair.modified()

	# ----------------------------------------------------------------------
	def copy(self, event=None):
		self.selectBranches()
		# remove spawned children
		for item in self.selection():
			run = self.item2info.get(item)
			if run and isinstance(run,RunInfo) and run("parent"):
				self.selection_remove(item)
		# Copy to clipboard
		bio = io.BytesIO()
		pickler = pickle.Pickler(bio)
		pickler.dump(self.selectedItems())
		self.flair.clipboard_clear()
		self.flair.clipboardAppend(bio, tkFlair._RUN_MIME)
		return "break"

	# ----------------------------------------------------------------------
	def cut(self, event=None):
		self.copy()
		self.delete(*self.selection())
		self.addUndo()
		return "break"

	# ----------------------------------------------------------------------
	# Paste below current item
	# ----------------------------------------------------------------------
	def paste(self, event=None, below=True):
		if tkFlair._RUN_MIME in self.flair.clipboardTargets():
			unpickler = self.flair.clipboardUnpickler(tkFlair._RUN_MIME)
			runs = unpickler.load()
			#-------------------------------------------------------
			# re-parent all pasted runs before inserting in the tree
			# below
			#-------------------------------------------------------
			def _reparent(item):
				if isinstance(item,GroupInfo):
					for item in item.childs:
						_reparent(item)
				else:
					item.parent(self.project)
					for child in item.family:
						child.parent(self.project)
			_reparent(runs)
		else:
			self.flair.notify("Cannot paste",
					"Invalid content in clipboard",
					tkFlair.NOTIFY_WARNING)
			return "break"

		active = self.focus()
		self.selectClear()
		run = self.item2info.get(active)
		#print("RUN=",run.name)
		if run and isinstance(run,RunInfo) and run("parent"):
			# ignore spawned runs
			active = self.parent(active)
		if active is None:
			parent = ""
			index  = 1
		elif self.item(active, "open") and \
		     isinstance(run, GroupInfo):
			parent = active
			index  = 0
		else:
			parent = self.parent(active)
			index  = self.index(active)+int(below)
			if parent=="" and index==0: index=1

		#print("PARENT=",self.item(parent,"text"))
		#print("INDEX=",index)
		selected = []
		for run in runs.childs:
			selected.append(self.insert(parent, index, run))
			index += 1
		if selected:
			self.selection_set(selected)
			self.focus(selected[0])
			self.see(selected[0])
		self.addUndo()
		#print("<<< RUNS=",type(runs))
		#self.project.runs.dump(runs)
		return "break"

	# ----------------------------------------------------------------------
	# Paste runs above current run
	# ----------------------------------------------------------------------
	def pasteAbove(self, event=None):
		return self.paste(event, False)
