# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   06-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import tkinter as tk
from tkinter import messagebox

import undo
import icons
import Ribbon
import RunTab
import DataTab
import Project
import RunList
import tkFlair
import tkExtra
import FilesTab
import FlairRibbon

#===============================================================================
# Run Frame
#===============================================================================
class RunPage(FlairRibbon.FlairPage):
	"""Execute a simulation, inspect the output files and merge the generated data."""

	_name_ = "Run"
	_icon_ = "run"

	#----------------------------------------------------------------------
	def init(self):
		self.selected  = []		# remember selected runs for when saving the information
		self.undoredo  = undo.UndoRedo(100)
		self.refreshOn = ("save","run","exe","files","data")

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Tools ===========
		group = Ribbon.LabelGroup(self.ribbon, "View")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		kw = {
			"compound" : tk.TOP,
			"variable" : self.tabVar,
			"command"  : self.changeTab,
			"padx"     : 5
		}

		# ---
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Runs",
				image=icons.get("run32"),
				anchor=tk.CENTER,
				value=0,
				**kw)
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
		tkExtra.Balloon.set(b, "Run setting page")

		b = Ribbon.LabelRadiobutton(group.frame,
				text="Files",
				image=icons.get("file-manager32"),
				anchor=tk.CENTER,
				value=1,
				**kw)
		tkExtra.Balloon.set(b, "Output files inspecting page")
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)

		b = Ribbon.LabelRadiobutton(group.frame,
				text="Data",
				image=icons.get("data32"),
				anchor=tk.CENTER,
				value=2,
				**kw)
		tkExtra.Balloon.set(b, "Merge output data page")
		b.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)

	#----------------------------------------------------------------------
	# Create Project page
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage(False)

		self.splitter = tk.PanedWindow(self.frame,
					orient=tk.HORIZONTAL,
					showhandle=0,
					sashrelief=tk.RAISED,
					sashwidth=4,
					opaqueresize=1)
		self.splitter.pack(expand=tk.YES, fill=tk.BOTH)

		# ======= Runs ========
		frame = tk.Frame(self.splitter)
		self.splitter.add(frame, minsize=50,
				width=tkFlair.config.getInt(tkFlair._PAGE_SECTION,
						f"{self.name}.split", 250))
		self.runList = RunList.RunListbox(frame, self, height=5)
		self.runList.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)

		sb = tkExtra.AutoScrollbar(frame, orient=tk.VERTICAL, command=self.runList.yview)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		self.runList.config(yscrollcommand=sb.set)

		# Bind events
		self.runList.bind('<<TreeviewSelect>>',	self.select)
		self.runList.bind("<Insert>",		self.add)
		self.runList.bind("<Delete>",		self.delete)
		self.runList.bind("<<MoveUp>>",		self.moveUp)
		self.runList.bind("<<MoveDown>>",	self.moveDown)
		self.runList.bind("<<MoveLeft>>",	self.moveLeft)
		self.runList.bind("<<MoveRight>>",	self.moveRight)

		# ======== Tabs =======
		tabFrame = tk.Frame(self.splitter, padx=0, pady=0)
		self.splitter.add(tabFrame)

		# --------- Run tab ---------
		self.runtab = RunTab.RunTab(tabFrame, self)
		self.addTab(self.runtab)
		# --------- Files tab ---------
		self.addTab(FilesTab.FilesTab(tabFrame, self))
		# --------- Data tab ---------
		self.addTab(DataTab.DataTab(tabFrame, self))

		self.runList.set(self.flair)

	# ----------------------------------------------------------------------
	def canUndo(self):	return self.undoredo.canUndo()
	def canRedo(self):	return self.undoredo.canRedo()
	def resetUndo(self):	self.undoredo.reset()

	# ----------------------------------------------------------------------
	def undo(self, event=None):
		if self.canUndo():
			self.undoredo.undo()
			self.refresh()

	# ----------------------------------------------------------------------
	def redo(self, event=None):
		if self.canRedo():
			self.undoredo.redo()
			self.refresh()

	# ----------------------------------------------------------------------
	def select(self, event=None):
		if self.activeTab is not None:
			self.page.after_idle(self.activeTab.activate)
		self.setValid()

	# ----------------------------------------------------------------------
	def get(self):
		self.runList.scanOpen()
		self.runtab.get()

	# ----------------------------------------------------------------------
	def selectAll(self):
		if self.page.focus_get() is not self.runList:
			return super().selectAll()
		else:
			self.runList.selectAll()
		return "break"

	# ----------------------------------------------------------------------
	def selectClear(self):
		if self.page.focus_get() is not self.runList:
			return super().selectClear()
		else:
			self.runList.selectClear()
		return "break"

	# ----------------------------------------------------------------------
	def selectInvert(self):
		if self.page.focus_get() is not self.runList:
			return super().selectInvert()
		else:
			self.runList.selectInvert()
		return "break"

	# ----------------------------------------------------------------------
	# Activate and set focus
	# ----------------------------------------------------------------------
	#def activate(self):
	#	super().activate()

	# ----------------------------------------------------------------------
	# Update run progress information
	# and Enable or Disable the fields
	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		super().refresh(what)
		if self.project is None: return

		# Refresh run listbox
		self.runtab.fillQueueList()	# If list is changed from ConfigDialog

		# Run list
		self.runList.refreshRuns()
		#self.runList.updateSelection()
		self.runList.fill()
		self.select()

	# ----------------------------------------------------------------------
	def refreshStatus(self):
		super().refreshStatus()
		if self.activeTab: self.activeTab.refreshStatus()

	# ----------------------------------------------------------------------
	def reset(self):
		self.selected = []

	# ----------------------------------------------------------------------
	def beforeDock(self):
		self.reset()

	# ----------------------------------------------------------------------
	def clean(self):
		self.activeTab.clean()

	# ----------------------------------------------------------------------
	def copy(self, event=None):
		self.get()
		if self.page.focus_get() is self.runList:
			return self.runList.copy(event)

	# ----------------------------------------------------------------------
	def cut(self, event=None):
		self.get()
		if self.page.focus_get() is self.runList:
			return self.runList.cut(event)

	# ----------------------------------------------------------------------
	# Paste below current item
	# ----------------------------------------------------------------------
	def paste(self, event=None):
		self.get()
		if self.page.focus_get() is self.runList:
			return self.runList.paste(event)

	# ----------------------------------------------------------------------
	# Move up select items by one
	# ----------------------------------------------------------------------
	def moveUp(self, event=None):
		self.get()
		self.runList.moveUp(True)
		self.undoredo.add(self.project.runs.setUndo(self.runList.items()))
		return "break"

	# ----------------------------------------------------------------------
	# Move down select items by one
	# ----------------------------------------------------------------------
	def moveDown(self, event=None):
		self.get()
		self.runList.moveDown(True)
		self.undoredo.add(self.project.runs.setUndo(self.runList.items()))
		return "break"

	# ----------------------------------------------------------------------
	# Move left (up one parent) selected items
	# ----------------------------------------------------------------------
	def moveLeft(self, event=None):
		self.get()
		self.runList.moveLeft(True)
		self.undoredo.add(self.project.runs.setUndo(self.runList.items()))
		return "break"

	# ----------------------------------------------------------------------
	# Move right (down one parent if next is group)
	# ----------------------------------------------------------------------
	def moveRight(self, event=None):
		self.get()
		self.runList.moveRight(True)
		self.undoredo.add(self.project.runs.setUndo(self.runList.items()))
		return "break"

	# ----------------------------------------------------------------------
	# Add new run
	# ----------------------------------------------------------------------
	def add(self, event=None):
		self.get()
		if self.runList.addRun():
			self.flair.refresh("run")
			self.runtab.selectDefaultDefines()

	# ----------------------------------------------------------------------
	def addGroup(self, event=None):
		self.get()
		self.runList.addGroup()

	# ----------------------------------------------------------------------
	# Delete runs
	# ----------------------------------------------------------------------
	def delete(self, event=None):
		#if self.page.focus_get() is not self.runList:
		#	return super().delete(event)
		self.get()	# get all information before pushing into undo

		selection = self.runList.selection()
		active    = self.runList.focus()
		path      = self.runList.path(active)
		idx       = [self.runList.index(x) for x in self.runList.dir(active)]
		self.runList.delete(*selection)
		self.runList.focus_set()

		# find item to select
		parent = ""
		child = None
		for i,p in zip(idx,path):
			try:
				child = self.runList.get_children(parent)[i]
				if self.runList.item(child, "text") != p:
					break
				parent = child
			except IndexError:
				while True:
					try:
						child = self.runList.get_children(parent)[-1]
						break
					except IndexError:
						if parent=="": break
						parent = self.runList.parent(parent)
		if child is not None:
			self.runList.focus(child)
			self.runList.selection_set(child)
			self.runList.see(child)

		self.selected = None
		self.undoredo.add(self.project.runs.setUndo(self.runList.items()))
		self.modified()
		self.select()
		self.flair.refresh("run")
		return "break"

	# ----------------------------------------------------------------------
	# Clone runs
	# ----------------------------------------------------------------------
	def clone(self, event=None):
		self.get()
		#if self.page.focus_get() is not self.runList: return super().clone(event)
		if not self.runList.selection(): return
		for item in reversed(self.runList.selection()):
			run = self.runList.item2info.get(item)
			if isinstance(run, Project.RunInfo):
				self.runList.selection_remove(item)
				run = run.clone()
				run.name = self.project.runs.uniqueName(run.name)
				index  = self.runList.index(item)
				parent = self.runList.parent(item)
				new    = self.runList.insert(parent, index+1, run)
				self.runList.selection_add(new)
				self.runList.focus(new)
			#else:
			#	breakpoint()
		self.undoredo.add(self.project.runs.setUndo(self.runList.items()))
		self.modified()
		self.select()
		self.flair.refreshGeometry()
		return "break"

	# ----------------------------------------------------------------------
	def executeRun(self):
		self.runtab.execute()

	# ----------------------------------------------------------------------
	def refreshRun(self):
		self.runtab.refresh()

	# ----------------------------------------------------------------------
	# Loop runs
	# ----------------------------------------------------------------------
	def loop(self):
		runs = []
		for run in self.runList.selectedRuns():
			if run("parent"): continue # ignore child nodes
			runs.append(run)
		if not runs: return
		if LoopRunsDialog(self.page, self.flair, runs, self.runList).show():
			self.undoredo.add(self.project.runs.setUndo(self.runList.items()))
			self.flair.refresh("run")

	# ----------------------------------------------------------------------
	# Rename selected runs
	# ----------------------------------------------------------------------
	def rename(self, event=None):
		self.get()
		self.runList.rename()

	# ----------------------------------------------------------------------
	# Return a list of the selected parents only
	# ----------------------------------------------------------------------
	def _selectedParentsList(self):
		runs = []
		for run in self.runList.selectedRuns():
			if run.isdefault(): continue
			if run("parent"): continue # ignore child nodes
			runs.append(run)
			if run.family:
				# Append all its family
				for child in run.family:
					runs.append(child)
		return runs

	# ----------------------------------------------------------------------
	def help(self):
		if str(self.page.focus_get()).startswith(str(self.runList)):
			super().help()
		else:
			self.activeTab.help()

#===============================================================================
# Create clone of Runs by looping on one variable
#===============================================================================
class LoopRunsDialog(tk.Toplevel):
	def __init__(self, master, flair, runs, runlist):
		super().__init__(master)
		self.transient(master)
		self.title("Loop runs")

		self.flair   = flair
		self.runs    = runs
		self.runList = runlist

		self.incparent = tk.BooleanVar(self)

		l = tk.Label(self, text="Create clone of the selected runs by\n"
					"looping the following variable over\n"
					"the defined range",
				justify=tk.LEFT,
				anchor=tk.NW,
				pady=2)
		l.pack(side=tk.TOP)

		frame = tk.Frame(self)
		frame.pack(fill=tk.BOTH)

		# ---
		row = 0
		tk.Label(frame, text="Variable:").grid(
			row=row, column=0, sticky=tk.E)

		self.variable = tkExtra.Combobox(frame, width=16)
		self.variable.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.variable,"Select the variable to loop")

		# Update defines
		lst = []
		for card in self.flair.project.input.cardsSorted("#define",2).values():
			if card._indent != 0: continue
			lst.append(card.name())
		self.variable.fill(lst)

		# ---
		row += 1
		tk.Label(frame, text="Start value:").grid(
			row=row, column=0, sticky=tk.E)
		self.start = tk.Entry(frame, background="White", justify=tk.LEFT)
		self.start.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.start, "Enter starting value or expression")

		# ---
		row += 1
		tk.Label(frame, text="End value:").grid(
			row=row, column=0, sticky=tk.E)
		self.stop = tk.Entry(frame, background="White", justify=tk.LEFT)
		self.stop.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.stop, "Enter ending value or expression")

		# ---
		row += 1
		tk.Label(frame, text="Step:").grid(
			row=row, column=0, sticky=tk.E)
		self.step = tk.Entry(frame, background="White", justify=tk.LEFT)
		self.step.grid(row=row, column=1, sticky=tk.EW)
		tkExtra.Balloon.set(self.step, "Step value or expression")

		# ---
		row += 1
		b = tk.Checkbutton(frame, text="Use Parent as 1st iteration",
				variable=self.incparent,
				onvalue=1, offvalue=0,
				anchor=tk.W)
		tkExtra.Balloon.set(b,
			"Check to use the parent run as the first iteration of the loop.\n"
			"Uncheck to create a new run (suffix -1) for the first iteration")
		b.grid(row=row, column=0, columnspan=2, sticky=tk.W)

		for run in runs:
			if run.isdefault():
				b.config(state=tk.DISABLED)
				break
		frame.grid_columnconfigure(1, weight=1)

		# ----
		frame = tk.Frame(self)
		tk.Button(frame, text="Cancel",
				image=icons.get("x"),
				compound=tk.LEFT,
				command=self.cancel).pack(side=tk.RIGHT)
		tk.Button(frame, text="Ok",
				image=icons.get("ok"),
				compound=tk.LEFT,
				command=self.ok).pack(side=tk.RIGHT)
		self.bind("<Return>",  self.ok)
		self.bind("<KP_Enter>",self.ok)
		self.bind("<Escape>",  self.cancel)
		frame.pack(side=tk.BOTTOM, fill=tk.X)
		self._ok = False

	# ----------------------------------------------------------------------
	# show dialog and return True if Ok is pressed
	# ----------------------------------------------------------------------
	def show(self):
		self.deiconify()
		# Get the focus
		self.focus_set()
		self.grab_set()
		self.wait_window()
		return self._ok

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		var = self.variable.get()
		if var=="":
			messagebox.showerror("No variable selected",
				"Please select a valid variable first",
				parent=self)
			return

		inp = self.flair.project.input
		try:
			start = float(eval(self.start.get(), inp.globalDict, inp.localDict))
		except:
			messagebox.showerror("Error start value",
				"Please enter a valid starting value",
				parent=self)
			return

		try:
			stop = float(eval(self.stop.get(), inp.globalDict, inp.localDict))
		except:
			messagebox.showerror("Error stop value",
				"Please enter a valid stopping value",
				parent=self)
			return

		step = self.step.get()
		if step == "": step="1"
		try:
			step = float(eval(step, inp.globalDict, inp.localDict))
		except:
			messagebox.showerror("Error step value",
				"Please enter a valid step value",
				parent=self)
			return

		if step==0.0: step = 1.0

		if start<stop and step<0.0:
			messagebox.showerror("Wrong step value",
				"Start<Stop and step<0. Please change sign of step",
				parent=self)
			return
		elif start>stop and step>0.0:
			messagebox.showerror("Wrong step value",
				"Start>Stop and step>0. Please change sign of step",
				parent=self)
			return

		# Create runs
		val = start
		cnt = 1
		if step > 0.0:
			while val <= stop:
				self.createRuns(var, val, cnt)
				val += step
				cnt += 1

		else:
			while val >= stop:
				self.createRuns(var, val, cnt)
				val += step
				cnt += 1

		self.flair.modified()
		self._ok = True
		self.destroy()

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		self._ok = False
		self.destroy()

	# ----------------------------------------------------------------------
	def createRuns(self, var, val, cnt):
		try:
			f = float(val)
			val = str(int(f)) if float(int(f))==f else str(f)
		except:
			val = str(val)
		project = self.flair.project
		for i,run in enumerate(self.runs):
			if cnt==1 and self.incparent.get():
				clone = run
			else:
				clone = run.clone()
				if run.isdefault():
					clone.name = run.inputBaseName()
					clone.defaultDefines()
				# FIXME uniqueName is not working properly since the RunTree
				# is not updated until the end of the loop
				clone.name = project.runs.uniqueName(f"{clone.name}-{cnt}")

				if clone("title"):
					title = clone("title")
				else:
					title = project.title
				clone["title"] = f"{title} [{var}={val}]"

			# Check defines and append the var,value
			for j,nv in enumerate(clone.defineList):
				if nv[0]==var:
					del clone.defineList[j]
					break
			clone.defineList.append((var,val))

			item   = self.runList.info2item(run)
			index  = self.runList.index(item)
			parent = self.runList.parent(item)
			self.runList.insert(parent, index+cnt, clone)
