#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch,
# Date:   02-Jul-2008

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import tkinter as tk

import Manual
import tkExtra
import BasePlot

#===============================================================================
# Double Differential Plot Frame (mainly for USRBDX)
#===============================================================================
class Usr2DPlotFrame(BasePlot.BasePlotFrame):
	def __init__(self, master, flair):
		super().__init__(master, flair)

		# === File & Detector ===
		lblframe = tkExtra.ExLabelFrame(self, text="Detector ")
		lblframe.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=3)
		frame = lblframe.frame

		row, col = 0, 0
		tk.Label(frame, text="File:", anchor=tk.E).grid(row=row,
				column=col, sticky=tk.E)
		col += 1
		self.datafile = tkExtra.FileEntry(frame,
					title="Load USRxxx file",
					initialdir=self.project.dir(),
					filetypes=[
						("Fluka _tab.lis files","*_tab.lis"),
						("All","*")],
					foreground="DarkBlue", background="White")
		self.datafile.bind("<<FileChanged>>", self.load)
		tkExtra.Balloon.set(self.datafile, "USRBDX tab.lis output file")
		self.datafile.grid(row=row, column=col, columnspan=2, sticky=tk.EW)
		self.widget["datafile"] = self.datafile

		col += 2
		tk.Label(frame, text="Det:", anchor=tk.W).grid(row=row,
				column=col, sticky=tk.W, pady=0)
		col += 1
		self.det = tkExtra.Combobox(frame, width=10)
		self.det.grid(row=row, column=col, sticky=tk.EW, pady=0)
		self.widget["@det"] = self.det

		frame.grid_columnconfigure(1, weight=1)
		frame.grid_columnconfigure(4, weight=1)

		self._datafile = None
		self.bindToggle()

	# ----------------------------------------------------------------------
	def type(self): return "USR-2D"

	# ----------------------------------------------------------------------
	# Set plot
	# ----------------------------------------------------------------------
	def set(self, plot):
		super().set(plot)
		if plot is None: return

		name = self.project.name()
		self.datafile.initialdir = self.project.dir()
		self.datafile.setFiletypes([
					("Fluka _tab.lis files","*_tab.lis"),
					(f"{name} _tab.lis files",f"{name}*_tab.lis"),
					("All","*")])
		self._datafile = plot("datafile")
		self.loadDataFile(self._datafile)
		try: det = int(plot("det"))
		except: det = 1
		self.detector(det)

	# ----------------------------------------------------------------------
	def get(self, plot):
		super().get(plot)
		if plot is None: return

		if self.det.cget("state") != tk.DISABLED:
			det = self.detector()
			if det>=0: plot["det"] = str(det)

	# ----------------------------------------------------------------------
	def detector(self, val=None):
		if val is None:
			try:
				# Return the value
				return int(self.det.get().split()[0])-1
			except:
				return 0
		else:
			self.det.set("")
			if val >= 0:
				val1 = val + 1
				for item in self.det.get(0,tk.END):
					if int(item.split()[0])==val1:
						self.det.set(item)

	# ----------------------------------------------------------------------
	def loadDataFile(self, datafile):
		# --- Load a _tab.lis file ---
		if not datafile:
			self._datafile = None
			return

		# Load detector file
		self.det.clear()
		cnt   = 1
		first = None
		half  = 0
		name  = ""
		try:
			with open(datafile, "r", encoding="UTF-8", errors="replace") as f:
				for line in f:
					if line=="\n":
						half += 1
						if half == 2:
							half = 0
							cnt += 1

					elif line.startswith(" # Detector n:"):
						name = line[17:]
						p = name.find("(")
						if p>0: name = name[:p]
						name = f"{cnt+1} {name.strip()}"

					elif line.startswith(" # double differential"):
						self.det.insert(tk.END, name)
						if not first: first = name

					else:
						half = 0
			if first: self.det.set(first)
		except (IOError, FileNotFoundError):
			pass

	# ----------------------------------------------------------------------
	def load(self, event=None):
		filename = self.datafile.filename()
		if filename:
			self._datafile = self.project.relpath(filename)
			self.datafile.set(self._datafile)
			self.loadDataFile(self._datafile)

	# ----------------------------------------------------------------------
	def datafileFocusOut(self, event=None):
		if self._datafile == self.datafile.get(): return
		self._datafile = self.project.relpath(self.datafile.get())
		self.datafile.delete(0,tk.END)
		self.datafile.insert(0,self._datafile)
		self.loadDataFile(self._datafile)

	# ----------------------------------------------------------------------
	def help(self):
		Manual.show("","2D Plot")
