# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   05-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import tkinter as tk

import tkFlair
import tkExtra

import log
import Plot
import Data
import Manual
import BasePlot
import PlotEngine

from fluka import FlukaParticle, FlukaLayout

_format = "%.7g"

#-------------------------------------------------------------------------------
# range of floating point numbers
#-------------------------------------------------------------------------------
def binrange(start,stop,step):
	"""formatted number range"""
	if abs(start)<1e-9: start = 0.0
	if abs(stop)<1e-9:   stop = 0.0
	lst = [""]
	try:
		n = int(round((stop-start)/step))
	except (ZeroDivisionError, OverflowError):
		return lst

	# set an upper limit to avoid huge lists
	for i in range(min(n+1,10000)):
		x = start + float(i)*step
		if abs(x)<1e-9:
			lst.append("0.")
		else:
			num = _format%(x)
			if "." not in num and "e" not in num: num += "."
			lst.append(num)
	return lst

#===============================================================================
# USRBIN Data Plot Frame
#===============================================================================
class UsrbinPlotFrame(BasePlot.BasePlotFrame):
	def __init__(self, master, flair):
		super().__init__(master, flair)

		# === File & Detector ===
		frame = tk.LabelFrame(self, text="Binning Detector ")
		frame.pack(side=tk.TOP, expand=tk.YES, fill=tk.X, padx=3)

		row, col = 0, 0
		tk.Label(frame, text="File:").grid(row=row,
				column=col, sticky=tk.E)
		col += 1; span = 3
		self.datafile = tkExtra.FileEntry(frame,
					title="Load USRBIN file",
					initialdir=self.project.dir(),
					filetypes=[
						("Fluka USRBIN files",("*usrbin*","*.bnn")),
						("Fluka fort files",("*_ftn.*","*_fort.*")),
						("All","*")],
						foreground="DarkBlue", background="White")
		self.datafile.bind("<<FileChanged>>", self.load)
		tkExtra.Balloon.set(self.datafile, "USRBIN binary output file")
		self.datafile.grid(row=row, column=col, columnspan=span, sticky=tk.EW)
		self.widget["datafile"] = self.datafile

		# ---
		col += span+1
		tk.Label(frame, text="Title:", anchor=tk.E).grid(row=row,
				column=col, sticky=tk.E)
		col += 1
		self.usr_title = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_title.grid(row=row, column=col, columnspan=3, sticky=tk.EW)

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="Cycles:", anchor=tk.E).grid(row=row,
				column=col, sticky=tk.E)
		col += 1
		self.usr_cycles = tk.Label(frame, anchor=tk.E,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_cycles.grid(row=row, column=col, sticky=tk.EW)

		# ---
		col += 1
		tk.Label(frame, text="Primaries:", anchor=tk.E).grid(row=row,
				column=col, sticky=tk.E)
		col += 1
		self.usr_ncase = tk.Label(frame, anchor=tk.E,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_ncase.grid(row=row, column=col, columnspan=2, sticky=tk.EW)

		# ---
		col += 2
		tk.Label(frame, text="Weight:", anchor=tk.E).grid(row=row,
				column=col, sticky=tk.E)
		col += 1
		self.usr_weight = tk.Label(frame, anchor=tk.E,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_weight.grid(row=row, column=col, sticky=tk.EW)

		# ---
		col += 1
		tk.Label(frame, text="Time:", anchor=tk.E).grid(row=row,
				column=col, sticky=tk.E)
		col += 1
		self.usr_time = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_time.grid(row=row, column=col, sticky=tk.EW)

		# ---
		frame.grid_columnconfigure(1, weight=1)	# <<<<-------
		frame.grid_columnconfigure(3, weight=1)
		frame.grid_columnconfigure(6, weight=1)
		frame.grid_columnconfigure(8, weight=5)

		# === Info ===
		frame = tk.LabelFrame(self, text="Binning Info ")
		frame.pack(expand=tk.YES, fill=tk.X, padx=3)

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="Det:", anchor=tk.W).grid(row=row,
				column=col, sticky=tk.W, pady=0)
		col += 1
		self.det = tkExtra.Combobox(frame, width=10, command=self.fillDetectorInfo)
		self.det.grid(row=row, column=col, sticky=tk.EW, pady=0)
		self.widget["@det"] = self.det

		# ---
		col += 1
		self.usr_x = tk.Label(frame, text="X:")
		self.usr_x.grid(row=row, column=col, sticky=tk.E, pady=1)
		col += 1
		self.usr_x_val = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_x_val.grid(row=row, column=col, sticky=tk.EW)

		# ---
		col += 1
		tk.Label(frame, text="Min:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.usr_min = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._CLABEL_FOREGROUND_COLOR,
				background=tkFlair._CLABEL_BACKGROUND_COLOR)
		self.usr_min.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(self.usr_min, "Current projection minimum value")

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="Type:", anchor=tk.W).grid(row=row,
				column=col, sticky=tk.W)
		col += 1
		self.usr_type = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_type.grid(row=row, column=col, sticky=tk.EW)

		# ---
		col += 1
		self.usr_y = tk.Label(frame, text="Y:")
		self.usr_y.grid(row=row,
				column=col, sticky=tk.E, pady=1)
		col += 1
		self.usr_y_val = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_y_val.grid(row=row, column=col, sticky=tk.EW)

		# ---
		col += 1
		tk.Label(frame, text="Max:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.usr_max = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._CLABEL_FOREGROUND_COLOR,
				background=tkFlair._CLABEL_BACKGROUND_COLOR)
		self.usr_max.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(self.usr_max, "Current projection maximum value")

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="Score:", anchor=tk.W).grid(row=row,
				column=col, sticky=tk.W)
		col += 1
		self.usr_score = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_score.grid(row=row, column=col, sticky=tk.EW)

		# ---
		col += 1
		self.usr_z = tk.Label(frame, text="Z:")
		self.usr_z.grid(row=row, column=col, sticky=tk.E, pady=1)
		col += 1
		self.usr_z_val = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._ILABEL_FOREGROUND_COLOR,
				background=tkFlair._ILABEL_BACKGROUND_COLOR)
		self.usr_z_val.grid(row=row, column=col, sticky=tk.EW)

		# ---
		col += 1
		tk.Label(frame, text="Int:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		self.usr_int = tk.Label(frame, anchor=tk.W,
				foreground=tkFlair._CLABEL_FOREGROUND_COLOR,
				background=tkFlair._CLABEL_BACKGROUND_COLOR)
		self.usr_int.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(self.usr_int, "Current projection integral value")

		# ---
		frame.grid_columnconfigure(1, weight=1)	# <<<<-------
		frame.grid_columnconfigure(3, weight=1)
		frame.grid_columnconfigure(5, weight=1)

		# === MultiFrame: Projection, plotting ... ===
		multiframe = tk.Frame(self)
		multiframe.pack(expand=tk.YES, fill=tk.BOTH)

		# === 1st Projection & Limits ===
		frame = tk.LabelFrame(multiframe, text="Projection & Limits ")
		frame.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X, padx=3)

		# ---
		row, col = 0, 0
		self.proj  = tk.StringVar()
		self.proj.set("Z")
		self.proj_x = tk.Radiobutton(frame, text="X",
				variable=self.proj, value='X')
		self.proj_x.grid(row=row, column=col, sticky=tk.W)

		col += 1
		self.x_min = tkExtra.Combobox(frame, label=True,
				width=8, command=self.xChanged)
		self.x_min._text.config(background="White")
		self.x_min.grid(row=row, column=col, sticky=tk.EW)
		self.widget["bxmin"] = self.x_min

		col += 1
		self.x_rebin = tk.IntVar()
		self.x_rebin.set(1)
		self.updating = False
		self.x_rebin.trace("w", self.xChanged)
		s = tk.Spinbox(frame, text=self.x_rebin, from_=1, to=1000,
				background="White",width=3)
		s.grid(row=row, column=col)
		tkExtra.Balloon.set(s, "Rebin factor")
		self.widget["xrebin"] = s
		self.variable["xrebin"] = self.x_rebin

		col += 1
		self.x_max = tkExtra.Combobox(frame, label=True, width=8)
		self.x_max._text.config(background="White")
		self.x_max.grid(row=row, column=col, sticky=tk.EW)
		self.widget["bxmax"] = self.x_max

		# ---
		col += 1
		b = tk.Button(frame, text="Get", padx=2, pady=0,
				command=self.get_limits)
		b.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Get limits from plot window")

		# ---
		row, col = row+1, 0
		self.proj_y = tk.Radiobutton(frame, text="Y",
				variable=self.proj, value='Y')
		self.proj_y.grid(row=row, column=col, sticky=tk.W)

		col += 1
		self.y_min = tkExtra.Combobox(frame, label=True,
				width=8, command=self.yChanged)
		self.y_min._text.config(background="White")
		self.y_min.grid(row=row, column=col, sticky=tk.EW)
		self.widget["bymin"] = self.y_min

		col += 1
		self.y_rebin = tk.IntVar()
		self.y_rebin.set(1)
		self.y_rebin.trace("w", self.yChanged)
		s = tk.Spinbox(frame, text=self.y_rebin, from_=1, to=1000,
				background="White",width=3)
		s.grid(row=row, column=col)
		tkExtra.Balloon.set(s, "Rebin factor")
		self.widget["yrebin"] = s
		self.variable["yrebin"] = self.y_rebin

		col += 1
		self.y_max = tkExtra.Combobox(frame, label=True, width=8)
		self.y_max._text.config(background="White")
		self.y_max.grid(row=row, column=col, sticky=tk.EW)
		self.widget["bymax"] = self.y_max

		# ---
		col += 1
		self.swap = tk.IntVar()
		self.swap.set(0)
		b = tk.Checkbutton(frame, text="swap", variable=self.swap)
		b.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(b, "Swap axes on binning data")
		self.widget["swap"] = b
		self.variable["swap"] = self.swap

		# ---
		row, col = row+1, 0

		self.proj_z = tk.Radiobutton(frame, text="Z",
				variable=self.proj, value='Z')
		self.proj_z.grid(row=row, column=col, sticky=tk.W)
		self.widget["proj"]   = (self.proj_x, self.proj_y, self.proj_z)
		self.variable["proj"] = self.proj

		col += 1
		self.z_min = tkExtra.Combobox(frame, label=True,
				width=8, command=self.zChanged)
		self.z_min._text.config(background="White")
		self.z_min.grid(row=row, column=col, sticky=tk.EW)
		self.widget["bzmin"] = self.z_min

		col += 1
		self.z_rebin = tk.IntVar()
		self.z_rebin.set(1)
		self.z_rebin.trace("w", self.zChanged)
		s = tk.Spinbox(frame, text=self.z_rebin, from_=1, to=1000,
				background="White",width=3)
		s.grid(row=row, column=col)
		tkExtra.Balloon.set(s, "Rebin factor")
		self.widget["zrebin"] = s
		self.variable["zrebin"] = self.z_rebin

		col += 1
		self.z_max = tkExtra.Combobox(frame, label=True, width=8)
		self.z_max._text.config(background="White")
		self.z_max.grid(row=row, column=col, sticky=tk.EW)
		self.widget["bzmax"] = self.z_max

		# ---
		col += 1
		var = tk.IntVar()
		var.set(0)
		b = tk.Checkbutton(frame, text="errors", variable=var)
		b.grid(row=row, column=col, sticky=tk.W)
		tkExtra.Balloon.set(b, "Show relative errors")
		self.widget["errors"] = b
		self.variable["errors"] = var

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="Norm:").grid(row=row, column=col, sticky=tk.E)
		col += 1; span=3
		b = tk.Entry(frame, background="White")
		tkExtra.Balloon.set(b, BasePlot._NORM_BALLOON)
		b.grid(row=row, column=col, columnspan=span, sticky=tk.EW)
		self.widget["norm"] = b

		# ---
		tkExtra.Balloon.set(self.x_min, "Project limit minimum")
		tkExtra.Balloon.set(self.y_min, "Project limit minimum")
		tkExtra.Balloon.set(self.z_min, "Project limit minimum")
		tkExtra.Balloon.set(self.x_max, "Project limit maximum")
		tkExtra.Balloon.set(self.y_max, "Project limit maximum")
		tkExtra.Balloon.set(self.z_max, "Project limit maximum")

		frame.grid_columnconfigure(1, weight=1)	# <<<<-------
		frame.grid_columnconfigure(3, weight=1)

		# === 2nd Tabs depending on type of plot ===

		frame = tk.Frame(multiframe)
		frame.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
		button_frame = tk.Frame(frame)
		button_frame.pack(side=tk.TOP, fill=tk.X)

		self.typeCombo = tkExtra.Combobox(button_frame,
			width=15, command=self.tabPageChange)
		self.typeCombo.pack(side=tk.RIGHT)
		tk.Label(button_frame, text="Type:").pack(side=tk.RIGHT)
		for i in Plot.USRBIN_HIST: self.typeCombo.insert(tk.END, i)
		self.widget["hist"] = self.typeCombo

		tab_frame = tk.Frame(frame, relief=tk.RAISED)
		tab_frame.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=tk.YES)

		self.frame2d = tk.Frame(tab_frame)
		self.frame2d.pack(expand=tk.YES, fill=tk.BOTH)

		# === Geometry ===
		frame = tk.LabelFrame(self.frame2d, text="Geometry ")
		frame.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X, padx=3)

		# ---
		row, col = 0, 0
		tk.Label(frame, text="Use:").grid(row=row, column=col, sticky=tk.E)

		col += 1
		b = tkExtra.Combobox(frame)
		b.fill((Plot.GEO_NO, Plot.GEO_AUTO))
		b.set(Plot.GEO_NO)
		b.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Geometry plot to superimpose to data")
		self.widget["geo"] = b

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="Pos:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		b = tkExtra.FloatEntry(frame, background="White", width=10)
		b.grid(row=row, column=col, sticky=tk.EW)
		tkExtra.Balloon.set(b,
			"Geometry position on the projecting plane to draw when Auto is selected")
		self.widget["geopos"] = b

		# ---
		row, col = row+1, 0
		tk.Label(frame, text="Axes:").grid(row=row, column=col, sticky=tk.E)
		col += 1
		b = tkExtra.Combobox(frame, width=5)
		b.fill(Plot.COORDAUTO)
		b.set(Plot.COORDAUTO[0])
		b.grid(row=row, column=col, sticky=tk.EW)
		self.widget["axes"] = b

		frame.grid_columnconfigure(1, weight=1)

		# === 2nd tab 1D Histogram ===
		self.frame1d = tk.LabelFrame(tab_frame, text="Options ")
		self.styleFrame(self.frame1d).pack(side=tk.LEFT,fill=tk.BOTH)

		# ---- variables -----
		self.inp  = []
		self.usr  = Data.Usrbin()
		self._datafile = None

		self.bindToggle()

	# ----------------------------------------------------------------------
	def type(self): return "USRBIN"

	# ----------------------------------------------------------------------
	# Show Frame
	# ----------------------------------------------------------------------
	def set(self, plot):
		try:
			cb = int(plot("cbtics"))
		except:
			cb = 0

		if cb==0 and (plot("cblabel")     or
				plot("cbmin")!="" or
				plot("cbmax")     or
				plot("hist")==Plot.USRBIN_HIST[0]):
			# Set cbtics if needed
			plot["cbtics"] = 1

		self.fillGeometry()

		super().set(plot)
		if plot is None: return

		self.updating = True

		# ---
		self.x_rebin.set(int(plot("xrebin")))
		self.y_rebin.set(int(plot("yrebin")))
		self.z_rebin.set(int(plot("zrebin")))

		# ---
		name = self.project.name()
		self.datafile.initialdir = self.project.dir()
		self.datafile.setFiletypes([	("Fluka USRBIN files",("*.bnn","*usrbin*")),
						(f"{name} .bnn files", f"{name}*.bnn"),
						("Fluka fort files",("*_ftn.*","*_fort.*")),
						("All","*")])
		self._datafile = plot("datafile")
		self.loadDataFile(self._datafile)
		try:
			d = int(plot("det"))
		except:
			d = 1
		self.detector(d)

		try:    pt = int(plot("pt.0")) % 14
		except: pt = 0
		self.plot_pt.set(PlotEngine.POINTS[pt])

		self.updating = False

		self.xChanged()
		self.yChanged()
		self.zChanged()
		self.afterPlot()

	# ----------------------------------------------------------------------
	def get(self, plot):
		super().get(plot)
		if plot is None: return

		self.updating = True
		if self.det.cget("state") != tk.DISABLED:
			det = self.detector()
			if det>=0: plot["det"] = str(det)

		if self.plot_pt["state"] != tk.DISABLED:
			plot["pt.0"] = PlotEngine.POINTS.index(self.plot_pt.get())
		self.updating = False

	# ----------------------------------------------------------------------
	def afterPlot(self):
		self.usr_min.config(text=self.plot("min"))
		self.usr_max.config(text=self.plot("max"))
		self.usr_int.config(text=self.plot("int"))

	# --------------------------------------------------------------------
	def xChanged(self, *abc): self.fillMaxLimit(0)
	def yChanged(self, *abc): self.fillMaxLimit(1)
	def zChanged(self, *abc): self.fillMaxLimit(2)

	# --------------------------------------------------------------------
	def fillMaxLimit(self, axis):
		if self.updating: return
		d = self.detector()
		if d<=0 or d>len(self.usr.detector): return
		det = self.usr.detector[d-1]

		if axis == 0:
			cmin   = self.x_min
			crebin = self.x_rebin
			cmax   = self.x_max
			dlow   = det.xlow
			dhigh  = det.xhigh
			#n      = det.nx
			dd     = det.dx
		elif axis == 1:
			cmin   = self.y_min
			crebin = self.y_rebin
			cmax   = self.y_max
			dlow   = det.ylow
			dhigh  = det.yhigh
			#n      = det.ny
			dd     = det.dy
		else:
			cmin   = self.z_min
			crebin = self.z_rebin
			cmax   = self.z_max
			dlow   = det.zlow
			dhigh  = det.zhigh
			#n      = det.nz
			dd     = det.dz

		try:
			vmin = float(cmin.get())
		except:
			cmax.fill(binrange(dlow, dhigh, dd))
			return

		assert dd > 0.0

		try:
			drebin = dd * float(crebin.get())
		except:
			drebin = dd

		# Round vmin to the next bin
		minbin = round((vmin - dlow)/dd)
		vmin = dlow + (minbin*dd + drebin)

		# Fill max list
		cmax.fill(binrange(vmin, dhigh, drebin))

		# Find closest max
		try:
			vmax   = float(cmax.get())
			maxbin = round((vmax - dlow)/dd)
		except:
			return

		if maxbin < minbin:
			cmax.set("")
		else:
			# Round vmax to closest re-bin
			try:
				cmax.set(_format%(round((vmax-vmin)/drebin)*drebin+vmin))
			except:
				pass

	# ----------------------------------------------------------------------
	def clearInfo(self, all_=True):
		if all_:
			self.usr_title.config(text="")
			self.usr_time.config(text="")
			self.usr_weight.config(text="")
			self.usr_ncase.config(text="")
			self.usr_cycles.config(text="")

		self.usr_type.config(text="")
		self.usr_score.config(text="")
		self.usr_x_val.config(text="")
		self.usr_y_val.config(text="")
		self.usr_z_val.config(text="")
		self.usr_min.config(text="")
		self.usr_max.config(text="")
		self.usr_int.config(text="")

	# ----------------------------------------------------------------------
	def loadDataFile(self, datafile):
		# --- Load USRBIN ---
		if not datafile:
			self._datafile = None
			self.usr.reset()
			self.clearInfo()
			self.det.clear()
			return

		datafile = self.project.abspath(datafile)
		try:
			self.usr.readHeader(datafile)
		except FileNotFoundError as ex:
			if self.winfo_ismapped():
				# FIXME shoud display a nice message
				self.flair.notify("USRBIN file not found",
					str(ex),
					tkFlair.NOTIFY_ERROR)
				log.error(ex)
			self.clearInfo()
			self.det.clear()
		except (IOError, MemoryError) as ex:
			if self.winfo_ismapped():
				# FIXME shoud display a nice message
				self.flair.notify("Error loading USRBIN",
					f"Error loading USRBIN file: {ex}",
					tkFlair.NOTIFY_ERROR)
				log.error(ex)
			self.clearInfo()
			self.det.clear()
			return

		self.usr_title.config(text=self.usr.title)
		self.usr_time.config(text=self.usr.time)
		self.usr_weight.config(text=self.usr.weight)
		self.usr_ncase.config(text=self.usr.ncase)
		self.usr_cycles.config(text=self.usr.nbatch)

		# Populate the detector menu
		self.det.delete(0,tk.END)
		for i,det in enumerate(self.usr.detector,1):
			self.det.insert(tk.END, f"{i} {det.name}")

	# --------------------------------------------------------------------
	def fillDetectorInfo(self):
		# --- Display Detector ---
		d = self.detector()
		if d<=0 or d>len(self.usr.detector):
			self.clearInfo(False)
			return

		det = self.usr.detector[d-1]
		for (n,i) in FlukaLayout.USRBIN_type:
			if i==det.type:
				self.usr_type.config(text=f"{i}: {n}")
				break
		self.usr_score.config(text=FlukaParticle.convert(det.score))

		b10 = det.type % 10
		if   b10 in (1, 7):
			labels = ("R:", "Φ:", "Z:")
		elif b10==2:
			labels = ("R1:", "R2:", "R3:")
		elif b10==8:
			labels = ("I1:", "I2:", "I3:")
		else:
			labels = ("X:", "Y:", "Z:")

		self.usr_x.config(text=labels[0])
		self.usr_y.config(text=labels[1])
		self.usr_z.config(text=labels[2])

		self.proj_x.config(text=labels[0])
		self.proj_y.config(text=labels[1])
		self.proj_z.config(text=labels[2])

		self.usr_x_val.config(text=f"[{det.xlow} .. {det.xhigh}] x {det.nx} ({det.dx})")
		self.usr_y_val.config(text=f"[{det.ylow} .. {det.yhigh}] x {det.ny} ({det.dy})")
		self.usr_z_val.config(text=f"[{det.zlow} .. {det.zhigh}] x {det.nz} ({det.dz})")

		# Fill Comboboxes
		self.x_min.fill(binrange(det.xlow, det.xhigh, det.dx))
		self.y_min.fill(binrange(det.ylow, det.yhigh, det.dy))
		self.z_min.fill(binrange(det.zlow, det.zhigh, det.dz))

		self.fillMaxLimit(0)
		self.fillMaxLimit(1)
		self.fillMaxLimit(2)

		self.selectProjection(det)

	# --------------------------------------------------------------------
	# Override the setState to correct for the correct projection
	# --------------------------------------------------------------------
	def setState(self, state):
		super().setState(state)
		if state!=tk.DISABLED:
			d = self.detector()
			if d<=0 or d>len(self.usr.detector): return
			self.selectProjection(self.usr.detector[d-1])

	# --------------------------------------------------------------------
	def selectProjection(self, det):
		page = self.typeCombo.get()

		self.proj_x.config(state=tk.NORMAL)
		self.proj_y.config(state=tk.NORMAL)
		self.proj_z.config(state=tk.NORMAL)

		# Check how many bins are requested
		# and active tab
		if page == Plot.USRBIN_HIST[0]:
			if det.type not in (2,12):
				if   det.nx == 1:
					self.proj.set('X')
					self.proj_y.config(state=tk.DISABLED)
					self.proj_z.config(state=tk.DISABLED)
				elif det.ny == 1:
					self.proj.set('Y')
					self.proj_x.config(state=tk.DISABLED)
					self.proj_z.config(state=tk.DISABLED)
				elif det.nz == 1:
					self.proj.set('Z')
					self.proj_x.config(state=tk.DISABLED)
					self.proj_y.config(state=tk.DISABLED)

				ones = (det.nx==1) + (det.ny==1) + (det.nz==1)
				if ones>1:
					self.flair.notify("USRBIN 2D Projection",
						"Two dimensions have ONLY 1 bin,\n"
						"Therefore 2D projection IS NOT possible ONLY 1D",1)
			else:
				self.proj_x.config(state=tk.DISABLED)
				self.proj_y.config(state=tk.DISABLED)
				self.proj_z.config(state=tk.DISABLED)

		#elif page == Plot.USRBIN_HIST[1]:
		else:
			if   det.ny == 1 and det.nz == 1:
				self.proj.set('X')
			elif det.nx == 1 and det.nz == 1:
				self.proj.set('Y')
			elif det.nx == 1 and det.ny == 1:
				self.proj.set('Z')

			if det.nx == 1:
				self.proj_x.config(state=tk.DISABLED)
				if self.proj.get()=="X": self.proj.set("Y")

			if det.ny == 1:
				self.proj_y.config(state=tk.DISABLED)
				if self.proj.get()=="Y": self.proj.set("Z")

			if det.nz == 1:
				self.proj_z.config(state=tk.DISABLED)
				if self.proj.get()=="Z":
					if det.nx != 1:
						self.proj.set("X")
					else:
						self.proj.set("Y")

	# --------------------------------------------------------------------
	def tabPageChange(self, event=None):
		if self.typeCombo.get() == Plot.USRBIN_HIST[0]:
			self.frame2d.pack(expand=tk.YES, fill=tk.BOTH)
			try: self.frame1d.pack_forget()
			except: pass
		else:
			self.frame1d.pack(expand=tk.YES, fill=tk.BOTH)
			try: self.frame2d.pack_forget()
			except: pass
		d = self.detector()
		if d<=0 or d>len(self.usr.detector): return
		self.selectProjection(self.usr.detector[d-1])

	# ----------------------------------------------------------------------
	def detector(self, val=None):
		if val is None:
			# Get the value
			try:
				return int(self.det.get().split()[0])
			except:
				return -1
		else:
			if 0 < val <= len(self.usr.detector):
				det = self.usr.detector[val-1]
				self.det.set(f"{val} {det.name}")
			else:
				self.clearInfo(False)

	# ----------------------------------------------------------------------
	def fillGeometry(self):
		geo = self.widget["geo"]
		geo.clear()
		geo.insert(tk.END, Plot.GEO_NO)
		geo.insert(tk.END, Plot.GEO_AUTO)

		for plot in self.project.plots():
			if plot.type == "Geometry":
				geo.insert(tk.END, plot.name)

	# ----------------------------------------------------------------------
	def load(self, event=None):
		fn = self.datafile.filename()
		if fn:
			det = self.detector()
			det = max(det, 1)
			self._datafile = self.project.relpath(fn)
			self.datafile.set(self._datafile)
			self.loadDataFile(self._datafile)
			self.detector(det)
			self.fillDetectorInfo()

	# ----------------------------------------------------------------------
	def get_limits(self):
		if not self.flair.plotEngine: return
		xlow, xhigh = self.flair.plotEngine.getRange("x")
		if xhigh is None: return
		ylow, yhigh = self.flair.plotEngine.getRange("y")
		if yhigh is None: return

		proj = self.proj.get()
		if proj == "X":
			xlow_field  = self.y_min
			xhigh_field = self.y_max
			ylow_field  = self.z_min
			yhigh_field = self.z_max

		elif proj == "Y":
			xlow_field  = self.x_min
			xhigh_field = self.x_max
			ylow_field  = self.z_min
			yhigh_field = self.z_max

		elif proj == "Z":
			xlow_field  = self.x_min
			xhigh_field = self.x_max
			ylow_field  = self.y_min
			yhigh_field = self.y_max

		if self.swap.get():
			xlow_field,  ylow_field  = ylow_field,  xlow_field
			xhigh_field, yhigh_field = yhigh_field, xhigh_field

		xlow_field.set( _format%(xlow))
		xhigh_field.set(_format%(xhigh))
		ylow_field.set( _format%(ylow))
		yhigh_field.set(_format%(yhigh))

	# ----------------------------------------------------------------------
	def help(self):
		Manual.show("","Mesh Plot")
