# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   07-Dec-2007

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

#import re
import math

import csg
import bmath
#import Global

import log
from deck  import Card
from fluka import Fluka
from mcnp  import Mcnp, McnpCard
from .convert import Convert, register

#===============================================================================
# Conversion class from Fluka to Mcnp
#===============================================================================
class Fluka2Mcnp(Convert):
	options = [		# user selectable options as in the ListboxDialog
			("macrobodies", "Use macrobodies (RCC, BOX...)", False, bool),
		]
	#-----------------------------------------------------------------------
	# Convert FLUKA natural material to isotopic composition
	#-----------------------------------------------------------------------
	_isotopes = {
		 1:  1001,
		 2:  2004,
		 3:((3007,0.925), (3006,0.075)),
		 4:  4009,
		 5:((5011,0.801), (5010,0.199)),
		 6:((6012,0.989), (6013,0.011)),
		 7:((7014,0.996), (7015,0.004)),
		 8:  8016,
		 9:  9019,
		11: 11023,
		13: 13027,
		15: 15031,
		25: 25055,
		27: 27059,
		33: 33075,
		35:((35079,0.507), (35081,0.493)),
		36:((36084,0.570), (36078,0.003), (36080,0.023),
		    (36082,0.116), (36083,0.115), (36086,0.173)),
		38: 38090,
		41: 41093,
		43: 43099,
		53: 53127,
		55: 55133,
		56: 56138,
		62: ((62149,0.2883), (62150,0.1537), (62152,0.558)),
		73: 73181,
		74: 74186,
		75:((75187,0.626), (75185,0.374)),
		79: 79197,
		82:((82208,0.524), (82207,0.221), (82206,0.241), (82204, 0.014)),
		83: 83209,
		90: 90232,
		92:((92235,0.00726), (92238,0.99274)),
		93: 93237,
	}

	#-----------------------------------------------------------------------
	# FIXME some automatic unit conversion system!
	#-----------------------------------------------------------------------
	def convert(self, **kwargs):
		super().convert(**kwargs)
		self.output = Mcnp()

		self.macroBodies = kwargs.get("macrobodies",False)
		self.surfaces   = []	# List of Mcnp surfaces
		self.bodies     = {}	# Dictionary of FLUKA bodies
		self.transform  = []
		self.materials  = []
		self._transformCache = {}

		try:
			card = self.input.cards["TITLE"][0]
			self.addCard("TITLE").setTitle(card.title().rstrip())
		except:
			pass

		self.input.cardProperties()
		self.input.addCardProperty("REGION","BIASING","@importance",3,4,5,6)

		self.convertBodies()
		self.convertRegions()

		for card in self.surfaces:
			self.add(card)
		for card in self.transform:
			self.add(card)

		self._convertMaterials()
		#self._convertImportance()
		return self.output

	#-----------------------------------------------------------------------
	def addCard(self, tag, what=None, comment=""):
		card = McnpCard(tag, what, comment)
		self.add(card)
		return card

	#-----------------------------------------------------------------------
	# Convert Bodies
	#-----------------------------------------------------------------------
	def convertBodies(self):
		for card in self.input.cardsSorted(Card.BODY,-1).values():
			if card is None: continue
			# Convert body card to Add a new surface [##, in, out, card]
			nsurf = len(self.surfaces)
			body = [0, "", "", card]
			self.bodies[card.name()] = body

			if card.tag == "RPP":
				self._convertRPP(card, body)
			elif card.tag == "BOX":
				self._convertBOX(card, body)
			elif card.tag == "SPH":
				self._addSphere(body,'-',card.bodyP(),card.bodyR(),
						card[Card.MATRIX], str(card))
			elif card.tag == "RCC":
				self._convertRCC(card, body)
			elif card.tag == "REC":
				self._convertREC(card, body)
			elif card.tag == "TRC":
				self._convertTRC(card, body)
			elif card.tag == "ELL":
				self._convertELL(card, body)
			elif card.tag == "ARB":
				self._convertARB(card, body)
			elif card.tag == "WED":
				self._convertWED(card, body)
			elif card.tag in ("XYP", "XZP", "YZP", "PLA"):
				self._addPlane(body,'-',card.bodyP(),card.bodyN(),
						card[Card.MATRIX], str(card))
			elif card.tag in ("XCC", "YCC", "ZCC"):
				self._convertInfCylinder(card, body)
			elif card.tag in ("XEC", "YEC", "ZEC"):
				self._convertInfEllipticalCylinder(card, body)
			elif card.tag == "QUA":
				self._convertQUA(card, body)
			elif card.tag == "END":
				pass	# ignore END card
			else:
				raise Exception(f"MCNP: unknown FLUKA body type {card.tag}")

			body[0] = len(self.surfaces) - nsurf

	#-----------------------------------------------------------------------
	def _convertRPP(self, card, body):
		M = card[Card.MATRIX]
		if self.macroBodies:
			self._addSurf(body,'-',"RPP",
					[card.numWhat(1), card.numWhat(2), card.numWhat(3),
					 card.numWhat(4), card.numWhat(5), card.numWhat(6)],
					str(card), Fluka2Mcnp.TR(M))
		else:
			P  = card.bodyP()
			Po = card.bodyPo()
			self._addPlane(body,'+',P, bmath.Vector.X, M, str(card))
			self._addPlane(body,'-',Po,bmath.Vector.X, M)
			self._addPlane(body,'+',P, bmath.Vector.Y, M)
			self._addPlane(body,'-',Po,bmath.Vector.Y, M)
			self._addPlane(body,'+',P, bmath.Vector.Z, M)
			self._addPlane(body,'-',Po,bmath.Vector.Z, M)

	#-----------------------------------------------------------------------
	def _convertBOX(self, card, body):
		P = card.bodyP()
		X = card.bodyX()
		Y = card.bodyY()
		Z = card.bodyZ()
		M = card[Card.MATRIX]

		if self.macroBodies:
			self._addSurf(body,'-',"BOX",
					[P.x, P.y, P.z, X.x, X.y, X.z,
					 Y.x, Y.y, Y.z, Z.x, Z.y, Z.z],
					str(card), Fluka2Mcnp.TR(M))
		else:
			# Correct the order
			if (X.cross(Y)).dot(Z)<0:
				X, Y = Y, X		# inverse X with Y

			# outer corner
			Po = P + X + Y + Z
			X.norm()
			Y.norm()
			Z.norm()

			# first 3 planes will be
			self._addPlane(body,'+',P,X,M, str(card))
			self._addPlane(body,'+',P,Y,M)
			self._addPlane(body,'+',P,Z,M)

			# last 3 planes will be
			self._addPlane(body,'-',Po,X,M)
			self._addPlane(body,'-',Po,Y,M)
			self._addPlane(body,'-',Po,Z,M)

	#-----------------------------------------------------------------------
	def _convertRCC(self, card, body):
		P = card.bodyP()
		Z = card.bodyZ()
		r = card.bodyR()
		M = card[Card.MATRIX]
		if self.macroBodies:
			self._addSurf(body,'-',"RCC",
					[P.x, P.y, P.z, Z.x, Z.y, Z.z, r],
					str(card), Fluka2Mcnp.TR(M))
		else:
			Z.norm()
			self._addSurf(body,'-',"CZ", [r],
					str(card),Fluka2Mcnp.TR(M,P,Z=Z))
			self._addPlane(body,'+',P,Z,M)
			self._addPlane(body,'-',card.bodyPo(),Z,M)

	#-----------------------------------------------------------------------
	def _convertREC(self, card, body):
		P = card.bodyP()
		X = card.bodyX()
		Y = card.bodyY()
		Z = card.bodyZ()
		M = card[Card.MATRIX]

		if self.macroBodies:
			self._addSurf(body,'-',"REC",
					[P.x, P.y, P.z, Z.x, Z.y, Z.z,
					 X.x, X.y, X.z, Y.x, Y.y, Y.z],
					str(card), Fluka2Mcnp.TR(M))
		else:
			radiusX = X.length()
			radiusY = Y.length()

			# correct system
			if (X.cross(Y)).dot(Z) < 0.0:
				X = -X	# from left-hand to right-hand system

			# normalize the coordinates
			Z = X.cross(Y)
			Y = Z.cross(X)

			X.norm()
			Y.norm()
			Z.norm()

			a = 1.0/(radiusX**2)
			b = 1.0/(radiusY**2)

			self._addSurf(body,'-',"SQ",
					[a, b, 0, 0, 0, 0, -1, 0, 0, 0],
					str(card), Fluka2Mcnp.TR(M,P=P,X=X,Y=Y))
			self._addPlane(body,'+',P,Z,M)
			self._addPlane(body,'-',card.bodyPo(),Z,M)

	#-----------------------------------------------------------------------
	def _convertTRC(self, card, body):
		P = card.bodyP()
		Z = card.bodyZ()
		M = card[Card.MATRIX]
		baseRadius = card.numWhat(7)
		apexRadius = card.numWhat(8)

		if self.macroBodies:
			self._addSurf(body,'-', "TRC",
					[P.x, P.y, P.z, Z.x, Z.y, Z.z,
					 baseRadius, apexRadius],
					str(card), Fluka2Mcnp.TR(M))
		else:
			Zlen = Z.length()
			Z.norm()
			s = (apexRadius-baseRadius)/Zlen
			if abs(s)<1e-10:
				# Normal cylinder
				self._addSurf(body,'-',"CZ", [baseRadius],
						str(card), Fluka2Mcnp.TR(M,P,Z=Z))
				self._addPlane(body,'+',P,Z,M)
				self._addPlane(body,'-',card.bodyPo(),Z,M)
			else:
				self._addSurf(body,'-',"KZ",
						[-baseRadius/s, s**2],
						str(card), Fluka2Mcnp.TR(M,P,Z=Z))
				self._addPlane(body,'+',P,Z,M)
				self._addPlane(body,'-',card.bodyPo(),Z,M)

	#-----------------------------------------------------------------------
	def _convertELL(self, card, body):
		P1 = card.bodyP1()
		P2 = card.bodyP2()
		M = card[Card.MATRIX]
		P  = 0.5 * (P1 + P2)
		a  = card.numWhat(7) / 2.0
		c  = (P1-P).length()
		b  = math.sqrt(a**2 - c**2)

		# Create axis system
		Z = P1-P2
		Z.norm()

		self._addSurf(body,'-', "SQ",
				[1.0/b**2, 1.0/b**2, 1.0/a**2, 0, 0, 0, -1, 0, 0, 0],
				str(card), Fluka2Mcnp.TR(M,P,Z=Z))

	#-----------------------------------------------------------------------
	def _convertQUA(self, card, body):
		M = card[Card.MATRIX]
		self._addSurf(body,'-', "GQ",
				[card.numWhat(1), card.numWhat(2), card.numWhat(3),
				 # Warning on the order xy yz zx
				 card.numWhat(4), card.numWhat(6), card.numWhat(5),
				 card.numWhat(7), card.numWhat(8), card.numWhat(9),
				 card.numWhat(10)],
				str(card), Fluka2Mcnp.TR(M))

	#-----------------------------------------------------------------------
	def _convertARB(self, card, body):
		M = card[Card.MATRIX]
		if self.macroBodies:
			self._addSurf(body,'-', "ARB",
					[card.bodyPn(1), card.bodyPn(2), card.bodyPn(3),
					 card.bodyPn(4), card.bodyPn(5), card.bodyPn(6),
					 card.bodyPn(7), card.bodyPn(8),
					 card.intWhat(25), card.intWhat(26),
					 card.intWhat(27), card.intWhat(28),
					 card.intWhat(29), card.intWhat(30)],
					 str(card), Fluka2Mcnp.TR(M))
		else:
			V = [None]*8
			for i in range(8):
				V[i] = card.bodyPn(i+1)
			#F = [None]*6
			s = str(card)
			for i in range(6):
				face = card.intWhat(i+25)
				if face == 0: continue
				v1,face = divmod(face, 1000)
				v2,face = divmod(face,  100)
				v3,v4   = divmod(face,   10)

				# use the first three to define the plane
				E1 = V[v2-1] - V[v1-1]
				E2 = V[v3-1] - V[v1-1]
				N = E1.cross(E2)
				N.norm()
				self._addPlane(body,'-',V[v1-1],N,M,s)
				s = None

	#-----------------------------------------------------------------------
	def _convertWED(self, card, body):
		P = card.bodyP()
		X = card.bodyX()
		Y = card.bodyY()
		Z = card.bodyZ()
		M = card[Card.MATRIX]

		if self.macroBodies:
			self._addSurf(body,'-',"WED",
					[P.x, P.y, P.z, X.x, X.y, X.z,
					 Y.x, Y.y, Y.z, Z.x, Z.y, Z.z],
					str(card), Fluka2Mcnp.TR(M))
		else:
			# Correct the order
			if (X.cross(Y)).dot(Z)<0:
				X, Y = Y, X		# inverse X with Y

			XC = P + X
			YC = P + Y
			ZC = P + Z

			# normal of the plane
			N = (YC-XC).cross(ZC-XC)
			N.norm()

			# first 3 planes will be
			self._addPlane(body,'+',P,X,M,str(card))
			self._addPlane(body,'+',P,Y,M)
			self._addPlane(body,'+',P,Z,M)

			# cutting plane
			self._addPlane(body,'-',XC,N,M)

	#-----------------------------------------------------------------------
	def _convertInfCylinder(self, card, body):
		u = card.numWhat(1)
		v = card.numWhat(2)
		r = card.numWhat(3)
		M = card[Card.MATRIX]
		axis = card.tag[0]

		if u==0.0 and v==0.0:
			self._addSurf(body,'-',f"C{axis}", [r],
					str(card), Fluka2Mcnp.TR(M))
		else:
			if axis=="Y": u,v = v,u		# Reverse for Y
			self._addSurf(body,'-',f"C/{axis}",
					[u, v, r],
					str(card), Fluka2Mcnp.TR(M))

	#-----------------------------------------------------------------------
	def _convertInfEllipticalCylinder(self, card, body):
		P     = card.bodyP()
		a,b,c = card.bodyR()
		M     = card[Card.MATRIX]
		#axis  = card.tag[0]
		if a!=0.0: a = 1.0/a**2
		if b!=0.0: b = 1.0/b**2
		if c!=0.0: c = 1.0/c**2

		# FIXME it should not be correct!!!
		self._addSurf(body,'-', "SQ",
				[a, b, c, 0, 0, 0, -1, P.x, P.y, P.z,],
				str(card), Fluka2Mcnp.TR(M))

	#-----------------------------------------------------------------------
	# add a plane and convert it if possible to best match
	#-----------------------------------------------------------------------
	def _addPlane(self, body, op, pos, normal, matrix, comment=""):
		normal.normalize()
		if matrix is not None:
			pos = matrix * pos
			normal = matrix.multNoTranslation(normal)

		dire = normal.direction()
		if dire == "X":
			self._addSurf(body, op, "PX", [pos.x], comment)
		elif dire == "Y":
			self._addSurf(body, op, "PY", [pos.y], comment)
		elif dire == "Z":
			self._addSurf(body, op, "PZ", [pos.z], comment)
		elif dire == "-X":
			op = "-" if op=="+" else "+"
			self._addSurf(body, op, "PX", [pos.x], comment)
		elif dire == "-Y":
			op = "-" if op=="+" else "+"
			self._addSurf(body, op, "PY", [pos.y], comment)
		elif dire == "-Z":
			op = "-" if op=="+" else "+"
			self._addSurf(body, op, "PZ", [pos.z], comment)
		else:
			self._addSurf(body, op, "P",
					[normal.x, normal.y, normal.z,
					 pos.dot(normal)],
					comment)

	#-----------------------------------------------------------------------
	# add sphere with the appropriate card
	#-----------------------------------------------------------------------
	def _addSphere(self, body, op, pos, r, matrix, comment=""):
		if matrix is not None:
			pos = matrix * pos
		x,y,z = pos
		small = bmath._accuracy

		if abs(x)<small and abs(y)<small and abs(z)<small:
			self._addSurf(body,'-',"SO", [r], comment)
		elif abs(y)<small and abs(z)<small:
			self._addSurf(body,'-',"SX", [x,r], comment)
		elif abs(x)<small and abs(z)<small:
			self._addSurf(body,'-',"SY", [y,r], comment)
		elif abs(x)<small and abs(y)<small:
			self._addSurf(body,'-',"SZ", [z,r], comment)
		else:
			self._addSurf(body,'-',"S", [x,y,z,r], comment)

	#-----------------------------------------------------------------------
	# Add a surface to MCNP
	#-----------------------------------------------------------------------
	def _addSurf(self, body, op, tag, whats, comment="", tr=None):
		nop = "-" if op=="+" else "+"
		n = len(self.surfaces)+1
		whats.insert(0,str(n))

		inside  = body[1]
		outside = body[2]
		if inside == "":
			inside  = f"{op}{n}"
			outside = f"{nop}{n}"
		else:
			inside  += f" {op}{n}"
			outside += f":{nop}{n}"

		body[1] = inside
		body[2] = outside

		surface = McnpCard(tag, whats, comment)
		if tr is not None:
			try:
				tn = self._transformCache[str(tr)]
			except KeyError:
				tn = len(self.transform)+1
				self._transformCache[str(tr)] = tn
				tr.insert(0,str(tn))
				self.transform.append(McnpCard("TR", tr))
			surface["tr"] = tn
		self.surfaces.append(surface)

	#-----------------------------------------------------------------------
	# Convert FLUKA REGION expressions to MCNP
	#-----------------------------------------------------------------------
	def convertRegions(self):
		region = None
		expr   = ""
		for card in self.input.cardsSorted(Card.REGION,-1).values():
			if card.name() == "&":
				# append to previous expression
				expr += card.expr()
				continue

			if region is None:
				# Remember starting region
				region = card
				expr   = region.expr()
				continue

			self._convertRegion(region, expr)

			# Remember starting region
			region = card
			expr   = region.expr()

		# Last one
		if region: self._convertRegion(region, expr)

		# Create universe
		# ... XXX assume is composed by only one body
		#self.cells.append(("%d 0 %s"%(len(self.cells)+1,body[2])
		#			,"Universe"))

	#-----------------------------------------------------------------------
	def _convertRegion(self, region, expr):
		# Find material
		material = region[Card.MATERIAL]
		if material is None:
			log.error(f"Region {region.name()} is not assigned material")
			return True

		density = material.numWhat(3)
		whats = [region["@n"]]
		if density <= 0.0:
			whats.append(0)
		else:
			n = material["@n"]
			if n<0:
				log.warning(f"Predefined material {material.name()} "
					"is substituted by BLCKHOLE")
				n=1
			whats.append(n)
			whats.append(-density)

			# Add material to the list
			n -= 1
			if n >= len(self.materials):
				self.materials.extend([None] *
					(n-len(self.materials)+1))
			self.materials[n] = material

		# remove spaces
		region_expr = "".join(expr.split())
		if not region_expr: return True
		region_expr = csg.tokenize(region_expr)

		op = "+"
		first = True
		# construct the MCNP region line
		cell_expr = ""
		prevlen = 0
		for item in region_expr:
			if item in ("+", "-"):
				op = item
			elif item == "|":
				op = "+"
				if first:
					first = False
				else:
					cell_expr += " :"
			elif item == "(":
				if op == "-":
					cell_expr += " #("
				else:
					cell_expr += " ("
			elif item == ")":
				cell_expr += ")"
			else:
				first = False
				body = self.bodies[item]
				n, inside, outside, bcard = body
				if op=="+":
					cell_expr += f" {inside}"
				elif n>1:
					cell_expr += f" ({outside})"
				else:
					cell_expr += f" {outside}"
			if len(cell_expr)-prevlen > McnpCard.MAXLEN:
				cell_expr += "\n"
				prevlen = len(cell_expr.strip())
		card = self.addCard("CELL", whats, region.comment())
		card.appendComment(region._toStr())
		card.setExpr(cell_expr.strip())

		imp = 1
		if material.intWhat(4)==1:
			imp = 0
		else:
			biasing = region["@biasingN"]
			if biasing is None:
				biasing = region["@biasingA"]
			if biasing is not None:
				imp = biasing.numWhat(3)
			else:
				imp = 1
		card["imp:n"] = imp

	#-----------------------------------------------------------------------
	# Add materials cards
	# Import Materials that are in use
	# FIXME converts everything to atom fractions
	#       maybe it be nice to keep mass if needed
	#-----------------------------------------------------------------------
	def _convertMaterials(self):
		for mat in self.materials:
			if mat is None: continue
			ZAID = {}

			#-------------------------------------------------------
			def add(z, f, am):
				nonlocal ZAID
				try:
					ZAID[z][0] += f
				except (KeyError, IndexError):
					ZAID[z] = [f,am]

			#-------------------------------------------------------
			for za,[f,am] in self.input.materialZAID(mat,False,-1).items():
				z,a = divmod(za,1000)
				if a==0:
					zaid = Fluka2Mcnp._isotopes.get(z)
				else:
					zaid = None
				if zaid is not None:
					if isinstance(zaid, int):
						add(zaid,f,am)
					else:
						for zz,ff in zaid:
							add(zz,ff*f,am)
				else:
					add(za,f,am)
			mat.zaid = list(ZAID.items())

			whats = [mat["@n"]]
			for z,(f,am) in mat.zaid:
				whats.append(z)
				whats.append(f)
			self.addCard("M", whats, str(mat))
			if mat.name() in ("WATER","H2O"):
				self.addCard("MT", [mat['@n'], 'lwtr.01t'])

	#-----------------------------------------------------------------------
	# Add biasing
	#-----------------------------------------------------------------------
	#def _writeBiasing(self, fout):
	#	fout.write("C ---------- Biasing ----------\n")
	#	line   = []
	#	repeat = 0
	#	prev   = -1.0
	#	for imp in self.importance:
	#		if imp == prev:
	#			repeat += 1
	#		else:
	#			if prev<0:
	#				pass
	#			else:
	#				line.append(f"{prev:.15g}")
	#				if repeat == 1:
	#					line.append(line[-1])
	#				elif repeat > 1:
	#					line.append(f"{repeat}r")
	#			prev = imp
	#			repeat = 0
	#	line.append(f"{prev:.15g}")
	#	if repeat == 1:
	#		line.append(line[-1])
	#	else:
	#		line.append(f"{repeat}r")
	#	self._writeLine(fout, f"IMP:n {' '.join(line)}")

	#-----------------------------------------------------------------------
	# Convert matrix to string
	# @param M	rotation matrix
	# @param P	[optional] position
	# @param X	[optional] X-axis vector
	# @param Y	[optional] Y-axis vector
	# @param Z	[optional] Z-axis vector
	# @return TR card whats
	#-----------------------------------------------------------------------
	@staticmethod
	def TR(M, P=None, Z=None, X=None, Y=None):
		if M is not None:
			if P is not None: P = M * P
			if X is not None: X = M.multNoTranslation(X)
			if Y is not None: Y = M.multNoTranslation(Y)
			if Z is not None: Z = M.multNoTranslation(Z)
		elif P is None and Z is None and X is None and Y is None:
			return None	 # M = P = Z = None

		if P is not None:	# P is defined
			if Z is not None and X is None and Y is None:		# Only Z
				return [P.x, P.y, P.z] + [""]*6 + [Z.x, Z.y, Z.z]
			elif Z is None and X is not None and Y is not None:	# Only X,Y
				return [P.x, P.y, P.z] + [X.x, X.y, X.z] + [Y.x, Y.y, Y.z]
			elif Z is None and X is None and Y is None:		# no-axis
				return [P.x, P.y, P.z]
			else:
				assert False	# We need either X&Y or Z
		elif Z is not None:
			return [""]*9 + [Z.x, Z.y, Z.z]
		elif X is not None and Y is not None:
			return [""]*3 + [X.x, X.y, X.z] + [Y.x, Y.y, Y.z]
		else:
			return  [M[0][3], M[1][3], M[2][3],
				 M[0][0], M[1][0], M[2][0],
				 M[0][1], M[1][1], M[2][1],
				 M[0][2], M[1][2], M[2][2]]

#-------------------------------------------------------------------------------
# Execute only once and register the conversion program
#-------------------------------------------------------------------------------
register(Fluka.MODE, Mcnp.MODE, Fluka2Mcnp)
