# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   8-Jan-2019

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import math

import log
import csg
import bmath
from Evaluation import fwhm

from deck    import Card
from fluka   import Fluka, FlukaParticle
from moira   import Moira, MoiraCard
from element import Element

from .convert import Convert, register

#-------------------------------------------------------------------------------
# Check if all items in a list are integers
#-------------------------------------------------------------------------------
def allIntegers(lst, acc):
	return all([abs(x-float(round(x)))<acc for x in lst])

#===============================================================================
# Conversion class from Fluka to Moira
#===============================================================================
class Fluka2Moira(Convert):
	src = Fluka.MODE
	dst = Moira.MODE
	INVMATRIX = "%s{-1}"	# pattern for inverse matrix lattices = []
	BEAMPOS_ORIENTATION = ("", "NEGATIVE")
	FIELDS = {1:"dipole", 2:"dipole", 4:"quadrupole", 6:"sextupole", 8:"octupole", 10:"decapole"}
	#BEAMPOS_VOLSOURCES  = ("FLOOD", "CART-VOL", "CYLI-VOL", "SPHE-VOL")
	mesh_geo = "mesh"
	volume_geo = "volume"
	surface_geo = "surface"

	#-----------------------------------------------------------------------
	def __init__(self, inp):
		super().__init__(inp)

		self._rotdefi = self.input.getTransformation()
		self.input.bodyProperties(self._rotdefi)
		self.input.regionProperties()#self._rotdefi)
		self._elements  = {}		# keep track of defined elements
		self._materials = {}		# keep track of defined materials
		self.compounds  = []
		self._reg       = []
		self._rotid     = {}
		self._last_unit  = None
		self._count_meshes_per_output_unit = {}

		# Cache lastly encountered geometry
		# (to be able to append further scorers to it)
		self._last_geo = ""
		self._last_reg = ""
		self._last_reg2 = ""
		self._last_norm = -1
		self._last_mesh_xmax = 0
		self._last_mesh_ymax = 0
		self._last_mesh_zmax = 0
		self._last_mesh_xmin = 0
		self._last_mesh_ymin = 0
		self._last_mesh_zmin = 0
		self._last_mesh_nx = 1
		self._last_mesh_ny = 1
		self._last_mesh_nz = 1

		self._has_scorer = False
		self.fields     = []
		self.includes   = []
		self.lattices   = []
		self.reg        = None
		self.beam       = None
		self.beamEn     = None
		self.output     = None
		self._comment   = None		# prefix comment to define regions
		self._enable    = True		# disable card after conversion
		self._applycuts = False		# apply production cuts

	#-----------------------------------------------------------------------
	# FIXME some automatic unit conversion system!
	#-----------------------------------------------------------------------
	def convert(self, **kwargs):
		super().convert(**kwargs)
		self.output = Moira()
		self.preprocess()
		try:
			card = self.input["TITLE"][0]
			self.used(card)
			self.addCard("/moira/title",[card.text()],card.comment())
		except IndexError:
			pass
		card = self.addCard("/run/numberOfThreads")
		card.disable()
		self.addCard("/physics_lists/select",["FTFP_BERT_HP", "EM0"])
		self.convertMaterials()
		self.convertTransformations()
		blkregion, world = self.findBlackBody()
		self.convertBodies(world)
		self.convertRegions(blkregion, world)
		self.convertAllOtherCards()
		return self.output

	#-----------------------------------------------------------------------
	def addCard(self, tag, what=None, comment=""):
		if self._comment:
			comment = f"{self._comment}\n{comment}"
			self._comment = None
		card = MoiraCard(tag, what, comment)
		self.add(card, self._enable)
		return card

	#-----------------------------------------------------------------------
	def material(self, name):
		try:
			return self._materials[name]
		except KeyError:
			try:
				g4mat = Moira.materialFromFluka(name)
				self._materials[name] = g4mat	# remember for next time
				return g4mat
			except KeyError:
				log.error(f"Material {name} was not converted to moira")
		return "G4_Galactic"

	#-----------------------------------------------------------------------
	def element(self, name):
		try:
			return self._elements[name]
		except KeyError:
			try:
				g4ele = Moira.elementFromFluka(name)
				self._elements[name] = g4ele	# remember for next time
				return g4ele
			except KeyError:
				try:
					self._elements[name] = name	# remember for next time
					return name
				except KeyError:
					log.error(f"Element {name} was not converted to moira")
					return "G4_Galactic"

	#-----------------------------------------------------------------------
	# Preprocessor commands
	#-----------------------------------------------------------------------
	def preprocess(self):
		for card in self.input.cardlist:
			if card.tag == "BEAM":
				particle = card.name() if card.name() != '' else "PROTON"
				self.beam = particle
				self.beamEn = abs(card.intWhat(1))
			if card.tag == "#define":
				self._enable = card.notIgnore()
				self.add(card.clone(), self._enable)
		self._enable = True

	#-----------------------------------------------------------------------
	# Convert materials from fluka to moira
	#-----------------------------------------------------------------------
	def convertMaterials(self):
		# 1st pass register all fluka defined materials
		self._comment = "\n--------------------------------------------------" \
				"\n   Materials" \
				"\n--------------------------------------------------"
		for mat in self.input["MATERIAL"]:
			if mat.notActive(): continue
			self._enable = mat.notIgnore()
			# Check if material is an element existing in G4
			if mat.what(1)!="" and mat.what(2)=="" and mat.what(6)=="" and \
			   Moira.materialFromFluka(mat.name()):
				self._materials[mat.name()] = Moira.materialFromFluka(mat.name())
				self.used(mat)
			else:
				self._materials[mat.name()] = mat.name()

		#elements       = {}
		doub_compounds = []
		isotlist       = []
		ellist         = []
		matlist        = []
		temperature    = []
		pressure       = []
		state          = []

		# Removing double definition of compounds
		for card in self.input["COMPOUND"]:
			if not card.what(2): continue
			if card.name() in self.compounds:
				if card.notActive(): continue
				doub_compounds.append(card.name())
			else:
				self.compounds.append(card.name())

		for comp in doub_compounds:
			for card in self.input["COMPOUND"]:
				if comp == card.name():
					self.input.delCard(card.pos())
					break
			for card in self.input["MATERIAL"]:
				if comp == card.name():
					self.input.delCard(card.pos())
					break

		for mat in self.input["MATERIAL"]:
			if mat.notActive() or self.isused(mat): continue
			self._enable = mat.notIgnore()
			self.used(mat)
			name = mat.name()
			Z    = mat.intWhat(1)
			Am   = mat.numWhat(2)
			rho  = mat.numWhat(3)
			if mat.what(5): self.notImplemented(mat, 5)
			A    = mat.intWhat(6)
			#if Z!=0 and Am==0 and A==0: # FIXME and rho==..
			if name not in self.compounds and Z != 0 and A != 0:
				element = Element.byZ(Z)
				if not Am: Am = element.Amass()
				if name in ellist: continue
				if name in matlist: continue

				self.addCard("ELEM", [name,element.symbol,Z,Am])
				self.addCard("MATE", [name,Z,Am,rho])
				ellist.append(name)
				matlist.append(name)

			elif name not in self.compounds and Z != 0 and A in [0, ""]:
				# Find A from database and create material
				element  = Element.byZ(Z)
				isotopes = element.isotopes(Z)
				liso     = [name, element.symbol]
				# liso   = [name+"_iso", element.symbol]
				niso     = 0
				isos     = []

				for iso in isotopes:
					if "%" in isotopes[iso].decayt:
						niso += 1
				# if niso == 1:
					# self.addCard("MATE", [name,Z,element.Amass(),rho], mat.comment())
				# else:
				for iso in isotopes:
					if "%" in isotopes[iso].decayt:
					# if isotopes[iso].decayt == "stable":
						A = isotopes[iso].A
						if "m" in A or "n" in A:
							A = A[:-1]
						symb = isotopes[iso].symbol
						Am   = isotopes[iso].Am
						if f"{symb}-{A}" not in isotlist:
							isotlist.append(f"{symb}-{A}")
							self.addCard("ISOT", [f"{symb}-{A}",Z,A,Am], mat.comment())
							mat.setComment("")
						isos.append(f"{symb}-{A}")
						lisoaux=(isotopes[iso].decayt.split()[0])[:-1]
						lisoaux=lisoaux.split(";")[0]
						isos.append(lisoaux)

				liso.append(niso)
				liso.extend(isos)

				if name in ellist: continue
				if name in matlist: continue

				self.addCard("ELEM_FROM_ISOT", liso, mat.comment())
				self.addCard("MIXT_BY_NATOMS", [name, rho, 1, name, 1.0], mat.comment())
				ellist.append(name)
				matlist.append(name)
				# Option 2?
				# if not Am: Am = element.Amass()
				# self.addCard("ELEM", [name+"_el",element.symbol,Z,Am])
				# self.addCard("MIXT_BY_NATOMS", [name, rho, 1, name+"_el", 1.0])
			else:
				# Create mixture by finding the compound
				tag   = None
				what  = [name, rho, 0 ]	# 3rd element is the # of components
				total = 0.0
				fracs = []
				for card in self.input["COMPOUND"]:
					self.used(card)
					if card.notActive(): continue
					if card.name() != name: continue
					for i in range(1,card.nwhats(),2):
						frac = card.numWhat(i)
						elem = card.what(i+1)
						if frac==0 or not elem: break
						if tag is None:
							if frac>0:
								tag = "MIXT_BY_NATOMS"
							elif elem[0]=="-":
								tag = "MIXT_BY_VOLUME"
							else:
								tag = "MIXT"

						elemname = self.material(elem)
						if tag == "MIXT_BY_NATOMS":
							#if elem not in elements:
							# FIXME bnv!!!!
							# try to properly identify element
							# from the definition and Z
							elemname = elemname.replace("G4_","")
						frac = abs(frac)
						total += frac
						fracs.append(frac)
						frac = bmath.format(frac)

						if elem[0]=="-": elem = elem[1:]
						if tag == "MIXT_BY_NATOMS":
							what.append(self.element(elem))
						else:
							what.append(self.material(elem))
						what.append(frac)

					# check if all are integers
					if tag == "MIXT_BY_NATOMS":
						acc = 0.001
						if not allIntegers(fracs, acc):
							# Convert to ints
							d = 0.1
							while d>1e-10:
								fracs_new = [x/d for x in fracs]
								if allIntegers(fracs_new, acc):
									fracs = fracs_new
									break
								d /= 10
							fracs = [round(x) for x in fracs]
							for i,j in enumerate(range(4,len(what),2)):
								what[j] = fracs[i]
					else:
						for i in range(len(fracs)):
							fracs[i] /= total
						for i,j in enumerate(range(4,len(what),2)):
							what[j] = fracs[i]

				if tag is not None:
					what[2] = (len(what)-3)//2
					self.addCard(tag, what, mat.comment())

			for t in self.input.cardlist:
				self._enable = t.notIgnore()
				if t.tag == "LOW-MAT" and t.what(1) == name: # Temperature
					if name not in temperature:
						self.addCard("MATE_TEMPERATURE", [name, t.what(4)])
						temperature.append(name)
					element = Element.byZ(Z)
					tm = element.melting
					tb = element.boiling
					if "atm" in tm or "atm" in tb: continue
					tm  = float((tm.split()[0]).replace("~", ""))
					tm += 273.15
					tb  = float((tb.split()[0]).replace("~", ""))
					tb += 273.15
					T   = float(t.what(4))

					if name not in state:
						if T < tm:
							self.addCard("MATE_STATE", [name, "Solid"])
						elif tm < T < tb:
							self.addCard("MATE_STATE", [name, "Liquid"])
						elif tb < T:
							self.addCard("MATE_STATE", [name, "Gas"])
						state.append(name)

				if t.tag == "MAT-PROP" and t.what(4) == name: # Pressure
					if name in pressure: continue
					self.addCard("MATE_PRESSURE", [name, t.what(1)])
					pressure.append(name)

	#-----------------------------------------------------------------------
	def convertTransformations(self):
		self._comment = "\n--------------------------------------------------" \
				"\n   Transformations" \
				"\n--------------------------------------------------"
		# find unique names
		self._rotid = {}
		for name, rot in self._rotdefi.items():
			if not isinstance(rot,int): continue
			if rot in self._rotid: continue
			self._rotid[rot] = name
			rotm = self._rotdefi[rot]
			#rx,ry,rz = [bmath.format(math.degrees(x))
			#		for x in rotm.getEulerRotation()]
			#self.addCard("ROTM", [name, rx, ry, rz])
			what = [name]
			for i in range(3):
				what.extend([bmath.format(x) for x in rotm[i][:3]])
			self.addCard("ROTM", what)
			#print()
			#print(name)
			#print("M=",rotm)
			#print("Rxyz=",rx,ry,rz, "T=",rotm[0][3],rotm[1][3],rotm[2][3])

			# and the inverse matrix
			roti = rotm.clone()
			roti.inv()
			#rx,ry,rz = [bmath.format(math.degrees(x),16)
			#		for x in roti.getEulerRotation()]
			#self.addCard("ROTM", [Fluka2Moira.INVMATRIX%(name), rx, ry, rz])
			what = [Fluka2Moira.INVMATRIX%(name)]
			for i in range(3):
				what.extend([bmath.format(x) for x in roti[i][:3]])
			self.addCard("ROTM", what)
			#print("I=",roti)
			#print("Rxyz=",rx,ry,rz, "T=",roti[0][3],roti[1][3],roti[2][3])
			#print()

	#-----------------------------------------------------------------------
	#def findBody(self, name):
	#	for t in CardInfo.bodies[self.input.MODE]:
	#		for x in self.input[t]:
	#			if x.notIgnore() and x.name()==name:
	#				return x

	#-----------------------------------------------------------------------
	# --- Find world and void ---
	# Find outer blackbody
	# FIXME very very primitive
	# it will search for:
	#  1. regions with BLCKHOLE as material
	#  2. name: BLKBODY
	#  3. in expression to find +BODY -[...]
	#     where BODY is referred only in this expression
	#  4. typically the first one or the last one
	# it should be the biggest region with
	# BLACKHOLE material and typically a simple expression +BLK -VOID
	# where both BLK and VOID are simple objects RPP or BOX.
	#-----------------------------------------------------------------------
	def findBlackBody(self):
		# build dictionary with biggest dimension
		bodyMax = {}
		bodies = self.input.cardsCache(Card.BODY)
		for name,body in bodies.items():
			try:
				m = max([abs(body.numWhat(x))
							for x in range(1,body.nwhats())])
			except ValueError:
				m = 0
			bodyMax[name] = m, body

		# For all regions with BLCKHOLE as material
		# find the biggest body
		regionMax = 0
		blkregion = None
		for region in self.input["REGION"]:
			if region.notActive(): continue
			material = region[Card.MATERIAL]
			if material is not None and \
			   material.name() == "BLCKHOLE":
				for token in csg.tokenize(region.expr()):
					if token in "+-|:()": continue
					try:
						r, body = bodyMax[token]
						if r>regionMax:
							regionMax = r
							blkregion = region
					except KeyError:
						pass

		if blkregion is None:
			raise Exception("Black body search didn't work out!")

		# Possible patterns of blkregion
		# 1. +blkbody -world
		expr = "".join(blkregion.expr().split())
		m = re.match(r"^\+(\w+)-(\w+)$", expr)
		if m:
			world = bodies[m.group(2)]
		else:
			# 2. -blkbody
			m = re.match(r"^-(\w+)$", blkregion.expr())
			if m:
				world = bodies[m.group(1)]
			else:
				# 3. +blkbody -body1 -body2 ....
				m = re.match(r"^\+(\w+)\b", expr)
				if m:
					world = bodies[m.group(1)]
				else:
					raise Exception("Black body search didn't work out!")

		if world.tag not in ("RPP","SPH"):
			raise Exception("Black body seems it is not RPP or SPH")
		log.say("Finding blackbody:")
		log.say("\tBlkregion=", blkregion)
		log.say("\tWorld=", world)
		return blkregion, world

	#-----------------------------------------------------------------------
	def convertBodies(self, world):
		matrix = None
		numMat = 0
		self._comment = "\n--------------------------------------------------" \
				"\n   Bodies" \
				"\n--------------------------------------------------"
		for card in self.input.cardlist:
			if card.notActive():    continue
			if not card.isGeo(): continue
			self._enable = card.notIgnore()
			#if card is world:
			#	self.used(card)
			#elif card.type() == Card.BODY:
			if card.type() == Card.BODY:
				if card[Card.MATRIX] is not matrix:
					matrix = card[Card.MATRIX]
					if matrix is not None:
						#FIXME Find a better idea..
						what = [f"MATRIX{numMat}"]
						numMat += 1
						for row in matrix[:-1]: # skip last 0 0 0 1 row
							what.extend(row)
						if self._enable: self.addCard("MATRIX", what)
					else:
						# if self._enable: self.addCard("MATRIX")
						self._enable = True
						self.addCard("MATRIX")
						self._enable = card.notIgnore()
				self.addCard(card.tag, card.whats()[:], card.comment())
				self.used(card)
			elif card.tag[0] == "#":
				self.add(card.clone())
				self.used(card)
			#elif card.tag[0] == "$":
			#	self.used(card)

	#-----------------------------------------------------------------------
	# Convert them to volumes
	#-----------------------------------------------------------------------
	def volumeCard(self, body, parent=None, mat="G4_Galactic", comment=""):
		name = body.name()
		if body.tag == "SPH":
			# Comm
			self.addCard("VOLU.ORB", [name, mat, body.numWhat(4)], comment)
			x = body.numWhat(1)
			y = body.numWhat(2)
			z = body.numWhat(3)
		elif body.tag == "RPP":
			dx = (body.numWhat(2)-body.numWhat(1))/2
			dy = (body.numWhat(4)-body.numWhat(3))/2
			dz = (body.numWhat(4)-body.numWhat(5))/2
			self.addCard("VOLU.BOX", [name, mat, dx, dy, dz], comment)
			x = bmath.format((body.numWhat(2)+body.numWhat(1))/2)
			y = bmath.format((body.numWhat(4)+body.numWhat(3))/2)
			z = bmath.format((body.numWhat(4)+body.numWhat(5))/2)
		else:
			raise Exception(name+" with type "+body.tag+" is not supported as volume")
		if parent:
			self.addCard("PLACE", [name, 1, parent, "", x, y, z], comment)

	#-----------------------------------------------------------------------
	# Regions & Volumes
	#-----------------------------------------------------------------------
	def convertRegions(self, blkregion, world):
		self._comment = "\n--------------------------------------------------" \
				"\n   Volumes" \
				"\n--------------------------------------------------"
		# Write the basic volumes
		self.volumeCard(world, comment=world.comment())
		#self.writeline(":VIS %s OFF"%(world.name()))

		#self.volumeCard(void, world.name(), comment=void.comment())
		#self.writeline(":VIS %s OFF"%(void.name()))

		card = None
		self.used(blkregion)
		try:
			self.used(blkregion[Card.ASSIGNMAT])
		except (KeyError, TypeError):
			pass
		for region in self.input["REGION"]:
			if region.notActive() or region==blkregion: continue
			self._reg.append(region.name())
			self._enable = region.notIgnore()
			material = region[Card.MATERIAL]
			try:
				self.used(region[Card.ASSIGNMAT])
			except (KeyError, TypeError):
				pass
			expr = "\n".join([x.strip() for x in region.expr().splitlines()])
			if region.name()=="&":
				self._reg.remove(region.name())
				# continuation of previous card
				card.setComment(card.comment()+"\n"+region.comment())
				card.setExpr(card.expr()+"\n"+expr)
				self.used(region)
			else:
				if expr and expr[0] == "|": expr = expr[1:]
				lattice = region[Card.LATTICE]
				if lattice:
					g4mat = ""
				elif material:
					g4mat = self.material(material.name())
				else:
					g4mat = "G4_Galactic"
				# FIXME
				if region[Card.ASSIGNMAT]:
					if region[Card.ASSIGNMAT].intWhat(5) == 1.0:
						self.fields.append(region.name())
				card = MoiraCard("CELL",[region.name(),
							world.name(),
							g4mat],
							#field],
						region.comment())
				card.setExpr(expr)
				self.add(card, self._enable)
				self.used(region)
				if lattice:
					name = lattice.name()
					inv = True
					if name[0] == "-":
						inv = not inv
						name = name[1:]
					rotm = self.input.getTransformation(name, self._rotdefi, inv)
					# try to get the real name used based on the id
					rid  = self._rotdefi[name]
					name = self._rotid[rid]
					if inv: name = Fluka2Moira.INVMATRIX%(name)
					dx = rotm[0][3]
					dy = rotm[1][3]
					dz = rotm[2][3]
					self.addCard("LATTICE",
							[region.name(), name,
							dx, dy, dz],
							lattice.comment())
					self.lattices.append(region.name())

	#-----------------------------------------------------------------------
	# Process all other cards (except geometry)
	#-----------------------------------------------------------------------
	def convertAllOtherCards(self):
		self._beamposShape = ""
		self._beamPos = bmath.Vector(0,0,0)
		self._beamDir = bmath.Vector(0,0,1)
		self._source  = False
		self._ww = []
		self._Z  = None
		self._A  = 1

		# Pre-process some cards
		for card in self.input.cardlist:
			if card.notActive(): continue
			if card.tag == "BEAMPOS":
				self._beamposShape = card.sdum()
			if card.tag == "SOURCE":
				self._source = True
			if card.tag == "WW-FACTO":
				self._ww.append(card.numWhat(1))
				self._ww.append(card.numWhat(2))
				self._ww.append(card.what(4))
			if card.tag == "HI-PROPE":
				self._Z = card.intWhat(1)
				self._A = card.intWhat(2)

		# -------------------------------------------------------------
		def convertCards(comment, convertFunc, inp = self.input):
			self._comment = comment
			for card in inp.cardlist:
				if self.isused(card): continue
				self._enable = card.notIgnore()

				try:
					func = convertFunc[card.tag]
					if not card.notActive():
						func(self, card)
					else:
						continue
						# clone = card.clone()
						# clone.setEnable(False)
						# self.add(clone)
					self.used(card)
				except KeyError:
					pass

		convertCards(	"\n--------------------------------------------------"
				"\n   Scoring"
				"\n--------------------------------------------------",
				{ "SCORE"    : Fluka2Moira._convertSCORE,
				  "USERDUMP" : Fluka2Moira._convertUSERDUMP,
				  "USRBDX"   : Fluka2Moira._convertUSRBDX,
				  "USRYIELD" : Fluka2Moira._convertUSRYIELD,
				  "USRBIN"   : Fluka2Moira._convertUSRBIN,
				  "USRTRACK" : Fluka2Moira._convertUSRTRACK })

		# CLOSE SCORING (if any scorer was set)
		if (self._has_scorer):
			self.addCard("/tally/close")

		convertCards(	"\n--------------------------------------------------"
				"\n   Fields Created"
				"\n--------------------------------------------------",
				{ "MGNCREAT" : Fluka2Moira._convertMGNCREAT,
				  "MGNDATA"  : Fluka2Moira._convertMGNDATA})

		convertCards(	"\n--------------------------------------------------"
				"\n   Fields Assigned"
				"\n--------------------------------------------------",
				{ "MGNFIELD" : Fluka2Moira._convertMGNFIELD })

		convertCards(	"\n--------------------------------------------------"
				"\n--------------------------------------------------",
				{ "USRGCALL" : Fluka2Moira._convertUSRGCALL })

		convertCards(	"\n--------------------------------------------------"
				"\n   Biasing"
				"\n--------------------------------------------------",
				{ "BIASING"   : Fluka2Moira._convertBIASING ,
				  "WW-THRES"  : Fluka2Moira._convertWW })

		convertCards(	"\n--------------------------------------------------"
				"\n   EM Cuts"
				"\n--------------------------------------------------",
				{ "EMFCUT"   : Fluka2Moira._convertEMCut
				  })

		convertCards(	"\n--------------------------------------------------"
				"\n   Hadrons/muons/Neutrinos Cuts"
				"\n--------------------------------------------------",
				{ "PART-THR"  : Fluka2Moira._convertHCut
				  # "DELTARAY"  : Fluka2Moira._convertHCut
				  })
		if (self._applycuts):
			self.addCard("/process/em/applyCuts", ['true'])

		self._comment = "\n--------------------------------------------------" \
				"\n   Verbose" \
				"\n--------------------------------------------------"

		self.addCard("/cuts/verbose", [0])
		self.addCard("/material/verbose", [0])
		self.addCard("/process/had/verbose", [0])
		self.addCard("/process/em/verbose", [0])
		self.addCard("/tracking/verbose", [0])

		self._comment = "\n--------------------------------------------------" \
				"\n   Initialize" \
				"\n--------------------------------------------------"

		self.addCard("/cuts/close")
		self.addCard("/run/initialize")

		# -------------------------------------------------------------
		# After /run/initialize
		# -------------------------------------------------------------

		# -------------------------------------------------------------
		# Loop over all cards to convert
		# -------------------------------------------------------------
		convertCards("\n--------------------------------------------------"
			"\n   Primary Source"
			"\n--------------------------------------------------",
			{ "BEAM"     : Fluka2Moira._convertBEAM,
			  "BEAMAXES" : Fluka2Moira._convertBEAMAXES,
			  "BEAMPOS"  : Fluka2Moira._convertBEAMPOS,
			  "SOURCE"   : Fluka2Moira._convertSOURCE})

		convertCards("\n--------------------------------------------------"
			"\n   Start run"
			"\n--------------------------------------------------",
			{ "RANDOMIZ" : Fluka2Moira._convertRANDOMIZ,
			  "START"    : Fluka2Moira._convertSTART })

		for card in self.input.cardlist:
			if isinstance(card, MoiraCard) and card.tag != "/run/numberOfThreads":
				self.add(card.clone())

	#-----------------------------------------------------------------------
	# Card conversion methods
	#-----------------------------------------------------------------------
	def _convertSCORE(self, card):
		self._has_scorer = True
		whats = [card.what(1), card.what(2), card.what(3), card.what(4)]
		whats = [x for x in whats if x] # Remove empty elements
		for i, w in enumerate(whats):
			if w in ("ENERGY"):
				if self._last_geo != Fluka2Moira.volume_geo \
				   or self._last_reg != "*":
					volumeCard = MoiraCard("VOLUME", ["*"])
					self.add(volumeCard, True)
					self._last_geo = Fluka2Moira.volume_geo
					self._last_reg = "*"

				self.addCard("/tally/file",["energy.hist"])
				name = "score"
				new = MoiraCard("TOTAL_ENERGY_DEPOSIT", [name, "GeV"], card.comment())
				self.add(new, self._enable)
			else: # XXX Stars scoring
				self.notImplemented(card, i+1)
				# clone = card.clone()
				# clone.setEnable(False)
				# self.add(clone)
				return

	#-----------------------------------------------------------------------
	def _convertBEAM(self, card):
		if self._source: return
		particle = card.name() if card.name() != '' else "PROTON"
		self.addCard("/primary/select",["gps"],
				"Primary particle generator")
		self.addCard("/gps/particle",
				[Moira.particleFromFluka(particle)],
				card.comment())
		if particle in ("ISOTOPE","HEAVYION"):
			self.addCard("/gps/ion", [self._Z, self._A])

		# We should keep the formulas if any
		# energy = card.numWhat(1)	# or momentum
		energy = card.numWhat(1) if card.numWhat(1) != 0.0 else -200 # 200 GeV Default
		dp     = card.numWhat(2) if card.numWhat(1) != '' else 0 # no spread in p
		dphi   = card.numWhat(3) if card.numWhat(1) != '' else 0 # no angular spread
		dx     = card.numWhat(4) if card.numWhat(1) != '' else 0
		dy     = card.numWhat(5) if card.numWhat(1) != '' else 0
		shape  = card.numWhat(6)
		m      = Moira.massParticle(Moira.particleFromFluka(particle))

		if particle == "HEAVYION":
			m_p = Moira.massParticle(Moira.particleFromFluka("PROTON")) * self._Z
			m_n = Moira.massParticle(Moira.particleFromFluka("NEUTRON")) * (self._A - self._Z)
			# breakpoint()
			isotope    = Element.iso(self._Z, self._A)
			# print(m_p, m_n, isotope.massex)
			#print(m_p)
			#print(m_n)
			#print(1e-3*float(isotope.massex))
			try:
				m = m_p + m_n + 1e-3*float(isotope.massex) # 1e-3: MeV->GeV
			except:
				#print("Mass excess not in Data base for "+"Z = "+str(self._Z)+" A = "+str(self._A)+".")
				raise Exception("Check if the isotope exists in database (or in the universe) | press ctrl+r to continue")

		elif particle == "ISOTOPE":
			self.addCard("/gps/energy",[0.0,"GeV"])
			return

		ep = ''
		if energy > 0:
			ep = "momentum"
			p  = energy * self._A
			# energy = math.sqrt(p**2 + m**2) - m
		else:
			ep     = "energy"
			energy = -energy * self._A
			p      = math.sqrt(energy * (energy+2*m))

		# Energy/Momentum spread
		self._comment = "Energy distribution"
		if dp < 0:
			# dE = p/(m+energy) * abs(dp)
			self.addCard("/gps/mom/type",["Gauss"])
			self.addCard("/gps/mom/mono",[bmath.format(p), "GeV"])
			# self.addCard(f"/gps/{ep[:3]}/sigma",[bmath.format(dE/fwhm), "GeV"])
			self.addCard("/gps/mom/sigma",[bmath.format(-dp/fwhm), "GeV"])
		elif dp == 0:
			self.addCard(f"/gps/{ep}",[bmath.format(energy),"GeV"])
		else:
			self.addCard(f"/gps/{ep}",[bmath.format(energy),"GeV"])
			self.addCard(f"/gps/{ep[:3]}/type",["Lin"])
			# minE = math.sqrt((p-dp/2)**2 + m**2) - m
			# maxE = math.sqrt((p+dp/2)**2 + m**2) - m
			minE = (p-dp)/2
			maxE = (p+dp)/2
			self.addCard(f"/gps/{ep[:3]}/min",[bmath.format(minE), "GeV"])
			self.addCard(f"/gps/{ep[:3]}/max",[bmath.format(maxE), "GeV"])
			self.addCard(f"/gps/{ep[:3]}/gradient", [0.0])
			self.addCard(f"/gps/{ep[:3]}/intercept",[2.0])
		# Divergence
		self._comment = "Angular distribution"
		if dphi < 0:
			self.addCard("/gps/ang/type",["beam1d"])
			self.addCard("/gps/ang/sigma_r",[-dphi/fwhm, "mrad"])
		elif dphi > 1000:
			self.addCard("/gps/ang/type",["iso"])
		elif dphi >= 0.0:
			self.addCard("/gps/ang/type",["iso"])
			self.addCard("/gps/ang/maxtheta",[dphi/2, "mrad"])
		# Beam Shape
		if self._beamposShape in Fluka2Moira.BEAMPOS_ORIENTATION:
			if shape < 0:
				self._comment = "Position distribution"
				self.addCard("/gps/pos/type",["Plane"])
				self.addCard("/gps/pos/shape",["Annulus"])
				self.addCard("/gps/pos/radius",[dx,"cm"])
				self.addCard("/gps/pos/inner_radius",[dy,"cm"])
			# elif shape > 0:
			elif shape == 0.0 and dx == 0 and dy == 0 :
				pass
			else:
				self._comment = "Position distribution"
				self.addCard("/gps/pos/type",["Beam"])
				if dx < 0:
					self.addCard("/gps/pos/sigma_x",[-dx/fwhm, "cm"])
				elif dx > 0:
					self.addCard("/gps/pos/halfx",[dx/2, "cm"])
				if dy < 0:
					self.addCard("/gps/pos/sigma_y",[-dy/fwhm, "cm"])
				elif dy > 0:
					self.addCard("/gps/pos/halfy",[dy/2, "cm"])
			# else:
				# self.addCard("/gps/pos/type",["Point"])

	#-----------------------------------------------------------------------
	def _convertSOURCE(self, card):
		self.addCard("/primary/select", ["data"])
		self.addCard("/primary/units", ["cm", "GeV", "s"])
		self.addCard("/primary/load", [card.sdum()+".dat"]) # FIXME Find a way to convert automatically

	#-----------------------------------------------------------------------
	def _convertBEAMPOS(self, card):
		if self._source: return
		if card.sdum() in Fluka2Moira.BEAMPOS_ORIENTATION:
			x = card.numWhat(1)
			y = card.numWhat(2)
			z = card.numWhat(3)
			self._beamPos = bmath.Vector(x,y,z)
			self.addCard("/gps/pos/centre",[x,y,z,"cm"],card.comment())
			u = card.numWhat(4)
			v = card.numWhat(5)
			tsqr = math.sqrt(u*u + v*v)	# do it ala fluka beacrd.f
			if abs(u)>1 or abs(v)>1 or abs(tsqr)>1:
				u /= tsqr
				v /= tsqr
				w = 0
			else:
				w = math.sqrt(max(0, (1-tsqr)*(1+tsqr)))
				if card.sdum()=="NEGATIVE":
					w = -w
			self._beamDir = bmath.Vector(u, v, w)
			self._beamDir.normalize()

			if self.input["BEAMAXES"]:
				# skip beam direction it will be set by the beam axes
				pass
			elif u or v:
				X = self._beamDir.orthogonal()
				X.normalize()
				Y = self._beamDir ^ X
				self.addCard("/gps/ang/rot1", [Y[0], Y[1], Y[2]])
				self.addCard("/gps/ang/rot2", [X[0], X[1], X[2]])
			elif card.sdum()=="":
				self.addCard("/gps/ang/rot1",[1,  0, 0])
				self.addCard("/gps/ang/rot2",[0, -1, 0])
		elif card.sdum() == "FLOOD":
			r = card.numWhat(1)
			self.addCard("/gps/pos/type",  ["Surface"])
			self.addCard("/gps/pos/shape", ["Sphere"])
			self.addCard("/gps/pos/radius",[r, "cm"])
			self.addCard("/gps/pos/centre",
					[self.beamPos[0], self.beamPos[1], self.beamPos[2], "cm"])
		elif card.sdum() == "CART-VOL":
			Vin  = bmath.Vector(card.numWhat(1),card.numWhat(3),card.numWhat(5))
			Vout = bmath.Vector(card.numWhat(2),card.numWhat(4),card.numWhat(6))
			half = (Vout - Vin)/2
			pos = self._beamPos + (Vout + Vin)/2
			# self.addCard("/gps/pos/type",["Volume"])
			self.addCard("/gps/pos/type",["Beam"], card.comment())
			self.addCard("/gps/pos/halfx",[half[0], "cm"])
			self.addCard("/gps/pos/halfy",[half[1], "cm"])
			self.addCard("/gps/pos/halfz",[half[2], "cm"])
			self.addCard("/gps/pos/centre",[pos[0], pos[1], pos[2], "cm"])
		elif card.sdum() == "CYLI-VOL":
			#rin  = card.numWhat(1)
			rout = card.numWhat(2)
			hin  = card.numWhat(3)
			hout = card.numWhat(4)
			# XXX Or volume?
			self.addCard("/gps/pos/type",["Surface"], card.comment())
			# self.addCard("/gps/pos/type",["Volume"]) # XXX Or volume?
			self.addCard("/gps/pos/shape",["Cylinder"])
			self.addCard("/gps/pos/radius",[rout, "cm"])
			self.addCard("/gps/pos/halfz",[(hout-hin)/2, "cm"])
			self.addCard("/gps/pos/centre",
					[self._beamPos[0],
					 self._beamPos[1],
					 self._beamPos[2]+(hout+hin)/2,
					 "cm"])
		elif card.sdum() == "SPHE-VOL":
			#rin  = card.numWhat(1)
			rout = card.numWhat(2)
			self.addCard("/gps/pos/type",["Volume"], card.comment())
			self.addCard("/gps/pos/shape",["Sphere"])
			self.addCard("/gps/pos/radius",[rout, "cm"])
			self.addCard("/gps/pos/centre",
					[self.beamPos[0], self.beamPos[1], self.beamPos[2], "cm"])
		else:
			self.notImplemented(card,0,f"uknown SDUM={card.sdum()}")

	#-----------------------------------------------------------------------
	def _convertBEAMAXES(self, card):
		X = bmath.Vector(card.numWhat(1), card.numWhat(2), card.numWhat(3))
		if X.normalize() < 1e-10:
			X = bmath.Vector.Xo
		Z = bmath.Vector(card.numWhat(4), card.numWhat(5), card.numWhat(6))
		if Z.normalize() < 1e-10:
			Z = bmath.Vector.Zo
		Y = Z ^ X
		X = Y ^ Z
		self.addCard("/gps/pos/rot1", [X[0], X[1], X[2]])
		self.addCard("/gps/pos/rot2", [Y[0], Y[1], Y[2]])
		# FIXME to be verified!!!
		self._beamDir = bmath.Vector(
				X[0]*self._beamDir[0]+Y[0]*self._beamDir[1]+Z[0]*self._beamDir[2],
				X[1]*self._beamDir[0]+Y[1]*self._beamDir[1]+Z[1]*self._beamDir[2],
				X[2]*self._beamDir[0]+Y[2]*self._beamDir[1]+Z[2]*self._beamDir[2])
		self._beamDir.normalize()
		X = self._beamDir.orthogonal()
		X.normalize()
		Y = self._beamDir ^ X
		self.addCard("/gps/ang/rot1", [Y[0], Y[1], Y[2]])
		self.addCard("/gps/ang/rot2", [X[0], X[1], X[2]])

	#-----------------------------------------------------------------------
	def _convertBIASING(self, card):
		part  = card.intWhat(1)
		imp   = card.what(3)
		reg   = card.what(4)
		if part == 0: part = "all"
		if part == 1: part = "hadrons"
		if part == 2: part = "em"
		if part == 3: part = "neutrons"
		self.addCard("/bias/region",[reg,part,imp],card.comment())

	#-----------------------------------------------------------------------
	def _convertWW(self, card):
		E2   = card.numWhat(1)
		E1   = card.numWhat(2)
		f    = card.numWhat(3)
		#part = card.what(4)
		w1   = self._ww[0]
		w2   = self._ww[1]
		reg  = self._ww[2]
		self.addCard("/bias/ww",[reg, "all", w1, w2, E1, E2, f, "Mev"], card.comment(), card.notIgnore())

	#-----------------------------------------------------------------------
	# Magnetic Fields
	# FIXME this specific to our BMI interpretation
	#       we should make somehow a plugin
	#-----------------------------------------------------------------------
	def _convertUSRGCALL(self, card):
		if card.intWhat(1) != 100: return
		reg   = card.intWhat(2) - 1
		#step  = card.what(4)
		grad  = card.what(5)
		offX  = card.what(7)
		offY  = card.what(8)
		offZ  = card.what(9)
		alpha = card.what(10)
		beta  = card.what(11)
		gamma = card.what(12)
		fldtype = card.name()
		# if fldtype not in self.includes:
			# self.add(Card("#include", [fldtype+".macro"] )) # XXX Find input with pattern?
			# self.add(Card("#endinclude"))
			# self.includes.append(fldtype)
		self.addCard("/field/magnetic/assign",   [self.lattices[reg], fldtype, grad, "T"])
		self.addCard("/field/magnetic/offset",   [offX, offY, offZ, "cm"])
		self.addCard("/field/magnetic/rotation", [alpha, beta, gamma, "rad"])
		self.used(card)

	#-----------------------------------------------------------------------
	def _convertUSRBDX(self, card):
		self._has_scorer = True

		name  = card.name()
		utype = card.intWhat(1,1)
		part  = card.what(2)
		unit  = abs(card.intWhat(3))
		reg1  = card.what(4)
		reg2  = card.what(5)
		area  = card.numWhat(6) if card.numWhat(6) > 0 else 1
		Emax  = card.numWhat(7)
		Emin  = card.numWhat(8)
		Ebins = card.intWhat(9)
		Amax  = card.numWhat(10)
		Amin  = card.numWhat(11)
		Abins = card.intWhat(12)

		if unit != self._last_unit:
			self.addCard("/tally/file", [f"unit{unit}.hist"])
			self._last_unit = unit

		if utype > 80:
			utype -= 100
			type_name = "flux"
		else:
			type_name = "current"

		direction = "bothWays"
		if utype >= 8:
			utype -= 10
		else:
			direction = "exitOnly"

		if self._last_geo != Fluka2Moira.surface_geo \
		   or reg1 != self._last_reg \
			      or reg2 != self._last_reg2 \
					 or area != self._last_norm:
			surfaceCard = MoiraCard("SURFACE", [reg1, reg2])
			surfaceCard["norm.normalizationFactor"] = bmath.format(area)
			surfaceCard["norm.Unit"] = "cm2"
			self.add(surfaceCard, True)
			self._last_geo = Fluka2Moira.surface_geo
			self._last_reg = reg1
			self._last_reg2 = reg2
			self._last_norm = area

		if type_name == "flux":
			new = MoiraCard("SURF_FLUENCE", [name, "/cm2/GeV"], card.comment())
		elif type_name == "current":
			new = MoiraCard("SURF_CURRENT", [name, "/cm2/GeV"], card.comment())
		else:
			self.notImplemented(card,1)
			clone = card.clone()
			clone.setEnable(False)
			self.add(clone)
			return

		new["particle"] = Moira.particleFromFluka(part)
		# NB: USRBDX is always on surface normal
		new["direction.directionKind"] = "surfaceNormal"
		if direction == "exitOnly":
			new["direction.isExitOnly"] = "exitOnly"
		self.add(new, self._enable)

		logE = utype < 0
		isLogE = "log" if logE else "linear"
		new = MoiraCard("DIFFERENTIAL_IN")
		new["differentialIn.DifferentialVariable"] = "kineticEnergy"
		new["differentialIn.Log"] = isLogE
		new["differentialIn.NumBins"] = Ebins
		new["differentialIn.Min"] = bmath.format(Emin)
		new["differentialIn.Max"] = bmath.format(Emax)
		new["differentialIn.ExtremaUnit"] = "GeV"
		self.add(new, self._enable)

		if Abins > 0:
			logA = abs(utype)==2
			isLogA = "log" if logA else "linear"
			if Amax == 0:
				Amax = 2*math.pi
			new = MoiraCard("DIFFERENTIAL_IN")
			new["differentialIn.DifferentialVariable"] = "solidAngle"
			new["differentialIn.Log"] = isLogA
			new["differentialIn.NumBins"] = Abins
			new["differentialIn.Min"] = bmath.format(Amin)
			new["differentialIn.Max"] = bmath.format(Amax)
			new["differentialIn.ExtremaUnit"] = "sr"
			self.add(new, self._enable)
			#self.addCard("DIFFERENTIAL_IN",
					#["solidAngle", isLogA, Abins, Amin, Amax, "sr"])

		self.used(card)

	#-----------------------------------------------------------------------
	# Convert USRYIELD card from FLUKA to MOIRA.
	#-----------------------------------------------------------------------
	def _convertUSRYIELD(self, card):
		self._has_scorer = True

		# PARSE THE WHATS
		name  = card.name()
		#yield_or_beam_def = card.sdum()
		diff_variables = card.intWhat(1)
		particle = card.what(2)
		output_unit = abs(card.intWhat(3))
		region1 = card.what(4)
		region2 = card.what(5)
		norm = card.numWhat(6) if card.numWhat(6) > 0 else 1
		first_diff_variable_max = card.numWhat(7)
		first_diff_variable_min = card.numWhat(8)
		first_diff_variable_num_bins = card.intWhat(9) if card.intWhat(9) > 0 else 50
		second_diff_variable_max = card.numWhat(10)
		second_diff_variable_min = card.numWhat(11)
		second_diff_variable_num_bins = 1
		weight = card.intWhat(12) if card.intWhat(12) > 0 else 1

		# FIRST AND SECOND DIFF VARIABLES
		# Get the first and second diff variables integers from `diff_variables`
		# `diff_variables` is second_diff_variable * 100 + first_diff_variable
		first_diff_variable_is_log = False
		second_diff_variable_is_log = False

		second_diff_variable = diff_variables // 100
		first_diff_variable = diff_variables - second_diff_variable * 100

		# Special FLUKA treatment to detect is_log for first diff variable
		if first_diff_variable >= 50:
			first_diff_variable_is_log = True
			first_diff_variable = 100 - first_diff_variable
			if first_diff_variable_min == 0:
				first_diff_variable_min = 0.001

		# Get the first and second diff variables names, units, and values from the FLUKA integers
		first_diff_variable_name, first_diff_variable_unit, first_diff_variable_min, first_diff_variable_max = \
					self._convertDiffVariableHelper(first_diff_variable,
									first_diff_variable_min,
									first_diff_variable_max)

		second_diff_variable_name, second_diff_variable_unit, second_diff_variable_min, second_diff_variable_max = \
					self._convertDiffVariableHelper(second_diff_variable,
									second_diff_variable_min,
									second_diff_variable_max)

		# PARTICLE(S)
		particle_name = "ALL-PART"
		if card.intWhat(2) < -80:
			particle = -(card.intWhat(2) + 100)
		particle_name = Moira.particleFromFluka(particle)

		# OUTPUT FILE
		if output_unit != self._last_unit:
			self.addCard("/tally/file", [f"unit{output_unit}.hist"])
			self._last_unit = output_unit

		# SURFACE CARD
		if self._last_geo != Fluka2Moira.surface_geo \
		   or region1 != self._last_reg \
			      or region2 != self._last_reg2 \
					 or norm != self._last_norm:
			surface_card = MoiraCard("SURFACE", [region1, region2])
			surface_card["norm.normalizationFactor"] = bmath.format(norm)
			surface_card["norm.Unit"] = "cm2"
			self.add(surface_card, True)
			self._last_geo = Fluka2Moira.surface_geo
			self._last_reg = region1
			self._last_reg2 = region2
			self._last_norm = norm

		# SCORER CARD
		weight_value = weight - (weight // 100) * 100
		scorer_name = "undefined"
		weight_name = "unsupported"
		# d2N/dx1dx2
		if weight_value == 3:
			scorer_name = "SURF_CURRENT"
		# sec(theta) d2N/dx1dx2
		elif weight_value == 6:
			scorer_name = "SURF_FLUENCE"
		# other weight
		else:
			scorer_name = "SURF_GENERIC_SCORER"
			# NB: Here, one can add any weight supported by Moira!
			#weight_name = "unsupported"

		if scorer_name != "undefined":
			scorer_card = MoiraCard(scorer_name, [name, ""], card.comment())
		else:
			self.notImplemented(card,1)
			clone = card.clone()
			clone.setEnable(False)
			self.add(clone)
			return

		if scorer_name == "SURF_GENERIC_SCORER":
			scorer_card["weight"] = weight_name
		scorer_card["particle"] = particle_name
		scorer_card["direction.isExitOnly"] = "bothWays"
		# NB: Only support beamDirection (USRYIELD default) for now
		scorer_card["direction.directionKind"] = "beamDirection"
		self.add(scorer_card, self._enable)

		# FIRST DIFF VARIABLE CARD
		first_diff = MoiraCard("DIFFERENTIAL_IN")
		first_diff["differentialIn.DifferentialVariable"] = first_diff_variable_name
		first_diff["differentialIn.Log"] = "log" if first_diff_variable_is_log else "linear"
		first_diff["differentialIn.NumBins"] = first_diff_variable_num_bins
		first_diff["differentialIn.Min"] = bmath.format(first_diff_variable_min)
		first_diff["differentialIn.Max"] = bmath.format(first_diff_variable_max)
		first_diff["differentialIn.ExtremaUnit"] = first_diff_variable_unit
		self.add(first_diff, self._enable)

		# SECOND DIFF VARIABLE CARD
		second_diff = MoiraCard("DIFFERENTIAL_IN")
		second_diff["differentialIn.DifferentialVariable"] = second_diff_variable_name
		second_diff["differentialIn.Log"] = "log" if second_diff_variable_is_log else "linear"
		second_diff["differentialIn.NumBins"] = second_diff_variable_num_bins
		second_diff["differentialIn.Min"] = bmath.format(second_diff_variable_min)
		second_diff["differentialIn.Max"] = bmath.format(second_diff_variable_max)
		second_diff["differentialIn.ExtremaUnit"] = second_diff_variable_unit
		self.add(second_diff, self._enable)

		self.used(card)

	#-----------------------------------------------------------------------
	# Convert FLUKA diff variables integers to info supported by MOIRA.
	# NB: Straightforward to add support for extra diffVariables here.
	#-----------------------------------------------------------------------
	def _convertDiffVariableHelper(self,
				       diff_variable,
				       param_min,
				       param_max):

		diff_variable_name = "unknown"
		diff_variable_unit = "unknown"
		diff_variable_min = param_min
		diff_variable_max = param_max

		if diff_variable == 1:
			diff_variable_name = "kineticEnergy"
			diff_variable_unit = "GeV"
		elif diff_variable == 14:
			diff_variable_name = "solidAngle"
			diff_variable_unit = "sr"
			# Convert FLUKA angle in radians, to solid angle: 2*PI (1 - cos(theta))
			diff_variable_min = 2 * math.pi * (1 - math.cos(param_min))
			diff_variable_max = 2 * math.pi * (1 - math.cos(param_max))
		elif diff_variable == 24:
			diff_variable_name = "solidAngle"
			diff_variable_unit = "sr"
			# Convert FLUKA angle in degrees to solid angle: 2*PI (1 - cos(theta))
			diff_variable_min = 2 * math.pi * (1 - math.cos(param_min * math.pi / 180))
			diff_variable_max = 2 * math.pi * (1 - math.cos(param_max * math.pi / 180))
		elif diff_variable == 22:
			diff_variable_name = "charge"
			diff_variable_unit = "eplus"
		elif diff_variable == 23:
			diff_variable_name = "LET"
			diff_variable_unit = "(MeV*cm2/mg)"
			# Convert FLUKA value from keV/(μm*g/cm3) to (MeV*cm2/mg)
			diff_variable_min = param_min * 0.01
			diff_variable_max = param_max * 0.01
		elif diff_variable == 33:
			diff_variable_name = "globalTime"
			diff_variable_unit = "s"

		return diff_variable_name, diff_variable_unit, diff_variable_min, diff_variable_max

	#-----------------------------------------------------------------------
	def _convertUSRTRACK(self, card):
		self._has_scorer = True

		name  = card.name()
		utype = card.intWhat(1)
		part  = card.what(2)
		unit  = abs(card.intWhat(3))
		reg   = card.what(4)
		vol   = card.numWhat(5) if card.numWhat(5) > 0 else 1
		Ebins = card.intWhat(6)
		Emax  = card.numWhat(7)
		Emin  = card.numWhat(8)

		if unit != self._last_unit:
			self.addCard("/tally/file", [f"unit{unit}.hist"])
			self._last_unit = unit

		if self._last_geo != Fluka2Moira.volume_geo \
		   or reg != self._last_reg \
			     or vol != self._last_norm:
			volumeCard = MoiraCard("VOLUME", [reg])
			volumeCard["norm.normalizationFactor"] = bmath.format(vol)
			volumeCard["norm.Unit"] = "cm3"
			self.add(volumeCard, True)
			self._last_geo = Fluka2Moira.volume_geo
			self._last_reg = reg
			self._last_norm = vol

		new = MoiraCard("TRACK_LENGTH", [name, "/cm2/GeV"], card.comment())
		new["particle"] = Moira.particleFromFluka(part)
		self.add(new, self._enable)

		logE = utype < 0
		isLogE = "log" if logE else "linear"
		new = MoiraCard("DIFFERENTIAL_IN")
		new["differentialIn.DifferentialVariable"] = "kineticEnergy"
		new["differentialIn.Log"] = isLogE
		new["differentialIn.NumBins"] = Ebins
		new["differentialIn.Min"] = bmath.format(Emin)
		new["differentialIn.Max"] = bmath.format(Emax)
		new["differentialIn.ExtremaUnit"] = "GeV"
		self.add(new, self._enable)
		#self.addCard("differentialIn", ["kineticEnergy", isLogE, Ebins, Emin, Emax, "GeV"])

		self.used(card)

	#-----------------------------------------------------------------------
	def _convertUSERDUMP(self, card):
		self.addCard("/tally/userdump", [card.sdum(),""], card.comment())
		self.used(card)

	#-----------------------------------------------------------------------
	def _convertUSRBIN(self, card):
		self._has_scorer = True

		name  = card.name()
		utype = card.intWhat(1)
		part  = card.what(2)
		unit  = abs(card.intWhat(3))
		xmax  = card.numWhat(4)
		ymax  = card.numWhat(5)
		zmax  = card.numWhat(6)
		xmin  = card.numWhat(7)
		ymin  = card.numWhat(8)
		zmin  = card.numWhat(9)
		nx    = max(card.intWhat(10),1)
		ny    = max(card.intWhat(11),1)
		nz    = max(card.intWhat(12),1)

		# MESH SCORING
		if utype in (0,1,10,11):

			# ADD MESH
			# Only when needed:
			# if last mesh is already correctly defined, just add scorers to it.
			if self._last_geo != Fluka2Moira.mesh_geo \
			   or xmax != self._last_mesh_xmax \
			   or ymax != self._last_mesh_ymax \
			   or zmax != self._last_mesh_zmax \
			   or xmin != self._last_mesh_xmin \
			   or ymin != self._last_mesh_ymin \
			   or zmin != self._last_mesh_zmin \
			   or nx   != self._last_mesh_nx \
			   or ny   != self._last_mesh_ny \
			   or nz   != self._last_mesh_nz:

				# Cannot dump different meshes into same output unit.
				try:
					self._count_meshes_per_output_unit[unit] += 1
					counter = self._count_meshes_per_output_unit[unit]
					self.addCard("/tally/file",
						     [f"unit{unit}_{counter}.hist"], "")
				except KeyError:
					self._count_meshes_per_output_unit[unit] = 1
					counter = self._count_meshes_per_output_unit[unit]
					self.addCard("/tally/file",
						     [f"unit{unit}_{counter}.hist"], "")

				# CARTESIAN MESH
				if utype in (0, 10):
					new = MoiraCard("MESH.BOX",[f"box_{counter}"],card.comment())
					x  = (xmin+xmax)/2
					y  = (ymin+ymax)/2
					z  = (zmin+zmax)/2
					new["xAxis.numBinsX"] = nx
					new["xAxis.minX"] = bmath.format(xmin - x)
					new["xAxis.maxX"] = bmath.format(xmax - x)
					new["xAxis.unit"] = "cm"
					new["yAxis.numBinsY"] = ny
					new["yAxis.minY"] = bmath.format(ymin - y)
					new["yAxis.maxY"] = bmath.format(ymax - y)
					new["yAxis.unit"] = "cm"
					new["zAxis.numBinsZ"] = nz
					new["zAxis.minZ"] = bmath.format(zmin - z)
					new["zAxis.maxZ"] = bmath.format(zmax - z)
					new["zAxis.unit"] = "cm"
					# CYLINDRICAL MESH
				elif utype in (1,11):
					new = MoiraCard("MESH.CYLINDER",[f"cylinder_{counter}"],card.comment())
					x  = ymin
					y  = ymax
					z  = (zmin+zmax)/2
					new["rAxis.numBinsR"] = nx
					new["rAxis.maxR"] = bmath.format(xmax)
					new["rAxis.unit"] = "cm"
					new["phiAxis"] = ny
					new["zAxis.numBinsZ"] = nz
					new["zAxis.minZ"] = bmath.format(zmin - z)
					new["zAxis.maxZ"] = bmath.format(zmax - z)
					new["zAxis.unit"] = "cm"

				# TRANSLATE MESH
				new["translate.transX"] = bmath.format(x)
				new["translate.transY"] = bmath.format(y)
				new["translate.transZ"] = bmath.format(z)
				new["translate.Unit"] = "cm"
				self.add(new, True)

				self._last_geo = Fluka2Moira.mesh_geo
				self._last_mesh_xmax = xmax
				self._last_mesh_ymax = ymax
				self._last_mesh_zmax = zmax
				self._last_mesh_xmin = xmin
				self._last_mesh_ymin = ymin
				self._last_mesh_zmin = zmin
				self._last_mesh_nx = nx
				self._last_mesh_ny = ny
				self._last_mesh_nz = nz

			# ADD QUANTITY TO MESH
			if part == "ENERGY":
				#self.addCard("/tally/totalEnergyDeposit/add", [name+"Energy", "GeV/cm3"])
				#new = MoiraCard("TOTAL_ENERGY_DEPOSIT", [name+"_Energy", "GeV/cm3"], card.comment())
				new = MoiraCard("MESH_TOTAL_ENERGY_DEPOSIT", [name, "GeV"], card.comment())
				self.add(new, self._enable)
			elif part == "ALL-PART":
				#self.addCard("/tally/trackLength/add", [name+"TL", "/cm2"])
				#new = MoiraCard("TRACK_LENGTH", [name+"_TL", "/cm2"], card.comment())
				new = MoiraCard("MESH_TRACK_LENGTH", [name, "cm"], card.comment())
				self.add(new, self._enable)
			elif part == "BEAMPART":
				particle = Moira.particleFromFluka(self.beam)
				new = MoiraCard("MESH_TRACK_LENGTH", [name + "_" + particle[0], "cm"], card.comment())
				new["particle"] = particle
				self.add(new, self._enable)
			else: # XXX Used momentarily for the fluence
				particle = Moira.particleFromFluka(part)
				if particle:
					new = MoiraCard("MESH_TRACK_LENGTH",
							[name + "_" + particle[0], "cm"],
							card.comment())
					new["particle"] = particle
					self.add(new, self._enable)

		# 0D SCORING
		elif utype in (2,8,12,18):
			if unit != self._last_unit:
				self.addCard("/tally/file", [f"unit{unit}.hist"], "")
				self._last_unit = unit

			if self._last_geo != Fluka2Moira.volume_geo \
			   or self._last_reg != "*":
				volumeCard = MoiraCard("VOLUME", ["*"])
				self.add(volumeCard, True)
				self._last_geo = Fluka2Moira.volume_geo
				self._last_reg = "*"

			if part == "ENERGY":
				new = MoiraCard("TOTAL_ENERGY_DEPOSIT", [name, "GeV"],
						card.comment())
				self.add(new, self._enable)
			else:
				try:
					if part == "ALL-PART":
						new = MoiraCard("TRACK_LENGTH", [name, "cm"],
								card.comment())
						self.add(new, self._enable)
					else:
						particle = Moira.particleFromFluka(part)
						new = MoiraCard("TRACK_LENGTH", [name, "cm"],
								card.comment())
						new["particle"] = particle
						self.add(new, self._enable)
				except IndexError:
					self.notImplemented(card,1)
					clone = card.clone()
					clone.setEnable(False)
					self.add(clone)
					return
		else:
			self.notImplemented(card,1)
			clone = card.clone()
			clone.setEnable(False)
			self.add(clone)
			return

	#-----------------------------------------------------------------------
	def _convertMGNCREAT(self, card):
		name  = card.name()
		ftype = card.intWhat(1)
		r     = card.numWhat(2)
		Xo    = card.numWhat(3)
		Yo    = card.numWhat(4)
		sym   = card.intWhat(5)

		def _sym(s):
			if s == 1:
				out = "-Bx"
			elif s == 2:
				out = "-By"
			elif s == 3:
				out = "-Bz"
			else:
				out = "none"
			return out

		Bx = _sym(sym%10)
		By = _sym((sym%100)//10)
		Bz = _sym(sym//100)

		#azm     = card.numWhat(7)
		rbend   = card.numWhat(8)
		dbend   = "-x" if rbend < 0.0 else "+x"
		sagitta = card.numWhat(9)
		Nx      = card.intWhat(10)
		Ny      = card.intWhat(11)
		Nz      = card.intWhat(12)
		Xmin    = card.numWhat(13)
		Xmax    = card.numWhat(16)
		Ymin    = card.numWhat(14)
		Ymax    = card.numWhat(17)
		Zmin    = card.numWhat(15)
		Zmax    = card.numWhat(18)
		if card.numWhat(1) < 100: # Multipolar Field
			# self.add(Card("#include", [name+".macro"], self._comment))
			self.addCard("/field/magnetic/create", [name, "2D"])
			self.addCard("/field/magnetic/symmetry", [Bx, By, Bz])
			self.addCard("/field/magnetic/core", [Fluka2Moira.FIELDS[ftype], r , Xo, Yo, "cm"])
			#self.addCard("/field/magnetic/xgrid", [2, -1 , 1, "cm"]) # XXX To Modify in Fluka++
			#self.addCard("/field/magnetic/ygrid", [2, -1 , 1, "cm"]) # Disabled check() and added minor changes
			#self.addCard("/field/magnetic/data", [0.0, 0.0 , 0.0])
			#self.addCard("/field/magnetic/data", [0.0, 0.0 , 0.0])
			#self.addCard("/field/magnetic/data", [0.0, 0.0 , 0.0])
			#self.addCard("/field/magnetic/data", [0.0, 0.0 , 0.0])
			# self.add(Card("#endinclude"))
		else:
			# self.add(Card("#include", [name+".macro"], self._comment))
			if card.numWhat(1) < 300: #2D Fields
				self.addCard("/field/magnetic/create", [name, "2D"])
				self.addCard("/field/magnetic/core", [Fluka2Moira.FIELDS[int(ftype%200)], r , Xo, Yo, "cm"])
				self.addCard("/field/magnetic/symmetry", [Bx, By, Bz])
				self.addCard("/field/magnetic/xgrid", [Nx, Xmin , Xmax, "cm"])
				self.addCard("/field/magnetic/ygrid", [Ny, Ymin , Ymax, "cm"])
				if rbend:
					self.addCard("/field/magnetic/bend" , [dbend, abs(rbend), "cm"])
					self.addCard("/field/magnetic/sagitta" , [sagitta, "cm"])
				(Fluka2Moira.FIELDS)[name] = [2, Nx*Ny]
			else: # 3D Fields
				self.addCard("/field/magnetic/create", [name, "3D"])
				self.addCard("/field/magnetic/symmetry", [Bx, By, Bz])
				self.addCard("/field/magnetic/xgrid", [Nx, Xmin , Xmax, "cm"])
				self.addCard("/field/magnetic/ygrid", [Ny, Ymin , Ymax, "cm"])
				self.addCard("/field/magnetic/zgrid", [Nz, Zmin , Zmax, "cm"])
				(Fluka2Moira.FIELDS)[name] = [3, Nx*Ny*Nz]
		self.used(card)

	#-----------------------------------------------------------------------
	def _convertMGNDATA(self, card):
		name = card.name()
		field = Fluka2Moira.FIELDS[name][0]
		lines = Fluka2Moira.FIELDS[name][1]
		if field == 2: #2D Fields
			for i in range(lines):
				self.addCard("/field/magnetic/data", [card.numWhat(i*2 + 1), card.numWhat(i*2 + 2)])
		elif field == 3: #3D Fields
			for i in range(lines):
				self.addCard("/field/magnetic/data", [card.numWhat(i*3 + 1), card.numWhat(i*3 + 2), card.numWhat(i*3 + 3)])
		# self.add(Card("#endinclude"))
		self.used(card)

	#-----------------------------------------------------------------------
	def _convertMGNFIELD(self, card):
		name     = card.name()
		strength = card.numWhat(1)
		rot      = card.what(2)
		#rpr      = card.intWhat(3) # Region, Prototype, Replica
		freg     = card.what(4)
		#toreg    = card.what(5)
		#step     = card.intWhat(6)
		if not name:
			Bx = card.numWhat(4)
			By = card.numWhat(5)
			Bz = card.numWhat(6)
			strength = math.sqrt(Bx*Bx + By*By + Bz*Bz)
			if strength == 0: return
			for freg in self.fields:
				self.addCard("/field/magnetic/assign", [freg, "dipole", strength, "T"])
				self.addCard("/field/magnetic/direction", [Bx/strength, By/strength, Bz/strength])
		else:
			self.addCard("/field/magnetic/assign", [freg, name, strength, "T"])
		for name, rt in self._rotdefi.items():
			if name != rot: continue
			self._rotid[rt] = name
			# rotm = self._rotdefi[rt]

			inv = True
			if name[0] == "-":
				inv = not inv
				name = name[1:]

			rotm = self.input.getTransformation(name, self._rotdefi, inv)
			# try to get the real name used based on the id
			rid  = self._rotdefi[name]
			name = self._rotid[rid]
			if inv: name = Fluka2Moira.INVMATRIX%(name)
			dx = rotm[0][3]
			dy = rotm[1][3]
			dz = rotm[2][3]

			rx,ry,rz = [bmath.format(math.degrees(x)) for x in rotm.getEulerRotation()]
			self.addCard("/field/magnetic/rotation", [rx, ry, rz, "deg"])
			self.addCard("/field/magnetic/offset", [dx, dy, dz, "cm"])
		self.used(card)

	#-----------------------------------------------------------------------
	def _convertEMCut(self, card):
		mats = list(Fluka.defaultMaterials.copy().keys())
		icrumats = list(Fluka.icruMaterials.copy().keys())
		nmats = list(self._materials)

		auxnmats = nmats.copy()
		for mt in auxnmats:
			if mt in mats:
				nmats.remove(mt)
			if mt in icrumats:
				nmats.remove(mt)
			if mt[0] == '-':
				nmats.remove(mt)
		if nmats:
			nmats.extend(icrumats)
			mats.extend(nmats)
		else: mats.extend(icrumats)
		mats.append('')

		name = card.name()
		m    = Moira.massParticle(Moira.particleFromFluka("ELECTRON"))
		ecut = abs(card.numWhat(1))
		if card.numWhat(1) > 0: ecut -= m # Kinetic energy is used in Moira for cuts
		gcut = card.numWhat(2)

		lregs = self._reg.copy()
		lregs.append('')

		if name == "":
			regi = card.what(4) if card.what(4) != "" else lregs[2]

			try:
				ri_ind = lregs.index(self._reg[0])
			except ValueError:
				ri_ind = lregs.index(lregs[2])

			regf   = card.what(5) if card.what(5) != "" else regi
			rf_ind = lregs.index(regf) + 1 if regf != "@LASTREG" else -1
			step   = card.intWhat(6) if card.intWhat(6) != 0 else 1
			for rg in lregs[ri_ind:rf_ind:step]:
				if rg =="&": continue
				self.addCard("/cuts/transport", [ecut, gcut, 'GeV', rg])
		elif name == "PROD-CUT":
			mati   = card.what(4) if card.what(4) != "" else mats[0]
			mi_ind = mats.index(mati)
			matf   = card.what(5) if card.what(5) != "" else mats[-1]
			mf_ind = mats.index(matf)+1 if matf != "@LASTMAT" else -1
			step   = card.intWhat(6) if card.intWhat(6) != 0 else 1
			for m in mats[mi_ind:mf_ind:step]:
				if m in list(self._materials):
					self.addCard("/cuts/production",
							[ecut, gcut, 0, 'GeV', self._materials[m]])
		else: #ANNH-THR, DELAYED, ELPO-THR, PHO2-THR, PHOT-THR, PROMPT
			# Not implemented yet
			return
		if name == "PROD-CUT" and not self._applycuts:
			self._applycuts = True

	#-----------------------------------------------------------------------
	def _convertHCut(self, card):
		parts = FlukaParticle.particles.copy()[:67] # just the particles
		ecut  = abs(card.numWhat(1))
		# if card.numWhat(1) > 0: ecut  # In case of momentum TODO

		parti  = card.what(2)
		pi_ind = parts.index(parti)
		partf  = card.what(3) if card.what(3) != "@LASTPAR" else parts[-1]
		partf  = card.what(2) if card.what(3) == "" else partf
		pf_ind = parts.index(partf)+1
		step   = card.intWhat(6) if card.intWhat(6) != 0 else 1

		for prt in parts[pi_ind:pf_ind:step]:
			if Moira.particleFromFluka(prt):
				self.addCard("/cuts/transporth",[Moira.particleFromFluka(prt), ecut, 'GeV'])

	#-----------------------------------------------------------------------
	def _convertRANDOMIZ(self, card):
		# FIXME doesn't work when a function is used e.g.  =seed
		# since it will setSeeds as as "=seed =seed" leading to g4 crash
		seed = card.numWhat(2)
		# Only add seed if it exists!
		if seed:
			self.addCard("/random/setSeeds",[f"{seed} {seed}"])

	#-----------------------------------------------------------------------
	def _convertSTART(self, card):
		no = card.intWhat(1) if card.intWhat(1) != 0.0 else 5000
		self.addCard("/run/beamOn",[no])

#-------------------------------------------------------------------------------
# Execute only once and register the conversion program
#-------------------------------------------------------------------------------
register(Fluka.MODE, Moira.MODE, Fluka2Moira)
