# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   07-Dec-2007

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import math

import rexx
import bmath
from element import Element

import log
from deck  import Card
from fluka import Fluka, FlukaCard
from mcnp  import Mcnp
from .convert import Convert, register

#===============================================================================
# Conversion class from Mcnp to Fluka
#===============================================================================
class Mcnp2Fluka(Convert):
	PAT_CELL     = re.compile(r"(.*?)#(\d+)(.*)", re.DOTALL | re.IGNORECASE)
	PAT_RMCELL   = re.compile(r"#\d+",            re.DOTALL | re.IGNORECASE)
	PAT_SURFTRCL = re.compile(r"^(S\d+)(C\d+)$")
	# Accept also surfaces from macro bodies
	PAT_SURFACE  = re.compile(r"^[+-]?\d*\.?\d$")

	#-----------------------------------------------------------------------
	# FIXME some automatic unit conversion system!
	#-----------------------------------------------------------------------
	def convert(self, **kwargs):
		super().convert(**kwargs)
		self.output = Fluka()
		rotdefi = self.input.cardProperties()

		try:
			card = self.input.cards["TITLE"][0]
			self.addCard("TITLE").setTitle(card.title().rstrip())
		except KeyError:
			pass
		self.addCard("DEFAULTS",["PRECISIO"])
		self.addCard("BEAM")
		self.addCard("BEAMPOS")

		self.addCard("GEOBEGIN",["COMBNAME"])
		self.convertSurfaces()
		self.addCard("END")
		self.convertCells()
		self.addCard("END")
		self.addCard("GEOEND")
		self.convertTransformations(rotdefi)
		self.convertMaterials()
		self.assignMaterial()
		#self.convertImportance()
		self.addCard("START")
		self.addCard("STOP")
		return self.output

	#-----------------------------------------------------------------------
	def addCard(self, tag, what=None, comment=""):
		# special treatment for PLA convert to XYP, XZP, YZP if needed
		if tag=="PLA":
			if what[1]==0.0 and what[2]==0.0 and what[3]>0.0:
				tag = "XYP"
				what = [what[0], what[6]]
			elif what[1]==0.0 and what[2]>0.0 and what[3]==0.0:
				tag = "XZP"
				what = [what[0], what[5]]
			elif what[1]>0.0 and what[2]==0.0 and what[3]==0.0:
				tag = "YZP"
				what = [what[0], what[4]]
		card = FlukaCard(tag, what, comment)
		self.add(card)
		return card

	#-----------------------------------------------------------------------
	# Convert MCNP surfaces to FLUKA bodies
	#-----------------------------------------------------------------------
	def convertSurfaces(self):
		prevtr = None
		for card in self.input.cardsSorted(Card.BODY,-1).values():
			if card is None: continue
			#Global.dump(">S>",card, level=2)
			tr   = card["tr"]
			name = f"S{card.name()}"
			arg  = [card.numWhat(x) for x in range(1,card.nwhats())]

			if prevtr is not None and prevtr != tr:
				self.addCard("$end_transform")
				prevtr = None
			if prevtr is None and tr is not None:
				self.addCard("$start_transform",[f"rot{tr}"])
				prevtr = tr

			if card.tag=="P":
				# Plane
				tag = "PLA"
				if len(arg)==4:
					a,b,c,d = arg
					t = d/(a*a + b*b + c*c)
					what = [name, a, b, c, a*t, b*t, c*t]
				elif len(arg)==9:
					A = bmath.Vector(arg[0], arg[1], arg[2])
					B = bmath.Vector(arg[3], arg[4], arg[5])
					C = bmath.Vector(arg[6], arg[7], arg[8])
					n = (B-A) ^ (C-A)
					n.normalize()
					d = n*A
					# From MCNP4B manual 3-18
					if abs(d)<1e-10:				# D=0
						if abs(n[2])<1e-10 and abs(n[1])<1e-10: # D=C=B=0
							n = bmath.Vector.X
						elif abs(n[2])<1e-10:			# D=C=0
							n = bmath.Vector.Y
						else:					# D=0
							n = bmath.Vector.Z
					elif d < 0.0:
						n = -n
					what = [name, n[0], n[1], n[2], A[0], A[1], A[2]]
				else:
					log.error(f"Plane {card.name()} has invalid number of arguments")
					what = [name] + arg
				self.addCard(tag, what, card.comment())
			elif card.tag=="PX":
				# Plane normal to X-axis
				tag  = "YZP"
				what = [name, arg[0]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="PY":
				# Plane normal to Y-axis
				tag  = "XZP"
				what = [name, arg[0]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="PZ":
				# Plane normal to Z-axis
				tag  = "XYP"
				what = [name, arg[0]]
				self.addCard(tag, what, card.comment())

			#elif card.tag=="X":
			#	pass	# Not implemented yet see notes
			#	self.addCard(tag, what, card.comment())

			#elif surface=="Y":
			#	pass	# Not implemented yet see notes
			#	self.addCard(tag, what, card.comment())

			#elif surface=="Z":
			#	pass	# Not implemented yet see notes
			#	self.addCard(tag, what, card.comment())

			elif card.tag=="SO":
				# Sphere at Origin
				tag  = "SPH"
				what = [name, 0., 0., 0., arg[0]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="S":
				# Sphere general
				tag  = "SPH"
				what = [name, arg[0], arg[1], arg[2], arg[3]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="SX":
				# Sphere centered on X-axis
				tag  = "SPH"
				what = [name, arg[0], 0., 0., arg[1]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="SY":
				# Sphere centered on Y-axis
				tag  = "SPH"
				what = [name, 0., arg[0], 0., arg[1]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="SZ":
				# Sphere centered on Z-axis
				tag  = "SPH"
				what = [name, 0., 0., arg[0], arg[1]]
				self.addCard(tag, what, card.comment())

			elif card.tag=="C/X":
				# Cylinder parallel to X-axis
				tag  = "XCC"
				what = [name, arg[0], arg[1], arg[2]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="C/Y":
				# Cylinder parallel to Y-axis
				tag  = "YCC"
				what = [name, arg[1], arg[0], arg[2]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="C/Z":
				# Cylinder parallel to Z-axis
				tag  = "ZCC"
				what = [name, arg[0], arg[1], arg[2]]
				self.addCard(tag, what, card.comment())

			elif card.tag=="CX":
				# Cylinder on X-axis
				tag  = "XCC"
				what = [name, 0., 0., arg[0]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="CY":
				# Cylinder on Y-axis
				tag  = "YCC"
				what = [name, 0., 0., arg[0]]
				self.addCard(tag, what, card.comment())
			elif card.tag=="CZ":
				# Cylinder on Z-axis
				tag  = "ZCC"
				what = [name, 0., 0., arg[0]]
				self.addCard(tag, what, card.comment())

			elif card.tag=="K/X":
				# Cone parallel to X-axis
				# FIXME the last variable, single is not taken into account
				tag = "QUA"
				x,y,z,t2  = arg[:4]
				try:
					#single = arg[4]
					log.warning("Single side cone is not implemented yet!")
					log.warningCont(card)
				except IndexError:
					pass
					#single = None
				what = [name,	-t2, 1.0, 1.0,			# Cxx Cyy Czz
						0.0, 0.0, 0.0,			# Cxy Cxz Cyz
						2.0*t2*x, -2.0*y, -2.0*z,	# Cx  Cy  Cz
						y*y + z*z - t2*x*x]
				self.addCard(tag, what, card.comment())

			elif card.tag=="K/Y":
				# Cone parallel to Y-axis
				# FIXME the last variable, single is not taken into account
				tag = "QUA"
				x,y,z,t2  = arg[:4]
				try:
					#single = arg[4]
					log.warning("single side cone is not implemented yet!")
					log.warningCont(card)
				except IndexError:
					pass
					#single = None
				what = [name,	1.0, -t2, 1.0,			# Cxx Cyy Czz
						0.0, 0.0, 0.0,			# Cxy Cxz Cyz
						-2.0*x, 2.0*t2*y, -2.0*z,	# Cx  Cy  Cz
						x*x + z*z - t2*y*y]
				self.addCard(tag, what, card.comment())

			elif card.tag=="K/Z":
				# Cone parallel to Z-axis
				# FIXME the last variable, single is not taken into account
				tag = "QUA"
				x,y,z,t2  = arg[:4]
				try:
					#single = arg[4]
					log.warning("single side cone is not implemented yet!")
					log.warningCont(card)
				except IndexError:
					pass
					#single = None
				what = [name,	1.0, 1.0, -t2,			# Cxx Cyy Czz
						0.0, 0.0, 0.0,			# Cxy Cxz Cyz
						-2.0*x, -2.0*y, 2.0*t2*z,	# Cx  Cy  Cz
						x*x + y*y - t2*z*z]
				self.addCard(tag, what, card.comment())

			elif card.tag=="KX":
				# Cone on X-axis
				tag  = "QUA"
				x  = arg[0]
				t2 = arg[1]
				try:
					#single = arg[2]
					log.warning("single side cone is not implemented yet!")
					log.warningCont(card)
				except IndexError:
					pass
					#single = None
				what = [name,	-t2, 1.0, 1.0,		# Cxx Cyy Czz
						0.0, 0.0, 0.0,		# Cxy Cxz Cyz
						2.0*t2*x, 0.0, 0.0,	# Cx  Cy  Cz
						-t2*x*x]
				self.addCard(tag, what, card.comment())

			elif card.tag=="KY":
				# Cone on Y-axis
				tag = "QUA"
				y  = arg[0]
				t2 = arg[1]
				try:
					#single = arg[2]
					log.warning("single side cone is not implemented yet!")
					log.warningCont(card)
				except IndexError:
					#single = None
					pass
				what = [name,	1.0, -t2, 1.0,		# Cxx Cyy Czz
						0.0, 0.0, 0.0,		# Cxy Cxz Cyz
						0.0, 2.0*t2*y, 0.0,	# Cx  Cy  Cz
						-t2*y*y]
				self.addCard(tag, what, card.comment())

			elif card.tag=="KZ":
				# Cone on Z-axis
				tag = "QUA"
				z  = arg[0]
				t2 = arg[1]
				try:
					#single = arg[2]
					log.warning("single side cone is not implemented yet!")
					log.warningCont(card)
				except IndexError:
					#single = None
					pass
				what = [name,	1.0, 1.0, -t2,		# Cxx Cyy Czz
						0.0, 0.0, 0.0,		# Cxy Cxz Cyz
						0.0, 0.0, 2.0*t2*z,	# Cx  Cy  Cz
						-t2*z*z]
				self.addCard(tag, what, card.comment())

			elif card.tag=="SQ":
				# Ellipsoid, Hyperboloid, Paraboloid axes parallel to X,Y,Z-axis
				tag  = "QUA"
				a,b,c, d,e,f, g, x,y,z = arg
				what = [name, a, b, c, 0, 0, 0,
						2*(d-a*x), 2*(e-b*y), 2*(f-c*z),
						g + a*x**2 + b*y**2 + c*z**2 - 2*(d*x + e*y + f*z)]
				self.addCard(tag, what, card.comment())

			elif card.tag=="GQ":
				# Cylinder, Cone, Ellipsoid, Hyperboloid, Paraboloid generic
				tag  = "QUA"
				what = [name, arg[0], arg[1], arg[2],
					      arg[3], arg[5], arg[4],	# Warning on the order xy yz zx
					      arg[6], arg[7], arg[8],
					      arg[9]]
				self.addCard(tag, what, card.comment())

			elif card.tag in ("TX","TY","TZ"):
				tag  = f"TR{card.tag[-1]}"
				what = [name] + arg
				self.addCard(tag, what, card.comment())

			elif card.tag in ("BOX", "REC", "WED"):
				tag  = card.tag
				what = [name, arg[0], arg[1], arg[2],
					      arg[3], arg[4], arg[5],
					      arg[6], arg[7], arg[8],
					      arg[9], arg[10], arg[11]]
				self.addCard(tag, what, card.comment())

				# Facet description BOX
				# 1 Plane normal to end       of a1x a1y a1z
				# 2 Plane normal to beginning of a1x a1y a1z
				# 3 Plane normal to end       of a2x a2y a2z
				# 4 Plane normal to beginning of a2x a2y a2z
				# 5 Plane normal to end       of a3x a3y a3z
				# 6 Plane normal to beginning of a3x a3y a3z

				# Facet description WED
				# 1 Slant plane including top and bottom hypotenuses
				# 2 Plane including vectors v2 and v3
				# 3 Plane including vectors v1 and v3
				# 4 Plane including vectors v1 and v2 at end of v3 (top triangle)
				# 5 Plane including vectors v1 and v2 at beginning of v3
				#   (bottom triangle, including vertex point)

				# Facet description REC
				# 1 Elliptical cylinder
				# 2 Plane normal to end of hx hy hz
				# 3 Plane normal to beginning of hx hy hz

			elif card.tag=="RPP":
				tag  = card.tag
				what = [name, arg[0], arg[1],
					      arg[2], arg[3],
					      arg[4], arg[5]]
				self.addCard(tag, what, card.comment())

				# Facet description
				# 1 Plane xmax
				self.addCard("YZP", [f"{name}.1", arg[1]])
				# 2 Plane xmin
				#self.addCard("YZP", [f"{name}.2", arg[0]])
				self.addCard("PLA", [f"{name}.2", -1,0,0, arg[0],0,0])
				# 3 Plane ymax
				self.addCard("XZP", [f"{name}.3", arg[3]])
				# 4 Plane ymin
				#self.addCard("XZP", [f"{name}.4", arg[2]])
				self.addCard("PLA", [f"{name}.4", 0,-1,0, 0,arg[2],0])
				# 5 Plane zmax
				self.addCard("XYP", [f"{name}.5", arg[5]])
				# 6 Plane zmin
				#self.addCard("XYP", [f"{name}.6", arg[4]])
				self.addCard("PLA", [f"{name}.6", 0,0,-1, 0,0,arg[4]])

			#elif card.tag in ("RHP", "HEX"):
				# Facet description
				# 1 Plane normal to end of r1 r2 r3
				# 2 Plane opposite facet 1
				# 3 Plane normal to end of s1 s2 s3
				# 4 Plane opposite facet 3
				# 5 Plane normal to end of t1 t2 t3
				# 6 Plane opposite facet 5
				# 7 Plane normal to end of h1 h2 h3
				# 8 Plane normal to beginning of h1 h2 h3

			elif card.tag=="SPH":
				tag  = card.tag
				what = [name, arg[0], arg[1], arg[2], arg[3]]
				self.addCard(tag, what, card.comment())

			elif card.tag=="RCC":
				tag  = card.tag
				what = [name, arg[0], arg[1], arg[2],
					      arg[3], arg[4], arg[5],
					      arg[6]]
				self.addCard(tag, what, card.comment())
				# Facet description
				# 1 Cylindrical surface of radius r
				# 2 Plane normal to end of       hx hy hz
				# 3 Plane normal to beginning of hx hy hz
				if arg[4]==0.0 and arg[5]==0.0:
					self.addCard("XCC", [f"{name}.1", arg[1], arg[2], arg[6]])
					self.addCard("YZP", [f"{name}.2", arg[0] + arg[3]])
					#self.addCard("YZP", [f"{name}.3", arg[0]])
					self.addCard("PLA", [f"{name}.3", -1,0,0, arg[0],0,0])
				elif arg[3]==0.0 and arg[5]==0.0:
					self.addCard("YCC", [f"{name}.1", arg[2], arg[0], arg[6]])
					self.addCard("XZP", [f"{name}.2", arg[1] + arg[4]])
					#self.addCard("XZP", [f"{name}.3", arg[1]])
					self.addCard("PLA", [f"{name}.3", 0,-1,0, 0,arg[1],0])
				elif arg[3]==0.0 and arg[4]==0.0:
					self.addCard("ZCC", [f"{name}.1", arg[0], arg[1], arg[6]])
					self.addCard("XYP", [f"{name}.2", arg[2] + arg[5]])
					#self.addCard("XYP", [f"{name}.3", arg[2]])
					self.addCard("PLA", [f"{name}.3", 0,0,-1, 0,0,arg[2]])
				else:
					pass

			elif card.tag=="TRC":
				tag  = card.tag
				what = [name] + arg
				self.addCard(tag, what, card.comment())
				# Facet description
				# 1 Conical surface
				# 2 Plane normal to end of hx hy hz
				# 3 Plane normal to beginning of hx hy hz

			elif card.tag=="ARB":
				tag  = card.tag
				what = [name] + arg
				self.addCard(tag, what, card.comment())
				self._arbFacets(name, what)
			else:
				raise Exception(card)

		if prevtr is not None:
			self.addCard("$end_transform")
		return True

	#-----------------------------------------------------------------------
	# Facet description
	# 1 Plane defined by corners n1
	# 2 Plane defined by corners n2
	# 3 Plane defined by corners n3
	# 4 Plane defined by corners n4
	# 5 Plane defined by corners n5
	# 6 Plane defined by corners n6
	#-----------------------------------------------------------------------
	def _arbFacets(self, name, what):
		vertices = []
		for i in range(1,3*8+1,3):
			vertices.append(bmath.Vector(what[i], what[i+1], what[i+2]))

		#---------------------------------------------------------------
		# return vertices of a face
		#---------------------------------------------------------------
		def faceVertices(f):
			V    = []
			pts  = set(range(8))
			face = int(what[f + 3*8+1])
			while face>0:
				v = face%10-1
				if 0<=v<8:
					V.append(vertices[v])
					try:
						pts.remove(v)
					except KeyError:
						pass
				face //= 10
			return V, pts
		#---------------------------------------------------------------
		# create faces
		#---------------------------------------------------------------
		for i in range(6):
			V,pts = faceVertices(i)
			if all([x is None for x in V]): continue
			N = (V[1] - V[0]) ^ (V[2] - V[0])
			N.normalize()
			d = N*V[0]
			# the normal sign is defined by the location of the other points of the ARB
			for j in pts:
				sense = -N*vertices[j] + d
				if sense > 1e-5:
					break
				elif sense < -1e-5:
					N = -N
					break
			self.addCard("PLA", [f"{name}.{i+1}", N.x, N.y, N.z, V[0].x, V[0].y, V[0].z])

	#-----------------------------------------------------------------------
	# Convert Mcnp cells expressions to FLUKA regions
	#-----------------------------------------------------------------------
	def convertCells(self):
		#trsurf  = {}		# transformed surfaces for "TRCL" cells
		self._density = {}
		for card in self.input.cardsSorted(Card.CELL,-1).values():
			#Global.dump(">C>", card, level=2)
			tokens = card.expr().replace("("," ( ")
			tokens = re.sub(r"#\s+", "#", tokens)
			tokens = re.sub(r"\s*=\s*", "= ", tokens)
			tokens = tokens.replace("#", " #")
			tokens = tokens.replace("# +(", " -( ")
			tokens = tokens.replace("#(", " -( ")
			tokens = tokens.replace(")", " ) ")
			tokens = tokens.replace(":", " : ")
			tokens = tokens.split()
			#log.say(">T>", tokens, level=2)

			region = self.addCard("REGION", [f"C{card.name()}"], card.comment())
			material = f"M{card.what(1)}"
			density  = str(card.numWhat(2))
			expr  = []
			#if material == "LIKE":
			#	like = tokens.pop(0)
			#	if tokens.pop(0) != "BUT":
			#		log.error(f"Cell {card.name()} 'BUT' was expected after LIKE")
			#		continue
			#	region["@ref"] = ref = cells[like]
			#	material = ref["@mat"]
			#	density  = ref["@density"]

			#	expr = ref.expr().split()	# copy expression
			#elif material != "0":
			#	density = tokens.pop(0)
			#else:
			#	density = 0

			densities = self._density.setdefault(material,[])
			if density not in densities:
				densities.append(density)
				if len(densities)>1:
					material = f"{material}.{len(densities)}"
			region["@mat"] = material
			region["@density"] = density

			# Parse normal expression
			if not expr:
				depth = 0
				while tokens:
					token = tokens.pop(0)
					if token==":":
						if depth>0:
							expr.append("|")
						else:
							expr.append("\n|")
					elif token == "(":
						expr.append("+(")
						depth += 1
					elif token == "-(":
						expr.append(token)
						depth += 1
					elif token == ")":
						expr.append(token)
						depth -= 1
					elif token[0] == "#":
						expr.append(token)
					elif Mcnp2Fluka.PAT_SURFACE.match(token):
						try:
							num = int(token)
						except:
							num = float(token)
						if num<0:
							op = "+"
							num = -num
						else:
							op = "-"
						expr.append(f"{op}S{num}")
					else:
						tokens.insert(0,token)
						break

			# Parse remaining tokens
			#while tokens:
			#	token = tokens.pop(0)
			#	# -- Importance ---
			#	if token == "IMP":
			#		colon = tokens.pop(0)	# :
			#		equal = tokens.pop(0)	# particle=
			#		try:
			#			value = float(tokens.pop(0))	# value
			#		except ValueError:
			#			value = None
			#		if colon != ":" or equal[-1] != "=" or value is None:
			#			log.error(f"CELL {name} bad importance IMP{colon}{equal}")
			#		else:
			#			region["@IMP:"+equal[:-1]] = value

			#	# -- Translation ---
			#	elif token in ("TRCL=", "*TRCL="):
			#		deg = token[0]=="*"
			#		token = tokens.pop(0)
			#		if token == "(":
			#			transform = []
			#			while True:
			#				token = tokens.pop(0)
			#				if token==")": break
			#				try:
			#					transform.append(float(token))
			#				except ValueError:
			#					log.error(f"CELL {name} bad value '{token}' for TRCL")
			#			rot = f"trcl{idx}"
			#			self._trcl[rot] = self.__parseTransformation(
			#						transform, deg)
			#		else:
			#			# Use existing transformation
			#			rot = f"rot{token}"

			#		for i,t in enumerate(expr):
			#			# Surfaces should be as: [+\-]S\d+ (C\d+)?
			#			if len(t)>2 and t[0] in "+-" and t[1] == "S":
			#				m = PAT_SURFTRCL.match(t[1:])
			#				if m:	# append cell name
			#					surface = m.group(1) + name
			#				else:
			#					surface = t[1:] + name
			#				trsurf[surface] = rot	# assign rotation
			#				expr[i] = t[0] + surface

			#	# --- Material override ---
			#	elif token == "MAT=":
			#		region["@mat"] = tokens.pop(0)

			#	# FIXME: tokens to be implemented
			#	# VOL, PWT, EXT, FCL, WWN, DXC, NONU, PD, TMP, U, LAT, FILL
			#	else:
			#		value = tokens.pop(0)
			#		log.error(f"CELL {name} Unsupported token {token}{value} found")

			# Set the expression
			region.setExpr(rexx.join(expr,70))

		# --------------------------------
		# Correct cell references # as -()
		# --------------------------------
		regions = self.output.cardsSorted(Card.REGION)
		for card in regions.values():
			expr = card.expr()
			if "#" not in expr: continue
			while True:
				m = Mcnp2Fluka.PAT_CELL.match(expr)
				if m is None: break
				try:
					ref = regions[f"C{m.group(2)}"]
					# Expand only first level by removing
					refexpr = re.sub(Mcnp2Fluka.PAT_RMCELL, "", ref.expr()).strip()
					expr = f"{m.group(1)}\n-({refexpr}){m.group(3)}"
				except KeyError:
					log.error(f"Cell {card.name()} refers to "
						f"non existing cell #{m.group(2)}")
					expr = f"{m.group(1)}\n{m.group(3)}"
			card.setText(expr)

		# Create clones of surfaces if needed
		#for name, rot in trsurf.items():
		#	m = PAT_SURFTRCL.match(name)
		#	surf = m.group(1)
		#	#cell = m.group(2)
		#	# Find body and clone it
		#	for card in self.input.cardlist:
		#		if len(card.tag) != 3 or card.name() != surf: continue
		#		pos = card.pos()
		#		clone = card.clone()
		#		clone.setName(name)
		#		self.input.add(FlukaCard("$start_transform",[rot]),pos+1)
		#		self.input.add(clone,pos+2)
		#		self.input.add(FlukaCard("$end_transform"),pos+3)
		#		self.input.renumber(pos)
		#		break

	#-----------------------------------------------------------------------
	def convertTransformations(self, rotdefi):
		for card in self.input.cardsSorted(Card.ROTATION,-1).values():
			comment = card.comment()
			rid = str(card.name())
			M = rotdefi[rid]
			name = f"rot{rid}"

			Rx,Ry,Rz = map(math.degrees, M.getEulerRotation())
			x = M[0][3]
			y = M[1][3]
			z = M[2][3]

			m = card.intWhat(13) >= 0
			EPS = 1e-15
			if not m and (abs(x)>EPS or abs(y)>EPS or abs(z)>EPS):
				Minv = M.inverse()
				x = Minv[0][3]
				y = Minv[1][3]
				z = Minv[2][3]
				self.addCard("ROT-DEFI",
						[name, "", "", "", -x,-y,-z], comment)
				comment = ""

			if abs(Rz) > EPS:
				self.addCard("ROT-DEFI",
						[name, "", "", bmath.format(Rz,10)], comment)
				comment=""

			if abs(Ry) > EPS:
				self.addCard("ROT-DEFI",
						[name, 200, "", bmath.format(Ry,10)], comment)
				comment=""

			if abs(Rx) > EPS:
				self.addCard("ROT-DEFI",
						[name, 100, "", bmath.format(Rx,10)], comment)
				comment=""

			if m and (abs(x)>EPS or abs(y)>EPS or abs(z)>EPS):
				# FIXME not correct
				#x,y,z = M.multNoTranslation([-x,-y,-z])
				self.addCard("ROT-DEFI",
						[name, "", "", "", x,y,z], comment)

	#-----------------------------------------------------------------------
	# FIXME densities I need a dictionary of material/density from the regions
	#-----------------------------------------------------------------------
	def convertMaterials(self):
		def za(x): return divmod(int(str(x).split(".")[0]), 1000)
		elems = {}
		toadd = []
		for card in self.input.cardsSorted(Card.MATERIALS).values():
			name = f"M{card.name()}"

			try:
				densities = self._density[name]
			except KeyError:
				log.warning(f"Material '{name}' is not used."
						"Setting density to 1.0")
				densities = [-1.0]

			for im, density in enumerate(map(float,densities)):
				if density>0.0:
					log.warning(f"Material '{name}' is defined with"
						"atom density, please correct it")
				else:
					density = -density

				matname = name if im==0 else f"{name}.{im+1}"

				if card.nwhats() == 3:
					# isotope
					z,a = za(card.what(1))
					elems[z*1000+a] = name
					if a==0: a=""
					self.addCard("MATERIAL",
							[matname, z, "", density, "", "", a],
							card.comment())
				else:
					# compound
					self.addCard("MATERIAL",
							[matname, "", "", density, "", "", ""],
							card.comment())
					what = [matname]
					for i in range(1,card.nwhats(),2):
						try:
							z,a = za(card.what(i))
						except:
							break
						frac = card.numWhat(i+1)

						try:
							n = elems[z*1000+a]
						except KeyError:
							# Add a new isotope
							element = Element.byZ(z)
							if a==0:
								n = element.inputName(Fluka.MODE)
							else:
								n = f"{element.symbol.upper()}-{a}"
							#n = f"M{z*1000+a}"
							elems[z*1000+a] = n
							if a==0: a=""
							toadd.append([n, z, "", "1.0", "", "", a])

						what.append(frac)
						what.append(n)
					# make what multiples of 6+1
					while (len(what)-1)%6 != 0: what.append("")
					self.addCard("COMPOUND", what)

		for what in toadd:
			self.addCard("MATERIAL", what)

	#-----------------------------------------------------------------------
	# Assign materials to regions
	#-----------------------------------------------------------------------
	def assignMaterial(self):
		for card in self.output.cardsSorted(Card.REGION,-1).values():
			material = str(card["@mat"])
			if material == "M0":
				material = "VACUUM"
			elif material == "M-1":
				material = "BLCKHOLE"
			self.addCard("ASSIGNMA",["", material, card.name()])

	#-----------------------------------------------------------------------
	# Add biasing
	#-----------------------------------------------------------------------
	#def _writeBiasing(self, fout):
	#	fout.write("C ---------- Biasing ----------\n")
	#	line   = []
	#	repeat = 0
	#	prev   = -1.0
	#	for imp in self.importance:
	#		if imp == prev:
	#			repeat += 1
	#		else:
	#			if prev<0:
	#				pass
	#			else:
	#				line.append(f"{prev:.15g}")
	#				if repeat == 1:
	#					line.append(line[-1])
	#				elif repeat > 1:
	#					line.append(f"{repeat}r")
	#			prev = imp
	#			repeat = 0
	#	line.append(f"{prev:.15g}")
	#	if repeat == 1:
	#		line.append(line[-1])
	#	else:
	#		line.append(f"{repeat}r")
	#	self._writeLine(fout, f"IMP:n {' '.join(line)}")

#-------------------------------------------------------------------------------
# Execute only once and register the conversion program
#-------------------------------------------------------------------------------
register(Mcnp.MODE, Fluka.MODE, Mcnp2Fluka)
