# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-May-2004

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

__all__ = [	"Card",
		"CSGCard",
		"CardInfo",
		"DeckConfig",
		"Input",
		"Layout" ]

import re
import os
import sys
import string
import configparser
from itertools import count, chain

import log
from .card   import Card, CardInfo, VarInfo, CSGCard
from .input  import Input
from .layout import Layout
from .config import DeckConfig

_RANGE_REPLACE_PAT = re.compile(r"^range\.([0-9]+)\.([0-9]+)$")
_GUIDE_NAME_PAT    = re.compile(r"^\"(.*?)\"\s*(.*)$", re.DOTALL)

DATABASE = "db/input.ini"
__loaded = False

#-------------------------------------------------------------------------------
def dbpath(filename):
	return os.path.join(os.path.dirname(os.path.dirname(__file__)), filename)

#-------------------------------------------------------------------------------
# Go through all cards and populate the appropriate CardInfo class
#-------------------------------------------------------------------------------
def readCardini(cardini, mode, cardinfo_class):
	#-----------------------------------------------------------------------
	# Convert to integer when possible
	#-----------------------------------------------------------------------
	def toInt(x):
		try:
			return int(x)
		except ValueError:
			return x

	clone_found = False		# no clone item is found
	for ipass in range(2):		# pass=0: read all normal cards
					# pass=1: read all clones
		for name in cardini.sections():
			# ignore special sections starting with '@'
			if name[0]=="@": continue
			try:
				clone = cardini.get(name, "clone")
				if ipass==0:
					clone_found = True
					continue	# ignore clones on first pass
				# replace () with []
				if "(" in clone:
					clone = clone.translate(CardInfo._paren_trans)
				# Initialize values with clones data
				source = CardInfo.get(mode, clone)
				assert source is not None
				if source is not CardInfo.NONE:
					cinfo = cardinfo_class(mode,name)
					cinfo.copy(source)
				else:
					log.error(f"Reading ini for mode='{mode}' "
						f"card='{name}' clone='{clone}' not found")
					sys.exit(0)
			except configparser.NoOptionError:
				if ipass==1: continue
				# Create card on first pass to be populated
				cinfo = cardinfo_class(mode,name)
				clone = None

			try:    cinfo.setMeaning(cardini.get(name, "meaning"))
			except: pass
			try:    cinfo.setAlias(cardini.get(name, "alias"))
			except: pass
			try:    cinfo.setIcon(cardini.get(name, "icon"))
			except: pass
			try:    cinfo.setOTag(cardini.get(name, "otag"))
			except: pass
			try:    cinfo.setCompactMode(cardini.get(name, "compactMode"))
			except: pass
			try:    cinfo.setIsExtension(cardini.get(name, "isExtension"))
			except: pass
			try:    cinfo.setOrder(cardini.get(name, "order"))
			except: pass
			#try:    cinfo.setRepeat(cardini.get(name, "repeat"))
			#except: pass

			# FIXME Only for FLUKA?
			if mode=="fluka":
				try:
					dc = cardini.get(name, "disable")
					cinfo.setDisableComment(dc!="ifendif")
				except configparser.NoOptionError:
					pass

			# Parse "whats"
			# --- Groups ---
			try:
				group = []
				items = cardini.get(name, "group").split()
				for item in items:
					# Capitalize each group
					item = string.capwords(item,"/")
					group.append(item)
				cinfo.setGroup(group)
			except configparser.NoOptionError:
				pass

			# --- Range ---
			range_ = []
			#range_found = False
			for i in count():
				try:
					# parse range assume groups in () and [] and break in spaces
					rng = []
					item = ""
					nest = 0	# parenthesis and bracket [] count
					backspace = False
					for ch in cardini.get(name, f"range.{i}"):
						# maybe convert to function?
						if ch.isspace() and nest==0 and not backspace:
							if item:
								rng.append(item)
								item = ""
						elif ch in "([":	# not very fail safe but faster
							nest += 1
							item += ch
						elif ch in ")]":
							nest -= 1
							item += ch
						elif ch == "\\":
							backspace = True
						elif ch == "\n":
							pass
						else:
							item += ch
							backspace = False
					assert nest==0
					if item: rng.append(item)
					range_.append(rng)
				except configparser.NoOptionError:
					break
			if range_:
				cinfo.setRange(range_)
				#range_found = True

			# --- Defaults ---
			default = []
			default_found = False
			for i in range(cinfo.getNRanges()):
				try:
					default.append(cardini.get(name, f"defaults.{i}").split())
					default_found = True
				except configparser.NoOptionError:
					default.append(None)
			if default_found: cinfo.setDefault(default)

			# ---other info ---
			guide   = []
			label   = []
			layout  = []
			dynamic = []
			gfound  = False
			lfound  = False
			yfound  = False
			dfound  = False
			for i in range(cinfo.getNRanges()):
				gde = []
				lbl = []
				# Check for text "-1" index and append it at the end
				for j in chain(range(cinfo.nwhats[i]),[-1]):
					# --- guidance info ---
					l = g = None
					try:
						g = cardini.get(name, f"guide.{i}.{j}")
						if g and g[0]=='"':
							pat = _GUIDE_NAME_PAT.match(g)
							if pat:
								l = pat.group(1)
								g = pat.group(2)
								lfound = True
						gfound = True
					except configparser.NoOptionError:
						pass

					# --- label info ---
					try:
						l = cardini.get(name, f"label.{i}.{j}")
						lfound = True
					except configparser.NoOptionError:
						pass

					# append to the list unless it is a clone
					# which we need to replace
					if clone is None:
						gde.append(g)
						lbl.append(l)
					else:
						# For clones replace item directly
						if gfound:
							cinfo.guide[i][j] = g
							gfound = False
						if lfound:
							cinfo.label[i][j] = l
							lfound = False
				if clone is None:
					guide.append(gde)
					label.append(lbl)

				try:
					layout.append([toInt(x)
							for x in cardini.get(name,
								f"layout.{i}").split()])
					yfound = True
				except configparser.NoOptionError:
					layout.append(None)

				try:
					dynamic.append([toInt(x)
							for x in cardini.get(name,
								f"dynamic.{i}").split()])
					dfound = True
				except configparser.NoOptionError:
					dynamic.append(None)

			if gfound: cinfo.setGuide(guide)
			if lfound: cinfo.setLabel(label)
			if yfound: cinfo.setLayoutTemplate(layout)
			if dfound: cinfo.setLayoutDynamic(dynamic)

			# --- assert info ---
			assert_ = []
			afound = False
			for i in range(cinfo.getNRanges()):
				a = []
				for j in count():
					try:
						a.append(cardini.get(name, f"assert.{i}.{j}"))
						afound = True
					except configparser.NoOptionError:
						break
				assert_.append(a)
			if afound: cinfo.setAssert(assert_)

			# --- Correct cloned types ---
			# Parse variables and insert items
			# Find variables
			properties = []
			for vname,value in cardini.items(name):
				if vname.startswith("property."):
					properties.append((vname[9:], value))
				elif vname[0]=="+":
					word = vname[1:].split(".")
					case  = int(word[1])
					index = int(word[2])
					if word[0]=="range":
						cinfo.range[case].insert(index, value)
						#range_found = True
					elif word[0] == "label":
						cinfo.label[case].insert(index, value)
					elif word[0] == "guide":
						cinfo.guide[case].insert(index, value)
					elif word[0] == "layout":
						# extend list at specific position
						cinfo.layoutTemplate[case][index:index] = \
							[toInt(x) for x in value.split()]
					else:
						raise Exception(f"Unknown insert value {vname!r} "
								f"in {name!r}")
				else:
					pat = _RANGE_REPLACE_PAT.match(vname)
					if pat:
						case  = int(pat.group(1))
						index = int(pat.group(2))
						try:
							cinfo.range[case][index] = value
						except IndexError as ex:
							log.error(f"readCardIni tag={mode}:{name} {ex}")
						#range_found = True
			#if range_found: cinfo.compileRange()
			cinfo.compileRange()

			# For all properties get parameters and populate cinfo
			for vname, meaning in properties:
				array = False		# variable is a list/array of
				try:
					varlist = cardini.get(name, f"var.{vname}").strip()
					if varlist[0] == "[":
						if varlist[-1] != ']':
							log.error(f"readCardIni tag={mode}:{name} "
								f"Property var.{vname} missing ]")
						array = True
						varlist = varlist[1:-1]
					varlist = varlist.split()
				except configparser.NoOptionError:
					varlist = [vname]
				try:    otag   = cardini.get(name, f"otag.{vname}")
				except: otag   = None
				try:
					layout = [toInt(x) for x in cardini.get(name,
								f"layout.{vname}").split()]
				except: layout = None
				variables = []
				for var in varlist:
					if len(varlist)==1:
						vn = vname
					else:
						vn = f"{vname}.{var}"
					try:    type_   = cardini.get(name, f"type.{vn}")
					except: type_   = ""
					try:    default = cardini.get(name, f"default.{vn}")
					except: default = ""
					try:    range_  = cardini.get(name, f"range.{vn}")
					except: range_  = ""
					try:
						candidates = cardini.get(name,
								f"candidates.{vn}").split()
						candidates.insert(0,"")
					except: candidates = []
					try:    guidance = cardini.get(name, f"guidance.{vn}")
					except: guidance = ""
					variables.append(VarInfo(var,
								type_,
								default,
								range_,
								candidates,
								guidance))
				try:    assert_ = cardini.get(name, f"assert.{vname}")
				except: assert_ = ""
				cinfo.setProperty(vname,
						meaning,
						variables,
						layout,
						assert_,
						otag,
						array)

			# Add the card info
			CardInfo.add(cinfo)

		if not clone_found: break

	# Finally populate all default/predefined values for typedef
	try:
		for item,value in cardini.items("@defaults"):
			CardInfo.addDefaults(mode, item, value)
	except configparser.NoSectionError:
		pass

#===============================================================================
# Load only once
def init(filename=None):
	global __loaded
	if __loaded: return
	__loaded = True

	Card.register_class(Input.MODE, Card)	# default generic card

	# Special card
	CardInfo.NONE = CardInfo._db[None] = CardInfo("","error")
	CardInfo.NONE.setGroup("Error")
	CardInfo.NONE.setMeaning("Unknown card")

	# Load fluka database
	if filename is None:
		filename = dbpath(DATABASE)

	# read card database and prepare CardInfo classes
	cardini = configparser.RawConfigParser()
	if not cardini.read(filename, encoding="UTF-8"): return False

	# Read the groups
	section = "@group"
	for i in count(1):
		n = f"order.{i}"
		try: name = cardini.get(section, n)
		except: break
		CardInfo.addGroup(name)

	# Read the cards
	readCardini(cardini, Input.MODE, CardInfo)

	# Create flair tags
	CardInfo.FLAIR_TAGS = [ x.tag for x in CardInfo.cards("")
					if "Flair" in x.group ]
	CardInfo.FLAIR_TAGS.sort()
	del cardini
#init()
