# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import copy
import math		# used in validate -> eval()
import time
import string
from operator import attrgetter
from collections import namedtuple, OrderedDict, defaultdict

import log
import bmath
import Global
import Evaluation
from stream import Stream

PropInfo = namedtuple("PropInfo", ["meaning", "variables", "layout", "assert_", "otag", "array"])
VarInfo  = namedtuple("VarInfo",  ["name", "type", "default", "range", "candidates", "guidance"])

#===============================================================================
# CardInfo class
#===============================================================================
class CardInfo:
	_db     = {}		# Cards informations dictionary
	_dbmode = {}		# Cards informations dictionary (list all cards per specific mode)
	_db_macros = {}         # Macro-cards informations dictionary
	_funcs  = {}		# Caching compiled functions

	NONE    = None		# Special "error" or "unknown" info

	_WNPAT     = re.compile(r"\b([wW])\((-?\d+)\)")
	_wPAT      = re.compile(r"\bw\(")
	_WPAT      = re.compile(r"\bW\(")
	_VOWELSPAT = re.compile("[AEYUIOaeyuio]")
	_PATDEF    = re.compile(r"^([si])(\d*):(.*)$")

	typedef    = defaultdict(dict)	# dynamic dictionary containing per mode all cards
					# defining a variable as (s:xxxx,what)
	defaults   = defaultdict(dict)	# predefined names per typedef
	bodies     = defaultdict(list)	# list of bodies for each mode, None for all
	volumes    = defaultdict(list)	# volumes/regions/cells
	materials  = defaultdict(list)
	rotations  = defaultdict(list)
	FLAIR_TAGS = []

	GEOMETRY_GROUP           = "Geometry"
	GEOMETRY_BODY_GROUP      = GEOMETRY_GROUP+"/Body"
	GEOMETRY_SURFACE_GROUP   = GEOMETRY_GROUP+"/Surface"
	GEOMETRY_SOLID_GROUP     = GEOMETRY_GROUP+"/Solid"
	GEOMETRY_VOLUME_GROUP    = GEOMETRY_GROUP+"/Volume"
	GEOMETRY_TRANSFORM_GROUP = GEOMETRY_GROUP+"/Transform"
	PREPROCESSOR_GROUP       = "Preprocessor"

	groups           = []
	groups_set       = set()
	groups_underline = []
	groups_icons  = { GEOMETRY_BODY_GROUP      : "TRC",
			  GEOMETRY_SOLID_GROUP     : "BOX",
			  GEOMETRY_VOLUME_GROUP    : "REGION",
			  GEOMETRY_TRANSFORM_GROUP : "ROT-DEFI",
			  PREPROCESSOR_GROUP       : "hash",
			 }

	# Special card
	EXPANSION  = "expansion"
	TRANSLAT   = "translat"
	TRANSFORM  = "transform"
	IFENDIF    = "#if..#endif"

	_paren_trans = str.maketrans("()","[]")	# translate [] -> () that config parser cannot read

	#-----------------------------------------------------------------------
	def __init__(self, mode, tag):
		if tag[-1]=="/":
			self.directory = True	# directory of cards
			tag = tag[:-1]
		else:
			self.directory = False

		# replace () with []
		if "(" in tag:
			tag = tag.translate(CardInfo._paren_trans)

		self.mode    = mode		# card mode
		self.tag     = tag		# tag should be unique, e.g. in FLUKA 8chars
		self.alias   = tag		# short name for display
		self.otag    = None		# override tag in case of macro-cards
		self.layout  = None		# layout for displaying
		self.layoutTemplate = None	# layout template from ini
		self.layoutDynamic = None	# Properties to display in geometry editor
						# icon name to use other than tag
		self.icon = tag.split("/")[-1] if tag[0]=="/" else tag
		self.meaning    = ""		# card meaning
		self.group      = []		# groups that card belong to
		self.define     = None		# what variable is defining
		self.compact_mode = False	# In compact mode: only properties names are used,
						# not their full paths.
		self.is_extension = False	# An extension card is a complement of another card
		                                # (it could not be used standalone).
		self.range      = []		# accepted range of parameters (whats)
		self.assert_    = []		# assertion code of parameters
		self.label      = []		# label for each parameter (whats)
		self.guide      = []		# guidance text for each parameter
		self.default    = [None]	# default value for each parameter
		self.nwhats     = []		# number of whats per case
		self.maxwhats   = 0		# max number of whats for parsing
		self._rangeCode = []		# list of range code per case
		self.vtype      = []		# variable type of each what per case
		self._assertCode= []		# assertion code

		# variables
		self.property   = OrderedDict()	# dictionary of variables/properties
						# each variable will contain a named tuple

		self.underline = -1		# underline for the menus
		self.order     = 99999		# ordering in file for finding best position
		self.array     = []

	# ----------------------------------------------------------------------
	# Copy info from src
	# ----------------------------------------------------------------------
	def copy(self, src):
		self.group      = src.group[:]
		self.meaning    = src.meaning
		# copy alias if not default
		if src.alias != src.tag:
			self.alias = src.alias
		self.otag       = src.otag
		self.directory  = src.directory
		self.define     = src.define
		self.range      = copy.deepcopy(src.range)
		self.assert_    = copy.deepcopy(src.assert_)
		self.label      = copy.deepcopy(src.label)
		self.guide      = copy.deepcopy(src.guide)
		self.default    = copy.deepcopy(src.default)
		if src.layoutTemplate:
			self.layoutTemplate = [x[:] for x in src.layoutTemplate] \
					if src.layoutTemplate else None
		if src.layoutDynamic:
			self.layoutDynamic = [x[:] for x in src.layoutDynamic]
		self.icon       = src.icon
		self.nwhats     = src.nwhats[:]
		self.maxwhats   = src.maxwhats
		self._rangeCode = [x[:] for x in src._rangeCode]
		self.vtype      = [x[:] for x in src.vtype]
		self._assertCode= [x[:] for x in src._assertCode]
		self.property   = src.property.copy()	# FIXME maybe deepcopy?
		self.order      = src.order
		self.underline  = src.underline
		self.array      = src.array[:]

	# ----------------------------------------------------------------------
	# Setters
	# ----------------------------------------------------------------------
	def setAlias(self, alias):
		self.alias = alias

	# ----------------------------------------------------------------------
	def setLayoutTemplate(self, layout):
		self.layoutTemplate = layout

	# ----------------------------------------------------------------------
	def setLayoutDynamic(self, dynamic):
		self.layoutDynamic = dynamic

	# ----------------------------------------------------------------------
	def setIcon(self, icon):
		self.icon = icon

	# ----------------------------------------------------------------------
	def setOTag(self, otag):
		self.otag = otag

	# ----------------------------------------------------------------------
	def setCompactMode(self, compact_mode):
		self.compact_mode = compact_mode

	# ----------------------------------------------------------------------
	def setIsExtension(self, is_extension):
		self.is_extension = is_extension

	# ----------------------------------------------------------------------
	def setGroup(self, group):
		if isinstance(group,str):
			self.group = [group]
		else:
			self.group = group
		for g in self.group:
			CardInfo.addGroup(g)

	# ----------------------------------------------------------------------
	def setRange(self, range_):
		self.range = range_
		self._nwhats()

	# ----------------------------------------------------------------------
	def setOrder(self, o):
		self.order = int(o)

	# ----------------------------------------------------------------------
	def _nwhats(self):
		self.nwhats   = []
		self.maxwhats = 0
		for r in self.range:
			w = len(r)
			for j in range(len(r)-1,-1,-1):
				if r[j] != "-": break
				w -= 1
			self.nwhats.append(w)
			self.maxwhats = max(self.maxwhats, w)

	# ----------------------------------------------------------------------
	def getNRanges(self):
		return len(self.range)

	# ----------------------------------------------------------------------
	def setAssert(self, assert_):
		self.assert_ = assert_
		# compile assertion code
		for i,assertList in enumerate(self.assert_):
			ass = []
			for a in assertList:
				ass.append(self.compileAssert(a,i))
			self._assertCode.append(ass)

	# ----------------------------------------------------------------------
	def setLabel(self, label):
		self.label = label

	# ----------------------------------------------------------------------
	def setGuide(self, guide):
		self.guide = guide

	# ----------------------------------------------------------------------
	def setMeaning(self, meaning):
		self.meaning = meaning #.replace("\n"," ")

	# ----------------------------------------------------------------------
	def setDefault(self, default):
		self.default = default

	# ----------------------------------------------------------------------
	# Set variable/property with the card
	# ----------------------------------------------------------------------
	def setProperty(self, name, meaning, variables, layout, assert_, otag, array):
		self.property[name] = PropInfo(meaning, variables, layout, assert_, otag, array)
		if array: self.array.append(name)

	# ----------------------------------------------------------------------
	# process range for units, particles, materials and regions
	# ----------------------------------------------------------------------
	def compileRange(self):
		self._nwhats()
		self._rangeCode = []
		self.vtype = []
		for i,r in enumerate(self.range):
			try:
				label = self.label[i]
			except IndexError:
				label = []
			try:
				rcode, vtype = self.compileConditions(r, label)
			except Exception as ex:
				log.error(f"CardInfo tag='{self.tag}' i={i} r={r}")
				raise ex
			self._rangeCode.append(rcode)
			self.vtype.append(vtype)
			# Check for definitions and add them to the dictionary
			for iw,w in enumerate(r):
				m = CardInfo._PATDEF.match(w)
				if m is None: continue
				var = m.group(3)
				dct = CardInfo.typedef[self.mode]
				try:
					lst = dct[var]
				except KeyError:
					dct[var] = lst = set()
				lst.add(self.tag)
				#lst.append((self.tag,iw))
				break

	# ----------------------------------------------------------------------
	# Compile conditions
	# The code which is generated is in the form i.e.:
	#	lambda x: x>
	# and should be executed with:
	#	eval(code)(x)
	# ----------------------------------------------------------------------
	def compileConditions(self, rng, label):
		# the switch condition is given by the f() fields and
		# by the what[0] if is defined

		# Code for each what
		# it will contain   (index, what-type, compiled-code, code)
		# what-type = s - string
		#             i - integer
		#             r - real/float
		#             f - flag to be checked for the case

		vtype = []	# variable types
		rcode = []	# range code

		# FIXME appending code for debugging!
		# Check if all [fri]() conditions are met
		for iw,w in enumerate(rng):
			if w[0] == "-":
				if iw==0:
					s = "lambda x:x==\"\""
					rcode.append((0, 'f', CardInfo._compileFunc(s), s))
					vtype.append('f')
				else:
					vtype.append('-')

			elif w in ("-","s"):
				vtype.append(w)

			elif w[0]=="[" or (len(w)>1 and w[1]=="["):	# list of choices
				if w[0] == "[":
					typ = 's'	# generic string
				else:
					typ = w[0]	# specific type
					w = w[1:]
				assert w[-1]=="]"
				choices  = []
				hasempty = False
				default  = []
				for c in map(str.strip, w[1:-1].split(",")):
					if c and c[0]=="@":
						adddef = True
						c = c[1:]
					else:
						adddef = False
					if ":" in c:
						a,b = map(str.strip,c.split(":"))
						if a=="": hasempty=True
						choices.append(b)
						if adddef: default.append(f"'{b}'")
					else:
						if c=="": hasempty=True
						choices.append(c)
						if adddef: default.append(f"'{c}'")
				if not hasempty:
					choices.insert(0,"")
				if default:
					if len(default)==1:
						s = f"lambda x:str(x)=={default[0]}"
					else:
						s = f"lambda x:str(x) in ({','.join(default)})"
				else:
					if self.mode=="fluka":
						assert max([len(x) for x in choices]) <= 8
						s = "lambda x:str(x)[:8] in ("
					else:
						s = "lambda x:str(x) in ("
					s += ",".join([f"'{x}'" for x in choices])
					s += ")"
				rcode.append((iw, typ, CardInfo._compileFunc(s), s))
				vtype.append(typ)

			elif w[0] in "\"'":
				if ":" in w:
					ws,num = w.split(":")
					s = f"lambda x:x in ({ws},{num})"
				else:
					s = f"lambda x:x=={w}"
				rcode.append((0, 'f', CardInfo._compileFunc(s), s))
				vtype.append('f')

			# type  r() l() L() D() E()
			elif len(w)>4 and w[1] == "(":
				expr = w[2:-1]
				name = "x"
				try:
					if label[iw] and re.search(r".*\b%s\b.*"%(label[iw]), expr):
						name = label[iw]
				except IndexError:
					pass
				expr = expr.replace(',', " or ")
				s = f"lambda {name}:{expr}"
				rcode.append((iw, w[0], CardInfo._compileFunc(s), s))
				vtype.append(w[0])

			elif CardInfo._PATDEF.match(w):
				m = CardInfo._PATDEF.match(w)
				if m.group(1)=="s":
					vtype.append(w)
					if m.group(2):
						s = f"lambda x:len(str(x))<={m.group(2)}"
						rcode.append((iw, w, CardInfo._compileFunc(s), s))
				elif m.group(1)=="i":
					vtype.append(w)
					if m.group(2):
						s = f"lambda x:len(str(int(x)))<={m.group(2)}"
						rcode.append((iw, w, CardInfo._compileFunc(s), s))
					else:
						s = "lambda x:int(x)"
						rcode.append((iw, w, CardInfo._compileFunc(s), s))
				self.define = m.group(3)

			else:
				rcode.append((iw, w, None, None))
				vtype.append(w)

		return rcode, vtype

	#-----------------------------------------------------------------------
	# Compile assert statement
	# First detect all what references w(#) and replace them with:
	#	w(#) -> c.numWhat(#)
	#	W(#) -> c.what(#)
	# identify all what-indexes to report as problems
	# and request in the code that are not zero!
	#
	# finally compile the code
	# example: w(8)>w(7)
	#	lambda c: c.numWhat(7)==0 or c.numWhat(8)==0 or c.numWhat(8)>c.numWhat(7)
	#-----------------------------------------------------------------------
	def compileAssert(self, assertStatement, ri):
		whats = set()
		for match in CardInfo._WNPAT.finditer(assertStatement):
			try: whats.add(int(match.group(2)))
			except: pass
		ass = CardInfo._wPAT.sub("c.numWhat(", assertStatement)
		ass = CardInfo._WPAT.sub("c.what(", ass)
		try:
			for i,label in enumerate(self.label[ri]):
				if label:
					ass = re.sub(rf"\b{label}\b",
							f"c.numWhat({i})",
							ass)
					whats.add(i)
		except IndexError:
			pass

		whats = list(whats)
		expr = "lambda c:"
		for i in whats:
			expr += f"c.what({i})=='' or "
		expr += f"({ass})"
		return (whats, CardInfo._compileFunc(expr))

	#-----------------------------------------------------------------------
	# Compile or use pre-defined function
	#-----------------------------------------------------------------------
	@staticmethod
	def _compileFunc(s):
		s = s.replace("&&"," and ")
		s = s.replace("||"," or ")
		try:
			return CardInfo._funcs[s]
		except:
			c = compile(s, "<string>", "eval")
			CardInfo._funcs[s] = c
			return c

	# ----------------------------------------------------------------------
	# Return name of compiled function (used for debugging)
	# ----------------------------------------------------------------------
	@staticmethod
	def _compileFuncName(code):
		for n,c in CardInfo._funcs.items():
			if code == c:
				return n

	# ----------------------------------------------------------------------
	# Find cards range(case)
	# ----------------------------------------------------------------------
	def findCase(self, card):
		#if card.tag == "SPECSOUR": breakpoint()
		if card.info.isNone() or not self.range: return None
		if len(self._rangeCode)==1: return 0
		#print()
		#print("findCase:",card.tag,card.whats())
		for i, code in enumerate(self._rangeCode):
			# Check all 'f'-flags
			for w,f,c,s in code:
				if f=='f' and c:
					v = card.evalWhat(w)
					if w>0 and v=="": v=0
					try: v=int(float(v))
					except: pass
					#print(f"   Checking {i=} {w=} {f=} {s=} {v=} RESULT={eval(c)(v)}")
					try:
						if not eval(c)(v):
							#print("   --->>failed")
							break
					except:
						pass
			else:
				# All conditions met
				#print("Passed",i)
				return i
		return 0 # return default case

	# ----------------------------------------------------------------------
	# @param target type to search
	# @return list of whats that contains type target
	#	  tuple - for a range
	# ----------------------------------------------------------------------
	def findType(self, target, case, include_definition=True):
		starget = target+"-"
		# Go through the various whats
		found = []
		if case is None: return found
		range_ = self.range[case]
		for i,r in enumerate(range_):
			if include_definition:		# s10:body
				m = CardInfo._PATDEF.match(r)
				if m: r = m.group(3)	# remove definitions
			if r in (target, starget):
				found.append(i)
			if r=="i" and len(found)==2:
				found.append(i)
				return tuple(found)
		return found

	# ----------------------------------------------------------------------
	# Validate card: return a list of failing variables.
	# ----------------------------------------------------------------------
	def validate(self, card, case=None):
		if self is self._db[None]:
			return ["Error card"]
		#if card.tag == "SPECSOUR": breakpoint()
		if card.input is None: return []
		if case is None:
			case = self.findCase(card)
			if case is None: return []	# No case found!

		# FIXME check special cases
		# 1. Regions check body names inside
		# 2. All references of names if they exist mi, region, bi, di, var, ti
		# 3. Check step for ranges to be >= 0
		# 4. Check limits in USRxxx cards
		# 5. Check logical units usage
		# 6. Check duplicate name definition of bodies, regions...
		# Check all fields
		pb = []
		for w,f,c,s in self._rangeCode[case]:
			x = card.evalWhat(w)
			#if x=="": continue	# Use default
			#print(f"{w=} {f=} {s=} {x=}")

			if f=='b':	# boolean
				if x=="": x = "0"
				if isinstance(x,int):
					if x not in (0,1): pb.append(w)
				elif x.upper() not in ("","0","1","TRUE","FALSE","ON","OFF"):
					pb.append(w)

			elif f=='i':	# integer
				if x=="": x = 0
				try:
					xi = int(float(x))
					if abs(float(x)-float(xi))>1e-9:
						pb.append(w)
						continue
					x = xi
				except (ValueError, TypeError):
					#x = 0
					pb.append(w)
					continue

			elif f=='f':	# integer flag used for selecting case
				if x=="" and w>0: x = 0
				try: x = int(float(x))
				except (ValueError, TypeError): pass

			elif f=='l':	# long integer
				if x=="": x = 0
				try:
					xi = int(float(x))
					if abs(float(x)-float(xi))>1e-9:
						pb.append(w)
						continue
					x = xi
					# Special patch for RADDECAY to restore
					# the long type instead of float
					card.setWhat(w, xi)
				except (ValueError, TypeError):
					#x = 0
					pb.append(w)
					continue

			elif f in ('r','d','L','D','A','E','T'):	# real number
				if x=="": x = 0.0
				try:
					x = float(x)
				except (ValueError, TypeError):
					# Check if it is a Fortran number with D or internal spaces
					try:
						x = float(str(x).upper().replace("D","E"))
					except (ValueError, TypeError):
						pb.append(w)
						continue

			elif f in ("particle","particle-"):
				# FIXME for FLUKA
				pass
				#try:
				#	x = float(x)
				#	if -100.0<x<1000.0 and abs(int(x)) not in FlukaParticle._db:
				#		pb.append(w)
				#except (ValueError, TypeError):
				#	if x:
				#		if x[0]=="-": x=x[1:]
				#		if x not in FlukaParticle._db:
				#			pb.append(w)

			try:
				# FIXME we should provide as dictionary the card
				# in case there is a cross reference to other cards via the labels
				if c and not eval(c)(x):
					pb.append(w)
			except Exception as ex:
				pb.append(w)
				pb.append(str(ex))

		# Check all assert statements
		if self.assert_:
			for i,(whats,code) in enumerate(self._assertCode[case]):
				if whats and not eval(code, card.globalDict)(card):
					pb.extend(whats)
					pb.append(f"Assertion failed: {self.assert_[case][i]}")
					log.error("Assertion FAILED")
					log.errorCont(card)

		return pb

	#-----------------------------------------------------------------------
	# Case insensitive search on the strings of the taginfo
	#-----------------------------------------------------------------------
	def find(self, pattern):
		for pat in pattern:
			found = False
			if pat in self.tag.upper(): found = True
			if self.alias:
				if pat in self.alias.upper(): found = True
			if pat in self.meaning.upper(): found = True
			for C in self.guide:
				if found: break
				for i in C:
					if i is not None and pat in i.upper():
						found = True
						break
			for C in self.label:
				if found: break
				for i in C:
					if i is not None and pat in i.upper():
						found = True
						break

			for name,var in self.property.items():
				if found: break
				for i,vi in enumerate(var.variables):
					if pat in vi.name.upper() or \
					   pat in vi.guidance.upper():
						found = True
						break
			if not found:
				return False
		return True

	# ----------------------------------------------------------------------
	# Add card info to general dictionary
	# ----------------------------------------------------------------------
	@staticmethod
	def add(info):
		CardInfo._db[f"{info.mode}:{info.tag}"] = info

		# Add card info to macro-cards dictionary (if it is a macro-card!))
		if info.otag:
			CardInfo._db_macros[f"{info.mode}:{info.otag}"] = info

		# Add card info to mode dictionary
		try:
			lst = CardInfo._dbmode[info.mode]
		except KeyError:
			if info.mode:
				lst = CardInfo.cards("")[:]	# include basic CardInfos FIXME?
			else:
				lst = []
			CardInfo._dbmode[info.mode] = lst
		lst.append(info)

	# ----------------------------------------------------------------------
	# get card information
	# ----------------------------------------------------------------------
	@staticmethod
	def get(mode, card):
		# Card is found
		if isinstance(card, Card):
			tag = card.tag
		# Card is not found: keep parsed line as tag...
		else:
			tag = card

		try:
			return CardInfo._db[f"{mode}:{tag}"]
		except KeyError:
			try:
				return CardInfo._db[f":{tag}"]
			# ... and return no card info!
			except KeyError:
				return CardInfo.NONE

	# ----------------------------------------------------------------------
	# Returns macro-card information from its tag.
	# ----------------------------------------------------------------------
	@staticmethod
	def getMacroCardInfo(mode, macro_tag):
		try:
			return CardInfo._db_macros[f"{mode}:{macro_tag}"]
		except KeyError:
			return CardInfo.NONE

	# ----------------------------------------------------------------------
	# return all cardinfo for specific mode
	# ----------------------------------------------------------------------
	@staticmethod
	def cards(mode):
		return CardInfo._dbmode.get(mode,[])

	# ----------------------------------------------------------------------
	@staticmethod
	def allcards():
		return CardInfo._db.values()

	# ----------------------------------------------------------------------
	def isNone(self):
		return self.group is None

	# ----------------------------------------------------------------------
	# Sort based on tag
	# ----------------------------------------------------------------------
	@staticmethod
	def sort(cards):
		cards.sort(key=attrgetter("tag"))
		return cards

	# ----------------------------------------------------------------------
	@staticmethod
	def addGroup(group):
		if group not in CardInfo.groups_set:
			if "/" not in group: CardInfo.groups.append(group)
			CardInfo.groups_set.add(group)

	# ----------------------------------------------------------------------
	@staticmethod
	def addDefaults(mode, item, value):
		group,idx = item.split(".")
		dct = CardInfo.defaults[mode]
		try:
			items = dct[group]
		except KeyError:
			dct[group] = items = {}
		items[value] = idx

	# ----------------------------------------------------------------------
	# Post initialization after loading other inputs init
	# ----------------------------------------------------------------------
	@staticmethod
	def post_init():
		#---------------------------------------------------------------
		# for all populate bodies/surfaces
		#---------------------------------------------------------------
		for mode in ("fluka","mcnp","moira","penelope","phits"):
			CardInfo.bodies[mode] = [ x.tag for x in CardInfo.cards(mode)
						if CardInfo.GEOMETRY_BODY_GROUP in x.group or
						   CardInfo.GEOMETRY_SURFACE_GROUP in x.group]
			CardInfo.bodies[mode].sort()
			CardInfo.bodies[None].extend(CardInfo.bodies[mode])
			CardInfo.bodies[None].sort()

		#---------------------------------------------------------------
		# For all
		#---------------------------------------------------------------
		# Cards with no group assign them one
		# or create one
		for info in CardInfo.allcards():
			if info.group: continue
			if info.tag[0]=="/":
				pos = info.tag.find("/",1)
				parent = info.tag[:pos] if pos>0 else info.tag
				pinfo = CardInfo.get(info.mode, parent)
				if pinfo is not CardInfo.NONE and pinfo.group:
					info.setGroup(pinfo.group)
					continue
				# Create a group if it doesn't have one
				group = string.capwords(parent[1:],"/")
			else:
				group = info.tag.capitalize()
			info.setGroup(group)

		# Groups underline
		gunderline = ""
		for group in CardInfo.groups:
			# Find underline
			u = 0
			for ch in group.upper():
				if gunderline.find(ch)==-1:
					gunderline += ch
					break
				u += 1
			else:
				u = None
			CardInfo.groups_underline.append(u)

#===============================================================================
# Base class of Card and Input
# it should be a tree of cards
#===============================================================================
class CardList: #(list)??? instead of self._children
	_INDENT_INC    = {"#if", "#ifdef", "#ifndef", "#elif", "#else", "#include"}
	_INDENT_DEC    = {"#elif", "#else", "#endif", "#endinclude"}

	#-----------------------------------------------------------------------
	def __init__(self):
		# Links to input, parent and children
		self.input     = None		# master input class holding card
		self._children = []		# list of sub-cards, include, voxel...
		self._parent   = None		# Pointer to parent card
		self._indent   =  0		# indent level (pre, include, geometry, transform)
		self._pos      = -1		# if -1 the card is most probably deleted

		# global card information
		self.tag       = None		# card tag
		self.enable    = True		# Enable or Disabled
		self.modified()

	# ----------------------------------------------------------------------
	def modified(self):
		"""set last time modified"""
		self._modifiedTime = time.time()

	# ----------------------------------------------------------------------
	def modifiedTime(self):
		"""return modified time"""
		return self._modifiedTime

	# ----------------------------------------------------------------------
	# UnPickle information to restore
	# ----------------------------------------------------------------------
	def __getstate__(self):
		# return the basic information without the unpicklable entries
		state = {
			"tag"      : self.tag,
			"enable"   : self.enable,
			"_pos"     : self._pos,
		}
		return state

	# ----------------------------------------------------------------------
	def __call__(self):
		return self._children

	# ----------------------------------------------------------------------
	def __getitem__(self, item):
		return self._children[item]

	#-----------------------------------------------------------------------
	def add(self, card, pos=None, renumber=False):
		assert isinstance(card.tag,str)
		card.input  = self.input		# Keep input file link
		card.parent = self
		# update global dictionary?
		if self.input is not None:
			try:
				self.input.cards[card.tag].append(card)
			except KeyError:
				self.input.cards[card.tag] = [card]

		prevCard = None
		children = self._children
		if pos is None or pos>=len(children):
			if children:
				prevCard = children[-1]	# Find prev card
			children.append(card)		# Global cardlist
			card._pos = len(children)-1
		else:
			if pos>0: prevCard = children[pos-1]
			children.insert(pos, card)
			card._pos = pos

		# Check indent level of previous card
		if prevCard:
			if card.enable and prevCard.enable and prevCard.tag in self._INDENT_INC:
				card._indent = prevCard._indent + 1
			else:
				card._indent = prevCard._indent
		else:
			card._indent = 0

		if card._children:
			ind = card._indent + 1
			for child in card._children:
				child._indent = ind

		if card.enable and card.tag in self._INDENT_DEC:
			card._indent = max(0,card._indent-1)

		if renumber:
			p = pos
			if p is None: p = 0
			self.renumber(p)
		#self.modified()
		return card

	#-----------------------------------------------------------------------
	# Put the correct index to cards in the range given by fromPos:toPos
	#-----------------------------------------------------------------------
	def renumber(self, fromPos=0, toPos=-1):
		children = self._children
		if toPos<0 or toPos>len(children):
			toPos = len(children)
		if fromPos <= 0:
			fromPos = 0
			indent  = 0
		else:
			try:
				card   = children[fromPos-1]
				indent = children[fromPos-1]._indent
				if card.enable and card.tag in self._INDENT_INC \
				   or card.tag[0]=="[":
					indent += 1
			except IndexError:
				fromPos = indent = 0

		for i,card in enumerate(self._children[fromPos:toPos], fromPos):
			card._pos = i
			if card.enable and card.tag in self._INDENT_DEC \
			   or card.tag[0]=="[":
				indent = max(0,indent-1)
			card._indent = indent
			if card._children:
				ind = card._indent + 1
				for child in card._children:
					child._indent = ind
			if card.enable and card.tag in self._INDENT_INC \
			   or card.tag[0]=="[":
				indent += 1

		self.modified()
		assert self.checkNumbering()

	#-----------------------------------------------------------------------
	# Verify the internal order of the cards
	#-----------------------------------------------------------------------
	def checkNumbering(self):
		err = indent = 0
		for i, card in enumerate(self._children):
			if card._pos != i:
				err += 1
				log.error(f"Card #{card.pos1()} out of order it should be #{i}")
				log.errorCont(card)
				card._pos = i
			if card.enable and \
			   (card.tag in self._INDENT_DEC or card.tag[0]=="["):
				indent = max(0,indent-1)
			if card.enable and card.indent() != indent:
				log.error(f"Card #{i+1} indent out of order "
					f"card._indent={card._indent} it should be {indent}")
				log.errorCont(card)
			if card.enable and \
			   (card.tag in self._INDENT_INC or card.tag[0]=="["):
				indent += 1
		if err>0:
			log.error(f"{err} cards found out of order. Correcting error")
			log.errorCont(f"*** Please contact {__email__} ***")
			import traceback
			traceback.print_stack()
			#self.renumber()
			return False
		return True

#===============================================================================
# Card base class
# FIXME:
#     subclass (dict) to replace the whats and properties
#     - whats / list arguments are assigned as [#] = value, starting from 0
#     - properties will be assigned as [name] = value
#     - all internal variables not to be saved will be prefixed
#       with "@" [@name] = value
#===============================================================================
class Card(CardList):
	"""Base class for a generic input card"""

	MODE     = ""		# Mode name
	COMMENT  = "*"
	DISABLE  = "!"

	GENERIC   = 0		# card types
	OBJECT    = 1
	BODY      = 2		# Body or Surface
	SURFACE   = 2		# WARNING: Evaluation refers to BODY by number!!!!
	REGION    = 3		# REGION / CELL / MODULE
	CELL      = 3
	MODULE    = 3
	SOLID     = 4
	MATERIALS = 5
	ROTATION  = 6		# rotation/transformation matrix definitions

	# Constants and properties
	ALPHA     = "alpha"

	# Temporary properties starting with @
	ASSIGNMAT = "@assignmat"	# card assigned material
	CACHE     = "@cache"		# cached information, typically processed whats or zones
	CHKSUM    = "@chksum"		# card whats checksum when expressions are present
	COLOR     = "@color"		# indirect displayed color (not saved)
	CONT      = "@cont"		# region has continuation region cards with name="&"
	FILE      = "@file"		# any file associated with card
	HIDDEN    = "@hidden"		# HiddenCardWidget when copy/paste
	ID        = "@id"		# card id in geometry viewer
	INCLUDE   = "@include"		# include input file
	LATTICE   = "@lattice"		# LATTICE assigned card
	MATERIAL  = "@material"		# Material assigned to region
	MATRIX    = "@matrix"		# transformation matrix associated with card
	NODE      = "@node"		# selected node of card
	ORDER     = "@order"		# Order of free variables in card
	PARENT    = "@parent"		# parent card in case of nested cards
	SELECT    = "@select"		# selected status of card in geometry viewer
	TYPE      = "@type"		# card type, as defined above
	USED      = "@used"		# flag if the card is being used or not during conversion
	VOLUME    = "@volume"		# volume (cm³) value associated with region
	ZONE      = "@zone"		# list of zones

	# Region types
	REGION_NORMAL    = 0
	REGION_BLACKHOLE = 1
	REGION_LATTICE   = 2
	REGION_VOXEL     = 3

	# Patterns
	NAMEPAT   = re.compile(r"^[A-Za-z_][A-Za-z0-9_.@\$]*$")
	VARVALPAT = re.compile(r"([A-Za-z@_][A-Za-z-0-9@_.-]*)\s*=\s*(.*)$")
	EXPRNUM   = re.compile(r"^= *(-?[\.0-9]*[eE]?[+\-]?[0-9]*) *$")
	SPLINEPAT = re.compile(r"=\s*spline\(['\"]([A-Za-z_][A-Za-z0-9_.:]*)['\"],\s*frame\s*\)\.[xyz]")

	cardInfo   = CardInfo	# Pointer to Card Info
	classes    = {}		# Subclasses to automatically create card in read()
	globalDict = None	# global dictionary has to be replaced by higher level classes

	# ----------------------------------------------------------------------
	def __init__(self, tag, what=None, comment="", text="", prop=None):
		super().__init__()
		self._init()
		self._what = what if what else []# what list (should be moved to properties)
						# Actually the whole Card should become a dictionary
		self._text = text		# addition information like title or region expression
						# FIXME should be moved to the properties
		if isinstance(comment,list):
			self._comment = '\n'.join(comment)
		else:
			self._comment = comment
		self.setTag(tag)
		if prop is not None:
			self.property.update(prop)

	# ----------------------------------------------------------------------
	# Initialize basic structures
	# ----------------------------------------------------------------------
	def _init(self):
		self.active   = True		# Preprocessor activate/deactivate
		self.property = {}		# properties dictionary
		self.invalid  = []		# Invalid whats & properties
		self._invalid = []		# Additional errors for the card
		self._geo     = False		# FIXME, is it a geometry card?
		self._type    = Card.GENERIC	# type of card

	# ----------------------------------------------------------------------
	# Pickle information to copy
	# ----------------------------------------------------------------------
	def __getstate__(self):
		# return the basic information without the unpicklable entries
		state = super().__getstate__()
		state = {
			"tag"      : self.tag,
			"_what"    : self._what,
			"_text"    : self._text,
			"_comment" : self._comment,
			"property" : self.property,
			"enable"   : self.enable,
		}
		return state

	# ----------------------------------------------------------------------
	# UnPickle information to restore
	# ----------------------------------------------------------------------
	def __setstate__(self, state):
		self.__init__(state["tag"])
		self.__dict__.update(state)
		self.setTag(self.tag)

	# ----------------------------------------------------------------------
	# Clone card
	# ----------------------------------------------------------------------
	def clone(self):
		"""clone card object"""
		card = self.__class__(self.tag,
				self._what[:],
				self._comment,
				self._text)
		card.enable = self.enable
		card._pos   = self._pos
		card.copyProperties(self)
		return card

	# ----------------------------------------------------------------------
	# I/O
	# ----------------------------------------------------------------------
	def args(self):		return self._what		# generic name of arguments
	def nargs(self):	return len(self._what)
	def whats(self):	return self._what		# fluka specific name
	def nwhats(self):	return len(self._what)
	def name(self):		return str(self.what(0))	# Force always a string
	def text(self):		return self._text
	def comment(self):	return self._comment
	def isGeo(self):	return self._geo
	def type(self):		return self._type
	def pos(self):		return self._pos
	def pos1(self):		return self._pos+1		# base 1 position for printing
	def indent(self):	return max(0,self._indent)
	expr  = text
	title = text

	# ----------------------------------------------------------------------
	# set
	# ----------------------------------------------------------------------
	def setName(self, name): self.setWhat(0, name)
	def setText(self, e):
		self._text = str(e)
		self.modified()
	setExpr  = setText
	setTitle = setText

	# ----------------------------------------------------------------------
	# Sort based on card position
	# ----------------------------------------------------------------------
	@staticmethod
	def sort(cards):
		cards.sort(key=attrgetter("_pos"))
		return cards

	# ----------------------------------------------------------------------
	# Return what value -1=text, 0=name
	# ----------------------------------------------------------------------
	def what(self, n):
		if n<0:
			return self._text
		else:
			try:	return self._what[n]
			except: return ""

	# ----------------------------------------------------------------------
	# Return what as integer
	# ----------------------------------------------------------------------
	def intWhat(self, n, default=0):
		"""try to return an integer from what"""
		try:
			return int(float(self.evalWhat(n)))
		except:
			return default

	# ----------------------------------------------------------------------
	# Return what as floating point number
	# ----------------------------------------------------------------------
	def numWhat(self, n, default=0.0):
		"""@return a number from what"""
		try:
			return float(self.evalWhat(n))
		except:
			return default

	# ----------------------------------------------------------------------
	# Return color hex code or string
	# ----------------------------------------------------------------------
	def colorWhat(self, n):
		color = self.evalWhat(n)
		try:
			color = int(float(color))
			return f"#{color:06X}"
		except:
			return color

	# ----------------------------------------------------------------------
	# Simplify a single what expression
	# ----------------------------------------------------------------------
	def _simplifyExpr(self, w):
		# Fixed number as expression
		pat = Card.EXPRNUM.match(w)
		if pat is not None:
			try:
				return float(pat.group(1))
			except:
				pass
		return w

	# ----------------------------------------------------------------------
	# Evaluate an expression
	# ----------------------------------------------------------------------
	def evalExpr(self, w, dollar=True):
		if not w or not isinstance(w,str):
			return w
		elif w[0] in "=([{" or self.tag=="#define":
			if self.input is None: return w
			expr = w[1:] if w[0]=="=" else w

			# Check for possible vector definitions
			# replace any { to Vector( and } to )
			if "{" in expr:
				expr = expr.replace("{","Vector(")
				expr = expr.replace("}",")")
				#expr = expr.replace("[[","Matrix(")
				#expr = expr.replace("]]",")")
			if self.MODE=="":
				globalDict = self.input.globalDict
			else:
				globalDict = self.globalDict

			# In case of recursive call save the card and restore it later on
			save = self.input.localDict.card
			self.input.localDict.card = self
			try:
				value = eval(expr, globalDict, self.input.localDict)
				self.input.localDict.card = save
				if expr[0] in "'\"" and isinstance(value,str):
					value = repr(value)
			except Exception as err:
				if Global.developer:
					import traceback; traceback.print_stack()
					log.error(f"Evaluate expr in card: {self.rawStr()}")
					log.errorCont(err)
					log.errorCont(f"what: {w}")
				return w #"?"
			self.input.localDict.card = save
			return Evaluation.float_or_int(value)
		elif dollar:
			if w[0] == "$":
				var  = w[1:]
				sign = 1.0
			elif w[:2] == "-$":
				var  = w[2:]
				sign = -1.0
			else:
				return w
			# check defines
			try:
				define = self.input.cardsCache("#define")[var]
				val = define.evalWhat(1)
				# Convert to int or float if possible
				try:
					f = sign*float(val)
					if float(int(f)) == f:
						return int(f)
					return f
				except:
					if sign<0.0:
						return "-"+str(val)
					else:
						return val
			except KeyError:
				pass
		return w

	#-----------------------------------------------------------------------
	# @return is the what string is an evaluated function
	#-----------------------------------------------------------------------
	@staticmethod
	def isEvalStr(w):
		return isinstance(w,str) and w and (w[0] in ("=","$") or w[:2]=="-$")

	# ----------------------------------------------------------------------
	# Is evaluated what
	# ----------------------------------------------------------------------
	def isEvalWhat(self, n):
		return Card.isEvalStr(self[n])

	# ----------------------------------------------------------------------
	# Return evaluated what, (substitute $dollars defines with values if needed)
	# ----------------------------------------------------------------------
	def evalWhat(self, n, dollar=True):
		if self.active:
			return self.evalExpr(self[n], dollar)
		else:
			return self[n]

	# ----------------------------------------------------------------------
	# Return evaluated what,
	# but NOT empty space (used on $xxx cards)
	# ----------------------------------------------------------------------
	def evalName(self, n):
		#Global.dump("Input.evalName",self.tag)
		p = int(self.tag[0] in ("#","$","!"))
		prefix = self.tag[p]+CardInfo._VOWELSPAT.sub("",self.tag[p+1:])
		num = f"{self._pos}.{n}"
		return prefix[:8-len(num)] + num

	# ----------------------------------------------------------------------
	def evalWhatOrName(self, n, zero=False):
		w = self[n]
		if isinstance(w,str) and w!="" and w[0]=="=":
			w = self._simplifyExpr(w)
			if isinstance(w,str) and w[0]!="=": return w
		w = self.evalWhat(n,False)
		if zero and w=="": return 0.
		return w

	# ----------------------------------------------------------------------
	# If expression string is a spline pattern
	#    =spline('spline_name',frame).[xyz]
	# @return the spline name if any
	# ----------------------------------------------------------------------
	@staticmethod
	def isSplineStr(w):
		try:
			m = Card.SPLINEPAT.match(w)
		except TypeError:
			return None
		if m: return m.group(1)
		return None

	# ----------------------------------------------------------------------
	# Check a cards checksum if it contains functions and they are modified
	# externally with an editor
	# ----------------------------------------------------------------------
	def checkSum(self):
		assert self[Card.CHKSUM] is not None
		chksum = 0
		for i,w in enumerate(self._what):
			# ignore the [-]$var
			if isinstance(w,str) and w and w[0]=="=":
				try:
					what = bmath.format(round(float(self.evalWhat(i)),10),10)
					chksum ^= hash(what)
					#Global.dump("checkSum", self.tag, self.name(), i,
					#	repr(what), repr(w), chksum)
				except ValueError:
					pass
		#if chksum != self[Card.CHKSUM]:
		#	Global.dump("*** CheckSum", self.tag, self.name())
		#	Global.dump("     card.CHKSUM=",self[Card.CHKSUM])
		#	Global.dump("     calc.CHKSUM=",chksum)
		return chksum == self[Card.CHKSUM]

	# ----------------------------------------------------------------------
	# set card state
	# ----------------------------------------------------------------------
	def setEnable(self, e):
		"""set state enable(True)/disable(False)"""
		self.enable = e
		for child in self._children:
			child.setEnable(e)
		self.modified()

	# ----------------------------------------------------------------------
	# Disable card
	# ----------------------------------------------------------------------
	def disable(self):
		"""disable card"""
		self.enable = False
		for child in self._children:
			child.disable()
		self.modified()

	# ----------------------------------------------------------------------
	# Set Active state
	# ----------------------------------------------------------------------
	def setActive(self, a):
		"""set active state of card"""
		self.active = a
		for child in self._children:
			child.setActive(a)

	# ----------------------------------------------------------------------
	# Ignore card if not active or not enabled
	# ----------------------------------------------------------------------
	def ignore(self):
		return not self.enable or not self.active

	# ----------------------------------------------------------------------
	def notIgnore(self):
		return self.enable and self.active

	# ----------------------------------------------------------------------
	# Check if card is active
	# ----------------------------------------------------------------------
	def notActive(self):
		return not self.active

	# ----------------------------------------------------------------------
	def __delitem__(self, item):
		try:
			del self.property[item]
		except KeyError:
			pass

	# ----------------------------------------------------------------------
	# get card information
	# ----------------------------------------------------------------------
	def __getitem__(self, n=None, default=None):
		"""get card information"""
		if isinstance(n, int):
			return self.what(n)
		else:
			return self.property.get(n, default)
	get = __getitem__

	# ----------------------------------------------------------------------
	# set information to card
	# ----------------------------------------------------------------------
	def __setitem__(self, n, v):
		"""set value to what(#, name, comment, text)"""
		if isinstance(n, int):
			self.setWhat(n, v)
		elif n:
			self.property[n] = v	# low level direct setting
			if n[0] != "@": self.modified()
	set = __setitem__

	# ----------------------------------------------------------------------
	# set comment
	# ----------------------------------------------------------------------
	def setComment(self, comment=""):
		"""set comment string"""
		if isinstance(comment,list):
			self._comment = "\n".join(comment)
		else:
			self._comment = comment
		self.modified()

	# ----------------------------------------------------------------------
	# Append comment
	# ----------------------------------------------------------------------
	def appendComment(self, comment):
		"""append comment"""
		if not comment: return
		if isinstance(comment,list):
			comment = "\n".join(comment)
		if self._comment:
			self._comment += f"\n{comment}"
		else:
			self._comment = comment
		self.modified()

	# ----------------------------------------------------------------------
	def appendText(self, txt):
		if self._text:
			self._text += f"\n{txt}"
		else:
			self._text += txt
	appendExpr = appendText

	# ----------------------------------------------------------------------
	# Set the whats list
	# ----------------------------------------------------------------------
	def setWhats(self, whats):
		self._what = whats
		self.modified()

	# ----------------------------------------------------------------------
	# Set whats list ONLY they are currently not an expression
	# ----------------------------------------------------------------------
	def setWhatsIfValue(self, whats):
		current = self._what
		self._what = whats
		for i,w in enumerate(current):	# replace expressions
			if Card.isEvalStr(w):
				self._what[i] = w
		self.modified()

	# ----------------------------------------------------------------------
	# Trim extra whats of card when we change the Tag
	# ----------------------------------------------------------------------
	def trimWhats(self):
		if len(self._what) > self.info.maxwhats:
			self.setNWhats(self.info.maxwhats)

	# ----------------------------------------------------------------------
	# Set a what value
	# ----------------------------------------------------------------------
	def setWhat(self, w, value):
		if w<0:
			self._text = value
		else:
			# Extend list if needed
			if w>=len(self._what):
				self._what.extend([""] * (w-len(self._what)+1))
			self._what[w] = value
		self.modified()

	# ----------------------------------------------------------------------
	# Set a what value if current value is not an expression
	# ----------------------------------------------------------------------
	def setWhatIfValue(self, w, value):
		if self.isEvalWhat(w): return False
		self.setWhat(w, value)
		return True

	# ----------------------------------------------------------------------
	# Set the total number of whats
	# Usually to remove whats
	# ----------------------------------------------------------------------
	def setNWhats(self, n):
		"""set the total number of whats"""
		l = len(self._what)
		if n < l:
			del self._what[n:]
		elif n > l:
			self._what += [""] * (n-l)
		self.modified()

	# ----------------------------------------------------------------------
	# Append whats
	# ----------------------------------------------------------------------
	def appendWhats(self, whats, pos=None):
		if pos is None:
			# append the whats
			# chksum remains the same since there are not fields to override
			self._what.extend(whats)
		else:
			# insert the modified fields
			# chksum to be updated: the fields that are set are from the
			# _addCard that set the expressions
			chksum = self.getProperty(Card.CHKSUM, 0)
			for w in whats:
				if self.what(pos)=="":
					self.setWhat(pos,w)
				else:
					try:
						what = bmath.format(round(float(w),10),10)
					except (ValueError, TypeError):
						what = w
					chksum ^= hash(what)
					#Global.dump("appendWhats CHKSUM", self.tag, self.name(),
					#		pos, repr(w), repr(what), chksum)
				pos += 1
			if chksum:
				self[Card.CHKSUM] = chksum

	# ----------------------------------------------------------------------
	# Pad whats to make them multiple of n*6+1
	# ----------------------------------------------------------------------
	def padWhats(self):
		rem = 6-(len(self._what)-1)%6
		if rem>1: self._what.extend([""]*rem)

	# ----------------------------------------------------------------------
	# Set a property value, IF the property exists in the card.
	# Returns a boolean, representing whether the property was found.
	# This is different from directly calling setProperty,
	# which, when property is not found, would create a new entry for it.
	# ----------------------------------------------------------------------
	def setPropertyIfExists(self, property_name, values):
		# Found property: set it!
		if property_name in self.info.property:
			self.setProperty(property_name, values)
			return True
		# Property was not found
		return False

	# ----------------------------------------------------------------------
	# set/get user property to be used from the Card.read()
	# WARNING:
	#    This is a bit more advanced than the simple __setitem__
	#    It checks for a possible group and splits it to components
	# ----------------------------------------------------------------------
	def setProperty(self, name, value):
		self.modified()
		try:
			pinfo = self.info.property[name]
		except KeyError:
			# default action
			self.property[name] = value
			return

		if pinfo.array:
			n = self.property.get(f"{name}:0",0)+1
			self.property[f"{name}:0"] = n
			aname = f"{name}:{n}"
		else:
			aname = name	# array name = name

		if len(pinfo.variables)==1:
			self.property[aname] = value
		else:
			if isinstance(value,str):
				lst = Stream.str2list(value)
			else:
				lst = [value]
			value = []
			for i,vi in enumerate(pinfo.variables):
				if name==vi.name:	# against original name (not aname)
					varname = aname
				else:
					varname = "%s.%s"%(aname,vi.name)
				try:
					self.property[varname] = lst[i]
				except IndexError:
					return

	# ----------------------------------------------------------------------
	# Set a property value if current value is not an expression
	# ----------------------------------------------------------------------
	def setPropertyIfValue(self, name, value):
		try:
			if Card.isEvalStr(self.property[name]): return False
		except KeyError:
			pass
		self.property[name] = value
		return True

	# ----------------------------------------------------------------------
	# @return property value or return default
	# ----------------------------------------------------------------------
	def getProperty(self, name, default=None):
		return self.property.get(name, default)

	# ----------------------------------------------------------------------
	# @return evaluated property expression if any
	# ----------------------------------------------------------------------
	def evalProperty(self, name, default=None):
		prop = self.property.get(name, default)
		if Card.isEvalStr(prop):
			return self.evalExpr(prop)
		else:
			return prop

	# ----------------------------------------------------------------------
	# @return evaluated property as a integer number
	# ----------------------------------------------------------------------
	def intProperty(self, name, default=0):
		try:
			return int(float(self.evalProperty(name, default)))
		except:
			return default

	# ----------------------------------------------------------------------
	# @return evaluated property as a floating point number
	# ----------------------------------------------------------------------
	def numProperty(self, name, default=0.0):
		try:
			return float(self.evalProperty(name, default))
		except:
			return default

	# ----------------------------------------------------------------------
	def clearProperties(self, all_=True):
		self.modified()
		if not self.property: return
		if all_:
			self.property.clear()
		else:
			# Clear only user properties,
			# ignore the system ones starting with @
			for n in list(self.property.keys()):
				if n[0]!="@":
					del self.property[n]

	# ----------------------------------------------------------------------
	# Copy properties from card
	# ----------------------------------------------------------------------
	def copyProperties(self, card):
		self.property = card.property.copy()	# Swallow copy

	# ----------------------------------------------------------------------
	# Return units used by card
	# ----------------------------------------------------------------------
	def fileunits(self, absolute=True):
		return ()

	# ----------------------------------------------------------------------
	# Change tag of card, reduce the number of whats to the new card
	# XXX Do not forget to update the Input.cards... before
	# ----------------------------------------------------------------------
	def setTag(self, tag):
		"""change tag of card"""
		self.tag  = tag if tag else "error"
		self.info = self.cardInfo.get(self.MODE, tag)
		if tag in CardInfo.bodies.get(self.MODE,[]):
			self._type = Card.BODY
			self._geo  = True
		elif tag in CardInfo.FLAIR_TAGS:
			self._type = Card.OBJECT
		else:
			self._type = Card.GENERIC
		self.modified()

	# ----------------------------------------------------------------------
	# Return card's cardInfo case (range)
	# ----------------------------------------------------------------------
	def case(self):
		return self.info.findCase(self)

	# ----------------------------------------------------------------------
	# Return a specific type from the card, see database from more info
	# ----------------------------------------------------------------------
	def findType(self, target):
		"""return whats with a specific type from the card, particle, material, ..."""
		case = self.info.findCase(self)
		lst  = self.info.findType(target, case)
		if isinstance(lst, list):
			return [self.what(x) for x in lst]
		else:
			return (self.what(lst[0]), self.what(lst[1]), lst[2])

	# ----------------------------------------------------------------------
	# Validate card values
	# ----------------------------------------------------------------------
	def validate(self, case=None):
		self.invalid = self.info.validate(self, case)
		for item in self._invalid:
			if  item not in self.invalid:
				self.invalid.append(item)
		return self.invalid

	# ----------------------------------------------------------------------
	# test card against a filter-function
	# Filter format  type [function] \n what \n value
	#
	# func: [type][function]
	#	type:	i=int
	#		f=float
	#		s=string
	#	function:
	#		a=abs
	# var:
	#	0..9 = whats
	#	  *  = anywhere
	#	  b  = body
	#	  d  = detector
	#	  s  = sdum
	#	  m  = material
	#	  p  = particle	(range if tuple)
	#	  r  = region		(range if tuple)
	#	  t  = transformation
	#	  u  = unit
	# value:
	#	anything int, float or string
	#
	# keepcase:
	#	True for matching letter case,
	#	otherwise search in UPPERcase
	# ----------------------------------------------------------------------
	def test(self, func, var, value, keepcase=False):
		"""check card against a filter function func"""
		# Search everywhere
		if var=="*":
			if keepcase:
				if self.tag.find(value)>=0:      return True
				if self._comment.find(value)>=0: return True
				if self._text.find(value)>=0:    return True
				for w in self._what:
					if str(w).find(value)>=0: return True
			else:
				if self.tag.upper().find(value)>=0:      return True
				if self._comment.upper().find(value)>=0: return True
				if self._text.upper().find(value)>=0:    return True
				for w in self._what:
					if str(w).upper().find(value)>=0: return True
			return False

		#if var is None: return False
		if func=="ia":		# Absolute integer
			return abs(self.intWhat(var)) == int(value)

		elif func=="f":		# Floating point
			return self.numWhat(var) == float(value)

		elif func=="i":		# Integer
			return self.intWhat(var) == int(value)

		elif func=='sa':	# "Absolute" string (remove the minus '-')
			if var=="s":
				w = self.name()
			else:
				w = str(self.what(var))
			if w and w[0]=='-': w=w[1:]
			if not keepcase: w = w.upper()
			return w == str(value)

		elif func=="s":
			if var=="s":
				return self.name() == str(value)
			else:
				return str(self.what(var)) == str(value)

		return False

	# ----------------------------------------------------------------------
	# FIXME this check should be done in the invalid
	# ----------------------------------------------------------------------
	def clearInvalid(self):
		self._invalid = []

	# ----------------------------------------------------------------------
	# FIXME this check should be done in the invalid
	# ----------------------------------------------------------------------
	def addInvalid(self, item):
		if item not in self._invalid:
			self._invalid.append(item)

		# Append to the invalid cards
		if self.invalid:
			if item not in self.invalid:
				self.invalid.append(item)
		else:
			self.invalid = [item]

	# ----------------------------------------------------------------------
	# Convert card to names or numbers
	# ----------------------------------------------------------------------
	def convert(self, tonames=True):
		pass

	# ----------------------------------------------------------------------
	# Return differences of the two cards as a list of integers
	#     -3 = enable/disable
	#     -2 = Comment
	#     -1 = Extra
	#      0 = Sdum
	#     1..n = What's
	# ----------------------------------------------------------------------
	def diff(self, card):
		df = []
		if self.enable != card.enable:
			df.append(-3)
		if self._comment != card._comment:
			df.append(-2)
		if self._text != card._text:
			df.append(-1)
		minwhats = min(self.nwhats(), card.nwhats())
		maxwhats = max(self.nwhats(), card.nwhats())
		for w in range(0, minwhats):
			if w > card.nwhats():
				df.append(w)
			elif self.what(w) != card.what(w):
				df.append(w)
		if minwhats != maxwhats:
			df.extend(range(minwhats, maxwhats))
		return df

	# ----------------------------------------------------------------------
	# @return error message string
	# ----------------------------------------------------------------------
	def errorMessage(self, labels=None):
		case   = self.case()
		range_ = self.info.range[case]
		try:
			extra  = self.info.guide[case]
		except IndexError:
			extra = []
		msg = ""
		displayed = set()
		for e in self.invalid:
			if e is None: continue
			if e in displayed: continue	# skip multiple messages
			displayed.add(e)

			try:    label = labels[e]
			except: label = ""

			if msg: msg += f"\n{log.ERRORCONT}"

			try:    extra_e = extra[e]
			except: extra_e = ""

			if isinstance(e,str):
				msg += e
			elif label:
				msg += f"What={e}  Label={label}  Type={range_[e]}: {extra_e}"
			else:
				msg += f"What={e}  Type={range_[e]}: {extra_e}"
		return msg

	# ======================================================================
	# String functions
	# ======================================================================
	# Return a formatted comment string
	# ----------------------------------------------------------------------
	def commentStr(self):
		"""return a comment string"""
		if not self._comment: return ""
		# if comment=="..\n" with an empty line in the end splitlines
		# is ignoring it.e.g.
		#      "\n".join(["","",""]) => "\n\n"
		#      "\n\n".splitlines()   => ["",""]
		#      "\n\n".split("\n")    => ["","",""]
		lines = [f"{self.COMMENT}{x}".rstrip() for x in self._comment.splitlines()]
		return "\n".join(lines)

	# ----------------------------------------------------------------------
	# Return a string of property combined sub-variables
	# @param name	property name
	# @param used	expects a set() to fill with used variables
	# @param sep	separator to use for variables
	# @param default if value is not supplied how to interpret the default one
	# @return property setting string, None if not found
	# ----------------------------------------------------------------------
	def propertyStr(self, name, used=None, sep="=", default='""', evaluate=False):
		try:
			pinfo = self.info.property[name]
		except KeyError:
			# Unknown property
			value = self.property.get(name,"")
			if value not in (None,""):
				return name+sep+Stream.quoteIfNeeded(self.property[name])
			return None

		if pinfo.array:
			n = self.intProperty(f"{name}:0",0)
			if n == 0: return None
		else:
			n = 1

		out = []
		for i in range(1,n+1):
			if pinfo.array:
				aname = f"{name}:{i}"
			else:
				aname = name
			value = []
			found = False
			for vi in pinfo.variables:
				if len(pinfo.variables)==1 or name==vi.name:
					varname = aname
				else:
					varname = f"{aname}.{vi.name}"
				v = self.get(varname,"")
				if v != "":	# can be also 0 or False
					if isinstance(v,str) and evaluate and v[0]=="=":
						value.append(str(self.evalExpr(v)))
					else:
						value.append(Stream.quoteIfNeeded(v))
					found = True
					if used is not None:
						used.add(varname)
				else:
					value.append(default)
			while value and value[-1]==default:
				value.pop()	# trim trailing empty values
			if value or found:
				out.append(name+sep+" ".join(value))

		if out: return "\n\t".join(out)
		return None

	# ----------------------------------------------------------------------
	# Low level conversion to string
	# @param mode is only used in writing inside the flair file
	# ----------------------------------------------------------------------
	def _toStr(self, prefix="", mode=None, properties=True, **kwargs):
		if mode is None or self.MODE in ("",mode):
			ctag = f"{prefix}{self.tag}"
		else:
			ctag = f"{prefix}{self.MODE}:{self.tag}"

		# WARNING:
		# if mode is None: then we are writing in the native .inp format
		# else:            we are writing in the .flair file
		if self.tag[0]=="#" and mode is None:
			if self.nwhats()>0:
				try:
					# FIXME undef?
					if self.tag == "#define":
						if self.nwhats()>1:
							w = self.evalWhat(1)
							s = f"{ctag} {str(self._what[0])} " \
								f"{bmath.format(w,-17,useDot=False)}"
						else:
							s = f"{ctag} {self._what[0]}"
					else:
						# FIXME if =expr?
						w = " ".join([str(self.evalExpr(x))
								for x in self._what])
						s = f"{ctag} {w}"
					s = s.strip()
				except IndexError:
					pass
			else:
				s = ctag
		else:
			# Parameters
			whats = [bmath.format(x,-17) for x in self._what]
			while whats and whats[-1]=='':
				whats.pop()	# trim trailing empty values
			s = ctag + " " + Stream.list2str(whats) if whats else ctag
			if self._text:
				for line in self._text.splitlines():
					if line and line[0] in " \t":
						s += f"\n\t{line!r}"
					else:
						s += f"\n\t{line}"

		# Write properties if any
		if self.property and properties:
			# Organize properties
			used = set()

			# Maybe go with the order that was defined
			for name in self.info.property.keys():
				v = self.propertyStr(name, used)
				if v is not None:
					s += f"\n\t{v}"

			#order = self.get(Card.ORDER,[])
			#for n in order:
			#	if n=="\n": continue
			#	v = self[n]
			#	if v in ("",None): continue
			#	s += "\t%s=%s\n"%(n,v)

			# For anything else
			for n,v in self.property.items():
				if n in used: continue
				if ":" in n: continue
				#if n in order: continue	# skip what ever was written
				if v in ("",None): continue
				#if n[0]=="@": continue
				if n[0]=="@" and (n!=Card.SELECT or v==0): continue
				s += f"\n\t{n}={v}"

		# Write Children if any
		if self._children:
			amp = prefix.find("&")
			if amp>=0:
				cprefix = f"{prefix[:amp]}\t& "
			else:
				cprefix = f"{prefix}\t& "
			for child in self._children:
				s += "\n"+Card._toStr(child, cprefix, mode, properties, **kwargs)
		return s

	# ----------------------------------------------------------------------
	# Return card without evaluating whats
	# ----------------------------------------------------------------------
	def rawStr(self):
		w = [str(x) for x in self._what]
		w.insert(0,self.tag)
		w.append(self._text)
		s = " ".join(w)
		if self._comment:
			return f"{self.commentStr()}\n{s}"
		else:
			return s

	# ---------------------------------------------------------------------
	# return string with only the cards
	# ----------------------------------------------------------------------
	def toStr(self, **kwargs):	# FIXME fluka uses fmt
		return self._toStr(self.DISABLE if not self.enable else "", **kwargs)
	__str__ = toStr

	# ----------------------------------------------------------------------
	# return string representation of card
	# ----------------------------------------------------------------------
	def __repr__(self):
		return f"{object.__repr__(self)}: {Card.__str__(self)}"

	# ======================================================================
	# I/O flair native format
	# ======================================================================
	# Write a card in flair native format
	# ----------------------------------------------------------------------
	def write(self, sout, mode):
		for l in self._comment.splitlines():
			sout.write(f"{sout.commentChar}{l.rstrip()}\n")
		# Make a low level call to Card._toStr() to avoid
		# a possible call on the FlukaCard or MoiraCard
		sout.writeline(Card._toStr(self, Card.DISABLE
				if not self.enable else "", mode))

	# ----------------------------------------------------------------------
	# Read and return a card from flair native format
	# ----------------------------------------------------------------------
	@staticmethod
	def read(sin, mode=""):
		comment = []
		while True:
			if sin.readline() is None: return None
			sin.skipSpaces()
			if sin.eol():
				comment.append("")
			elif sin.line[sin.column] == sin.commentChar:
				comment.append(sin.line[sin.column+1:].rstrip())
			else:
				if sin.line[sin.column] == Card.DISABLE:
					enable = False
					sin.column += 1
					sin.skipSpaces()
				else:
					enable = True
				if sin.line[sin.column] == "&":
					sin.column += 1
					sin.skipSpaces()
				tag = sin.parseIdentifier()
				if tag: break
				return None
		if tag=="EndInput" and sin.prefix==0: return None
		#Global.dump(f"+++ line={sin.line}")
		#Global.dump(f">>>\ttag={tag}\tcol={sin.column}")

		# Check for disabled card starting with !
		# * -> comments
		# # -> preprocessor
		# ! -> disabled cards (WARNING old flair cards were starting with !)
		# $ -> start/stop_transform
		#if tag[0]==Card.DISABLE:
		#	enable = False
		#	tag = tag[1:]
		#else:
		#	enable = True

		# Check for a different type of card
		card = Card.new(tag, mode)
		card.enable = enable
		if comment: card._comment = "\n".join(comment) + "\n"
		prefix = sin.prefix
		card._what = sin.parseList()
		#Global.dump(f">>>\twhat={card._what}")
		text    = []
		order   = []	# remember variable order
		comment = []
		while True:
			if sin.eol():
				# Loop to ignore empty lines
				while True:
					sin.readline()
					if sin.line: break
					text.append("")
				if sin.prefix <= prefix:
					sin.saveline()
					break
			if sin.line is None: break
			if sin.prefix <= prefix: break
			if sin.column<len(sin.line) and \
			   sin.line[sin.column] in (sin.commentChar, "&"):
				## remember prefix to abort if needed
				#prefix = sin.prefix
				# start a child command
				#if "phits.out" in sin.line: breakpoint()
				sin.saveline()
				#if tag == "[Material]": breakpoint()
				card.add(Card.read(sin,mode))
				#Global.dump(">c>",f"Card: {card.tag} child: #{len(card._children)} {self._children[-1].tag}")
				continue
			sin.skipSpaces()
			# FIXME make more intelligent
			# parse directly stream and check if the
			# variable name is a "group"
			m = Card.VARVALPAT.match(sin.line[sin.column:])
			if m is not None:
				name  = m.group(1)
				card.setProperty(name, Evaluation.float_or_int(m.group(2)))
				order.append(name)
				order.append("\n")
			elif sin.eol():
				text.append("")
				continue
			elif sin.line[sin.column] in "'\"":
				text.append(sin.parseString())
			else:
				text.append(sin.line[sin.column:])
			sin.column = len(sin.line)	# temporary hack!!!
		card.setText("\n".join(text))
		card[Card.ORDER] = order		# to be used for the layout
							# and latter writing
		return card

	# ----------------------------------------------------------------------
	# Return a new card based on mode
	# ----------------------------------------------------------------------
	@staticmethod
	def new(tag, defmode=""):
		# Check for a different type of card
		if tag[0] in "#" or tag in CardInfo.FLAIR_TAGS:
			return Card(tag)	# flair card
		fcolon = tag.find(":")
		if fcolon<1:
			return Card.classes[defmode](tag)
		else:
			return Card.classes[tag[:fcolon]](tag[fcolon+1:])

	# ----------------------------------------------------------------------
	# @see new()
	# ----------------------------------------------------------------------
	@staticmethod
	def splitTag(tag, defmode=""):
		if tag[0] in "#!":
			return "",tag
		fcolon = tag.find(":")
		if fcolon<1:
			return defmode, tag
		else:
			return tag[:fcolon], tag[fcolon+1:]

	# ----------------------------------------------------------------------
	# register class for automatic creation of card
	# ----------------------------------------------------------------------
	@staticmethod
	def register_class(name, class_):
		Card.classes[name] = class_

	# ----------------------------------------------------------------------
	# return all registered mode names
	# ----------------------------------------------------------------------
	@staticmethod
	def modes():
		return Card.classes.keys()

#===============================================================================
# CSG geometry: Body and regions cards
# Special methods for bodies
# @see Card
#===============================================================================
class CSGCard:
	# ----------------------------------------------------------------------
	# Special body method returning various geometrical information
	# ... if they are defined
	# ----------------------------------------------------------------------
	def bodyP(self):
		"""Position of body"""
		if   self.tag == "RPP":
			return bmath.Vector(	self.numWhat(1),
						self.numWhat(3),
						self.numWhat(5))
		elif self.tag == "PLA":
			return bmath.Vector(	self.numWhat(4),
						self.numWhat(5),
						self.numWhat(6))
		elif self.tag == "XYP":
			return bmath.Vector(0.0, 0.0, self.numWhat(1))
		elif self.tag == "XZP":
			return bmath.Vector(0.0, self.numWhat(1), 0.0)
		elif self.tag == "YZP":
			return bmath.Vector(self.numWhat(1), 0.0, 0.0)
		elif self.tag in ("XCC", "XEC"):
			return bmath.Vector(0.0, self.numWhat(1), self.numWhat(2))
		elif self.tag in ("YCC", "YEC"):
			return bmath.Vector(self.numWhat(2), 0.0, self.numWhat(1))
		elif self.tag in ("ZCC", "ZEC"):
			return bmath.Vector(self.numWhat(1), self.numWhat(2), 0.0)
		elif self.tag in ("BOX", "SPH", "RCC", "TRC", "WED", "REC", "ELL", "ARB") or \
		     self._type == Card.OBJECT:
			return bmath.Vector(	self.numWhat(1),
						self.numWhat(2),
						self.numWhat(3))
		else:
			return None
	bodyP1 = bodyP

	# ----------------------------------------------------------------------
	def bodyPo(self):
		"""Outer position of body"""
		if   self.tag == "RPP":
			return bmath.Vector(	self.numWhat(2),
						self.numWhat(4),
						self.numWhat(6))
		elif self.tag in ("RCC", "TRC", "REC"):
			return bmath.Vector(self.numWhat(1)+self.numWhat(4),
					    self.numWhat(2)+self.numWhat(5),
					    self.numWhat(3)+self.numWhat(6))
		else:
			return None

	# ----------------------------------------------------------------------
	# Second point of body
	# ----------------------------------------------------------------------
	def bodyP2(self):
		if   self.tag in ("ELL", "ARB"):
			return bmath.Vector(	self.numWhat(4),
						self.numWhat(5),
						self.numWhat(6))
		else:
			return None

	# ----------------------------------------------------------------------
	# WARNING index starting from 1
	# ----------------------------------------------------------------------
	def bodyPn(self, n):
		if   self.tag == "ARB":
			i = 3*n-2
			return bmath.Vector(	self.numWhat(i),
						self.numWhat(i+1),
						self.numWhat(i+2))
		else:
			return None

	# ----------------------------------------------------------------------
	def bodyN(self):
		"""Normal of body"""
		if   self.tag in ("PLA","P"):
			return bmath.Vector(	self.numWhat(1),
						self.numWhat(2),
						self.numWhat(3))
		elif self.tag in ("XYP", "ZCC", "ZEC"):
			return bmath.Vector.Z
		elif self.tag in ("XZP", "YCC", "YEC"):
			return bmath.Vector.Y
		elif self.tag in ("YZP", "XCC", "XEC"):
			return bmath.Vector.X
		else:
			return None

	# ----------------------------------------------------------------------
	def bodyX(self):
		"""X-vector of body"""
		if   self.tag == "RPP":
			return bmath.Vector(self.numWhat(2)-self.numWhat(1), 0.0, 0.0)
		elif self.tag in ("BOX", "WED"):
			return bmath.Vector(	self.numWhat(4),
						self.numWhat(5),
						self.numWhat(6))
		elif self.tag == "REC":
			return bmath.Vector(	self.numWhat(7),
						self.numWhat(8),
						self.numWhat(9))
		elif self.tag == "XEC":
			return bmath.Vector(0.0, self.numWhat(3), 0.0)
		elif self.tag == "YEC":
			return bmath.Vector(0.0, 0.0, self.numWhat(3))
		elif self.tag == "ZEC":
			return bmath.Vector(self.numWhat(3), 0.0, 0.0)
		else:
			return None

	# ----------------------------------------------------------------------
	def bodyY(self):
		"""Y-vector of body"""
		if   self.tag == "RPP":
			return bmath.Vector(0.0, self.numWhat(4)-self.numWhat(3), 0.0)
		elif self.tag in ("BOX", "WED"):
			return bmath.Vector(	self.numWhat(7),
						self.numWhat(8),
						self.numWhat(9))
		elif self.tag == "REC":
			return bmath.Vector(	self.numWhat(10),
						self.numWhat(11),
						self.numWhat(12))
		elif self.tag == "XEC":
			return bmath.Vector(0.0, 0.0, self.numWhat(4))
		elif self.tag == "YEC":
			return bmath.Vector(self.numWhat(4), 0.0, 0.0)
		elif self.tag == "ZEC":
			return bmath.Vector(0.0, self.numWhat(4), 0.0)
		else:
			return None

	# ----------------------------------------------------------------------
	def bodyZ(self):
		"""Z-vector of body"""
		if   self.tag == "RPP":
			return bmath.Vector(0.0, 0.0, self.numWhat(6)-self.numWhat(5))
		elif self.tag in ("BOX", "WED"):
			return bmath.Vector(	self.numWhat(10),
						self.numWhat(11),
						self.numWhat(12))
		elif self.tag in ("RCC", "TRC", "REC"):
			return bmath.Vector(	self.numWhat(4),
						self.numWhat(5),
						self.numWhat(6))
		elif self._type == Card.OBJECT:
			return bmath.Vector(	self.numWhat(7),
						self.numWhat(8),
						self.numWhat(9))
		else:
			return None

	# ----------------------------------------------------------------------
	def bodyR(self):
		"""Radius (or x-radius) of body"""
		if   self.tag == "SPH":
			return self.numWhat(4)
		elif self.tag == "RCC":
			return self.numWhat(7)
		elif self.tag in ("XCC", "YCC", "ZCC"):
			return self.numWhat(3)
		elif self.tag == "XEC":
			return bmath.Vector(0.0, self.numWhat(3), self.numWhat(4))
		elif self.tag == "YEC":
			return bmath.Vector(self.numWhat(4), 0.0, self.numWhat(3))
		elif self.tag == "ZEC":
			return bmath.Vector(self.numWhat(3), self.numWhat(4), 0.0)
		else:
			return None

	# ----------------------------------------------------------------------
	# Add a Zone to CELL/REGION
	# ----------------------------------------------------------------------
	def addZone(self, zone):
		"""Add zone to card"""
		zone = str(zone)
		if self._text.strip():
			self._text += f"\n| {zone}"
		else:
			self._text = zone
		self.modified()

	# ----------------------------------------------------------------------
	# Set a list of zones as expression
	# ... removing empty zones
	# ----------------------------------------------------------------------
	def setZones(self, zones):
		if zones:
			self.setText("\n| ".join(
				[x.strip() for x in zones if x.strip()]))
		else:
			self.setText("")
