# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   23-Nov-2018

import os
import re
import log
import tool

#===============================================================================
# A container class to keep the programs
#===============================================================================
class DeckConfig:
	SECTION     = ""		# Configuration section name
	version     = ""		# program version
	released    = ""
	directory   = ""		# installation directory
	config      = None		# Configuration file
	CORE_PAT    = re.compile(r"^core\.?\d*$")
	DEFAULT_PRG = ""		# default running program script
	DEFAULT_EXE = ""		# default executable
	EXT         = ".inp"		# default input extension
	EXTENSIONS  = [EXT]		# list of alternative extensions
	programs    = { "prg"        : "",
			"exe"        : "",
			"usrbin2dvh" : "",
			"vtk2usrbin" : "",
			"usrbin2rtdose.py" : ""}

	#-------------------------------------------------------------------------------
	@staticmethod
	def getInt(section, name, default):
		try:
			return DeckConfig.config.getint(section, name)
		except:
			DeckConfig.config.set(section, name, str(default))
			return default

	#-------------------------------------------------------------------------------
	@staticmethod
	def getFloat(section, name, default):
		try:
			return DeckConfig.config.getfloat(section, name)
		except:
			DeckConfig.config.set(section, name, str(default))
			return default

	#-------------------------------------------------------------------------------
	@staticmethod
	def getStr(section, name, default):
		try:
			return DeckConfig.config.get(section, name)
		except:
			DeckConfig.config.set(section, name, str(default))
			return default

	#-------------------------------------------------------------------------------
	@staticmethod
	def getBool(section, name, default):
		try:
			return DeckConfig.config.getboolean(section, name)
		except:
			DeckConfig.config.set(section, name, str(bool(default)))
			return default

	#-------------------------------------------------------------------------------
	@staticmethod
	def set(section, name, value):
		DeckConfig.config.set(section, name, str(value))
	setStr   = set
	setInt   = set
	setFloat = set

	#-------------------------------------------------------------------------------
	@staticmethod
	def setBool(section, name, value):
		DeckConfig.config.set(section, name, str(bool(value)))

	#=======================================================================
	# Configuration related methods
	#=======================================================================
	@staticmethod
	def findVersion(directory=None):
		pass

	#-----------------------------------------------------------------------
	# Directories
	#-----------------------------------------------------------------------
	@classmethod
	def binDir(cls):	return os.path.join(cls.directory,"bin")
	@classmethod
	def dataDir(cls):	return os.path.join(cls.directory,"data")
	@classmethod
	def documentDir(cls):	return os.path.join(cls.directory,"doc")
	@classmethod
	def includeDir(cls):	return os.path.join(cls.directory,"include")
	@classmethod
	def checkDir(cls):
		if os.path.isdir(cls.directory):
			return True
		else:
			log.error(f"'{cls.directory}' directory is not found or not set")
			return False

	#-----------------------------------------------------------------------
	@classmethod
	def setDir(cls, path=""):
		cls.directory = path
		if cls.directory == "":
			path = tool.findExe(cls.DEFAULT_EXE)
			if path:
				if os.path.basename(path)=="bin":
					cls.directory = os.path.dirname(path)
				else:
					cls.directory = path
		else:
			# Check if the bin was provided instead of the master dir
			exefile = os.path.join(cls.directory, cls.DEFAULT_EXE)
			if tool.isexecutable(exefile):
				if os.path.basename(cls.directory)=="bin":
					cls.directory = os.path.dirname(cls.directory)

	#-----------------------------------------------------------------------
	# Set command
	#-----------------------------------------------------------------------
	@classmethod
	def setCmd(cls, cmd, val):
		val = val.strip()
		if not val:
			cls.programs[cmd] = ""
			return
		path = os.path.dirname(os.path.abspath(val))
		if not path or \
		   path==cls.directory or \
		   path==os.path.join(cls.directory, "bin"):
			cls.programs[cmd] = os.path.basename(val)
		else:
			cls.programs[cmd] = val

	#-----------------------------------------------------------------------
	# @return any simulation command
	#-----------------------------------------------------------------------
	@classmethod
	def command(cls, prg):
		if prg == "prg":
			prg = cls.DEFAULT_PRG
		elif prg == "exe":
			prg = cls.DEFAULT_EXE

		cmd = cls.programs.get(prg, "")
		if not cmd: cmd = prg
		# directory do not exist (error)
		#if not cls.checkDir():
		#	return cmd

		# check if cmd it is executable or absolute path
		if tool.isexecutable(cmd) or cmd[0] == os.path.sep:
			return cmd

		# directory ending in "bin"
		if os.path.basename(cls.directory) == "bin":
			# check "directory[/bin]/cmd"
			newcmd = os.path.join(cls.directory, cmd)
			if tool.isexecutable(newcmd):
				return newcmd
		else:
			# check "directory/bin/cmd"
			newcmd = os.path.join(cls.directory,"bin",cmd)
			if tool.isexecutable(newcmd):
				return newcmd

			# check "directory/cmd"
			newcmd = os.path.join(cls.directory,cmd)
			if tool.isexecutable(newcmd):
				return newcmd

		# check flair directory
		newcmd = os.path.join(DeckConfig.directory,cmd)
		if tool.isexecutable(newcmd):
			return newcmd

		return cmd

	#-----------------------------------------------------------------------
	# @return default submission program/script, executable
	#-----------------------------------------------------------------------
	@classmethod
	def program(cls):
		return cls.command("prg")

	#-----------------------------------------------------------------------
	# @return default executable
	#-----------------------------------------------------------------------
	@classmethod
	def executable(cls):
		return cls.command("exe")

	#-----------------------------------------------------------------------
	@classmethod
	def isDefaultExecutable(cls):
                return cls.programs.get("exe","") == cls.DEFAULT_EXE

	#-----------------------------------------------------------------------
	# Load from config
	#-----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		pass

	#-----------------------------------------------------------------------
	# Save to config
	#-----------------------------------------------------------------------
	@staticmethod
	def saveConfig(config):
		pass
