# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-May-2004

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

__all__ = [ "Input", "Card", "CardInfo" ]

import os
import re
import time
import hashlib
import traceback

from collections import OrderedDict, defaultdict

import log
import tool
import bmath
import Global
import Evaluation

from .config import DeckConfig
from .card import Card, CardList, CardInfo

#===============================================================================
# Input class
#===============================================================================
class Input(CardList):
	MODE = ""		# no name

	# Preprocessor cards that increase or decrease the indent
	# FIXME INDENT_INC/DEC move to Card
	_INDENT_INC    = {"#if", "#ifdef", "#ifndef", "#elif", "#else", "#include"}
	_INDENT_DEC    = {"#elif", "#else", "#endif", "#endinclude"}
	# XXX XXX XXX
	_PREPRO_BLOCK  = {"#if", "#ifdef", "#ifndef", "#elif", "#else", "#endif"}

	config  = DeckConfig()			# configuration information

	# transformation values
	zero     = 1.0E-10			# accuracy for zero
	infinite = 1.0E10			# infinite distance
	_useQUA  = True				# enable the use of QUA on transformations
	_useBOX  = False			# enable the use of BOX/ARB/WED

	# ----------------------------------------------------------------------
	def __init__(self):
		super().__init__()
		# Input file information
		self.input  = self		# self reference for all cards
		self.mode   = Input.MODE
		self.inputs = [self]

		# Cards
		#self.cardlist = []		# Actual list of cards
		# FIXME can be converted to defaultdict(list)
		self.cards    = {}		# dictionary for easier accessing the cardList

		# cache card lists
		self.cache = {}
		self.modified()
		self.localDict  = Evaluation.LocalDict(self)
		self._localDict = {			# base local dict to be copied on
				"frame" : 0,		# camera frame number
				}
		self.localDict.update(self._localDict)	# ... each update
		self.globalDict = None

		# I/O variables
		self._files    = []		# File handles used during I/O
		self._lineNo   = []		# file line number
		self._type     = []		# card opened file type
		self._comment  = []		# Comment preceding card
		self._md5      = {}		# dictionary of md5
		self._md5cur   = None		# current md5
		self._backup   = True
		self._saveline = None		# save line during reading

	# ----------------------------------------------------------------------
	# Pickle information to copy
	# ----------------------------------------------------------------------
	def __getstate__(self):
		# return the basic information without the unpicklable entries
		state = {
			"mode"     : self.mode,
			"cardlist" : self.cardlist,
		}
		return state

	# ----------------------------------------------------------------------
	# UnPickle information to restore
	# ----------------------------------------------------------------------
	def __setstate__(self, state):
		self.__init__()
		self.__dict__.update(state)
		for card in self.cardlist:
			try:
				self.cards[card.tag].append(card)
			except KeyError:
				self.cards[card.tag] = [card]
		self.renumber()

	# ----------------------------------------------------------------------
	def __call__(self):
		return self._children

	# ----------------------------------------------------------------------
	@property
	def cardlist(self):
		return self._children

	@cardlist.setter
	def cardlist(self, val):
		assert False

	# ----------------------------------------------------------------------
	def __getitem__(self, item):
		return self.cards.get(item,[])

	# ----------------------------------------------------------------------
	# @return input name
	# ----------------------------------------------------------------------
	#def name(self):
	#	return os.path.splitext(os.path.basename(self.filename))[0]

	# ----------------------------------------------------------------------
	# @return input extension (can be different than default)
	# ----------------------------------------------------------------------
	#def ext(self):
	#	return os.path.splitext(self.filename)[1]

	# ----------------------------------------------------------------------
	# @return input extension (can be different than default)
	# ----------------------------------------------------------------------
	@classmethod
	def EXT(cls):
		return cls.config.EXT

	# ----------------------------------------------------------------------
	def md5(self, filename):
		return self._md5[filename].hexdigest()

	# ----------------------------------------------------------------------
	# set last time modified
	# ----------------------------------------------------------------------
	def modified(self):
		self._modifiedTime = time.time()

	# ----------------------------------------------------------------------
	def modifiedTime(self):
		"""return modified time"""
		return self._modifiedTime

	# ----------------------------------------------------------------------
	# Check if a newer input file exists than current buffer
	# ----------------------------------------------------------------------
	def checkFileTimestamp(self, filename):
		# FIXME Scan all #include cards
		for fn in self.filenames(filename):
			try:
				if os.path.getmtime(fn) > self._modifiedTime:
					return True
			except:
				pass
		return False

	# ----------------------------------------------------------------------
	# Set as input/modified time the latest time from files
	# ----------------------------------------------------------------------
	#def setFileTime(self):
	#	for filename in self.filenames():
	#		try:
	#			mtime = os.path.getmtime(filename)
	#			self._modifiedTime = max(self._modifiedTime, mtime)
	#		except OSError:
	#			pass

	# ----------------------------------------------------------------------
	def addInput(self, inp):
		self.inputs.append(inp)

	# ----------------------------------------------------------------------
	# @return a clone of current input
	# ----------------------------------------------------------------------
	def clone(self):
		input_ = self.__class__()
		input_.mode = self.mode
		for card in self.cardlist:
			input_.add(card.clone())
		input_.renumber()

		# Clone child
		for inp in self.inputs[1:]:
			input_.inputs.append(inp.clone())
		return input_

	# ----------------------------------------------------------------------
	# Override defines with defines provided as a list
	# ----------------------------------------------------------------------
	def enableDefines(self, defines=None):
		if defines is None: return
		# if there are defines and we unselect all
		# then the defines will be empty and don't work!!
		# Enable/disable #define cards
		for card in self["#define"]:
			if card._indent != 0: continue
			# Search if it exists on defines
			name = card.name()
			for n,v in defines:
				if n == name:
					card.setEnable(True)
					if v != "": card.setWhat(1,v)
					break
			else:
				card.setEnable(False)

	# ----------------------------------------------------------------------
	# extend self input cardlist with another input cardlist
	# ----------------------------------------------------------------------
	def extend(self, inp):
		# Keep a list of include inputs
		# For the moment embed the input in the cardlist
		for card in inp.cardlist:
			card.input = self
		self.cardlist.extend(inp.cardlist)
		for tag,cards in inp.cards.items():
			try:
				self.cards[tag].extend(cards)
			except KeyError:
				self.cards[tag] = cards[:]

	#-----------------------------------------------------------------------
	# Return input filenames
	#-----------------------------------------------------------------------
	def filenames(self, filename):
		files = [filename]
		files.extend([card.name() for card in self["#include"]])
		return files

	# ----------------------------------------------------------------------
	# Create bare minimum input file
	# ----------------------------------------------------------------------
	def minimumInput(self):
		pass

	#-----------------------------------------------------------------------
	# #include card has been modified take appropriate action
	# @return	True if #include has modified the input
	#		False if nothing happened
	# FIXME Should be moved to InputPage with proper undo/redo
	#
	# FIXME it doesn't work
	#-----------------------------------------------------------------------
	def include(self, includecard):
		return True

	#-----------------------------------------------------------------------
	# Cards
	#-----------------------------------------------------------------------
	# Return an extend list with all cards including the ones from the voxel
	# ----------------------------------------------------------------------
	def allcards(self):
		return self.cardlist
		# FIXME make with generators
		#for card in self.cardlist:
		#	if card.tag == "#include":
		#		# call inputs allcards
		#	else:
		#		yield card

	# ----------------------------------------------------------------------
	# Return an ordered dictionary of cards of a specific tag sorted with _pos
	# @param tag    string tag
	#               list of tags
	#               Card.type(), BODY, SOLID, VOLUME...
	# @param which  which cards to show
	#        Possibilities on which:
	#             0: enabled and active (requires a call to pre-process before)
	#             1: enabled
	#             2: all
	#
	#            key will be the _pos, to accepted duplicate cards
	#            -1: enabled and active
	#            -2: enabled
	#            -3: all
	#-----------------------------------------------------------------------
	def cardsSorted(self, tag, which=0):
		if isinstance(tag,int):
			# Special type of tags
			if tag == Card.BODY:
				tag = CardInfo.bodies[self.MODE]
			elif tag == Card.SOLID:
				try:
					tag = CardInfo.typedef[self.MODE]["solid"]
				except KeyError:
					return {}
			elif tag == Card.REGION:	# Card.CELL
				tag = CardInfo.volumes[self.MODE]
			elif tag == Card.MATERIALS:
				tag = CardInfo.materials[self.MODE]
			elif tag == Card.ROTATION:
				tag = CardInfo.rotations[self.MODE]
			else:
				raise Exception(f"Unknown group tag {tag} provided")

		# sort the list for faster processing
		if isinstance(tag,str):
			cards = self[tag][:]
		else:
			cards = []
			for t in tag:
				cards.extend(self[t])

		Card.sort(cards)

		# Sorted by name
		if which == 0:
			return OrderedDict([(x.name(),x) for x in cards if x.notIgnore()])
		elif which == 1:
			return OrderedDict([(x.name(),x) for x in cards if x.enable])
		elif which == 2:
			return OrderedDict([(x.name(),x) for x in cards])

		# Sorted by position in input
		elif which == -1:
			return OrderedDict([(x.pos(),x) for x in cards if x.notIgnore()])
		elif which == -2:
			return OrderedDict([(x.pos(),x) for x in cards if x.enable])
		elif which == -3:
			return OrderedDict([(x.pos(),x) for x in cards])

		raise Exception("Unknown which parameter")

	#-----------------------------------------------------------------------
	# Cache lists when they are requested
	#-----------------------------------------------------------------------
	def cardsCache(self, tag, which=0):
		try:
			return self.cache[tag]
		except KeyError:
			if which>=0:
				self.cache[tag] = cards = self.cardsSorted(tag, which)
				return cards
			else:
				# Print do not cache cards sorted by position!!!
				if Global.developer:
					log.error("Calling input.cardsCache with indexing by position (which<0)")
					log.errorCont("Use .cardsSorted() instead")
					traceback.print_stack()
				return self.cardsSorted(tag, which)

	#-----------------------------------------------------------------------
	def clearCache(self, tag=None):
		self.clearUnits()
		if tag is not None:
			try: del self.cache[tag]
			except KeyError: pass
		else:
			self.cache.clear()

	#-----------------------------------------------------------------------
	# Find best position from the sorting order in the file
	# search which card is closer to one we've asked
	#-----------------------------------------------------------------------
	def bestPosition(self, tag):
		ci  = CardInfo.get(self.mode, tag)
		if ci.tag[0] == "#": return None
		cards = [x for x in self[tag] if x.notIgnore()]
		if cards:
			# Similar cards exists then add it after the last one
			pos = max([x._pos for x in cards]) + 1
			# but with indent=0
			while pos<len(self.cardlist) and \
			      (self.cardlist[pos].indent()>0 or self.cardlist[pos].tag == "#endif"):
				pos += 1
			return pos

		cur = ci.order
		mindist = 10000
		mintag  = ""
		mincl   = []
		for c in CardInfo.cards(self.mode):
			cards = [x for x in self[c.tag] if x.notIgnore()]
			if cards:
				d = cur - c.order
				if abs(d)<abs(mindist):
					mindist = d
					mintag  = c.tag
					mincl   = cards

		if mintag=="" or not mincl:
			pos = 0
		elif mindist>=0:
			pos = max([x._pos for x in mincl])+1
		else:
			pos = min([x._pos for x in mincl])

		# but with indent=0
		while pos<len(self.cardlist) and \
		      (self.cardlist[pos].indent()>0 or self.cardlist[pos].tag == "#endif"):
			pos += 1

		# skip backwards any START, STOP card
		if pos<len(self.cardlist)-1:
			while pos>0 and self.cardlist[pos].tag in ("START","STOP"):
				pos -=1

		return pos

	#-----------------------------------------------------------------------
	# Delete all cards
	#-----------------------------------------------------------------------
	def delAll(self):
		del self.cardlist[:]
		self.cards.clear()
		self.modified()

	#-----------------------------------------------------------------------
	# Delete card by position
	#-----------------------------------------------------------------------
	def delCard(self, pos, renumber=True):
		card = self.cardlist[pos]
		card.input = None
		card._pos  = -1
		tag = card.tag

		# Delete from the main list
		del self.cardlist[pos]
		if renumber: self.renumber(pos)

		# Find tag list
		taglist = self.cards[tag]
		taglist.remove(card)
		if not taglist: del self.cards[tag]
		self.modified()
	delete = delCard

	#-----------------------------------------------------------------------
	# Delete all cards with a specific tag
	#-----------------------------------------------------------------------
	def delTag(self, tag, renumber=True):
		try:
			for card in self.cards[tag]:
				card._pos  = -1
				card.input = None
				self.cardlist.remove(card)
			del self.cards[tag]
		except KeyError:
			return

		if renumber: self.renumber()
		self.modified()

	#-----------------------------------------------------------------------
	# Change card tag
	#-----------------------------------------------------------------------
	def changeTag(self, card, newtag):
		# Remove from previous list
		try:
			self.cards[card.tag].remove(card)
		except ValueError:
			pass

		# Check for a different type of card
		mode,tag = Card.splitTag(newtag, card.MODE)
		if mode == card.MODE:
			# change the tag
			card.setTag(tag)
			card.trimWhats()
		else:
			# create a new card
			assert False

		# Add to new list
		try:
			taglist = self.cards[card.tag]
		except:
			taglist = []
			self.cards[card.tag] = taglist
		taglist.append(card)
		self.modified()

	#-----------------------------------------------------------------------
	# Changing all cards tag from old to new
	#-----------------------------------------------------------------------
	def changeAllTags(self, old, new, show=True):
		if show:
			log.say(f"Changing card(s) '{old}' to '{new}'")
		try:
			cl = self.cards[old]
			self.cards[new] = cl
			del self.cards[old]
			for card in cl:
				card.setTag(new)
				card.trimWhats()
		except KeyError:
			pass
		self.modified()

	#-----------------------------------------------------------------------
	# Move card (src) at position (dest)
	#-----------------------------------------------------------------------
	def moveCard(self, src, dest, renumber=True):
		card = self.cardlist[src]
		del self.cardlist[src]
		if dest >= src: dest -= 1
		self.cardlist.insert(dest, card)
		if renumber:
			self.renumber(min(dest, src), max(dest,src)+1)
		self.modified()

	#-----------------------------------------------------------------------
	# Replace card at position (pos) with card
	#-----------------------------------------------------------------------
	def replaceCard(self, pos, card):
		"""replace card at position (pos) with card"""
		# remember old card
		old = self.cardlist[pos]

		# replace in the cardlist
		self.cardlist[pos] = card

		# remove old from the taglist
		taglist = self.cards[old.tag]
		taglist.remove(old)
		if not taglist: del self.cards[old.tag]

		# add the new to the taglist
		try:
			taglist = self.cards[card.tag]
			taglist.append(card)
		except KeyError:
			taglist = [card]
			self.cards[card.tag] = taglist
		card._pos  = pos
		card.input = self
		self.modified()

		return old

	#=======================================================================
	# Transform a body card, using a 4x4 transformation matrix
	#=======================================================================
	@staticmethod
	def _round0(num):
		if abs(num)<Input.zero: return 0.0
		return num

	#-----------------------------------------------------------------------
	def transformBody(self, card, matrix):
		tag = card.tag

		if tag=="SPH":
			center = matrix * card.bodyP()
			card.setWhat(1, Input._round0(center[0]))
			card.setWhat(2, Input._round0(center[1]))
			card.setWhat(3, Input._round0(center[2]))
		elif tag in ("XCC", "YCC", "ZCC"):
			self._transformInfCyl(card, matrix)
		elif tag in ("RCC", "TRC"):
			# XXX MM comment: add a check to switch to
			# infinite cylinders if it's the case?
			center = matrix * card.bodyP()
			card.setWhat(1, Input._round0(center[0]))
			card.setWhat(2, Input._round0(center[1]))
			card.setWhat(3, Input._round0(center[2]))

			length = matrix.multNoTranslation(card.bodyZ())
			card.setWhat(4, Input._round0(length[0]))
			card.setWhat(5, Input._round0(length[1]))
			card.setWhat(6, Input._round0(length[2]))
		elif tag in ("XYP", "XZP", "YZP", "PLA"):
			self._transformPlane(card, matrix)
		elif tag in ("RPP","BOX"):
			self._transformBox(card, matrix)
		elif tag=="REC":
			self._transformREC(card, matrix)
		elif tag in ("WED", "RAW"):
			self._transformWED(card, matrix)
		elif tag=="ELL":
			focus1 = matrix * card.bodyP()
			focus2 = matrix * card.bodyP2()
			card.setWhat(1, Input._round0(focus1[0]))
			card.setWhat(2, Input._round0(focus1[1]))
			card.setWhat(3, Input._round0(focus1[2]))
			card.setWhat(4, Input._round0(focus2[0]))
			card.setWhat(5, Input._round0(focus2[1]))
			card.setWhat(6, Input._round0(focus2[2]))
		elif tag in ("XEC", "YEC", "ZEC"):
			self._transformInfEllCyl(card, matrix)
		elif tag=="ARB":
			for i in range(1,9):
				point = matrix * card.bodyPn(i)
				n = 3*i-2
				card.setWhat(n  , Input._round0(point[0]))
				card.setWhat(n+1, Input._round0(point[1]))
				card.setWhat(n+2, Input._round0(point[2]))
		elif tag=="QUA":
			self._transformQUA(card, matrix)
		else:
			log.error(f"Invalid card {card.tag} for transformation")
			return True

		self.modified()
		return False

	# ----------------------------------------------------------------------
	def _transformPlane(self, card, matrix):
		point  = matrix * card.bodyP()
		normal = matrix.multNoTranslation(card.bodyN())
		dire   = normal.direction(Input.zero)
		if dire == "X":
			self.changeTag(card, "YZP")
			card.setWhat(1, Input._round0(point[0]))
		elif dire == "Y":
			self.changeTag(card, "XZP")
			card.setWhat(1, Input._round0(point[1]))
		elif dire == "Z":
			self.changeTag(card, "XYP")
			card.setWhat(1, Input._round0(point[2]))
		else:
			self.changeTag(card, "PLA")
			card.setWhat(1, Input._round0(normal[0]))
			card.setWhat(2, Input._round0(normal[1]))
			card.setWhat(3, Input._round0(normal[2]))
			card.setWhat(4, Input._round0(point[0]))
			card.setWhat(5, Input._round0(point[1]))
			card.setWhat(6, Input._round0(point[2]))

	# ----------------------------------------------------------------------
	def _transformInfCyl(self, card, matrix):
		point     = matrix * card.bodyP()
		direction = matrix.multNoTranslation(card.bodyN())
		# XXX no scaling for the moment
		radius = card.numWhat(3)
		dirCheck = direction.direction(Input.zero)

		if   dirCheck in ("X","-X"):
			self.changeTag(card, "XCC")
			card.setWhat(1, Input._round0(point[1]))
			card.setWhat(2, Input._round0(point[2]))

		elif dirCheck in ("Y","-Y"):
			self.changeTag(card, "YCC")
			card.setWhat(1, Input._round0(point[2]))
			card.setWhat(2, Input._round0(point[0]))

		elif dirCheck in ("Z","-Z"):
			self.changeTag(card, "ZCC")
			card.setWhat(1, Input._round0(point[0]))
			card.setWhat(2, Input._round0(point[1]))

		# XXX transform to old version
		elif Input._useQUA:
			# convert card to "QUA" and then transform it
			tag = card.tag
			p = card.bodyP()
			self.changeTag(card, "QUA")
			card.setWhat(1, (1.0, 0.0)[tag=="XCC"])
			card.setWhat(2, (1.0, 0.0)[tag=="YCC"])
			card.setWhat(3, (1.0, 0.0)[tag=="ZCC"])
			card.setWhat(4, 0.0)
			card.setWhat(5, 0.0)
			card.setWhat(6, 0.0)
			card.setWhat(7, -2.0*p.x)
			card.setWhat(8, -2.0*p.y)
			card.setWhat(9, -2.0*p.z)
			card.setWhat(10, -radius**2 + p.length2())
			self._transformQUA(card, matrix)
		else:
			point -= direction  * Input.infinite
			direction  = (2.0 * Input.infinite) * direction
			self.changeTag(card, "RCC")
			card.setWhat(1, Input._round0(point[0]))
			card.setWhat(2, Input._round0(point[1]))
			card.setWhat(3, Input._round0(point[2]))
			card.setWhat(4, Input._round0(direction[0]))
			card.setWhat(5, Input._round0(direction[1]))
			card.setWhat(6, Input._round0(direction[2]))
			card.setWhat(7, Input._round0(radius))

	# ----------------------------------------------------------------------
	def _transformWED(self, card, matrix):
		point = matrix * card.bodyP()
		X = matrix.multNoTranslation(card.bodyX())
		Y = matrix.multNoTranslation(card.bodyY())
		Z = matrix.multNoTranslation(card.bodyZ())
		card.setWhat( 1, Input._round0(point[0]))
		card.setWhat( 2, Input._round0(point[1]))
		card.setWhat( 3, Input._round0(point[2]))
		card.setWhat( 4, Input._round0(X[0]))
		card.setWhat( 5, Input._round0(X[1]))
		card.setWhat( 6, Input._round0(X[2]))
		card.setWhat( 7, Input._round0(Y[0]))
		card.setWhat( 8, Input._round0(Y[1]))
		card.setWhat( 9, Input._round0(Y[2]))
		card.setWhat(10, Input._round0(Z[0]))
		card.setWhat(11, Input._round0(Z[1]))
		card.setWhat(12, Input._round0(Z[2]))

	# ----------------------------------------------------------------------
	def _transformREC(self, card, matrix):
		point = matrix * card.bodyP()
		X = matrix.multNoTranslation(card.bodyX())
		Y = matrix.multNoTranslation(card.bodyY())
		Z = matrix.multNoTranslation(card.bodyZ())
		card.setWhat( 1, Input._round0(point[0]))
		card.setWhat( 2, Input._round0(point[1]))
		card.setWhat( 3, Input._round0(point[2]))
		card.setWhat( 4, Input._round0(Z[0]))
		card.setWhat( 5, Input._round0(Z[1]))
		card.setWhat( 6, Input._round0(Z[2]))
		card.setWhat( 7, Input._round0(X[0]))
		card.setWhat( 8, Input._round0(X[1]))
		card.setWhat( 9, Input._round0(X[2]))
		card.setWhat(10, Input._round0(Y[0]))
		card.setWhat(11, Input._round0(Y[1]))
		card.setWhat(12, Input._round0(Y[2]))

	# ----------------------------------------------------------------------
	def _transformInfEllCyl(self, card, matrix):
		center = matrix * card.bodyP()
		X      = matrix.multNoTranslation(card.bodyX())
		Y      = matrix.multNoTranslation(card.bodyY())
		Z      = matrix.multNoTranslation(card.bodyN())
		dirX   = X.direction(Input.zero)
		dirY   = Y.direction(Input.zero)
		dirZ   = Z.direction(Input.zero)

		if dirX=="N" or dirY=="N" or dirZ=="N":
			if Input._useQUA:
				# convert card to "QUA" and then transform it
				tag = card.tag
				p   = card.bodyP()
				r   = card.bodyX() + card.bodyY()
				rx2 = r.x**2
				ry2 = r.x**2
				rz2 = r.x**2
				self.changeTag(card, "QUA")
				card.setWhat(1, (1.0/rx2, 0.0)[tag=="XEC"])
				card.setWhat(2, (1.0/ry2, 0.0)[tag=="YEC"])
				card.setWhat(3, (1.0/rz2, 0.0)[tag=="ZEC"])
				card.setWhat(4, 0.0)
				card.setWhat(5, 0.0)
				card.setWhat(6, 0.0)
				card.setWhat(7, -2.0*p.x/rx2)
				card.setWhat(8, -2.0*p.y/ry2)
				card.setWhat(9, -2.0*p.z/rz2)
				card.setWhat(10, -1.0 + p.x**2/rx2
					+ p.y**2/ry2 + p.z**2/rz2)
				self._transformQUA(card, matrix)
			else:
				# Transforming in REC
				self.changeTag(card, "REC")

				center -= Z   * Input.infinite
				Z       = 2.0 * Input.infinite * Z

				card.setWhat( 1, Input._round0(center[0]))
				card.setWhat( 2, Input._round0(center[1]))
				card.setWhat( 3, Input._round0(center[2]))
				card.setWhat( 4, Input._round0(Z[0]))
				card.setWhat( 5, Input._round0(Z[1]))
				card.setWhat( 6, Input._round0(Z[2]))
				card.setWhat( 7, Input._round0(X[0]))
				card.setWhat( 8, Input._round0(X[1]))
				card.setWhat( 9, Input._round0(X[2]))
				card.setWhat(10, Input._round0(Y[0]))
				card.setWhat(11, Input._round0(Y[1]))
				card.setWhat(12, Input._round0(Y[2]))

		elif dirZ in ("X","-X"):
			self.changeTag(card, "XEC")
			card.setWhat(1, Input._round0(center[1]))
			card.setWhat(2, Input._round0(center[2]))
			XY = X+Y
			card.setWhat(3, Input._round0(abs(XY.y)))
			card.setWhat(4, Input._round0(abs(XY.z)))

		elif dirZ in ("Y","-Y"):
			self.changeTag(card, "YEC")
			card.setWhat(1, Input._round0(center[2]))
			card.setWhat(2, Input._round0(center[0]))
			XY = X+Y
			card.setWhat(3, Input._round0(abs(XY.z)))
			card.setWhat(4, Input._round0(abs(XY.x)))

		else:
			self.changeTag(card, "ZEC")
			card.setWhat(1, Input._round0(center[0]))
			card.setWhat(2, Input._round0(center[1]))
			XY = X+Y
			card.setWhat(3, Input._round0(abs(XY.x)))
			card.setWhat(4, Input._round0(abs(XY.y)))

	# ----------------------------------------------------------------------
	def _transformBox(self, card, matrix):
		point = matrix * card.bodyP()
		X = matrix.multNoTranslation(card.bodyX())
		Y = matrix.multNoTranslation(card.bodyY())
		Z = matrix.multNoTranslation(card.bodyZ())

		XD = X.direction(Input.zero)
		YD = Y.direction(Input.zero)
		ZD = Z.direction(Input.zero)

		if XD!="N" and YD!="N" and ZD!="N":
			self.changeTag(card, "RPP")
			diagonal = X+Y+Z
			card.setWhat(1, Input._round0(min(point[0], point[0]+diagonal[0])))
			card.setWhat(2, Input._round0(max(point[0], point[0]+diagonal[0])))

			card.setWhat(3, Input._round0(min(point[1], point[1]+diagonal[1])))
			card.setWhat(4, Input._round0(max(point[1], point[1]+diagonal[1])))

			card.setWhat(5, Input._round0(min(point[2], point[2]+diagonal[2])))
			card.setWhat(6, Input._round0(max(point[2], point[2]+diagonal[2])))
		elif Input._useBOX or card.tag == "BOX":
			self.changeTag(card, "BOX")
			card.setWhat( 1, Input._round0(point[0]))
			card.setWhat( 2, Input._round0(point[1]))
			card.setWhat( 3, Input._round0(point[2]))
			card.setWhat( 4, Input._round0(X[0]))
			card.setWhat( 5, Input._round0(X[1]))
			card.setWhat( 6, Input._round0(X[2]))
			card.setWhat( 7, Input._round0(Y[0]))
			card.setWhat( 8, Input._round0(Y[1]))
			card.setWhat( 9, Input._round0(Y[2]))
			card.setWhat(10, Input._round0(Z[0]))
			card.setWhat(11, Input._round0(Z[1]))
			card.setWhat(12, Input._round0(Z[2]))

	# ----------------------------------------------------------------------
	def _transformQUA(self, card, matrix):
		A = bmath.Matrix(4)
		A[0][0] = card.numWhat(1)		# Cxx
		A[1][1] = card.numWhat(2)		# Cyy
		A[2][2] = card.numWhat(3)		# Czz

		A[0][1] = A[1][0] = card.numWhat(4)/2.0	# Cxy
		A[0][2] = A[2][0] = card.numWhat(5)/2.0	# Cxz
		A[1][2] = A[2][1] = card.numWhat(6)/2.0	# Cyz

		A[0][3] = A[3][0] = card.numWhat(7)/2.0	# Cx
		A[1][3] = A[3][1] = card.numWhat(8)/2.0	# Cy
		A[2][3] = A[3][2] = card.numWhat(9)/2.0	# Cz

		A[3][3] = card.numWhat(10)		# C

		minv = matrix.inverse()

		# New = matrix^T * A * matrix
		AN = minv.T() * (A * minv)

		# XXX optimize (convert to other objects if needed)
		card.setWhat(1, Input._round0(AN[0][0]))
		card.setWhat(2, Input._round0(AN[1][1]))
		card.setWhat(3, Input._round0(AN[2][2]))

		card.setWhat(4, Input._round0(AN[0][1]*2.0))
		card.setWhat(5, Input._round0(AN[0][2]*2.0))
		card.setWhat(6, Input._round0(AN[1][2]*2.0))

		card.setWhat(7, Input._round0(AN[0][3]*2.0))
		card.setWhat(8, Input._round0(AN[1][3]*2.0))
		card.setWhat(9, Input._round0(AN[2][3]*2.0))

		card.setWhat(10, Input._round0(AN[3][3]))

	#=======================================================================
	# Utilities
	#=======================================================================
	#-----------------------------------------------------------------------
	# Scan for units
	#-----------------------------------------------------------------------
	def scanUnits(self):
		pass

	#-----------------------------------------------------------------------
	def clearUnits(self):
		pass

	#-----------------------------------------------------------------------
	# Validate all active cards
	#-----------------------------------------------------------------------
	def validate(self):
		for card in self.cardlist:
			if card.ignore(): continue
			card.validate()

	#-----------------------------------------------------------------------
	# Check expressions in input (must be preprocessed before)
	#-----------------------------------------------------------------------
	def checkExpr(self):
		ok = True
		for card in self.cardlist:
			if card[Card.CHKSUM] and not card.checkSum():
				log.warning("Possibly modified expressions in card")
				for i,w in enumerate(card.whats()):
					if isinstance(w,str) and w and w[0]=="=":
						log.warningCont(f"what[{i}] {w}")
				log.warningCont(card)
				ok = False
		return ok

	#-----------------------------------------------------------------------
	# Find all cards with a specific type generation
	#-----------------------------------------------------------------------
	def cardsDefineType(self, target):
		target = ":" + target
		lst = []
		for card in self.cardlist:
			case = card.info.findCase(card)
			if case is not None:
				try:
					if target in card.info.range[case][0]:
						lst.append(card)
				except (TypeError,IndexError):
					pass
		return lst

	#-----------------------------------------------------------------------
	# Change name to all cards
	#-----------------------------------------------------------------------
	def changeName(self, kind, old, new):
		# For bodies use special treatment
		if kind == "body":
			pat = re.compile(rf"(\b{re.escape(old)}\b)")
			for card in self.cardsSorted(Card.REGION,-1).values():
				newtext = pat.sub(new, card.text())
				if newtext != card.text():
					card.setText(newtext)
		else:
			for card in self.cardlist:
				case = card.info.findCase(card)
				lst  = card.info.findType(kind, case, False)
				if isinstance(lst, tuple):
					lst = list(lst)
					lst.pop()	# skip step
				for w in lst:
					what = card.what(w)
					if not what:
						pass
					elif what == old:
						card.setWhat(w, new)
					elif what[0]=="-" and what[1:] == old:
						card.setWhat(w, "-"+new)

				if kind == "var":
					# FIXME
					# change variable in all expression!!!
					pass
		self.modified()

	#-----------------------------------------------------------------------
	# Evaluate expression as floating point
	#-----------------------------------------------------------------------
	def eval(self, expr):
		if not isinstance(expr,float):
			if expr and isinstance(expr,str):
				if expr[0]=="=": expr=expr[1:]
			try:
				return float(expr)
			except ValueError:
				pass
			try:
				return float(eval(str(expr),
						Evaluation._globalDict,
						self.localDict))
			except:
				return 1.0
		return expr

	#-----------------------------------------------------------------------
	# Update base local dict
	#-----------------------------------------------------------------------
	def setLocalDict(self, **kwargs):
		self._localDict.update(kwargs)
		# update also localDict if needed
		#self.localDict[var] = value

	#-----------------------------------------------------------------------
	# Pre-process all cards and set correspondingly the card.active flag
	# stack[] contains the evaluation of the stack nesting
	#      -1: inactive - do not include any subsequent #if..
	#       0: false & active - include substitute #if..
	#       1: true  & active
	#       2: false & active after else
	#       3: true  & active after else
	# @return list of error cards
	# FIXME values do not WORK!
	#-----------------------------------------------------------------------
	def preprocess(self, activeDefines=None, localVars=None):
		self.clearCache()
		defined = defaultdict(int)	# is variable defined
		errors  = []
		self.localDict.clear()
		self.localDict.update(self._localDict)
		#if activeDefines exists and has something
		defs = {}
		if activeDefines is not None:
			useInputDefines = False
			for var,val in activeDefines:
				defined[var] = 1
				if isinstance(val,str) and val:
					if val[0]=="=": val=val[1:]
					try:
						val = float(eval(val,
								Evaluation._globalDict,
								self.localDict))
					except: pass
				try:    self.localDict[var] = float(val)
				except: self.localDict[var] = val
				defs[var] = val
		else:
			# if none or empty list e.g. default run = []
			useInputDefines = True

		stack = [1]
		active13 = True
		for i,card in enumerate(self.cardlist):
			tag    = card.tag
			var    = card.name()
			active = stack[-1]
			if not card.enable:
				card.setActive(active13)
				continue
			active13 = active in (1,3)
			card.setActive(active13)

			if active13 and tag == "#define":
				if useInputDefines or card._indent!=0:
					defined[var] = 1
				else:
					if self.localDict.get(var,"") == "":
						self.localDict[var] = card.evalWhat(1)
					elif var not in defs:
						err = "Duplicate #define with same name is not permitted"
						errors.append((card,err))
						card.addInvalid(err)
						log.error(err)
						log.errorCont(card)

			elif active13 and tag == "#undef" and useInputDefines:
				defined[var] = 0

			elif tag in ("#if","#ifdef"):
				if active13:
					stack.append(defined[var])
				else:
					stack.append(-1)	# ignore this nesting

			elif tag == "#ifndef":
				if active13:
					stack.append(1-defined[var])
				else:
					stack.append(-1)	# ignore this nesting

			elif tag == "#elif":
				if len(stack)<=1:
					err = f"Unexpected #elif, card: {i+1}"
					errors.append((card,err))
					card.addInvalid(err)
					log.error(err)
				elif active==1:
					stack[-1] = -1	# terminate this nesting
				elif active==0:
					stack[-1] = defined[var]
				elif active>1:
					err = f"Misplaced #elif after #else, card: {i+1}"
					errors.append((card,err))
					card.addInvalid(err)
					log.error(err)
				active = stack[-1]
				active13 = active in (1,3)
				card.setActive(active13)

			elif tag == "#else":
				if len(stack)<=1:
					err = f"Unexpected #else, card: {i+1}"
					errors.append((card,err))
					card.addInvalid(err)
					log.error(err)
				elif active>1:
					err = f"Misplaced #else, card: {i+1}"
					errors.append((card,err))
					card.addInvalid(err)
					log.error(err)
				elif active>=0:
					stack[-1] = 3-stack[-1]
				active = stack[-1]
				active13 = active in (1,3)
				card.setActive(active13)

			elif tag == "#endif":
				if len(stack)<=1:
					err = f"Unexpected #endif, card: {i+1}"
					errors.append((card,err))
					card.addInvalid(err)
					log.error(err)
				else:
					stack.pop()
				active = stack[-1]
				active13 = active in (1,3)
				card.setActive(active13)
		return errors

	#-----------------------------------------------------------------------
	# Process card properties
	#-----------------------------------------------------------------------
	def cardProperties(self):
		rotdefi = self.getTransformation()
		self.bodyProperties(rotdefi)
		self.regionProperties() #rotdefi)
		return rotdefi

	#-----------------------------------------------------------------------
	# Return the body properties as a dictionary of the active/enabled
	# bodies
	#-----------------------------------------------------------------------
	def bodyProperties(self, rotdefi):
		pass

	#-----------------------------------------------------------------------
	# Return the region properties as a dictionary of the active/enabled
	# regions
	#-----------------------------------------------------------------------
	def regionProperties(self):
		pass
	cellProperties = regionProperties

	#-----------------------------------------------------------------------
	# Append to card's properties the additional value from tag's
	# @param tag		cards to append the extra property
	# @param proptag	cards from where to get the property
	# @param prop		property name to add in cards
	# @param valueWhat	which value to add as property
	# @param fromWhat,toWhat,stepWhat	range of cards
	# @param sdum		filter cards with the specific sdum
	# @param func		apply a function to value
	#-----------------------------------------------------------------------
	def addCardProperty(self, tag, proptag, prop, valueWhat,
			fromWhat, toWhat=-1, stepWhat=-1, sdum=None, func=None):

		# create a dictionary of cards for faster accessing
		cardlist = list(self.cardsSorted(tag,-1).values())
		if tag=="REGION":
			# Remove continuations of regions
			cardlist = [x for x in cardlist if x.name()!="&"]

		dictionary = {}
		for card in cardlist:
			dictionary[card.name()] = card
			card[prop] = 0.0

		for card in self.cardsSorted(proptag,-1).values():
			if sdum is not None and card.name() != sdum: continue
			try: from_ = dictionary[card.what(fromWhat)]["@n"]
			except: continue
			if from_ is None: continue
			to_  = from_
			step = 1
			if toWhat>=0:
				try:
					toW = card.what(toWhat)
					if toW[0]=='@':
						to_ = len(cardlist)
					else:
						to_  = dictionary[toW]["@n"]
					step = card.intWhat(stepWhat)
				except:
					pass
			if step<=0: step=1

			value = card.numWhat(valueWhat)
			if func: value = func(value)

			for c in cardlist[from_-1:to_:step]:
				c[prop] = value

	#-----------------------------------------------------------------------
	# Dynamic list based on name definition from database as s##:name
	#-----------------------------------------------------------------------
	def typedefDict(self, name):
		dct = {}
		try:
			defaults = CardInfo.defaults[self.MODE][name]
			dct.update(defaults)
		except KeyError:
			pass
		try:
			for tag in CardInfo.typedef[self.MODE][name]:
				dct.update(self.cardsSorted(tag))
		except KeyError:
			pass
			#if Global.developer:
			#	import traceback; traceback.print_stack()
			#	log.error(f"typedef['{self.MODE}']['{name}'] not found!")
		return dct

	#-----------------------------------------------------------------------
	# Return material list
	#-----------------------------------------------------------------------
	def materialList(self, which=0, *kw):
		return list(self.defaultMaterials.keys())

	#-----------------------------------------------------------------------
	# Return material dictionary
	#-----------------------------------------------------------------------
	def materialDict(self):
		matDict = self.defaultMaterials.copy()
		matDict.update(self.cardsSorted(Card.MATERIALS))
		return matDict

	#-----------------------------------------------------------------------
	# Return element list
	#-----------------------------------------------------------------------
	def elementList(self, which=0):
		return [""]

	#-----------------------------------------------------------------------
	# material ZAID description
	# ftype: fraction type 0=atom, 1=mass, -1=original
	# @return 'atom' fraction composition. Normalized to 1.0 if norm is True
	#-----------------------------------------------------------------------
	def materialZAID(self, material, norm=False, ftype=0):
		Global.dump("\n*-* Input.materialZAID: not implemented")
		#traceback.print_stack()
		return []

	#-----------------------------------------------------------------------
	# Rotation/Transformation list
	# ----------------------------------------------------------------------
	def rotationList(self, negative=False):
		return [""] + sorted(self.typedefDict("rotation").keys())

	#-----------------------------------------------------------------------
	# @param idx		rotation to return
	#			Can be prefixed with "-"
	#			None to return a dictionary
	# @param rotdefi	cached dictionary
	# @param inv		return the inverse of the idx matrix.
	#			Accepted even if idx is prefixed with "-"
	# @return rotation matrix from a rot-defi cards or a dictionary with all
	#         transformations
	# FIXME I should combine the rotdefi with the self.cache!
	#-----------------------------------------------------------------------
	def getTransformation(self, idx=None, rotdefi=None, inv=False):
		Global.dump("\n*-* Input.getTransformation: not implemented")
		return bmath.Matrix(4, type=1)

	# ----------------------------------------------------------------------
	# Refresh cards information
	# 1. ?? (renumber??)
	# 2. Reload voxels
	# 3. Properties?
	# ----------------------------------------------------------------------
	def refresh(self):
		pass

	# ----------------------------------------------------------------------
	# Disable all active duplicate defines, since fluka doesn't like them
	# return errors when the duplicate defines had a different value
	# ----------------------------------------------------------------------
	def disableDuplicateDefines(self):
		errors  = []
		defines = {}
		for card in self.cardlist:
			if card.ignore() or card.tag!="#define": continue
			try:
				d = defines[card.name()]
				w = card.numWhat(1)
				if abs(d-w)>abs(d+w+1.0)*1e-10:
					errors.append((card,
						f"{log.ERROR}Card #{card.pos1()}"
						f" Duplicate #define {card.name()}"
						f" found with different values Old={d} New={w}"))
				card.disable()
			except:
				defines[card.name()] = card.numWhat(1)
		return errors

	#=======================================================================
	# I/O
	# methods for Input IO operations
	#=======================================================================
	#-----------------------------------------------------------------------
	# open file and append it to the file handle list
	# @param tag    what tag opened the file (during nesting #include...)
	# @param backup	keep or not backup of the present file
	#-----------------------------------------------------------------------
	def open(self, filename, mode="r", tag=None, backup=None, showerror=True):
		if backup is not None: self._backup = backup
		if filename is None:
			f = None	# don't open anything, for disabled #include
		elif isinstance(filename, str):
			# FIXME use relative wrt to parent or not?
			#       @see Project._saveInputToProject
			# fluka2011 was always wrt to the current directory
			#           instead of the parent input
			# Convert relative filename wrt to parent
			if Global.experimental:
				if filename[0] != os.path.sep and self._files:
					prevdir = os.path.dirname(self._files[-1].name)
					filename = os.path.abspath(os.path.join(prevdir, filename))
			else:
				# FIXME!!! Absolute path versus current directory
				filename = os.path.abspath(filename)

			# check for infinite loop
			if filename in [x.name for x in self._files]:
				log.error(f"File '{filename}' is recursively loaded")
				return None
			if mode=="w" and self._backup:
				tool.backup(filename)
			try:	# open file
				f = open(filename, mode, encoding="UTF-8",errors="replace")
			except:
				if showerror:
					ms = "writing" if mode=="w" else "reading"
					log.error(f"Cannot open file '{filename}' for {ms}\n")
				f = open("/dev/null",mode)
		else:
			f = filename
		self._files.append(f)
		self._lineNo.append(0)
		self._type.append(tag)
		if f is not None:
			self._md5cur = hashlib.md5()
			try:
				self._md5[os.path.abspath(f.name)] = self._md5cur
			except AttributeError:
				# From copy&paste StringIO
				self._md5[None] = self._md5cur
		else:
			self._md5cur = None
		return f

	#-----------------------------------------------------------------------
	# Close topmost file
	#-----------------------------------------------------------------------
	def close(self):
		if self._files:
			f = self._files.pop()
			if f: f.close()
			self._lineNo.pop()
			self._type.pop()
			if self._files:
				try:
					self._md5cur = self._md5[self._files[-1].name]
				except AttributeError:
					# From copy&paste StringIO
					self._md5cur = self._md5[None]
			else:
				self._md5cur = None

	#-----------------------------------------------------------------------
	# Open and read on input file, return None if error, else md5 checksum
	#-----------------------------------------------------------------------
	def load(self, filename):
		#if isinstance(filename,str):
		#	self.filename = filename
		self._md5 = {}
		self._md5cur = None
		if self.open(filename,"r") is None: return False
		if self._load(): return False
		assert not self._files
		self.renumber()
		self.preprocess()
		#self.setFileTime()
		self.modified()
		if self._postLoad(): return False
		return True

	#-----------------------------------------------------------------------
	# Open and write the input to file (specified either by filename,
	# or by file pointer). Optionally use a different geometry
	# file and a different format
	#-----------------------------------------------------------------------
	def save(self, filename, backup=None):
		self._md5 = {}
		self._md5cur = None
		del self._files[:]
		del self._lineNo[:]
		if self.open(filename, "w", backup=backup) is None: return False
		self.preprocess()
		if self._save(): return False
		self.close()
		#self.setFileTime()
		self.modified()
		assert not self._files
		if self._postSave(): return False
		return True

	# ----------------------------------------------------------------------
	# Write an input file to be used for plotting
	# ----------------------------------------------------------------------
	def writePlotgeomInput(self, plotgeom, directory=".", title=""):
		pass

	#-----------------------------------------------------------------------
	# Abstract import (parse) method
	#-----------------------------------------------------------------------
	def _load(self):
		pass

	#-----------------------------------------------------------------------
	# Abstract post import functions
	#-----------------------------------------------------------------------
	def _postLoad(self):
		self.validate()

	#-----------------------------------------------------------------------
	# Abstract export method
	#-----------------------------------------------------------------------
	def _save(self):
		pass

	#-----------------------------------------------------------------------
	# Abstract post export functions
	#-----------------------------------------------------------------------
	def _postSave(self):
		pass

	#-----------------------------------------------------------------------
	# read line from the last opened file
	# @return line.rstrip() or None on end of all files
	#-----------------------------------------------------------------------
	def readline(self):
		if self._saveline:
			line = self._saveline
			self._saveline = None
			return line
		while True:
			try:
				self._lineNo[-1] += 1
				line = self._files[-1].readline()
			except IndexError:
				return None
			except AttributeError:
				line = ""	# from None file
			if not line:
				self.close()
				if not self._files:
					return None
			else:
				if self._md5cur is not None:
					self._md5cur.update(line.encode())
				#print("MD5.READ",self._files[-1].name,
				#	self._md5cur.hexdigest(),line.rstrip())
				return line.rstrip()

	#-----------------------------------------------------------------------
	def saveline(self, line):
		assert self._saveline is None
		self._saveline = line

	#-----------------------------------------------------------------------
	# Write a string to file, calculate as well the md5 sum
	#-----------------------------------------------------------------------
	def write(self, s):
		"""Write string in utf-8 encoding"""
		if self._md5cur is not None:
			self._md5cur.update(s.encode())
		#print("MD5.WRITE",self._files[-1].name, self._md5cur.hexdigest(),s.rstrip())
		self._files[-1].write(s)

	#-----------------------------------------------------------------------
	def writeline(self, s=""):
		"""Write string with new line appended"""
		self.write(f"{s}\n")

	#-----------------------------------------------------------------------
	@classmethod
	def newCard(cls, *args, **kwargs):
		return Card(*args, **kwargs)

#-------------------------------------------------------------------------------
#if __name__ == "__main__":
#	import sys
#	Element.init("db/isotopes.ini")
#
#	inp = Input()
#	inp.verbose = True
#	inp.read(sys.argv[1])
#	inp.regionProperties()
##	log.say(inp.materialList())
#	if len(sys.argv)>=3:
#		#f = open(sys.argv[3],"w")
#		#inp.writeGeometry(f, FlukaCard.FORMAT_FREE, FlukaCard.FORMAT_FREE)
#		#f.close()
#		inp.write(sys.argv[2])
#
#	for card in inp["MATERIAL"]:
#		print()
#		print("-"*50)
#		print(card.sdum())
#		zaid = inp.materialZAID(card, True)
#		print("Compound:")
#		for za,(f,am) in zaid.items():
#			print("\t",za,f)
#		zaid = inp.materialZAID(card)
#		print("Compound:")
#		for za,(f,am) in zaid.items():
#			print("\t",za,f)
#
##	import bz2, pickle
##	f = bz2.BZ2File(sys.argv[2],'w')
##	inp.dump(pickle.Pickler(f))
##	f.close()
#
##	f = bz2.BZ2File(sys.argv[2],'r')
##	inp2 = Input()
##	inp2.load(pickle.Unpickler(f))
##	f.close()
#	#log.say(inp.getTransformation(1))
#	#log.say(inp.getTransformation("rot#001"))
#	#print inp.getTransformation("CDQ4R1")
#	#inp.write(sys.argv[2])
