#!/bin/env python3
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Jun-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import Global
from .card  import CardInfo, Card
from .input import Input

#===============================================================================
# Layout class for the specific input/canvas display
#===============================================================================
class Layout:
	MODE = Card.MODE
	name            = (None,      0)
	tab             = (None,     None)
	newline         = ("\n",     None)
	empty_what      = ("",       "")
	default_what    = ("default", 0)
	ignore_what     = ("ignore",  0)
	reset_what      = ("reset",  -1)
	on_what         = ("On",      1)
	off_what        = ("off",     0)
	false_what      = ("false",   0)
	true_what       = ("true",    1)
	bool_what       = [empty_what, false_what, true_what]
	onoff_what      = [on_what,    off_what]
	onoffempty_what = [empty_what, on_what,  off_what]
	onoffreset_what = [on_what,    off_what, reset_what]
	onoff3_what     = [empty_what, on_what, ("off", -1), ignore_what]
	empty_layout    = [tab]
	BODY_PARENTHESIS= "()"
	FUNCTION        = 999

	# ----------------------------------------------------------------------
	def __init__(self, flair):
		self.flair = flair
		self.setInput(self.flair.project.input)
		self.init()
		# lists that are dynamically linked to an Input File
		self._cache = {}

	# ----------------------------------------------------------------------
	# set input file to use
	# ----------------------------------------------------------------------
	def setInput(self, inp):
		self.input = inp

	# ----------------------------------------------------------------------
	# Automatically detect type from range
	# @return (name, list-of-values)
	# ----------------------------------------------------------------------
	def iType(self, name, rng):
		try:
			if "[" in rng[:2]:
				choices = []
				empty = False
				for c in rng[1+int(rng[1]=="["):-1].split(","):
					if c and c[0]=="@": c=c[1:]
					if ":" in c:
						a,b = map(str.strip, c.split(":"))
						if a=="": empty = True
						choices.append((a,int(b)))
					else:
						c = c.strip()
						if c=="": empty = True
						choices.append(c)
				if not empty: choices.insert(0,"")
				return (name, choices)
			elif rng=="b":
				return (name, Layout.bool_what)
			elif rng=="rotation":
				return (name, self.__class__.rotationList)
			elif rng=="material":
				return (name, self.__class__.materialList)
			elif rng=="region":
				return (name, self.__class__.regionList)
			elif rng=="particle":
				return (name, self.__class__.particleList)
			elif rng.startswith("fn("):
				# expect patterns in fn(Data:*.dat,Include:*.inc)
				filetypes = tuple(x.split(":") for x in rng[3:-1].split(","))
				if not filetypes: filetypes=(("All","*"),)
				return (name, filetypes)
			elif rng in CardInfo.typedef[self.MODE]:
				return (name, lambda s=self,n=rng : s.typedefList(n))
			else:
				return name
		except (TypeError,IndexError):
			return name

	# ----------------------------------------------------------------------
	# @param card   card to return layout for
	# @param info   optional cardinfo to create a static layout
	#               Static layout is created only if the *array*
	#               is not True
	# @return generic layout of card
	# ----------------------------------------------------------------------
	def generic_layout(self, card, info=None):
		#---------------------------------------------------------------
		# name
		#---------------------------------------------------------------
		def iName(i):
			nonlocal info, case
			if isinstance(i,int):
				try:
					label = info.label[case][i]
					if i==0 and (not label or label.lower()=="name"):
						return None
					return label
				except (TypeError,IndexError):
					if i==0: return None
					return f"#{i}"
			else:
				return i

		#---------------------------------------------------------------
		# add to layout 'item'
		#---------------------------------------------------------------
		def add2layout(item):
			nonlocal layout
			if item=="n":
				layout.append(Layout.newline)
			elif item=="t":
				layout.append(Layout.tab)
			elif item==-1:
				layout.append((iName(item),-1))
			elif isinstance(item, int):
				try:
					layout.append((iName(item),
						self.iType(item, info.range[case][item])))
				except:
					#print(f"+++ {info.tag=} {case=} {item=}")
					raise
			else:
				# FIXME needs range also
				layout.append((iName(item), item))

		#---------------------------------------------------------------
		# add the property
		#---------------------------------------------------------------
		def addproperty(name, pinfo):
			nonlocal layout
			layoutTemplate = pinfo.layout if pinfo.layout else range(len(pinfo.variables))
			layout.append(Layout.newline)
			for i in layoutTemplate:
				if i=="n":
					layout.append(Layout.newline)
					continue
				if i=="t":
					layout.append(Layout.tab)
					continue
				#if isinstance(i,str): i = int(i)
				vi = pinfo.variables[i]
				if not vi.type:
					raise Exception(f"vi.name='{vi.name}' "
						f"pinfo='{pinfo}' has no type")
				if len(pinfo.variables)==1 or name==vi.name:
					varname = name
				else:
					varname = f"{name}.{vi.name}"
				if i==0:
					# FIXME completly arbitrary
					label = varname if len(vi.name)<8 else name
				else:
					label = vi.name
				if vi.guidance:
					guidance = vi.guidance
				else:
					guidance = pinfo.meaning
				if vi.range:
					guidance += f"\nRange: {vi.range}"
				if Global.developer:
					if pinfo.assert_:
						guidance += f"\nAssert: {pinfo.assert_}"
				if vi.default:
					guidance += f"\nDefault: {vi.default}"
				candidates = vi.candidates
				vtype = self.iType(varname, vi.type)
				if isinstance(vtype,tuple):
					layout.append((label, vtype, guidance))
				#if vi.type=="b":
				#	candidates = Layout.bool_what
				elif candidates:
					layout.append((label, (varname, candidates), guidance))
				else:
					layout.append((label, varname, guidance))

		#---------------------------------------------------------------
		# Create a layout for the whats
		#---------------------------------------------------------------
		if card is not None:
			#if card.tag=="/SURFACE/Y-SCALE": breakpoint()
			info = card.info
			case = card.case()
			try:
				nwhats = info.nwhats[case]
			except (TypeError, IndexError):
				nwhats = card.nwhats()
		else:
			# In case array is defined created as dynamic one
			if info.array: return self.generic_layout
			case = 0
			try:
				nwhats = info.nwhats[case]
			except (TypeError, IndexError):
				nwhats = 6
		layout = []

		# --- Scan layout ---
		if info.layoutTemplate:
			for x in info.layoutTemplate[case]:
				add2layout(x)
			if not layout: layout.append((None,None))
		else:
			# Positional arguments
			if info.tag[0]=="/":
				if iName(0): layout.append(Layout.tab)
			for i in range(nwhats):
				# FIXME
				if i and i%4==0: layout.append(Layout.newline)
				try:
					vtype = info.range[case][i]
					layout.append((iName(i),self.iType(i,vtype)))
				except (TypeError,IndexError):
					layout.append((iName(i),i))
			# Empty layout?
			if not layout:
				layout.append(Layout.tab)

		# --- Add the variables ---
		for name,pinfo in info.property.items():
			if not pinfo.variables:
				layout.append(Layout.newline)
				layout.append((name,None))
				continue
			if pinfo.array: continue
			addproperty(name, pinfo)

		# --- Add the arrays ---
		if info.array:
			layout.append(Layout.newline)
			layout.append((f"# {name}(s)",
					(f"{name}:0", list(range(101))),
					f"Number of {name} items"))
			for name in info.array:
				pinfo = info.property[name]
				for i in range(1, card.intProperty(f"{name}:0",0) +1):
					addproperty(f"{name}:{i}", pinfo)
		return layout

	# ----------------------------------------------------------------------
	# get layout for card
	# ----------------------------------------------------------------------
	def get(self, card):
		if card.info:
			layout = card.info.layout
			if callable(layout):
				layout = layout(card)
			if layout is not None:
				return layout
		return self.generic_layout(card)

	# ----------------------------------------------------------------------
	# Return what to labels conversion
	# ----------------------------------------------------------------------
	def what2labels(self, layout):
		# What to labels conversion
		# Find fields to highlight
		labels = [""]*6
		for lbl,dat,*_ in layout:
			if isinstance(dat,tuple):
				dat = dat[0]
			elif isinstance(dat,int):
				while dat>=len(labels): labels.append("")
				labels[dat] = lbl
		return labels

	# ----------------------------------------------------------------------
	# Enable/disable caching of lists
	# ----------------------------------------------------------------------
	def cache(self, on):
		self._cache.clear()
		if on:
			self._cache["define"] = self.defineList()

	#-----------------------------------------------------------------------
	def _cacheTypedef(self):
		for name in CardInfo.typedef[self.MODE].keys():
			if name not in self._cache:
				self._cache[name] = self.typedefList(name)

	#-----------------------------------------------------------------------
	# Dynamic list based on name definition from database as s##:name
	#-----------------------------------------------------------------------
	def typedefList(self, name):
		try: return self._cache[name]
		except KeyError: pass
		if self.MODE == self.input.MODE:	# only for the same MODE
			return [""] + sorted(self.input.typedefDict(name).keys())
		else:
			return [""]

	# ----------------------------------------------------------------------
	def particleList(self):
		pass

	# ----------------------------------------------------------------------
	def bodyList(self, inctag=True):
		try: return self._cache["body"]
		except KeyError: pass
		bodies = set()
		if inctag:
			bodies.add("")
			for card in self.input.cardsSorted(Card.BODY).values():
				if card is not None:
					bodies.add(f"{card.name()} [{card.tag}]")
		else:
			for card in self.input.cardsSorted(Card.BODY).values():
				bodies.add(card.name())
		return sorted(list(bodies), key=str.upper)

	# ----------------------------------------------------------------------
	# Return define variable list
	# ----------------------------------------------------------------------
	def defineList(self):
		try: return self._cache["define"]
		except KeyError: pass
		try: defines = {x.name() for x in self.input.cards["#define"]}
		except KeyError: return [""]
		lst = sorted(list(defines), key=str.upper)
		lst.insert(0,"")
		return lst

	# ----------------------------------------------------------------------
	# Return material list, excluding material 'e' if defined
	#-----------------------------------------------------------------------
	def materialList(self, e=None):
		try: return self._cache["material"]
		except KeyError: pass
		lst = [""] + self.input.materialList(1)
		if e:
			try: lst.remove(e)
			except ValueError: pass
		return lst

	#-----------------------------------------------------------------------
	def regionList(self):
		try: return self._cache["region"]
		except KeyError: pass
		lst = [""]
		# Add defined regions
		lst.extend([x.name()
			for x in self.input.cardsSorted(Card.REGION).values()
					if x.name()!="&"])
		return lst

	cellList = regionList

	# ----------------------------------------------------------------------
	# Rotation matrix list
	# ----------------------------------------------------------------------
	def rotationList(self):
		try: return self._cache["rotation"]
		except KeyError: pass
		return self.input.rotationList(True)

	# ----------------------------------------------------------------------
	def setWhat(self, card, what, val):
		self.flair.addUndo(self.flair.setWhatUndo(card, what, val))

	# ----------------------------------------------------------------------
	def setAbsWhat(self, card, what, val):
		self.flair.addUndo(self.flair.setAbsWhatUndo(card, what, val))

	# ----------------------------------------------------------------------
	def setSign(self, card, what, sign):
		self.flair.addUndo(self.flair.setSignUndo(card, what, sign))

	# ----------------------------------------------------------------------
	def setCard(self, card, name, value, default=None):
		self.flair.addUndo(self.flair.setCardUndo(card, name, value, default))

	# ----------------------------------------------------------------------
	def relpath(self, path):
		return self.flair.project.relpath(path)

	# ----------------------------------------------------------------------
	# initialize layout and order cards...
	# ----------------------------------------------------------------------
	def _init_set(self, order, mode):
		# Set ordering and underline
		for i,(t,l) in enumerate(order):
			underline = t.find('&')
			if underline>=0:
				tag = t.replace('&','')
			else:
				tag = t
			ci = CardInfo.get(mode,tag)
			if l is not Layout.generic_layout:
				ci.layout = l
			else:
				ci.layout = l(self,None,info=ci)
			ci.order  = i
			ci.underline = underline

	# ----------------------------------------------------------------------
	def init(self):
		order = [("#&define"  , _P_DEFINE_layout ),
			( "#&undef"   , _P_UNDEF_layout  ),
			( "#inc&lude" , _P_INCLUDE_layout),
			( "#&if"      , _P_IF_layout     ),
			( "#ifdef"    , _P_IFDEF_layout  ),
			( "#ifndef"   , _P_IFNDEF_layout ),
			( "#eli&f"    , _P_ELIF_layout   ),
			( "#&else"    , _P_ELSE_layout   ),
			( "#e&ndif"   , _P_ENDIF_layout  ),
			( "#endinclude", Layout.empty_layout),

			( "point"     , _POINT_layout    ),
			( "arrow"     , _ARROW_layout    ),
			#( "spline"    , _SPLINE_layout   ),
			( "ruler"     , _RULER_layout    ),
			( "mesh"      , _MESH_layout     ),
			( "light"     , _LIGHT_layout    ) ]
		self._init_set(order, Input.MODE)

#================================ Preprocessor =================================
_P_DEFINE_layout = [ (None,0), ("",1)]
_P_UNDEF_layout  = [ (None,(0, Layout.defineList))]
_P_IF_layout     = _P_UNDEF_layout
_P_IFDEF_layout  = _P_IF_layout
_P_IFNDEF_layout = _P_IF_layout
_P_ELIF_layout   = _P_UNDEF_layout
_P_ELSE_layout   = Layout.empty_layout
_P_ENDIF_layout  = Layout.empty_layout

_P_INCLUDE_layout = [	#(None, 0),
			(None, (0, (("All inputs", ("*.flair","*.inp","*.fluka","*.moira")),
				    ("Flair files", "*.flair"),
				    ("Fluka input", ("*.inp","*.fluka")),
				    ("Include", "*.inc"),
				    ("All", "*")))),
		]

#-------------------------------------------------------------------------------
# FIXME TO BE CHANGED with variables and no longer with whats!!!!
#-------------------------------------------------------------------------------

#------------------------------- POINT -----------------------------------------
__ANCHOR_opt  = [ ("none",0),
		  ("C", 1), ("N", 2),  ("NE",3),
		  ("E", 4), ("SE",5),  ("S", 6),
		  ("SW",7), ("W",  8), ("NW",9) ]
__POINT_opt   = [ ("cross",0), ("dot",1), ("square",2), ("x",3), ("diamond",4), ("circle",5) ]
__ARROW_opt   = [ ("head",0), ("tail",1), ("headtail",2), ("line",3) ]
__SPLINE_opt  = [ ("line", 0), ("cardinal",1) ]
__RULER_opt   = [ ("simple",0), ("angle",1) ]
__LIGHT_opt   = [ ("sun",0), ("omni",1), ("spot",2) ]
__FALLOFF_opt = [ ("constant",0), ("1/r",1), ("1/r2",2),
		  ("const+shadow", 3), ("1/r+shadow",4), ("1/r2+shadow",5) ]

def __ANCHOR_type(layout, card, val=None):
	s,a = divmod(card.intWhat(5),100)
	if val is None:
		return a
	else:
		try: aval = int(val)
		except: aval = 0
		layout.setWhat(card, 5, s*100 + aval)

def __SIZE_type(layout, card, val=None):
	s,a = divmod(card.intWhat(5),100)
	if val is None:
		return s
	else:
		try: sval = int(val)
		except: sval = 0
		layout.setWhat(card, 5, sval*100 + a)

_POINT_layout = [ (None, 0),
		("x",	1),
		("y",	2),
		("z",	3),
		Layout.newline,
	("Option",		(4, __POINT_opt)),
		("Size",	__SIZE_type, "Pixel/drawing size of object"),
		("Anchor",	(__ANCHOR_type, __ANCHOR_opt)),
		("Color",	6)]

#------------------------------- ARROW ----------------------------------------
_ARROW_layout = [ (None, 0),
		("x",	1),
		("y",	2),
		("z",	3),
		Layout.newline,
	Layout.tab,
		("dx",	7),
		("dy",	8),
		("dz",	9),
		Layout.newline,
	("Option",		(4, __ARROW_opt)),
		("Size",	__SIZE_type, "Pixel/drawing size of object"),
		("Anchor",	(__ANCHOR_type, __ANCHOR_opt)),
		("Color",	6)]

#------------------------------- RULER ----------------------------------------
_RULER_layout = [ (None, 0),
		("x",	1),
		("y",	2),
		("z",	3),
		Layout.newline,
	Layout.tab,
		("xe",	7),
		("ye",	8),
		("ze",	9),
		Layout.newline,
	Layout.tab,
		("xa",	10),
		("ya",	11),
		("za",	12),
		Layout.newline,
	("Option",		(4, __RULER_opt)),
		("Size",	__SIZE_type, "Pixel/drawing size of object"),
		("Anchor",	(__ANCHOR_type, __ANCHOR_opt)),
		("Color",	6)]

#------------------------------- LIGHT ----------------------------------------
_LIGHT_layout = [ (None, 0),
		("x",	1),
		("y",	2),
		("z",	3),
		Layout.newline,
	Layout.tab,
		("dx",	7),
		("dy",	8),
		("dz",	9),
		Layout.newline,
	("Option",		(4, __LIGHT_opt)),
		("Size",	__SIZE_type, "Pixel/drawing size of object"),
		("Anchor",	(__ANCHOR_type, __ANCHOR_opt)),
		("Color",	6),
		Layout.newline,
	Layout.tab,
		("Power",	10),
		("Falloff",	(11,__FALLOFF_opt)),
		("Specular",	12),
		Layout.newline,
	Layout.tab,
		("Radius",	13),
		("# Samples",	14)]

#------------------------------- MESH ------------------------------------------
_MESH_layout = [ (None, 0),
		("x",	1),
		("y",	2),
		("z",	3),
		Layout.newline,
	("Option",		(4, __POINT_opt)),
		("Size",	__SIZE_type, "Pixel/drawing size of object"),
		("Anchor",	(__ANCHOR_type, __ANCHOR_opt)),
		("Color",	6),
		Layout.newline,
	("File",		(-1, ("STL files","*.stl")))]
