#!/bin/env python3
# -*- coding: utf-8 -*-
#
#
#                 *****   ****
#                **  **     **               **
#                **         **               **
#               **         **
#             *******      **     **** **  ***   ***  **
#               **        **     **  ***    **    ** ****
#              **         **    **    **    **    **** *
#              **         **    **    **   **    ***
#              **        **    **    **    **    **
#             **         **    **    **    **    **
#             **        **     **   ***   **    **
#             **        ** *   **  **** * ** *  **
#             **        ****    **** ***  ****  **
#            **
#         ** **
#         ****                                   (c) CERN
#
#
#           www:      http://cern.ch/flair
#           email:    Vasilis.Vlachoudis@cern.ch
#           authors:  Vasilis Vlachoudis
#
#  About
#  ~~~~~
#  /flair/ n [U,C] natural or instinctive ability to do something well
#             to select or recognize what is best, more useful, etc.
#                                         Oxford Advanced Dictionary
#
#  flair is an advanced user friendly interface for various Monte Carlo
#  programs. Originally it was developed for fluka, where its unique
#  features strongly contributed to the success of both codes fluka and
#  flair. From version 3 the interface was separated from the
#  functionality permitting an easy integration of other simulation
#  packages. Currently is supporting the CERN fluka http://fluka.cern
#  and moira (http://cern.ch/moira) a Geant4 (cern.ch/geant4) based
#  application, permitting the user to run Geant4 simulations using the
#  same interactive experience. The future plans includes the support
#  of MCNP, PHITS...
#
#  Use of Flair must be acknowledged using the following reference:
#
#  V. Vlachoudis, Proc. Int. Conf. on Mathematics, Computational
#  Methods & Reactor Physics (M&C 2009), Saratoga Springs,
#  New York, 2009
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   12-Sep-2013

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import io
import os
import re
import bz2
import sys
import stat
import time
import getopt
import pickle
import fnmatch
import binascii
import traceback
import subprocess
import configparser
import tkinter as tk
from tkinter import messagebox

sys.path.insert(0,os.path.join(os.path.abspath(os.path.dirname(__file__)), "lib"))

# First include to setup global variables
import Global
# Fast scan for developer option and initialize it if necessary
if   "-d" in sys.argv: Global.developer    = True
elif "-D" in sys.argv: Global.developer    = False
if   "-e" in sys.argv: Global.experimental = True
elif "-E" in sys.argv: Global.experimental = False
if   "-p" in sys.argv: Global.profiler     = True
elif "-P" in sys.argv: Global.profiler     = False

import log
import undo
import icons
import inputs
import convert
import tkFlair
import tkExtra
import tkDialogs
import bFileDialog

from deck     import Card, CardInfo
from fluka    import Fluka, FlukaCard, FlukaConfig
from mcnp     import Mcnp
from moira    import Moira, MoiraConfig, Gdml, Gdml2Moira, Gdml2Fluka

if sys.version_info < (3,0):
	sys.stderr.write("FATAL ERROR: flair only works with only with python V3+\n")
	sys.exit(0)

import Manual
import Ribbon
import Project
import Updates
import FlairRibbon

# Plot Engines
import GPPlot
import MPPlot
import Gnuplot

# Importers/Exporters
import Povray
import OpenSCAD

# Frames
import Output
import RunPage
import PetPage
import PlotPage
import DicomPage
import InputPage
import Calculator
import ViewerPage
import CompilePage
import ProjectPage
import MaterialsPage
#import OptimizerPage
import PeriodicTable
import GeometryEditor
from element import Element
from material import Material

#import EmptyPage

PLOT_ENGINES = ["Gnuplot", "Matplotlib"]

_JOB_REFRESHTIME = 1000	# ms
_AUTOSAVE_FIlENAME = "flair_autosave.pickle"

#===============================================================================
root          = None
checkFlukaExe = True
checkMoiraExe = True

#-------------------------------------------------------------------------------
def malloc_trim():
	return
	import ctypes
	ctypes.CDLL('libc.so').malloc_trim(0)

#-------------------------------------------------------------------------------
def garbage_collect(event=None):
	import gc
	gc.set_debug(gc.DEBUG_STATS | gc.DEBUG_COLLECTABLE | gc.DEBUG_UNCOLLECTABLE
			| gc.DEBUG_LEAK)
	log.say(f"gc isenabled={gc.isenabled()}\n")
	log.say(f"gc threshold={gc.get_threshold()}\n")
	log.say(f"gc collect={gc.collect()}\n")
	malloc_trim()

#-------------------------------------------------------------------------------
# trace function calls
#-------------------------------------------------------------------------------
def tracefunc(frame, event, arg, _indent=[0]):
	try:
		classname = frame.f_globals["__name__"]
	except KeyError:
		classname = ""
	if not frame.f_code.co_filename.startswith("/home"): return
	funcname = frame.f_code.co_name
	#if arg:
	#	args = repr(arg)
	#else:
	args = ""
	if event == "call":
		_indent[0] += 1
		print(f"{_indent[0]} {'-'*(_indent[0]*2)}> {classname}.{funcname}({args})")
	elif event == "return":
		#print(f"{_indent[0]} <{'-'*(_indent[0]*2)} {classname}.{funcname}({args}) [exit]")
		_indent[0] -= 1
	return tracefunc
#sys.setprofile(tracefunc)

#===============================================================================
if not Global.experimental:
	_PAGES =  [	#  Class
			Calculator.CalculatorPage,
			CompilePage.CompilePage,
			DicomPage.DicomPage,
			ProjectPage.ProjectPage,
			InputPage.InputPage,
			MaterialsPage.MaterialPage,
			Output.OutputPage,
			PetPage.PetPage,
			PlotPage.PlotListPage,
			RunPage.RunPage,
			PeriodicTable.PeriodicTablePage,
			ViewerPage.ViewerPage
		]
else:
	_PAGES =  [	#  Class
			Calculator.CalculatorPage,
			CompilePage.CompilePage,
			DicomPage.DicomPage,
			ProjectPage.ProjectPage,
			InputPage.InputPage,
			MaterialsPage.MaterialPage,
			Output.OutputPage,
			PetPage.PetPage,
			#OptimizerPage.OptimizerPage,	# <<<
			PlotPage.PlotListPage,
			RunPage.RunPage,
			PeriodicTable.PeriodicTablePage,
			ViewerPage.ViewerPage
			#EmptyPage.EmptyPage
		]
if GeometryEditor.installed():
	_PAGES.append(GeometryEditor.GeometryEditor)
else:
	_PAGES.append(GeometryEditor.GeometryEditorEmpty)

#===============================================================================
# FlairTabRibbonFrame
#===============================================================================
class FlairTabRibbonFrame(Ribbon.TabRibbonFrame):
	# ----------------------------------------------------------------------
	def __init__(self, master, mastertab=None, **kw):
		super().__init__(master, mastertab, **kw)
		self._inundo = False

		self.createNotify()

		# Create bindings for events that can be of general interest
		# in the whole application
		# All private events are handled by their respective page
		toplevel = self.winfo_toplevel()

		# Event generators
		# Order is Widget -> Class -> TopLevel
		# Therefore all bindings are the last resort
		toplevel.event_add("<<Help>>",		"<F1>")
		toplevel.event_add("<<Rename>>",	"<F2>")
		toplevel.event_add("<<Refresh>>",	"<F5>")
		toplevel.event_add("<<Splitter>>",	"<F11>")
		toplevel.event_add("<<Copy>>",		"<Control-Key-c>")
		toplevel.event_add("<<Copy>>",		"<Control-Key-C>")
		toplevel.event_add("<<Clone>>",		"<Control-Key-d>")
		toplevel.event_add("<<CloneSpecial>>",	"<Control-Key-D>")
		toplevel.event_add("<<Output>>",	"<Control-Key-o>")
		toplevel.event_add("<<Output>>",	"<Control-Key-O>")
		toplevel.event_add("<<Hardcopy>>",	"<Control-Key-p>")
		toplevel.event_add("<<Refresh>>",	"<Control-Key-r>")
		toplevel.event_add("<<Refresh>>",	"<Control-Key-R>")
		toplevel.event_add("<<Save>>",		"<Control-Key-s>")
		toplevel.event_add("<<Save>>",		"<Control-Key-S>")
		toplevel.event_add("<<Paste>>",		"<Control-Key-v>")
		toplevel.event_add("<<PasteSpecial>>",	"<Control-Key-V>")
		toplevel.event_add("<<Cut>>",		"<Control-Key-x>")
		toplevel.event_add("<<Cut>>",		"<Control-Key-X>")
		# Correct Ctrl-y behavior
		toplevel.event_delete("<<Paste>>",	"<Control-Key-y>")
		toplevel.event_delete("<<Paste>>",	"<Control-Key-Y>")
		toplevel.event_add("<<Redo>>",		"<Control-Key-y>")
		toplevel.event_add("<<Redo>>",		"<Control-Key-Y>")
		toplevel.event_add("<<Undo>>",		"<Control-Key-z>")
		toplevel.event_add("<<Undo>>",		"<Control-Key-Z>")
		toplevel.event_add("<<MoveLeft>>",	"<Control-Key-Left>")
		toplevel.event_add("<<MoveRight>>",	"<Control-Key-Right>")
		toplevel.event_add("<<MoveUp>>",	"<Control-Key-Up>")
		toplevel.event_add("<<MoveDown>>",	"<Control-Key-Down>")
		toplevel.event_add("<<Execute>>",	"<Control-Key-Return>")
		toplevel.event_add("<<Execute>>",	"<Control-Key-KP_Enter>")

		# Event bindings
		toplevel.bind("<<Clone>>",		self.clone)
		toplevel.bind("<<Copy>>",		self.copy)
		toplevel.bind("<<Cut>>",		self.cut)
		toplevel.bind("<<Help>>",		self.help)
		toplevel.bind("<<NewWindow>>",		Flair.new)
		toplevel.bind("<<Paste>>",		self.paste)
		#toplevel.bind("<<PasteSelection>>",	self.pasteSelection)
		toplevel.bind("<<Redo>>",		self.redo)
		toplevel.bind("<<Refresh>>",		self.refreshKey)
		toplevel.bind("<<Save>>",		self.save)
		toplevel.bind("<<Undo>>",		self.undo)
		toplevel.bind("<<Execute>>",		self.execute)
		toplevel.bind("<<RibbonSelectAll>>",	self.selectAll)
		toplevel.bind("<<RibbonSelectClear>>",	self.selectClear)
		toplevel.bind("<<RibbonSelectInvert>>",	self.selectInvert)

		# flair keyboard bindings
		if Global.developer:
			toplevel.bind("<Control-Key-G>",garbage_collect)
		toplevel.bind("<Control-Key-q>",	programExit)
		toplevel.bind("<Control-Key-Q>",	programExit)
		toplevel.bind("<Control-Key-w>",	self.close)
		toplevel.bind("<Control-Key-W>",	self.close)
		#toplevel.bind("<Control-Key-Return>",	self.execute)

		toplevel.bind("<Control-Key-plus>",	self.fontIncrease)
		toplevel.bind("<Control-Key-equal>",	self.fontIncrease)
		toplevel.bind("<Control-Key-KP_Add>",	self.fontIncrease)
		toplevel.bind("<Control-Key-minus>",	self.fontDecrease)
		toplevel.bind("<Control-Key-underscore>",self.fontDecrease)
		toplevel.bind("<Control-Key-KP_Subtract>",self.fontDecrease)

		toplevel.bind("<Control-Tab>",		self.activatePrevious)
		#toplevel.bind("<Control-Tab>",		self.nextWindow)
		#toplevel.bind("<Shift-Control-Tab>",	self.prevWindow)
		#toplevel.bind("<Control-Tab>",		self.tabRight)
		#toplevel.bind("<Control-Shift-Tab>",	self.tabLeft)

		toplevel.bind("<Menu>",			self.tabFocus)
		toplevel.bind("<F10>",			self.tabFocus)
		toplevel.bind("<F12>",			self.minmax)
		toplevel.bind("<Alt-Control-Key-Left>",	self.tabLeft)
		toplevel.bind("<Alt-Control-Key-Right>",self.tabRight)
		toplevel.bind("<Alt-Control-Key-h>",	self.tabLeft)
		toplevel.bind("<Alt-Control-Key-l>",	self.tabRight)

	# ----------------------------------------------------------------------
	# Create pages on initialization
	# ----------------------------------------------------------------------
	def create(self, flair):
		# Create pages
		#for name, iconname, PageClass in _PAGES:
		for PageClass in _PAGES:
			name = PageClass._name_
			iconname = PageClass._icon_
			if isinstance(iconname,str):
				icon = icons.get(iconname)
			else:
				icon = iconname
			page = PageClass(flair, name, icon, False)
			# page settings
			try:
				settings = tkFlair.config.get(tkFlair._PAGE_SECTION,name).split()
				page.width  = int(settings[0])
				page.height = int(settings[1])
				page.pinned = bool(int(settings[2]))
				# shortcut?
				# undocked?
			except: # configparser.NoOptionError:
				pass
			self.addPage(page)

		# Set tab order
		try:
			tabs = tkFlair.config.get(tkFlair._PAGE_SECTION,"tabs").split()
		except: # configparser.NoOptionError:
			tabs = []
		if not tabs:
			tabs = ["Flair", "Input", "Geometry", "Run", "Data", "Plot"]
		self.tabs(tabs)

		# Remember setup
		self.changeDynamic()	#<--- must ensure that is not shown
		try:
			rpn = tkFlair.config.get(tkFlair._PAGE_SECTION,"pages").split()
		except configparser.NoOptionError:
			rpn = []

		if rpn:
			# Checked for errors
			for page in self.pages.values():
				if page.pinned and page.name not in rpn:
					#log.error("page",page.name,"was pinned and not in the rpn")
					#log.errorCont("RPN:", rpn)
					page.pinned = False

			# Count doubles
			for item in rpn:
				try:
					float(item)
				except ValueError:
					if rpn.count(item) > 1:
						#log.error("duplicate page",item," found in rpn")
						#log.errorCont("RPN:", rpn)
						rpn = []
						break

		# Apply rpn
		if rpn: self.RPN(rpn)

		# If project file exist then go to the right page else to Flair
		self.changePage(flair.project.page)	#<--- must ensure that exists

		# Undock pages
		try:
			for name in tkFlair.config.get(tkFlair._PAGE_SECTION,"undocked").split():
				self.undock(name)
		except configparser.NoOptionError:
			pass

	# ----------------------------------------------------------------------
	# retrieve page settings
	# ----------------------------------------------------------------------
	def saveSettings(self):
		# Find Undocked
		undocked = []
		for page in self.pages.values():
			if page.isundocked():
				undocked.append(page.name)
		tkFlair.config.set(tkFlair._PAGE_SECTION, "undocked"," ".join(undocked))

		# Remember page setup
		rpn = self.RPN()
		tkFlair.config.set(tkFlair._PAGE_SECTION, "pages"," ".join(map(str,rpn)))

		# dock all un docked windows
		self.dockAll()

		# Tabs
		tabs = self.tabs()
		tkFlair.config.set(tkFlair._PAGE_SECTION, "tabs"," ".join(tabs))

		# Write individual settings
		for page in self.pages.values():
			tkFlair.config.set(tkFlair._PAGE_SECTION, page.name,
				f"{page.width} {page.height} {int(page.pinned)}")
			page.saveConfig(tkFlair.config)

	# ----------------------------------------------------------------------
	def flair(self):
		return self.getActivePage().flair

	# ----------------------------------------------------------------------
	def close(self, event=None):
		page = self.getActivePage()
		if page.isundocked():
			page.dock()
		else:
			page.flair.close()

	# ----------------------------------------------------------------------
	def save(self, event=None):
		self.flair().saveProject()

	# ----------------------------------------------------------------------
	def cut(self,   event=None):	return self.getActivePage().cut(event)
	def copy(self,  event=None):	return self.getActivePage().copy(event)
	def paste(self, event=None):	return self.getActivePage().paste(event)
	def clone(self, event=None):	return self.getActivePage().clone(event)
	def help(self,  event=None):	return self.getActivePage().help()
	def minmax(self, event=None):	return self.getActivePage().minmaxButton()

	# ----------------------------------------------------------------------
	# undo command
	# ----------------------------------------------------------------------
	def undo(self, event=None):
		page = self.getActivePage()
		self._inundo = True
		page.undo(event)
		self._inundo = False
		page.refreshUndoButton()
		return "break"

	# ----------------------------------------------------------------------
	# redo command
	# ----------------------------------------------------------------------
	def redo(self, event=None):
		page = self.getActivePage()
		self._inundo = True
		page.redo(event)
		self._inundo = False
		page.refreshUndoButton()
		return "break"

	# ----------------------------------------------------------------------
	def undolist(self, event=None):
		if tkFlair.UndoListbox.posted is not None:
			tkFlair.UndoListbox.posted.cancel()
		else:
			b = self.tool["undolist"]
			x = b.winfo_rootx()
			y = b.winfo_rooty() + b.winfo_height()
			tkFlair.UndoListbox(self, self.getActivePage().flair, x, y)

	# ----------------------------------------------------------------------
	def selectAll(self, event=None):
		self.getActivePage().selectAll()

	#-----------------------------------------------------------------------
	def selectClear(self, event=None):
		self.getActivePage().selectClear()

	#-----------------------------------------------------------------------
	def selectInvert(self, event=None):
		self.getActivePage().selectInvert()

	# ----------------------------------------------------------------------
	def refreshUndoButton(self):
		page = self.getActivePage()
		page.refreshUndoButton()

	# ----------------------------------------------------------------------
	# refresh a page on specific events
	# If what is None then refresh all viewports
	# otherwise check if the what is in the refreshOn list of
	# each page
	#	therefore only the pages dealing with the
	#	executable will be refreshed
	# ----------------------------------------------------------------------
	def refresh(self, what=True, ignore=None):
		#log.say("Tabs.refresh=",what)
		for page in self.pages.values():
			if page is ignore: continue
			if what is True or what in page.refreshOn:
				page.setInvalid(what)

	# ----------------------------------------------------------------------
	def refreshKey(self, event=None):
		self.refresh()

	# ----------------------------------------------------------------------
	# redraw all frames
	# ----------------------------------------------------------------------
	def redraw(self, event=None):
		for page in self.pages.values():
			if page.created():
				page.redraw()
		self.refreshUndoButton()

	# ----------------------------------------------------------------------
	# Reset settings on each page (before loading new project)
	# ----------------------------------------------------------------------
	def reset(self):
		for page in self.pages.values():
			page.reset()

	# ----------------------------------------------------------------------
	# execute default command in current page
	# ----------------------------------------------------------------------
	def execute(self, event=None):
		self.getActivePage().execute()
		return "break"

	# ----------------------------------------------------------------------
	# active frame update
	# ----------------------------------------------------------------------
	def get(self):
		for page in self.pages.values():
			if page.ismapped():
				page.get()

	# ----------------------------------------------------------------------
	# Dock all undocked windows
	# ----------------------------------------------------------------------
	def dockAll(self):
		for page in self.pages.values():
			if page.page and page.isundocked():
				page.dock()

	#-----------------------------------------------------------------------
	# Create notification frame
	#-----------------------------------------------------------------------
	def createNotify(self):
		self.notifyFrame = tk.Frame(self, relief=tk.GROOVE, background="#FFA0A0")

		self._notifyTitle = tk.Label(self.notifyFrame, relief=tk.RAISED, font="bold")
		self._notifyTitle.grid(row=0, column=0, columnspan=2, sticky=tk.NSEW)

		self._notifyIcon = tk.Label(self.notifyFrame)
		self._notifyIcon.grid(row=1, column=0, sticky=tk.NSEW)

		self._notifyMsg = tk.Message(self.notifyFrame, aspect=500)
		self._notifyMsg.grid(row=1, column=1, sticky=tk.NSEW)

		self._notifyAction = tk.Button(self.notifyFrame)

		self.notifyFrame.grid_rowconfigure(1, weight=1)
		self.notifyFrame.grid_columnconfigure(1, weight=1)

		self._notifyTitle.bind("<Button-1>", self.notifyHide)
		self._notifyMsg.bind("<Button-1>",   self.notifyHide)
		self._notifyTitle.bind("<Enter>",    self._notifyShow)
		self._notifyMsg.bind("<Motion>",     self._notifyShow)
		self._notifyMsg.bind("<Enter>",      self._notifyShow)
		self._notifyMsg.bind("<Motion>",     self._notifyShow)
		self._notifyId = None
		self._notifyY = 0
		self._notifyT = 0

	#-----------------------------------------------------------------------
	def notify(self, title, msg, level=0, action=None, actionCmd=None, timeout=None):
		self._notifyTitle.config(text=title)
		self._notifyMsg.config(text=str(msg))
		if action:
			self._notifyAction.config(text=action, command=actionCmd)
			self._notifyAction.grid(row=2, column=0, columnspan=2)
		else:
			self._notifyAction.grid_forget()

		color = ("LightGreen", "#FFDD70", "#FFA0A0")[level]
		icon  = ("info32", "warning32", "error32")[level]
		self.notifyFrame.config(background=color)
		self._notifyTitle.config(background=color)
		self._notifyIcon.config(background=color, image=icons.get(icon))
		self._notifyMsg.config(background=color)
		self._notifyAction.config(background=color, activebackground=color)

		# Increase timeout by the size of the string
		if timeout is None:
			timeout = 15000
			if level == tkFlair.NOTIFY_WARNING:
				timeout *= 2
			if level == tkFlair.NOTIFY_ERROR:
				timeout *= 4
		timeout += (len(str(msg))*1000)//20

		if self._notifyId:
			self.notifyFrame.after_cancel(self._notifyId)

		self._notifyShow(timeout=timeout)

	#----------------------------------------------------------------------
	def _notifyShow(self, event=None, timeout=5000):
		if self._notifyId: self.notifyFrame.after_cancel(self._notifyId)
		width  = max(300,self.notifyFrame.winfo_reqwidth())
		height = max(100,self.notifyFrame.winfo_reqheight())
		self.notifyFrame.place(x=-width, width=width, relx=1.0,
				y=-height-30, height=height, rely=1.0)
		self._notifyY = -height
		self._notifyT = 0
		self._notifyId = self.notifyFrame.after(timeout, self._notifyAfter)

	#----------------------------------------------------------------------
	def _notifyAfter(self):
		if self._notifyY < 0:
			if self._notifyT == 0:
				self._notifyT = time.time()
			else:
				dt = time.time() - self._notifyT
				dy = int(100. * dt)	# speed * dt 1pixel/10ms
				if dy > 1:
					self._notifyY += dy
					self._notifyT = time.time()
			self.notifyFrame.place(y=self._notifyY-30)
			self._notifyId = self.notifyFrame.after(10, self._notifyAfter)
		else:
			self._notifyId = None
			self.notifyHide()

	#----------------------------------------------------------------------
	def notifyHide(self, event=None):
		if self._notifyId: self.notifyFrame.after_cancel(self._notifyId)
		self.notifyFrame.place_forget()
		self._notifyY  = 0
		self._notifyT  = 0
		self._notifyId = None

	#----------------------------------------------------------------------
	def notifyIsVisible(self):
		return self._notifyId is not None

	#----------------------------------------------------------------------
	def fontIncrease(self, event):
		tkFlair.fontIncrease(+1)
		return "break"

	#----------------------------------------------------------------------
	def fontDecrease(self, event):
		tkFlair.fontIncrease(-1)
		return "break"

#===============================================================================
# Flair Project Frame
#===============================================================================
class Flair(tk.Toplevel):
	# Output patterns used in double clicking and in GeometryErrors
	_CARD_PAT        = re.compile(r"[Cc]ard #(\d+)")
	_BODY_PAT        = re.compile(r"([Bb]ody|[Ss]urface) ['""](.*?)['""]")
	_MATERIAL_PAT    = re.compile(r"([Mm]aterial) ['""](.*?)['""]")
	_REGION_PAT      = re.compile(r"([Rr]egion|[Cc]ell|[Zz]one|[Mm]odule|[Cc]lone) ['""](.*?)['""]")
	_ALL_PATS = [	(_BODY_PAT,     "B"),
			(_MATERIAL_PAT, "M"),
			(_REGION_PAT,   "R"),
		]
	_REGION_ZONE_PAT = re.compile(r"(Region|Cell|Zone|Module|Clone) ['""](.*?):(zone)?(\d+)['""]")

	#----------------------------------------------------------------------
	# Initialize Flair
	#----------------------------------------------------------------------
	def __init__(self, master=None, project=None):
		super().__init__(master, class_="flair")

		try:
			geom = "%sx%s" % (tkFlair.config.get(tkFlair.__program__, "width"),
					  tkFlair.config.get(tkFlair.__program__, "height"))
		except configparser.NoOptionError:
			geom = "900x700"
		try:
			self.geometry(geom)
		except tk.TclError:
			pass

		try:
			self.tk.call('wm','iconphoto',self._w, icons.get("flair32"))
		except tk.TclError:
			self.iconbitmap(f"@{tkFlair.prgDir}/icons/flair32.xbm")

		self.log            = None
		self.geometryPage   = None
		self.plotEngine     = None
		self.undoredo       = undo.UndoRedo()
		self.jobs           = []		# list of active jobs
		self._renumber      = True		# toggle for renumbering cards
		self._refresh       = True		# toggle for refreshing display
		self._plotEngines   = {}
		self._autosave      = time.time()
		self._autosaved     = False
		self._renumberPos   = sys.maxsize
		self._nextsaveas    = False		# next save should be saveProjectAs

		self.__afterActivate= None
		self.__afterProcess = None
		self.__afterRefresh = None
		self._inRefresh     = False

		# Attach a project file
		if project is None:
			self.project = Project.Project()
		else:
			self.project = project
			self._scanProjectDirs()

		# Custom Tab
		self.tabs = FlairTabRibbonFrame(self)
		self.tabs.create(self)
		self.tabs.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)

		# Create frame
		self.refreshTitle()

		self.inputPage    = self.tabs.pages["Input"]
		self.viewerPage   = self.tabs.pages["Viewer"]
		self.geometryPage = self.tabs.pages["Geometry"]
		self.outputPage   = self.tabs.pages["Output"]

		# Check if there are messages in the output
		errors = bool(log.buffer())
		self.log = self.newLog("flair","flair")
		log.set(self.log)

		# do not double flush the messages when in developer mode
		if Global.developer: log.stopBuffering()
		self.flush()
		self.after_idle(self.tabs.refresh)

		self.protocol("WM_DELETE_WINDOW", self.close)
		self.bind("<<Output>>",           self.showOutput)
		self.bind("<<Preferences>>",      self.preferences)

		# focusIn semaphore
		self._inFocus = False
		self.bind("<FocusIn>", self.focusIn)

		if errors:
			self.errorsDuringLoading()

	#-------------------------------------------------------------------------------
	# New flair window
	#-------------------------------------------------------------------------------
	@staticmethod
	def new(event=None):
		return Flair()

	# --------------------------------------------------------------------
	def notify(self, title, msg, level=0, action=None,
			actionCmd=None, timeout=10000, showOnOutput=True):
		self.tabs.notify(title, msg, level, action, actionCmd, timeout)
		if showOnOutput:
			if level == tkFlair.NOTIFY_ERROR:
				if self.log:
					self.log.error(msg)
				else:
					print(msg)
			elif level == tkFlair.NOTIFY_WARNING:
				if self.log:
					self.log.warning(msg)
				else:
					print(msg)

	# --------------------------------------------------------------------
	def notifyHide(self):      self.tabs.notifyHide()
	def notifyIsVisible(self): return self.tabs.notifyIsVisible()

	# --------------------------------------------------------------------
	# Add a process to the watch list
	# --------------------------------------------------------------------
	def addProcess(self, process, logger, *args, **kw):
		self.jobs.append((process,logger))
		assert not process.isAlive()
		process.start(*args, **kw)
		if self.__afterProcess is None:
			self.__afterProcess = self.after(_JOB_REFRESHTIME,
							self._watchProcesses)

	# --------------------------------------------------------------------
	# Timer loop to watch processes for output and/or termination
	# --------------------------------------------------------------------
	def _watchProcesses(self):
		try: os.waitpid(-1,os.WNOHANG)	# Cleanup child processes
		except OSError: pass
		alive = False
		for i in range(len(self.jobs)-1, -1, -1):
			process, logger = self.jobs[i]
			if process.isAlive():
				logger.write(process.output())
				logger.status(process.status, process.percent)
				self.refreshStatus()
				alive = True
			else:
				logger.write(process.output())
				logger.status(process.status, process.percent)
				self.refreshStatus()
				if process.notify is not None: process.notify()
				logger.timestamp("Ended:")
				if process.message[0]:
					logger.write("\n%s\n%s\n" % process.message)
					if process.rc:
						self.notify(process.message[0],
							process.message[1],
							tkFlair.NOTIFY_ERROR,
							"Output", self.showOutput)
					else:
						self.notify(process.message[0],
							process.message[1])
				del self.jobs[i]
		if alive:
			self.__afterProcess = self.after(_JOB_REFRESHTIME, self._watchProcesses)
		else:
			self.__afterProcess = None

	#----------------------------------------------------------------------
	# Create a new log entry
	#----------------------------------------------------------------------
	def newLog(self, typ, name, append=False):
		if self.outputPage.page is None:
			self.outputPage.create(self.tabs)
		newlog = self.outputPage.newLog(typ, name, append)
		log.set(newlog)
		return newlog

	#----------------------------------------------------------------------
	# Restore log to flair
	#----------------------------------------------------------------------
	def restoreLog(self):
		log.set(self.log)

	#-----------------------------------------------------------------------
	# Write a string to log entry
	#-----------------------------------------------------------------------
	def write(self, msg):
		if self.log:
			self.flush()
			self.log.write(msg)

	# ----------------------------------------------------------------------
	def flush(self):
		if self.log and log.buffer():
			self.log.writeln("\n".join(log.buffer()))
			log.stopBuffering()

	#-----------------------------------------------------------------------
	# Project is modified (not the input)
	#-----------------------------------------------------------------------
	def modified(self, m=True):
		self.project.modified(m)
		if self.__afterRefresh is None:
			self.__afterRefresh = self.after_idle(self.refreshLate)

	#-----------------------------------------------------------------------
	# Input is modified (thus project as well)
	#-----------------------------------------------------------------------
	def inputModified(self, m=True):
		self.project.inputModified()
		self.modified(m)

	# ----------------------------------------------------------------------
	# Late refreshing of information after the input was modified
	# ----------------------------------------------------------------------
	def refreshLate(self):
		self.__afterRefresh = None
		self.refreshTitle()
		self.refreshStatus()
		if self.inputPage.page: self.inputPage.treeRefresh()
		if int(time.time() - self._autosave) > tkFlair._autoSaveTime:
			self._savePickle(_AUTOSAVE_FIlENAME, False)
			self._autosaved = True
			self._autosave  = time.time()
		self.tabs.getActivePage().refreshUndoButton()

	#-----------------------------------------------------------------------
	# Refresh frame title
	#-----------------------------------------------------------------------
	def refreshTitle(self):
		if self.project.isModified():
			c = "+ "
		else:
			c = ""

		if self.project.filename == "":
			title = "[untitled]"
		else:
			title = self.project.filename

		try: self.title(f"{c}{title} - {tkFlair.__program__}")
		except tk.TclError: pass

	# ----------------------------------------------------------------------
	# Refresh status bar
	# ----------------------------------------------------------------------
	def refreshStatus(self):
		if FlairRibbon.FlairPage.active:
			FlairRibbon.FlairPage.active.refreshStatus()

	# ----------------------------------------------------------------------
	# Set status message
	# ----------------------------------------------------------------------
	def setStatus(self, txt, fg="#000000", bg=None, update=False):
		if FlairRibbon.FlairPage.active:
			FlairRibbon.FlairPage.active.setStatus(txt,fg,bg,update)

	# ----------------------------------------------------------------------
	def refreshGeometry(self):
		if self.geometryPage.created():
			self.geometryPage.loadProject()

	# ----------------------------------------------------------------------
	def inundo(self):
		return self.tabs._inundo

	# ----------------------------------------------------------------------
	def resetUndo(self):
		# Reset undo info on all frames
		for page in self.tabs.pages.values():
			try:
				page.resetUndo()
				page.refreshUndoButton()
			except AttributeError:
				pass

	# ----------------------------------------------------------------------
	def busy(self):
		tkExtra.busy(self)

	# ----------------------------------------------------------------------
	def notBusy(self):
		tkExtra.notBusy(self)

	# ======================================================================
	# Event handlers
	# ======================================================================
	# Focus In event
	# ----------------------------------------------------------------------
	def focusIn(self, event):
		if self._inFocus: return
		# FocusIn is generated for all sub-windows, handle only the main window
		if self is not event.widget: return
		self._inFocus = True
		self.checkProjectFile()
		self._inFocus = False
		if self.project.dir():
			try:
				os.chdir(self.project.dir())
			except:
				pass

	# ----------------------------------------------------------------------
	# Check if input file has changed during editing
	# ----------------------------------------------------------------------
	def checkProjectFile(self, parent=None):
		if not self.project.checkFileTimestamp(): return
		if self.project.isModified():
			if parent is None: parent = self
			if tkFlair.askyesno("Warning",
					f"Project file {self.project.filename} has "
					"changed since editing started\n"
					"Reload new version",
					parent=parent):
				self.logReset()
				try:
					self.project.load()
				except (FileNotFoundError,PermissionError,OSError):
					self.notify("Error loading project",
						f"Cannot load project: '{self.project.filename}' "
						"either the file does not exist or you have "
						"insufficient permissions to access it",
						tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
				self.refresh("project")
				if not self.errorsDuringLoading():
					self.notify("Loaded project",
						"Project: %s\nInput: %s\nDir: %s" %
							(os.path.basename(self.project.filename),
							 self.project.inputName(),
							 self.project.dir()),
						timeout=2000)
			else:
				self.inputModified()
		else:
			try:
				self.project.load()
			except (FileNotFoundError,PermissionError,OSError):
				self.notify("Error loading project",
					f"Cannot load project: '{self.project.filename}' "
					"either the file does not exist or you have "
					"insufficient permissions to access it",
					tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
			self.refresh("project")
			self.geometryPage.loadState()

	#-----------------------------------------------------------------------
	# low level input conversion
	#-----------------------------------------------------------------------
	def _convertInput(self, inp, mode):
		options = convert.options(inp, mode)
		if options:
			opt = tkDialogs.ListboxDialog(self,
					f"{self.project.mode()} \N{RIGHTWARDS ARROW} {mode}",
					"Conversion options").show(options)
		else:
			opt = {}
		return convert.input(inp, mode, None, **opt)

	#-----------------------------------------------------------------------
	# convert project to another type
	#-----------------------------------------------------------------------
	def convert(self, mode):
		if self.project.mode() == mode: return True
		# -------------------------------------------------
		# FIXME TODO convert input to new format and import
		# all possible cards
		# For the moment just show the icons
		# -------------------------------------------------
		if not self.project.input.cardlist:
			# Empty file
			self.project.newInput(mode)
			self.refresh("mode")
			return True

		try:
			ans = tkDialogs.Dialog(self,
					{"title":"Input Convert",
					"text" : "Convert input mode:\n"
						f"{self.project.mode()} \N{RIGHTWARDS ARROW} {mode}.\n"
						"Warning: the change is irreversible!",
					"bitmap": "questhead",
					"default": 2,
					"strings": ("To New Project", "Save and Replace Project", "Cancel")})
		except tk.TclError:
			return

		if ans.num in (2,-1): return
		if ans.num == 1: self.saveProject()

		#if self.project.input.cardlist and self.project.isModified():
		#	if not tkFlair.askyesno("Warning",
		#		    f"Changing input from '{self.project.mode()}' to '{mode}'.\n"
		#		    "The change can be irreversible!\n"
		#		    "Please save the project before.\n"
		#		    "Save and perform the change?",
		#		    parent=self):
		#		return False
		#	self.saveProject()

		#elif not Global.experimental and not tkFlair.askyesno("Warning",
		#	    f"Changing input from '{self.project.mode()}' to '{mode}'.\n"
		#	    "The change can be irreversible!\n"
		#	    "Continue?",
		#	    parent=self):
		#	return False

		# -------------------------------------------------
		# Do the change
		# -------------------------------------------------
		self.busy()
		self._nextsaveas = True
		try:
			newinput = self._convertInput(self.project.input, mode)
		except convert.NotSupported as err:
			self.notify("ERROR Changing mode", str(err),
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
			self.notBusy()
			return False
		if ans.num==0:
			newflair = Flair()
			newflair.project.input = newinput
			newflair.modified()
			newflair.refresh("mode")
			newflair.tabs.changePage("Input")
		else:
			self.project.input = newinput
			self.modified()
			self.refresh("mode")

		self.notify("Input Mode changed",
			f"Input file mode has changed to '{mode}'",
			tkFlair.NOTIFY_WARNING)
		self.notBusy()
		return True

	# ----------------------------------------------------------------------
	# Project Files
	# ----------------------------------------------------------------------
	def newProject(self, event=None):
		self.get()
		if self.project.isModified():
			if not tkFlair.askyesno("New Project",
					"Current Project is modified\nDiscard changes",
					parent=self):
				return
		self.reset()
		self.project.runs.stopRefresh()
		self.project.init()
		self.project.input.minimumInput()
		self.tabs.refresh()
		self.refreshStatus()
		self.refreshTitle()
		self.resetUndo()

	# ----------------------------------------------------------------------
	# Scan all project directories and append them to the bFileDialog history
	# ----------------------------------------------------------------------
	def _scanProjectDirs(self):
		def add(name):
			if not isinstance(name,str): return
			if name == Project.RunInfo.DEFAULT_INPUT: return
			try:
				bFileDialog.FileDialog.append2History(
					os.path.dirname(os.path.abspath(name)))
			except:
				pass

		add(self.project.compile.name)
		for fn in self.project.compile("source"):
			add(fn)

		for run in self.project.runs():
			add(run.name)
			for usr in run.usrinfo:
				add(usr.name)

		for plot in self.project.plots():
			add(plot.name)
			for name,value in plot.items():
				if "file" in name:
					add(value)

		for layer in self.project.geometry.layers.values():
			for name,value in layer.items():
				if "file" in name:
					add(value)

	# ----------------------------------------------------------------------
	# Load project
	# ----------------------------------------------------------------------
	def loadProject(self, filename):
		self.logReset()
		self.reset()
		self.project.runs.stopRefresh()
		self.project.init()
		try:
			self.project.load(filename)
		except (FileNotFoundError,PermissionError,OSError):
			self.notify("Error loading project",
				f"Cannot load project: '{filename}' either the file "
				"does not exist or you have insufficient "
				"permissions to access it",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
		self._scanProjectDirs()
		self.reset()
		self._autosave = time.time()
		self.refreshStatus()
		self.inputModified(False)
		self.modified(False)
		self.refreshTitle()
		self.resetUndo()
		tkFlair.addRecent(filename)
		self.tabs.refresh()
		self.geometryPage.reset()
		#self.geometryPage.loadState()
		self.tabs.changePage(self.project.page)	#<--- must ensure that exists
		if not self.errorsDuringLoading():
			self.notify("Loaded project",
				"Project: %s\nInput: %s\nDir: %s" %
					(os.path.basename(filename),
					 self.project.inputName(),
					 self.project.dir()),
				timeout=2000)
		return True

	# ----------------------------------------------------------------------
	# Open Project
	# ----------------------------------------------------------------------
	def openProject(self):
		self.get()
		if self.project.isModified():
			if not tkFlair.askyesno("Open Project",
					"Current Project is modified\nDiscard changes",
					parent=self):
				return
		filename = bFileDialog.askopenfilename(master=self,
					title="Open flair project",
					initialdir=self.project.dir(),
					filetypes=[
						("Flair Project files","*.flair"),
						#("Flair autosave pickle","*.pickle"),
						("All","*")])
		if filename:
			_, ext = os.path.splitext(filename)
			if ext in inputs.ALL_EXTS:
				# Just as a fail safe precaution
				self.project.newInput(inputs.EXT2MODE[ext])
				try:
					self.project.loadInput(filename, onlyinput=True)
				except (FileNotFoundError,PermissionError,OSError):
					self.notify("Error loading project",
						f"Cannot load project: '{filename}' either the file "
						"does not exist or you have insufficient "
						"permissions to access it",
						tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
				self.redraw()
			elif ext == ".pickle":
				self.notify("Load pickle",
					"Pickle can only be loaded at startup\n"
					"Exit flair and issue the command:\n"
					f"flair {filename}",
					tkFlair.NOTIFY_WARNING)
				#self._loadPickle(filename)
			else:
				if self.loadProject(filename):
					tkFlair.addRecent(self.project.abspath())

	# ----------------------------------------------------------------------
	# low level save project, check for errors and report
	# ----------------------------------------------------------------------
	def _saveProject(self, filename=None):
		oldSize = self.log.size()
		try:
			self.project.save(filename)
			error = None
		except (PermissionError,OSError) as ex:
			error = str(ex)
			if Global.developer:
				traceback.print_exc()
		else:
			if self.log.size() != oldSize:
				error = "Look the output"

		if error is not None:
			self.notify("Error saving project file",
				"Error during saving the input\n"
				f"File: {self.project.abspath()}\n" + error,
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
			return True

		self.delPickle()
		return False

	# ----------------------------------------------------------------------
	# Save current project
	# ----------------------------------------------------------------------
	def saveProject(self):
		self.get()
		self.project.page = self.tabs.activePage.get()
		if not self.project.filename or self._nextsaveas:
			self._nextsaveas = False
			if self.saveProjectAs(): return True
		else:
			if self._saveProject(): return True
		self.refreshTitle()
		self.refreshStatus()
		self.delPickle()
		return False

	# ----------------------------------------------------------------------
	# Save Project As: current project with different filename
	# ----------------------------------------------------------------------
	def saveProjectAs(self):
		self.get()
		initfile = ""
		if not self.project.filename and self.project.external:
			initfile = self.project.external.abspath()
		filename = bFileDialog.asksaveasfilename(master=self,
					title="Save flair project as",
					initialdir=self.project.dir(),
					initialfile=initfile,
					filetypes=[
						("Flair Project files",("*.flair")),
						#("Flair autosave pickle","*.pickle"),
						("All","*")])
		if not filename: return True

		# Correct for accidental mistype of extension
		ext = os.path.splitext(filename)[1]
		if ext == ".pickle":
			if self._savePickle(filename): return True
		else:
			if ext != Project.EXT: filename += Project.EXT
			if self._saveProject(filename): return True
		self.refreshTitle()
		tkFlair.addRecent(self.project.abspath())
		self.project.modified(False)
		self.refresh("save")
		self.refreshTitle()
		self.refreshStatus()
		return False

	## ----------------------------------------------------------------------
	#def _loadPickle(self, filename=None):
	#	try:
	#		with bz2.BZ2File(filename,'r') as f:
	#			self.project = pickle.Unpickler(f).load()
	#		return False
	#	except Exception as ex:
	#		self.notify("Error loading input pickled file",
	#			"Error during saving the input\n"
	#			f"file: {filename}\n"
	#			f"Error: {ex}",
	#			tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
	#		if Global.developer:
	#			traceback.print_exc()
	#		return True

	# ----------------------------------------------------------------------
	def _savePickle(self, filename=None, showerror=True):
		try:
			with bz2.BZ2File(filename,'w') as f:
				pickle.Pickler(f).dump(self.project)
			return False
		except Exception as ex:
			if showerror:
				self.notify("Error saving pickled file",
					"Error during saving the project as pickle\n"
					f"file: {self.project.abspath()}.pickle\n"
					f"Error: {ex}",
					tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
			return True

	#-----------------------------------------------------------------------
	# Cleanup possible autosaved pickle file
	#-----------------------------------------------------------------------
	def delPickle(self):
		if self._autosaved:
			try: os.remove(_AUTOSAVE_FIlENAME)
			except: pass
			self._autosaved = False

	# ----------------------------------------------------------------------
	# Load input file
	# ----------------------------------------------------------------------
	def loadInput(self):
		self.get()
		filename = bFileDialog.askopenfilename(master=self,
					title="Load input file",
					initialdir=self.project.dir(),
					initialfile=self.project.inputName(),
					filetypes=inputs.filetypes(self.project.mode()))
		if filename:
			self.logReset()
			_, ext = os.path.splitext(filename)
			# Correct for accidental mistype of extension
			if ext in inputs.ALL_EXTS:
				self.project.newInput(inputs.EXT2MODE[ext])
			try:
				self.project.loadInput(filename, onlyinput=True)
			except Exception as ex:
				self.notify("Error loading project",
					f"Cannot load project: '{filename}' either the file "
					"does not exist or you have insufficient "
					"permissions to access it.\n"
					f"Error: {ex}",
					tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
			self.resetUndo()	# << before the refresh since it deletes all cards
			self.refresh()
			self.errorsDuringLoading()

	# ----------------------------------------------------------------------
	def _saveInput(self, filename=None):
		try:
			self.project.saveInput(filename)
			return False
		except Exception as ex:
			if filename is None: filename = self.project.inputName()
			self.notify("Error saving input file",
				"Error during saving the input\n"
				f"file: {self.project.dir()}/{filename}\n"
				f"Error: {ex}",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
			return True

	# ----------------------------------------------------------------------
	# Should be replaced with calls to saveProject() instead
	# ----------------------------------------------------------------------
	#def saveInput(self):
	#	# FIXME only when export is set
	#	self.get()
	#	self._autosave = time.time()
	#	if self.project.inputName():
	#		return self._saveInput()
	#	else:
	#		return self.saveInputAs()

	# ----------------------------------------------------------------------
	# Export input file
	# ----------------------------------------------------------------------
	def saveInputAs(self):
		self.tabs.get()
		self._autosave = time.time()
		initialfile = self.project.inputName()
		filename = bFileDialog.asksaveasfilename(master=self,
				title="Save Input file as",
				initialdir=self.project.dir(),
				initialfile=initialfile,
				filetypes=inputs.filetypes(self.project.mode()))
		if not filename: return True

		# Correct for accidental mistype of extension
		if filename == initialfile: return self.saveProject()
		if self._saveInput(filename): return True
		self.tabs.refresh("save")
		self.refreshTitle()
		self.refreshStatus()
		return False

	# ----------------------------------------------------------------------
	# Open template and set it as the new project
	# @param template	template to use, else use the minimumInput
	# @param onlyInput	if True load only the input from the project
	#			else load and replace the whole project
	# ----------------------------------------------------------------------
	def openTemplate(self, template=None, onlyInput=False):
		if self.project.isModified() and self.project.input.cardlist:
			if not tkFlair.askyesno("Exit",
				"Project is modified.\n"
				"Create a new input file?",
				parent=self):
				return

		if onlyInput:
			# replace input the existing project
			# open first a new project and use only the input
			project = Project.Project()
		else:
			self.reset()
			project = self.project
			project.init()
		self._autosave  = time.time()

		# open template based on extension
		if template is not None:
			_,ext = os.path.splitext(template)
			if ext == Project.EXT:
				cwd = os.getcwd()	# remember directory
				project.load(template)
				os.chdir(cwd)		# and restore it
				project.filename = ""
				self._autosave = time.time()
				self.refreshStatus()
				self.inputModified(False)
				self.modified(False)
				self.resetUndo()
				self.tabs.refresh()
				self.notify("New project",
					"New project file has been created based on "
					f"template: {os.path.basename(template)}")
			else:
				project.input.load(template)
				self.notify("New Input",
					"New Input file has been created based on "
					f"template: {os.path.basename(template)}")
		else:
			project.input = Fluka()
			project.input.minimumInput()

		if onlyInput:
			self.project.input = project.input
			project.input = None
			del project

		self.refreshTitle()
		self.refresh()

	# ----------------------------------------------------------------------
	# Import cards from another fluka file
	# ----------------------------------------------------------------------
	def importFluka(self):
		self.notify("Import",
			"Fluka input importing is not implemented for the moment",
			tkFlair.NOTIFY_WARNING)
		#inputFrame = self.frames["input"][0]
		#inputFrame.show()
		#inputFrame.insert2Input()

	# ----------------------------------------------------------------------
	def importMoira(self):
		self.notify("Import",
			"Moira input importing is not implemented for the moment",
			tkFlair.NOTIFY_WARNING)

	# ----------------------------------------------------------------------
	def importGdml(self):
		filename = bFileDialog.askopenfilename(master=self,
					title="Import GDML file",
					initialdir=self.project.dir(),
					filetypes=[("Gdml files","*.gdml"),
						("All","*")])
		if not filename: return
		oldSize = self.log.size()	# remember size before loading
		filename = self.project.relpath(filename)

		if self.project.ismode(Moira):
			gdml = Gdml(filename)
			Gdml2Moira(gdml, self.project.input).convert()
		elif self.project.ismode(Fluka):
			gdml = Gdml(filename)
			Gdml2Fluka(gdml, self.project.input).convert()

		if self.log.size() != oldSize:
			self.notify("GDML Import",
				"Errors or warnings during GDML import.\n"
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
		else:
			self.notify("GDML Import",
				f"GDML file '{filename}' imported")
		self.inputModified()
		self.redraw()

	# ----------------------------------------------------------------------
	def exportFluka(self):
		self.notify("Export",
			"Fluka input exporting is not implemented for the moment",
			tkFlair.NOTIFY_WARNING)

	# ----------------------------------------------------------------------
	def importMcnp(self):
		filename = bFileDialog.askopenfilename(master=self,
					title="Import MCNP file",
					initialdir=self.project.dir(),
					filetypes=inputs.filetypes(Mcnp.MODE, other=False))
		if not filename: return
		oldSize = self.log.size()	# remember size before loading
		mode = self.project.mode()
		mcnp = Mcnp()
		mcnp.load(filename)
		try:
			self.project.input = self._convertInput(mcnp, mode)
		except KeyError:
			self.notify("MCNP Import",
				"Errors converting from MCNP input.\n"
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
			return
		if self.log.size() != oldSize:
			self.notify("MCNP Import",
				"Errors or warnings during MCNP import.\n"
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
		else:
			self.notify("MCNP Import",
				f"MCNP file '{self.project.relpath(filename)}' imported")
		self.inputModified()
		self.redraw()

	# ----------------------------------------------------------------------
	def exportMcnp(self):
		if self.project.filename:
			initfile = self.project.abspath(self.project.name())+".mcnp"
		else:
			initfile = ""
		filename = bFileDialog.asksaveasfilename(master=self,
					title="Save Input as MCNP file",
					initialdir=self.project.dir(),
					initialfile=initfile,
					filetypes=inputs.filetypes(Mcnp.MODE, other=False))
		if not filename: return
		oldSize = self.log.size()	# remember size before loading
		try:
			mcnp = self._convertInput(self.project.input, Mcnp.MODE)
		except KeyError:
			self.notify("MCNP Export",
				"Failed to convert to MCNP.\n"
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
			return
		#mcnp.setMacroBodies(tkFlair.askyesno("Export to MCNP",
		#	"Do you want to use macro-bodies in MCNP/X like RPP, BOX, RCC...?",
		#	parent=self))
		#mcnp.fromFluka(self.project.input)
		mcnp.save(filename)
		if self.log.size() != oldSize:
			self.notify("MCNP Export",
				"Errors or warnings during MCNP export.\n"
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
		else:
			self.notify("MCNP Export",
				f"Input exported to MCNP format as {self.project.relpath(filename)}")

	# ----------------------------------------------------------------------
	def exportMoira(self):
		if self.project.filename:
			initfile = self.project.abspath(self.project.name())+".mcnp"
		else:
			initfile = ""
		filename = bFileDialog.asksaveasfilename(master=self,
					title="Save Input as Moira file",
					initialdir=self.project.dir(),
					initialfile=initfile,
					filetypes=[("Moira files","*.moira"),
						("All","*")])
		if not filename: return
		oldSize = self.log.size()	# remember size before loading
		self.project.input.writeMoira(filename)
		if self.log.size() != oldSize:
			self.notify("Moira Export",
				"Errors or warnings during Moira export.\n"
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
		else:
			self.notify("Moira Export",
				"Input exported to Moira format "
				f"as {self.project.relpath(filename)}")

	# ----------------------------------------------------------------------
	def exportPovray(self):
		if self.project.filename:
			initfile = self.project.abspath(self.project.name())+".pov"
		else:
			initfile = ""
		filename = bFileDialog.asksaveasfilename(master=self,
					title="Save Input as Povray file",
					initialdir=self.project.dir(),
					initialfile=initfile,
					filetypes=[("Povray files",("*.pov", "*.inc")),
							("All","*")])
		if not filename: return
		oldSize = self.log.size()	# remember size before loading
		relfn = self.project.relpath(filename)
		Povray.export(self.project.input, relfn)
		if self.log.size() != oldSize:
			self.notify("Povray Export",
				"Errors or warnings during Povray export.\n"
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
		else:
			self.notify("Povray Export",
				f"Input exported to Povray format as {relfn}")

	# ----------------------------------------------------------------------
	def exportOpenscad(self):
		if self.project.filename:
			initfile = self.project.abspath(self.project.name())+".pov"
		else:
			initfile = ""
		filename = bFileDialog.asksaveasfilename(master=self,
					title="Save Input as OpenSCAD file",
					initialdir=self.project.dir(),
					initialfile=initfile,
					filetypes=[("Povray files","*.scad"),
							("All","*")])
		if not filename: return
		oldSize = self.log.size()	# remember size before loading
		relfn = self.project.relpath(filename)
		OpenSCAD.export(self.project.input, relfn)
		if self.log.size() != oldSize:
			self.notify("OpenSCAD Export",
				"Errors or warnings during OpenSCAD export.\n"
				"Open the Output window for more information.",
				tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
		else:
			self.notify("OpenSCAD Exporting",
				f"Input exported to OpenSCAD format as {relfn}")

	# ----------------------------------------------------------------------
	def exportMakefile(self):
		initfile = "Makefile"
		fmake = bFileDialog.asksaveasfilename(master=self,
					title="Save makefile as",
					initialdir=self.project.dir(),
					initialfile=initfile,
					filetypes=[ ("All","*"),
						("Make files",("*.mak","*.make"))])
		if fmake:
			self.project.compile.writeMakefile(fmake)

	# ----------------------------------------------------------------------
	# insert/remove #include file for includecard
	# Process #include card, either load or remove code
	# depending on the card perform the appropriate action
	# through the undo/redo buffer
	# ----------------------------------------------------------------------
	def include(self, includecard):
		Global.dump(f"*-* Flair.include {includecard.name()!r}")
		# Check if #include contains already imported cards
		fromid = self.project.input.cardlist.index(includecard)
		toid   = fromid+1	# Find ending of #include
		level  = 1
		while toid<len(self.project.input.cardlist):
			card = self.project.input.cardlist[toid]
			if card.tag == "#include":
				level += 1
			elif card.tag == "#endinclude":
				card.setEnable(includecard.enable)
				level -= 1
				if level==0: break
			toid += 1

		# Check for disabling the #include card
		if not includecard.enable or \
		  (os.path.isfile(includecard.name()) and toid-fromid>1):
			for card in range(toid-1, fromid, -1):
				self.project.input.delCard(card, False)
			self.project.input.renumber(fromid)
			return

		# If range is not empty then return
		if toid-fromid>1: return

		# Create a temporary input to load the included lines
		_, ext = os.path.splitext(includecard.name())
		if ext == Project.EXT:
			# FIXME???
			p = Project.Project()
			p.load(includecard.name())
			incInput = p.input
		else:
			incInput = Fluka()
			location = 0	# 0=input, 1=bodies, 2=regions
			for i in range(fromid):
				card = self.project.input.cardlist[i]
				if not card.enable: continue
				if card.tag == "FREE":
					incInput.format = FlukaCard.FORMAT_FREE
				elif card.tag == "FIXED":
					incInput.format = FlukaCard.FORMAT_SINGLE
				elif card.tag == "GLOBAL":
					w4 = card.intWhat(4)
					if w4 in (2,3):
						incInput.format= FlukaCard.FORMAT_FREE
					if card.intWhat(5)>0:
						incInput.geoFormat = FlukaCard.FORMAT_FREE
				elif card.tag == "GEOBEGIN":
					if card.sdum()[:8]=="COMBNAME":
						incInput.geoFormat = FlukaCard.FORMAT_FREE
					elif card.sdum()[:8]=="COMBINAT":
						incInput.geoFormat = FlukaCard.FORMAT_SINGLE
					location += 1
				elif card.tag == "END":
					location += 1
				elif card.tag == "GEOEND":
					location = 0
			try:	# Check if file exist
				os.stat(includecard.name())
			except OSError:
				# File not exist do nothing
				return False

			incInput.open(includecard.name(),"r")
			if location==0:
				incInput._load()
			elif incInput==1:
				incInput._parseBodies()
			elif location==2:
				incInput._parseRegions()
			elif location==3:
				incInput._parseVolume()
			else:
				assert False
			assert not incInput._files

		pos = fromid+1
		for card in incInput.cardlist:
			card.input = self.project.input		# transfer ownership
			self.project.input.add(card, pos, False)
			pos += 1
		self.project.input.renumber(fromid)
		self.project.input.clearCache()
		self.refresh()

	# ----------------------------------------------------------------------
	def logReset(self):
		if self.log: self.log.reset()

	# ----------------------------------------------------------------------
	def errorsDuringLoading(self, warn=None):
		if self.log and self.log.size():
			msgtitle = "Warnings" if warn else "Errors"
			if warn is None:
				notify  = tkFlair.NOTIFY_ERROR
			elif warn:
				notify  = tkFlair.NOTIFY_WARNING
			else:
				notify  = tkFlair.NOTIFY_ERROR
			self.notify(f"{msgtitle} during loading",
				"Error or warning messages appeared while loading input. "
				"Open the Output window for more information.",
				notify, "Output", self.showOutput,
				showOnOutput=False)
			return True
		else:
			return False

	# ----------------------------------------------------------------------
	# redraw all frames
	# ----------------------------------------------------------------------
	def redraw(self, event=None):
		self.tabs.redraw()

	# ----------------------------------------------------------------------
	# refresh
	# @param what = True to force refresh everywhere, else string with refreshOn
	# @param ignore	pages to ignore
	# ----------------------------------------------------------------------
	def refresh(self, what=True, ignore=None):
		if self._inRefresh: return	# block re-entry of refresh
		self._inRefresh = True
		self.tabs.refresh(what, ignore)
		self._inRefresh = False

	# ----------------------------------------------------------------------
	def reset(self):
		self.tabs.reset()

	# ----------------------------------------------------------------------
	def get(self): self.tabs.get()

	#-----------------------------------------------------------------------
	# Open files in viewer
	#-----------------------------------------------------------------------
	def viewer(self, *files):
		Global.dump("*-* viewer:",files)
		if not files: return
		for fn in files:
			if isinstance(fn,tuple):
				# (name,data)
				self.tabs.changePage(self.viewerPage)
				self.viewerPage.show(*fn)
			else:
				# filename
				try:
					s = os.stat(fn)
				except:
					self.notify("Viewer error",
						f"Viewer cannot open file '{fn}'",
						tkFlair.NOTIFY_ERROR,
						"Output", self.showOutput)
					continue

				filename = os.path.basename(fn)

				# Check for a directory
				if stat.S_ISDIR(s.st_mode): continue

				if filename=="core" or FlukaConfig.CORE_PAT.match(filename):
					# Try to find the executable from the core file
					# FIXME needs to find Run-executable and not the master
					# Check for core dump
					try:
						out = subprocess.check_output(['file',fn])
						exe = re.search(r" execfn: '(.*?)', ",
							out.decode(errors="replace")).group(1)
					except Exception:
						exe = self.project.input.config.executable()
					self.write(f"{Project.terminal} -e {Project.debugger} {exe} {fn}&")
					os.system(f"{Project.terminal} -e {Project.debugger} {exe} {fn}&")
				else:
					self.tabs.changePage(self.viewerPage)
					if self.viewerPage.hasFile(fn):
						self.viewerPage.reload(fn)
					else:
						self.viewerPage.load(fn)

	# ----------------------------------------------------------------------
	# FIXME Somehow register associated application
	# ----------------------------------------------------------------------
	def application(self, *args):
		if not args: return
		exe = self.project.input.config.executable()
		try:
			if not os.path.isfile(exe):
				raise Exception(f"Executable '{exe}' not found")
			os.spawnv(os.P_NOWAIT, exe, [exe] + list(args) )
		except Exception as err:
			self.notify("Error executing command",
				f"Error executing command:\n{exe} {' '.join(args)}\n"
				f"{err}",
				tkFlair.NOTIFY_ERROR)

	# ----------------------------------------------------------------------
	def showPlot(self, name):
		# Search plots for the appropriate one
		for plot in self.project.plots():
			if name == plot.name:
				# Switch to Plots page and activate the plot
				plotPage = self.tabs.pages["Plot"]
				self.tabs.changePage(plotPage)
				plotPage.showPlot(name)
				break

	# ----------------------------------------------------------------------
	def doAllPlots(self):
		# Do all plots and save them
		for plot in self.project.plots():
			log.say("\nPlotting:",plot.name)
			plotter = GPPlot.GPPlot(self.plotEngine, self.project, plot)
			self.reset()
			err = plotter.show()
			if err:
				log.error(err)
			plotter.save(plot("format"))

	# ----------------------------------------------------------------------
	def showOutput(self, event=None):
		self.notifyHide()
		if self.outputPage.isdocked() and self.tabs.getActivePage() is self.outputPage:
			self.outputPage.hide(True)
		else:
			self.tabs.changePage(self.outputPage)

	#-------------------------------------------------------------------------------
	# Show error card in Input Editor or Geometry Editor
	#-------------------------------------------------------------------------------
	def showCard(self, errline):
		self.setStatus(errline, "DarkRed")

		# For input editor
		m = Flair._CARD_PAT.search(errline)
		if m:
			self.tabs.changePage("Input")
			self.inputPage.editor().goto(int(m.group(1))-1)
			return

		# Ask editor to select the object with the error
		m = Flair._REGION_ZONE_PAT.search(errline)
		if m:
			self.tabs.changePage("Geometry")
			self.geometryPage.selectGeometryPropertiesTab()
			self.geometryPage.select(("R",m.group(2),0), True)
			self.geometryPage.select(("Z",int(m.group(4))-1,0), True)
			return

		for pat,obj in Flair._ALL_PATS:
			m = pat.search(errline)
			if m:
				self.tabs.changePage("Geometry")
				self.geometryPage.selectGeometryPropertiesTab()
				self.geometryPage.select((obj,m.group(2),0), True)
				return

	# ----------------------------------------------------------------------
	# Preferences have changed
	# ----------------------------------------------------------------------
	def preferences(self, event=None, page=None):
		global checkFlukaExe
		self.notifyHide()
		if page is None:
			# Guess page
			page = self.tabs.getActivePage()
			if page.activeTab:
				page = page.activeTab.name
			else:
				page = page.name
			if page == "Compile":
				page = self.project.mode().capitalize()
		if tkFlair.preferences(self, page):
			self.redraw()
			checkFlukaExe = True
			self.checkFluka()

	def preferences_fluka(self): self.preferences(page="Fluka")
	def preferences_moira(self): self.preferences(page="Moira")

	# ----------------------------------------------------------------------
	# Close Window
	# ----------------------------------------------------------------------
	def close(self, event=None):
		global root
		self.get()
		if self.project.isModified():
			title = self.project.filename
			if title=="": title = "[untitled]"
			try:
				ans = tkDialogs.Dialog(self,
					{"title":f"Close: {title}",
					"text":f"Project {title} has been modified. "
						"Do you want to save it before "
						"exiting the application?",
					"bitmap": "questhead",
					"default": 2,
					"strings": ('Save and Close','Discard and Close','Cancel')})
			except tk.TclError:
				return
			if ans.num==0:
				if self.saveProject(): return
			elif ans.num in (2,-1):
				return

		self.project.runs.stopRefresh()
		self.delPickle()

		# FIXME check if there are unfinished jobs in the queue and ask the user
		# Save page settings
		self.tabs.saveSettings()

		# special I/O
		# splitters
		for name in ("Input", "Geometry", "Run", "Plot"):
			page = self.tabs.pages[name]
			if not page.created(): continue
			if not hasattr(page,"splitter"): continue
			if isinstance(page.splitter, tk.PanedWindow):
				tkFlair.config.set(tkFlair._PAGE_SECTION,
						f"{name}.split",
						page.splitter.sash_coord(0)[0]-page.splitter["bd"])
			else:
				try:
					tkFlair.config.set(tkFlair._PAGE_SECTION,
							f"{name}.split",
							page.splitter.split)
				except AttributeError:
					pass

			# Save viewer splitters
			if name == "Geometry":
				try:
					tkFlair.config.set(tkFlair._PAGE_SECTION,
							f"{name}.vsplit",
							page.vsplit())
					tkFlair.config.set(tkFlair._PAGE_SECTION,
							f"{name}.tsplit",
							page.tsplit())
					tkFlair.config.set(tkFlair._PAGE_SECTION,
							f"{name}.bsplit",
							page.bsplit())
				except AttributeError:
					pass

		# Save window info on config
		try:
			tkFlair.config.set(tkFlair.__program__, "width",  self.winfo_width())
			tkFlair.config.set(tkFlair.__program__, "height", self.winfo_height())
		except tk.TclError:
			pass

		flairCount = 0
		for child in root.winfo_children():
			if isinstance(child, Flair):
				flairCount += 1

		FlairRibbon.FlairPage.clearActive()
		self.destroy()
		if root is not None and flairCount==1:
			# Before closing the program
			check4Errors()
			# Write initialization files
			tkFlair.writeIni()
			saveMaterial()
			Manual.destroy()
			root.destroy()
			root = None

	# ----------------------------------------------------------------------
	# Quit program
	# ----------------------------------------------------------------------
	@staticmethod
	def quit():
		programExit()

	# ----------------------------------------------------------------------
	# Check for updates and license
	# ----------------------------------------------------------------------
	def checkUpdates(self):
		# Find fluka version
		Updates.CheckUpdateDialog(self, tkFlair.__version__,
					FlukaConfig.version)

	#-----------------------------------------------------------------------
	def check(self):
		if not Updates.License.verify():
			Updates.License(self)
			if not Updates.License.verify():
				sys.exit()
		if not Updates.CheckUpdateDialog.verify():
			self.checkUpdates()

	# ----------------------------------------------------------------------
	# Check fluka installation
	# ----------------------------------------------------------------------
	def checkFluka(self):
		# Verify if fluka exists in this directory
		global checkFlukaExe
		if not checkFlukaExe: return
		checkFlukaExe = False

		if FlukaConfig.directory != "":
			exefile = FlukaConfig.command("fluka")
		else:
			self.notify("Set Fluka Directory",
				"Fluka directory is not found in the settings nor in the PATH. "
				"Please set the Fluka Directory in the Preferences Dialog.",
				tkFlair.NOTIFY_WARNING,
				"Preferences", self.preferences_fluka)
			return
		FlukaConfig.findVersion()
		if os.path.isfile(exefile) and os.access(exefile, os.X_OK):
			return

		self.write(f"Warning: {exefile!r} do not exist\n")
		self.write("\tPlease set the correct FLUKA directory in Preferences\n")
		self.notify("Warning FLUKA executable not found",
			f"Cannot find FLUKA executable {exefile!r}\n"
			"Please set the correct FLUKA directory in Preferences",
			tkFlair.NOTIFY_WARNING,
			"Preferences", self.preferences_fluka)

	# ----------------------------------------------------------------------
	def checkMoira(self):
		# Verify if fluka exists in this directory
		global checkMoiraExe
		if not checkMoiraExe: return
		checkMoiraExe = False

		if MoiraConfig.directory != "":
			exefile = MoiraConfig.command("fluka")
		else:
			self.notify("Set Moira Directory",
				"Moira directory is not found in the settings nor in the PATH. "
				"Please set the Moira Directory in the Preferences Dialog.",
				tkFlair.NOTIFY_WARNING,
				"Preferences", self.preferences_fluka)
			return

		MoiraConfig.findVersion()

		if os.path.isfile(exefile) and os.access(exefile, os.X_OK):
			return

		self.write(f"Warning: {exefile!r} do not exist\n")
		self.write("Please set the correct Moira directory in Preferences\n")
		self.notify("Warning MOIRA executable not found",
			f"Cannot find Moira executable {exefile!r}\n"
			"Please set the correct Moira directory in Preferences",
			tkFlair.NOTIFY_WARNING,
			"Preferences", self.preferences_fluka)

	#-----------------------------------------------------------------------
	def clipboard(self, target='STRING', selection='CLIPBOARD'):
		try:
			return self.selection_get(selection=selection, type=target)
		except tk.TclError:
			return ""

	#-----------------------------------------------------------------------
	def clipboardUnpickler(self, target='STRING'):
		return pickle.Unpickler(io.BytesIO(binascii.a2b_hex(
					self.clipboard(target))))

	#-----------------------------------------------------------------------
	# Return clipboard targets
	#-----------------------------------------------------------------------
	def clipboardTargets(self):
		try:
			return self.clipboard('TARGETS').split()
		except tk.TclError:
			return []

	#-----------------------------------------------------------------------
	def clipboardAppend(self, bio, target='STRING'):
		self.clipboard_append(binascii.b2a_hex(bio.getvalue()), type=target)

	# ----------------------------------------------------------------------
	# Copy selected range to clipboard
	# ----------------------------------------------------------------------
	def cards2Clipboard(self, cards):
		self.clipboard_clear()

		# String representation
		cardstr = "\n".join([str(x) for x in cards])
		self.clipboard_append(cardstr, type='STRING')

		# Internal representation
		bio = io.BytesIO()
		pickler = pickle.Pickler(bio)
		for card in cards:
			pickler.dump(card)

		# copy to clipboard as string/hex format
		self.clipboardAppend(bio, tkFlair._CARD_MIME)

	#-----------------------------------------------------------------------
	# Return cards from clipboard
	#-----------------------------------------------------------------------
	def clipboard2Cards(self):
		cards = []
		Global.dump("*-* Flair.clipboard2Cards: targets=",self.clipboardTargets())
		if tkFlair._CARD_MIME in self.clipboardTargets():
			clipboard = self.clipboard(tkFlair._CARD_MIME)
			if clipboard:
				unpickler = pickle.Unpickler(io.BytesIO(
						binascii.a2b_hex(clipboard)))
				while True:
					try:
						card = unpickler.load()
						cards.append(card)
					except EOFError:
						break
			else:
				self.notify("Cannot paste",
						"Invalid content in clipboard",
						tkFlair.NOTIFY_WARNING)
				return []
		else:
			# We are in trouble, text format
			# Try to guess the format of the cards
			# Create a temporary input
			inp = Fluka()
			try:
				inp.load(io.StringIO(self.clipboard()))
			except TypeError:
				return None
			cards = inp.cardlist
			# Remove error cards
			#cards = [x for x in inp.cardlist
			#		if CardInfo.get("fluka",x).name != Card.ERROR]
			del inp
		return cards

	# ======================================================================
	# Undo/redo input functions
	# ======================================================================
	def addUndo(self, undoinfo, msg=None):
		if not undoinfo: return
		self.undoredo.add(undoinfo, msg)
		self.inputModified()

	# ----------------------------------------------------------------------
	# Create undo list with refresh and renumber calls
	# ----------------------------------------------------------------------
	def undolistRefresh(self, undoinfo=None, refresh=True, renumber=True):
		if undoinfo is None:
			# first call
			undoinfo = []
			if refresh:
				undoinfo.append(self.refreshUndo())
			else:
				undoinfo.append((self.refreshUndo,))
			if renumber:
				undoinfo.append(self.renumberUndo())
		else:
			# last call with inverse pattern
			if renumber:
				undoinfo.append(self.renumberUndo())
			if refresh:
				undoinfo.append(self.refreshUndo())
			else:
				undoinfo.append((self.refreshUndo,))
		return undoinfo

	# ----------------------------------------------------------------------
	# add undolist with refreshUndo and renumberUndo
	# ----------------------------------------------------------------------
	def addUndolistRefresh(self, undoinfo, msg):
		self.undolistRefresh(undoinfo)
		self.addUndo(undoinfo,msg)

	# ----------------------------------------------------------------------
	# Renumber only
	# ----------------------------------------------------------------------
	def undolistRenumberOnly(self, undoinfo=None):
		return self.undolistRefresh(undoinfo, False, True)

	#-----------------------------------------------------------------------
	def addUndolistRenumberOnly(self, undoinfo, msg):
		self.undolistRenumberOnly(undoinfo)
		if len(undoinfo)>4:	# add if there was something extra
			self.addUndo(undoinfo,msg)

	# ----------------------------------------------------------------------
	# Refresh only
	# ----------------------------------------------------------------------
	def undolistRefreshOnly(self, undoinfo=None):
		return self.undolistRefresh(undoinfo, True, False)

	#-----------------------------------------------------------------------
	def addUndolistRefreshOnly(self, undoinfo, msg):
		self.undolistRefreshOnly(undoinfo)	# execute the refreshUndo
		if len(undoinfo)>2:	# if only refreshUndo is present then don't add
			self.addUndo(undoinfo,msg)

	# ----------------------------------------------------------------------
	def canUndo(self):	return self.undoredo.canUndo()
	def canRedo(self):	return self.undoredo.canRedo()

	# ----------------------------------------------------------------------
	def undoInput(self):
		self.inputPage.commit()
		if self.canUndo():
			self.undoredo.undo()
			self.refresh()
			self.project.input.clearCache(Card.BODY)

	# ----------------------------------------------------------------------
	def redoInput(self):
		self.inputPage.commit()
		if self.canRedo():
			self.undoredo.redo()
			self.refresh()
			self.project.input.clearCache(Card.BODY)

	# ----------------------------------------------------------------------
	# Activate card in InputPage editor
	# activate card if needed in all Input editor and Geometry Viewer
	# ----------------------------------------------------------------------
	def activateIfNeeded(self, card):
		Global.dump("*-* Flair.activateIfNeeded")
		if isinstance(card,int):
			try:
				card = self.project.input.cardlist[card]
			except IndexError:
				card = self.project.input.cardlist[-1]
		if card.input is None: return
		if self.inputPage.page and self._refresh:
			# InputPage activate if needed should be executed immediately
			# if redrawing is required for the card
			# e.g. the comment number of lines has changed
			# it can create problems of repositioning
			self.inputPage.activateIfNeeded(card)
		if self.__afterActivate is None:
			self.__afterActivate = self.after_idle(self._activateIfNeededLate, card)

	#-----------------------------------------------------------------------
	# Activate card on idle
	#-----------------------------------------------------------------------
	def _activateIfNeededLate(self, card):
		Global.dump(f"*-* Flair.activateIfNeededLate card={card}")
		self.__afterActivate = None
		# We are still in undo but late
		self._undo = True
		if self.inputPage.page:
			self.inputPage.activateIfNeeded(card)
		if self.geometryPage and self.geometryPage.created():
			self.geometryPage.activateIfNeeded(card)
		self._undo = False

	# ----------------------------------------------------------------------
	# set what value to card and return undo info
	# ----------------------------------------------------------------------
	def setWhatUndo(self, card, w, v):
		undoinfo = (f"Card {card.tag}:{card.name()} what({w})={v}",
				self.setWhatUndo, card, w, card.what(w))
		card.setWhat(w, v)
		self.inputModified()
		self.activateIfNeeded(card)
		# FIXME maybe create a card method to update whatever is needed!
		if isinstance(card, FlukaCard) and card.info.useUnits:
			self.project.input.clearUnits()
		return undoinfo

	# ----------------------------------------------------------------------
	# set absolute what value and return undo info
	# ----------------------------------------------------------------------
	def setAbsWhatUndo(self, card, w, v):
		undoinfo = (f"Card {card.tag}:{card.name()} abs-what({w})={v}",
				self.setAbsWhatUndo, card, w, card.absWhat(w))
		card.setAbsWhat(w, v)
		self.inputModified()
		self.activateIfNeeded(card)
		return undoinfo

	# ----------------------------------------------------------------------
	# Change the sign of a card.what, and return undo info
	# ----------------------------------------------------------------------
	def setSignUndo(self, card, w, s):
		undoinfo = (f"Card {card.tag}:{card.name()} sign-what({w})",
				self.setSignUndo, card, w, card.sign(w))
		if not card.setSign(w, s): return None
		self.inputModified()
		self.activateIfNeeded(card)
		return undoinfo

	# ----------------------------------------------------------------------
	# set comment and return undo information
	# ----------------------------------------------------------------------
	def setCommentUndo(self, card, comment):
		undoinfo = (f"Card {card.tag}:{card.name()} comment",
				self.setCommentUndo, card, card.comment())
		card.setComment(comment)
		self.inputModified()
		self.activateIfNeeded(card)
		return undoinfo

	# ----------------------------------------------------------------------
	# set an associated variable var with value and return undo information
	# ----------------------------------------------------------------------
	def setCardUndo(self, card, name, value, default=None):
		Global.dump(f"*.* Flair.setCardUndo {card.tag}:{card.name()}[{name}]={value}")
		undoinfo = (f"Card {card.tag}:{card.name()} variable name={name}",
				self.setCardUndo, card, name,
				card.get(name,default), default)
		card[name] = value
		# Check for inclusion of a new file
		if card.tag == "#include":
			self.include(card)
		self.inputModified()
		self.activateIfNeeded(card)
		return undoinfo

	# ----------------------------------------------------------------------
	# Create the undo sequence for modifying a region with multiple cards
	# as well delete the cards and merge the comments
	# ----------------------------------------------------------------------
	def _modifyContRegionZoneUndo(self, region, msg):
		regions = list(self.project.input.cardsSorted("REGION",-1).values())
		try:
			pos = regions.index(region)
		except ValueError:
			return None

		undoinfo = self.undolistRefresh(refresh=False)
		undoinfo.append((self.setWhatUndo, region, -1, region.expr()))
		undoinfo.append((self.setCommentUndo, region, region.comment()))

		delcards = []
		expr     = region.expr()
		comment  = region.comment()
		for card in regions[pos+1:]:
			if card.name()=="&":
				cmt = card.comment()
				if cmt:
					if comment == "":
						comment = cmt
					else:
						comment += "\n"+cmt
				expr += "\n" + card.text()
				delcards.append(card)
			else:
				break
		region.setComment(comment)
		region.setText(expr)

		for card in reversed(delcards):
			undoinfo.append(self.delCardUndo(card.pos(), False))

		region[Card.CONT] = False
		self.undolistRefresh(undoinfo, False)
		return undoinfo #, msg)

	# ----------------------------------------------------------------------
	# add zone to region and return undo information
	# ----------------------------------------------------------------------
	def addZoneUndo(self, card, zone):
		msg = f"Card {card.tag}:{card.name()} add zone"
		if card[Card.CONT]:
			undoinfo = self._modifyContRegionZoneUndo(card, msg)
			if undoinfo is None:
				return undoinfo
		else:
			undoinfo = (msg, self.setWhatUndo, card, -1, card.text())

		card.addZone(zone)
		self.inputModified()
		self.activateIfNeeded(card)
		return undoinfo

	# ----------------------------------------------------------------------
	# set zones to region and return undo information
	# ----------------------------------------------------------------------
	def setZonesUndo(self, card, zones):
		msg = f"Card {card.tag}:{card.name()} set zones"
		if card[Card.CONT]:
			undoinfo = self._modifyContRegionZoneUndo(card, msg)
			if undoinfo is None:
				return undoinfo
		else:
			undoinfo = (msg, self.setWhatUndo, card, -1, card.text())

		card.setZones(zones)
		self.inputModified()
		self.activateIfNeeded(card)
		return undoinfo

	# ----------------------------------------------------------------------
	# set the number of card whats
	# ----------------------------------------------------------------------
	def setNWhatsUndo(self, card, n):
		if n > card.nwhats():
			undoinfo = (f"Card {card.tag}:{card.name()} set nwhats={n}",
					self.setNWhatsUndo, card, card.nwhats())
			card.setNWhats(n)
			self.inputModified()
			self.activateIfNeeded(card)
			return undoinfo
		elif n < card.nwhats():
			undoinfo = []
			for w in range(n+1, card.nwhats()):
				undoinfo.append(self.setWhatUndo(card, w, ""))
			undoinfo.append((self.setNWhatsUndo, card, card.nwhats()))
			card.setNWhats(n)
			self.inputModified()
			self.activateIfNeeded(card)
			return undoinfo
		else:
			#if card is not self.active.card: self.activate(card)
			return None

	# ----------------------------------------------------------------------
	# Add/Delete a card and return undo info
	# WARNING: For successive add/del cards disable renumbering and call
	#          twice renumberUndo() before and after the commands.
	#          The double call is needed since redo will change the order
	#          of the commands
	# ----------------------------------------------------------------------
	def addCardUndo(self, card, pos, renumber):
		#Global.dump(f"*-* addCardUndo card={card.tag} pos={pos}, renumber={renumber}")
		undoinfo = (f"Add card {card.tag}:{card.name()}",
				self.delCardUndo, pos, renumber)
		self.project.input.add(card, pos, renumber)
		self.inputModified()
		self.activateIfNeeded(pos)
		if not renumber:
			self._renumberPos = min(self._renumberPos, pos)
		return undoinfo

	# ----------------------------------------------------------------------
	# delete a card from input and return undo info
	# ----------------------------------------------------------------------
	def delCardUndo(self, pos, renumber):
		#Global.dump(f"*-* delCardUndo pos={pos}, renumber={renumber}")
		card  = self.project.input.cardlist[pos]
		undoinfo = (f"Delete card {card.tag}:{card.name()}",
				self.addCardUndo, card, pos, renumber)
		self.project.input.delCard(pos, renumber)
		self.inputModified()
		if not renumber:
			self._renumberPos = min(self._renumberPos, pos)
		return undoinfo

	# ----------------------------------------------------------------------
	# Remove all body references
	# ----------------------------------------------------------------------
	def removeBodyRefUndo(self, body):
		# Search all regions for possibly a reference to this body
		# and remove it
		pat = re.compile(r"([\+\-]\b%s\b)" % (re.escape(body)))
		undoinfo = []
		for card in self.project.input["REGION"]:
			if card.ignore(): continue
			expr = pat.sub("", card.text())
			if expr != card.text():
				undoinfo.append(self.setWhatUndo(card, -1, expr))
		return undoinfo

	# ----------------------------------------------------------------------
	# Remove all body references
	# Warning expects input to be correctly numbered
	# ----------------------------------------------------------------------
	def removeRegionRefUndo(self, region):
		# Search all cards referring to region and remove them
		remove = []
		for card in self.project.input.cardlist:
			if card.ignore(): continue
			case = card.info.findCase(card)
			lst  = card.info.findType("region", case)
			if isinstance(lst, tuple):
				lst = list(lst)
				lst.pop()	# skip step
			for w in lst:
				if card.what(w) == region:
					remove.append(card.pos())
					break
		undoinfo = []
		remove.reverse()
		for pos in remove:
			undoinfo.append(self.delCardUndo(pos, True))
		return undoinfo

	# ----------------------------------------------------------------------
	# Every second call it refreshes the display
	# It needs to be called twice on each undo:
	# - in the beginning
	# - in the end
	# ----------------------------------------------------------------------
	def refreshUndo(self):
		self._refresh = not self._refresh
		if self._refresh:
			if self.geometryPage.created():
				self.geometryPage.refreshUndo()
			self.redraw()
		return (self.refreshUndo,)

	# ----------------------------------------------------------------------
	# WARNING: renumberUndo should always come in pairs
	# one before and one in the end of the operation
	#
	# Every second call it renumbers the cards
	# It needs to be called twice on each undo:
	# - in the beginning
	# - in the end
	# ----------------------------------------------------------------------
	def renumberUndo(self):
		self._renumber = not self._renumber
		if not self._renumber:
			self._renumberPos = sys.maxsize	# an invalid number
		undoinfo = (self.renumberUndo,)
		if self._renumber and self._renumberPos<sys.maxsize:
			self.project.input.renumber(self._renumberPos-1)
			if isinstance(self.project.input, Fluka):
				self.project.input.clearUnits()
		return undoinfo

	# ----------------------------------------------------------------------
	def saveCardUndoInfo(self, pos):
		card    = self.project.input.cardlist[pos]
		comment = card.comment()
		tag     = card.tag
		whats   = card.whats()[:]	# clone list
		text    = card.text()
		prop    = card.property.copy()
		return ("Save card", self.restoreCardUndo,
				pos, comment, tag, whats, text, prop)

	# ----------------------------------------------------------------------
	def restoreCardUndo(self, pos, comment, tag, whats, text, prop):
		undoinfo = self.saveCardUndoInfo(pos)
		card = self.project.input.cardlist[pos]
		card.setComment(comment)
		if card.tag != tag:
			self.project.input.changeTag(card, tag)
		card.setWhats(whats)
		card.setText(text)
		card.property = prop
		return undoinfo

	# ----------------------------------------------------------------------
	def setEnableUndo(self, card, enable):
		if card.tag == "#include":
			undoinfo = self.undolistRefresh()
			undoinfo.append((self.setEnableUndo, card, card.enable))
			card.setEnable(enable)
			self.include(card)
			self.undolistRefresh(undoinfo)
		else:
			undoinfo = (self.setEnableUndo, card, card.enable)
			card.setEnable(enable)
			if card.isGeo() and len(card.tag)==3:
				self.project.input.clearCache(Card.BODY)
			elif card.tag in self.project.input._INDENT_INC or \
			     card.tag in self.project.input._INDENT_DEC:
				self.project.input.renumber(card.pos())
		self.inputModified()
		return undoinfo

	# ----------------------------------------------------------------------
	def changeTagUndo(self, card, tag):
		undoinfo = (self.changeTagUndo, card, f"{card.MODE}:{card.tag}")
		self.project.input.changeTag(card, tag)
		self.inputModified()
		return undoinfo

	# ----------------------------------------------------------------------
	def changeNameUndo(self, kind, old, new):
		self.project.input.changeName(kind, old, new)
		return (self.changeNameUndo, kind, new, old)

	#-----------------------------------------------------------------------
	# Insert or delete $start_xxx $end_xxx cards in order to make an input
	# that reflects body properties
	#-----------------------------------------------------------------------
	def correctBodyPropertiesUndo(self):
		translat  = None
		transform = None
		expansion = None
		idx = 0

		undoinfo = self.undolistRefresh()

		# -------------------------------------------------------------
		# insert a card in input
		# WARNING:
		#    default: is to insert above current card
		#    idx is pointing the next one
		# -------------------------------------------------------------
		def insCard(tag, whats=None, pos=None):
			nonlocal idx
			c = FlukaCard(tag, whats)
			if pos is None: pos = idx-1
			undoinfo.append(self.addCardUndo(c, pos, False))
			idx += 1
			return c

		# -------------------------------------------------------------
		# Remove card above if it is 'tag'
		# return True on success
		# WARNING:
		#    default: is to remove above the current card
		# -------------------------------------------------------------
		def removeAbove(tag, pos=None):
			nonlocal idx
			if pos is None: pos = idx-2
			if cardlist[pos].tag == tag:
				undoinfo.append(self.delCardUndo(pos, False))
				idx -= 1
				return True
			return False

		cardlist = self.project.input.cardlist
		while idx < len(cardlist):
			card = cardlist[idx]
			idx += 1
			prepro = card.tag[0]=="#"
			if card.ignore() or not (card.isGeo() or prepro): continue

			if (len(card.tag)==3 and card.isGeo()) or prepro:
				# if body has changed translation
				try:    dx = float(card.get("@dx",0.0))
				except: dx = 0.0
				try:    dy = float(card.get("@dy",0.0))
				except: dy = 0.0
				try:    dz = float(card.get("@dz",0.0))
				except: dz = 0.0
				try:    e = float(card.get("@expansion",0.0))
				except: e = 0.0
				t = card.get("@transform","")

				# ---------------------------------------------
				# Check/insert for {start/end}_translat
				# ---------------------------------------------
				if translat is None:
					# no pending translat card
					if dx!=0.0 or dy!=0.0 or dz!=0.0:
						translat = insCard("$start_translat",
								[dx, dy, dz])
				elif dx==0.0 and dy==0.0 and dz==0.0:
					# insert end_translat to close previous one
					if not removeAbove("$start_translat"):
						insCard("$end_translat")
					translat = None
				elif abs(dx-translat.numWhat(0)) > 1e-10 or \
				     abs(dy-translat.numWhat(1)) > 1e-10 or \
				     abs(dz-translat.numWhat(2)) > 1e-10:
					# modify previous translation
					if not removeAbove("$start_translat"):
						insCard("$end_translat")
					translat = insCard("$start_translat",
							[dx, dy, dz])

				# ---------------------------------------------
				# Check/insert for {start/end}_expansion
				# ---------------------------------------------
				if expansion is None:
					# no pending expansion card
					if e not in (0.0, 1.0):
						expansion = insCard("$start_expansion", [e])
				elif e in (0.0, 1.0):
					# insert end_expansion to close previous one
					if not removeAbove("$start_expansion"):
						insCard("$end_expansion")
					expansion = None
				elif abs(e-expansion.numWhat(0)) > 1e-10:
					# modify previous expansion
					if not removeAbove("$start_expansion"):
						insCard("$end_expansion")
					expansion = insCard("$start_expansion", [e])

				# ---------------------------------------------
				# Check/insert for {start/end}_transform
				# WARNING in case a start_transform is already present
				# ---------------------------------------------
				if transform is None:
					# no pending transform card
					if t != "":
						# check for a possible start_translat card
						if cardlist[idx-2].tag == "$start_translat":
							transform = insCard("$start_transform",
										[t], idx-2)
						else:
							transform = insCard("$start_transform",
										[t])
				elif t == "":
					# check immediately above
					if not removeAbove("$start_transform"):
						# or one card before in case of start_translat
						# is present above
						if not removeAbove("$start_transform", idx-3):
							insCard("$end_transform")
					transform = None
				elif t != transform.what(0):
					# modify previous transform
					if not removeAbove("$start_transform"):
						if not removeAbove("$start_transform", idx-3):
							insCard("$end_transform")
					if cardlist[idx-2].tag == "$start_translat":
						transform = insCard("$start_transform", [t], idx-2)
					else:
						transform = insCard("$start_transform", [t])

				# ---------------------------------------------
				if card.tag=="END": break

			elif card.tag[0]=="$":
				if card.tag=="$start_translat":
					if translat is not None:
						# Close any pending translat
						insCard("$end_translat")
					translat = card
				elif card.tag=="$start_transform":
					if transform is not None:
						insCard("$end_transform")
					transform = card
				elif card.tag=="$start_expansion":
					if expansion is not None:
						insCard("$end_expansion")
					expansion = card

				elif card.tag=="$end_translat":
					if translat is None:
						idx -= 1
						undoinfo.append(self.delCardUndo(idx, False))
					translat = None
				elif card.tag=="$end_transform":
					if transform is None:
						idx -= 1
						undoinfo.append(self.delCardUndo(idx, False))
					transform = None
				elif card.tag=="$end_expansion":
					if expansion is None:
						idx -= 1
						undoinfo.append(self.delCardUndo(idx, False))
					expansion = None

		self.undolistRefresh(undoinfo)

		return undoinfo

	# ----------------------------------------------------------------------
	# Base class to start plotting (Launches gnuplot)
	# ----------------------------------------------------------------------
	def startPlotEngine(self, engine):
		self.plotEngine = self._plotEngines.get(engine)
		if self.plotEngine is None:
			if engine == PLOT_ENGINES[0]:
				log.say("Starting gnuplot")
				self.plotEngine = GPPlot.start()
				if self.plotEngine is None:
					self.notify("Gnuplot not responding",
						"Gnuplot program is not responding. Please "
						"ensure that the path is correct or the program exist",
						tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
					return None
				log.say(f"Gnuplot version = {self.plotEngine.version()}")
				log.say(f"Default terminal = {Gnuplot.defaultTerminal}")

			elif engine == PLOT_ENGINES[1]:
				self.plotEngine = MPPlot.start()
				if self.plotEngine is None:
					self.notify("Matplotlib not responding",
						"Matplotlib python library is not responding. Please "
						"ensure that the necessary files are installed",
						tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
					return None
				log.say(f"Matplotlib version = {self.plotEngine.version()}")

			else:
				self.notify("Plot engine",
					f"Cannot find plot engine {engine}",
					tkFlair.NOTIFY_ERROR, "Output", self.showOutput)
				return None

			self._plotEngines[engine] = self.plotEngine

		if self.plotEngine.version() == 0:
			return None

		self.plotEngine.cd(self.project.dir())
		return self.plotEngine

#-------------------------------------------------------------------------------
# Material Database I/O
#-------------------------------------------------------------------------------
def loadMaterialIni():
	Material.init([os.path.join(tkFlair.prgDir,"db","material.ini"),
			tkFlair.iniMaterial])

#-------------------------------------------------------------------------------
def saveMaterial():
	if not Material.changed: return
	tkFlair.createIniDir()
	matini = Material.makeIni()
	try:
		with open(tkFlair.iniMaterial, "w", encoding="UTF-8") as fini:
			matini.write(fini)
	except IOError:
		messagebox.showerror("Cannot open file",
			f"Cannot open file {tkFlair.iniMaterial} for writing")
		log.error(f"Error: {tkFlair.__program__} cannot write '{tkFlair.iniMaterial}' file")
		return
	Material.changed = False

#-------------------------------------------------------------------------------
# check for errors
#-------------------------------------------------------------------------------
def check4Errors():
	if tkFlair.errors and tkFlair._errorReport:
		tkFlair.ReportDialog.show(root)

#-------------------------------------------------------------------------------
# Global Program Exit
#-------------------------------------------------------------------------------
def programExit(event=None):
	for child in root.winfo_children():
		try:
			if isinstance(child, Flair):
				child.close()
			else:
				child.destroy()
		except tk.TclError:
			pass

#-------------------------------------------------------------------------------
# show short help
# @see desktop/flair - the flair bash completion script
#-------------------------------------------------------------------------------
def usage():
	tkFlair.findRevision()
	print()
	print(f"Usage: {tkFlair.__program__} [options] [filename]...")
	print()
	print(" Options:")
	print("\t-1\t\tLoad the first flair file in the folder")
	print("\t--compile\tCompile executable")
	print("\t-c #|--convert\tConvert to simulation mode")
	print("\t-d\t\tActivate the beta-development features")
	print("\t-D\t\tDeactivate beta-development features (default)")
	print("\t--data #\tProcess/Merge the data files of all or specific runs (no default)")
	print("\t\t\tAccepts patterns or comma separated names e.g. *  or foo* or foo,bar*")
	print("\t--exe file\tFluka executable. (default: {FlukaDir}/bin/fluka)")
	print("\t-h|-?|--help\tprint this help page")
	print("\t-i/--ini file\tAlternative configuration file (default: $HOME/.flair/flair.ini)")
	print("\t-l | --list\tList recent projects")
	if Global.developer:
		print("\t-m #\t\tForce simulation mode of new project/input")
		print("\t\t\tAccepted types:", ", ".join(inputs.MODES))
		print("\t--noansi\tdisable ansi coloring on dumps")
		print("\t--nocleanup\tdisable cleanup of temporary files")
		print("\t-p\t\tenable profiling")
		print("\t-P\t\tdisable profiling (default)")
	print("\t--page #\tStart flair in specified page")
	print("\t--plots\t\tDo all plots and save the files")
	print("\t-r|--recent\tLoad most recent project")
	print("\t-R #\t\tLoad recent project (number 1..10 or filename)")
	print("\t--run #\t\tStart runs given as (name, pattern or a comma-list of runs)")
	print("\t-s | --skip\tSkip About dialog")
	print("\t-u | --update\tRecalculate and save input file variables")
	print("\t-v | --verbose\tIncrease verbosity level by one")
	print("\t-V\t\tDecrease verbosity level by one")
	print()
	print("\tfilename[s]\tflair project, input, imported files or directory")
	print("\t\t\tautomatically detect type from file extension")
	print("\t\t\textensions supported: <none|.flair>, .inp, .fluka,")
	print("\t\t\t                      .moira, .mac[ro], .gdml")
	print("\t\t\t                      .mcnp, .m, .phits, .pht,")
	print("\t\t\t                      .pickle")
	print()
	print(" Environment variables:")
	print("\tPYTHON\tSet this variable only in the case a different")
	print("\t\tpython installation is needed. The variable should")
	print("\t\tpoint the python (V3) executable")
	print()
	print(f" Authors:\t{tkFlair.__author__}")
	print(f" Email:\t\t{tkFlair.__email__}")
	print(f" Web:\t\t{tkFlair.__www__}")
	print(f" Created:\t{tkFlair.__created__}")
	print(f" Last Change:\t{tkFlair.__lastchange__}")
	print(f" Version:\t{tkFlair.__version__} [{tkFlair.__revision__} {tkFlair.__refname__}]")

#-------------------------------------------------------------------------------
def _getopt(arglist):
	try:
		return getopt.getopt(arglist,
					"?hc:dDi:eER:rlm:g1pPst:uvV",
					["help",
					 "compile",
					 "convert=",
					 "data=",
					 "exe=",
					 "ini=",
					 "list",
					 "mode=",
					 "noansi",
					 "nocleanup",
					 "page=",
					 "plots",
					 "recent",
					 "run=",
					 "skip",
					 "update",
					 "verbose"])
	except getopt.GetoptError:
		usage()
		sys.exit(1)

#-------------------------------------------------------------------------------
# parse command line arguments
# @return project
#-------------------------------------------------------------------------------
def parseArgs(arglist):
	project = Project.Project()

	data       = None
	recent     = None
	update     = False
	compileExe = False
	firstMatch = False
	forceMode  = None
	convertTo  = None
	gui        = True	# Is X11 gui needed?
	run        = ""

	#-----------------------------------------------------------------------
	# Scan in names
	#-----------------------------------------------------------------------
	def scanRecent(val):
		# first loop start for matching prefix
		# second pass check for any match
		for loop in range(2):
			for r in range(tkFlair._maxRecent):
				try: filename = tkFlair.getRecent(r)
				except: continue
				if not os.path.isfile(filename):
					continue
				fn, _ = os.path.splitext(os.path.basename(filename))
				if (loop==0 and fn.startswith(val)) or \
				   (loop==1 and val in filename):
					return r
		return 0

	optlist, args = _getopt(arglist)

	for opt, val in optlist:
		if opt in ("-h", "-?", "--help"):
			usage()
			sys.exit(0)
		elif opt in ("-i", "--ini"):
			pass
		elif opt == "--compile":
			compileExe = True
			gui = False
		elif opt in ("-c", "--convert"):
			convertTo = val
			gui = False
		elif opt == "--exe":
			project.compile.name = val
		elif opt == "-d":
			Global.developer = True
		elif opt == "-D":
			Global.developer = False
		elif opt == "--data":
			data = val
			gui = False
		elif opt == "-e":
			Global.experimental = True
		elif opt == "-E":
			Global.experimental = False
		elif opt == "-p":
			Global.profiler = True
		elif opt == "-P":
			Global.profiler = False
		elif opt == "-1":
			firstMatch = True
		elif opt in ("-m", "--mode"):
			if val in inputs.MODES:
				project.newInput(val)
				forceMode = val
			elif val in ("flair", "gdml", "pickle"):
				forceMode = val
			else:
				print(f"Error: unknown file mode {val!r}", file=sys.stderr)
				sys.exit(-1)
		elif opt in ("-r", "-R", "--recent", "-l", "--list"):
			if opt in ("-r","--recent"):
				r = 0
			elif opt in ("-l","--list"):
				r = -1
			else:
				try:
					r = int(val)-1
				except ValueError:
					r = scanRecent(val)
			if r<0:
				# display list of recent files
				maxlen = 10
				for i in range(tkFlair._maxRecent):
					try: filename = tkFlair.getRecent(i)
					except: continue
					if not filename: break
					maxlen = max(maxlen, len(os.path.basename(filename)))

				print("Recent files:")
				for i in range(tkFlair._maxRecent):
					try: filename = tkFlair.getRecent(i)
					except: continue
					if not filename: break
					d  = os.path.dirname(filename)
					fn = os.path.basename(filename)
					if os.path.isfile(filename):
						print(f"  {i+1:2d}: {fn:<{maxlen}}  {d}")
					else:
						print(f"  -x- {fn:<{maxlen}}  {d} [NOT FOUND]")
				try:
					r = input("Select one: ")
					if not r: sys.exit(0)
					r = int(r)-1
				except ValueError:
					r = scanRecent(r)
					pass
				except (KeyboardInterrupt, EOFError):
					sys.exit(0)
			try:
				recent = tkFlair.getRecent(r)
			except:
				pass
		elif opt in ("-s", "--skip"):
			tkFlair._SKIP_INTRO = True
		elif opt == "--page":
			pass
		elif opt == "--plots":
			pass		# special treatment we need the flair window
					# it will be executed in later in the main
		elif opt == "--run":
			run = val
			gui = False
		elif opt in ("-u","--update"):
			update = True
			gui = False
		elif opt in ("-v","--verbose"):
			Global.verbose = Global.verbose+1
		elif opt == "-V":
			Global.verbose = Global.verbose-1
		elif opt == "--noansi":
			Global._ansi = False
		elif opt == "--nocleanup":
			Project.cleanup = False
		else:
			log.error(f"Option {opt} = \'{val}\'")

	# Parse remaining arguments except files
	if recent: args.append(recent)
	loaded = False

	#-----------------------------------------------------------------------
	# Load first flair file from a given directory
	#-----------------------------------------------------------------------
	def loadFirstFlair(directory):
		for filename in os.listdir(directory):
			_,ext = os.path.splitext(filename)
			if ext == Project.EXT:
				try:
					project.load(os.path.join(directory,filename))
				except (FileNotFoundError,PermissionError,OSError) as ex:
					project.init()
					log.error(ex)
	#-----------------------------------------------------------------------

	if not args and firstMatch:
		loadFirstFlair(".")

	for filename in args:
		filename = os.path.abspath(filename)
		if os.path.isdir(filename):
			if not loaded:
				if firstMatch:
					loadFirstFlair(filename)
					loaded = True
				else:
					project.chdir(filename)
			continue

		name, ext = os.path.splitext(filename)
		try:
			project.chdir(os.path.dirname(filename))
		except Exception as ex:
			log.error(ex)
			sys.exit(1)

		# If forceMode is set and flair is requested then ignore it
		if ext == Project.EXT and forceMode != "flair": forceMode = None

		# if filename is not a complete path, then search for obvious candidates
		# assuming that the shell tab completion left it half
		if not os.path.exists(filename) and ext in ("",".",".fl") or forceMode == "flair":
			# try with .flair extension
			ext = Project.EXT
			if os.path.isfile(name + Project.EXT):
				ext = Project.EXT
				filename = name + ext
			elif os.path.isfile(name + FlukaConfig.EXT):
				ext = FlukaConfig.EXT
				filename = name + ext
			else:
				# open an empty project and assign the name
				# to what the user asked
				ext = Project.EXT
				project.init()
				project.filename = name + ext
				loaded = True
		if loaded:
			continue
		elif ext in inputs.ALL_EXTS or forceMode in inputs.MODES:
			if not forceMode:
				project.newInput(inputs.EXT2MODE[ext])
			try:
				project.loadInput(filename, mode=forceMode, onlyinput=True)
				loaded = True
			except (FileNotFoundError,PermissionError,OSError) as ex:
				project.init()
				log.error(ex)
				traceback.print_exc()
		elif ext == ".gdml" or forceMode == "gdml":
			gdml = Gdml(filename)
			if project.mode() == Moira.MODE:
				Gdml2Moira(gdml, project.input).convert()
			elif project.mode() == Fluka.MODE:
				Gdml2Fluka(gdml, project.input).convert()
			loaded = True
		elif ext == ".pickle" or forceMode == "pickle":
			try:
				with bz2.BZ2File(filename,'r') as f:
					project = pickle.Unpickler(f).load()
					loaded = True
			except Exception as ex:
				project.init()
				log.error(ex)
				traceback.print_exc()
		elif ext == Project.EXT or forceMode == "flair":
			try:
				project.load(filename)
				loaded = True
			except (FileNotFoundError,PermissionError,OSError):
				# open an empty project and assign the name
				# to what the user asked
				project.init()
				project.filename = filename
		else:
			log.error(f"Unrecognised file '{filename}'")
			log.errorCont("The '--mode' option could be used to specify format")
			sys.exit(1)

	# exit after processing args, with return code
	exitRC = None

	if convertTo is not None:
		newInp = convert.input(project.input, convertTo)
		project.input = newInp

	if update:
		exitRC = 0
		project.input.preprocess()
		project.saveInput()

	if run:
		exitRC = 0
		for name in run.split(","):
			for run in project.getRunByPattern(name):
				if run.family:
					for child in run.family:
						r = project.runs.getRunByName(child, run.name)
						print(f">>> Starting run: {r.name}")
						r.start()
				else:
					print(f">>> Starting run: {run.name}")
					run.start()
				#run.attach()

	if data:
		data = data.split(",")
		import DataProcess
		process = DataProcess.DataProcess()
		process.output = lambda x : sys.stdout.write(x.decode() if isinstance(x,bytes) else x)
		for run in project.runs():
			if run("parent"): continue	# skip child runs
			for d in data:
				if fnmatch.fnmatch(run.name, d):
					break
			else:
				continue
			print("\nProcessing:", run.name)
			project.findRunUsrInfo(run)
			process.init()
			process.execute(run.usrinfo)
			process.wait()
			process.output("%s\n%s\n" % process.message)
			if process.rc:
				exitRC = process.rc

	if compileExe:
		import CompileProcess
		process = CompileProcess.CompileProcess(project)
		process.output = lambda x : sys.stdout.write(x.decode()
						if isinstance(x,bytes) else x)
		process.init()
		process.execute()
		process.output("%s\n%s\n" % process.message)
		exitRC = process.rc

	if exitRC is not None: sys.exit(exitRC)

	return project, gui

#-------------------------------------------------------------------------------
# Post process arguments after the flair project is created
#-------------------------------------------------------------------------------
def postParseArg(arglist, flair, project):
	for opt, val in _getopt(arglist)[0]:
		if opt == "--page":
			# show page requested
			# find closest match
			val = val.upper()
			for page in flair.tabs.pages.keys():
				if page.upper().startswith(val):
					flair.tabs.changePage(page)
					break
		elif opt == "--plots":
			# do plots is requested
			if flair is None:
				log.error("Cannot connect to X11")
				sys.exit(1)
			flair.withdraw()
			if flair.startPlotEngine(PlotPage.PLOT_ENGINES[0]) is None:
				log.error("Cannot start plot engine")
				sys.exit(2)
			flair.doAllPlots()
			sys.exit(0)
		elif opt == "--run":
			for name in val.split(","):
				for run in project.runs.pattern(name):
					if run.family:
						for child in run.family:
							print(f">>> Starting run: {child.name}")
							child.start()
					else:
						print(f">>> Starting run: {run.name}")
						run.start()
					#run.attach()
			sys.exit(0)

#-------------------------------------------------------------------------------
# Initialize all databases with the correct order (for some of them)
#-------------------------------------------------------------------------------
def initialize(ini, gui):
	# Initialise options
	if gui:
		Element.init(os.path.join(tkFlair.prgDir, "db", "isotopes.ini"))
		loadMaterialIni()
	tkFlair.init(ini, gui)
	inputs.init()

	if gui:
		# --------------------------------------------------------------
		# Do aliases of all icons
		# --------------------------------------------------------------
		for cinfo in CardInfo.allcards():
			if cinfo.icon:
				try:
					icons.alias(cinfo.tag, cinfo.icon)
				except KeyError:
					cinfo.icon = None

		# --------------------------------------------------------------
		# Load all recent directories to history of bFileDialog
		# --------------------------------------------------------------
		for i in range(tkFlair._maxRecent):
			try: filename = tkFlair.getRecent(i)
			except: continue
			if not filename: break
			try:
				bFileDialog.FileDialog.append2History(
					os.path.dirname(os.path.abspath(filename)))
			except:
				pass

#-------------------------------------------------------------------------------
# Move main loop here to avoid any local variables exposed to the whole program
#-------------------------------------------------------------------------------
def main():
	global root
	#signal.signal(signal.SIGCHLD, signal.SIG_IGN)
	# Convert path to absolute
	sys.argv[0] = os.path.abspath(sys.argv[0])

	# ibus problem with tkinter, which causes a lag on flair opening
	# and closing as well memory leaks
	# https://github.com/ibus/ibus/issues/2324
	os.environ['XMODIFIERS'] = "@im=none"

	# Start buffering output in case of error message
	log.buffering()

	# Start X if possible
	#tk.NoDefaultRoot()
	try:
		root = tk.Tk()
		root.title(tkFlair.__program__)
		tk.CallWrapper = tkFlair.CallWrapper
		tkExtra.changeBehaviourTextClasses(root)
		root.withdraw()
		gui = True
	except tk.TclError as ex:
		gui = False
		guierror = str(ex)
		#print(ex, file=sys.stderr)
		#sys.exit(-1)

	# Initialise options
	# Check for alternative ini file
	ini = None
	for opt in ("-i","--ini"):
		if opt in sys.argv:
			try:
				ini = sys.argv[sys.argv.index(opt)+1]
				break
			except IndexError:
				pass
	initialize(ini, gui)

	if gui:
		tkFlair.addOptions(tkFlair.__program__, root)
		tkFlair.addOptions(tkFlair._FONT_SECTION, root)
		tkFlair.addOptions(tkFlair._COLOR_SECTION, root)
		# Check for dependencies
		if not GeometryEditor.installed():
			log.warning("Geometry Viewer not found")
			log.error(GeometryEditor.geoviewer_error())
		if not GeometryEditor.PILInstalled():
			log.warning("PIL.Image and PIL.ImageTk not found")

	project,_ = parseArgs(sys.argv[1:])
	if project.filename: tkFlair.addRecent(project.abspath())

	flair = Flair(root, project) if gui else None
	postParseArg(sys.argv[1:], flair, project)

	if not gui:
		print(guierror, file=sys.stderr)
		sys.exit(-1)

	if not tkFlair._SKIP_INTRO: tkFlair.aboutDialog(flair, 2500)
	flair.checkFluka()
	#flair.checkMoira()
	Manual.init([	("", os.path.join(tkFlair.prgDir,"manual")),
			(Fluka.MODE, FlukaConfig.documentDir()),
			(Moira.MODE, MoiraConfig.documentDir())])
	flair.check()
	GeometryEditor.checkVersion(flair)

	# Main loop
	if root is not None: root.mainloop()
		#count = 0
		#while True:
		#	try:
		#		root.mainloop()
		#		break
		#	except KeyboardInterrupt:
		#		count += 1
		#		log.say("***Interrupt***")
	Manual.Manual.done()
	icons.delete()

	del flair
	del root

#===============================================================================
# Main subroutine
#===============================================================================
if __name__ == "__main__":
	main()
