# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

import re

import csg
import log
import bmath
import Evaluation
from deck import CardInfo, Card, CSGCard

from .particle import FlukaParticle

#-------------------------------------------------------------------------------
# Patterns
_EXPR1VAR  = re.compile(r"^= *(-?) *([A-Za-z_][A-Za-z0-9_.:!\$]{1,8}) *$")

#===============================================================================
class FlukaGlobalDict(Evaluation.GlobalDict):
	pass

#===============================================================================
# Fluka CardInfo class
#===============================================================================
class FlukaCardInfo(CardInfo):
	BODY_NOVXL_TAGS = []	# Without VOXELS
	TRANSFORM_TAGS  = []
	DETECTOR_TAGS   = ("DETECT", "EVENTBIN","RESNUCLE","USRBDX",
			   "USRBIN", "USRCOLL", "USRTRACK","USRYIELD")

	#-----------------------------------------------------------------------
	def __init__(self, *args):
		super(). __init__(*args)
		if self.tag[0] not in "$#":
			self.tag = self.tag[0:8]	# truncate to 8 chars
		self.disableComment = True		# Comment disabled cards instead
							# of #if 0..#endif
		self.useUnits   = False			# True if any case is using units

	# ----------------------------------------------------------------------
	# Copy info from src
	# ----------------------------------------------------------------------
	def copy(self, src):
		super().copy(src)
		self.disableComment = src.disableComment
		self.useUnits = src.useUnits

	# ----------------------------------------------------------------------
	def setRange(self, range_):
		super().setRange(range_)
		self.useUnits = False
		for r in self.range:
			if "lunit" in r or "lunit-" in r:
				self.useUnits = True
				break

	# ----------------------------------------------------------------------
	def setDisableComment(self, c=False):
		self.disableComment = c

	# ----------------------------------------------------------------------
	# Validate card: return a list of failing variables.
	# ----------------------------------------------------------------------
	def validate(self, card, case=None):
		pb = super().validate(card, case)

		if card.input is None: return pb

		if card.isGeo() and len(card.tag) == 3 and card.tag!="END":
			if not card.NAMEPAT.match(card.name()):
				pb.append(0)

		elif card.tag == "COMPOUND":
			# Check for nesting compounds
			# XXX for them moment on level checking
			for w in range(2,7,2):
				if card.evalWhat(w) == card.name():
					pb.append(w)

		# Special check for regions
		elif card.type()==Card.REGION:
			name = card.name()
			if name!="&" and not card.NAMEPAT.match(name):
				pb.append(0)

			exp = csg.tokenize(card.expr())
			# Check expression
			try:
				csg.exp2rpn(exp)
				if exp and not csg.check(exp):
					pb.append(-1)
					pb.append("Invalid expression")
				else:
					bodies  = card.input.cardsCache(Card.BODY)
					regions = card.input.cardsCache(Card.REGION)
					notadded = True
					for token in exp:
						if token in "-+|@": continue
						if token[0] == '#':
							token = token[1:]
							if token not in regions:
								if notadded:
									pb.append(-1)
									notadded = False
								pb.append(f"Invalid region '{token}'")
						elif card.NAMEPAT.match(token):
							if token not in bodies:
								if notadded:
									pb.append(-1)
									notadded = False
								pb.append(f"Invalid body '{token}'")
						else:
							if notadded:
								pb.append(-1)
								notadded = False
							pb.append(f"Invalid token '{token}'")
			except csg.CSGException as ex:
				pb.append(-1)
				pb.append(str(ex))

		return pb

	# ----------------------------------------------------------------------
	# convert to names
	# ----------------------------------------------------------------------
	def toNames(self, card, case=None):
		if card.input is None: return

		if case is None:
			case = self.findCase(card)

		if case is None: return
		range_ = self.range[case]

		# Special cases
		if card.tag == "LATTICE":
			# Check the sdum and convert it to something that flair
			# and FREE format of FLUKA likes
			sdum = card.name().upper()
			if sdum[:4] == "ROT#" or sdum[:3] == "RO#":
				try:
					idx = int(sdum[sdum.find('#')+1:])
				except:
					idx = 999
				card.setSdum(f"rot#{idx:03d}")
		elif card.tag == "AUXSCORE":
			try:
				# we don't use the intWhat() since it will
				# always return 0 in case of names
				w = int(card.evalWhat(1))
				card.setWhat(1, ("USRBDX",
						 "USRBIN",
						 "USRTRACK",
						 "USRCOLL",
						 "USRYIELD")[w-1])
			except ValueError:
				pass

		# Go through the various whats
		for i in range(min(card.nwhats(), len(range_))):
			r = range_[i]
			w = card.what(i)
			if w == "": continue
			try: aw = int(float(w))
			except: continue
			if r == "material":
				aw = abs(aw)
				if aw==0:
					card.setWhat(i, "")
					continue
				# FIXME to correct treatment of defines...
				lst = card.input.materialList()
				aw -= 1
				if aw >= len(lst):
					card.setAbsWhat(i, "@LASTMAT")
				else:
					card.setAbsWhat(i, lst[aw])

			elif r in ("region", "region-"):
				if aw==0:
					card.setWhat(i, "")
					continue
				if r == "region":
					if aw==-1:
						card.setWhat(i, "@ALLREGS")
						continue
					if aw<0:
						continue
				aw = abs(aw)
				# FIXME to correct treatment of defines...
				try:
					regions = list(card.input.cardsCache("REGION").values())
					aw -= 1
					# Normally should not enter here unless if the voxel is not found
					if aw >= len(regions):
						if "VOXELS" in card.input.cards:
							# FIXME no way to know about @LASTREG!
							if aw == len(regions):
								card.setAbsWhat(i, "VOXEL")
							else:
								if (aw-len(regions)) < 1000:
									card.setAbsWhat(i,
										"VOXEL%03d"%(aw-len(regions)))
								elif (aw-len(regions)) < 10000:
									card.setAbsWhat(i,
										"VOXE%04d"%(aw-len(regions)))
								else:
									card.setAbsWhat(i,
										"VOX%05d"%(aw-len(regions)))
						else:
							card.setAbsWhat(i, "@LASTREG")
					else:
						card.setAbsWhat(i, regions[aw].name())
				except KeyError:
					pass

			elif r in ("particle", "particle-"):
				if aw==0:
					card.setWhat(i, "RAY")
					continue
				if r[-1]=="-":
					# Special treatment for DISCARD and negative particle id's
					if aw > 980:
						if aw<1000: card.setSign(i,True)
						aw = -abs(1000-aw)
						card.setAbsWhat(i, FlukaParticle.convert(aw, True))
					else:
						card.setAbsWhat(i,
							FlukaParticle.convert(abs(aw), True))
				else:
					card.setWhat(i, FlukaParticle.convert(aw, True))

			elif r=="detector":
				aw = abs(aw)
				if aw==0:
					card.setWhat(i, "")
					continue
				# Type of detector in sdum
				detectors = list(card.input.cardsCache(card.name()).values())
				if detectors:
					aw -= 1
					if 0 <= aw < len(detectors):
						name = detectors[aw]
						# Ensure that there is only one reference
						if detectors.index(name) == aw:
							card.setWhat(i, name)

			elif r=="binning":
				aw = abs(aw)
				if aw==0:
					card.setWhat(i, "")
					continue
				# FIXME detectors EVENTBIN & USRBIN
				detectors = list(card.input.cardsCache("USRBIN").values())
				if detectors:
					aw -= 1
					if 0 <= aw < len(detectors):
						name = detectors[aw]
						# Ensure that there is only one reference
						if detectors.index(name) == aw:
							card.setWhat(i, name)

	# ----------------------------------------------------------------------
	# Post initialization after loading other inputs init
	# ----------------------------------------------------------------------
	@staticmethod
	def post_init():
		mode = FlukaCard.MODE
		CardInfo.volumes[mode]   = "REGION"
		CardInfo.materials[mode] = "MATERIAL"
		CardInfo.rotations[mode] = "ROT-DEFI"

#===============================================================================
# FlukaCard base class
# FIXME:
#	subclass (dict) to replace the whats and properties
#	- whats / list arguments are assigned as [#] = value, starting from 0
#	- properties will be assigned as [name] = value
#	- all internal variables not to be saved will be prefixed
#	with "@" [@name] = value
#===============================================================================
class FlukaCard(Card, CSGCard):
	"""fluka input card"""

	MODE     = "fluka"	# Mode name
	COMMENT  = "*"

	EXPR         = "!@"
	EXPR_OLD     = 0	# Use old format !@what.=expr
	EXPR_EXPAND  = 1	# Do not write any expression

	SCALE     = "*...+....1....+....2....+....3....+....4....+....5....+....6....+....7....+....\n"

	IVOPT   = "@ivopt"
	IDBG    = "@idbg"
	VOXEL   = "@voxel"

	GEOFILE = "file"
	GEOOUT  = "output"

	# File formats
	FORMAT_SINGLE  = 0
	FORMAT_DOUBLE  = 1
	FORMAT_FREE    = 2
	FORMAT_VOXEL   = 3

	cardInfo   = FlukaCardInfo		# Pointer to Card Info
	globalDict = FlukaGlobalDict()

	# ----------------------------------------------------------------------
	def __init__(self, tag=None, what=None, comment="", text=""):
		"""Initialise a card
		Each card contains a tag, whats, one multi-line comment and sometimes
		some text information
		sdum=what[0]
		continuation cards have multiple whats of 1+6*lines.
		text is used for TITLE, GEOBEGIN and REGION
		"""
		super().__init__(tag, what, comment, text)
		self._sign    = []		# remember sign of what if empty
		assert self.tag[0] != "#"

	# ----------------------------------------------------------------------
	# Change tag of card, reduce the number of whats to the new card
	# XXX Do not forget to update the Input.cards... before
	# ----------------------------------------------------------------------
	def setTag(self, tag):
		super().setTag(tag)
		if not tag: return
		if tag == "REGION":
			self._type = Card.REGION
			self._geo  = True
		elif tag in ("LATTICE", "END", "LATTSNGL", "VOXELS") or tag[0]=="$":
			self._geo = True

	# ----------------------------------------------------------------------
	# I/O
	# ----------------------------------------------------------------------
	sdum = Card.name

	# ----------------------------------------------------------------------
	# Set SDUM to card (what[0])
	# ----------------------------------------------------------------------
	def setSdum(self, s):
		self.setWhat(0, s)

	# ----------------------------------------------------------------------
	# Set absolute information to card. The "what" will keep the same sign
	# ----------------------------------------------------------------------
	def setAbsWhat(self, w, v):
		"""change the absolute value of what keeping the same sign"""
		s = self.sign(w)

		if v in ("",0.0):
			self.setWhat(w, v)
			return
		if not Card.isEvalStr(v):
			s = self.sign(w)
			try:
				v = abs(float(v))
				if s: v = -v
			except:
				if v[0]=="+" or v[0]=="-":
					if s: v = "-" + v[1:]
					else: v = v[1:]
				else:
					if s: v = "-" + v
		self.setWhat(w, v)

	# ----------------------------------------------------------------------
	# Return absolute value of what
	# ----------------------------------------------------------------------
	def absWhat(self, w):
		"""return the absolute value of what[w]"""
		n = self.what(w)
		if n in ("",0.0): return n
		try:
			return abs(float(n))
		except:
			n = n.strip()
			if n and n[0] in ("+", "-"):
				return n[1:]
			return n

	# ----------------------------------------------------------------------
	# Signs are used to avoid cases when the what[w] is ZERO
	# Should not be used by the user... only in Layout.py
	# @return True on success (not an expression)
	# ----------------------------------------------------------------------
	def setSign(self, w, s):
		# only ignore the flair-expressions
		if self.isEvalWhat(w): return False

		if w >= len(self._sign):
			self._sign += [None] * (w-len(self._sign)+1)

		self._sign[w] = s
		n = self.what(w)

		if n in ("",0.0): return
		try:
			n = abs(float(n))
			if s: n = -n
		except:
			n = n.strip()
			if n[0] in ("+", "-"):
				if s: n = "-" + n[1:]
				else: n = n[1:]
			else:
				if s: n = "-" + n
		self.setWhat(w, n)
		return True

	# ----------------------------------------------------------------------
	# return (minus) sign of a what
	# ----------------------------------------------------------------------
	def sign(self, w):
		"""return (minus) sign of a number: False=+, True=-"""
		if self.isEvalWhat(w):
			# special case
			n = self.evalWhat(w)
		else:
			# Default numeric value
			try:
				n = self._what[w]
			except:
				try: return self._sign[w]
				except: return False

		if n=="": n=0.0
		try:
			n = float(n)
			if n<0.0: return True
			if n>0.0: return False

			if w<len(self._sign):
				s = self._sign[w]
				if s is not None: return s
			return False
		except:
			n = n.strip()
			if not n: return False
			if n[0] == "-": return True
			if n[0] == "+": return False

			if w<len(self._sign):
				s = self._sign[w]
				if s is not None: return s
			return False

	# ----------------------------------------------------------------------
	# test card against a filter-function
	# ----------------------------------------------------------------------
	def test(self, func, var, value, keepcase=False):
		"""check card against a filter function func"""
		# Units
		if var=="lunit":
			uval = abs(int(value))
			try:
				self.fileunits().index(uval)
				return True
			except:
				return False

		# Region
		elif var == "region":
			if self.tag == "REGION":
				return (str(self.name())
					if keepcase
					else str(self.name()).upper()) == value
			varlist = self.findType(var)
			if isinstance(varlist, list):
				# FIXME check for signed...
				if not keepcase:
					varlist = [str(x).upper() for x in varlist]
				return value in varlist
			else:
				# Check for a range
				reg         = self.input.region(value, False)
				reg_from    = self.input.region(varlist[0], False)
				if reg_from == reg: return True
				reg_to      = self.input.region(varlist[1], False)
				if reg_to   == reg: return True
				reg_step    = varlist[2]
				if reg_from < reg < reg_to:
					return ((reg - reg_from) % reg_step) == 0
			return False

		# Body
		elif var=="body":
			if not self._geo: return
			if self.tag == "REGION":
				opt = re.MULTILINE|re.DOTALL
				if not keepcase: opt |= re.I
				return re.match(rf".*\b{value}\b.*", self._text,opt) is not None
			if self.tag not in CardInfo.bodies[self.MODE]: return False
			return (str(self.name()) if keepcase else str(self.name()).upper()) == value

		# Material
		elif var=="material":
			if self.tag=="MATERIAL":
				if keepcase:
					return str(self.name()) == value
				else:
					return str(self.name()).upper() == value
			varlist = self.findType(var)
			if isinstance(varlist, list):
				# FIXME check for signed...
				if not keepcase: varlist = [str(x).upper() for x in varlist]
				return value in varlist
			elif self.input.MODE == self.MODE:
				mat         = self.input.material(value, False)
				mat_from    = self.input.material(varlist[0], False)
				if mat_from == mat: return True
				mat_to      = self.input.material(varlist[1], False)
				if mat_to   == mat: return True
				mat_step    = varlist[2]
				if mat_from < mat < mat_to:
					return ((mat - mat_from) % mat_step) == 0
			return False

		# Transformation
		elif var=="transform":
			# FIXME check also what(1) and Rot#\d+
			if self.tag=="ROT-DEFI":
				if keepcase:
					return str(self.what(0)) == value
				else:
					return str(self.what(0)).upper() == value
			varlist = self.findType(var)
			if not keepcase: varlist = [str(x).upper() for x in varlist]
			if value in varlist: return True
			# Check also -value
			if isinstance(value, (int, float)):
				if -value in varlist: return True
			elif isinstance(value,str):
				if f"-{value}" in varlist: return True
			return False

		# Detector
		elif var=="detector":
			if self.tag in FlukaCardInfo.DETECTOR_TAGS:
				if keepcase:
					return str(self.what(0)) == value
				else:
					return str(self.what(0)).upper() == value

			# FIXME ranges are not supported yet!
			varlist = list(map(str,self.findType(var)))
			varlist.extend(list(map(str,self.findType("binning"))))
			if not keepcase: varlist = [x.upper() for x in varlist]
			return value in varlist
		# Particle
		elif var=="particle":
			varlist = self.findType(var)
			if isinstance(varlist, list):
				# FIXME check for signed...
				if not keepcase: varlist = [str(x).upper() for x in varlist]
				return value in varlist
			else:
				part         = FlukaParticle.convert(value, False)
				part_from    = FlukaParticle.convert(varlist[0], False)
				if part_from == part: return True
				part_to      = FlukaParticle.convert(varlist[1], False)
				if part_to == part: return True
				part_step    = varlist[2]
				if part_from < part < part_to:
					return ((part - part_from) % part_step) == 0
				return False

		else:
			return super().test(func, var, value, keepcase)

	#-----------------------------------------------------------------------
	# Check COMPOUND cards for correct mixture
	#-----------------------------------------------------------------------
	def checkCompound(self, all_=False):
		assert self.tag == "COMPOUND"
		# check sign of whats
		s1 = self.sign(1)
		s2 = self.sign(2)
		n  = self.nwhats()
		if all_:
			start = 1
		else:
			start = n-6
		error = False
		for i in range(start, n, 2):
			j = i+1
			if self.numWhat(i)!=0.0:
				if self.sign(i)!=s1 or \
				   self.sign(j)!=s2:
					log.error("Inconsistent compound mixture (Check signs)")
					log.errorCont(f"What({i}) = {self.what(i)}")
					log.errorCont(f"What({j}) = {self.what(j)}")
					log.errorCont(self)
					error = True
			else:
				self.setSign(i, s1)
				self.setSign(j, s2)
		return error

	# ----------------------------------------------------------------------
	# Convert card to names or numbers
	# ----------------------------------------------------------------------
	def convert(self, tonames=True):
		if tonames:
			self.info.toNames(self)
		else:
			self.info.toNumbers(self)
		self.modified()

	# ----------------------------------------------------------------------
	# Return units used by card
	# ----------------------------------------------------------------------
	def fileunits(self, absolute=True):
		"""return unit from card"""
		case = self.info.findCase(self)
		lst  = self.info.findType("lunit", case)
		if absolute:
			if isinstance(lst, list):
				return [abs(self.intWhat(x)) for x in lst]
			else:
				return (abs(self.intWhat(lst[0])),
					abs(self.intWhat(lst[1])), lst[2])
		else:
			if isinstance(lst, list):
				return [self.intWhat(x) for x in lst]
			else:
				return (self.intWhat(lst[0]),
					self.intWhat(lst[1]), lst[2])

	# ----------------------------------------------------------------------
	# return bodies in single/double/free format
	# ----------------------------------------------------------------------
	def _bodyStr(self, prefix="", fmt=None):
		i = 1
		if fmt is None or fmt==FlukaCard.FORMAT_FREE:
			name = str(self.what(0)).strip()
			if name == "":
				name = "<name>"
			line = "%s%s %-10s" % (prefix, self.tag, name)
			prevlen = 0
			while i<self.info.maxwhats:
				# use what to keep it as string
				if fmt is None:
					w = self.evalWhat(i,False)
				else:
					w = self.evalWhatOrName(i)
				try:
					float(w)	# Force to number
					w = bmath.format(w, 19)
				except (ValueError,TypeError):
					if not w or (not isinstance(w,str) or (w[0]!="$"
					   and w[:2] not in ("-$","+$"))):
						w = "0.0"
				if len(line)+len(w)+1 > prevlen+80:
					prevlen = len(line)
					line += f"\n{prefix}              {w}"
				else:
					line += f" {w}"
				i += 1
			return line

		elif fmt==FlukaCard.FORMAT_SINGLE:
			width = 10
			step  = 6

		else:
			width = 21
			step  = 3

		line = "%s  %3s%5s" % (prefix, self.tag, str(self.what(0)))
		while i<len(self._what):
			inext = i+step
			imin  = min(inext, len(self._what))
			while i<imin:
				w = self._what[i]
				try:
					float(w)
					w = bmath.format(w, width)
				except ValueError:
					w = "0.0"
				line += w.rjust(width)
				i += 1
			if i<len(self._what): line += "\n%s          "%(prefix)
		return line

	# ----------------------------------------------------------------------
	# return end string
	# ----------------------------------------------------------------------
	@staticmethod
	def _endStr(fmt=None):
		if fmt is None or fmt==FlukaCard.FORMAT_FREE:
			return "END"
		else:
			return "  END"

	# ----------------------------------------------------------------------
	# Split a region expression into multiple lines
	# ----------------------------------------------------------------------
	@staticmethod
	def splitExpr(expr, maxlength=100):
		line = ""
		first = True
		for l in expr.splitlines():
			l = l.rstrip()	# remove trailing spaces
			if not l: continue
			while len(l)>maxlength:
				for i in range(maxlength-1, 1, -1):
					if l[i] not in ('+','-','|','(',')','#'):
						continue
					# split line on char before
					if first:
						line += l[0:i].rstrip()
						first = False
					else:
						line += "\n%s" % (l[0:i])
					l = l[i:]
					break
				if i<=2: break
			if first:
				line += l
				first = False
			else:
				line += "\n%s" % (l)
		return line

	# ----------------------------------------------------------------------
	# @return region expression with expanded #REGION tokens
	# ----------------------------------------------------------------------
	def expandRegions(self):
		REGION = re.compile(r"(.*)#([A-Za-z_][A-Za-z0-9_.@\$]*)(.*)")
		expr = self._text
		regions = self.input.cardsCache(Card.REGION)
		# to avoid infinite loops
		for i in range(10000):
			pat = REGION.match(expr)
			if pat is None: return expr
			try:
				region = regions[pat.group(2)]
			except KeyError:
				log.error(f"expandRegions: Error in REGION {self.name()!r}"
					f"cannot find: #{pat.group(2)!r}")
				return self._text
			expr = f"{pat.group(1)}-({region.text()}){pat.group(3)}"
		log.error(f"expandRegions: Error in REGION {self.name()!r} infinite loop detected")
		return self._text

	# ----------------------------------------------------------------------
	# return region string
	# ----------------------------------------------------------------------
	def _regionStr(self, fmt=None):
		try:    neighbors = int(self._what[1])
		except: neighbors = 5

		if fmt is None or fmt==FlukaCard.FORMAT_FREE:
			name = str(self.name()).strip()
			if name == "":
				name = "<name>"

			if name == "&":	# Continuation of a region
				line = "              "
			else:
				line = f"{name:<10}{neighbors:4d} "

			first = True
			if '#' in self._text:
				expr = self.expandRegions()
			else:
				expr = self._text
			for l in expr.splitlines():
				l = l.rstrip()	# remove trailing spaces
				if not l: continue
				while len(l)>100:
					for i in range(99, 1, -1):
						if l[i] not in '+-|()#':
							continue
						# split line on char before
						if first:
							line += l[0:i].rstrip()
							first = False
						else:
							line += f"\n               {l[0:i]}"
						l = l[i:]
						break
					if i<=2: break
				if first:
					line += l
					first = False
				else:
					line += f"\n               {l}"
			return line
		else:	# fixed format
			expr = ""
			first = True
			if self._text.find("(")>=0 or self._text.find(")")>=0:
				raise TypeError(f"Region '{self.name()}' "
					"no parenthesis are allowed in fixed format")
			nop = 0
			for l in self._text.splitlines():
				# Remove spaces
				l = "".join(l)
				l = l.replace("+", " +")
				l = l.replace("-", " -")
				l = l.replace("|", " | ")
				if nop>0 and expr!="":
					expr += "\n          "
					nop = 0
				orbefore = False
				for op in l.split():
					if op=="|":
						expr += "OR"
						orbefore = True
						continue
					if not orbefore: expr += "  "
					orbefore = False
					sop = str(op).rjust(5)
					if len(sop)>5: sop = sop[0:5]
					expr += sop
					nop  += 1
					if nop==9:
						expr += "\n          "
						nop = 0
			return "%5s%5d%s" % (str(self.name()), neighbors, expr)

	# ----------------------------------------------------------------------
	def _toStr(self, prefix="", fmt=None, **kwargs):
		"""return card as a string with fluka format fmt"""
		# special cards
		ctag = prefix + self.tag
		try:
			if self.tag[0] == "$":
				if self.nwhats()>0:
					return ("%s %s" %
						(ctag, " ".join(
							[str(self.evalWhatOrName(i,True))
								for i in range(0,self.nwhats())]))).rstrip()
				else:
					return ctag
		except IndexError:
			pass

		if self._geo:
			if self.tag == "END":
				return FlukaCard._endStr()
			elif self._type == Card.BODY and self.tag != "VOXELS":
				return self._bodyStr(prefix,fmt)
			elif self._type == Card.REGION:
				return self._regionStr(fmt)

		if fmt==FlukaCard.FORMAT_FREE:
			width = 22
		else:
			width = 10

		# Number of whats
		nwhats = len(self._what)
		if self.tag == "PLOTGEOM":
			#Another card with special treatment
			nwhats = min(6,nwhats)
		elif fmt==FlukaCard.FORMAT_FREE:
			# Count real whats
			while nwhats>0:
				i = nwhats-1
				if self._what[i] == "":
					nwhats = i
				else:
					break

		# Add whats in blocks of 6
		line  = ""
		first = True
		i     = 1

		if prefix:
			ctag = "%s%-10s"%(prefix,self.tag)
		else:
			ctag = self.tag.ljust(10)

		# Cards with no whats
		if nwhats <= 1:
			if line: line += "\n"
			line += ctag
			if nwhats == 1:
				if fmt==FlukaCard.FORMAT_FREE:
					line += (" , "*7) + str(self.evalWhatOrName(0))
				elif fmt is None:
					line += (" "*(width*6)) + str(self.evalWhat(0))
				else:
					line += (" "*(width*6)) + str(self.evalWhatOrName(0))
			line = line.rstrip()
		else:
			# Cards with many whats
			cont_sdum = str(self.evalWhatOrName(0)) \
					if self.tag == "COMPOUND" else " &"

			while i < nwhats:
				if line: line += "\n"
				line += ctag
				if fmt==FlukaCard.FORMAT_FREE:
					line += ", "

				inext = i+6
				imin  = min(inext, nwhats)
				while i<imin:
					if fmt is None:
						w = self.evalWhat(i)
					else:
						w = self.evalWhatOrName(i)
					if isinstance(w, int) and \
					   len(str(w))<=width:	# for RADDECAY
						w = str(w)
					else:
						try:
							float(w)
							w = bmath.format(w, width)
						except:
							if isinstance(w,str) and w and w[0]=="=":
								w = ""
							else:
								w = str(w)[:width]
					if fmt==FlukaCard.FORMAT_FREE:
						line += w + ", "
					else:
						line += w.rjust(width)
					i += 1

				# Check for continuation lines and sdum
				if first:
					if fmt is None:
						sdum = str(self.evalWhat(0))
					else:
						sdum = str(self.evalWhatOrName(0))
				elif i>6:
					sdum = cont_sdum
					if sdum.strip()=="&": cont_sdum += "&"
				first = False

				# Do we need to pad with the line
				if sdum != "":
					if i<inext:
						if fmt==FlukaCard.FORMAT_FREE:
							line += " ,"*(inext-i)
						else:
							line += " "*(width*(inext-i))
					line += sdum
				elif fmt==FlukaCard.FORMAT_FREE:
					# Chop the last comma
					line = line.rstrip()
					if line[-1] == ',':
						line = line[:-1]
				line = line.rstrip()

		# XXX Double format doesn't exist for GEOBEGIN
		# simply triggers the writing of the header inline the input
		# or in an external file
		if self.tag in ("TITLE", "OPEN", "mesh"):
			txt = self._text.rstrip()
			if self.isEvalStr(txt):
				txt = self.evalExpr(txt)
			line += f"\n{txt}"

		elif self.tag == "GEOBEGIN":
			try:    int(self[FlukaCard.IVOPT])
			except:     self[FlukaCard.IVOPT] = 0
			try:    int(self[FlukaCard.IDBG])
			except:     self[FlukaCard.IDBG] = 0
			if fmt==FlukaCard.FORMAT_FREE:
				line += "\n%d,%d,%s" \
					% (self[FlukaCard.IVOPT], self[FlukaCard.IDBG],
					   self._text.rstrip())
			else:
				line += "\n%5d%5d          %s" \
					% (self[FlukaCard.IVOPT], self[FlukaCard.IDBG],
					   self._text.rstrip())

		elif self.tag == "PLOTGEOM":
			w6 = self.intWhat(6)
			if w6 in (0,5):
				line += f"\n{self._text}"
				line += "\n%10s%10s%10s%10s%10s%10s" \
					% (bmath.format(self.numWhat( 7), 10),
					   bmath.format(self.numWhat( 8), 10),
					   bmath.format(self.numWhat( 9), 10),
					   bmath.format(self.numWhat(10), 10),
					   bmath.format(self.numWhat(11), 10),
					   bmath.format(self.numWhat(12), 10) )
				line += "\n%10s%10s%10s%10s%10s%10s" \
					% (bmath.format(self.numWhat(13), 10),
					   bmath.format(self.numWhat(14), 10),
					   bmath.format(self.numWhat(15), 10),
					   bmath.format(self.numWhat(16), 10),
					   bmath.format(self.numWhat(17), 10),
					   bmath.format(self.numWhat(18), 10) )
				line += "\n%10s%10s%10s%10s" \
					% (bmath.format(self.numWhat(19), 10),
					   bmath.format(self.numWhat(20), 10),
					   bmath.format(self.numWhat(21), 10),
					   bmath.format(self.numWhat(22), 10) )
		return line

	# ---------------------------------------------------------------------
	# return string with only the cards
	# ----------------------------------------------------------------------
	def toStr(self, fmt=None, **kwargs):
		if not self.enable and self.info is not CardInfo.NONE \
		   and self.info.disableComment:
			prefix = self.COMMENT
		else:
			prefix = ""
		return self._toStr(prefix, fmt)

	# ----------------------------------------------------------------------
	# Convert to string using default format
	# ----------------------------------------------------------------------
	def __str__(self):
		"""return a card string with the comment"""
		if self._comment:
			return f"{self.commentStr()}\n{self.toStr()}"
		else:
			return self.toStr()

	# ----------------------------------------------------------------------
	# Simplify a single what expression only if it is defined as #define
	#	=VAR  ->  $VAR
	#	=-VAR -> -$VAR
	# ----------------------------------------------------------------------
	def _simplifyExpr(self, w):
		# Single variables convert to $
		pat = _EXPR1VAR.match(w)
		if pat is not None and \
		   pat.group(2) in self.input.cardsCache("#define"):
			if self.tag == "#define":
				return "%s%s"%(pat.group(1),pat.group(2))
			else:
				return "%s$%s"%(pat.group(1),pat.group(2))
		return super()._simplifyExpr(w)

	# ----------------------------------------------------------------------
	def evalWhatStr(self, expr):
		"""return string for special evaluated whats"""
		es = ""
		for i,w in enumerate(self._what):
			# Same as isEvalStr but without the "$" search
			if isinstance(w,str) and w!="" and w[0]=="=":
				if self.tag[0] != "#":
					w = self._simplifyExpr(w)
					if not isinstance(w,str) or w[0]!="=":
						continue
				if expr==FlukaCard.EXPR_OLD:
					es += f"{FlukaCard.EXPR}what.{i}{w}\n"
		if expr==FlukaCard.EXPR_OLD and self.property:
			for n,v in self.property.items():
				if n[0] != '@' and v:	# skip system values
					es += f"{FlukaCard.EXPR}{n}={v}\n"
		return es

	# ----------------------------------------------------------------------
	# Set a list of zones as expression
	# ... removing empty zones
	# ----------------------------------------------------------------------
	def setZones(self, zones):
		if zones:
			self.setText("\n| ".join(
				[x.strip() for x in zones if x.strip()]))
		else:
			self.setText("")

	# ----------------------------------------------------------------------
	# Load voxel information
	# ----------------------------------------------------------------------
	def loadVoxel(self):
		from .voxel import FlukaVoxel
		fn = self.name().strip()
		if fn: self[FlukaCard.VOXEL] = FlukaVoxel(f"{fn}.vxl")
