#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Oct-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re
import math
import tempfile
from collections import OrderedDict, defaultdict

import log
import bmath
import Global
from element import Element
from deck    import Input, Card, CardInfo
from .card   import FlukaCard
from .config import FlukaConfig

#-------------------------------------------------------------------------------
# Constants
BODY_PREFIX    = "B"
REGION_PREFIX  = "R"

#-------------------------------------------------------------------------------
# Patterns
_PRPPAT    = re.compile(r"^(#\w+)\b *([A-Za-z0-9_.@-]*) *(.*)$")
_REGIONPAT = re.compile(r"^\s*([A-Za-z_][A-Za-z0-9_.]*)\s*(-?\d+)\s*(.*)$")
_CWHAT     = re.compile(r"^[*!]@what\.(-?\d+)(=.*)$")
_OPT       = re.compile(r"^!@(\w+)=(.*)$")

#===============================================================================
# Internal Utilities
#===============================================================================
def _str2num(w):
	"""Check if argument w is a valid Fortran number and
	return argument converted to a valid python number"""
	try:
		if len(w)>1:
			# Check for IndexError exception
			"+-.0123456789".index(  w[ 0])
			".0123456789dDeE".index(w[ 1])
			".0123456789".index(    w[-1])
			we = w.replace( "d", "E")
			we = we.replace("D", "E")
			we = we.replace(" ", "")
			# Check for ValueError exception
			wn = float(we)
			# Check for ValueError exception
			wi = int(wn)
			if abs(wi)<10000000000 and float(wi) == wn:
				# Make sure that it was a pure integer
				if "." not in we and "E" not in we:
					return wi
			return we
		return int(w)
	except:
		return w

#-------------------------------------------------------------------------------
def _intWhat(what, n):
	try: return int(float(_str2num(what[n])))
	except: return 0

#-------------------------------------------------------------------------------
def _body2name(name):
	try:
		return BODY_PREFIX + str(int(name))
	except:
		return str(name)

#-------------------------------------------------------------------------------
def _region2name(name):
	try:
		return REGION_PREFIX + str(int(name))
	except:
		pass
	name = str(name)
	if Card.NAMEPAT.match(name): return name
	return REGION_PREFIX + name

#===============================================================================
# Structures as placeholders
#===============================================================================
class LowNeutMaterial:
	def __init__(self):
		self.elem = ""
		self.desc = ""
		self.name = ""
		self.temp =  0
		self.db   = ""
		self.id1  = -1
		self.id2  = -1
		self.id3  = -1
		self.g    =  0

#===============================================================================
# Logical Units
#===============================================================================
class LogicalUnits:
	def __init__(self):
		self.reset()

	# ----------------------------------------------------------------------
	# Reset to default units
	# ----------------------------------------------------------------------
	def reset(self):
		self.list = [None]*100
		# Add some default units:
		self.list[ 1] = (False, "RANDOMIZ",	None)
		self.list[ 2] = (False, "system",	None)
		self.list[ 3] = (False, "system",	None)
		self.list[ 4] = (False, "system",	None)
		self.list[ 5] = (False, "GEOBEGIN",	None)	# stdin
		self.list[ 6] = (False, "stderr",	None)
		self.list[ 7] = (False, "system",	None)
		self.list[ 8] = (False, "system",	None)
		self.list[ 9] = (False, "LOW-NEUT",	None)
		self.list[10] = (False, "system",	None)
		self.list[11] = (False, "GEOBEGIN",	None)
		self.list[12] = (False, "system",	None)
		self.list[13] = (False, "EMFFLUO",	None)
		self.list[14] = (False, "data",		None)
		self.list[15] = (False, "system",	None)
		self.list[16] = (False, "system",	None)
		self.list[17] = (False, "DETECT",	None)
		self.list[18] = (False, "system",	None)
		self.list[19] = (False, "DPMJET",	None)
		self.list[20] = (False, "system",	None)
		#self.list[49] = (True,  "USERDUMP",	None)

	# ----------------------------------------------------------------------
	# Assign a new unit type
	# ----------------------------------------------------------------------
	def assign(self, unit, tag, fn=None):
		"""Assign a new unit to tag"""
		try:
			unit   = int(unit)
			binary = unit<0
			unit   = abs(unit)
			self.list[unit] = (binary, tag, fn)
		except:
			pass

	# ----------------------------------------------------------------------
	def clear(self, unit=None):
		"""Clear a unit from list"""
		if unit is None:
			del self.list[:]
		else:
			unit = abs(unit)
			self.list[unit] = None

	# ----------------------------------------------------------------------
	def empty(self):
		return not bool(self.list)

	# ----------------------------------------------------------------------
	# Return unit type
	# ----------------------------------------------------------------------
	def get(self, u):
		"""Return unit information"""
		try:
			u = abs(int(u))
			return self.list[u]
		except:
			return None

	# ----------------------------------------------------------------------
	# Find first free
	# ----------------------------------------------------------------------
	#def free(self):
	#	"""Return first free unit"""
	#	for i in range(21, len(self.list)):
	#		u = self.list[i]
	#		if u is None: return i

	#-----------------------------------------------------------------------
	# FIXME needs optimisation, we should find all cards using units in the
	# CardInfo and search only those cards
	#-----------------------------------------------------------------------
	def scan(self, input_):
		self.reset()
		for card in input_.cardlist:
			if card.ignore(): continue
			# Ignore cards that use only as readonly SPECSOUR, OPEN is special
			if card.tag in ("OPEN","SPECSOUR"): continue
			for u in card.fileunits():
				if u == 0: continue
				unit = self.get(u)
				# check for a possible conflict!
				if unit is not None:
					(b, c, _) = unit
					if c!=card.tag or (b ^ (int(u)<0)):
						# Find what with units
						case = card.case()
						card.invalid = card.info.findType("lu", case)
						log.warning(f"Possible conflict with unit: {u}")
						log.warningCont("is used by: %s card #%d" %
							(str(self.get(u)[1]),card.pos1()))
						log.warningCont(card)
						break
				self.assign(u, card.tag)

		# For all open cards assign the correct name
		for card in input_["OPEN"]:
			unit = card.intWhat(1)
			u    = self.list[abs(unit)]
			if u is None:
				self.assign(unit, "OPEN", card.text())
			else:
				self.assign(unit, u[1], card.text())

	# ----------------------------------------------------------------------
	# Return a list of all units used
	# ----------------------------------------------------------------------
	def filterList(self, tag=None, binary=True, onlymatches=False):
		"""Return a list of all units, filtered for a card"""
		lst = [("", "")]
		for i in range(1, 100):
			u = self.list[i]

			if u is not None and tag is not None:
				if u[1]!=tag: continue
			elif u is None and onlymatches:
				continue

			if binary:
				lst.append((f"{i:02d} ASC",  i))
				lst.append((f"{i:02d} BIN", -i))
			else:
				lst.append((f"{i:02d}", i))
		lst.append(("0", 0))
		return lst

#===============================================================================
# Fluka Input class
#===============================================================================
class Fluka(Input):
	MODE = FlukaCard.MODE

	_INDENT_INC = Input._INDENT_INC | {"$start_expansion", "$start_translat",
				"$start_transform", "GEOBEGIN", "END"}
	_INDENT_DEC = Input._INDENT_DEC | {"$end_expansion", "$end_translat",
				"$end_transform", "GEOEND", "END"}

	config  = FlukaConfig()			# configuration information

	defaultMaterials = OrderedDict()	# Default material list (cards)
	icruMaterials    = OrderedDict()	# ICRU special compounds (cards)
	neutronGroups    = {}			# Neutron energy groups as numbers
	neutronGroupsS   = {}			# Neutron energy groups as strings with units
	lowMaterials     = {}			# LOW-NEUT material cross sections
	NGROUPS          = []			# Neutron groups

	warnMultMat = True			# Warn user on multiple materials
	_VOXELPAT   = re.compile(r"^VOX[E]?[L]?(\d+)$")
	_ROTPAT     = re.compile(r"^ROT?#(\d+)$",re.I)

	# ----------------------------------------------------------------------
	def __init__(self, filename=None):
		super().__init__()
		self.mode       = Fluka.MODE
		self.globalDict = FlukaCard.globalDict
		self.fileunits  = LogicalUnits()		# Logical Units

		# I/O variables
		self.format      = FlukaCard.FORMAT_SINGLE
		self.geoFormat   = FlukaCard.FORMAT_SINGLE
		self._exprFmt    = FlukaCard.EXPR_OLD
		self._cwhat      = []		# Commented special whats
		self._properties = []		# Commented card properties
		self._nwhats     = 0		# Number of whats of last added card
		self._ivopt      = 0
		self._idbg       = 0

		# state of card read during parsing of input file
		self._cardDisabled = 0		# how it is disabled:
						# 0 = enabled  (00)
						# 1 = comment  (01)
						# 2 = #if 0    (10)
						# 3 = both     (11)
		if filename: self.load(filename)

	#-----------------------------------------------------------------------
	# Return input filename
	#-----------------------------------------------------------------------
	def filenames(self, filename):
		files = super().filenames(filename)
		files.extend([card[FlukaCard.GEOFILE] for card in self["GEOBEGIN"]
				if card[FlukaCard.GEOFILE]])
		return files

	#-----------------------------------------------------------------------
	# #include card has been modified take appropriate action
	# @return	True if #include has modified the input
	#		False if nothing happened
	# FIXME Should be moved to InputPage with proper undo/redo
	#-----------------------------------------------------------------------
	#def include(self, includecard):
	#	super().include(includecard)

	#	# Check if #include contains already imported cards
	#	fromid = self.cardlist.index(includecard)
	#	# Finding ending of #include
	#	toid   = fromid+1
	#	level  = 1
	#	while toid<len(self.cardlist):
	#		card = self.cardlist[toid]
	#		if card.tag == "#include":
	#			level += 1
	#		elif card.tag == "#endinclude":
	#			card.setEnable(includecard.enable)
	#			level -= 1
	#			if level==0: break
	#		toid += 1

	#	# Check for disabling the #include card
	#	if not includecard.enable:
	#		# Remove all intermediate cards
	#		for i in range(toid-1, fromid, -1):
	#			self.delCard(i, False)
	#		self.renumber(fromid)
	#		return False

	#	# If range is not empty then return
	#	if toid-fromid>1: return False

	#	# Create a temporary input to load the included lines
	#	fluka = Fluka()
	#	location = 0	# 0=input, 1=bodies, 2=regions
	#	for i in range(fromid):
	#		card = self.cardlist[i]
	#		if not card.enable: continue
	#		if card.tag == "FREE":
	#			fluka.format = FlukaCard.FORMAT_FREE

	#		elif card.tag == "FIXED":
	#			fluka.format = FlukaCard.FORMAT_SINGLE

	#		elif card.tag == "GLOBAL":
	#			w4 = card.intWhat(4)
	#			if w4 in (2,3):
	#				fluka.format= FlukaCard.FORMAT_FREE
	#			if card.intWhat(5)>0:
	#				fluka.geoFormat = FlukaCard.FORMAT_FREE

	#		elif card.tag == "GEOBEGIN":
	#			if card.sdum()[:8]=="COMBNAME":
	#				fluka.geoFormat = FlukaCard.FORMAT_FREE
	#			elif card.sdum()[:8]=="COMBINAT":
	#				fluka.geoFormat = FlukaCard.FORMAT_SINGLE
	#			location += 1

	#		elif card.tag == "END":
	#			location += 1

	#		elif card.tag == "GEOEND":
	#			location = 0

	#	try:	# Check if file exist
	#		os.stat(includecard.name())
	#	except OSError:
	#		# File not exist do nothing
	#		return False

	#	fluka.open(includecard.name(),"r")
	#	if location==0:
	#		fluka._load()
	#	elif fluka==1:
	#		fluka._parseBodies()
	#	elif location==2:
	#		fluka._parseRegions()
	#	elif location==3:
	#		fluka._parseVolume()
	#	else:
	#		assert False
	#	assert not fluka._files

	#	pos = fromid+1
	#	for card in fluka.cardlist:
	#		card.input = self
	#		self.add(card, pos)
	#		pos += 1

	#	self.clearUnits()
	#	self.renumber(fromid)
	#	self.modified()
	#	return True

	#=======================================================================
	# Utilities
	#=======================================================================
	#-----------------------------------------------------------------------
	# Scan for units
	#-----------------------------------------------------------------------
	def scanUnits(self):
		self.fileunits.scan(self)

	#-----------------------------------------------------------------------
	def clearUnits(self):
		self.fileunits.clear()

	# ----------------------------------------------------------------------
	# Create bare minimum input file
	# ----------------------------------------------------------------------
	def minimumInput(self):
		super().minimumInput()
		self.add(FlukaCard("GEOBEGIN",["COMBNAME"]))
		self.add(FlukaCard("END"))
		self.add(FlukaCard("END"))
		self.add(FlukaCard("GEOEND"))
		self.add(FlukaCard("STOP"))
		self.renumber()

	#-----------------------------------------------------------------------
	# Cards
	#-----------------------------------------------------------------------
	# Return cards of a specific tag sorted with _pos
	# @param tag     string tag
	#                list of tags
	#                Card.type(), BODY, SOLID, VOLUME...
	# @param which   look the Input.cardsSorted()
	# WARNING on REGION maybe the "&" you want to get rid off
	#-----------------------------------------------------------------------
	def cardsSorted(self, tag, which=0):
		cards = super().cardsSorted(tag, which)

		# Special treatment
		if tag == Card.BODY:
			# Append the only VOXEL card if exists
			for card in self["VOXELS"]:
				if card.notIgnore():
					cards["VOXEL"] = card
					break

		elif tag in (Card.REGION, Card.MATERIALS,
				"REGION", "ASSIGNMA", "MATERIAL", "COMPOUND", "CORRFACT"):
			if isinstance(tag,int):
				if tag == Card.MATERIALS:
					tag = CardInfo.materials[self.MODE]
				elif tag == Card.REGION:
					tag = CardInfo.volumes[self.MODE]
			if tag == "MATERIAL":
				for card in self.defaultMaterials.values():
					# Warning on positioning!!!!!
					if card.name() not in cards:
						cards[card.pos() if which<0 else card.name()] = card

				# FIXME Must append at the end ONLY the ICRU that are assigned
				for card in self.cardsSorted("ASSIGNMA", -1).values():
					name = card.what(1)
					if name not in cards:
						try:
							cards[card.pos() if which<0 else name] = Fluka.icruMaterials[name]
						except KeyError:
							pass
			elif tag in (Card.REGION, "REGION"):
				# delete all continuation regions
				if which>=0:
					try:
						del cards["&"]
					except KeyError:
						pass

			# Append/Insert voxel cards
			pos = len(self.cardlist)
			for voxel in self["VOXELS"]:
				if voxel.ignore() or voxel[FlukaCard.VOXEL] is None: continue
				if tag in (Card.REGION, "REGION"):
					# add to the end
					voxelregions = voxel[FlukaCard.VOXEL]["REGION"]
					for x in voxelregions:
						x[Card.PARENT] = voxel
					cards.update([(x.name(),x) for x in voxelregions])
				else:
					# insert to the beginning
					ocards = cards
					if which<0:
						cards = OrderedDict([(pos+x.pos(),x)
								for x in voxel[FlukaCard.VOXEL][tag]])
					else:
						cards = OrderedDict([(x.name(),x)
								for x in voxel[FlukaCard.VOXEL][tag]])
					cards.update(ocards)
				pos += len(voxel[FlukaCard.VOXEL].cardlist)
		return cards

	# ----------------------------------------------------------------------
	# Return an extend list with all cards including the ones from the voxel
	# ----------------------------------------------------------------------
	def allcards(self):
		clist = super().allcards()
		for voxel in self["VOXELS"]:
			if voxel.ignore() or voxel[FlukaCard.VOXEL] is None: continue
			clist = clist[:]
			clist.extend(voxel[FlukaCard.VOXEL].cardlist)
			break	# treat only the first VOXELS card
		return clist

	# ----------------------------------------------------------------------
	# Refresh cards information
	# 1. ?? (renumber??)
	# 2. Reload voxels
	# 3. Properties?
	# ----------------------------------------------------------------------
	def refresh(self):
		for card in self["VOXELS"]:
			card.loadVoxel()

	#-----------------------------------------------------------------------
	# Return the body properties as a dictionary of the active/enabled
	# bodies
	#-----------------------------------------------------------------------
	def bodyProperties(self, rotdefi):
		translat   = None
		transform  = None
		transform2 = None
		expansion  = None
		matrix = [bmath.Matrix(4,type=1)]	# list of transformation matrices

		for card in self.cardlist:
			if not card.isGeo(): continue
			if card.ignore(): continue

			if len(card.tag)==3:
				if card.tag=="END": break
				if translat:
					card["@dx"] = translat.numWhat(0)
					card["@dy"] = translat.numWhat(1)
					card["@dz"] = translat.numWhat(2)
				else:
					del card["@dx"]
					del card["@dy"]
					del card["@dz"]
				if expansion:
					card["@expansion"]  = expansion.numWhat(0)
				else:
					del card["@expansion"]
				if transform:
					card["@transform"]  = transform.what(0)
				else:
					del card["@transform"]
				if transform2:
					card["@transform2"] = transform2.what(0)
				else:
					del card["@transform2"]
				if len(matrix)>1:
					card[Card.MATRIX] = matrix[-1]
				else:
					del card[Card.MATRIX]

			elif card.tag[0]=="$":
				if card.tag=="$start_translat":
					if translat is not None:
						log.error(f"Card #{card.pos1()} double '{card.tag}' found")
						log.errorCont(card)
					else:
						matrix.append(matrix[-1]*bmath.Matrix.translate(
							card.numWhat(0), card.numWhat(1), card.numWhat(2)))
						translat = card
				elif card.tag=="$start_transform":
					if translat is not None:
						log.error(f"Card #{translat.pos1()} "
							f"'{card.tag}' must be the toplevel transformation")
					else:
						if transform is not None: # and transform2 is not None:
							log.warning(f"Card #{card.pos1()} multiple "
								f"'{card.tag}' found")
							log.warningCont(card)
						# Fluka applies $start_translat before $start_transform
						#matrix.append(inp.getTransformation(
						#		card.evalWhat(0),rotdefi)*matrix[-1])
						# Correct mathematical approach
						matrix.append(matrix[-1]*self.getTransformation(
								card.evalWhat(0),rotdefi))
						if transform is None:
							transform = card
						else:
							transform2 = card
				elif card.tag=="$start_expansion":
					if expansion is not None:
						log.error(f"Card #{card.pos1()} double '{card.tag}' found")
						log.errorCont(card)
					else:
						matrix.append(matrix[-1]*bmath.Matrix.scale(card.numWhat(0)))
						expansion = card
				elif card.tag=="$end_translat":
					if translat is None:
						log.error(f"Card #{card.pos1()} '{card.tag}' "
							"without $start_xxx")
						log.errorCont(card)
					translat = None
					if len(matrix)>1: matrix.pop()
				elif card.tag=="$end_transform":
					if transform is None:
						log.error(f"Card #{card.pos1()} '{card.tag}' "
							"without $start_xxx")
						log.errorCont(card)
					if transform2 is not None:
						transform2 = None
					else:
						transform = None
					if len(matrix)>1: matrix.pop()
				elif card.tag=="$end_expansion":
					if expansion is None:
						log.error(f"Card #{card.pos1()} '{card.tag}' "
							"without $start_xxx")
						log.errorCont(card)
					expansion = None
					if len(matrix)>1: matrix.pop()

	#-----------------------------------------------------------------------
	# Return the region properties as a dictionary of the active/enabled
	# regions
	#-----------------------------------------------------------------------
	def regionProperties(self):
		matList = self.materialList()
		rotDefi = {}

		# Create a material dictionary including default materials
		matDict = self.defaultMaterials.copy()
		matDict.update(Fluka.icruMaterials)

		# ... finally materials in voxels and input
		for card in self.cardsSorted(Card.MATERIALS).values():
			try:
				card["@n"] = matList.index(card.name())+1
			except ValueError:
				card["@n"] = 1
			matDict[card.name()] = card

		# Find regions
		blckhole = matDict["BLCKHOLE"]
		vacuum   = matDict["VACUUM"]

		#---------------------------------------------------------------
		# Build dictionaries and clean up regions from Input and voxel
		n = 0
		regionList = []
		regionDict = self.cardsSorted("REGION")
		for region in regionDict.values():
			if region.name() == "&": continue
			n += 1
			region["@n"] = n		# Assign default values to region
			region[Card.ASSIGNMAT]= None	# Card that assigned the material
			region[Card.MATERIAL]= None	# no material assigned
			region["@biasingA"]  = None	# All particles biasing
			region["@biasingH"]  = None	# Hadrons
			region["@biasingE"]  = None	# emf
			region["@biasingN"]  = None	# low-neutrons
			region[Card.LATTICE] = None
			region["@rotdefi"]   = ""
			region["@rotdefi2"]  = ""
			region[Card.CONT]    = False	# has continuation cards
			regionList.append(region)

		# add last region
		if regionList: regionDict["@LASTREG"] = regionList[-1]

		# Find transformations
		for card in self.cardsSorted("ROT-DEFI",-1).values():
			w1 = card.intWhat(1)
			if w1 < 1000:
				(_,i) = divmod(w1, 100)
			else:
				(i,_) = divmod(w1, 1000)

			rotDefi[i] = card.name()
			rotDefi[card.name()] = i

		# Check all assignmat
		for card in self.cardsSorted("ASSIGNMA",-1).values():
			material = matDict.get(card.evalWhat(1))
			if material is None:
				log.error(f"Card #{card.pos1()} material "
					f"{card.evalWhat(1)} is not defined")
				log.errorCont(card)
				material = self.defaultMaterials["BLCKHOLE"]

			matDecay = matDict.get(card.what(6), material)

			fromReg = regionDict.get(card.evalWhat(2))
			if fromReg is None:
				pat = Fluka._VOXELPAT.match(card.what(2))
				if pat is None:
					log.error(f"Card #{card.pos1()} "
						f"Region '{card.what(2)}' is not defined")
					log.errorCont(card)
				continue
			fromReg = fromReg["@n"]

			try:
				toReg = regionDict.get(card.evalWhat(3))["@n"]
			except:
				toReg = fromReg

			step = card.intWhat(4)
			if step == 0: step = 1

			for rg in range(fromReg, toReg+1, step):
				region = regionList[rg-1]
				if "VOXELS" in self.cards and \
				  (region.name()=="VOXEL" or Fluka._VOXELPAT.match(region.name())):
					region[Card.TYPE] = Card.REGION_VOXEL
				else:
					region[Card.TYPE] = int(material is blckhole)
				if region[Card.ASSIGNMAT] is not None and \
				   Fluka.warnMultMat and \
				   material.name() not in ("BLCKHOLE", "VACUUM"):
					log.warning(f"Multiple ASSIGNMAT for region {region.name()}:")
					log.warningCont("Previous card #%d\n%s"%
						(region[Card.ASSIGNMAT].pos1(), region[Card.ASSIGNMAT]))
					log.warningCont("Current card #%d\n%s\n"%
						(card.pos1(), str(card)))
				region[Card.ASSIGNMAT]= card
				region[Card.MATERIAL] = material
				region["@matDecay"]   = matDecay

		# Check Biasing cards
		for card in self.cardsSorted("BIASING",-1).values():
			what1 = card.intWhat(1)
			if what1<0: continue
			if what1 == 1:
				tag = "@biasingH"
			elif what1 == 2:
				tag = "@biasingE"
			elif what1 == 3:
				tag = "@biasingN"
			else:
				tag = "@biasingA"

			try:
				fromReg = regionDict.get(card.evalWhat(4))["@n"]
			except:
				continue
			try:
				toReg = regionDict.get(card.evalWhat(5))["@n"]
			except:
				toReg = fromReg
			try:    step = int(card.what(6))
			except: step = 1
			for rg in range(fromReg, toReg+1, step):
				region = regionList[rg-1]
				region[tag] = card

		# Check Lattices
		for card in self.cardsSorted(("LATTICE","LATTSNGL"),-1).values():
			fromReg  = regionDict.get(card.evalWhat(1))
			if fromReg is None: continue
			fromReg  = fromReg["@n"]
			if card.tag == "LATTICE":
				try:
					toReg = regionDict.get(card.evalWhat(2))
					toReg = toReg["@n"]
				except:
					toReg = fromReg
				try: step = int(card.what(3))
				except: step = 1
				rotdefi  = card.name()
				rotdefi2 = ""
			else:
				toReg = fromReg
				rotdefi  = card.what(2)
				rotdefi2 = card.what(3)

			for rg in range(fromReg, toReg+1, step):
				region = regionList[rg-1]
				region[Card.TYPE]     = Card.REGION_LATTICE
				region[Card.MATERIAL] = vacuum
				region[Card.LATTICE]  = card
				region["@rotdefi"]    = rotdefi
				region["@rotdefi2"]   = rotdefi2

		# Verify if all regions have material assigned
		# FIXME double checking also in GeometryViewer
		#for region in regionList:
		#	if region[Card.MATERIAL] is None:
		#		log.warning(f"Region {region.name()} not assigned any material")

	#-----------------------------------------------------------------------
	# Convert material to name/number
	#-----------------------------------------------------------------------
	def material(self, mat, toName):
		lst = self.materialList(0,True)
		if toName:
			mat -= 1		# 1 based
			if mat<0 or mat>=len(lst):
				if mat != -1:	# Accept -1 == 0 index
					raise Exception(f"Invalid material index requested idx={mat+1}")
				return ""
			return lst[mat]
		else:
			try:
				return lst.index(mat)+1
			except:
				return 0

	#-----------------------------------------------------------------------
	# Return material list
	#-----------------------------------------------------------------------
	def materialList(self, which=0, extended=False, *kw):
		lst = super().materialList(which)

		# Add first voxel materials
		# FIXME maybe apply it on the card and not on the Input!!!!
		for voxel in self["VOXELS"]:
			if voxel.ignore() or voxel[FlukaCard.VOXEL] is None: continue
			lst.extend([m.name() for m in voxel[FlukaCard.VOXEL]["MATERIAL"]])

		# Add user defined materials
		for card in self.cardsSorted(Card.MATERIALS, which).values():
			# Check for possible icru material defined as alternate dE/dx material
			if extended:
				altmat = card.what(5)
				if altmat and altmat not in lst and altmat in Fluka.icruMaterials:
					# we need to insert it before creating the current material
					lst.append(altmat)

			# process current material
			name  = card.name()
			index = card.intWhat(4)-1	# 0 based!
			# Normal material
			if index<0:
				try:    index = lst.index(name)
				except: index = len(lst)
			# Append or replace
			if index == len(lst):
				lst.append(name)
			elif index < len(lst):
				if lst[index] != "" and lst[index] != name:
					log.warning( "overriding material "
						f"index {index}={lst[index]} by {name}")
				lst[index] = name
			else:
				lst.extend([""] * (index-len(lst)+1))
				lst[index] = name

		# Add ICRU materials
		if extended:
			# Add all icru materials regardless of their order
			for mat in Fluka.icruMaterials.keys():
				if mat not in lst:
					lst.append(mat)
		else:
			# Add only the assigned ones
			# correct index assigned by FLUKA
			# FIXME Not very intelligent.
			for card in self.cardsSorted("ASSIGNMA", which).values():
				mat = card.what(1)
				if mat not in lst:
					lst.append(mat)
		return lst

	#-----------------------------------------------------------------------
	# Return material dictionary
	#-----------------------------------------------------------------------
	def materialDict(self):
		matDict = super().materialDict()
		matDict.update(Fluka.icruMaterials)

		# Voxel materials
		for voxel in self["VOXELS"]:
			if voxel.ignore() or voxel[FlukaCard.VOXEL] is None: continue
			for card in voxel[FlukaCard.VOXEL]["MATERIAL"]:
				matDict[card.name()] = card

		# ... and finally from the input
		for card in self.cardsSorted(Card.MATERIALS).values():
			matDict[card.name()] = card

		return matDict

	#-----------------------------------------------------------------------
	# material ZAID description
	# ftype: fraction type 0=atom, 1=mass, -1=original
	# @return 'atom' fraction composition. Normalized to 1.0 if norm is True
	#
	# FIXME move to fluka2->mcnp conversion
	#-----------------------------------------------------------------------
	def materialZAID(self, material, norm=False, ftype=0):
		# Check if it is a compound or a single material
		# Single isotope?
		Z  = material.intWhat(1)
		Am = material.numWhat(2)
		A  = material.intWhat(6)
		if Z>0:
			if Am==0.0:
				try:
					element = Element.byZ(Z)
					Am = element.Amass()
				except IndexError:
					Am = A
			return {Z*1000+A : [1.0, Am]}

		rho = material.numWhat(3)		# Density

		# Compound search
		ZAID = {}

		# FIXME could use the cardsSorted as dictionary
		# to avoid the search
		for compound in self["COMPOUND"]:
			if compound.ignore(): continue
			if compound.name() != material.name(): continue

			for i in range(1,len(compound.whats()),2):
				frac = compound.numWhat(i)
				mat  = compound.what(i+1)
				if mat == "": continue
				if mat[0]=="-":
					volume = True
					mat = mat[1:]
				else:
					volume = False

				# if we should return the original type
				# and compound fraction is in mass set ftype to 1
				if frac<0.0 and ftype<0:
					ftype=1

				# Find material
				for card in self["MATERIAL"]:
					if card.ignore(): continue
					if card.name() == mat: break
				else:
					# Search default materials
					for card in self.defaultMaterials.values():
						if card.name() == mat: break

				# Scan recursively
				zaid  = self.materialZAID(card)
				total = sum([f for f,am in zaid.values()])
				aw    = sum([f*am for f,am in zaid.values()])/total

				for za,[f,am] in zaid.items():
					f /= total
					# Scale fractions
					if volume:
						# From volume fraction
						f *= frac * rho / aw
						assert False	# Not tested yet!
					elif frac<0.0:
						# From mass fraction
						f *= -frac / aw
					else:
						# From atom fraction
						f *= frac

					try:
						ZAID[za][0] += f
					except (KeyError, IndexError):
						ZAID[za] = [f, am]
					#ZAID.append((za,f))

		if ftype==1:
			# convert to mass fraction
			for value in ZAID.values():
				value[0] *= -value[1]

		if norm:
			# normalize to 1.0 after we processed all compound cards
			total = sum([f for f,am in ZAID.values()])
			for value in ZAID.values():
				value[0] /= total
		return ZAID

	#-----------------------------------------------------------------------
	# @param idx		rotation to return
	#			Can be prefixed with "-"
	#			None to return a dictionary
	# @param rotdefi	cached dictionary
	# @param inv		return the inverse of the idx matrix.
	#			Accepted even if idx is prefixed with "-"
	# @return rotation matrix from a rot-defi cards or a dictionary with all
	#         transformations
	# FIXME I should combine the rotdefi with the self.cache!
	#-----------------------------------------------------------------------
	def getTransformation(self, idx=None, rotdefi=None, inv=False):
		neg = False
		if isinstance(idx,str):
			if idx and idx[0]=='-':
				neg = True
				idx = idx[1:]
			match = Fluka._ROTPAT.match(idx)
			if match:
				idx = int(match.group(1))
		elif isinstance(idx,int):
			if idx<0:
				neg = True
				idx = -idx
		if inv: neg = not neg

		# return value if cached
		if idx is not None:
			if rotdefi is not None:
				matrix = rotdefi.get(idx)
				if isinstance(matrix,int): matrix = rotdefi.get(matrix)
				if matrix is None:
					return bmath.Matrix(4, type=1)
				elif neg:
					matrix = matrix.inverse()
				return matrix
			matrix = bmath.Matrix(4, type=1)

		# ------------------------------------------------------
		# Scan transformations cards
		# ------------------------------------------------------
		rotdefi = {}
		last = 0
		for card in self.cardsSorted("ROT-DEFI",-1).values():
			w1 = card.intWhat(1)
			if w1 < 1000:
				(j, i) = divmod(w1, 100)
			else:
				(i, j) = divmod(w1, 1000)

			if idx is None:
				if i==0: # try with the name
					i = rotdefi.get(card.name())
					matrix = rotdefi.get(i)
				else:
					matrix = rotdefi.get(i)
					if matrix is not None and rotdefi.get(card.name())!=i:
						# FIXME store the error somewhere...
						log.error(f"Card #{card.pos1()} ROT-DEFI "
							"Index and name do not match")
						log.errorCont(card)

				if matrix is None:
					matrix = bmath.Matrix(4, type=1)
					if i is None:
						last += 1
						i = last
					else:
						last = max(last,i)
					rotdefi[i] = None
					if card.name()!="": rotdefi[card.name()] = i
					rotdefi[f"rot#{i:03d}"] = i

			else:
				if i!=idx and card.name()!=idx: continue

			if j==0: j=3
			theta = card.numWhat(2)
			phi   = card.numWhat(3)
			x = card.numWhat(4)
			y = card.numWhat(5)
			z = card.numWhat(6)

			if x!=0.0 or y!=0.0 or z!=0.0:
				matrix = bmath.Matrix.translate(x,y,z)*matrix

			if phi!=0.0:
				m = bmath.Matrix(4,type=1)
				m.rotate(math.radians(-phi), j-1)
				matrix = m*matrix

			if theta!=0.0:
				m = bmath.Matrix(4,type=1)
				m.rotate(math.radians(-theta), (j+1)%3)
				matrix = m*matrix

			if idx is None:
				rotdefi[i] = matrix

		if neg:
			matrix = matrix.inverse()

		if idx is None:
			return rotdefi
		return matrix

	# ----------------------------------------------------------------------
	# Rotation matrix list
	# ----------------------------------------------------------------------
	def rotationList(self, negative=False):
		rot = {}
		try:
			for card in self.cards["ROT-DEFI"]:
				w1 = card.intWhat(1)
				if w1 < 1000:
					(_,i) = divmod(w1, 100)
				else:
					(i,_) = divmod(w1, 1000)
				if i>0:
					rot[f"rot#{i:03d}"] = True
				if card.name():
					rot[card.name()] = True
		except KeyError:
			return [""]

		l = sorted(rot)
		if negative:
			lst = [""]
			for i in l:
				lst.append(i)
				lst.append(f"-{i}")
		else:
			lst = l
			lst.insert(0,"")

		del rot
		return lst

	#-----------------------------------------------------------------------
	# Convert input to names and check for obsolete and/or non-valid cards
	#-----------------------------------------------------------------------
	def convert2Names(self):
		# Check geometry for FLUGG
		try:
			geobegin = self.cards["GEOBEGIN"][0]
			notflugg = geobegin.sdum()!="FLUGG"
		except KeyError:
			notflugg = True

		matDict = {}
		for i,n in enumerate(self.materialList(0,True)):
			if n not in matDict:
				matDict[n] = [i+1]
			else:
				matDict[n].append(i+1)

		for card in self.cardlist:
			if not card.enable: continue

			# Check for obsolete cards
			tag = card.tag
			if tag in ("ACCURACY", "OUTLEVEL", "EGSCUT",
				   "EGSFIX",   "EGSFLUO",  "EGSRAY",
				   "EGS",      "LANDAU"):
				if card.enable:
					log.warning(f"Obsolete card {tag}")
				continue
			if tag=="EXTRAWEI":
				log.warning(f"Changing: {tag} to USERWEIG")
				self.changeAllTags(tag, "USERWEIG")
			elif tag=="DUMPTHEM":
				log.warning(f"Changing: {tag} to USERDUMP")
				self.changeAllTags(tag, "USERDUMP")
			elif tag=="MATERIAL":
				index = matDict.get(card.sdum(),[])
				if len(index)>1 and card.intWhat(4) in (0, index[0]):
					log.warning(f"Material {card.name()} already exists with index={index}")
					log.errorCont(card)

			# Check input
			try:	case  = card.case()
			except:
				log.error(f"Card #{card.pos1()} Unknown card {card.tag!r}")
				log.errorCont(card)
				continue

			if card.info is CardInfo.NONE:
				log.error(f"Card #{card.pos1()} Invalid tag {card.tag!r}")
				log.errorCont(card)
				card.invalid = ["Invalid tag name"]
				card.enable = False
				continue

			if card.notIgnore():
				invalid = card.validate(case)
				if invalid:
					if invalid[0] is None:
						log.warning(f"Card #{card.pos1()} No matching case found")
					else:
						log.warning(f"Invalid what(s)={invalid}")
					log.warningCont(card)

			# Convert card to names
			if notflugg: card.convert(True)

	#=======================================================================
	# I/O
	# methods for Input IO operations
	#=======================================================================
	#-----------------------------------------------------------------------
	# Write input, detect by file extension the type
	#-----------------------------------------------------------------------
	def save(self, filename, backup=None, expr=FlukaCard.EXPR_OLD):
		self._exprFmt = expr
		super().save(filename, backup)

	#-----------------------------------------------------------------------
	# Close topmost file
	#-----------------------------------------------------------------------
	def close(self):
		# end of file?
		if self._type[-1] == "#include":
			if self._cardDisabled & 1:
				self._cardDisabled -= 1
			card = Card("#endinclude", comment=self._comment)
			self._addCard(card)
			if self._files[-1] is None:
				card.disable()
		super().close()

	#-----------------------------------------------------------------------
	# FIXME override READLINE for regions
	# since the region should be added at the end.
	# It should be corrected and add the region immediately
	# and simply append the comments and the expression to the last card
	# if needed
	#-----------------------------------------------------------------------
	def readlineRegion(self):
		while True:
			try:
				self._lineNo[-1] += 1
				line = self._files[-1].readline()
			except IndexError:
				return None
			except AttributeError:
				line = ""	# from None file

			if self._md5cur is not None:
				self._md5cur.update(line.encode())

			if not line:
				# end of file?
				if self._type[-1] == "#include":
					return "#endinclude"
				self.close()
				if not self._files: return None
				continue

			return line.rstrip()

	#-----------------------------------------------------------------------
	# Used internally while parsing to add a card
	# @see Card.appendWhats() and Card.checkSum
	#-----------------------------------------------------------------------
	def _addCard(self, card):
		self._nwhats = card.nwhats()

		# Replace special whats with card whats
		chksum = 0
		for i,w in enumerate(self._cwhat):
			if w:	# Calculate chksum
				try:
					what = bmath.format(round(float(card._what[i]),10),10)
					chksum ^= hash(what)
					#Global.dump("_addCard CHKSUM",card.tag,card.name(),i,
					#		repr(card._what[i]), repr(what), w, chksum)
				except (IndexError,ValueError):
					pass
				# Set new what
				card.setWhat(i,w)
		if chksum:
			card[Card.CHKSUM] = chksum

		# add properties if any
		for n,v in self._properties:
			card[n] = v

		self.add(card)
		if self._cardDisabled: card.disable()

		# Reset comment and special whats
		self._comment = []
		del self._cwhat[:]
		del self._properties[:]

	#-----------------------------------------------------------------------
	# Parse a comment line
	#-----------------------------------------------------------------------
	def _parseComment(self, line, check4Card=True):
		line = line.rstrip()
		if not line: return None

		#---------------------------------------------------------------
		def addCWhat(w,value):
			if w<0: return
			if w==len(self._cwhat):
				self._cwhat.append(value)
			else:
				if w>len(self._cwhat):
					self._cwhat.extend([None]*(w-len(self._cwhat)+1))
				self._cwhat[w] = value

		if self._cardDisabled & 1:
			self._cardDisabled -= 1

		if line[0] in "!*":
			# if next character is not space
			# check if it is a valid card
			#cstart = 1
			if len(line)>1:
				#if line[1]==" ": cstart = 2

				#if line[1:4] in ("...", "234", "===", "---",
				#			"***", "___", "_..", "+++", "_AA" ):
				#	cstart = 2

				if line[0]=="!" and line[1:9].strip() in CardInfo.FLAIR_TAGS:
					# remove the ! and replace it with space at the end
					card = line[1:9].strip()
					return f"{card:10}{line[10:]}"

				elif line[:2] == FlukaCard.EXPR or \
				     line[1:].startswith("@what."):
					# treat special commented whats
					m = _CWHAT.match(line)
					if m:
						try:
							addCWhat(int(m.group(1)), m.group(2))
							return None
						except ValueError:
							pass
					if line[0]=="!":
						m = _OPT.match(line)
						if m:
							self._properties.append((m.group(1),m.group(2)))
							return None

				elif check4Card and (line[1] in "#$/" or line[1].isalpha()):
					# Check for commented cards
					try:
						lline = line[1:]
						tag = lline.split(maxsplit=1)[0]
						# Check tag if it exists in CardInfo
						ci = CardInfo.get(self.MODE, tag)
						if ci.tag == tag:
							# Commented card found
							self._cardDisabled += 1
							return lline
					except:
						pass

			cline = line[1:]
			if cline=="": cline=" "
			self._comment.append(cline)
			return None

		# Remove inline comments and check for expressions
		p = line.find("!")
		if p>0:
			try:
				if line[p+1]=='=':	# inline expressions
					addCWhat(1, line[p+1:])
				else:
					self._comment.append(line[p+1:])
			except IndexError:
				pass
			line = line[0:p]

		if line[0] == "#":
			if line=="#if 0" and not self._cardDisabled:
				# FIXME multiple #if 0 #if 0
				self._cardDisabled += 2
			elif line=="#endif" and self._cardDisabled&2:
				self._cardDisabled -= 2
			else:
				return line
			return None

		return line

	#-----------------------------------------------------------------------
	# Parse a single line to card, whats and sdum
	#-----------------------------------------------------------------------
	def _parseLine(self, line, enable=True):
		if line[0] == "$":
			what = Fluka._parseFreeFormat(line)
			tag = what.pop(0).lower()

		elif line[0] == "#":
			m = _PRPPAT.match(line)
			if m:
				tag = m.group(1)
				if tag == "#include":
					# replace the whats
					what = [line[9:].strip()]
					if enable:
						if self._cardDisabled:
							self.open(None,"r",tag)
						else:
							self.open(what[0],"r",tag)
				else:
					what = [m.group(2), m.group(3)]
			else:
				tag = line.strip()
				what = [""]

		elif self.format == FlukaCard.FORMAT_SINGLE:
			tag = line[0:8].rstrip()
			what = [line[70:].strip()]	# sdum
			for i in range(10, 61, 10):
				w = line[i:i+10].strip()
				what.append(_str2num(w))

		else:
			what = Fluka._parseFreeFormat(line)
			tag = what.pop(0)
			if len(tag)>8: tag=tag[0:8]
			if len(what)==7:
				sdum = what.pop()
			else:
				sdum = ""
			what.insert(0, sdum)

		return (tag, what)

	#-----------------------------------------------------------------------
	# Parse a free format line
	# Remove unwanted characters from free format line and
	# separate everything with comma ,
	#-----------------------------------------------------------------------
	@staticmethod
	def _parseFreeFormat(line):
		"""Remove unwanted characters from a free format line
		and separate everything with comma ,"""
		line = re.sub(" *[;:\\,] *",",",line.strip())
		what = re.sub(" +",",",line).split(',')
		return list(map(_str2num, what))

	#-----------------------------------------------------------------------
	# Parse Geometry
	#-----------------------------------------------------------------------
	def _parseGeometry(self, geobegin):
		voxel = False
		while True:
			line = self.readline()
			if line is None:
				raise Exception("Error parsing geometry")

			line = self._parseComment(line)
			if line is None: continue

			if line[0] in '#!':
				tag, what = self._parseLine(line)
				self._addCard(Card(tag, what, self._comment))

			elif line[0:6] == "GEOEND":
				tag, what = self._parseLine(line)
				self._addCard(FlukaCard(tag, what, self._comment))
				return

			elif line[0:6] == "VOXELS":
				tag, what = self._parseLine(line)
				card = FlukaCard(tag, what)
				self._addCard(card)
				try:
					card.loadVoxel()
				except Exception as err:
					log.error(f"Card #{card.pos1()}"
						f"loading voxel file {card.name()}.vxl")
					log.errorCont(str(err))
				voxel = True

			else:
				break

		# Except the title of the GEOBEGIN
		try:    self._ivopt = int(line[0:5].strip())
		except: self._ivopt = 0
		try:    self._idbg  = int(line[5:10].strip())
		except: self._idbg  = 0

		# Special treatment for GEOBEGIN
		geobegin.setText(line[20:].strip())
		# Save vars for use in parsing
		geobegin[FlukaCard.IVOPT] = self._ivopt
		geobegin[FlukaCard.IDBG]  = self._idbg

		if self._idbg<0 and self.geoFormat==FlukaCard.FORMAT_SINGLE:
			self.geoFormat = FlukaCard.FORMAT_DOUBLE

		self._parseBodies()
		self._parseRegions()
		self._parseVolume()

		# Correct last body name to VOXEL
		if voxel and self.geoFormat != FlukaCard.FORMAT_FREE:
			# Count all bodies
			nbodies = 1	# +1
			for body in CardInfo.bodies[FlukaCard.MODE]:
				try:
					nbodies += len(self.cards[body])
				except KeyError:
					pass
			self.changeName("body", _body2name(str(nbodies)), "VOXEL")

	#-----------------------------------------------------------------------
	# Parse bodies
	#-----------------------------------------------------------------------
	def _parseBodies(self):
		while True:
			line = self.readline()
			if line is None: return
			line = self._parseComment(line)
			if line is None: continue

			# Preprocessor or flair cards
			if line[0] in "#!":
				tag, what = self._parseLine(line)
				self._addCard(Card(tag, what, self._comment))
				continue
			if line[0] == "$":
				# FIXME WRONG place
				tag, what = self._parseLine(line)
				self._addCard(FlukaCard(tag, what, self._comment))
				continue

			what = []
			tag, name = self._parseBodyLine(line, what)
			name = _body2name(name)

			if not self._cardDisabled and (tag=="END" or line=="END"):
				self._addEND()
				return

			_cardDisabled = self._cardDisabled

			info = CardInfo.get(self.MODE, tag)
			if info is None:
				log.error(f"Line {self._lineNo[-1]}: "
					f"Unknown body type '{tag}' for '{name}'")
				log.errorCont(line)
				return
			needwhat = info.maxwhats-1
			# Read and parse needed whats
			while len(what)<needwhat:
				line = self.readline()
				if line is None:
					raise Exception("Error parsing geometry")
				if _cardDisabled&1:
					# check the comment contains only numbers
					try:
						# use max() to evaluate the map object
						max(map(float, line[1:].split()))
						line = line[1:]
					except ValueError:
						self.saveline(line)
						break
				line = self._parseComment(line)
				if line is None: continue
				self._parseBodyLine(line, what)
			# delete extra whats if any
			if len(what)>needwhat: del what[needwhat:]
			what.insert(0, name)
			self._cardDisabled = _cardDisabled
			self._addCard(FlukaCard(tag, what, self._comment))
			if self._cardDisabled&1:
				self._cardDisabled -= 1

	#-----------------------------------------------------------------------
	# Parse a body line
	#-----------------------------------------------------------------------
	def _parseBodyLine(self, line, what):
		startWhat = len(what)
		if self.geoFormat == FlukaCard.FORMAT_SINGLE:
			body = line[2:5].strip()
			name = line[5:10].strip()
			if startWhat>0 and (body or name):
				raise Exception("Invalid body definition at line: "
					+self._lineNo[-1])
			for i in range(10, 61, 10):
				w = line[i:i+10].strip()
				what.append(_str2num(w))

		elif self.geoFormat == FlukaCard.FORMAT_DOUBLE:
			body = line[2:5].strip()
			name = line[5:10].strip()
			if startWhat>0 and (body or name):
				raise Exception("Invalid body definition at line: "
					+self._lineNo[-1])
			for i in range(10, 55, 22):
				w = line[i:i+22].strip()
				what.append(_str2num(w))

		else:
			wlist = Fluka._parseFreeFormat(line)
			body = name = ""
			if startWhat==0:
				body = str(wlist[0])
				if len(wlist)>1: name = wlist[1]
				i = 2
			else:
				i = 0

			for i in range(i, len(wlist)):
				what.append(_str2num(wlist[i]))

		return (body, name)

	#-----------------------------------------------------------------------
	# Parse Regions
	#-----------------------------------------------------------------------
	def _parseRegions(self):
		name          = ""
		neighbors     = ""
		expstr        = ""
		comment       = []
		_cardDisabled = 0
		properties    = []
		regions       = set()	# check for duplicate names in FIXED format
		self._comment = []

		while True:
			# FIXME dirty hack
			# the problem comes that the addRegion is executed in the end
			# and not immediately when it appears
			# Maybe the correct way will be to add immediately a region
			# and simply append to its last expression
			#
			# For the moment create a special readline!!
			line = self.readlineRegion()

			if line is None:
				if name!="":
					self._addRegion(name, neighbors, comment, expstr, _cardDisabled)
				return
			line = self._parseComment(line, False)
			if line is None: continue

			if line[:6] == "GEOEND":
				self._addRegion(name, neighbors, comment, expstr, _cardDisabled)
				self._addEND()
				log.error("Parsing geometry END is missing")
				return

			# Preprocessor or flair cards
			if line[0] in "#!":
				# Add current region if any
				oldcomment = self._comment	# Save the comment
				self._addRegion(name, neighbors, comment, expstr, _cardDisabled)
				# Possible region continuation
				name      = "&"
				neighbors = 0
				expstr    = ""
				comment   = []
				self._comment = oldcomment	# Restore it

				# Add current card
				tag, what = self._parseLine(line)
				if tag == "#endinclude":
					self.close()
					self._cardDisabled = 0
				else:
					self._addCard(Card(tag, what, self._comment))

			# Free format
			elif self.geoFormat == FlukaCard.FORMAT_FREE:
				# Free format
				line = line.strip()

				if line=="END":
					oldcomment = self._comment	# remember comment
					self._addRegion(name, neighbors, comment,
							expstr, _cardDisabled)
					self._comment = oldcomment
					self._addEND()
					return

				# Check for region continuation
				elif line and "+-|()".find(line[0]) >= 0:
					if self._comment:
						oldcomment = self._comment	# remember comment
						# add region
						self._addRegion(name, neighbors, comment,
								expstr, _cardDisabled)
						name      = "&"			# continuation card
						neighbors = 0
						expstr    = line
						comment   = oldcomment		# restore comment
						self._comment = []		# just for precaution
					else:
						# append to region expression
						expstr += f"\n{line}"

				# New region definition
				else:
					match = _REGIONPAT.match(line)
					if match is None:
						log.error(f"Line #{self._lineNo[-1]} in region: {line}")
						continue

					oldcomment    = self._comment		# remember comment
					thisProp      = self._properties	# and properties
					self._properties = properties		# restore old properties
					# New region definition
					self._addRegion(name, neighbors, comment,
							expstr, _cardDisabled)
					name          = match.group(1)
					neighbors     = match.group(2)
					expstr        = match.group(3)
					comment       = oldcomment		# remember comment
					properties    = thisProp		# and properties
					self._comment = []			# just for precaution
					_cardDisabled = self._cardDisabled

			# Single precision fixed format
			else:
				n = line[1:6].strip()	# Name

				# New region definition
				if n!="":
					thisComment = self._comment
					# Save previous region
					self._addRegion(name, neighbors, comment,
							expstr, _cardDisabled)
					if n=="END":
						self._addEND()
						return

					# New region
					name = _region2name(n)
					if name in regions:
						name = f"R{len(regions)}"
					regions.add(name)

					neighbors = line[6:10].strip()
					expstr    = ""
					comment   = thisComment
					_cardDisabled = self._cardDisabled
				elif self._comment:
					comment.extend(self._comment)
				self._comment = []

				# get expression
				e = line[10:]

				# Convert to free format expression string
				efree = ""

				if abs(self._idbg) == 100:
					width = 9
				else:
					width = 7
				if len(e)%width != 0: e += " "*(width-len(e)%width)

				for i in range(0, len(e), width):
					orop = e[i:i+2]
					if orop=="OR":
						efree += " |"

					body = e[i+2:i+width].strip()
					if body=="":
						log.error(f"Bad alignment in region definition: '{name}'")
						raise Exception(f"Error parsing region {name}: "
							f"Invalid body name found '{e[i+2:i+width]}' "
							f"column={i+2}")
					sign = body[0]
					if sign in ("+", "-"):
						efree += " %s%s" % (sign,
							_body2name(body[1:]))
					else:
						efree += " +%s" % (_body2name(body))

				if expstr=="":
					expstr = efree
				else:
					expstr += "\n" + efree	# Continue region

	#-----------------------------------------------------------------------
	def _addEND(self):
		self._addCard(FlukaCard("END", [], self._comment))

	#-----------------------------------------------------------------------
	def _addRegion(self, name, neighbors, comment, expstr, cardDisabled):
		if name=="" or (name=="&" and expstr==""): return None
		saveDisabled = self._cardDisabled
		self._cardDisabled = cardDisabled

		try: neighbors = int(neighbors)
		except: neighbors = 5

		what = [ name, neighbors ]
		# XXX should not come where with expstr=""
		if len(expstr)>1 and expstr[0]=="\n": expstr = expstr[1:]
		card = FlukaCard("REGION", what, comment, expstr)
		card[Card.VOLUME] = ""
		self._addCard(card)
		self._cardDisabled = saveDisabled
		return card

	#-----------------------------------------------------------------------
	@staticmethod
	def _parseTransformation(what):
		if "#" in what:
			uwhat = what.upper()
			if uwhat[:4] == "ROT#" or uwhat[:3] == "RO#":
				try: idx = int(what.split("#")[-1])
				except: idx = 1
				what = f"rot#{idx:03d}"
		return what

	#-----------------------------------------------------------------------
	# Parse Volume data if any
	#-----------------------------------------------------------------------
	def _parseVolume(self):
		if self._ivopt != 3: return
		# Expect data in the format 7E10.5
		self.preprocess()		# << XXX very nasty on half file
		lst = self["REGION"]
		Card.sort(lst)
		regionList = [x for x in lst
			if not x.ignore() and x.name()!="&"]

		nlines, rem = divmod(len(regionList), 7)
		if rem > 0: nlines += 1
		nr = 0
		while nlines>0:
			# peek next line
			pos  = self._files[-1].tell()
			line = self.readline()
			if line is None:
				raise Exception("Error parsing geometry")
			if not line: continue
			if line[0] in ("*", "G"):
				# FIXME!!!!! will break the md5
				# go back and exit
				self._files[-1].seek(pos)
				break
			for i in range(0, 61 ,10):
				vol = _str2num(line[i:i+10].strip())
				if nr < len(regionList):
					regionList[nr][Card.VOLUME] = vol
				nr += 1
			nlines -= 1

	#-----------------------------------------------------------------------
	# Import/load a fluka input file including geometry
	#-----------------------------------------------------------------------
	def _load(self):
		tag  = ""
		sdum = ""
		self._comment = []			# current comment
		self._cwhat   = []			# Commented special whats
		self._nwhats  = 0			# Number of whats of last added card
		self._stack   = [1]			# track active #if nesting during reading
							# look Input.preprocess for values
		self._defined = defaultdict(int)	# active #define names
		self._cardDisabled = 0
		self.format    = FlukaCard.FORMAT_SINGLE
		self.geoFormat = FlukaCard.FORMAT_SINGLE
		while True:
			line = self.readline()
			if line is None: break
			line = self._parseComment(line)
			if line is None: continue

			prev_tag  = tag
			prev_sdum = sdum
			tag, what = self._parseLine(line)
			try:
				sdum = what[0]
			except IndexError:
				sdum = ""
			text      = ""

			# Handle card exceptions
			if not tag: pass

			# 1. Continuation
			elif isinstance(sdum,str) and sdum in ("&","&&") and tag==prev_tag:
				# Find last card and append whats
				try:
					card = self.cards[tag][-1]
					card.appendComment(self._comment)
					if card.nwhats() == self._nwhats:
						# append cards at the end
						card.appendWhats(what[1:])
					else:
						# append cards in the middle in case
						# that functions are used
						card.appendWhats(what[1:], self._nwhats)
					self._nwhats += len(what)-1
					self._comment = []
					continue
				except KeyError:
					pass

			# 2. Compound continuation (special case)
			elif tag == "COMPOUND":
				# check if previous line is COMPOUND same state
				# and same sdum
				if prev_tag == tag and prev_sdum == sdum:
					# find last COMPOUND card with that sdum
					card = self.cards[tag][-1]
					# check state
					if card.enable == bool(self._cardDisabled==0):
						card.appendComment(self._comment)
						if card.nwhats() == self._nwhats:
							card.appendWhats(what[1:])
						else:
							card.appendWhats(what[1:], self._nwhats)
						self._nwhats += len(what)-1
						self._comment = []
						card.checkCompound()
						continue	# skip the rest

			# 3. Extra line readout
			if tag in ("TITLE", "OPEN", "mesh"):
				text = self.readline()
				if text is None: break

			# Handling of special cases needed for parsing
			# the rest of the file
			elif tag == "FREE":
				if not self._cardDisabled and self._stack[-1] in (1,3):
					self.format = FlukaCard.FORMAT_FREE

			elif tag == "FIXED":
				if not self._cardDisabled and self._stack[-1] in (1,3):
					self.format = FlukaCard.FORMAT_SINGLE

			elif tag=="GLOBAL" and not self._cardDisabled and self._stack[-1] in (1,3):
				w4 = _intWhat(what, 4)
				if w4 in (2, 3):
					self.format= FlukaCard.FORMAT_FREE
				if _intWhat(what, 5)>0:
					self.geoFormat = FlukaCard.FORMAT_FREE

			# Handle special card PLOTGEOM
			elif tag == "PLOTGEOM":
				w6 = _intWhat(what, 6)
				if w6 and self.open("PLG.GFXINDAT","r",tag) is None:
					log.error("Cannot open file PLG.GFXINDAT")
					continue

				text = self.readline()
				line = self.readline()
				what.extend([_str2num(line[x:x+10])
					for x in range(0, 60, 10)])

				line = self.readline()
				what.extend([_str2num(line[x:x+10])
					for x in range(0, 60, 10)])

				line = self.readline()
				what.extend([_str2num(line[x:x+10])
					for x in range(0, 40, 10)])

			elif tag=="COMMENT":
				# Add to comments the following lines
				for _ in range(max(1, _intWhat(what, 1))):
					line = self._parseComment("*"+self.readline())
				continue

			elif tag == "GEOBEGIN" and sdum!="FLUGG":
				# Starting position of GEO
				if self.geoFormat == FlukaCard.FORMAT_SINGLE:
					if sdum[:8]=="COMBNAME" and self._stack[-1] in (1,3):
						self.geoFormat = FlukaCard.FORMAT_FREE
				card = FlukaCard(tag, what, self._comment)
				self._addCard(card)

				w3 = _intWhat(what, 3)
				w4 = _intWhat(what, 4)

				if w3 not in (0,5):
					fn = self.readline()
					if fn is None:
						raise Exception("Error parsing geometry")
					card[FlukaCard.GEOFILE] = fn.strip()

				if w4 not in (0,11):
					fn = self.readline()
					if fn is None:
						raise Exception("Error parsing geometry")
					card[FlukaCard.GEOOUT] = fn.strip()

				if card[FlukaCard.GEOFILE]:
					self.open(card[FlukaCard.GEOFILE],"r",tag)
				self._parseGeometry(card)
				continue

			elif tag == "LATTICE":
				what[0] = Fluka._parseTransformation(what[0])

			elif tag == "LATTSNGL":
				try:
					what[2] = Fluka._parseTransformation(what[2])
					what[3] = Fluka._parseTransformation(what[3])
				except IndexError:
					pass

			# Add current card
			if tag and (tag[0]=="#" or tag in CardInfo.FLAIR_TAGS):		# flair cards
				self._addCard(Card(tag, what, self._comment, text))
				self._parsePreprocess(tag, what[0])
			else:
				self._addCard(FlukaCard(tag, what, self._comment, text))

		# what about nested files???
		if len(self._stack)>1:
			log.error("Non-terminated #if")
		del self._stack
		del self._defined

	#-----------------------------------------------------------------------
	# Preprocess during loading and keep information if card is active or not
	# values of the stack like in the Input.preprocess
	#-----------------------------------------------------------------------
	def _parsePreprocess(self, tag, var):
		if self._cardDisabled: return
		active = self._stack[-1]
		active13 = active in (1,3)
		cardno = len(self.cardlist)+1
		if active13 and tag == "#define":
			self._defined[var] = 1

		elif active13 and tag == "#undef":
			self._defined[var] = 0

		elif tag in ("#if","#ifdef"):
			if active13:
				self._stack.append(self._defined[var])
			else:
				self._stack.append(-1)	# ignore this nesting

		elif tag == "#ifndef":
			if active13:
				self._stack.append(1-self._defined[var])
			else:
				self._stack.append(-1)	# ignore this nesting

		elif tag == "#elif":
			if len(self._stack)<=1:
				log.error(f"Card #{cardno} unexpected #elif")
			elif active==1:
				self._stack[-1] = -1	# terminate this nesting
			elif active==0:
				self._stack[-1] = self._defined[var]
			elif active>1:
				log.error(f"Card #{cardno} misplaced #elif after #else")

		elif tag == "#else":
			if len(self._stack)<=1:
				log.error(f"Card #{cardno} unexpected #else")
			elif active>1:
				log.error(f"Card #{cardno} misplaced #else")
			elif active>=0:
				self._stack[-1] = 3-self._stack[-1]

		elif tag == "#endif":
			if len(self._stack)<=1:
				log.error(f"Card #{cardno} unexpected #endif")
			else:
				self._stack.pop()

	#-----------------------------------------------------------------------
	def _postLoad(self):
		self.convert2Names()
		super()._postLoad()

	#-----------------------------------------------------------------------
	# Save input file with the expression format
	# return md5 checksum or None on error
	#-----------------------------------------------------------------------
	def _save(self):
		# Initialize variables
		self.format    = FlukaCard.FORMAT_SINGLE
		self.geoFormat = FlukaCard.FORMAT_SINGLE

		if "#include" in self.cards:
			return self._saveWithInclude()

		# Expressions
		exprNotGeo = lambda x: not x._geo

		# 1st. Write all input cards before GEOBEGIN
		#      skipping geometry cards
		try:
			geobegin = self["GEOBEGIN"][0]
			begin_   = self.cardlist.index(geobegin)
		except:
			geobegin = FlukaCard("GEOBEGIN", ["COMBNAME"])
			# Make a guess of the position before the first geometry card
			for pos,card in enumerate(self.cardlist):
				if card.isGeo():
					begin_ = pos
					break
			else:
				begin_ = len(self.cardlist)

		# Check for volume data
		self._ivopt = 0
		for region in self["REGION"]:
			vol = region.getProperty(Card.VOLUME,"")
			if vol != "":
				self._ivopt = 3
				break

		geobegin[FlukaCard.IVOPT] = self._ivopt

		# Do not write any defines!
		# geolevel defines the position in the input file look the writeCards()
		self.writeCards(exprNotGeo, None, 0, 0, begin_)

		# if there is no geometry... then exit
		if begin_ == len(self.cardlist): return

		# 2nd. Write all non-geometry cards which are inside GEOBEGIN .. GEOEND
		try:
			# Find first GEOEND
			geoend = self["GEOEND"][0]
			end_ = self.cardlist.index(geoend)
			self.writeCards(exprNotGeo, None, 1, begin_+1, end_)
		except:
			geoend = None
			end_ = len(self.cardlist)

		# 3th. Write the geometry cards
		# correct geobegin card if needed
		if self._exprFmt==FlukaCard.EXPR_EXPAND:
			geobegin = geobegin.clone()
			geobegin.setWhat(3,"")
			geobegin.setWhat(4,"")

		if not geobegin[FlukaCard.GEOFILE] and geobegin.intWhat(3) not in (0,5):
			geobegin.setWhat(3,"")
		if not geobegin[FlukaCard.GEOOUT] and geobegin.intWhat(4) not in (0,11):
			geobegin.setWhat(4,"")

		flugg = geobegin.sdum()=="FLUGG"
		self.checkFormat(geobegin)
		geoline = geobegin.toStr(fmt=self.format).splitlines()
		if geobegin.comment():
			self.writeline(geobegin.commentStr())
		self.writeline(geoline[0])

		# Write the geometry filename
		if not flugg and geobegin[FlukaCard.GEOFILE] and geobegin.intWhat(3) not in (0,5):
			self.writeline(geobegin[FlukaCard.GEOFILE])

		# Write geo output file before the title
		if geobegin[FlukaCard.GEOOUT] and geobegin.intWhat(4) not in (0,11):
			self.writeline(geobegin[FlukaCard.GEOOUT])

		# Open the geometry file
		fgeo = False
		if not flugg and geobegin[FlukaCard.GEOFILE] and geobegin.intWhat(3) not in (0,5):
			try:
				fgeo = True
				self.open(geobegin[FlukaCard.GEOFILE], "w")
			except:
				log.error(f"Cannot open output geometry file '{geobegin[FlukaCard.GEOFILE]}'")
				log.errorCont("Write geometry to input")

		# Handle VOXELS card
		if not flugg:
			# Write VOXELS are present
			if "VOXELS" in self.cards:
				self.writeCards(lambda x:x.tag=="VOXELS", self.format, 1)

			# Write title
			self.writeline(geoline[1])

			# 4th Write geometry
			self._writeGeometry(self.geoFormat)

		if fgeo: self.close()

		# 6th. Write remaining cards including the GEOEND if any
		if geoend is None:
			geoend = FlukaCard("GEOEND")
			self.writeCard(geoend, self.format)
			self.writeCards(exprNotGeo, None, 4, begin_+1)
		else:
			self.writeCards(exprNotGeo, None, 4, end_)

		# Check for STOP at the end
		if self.cardlist[-1].tag != "STOP":
			self.writeCard(FlukaCard("STOP"), self.format)

	#-----------------------------------------------------------------------
	# If include are present write in a strict sequential format
	#-----------------------------------------------------------------------
	def _saveWithInclude(self):
		"""Open and write the input to file (specified either by filename,
		or by file pointer). Optionally use a different geometry
		file and a different format"""

		# Check for volume data
		self._ivopt = 0
		for region in self["REGION"]:
			vol = region.getProperty(Card.VOLUME,"")
			if vol != "":
				self._ivopt = 3
				break

		# main writing loop
		geolevel = 0	# 0 before GEOBEGIN
				# 1 bodies and VOXELS
				# 2 REGIONs
				# 3 LATTICEs
				# 4 GEOEND
				# 5 after
		geoExternal = False
		includeoff  = 0	# counter of how many includes to exclude when the #include is disabled
		for card in self.cardlist:
			tag = card.tag

			# if we are under an include off do not write the file
			if includeoff > 0:
				if tag=="#include":
					includeoff += 1
				elif tag=="#endinclude":
					includeoff -= 1

			elif tag=="#include" :
				# Write the card even if disabled
				if self._exprFmt != FlukaCard.EXPR_EXPAND:
					self.writeCard(card, self.format)
				if card.ignore():
					includeoff = 1
				elif card.notIgnore() and self._exprFmt!=FlukaCard.EXPR_EXPAND:
					self.open(card.what(0),"w")

			elif tag=="#endinclude":
				if card.comment():
					self.writeline(card.commentStr())
				if card.notIgnore() and self._exprFmt!=FlukaCard.EXPR_EXPAND:
					self.close()

			elif tag=="GEOBEGIN":
				# correct geobegin card if needed
				if self._exprFmt==FlukaCard.EXPR_EXPAND:
					card = card.clone()
					card.setWhat(3,"")
					card.setWhat(4,"")
				if not card[FlukaCard.GEOFILE] and card.intWhat(3) not in (0,5):
					card.setWhat(3,"")
				if not card[FlukaCard.GEOOUT]  and card.intWhat(4) not in (0,11):
					card.setWhat(4,"")
				card[FlukaCard.IVOPT] = self._ivopt
				self.checkFormat(card)
				flugg = card.sdum()=="FLUGG"
				geoline = card.toStr(fmt=self.format).splitlines()
				geolevel = 1
				if card.comment():
					self.writeline(card.commentStr())
				self.writeline(geoline[0])

				if flugg:
					# Do nothing, skip the geometry
					pass
				else:
					if card[FlukaCard.GEOFILE] and card.intWhat(3) not in (0,5):
						self.writeline(card[FlukaCard.GEOFILE])
						geoExternal = True
					# Write geo out file before the title
					if card[FlukaCard.GEOOUT] !="" and card.intWhat(4) not in (0,11):
						self.writeline(card[FlukaCard.GEOOUT])
					if geoExternal and self._exprFmt!=FlukaCard.EXPR_EXPAND:
						self.open(card[FlukaCard.GEOFILE],"w")

				# If VOXELS are present
				# FIXME not good. Needs to write the cards as is
				try:
					for c in self.cards["VOXELS"]:
						self.writeCard(c, self.format)
				except KeyError:
					pass

				# Write title
				if not flugg:
					self.writeline(geoline[1])

			elif geolevel>0:
				if geolevel==1 and tag=="REGION":
					# XXX Could be problematic with the #if's
					geolevel = 2

				elif geolevel==2 and (tag!="REGION" and tag[0]!='#'):
					# XXX could be problematic if #include is before
					# then the END is written in the #include
					geolevel = 3

				# XXX insert region volumes? LATTICE, LATTSNGL or GEOEND
				if tag in ("LATTICE","LATTSNGL"):
					self.writeCard(card, self.format)
				elif tag == "VOXELS":
					pass
				elif card._geo:
					self.writeCard(card, self.geoFormat)
				else:
					if tag=="GEOEND": geolevel = 0
					if geoExternal and self._exprFmt!=FlukaCard.EXPR_EXPAND:
						self.close()
					self.writeCard(card, self.format)

			else:
				self.writeCard(card, self.format)
				self.checkFormat(card)

	# ----------------------------------------------------------------------
	# Find current format
	# FIXME #defines and state!
	# ----------------------------------------------------------------------
	def checkFormat(self, card):
		if card.ignore(): return
		tag = card.tag
		if   tag == "FREE":
			self.format = FlukaCard.FORMAT_FREE
		elif tag == "FIXED":
			self.format = FlukaCard.FORMAT_SINGLE
		elif tag == "GLOBAL":
			w = card.intWhat(4)
			if w in (2, 3):
				self.format= FlukaCard.FORMAT_FREE
			w = card.intWhat(5)
			if w > 0:
				self.geoFormat = FlukaCard.FORMAT_FREE
			elif w < 0:
				# XXX XXX Accept only FREE format geometry XXX XXX
				log.warning("Changing GLOBAL what(5)=1")
				card.setWhat(5,1)
				self.geoFormat = FlukaCard.FORMAT_FREE
		elif tag == "GEOBEGIN":
			# Force free geometry format
			if card.sdum()[:8]!="COMBNAME" and card.sdum()!="FLUGG":
				# XXX XXX Accept only COMBNAME! XXX XXX
				# To avoid stupid user errors
				log.warning(f"Changing GEOBEGIN sdum={card.sdum()} to COMBNAME")
				card.setSdum("COMBNAME")
			self.geoFormat = FlukaCard.FORMAT_FREE
			#if self.geoFormat == FlukaCard.FORMAT_SINGLE:
			#	if card.sdum()=="COMBNAME":
			#		self.geoFormat = FlukaCard.FORMAT_FREE
			#	elif card[FlukaCard.IDBG] < 0:
			#		self.geoFormat = FlukaCard.FORMAT_DOUBLE

	#-----------------------------------------------------------------------
	# write a single card
	#-----------------------------------------------------------------------
	def writeCard(self, card, fmt):
		# Skip flair cards
		# FIXME Temporary I can embed them adding an ! as prefix
		if card.tag in CardInfo.FLAIR_TAGS or \
		   card.MODE not in ("",Fluka.MODE): return
		if0 = False
		if not card.enable and isinstance(card, FlukaCard) and \
		   (card.info is CardInfo.NONE or not card.info.disableComment):
			if0 = True
			self.write("#if 0\n")
		else:
			if0 = False
		if card.comment():
			self.writeline(card.commentStr())
		es = FlukaCard.evalWhatStr(card, self._exprFmt)
		if es: self.write(es)
		self.writeline(card.toStr(fmt=fmt, properties=False))

		if if0: self.write("#endif\n")

	#-----------------------------------------------------------------------
	# Write cards of that fulfill a given condition(lambda)
	# with all the preprocessor cards around them
	# FIXME Write also non empty blocks: #if .. #define/#undef .. #endif
	#       Might end up in duplicate blocks when only #define/#undef
	#       exist inside the geometry
	# geolevel
	#   -2 ignore defines at all levels
	#   -1 ignore defines at level=0 (write higher nesting ones)
	#    0 before geobegin
	#    1 bodies and voxels
	#    2 regions
	#    3 lattices
	#    4 after geoend
	# from_ .. to_   (to_ is not included)
	#-----------------------------------------------------------------------
	def writeCards(self, condition, format_=None, geolevel=0, from_=0, to_=-1):
		nest    =  0
		start   = -1
		level   =  0
		writing = False

		# Force a fixed format
		if format_ is not None: fmt = format_

		# Loop over all cards
		if to_ == -1: to_ = len(self.cardlist)
		for cid in range(from_, to_):
			card = self.cardlist[cid]
			tag  = card.tag
			if tag == "GEOBEGIN":
				level = 1
			elif tag == "GEOEND":
				level = 4

			if format_ is None:	# find format
				fmt = self.format

			# Preprocessor nesting
			if card.enable and tag[0]=="#":
				if writing and \
				   (self._exprFmt != FlukaCard.EXPR_EXPAND or
				    card.tag not in ("#include", "#endinclude")):
					self.writeCard(card, fmt)

				if tag in ("#if", "#ifdef", "#ifndef"):
					if start < 0 and nest == 0:
						start = cid
					nest += 1

				elif tag == "#endif":
					nest -= 1
					if nest == 0:
						start  = -1
						writing = False

					elif nest < 0:
						# misplaced #elif or #else
						start = cid
						nest = 0

				elif tag in ("#elif", "#else") and nest == 0:
					# misplaced #elif or #else
					start = cid
					nest = 1

				elif tag in ("#define", "#undef"):
					if level == geolevel:
						if nest == 0:
							# Write cards on top nesting correct level
							self.writeCard(card, fmt)
						elif start>=0 and not writing:
							# Write everything up to now
							for c in self.cardlist[start:cid+1]:
								if c.tag[0]=='#' or condition(c):
									self.writeCard(c, fmt)
							start   = -1
							writing = True

					elif geolevel==-1:
						# For anything with high nesting write them anyway
						# when defines are not required
						if nest != 0 and start>=0 and not writing:
							# Write everything up to now
							for c in self.cardlist[start:cid+1]:
								if c.tag[0]=='#' or condition(c):
									self.writeCard(c, fmt)
							start   = -1
							writing = True

			# Target cards
			elif condition(card):
				if nest == 0:
					self.writeCard(card, fmt)
				else:
					if not writing:
						# Write everything up to now
						for c in self.cardlist[start:cid]:
							if c.tag[0]=='#' or condition(c):
								self.writeCard(c, fmt)
						start = -1
						writing = True
					self.writeCard(card, fmt)

			# Change the format after the current card if needed
			if card.enable:
				if format_ is None:	# find format
					self.checkFormat(card)
				if card.tag=="END":
					level += 1

		# In the error case where a #if..#endif is not closed
		if start >= 0:
			for c in self.cardlist[start:]:
				if c.tag[0]=='#' or condition(c):
					self.writeCard(c, fmt)

	#-----------------------------------------------------------------------
	# Write geometry only
	#      preprocessor cards are written only when cards
	#      are in the correct position
	#-----------------------------------------------------------------------
	def _writeGeometry(self, geoFmt=None, fmt=None):
		if geoFmt is None: geoFmt = self.geoFormat

		# Use the formatting of the input
		if fmt is None: fmt = self.format

		# 1st. Write bodies
		self.writeCards(lambda x:x.tag in CardInfo.bodies[FlukaCard.MODE] or x.tag[0]=="$",
				geoFmt, 1)
		ends = self.cards.get("END", None)
		if ends:
			# write all END until the first active
			for lastEnd,card in enumerate(ends):
				self.writeCard(card, geoFmt)
				if card.notIgnore():
					break
		else:
			self.writeline(FlukaCard._endStr(geoFmt))

		# 2nd. Write regions
		self.writeCards(lambda x:x.tag=="REGION", geoFmt, 2)
		if ends and len(ends)>1:
			# write all END after the last one
			for card in ends[lastEnd+1:]:
				self.writeCard(card, geoFmt)
		else:
			self.writeline(FlukaCard._endStr(geoFmt))

		# Write volume data if needed
		try:
			self._ivopt = int(self._ivopt)
		except:
			self._ivopt = 0

		if self._ivopt == 3:
			line = ""
			for region in self.cardsSorted("REGION",-1).values():
				if region.name() == "&": continue
				vol = region.getProperty(Card.VOLUME,"")
				if vol=="": vol="1.0"
				line += "%10s"%(bmath.format(vol, 10))
				if len(line)==70:
					self.writeline(line)
					line = ""
			if line: self.writeline(line)

		# 3rd. Write lattices
		if "LATTICE"  in self.cards or \
		   "LATTSNGL" in self.cards:
			self.writeCards(
				lambda x:x.tag in ("LATTICE","LATTSNGL"),
				fmt, 3)

	# ----------------------------------------------------------------------
	# Write an input file to be used with plotgeom
	# ----------------------------------------------------------------------
	def writePlotgeomInput(self, plotgeom, directory=".", title=""):
		"""Write a temporary file for plotting the geometry"""
		# Write a minimized version of the file
		# FIXME preprocessor defines are missing
		# GEOBEGIN
		# ...Bodies
		# ...Regions
		# ...Lattices
		# GEOEND
		# ROT-DEFI if any
		# MATERIAL
		# ASSIGNMA
		# PLOTGEOM card
		# STOP
		ftmp = tempfile.NamedTemporaryFile(mode="wt",
					suffix=".inp",
					prefix=Global.tmpPrefix,
					dir=directory,
					delete=False)
		self.open(ftmp, "w", backup=False)
		self.preprocess()

		# ------------------------------------------------------
		# Write a minimized file with only the geometry and
		# what it is necessary
		# ------------------------------------------------------
		self.write("* Temporary input file only for geometry plotting\n")

		# XXX Lattice card should be saved with FIXED FORMAT, in this case!
		# write active defines
		self.writeCards(lambda x:False, FlukaCard.FORMAT_FREE)

		# Write global card
		glcard = FlukaCard("GLOBAL", ["","","","",3.0,1.0])
		try:
			for card in self.cards["GLOBAL"]:
				if card.notIgnore():
					for i in range(4):
						glcard.setWhat(i, card.what(i))
		except KeyError:
			pass
		self.writeCard(glcard, FlukaCard.FORMAT_SINGLE)

		# GEOBEGIN More tricky
		geobegin = FlukaCard("GEOBEGIN", ["COMBNAME"], text=title)
		geobegin[FlukaCard.IVOPT] = 0
		geobegin[FlukaCard.IDBG]  = 0
		# Copy important options from real GEOBEGIN
		flugg = False
		try:
			card = self.cards["GEOBEGIN"][0]
			geobegin.setWhat(1, card.numWhat(1))
			geobegin.setWhat(2, card.numWhat(2))
			geobegin.setWhat(5, card.numWhat(5))
			geobegin.setWhat(6, card.numWhat(6))
			if card.sdum()=="FLUGG":
				geobegin.setSdum(card.sdum())
				flugg = True
		except (KeyError, TypeError):
			pass
		geoline = geobegin.toStr(fmt=FlukaCard.FORMAT_FREE).splitlines()
		self.write(geoline[0]+"\n")
		if flugg:
			self.close()
			return None

		# If VOXELS are present, write the first active one
		if "VOXELS" in self.cards:
			for card in self.cards["VOXELS"]:
				if card.notIgnore():
					self.write(card.toStr(fmt=FlukaCard.FORMAT_FREE)+"\n")
					break
		self.write(geoline[1]+"\n")		# Write title

		self._writeGeometry(FlukaCard.FORMAT_FREE, FlukaCard.FORMAT_FREE)

		geoend = FlukaCard("GEOEND")
		self.writeCard(geoend, FlukaCard.FORMAT_FREE)
		self.writeCards(lambda x:x.tag in ("ROT-DEFI", "MATERIAL", "ASSIGNMA",
						   "USRICALL", "USRGCALL",
						   "MGNFIELD", "MGNCREAT", "MGNDATA"),
				FlukaCard.FORMAT_FREE,
				geolevel=-2)

		# -------------------------------------
		# Create the PLOTGEOM card and write it
		# -------------------------------------
		what = [""]*24
		what[ 0] = "FORMAT"
		what[ 2] = float(plotgeom.get("boundaries",0))

		O = plotgeom.get("O", bmath.Vector.O)
		U = plotgeom.get("U", bmath.Vector.Z)
		V = plotgeom.get("V", bmath.Vector.X)
		extu = plotgeom.get("extu", 50.0)
		extv = plotgeom.get("extv", 50.0)

		loP  = -extu*U -extv*V + O
		hiP  =  extu*U +extv*V + O

		what[ 7] = loP.x
		what[ 8] = loP.y
		what[ 9] = loP.z
		what[10] = hiP.x
		what[11] = hiP.y
		what[12] = hiP.z

		what[13] = V.x
		what[14] = V.y
		what[15] = V.z
		what[16] = U.x
		what[17] = U.y
		what[18] = U.z

		what[19] = 1.0
		what[20] = 1.0
		what[21] = plotgeom.get("nu",200)
		what[22] = plotgeom.get("nv",200)
		self.writeCard(FlukaCard("PLOTGEOM", what, title, text=title),
				FlukaCard.FORMAT_FREE)

		self.write("STOP\n")
		self.close()
		return os.path.splitext(os.path.abspath(ftmp.name))[0]

	#-----------------------------------------------------------------------
	@classmethod
	def newCard(cls, *args, **kwargs):
		return FlukaCard(*args, **kwargs)
