# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   8-Nov-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import math
from math import ceil

import log
import bmath
import Global

from deck import Card
from deck.layout import Layout

from .input    import Fluka, FlukaConfig
from .particle import FlukaParticle
from .card     import FlukaCard

_pwxs_databases = None
_Sab_materials  = None

#===============================================================================
# Layout class for the specific input/canvas display
#===============================================================================
class FlukaLayout(Layout):
	MODE = FlukaCard.MODE

	# Used from usrbin-plot
	USRBIN_type = [ Layout.tab,
			("X-Y-Z"             , 10),
			("R-Φ-Z"             , 11),
			("Region"            , 12),
			("|X|-Y-Z"           , 13),
			("X-|Y|-Z"           , 14),
			("X-Y-|Z|"           , 15),
			("|X|-|Y|-|Z|"       , 16),
			("R-Φ-|Z|"           , 17),
			("Special"           , 18),
			("X-Y-Z point"       ,  0),
			("R-Φ-Z point"       ,  1),
			("Region point"      ,  2),
			("|X|-Y-Z point"     ,  3),
			("X-|Y|-Z point"     ,  4),
			("X-Y-|Z| point"     ,  5),
			("|X|-|Y|-|Z| point" ,  6),
			("R-Φ-|Z| point"     ,  7),
			("Special point"     ,  8) ]

	# Used from resnuclei-plot
	RESNUCLE_type = [("Spallation",  1),
			 ("Low Energy",  2),
			 ("All",  3)]

	# Used in GeoViewer
	ASSIGNMA_field = [	Layout.empty_what,
				("Magnetic", 1),
				("Electric", 2),
				("Magnetic+Electric", 3),
				("Prompt Magnetic", 4),
				("Prompt Electric", 5),
				("Prompt Magnetic+Electric", 6),
				("Decay Magnetic", 7),
				("Decay Electric", 8),
				("Decay Magnetic+Electric", 9),
				Layout.off_what,
				Layout.reset_what]

	# ----------------------------------------------------------------------
	#def __init__(self, flair):
	#	super().__init__(flair)

	# ----------------------------------------------------------------------
	# Enable/disable caching of lists
	# ----------------------------------------------------------------------
	def cache(self, on):
		super().cache(on)
		if on:
			self._cache["body"]        = self.bodyList()
			self._cache["region"]      = self.regionList()
			self._cache["material"]    = self.materialList()
			self._cache["materialall"] = self.materialAllList()
			self._cache["binning"]     = self.binningList()
			self._cache["rotation"]    = self.rotationList()
			self._cache["rotation2"]   = self.rotationList2()
			self._cacheTypedef()

	# ----------------------------------------------------------------------
	# Return body list
	# ----------------------------------------------------------------------
	def bodyList(self, inctag=True):
		try: return self._cache["body"]
		except KeyError: pass
		bodies = super().bodyList(inctag)
		if "VOXELS" in self.input.cards:
			bodies.append("VOXEL")
		return sorted(bodies, key=str.upper)

	# ----------------------------------------------------------------------
	def magneticList(self):		return self.typedefList("magnetic")

	# ----------------------------------------------------------------------
	# Return region list
	# ----------------------------------------------------------------------
	def regionList(self):
		try: return self._cache["region"]
		except KeyError: pass
		lst = super().regionList()
		lst.append("@LASTREG")
		return lst

	# ----------------------------------------------------------------------
	def allRegionList(self):
		lst = self.regionList()[:]
		lst.append("@ALLREGS")
		return lst

	# ----------------------------------------------------------------------
	def signRegionList(self):
		lst = self.regionList()[:]
		for i in range(len(lst)-2,0,-1):
			reg = lst[i]
			lst.insert(i,f"-{reg}")
		return lst

	# ----------------------------------------------------------------------
	# Rot-Defi list for ROTPRBIN
	# ----------------------------------------------------------------------
	def rotationList2(self):
		try: return self._cache["rotation2"]
		except KeyError: pass

		lst = self.input.rotationList(True)
		for i,item in enumerate(lst):
			if item.startswith("rot#"):
				lst[i] = (item,int(item[4:]))
		lst.append(("Ignore",0))
		lst.append(("Ignore",-1))
		return lst

	# ----------------------------------------------------------------------
	# Return material list, excluding material if defined
	#-----------------------------------------------------------------------
	def materialList(self, e=None):
		try: return self._cache["material"]
		except KeyError: pass
		lst = super().materialList(e)
		lst.append("@LASTMAT")
		return lst

	# ----------------------------------------------------------------------
	# Return material list, excluding material if defined
	#-----------------------------------------------------------------------
	def materialAllList(self, e=None):
		try: return self._cache["materialall"]
		except KeyError: pass
		lst = [""] + self.input.materialList(1,True)
		if e:
			try: lst.remove(e)
			except ValueError: pass
		return lst

	# ----------------------------------------------------------------------
	# Binning list
	# ----------------------------------------------------------------------
	def binningList(self):
		try: return self._cache["binning"]
		except KeyError: pass

		lst = [""]
		try: lst.extend([x.name() for x in self.input.cards["USRBIN"]])
		except: pass
		try: lst.extend([x.name() for x in self.input.cards["EVENTBIN"]])
		except: pass
		lst.extend(range(1,len(lst)))
		return lst

	# ----------------------------------------------------------------------
	# Detector list
	# ----------------------------------------------------------------------
	def detectorList(self, tag):
		tag = tag[:8]
		if tag in ("USRBIN","EVENTBIN"): tag = ("USRBIN","EVENTBIN")
		elif tag in ("USRTRACK","USRCOLL"): tag = ("USRTRACK","USRCOLL")
		lst = [""]
		lst.extend(self.input.cardsSorted(tag[:8]).keys())
		lst.extend(range(1,len(lst)))
		return lst

	# ----------------------------------------------------------------------
	# DcyTimes list
	# ----------------------------------------------------------------------
	def dcytimesList(self):
		cnt = 1
		lst = []
		dcy = {}
		for card in self.input.cardsSorted("DCYTIMES",-1).values():
			for i in range(1,7):
				w = card.numWhat(i)
				if w not in dcy:
					dcy[w] = cnt
					w = card.what(i)
					if w == "": w = "0.0"
					lst.append((str(w),cnt))
					cnt += 1
		lst.append(("Semi-Analogue",-1))
		lst.append("")
		return lst

	# ----------------------------------------------------------------------
	# Return unit list
	# ----------------------------------------------------------------------
	def fileUnitList(self, tag=None, binary=True, onlymatches=False):
		try:
			if self.input.fileunits.empty():
				self.input.scanUnits()
			return self.input.fileunits.filterList(tag, binary, onlymatches)
		except AttributeError:
			# In case of Moira
			return []

	# ----------------------------------------------------------------------
	# Return RADBIOL list
	# ----------------------------------------------------------------------
	def radbiolList(self):
		lst = [""]
		try:
			i = 1
			for card in self.input.cards["RAD-BIOL"]:
				if card.ignore(): continue
				lst.append((card.name(),i))
				i += 1
		except KeyError:
			pass
		return lst

	# ----------------------------------------------------------------------
	def neutronGroupsUsed(self):
		groups = set()	# List of groups to display
		try:
			# Check all LOW-BIAS cards
			for card in self.input.cards["LOW-NEUT"]:
				if card.ignore(): continue
				w1 = card.intWhat(1)
				if w1 == 0:
					g = 260
				else:
					g = w1
				groups.add(g)
		except KeyError:
			pass
		if groups:
			return list(groups)
		else:
			return [260]

	# ----------------------------------------------------------------------
	# Low energy neutron group list
	# ----------------------------------------------------------------------
	def neutronGroupList(self):
		lst = [ ("No cut-off",      0) ]
		groups = self.neutronGroupsUsed()
		for g in groups:
			i = 1
			for e in Fluka.neutronGroupsS.get(g, []):
				if len(groups)==1:
					lst.append((e, i))
				else:
					lst.append((f"{g}: {e}", i))
				i += 1
		return lst

	# ----------------------------------------------------------------------
	# Low energy neutron materials
	# ----------------------------------------------------------------------
	def lowmatList(self):
		groups = self.neutronGroupsUsed()
		lst = [("?Error",0)]
		for g in groups:
			i = 1
			for mat in Fluka.lowMaterials.get(g,[]):
				if len(groups)==1:
					lst.append((f"{mat.elem}. {mat.desc}, {mat.temp}", i))
				else:
					lst.append((f"{g}: {mat.elem}. {mat.desc}, {mat.temp}", i))
				i += 1
		return lst

	# ----------------------------------------------------------------------
	# Papro (Particle properties or heavy ion)
	# Used in AUXSCORE and SPECSOU
	# ----------------------------------------------------------------------
	def _papro_what(self, card, what, val=None):
		w = card.intWhat(what)
		if val is None:
			if w <= -100:
				return ""
			else:
				return card.what(what)
		else:
			if val!="" or w>-100:
				self.setWhat(card, what, val)

	# ----------------------------------------------------------------------
	def _papro_Z(self, card, what, val=None):
		w = card.intWhat(what)
		Z = 0
		A = 0
		M = 0
		if w<=-100:
			i = -w // 100
			i, Z = divmod(i, 1000)
			M, A = divmod(i, 1000)
		if val is None:
			return Z
		else:
			try: Z = int(val)
			except: pass
			self.setWhat(card, what, -(Z*100 + A*100000 + M*100000000))

	# ----------------------------------------------------------------------
	def _papro_A(self, card, what, val=None):
		w = card.intWhat(what)
		Z = 0
		A = 0
		M = 0
		if w<=-100:
			i = -w // 100
			i, Z = divmod(i, 1000)
			M, A = divmod(i, 1000)
		if val is None:
			return A
		else:
			try: A = int(val)
			except: pass
			self.setWhat(card, what, -(Z*100 + A*100000 + M*100000000))

	# ----------------------------------------------------------------------
	def _papro_M(self, card, what, val=None):
		w = card.intWhat(what)
		Z = 0
		A = 0
		M = 0
		if w<=-100:
			i = -w // 100
			i, Z = divmod(i, 1000)
			M, A = divmod(i, 1000)
		if val is None:
			return M
		else:
			try: M = int(val)
			except: pass
			self.setWhat(card, what, -(Z*100 + A*100000 + M*100000000))

	# ----------------------------------------------------------------------
	@staticmethod
	def _papro_isotope(card, what):
		w = card.what(what)
		if w != "":
			try:    return int(w)<=-100
			except: return False
		return True

	# ----------------------------------------------------------------------
	# Papro (Particle properties or heavy ion)
	# Used in SPOTBEAM
	def _paprospot_what(self, card, val=None):
		w = card.what(1)
		if val is None:
			if w != "":
				try:
					w = int(w)
					if w==0 or w>1000: return ""
				except:
					return w
			else:
				return w
		else:
			self.setWhat(card, 1, val)

	# ----------------------------------------------------------------------
	@staticmethod
	def _paprospot_isotope(card):
		w = card.what(1)
		if w != "":
			try:
				w = int(w)
				return w==0 or w>1000
			except:
				return False
		return True

	# ----------------------------------------------------------------------
	def _paprospot_Z(self, card, val=None):
		w = card.intWhat(1)
		Z = 0
		A = 0
		M = 0
		if w>=1000:
			i, A = divmod(w,1000)
			M, Z = divmod(i, 100)
		if val is None:
			return Z
		else:
			try: Z = int(val)
			except: pass
			self.setWhat(card, 1, A + 1000*Z + 100000*M)

	# ----------------------------------------------------------------------
	def _paprospot_A(self, card, val=None):
		w = card.intWhat(1)
		Z = 0
		A = 0
		M = 0
		if w>=1000:
			i, A = divmod(w,1000)
			M, Z = divmod(i, 100)
		if val is None:
			return A
		else:
			try: A = int(val)
			except: pass
			self.setWhat(card, 1, A + 1000*Z + 100000*M)

	# ----------------------------------------------------------------------
	def _paprospot_M(self, card, val=None):
		w = card.intWhat(1)
		Z = 0
		A = 0
		M = 0
		if w>=1000:
			i, A = divmod(w,1000)
			M, Z = divmod(i, 100)
		if val is None:
			return M
		else:
			try: M = int(val)
			except: pass
			self.setWhat(card, 1, A + 1000*Z + 100000*M)

	# ----------------------------------------------------------------------
	# What type/val Lambdas
	# ----------------------------------------------------------------------
	def _what1_papro_what(self, c, v=None):	return self._papro_what(c,1,v)
	def _what1_papro_Z(self, c, v=None):	return self._papro_Z(c,1,v)
	def _what1_papro_A(self, c, v=None):	return self._papro_A(c,1,v)
	def _what1_papro_M(self, c, v=None):	return self._papro_M(c,1,v)

	def _what2_papro_what(self, c, v=None):	return self._papro_what(c,2,v)
	def _what2_papro_Z(self, c, v=None):	return self._papro_Z(c,2,v)
	def _what2_papro_A(self, c, v=None):	return self._papro_A(c,2,v)
	def _what2_papro_M(self, c, v=None):	return self._papro_M(c,2,v)

	# ----------------------------------------------------------------------
	# Return what to labels conversion
	# ----------------------------------------------------------------------
	def what2labels(self, layout):
		# What to labels conversion
		# Find fields to highlight
		labels = [""]*6
		for lbl,dat,*desc in layout:
			if isinstance(dat,tuple):
				dat = dat[0]
			elif isinstance(dat,int):
				while dat>=len(labels): labels.append("")
				labels[dat] = lbl
		return labels

	# ----------------------------------------------------------------------
	# Absolute WHATS
	# ----------------------------------------------------------------------
	def _what_abs_type(self, card, what, val):
		if val is None:
			if card.isEvalWhat(what):
				e = card.numWhat(what)
				if e > 0.0:
					return 0
				elif e < 0.0:
					return 1
				else:
					return Layout.FUNCTION
				return Layout.FUNCTION
			elif card.sign(what):
				return 1
			else:
				return 0
		else:
			self.setSign(card, what, bool(val))

	# ----------------------------------------------------------------------
	def _what_abs_val(self, card, what, val):
		if val is None:
			return card.absWhat(what)
		else:
			self.setAbsWhat(card, what, val)

	# ----------------------------------------------------------------------
	def _what_bit_val(self, card, what, bit, val, max_=1022):
		w = min(card.intWhat(what), max_)
		mask = 1<<bit
		if val is None:
			if w<0:
				return w
			elif w & mask > 0:
				return 1
			else:
				return 0
		else:
			if val<0:
				self.setWhat(card, what, val)
				return
			w = max(w, 0)
			if val:
				self.setWhat(card, what, w | mask)
			else:
				self.setWhat(card, what, w & ~mask)

	# ----------------------------------------------------------------------
	# WARNING reverse order!
	# ----------------------------------------------------------------------
	def _what_digit_val(self, card, what, digit, val, length=9):
		w = "%0*d"%(length, card.intWhat(what))
		if val is None:
			return int(w[digit])
		else:
			self.setWhat(card, what,
				"%s%d%s" % (w[:digit], val, w[digit+1:]))

	# ----------------------------------------------------------------------
	# What type/val Lambdas
	# ----------------------------------------------------------------------
	def _what1_type(self,c,v=None): return self._what_abs_type(c,1,v)
	def _what1_val( self,c,v=None): return self._what_abs_val( c,1,v)
	def _what2_type(self,c,v=None): return self._what_abs_type(c,2,v)
	def _what2_val( self,c,v=None): return self._what_abs_val( c,2,v)
	def _what3_type(self,c,v=None): return self._what_abs_type(c,3,v)
	def _what3_val( self,c,v=None): return self._what_abs_val( c,3,v)
	def _what4_type(self,c,v=None): return self._what_abs_type(c,4,v)
	def _what4_val( self,c,v=None): return self._what_abs_val( c,4,v)
	def _what5_type(self,c,v=None): return self._what_abs_type(c,5,v)
	def _what5_val( self,c,v=None): return self._what_abs_val( c,5,v)
	def _what6_type(self,c,v=None): return self._what_abs_type(c,6,v)
	def _what6_val( self,c,v=None): return self._what_abs_val( c,6,v)

	# ----------------------------------------------------------------------
	# initialize layout and order cards...
	# ----------------------------------------------------------------------
	def init(self):
		order = [("TITLE"     , Layout.generic_layout ),
			( "FIXED"     , Layout.empty_layout),
			( "FREE"      , Layout.empty_layout),
			( "GLOBAL"    , _GLOBAL_layout4   ),
			( "DEFAULTS"  , _DEFAULTS_layout4 ),
			( "&BEAM"     , _BEAM_layout4     ),
			( "BEAM&POS"  , _BEAMPOS_layout4  ),
			( "BEAM&AXES" , _BEAMAXES_layout4 ),
			( "HI-PROPE"  , _HI_PROPE_layout4 ),
			( "SO&URCE"   , _SOURCE_layout4   ),
			( "SPECSOUR"  , _SPECSOUR_layout4 ),
			( "SPOTBEAM"  , _SPOTBEAM_layout4 ),
			( "SPOTPOS"   , _SPOTPOS_layout4  ),
			( "SPOTDIR"   , _SPOTDIR_layout4  ),
			( "SPOTTRAN"  , _SPOTTRAN_layout4 ),
			( "GCR-SPE"   , _GCR_SPE_layout4  ),
			( "POLARIZA"  , _POLARIZA_layout4 ),
			( "DISCARD"   , _DISCARD_layout4  ),
			( "BME"       , _BME_layout4      ),
			( "DELTARAY"  , _DELTARAY_layout4 ),
			( "DPMJET"    , _DPMJET_layout4   ),
			( "EMFFLUO"   , _EMFFLUO_layout4  ),
			( "EMFRAY"    , _EMFRAY_layout4   ),
			( "EVENTYPE"  , _EVENTYPE_layout4 ),
			( "IONFLUCT"  , _IONFLUCT_layout4 ),
			( "IONTRANS"  , _IONTRANS_layout4 ),
			( "MULSOPT"   , _MULSOPT_layout4  ),
			( "MUPHOTON"  , _MUPHOTON_layout4 ),
			( "MYRQMD"    , _MYRQMD_layout4   ),
			( "OPT-PROD"  , _OPT_PROD_layout4 ),
			( "PAIRBREM"  , _PAIRBREM_layout4 ),
			( "PHOTONUC"  , _PHOTONUC_layout4 ),
			( "PHYSICS"   , _PHYSICS_layout4  ),
			( "SYRASTEP"  , _SYRASTEP_layout4 ),
			( "RQMD"      , _RQMD_layout4     ),
			( "GEOBEGIN"  , _GEOBEGIN_layout4 ),
			( "VOXELS"    , _VOXELS_layout4   ),
			( "&RPP"      , Layout.generic_layout ),
			( "&BOX"      , Layout.generic_layout ),
			( "&SPH"      , Layout.generic_layout ),
			( "R&CC"      , Layout.generic_layout ),
			( "R&EC"      , Layout.generic_layout ),
			( "&TRC"      , Layout.generic_layout ),
			( "&TRX"      , Layout.generic_layout ),
			( "&TRY"      , Layout.generic_layout ),
			( "&TRZ"      , Layout.generic_layout ),
			( "E&LL"      , Layout.generic_layout ),
			( "&WED"      , Layout.generic_layout ),
			( "RAW"       , Layout.generic_layout ),
			( "&ARB"      , Layout.generic_layout ),
			( "&PLA"      , Layout.generic_layout ),
			( "&QUA"      , Layout.generic_layout ),
			( "&YZP"      , Layout.generic_layout ),
			( "X&ZP"      , Layout.generic_layout ),
			( "&XYP"      , Layout.generic_layout ),
			( "XCC"       , Layout.generic_layout ),
			( "YCC"       , Layout.generic_layout ),
			( "ZCC"       , Layout.generic_layout ),
			( "XEC"       , Layout.generic_layout ),
			( "YEC"       , Layout.generic_layout ),
			( "ZEC"       , Layout.generic_layout ),
			( "END"       , Layout.empty_layout),
			( "$start_expansion", _START_EXPANSION_layout4 ),
			( "$start_translat",  _START_TRANSLAT_layout4  ),
			( "$start_transform", _START_TRANSFORM_layout4 ),
			( "$end_expansion",   _END_EXPANSION_layout4   ),
			( "$end_translat",    _END_TRANSLAT_layout4    ),
			( "$end_transform",   _END_TRANSFORM_layout4   ),
			( "REGION"    , _REGION_layout4   ),
			( "LATTICE"   , _LATTICE_layout4  ),
			( "LATTSNGL"  , _LATTSNGL_layout4 ),
			( "GEO&END"   , _GEOEND_layout4   ),
			( "&MATERIAL" , _MATERIAL_layout4 ),
			( "COMPOUND"  , _COMPOUND_layout4 ),
			( "LOW-MAT"   , _LOW_MAT_layout4  ),
			( "&OPT-PROP" , _OPT_PROP_layout4 ),
			( "MAT-&PROP" , _MAT_PROP_layout4 ),
			( "CORRFACT"  , _CORRFACT_layout4 ),
			( "STERNHEI"  , _STERNHEI_layout4 ),
			( "ASSIGNMA"  , _ASSIGNMA_layout4 ),
			( "EMF"       , _EMF_layout4      ),
			( "EMF&CUT"   , _EMFCUT_layout4   ),
			( "EMFFIX"    , _EMFFIX_layout4   ),
			( "ELCFIELD"  , _ELCFIELD_layout4 ),
			( "EXPTRANS"  , _EXPTRANS_layout4 ),
			( "&FLUKAFIX" , _FLUKAFIX_layout4 ),
			( "&LOW-NEUT" , _LOW_NEUT_layout4 ),
			( "LOW-PWXS"  , _LOW_PWXS_layout4  ),
			( "MCSTHRES"  , _MCSTHRES_layout4 ),
			( "MGNFIELD"  , _MGNFIELD_layout4 ),
			( "&PART-THR" , _PART_THR_layout4 ),
			( "STEPSIZE"  , _STEPSIZE_layout4 ),
			( "THRESHOL"  , _THRESHOL_layout4 ),
			( "TIME-&CUT" , _TIME_CUT_layout4 ),
			( "BIASING"   , _BIASING_layout4  ),
			( "EMF-BIAS"  , _EMF_BIAS_layout4 ),
			( "LAM-BIAS"  , _LAM_BIAS_layout4 ),
			( "LOW-BIAS"  , _LOW_BIAS_layout4 ),
			( "LOW-DOWN"  , _LOW_DOWN_layout4 ),
			( "WW-FACTO"  , _WW_FACTO_layout4 ),
			( "WW-PROFI"  , _WW_PROFI_layout4 ),
			( "WW-THRES"  , _WW_THRES_layout4 ),
			( "BAMJET"    , _BAMJET_layout4   ),
			( "DPM-PARA"  , _DPM_PARA_layout4 ),
			( "EVXTEST"   , _EVXTEST_layout4  ),
			( "DCYSCORE"  , _DCYSCORE_layout4 ),
			( "DCYTIMES"  , _DCYTIMES_layout4 ),
			( "IRRPROFI"  , _IRRPROFI_layout4 ),
			( "RADDECAY"  , _RADDECAY_layout4 ),
			( "ROT-&DEFI" , _ROT_DEFI_layout4 ),
			( "ROTPRBIN"  , _ROTPRBIN_layout4 ),
			( "AUXSCORE"  , _AUXSCORE_layout4 ),
			( "TCQUENCH"  , _TCQUENCH_layout4 ),
			( "DETECT"    , _DETECT_layout4   ),
			( "DETGEB"    , _DETGEB_layout4   ),
			( "USR&BIN"   , _USRBIN_layout4   ),
			( "USRBD&X"   , _USRBDX_layout4   ),
			( "USR&COLL"  , _USRCOLL_layout4  ),
			( "USR&TRACK" , _USRTRACK_layout4 ),
			( "USR&YIELD" , _USRYIELD_layout4 ),
			( "&EVENTBIN" , _EVENTBIN_layout4 ),
			( "EVENTDAT"  , _EVENTDAT_layout4 ),
			( "&USERDUMP" , _USERDUMP_layout4 ),
			( "USER&WEIG" , _USERWEIG_layout4 ),
			( "&RESNUCLE" , _RESNUCLE_layout4 ),
			( "RAD-BIOL"  , _RAD_BIOL_layout4 ),
			( "TPSSCORE"  , _TPSSCORE_layout4 ),
			( "&SCORE"    , _SCORE_layout4    ),
			( "OPEN"      , _OPEN_layout4     ),
			( "USRGCALL"  , _USRGCALL_layout4 ),
			( "USRICALL"  , _USRICALL_layout4 ),
			( "USROCALL"  , _USROCALL_layout4 ),
			( "PLOTGEOM"  , _PLOTGEOM_layout4 ),
			( "&RANDOMIZ" , _RANDOMIZ_layout4 ),
			( "&START"    , _START_layout4    ),
			( "STOP"      , _STOP_layout4     )]
		self._init_set(order, Fluka.MODE)

#-------------------------------------------------------------------------------
# Common types
#-------------------------------------------------------------------------------
_log_what      = [("Linear", 1), ("Log",  -1)]

__mat456_range = [ ("Mat",	(4, FlukaLayout.materialList)),
		   ("to Mat",	(5, FlukaLayout.materialList)),
		   ("Step",	 6)]
__reg456_range = [ ("Reg",	(4, FlukaLayout.regionList)),
		   ("to Reg",	(5, FlukaLayout.regionList)),
		   ("Step",	 6)]
_score_type = [	"",
		"EVENTBIN",
		"RESNUCLE",
		"USRBDX",
		"USRBIN",
		"USRCOLL",
		"USRTRACK",
		"USRYIELD" ]

#===============================================================================
# Layouts
#===============================================================================
_generic_layout4_1 = [
		Layout.tab,
		("#1", 1),
		("#2", 2),
		("#3", 3),
		Layout.newline,
	("sdum", 0),
		("#4", 4),
		("#5", 5),
		("#6", 6)]

#-------------------------------------------------------------------------------
def _generic_layoutN(nwhats):
	n,m = divmod(nwhats-1,6)
	if m==0: n -= 1
	layout = _generic_layout4_1[:]
	i = 7
	while n>0:
		n -= 1

		layout.append(Layout.newline)
		layout.append(Layout.tab)
		for _ in range(3):
			layout.append((f"#{i}",i))
			i += 1

		layout.append(Layout.newline)
		layout.append(Layout.tab)

		for _ in range(3):
			layout.append((f"#{i}",i))
			i += 1
	return layout

#-------------------------------------------------------------------------------
def _generic_layout4(card):
	if card.nwhats()<=7:
		return _generic_layout4_1
	else:
		return _generic_layoutN(card.nwhats())

#-------------------------------------------------------------------------------
_generic_layout = _generic_layout4

#-------------------------------------------------------------------------------
_BAMJET_layout4   = _generic_layout4	# Developers
_DPM_PARA_layout4 = _generic_layout4	# Developers
_RQMD_layout4     = _generic_layout4	# Developers

_BME_layout4      = _generic_layout4	# not implemented yet
_EXPTRANS_layout4 = _generic_layout4	# not implemented yet
_MYRQMD_layout4   = _generic_layout4	# not implemented yet

#================================ Special Cards ================================
_SOURCE_layout4   = _generic_layoutN(18)
_USRGCALL_layout4 = _generic_layout4
_USRICALL_layout4 = _generic_layout4
_USROCALL_layout4 = _generic_layout4

#================================ Geometry Cards ===============================
#-------------------------------- $Start_expansion -----------------------------
_START_EXPANSION_layout4 = [
	Layout.tab,
		("f",		0)]

#-------------------------------- $End_expansion -------------------------------
_END_EXPANSION_layout4 = Layout.empty_layout

#-------------------------------- $Start_translat ------------------------------
_START_TRANSLAT_layout4 = [
	Layout.tab,
		("dx",		0),
		("dy",		1),
		("dz",		2) ]

#-------------------------------- $End_translat --------------------------------
_END_TRANSLAT_layout4 = Layout.empty_layout

#-------------------------------- $Start_transform -----------------------------
_START_TRANSFORM_layout4 = [
	Layout.tab,
		( "Trans",	(0, Layout.rotationList))]

#-------------------------------- $End_transform -------------------------------
_END_TRANSFORM_layout4 = Layout.empty_layout

#-------------------------------- REGION ---------------------------------------
def __REGION_volume(layout, card, val=None):
	if val is None:
		try: return card[Card.VOLUME]
		except: return ""
	else:
		card[Card.VOLUME] = val

__REGION_layout4_definition = [
		Layout.name,
		Layout.tab,
		("Neigh",  1),
		Layout.newline,
		("expr",  -1)]

__REGION_layout4_cont = [ (None, -1) ]

def _REGION_layout4(card):
	if card.name() == "&":
		return __REGION_layout4_cont
	else:
		return __REGION_layout4_definition

#===================== Normal Cards ===================
#------------------------------- ASSIGNMAt -------------------------------------
_ASSIGNMA_layout4 = [ Layout.tab,
		("Mat",		(1, FlukaLayout.materialAllList)),
		("Reg",		(2, FlukaLayout.regionList)),
		("to Reg",	(3, FlukaLayout.regionList)),
		Layout.newline,
	Layout.tab,
		("Mat(Decay)",	(6, FlukaLayout.materialList)),
		("Step",	 4),
		("Field",	(5, FlukaLayout.ASSIGNMA_field))]

#------------------------------- AUXSCORE --------------------------------------
def _AUXSCORE_layout4(card):
	layout = [ Layout.tab,
		("Type",	(1, _score_type)),
		("Part",       (FlukaLayout._what2_papro_what, FlukaParticle.particles),
				"Particle type or heavy ion"),
		("Set",		(0, [
		            "",
		            "AMB74",
					"AMBGS",
					"EAP74",
					"ERT74",
					"EWT74",
					"EAPMP",
					"ERTMP",
					"EWTMP",
					"EDAP",
					"EDPA",
					"EDISO",
					"EDLLAT",
					"EDRLAT",
					"EDROT",
					"EDWORST",
					"HS",
					"HP000",
					"HP015",
					"HP030",
					"HP045",
					"HP060",
					"HP075",
					"HP090",
					"HP180",
					"HPISO",
					"HPISI",
					"HPSSI",
					"HPROT",
					"DE000",
					"DE015",
					"DE030",
					"DE045",
					"DE060",
					"DE075",
					"DE090",
					"DEROT",
					"DS000",
					"DS015",
					"DS030",
					"DS045",
					"DS060",
					"DS075"
					])),
		Layout.newline,
		("Delta Ray",	(3, [	Layout.empty_what,
					("assign to self",0),
					("assign to parent", 1),
					Layout.reset_what])) ]

	if FlukaLayout._papro_isotope(card, 2):
		layout.extend([	("Z",		FlukaLayout._what2_papro_Z, "Heavy ion charge"),
				("A",		FlukaLayout._what2_papro_A, "Heavy ion mass"),
				("Isomer",	FlukaLayout._what2_papro_M, "Heavy ion isomer"),
				Layout.newline,
				Layout.tab ])

	layout.extend([
		    ("Det",	(4, lambda l,c=card:l.detectorList(c.what(1)))),
		    ("to Det",	(5, lambda l,c=card:l.detectorList(c.what(1)))),
		    ("Step",	 6)])
	return layout

#------------------------------- BEAM ------------------------------------------
__BEAM_type0 = [ ("Momentum",0), ("Energy",1), ("Function",Layout.FUNCTION) ]
__BEAM_type1 = [ ("Flat",0), ("Gauss",1), ("Function",Layout.FUNCTION) ]
__BEAM_type2 = [ ("Flat",0), ("Gauss",1), ("Isotropic",2), ("Function", Layout.FUNCTION) ]
__BEAM_type4 = [ ("Rectangular",0), ("Annular",1), ("Gauss",2), ("Function", Layout.FUNCTION) ]
__BEAM_type5 = [ ("Rectangular",0), ("Gauss",2), ("Function", Layout.FUNCTION) ]

#-------------------------------------------------------------------------------
# Beam divergence
#-------------------------------------------------------------------------------
def __BEAM_div_type(layout, card, val=None, w=3):
	if val is None:
		if card.isEvalWhat(w):
			e = card.numWhat(w)
			if e > 0.0:
				#if e >= 2000.0*math.pi:
				#	return 2
				return 0
			elif e < 0.0:
				return 1
			else:
				return Layout.FUNCTION
		elif card.sign(w):
			return 1
		elif abs(card.numWhat(w)) >= 2000.0*math.pi:
			return 2
		return 0
	else:
		n = card.absWhat(w)
		try: n = float(n)
		except: n = 0.0
		if val==2:
			if n <= 2000.0*math.pi:
				layout.setWhat(card, w, 10000.0+n)
		else:
			if n >= 2000.0*math.pi:
				n -= 10000.0
				if n >= 2000.0*math.pi: n = 0.0
				layout.setWhat(card, w, n)

			if val==0:
				layout.setSign(card, w, False)
			elif val==1:
				layout.setSign(card, w, True)

#-------------------------------------------------------------------------------
# Horizontal beam shape
#-------------------------------------------------------------------------------
def __BEAM_shape_type4(layout, card, val=None):
	if val is None:
		if card.isEvalWhat(4):
			e = card.numWhat(4)
			if e < 0.0:
				return 2
			elif e > 0.0:
				if card.numWhat(6)<0.0:
					return 1
				return 0
			else:
				return Layout.FUNCTION
		elif card.sign(4):
			return 2
		else:
			if card.sign(6):
				return 1
			return 0
	else:
		if val==0:	# Flat
			undoinfo = [
				layout.flair.setSignUndo(card, 4,  False),
				layout.flair.setSignUndo(card, 6,  False)]
		elif val==1:	# Annular
			if card.intWhat(6) == 0:
				undoinfo = [layout.flair.setWhatUndo(card, 6, 1.0)]
			else:
				undoinfo = []
			undoinfo.extend([
				layout.flair.setSignUndo(card, 4,  False),
				layout.flair.setSignUndo(card, 5,  False),
				layout.flair.setSignUndo(card, 6,  True) ])
		else:		# Gauss
			undoinfo = [
				layout.flair.setSignUndo(card, 4,  True),
				layout.flair.setSignUndo(card, 6,  False)]
		layout.flair.addUndo(undoinfo)

#-------------------------------------------------------------------------------
# Vertical beam shape
#-------------------------------------------------------------------------------
def __BEAM_shape_type5(layout, card, val=None):
	if val is None:
		if card.isEvalWhat(5):
			e = card.numWhat(5)
			if e < 0.0:
				return 2
			elif e > 0.0:
				return 0
			else:
				return Layout.FUNCTION
		elif card.sign(5):
			return 2
		else:
			return 0
	else:
		if val==0:	# Flat
			undoinfo = [
				layout.flair.setSignUndo(card, 5,  False),
				layout.flair.setSignUndo(card, 6,  False)]
		else:
			undoinfo = [
				layout.flair.setSignUndo(card, 5,  True),
				layout.flair.setSignUndo(card, 6,  False)]
		layout.flair.addUndo(undoinfo)

#-------------------------------------------------------------------------------
# Extend layout with the beam shape
#-------------------------------------------------------------------------------
def __BEAM_shape_layout(card, layout):
	layout.append(("Shape(X)", (__BEAM_shape_type4, __BEAM_type4),
			"Beam shape in X axis"))
	if card.sign(6):
		layout.append(("Rmin",	 FlukaLayout._what5_val,
				"WHAT(5): Minimum radius of annular beam [cm]"))
		layout.append(("Rmax",	 FlukaLayout._what4_val,
				"WHAT(4): Maximum radius of annular beam [cm]"))
		layout.append(Layout.tab)
	else:
		if card.sign(4):
			layout.append(("x(FWHM)",  FlukaLayout._what4_val,
					"WHAT(4): Gaussian FWHM in x-direction [cm]"))
		else:
			layout.append(("Δx",	 FlukaLayout._what4_val,
					"WHAT(4): Beam width in x-direction [cm]"))

		layout.append(("Shape(Y)", (__BEAM_shape_type5, __BEAM_type5),
				"Beam shape in Y axis"))
		if card.sign(5):
			layout.append(("y(FWHM)",	 FlukaLayout._what5_val,
					"WHAT(5): Gaussian FWHM in y-direction [cm]"))
		else:
			layout.append(("Δy",	 FlukaLayout._what5_val,
					"WHAT(5): Beam width in y-direction [cm]"))

#-------------------------------------------------------------------------------
# Generate a dynamic layout
#-------------------------------------------------------------------------------
def _BEAM_layout4(card):
	# 1st line
	layout = [ Layout.tab,
		("Beam", (FlukaLayout._what1_type, __BEAM_type0),
			"WHAT(1) sign: Energy or Momentum") ]

	if card.sign(1):
		layout.append(("E",	 FlukaLayout._what1_val,
				"WHAT(1): Energy [GeV or GeV/n for Heavy Ions]"))
	else:
		layout.append(("p",	 FlukaLayout._what1_val,
				"WHAT(1): Momentum [GeV/c]"))

	layout.append(("Part",	(0, FlukaParticle.beam)))
	layout.append(Layout.newline)

	# 2nd line
	layout.append(("Δp",	(FlukaLayout._what2_type, __BEAM_type1),
			"WHAT(2): Momentum spread type"))
	if card.sign(2):
		layout.append(("Δp(FWHM)", FlukaLayout._what2_val,
			"WHAT(2): Momentum spread Gaussian [GeV/c] FWHM=2.355*sigma"))
	else:
		layout.append(("Δp",	FlukaLayout._what2_val,
			"WHAT(2): Momentum spread FLAT [GeV/c]"))

	layout.append(("Δϕ",	(__BEAM_div_type, __BEAM_type2),
			"WHAT(3): Divergence type"))
	if card.numWhat(3) >= 2000.0*math.pi:
		layout.append(Layout.tab)
	elif card.sign(3):
		layout.append(("Δϕ (FWHM)", FlukaLayout._what3_val,
				"WHAT(3): Divergence [mrad]"))
	else:
		layout.append(("Δϕ", FlukaLayout._what3_val,
				"WHAT(3): Divergence [mrad]"))
	layout.append(Layout.newline)

	# 3rd line
	__BEAM_shape_layout(card, layout)

	return layout

#------------------------------- BEAMAXES --------------------------------------
_BEAMAXES_layout4 = [ Layout.tab,
		("cosBxx",	1),
		("cosBxy",	2),
		("cosBxz",	3),
		Layout.newline,
	Layout.tab,
		("cosBzx",	4),
		("cosBzy",	5),
		("cosBzz",	6) ]

#------------------------------- BEAMPOS ---------------------------------------
__BEAMPOS_sdum = [("POSITIVE",""),
		"NEGATIVE",
		"FLOOD",
		"CART-VOL",
		"CYLI-VOL",
		"SPHE-VOL" ]
__BEAMPOS_layout4 = [ Layout.tab,
		("x",		1),
		("y",		2),
		("z",		3),
		Layout.newline,
	Layout.tab,
		("cosx",	4),
		("cosy",	5),
		("Type",	(0, __BEAMPOS_sdum))]

__BEAMPOS_flood_layout4 = [ Layout.tab,
		("R",		1),
		Layout.tab,
		("Type",	(0, __BEAMPOS_sdum))]

__BEAMPOS_cart_layout4 = [ Layout.tab,
		("Xin",		1),
		("Xout",	2),
		("Type",	(0, __BEAMPOS_sdum)),
		Layout.newline,
	Layout.tab,
		("Yin",		3),
		("Yout",	4),
		Layout.newline,
	Layout.tab,
		("Zin",		5),
		("Zout",	6),
		Layout.tab ]

__BEAMPOS_cyli_layout4 = [ Layout.tab,
		("Rin",		1),
		("Rout",	2),
		("Type",	(0, __BEAMPOS_sdum)),
		Layout.newline,
	Layout.tab,
		("Hin",		3),
		("Hout",	4),
		Layout.tab ]

__BEAMPOS_sphe_layout4 = [ Layout.tab,
		("Rin",		1),
		("Rout",	2),
		("Type",	(0, __BEAMPOS_sdum))]

def _BEAMPOS_layout4(card):
	if card.sdum() == "FLOOD":
		return __BEAMPOS_flood_layout4
	elif card.sdum() == "CART-VOL":
		return __BEAMPOS_cart_layout4
	elif card.sdum() == "CYLI-VOL":
		return __BEAMPOS_cyli_layout4
	elif card.sdum() == "SPHE-VOL":
		return __BEAMPOS_sphe_layout4
	else:
		return __BEAMPOS_layout4

#------------------------------- BIASING ---------------------------------------
__BIASING_types = [	Layout.empty_what,
			("All regions",    -1),
			("All particles",   0),
			("Hadrons & Muons", 1),
			("e-e+,γ",          2),
			("Low neutrons",    3) ]
__BIASING_sdum = [ "", "PRINT", "NOPRINT", "USER", "NOUSER", "RRPRONLY" ]
__BIASING_prim = [ "", "PRIMARY", "NOPRIMAR" ]

__BIASING_layout4_regions = [ Layout.tab,
		("Type",	(1, __BIASING_types)),
		("RR",		2),
		("Imp",		3),
		Layout.newline,
	("Opt",		(0, __BIASING_sdum)) ] \
		+ __reg456_range

__BIASING_layout4_particles   = [ Layout.tab,
		("Type",	(1, __BIASING_types)),
		("Mod. M",	2),
		Layout.newline,
	("Opt",		(0, __BIASING_prim)),
		("Part",	(3, FlukaParticle.list)),
		("to Part",	(4, FlukaParticle.list)),
		("Step",	 5) ]

def _BIASING_layout4(card):
	if card.intWhat(1) < 0:
		return __BIASING_layout4_particles
	else:
		return __BIASING_layout4_regions

#------------------------------- COMPOUND --------------------------------------
__COMPOUND_list = [ ("Atom", 0), ("Mass", 1), ("Volume", 2) ]

def __COMPOUND_components(layout, card, val=None):
	n = (card.nwhats()-1)//2
	if val is None:
		n = (n//3)*3+3	# round to the next 3 (base 0)
		return n
	elif val != n:
		nn = int(val)*2-1

		s1 = card.sign(1)
		s2 = card.sign(2)

		undoinfo = [ layout.flair.setNWhatsUndo(card, nn) ]
		for i in range(3,card.nwhats(),2):
			undoinfo.append( layout.flair.setSignUndo(card, i,   s1))
			undoinfo.append( layout.flair.setSignUndo(card, i+1, s2))
		layout.flair.addUndo(undoinfo, "Compound no. elements")

def __COMPOUND_type(layout, card, val=None):
	if val is None:
		s1 = card.sign(1)
		s2 = card.sign(2)
		if not s1 and not s2:
			return 0
		elif s1 and not s2:
			return 1
		return 2
	else:
		if val==0:
			s1 = False
			s2 = False
		elif val==1:
			s1 = True
			s2 = False
		else:
			s1 = True
			s2 = True

		undoinfo = []
		for i in range(1,card.nwhats(),2):
			undoinfo.append(layout.flair.setSignUndo(card, i,   s1))
			undoinfo.append(layout.flair.setSignUndo(card, i+1, s2))
		layout.flair.addUndo(undoinfo, "Compound type")

__COMPOUND_elements = [ (f"{x-2}..{x}",x) for x in range(3,82,3) ]

def _COMPOUND_layout4(card):
	if card.nwhats()<7:
		card.setNWhats(7)

	layout = [
		("Name",	(0, FlukaLayout.materialList)),
		Layout.tab,
		("Mix",		(__COMPOUND_type, __COMPOUND_list),
					"Type of fraction: Atom/Mass/Volume"),
		("Elements",	(__COMPOUND_components, __COMPOUND_elements),
					"Number of elements defining compound") ]

	n = ((card.nwhats()-1)//6)*6+6
	i = what = 0
	while what < n:
		if i%2==0: layout.append(Layout.newline)
		i    += 1
		what += 1
		layout.append((f"f{i}",
				lambda l,c,v=None,w=what:l._what_abs_val(c,w,v),
				f"Fraction of material #{i}"))
		what += 1
		layout.append(
			(f"M{i}",
				(lambda l,c,v=None,w=what:l._what_abs_val(c,w,v),
				lambda l,c=card: l.materialAllList(c.name())),
				f"Material #{i} name"))
	if (n//2)%2:
		layout.extend([Layout.tab, Layout.tab])

	return layout

#------------------------------- CORRFACT --------------------------------------
_CORRFACT_layout4 = [ Layout.tab,
		("xρ (dE/dx)",	1),
		("xρ other",	2),
		Layout.newline,
	Layout.tab,
		("Reg",		(4, FlukaLayout.regionList)),
		("to Reg",	(5, FlukaLayout.regionList)),
		("Step",	 6)]

#------------------------------- DCYTIMES --------------------------------------
_DCYTIMES_layout4 = [ Layout.tab,
		("t1",		1),
		("t2",		2),
		("t3",		3),
		Layout.newline,
	Layout.tab,
		("t4",		4),
		("t5",		5),
		("t6",		6) ]

#------------------------------- DCYSCORE --------------------------------------
def _DCYSCORE_layout4(card):
	return [ Layout.tab,
		("Cooling t",	(1, FlukaLayout.dcytimesList)),
		("Kind",	(0, _score_type)),
		Layout.newline,
	Layout.tab,
		("Det",		(4, lambda l,c=card:l.detectorList(c.name()))),
		("to Det",	(5, lambda l,c=card:l.detectorList(c.name()))),
		("Step",	6) ]

#------------------------------- DETECT - ---------------------------------------
if Global.developer:
	__DETECT_types = [ ("Detector", 0), ("Combination", 1) ]
else:
	__DETECT_types = [ ("Detector", 0), ]
__DETECT_coin = [ Layout.empty_what, ("Coincidence", 1), ("AntiCoincidence", -1) ]

# what(1)<=0 for detector, >0.0 for combination
# |what(1)|>1024: set as well the number of channels
def __DETECT_type(layout, card, val=None):
	w1 = card.intWhat(1)
	if val is None:
		return 0 if w1 <= 0 else 1
	else:
		try:    val = int(val)
		except: val = 0
		if abs(w1)<=1024:
			layout.setWhat(card, 1, val)
		elif val<=0:
			layout.setWhat(card, 1, -abs(w1))
		else:
			layout.setWhat(card, 1,  abs(w1))

def __DETECT_channels(layout, card, val=None):
	w1 = card.intWhat(1)
	if val is None:
		return 1024 if abs(w1)<=1024 else abs(w1)
	else:
		try:    val = int(val)
		except: val = 0
		if val<1024:
			layout.setWhat(card, 1, 0 if w1<=0 else 1)
		elif w1<=0:
			layout.setWhat(card, 1, -val)
		else:
			layout.setWhat(card, 1, val)

def __DETECT_regions(layout, card, val=None):
	# sdum whats first regs firstreg
	numberOfContinuationCards = ceil((card.nwhats() -1) /6) -1
	numberOfRegions = numberOfContinuationCards * 5 + 1
	if val is None:
		return max(numberOfRegions, 1)
	# Change number of regions
	else:
		val = int(val)
		if val != numberOfRegions:
			nn = ((val-1)//5+1)*6+1
			undoinfo = [ layout.flair.setNWhatsUndo(card, nn) ]
			for i in range(7, nn, 6):
				undoinfo.append(layout.flair.setWhatUndo(card, i, 0))
			layout.flair.addUndo(undoinfo)

def __DETECT_regions2(layout, card, val=None):
	# sdum whats regs
	numberOfRegions = ceil((card.nwhats() -1) /6) * 5
	if val is None:
		return numberOfRegions
	# Change number of regions
	else:
		val = int(val)
		if val != numberOfRegions:
			nn = (val//5)*6+1
			undoinfo = [ layout.flair.setNWhatsUndo(card, nn) ]
			w1 = card.what(1)
			for i in range(7, nn, 6):
				undoinfo.append(layout.flair.setWhatUndo(card, i, w1))
			layout.flair.addUndo(undoinfo)

def _DETECT_layout4(card):
	if card.intWhat(1)<=0:		# Detector
		layout = [ Layout.tab,
			("Type",	(__DETECT_type, __DETECT_types)),
			("Channels",	__DETECT_channels,
					"Number of channels if >1024 (default: 1024)"),
			("Name",	0),
			Layout.newline,
			("Emin",	2),
			("Emax",	3),
			("Ecut",	4),
			("Trigger",	(5, __DETECT_coin)),
			Layout.newline,
			("Regions",	(__DETECT_regions, list(range(1,12,5)))),
					#"Maximum number of regions/detectors to be considered"),

			("Reg1",	(6, FlukaLayout.signRegionList))]
		ic = 2	# column
		iw = 8	# what #
		numberOfContinuationCards = ceil((card.nwhats() -1) /6) -1
		# If card.nwhats() = 7 or less, there is 0 continuation card.
		# If card.nwhats() = 8, 9, etc, or 13, there is 1 continuation card.
		# etc...
		regionIndexMaxExclusive = numberOfContinuationCards * 5 + 2
		for regionIndex in range(2, regionIndexMaxExclusive):
			# Add new line when ic = 4 + k*4
			if ic%4 == 0: layout.append(Layout.newline)
			ic += 1
			# Region name
			layout.append((f"Reg{regionIndex}", (iw, FlukaLayout.signRegionList)))
			iw += 1
			# Skip `What(7 + k*6)`: only interested in `What` potentially adding a region!
			if (iw-1)%6 == 0: iw += 1
		return layout

	elif Global.developer:		# Combination
		layout = [ Layout.tab,
			("Type",	(__DETECT_type, __DETECT_types)),
			("Channels",	__DETECT_channels,
					"Number of channels if >1024 (default: 1024)"),
			("Name",	0),
			Layout.newline,
			("Regions",	(__DETECT_regions2, list(range(5,11,5))))]
		ic = 1	# column
		iw = 2	# what #
		numberOfRegionsGroups = ceil((card.nwhats() -1) /6)
		regionIndexMaxExclusive = numberOfRegionsGroups * 5 + 1
		for regionIndex in range(1, regionIndexMaxExclusive):
			# Add new line when ic = 4 + k*4
			if ic%4 == 0: layout.append(Layout.newline)
			ic += 1
			# Region name
			layout.append((f"Reg{regionIndex}", (iw, FlukaLayout.signRegionList)))
			iw += 1
			# Skip `What(7 + k*6)`: only interested in `What` potentially adding a region!
			if (iw-1)%6 == 0: iw += 1
		return layout

#------------------------------- DETGEB - ---------------------------------------
_DETGEB_layout4 = [ Layout.tab,
		("a",  1),
		("b",  2),
		("c",  3),
		Layout.newline,
	("where", (0, [("Both",""), "TRIGGER", "DETECTOR"])),
		("Det",    (4, lambda l:l.detectorList("DETECT"))),
		("to Det", (5, lambda l:l.detectorList("DETECT"))),
		("Step",    6)]

#------------------------------- DISCARD ---------------------------------------
# FIXME Heavy ions
_DISCARD_layout4 = [ Layout.tab,
		("p1",	(1, FlukaParticle.signedListAll)),
		("p2",	(2, FlukaParticle.signedListAll)),
		("p3",	(3, FlukaParticle.signedListAll)),
		Layout.newline,
	Layout.tab,
		("p4",	(4, FlukaParticle.signedListAll)),
		("p5",	(5, FlukaParticle.signedListAll)),
		("p6",	(6, FlukaParticle.signedListAll)) ]

#------------------------------- DEFAULTS --------------------------------------
__DEFAULTS_sdum = [ "CALORIME", "DAMAGE", "EM-CASCA", "EET/TRAN", "HADROTHE",
		"ICARUS", "NEUTRONS", "NEW-DEFA", "PRECISIO",
		"SHIELDIN" ]

_DEFAULTS_layout4 = [	Layout.tab,
			("",	(0, __DEFAULTS_sdum))]

#------------------------------- DELTARAY --------------------------------------
_DELTARAY_layout4 = [ Layout.tab,
		("E thres",	1),
		("# Log dp/dx", 2),
		("Log width dp/dx", 3),
		Layout.newline,
	("Print",	(0, ["NOPRINT", "PRINT"])) ] \
		+ __mat456_range

#------------------------------- DPMJET ----------------------------------------
__DPMJET_verb = [ Layout.empty_what,
		("Minimum", 0),
		("Minor Errors", 1),
		("Common Block", 2),
		("Final Residual", 3),
		("Final State Obj", 4)]
_DPMJET_layout4 = [
		Layout.tab,
		("Format", 1),
		("Unit", (2, lambda l: l.fileUnitList("DPMJET",False))),
		("Verbose", (3, __DPMJET_verb)) ]

#------------------------------- ELCFIELD --------------------------------------
_ELCFIELD_layout4 = [ Layout.tab,
		("Max Ang (deg)",	1),
		("Bound Acc. (cm)",	2),
		("Min step (cm)",	3),
		Layout.newline,
	Layout.tab,
		("Ex",		4),
		("Ey",		5),
		("Ez",		6),
		Layout.newline,
	Layout.tab,
		("f(dp/dx)",	0)]

#------------------------------- EMF -------------------------------------------
__EMF_switch = [ ("ON", ""), ("OFF", "EMF-OFF") ]
_EMF_layout4 = [ Layout.tab,
		("",		(0, __EMF_switch))]

#------------------------------- EMF-BIAS --------------------------------------
__EMF_BIAS_sdum = [ "", "LPBEMF", "LAMBEMF", "LAMBBREM", "LBRREMF", "LBRRBREM" ]

__EMF_BIAS_layout4_header = [ Layout.tab,
			("Type",	(0, __EMF_BIAS_sdum)) ]

__EMF_BIAS_layout4_LPB = __EMF_BIAS_layout4_header \
		+ [("Ethr e-e+", 2),
		   ("Ethr γ",    3),
		Layout.newline,
		("Old bremss.", (lambda l,c,v=None,b=0:l._what_bit_val(c,1,b,v),
					Layout.onoffreset_what),
				"LPB activated for bremsstrahlung and pair production (old default)"),
		("Bremsstrahlung",	(lambda l,c,v=None,b=1:l._what_bit_val(c,1,b,v),
						Layout.onoffreset_what),
				"LPB activated for bremsstrahlung"),
		("Pair Prod.",		(lambda l,c,v=None,b=2:l._what_bit_val(c,1,b,v),
						Layout.onoffreset_what),
				"LPB activated for pair production"),
		("e+ ann @rest",	(lambda l,c,v=None,b=3:l._what_bit_val(c,1,b,v),
						Layout.onoffreset_what),
				"LPB activated for positron annihilation at rest"),
		Layout.newline,
		("Compton",		(lambda l,c,v=None,b=4:l._what_bit_val(c,1,b,v),
						Layout.onoffreset_what),
				"LPB activated for Compton scattering"),
		("Bhabha&Moller",	(lambda l,c,v=None,b=5:l._what_bit_val(c,1,b,v),
						Layout.onoffreset_what),
				"LPB activated for Bhabha & Moller scattering"),
		("Photo-electric",	(lambda l,c,v=None,b=6:l._what_bit_val(c,1,b,v),
						Layout.onoffreset_what),
				"LPB activated for photoelectric effect"),
		("e+ ann @flight",	(lambda l,c,v=None,b=7:l._what_bit_val(c,1,b,v),
						Layout.onoffreset_what),
				"LPB activated for positron annihilation in flight"),
		Layout.newline, Layout.tab ] + __reg456_range

__EMF_BIAS_layout4_other = __EMF_BIAS_layout4_header + [
		("e <Path>",	 1),
		("γ <Path>",	 2),
		Layout.newline,
	("Generation",		 3) ] \
		+ __mat456_range

def _EMF_BIAS_layout4(card):
	sdum = card.sdum()
	if sdum in ("LPBEMF", ""):
		return __EMF_BIAS_layout4_LPB
	else:
		return __EMF_BIAS_layout4_other

#------------------------------- EMFCUT ----------------------------------------
__EMFCUT_sdum = [ ("transport",""),
			"ANNH-THR",
			"DELAYED",
			"ELPO-THR",
			"PHO2-THR",
			"PHOT-THR",
			"PROD-CUT",
			"PROMPT"]

__EMFCUT_ee = [ ("Total",0), ("Kinetic",1), ("Function",Layout.FUNCTION) ]

__EMFCUT_layout4_elpothr = [ Layout.tab,
		("Type", (0, __EMFCUT_sdum)),
		("e-e+ Brem",	1),
		("Bhabha/Moller",2),
		Layout.newline,
	("Photonuc",	3) ] \
		+ __mat456_range

__EMFCUT_layout4_annhthr = [ Layout.tab,
		("Type", (0, __EMFCUT_sdum)),
		("Ekin e+ ann",	1),
		Layout.newline,
	Layout.tab ] + __mat456_range

__EMFCUT_layout4_photthr = [ Layout.tab,
		("Type", (0, __EMFCUT_sdum)),
		("Compton",	1),
		("Photo-elec",	2),
		Layout.newline,
	("Pair prod",		3) ] \
		+ __mat456_range

__EMFCUT_layout4_pho2thr = [ Layout.tab,
		("Type",	 (0, __EMFCUT_sdum)),
		("Rayleigh",	1),
		("Photonuc",	2),
		Layout.newline,
	Layout.tab ] \
		+ __mat456_range

def _EMFCUT_layout4(card):
	sdum = card.sdum()
	if sdum=="ELPO-THR":
		return __EMFCUT_layout4_elpothr
	elif sdum=="ANNH-THR":
		return __EMFCUT_layout4_annhthr
	elif sdum=="PHOT-THR":
		return __EMFCUT_layout4_photthr
	elif sdum=="PHO2-THR":
		return __EMFCUT_layout4_pho2thr
	else:
		layout = [ Layout.tab,
			("Type",	(0, __EMFCUT_sdum)),
			Layout.newline,
		Layout.tab,
			("e-e+ Threshold",	(FlukaLayout._what1_type, __EMFCUT_ee))]

		if card.sign(1):
			layout.append(("e-e+ Ekin",	 FlukaLayout._what1_val,
					"WHAT(1): Electrons Kinetic Energy [GeV]"))
		else:
			layout.append(("e-e+ E",	 FlukaLayout._what1_val,
					"WHAT(1): Electrons Total Energy [GeV]"))
		layout.append(("γ",	2))
		layout.append(Layout.newline)
		if sdum=="PROD-CUT":
			layout.append(("Fudgem", 3))
			layout.extend(__mat456_range)
		else:
			layout.append(Layout.tab)
			layout.extend(__reg456_range)
		return layout

#------------------------------- EMFFIX ----------------------------------------
_EMFFIX_layout4 = [Layout.tab,
		("Mat1",	(1, FlukaLayout.materialList)),
		("Max Frac.1",	2),
		("Print",	(0, ["", "PRINT", "NOPRINT" ])),
		Layout.newline,
	Layout.tab,
		("Mat2",	(3, FlukaLayout.materialList)),
		("Max Frac.2",	4),
		Layout.newline,
	Layout.tab,
		("Mat3",	(5, FlukaLayout.materialList)),
		("Max Frac.3",	6) ]

#------------------------------- EMFFLUO ---------------------------------------
_EMFFLUO_layout4 = [ Layout.tab,
		("Fluorescence",(1, Layout.onoff3_what)),
		Layout.newline,
	Layout.tab,
		("Mat",		(2, FlukaLayout.materialList)),
		("to Mat",	(3, FlukaLayout.materialList)),
		("Step",	4) ]

#------------------------------- EMFRAY ----------------------------------------
_EMFRAY_corr = [ Layout.empty_what,
		Layout.ignore_what,
		("Rayleigh+Compton(bind)",	1),
		("Rayleight",		2),
		("Compton(bind)",	3),
		("R&C (bind+prof)",	4),
		("Rayleight (prof)",	5),
		("Compton (bind+prof)",	6),
		("Nothing",		-1) ]
_EMFRAY_layout4 = [ Layout.tab,
		("Activate",	(1, _EMFRAY_corr)),
		Layout.newline,
	Layout.tab,
		("Reg",		(2, FlukaLayout.regionList)),
		("to Reg",	(3, FlukaLayout.regionList)),
		("Step",	4) ]

#------------------------------- EVENTDAT --------------------------------------
_EVENTDAT_layout4 = [ Layout.tab,
		("Unit", (1, lambda l: l.fileUnitList("EVENTDAT"))),
		("File", 0) ]

#------------------------------- EVENTYPE --------------------------------------
__EVENTYPE_list = [ Layout.empty_what,
		("Reset", -2),
		("dE/dx only", 1),
		("All recoils", 2),
		("Recoils pid<=d (3)", 3),
		("Recoils pid<=t (4)", 4),
		("Recoils pid<=3-He (5)", 5),
		("Recoils pid<=4-He (6)", 6) ]
__EVENTYPE_model = [ "", "EVAP", "DPMJET" ]

_EVENTYPE_layout4 = [ Layout.tab,
		("Ions",	(3, __EVENTYPE_list)),
		("Model",	(0, __EVENTYPE_model)) ]

#------------------------------- EVXTEST ---------------------------------------
__EVXTEST_what5 = [ Layout.empty_what,
		("Include cascade+excitation", 0),
		("Exclude cascade+excitation", 1),
		("Evaporation", -1),
		("gamma Deexcitation", -2)]
__EVXTEST_what6 = [ Layout.empty_what,
		("Included", 0),
		("Excluded", 1) ]
__EVXTEST_sdum = ["PRINT", "" ]
_EVXTEST_layout4 = [ Layout.tab,
		("Events",	1),
		("Projectile",	(2, FlukaParticle.list)),
		("Momentum",	3),
		Layout.newline,
	("Mat",			(4, FlukaLayout.materialList)),
		("Cascade",	(5, __EVXTEST_what5)),
		("Diffractive",(6, __EVXTEST_what6)),
		("Print",	(0, __EVXTEST_sdum)) ]

#------------------------------- FLUKAFIX --------------------------------------
_FLUKAFIX_layout4 = [Layout.tab,
		("Ekin frac",	1),
		Layout.newline,
	Layout.tab ] + __mat456_range

#------------------------------- IONFLUCT --------------------------------------
__IONFLUCT_sdum = [ "", "PRIM-ION" ]

__IONFLUCT_accuracy = [ Layout.empty_what,
			("Minimal", 1), ("Low",     2),
			("High",    3), ("Maximum", 4),
			Layout.reset_what ]

__IONFLUCT_model = [ Layout.empty_what,
			("Model-1", 1), ("Model-2", 2),
			("Model-3", 3), ("Model-4", 4),
			Layout.reset_what ]

__IONFLUCT_other_layout4 = [ Layout.tab,
		("Type",	(0, __IONFLUCT_sdum)),
		("hadron",	(1, Layout.onoff3_what)),
		("e+e-",	(2, Layout.onoff3_what)),
		Layout.newline,
	("Accuracy",	(3, __IONFLUCT_accuracy))] \
		+ __mat456_range

__IONFLUCT_primion_layout4 = [ Layout.tab,
		("Type",	(0, __IONFLUCT_sdum)),
		("1st Vion",	1),
		("Prim Ion/cm", 2),
		Layout.newline,
	("Model",	(3, __IONFLUCT_model))] \
		+ __mat456_range

def _IONFLUCT_layout4(card):
	if card.sdum()=="PRIM-ION":
		return __IONFLUCT_primion_layout4
	else:
		return __IONFLUCT_other_layout4

#------------------------------- IONTRANS --------------------------------------
__IONTRANS_what1 = [ Layout.empty_what,
		Layout.ignore_what,
		("No transport of ions", 1),
		("Approximate", -1),
		("Full transport", -2),
		"HEAVYION",
		"DEUTERON",
		"TRITON",
		"3-HELIUM",
		"4-HELIUM"
		]

_IONTRANS_layout4 = [ Layout.tab,
	("Transport",	(1, __IONTRANS_what1)) ]

#------------------------------- IRRPROFI --------------------------------------
_IRRPROFI_layout4 = [ Layout.tab,
		("Δt",	1),
		("p/s",	2),
		Layout.newline,
	Layout.tab,
		("Δt",	3),
		("p/s",	4),
		Layout.newline,
	Layout.tab,
		("Δt",	5),
		("p/s",	6)]

#------------------------------- GEOBEGIN --------------------------------------
def __GEOBEGIN_geofile(layout, card, val=None):
	inp = layout.flair.project.input
	if val is None:
		return card.get(FlukaCard.GEOFILE,"")
	else:
		# WARNING: Displaying the dialog below sends a FocusOut event
		# to the input editor which will trigger a new commit
		# To avoid re-entry is protected with the following variable
		#global __GEOBEGIN_geofile_protection
		#try:
		#	if __GEOBEGIN_geofile_protection: return
		#except AttributeError:
		#	pass
		#__GEOBEGIN_geofile_protection = True

		# if geometry is empty then load the file
		geoFile = layout.relpath(val)
		if geoFile == card.get(FlukaCard.GEOFILE,""):
			#__GEOBEGIN_geofile_protection = False
			return
		try:
			empty = inp["GEOEND"][0].pos()-card.pos() <= 3
		except:
			#__GEOBEGIN_geofile_protection = False
			return
		if not empty:
			try:
				s = os.stat(geoFile)
				if s.st_size == 0:
					#__GEOBEGIN_geofile_protection = False
					return	# size is zero
			except OSError:
				layout.setCard(card, FlukaCard.GEOFILE, geoFile, "")
				#__GEOBEGIN_geofile_protection = False
				return	# or doesn't exist

			#ans = tkDialogs.Dialog(layout.flair,
			#	{"title" : "Replace with external geometry",
			#	"text"   : "You have selected an external geometry file, "
			#		   "while your input contains already a geometry.\n",
			#	"bitmap" : "questhead",
			#	"default": 2,
			#	"strings": ('Load from external','Save current to external','Cancel')})
			#if ans.num==1:
			#	layout.setCard(card, FlukaCard.GEOFILE, geoFile, "")
			#	__GEOBEGIN_geofile_protection = False
			#	return
			#elif ans.num==2:
			#	__GEOBEGIN_geofile_protection = False
			#	return

		undoinfo = [layout.flair.refreshUndo(), layout.flair.renumberUndo()]
		undoinfo.append(layout.flair.setCardUndo(card, FlukaCard.GEOFILE, geoFile, ""))

		# delete everything between GEOBEGIN GEOEND
		pos = card.pos()
		pos += 1
		while True:
			if inp.cardlist[pos].tag == "GEOEND":
				break
			undoinfo.append(layout.flair.delCardUndo(pos, False))
			#inp.delCard(pos)

		geoinp = Fluka()
		# Copy the old setting... correct it before!
		geoinp.geoFormat = layout.flair.project.input.geoFormat
		if card.name()[:8]=="COMBNAME":
			geoinp.geoFormat = FlukaCard.FORMAT_FREE

		try:	# prepare to load file
			geoinp.open(val)
			geoinp._parseGeometry(card)
			for card in geoinp.cardlist:
				undoinfo.append(layout.flair.addCardUndo(card, pos, False))
				#inp.add(card, pos)
				pos += 1
			geoinp.close()
		except:
			pass
		undoinfo.append(layout.flair.renumberUndo())
		undoinfo.append(layout.flair.refreshUndo())
		layout.flair.addUndo(undoinfo)
		#inp.renumber()
		#layout.flair.refresh("card")
		#__GEOBEGIN_geofile_protection = False

# -----------------------------------------------------------------------------
def __GEOBEGIN_geoout(layout, card, val=None):
	if val is None:
		return card.get(FlukaCard.GEOOUT,"")
	else:
		layout.setCard(card, FlukaCard.GEOOUT, layout.relpath(val), "")

__GEOBEGIN_layout4 = [Layout.tab,
		("Accuracy",	2),
		("Option",	(5, [Layout.empty_what, ("Default",0),
			("Logical", 1),
			("L+Plane", 2),
			("L+P+Bbox", 3)])),
		("Paren",	1),
		Layout.newline,
	Layout.tab,
		("Geometry",	(3, lambda l: l.fileUnitList("GEOBEGIN",False))),
		("Out",		(4, lambda l: l.fileUnitList("GEOBEGIN",False))),
		("Fmt",		(0, ["", "COMBNAME", "COMBINAT", "FLUGG"]))]

__GEOBEGIN_title = [Layout.newline, ("Title",	 -1) ]

def _GEOBEGIN_layout4(card):
	layout = __GEOBEGIN_layout4[:]
	if card.intWhat(3) not in (0,5) or card.intWhat(4) not in (0,11):
		if card.intWhat(3) not in (0,5):
			layout.append(Layout.newline)
			layout.append(Layout.tab)
			layout.append(("File",
				(__GEOBEGIN_geofile, (("Geometry file","*.geo"),("All","*"))),
				"External file for saving the geometry"))
		else:
			layout.append(Layout.tab)
		layout.append(Layout.tab)

		if card.intWhat(4) not in (0,11):
			layout.append(Layout.newline)
			layout.append(Layout.tab)
			layout.append(("Output",
				(__GEOBEGIN_geoout,  (("Geometry file","*.out"),("All","*"))),
				"Write geometry output to external file"))
		else:
			layout.append(Layout.tab)
		layout.append(Layout.tab)
	layout.extend(__GEOBEGIN_title)
	return layout

#------------------------------- GEOEND ----------------------------------------
__GEOEND_type = ["", "DEBUG"]

__GEOEND_layout_normal = [ Layout.tab,
		("",		(0, __GEOEND_type)),
		Layout.tab,
		Layout.tab ]

__GEOEND_layout_debug = [ Layout.tab,
		("",		(0, __GEOEND_type)),
		Layout.tab,
		("Max. Errors",	10),
		Layout.newline,
	Layout.tab,
		("Xmin",	 4),
		("Xmax",	 1),
		("NX",		 7),
		Layout.newline,
	Layout.tab,
		("Ymin",	 5),
		("Ymax",	 2),
		("NY",		 8),
		Layout.newline,
	Layout.tab,
		("Zmin",	 6),
		("Zmax",	 3),
		("NZ",		 9) ]

def _GEOEND_layout4(card):
	# depending on the value of what(1) there are various layouts
	if card.sdum() == "DEBUG":
		return __GEOEND_layout_debug
	else:
		return __GEOEND_layout_normal

#------------------------------- _GLOBAL ---------------------------------------
__GLOBAL_analogue = [ Layout.empty_what,	("Input", 0),
		("Analogue", -1),	("Biased", 2) ]
__GLOBAL_dnear    = [ Layout.empty_what,	Layout.default_what,
		("Always use",          1),
		("Don't use on steps", -1),
		("Do not use at all",  -2) ]
__GLOBAL_input    = [ Layout.empty_what,	Layout.default_what,
		("Names", 1),	("Free", 2),
		("Names+Free", 3),	("Format", 4),	Layout.reset_what]
__GLOBAL_geometry = [ Layout.empty_what,	Layout.default_what,
		("Free", 1),	("Fixed", -1) ]

_GLOBAL_layout4 = [ Layout.tab,
		("Max #reg",	 1),
		("Analogue",	(2, __GLOBAL_analogue)),
		("DNear",	(3, __GLOBAL_dnear)),
		Layout.newline,
	("Input",	(4, __GLOBAL_input)),
		("Geometry",	(5, __GLOBAL_geometry)),
		("Memory",	6),
		("Flag",       (0, ["","DEPRBODY","OLDFLAIR","OLDFLBDY"])) ]

#------------------------------- GCR-SPE ---------------------------------------
__GCR_sdum = ["Spectra", "DIPCOORD", "NO-NORM", "NORMALIZ"]
__GCR_field = [ ("naive dipole",	0),
		("exact multipole",	1),
		("MAGFLD routine",	2),
		("dipole @ infinite",	3),
		("undefined",		4) ]

def __GCR_SPE_sdum(layout, card, val=None):
	if val is None:
		if card.sdum() in ("DIPCOORD","NO-NORM","NORMALIZ"):
			return card.sdum()
		else:
			return "Spectra"
	else:
		if val in ("DIPCOORD","NO-NORM","NORMALIZ"):
			layout.setWhat(card, 0, val)
		else:
			layout.setWhat(card, 0, "")

def __GCR_what1_i0(layout, card, val=None):
	w1 = card.intWhat(1)
	i1, i0 = divmod(w1, 1000)
	if val is None:
		return i0
	else:
		try: ival = int(val)
		except: ival = 0
		layout.setWhat(card, 1, i1*1000 + ival)

def __GCR_what1_i1(layout, card, val=None):
	w1 = card.intWhat(1)
	i1, i0 = divmod(w1, 1000)
	if val is None:
		return i1
	else:
		try: ival = int(val)
		except: ival = 0
		layout.setWhat(card, 1, ival*1000 + i0)

def __GCR_what3_j0(layout, card, val=None):
	w3 = card.intWhat(3)
	j1, j0 = divmod(w3, 1000)
	if val is None:
		return j0
	else:
		try: jval = int(val)
		except: jval = 0
		layout.setWhat(card, 3, j1*1000 + jval)

def __GCR_what6_k0(layout, card, val=None):
	w6 = card.intWhat(6)
	k1, k0 = divmod(w6, 10000)
	if val is None:
		return k0
	else:
		try: kval = int(val)
		except: kval = 0
		layout.setWhat(card, 6, k1*1000 + kval)

def __GCR_what6_k1(layout, card, val=None):
	w6 = card.intWhat(6)
	k1, k0 = divmod(w6, 10000)
	if val is None:
		return k1
	else:
		try: kval = int(val)
		except: kval = 0
		layout.setWhat(card, 6, kval*1000 + k0)

__GCR_common = [Layout.tab, ("Type", (__GCR_SPE_sdum,__GCR_sdum))]

__GCR_SPE_layout4_dipcoord = __GCR_common[:] + \
		[Layout.newline,
	Layout.tab,
		("Rdip",   1),
		("θdip",   2),
		("φdip",   3),
		Layout.newline,
	Layout.tab,
		("Date",   6),
		("θnorth", 4),
		("φnorth", 5)]

__GCR_SPE_layout4_default = __GCR_common[:] + \
		[Layout.newline,
	("Field",	(__GCR_what1_i0, __GCR_field), "Magnetic field type"),
		("Shells",	__GCR_what1_i1, "Number of atmospheric shells"),
		("Radius",	4),
		Layout.newline,
	("Equatorial Field",	5),
		("Dump shell",	__GCR_what6_k0),
		("Unit",	__GCR_what6_k1),
		("DateFile",	0)]

def _GCR_SPE_layout4(card):
	sdum = card.sdum()
	if sdum == "DIPCOORD":
		return __GCR_SPE_layout4_dipcoord
	elif sdum in ("NO-NORM", "NORMALIZ"):
		return __GCR_common
	else:
		return __GCR_SPE_layout4_default

#------------------------------- HI-PROPE --------------------------------------
_HI_PROPE_layout4 = [ Layout.tab,
		("Z",		1),
		("A",		2),
		("Isom",	3) ]

#------------------------------- LAM-BIAS --------------------------------------
__LAM_BIAS_sdum = ["",
		"DCDRBIAS",
		"DCY-DIRE",
		"DECALL",
		"DECPRI",
		"GDECAY",
		"INEALL",
		"INEPRI",
		"N1HSCBS" ]

__LAM_BIAS_layout4_dcydire = [ Layout.tab,
		("Type",	(0, __LAM_BIAS_sdum)),
		("Distribution", (5,
				[Layout.empty_what,
				 ("Point-like  exp(1-cosθ)", 0),
				 ("Cylindrical exp(1-cosθ)", 1),
				 ("Point-like  1/(1+(1-cosθ)²)", 100),
				 ("Cylindrical 1/(1+(1-cosθ)²)", 101) ])),
		Layout.newline,
	Layout.tab,
		("λ decay", 4),
		("Background", 6),
		Layout.newline,
	Layout.tab,
		("cosX",	1),
		("cosY",	2),
		("cosZ",	3) ]

__LAM_BIAS_layout4_dcdrbias = [ Layout.tab,
		("Type",	(0, __LAM_BIAS_sdum)),
		("Activate",	(1, Layout.onoff3_what)),
		Layout.newline,
	Layout.tab,
		("Part",	(4, FlukaParticle.list)),
		("to Part",	(5, FlukaParticle.list)),
		("Step",	6) ]

__LAM_BIAS_layout4_gdecay = [ Layout.tab,
		("Type",	(0, __LAM_BIAS_sdum)),
		("<λ>",		1),
		("× λ inelastic",2),
		Layout.newline,
	("Mat",			(3, FlukaLayout.materialList)),
		("Part",	(4, FlukaParticle.list)),
		("to Part",	(5, FlukaParticle.list)),
		("Step",	6) ]

__LAM_BIAS_layout4_n1hscbs = [ Layout.tab,
		("Type",	(0, __LAM_BIAS_sdum)),
		("Activate",	(1, Layout.onoff3_what)) ]

__LAM_BIAS_layout4_other = [ Layout.tab,
		("Type",	(0, __LAM_BIAS_sdum)),
		("× mean life",	1),
		("× λ inelastic",	2),
		Layout.newline,
	("Mat",			(3, FlukaLayout.materialList)),
		("Part",	(4, FlukaParticle.list)),
		("to Part",	(5, FlukaParticle.list)),
		("Step",	6) ]

def _LAM_BIAS_layout4(card):
	sdum = card.sdum()
	if   sdum == "DCY-DIRE":
		return __LAM_BIAS_layout4_dcydire
	elif sdum == "DCDRBIAS":
		return __LAM_BIAS_layout4_dcdrbias
	elif sdum == "N1HSCBS":
		return __LAM_BIAS_layout4_n1hscbs
	elif sdum == "GDECAY":
		return __LAM_BIAS_layout4_gdecay
	else:
		return __LAM_BIAS_layout4_other

#------------------------------- LATTICE ---------------------------------------
_LATTICE_layout4 = [ #Layout.tab,
	( "Trans",		(0, Layout.rotationList)),
		("Reg",		(1, FlukaLayout.regionList)),
		("to Reg",	(2, FlukaLayout.regionList)),
		("Step",	 3),
		Layout.newline,
	Layout.tab,
		("Lat",		 4),
		("to Lat",	 5),
		("Step",	 6) ]

#------------------------------- LATTSNGL --------------------------------------
_LATTSNGL_layout4 = [ (None,     4),
		("Reg",		(1, FlukaLayout.regionList)),
		("R1",		(2, Layout.rotationList)),
		("R2",		(3, Layout.rotationList))]

#------------------------------- LOW-BIAS --------------------------------------
_LOW_BIAS_layout4 = [ Layout.tab,
		("Ecut",	(1, FlukaLayout.neutronGroupList)),
		("NonAnalogue",	(2, FlukaLayout.neutronGroupList)),
		("Survival",	 3),
		Layout.newline,
	Layout.tab ] + __reg456_range

#------------------------------- LOW-DOWN --------------------------------------
_LOW_DOWN_layout4 = [ Layout.tab,
		("Limit",	(1, FlukaLayout.neutronGroupList)),
		("Bias",	 2),
		Layout.newline,
	Layout.tab ] + __reg456_range

#------------------------------- LOW-MAT ---------------------------------------
def __lowmatFunc(layout, card, val=None):
	groups = layout.neutronGroupsUsed()
	if val is None:
		# Find value
		w2 = card.intWhat(2)
		w3 = card.intWhat(3)
		w4 = card.intWhat(4)
		s  = card.sdum()
		i = 1
		for g in groups:
			for mat in Fluka.lowMaterials[g]:
				if w2 == mat.id1 and \
				   w3 == mat.id2 and \
				   w4 == mat.id3 and \
				   s  == mat.name:
					return i
				i += 1
		return 0
	else:
		# Set value
		try:
			val = int(val)
		except:
			return
		if val==0:
			undoinfo = [
				layout.flair.setWhatUndo(card, 2, ""),
				layout.flair.setWhatUndo(card, 3, ""),
				layout.flair.setWhatUndo(card, 4, ""),
				layout.flair.setWhatUndo(card, 0, "") ]
		else:
			val -= 1
			group = -1
			for g in groups:
				l = len(Fluka.lowMaterials[g])
				if val < l:
					group = g
					break
			mat = Fluka.lowMaterials[group][val]
			undoinfo = [
				layout.flair.setWhatUndo(card, 2, mat.id1),
				layout.flair.setWhatUndo(card, 3, mat.id2),
				layout.flair.setWhatUndo(card, 4, mat.id3),
				layout.flair.setWhatUndo(card, 0, mat.name) ]
		layout.flair.addUndo(undoinfo)

_LOW_MAT_layout4 = [ Layout.tab,
		("Mat",		(1, FlukaLayout.materialList)),
		("LowMat",	(__lowmatFunc, FlukaLayout.lowmatList),
			"Combined identifier WHAT(1,2,3) of the corresponding "
			"low-energy neutron material") ]

#------------------------------- LOW-NEUT --------------------------------------
__LOW_NEUT_ng   = [ ("260", 260), ("72", 72) ]
__LOW_NEUT_gg   = [ ("42",   42), ("22", 22), ]
__LOW_NEUT_maxe = [ "0.02", "0.0196" ]

__LOW_NEUT_print = [ Layout.empty_what,
		("Minimum",      0),
		("Integral XS",  1),
		("+Downscatter", 2),
		("+Angles",      3),
		("+Residual",    4)]

__LOW_NEUT_pt    = [ Layout.empty_what,
		Layout.default_what,
		("Basic",        1),
		(">3eV S(a,b)",  2),
		(">3eV all",     3),
		("Everywhere",   4),
		("Disable",     -1)
		]
__LOW_NEUT_fiss = [ Layout.empty_what,
		Layout.default_what,
		("Forced 1",  1),
		("Normal",   -1)
		]

def __LOW_NEUT_what6_pt(layout, card, val=None):
	w = card.intWhat(6)
	i1,i0 = divmod(abs(w),10)
	if w<0:
		i0 = -i0
		i1 = -i1
	if val is None:
		return i0
	else:
		try:
			i0 = int(val)
			if i0>=0:
				i1 = max(i1, 0)
				nw = i1*10 + i0
			else:
				nw = i0
			if nw != w:
				layout.setWhat(card, 6, nw)
		except ValueError:
			pass

def __LOW_NEUT_what6_fiss(layout, card, val=None):
	w = card.intWhat(6)
	i1,i0 = divmod(abs(w),10)
	if w<0:
		i0 = -i0
		i1 = -i1
	if val is None:
		return i1
	else:
		try:
			i1 = int(val)
			if i1>=0:
				i0 = max(i0, 0)
				nw = i1*10 + i0
			else:
				nw = i1*10
			if nw != w:
				layout.setWhat(card, 6, nw)
		except ValueError:
			pass

_LOW_NEUT_layout4 = [ Layout.tab,
		("n-groups",	(1, __LOW_NEUT_ng)),
		("γ-groups",	(2, __LOW_NEUT_gg)),
		("Emax",	(3, __LOW_NEUT_maxe)),
		Layout.newline,
	Layout.tab,
		("Print",	(4, __LOW_NEUT_print)),
		("PtWise",	(__LOW_NEUT_what6_pt,   __LOW_NEUT_pt),
				"Point wise treatment"),
		("(n,f)",	(__LOW_NEUT_what6_fiss, __LOW_NEUT_fiss),
					"Neutron induced fission multiplicity")]

#------------------------------- LOW-PWXS --------------------------------------
def __LOW_PWXS_databases(layout):
	# Cache databases on first call
	global _pwxs_databases
	if _pwxs_databases is not None:
		return _pwxs_databases

	path = os.path.join(FlukaConfig.dataDir(),"neutron")
	try:
		_pwxs_databases = [x for x in os.listdir(path)
				if os.path.isdir(os.path.join(path,x))]
	except FileNotFoundError:
		_pwxs_databases = ["default"]
	_pwxs_databases.sort()
	_pwxs_databases.insert(0,"")
	return _pwxs_databases

def __LOW_PWXS_Sab(layout):
	global _Sab_materials
	if _Sab_materials is not None:
		return _Sab_materials
	sab = set()
	for p,dirs,files in os.walk(os.path.join(FlukaConfig.dataDir(),"neutron"), followlinks=True):
		if "ThermalScattering" in p:
			for f in files:
				if f.endswith(".z"):
					f = f[:-2]
				sab.add(f)
	_Sab_materials = list(sab)
	_Sab_materials.sort()
	_Sab_materials.insert(0,"")
	_Sab_materials.append("unbind")
	return _Sab_materials

_LOW_PWXS_layout4 = [ Layout.tab ] + __mat456_range + [
	Layout.newline,
		("db",    (0,__LOW_PWXS_databases)),
		("IAZ",    1),
		("S(α,β)",(2,__LOW_PWXS_Sab)),
		("T",      3) ]

#------------------------------- MATERIAL --------------------------------------
_MATERIAL_layout4 = [
		Layout.name,
		Layout.tab,
		("#",		4),
		("ρ",		3),
		Layout.newline,
	("Z",			1),
		("Am",		2),
		("A",		6),
		("dE/dx",	(5, FlukaLayout.materialList)) ]

#------------------------------- MAT-PROP --------------------------------------
if Global.developer:
	__MAT_PROP_sdum = [ "", "DPA-ENER", "LOWNTEMP", "USERDIRE" ]
else:
	__MAT_PROP_sdum = [ "", "DPA-ENER", "USERDIRE" ]
__MAT_PROP_call = [ Layout.empty_what, ("USRMED", 1), Layout.reset_what ]

__MAT_PROP_layout4_default = [ Layout.tab,
		("Type",	(0, __MAT_PROP_sdum)),
		("Gas pressure", 1),
		("RHOR",	2),
		Layout.newline,
	("Ionization",	3)] + __mat456_range

__MAT_PROP_layout4_dpaener = [ Layout.tab,
		("Type",	(0, __MAT_PROP_sdum)),
		("DPA Eth",  1),
		Layout.newline ,
		Layout.tab     ,
		("b_arcdpa", 2),
		("c_arcdpa", 3),
		Layout.newline ,
		Layout.tab ] + __mat456_range

__MAT_PROP_layout4_lowntemp = [ Layout.tab,
		("Type",	(0, __MAT_PROP_sdum)),
		("T_act/T_xs",	1),
		("Thermal groups",2),
		Layout.newline,
		("Tnominal",	3) ] + __mat456_range

__MAT_PROP_layout4_userdire = [ Layout.tab,
		("Type",	(0, __MAT_PROP_sdum)),
		("Call",	(1, __MAT_PROP_call)),
		Layout.newline,
		Layout.tab ] + __mat456_range

def _MAT_PROP_layout4(card):
	sdum = card.sdum()[:8]
	if sdum=="DPA-ENER":
		return __MAT_PROP_layout4_dpaener
	elif sdum=="LOWNTEMP":
		return __MAT_PROP_layout4_lowntemp
	elif sdum=="USERDIRE":
		return __MAT_PROP_layout4_userdire
	else:
		return __MAT_PROP_layout4_default

#------------------------------- MCSTHRESh -------------------------------------
_MCSTHRES_layout4 = [Layout.tab,
		("Primary",   1),
		("Secondary", 2) ]

#------------------------------- MGNFIELD --------------------------------------
__MGNFIELD_layout4_default = \
	[ ("",	(0, FlukaLayout.magneticList)),
		("Max Ang (deg)",	1),
		("Bound Acc. (cm)",	2),
		("Min step (cm)",	3),
		Layout.newline,
	Layout.tab,
		("Bx",		4),
		("By",		5),
		("Bz",		6)]

__MGNFIELD_layout4_assign = \
	[ ("",	(0, FlukaLayout.magneticList)),
		("Strength",	1, "Magnetic field strength or gradient"),
		("Rotdefi",	(2, FlukaLayout.rotationList)),
		("On",		(3, [("Region",0),("Prototype",1),("Replica",2)])),
		Layout.newline,
	Layout.tab,
		("Reg"	,	(4, FlukaLayout.regionList)),
		("to Reg",	(5, FlukaLayout.regionList)),
		("Step",	 6) ]

def _MGNFIELD_layout4(card):
	if card.sdum()=="":
		return __MGNFIELD_layout4_default
	else:
		return __MGNFIELD_layout4_assign

#------------------------------- MULSOPT ---------------------------------------
__MULSOPT_sdum = [ "", "FANO-ON", "FANO-OFF", "MLSH-ON", "MLSH-OFF",
		"GLOBAL", "GLOBEMF", "GLOBHAD" ]
__MULSOPT_strech = ["",  "1",  "2",  "3",  "4",
		        "11", "12", "13", "14",
		       "101","102","103","104",
		       "111","112","113","114"]

__MULSOPT_what1 = [ Layout.empty_what, Layout.on_what, Layout.off_what, Layout.reset_what ]

def __MULSOPT_opt(layout, card, val=None):
	w1 = card.intWhat(1)
	if w1<0:
		i0   = -1
		i2i1 =  0
	else:
		i2i1, i0 = divmod(w1, 10)
	if val is None:
		if card.what(1)=="":
			return ""
		else:
			return i0
	else:
		if val=="":
			layout.setWhat(card, 1, "")
		else:
			try:
				i0 = int(val)
				if i0<0:
					w1 = -1
				else:
					w1 = i0 + i2i1*10
				layout.setWhat(card, 1, w1)
			except ValueError:
				pass

def __MULSOPT_steps_i1(layout, card, val=None):
	w1 = card.intWhat(1)
	w1, i0 = divmod(w1, 10)
	i2, i1 = divmod(w1, 10000)
	if val is None:
		return i1
	else:
		try:
			layout.setWhat(card, 1, i0 + int(val)*10 + i2*100000)
		except ValueError:
			pass

def __MULSOPT_steps_i2(layout, card, val=None):
	i2,i1i0 = divmod(card.intWhat(1), 100000)
	if val is None:
		return i2
	else:
		try:
			layout.setWhat(card, 1, i1i0 + int(val)*100000)
		except ValueError:
			pass

__MULSOPT_what2 = [ ("No corrections",	0),
		   ("1st Born",		1),
		   ("2nd Born",		2),
		   ("1st+Finite size", -1),
		   ("2nd+Finite size", -2),
		   ("no MCS",		3),
		   ("Form factors",    -3) ]

__MULSOPT_what3 = [ ("No corrections",	0),
		   ("1st Born",		1),
		   ("2nd Born",		2),
		   ("1st+Finite size", -1),
		   ("2nd+Finite size", -2),
		   ("no MCS",		3) ]

__MULSOPT_fano_off_layout4 = [ Layout.tab,
		("Type",      (0, __MULSOPT_sdum)),
		("Optimize",  (__MULSOPT_opt, __MULSOPT_what1),
			"Optimization switch for single scattering"),
		Layout.newline,
	Layout.tab,
		("h/μ Corr",  (2, __MULSOPT_what2)),
		("e-e+ Corr", (3, __MULSOPT_what3)),
		Layout.newline,
	Layout.tab ] + __mat456_range

__MULSOPT_fano_on_layout4 = [ Layout.tab,
		("Type",      (0, __MULSOPT_sdum)),
		("Optimize",  (__MULSOPT_opt, __MULSOPT_what1),
			"WHAT(1) I0: Optimization switch for single scattering"),
		Layout.newline,
	("h/μ steps",	__MULSOPT_steps_i1,
				"WHAT(1) I1: Number of single scattering steps for hadrons and muons"),
		("e-e+ steps", __MULSOPT_steps_i2,
				"WHAT(1) I2: Number of single scattering steps for electrons and positrons"),
		("h/μ Corr",    (2, __MULSOPT_what2)),
		("e-e+ Corr",   (3, __MULSOPT_what3)),
		Layout.newline,
	Layout.tab ] + __mat456_range

__MULSOPT_global_layout4 = [ Layout.tab,
		("Type",      (0, __MULSOPT_sdum)),
		("Min step",   1),
		("Stretching", (2, __MULSOPT_strech)),
		Layout.newline,
	("Optimal",    3),
		("Single scat",	(4, Layout.onoff3_what)),
		("E<Moliere",	(5, Layout.onoff3_what)),
		("# scatterings", 6)]

def _MULSOPT_layout4(card):
	if card.sdum() in __MULSOPT_sdum[5:]:
		return __MULSOPT_global_layout4
	else:
		if card.what(1)=="" or card.intWhat(1)<0:
			return __MULSOPT_fano_off_layout4
		else:
			return __MULSOPT_fano_on_layout4

#------------------------------- MUPHOTON --------------------------------------
__MUPHOTON_inter = [	Layout.empty_what,
			Layout.ignore_what,
			("Full", 1),
			("No secondaries", 2),
			Layout.reset_what ]
_MUPHOTON_layout4 = [ Layout.tab,
		("μ Inter",     (1, __MUPHOTON_inter)),
		("σ long/trans", 2),
		("ρ inter",      3),
		Layout.newline,
	Layout.tab ] + __mat456_range

#------------------------------- OPEN ------------------------------------------
__OPEN_status  = [ "", "OLD", "NEW", "UNKNOWN", "READONLY", "SCRATCH" ]
__OPEN_layout4 = [ Layout.tab,
		("Unit",	(1, lambda l: l.fileUnitList(None,True))),
		("Status",	(0, __OPEN_status)),
		Layout.newline,
	Layout.tab,
		("File",	(-1,("All","*")))]
__OPEN_scratch_layout4 = [ Layout.tab,
		("Unit",	(1, lambda l: l.fileUnitList(None,True))),
		("Status",	(0, __OPEN_status))]

def _OPEN_layout4(card):
	if card.sdum() == "SCRATCH":
		return __OPEN_scratch_layout4
	else:
		return __OPEN_layout4

#------------------------------- OPT-PROD --------------------------------------
__OPT_PROD_sdum = [ "", "CERE-OFF", "TRD-OFF", "SCIN-OFF",
			"CERENKOV", "CEREN-WV", "CEREN-OM", "TR-RADIA",
			"SCINTILL", "SCINT-WV", "SCINT-OM"]

__OPT_PROD_off_layout4 = [ Layout.tab,
		("Type",	(0, __OPT_PROD_sdum)),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROD_cerenkov_layout4 = [ Layout.tab,
		("Type",	(0, __OPT_PROD_sdum)),
		("Emin",	1),
		("Emax",	2),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROD_cerenwv_layout4 = [ Layout.tab,
		("Type",	(0, __OPT_PROD_sdum)),
		("λmin",	1),
		("λmax",	2),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROD_cerenom_layout4 = [ Layout.tab,
		("Type",	(0, __OPT_PROD_sdum)),
		("Ωmin",	1),
		("Ωmax",	2),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROD_scintill_layout4 = [ Layout.tab,
		("Type",	(0, __OPT_PROD_sdum)),
		("Ei",		1),
		("fraction",	2),
		Layout.newline,
	("Time",	3)] + __mat456_range

__OPT_PROD_scintwv_layout4 = [ Layout.tab,
		("Type",	(0, __OPT_PROD_sdum)),
		("λi",		1),
		("fraction",	2),
		Layout.newline,
	("Time",	3)] + __mat456_range

__OPT_PROD_scintom_layout4 = [ Layout.tab,
		("Type",	(0, __OPT_PROD_sdum)),
		("Ωi",		1),
		("fraction",	2),
		Layout.newline,
	("Time",	3)] + __mat456_range

def _OPT_PROD_layout4(card):
	s = card.sdum()
	if   s in ("CERENKOV", "TR-RADIA"):
		return __OPT_PROD_cerenkov_layout4
	elif s == "CEREN-WV":
		return __OPT_PROD_cerenwv_layout4
	elif s == "CEREN-OM":
		return __OPT_PROD_cerenom_layout4
	elif s == "SCINTILL":
		return __OPT_PROD_scintill_layout4
	elif s == "SCINT-WV":
		return __OPT_PROD_scintwv_layout4
	elif s == "SCINT-OM":
		return __OPT_PROD_scintom_layout4
	else:
		return __OPT_PROD_off_layout4

#------------------------------- OPT-PROP --------------------------------------
__OPT_PROP_sdum = [	"",
			"METAL",
			"OM-LIMIT",
			"OM-SENSI",
			"RESET",
			"SENSITIV",
			"SPEC-BDX",
			"WV-LIMIT",
			"WV-SENSI",
			"&1",
			"&2",
			"&3"
			]

__OPT_PROP_reset_layout4 = [ Layout.tab,
		("Type", (0, __OPT_PROP_sdum)),
		Layout.newline,
		Layout.tab] + __mat456_range

__OPT_PROP_empty_layout4 = [ Layout.tab,
		("Type", (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("Refraction",	1),
		("Absorption",	2),
		("Diffusion",	3),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROP_wvlimit_layout4 = [ Layout.tab,
		("Type", (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("λmin",	1),
		("λ",		2),
		("λmax",	3),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROP_omlimit_layout4 = [ Layout.tab,
		("Type", (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("Ωmin",	1),
		("Ω",		2),
		("Ωmax",	3),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROP_metal_layout4 = [ Layout.tab,
		("Type", (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("1st",		1),
		("2nd",		2),
		("3rd 1-r",	3),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROP_amp1_layout4 = [ Layout.tab,
		("Type",	 (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("4th",		1),
		("5th",		2),
		("6th",		3),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROP_amp2_layout4 = [ Layout.tab,
		("Type",	 (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("7th",		1),
		("8th",		2),
		("9th",		3),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROP_amp3_layout4 = [ Layout.tab,
		("Type", (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("10th",	1),
		("11th",	2),
		("12th",	3),
		Layout.newline,
	Layout.tab ] + __mat456_range

__OPT_PROP_sensitiv_layout4 = [ Layout.tab,
		("Type",	 (0, __OPT_PROP_sdum)),
		("0th",		1),
		("1st",		2),
		Layout.newline,
	Layout.tab,
		("2nd",		3),
		("3rd",		4),
		("Max sensitivity", 5)]

__OPT_PROP_wvsensi_layout4 = [ Layout.tab,
		("Type",	 (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("λmin",	1),
		("λ",		2),
		("λmax",	3)]

__OPT_PROP_omsensi_layout4 = [ Layout.tab,
		("Type",	 (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("Ωmin",	1),
		("Ω",		2),
		("Ωmax",	3) ]

__OPT_PROP_specbdx_layout4 = [ Layout.tab,
		("Type", (0, __OPT_PROP_sdum)),
		Layout.newline,
	Layout.tab,
		("Activate",	(1, Layout.onoff3_what)),
		("Reg",		(2, FlukaLayout.regionList)),
		("to Reg",	(3, FlukaLayout.regionList)),
		Layout.newline,
	Layout.tab,
		("Activate",	(4, Layout.onoff3_what)),
		("Reg",		(5, FlukaLayout.regionList)),
		("to Reg",	(6, FlukaLayout.regionList)) ]

def _OPT_PROP_layout4(card):
	s = card.sdum()
	if   s == "RESET":
		return __OPT_PROP_reset_layout4
	elif s == "WV-LIMIT":
		return __OPT_PROP_wvlimit_layout4
	elif s == "OM-LIMIT":
		return __OPT_PROP_omlimit_layout4
	elif s == "METAL":
		return __OPT_PROP_metal_layout4
	elif s == "&1":
		return __OPT_PROP_amp1_layout4
	elif s == "&2":
		return __OPT_PROP_amp2_layout4
	elif s == "&3":
		return __OPT_PROP_amp3_layout4
	elif s == "SENSITIV":
		return __OPT_PROP_sensitiv_layout4
	elif s == "WV-SENSI":
		return __OPT_PROP_wvsensi_layout4
	elif s == "OM-SENSI":
		return __OPT_PROP_omsensi_layout4
	elif s == "SPEC-BDX":
		return __OPT_PROP_specbdx_layout4
	else:
		return __OPT_PROP_empty_layout4

#------------------------------- PAIRBREM --------------------------------------
__PAIRBREM_activate = [
		Layout.empty_what,
		Layout.ignore_what,
		("Pair Prod",		 1),
		("Bremsstrahlung",	 2),
		("Both",		 3),
		("No Pair",		-1),
		("No Bremsstrahlung",	-2),
		("Inhibit both",	-3) ]

_PAIRBREM_layout4 = [ Layout.tab,
		("Act",		(1, __PAIRBREM_activate)),
		("e-e+ Thr",	2),
		("γ Thr",	3),
		Layout.newline,
	Layout.tab ] + __mat456_range

#------------------------------- SYRASTEP --------------------------------------
_SYRASTEP_layout4 = [ Layout.tab,
		("γ Thr",   1),
		("reverse flag",	3),
		Layout.newline,
	Layout.tab ] + __reg456_range

#------------------------------- PART-THR --------------------------------------
__PART_THR_types = [
		("Momentum",         0),
		("Energy",           1),
		("γ gamma (E/mc^2)", 2),
		("η eta (β*γ)",      3),
		("Function",  Layout.FUNCTION)]

def _PART_THR_layout4(card):
	layout = [ Layout.tab,
		("Type",	(FlukaLayout._what1_type, __PART_THR_types),
				"WHAT(1) sign: Energy or Momentum") ]

	if card.intWhat(5)>0:
		labels = ( "γ", "η" )
	else:
		labels = ( "p", "E" )
	if card.sign(1):
		label = labels[1]
	else:
		label = labels[0]

	layout += [ (label,	 FlukaLayout._what1_val),
		Layout.newline,
		Layout.tab,
		("Part",	(2, FlukaParticle.list)),
		("to Part",	(3, FlukaParticle.list)),
		("Step",	 4) ]

	return layout

#------------------------------- PHOTONUC --------------------------------------
__PHOTONUC_sdum = [ "", "ELECTNUC", "MUMUPAIR", "MUMUPRIM" ]

def __PHOTONUC_int_what1(layout, card, val=None, sub=1):
	ia, ih, id, iq, ig = 0, 0, 0, 0, 0
	w1 = card.intWhat(1)
	if   w1 == -1: ia = -1
	elif w1 ==  1: ia = 1
	elif w1 ==  2: ih = 1
	elif w1 ==  3: id = 1
	elif w1 ==  4: iq = 1
	elif w1 ==  5: ig = 1
	else:
		ia = 0
		ig,w1 = divmod(w1, 1000)
		iq,w1 = divmod(w1,  100)
		id,ih = divmod(w1,   10)

	if val is None:
		if   sub==1: return ia
		elif sub==2: return ih
		elif sub==3: return id
		elif sub==4: return iq
		else:        return ig
	else:
		try: val = int(val)
		except: val = 0
		if val == -1:
			layout.setWhat(card, 1, val)
			return

		ia = max(ia, 0)
		if   sub==1:
			layout.setWhat(card, 1, val)
			return
		elif sub==2: ih=val
		elif sub==3: id=val
		elif sub==4: iq=val
		else:        ig=val
		val = ih + id*10 + iq*100 + ig*1000
		if   val ==    1: val=2
		elif val ==   10: val=3
		elif val ==  100: val=4
		elif val == 1000: val=5
		layout.setWhat(card, 1, val)

__PHOTONUC_layout4_default = [ Layout.tab,
		("Type",	(0, __PHOTONUC_sdum)),
		Layout.tab,
		("All E",	(lambda l,c,v=None,s=1:
				__PHOTONUC_int_what1(l,c,v,s), Layout.onoffreset_what),
				"WHAT(1)=1.0: photo-nuclear interactions are activated at all energies"),
		Layout.newline,
		("E>0.7GeV",	(lambda l,c,v=None,s=2:
				__PHOTONUC_int_what1(l,c,v,s), Layout.onoffreset_what),
				"WHAT(1) ih=1: photo-nuclear interactions are "
				"activated only in the high energy range (above 0.7 GeV)"),
		("Δ resonance", (lambda l,c,v=None,s=3:
				__PHOTONUC_int_what1(l,c,v,s), Layout.onoffreset_what),
				"WHAT(1) id=1: photo-nuclear interactions are activated "
				"only in the region of the Delta resonance"),
		("Quasi D",	(lambda l,c,v=None,s=4:
				__PHOTONUC_int_what1(l,c,v,s), Layout.onoffreset_what),
				"WHAT(1) iq=1: only quasi-deuteron photo-nuclear interactions are activated"),
		("Giant Dipole", (lambda l,c,v=None,s=5:
				__PHOTONUC_int_what1(l,c,v,s), Layout.onoffreset_what),
				"WHAT(1) ig=1: only Giant Dipole Resonance photo-nuclear "
				"interactions are activated"),
		Layout.newline,
		Layout.tab] + __mat456_range

def __PHOTONUC_mu_what1(layout, card, val=None, sub=1):
	w1 = card.intWhat(1)
	if w1 == -1:
		ich, iqe, iin, ids = -1, 0, 0, 0
	else:
		ids,w1  = divmod(w1, 1000)
		iin,w1  = divmod(w1,  100)
		iqe,ich = divmod(w1,   10)

	if val is None:
		if   sub==1: return ich
		elif sub==2: return iqe
		elif sub==3: return iin
		else:        return ids
	else:
		val = int(val)
		if val == -1:
			layout.setWhat(card, 1, val)
		else:
			ich = max(ich, 0)
			if   sub==1: ich = val
			elif sub==2: iqe = val
			elif sub==3: iin = val
			else:        ids = val
			layout.setWhat(card, 1, ich + iqe*10 + iin*100 + ids*1000)

__PHOTONUC_layout4_mumupair = [ Layout.tab,
		("Type",	(0, __PHOTONUC_sdum)),
		#("μ+μ- prod", (1, __PHOTONUC_mupair_prod)),
		Layout.newline,
		("Coherent",		(lambda l,c,v=None,s=1:
				__PHOTONUC_mu_what1(l,c,v,s), Layout.onoffreset_what),
				"activate muon pair coherent production"),
		("InC. quasielastic",	(lambda l,c,v=None,s=2:
				__PHOTONUC_mu_what1(l,c,v,s), Layout.onoffreset_what),
				"activate muon pair incoherent quasielastic production"),
		("Inc. Inelastic",	(lambda l,c,v=None,s=3:
				__PHOTONUC_mu_what1(l,c,v,s), Layout.onoffreset_what),
				"activate muon pair incoherent inelastic production"),
		("Deep Inelastic",	(lambda l,c,v=None,s=4:
				__PHOTONUC_mu_what1(l,c,v,s), Layout.onoffreset_what),
				"activate muon pair deep inelastic production"),
		Layout.newline,
		("Bias inter-λ", 2) ] + __mat456_range

def _PHOTONUC_layout4(card):
	sdum = card.sdum()
	if sdum in __PHOTONUC_sdum[2:3]:
		return __PHOTONUC_layout4_mumupair
	else:
		return __PHOTONUC_layout4_default

#------------------------------- PHYSICS ---------------------------------------
__PHYSICS_sdum = ["",
		  "CHARMDEC",
		  "COALESCE",
		  "DECAYS",
		  "DPMTHRES",
		  "EM-DISSO",
		  "EVAPORAT",
		  "IONBRPAI",
		  "IONSPLIT",
		  "LIMITS",
		  "LOWDEU",
		  "NEUTRINO",
		  "PEATHRES",
		  "QMDTHRES" ]
if Global.developer:
	__PHYSICS_sdum.extend([
		  "CAPTPROB",
		  "COAPARAM",
		  "DEDX-SET",
		  "RQMDFLAG",
		  "EXPNUCLV",
		  "FISSION",
		  "FL2005.6",
		  "FL2006.3",
		  "FL2008.3",
		  "FL2011.2",
		  "IASLEVEL",
		  "LEVEL-DE",
		  "PISIGMAS",
		  "PREEQUIL",
		  "PREFTCMP",
		  "RESO-REI",
		  "RESTODPM",
		  "TAUFOR"
		  ])
__PHYSICS_sdum.sort()

__PHYSICS_layout4_generic = [ Layout.tab,
			("Type", (0, __PHYSICS_sdum)),
			("#1",	1),
			("#2",	2),
			Layout.newline,
		("#3",		3),
			("#4",	4),
			("#5",	5),
			("#6",	6)]

__PHYSICS_coalesce_what1 = [Layout.empty_what,
			Layout.on_what,
			("+Deuteron", 1001),
			("off", -1),
			Layout.ignore_what]

__PHYSICS_layout4_captprob = [ Layout.tab,
		("Type",	(0, __PHYSICS_sdum)),
		("Law",	(1,	[Layout.empty_what,
				Layout.ignore_what,
				("Z-Law", 1),
				("Daniel's law", 2)])),
		("MNTHCP",	2)]

__PHYSICS_layout4_charmdec = [ Layout.tab,
		("Type",	(0, __PHYSICS_sdum)),
		("Charmed",	(1, Layout.onoff3_what)),
		Layout.tab ]

if Global.developer:
	__PHYSICS_layout4_coalesce = [ Layout.tab,
			("Type",	(0, __PHYSICS_sdum)),
			("Activate",	(1,Layout.onoff3_what)),
			Layout.newline,
			("Model 1",	2),
			("Model 2",	3),
			("Model 3",	5),
			("Model 4",	6)]
else:
	__PHYSICS_layout4_coalesce = [ Layout.tab,
			("Type",	(0, __PHYSICS_sdum)),
			("Activate",	(1,Layout.onoff3_what))]

def __PHYSICS_decays_what1_flag1(layout, card, val=None):
	w = card.intWhat(1)
	f100,f1 = divmod(w,100)
	if val is None:
		if w<=-1:
			return -1
		else:
			return f1
	else:
		try: val = int(val)
		except: val = 0
		if w<0 or val<0:
			layout.setWhat(card, 1, val)
			return
		layout.setWhat(card, 1, f100*100 + val)

def __PHYSICS_decays_what1_flag100(layout, card, val=None):
	w = card.intWhat(1)
	f100,f1 = divmod(w,100)
	if val is None:
		if w<100:
			return 0
		else:
			return f100
	else:
		try: val = int(val)
		except: val = 0
		f1 = max(f1, 0)
		layout.setWhat(card, 1, val*100 + f1)

__PHYSICS_decays1 = [ Layout.empty_what, Layout.ignore_what,
		("Max Acc. with Polarization", 1),
		("Max Acc. no Polarization", 2),
		("Phase space like", 3),
		Layout.reset_what]

__PHYSICS_decays100 = [ ("", 0),
		("Leptonic only", 1),
		("Hadronic only", 2)]

__PHYSICS_layout4_decays = [ Layout.tab,
		("Type",	(0, __PHYSICS_sdum)),
		("Decays",	(__PHYSICS_decays_what1_flag1, __PHYSICS_decays1),
				"Flag for particle decay"),
		("Allowed",	(__PHYSICS_decays_what1_flag100, __PHYSICS_decays100),
				"Allowed decays"),
		Layout.newline,
	Layout.tab,
		("Part",	(4, FlukaParticle.list)),
		("to Part",	(5, FlukaParticle.list)),
		("Step",	 6) ]

__PHYSICS_layout4_dedxset = [ Layout.tab,
		("Type",	(0, __PHYSICS_sdum)),
		Layout.newline,
	("dE/dx Nuc",		(1, Layout.onoff3_what)),
		("Pred.Mat.",	(2, Layout.onoff3_what)),
		("Barkas",	(3, Layout.onoff3_what)),
		("Bloch",	(4, Layout.onoff3_what))]

__PHYSICS_layout4_dpmthres = [ Layout.tab,
		("Type",		(0, __PHYSICS_sdum)),
		("Emin h",		1),
		("Emin ions DPM",	2),
		Layout.newline,
	("Emin Ions RQMD",		3),
		("Smearing +/-ΔE",	4),
		("Smearing +/-ΔE hA",	5),
		("h-A primaries",	(6, Layout.onoff3_what))]

__PHYSICS_emdisso1 = [ Layout.empty_what, Layout.ignore_what,
		("No EM-Dissociation",	1),
		("Proj&Target EM-Disso",2),
		("Proj EM-Disso",	3),
		("Target EM-Disso",	4),
		Layout.reset_what]

__PHYSICS_emdisso2 = [	Layout.empty_what,
			Layout.ignore_what,
			Layout.on_what,
			("off", 2),
			Layout.reset_what]

if Global.developer:
	__PHYSICS_layout4_emdisso = [ Layout.tab,
			("Type",	(0, __PHYSICS_sdum)),
			("EM Disso",	(1, __PHYSICS_emdisso1)),
			Layout.newline,
		Layout.tab,
			("Muon+/-",	(2, __PHYSICS_emdisso2)),
			("Deuteron",	(3, __PHYSICS_emdisso2)),
			("Pre-Tabulation",(4, __PHYSICS_emdisso2))]
else:
	__PHYSICS_layout4_emdisso = [ Layout.tab,
			("Type",	(0, __PHYSICS_sdum)),
			("EM Disso",	(1, __PHYSICS_emdisso1)),
			Layout.tab ]

__PHYSICS_layout4_peathres = [ Layout.tab,
		("Type",	(0, __PHYSICS_sdum)),
		("Nucleons",	1),
		("Pions",	2),
	Layout.newline,
		("Kaons",	3),
		("Kaonbars",	4),
		("AntiNucleon",5),
		("(Anti)Hyperons",6)]

if Global.developer:
	__PHYSICS_layout4_ionbrpai = [ Layout.tab,
			("Type",	(0, __PHYSICS_sdum)),
			("Direct Pair",	(1, Layout.onoff3_what)),
			("Bremss.",	(2, Layout.onoff3_what)),
			Layout.newline,
		Layout.tab,
			("Form fact",	(3, Layout.onoff3_what))]
else:
	__PHYSICS_layout4_ionbrpai = [ Layout.tab,
			("Type",	(0, __PHYSICS_sdum)),
			("Direct Pair",	(1, Layout.onoff3_what)),
			Layout.newline,
		Layout.tab,
			("Form fact",	(3, Layout.onoff3_what))]

def __PHYSICS_evaporat_what1(card):
	w = card.intWhat(1)
	if w<0:
		return 0, 0, -1
	AZ,i0 = divmod(w, 100)
	Amax, Zmax = divmod(AZ, 100)
	return Amax, Zmax, i0

def __PHYSICS_evaporat_what1_i0(layout, card, val=None):
	Amax, Zmax,i0 = __PHYSICS_evaporat_what1(card)
	if val is None:
		return i0
	else:
		try: val = int(val)
		except: val = 0
		if val<0:
			layout.setWhat(card, 1, -1)
		else:
			layout.setWhat(card, 1, Amax*10000 + Zmax*100 + val)

def __PHYSICS_evaporat_what1_Zmax(layout, card, val=None):
	Amax, Zmax,i0 = __PHYSICS_evaporat_what1(card)
	if val is None:
		return Zmax
	else:
		try: val = int(val)
		except: val = 0
		layout.setWhat(card, 1, Amax*10000 + val*100 + i0)

def __PHYSICS_evaporat_what1_Amax(layout, card, val=None):
	Amax, Zmax,i0 = __PHYSICS_evaporat_what1(card)
	if val is None:
		return Amax
	else:
		try: val = int(val)
		except: val = 0
		layout.setWhat(card, 1, val*10000 + Zmax*100 + i0)

__PHYSICS_evaporat = [ Layout.ignore_what,
			("Old Evaporation",          1),
			("New Evap no heavy frag",   2),
			("New Evap with heavy frag", 3),
			Layout.reset_what ]
if Global.developer:
	__PHYSICS_layout4_evaporat = [ Layout.tab,
			("Type",	(0, __PHYSICS_sdum)),
			("Model",	(__PHYSICS_evaporat_what1_i0, __PHYSICS_evaporat),
					"FLUKA evaporation model"),
			Layout.newline,
		Layout.tab,
			("Zmax",	__PHYSICS_evaporat_what1_Zmax, "Maximum fragment Z"),
			("Amax",	__PHYSICS_evaporat_what1_Amax, "Maximum fragment A"),
			("#2",		2),
			Layout.newline,
		("#3", 3),
			("#4",		4),
			("#5",		5),
			("#6",		6)]
else:
	__PHYSICS_layout4_evaporat = [ Layout.tab,
			("Type",	(0, __PHYSICS_sdum)),
			("Model",	(__PHYSICS_evaporat_what1_i0, __PHYSICS_evaporat),
					"FLUKA evaporation model"),
			Layout.newline,
		Layout.tab,
			("Zmax",	__PHYSICS_evaporat_what1_Zmax, "Maximum fragment Z"),
			("Amax",	__PHYSICS_evaporat_what1_Amax, "Maximum fragment A") ]

__PHYSICS_layout4_ionsplit = [ Layout.tab,
		("Type",	(0, __PHYSICS_sdum)),
		("Ion Split",	(1, Layout.onoff3_what)),
		("Splitting",	(6, [("Threshold",0),
				     ("Probabilistic",1),
				     ("On nonelastic",2),
				     ("Nonelastic",3)])),
		Layout.newline,
	("Emin",		2),
		("Emax",	3),
		("Amin",	4),
		("Amax",	5)]

__PHYSICS_layout4_limits = [ Layout.tab,
		("Type"	,	(0, __PHYSICS_sdum)),
		("Max.Pcms (pp)",	1) ]

__PHYSICS_neutrino_flag = [ ("ignored", 0),
			    ("NC", 1),
			    ("CC", 2),
			    ("NC+CC", 3),
			    ("No interaction", -1) ]

__PHYSICS_neutrino_flag2 = [Layout.empty_what,
			Layout.ignore_what,
			("Performed", 1),
			("Not performed", -1)]

__PHYSICS_layout4_neutrino = [ Layout.tab,
		("Type",		(0, __PHYSICS_sdum)),
		("Quasielastic",	(1,__PHYSICS_neutrino_flag)),
		("Resonant",		(2,__PHYSICS_neutrino_flag)),
		Layout.newline,
	Layout.tab,
		("Deep Inelastic",	(3,__PHYSICS_neutrino_flag)),
		("Charm production",	(4,__PHYSICS_neutrino_flag)),
		("Forced Interactions",	(6,__PHYSICS_neutrino_flag2)) ]

if Global.developer:
	__PHYSICS_layout4_qmdthres = [ Layout.tab,
			("Type",		(0, __PHYSICS_sdum)),
			("Emin(QMD)",		1),
			("Emin(BME)",		2),
			("QMD-RQMD",		3),
			("BME-QMD",		4),
			("Emax(fusion)",	5),
			("BME-RQMD",		6)]
else:
	__PHYSICS_layout4_qmdthres = [ Layout.tab,
			("Type",		(0, __PHYSICS_sdum)),
			Layout.tab,
			("Emin(BME)",		2),
			Layout.newline,
		Layout.tab,
			Layout.tab,
			("Emax(fusion)",	5),
			("BME-RQMD",		6)]

__PHYSICS_layout4_rqmdflag_what1 = [ ("Fast cascade", 1), Layout.ignore_what, ("QMD mode", -1) ]
__PHYSICS_layout4_rqmdflag = [ Layout.tab,
		("Type",	(0, __PHYSICS_sdum)),
		("QMD flag",	(1,__PHYSICS_layout4_rqmdflag_what1)),
		("Preequilibrium step", (2,Layout.onoffreset_what))]

__PHYSICS_layout4_lowdeu = [ Layout.tab,
		("Type",	(0, __PHYSICS_sdum)),
		("Activate",	(1,[("On",0),("Off",-1)]))]

def _PHYSICS_layout4(card):
	sdum = card.sdum()[:8]
	if   sdum == "CAPTPROB":
		return __PHYSICS_layout4_captprob
	elif sdum == "CHARMDEC":
		return __PHYSICS_layout4_charmdec
	elif sdum == "COALESCE":
		return __PHYSICS_layout4_coalesce
	elif sdum == "DECAYS":
		return __PHYSICS_layout4_decays
	elif sdum == "DEDX-SET":
		return __PHYSICS_layout4_dedxset
	elif sdum == "DPMTHRES":
		return __PHYSICS_layout4_dpmthres
	elif sdum == "EM-DISSO":
		return __PHYSICS_layout4_emdisso
	elif sdum == "IONBRPAI":
		return __PHYSICS_layout4_ionbrpai
	elif sdum == "IONSPLIT":
		return __PHYSICS_layout4_ionsplit
	elif sdum == "EVAPORAT":
		return __PHYSICS_layout4_evaporat
	elif sdum == "LIMITS":
		return __PHYSICS_layout4_limits
	elif sdum == "PEATHRES":
		return __PHYSICS_layout4_peathres
	elif sdum == "NEUTRINO":
		return __PHYSICS_layout4_neutrino
	elif sdum == "QMDTHRES":
		return __PHYSICS_layout4_qmdthres
	elif sdum == "RQMDFLAG":
		return __PHYSICS_layout4_rqmdflag
	elif sdum == "LOWDEU":
		return __PHYSICS_layout4_lowdeu
	else:
		return __PHYSICS_layout4_generic

#------------------------------- PLOTGEOM --------------------------------------
_PLOTGEOM_layout4 = [ Layout.tab,
		("Axes",	(1, [Layout.empty_what, ("Yes", 0), ("No", 1)])),
		("Boundaries",	(2, [Layout.empty_what, ("Yes", 0), ("No", 1)])),
		("Numbering",	(3, [Layout.empty_what, ("No", 0), ("Yes", 1)])),
		Layout.newline,
	("Worm Length",		4),
		("Verbose",	(5, [Layout.empty_what, ("No", 0), ("Yes", 1)])),
		("Format",	(0, [("Ascii","FORMAT"), ("Binary","")])),
		Layout.newline,
	("Title",	-1),
		Layout.newline,
	Layout.tab,
		("Xmin", 7),
		("Ymin", 8),
		("Zmin", 9),
		Layout.newline,
	Layout.tab,
		("Xmax",10),
		("Ymax",11),
		("Zmax",12),
		Layout.newline,
	Layout.tab,
		("Yx",	13),
		("Yy",	14),
		("Yz",	15),
		Layout.newline,
	Layout.tab,
		("Xx",	16),
		("Xy",	17),
		("Xz",	18),
		Layout.newline,
	("ExpY",	19),
		("ExpX",20),
		("NX",	21),
		("NY",	22) ]

#------------------------------- POLARIZA --------------------------------------
__POLARIZA_flag = [	Layout.empty_what,
			("Not polarized", 0),
			("Orthogonal",    1)
		]
_POLARIZA_layout4 = [ Layout.tab,
		("cosx",	 1),
		("cosy",	 2),
		("cosz",	 3),
		Layout.newline,
	Layout.tab,
		("Relative",	(4, Layout.onoffempty_what)),
		("Fraction",	 5),
		("1-Fraction",	(6,__POLARIZA_flag))]

#------------------------------- RADDECAY --------------------------------------
__RADDECAY_layout4_what1 = [ Layout.empty_what, ("Activation", 1),
		("Semi-Analogue", 2), Layout.reset_what ]
__RADDECAY_layout4_digit = [ Layout.ignore_what, ("Prompt", 1), ("Decay", 2), ("Both",3)]

def __RADDECAY_decay_cut(layout, card, val=None):
	wh,wl = divmod(card.intWhat(5), 100000)
	if val is None:
		return float(wh)
	else:
		try: val = bmath.limit(0, int(float(val)), 99999)
		except ValueError: val = 0
		layout.setWhat(card, 5, val*100000 + wl)

def __RADDECAY_prompt_cut(layout, card, val=None):
	wh,wl = divmod(card.intWhat(5), 100000)
	if val is None:
		return float(wl)
	else:
		try: val = bmath.limit(0, int(float(val)), 99999)
		except ValueError: val = 0
		layout.setWhat(card, 5, wh*100000 + val)

_RADDECAY_layout4 = [ Layout.tab,
		("Decays",	(1, __RADDECAY_layout4_what1)),
		("Patch Isom",	(2, Layout.onoff3_what)),
		("Replicas",	 3),
		Layout.newline,
	# a
	("h/μ Int",	(lambda l,c,v=None:l._what_digit_val(c,4,0,v),
					__RADDECAY_layout4_digit),
			"WHAT(4) a: hadron/muon interaction length or decay biasing, "\
			"as defined by command LAM-BIAS"),
	# b
	("h/μ LPB",	(lambda l,c,v=None:l._what_digit_val(c,4,1,v),
					__RADDECAY_layout4_digit),
			"WHAT(4) b: hadron/muon leading particle  biasing"),
	# c
	("h/μ WW",	(lambda l,c,v=None:l._what_digit_val(c,4,2,v),
					__RADDECAY_layout4_digit),
			"WHAT(4) c: hadron/muon importance and Weight Window biasing, "\
			"as defined by commands BIASING and WW-FACTOr"),
	# d
	("e-e+ Int",	(lambda l,c,v=None:l._what_digit_val(c,4,3,v),
				__RADDECAY_layout4_digit),
			"WHAT(4) d: e+/e-/gamma interaction length biasing, "\
			"as defined by command EMF-BIAS"),
		Layout.newline,
	# e
	("e-e+ LPB",	(lambda l,c,v=None:l._what_digit_val(c,4,4,v),
				__RADDECAY_layout4_digit),
			"WHAT(4) e: e+/e-/gamma leading particle biasing, "\
			"as defined by command EMF-BIAS"),
	# f
	("e-e+ WW",	(lambda l,c,v=None:l._what_digit_val(c,4,5,v),
				__RADDECAY_layout4_digit),
			"WHAT(4) f: e+/e-/gamma importance and Weight Window biasing, "\
			"as defined by commands BIASING and WW-FACTOr"),
	# g
	("Low-n Bias",	(lambda l,c,v=None:l._what_digit_val(c,4,6,v),
				__RADDECAY_layout4_digit),
			"WHAT(4) g: low-energy neutron biased down-scattering, "\
			"as defined by command LOW-DOWN, and non-analogue absorption, "\
			"as defined by LOW-BIAS"),
	# h...
	# i
	("Low-n WW",	(lambda l,c,v=None:l._what_digit_val(c,4,8,v),
				__RADDECAY_layout4_digit),
			"WHAT(4) i: low-energy neutron importance and Weight Window biasing, "\
			"as defined by commands BIASING and WW-FACTOr"),
		Layout.newline,
	Layout.tab,
		("decay cut",	__RADDECAY_decay_cut,
			"WHAT(5) xxxxx*0.1: transport energy cut-off multiplication factor "\
			"for beta+, beta- and gamma decay radiation"),
		("prompt cut",	__RADDECAY_prompt_cut,
			"WHAT(5) yyyyy*0.1: transport energy cut-off multiplication factor "\
			"for prompt e+, e- and gamma radiation"),
		("Coulomb corr",	(6, Layout.onoff3_what))]

#------------------------------- RAD-BIOL --------------------------------------
_RAD_BIOL_layout4 = [ Layout.tab, ("File", 0)]

#------------------------------- RANDOMIZE -------------------------------------
_RANDOMIZ_layout4 = [ Layout.tab,
		("Unit",	(1, lambda l: l.fileUnitList("RANDOMIZ",False))),
		("Seed",	 2) ]

#------------------------------- RESNUCLEi -------------------------------------
_RESNUCLE_layout4 = [ Layout.tab,
		("Type",	(1, FlukaLayout.RESNUCLE_type)),
		("Unit",	(2, lambda l: l.fileUnitList("RESNUCLE"))),
		("Name",	 0),
		Layout.newline,
	("Max Z",		 3),
		("Max M",	 4),
		("Reg",		(5, FlukaLayout.allRegionList)),
		("Vol",		 6) ]

#------------------------------- ROT-DEFI --------------------------------------
__ROT_DEFI_axis = [ ("X", 1), ("Y", 2), ("Z", 3), ("Z", 0) ]
def __ROT_DEFI_id(layout, card, val=None):
	w1 = card.intWhat(1)
	if w1 < 1000:
		(j, i) = divmod(w1, 100)
	else:
		(i, j) = divmod(w1, 1000)
	if val is None:
		return i
	else:
		try: ival = int(val)
		except: ival = 0
		if ival<100 and w1<1000:
			layout.setWhat(card, 1, j*100 + ival)
		else:
			layout.setWhat(card, 1, ival*1000 + j)

def __ROT_DEFI_rot(layout, card, val=None):
	w1 = card.intWhat(1)
	if w1 < 1000:
		(j, i) = divmod(w1, 100)
	else:
		(i, j) = divmod(w1, 1000)
	if val is None:
		return j
	else:
		try: ival = int(val)
		except: ival = 0
		if w1<1000:
			layout.setWhat(card, 1, ival*100 + i)
		else:
			layout.setWhat(card, 1, i*1000 + ival)

_ROT_DEFI_layout4 = [ Layout.tab,
		("Axis",	(__ROT_DEFI_rot, __ROT_DEFI_axis),
				"WHAT(1) j: Transformation axis"),
		("Id",		__ROT_DEFI_id, "WHAT(1) i: Transformation index"),
		("Name",	 0),
		Layout.newline,
	Layout.tab,
		("Polar",	 2),
		("Azm",		 3),
		Layout.newline,
	Layout.tab,
		("Δx",	 4),
		("Δy",	 5),
		("Δz",	 6) ]

#------------------------------- ROTPRBIN --------------------------------------
__ROTPRBIN_type = [ "", "RHO-FACT" ]

_help = "set transformation or density correction factor on binning"

def __ROTPRBIN_type_sdum(layout, card, val=None):
	if val is None:
		if card.sdum()==__ROTPRBIN_type[1]:
			return card.sdum()
		else:
			return ""
	else:
		if val==__ROTPRBIN_type[1] or card.sdum()==__ROTPRBIN_type[1]:
			layout.setWhat(card, 0, val)
		elif val == "":
			pass	# do nothing not to change the rotation

__ROTPRBIN_layout4 = [ Layout.tab,
		("Type",	(__ROTPRBIN_type_sdum, __ROTPRBIN_type), _help),
		("Storage",	 1),
		("# Events",	 3),
		Layout.newline,
	Layout.tab,
		("Rot",		(2, FlukaLayout.rotationList2)),
		("Rot2",	(0, FlukaLayout.rotationList2)),
		Layout.newline,
	Layout.tab,
		("Bin",		(4, FlukaLayout.binningList)),
		("to Bin",	(5, FlukaLayout.binningList)),
		("Step",	 6) ]

__ROTPRBIN_rhofact_layout4 = [ Layout.tab,
		("Type",	(__ROTPRBIN_type_sdum, __ROTPRBIN_type), _help),
		("Density",	(1, [	Layout.empty_what, Layout.default_what,
					("Density correction for all",1),
					("dE/dx correction", 2),
					Layout.reset_what])),
		Layout.newline,
	Layout.tab,
		("Bin",		(4, FlukaLayout.binningList)),
		("to Bin",	(5, FlukaLayout.binningList)),
		("Step",	 6) ]

def _ROTPRBIN_layout4(card):
	if card.sdum() == __ROTPRBIN_type[1]:
		return __ROTPRBIN_rhofact_layout4
	else:
		return __ROTPRBIN_layout4

#------------------------------- SCORE -----------------------------------------
_SCORE_layout4 = [
	Layout.tab,
		("Part1",	(1, FlukaParticle.listAll)),
		("Part2",	(2, FlukaParticle.listAll)),
		Layout.newline,
	Layout.tab,
		("Part3",	(3, FlukaParticle.listAll)),
		("Part4",	(4, FlukaParticle.listAll)) ]

#------------------------------- START -----------------------------------------
_START_layout4 = [ Layout.tab,
		("No.",		 1),
		("Core",	(4,[Layout.empty_what, Layout.on_what, Layout.off_what])),
		Layout.newline,
	Layout.tab,
		("Time",	 6),
		("Report",	(5, [Layout.default_what, ("Every history",1)])) ]

#------------------------------- STEPSIZE --------------------------------------
_STEPSIZE_layout4 = [ Layout.tab,
		("Min (cm)",	 1),
		("Max (cm)",	 2),
		Layout.newline,
	Layout.tab,
		("Reg",		(3, FlukaLayout.regionList)),
		("to Reg",	(4, FlukaLayout.regionList)),
		("Step",	 5) ]

#------------------------------- STERNHEIme ------------------------------------
_STERNHEI_layout4 = [ Layout.tab,
		("Cbar",	 1),
		("X0",		 2),
		("X1",		 3),
		Layout.newline,
	("Mat",			(0, FlukaLayout.materialList)),
		("a",		 4),
		("m",		 5),
		("δ0",		 6) ]

#------------------------------- SPECSOUR --------------------------------------
__SPECSOUR_types = [	"",
			"BEAMSPOT",
			"BIN-SOUR",
			"CROSSASY",
			"CROSSSYM",
			"GCR-ALLF",
			"GCR-IONF",
			"GCR-SPEC",
			"PPSOURCE",
			"SPE-2003",
			"SPE-2005",
			"SPE-SPEC",
			"SYNC-RAD",
			"SYNC-RDN",
			"SYNC-RAS",
			"SYNC-RDS",
		]

__SPECSOUR_line1 = [ Layout.tab,
		("Type",	(0, __SPECSOUR_types)),
		Layout.newline ]

__SPECSOUR_lastlines = [
		Layout.newline,
	Layout.tab,
		("σx",	 7),
		("σy",	 8),
		("σz",	 9),
		Layout.newline,
	("limit-σ",	10),
		("Part",(11, FlukaParticle.hadron)),
		("A",	12),
		("Z",	13),
		Layout.newline,
	("σθC_1",	14),
		("σ0_1",	15),
		("σθC_2",	16),
		("σ0_2",	17),
		Layout.newline,
	Layout.tab,
		("NonElastic", (lambda l,c,v=None:l._what_digit_val(c,18,2,v,3), Layout.onoff_what),
			"WHAT(18) : flag for nuclear nonelastic interactions"),
		("Elastic",	(lambda l,c,v=None:l._what_digit_val(c,18,1,v,3), Layout.onoff_what),
			"WHAT(18) : flag for nuclear elastic interactions"),
		("EM dissociation",(lambda l,c,v=None:l._what_digit_val(c,18,0,v,3), Layout.onoff_what),
			"WHAT(18) : flag for electromagnetic dissociation interactions")]

__SPECSOUR_ppsource_layout4 = __SPECSOUR_line1[:] + \
	[Layout.tab,
		("P1x",		 1),
		("P1y",		 2),
		("P1z",		 3),
		Layout.newline,
	Layout.tab,
		("P2x",		 4),
		("P2y",		 5),
		("P2z",		 6) ] + __SPECSOUR_lastlines

__SPECSOUR_crossasy_layout4 = __SPECSOUR_line1[:] + \
	[ Layout.tab,
		("P1lab",	 1),
		("Polar1",	 2),
		("Azimuthal",	 3),
		Layout.newline,
	Layout.tab,
		("P2-lab",	 4),
		("Polar2",	 5),
		Layout.tab ] + __SPECSOUR_lastlines

__SPECSOUR_crosssym_layout4 = __SPECSOUR_line1[:] + \
	[ Layout.tab,
		("Plab",	 1),
		("CrossAng/2",	 2),
		("Azimuthal",	 3)] + __SPECSOUR_lastlines

__SPECSOUR_other_layout4 = __SPECSOUR_line1[:]
for iii in range(0,18):
	if iii%3 == 0:
		if iii: __SPECSOUR_other_layout4.append(Layout.newline)
		__SPECSOUR_other_layout4.append(Layout.tab)
	__SPECSOUR_other_layout4.append((f"#{iii+1}", iii+1))

__SPECSOUR_type0 = [ ("Radius",0), ("MagField",1), ("Function",Layout.FUNCTION) ]

__SPECSOUR_beamspot_layout4 =  __SPECSOUR_line1[:2] + \
		[("# spots",    1),
		 ("Sampling",	(2,	[("Random", 0),
					 ("Sequentially weighted", 1),
					 ("Sequentially equal", 2)]))]

__SPECSOUR_binsour_layout4 =  __SPECSOUR_line1[:2] + \
		[("Unit",	(1, lambda l: l.fileUnitList("SPECSOUR", False, False))),
		 ("Det Id",	 2),
		Layout.newline,
	 Layout.tab,
		("Rot-before",	(3, FlukaLayout.rotationList2)),
		("Rot-after",	(4, FlukaLayout.rotationList2)) ]

def _SPECSOUR_layout4(card):
	# depending on the value of sdum
	sdum = card.sdum().upper()

	if sdum == "PPSOURCE":
		return __SPECSOUR_ppsource_layout4

	elif sdum == "CROSSASY":
		return __SPECSOUR_crossasy_layout4

	elif sdum == "CROSSSYM":
		return __SPECSOUR_crosssym_layout4

	elif sdum == "BIN-SOUR":
		return __SPECSOUR_binsour_layout4

	elif sdum == "BEAMSPOT":
		return __SPECSOUR_beamspot_layout4

	elif sdum in ("SYNC-RAD", "SYNC-RDN", "SYNC-RAS", "SYNC-RDS"):
		# First line
		layout = [ Layout.tab,
			("Type",	(0, __SPECSOUR_types)),
			("Part",	(FlukaLayout._what1_papro_what, FlukaParticle.particles),
					"Particle type or heavy ion"),
			Layout.newline]

		# Heavy ion selection
		if FlukaLayout._papro_isotope(card, 1):
			layout.extend([Layout.tab,
				("Z",		FlukaLayout._what1_papro_Z, "Heavy ion charge"),
				("A",		FlukaLayout._what1_papro_A, "Heavy ion mass"),
				("Isomer",	FlukaLayout._what1_papro_M, "Heavy ion isomer"),
				Layout.newline])

		# what(2) Energy or momentum
		layout.append(("E/p",	(FlukaLayout._what2_type, __BEAM_type0),
					"WHAT(2) sign: Energy or Momentum"))
		if card.sign(2):
			layout.append(("E",	 FlukaLayout._what2_val,
					"WHAT(2): Energy [GeV]"))
		else:
			layout.append(("p",	 FlukaLayout._what2_val,
					"WHAT(2): Momentum [GeV/c]"))

		# what(3) Radius or Magnetic field
		layout.append(("R/B", (FlukaLayout._what3_type, __SPECSOUR_type0),
					"WHAT(3) sign: Radius or Magnetic Field"))
		if card.sign(3):
			layout.append(("B",	 FlukaLayout._what3_val,
					"WHAT(3): Magnetic Field [T]"))
		else:
			layout.append(("R",	 FlukaLayout._what3_val,
					"WHAT(3): Radius [cm]"))

		# what(4) = minimum photon energy for sampling (GeV)
		# what(5) = x-component of the magnetic field versor
		# what(6) = y-component of the magnetic field versor
		# what(7) = length (cm) of the emission segment (arc)
		layout.extend([
			Layout.newline,
			("Eγmin",	 4),
			("Bx",		 5),
			("By",		 6),
			("Length",	 7),
			Layout.newline,
		Layout.tab,
			("x2",		 8),
			("y2",		 9),
			("z2",		10),
			Layout.newline,
		Layout.tab,
			("cosx2",	11),
			("cosy2",	12) ])
		return layout
	else:
		return __SPECSOUR_other_layout4

#------------------------------- SPOTBEAM --------------------------------------
def __SPOTBEAM_divX_type(layout, card, val=None):
	return __BEAM_div_type(layout, card, val, w=5)

def __SPOTBEAM_divY_type(layout, card, val=None):
	return __BEAM_div_type(layout, card, val, w=6)

def _SPOTBEAM_layout4(card):
	# 1st line
	layout = [ Layout.name,	# spot number
		("Beam", (FlukaLayout._what2_type, __BEAM_type0),
			"WHAT(2) sign: Energy or Momentum") ]

	if card.sign(2):
		layout.append(("E",	 FlukaLayout._what2_val,
				"WHAT(2): Energy [GeV or GeV/n for HI]"))
	else:
		layout.append(("p",	 FlukaLayout._what2_val,
				"WHAT(2): Momentum [GeV/c]"))

	layout.append(("Part",		(FlukaLayout._paprospot_what, FlukaParticle.particles),
			"Particle type or heavy ion"))
	layout.append(Layout.newline)

	# 1bis line
	if FlukaLayout._paprospot_isotope(card):
		layout.extend([ Layout.tab,
				("Z",		FlukaLayout._paprospot_Z, "Heavy ion charge"),
				("A",		FlukaLayout._paprospot_A, "Heavy ion mass"),
				("Isomer",	FlukaLayout._paprospot_M, "Heavy ion isomer"),
				Layout.newline,
				])

	# 2nd line
	layout.append(("Δp",	(FlukaLayout._what4_type, __BEAM_type1),
			"WHAT(4): Momentum spread type"))
	if card.sign(4):
		layout.append(("Δp(FWHM)",FlukaLayout._what4_val,
			"WHAT(4): Momentum spread Gaussian [GeV/c] FWHM=2.355*sigma"))
	else:
		layout.append(("Δp",	FlukaLayout._what4_val,
			"WHAT(4): Momentum spread FLAT [GeV/c]"))

	layout.append(("Spot Weight", 3))
		#("Printout",   (6, [Layout.empty_what, Layout.on_what,
		#		Layout.default_what, Layout.reset_what])),
	layout.append(Layout.newline)

	# 3rd line (beam divergence)
	layout.append(("Δϕx",	(__SPOTBEAM_divX_type, __BEAM_type2),
			"WHAT(5): X Divergence type"))
	if card.numWhat(5) >= 2000.0*math.pi:
		layout.append(Layout.tab)
	else:
		layout.append(("Δϕx", FlukaLayout._what5_val,
				"WHAT(5): X Divergence [mrad]"))

	layout.append(("Δϕy",	(__SPOTBEAM_divY_type, __BEAM_type2),
			"WHAT(6): Y Divergence type"))
	if card.numWhat(6) >= 2000.0*math.pi:
		layout.append(Layout.tab)
	else:
		layout.append(("Δϕy", FlukaLayout._what6_val,
				"WHAT(6): Y Divergence [mrad]"))
	return layout

#------------------------------- SPOTPOS  --------------------------------------
__SPOTPOS_layout4 = [
		Layout.name,	# spot number
		("x",		 1),
		("y",		 2),
		("z",		 3),
	]

def _SPOTPOS_layout4(card):
	# 1st line
	layout = __SPOTPOS_layout4[:]
	layout.append(Layout.newline)

	# 2nd line (use the same from BEAM card)
	__BEAM_shape_layout(card, layout)
	return layout

#------------------------------- SPOTDIR  --------------------------------------
_SPOTDIR_layout4 = [
		Layout.name,	# spot number
		("cosx",	 1),
		("cosy",	 2),
		("cosz",	 3),
		Layout.newline,
	Layout.tab,
		("cosBxx",	 4),
		("cosBxy",	 5),
		("cosBxz",	 6),
	]

#------------------------------- SPOTTRAN --------------------------------------
_SPOTTRAN_layout4 = [
		Layout.tab,
		("Trans",	(1, Layout.rotationList)),
		("TransBox",	(2, Layout.rotationList)),
		("Type",	(0, ["", "DIRECT", "INVERSE"])),
		Layout.newline,
	Layout.tab,
		("Spot",	 4),
		("to Spot",	 5),
		("Step",	 6) ]

#------------------------------- STOP  -----------------------------------------
_STOP_layout4 = Layout.empty_layout

#------------------------------- TCQUENCH --------------------------------------
_TCQUENCH_layout4 = [ Layout.tab,
		("t cut-off",	 1),
		("Birks c1",	 2),
		("Birks c2",	 3),
		Layout.newline,
	Layout.tab,
		("Bin",		(4, FlukaLayout.binningList)),
		("to Bin",	(5, FlukaLayout.binningList)),
		("Step",	 6) ]

#------------------------------- TIME-CUT --------------------------------------
_TIME_CUT_layout4 = [ Layout.tab,
		("Tcut",	 1),
		("Tdelay",	 2),
		("Start Mat",	(3, FlukaLayout.materialList)),
		Layout.newline,
	Layout.tab,
		("Part",	(4, FlukaParticle.list)),
		("to Part",	(5, FlukaParticle.list)),
		("Step",	 6) ]

#------------------------------- TITLE -----------------------------------------
_TITLE_layout4 = [ (None,	-1) ]

#------------------------------- THRESHOLD -------------------------------------
_THRESHOL_layout4 = [ Layout.tab,
		("Elastic",	 3),
		("Inelastic",	 4),
		("Star",	 6) ]

#------------------------------- TPSSCORE --------------------------------------
__TPSSCORE_sdum = ["","BIOTOBIN"]

_TPSSCORE_layout4_biotobin = [ Layout.tab,
		("Type",	(0, __TPSSCORE_sdum)),
		("Bio set",	(1, FlukaLayout.radbiolList)),
		Layout.newline,
	Layout.tab,
		("Bin",		(4, FlukaLayout.binningList)),
		("to Bin",	(5, FlukaLayout.binningList)),
		("Step",	 6) ]

def _TPSSCORE_layout4(card):
	if card.sdum() == "BIOTOBIN":
		return _TPSSCORE_layout4_biotobin
	else:
		return [Layout.tab,
			("Type", (0, __TPSSCORE_sdum))]

#------------------------------- USERWEIGHT ------------------------------------
__USERWEIG_what3 = [	Layout.empty_what,
			Layout.ignore_what,
			("No weight", -1),
			("FLUSCW", 1),
			("FLUSCW,FLDSCP", 2),
			("FLUSCW+",  3),
			("FLUSCW,FLDSCP+",  4) ]
__USERWEIG_what6 = [	Layout.empty_what,
			Layout.ignore_what,
			("No weight", -1),
			("COMSCW", 1),
			("COMSCW,ENDSCP", 2),
			("COMSCW+", 3),
			("COMSCW, ENDSCP+", 4) ]

_USERWEIG_layout4 = [ Layout.tab,
		("ϕ Weight",		(3, __USERWEIG_what3)),
		("Resnuclei",		(5, [("No weight",0,),("USRRNC",1)])),
		("Density Weight",	(6, __USERWEIG_what6)) ]

#------------------------------- USRBIN ----------------------------------------
__USRBIN_layout4_XYZ = [ Layout.tab, # Layout.name,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("USRBIN"))),
		("Name",	0),
		Layout.newline,
	("Type",		(1, FlukaLayout.USRBIN_type)),
		("Xmin",	 7),
		("Xmax",	 4),
		("NX",		10),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("Ymin",	 8),
		("Ymax",	 5),
		("NY",		11),
		Layout.newline,
	Layout.tab,
		("Zmin",	 9),
		("Zmax",	 6),
		("NZ",		12) ]

__USRBIN_layout4_RZPhi = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("USRBIN"))),
		("Name",	0),
		Layout.newline,
	("Type",		(1, FlukaLayout.USRBIN_type)),
		("Rmin",	 7),
		("Rmax",	 4),
		("NR",		10),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("X",		 8),
		("Y",		 5),
		("NΦ",		11),
		Layout.newline,
	Layout.tab,
		("Zmin",	 9),
		("Zmax",	 6),
		("NZ",		12) ]

__USRBIN_layout4_Region = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("USRBIN"))),
		("Name",	0),
		Layout.newline,
	("Type",		(1, FlukaLayout.USRBIN_type)),
		("R1from",	(7, FlukaLayout.regionList)),
		("R1to",	(4, FlukaLayout.regionList)),
		("Step1",	10),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("R2from",	(8, FlukaLayout.regionList)),
		("R2to",	(5, FlukaLayout.regionList)),
		("Step2",	11),
		Layout.newline,
	Layout.tab,
		("R3from",	(9, FlukaLayout.regionList)),
		("R3to",	(6, FlukaLayout.regionList)),
		("Step3",	12) ]

__USRBIN_layout4_Special = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("USRBIN"))),
		("Name",	0),
		Layout.newline,
	("Type",		(1, FlukaLayout.USRBIN_type)),
		("I1min",	 7),
		("I1max",	 4),
		("Step1",	10),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("I2min",	 8),
		("I2max",	 5),
		("Step2",	11),
		Layout.newline,
	Layout.tab,
		("F3min",	 9),
		("F3max",	 6),
		("N3",		12) ]

def _USRBIN_layout4(card):
	# depending on the value of what(1) there are various layouts
	w1 = card.intWhat(1) % 10
	if w1 == 2:
		return __USRBIN_layout4_Region
	elif w1 == 8:
		return __USRBIN_layout4_Special
	elif w1 in (1, 7):
		return __USRBIN_layout4_RZPhi
	else:
		return __USRBIN_layout4_XYZ

#------------------------------- EVENTBIN --------------------------------------
__EVENTBIN_type  = ("Type", (lambda l,c,v=None,w=1:l._what_abs_val(c,w,v),
						FlukaLayout.USRBIN_type))
__EVENTBIN_print_choice = [("All Cells", 0), ("Non-Zero Cells", 1), ("Function", Layout.FUNCTION) ]
__EVENTBIN_print        = ("Print", (lambda l,c,v=None,w=1:l._what_abs_type(c,w,v),
				__EVENTBIN_print_choice))

__EVENTBIN_layout4_XYZ = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("EVENTBIN"))),
		("Name",	 0),
		Layout.newline,
	__EVENTBIN_type,
		("Xmin",	 7),
		("Xmax",	 4),
		("NX",		10),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("Ymin",	 8),
		("Ymax",	 5),
		("NY",		11),
		Layout.newline,
	__EVENTBIN_print,
		("Zmin",	 9),
		("Zmax",	 6),
		("NZ",		12) ]

__EVENTBIN_layout4_RZPhi = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("EVENTBIN"))),
		("Name",	 0),
		Layout.newline,
	__EVENTBIN_type,
		("Rmin",	 7),
		("Rmax",	 4),
		("NR",		10),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("X",		 8),
		("Y",		 5),
		("NΦ",		11),
		Layout.newline,
	__EVENTBIN_print,
		("Zmin",	 9),
		("Zmax",	 6),
		("NZ",		12) ]

__EVENTBIN_layout4_Region = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("EVENTBIN"))),
		("Name",	 0),
		Layout.newline,
	__EVENTBIN_type,
		("R1from",	(7, FlukaLayout.regionList)),
		("R1to",	(4, FlukaLayout.regionList)),
		("Step1",	10),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("R2from",	(8, FlukaLayout.regionList)),
		("R2to",	(5, FlukaLayout.regionList)),
		("Step2",	11),
		Layout.newline,
	__EVENTBIN_print,
		("R3from",	(9, FlukaLayout.regionList)),
		("R3to",	(6, FlukaLayout.regionList)),
		("Step3",	12) ]

__EVENTBIN_layout4_Special = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("EVENTBIN"))),
		("Name",	 0),
		Layout.newline,
	__EVENTBIN_type,
		("I1min",	 7),
		("I1max",	 4),
		("Step1",	10),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("I2min",	 8),
		("I2max",	 5),
		("Step2",	11),
		Layout.newline,
	__EVENTBIN_print,
		("F3min",	 9),
		("F3max",	 6),
		("N3",		12) ]

def _EVENTBIN_layout4(card):
	# depending on the value of what(1) there are various layouts
	w1 = abs(card.intWhat(1)) % 10
	if w1 == 2:
		return __EVENTBIN_layout4_Region
	elif w1 == 8:
		return __EVENTBIN_layout4_Special
	elif w1 in (1, 7):
		return __EVENTBIN_layout4_RZPhi
	else:
		return __EVENTBIN_layout4_XYZ

#------------------------------- USRCOLL ---------------------------------------
_USRCOLL_layout4 = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("USRCOLL"))),
		("Name",	 0),
		Layout.newline,
	("Type",		(1, _log_what)),
		("Reg",		(4, FlukaLayout.allRegionList)),
		Layout.tab,
		("Vol",		 5),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("Emin",	 8),
		("Emax",	 7),
		("Bins",	 6) ]

#------------------------------- USRTRACK --------------------------------------
_USRTRACK_layout4 = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("USRTRACK"))),
		("Name",	 0),
		Layout.newline,
	("Type",		(1, _log_what)),
		("Reg",		(4, FlukaLayout.allRegionList)),
		Layout.tab,
		("Vol",		 5),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("Emin",	 8),
		("Emax",	 7),
		("Bins",	 6) ]

#------------------------------- USRBDX ----------------------------------------
__linElinW =  1
__logElinW = -1
__linElogW =  2
__logElogW = -2

_USRBDX_type = [ Layout.empty_what,
		("Φ2,LogE,LinΩ", 110 + __logElinW),
		("Φ2,LogE,LogΩ", 110 + __logElogW),
		("Φ2,LinE,LinΩ", 110 + __linElinW),
		("Φ2,LinE,LogΩ", 110 + __linElogW),

		("Φ1,LogE,LinΩ", 100 + __logElinW),
		("Φ1,LogE,LogΩ", 100 + __logElogW),
		("Φ1,LinE,LinΩ", 100 + __linElinW),
		("Φ1,LinE,LogΩ", 100 + __linElogW),

		("I2,LogE,LinΩ", 10 + __logElinW),
		("I2,LogE,LogΩ", 10 + __logElogW),
		("I2,LinE,LinΩ", 10 + __linElinW),
		("I2,LinE,LogΩ", 10 + __linElogW),

		("I1,LogE,LinΩ", __logElinW),
		("I1,LogE,LogΩ", __logElogW),
		("I1,LinE,LinΩ", __linElinW),
		("I1,LinE,LogΩ", __linElogW)]
if Global.developer:
	_USRBDX_type.extend([
		("Φ2,LogE,LinΩ+GW", 10110 + __logElinW),
		("Φ2,LogE,LogΩ+GW", 10110 + __logElogW),
		("Φ2,LinE,LinΩ+GW", 10110 + __linElinW),
		("Φ2,LinE,LogΩ+GW", 10110 + __linElogW),

		("Φ1,LogE,LinΩ+GW", 10100 + __logElinW),
		("Φ1,LogE,LogΩ+GW", 10100 + __logElogW),
		("Φ1,LinE,LinΩ+GW", 10100 + __linElinW),
		("Φ1,LinE,LogΩ+GW", 10100 + __linElogW),

		("I2,LogE,LinΩ+GW", 10010 + __logElinW),
		("I2,LogE,LogΩ+GW", 10010 + __logElogW),
		("I2,LinE,LinΩ+GW", 10010 + __linElinW),
		("I2,LinE,LogΩ+GW", 10010 + __linElogW),

		("I1,LogE,LinΩ+GW", 10000 +__logElinW),
		("I1,LogE,LogΩ+GW", 10000 +__logElogW),
		("I1,LinE,LinΩ+GW", 10000 +__linElinW),
		("I1,LinE,LogΩ+GW", 10000 +__linElogW) ])

_USRBDX_layout4 = [ Layout.tab,
		Layout.tab,
		("Unit",	(3, lambda l: l.fileUnitList("USRBDX"))),
		("Name",	 0),
		Layout.newline,
	("Type",		(1, _USRBDX_type)),
		("Reg",		(4, FlukaLayout.regionList)),
		("to Reg",	(5, FlukaLayout.regionList)),
		("Area",	 6),
		Layout.newline,
	("Part",		(2, FlukaParticle.listAll)),
		("Emin",	 8),
		("Emax",	 7),
		("Ebins",	 9),
		Layout.newline,
	Layout.tab,
		("Ωmin",	11),
		("Ωmax",	10),
		("Ωbins",	12) ]

#------------------------------- USRYIELD --------------------------------------
__USRYIELD_type = [ "Yield", "BEAMDEF" ]
__USRYIELD_iea_type  = [
		(""                  ,  0),
		("Ekin GeV"          ,  1),
		("Ptot GeV/c"        ,  2),
		("Rapidity lab"      ,  3),
		("Rapidity CMS"      ,  4),
		("Pseudorap. lab"    ,  5),
		("Pseudorap. CMS"    ,  6),
		("Feynman-x lab"     ,  7),
		("Feynman-x CMS"     ,  8),
		("Ptr GeV/c"         ,  9),
		("Mtr GeV"           , 10),
		("Plong lab"         , 11),
		("Plong CMS"         , 12),
		("Etot GeV"          , 13),
		("Polar θ lab"       , 14),
		("Polar θ CMS"       , 15),
		("Ptr^2 (GeV/c)**2"  , 16),
		("Weighted Angle lab", 17),
		("Weighted Ptr GeV/c", 18),
		("Plab/Pbeam"        , 19),
		("Etr"               , 20),
		("Eexit"             , 21),
		("Particle charge"   , 22),
		("Particle LET"      , 23),
		("Polar θ lab deg"   , 24),
		("Polar θ CMS deg"   , 25),
		("Ekin/nucl lab"     , 26),
		("P/nucl lab"        , 27),
		("Part.Bar.Charge"   , 28),
		("4 Mom.Tran. -t"    , 29),
		("X_F_long CMS"      , 30),
		("M_x squared"       , 31),
		("M_x squared / s"   , 32),
		("Time (s)"          , 33),
		("Sin avg lab"       , 34),
		("Ptot (GeV/c) cms"  , 35),
		("Etot (GeV/c) cms"  , 36),
		("Wei.Lab.Ang deg"   , 37),
		("Sin avg lab deg"   , 38) ]

__USRYIELD_yield = [ Layout.empty_what,
		#("Plain Double"         ,  0),
		("d2σ/dx1dx2"            ,  0),
		("d2σ/dx1dx2"            ,  1),
		("E d3σ/dp3"             ,  2),
		("d2N/dx1dx2"            ,  3),
		("d2(x2*N)/dx1dx2"       ,  4),
		("d2(x1*N)/dx1dx2"       ,  5),
		("d2N/dx1dx2/cos(θ)"     ,  6),
		("d2(x2^2*N)/dx1dx2"     ,  7),
		("d2(x1^2*N)/dx1dx2"     ,  8),
		("d2(x2*N)/dx1dx2/cos(θ)", 16),
		("d2(x1*N)/dx1dx2/cos(θ)", 26),
		("d2σ/dx1dx2"            , 53),
		("d2(x2*σ)/dx1dx2"       , 54),
		("d2(x1*σ)/dx1dx2"       , 55),
		("d2σ/dx1dx2/cos(θ)"     , 56),
		("d2(x2^2*σ)/dx1dx2"     , 57),
		("d2(x1^2*σ)/dx1dx2"     , 58),
		("d2(x2*σ)/dx1dx2/cos(θ)", 66),
		("d2(x1*σ)/dx1dx2/cos(θ)", 76) ]

__USRYIELD_yield2 = [ Layout.empty_what,
		#("Plain Double"    , 0),
		("d2σ/dx1dx2"       , 0),
		("d2σ/dx1dx2"       , 1),
		("E d3σ/dp3"        , 2),
		("d2N/dx1dx2"       , 3),
		("d2(x2*N)/dx1dx2"  , 4),
		("d2(x1*N)/dx1dx2"  , 5),
		("d2(x2^2*N)/dx1dx2", 7),
		("d2(x1^2*N)/dx1dx2", 8) ]

# Will only allow the change from BEAMDEF to something else
def __USRYIELD_sdum(layout, card, val=None):
	if val is None:
		sdum = card.sdum()[:8]
		if sdum=="BEAMDEF":
			return sdum
		else:
			return "Yield"
	else:
		if val=="BEAMDEF":
			# Remove extra whats
			undoinfo = [ layout.flair.setNWhatsUndo(card, 7) ]
			undoinfo.append(layout.flair.setWhatUndo(card, 0, val))
			layout.flair.addUndo(undoinfo)
		elif card.sdum()=="BEAMDEF" and val=="Yield":
			layout.setWhat(card, 0, val)

__USRYIELD_layout4_beamdef = [
	Layout.tab,
		("Type",	(__USRYIELD_sdum, __USRYIELD_type)),
		("Proj",	(1, FlukaParticle.listAll)),
		("Target",	(2, FlukaParticle.listAll)),
		Layout.newline,
	("p",			 3),
		("cosx",	 4),
		("cosy",	 5),
		("cosz",	 6) ]

# Set the WHAT(1) = ie + ia*100
def __iea(card):
	w1 = card.intWhat(1)
	if w1==0: w1=101
	(ia, ie) = divmod(w1, 100)
	if ie>50:
		ia += 1
		ie = ie-100
	return (ia,ie)

def __USRYIELD_ie(layout, card, val=None):
	(ia,ie) = __iea(card)
	if val is None:
		return abs(ie)
	else:
		if ie<0: val = -val
		layout.setWhat(card, 1, ia*100 + val)

def __USRYIELD_ia(layout, card, val=None):
	(ia,ie) = __iea(card)
	if val is None:
		return ia
	else:
		layout.setWhat(card, 1, val*100 + ie)

def __USRYIELD_log(layout, card, val=None):
	(ia,ie) = __iea(card)
	if val is None:
		if ie<0: return -1
		return 1
	else:
		layout.setWhat(card, 1, ia*100 + val*abs(ie))

def __USRYIELD_kind(layout, card, val=None):
	(ixm, ixa) = divmod(card.intWhat(12), 100)
	if val is None:
		return ixa
	else:
		if val == "":
			layout.setWhat(card, 12, val)
		elif val == 0:
			layout.setWhat(card, 12, 0)
		else:
			layout.setWhat(card, 12, ixm*100 + int(val))

def __USRYIELD_kindMat(layout, card, val=None):
	(ixm, ixa) = divmod(card.intWhat(12), 100)
	if val is None:
		try:
			mat = layout.input.material(ixm, True)
		except Exception as err:
			log.error(err)
			log.errorCont(card)
			return ""
		return mat
	else:
		if ixa != 0:
			ixm = layout.input.material(val, False)
			layout.setWhat(card, 12, ixm*100 + ixa)

def __USRYIELD_particle(card):
	try:
		w = int(card.what(2))
		if w<-80:
			# w = -100-ij <=> ij = -100-w
			entering = True
			ij = -100-w
			part = FlukaParticle.get(ij)
		else:
			entering = False
			part = FlukaParticle.get(w)
	except ValueError:
		entering = False
		part = FlukaParticle.get(card.what(2))
	return part, entering

def __USRYIELD_what2(layout, card, val=None):
	part,entering = __USRYIELD_particle(card)

	if val is None:
		if part is not None:
			return part.name
		else:
			return ""
	else:
		newpart = FlukaParticle.get(val)
		if newpart is None:
			layout.setWhat(card, 2, "")
		elif card.intWhat(4)==-1 and card.intWhat(5)==-2 and entering:
			layout.setWhat(card, 2, -100-newpart.id)
		else:
			layout.setWhat(card, 2, val)

def __USRYIELD_emerging(layout, card, val=None):
	part,entering = __USRYIELD_particle(card)
	if val is None:
		if card.intWhat(4)==-1 and card.intWhat(5)==-2:
			if entering:
				return "ENTERING"
			else:
				return "EMERGING"
		else:
			return ""
	else:
		undoinfo = []
		if val in ("EMERGING", "ENTERING"):
			undoinfo.append(layout.flair.setWhatUndo(card, 4, -1))
			undoinfo.append(layout.flair.setWhatUndo(card, 5, -2))
			if part is not None:
				if val=="EMERGING":
					undoinfo.append(layout.flair.setWhatUndo(card, 2, part.name))
				else:
					undoinfo.append(layout.flair.setWhatUndo(card, 2, -100-part.id))
		else:
			if part is not None and card.intWhat(2)<-80:
				undoinfo.append(layout.flair.setWhatUndo(card, 2, part.name))
			undoinfo.append(layout.flair.setWhatUndo(card, 4, ""))
			undoinfo.append(layout.flair.setWhatUndo(card, 5, ""))
		layout.flair.addUndo(undoinfo)

__USRYIELD_layout4_score = [
	Layout.tab,
		("Type",	(__USRYIELD_sdum, __USRYIELD_type),
				"SDUM: Yield for user yield or BEAMDEF to define beam axes"),
		("Unit",	(3, lambda l: l.fileUnitList("USRYIELD"))),
		("Name",	 0),
		Layout.newline,
	("ie",			(__USRYIELD_ie, __USRYIELD_iea_type),
				"WHAT(1) ie: quantity to score"),
		("ia",		(__USRYIELD_ia, __USRYIELD_iea_type),
				"WHAT(1) ia: quantity to score"),
		("Log",		(__USRYIELD_log, _log_what),
				"WHAT(1) sign: logarithmic or linear scoring"),
		("Norm",	 6),
				# FIXME special <80 means ENTERING in interaction
		Layout.newline,
	("Part",		(__USRYIELD_what2, FlukaParticle.listAll)),
		("Yield",	(__USRYIELD_emerging, ["", "EMERGING", "ENTERING"]),
				"WHAT(4): Emerging from inelastic hadronic interactions is scored"),
		("Reg",		(4, FlukaLayout.regionList)),
		("to Reg",	(5, FlukaLayout.regionList)),
		Layout.newline,
	("Min1",		 8),
		("Max1",	 7),
		("Nbins1",	 9),
		Layout.newline,
	("Min2",		11),
		("Max2",	10),
		("Kind",	(__USRYIELD_kind, __USRYIELD_yield),
				"WHAT(12) ixa: kind of yield or cross section desired"),
		("Mat",		(__USRYIELD_kindMat, FlukaLayout.materialList),
				"WHAT(12) ixm: material of the target for cross section or LET calculations") ]

__USRYIELD_layout4_score_EMERGING = [
	Layout.tab,
		("Type",	(__USRYIELD_sdum, __USRYIELD_type)),
		("Unit",	(3, lambda l: l.fileUnitList("USRYIELD"))),
		("Name",	 0),
		Layout.newline,
	("ie",			(__USRYIELD_ie, __USRYIELD_iea_type),
				"WHAT(1) ie: quantity to score"),
		("ia",		(__USRYIELD_ia, __USRYIELD_iea_type),
				"WHAT(1) ia: quantity to score"),
		("Log",		(__USRYIELD_log, _log_what),
				"WHAT(1) sign: logarithmic or linear scoring"),
		("Norm",	 6),
		Layout.newline,
	("Part",		(__USRYIELD_what2, FlukaParticle.listAll)),
		("Yield",	(__USRYIELD_emerging, ["", "EMERGING", "ENTERING"]),
				"WHAT(4): Emerging from inelastic hadronic interactions is scored"),
		Layout.tab,
		Layout.tab,
		Layout.newline,
	("Min1",		 8),
		("Max1",	 7),
		("Nbins1",	 9),
		Layout.tab,
		Layout.newline,
	("Min2",		11),
		("Max2",	10),
		("Kind",	(__USRYIELD_kind, __USRYIELD_yield2),
				"WHAT(12) ixa: kind of yield or cross section desired"),
		("Mat",		(__USRYIELD_kindMat, FlukaLayout.materialList),
				"WHAT(12) ixm: material of the target for cross section or LET calculations") ]

def _USRYIELD_layout4(card):
	# depending on the value of sdum
	if card.sdum() == "BEAMDEF":
		return __USRYIELD_layout4_beamdef
	elif card.intWhat(4)==-1 and card.intWhat(5)==-2:
		return __USRYIELD_layout4_score_EMERGING
	else:
		return __USRYIELD_layout4_score

#------------------------------- USERDUMP --------------------------------------
__USERDUMP_type = [ "Dump", "UDQUENCH" ]

__USERDUMP_what1 = [ Layout.empty_what, Layout.ignore_what, ("Complete",100), Layout.reset_what ]
__USERDUMP_what3 = [ Layout.empty_what,
			("All",			0),
			("Source",		1),
			("Traj&Cont Losses",	2),
			("Local Losses",	3),
			("Source, Traj&Cont",	4),
			("Source, Local",	5),
			("Traj & All Losses",	6),
			("No Source, Traj, Losses",	7) ]
__USERDUMP_what4 = [	Layout.empty_what,
			Layout.ignore_what,
			("User Defined", 1),
			Layout.reset_what]

def __USERDUMP_sdum(layout, card, val=None):
	if val is None:
		if card.sdum()=="UDQUENCH":
			return "UDQUENCH"
		else:
			return "Dump"
	else:
		layout.setWhat(card, 0, val)

__USERDUMP_layout4_dump = [ Layout.tab,
		("Type",	(__USERDUMP_sdum, __USERDUMP_type),
				"Dump: Creates a collision tape\nUDQUENCH to specify Birks parameter for quenching"),
		("Unit",	(2, lambda l: l.fileUnitList("USERDUMP",False))),
		("File",	 0),
		Layout.newline,
	Layout.tab,
		("What",	(1, __USERDUMP_what1)),
		("Score",	(3, __USERDUMP_what3)),
		("Dump",	(4, __USERDUMP_what4)) ]

__USERDUMP_layout4_udquench = [ Layout.tab,
		("Type",	(__USERDUMP_sdum, __USERDUMP_type)),
		("1st Birks m1", 1),
		("2 Birks m1",	 2),
		Layout.newline,
	("1st Birks m2",	 3),
		("2nd Birks m2", 4),
		("1st Birks m3", 5),
		("2nd Birks m3", 6)]

def _USERDUMP_layout4(card):
	if card.sdum() == "UDQUENCH":
		return __USERDUMP_layout4_udquench
	else:
		return __USERDUMP_layout4_dump

#------------------------------- VOXELS ----------------------------------------
_VOXELS_layout4 = [ Layout.tab,
		("Filename",	(0, ("Voxel file","*.vxl"))),
		Layout.newline,
	Layout.tab,
		("x",		 1),
		("y",		 2),
		("z",		 3),
		Layout.newline,
	Layout.tab,
		("Trans.",	(4, Layout.rotationList)),
		("Debug",	(6, [Layout.empty_what, Layout.default_what, ("Debug", 1)])),
		Layout.tab]

#------------------------------- WW-FACTOr -------------------------------------
__WW_FACTO_profile = [ "", 1, 2, 3, 4, 5, -1 ]
_WW_FACTO_layout4 = [ Layout.tab,
		("RR",		 1),
		("Split",	 2),
		("mult f",	 3),
		Layout.newline,
	("LowE n",		(0, __WW_FACTO_profile)) ] + __reg456_range

#------------------------------- WW-PROFIle ------------------------------------
_WW_PROFI_layout4 = [ Layout.tab,
		("WW extra f",	 1),
		("Importa f",	 2),
		("Profile",	(5, list(range(1,6)))),
		Layout.newline,
	Layout.tab,
		("En low",	(3, FlukaLayout.neutronGroupList)),
		("En high",	(4, FlukaLayout.neutronGroupList)),
		("Step",	 5) ]

#------------------------------- WW-THRESh -------------------------------------
__WW_THRES_layout4_normal = [ Layout.tab,
		("E upper",	 1),
		("E lower",	 2),
		("amp f",	 3),
		Layout.newline,
	("Opt",			(0, __BIASING_prim)),
		("Part",	(4, FlukaParticle.list)),
		("to Part",	(5, FlukaParticle.list)),
		("Step",	 6) ]

__WW_THRES_layout4_lowneut = [ Layout.tab,
		("En upper",	(1, FlukaLayout.neutronGroupList)),
		("En lower",	(2, FlukaLayout.neutronGroupList)),
		("amp f",	 3),
		Layout.newline,
	("Opt",			(0, __BIASING_prim)),
		("Part",	(4, FlukaParticle.list)),
		("to Part",	(5, FlukaParticle.list)),
		("Step",	 6) ]

def _WW_THRES_layout4(card):
	if card.what(4)=="WWLOWNEU" or card.what(5)=="WWLOWNEU":
		return __WW_THRES_layout4_lowneut
	else:
		return __WW_THRES_layout4_normal
