# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

from log import say

#===============================================================================
# FlukaParticle class
#===============================================================================
class FlukaParticle:
	_db           = {}	# bi-directional particle dictionary particle<->id
	list          = []	# list of particles
	hadron        = []	# list of hadrons
	lepton        = []	# list of leptons
	beam          = []	# list of particles accepted in beam card
	listAll       = []	# list of every known particle including generalized
	signedList    = []	# signed list of particles
	signedListAll = []	# signed list of all known particles
	particles     = []	#

	# ----------------------------------------------------------------------
	def __init__(self, idx, name, mass=0.0, comment=None, pdg=0):
		self.id      = idx
		self.name    = name
		self.mass    = mass
		self.comment = comment
		self.pdg     = pdg

	# ----------------------------------------------------------------------
	# add particle
	# ----------------------------------------------------------------------
	@staticmethod
	def add(idx, name, mass=0.0, comment=None, pdg=0):
		from .card import FlukaCard
		p = FlukaParticle(idx, name, mass, comment, pdg)
		FlukaParticle._db[idx]  = p
		FlukaParticle._db[name] = p

		# Populate also _baseLocalDict for the function evaluation
		if mass>0.0:
			n = name
			if name[-1] == "-":
				n = n[:-1]+"n"
			elif name[-1] == "+":
				n = n[:-1]+"p"
			if n.find("-")>=0:
				n = n.replace("-","")
			FlukaCard.globalDict[f"m{n}"] = mass
		return p

	# ----------------------------------------------------------------------
	def __getitem__(self, idx):
		return FlukaParticle.particles[idx]

	# ----------------------------------------------------------------------
	@staticmethod
	def get(idx):
		return FlukaParticle._db.get(idx)

	# ----------------------------------------------------------------------
	# create lists
	# ----------------------------------------------------------------------
	@staticmethod
	def makeLists():
		if FlukaParticle.list: return
		lastpar = FlukaParticle.convert("@LASTPAR",False)
		# All particles
		FlukaParticle.list.append("")
		FlukaParticle.hadron.append("")
		FlukaParticle.lepton.append("")
		FlukaParticle.listAll.append("")
#		for p in FlukaParticle._db:
#			if isinstance(p, str):
		# Go through in ordered by id list
		for i in sorted([FlukaParticle._db[x].id
				for x in FlukaParticle._db
					if isinstance(x,str)]):
			part = FlukaParticle._db[i]
			if part.name != "@LASTPAR":
				FlukaParticle.listAll.append(part.name)
			if part.id < lastpar:
				FlukaParticle.list.append(part.name)
			if abs(part.pdg)<100:
				FlukaParticle.lepton.append(part.name)
			else:
				FlukaParticle.hadron.append(part.name)

		# Beam particles
		FlukaParticle.beam = FlukaParticle.list[:]
		FlukaParticle.beam.append("ISOTOPE")

		FlukaParticle.hadron.append("4-HELLIUM")
		FlukaParticle.hadron.append("3-HELLIUM")
		FlukaParticle.hadron.append("TRITON")
		FlukaParticle.hadron.append("DEUTERON")
		FlukaParticle.hadron.append("HEAVYION")

		# Sort lists
		FlukaParticle.beam.sort()
		FlukaParticle.hadron.sort()
		FlukaParticle.lepton.sort()

		# Only real particles + generalized ones
		FlukaParticle.particles = FlukaParticle.list[:]
		for p in range(201, 228):
			if p in (208, 211, 219, 220, 221, 222): continue
			FlukaParticle.particles.append(FlukaParticle._db[p].name)

		# Sort
		FlukaParticle.list.append("@LASTPAR")
		FlukaParticle.listAll.append("@LASTPAR")

		# Create signed list
		FlukaParticle.signedList += FlukaParticle.list
		for i in range(len(FlukaParticle.signedList)-1, 0, -1):
			p = FlukaParticle.signedList[i]
			if p and p[0] != '@':
				FlukaParticle.signedList.insert(i+1, '-'+p)

		FlukaParticle.signedListAll += FlukaParticle.listAll
		for i in range(len(FlukaParticle.signedListAll)-1, 0, -1):
			p = FlukaParticle.signedListAll[i]
			if p and p[0] != '@':
				FlukaParticle.signedListAll.insert(i+1, '-'+p)

	# ----------------------------------------------------------------------
	# Convert to integer or string
	# ----------------------------------------------------------------------
	@staticmethod
	def convert(pid, tonames=True):
		"""Convert particle from name to number
		pid particle type
		tonames True/False"""

		try:
			idx = int(pid)
			isInt = True
		except:
			if pid == "":
				idx = 0
				isInt = True
			else:
				isInt = False

		if tonames:
			if isInt:
				if idx<=-100 or idx>1000: return str(pid)
				#if 988 <= idx <= 1012:
				#	# Only for DISCARD card
				#	idx = -abs(1000 - idx)
				try:
					return FlukaParticle._db[idx].name
				except:
					say(f"Particle {pid} not found!")
					return "UNKNOWN"
			else:
				return pid
		else:
			if isInt:
				return idx
			try:
				return FlukaParticle._db[pid].id
			except:
				raise Exception(f"Particle {pid} not found!")

	# ----------------------------------------------------------------------
	# scan input for particles
	# ----------------------------------------------------------------------
	@staticmethod
	def scan(input_):
		# FIXME for the moment return everything
		return FlukaParticle.listAll
