# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

import struct

import log
import fortran
from deck   import Card
from .input import Fluka
from .card  import FlukaCard

#===============================================================================
# Fluka Voxel input class
#	parse the voxel file and create fake cards for the
#		materials, assignmats and corrfact
#	that might be embedded in the voxel file
# FIXME: I/O reading should be moved to inout.py (maybe writing in the future)
#===============================================================================
class FlukaVoxel(Fluka):
	MODE = "FlukaVoxel"

	#-----------------------------------------------------------------------
	def __init__(self, filename=None):
		super().__init__()
		self.mode = FlukaVoxel.MODE
		self.filename = ""
		self.cleanup()
		if filename is not None: self.load(filename)

	# ----------------------------------------------------------------------
	# Clean up voxel structure
	# ----------------------------------------------------------------------
	def cleanup(self):
		self.title     = ""
		self.nx        = 0
		self.ny        = 0
		self.nz        = 0
		self.nregions  = 0		# no. of regions (no in fluka)
		self.nvoxels   = 0		# no. unique voxels (mo in fluka)
		self.dx        = 0.0
		self.dy        = 0.0
		self.dz        = 0.0
		self.kreg      = None
		self.roiN      = 0
		self.roiName   = {}		# roi structure names
		self.roiColor  = {}		# colors of rois
		self.roiVolume = {}
		#self.roiComb  = {}		# roi dictionary of combination of structures
		#self.roiData  = None		# roi structure data

	# ----------------------------------------------------------------------
	# Read voxel information
	# ----------------------------------------------------------------------
	def load(self, filename):
		self.filename = filename
		try:
			with open(filename, "rb") as f:
				# Title CHAR*80
				record = fortran.read(f)
				if record is None or len(record) != 80:
					log.error("Invalid voxel file. No title found")
					return False
				self.title = record.strip().decode(errors="replace")

				# Memory Size nx,ny,nz,nregions,nvoxels
				record = fortran.read(f)
				if record is None or len(record) not in (20,32,36):
					log.error("Invalid voxel file. Wrong memory size")
					return False

				# First 5 bytes
				if len(record) < 20:
					log.error("Invalid voxel file. Wrong header size")
					return False

				self.nx, self.ny, self.nz, self.nregions, self.nvoxels = \
						struct.unpack("=5i", record[:20])
				self.roiN = 0
				roiCombN  = 0
				roiPlanar = 0
				record = record[20:]
				if len(record)>=3*4:
					self.roiN, roiCombN, roiMaxLen = struct.unpack("=3i", record[:3*4])
					record = record[3*4:]
					if len(record)>=4:
						roiPlanar = struct.unpack("=i", record[:4])[0]

				# Create voxel regions
				card = FlukaCard("REGION",["VOXEL"])
				#card["@id"] = 0
				card[Card.TYPE] = Card.REGION_VOXEL
				card.setExpr("+VOXEL")
				self.add(card)

				for i in range(1,self.nregions+2):	# +1 to include organ 0
					card = FlukaCard("REGION",[FlukaVoxel.regionName(i)])
					#card["@id"] = i+1
					self.add(card)
					card[Card.TYPE] = Card.REGION_VOXEL

				# Voxel dimension
				record = fortran.read(f)
				if record is None or len(record) != 3*8:
					log.error("Invalid voxel file. Wrong dimensions")
					return False
				self.dx, self.dy, self.dz = struct.unpack("=3d", record)

				# Voxel data (skip)
				skip = fortran.skip(f)
				if skip != self.nx*self.ny*self.nz*2:
					log.error("Invalid voxel file. Data size do not match")
					return False

				# kreg
				record = fortran.read(f)
				if record is None or len(record) != 2*self.nvoxels:
					log.error("Invalid voxel file. "
						"Wrong voxel-2-region (kreg) array")
					return False
				self.kreg = struct.unpack(f"={len(record)//2}H", record)

				# roiStruct
				if self.roiN>0:
					# structures
					for i in range(1,self.roiN+1):
						record = fortran.read(f)
						color, volume, name = struct.unpack("=id64s",record)
						name = name.strip().decode(errors="replace")
						if name:
							self.roiColor[i]  = color
							self.roiVolume[i] = volume
							self.roiName[i]   = name

					# for the moment skip everything
					for i in range(roiCombN):
						record = fortran.read(f)	# skip index
					record = fortran.read(f)		# skip data

					# Read planar structures
					for i in range(roiPlanar):
						record = fortran.read(f)	# skip index
						roi,n = struct.unpack("=ii", record)
						for j in range(n):
							record = fortran.read(f)# skip data

				# Read embedded cards
				while True:
					try:
						record = fortran.read(f)
					except IOError:
						break
					if record is None: break
					if len(record) != 80:
						log.error("Invalid voxel file. "
							f"Wrong card found '{record}'")
						return False
					record = record.decode(errors="replace")
					tag  = record[:8].strip()
					sdum = record[-10:].strip()
					what = [sdum,			# sdum
						record[10:20].strip(),	# what(1)
						record[20:30].strip(),	# what(2)
						record[30:40].strip(),	# what(3)
						record[40:50].strip(),	# what(4)
						record[50:60].strip(),	# what(5)
						record[60:70].strip()]	# what(6)
					if tag=="COMPOUND" and card.tag==tag and card.name()==sdum:
						# append whats to previous card
						card.appendWhats(what[1:])
					else:
						# if COMPOUND check for existing to append
						card = FlukaCard(tag, what)
						self.add(card)
			return True

		except IOError as ex:
			log.error(f"reading voxel file {self.filename!r}")
			log.errorCont(ex)
			self.cleanup()
			return False

	# ----------------------------------------------------------------------
	def __str__(self):
		return "Voxel\n" \
			f"File:\t{self.filename}\n" \
			f"Title:\t{self.title}\n" \
			f"nxyz:\t{self.nx} x {self.ny} x {self.nz}\n" \
			f"nregions,nvoxels:\t{self.nregions}\t{self.nvoxels}\n" \
			f"dxyz:\t{self.dx} x {self.dy} {self.dz}\n"

	# ----------------------------------------------------------------------
	# return a valid region name
	# ----------------------------------------------------------------------
	@staticmethod
	def regionName(idx):
		if idx==0: return "VOXEL"
		num = f"{idx:03}"
		return "VOXEL"[:(8-len(num))] + num
