# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Import once to initialize all codes
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

from collections import defaultdict

import deck
import mcnp
import fluka
import moira
import phits
import Global
import penelope

# default predefined extensions
EXT2MODE = {
		fluka.FlukaConfig.EXT           : fluka.Fluka.MODE,
		fluka.FlukaConfig.EXTENSIONS[1] : fluka.Fluka.MODE,
		mcnp.McnpConfig.EXT             : mcnp.Mcnp.MODE,
		mcnp.McnpConfig.EXTENSIONS[1]   : mcnp.Mcnp.MODE,
		moira.MoiraConfig.EXT           : moira.Moira.MODE,
		moira.MoiraConfig.EXTENSIONS[1] : moira.Moira.MODE,
		moira.MoiraConfig.EXTENSIONS[2] : moira.Moira.MODE,
		penelope.PenelopeConfig.EXT     : penelope.Penelope.MODE,
		phits.PhitsConfig.EXT           : phits.Phits.MODE,
		phits.PhitsConfig.EXTENSIONS[1] : phits.Phits.MODE,
	}

MODES    = []				# all available mods (@see init)
ALL_EXTS = set()			# All input extensions (@see init)
MODE2EXT = defaultdict(list)		# mode to file type (@see init)

MODE2INPUT = {
		fluka.Fluka.MODE       : fluka.Fluka,
		mcnp.Mcnp.MODE         : mcnp.Mcnp,
		moira.Moira.MODE       : moira.Moira,
		penelope.Penelope.MODE : penelope.Penelope,
		phits.Phits.MODE       : phits.Phits,
	}

MODE2CONFIG = {
		fluka.Fluka.MODE       : fluka.FlukaConfig,
		mcnp.Mcnp.MODE         : mcnp.McnpConfig,
		moira.Moira.MODE       : moira.MoiraConfig,
		penelope.Penelope.MODE : penelope.PenelopeConfig,
		phits.Phits.MODE       : phits.PhitsConfig,
	}

#-------------------------------------------------------------------------------
# @param mode   project default mode to give higher priority
# @param name   optionally existing input
# @param other  include also other file types
# @return a preferential list of input files depending the mode and a filename
#-------------------------------------------------------------------------------
def filetypes(mode, name=None, other=True):
	filetypes = []

	#---------------------------------------------------------------
	def addExt(mode):
		nonlocal filetypes
		exts = MODE2EXT[mode]
		mode = mode.capitalize()
		filetypes.append((f"{mode} input files", tuple(f"*{x}" for x in exts)))
		if name is not None:
			filetypes.append((f"{mode} '{name}*' files",
				tuple(f"{name}*{x}" for x in exts)))
	#---------------------------------------------------------------

	# first add current mode
	addExt(mode)
	# all the reset
	if other:
		for m in MODE2EXT.keys():
			if m == mode: continue
			addExt(m)

	filetypes.append(("All","*"))
	return filetypes

#-------------------------------------------------------------------------------
# Initialize all CardInfo
#-------------------------------------------------------------------------------
def init():
	deck.init()
	fluka.init()
	moira.init()
	mcnp.init()
	phits.init()
	penelope.init()

	MODES.append(fluka.Fluka.MODE)
	if Global.developer:
		MODES.append(moira.Moira.MODE)
		MODES.append(mcnp.Mcnp.MODE)
		MODES.append(phits.Phits.MODE)
		MODES.append(penelope.Penelope.MODE)

	deck.CardInfo.post_init()
	fluka.FlukaCardInfo.post_init()
	moira.MoiraCardInfo.post_init()
	mcnp.McnpCardInfo.post_init()
	phits.PhitsCardInfo.post_init()
	penelope.PenelopeCardInfo.post_init()

	# Make the invert MODE2EXT
	# FIXME make it user configurable from the .ini
	for ext,mode in EXT2MODE.items():
		MODE2EXT[mode].append(ext)
		ALL_EXTS.add(ext)
#init()
