# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20-Apr-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import select
import subprocess

import ansi
import PlotEngine

_range_pat  = re.compile(r"set ([xy]2?)range \[(.*?):(.*?)]")
_range_pat2 = re.compile(r".*currently \[(.*?):(.*?)]")

STYLE = [	"lines",
		"points",
		"linespoints",
		"impulses",
		"dots",
		"steps",
		"fsteps",
		"histeps",
		"errorbars",
		"yerrorbars",
		"xerrorbars",
		"xyerrorbars",
		"errorlines",
		"yerrorlines",
		"xerrorlines",
		"xyerrorlines",
		"filledcurves",
		"boxes",
		"boxerrorbars",
		"boxxyerrorbars" ]

SPEC = [	2,	#"lines",	To be checked!!!
		2,	#"points",
		2,	#"linespoints",
		2,	#"impulses",
		2,	#"dots",
		2,	#"steps",
		2,	#"fsteps",
		2,	#"histeps",
		3,	#"errorbars",
		3,	#"yerrorbars",
		3,	#"xerrorbars",
		4,	#"xyerrorbars",
		3,	#"errorlines",
		3,	#"yerrorlines",
		3,	#"xerrorlines",
		4,	#"xyerrorlines",
		2,	#"filledcurves",
		2,	#"boxes",
		4,	#"boxerrorbars",
		4 ]	#"boxxyerrorbars"

SMOOTH = [	"",
		"unique",
		"frequency",
		"cumulative",
		"csplines",
		"acsplines",
		"bezier" ]

defaultTerminal = ""
program = "gnuplot"

#===============================================================================
# Gnuplot class
#===============================================================================
class Gnuplot(PlotEngine.PlotEngine):
	"""
	Interface class to call gnuplot
	"""
	def __init__(self, args=None):
		PlotEngine.PlotEngine.__init__(self)
		self._logger = None
		self.init(args)

	# ----------------------------------------------------------------------
	def init(self, args=None):
		if args:
			prg = f"{program} {args}"
		else:
			prg = program
			if not prg: prg = "gnuplot"
		self.pipe = subprocess.Popen(prg,
					shell=True,
					bufsize=0,
					stdin=subprocess.PIPE,
					stdout=subprocess.PIPE,
					stderr=subprocess.STDOUT,
					close_fds=True)
		self.pipe_in  = self.pipe.stdin
		self.pipe_out = self.pipe.stdout
		self("set mouse")	# Initialize mouse

	# ----------------------------------------------------------------------
	def setLogger(self, logger=None):
		"""Set function to print or log the commands send to gnuplot"""
		self._logger = logger

	# ----------------------------------------------------------------------
	def logMessage(self, msg):
		if self._logger:
			for line in msg.splitlines():
				#self._logger(f"gnuplot> {line}")
				self._logger(line)
		else:
			for line in msg.splitlines():
				#print(f"gnuplot> {line}")
				print(line)

	# ----------------------------------------------------------------------
	def poll(self):
		"""Check if the process has ended and return the returncode else None"""
		self.pipe.poll()
		return self.pipe.returncode

	# ----------------------------------------------------------------------
	def close(self):
		"""Close the pipe"""
		self.pipe_in.close()
		self.pipe_out.close()
		self.pipe.wait()

	# ----------------------------------------------------------------------
	def __del__(self):
		try:
			self.quit()
			self.close()
		except:
			pass

	# ----------------------------------------------------------------------
	def readline(self):
		"""read a line returned from gnuplot"""
		return self.pipe_out.readline().decode(errors="backslashreplace")

	# ----------------------------------------------------------------------
	def writeline(self, s):
		"""Send a string to command, followed by newline"""
		self.pipe_in.write((s+'\n').encode())
		self.pipe_in.flush()

	# ----------------------------------------------------------------------
	# Send a string to command, followed by newline.
	# ----------------------------------------------------------------------
	def write(self, s):
		self.logMessage(s)
		try:
			self.writeline(s)
		except IOError:
			# Try to restart gnuplot
			self.init()
			# Re-execute commands...
			self.writeline(s)

	# ----------------------------------------------------------------------
	def __call__(self, s):
		self.write(s)

	# ----------------------------------------------------------------------
	def flush(self):
		"""Flush the buffer to the piped command"""
		self.pipe_in.flush()

	# ----------------------------------------------------------------------
	def check(self, timeout=None):
		#ready = select.select([self.pipe_out], [], [], 0.0001)
		#if self.pipe_out in ready[0]: return True

		# give a little time for buffers to fill
		try:
			ready = select.select([self.pipe_out], [], [], timeout)
		except KeyboardInterrupt:
			return False
		return self.pipe_out in ready[0]

	# ----------------------------------------------------------------------
	def reset(self):
		"""Perform a reset"""
		self("reset")

	# ----------------------------------------------------------------------
	def quit(self):
		"""Quit the program"""
		self("\nquit")

	# ----------------------------------------------------------------------
	def title(self, title, options=""):
		"""Set the plot title"""
		self(f"set title '{self.escape(title)}' {options}")

	# ----------------------------------------------------------------------
	def label(self, axis, label, options=""):
		"""Set an axis label followed by the options"""
		self(f"set {axis}label '{self.escape(label)}' {options}")

	# ----------------------------------------------------------------------
	def grid(self, on=True):
		"""Turn on/off the grid"""
		if on:
			self("set grid front")
		else:
			self("unset grid")

	# ----------------------------------------------------------------------
	def aspect(self, ratio):
		"""Set the aspec ratio"""
		self(f"set size ratio {ratio}")

	# ----------------------------------------------------------------------
	def tics(self, axis, options=""):
		"""Turn on/off the ticks"""
		self(f"set {axis}tics {options}")

	# ----------------------------------------------------------------------
	def log(self, axis="y", log=True):
		"""Set log or linear scale on axis"""
		if log:
			self(f"set logscale {axis}")
		else:
			self(f"unset logscale {axis}")

	# ----------------------------------------------------------------------
	def range(self, axis, amin="", amax=""):
		"""Set axis range"""
		if amin=="" and amax=="": return
		self(f"set {axis}range [{amin}:{amax}]")

	# ----------------------------------------------------------------------
	def plot(self, cmd, data=None):
		"""Perform a plot command, optionally followed by the data
		e.g. gnuplot.plot("w lines",numpyarray.tolist())
		"""
		if data is None:
			self(f"plot {cmd}")
		else:
			self(f"plot '-' {cmd}")
			# accept a list-of-lists of data and plot it interactively
			if isinstance(data,list):
				for line in data:
					self(" ".join(map(str,line)))
			else:
				self(data)
			self("e")

	# ----------------------------------------------------------------------
	def splot(self, cmd, data=None):
		"""Perform an splot command, optionally followed by the data
		e.g. gnuplot.splot("matrix",numpyarray.tolist())
		"""
		if data is None:
			self(f"splot {cmd}")
		else:
			self(f"splot '-' {cmd}")
			# accept a list-of-lists of data and plot it interactively
			if isinstance(data,list):
				for line in data:
					self(" ".join(map(str,line)))
			else:
				self(data)
			self("e")

	# ----------------------------------------------------------------------
	def replot(self):
		"""Replot last plot"""
		self("replot")

	# ----------------------------------------------------------------------
	def x11(self):
		"""Set x11 as default terminal and raise the plot window"""
		self("set terminal x11 raise")

	# ----------------------------------------------------------------------
	def output(self, output=None):
		"""Set the output file of gnuplot"""
		if output is None:
			self("unset output")
		else:
			self(f"set output '{output}'")

	# ----------------------------------------------------------------------
	def set(self, cmd=""):
		"""Generic gnuplot set command"""
		self(f"set {cmd}")

	# ----------------------------------------------------------------------
	# Send end data marker
	# ----------------------------------------------------------------------
	def end(self):
		"""Send an 'end' of data marker"""
		self("e")

	# ----------------------------------------------------------------------
	def empty(self):
		"""Read all pending output lines from gnuplot"""
		while self.check(1.0):
			print(f"{ansi.BOLD}>e>{self.readline()[:-1]}{ansi.NORMAL}")

	# ----------------------------------------------------------------------
	def version(self):
		"""return version number of gnuplot"""
		if Gnuplot._version is None:
			self("show version")
			cnt = 0
			while self.check(1.0):
				words = self.readline().strip().split()
				if words and words[0] == "Version":
					Gnuplot._version = float(words[1])
				cnt += 1
				if cnt > 1000:
					Gnuplot._version = None
					raise Exception("Gnuplot is not responding")
		return Gnuplot._version

	# ----------------------------------------------------------------------
	def getTerminal(self):
		"""return the default terminal of gnuplot"""
		self("show terminal")
		while self.check(1.0):
			words = self.readline().strip().split()
			if words and words[0] == "terminal":
				return words[3]
		return None

	# ----------------------------------------------------------------------
	def getRange(self, axis="x"):
		"""Return a tuple with the selected axis range ({axis}low, {axis}high)"""
		self(f"show {axis}range")
		low, high  = None, None
		timeout = 1.0
		while self.check(timeout):
			line = self.readline().strip()
			m = _range_pat.match(line)
			if m:
				a = m.group(1)
				if a != axis: continue
				l = m.group(2)
				h = m.group(3)
				try:
					low  = float(l)
					high = float(h)
					if low > high: high,low = low,high
				except:
					pass

			if low is None or high is None:
				m = _range_pat2.match(line)
				if m:
					l = m.group(1)
					h = m.group(2)
					try:
						low  = float(l)
						high = float(h)
						if low > high: high,low = low,high
					except:
						pass
			timeout = 0.1
		return low, high

	# ----------------------------------------------------------------------
	def getRange2(self, axes):
		"""return a 4-tuple with the x and y axis range (xlow,xhigh,ylow,yhigh)"""
		self("show xrange")
		self("show yrange")
		xlow, xhigh  = None, None
		ylow, yhigh  = None, None
		timeout = 1.0
		while self.check(timeout):
			line = self.readline()
			m = _range_pat.match(line.strip())
			if m:
				axis = m.group(1)
				low  = m.group(2)
				high = m.group(3)
				try: low = float(low)
				except: continue
				try: high = float(high)
				except: continue
				if low>high: high,low = low,high
				if axis=="x":
					xlow, xhigh = low, high
				else:
					ylow, yhigh = low, high
			timeout = 0.1
		return xlow, xhigh, ylow, yhigh

	# ----------------------------------------------------------------------
	def pause(self, arg="-1"):
		"""Send a pause command"""
		self(f"pause {arg}")

	# ----------------------------------------------------------------------
	def cd(self, path):
		"""Change plot engine directory to path"""
		if path:
			self(f"cd '{path}'")

	# ----------------------------------------------------------------------
	def test(self):
		"""Send a test command to gnuplot"""
		self("test")

#-------------------------------------------------------------------------------
if __name__ == "__main__":
	#import sys
	gp = Gnuplot()
	mx = 10.0
	#print("Version=",gp.version())
	while mx<10000.0:
		#print(mx)
		gp.reset()
		gp(f"plot [0:{mx}] sin(x)")
		gp("show xrange")
		#gp("show version")
		while gp.check(0.05):
			print(gp.readline())
		#	print("%...time.time(), gp.readline()
		j = 0
		#for i in range(0,10):
		#	j += 1
		#	sys.stdin.readline()
		mx += 1
