# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20-Apr-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import tkinter as tk
from collections import namedtuple

try:
	import matplotlib
	#import matplotlib.pyplot as pyplot
	#matplotlib.use('TkAgg')
	from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
	from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
	from matplotlib.figure import Figure
except ImportError:
	matplotlib = None

import PlotEngine
# Implement the default Matplotlib key bindings.

TkCanvas = namedtuple("TkCanvas", ["toplevel", "canvas", "toolbar"])

# Colors with different name in Matplotlib than gnuplot
_COLOR_TRANSLATE = {
	"dark-blue"         : "darkblue",
	"dark-cyan"         : "darkcyan",
	"dark-goldenrod"    : "darkgoldenrod",
	"dark-gray"         : "darkgray",
	"dark-green"        : "darkgreen",
	"dark-khaki"        : "darkkhaki",
	"dark-magenta"      : "darkmagenta",
	"dark-olivegreen"   : "darkolivegreen",
	"dark-orange"       : "darkorange",
	"dark-red"          : "darkred",
	"dark-salmon"       : "darksalmon",
	"dark-spring-green" : "darkseagreen",
	"dark-turquoise"    : "darkturquoise",
	"light-blue"        : "lightblue",
	"light-coral"       : "lightcoral",
	"light-cyan"        : "lightcyan",
	"light-goldenrod"   : "lightgoldenrodyellow",
	"light-gray"        : "lightgray",
	"light-green"       : "lightgreen",
	"light-pink"        : "lightpink",
	"light-salmon"      : "lightsalmon",
	"orange-red"        : "orangered",
}

_FONT_TRANSLATE = {
		"Arial"     : "Arial",
		"Courier"   : "Courier New",
		"Helvetica" : "Arial",
		"Sans"      : "Comic Sans MS",
		"Serif"     : "MS Reference Sans Serif",
		"Symbol"    : "Symbol",
		"Times"     : "Times New Roman",
	}
'''
_MARKET_STYLE = {
		"dot"                  : ".",
		"+"                    : "+",
		"x"                    : "x",
		"square"               : "s",
		"square-filled"        : "s",
		"triangle-up"          : "^",
		"triangle-up-filled"   : "^",
		"triangle-down"        : "v",
		"triangle-down-filled" : "v",
		"diamond"              : "D",
		"diamond-filled"       : "D",
		}
'''

_MARKET_STYLE = [
		".",
		"+",
		"x",
		"*",
		"s",
		"s",
		"o",
		"o",
		"^",
		"^",
		"v",
		"v",
		"D",
		"D",
		]

#===============================================================================
# Matplotlib class
#===============================================================================
class Matplotlib(PlotEngine.PlotEngine):
	"""
	Interface class to call matplotlib
	"""

	_tkdisplays = {}		# active tk displays

	#-----------------------------------------------------------------------
	def __init__(self):
		PlotEngine.PlotEngine.__init__(self)
		self.fig = Figure() if matplotlib else None

	# ----------------------------------------------------------------------
	def version(self):
		if Matplotlib._version is None:
			a,b,_ = matplotlib.__version__.split(".")
			Matplotlib._version = float(a) + float(b)/10.0
		return Matplotlib._version

	# ----------------------------------------------------------------------
	def reset(self):
		try:
			self.fig.clear()
			self.ax = self.fig.add_subplot(111)
		except tk.TclError:
			pass

	# ----------------------------------------------------------------------
	@staticmethod
	def _correctOptions(options):
		try:
			options["color"] = _COLOR_TRANSLATE[options["color"]]
		except KeyError:
			pass

	# ----------------------------------------------------------------------
	def title(self, title,  options):
		Matplotlib._correctOptions(options)
		# csfont  = options['fontname']
		#cssize  = options.get('fontsize')
		#cscolor = options.get('color')
		#pyplot.title(title,
		#		fontsize = cssize,
		#		color = cscolor,
		#		fontname = csfont)	# FIXME fontname fails
		self.ax.set_title(title, **options)

	# ----------------------------------------------------------------------
	def label(self, axis, label, options):
		if axis=="x":
			self.ax.set_xlabel(label, options)
		elif axis=="y":
			self.ax.set_ylabel(label, options)
		elif axis=="z":
			self.ax.set_zlabel(label, options)

	# ----------------------------------------------------------------------
	def grid(self, on=True):
		self.ax.grid(on)
		if on:
			self.ax.grid(True, linestyle='--', alpha=.3)

	# ----------------------------------------------------------------------
	def aspect(self, ratio):
		self.ax.set_aspect(ratio)

	# ----------------------------------------------------------------------
	def tics(self, axis, options):
		if axis == "x":
			#self.ax.set_xticks()
			self.ax.tick_params(axis, **options) #labelcolor=cscolor)
					# labelsize = cssize)
		elif axis=="y":
			#self.ax.set_yticks()
			self.ax.tick_params(axis, **options) #labelcolor=cscolor)
					# labelsize = cssize)

	# ----------------------------------------------------------------------
	def log(self, axis="y", log=True):
		axes = self.ax #.gca()
		if axis=="x":
			#pyplot.gca().set_xscale(log and 'log' or 'linear')	# symlog?
			axes.set_xscale("log" if log else "linear")
		elif axis=="y":
			#pyplot.gca().set_yscale(log and 'log' or 'linear')
			axes.set_yscale("log" if log else "linear")
		#elif axis=="z":
		#	pyplot.gca().set_zscale(log and 'log' or 'linear')
		#	pyplot.gca().set_zscale("log" if log else "linear")

	# ----------------------------------------------------------------------
	def range(self, axis, amin="", amax=""):
		axes = self.ax #.gca()
		#args = {}
		if axis=="x":
			if amin:
				try: axes.set_xlim(left=float(amin))
				except: pass
			if amax:
				try: axes.set_xlim(right=float(amax))
				except: pass
			#axes.set_xlim(**args)
			#axes.set_xlim((amin, amax))
		elif axis=="y":
			if amin:
				try: axes.set_ylim(bottom=float(amin))
				except: pass
			if amax:
				try: axes.set_ylim(top=float(amax))
				except: pass
			#axes.set_ylim((amin, amax))
		#elif axis=="z":
		#	axes.set_zlim((amin, amax))

	# ----------------------------------------------------------------------
	def show(self, block=False):
		pass
		# pyplot.show()
		# pyplot.show(block)

	#-----------------------------------------------------------------------
	def font(self, name):
		if name in _FONT_TRANSLATE:
			return _FONT_TRANSLATE[name]
		return name

	#-----------------------------------------------------------------------
	def color(self, name):
		if not name: return None
		if name in _COLOR_TRANSLATE:
			return _COLOR_TRANSLATE[name]
		return name

	#-----------------------------------------------------------------------
	def marker(self, name):
		# if name in _MARKET_STYLE:
		return _MARKET_STYLE[name]
		# return name

	# ----------------------------------------------------------------------
	# Prepare a tk display and return active canvas
	# ----------------------------------------------------------------------
	def tkdisplay(self, idx, title=None):
		try:
			toplevel,canvas,_ = Matplotlib._tkdisplays[idx]
		except KeyError:
			toplevel = tk.Toplevel()
			toplevel.protocol("WM_DELETE_WINDOW",
					lambda s=self,idx=idx: s.tkdestroy(idx))

			canvas = FigureCanvasTkAgg(self.fig, toplevel)
			toolbar = NavigationToolbar2Tk(canvas, toplevel)
			toolbar.update()
			canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

			Matplotlib._tkdisplays[idx] = TkCanvas(toplevel, canvas, toolbar)

		# Set the toplevel title if requested
		if title is not None:
			toplevel.title(title)

		return canvas

	# ----------------------------------------------------------------------
	def tkdestroy(self, idx):
		Matplotlib._tkdisplays[idx].toplevel.destroy()
		del Matplotlib._tkdisplays[idx]

#===============================================================================
## Create a matplotlib tk toplevel window
##===============================================================================
#class MPToplevel(tk.Toplevel)
#	#-----------------------------------------------------------------------
#	def __init__(self, master, *args, **kwargs):
#		super().__init__(master, *args, **kwargs)

#-------------------------------------------------------------------------------
def hasMatPlotLib():
	return matplotlib is not None
