#!/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   03-Apr-2014

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re

import tkinter as tk
from tkinter import ttk
import tkinter.font as tkFont
from tkinter.colorchooser import askcolor

import tkExtra

_FONT_SIZE = (   '6',  '7',  '8',  '9', '10', '11', '12', '13',
		'14', '16', '18', '20', '22', '24', '26', '28',
		'32', '36', '40', '48', '56', '64', '72' )

_FONT_JUSTIFY = (tk.LEFT, tk.CENTER, tk.RIGHT)

try:
	import PIL.Image
	import PIL.ImageTk
	have_PIL = True
except:
	have_PIL = False

# User readable styles
ABBREV = {
	"Heading 1"  : "h1",
	"Heading 2"  : "h2",
	"Heading 3"  : "h3",
	"Bold"       : "b",
	"Italic"     : "i",
	"BoldItalic" : "bi",
	"Underline"  : "u",
	"Overstrike" : "o",
	"Normal"     : "n",
	"Link"       : "link",
	"Caption"    : "caption",

	"Left"       : "l",
	"Right"      : "r",
	"Center"     : "c",
	}
# and inverted ones
INV_ABBREV = {y:x for (x,y) in ABBREV.items()}

#===============================================================================
# RichText widget
#===============================================================================
class RichText(tk.Text):
	TAGPAT = re.compile(r"{(\w+):\s*(.*?)}", re.MULTILINE|re.DOTALL)
	TAGS = {
		"n"   : {"font"       : ("Sans",10,""),
			 "justify"    : tk.LEFT,
			 "foreground" : "#000000",
			 "background" : "#FFFFFF"},
		"h1"  : {"font"       : (None,16,"bold"),
			 "foreground" : "DarkBlue",
			 "justify"    : tk.LEFT},
		"h2"  : {"font"       : (None,14,"bold"),
			 "foreground" : "DarkBlue",
			  "justify"   : tk.LEFT},
		"h3"  : {"font"       : (None,14,"italic"),
			 "foreground" : "DarkBlue",
			 "justify"    : tk.LEFT},
		"b"   : {"font"       : (None,None,"bold")},
		"i"   : {"font"       : (None,None,"italic")},
		"bi"  : {"font"       : (None,None,"bold italic")},
		"u"   : {"font"       : (None,None,"underline")},
		"o"   : {"font"       : (None,None,"overstrike")},
		"l"   : {"justify"    : tk.LEFT},
		"r"   : {"justify"    : tk.RIGHT},
		"c"   : {"justify"    : tk.CENTER},
		"img" : {},
		"link": {"font"         : (None,None,"underline"),
			 "foreground"   : "Blue"},
		"caption":{"foreground" : "DarkBlue"}
		}
	LINETAGS = ["h1", "h2", "h3", "l", "c", "r"]

	# --------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, *args, **kwargs)
		self.clear()

	# --------------------------------------------------------------------
	def clear(self):
		self.images = {}
		self._img   = {}

	# --------------------------------------------------------------------
	# Open file and highlight
	# @return list of errors
	# --------------------------------------------------------------------
	def open(self, filename):
		try:
			f = open(filename, "r")
		except IOError:
			# Error file not found!!!
			self.clear()
			return f"File {filename} not found"
		data = f.read()
		f.close()
		return self.setText(data)

	# --------------------------------------------------------------------
	# Return style with inheritance
	# --------------------------------------------------------------------
	@staticmethod
	def getStyle(tag):
		style = RichText.TAGS.get(tag)
		if style is None: return None

		based = style.get("based")
		if based is None:
			based = RichText.TAGS.get("n")
		else:
			based = RichText.getStyle(based)
		if based is None: return None

		newstyle = based.copy()
		try: del newstyle["background"]
		except: pass
		for tag,value in style.items():
			if tag == "based": continue
			if value is None: continue

			if isinstance(value,tuple):
				newvalue = []
				for i,v in enumerate(value):
					if v is None:
						v = based[tag][i]
					newvalue.append(v)
				newstyle[tag] = tuple(newvalue)
			else:
				# Replace with the new value
				newstyle[tag] = value

		return newstyle

	# --------------------------------------------------------------------
	def configureTags(self):
		# Configure tags
		for tag in RichText.TAGS:
			style = RichText.getStyle(tag)
			if style: self.tag_config(tag, style)

		# Basic colors of text window
		style = RichText.TAGS["n"]
		for n in ("font", "background", "foreground"):
			self[n] = style[n]

		self.tag_bind("image","<Double-1>", self.imageDouble)
		self.tag_bind("link", "<Double-1>", self.linkDouble)

	# --------------------------------------------------------------------
	def imageDouble(self, event):
		self.event_generate("<<Image>>")
		return "break"

	# --------------------------------------------------------------------
	def linkDouble(self, event):
		self.event_generate("<<Link>>")
		return "break"

	# --------------------------------------------------------------------
	def insertImage(self, filename):
		image = self.image(filename)
		if image:
			img = self.image_create(tk.INSERT, image=image)
			self.tag_add("image",tk.INSERT,tk.INSERT)
			self._img[img] = filename

	# --------------------------------------------------------------------
	def imageName(self, img): return self._img.get(img)

	# --------------------------------------------------------------------
	# Set data to widget and highlight it
	# --------------------------------------------------------------------
	def setText(self, data):
		self.clear()
		last = 0
		self.delete("0.0",tk.END)
		for pat in re.finditer(RichText.TAGPAT, data):
			# text outside tag
			self.insert(tk.END, data[last:pat.start()])
			tag = pat.group(1)
			value = pat.group(2)
			if tag == "img":
				image = self.image(value)
				if image:
					img = self.image_create(tk.END, image=image)
					self.tag_add("image",self.index(img))
					self._img[img] = value
				else:
					self.insert(tk.END,pat.group(0))
			else:
				self.insert(tk.END, value, tag)
			last = pat.end()
		self.insert(tk.END, data[last:])

		self.configureTags()

	# --------------------------------------------------------------------
	# Get rich text from widget
	# --------------------------------------------------------------------
	def getText(self):
		data = ""
		tag = None
		for key,value,index in self.dump("0.0",tk.END):
			if key=="tagon":
				if value not in ("sel","image"):
					if tag is not None: data += "}"
					tag = value
					data += f"{{{tag}:"
			elif key=="text":
				data += value
			elif key=="tagoff":
				if tag is not None: data += "}"
				tag  = None
			elif key=="image":
				if tag is not None: data += "}"
				name = self._img.get(value)
				data += f"{{img:{name}}}"
				if tag is not None: data += f"{{{tag}: "
			elif key=="mark":
				pass
			elif key=="window":
				pass
		return data

	# --------------------------------------------------------------------
	# Find image
	# --------------------------------------------------------------------
	def image(self, filename):
		if not have_PIL: return None
		try:
			s = os.stat(filename)
		except:
			return None

		image = self.images.get(filename)
		if image is not None:
			image, mt = image
			if s.st_mtime > mt: image = None

		if image is None:
			try:
				image = PIL.ImageTk.PhotoImage(PIL.Image.open(filename))
				self.images[filename] = image, s.st_mtime
			except:
				image = None
		return image

	# --------------------------------------------------------------------
	def setStyle(self, style):
		insert = self.index(tk.INSERT)
		style = ABBREV.get(style,style)
		try:
			sel_first = self.index(tk.SEL_FIRST)
			sel_last  = self.index(tk.SEL_LAST)

			if style in RichText.LINETAGS:
				sel_first = f"{sel_first} linestart"
				sel_last  = f"{sel_last} lineend"
		except tk.TclError:
			cur = self.index(tk.INSERT)
			if style in RichText.LINETAGS:
				sel_first = f"{cur} linestart"
				sel_last  = f"{cur} lineend"
			else:
				sel_first = f"{cur} - 1 char"
				sel_last  = cur

		names = self.tag_names(sel_first)

		for n in names:
			if n in RichText.LINETAGS:
				first = f"{sel_first} linestart"
				last  = f"{sel_last} lineend"
				break
		else:
			first = sel_first
			last  = sel_last

		for tag in RichText.TAGS:
			self.tag_remove(tag, first, last)

		if style!="n" and style not in names:
			self.tag_add(style, sel_first, sel_last)

		# Why? + 1 chars (but it works)
		self.mark_set(tk.INSERT,f"{insert} + 1 chars")

	# --------------------------------------------------------------------
	def styleList(self):
		styles = list(ABBREV)
		# Add custom styles here
		styles.sort()
		return styles

	# --------------------------------------------------------------------
	def styleAtCursor(self):
		for n in self.tag_names(tk.INSERT):
			if n in RichText.TAGS:
				return INV_ABBREV.get(n,n)
		return "Normal"

#===============================================================================
# Style configuration
#===============================================================================
class StyleConfigurationDialog(tk.Toplevel):
	def __init__(self, master, text, active=None):
		super().__init__(master)
		self.title("Style configuration")
		self.transient(master)
		self.text = text

		# Variables
		self.fontChanged = False
		self.fontSize   = tk.StringVar(self)
		self.fontAbs    = tk.BooleanVar(self)
		self.fontBold   = tk.BooleanVar(self)
		self.fontItalic = tk.BooleanVar(self)
		self.fontUnderline = tk.BooleanVar(self)
		self.fontStrike = tk.BooleanVar(self)
		self.fontFamily = tk.StringVar(self)
		self.font       = tkFont.Font(self, ('helvetica', 12, 'normal'))
		self.fontSize.set('12')
		self.fontFamily.set('helvetica')
		self.fontJustify = tk.StringVar(self)
		self.fontJustify.set(tk.LEFT)

		# Style selection frame
		frame = tk.Frame(self, relief=tk.RAISED)
		frame.pack(side=tk.TOP, fill=tk.X, pady=2)

		tk.Label(frame, text="Style:").pack(side=tk.LEFT)

		self.styleCombo = tkExtra.Combobox(frame, #False,
					width=12,
					background="White",
					command=self.setStyleFromCombo)
		self.styleCombo.pack(side=tk.LEFT)
		self.styleCombo.bindWidgets("<Button-3>", self.toggleField)
		tkExtra.Balloon.set(self.styleCombo, "Select style to edit")

		# Style editing frame
		frame = tk.Frame(self, relief=tk.SUNKEN, pady=2, padx=1)
		frame.pack(expand=tk.YES, fill=tk.BOTH)

		f = tk.Frame(frame)
		f.pack(side=tk.TOP, padx=2, pady=2, fill=tk.BOTH, expand=tk.YES)

		#tk.Label(f, justify=tk.LEFT, text='Font:').pack(side=tk.TOP, anchor=tk.W)
		self.listFontName = tkExtra.ExListbox(f,
					height=5,
					takefocus=tk.FALSE,
					exportselection=tk.FALSE)
		self.listFontName.pack(side=tk.LEFT, expand=tk.TRUE, fill=tk.BOTH)
		self.listFontName.bind('<ButtonRelease-1>',
					self.listFontButtonRelease)
		#self.listFontName.bind("<<ListboxSelect>>",
		#			self.listFontButtonRelease)
		self.listFontName.bind("<KeyRelease>",
					self.listFontKeyRelease)
		self.listFontName.bind("<Button-3>", self.toggleField)
		sb = ttk.Scrollbar(f)
		sb.pack(side=tk.LEFT, fill=tk.Y)
		sb.config(command=self.listFontName.yview)
		self.listFontName.config(yscrollcommand=sb.set)

		# --- Param ---
		f = tk.Frame(frame)
		f.pack(padx=5, pady=2, fill=tk.X)
		tk.Label(f, text='Size:').pack(side=tk.LEFT, anchor=tk.W)
		self.menuFontSize = tkExtra.ExOptionMenu(f,
					self.fontSize,
					command=self.setFontSample,
					*_FONT_SIZE)
		self.menuFontSize.pack(side=tk.LEFT, anchor=tk.W)
		self.menuFontSize.bind("<Button-3>", self.toggleField)

		self.fontAbs.set(1)
		self.checkFontAbs = tk.Checkbutton(f, text='DPI',
					variable=self.fontAbs,
					onvalue=1, offvalue=0,
					command=self.setFontSample)
		self.checkFontAbs.pack(side=tk.LEFT, anchor=tk.W)
		self.checkFontAbs.bind("<Button-3>", self.toggleField)

		self.checkFontBold = tk.Checkbutton(f, text='Bold',
					variable=self.fontBold,
					onvalue=1, offvalue=0,
					command=self.setFontSample)
		self.checkFontBold.pack(side=tk.LEFT, anchor=tk.W)
		self.checkFontBold.bind("<Button-3>", self.toggleField)

		self.checkFontItalic = tk.Checkbutton(f, text='Italic',
					variable=self.fontItalic,
					onvalue=1, offvalue=0,
					command=self.setFontSample)
		self.checkFontItalic.pack(side=tk.LEFT, anchor=tk.W)
		self.checkFontItalic.bind("<Button-3>", self.toggleField)

		self.checkFontUnderline = tk.Checkbutton(f, text='Underline',
					variable=self.fontUnderline,
					onvalue=1, offvalue=0,
					command=self.setFontSample)
		self.checkFontUnderline.pack(side=tk.LEFT, anchor=tk.W)
		self.checkFontUnderline.bind("<Button-3>", self.toggleField)

		self.checkFontStrike = tk.Checkbutton(f, text='Strike',
					variable=self.fontStrike,
					onvalue=1, offvalue=0,
					command=self.setFontSample)
		self.checkFontStrike.pack(side=tk.LEFT, anchor=tk.W)
		self.checkFontStrike.bind("<Button-3>", self.toggleField)

		# ---
		f = tk.Frame(frame)
		f.pack(padx=5, pady=2, fill=tk.X)
		tk.Label(f, text="Color: Foreground").pack(side=tk.LEFT)
		self.fontForeground = tk.Button(f, text="         ", command=self.foregroundColor,
					background="Black", activebackground="Black")
		self.fontForeground.pack(side=tk.LEFT)
		self.fontForeground.bind("<Button-3>", self.toggleField)

		tk.Label(f, text="Background").pack(side=tk.LEFT)
		self.fontBackground = tk.Button(f, text="         ", command=self.backgroundColor,
					background="White", activebackground="White")
		self.fontBackground.pack(side=tk.LEFT)
		self.fontBackground.bind("<Button-3>", self.toggleField)

		# ---
		f = tk.Frame(frame)
		f.pack(padx=5, pady=2, fill=tk.X)
		tk.Label(f, text="Justify:").pack(side=tk.LEFT)
		self.menuJustify = tkExtra.ExOptionMenu(f,
					self.fontJustify,
					command=self.setFontSample,
					*_FONT_JUSTIFY)
		self.menuJustify.pack(side=tk.LEFT, anchor=tk.W)

		# --- Sample ---
		f = tk.Frame(frame, relief=tk.SOLID, borderwidth=1)
		f.pack(side=tk.TOP, padx=5, pady=5, expand=tk.TRUE, fill=tk.BOTH)

		self.labelFontSample = tk.Label(f,
			text=   "AaBbCcDdEe\nFfGgHhIiJjK\n"
				"AαBβΓγΔδEε\n"
				"ZζHηΘθIιKκ\n"
				"ΛλMμNνΞξOo\n"
				"ΠπPρΣσTτYυΦ\n"
				"ϕXχΨψΩω\n"
				"1234567890\n#:+=(){}[]",
			justify=tk.LEFT, font=self.font)
		self.labelFontSample.pack(expand=tk.TRUE, fill=tk.BOTH)

		self.protocol("WM_DELETE_WINDOW", self.close)
		self.bind('<Escape>', self.close)

		# Button frame
		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, fill=tk.X)
		tk.Button(frame, text="Help", state=tk.DISABLED, command=self.help).pack(side=tk.RIGHT)
		tk.Button(frame, text="Close", command=self.close).pack(side=tk.RIGHT)
		tk.Button(frame, text="Apply", command=self.apply).pack(side=tk.RIGHT)

		# Load fonts
		self.loadFonts()
		self.styleCombo.fill(self.text.styleList())
		self._current = None
		if active:
			self.styleCombo.set(active)
		else:
			self.styleCombo.set("Normal")

		# Wait action
		if not self.winfo_ismapped():
			self.wait_visibility()
		self.grab_set()
		self.focus_set()
		self.wait_window()

	# --------------------------------------------------------------------
	def apply(self, event=None):
		if self.fontChanged: self.saveStyle()
		if self.text is not None:
			self.text.configureTags()

	# --------------------------------------------------------------------
	def close(self, event=None):
		self.apply()
		self.destroy()

	# --------------------------------------------------------------------
	def saveStyle(self):
		name  = self.styleCombo.get()
		short = ABBREV.get(name, name)
		style = RichText.getStyle(short)

		if self.listFontName["state"] != tk.DISABLED:
			family = self.fontFamily.get()
		else:
			family = None

		if self.menuFontSize["state"] != tk.DISABLED:
			size = int(self.fontSize.get())
			if not self.fontAbs.get():
				size = -size
		else:
			size = None

		options = ""
		if self.checkFontBold["state"] != tk.DISABLED or \
		   self.checkFontItalic["state"] != tk.DISABLED or \
		   self.checkFontUnderline["state"] != tk.DISABLED or \
		   self.checkFontStrike["state"] != tk.DISABLED:
			if self.fontBold.get(): options += " bold"
			if self.fontItalic.get(): options += " italic"
			if self.fontUnderline.get(): options += " underline"
			if self.fontStrike.get(): options += " overstrike"

		style["font"] = (family, size, options.strip())

		if self.menuJustify["state"] != tk.DISABLED:
			style["justify"] = self.fontJustify.get()

		if self.fontForeground["state"] != tk.DISABLED:
			style["foreground"] = self.fontForeground["background"]

		if self.fontBackground["state"] != tk.DISABLED:
			b = self.fontBackground["background"]
			if short != "n" and RichText.TAGS["n"]["background"] == b:
				try: del style["background"]
				except: pass
			else:
				style["background"] = b
		RichText.TAGS[short] = style
		self.fontChanged = False

	# --------------------------------------------------------------------
	def setStyleFromCombo(self):
		# Save current!
		if self._current and self.fontChanged:
			self.saveStyle()

		# New style
		name  = self.styleCombo.get()
		short = ABBREV.get(name, name)
		if short == self._current: return
		style = RichText.getStyle(short)

		self._current = short

		# Font
		family, size, options = style["font"]
		self.fontFamily.set(family)
		self.fontSize.set(size)
		self.fontBold.set("bold" in options)
		self.fontItalic.set("italic" in options)
		self.fontUnderline.set("underline" in options)
		self.fontStrike.set("overstrike" in options)

		self.fontJustify.set(style.get("justify",tk.LEFT))

		c = style.get("foreground","#000000")
		self.fontForeground.config(background=c, activebackground=c)
		c = style.get("background","#FFFFFF")
		self.fontBackground.config(background=c, activebackground=c)

		self.listFontName.selection_clear(0, tk.END)
		for i in range(self.listFontName.size()):
			if self.listFontName.get(i) == family:
				self.listFontName.selection_set(i)
				self.listFontName.selection_anchor(i)
				self.listFontName.activate(i)
				self.listFontName.see(i)
				break
		self.setFontSample()
		self.fontChanged = False

	# --------------------------------------------------------------------
	def loadFonts(self):
		fonts = list(tkFont.families(self))
		fonts.sort()
		# remove duplicates
		self.listFontName.insert(tk.END,'')
		for font in fonts:
			self.listFontName.insert(tk.END, font)
		self.listFontName.see(0)
		self.listFontName.activate(0)
		self.listFontName.selection_set(0)

	# --------------------------------------------------------------------
	def listFontButtonRelease(self, event=None):
		font = self.listFontName.get(tk.ANCHOR)
		self.fontFamily.set(font)
		self.setFontSample()

	# --------------------------------------------------------------------
	def listFontKeyRelease(self, event=None):
		self.after_idle(self.listFontButtonRelease)

	# --------------------------------------------------------------------
	def setFontSample(self, event=None):
		self.fontChanged = True
		fontFamily = self.fontFamily.get()

		if self.fontAbs.get():
			fontSign = 1
		else:
			fontSign = -1

		if self.fontBold.get():
			fontWeight = tkFont.BOLD
		else:
			fontWeight = tkFont.NORMAL

		if self.fontItalic.get():
			fontSlant = tkFont.ITALIC
		else:
			fontSlant = tkFont.ROMAN

		fontUnder  = bool(self.fontUnderline.get())
		fontStrike = bool(self.fontStrike.get())

		try:
			self.font.config(
				size=fontSign*int(self.fontSize.get()),
				weight=fontWeight,
				slant=fontSlant,
				underline=fontUnder,
				overstrike=fontStrike,
				family=fontFamily)
		except UnicodeEncodeError:
			pass

		j = self.fontJustify.get()
		self.labelFontSample["justify"] = j
		if   j==tk.LEFT:  self.labelFontSample["anchor"] = tk.W
		elif j==tk.RIGHT: self.labelFontSample["anchor"] = tk.E
		else:          self.labelFontSample["anchor"] = tk.CENTER

		self.labelFontSample["foreground"] = self.fontForeground["background"]
		self.labelFontSample["background"] = self.fontBackground["background"]

	# --------------------------------------------------------------------
	def chooseColorFromButton(self, button, title):
		try:
			rgb, colorStr = askcolor(title=title,
						initialcolor=button["background"],
						parent=self)
		except tk.TclError:
			colorStr = None
		if colorStr is not None:
			colorStr = str(colorStr)
			button.config(background=colorStr, activebackground=colorStr)
			self.setFontSample()

	# --------------------------------------------------------------------
	def foregroundColor(self, event=None):
		self.chooseColorFromButton(self.fontForeground, "Foreground color")

	# --------------------------------------------------------------------
	def backgroundColor(self, event=None):
		self.chooseColorFromButton(self.fontBackground, "Background color")

	# ----------------------------------------------------------------------
	def toggleField(self, event):
		if event.widget["state"] != tk.DISABLED:
			state = tk.DISABLED
		else:
			state = tk.NORMAL

		if isinstance(event.widget.master, tkExtra.Combobox):
			event.widget.master.config(state=state)
		else:
			event.widget.config(state=state)

	# ----------------------------------------------------------------------
	def help(self):
		pass

# -----------------------------------------------------------------------------
if __name__ == "__main__":
	import sys
	def close(event=None):
		global root
		root.destroy()

	root = tk.Tk()
	root.protocol("WM_DELETE_WINDOW", root.destroy)
	root.bind("<Escape>", close)
	root.bind("<Control-Key-q>", close)
	root.bind("<Control-Key-w>", close)

	toolbar = tk.Frame(root)
	toolbar.pack(side=tk.TOP, fill=tk.X)

	text = RichText(root)
	text.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

	for name in RichText.TAGS:
		cmd = lambda w=text,n=name: w.setStyle(n)
		b = tk.Button(toolbar, text=name, command=cmd)
		b.pack(side=tk.LEFT)

	text.open(sys.argv[1])
	#print text.getText()

	root.mainloop()
