# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   18-May-2004

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re

class CSGException(Exception): pass

MAXEXPR = 10000

# Zone formats in sync with gzone.h
STD       = 0	# standard format in products
RPN       = 1	# Reverse Polish format
NORMAL    = 2	# Normal expression
#RPNEXP    = 3	# Reverse Polish expanded form after rpnorm with no nesting
#NORMALEXP = 4	# Normal expression as product | product...

PAT_HASH = re.compile(r"#\s*")

# ----------------------------------------------------------------------
def tokenize(expr):
	"""return a list of expression tokens"""
	line = "".join(expr)
	line = line.replace("+", " + ")
	line = line.replace("-", " - ")
	line = line.replace("|", " | ")
	line = line.replace(":", " : ")
	line = line.replace("(", " ( ")
	line = line.replace(")", " ) ")
	if "#" in line:
		line = PAT_HASH.sub(" #", line)
		#line = line.replace("#+(", " - ( ")
		line = line.replace("#(", " - ( ")
	return line.split()

# ----------------------------------------------------------------------
def splitZones(expr):
	"""split a tokenized expression into zones"""
	zones = []
	zone  = []
	depth = 0
	for token in expr:
		if token in "|:" and depth==0:
			if zone:
				zones.append(zone)
				zone = []
			continue
		if token == "(":
			depth += 1
		elif token == ")":
			depth -= 1
			if depth<0:
				raise CSGException("Too many closing parenthesis")
		zone.append(token)
	if zone: zones.append(zone)	# normal format
	return zones

# ------------------------------------------------------------------------------
# Convert a list of STD expression to a fluka zone string
#   +  -> inside
#   -  -> outside
# ------------------------------------------------------------------------------
def toStringFluka(fmt, expr, maxlen=70):
	"""return a string version of a tokenize expression"""
	output = ""
	newline = 0		# last new line position
	if fmt==STD:
		op = "+"
		for b in expr:
			if b=="-":
				op="-"
			else:
				output += f"{op}{b}"
				if len(output)-newline > maxlen:
					output += "\n"
					newline = len(output)
				else:
					output += " "
		if output: output = output[:-1]
	else:
		if fmt==RPN: expr = rpn2exp(expr)
		prev = "("
		depth = 0
		for token in expr:
			if prev == "(":
				depth += 1
				output += token
			elif token == "|":
				if depth==1:
					output += "\n| "
					newline = len(output)
				else:
					output += " | "
			elif token in ("+", "-"):
				output += f" {token}"
			elif token==")":
				output += token
				depth -= 1
			else:
				output += token
			if len(output)-newline > maxlen:
				output += "\n"
				newline = len(output)
			prev = token
	return output

# ------------------------------------------------------------------------------
# Convert a list of STD expression to an MCNP expression
#   +  -> outside
#   -  -> inside
#   #  -> complement
#   () -> +()
# ------------------------------------------------------------------------------
def toStringMcnp(fmt, expr, maxlen=70):
	"""return a string version of a tokenize expression"""
	output = ""
	newline = 0		# last new line position
	if fmt==STD:
		op = "-"
		for b in expr:
			if b=="-":
				op="+"
			else:
				output += f"{op}{b}"
			if len(output)-newline > maxlen:
				output += "\n"
				newline = len(output)
			else:
				output += " "
		if output: output = output[:-1]
	else:
		if fmt==RPN: expr = rpn2exp(expr)
		start = True
		depth = 0
		sign  = "+"
		for token in expr:
			if token == "(":
				if sign == "+":
					output += " ("
				else:
					output += " #("
				depth += 1
				start = True

			elif token in "|:":
				if depth == 0:
					output += "\n: "
					newline = len(output)
				else:
					output += " : "
				sign  = "+"
				start = True

			elif token in "+-":
				sign = token

			elif token == ")":
				output += token
				depth -= 1
				start = False

			else:
				if not start: output += " "
				output += "-" if sign=="+" else "+"
				output += token
				start = False

			if len(output)-newline > maxlen:
				output += "\n"
				newline = len(output)

	return output

#-------------------------------------------------------------------------------
# Convert a tokenized expression into string
# @param inv	invert +/- False for FLUKA, True for MCNP (+,- inverted)
# @apram fmt	0=STD for product, 1=RPN, 2=NORMAL
# @param expr	tokenized expression to format
# @param maxlen	truncate to maximum length each line string
# @return string expression
#-------------------------------------------------------------------------------
def toString(inv, fmt, expr, maxlen=70):
	if not inv:
		return toStringFluka(fmt, expr, maxlen)
	else:
		return toStringMcnp(fmt, expr, maxlen)

# ----------------------------------------------------------------------
def check(expr):
	"""check depth of an rpn expression"""
	depth = 0
	for token in expr:
		if token in "+-|:#":
			depth -= 1
		else:
			depth += 1
	return depth == 1

# ----------------------------------------------------------------------
#def checkExpr(expr):
#	"""check a normal expression for correctness"""
#	depth = 0
#	prev  = None
#	for token in expr:
#		if token in "|:+-#":
#			if prev in

# ----------------------------------------------------------------------
# Subroutine:	exp2rpn
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20/4/2004
#
# Convert the FLUKA Boolean expression to Reverse Polish Notation (RPN)
# Since the normalization routine does not accept the +/- as signs to
# objects, the routine is converting the leading - to @- (Universe minus)
# where the special symbol @ is treated as the universe.
# Furthermore the leading + is ignored as well as the leading | which is
# accepted by fluka.
#
# WARNING: No check is done for the correctness of the expression apart
#          from the parenthesis nesting
#
# ie.
#        A+B         -> A B +
#       (A+B)|C      -> A B + C |
#       (A|B)|C+D|E  -> A B | C D + | E |
#       -A           -> @ A -
#       -(-A)        -> @ @ A - -
#
# The routine is using the same array for returning the Reverse Polish
# expression, since the format is more compact.
# This is generally true apart one case -A -> @ A -
#
# Priorities are treated as:
#     Operator  Priority         In     Out
#     --------  --------        ---     ---
#       |       lower             1       2
#       +       high              3       4
#       -       high              3       4
#       (       higher           99       0
#       )       higher            0      99
#       object  highest         101     100
#
# Algorithm
#  Consider the expression as a train moving on a railroad with a
#  T-shape, where each token is one wagon
#
#                      <-  (A|B)|C+D
#      ------------.   .------------
#      RPN-End      \ /      Exp-End
#                    |
#                   S|
#                   t|
#                   a|
#                   c|
#                   k|
#
#  Each wagon to move from the Exp-End to the RPN-End it has to make
#  a stop first in the Stack-End. Before entering in the stack, the
#  priority-IN will be checked against the objects in the stack.
#  All top-most objects currently present in the stack with
#  higher priority-OUT will be transferred from the stack to the
#  RPN-End. Apart from the opening parenthesis ( which is discarded.
#
#  Example:
#  (1)                         A+B|C (1)
#  (2) A                        +B|C (2)
#  (3) A                         B|C (3)
#  (4) A                          |C (4)
#  (5) A B +                       C (5)
#  (6) A B + C
#  (7) A B + C |
#      ------------.   .------------
#      RPN-End      \ /      Exp-End
#                    |
#                   S|
#                   t|
#                   a|
#                   c|     B   C
#                   k| A + + | | |
#                      1 2 3 4 5 6 7
# ----------------------------------------------------------------------
def exp2rpn(expr):
	"""Replace expression list expr to rpn format"""
	#              Operator In Out
	priorities = {	"(" : [ 99,  0],
			"|" : [  1,  2],
			":" : [  1,  2],
			"+" : [  3,  4],
			"-" : [  3,  4],
			")" : [  0, 99],
			"V" : [101,100]}	# variable
	stack = []
	newproduct = True
	prevop = 0				# previous token input=identifier
	i = m = 0
	while i<len(expr):
		token = expr[i]
		# Check for special leading chars
		if newproduct and token in "+|:":
			newproduct = token in "|:"
			i += 1
			continue

		# Find priorities
		ip,op = priorities.get(token, [101,100])	# priorities["V"]

		if newproduct and token in "-":
			expr.insert(i,'*')	# insert space in ith position
			token = '@'      # Universe
		elif (token=="(" or op==100) and prevop >= 99:	# ') (' or 'V ('
			# insert a + before parenthesis if needed in the case of mcnp
			expr.insert(i,'+')
			prevop = 4
			continue

		newproduct = token in "(|:"

		# Remove from the stack everything with higher priority
		while stack and ip<stack[-1][1]:
			if stack[-1][0] != '(':
				expr[m] = stack.pop()[0]
				m += 1
			else:
				stack.pop()

		# Push it into the stack
		if token!=')':
			stack.append([token, op])
		else:
			# Should be an opening parenthesis
			if not stack:
				raise CSGException("Unbalanced parenthesis")
			stack.pop()

		i += 1
		prevop = op	# remember previous token
		#print("RPN=",expr[0:m])
		#print("STK=",stack)

	# Empty Stack
	while stack:
		if stack[-1][0]=='(':
			raise CSGException("Unbalanced parenthesis")

		if m>=len(expr): expr.append("")
		expr[m] = stack.pop()[0]
		m += 1

	# Delete unwanted items
	for i in range(len(expr)-1,m-1,-1):
		del expr[i]

# ----------------------------------------------------------------------
# Subroutine:	rpnorm
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20/4/2004
#
# Normalize a CG expression given in Reverse Polish Notation.
# Normalized CG expression is an expression given as sum (Boolean OR) of
# products (Boolean intersection or subtraction).
# The normalization (expansion of parenthesis and operator priorities)
# should be performed by recursively calling the RPNRULE subroutine.
# Since Fortran-77 doesn't have recursion, call the RPNRULE for every
# operator starting from the right-most one, until no rule is found.
# ----------------------------------------------------------------------
def rpnorm(expr):
	# Loop until there is no any extra change needed

	# Scan to find the first operators
	changed = 1
	while changed:
		changed = 0
		i = len(expr)-1
		while i>=4:
			tx = expr[i]
			if tx in ('+', '-', '|'):
				l = len(expr)
				rule = _rpnrule(expr,i)
				if rule>0:
					changed = 1
					i += len(expr)-l+1
			i -= 1
			if len(expr) > MAXEXPR:
				raise CSGException("Expansion failed. Too many terms")

# ----------------------------------------------------------------------
# Subroutine:	_subTerms
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20/4/2004
#
# This routine returns the pointers in the RPN terms array of the
# starting point of the left sub-expression LOWLEFT and right
# sub-expression LOWRIGHT given the pointer of the expr-operator NTX.
# The searching is performed by scanning from right to left the number
# of operators and objects pushed into the stack
#
# EXP          (expr-left) op (expr-right)
#
# RPN ...... | expr-left | expr-right | op | .......
#           Lowleft      LowRight  op-1 ntx
# ----------------------------------------------------------------------
def _subTerms(expr,n):

	nop = 0
	lowRight = 0
	lowLeft  = 0

	while n>=0:
		t = expr[n]
		if t in "+-#|:":
			nop += 1
		else:
			nop -= 1

		n -= 1
		if nop==0:
			if lowRight==0:
				lowRight = n+1
				nop += 1
				continue
			lowLeft = n+1
			return lowLeft,lowRight
	return lowLeft,lowRight

# ----------------------------------------------------------------------
# Subroutine:	rpnrule
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20/4/2004
#
# Find a matching rule and apply it on the sub-expression starting from
# the N position in the Reverse Polish Notation
#
# An expression is in normal form when all the parenthesis are expanded
# and the expression is described as a sum (UNIONS) of products
# (INTERSECTIONS and/or SUBTRACTIONS)
#
# An expression can be converted to normal form by repeatedly applying
# the following set of production rules to the expression and then to its
# sub-expressions:
#
#    Normal Form                        Reverse Polish Notation
#  1. X-(Y|Z) -> (X-Y)-Z                X Y Z | -  ->  X Y - Z -
#  2. X+(Y|Z) -> (X+Y)|(X+Z)            X Y Z | +  ->  X Y + X Z + |
#  3. X-(Y+Z) -> (X-Y)|(X-Z)            X Y Z + -  ->  X Y - X Z - |
#  4. X+(Y+Z) -> (X+Y)+Z                X Y Z + +  ->  X Y + Z +
#  5. X-(Y-Z) -> (X-Y)|(X+Z)            X Y Z - -  ->  X Y - X Z + |
#  6. X+(Y-Z) -> (X+Y)-Z                X Y Z - +  ->  X Y + Z -
#  7. X|(Y|Z) -> (X|Y)|Z                X Y Z | |  ->  X Y | Z |
#  8. (X-Y)+Z -> (X+Z)-Y                X Y - Z +  ->  X Z + Y -
#  9. (X|Y)-Z -> (X-Z)|(Y-Z)            X Y | Z -  ->  X Z - Y Z - |
# 10. (X|Y)+Z -> (X+Z)|(Y+Z)            X Y | Z +  ->  X Z + Y Z + |
# X,Y, and Z here match both primitives or sub-expressions.
# ----------------------------------------------------------------------
def _rpnrule(expr,n):
	# Reset rule
	rule = 0

	# Top-most operator
	op  = expr[n]
	if op not in "+-|": return rule

	# Right operator
	rop = expr[n-1]

	# Find left and right sub-trees
	ll,lr = _subTerms(expr,n)

	# Left operator
	lop = ' '
	if lr>0: lop = expr[lr-1]

	# Find Rule
	if   op=="-" and rop=="|":
		rule = 1
	elif op=="+" and rop=="|":
		rule = 2
	elif op=="-" and rop=="+":
		rule = 3
	elif op=="+" and rop=="+":
		rule = 4
	elif op=="-" and rop=="-":
		rule = 5
	elif op=="+" and rop=="-":
		rule = 6
	elif op=="|" and rop=="|":
		rule = 7
	elif op=="+" and lop=="-":
		rule = 8
	elif op=="-" and lop=="|":
		rule = 9
	elif op=="+" and lop=="|":
		rule = 10
	else:
		return rule

	# Find sub expressions X Y Z
	if rule<=7:    # X op (Y rop Z)
		Xu = lr-1
		Xl = ll

		ll,lr = _subTerms(expr,n-1)
		#Yu = lr-1
		#Yl = ll
		Zu = n-2
		Zl = lr
	else:		    # (X lop Y) op Z
		Zu = n-1
		Zl = lr
		L  = lr-1
		ll,lr = _subTerms(expr,L)
		Xu = lr-1
		Xl = ll
		#Yu = L-1
		#Yl = lr

	#print("Rule: ",rule)
	#print("X[",Xl,Xu,"]=",expr[Xl:Xu+1])
	#print("Y[",Yl,Yu,"]=",expr[Yl:Yu+1])
	#print("Z[",Zl,Zu,"]=",expr[Zl:Zu+1])

	# Expand the rule

	# 1. X-(Y|Z) -> (X-Y)-Z	 X Y Z | -  ->  X Y - Z -
	if    rule==1:
		# Leave X Y
		# Insert a - operator after Y
		expr.insert(Zl,"-")
		# Chop length by 1
		del expr[Zu+2]
		# Change the last operator to -
		expr[Zu+2] = '-'

	# 2. X+(Y|Z) -> (X+Y)|(X+Z)     X Y Z + |  ->  X Y + X Z + |
	elif rule==2:
		# Leave X Y
		# Insert a + operator after Y
		expr.insert(Zl,"+")
		# Copy X after the + operator
		to = Zl+1
		expr[to:to] = expr[Xl:Xu+1]
		Zu += Xu-Xl+2
		# Change last 2 operators to + |
		expr[Zu+1] = '+'
		expr[Zu+2] = '|'

	# 3. X-(Y+Z) -> (X-Y)|(X-Z)     X Y Z + -  ->  X Y - X Z - |
	elif rule==3:
		# Leave X Y
		# Insert a - operator after Y
		expr.insert(Zl,"-")
		# Copy X after the - operator
		to = Zl+1
		expr[to:to] = expr[Xl:Xu+1]
		Zu += Xu-Xl+2
		# Change last 2 operators to - |
		expr[Zu+1] = '-'
		expr[Zu+2] = '|'

	# 4. X+(Y+Z) -> (X+Y)+Z	 X Y Z + +  ->  X Y + Z +
	elif rule==4:
		# Leave X Y
		# Insert a + operator after Y
		expr.insert(Zl,"+")
		# Chop length by 1
		del expr[Zu+2]
		# Change the last operator to +
		expr[Zu+2] = '+'

	# 5. X-(Y-Z) -> (X-Y)|(X+Z)     X Y Z - -  ->  X Y - X Z + |
	elif rule==5:
		# Leave X Y
		# Insert a - operator after Y
		expr.insert(Zl,"-")
		# Copy X after the - operator
		to = Zl+1
		expr[to:to] = expr[Xl:Xu+1]
		Zu += Xu-Xl+2
		# Change last 2 operators to + |
		expr[Zu+1] = '+'
		expr[Zu+2] = '|'

	# 6. X+(Y-Z) -> (X+Y)-Z	 X Y Z - +  ->  X Y + Z -
	elif rule==6:
		# Leave X Y
		# Insert a + operator after Y
		expr.insert(Zl,"+")
		# Chop length by 1
		del expr[Zu+2]
		# Change the last operator to -
		expr[Zu+2] = '-'

	# 7. X|(Y|Z) -> (X|Y)|Z	 X Y Z | |  ->  X Y | Z |
	elif rule==7:
		# Leave X Y
		# Insert a | operator after Y
		expr.insert(Zl,"|")
		# Chop length by 1
		del expr[Zu+2]
		# Change the last operator to |
		expr[Zu+2] = '|'

	# 8. (X-Y)+Z -> (X+Z)-Y	 X Y - Z +  ->  X Z + Y -
	elif rule==8:
		# Leave X
		# Copy "Z +" after X
		L = Zu-Zl+2
		to = Xu+1
		expr[to:to] = expr[Zl:Zl+L]
		# Delete old "Z +"
		del expr[Zl+L:Zl+L+L]

	# 9. (X|Y)-Z -> (X-Z)|(Y-Z)     X Y | Z -  ->  X Z - Y Z - |
	#10. (X|Y)+Z -> (X+Z)|(Y+Z)     X Y | Z +  ->  X Z + Y Z + |
	elif rule in (9,10):
		# Leave X
		# Copy "Z -" or "Z +" after X
		L = Zu-Zl+2
		to = Xu+1
		expr[to:to] = expr[Zl:Zu+2]
		# Correct Z position
		Zl += L
		Zu += L
		# Delete the | infront of Z
		del expr[Zl-1]
		# Add | at the end
		expr.insert(Zu+1,"|")

	return rule

# ----------------------------------------------------------------------
# Convert rpn(postfix) expression to normal(infix) notation
# ----------------------------------------------------------------------
def rpn2exp(expr):
	stack = []

	#-----------------------------------------------------------------------
	# Add expression with parenthesis around
	#-----------------------------------------------------------------------
	def addWithParenthesis(aob, a):
		aob.append("(")
		aob.extend(a)
		aob.append(")")

	#-----------------------------------------------------------------------
	# check for operator in top level only
	#-----------------------------------------------------------------------
	def needParenthesis(expr):
		depth = 0
		for t in expr:
			if t=="(":
				depth += 1
			elif t==")":
				depth -= 1
			elif t in "|:":
				if depth==0: return True
				#return depth==0
		return False
	#-----------------------------------------------------------------------

	for token in expr:
		if token in ":|+-":
			b = stack.pop()
			a = stack.pop()
			aob = []
			if len(a)==1:
				if a[0]!='@':
					aob.extend(a)
			elif token in "|:":
				aob.extend(a)
			elif needParenthesis(a):
				addWithParenthesis(aob,a)
			else:
				aob.extend(a)
			aob.append(token)
			if len(b)==1:
				aob.extend(b)
			elif token in "|:":
				aob.extend(b)
			elif token == "-":
				addWithParenthesis(aob,b)
			elif needParenthesis(b) or b[0]=='-':
				addWithParenthesis(aob,b)
			else:
				aob.extend(b)
			stack.append(aob)
		else:
			stack.append([token])
	return stack.pop()

# ----------------------------------------------------------------------
# Subroutine:	rpn2exp
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20/4/2004
#
# Convert a NORMALIZED Reverse Polish notation to a standard expression
# WARNING: The routine expects an expression where for each UNION
# operator the right-sub-expression is a product while the left can be
# UNION or a product
# ----------------------------------------------------------------------
def rpnorm2exp(rpn):
	zones = []
	plus  = []
	minus = []

	i=0
	nstack = 0
	endprod = 0
	#print("RPN=",rpn)
	#print("LEN=",len(rpn))
	while i<len(rpn):
		tx = rpn[i]
		if tx in "+-|":
			nstack -= 1
		else:
			# First term is always a plus
			# .. peek then next operator to check for sign
			if not plus or rpn[i+1]=='+':
				plus.append(tx)
				lastPlus = 1
			elif rpn[i+1]=='-':
				minus.append(tx)
				lastPlus = 0
			nstack += 1

		if nstack==0:
			endprod = 1
		elif nstack==3:
			if lastPlus:
				del plus[-1]
			else:
				del minus[-1]
			i -= 2
			endprod = 1
		elif tx=='|':
			i -= 2
			endprod = 1
		elif i==len(rpn)-1:
			#print(*,"END i=ntx")
			endprod = 1

		if endprod:
			optZone(plus,minus)
			# remove None terms
			plus  = [x for x in plus  if x]
			minus = [x for x in minus if x]

			# Perform a bubble sort on product terms
			plus.sort()
			minus.sort()

			if plus or minus:
				zones.append((plus,minus))
			plus    = []
			minus   = []
			nstack  = 0
			endprod = 0
		i += 1

	# Remove duplicates of products
	rmDoubles(zones)

	# Reconstruct expression
	expr = []
	for plus,minus in zones:
		if expr and expr[-1]!="|": expr.append("|")
		# Fill the new array
		for j in plus:
			expr.append("+")
			expr.append(j)
		for j in minus:
			expr.append("-")
			expr.append(j)

	return expr

# ----------------------------------------------------------------------
# Subroutine:	optZone
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20/4/2004
#
# Optimize a product, by removing the universe @, duplicated terms like
# A+A, A-A, -A-A, and finally calling the geometrical optimizations and
# sorting by name the primitives in both the plus and minus terms.
#
# The product is described by 2 arrays the PLUS,NPLUS and MINUS,NMINUS
# with all the plus and minus terms of the product
#
# WARNING: It doesn't delete the term from the arrays but changes the
# name to space.
# ----------------------------------------------------------------------
def optZone(plus,minus):
	# Remove Universe @ from PLUS
	# and
	# remove duplicate terms A+A=A, A-A={}
	for i,pi in enumerate(plus):
		if pi=='@':
			plus[i] = None
		elif pi is not None:
			for j in range(i+1,len(plus)):
				if pi==plus[j]: plus[j]=None
			for m in minus:
				if pi==m:
					# Remove everything
					del plus[:]
					del minus[:]
					return

	# Discard product if universe @ exists in MINUS
	# Check for duplicates in MINUS like -A-A=-A
	for i,mi in enumerate(minus):
		if mi=='@':
			# Remove everything
			del plus[:]
			del minus[:]
			return
		if mi is not None:
			for j in range(i+1,len(minus)):
				if mi==minus[j]: minus[j]=None

	# Perform the Geometrical optimization in the product
	#call OptGeo(nplus,plus,nminus,minus)

# ----------------------------------------------------------------------
# Subroutine:	rmDoubles
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   20/4/2004
#
# Remove duplicates of products A+B|A+B|C+D  ->  A+B|C+D
# ----------------------------------------------------------------------
def rmDoubles(zones):
	i = -1
	while i<len(zones)-1:
		i += 1
		plus1, minus1 = zones[i]
		j = i
		while j<len(zones)-1:
			j += 1
			plus2, minus2 = zones[j]

			# Check if they are the same
			if len(plus1)!=len(plus2) or len(minus1)!=len(minus2): continue
			diffplus = -1
			for k,p1 in enumerate(plus1):
				if p1 != plus2[k]:
					diffplus = k
					break
			#if not same: continue

			diffminus = -1
			for k,m1 in enumerate(minus1):
				if m1 != minus2[k]:
					diffminus = k
					break
			#if not same: continue

			if diffplus==-1 and diffminus==-1:
				del zones[j]
				j -= 1
				continue

# ----------------------------------------------------------------------
# Subroutine:	split
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   18/5/2004
#
# Break products into lists
# ----------------------------------------------------------------------
def split(expr):
	brk=[]
	expr=expr[:]	# Make a copy of the expression
	while 1:
		try:
			p = expr.index("|")
			brk.append(expr[0:p])
			del expr[0:p+1]
		except:
			brk.append(expr)
			return brk

#-------------------------------------------------------------------------------
if __name__=="__main__":
	import sys
	from pprint import pprint

	exp = tokenize(" ".join(sys.argv[1:]))
	print("Tokens=")
	pprint(exp)

	print("Zones=")
	pprint(splitZones(exp))

	print("ZoneString=")
	pprint([toString(0, NORMAL,x) for x in splitZones(exp)])

	print("ORG=",exp)

	exp2rpn(exp)
	print("RPN=",exp)
	print("check=",check(exp))
	sys.exit(0)

	#exp = rpn2exp(exp)
	#print("NORM=",exp)
	#print("EXP=","".join(exp))

	rpnorm(exp)
	print("RPNORM=",exp)

	exp = rpn2exp(exp)
	print("NORM=",exp)

#	list = breakprod(exp)
	print("EXP=",toString(0, NORMAL,exp))
