# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   11-Jun-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import configparser
from legacy import cmp #, cmp_to_key

#-------------------------------------------------------------------------------
def _cmpA(x, y):
	x = x[0]
	y = y[0]
	try:
		ix = int(x)
		nx = ""
	except:
		ix = int(x[:-1])
		nx = x[-1:]
	try:
		iy = int(y)
		ny = ""
	except:
		iy = int(y[:-1])
		ny = y[-1:]

	c = cmp(ix, iy)
	if c == 0:
		return cmp(nx, ny)
	return c

# ----------------------------------------------------------------------
class Isotope:
	def __init__(self, symbol, Z, A):
		self.Z      = Z
		self.symbol = symbol
		self.A      = A
		self.spin   = ""
		self.massex = ""
		self.decayt = ""
		self.decayp = ""
		self.Am     = ""

#===============================================================================
# Element Class
#===============================================================================
class Element:
	elements = []		# database of elements sorted by Z-1
	footnote = {}

	# ----------------------------------------------------------------------
	def __init__(self, symbol, Z):
		self.Z          = Z
		self.symbol     = symbol
		self.name       = ""
		self.AmassStr   = ""
		self.densityStr = ""
		self.melting    = ""
		self.boiling    = ""
		self.oxidation  = ""
		self.isotope    = {}

	# ----------------------------------------------------------------------
	def inputName(self, mode=None):
		if mode == "fluka":
			ename = self.name.upper()[:8]
			if ename=="PHOSPHOR": ename = "PHOSPHO"
			return ename
		return self.name

	# ----------------------------------------------------------------------
	def Amass(self):
		return float(self.AmassStr.split()[0])

	# ----------------------------------------------------------------------
	def density(self):
		return float(self.densityStr.split()[0])

	# ----------------------------------------------------------------------
	def densityNote(self):
		d = self.densityStr.split()
		if d: return d[-1]
		return None

	#-------------------------------------------------------------------------------
	@staticmethod
	def byZ(Z):
		try:
			return Element.elements[Z-1]
		except IndexError:
			return None

	#-------------------------------------------------------------------------------
	@staticmethod
	def bySymbol(symbol):
		symbol = symbol.upper()
		for element in Element.elements:
			if element and element.symbol.upper() == symbol:
				return element
		return None

	#-------------------------------------------------------------------------------
	@staticmethod
	def byName(name):
		name = name.upper()
		for element in Element.elements:
			if element and element.name.upper() == name:
				return element
		return None

	#-------------------------------------------------------------------------------
	@staticmethod
	def isotopes(Z):
		elem = Element.elements[Z-1]
		eldic = elem.isotope.copy()
		return eldic

	#-------------------------------------------------------------------------------
	@staticmethod
	def iso(Z, N):
		elem = Element.elements[Z-1]
		return elem.isotope.get(str(N))

	# ----------------------------------------------------------------------
	@staticmethod
	def init(filename):
		c = configparser.RawConfigParser()
		c.read(filename, encoding="UTF-8")

		# Read Elements
		Element.elements = [None] * len(c.sections())
		for symb in c.sections():
			if symb[0].islower(): continue
			Z         = c.getint(symb, "Z")
			elem      = Element(symb, Z)
			elem.name = c.get(symb, "name")
			try:	elem.AmassStr   = c.get(symb, "Amass")
			except: pass
			try:	elem.densityStr = c.get(symb, "density")
			except: pass
			try:	elem.melting = c.get(symb, "melting")
			except: pass
			try:	elem.boiling = c.get(symb, "boiling")
			except: pass
			try:	elem.oxidation = c.get(symb, "oxidation")
			except: pass

			elem.isotope = {}
			for item,value in c.items(symb):
				if item.startswith("isotope."):
					A     = item.split(".")[1]
					iso   = Isotope(symb, Z, A)
					words = value.split(";")
					iso.spin   = words[0].strip()
					iso.massex = words[1].strip()
					iso.decayt = words[2].strip()
					iso.Am     = words[3].strip()
					iso.decayp = words[4].strip()
					elem.isotope[A] = iso
			# elem.isotope.sort(key=cmp_to_key(_cmpA))
			Element.elements[Z-1] = elem

		# Remove all None parts (Do I need that?)
		#try:
		#	while True: Element.elements.remove(None)
		#except ValueError:
		#	pass

		for (i,n) in c.items("footnotes"):
			Element.footnote[i] = n #.replace("\n"," ")

		del c
