#!/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   06-Dec-2019

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re

#===============================================================================
# Basic element of Markdown syntax
#===============================================================================
class MDElement():
	"""
	Basic markdown syntax element
	it holds a list of items and options
	"""
	#-----------------------------------------------------------------------
	def __init__(self, items, opt=None):
		self.items = items
		self.opt   = opt

	#-----------------------------------------------------------------------
	def __repr__(self):
		items = "[..]" if isinstance(self.items, list) else repr(self.items)
		if self.opt is not None:
			return f"{self.__class__.__name__}({self.opt!r}, {items})"
		else:
			return f"{self.__class__.__name__}({items})"

	#-----------------------------------------------------------------------
	def __str__(self):
		out = []
		def tostr(data, prefix=0):
			nonlocal out
			out.append((" "*prefix)+repr(data))
			if isinstance(data.items,list):
				for item in data.items:
					tostr(item, prefix+3)
				out.append("")
		tostr(self)
		return "\n".join(out)

#===============================================================================
# Definition of markdown elements
#===============================================================================
class Block(MDElement):		pass
class Paragraph(MDElement):	pass
class List(MDElement):		pass
class Text(MDElement):		pass
class Emphasis(MDElement):	pass
class Underline(MDElement):	pass
class Bold(MDElement):		pass
class Overstrike(MDElement):	pass
class Header(MDElement):	pass
class Link(MDElement):		pass
class Image(MDElement):		pass
class Code(MDElement):		pass

#===============================================================================
# Mark down document is the base class holding a collection of mark down
# elements and some basic parsing static methods
#===============================================================================
class Document:
	#-----------------------------------------------------------------------
	def __init__(self, text=None):
		self.text = None
		self.data = None
		if text is not None:
			self.parse(text)

	#-----------------------------------------------------------------------
	def __str__(self):
		return str(self.data)

	#-----------------------------------------------------------------------
	# Parse text to markdown elements
	#-----------------------------------------------------------------------
	def parse(self, text):
		self.text = text
		self.data = None

#===============================================================================
# Parses Markdown strings and return a type and items
#===============================================================================
class Markdown(Document):
	"""
	First split strings into blocks as paragraphs. For each paragraph,
	parse it when it is a List item, a Header, an Emphasis, a Bold, or just simple text.

	Usage:

	>>> Markdown.parse('This is a paragraph.')
	MDElement([Paragraph([Text('This is a paragraph.')])])

	#>>> Markdown.parse('The first paragraph.\\n\\nThe second paragraph.')
	#MDElement([Paragraph([Text('The first paragraph.')]), Paragraph([Text('The second paragraph.')])])

	#>>> Markdown.parse('Emphasis *is* supported.')
	#MDElement([Paragraph([Text('Emphasis '), Emphasis('is'), Text(' supported.')])])

	#>>> Markdown.parse('Bold **is** supported.')
	#MDElement([Paragraph([Text('Bold '), Bold('is'), Text(' supported.')])])

	#>>> Markdown.parse('- a list')
	#MDElement([List([Paragraph([Text('a list')])])])

	#>>> Markdown.parse('- the first item\\n- the second item')
	#MDElement([List([Paragraph([Text('the first item')]), Paragraph([Text('the second item')])])])

	#>>> Markdown.parse('- Emphasis *is* supported.')
	#MDElement([List([Paragraph([Text('Emphasis '), Emphasis('is'), Text(' supported.')])])])

	#>>> Markdown.parse('* the first item\\n* the second item')
	#MDElement([List([Paragraph([Text('the first item')]), Paragraph([Text('the second item')])])])

	#>>> Markdown.parse('- item1\\n- item2\\n - item2.1\\n - item2.2\\n- item3')
	#MDElement([List([Paragraph([Text('item1')]), Paragraph([Paragraph([Text('item2')]), List([Paragraph([Text('item2.1')]), Paragraph([Text('item2.2')])])]), Paragraph([Text('item3')])])])

	#>>> Markdown.parse('#######this is header')
	#MDElement([Paragraph([Header(6,[Text('#this is header')])])])

	#>>> Markdown.parse('#######this is *header* and **header**')
	#MDElement([Paragraph([Header(6,[Text('#this is '), Emphasis('header'), Text(' and '), Bold('header'), Text('')])])])

	#>>> Markdown.parse('#header1\\n##header2\\n\\nanother paragraph')
	#MDElement([Paragraph([Header(1,[Text('header1')]), Header(2,[Text('header2')])]), Paragraph([Text('another paragraph')])])

	#>>> Markdown.parse('hello ##this is not *header* and **header**')
	#MDElement([Paragraph([Text('hello ##this is not '), Emphasis('header'), Text(' and '), Bold('header'), Text('')])])
	"""

	BLOCK_PAT    = re.compile(r"\n{2,}")
	LIST_PAT     = re.compile(r"^[-|\*]\s+", re.M)
	SUBLIST_PAT  = re.compile(r"\s+[-|\*]\s+", re.M)
	HEADER_SPLIT = re.compile(r"(^#{1,6}\s+.*$)", re.M)
	HEADER_PAT   = re.compile(r"^(#{1,6})\s+(.*)$")
	LINK_SPLIT   = re.compile(r"(!?\[.*?\]\(.*?\))", re.M)
	PART_SPLIT   = re.compile(r"(\*\*[^*]*\*\*|\*[^\*]*\*|__[^_]*__|_[^_]*_|~~[^~]*~~)")
	INLINE_ELEMENTS = [
			(re.compile(r'\*\*([^ *][^*]*)\*\*'), Bold),
			(re.compile(r'__([^ _][^_]*)__'),     Underline),
			(re.compile(r'\*([^ *][^*]*)\*'),     Emphasis),
			(re.compile(r'_([^ _][^_]*)_'),       Emphasis),
			(re.compile(r'~~([^ ~][^~]*)~~'),     Overstrike)
		]
	LINK_PAT    = re.compile(r"(!)?\[(.*?)\]\((.*?)\)")

	#-----------------------------------------------------------------------
	# Parse text to markdown elements
	#-----------------------------------------------------------------------
	def parse(self, text):
		super().parse(text)
		self.data = MDElement([Markdown.parse_block(block)
				for block in Markdown.BLOCK_PAT.split(self.text)])

	#-----------------------------------------------------------------------
	# Split block to headers and paragraphs
	#-----------------------------------------------------------------------
	@staticmethod
	def parse_block(block):
		items = []
		for item in Markdown.HEADER_SPLIT.split(block):
			if not item: continue
			mat = Markdown.HEADER_PAT.match(item)
			if mat is not None:
				if Markdown.PART_SPLIT.search(mat.group(2)):
					title = [Markdown.parse_formatting(x)
							for x in Markdown.PART_SPLIT.split(mat.group(2))
							if x]
				else:
					title = mat.group(2)
				items.append(Header(title, len(mat.group(1))))
			else:
				items.append(Markdown.parse_paragraph(item.strip()))
		return Block(items)

	#-----------------------------------------------------------------------
	@staticmethod
	def parse_paragraph(text):
		# List items begins with ' - ' or ' * '
		mat = Markdown.LIST_PAT.match(text)
		if mat is not None:
			# '^' should match at the beginning of the string
			# and at the beginning of each line
			items = [item.strip()
				for item in re.split(Markdown.LIST_PAT, text)[1:]]
			lists = List([Markdown.parse_paragraph(x) for x in items if x])
			for item in lists.items:
				mat = Markdown.SUBLIST_PAT.search(item.items[0].items)
				if mat is not None:
					inner_items = [x.strip()
							for x in Markdown.SUBLIST_PAT.split(
									item.items[0].items)]
					item.items = []
					item.items.append(Markdown.parse_paragraph(inner_items[0]))
					inner_lists = List([Markdown.parse_paragraph(x)
								for x in inner_items[1:]])
					item.items.append(inner_lists)
			return lists
		return Markdown.parse_links(text)

	#-----------------------------------------------------------------------
	# split out Emphasis, Bold, Overstrike
	#-----------------------------------------------------------------------
	@staticmethod
	def parse_links(block):
		items = []
		for item in Markdown.LINK_SPLIT.split(block):
			mat = Markdown.LINK_PAT.match(item)
			if mat is not None:
				if Markdown.PART_SPLIT.search(mat.group(2)):
					title = [Markdown.parse_formatting(x)
							for x in Markdown.PART_SPLIT.split(mat.group(2))
								if x]
				else:
					title = mat.group(2)

				if mat.group(1) is not None:
					items.append(Image(title, mat.group(3)))
				else:
					items.append(Link(title, mat.group(3)))
			else:
				items.extend([Markdown.parse_formatting(x)
					for x in Markdown.PART_SPLIT.split(item) if x])
		return Paragraph(items)

	#-----------------------------------------------------------------------
	# Parse inline elements
	# Emphasis Bold, Overstrike, Underline, ...
	# images, links
	#-----------------------------------------------------------------------
	@staticmethod
	def parse_formatting(text):
		for regexp, klass in Markdown.INLINE_ELEMENTS:
			m = regexp.match(text)
			if m is not None:
				return klass(m.group(1))
		return Text(text)

#===============================================================================
# Parses old flair Rich Text format
#===============================================================================
class RTF(Document):
	TAGPAT = re.compile(r"{(\w+):\s*(.*?)}", re.MULTILINE|re.DOTALL)
	TAG2MD = { "h1" : ("#",   True),	# new tag, line (True), enclose(False)
		   "h2" : ("##",  True),
		   "h3" : ("###", True),
		   "l"  : ("<<<", True),
		   "r"  : (">>>", True),
		   "c"  : ("^^^", True),
		   "i"  : ("_",   False),
		   "b"  : ("**",  False),
		   "bi" : ("_",   False),
		   "u"  : ("__",  False),
		   "o"  : ("~~",  False),
	}

	#-----------------------------------------------------------------------
	# Convert from old rich text to markdown
	#-----------------------------------------------------------------------
	def parse(self, text):
		super().parse(text)
		items = []
		last = 0
		for pat in re.finditer(RTF.TAGPAT, text):
			items.append(text[last:pat.start()])
			last = pat.end()
			tag = pat.group(1)
			value = pat.group(2)
			if tag == "img":
				fn = os.path.basename(value)
				items.append(f"![{fn}]({value})")
			elif tag == "link":
				items.append(f"[{value}]({value})")
			else:
				try:
					ntag,line = RTF.TAG2MD[tag]
					if line:
						items.append(f"{ntag} {value}")
					else:
						items.append(f"{ntag}{value}{ntag}")
				except KeyError:
					items.append(value)
		self.parse("".join(items))

#===============================================================================
# Parses Restructured text output
#===============================================================================
class RST(Document):
	HEADERS = [	re.compile(r'#{3,}'),		# h1
			re.compile(r'\*{3,}'),		# h2
			re.compile(r'={3,}'),		# h3
			re.compile(r'-{3,}'),		# h4
			re.compile(r'\^{3,}')]		# h5

	#-----------------------------------------------------------------------
	# Convert output of RST to markdown
	#-----------------------------------------------------------------------
	def parse(self, text):
		super().parse(text)

		prevline = ""

		self.data = []
		section = []
		for line in text.splitlines():
			line = line.rstrip()
			if not line:
				# starting a new section?
				if not prevline:
					section.append(Text("\n"))
				section.append(Markdown.parse_paragraph(prevline))
				prevline = ""

			if line[0] in "#*=-":
				for i,pat in enumerate(RST.HEADERS):
					if pat.match(line):
						if section: self.data.append(section)
						# parse possible formatting
						# use from Markdown
						if Markdown.PART_SPLIT.search(prevline):
							title = [Markdown.parse_formatting(x)
								for x in Markdown.PART_SPLIT.split(prevline)
								if x]
						else:
							title = prevline
						prevline = ""
						section = [ Header(title, i+1) ]
						break
				else:
					prevline += f"\n{line}"

			else:
				prevline += f"\n{line}"

#-------------------------------------------------------------------------------
if __name__ == "__main__":
	import sys
	#import doctest
	#doctest.testmod()
	text = """
#Header
##Header in **continuous** lines

#######Test header with 7 '#'

## header *emphasis* and **bold**

This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph.


*This is emphasis.*

- a*emphasis*
- b**bold**
- c

- a_emphasis_
- b__emphasis__
- c

- A_em*pha*sis_
- B__em*pha*sis__
- c

Another paragraph.

- item1
 - item1.1
 - item1.2
- item2
- item3

This is a paragraph. This is a paragraph. This is a paragraph. This is a paragraph. *This is a paragraph.* This is a paragraph. This is a paragraph. **This is a paragraph.** This is a paragraph. This is a paragraph.


- item1
- item2
 - item2.1
 - item2.2
- item3
- item4

* http://cern.ch/flair/flair.html
* http1://cern.ch/flair/flair.html
* [Link](http://cern.ch/flair/flair.html)
* ![Image](http://cern.ch/flair/images/flair.png)

<p>this is html</p>

<h2>this is html</h2>


<em>this is html</em>
"""
	md = Markdown()
	if len(sys.argv)>1:
		fn,ext = os.path.splitext(sys.argv[1])
		with open(sys.argv[1],"r") as f:
			text = f.read()
		md.parseRST(text)
		from pprint import pprint
		pprint(md.data)
	else:
		def testparse(text):
			md.parse(text)
			print("<<<", text)
			print(">>>", md.data)
		#print(md.parse(text))
		#print(md.parse("Line **bold1** normal __bold2__"))
		#print(md.parse("Line **bold1** normal __bold2__"))
		#print(md.parse("Line **bold** [Link text](http://cern.ch) *emphasis*"))
		#print(md.parse("Line **bold** ![Link *text*](http://cern.ch) *emphasis*"))
		testparse("# Header1\nLine **bold** ![Link *text*](http://cern.ch) *emphasis*")
		testparse("# Header1\n\nLine **bold** ![Link *text*](http://cern.ch) *emphasis*")
		testparse("# Header1\n- Item1\n- Item 2\n- Item 3")
		testparse("# Header1\nBlah blah *bold\n- Item1\n- Item 2\n- Item 3")
