#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   06-Jan-2021

import os
import glob
from math import sqrt, sin, cos, radians

# ------------------------- Palettes -------------------------------------------
# Standard fluka palette
_FLUKA_PALETTE = [
	(1.0, 1.0, 1.0),
	(0.8, 0.8, 0.8),
	(0.9, 0.6, 0.9),
	(1.0, 0.4, 1.0),
	(0.9, 0.0, 1.0),
	(0.7, 0.0, 1.0),
	(0.5, 0.0, 0.8),
	(0.0, 0.0, 0.8),
	(0.0, 0.0, 1.0),
	(0.0, 0.6, 1.0),
	(0.0, 0.8, 1.0),
	(0.0, 0.7, 0.5),
	(0.0, 0.9, 0.2),
	(0.5, 1.0, 0.0),
	(0.8, 1.0, 0.0),
	(1.0, 1.0, 0.0),
	(1.0, 0.8, 0.0),
	(1.0, 0.5, 0.0),
	(1.0, 0.0, 0.0),
	(0.8, 0.0, 0.0),
	(0.6, 0.0, 0.0),
	(0.4, 0.0, 0.0),
	(0.0, 0.0, 0.0) ]

# Standard palette without the final black color
_FLUKA2_PALETTE = _FLUKA_PALETTE[:]
del _FLUKA2_PALETTE[-1]

# Gray scale palette
_GRAY_PALETTE = [(0.0, 0.0, 0.0), (1.0, 1.0, 1.0)]

# Inverse gray scale palette
_INVGRAY_PALETTE = [(1.0, 1.0, 1.0), (0.0, 0.0, 0.0)]

# Difference Blue-White-Red
_DIFF_PALETTE =  [(0.0, 0.0, 1.0), (1.0,  1.0,  1.0), (1.0, 0.0, 0.0)]

# region plotting palette
_GEOMETRY_PALETTE = [
	(0.5, 0.5, 0.5), # grey
	(1.0, 0.0, 1.0), # magenta
	(0.0, 0.0, 1.0), # blue
	(0.0, 1.0, 0.0), # green
	(1.0, 1.0, 0.0), # yellow
	(1.0, 0.0, 0.0), # red
	(1.0, 1.0, 1.0), # white
	]

# ------------------------------------------------------------------------------
def RGB(r,g,b): return r<<16 | g<<8 | b
#def RGB2str(r,g,b): return f"#{r:02X}{g:02X}{b:02X}"
def RGB2str(rgb): return f"#{rgb:06X}"

# ------------------------------------------------------------------------------
# Return a pseudo-random color based on 'x' as seed
# ------------------------------------------------------------------------------
def rndColor(x):
	def rnd(zw):
		w = zw & 0xffff
		z = (zw >> 16) & 0xffff
		z = 36969 * (z & 0xffff) + (z >> 16)
		w = 18000 * (w & 0xffff) + (w >> 16)
		return (z << 16) + (w & 0xffff)

	x = rnd(x)
	R = (x % 224) + 16
	x = rnd(x)
	G = (x % 224) + 16
	x = rnd(x)
	B = (x % 224) + 16
	return R<<16 | G<<8 | B

#===============================================================================
# Color Palette class
#===============================================================================
class Palette:
	# RGB formulae as in gnuplot
	_rgbformulae = [
			lambda x: 0,				# 0
			lambda x: 0.5,				# 1
			lambda x: 1,				# 2
			lambda x: x,				# 3
			lambda x: x**2,				# 4
			lambda x: x**3,				# 5
			lambda x: x**4,				# 6
			lambda x: sqrt(x),			# 7
			lambda x: sqrt(sqrt(x)),		# 8
			lambda x: sin(radians(90*x)),		# 9
			lambda x: cos(radians(90*x)),		# 10
			lambda x: abs(x-0.5),			# 11
			lambda x: (2*x-1)**2,			# 12
			lambda x: sin(radians(180*x)),		# 13
			lambda x: abs(cos(radians(180*x))),	# 14
			lambda x: sin(radians(360*x)),		# 15
			lambda x: cos(radians(360*x)),		# 16
			lambda x: abs(sin(radians(360*x))),	# 17
			lambda x: abs(cos(radians(360*x))),	# 18
			lambda x: abs(sin(radians(720*x))),	# 19
			lambda x: abs(cos(radians(720*x))),	# 20
			lambda x: 3*x,				# 21
			lambda x: 3*x-1,			# 22
			lambda x: 3*x-2,			# 23
			lambda x: abs(3*x-1),			# 24
			lambda x: abs(3*x-2),			# 25
			lambda x: (3*x-1)/2,			# 26
			lambda x: (3*x-2)/2,			# 27
			lambda x: abs((3*x-1)/2),		# 28
			lambda x: abs((3*x-2)/2),		# 29
			lambda x: 0 if x<=0.25 else (1 if x>=0.57 else x/0.32-0.78125),		# 30
			lambda x: 0 if x<=0.42 else (1 if x>=0.92 else 2*x-0.84),		# 31
			lambda x: x*4 if x<=0.42 else (-2*x+1.84 if x<=0.92 else x/0.08-11.5),	# 32
			lambda x: abs(2*x - 0.5),		# 33
			lambda x: 2*x,				# 34
			lambda x: 2*x - 0.5,			# 35
			lambda x: 2*x - 1			# 36
		]

	#-----------------------------------------------------------------------
	def __init__(self, palette, ncolors=None):
		if isinstance(palette,tuple):
			self.rgb = palette
			self.palette = self.formulae(32 if ncolors is None else ncolors)

		elif isinstance(palette,str):
			self.rgb = None
			self.palette = paletteStore[palette][:]
		else:
			self.rgb = None
			self.palette = palette

		if ncolors is not None:
			self.palette = self.resize(ncolors)

	#-----------------------------------------------------------------------
	def __len__(self):
		return len(self.palette)

	#-----------------------------------------------------------------------
	def __getitem__(self, item):
		return self.palette[item]

	#-----------------------------------------------------------------------
	def clone(self):
		palette = Palette(self.palette[:])
		palette.rgb = self.rgb
		return palette

	#-----------------------------------------------------------------------
	# From rgb formulae
	#-----------------------------------------------------------------------
	def formulae(self, n):
		R = Palette._rgbformulae[abs(self.rgb[0])]
		G = Palette._rgbformulae[abs(self.rgb[1])]
		B = Palette._rgbformulae[abs(self.rgb[2])]
		iR = self.rgb[0]>=0
		iG = self.rgb[1]>=0
		iB = self.rgb[2]>=0
		palette = []
		for i in range(0,n+1):
			x = i/n
			r = min(1, max(0, R(x) if iR else R(1-x)))
			g = min(1, max(0, G(x) if iG else G(1-x)))
			b = min(1, max(0, B(x) if iB else B(1-x)))
			palette.append((r,g,b))
		return palette

	#-----------------------------------------------------------------------
	# return one formula color for 0<=x<=1
	#-----------------------------------------------------------------------
	def formulaColor(self, x):
		R = Palette._rgbformulae[abs(self.rgb[0])]
		G = Palette._rgbformulae[abs(self.rgb[1])]
		B = Palette._rgbformulae[abs(self.rgb[2])]
		iR = self.rgb[0]>=0
		iG = self.rgb[1]>=0
		iB = self.rgb[2]>=0
		r = min(1, max(0, R(x) if iR else R(1-x)))
		g = min(1, max(0, G(x) if iG else G(1-x)))
		b = min(1, max(0, B(x) if iB else B(1-x)))
		return r,g,b

	#-----------------------------------------------------------------------
	# Resize palette to n colors
	#-----------------------------------------------------------------------
	def resize(self, n):
		if n==len(self.palette): return
		n = max(n,2)
		if self.rgb:
			return self.formulae(n)
		else:
			def interpolate(a,x,b): return (b-a)*x+a
			npal    = len(self.palette)
			palstep = 1.0 / (npal-1)
			npal   -= 1
			step    = 1.0 / (n-1)
			palette = []
			for i in range(n):
				ic,rem = divmod(step*float(i), palstep)
				ic = int(ic)
				rem  /= palstep
				if ic<0:
					R,G,B = self.palette[0]
				elif ic>=npal:
					R,G,B = self.palette[-1]
				else:
					lo = self.palette[ic]
					hi = self.palette[ic+1]
					R = interpolate(lo[0], rem, hi[0])
					G = interpolate(lo[1], rem, hi[1])
					B = interpolate(lo[2], rem, hi[2])
				palette.append((R,G,B))
			return palette

	#-----------------------------------------------------------------------
	# Resize palette to n colors.
	# Warning: Convert palette to integers: RRGGBB
	#-----------------------------------------------------------------------
	def asRGB(self, n):
		newpalette = []
		n = max(n,2)
		if self.rgb:
			for i in range(n):
				r,g,b = self.formulaColor(i / (n-1))
				newpalette.append(int(r*255)<<16 | int(g*255)<<8 | int(b*255))
		else:
			def interpolate(a,x,b): return (b-a)*x+a
			npal       = len(self)
			palstep    = 1.0 / (npal-1)
			npal      -= 1
			step       = 1.0 / (n-1)
			for i in range(n):
				ic,rem = divmod(step*float(i), palstep)
				ic = int(ic)
				rem  /= palstep
				if ic<0:
					r,g,b = self.palette[0]
				elif ic>=npal:
					r,g,b = self.palette[-1]
				else:
					lo = self.palette[ic]
					hi = self.palette[ic+1]
					r = interpolate(lo[0], rem, hi[0])
					g = interpolate(lo[1], rem, hi[1])
					b = interpolate(lo[2], rem, hi[2])
				newpalette.append(int(r*255)<<16 | int(g*255)<<8 | int(b*255))
		return newpalette

	#-----------------------------------------------------------------------
	# Convert to dword 0x00RRGGBB
	#-----------------------------------------------------------------------
	def dword(self):
		return [(int(R*255)<<16 | int(G*255)<<8 | int(B*255))
				for R,G,B in self.palette]

	#-----------------------------------------------------------------------
	# @return a palette for tk as #RRGGBB
	#-----------------------------------------------------------------------
	def tk(self):
		return [f"#{int(R*255):02X}{int(G*255):02X}{int(B*255):02X}"
				for R,G,B in self.palette]

	#-----------------------------------------------------------------------
	# @return a PIL formatted palette
	#-----------------------------------------------------------------------
	def pil(self, smooth=False):
		pal = []
		if smooth:
			if len(self)!=256:
				palette = self.palette.clone()
				palette.palette = palette.resize(256)
			else:
				palette = self
			for R,G,B in palette:
				pal.append(int(R*255))
				pal.append(int(G*255))
				pal.append(int(B*255))
		else:
			repeat = 256/len(self)
			for i,(R,G,B) in enumerate(self.palette):
				R = int(R*255)
				G = int(G*255)
				B = int(B*255)
				n = 3*min(256,int(round(repeat*(i+1))))
				while len(pal) < n:
					pal.append(R)
					pal.append(G)
					pal.append(B)
			assert len(pal)==3*256
		return pal

	#-----------------------------------------------------------------------
	# @return a gnuplot string with the palette defined
	#-----------------------------------------------------------------------
	def gnuplot(self):
		pal = [f"{i}. {R} {G} {B}"
			for i,(R,G,B) in enumerate(self.palette,1)]
		return f"set palette defined ({','.join(pal)})"

#-------------------------------------------------------------------------------
# Initialize palette store
#-------------------------------------------------------------------------------
paletteStore = {
	""         : Palette(_FLUKA_PALETTE),		# default
	"diff"     : Palette(_DIFF_PALETTE),
	"fluka"    : Palette(_FLUKA_PALETTE),
	"fluka2"   : Palette(_FLUKA2_PALETTE),
	"gray"     : Palette(_GRAY_PALETTE),
	"invgray"  : Palette(_INVGRAY_PALETTE),
	"geometry" : Palette(_GEOMETRY_PALETTE),
	"gnuplot"  : Palette((7,5,15)),
	"gnuplot2" : Palette((30,31,32)),
	"rainbow"  : Palette((33,13,10)),
	"hot"      : Palette((21,22,23)),
	"afmhot"   : Palette((34,35,36)),
	"ocean"    : Palette((23,28,3)),
	"GnRdVi"   : Palette((3,11,6)),
}

#-------------------------------------------------------------------------------
# load palette from filename
#-------------------------------------------------------------------------------
def loadPalette(filename, name=None):
	if name is None:
		name,_ = os.path.splitext(os.path.basename(filename))

	with open(filename) as f:
		pal = []
		for line in f:
			line = line.strip()
			if not line or line[0]=='#': continue
			r,g,b = map(float,line.split())
			pal.append((r,g,b))

	if len(pal)==1:
		# assume an rgb formulae
		r,g,b = map(int, pal[0])
		palette = Palette((r,g,b))
	else:
		palette = Palette(pal)

	paletteStore[name] = palette

#-------------------------------------------------------------------------------
def load(dirs, ext=".pal"):
	"""Read all palettes from directories"""
	for folder in dirs:
		for fn in glob.glob(f"{folder}/*{ext}"):
			loadPalette(fn)
