# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

__author__  = "Vasilis Vlachoudis"
__email__   = "Vasilis.Vlachoudis@cern.ch"

import threading
import subprocess
from log import say

#===============================================================================
# Base Process
# The base class for all offering monitoring tools for the state of a system
# process e.g. Compile, Running, Data merging, etc.
#===============================================================================
class Process:
	# Constants defining the status
	STATUS_IDLE     = 0
	STATUS_WAITING  = 1
	STATUS_RUNNING  = 2
	STATUS_FINISHED = 3
	STATUS_ERROR    = 4
	STATUS_KILLED   = 5

	#           0        1           2          3          4         5
	STATUS = ["Idle", "Waiting", "Running", "Finished", "Error", "Killed"]

	#----------------------------------------------------------------------
	def __init__(self):
		self._mutex = threading.Semaphore()
		self.notify = None	# notify call on end FIXME maybe superclass as FlairProcess
		self.init()

	#----------------------------------------------------------------------
	def init(self):
		self._thread = None			# thread id
		self.rc      = 0			# return code
		self._out    = []			# output string buffer
		self.message = ("","")			# current message
		self.percent = 0			# percentage done
		self.status  = Process.STATUS_IDLE	# running status

	#----------------------------------------------------------------------
	def statusMessage(self):
		return Process.STATUS[self.status]

	#----------------------------------------------------------------------
	def start(self, *args, **kwargs):
		if self.status != Process.STATUS_IDLE:
			say("Process is not idle")
			return False
		if self._thread and self._thread.is_alive():
			say("Process is still alive")
			return
		self.init()
		self.rc      = None
		#self.percent = 0
		self.status  = Process.STATUS_WAITING
		self._thread = threading.Thread(target=self._executeThread, args=args, kwargs=kwargs)
		self._thread.start()
		return True

	#----------------------------------------------------------------------
	def _executeThread(self, *args, **kwargs):
		try:
			self.rc = self.execute(*args, **kwargs)
		except Exception as ex:
			self.message = ("Exception", str(ex))
			self.status  = Process.STATUS_ERROR
			self.rc = -1
			raise
		else:
			if self.rc != 0:
				self.status = Process.STATUS_ERROR
			else:
				self.percent = 100
				self.status  = Process.STATUS_FINISHED

	#----------------------------------------------------------------------
	def isAlive(self):
		if self.status in (Process.STATUS_WAITING, Process.STATUS_RUNNING):
			return True
		if self._thread is not None:
			return self._thread.is_alive()
		return False

	#----------------------------------------------------------------------
	def kill(self):
		self.status = Process.STATUS_KILLED

	#----------------------------------------------------------------------
	def isKilled(self):
		return self.status == Process.STATUS_KILLED

	#----------------------------------------------------------------------
	def wait(self, timeout=None):
		if self._thread is not None:
			self._thread.join(timeout)

	#----------------------------------------------------------------------
	# Method to override with the cleaning of the execution output
	#----------------------------------------------------------------------
	def clean(self, *args, **kwargs):
		self.init()
		return 0

	#----------------------------------------------------------------------
	# Methods to override with the execution of the task
	#----------------------------------------------------------------------
	def execute(self, *args, **kwargs):
		self.status = Process.STATUS_RUNNING
		return 0

	#----------------------------------------------------------------------
	def lock(self):
		self._mutex.acquire()

	#----------------------------------------------------------------------
	def release(self):
		self._mutex.release()

	#----------------------------------------------------------------------
	# Return/Set output
	#----------------------------------------------------------------------
	def output(self, out=None):
		self.lock()
		if out is None:
			out = "".join(self._out)
			del self._out[:]
		else:
			if isinstance(out,bytes):
				self._out.append(out.decode(errors="replace"))
			else:
				self._out.append(str(out))
			out = None
		self.release()
		return out

#===============================================================================
# PopenProcess - execute a system command given as a list argument
#===============================================================================
class PopenProcess(Process):
	def execute(self, cmd):
		super().execute()
		self.pipe = subprocess.Popen(cmd,
				stdin =subprocess.PIPE,
				stdout=subprocess.PIPE,
				stderr=subprocess.STDOUT,
				close_fds=True)
		while True:
			line = self.pipe.stdout.readline()
			if not line: break
			self.output(line)
		return self.pipe.wait()

#------------------------------------------------------------------------------
if __name__=="__main__":
	class MyProcess(Process):
		#----------------------------------------------------------------------
		def execute(self, path):
			super().execute()
			self.p = subprocess.Popen(["/bin/ls","-lR",path],
					stdout=subprocess.PIPE)
			self.stdout = self.p.stdout
			return self.p.wait()

		#----------------------------------------------------------------------
		def kill(self):
			super().kill()
			self.p.terminate()
