# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   02-Mar-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

#_INDENT  = "\t"

#===============================================================================
class Stream:
	QUOTENEED = " ,'\"#\t\n[]\\"
	#PREFIX   = re.compile(r"^\s*")
	#IDENT    = re.compile(r"[A-Za-z_][A-Za-z0-9._]*$")

	#-----------------------------------------------------------------------
	def __init__(self, filename, mode="r"):
		super().__init__()
		self.open(filename, mode)
		self.skipcolon    = True	# old format skip colon
		self.commentChar  = "#"
		self.continueChar = "\\"
		self.escapeChar   = "\\"
		self.tabsize      = 8		# standard tab size

	#-----------------------------------------------------------------------
	def open(self, filename, mode="r"):
		self.filename = filename
		if isinstance(filename,str):
			self.f = open(filename, mode,
						encoding="UTF-8",
						errors="replace")
		else:
			self.f = filename
		self.reset()

	#-----------------------------------------------------------------------
	def close(self):
		self.f.close()

	#-----------------------------------------------------------------------
	def name(self):
		"""return filename"""
		return self.f.name

	#-----------------------------------------------------------------------
	def reset(self):
		self.lineno  = 0	# line number
		self.comment = []
		self._save   = None	# saved line
		self.line    = ""	# current line
		self.prefix  = 0	# prefix spaces
		self.column  = 0	# column number
		#self.inStr   = False	# inside a string

	#-----------------------------------------------------------------------
	def readline(self):
		"""return read and return one line"""
		if self._save:
			self.line = self._save
			self._save = None
			self.column = 0
		else:
			self.line = self.f.readline()
			self.lineno += 1
			if not self.line:
				self.line = None
				return self.line
			if self.line[-1]=="\n":		# strip trailing \n
				self.line = self.line[:-1]

		# Find prefix
		self.column = 0
		self.prefix = 0
		for c in self.line:
			if c == " ":
				self.prefix += 1
			elif c == "\t":
				self.prefix += (self.prefix+self.tabsize)//self.tabsize * self.tabsize
			else:
				break
			self.column += 1
		return self.line

	#-----------------------------------------------------------------------
	def eol(self):
		return self.column >= len(self.line)

	#-----------------------------------------------------------------------
	def saveline(self, line=None):
		if self._save:
			raise Exception("Already a line is present in save buffer")
		if line is None:
			self._save = self.line
		else:
			self._save = line
		self.column = len(self.line)

	#-----------------------------------------------------------------------
	def skipSpaces(self, spaces=" \t"):
		while self.column<len(self.line):
			if self.line[self.column] not in spaces:
				return
			self.column += 1

#	#-----------------------------------------------------------------------
#	def skipSep(self):
#		return self.skipSpaces(": \t")
#
#	#-----------------------------------------------------------------------
#	def parseValue(self):
#		start = self.column
#		if self.line[self.column]=="'\"":
#		else:
#			val = self.line[self.column:]
#			while val[-1]==self.continueChar:
#				val = val[:-1]
#				self.readline()
#				if self.line is None: break
#				val += self.line
#			return val

	#-----------------------------------------------------------------------
	def parseString(self):
		"""parse a string from line starting at column"""
		quote = self.line[self.column]
		self.column += 1
		value = ""
		while True:
			c = self.line[self.column]
			self.column += 1
			if self.column >= len(self.line): break
			if c == quote:
				# check for a double quote
				if self.line[self.column]==quote:
					value += quote
					self.column += 1
					if self.column >= len(self.line): break
				else:
					break
			elif c=="\\":
				# accept literally next character as such
				value += self.line[self.column]
				self.column += 1
				if self.column >= len(self.line): break
			else:
				value += c
		return value

	#-----------------------------------------------------------------------
	def parseVarValue(self, separators=" \t,:"):
		"""Read line and split it to variable and value"""
		line = ""
		self.comment = []
		while True:
			curline = self.readline()
			if curline is None: return (None, None)
			curline = curline.strip()
			if not curline: continue
			if curline[0]==self.commentChar:
				self.comment.append(curline[1:])
				continue

			if curline[-1]==self.continueChar:
				line += curline[:-1] + " "
				continue

			line += curline

			# Find separator
			col = len(line)
			for sep in separators:
				pos = line.find(sep)
				if 0<=pos<col: col = pos

			# split
			if col>=0:
				tag = line[0:col].strip()
				val = line[col+1:].strip()
				return (tag, val)
			return (line, "")

	#-----------------------------------------------------------------------
	# read next identifier until separator.
	# if last character is in suffix ignore it
	#-----------------------------------------------------------------------
	def parseIdentifier(self, separators=" \t,", suffix=":"):
		"""Return tag from current column position"""
		start = self.column
		while self.column<len(self.line):
			if self.line[self.column] in separators: break
			self.column += 1
		if self.line[self.column-1] in suffix:
			self.column -= 1
		return self.line[start:self.column]

	#-----------------------------------------------------------------------
	# parse a section title like [name with space]
	#-----------------------------------------------------------------------
	#def parseSection(self):
	#	"""Return tag as a section title"""
	#	start = self.column
	#	# mustbe [
	#	while self.column<len(self.line):
	#		self.column += 1
	#		if self.line[self.column] == "]":
	#			self.column += 1
	#			break
	#	return self.line[start:self.column]

	#-----------------------------------------------------------------------
	def readBlock(self, prefix="|"):
		"""read a block of lines until ^L character"""
		block = []
		while True:
			line = self.readline()
			if line is None: break
			if prefix is not None:
				line = line.lstrip()
				if line.startswith(prefix):
					line = line[len(prefix):]
				elif not block:
					prefix = None	# compatibility with past
				else:
					self.saveline(line)
					break

			ep = line.find('\f')
			if ep >= 0:
				block.append(line[0:ep])
				break
			block.append(line)
		return "\n".join(block)

	#-----------------------------------------------------------------------
	# Parse a list from the stream
	#-----------------------------------------------------------------------
	def parseList(self, maxitems=9999999):
		"""@return a list from current position"""
		lst = []
		self.skipSpaces()
		if self.column>=len(self.line): return lst
		if self.line[self.column] == "[":
			bracket = True			# starting bracket
			self.column += 1
		else:
			bracket = False
		value   = ""
		space   = False				# space separator special treatment
		inquote = None
		self.line = self.line.rstrip()
		while self.column<len(self.line):
			c = self.line[self.column]
			self.column += 1
			if inquote is not None:
				# Inside a string
				if inquote==c:
					# end of string
					inquote = None
				elif c=="\\":
					if self.column>=len(self.line):
						self.readline()
						self.line = self.line.rstrip()
					else:
						value += self.line[self.column]
						self.column += 1
				else:
					value += c
			elif c in "'\"":
				# Start of string
				if value: # FIXME could be an error haha"bingo"
					lst.append(value)
					space = False
					value = ""
					if len(lst)>=maxitems:
						self.column -= 1
						return lst
				inquote = c
			elif c in " \t":
				space = True
			elif c == ",":		# separator
				lst.append(value)
				space = False
				value = ""
				if len(lst)>=maxitems: return lst
			elif c == "]":
				if bracket: break	# end of list
				value += c
			elif c in ":;":	# end of list
				break
			elif c == "\\":
				self.column += 1
				if self.column>=len(self.line):
					# end of line = continuation
					self.readline()
					self.line = self.line.rstrip()
					space = True
				else:
					# else accept literally next character
					value += self.line[self.column]
					self.column += 1
			elif space:
				# any other character append to value
				if value: lst.append(value)
				space = False
				if len(lst)>=maxitems:
					self.column -= 1
					return lst
				value = c
			else:
				value += c
		if space or value: lst.append(value)
		return lst

	#-----------------------------------------------------------------------
	# Convert a string to list equivalent to the parseList but from a string
	#-----------------------------------------------------------------------
	@staticmethod
	def str2list(line, maxitems=9999999):
		"""@return a list parsing a string"""
		lst     = []
		line    = line.strip()
		column  = 0
		bracket = 0
		if column>=len(line): return lst

		#---------------------------------------------------------------
		# Skip spaces and tabs
		#---------------------------------------------------------------
		def _skipSpaces():
			nonlocal column, line
			while column<len(line):
				if line[column] in " \t":
					column += 1
				else:
					break

		#---------------------------------------------------------------
		# parse an identifier accept anything except:
		#  space tab comma [ ] : ; \
		#---------------------------------------------------------------
		def _parseIdentifier(value):
			nonlocal column, line
			while column<len(line):
				if line[column] in " \t,[]":
					return value
				value += line[column]
				column += 1
			return value

		#---------------------------------------------------------------
		# parse a quoted string accept ' or " as start
		# double '' "" or escaped \' \"
		# inside the string
		# return the value
		# FIXME merge with the static method below
		#---------------------------------------------------------------
		def _parseString(quote):
			nonlocal column, line
			value = ""
			while True:
				c = line[column]
				column += 1
				if column >= len(line): return value
				if c == quote:
					# check for a double quote
					if line[column]==quote:
						value += quote
					else:
						return value
				elif c=="\\":
					# accept literally next character as such
					value += line[column]
					column += 1
					if column >= len(line): return value
				else:
					value += c

		#---------------------------------------------------------------
		# Main loop
		#---------------------------------------------------------------
		if line[column] == "[":	# FIXME count brackets!
			bracket += 1
			column  += 1
		valueAppended = False
		currentList = lst
		stack = []
		while column<len(line):
			_skipSpaces()

			c = line[column]
			column += 1

			if c in "'\"":
				currentList.append(_parseString(c))
				valueAppended = True

			elif c == ",":
				if not valueAppended:
					currentList.append("")
				valueAppended = False

			elif c == "[":
				bracket += 1
				currentList.append([])
				stack.append(currentList)
				currentList = currentList[-1]

			elif c == "]":
				bracket -= 1
				if bracket == 0: break
				currentList = stack.pop()

			#elif c == ":;":

			else:
				currentList.append(_parseIdentifier(c))
				valueAppended = True

		if bracket:
			assert line[column]==']'
		return lst

	#-----------------------------------------------------------------------
	# Convert a list to string
	#-----------------------------------------------------------------------
	@staticmethod
	def list2str(lst, maxlen=70, prefix="\t", continuation=" \\"):
		"""@return a list as a string starting with prefix if needed"""
		comma   = ""
		line    = ""
		nextmax = 70
		for item in lst:
			if isinstance(item,str) and \
			   any(x in item for x in Stream.QUOTENEED):
				if len(line)>nextmax:
					line += continuation+"\n"+prefix
					nextmax += maxlen
				if line or comma:
					line += f"{comma} {item!r}"
				else:
					line += repr(item)
				comma = ""
			else:
				s = str(item).strip()
				if s:
					if len(line)>nextmax:
						line += continuation+"\n"+prefix
						nextmax += maxlen
					if line or comma:
						line += f"{comma} {s}"
					else:
						line = s
					comma = ""
				else:
					if line and not comma: comma = ","
					comma += ","
		return line

	#-------------------------------------------------------------------------------
	# Quote a value if necessary
	#-----------------------------------------------------------------------
	@staticmethod
	def quoteIfNeeded(value):
		if isinstance(value,str) and \
		   any(x in value for x in Stream.QUOTENEED):
			return repr(value)
		else:
			return str(value).strip()

	#-----------------------------------------------------------------------
	def write(self, s):
		self.f.write(s)

	#-----------------------------------------------------------------------
	def writeline(self, s):
		self.f.write(s+"\n")

	#-----------------------------------------------------------------------
	def writeValue(self, item):
		"""Write a single value"""
		self.f.write(Stream.quoteIfNeeded(item))

	#-----------------------------------------------------------------------
	def writeList(self, lst, prefix="\t"):
		"""Write a list starting with prefix if needed"""
		self.f.write(Stream.list2str(lst, prefix))

	#-----------------------------------------------------------------------
	def writeBlock(self, block, prefix="\t|"):
		for line in block.splitlines():
			self.writeline(f"{prefix}{line}")

#	#-----------------------------------------------------------------------
#	# Return next item
#	#-----------------------------------------------------------------------
#	def nextToken(self, schema):
#		c = self.line[self.column]
#
#		if "A"<=c<="Z" or "a"<=c<="z" or c in ("_","#","@"):
#			# parse identifier
#			token = c
#			while self.column<len(self.line):
#				self.column += 1
#				c = self.line[self.column]
#				if "A"<=c<="Z" or "a"<=c<="z" or "0"<=c<="9" or c in ("_","#","@"):
#					token += c
#				else:
#					break
#			return Parser.IDENT, token
#
#		elif c in ("'",'"'):
#			# parse string
#			quote = c
#			token = ""
#			while True:
#				self.column += 1
#				c = self.line[self.column]
#				if c=="\\":
#					self.column += 1
#					if self.column>=len(self.line):
#						self.readline()
#					else:
#						token += self.line[self.column]
#				elif c==quote:
#					self.column += 1
#					break
#				else:
#					token += c
#			return Parser.STR, token
#
#		elif "0"<=c<="9" or c in ("+","-","."):
#			# parse number
#			pass
#
#	#-----------------------------------------------------------------------
#	def read(self):
#		pass
#
##===============================================================================
## Schema format is dictionary with entries
##	<name> : <value>
## <name>:
##	any string, use * for global matching
##
## <value>:
##	[<type>...]/[<options>...] (,default-value)
##	<type>:
##		i = integer   (default=0)
##		r = real      (default=0.0)
##		v = variable  (no-default)
##		s = string    (default="")
##		l = list      (default=[])
##		d = dictionary
##		D = list + dictionary
##
##	<options>
##		b = use-brackets
##		B = don't use brackets(default)
##		c = use-comma
##		C = don't use comma(default)
##		m = multi-line string (like notes start from next line
##		M = single line or break at 80chars
##
##	<default-value>
##		- either a fixed value, string, int, float
##		- another schema for sub-content
##===============================================================================
#_DEFAULT = "D"
#_DEFAULT_VALUES = {
#	"i" : 0,
#	"r" : 0.0,
#	"s" : "",
#	"l" : [],
#}
#
#input_schema = {
#	"REGION" : "vDS",
#	"*"      : "D",
#}
#
#run_schema = {
#	"prev"     : "i",
#	"last"     : ("i",5),
#	"status"   : "i",
#	"pid"      : "i",
#	"startrun" : "i",
#	"family"   : "l",
#	"parent"   : "s",
#}
#
#main_schema = {
#	"version" : "r",
#	"timeout" : ("r","0.25"),
#	"title"   : "s",
#	"notes"   : "s/m",
#	"list"    : "l",
#	"list2"   : "l/c",
#	"input"   : ("v",input_schema),
#	"run"     : ("v",run_schema),
#}
#
##===============================================================================
##class Schema(dict):
##	pass
#
##===============================================================================
## flair markup language
## Dictionary:
##	name [value[, *]]...	list of values
##	namekkk
##===============================================================================
#class FML:
#	def __init__(self, schema):
#		super().__init__()
#		self.schema = schema
#
#	#-----------------------------------------------------------------------
#	def read(self, filename, ):
#		self._f = open(filename,"r")
#		self._f.close()
#
#	#-----------------------------------------------------------------------
#	def write(self, filename, data):
#		self._f = open(filename,"w")
#		self._col = 0
#		self.writeBranch(data, self.schema, "")
#		self._f.close()
#
#	#-----------------------------------------------------------------------
#	# low level write command which records the line length
#	# and breaks long string if needed
#	#-----------------------------------------------------------------------
#	def _write(self, line, prefix=""):
#		if line=="\n":
#			self._f.write(line)
#			self._col = 0
#		elif "\n" in line:
#			first = True
#			for s in line.splitlines():
#				if not first: self._f.write(prefix)
#				self._f.write(s)
#				self._f.write(" \n")
#				self._col = len(prefix)+len(s)
#		else:
#			self._f.write(line)
#			self._col += len(line)
#
#	#-----------------------------------------------------------------------
#	def writeBranch(self, data, schema, prefix):
#		for n,v in data.items():
#			try:
#				fmt = schema[n]
#			except KeyError:
#				print(f"Schema {n} not found")
#				fmt = _DEFAULT
#
#			if isinstance(fmt, tuple):
#				fmt, default = fmt
#			else:
#				default = _DEFAULT_VALUES.get(fmt[0],"")
#
#			# skip default values
#			if v==default: continue
#
#			if isinstance(v, tuple):
#				self._write(prefix)
#				self._writeValue(n, prefix)
#				self._write("\t")
#				self._writeValue(v[0], prefix, "b" in fmt, "c" in fmt)
#				self._write("\n")
#
#				self.writeBranch(v[1], default, prefix+_INDENT)
#			else:
#				self._write(prefix)
#				self._writeValue(n, prefix)
#				self._write(_INDENT)
#				self._writeValue(v, prefix, "b" in fmt, "c" in fmt)
#				self._write("\n")
#
#	#-----------------------------------------------------------------------
#	def _writeValue(self, value, prefix, usebrackets=True, usecomma=True, multilines=False):
#		# --- number ---
#		if isinstance(value, (int,float)):
#			self._write(str(value))
#
#		# --- string ---
#		elif isinstance(value, str):
#			if _VAR_PAT.match(value):
#				self._write(str(value))
#			else:
#				self._write(repr(value))
#
#		# --- list ---
#		elif isinstance(value, list):
#			if usebrackets: self._write("[")
#			comma = False
#			commastr = ", " if usecomma else " "
#			for item in value:
#				if comma: self._write(commastr)
#				comma = True
#				self._writeValue(item, prefix+_INDENT)
#			if usebrackets: self._write("]")
#
#		# --- dictionary ---
#		elif isinstance(value, dict):
#			pass
#
#		else:
#			raise Exception("Unknown variable type %s"%(type(value)))

if __name__ == "__main__":
	import io
	def testList(s):
		stream = Stream(io.StringIO(s))
		print(f"\nLine='{s}'")
		print(f"Read=<{stream.readline()}>")
		print("List=",stream.parseList())
	testList(" abc  def  ,   efg '\\''")
	testList(" [abc  def  , ]  efg '\\''")
	testList(" [abc  def  ]  efg '\\''")
	testList(" abc  def  :  efg '\\''")
	testList(" abc  def   efg '\\''\\      \n \\ \n xyz")
	testList(" abc  def ,  efg '\\''\\      \n  , \\ \n xyz")

#	lst=["ene dep",1.,"NEUTRON",-21,"","","",1.,"","",2.,3.,4.]
#	print("list=",lst)
#	line = Stream.list2str(lst)
#	print("write=",line)
#	print("parse=",Stream.parseList(line.strip()))
