#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Author:	Vasilis.Vlachoudis@cern.ch
# Date:	29-Dec-2019
#from __future__ import print_function

__author__   = "Vasilis Vlachoudis"
__email__    = "Vasilis.Vlachoudis@cern.ch"
__version__  = "0.0"

import bisect

import tkinter as tk

import tkExtra
from tkconst import STATE

#===============================================================================
class TimeFrameCanvas(tk.Canvas):
	BACKGROUND    = "#404040"
	TEXT_COLOR    = "#FFFFFF"
	ROW_COLOR     = ("#505050", "#606060")
	GRID0_COLOR   = "#A0A0A0"
	GRID5_COLOR   = "#303030"
	CURRENT_COLOR = "#FF8C00"
	KEY_COLOR     = "#FFFF00"
	#FONT          = ("Helvetica",-12)
	FRAME_LABEL   = "frame:"

	#-----------------------------------------------------------------------
	def __init__(self, master, *args, **kwargs):
		super().__init__(master, *args, **kwargs)
		self["background"] = TimeFrameCanvas.BACKGROUND
		self.bind("<Configure>",	self.show)
		self.bind("<Button-1>",		self.button1)
		self.bind("<Control-Button-4>",	self.zoomIn)
		self.bind("<Control-Button-5>",	self.zoomOut)
		self.bind("<Button-4>",		self.scrollLeftFast)
		self.bind("<Button-5>",		self.scrollRightFast)
		self.bind("<B1-Motion>",	self.button1Motion)
		self.bind("<B2-Motion>",	self.button2Motion)
		self.bind("<ButtonRelease-1>",	self.buttonRelease)
		self.bind("<ButtonRelease-2>",	self.buttonRelease)
		self.bind("<Key>",		self.handleKey)
		self.bind("<Enter>",		self.focus)

		self.startVar = tk.IntVar()
		self.stopVar  = tk.IntVar()

		e = tk.Entry(self,	textvariable=self.startVar,
					width=5,
					justify=tk.LEFT,
					insertbackground="White",
					foreground=TimeFrameCanvas.KEY_COLOR,
					background=TimeFrameCanvas.ROW_COLOR[0],
					borderwidth=0)
		e.place(relx=0, rely=1, anchor=tk.SW)
		e.bind("<Return>",   self.focus)
		e.bind("<KP_Enter>", self.focus)
		tkExtra.Balloon.set(e, "Start frame number")

		e = tk.Entry(self,	textvariable=self.stopVar,
					width=5,
					justify=tk.RIGHT,
					insertbackground="White",
					foreground=TimeFrameCanvas.KEY_COLOR,
					background=TimeFrameCanvas.ROW_COLOR[1],
					borderwidth=0)
		e.place(relx=1, rely=1, anchor=tk.SE)
		e.bind("<Return>",   self.focus)
		e.bind("<KP_Enter>", self.focus)
		tkExtra.Balloon.set(e, "Last frame number")

		self.key_width = 0
		self.scan_drag = None
		self.current   = 0		# current frame
		self.start     = 0		# starting frame number
		self.stop      = 250		# stopping frame number
		self.left      = self.start	# left side displayed
		self.zoom      = 6		# pixel zooming for each frame
		self.maxkeys   = 3		# constraint plotting of keys
		self.keys      = [(TimeFrameCanvas.FRAME_LABEL,[])]
		self.selected  = set()		# selected keys
		self._keysId   = {}		# keys drawing id
		self.active    = None		# key last clicked
		self.startVar.trace("w", self.entryChanged)
		self.stopVar.trace("w",  self.entryChanged)
		self.setLimits(self.start, self.stop)

	#-----------------------------------------------------------------------
	def focus(self, event=None):
		self.focus_set()

	#-----------------------------------------------------------------------
	# Set the frame limits
	#-----------------------------------------------------------------------
	def setLimits(self, start, stop, zoom=None):
		self.startVar.set(start)	# set the variables
		self.stopVar.set(stop)		# which will update the limits

	#-----------------------------------------------------------------------
	# low level call without updating the variables
	# used only from the entryChanged when the limit variables are
	# manually edited
	#-----------------------------------------------------------------------
	def _setLimits(self, start, stop, zoom=None):
		if stop < start:
			start, stop = stop, start
		if stop < start+10:	# minimum 10 frames
			stop = start + 10
		self.start = start	# first key frame
		self.stop  = stop	# last key frame
		self.left  = start	# left-most keyframe displayed
		if zoom:
			self.zoom = zoom # zoom level
		self.current = max(self.start, min(self.current, self.stop))

	#-----------------------------------------------------------------------
	def entryChanged(self, a, b, c):
		try:
			start = int(self.startVar.get())
			stop  = int(self.stopVar.get())
		except:
			return
		if start != self.start or stop != self.stop:
			self._setLimits(start,stop)
			self.draw()

	#-----------------------------------------------------------------------
	def setCurrent(self, current):
		self.current = max(self.start, min(current, self.stop))
		self.see()
		self.draw()

	#-----------------------------------------------------------------------
	def nframes(self):
		return self.stop - self.start + 1

	#-----------------------------------------------------------------------
	# Set the keys and redraw
	#-----------------------------------------------------------------------
	def setKeys(self, keys=[], selected=None):
		if keys:
			self.keys = keys
		else:
			self.keys = [(TimeFrameCanvas.FRAME_LABEL,[])]
		if selected is not None:
			self.selected = selected
		else:
			self.selected = set()
		self.show()

	#-----------------------------------------------------------------------
	def select(self, name, key):
		self.selected.add((name,key))

	#-----------------------------------------------------------------------
	def unselect(self, name, key=None):
		if key is not None:
			self.selected.discard(name,key)
		else:
			for nk in list(self.selected):
				self.selected.remove(nk)

	#-----------------------------------------------------------------------
	def zoomIn(self, event=None):
		self.zoom = min(20,self.zoom+1)
		self.draw()

	#-----------------------------------------------------------------------
	def zoomOut(self, event=None):
		self.zoom = max(3, self.zoom-1)
		self.draw()

	#-----------------------------------------------------------------------
	def notify(self):
		self.event_generate("<<TimeFrameChange>>", data=self.current)

	#-----------------------------------------------------------------------
	def width(self):
		return self.winfo_width()-self.key_width

	#-----------------------------------------------------------------------
	# Horizontal scroll keyframe
	#-----------------------------------------------------------------------
	def scroll(self, dx):
		self.left += dx
		self.left = max(self.start,
				min(self.left,
					self.stop - self.width()//self.zoom+10))
		self.show()

	#-----------------------------------------------------------------------
	def scrollLeftFast(self, event=None):	self.scroll(-10)
	def scrollRightFast(self, event=None):	self.scroll( 10)

	#-----------------------------------------------------------------------
	# Make self.current visible
	# @return True if current was visible, else False
	#-----------------------------------------------------------------------
	def see(self):
		if self.current < self.left:
			while self.current < self.left:
				self.scroll(self.current - self.left - 5)
			return False

		if self.current > self.left + self.width()//self.zoom - 5:
			while self.current > self.left + self.width()//self.zoom - 5:
				self.scroll(5)
			return False

		return True

	#-----------------------------------------------------------------------
	# move current position and make it visible if necessary
	#-----------------------------------------------------------------------
	def moveCurrent(self, dx):
		self.current += dx
		self.current = max(self.start, min(self.current, self.stop))
		if self.see():
			self.show()
		self.notify()

	#-----------------------------------------------------------------------
	# Add current frame to keys
	#-----------------------------------------------------------------------
	def addCurrent(self):
		for key,frames in self.keys:
			if key==TimeFrameCanvas.FRAME_LABEL: break
			if self.current not in frames:
				bisect.insort(frames, self.current)
		self.show()

	#-----------------------------------------------------------------------
	# Delete current frame from keys
	#-----------------------------------------------------------------------
	def delCurrent(self):
		for key,frames in self.keys:
			if key==TimeFrameCanvas.FRAME_LABEL: break
			try:
				frames.remove(self.current)
			except ValueError:
				pass
		self.show()

	#-----------------------------------------------------------------------
	# Keyboard events
	#-----------------------------------------------------------------------
	def handleKey(self, event):
		#print("keysym=",event.keysym,"char=",event.char)
		if event.keysym in ("space", "a"):
			self.addCurrent()
		elif event.keysym in ("Delete", "BackSpace", "KP_Delete", "x"):
			self.delCurrent()
		elif event.keysym == "Left":
			self.moveCurrent(-10 if event.state&STATE.CONTROL else -1)
		elif event.keysym == "Right":
			self.moveCurrent(10 if event.state&STATE.CONTROL else 1)
		elif event.keysym == "Prior":
			dw = self.width()//(self.zoom*2)
			self.scroll(-dw)
		elif event.keysym == "Next":
			dw = self.width()//(self.zoom*2)
			self.scroll(dw)
		elif not event.char:
			return
		elif event.char in ("=","+"):
			self.zoomIn()
		elif event.char in ("-","_"):
			self.zoomOut()

	#-----------------------------------------------------------------------
	# @return True if current didn't change
	#-----------------------------------------------------------------------
	def _findCurrent(self, event):
		old = self.current
		self.current = (event.x+self.zoom//2 - self.key_width) // self.zoom + self.left
		self.current = max(self.start, min(self.current, self.stop))
		return old == self.current

	#-----------------------------------------------------------------------
	# We don't use canvasx,y = event.x no shift
	#-----------------------------------------------------------------------
	def button1(self, event=None):
		if self._findCurrent(event): return
		self.active = None
		for tag in self.find_overlapping(event.x, event.y, event.x+1, event.y+1):
			try:
				c = self._keysId[tag]
				if self.active  is None:
					self.active  = c
				elif c in self.selected:
					self.active  = c
			except KeyError:
				continue
		if self.see(): self.show()
		self.notify()

	#-----------------------------------------------------------------------
	# Draw active if any
	#-----------------------------------------------------------------------
	def button1Motion(self, event):
		if self._findCurrent(event): return
		if self.active is not None:
			self.event_generate("<<TimeFrameActive>>")
		if self.see(): self.show()
		self.notify()

	#-----------------------------------------------------------------------
	# Drag keyframe
	#-----------------------------------------------------------------------
	def button2Motion(self, event=None):
		if self.scan_drag is None:
			self.scan_drag = event.x
		else:
			dx = self.scan_drag - event.x
			s = dx // self.zoom
			if s:
				self.scroll(s)
				self.scan_drag = event.x

	#-----------------------------------------------------------------------
	def buttonRelease(self, event=None):
		self.scan_drag = None

	#-----------------------------------------------------------------------
	# Draw keyframe
	#-----------------------------------------------------------------------
	def draw(self):
		self.delete(tk.ALL)
		start = max(self.start, self.left)
		start10 = (start//10)*10
		stop10  = (self.stop//10)*10
		width = self.winfo_width()
		x  = 0
		y  = 0
		dy = 0		# line height
		self.key_width = 0

		# Draw Key names on left side
		icolor = 0
		for key,_ in self.keys[:self.maxkeys]:
			t = self.create_text(x+2, y,
					text=key,
					#font=self.font,
					fill=TimeFrameCanvas.TEXT_COLOR,
					anchor=tk.NW)
			x1,y1,x2,y2 = self.bbox(t)
			self.key_width = max(self.key_width, x2)
			dy = y2-y1
			b = self.create_rectangle((x, y, width, y2),
					fill=TimeFrameCanvas.ROW_COLOR[icolor],
					width=0)
			icolor = 1-icolor
			self.lower(b)
			y = y2 + 1
		self.key_width += 10

		# Frame number and vertical grid
		y  += 2
		skipodd = self.zoom<=4
		for i in range(start10, stop10+1, 10):
			x = self.key_width + self.zoom*(i-self.left)
			if x>=self.key_width:
				self.create_line(x,0,x,y, fill=TimeFrameCanvas.GRID0_COLOR)
			x2 = x + self.zoom*5
			if x2>=self.key_width:
				self.create_line(x2,0,x2,y-2, fill=TimeFrameCanvas.GRID5_COLOR)
			if i<self.start:
				i = self.start
				x = self.key_width + self.zoom*(i-self.left)
			if x > width: break
			if i==start or (i-start)*self.zoom > 15 and abs(i-self.current)*self.zoom>15:
				if skipodd and (i//10) & 1: continue
				self.create_text(x, y,
						text=str(i),
						#font=self.font,
						fill=TimeFrameCanvas.TEXT_COLOR,
						anchor=tk.N,
						justify=tk.CENTER)

		# Draw current cursor
		x = self.key_width + self.zoom*(self.current-self.left)
		cur = self.create_text(x, y,
				text=str(self.current),
				fill=TimeFrameCanvas.CURRENT_COLOR,
				anchor=tk.N,
				justify=tk.CENTER)

		# cleanup overlaps
		x1,y1,x2,y2 = self.bbox(cur)
		y12 = (y1+y2)//2
		for i in self.find_overlapping(x1,y12,x2,y12):
			if i != cur: self.delete(i)

		y = 0
		x -= self.zoom//2+1
		self.create_rectangle((x, y, x+self.zoom+3, len(self.keys[:self.maxkeys])*dy),
					fill=TimeFrameCanvas.CURRENT_COLOR,
					width=0)

		# Draw keys
		self._keysId = {}
		for (key,frames) in self.keys[:self.maxkeys]:
			for i,f in enumerate(frames):
				if f<self.left: continue
				x = self.key_width + self.zoom*(f-self.left)
				if x>width: break
				if (key,i) in self.selected:
					symbol="\N{BLACK DIAMOND}"
				else:
					symbol="\N{WHITE DIAMOND}" #CONTAINING BLACK SMALL DIAMOND,
				idx = self.create_text(x,y,
					text=symbol,
					fill=TimeFrameCanvas.KEY_COLOR,
					anchor=tk.N,
					justify=tk.CENTER)
				self._keysId[idx] = (key,i)
			y += dy

	#-----------------------------------------------------------------------
	# adjust height to display everything
	#-----------------------------------------------------------------------
	def show(self, event=None):
		self.draw()
		try: x1,y1,x2,y2 = self.bbox(tk.ALL)
		except: return
		self.configure(height=y2-y1)

#-------------------------------------------------------------------------------
if __name__ == "__main__":
	root = tk.Tk()

	tframe = TimeFrameCanvas(root)
	tframe.pack(fill=tk.BOTH, expand=tk.YES)

	tframe.setKeys(["Camera","Light","Test"])
	#tframe.update_idletasks()
	tframe.draw()

	tframe.bind("<Key-q>", lambda e : root.quit())
	tframe.focus_set()

	root.mainloop()
