#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   12-Oct-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re
import sys
import glob
import time
import tool
import tkinter as tk
from tkconst import STATE
from fnmatch import fnmatch
from tkinter import ttk, dnd
from tkinter.colorchooser import askcolor

import log
import icons
import bFileDialog

#-------------------------------------------------------------------------------
# Display busy cursor
#-------------------------------------------------------------------------------
def busy(widget, update=True):
	try:
		widget.tk.call("tk","busy","hold",widget.winfo_toplevel())
	except tk.TclError:
		widget.config(cursor="watch")
	if update: widget.update_idletasks()

#-------------------------------------------------------------------------------
# Restore to default
#-------------------------------------------------------------------------------
def notBusy(widget, update=True):
	try:
		widget.tk.call("tk","busy","forget",widget.winfo_toplevel())
	except tk.TclError:
		widget.config(cursor="")
	if update: widget.update_idletasks()

#-------------------------------------------------------------------------------
# Center window under parent
#-------------------------------------------------------------------------------
def placeWindow(window, parent):
	window.tk.call("::tk::PlaceWindow", window, "widget", parent)

#-------------------------------------------------------------------------------
# bind event with data as a replacement of the tkinter bind for virtual
# events to send data <<...>>
#
# Example, instead of binding like
#	widget.bind("<<VirtualEvent>>", function)
# use it as
#	bindData(widget, "<<VirtualEvent>>", function)
#
# def function(event):
#	print event.serial, event.widget, event.data
#
# Send message as
#	widget.event_generate("<<VirtualEvent>>", data="Hello")
#	widget.event_generate("<<VirtualEvent>>", data=("One","Two"))
#	widget.event_generate("<<VirtualEvent>>", serial=10, data=("One","Two"))
#
# WARNING: Unfortunately it will convert data to STRING!!!
#-------------------------------------------------------------------------------
def bindData(widget, sequence, func, add=None):
	def _substitute(*args):
		e = tk.Event()
		nsign, b, t, T, d, W = args
		try:    e.serial = int(nsign)
		except: e.serial = nsign
		try:    e.num    = int(b)
		except: e.num    = b
		try:    e.time   = int(t)
		except: e.time   = t
		e.type = T
		e.data = d
		try:
			e.widget = widget._nametowidget(W)
		except KeyError:
			e.widget = W
		return (e,)

	funcid = widget._register(func, _substitute, needcleanup=1)
	cmd = '{0}if {{"[{1} %# %b %t %T %d %W]" == "break"}} break\n'.format('+' if add else '', funcid)
	widget.tk.call('bind', widget._w, sequence, cmd)

#===============================================================================
# Sort Assist class for MultiListbox
#===============================================================================
class SortAssist:
	def __init__(self, column):
		self.column = column

	def __call__(self, x):
		return x[self.column]

#-------------------------------------------------------------------------------
# Multiple configuration of many widgets given in a list
# lst = list of widgets
#-------------------------------------------------------------------------------
def multiConfig(lst, **opts):
	"""Multiple configuration of many widgets"""
	for w in lst:
		w.config(**opts)

#-------------------------------------------------------------------------------
# Toggle toplevel window height
#-------------------------------------------------------------------------------
def toggleHeight(root, oldHeight):
	"""Toggle window height"""
	m = re.match(r"(\d+)x(\d+)\+(-?\d+)\+(-?\d+)", root.wm_geometry())
	if not m:
		root.bell()
		return oldHeight

	width, height, x, y = map(int, m.groups())
	maxheight = root.winfo_screenheight()

	if sys.platform in ("win32", "win64"):
		newy = 0
		newheight = maxheight - 72
	else:
		#newy = 24
		newy = 0
		#newheight = maxheight - 96
		newheight = maxheight - 88

	if height >= newheight:
		try:
			newheight = oldHeight
		except:
			return oldHeight

	root.wm_geometry(f"{width}x{newheight}+{x}+{newy}")
	return height

#===============================================================================
def _entryPaste(event):
	"""global replacement for the Entry.paste"""
	try:
		event.widget.delete('sel.first', 'sel.last')
	except tk.TclError:
		pass	# nothing is selected

	# in tk.call() use the widget's string representation event.widget._w
	# instead of event.widget, which is the widget instance itself
	try:
		text = event.widget.tk.call('::tk::GetSelection', event.widget._w, 'CLIPBOARD')
	except tk.TclError:
		return
	event.widget.insert(tk.INSERT, text)
	event.widget.tk.call('tk::EntrySeeInsert', event.widget._w)
	return "break"

#-------------------------------------------------------------------------------
def _textPaste(event):
	"""global replacement for the Text.paste"""
	oldSeparator = event.widget.cget("autoseparators")
	if oldSeparator:
		event.widget.config(autoseparators=0)
		event.widget.edit_separator()
	try:
		event.widget.delete('sel.first', 'sel.last')
	except tk.TclError:
		pass	# nothing is selected

	# in tk.call() use the widget's string representation event.widget._w
	# instead of event.widget, which is the widget instance itself
	try:
		text = event.widget.tk.call('::tk::GetSelection', event.widget._w, 'CLIPBOARD')
	except tk.TclError:
		return
	event.widget.insert(tk.INSERT, text)
	if oldSeparator:
		event.widget.edit_separator()
		event.widget.config(autoseparators=1)
	event.widget.see('insert')
	return "break"

#-------------------------------------------------------------------------------
def _selectAll(event):
	try:
		event.widget.selection_range(0,tk.END)
	except AttributeError:
		# XXX in the default color selection dialog the event.widget is 'str'
		pass

#-------------------------------------------------------------------------------
def changeBehaviourTextClasses(root):
	root.bind_class('Entry', '<<Paste>>', _entryPaste)
	root.bind_class('Text',  '<<Paste>>', _textPaste)
	root.bind_class('Entry', '<Control-Key-a>', _selectAll)

#===============================================================================
# Entry having a set value
#===============================================================================
class Entry(tk.Entry):
	def set(self, value):
		self.delete(0,tk.END)
		self.insert(0, value)

#===============================================================================
# LabelEntry. display a label when entry field is empty
#===============================================================================
class LabelEntry(tk.Entry):
	def __init__(self, master, label=None, labelcolor=None, **kw):
		super().__init__(master, **kw)
		self.label  = label
		self._empty = True
		self._fg    = self["foreground"]
		if labelcolor is not None:
			self.labelcolor = labelcolor
		else:
			self.labelcolor = self._fg

		self._clear = tk.Button(self,
					text="\N{VECTOR OR CROSS PRODUCT}",
					command=self.clear,
					foreground=self["background"],
					relief=tk.FLAT,
					cursor="arrow",
					borderwidth=0,
					padx=-1, pady=-1)
		self._clear.pack(side=tk.RIGHT, padx=0, pady=0)
		self.pack_propagate(0)
		Balloon.set(self._clear, "Clear entry")

		self.bind("<FocusIn>",  self._focusIn)
		self.bind("<FocusOut>", self._focusOut)
		self["validate"] = "key"
		self["validatecommand"] = (self.register(self.validate), '%P')
		self.showLabel()

	# ----------------------------------------------------------------------
	def showLabel(self):
		self.delete(0, tk.END)
		self.insert(0, self.label)
		self["foreground"] = self.labelcolor
		self._empty = True	# Restore empty since validation will destroy it

	# ----------------------------------------------------------------------
	def removeLabel(self):
		self.delete(0,tk.END)
		self["foreground"] = self._fg

	# ----------------------------------------------------------------------
	def _focusIn(self, event):
		if self._empty:
			self.removeLabel()

	# ----------------------------------------------------------------------
	def _focusOut(self, event):
		if self._empty or self.get()=="":
			self.showLabel()

	# ----------------------------------------------------------------------
	def validate(self, value):
		self._empty = value == ""
		self["foreground"] = self._fg
		return True

	# ----------------------------------------------------------------------
	def set(self, value):
		self._empty = value==""
		if self._empty:
			self.showLabel()
			self.master.focus_set()	# lose focus
		else:
			self.removeLabel()
			self.insert(0, value)

	# ----------------------------------------------------------------------
	def get(self):
		if self._empty:
			return ""
		else:
			return tk.Entry.get(self)

	# ----------------------------------------------------------------------
	def clear(self):
		if self.focus() is self:
			self.delete(0, tk.END)
		else:
			self.removeLabel()
		self.event_generate("<<LabelEntryClear>>")

#===============================================================================
# _ValidatingEntry
#===============================================================================
class _ValidatingEntry(tk.Entry):
	"""base class for validating entry widgets"""
	# ----------------------------------------------------------------------
	def __init__(self, master, value="", **kw):
		super().__init__(master, **kw)
		self["validate"] = "key"
		self["validatecommand"] = (self.register(self.validate), '%P')

	# ----------------------------------------------------------------------
	def validate(self, value):
		# override: return True if valid False if invalid
		return True

	# ----------------------------------------------------------------------
	def set(self, value):
		self.delete(0, tk.END)
		self.insert(0, value)

	# ----------------------------------------------------------------------
	def getint(self, default=0):
		try:
			return int(self.get())
		except:
			return default

	# ----------------------------------------------------------------------
	def getfloat(self, default=0.0):
		try:
			return float(self.get())
		except:
			return default

#===============================================================================
# Maximum Length Entry
#===============================================================================
class MaxLengthEntry(_ValidatingEntry):
	"""MaxLengthEntry limit entry length maximum maxlength characters"""
	def __init__(self, master, value="", maxlength=None, **kw):
		super().__init__(master, value, **kw)
		self.maxlength = maxlength

	# ----------------------------------------------------------------------
	def insert(self, idx, value):
		m = self.maxlength
		self.maxlength = None
		super().insert(idx, value)
		self.maxlength = m

	# ----------------------------------------------------------------------
	def validate(self, value):
		if self.maxlength is not None:
			return len(value) <= self.maxlength
		return True

#===============================================================================
# Integer Validating Entry
#===============================================================================
class IntegerEntry(_ValidatingEntry):
	"""IntegerEntry accepting only integers"""
	# ----------------------------------------------------------------------
	def validate(self, value):
		try:
			if value: int(value)
			return True
		except ValueError:
			if value in ("+","-"): return True
		return False

#===============================================================================
# Floating Point Validating Entry
#===============================================================================
class FloatEntry(_ValidatingEntry):
	"""accept only floating point numbers"""
	# ----------------------------------------------------------------------
	def validate(self, value):
		try:
			if value: float(value)
			return True
		except ValueError:
			if value in ("+", "-", ".", "+.", "-."): return True
			if len(value)>1:
				last = value[-1]
				if last in ("e","E"):
					if "e" in value[:-1] or \
					   "E" in value[:-1]: return False
					return True
				plast = value[-2]
				if plast in ("e","E") and \
				    last in ("-","+"): return True
		return False

#===============================================================================
# Vector Validating Entry
#===============================================================================
class VectorEntry(_ValidatingEntry):
	"""accept only vectors"""
	PAT = re.compile(r"[(),;\[\]]")

	# ----------------------------------------------------------------------
	def validate(self, value):
		# remove from value comma, semicolon, and parenthesis () []
		for token in VectorEntry.PAT.sub(" ",value).split():
			try:
				float(token)
			except ValueError:
				if token in ("+","-",".","+.","-."): continue
				if len(token)>1:
					last = token[-1]
					if last in ("e","E"):
						if "e" in token[:-1] or \
						   "E" in token[:-1]: return False
						return True
					plast = token[-2]
					if plast in ("e","E") and \
					    last in ("-","+"): return True
				return False
		return True

	# ----------------------------------------------------------------------
	# Get contents as a list
	# ----------------------------------------------------------------------
	def getlist(self):
		return VectorEntry.PAT.sub(" ",self.get()).split()

	# ---------------------------------------------------------------------
	# Split vector in to a list of widgets
	# ----------------------------------------------------------------------
	def split(self, widgets):
		value = self.get()
		for ch in " ,()[];":
			if ch in value:
				xyz = self.getlist()
				if xyz:
					self.set(xyz[0])
					for i,w in enumerate(widgets):
						if len(xyz)>i+1: w.set(xyz[i+1])
				return

#===============================================================================
# Based on implementation of Mitja Martini on 2008-11-29.
# updated by Russell Adams, 2011/01/24 to support Python 3 and Combobox.
#===============================================================================
class AutocompleteEntry(tk.Entry):
	"""
	Subclass of Tkinter.Entry that features autocompletion.

	To enable autocompletion use set_completion_list(list) to define
	a list of possible strings to hit.
	To cycle through hits use down and up arrow keys.
	"""
	#tkinter_umlauts=['odiaeresis', 'adiaeresis', 'udiaeresis', 'Odiaeresis',
	#		'Adiaeresis', 'Udiaeresis', 'ssharp']

	#-----------------------------------------------------------------------
	def set_completion_list(self, completion_list):
		# Work with a sorted list
		self._completion_list = sorted(completion_list, key=str.lower)
		self._hits      = []
		self._hit_index = 0
		self.position   = 0
		self.bind('<KeyRelease>', self.handle_keyrelease)

	#-----------------------------------------------------------------------
	# autocomplete the Entry, delta may be 0/1/-1 to cycle through
	# possible hits
	#-----------------------------------------------------------------------
	def autocomplete(self, delta=0):
		if delta: # need to delete selection otherwise we would fix the current position
			self.delete(self.position, tk.END)
		else: # set position to end so selection starts where textentry ended
			self.position = len(self.get())
		# collect hits
		_hits = []
		for element in self._completion_list:
			if element.lower().startswith(self.get().lower()):  # Match case-insensitively
				_hits.append(element)
		# if we have a new hit list, keep this in mind
		if _hits != self._hits:
			self._hit_index = 0
			self._hits=_hits
		# only allow cycling if we are in a known hit list
		if _hits == self._hits and self._hits:
			self._hit_index = (self._hit_index + delta) % len(self._hits)
		# now finally perform the auto completion
		if self._hits:
			self.delete(0,tk.END)
			self.insert(0,self._hits[self._hit_index])
			self.select_range(self.position,tk.END)

	#-----------------------------------------------------------------------
	# event handler for the keyrelease event on this widget
	#-----------------------------------------------------------------------
	def handle_keyrelease(self, event):
		#print("handle_keyrelease")
		if event.keysym == "BackSpace":
			self.delete(self.index(tk.INSERT), tk.END)
			self.position = self.index(tk.END)
		elif event.keysym == "Left":
			if self.position < self.index(tk.END): # delete the selection
				self.delete(self.position, tk.END)
			else:
				self.position = self.position-1 # delete one character
				self.delete(self.position, tk.END)
		elif event.keysym == "Right":
			self.position = self.index(tk.END) # go to end (no selection)
		elif event.keysym == "Down":
			self.autocomplete(1) # cycle to next hit
		elif event.keysym == "Up":
			self.autocomplete(-1) # cycle to previous hit
		elif len(event.keysym) == 1: # or event.keysym in AutocompleteEntry.tkinter_umlauts:
			self.autocomplete()

#===============================================================================
# File Entry
#===============================================================================
class FileEntry(tk.Entry):
	#-----------------------------------------------------------------------
	def __init__(self, master,
			title=None,
			icon=None,
			initialdir=None,
			filetypes=None,
			multiple=False,
			**kwargs):
		super().__init__(master, **kwargs)

		self.title      = title
		self.initialdir = initialdir
		self.multiple   = multiple
		self.setFiletypes(filetypes)

		if icon is None: icon = icons.get("load")
		self._button = tk.Button(self,
				image=icon,
				command=self.invoke,
				foreground=self["background"],
				relief=tk.FLAT,
				cursor="arrow",
				borderwidth=1,
				padx=-1, pady=-1)
		self._button.pack(side=tk.RIGHT, padx=0, pady=0)
		self.pack_propagate(0)
		Balloon.set(self._button, title)
		self.bind('<Return>',     self.fileChanged)
		self.bind('<FocusOut>',   self.fileChanged)

		# autocompletion information
		self._hits      = []
		self._hit_index = 0
		self._prev      = 0
		self.bind('<KeyRelease>', self.handle_keyrelease)

	#-----------------------------------------------------------------------
	def setFiletypes(self, filetypes):
		self._filetypes = filetypes
		self._pattern   = []
		if filetypes is None: return
		for filetype in filetypes:
			if isinstance(filetype, tuple):
				_, fts = filetype
				if isinstance(fts,str): fts = fts.split()
				for ft in fts:
					if ft and ft!="*":
						self._pattern.append(ft)
			#elif isinstance(filetype, str):
			else:
				assert False

	#-----------------------------------------------------------------------
	def invoke(self, event=None):
		if self["state"] != tk.NORMAL: return
		fn = bFileDialog.askopenfilename(master=self,
				title       = self.title,
				initialdir  = self.directory(),
				initialfile = os.path.basename(self.get()),
				filetypes   = self._filetypes,
				multiple    = self.multiple)
		if fn:
			if isinstance(fn,list) and len(fn)==1: fn = fn[0]

			if isinstance(fn,str):
				filename = tool.relativePath(fn, self.initialdir) \
						if self.initialdir is not None else fn
			else:
				filename = ",".join([f'"{tool.relativePath(x, self.initialdir)}"'
								if self.initialdir is not None else x
							for x in fn])
			self.delete(0,tk.END)
			self.insert(0,filename)
			self.fileChanged()

	#-----------------------------------------------------------------------
	def set(self, fn):
		self.delete(0,tk.END)
		self.insert(0,fn)

	#-----------------------------------------------------------------------
	def filename(self):
		if self.initialdir:
			return os.path.join(self.initialdir, self.get())
		else:
			return self.get()

	#-----------------------------------------------------------------------
	def files(self):
		assert self.multiple
		files = self.get()
		if files and files[0]=='"':
			return [os.path.join(self.initialdir, x) for x in files[1:-1].split('","')]
		else:
			return [files]

	#-----------------------------------------------------------------------
	def directory(self):
		if self.initialdir:
			return os.path.dirname(os.path.join(self.initialdir, self.get()))
		else:
			return os.path.dirname(self.get())

	#-----------------------------------------------------------------------
	def fileChanged(self, event=None):
		self.event_generate("<<FileChanged>>")

	#-----------------------------------------------------------------------
	def insert(self, pos, text):
		super().insert(pos, text)

	#-----------------------------------------------------------------------
	def delete(self, from_, to_):
		super().delete(from_, to_)

	#-----------------------------------------------------------------------
	# autocomplete the Entry, delta may be 0/1/-1 to cycle through
	# possible hits
	#-----------------------------------------------------------------------
	def autocomplete(self, delta=0):
		#print(f"*** autocomplete {self.get()}")
		cursor = self.index(tk.INSERT)
		if delta:
			# need to delete selection otherwise we would fix the current position
			super().delete(cursor, tk.END)

		if self._prev != hash(self.get()) and self.get():
			if delta==0 or not self._hits:
				# collect hits
				self._hits = glob.glob(self.get()+"*")
				if self._pattern:
					self._hits = [x for x in self._hits
							if any([fnmatch(x,p) for p in self._pattern])]
				self._hit_index = 0

			# now finally perform the auto completion
			if self._hits:
				self._hit_index = (self._hit_index + delta) % len(self._hits)
				super().delete(0,tk.END)
				super().insert(0,self._hits[self._hit_index])
				self.icursor(cursor)

		if self._hits:
			self.select_range(cursor,tk.END)
		self._prev = hash(self.get)
		#print(f"<<<<< {self.get()}")

	#-----------------------------------------------------------------------
	# event handler for the keyrelease event on this widget
	#-----------------------------------------------------------------------
	def handle_keyrelease(self, event):
		#print(f"*** handle_keyrelease {event.keysym=} {event.char=}")
		if event.keysym == "BackSpace":
			self.delete(self.index(tk.INSERT), tk.END)
		elif event.keysym in ("Left","Right","Shift_L","Shift_R",
				"Control_L","Control_R","Alt_L","Alt_R"):
			return
		elif event.keysym == "Down":
			self.autocomplete(1) # cycle to next hit
		elif event.keysym == "Up":
			self.autocomplete(-1) # cycle to previous hit
		elif event.keysym in ("Return", "KP_Enter"):
			self.select_clear()
			self.icursor(tk.END)
			self.fileChanged()
		else:
			self.autocomplete()

#===============================================================================
# Auto complete combobox based on ttk
#===============================================================================
class AutocompleteCombobox(ttk.Combobox):
	#-----------------------------------------------------------------------
	def set_completion_list(self, completion_list):
		"""Use our completion list as our drop down selection menu, arrows move through menu."""
		self._completion_list = sorted(completion_list, key=str.lower) # Work with a sorted list
		self._hits = []
		self._hit_index = 0
		self.position = 0
		self.bind('<KeyRelease>', self.handle_keyrelease)
		self['values'] = self._completion_list  # Setup our popup menu

	#-----------------------------------------------------------------------
	def autocomplete(self, delta=0):
		"""autocomplete the Combobox, delta may be 0/1/-1 to cycle through possible hits"""
		if delta: # need to delete selection otherwise we would fix the current position
			self.delete(self.position, tk.END)
		else: # set position to end so selection starts where textentry ended
			self.position = len(self.get())
		# collect hits
		_hits = []
		for element in self._completion_list:
			if element.lower().startswith(self.get().lower()): # Match case insensitively
				_hits.append(element)
		# if we have a new hit list, keep this in mind
		if _hits != self._hits:
			self._hit_index = 0
			self._hits=_hits
		# only allow cycling if we are in a known hit list
		if _hits == self._hits and self._hits:
			self._hit_index = (self._hit_index + delta) % len(self._hits)
		# now finally perform the auto completion
		if self._hits:
			self.delete(0,tk.END)
			self.insert(0,self._hits[self._hit_index])
			self.select_range(self.position,tk.END)

	#-----------------------------------------------------------------------
	def handle_keyrelease(self, event):
		"""event handler for the keyrelease event on this widget"""
		if event.keysym == "BackSpace":
			self.delete(self.index(tk.INSERT), tk.END)
			self.position = self.index(tk.END)
		if event.keysym == "Left":
			if self.position < self.index(tk.END): # delete the selection
				self.delete(self.position, tk.END)
			else:
				self.position = self.position-1 # delete one character
				self.delete(self.position, tk.END)
		if event.keysym == "Right":
			self.position = self.index(tk.END) # go to end (no selection)
		if len(event.keysym) == 1:
			self.autocomplete()
		# No need for up/down, we'll jump to the popup
		# list at the position of the autocompletion

#===============================================================================
# Button that accepts any mouse click left or right with the
# command2=, command3= configuration arguments
#===============================================================================
class AnyButton(tk.Button):
	def __init__(self, master, *args, **kwargs):
		self._command2 = None
		self._command3 = None
		kwargs = self._extraCommand(kwargs)

		super().__init__(master, *args, **kwargs)

		self.bind("<Button-2>", self._press)
		self.bind("<Button-3>", self._press)
		self.bind("<ButtonRelease-2>", self._click2)
		self.bind("<ButtonRelease-3>", self._click3)

	#-----------------------------------------------------------------------
	# Accept two new configuration parameters
	#	command2 and command3
	#-----------------------------------------------------------------------
	def _extraCommand(self, kwargs):
		if "command2" in kwargs:
			self._command2 = kwargs["command2"]
			del kwargs["command2"]
		if "command3" in kwargs:
			self._command3 = kwargs["command3"]
			del kwargs["command3"]
		return kwargs

	#-----------------------------------------------------------------------
	def config(self, **kwargs):
		kwargs = self._extraCommand(kwargs)
		if kwargs:
			super().config(**kwargs)

	#-----------------------------------------------------------------------
	def _press(self, event):
		self.config(relief="sunken")
		self.update_idletasks()

	#-----------------------------------------------------------------------
	def _release(self):
		self.config(relief="raised")
		self.update_idletasks()

	#-----------------------------------------------------------------------
	def _click2(self, event):
		self._release()
		if self._command2: self._command2()

	#-----------------------------------------------------------------------
	def _click3(self, event):
		self._release()
		if self._command3: self._command3()

#===============================================================================
# Auto Scroll Bar
# Author: Fredrik Lundh <www.pythonware.com>
#===============================================================================
class AutoScrollbar(ttk.Scrollbar):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.prev = (0,-1.0)		# remember previous settings to avoid infinite loop
		self.kw = {}

	# ----------------------------------------------------------------------
	def forget(self):
		if self._method==0:
			self.grid_remove()
		else:
			self.pack_forget()

	#-----------------------------------------------------------------------
	def restore(self):
		if self._method==0:
			ttk.Scrollbar.grid(self, **self.kw)
		else:
			ttk.Scrollbar.pack(self, **self.kw)

	# ----------------------------------------------------------------------
	# a scrollbar that hides itself if it's not needed.
	# ----------------------------------------------------------------------
	def set(self, lo, hi):
		flo = float(lo)
		fhi = float(hi)
		try:
			g = self.get()
		except tk.TclError:
			return
		# Avoid infinite loop
		if self.prev[0]==flo and self.prev[1]==fhi:
			return
#		if float(g[0])<=0.001 and float(g[1])<=0.001:
#			self.forget()
#			return
		if abs(flo-float(g[0]))<=0.001 and abs(fhi-float(g[1]))<=0.001:
			return
		if flo <= 0.001 and fhi >= 0.999:
			self.forget()
		elif flo > 0.001 or fhi < 0.999:
			self.restore()
		ttk.Scrollbar.set(self, lo, hi)
		self.prev = g

	# ----------------------------------------------------------------------
	def grid(self, **kw):
		self._method = 0
		ttk.Scrollbar.grid(self, **kw)
		self.kw = kw

	# ----------------------------------------------------------------------
	def pack(self, **kw):
		self._method = 1
		ttk.Scrollbar.pack(self, **kw)
		self.kw = kw

	# ----------------------------------------------------------------------
	def place(self, **kw):
		raise tk.TclError("cannot use place with this widget")

#===============================================================================
# ProgressBar Canvas
#===============================================================================
class ProgressBar(tk.Canvas):
	def __init__(self, master=None, **kw):
		super().__init__(master, **kw)
		#self.config(background="DarkGray")
		self.currBox = self.create_rectangle(0, 0, 0, 0,
					fill='Orange',
					width=0)
		self.doneBox = self.create_rectangle(0, 0, 0, 0,
					fill='DarkGreen',
					width=0)
		self.text = self.create_text(0,0,
					text="",
					fill="White",
					anchor=tk.CENTER,
					justify=tk.CENTER)
		self.auto = True
		self.showTime = True

		self.bind('<Configure>', self.draw)
		self.setLimits()

	# ----------------------------------------------------------------------
	def setAuto(self, auto):
		self.auto = auto

	# ----------------------------------------------------------------------
	def setShowTime(self, b):
		self.showTime = b

	# ----------------------------------------------------------------------
	def setStartTime(self, t0):
		self.t0 = t0

	# ----------------------------------------------------------------------
	def setLimits(self, low=0, high=100, step=1):
		self.low    = float(low)
		self.high   = float(high)
		self.length = float(high-low)
		self.step   = float(step)
		self.done   = float(low)
		self.now    = float(low)
		self.t0     = time.time()
		self.msg    = ""

	# ----------------------------------------------------------------------
	def setProgress(self, now, done=None, txt=None):
		self.now = now
		if self.now < self.low:
			self.now = self.low
		elif self.now > self.high:
			self.now = self.high

		if done is None:
			self.done = now - self.step
		else:
			self.done = done

		if self.done < self.low:
			self.done = self.low
		elif self.done > self.high:
			self.done = self.high

		# calculate remaining time
		dt = time.time() - self.t0
		p  = now - self.low
		if p>0:
			tot = dt/p*(self.high-self.low)
		else:
			tot = 0.0

		# elapsed time
		dh,s  = divmod(int(dt),3600)
		dm,ds = divmod(s,60)
		if dh > 0:
			elapsedTxt = f"{dh}h{dm:02d}m"
		elif dm > 0:
			elapsedTxt = f"{dm}m{ds:02d}s"
		else:
			elapsedTxt = f"{ds}s"

		# total time
		th,s  = divmod(int(tot),3600)
		tm,ts = divmod(s,60)
		if th > 0:
			totalTxt = f"Tot: {th}h{tm:02d}m"
		elif tm > 0:
			totalTxt = f"Tot: {tm}m{ts:02d}s"
		else:
			totalTxt = f"Tot: {ts}s"

		# remain time
		remain = tot - dt
		if remain>0:
			rh,s  = divmod(int(remain),3600)
			rm,rs = divmod(s,60)
			if rh > 0:
				remainTxt = f"Rem: {rh}h{rm:02d}m"
			elif rm > 0:
				remainTxt = f"Rem: {rm}m{rs:02.1f}s"
			else:
				remainTxt = f"Rem: {rs}s"
		else:
			remainTxt = ""

		self.draw()
		if txt is not None:
			self.setText(txt)

		elif self.auto:
			if self.showTime:
				self.autoText(f"[{elapsedTxt} {totalTxt} {remainTxt}]")
			else:
				self.autoText("")

	# ----------------------------------------------------------------------
	def clear(self):
		self.setProgress(0, 0)

	# ----------------------------------------------------------------------
	def setText(self, txt):
		self.itemconfig(self.text, text=txt)

	# ----------------------------------------------------------------------
	def configText(self, **args):
		self.itemconfig(self.text, **args)

	# ----------------------------------------------------------------------
	def autoText(self, tmsg):
		completed = self.done - self.low
		if self.low != 0:
			low = f"{self.low} - "
		else:
			low = ""
		if self.step == 1:
			self.msg = f"Current: {self.now:.9g} [{low}{self.high:.9g}]" \
				f" Completed: {int((100*completed)/self.length)}% {tmsg}"
		else:
			self.msg = f"Current: {self.now-self.step+1:.9g}-{self.now:.9g} [{low}{self.high:.9g}]" \
				f" Completed: {int((100*completed)/self.length)}% {tmsg}"
		self.setText(self.msg)

	# ----------------------------------------------------------------------
	def getProgress(self):
		return (self.done, self.now)

	# ----------------------------------------------------------------------
	def draw(self, event=None):
		width  = self.winfo_width()
		height = self.winfo_height()

		if self.length==0: return
		wn = int(width * (self.now  - self.low) / self.length)
		wd = int(width * (self.done - self.low) / self.length)
		if wd >= wn: wd = wn - 1

		self.coords(self.currBox, 0, 0, wn, height)
		self.coords(self.doneBox, 0, 0, wd, height)

		if self.itemcget(self.text, "justify") == tk.CENTER:
			self.coords(self.text, width/2, height/2)
		else:
			self.coords(self.text, 1,height/2)

#===============================================================================
# Static class providing the search textbox toplevel window
# with common functionality
#===============================================================================
class SearchBox(tk.Toplevel):
	TIME_THRESHOLD = 1000	# ms time threshold for key presses
	top     = None		# toplevel window
	label   = None		# Label widget to update
	search  = ""		# current search string
	orig    = ""		# original search string
	time    = 0		# last event time

	additionalChar = "-+._:$%#*"
	ignoreNonAlpha = True	# Ignore non-alpha characters
	ignoreCase     = True	# Ignore case of letters

	# ----------------------------------------------------------------------
	# Add a new keyboard event
	# @return True if the calling listbox needs update
	# ----------------------------------------------------------------------
	@staticmethod
	def addkey(event):
		# Shift key -> ignore them
		if event.keysym in ("Shift_L","Shift_R"):
			return False
		elif not event.char:
			SearchBox.time = 0
			return False

		ch = event.char.upper() if SearchBox.ignoreCase else event.char

		# Delete search
		if event.keysym in ("Delete","Escape","Return","KP_Enter"):
			SearchBox.reset()
		# Search Again	Ctrl-G
		elif event.char=='\007':
			pass
		# Backspace
		elif event.keysym == "BackSpace":
			SearchBox.search = SearchBox.search[:-1]
			SearchBox.orig   = SearchBox.orig[:-1]
		# Ignore non-printable characters
		elif SearchBox.ignoreNonAlpha and \
		     not (ch.isalnum() or
		     SearchBox.additionalChar.find(ch)>=0):
			return False
		# Timeout
		elif event.time - SearchBox.time > SearchBox.TIME_THRESHOLD:
			# Start a new search
			SearchBox.search = ch
			SearchBox.orig   = event.char
		else:
			SearchBox.search += ch
			SearchBox.orig   += event.char
		SearchBox.time = event.time
		return True

	#-----------------------------------------------------------------------
	# @return string 'text' as upper case when case is to be ignored
	#-----------------------------------------------------------------------
	@staticmethod
	def case(text):
		return text.upper() if SearchBox.ignoreCase else text

	#-----------------------------------------------------------------------
	# @return True if 'text' matches search pattern
	#-----------------------------------------------------------------------
	@staticmethod
	def match(text):
		if not text: return False
		text = SearchBox.case(text)
		if SearchBox.ignoreNonAlpha:
			for pos,a in enumerate(text):
				if a.isalnum() or SearchBox.additionalChar.find(a)>=0:
					break
		else:
			pos = 0
		return SearchBox.search == text[pos:pos+len(SearchBox.search)]

	#-----------------------------------------------------------------------
	# Reset search
	#-----------------------------------------------------------------------
	@staticmethod
	def reset():
		"""Reset search string"""
		SearchBox.search = ""
		SearchBox.orig   = ""
		SearchBox.time   = 0
		if SearchBox.top is not None:
			try:
				SearchBox.top.withdraw()
			except tk.TclError:
				SearchBox.top = None

	#-----------------------------------------------------------------------
	# Hide toplevel search window
	#-----------------------------------------------------------------------
	@staticmethod
	def hide(event=None):
		if SearchBox.top is not None:
			try: SearchBox.top.withdraw()
			except: pass

	# ----------------------------------------------------------------------
	# Show toplevel search window
	# ----------------------------------------------------------------------
	@staticmethod
	def show(widget):
		if SearchBox.top is None:
			SearchBox.top = tk.Toplevel()
			SearchBox.top.overrideredirect(1)
			SearchBox.label = tk.Label(SearchBox.top,
						anchor=tk.E,
						relief=tk.SOLID,
						background="Yellow",
						takefocus=False,
						borderwidth=1)
			SearchBox.label.pack(fill=tk.BOTH)

		if SearchBox.orig == "":
			SearchBox.hide()
			return

		SearchBox.label["text"] = SearchBox.orig

		# Guess position
		x = widget.winfo_rootx() + widget.winfo_width() \
				       - SearchBox.label.winfo_width()
		y = widget.winfo_rooty() + widget.winfo_height()-12
		SearchBox.top.wm_geometry(f"+{x}+{y}")
		SearchBox.top.deiconify()
		SearchBox.top.lift()
		SearchBox.top.update_idletasks()

#===============================================================================
# Extended Listbox
#===============================================================================
class ExListbox(tk.Listbox):
	"""Listbox that allows keyboard scanning, and a popup menu"""

	#-----------------------------------------------------------------------
	def __init__(self, master, **kw):
		super().__init__(master, **kw)
		self._single = kw.get('selectmode','') in [tk.SINGLE, tk.BROWSE]
		self.bind('<Key>',	self._key)
		self.bind('<Home>',	lambda e,s=self:s._scrollTo(0))
		self.bind('<Prior>',	lambda e,s=self:s._scrollTo(-1, tk.PAGES))
		self.bind('<Next>',	lambda e,s=self:s._scrollTo( 1, tk.PAGES))
		self.bind('<End>',	lambda e,s=self:s._scrollTo(tk.END))
		self.bind('<FocusOut>',	SearchBox.hide)
		self.bind('<Unmap>',	SearchBox.hide)
		self.bind('<<Cut>>',	self.copy)
		self.bind('<<Copy>>',	self.copy)
		self.bind('<<SelectInvert>>', self.selectInvert)
		self.bind('<Control-Key-space>',self.popupMenu)

		if not self._single:
			self.bind('<Control-Key-a>',	self.selectAll)
			self.bind('<Control-Shift-A>',	self.selectClear)
			self.bind('<Button-3>',		self.popupMenu)

		# User variables to modify
		self._usermenu      = None	# Assign a user-popup menu
						# Should be a list with tuples
						#  in the form:
						#  (label, underline, command)
						# or None for separator

	# ----------------------------------------------------------------------
	def setPopupMenu(self, menu=None):
		"""Setup a popup menu list it should be in the form
		   [ (label, underline, command), ... ]"""
		self._usermenu = menu
		self.bind('<Button-3>',		self.popupMenu)
		self.bind('<Control-Key-space>',self.popupMenu)

	# ----------------------------------------------------------------------
	# Handle key events for quick searching
	# ----------------------------------------------------------------------
	def _key(self, event):
		"""handle key events for quick searching"""
		if SearchBox.addkey(event):
			SearchBox.show(self)
		else:
			return
		item = self.index(tk.ACTIVE)
		# Search Again	Ctrl-G?
		if event.char=='\007': item += 1
		for loop in range(2):
			if loop>0: item = 0
			for i in range(item, self.size()):
				if SearchBox.match(str(self.get(i))):
					if self._single:
						self.selection_clear(0, tk.END)
						self.selection_set(i)
					self.activate(i)
					self.see(i)
					self.event_generate("<<ListboxSelect>>")
					return "break"

	# ----------------------------------------------------------------------
	# Create the popup menu
	# ----------------------------------------------------------------------
	def popupMenu(self, event):
		"""Create popup menu with default actions"""
		if self["state"] == tk.DISABLED: return
		self.focus_set()
		menu=tk.Menu(self, tearoff=0)
		if self._usermenu:
			for entry in self._usermenu:
				if entry is None:
					menu.add_separator()
				else:
					name,und,cmd = entry[:3]
					icon = entry[3] if len(entry)>3 else None
					acc  = entry[4] if len(entry)>4 else ""
					menu.add_command(label=name, underline=und,
							image=icon, compound=tk.LEFT,
							accelerator=acc,
							command=cmd)
			if not self._single: menu.add_separator()

		if not self._single:
			menu.add_command(label='All', underline=0,
					image=icons.get("select_all"),
					compound=tk.LEFT,
					command=self.selectAll)
			menu.add_command(label='Clear', underline=0,
					image=icons.get("select_none"),
					compound=tk.LEFT,
					command=self.selectClear)
			menu.add_command(label='Invert', underline=0,
					image=icons.get("select_invert"),
					compound=tk.LEFT,
					command=self.selectInvert)

		menu.tk_popup(event.x_root, event.y_root)
		return "break"

	# ----------------------------------------------------------------------
	# Selection
	# ----------------------------------------------------------------------
	def selectAll(self, event=None):
		"""Select all items"""
		self.selection_set(0, tk.END)
		self.event_generate("<<ListboxSelect>>")
		return "break"

	# ----------------------------------------------------------------------
	def selectClear(self, event=None):
		"""Selection Clear"""
		self.selection_clear(0, tk.END)
		self.event_generate("<<ListboxSelect>>")
		return "break"

	# ----------------------------------------------------------------------
	def selectInvert(self, event=None):
		"""Invert selection"""
		for i in range(self.size()):
			if self.select_includes(i):
				self.selection_clear(i)
			else:
				self.selection_set(i)
		self.event_generate("<<ListboxSelect>>")
		return "break"

	# ----------------------------------------------------------------------
	# select active and selected items
	# ----------------------------------------------------------------------
	def selectSaved(self, save, default=None):
		"""selected the saved items.
		   If list has changed then selected the default item"""
		self.selection_clear(0,tk.END)

		if save is not None:
			self.activate(save[0])
			for s in save[1]:
				self.selection_set(s)
			self.see(save[0])

		if default is not None:
			if save is None or \
			   (save is not None and save[0] >= self.size()):
				if isinstance(default, tuple):
					self.selection_set(default[0], default[1])
					self.activate(default[0])
				else:
					self.selection_set(default)
					self.activate(default)
		self.event_generate("<<ListboxSelect>>")

	# ----------------------------------------------------------------------
	def _scrollTo(self, pos, unit=None):
		if unit:
			self.yview_scroll(pos, unit)
		else:
			if self._single:
				self.selection_clear(0, tk.END)
				self.selection_set(pos)
			self.activate(pos)
			self.see(pos)
		self.event_generate("<<ListboxSelect>>")
		return 'break'

	# ----------------------------------------------------------------------
	# Change the value of a list item
	# and return the value of the old one
	# ----------------------------------------------------------------------
	def set(self, index, value):
		"""Set/Change the value of a list item"""
		try:
			sel = self.selection_includes(index)
			act = self.index(tk.ACTIVE)
			self.delete(index)
		except tk.TclError:
			return
		self.insert(index, value)
		if sel: self.selection_set(index)
		self.activate(act)
		#self.event_generate("<<ListboxSelect>>")

	# ----------------------------------------------------------------------
	# Swap two items in the list
	# ----------------------------------------------------------------------
	def swap(self, a, b):
		"""Swap two items in the list"""
		if a>b: a, b = b, a

		at = self.get(a)
		bt = self.get(b)

		self.delete(b)
		self.delete(a)

		self.insert(a, bt)
		self.insert(b, at)

	# ----------------------------------------------------------------------
	# Move up select items by one
	# ----------------------------------------------------------------------
	def moveUp(self):
		"""Move selected items up"""
		for i in map(int,self.curselection()):
			if i==0: continue
			prev = i-1
			if not self.selection_includes(prev):
				act = self.index(tk.ACTIVE)
				self.swap(prev,i)
				self.selection_set(prev)
				if act == i: self.activate(prev)
		self.event_generate("<<ListboxSelect>>")

	# ----------------------------------------------------------------------
	# Move down select items by one
	# ----------------------------------------------------------------------
	def moveDown(self):
		"""Move selected items down"""
		sz  = self.size()-1
		lst = list(map(int,self.curselection()))
		lst.reverse()
		for i in lst:
			if i >= sz: continue
			next_ = i+1
			if not self.selection_includes(next_):
				act = self.index(tk.ACTIVE)
				self.swap(i,next_)
				self.selection_set(next_)
				if act == i: self.activate(next_)
		self.event_generate("<<ListboxSelect>>")

	# ----------------------------------------------------------------------
	def deleteByName(self, item):
		"""delete entry by name"""
		act = self.index(tk.ACTIVE)
		for i in range(self.size()-1,-1,-1):
			it = self.get(i)
			if it == item:
				self.delete(i)
		self.activate(act)

	# ----------------------------------------------------------------------
	# Fill the listbox
	# ----------------------------------------------------------------------
	def fill(self, items=None):
		self.delete(0,tk.END)
		for item in items: self.insert(tk.END, item)

	# ----------------------------------------------------------------------
	# Copy current elements to clipboard
	# ----------------------------------------------------------------------
	def copy(self, event=None):
		sel = self.curselection()
		if not sel: return
		items = []
		for i in sel:
			items.append(self.get(i))
		self.clipboard_clear()
		self.clipboard_append("\n".join(items))

#===============================================================================
# Search Listbox
# A listbox that the list is narrowing down to the matching items
#===============================================================================
class SearchListbox(ExListbox):
	def __init__(self, master, **kw):
		super().__init__(master, **kw)
		self.prefixSearch = False
		self._items = []
		self._pos   = []

	# ----------------------------------------------------------------------
	# Fill the listbox
	# ----------------------------------------------------------------------
#	def fill(self, items=None):
#		del self._items[:]
#		if items is None:
#			for item in tk.Listbox.get(self,0,tk.END):
#				self._items.append(item)
#		else:
#			self.delete(0,tk.END)
#			for item in items:
#				self._items.append(item)
#				self.insert(tk.END, item)
#		self._pos = range(len(self._items))

	# ----------------------------------------------------------------------
	def reset(self):
		if self._items and SearchBox.search:
			SearchBox.reset()
			tk.Listbox.delete(self, 0, tk.END)
			for item in self._items:
				tk.Listbox.insert(self, tk.END, item)
			del self._items[:]
			del self._pos[:]

	# ----------------------------------------------------------------------
	def _key(self, event):
		"""handle key events for quick searching"""
		if SearchBox.addkey(event):
			SearchBox.show(self)
			if event.keysym not in ("Escape","Return","KP_Enter"):
				self.filter(refill=event.keysym in ("Delete","BackSpace"))

	# ----------------------------------------------------------------------
	def filter(self, search=None, refill=False):
		activepos = -1		# previous active match
		searchpos = -1		# exact match
		prefixpos = -1		# closest prefix match
		active    = tk.Listbox.get(self,tk.ACTIVE)
		if search is not None:
			SearchBox.search = SearchBox.case(search)
		search = SearchBox.search
		prefix = self.prefixSearch
		if search and search[0]=="*":
			search = search[1:]
			prefix = not prefix

		# Fill up the list of items
		if not self._items:
			for item in tk.Listbox.get(self,0,tk.END):
				self._items.append(item)
			self._pos = list(range(len(self._items)))

		# if Search string is empty, fill the entire list
		if not search:
			tk.Listbox.delete(self, 0, tk.END)
			for i,item in enumerate(self._items):
				if active == item: activepos = i
				if search == item: searchpos = i
				elif item.startswith(search): prefixpos = i
				tk.Listbox.insert(self, tk.END, item)
			self._pos = list(range(len(self._items)))

		# Backspace removes one character then we need to expand the list
		elif refill or not self._pos:
			# FIXME I could find the correct position and insert it
			# instead of delete all and repopulate
			tk.Listbox.delete(self, 0, tk.END)
			del self._pos[:]
			for i,item in enumerate(map(str,self._items)):
				if prefix:
					if SearchBox.ignoreCase:
						uitem = item.upper()
						if uitem.startswith(search):
							if active ==  item: activepos = tk.Listbox.size(self)
							if search == uitem: searchpos = tk.Listbox.size(self)
							elif item.startswith(search):
								prefixpos = tk.Listbox.size(self)
							tk.Listbox.insert(self, tk.END, item)
							self._pos.append(i)
					else:
						if item.startswith(search):
							if active == item: activepos = tk.Listbox.size(self)
							if search == item: searchpos = tk.Listbox.size(self)
							elif item.startswith(search):
								prefixpos = tk.Listbox.size(self)
							tk.Listbox.insert(self, tk.END, item)
							self._pos.append(i)
				else:
					if SearchBox.ignoreCase:
						uitem = item.upper()
						if uitem.find(search)>=0:
							if active ==  item: activepos = tk.Listbox.size(self)
							if search == uitem: searchpos = tk.Listbox.size(self)
							elif item.startswith(search):
								prefixpos = tk.Listbox.size(self)
							tk.Listbox.insert(self, tk.END, item)
							self._pos.append(i)
					else:
						if item.find(search)>=0:
							if active == item: activepos = tk.Listbox.size(self)
							if search == item: searchpos = tk.Listbox.size(self)
							elif item.startswith(search):
								prefixpos = tk.Listbox.size(self)
							tk.Listbox.insert(self, tk.END, item)
							self._pos.append(i)
		else:
			# FIXME I could use the fnmatch or re to allow * and ? as pattern
			# If a new character added then shrink the existing list
			# Scan in reverse order
			for i in range(tk.Listbox.size(self)-1, -1, -1):
				item = SearchBox.case(str(tk.Listbox.get(self, i)))
				if active == item: activepos = i
				if search == item: searchpos = i
				elif item.startswith(search): prefixpos = i

				if (prefix and not item.startswith(search)) or \
				   item.find(search)<0:
					tk.Listbox.delete(self, i)
					del self._pos[i]
					if i<activepos: activepos -= 1
					if i<searchpos: searchpos -= 1
					if i<prefixpos: prefixpos -= 1

		if prefixpos >= 0: activepos = prefixpos
		if searchpos >= 0: activepos = searchpos

		tk.Listbox.selection_clear(self, 0, tk.END)
		tk.Listbox.selection_set(self, activepos)
		tk.Listbox.activate(self, activepos)
		tk.Listbox.see(self, activepos)

	# ----------------------------------------------------------------------
	def insert(self, index, *elements):
		del self._items[:]
		return tk.Listbox.insert(self, index, *elements)

	# ----------------------------------------------------------------------
	def delete(self, first, last=None):
		del self._items[:]
		return tk.Listbox.delete(self, first, last)

	# ----------------------------------------------------------------------
	def curselection(self):
		if self._items:
			return [self._pos[int(x)] for x in tk.Listbox.curselection(self)]
		else:
			return tk.Listbox.curselection(self)

	# ----------------------------------------------------------------------
	# FIXME needs work to handle, tk.ACTIVE, tk.END...
	# ----------------------------------------------------------------------
	def get(self, first, last=None):
		#log.say("SearchListbox.get",first,type(first),last,type(last))
		if not self._items:
			return tk.Listbox.get(self, first, last)

		elif first == tk.ACTIVE:
			return tk.Listbox.get(self, first, last)

		elif last is None:
			return self._items[first]

		elif last == tk.END:
			last = len(self._items)

		else:
			last = int(last)+1

		if not self._items: return ""
		return self._items[int(first):last]

#===============================================================================
# Extension of tree view with sorting capabilities
# see the treeview.tcl
# https://stackoverflow.com/questions/1966929/tk-treeview-column-sort
#===============================================================================
class Treeview(ttk.Treeview):
	def __init__(self, master, *kw, **args):
		super().__init__(master, *kw, **args)
		self._sortConvert = {}	# sort convert function e.g. float()
		#self.bind("<Key-Escape>",	self.escape)
		#self.unbind("<B1-Motion>")
		self.bind("<B1-Motion>",	self.b1motion)
		#self.bind("<ButtonRelease-1>",	self.release)
		#self.bind("<Up>",		self.up)
		#self.bind("<Down>",		self.down)
		self.bind("<Shift-Key-Up>",	self.selectUp)
		self.bind("<Shift-Key-Down>",	self.selectDown)
		self.bind("<KeyRelease>",	self.keyRelease)
		#self.bind("<KeyRelease-Shift_L>", self.keyRelease)
		#self.bind("<KeyRelease-Shift_R>", self.keyRelease)
		#self.bind("<KeyRelease-ISO_Next_Group>", self.keyRelease)
		self.bind('<Key-space>',	self.toggleOpen)
		self.bind('<Control-Key-space>',self.popupMenu)

		single = args.get('selectmode','') in [tk.SINGLE, tk.BROWSE]
		if not single:
			self.bind('<Control-Key-a>',	self.selectAll)
			self.bind('<Control-Shift-A>',	self.selectClear)
			self.bind('<Button-3>',		self.popupMenu)

		self._dnd       = None
		self._dnd_label = None
		self._dnd_item  = None
		self._dnd_event = None
		self._usermenu  = None
		self._anchor    = ""

	# ----------------------------------------------------------------------
	def tag_names(self):
		return self.tk.call(self._w, "tag", "names")

	# ----------------------------------------------------------------------
	def tag_add(self, tag, items):
		if len(items) == 1 and isinstance(items[0], (tuple, list)):
			items = items[0]
		self.tk.call(self._w, "tag", "add", tag, items)

	# ----------------------------------------------------------------------
	def tag_remove(self, tag, items=None):
		if items is None:
			self.tk.call(self._w, "tag", "remove", tag)
		else:
			if len(items) == 1 and isinstance(items[0], (tuple, list)):
				items = items[0]
			self.tk.call(self._w, "tag", "remove", tag, items)

	# ----------------------------------------------------------------------
	# @return dir of until current item
	# ----------------------------------------------------------------------
	def dir(self, item):
		path = []
		while item:
			path.insert(0, item)
			item = self.parent(item)
		return path

	# ----------------------------------------------------------------------
	# @return path of until current item
	# ----------------------------------------------------------------------
	def path(self, item):
		path = []
		while item:
			path.insert(0, self.item(item,"text"))
			item = self.parent(item)
		return path

	# ----------------------------------------------------------------------
	# Select all descendants of a branch
	# @param item	branch to select
	# ----------------------------------------------------------------------
	def selectBranch(self, item):
		if isinstance(item,(list,tuple)):
			for i in item:
				self.selectBranch(i)
		else:
			children = self.get_children(item)
			self.selection_add(children)
			self.selectBranch(children)

	# ----------------------------------------------------------------------
	# Select all descendants of branches if only the root-element of the
	# branch is selected and not any descendant
	# ----------------------------------------------------------------------
	def selectBranches(self):
		selection = set(self.selection())
		for item in reversed(self.selection()):
			if not self.isChildSelected(item, selection):
				self.selectBranch(item)

	# ----------------------------------------------------------------------
	# Select everything
	# ----------------------------------------------------------------------
	def selectAll(self, event=None):
		self.selectBranch("")

	# ----------------------------------------------------------------------
	# Clear the selection
	# ----------------------------------------------------------------------
	def selectClear(self, event=None):
		self.selection_remove(self.selection())
	selection_clear = selectClear

	# ----------------------------------------------------------------------
	# Invert selection
	# ----------------------------------------------------------------------
	def selectInvert(self, event=None):
		def selInvert(pid):
			for child in self.get_children(pid):
				selInvert(child)
			self.selection_toggle(pid)
		selInvert("")

	# ----------------------------------------------------------------------
	# @return true if item is selected
	# ----------------------------------------------------------------------
	def isselected(self, item):
		return item in self.selection()

	# ----------------------------------------------------------------------
	# @return true if any of the descendants of the branch is selected
	# ----------------------------------------------------------------------
	def isChildSelected(self, item, selection=None):
		if selection is None: selection = set(self.selection())
		for child in self.get_children(item):
			if child in selection or \
			   self.isChildSelected(child, selection):
				return True
		return False

	# ----------------------------------------------------------------------
	# Setup a popup menu list it should be in the form
	#  [ (label, underline, command), ... ]
	# ----------------------------------------------------------------------
	def setPopupMenu(self, menu=None):
		self._usermenu = menu
		self.bind('<Button-3>',		self.popupMenu)
		self.bind('<Control-Key-space>',self.popupMenu)

	# ----------------------------------------------------------------------
	# Create the popup menu
	# ----------------------------------------------------------------------
	def popupMenu(self, event):
		"""Create popup menu with default actions"""
		#if self["state"] == tk.DISABLED: return
		self.focus_set()
		menu=tk.Menu(self, tearoff=0)
		single = self.cget('selectmode') in [tk.SINGLE, tk.BROWSE]
		if self._usermenu:
			for entry in self._usermenu:
				if entry is None:
					menu.add_separator()
				else:
					name,und,cmd = entry[:3]
					icon = entry[3] if len(entry)>3 else None
					acc  = entry[4] if len(entry)>4 else ""
					menu.add_command(label=name, underline=und,
							image=icon, compound=tk.LEFT,
							accelerator=acc,
							command=cmd)
			if not single: menu.add_separator()

		if not single:
			menu.add_command(label='All', underline=0,
					image=icons.get("select_all"),
					compound=tk.LEFT,
					command=self.selectAll)
			menu.add_command(label='Clear', underline=0,
					image=icons.get("select_none"),
					compound=tk.LEFT,
					command=self.selectClear)
			menu.add_command(label='Invert', underline=0,
					image=icons.get("select_invert"),
					compound=tk.LEFT,
					command=self.selectInvert)

		menu.tk_popup(event.x_root, event.y_root)
		return "break"

	#-----------------------------------------------------------------------
	# Open close all selected items
	#-----------------------------------------------------------------------
	def toggleOpen(self, menu=None):
		open_ = None
		for item in self.selection():
			if open_ is None:
				open_ = not self.item(item, "open")
			self.item(item, open=open_)
		return "break"

	# ----------------------------------------------------------------------
	# Select by clicking and dragging the B1 mouse
	# or do DND if active
	# ----------------------------------------------------------------------
	def b1motion(self, event):
		if self._dnd is not None:
			self._dnd.on_motion(event)
			return "break"
		# If selected item is the last one focused
		# then ignore, otherwise Select is selecting everything
		item = self.identify_row(event.y)
		if item=="":
			children = self.get_children()
			if not children: return
			try:
				x,y,w,h = self.bbox(children[0])
			except ValueError:
				return
			if event.y < y:
				item = children[0]
				event.y = y
			else:
				x,y,w,h = self.bbox(children[-1])
				if event.y > y+h:
					item = children[-1]
					event.y = y+h
					while True:
						children = self.get_children(item)
						if not children: break
						try:
							x,y,w,h = self.bbox(children[-1])
						except:
							break
						item = children[-1]
						event.y = y+h
		if item == self.focus(): return
		if event.state & STATE.CONTROL:
			# copy state of focused item
			if self.isselected(self.focus()):
				self.selection_add(item)
			else:
				self.selection_remove(item)
			return "break"
		else:
			self.tk.call("ttk::treeview::Select", self._w, event.x, event.y, "extend")
			return "break"

	# ----------------------------------------------------------------------
	#def release(self, event):
	#	self.tk.call("ttk::treeview::Release", self._w, event.x, event.y)
	#	self._anchor = self.focus()
	#	return "break"

	## ----------------------------------------------------------------------
	#def up(self, event):
	#	self.tk.call("ttk::treeview::Keynav", self._w, "up")
	#	self._anchor = self.focus()
	#	return "break"

	## ----------------------------------------------------------------------
	#def down(self, event):
	#	self.tk.call("ttk::treeview::Keynav", self._w, "down")
	#	self._anchor = self.focus()
	#	return "break"

	# ----------------------------------------------------------------------
	# add vi bindings
	# ----------------------------------------------------------------------
	def bindVi(self):
		pass
		#self.bind("<Key-k>",	lambda e
		#self.event_add("<Key-Up>", "<Key-k>")

	# ----------------------------------------------------------------------
	# Enable SearchBox for widget
	# ----------------------------------------------------------------------
	def bindSearch(self, column=None):
		self.bind('<FocusOut>',	SearchBox.hide)
		self.bind('<Unmap>',	SearchBox.hide)
		self.bind('<Key>',	self._key)
		self._searchColumn = column

	# ----------------------------------------------------------------------
	def _key(self, event):
		"""handle key events for quick searching"""
		#---------------------------------------------------------------
		# @return next item either moving down the children or
		# up the parent
		#---------------------------------------------------------------
		def _next(item):
			# Check children
			children = self.get_children(item)
			if children: return children[0]
			# Check next item or next parent
			while True:
				nxt = self.next(item)
				if nxt: return nxt
				item = self.parent(item)
				if item=="": return ""

		#---------------------------------------------------------------
		# @return text or column of item
		#---------------------------------------------------------------
		def _text(item):
			if self._searchColumn is None:
				return str(self.item(item,"text"))
			else:
				try:
					return str(self.item(item, "values")[self._searchColumn])
				except IndexError:
					return ""
		#---------------------------------------------------------------
		if SearchBox.addkey(event):
			SearchBox.show(self)
		else:
			return
		item = self.focus()
		# Search Again	Ctrl-G?
		if event.char=='\007': item = _next(item)
		for loop in range(2):
			if loop>0: item = _next("")
			while item != "":
				if SearchBox.match(_text(item)):
					self.focus(item)
					self.selection_set(item)
					self.see(item)
					return 'break'
				item = _next(item)

	# ----------------------------------------------------------------------
	def keyRelease(self, event):
		#print(f"{event.keysym=} {event.keycode=} {event.char=}")
		if event.keycode in (50,62):	# Shift_L, Shift_R
			self._anchor = ""

	# ----------------------------------------------------------------------
	def selectUp(self, event):
		if not self._anchor:
			self._anchor = self.focus()
		self.tk.call("ttk::treeview::Keynav", self._w, "up")
		if self.focus() != self._anchor:
			self.tk.call("ttk::treeview::SelectOp", self._w, self._anchor, "extend")
		return "break"

	# ----------------------------------------------------------------------
	def selectDown(self, event):
		if not self._anchor:
			self._anchor = self.focus()
		self.tk.call("ttk::treeview::Keynav", self._w, "down")
		if self.focus() != self._anchor:
			self.tk.call("ttk::treeview::SelectOp", self._w, self._anchor, "extend")
		return "break"

	# ----------------------------------------------------------------------
	# Sort column
	# ----------------------------------------------------------------------
	def sortColumn(self, col, convertFunc=None):
		self._sortConvert[col] = convertFunc
		self.heading(col,
			command=lambda s=self,col=col:self._sortColumn(col, False))

	# ----------------------------------------------------------------------
	def _sortColumn(self, col, reverse):
		l = []
		cnv = self._sortConvert.get(col)
		for k in self.get_children(''):
			if col=="#0":
				v = self.item(k,"text")
			else:
				v = self.set(k,col)	# set = get with only column
			if cnv:
				try:
					v = cnv(v)
				except ValueError:
					pass
			l.append((v,k))
		#l = [(self.set(k, col), k)
		#	for k in self.get_children('')]
		l.sort(reverse=reverse)

		# rearrange items in sorted positions
		for index, (val, k) in enumerate(l):
			self.move(k, '', index)

		# reverse sort next time
		heading = self.heading(col,"text")
		if heading and heading[-1] in ("\N{BLACK UP-POINTING TRIANGLE}",
						"\N{BLACK DOWN-POINTING TRIANGLE}"):
			heading = heading[:-2]
		heading += ' '
		heading += "\N{BLACK DOWN-POINTING TRIANGLE}" if reverse \
				else "\N{BLACK UP-POINTING TRIANGLE}"
		self.heading(col, text=heading,
				command=lambda s=self,col=col:self._sortColumn(col, not reverse))

	# ----------------------------------------------------------------------
	# Start DND
	# ----------------------------------------------------------------------
	def dnd_start(self, event):
		self._dnd = dnd.dnd_start(self, event)
		if not self._dnd: return False
		selection = self.selection()
		if len(selection)==1:
			self._dnd_label = self.item(selection[0],"text")
		else:
			self._dnd_label = f"{self.item(selection[0],'text')}"\
				f" .. {self.item(selection[-1],'text')}"
		self._dnd_event = "<<Copy>>" if event.state & STATE.CONTROL else "<<Cut>>"
		return True

	# ----------------------------------------------------------------------
	def dnd_accept(self, source, event):
		#print("*-* dnd_accept")
		return self if isinstance(self, source.__class__) else None

	# ----------------------------------------------------------------------
	# Create drag object
	# ----------------------------------------------------------------------
	def dnd_enter(self, source, event):
		#print("*-* dnd_enter", id(self), "source=",id(source), "label=",id(source._dnd_label))
		if self._dnd_event:
			# On first motion send Cut or Copy message
			assert source is self
			self.event_generate(self._dnd_event)
			self.selectClear()
			self._dnd_event = None
		self.focus_set()
		if self._dnd_item is None:
			self._dnd_item = tk.Label(self, text=source._dnd_label)
		self.dnd_motion(source, event)

	# ----------------------------------------------------------------------
	def dnd_motion(self, source, event):
		#print("*-* dnd_motion",id(self), "source=",id(source))
		x = event.x_root - self.winfo_rootx()
		y = event.y_root - self.winfo_rooty()
		self._dnd_item.place(x=x, y=y)

	# ----------------------------------------------------------------------
	def dnd_leave(self, source, event):
		#print("*-* dnd_leave",id(self))
		self._dnd_item.place_forget()
		self._dnd_item = None

	# ----------------------------------------------------------------------
	def dnd_commit(self, source, event):
		#print("*-* dnd_commit",id(self))
		self._dnd_item.place_forget()
		self._dnd_item = None
		item = self.identify_row(event.y)
		if item:
			self.focus(item)
			bbox = self.bbox(item)
			ymid = bbox[1]+bbox[3]//2
			if event.y >= ymid:
				self.event_generate("<<Paste>>")
			else:
				self.event_generate("<<PasteSpecial>>")
		else:
			try:
				self.focus(self.get_children()[-1])
			except IndexError:
				pass
			self.event_generate("<<Paste>>")

	# ----------------------------------------------------------------------
	def dnd_end(self, target, event):
		#print("*-* dnd_end", id(self), "target=",id(target))
		self._dnd = None
		if self._dnd_event or target is None:
			# No DND happened
			self.selectClear()
			self.selection_set(self.focus())
			self._dnd_event = None
			self.event_generate("<<Undo>>")

	# ----------------------------------------------------------------------
	# Escape key pressed during dnd
	# ----------------------------------------------------------------------
	#def escape(self, event):
	#	if self._dnd:
	#		self.dnd_end(None,None)
	#		return "break"

#===============================================================================
# MultiListbox based on recipe from
#	http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52266
# Author: Brent Burley
# Date:   2001/03/14
#===============================================================================
class MultiListbox(tk.Frame):
	"""Multilistbox class"""

	# Add default options if not supplied
	defopt = (("borderwidth",	0),
		  ("selectmode",	tk.EXTENDED),
		  ("selectborderwidth",	0),
		  ("relief",		tk.FLAT),
		  ("exportselection",	tk.FALSE),
		  ("takefocus",		tk.FALSE))

	def __init__(self, master, lists, **options):
		super().__init__(master)
		self.paneframe = tk.PanedWindow(self, orient=tk.HORIZONTAL,
				showhandle=0, handlepad=0, handlesize=0,
				sashwidth=2, opaqueresize=1)
		self.paneframe.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		self.paneframe.bind("<Button-1>",	self._sashMark)
		self.paneframe.bind("<B1-Motion>",	self._sashDrag)
		self.paneframe.bind("<ButtonRelease-1>",self._sashRelease)

		self._lists  = []
		self._labels = []
		col = 0
		if "header" in options:
			header = options["header"]
			del options["header"]
		else:
			header = 1

		if "stretch" in options:
			stretch = options["stretch"]
			del options["stretch"]
		else:
			stretch = "always"

		for n,o in MultiListbox.defopt:
			if n not in options:
				options[n] = o

		for l, w, a in lists:
			#if header:
			frame = tk.Frame(self.paneframe, border=0)
			try: self.paneframe.add(frame, minsize=16, stretch=stretch)
			except: self.paneframe.add(frame, minsize=16)	# tk8.4
			if header:
				lbl = tk.Label(frame, text=l, borderwidth=1,
						relief=tk.RAISED)
				lbl.pack(fill=tk.X)
				lbl.bind('<Button-1>', lambda e, s=self, c=col:
						s.sort(c))
				self._labels.append(lbl)
			#else:
			#	frame = self

			lb = ExListbox(frame, width=w, **options)
			#if header:
			#	lb.pack(expand=tk.YES, fill=tk.BOTH)
			#else:
			lb.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
			self._lists.append(lb)

			lb.bind('<B2-Motion>', lambda e, s=self:
						s._b2motion(e.x, e.y))
			lb.bind('<Button-2>', lambda e, s=self:
						s._button2(e.x, e.y))
			lb.bind('<Button-4>', lambda e, s=self:
						s._scroll(tk.SCROLL, -1, tk.UNITS))
			lb.bind('<Button-5>', lambda e, s=self:
						s._scroll(tk.SCROLL, 1, tk.UNITS))
			lb.bind('<<ListboxSelect>>', lambda e, s=self, l=lb:
						s._updateSelect(l))
			col += 1

		self._lists[0]["takefocus"] = True

		if header:
			frame = tk.Frame(self)
			frame.pack(side=tk.RIGHT, fill=tk.Y)
			tk.Label(frame, borderwidth=1, relief=tk.RAISED).pack(fill=tk.X)

		sb = ttk.Scrollbar(frame, orient=tk.VERTICAL,
				takefocus=False,
				command=self._scroll)
		if header:
			sb.pack(fill=tk.Y, expand=tk.YES)
		else:
			sb.pack(side=tk.RIGHT, fill=tk.Y)

		self._lists[0]['yscrollcommand'] = sb.set
		self.activeList   = self._lists[0]
		self.sortAssist   = SortAssist
		self._sortOrder   = None	# Array containing the previous order of the list after sort
		self._sortColumn  = -1
		self._sashIndex   = -1
		self._sortReverse = False
		self._sashx       = None

	# ----------------------------------------------------------------------
	# Bind left/right arrow to focusing different list
	# ----------------------------------------------------------------------
	def bindLeftRight(self):
		""" Default bindings of left/right arrow to focusing different lists"""
		self.bindList('<Left>',  self.focusLeft)
		self.bindList('<Right>', self.focusRight)

	# ----------------------------------------------------------------------
	def bindList(self, event, func):
		self.bind(event, func)
		for l in self._lists:
			l.bind(event, func)

	# ----------------------------------------------------------------------
	def unbindList(self, event):
		self.unbind(event)
		for l in self._lists:
			l.bind(event)

	# ----------------------------------------------------------------------
	def _updateSelect(self, lst=None):
		if lst is None: lst = self.activeList
		ypos = lst.yview()[0]
		sel = lst.curselection()
		act = lst.index(tk.ACTIVE)
		for l in self._lists:
			if l is lst: continue
			l.selection_clear(0, tk.END)
			for s in sel:
				l.selection_set(s)
			l.activate(act)
			l.yview_moveto(ypos)
		self.event_generate('<<ListboxSelect>>')

	# ----------------------------------------------------------------------
	# Update header labels
	# ----------------------------------------------------------------------
	def labels(self, names):
		for i,n in enumerate(names):
			if i>len(self._labels): return
			self._labels[i].config(text=n)

	# ----------------------------------------------------------------------
	def _button2(self, x, y):
		for l in self._lists:
			l.scan_mark(x, y)
		return 'break'

	# ----------------------------------------------------------------------
	def _b2motion(self, x, y):
		for l in self._lists:
			l.scan_dragto(x, y)
		return 'break'

	# ----------------------------------------------------------------------
	def _sashMark(self, event):
		self._sashIndex = -1
		try:
			self._sashIndex,which = self.paneframe.identify(event.x, event.y)
			if which == "sash":
				self._sashx = [self.paneframe.sash_coord(i)[0]
						for i in range(len(self._lists)-1)]
				self._sashdx = self._sashx[self._sashIndex] - event.x
				self._sashDrag(event)
			else:
				self._sashIndex = -1
		except:
			return
		return 'break'

	# ----------------------------------------------------------------------
	def _sashDrag(self, event):
		if self._sashx and self._sashIndex >= 0:
			ddx = event.x - self._sashdx - self._sashx[self._sashIndex]
			self.paneframe.sash_place(self._sashIndex, event.x-self._sashdx, 1)
			for i in range(self._sashIndex+1, len(self._lists)-1):
				self.paneframe.sash_place(i, self._sashx[i]+ddx, 1)
		return 'break'

	# ----------------------------------------------------------------------
	def _sashRelease(self, event):
		if self._sashIndex >= 0:
			self._sashx = None
		return 'break'

	# ----------------------------------------------------------------------
	def _scroll(self, *args):
		for l in self._lists:
			l.yview(*args)
		return 'break'

	# ----------------------------------------------------------------------
	def curselection(self):
		return self._lists[0].curselection()

	# ----------------------------------------------------------------------
	def delete(self, first, last=None):
		for l in self._lists:
			l.delete(first, last)

	# ----------------------------------------------------------------------
	def get(self, first, last=None):
		result = []
		for l in self._lists:
			result.append(l.get(first, last))
		if last: return zip(*result)
		return result

	# ----------------------------------------------------------------------
	def listbox(self, i):
		return self._lists[i]

	# ----------------------------------------------------------------------
	def listboxes(self):
		return self._lists

	# ----------------------------------------------------------------------
	def listboxIndex(self, widget):
		return self._lists.index(widget)

	# ----------------------------------------------------------------------
	def index(self, item):
		return self._lists[0].index(item)

	# ----------------------------------------------------------------------
	def insert(self, index, *elements):
		for e in elements:
			for i,l in enumerate(self._lists):
				l.insert(index, e[i])
			if len(e) < len(self._lists):
				for l in self._lists[len(e) : len(self._lists)]:
					l.insert(index, "")

		if self._sortColumn>=0:
			txt = self._labels[self._sortColumn]["text"]
			self._labels[self._sortColumn].config(text=txt[:-1])
			self._sortColumn = -1

	# ----------------------------------------------------------------------
	# Change the value of a list item
	# and return the value of the old one
	# ----------------------------------------------------------------------
	def set(self, index, value):
		"""Set the value of a list item."""
		self.delete(index)
		self.insert(index, value)

	# ----------------------------------------------------------------------
	def size(self):
		return self._lists[0].size()

	# ----------------------------------------------------------------------
	def setPopupMenu(self, menu=None):
		"""Setup a popup menu list it should be in the form
		   [ (label, underline, command), ... ]"""
		for l in self._lists:
			l.setPopupMenu(menu)

	# ----------------------------------------------------------------------
	def nearest(self, y):
		return self._lists[0].nearest(y)

	# ----------------------------------------------------------------------
	def see(self, index):
		for l in self._lists:
			l.see(index)

	# ----------------------------------------------------------------------
	def configure(self, **kw):
		for l in self._lists:
			l.configure(**kw)
	config = configure

	# ----------------------------------------------------------------------
	def itemcget(self, index, option):
		"""Return the resource value for an ITEM and an OPTION."""
		return self._lists[0].itemcget(index, option)

	# ----------------------------------------------------------------------
	def itemconfigure(self, index, cnf=None, **kw):
		"""Configure resources of an ITEM.
		The values for resources are specified as keyword arguments.
		To get an overview about the allowed keyword arguments
		call the method without arguments.
		Valid resource names: background, bg, foreground, fg,
		selectbackground, selectforeground."""
		for l in self._lists:
			l.itemconfigure(index, cnf, **kw)
	itemconfig = itemconfigure

	# ----------------------------------------------------------------------
	# Override of the standard tkinter cget() routine
	# ----------------------------------------------------------------------
	def __getitem__(self, key):
		return self._lists[0].cget(key)

	# ----------------------------------------------------------------------
	# Override of the standard tkinter config() routine
	# ----------------------------------------------------------------------
	def __setitem__(self, key, value):
		for l in self._lists:
			l[key] = value

	# ----------------------------------------------------------------------
	# Selection
	# ----------------------------------------------------------------------
	def selection_anchor(self, index):
		for l in self._lists:
			l.selection_anchor(index)

	# ----------------------------------------------------------------------
	def selection_includes(self, index):
		return self._lists[0].selection_includes(index)

	# ----------------------------------------------------------------------
	def selection_clear(self, first, last=None):
		for l in self._lists:
			l.selection_clear(first, last)

	# ----------------------------------------------------------------------
	def selection_set(self, first, last=None):
		for l in self._lists:
			l.selection_set(first, last)

	# ----------------------------------------------------------------------
	def selectAll(self, event=None):
		"""Select all items"""
		self.selection_set(0, tk.END)
		self.event_generate("<<ListboxSelect>>")
		return "break"

	# ----------------------------------------------------------------------
	def selectClear(self, event=None):
		"""Unselect all items"""
		self.selection_clear(0, tk.END)
		self.event_generate("<<ListboxSelect>>")
		return "break"

	# ----------------------------------------------------------------------
	def selectInvert(self, event=None):
		"""Invert selection"""
		l = self._lists[0]
		for i in range(l.size()):
			if l.select_includes(i):
				self.selection_clear(i)
			else:
				self.selection_set(i)
		self.event_generate("<<ListboxSelect>>")
		return "break"

	# ----------------------------------------------------------------------
	def activate(self, index):
		for l in self._lists:
			l.activate(index)

	# ----------------------------------------------------------------------
	def focus_set(self):
		self._lists[0].focus_set()

	# ----------------------------------------------------------------------
	def focusLeft(self, event=None):
		listbox = self.focus_get()
		if listbox is None: return
		active = listbox.index(tk.ACTIVE)
		try:
			lid = self._lists.index(listbox) - 1
			if lid>=0:
				self._lists[lid].activate(active)
				self._lists[lid].focus_set()
		except:
			pass

	# ----------------------------------------------------------------------
	def focusRight(self, event=None):
		listbox = self.focus_get()
		if listbox is None: return
		active = listbox.index(tk.ACTIVE)
		try:
			lid = self._lists.index(listbox) + 1
			if lid < len(self._lists):
				self._lists[lid].activate(active)
				self._lists[lid].focus_set()
		except:
			pass

	# ----------------------------------------------------------------------
	def sort(self, column, reverse=None):
		""" Sort by a given column."""
		if self._lists[0].cget("state") == tk.DISABLED: return
		if self.sortAssist is None: return
		if column == self._sortColumn:
			txt = self._labels[self._sortColumn]["text"][:-1]
			if reverse is None:
				reverse = not self._sortReverse
		else:
			if self._sortColumn>=0:
				txt = self._labels[self._sortColumn]["text"][:-1]
				self._labels[self._sortColumn].config(text=txt)
				self._sortColumn = -1
			txt = self._labels[column]["text"]
			if reverse is None:
				reverse = False

		#elements = self.get(0, tk.END)
		elements = []
		lst = self._lists[0]
		for i in range(self.size()):
			item = []
			for l in self._lists:
				item.append(l.get(i))
			item.append(lst.selection_includes(i))	# Include selection
			item.append(i)				# Include position
			elements.append(item)

		try: active = int(self.index(tk.ACTIVE))
		except: active = -1

		self.delete(0, tk.END)

		elements.sort(key=self.sortAssist(column), reverse=reverse)

		# get selection status
		status = []
		self._sortOrder = []
		newactive = -1
		for i,item in enumerate(elements):
			idx = item.pop()
			if active == idx: newactive = i
			self._sortOrder.append(idx)
			status.append(item.pop())

		self.insert(tk.END, *elements)

		for i,s in enumerate(status):
			if s:
				self.selection_set(i)
				if newactive<0: newactive = i
		if newactive>=0:
			self.activate(newactive)

		self._sortColumn  = column
		self._sortReverse = reverse

		if reverse:
			self._labels[column].config(text=txt+"\N{BLACK DOWN-POINTING TRIANGLE}")
		else:
			self._labels[column].config(text=txt+"\N{BLACK UP-POINTING TRIANGLE}")
		self.event_generate("<<ListboxSort>>")

	# ----------------------------------------------------------------------
	def saveSort(self):
		return self._sortColumn, self._sortReverse

	# ----------------------------------------------------------------------
	def restoreSort(self, arg):
		if arg[0] >= 0:
			self.sort(*arg)

	# ----------------------------------------------------------------------
	def yview(self):
		return self._lists[0].yview()

	# ----------------------------------------------------------------------
	def yview_moveto(self, fraction):
		for l in self._lists:
			l.yview_moveto(fraction)

	# ----------------------------------------------------------------------
	def moveUp(self):
		for l in self._lists:
			l.moveUp()

	# ----------------------------------------------------------------------
	def moveDown(self):
		for l in self._lists:
			l.moveDown()

#===============================================================================
# A MultiListbox that remembers the color of items
#===============================================================================
class ColorMultiListbox(MultiListbox):
	# ----------------------------------------------------------------------
	def sort(self, column, direction=None):
		# remember colors
		colors = {}
		for i in range(self.size()):
			colors[self._lists[0].get(i)] = \
				self._lists[0].itemcget(i, "foreground")

		MultiListbox.sort(self, column, direction)

		# set colors
		for i in range(self.size()):
			self.setColor(i, colors[self._lists[0].get(i)])

		del colors

	# ----------------------------------------------------------------------
	def setColor(self, idx, color):
		for l in self._lists:
			l.itemconfigure(idx, foreground=color)

#===============================================================================
# Image list
#===============================================================================
class ImageListbox(tk.Text):
	"""ImageListbox widget which can display a list of strings and images"""
	def __init__(self, master, **options):
		super().__init__(master, **options)
		self.config(cursor="arrow",
			tabs="20p",
			#insertofftime=0,
			#insertontime=0,
			wrap=tk.NONE,
			insertwidth=0,
			takefocus=tk.TRUE,
			exportselection=0)
			#state=tk.DISABLED)
		self.bind("<Button-1>",		self._button1)
		self.bind("<Control-Button-1>",	self._controlButton1)
		self.bind("<Shift-Button-1>",	self._motion1)
		self.bind("<B1-Motion>",	self._motion1)
		self.bind("<Control-B1-Motion>",self._controlMotion1)
		self.bind("<Key>",		self._key)
		self.bind("<Delete>",		self._break)
		self.bind("<Return>",		self._break)
		self.bind("<KeyRelease>",	self._break)

		self.bind("<<Cut>>",		self.cut)
		self.bind("<<Copy>>",		self.copy)
		self.bind("<<Paste>>",		self.paste)
		self.bind("<<SelectAll>>",	self.selectAll)
		self.bind("<<SelectNone>>",	self.selectNone)

		self._selection = []
		self._anchor = 0
		self._active = 0

	# ----------------------------------------------------------------------
	def insert(self, index, icon, text):
		"""Insert ELEMENTS at INDEX."""
		#self.config(state=tk.NORMAL)
		if index != tk.END:
			index = int(index)
			sindex = f"{index+1}.0"
			super().insert(sindex, f"\t{text}\n")
			self.image_create(sindex, image=icon)
			self._selection.insert(index,False)
		else:
			self.image_create(tk.END, image=icon)
			super().insert(tk.END, f"\t{text}\n")
			self._selection.append(False)
		#self.config(state=tk.DISABLED)

	# ----------------------------------------------------------------------
	def delete(self, first, last=None):
		"""Delete items from FIRST to LAST (not included)."""
		if first == tk.END:
			super().delete("end.0", tk.END)
			self._selection.pop()
			return "break"
		if first == tk.ACTIVE:
			first = self.index(tk.ACTIVE)
		if last is None:
			i = int(first)
			if 0 <= i < len(self._selection):
				super().delete(f"{i+1}.0", f"{i+1}.0 + 1 lines")
				del self._selection[i]
			return "break"

		if last == tk.END:
			last = self.size()

		first = int(first)
		lines = int(last) - first
		super().delete(f"{first+1}.0", f"{first+1}.0 + {lines} lines")
		try:
			del self._selection[first:last]
		except IndexError:
			pass
		return "break"

	# ----------------------------------------------------------------------
	def size(self):
		"""Return the number of elements in the listbox."""
		return len(self._selection)

	# ----------------------------------------------------------------------
	def curselection(self):
		"""Return list of indices of currently selected item."""
		sel = []
		for i,x in enumerate(self._selection):
			if x: sel.append(i)
		return sel

	# ----------------------------------------------------------------------
	def nearest(self, y):
		"""Get index of item which is nearest to y coordinate Y."""
		index = super().index(f"@1,{y}")
		i = int(index.split(".", maxsplit=1)[0])-1
		if i>= self.size(): i -= 1
		return i

	# ----------------------------------------------------------------------
	def _button1(self, event):
		self.focus()
		self.selection_anchor(self.nearest(event.y))
		self._active = self._anchor
		self._select()
		return "break"

	# ----------------------------------------------------------------------
	def _motion1(self, event):
		y = self.nearest(event.y)
		if self._active != y:
			self._active = y
			self._selectRange()
		return "break"

	# ----------------------------------------------------------------------
	def _controlButton1(self, event):
		self.selection_anchor(self.nearest(event.y))
		self._active = self._anchor
		self._selection[self._anchor] = not self._selection[self._anchor]
		self._tagSelection()
		self.event_generate("<<ListboxSelect>>")
		return "break"

	# ----------------------------------------------------------------------
	def _controlMotion1(self, event):
		self._active = self.nearest(event.y)
		last = self._selection[self._anchor]
		if self._active < self._anchor:
			for i in range(self._active, self._anchor):
				self._selection[i] = last
		elif self._active > self._anchor:
			for i in range(self._anchor, self._active+1):
				self._selection[i] = last
		self._tagSelection()
		return "break"

	# ----------------------------------------------------------------------
	def _key(self, event):
		if event.keysym == "Up":
			if self._active == 0: return "break"
			self._active -= 1
			if event.state & STATE.SHIFT:
				self._selectRange()
			else:
				self._anchor = self._active
				self._select()

		elif event.keysym == "Down":
			self._active += 1
			if self._active >= self.size():
				self._active = self.size()-1
				return "break"
			if event.state & STATE.SHIFT:
				self._selectRange()
			else:
				self._anchor = self._active
				self._select()

		elif event.keysym in ("Prior", "Next", "Delete"):
			return

		if event.state & STATE.CONTROL == 0:
			# Ignore all normal keys
			return "break"
		# else let system handle all Control keys

	# ----------------------------------------------------------------------
	def _break(self, event):
		return "break"

	# ----------------------------------------------------------------------
	def _select(self):
		self._selection = [False] * len(self._selection)
		self.selection_set(self._active)
		idx = f"{self._active+1}.0"
		super().see(idx)
		super().index(idx)
		self.event_generate("<<ListboxSelect>>")

	# ----------------------------------------------------------------------
	def _selectRange(self):
		self._selection = [False] * len(self._selection)
		if self._active < self._anchor:
			for i in range(self._active, self._anchor):
				try:
					self._selection[i] = True
				except IndexError:
					pass
		elif self._active > self._anchor:
			for i in range(self._anchor, self._active+1):
				try:
					self._selection[i] = True
				except IndexError:
					pass
		try:
			self._selection[self._anchor] = True
		except IndexError:
			pass
		self._tagSelection()
		self.event_generate("<<ListboxSelect>>")
		return "break"

	# ----------------------------------------------------------------------
	def selectAll(self, event=None):
		self.selection_set(0,tk.END)
		return "break"

	# ----------------------------------------------------------------------
	def selectNone(self, event=None):
		self.selection_clear(0,tk.END)
		return "break"

	# ----------------------------------------------------------------------
	def selection_anchor(self, index):
		"""Set the fixed end oft the selection to INDEX."""
		self._anchor = index

	select_anchor = selection_anchor

	# ----------------------------------------------------------------------
	def selection_includes(self, index):
		"""Return 1 if INDEX is part of the selection."""
		return self._selection[index]

	select_includes = selection_includes

	# ----------------------------------------------------------------------
	def selection_clear(self, first=None, last=None):
		"""Clear the selection from FIRST to LAST (not included)."""
		if first is None:
			self._selection = [False] * len(self._selection)
		elif first == tk.END:
			self._selection[-1] = False
		elif last is None:
			if 0 <= int(first) < len(self._selection):
				self._selection[int(first)] = False
		else:
			if last == tk.END: last = self.size()
			for i in range(int(first), last):
				self._selection[i] = False
		self._tagSelection()

	select_clear = selection_clear

	# ----------------------------------------------------------------------
	def selection_set(self, first, last=None):
		"""Set the selection from FIRST to LAST (not included) without
		changing the currently selected elements."""
		if first == tk.END:
			self._selection[-1] = True
		elif last is None:
			if 0 <= int(first) < len(self._selection):
				self._selection[int(first)] = True
		else:
			if last == tk.END:
				last = self.size()
			for i in range(int(first), last):
				self._selection[i] = True
		self._tagSelection()

	select_set = selection_set

	# ----------------------------------------------------------------------
	def see(self, index):
		"""Scroll such that INDEX is visible."""
		if index == tk.END:
			super().see(index)
		else:
			super().see(f"{int(index)+1}.0")

	# ----------------------------------------------------------------------
	def _tagSelection(self):
		self.tag_delete("lola")
		for i,x in enumerate(self._selection):
			if x:
				self.tag_add("lola", f"{i+1}.0", f"{i+1}.0 +1 lines")
		self.tag_configure("lola", foreground="White", background="SteelBlue2")
		super().selection_clear()

	# ----------------------------------------------------------------------
	def bbox(self, index):
		"""Return a tuple of X1,Y1,X2,Y2 coordinates for a rectangle
		which encloses the item identified by index in ARGS."""
		if index == tk.END:
			return super().dlineinfo(index)[:4]
		if index == tk.ACTIVE:
			index = self.index(index)
		return super().bbox(f"{int(index)+1}.2")

	# ----------------------------------------------------------------------
	def dlineinfo(self,index):
		if index == tk.END:
			return super().dlineinfo(index)
		if index == tk.ACTIVE:
			index = self.index(index)
		return super().dlineinfo(f"{int(index)+1}.0")

	# ----------------------------------------------------------------------
	def activate(self, index):
		"""Activate item identified by INDEX."""
		if index == tk.END:
			self._active = self.size()-1
		else:
			self._active = int(index)

	# ----------------------------------------------------------------------
	def get(self, first, last=None):
		"""Get list of items from FIRST to LAST (not included)."""
		if first == tk.END:
			first = self.size()-1
		elif first == tk.ACTIVE:
			first = self._active
		else:
			first = int(first)

		if last is None:
			if 0 <= first < len(self._selection):
				first += 1
				#img = super().image_cget(f"{first}.0", "image")
				return super().get(f"{first}.2", f"{first}.end")
			return None

		if last == tk.END:
			last = self.size()
		elif last == tk.ACTIVE:
			last = self._active
		else:
			last = int(last)

		# FIXME....
		# return list of text items

	# ----------------------------------------------------------------------
	# get both image and text
	# ----------------------------------------------------------------------
	def elicit(self, first, last=None):
		"""Get list of items from FIRST to LAST (not included)."""
		if first == tk.END:
			first = self.size()-1
		elif first == tk.ACTIVE:
			first = self._active
		else:
			first = int(first)

		if last is None:
			if 0 <= first < len(self._selection):
				first += 1
				img = super().image_cget(f"{first}.0", "image")
				txt = super().get(f"{first}.2", f"{first}.end")
				return img,txt
			return None,None

		if last == tk.END:
			last = self.size()
		elif last == tk.ACTIVE:
			last = self._active
		else:
			last = int(last)

		# FIXME....
		# return list of (image,text) items

	# ----------------------------------------------------------------------
	def index(self, index):
		"""Return index of item identified with INDEX."""
		if index == tk.ACTIVE:
			return self._active
		else:
			return super().index(index)

	# ----------------------------------------------------------------------
	def scan_mark(self, x, y):
		"""Remember the current X, Y coordinates."""
		pass

	# ----------------------------------------------------------------------
	def scan_dragto(self, x, y):
		"""Adjust the view of the listbox to 10 times the
		difference between X and Y and the coordinates given in
		scan_mark."""
		pass

	# ----------------------------------------------------------------------
	def itemcget(self, index, option):
		"""Return the resource value for an ITEM and an OPTION."""
		pass

	# ----------------------------------------------------------------------
	def itemconfigure(self, index, cnf=None, **kw):
		"""Configure resources of an ITEM.

		The values for resources are specified as keyword arguments.
		To get an overview about the allowed keyword arguments
		call the method without arguments.
		Valid resource names: background, bg, foreground, fg,
		selectbackground, selectforeground."""
		pass

	itemconfig = itemconfigure

	# ----------------------------------------------------------------------
	# Override cut,copy,paste to do nothing
	# ----------------------------------------------------------------------
	def cut(self, event=None):
		return "break"
	copy  = cut
	paste = cut

#===============================================================================
# Class to edit in place the contents of a listbox
#===============================================================================
class InPlaceEdit:
	def __init__(self, listbox,
				item=tk.ACTIVE,
				value=None,
				x=None,
				select=True,
				column=None,
				**kw):
		# Return value
		self.value   = None	# Result
		self.frame   = None
		self.lastkey = None	# Last key that exited the editbox
		self.column  = column
		self.kw      = kw
		self._x      = x
		self._select = select

		# Find active
		try: self.active = listbox.index(item)
		except: return

		self.item    = item
		self.listbox = listbox

		# Create and set value
		self.frame = tk.Frame(listbox, relief=None)
		self.createWidget()
		self.old = self.set(value)
		self.defaultBinds()

		# Bindings
		self.frame.bind("<FocusOut>",        self.focusOut)
		# Unmap creates core dump when Fn key is pressed
		self.frame.bind("<ButtonRelease-1>", self.clickOk)
		self.frame.bind("<ButtonRelease-3>", self.clickCancel)
		self.listbox.bind("<Configure>",     self.resize)
		#self.frame.bind("<Unmap>",          self._destroy)

		try:
			self._grab_window = self.frame.grab_current()
		except tk.TclError:
			self._grab_window = None
		self.resize()
		self.show()

	# ----------------------------------------------------------------------
	def show(self):
		# Show and wait to be destroyed
		try:
			if not self.frame.winfo_ismapped():
				self.frame.wait_visibility()
			self.frame.grab_set()
			self.icursor()
			self.frame.wait_window()
			self.frame.place_forget()
		except tk.TclError:
			pass
		#self.listbox.focus_set()

	# ----------------------------------------------------------------------
	# Override method if another widget is requested
	# ----------------------------------------------------------------------
	def createWidget(self):
		self.edit = tk.Entry(self.frame, **self.kw)
		self.edit.pack(expand=tk.YES, fill=tk.BOTH)
		self.edit.focus_set()

	# ----------------------------------------------------------------------
	# set insert cursor at location
	# ----------------------------------------------------------------------
	def icursor(self):
		if self._x is not None:
			self.edit.icursor(f"@{self._x}")

	# ----------------------------------------------------------------------
	# Set default bindings
	# ----------------------------------------------------------------------
	def defaultBinds(self):
		try:
			self.edit.bind("<Return>",   self.ok)
			self.edit.bind("<KP_Enter>", self.ok)
			self.edit.bind("<Up>",       self.ok)
			self.edit.bind("<Down>",     self.ok)
			self.edit.bind("<Escape>",   self.cancel)
		except AttributeError:
			pass

	# ----------------------------------------------------------------------
	def resize(self, event=None):
		if self.frame is None: return
		if isinstance(self.listbox, tk.Listbox):
			bbox = self.listbox.bbox(self.item)
		elif isinstance(self.listbox, ttk.Treeview):
			try:
				if self.column is None:
					bbox = self.listbox.bbox(self.item)
				else:
					bbox = self.listbox.bbox(self.item, self.column)
			except tk.TclError:
				return
		try:
			x, y, w, h = bbox
		except (TypeError,ValueError):
			return
		w = self.listbox.winfo_width() - x
		h += 3
		try:
			self.frame.place(in_=self.listbox,
					x=x-1, y=y-1,
					width=w, height=h,
					bordermode=tk.OUTSIDE)
			self.frame.update_idletasks()
		except tk.TclError:
			pass

	# ----------------------------------------------------------------------
	# Override method to set the value
	# ----------------------------------------------------------------------
	def set(self, value):
		if self.frame is None: return
		if value is None:
			if isinstance(self.listbox, tk.Listbox):
				value = self.listbox.get(self.item)
			elif isinstance(self.listbox, ttk.Treeview):
				if self.column is None:
					value = self.listbox.item(self.item, "text")
				else:
					try:
						value = self.listbox.item(self.item, "values")[self.column]
					except IndexError:
						value = ""
		self.edit.delete(0, tk.END)
		self.edit.insert(0, value)
		if self._select:
			self.edit.selection_range(0, tk.END)
		return value

	# ----------------------------------------------------------------------
	# Override method to get the value
	# ----------------------------------------------------------------------
	def get(self):
		if self.frame is None: return None
		return self.edit.get()

	# ----------------------------------------------------------------------
	def reset_grab(self):
		if self.frame is None: return
		self.frame.grab_release()
		if self._grab_window is not None:
			try:
				self._grab_window.grab_set()
			except tk.TclError:
				pass

	# ----------------------------------------------------------------------
	def clickOk(self, event):
		# If clicked outside return ok
		if event.x < 0 or \
		   event.y < 0 or \
		   event.x > self.frame.winfo_width() or \
		   event.y > self.frame.winfo_height():
			self.ok(event)

	# ----------------------------------------------------------------------
	def clickCancel(self, event):
		# If clicked outside return cancel
		if event.x < 0 or \
		   event.y < 0 or \
		   event.x > self.frame.winfo_width() or \
		   event.y > self.frame.winfo_height():
			self.cancel(event)

	# ----------------------------------------------------------------------
	def focusOut(self, event=None):
		self.ok()

	# ----------------------------------------------------------------------
	def updateValue(self):
		if isinstance(self.listbox, tk.Listbox):
			self.listbox.delete(self.active)
			self.listbox.insert(self.active, self.value)
		elif isinstance(self.listbox, ttk.Treeview):
			if self.column is None:
				self.listbox.item(self.item, text=self.value)
			else:
				values = list(self.listbox.item(self.item, "values"))
				try:
					values[self.column] = self.value
				except IndexError:
					while len(values) < self.column:
						values.append("")
					values.append(self.value)
				self.listbox.item(self.item, values=values)

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		if event: self.lastkey = event.keysym
		self.value = self.get()
		self.frame.unbind('<FocusOut>')

		if isinstance(self.listbox, ttk.Treeview):
			act = self.listbox.focus()
			sel = False
		else:
			act = self.listbox.index(tk.ACTIVE)
			sel = self.listbox.selection_includes(self.active)
		self.updateValue()
		self.listbox.see(self.item)

		if sel:
			self.listbox.selection_set(self.active)
		if isinstance(self.listbox, ttk.Treeview):
			self.listbox.focus(act)
		else:
			self.listbox.activate(act)
		if self.value == self.old: self.value = None

		self.reset_grab()
		self.listbox.focus_set()
		self.frame.place_forget()
		self.frame.destroy()
		return "break"

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		self.reset_grab()
		self.listbox.focus_set()
		self.frame.place_forget()
		self.frame.destroy()
		return "break"

	# ----------------------------------------------------------------------
	def ignore(self, event=None):
		pass

#===============================================================================
class InPlaceSpinbox(InPlaceEdit):
	# ----------------------------------------------------------------------
	def createWidget(self):
		self.edit = tk.Spinbox(self.frame, **self.kw)
		self.edit.pack(expand=tk.YES, fill=tk.BOTH)
		self.edit.focus_set()

	# ----------------------------------------------------------------------
	def set(self, value):
		if self.frame is None: return
		if value is None:
			value = self.listbox.get(self.item)
		self.edit.delete(0, tk.END)
		self.edit.insert(0, value)
		return value

#===============================================================================
class InPlaceInteger(InPlaceEdit):
	# ----------------------------------------------------------------------
	def createWidget(self):
		self.edit = IntegerEntry(self.frame, **self.kw)
		self.edit.pack(expand=tk.YES, fill=tk.BOTH)
		self.edit.focus_set()

#===============================================================================
class InPlaceFloat(InPlaceEdit):
	# ----------------------------------------------------------------------
	def createWidget(self):
		self.edit = FloatEntry(self.frame, **self.kw)
		self.edit.pack(expand=tk.YES, fill=tk.BOTH)
		self.edit.focus_set()

#===============================================================================
class InPlaceList(InPlaceEdit):
	def __init__(self, listbox, item=tk.ACTIVE, value=None, height=None, values=None, **kw):
		self.values = values
		self.height = height
		if values is None: values = []
		super().__init__(listbox, item, value, **kw)

	# ----------------------------------------------------------------------
	def createWidget(self):
		self.frame.config(relief=tk.RAISED)
		sb = ttk.Scrollbar(self.frame)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		if self.height is None:
			if len(self.values)<10:
				self.height = max(len(self.values)+1,3)
			else:
				self.height = 10
		self.edit = ExListbox(self.frame,
				selectmode=tk.BROWSE,
				height=self.height,
				#background="White",
				yscrollcommand=sb.set)
		sb.config(command=self.edit.yview)
		self.edit.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
		self.edit.bind('<ButtonRelease-1>', self.ok)
		self.edit.focus_set()

	# ----------------------------------------------------------------------
	def set(self, value):
		if value is None:
			value = self.listbox.get(self.item)
		# Fill&select listbox
		for item in self.values:
			self.edit.insert(tk.END, item)
			if item == value:
				self.edit.activate(tk.END)
				self.edit.selection_set(tk.END)
		if not self.edit.curselection():
			self.edit.activate(0)
		self.edit.see(tk.ACTIVE)
		return value

	# ----------------------------------------------------------------------
	def get(self):
		cur = self.edit.curselection()
		if cur:
			return self.edit.get(cur[0])
		else:
			return ""

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		cur = self.edit.curselection()
		if cur:
			InPlaceEdit.ok(self, event)
		else:
			InPlaceEdit.cancel(self)

	# ----------------------------------------------------------------------
	def defaultBinds(self):
		InPlaceEdit.defaultBinds(self)
		try:
			self.edit.unbind("<Up>")
			self.edit.unbind("<Down>")
		except AttributeError:
			pass

	# ----------------------------------------------------------------------
	def resize(self, event=None):
		if self.frame is None: return
		bbox = self.listbox.bbox(self.item)
		if bbox is None: return
		x, y, item_width, item_height = bbox
		list_width  = self.listbox.winfo_width()
		list_height = self.listbox.winfo_height()
		h = item_height*self.height + 2

		if y+h > list_height:
			y = list_height - h
			if y <= 0:
				y = 0
				h = list_height

		try:
			self.frame.place(in_=self.listbox,
					x=x-1, y=y,
					width=list_width, height=h,
					bordermode=tk.OUTSIDE)
			self.frame.update_idletasks()
		except tk.TclError:
			pass

#===============================================================================
class InPlaceSearchList(InPlaceList):
	# ----------------------------------------------------------------------
	def createWidget(self):
		self.frame.config(relief=tk.RAISED)
		sb = ttk.Scrollbar(self.frame)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		if self.height is None:
			if len(self.values)<10:
				self.height = max(len(self.values)+1,3)
			else:
				self.height = 10
		self.edit = SearchListbox(self.frame,
				selectmode=tk.BROWSE,
				height=self.height,
				#background="White",
				yscrollcommand=sb.set)
		sb.config(command=self.edit.yview)
		self.edit.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
		self.edit.bind('<ButtonRelease-1>', self.ok)
		self.edit.focus_set()

#===============================================================================
class InPlaceColor(InPlaceEdit):
	# ----------------------------------------------------------------------
	def createWidget(self):
		b = tk.Button(self.frame, text="x",
			padx=0, pady=0, command=self.clearColor)
		b.pack(side=tk.LEFT)
		self.edit = tk.Button(self.frame, command=self.selectColor)
		self.edit.pack(side=tk.RIGHT, expand=tk.YES, fill=tk.BOTH)
		self.edit.focus_set()

	# ----------------------------------------------------------------------
	def set(self, value):
		if value is None:
			value = self.listbox.get(self.item)
		try:
			self.value = f"#{int(float(value)):06X}"
		except ValueError:
			self.value = value
		if self.value is None or self.value=="": self.value = "White"
		if value != "":
			self.edit.config(text=self.value,
					background=self.value,
					activebackground=self.value)
		return self.value

	# ----------------------------------------------------------------------
	def get(self):
		return self.edit["text"]

	# ----------------------------------------------------------------------
	def selectColor(self):
		self.frame.unbind("<FocusOut>")
		try:
			rgb, colorStr = askcolor(
					title="Color",
					initialcolor=self.value,
					parent=self.listbox.master)
		except tk.TclError:
			colorStr = None
		if colorStr is not None:
			colorStr = str(colorStr)
			self.value = colorStr
			self.edit.config(text=colorStr,
					background=colorStr,
					activebackground=colorStr)
		self.frame.bind("<FocusOut>", self.cancel)
		self.edit.focus_set()
		self.ok()

	# ----------------------------------------------------------------------
	def clearColor(self):
		self.frame.unbind("<FocusOut>")
		self.value = None
		self.edit.config(text="",
				background="White",
				activebackground="White")
		self.frame.bind("<FocusOut>", self.cancel)
		self.edit.focus_set()

#===============================================================================
class InPlaceMaxLength(InPlaceEdit):
	def __init__(self, listbox, item=tk.ACTIVE, value=None, maxlength=None, **kw):
		self.maxlength = maxlength
		super().__init__(listbox, item, value, **kw)

	# ----------------------------------------------------------------------
	# Override method if another widget is requested
	# ----------------------------------------------------------------------
	def createWidget(self):
		self.edit = MaxLengthEntry(self.frame,
					maxlength=self.maxlength,
					**self.kw)
		self.edit.pack(expand=tk.YES, fill=tk.BOTH)
		self.edit.focus_set()

#===============================================================================
class InPlaceText(InPlaceEdit):
	# ----------------------------------------------------------------------
	def show(self):
		self.toplevel.bind("<FocusOut>", self.focusOut)
		try:
			if not self.toplevel.winfo_ismapped():
				self.toplevel.wait_visibility()
			self.toplevel.grab_set()
			self.toplevel.wait_window()
		except tk.TclError:
			pass

	# ----------------------------------------------------------------------
	def defaultBinds(self):
		InPlaceEdit.defaultBinds(self)
		self.toplevel.bind("<ButtonRelease-1>", self.clickOk)
		self.toplevel.bind("<ButtonRelease-3>", self.clickCancel)
		#self.edit.bind("<ButtonRelease-1>",    self.clickOk)
		#self.edit.bind("<ButtonRelease-3>",    self.clickCancel)
		self.edit.bind("<Shift-Return>",        self.shiftReturn)
		self.edit.bind("<Escape>",              self.cancel)
		# bind to an empty function, so it will be consumed
		# by the Text widget and not send to parent
		self.edit.bind("<Up>",                  self.ignore)
		self.edit.bind("<Down>",                self.ignore)
		self.edit.bind("<Left>",                self.ignore)
		self.edit.bind("<Right>",               self.ignore)

	# ----------------------------------------------------------------------
	def createWidget(self):
		self.toplevel = tk.Toplevel(self.listbox)
		self.toplevel.transient(self.listbox)
		if sys.platform in ("win32", "win64"):
			self.toplevel.update_idletasks()
		self.toplevel.overrideredirect(1)
		self.edit = tk.Text(self.toplevel, width=80, height=10,
					background="White", undo=True)
		self.edit.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		self.edit.focus_set()

	# ----------------------------------------------------------------------
	def resize(self, event=None):
		if self.frame is None: return
		bbox = self.listbox.bbox(self.item)
		if bbox is None: return
		x, y, w, h = bbox
		x += self.listbox.winfo_rootx()
		y += self.listbox.winfo_rooty()
		#w  = self.listbox.winfo_width()
		try:
			self.toplevel.wm_geometry(f"+{x}+{y}")
		except tk.TclError:
			pass

	# ----------------------------------------------------------------------
	def set(self, value):
		if self.frame is None: return
		if value is None:
			value = self.listbox.get(self.item)
		self.edit.delete("0.0", tk.END)
		self.edit.insert("0.0", value)
		self.edit.tag_add(tk.SEL, "0.0", tk.END)
		return value

	# ----------------------------------------------------------------------
	def get(self):
		if self.frame is None: return None
		return self.edit.get("0.0", tk.END).strip()

	# ----------------------------------------------------------------------
	def shiftReturn(self, event):
		# Empty binding to avoid the Shift-Return to trigger the "ok"
		pass

	# ----------------------------------------------------------------------
	def clickOk(self, event):
		# If clicked outside return ok
		if event.x < 0 or \
		   event.y < 0 or \
		   event.x > self.toplevel.winfo_width() or \
		   event.y > self.toplevel.winfo_height():
			self.ok(event)

	# ----------------------------------------------------------------------
	def clickCancel(self, event):
		# If clicked outside return cancel
		if event.x < 0 or \
		   event.y < 0 or \
		   event.x > self.toplevel.winfo_width() or \
		   event.y > self.toplevel.winfo_height():
			self.cancel(event)

	# ----------------------------------------------------------------------
	def ok(self, event=None):
		InPlaceEdit.ok(self, event)
		self.toplevel.destroy()
		return "break"

	# ----------------------------------------------------------------------
	def cancel(self, event=None):
		InPlaceEdit.cancel(self, event)
		self.toplevel.destroy()
		return "break"

#===============================================================================
class InPlaceFile(InPlaceEdit):
	# ----------------------------------------------------------------------
	# @param title		dialog title
	# @param filetypes	filetypes to be passed to Dialog
	# @param save		save or load dialog
	# @param showext	show extension in the text input
	# ----------------------------------------------------------------------
	def __init__(self, listbox, item=tk.ACTIVE,
				value=None,
				title=None,
				filetypes=None,
				save=True,
				showext=True,
				**kw):
		self.title     = title
		self._filetypes= filetypes
		self._save     = save
		self._icon     = icons.get("load")
		if not showext:
			value, self._ext = os.path.splitext(value)
		else:
			self._ext = None
		super().__init__(listbox, item, value, **kw)

	# ----------------------------------------------------------------------
	def createWidget(self):
		self.edit = tk.Entry(self.frame, width=5, **self.kw)
		self.edit.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		b = tk.Button(self.frame, image=self._icon,
			padx=0, pady=0, command=self.fileDialog)
		b.pack(side=tk.RIGHT)
		self.edit.focus_set()

	# ----------------------------------------------------------------------
	def fileDialog(self):
		self.frame.unbind("<FocusOut>")
		self.frame.grab_release()
		if self.value is None:
			if isinstance(self.listbox, ttk.Treeview):
				filename = self.listbox.item(self.item,"text")
			else:
				filename = self.listbox.get(self.item)
		else:
			filename = self.value
		if self._ext is not None:
			filename += self._ext
		if self._save:
			fn = bFileDialog.asksaveasfilename(master=self.listbox,
				title       = self.title,
				initialfile = filename,
				filetypes   = self._filetypes)
		else:
			fn = bFileDialog.askopenfilename(master=self.listbox,
				title       = self.title,
				initialfile = filename,
				filetypes   = self._filetypes)
		self.frame.grab_set()
		#self.frame.bind("<FocusOut>", self.cancel)
		self._icon = None
		if fn:
			if self._ext is not None:
				self._ext = os.path.splitext(fn)[1]
			self.edit.delete(0, tk.END)
			self.edit.insert(0, fn)
			self.ok()
		else:
			self.cancel()

#=============================================================================
# PopupList
# Show a popup list on a top level and return selected item
#=============================================================================
class PopupList(tk.Toplevel):
	def __init__(self, master, items=None, selected=None, **kw):
		super().__init__(master, **kw)
		self.selected = selected
		self.overrideredirect(1)
		self.transient(master)

		# Create the listbox inside the dropdown window
		sb = ttk.Scrollbar(self)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		self._listbox = SearchListbox(self,
					selectmode=tk.BROWSE,
					yscrollcommand=sb.set)
		self._listbox.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		sb.config(command=self._listbox.yview)

		if items:
			for item in items:
				self._listbox.insert(tk.END, item)
				if selected == item:
					self._listbox.selection_set(tk.END)
					self._listbox.activate(tk.END)
					self.see(tk.ACTIVE)

		self._listbox.bind('<Escape>',		self.close)
		self._listbox.bind('<Return>',		self._select)
		self._listbox.bind('<KP_Enter>',	self._select)
		self._listbox.bind('<Tab>',		self._select)
		self._listbox.bind('<Shift-Tab>',	self._select)
		self._listbox.bind('<ButtonRelease-1>',	self._release)
		self.bind('<FocusOut>',			self.close)

	# ----------------------------------------------------------------------
	def show(self, x, y):
		self.deiconify()
		if x is not None and y is not None:
			self.geometry(f"+{x}+{y}")
		self._listbox.focus_set()
		#if not self.winfo_ismapped():
		#	self.wait_visibility()
		#self.grab_set()
		self.wait_window()
		return self.selected

	# ----------------------------------------------------------------------
	def close(self, event=None):
		self.grab_release()
		self.destroy()

	# ----------------------------------------------------------------------
	def _select(self, event=None):
		self.selected = self._listbox.get(tk.ACTIVE)
		self.close()

	# ----------------------------------------------------------------------
	def _release(self, event):
		act = self._listbox.nearest(event.y)
		self._listbox.activate(act)
		self._select()

#=============================================================================
# Combobox
#=============================================================================
class Combobox(tk.Frame):
	def __init__(self, master, label=True, *args, **kwargs):
		super().__init__(master, class_="Combobox")
		tk.Frame.config(self, padx=0, pady=0)

		if "command" in kwargs:
			self.command = kwargs.get("command")
			del kwargs["command"]
		else:
			self.command = None

		# Create entry and button
		if label:
			self._text = tk.Label(self, relief=tk.GROOVE, anchor=tk.W, *args, **kwargs)
		else:
			self._text = tk.Entry(self, *args, **kwargs)
		self._text.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

		# Arrow button
		self._post = tk.IntVar()
		self._post.trace("w", self._showList)
		self._arrowBtn = tk.Checkbutton(self,
			text="\N{BLACK DOWN-POINTING TRIANGLE}",
			variable=self._post,
			indicatoron=False,
			padx=2, pady=0)
		self._arrowBtn.pack(side=tk.RIGHT, fill=tk.Y)

		# Bindings
		self._text.bind('<Up>',       self.postList)
		self._text.bind('<Down>',     self.postList)
		self._text.bind('<Return>',   self.postList)
		self._text.bind('<KP_Enter>', self.postList)
		self.bind('<Up>',             self.postList)
		self.bind('<Down>',           self.postList)
		self.bind('<Return>',         self.postList)
		self.bind('<KP_Enter>',       self.postList)
		if label:
			self._text.bind('<Key-space>',       self.postList)
			self._text.bind('<Button-1>',        self._togglePost)
		else:
			self._text.bind('<Button-1>',        self.button1)
			self._text.bind('<ButtonRelease-1>', self.release1)
		self._arrowBtn.bind("<Key>", self._key)

		# Need to unpost the popup if the entry field is unmapped (eg:
		# its toplevel window is withdrawn) while the popup list is
		# displayed.
		self._text.bind('<Unmap>', self.unpostList)

		# Create a static popup window with dropdown list
		self._popup = tk.Toplevel(master)
		self._popup.overrideredirect(1)
		self._popup.transient(master)
		self._popup.withdraw()

		# Create the listbox inside the dropdown window
		sb = ttk.Scrollbar(self._popup)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		for k in ("anchor","justify"):
			try: del kwargs[k]
			except KeyError: pass
		self._listbox = SearchListbox(self._popup,
					selectmode=tk.BROWSE,
					yscrollcommand=sb.set,
					*args,
					**kwargs)
		self._listbox.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
		sb.config(command=self._listbox.yview)

		# Bind events to the dropdown window.
		self._listbox.bind('<Escape>',          self.unpostList)
		self._listbox.bind('<Return>',          self._selectUnpost)
		self._listbox.bind('<KP_Enter>',        self._selectUnpost)
		self._listbox.bind('<Tab>',             self._selectUnpost)
		self._listbox.bind('<Shift-Tab>',       self._selectUnpost)
		self._listbox.bind('<ButtonRelease-1>', self._release)
		self._popup.bind('<FocusOut>',          self._focusOut)
		self._popup.bind('<Button-1>',          self.popupClick)
		self._popup.bind('<Button-3>',          self.popupClick)
		#self._popup.bind('<Shift-Tab>',        self._selectUnpostPrev)
		#self._popup.bind('<Tab>',              self._selectUnpostNext)
		#if sys.platform in ("linux","linux2"):
		#	self._popup.bind('<ISO_Left_Tab>', self._selectUnpostPrev)

		# grab_set redirects all mouse events to the list even
		# when the list is posted with a mouse click
		#self._hide_on_next_release = False
		self._grab_window = None

	# ----------------------------------------------------------------------
	# Unhide and post the list
	# ----------------------------------------------------------------------
	def _showList(self, a=False, b=False, c=False):
		if self._post.get():
			Balloon.hide()
			self._grab_window = None
			try:
				self._grab_window = self.grab_current()
				if self._grab_window is not None:
					self._grab_window.grab_release()
			except KeyError:
				pass
			if self._text.cget("state") == tk.DISABLED:
				self._post.set(False)
				return

			SearchBox.reset()
			self.beforeShow()

			h = self._popup.winfo_height()
			if h == 1:
				self._popup.deiconify()
				self._popup.lift()
				self._popup.update_idletasks()
				h = self._popup.winfo_height()
			w = self._text.winfo_width() + self._arrowBtn.winfo_width()

			x = self._text.winfo_rootx()
			y = self._text.winfo_rooty() + self._text.winfo_height()
			sh = self.winfo_screenheight()
			if y + h > sh and y > sh / 2:
				y = self._text.winfo_rooty() - h

			self._popup.deiconify()
			self._popup.geometry(f"{w}x{h}+{x}+{y}")
			self._popup.lift()
			try:
				# Grab the popup, so that all events are delivered to it, and
				# set focus to the listbox, to make keyboard navigation
				# easier.
				self._popup.grab_set()
			except tk.TclError:
				pass
			self._popup.update_idletasks()
			self._listbox.focus_set()
			self._showSelection()

		elif self._popup.winfo_ismapped():
			self._popup.grab_release()
			if self._grab_window:
				self._grab_window.grab_set()
				self._grab_window = None
			self._popup.withdraw()
			self._arrowBtn.focus_set()
			self.afterHide()

	# ----------------------------------------------------------------------
	def _showSelection(self):
		lb = self._listbox
		lb.selection_clear(0,tk.END)
		item = self.get()
		# Test active
		if lb.get(tk.ACTIVE) != item:
			# Scan list
			for i in range(lb.size()):
				if item == lb.get(i):
					lb.activate(i)
		lb.selection_set(tk.ACTIVE)
		lb.see(tk.ACTIVE)

	# ----------------------------------------------------------------------
	# Post list on click
	# ----------------------------------------------------------------------
	def button1(self, event):
		if self.focus_get() is not self._text:
			self.postList()
			return "break"

	# ----------------------------------------------------------------------
	# Edit on release
	# ----------------------------------------------------------------------
	def release1(self, event):
		if self.focus_get() is not self._text:
			self._text.focus_set()
			self._text.icursor(f"@{event.x}")

	# ----------------------------------------------------------------------
	def postList(self, event=None):
		if self._arrowBtn.cget("state") != tk.DISABLED:
			self._post.set(True)
		return "break"

	# ----------------------------------------------------------------------
	def unpostList(self, event=None):
		self._listbox.reset()
		if self._arrowBtn.cget("state") != tk.DISABLED:
			self._post.set(False)
		return "break"

	# ----------------------------------------------------------------------
	def _togglePost(self, event):
		if self._text.cget("state") != tk.DISABLED:
			self._post.set( not self._post.get() )
		return "break"

	# ----------------------------------------------------------------------
	def _key(self, event):
		if event.char.isalnum():
			self.postList()
			self._listbox._key(event)
			return "break"

	# ----------------------------------------------------------------------
	def _focusOut(self, event):
		try:
			f = self._popup.focus_get()
		except KeyError:
			pass
		else:
			if f in (self._popup, self._listbox): return
		self._focus = None
		self.unpostList()

	# ----------------------------------------------------------------------
	def _selectUnpost(self, event=None):
		if self._post.get():
			if self._listbox.size():
				sel = self._listbox.get(tk.ACTIVE)
				self.set(sel)
			# FIXME to be equivalent to the ttk.Combobox behavior
			#if isinstance(self._text, tk.Entry):
			#	self._text.selection_range(0, tk.END)
			self.unpostList()

	# ----------------------------------------------------------------------
	def invoke(self):
		if self.command is not None:
			self.command()

	# ----------------------------------------------------------------------
	def _release(self, event):
		act = self._listbox.nearest(event.y)
		self._listbox.activate(act)
		self._selectUnpost()

	# ----------------------------------------------------------------------
	def popupClick(self, event):
		if event.x < 0 or event.y < 0 or \
		   event.x > self._popup.winfo_width() or \
		   event.y > self._popup.winfo_height():
			self.unpostList()

	# ----------------------------------------------------------------------
	# The following methods are called before the show of the list ...
	# ----------------------------------------------------------------------
	def beforeShow(self):
		pass

	# ----------------------------------------------------------------------
	# ... and after hide it
	# The user should override them in case some special treatment is needed
	# ----------------------------------------------------------------------
	def afterHide(self):
		pass

	# ----------------------------------------------------------------------
	# Public methods
	# ----------------------------------------------------------------------
	def get(self, first=None, last=None):
		if first is None:
			if isinstance(self._text, tk.Label):
				return self._text.cget("text")
			else:
				return self._text.get()
		else:
			return self._listbox.get(first, last)

	# ----------------------------------------------------------------------
	def set(self, txt, invoke=True):
		if isinstance(self._text, tk.Label):
			self._text.config(text=txt)
		else:
			self._text.delete(0, tk.END)
			self._text.insert(0, txt)
		if invoke: self.invoke()

	# ----------------------------------------------------------------------
	def size(self):
		return self._listbox.size()

	# ----------------------------------------------------------------------
	def clear(self):
		self.clearLabel()
		self.clearList()

	# ----------------------------------------------------------------------
	def clearLabel(self):
		if isinstance(self._text, tk.Label):
			self._text.config(text="")
		else:
			self._text.delete(0, tk.END)

	# ----------------------------------------------------------------------
	def clearList(self):
		self._listbox.delete(0, tk.END)

	# ----------------------------------------------------------------------
	def insert(self, index, *elements):
		self._listbox.insert(index, *elements)

	# ----------------------------------------------------------------------
	def delete(self, first, last=None):
		self._listbox.delete(first, last)

	# ----------------------------------------------------------------------
	def fill(self, items):
		self.clearList()
		for item in items:
			self._listbox.insert(tk.END, item)

	# ----------------------------------------------------------------------
	def select(self, index=None):
		if index is None:
			txt = self.get()
			for i in range(self.size()):
				if txt == self._listbox.get(i):
					return i
			return -1
		elif 0 <= index < self._listbox.size():
			self.set(self._listbox.get(index))

	# ----------------------------------------------------------------------
	def configure(self, **kw):
		if "command" in kw:
			self.command = kw.get("command")
			del kw["command"]
		self._text.configure(**kw)
		self._arrowBtn.configure(**kw)
	config = configure

	# ----------------------------------------------------------------------
	def __setitem__(self, key, value):
		self.configure(**{key:value})

	# ----------------------------------------------------------------------
	def cget(self, key):
		return self._text.cget(key)
	__getitem__ = cget

	# ----------------------------------------------------------------------
	def bindWidgets(self, event, func):
		self._text.bind(event, func)
		self._arrowBtn.bind(event, func)

#===============================================================================
# ExOptionMenu
#===============================================================================
class ExOptionMenu(tk.OptionMenu):
	def __init__(self, master, variable, value, *values, **kwargs):
		super().__init__(master, variable, value,
			*values, **kwargs)
		self.variable = variable
		self.command  = kwargs.get("command")

	# ----------------------------------------------------------------------
	def delete(self, from_=0, to_=tk.END):
		"""Delete items from menu"""
		self["menu"].delete(from_, to_)

	# ----------------------------------------------------------------------
	def add(self, value):
		"""Add an extra value to the menu"""
		menu = self["menu"]
		menu.add_command(label=value,
			command=tk._setit(self.variable, value, None))

	# ----------------------------------------------------------------------
	def set(self, valueList, value=None):
		"""
		clear and reload the menu with a new set of options.
		valueList - list of new options
		value - initial value to set the optionmenu's menubutton to
		"""
		self['menu'].delete(0, tk.END)
		for item in valueList:
			self['menu'].add_command(label=item,
				command=tk._setit(self.variable, item, self.command))
		if value:
			self.variable.set(value)

#===============================================================================
# Splitter Frame
#===============================================================================
class Splitter(tk.Frame):
	"""Base class for horizontal or vertical frame splitter"""
	def __init__(self, master, split=0.5, horizontal=True,
				absolute=False, sash=False, sash_width=8):
		super().__init__(master, class_="Splitter")

		self.f1 = tk.Frame(self, bd=0, relief=tk.SUNKEN)
		self.f2 = tk.Frame(self, bd=0, relief=tk.SUNKEN)
		self.dragFrame = tk.Frame(self, bd=2, relief=tk.GROOVE)
		self.dragFrame.bind("<B1-Motion>",       self.motion)      # Overridden
		self.dragFrame.bind("<ButtonRelease-1>", self.stopMotion)  # Overridden
		self.dragFrame.bind("<Double-Button-1>", self.toggle)

		if sash:
			self.sash = tk.Button(self,
						bd=0,
						activebackground="LightYellow",
						takefocus=False,
						command=self.toggle)
			self.sash.bind("<B1-Motion>",       self.motion)      # Overridden
			self.sash.bind("<ButtonRelease-1>", self.stopMotion)  # Overridden
		else:
			self.sash = False

		self.split    = split
		self.save     = split
		self.absolute = absolute
		self.dragFrame_size = sash_width
		self.sash_size      = 18
		self._inmotion = False
		self.setRange()
		self.setOrientation(horizontal)
		if self.absolute:
			self.bind("<Configure>", self.placeChilds)

	# ----------------------------------------------------------------------
	def orient(self):      return self._hori
	def firstFrame(self):  return self.f1
	def secondFrame(self): return self.f2

	# ----------------------------------------------------------------------
	def setOrientation(self, horizontal=True):
		self._hori = horizontal	# True horizontal / False vertical
		self.f1.place_forget()
		self.f2.place_forget()
		if self._hori:
			self.dragFrame["cursor"] = "sb_h_double_arrow"
		else:
			self.dragFrame["cursor"] = "sb_v_double_arrow"
		self.placeChilds()

	# ----------------------------------------------------------------------
	def swapOrient(self):
		self.dragFrame.place_forget()
		self.setOrientation(not self._hori)

	# ----------------------------------------------------------------------
	def equal(self):
		self._setSplit(0.5)
		self.placeChilds()

	# ----------------------------------------------------------------------
	def minimize(self):
		self._setSplit(0.0)
		self.placeChilds()

	# ----------------------------------------------------------------------
	def maximize(self):
		self._setSplit(1.0)
		self.placeChilds()

	# ----------------------------------------------------------------------
	# Toggle position normally with double click
	# ----------------------------------------------------------------------
	def toggle(self, event=None):
		if self.absolute:
			if self.save == self.split: self.save = 100
			if self.split > self.sash_size:
				self.save = self.split
				self.split = 1
			else:
				self.split = self.save
		else:
			if self.save == self.split: self.save = 0.3
			if self.split <= self.min or self.split >= self.max:
				self.split = self.save
			elif self.split < 0.5:
				self.split = self.min
			else:
				self.split = self.max
		self.placeChilds()

	# ----------------------------------------------------------------------
	# Set acceptable range
	# ----------------------------------------------------------------------
	def setRange(self, _min=0.0, _max=1.0):
		if _min<0.01: _min=0.0
		if _max>0.99: _max=1.0
		self.margin = 5	# pixels on absolute
		self.min = _min
		self.max = _max

	# ----------------------------------------------------------------------
	def _setSplit(self, newSplit):
		if newSplit == self.split: return
		if self.absolute:
			if newSplit <= self.margin: newSplit = self.margin
			if self._hori:
				if newSplit + self.margin >= self.winfo_width():
					newSplit = self.winfo_width() - self.margin
			else:
				if newSplit + self.margin >= self.winfo_height():
					newSplit = self.winfo_height() - self.margin
		else:
			if newSplit <= self.min: newSplit = self.min
			if newSplit >= self.max: newSplit = self.max
		self.save  = self.split
		self.split = newSplit

	# ----------------------------------------------------------------------
	# Set the split position
	# ----------------------------------------------------------------------
	def setSplit(self, newSplit):
		"""Change the spliting position"""
		self._setSplit(newSplit)
		self.placeChilds()

	# ----------------------------------------------------------------------
	def motion(self, event):
		self._inmotion = True
		if self.absolute:
			if self._hori:
				# Horizontal
				self._setSplit(event.x_root - self.winfo_rootx())
				self.dragFrame.place(x=self.split,
						relheight=1.0,
						width=self.dragFrame_size)
				if self.sash:
					self.sash.place(
						x=max(0,self.split+self.dragFrame_size-self.sash_size),
						y=4,
						width=self.sash_size,
						height=self.sash_size)
			else:
				pass
		else:
			if self._hori:
				# Horizontal
				self._setSplit(float(event.x_root - self.winfo_rootx()) /
						float(self.winfo_width()))
				self.dragFrame.place(relx=self.split, x=-2,
						relheight=1.0,
						width=self.dragFrame_size)
			else:
				# Vertical
				self._setSplit(float(event.y_root - self.winfo_rooty()) /
						float(self.winfo_height()))
				self.dragFrame.place(rely=self.split, y=-2,
						relwidth=1.0,
						height=self.dragFrame_size)

	# ----------------------------------------------------------------------
	def stopMotion(self, event):
		if self._inmotion:
			self._inmotion = False
			self.placeChilds()
			return "break"

	# ----------------------------------------------------------------------
	# Place the two frames
	# ----------------------------------------------------------------------
	def placeChilds(self, event=None):
		"""(Re)place the two frames"""
		self._inmotion = False
		if self.absolute:
			if self._hori:
				# Horizontal
				self.f1.place(	relx=0.0,
						width=self.split-1,
						relheight=1.0)
				self.f2.place(	x=self.split+self.dragFrame_size,
						width=self.winfo_width()-self.split-self.dragFrame_size,
						relheight=1.0)
				self.dragFrame.place(x=self.split,
						relheight=1.0,
						width=self.dragFrame_size)
				if self.sash:
					self.sash.place(x=max(0,self.split+self.dragFrame_size-self.sash_size),
						y=4,
						width=self.sash_size,
						height=self.sash_size)
					self.sash["text"] = "\N{BLACK LEFT-POINTING SMALL TRIANGLE}" \
							if self.split>=self.sash_size \
							else "\N{BLACK RIGHT-POINTING SMALL TRIANGLE}"
			else:
				# Vertical
				self.f1.place(	rely=0.0,
						height=self.split,
						relwidth=1.0)
				self.f2.place(	y=self.split+self.dragFrame_size,
						height=self.winfo_height()-self.split()-self.dragFrame_size,
						relwidth=1.0)
				self.dragFrame.place(y=self.split-1,
						relwidth=1.0,
						height=self.dragFrame_size)
				if self.sash:
					self.sash.place(x=4,
						y=max(0,self.split+self.dragFrame_size-self.sash_size),
						width=self.sash_size,
						height=self.sash_size)
					self.sash["text"] = "\N{BLACK UP-POINTING SMALL TRIANGLE}" \
							if self.split>1 \
							else "\N{BLACK DOWN-POINTING SMALL TRIANGLE}"
		else:
			if self._hori:
				# Horizontal
				self.f1.place(	relx=0.0,
						relwidth=self.split,
						relheight=1.0)
				self.f2.place(	relx=self.split,
						x=3,
						relwidth=1.0-self.split,
						relheight=1.0)
				self.dragFrame.place(relx=self.split,
						x=-1,
						relheight=1.0,
						width=self.dragFrame_size)
			else:
				# Vertical
				self.f1.place(	rely=0.0,
						relheight=self.split,
						relwidth=1.0)
				self.f2.place(	rely=self.split,
						y=2,
						relheight=1.0-self.split,
						relwidth=1.0)
				self.dragFrame.place(rely=self.split,
						y=-2,
						relwidth=1.0,
						height=self.dragFrame_size)

#===============================================================================
# Horizontal Splitter
#===============================================================================
class HSplitter(Splitter):
	"""Horizontal frame splitter"""
	def __init__(self, master, split=0.5, absolute=False, sash=False):
		super().__init__(master, split, True, absolute, sash)

	# ----------------------------------------------------------------------
	def leftFrame(self):  return self.firstFrame()
	left = leftFrame
	def rightFrame(self): return self.secondFrame()
	right = rightFrame

#===============================================================================
# Vertical Splitter
#===============================================================================
class VSplitter(Splitter):
	"""Vertical frame splitter"""
	def __init__(self, master, split=0.5, absolute=False, sash=False):
		super().__init__(master, split, False, absolute, sash)

	# ----------------------------------------------------------------------
	def topFrame(self):    return self.firstFrame()
	top = topFrame
	def bottomFrame(self): return self.secondFrame()
	bottom = bottomFrame

#===============================================================================
# Splitter Node
#-------------------------------------------------------------------------------
class _SplitNode:
	def __init__(self, parent, widget, pos=0.5, hori=True):
		self.parent = parent	# Parent of node
		self.left   = None	# Left child node
		self.right  = None	# Right child node (None if end node)
		self.pos    = pos	# Splitting position (<0.0 inverts hori)
		self.hori   = hori	# Horizontal splitting (Vertical frames)
		if self.pos < 0.0:
			self.pos = -self.pos
			self.hori = not self.hori
		self.child  = widget
		self.split  = None	# drawing frame
		self._xy    = 0.0	# Absolute limits of window size for splitter
		self._wh    = 1.0

	# ----------------------------------------------------------------------
	def end(self):	return self.child is not None
	def full(self):	return self.left is not None and self.right is not None

	# ----------------------------------------------------------------------
	def getpos(self):
		if self.hori:
			return self.pos
		else:
			return -self.pos

	# ----------------------------------------------------------------------
	def setCursor(self):
		if self.split:
			if self.hori:
				self.split["cursor"] = "sb_h_double_arrow"
			else:
				self.split["cursor"] = "sb_v_double_arrow"

	# ----------------------------------------------------------------------
	def makeSplit(self, master, drag):
		self.split = tk.Frame(master, bd=1, relief=tk.GROOVE)
		self.split.bind("<B1-Motion>",drag)
		#split.bind("<ButtonRelease-1>", self.placeChilds)
		#split.bind("<Double-Button-1>", self.toggle)
		self.setCursor()

	# ----------------------------------------------------------------------
	def printNode(self, depth):
		if self.left: self.left.printNode(depth+1)
		if self.child:
			log.say("   "*depth, self.child, self.child["bg"])
		else:
			log.say("   "*depth, " ======== H=",self.hori," pos=",self.pos)
		if self.right: self.right.printNode(depth+1)

#===============================================================================
# Tree Splitter allows any nesting of splitting using a tree structure
#===============================================================================
class TreeSplitter(tk.Frame):
	"""Splitter using a tree structure"""
	def __init__(self, master, **kw):
		super().__init__(master, class_="TreeSplitter", **kw)
		self.tree      = None
		self.width     =  3
		self.border    = 0.01
		self._maxchild = None
		self._splitters= {}		# Dictionary of splitters for faster lookup
		self._drag        = None
		self._dragFrame   = None
		self._drag_x_root = -1
		self._drag_y_root = -1

	# ----------------------------------------------------------------------
	def isempty(self): return self.tree is None

	# ----------------------------------------------------------------------
	# Add a new node under a parent
	# ----------------------------------------------------------------------
	def add(self, parent, child, pos=0.5, hori=True):
		node = self.node(child)
		if node is not None: return node
		if isinstance(parent, tk.Widget):
			parent = self.node(parent)
		node = self._add(parent, child, pos, hori)
		self.placeChilds()
		return node

	# ----------------------------------------------------------------------
	def _add(self, parent, child, pos=0.5, hori=True):
		node = _SplitNode(parent, child, pos, hori)

		if parent is None:
			# Add to top
			if self.tree is None:
				self.tree = node	# Set up the root node
				self._maxchild = child
			else:
				oldtree = self.tree
				self.tree = parent = _SplitNode(None, None, pos, hori)

				parent.left = node
				node.parent = parent

				parent.right   = oldtree
				oldtree.parent = parent
				self._maxchild = None

		else:
			if parent.end():	# End node with only a child
				# Keep parent the same and make a new node for both children
				parent.left  = _SplitNode(parent, parent.child, pos, hori)
				parent.right = node
				parent.child = None
			else:
				raise Exception("Parent node is full")
			self._maxchild = None

		if parent and parent.child is None:
			parent.makeSplit(self, self.dragSplitter)
			self._splitters[parent.split] = parent

		self.placeChilds()
		return node

	# ----------------------------------------------------------------------
	# Remove a child from the frame
	# ----------------------------------------------------------------------
	def remove(self, node):
		if isinstance(node,tk.Widget):
			node = self.node(node)
			if node is None: return
		if node.child is self._maxchild: self._maxchild = None
		self._remove(node)
		self.placeChilds()

	# ----------------------------------------------------------------------
	def _remove(self, node):
		if not node.end():
			raise Exception("Only end nodes can be removed")

		if node.child is not None: node.child.place_forget()
		if node.split is not None:
			del self._splitters[node.split]
			node.split.place_forget()

		parent = node.parent
		if parent is None:
			self.tree = None

		elif parent.right is node:
			# re-parent left node
			grandpa = parent.parent
			if grandpa is None:	# root tree
				self.tree = parent.left
			elif grandpa.left is parent:
				grandpa.left = parent.left
			else:
				grandpa.right = parent.left
			parent.left.parent = grandpa
			if parent.split is not None:
				del self._splitters[parent.split]
				parent.split.place_forget()

		elif parent.left is node:
			# re-parent right node
			grandpa = parent.parent
			if grandpa is None:	# root tree
				self.tree = parent.right
			elif grandpa.left is parent:
				grandpa.left = parent.right
			else:
				grandpa.right = parent.right
			parent.right.parent = grandpa
			if parent.split is not None:
				del self._splitters[parent.split]
				parent.split.place_forget()

		else:
			raise Exception("TreeSplitter is broken")

	# ----------------------------------------------------------------------
	# Replace the child of the node
	# ----------------------------------------------------------------------
	def replace(self, node, child):
		place = node.child.place_info()
		if self._maxchild is node.child:
			self._maxchild = child
		node.child.place_forget()
		node.child = child
		node.child.place(**place)

	# ----------------------------------------------------------------------
	# Clean up the whole tree
	# ----------------------------------------------------------------------
	def removeAll(self):
		self.__remove(self.tree)
		self.tree = None
		self._splitters = {}
		self._maxchild = None

	# ----------------------------------------------------------------------
	def __remove(self, node):
		if node is None: return
		self.__remove(node.left)
		self.__remove(node.right)

		if node.child is not None: node.child.place_forget()
		if node.split is not None: node.split.place_forget()

	# ----------------------------------------------------------------------
	# Depending on rpn
	# if None: Return RPN expression of the tree
	# else:    Create the tree from the rpn expression
	# ----------------------------------------------------------------------
	def RPN(self, rpn=None):
		if rpn is not None:
			self.removeAll()
			stack = []
			for item in rpn:
				if isinstance(item, tk.Widget):
					stack.append(_SplitNode(None, item))
				else:
					try:
						right = stack.pop()
						left  = stack.pop()
					except IndexError:
						break

					node = _SplitNode(None, None, item)

					node.makeSplit(self, self.dragSplitter)
					self._splitters[node.split] = node

					node.left    = left
					left.parent  = node
					node.right   = right
					right.parent = node

					stack.append(node)
			try:
				self.tree = stack.pop()
				self.placeChilds()
			except IndexError:
				pass
		else:
			rpn = []
			self.__rpn(self.tree, rpn)
			return rpn

	# ----------------------------------------------------------------------
	def __rpn(self, node, rpn):
		if node is None: return
		self.__rpn(node.left,  rpn)
		self.__rpn(node.right, rpn)

		if node.child is not None:
			rpn.append(node.child)
		else:
			rpn.append(node.getpos())

	# ----------------------------------------------------------------------
	def printTree(self):
		if self.tree: self.tree.printNode(0)

	# ----------------------------------------------------------------------
	def childs(self):
		"""Return list of child nodes"""
		lst = []
		self.__childNode(self.tree, lst)
		return lst

	# ----------------------------------------------------------------------
	def __childNode(self, node, lst):
		if node is None: return
		self.__childNode(node.left,  lst)
		self.__childNode(node.right, lst)
		if node.child is not None: lst.append(node.child)

	# ----------------------------------------------------------------------
	# return node that has as child the widget
	# ----------------------------------------------------------------------
	def __searchWidget(self, node, widget):
		if node is None: return None
		n = self.__searchWidget(node.left, widget)
		if n is not None: return n
		n = self.__searchWidget(node.right, widget)
		if n is not None: return n
		if node.child is widget: return node
		return None

	# ----------------------------------------------------------------------
	def node(self, widget):
		if widget is None: return None
		return self.__searchWidget(self.tree, widget)

	# ----------------------------------------------------------------------
	def __placeForget(self, node):
		if node is None: return
		if node.split is not None: node.split.place_forget()
		if node.child is not None: node.child.place_forget()
		self.__placeForget(node.left)
		self.__placeForget(node.right)

	# ----------------------------------------------------------------------
	def __placeNode(self, node, x, y, w, h):
		if node is None: return
		#log.say("placeNode", node, node.child)
		#log.say("    ",x, y, w, h)
		if node.end():
			# Place the child
			if x>0.0:   xx =  self.width
			else:	    xx =  0
			if x+w<1.0: ww = -self.width-xx
			else:	    ww =  0
			if y>0.0:   yy =  self.width
			else:	    yy =  0
			if y+h<1.0: hh = -self.width-yy
			else:	    hh =  0
			node.child.place(in_  = self,
					relx      = x,
					x         = xx,
					relwidth  = w,
					width     = ww,
					rely      = y,
					y         = yy,
					relheight = h,
					height    = hh)
			return

		# Place the splitter
		if node.hori:	# Splitting along X => Vertical frames
			node._xy = x
			node._wh = w
			pos = x + w*node.pos
			sw  = pos - x
			if sw <= self.border:
				pos = min(x+self.border, x+w-self.border)
				sw  = pos - x
			if y>0.0:   yy =  self.width
			else:	    yy =  0
			if y+h<1.0: hh = -self.width
			else:	    hh =  0
			node.split.place(in_   = self,
					relx      = pos,
					x         = -self.width,
					relwidth  = 0.0,
					width     = 2*self.width,
					rely      = y,
					y         = yy,
					relheight = h,
					height    = hh)

			self.__placeNode(node.left,  x, y, sw, h)
			self.__placeNode(node.right, x+sw, y, w-sw, h)

		else:	# Splitting along Y => Horizontal frames
			node._xy = y
			node._wh = h
			pos = y + h*node.pos
			sh  = pos - y
			if sh <= self.border:
				pos = min(y+self.border, y+w-self.border)
				sh  = pos - y
			if x>0.0:   xx =  self.width
			else:	    xx =  0
			if x+w<1.0: ww = -self.width
			else:	    ww =  0
			node.split.place(in_   = self,
					relx      = x,
					x         = xx,
					relwidth  = w,
					width     = ww,
					rely      = pos,
					y         = -self.width,
					relheight = 0.0,
					height    = 2*self.width)

			self.__placeNode(node.left,  x, y, w, sh)
			self.__placeNode(node.right, x, y+sh, w, h-sh)

	# ----------------------------------------------------------------------
	# Place the frames [1..3]
	# ----------------------------------------------------------------------
	def placeChilds(self):
		if self.tree is None: return
		if self._maxchild is not None:
			self.__placeForget(self.tree)
			self._maxchild.place(in_ = self,
				relx=0.0, x=0, relwidth=1.0, width=0,
				rely=0.0, y=0, relheight=1.0, height=0)
		else:
			self.__placeNode(self.tree, 0.0, 0.0, 1.0, 1.0)

	# ----------------------------------------------------------------------
	# drag splitter and reposition childs
	# ----------------------------------------------------------------------
	def dragSplitter(self, event):
		node = self._splitters[event.widget]
		if node.hori:
			pos = float(event.x_root - self.winfo_rootx()) / \
				float(self.winfo_width())
		else:
			pos = float(event.y_root - self.winfo_rooty()) / \
				float(self.winfo_height())

		# Absolute positioning
		pos = min(max(pos,self.border),1.0-self.border)

		# Convert to relative
		node.pos = (pos - node._xy) / node._wh

		self.placeChilds()

	# ----------------------------------------------------------------------
	def maximize(self, child=None):
		if self._maxchild is child:
			self._maxchild = None
		else:
			self._maxchild = child
		self.placeChilds()

	# ----------------------------------------------------------------------
	def maxchild(self):
		return self._maxchild

	# ----------------------------------------------------------------------
	# return node containing x,y position in absolute coordinates
	# ----------------------------------------------------------------------
	def nodeContaining(self, node, x, y):
		if node is None: return None

		n = self.nodeContaining(node.left, x, y)
		if n is not None: return n

		n = self.nodeContaining(node.right, x, y)
		if n is not None: return n

		if node.child is not None:
			rx = node.child.winfo_rootx()
			ry = node.child.winfo_rooty()
			if rx <= x <= rx+node.child.winfo_width() and \
			   ry <= y <= ry+node.child.winfo_height():
				return node
		return None

	# ----------------------------------------------------------------------
	# Find position in screen
	# ----------------------------------------------------------------------
	def reposition(self, node, event):
		if self._maxchild is not None: return False
		# First check absolute placement
		x = float(event.x_root - self.winfo_rootx()) / float(self.winfo_width())
		y = float(event.y_root - self.winfo_rooty()) / float(self.winfo_height())
		if x<0.0 or x>1.0 or y<0.0 or y>1.0:
			return
		if x<0.1 or x>0.9 or y<0.1 or y>0.9:
			x1 = 1.0-x
			y1 = 1.0-y

			self._remove(node)
			newnode = self._add(None, node.child)
			parent = self.tree

			if x>y and x>y1:
				# Move to right
				parent.hori = True
				if parent.right is not None: # Swap to move to TOP
					parent.left, parent.right = parent.right, parent.left

			elif x1>y and x1>y1:
				# Move to left
				parent.hori = True

			elif y>x and y>x1:
				# Move to bottom
				parent.hori = False
				if parent.right is not None: # Swap to move to TOP
					parent.left, parent.right = parent.right, parent.left

			else:
				# Move to top
				parent.hori = False
			parent.setCursor()
			self.placeChilds()
			return True

		# Place inside another widget
		overnode = self.nodeContaining(self.tree, event.x_root, event.y_root)
		if overnode is None or overnode is node: return False
		overwidget = overnode.child

		# Then inside other widgets
		x = float(event.x_root - overwidget.winfo_rootx()) / float(overwidget.winfo_width())
		y = float(event.y_root - overwidget.winfo_rooty()) / float(overwidget.winfo_height())
		x1 = 1.0-x
		y1 = 1.0-y
		if 0.4<x<0.6 and 0.4<y<0.6:
			# Swap children
			overnode.child, node.child = node.child, overnode.child

		else:
			self._remove(node)
			overnode = self.node(overwidget)	# Maybe it has changed
			newnode  = self._add(overnode, node.child)
			parent = newnode.parent

			if x>y and x>y1:
				# Move to right
				parent.hori = True

			elif x1>y and x1>y1:
				# Move to left
				parent.hori = True
				if parent.right is not None: # Swap to move to TOP
					parent.left, parent.right = parent.right, parent.left

			elif y>x and y>x1:
				# Move to bottom
				parent.hori = False

			else:
				# Move to top
				parent.hori = False
				if parent.right is not None: # Swap to move to TOP
					parent.left, parent.right = parent.right, parent.left
			parent.setCursor()

		self.placeChilds()
		return True

	# ----------------------------------------------------------------------
	# Event handlers for dragging and placing
	# Bind to <Button-1>
	# ----------------------------------------------------------------------
	def dragStart(self, event):
		self._drag  = None
		self._dragFrame = None
		self._drag_x_root = event.x_root
		self._drag_y_root = event.y_root

	# ----------------------------------------------------------------------
	# Bind to <B1-Motion>
	# ----------------------------------------------------------------------
	def dragMove(self, event):
		if self._maxchild is not None: return
		if self.tree.child is not None: return	# Only one node
		if self._drag is None:
			if abs(self._drag_x_root - event.x_root)>10 or \
			   abs(self._drag_y_root - event.y_root)>10:
				self["cursor"] = "hand1"
				self._drag = self.nodeContaining(self.tree,
							self._drag_x_root,
							self._drag_y_root)
				if self._drag:
					self._dragFrame = tk.Frame(self._drag.child.master,
								relief=tk.RIDGE,
								borderwidth=2*self.width,
								bg="LightYellow")

		elif self._dragFrame is not None:
			# First check absolute placement
			sx = float(self.winfo_rootx())
			sy = float(self.winfo_rooty())
			sw = float(self.winfo_width())
			sh = float(self.winfo_height())
			x = (float(event.x_root) - sx) / sw
			y = (float(event.y_root) - sy) / sh

			if x<0.0 or x>1.0 or y<0.0 or y>1.0:
				self._dragFrame.place_forget()
				return

			if x<0.1 or x>0.9 or y<0.1 or y>0.9:
				x1 = 1.0-x
				y1 = 1.0-y

				if x>y and x>y1:
					# Move to right
					self._dragFrame.place(in_=self,
						relx = 0.5,
						relwidth = 0.5,
						rely = 0.0,
						relheight = 1.0)

				elif x1>y and x1>y1:
					# Move to left
					self._dragFrame.place(in_=self,
						relx = 0.0,
						relwidth = 0.5,
						rely = 0.0,
						relheight = 1.0)

				elif y>x and y>x1:
					# Move to bottom
					self._dragFrame.place(in_=self,
						relx = 0.0,
						relwidth = 1.0,
						rely = 0.5,
						relheight = 0.5)

				else:
					# Move to top
					self._dragFrame.place(in_=self,
						relx = 0.0,
						relwidth = 1.0,
						rely = 0.0,
						relheight = 0.5)
				self._dragFrame.lift()
				return

			# Test inside a widget
			over = self.nodeContaining(self.tree,
					event.x_root,
					event.y_root)

			if over is None or over is self._drag:
				self._dragFrame.place_forget()
				return

			overwidget = over.child

			# Then inside other widgets
			wx = float(overwidget.winfo_rootx())
			wy = float(overwidget.winfo_rooty())
			ww = float(overwidget.winfo_width())
			wh = float(overwidget.winfo_height())

			x = (float(event.x_root) - wx) / ww
			y = (float(event.y_root) - wy) / wh
			x1 = 1.0-x
			y1 = 1.0-y
			if 0.4<x<0.6 and 0.4<y<0.6:
				# Swap children
				self._dragFrame.place(in_=self,
					relx      = (wx-sx)/sw,
					relwidth  = ww/sw,
					rely      = (wy-sy)/sh,
					relheight = wh/sh)

			else:
				if x>y and x>y1:
					# Move to left
					self._dragFrame.place(in_=self,
						relx      = (wx+ww/2.0-sx)/sw,
						relwidth  = ww/sw/2.0,
						rely      = (wy-sy)/sh,
						relheight = wh/sh)

				elif x1>y and x1>y1:
					# Move to right
					self._dragFrame.place(in_=self,
						relx      = (wx-sx)/sw,
						relwidth  = ww/sw/2.0,
						rely      = (wy-sy)/sh,
						relheight = wh/sh)

				elif y>x and y>x1:
					# Move to bottom
					self._dragFrame.place(in_=self,
						relx      = (wx-sx)/sw,
						relwidth  = ww/sw,
						rely      = (wy+wh/2.0-sy)/sh,
						relheight = wh/sh/2.0)

				else:
					# Move to top
					self._dragFrame.place(in_=self,
						relx      = (wx-sx)/sw,
						relwidth  = ww/sw,
						rely      = (wy-sy)/sh,
						relheight = wh/sh/2.0)

	# ----------------------------------------------------------------------
	# Bind to <ButtonRelease-1>
	# ----------------------------------------------------------------------
	def dragEnd(self, event):
		if self._maxchild is not None: return
		if self._dragFrame is None: return
		if self._drag:
			self["cursor"] = ""
			self._dragFrame.place_forget()
			self._dragFrame = None
			return self.reposition(self._drag, event)
		return False

#=============================================================================
# Display a balloon message (only static methods)
#=============================================================================
class Balloon:
	_top       = None
	_widget    = None
	font       = ("Helvetica","-12")
	foreground = "Black"
	background = "LightYellow"
	delay      = 1500
	x_mouse    = 0
	y_mouse    = 0
	hideafter  = None
	_first     = True

	# ----------------------------------------------------------------------
	# set a balloon message to a widget
	# ----------------------------------------------------------------------
	@staticmethod
	def set(widget, help_):
		widget._help = help_
		if Balloon._first:
			Balloon._first = False
			widget.bind_all('<Any-Enter>',  Balloon.enter)
			widget.bind_all('<Any-Leave>',  Balloon.leave)
			widget.bind_all('<KeyPress>',   Balloon.hide)

	# ----------------------------------------------------------------------
	@staticmethod
	def sethelp(widget, help_):
		widget._help = help_

	# ----------------------------------------------------------------------
	# Mouse entered widget show the balloon after a delay
	# ----------------------------------------------------------------------
	@staticmethod
	def enter(event):
		if Balloon._widget is event.widget: return
		if not hasattr(event.widget, "_help"): return
		Balloon._widget = event.widget
		Balloon.x_mouse = event.x_root
		Balloon.y_mouse = event.y_root
		return event.widget.after(Balloon.delay, Balloon.show)

	# ----------------------------------------------------------------------
	# Mouse left widget hide the balloon
	# ----------------------------------------------------------------------
	@staticmethod
	def leave(event=None):
		if Balloon.hideafter:
			if Balloon._top: Balloon._top.after_cancel(Balloon.hideafter)
			Balloon.hideafter = None
		Balloon._widget = None
		if Balloon._top is None: return
		try:
			if Balloon._top.winfo_ismapped():
				Balloon._top.withdraw()
		except tk.TclError:
			Balloon._top = None
	hide=leave

	# ----------------------------------------------------------------------
	@staticmethod
	def setWidget(widget, x, y):
		Balloon._widget = widget
		Balloon.x_mouse = x
		Balloon.y_mouse = y

	# ----------------------------------------------------------------------
	# @return true if balloon is shown
	# ----------------------------------------------------------------------
	@staticmethod
	def isshown():
		try:
			return Balloon._top is not None and Balloon._top.winfo_ismapped()
		except tk.TclError:
			return False

	# ----------------------------------------------------------------------
	# display the balloon help
	# ----------------------------------------------------------------------
	@staticmethod
	def show(hideafter=None):
		try:
			if Balloon._widget is None: return
			if Balloon._top is None:
				Balloon._top = tk.Toplevel()
				Balloon._top.overrideredirect(1)
				Balloon._msg = tk.Message(Balloon._top,
						aspect=300,
						foreground=Balloon.foreground,
						background=Balloon.background,
						relief=tk.SOLID,
						borderwidth=1,
						font=Balloon.font)
				Balloon._msg.pack()
				Balloon._top.bind("<1>",Balloon.hide)
			Balloon._msg.config(text=Balloon._widget._help)
			# Guess position
			x = Balloon._widget.winfo_rootx() + Balloon._widget.winfo_reqwidth()//2
			y = Balloon._widget.winfo_rooty() + Balloon._widget.winfo_reqheight()+5
			# if too far away use mouse
			if abs(x - Balloon.x_mouse) > 30:
				x = Balloon.x_mouse + 20
			if abs(y - Balloon.y_mouse) > 30:
				y = Balloon.y_mouse + 10
			# Force size recalculation
			#Balloon._top.wm_geometry("")
			Balloon._top.wm_geometry(f"+{x}+{y}")
			Balloon._top.deiconify()
			Balloon._top.lift()
			Balloon._top.update_idletasks()

			# Check if it is hidden on bottom-right sides
			move = False
			if Balloon._top.winfo_rootx() + Balloon._top.winfo_width() >= \
			   Balloon._top.winfo_screenwidth():
				x = Balloon._top.winfo_screenwidth() \
					- Balloon._top.winfo_width() - 20
				move = True
			if Balloon._top.winfo_rooty() + Balloon._top.winfo_height() >= \
			   Balloon._top.winfo_screenheight():
				y = Balloon._top.winfo_screenheight() \
					- Balloon._top.winfo_height() - 10
				move = True
			if move:
				Balloon._top.wm_geometry(f"+{x}+{y}")
			if Balloon.hideafter:
				Balloon._top.after_cancel(Balloon.hideafter)
				Balloon.hideafter = None
			if hideafter is not None:
				Balloon._hideafter = Balloon._top.after(hideafter, Balloon.hide)
		except tk.TclError:
			Balloon._top = None

#===============================================================================
# A LabelFrame that can collapse/expand
#===============================================================================
class ExLabelFrame(tk.LabelFrame):
	def __init__(self, master, *args, **kwargs):
		if "command" in kwargs:
			self.command = kwargs.get("command")
			del kwargs["command"]
		else:
			self.command = None

		super().__init__(master, *args, **kwargs)
		self.frame = tk.Frame(self)
		self.frame.pack(expand=tk.YES, fill=tk.BOTH)
		self.bind("<Button-1>", self. click)
		if self["height"]==0:
			self["height"] = 20
		self.width   = self["width"]

	# ----------------------------------------------------------------------
	def click(self, event=None):
		if self.frame.winfo_ismapped():
			self.collapse()
		else:
			self.expand()
		if self.command is not None:
			self.command(event)

	# ----------------------------------------------------------------------
	def collapse(self):
		self["width"] = self.winfo_width()
		self.frame.pack_forget()
		lbl = self["text"]
		if lbl[-1] in ("\N{BLACK UP-POINTING TRIANGLE}", "\N{BLACK DOWN-POINTING TRIANGLE}"):
			lbl = lbl[:-1]
		self["text"] = lbl+"\N{BLACK UP-POINTING TRIANGLE}"

	# ----------------------------------------------------------------------
	def expand(self):
		self["width"] = self.width
		self.frame.pack(fill=tk.BOTH)
		lbl = self["text"]
		if lbl[-1] in ("\N{BLACK UP-POINTING TRIANGLE}", "\N{BLACK DOWN-POINTING TRIANGLE}"):
			self["text"] = lbl[:-1]

	# ----------------------------------------------------------------------
	def isexpanded(self):
		return self.frame.winfo_ismapped()

	# ----------------------------------------------------------------------
	def __call__(self): return self.frame

#================================================================================
# ScrollFrame based on Bruno's implementation
#================================================================================
class ScrollFrame(tk.Frame):
	# ----------------------------------------------------------------------
	def __init__(self, master=None, stretch=True, cnf=None, **kw):
		if cnf is None: cnf={}
		super().__init__(master, cnf, **kw)
		self.client = tk.Frame(self, border=0)

		# width and height of Scrollframe
		self.W = 1.0
		self.H = 1.0

		# top left corner coordinates of client frame
		self.client_x = 0
		self.client_y = 0

		# width and height of client frame
		self.client_w = 1.0
		self.client_h = 1.0

		# scroll commands (default)
		self.xscrollcommand=lambda *args:None
		self.yscrollcommand=lambda *args:None

		# scroll increments
		self.xscrollincrement = 15
		self.yscrollincrement = 15

		# stretches
		self.stretch   = stretch
		self.stretch_x = stretch
		self.stretch_y = stretch

		#self.bind("<Expose>",		self.updateScrollRegion)
		self.bind("<Configure>",	self.updateScrollRegion)

		self.defaultBinds()

		#w = self.client.winfo_toplevel()
		self.mult  = 1.0
		self._drag = None
		self._startx = self._startx = 0

	# ----------------------------------------------------------------------
	def cget(self,item):
		if not hasattr(self,item):
			return tk.Frame.cget(self,item)
		else:
			getattr(self,item)
	__getitem__ = cget

	def __setitem__(self,item,value):self.configure({item:value})

	# ----------------------------------------------------------------------
	def configure(self,cnf=None,**kw):
		if kw: cnf=tk._cnfmerge((cnf,kw))
		for key in cnf.keys():
			if not hasattr(self,key):
				tk.Frame.configure(self,cnf)
			else:
				setattr(self,key,cnf[key])
	config=configure

	# ----------------------------------------------------------------------
	# Use this method to get the parent widget of the frame
	# ----------------------------------------------------------------------
	def __call__(self): return self.client

	# ----------------------------------------------------------------------
	def position(self):
		return self.client_x, self.client_y

	#-------------------------------------------------------------------------------
	@staticmethod
	def bindChilds(widget, event, function, ignore=None):
		if ignore is None: ignore = [event]
		for child in widget.winfo_children():
			ScrollFrame.bindChilds(child, event, function, ignore)
			for e in ignore:
				if child.bind(e) or child.bind_class(child.__class__.__name__,e):
					break
			else:
				child.bind(event, function)

	#-------------------------------------------------------------------------------
	def defaultBinds(self):
		ignore = ["<2>", "<B2-Motion>", "<ButtonRelease-2>"]
		ScrollFrame.bindChilds(self.client, "<B2-Motion>",	self.drag, ignore)
		del ignore[1]	# delete motion that was already assigned
		ScrollFrame.bindChilds(self.client, "<ButtonRelease-2>",self.dragRelease, ignore)
		ScrollFrame.bindChilds(self.client, "<Button-4>",	self.scrollUp)
		ScrollFrame.bindChilds(self.client, "<Button-5>",	self.scrollDown)
		ScrollFrame.bindChilds(self.client, "<Shift-Button-4>",	self.scrollLeft)
		ScrollFrame.bindChilds(self.client, "<Shift-Button-5>",	self.scrollRight)

	# ----------------------------------------------------------------------
	def ischild(self, widget):
		if widget is None: return False
		if widget is self.client: return True
		return self.ischild(widget.master)

	# ----------------------------------------------------------------------
	def drag(self, event):
		if self._drag is not None:
			dx = (event.x_root - self._drag[0])*self.mult
			dy = (event.y_root - self._drag[1])*self.mult
			self.client_x = int(self._start_x + dx)
			self.client_y = int(self._start_y + dy)
			self.updateScrollx()
			self.updateScrolly()
			self.client.place_configure(x=self.client_x, y=self.client_y)
		else:
			if not self.ischild(event.widget): return
			self.config(cursor="hand2")
			self._drag = event.x_root, event.y_root
			self._start_x = self.client_x
			self._start_y = self.client_y
		return "break"

	# ----------------------------------------------------------------------
	def dragRelease(self, event):
		self._drag = None
		self.config(cursor="")
		return "break"

	# ----------------------------------------------------------------------
	def scrollUp(self, event):
		if not self.ischild(event.widget): return
		self.yview(tk.SCROLL, -1, tk.UNITS)
		return "break"

	# ----------------------------------------------------------------------
	def scrollDown(self, event):
		if not self.ischild(event.widget): return
		self.yview(tk.SCROLL,  1, tk.UNITS)
		return "break"

	# ----------------------------------------------------------------------
	def scrollLeft(self, event):
		if not self.ischild(event.widget): return
		self.xview(tk.SCROLL, -1, tk.UNITS)
		return "break"

	# ----------------------------------------------------------------------
	def scrollRight(self, event):
		if not self.ischild(event.widget): return
		self.xview(tk.SCROLL,  1, tk.UNITS)
		return "break"

	# ----------------------------------------------------------------------
	def xview(self, action, value=None, units='pages'):
		if action == "moveto":
			fraction = float(value)
			if fraction <= 0.0:
				self.client_x = 0
			elif fraction >= float(self.client_w-self.W)/self.client_w:
				self.client_x = self.W-self.client_w
			else:
				self.client_x = int(-self.client_w*fraction)

		elif action == "scroll":
			amount=int(value)
			if self.client_x == 0 and amount < 0:return
			if self.W >= self.client_w: return
			if self.client_x == self.W-self.client_w and amount > 0:return
			if units == "units":
				dx = self.xscrollincrement
			else:
				dx = amount*self.W*0.99
			self.client_x -= amount*dx
		else:
			return

		self.updateScrollx()
		self.client.place_configure(x=self.client_x)

	# ----------------------------------------------------------------------
	def yview(self, action, value=None, units='pages'):
		if value is None:
#			return
			value  = action
			action = "moveto"

		if action == "moveto":
			fraction=float(value)
			if fraction <= 0.0:
				self.client_y = 0
			elif fraction >= float(self.client_h-self.H)/self.client_h:
				self.client_y = self.H-self.client_h
			else:
				self.client_y = int(-self.client_h*fraction)

		elif action == "scroll":
			amount=int(value)
			if self.client_y == 0 and amount < 0:return
			if self.H >= self.client_h: return
			if self.client_y == self.H-self.client_h and amount > 0:return
			if units == "units":
				dy = self.yscrollincrement
			else:
				dy = self.H
			self.client_y -= amount*dy

		else:
			return

		self.updateScrolly()
		self.client.place_configure(y=self.client_y)

	# ----------------------------------------------------------------------
	def moveto(self, x, y):
		if x >= 0:
			self.client_x = 0
		elif x <= self.W - self.client_w:
			self.client_x = self.W-self.client_w
		else:
			self.client_x = x

		if y >= 0:
			self.client_y = 0
		elif y <= self.H - self.client_h:
			self.client_y = self.H-self.client_h
		else:
			self.client_y = y

		self.updateScrollx()
		self.updateScrolly()
		self.client.place_configure(x=self.client_x,y=self.client_y)

	# ----------------------------------------------------------------------
	def updateScrollx(self, *args):
		if self.client_x >= 0:
			low = 0.0
		else:
			low = -float(self.client_x)/self.client_w

		if self.client_x+self.client_w <= self.W:
			high = 1.0
		else:
			high = low+float(self.W)/self.client_w

		if low <= 0.0:
			self.client_x=0
		elif high >= 1.0:
			if self.client_w > self.W:
				self.client_x = self.W-self.client_w
				low  = -float(self.client_x)/self.client_w
			else:
				self.client_x = 0
				low = 0.0
			high = low+float(self.W)/self.client_w
		if self.client_w < self.W:
			self.stretch_x = self.stretch
		else:
			self.stretch_x = False
		self.xscrollcommand(low,high)

	# ----------------------------------------------------------------------
	def updateScrolly(self, *args):
		if self.client_y >= 0:
			low = 0.0
		else:
			low = -float(self.client_y)/self.client_h
		if self.client_y+self.client_h <= self.H:
			high = 1.0
		else:
			high = low+float(self.H)/self.client_h

		if low <= 0.0:
			self.client_y = 0
		elif high >= 1.0:
			if self.client_h > self.H:
				self.client_y = self.H-self.client_h
				low  = -float(self.client_y)/self.client_h
			else:
				self.client_y = 0
				low = 0.0
			high = low+float(self.H)/self.client_h

		if self.client_h < self.H:
			self.stretch_y = self.stretch
		else:
			self.stretch_y = False
		self.yscrollcommand(low,high)

	# ----------------------------------------------------------------------
	def updateScrollRegion(self, *args):
		if self.client.children:
			self.client_w = self.client.winfo_reqwidth()
			self.client_h = self.client.winfo_reqheight()
			self.W = self.winfo_width()
			self.H = self.winfo_height()

			self.updateScrolly()
			self.updateScrollx()

			if self.stretch_y:
				h = self.H
			else:
				h = self.client_h

			if self.stretch_x:
				w = self.W
			else:
				w = self.client_w

			self.client.place_configure(
				x=self.client_x,
				y=self.client_y,
				height=h,
				width=w,
				anchor="nw")
		else:
			self.xscrollcommand(0.0,1.0)
			self.yscrollcommand(0.0,1.0)
			self.client.place_forget()

#================================================================================
# The following is from idlelib (tabpage.py)
#================================================================================
class InvalidTabPage(Exception): pass
class AlreadyExists(Exception): pass

#===============================================================================
# A page tab frame button
#===============================================================================
class PageTab(tk.Frame):
	"""
	a 'page tab' like framed button
	"""

	# ----------------------------------------------------------------------
	def __init__(self, parent):
		super().__init__(parent, borderwidth=2, relief=tk.RIDGE)
		self.button=tk.Radiobutton(self, padx=5, pady=2, takefocus=tk.FALSE,
			indicatoron=tk.FALSE, highlightthickness=0,
			borderwidth=0, selectcolor=self.cget('bg'))
		self.button.pack(fill=tk.BOTH)

#===============================================================================
# Tab pages
#===============================================================================
class TabPageSet(tk.Frame):
	"""
	a set of 'pages' with TabButtons for controlling their display
	"""

	# ----------------------------------------------------------------------
	def __init__(self, parent, pageNames=None, top=True, hidetext=False, **kw):
		"""
		pageNames - a list of strings, each string will be the dictionary key
		to a page's data, and the name displayed on the page's tab. Should be
		specified in desired page order. The first page will be the default
		and first active page.
		"""
		super().__init__(parent, kw)
		self.grid_location(0, 0)

		self.tabBar=tk.Frame(self)
		self.top = top
		self.hidetext = hidetext

		if top:
			self.columnconfigure(0, weight=1)
			self.rowconfigure(1, weight=1)
			self.tabBar.grid(row=0, column=0, sticky=tk.EW)
		else:
			self.columnconfigure(1, weight=1)
			self.rowconfigure(0, weight=1)
			self.tabBar.grid(row=0, column=0, sticky=tk.NSEW)

		self.activePage=tk.StringVar(self)
		self.defaultPage=''
		self.pages={}
		if pageNames is not None:
			for name in pageNames:
				if isinstance(name,tuple):
					self.addPage(*name)
				else:
					self.addPage(name)

	# ----------------------------------------------------------------------
	def page(self, name):
		return self.pages[name]['page']

	# ----------------------------------------------------------------------
	def __getitem__(self, name): return self.page(name)

	# ----------------------------------------------------------------------
	def changePage(self, pageName=None):
		if pageName:
			if pageName in self.pages:
				self.activePage.set(pageName)
			else:
				raise InvalidTabPage("Invalid TabPage Name")
		# pop up the active 'tab' only
		for page in self.pages:
			tab = self.pages[page]['tab']
			tab.config(relief=tk.RIDGE)
			tab.button.config(background="DarkGray",
					activebackground="DarkGray")
			if self.hidetext: tab.button.config(text="")

		tab = self.pages[self.getActivePage()]['tab']
		tab.config(relief=tk.RAISED)
		tab.button.config(
			background="LightGray",
			activebackground="LightGray")
		if self.hidetext:
			tab.button.config(text=self.getActivePage())
		# switch page
		self.pages[self.getActivePage()]['page'].lift()

		self.event_generate("<<ChangePage>>") #, data=pageName)

	# ----------------------------------------------------------------------
	def getActivePage(self):
		return self.activePage.get()

	# ----------------------------------------------------------------------
	def addPage(self, pageName, icon=None):
		if pageName in self.pages:
			raise AlreadyExists("TabPage Name Already Exists")

		self.pages[pageName]={
			'tab' : PageTab(self.tabBar),
			'page': tk.Frame(self, borderwidth=2, relief=tk.RAISED) }
		if icon:
			self.pages[pageName]['tab'].button.config(text=pageName,
				image=icon, compound=tk.LEFT)
			self.icons = True
		else:
			self.pages[pageName]['tab'].button.config(text=pageName)
		self.pages[pageName]['tab'].button.config(
				command=self.changePage,
				variable=self.activePage,
				value=pageName)
		if self.top:
			self.pages[pageName]['tab'].pack(side=tk.LEFT)
			self.pages[pageName]['page'].grid(row=1, column=0, sticky=tk.NSEW)
		else:
			self.pages[pageName]['tab'].pack(side=tk.TOP, fill=tk.X)
			self.pages[pageName]['page'].grid(row=0, column=1, sticky=tk.NSEW)

		if len(self.pages)==1: # adding first page
			self.defaultPage=pageName
			self.activePage.set(self.defaultPage)
			self.changePage()

	# ----------------------------------------------------------------------
	def removePage(self, pageName):
		if pageName not in self.pages:
			raise InvalidTabPage("Invalid TabPage Name")
		self.pages[pageName]['tab'].pack_forget()
		self.pages[pageName]['page'].grid_forget()
		self.pages[pageName]['tab'].destroy()
		self.pages[pageName]['page'].destroy()
		del self.pages[pageName]
		# handle removing last remaining, or default, or active page
		if not self.pages: # removed last remaining page
			self.defaultPage=''
			return
		if pageName==self.defaultPage: # set a new default page
			self.defaultPage=\
				self.tabBar.winfo_children()[0].button.cget('text')
		if pageName==self.getActivePage(): # set a new active page
			self.activePage.set(self.defaultPage)
		self.changePage()

	# ----------------------------------------------------------------------
	def renamePage(self, old, new):
		if old not in self.pages:
			raise InvalidTabPage("Invalid TabPage Name")
		self.pages[new] = self.pages[old]
		del self.pages[old]
		self.pages[new]['tab'].button.config(text=new, value=new)
		if old == self.getActivePage():
			self.activePage.set(new)

#===============================================================================
def _test():
	root = tk.Tk()
	frame = tk.Frame(root)
	frame.pack(side=tk.TOP, fill=tk.X)

	p = ProgressBar(frame, background="DarkGray", height=24)
	p.pack(side=tk.TOP, fill=tk.X)

	def addProg(ev):
		p.setProgress(p.getProgress()[0]+10.0)
		p.autoText("Test")
	p.bind('<1>', addProg)

	frame = tk.Frame(root)
	frame.pack(side=tk.BOTTOM, expand=tk.YES, fill=tk.BOTH)
	hsplit  = HSplitter(frame, 0.7)
	vsplitL = VSplitter(hsplit.leftFrame(), 0.5)
	vsplitR = VSplitter(hsplit.rightFrame(), 0.3)

	tk.Label(vsplitL.topFrame(), text='MultiListbox').pack()
	mlb = MultiListbox(vsplitL.topFrame(),
			(('Subject', 40, None),
			 ('Sender',  20, None),
			 ('Date',    10, None)))
	for ii in range(100):
		mlb.insert(tk.END, (f"{ii} Important Message",
				'John Doe', f"10/10/{1900+ii:04d}"))
	mlb.pack(expand=tk.YES, fill=tk.BOTH)

	l = tk.Label(vsplitL.bottomFrame(), text="Combobox")
	l.pack(side=tk.TOP)
	cb = Combobox(vsplitL.bottomFrame(), label=True)
	cb.pack(side=tk.BOTTOM, expand=tk.YES, fill=tk.X)
	cb.fill(("one","two","three","four", "fix-six-seven-eight-nine-ten"))
	cb.select(0)

	tk.Label(vsplitR.topFrame(), text='SearchListbox').pack()
	lb = SearchListbox(vsplitR.topFrame(), selectmode=tk.BROWSE,
		exportselection=tk.FALSE)
	lb.insert(tk.END,"Starting")
	lb.insert(tk.END,"Loading card database")
	lb.insert(tk.END,"Loading isotopes database")
	lb.insert(tk.END,"Layout initialization")
	lb.insert(tk.END,"Layout create Tree list")
	lb.insert(tk.END,"--Initialize Tk")
	lb.insert(tk.END,"After initialization of Tk")
	lb.insert(tk.END,"Creating frames")
	lb.insert(tk.END,"Creation of windows...")
	lb.insert(tk.END,"Writing ini file")
	lb.insert(tk.END,"Exiting program")
#	lb.fill()
	lb.pack(expand=tk.YES, fill=tk.BOTH)
	lb.focus_set()
	SearchBox.ignoreCase     = True
#	SearchBox.ignoreNonAlpha = False

#	v = tk.StringVar()
#	lst = ["One", "Two", "Three", "Four"]
#	v.set("One")
#	o = ExOptionMenu(vsplitR.bottomFrame(), v, *lst)
#	o.pack()
#	o.delete()
#	lst.reverse()
#	for i in lst:
#		o.add(i)

	#test dialog
	frame = vsplitR.bottomFrame()
	tabPage=TabPageSet(frame, pageNames=['Foobar','Baz'])
	tabPage.pack(expand=tk.TRUE, fill=tk.BOTH)
	tk.Label(tabPage['Foobar'], text='Foo', pady=20).pack()
	tk.Label(tabPage['Foobar'], text='Bar', pady=20).pack()
	tk.Label(tabPage['Baz'], text='Baz').pack()
	entryPgName=tk.Entry(frame)
	buttonAdd=tk.Button(frame, text='Add Page',
		command=lambda:tabPage.addPage(entryPgName.get()))
	buttonRemove=tk.Button(frame, text='Remove Page',
		command=lambda:tabPage.removePage(entryPgName.get()))
	labelPgName=tk.Label(frame, text='name of page to add/remove:')
	buttonAdd.pack(padx=5, pady=5)
	buttonRemove.pack(padx=5, pady=5)
	labelPgName.pack(padx=5)
	entryPgName.pack(padx=5)
	tabPage.changePage()
	b = tk.Button(root, text="Exit", command=root.destroy)
	Balloon.set(b, "Push me to exit")
	b.pack()
	e = FloatEntry(root)
	Balloon.set(e, "Enter a floating point number")
	e.pack()
	e = IntegerEntry(root)
	Balloon.set(e, "Enter an integer number")
	e.pack()
	root.geometry("800x600")
	root.mainloop()

#===============================================================================
if __name__ == "__main__":
	_test()
