# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   14-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import time
import tkinter as tk
from tkinter import messagebox

import icons
import tkExtra
import bFileDialog
import tkTextEditor

_DEFAULT_SPLIT = 150

#===============================================================================
# File Data Viewer
#===============================================================================
class FileData:
	def __init__(self, filename, maxsize=None):
		self.filename = filename
		self.fn, self.ext = os.path.splitext(filename)
		self.maxsize  = maxsize
		self.filedata = ""
		self.mtime    = 0
		self.ypos     = 0.0

		# Sections contains a dictionary with the range in the file
		# to display: self.sections[idname] = (from, to)
		self.sections = {}
		# Tree contains a nested list of lists to construct the
		# displayed tree
		# [ (title, idname, None|another-tree),... ]
		self.tree     = []

	# ----------------------------------------------------------------------
	# Returns True on failure
	# ----------------------------------------------------------------------
	def load(self):
		try:
			stat = os.stat(self.filename)
			size = stat.st_size
			self.mtime = stat.st_mtime
			if self.maxsize is not None:
				if size >= self.maxsize: return True
			with open(self.filename, "r",
					encoding="UTF-8",
					errors="replace") as f:
				self.filedata = f.read()
		except (OSError, IOError):
			return True
		except (OverflowError, MemoryError):
			return True
		return False

	# ----------------------------------------------------------------------
	def set(self, text):
		self.filedata = text
		self.mtime    = time.time()

	# ----------------------------------------------------------------------
	# Override routine for creating the sections table
	# Sections dictionary should contain tuples with the range start:end
	# ----------------------------------------------------------------------
	def makeTOC(self):
		pass

	# ----------------------------------------------------------------------
	# Return icon to display
	# ----------------------------------------------------------------------
	def icon(self, name):
		return None

	# ----------------------------------------------------------------------
	# Override routine to highlight messages
	# @param widget the text widget to highlight
	# ----------------------------------------------------------------------
	def highlight(self, widget):
		pass

#===============================================================================
# File Viewer frames
#===============================================================================
class FileViewer(tk.Frame):
	split  = _DEFAULT_SPLIT

	def __init__(self, master, **kw):
		"""Open a viewer assuming that the file exists."""
		super().__init__(master, **kw)

		self._inFocus = False
		self.splitter  = tkExtra.HSplitter(self, FileViewer.split, True)
		self.splitter.pack(expand=tk.YES, fill=tk.BOTH)

		treeFrame = self.splitter.leftFrame()

		# Tree
		self.tree = tkExtra.Treeview(treeFrame,
					show="tree",
					selectmode="browse")
		self.tree.grid(row=0, column=0, sticky=tk.NSEW)
		self.tree.bind("<<TreeviewSelect>>", self.treeSelect)
		self.tree.setPopupMenu([
			("Reload",     0, self.reloadSelected,icons.get("load")),
			("Close",      0, self.close,         icons.get("close"))
			])

		treeFrame.grid_rowconfigure(0, weight=1)
		treeFrame.grid_columnconfigure(0, weight=1)

		# Scroll bars
		sb=tkExtra.AutoScrollbar(treeFrame, takefocus=0)
		sb.grid(row=0, column=1, sticky=tk.NS)
		self.tree.configure(yscrollcommand=sb.set)
		sb.configure(command=self.tree.yview)

		sb=tkExtra.AutoScrollbar(treeFrame,
				orient=tk.HORIZONTAL, takefocus=0)
		sb.grid(row=1, column=0, sticky=tk.EW)
		self.tree.configure(xscrollcommand=sb.set)
		sb.configure(command=self.tree.xview)

		self.tree.focus_set()

		self.editor = tkTextEditor.TextEditor(self.splitter.rightFrame())
		self.editor.txt.config(state=tk.DISABLED)
		self.editor.txt.focus_set()
		self.editor.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

		self.editor.txt.bind('<Up>',   lambda e, s=self.editor.txt:
					s.yview(tk.SCROLL, -1, tk.UNITS))
		self.editor.txt.bind('<Down>', lambda e, s=self.editor.txt:
					s.yview(tk.SCROLL,  1, tk.UNITS))
		self.editor.txt.bind('<Button-1>', lambda e, s=self.editor.txt:
					s.focus_set())
		self.editor.txt.bind("<F11>",	self.toggleSplitter)
		self.tree.bind("<F11>",		self.toggleSplitter)
		self.bind("<F11>",	self.toggleSplitter)
		self.bind('<FocusIn>', self._focusIn)

		self.files = []
		self.cwd   = ""		# current working directory
		self.curid = None

	# ----------------------------------------------------------------------
	def setCwd(self, cwd):
		self.cwd = cwd

	# ----------------------------------------------------------------------
	def _append(self, curfile):
		self.files.append(curfile)
		self.treeFill()
		item = f"file:{len(self.files)-1}"
		self.tree.selection_set(item)
		self.tree.focus(item)
		self.tree.see(item)

	# ----------------------------------------------------------------------
	# Routine to be overwritten
	# @return the FileData structure
	# ----------------------------------------------------------------------
	def loadData(self, filename, maxsize):
		curfile = FileData(filename, maxsize)
		if curfile.load():
			return None
		curfile.makeTOC()
		return curfile

	# ----------------------------------------------------------------------
	# Load file, returns True on failure
	# ----------------------------------------------------------------------
	def load(self, filename, maxsize=None):
		curfile = self.loadData(filename, maxsize)
		if curfile is None: return True
		self._append(curfile)
		return False

	# ----------------------------------------------------------------------
	# Show a text(file) from memory
	# ----------------------------------------------------------------------
	def show(self, name, text):
		curfile = FileData(name)
		curfile.set(text)
		self._append(curfile)

	# ----------------------------------------------------------------------
	def hasFile(self, filename):
		return filename in [x.filename for x in self.files if x]

	# ----------------------------------------------------------------------
	# Return selected file
	# ----------------------------------------------------------------------
	def selected(self):
		if not self.files: return None
		try:
			_,idx,*_ = self.tree.focus().split(":")
		except:
			return None
		return int(idx)

	# ----------------------------------------------------------------------
	def reload(self, filename):
		for i,f in enumerate(self.files):
			if f is not None and f.filename == filename:
				item = f"file:{i}"
				break
		else:
			self.tree.selection_set(item)
			self.tree.focus(item)
			self.reloadSelected()

	# ----------------------------------------------------------------------
	def reloadSelected(self, event=None):
		idx = self.selected()
		if idx is None: return
		curfile = self.files[idx]
		curfile.load()
		self.showSection(self.tree.focus())

	# ----------------------------------------------------------------------
	def openFile(self, event=None):
		fn = bFileDialog.askopenfilename(master=self)
		if fn:
			self.load(fn)

	# ----------------------------------------------------------------------
	def close(self, event=None):
		idx = self.selected()
		if idx is None: return
		del self.files[idx]
		if idx>=len(self.files):
			idx = len(self.files)-1
		self.curid = None
		self.treeFill()
		if idx>=0: self.showSection(idx)

	# ----------------------------------------------------------------------
	def _focusIn(self, event=None):
		if self._inFocus: return
		# FocusIn is generated for all sub-windows, handle only the main window
		if self is not event.widget: return
		self._inFocus = True

		# Check time stamp for all files
		changed = []
		fn      = []
		for i,f in enumerate(self.files):
			if f is None: continue
			try:
				mtime = os.path.getmtime(f.filename)
				if mtime > f.mtime:
					changed.append(i)
					fn.append(f.filename)
			except OSError:
				pass

		if not changed:
			pass
		elif messagebox.askyesno("File has changed",
				f"Files:\n{', '.join(fn)}\nhave changed. Reload?",
				parent=self):
			for i in changed:
				self.tree.focus(f"file:{i}")
				self.reloadSelected()
		else:
			for i in changed:
				f = self.files[i]
				try:
					f.mtime = os.path.getmtime(f.filename)
				except OSError:
					pass

		self._inFocus = False

	# ----------------------------------------------------------------------
	def toggleTree(self, event=None):
		if self.splitter.split > self.splitter.min:
			self.splitter.toggle()
			self.splitter.rightFrame().focus_set()
			self.focus_set()
		else:
			self.splitter.toggle()
			if self.splitter.split <= self.splitter.min:
				self.splitter.split = _DEFAULT_SPLIT
				self.splitter.place()

	# ----------------------------------------------------------------------
	# Callback routine to fill tree contents
	# ----------------------------------------------------------------------
	def treeFill(self):
		#---------------------------------------------------------------
		# scan for open items
		#---------------------------------------------------------------
		def _scanOpen(parent=""):
			nonlocal open_
			if self.tree.item(parent,"open"):
				open_.add(parent)
			for item in self.tree.get_children(parent):
				_scanOpen(item)

		#---------------------------------------------------------------
		# Fill a branch
		#---------------------------------------------------------------
		def _fillBranch(curfile, parent, treelist):
			nonlocal open_
			for name,idx,sublist in treelist:
				if curfile.sections.get(idx) is None: continue
				icon = f.icon(idx)
				if icon is None: icon=icons.get(name,idx,"section")
				item = self.tree.insert(parent, tk.END,
								id=f"{parent}:{idx}",
								text=name,
								image=icon,
								open=idx in open_)
				if sublist:
					_fillBranch(curfile, item, sublist)
		#---------------------------------------------------------------

		open_ = set()
		_scanOpen()
		selection = self.tree.selection()
		self.tree.delete(*self.tree.get_children())
		for idx,f in enumerate(self.files):
			if f is None: continue
			parent = self.tree.insert("", tk.END,
						id=f"file:{idx}",
						text=os.path.relpath(f.filename, self.cwd),
						image=icons.get("file"),
						tag="file",
						open=True)
			_fillBranch(f, parent, f.tree)
		self.tree.tag_configure("file", foreground="DarkRed")
		#self.tree.selection_set(selection)

	# ----------------------------------------------------------------------
	def treeSelect(self, event):
		self.showSection(self.tree.focus())

	# ----------------------------------------------------------------------
	def showSection(self, item):
		if isinstance(item,int):
			idx  = item
			node = None
			item = f"file:{idx}"
			self.tree.selection_set(item)
			self.tree.focus(item)
			self.tree.see(item)
		else:
			try:
				_,idx,*node = item.split(":")
			except ValueError:
				return
			idx = int(idx)
		if idx>=len(self.files): return
		self.editor.txt.config(state=tk.NORMAL)

		# Remember position
		if isinstance(self.curid, int):
			self.files[self.curid].ypos = self.editor.txt.yview()[0]

		# Delete everything
		self.editor.txt.delete(1.0, tk.END)

		# find section infos
		self.curid = idx
		curfile = self.files[self.curid]
		if curfile is None: return
		if isinstance(self.master,tk.Toplevel):
			self.master.title(f"FileViewer: {curfile.filename}")
		if not node:
			try:
				self.editor.txt.insert(1.0, curfile.filedata)
			except OverflowError:
				self.editor.txt.insert(1.0, "File too big to load...")
			except Exception as err:
				self.editor.txt.insert(1,0, str(err))
			self.editor.txt.yview_moveto(curfile.ypos)
			curfile.highlight(self.editor.txt)
		else:
			section = curfile.sections.get(node[-1])
			if section is not None:
				self.editor.txt.insert(tk.INSERT,
					curfile.filedata[section[0]:section[1]])
				curfile.highlight(self.editor.txt)
			self.curid = None	# forget it

		self.editor.txt.config(state=tk.DISABLED)

	# ----------------------------------------------------------------------
	def clear(self):
		self.editor.txt.config(state=tk.NORMAL)
		self.editor.txt.delete(1.0, tk.END)
		self.editor.txt.config(state=tk.DISABLED)

	# ----------------------------------------------------------------------
	def toggleSplitter(self, event=None):
		self.splitter.toggle()
