#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   06-Jan-2021

import math
import tkinter as tk

import bmath
from palette import Palette

# winfo_rgb is slow! caching the colors
_TK2INT = {}

#-------------------------------------------------------------------------------
# Convert tk string color to integer color
#-------------------------------------------------------------------------------
def tk2int(color):
	if not isinstance(color,str): return color

	if color and color[0] == '#':
		return int(color[1:],16)

	try:
		return _TK2INT[color]
	except KeyError:
		try:
			r,g,b = tk._default_root.winfo_rgb(color)
			rgb = (r>>8)<<16 | (g>>8)<<8 | (b>>8)
		except tk.TclError:
			rgb = 0
		_TK2INT[color] = rgb
		return rgb

#===============================================================================
# PaletteCanvas
#===============================================================================
class PaletteCanvas(tk.Canvas):
	def __init__(self, master, *kw, **kwargs):
		super().__init__(master, *kw, **kwargs)
		self.min     = 0		# minimum value
		self.max     = 1		# maximum value
		self.log     = False		# log or linear plot
		self.stops   = None		# absolute value of stops if any
		self.palette = None

	#-----------------------------------------------------------------------
	def setPalette(self, palette, ncolors=None):
		if isinstance(palette,Palette):
			self.palette = palette
		else:
			self.palette = Palette(palette, ncolors)
		self.tkcolors = self.palette.tk()

	#-----------------------------------------------------------------------
	def setLimits(self, mi, ma, log=False):
		self.log  = log
		if log:
			self.min  = math.log(mi)
			self.max  = math.log(ma)
		else:
			self.min  = mi
			self.max  = ma

	#-----------------------------------------------------------------------
	# Predefined stops as fraction of total range or absolute
	#-----------------------------------------------------------------------
	#def stops(self, s=None, absolute=True)
	#	self.stops = s

	#-----------------------------------------------------------------------
	def draw(self):
		try:
			self.delete(tk.ALL)
		except tk.TclError:
			return

		width  = self.winfo_width()
		height = self.winfo_height()
		margin = 10
		dheight = height - 2*margin
		drange  = self.max - self.min
		ymin    = height - margin

		#r     = 1.0
		#steps = 0
		#while True:
		#	steps = round((paletteLev[1]-paletteLev[0])/r)
		#	if steps < 10:
		#		r /= 10.0
		#	elif steps >= 95:
		#		r *= 10.0
		#	elif steps >= 20:
		#		r *= 5.0
		#	else:
		#		break

		# --------------------------------------------------------------
		# Draw grid
		# For linear space
		# Find powers to show ticks
		d10  = 10**(round(math.log10(drange))-1)
		if drange//d10 > 20: d10 *= 10
		dmin = math.floor(abs(self.min) / d10) * d10
		if self.min<0: dmin = -dmin
		if dmin < self.min-1e-10: dmin += d10
		ds = dheight / drange
		d102 = d10/2

		# Draw labels
		d = dmin
		while d <= self.max:
			y = round((self.min-d)*ds + ymin)
			if abs(d)<1e-10: d = 0
			self.create_text(width-2, y,
						#text="%.3g"%(d),
						text=bmath.format(d,5,useDot=False),
						tag="label",
						justify=tk.RIGHT,
						anchor=tk.E)
			d += d10
		x1,_,_,_ = self.bbox("label")

		# Draw boxes
		boxes = []
		if self.stops:
			pass
		else:
			w  = min(20,x1-8)
			dy = dheight / len(self.palette)
			#ds = drange / (len(self.palette)-1)
			y  = ymin
			#s  = self.min
			for ic in range(len(self.palette)):
				yn = y-dy
				boxes.append(self.create_rectangle((0,round(yn),w,round(y)),
							width=0,
							fill=self.tkcolors[ic]))
				y  = yn
				#s += ds

		# Draw ticks
		d = dmin
		while d <= self.max:
			y = round((self.min-d)*ds + ymin)
			self.create_line(0, y, x1-2, y)
			y = round((self.min-(d+d102))*ds + ymin)
			self.create_line(w, y, w+(x1-w)//2, y)
			d += d10
