#!/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   03-Apr-2014

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os

import tkinter as tk
#import tkinter.font as tkFont
#from tkinter.colorchooser import askcolor

try:
	import PIL.Image
	import PIL.ImageTk
	have_PIL = True
except:
	have_PIL = False

import markdown

_FONT_SIZE = (   '6',  '7',  '8',  '9', '10', '11', '12', '13',
		'14', '16', '18', '20', '22', '24', '26', '28',
		'32', '36', '40', '48', '56', '64', '72' )

_FONT_JUSTIFY = (tk.LEFT, tk.CENTER, tk.RIGHT)

# User readable styles
ABBREV = {
	"Heading 1"  : "h1",
	"Heading 2"  : "h2",
	"Heading 3"  : "h3",
	"Bold"       : "b",
	"Italic"     : "i",
	"BoldItalic" : "bi",
	"Underline"  : "u",
	"Overstrike" : "o",
	"Normal"     : "n",
	"Link"       : "link",
	"Caption"    : "caption",

	"Left"       : "l",
	"Right"      : "r",
	"Center"     : "c",
	}
# and inverted ones
INV_ABBREV = {y:x for (x,y) in ABBREV.items()}

#===============================================================================
# Markdown formatter
# class to parse/load/save in markdown format
# insert/get to a tkinter Text() widget a markdown text
#===============================================================================
class MarkdownFormatter:
	TAGS = {
		"md:n"    : {"font"       : ("Sans",10,""),
		             "justify"    : tk.LEFT,
		             "foreground" : "#000000"},
		"md:h1"   : {"font"       : (None,18,("bold","italic")),
		             "foreground" : "DarkBlue",
		             "justify"    : tk.LEFT},
		"md:h2"   : {"font"       : (None,16,("bold","italic")),
		             "foreground" : "DarkBlue",
		             "justify"   : tk.LEFT},
		"md:h3"   : {"font"       : (None,14,("bold","italic")),
		             "foreground" : "DarkBlue",
		             "justify"    : tk.LEFT},
		"md:h4"   : {"font"       : (None,13,("bold","italic")),
		             "foreground" : "DarkBlue",
		             "justify"    : tk.LEFT},
		"md:h5"   : {"font"       : (None,12,("bold","italic")),
		             "foreground" : "DarkBlue",
		             "justify"    : tk.LEFT},
		"md:h6"   : {"font"       : (None,11,("bold","italic")),
		             "foreground" : "DarkBlue",
		             "justify"    : tk.LEFT},
		"md:b"    : {"font"       : (None,None,"bold")},
		"md:i"    : {"font"       : (None,None,"italic")},
		"md:u"    : {"font"       : (None,None,"underline")},
		#"md:b:i" : {"font"       : (None,None,"bold italic")},
		#"md:b:u" : {"font"       : (None,None,"bold underline")},
		"md:o"    : {"font"       : (None,None,"overstrike")},
		"md:l"    : {"justify"    : tk.LEFT},
		"md:r"    : {"justify"    : tk.RIGHT},
		"md:c"    : {"justify"    : tk.CENTER},
		"md:img"  : {"foreground" : "Red"},
		"md:link" : {"font"       : (None,None,"underline"),
		             "foreground" : "Blue"},
		"md:caption":{"foreground" : "DarkBlue"},
		"md:url"  : {"elide"      : True},
		"md:code" : {"font"      : ("Fixed",10,"")},
		}
	TAG2MD = {
		"md:n"    : (markdown.Text,       None),
		"md:b"    : (markdown.Bold,       None),
		"md:i"    : (markdown.Emphasis,   None),
		"md:u"    : (markdown.Underline,  None),
		"md:o"    : (markdown.Overstrike, None),
		"md:h1"   : (markdown.Header,     1),
		"md:h2"   : (markdown.Header,     2),
		"md:h3"   : (markdown.Header,     3),
		"md:h4"   : (markdown.Header,     4),
		"md:h5"   : (markdown.Header,     5),
		"md:h6"   : (markdown.Header,     6),
		"md:img"  : (markdown.Image,      None),
		"md:link" : (markdown.Link,      None),
		"md:code" : (markdown.Code,      None),
		#"md:c"   : (markdown.Center,    None),
		#"md:l"   : (markdown.Left,      None),
		#"md:r"   : (markdown.Right,     None),
		}
	FORMAT = {
		markdown.Bold      : "md:b",
		markdown.Underline : "md:u",
		markdown.Emphasis  : "md:i",
		markdown.Overstrike: "md:o",
		markdown.Link      : "md:link",
		markdown.Image     : "md:img",
		}
	LINETAGS = ["md:h1", "md:h2", "md:h3", "md:h4", "md:h5", "md:h6", "md:l", "md:c", "md:r"]

	#-----------------------------------------------------------------------
	def __init__(self, text):
		self.text = text
		self.configureTags()
		self.reset()

	# --------------------------------------------------------------------
	def reset(self):
		self.images = {}
		self._img   = {}

	#-----------------------------------------------------------------------
	# Insert to text widget at position the data
	#-----------------------------------------------------------------------
	def insert(self, position, md):
		self.text.mark_set(tk.INSERT, position)
		self._insert(md.data)

	#-----------------------------------------------------------------------
	def _insert(self, data, fmt="md:n"):
		try:
			f2 = MarkdownFormatter.FORMAT[data.__class__]
			fmt = f2
		except KeyError:
			if isinstance(data, markdown.Header):
				fmt = f"md:h{data.opt}"

		if isinstance(data, markdown.Image):
			image = self.image(data.opt)
			if image:
				img = self.text.image_create(tk.INSERT, image=image)
				self.text.tag_add("md:img",self.text.index(img))
				self._img[img] = data.items
				self.text.insert(tk.INSERT, data.opt, "md:url")
			else:
				self.text.insert(tk.INSERT, f"![data.items]({data.opt})", "md:img")

		elif isinstance(data, markdown.Link):
			self.text.insert(tk.INSERT, data.items, "md:link")
			self.text.insert(tk.INSERT, data.opt, "md:url")

		elif isinstance(data, markdown.Header):
			self._insert(data.items, f"md:h{data.opt}")

		elif isinstance(data, str):
			self.text.insert(tk.INSERT, data, fmt)

		elif isinstance(data, list):
			for item in data:
				self._insert(item, fmt)

		elif isinstance(data.items, list):
			for item in data.items:
				self._insert(item, fmt)

		elif isinstance(data.items, str):
			self.text.insert(tk.INSERT, data.items, fmt)

		else:
			#print(f"{data=}")
			assert False

		if isinstance(data, (markdown.Block, markdown.Header, markdown.Paragraph)):
			self.text.insert(tk.INSERT, "\n")

	# --------------------------------------------------------------------
	# Find image
	# --------------------------------------------------------------------
	def image(self, filename):
		if not have_PIL: return None
		try:
			s = os.stat(filename)
		except:
			return None

		image = self.images.get(filename)
		if image is not None:
			image, mt = image
			if s.st_mtime > mt: image = None

		if image is None:
			try:
				image = PIL.ImageTk.PhotoImage(PIL.Image.open(filename))
				self.images[filename] = image, s.st_mtime
			except:
				image = None
		return image

	# --------------------------------------------------------------------
	# Get rich text from widget
	# --------------------------------------------------------------------
	def get(self, from_, to_):
		items = []
		tag = []

		for key,value,index in self.dump(from_, to_):
			if key=="tagon":
				if value.startswith("md:"):
					try:
						klass,opt = MarkdownFormatter.TAG2MD[value]
						tag.append(klass(value))
					except KeyError:
						print("Unknown tag",tag)
			elif key=="text":
				if tag:
					klass,opt = tag[-1]
				items.append(klass(value,opt))
			elif key=="tagoff":
				tag.pop()
			#elif key=="image":
			#	if tag is not None: data += "}"
			#	name = self._img.get(value)
			#	data += f"{{img:{name}}}"
			#	if tag is not None: data += f"{{{tag}: "
			elif key=="mark":
				pass
			elif key=="window":
				pass
		#return data

	# --------------------------------------------------------------------
	# Return style with inheritance
	# --------------------------------------------------------------------
	@staticmethod
	def getStyle(tag):
		style = MarkdownFormatter.TAGS.get(tag)
		if style is None: return None

		based = style.get("based")
		if based is None:
			based = MarkdownFormatter.TAGS.get("md:n")
		else:
			based = MarkdownFormatter.getStyle(based)
		if based is None: return None

		newstyle = based.copy()
		try: del newstyle["background"]
		except: pass
		for tag,value in style.items():
			if tag == "based": continue
			if value is None: continue
			if isinstance(value,tuple):
				newvalue = []
				for i,v in enumerate(value):
					if v is None:
						v = based[tag][i]
					newvalue.append(v)
				newstyle[tag] = tuple(newvalue)
			else:
				# Replace with the new value
				newstyle[tag] = value

		return newstyle

	# --------------------------------------------------------------------
	def configureTags(self):
		# Configure tags
		for tag in MarkdownFormatter.TAGS:
			style = MarkdownFormatter.getStyle(tag)
			#print("Markdown: ",tag,style)
			if style: self.text.tag_config(tag, style)

		# Basic colors of text window
		style = MarkdownFormatter.TAGS["md:n"]
		for n in ("font", "foreground"):
			self.text[n] = style[n]

	# --------------------------------------------------------------------
	def setStyle(self, style):
		insert = self.text.index(tk.INSERT)
		style = ABBREV.get(style,style)
		try:
			sel_first = self.text.index(tk.SEL_FIRST)
			sel_last  = self.text.index(tk.SEL_LAST)

			if style in MarkdownFormatter.LINETAGS:
				sel_first = f"{sel_first} linestart"
				sel_last  = f"{sel_last} lineend"
		except tk.TclError:
			cur = self.text.index(tk.INSERT)
			if style in MarkdownFormatter.LINETAGS:
				sel_first = f"{cur} linestart"
				sel_last  = f"{cur} lineend"
			else:
				sel_first = f"{cur} - 1 char"
				sel_last  = cur

		names = self.text.tag_names(sel_first)

		for n in names:
			if n in MarkdownFormatter.LINETAGS:
				first = f"{sel_first} linestart"
				last  = f"{sel_last} lineend"
				break
		else:
			first = sel_first
			last  = sel_last

		for tag in MarkdownFormatter.TAGS:
			self.text.tag_remove(tag, first, last)

		if style!="md:n" and style not in names:
			self.text.tag_add(style, sel_first, sel_last)

		# Why? + 1 chars (but it works)
		self.text.mark_set(tk.INSERT,f"{insert} + 1 chars")

# -----------------------------------------------------------------------------
if __name__ == "__main__":
	import sys
	def close(event=None):
		root.destroy()

	root = tk.Tk()
	root.protocol("WM_DELETE_WINDOW", root.destroy)
	root.bind("<Escape>", close)
	root.bind("<Control-Key-q>", close)
	root.bind("<Control-Key-w>", close)

	toolbar = tk.Frame(root)
	toolbar.pack(side=tk.TOP, fill=tk.X)

	text = tk.Text(root)
	text.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
	sb = tk.Scrollbar(root, command=text.yview)
	sb.pack(side=tk.RIGHT, fill=tk.Y)
	text.config(yscrollcommand=sb.set)

	text.bind("<Key-q>", lambda e: root.destroy())
	text.bind("<Key-k>", lambda e,t=text: t.yview(tk.SCROLL, -1, tk.UNITS))
	text.bind("<Key-j>", lambda e,t=text: t.yview(tk.SCROLL, +1, tk.UNITS))
	text.bind("<Control-Key-u>", lambda e,t=text: t.yview(tk.SCROLL, -1, tk.PAGES))
	text.bind("<Control-Key-d>", lambda e,t=text: t.yview(tk.SCROLL, +1, tk.PAGES))
	text.bind("<g>", lambda e,t=text: t.yview(tk.MOVETO, 0.0))
	text.bind("<G>", lambda e,t=text: t.yview(tk.MOVETO, 1.0))

	mdf = MarkdownFormatter(text)
	for name in MarkdownFormatter.TAGS:
		cmd = lambda n=name: mdf.setStyle(n)
		b = tk.Button(toolbar, text=name[3:], command=cmd)
		b.pack(side=tk.LEFT)

	fn,ext = os.path.splitext(sys.argv[1])
	with open(sys.argv[1],"r") as f:
		document = f.read()
	if ext == ".md":
		md = markdown.Markdown(document)
	elif ext == ".rtf":
		md = markdown.RTF(document)
	else:
		md = markdown.RST(document)

	#md = markdown.Markdown("""# Header1 **bold**
#Line **bold** [Link *text*](http://cern.ch) *emphasis* __underline__ ~~strikethrough~~
#
#
## Heading 1
#
#Heading 2
#---------
#
#### Heading 3
#
##### Heading 4
#
###### Heading 5
####### Heading 6
######## Heading 7
#
#Flair log
#![Flair logo](../images/flair-logo.gif)
#New paragraph
#""")
	mdf.insert(tk.END, md)
	text.focus_set()
	text.mark_set(tk.INSERT, "1.0")
	text.config(state=tk.DISABLED)
	root.mainloop()
