#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   03-Mar-2015

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import time

#===============================================================================
# Cached listdir
#===============================================================================
_cache = {}

#-------------------------------------------------------------------------------
def listdir_reset():
	global _cache
	_cache = {}

#-------------------------------------------------------------------------------
def listdir(path):
	"""List directory contents, using cache."""
	try:
		cached_mtime, lst = _cache[path]
		del _cache[path]
	except KeyError:
		cached_mtime, lst = -1, []

	try:
		mtime = os.path.getmtime(path)
	except IOError:
		return []
	if mtime != cached_mtime:
		try:
			lst = os.listdir(path)
		except IOError:
			return []
		#lst.sort()

	_cache[path] = mtime, lst
	return lst

#-------------------------------------------------------------------------------
# substitute of os.walk up to a certain level
#-------------------------------------------------------------------------------
def walk(path, level=1):
	path = path.rstrip(os.path.sep)
	assert os.path.isdir(path)
	num_sep = path.count(os.path.sep)
	for root, dirs, files in os.walk(path):
		yield root, dirs, files
		num_sep_this = root.count(os.path.sep)
		if num_sep + level <= num_sep_this:
			del dirs[:]

#-------------------------------------------------------------------------------
def abbrevPath(path, maxlen=40):
	"""Abbreviate a large path to fit in small entry"""
	if len(path)<maxlen: return path
	prefix = ""
	if path.find(os.sep)>=0:
		d = os.path.dirname(path).split(os.sep)
		# keep only the first and last
		del d[1:-2]
		d[1] = "..."
		prefix = os.sep.join(d) + os.sep
		if len(prefix) > maxlen:
			del d[-1]
			prefix = os.sep.join(d) + os.sep

	filename = os.path.basename(path)
	maxlen -= len(prefix)
	if len(filename) >= maxlen:
		# remove middle characters
		rm = max((maxlen - 3)//2, 1)
		filename = f"{filename[:rm]}...{filename[-rm:]}"

	return f"{prefix}{filename}"

#-------------------------------------------------------------------------------
# Make path relative or smallest representation
# os.path.relpath() fails to return absolute paths, it makes everything relative
#-------------------------------------------------------------------------------
def relativePath(path, directory):
	"""Convert path to relative to the directory"""
	if not path: return path
	dirpath = os.path.dirname(os.path.normpath(path))

	# if already relative => return
	if not dirpath or (dirpath[0] not in (os.sep,".")):
		return path

	if path==directory: return "."

	# Find common path
	d = directory.split(os.sep)
	p = path.split(os.sep)
	cp = ""
	for i in range(min(len(d),len(p))):
		if d[i]:
			if d[i] == p[i]:
				cp += os.sep + d[i]
			else:
				break
	if cp: cp += os.sep
	dirsuffix = directory[len(cp):]
	if dirsuffix:
		relpath = (os.pardir+os.sep) * len(dirsuffix.split(os.sep))
	else:
		relpath = ""
	relpath += path[len(cp):]
	if len(relpath)<len(path):
		return relpath if relpath else "."
	return path

#-------------------------------------------------------------------------------
def friendlyTime(t):
	"""Display time in a human friendly way"""
	if t<1.0:
		return f"{t*1000.0:.3g} ms"
	elif t<60.0:
		return f"{t:.3g} s"

	(days,  tm) = divmod(t, 86400)
	(hours, tm) = divmod(tm, 3600)
	(minu, sec) = divmod(tm,   60)

	if days>0:
		s = f"{days}d"
	else:
		s = ""
	if hours>0 or s!="":
		if s!="": s += " "
		s += f"{int(hours)}h"
	if minu>0 or s!="":
		if s!="": s += " "
		s += f"{int(minu)}m"
	if s!="": s += " "

	s += f"{int(sec)}s"
	return s

#-------------------------------------------------------------------------------
# Check if file is newer than another
#-------------------------------------------------------------------------------
def isnewer(newfile, oldfile):
	"""Check if the first file is newer than the second"""
	try:    n = os.path.getmtime(newfile)
	except: return False
	try:    o = os.path.getmtime(oldfile)
	except: return True
	return n > o

#-------------------------------------------------------------------------------
def getmtime(filename):
	try:    return os.path.getmtime(filename)
	except: return 0

#-------------------------------------------------------------------------------
# Create backup of file if needed
#-------------------------------------------------------------------------------
def backup(filename, suffix="~"):
	try:
		s = os.stat(filename)
	except FileNotFoundError:
		return False
	if (s.st_mode & (1<<7)) == 0: return False
	backupname = filename+suffix
	try: os.remove(backupname)
	except OSError: pass
	try:
		os.rename(filename, backupname)
	except OSError:
		return False
	return True

#-------------------------------------------------------------------------------
# @return true if filename is executable
#-------------------------------------------------------------------------------
def isexecutable(filename):
	return os.path.isfile(filename) and os.access(filename, os.X_OK)

#-------------------------------------------------------------------------------
# @return path from environment PATH that contains file and it is executable
#-------------------------------------------------------------------------------
def findExe(exefile):
	for path in os.getenv("PATH").split(":"):
		filename = os.path.join(path, exefile)
		if os.path.isfile(filename) and os.access(filename, os.X_OK):
			return path
	return None

#-------------------------------------------------------------------------------
# timeit decorator
#-------------------------------------------------------------------------------
def timeit(func):
	def wrapper(*args, **kwargs):
		t = time.perf_counter()
		func(*args, **kwargs)
		print("timeit",func.__class__.__name__,func.__name__,
				"=",time.perf_counter()-t)
	return wrapper

#===============================================================================
# Simple timer keeping class
#===============================================================================
class Timer:
	def __init__(self):
		self._start = time.perf_counter()
		self._stop  = self._start

	#-----------------------------------------------------------------------
	def lap(self, msg=None):
		if msg is not None:
			print(msg, time.perf_counter()-self._stop, "s")
		self._stop = time.perf_counter()

	#-----------------------------------------------------------------------
	def stop(self, msg=None):
		self._stop = time.perf_counter()
		if msg is not None:
			print(msg, self(), "s")

	#-----------------------------------------------------------------------
	def __call__(self):
		return self._stop - self._start
