# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   15-Oct-2017

#===============================================================================
# Undo Redo Class
#===============================================================================
class UndoRedo:
	#-----------------------------------------------------------------------
	def __init__(self, maxlimit=None):
		self.undoList = []
		self.redoList = []
		self.maxLimit = maxlimit

	#-----------------------------------------------------------------------
	def reset(self):
		del self.undoList[:]
		del self.redoList[:]

	#-----------------------------------------------------------------------
	# Add undoinfo as (msg, func/list, args)
	#-----------------------------------------------------------------------
	def add(self, undoinfo, msg=None):
		if not undoinfo: return
		if msg is not None:
			if isinstance(undoinfo[0], str):
				# replace message
				undoinfo = (msg,) + undoinfo[1:]
			elif isinstance(undoinfo, tuple):
				undoinfo = (msg,) + undoinfo
			else:
				undoinfo = (msg, undoinfo)
			f = 1
		else:
			f = int(isinstance(undoinfo[0], str))
		assert isinstance(undoinfo,list) or \
		       callable(undoinfo[f]) or \
		       isinstance(undoinfo[f],list)
		self.undoList.append(undoinfo)
		if self.maxLimit is not None:
			while len(self.undoList)>=self.maxLimit:
				del self.undoList[0]
		del self.redoList[:]

	#-----------------------------------------------------------------------
	# Split the undoinfo into [msg, ]func/list [, args]
	# msg can exists or not check for str
	# func can be a list or an executable
	# if func is a list then there are no args
	# @return always a tuple of 3 with msg, func, args
	#-----------------------------------------------------------------------
	@staticmethod
	def _split(undoinfo):
		if isinstance(undoinfo, list):
			return None, undoinfo, None
		elif undoinfo[0] is None or isinstance(undoinfo[0], str):
			assert callable(undoinfo[1]) or isinstance(undoinfo[1],list)
			return undoinfo[0], undoinfo[1], undoinfo[2:]
		else:
			assert callable(undoinfo[0]) or isinstance(undoinfo[0],list)
			return None, undoinfo[0], undoinfo[1:]

	#-----------------------------------------------------------------------
	# Execute the undoinfo and return the redoinfo
	#-----------------------------------------------------------------------
	def _execute(self, undoinfo):
		if undoinfo is None: return None
		msg, func, args = UndoRedo._split(undoinfo)
		if isinstance(func, list):
			redolist = []
			while func:
				redolist.append(self._execute(func.pop()))
			if msg:
				return msg, redolist
			else:
				return redolist
		else:
			#print('<<< Undo:', type(undoinfo))
			redoinfo = func(*args)
			#print(">>> Redo:", redoinfo)
			if redoinfo is None or \
			   isinstance(redoinfo[0], str):
				return redoinfo
			elif msg:
				if isinstance(redoinfo, tuple):
					return (msg,) + redoinfo
				else:
					return (msg, redoinfo)
			else:
				return redoinfo

	#-----------------------------------------------------------------------
	def undo(self):
		if not self.undoList: return
		self.redoList.append(self._execute(self.undoList.pop()))

	#-----------------------------------------------------------------------
	def redo(self):
		if not self.redoList: return
		self.undoList.append(self._execute(self.redoList.pop()))

	#-----------------------------------------------------------------------
	def canUndo(self):
		return bool(self.undoList)

	#-----------------------------------------------------------------------
	def canRedo(self):
		return bool(self.redoList)

	#-----------------------------------------------------------------------
	def undoText(self):
		u = self.undoList[-1]
		if isinstance(u[0], str):
			return u[0]
		else:
			return "undo"

	#-----------------------------------------------------------------------
	def undoTextList(self):
		lst = []
		for u in self.undoList:
			if isinstance(u[0], str) and u[0]:
				lst.append(u[0])
			else:
				lst.append("undo")
		lst.reverse()
		return lst
