# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   09-Jun-2020

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import time
import configparser
from itertools import count

from element import Element
from fluka import Fluka, FlukaCard
from phits import Phits, PhitsCard

#===============================================================================
# Material class
#===============================================================================
class Material:
	groups    = []
	materials = {}
	names     = {}
	changed   = False

	_COMPOSITION = ["atom", "mass", "volume"]
	_STATE       = ["solid", "liquid", "gas"]

	# ----------------------------------------------------------------------
	def __init__(self, title):
		self.title         = title
		self.name          = ""
		self.group         = []
		self.density       = 0.0
		self.composition   = 0
		self.notes         = ""
		self.elements      = []
		self.stoichiometry = ""
		self.state         = 0
		self.pressure      = ""   # Pressure
		self.rhor          = ""   # Density factor
		self.ionization    = ""   # Average Ionization potential
		self.dpa_threshold = ""   # DPA Damage threshold
		self.barcdpa       = ""   # b arc-dpa Nordlund fit parameter
		self.carcdpa       = ""   # c arc-dpa Nordlund fit parameter
		self.s_cbar        = ""   # Sternheimer parameters
		self.s_x0          = ""
		self.s_x1          = ""
		self.s_a           = ""
		self.s_m           = ""
		self.s_d0          = ""
		self._modified     = ""

	# ----------------------------------------------------------------------
	def make(self):
		# Make stoichiometry
		self.makeStoichiometry()

		# Calculate chksum
		self.chksum  = 0.0
		try: self.chksum  = self.density
		except: pass

		try: self.chksum += float(self.pressure)
		except: pass
		try: self.chksum += float(self.rhor)
		except: pass
		try: self.chksum += float(self.ionization)
		except: pass
		try: self.chksum += float(self.dpa_threshold)
		except: pass
		try: self.chksum += float(self.barcdpa)
		except: pass
		try: self.chksum += float(self.carcdpa)
		except: pass

		try: self.chksum += float(self.s_cbar)
		except: pass
		try: self.chksum += float(self.s_x0)
		except: pass
		try: self.chksum += float(self.s_x1)
		except: pass
		try: self.chksum += float(self.s_a)
		except: pass
		try: self.chksum += float(self.s_m)
		except: pass
		try: self.chksum += float(self.s_d0)
		except: pass

		for Z,A,f in self.elements:
			try: self.chksum += float(Z)
			except: pass
			try: self.chksum += float(A)
			except: pass
			try: self.chksum += float(f)
			except: pass

	# ----------------------------------------------------------------------
	def modified(self):
		self._modified = time.strftime("%d/%m/%Y",time.localtime())

	# ----------------------------------------------------------------------
	def inputName(self, mode=None):
		if mode == Fluka.MODE:
			matname = self.name.split()[:8]
			if matname=="PHOSPHOR": matname = "PHOSPHO"
			return matname
		else:
			return self.name

	# ----------------------------------------------------------------------
	# Set the stoichiometry string
	# ----------------------------------------------------------------------
	def makeStoichiometry(self):
		if len(self.elements)==1:
			Z, A, f = self.elements[0]
			A = "" if A==0 else str(A)
			element = Element.byZ(Z)
			if element is not None:
				self.stoichiometry = f" {A}{element.symbol}"
			else:
				self.stoichiometry = ""
			return

		# Percentage or absolute
		mult = 1.0
		try:
			if max([f for Z, A, f in self.elements])<1.0:
				mult = 100.0
		except ValueError:
			pass

		cnt = 0
		stoichiometry = ""
		for Z, A, f in self.elements:
			if A==0: A=""
			else: A=str(A)

			f *= mult
			if float(int(f)) == f:
				f = str(int(f))
			else:
				if mult>1.0:
					f = f"{f:.3g}".strip()
				else:
					f = str(f)

			element = Element.byZ(Z)
			if element:
				stoichiometry += f" {A}{element.symbol}-{f},"
				cnt += 1
				if cnt == 10: break

		if stoichiometry:
			self.stoichiometry = stoichiometry[1:-1]
		else:
			self.stoichiometry = ""

	#-----------------------------------------------------------------------
	# Generate the list of MATERIAL/COMPOUND needed to define material
	# @param ignore	dict/set of materials already existing to be ignored
	# @return a list of material cards (MATERIAL/COMPOUND)
	#-----------------------------------------------------------------------
	def materialCard(self, matname, ignore=None, mode=''):
		# Compound, or a single element?
		cards = []
		if len(self.elements)==1:
			# Single element
			Z, A, f = self.elements[0]
			element = Element.byZ(Z)
			if A == 0: A = ""
			if mode == Fluka.MODE:
				what = [matname, Z, "", self.density, "", "", A]
				cards.append(FlukaCard("MATERIAL", what))
			elif mode == Phits.MODE:
				what = [matname, f"{A}{element.symbol}", 1.0]
				cards.append(PhitsCard("MAT", what))
		else:
			# Compound
			comment = f"{self.title}\n{self.notes}"
			if mode == Fluka.MODE:
				for Z, A, f in self.elements:
					element = Element.byZ(Z)
					if A==0:
						name = element.inputName(mode)
					else:
						name = f"{element.symbol.upper()}-{A}"
					if ignore and name in ignore: continue
					if A == 0: A = ""
					if mode == Fluka.MODE:
						what = [name, Z, "", element.density(), "", "", A]
						cards.append(FlukaCard("MATERIAL", what))
					#elif mode == Phits.MODE:
					#	what = [matname, f"{A}{element.symbol}", 1.0]
					#	cards.append(PhitsCard("MAT", what))

				what = [matname, "", "", self.density]
				cards.append(FlukaCard("MATERIAL", what, comment))

				# Create the compound
				if self.composition == 0:
					s1 = 1.0
					s2 = ""
				elif self.composition == 1:
					s1 = -1.0
					s2 = ""
				else:
					s1 = -1.0
					s2 = ""
				what = [matname]
				for e in self.elements:
					try: Z, A, f = e
					except: break
					element = Element.byZ(Z)
					if A == 0:
						ename = element.inputName(mode)
					else:
						ename = f"{element.symbol.upper()}-{A}"
						ename = ename[:8]
					what.append(s1*f)
					what.append(s2+ename)
				remain = len(self.elements)%3
				if remain > 0:
					while remain < 3:
						what.append("")
						what.append("")
						remain += 1
				cards.append(FlukaCard("COMPOUND", what))

			#-------------------------------------------------------
			elif mode == Phits.MODE:
				what = [matname]
				for Z, A, f in self.elements:
					element = Element.byZ(Z)
					if A==0:
						name = element.inputName(mode)
					else:
						name = f"{element.symbol.upper()}-{A}"
					if ignore and name in ignore: continue
					if A == 0: A = ""
					what.append(f"{A}{element.symbol}")
					if self.composition == 0:
						what.append(f)
					elif self.composition == 1:
						what.append(-f)
				cards.append(PhitsCard("MAT", what, comment))

		return cards

	#-----------------------------------------------------------------------
	def matPropCard(self, matname, mode):
		if not self.pressure and \
		   not self.rhor     and \
		   not self.ionization: return None
		if mode == "fluka":
			return FlukaCard("MAT-PROP",
				["", self.pressure, self.rhor, self.ionization, matname])

	#-----------------------------------------------------------------------
	def dpaCard(self, matname, mode):
		if not self.dpa_threshold: return None
		if mode == "fluka":
			return FlukaCard("MAT-PROP",
				["DPA-ENER", self.dpa_threshold, self.barcdpa,
					self.carcdpa, matname])

	#-----------------------------------------------------------------------
	def sternheimerCard(self, matname, mode):
		if not (self.s_cbar or self.s_x0 or
		        self.s_x1   or self.s_a  or
		        self.s_m    or self.s_d0): return None
		if mode == "fluka":
			return FlukaCard("STERNHEI",
					[matname, self.s_cbar, self.s_x0,
					self.s_x1, self.s_a, self.s_m, self.s_d0])

	#-----------------------------------------------------------------------
	@staticmethod
	def get(name):
		try:
			return Material.materials[name]
		except KeyError:
			pass
		return Material.names.get(name)

	#-----------------------------------------------------------------------
	# Initialize
	#-----------------------------------------------------------------------
	@staticmethod
	def init(filenames):
		# read card database and prepare CardInfo classes
		matini = configparser.RawConfigParser()
		if not matini.read(filenames, encoding="UTF-8"):
			raise Exception(f"Cannot open file {filenames}")

		# Read groups...
		try:
			for i in count(1):
				g = matini.get("groups",f"group.{i}")
				Material.groups.append(g)
		except configparser.NoOptionError:
			pass

		# Read Materials
		for material in matini.sections():
			if material == "groups": continue
			mat = Material(material)

			mat.name        = matini.get(material, "name")
			mat.group       = sorted(map(int, matini.get(material, "groups").split()))
			mat.density     = matini.getfloat(material, "density")
			mat.composition = Material._COMPOSITION.index(matini.get(material,
					"composition"))

			try:
				notes = matini.get(material, "notes")
				mat.notes = notes[1:].replace("\n$", "\n").strip()
			except configparser.NoOptionError: pass

			try: mat._modified = matini.get(material, "modified")
			except configparser.NoOptionError: pass

			mat.state  = Material._STATE.index(matini.get(material, "state"))

			try: mat.pressure = matini.getfloat(material, "pressure")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.rhor     = matini.getfloat(material, "rhor")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.ionization = matini.getfloat(material, "ionization")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.dpa_threshold = matini.getfloat(material, "dpa")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.barcdpa = matini.getfloat(material, "barcdpa")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.carcdpa = matini.getfloat(material, "carcdpa")
			except (configparser.NoOptionError, ValueError): pass

			try: mat.s_cbar = matini.getfloat(material, "s_cbar")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.s_x0   = matini.getfloat(material, "s_x0")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.s_x1   = matini.getfloat(material, "s_x1")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.s_a    = matini.getfloat(material, "s_a")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.s_m    = matini.getfloat(material, "s_m")
			except (configparser.NoOptionError, ValueError): pass
			try: mat.s_d0   = matini.getfloat(material, "s_d0")
			except (configparser.NoOptionError, ValueError): pass

			n = matini.getint(material, "elements")
			for i in range(1, n+1):
				Z = matini.getint(material, f"z.{i}")
				try:    A = matini.getint(material, f"a.{i}")
				except (configparser.NoOptionError, ValueError):
					A = 0
				f = matini.getfloat(material, f"f.{i}")
				mat.elements.append((Z, A, f))
			mat.make()
			Material.materials[mat.title] = mat
			for name in mat.name.split():
				Material.names[name] = mat

		# Cleanup
		del matini

	#-------------------------------------------------------------------------------
	# Create a new ini file only with the modified materials
	#-------------------------------------------------------------------------------
	@staticmethod
	def makeIni(ismodified=True):
		matini = configparser.RawConfigParser()

		# Create groups
		matini.add_section("groups")
		for i,g in enumerate(Material.groups,1):
			matini.set("groups", f"group.{i}", g)

		# Add materials
		for mat in Material.materials.values():
			if ismodified and mat._modified == "": continue
			section = mat.title
			matini.add_section(section)
			matini.set(section, "name",     mat.name)
			matini.set(section, "notes",
				"$"+mat.notes.replace("\n","\n$"))
			matini.set(section, "groups",   " ".join(map(str,mat.group)))
			matini.set(section, "density",  str(mat.density))
			matini.set(section, "state",    Material._STATE[mat.state])
			matini.set(section, "composition", Material._COMPOSITION[mat.composition])
			matini.set(section, "elements", len(mat.elements))
			matini.set(section, "dpa",      str(mat.dpa_threshold))
			matini.set(section, "barcdpa",  str(mat.barcdpa))
			matini.set(section, "carcdpa",  str(mat.carcdpa))
			matini.set(section, "pressure", str(mat.pressure))
			matini.set(section, "ionization", str(mat.ionization))
			for i,(Z,A,f) in enumerate(mat.elements,1):
				matini.set(section, f"z.{i}", Z)
				if A!=0: matini.set(section, f"a.{i}", A)
				matini.set(section, f"f.{i}", f)
			matini.set(section, "rhor",     str(mat.rhor))
			matini.set(section, "s_cbar",   str(mat.s_cbar))
			matini.set(section, "s_x0",     str(mat.s_x0))
			matini.set(section, "s_x1",     str(mat.s_x1))
			matini.set(section, "s_a",      str(mat.s_a))
			matini.set(section, "s_m",      str(mat.s_m))
			matini.set(section, "s_d0",     str(mat.s_d0))
			if mat._modified != "":
				matini.set(section, "modified", mat._modified)
		return matini
