# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Feb-2021

import re

import csg
import bmath
import Evaluation
from deck import CardInfo, Card, CSGCard

#===============================================================================
# Global dictionary containing all units constants
#===============================================================================
class McnpGlobalDict(Evaluation.GlobalDict):
	def __init__(self):
		super().__init__()
		self.redefineDistance_cm(1)	# in cm
		self.redefineEnergy_GeV(1e6)	# in eV	(MeV)
		self.redefineTime_s(1e8)	# in s	(shakes)

#===============================================================================
# Mcnp CardInfo class
#===============================================================================
class McnpCardInfo(CardInfo):
	# ----------------------------------------------------------------------
	# Validate card: return a list of failing variables.
	# ----------------------------------------------------------------------
	def validate(self, card, case=None):
		pb = super().validate(card, case)

		# Special check for regions
		if card.type()==Card.REGION:
			exp = csg.tokenize(card.expr())
			# Check expression
			try:
				csg.exp2rpn(exp)
				if exp and not csg.check(exp):
					pb.append(-1)
					pb.append("Invalid expression")
				else:
					surfaces = card.input.cardsCache(Card.BODY)
					cells    = card.input.cardsCache(Card.REGION)
					notadded = True
					for token in exp:
						if token in "-+:#@": continue
						if card.NAMEPAT.match(token):
							if token not in surfaces:
								if notadded:
									pb.append(-1)
									notadded = False
								pb.append(f"Invalid surface '{token}'")
						elif token[0]=="#":
							token = token[1:]
							if card.NAMEPAT.match(token):
								if token not in cells:
									if notadded:
										pb.append(-1)
										notadded = False
									pb.append(f"Invalid cell '{token}'")
						else:
							if notadded:
								pb.append(-1)
								notadded = False
							pb.append(f"Invalid token '{token}'")
			except csg.CSGException as ex:
				pb.append(-1)
				pb.append(str(ex))
		return pb

	# ----------------------------------------------------------------------
	# Post initialization after loading other inputs init
	# ----------------------------------------------------------------------
	@staticmethod
	def post_init():
		mode = McnpCard.MODE
		CardInfo.volumes[mode]   = ["CELL","cell"]
		CardInfo.materials[mode] = ["M", "m"]
		CardInfo.rotations[mode] = ["TR","TR*", "tr", "tr*"]

#===============================================================================
# McnpCard card class
#===============================================================================
class McnpCard(Card, CSGCard):
	"""MCNP input card"""
	MODE    = "mcnp"
	COMMENT = "C "
	MAXLEN  = 70

	# Patterns
	NAMEPAT    = re.compile(r"^[0-9]*$")

	cardInfo   = McnpCardInfo	# Pointer to Card Info
	globalDict = McnpGlobalDict()

	title    = Card.name
	setTitle = Card.setName

	# ----------------------------------------------------------------------
	# Change tag of card, reduce the number of whats to the new card
	# XXX Do not forget to update the Input.cards... before
	# ----------------------------------------------------------------------
	def setTag(self, tag):
		super().setTag(tag)
		if tag == "CELL":
			self._type = Card.CELL
			self._geo  = True

	# ---------------------------------------------------------------------
	# return string with only the cards
	# ----------------------------------------------------------------------
	def _toStr(self, prefix="", comments=False, **kwargs):
		#---------------------------------------------------------------
		# replace empty whats with "j"
		#---------------------------------------------------------------
		def empty2jumps(whats):
			# trim trailing empty whats
			while whats and whats[-1]=="":
				whats.pop()
			# replace empty whats with "j"
			i = 0
			count = 0
			while i<len(whats):
				if whats[i]=="":
					count += 1
					whats[i]="j"
				else:
					if count >= 3:
						i -= count
						whats[i] = f"{count}j"
						del whats[i+1:i+count]
					count = 0
				i += 1

		#---------------------------------------------------------------
		if self._type == Card.CELL:
			mat = self.what(1)
			out = f"{self.name()} {mat}"
			if mat not in ("-1","0"):
				out += f" {self.evalWhat(2)}"
			lines = [x.rstrip() for x in self.expr().splitlines()]
			if lines:
				lines[0] = f"{out} {lines[0]}"
			else:
				lines.append(out)
			# check line length for >MAXLEN and break it
			i = 0
			prefix = "\n     "
			while i<len(lines):
				if len(lines[i]) > McnpCard.MAXLEN-len(prefix):
					line = lines[i]
					insert = False
					while len(line) > McnpCard.MAXLEN-len(prefix):
						# find last space before the MAXLEN column
						try:
							space = line.rindex(' ',0,
								McnpCard.MAXLEN-len(prefix))
						except ValueError:
							space = McnpCard.MAXLEN-1

						if insert:
							lines.insert(i,line[:space])
						else:
							lines[i] = line[:space]
							insert = True
						i += 1
						line = line[space+1:].strip()
					lines.insert(i,line)
				i += 1

			prop = []
			for name,value in self.property.items():
				if name.startswith("imp:"):
					prop.append(f"{name}={value}")
			if prop: lines.append(" ".join(prop))
			return prefix.join(lines)

		#---------------------------------------------------------------
		elif self._type == Card.SURFACE:
			if self["tr"] is not None:
				out = f"{self.name()} {self['tr']} {self.tag}"
			else:
				out = f"{self.name()} {self.tag}"

			prevlen = 0
			#for i in range(1,self.info.maxwhats):
			for i in range(1,self.nwhats()):
				w = bmath.format(self.evalWhat(i))
				if w=="": w = "j"
				if len(out)+1+len(w)-prevlen > McnpCard.MAXLEN:
					out += "\n     "
					prevlen = len(out)-5	# minus 5 space chars before
				out += f" {w}"
			return out
		else:
			if self.tag in ("M","MT","TR","TR*"):
				out = f"{self.tag}{self.intWhat(0)}"
			else:
				out = self.tag
			whats = [bmath.format(self.evalWhat(i),useDot=False)
					for i in range(1,self.nwhats())]
			empty2jumps(whats)

			prevlen = 0
			for i,w in enumerate(whats,1):
				if len(out)+1+len(w)-prevlen > McnpCard.MAXLEN:
					out += "\n     "
					prevlen = len(out)-5	# minus 5 space chars before
				out += f" {w}"
			return out

	# ----------------------------------------------------------------------
	# Set a list of zones as expression
	# ... removing empty zones
	# ----------------------------------------------------------------------
	def setZones(self, zones):
		if zones:
			self.setText("\n: ".join(
				[x.strip() for x in zones if x.strip()]))
		else:
			self.setText("")
