# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Feb-2021

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

from collections import OrderedDict

import re
import math

import log
import bmath
import Global

from deck    import Input, Card
from .card   import McnpCard
from .config import McnpConfig

#===============================================================================
class McnpGeometry(Input):
	PAT_TRANS = re.compile(r"^(\*?)TR(\d+)(\*?) (.*)", re.DOTALL | re.IGNORECASE)

	#-----------------------------------------------------------------------
	# Scan transformations
	# @return rotation matrix from a rot-defi cards or a dictionary with all
	#         transformations
	# FIXME add the other matrix definitions (Euler, etc)
	#-----------------------------------------------------------------------
	def getTransformation(self, idx=None, rotdefi=None, inv=False):
		neg = False
		if isinstance(idx,str):
			if idx and idx[0]=='-':
				# FIXME do I need negative for mcnp?
				neg = True
				idx = idx[1:]
		elif isinstance(idx,int):
			if idx<0:
				# FIXME do I need negative for mcnp?
				neg = True
				idx = -idx
		if inv: neg = not neg

		# return value if cached
		if idx is not None:
			if rotdefi is not None:
				matrix = rotdefi.get(idx)
				if matrix is None:
					return bmath.Matrix(4, type=1)
				elif neg:
					matrix = matrix.inverse()
				return matrix
			matrix = bmath.Matrix(4, type=1)
			assert False	# FIXME temporary for debugging to check if needed

		#---------------------------------------------------------------
		# code similar to mcnpf/arbobv.F
		# provide 2 arbitrary base vectors b and c perpendicular to a.
		# the length of a is assumed to be 1.
		# make abc a right-handed system.
		# @return None if a=0.
		#         b,c otherwise
		# b is made catawampous to avoid singularities.
		#---------------------------------------------------------------
		def arbobv(a):
			j  = 0
			if abs(a[1]) > abs(a[0]): j=1
			if abs(a[2]) > abs(a[j]): j=2
			k = (j+1)%3
			l = 3-j-k
			t = a[j]**2 + a[k]**2
			if t==0: return
			b = bmath.Vector()
			b[l] = .0637922
			b[k] = (a[j] * math.sqrt(t-(t+a[l]**2)*b[l]**2) - a[k]*a[l]*b[l]) / t
			b[j] = -(b[k]*a[k]+b[l]*a[l]) / a[j]
			b.normalize()
			c = a^b
			c.normalize()
			return b,c

		#---------------------------------------------------------------
		# Scan transformations cards
		#---------------------------------------------------------------
		rotdefi = {}
		for card in self.cardsSorted(Card.ROTATION,-1).values():
			if not card.nwhats(): continue
			matrix = bmath.Matrix(4, type=1)
			df = 0 if card.tag=="TR" else 90	# cos(90)=0
			U = bmath.Vector(card.numWhat( 4,df), card.numWhat( 5,df), card.numWhat( 6,df))
			V = bmath.Vector(card.numWhat( 7,df), card.numWhat( 8,df), card.numWhat( 9,df))
			W = bmath.Vector(card.numWhat(10,df), card.numWhat(11,df), card.numWhat(12,df))
			if card.tag == "TR*":
				for i in range(3):
					U[i] = math.cos(math.radians(U[i]))
					V[i] = math.cos(math.radians(V[i]))
					W[i] = math.cos(math.radians(W[i]))
			#print("card=",card)
			#print("U=",U)
			#print("V=",V)
			#print("W=",W)
			if card.intWhat(13) == 2:
				matrix = bmath.Matrix.eulerRotation(U[2],U[1],U[0])
			else:
				zeroU = U.normalize()<1e-10
				zeroV = V.normalize()<1e-10
				zeroW = W.normalize()<1e-10
				#print(f"zero U={zeroU} V={zeroV} W={zeroW}")
				if zeroU and zeroV and zeroW:
					pass	# Unit matrix do nothing
				else:
					if zeroU or zeroV or zeroW:
						# at least one vector is not zero
						# calculate the zero components
						try:
							if zeroU and not zeroV and not zeroW: # U=0
								#print(">>>>> U=0")
								U = V.cross(W)
								U.normalize()
							elif not zeroU and zeroV and not zeroW: # V=0
								#print(">>>>> V=0")
								V = W.cross(U)
								V.normalize()
							elif not zeroU and not zeroV and zeroW: # W=0
								#print(">>>>> W=0")
								W = U.cross(V)
								W.normalize()
							elif not zeroU:
								#print(">>>>> V,W=0")
								V,W = arbobv(U)
							elif not zeroV:
								#print(">>>>> U,W=0")
								W,U = arbobv(V)
							elif not zeroW:
								#print(">>>>> U,V=0")
								U,V = arbobv(W)
						except:
							log.error(f"TR{card.name()} invalid orthogonal system")
							U = bmath.Vector.X
							V = bmath.Vector.Y
							W = bmath.Vector.Z
					else:
						# Check for orthogonality
						V,W = U.gramSchmidt(V,W)
						V.normalize()
						W.normalize()

					for i in range(3):
						matrix[i][0] = U[i]
						matrix[i][1] = V[i]
						matrix[i][2] = W[i]
					#print(f"Matrix1=\n{matrix}")

				if card.intWhat(13) >= 0:
					matrix[0][3] = card.numWhat(1)
					matrix[1][3] = card.numWhat(2)
					matrix[2][3] = card.numWhat(3)
				else:
					matrix *= bmath.Matrix.translate(-card.numWhat(1),
									 -card.numWhat(2),
									 -card.numWhat(3))
			rotdefi[str(card.name())] = matrix

		return rotdefi

	#-----------------------------------------------------------------------
	# Return the body properties as a dictionary of the active/enabled
	# bodies
	#-----------------------------------------------------------------------
	def bodyProperties(self, rotdefi):
		for card in self.cardsSorted(Card.SURFACE, -1).values():
			if card.ignore(): continue
			if card["tr"]:
				try:
					card[Card.MATRIX] = rotdefi[str(card["tr"])]
				except KeyError:
					log.error(f"Surface '{card.name()}' "
						f"transformation TR{card['tr']} not found")
			else:
				card[Card.MATRIX] = None

	#-----------------------------------------------------------------------
	# @return (line,prop) clean line and dictionary with properties
	#         prop[tag] = value
	#-----------------------------------------------------------------------
	def _parseProperties(self, line):
		prop = {}

		m = Mcnp.PAT_PROP.search(line)
		if m is None: return line,{}

		pstr = line[m.start():]	# property string
		line = line[:m.start()].rstrip()
		while pstr:
			m = Mcnp.PAT_PROP.search(pstr)
			if m is None: break
			tag = m.group(1).lower()
			pstr = pstr[m.end():]
			m = Mcnp.PAT_PROP.search(pstr)
			if m is not None:
				value = pstr[:m.start()].strip()
				pstr = pstr[m.start():]
			else:
				value = pstr.strip()
				pstr = None
			prop[tag] = value
		return line,prop

	#-----------------------------------------------------------------------
	# nR: repeat the proceeding entry n times
	# nI: insert n linear interpolate entries
	# xM: multiply proceeding entry by x
	# nJ: replace by the default value (or jump) n entries (J alone is accepted)
	#-----------------------------------------------------------------------
	def _parseFeatures(self, command):
		# Parse nR
		while True:
			m = Mcnp.PAT_nR.match(command)
			if m is None: break
			command = m.group(1) \
				  + " ".join([m.group(2)]*(int(m.group(3))+1)) \
				  + m.group(4)

		# Parse nI
		while True:
			m = Mcnp.PAT_nI.match(command)
			if m is None: break
			try:
				command = m.group(1) + m.group(2)
				a = float(m.group(2))
				n = int(m.group(3))
				b = float(m.group(4))
				d = (b-a)/(n+1)
				x = a
				for _ in range(n):
					x += d
					command += " %g"%(x)
				command += " " + m.group(4) + m.group(5)
			except ValueError:
				break

		# Parse xM
		while True:
			m = Mcnp.PAT_xM.match(command)
			if m is None: break
			try:
				command = m.group(1) + m.group(2) + " " + \
					  str(float(m.group(2))*float(m.group(3))) + \
					  m.group(4)
			except ValueError:
				break

		# Parse single J
		while True:
			m = Mcnp.PAT_J.match(command)
			if m is None: break
			command = m.group(1) + "0" + m.group(2)

		# Parse nJ
		while True:
			m = Mcnp.PAT_nJ.match(command)
			if m is None: break
			command = m.group(1) + " ".join(["0"]*int(m.group(2))) + m.group(3)

		return command

#===============================================================================
# Mcnp Input class
#===============================================================================
class Mcnp(McnpGeometry):
	MODE    = McnpCard.MODE

	SECTION_TITLE = 0
	SECTION_CELL  = 1
	SECTION_SURF  = 2
	SECTION_MAIN  = 3
	SECTION_END   = 4

	PAT_nR       = re.compile(r"^(.*?\s)(\S+)\s*\b(\d+)R\b(.*)", re.DOTALL | re.IGNORECASE)
	PAT_nI       = re.compile(r"^(.*?\s)(\S+)\s*\b(\d+)I\s*(\S*)\s*(.*)", re.DOTALL | re.IGNORECASE)
	PAT_xM       = re.compile(r"^(.*?\s)(\S+)\s*\b(\d+)M\b(.*)", re.DOTALL | re.IGNORECASE)
	PAT_J        = re.compile(r"^(.*?)\bJ\b(.*)", re.DOTALL | re.IGNORECASE)
	PAT_nJ       = re.compile(r"^(.*?)\b(\d+)J\b(.*)", re.DOTALL | re.IGNORECASE)

	#PAT_PROP     = re.compile(r"([a-z:]+) *=", re.IGNORECASE)
	PAT_PROP      = re.compile(r"([a-z][a-z:,]*)[ =]*", re.IGNORECASE)

	PAT_MATERIAL = re.compile(r"^M(\d+) (.*)", re.DOTALL | re.IGNORECASE)
	#PAT_CELL     = re.compile(r"(.*?)#(\d+)(.*)", re.DOTALL | re.IGNORECASE)
	#PAT_RMCELL   = re.compile(r"#\d+", re.DOTALL | re.IGNORECASE)
	#PAT_IMP      = re.compile(r"^IMP:(.) (.*)", re.DOTALL | re.IGNORECASE)
	#PAT_SURFTRCL = re.compile(r"^(S\d+)(C\d+)$", re.IGNORECASE)

	config  = McnpConfig()		# configuration information
	defaultMaterials = OrderedDict([
				("-1", McnpCard("M",[-1])),	# black hole
				( "0", McnpCard("M",[0]))	# vacuum
			])
	zaid = []		# list of zaid
	sab  = []		# list of Sab elements

	# ----------------------------------------------------------------------
	def __init__(self):
		super().__init__()
		self.mode       = Mcnp.MODE
		self.globalDict = McnpCard.globalDict

	#-----------------------------------------------------------------------
	# Cards
	#-----------------------------------------------------------------------
	# Return cards of a specific tag sorted with _pos
	# @param tag     string tag
	#                list of tags
	#                Card.type(), BODY, SOLID, VOLUME...
	# @param which   look the Input.cardsSorted()
	#-----------------------------------------------------------------------
	def cardsSorted(self, tag, which=0):
		return super().cardsSorted(tag, which)

	#-----------------------------------------------------------------------
	# Return material list
	#-----------------------------------------------------------------------
	def materialList(self, which=0, *kw):
		lst = super().materialList(which)
		lst.extend([x.name()
			for x in self.cardsSorted(Card.MATERIALS,-1).values()])
		return lst

	#-----------------------------------------------------------------------
	# Return the region properties as a dictionary of the active/enabled
	# regions
	#-----------------------------------------------------------------------
	def regionProperties(self):
		#matList = self.materialList()
		matDict = self.materialDict()
		#rotDefi = {}

		# special regions
		blckhole = matDict["-1"]
		#vacuum   = matDict["0"]

		#---------------------------------------------------------------
		# Build dictionaries and clean up regions from Input and voxel
		n = 0
		regionList = []
		for region in self.cardsSorted(Card.REGION,-1).values():
			material = matDict.get(str(region.what(1)),None)
			n += 1
			region["@n"] = n		# Assign default values to region
			region[Card.TYPE]     = int(material is blckhole)
			region[Card.ASSIGNMAT]= region	# Card that assigned the material
			region[Card.MATERIAL] = material
			#region["@biasingA"]  = None	# All particles biasing
			#region["@biasingH"]  = None	# Hadrons
			#region["@biasingE"]  = None	# emf
			#region["@biasingN"]  = None	# low-neutrons
			#region[Card.LATTICE] = None
			#region["@rotdefi"]   = ""
			#region[Card.CONT]    = False	# has continuation cards
			regionList.append(region)

	#=======================================================================
	# I/O
	# methods for Input IO operations
	#=======================================================================
	#-----------------------------------------------------------------------
	# Export method
	#-----------------------------------------------------------------------
	def _save(self):
		Global.dump("\n*-* Writing mcnp file")

		# title
		try:
			self.writeline(self.cards("TITLE")[0].text())
		except:
			self.writeline("flair input")

		# cells
		for card in self.cardsSorted("CELL",-1).values():
			if not isinstance(card, McnpCard): continue
			if card.comment():
				self.writeline(card.commentStr().expandtabs())
			self.writeline(card.toStr())
		self.writeline()

		# surfaces
		for card in self.cardsSorted(Card.SURFACE,-1).values():
			if not isinstance(card, McnpCard): continue
			if card.comment():
				self.writeline(card.commentStr().expandtabs())
			self.writeline(card.toStr())
		self.writeline()

		# all other cards
		for card in self.allcards():
			if card.ignore(): continue
			if not isinstance(card, McnpCard): continue
			if card.type() in (Card.CELL, Card.SURFACE): continue
			if card.tag == "TITLE": continue
			if card.comment():
				self.writeline(card.commentStr().expandtabs())
			self.writeline(card.toStr())

	#-----------------------------------------------------------------------
	# Import/load a macro and the related moira input files
	#-----------------------------------------------------------------------
	def _load(self):
		Global.dump("\n*-* Reading mcnp file")
		# Initialize internal variables for reading
		self._density   = {}			# material density combination
		self._trcl      = {}			# list of inlined TRCL
		self.cells      = []
		self.regions    = []
		self.materials  = []
		self.importance = []

		cardcomment = ""# comment of the card
		comment = ""	# comments at the end of the card which are meant for
				# the next one
				# first populate the last comment and if the card
				# continuous append it to the comment
		cmdline = ""

		def appendComment(comment, line):
			if comment:
				return f"{comment}\n{line}"
			else:
				return line

		cont = False
		self._section = Mcnp.SECTION_TITLE	# section of file during reading
		while True:
			line = self.readline()
			if line is None: break

			# Check for comments
			if line and (line in ("c","C") or line[:2] in ("c ", "C ")):
				comment = appendComment(comment, line[1:])
				continue
			dollar = line.find("$")
			if dollar>=0:
				comment = appendComment(comment, line[dollar+1:])
				line = line[:dollar].rstrip()

			# Empty line
			if not line:
				# and no comment then change section
				if dollar<0:
					self.__parseCommand(cmdline, cardcomment)
					self.__parseCommand(None)	# Advance section
				cardcomment = comment
				comment = ""
				cmdline = ""

			elif cont or line[:5]=="     " or line[0]=="\t":
				if line[-1]=="&":
					line = line[:-1].rstrip()
					cont = True
				else:
					cont = False
				# Continuation
				cmdline += f"\n{line.strip()}"
				if comment:
					cardcomment = appendComment(cardcomment, comment)
					comment = ""

			else:
				if line[-1]=="&":
					line = line[:-1].rstrip()
					cont = True
				else:
					cont = False
				# Execute command
				if cmdline:
					if self._section == Mcnp.SECTION_TITLE:
						self.add(McnpCard("TITLE",[cmdline], cardcomment))
						self._section += 1
					else:
						self.__parseCommand(cmdline, cardcomment)
				cardcomment = comment
				comment = ""
				cmdline = line

		# execute last command
		self.__parseCommand(cmdline, comment)

		# Write region properties
		#self.__regionProperties()

		del self._density
		del self._trcl

	#-----------------------------------------------------------------------
	# parse Mcnp command line
	#-----------------------------------------------------------------------
	def __parseCommand(self, command, comment=""):
		if command:
			command = self._parseFeatures(command)

		if command is None:
			self._section += 1	# advance section

		elif not command:
			return

		elif self._section == Mcnp.SECTION_CELL:
			self.__parseCell(command, comment)

		elif self._section == Mcnp.SECTION_SURF:
			self.__parseSurface(command, comment)

		elif self._section == Mcnp.SECTION_MAIN:
			# Generic commands
			ucommand = command.upper()
			m = Mcnp.PAT_MATERIAL.match(ucommand)
			if m:
				self.add(McnpCard("M",[m.group(1)] + m.group(2).split(), comment))
				return

			m = McnpGeometry.PAT_TRANS.match(ucommand)
			if m:
				deg = "*" if "*" in (m.group(1), m.group(3)) else ""
				name = m.group(2)
				try:
					whats = list(map(float,m.group(4).split()))
					whats.insert(0,name)
				except:
					log.error(f"TR{m.group(2)} parsing transformation '{ucommand}'")
				self.add(McnpCard(f"TR{deg}", whats))
				return

			#m = Mcnp.PAT_IMP.match(ucommand)
			#if m:
			#	var = f"@IMP:{m.group(1)}"
			#	for region,imp in zip(self.regions, map(float, m.group(2).split())):
			#		region[var] = imp

			token = command.split()
			# FIXME replace : to avoid mode:tag problem
			self.add(McnpCard(token[0].upper().replace(":","%"),token[1:],comment))

		#if comment: log.say("C",comment)
		#log.say(self._section, command)

	#-----------------------------------------------------------------------
	def __parseCell(self, command, comment):
		#Global.dump(">C>",command, level=2)
		try:
			name,mat,expr = command.split(maxsplit=2)
		except ValueError:
			log.error(f"parsing cell '{command}'")
			return

		if mat not in ("0","-1"):
			density,expr = expr.split(maxsplit=1)
		else:
			density = ""
		expr,prop = self._parseProperties(expr)
		# FIXME like but...
		self.add(McnpCard("CELL",[name,mat,density],comment,expr,prop=prop))

	#-----------------------------------------------------------------------
	def __parseSurface(self, command, comment):
		#Global.dump(">S>",command, level=2)
		token = command.split()
		try:
			name = token.pop(0)	# split into 2 instructions
		except ValueError:
			log.error(f"Surface {name} invalid id")
			return False
		surface = token.pop(0)
		try:
			prop = {"tr" : str(int(surface))}
			surface = token.pop(0)
		except ValueError:
			prop = None
		arg = [name] + token
		# TODO
		#   * If a surface is preceded by an asterisk * then an reflective
		#     surface is created
		self.add(McnpCard(surface.upper(),arg,comment,prop=prop))

	#-----------------------------------------------------------------------
	@classmethod
	def newCard(cls, *args, **kwargs):
		return McnpCard(*args, **kwargs)
