# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

import os
import re
import csg
import bmath
import Evaluation
from deck import CardInfo, Card, CSGCard

#===============================================================================
# Global dictionary for moira, containing all units constants
#===============================================================================
class MoiraGlobalDict(Evaluation.GlobalDict):
	def __init__(self):
		super().__init__()
		#self.redefineDistance_cm(10)	# in mm
		self.redefineEnergy_GeV(1000)	# in MeV
		self.redefineTime_s(1e9)	# in ns

#===============================================================================
# Moira CardInfo class
#===============================================================================
class MoiraCardInfo(CardInfo):
	# ----------------------------------------------------------------------
	# Post initialization after loading other inputs init
	# ----------------------------------------------------------------------
	@staticmethod
	def post_init():
		mode = MoiraCard.MODE
		# Populate volumes
		CardInfo.volumes[mode] = [ x.tag for x in CardInfo.cards(mode)
					if CardInfo.GEOMETRY_VOLUME_GROUP in x.group and
					   x.tag[:4] in ("VOLU","CELL")]
		CardInfo.volumes[mode].sort()

		# Populate material definition
		CardInfo.materials[mode] = ["ELEM", "ELEM_FROM_ISOT",
					"ISOT", "MATE",
					"MIXT", "MIXT_BY_WEIGHT",
					"MIXT_BY_NATOMS", "MIXT_BY_VOLUME"]
		CardInfo.materials[mode].sort()
		CardInfo.rotations[mode] = "ROTM"

	# ----------------------------------------------------------------------
	# Validate card: return a list of failing variables.
	# ----------------------------------------------------------------------
	def validate(self, card, case=None):
		pb = super().validate(card, case)
		if card.input is None: return pb

		# Special check for regions
		if card.type()==Card.CELL:
			exp = csg.tokenize(card.expr())
			# Check expression
			try:
				csg.exp2rpn(exp)
				if exp and not csg.check(exp):
					pb.append(-1)
					pb.append("Invalid expression")
				else:
					bodies = card.input.cardsCache(Card.BODY)
					notadded = True
					for token in exp:
						if token in "-+|@": continue
						if card.NAMEPAT.match(token):
							# XXX Check in body list
							if token not in bodies:
								if notadded:
									pb.append(-1)
									notadded = False
								pb.append(f"Invalid body '{token}'")
						else:
							if notadded:
								pb.append(-1)
								notadded = False
							pb.append(f"Invalid token '{token}'")
			except csg.CSGException as ex:
				pb.append(-1)
				pb.append(str(ex))

		return pb

#===============================================================================
# MoiraCard base class
# FIXME:
#	subclass (dict) to replace the whats and properties
#	- whats / list arguments are assigned as [#] = value, starting from 0
#	- properties will be assigned as [name] = value
#	- all internal variables not to be saved will be prefixed
#         with "@" [@name] = value
#===============================================================================
class MoiraCard(Card, CSGCard):
	"""moira input card"""
	MODE       = "moira"
	COMMENT    = "// "
	COMMENTMAC = "# "

	DEF_ROTM   = "__IDENT__"	# default rotation that will be generated always

	cardInfo   = MoiraCardInfo	# Pointer to Card Info
	globalDict = MoiraGlobalDict()

	unitConv   = {  "L"  : 10,	# length cm in mm
			"iL" : 0.1,	# inverse length 1/cm in 1/mm
			"D"  : 10*10,	# length² cm² in mm²
			"iD" : 0.1*0.1,	# inverse length²  1/cm² in 1/mm²
			"E"  : 1000,	# energy GeV in MeV
			#"T" : 1e9,	# time s in ns
			#"A" : 1
		 }

	# ----------------------------------------------------------------------
	# Change tag of card, reduce the number of whats to the new card
	# XXX Do not forget to update the Input.cards... before
	# ----------------------------------------------------------------------
	def setTag(self, tag):
		super().setTag(tag)
		if tag == "PLACE" or tag.startswith("PLACE."):
			self._geo = True
		elif tag in CardInfo.volumes.get(MoiraCard.MODE,[]):
			self._type = Card.CELL
			self._geo  = True
		elif tag == "LATTICE":
			self._geo  = True
		elif tag in CardInfo.typedef[MoiraCard.MODE]["solid"]:
			self._type = Card.SOLID
			self._geo  = True

	# ----------------------------------------------------------------------
	setTitle = Card.setName

	# ----------------------------------------------------------------------
	# test card against a filter-function
	# ----------------------------------------------------------------------
	def test(self, func, var, value, keepcase=False):
		# Solid, Volume or Cell
		if var in ("solid", "volume"):
			if not self._geo: return
			varlist = self.findType(var)
			if varlist:
				if not keepcase: varlist = [str(x).upper() for x in varlist]
				return value in varlist

		# particle
		#elif var=="particle":

		# Body
		elif var=="body":
			if not self._geo: return
			if self.tag == "CELL":
				opt = re.MULTILINE|re.DOTALL
				if not keepcase: opt |= re.I
				return re.match(rf".*\b{value}\b.*", self._text,opt) is not None
			if self.tag not in CardInfo.bodies[self.MODE]: return False
			return (str(self.name()) if keepcase else str(self.name()).upper()) == value

		else:
			return super().test(func, var, value, keepcase)

	# ----------------------------------------------------------------------
	# String functions
	# ----------------------------------------------------------------------
	# Return a formatted comment string
	# ----------------------------------------------------------------------
	def commentStr(self):
		tag = self.tag if self.info.otag is None else self.info.otag
		cs = (self.COMMENTMAC if (tag[0] == "/" or self.info.otag) else self.COMMENT)

		# Remove trailing spaces from lines.
		lines = [x.rstrip() for x in self._comment.split('\n')]

		# Prepend with #, when line is not empty.
		lines = [(cs+x if x else x) for x in lines]

		return "\n".join(lines)

	# ----------------------------------------------------------------------
	def evalWhatOrName(self, n, zero=False):
		w = super().evalWhatOrName(n, zero)
		if isinstance(w,str) and " " in w:
			return f'"{w}"'
		return w

	# ---------------------------------------------------------------------
	# return string with only the cards
	# ----------------------------------------------------------------------
	def _toStr(self, prefix="", **kwargs):
		case = self.case()
		try:    vtype = self.info.vtype[case]
		except: vtype = None

		# replace empty whats with default values "!"
		try:    default = self.info.default[case]
		except: default = None

		# FIXME optimize the following loops both the whats
		# and the correction for the default, length

		# Card with parameters
		whats = [self.evalWhat(i) for i in range(self.nwhats())]

		# special cases
		if self.tag == "CELL":	# FIXME maybe add an option in ini
			if len(whats)<self.info.maxwhats:
				whats.extend(["!"]*(self.info.maxwhats-len(whats)))
		n    = len(whats)-1
		last = not self.text()

		# correct whats for units and for empty/!
		for i,w in enumerate(reversed(whats)):
			j = n-i
			if w in ("","!"):
				try:
					whats[j] = default[j]
				except:
					if not last: whats[j] = "!"
			else:
				last = False
				# check for default values and special conversion
				# for length/angle
				try:	# Convert units
					whats[j] = bmath.format(float(w) * MoiraCard.unitConv[vtype[j]],
						useDot=False)
				except (KeyError, ValueError, IndexError, TypeError):
					whats[j] = str(w)

		# Extension macro-cards, in compact mode:
		# they will not be printed with a prepended path!
		is_compact_extension = self.info.compact_mode and self.info.is_extension
		# Standard mode
		if self.info.otag is not None:
			tag = self.info.otag
		# Compact mode
		# Base card: its path is printed
		elif not is_compact_extension:
			tag = self.tag
		# Extension card: do not print any path!
		else:
			tag = ""

		# Extension macro-cards: no prepended path!
		if is_compact_extension:
			s = ""
		elif tag[0] == "/":
			s = prefix + tag
		else:
			try:
				# Special case for SOLID.xxx BODY.xxx VOLU.xxx
				# that has special syntax
				pdot = tag.index(".")
				tag  = tag[:pdot]
				if self.type() == Card.SOLID:
					whats.insert(1, self.tag[pdot+1:])
				elif tag == "VOLU":
					# Move material to the end
					if len(whats)>1:
						whats.append(whats.pop(1))
						whats.insert(1, self.tag[pdot+1:])
				s = f"{prefix}:{tag}"
			except ValueError:
				if self._type == Card.BODY:
					whats.insert(1, tag)
					tag = "BODY"
				s = f"{prefix}:{tag}"

		if whats:
			# special treatment
			if tag.startswith("MIXT"):
				s += " " + " ".join(whats[:3])
				if len(whats)>3:
					maxlen = max([len(x) for x in whats[3::2]])
					for i in range(3,len(whats),2):
						try:
							s += "\n\t\t%-*s %s"%(maxlen, whats[i], whats[i+1])
						except IndexError:
							s += "\n\t\t%-*s"%(maxlen, whats[i])
			else:
				s += " " + " ".join(whats)

		if self.text():
			for i,line in enumerate(self.text().splitlines()):
				if i==0:
					s += " "
				else:
					s += "\n\t"
				s += line

		# Organize properties
		used = set()
		# Maybe go with the order that was defined
		for name,var in self.info.property.items():
			v = self.propertyStr(name, used, " ", "!", True)
			if v is not None:
				if s: s += "\n"
				# Standard mode
				if not self.info.compact_mode:
					# The property has its own tag: use it!
					if var.otag is not None:
						s += var.otag
					# Property has no tag: use dir name.
					else:
						s += os.path.dirname(tag)
					# Append property name
					s += "/" + v
				# Compact mode
				else:
					# Prepend spaces
					# For extension macro-cards, need to look at the directory
					# from the previously printed card (when printing macro file)!
					directory_name = kwargs.pop("previous_card_directory", "") \
							 if is_compact_extension \
							    else os.path.dirname(tag)
					numSpaces = len(directory_name) + 1
					# When clicking on extension card, no directory can be available.
					if is_compact_extension and not directory_name:
						numSpaces = 22
					s += " " * numSpaces

					# The property has its own tag: use it!
					if var.otag is not None:
						s += var.otag + "/"

					# Append property name
					s += v

		# For anything else
		for n,v in self.property.items():
			if n in used: continue
			if n[0]=="@": continue
			if v in ("",None): continue
			if s: s += "\n"
			s += tag + f"/{n}={v}"
		return s
