# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   21-Mar-2013

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
import math
from xml.dom import minidom
from operator import itemgetter
from collections import namedtuple, OrderedDict

import bmath
import Global

import log
from polygon import Polygon

from deck  import Card
from fluka import Fluka, FlukaCard

from .card  import MoiraCard, MoiraGlobalDict
from .input import Moira

# VecMatrix for position and rotation
VecMatrix = namedtuple("VecMatrix", ["vector", "matrix"])
PhysVol   = namedtuple("PhysVol", ["name", "cname", "position", "rotation", "param"])

#===============================================================================
class GdmlException(Exception): pass

#===============================================================================
# Gdml Importer/Exporter class
#===============================================================================
class Gdml:
	NAMEHEX = re.compile(r"^(.*?)(0x[0-9a-fA-F]+)?(?:|_rot|_pos)$")
	SYMBOL  = re.compile(r"^([A-Za-z][A-Za-z]*)(\d+)?(0x[0-9a-fA-F]+)?$")
	TRANS   = str.maketrans(":","_")
	SMALL   = 1e-10
	UNITS   = MoiraGlobalDict()

	#-----------------------------------------------------------------------
	# Gdml constructor
	#-----------------------------------------------------------------------
	def __init__(self, filename=None):
		self.init()
		log.repeat_clear()
		if filename is not None:
			self.read(filename)

	#-----------------------------------------------------------------------
	# FIXME temporary use all dictionaries as such
	#-----------------------------------------------------------------------
	def init(self):
		self.constant = Gdml.UNITS.copy()
		self.position = OrderedDict()
		self.positionMatrix = OrderedDict()
		self.rotation = OrderedDict()
		self.rotationMatrix = OrderedDict()
		self.isotope  = OrderedDict()
		self.element  = OrderedDict()
		self.material = OrderedDict()
		self.solids   = OrderedDict()
		self.volumes  = OrderedDict()
		self.variable = OrderedDict()
		self.world    = None

	#-------------------------------------------------------------------------------
	@staticmethod
	def unit(x):
		try:
			return Gdml.UNITS[x]
		except KeyError:
			log.rerror(f"Gdml: unit {x} not found")
			raise

	#-----------------------------------------------------------------------
	def eval(self, x, u=1.0):
		return eval(x, self.constant, self.variable)*u

	#-------------------------------------------------------------------------------
	@staticmethod
	def uniqueName(name, dictionary, prefix):
		name = name.translate(Gdml.TRANS)
		pat = Gdml.NAMEHEX.match(name)
		if pat is not None:
			newname = oname = pat.group(1)
		else:
			newname = oname = name
		if not Card.NAMEPAT.match(newname):
			newname = prefix

		i = 0
		while newname in dictionary:
			i += 1
			newname = oname + str(i)
		dictionary[newname] = name
		if name != newname:
			dictionary[name] = newname
		return newname

	#-----------------------------------------------------------------------
	# read Gdml input
	#-----------------------------------------------------------------------
	def read(self, filename, setup=True):
		self.filename = filename
		if Global.verbose: log.say(">>> Gdml::read",filename)
		with open(self.filename,"r") as f:
			xml = minidom.parse(f)

		if Global.verbose: log.say(">>> Gdml::parseDefine")
		for node in xml.getElementsByTagName("define"):
			self._parseDefinitions(node)
		if Global.verbose: log.say(">>> Gdml::parseMaterials")
		for node in xml.getElementsByTagName("materials"):
			self._parseMaterials(node)
		if Global.verbose: log.say(">>> Gdml::parseSolids")
		for node in xml.getElementsByTagName("solids"):
			self._parseSolids(node)
		if Global.verbose: log.say(">>> Gdml::parseStructure")
		for node in xml.getElementsByTagName("structure"):
			self._parseStructure(node)
		if Global.verbose: log.say(">>> Gdml::parseSetup")
		if setup: self._parseSetup(xml)

	#-----------------------------------------------------------------------
	# parse constant
	#-----------------------------------------------------------------------
	def _parseNameValue(self, node):
		name = node.attributes["name"].value
		try: value = self.eval(node.attributes["value"].value)
		except KeyError: value = 0.0
		return name, value

	#-----------------------------------------------------------------------
	# parse a value with a unit
	#-----------------------------------------------------------------------
	def _parseValue(self, node):
		try:
			u = Gdml.unit(node.attributes["unit"].value)
		except KeyError:
			u = 1.0
		try:
			return self.eval(node.attributes["value"].value, u)
		except KeyError:
			return 0.0

	#-----------------------------------------------------------------------
	# parse expression
	#-----------------------------------------------------------------------
	def _parseExpression(self, node):
		name = node.attributes["name"].value
		for child in node.childNodes:
			if child.nodeType != child.TEXT_NODE: continue
			expr = child.data
			break
		try: value = self.eval(expr)
		except KeyError: value = 0.0
		return name, value

	#-----------------------------------------------------------------------
	# Parse vector
	#-----------------------------------------------------------------------
	def _parseVector(self, node, u):
		name = str(node.attributes["name"].value)

		try:
			u = Gdml.unit(node.attributes["unit"].value)
		except KeyError:
			u = 1.0

		try:
			x = self.eval(node.attributes["x"].value, u)
		except KeyError:
			x = 0.0

		try:
			y = self.eval(node.attributes["y"].value, u)
		except KeyError:
			y = 0.0

		try:
			z = self.eval(node.attributes["z"].value, u)
		except KeyError:
			z = 0.0

		return name,bmath.Vector(x,y,z)

	#-----------------------------------------------------------------------
	# Parse position
	#-----------------------------------------------------------------------
	def _parsePosition(self, node):
		name, pos = self._parseVector(node, 1.0)
		self.position[name] = VecMatrix(pos*Gdml.UNITS["mm"], bmath.Matrix.translate(pos))
		return name

	#-----------------------------------------------------------------------
	# Parse rotation
	#-----------------------------------------------------------------------
	def _parseRotation(self, node):
		name, rot = self._parseVector(node, 1.0)

		if rot[2] != 0.0:
			matrix = bmath.Matrix.rotZ(-rot[2])
		else:
			matrix = None

		if rot[1] != 0.0:
			m = bmath.Matrix.rotY(-rot[1])
			if matrix is None:
				matrix = m
			else:
				matrix = m * matrix

		if rot[0] != 0.0:
			m = bmath.Matrix.rotX(-rot[0])
			if matrix is None:
				matrix = m
			else:
				matrix = m * matrix

		if matrix is None:
			matrix = bmath.Matrix(4,type=1)

		self.rotation[name] = VecMatrix([math.degrees(x) for x in rot], matrix)
		return name

	#-----------------------------------------------------------------------
	# Parse definitions
	#-----------------------------------------------------------------------
	def _parseDefinitions(self, node):
		for child in node.childNodes:
			if child.localName is None:
				continue

			if child.localName == "constant":
				name, value = self._parseNameValue(child)
				self.constant[name] = value

			elif child.localName == "quantity":
				name = child.attributes["name"].value
				value = self._parseValue(child)
				self.constant[name] = value

			elif child.localName == "expression":
				name, value = self._parseExpression(child)
				self.constant[name] = value

			elif child.localName == "position":
				self._parsePosition(child)

			elif child.localName == "rotation":
				self._parseRotation(child)

			elif child.localName == "variable":
				name, value = self._parseNameValue(child)
				self.variable[name] = value

			#elif child.localName == "scale":
			#	pass
			#elif child.localName == "matrix":
			#	pass
			else:
				log.rerror(f"define: '{child.localName}' is not implemented")

	#-----------------------------------------------------------------------
	# Parse materials
	#-----------------------------------------------------------------------
	def _parseMaterials(self, node):
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "define":
				self._parseDefinitions(child)
			elif child.localName == "element":
				self._parseElement(child)
			elif child.localName == "isotope":
				self._parseIsotope(child)
			elif child.localName == "material":
				self._parseMaterial(child)
			else:
				log.rerror(f"material: '{child.localName}' is not implemented")

	#-----------------------------------------------------------------------
	def _parseIsotope(self, node):
		isotope = {}
		isotope["name"] = name = str(node.attributes["name"].value)
		isotope["Z"] = node.attributes["Z"].value
		isotope["N"] = node.attributes["N"].value
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "atom":
				isotope["atom"] = self._parseValue(child)
		self.isotope[name] = isotope

	#-----------------------------------------------------------------------
	# Parse an element
	#-----------------------------------------------------------------------
	def _parseElement(self, node):
		element = {}
		element["name"] = name = str(node.attributes["name"].value)
		try:
			element["formula"] = str(node.attributes["formula"].value)
		except KeyError:
			element["formula"] = ""
		try:
			element["Z"] = int(float(node.attributes["Z"].value))
		except KeyError:
			element["Z"] = ""

		fraction = []

		for child in node.childNodes:
			if child.localName is None:
				continue

			if child.localName == "atom":
				if "mass" in element:
					raise GdmlException("atom and mass in element")
				# XXX FIXME WRONG to be verified!!!
				element["mass"] = self._parseValue(child)

			elif child.localName == "fraction":
				try:
					n = eval(child.attributes["n"].value)
				except KeyError:
					log.rerror(f"Gdml: no 'n' attribute in element {name}")
					return None
				try:
					isotope = str(child.attributes["ref"].value)
				except KeyError:
					continue
				fraction.append((n,isotope))

			else:
				log.rerror(f"element {name}: '{child.localName}' is not implemented")
		element["fraction"] = fraction
		self.element[name]  = element
		return element

	#-----------------------------------------------------------------------
	# Parse material
	#-----------------------------------------------------------------------
	def _parseMaterial(self, node):
		material = {}
		material["name"] = name = str(node.attributes["name"].value)
		try: material["formula"] = str(node.attributes["formula"].value)
		except KeyError: pass
		try: material["Z"] = int(float(node.attributes["Z"].value))
		except KeyError: material["Z"] = ""
		material["mass"] = ""

		fraction = []

		mixt = "weight"	# can be "atom", "weight", "volume"
		for child in node.childNodes:
			if child.localName is None:
				continue

			if child.localName == "D":
				material["D"] = self._parseValue(child)

			elif child.localName == "atom":
				material["mass"] = self._parseValue(child)

			elif child.localName == "fraction":
				mixt = "weight"
				try:
					n = eval(child.attributes["n"].value)
				except KeyError:
					log.rerror(f"Gdml no 'n' attribute in material/fraction {name}")
				try: isotope = str(child.attributes["ref"].value)
				except KeyError: continue
				fraction.append((n,isotope))

			elif child.localName == "composite":
				mixt = "atom"
				try:
					n = eval(child.attributes["n"].value)
				except KeyError:
					log.rerror(f"no 'n' attribute in material/composite {name}")
				try: isotope = str(child.attributes["ref"].value)
				except KeyError: continue
				fraction.append((n,isotope))

			# FIXME volume?

			elif child.localName == "MEE":
				material["MEE"] = self._parseValue(child)

			elif child.localName == "P":
				material["P"] = self._parseValue(child)

			elif child.localName == "T":
				material["T"] = self._parseValue(child)

			else:
				log.rerror(f"material {name}: '{child.localName}' is not implemented")

		material["mixt"]     = mixt
		material["fraction"] = fraction
		self.material[name] = material

	#-----------------------------------------------------------------------
	# Parse solids
	#-----------------------------------------------------------------------
	def _parseSolids(self, node):
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "define":
				self._parseDefinitions(child)
			elif child.localName == "xtru":
				self._parseXtru(child)
			elif child.localName in ("union", "intersection", "subtraction"):
				self._parseBoolean(child)
			elif child.localName in ("tessellated", "tet"):
				log.rerror(f"solids {child.attributes['name'].value}:"
					f"'{child.localName}' is not implemented.\n")
			else:
				self._parseSolid(child)

	#-----------------------------------------------------------------------
	# Parse all parameters of solid and add it to dictionary
	#-----------------------------------------------------------------------
	def _parseSolid(self, node):
		solid = {}
		solid["name"] = name = str(node.attributes["name"].value)
		solid["type"] = node.localName
		try:
			aunit = Gdml.unit(node.attributes["aunit"].value)
		except KeyError:
			aunit = Gdml.UNITS["rad"]	# default unit
		try:
			lunit = Gdml.unit(node.attributes["lunit"].value)
		except KeyError:
			lunit = Gdml.UNITS["mm"]	# default unit

		# Convert to standard units (cm, radians,...)
		for a,val in node.attributes.items():
			al = a.lower()
			if a in ("name","aunit","lunit"):
				# units ignore
				continue
			if a in ("numsides",):
				# no unit applied
				pass
			elif "phi" in al or "theta" in al or "alph" in al or "angle" in al:
				# angles
				val = float(bmath.format(self.eval(val, aunit),15))
			else:
				# distance
				val = float(bmath.format(self.eval(val, lunit),15))
			solid[a] = val

		# Additional information e.g. zplane
		zplanes = []
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "zplane":
				zplane = {}
				for a,val in child.attributes.items():
					if a in ("name","aunit","lunit"):
						continue
					zplane[a] = self.eval(val, lunit)
				zplanes.append(zplane)

		if zplanes: solid["zplane"] = zplanes

		self.solids[name] = solid

	#-----------------------------------------------------------------------
	# Parse extruded objects
	#-----------------------------------------------------------------------
	def _parseXtru(self, node):
		solid = {}
		solid["name"] = name = str(node.attributes["name"].value)
		solid["type"] = node.localName
		try:
			lunit = Gdml.unit(node.attributes["lunit"].value)
		except KeyError:
			lunit = 1.0	# mm default unit

		polygon = Polygon()
		solid["sections"] = sections = []

		for child in node.childNodes:
			if child.localName is None:
				continue

			if child.localName == "twoDimVertex":
				try: x = self.eval(child.attributes["x"].value, lunit)
				except KeyError: x = 0.0
				try: y = self.eval(child.attributes["y"].value, lunit)
				except KeyError: y = 0.0
				polygon.add(bmath.Vector(x,y,0.0))

			elif child.localName == "section":
				#try: order = self.eval(child.attributes["zOrder"].value)
				#except KeyError: order = 0.0
				try: z = self.eval(child.attributes["zPosition"].value, lunit)
				except KeyError: z = 0.0
				try: dx = self.eval(child.attributes["xOffset"].value, lunit)
				except KeyError: dx = 0.0
				try: dy = self.eval(child.attributes["yOffset"].value, lunit)
				except KeyError: dy = 0.0
				try: scale = self.eval(child.attributes["scalingFactor"].value)
				except KeyError: scale = 1.0
				sections.append((z, dx, dy, scale))
				if dx!=0.0 or dy!=0.0 or scale!=1.0:
					log.rerror(f"xtru {child.localName}: dx,dy,scale are not supported")

			else:
				log.rerror(f"xtru {node.localName}: '{child.localName}' is not implemented")

#		print ">>> xtru",name
#		f = open(f"{name}.xtru","w")
#		for x,y,z in polygon.P:
#				print>>f, x,y,z
#		print>>f, polygon.P[0][0], polygon.P[0][1], polygon.P[0][2]
#		print>>f
#		print>>f
		sections.sort(key=itemgetter(0))	# sort on z order
		solid["orientation"] = polygon.orientation()
		solid["polygon"]     = polygon.split2Convex()
		self.solids[name] = solid
#		for polygon in solid["polygon"]:
#			for x,y,z in polygon:
#				print>>f, x,y,z
#			print>>f, polygon[0][0], polygon[0][1], polygon[0][2]
#			print>>f
#		f.close()

	#-----------------------------------------------------------------------
	# Parse boolean solid operation
	#-----------------------------------------------------------------------
	def _parseBoolean(self, node):
		solid = {}
		solid["name"] = name = str(node.attributes["name"].value)
		solid["type"] = node.localName
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "first":
				solid["first"] = str(child.attributes["ref"].value)
			elif child.localName == "second":
				solid["second"] = str(child.attributes["ref"].value)

			elif child.localName == "positionref":
				solid["position"] = str(child.attributes["ref"].value)
			elif child.localName == "position":
				solid["position"] = self._parsePosition(child)
			elif child.localName == "rotationref":
				solid["rotation"] = str(child.attributes["ref"].value)
			elif child.localName == "rotation":
				solid["rotation"] = self._parseRotation(child)

			elif child.localName == "firstpositionref":
				solid["firstposition"] = str(child.attributes["ref"].value)
			elif child.localName == "firstposition":
				solid["firstposition"] = self._parsePosition(child)
			elif child.localName == "firstrotationref":
				solid["firstrotation"] = str(child.attributes["ref"].value)
			elif child.localName == "firstrotation":
				solid["firstrotation"] = self._parseRotation(child)
			else:
				log.rerror(f"Boolean {node.localName}: '{child.localName}' is not implemented")
		self.solids[name] = solid

	#-----------------------------------------------------------------------
	# Parse Structure
	#-----------------------------------------------------------------------
	def _parseStructure(self, node):
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "assembly":
				self._parseAssembly(child)
			elif child.localName == "volume":
				self._parseVolume(child)
			else:
				log.rerror(f"structure {child.attributes['name'].value}: "
					f"'{child.localName}' is not implemented.")

	#-----------------------------------------------------------------------
	# Parse volume and add it to dictionary
	#-----------------------------------------------------------------------
	def _parseVolume(self, node):
		volume = {}
		volume["name"] = name = str(node.attributes["name"].value)
		volume["physvol"] = []

		for child in node.childNodes:
			if child.localName is None: continue

			if child.localName == "materialref":
				volume["material"] = str(child.attributes["ref"].value)

			elif child.localName == "solidref":
				volume["solid"] = str(child.attributes["ref"].value)

			elif child.localName == "physvol":
				volume["physvol"].append(self._parsePhysvol(child))

			elif child.localName == "divisionvol":
				volume["divisionvol"] = self._parseDivisionvol(child)

			elif child.localName == "replicavol":
				volume["replicavol"] = self._parseReplicavol(child)

			elif child.localName == "paramvol":
				volume["paramvol"] = self._parseParamvol(child)

			else:
				log.rerror(f"volume {node.localName}: '{child.localName}' is not implemented")

#		# Count immediate childs
#		volume["nchlids"] = len(volume["physvol"])
#		# ... and all family
#		for x in volume["physvol"]: print x
#		volume["nfamily"] = sum([x["nfamily"] for x in volume["physvol"]])
#		for i in ("divisionvol","replicavol","paramvol"):
#			# FIXME
#			if i in volume:
#				k = volume[n].get("number",0)
#				volume["nchlids"] += k
#				volume["nfamily"] *= k
		self.volumes[name] = volume

	#-----------------------------------------------------------------------
	def _parsePhysvol(self, node):
		# Parse child nodes
		cname  = None
		cpos   = None
		crot   = None

		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "file":
				self.read(str(child.attributes["name"].value), False)
				cname = str(child.attributes["volname"].value)
			elif child.localName == "volumeref":
				cname = str(child.attributes["ref"].value)
			elif child.localName == "positionref":
				cpos = str(child.attributes["ref"].value)
			elif child.localName == "position":
				cpos = self._parsePosition(child)
			elif child.localName == "rotationref":
				crot = str(child.attributes["ref"].value)
			elif child.localName == "rotation":
				crot = self._parseRotation(child)
			else:
				log.rerror(f"{node.localName}: '{child.localName}' is not implemented")

		if node.hasAttribute("name"):
			name = str(node.getAttribute("name"))
		else:
			name = cname
		return PhysVol(name, cname, cpos, crot, None)

	#-----------------------------------------------------------------------
	def _parseReplicavol(self, node):
		try: number = int(eval(node.attributes["number"].value))
		except: number = 0
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "volumeref":
				cname = str(child.attributes["ref"].value)
			elif child.localName == "replicate_along_axis":
				axis,width,offset = self._replicateAlongAxis(child)
			else:
				log.rerror(f"{node.localName}: '{child.localName}' is not implemented")
		return (cname, axis, number, width, offset)

	#-----------------------------------------------------------------------
	# Parse replicate_along_axis of replicavol
	#-----------------------------------------------------------------------
	def _replicateAlongAxis(self, node):
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "direction":
				try: x = int(float(child.attributes["x"].value))
				except KeyError: x = 0
				try: y = int(float(child.attributes["y"].value))
				except KeyError: y = 0
				try: z = int(float(child.attributes["z"].value))
				except KeyError: z = 0
				if   x!=0: axis = "x"
				elif y!=0: axis = "y"
				elif z!=0: axis = "z"
				else:
					log.rerror("replicate_along_axis: axis not found")
					axis = "p"
			elif child.localName == "width":
				width = self._parseValue(child)
			elif child.localName == "offset":
				offset = self._parseValue(child)
		return (axis, width, offset)

	#-----------------------------------------------------------------------
	# Parse a parametrized volume
	#-----------------------------------------------------------------------
	def _parseParamvol(self, node):
		try: ncopies = int(float(node.attributes["ncopies"].value))
		except KeyError: ncopies = 1

		params = []
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "volumeref":
				cname = str(child.attributes["ref"].value)
			elif child.localName == "parameterised_position_size":
				for child2 in child.childNodes:
					if child2.localName is None:
						continue
					if child2.localName == "parameters":
						params.append(self._parseParametrizedPositionSize(child2))
					else:
						log.rerror(f"{node.localName}.{child.localName}: '{child2.localName}' is not implemented")
			else:
				log.rerror(f"{node.localName}: '{child.localName}' is not implemented")

		if len(params) != ncopies:
			log.rerror(f"{node.localName}: ncopies={ncopies} do not match the number of parameters")

		return (cname, params)

	#-----------------------------------------------------------------------
	def _parseParametrizedPositionSize(self, node):
		param = {}

		try:
			aunit = Gdml.unit(node.attributes["aunit"].value)
		except KeyError:
			aunit = Gdml.UNITS["rad"]	# default unit
		try:
			lunit = Gdml.unit(node.attributes["lunit"].value)
		except KeyError:
			lunit = Gdml.UNITS["mm"]	# default unit

		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "position":
				for a,val in child.attributes.items():
					if a in ("aunit","lunit","name"):
						continue
					if a in ("startphi","deltaphi","starttheta","deltatheta"):
						val = self.eval(val, aunit)
					else:
						val = self.eval(val, lunit)
					param[f"pos.{a}"] = val
			else:
				for a,val in child.attributes.items():
					if a in ("aunit","lunit"):
						continue
					if a in ("startphi","deltaphi","starttheta","deltatheta"):
						val = self.eval(val, aunit)
					else:
						val = self.eval(val, lunit)
					param[a] = val
		return param

	#-----------------------------------------------------------------------
	# Parse a division volume and convert it to similar parametrized volumes
	#-----------------------------------------------------------------------
	def _parseDivisionvol(self, node):
		try:
			u = Gdml.unit(node.attributes["unit"].value)
		except KeyError:
			u = 1.0

		try:    axis = node.attributes["axis"].value[1].lower()
		except: axis = "z"
		try:    number = int(eval(node.attributes["number"].value))
		except: number = 0
		try:    width = eval(node.attributes["width"].value, u)
		except: width = 0.0
		try:    offset = eval(node.attributes["offset"].value, u)
		except: offset = 0.0
		for child in node.childNodes:
			if child.localName is None:
				continue
			if child.localName == "volumeref":
				cname = str(child.attributes["ref"].value)
			else:
				log.rerror(f"{node.localName}: '{child.localName}' is not implemented")
		return (cname,axis,number,width,offset)

	#-----------------------------------------------------------------------
	def _parseAssembly(self, node):
		volume = {}
		volume["name"] = name = str(node.attributes["name"].value)
		volume["assembly"] = []

		for child in node.childNodes:
			if child.localName is None: continue

			if child.localName == "physvol":
				volume["assembly"].append(self._parsePhysvol(child))

			else:
				log.rerror(f"{node.localName} {name}: '{child.localName}' is not implemented")
		self.volumes[name] = volume

	#-----------------------------------------------------------------------
	# Parse Setup
	#-----------------------------------------------------------------------
	def _parseSetup(self, xml):
		for p in xml.getElementsByTagName("setup"):
			for child in p.childNodes:
				if child.localName is None:
					continue
				if child.localName == "world":
					self.world = str(child.attributes["ref"].value)
				else:
					log.rerror(f"setup: '{child.localName}' is not implemented")

#===============================================================================
# Translator class from Gdml to Moira
#===============================================================================
class Gdml2Moira:
	def __init__(self, gdml, inp=None):
		# FIXME we should move it to translator
		self.gdml  = gdml
		self.input = Moira() if inp is None else inp

	#-----------------------------------------------------------------------
	def convert(self):
		self._matNames = {}
		self.addMaterials()

		self._rotmNames = {}
		self.addRotations()

		self._solidNames = {}
		comment = "\nSolids"
		for solid in self.gdml.solids.values():
			self.addSolid(solid, comment)
			comment = ""

		self._volumeNames = {}
		self.addVolume(self.gdml.world, "\nVolumes")

		self.input.renumber()
		return self.input

	#-----------------------------------------------------------------------
	def add(self, card):
		for i,w in enumerate(card._what):
			if isinstance(w,float):
				card._what[i] = bmath.format(w)
		self.input.add(card, renumber=False)
		return card

	#-----------------------------------------------------------------------
	def addMaterials(self):
		# Write isotopes
		comment = "\nIsotopes"
		for name,isotope in self.gdml.isotope.items():
			uname = Gdml.uniqueName(name, self._matNames, "MAT")
			self.add(MoiraCard("ISOT",
					[uname,
					 isotope["Z"],
					 isotope["N"],
					 isotope["atom"]], comment))
			comment = ""

		# Write elements
		comment = "\nElements"
		for name,element in self.gdml.element.items():
			uname = Gdml.uniqueName(name, self._matNames, "MAT")
			fraction = element.get("fraction",None)
			if not fraction:
				self.add(MoiraCard("ELEM",
						[uname,
						 element.get("formula",""),
						 element.get("Z",""),
						 element.get("mass","")], comment))
			else:
				m = Gdml.SYMBOL.match(fraction[0][1])
				if m:
					symbol = m.group(1)
				else:
					symbol = uname
				what = [uname, symbol, len(fraction)]
				for f,n in fraction:
					un = self._matNames[n]
					what.append(un)
					what.append(f)
				self.add(MoiraCard("ELEM_FROM_ISOT", what, comment))
			comment = ""
#			whats = [name8]
#			for n,iso in element["fraction"]:
#				whats.append(bmath.format(n))
#				whats.append(self._matNames[iso])
#			if len(whats)>1:
#				self.add(MoiraCard("COMPOUND",whats), pos, False)
#			pos += 1

		# Write materials
		comment = "\nMaterials"
		for name, mat in self.gdml.material.items():
			uname = Gdml.uniqueName(name, self._matNames, "MAT")
			Z  = mat.get("Z","")
			Am = mat.get("mass","")
			D  = mat.get("D","")
			if Z and Am:
				self.add(MoiraCard("MATE",
						[uname, Z, Am, D], comment))
			else:
				whats = [uname, D, len(mat["fraction"])]
				for n,iso in mat["fraction"]:
					try:
						# find component name
						m = self._matNames[iso]
					except KeyError:
						# if it is defined later then add it first
						m = Gdml.uniqueName(iso, self._matNames, "MAT")
					whats.append(m)
					whats.append(bmath.format(n))
				if mat["mixt"] == "weight":
					tag = "MIXT"
				elif mat["mixt"] == "atom":
					tag = "MIXT_BY_NATOMS"
				elif mat["mixt"] == "volume":
					tag = "MIXT_BY_VOLUME"
				self.add(MoiraCard(tag, whats, comment))
			comment = ""

#			P   = mat.get("P","")
#			mee = mat.get("MEE","")
#			if P or mee:
#				try: mee *= 1e9	# convert to eV
#				except: pass
#				inp.add(FlukaCard("MAT-PROP",
#						["", P, "", mee, name8]),
#						pos, False)

	#-----------------------------------------------------------------------
	# Add rotation matrices
	#-----------------------------------------------------------------------
	def addRotations(self):
		comment = "\nRotations"
		for name,rot in self.gdml.rotation.items():
			uname = Gdml.uniqueName(name, self._rotmNames, "R")
			rotv = rot.vector
			self.add(MoiraCard("ROTM",
					[ uname, rotv[0], rotv[1], rotv[2]],
					comment))
			comment = ""

	#-----------------------------------------------------------------------
	# Convert gdml solid to moira
	#-----------------------------------------------------------------------
	def addSolid(self, solid, comment):
		name = Gdml.uniqueName(solid["name"], self._solidNames, "S")
		# Parametrized solid, create a new dictionary
#		if param is not None:
#			solid = solid.copy()
#			solid.update(param)

		#lunit = Gdml.UNITS["mm"]
		#aunit = Gdml.UNITS["deg"]

		t = solid["type"]
		x = solid.get("x",0)/2
		y = solid.get("y",0)/2
		z = solid.get("z",0)/2
		sphi = math.degrees(solid.get("startphi", 0.0))
		dphi = math.degrees(solid.get("deltaphi", math.pi*2.))

		# FIXME possibly angles we need them in degrees!!!!

		card = None
		#if t == "arb8":
		#	log.rerror("solid: 'arb8' not implemented!!!")
		# FIXME change to elif once implemented
		if t == "box":
			card = MoiraCard("SOLID.BOX", [name, x, y, z])
		elif t == "cone":
			if sphi>0 or dphi<360:
				card = MoiraCard("SOLID.CONS",
						[name,
						 solid.get("rmin1",0),
						 solid.get("rmin2",0),
						 solid.get("rmax1",0),
						 solid.get("rmax2",0),
						 z,
						 sphi,
						 dphi])
			else:
				card = MoiraCard("SOLID.CONE",
						[name,
						 solid.get("rmin1",0),
						 solid.get("rmin2",0),
						 solid.get("rmax1",0),
						 solid.get("rmax2",0),
						 z])
#		elif t == "cutTube":
#			expr = self._writeFlukaCutTube(inp, solid, matrix, comment)
		elif t == "elcone":
			card = MoiraCard("SOLID.ELLIPTICAL_CONE",
					[name,
					 solid.get("dx",0),
					 solid.get("dy",0),
					 solid.get("zmax",0),
					 solid.get("zcut",0)])
		elif t == "eltube":
			card = MoiraCard("SOLID.ELLIPTICAL_TUBE",
					[name,
					 solid.get("dx",0),
					 solid.get("dy",0),
					 solid.get("dz",0)])
		elif t == "ellipsoid":
			card = MoiraCard("SOLID.ELLIPSOID",
					[name,
					 solid.get("ax",0),
					 solid.get("by",0),
					 solid.get("cz",0),
					 solid.get("zcut1",0),	# FIXME only if they exist!!!!
					 solid.get("zcut2",0)])
		elif t == "hype":
			card = MoiraCard("SOLID.HYPE",
					[name,
					 solid.get("rmin",0),
					 solid.get("rmax",0),
					 solid.get("inst",0),
					 solid.get("outst",0),
					 z])
		elif t == "orb":
			card = MoiraCard("SOLID.ORB", [name, solid["r"]])
		elif t == "para":
			card = MoiraCard("SOLID.PARA",
					[name,
					 x, y, z,
					 math.degrees(solid.get("alpha", 0)),
					 math.degrees(solid.get("theta", 0)),
					 math.degrees(solid.get("phi", 0))])
#		elif t == "polycone":
#			whats = [name,
#				 sphi,
#				 dphi,
#				 len(solid["zplane"])]
#			for zplane in solid["zplane"]:		# FIXME RZ
#				whats.append(zplane.get("rmax",0))
#				whats.append(zplane.get("z",0))
#			card = MoiraCard("SOLID.POLYCONE", whats)
		elif t == "sphere":
			card = MoiraCard("SOLID.SPHERE",
					[name,
					 solid.get("rmin",0),
					 solid.get("rmax",0),
					 sphi,
					 dphi,
					 math.degrees(solid.get("starttheta",0)),
					 math.degrees(solid.get("deltatheta",math.pi))])
		elif t == "torus":
			card = MoiraCard("SOLID.TORUS",
					[name,
					 solid.get("rmin",0),
					 solid.get("rmax",0),
					 solid.get("rtor",0),
					 sphi,
					 dphi])
		elif t == "trap":
			card = MoiraCard("SOLID.TRAP",
					[name,
					 z,
					 math.degrees(solid.get("theta" ,0)),
					 math.degrees(solid.get("phi"   ,0)),
					 solid.get("y1",0)/2,
					 solid.get("x1",0)/2,
					 solid.get("x2",0)/2,
					 math.degrees(solid.get("alpha1",0)),
					 solid.get("y2",0)/2,
					 solid.get("x3",0)/2,
					 solid.get("x4",0)/2,
					 math.degrees(solid.get("alpha2",0))])
		elif t == "trd":
			card = MoiraCard("SOLID.TRD",
					[name,
					 solid.get("x1",0)/2,
					 solid.get("x2",0)/2,
					 solid.get("y1",0)/2,
					 solid.get("y2",0)/2,
					 z])
		elif t == "tube":
			if sphi>0 or dphi<360:
				card = MoiraCard("SOLID.TUBS",
						[name,
						solid.get("rmin",0),
						solid.get("rmax",0),
						z/2,
						sphi,
						dphi])
			else:
				card = MoiraCard("SOLID.TUBE",
						[name,
						solid.get("rmin",0),
						solid.get("rmax",0),
						z])

		elif t == "twistedbox":
			card = MoiraCard("SOLID.TWISTEDBOX",
					[name,
					 math.degrees(solid.get("PhiTwist",0)),
					 x, y, z])
		elif t == "twistedtrd":
			card = MoiraCard("SOLID.TWISTEDTRD",
					[name,
					 solid.get("x1",0)/2,
					 solid.get("x2",0)/2,
					 solid.get("y1",0)/2,
					 solid.get("y2",0)/2,
					 z,
					 math.degrees(solid.get("PhiTwist",0))])
		elif t == "twistedtrap":
			card = MoiraCard("SOLID.TWISTEDTRAP",
					[name,
					 math.degrees(solid.get("PhiTwist",0)),
					 z,
					 math.degrees(solid.get("Theta",0)),
					 math.degrees(solid.get("Phi",0)),
					 solid.get("y1",0)/2,
					 solid.get("x1",0)/2,
					 solid.get("x2",0)/2,
					 solid.get("y2",0)/2,
					 solid.get("x3",0)/2,
					 solid.get("x4",0)/2,
					 math.degrees(solid.get("Alph",0))])
		elif t == "twistedtubs":
			card = MoiraCard("SOLID.TWISTEDTUBS",
					[name,
					 math.degrees(solid.get("twistedangle",0)),
					 solid.get("endinnerrad",0),
					 solid.get("endouterrad",0),
					 solid.get("zlen",0)/2,
					 math.degrees(solid.get("phi",0))])

#		elif t == "xtru":
#			expr = self._writeFlukaXtru(solid, matrix, comment)
		elif t in ("union", "intersection", "subtraction"):
			if "firstposition" in solid or "firstrotation" in solid:
				log.rerror("Solid boolean operation with first"
					" position/rotation is not supported yet.")
			try:
				pos = self.gdml.position[solid["position"]].vector
			except KeyError:
				pos = bmath.Vector.O

			first  = self._solidNames[solid["first"]]
			second = self._solidNames[solid["second"]]
			card = MoiraCard(f"SOLID.{t.upper()}",
					[name,
					first, second,
					self._rotmNames.get(solid.get("rotation"),""),
					pos[0], pos[1], pos[2]])
		else:
			log.rerror(f"solid: '{t}' is not implemented")
		if card:
			if comment: card.setComment(comment)
			self.add(card)

	#-----------------------------------------------------------------------
	# Convert gdml volume to moira
	#-----------------------------------------------------------------------
	def addVolume(self, name, comment=""):
		volume = self.gdml.volumes[name]
		vname = Gdml.uniqueName(volume["name"], self._volumeNames, "V")
		err = False

		try:
			mat = volume["material"]
		except KeyError:
			err = True
			mat = "Vacuum"

		try:
			mat = self._matNames[mat]
		except KeyError:
			err = True
		try:
			sname = self._solidNames[volume["solid"]]
		except KeyError:
			sname = volume["solid"]
			err = True

		self.add(MoiraCard("VOLU", [ vname, sname, mat], comment))
		for child in volume["physvol"]:
			cname = self.addVolume(child.cname)
			try:
				pos = self.gdml.position[child.position].vector
			except KeyError:
				pos = bmath.Vector.O
			card = MoiraCard("PLACE",
					[cname,
					 1,
					 vname,
					 self._rotmNames.get(child.rotation,""),
					 pos[0], pos[1], pos[2]])
			if err: card.setEnable(False)
			self.add(card)
		return vname

#===============================================================================
# Gdml Importer/Exporter  to fluka
# FIXME: XXX OLD OLD XXX
#        load with the new Gdml() class before and perform here only
#        the conversion
#===============================================================================
class Gdml2Fluka:
	def __init__(self, gdml, inp=None):
		# FIXME we should move it to translator
		self.gdml  = gdml
		self.input = Fluka() if inp is None else inp
		log.repeat_clear()

	#-----------------------------------------------------------------------
	def convert(self):
		if not self.input.cardlist: self.input.minimumInput()

		# Initialize dictionaries of used names
		self._matNames    = {}
		self._bodyNames   = {}
		self._regionNames = {}
		self.bodyPos   = self.input["END"][0].pos()-1	# Find first end position
		self.regionPos = self.input["END"][1].pos()-1
		self.matPos    = self.input["GEOEND"][0].pos()+1

		# Write first the materials
		self.addMaterials()
		# Start with world
		world = self.addVolume(self.gdml.world, self.gdml.world, 0)
		self.addBlackVolume(world)
		self.input.renumber()

		return self.input

	#-------------------------------------------------------------------------------
	@staticmethod
	def uniqueName(name, dictionary, prefix):
		name = name.translate(Gdml.TRANS)
		pat = Gdml.NAMEHEX.match(name)
		if pat is not None:
			name8 = oname8 = pat.group(1)[:8]
		else:
			name8 = oname8 = name[:8]
		if not Card.NAMEPAT.match(name8):
			name8 = prefix
		i = 0

		while name8 in dictionary:
			i += 1
			si = str(i)
			if len(oname8) + len(si) <= 8:
				name8 = oname8 + si
			else:
				name8 = oname8[:8-len(si)]+si
		dictionary[name8] = name
		if name != name8:
			dictionary[name] = name8
		return name8

	#-----------------------------------------------------------------------
	def add(self, card, pos=None):
		for i,w in enumerate(card._what):
			if isinstance(w,float):
				card._what[i] = bmath.format(w)
		self.input.add(card, pos, renumber=False)
		return card

	#-----------------------------------------------------------------------
	# Write Materials in FLUKA format
	#-----------------------------------------------------------------------
	def addMaterials(self):
		# Write isotopes
		for name,isotope in self.gdml.isotope.items():
			name8 = Gdml2Fluka.uniqueName(name, self._matNames, "MAT")
			Z = int(float(isotope["Z"]))
			A = int(float(isotope["N"]))
			self.add(FlukaCard("MATERIAL",
					[name8, Z, "", 1.0, "", "", A]),
				self.matPos)
			self.matPos += 1

		# Write elements
		for name,element in self.gdml.element.items():
			name8 = Gdml2Fluka.uniqueName(name, self._matNames, "MAT")
			self.add(FlukaCard("MATERIAL",
					[name8, element.get("Z",""),
						element.get("mass",""),
						1.0],
					element.get("formula","")),
				self.matPos)
			self.matPos += 1

			whats = [name8]
			for n,iso in element["fraction"]:
				whats.append(bmath.format(n))
				whats.append(self._matNames[iso])
			if len(whats)>1:
				self.add(FlukaCard("COMPOUND",whats), self.matPos)
			self.matPos += 1

		# Write materials
		for name,mat in self.gdml.material.items():
			name8 = Gdml2Fluka.uniqueName(mat["name"], self._matNames, "MAT")
			self.add(FlukaCard("MATERIAL",
					[name8, mat.get("Z",""),
						mat.get("mass",""),
						mat.get("D","")]),
				self.matPos)
			self.matPos += 1

			P   = mat.get("P","")
			mee = mat.get("MEE","")
			if P or mee:
				try: mee *= 1e9	# convert to eV
				except: pass
				self.add(FlukaCard("MAT-PROP",
						["", P, "", mee, name8]),
					self.matPos)
				self.matPos += 1

			whats = [name8]
			for n,iso in mat["fraction"]:
				whats.append(bmath.format(-n))
				try:
					# find component name
					m = self._matNames[iso]
				except KeyError:
					# if it is defined later then add it first
					m = Gdml2Fluka.uniqueName(iso, self._matNames, "MAT")
				whats.append(m)
			card = self.add(FlukaCard("COMPOUND",whats), self.matPos)
			card.padWhats()
			self.matPos += 1

	#-----------------------------------------------------------------------
	# Add volume as FLUKA region
	#-----------------------------------------------------------------------
	def addVolume(self, name, vname, depth, matrix=None, path="", param=None):
		if Global.verbose:
			Global.dump(f"*-* writeFlukaVolume (depth={depth}) {name} {param}")
		try:
			volume = self.gdml.volumes[vname]
		except KeyError:
			log.rerror(f"Volume {vname} not found")
			return ""

		newpath = path + "/" + vname

		# Write necessary bodies
		try:
			mother = self.addSolid(volume["solid"], matrix, newpath, param)
			if mother == "": return ""
		except KeyError:
			return ""

		# Create division volumes if any
		if "replicavol" in volume:
			self.replicavol2physvol(vname, volume)
		elif "divisionvol" in volume:
			self.divisionvol2physvol(vname, volume)
		elif "paramvol" in volume:
			self.paramvol2physvol(vname, volume)

		# Write child bodies
		child = ""
		if "assembly" in volume:
			# Write assembly (assuming as world volume)
			# FIXME not very intelligent, maybe merge with the one below
			for aname, acname, apos, arot, cparam in volume["assembly"]:
				amatrix = self.transformMatrix(matrix, apos, arot)
				apath = newpath + "/" + acname
				m = self.addVolume(aname, acname, depth+1,
						amatrix, apath, cparam)
				if m: child += f"-{self.parenthesis(m)} "
		else:
			for pname, cname, cpos, crot, cparam in volume["physvol"]:
				cmatrix = self.transformMatrix(matrix, cpos, crot)
				cvol = self.gdml.volumes.get(cname,{})
				if "assembly" in cvol:
					# Inline expand assembly
					apath = newpath + "/" + cname
					for aname, acname, apos, arot, acparam in cvol["assembly"]:
						amatrix = self.transformMatrix(cmatrix, apos, arot)
						m = self.addVolume(aname, acname, depth+1,
								amatrix, apath, cparam)
						if m: child += f"-{self.parenthesis(m)} "
				else:
					# Insert physvol
					m = self.addVolume(pname, cname, depth+1,
							cmatrix, newpath, cparam)
					if m: child += f"-{self.parenthesis(m)} "

		# Write expression
		rname = self.addRegionName(name)
		if child!="" and "|" in mother: mother = f"({mother})"
		expr = FlukaCard.splitExpr(f"{mother} {child}",80)
		card = FlukaCard("REGION",[rname],f"{name} [{path}]\nref: {vname}",expr)
		self.add(card, self.regionPos)

		# Write material...
		try:
			mat = self._matNames[volume["material"]]
		except:
			mat = volume.get("material","VACUUM")
		self.add(FlukaCard("ASSIGNMA",["", mat, rname]),
				self.matPos)
		self.matPos += 1
		return mother

	#-----------------------------------------------------------------------
	# Write everything outside world as the blackhole
	#-----------------------------------------------------------------------
	def addBlackVolume(self, world):
		expr = f"-({world})"
		name = self.addRegionName("BLKBODY")
		card = FlukaCard("REGION", [name],
				f"Blackbody everything outside world: {name}",
				expr)
		self.add(card, self.regionPos)
		# Write material...
		self.add(FlukaCard("ASSIGNMA",["","BLCKHOLE",name]),
				self.matPos)
		self.matPos += 1

	#-----------------------------------------------------------------------
	# Add SOLID as FLUKA bodies
	# @return expression to include as region
	#-----------------------------------------------------------------------
	def addSolid(self, name, matrix=None, path="", param=None):
		try:
			solid = self.gdml.solids[name]
		except KeyError:
			log.rerror(f"Solid {name} not found")
			return ""

		# Parametrized solid, create a new dictionary
		if param is not None:
			solid = solid.copy()
			solid.update(param)

		t = solid["type"]
		comment = f"{t} {name} [{path}]"

		if t == "arb8":
			expr = self._addArb8(solid, matrix, comment)

		elif t == "box":
			expr = self._addBox(solid, matrix, comment)

		elif t == "cone":
			expr = self._addCone(solid, matrix, comment)

		elif t == "cutTube":
			expr = self._addCutTube(solid, matrix, comment)

		elif t == "eltube":
			expr = self._addElTube(solid, matrix, comment)

		elif t == "ellipsoid":
			expr = self._addEllipsoid(solid, matrix, comment)

		elif t == "orb":
			expr = self._addOrb(solid, matrix, comment)

		elif t == "polycone":
			expr = self._addPolycone(solid, matrix, comment)

		elif t == "polyhedra":
			expr = self._addPolyhedra(solid, matrix, comment)

		elif t == "sphere":
			expr = self._addSphere(solid, matrix, comment)

		elif t == "trd":
			expr = self._addTrd(solid, matrix, comment)

		elif t == "tube":
			expr = self._addTube(solid, matrix, comment)

		elif t == "xtru":
			expr = self._addXtru(solid, matrix, comment)

		elif t in ("union", "intersection", "subtraction"):
			expr = self._addBoolean(solid, matrix, comment)

		else:
			log.rerror(f"solid: '{t}' not implemented")
			return ""
		return expr.strip()

	#-----------------------------------------------------------------------
	# Transform matrix with rot,pos
	# WARNING: be careful on operations to return a NEW matrix and not
	#          to modify the original
	#-----------------------------------------------------------------------
	def transformMatrix(self, matrix, pos, rot, inverse=False):
		cmatrix = matrix

		if pos is not None:
			if isinstance(pos,str):
				m = self.gdml.position[pos].matrix
			else:
				m = pos
			if cmatrix is None:
				cmatrix = m.clone()
			else:
				cmatrix = cmatrix * m

		if rot is not None:
			if isinstance(rot,str):
				m = self.gdml.rotation[rot].matrix
			else:
				m = rot
			if inverse:
				m = m.inverse()
				if cmatrix is None:
					cmatrix = m
				else:
					cmatrix = cmatrix * m
			elif cmatrix is None:
				cmatrix = m.clone()
			else:
				cmatrix = cmatrix * m

		return cmatrix

	#-----------------------------------------------------------------------
	# Use or strip the parentheses of a sub-expression
	#-----------------------------------------------------------------------
	def parenthesis(self, expr):
		if expr.find(" ") >= 0:
			return f"({expr})"
		else:
			return expr[1:]

	#-----------------------------------------------------------------------
	# Convert replicavol to physvol
	#-----------------------------------------------------------------------
	def replicavol2physvol(self, name, volume):
		solid  = self.gdml.solids[volume["solid"]]
		t = solid["type"]

		cname, axis, number, width, offset = volume["replicavol"]

		def calcDimensions(low,high,number,width,offset):
			# Calculate dimensions
			low += offset
			if width>0.0:
				n = int((high-low)/width)
				if number == 0: number = n
			else:
				width = (high-low)/number
			return low,high,number,width

		if axis in ("x","y","z"):
			h = solid[axis]/2.0
			low,high,number,width = calcDimensions(-h,h,number,width,offset)
			pos = bmath.Vector()
			idx = ord(axis) - ord('x')
			for i in range(number):
				pos[idx] = low + (i+0.5)*width
				m = bmath.Matrix.translate(pos)
				volume["physvol"].append(PhysVol(name, cname, m, None, None))
		elif axis=="p":
			l = solid.get("startphi",0.0)
			h = l + solid.get("deltaphi",2*math.pi)
			low,high,number,width = calcDimensions(l,h,number,width,offset)
			for i in range(number):
				rot = low + (i+0.5)*width
				m = bmath.Matrix.rotZ(rot)
				volume["physvol"].append(PhysVol(name, cname, None, m, None))

		elif axis=="r":
			rmin = solid["rmin"]
			rmax = solid["rmax"]
			low,high,number,width = calcDimensions(rmin,rmax,number,width,offset)
			for i in range(number):
				param = {"rmin" : low + i*width,
					 "rmax" : low + (i+1)*width}
				volume["physvol"].append(PhysVol(name, cname, None, None, param.copy()))

		else:
			log.rerror(f"replicavol {name}: solid {solid} type={t} axis={axis} not implemented")

		del volume["replicavol"]	# delete not to reprocess next time

	#-----------------------------------------------------------------------
	# Convert paramvol to physvol
	#-----------------------------------------------------------------------
	def paramvol2physvol(self, name, volume):
		cname, params = volume["paramvol"]
		for p in params:
			tx = p.get("pos.x",0.0)
			ty = p.get("pos.y",0.0)
			tz = p.get("pos.z",0.0)
			if tx!=0.0 or ty!=0.0 or tz!=0.0:
				m = bmath.Matrix.translate(tx,ty,tz)
			else:
				m = None
			volume["physvol"].append((PhysVol(name, cname, m, None, p)))
		del volume["paramvol"]	# delete it so not to reprocess next time

	#-----------------------------------------------------------------------
	# Convert divisionvol to physvol
	#
	# List of volumes that support divisioning
	# G4Box		kXAxis, kYAxis, kZAxis
	# G4Tubs	kRho,   kPhi,   kZAxis
	# G4Cons	kRho,   kPhi,   kZAxis
	# G4Trd		kXAxis, kYAxis, kZAxis
	# G4Para	kXAxis, kYAxis, kZAxis
	# G4Polycone	kRho,   kPhi,   kZAxis	FIXME not implemented
	# G4Polyhedra	kRho,   kPhi,   kZAxis (*)
	#-----------------------------------------------------------------------
	def divisionvol2physvol(self, name, volume):
		solid  = self.gdml.solids[volume["solid"]]
		t = solid["type"]
		param = solid.copy()
		del param["name"]
		del param["type"]

		cname, axis, number, width, offset = volume["divisionvol"]
		if t in ("polycone","polyhedra"):
			log.rerror(f"divisionvol {name}: solid {solid} type={t} axis={axis} not implemented")

		def calcDimensions(low,high,number,width,offset):
			# Calculate dimensions
			low += offset
			if width>0.0:
				n = int((high-low)/width)
				if number == 0: number = n
			else:
				width = (high-low)/number
			return low,high,number,width

		if axis in ("x","y","z"):
			h = solid[axis]/2.0
			low,high,number,width = calcDimensions(-h,h,number,width,offset)
			param[axis] = width
			pos = bmath.Vector()
			idx = ord(axis) - ord('x')
			for i in range(number):
				pos[idx] = low + (i+0.5)*width
				m = bmath.Matrix.translate(pos)
				volume["physvol"].append((name, cname, m, None, param.copy()))
		elif axis=="p":
			l = solid.get("startphi",0.0)
			h = l + solid.get("deltaphi",2*math.pi)
			low,high,number,width = calcDimensions(l,h,number,width,offset)
			param["deltaphi"] = width
			for i in range(number):
				param["startphi"] = low + i*width
				volume["physvol"].append((name, cname, None, None, param.copy()))
		elif axis=="r":
			rmin = solid.get("rmin",0.0)
			rmax = solid.get("rmax",0.0)
			low,high,number,width = calcDimensions(rmin,rmax,number,width,offset)
			for i in range(number):
				param["rmin"] = low + i*width
				param["rmax"] = low + (i+1)*width
				volume["physvol"].append((name, cname, None, None, param.copy()))
		else:
			log.rerror("divisionvol %s: solid %s type=%s axis=%s not implemented"%(name,solid,t,axis))

		del volume["divisionvol"]	# delete not to reprocess next time

	#-----------------------------------------------------------------------
	# Add a body to input with comment and matrix transformation
	#-----------------------------------------------------------------------
	def _addBody(self, tag, name, what, matrix, comment=""):
		bname = self.addBodyName(name)
		what.insert(0, bname)
		card = FlukaCard(tag,what,comment)
		self.add(card, self.bodyPos)
		if matrix: self.input.transformBody(card, matrix)
		return bname

	#-----------------------------------------------------------------------
	def _addZCutPlanes(self, solid, matrix, comment=""):
		z = solid.get("z",0.0)
		# Low cutting planes
		lowX = solid.get("lowX",0.0)
		lowY = solid.get("lowY",0.0)
		lowZ = solid.get("lowZ",0.0)
		bname1 = self._addBody("PLA", solid["name"],
				[lowX, lowY, lowZ, 0.0, 0.0, -z], matrix, comment)
		# High cutting planes
		highX = solid.get("highX",0.0)
		highY = solid.get("highY",0.0)
		highZ = solid.get("highZ",0.0)
		bname2 = self._addBody("PLA",solid["name"],
				[highX, highY, highZ, 0.0, 0.0, z], matrix, comment)
		return f"+{bname1} +{bname2}"

	#-----------------------------------------------------------------------
	def _addPhiCutPlanes(self, solid, matrix, comment=""):
		# Cutting planes
		sphi = solid.get("startphi",0.0)
		dphi = solid.get("deltaphi",2*math.pi)
		if sphi>0.0 or dphi<2*math.pi:
			c = math.cos(sphi)
			s = math.sin(sphi)
			bname1 = self._addBody("PLA", solid["name"],
					[-s, c, 0.0, 0.0, 0.0, 0.0], matrix, comment)

			f = sphi+dphi
			c = math.cos(f)
			s = math.sin(f)
			bname2 = self._addBody("PLA", solid["name"],
					[-s, c, 0.0, 0.0, 0.0, 0.0],matrix, comment)
			return f"+{bname2} -{bname1}"
		else:
			return ""

	#-----------------------------------------------------------------------
	def _addArb8(self, solid, matrix, comment):
		# Find points
		p = []
		dz = solid["dz"]
		for i in range(1,9):
			x = solid.get(f"v{i}x",0.0)
			y = solid.get(f"v{i}y",0.0)
			if i<=4:
				z= -dz
			else:
				z = dz
			p.append(bmath.Vector(x,y,z))

		# Top plane
		same = 0
		for i in range(4,7):
			for j in range(i+1,8):
				if (p[i]-p[j]).length2() < Gdml.SMALL:
					same += 1
		if same<2:
			bname = self._addBody("XYP", solid["name"],
				[dz], matrix, comment)
			expr = f"+{bname}"
		else:
			expr = ""

		# Bottom plane
		same = 0
		for i in range(0,3):
			for j in range(i+1,4):
				if (p[i]-p[j]).length2() < Gdml.SMALL:
					same += 1

		if same<2:
			bname = self._addBody("XYP", solid["name"],
					[-dz], matrix, comment)
			expr += f"-{bname}"

		#---------------------------------------------------------------
		# Add a face with two planes abc and acd
		# FIXME order! + or - or I have to find runtime?
		#---------------------------------------------------------------
		def addFace(a,b,c,d,sign):
			x,y,z = p[a]

			n1 = (p[b]-p[a]) ^ (p[c]-p[a])
			n1.normalize()

			if abs(n1[0]) >= Gdml.SMALL or \
			   abs(n1[1]) >= Gdml.SMALL or \
			   abs(n1[2]) >= Gdml.SMALL:
				bname = self._addBody("PLA", solid["name"],
					[n1[0],n1[1],n1[2],x,y,z], matrix)
#				log.say(">>>",n1,x,y,z)
				e = f"{sign}{bname} "
			else:
				e = ""

			n2 = (p[c]-p[a]) ^ (p[d]-p[a])
			n2.normalize()

			if (abs(n2[0]) >= Gdml.SMALL or
			    abs(n2[1]) >= Gdml.SMALL or
			    abs(n2[2]) >= Gdml.SMALL) and \
			    n1*n2 < 0.999999:
				log.rwarning(f"twisted arb8 {solid['name']} is approximated ({n1*n2})")
				bname = self._addBody("PLA", solid["name"],
					[n2[0],n2[1],n2[2],x,y,z], matrix)
#				log.say(">>>",n2,x,y,z)
				e += f"{sign}{bname} "

			# Check with fitPlane
#			plane = bmath.fitPlane(p)
#			log.say("+++",plane)
			return e

		#---------------------------------------------------------------
		# Check sign of operation
		A = p[1]-p[0]
		B = p[3]-p[0]
		if abs(A*B) < Gdml.SMALL:
			A = p[5]-p[4]
			B = p[7]-p[4]

		if (A^B)[2] > 0.0:
			sign = "+"
		else:
			sign = "-"

		expr += addFace(0,1,5,4,sign)
		expr += addFace(1,2,6,5,sign)
		expr += addFace(2,3,7,6,sign)
		expr += addFace(3,0,4,7,sign)

		return expr

	#-----------------------------------------------------------------------
	def _addBox(self, solid, matrix, comment):
		return "+%s"%(self._addBody("RPP", solid["name"],
				[-solid["x"]/2.0, solid["x"]/2.0,
				 -solid["y"]/2.0, solid["y"]/2.0,
				 -solid["z"]/2.0, solid["z"]/2.0], matrix, comment))

	#-----------------------------------------------------------------------
	def _addCone(self, solid, matrix, comment):
		# outer cone
		bname = self._addBody("TRC", solid["name"],
				[0.0, 0.0, -solid["z"]/2.0,
				 0.0, 0.0,  solid["z"],
				 solid["rmax1"],solid["rmax2"]], matrix, comment)
		expr = f"+{bname} "

		if solid.get("rmin1",0.0)>0.0 or solid.get("rmin2",0.0)>0.0:
			# Inner cone
			bname = self._addBody("TRC", solid["name"],
					[0.0, 0.0, -solid["z"]/2.0,
					 0.0, 0.0,  solid["z"],
					 solid["rmin1"],solid["rmin2"]], matrix, comment)
			expr += f"-{bname} "
		expr += self._addPhiCutPlanes(solid, matrix)
		return expr

	#-----------------------------------------------------------------------
	def _addCutTube(self, solid, matrix, comment):
		# outer cylinder
		bname =self._addBody("ZCC", solid["name"],
				[0.0, 0.0, solid["rmax"]], matrix, comment)
		expr = f"+{bname} "

		if solid.get("rmin",0.0) > 0.0:
			# Inner cylinder
			bname = self._addBody("ZCC", solid["name"],
					[0.0, 0.0, solid["rmin"]], matrix, comment)
			expr += f"-{bname} "

		expr += self._addPhiCutPlanes(solid, matrix, comment)
		expr += self._addZCutPlanes(solid, matrix, comment)
		return expr

	#-----------------------------------------------------------------------
	def _addElTube(self, solid, matrix, comment):
		bname = self._addBody("REC", solid["name"],
				[0.0, 0.0, -solid["dz"]/2.0,
				 0.0, 0.0,  solid["dz"],
				 solid["dx"], 0.0, 0.0,
				 0.0, solid["dy"], 0.0], matrix, comment)
		return f"+{bname}"

	#-----------------------------------------------------------------------
	def _addEllipsoid(self, solid, matrix, comment):
		bname = self._addBody("QUA", solid["name"],
				[1.0/solid["ax"]**2, 1.0/solid["by"]**2, 1.0/solid["cz"]**2,
				 0.0, 0.0, 0.0,
				 0.0, 0.0, 0.0, -1.0],matrix,comment)
		expr = f"+{bname}"

		if "zcut1" in solid:
			bname = self._addBody("XYP", solid["name"],
				[solid["zcut1"]], matrix, comment)
			expr += f"-{bname}"

		if "zcut2" in solid:
			bname = self._addBody("XYP", solid["name"],
				[solid["zcut2"]], matrix, comment)
			expr += f"+{bname}"

		return expr

	#-----------------------------------------------------------------------
	def _addOrb(self, solid, matrix, comment):
		bname = self._addBody("SPH", solid["name"],
				[0.0, 0.0, 0.0, solid["r"]], matrix, comment)
		return f"+{bname}"

	#-----------------------------------------------------------------------
	def _addPolycone(self, solid, matrix, comment):
		# remember first z-section
		zplane = solid["zplane"][0]
		z    = zplane.get("z",0.0)	# z-location
		rmin = zplane.get("rmin",0.0)	# inner radius
		rmax = zplane.get("rmax",0.0)	# outer radius
		if rmin > rmax:
			log.rwarning(f"polycone {solid['name']} plane z={zplane['z']} "
					f"rmin>rmax  rmin={rmin} rmax={rmax}")
			rmin, rmax = rmax, rmin

		# First plane
		zname = self._addBody("XYP", solid["name"],
				[z], matrix, comment)
		expr = ""
		phicut = self._addPhiCutPlanes(solid, matrix, comment)

		for zplane in solid["zplane"][1:]:
			# write something if there is a change in z
			zn    = zplane.get("z",0.0)	# next z plane
			rnmin = zplane.get("rmin",0.0)	# next inner radius
			rnmax = zplane.get("rmax",0.0)	# next outer radius
			if rnmin > rnmax:
				log.rwarning(f"polycone {solid['name']} plane z={zn} "
						f"rmin>rmax  rmin={rnmin} rmax={rnmax}")
				rnmin, rnmax = rnmax, rnmin

			# WARNING polycone is the G4GenericPolycone which can decrease as well
			if zn < z:
				swap = True
				zn, z = z, zn
				rnmin, rmin = rmin, rnmin
				rnmax, rmax = rmax, rnmax
			else:
				swap = False

			if abs(zn-z) >= Gdml.SMALL:
				if expr: expr += " | "
				if swap:
					bname = self._addBody("XYP", solid["name"],
							[z], matrix, comment)
					expr += f"+{zname} -{bname}"
				else:
					bname = self._addBody("XYP", solid["name"],
							[zn], matrix, comment)
					expr += f"+{bname} -{zname}"
				zname = bname	# remember last z-plane

				# Check inner radius
				if abs(rnmin-rmin) < Gdml.SMALL:
					# Cylinder
					if rnmin > Gdml.SMALL:
						# with non-zero radius
						bname = self._addBody("ZCC", solid["name"],
								[0.,0.,rmin], matrix, comment)
						expr += f"-{bname} "
				else:
					# Cone
					Ra = rnmin
					Rb = rmin
					za = zn
					zb = z

					zc = (Rb*za-Ra*zb)/(Rb-Ra)	# zero point
					s  = (Rb-Ra)/(zb-za)		# slope
					s2 = s**2
					bname = self._addBody("QUA", solid["name"],
							[1., 1., -s2,
							 0., 0., 0.,
							 0., 0., 2.*s2*zc,
							 -s2*zc*zc], matrix, comment)
					expr += f"-{bname} "

				# Check inner radius
				if abs(rnmax-rmax) < Gdml.SMALL:
					# Cylinder
					bname = self._addBody("ZCC", solid["name"],
							[0.,0.,rmax], matrix, comment)
				else:
					# Cone
					Ra = rnmax
					Rb = rmax
					za = zn
					zb = z

					zc = (Rb*za-Ra*zb)/(Rb-Ra)	# zero point
					s  = (Rb-Ra)/(zb-za)		# slope
					s2 = s**2
					bname = self._addBody("QUA", solid["name"],
							[1., 1., -s2,
							 0., 0., 0.,
							 0., 0., 2.*s2*zc,
							 -s2*zc*zc], matrix, comment)
					#bname = self._addBody("TRC", solid["name"],
					#		[0., 0., zb,
					#		 0., 0., za-zb,
					#		 Rb, Ra], matrix, comment)
				expr += f"+{bname}{phicut}"

			# remember last z-section
			if not swap:
				z    = zn	# z-location
				rmin = rnmin	# inner radius
				rmax = rnmax	# outer radius
		return expr

	#-----------------------------------------------------------------------
	def _addPolyhedra(self, solid, matrix, comment):
		# remember first z-section
		zplane = solid["zplane"][0]
		z    = zplane.get("z",0.0)	# z-location
		rmin = zplane.get("rmin",0.0)	# inner radius
		rmax = zplane.get("rmax",0.0)	# outer radius
		if rmin > rmax:
			log.rwarning(f"polyhedra {solid['name']} plane z={zplane['z']} "
					f"rmin>rmax  rmin={rmin} rmax={rmax}")
			rmin, rmax = rmax, rmin

		try:
			numsides = int(solid["numsides"])
		except (ValueError, KeyError):
			log.rerror(f"polyhedra {solid['name']} no numsides specified")
			return ""
		try:
			dphi = math.pi*2 / float(numsides)
		except ZeroDivisionError:
			log.rerror(f"polyhedra {solid['name']} invalid numsides specified")
			return ""
		cosphi2 = math.cos(dphi/2)

		# First plane
		zname  = self._addBody("XYP", solid["name"], [z], matrix, comment)
		expr   = ""
		phicut = self._addPhiCutPlanes(solid, matrix, comment)

		rcut = []	# r-cutting planes, create only once

		for zplane in solid["zplane"][1:]:
			# write something if there is a change in z
			zn    = zplane.get("z",0.0)	# next z plane
			rnmin = zplane.get("rmin",0.0)	# next inner radius
			rnmax = zplane.get("rmax",0.0)	# next outer radius
			if rnmin > rnmax:
				log.rwarning(f"polyhedra {solid['name']} plane z={zn} "
						f"rmin>rmax  rmin={rnmin} rmax={rnmax}")
				rnmin, rnmax = rnmax, rnmin

			# WARNING polyhedra is the G4Polyhedra FIXME can it decrease?
			if zn < z:
				log.rwarning(f"polyhedra {solid['name']} wrong z slice order z={z}, znext={zn}")
				return ""

			znname = self._addBody("XYP", solid["name"], [zn], matrix, comment)

			if rmin<=0.0:
				if expr: expr += " | "
				expr += f"+{znname} -{zname}"

			# reduce to phicut if any
			for i in range(numsides):
				nphi = i * dphi
				phi  = nphi - dphi/2
				x = math.cos(phi)
				y = math.sin(phi)

				nx = math.cos(nphi)	# not normalized yet to nz
				ny = math.sin(nphi)

				fz = math.atan2((rnmax-rmax)*cosphi2, zn-z) + math.pi/2
				cfz = math.cos(fz)
				sfz = math.sin(fz)

				# Create r-cutting planes if needed the first time
				if rmin>0.0 and not rcut:
					for j in range(numsides):
						f = j*dphi - dphi/2 + math.pi/2
						rcut.append(self._addBody("PLA", solid["name"],
								[math.cos(f), math.sin(f), 0, 0, 0, 0],
								matrix, comment))

				# outside plane
				bdout = self._addBody("PLA", solid["name"],
						[sfz*nx, sfz*ny, cfz, rmax*x, rmax*y, z],
						matrix, comment)

				if rmin>0.0:
					# inside plane
					fz = math.atan2((rnmin-rmin)*cosphi2, zn-z) + math.pi/2
					cfz = math.cos(fz)
					sfz = math.sin(fz)
					bdin = self._addBody("PLA", solid["name"],
							[sfz*nx, sfz*ny, cfz, rmin*x, rmin*y, z],
							matrix, comment)

					j = (i+1)%numsides
					if expr: expr += " | "
					expr += f"+{znname} -{zname}"
					expr += f"+{bdout} -{bdin} -{rcut[i]} +{rcut[j]}"
#					expr += "+%s%s"%(bname,phicut)
				else:
					expr += f"+{bdout} "
#			expr += "+%s%s"%(bname,phicut)

			# remember last parameters
			z = zn
			rmin = rnmin
			rmax = rnmax
			zname = znname	# remember last z-plane
		return expr

	#-----------------------------------------------------------------------
	def _addSphere(self, solid, matrix, comment):
		bname = self._addBody("SPH", solid["name"],
				[0.0, 0.0, 0.0, solid["rmax"]], matrix, comment)
		expr = f"+{bname}"

		if "rmin" in solid and solid["rmin"]>0.0:
			bname = self._addBody("SPH", solid["name"],
					[0.0, 0.0, 0.0, solid["rmin"]], matrix, comment)
			expr += f"-{bname}"

		expr += self._addPhiCutPlanes(solid, matrix, comment)

		# Theta Cutting planes
		stheta = solid.get("starttheta",0.0)
		dtheta = solid.get("deltatheta",math.pi)
		if stheta > 0.0 or dtheta<math.pi:
			c = math.cos(stheta)
			s = math.sin(stheta)
			bname1 = self._addBody("PLA", solid["name"],
					[-c, s, 0.0, 0.0, 0.0, 0.0], matrix, comment)

			f = stheta+dtheta
			c = math.cos(f)
			s = math.sin(f)
			bname2 = self._addBody("PLA", solid["name"],
					[-c, s, 0.0, 0.0, 0.0, 0.0], matrix, comment)
			expr += f"+{bname2} -{bname1}"
		return expr

	#-----------------------------------------------------------------------
	def _addTube(self, solid, matrix, comment):
		rmin = solid.get("rmin",0.0)
		rmax = solid.get("rmax",0.0)
		if rmin > rmax:
			log.rerror(f"tube {solid['name']} rmin>rmax  rmin={rmin} rmax={rmax}")
			rmin, rmax = rmax, rmin

		# outer cylinder
		bname = self._addBody("RCC", solid["name"],
				[0.0, 0.0, -solid["z"]/2.0,
				 0.0, 0.0,  solid["z"],
				 rmax], matrix, comment)
		expr = f"+{bname} "

		if rmin > 0.0:
			# Inner cylinder
			bname = self._addBody("ZCC", solid["name"],
					[0.0, 0.0, rmin], matrix, comment)
			expr += f"-{bname} "

		expr += self._addPhiCutPlanes(solid, matrix, comment)
		return expr

	#-----------------------------------------------------------------------
	def _addTrd(self, solid, matrix, comment):
		z  = solid["z"] /2
		x1 = solid["x1"]/2
		y1 = solid["y1"]/2
		x2 = solid["x2"]/2
		y2 = solid["y2"]/2

		# cutting z-planes
		bname = self._addBody("XYP", solid["name"],
				[ z ], matrix, comment)
		expr = f"+{bname} "

		bname = self._addBody("XYP", solid["name"],
				[-z], matrix, comment)
		expr += f"-{bname} "

		# -X
		bname = self._addBody("PLA", solid["name"],
				[-2.0*z, 0., (x1-x2), -x1, -y1, -z],
				matrix, comment)
		expr += f"+{bname} "

		# -Y
		bname = self._addBody("PLA", solid["name"],
				[0., -2.0*z, (y1-y2), -x1, -y1, -z],
				matrix, comment)
		expr += f"+{bname} "

		# +X
		bname = self._addBody("PLA", solid["name"],
				[2.0*z, 0., (x1-x2), x1, y1, -z],
				matrix, comment)
		expr += f"+{bname} "

		# +Y
		bname = self._addBody("PLA", solid["name"],
				[0., 2.0*z, (y1-y2), x1, y1, -z],
				matrix, comment)
		expr += f"+{bname} "

		return expr

	#-----------------------------------------------------------------------
	# FIXME VERY PRIMITIVE!!!!
	#-----------------------------------------------------------------------
	def _addXtru(self, solid, matrix, comment):
		# Write Z plane
		sign = "-"
		zexpr = ""
		for z,dx,dy,scale in solid["sections"]:
			bname = self._addBody("XYP", solid["name"],
				[z], matrix, comment)
			zexpr += f"{sign}{bname} "
			sign = "+"

		# Write polygones
		Z = bmath.Vector(0.0,0.0,solid["orientation"])
		bname = None
		expr = ""
		lastPoint = None
		for P in solid["polygon"]:
			zone = zexpr
			for i,Pi in enumerate(P):
				j = i+1
				if j>=len(P): j=0

				# Check last polygon-plane with current one
				if lastPoint and (lastPoint[0]-P[j]).length()<Gdml.SMALL:
					zone += f"-{lastPoint[1]} "
					lastPoint = None
					continue

				N = (P[j]-Pi) ^ Z
				N.normalize()

				bname = self._addBody("PLA", solid["name"],
					[N[0],N[1],N[2], Pi[0],Pi[1],Pi[2]], matrix, comment)
				last = (Pi,bname)
				zone += f"+{bname} "
			lastPoint = last

			if expr:
				expr += f" | {zone}"
			else:
				expr = zone

		return expr

	#-----------------------------------------------------------------------
	def _addBoolean(self, solid, matrix, comment):
		comment = f"{solid['name']} [{comment}]"

		# first body
		if "firstposition" in solid or "firstrotation" in solid:
			cmatrix = self.transformMatrix(matrix,
						solid.get("firstposition"),
						solid.get("firstrotation"),
						True)
		else:
			cmatrix = matrix
		first = self.addSolid(solid["first"], cmatrix, comment+" Boolean: First")

		# second body
		cmatrix = self.transformMatrix(matrix,
					solid.get("position"),
					solid.get("rotation"),
					True)
		second = self.addSolid(solid["second"], cmatrix, comment+" Boolean: Second")

		if solid["type"] == "union":
			return f"{self.parenthesis(first)} | {self.parenthesis(second)}"
		elif solid["type"] == "intersection":
			return f"{self.parenthesis(first)} + {self.parenthesis(second)}"
		elif solid["type"] == "subtraction":
			return f"{self.parenthesis(first)} - {self.parenthesis(second)}"

	#-----------------------------------------------------------------------
	# Create a unique fluka name out of a solid name
	#-----------------------------------------------------------------------
	def addBodyName(self, name=""):
		self.bodyPos   += 1
		self.regionPos += 1
		self.matPos    += 1
		name = Gdml2Fluka.uniqueName(name, self._bodyNames, "B")
		return name

	#-----------------------------------------------------------------------
	# Return the body name from a gdml solid
	#-----------------------------------------------------------------------
	def bodyName(self, name):
		return self._bodyNames[name]

	#-----------------------------------------------------------------------
	# Create a unique fluka name out of a volume name
	#-----------------------------------------------------------------------
	def addRegionName(self, name):
		self.regionPos += 1
		self.matPos    += 1
		name = Gdml2Fluka.uniqueName(name, self._regionNames, "R")
		return name

#===============================================================================
#if __name__ == "__main__":
#	import sys
#	Input.init()
#	Global.verbose = True
#	gdml = Gdml(sys.argv[1])
#
#	inp = Input.Input()
#	gdml.toFluka(inp)
#	inp.write(sys.argv[2])
#
#	sys.exit(0)
