# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Oct-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import tempfile
from collections import namedtuple, OrderedDict

import bmath
import Global

from deck    import Input, Card, CardInfo
from .card   import MoiraCard
from .config import MoiraConfig

# FIXME should be in a database or better get dynamically from G4
G4Mat = namedtuple("G4Mat", ["density", "Z", "I"])

#===============================================================================
# Moira Input class
#===============================================================================
class Moira(Input):
	MODE    = MoiraCard.MODE

	EXTMAC  = ".macro"		# macro file extension
	config  = MoiraConfig()		# configuration information

	defaultMaterials       = OrderedDict()
	defaultElements        = OrderedDict()
	FLUKA_MAT_2_MOIRA      = {}
	FLUKA_ELE_2_MOIRA      = {}
	FLUKA_PARTICLE_2_MOIRA = {}
	MASS_PARTICLES_MOIRA   = {}

	# ----------------------------------------------------------------------
	def __init__(self, filename=None):
		super().__init__()
		self.mode       = Moira.MODE
		self.globalDict = MoiraCard.globalDict
		if filename: self.load(filename)

	#-----------------------------------------------------------------------
	# Return input filename
	#-----------------------------------------------------------------------
	def filenames(self, filename):
		files = super().filenames(filename)
		files.append(self.macrofile(filename))
		return files

	#-----------------------------------------------------------------------
	def macrofile(self, filename):
		fn,ext = os.path.splitext(filename)
		if ext != self.EXT():
			fn = filename
		return fn + Moira.EXTMAC

	#-----------------------------------------------------------------------
	# Cards
	#-----------------------------------------------------------------------
	# Return cards of a specific tag sorted with _pos
	# @param tag     string tag
	#                list of tags
	#                Card.type(), BODY, SOLID, VOLUME...
	# @param which   look the Input.cardsSorted()
	#-----------------------------------------------------------------------
	def cardsSorted(self, tag, which=0):
		if tag == Card.BODY:
			# include also all CELL parents
			cards = super().cardsSorted(Card.BODY, which)
			cellparents = {x.what(1) : None
					for x in self.cardsCache("CELL").values()}
			cards.update(cellparents)
			return cards
		else:
			return super().cardsSorted(tag, which)

	# ----------------------------------------------------------------------
	# Return an extend list with all cards including the ones from the voxel
	# ----------------------------------------------------------------------
	#def allcards(self):
	#	cl = super().allcards()
	#	for voxel in self["VOXELS"]:
	#		if voxel.ignore() or voxel[MoiraCard.VOXEL] is None: continue
	#		cl.extend(voxel[MoiraCard.VOXEL].cardlist)
	#	return cl

	#-----------------------------------------------------------------------
	# Return material list
	#-----------------------------------------------------------------------
	def materialList(self, which=0, *kw):
		lst = super().materialList(which)
		lst.extend([x.name()
			for x in self.cardsSorted(Card.MATERIALS,-1).values()])
		return lst

	#-----------------------------------------------------------------------
	# Return element list
	#-----------------------------------------------------------------------
	def elementList(self, which=0):
		return list(self.defaultElements.keys())

	#-----------------------------------------------------------------------
	# Scan transformations
	# @return rotation matrix from a rot-defi cards or a dictionary with all
	#         transformations
	# FIXME add the other matrix definitions (Euler, etc)
	#-----------------------------------------------------------------------
	def getTransformation(self, idx=None, rotdefi=None, inv=False):
		neg = False
		if isinstance(idx,str):
			if idx and idx[0]=='-':
				neg = True
				idx = idx[1:]
		elif isinstance(idx,int):
			if idx<0:
				neg = True
				idx = -idx
		if inv: neg = not neg

		# return value if cached
		if idx is not None:
			if rotdefi is not None:
				matrix = rotdefi.get(idx)
				if matrix is None:
					return bmath.Matrix(4, type=1)
				elif neg:
					matrix = matrix.inverse()
				return matrix
			matrix = bmath.Matrix(4, type=1)
			assert False	# FIXME temporary for debugging to check if needed

		# ------------------------------------------------------
		# Scan transformations cards
		# ------------------------------------------------------
		# ROTM cards
		# ------------------------------------------------------
		rotdefi = {}
		for card in self.cardsSorted("ROTM",-1).values():
			if not card.nwhats(): continue
			matrix = bmath.Matrix(4, type=1)
			# Set only rotation component
			matrix[0][0] = card.numWhat(1)
			matrix[0][1] = card.numWhat(2)
			matrix[0][2] = card.numWhat(3)
			matrix[1][0] = card.numWhat(4)
			matrix[1][1] = card.numWhat(5)
			matrix[1][2] = card.numWhat(6)
			matrix[2][0] = card.numWhat(7)
			matrix[2][1] = card.numWhat(8)
			matrix[2][2] = card.numWhat(9)
			if idx is None:
				rotdefi[card.name()] = matrix
			#else:
			#	pass

		# ------------------------------------------------------
		# Add translation from LATTICE card
		# FIXME wrong!!!!
		# we can have multiple LATTICE with the same rotdefi
		# but with different translation
		# A unique identifier should be found rot/lattice
		# ------------------------------------------------------
		#for card in self.cardsSorted("LATTICE",-1).values():
		#	name = card.what(1)
		#	matrix = rotdefi.get(name)
		#	if matrix is None:
		#		matrix = bmath.Matrix(4, type=1)
		#	matrix[0][3] = card.numWhat(2)
		#	matrix[1][3] = card.numWhat(3)
		#	matrix[2][3] = card.numWhat(4)
		#	rotdefi[name] = matrix

		# ------------------------------------------------------
		# MATRIX cards
		# ------------------------------------------------------
		#for card in self.cardsSorted("MATRIX",-1).values():
		#	if not card.whats(): continue
		#	matrix = bmath.Matrix(4, type=1)
		#	matrix[0][0] = card.numWhat(1)
		#	matrix[0][1] = card.numWhat(2)
		#	matrix[0][2] = card.numWhat(3)
		#	matrix[0][3] = card.numWhat(4)
		#	matrix[1][0] = card.numWhat(5)
		#	matrix[1][1] = card.numWhat(6)
		#	matrix[1][2] = card.numWhat(7)
		#	matrix[1][3] = card.numWhat(8)
		#	matrix[2][0] = card.numWhat(9)
		#	matrix[2][1] = card.numWhat(10)
		#	matrix[2][2] = card.numWhat(11)
		#	matrix[2][3] = card.numWhat(12)
		#	if idx is None:
		#		rotdefi[card.name()] = matrix
		#	#else:
		#	#	pass

		#if idx is None:
		return rotdefi
		#return matrix

	#-----------------------------------------------------------------------
	# Rotation/transformation list
	# ----------------------------------------------------------------------
	def rotationList(self, negative=False):
		rot = set()
		try:
			for card in self.cards["MATRIX"]:
				rot.add(card.name())
		except:
			pass
		try:
			for card in self.cards["ROTM"]:
				rot.add(card.name())
		except:
			pass

		l = sorted(rot)
		if negative:
			lst = ["",MoiraCard.DEF_ROTM]
			for i in l:
				lst.append(i)
		else:
			lst = l
			lst.insert(0,MoiraCard.DEF_ROTM)
			lst.insert(0,"")

		del rot
		return lst

	#-----------------------------------------------------------------------
	# Return the body properties as a dictionary of the active/enabled
	# bodies
	#-----------------------------------------------------------------------
	def bodyProperties(self, rotdefi):
		matrix  = None	# current transformation matrix
		#matname = None
		for card in self.allcards():
			if card.ignore(): continue
			if card.tag == "MATRIX":
				if not card.whats():
					#matname = None
					matrix  = None
				else:
					# current transformation matrix
					#matname = card.name()
					matrix  = bmath.Matrix(4,type=1)
					matrix[0][0] = card.numWhat(1)
					matrix[0][1] = card.numWhat(2)
					matrix[0][2] = card.numWhat(3)
					matrix[0][3] = card.numWhat(4)
					matrix[1][0] = card.numWhat(5)
					matrix[1][1] = card.numWhat(6)
					matrix[1][2] = card.numWhat(7)
					matrix[1][3] = card.numWhat(8)
					matrix[2][0] = card.numWhat(9)
					matrix[2][1] = card.numWhat(10)
					matrix[2][2] = card.numWhat(11)
					matrix[2][3] = card.numWhat(12)
			elif card.type() == Card.BODY:
				if matrix:
					card[Card.MATRIX]  = matrix
					#card["@transform"] = matname
				else:
					del card[Card.MATRIX]

	#-----------------------------------------------------------------------
	# Return the region properties as a dictionary of the active/enabled
	# regions
	#-----------------------------------------------------------------------
	def regionProperties(self):
		matDict = self.materialDict()

		#---------------------------------------------------------------
		# Build dictionaries and clean up regions from Input and voxel
		regionDict = self.cardsSorted(Card.CELL)
		regionList = []
		for region in regionDict.values():
			if region.name() == "&": continue
			# Assign default values to region
			region[Card.TYPE]      = Card.REGION_NORMAL
			region[Card.ASSIGNMAT] = region	# self
			if region.tag == "CELL":
				matname = region.evalWhat(2)
				region[Card.PARENT] = regionDict.get(region.what(1))
			else:
				matname = region.evalWhat(1)
				region[Card.PARENT] = None #regionDict.get(region.what())
			region[Card.MATERIAL]= matDict.get(matname)
			region["@biasingA"]  = None	# All particles biasing
			region["@biasingH"]  = None	# Hadrons
			region["@biasingE"]  = None	# emf
			region["@biasingN"]  = None	# low-neutrons
			region[Card.LATTICE] = None
			region["@rotdefi"]   = None
			region["@rotdefi2"]  = None
			region[Card.CONT]    = False	# has continuation cards
			regionList.append(region)
		for card in self.cardsSorted(("LATTICE"),-1).values():
			reg = regionDict.get(card.evalWhat(0))
			reg[Card.TYPE]     = Card.REGION_LATTICE
			reg[Card.MATERIAL] = matDict.get("G4_Galactic")
			reg[Card.LATTICE]  = card
			reg["@rotdefi"]    = card.evalWhat(1)
			reg["@dx"]  = card.numWhat(2)
			reg["@dy"]  = card.numWhat(3)
			reg["@dz"]  = card.numWhat(4)
			#reg["@rotdefi2"] = rotdefi2

	#-----------------------------------------------------------------------
	@staticmethod
	def addFlukaMat2Moira(flukaname, g4name):
		Moira.FLUKA_MAT_2_MOIRA[flukaname] = g4name

	#-----------------------------------------------------------------------
	@staticmethod
	def addFlukaEle2Moira(flukaname, g4name):
		Moira.FLUKA_ELE_2_MOIRA[flukaname] = g4name

	#-----------------------------------------------------------------------
	@staticmethod
	def materialFromFluka(material):
		# FIXME very crude
		# check on predefined ones
		try:
			return Moira.FLUKA_MAT_2_MOIRA[material]
		except KeyError:
			pass
		# check if by appending G4_ exists
		g4mat = f"G4_{material}"
		if g4mat in Moira.defaultMaterials:
			return g4mat

	#-----------------------------------------------------------------------
	@staticmethod
	def elementFromFluka(element):
		# check on predefined ones
		try:
			return Moira.FLUKA_ELE_2_MOIRA[element]
		except KeyError:
			pass
		g4ele = f"{element}"
		if g4ele in Moira.defaultElements:
			return g4ele
		return element
		# raise KeyError(f"Element {element} not implemented yet!")

	#-----------------------------------------------------------------------
	#@staticmethod
	#def materialDefaultInFluka(material):
	#	return material in Moira.FLUKA_MAT_2_MOIRA

	#-----------------------------------------------------------------------
	@staticmethod
	def addFlukaParticle2Moira(flukaname, g4name):
		Moira.FLUKA_PARTICLE_2_MOIRA[flukaname] = g4name

	#-----------------------------------------------------------------------
	@staticmethod
	def particleFromFluka(particle):
		return Moira.FLUKA_PARTICLE_2_MOIRA.get(particle,'')

	#-----------------------------------------------------------------------
	@staticmethod
	def addMassParticleMoira(particle, mass):
		Moira.MASS_PARTICLES_MOIRA[particle] = mass

	#-----------------------------------------------------------------------
	@staticmethod
	def massParticle(particle):
		return Moira.MASS_PARTICLES_MOIRA.get(particle,'')

	#=======================================================================
	# I/O
	# methods for Input IO operations
	#=======================================================================
	#-----------------------------------------------------------------------
	# Export method
	#-----------------------------------------------------------------------
	def _save(self):
		self._saveMoira()
		self._saveMacro()

	#-----------------------------------------------------------------------
	def _saveMoira(self):
		# unique name for identity rotation matrix
		self.writeline("// Default identity rotation matrix")
		self.writeline(f":ROTM {MoiraCard.DEF_ROTM} 0 0 0")
		self.writeline()

		# Loop all cards (write the geometry ones)
		for card in self.cardlist:
			if not isinstance(card, MoiraCard): continue
			if card.ignore(): continue
			tag = card.tag if card.info.otag is None else card.info.otag
			is_compact_extension = card.info.compact_mode and card.info.is_extension
			# Do not dump macro commands or cards extensions into moira file!
			if tag[0] == "/" or is_compact_extension: continue
			if card.comment(): self.writeline(card.commentStr())
			self.writeline(card.toStr())

	#-----------------------------------------------------------------------
	def _saveMacro(self):
		# Second pass write all macro cards
		# Take the filename from the first of the list
		# it could potentially be different than what the input has
		self.open(self.macrofile(self._files[0].name), "w")

		# keep track of all MESH.xxx and dump them at the end
		meshes = []
		previous_card_directory = ""
		for card in self.cardlist:
			if not isinstance(card, MoiraCard): continue
			if card.ignore(): continue
			tag = card.tag if card.info.otag is None else card.info.otag

			# Keep compatibility with G4 mesh cards for now
			if card.tag in ("G4.MESH.BOX", "G4.MESH.CYLINDER"):
				if meshes: # automatically close
					self.writeline("/score/close")
				meshes.append(card.name())
			elif card.tag == "/run/beamOn" and meshes:
				self.writeline("/score/close")

			is_compact_extension = card.info.compact_mode and card.info.is_extension
			# Do not dump non-macro commands in macro file!
			if tag[0] != "/" and not is_compact_extension: continue

			# Print card comment
			if card.comment():
				self.writeline(card.commentStr())

			# Print card
			self.writeline(card.toStr(previous_card_directory = previous_card_directory))

			# Keep base card directory path
			# (can be needed for extension cards)
			is_base_macro_card = card.info.otag and not card.info.is_extension
			if is_base_macro_card:
				previous_card_directory = os.path.dirname(card.info.otag)

		# Keep compatibility with G4 mesh cards for now
		for mesh in meshes:
			self.writeline(f"/score/dumpAllQuantitiesToFile {mesh} {mesh}.bnn")

		# Close macro
		self.close()

	#-----------------------------------------------------------------------
	# Import/load a macro, and the related moira input files.
	#-----------------------------------------------------------------------
	def _load(self):
		# Takes first filename.
		macrofile = self.macrofile(self._files[0].name)

		# Moira file
		self._loadMoira()
		# Macro file.
		self._loadMacro(macrofile)

	#-----------------------------------------------------------------------
	# Import/load a moira input file
	#-----------------------------------------------------------------------
	def _loadMoira(self):
		card = None
		self._comment = []		# current comment
		self._expr = ""
		tag = ""
		while True:
			line = self.readline()
			if line is None: break
			line = self._parseComment(line, "//")
			if line is None: continue
			words = line.split()
			if words[0][0]==":":
				tag = words[0][1:]
				del words[0]
				what = words
				if tag == "CELL":
					what = words[:3]
					self._expr = " ".join(words[3:])
				else:
					what = words
				card = self._addCard(tag, what)
				if not card: continue

				# Correct units
				case = card.case()
				if case is None: continue

				vtype = card.info.vtype[case]
				for i,w in enumerate(card.whats()):
					try:	# Convert units
						w = str(float(w) / MoiraCard.unitConv[vtype[i]])
						card.setWhat(i,w)
					except (KeyError, ValueError, IndexError, TypeError):
						pass
			else:
				if self._comment and card:
					card.appendComment(self._comment)
					self._comment = []
				if tag=="CELL":
					card.appendExpr(line.strip())
				elif card:
					card.appendWhats(words)

	#-----------------------------------------------------------------------
	# Load a G4 macro file (visualize it with Flair).
	# Create macro cards.
	#-----------------------------------------------------------------------
	def _loadMacro(self, filename):
		# Open file
		self.open(filename, "r", showerror = False)

		line = self.readline()
		next_line = ""

		# Go through entire macro file.
		while line is not None:

			# Skip commented lines.
			line = self._parseComment(line, "#")
			if line is None:
				line = self.readline()
				continue

			# Split line: command + parameters (if exist).
			line_command, *line_parameters = line.split()

			# Look whether a macro Flair card exists for the parsed command.
			is_macro_card = False
			macro_card_info = CardInfo.getMacroCardInfo(MoiraCard.MODE,
								    line_command)
			# Found macro card: use its tag!
			if macro_card_info is not CardInfo.NONE:
				is_macro_card = True
				new_card_tag = macro_card_info.tag
			# Otherwise, directly consider the command as a card.
			else:
				new_card_tag = line_command

			# Add new card, corresponding to the parsed line
			new_card = MoiraCard(new_card_tag,
					     line_parameters,
					     self._comment)
			self._comment = []
			all_extension_cards = []

			# If the new card is a macro card (instead of just 1 command):
			# parse the next lines, to be able to:
			# - add relevant parameters to the card,
			# - add extension cards.
			if is_macro_card:
				found_property = True
				found_extension_card = True

				# Go on as long as the next lines are related to the present macro card.
				while found_property or found_extension_card:

					next_line = self.readline()

					# Empty line means end of file
					if next_line is None:
						break
					# Ignore commented lines
					next_line = self._parseComment(next_line, "#")
					if next_line is None:
						continue

					# Split next line: command + parameters (if exist).
					try:
						next_line_command, next_line_parameters = next_line.split(None, 1)
					# Properties and extension cards would add info (by definition),
					# hence exit!
					except ValueError:
						break

					# Try to see whether an extension card exists for the parsed next line.
					# Properties / extension cards detect command name only,
					# not full path:
					next_line_command = next_line_command.rsplit("/", 1)[-1]
					extension_card_info = CardInfo.getMacroCardInfo(MoiraCard.MODE,
											next_line_command)

					# Found extension card.
					found_extension_card = (extension_card_info is not CardInfo.NONE) \
							       and extension_card_info.is_extension
					if found_extension_card:
						# Create extension card.
						extension_card = MoiraCard(extension_card_info.tag,
									   "",
									   self._comment)
						self._comment = []
						# Extension card: its properties are in the same line.
						extension_card.setPropertyIfExists(next_line_command,
										   next_line_parameters)
						# Store it.
						all_extension_cards.append(extension_card)
					# No extension card: look whether the line
					# holds property info for the base card.
					else:
						found_property = new_card.setPropertyIfExists(next_line_command,
											      next_line_parameters)
				# End: while found_property or found_extension_card

				# Update line, since next line has already been parsed.
				line = next_line

			# The new card is not a macro card:
			# nothing to do, move on directly to the next line.
			else:
				line = self.readline()

			# Finally, store the base card
			# and its extensions cards, if they exist.
			self.add(new_card)
			for extension_card in all_extension_cards:
				self.add(extension_card)

		# End: while line is not None

		# Close macro file.
		self.close()

	#-----------------------------------------------------------------------
	def _addCard(self, tag, what):
		if tag == "BODY":
			tag = what[1]
			del what[1]

		elif tag=="SOLID":
			tag = f"{tag}.{what[1]}"
			del what[1]

		elif tag=="VOLU":
			if len(what)>3:
				tag = f"{tag}.{what[1]}"
				del what[1]
				# Move material
				what.insert(1, what.pop())

		elif tag == "ROTM":
			if what[0]==MoiraCard.DEF_ROTM:
				self._comment = []
				return None

		elif tag == "PLACE":
			if what[3]==MoiraCard.DEF_ROTM:
				what[3] = ""

		card = MoiraCard(tag, what, self._comment)
		self.add(card)
		self._comment = []
		if tag=="CELL":
			card.setExpr(self._expr)
			self._expr = ""

		return card

	#-----------------------------------------------------------------------
	def _parseComment(self, line, prefix):
		line = line.rstrip()
		if not line:
			self._comment.append("")
			return None
		if line.lstrip().startswith(prefix):
			self._comment.append(line.replace(prefix, "", 1))
			return None
		return line

	# ----------------------------------------------------------------------
	# Write an input file to be used for plotting
	# ----------------------------------------------------------------------
	def writePlotgeomInput(self, plotgeom, directory=".", title=""):
		ftmp = tempfile.NamedTemporaryFile(mode="wt",
					suffix=self.EXT(),
					prefix=Global.tmpPrefix,
					dir=directory,
					delete=False)
		self.open(ftmp, "w", backup=False)
		self.preprocess()

		# Write a minimized file with only the geometry and
		# what it is necessary
		# ------------------------------------------------------
		self.write("// Temporary input file only for geometry plotting\n")
		self._saveMoira()

		# -------------------------------------------------------
		# Create a temporary macro file for plotting the geometry
		# -------------------------------------------------------
		fn = os.path.splitext(ftmp.name)[0]
		self.open(fn+Moira.EXTMAC,"w")

		O    = plotgeom.get("O", bmath.Vector.O)
		U    = plotgeom.get("U", bmath.Vector.Z)
		V    = plotgeom.get("V", bmath.Vector.X)
		extu = plotgeom.get("extu", 50.0)
		extv = plotgeom.get("extv", 50.0)
		nu   = int(plotgeom.get("nu",200))
		nv   = int(plotgeom.get("nv",200))

		prefix = ["/field/", "/control/", "FIELD", "&FIELD"]
		for card in self.cardlist:
			if not isinstance(card, MoiraCard): continue
			if card.ignore(): continue
			tag = card.tag if card.info.otag is None else card.info.otag
			if any(tag.startswith(x) for x in prefix):
				Global.dump(f"Moira.writePlotgeomInput field={card}")
				self.writeline(card.toStr())
		unit="cm"
		self.write("# Temporary macro file only for geometry plotting\n")
		self.write("/run/initialize\n")
		self.write(f"/plotter/origin {O.x} {O.y} {O.z} {unit}\n")
		self.write(f"/plotter/u      {U.x} {U.y} {U.z}\n")
		self.write(f"/plotter/v      {V.x} {V.y} {V.z}\n")
		self.write(f"/plotter/extend {extu} {extv} {unit}\n")
		self.write(f"/plotter/grid   {nu} {nv}\n")
		self.write(f"/plotter/plot ! 0 {unit}\n")
		self.close()	# for the macro
		self.close()	# and for moira
		return fn

	#-----------------------------------------------------------------------
	@classmethod
	def newCard(cls, *args, **kwargs):
		return MoiraCard(*args, **kwargs)
