# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   9-Nov-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

from deck import Card, CardInfo, Layout

from .input import Moira
from .card import MoiraCard

#===============================================================================
# Layout class for the specific input/canvas display
#===============================================================================
class MoiraLayout(Layout):
	MODE = MoiraCard.MODE

	# ----------------------------------------------------------------------
	def init(self):
		order = [
			("ISOT",      None),

			("ELEM",      None),
			("ELEM_FROM_ISOT", _ELEM_FROM_ISOT),

			("MIXT",           _MIXT_layout),
			("MIXT_BY_NATOMS", _MIXT_layout),
			("MIXT_BY_WEIGHT", _MIXT_layout),
			("MIXT_BY_VOLUME", _MIXT_layout),

			("ROTM"     , _ROTM_layout ),
			("BODY.&RPP", Layout.generic_layout ),
			("BODY.&BOX", Layout.generic_layout ),
			("BODY.&SPH", Layout.generic_layout ),
			("BODY.R&CC", Layout.generic_layout ),
			("BODY.R&EC", Layout.generic_layout ),
			("BODY.&TRC", Layout.generic_layout ),
			("BODY.&TRX", Layout.generic_layout ),
			("BODY.&TRY", Layout.generic_layout ),
			("BODY.&TRZ", Layout.generic_layout ),
			("BODY.E&LL", Layout.generic_layout ),
			("BODY.&WED", Layout.generic_layout ),
			("BODY.RAW" , Layout.generic_layout ),
			("BODY.&ARB", Layout.generic_layout ),
			("BODY.&PLA", Layout.generic_layout ),
			("BODY.&QUA", Layout.generic_layout ),
			("BODY.&YZP", Layout.generic_layout ),
			("BODY.X&ZP", Layout.generic_layout ),
			("BODY.&XYP", Layout.generic_layout ),
			("BODY.XCC" , Layout.generic_layout ),
			("BODY.YCC" , Layout.generic_layout ),
			("BODY.ZCC" , Layout.generic_layout ),
			("BODY.XEC" , Layout.generic_layout ),
			("BODY.YEC" , Layout.generic_layout ),
			("BODY.ZEC" , Layout.generic_layout ),

			("CELL"     , _CELL_layout  ),
			("PLACE"    , _PLACE_layout ),
		]
		self._init_set(order, Moira.MODE)

	# ----------------------------------------------------------------------
	# Enable/disable caching of lists
	# ----------------------------------------------------------------------
	def cache(self, on):
		super().cache(on)
		if on:
			self._cache["body"]         = self.bodyList()
			self._cache["material"]     = self.materialList()
			self._cache["element"]      = self.elementList()
			self._cache["rotation"]     = self.rotationList()
			self._cache["region"]       = self.regionList()
			self._cache["volumenocell"] = self.volumeListNoCell()
			self._cacheTypedef()

	# ----------------------------------------------------------------------
	# Return body list
	# ----------------------------------------------------------------------
	def bodyList(self, inctag=True):
		try: return self._cache["body"]
		except KeyError: pass
		bodies = set()
		if inctag:
			bodies.add("")
			for card in self.input.cardsSorted(Card.BODY).values():
				if card is not None:
					bodies.add(f"{card.name()} [{card.tag}]")
			for card in self.input.cardsSorted(Card.CELL).values():
				bodies.add(f"{card.what(1)} [{card.tag}]")
		else:
			# FIXME using Card.BODY does not return body cards. 
			for card in self.input.cardsSorted("BODY").values():
				bodies.add(card.name())
			for card in self.input.cardsSorted(Card.CELL).values():
				bodies.add(card.what(1))
		#if "VOXELS" in self.input.cards:
		#	bodies.add("VOXEL")
		return sorted(list(bodies), key=str.upper)

	# ----------------------------------------------------------------------
	def isotopeList(self):	return self.typedefList("isotope")
	def volumeList(self):	return self.typedefList("volume")
	cellList = volumeList

	# ----------------------------------------------------------------------
	# Return element list
	#-----------------------------------------------------------------------
	def elementList(self, e=None):
		try: return self._cache["element"]
		except KeyError: pass
		lst = [""] + self.input.elementList(1)
		if e:
			try: lst.remove(e)
			except ValueError: pass
		lst = lst + self.typedefList("element")
		return lst

	# ----------------------------------------------------------------------
	# @return list of volumes excluding CELL, as possible parents
	# ----------------------------------------------------------------------
	def volumeListNoCell(self):
		try: return self._cache["volumenocell"]
		except KeyError: pass
		vol = set()
		for tag in CardInfo.volumes.get(self.input.mode,[]):
			if tag == "CELL": continue
			try:
				for n in [x.name() for x in self.input.cards[tag]]:
					vol.add(n)
			except KeyError:
				pass
		lst = list(vol)
		lst.sort(key=str.upper)
		lst.insert(0,"")
		return lst

	# ----------------------------------------------------------------------
	# @return G4 region list
	# ----------------------------------------------------------------------
	def regionList(self):
		try: return self._cache["region"]
		except KeyError: pass
		lst = self.volumeList()	# << include also all volumes
		lst.extend(self.input.cardsSorted(Card.REGION).keys())
		lst.sort(key=str.upper)
		return lst

#================================ Geometry Cards ===============================
#---------------------------------- MIXT----------------------------------------
def _ELEM_FROM_ISOT(card):
	layout = [ Layout.name,
		   ("Symbol", 1),
		   ("n",     (2, list(range(60))), "Number of isotopes")]
	for i in range(card.intWhat(2)):
		layout.append(Layout.newline)
		layout.append(Layout.tab)
		layout.append((f"Elem.{i+1:02d}", (2*i+3, MoiraLayout.isotopeList)))
		layout.append((f"f.{i+1:02d}",     2*i+4))
	return layout

#---------------------------------- MIXT----------------------------------------
def _MIXT_layout(card):
	layout = [ Layout.name,
		   ("Density", 1),
		   ("n", (2, list(range(60))), "Number of components")]
	n = max(card.intWhat(2), (card.nwhats()-3)//2)
	for i in range(n):
		layout.append(Layout.newline)
		layout.append(Layout.tab)
		if card.tag == "MIXT_BY_NATOMS":
			layout.append((f"Elem.{i+1:02d}", (2*i+3, MoiraLayout.elementList)))
		else:
			layout.append((f"Mat.{i+1:02d}", (2*i+3, Layout.materialList)))
		layout.append((f"f.{i+1:02d}",    2*i+4))
	return layout

#---------------------------------- ROTM----------------------------------------
def __ROTM_type(layout, card, val=None):
	if val is None:
		n = card.nwhats()
		if   n<=4: return 4
		elif n<=7: return 7
		else: return 10
	elif card.nwhats() != val:
		layout.flair.addUndo(layout.flair.setNWhatsUndo(card, val))

__ROTM_header = [ (None, 0),
		   ("Type", (__ROTM_type,
				[("Euler",   4),
				 ("Polar",   7),
				 ("Matrix", 10 )]),
			     "Select rotation type"),
		   Layout.newline]

__ROTM3_layout = __ROTM_header[:] + \
		[ Layout.tab,
		  ("Rx", 1),
		  ("Ry", 2),
		  ("Rz", 3) ]

__ROTM6_layout = __ROTM_header[:] + \
		[ Layout.tab,
		  ("Polar.x", 1),
		  ("Azm.x",   2),
		  Layout.newline,
		  Layout.tab,
		  ("Polar.y", 3),
		  ("Azm.y",   4),
		  Layout.newline,
		  Layout.tab,
		  ("Polar.z", 5),
		  ("Azm.z",   6)]

__ROTM9_layout = __ROTM_header[:] + \
		[ Layout.tab,
		  ("Mxx", 1),
		  ("Mxy", 2),
		  ("Mxz", 3),
		  Layout.newline,
		  Layout.tab,
		  ("Myx", 4),
		  ("Myy", 5),
		  ("Myz", 6),
		  Layout.newline,
		  Layout.tab,
		  ("Myx", 7),
		  ("Myy", 8),
		  ("Myz", 9)]

def _ROTM_layout(card):
	if card.nwhats()==10:
		return __ROTM9_layout
	elif card.nwhats()==7:
		return __ROTM6_layout
	else:
		return __ROTM3_layout

#---------------------------------- CELL----------------------------------------
__CELL_layout_definition = [
		(None,           0),
		("Parent",      (1, MoiraLayout.volumeListNoCell)),
		("Mat",         (2, Layout.materialList)),
	Layout.newline,
		("expr",        -1)]

__CELL_layout_cont = [ (None, -1) ]

def _CELL_layout(card):
	if card.name() == "&":
		return __CELL_layout_cont
	else:
		return __CELL_layout_definition

#--------------------------------- PLACE ---------------------------------------
_PLACE_layout = [
		(None,          (0, MoiraLayout.volumeListNoCell)),
		("#",            1),
		("Parent",      (2, MoiraLayout.volumeListNoCell)),
	Layout.newline,
	Layout.tab,
		("X",            4),
		("Y",            5),
		("Z",            6),
	Layout.newline,
	Layout.tab,
		("Rot",         (3, Layout.rotationList))]

__CELL_layout_cont = [ (None, -1) ]
