# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

import re
import math
import bmath
import Evaluation
from deck import CardInfo, Card, CSGCard

#===============================================================================
# Global dictionary for penelope, containing all units constants
#===============================================================================
class PenelopeGlobalDict(Evaluation.GlobalDict):
	def __init__(self):
		super().__init__()
		self.redefineDistance_cm(1)	# in cm
		self.redefineEnergy_GeV(1e9)	# in eV
		self.redefineTime_s(1)		# in s

#===============================================================================
# Penelope CardInfo class
#===============================================================================
class PenelopeCardInfo(CardInfo):
	# ----------------------------------------------------------------------
	# Post initialization after loading other inputs init
	# ----------------------------------------------------------------------
	@staticmethod
	def post_init():
		mode = PenelopeCard.MODE
		info = CardInfo.get(mode, "MFNAME")
		info.define = "material"
		CardInfo.materials[mode] = "MFNAME"
		CardInfo.volumes[mode]   = list(CardInfo.typedef[mode]["region"])
		CardInfo.rotations[mode] = ""

#===============================================================================
#---------------------------------------------------------------
# Reduced form                 Indices       Quadric
#                           x² y² z² z  C
#---------------------------------------------------------------
# z −1 = 0                  0  0  0  1 -1    plane
# z² − 1 = 0                0  0  1  0 -1    pair of parallel planes
# x² + y² + z² − 1 = 0      1  1  1  0 -1    sphere
# x² + y² − 1 = 0           1  1  0  0 -1    cylinder
# x² − y² − 1 = 0           1 -1  0  0 -1    hyperbolic cylinder
# x² + y² − z² = 0          1  1 -1  0  0    cone
# x² + y² − z² − 1 = 0      1  1 -1  0 -1    one sheet hyperboloid
# x² + y² − z² + 1 = 0      1  1 -1  0  1    two sheet hyperboloid
# x² + y² − z = 0           1  1  0 -1  0    paraboloid
# x² − z = 0                1  0  0 -1  0    parabolic cylinder
# x² − y² − z = 0           1 -1  0 -1  0    hyperbolic paraboloid
#---------------------------------------------------------------
# unique id for reduced quadric
#---------------------------------------------------------------
#@staticmethod
# FIXME unfortunately older python versions do not accept it as static_method
# of PenelepeCard
def surface_id(x2,y2,z2,z,c):
	return (x2+1)*3**4 + \
	       (y2+1)*3**3 + \
	       (z2+1)*3**2 + \
	       (z+1)*3 + \
	       (c+1)

#===============================================================================
# PenelopeCard base class
#===============================================================================
class PenelopeCard(Card, CSGCard):
	"""Penelope input card"""
	MODE       = "penelope"
	COMMENT    = "       "
	DISABLE    = COMMENT

	cardInfo   = PenelopeCardInfo	# Pointer to Card Info
	globalDict = PenelopeGlobalDict()

	PAT_INLINE_CMT = re.compile(r"^\s*(\[.*\])\s*", re.M)

	#---------------------------------------------------------------
	_surface_form = {
			surface_id(0, 0, 0, 0, 0) : "quadric0",
			surface_id(0, 0, 0, 1, 0) : "plane",
			surface_id(0, 0, 0, 1,-1) : "plane",
			surface_id(0, 0, 0, 1, 1) : "plane",
			surface_id(0, 0, 1, 0,-1) : "pair_planes",
			surface_id(1, 1, 1, 0,-1) : "sphere",
			surface_id(1, 1, 0, 0,-1) : "cylinder",
			surface_id(1,-1, 0, 0,-1) : "hyp_cylinder",
			surface_id(1, 1,-1, 0, 0) : "cone",
			surface_id(1, 1,-1, 0,-1) : "hyper1",
			surface_id(1, 1,-1, 0, 1) : "hyper2",
			surface_id(1, 1, 0,-1, 0) : "paraboloid",
			surface_id(1, 0, 0,-1, 0) : "parabolic_cylinder",
			surface_id(1,-1, 0,-1, 0) : "hyperbolic_cylinder",
			#"plane"               : surface_id(0, 0, 0, 1, 0),
			#"plane"               : surface_id(0, 0, 0, 1,-1),
			#"pair_planes"         : surface_id(0, 0, 1, 0,-1),
			#"sphere"              : surface_id(1, 1, 1, 0,-1),
			#"cylinder"            : surface_id(1, 1, 0, 0,-1),
			#"hyp_cylinder"        : surface_id(1,-1, 0, 0,-1),
			#"cone"                : surface_id(1, 1,-1, 0, 0),
			#"hyper1"              : surface_id(1, 1,-1, 0,-1),
			#"hyper2"              : surface_id(1, 1,-1, 0, 1),
			#"paraboloid"          : surface_id(1, 1, 0,-1, 0),
			#"parabolic_cylinder"  : surface_id(1, 0, 0,-1, 0),
			#"hyperbolic_cylinder" : surface_id(1,-1, 0,-1, 0),
		}

	# ----------------------------------------------------------------------
	def surfaceForm(self):
		return PenelopeCard._surface_form.get(
				surface_id(
					self.intWhat(1), self.intWhat(2), self.intWhat(3),
					self.intWhat(4), self.intWhat(5)),
				"quadric")

	# ----------------------------------------------------------------------
	# Change tag of card, reduce the number of whats to the new card
	# XXX Do not forget to update the Input.cards... before
	# ----------------------------------------------------------------------
	def setTag(self, tag):
		super().setTag(tag)
		if tag in CardInfo.volumes.get(PenelopeCard.MODE,[]):
			self._type = Card.MODULE
			self._geo  = True

	# ======================================================================
	# String functions
	# ======================================================================
	# Return a formatted comment string
	# ----------------------------------------------------------------------
	def commentStr(self):
		"""return a comment string"""
		lines = []
		first_inline = True
		for line in self.comment().splitlines():
			m = PenelopeCard.PAT_INLINE_CMT.match(line)
			if m and first_inline:
				first_inline = False
				continue
			line = line.rstrip()
			if line:
				lines.append(f"{self.COMMENT}{line}")
			else:
				lines.append(f"{self.COMMENT}.")
		#if lines and lines[-1]: lines.append("")	# to append a last new line
		return "\n".join(lines)

	#-----------------------------------------------------------------------
	# return string with only the cards
	# ----------------------------------------------------------------------
	def _toStr(self, prefix="", comments=False, firstgeo=False, **kwargs):
		tag = self.info.alias if self.info is CardInfo.NONE else self.tag
		inline_cmt = ""
		#---------------------------------------------------------------
		# Geometry card
		#---------------------------------------------------------------
		if self.isGeo():
			WIDTH=64
			comments = self.comment().splitlines()
			if firstgeo:
				line = "\n".join(comments[:-1])
				del comments[0:-1]
				line += f"\n{'0'*WIDTH}"
			else:
				line  = '0'*WIDTH

			if comments:
				if len(comments[-1]) < WIDTH-19:
					inline_cmt = f"   {comments[-1]}".rstrip()
					comments.pop()
				for c in comments:
					line += f"\nC {c}".rstrip()

			line += f"\n{tag:8}({self.what(0):>4})"
			# inline comment
			line += inline_cmt

			if tag in ("SURFACE","SURFACE*"):
				line += f"\nINDICES=({self.intWhat(1):2d}," \
					f"{self.intWhat(2):2d}," \
					f"{self.intWhat(3):2d}," \
					f"{self.intWhat(4):2d}," \
					f"{self.intWhat(5):2d})"

			elif tag in ("BODY","MODULE"):
				line += f"\nMATERIAL({self.what(1):>4})"
				modules = self.input.cardsCache(Card.MODULE)
				for item in self.expr().split():
					name = item[1:]
					if item[0] == '#':
						try:
							t = modules[name].tag
						except KeyError:
							t = "module"
						line += f"\n{t:<8}({name:>4})"
					else:
						side = 1 if item[0]=='+' else -1
						line += f"\nSURFACE ({name:>4}), " \
							f"SIDE POINTER=({side:+2d})"

			# children
			if self._children and self.tag=="MODULE":
				line += f"\n{'1'*WIDTH}"
			for child in self._children:
				comments = child.comment().splitlines()
				inline_cmt = ""
				if comments:
					if len(comments)==1:
						inline_cmt = f"   {comments[0]}".rstrip()
				else:
					for c in comments:
						line += f"\n# {c}".rstrip()
				tag = child.info.alias if child.info is not CardInfo.NONE else child.tag
				line += f"\n{tag:>7}=({bmath.format(child.numWhat(0),22):>22}," \
					f"{child.intWhat(1):4})"
				if child.nwhats()>2:
					line += f"  {child.what(2)}"
				line += inline_cmt
		#---------------------------------------------------------------
		# Normal card
		#---------------------------------------------------------------
		else:
			# find inline comment and append it
			WIDTH=69
			for cmt in self.comment().splitlines():
				pat = PenelopeCard.PAT_INLINE_CMT.match(cmt)
				if pat:
					inline_cmt = pat.group(1)
					break
			line = f"{prefix}{tag:<7}{' '.join(map(str,self.whats()))}"
			if inline_cmt:
				line += " "*(max(1,WIDTH-len(inline_cmt)-len(line)))
				line += inline_cmt
			if comments:
				cmt = self.commentStr()
				if cmt:
					return f"{cmt}\n{line.rstrip()}"
		return line.rstrip()

	# ----------------------------------------------------------------------
	# Return what or the parameter if it is overridden by
	# ----------------------------------------------------------------------
	def paramWhat0(self):
		if self.intWhat(1) == 0 or self.input is None:
			return self.numWhat(0)
		else:
			params = self.input.cardsCache("PARINP")
			return params[str(self.intWhat(1))].numWhat(1)

	# ----------------------------------------------------------------------
	# @return transformation matrix for SURFACE, MODULE, CLONE
	#         None if not existing
	# ----------------------------------------------------------------------
	def matrix(self):
		if not self._children: return None,None
		translate = False
		T = bmath.Vector()
		rw = rt = rp = 0.0
		scale = [1.0, 1.0, 1.0]
		for child in self._children:
			if child.ignore(): continue
			try:
				pos = child.tag.rindex("/")
			except KeyError:
				continue
			tag = child.tag[pos+1:]
			if   tag == "X-SHIFT":
				T.x = child.paramWhat0()
				if T.x != 0.0: translate = True
			elif tag == "Y-SHIFT":
				T.y = child.paramWhat0()
				if T.y != 0.0: translate = True
			elif tag == "Z-SHIFT":
				T.z = child.paramWhat0()
				if T.z != 0.0: translate = True
			elif tag == "X-SCALE":
				scale[0] = child.paramWhat0()
			elif tag == "Y-SCALE":
				scale[1] = child.paramWhat0()
			elif tag == "Z-SCALE":
				scale[2] = child.paramWhat0()
			elif tag == "OMEGA":
				rw = child.paramWhat0()
				if child.what(2)!="RAD":
					rw = math.radians(rw)
			elif tag == "THETA":
				rt = child.paramWhat0()
				if child.what(2)!="RAD":
					rt = math.radians(rt)
			elif tag == "PHI":
				rp = child.paramWhat0()
				if child.what(2)!="RAD":
					rp = math.radians(rp)

		if rp==0 and rt==0 and rw==0 and not translate:
			return None, scale

		matrix = bmath.Matrix(type=1)
		if translate:
			matrix *= bmath.Matrix.translate(T)
		if rp != 0: matrix *= bmath.Matrix.rotZ(rp)
		if rt != 0: matrix *= bmath.Matrix.rotY(rt)
		if rw != 0: matrix *= bmath.Matrix.rotZ(rw)
		return matrix, scale
