# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Oct-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import re
from collections import OrderedDict

import log
import Global

from deck    import Input, Card, CardInfo
from .card   import PenelopeCard, PenelopeCardInfo
from .config import PenelopeConfig

#===============================================================================
# Penelope Input class
#===============================================================================
class Penelope(Input):
	MODE    = PenelopeCard.MODE
	config  = PenelopeConfig()		# configuration information
	defaultMaterials = OrderedDict([
				("vacuum", PenelopeCard("MFNAME",["VACUUM"]))	# vacuum
			])

	PAT_A4      = re.compile(r"^\S+\s*\((.{4})\)(.*)$")
	PAT_A4SIDE  = re.compile(r"^SURFACE \((.{4})\) *, *SIDE POINTER= *\((.*?)\)$")
	#PAT_PARINP  = re.compile(r"^PARINP\((\d\+)\)=\
	PAT_INDICES = re.compile(r"^INDICES=\((.*?),(.*?),(.*?),(.*?),(.*?)\)$")
	PAT_GEOVAR  = re.compile(r"^.{,7}=\((.*?),([ 0-9]*?)\) *(.*)$")
	PAT_GEOVAR2 = re.compile(r"^.{,7}=\((.*?),(.*?)\) *(DEG|RAD)? *(.*)$")
	PAT_INCLUDE = re.compile(r"^ *FILE=\((.*?)\).*")

	PAT_INLINE1 = re.compile(r"^(.*)(\[.*?])\s*")

	_INDENT_INC = {"GEOMFN","INCLUDE","INCLUDE*"}
	_INDENT_DEC = {"END"}

	_MAX_EXPR_LEN = 60

	# ----------------------------------------------------------------------
	def __init__(self, filename=None):
		super().__init__()
		self.mode       = Penelope.MODE
		self.globalDict = PenelopeCard.globalDict
		if filename: self.load(filename)

	#-----------------------------------------------------------------------
	# Return input filename
	#-----------------------------------------------------------------------
	def filenames(self, filename):
		files = super().filenames(filename)
		files.extend([card.what(0) for card in self["GEOMFN"]])
		return files

	#-----------------------------------------------------------------------
	# Cards
	#-----------------------------------------------------------------------
	# Return cards of a specific tag sorted with _pos
	# @param tag     string tag
	#                list of tags
	#                Card.type(), BODY, SOLID, VOLUME...
	# @param which   look the Input.cardsSorted()
	#-----------------------------------------------------------------------
	def cardsSorted(self, tag, which=0):
		cards = super().cardsSorted(tag, which)
		if tag == Card.MATERIALS:
			for card in self.defaultMaterials.values():
				# FIXME Warning on positioning!!!!!
				if card.name() not in cards:
					cards[card.name()] = card
		return cards

	#-----------------------------------------------------------------------
	# Return material list
	#-----------------------------------------------------------------------
	def materialList(self, which=0, *kw):
		lst = super().materialList(which)
		lst.extend([x.name().split(".")[0]
			for x in self.cardsSorted(Card.MATERIALS,-1).values()])
		return lst

	#-----------------------------------------------------------------------
	# Return the body properties as a dictionary of the active/enabled
	# bodies
	#-----------------------------------------------------------------------
	def bodyProperties(self, rotdefi):
		pass

	#-----------------------------------------------------------------------
	# Return the region properties as a dictionary of the active/enabled
	# regions
	#-----------------------------------------------------------------------
	def regionProperties(self):
		matList = list(self.cardsSorted(Card.MATERIALS,-1).values())
		#matDict = self.materialDict()
		vacuum  = matList[-1]
		#print(f"{matList=}")
		#print(f"{matDict=}")
		#---------------------------------------------------------------
		# Build dictionaries and clean up regions from Input and voxel
		n = 0
		regionList = []
		for region in self.cardsSorted(Card.REGION,-1).values():
			try:
				idx = region.intWhat(1)-1
				if idx>=0:
					material = matList[idx]
				else:
					material = vacuum
			except IndexError:
				material = vacuum
			n += 1
			region["@n"] = n		# Assign default values to region
			region[Card.TYPE]     = Card.REGION_NORMAL
			region[Card.ASSIGNMAT]= region	# Card that assigned the material
			region[Card.MATERIAL] = material
			#region["@biasingA"]  = None	# All particles biasing
			#region["@biasingH"]  = None	# Hadrons
			#region["@biasingE"]  = None	# emf
			#region["@biasingN"]  = None	# low-neutrons
			#region[Card.LATTICE] = None
			#region["@rotdefi"]   = ""
			#region[Card.CONT]    = False	# has continuation cards
			regionList.append(region)

	#=======================================================================
	# I/O
	# methods for Input IO operations
	#=======================================================================
	#-----------------------------------------------------------------------
	# Export/save method
	#-----------------------------------------------------------------------
	def _save(self):
		Global.dump("\n*-* Writing penelope file")
		# Loop all cards (write the geometry ones)
		ingeo    = False
		firstgeo = False	# special comment writing for first geo card
		for card in self.cardlist:
			if not isinstance(card, PenelopeCard): continue
			if not card.active: continue

			if ingeo and card.tag=="END":
				self.writeline("0"*64)
				self.writeline("END      0000000000000000000000000000000000000000000000000000000")
				ingeo = False
				self.close()
			else:
				self.writeline(card.toStr(comments=True, firstgeo=firstgeo))
				firstgeo = False

			if card.tag == "GEOMFN":
				self.open(card.what(0),"w")
				ingeo    = True
				firstgeo = True

	#-----------------------------------------------------------------------
	# Import/load a macro and the related moira input files
	#-----------------------------------------------------------------------
	def _load(self):
		Global.dump("\n*-* Reading penelope file")
		card = None
		self._comment = []		# current comment
		self._cardEnable = True
		while True:
			line = self.readline()
			if line is None: break
			line = self._parseComment(line, "      ")	# 6 spaces
			if line is None: continue

			# Get inline comments
			m = Penelope.PAT_INLINE1.match(line)
			if m:
				self._comment.append(m.group(2))
				line = m.group(1)

			tag, what = self._parseLine(line)
			# Check for only a geometry file
			if not self.cardlist and tag != "TITLE":
				self.saveline(line)
				self._parseGeometry()
				break
			card = PenelopeCard(tag, what, self._comment)
			self.add(card)
			if not self._cardEnable:
				card.disable()
			elif tag == "GEOMFN":
				self.open(what[0], "r")
				self._comment = []
				self._parseGeometry()

			self._comment = []
			self._cardEnable = True
			#print(f"line={line}")
			#print(f"tag={tag}, taginfo={nwhats}")

	#-----------------------------------------------------------------------
	# Parse a comment line
	#-----------------------------------------------------------------------
	def _parseComment(self, line, prefix):
		line = line.rstrip()
		if not line: return None
		if isinstance(prefix,str):
			if line.startswith(prefix):
				# Try for a commented card
				lline = line[7:]
				tag, what = self._parseLine(lline)
				ci = CardInfo.get(self.MODE, tag)
				if ci.tag == tag:
					self._cardEnable = False
					return lline
				self._comment.append(lline)
				return None
		for p in prefix:
			if line.startswith(p):
				self._comment.append(line[len(p):])
				return None
		if line in ("c","C"):
			self._comment.append("")
			return None
		return line

	#-----------------------------------------------------------------------
	# Parse a line and split into tag and what
	#-----------------------------------------------------------------------
	def _parseLine(self, line):
		tag = line[:6].strip()
		#if tag == "GEOMFN": breakpoint()
		#if tag in ("TITLE","GEOMFN"):
		if tag == "TITLE":
			what = [line[7:].rstrip()]
		else:
			ci = PenelopeCardInfo.get(self.MODE, tag)
			what = line[7:].split()[:ci.maxwhats]
		return tag, what

	#-----------------------------------------------------------------------
	# Load Constructive Quadratic geometry of Penelope
	#-----------------------------------------------------------------------
	def _parseGeometry(self):
		#---------------------------------------------------------------
		# Show an error on invalid syntax
		#---------------------------------------------------------------
		def invalidSyntax():
			nonlocal card, card_tag, line, tag
			prefix = f"/{card_tag}/" if card is not None else ""
			log.error(f"Line {self._lineNo[-1]} "
				f"Invalid syntax of '{prefix}{tag}'")
			log.errorCont(line)

		# Read cards separated by 0000 zero lines
		card     = None
		card_tag = ""
		include_level = 0		# nesting of includes to count END statements
		prefix_comments = True		# Process header as comments
		while True:
			line = self.readline()
			if prefix_comments:
				if line[:8] == "00000000":
					prefix_comments = False
				else:
					self._comment.append(line)
				continue
			#if line == "# The module is rotated and shifted.": breakpoint()
			line = self._parseComment(line,("#","c ","C "))
			if line is None: continue
			tag = line[:8].strip()
			processed = False

			#-------------------------------------------------------
			# 0*: mark new definition block
			# 1*: mark starting of transformation parameters
			#-------------------------------------------------------
			if tag in ("00000000", "11111111"):
				if tag[0]=="0":
					# New definition block
					card = None
				processed = True

			#-------------------------------------------------------
			# New definition block
			#-------------------------------------------------------
			elif card is None:
				if tag in ("SURFACE", "SURFACE*", "BODY", "MODULE", "CLONE"):
					pat = Penelope.PAT_A4.match(line)
					if pat:
						what = pat.group(1).strip()
						inline_comment = pat.group(2).strip()
						if inline_comment:
							self._comment.append(inline_comment)
						card = PenelopeCard(tag, [what], self._comment)
						self.add(card)
						card_tag = "SURFACE" if card.tag=="SURFACE*" \
								else card.tag
						processed = True
				elif tag in ("INCLUDE","INCLUDE*"):
					line = self.readline()
					pat = Penelope.PAT_INCLUDE.match(line)
					if pat:
						what = pat.group(1).strip()
						self.add(PenelopeCard(tag, [what], self._comment))
						self.open(pat.group(1).strip(),"r")
						prefix_comments = True
						include_level += 1
						processed = True
				elif tag == "END":
					self.add(PenelopeCard(tag, [], self._comment))
					processed = True
					if include_level == 0: return
					include_level -= 1
				else:
					log.error(f"Unknown tag {tag!r} found line: ",
							self._lineNo[-1])
					log.errorCont(line)
					continue

			#-------------------------------------------------------
			elif tag=="INDICES=" and card_tag=="SURFACE":
				pat = Penelope.PAT_INDICES.match(line)
				if pat and card.nwhats()<2:
					card.appendWhats([int(x) for x in pat.groups()])
					processed = True

			#-------------------------------------------------------
			elif tag in ("X-SHIFT=","Y-SHIFT=","Z-SHIFT=") and \
			     card_tag in ("SURFACE", "MODULE", "CLONE"):
				pat = Penelope.PAT_GEOVAR.match(line)
				if pat:
					child = PenelopeCard(f"/{card_tag}/{tag[:-1]}",
							[float(pat.group(1)), int(pat.group(2))])
							#list(pat.group(1,2)))
					if pat.group(3):
						child.appendComment(pat.group(3))
					card.add(child)
					processed = True

			#-------------------------------------------------------
			elif tag in ("AXX=", "AYY=", "AZZ=",
			             "AXY=", "AYZ=", "AXZ=",
			             "AX=",  "AY=",  "AZ=", "A0=") and \
			     card_tag == "SURFACE":
				pat = Penelope.PAT_GEOVAR.match(line)
				if pat:
					child = PenelopeCard(f"/{card_tag}/{tag[:-1]}",
							[float(pat.group(1)), int(pat.group(2))])
							#list(pat.group(1,2)))
					if pat.group(3):
						child.appendComment(pat.group(3))
					card.add(child)
					processed = True
			#-------------------------------------------------------
			elif tag in ("X-SCALE=","Y-SCALE=","Z-SCALE=") and \
			     card_tag == "SURFACE":
				pat = Penelope.PAT_GEOVAR.match(line)
				if pat:
					child = PenelopeCard(f"/{card_tag}/{tag[:-1]}",
							[float(pat.group(1)), int(pat.group(2))])
							#list(pat.group(1,2)))
					if pat.group(3):
						child.appendComment(pat.group(3))
					card.add(child)
					processed = True

			#-------------------------------------------------------
			elif tag in ("OMEGA=","THETA=","PHI=") and \
			     card_tag in ("SURFACE", "MODULE", "CLONE"):
				pat = Penelope.PAT_GEOVAR2.match(line)
				if pat:
					child = PenelopeCard(f"/{card_tag}/{tag[:-1]}",
							[float(pat.group(1)),
							 int(pat.group(2)),
							 pat.group(3)])
							#list(pat.group(1,3)))
					if pat.group(3):
						child.appendComment(pat.group(4))
					card.add(child)
					processed = True

			#-------------------------------------------------------
			elif tag=="MATERIAL" and card_tag in ("BODY","MODULE"):
				pat = Penelope.PAT_A4.match(line)
				if pat:
					card.setWhat(1, int(pat.group(1)))
					processed = True

			#-------------------------------------------------------
			elif tag=="SURFACE" and card_tag in ("BODY", "MODULE"):
				pat = Penelope.PAT_A4SIDE.match(line)
				if pat:
					if int(pat.group(2))>0:
						surface = f"+{pat.group(1).strip()}"
					else:
						surface = f"-{pat.group(1).strip()}"
					expr = card.expr()
					if not expr:
						card.setExpr(surface)
					elif len(expr)-expr.rfind("\n")>Penelope._MAX_EXPR_LEN:
						card.setExpr(f"{expr}\n#{surface}")
					else:
						card.setExpr(f"{card.expr()} {surface}")
					processed = True

			#-------------------------------------------------------
			elif tag in ("BODY", "MODULE") and \
			     card_tag in ("BODY", "MODULE"):
				pat = Penelope.PAT_A4.match(line)
				if pat:
					cell = pat.group(1).strip()
					expr = card.expr()
					if not expr:
						card.setExpr(f"#{cell}")
					elif len(expr)-expr.rfind("\n")>Penelope._MAX_EXPR_LEN:
						card.setExpr(f"{expr}\n#{cell}")
					else:
						card.setExpr(f"{expr} #{cell}")
					processed = True

			#-------------------------------------------------------
			elif card_tag=="CLONE" and tag=="MODULE":
				pat = Penelope.PAT_A4.match(line)
				if pat:
					card.setWhat(1, pat.group(1).strip())
					processed = True

			#-------------------------------------------------------
			if not processed:
				invalidSyntax()

			self._comment = []

	#-----------------------------------------------------------------------
	@classmethod
	def newCard(cls, *args, **kwargs):
		return PenelopeCard(*args, **kwargs)
