# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Feb-2021

import re

import csg
import bmath
import Evaluation
from deck import CardInfo, Card, CSGCard

#===============================================================================
# Global dictionary containing all units constants
#===============================================================================
class PhitsGlobalDict(Evaluation.GlobalDict):
	def __init__(self):
		super().__init__()
		self.redefineDistance_cm(1)	# in cm
		self.redefineEnergy_GeV(1e6)	# in eV	(MeV)
		self.redefineTime_s(1e8)	# in s	(shakes)

#===============================================================================
# Phits CardInfo class
#===============================================================================
class PhitsCardInfo(CardInfo):
	sectionsMap = {}	# section mapping from small letters to more reasonable from ini

	# ----------------------------------------------------------------------
	# Validate card: return a list of failing variables.
	# ----------------------------------------------------------------------
	def validate(self, card, case=None):
		pb = super().validate(card, case)

		# Special check for regions
		if card.type()==Card.REGION:
			exp = csg.tokenize(card.expr())
			# Check expression
			try:
				csg.exp2rpn(exp)
				if exp and not csg.check(exp):
					pb.append(-1)
					pb.append("Invalid expression")
				else:
					surfaces = card.input.cardsCache(Card.BODY)
					cells    = card.input.cardsCache(Card.REGION)
					notadded = True
					for token in exp:
						if token in "-+:#@": continue
						if card.NAMEPAT.match(token):
							if token not in surfaces:
								if notadded:
									pb.append(-1)
									notadded = False
								pb.append(f"Invalid surface '{token}'")
						elif token[0]=="#":
							token = token[1:]
							if card.NAMEPAT.match(token):
								if token not in cells:
									if notadded:
										pb.append(-1)
										notadded = False
									pb.append(f"Invalid cell '{token}'")
						else:
							if notadded:
								pb.append(-1)
								notadded = False
							pb.append(f"Invalid token '{token}'")
			except csg.CSGException as ex:
				pb.append(-1)
				pb.append(str(ex))
		return pb

	# ----------------------------------------------------------------------
	# Post initialization after loading other inputs init
	# ----------------------------------------------------------------------
	@staticmethod
	def post_init():
		mode = PhitsCard.MODE
		CardInfo.volumes[mode]   = ["CELL","cell"]
		CardInfo.materials[mode] = "/[material]/mat"
		CardInfo.rotations[mode] = ["TR","TR*","tr","tr*"]

		# Make association from [sectioninlower] -> [Section In Lower]
		for card in CardInfo.cards(mode):
			if card.tag[0]=="[":
				short = card.tag.lower().replace(" ","")
				PhitsCardInfo.sectionsMap[short] = card.tag

#===============================================================================
# PhitsCard card class
#===============================================================================
class PhitsCard(Card, CSGCard):
	"""PHITS input card"""
	MODE       = "phits"
	COMMENT    = "$"
	MAXLEN     = 70

	# Patterns
	NAMEPAT    = re.compile(r"^[0-9]*$")
	ETYPE      = re.compile(r"\b[exyzrat]-type$")

	cardInfo   = PhitsCardInfo	# Pointer to Card Info
	globalDict = PhitsGlobalDict()

	# ----------------------------------------------------------------------
	# Change tag of card, reduce the number of whats to the new card
	# XXX Do not forget to update the Input.cards... before
	# ----------------------------------------------------------------------
	def setTag(self, tag):
		super().setTag(tag)
		if tag == "CELL":
			self._type = Card.CELL
			self._geo  = True

	# ---------------------------------------------------------------------
	# return string with only the cards
	# ----------------------------------------------------------------------
	def _toStr(self, prefix="", comments=False, **kwargs):
		#---------------------------------------------------------------
		# Append whats to output string starting from 'start'
		#---------------------------------------------------------------
		def appendWhats(out, start, stop):
			nonlocal inline_comment
			prevlen = len(out)
			for i in range(start, stop):
				try:
					w = self.evalWhat(i)
					w = bmath.format(float(w),useDot=False)
				except (ValueError, TypeError):
					pass
				if len(out)+1+len(w)-prevlen > PhitsCard.MAXLEN:
					out += "\\\n"
					prevlen = len(out)-5	# minus 5 space chars before
				out += f" {w}"
			if inline_comment:
				if out.count("\n")==0 and max(40,len(out))+len(inline_comment)<100:
					return f"{out:<40} {inline_comment}"
				else:
					return f"{inline_comment}\n{out}"
			return out

		#---------------------------------------------------------------
		def appendText(out, width):
			for line in self.text().splitlines():
				out += f"\n{' ':<{width}} {line}"
			return out

		#---------------------------------------------------------------
		# Show comments if necessary
		# inline if possible
		inline_comment = ""
		if comments and self.comment():
			out = self.commentStr()
			if out[-1] != "\n":
				if len(out)<60:
					# potential inline comment
					inline_comment = out
					out = ""
		else:
			out = ""

		#---------------------------------------------------------------
		# Special cases
		# Cells:
		#---------------------------------------------------------------
		if self._type == Card.CELL:
			mat = self.what(1)
			out += f" {self.name():<10} {mat}"
			if mat not in ("-1","0"):
				out += f" {self.evalWhat(2)}"
			lines = [x.rstrip() for x in self.expr().splitlines()]
			if lines:
				lines[0] = f"{out} {lines[0]}"
			else:
				lines.append(out)

			# check line length for >MAXLEN and break it
			i = 0
			prefix = "\n     "
			while i<len(lines):
				if len(lines[i]) > PhitsCard.MAXLEN-len(prefix):
					line = lines[i]
					insert = False
					while len(line) > PhitsCard.MAXLEN-len(prefix):
						# find last space before the MAXLEN column
						try:
							space = line.rindex(' ',0,
								PhitsCard.MAXLEN-len(prefix))
						except ValueError:
							space = PhitsCard.MAXLEN-1
						if insert:
							lines.insert(i,line[:space])
						else:
							lines[i] = line[:space]
							insert = True
						i += 1
						line = line[space+1:].strip()
					lines.insert(i,line)
				i += 1
			return prefix.join(lines)
		#---------------------------------------------------------------
		# Surfaces:
		#---------------------------------------------------------------
		elif self._type == Card.SURFACE:
			if self["tr"] is not None:
				out += f" {self.name():<10} {self['tr']} {self.tag}"
			else:
				out += f" {self.name():<10} {self.tag}"

			prevlen = 0
			#for i in range(1,self.info.maxwhats):
			for i in range(1,self.nwhats()):
				w = bmath.format(self.evalWhat(i))
				if w=="": w = "0"
				if len(out)+1+len(w)-prevlen > PhitsCard.MAXLEN:
					out += "\n     "
					prevlen = len(out)-5	# minus 5 space chars before
				out += f" {w}"
			return out
		#---------------------------------------------------------------
		# Sections:
		#---------------------------------------------------------------
		elif self.tag[0] == "[":
			out += self.info.alias
			if not self.enable:
				out += " off"
			if self.text():
				out += f"\n{self.text()}"
			for child in self._children:
				out += f"\n{child.toStr(comments=comments)}"
			return out
		#---------------------------------------------------------------
		# Other:
		#---------------------------------------------------------------
		else:
			width = 12
			stop  = self.nwhats()
			etype = False
			if self.tag == "/[material]/mat":
				tag   = f"{self.info.alias}[{self.intWhat(0)}]"
				start = 1
			elif self.tag == "/[material]/mat/elem":
				tag   = ""
				start = 0
			elif self.tag in ("tr","tr*"):
				tag   = f"{self.info.alias}{self.intWhat(0)} ="
				start = 1
			elif self.tag in ("/[parameters]/emin", "/[parameters]/dmax",
					"/[parameters]/cmin",   "/[parameters]/file"):
				tag   = f"{self.info.alias}({self.intWhat(0)}) ="
				start = 1
			elif PhitsCard.ETYPE.search(self.tag):
				tag   = f"{self.info.alias} ="
				start = 0
				stop  = 1
				etype = True
			else:
				if self.info.alias!="error":
					tag = f"{self.info.alias} ="
				else:
					tag = f"{self.tag.split('/')[-1]} ="
				start = 0

			if out:
				out += f"\n{tag:>{width}}"
			else:
				out += f"{tag:>{width}}"
			out = appendWhats(out, start, stop)

			# post processing for [exyzlsad]-type
			if etype:
				elet  = self.tag[-6]
				etype = self.intWhat(0)
				if etype in (1,2,3):
					tag = f"n{elet} ="
					out += f"\n{tag:>{width}} {self.intWhat(1)}"
					if etype==1:
						out = appendText(out, width)
				elif etype in (4,5):
					try:
						w = self.evalWhat(1)
						w = bmath.format(float(w),useDot=False)
					except (ValueError, TypeError):
						pass
					tag = f"{elet}del ="
					out += f"\n{tag:>{width}} {w}"

				if etype in (2,3,4,5):
					try:
						w = self.evalWhat(2)
						w = bmath.format(float(w),useDot=False)
					except (ValueError, TypeError):
						pass
					tag = f"{elet}min ="
					out += f"\n{tag:>{width}} {w}"

					try:
						w = self.evalWhat(3)
						w = bmath.format(float(w),useDot=False)
					except (ValueError, TypeError):
						pass
					tag = f"{elet}max ="
					out += f"\n{tag:>{width}} {w}"
			elif self.text():
				out = appendText(out, width)

			for child in self._children:
				out += f"\n{child.toStr(comments=comments)}"
			return out

	# ----------------------------------------------------------------------
	# Set a list of zones as expression
	# ... removing empty zones
	# ----------------------------------------------------------------------
	def setZones(self, zones):
		if zones:
			self.setText("\n: ".join(
				[x.strip() for x in zones if x.strip()]))
		else:
			self.setText("")
