# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Feb-2021

import os
import re

from deck import DeckConfig
from element import Element

#===============================================================================
# A container class to keep the programs
# It should contain only static and class methods
#===============================================================================
class PhitsConfig(DeckConfig):
	SECTION     = "Phits"		# Configuration section name
	version     = "Unknown"
	released    = ""
	directory   = ""		# Executable installation directory
	datapath    = ""		# DATAPATH directory
	CYCLE       = "phits_"		# prefix of temporary cycle directory
	DEFAULT_PRG = "phits.sh"	# [flair.ini] Main application program/script
	DEFAULT_EXE = "phits.exe"	# [flair.ini]
	EXT         = ".pht"		# default input extension
	EXTENSIONS  = [EXT, ".phits", ".inp"]
	programs    = { "prg" : "",
			"exe" : ""
		}
	CYCLEDIR_PAT = re.compile(rf"^{CYCLE}(\d+)$")

	#-----------------------------------------------------------------------
	@staticmethod
	def loadXSdir():
		from .input import Phits
		datapath = PhitsConfig.datapath if PhitsConfig.datapath else os.getenv("DATAPATH")
		if not datapath: return
		try:
			with open(os.path.join(datapath,"xsdir.jnd"),"r") as f:
				del Phits.zaid[:]
				del Phits.sab[:]
				exist = set()
				Phits.zaid.append("")
				Phits.sab.append("")
				for line in f:	# skip until directory line
					if line.rstrip()=="directory": break
				for line in f:
					if line.startswith("     "): continue
					zaid_full = line.split(maxsplit=1)[0]
					zaid = zaid_full.split(".")[0]
					if zaid_full[-1]=="t":
						Phits.sab.append(zaid_full)
					else:
						if zaid not in exist:
							exist.add(zaid)
							Phits.zaid.append(zaid)
						Phits.zaid.append(zaid_full)
		except (FileNotFoundError, PermissionError):
			pass

		elements = set()
		for za in Phits.zaid:
			try:
				za = za.split(".")[0]
			except:
				pass
			try:
				za = int(za)
			except ValueError:
				continue
			z,a = divmod(za,1000)
			try:
				symbol = Element.byZ(z).symbol
			except:
				continue
			elements.add(symbol)
			if a>0:
				elements.add(f"{a}{symbol}")
				elements.add(f"{symbol}-{a}")

		Phits.zaid.extend(sorted(elements))

	#-----------------------------------------------------------------------
	# Load from config
	#-----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		section = PhitsConfig.SECTION
		PhitsConfig.setDir(config.getStr(section, "directory", ""))
		PhitsConfig.setCmd("prg", config.getStr(section, "program",   PhitsConfig.DEFAULT_PRG))
		PhitsConfig.setCmd("exe", config.getStr(section, "executable",PhitsConfig.DEFAULT_EXE))
		PhitsConfig.datapath = config.getStr(section, "datapath",  PhitsConfig.datapath)
		if not PhitsConfig.datapath and PhitsConfig.directory:
			PhitsConfig.datapath = os.path.join(PhitsConfig.directory,"data")
		PhitsConfig.loadXSdir()
