# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Feb-2021

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

from collections import OrderedDict

import re

import log
import Global

from deck    import Card
from mcnp    import McnpGeometry
from .card   import PhitsCard, PhitsCardInfo
from .config import PhitsConfig

#===============================================================================
# Phits Input class
#===============================================================================
class Phits(McnpGeometry):
	MODE    = PhitsCard.MODE

	PAT_SECTION = re.compile(r"^(\[.*?\])(off)?$")
	PAT_SET     = re.compile(r"\s*([A-Za-z0-9_]+)\[(.*?)\]")
	PAT_KEYNUM  = re.compile(r"^([A-Za-z]+)(\d+) *(.*)", re.DOTALL | re.IGNORECASE)
	PAT_MAT     = re.compile(r"^mat\[ *(\d+) *\] *(.*)", re.DOTALL | re.IGNORECASE)
	PAT_INFL    = re.compile(r"^infl:{(.*?)}", re.IGNORECASE)
	#PAT_VAR     = re.compile(r"^\w[A-Za-z0-9()[]-]*\s*=.*")

	# Mesh patterns
	PAT_ETYPE   = re.compile(r"\b[etxyzral]-type\b", re.IGNORECASE)
	PAT_ENUM    = re.compile(r"\bn([etxyzral])\b",   re.IGNORECASE)
	PAT_ELIMIT  = re.compile(r"\b([etxyzral])(min|max|del)\b", re.IGNORECASE)

	config      = PhitsConfig()		# configuration information
	defaultMaterials = OrderedDict([
				("-1", PhitsCard("/[material]/mat",[-1])),	# black hole
				( "0", PhitsCard("/[material]/mat",[ 0]))	# vacuum
			])
	zaid = []		# list of zaid
	sab  = []		# list of Sab elements

	_paren_trans = str.maketrans("()","  ")

	# ----------------------------------------------------------------------
	def __init__(self):
		super().__init__()
		self.mode       = Phits.MODE
		self.globalDict = PhitsCard.globalDict

	#-----------------------------------------------------------------------
	# Return input filename
	#-----------------------------------------------------------------------
	def filenames(self, filename):
		files = super().filenames(filename)
		return files

	#-----------------------------------------------------------------------
	# Cards
	#-----------------------------------------------------------------------
	# Return cards of a specific tag sorted with _pos
	# @param tag     string tag
	#                list of tags
	#                Card.type(), BODY, SOLID, VOLUME...
	# @param which   look the Input.cardsSorted()
	#-----------------------------------------------------------------------
	#def cardsSorted(self, tag, which=0):
	#	return super().cardsSorted(tag, which)

	#-----------------------------------------------------------------------
	# Return material list
	#-----------------------------------------------------------------------
	def materialList(self, which=0, *kw):
		lst = super().materialList(which)
		lst.extend([x.name()
			for x in self.cardsSorted(Card.MATERIALS,-1).values()])
		return lst

	#-----------------------------------------------------------------------
	# Return the region properties as a dictionary of the active/enabled
	# regions
	#-----------------------------------------------------------------------
	def regionProperties(self):
		#matList = self.materialList()
		matDict = self.materialDict()
		#rotDefi = {}

		# special regions
		blckhole = matDict["-1"]
		#vacuum   = matDict["0"]

		#---------------------------------------------------------------
		# Build dictionaries and clean up regions from Input and voxel
		n = 0
		regionList = []
		for region in self.cardsSorted(Card.REGION,-1).values():
			material = matDict.get(str(region.what(1)),None)
			n += 1
			region["@n"] = n		# Assign default values to region
			region[Card.TYPE]     = int(material is blckhole)
			region[Card.ASSIGNMAT]= region	# Card that assigned the material
			region[Card.MATERIAL] = material
			#region["@biasingA"]  = None	# All particles biasing
			#region["@biasingH"]  = None	# Hadrons
			#region["@biasingE"]  = None	# emf
			#region["@biasingN"]  = None	# low-neutrons
			#region[Card.LATTICE] = None
			#region["@rotdefi"]   = ""
			#region[Card.CONT]    = False	# has continuation cards
			regionList.append(region)

	#=======================================================================
	# I/O
	# methods for Input IO operations
	#=======================================================================
	#-----------------------------------------------------------------------
	# Export method
	#-----------------------------------------------------------------------
	def _save(self):
		Global.dump("\n*-* Writing phits file")

		section = None		# current section
		for card in self.cardlist:
			if not isinstance(card, PhitsCard): continue
			if card.ignore(): continue
			if card.tag[0] == "[":
				section = card.tag
				if section != "[title]":
					self.writeline()
			self.writeline(card.toStr(comments=True))

	#-----------------------------------------------------------------------
	# Import/load a macro and the related moira input files
	#-----------------------------------------------------------------------
	def _load(self):
		Global.dump("\n*-* Reading phits file")

		#---------------------------------------------------------------
		# all pending commands from section merge to create macro commands
		#---------------------------------------------------------------
		def sectionMacroCommands():
			nonlocal section, mesh
			if section is None: return
			for tag, args in mesh.items():
				if isinstance(args,PhitsCard): continue
				try:
					pat = Phits.PAT_ENUM.match(tag)
					# Check for ne, nx, n?
					if pat is not None:
						e = pat.group(1)
						card = mesh[f"{e}-type"]
						card.setWhat(1, args[0])
						continue
					# Check for e{del,min,max} pattern
					pat = Phits.PAT_ELIMIT.match(tag)
					if pat is not None:
						e = pat.group(1)
						card = mesh[f"{e}-type"]
						if pat.group(2)=="max":
							card.setWhat(3, args[0])
						else:
							card.setWhat(2, args[0])
						continue
				except KeyError:
					log.error(f"In Section '{section.tag}' '{tag}'"
						f"card found with no corresponding '{e}-type'")
				assert False
			mesh.clear()

		#---------------------------------------------------------------
		# Parse a command line
		#---------------------------------------------------------------
		def parseCmdLine(cmdline, cardcomment):
			nonlocal card, section, cont_space, comment_char, add_child, mesh, c_comment
			# FIXME qp: within a section place #if 0 .. #endif
			# FIXME q: from this point to the end of the file
			if cmdline[0] == '[':
				cmdline = cmdline.lower().replace(" ","")
				m = Phits.PAT_SECTION.match(cmdline)
				#print(f"CMDLINE={cmdline!r} {m=}")
				# call str.title() to capitalize every word first letter
				name = m.group(1) #.title()
				sectionMacroCommands()
				section = PhitsCard(
					PhitsCardInfo.sectionsMap.get(name,name), None)
				self.add(section)
				if m.group(2)=="off":
					section.disable()
				if name in ("[surface]", "[cell]"):
					comment_char = "$"
					add_child = False
				else:
					cont_space = name in ("[material]", "[transform]")
					comment_char = "#%!$"
					if name == "[material]":
						c_comment = self._icommat
					add_child = name not in ("[parameters]", "[transform]")
			elif cmdline.startswith("set:"):
				self.__parseSet(cmdline, cardcomment)
			elif section is None:
				log.error(f"Line {self._lineNo[-1]}: "
					"Command with no section")
				log.errorCont(cmdline)
			elif section.tag == "[title]":
				self.cardlist[-1].appendText(cmdline)
			elif section.tag == "[surface]":
				self.__parseSurface(cmdline, cardcomment)
			elif section.tag == "[cell]":
				self.__parseCell(cmdline, cardcomment)
			elif section.tag == "[transform]":
				self.__parseTransform(cmdline, cardcomment)
			elif section.tag == "[material]":
				self.__parseMaterial(cmdline, cardcomment)
			elif section.tag == "[matnamecolor]":
				# FIXME create cards for material colors
				self.cardlist[-1].appendText(cmdline)
			elif section.tag == "[volume]":
				# FIXME create cards for material colors
				self.cardlist[-1].appendText(cmdline)
			else:
				# FIXME special cases to be handled here
				# Check for possible continuation
				if "=" not in cmdline:
					assert card is not None
					card.appendText(cmdline)
					return

				tag, args = self.__parseTokenize(cmdline)
				# Check for special mesh groups
				if section.tag.startswith("[t-") and \
				   (Phits.PAT_ENUM.match(tag) or Phits.PAT_ELIMIT.match(tag)):
					try:
						pat = Phits.PAT_ENUM.match(tag)
						# Check for ne, nx, n?
						if pat is not None:
							e = pat.group(1)
							card = mesh[f"{e}-type"]
							card.setWhat(1, args[0])
							return
						# Check for e{del,min,max} pattern
						pat = Phits.PAT_ELIMIT.match(tag)
						if pat is not None:
							e = pat.group(1)
							card = mesh[f"{e}-type"]
							if pat.group(2)=="max":
								card.setWhat(3, args[0])
							else:
								card.setWhat(2, args[0])
							return
					except KeyError:
						log.error(f"In Section '{section.tag}' '{tag}'"
							f"card found with no corresponding '{e}-type'")
					assert False
				else:
					# Create card
					card = PhitsCard(f"/{section.tag}/{tag}",
							args, cardcomment)
					if add_child:
						section.add(card)
					else:
						self.add(card)
					if not section.enable: card.disable()
					# ?-type card
					if Phits.PAT_ETYPE.search(tag):
						mesh[tag] = card

		#---------------------------------------------------------------
		cont          = False		# continuation with suffix character
		cont_space    = False		# continuation with space prefix
		c_comment     = True		# c as comment
		comment_char  = "#%!$"		# comment characters
		#file_off     = False		# skip until end of file
		self._icommat = False		# c as comment in material section
		self._dennuc  = False		# materials den + nuc instead of the opposite
		section       = None		# current section
		add_child     = False		# add as child
		mesh          = {}		# merge mesh variables e-type,ne,emin,emax,edel
		cmdline       = ""		# current command line
		card          = None		# last card added
		cardcomment   = []		# comment of the card
		comment       = []		# comments at the end of the card which are
						# meant for the next one
						# first populate the last comment and if the card
						# continuous append it to the comment
		while True:
			line = self.readline()
			if line is None: break

			# Empty line
			if not line: continue
			Global.dump(">>>line=",line, level=2)

			# Check for leading comments
			if line[0] in "#%!$":
				comment.append(line[1:])
				continue
			if c_comment and line[:2] in ("c ","C "):
				comment.append(line[2:])
				continue
			if line[0] == "[":
				# new section restore defaults
				cont_space     = False
				c_comment      = True

			# Strip inline !, $ and # comments
			for ch in comment_char:
				try:
					pos = line.index(ch)
					comment.append(line[pos+1:])
					line = line[:pos].rstrip()
				except ValueError:
					pass
				if not line: break
			if not line: continue

			#if not cont_space:
			#	pos = line.find("#")
			#	if pos>=0:
			#		comment.append(line[pos+1:])
			#		line = line[:pos].rstrip()
			#		if not line: continue

			# handle possible line continuation
			if cont or (cont_space and line[:5]=="     " or line[0]=="\t"):
				if line[-1]=="&":
					line = line[:-1].rstrip()
					cont = True
				else:
					cont = False
				# Continuation
				cmdline += f"\n{line.strip()}"
				if comment:
					cardcomment.extend(comment)
					comment = []
				continue

			# process line
			line = line.strip()
			if line[-1] in "&\\":
				line = line[:-1].rstrip()
				cont = True
			else:
				cont = False
				if cmdline:
					parseCmdLine(cmdline, "\n".join(cardcomment))
				cardcomment = comment
				comment = []
				if line.startswith("infl:"):
					self.__parseIncludeFile(line,"\n".join(cardcomment))
					cmdline = ""
				else:
					cmdline = line

		# execute last command
		if cmdline:
			parseCmdLine(cmdline, "\n".join(cardcomment))
		sectionMacroCommands()

	#-----------------------------------------------------------------------
	# Parse set variable command => converted to flair '#define'
	#-----------------------------------------------------------------------
	def __parseSet(self, command, comment):
		command = command.strip()[4:]	# strip 'set:'
		while command:
			m = Phits.PAT_SET.search(command)
			if m:
				name = m.group(1)
				expr = m.group(2)
				try:
					float(expr)
				except ValueError:
					expr = f"={expr}"
				command = command[m.end():]
				#print(f"Phits._parseSet {name=} {expr=} {command=}")
				self.add(Card("#define",[name,expr],comment))
				comment = ""
			else:
				break

	#-----------------------------------------------------------------------
	def __parseTokenize(self, command):
		args = command.replace("="," ").split()
		tag = args.pop(0).lower()
		if "(" in tag:
			tag,*arg = tag.translate(Phits._paren_trans).split()
			return tag, arg + args
		else:
			return tag, args

	#-----------------------------------------------------------------------
	def __parseIncludeFile(self, command, comment):
		m = Phits.PAT_INFL.match(command)
		if m:
			self.add(PhitsCard("infl",[m.group(1)],comment))
			self.open(m.group(1),"r","infl")

	#-----------------------------------------------------------------------
	def __parseSurface(self, command, comment):
		Global.dump(">S>",repr(command))
		token = command.split()
		try:
			name = token.pop(0)	# split into 2 instructions
		except ValueError:
			log.error(f"Line {self._lineNo[-1]}: "
				 f"Surface '{name}' invalid id")
			log.errorCont(command)
			return False
		surface = token.pop(0)
		try:
			prop = {"tr" : str(int(surface))}
			surface = token.pop(0)
		except ValueError:
			prop = None
		arg = [name] + list(map(Phits._floatOrExpr, token))
		# TODO (from MCNP?)
		#   * If a surface is preceded by an asterisk * then an reflective
		#     surface is created
		self.add(PhitsCard(surface.upper(),arg,comment,prop=prop))

	#-----------------------------------------------------------------------
	def __parseCell(self, command, comment):
		Global.dump(">C>",repr(command))
		name,mat,expr = command.split(maxsplit=2)
		if mat.lower() == "like":
			like,_,expr = expr.split(maxsplit=2)
			_,prop = self._parseProperties(expr)
			self.add(PhitsCard("CELL_LIKE",[name,like],comment,prop=prop))
		else:
			if mat not in ("0","-1"):
				density,expr = expr.split(maxsplit=1)
			else:
				density = ""
			expr,prop = self._parseProperties(expr)
			self.add(PhitsCard("CELL",[name,mat,density],comment,expr,prop=prop))

	#-----------------------------------------------------------------------
	def __parseTransform(self, command, comment):
		Global.dump(">T>",repr(command))
		m = McnpGeometry.PAT_TRANS.match(command)
		if m:
			deg = "*" if "*" in (m.group(1), m.group(3)) else ""
			name = m.group(2)
			try:
				whats = list(map(Phits._floatOrExpr, m.group(4).split()))
				whats.insert(0,name)
			except:
				log.error(f"Line {self._lineNo[-1]}: "
					f"tr{m.group(2)} parsing transformation '{command}'")
				log.errorCont(command)
			self.add(PhitsCard(f"TR{deg}", whats))
			return

	#-----------------------------------------------------------------------
	def __parseMaterial(self, command, comment):
		Global.dump(">M>",repr(command))
		if command.startswith("den ") and command.split()[1]=="nuc":
			# FIXME add a PhitsCard()
			self._dennuc = True
			return

		Global.dump("C",comment)
		Global.dump(">>>",command)
		pat = Phits.PAT_MAT.match(command)
		if pat:
			card = PhitsCard("/[material]/mat", [pat.group(1)], comment)
			self.add(card)
			command = pat.group(2)
		else:
			pat = Phits.PAT_KEYNUM.match(command)
			if pat:
				cmd = pat.group(1).upper()
				if cmd=="M":
					card = PhitsCard("/[material]/mat",
							[pat.group(2)], comment)
					self.add(card)
					command = pat.group(3)
				else:
					return

		if self.cardlist[-1].tag != "/[material]/mat": return
		words = command.split()
		for a,b in zip(words[0::2], words[1::2]):
			if self._dennuc: a,b = b,a
			card.add(PhitsCard("/[material]/mat/elem", [a, b]))
			#self.cardlist[-1]._what.append(a)
			#self.cardlist[-1]._what.append(b)

	# ----------------------------------------------------------------------
	# Create bare minimum input file
	# ----------------------------------------------------------------------
	def minimumInput(self):
		super().minimumInput()
		self.add(PhitsCard("[title]"))
		self.add(PhitsCard("[parameters]"))
		self.add(PhitsCard("[source]"))
		self.add(PhitsCard("[material]"))
		self.add(PhitsCard("[cell]"))
		self.add(PhitsCard("[surface]"))
		self.add(PhitsCard("[end]"))
		self.renumber()

	#-----------------------------------------------------------------------
	@staticmethod
	def _floatOrExpr(value):
		try:
			return float(value)
		except:
			return f"={value}"

	#-----------------------------------------------------------------------
	@classmethod
	def newCard(cls, *args, **kwargs):
		return PhitsCard(*args, **kwargs)
