# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   4-Sep-2006

# version should be updated with the *update_version.r* program
# which updates the relative tkFlair, flair.spec, geoviewer.cc and geoviewer.spec
# The __revision__, __lastchange__ is updated from the packager otherwise on
# local runs from the current git repository

__author__     = "Vasilis Vlachoudis (lead author)," \
                 "Christian Theis," \
                 "Wioletta Kozlowska"
__email__      = "Vasilis.Vlachoudis@cern.ch"
__version__    = "3.3-0.1"

__program__    = "flair"
__created__    = "Jun 2006"
__www__        = "http://flair.cern"
__revision__   = "b919bc35"
__refname__    = ""
__lastchange__ = "Sat Nov 4 14:55:31 2023 +0100"

import os
import re
import sys
import stat
import time
import shutil
import gettext
import builtins
import traceback
import subprocess
import webbrowser
import configparser

import tkinter as tk
import tkinter.font as tkFont
from tkconst import STATE
from tkinter import ttk, messagebox

import log
import tool
import icons
import Global
import Ribbon
import palette
import Gnuplot
import Project
import tkExtra
import RichText
import tkDialogs
import bFileDialog
import tkFileViewer

from config   import Config
from deck     import Input, Card, CardInfo, DeckConfig
from fluka    import Fluka, FlukaConfig
from mcnp     import Mcnp,  McnpConfig
from moira    import Moira, MoiraConfig
from penelope import Penelope, PenelopeConfig
from phits    import Phits, PhitsConfig

# --- Global Variables ---
prgDir      = os.path.dirname(os.path.abspath(__file__))
iniDir      = os.path.expanduser(f"~/.config/{__program__}")
iniFile     = os.path.join(iniDir, f"{__program__}.ini")
iniMaterial = os.path.join(iniDir, "material.ini")

# dirty way of substituting the "_" on the builtin namespace
builtins._  = gettext.translation('flair',
				os.path.join(prgDir,'locale'),
				fallback=True).gettext
builtins.N_ = lambda message: message

# Variables
config           = None			# Global configuration (ini)
style            = None			# ttk global style
_SKIP_INTRO      = False
_SAVE_FRAME      = False
_errorReport     = True
_showFlukaFiles  = False
_gnuplotTerminal = ""
_maxRecent       = 10
_showToolbar     = True
_showStatusbar   = True
_autoSaveTime    = 120			# auto save to input.pickle every # seconds
dpiscale         = 1			# DPI scaling (integer value)

# --- Sections ---
_BATCH_SECTION       = Project.RunInfo.section
_COLOR_SECTION       = "Color"
_DATA_SECTION        = Project.Command.section
_FLAIR_SECTION       = "flair"
_FONT_SECTION        = "Font"
_GEOMETRY_SECTION    = "Geometry"
_INPUT_SECTION       = "Input"
_LAYERS_SECTION      = "Layers"
_MATERIAL_SECTION    = Project.MaterialInfo.CLASS
_PAGE_SECTION        = "Page"
_PLOT_SECTION        = "Plot"
_PROJECT_SECTION     = Project.section
_STYLE_SECTION       = "Style"
_ORIENTATION_SECTION = "Orientation"

__sections = [	_BATCH_SECTION,
		_COLOR_SECTION,
		_DATA_SECTION,
		_FLAIR_SECTION,
		_FONT_SECTION,
		_GEOMETRY_SECTION,
		_INPUT_SECTION,
		_LAYERS_SECTION,
		_MATERIAL_SECTION,
		_PAGE_SECTION,
		_PLOT_SECTION,
		_PROJECT_SECTION,
		_STYLE_SECTION,
		_ORIENTATION_SECTION]

# --- Global Fonts ---
_TREE_FONT  = ("helvetica", "11")
_TITLE_FONT = ("helvetica", "12", "bold")

# --- Global Colors ---
_TITLE_FOREGROUND_COLOR = None
_TITLE_BACKACTIVE_COLOR = None
_TITLE_BACKINACTIVE_COLOR = None

_ELABEL_FOREGROUND_COLOR = None		# Entry label
_ELABEL_BACKGROUND_COLOR = None

_ILABEL_FOREGROUND_COLOR = None		# Info label
_ILABEL_BACKGROUND_COLOR = None

_CLABEL_FOREGROUND_COLOR = None		# Calculated label
_CLABEL_BACKGROUND_COLOR = None

_LABEL_ENTRY_COLOR = "DarkGray"

# Special tags
_ALL_GROUP    = "*all*"

_NOTEBOOK_STYLE    = "InputEditor.TNotebook"
_NOTEBOOKTAB_STYLE = _NOTEBOOK_STYLE + ".Tab"
_TREEVIEW_STYLE    = "InputEditor.Treeview"

# Arrays
errors = []

# Plot formats types .eps, .ps, ... read from .ini
_PLOT_FORMAT   = {}
_PLOT_COMMANDS = ""

# Flair clipboard pattern
_CARD_MIME     = f"{__program__}/card"
_RUN_MIME      = f"{__program__}/run"
_PLOT_MIME     = f"{__program__}/plot"
_USR1D_MIME    = f"{__program__}/usr1d"
_LAYER_MIME    = f"{__program__}/layer"

NOTIFY_NORMAL  = 0
NOTIFY_WARNING = 1
NOTIFY_ERROR   = 2

_TEMPLATES_DIR = "templates"
_PALETTE_DIR   = "palette"
_DEFAULT_INP   = "basic.flair"

_SEND_EMAIL    = "http://flair.web.cern.ch/flair/cgi-bin/send_email.php"

_FONT_SIZE = (   '6',  '7',  '8',  '9', '10', '11', '12', '13',
		'14', '16', '18', '20', '22', '24', '26', '28',
		'32', '36', '40', '48', '56', '64', '72' )

_FONT_LIST = [
		"balloon",
		"input.comment",
		"input.fixed",
		"input.hidden",
		"input.label",
		"input.tag",
		"input.value",
		"ribbon",
		"ribbontab",
		"title",
		"tree",
		"TkDefaultFont",
		"TkFixedFont",
		"TkMenuFont",
		"TkTextFont" ]

_ON    = " \N{BALLOT BOX WITH X} "
_OFF   = " \N{BALLOT BOX} "
_ON    = " \N{BLACK LARGE SQUARE} "
_OFF   = " \N{WHITE LARGE SQUARE} "
_ON    = " \N{BLACK SQUARE} "
_OFF   = " \N{WHITE SQUARE} "
_MULTI = " \N{DIE FACE-6} "

_bodiesLayout = [
	("Plane",    ""    ),
	(  "&YZP",   "X&ZP"),
	(  "&XYP",   "&PLA"),
	("Box",      ""    ),
	(  "&RPP",   "&BOX"),
	(  "&WED",   "ARB" ),
	("Sphere",   ""    ),
	(  "&SPH",   "E&LL"),
	("Cylinder", ""    ),
	(  "R&CC",   "R&EC"),
	(  "XCC",    "XEC" ),
	(  "YCC",    "YEC" ),
	(  "ZCC",    "ZEC" ),
	("Cone",     ""    ),
	(  "&TRC",    ""   ),
	("Quad",     ""    ),
	(  "&QUA",    ""   )]
if Global.developer:
	_bodiesLayout.extend([
		("Torus",    ""   ),
		(  "TRX",    "TRY"),
		(  "TRZ",    ""   )])

_surfaceLayout = [
	("Plane",     "",       ""   ),
	(  "P&X",     "P&Y",    "P&Z" ),
	(  "&P",      "P&3",    ""   ),
	("Box",       "",       ""   ),
	(  "&RPP",    "&BOX",   "WED"),
	(  "ARB",     "",       ""   ),
	("Sphere",    "",       ""   ),
	(  "&S",      "S&O",    "SPH"),
	(  "SX",      "SY",     "SZ" ),
	(  "ELL",     "",       ""   ),
	("Cylinder",  "",       ""   ),
	(  "C/X",     "C/Y",    "C/Z"),
	(  "CX",      "CY",     "CZ" ),
	(  "RCC",     "REC",    ""   ),
	("Cone",      "",       ""   ),
	(  "K/X",     "K/Y",    "K/Z"),
	(  "KX",      "KY",     "KZ" ),
	(  "TRC",     "",       ""   ),
	("Quad",      "",       ""   ),
	(  "S&Q",     "&GQ",    ""   ),
	("Torus",     "",       ""   ),
	(  "TX",      "TY",     "TZ" ),
	("Generic",   "",       ""   ),
	(  "X",       "Y",      "Z"  ),
	("Prism",     "",       ""   ),
	(  "RHP",     "HEX",    ""   )]

# vim: setlocal tw=70
LICENSE_VERSION = 300
LICENSE = """{h1:FLAIR SINGLE USER LICENCE AGREEMENT}
{c:Last updated: 4 December 2019}

{b:PLEASE READ THIS SINGLE USER LICENCE AGREEMENT FOR FLAIR CAREFULLY:
This Licence is an agreement between You and the European Organization
for Nuclear Research (“CERN”) and governs Your Use of the Software (as
defined below).

Flair Version 3, an advanced user-friendly interface suitable for
FLUKA and other simulation codes, results from the work performed by
Vasilis Vlachoudis and other contributors. The copyright in Flair
Version 3 is entirely vested in CERN.

By clicking on “I Agree” or by Using the Software, You agree to be
bound by the terms of this Licence. If You do not agree with the terms
of this Licence, You may not Use the Software.}

{h2:1. Definitions}
{u:Copy} means reproduction in any material form, and storage in any medium.
“Copies” shall be interpreted accordingly;

{u:Export Control Regulations} means Laws that control the export, transit
and brokering of dual-use items, i.e. goods, software and technology that
can be used for both civilian and military applications;

{u:Flair} means version 3 of the Flair computer programme, an advanced
user-friendly interface suitable for FLUKA and other simulation codes,
available from http://flair.cern;

{u:Laws} means all CERN rules and regulations and all laws, treaties, rules,
regulations and orders of any local, national, or other authority having
jurisdiction over You, relevant for the execution of this Licence;

{u:Make Available} means to market, distribute, transfer, license or
sub-license, or in any way dispose of or make available;

{u:Modifications} means any modification, enhancement, adaptation,
translation, abridgement, condensation, revision or re-write. “Modifies” and
“Modified” shall be interpreted accordingly;

{u:Software} means Flair, as well as accompanying documentation and any other
information CERN may provide to You in relation thereto;

{u:Use} means to download, install, Copy, access or otherwise use Flair.
“Using” shall be interpreted accordingly;

{u:You} means the individual licensing the Software under this Licence.
“Your” shall be interpreted accordingly.

{h2:2. Licence Grant}
CERN hereby grants You a personal, non‑transferable, non‑sublicensable,
non‑exclusive, royalty-free source code licence to Use the Software for
internal scientific, non-military and non‑commercial purposes only. You may
not Modify nor Make Available the Software.

CERN reserves all rights not expressly granted to You under this Licence.
Without prejudice to the generality of the foregoing CERN grants no rights
to any intellectual property other than the copyright in the Software.

{h2:3. Technical Assistance}
CERN operates a User Forum to provide assistance related to the Use of
the Software. You shall address all requests for assistance through such
FLUKA User Forum at CERN, available at https://fluka-forum.web.cern.ch

{h2:4. Your Obligations}
You shall use the Software in accordance with best practice and
professional standards and always for non-military purposes only.

In exercising Your rights under this Licence, You shall comply with all
applicable Laws and Export Control Regulations, and shall at all times be
solely responsible and liable for such due compliance.

You shall report as soon as practically possible any errors or bugs found
in any portion of Software to the FLUKA User Forum at CERN.

{h2:5. Limitations and Restrictions}
You shall not:
  (a) Make Available the Software to, or allow the Use of the Software
      by, any other person or entity;

  (b) Modify the Software;

  (c) remove, modify, or conceal any copyright notice or other marks on
      or within the Software; and

  (d) transfer, sublicense, or assign Your rights under this Licence to
      any other person or entity.

{h2:6. Acknowledgement}
You shall explicitly acknowledge Your Use of the Software in any
publication or communication, scientific or otherwise, relating to such Use,
by citing the Flair set of references found in the Flair manual.

You shall ensure that the Software set of references, the relevant
copyright notices and the terms of this Licence are not altered or removed
from the Software and that all embodiments of the Software contain in full
the Software set of references, the relevant copyright notice, and the terms
of this Licence.

{h2:7. Warranties and Limitation of Liability}
THE SOFTWARE, AND ANY TECHNICAL ASSISTANCE, AS THE CASE MAY BE, IS
PROVIDED “AS IS”, AND CERN DOES NOT PROVIDE ANY EXPRESS OR IMPLIED
WARRANTIES OF ANY KIND CONCERNING THE SAME, INCLUDING BUT NOT LIMITED
TO NON-INFRINGEMENT OF RIGHTS OF THIRD PARTIES, MERCHANTABILITY,
SATISFACTORY QUALITY OR FITNESS FOR A PARTICULAR PURPOSE. CERN DOES
NOT PROVIDE ANY WARRANTIES REGARDING THE CONTINUOUS OR FUTURE
PROVISION OF THE SOFTWARE, ANY UPDATE OR NEW VERSION, OR ANY TECHNICAL
ASSISTANCE RELATED THERETO.

CERN DOES NOT ACCEPT ANY LIABILITY WHATSOEVER FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL, CONSEQUENTIAL, EXEMPLARY, PUNITIVE OR OTHER
DAMAGES OF ANY CHARACTER INCLUDING, WITHOUT LIMITATION, DAMAGES FOR
LOSS OF DATA, OR INTERRUPTION OF ACTIVITIES, ARISING OUT OF OR RELATED
TO (I) YOUR USE OF OR INABILITY TO USE THE SOFTWARE, HOWEVER CAUSED,
REGARDLESS OF THE THEORY OF LIABILITY (CONTRACT, TORT, OR OTHERWISE)
AND EVEN IF CERN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES,
OR (II) YOUR COMPLIANCE OR NON-COMPLIANCE WITH LAWS AND EXPORT CONTROL
REGULATIONS. YOU SHALL HOLD CERN FREE AND HARMLESS FROM AND INDEMNIFY
IT AGAINST ALL CLAIMS AND LIABILITY WITH REGARD TO THE SUBJECT MATTER
OF THIS SECTION.

{h2:8. Duration and Termination}
This Licence is effective upon You clicking on “I Agree” or Using the
Software, and shall remain in effect until expired or otherwise
terminated.

This Licence expires two years after the release date of the Software.
On such expiration date, You will no longer be able to Use the
Software.

You may terminate this Licence at any time by ceasing Use of or
destroying all copies of the Software. This Licence will terminate
automatically and without notice if You fail to comply with any of its
terms.

Upon expiration or termination of this Licence, You shall destroy all
copies of the Software in Your possession or control.

{h2:9. Survival}
Sections 5 (Limitations and Restrictions), 6 (Acknowledgement),
7 (Warranties and Limitation of Liability), and 10 (Applicable Law and
Dispute Settlement) shall survive expiration or termination of this
Licence.

{h2:10. Applicable Law and Dispute Settlement}
The terms of this Licence shall be interpreted in accordance with
their true meaning and effect, and independently of any national or
other laws. Without prejudice to CERN’s status as an Intergovernmental
Organization, reference shall be made to Swiss substantive law where:
(a) a matter is not specifically covered by this Licence; or (b) a
term of this Licence is ambiguous or unclear.

Reference to Swiss substantive law shall be made exclusively for the
matter or the term concerned, and shall in no event apply to the other
terms of this Licence.

Any dispute under this Licence that fails to be settled amicably shall
be referred to arbitration on the basis of CERN’s status as an
Intergovernmental Organization, and in accordance with the procedure
defined at:  http://legal.web.cern.ch/procedures/arbitration

BY CLICKING ON {b:“I AGREE”} YOU ACKNOWLEDGE THAT YOU HAVE READ,
UNDERSTAND AND AGREE TO BE BOUND BY THIS LICENCE.
"""

#-------------------------------------------------------------------------------
# Find revision from git
#-------------------------------------------------------------------------------
def findRevision():
	global __revision__, __lastchange__, __refname__
	if __revision__: return
	# Find revision from git
	try:
		__revision__,__lastchange__,refs = subprocess.check_output(
					["git","-C",prgDir,"log","-1","--format=%h|%cd|%D"],
					stderr=subprocess.DEVNULL
				).decode(errors="replace").rstrip().split("|")
		__refname__ = refs.split()[-1]
	except:      # subprocess.CalledProcessError does not catch the case if git is not installed!
		__revision__ = "unknown"
		__lastchange__ = "unknown"

#-------------------------------------------------------------------------------
# Create a font string
#-------------------------------------------------------------------------------
def fontString(font):
	name  = str(font[0])
	size  = str(font[1])
	if name.find(' ')>=0:
		s = f'"{name}" {size}'
	else:
		s = f"{name} {size}"

	try:
		if font[2] == tkFont.BOLD: s += " bold"
	except: pass
	try:
		if font[3] == tkFont.ITALIC: s += " italic"
	except: pass
	return s

#-------------------------------------------------------------------------------
# @return font dictionary from current config
#-------------------------------------------------------------------------------
def fontDictionary():
	global config
	fontDict = {}

	# and populate font list with extra fonts
	for n,v in config.items(_FONT_SECTION):
		if n not in _FONT_LIST:
			_FONT_LIST.append(n)

		vt = tuple(v.split(','))
		try:    size = int(vt[1])
		except: size = 12

		weight = tkFont.NORMAL
		slant  = tkFont.ROMAN

		try:
			if vt[2].find(tkFont.BOLD)>=0:
				weight = tkFont.BOLD

			# FIXME old version
			if vt[2].find(tkFont.ITALIC)>=0:
				slant = tkFont.ITALIC
		except:
			pass

		try:
			slant = vt[3]
		except:
			try:
				if vt[2].find(tkFont.ITALIC)>=0:
					slant = tkFont.ITALIC
			except:
				pass

		fontDict[n] = (vt[0], size, weight, slant)
	return fontDict

#-------------------------------------------------------------------------------
# Update config with fonts from dictionary
#-------------------------------------------------------------------------------
def updateConfigFonts(fontDict):
	section = _FONT_SECTION
	for n,v in fontDict.items():
		if not v[0]:	# Delete font
			try:
				config.remove_option(section, n)
			except: pass
		else:
			config.updateFont(section, n, v)

#-------------------------------------------------------------------------------
# Increase/decrease fonts by incr
#-------------------------------------------------------------------------------
def fontIncrease(incr):
	fontDict = fontDictionary()
	for n,v in fontDict.items():
		if isinstance(v,tuple):
			family, size, *rest = v
			if size<0:
				size -= incr
				if size > 5: size = -5
			else:
				size += incr
				size = max(size, 5)
			fontDict[n] = (family, size, *rest)
	updateConfigFonts(fontDict)

#-------------------------------------------------------------------------------
# Add Recent
#-------------------------------------------------------------------------------
def addRecent(filename):
	sfn = str(os.path.abspath(filename))
	if not os.path.isfile(sfn): return
	try:
		last = _maxRecent-1
		for i in range(_maxRecent):
			rfn = getRecent(i)
			if rfn is None:
				last = i-1
				break
			if rfn == sfn:
				if i==0: return
				last = i-1
				break
	except configparser.NoOptionError:
		pass

	# Shift everything by one
	for i in range(last, -1, -1):
		config.set(__program__,
				f"recent.{i+1}",
				getRecent(i))
	config.set(__program__, "recent.0", sfn)

#-------------------------------------------------------------------------------
def getRecent(recent):
	try:
		return config.get(__program__,f"recent.{recent}")
	except configparser.NoOptionError:
		return None

#-------------------------------------------------------------------------------
# Open and Read Initialization file
#-------------------------------------------------------------------------------
def openIni(ini=None, loaduserini=True):
	global config, iniDir, iniFile

	systemIni = os.path.join(prgDir, f"{__program__}.ini")

	if ini is not None:
		iniFile = os.path.abspath(ini)
		iniDir  = os.path.dirname(iniFile)

	config = Config()
	if loaduserini:
		config.read([systemIni, iniFile], encoding="UTF-8")
	else:
		config.read(systemIni, encoding="UTF-8")

	# Check for old configuration file....
	try:
		config.get(__program__, "version")
		# Check version if needed
	except:
		# Patch...
		config.read(systemIni, encoding="UTF-8")

	# Add the necessary sections
	for s in __sections:
		try: config.add_section(s)
		except: pass

#-------------------------------------------------------------------------------
# Read Initialization file
#-------------------------------------------------------------------------------
def readIni(gui=True):
	global _SKIP_INTRO, _SAVE_FRAME
	global _TREE_FONT, _TITLE_FONT
	global _TITLE_FOREGROUND_COLOR, _TITLE_BACKACTIVE_COLOR, _TITLE_BACKINACTIVE_COLOR
	global _ELABEL_FOREGROUND_COLOR, _ELABEL_BACKGROUND_COLOR
	global _ILABEL_FOREGROUND_COLOR, _ILABEL_BACKGROUND_COLOR
	global _CLABEL_FOREGROUND_COLOR, _CLABEL_BACKGROUND_COLOR
	global _PLOT_COMMANDS
	global _showFlukaFiles, _errorReport
	global _gnuplotTerminal, _autoSaveTime
	global _showToolbar, _showStatusbar
	global dpiscale

	import Manual

	if gui:
		section = _FONT_SECTION
		# Read basic fonts from the font section
		_TREE_FONT      = config.getFont(section, "tree",      _TREE_FONT)
		_TITLE_FONT     = config.getFont(section, "title",     _TITLE_FONT)
		Ribbon._TABFONT = config.getFont(section, "ribbontab", Ribbon._TABFONT)
		Ribbon._FONT    = config.getFont(section, "ribbon",    Ribbon._FONT)

		# read Tk fonts to initialize them
		config.getFont(section, "TkDefaultFont")
		config.getFont(section, "TkFixedFont")
		config.getFont(section, "TkMenuFont")
		config.getFont(section, "TkTextFont")
		tkExtra.Balloon.font = config.getFont(section, "balloon", tkExtra.Balloon.font)

		section = _COLOR_SECTION
		_TITLE_FOREGROUND_COLOR  = config.getColor(section, "title.foreground",   "white")
		_TITLE_BACKACTIVE_COLOR  = config.getColor(section, "title.background",   "dark slate blue")
		_TITLE_BACKINACTIVE_COLOR= config.getColor(section, "title.backinactive", "darkgray")

		_ELABEL_FOREGROUND_COLOR = config.getColor(section, "entrylabel.foreground", "darkblue")
		_ELABEL_BACKGROUND_COLOR = config.getColor(section, "entrylabel.background", "white")

		_ILABEL_FOREGROUND_COLOR = config.getColor(section, "infolabel.foreground", "darkblue")
		_ILABEL_BACKGROUND_COLOR = config.getColor(section, "infolabel.background", "lightgray")

		_CLABEL_FOREGROUND_COLOR = config.getColor(section, "calclabel.foreground", "darkred")
		_CLABEL_BACKGROUND_COLOR = config.getColor(section, "calclabel.background", "lightgray")

		_LABEL_ENTRY_COLOR       = config.getColor(section, "labelentry.foreground", "darkgray")

		tkExtra.Balloon.foreground = config.getColor(section, "balloon.foreground",
								tkExtra.Balloon.foreground)
		tkExtra.Balloon.background = config.getColor(section, "balloon.background",
								tkExtra.Balloon.background)

		# Correct materials
		for name, color in config.items(_COLOR_SECTION):
			# Correct material location
			if name.startswith("material."):
				config.set(_MATERIAL_SECTION, name.replace("material.","color."), color)
				config.remove_option(_COLOR_SECTION, name)
			elif name.startswith("alpha."):
				config.set(_MATERIAL_SECTION, name, color)
				config.remove_option(_COLOR_SECTION, name)

		# Read global colors
		for name, color in config.items(_COLOR_SECTION):
			if name.startswith("file."):
				bFileDialog.COLORS[name[5:]] = color

	# Flair
	section = _FLAIR_SECTION
	_showFlukaFiles = config.getBool(section, "showflukafiles", _showFlukaFiles)
	_errorReport    = config.getBool(section, "errorreport",    _errorReport)
	_showToolbar    = config.getBool(section, "showtoolbar",    _showToolbar)
	_showStatusbar  = config.getBool(section, "showstatusbar",  _showStatusbar)
	if config.has_option(section,"iconzoom"):
		dpiscale   = config.getInt(section,  "iconzoom",    dpiscale)
	else:
		dpiscale   = config.getInt(section,  "dpiscale",    dpiscale)
	tkFileViewer._DEFAULT_SPLIT *= dpiscale

	if "-d" not in sys.argv and "-D" not in sys.argv:
		Global.developer = config.getBool(section, "developer", Global.developer)
	Global.verbose   += max(0,min(config.getInt(section, "verbose",0),5))
	Fluka.warnMultMat = config.getBool(section, "warnmultmat",    Fluka.warnMultMat)

	bFileDialog._TIME_FORMAT  = config.getStr( section, "timeformat",bFileDialog._TIME_FORMAT)

	# Project
	Project.loadConfig(config)
	# Load materials
	Project.MaterialInfo.loadConfig(config)
	# Load all commands
	Project.Command.loadConfig(config)
	Project.UsrInfo.loadConfig(config)
	Project.RunInfo.loadConfig(config)
	Project.DicomInfo.loadConfig(config)

	# Printer
	tkDialogs.Printer.cmd   = config.getStr(_FLAIR_SECTION, "printercmd", "lpr")
	tkDialogs.Printer.paper = config.getStr(_FLAIR_SECTION, "paper",      "A4")
	tkDialogs.Printer.landscape = config.getBool(_FLAIR_SECTION,"landscape","True")

	_SKIP_INTRO = config.getBool(_FLAIR_SECTION, "skipintro", _SKIP_INTRO)
	_SAVE_FRAME = config.getBool(_FLAIR_SECTION, "rememberframe", _SAVE_FRAME)

	try: _autoSaveTime = config.getInt(_FLAIR_SECTION, "autosave", _autoSaveTime)
	except: pass

	try: tkExtra.SearchBox.TIME_THRESHOLD = \
		config.getInt(_FLAIR_SECTION, "keytimethreshold",
			tkExtra.SearchBox.TIME_THRESHOLD)
	except: pass

	try: tkExtra.Balloon.delay = \
		config.getInt(_FLAIR_SECTION, "balloondelay", tkExtra.Balloon.delay)
	except: pass

	Manual.VIEWER = config.getStr(_FLAIR_SECTION, "manualviewer", Manual.VIEWER)

	# Check plot formats
	section = _PLOT_SECTION
	items = config.items(section)
	if not items:
		for k,v in _PLOT_FORMAT.items():
			if k[0] == ".": config.set(section, k, v)
	else:
		_PLOT_FORMAT.clear()
		for k,v in items:
			if k[0] == ".":
				_PLOT_FORMAT[k] = v
	_PLOT_COMMANDS   = config.getStr(section, "commands","")
	_gnuplotTerminal = config.getStr(section, "terminal", _gnuplotTerminal)
	Gnuplot.program  = config.getStr(section, "program",  Gnuplot.program)

	if gui:
		import InputPage
		import GeometryTool
		import GeometryEditor
		InputPage.loadConfig(config)
		GeometryTool.loadConfig(config)
		GeometryEditor.loadConfig(config)

	# Load MC configuration
	DeckConfig.directory = prgDir
	FlukaConfig.loadConfig(config)
	MoiraConfig.loadConfig(config)
	McnpConfig.loadConfig(config)
	PenelopeConfig.loadConfig(config)
	PhitsConfig.loadConfig(config)

	section = _GEOMETRY_SECTION
	Input.zero     = config.getFloat(section, "zero",	Input.zero)
	Input.infinite = config.getFloat(section, "infinite",	Input.infinite)
	Input._useQUA  = config.getBool(section,  "usequa",	Input._useQUA)
	Input._useBOX  = config.getBool(section,  "usebox",	Input._useBOX)

	# Layers section - load default layers if any
	initGlobalLayers()

	# ------------- Style Section ------------------
	section = _STYLE_SECTION
	for k,v in config.items(section):
		name, var = k.split(".")
		if "," in v:
			nv = []
			for i in v.split(","):
				if i == "None": i=None
				nv.append(i)
			v = tuple(nv)
		else:
			if v=="None": v=None
		RichText.RichText.TAGS[name][var] = v

#-------------------------------------------------------------------------------
def initGlobalLayers():
	global config
	# Layers section - load default layers if any
	Project.initGlobalLayers()
	section = _LAYERS_SECTION
	names = []
	for i in range(100):
		try: name = config.get(section, f"layer.{i}")
		except: break
		names.append(name)
		Project._globalLayers[name] = Project.GeometryLayerInfo(None, name, True)

	# Scan all variables in the layers section
	pat = re.compile(r"^(.*)\.(\d+)$")
	for option in config.options(section):
		m = pat.match(option)
		if m:
			var = m.group(1)
			idx = int(m.group(2))
			if var == "layer": continue
			try: name = names[idx]
			except KeyError: name= "Media"
			val = config.get(section, option)
			if val in ("True","False"): val = (val=="True")
			try: val=int(val)
			except ValueError: pass
			Project._globalLayers[name][var] = val

#-------------------------------------------------------------------------------
# Update the layers from Project -> config
#-------------------------------------------------------------------------------
def updateLayers():
	section = _LAYERS_SECTION
	config.remove_section(section)
	config.add_section(section)

	for i,layer in enumerate(Project._globalLayers.values()):
		config.set(section, f"layer.{i}", layer.name)
		for key,val in layer.items():
			d = layer.default(key)
			if isinstance(val,bool): val = int(val)
			if val == d: continue
			config.set(section, f"{key}.{i}", str(val))

#-------------------------------------------------------------------------------
# Create the flair directory if needed
#-------------------------------------------------------------------------------
def createIniDir():
	# Create first the ini directory if doesn't exist
	try:
		os.stat(iniDir)
	except OSError:
		os.mkdir(iniDir)

#-------------------------------------------------------------------------------
# Remove from config all the unchanged items
#-------------------------------------------------------------------------------
def cleanIni():
	Global.dump("*-* tkFlair.cleanIni")
	global config
	newconfig = config	# Remember config
	config = Config()
	openIni(None,False)	# Open only system ini
	readIni()		# Reading will create all default sections and variables

	# Compare items
	for section in config.sections():
		for item, value in config.items(section):
			try:
				new = newconfig.get(section, item)
				if value==new:
					newconfig.remove_option(section, item)
			except configparser.NoOptionError:
				pass
		if not newconfig.items(section):
			# remove empty sections
			newconfig.remove_section(section)
	del config
	config = newconfig

#-------------------------------------------------------------------------------
# Write initialization file
#-------------------------------------------------------------------------------
def writeIni():
	Global.dump("*-* tkFlair.writeIni")
	createIniDir()
	tool.backup(iniFile)
	try:
		f = open(iniFile, "w")
	except IOError:
		log.error(f"Cannot write {iniFile} file\n")
		return
	config.set(__program__, "version", __version__)

	try:
		# write unsaved variables
		# 1. global colors	(XXX maybe move to Project)
		Project.MaterialInfo.saveConfig(config)
		import GeometryTool
		import GeometryEditor
		GeometryTool.saveConfig(config)
		GeometryEditor.saveConfig(config)
		Project.saveConfig(config)
	except (configparser.NoOptionError, configparser.NoSectionError) as ex:
		# FIXME fatal error. It should never come here
		# print the exception and restore the backup ini
		f.close()
		log.error("Fatal error. Please report it")
		log.errorCont(f"Restoring {iniFile}")
		log.error(ex)
		try:
			shutil.copyfile(f"{iniFile}~",iniFile)
		except Exception as ex:
			log.error(ex)
	else:
		# Cleanup
		cleanIni()
		# Write file
		config.write(f)
		f.close()

#-------------------------------------------------------------------------------
# Accept global font/color/setting options for all tkinter widgets
# They have to contain with a star '*' in their name
#	*font = Times 12		Set the global for everything
#	*text*font = Courier -14	Set the Text() widget font (help/viewer..)
#	*text*foreground = White	Set the Text() widget font (help/viewer..)
#-------------------------------------------------------------------------------
def addOptions(section, root):
	if root is None: return
	for name,value in config.items(section):
		if name.find("*")>=0:
			if section == _FONT_SECTION:
				value = fontString(value.split(','))
			root.option_add(name, value)

#-------------------------------------------------------------------------------
def writePrintIni():
	config.set(_FLAIR_SECTION, "printercmd", tkDialogs.Printer.cmd)
	config.set(_FLAIR_SECTION, "paper",      tkDialogs.Printer.paper)
	config.set(_FLAIR_SECTION, "landscape",  str(tkDialogs.Printer.landscape))

#-------------------------------------------------------------------------------
# Return an input template
#-------------------------------------------------------------------------------
def inputTemplate(mode):
	# Try mode as template
	template = os.path.join(prgDir, _TEMPLATES_DIR, f"{mode}.flair")
	if os.path.isfile(template):
		return template
	template = os.path.join(prgDir, _TEMPLATES_DIR, _DEFAULT_INP)
	if os.path.isfile(template):
		return template
	return None

#===============================================================================
# Menu commands
#===============================================================================
# ------------------------------------------------------------------------------
# Initialize underline string
# ------------------------------------------------------------------------------
def initUnderline(cards, mode):
	underline = {}
	for t in cards:
		if isinstance(t, str):
			c = CardInfo.get(mode,t)
		else:
			c = t
		if c.underline>=0:
			underline[c.tag[c.underline]] = True
	return underline

# ------------------------------------------------------------------------------
def commandLambda(mode, tag, callback):
	if tag[0] == "#" or tag in CardInfo.FLAIR_TAGS:
		t = tag
	else:
		t = mode + ":" + tag
	return lambda t=t, c=callback: c(t)

# ------------------------------------------------------------------------------
def findUnderline(ci, underline):
	# find character to underline
	u = ci.underline
	if u < 0:
		u = 0
		for ch in ci.tag.upper():
			if ch not in underline:
				underline[ch] = True
				break
			u += 1
		else:
			u = None
	return u

# ------------------------------------------------------------------------------
# create menu items for FLUKA cards
# ------------------------------------------------------------------------------
def addCardCommand(menu, mode, tag, cmd, underline, alias=None):
	if isinstance(tag, str):
		ci = CardInfo.get(mode, tag)
	else:
		ci = tag

	if underline is not None:
		u = findUnderline(ci, underline)
	else:
		u = None

	try:
		icon = icons.get(ci.tag)
	except KeyError:
		try:
			if ci.tag[0]=="/":
				icon = icons.get(ci.tag[ci.tag.rfind("/")+1:])
			else:
				icon = icons.get("tag")
		except KeyError:
			icon = icons.get("tag")

	menu.add_command(label=ci.tag if alias is None else alias,
			underline=u,
			image=icon,
			compound=tk.LEFT,
			command=cmd)
	return u

# ------------------------------------------------------------------------------
# Create a nice bodies menu @see _bodiesLayout
#   Plane
#      YZP     XZP
#      XYP     PLA
#   Box
#      RPP     BOX
#      WED     ARB
#   Sphere
#      SPH     ELL
#   Cylinder
#      RCC     REC
#      XCC     XEC
#      YCC     YEC
#      ZCC     ZEC
#   Cone
#      TRC
#   Quad
#      QUA
#   Torus
#      TRX     TRY
#      TRZ
#
# and surface menu
# ------------------------------------------------------------------------------
#   Plane
#      PX      PY      PZ
#      P       P3
#   Box
#      RPP     BOX     WED
#      ARB
#   Sphere
#      S       SO      SPH
#      SX      SY      SZ
#      ELL
#   Cylinder
#      C/X     C/Y     C/Z
#      CX      CY      CZ
#      RCC             REC
#   Cone
#      K/X     K/Y     K/Z
#      KX      KY      KZ
#      TRC
#   Quad
#      SQ      GQ
#   Torus
#      TX      TY      TZ
#   Generic
#      X       Y       Z
#   Hexagonal
#      RHP     HEX
# ------------------------------------------------------------------------------
def _layoutMenu(menu, mode, callback, layout, title):
	for column in range(len(layout[0])):
		if column>0 and menu["tearoff"]:
			menu.add_command(label="",font="Sans 7",state=tk.DISABLED)
		for row in layout:
			label = row[column]
			if label=="" or "a"<=label[-1]<="z":
				menu.add_command(label=label,state=tk.DISABLED)
				continue
			tag = label.replace("&","")
			underline = label.find("&")
			menu.add_command(label=tag,
					command=commandLambda(mode,tag,callback),
					underline=underline,
					image=icons.get(tag,"tag"),
					compound=tk.LEFT)
	for column in range(1,len(layout[0])):
		menu.entryconfigure(column*(len(layout)+int(menu["tearoff"])), columnbreak=1)
	if menu["tearoff"]:
		menu["tearoffcommand"] = lambda m,t,l=layout,ti=title : _layoutMenuTearOff(m,t,l,ti)

#-------------------------------------------------------------------------------
# Call the appropriate bodies/surface menu depending on mode
#-------------------------------------------------------------------------------
def bodiesMenu(menu, mode, callback):
	if mode in (Fluka.MODE, Moira.MODE):
		_layoutMenu(menu, mode, callback, _bodiesLayout, "Bodies")
	elif mode in (Mcnp.MODE, Phits.MODE):
		_layoutMenu(menu, mode, callback, _surfaceLayout, "Surfaces")

# ------------------------------------------------------------------------------
# Create a nice two column body menu
# ------------------------------------------------------------------------------
def _layoutMenuTearOff(menuname, top, layout, title):
	# Remove the tearoff bar
	try:
		menu = tk._default_root.nametowidget(menuname)
	except KeyError:
		# Apparently tear of from the std menu gives a name
		# with '#' instead of '.' which generates a keyerror
		return
	for column in range(len(layout[0])-1,-1,-1):
		menu.entryconfigure(column*(len(layout)+1)+1, columnbreak=1)
	# as tkinter doesn't know about the toplevel nametowidget doesn't work
	tk._default_root.tk.call('wm', 'title', top,  title)

#-------------------------------------------------------------------------------
def __addABCgroup(menu, mode, callback, abc, cards, **kw):
	# Group Sub Menu
	groupmenu = tk.Menu(menu, **kw)
	menu.add_cascade(label=abc,
			menu=groupmenu,
			image=icons.get("empty"),
			compound=tk.LEFT)
	underline = initUnderline(cards, mode)
	for c in cards:
		if c.directory: continue
		addCardCommand(groupmenu, mode, c,
				commandLambda(mode,c.tag,callback),
				underline)

#-------------------------------------------------------------------------------
def cardsABCMenu(menu, mode, callback, **kw):
	cards = [x for x in CardInfo.cards(mode)
			if "Obsolete" not in x.group ]
	CardInfo.sort(cards)

	# FIXME FIXME FIXME
	# Create menus and submenus
	# First level with the initial character
	# sublevels with ranges

	#-----------------------------------------------------------------------
	# return position starting with a new character
	#-----------------------------------------------------------------------
	def newCharacter(i):
		ch = cards[i].tag[0]
		i += 1
		for i in range(i,len(cards)):
			if cards[i].tag[0] != ch: break
		if i<len(cards) and cards[i].tag[0] == ch: i += 1
		return i

	MAXCARDS = 30	# up to 30 cards per menu

	start = 0
	category = ""
	while start < len(cards):
		category += cards[start].tag[0]
		stop = newCharacter(start)
		while stop-start < MAXCARDS and stop<len(cards):
			stop2 = newCharacter(stop)
			if stop2-start >= MAXCARDS: break
			if stop2<len(cards):
				category += cards[stop].tag[0]
			stop = stop2

		# Split to sub menus
		if stop-start > 25:
			while start<stop:
				to = min(start + MAXCARDS, stop)
				label = f"{cards[start].tag[:10]} .. {cards[to-1].tag[:10]}"
				__addABCgroup(menu,
						mode,
						callback,
						label,
						cards[start : to],
						**kw)
				start = to
		else:
			__addABCgroup(menu,
					mode,
					callback,
					category,
					cards[start : stop],
					**kw)
		category = ""
		start = stop

	if start < stop:
		__addABCgroup(menu,
				mode,
				callback,
				category,
				cards[start : stop],
				**kw)

#-------------------------------------------------------------------------------
# Create a tree structure with all cards
#-------------------------------------------------------------------------------
def cardsTreeMenu(menu, mode, cards, callback, **kw):
	# Added the sorted list of cards
	CardInfo.sort(cards)
	underline = initUnderline(cards, mode)
	prev = ""
	submenus = [("",menu)]
	for c in cards:
		tag = c.tag
		if tag[0]=="/":
			*words,tag = tag.split("/")
			if len(words) < len(submenus):
				del submenus[len(words):]
			# Find possible splitting point
			for i in range(min(len(submenus), len(words))):
				if words[i] != submenus[i][0]:
					del submenus[i:]
					break
			while len(submenus) < len(words):
				# create sub-menus
				sm = tk.Menu(submenus[-1][1], **kw)
				label = words[len(submenus)]
				icon = icons.get("/".join(words), label, "tags")
				submenus[-1][1].add_cascade(label=label,
						image=icon,
						compound=tk.LEFT,
						menu=sm)
				submenus.append((label, sm))
		else:
			if len(submenus)>1: del submenus[1:]
			if "." in tag:
				p = tag.index(".",0)
				if prev == tag[:p]:
					tag = "... "+tag[p:]
				else:
					prev = tag[:p]
			else:
				prev = tag

		if c.directory: continue
		addCardCommand(submenus[-1][1],
				mode, c,
				commandLambda(mode, c.tag, callback),
				underline if tag==c.tag else None,
				alias=tag)

#-------------------------------------------------------------------------------
# Create a submenu for a specific group. Call recursively for sub groups
#-------------------------------------------------------------------------------
def cardsGroupMenu(menu, mode, group, callback, **kw):
	# scan all cards belonging to a group
	cards = [x for x in CardInfo.cards(mode) if group in x.group]
	if not cards: return False	# is empty!

	# special treatment for Geometry/Body
	if group in (CardInfo.GEOMETRY_BODY_GROUP, CardInfo.GEOMETRY_SURFACE_GROUP):
		bodiesMenu(menu, mode, callback)
		return True

	# ------------------------
	# Find possible sub-groups
	# ------------------------
	for g in sorted(CardInfo.groups_set):
		if g.startswith(group+"/"):
			# add a submenu
			subgroup = g[len(group)+1:]	# FIXME until next "/"
			submenu = tk.Menu(menu, **kw)
			fullgroup = group+"/"+subgroup
			if cardsGroupMenu(submenu,
					mode,
					fullgroup,
					callback, **kw):
				icon = icons.get(CardInfo.groups_icons[fullgroup], "tags")
				menu.add_cascade(label=subgroup,
						image=icon,
						compound=tk.LEFT,
						menu=submenu)

	# Special groups (before the cards)
	if group==CardInfo.GEOMETRY_TRANSFORM_GROUP and mode=="fluka":
		# Transformations list
		#trans = [x for x in cards if x.tag[0]=="$"]
		icon  = icons.get("tags")
		menu.add_command(label=CardInfo.EXPANSION,
				command=commandLambda(mode,CardInfo.EXPANSION,callback),
				image=icon,
				compound=tk.LEFT,
				underline=0)
		menu.add_command(label=CardInfo.TRANSLAT,
				command=commandLambda(mode,CardInfo.TRANSLAT,callback),
				image=icon,
				compound=tk.LEFT,
				underline=0)
		menu.add_command(label=CardInfo.TRANSFORM,
				command=commandLambda(mode,CardInfo.TRANSFORM,callback),
				image=icon,
				compound=tk.LEFT,
				underline=1)

	# Added the sorted list of cards
	cardsTreeMenu(menu, mode, cards, callback, **kw)

	# Special groups (after the cards)
	if group==CardInfo.PREPROCESSOR_GROUP:
		menu.add_command(label=CardInfo.IFENDIF,
				command=commandLambda("",CardInfo.IFENDIF,callback),
				compound=tk.LEFT, image=icons.get(CardInfo.IFENDIF),
				underline=0)

	# Other special groups
	return True	# something was added

#-------------------------------------------------------------------------------
# Display cards menu
#-------------------------------------------------------------------------------
def cardsAllMenu(menu, mode, callback, **kw):
	for group in CardInfo.groups:
		if group == "Developers" and not Global.developer: continue
		# Find underline
		try:
			underline = CardInfo.groups_underline[CardInfo.groups.index(group)]
		except ValueError:
			underline = None

		# Group Sub Menu
		groupmenu = tk.Menu(menu, **kw)
		if cardsGroupMenu(groupmenu, mode, group, callback, **kw):
			menu.add_cascade(label=group,
					menu=groupmenu,
					image=icons.get(group,"empty"),
					compound=tk.LEFT,
					underline=underline)

#===============================================================================
# Manually add exception
#-------------------------------------------------------------------------------
def addException():
	try:
		typ, val, tb = sys.exc_info()
		traceback.print_exception(typ, val, tb)
		if errors: errors.append("")
		exception = traceback.format_exception(typ, val, tb)
		errors.extend(exception)
		for line in exception:
			log.say(line.strip())
		if len(errors) > 100:
			# If too many errors are found send the error report
			ReportDialog()
	except Exception as ex:
		log.error(ex)

#-------------------------------------------------------------------------------
# Exception handling decorator
# Should be used when dealing with threads and we want to register the exception
#-------------------------------------------------------------------------------
def exception(function):
	def wrapper(*args, **kwargs):
		try:
			return function(*args, **kwargs)
		except Exception as ex:
			addException()
			raise ex
	return wrapper

#===============================================================================
# Override default CallWrapper of tkinter
#===============================================================================
class CallWrapper:
	"""Replaces the tkinter.CallWrapper with extra functionality"""
	def __init__(self, func, subst, widget):
		"""Store FUNC, SUBST and WIDGET as members."""
		self.func   = func
		self.subst  = subst
		self.widget = widget

	# ----------------------------------------------------------------------
	def __call__(self, *args):
		"""Apply first function SUBST to arguments, than FUNC."""
		try:
			if self.subst:
				args = self.subst(*args)
			return self.func(*args)
		# One possible fix is to make an external file for the wrapper
		# and import depending the version
		except SystemExit:
			raise
		except KeyboardInterrupt:
			pass
		except:
			addException()

#===============================================================================
# Error message reporting dialog
#===============================================================================
class ReportDialog(tk.Toplevel):
	_shown = False	# avoid re-entry when multiple errors are displayed

	#-----------------------------------------------------------------------
	def __init__(self, master=None):
		if ReportDialog._shown: return
		ReportDialog._shown = True

		super().__init__(master)
		self.transient(master)
		self.title(f"{__program__} Error Reporting")

		# Label Frame
		frame = tk.LabelFrame(self, text="Report")
		frame.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

		l = tk.Label(frame, text="The following report is about to be send "
					f"to the author of {__program__}",
					justify=tk.LEFT, anchor=tk.W)
		l.pack(side=tk.TOP)

		self.text = tk.Text(frame, background="White")
		self.text.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

		sb = ttk.Scrollbar(frame, orient=tk.VERTICAL, command=self.text.yview)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		self.text.config(yscrollcommand=sb.set)

		# email frame
		frame = tk.Frame(self)
		frame.pack(side=tk.TOP, fill=tk.X)

		l = tk.Label(frame, text="Your email")
		l.pack(side=tk.LEFT)

		self.email = tk.Entry(frame, background="White")
		self.email.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X)
		tkExtra.Balloon.set(self.email,
			"Fill in your email address if you want to be informed on the bug resolution")

		# Automatic error reporting
		self.err = tk.BooleanVar()
		self.err.set(_errorReport)
		b = tk.Checkbutton(frame, text="Automatic error reporting",
			variable=self.err, anchor=tk.E, justify=tk.RIGHT)
		b.pack(side=tk.RIGHT)
		tkExtra.Balloon.set(b, "When enabled it will popup this dialog in case of an error")

		# Buttons
		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, fill=tk.X)

		b = tk.Button(frame, text="Close",
				image=icons.get("x"),
				compound=tk.LEFT,
				command=self.cancel)
		b.pack(side=tk.RIGHT)
		b = tk.Button(frame, text="Send report",
				image=icons.get("debug"),
				compound=tk.LEFT,
				command=self.sendReport)
		b.pack(side=tk.RIGHT)

		import Updates
		import GeometryEditor

		# FIXME check version PIL, numpy, pydicom
		try:
			import numpy
			numpy_version = numpy.__version__
		except ImportError:
			numpy_version = "?"

		try:
			import pydicom
			pydicom_version = pydicom.__version__
		except ImportError:
			try:
				import dicom
				pydicom_version = dicom.__version__
			except ImportError:
				pydicom_version = "?"

		findRevision()

		# Fill report
		txt = [ f"version     : {__version__}",
			f"geoviewer   : {GeometryEditor.version()} "
					f"{__version__!=GeometryEditor.version() and 'WARNING'}",
			f"revision    : {__revision__}",
			f"last change : {__lastchange__}",
			f"kleidi      : {config.getStr(_FLAIR_SECTION, Updates.License.key(),'0.0')}",
			 "python      : %s"%(sys.version.replace('\n',' ')),
			f"tk          : {tk.TkVersion}",
			f"tcl         : {tk.TclVersion}",
			f"numpy       : {numpy_version}",
			f"pydicom     : {pydicom_version}",
			f"platform    : {sys.platform}",
			"\nDescription:",
			"...please describe the problem...",
			"\nTraceback:" ]
		for e in errors:
			if e!="" and e[-1] == "\n":
				txt.append(e[:-1])
			else:
				txt.append(e)

		self.text.insert('0.0', "\n".join(txt))

		# Guess email
		user = os.getenv("USER")
		host = os.getenv("HOSTNAME")
		if user and host:
			email = f"{user}@{host}"
		else:
			email = ""
		email = config.getStr(_FLAIR_SECTION, "email", email)
		self.email.insert(0,email)

		self.protocol("WM_DELETE_WINDOW", self.close)
		self.bind('<Escape>', self.close)

		# Wait action
		self.grab_set()
		self.focus_set()
		self.wait_window()

	# ----------------------------------------------------------------------
	def close(self, event=None):
		ReportDialog._shown = False
		self.destroy()

	# ----------------------------------------------------------------------
	def sendReport(self):
		from urllib import request, parse
		global errors

		email = self.email.get()
		desc  = self.text.get('1.0', tk.END).strip()
		config.set(_FLAIR_SECTION, "email", self.email.get())

		# Send information
		self.config(cursor="watch")
		self.update_idletasks()

		data = parse.urlencode({"email":email, "desc":desc}).encode()
		try:
			handler = request.urlopen(_SEND_EMAIL, data=data)
		except Exception as ex:
			messagebox.showwarning("Error sending report",
				f"There was a problem connecting to the FLUKA web site.\n{ex}",
				parent=self)
		else:
			if handler.status == 200:
				messagebox.showinfo("Report successfully send",
					f"Report was successfully uploaded to {__program__} web site",
					parent=self)
				del errors[:]
			else:
				messagebox.showwarning("Error sending report",
					f"There was an error sending the report\n"
					f"Code={handler.status} {handler.reason}",
					parent=self)
			handler.close()
		self.config(cursor="")
		self.cancel()

	# ----------------------------------------------------------------------
	def cancel(self):
		global _errorReport, errors
		_errorReport = self.err.get()
		config.set(_FLAIR_SECTION, "errorreport", str(bool(self.err.get())))
		del errors[:]
		self.close()

	# ----------------------------------------------------------------------
	@staticmethod
	def show(master):
		ReportDialog(master)

#===============================================================================
def preferences(window, page=None):
	global config
	import ConfigDialog
	cd = ConfigDialog.ConfigDialog(window, config)
	if cd.show(page):
		# Load fonts...
		readIni()
		addOptions(_FONT_SECTION,  window.winfo_toplevel())
		addOptions(_COLOR_SECTION, window.winfo_toplevel())
		return True
	return False

#===============================================================================
# About dialog box
#===============================================================================
def aboutDialog(parent, timer=None):
	toplevel = tk.Toplevel(parent)
	toplevel.transient(parent)
	toplevel.title(f"About {__program__}")

	bg = "#707070"
	fg = "#ffffff"

	findRevision()

	font1 = 'Helvetica 30 bold'
	font2 = 'Helvetica 10'
	font3 = 'Helvetica 8'

	frame = tk.Frame(toplevel, borderwidth=2,
			relief=tk.SUNKEN, background=bg)
	frame.pack(side=tk.TOP, expand=tk.TRUE, fill=tk.BOTH, padx=5, pady=5)

	row = 0
	try:
		fn = os.path.join(prgDir,"images","flair-logo.gif")
		img = tk.PhotoImage(file=fn)
		if dpiscale>1: img = img.zoom(dpiscale)
		l = tk.Label(frame, image=img,
				foreground=fg, background=bg,
				relief=tk.RAISED,
				padx=0, pady=0)
		l.grid(row=row, column=0, columnspan=2, padx=5, pady=5)
	except tk.TclError:
		pass

	row += 1
	l = tk.Label(frame, text=__program__,
			foreground=fg, background=bg,
			font=font1)
	l.grid(row=row, column=0, columnspan=2, sticky=tk.W, padx=10, pady=5)

	row += 1
	l = tk.Label(frame, text="FLUKA Advanced Interface",
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=0, columnspan=2, sticky=tk.W, padx=10, pady=2)

	row += 1
	l = tk.Label(frame, text="/fleə(r)/\tn [U,C] natural or "
			"instinctive ability - to do something\n"
			"\twell, to select or recognize what is best, "
			"more useful, etc.\n"
			"\t\t\t[Oxford Advanced Dictionary]",
			font = font3,
			foreground=fg, background=bg, justify=tk.LEFT)
	l.grid(row=row, column=0, columnspan=2, sticky=tk.W, padx=10, pady=1)

	row += 1
	f = tk.Frame(frame, borderwidth=1, relief=tk.SUNKEN,
			height=2, background=bg)
	f.grid(row=row, column=0, columnspan=2, sticky=tk.EW, padx=5, pady=5)

	# -----
	row += 1
	l = tk.Label(frame, text='www:',
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=0, sticky=tk.E, padx=10, pady=2)

	l = tk.Label(frame, text=__www__,
			foreground=fg, background=bg, justify=tk.LEFT,
			activeforeground="Blue",
			font=font2, cursor="hand1")
	l.grid(row=row, column=1, sticky=tk.W, padx=2, pady=2)
	l.bind('<Button-1>', aboutWebsite)

	# -----
	row += 1
	l = tk.Label(frame, text='email:',
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=0, sticky=tk.E, padx=10, pady=2)

	l = tk.Label(frame, text=__email__,
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=1, sticky=tk.W, padx=2, pady=2)

	# -----
	row += 1
	l = tk.Label(frame, text='author:',
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=0, sticky=tk.NE, padx=10, pady=2)

	l = tk.Label(frame, text=__author__.replace(",","\n"),
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=1, sticky=tk.NW, padx=2, pady=2)

	# -----
	row += 1
	l = tk.Label(frame, text='version:',
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=0, sticky=tk.E, padx=10, pady=2)

	l = tk.Label(frame, text=f"{__version__} [R{__revision__}]",
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=1, sticky=tk.NW, padx=2, pady=2)

	# -----
	row += 1
	l = tk.Label(frame, text='last change:',
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=0, sticky=tk.E, padx=10, pady=2)

	l = tk.Label(frame, text=__lastchange__,
			foreground=fg, background=bg, justify=tk.LEFT,
			font=font2)
	l.grid(row=row, column=1, sticky=tk.NW, padx=2, pady=2)

	closeFunc = lambda e=None,t=toplevel: t.destroy()
	b = tk.Button(toplevel, text="Close", command=closeFunc)
	b.pack(pady=5)
	frame.grid_columnconfigure(1, weight=1)

	toplevel.bind('<Escape>',   closeFunc)
	toplevel.bind('<Return>',   closeFunc)
	toplevel.bind('<KP_Enter>', closeFunc)

	toplevel.deiconify()
	if not toplevel.winfo_ismapped():
		toplevel.wait_visibility()
	toplevel.resizable(False, False)
	try: toplevel.grab_set()
	except: pass
	b.focus_set()
	toplevel.lift()
	if timer:
		toplevel.after(timer, closeFunc)
	toplevel.wait_window()

#-------------------------------------------------------------------------------
def aboutWebsite(event=None):
	if openurl(__www__):
		messagebox.showwarning("No valid browser",
			"Cannot find any valid browser to open web page")

#===============================================================================
# Run Files list list
#===============================================================================
class RunFilesList(tkExtra.ColorMultiListbox):
	def __init__(self, master, flair):
		super().__init__(master,
				(('File', 30, None),
				 ('Type', 10, None),
				 ('Size',  5, None),
				 ('Date', 13, None)))
		self.bindList('<Double-1>', self.view)
		self.bindList('<Return>',   self.view)
		self.bindList('<KP_Enter>', self.view)
		self.flair = flair

	# ----------------------------------------------------------------------
	# Fill with files generated by the running script
	# ----------------------------------------------------------------------
	def fill(self, inpname, pid=None, cycle=None, project=None):
		self.project = project
		# Scan directory
		if isinstance(cycle, int):
			pat = re.compile(rf"^{inpname}{cycle:03d}[._].+$")
		else:
			pat = re.compile(rf"^{inpname}\d\d\d[._].+$")

		for filename in os.listdir("."):
			if pat.match(filename):
				try:
					s = os.lstat(filename)
					if stat.S_ISLNK(s.st_mode):
						continue
					ext,color = bFileDialog.fileTypeColor(filename)
					self.insert(tk.END,
						(filename, ext, s.st_size,
						 time.strftime(bFileDialog._TIME_FORMAT,
							time.localtime(s.st_mtime))))
					if color is not None:
						self.itemconfig(tk.END,foreground=color)
				except OSError:
					pass

		# Scan temporary directory
		if pid:
			# FIXME check for the cycle
			for ext,cfg in ((FlukaConfig.EXT, FlukaConfig),
					(MoiraConfig.EXT, MoiraConfig)):
				if os.path.isfile(inpname+ext):
					break
			else:
				cfg = None
			if cfg:
				try:
					for fn in os.listdir(f"{cfg.CYCLE}{pid}"):
						filename = f"{cfg.CYCLE}{pid}/{fn}"
						try:
							s = os.lstat(filename)
							if stat.S_ISLNK(s.st_mode):
								continue
							ext,color = bFileDialog.fileTypeColor(filename)
							self.insert(tk.END,
								(filename, ext, s.st_size,
								 time.strftime(bFileDialog._TIME_FORMAT,
									time.localtime(s.st_mtime))))
							if color is not None:
								self.itemconfig(tk.END,foreground=color)
						except OSError:
							pass
				except OSError:
					pass

		if self.size() == 0:
			messagebox.showerror("Cannot open output",
				"Cannot find any output file.",
				parent=self)
			return

	# ----------------------------------------------------------------------
	# View document
	# ----------------------------------------------------------------------
	def view(self, event=None):
		files = []
		for sel in self.curselection():
			files.append(self.get(sel)[0])
		self.flair.viewer(*files)

	# ----------------------------------------------------------------------
	# Edit document
	# ----------------------------------------------------------------------
	def edit(self, event=None):
		files = []
		for sel in self.curselection():
			files.append(self.get(sel)[0])
		editor(*files)

#===============================================================================
# OutputFilesDialog
#===============================================================================
class OutputFilesDialog(tk.Toplevel):
	def __init__(self, flair, inpname, pid=None, cycle=None, project=None):
		super().__init__(flair)
		self.transient(flair)
		self.title("Output File Viewer")
		self.flair = flair

		listbox = RunFilesList(self, flair)
		listbox.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
		listbox.fill(inpname, pid, cycle, project)

		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM, fill=tk.X)

		btn = tk.Button(frame,
				text="View", underline=0,
				image=icons.get("view"), compound=tk.LEFT,
				command=listbox.view)
		btn.pack(side=tk.LEFT)

		btn = tk.Button(frame,
				text="Edit", underline=0,
				image=icons.get("edit"), compound=tk.LEFT,
				command=listbox.edit)
		btn.pack(side=tk.LEFT)

		btn = tk.Button(frame, text="Close", underline=0,
				image=icons.get("x"), compound=tk.LEFT,
				command=self.close)
		btn.pack(side=tk.LEFT)

		self.bind("<Escape>", self.close)
		self.protocol("WM_DELETE_WINDOW", self.close)

		if listbox.size():
			# Wait action
			try:
				self.focus_set()
				self.wait_window()
			except tk.TclError:
				pass
		else:
			self.close()

	# ----------------------------------------------------------------------
	def close(self, event=None):
		self.destroy()

#===============================================================================
# Undo listbox
#===============================================================================
class UndoListbox(tk.Toplevel):
	posted = None

	def __init__(self, master, flair, x, y):
		super().__init__(master)
		self.overrideredirect(1)
		self.transient(master)

		self.flair = flair

		# Listbox
		frame = tk.Frame(self)
		frame.pack(expand=tk.YES,fill=tk.BOTH)

		self.listbox = tkExtra.ExListbox(frame, borderwidth=0,
				selectmode=tk.SINGLE,
				selectborderwidth=0,
				background="White",
				takefocus=True,
				exportselection=tk.FALSE)
		self.listbox.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

		sb = ttk.Scrollbar(frame, orient=tk.VERTICAL, command=self.listbox.yview)
		sb.pack(side=tk.RIGHT, fill=tk.Y)
		self.listbox.config(yscrollcommand=sb.set)

		frame = tk.Frame(self)
		frame.pack(side=tk.BOTTOM)

		self.listbox.bind("<Motion>",          self.motion)
		self.listbox.bind("<ButtonRelease-1>", self.release)
		self.listbox.bind('<FocusOut>',        self.cancel)
		self.bind("<Escape>",   self.cancel)
		self.bind("<Return>",   self.ok)
		self.bind("<KP_Enter>", self.ok)

		for lbl in self.flair.undoredo.undoTextList():
			self.listbox.insert(tk.END, lbl)

		self.listbox.activate(0)
		self.listbox.selection_set(0)
		self.listbox.focus_set()
		self.focus_set()

		self.geometry(f"+{x}+{y}")
		UndoListbox.posted = self

	#-----------------------------------------------------------------------
	def motion(self, event):
		y = self.listbox.nearest(event.y)
		self.listbox.selection_clear(0,tk.END)
		self.listbox.selection_set(0,y)
		self.listbox.activate(y)

	#-----------------------------------------------------------------------
	def release(self, event):
		y = self.listbox.nearest(event.y)
		self.listbox.selection_clear(0,tk.END)
		self.listbox.selection_set(y)
		self.listbox.activate(y)
		self.ok()

	#-----------------------------------------------------------------------
	def ok(self, event=None):
		sel = self.listbox.curselection()
		if not sel: return
		for _ in range(int(sel[0])+1):
			self.flair.write(self.flair.undoredo.undoText()+"\n")
			self.flair.undoredo.undo()
			self.flair.refresh()
		self.flair.project.input.clearCache(Card.BODY)
		self.destroy()
		UndoListbox.posted = None
		self.master.refreshUndoButton()

	#-----------------------------------------------------------------------
	def cancel(self, event=None):
		self.destroy()
		UndoListbox.posted = None

#===============================================================================
# Implement common functionality of a Treeview displaying an InfoTree
# Common for RunList and PlotPage
#===============================================================================
class InfoTreeview(tkExtra.Treeview):
	#-----------------------------------------------------------------------
	def __init__(self, master, **kw):
		super().__init__(master, **kw)
		self.bind("<Button-1>",		self.button1)
		self.bind("<ButtonRelease-1>",	self.release1)
		self.bind("<Double-1>",		self.double1)

		#self.infotree  = tree
		self.item2info  = {}		# dictionary from item to info class
		self._lastfocus = None
		self._lastclick = 0		# to discriminate click for selecting from double click for editing
		self._groupicon = icons.get("group")

	# ----------------------------------------------------------------------
	# return item from info
	# Not the most efficient, but we avoid the use of an extra dictionary
	# ----------------------------------------------------------------------
	def info2item(self, info):
		for i,p in self.item2info.items():
			if info is p: return i
		return None

	# ----------------------------------------------------------------------
	def clear(self):
		self.item2info.clear()

	# ----------------------------------------------------------------------
	# @return a tree(list of lists) of all items in the tree
	# ----------------------------------------------------------------------
	def items(self):
		def _scanBranch(parent, path):
			branch = Project.GroupInfo(self.item(parent,"text"))
			for item in self.get_children(parent):
				info = self.item2info.get(item)
				if isinstance(info, Project.GroupInfo):
					name = self.item(item, "text")
					newpath = f"{path}/{name}" if path else name
					branch.append(_scanBranch(item, newpath))
				else:
					info.dirname(path)
					branch.append(info)
			return branch
		return _scanBranch("","")

	# ----------------------------------------------------------------------
	# @return a tree of selected runs
	# @see PlotPage.PlotTreeview.selectedPlotTree
	# ----------------------------------------------------------------------
	def selectedItems(self):
		def _scanBranch(parent, branch):
			nonlocal selection
			for item in self.get_children(parent):
				info = self.item2info[item]
				if not isinstance(info, Project.GroupInfo) and \
				   item in selection:
					branch.append(info)
				else:
					name = self.item(item, "text")
					if item in selection:
						branch.append(_scanBranch(item,
								Project.GroupInfo(name)))
					else:
						_scanBranch(item, branch)
			return branch
		selection = set(self.selection())
		return _scanBranch("",Project.GroupInfo())

	# ----------------------------------------------------------------------
	# Button 1 pressed
	# ----------------------------------------------------------------------
	def button1(self, event):
		self.focus_set()
		if event.state & STATE.CONTROL:
			self.focus(self.identify_row(event.y))
			return
		if event.state & STATE.SHIFT: return
		#print()
		#print("identify_row=",self.identify_row(event.y))
		#print("identify_column=",self.identify_column(event.x))
		#print("identify_region=",self.identify_region(event.x,event.y))
		#print("identify_element=",self.identify_element(event.x,event.y))
		# previous focused
		self._lastfocus = self.focus()
		item = self.identify_row(event.y)
		if item: self.focus(item)
		element = self.identify_element(event.x, event.y)
		if item not in self.selection() or element=="Treeitem.indicator":
			# if clicked item is NOT selected revert to the default handler
			#print("NOT IN SELECTION RETURN TO DEFAULT")
			self._lastclick = event.time
			return

		#print("SELECTED=true")
		column    = self.identify_column(event.x)
		selection = self.selection()
		#print(f"{element=} {column=}")
		#print("image=",element in ("image","Treeitem.indicator"))
		#print("len=",len(selection))
		#print("column=",column)
		#print("in selection=",item in selection)
		#print("lastclick=", self.lastclick)
		#print("dt=", event.time-self._lastclick)

		if event.time - self._lastclick > Global.DND_TIME:
			#if (element=="image" or len(selection)>1) and column == "#0" and \
			#    self.dnd_start(event):
			if element=="image" and self.dnd_start(event):
				return "break"

		if len(selection)>1 and column != "#0":
			return "break"

	# ----------------------------------------------------------------------
	# Button 1 released
	# ----------------------------------------------------------------------
	def release1(self, event):
		if event.state & (STATE.CONTROL | STATE.SHIFT): return
		item = self.identify_row(event.y)
		if item: self.focus(item)
		selection = self.selection()
		element   = self.identify_element(event.x, event.y)
		column    = self.identify_column(event.x)

		if len(selection)==1 and \
		   item==self._lastfocus and \
		   element=="text" and \
		   column=="#0":
			#self.event_generate("<<Rename>>")
			return self.editInPlace(event)

		elif column!="#0":
			#self.event_generate("<<Rename>>")
			return self.editInPlace(event)

	# ----------------------------------------------------------------------
	def double1(self, event):
		if self.identify_element(event.x, event.y) == "text":
			#self.event_generate("<<Rename>>")
			return self.editInPlace(event)

	# ----------------------------------------------------------------------
	def editInPlace(self, event=None):
		...

	# ----------------------------------------------------------------------
	# Move up select items by one
	# @param default   ignore default item (first entry)
	# ----------------------------------------------------------------------
	def moveUp(self, default=False):
		for item in self.selection():
			index  = self.index(item)
			parent = self.parent(item)
			if index==0 or (default and parent=="" and index==1): continue
			if parent and not isinstance(self.item2info[parent],Project.GroupInfo): continue
			self.detach(item)
			self.reattach(item, parent, index-1)
		return "break"

	# ----------------------------------------------------------------------
	# Move down select items by one
	# @param default   ignore default item (first entry)
	# ----------------------------------------------------------------------
	def moveDown(self, default=False):
		for item in reversed(self.selection()):
			index  = self.index(item)
			parent = self.parent(item)
			if default and parent=="" and index==0: continue
			if parent and not isinstance(self.item2info[parent],Project.GroupInfo): continue
			if index == len(self.get_children(parent))-1: continue
			self.detach(item)
			self.reattach(item, parent, index+1)
		return "break"

	# ----------------------------------------------------------------------
	# Move left (up one parent)
	# ----------------------------------------------------------------------
	def moveLeft(self, default=False):
		for item in reversed(self.selection()):
			parent = self.parent(item)
			if parent=="" or not isinstance(self.item2info[parent],Project.GroupInfo): continue
			grand  = self.parent(parent)
			if default and self.index(parent)==0: continue
			index = self.index(parent)
			#if index == len(self.get_children(parent))-1: continue
			self.detach(item)
			self.reattach(item, grand, index)
		return "break"

	# ----------------------------------------------------------------------
	# Move right (down one parent if next is group)
	# ----------------------------------------------------------------------
	def moveRight(self, default=False):
		for item in reversed(self.selection()):
			index = self.index(item)
			if default and self.parent(item)=="" and index==0: continue
			nextitem = self.next(item)
			if nextitem and isinstance(self.item2info[nextitem], Project.GroupInfo):
				self.detach(item)
				self.reattach(item, nextitem, 0)
		return "break"

	# ----------------------------------------------------------------------
	# scan for open items
	# ----------------------------------------------------------------------
	def scanOpen(self, parent=""):
		for item in self.get_children(parent):
			info = self.item2info.get(item)
			if info and not isinstance(info, Project.GroupInfo):
				info["open"] = bool(self.item(parent,"open"))
			self.scanOpen(item)

	# ----------------------------------------------------------------------
	# Add one new info item at location
	# ----------------------------------------------------------------------
	def add(self, new):
		selection = self.selection()
		if len(selection)==1:
			active = self.focus()
			info   = self.item2info[active]
			if isinstance(info, Project.GroupInfo) and self.item(active,"open"):
				parent = active
				index  = len(self.get_children(active))
			else:
				parent = self.parent(active)
				index  = self.index(active)+1
		elif selection:
			active = selection[0]
			parent = self.parent(active)
			index  = self.index(active)
		else:
			parent = ""
			index  = 0

		item = self.insert(parent, index, new)
		self.focus(item)
		self.see(item)
		self.selection_set(item)
		return item

	# ----------------------------------------------------------------------
	# Add a group entry
	# ----------------------------------------------------------------------
	def addGroup(self):
		selection = self.selection()
		if len(selection)==1:
			active = self.focus()
			info   = self.item2info[active]
			if isinstance(info, Project.GroupInfo) and self.item(active,"open"):
				parent = active
				index  = len(self.get_children(active))
			else:
				parent = self.parent(active)
				index  = self.index(active)+1
		elif selection:
			active = selection[0]
			parent = self.parent(active)
			index  = self.index(active)
		else:
			parent = ""
			index  = 0

		group = Project.GroupInfo("group")
		item  = self.insert(parent, index, image=self._groupicon, open=True)
		self.selectClear()
		self.focus(item)
		self.see(item)
		edit = tkExtra.InPlaceEdit(self, item=item)
		if edit.value:
			group.name = edit.value
			self.item2info[item] = group
			self.selection_set(item)
			if len(selection)>1:
				# reattach all selected items under then new one
				for i,s in enumerate(selection):
					self.detach(s)
					self.reattach(s, item, i)
			return True
		else:
			ttk.Treeview.delete(self,item)
			self.focus(active)
			self.selection_set(active)
			return False

#-------------------------------------------------------------------------------
# Edit files
#-------------------------------------------------------------------------------
def editor(*files):
	if not files: return
	try:
		os.spawnv(os.P_NOWAIT, Project.editor,
				[ Project.editor ] + list(files) )
	except:
		messagebox.showerror("Error executing command",
			"Error executing command:\n"
			f"{Project.editor} {' '.join(files)}")

#-------------------------------------------------------------------------------
def openurl(url, new=False, parent=None):
	try:
		if webbrowser.get().basename in ("links", "elinks", "lynx"):
			raise Exception("no browser")
	except:
		log.error("No valid web browser found")
		messagebox.showerror("Web browser not found",
			"Error no valid web browser found",
			parent=parent)
		return True
	if new:
		webbrowser.open_new(url)
	else:
		webbrowser.open(url)
	return False

#-------------------------------------------------------------------------------
def printer(master):
	prt = tkDialogs.Printer(master)
	if prt.show():
		writePrintIni()
		return prt
	else:
		return None

#-------------------------------------------------------------------------------
# Stupid fix for python 2.3 the askyesno
# returns sometimes "yes" and sometimes "True"
#-------------------------------------------------------------------------------
def askyesno(title, message, **options):
	try:
		ans = messagebox.askquestion(title, message, **options)
		return str(ans)==messagebox.YES or ans is True
	except tk.TclError:
		return False

#-------------------------------------------------------------------------------
def init(ini=None, gui=True):
	global style
	openIni(ini)
	readIni(gui)
	if gui:
		icons.load(os.path.join(prgDir,"icons"))
		#icons.load(os.path.join(prgDir,"icons"), ".png")
		icons.loadAliases(os.path.join(prgDir,"icons","aliases.db"))
		palette.load([ os.path.join(prgDir, _PALETTE_DIR),
				os.path.join(iniDir, _PALETTE_DIR) ])

		style = ttk.Style()
		#Global.dump(style.theme_names())
		# ('clam', 'alt', 'default', 'classic'
		#style.theme_use("clam")
		#style.theme_use("alt")
		#style.theme_use("default")
		#style.theme_use("classic")
		#style.configure("TTreeview", background="Yellow")
		style.configure("TNotebook.Tab", background="Gray")
		style.configure(_NOTEBOOK_STYLE, background="LightGray")
		#style.configure(_TREEVIEW_STYLE, background="LightGray")
		#style.configure(_TREEVIEW_STYLE, background="Yellow")

		# For 4k HiDPI display
		if dpiscale > 1:
			#global root
			icons.zoom(dpiscale)
			#root.tk.call('tk', 'scaling', '-displayof', '.', dpiscale)
			style.configure("TScrollbar", #width="5m", arrowsize="5m")
					width=12*dpiscale,
					arrowsize=12*dpiscale)
			f=tkFont.Font(name='TkDefaultFont', exists=1)
			style.configure("Treeview", rowheight=f.metrics('linespace')+2)

#-------------------------------------------------------------------------------
def fini():
	icons.delete()
