#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   21-Jun-2020

import os
import sys
import getopt
import random
import datetime
try:
	import pydicom
except ImportError:
	try:
		import dicom as pydicom
	except ImportError:
		pydicom = None
try:
	import numpy as np
except ImportError:
	np = None

sys.path.insert(0,os.path.join(os.path.abspath(os.path.dirname(__file__)), "lib"))

import log
import tool
import Data
import tkFlair
from Evaluation import Qe

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"
__date__   = "2020-06-21"

#===============================================================================
class RTException(Exception): pass

#-------------------------------------------------------------------------------
def usrbin2rtdose(usrbin_file, det, rtdose_file, rtplan_file, output_file,
		norm=None, instance=None, sop=None, series=None):
	log.say("Usrbin to RTDOSE conversion")
	log.say(f"author: {__author__}")
	log.say(f"email:  {__email__}")
	log.say(f"\tUsrbin: {usrbin_file}")
	# Load usrbin and converted it to words
	try:
		usrbin = Data.Usrbin(usrbin_file)
	except Exception as ex:
		log.error(f"Cannot open USRBIN {usrbin_file!r}")
		log.errorCont(ex)
		raise RTException(f"Cannot open USRBIN {usrbin_file!r}")

	array = np.transpose(np.array(usrbin.readArray(det)),(2,1,0))
	array[array<0.0] = 0.0		# Filter possibly negative values
	scale  = 0xFFFF / array.max()	# 16 bit word max value 0xffff
	array *= scale			# Values are in range 0..0xffff

	# Use the default rtdose if None is supplied
	if rtdose_file is None:
		rtdose_file = os.path.join(tkFlair.prgDir,"templates","rtdose.dcm")
		if sop is None:
			sop = pydicom.uid.generate_uid(prefix="2.25.")
		if series is None:
			series = pydicom.uid.generate_uid(prefix="2.25.")

	# Load dicom files
	log.say(f"\tRTDOSE: {rtdose_file}")
	try:
		rtdose = pydicom.read_file(rtdose_file, force=True)
	except Exception as ex:
		log.error(f"Cannot open RTDOSE {rtdose_file!r}")
		log.errorCont(ex)
		raise RTException(f"Cannot open RTDOSE {rtdose_file!r}")
	try:
		instanceNumber = rtdose.InstanceNumber
		log.say(f"\tRTDOSE Instance Number: {instanceNumber}")
	except:
		instanceNumber = 0
		log.warning("RTDOSE Instance Number not found assumed 0")

	log.say(f"\tRTPLAN: {rtplan_file}")
	if rtplan_file:
		try:
			rtplan = pydicom.read_file(rtplan_file, force=True)
		except Exception as ex:
			log.error(f"Cannot open RTPLAN {rtdose_file!r}")
			log.errorCont(ex)
			raise RTException(f"Cannot open RTPLAN {rtdose_file!r}")

		# Check if rtplan is the correct one
		refplan_uid = rtdose.ReferencedRTPlanSequence[0].ReferencedSOPInstanceUID
		if refplan_uid != rtplan.SOPInstanceUID:
			log.warning("The RTPLAN loaded is not the one referenced from RTDOSE")
			log.warningCont(f"RTDOSE reference: {refplan_uid}")
			log.warningCont(f"RTPLAN SOP UID:   {rtplan.SOPInstanceUID}")
	else:
		rtplan = None
		del rtdose.ReferencedRTPlanSequence
		if norm is None:
			log.warning("No RTPLAN found. Assuming 1.0 as normalization")
			norm = 1.0

	if sop is not None:
		log.say(f"\tSOP Instance UID: {sop}")
		rtdose.SOPInstanceUID = sop
	if series is not None:
		log.say(f"\tSeries Instance UID: {series}")
		rtdose.SeriesInstanceUID = series
	else:
		series = rtdose.SeriesInstanceUID

	rtdose.SeriesDescription = "Monte Carlo recalculation"
	rtdose.Manufacturer = "http://flair.cern"
	rtdose.ManufacturerModelName = "fluka.4"
	rtdose.SoftwareVersions = f"usrbin2rtdose.py v{tkFlair.__version__}"
	if instance is None:
		rtdose.InstanceNumber = instanceNumber + 100
	else:
		rtdose.InstanceNumber = instance
	log.say(f"\tInstance Number: {rtdose.InstanceNumber}")
	sop = rtdose.SOPInstanceUID.split(".")
	# make a random new last part with the same number of digits
	sop[-1] = "".join([str(random.randint(0,9)) for x in range(len(sop[-1]))])
	rtdose.SOPInstanceUID = ".".join(sop)
	rtdose.StudyDescription = f"{os.path.basename(usrbin_file)}[{det}] {usrbin.title}".strip()
	now = datetime.datetime.today()
	rtdose.StudyDate   = now.strftime("%Y%m%d")
	rtdose.StudyTime   = now.strftime("%H%M%S")
	udt = datetime.datetime.fromtimestamp(os.path.getmtime(usrbin_file))
	rtdose.ContentDate = udt.strftime("%Y%m%d")
	rtdose.ContentTime = udt.strftime("%Y%m%d")

	# delete private elements (odd groups)
	for element in rtdose:
		if element.tag.group&1:
			del rtdose[element.tag]

	datatype = rtdose.pixel_array.dtype

	# UID prefixes
	#"1.2.300"	# Greece
	#"1.2.756"	# Switzerland

	# Defined Terms:
	#PLAN
	#	dose calculated for entire delivery of all fraction groups of RT Plan
	#MULTI_PLAN
	#	dose calculated for entire delivery of 2 or more RT Plans
	#FRACTION
	#BEAM
	#BRACHY
	#	dose calculated for entire delivery of one or more Brachy Application Setups
	#	within RT Plan
	#FRACTION_SESSION
	#	dose calculated for a single session ("fraction") of a single Fraction Group
	#	within RT Plan
	#BEAM_SESSION
	#	dose calculated for a single session ("fraction") of one or more Beams
	#	within RT Plan
	#BRACHY_SESSION
	#	dose calculated for a single session ("fraction") of one or more
	#	Brachy Application Setups within RT Plan
	#CONTROL_POINT
	#	dose calculated for one or more Control Points within a Beam for a single fraction
	#RECORD
	#	dose calculated for RT Beams Treatment Record

	# dose calculated for entire delivery of a single Fraction Group within RT Plan
	# BNV one fraction contains multiple beams
	log.say(f"\tDose Summation Type: {rtdose.DoseSummationType}")
	beamMet = 0.0
	if rtplan is None:
		pass
	elif rtdose.DoseSummationType == "PLAN":
		for fracSeq in rtplan.FractionGroupSequence:
			for seq in fracSeq.ReferencedBeamSequence:
				beamMet += float(seq.BeamMeterset)
	elif rtdose.DoseSummationType == "FRACTION":
		fracId = int(rtdose.ReferencedRTPlanSequence[0].
				ReferencedFractionGroupSequence[0].
				ReferencedFractionGroupNumber)
		log.say(f"\tFractionId: {fracId}")
		assert fracId == int(rtplan.FractionGroupSequence[fracId-1].FractionGroupNumber)
		for seq in rtplan.FractionGroupSequence[fracId-1].ReferencedBeamSequence:
			beamMet += float(seq.BeamMeterset)

	# dose calculated for entire delivery of one or more Beams within RT Plan
	elif rtdose.DoseSummationType == "BEAM":
		fracId = int(rtdose.ReferencedRTPlanSequence[0]
				.ReferencedFractionGroupSequence[0]
				.ReferencedFractionGroupNumber)
		log.say(f"\tFractionId: {fracId}")
		beamId = int(rtdose.ReferencedRTPlanSequence[0]
				.ReferencedFractionGroupSequence[fracId-1]
				.ReferencedBeamSequence[0]
				.ReferencedBeamNumber)
		log.say(f"\tBeamId: {beamId}")
		assert fracId == int(rtplan.FractionGroupSequence[fracId-1].FractionGroupNumber)
		assert beamId == int(rtplan.FractionGroupSequence[fracId-1]
				.ReferencedBeamSequence[beamId-1]
				.ReferencedBeamNumber)
		beamMet = float(rtplan.FractionGroupSequence[fracId-1]
				.ReferencedBeamSequence[beamId-1]
				.BeamMeterset)

	else:
		log.error(f"Cannot handle DoseSummationType={rtdose.DoseSummationType}")
		raise RTException(f"Cannot handle DoseSummationType={rtdose.DoseSummationType}")

	log.say(f"\tTotal Meterset: {beamMet}")

	# Dose type
	if rtdose.DoseType == "EFFECTIVE":
		rbe = 1.1
	elif rtdose.DoseType == "PHYSICAL":
		rbe = 1.0
	else:
		raise Exception(f"Unknown Dose Type {rtdose.DoseType}")
	log.say(f"\tRBE: {rbe}")

	# Monitor units or number of particles
	if rtplan:
		ionplan = "IonBeamSequence" in rtplan
		beamSequence = rtplan.IonBeamSequence if ionplan \
					else rtplan.BeamSequence
		if beamSequence[0].PrimaryDosimeterUnit == "MU":	# Monitor Units
			log.error("MU (Monitor Units) is not yet implemented.")
			log.errorCont("Assuming 1.0 as normalization")
		elif beamSequence[0].PrimaryDosimeterUnit == "NP":	# Number of Particles
			# Number of fractions
			#no_of_fractions = int(rtplan.FractionGroupSequence[0]
			#			.NumberOfFractionsPlanned)
			# GeV/g -> eV/kg = 1e12
			norm = beamMet * rbe * Qe * 1e12
		else:
			raise ValueError("Unknown Primary Dosimeter Unit "
					f"{beamSequence[0].PrimaryDosimeterUnit}")
	log.say(f"\tNormalization: {norm}")

	# Add normalization to Gy or RBE
	detector = usrbin.detector[0]
	rtdose.Columns        = detector.nx
	rtdose.Rows           = detector.ny
	rtdose.NumberOfFrames = detector.nz
	rtdose.PixelSpacing   = [round(detector.dx*10,6), round(detector.dy*10,6)]
	rtdose.SliceThickness = round(detector.dz*10,6)
	rtdose.ImagePositionPatient = [ round((detector.xlow + detector.dx/2)*10,6),
					round((detector.ylow + detector.dy/2)*10,6),
					round((detector.zlow + detector.dz/2)*10,6)]
	rtdose.GridFrameOffsetVector = [f"{round(z*detector.dz*10,6):g}"
						for z in range(detector.nz)]
	rtdose.DoseGridScaling = norm/scale
	rtdose.PixelData = np.array(array, dtype=datatype).tobytes()

	tool.backup(output_file)
	log.say(f"\tOutput RTDOSE file: {output_file}")
	try:
		rtdose.save_as(output_file)
	except Exception as ex:
		log.error(f"Cannot save RTDOSE {output_file!r}")
		log.errorCont(ex)
		raise RTException(f"Cannot save RTDOSE {output_file!r}")

	return series

#-------------------------------------------------------------------------------
# Display usage notice and exit
#-------------------------------------------------------------------------------
def usage(rc=0):
	print()
	print("syntax usrbin2rtdose.py -u usrbin -d rtdoseIN -p plan -o rtdoseOUT [args...]")
	print()
	print(" Options:")
	print("\t-h|-?|--help\tprint this help page")
	print("\t-u #|--usrbin #\tUSRBIN file to load (no-default)")
	print("\t-d #|--det #\tUSRBIN detector to load (default: 1)")
	print("\t-n #|--norm #\tNormalization value (no-default, rtplan calculated or 1.0)")
	print("\t-r #|--rtdose #\tRTDOSE file to load as template (no-default)")
	print("\t-p #|--rtplan #\tRTPLAN file to load (no-default)")
	print("\t-o #|--output #\tOutput RTDOSE file to save (no-default)")
	#print("\t\t\tIf only the filename is specified it will be saved")
	#print("\t\t\tin the same directory of rtdose")
	print("\t-i #|--instance #\tSpecify instance number to be used (default 100+current)")
	print("\t--sop #\t\tSpecify SOP id to be used (default current+random last digits)")
	print()
	print(f" Author:\t{__author__}")
	print(f" Email:\t\t{__email__}")
	print(f" Date:\t\t{__date__}")
	sys.exit(rc)

#-------------------------------------------------------------------------------
if __name__ == "__main__":
	if len(sys.argv) == 1:
		usage()
	try:
		optlist, args = getopt.getopt(sys.argv[1:],
					"?hu:n:d:r:p:o:i:",
					["help",
					"usrbin=",
					"det=",
					"norm=",
					"rtdose=",
					"rtplan=",
					"output=",
					"instance=",
					"sop=",
					"series=",
					])
	except getopt.GetoptError:
		usage(1)

	usrbin_file = None
	usrbin_det  = 1
	rtdose_file = None
	rtplan_file = None
	output_file = None
	instance    = None
	norm        = None
	sop         = None
	series      = None

	if pydicom is None:
		log.error("'pydicom' package is not found")
		sys.exit(2)

	for opt, val in optlist:
		if opt in ("-h", "-?", "--help"):
			usage(0)
		elif opt in ("-u", "--usrbin"):
			usrbin_file = val
		elif opt in ("-d", "--det"):
			try:
				usrbin_det = int(val)
			except ValueError:
				log.error(f"Invalid detector number {val!r} specified")
				sys.exit(3)
		elif opt in ("-n", "--norm"):
			try:
				norm = float(val)
			except ValueError:
				log.error(f"Invalid normalization {val!r} specified")
				sys.exit(3)
		elif opt in ("-r", "--rtdose"):
			rtdose_file = val
		elif opt in ("-p", "--rtplan"):
			rtplan_file = val
		elif opt in ("-o", "--output"):
			output_file = val
		elif opt in ("-i", "--instance"):
			instance = int(val)
		elif opt == "--sop":
			sop = val
		elif opt == "--series":
			series = val

	if usrbin_file is None:
		log.error("Usrbin file not specified")
		sys.exit(3)
	#if rtplan_file is None:
	#	log.error("RTPLAN file not specified")
	#	sys.exit(4)
	if output_file is None:
		log.error("Output RTDOSE file not specified")
		sys.exit(5)

	#if not os.path.dirname(output_file):
	#	output_file = os.path.join(os.path.dirname(rtdose_file), output_file

	try:
		uid = usrbin2rtdose(usrbin_file, usrbin_det-1,
				rtdose_file, rtplan_file, output_file,
				norm=norm, instance=instance, sop=sop, series=series)
	except RTException as ex:
		log.error(ex)
