#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import re
import math
import time
import tkinter as tk
from tkinter import ttk

import log
import csg
import tool
import bmath
import icons
import Global
import Manual
import Ribbon
import tkFlair
import tkExtra
import Project
import FlairRibbon
import GeometryBody
import GeometryViewer
import GeometryLayers
import GeometryErrors

from tkconst import STATE
from Global  import dump, BIT, ACTION, VIEWER
from deck    import Input, Card, CardInfo, Layout
from fluka   import Fluka, FlukaCard, FlukaVoxel
from moira   import Moira, MoiraCard
from mcnp    import Mcnp, McnpCard
from phits   import Phits, PhitsCard

# Body colors for Visible, seemed similar to the ones in GeometryViewer
# BUT for tkinter notation (Geometry Viewer = int) (tkinter = string #xxxxxx)
VISIBLE_COLOR   = "DarkMagenta"
LOCK_COLOR      = "Orange"
FREEZE_COLOR    = "#00A0A0"
ERROR_COLOR     = "Red"
WIREFRAME_COLOR = "#007070"
ADDZONE         = "+zone"
OPER_TRANS      = str.maketrans("+-|()\n\t","       ")

# Tuple of card tags that contain gobject information to be drawn
_CARD_OBJECTS = ("BEAM", "ROT-DEFI", "SPOTBEAM", "MATRIX","ROTM", "/primary/select")

_POINT_prop = [ ("option",	4),
		("anchor",	5),
		("size",	5),
		("color",	6),
		("x",		1),
		("y",		2),
		("z",		3)]

_ARROW_prop = _POINT_prop[:]
_ARROW_prop.extend([	("dx",	7),
			("dy",	8),
			("dz",	9),
			("@length", -1)])

_SPLINE_prop = _POINT_prop[:]
_SPLINE_prop.extend([	("tension",   7),
			("#segments", 8)])
#for _i in range(0, 19):
#	_SPLINE_prop.append(("x%d"%(_i+2), _i*3+7))
#	_SPLINE_prop.append(("y%d"%(_i+2), _i*3+8))
#	_SPLINE_prop.append(("z%d"%(_i+2), _i*3+9))

_RULER_prop = _POINT_prop[:]
_RULER_prop.extend([	("xe",	7),
			("ye",	8),
			("ze",	9),
			("xa",	10),
			("ya",	11),
			("za",	12),
			("@length",-1),
			("@angle", -2)])

_LIGHT_prop = _POINT_prop[:]
_LIGHT_prop.extend([	("dx",		 7),
			("dy",		 8),
			("dz",		 9),
			("power",	10),
			("falloff",	11),
			("spec",	12),
			("radius",	13),
			("# samples",	14)
		])

_CAMERA_prop = _POINT_prop[:]
_CAMERA_prop.extend([	("atx",    7),
			("aty",    8),
			("atz",    9),
			("upx",   10),
			("upy",   11),
			("upz",   12),
			("focal", 13),
			("fov",   14),
			("aspect",15)
		])

_MESH_prop = _POINT_prop[:]
_MESH_prop.append(("file", -1))

_MESH_prop.append(("file", -1))

_ROTDEFI_axes = ["", "X", "Y", "Z"]

_ARROW_opt  = [ "head", "tail", "headtail", "line" ]
_CAMERA_opt = [ "ortho", "perspective" ]
_LIGHT_opt  = [ "sun", "omni", "spot" ]
_MESH_opt   = [ "stl" ]
_POINT_opt  = [ "cross", "dot", "square", "x", "diamond", "circle" ]
_RULER_opt  = [ "simple", "angle" ]
_SPLINE_opt = [ "open", "closed", "lines", "lines-closed" ]
#_SPLINE_opt = [ "line", "cardinal" ]

_OBJECTS = {
		"arrow"  : (_ARROW_prop,  _ARROW_opt),
		"camera" : (_CAMERA_prop, _CAMERA_opt),
		"light"  : (_LIGHT_prop,  _LIGHT_opt),
		"mesh"   : (_MESH_prop,   _MESH_opt),
		"point"  : (_POINT_prop,  _POINT_opt),
		"ruler"  : (_RULER_prop,  _RULER_opt),
		"spline" : (_SPLINE_prop, _SPLINE_opt),
	}

_ANCHOR_list = [ "none", "C", "N", "NE", "E", "SE", "S", "SW", "W", "NW" ]

_ONOFF_list  = [ "off", "on"]
_FALLOFF_list = ["const+shadow", "1/r+shadow", "1/r2+shadow", "constant", "1/r", "1/r2" ]

# Patterns
_SPLIT_PAT  = re.compile(r"[(),;\[\]]")
_FIRST_WORD = re.compile(r"^\s*(\S+)\s+(.*)$",re.DOTALL)
_LAST_NUM   = re.compile(r"^(.*?)(\d+)$")

_ASSIGNMA_field = [ "",
		"Magnetic",
		"Electric",
		"Magnetic+Electric",
		"Prompt Magnetic",
		"Prompt Electric",
		"Prompt Magnetic+Electric",
		"Decay Magnetic",
		"Decay Electric",
		"Decay Magnetic+Electric" ]

#===============================================================================
# Show in place text
#===============================================================================
class InPlaceText(tkExtra.InPlaceText):
	def defaultBinds(self):
		super().defaultBinds()
		self.edit.bind("<Escape>", self.ok)

#===============================================================================
# Edit in place a region
#===============================================================================
class InPlaceRegion(InPlaceText):
	# ----------------------------------------------------------------------
	def __init__(self, listbox, flair, item=tk.ACTIVE, value=None, **kw):
		self.flair = flair
		self._inList = False
		super().__init__(listbox, item, value, **kw)

	# ----------------------------------------------------------------------
	def defaultBinds(self):
		super().defaultBinds()
		self.edit.bind("<Key>",  self._handleKey)

	# ----------------------------------------------------------------------
	def focusOut(self, event=None):
		if self._inList:
			return "break"
		else:
			self.ok()

	# ----------------------------------------------------------------------
	def _insertList(self, lst):
		try:
			x,y,x2,y2 = self.edit.bbox(tk.INSERT)
		except TypeError:
			x = y = 0
		x += self.edit.winfo_rootx()
		y += self.edit.winfo_rooty()
		self._inList = True
		self.toplevel.grab_release()
		selected = tkExtra.PopupList(self.toplevel, lst).show(x,y)
		self._inList = False
		if selected:
			try: selected = selected.split('[')[0]
			except: pass
			self.edit.insert(tk.INSERT,selected)
		# After closing a toplevel with overrideredirect(1)
		# the whole application is loosing the focus so a
		# simple focus_set() is not sufficient. I have to use
		# focus_force() on the toplevel
		try:
			self.toplevel.focus_force()
		except tk.TclError:
			pass
		self.edit.focus_set()
		self.toplevel.grab_set()

	# ----------------------------------------------------------------------
	def _handleKey(self, event):
		if event.char in ("+","-") or event.keysym == "Insert":
			self.edit.insert(tk.INSERT, event.char)
			lst = sorted(self.flair.project.input.cardsCache(Card.BODY).keys())
			lst.insert(0,Layout.BODY_PARENTHESIS)
			lst.insert(0,"")
			self._insertList(lst)
			return "break"

		elif event.char == "#":
			self.edit.insert(tk.INSERT, event.char)
			lst = sorted(self.flair.project.input.cardsCache(Card.REGION).keys())
			lst.insert(0,Layout.BODY_PARENTHESIS)
			lst.insert(0,"")
			self._insertList(lst)
			return "break"

#===============================================================================
# Generic properties for any card
#===============================================================================
class Properties:
	_lastActive  = 0	# last element active in listbox(s)
	_lastActive2 = 0	# remember them in case for a moment everything is
				# unselected, when an object is selected back
				# to restore the active element

	# ----------------------------------------------------------------------
	def __init__(self, editor, cards, freeze=False, visible=False):
		self.editor   = editor
		self.flair    = editor.flair
		self.listbox  = editor.propList
		self.listbox2 = editor.attrList
		self.cards    = cards
		self.many     = len(cards)>1
		self.input    = self.editor.project.input
		self.items    = ["name", "comment"]
		self.items2   = ["name"]
		self.whats    = {}	# translation of label to what#, name
		self._modified = 0	# last time card was modified

		for card in cards:
			if card.input is not None and \
			   isinstance(card.input, FlukaVoxel):
				self.voxel = True
				break
		else:
			self.voxel = False

		if freeze:  self.items2.append("freeze")
		if visible: self.items2.append("visible")

	# ----------------------------------------------------------------------
	def type(self): return self.cards[0].type()
	def name(self): return self.cards[0].name()

	# ----------------------------------------------------------------------
	# Return common tag of all cards if any otherwise None
	# ----------------------------------------------------------------------
	def tag(self):
		tag = self.cards[0].tag
		for c in self.cards:
			if tag != c.tag:
				return None
		return tag

	# ----------------------------------------------------------------------
	def fill(self):
		if self.listbox.size()  > 0: Properties._lastActive  = self.listbox.index(tk.ACTIVE)
		if self.listbox2.size() > 0: Properties._lastActive2 = self.listbox2.index(tk.ACTIVE)
		self.listbox.delete(0,tk.END)
		self.listbox2.delete(0,tk.END)
		for item in self.items:
			self.listbox.insert(tk.END, (item, self.value(item)))
		for item in self.items2:
			self.listbox2.insert(tk.END, (item, self.value(item)))
		self.listbox.activate(Properties._lastActive)
		self.listbox2.activate(Properties._lastActive2)
		self.modified()

	# ----------------------------------------------------------------------
	def modified(self):
		self._modified = self.cards[0].modifiedTime()

	# ----------------------------------------------------------------------
	def isModified(self):
		return self._modified < self.cards[0].modifiedTime()

	# ----------------------------------------------------------------------
	def value(self, name=None, val=None):
		if name == "name":
			if self.many:
				return "-"
			elif val is None:
				return self.cards[0].name()
			else:
				self.editor.changeName(self.cards[0].name(), val)
				return None

		elif name == "comment":
			if val is None:
				v = self.cards[0].comment().rstrip()
				for c in self.cards:
					if v != c.comment().rstrip():
						return "-"
				return v
			else:
				undoinfo = []
				for c in self.cards:
					undoinfo.append(self.flair.setCommentUndo(c,val))
				self.addUndo(undoinfo)

		elif name=="visible":
			if val is None:
				v = self.cards[0].get(Card.SELECT,0)&BIT.VISIBLE
				for c in self.cards:
					if v != c.get(Card.SELECT,0)&BIT.VISIBLE:
						return tkFlair._MULTI
				if v: return tkFlair._ON
				else: return tkFlair._OFF
			else:
				for c in self.cards:
					if val: c[Card.SELECT] = c.get(Card.SELECT,0) | BIT.VISIBLE
					else:   c[Card.SELECT] = c.get(Card.SELECT,0) & BIT.UNVISIBLE

		elif name=="freeze":
			if val is None:
				v = self.cards[0].get(Card.SELECT,0)&BIT.FREEZE
				for c in self.cards:
					if v != c.get(Card.SELECT,0)&BIT.FREEZE:
						return tkFlair._MULTI
				if v: return tkFlair._ON
				else: return tkFlair._OFF
			else:
				for c in self.cards:
					if val: c[Card.SELECT] = c.get(Card.SELECT,0) | BIT.FREEZE
					else:   c[Card.SELECT] = c.get(Card.SELECT,0) & BIT.UNFREEZE

		else:
			what = self.whats[name]
			if val is None:
				v = self.cards[0].what(what)
				for c in self.cards:
					if v != c.what(what): return ""
				return v
			else:
				if val=="": return
				undoinfo = []
				for c in self.cards:
					undoinfo.append(self.flair.setWhatUndo(c,what,val))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.refreshLate()
		return ""

	# ----------------------------------------------------------------------
	def edit(self, listbox):
		name = listbox.get(tk.ACTIVE)[0]
		if name == "name" and self.many:
			return None
		elif name == "comment":
			return InPlaceText(listbox.listbox(1))
		else:
			return tkExtra.InPlaceEdit(listbox.listbox(1),
						exportselection=False)

	# ----------------------------------------------------------------------
	# return true if fields accepts only one number
	# then multiple values to be split over the next fields
	# e.g. a copy paste of a vector ##x## ##y## ##z##
	# will be split in to the next fields
	# ----------------------------------------------------------------------
	def split(self):
		return False

	# ----------------------------------------------------------------------
	def split2(self):
		return False

	# ----------------------------------------------------------------------
	def addUndo(self, undoinfo):
		self.flair.addUndo(undoinfo)

	# ----------------------------------------------------------------------
	# Initialize standard layout from db
	# ----------------------------------------------------------------------
	def initLayout(self, card):
		if not self.tag(): return
		whats = [(x,i) for i,x in enumerate(card.info.label[0][1:-1],1)]
		if card.info.layoutDynamic:
			whats.extend([(x,-i)
				for i,x in enumerate(card.info.layoutDynamic[0],1)])
		self.items.extend([x for x,y in whats])
		self.whats = dict(whats)

#===============================================================================
# Body properties class
#===============================================================================
class BodyProperties(Properties):
	DIAM =  {"RCC" : 7,
		 "SPH" : 4,
		 "S"   : 4,
		 "XCC" : 3,
		 "YCC" : 3,
		 "ZCC" : 3,
		 "C/X" : 3,
		 "C/Y" : 3,
		 "C/Z" : 3,
		 "SX"  : 2,
		 "SY"  : 2,
		 "SZ"  : 2,
		 "SO"  : 1,
		 "CX"  : 1,
		 "CY"  : 1,
		 "CZ"  : 1}

	# ----------------------------------------------------------------------
	def __init__(self, editor, cards):
		super().__init__(editor, cards, freeze=True, visible=True)
		self.items.append("type")
		self.initLayout(cards[0])
		self.items2.extend(["bbox","wireframe",
				"dx","dy","dz","expansion","transform"])
				#"dx","dy","dz","expansion","transform","transform2"])

	# ----------------------------------------------------------------------
	def value(self, name=None, val=None):
		if name=="type":
			if val is None:
				#if self.bodytype: return self.cards[0].tag
				return self.cards[0].tag if self.whats else "-"

		elif name=="wireframe":
			if val is None:
				v = self.cards[0].get(Card.SELECT,0)&BIT.WIREFRAME
				for c in self.cards:
					if v != c.get(Card.SELECT,0)&BIT.WIREFRAME:
						return tkFlair._MULTI
				if v: return tkFlair._ON
				else: return tkFlair._OFF
			else:
				for c in self.cards:
					if val: c[Card.SELECT] = c.get(Card.SELECT,0) | BIT.WIREFRAME
					else:   c[Card.SELECT] = c.get(Card.SELECT,0) & BIT.UNWIREFRAME

		elif name=="bbox":
			if val is None:
				v = self.cards[0].get(Card.SELECT,0)&BIT.BBOX
				for c in self.cards:
					if v != c.get(Card.SELECT,0)&BIT.BBOX:
						return tkFlair._MULTI
				if v: return tkFlair._ON
				else: return tkFlair._OFF
			else:
				for c in self.cards:
					if val: c[Card.SELECT] = c.get(Card.SELECT,0) | BIT.BBOX
					else:   c[Card.SELECT] = c.get(Card.SELECT,0) & BIT.UNBBOX

		elif name in ("dx","dy","dz","expansion","transform","transform2"):
			key = f"@{name}"
			if val is None:
				v = self.cards[0][key]
				for c in self.cards:
					if v != c[key]:
						return ""
				if v is None:
					return ""
				else:
					return v
			else:
				for c in self.cards: c[key] = val
				self.flair.addUndo(self.flair.correctBodyPropertiesUndo())
				self.editor.forceRefresh()

		elif name in ("@dx", "@h"):
			if val is None:
				v = None
				for c in self.cards:
					try:
						if c.tag=="RPP":
							d = str(c.numWhat(2) - c.numWhat(1))
						else:
							d = str(math.sqrt(c.numWhat(4)**2
									+ c.numWhat(5)**2
									+ c.numWhat(6)**2))
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					if c.tag == "RPP":
						undoinfo.append(
							self.flair.setWhatUndo(c, 2, val+c.numWhat(1)))
					else:
						if c.tag in ("BOX", "WED", "RAW"):
							V = c.bodyX()
						else:	# RCC, TRC, REC
							V = c.bodyZ()
						V.normalize()
						V = V*val
						undoinfo.append(self.flair.setWhatUndo(c, 4, V[0]))
						undoinfo.append(self.flair.setWhatUndo(c, 5, V[1]))
						undoinfo.append(self.flair.setWhatUndo(c, 6, V[2]))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		elif name in ("@dy", "@rx"):
			if val is None:
				v = None
				for c in self.cards:
					try:
						if c.tag=="RPP":
							d = str(c.numWhat(4) - c.numWhat(3))
						else:
							d = str(math.sqrt(c.numWhat(7)**2
									+ c.numWhat(8)**2
									+ c.numWhat(9)**2))
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					if c.tag == "RPP":
						undoinfo.append(
							self.flair.setWhatUndo(
								c, 4, val+c.numWhat(3)))
					else:
						if c.tag in ("BOX", "WED", "RAW"):
							V = c.bodyY()
						else:
							V = c.bodyX()
						V.normalize()
						V = V*val
						undoinfo.append(self.flair.setWhatUndo(c, 7, V[0]))
						undoinfo.append(self.flair.setWhatUndo(c, 8, V[1]))
						undoinfo.append(self.flair.setWhatUndo(c, 9, V[2]))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		elif name in ("@dz", "@ry"):
			if val is None:
				v = None
				for c in self.cards:
					try:
						if c.tag=="RPP":
							d = str(c.numWhat(6) - c.numWhat(5))
						else:
							d = str(math.sqrt(c.numWhat(10)**2
									+ c.numWhat(11)**2
									+ c.numWhat(12)**2))
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					if c.tag == "RPP":
						undoinfo.append(
							self.flair.setWhatUndo(
								c, 6, val+c.numWhat(5)))
					else:
						if c.tag in ("BOX", "WED", "RAW"):
							V = c.bodyZ()
						else:
							V = c.bodyY()
						V.normalize()
						V = V*val
						undoinfo.append(self.flair.setWhatUndo(c, 10, V[0]))
						undoinfo.append(self.flair.setWhatUndo(c, 11, V[1]))
						undoinfo.append(self.flair.setWhatUndo(c, 12, V[2]))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		elif name == "@diam":
			if val is None:
				v = None
				for c in self.cards:
					try:
						d = str(2*c.numWhat(BodyProperties.DIAM[c.tag]))
					except:
						continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					try:
						w = BodyProperties.DIAM[c.tag]
						undoinfo.append(self.flair.setWhatUndo(c, w, val/2))
					except KeyError:
						pass
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		elif name == "@xmid":
			if val is None:
				v = None
				for c in self.cards:
					try:
						if c.tag == "RPP":
							d = str((c.numWhat(1) + c.numWhat(2))/2.0)
						elif c.tag in ("RCC","TRC","REC"):
							h = c.bodyZ()
							d = str(c.numWhat(1) + h[0]/2.0)
						elif c.tag == "BOX":
							h = c.bodyX() + c.bodyY() + c.bodyZ()
							d = str(c.numWhat(1) + h[0]/2.0)
						else:
							continue
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					if c.tag == "RPP":
						d = (c.numWhat(2) - c.numWhat(1))/2.0
						undoinfo.append(self.flair.setWhatUndo(c, 1, val-d))
						undoinfo.append(self.flair.setWhatUndo(c, 2, val+d))
					elif c.tag in ("RCC","TRC","REC"):
						h = c.bodyZ()
						undoinfo.append(self.flair.setWhatUndo(c, 1, val-h[0]/2.0))
					elif c.tag == "BOX":
						h = c.bodyX() + c.bodyY() + c.bodyZ()
						undoinfo.append(self.flair.setWhatUndo(c, 1, val-h[0]/2.0))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		elif name == "@ymid":
			if val is None:
				v = None
				for c in self.cards:
					try:
						if c.tag == "RPP":
							d = str((c.numWhat(3) + c.numWhat(4))/2.0)
						elif c.tag in ("RCC","TRC","REC"):
							h = c.bodyZ()
							d = str(c.numWhat(2) + h[1]/2.0)
						elif c.tag == "BOX":
							h = c.bodyX() + c.bodyY() + c.bodyZ()
							d = str(c.numWhat(2) + h[1]/2.0)
						else:
							continue
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					if c.tag == "RPP":
						d = (c.numWhat(4) - c.numWhat(3))/2.0
						undoinfo.append(self.flair.setWhatUndo(c, 3, val-d))
						undoinfo.append(self.flair.setWhatUndo(c, 4, val+d))
					elif c.tag in ("RCC","TRC","REC"):
						h = c.bodyZ()
						undoinfo.append(self.flair.setWhatUndo(c, 2, val-h[1]/2.0))
					elif c.tag == "BOX":
						h = c.bodyX() + c.bodyY() + c.bodyZ()
						undoinfo.append(self.flair.setWhatUndo(c, 2, val-h[1]/2.0))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		elif name == "@zmid":
			if val is None:
				v = None
				for c in self.cards:
					try:
						if c.tag == "RPP":
							d = str((c.numWhat(5) + c.numWhat(6))/2.0)
						elif c.tag in ("RCC","TRC","REC"):
							h = c.bodyZ()
							d = str(c.numWhat(3) + h[2]/2.0)
						elif c.tag == "BOX":
							h = c.bodyX() + c.bodyY() + c.bodyZ()
							d = str(c.numWhat(3) + h[2]/2.0)
						else:
							continue
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					if c.tag == "RPP":
						d = (c.numWhat(6) - c.numWhat(5))/2.0
						undoinfo.append(self.flair.setWhatUndo(c, 5, val-d))
						undoinfo.append(self.flair.setWhatUndo(c, 6, val+d))
					elif c.tag in ("RCC","TRC","REC"):
						h = c.bodyZ()
						undoinfo.append(self.flair.setWhatUndo(c, 3, val-h[2]/2.0))
					elif c.tag == "BOX":
						h = c.bodyX() + c.bodyY() + c.bodyZ()
						undoinfo.append(self.flair.setWhatUndo(c, 3, val-h[2]/2.0))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		elif name == "@azm":
			if val is None:
				v = None
				for c in self.cards:
					try:
						if c.tag in ("PLA","P"):
							d = str(math.degrees(c.bodyN().phi()))
						elif c.tag in ("RCC","TRC","REC"):
							d = str(math.degrees(c.bodyZ().phi()))
						else:
							continue
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					if c.tag in ("PLA","P"):
						n     = c.bodyN()
						ma    = n.length()
						phi   = n.phi()
						theta = n.theta()
						n.setPolar(ma, math.radians(val), theta)
						undoinfo.append(self.flair.setWhatUndo(c, 1, n[0]))
						undoinfo.append(self.flair.setWhatUndo(c, 2, n[1]))
						undoinfo.append(self.flair.setWhatUndo(c, 3, n[2]))
					elif c.tag in ("RCC","TRC","REC"):
						n     = c.bodyZ()
						ma    = n.length()
						phi   = n.phi()
						theta = n.theta()
						n.setPolar(ma, math.radians(val), theta)
						undoinfo.append(self.flair.setWhatUndo(c, 4, n[0]))
						undoinfo.append(self.flair.setWhatUndo(c, 5, n[1]))
						undoinfo.append(self.flair.setWhatUndo(c, 6, n[2]))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		elif name == "@polar":
			if val is None:
				v = None
				for c in self.cards:
					try:
						if c.tag in ("PLA","P"):
							d = str(math.degrees(c.bodyN().theta()))
						elif c.tag in ("RCC","TRC","REC"):
							d = str(math.degrees(c.bodyZ().theta()))
						else:
							continue
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					if c.tag in ("PLA","P"):
						n = c.bodyN()
						ma    = n.length()
						phi   = n.phi()
						theta = n.theta()
						n.setPolar(ma, phi, math.radians(val))
						undoinfo.append(self.flair.setWhatUndo(c, 1, n[0]))
						undoinfo.append(self.flair.setWhatUndo(c, 2, n[1]))
						undoinfo.append(self.flair.setWhatUndo(c, 3, n[2]))
					elif c.tag in ("RCC","TRC","REC"):
						n     = c.bodyZ()
						ma    = n.length()
						phi   = n.phi()
						theta = n.theta()
						n.setPolar(ma, phi, math.radians(val))
						undoinfo.append(self.flair.setWhatUndo(c, 4, n[0]))
						undoinfo.append(self.flair.setWhatUndo(c, 5, n[1]))
						undoinfo.append(self.flair.setWhatUndo(c, 6, n[2]))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		# Special for ELL!
		elif name in ("@cx", "@cy", "@cz"):
			idx = ord(name[-1])-ord('x')
			if val is None:
				v = None
				for c in self.cards:
					try:
						F1 = c.bodyP()
						F2 = c.bodyP2()
						C  = 0.5*(F1 + F2)
						d  = C[idx]
						if abs(d)<1e-10: d=0.0
						d  = bmath.format(d,15)
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return v if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					F1 = c.bodyP()
					F2 = c.bodyP2()
					C  = 0.5*(F1 + F2)
					d  = val - C[idx]
					undoinfo.append(self.flair.setWhatUndo(c, 1+idx, F1[idx]+d))
					undoinfo.append(self.flair.setWhatUndo(c, 4+idx, F2[idx]+d))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		# Special for ELL!
		elif name == "@rmajor":
			if val is None:
				v = None
				for c in self.cards:
					try: d = str(c.numWhat(7)/2.0)
					except: continue
					if v is None: v = d
					elif v != d: return ""
				return bmath.format(v,15) if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					# Current
					F1 = c.bodyP()
					F2 = c.bodyP2()
					C  = 0.5*(F1 + F2)
					Z  = F2 - F1
					f  = Z.normalize()/2.0
					RM = c.numWhat(7)/2.0
					rm2 = (RM-f)*(RM+f)
					# rm2 = RM2 - f2

					# New values
					# RM2 = rm2 - f2
					try: f  = math.sqrt(val**2 - rm2)
					except: f = 0.0
					F1 = C + f * Z
					F2 = C - f * Z

					for i in range(3):
						undoinfo.append(self.flair.setWhatUndo(c, 1+i, F1[i]))
						undoinfo.append(self.flair.setWhatUndo(c, 4+i, F2[i]))
					undoinfo.append(self.flair.setWhatUndo(c, 7, 2.0*val))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		# Special for ELL!
		elif name == "@rminor":
			if val is None:
				v = None
				for c in self.cards:
					try:
						F1 = c.bodyP()
						F2 = c.bodyP2()
						Z  = F2 - F1
						f  = Z.normalize()/2.0
						RM = c.numWhat(7)/2.0
						rm = (RM-f)*(RM+f)
						# rm2 = RM2 - f2
						if rm<0.0:
							rm = "0"
						else:
							rm = bmath.format(math.sqrt(rm), 15)
					except: continue
					if v is None: v = rm
					elif v != rm: return ""
				return v if v is not None else ""
			else:
				if val=="": return
				try: val = float(val)
				except ValueError: return
				undoinfo = []
				for c in self.cards:
					# Current
					F1 = c.bodyP()
					F2 = c.bodyP2()
					C  = 0.5*(F1 + F2)
					Z  = F2 - F1
					f  = Z.normalize()/2.0
					RM = c.numWhat(7)/2.0
					rm2 = (RM-f)*(RM+f)
					# New values
					# RM2 = rm2 - f2
					try: f  = math.sqrt(RM**2 - val**2)
					except: f = 0.0
					F1 = C + f * Z
					F2 = C - f * Z
					for i in range(3):
						undoinfo.append(self.flair.setWhatUndo(c, 1+i, F1[i]))
						undoinfo.append(self.flair.setWhatUndo(c, 4+i, F2[i]))
					undoinfo.append(self.flair.setWhatUndo(c, 7, 2.0*RM))
				self.addUndo(undoinfo)
				# FIXME do not reload input but update/reload only the necessary bodies
				self.editor.forceRefresh()

		else:
			return super().value(name, val)

	# ----------------------------------------------------------------------
	def edit(self, listbox):
		name = listbox.get(tk.ACTIVE)[0]
		if name == "type":
			return None
		elif name == "name":
			if self.many: return None
			return tkExtra.InPlaceMaxLength(listbox.listbox(1), maxlength=8, exportselection=True)
		elif name in ("transform", "rotdefi"):
			return tkExtra.InPlaceList(listbox.listbox(1),
				values=self.input.rotationList(True))
		else:
			return super().edit(listbox)

	# ----------------------------------------------------------------------
	def split(self):
		return int(self.listbox.index(tk.ACTIVE)) >= 3

	# ----------------------------------------------------------------------
	def split2(self):
		return int(self.listbox2.index(tk.ACTIVE)) >= 4

#===============================================================================
# Voxel Properties
#===============================================================================
class VoxelProperties(Properties):
	def __init__(self, editor, cards):
		super().__init__(editor, cards, freeze=True, visible=True)
		self.items.extend([ "x", "y", "z", "Trans"])
		self.whats = {"x":1, "y":2, "z":3, "Trans":4}

	# ----------------------------------------------------------------------
	def edit(self, listbox):
		name = listbox.get(tk.ACTIVE)[0]
		if name == "Trans":
			return tkExtra.InPlaceList(listbox.listbox(1),
				values=self.input.rotationList(True))
		else:
			return super().edit(listbox)

	# ----------------------------------------------------------------------
	def split(self):
		return int(self.listbox.index(tk.ACTIVE)) >= 3

	# ----------------------------------------------------------------------
	def split2(self):
		return int(self.listbox2.index(tk.ACTIVE)) >= 4

#===============================================================================
# FIXME
# Allow linking bodies. Bodies refer to other bodies and the only thing that
# changes is the transformation of the body
#===============================================================================
class LinkBodyProperties(Properties):
	def __init__(self, editor, cards):
		super().__init__(editor, cards)

#===============================================================================
# Region properties
#===============================================================================
class RegionProperties(Properties):
	def __init__(self, editor, cards):
		super().__init__(editor, cards, freeze=True)

		if self.many or self.voxel:
			self.items.extend(["material"])
		else:
			self.items.extend(["material", ADDZONE])
		self.items2.extend(["bbox", "alpha", "rotdefi", "field",
				"imp-All", "imp-H", "imp-E", "imp-N"])
		# move material, rotdefi to attributes
		self.nzones = 0

	# ----------------------------------------------------------------------
	def fill(self):
		# Dynamically adjust number of zones
		if not self.many and not self.voxel:
			n = len(self.cards[0].get(Card.ZONE,[]))
			if n:
				if self.nzones != n:
					del self.items[-self.nzones-1:]
					self.nzones = n
					for i in range(1,self.nzones+1):
						self.items.append(f"zone{i:02d}")
					self.items.append(ADDZONE)
			else:
				del self.items[3:]
				self.items.append(ADDZONE)
				self.nzones = 0
		super().fill()
		self.setColor()

	# ----------------------------------------------------------------------
	def setColor(self):
		color = self.editor.viewers.materialColor(self.cards[0][Card.MATERIAL])
		self.listbox.listbox(1).itemconfig(2,background=f"#{color:06X}")
		if not self.many and not self.voxel:
			self.listbox.itemconfig(tk.END, foreground=tkFlair._CLABEL_FOREGROUND_COLOR)

	# ----------------------------------------------------------------------
	def value(self, name=None, val=None):
		if name=="bbox":
			if val is None:
				v = self.cards[0].get(Card.SELECT,0)&BIT.BBOX
				for c in self.cards:
					if v != c.get(Card.SELECT,0)&BIT.BBOX:
						return tkFlair._MULTI
				if v: return tkFlair._ON
				else: return tkFlair._OFF
			else:
				for c in self.cards:
					if val: c[Card.SELECT] = c.get(Card.SELECT,0) | BIT.BBOX
					else:   c[Card.SELECT] = c.get(Card.SELECT,0) & BIT.UNBBOX

		elif name == "material":
			if val is None:
				try:
					v = self.cards[0][Card.MATERIAL].name()
					for c in self.cards:
						if v != c[Card.MATERIAL].name():
							return "-"
					return v
				except:
					return "error"
			else:
				# Set last material
				self.editor._lastMaterial = val

				# find the new material card to set the color
				for card in self.input["MATERIAL"]:
					if card.ignore(): continue
					if card.name()==val:
						for c in self.cards:
							c[Card.MATERIAL] = card
						self.setColor()
						break
				else:
					for card in list(self.input.defaultMaterials.values()) \
						  + list(Fluka.icruMaterials.values()):
						if card.name()==val:
							for c in self.cards:
								c[Card.MATERIAL] = card
							self.setColor()
							break

				# find which ASSIGNMAt card set the material
				# FIXME collect all undo in a list
				undoinfo = []
				for c in self.cards:
					assignmat = c[Card.ASSIGNMAT]
					if assignmat is None:
						# Create a new assignmat card in input
						pos = self.input.bestPosition("ASSIGNMA")
						card = FlukaCard("ASSIGNMA", ["",val,c.name()])
						undoinfo.append(
							self.flair.addCardUndo(card, pos, True))
						c[Card.ASSIGNMAT] = card

					elif assignmat.what(3)=="" or \
					     assignmat.what(2)==assignmat.what(3):
						# Set region assignmat
						undoinfo.append(
							self.flair.setWhatUndo(assignmat, 1, val))

					else:
						# Leave range untouched and save it below
						card = FlukaCard("ASSIGNMA", ["",val,c.name()])
						undoinfo.append(self.flair.addCardUndo(card,
								assignmat.pos()+1, True))
						c[Card.ASSIGNMAT] = card

				if undoinfo:
					self.addUndo(undoinfo)
					self.flair.refresh("card", self.editor)
					self.input.regionProperties()
					self.editor.updateGeometry()
					self.editor.redraw()	# FIXME maybe unnecessary call

		elif name == "field":
			if val is None:
				try:
					v = self.cards[0][Card.ASSIGNMAT].intWhat(5)
					for c in self.cards:
						if v != c[Card.ASSIGNMAT].intWhat(5):
							return "-"
					return _ASSIGNMA_field[v]
				except:
					return ""
			else:
				try:
					ival = _ASSIGNMA_field.index(val)
				except IndexError:
					return
				# find which ASSIGNMAt card set the material
				# FIXME collect all undo in a list
				undoinfo = []
				for c in self.cards:
					assignmat = c[Card.ASSIGNMAT]
					if assignmat is None:
						# Create a new assignmat card in input
						pos = self.input.bestPosition("ASSIGNMA")
						card = FlukaCard("ASSIGNMA",
								["","",c.name(),"","",ival])
						undoinfo.append(
							self.flair.addCardUndo(card, pos, True))
						c[Card.ASSIGNMAT] = card

					elif assignmat.what(3)=="" or \
					     assignmat.what(2)==assignmat.what(3):
						# Set region assignmat
						undoinfo.append(
							self.flair.setWhatUndo(assignmat, 5, ival))

					else:
						# Leave range untouched and save it below
						card = FlukaCard("ASSIGNMA",
								["",c.what(1),c.name(),c.what(3),
								c.what(4),ival,c.what(6)])
						undoinfo.append(
							self.flair.addCardUndo(card,
								assignmat.pos()+1, True))
						c[Card.ASSIGNMAT] = card

				if undoinfo:
					self.addUndo(undoinfo)
					self.flair.refresh("card", self.editor)
					self.input.regionProperties()
					self.editor.updateGeometry()
					self.editor.redraw()	# FIXME maybe unnecessary call

		elif name == "rotdefi":
			if val is None:
				v = self.cards[0].get("@rotdefi","")
				for c in self.cards:
					if v != c.get("@rotdefi",""):
						return "-"
				return v
			else:
				undoinfo = []
				for c in self.cards:
					lattice = c[Card.LATTICE]

					if val=="":
						if lattice is None: continue
						if lattice.what(2)=="" or \
						   lattice.what(1)==lattice.what(2):
							undoinfo.append(
								self.flair.delCardUndo(lattice.pos(), True))
							c["@rotdefi"]   = ""
							c[Card.LATTICE] = None
						else:
							# cannot  rotdefi use flair
							log.say("Cannot remove LATTICE card for REGION ",c.name())
					else:
						try:	# for integer
							rotdefi = f"Rot{int(val):03d}"
						except:
							rotdefi = val

						c["@rotdefi"] = rotdefi
						if lattice is None:
							# Create a new lattice card in input
							pos = self.input.bestPosition("LATTICE")
							card = FlukaCard("LATTICE", [rotdefi,c.name()])
							undoinfo.append(self.flair.addCardUndo(card, pos, True))
							c[Card.LATTICE] = card
						elif lattice.what(2)=="" or lattice.what(1)==lattice.what(2):
							# Set region lattice
							undoinfo.append(
								self.flair.setWhatUndo(lattice, 0, val))
						else:
							# Leave range untouched and save it below
							card = FlukaCard("LATTICE",[rotdefi,c.name()])
							undoinfo.append(
								self.flair.addCardUndo(card, lattice.pos()+1, True))
							c[Card.LATTICE] = card

				if undoinfo:
					self.addUndo(undoinfo)
					self.flair.refresh("card", self.editor)
					self.input.regionProperties()
					self.editor.updateGeometry()
					self.editor.redraw()	# FIXME maybe unnecessary call

		elif name in ("imp-All", "imp-H", "imp-E", "imp-N"):
			if name == "imp-All":
				tag   = "@biasingA"
				what1 = 0
			elif name == "imp-H":
				tag = "@biasingH"
				what1 = 1
			elif name == "imp-E":
				tag = "@biasingE"
				what1 = 2
			elif name == "imp-N":
				tag = "@biasingN"
				what1 = 3

			if val is None:
				biasing = self.cards[0][tag]
				if biasing is not None:
					v = biasing.what(3)
				else:
					v = ""
				for c in self.cards:
					biasing = c[tag]
					if biasing is not None:
						vi = biasing.what(3)
					else:
						vi = ""
					if v != vi: return "-"
				return v
			else:
				undoinfo = []

				for c in self.cards:
					biasing = c[tag]
					if val=="":
						if biasing is None: continue
						if biasing.what(2)=="":
							undoinfo.append(
								self.flair.delCardUndo(biasing.pos(), True))
							c[tag] = None
						else:
							# cannot delete biasing use flair
							log.say("Cannot remove BIASING card for REGION", c.name())
					else:
						if biasing is None:
							# Create a new biasing card in input
							pos = self.input.bestPosition("BIASING")
							card = FlukaCard("BIASING", ["",what1,"",val,c.name()])
							undoinfo.append(self.flair.addCardUndo(card, pos, True))
							c[tag] = card
						elif biasing.what(5)=="" or biasing.what(4)==biasing.what(5):
							# Set region importance
							undoinfo.append( self.flair.setWhatUndo(biasing, 3, val))
						else:
							# Leave range untouched and save it below
							card = FlukaCard("BIASING", ["",what1,"",val,c.name()])
							undoinfo.append(
								self.flair.addCardUndo(card, biasing.pos()+1, True))
							c[tag] = card

				if undoinfo:
					self.addUndo(undoinfo)
					self.flair.refresh("card", self.editor)
					# FIXME do not reload input but update/reload only the necessary bodies
					self.editor.forceRefresh()

		elif name.startswith("zone"):
			if self.many:
				return "-"
			i = int(name[4:])-1
			zones = self.cards[0][Card.ZONE]
			if i >= len(zones): return ""
			zone = zones[i]
			#rid = self.cards[0][Card.ID]
			#fmt,zone = self.editor.geometry().region(rid,"zone",i)
			#zone = csg.toString(self.project.ismode(Mcnp,Phits),fmt,zone)
			if val is None:
				return zone
			else:
				if val != zone:
					# create new expression by substituting the zone
					zones[i] = val
					self.flair.addUndo(self.flair.setZonesUndo(self.cards[0],zones))
					self.flair.refresh("redraw", self.editor)
					self.editor.forceRefresh()

		elif name == ADDZONE:
			if self.many:
				return "-"
			elif val is None:
				return ""
			elif val:
				self.flair.addUndo(
					self.flair.addZoneUndo(self.cards[0], val))
				self.flair.refresh("redraw", self.editor)
				self.editor.forceRefresh()

		elif name == "alpha":
			if val is None:
				v = self.cards[0].get(Card.ALPHA,0)
				for c in self.cards:
					if v != c.get(Card.ALPHA,0):
						return ""
				return v
			else:
				try: val = int(val)
				except: val = 0
				if val<0: val=0
				elif val>100: val=100

				undoinfo = []
				for c in self.cards:
					undoinfo.append(self.flair.setCardUndo(c, Card.ALPHA, val, 0))

				if undoinfo:
					self.addUndo(undoinfo)
					self.editor.redraw()
					self.editor.inputModified()

		else:
			return super().value(name, val)

	# ----------------------------------------------------------------------
	def edit(self, listbox):
		name = listbox.get(tk.ACTIVE)[0]
		if name == "material":
			return tkExtra.InPlaceList(listbox.listbox(1),
				values=self.input.materialList(extended=True))
		elif name == "rotdefi":
			return tkExtra.InPlaceList(listbox.listbox(1),
				values=self.input.rotationList(True))
		elif name.startswith("zone") or name==ADDZONE:
			if self.many: return None
			return InPlaceRegion(listbox.listbox(1), self.flair)
		elif name == "name":
			if self.many: return None
			return tkExtra.InPlaceMaxLength(listbox.listbox(1),
						maxlength=8, exportselection=True)
		elif name == "field":
			return tkExtra.InPlaceList(listbox.listbox(1), values=_ASSIGNMA_field)
		elif name == "alpha":
			return tkExtra.InPlaceInteger(listbox.listbox(1), exportselection=True)
		else:
			return super().edit(listbox)

#===============================================================================
class BeamProperties(Properties):
	def __init__(self, editor, cards):
		super().__init__(editor, cards, visible=True)
		self.items.extend(["energy","scale","color"])

	# ----------------------------------------------------------------------
	def fill(self):
		super().fill()
		self.setColor()

	# ----------------------------------------------------------------------
	def setColor(self):
		try:
			self.listbox.listbox(1).itemconfig(4,
				background=self.cards[0].get("color",""))
		except tk.TclError:
			pass

	# ----------------------------------------------------------------------
	def value(self, name=None, val=None):
		if name == "name":
			return super().value(name,val)

		elif name == "energy":
			if val is None:
				e = None
				for c in self.cards:
					if c.tag=="BEAM":
						energy = c.numWhat(1)
					else:
						energy = c.numWhat(2)
					if e is None:
						e = energy
					elif e != energy: return ""
				return e
			else:
				undoinfo = []
				for c in self.cards:
					if c.tag=="BEAM":
						w = 1
					else:
						w = 2
					undoinfo.append(self.flair.setWhatUndo(c, w, val))
				if undoinfo:
					self.addUndo(undoinfo)
					self.flair.refresh("card", self.editor)
					# FIXME do not reload input but update/reload
					# only the necessary bodies
					self.editor.forceRefresh()

		elif name == "scale":
			if val is None:
				v = self.cards[0].get("scale",1.0)
				for c in self.cards:
					if v != c.get("scale",1.0): return ""
				return v
			else:
				undoinfo = []
				for c in self.cards:
					undoinfo.append(self.flair.setCardUndo(c, "scale", val, 1.0))
				if undoinfo:
					self.addUndo(undoinfo)
					# FIXME do not reload input but update/reload
					# only the necessary bodies
					self.editor.forceRefresh()

		elif name == "color":
			if val is None:
				v = self.cards[0].get("color","")
				for c in self.cards:
					if v != c.get("color",""): return ""
				return v
			else:
				try:
					for c in self.cards:
						self.flair.addUndo(
							self.flair.setCardUndo(c, "color", val, 0))
				except:
					pass
				self.editor.forceRefresh()

		else:
			return super().value(name, val)

	# ----------------------------------------------------------------------
	def edit(self, listbox):
		name = listbox.get(tk.ACTIVE)[0]
		if name == "color":
			return tkExtra.InPlaceColor(listbox.listbox(1))
		else:
			return super().edit(listbox)

#===============================================================================
class RotdefiProperties(Properties):
	def __init__(self, editor, cards):
		super().__init__(editor, cards, visible=True)
		name = self.cards[0].name()
		self._rotdefi = [x
			for x in self.input.cardsSorted(Card.ROTATION,-1).values()
				if x.name()==name]

	# ----------------------------------------------------------------------
	def fill(self):
		# Dynamically adjust number of cards
		if not self.many:
			n = (len(self.items)-2) // 6
			if n != len(self._rotdefi):
				del self.items[2:]
				for i in range(1,len(self._rotdefi)+1):
					self.items.append(f"dx.{i}")
					self.items.append(f"dy.{i}")
					self.items.append(f"dz.{i}")
					self.items.append(f"axis.{i}")
					self.items.append(f"azm.{i}")
					self.items.append(f"polar.{i}")
				self.items.append("<add>")
		super().fill()

	# ----------------------------------------------------------------------
	def value(self, name=None, val=None):
		# Search all patterns with prefix and what#
		for prefix,what in (("dx.",4), ("dy.",5), ("dz.",6), ("azm.",3), ("polar.",2)):
			if not name.startswith(prefix): continue
			if self.many: return "-"
			try:
				idx = int(name[len(prefix):])-1
				card = self._rotdefi[idx]
			except:
				if val is None: return ""
				# Add a new card!

			if val is None:
				return card.what(what)
			else:
				self.flair.addUndo(
					self.flair.setWhatUndo(
						card, what, val))
				# FIXME do not reload input but update/reload only the necessary cards
				self.editor.forceRefresh()
			return

		if name.startswith("axis."):
			if self.many: return "-"
			idx = int(name[5:])-1
			card = self._rotdefi[idx]

			w1 = card.intWhat(1)
			if w1 < 1000:
				(j, i) = divmod(w1, 100)
			else:
				(i, j) = divmod(w1, 1000)

			if val is None:
				try: return _ROTDEFI_axes[j]
				except: return "?"

			else:
				try: j = _ROTDEFI_axes.index(val)
				except: return
				if w1 < 1000:
					w1 = i + j*100
				else:
					w1 = j + i*1000
				self.flair.addUndo(self.flair.setWhatUndo(card, 1, w1))
				# FIXME do not reload input but update/reload only the necessary cards
				self.editor.forceRefresh()

		elif name == "<add>":
			if val is None:
				return ""
			else:
				# add a new ROTDEFI card
				pos = self._rotdefi[-1].pos()+1
				if self.flair.project.ismode(Moira):
					if self._rotdefi[0].name().find("MATRIX") != -1:
						tag = "MATRIX" # FIXME awful implementation
					else:
						tag = "ROTM"
					card = MoiraCard(tag,[self._rotdefi[0].name()])
				elif self.flair.project.ismode(Fluka):
					card = FlukaCard("ROT-DEFI",[self._rotdefi[0].name()])
				else:
					assert False
				self.flair.addUndo(self.flair.addCardUndo(card, pos, True))
				self.flair.refresh("card", self.editor)
				self._rotdefi.append(card)
				self.fill()

		else:
			return super().value(name, val)

	# ----------------------------------------------------------------------
	def edit(self, listbox):
		name = listbox.get(tk.ACTIVE)[0]
		if name.startswith("axis."):
			return tkExtra.InPlaceList(listbox.listbox(1),
				values=_ROTDEFI_axes)
		elif name == "<add>":
			return tkExtra.InPlaceList(listbox.listbox(1),
				values=["","add"])
		return super().edit(listbox)

	# ----------------------------------------------------------------------
	def split(self):
		return int(self.listbox.index(tk.ACTIVE)) >= 2

#===============================================================================
class TRProperties(Properties):
	def __init__(self, editor, cards):
		super().__init__(editor, cards, visible=True)
		#name = self.cards[0].name()
		self.initLayout(cards[0])

#===============================================================================
class ObjectProperties(Properties):
	def __init__(self, editor, cards):
		super().__init__(editor, cards, visible=True)
		self.items.append("type")
		self.objtype = cards[0].tag
		for c in cards:
			if self.objtype != c.tag:
				self.objtype = None
				break
		if self.objtype:
			self.objprop, self.option = _OBJECTS[self.objtype]
			self.items.extend([x for x,y in self.objprop])
		else:
			self.objprop = self.option = None

	# ----------------------------------------------------------------------
	def fill(self):
		super().fill()
		self.setColor()

	# ----------------------------------------------------------------------
	def setColor(self):
		if self.objprop is not None:
			try:
				self.listbox.listbox(1).itemconfig(6,
					background=self.cards[0].colorWhat(6))
			except tk.TclError:
				pass

	# ----------------------------------------------------------------------
	def value(self, name=None, val=None):
		if name == "name":
			return super().value(name,val)

		elif name == "type":
			if val is None:
				if self.objprop: return self.cards[0].tag
				else: return "-"

		elif name == "option":
			if val is None:
				v = self.cards[0].intWhat(4)
				for c in self.cards:
					if v != c.intWhat(4): return ""
				try: return self.option[v]
				except: return self.option[0]
			else:
				try:
					opt = self.option.index(val)
					# FIXME undoinfo!!!
					undoinfo = []
					for c in self.cards:
						undoinfo.append(self.flair.setWhatUndo(c, 4, opt))
					self.addUndo(undoinfo)
				except IndexError:
					pass
				self.editor.forceRefresh()

		elif name == "size":
			if val is None:
				s,a = divmod(self.cards[0].intWhat(5),100)
				for c in self.cards:
					if s != c.intWhat(5) // 100: return ""
				return s
			else:
				s,a = divmod(self.cards[0].intWhat(5),100)
				try:
					ival = int(val)
					for c in self.cards:
						self.flair.addUndo(
							self.flair.setWhatUndo(
								c, 5, ival*100+a))
				except:
					pass
				self.editor.forceRefresh()

		elif name == "anchor":
			if val is None:
				s,a = divmod(self.cards[0].intWhat(5),100)
				for c in self.cards:
					if a != c.intWhat(5) % 100: return ""
				try:
					return _ANCHOR_list[a]
				except:
					return _ANCHOR_list[0]
			else:
				s,a = divmod(self.cards[0].intWhat(5),100)
				try:
					ival = _ANCHOR_list.index(val)
					for c in self.cards:
						self.flair.addUndo(
							self.flair.setWhatUndo(
								c, 5, s*100+ival))
				except IndexError:
					pass
				self.editor.forceRefresh()

		elif name == "color":
			if val is None:
				v = self.cards[0].what(6)
				for c in self.cards:
					if v != c.what(6): return ""
				return v
			else:
				try:
					for c in self.cards:
						self.flair.addUndo(
							self.flair.setWhatUndo(c, 6, val))
				except:
					pass
				self.editor.forceRefresh()

		elif name == "falloff":
			if val is None:
				v = self.cards[0].intWhat(11)
				for c in self.cards:
					if v != c.intWhat(11): return ""
				try:
					return _FALLOFF_list[v]
				except IndexError:
					return _FALLOFF_list[0]
			else:
				try:
					for c in self.cards:
						self.flair.addUndo(
							self.flair.setWhatUndo(
								c, 11, _FALLOFF_list.index(val)))
				except:
					pass
				self.editor.forceRefresh()

		elif name == "@length":
			if self.many: return None
			card = self.cards[0]
			if card.tag not in ("arrow", "ruler"): return None
			P = bmath.Vector(card.numWhat(1), card.numWhat(2), card.numWhat(3))
			D = bmath.Vector(card.numWhat(7), card.numWhat(8), card.numWhat(9))
			if card.tag == "ruler": D -= P
			if val is None:
				return bmath.format(D.length(),15)
			else:
				D.normalize()
				try:
					undoinfo = [self.flair.refreshUndo(),
						self.flair.saveCardUndoInfo(card.pos())]
					D = float(val)*D
					if card.tag == "ruler": D += P
					card.setWhat(7, bmath.format(D[0],15))
					card.setWhat(8, bmath.format(D[1],15))
					card.setWhat(9, bmath.format(D[2],15))
					undoinfo.append(self.flair.refreshUndo())
					self.flair.addUndo(undoinfo,"object length")
				except:
					pass

		elif name == "@angle":
			if self.many: return None
			card = self.cards[0]
			P = bmath.Vector(card.numWhat( 1), card.numWhat( 2), card.numWhat( 3))
			E = bmath.Vector(card.numWhat( 7), card.numWhat( 8), card.numWhat( 9))
			A = bmath.Vector(card.numWhat(10), card.numWhat(11), card.numWhat(12))
			D  = E-P
			D2 = A-P
			l = D.normalize()
			D2.normalize()
			if val is None:
				return bmath.format(math.degrees(math.acos(D*D2)),15)
			else:
				try: ang = math.radians(float(val))
				except: return
				W = D2 ^ D		# Find perpendicular vector
				l = W.normalize()
				if l>0.001:
					M = bmath.Matrix(4)	# Create rotation matrix
					M.rotate(ang,W)
					E = l*(M*D2) + P	# Find new ending position
					undoinfo = [self.flair.refreshUndo(),
						self.flair.saveCardUndoInfo(card.pos())]
					card.setWhat(7, bmath.format(E[0],15))
					card.setWhat(8, bmath.format(E[1],15))
					card.setWhat(9, bmath.format(E[2],15))
					undoinfo.append(self.flair.refreshUndo())
					self.flair.addUndo(undoinfo,"object length")

		else:
			if self.objprop is None:
				return super().value(name, val)

			for lbl,what in self.objprop:
				if name==lbl: break
			else:
				return super().value(name, val)

			if val is None:
				v = self.cards[0].what(what)
				for c in self.cards:
					if v != c.what(what): return ""
				return v
			else:
				#if val=="": return
				for c in self.cards:
					self.flair.addUndo(
						self.flair.setWhatUndo(c, what, val))
				self.editor.refreshLate()

	# ----------------------------------------------------------------------
	def edit(self, listbox):
		name = listbox.get(tk.ACTIVE)[0]
		if name == "type":
			return None
		elif name == "option":
			return tkExtra.InPlaceList(listbox.listbox(1), values=self.option)
		elif name == "anchor":
			return tkExtra.InPlaceList(listbox.listbox(1), values=_ANCHOR_list)
		elif name == "falloff":
			return tkExtra.InPlaceList(listbox.listbox(1), values=_FALLOFF_list)
		elif name == "color":
			return tkExtra.InPlaceColor(listbox.listbox(1))
		elif name == "file":
			edit = tkExtra.InPlaceFile(listbox.listbox(1),
				title="Load file",
				filetypes=(("STL files","*.stl"),("All","*")),
				save=False)
			if edit.value is not None:	# convert to int
				edit.value = self.flair.project.relpath(edit.value)
			return edit
		else:
			return super().edit(listbox)

	# ----------------------------------------------------------------------
	def split(self):
		return int(self.listbox.index(tk.ACTIVE)) >= 7

#===============================================================================
class MaterialProperties(Properties):
	BXDF  = [
			"Default",
			"Plastic smooth",
			"Plastic rough",
			"Plastic rough 2",
			"Shiny metal",
			"Shiny metal rough",
			"Shiny metal rough 2",
			"Brushed metal",
			"Substrate rough",
			"Substrate polished",
			"Metallic paint",
			"Concrete rough",
			"Concrete polished",
			"Minnaert",
			"Toon",
			"Translucent",
			"Refractive glass",
			"Clear glass/Water",
			"Dispersive glass",
			"High dispersive glass",
		]

	#-----------------------------------------------------------------------
	def __init__(self, editor, cards):
		super().__init__(editor, cards, visible=True)
		self.items.extend(Project.MaterialInfo.DEFAULT.keys())
		self.items2.extend(["global"])

	# ----------------------------------------------------------------------
	def fill(self):
		super().fill()
		self.setColor()

	# ----------------------------------------------------------------------
	def setColor(self):
		color = self.flair.project.getMaterial(self.cards[0].name())("color")
		self.listbox.listbox(1).itemconfig(2, background=color)

	# ----------------------------------------------------------------------
	def value(self, name=None, val=None):
		if name == "name":
			return super().value(name,val)

		elif name in Project.MaterialInfo.DEFAULT:
			project = self.flair.project
			if val is None:
				v = project.getMaterial(self.cards[0].name())(name)
				for c in self.cards:
					if v != project.getMaterial(c.name())(name): return ""
				return v
			else:
				for c in self.cards:
					mat = project.getMaterial(c.name())
					if mat.isGlobal():
						project.material2Local(c.name())
						mat = project.getMaterial(c.name())
					mat[name] = val
				self.editor.loadMaterialColors()
				self.editor.inputModified()

		elif name=="global":
			project = self.flair.project
			if val is None:
				v = project.getMaterial(self.cards[0].name()).isGlobal()
				for c in self.cards:
					if v != project.getMaterial(c.name()).isGlobal(): return ""
				if v: return tkFlair._ON
				else: return tkFlair._OFF
			else:
				for c in self.cards:
					if val: project.material2Global(c.name())
					else:   project.material2Local(c.name())

		else:
			return super().value(name, val)

	# ----------------------------------------------------------------------
	def edit(self, listbox):
		name = listbox.get(tk.ACTIVE)[0]
		if name == "color":
			return tkExtra.InPlaceColor(listbox.listbox(1))
		elif name == "bxdf":
			return tkExtra.InPlaceList(listbox.listbox(1),
						values=MaterialProperties.BXDF)
		else:
			return super().edit(listbox)

#===============================================================================
# Body Menu
#===============================================================================
class _BodyMenuButton(FlairRibbon.FlairMenu):
	def createMenu(self):
		if self.page.project is None: return
		menu = tk.Menu(self.page.ribbon, tearoff=1)
		tkFlair.bodiesMenu(menu, self.page.project.mode(),
					self.page.viewers.newBody)
		return menu

#===============================================================================
class _ObjectMenuButton(FlairRibbon.FlairMenu):
	def createMenu(self):
		if self.page.project is None: return
		menu = tk.Menu(self.page.ribbon, tearoff=1)
		self.page.viewers._addObjectMenu(menu)
		return menu

#===============================================================================
# Layer Menu
#===============================================================================
class _LayerMenuButton(Ribbon.MenuButton):
	def __init__(self, master, page, **kw):
		super().__init__(master, None, **kw)
		self.page = page

	#----------------------------------------------------------------------
	def showMenu(self):
		if self.page.project is None: return
		menu = tk.Menu(self, tearoff=0)
		for i,name in enumerate(self.page.project.geometry.layersList()):
			menu.add_command(label=name,
				command=lambda v=self.page.viewers,i=i: v.setLayers(i))
		self._showMenu(menu)

#===============================================================================
# Geometry Editor Empty
#===============================================================================
class GeometryEditorEmpty(FlairRibbon.FlairPage):
	_name_ = "Geometry"
	_icon_ = "geoedit"

	"""Graphical geometry editor"""
	# ----------------------------------------------------------------------
	def createPage(self):
		super().createPage(False)
		fn=os.path.join(tkFlair.prgDir,"images","grab-geometry-page.gif")
		self.img = tk.PhotoImage(file=fn)
		l = tk.Label(self.frame,
			image=self.img,
			text="Please install the geoviewer module",
			foreground="Darkred",
			background="LightYellow",
			compound=tk.BOTTOM)
		l.pack(expand=tk.YES, fill=tk.BOTH)
		self.splitter = None

	# ----------------------------------------------------------------------
	# Mockup methods
	# ----------------------------------------------------------------------
	def _empty_(self, *args): pass
	loadProject      = _empty_
	refreshUndo      = _empty_
	loadState        = _empty_
	saveState        = _empty_
	activateIfNeeded = _empty_

#===============================================================================
# Actions group menu
#===============================================================================
class _ActionsGroup(FlairRibbon.FlairMenuGroup):
	#----------------------------------------------------------------------
	def createMenu(self):
		menu = tk.Menu(self, tearoff=0)

		menu.add_radiobutton(
				label="Center",
				image=icons.get("center"), compound=tk.LEFT,
				variable=self.page._mouseAction,
				accelerator="C",
				value=ACTION.CENTER)

		menu.add_radiobutton(
				label="Correct",
				image=icons.get("edit"), compound=tk.LEFT,
				accelerator="E",
				variable=self.page._mouseAction,
				value=ACTION.EDIT)
		#tkExtra.Balloon.set(b, "Correct region by removing overlaps")

		menu.add_radiobutton(
				label="Set Pivot",
				image=icons.get("orbit"), compound=tk.LEFT,
				accelerator="P",
				variable=self.page._mouseAction,
				value=ACTION.PIVOT)

		#menu.add_radiobutton(
		#		label="Pencil",
		#		image=icons.get("pencil"), compound=tk.LEFT,
		#		state=tk.DISABLED,
		#		accelerator="P",
		#		variable=self.page._mouseAction,
		#		value=ACTION.PEN)

		menu.add_checkbutton(
				label="Snap",
				image=icons.get("snap"), compound=tk.LEFT,
				variable = self.page.snap2Grid)
		return menu

#===============================================================================
# Tool group menu
#===============================================================================
class _ToolGroup(FlairRibbon.FlairMenuGroup):
	#----------------------------------------------------------------------
	def createMenu(self):
		menu = tk.Menu(self, tearoff=0)
		menu.add_command(label="Link",
			underline=0,
			image=icons.get("link"),
			compound=tk.LEFT,
			command=self.page.viewers.setActionLink)
		menu.add_command(label="Planner",
			underline=0,
			image=icons.get("planner"),
			state=tk.NORMAL,
			compound=tk.LEFT,
			command=self.page.viewers.setActionPlanner)
		return menu

#===============================================================================
# View group menu
#===============================================================================
class _ViewGroup(FlairRibbon.FlairMenuGroup):
	#----------------------------------------------------------------------
	def createMenu(self):
		menu = tk.Menu(self, tearoff=0)

		menu.add_command(label="Time frame",
			underline=0,
			image=icons.get("frame"),
			compound=tk.LEFT,
			state=tk.NORMAL,
			accelerator="k",
			command=self.page.toggleTimeframe)

		menu.add_separator()

		menu.add_command(label="Reset",
				underline=0,
				image=icons.get("reset"),
				compound=tk.LEFT,
				command=self.page.resetEditor)

		# --- Orientation systems cascade ---
		orient_menu = tk.Menu(menu, tearoff=0)
		menu.add_cascade(label="Orientation",
				image=icons.get("axes"),
				compound=tk.LEFT,
				menu=orient_menu)
		for name in sorted(GeometryViewer.ORIENTATIONS.keys()):
			orient_menu.add_radiobutton(label=name,
					variable=self.page.orientSystem,
					command=self.page.orientationSystemChange)
		return menu

#===============================================================================
# Geometry Editor Window
#===============================================================================
class GeometryEditor(FlairRibbon.FlairPage):
	"""Graphical geometry editor"""

	_name_ = "Geometry"
	_icon_ = "geoedit"

	EXPANSION_LIMIT = 20000

	#----------------------------------------------------------------------
	def init(self):
		# project variables
		self._inFocus     = False
		self._projectName = None	# XXX FIXME should be eliminated
						# check the project id
						# since now we have a new project
						# every time we load one
		self.editObject     = None	# Object currently being edited
		self.__afterRefresh = None	# Undo after
		self._loadStatePending = True	# Make a refresh on first load
		self._loadingProject = False	# Do not send any update to flair while loading project
		self._late_messages  = []	# late messages to be displayed after reload

		self.refreshOn  = ("input", "card", "voxel", "plot", "project", "run", "mode")

	#----------------------------------------------------------------------
	def _geometry(self):
		return self.viewers._geometry

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Actions ==========
		group = _ActionsGroup(self.ribbon, self, "Actions")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ========== Select-Tool ==========
		col,row = 0,0
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Select",
				image=icons.get("select32"),
				compound=tk.TOP,
				variable=self._mouseAction,
				value=ACTION.SELECT,
				padx=5)
		b.grid(row=row, rowspan=3, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Select [s]")

		# ======== Pan/Rotate/Select ========
		col,row=1,0
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Pan",
				image=icons.get("pan"),
				variable=self._mouseAction,
				value=ACTION.PAN)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Pan viewport [x]")

		# ---
		col,row=1,1
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Orbit",
				image=icons.get("orbit"),
				variable=self._mouseAction,
				value=ACTION.ORBIT)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Orbit viewport [t]")

		# ---
		col,row=1,2
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Info",
				image=icons.get("target"),
				variable=self._mouseAction,
				value=ACTION.INFO)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Display body and region information on cursor location [i]")

		# ======== Zoom ========
		col,row = 2,0
		b = Ribbon.LabelRadiobutton(group.frame,
				#text="Zoom In",
				image=icons.get("zoom_in"),
				variable=self._mouseAction,
				value=ACTION.ZOOM_IN)
		b.grid(row=row, column=col, padx=5, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Drag a rectangle to zoom in")

		# ----
		col,row = 2,1
		b = Ribbon.LabelRadiobutton(group.frame,
				#text="Zoom Out",
				image=icons.get("zoom_out"),
				variable=self._mouseAction,
				value=ACTION.ZOOM_OUT)
		b.grid(row=row, column=col, padx=5, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Drag a rectangle to zoom out")

		# ----
		col,row = 2,2
		b = Ribbon.LabelButton(group.frame,
				#text="Zoom On",
				image=icons.get("zoom_on"),
				command=self.zoomOnSelected)
		b.grid(row=row, column=col, padx=5, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Zoom on selected objects [f]")

		# ========== Card ==========
		group = Ribbon.LabelGroup(self.ribbon, "Card")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ----
		col,row = 0,0
		b = _BodyMenuButton(group.frame, self,
				text="Body",
				image=icons.get("cone32"),
				compound=tk.TOP)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add Body [b]. Repeat body [B]")

		# ----
		col,row = 1,0
		b = Ribbon.LabelButton(group.frame,
				text="Zone",
				image=icons.get("zone32"),
				compound=tk.TOP,
				command=self.viewers.setActionZone)
		b.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add Zone to select region, else create "
					"a new region to add zone [d]\n"
					"Paint operation with [D]")

		# ---
		col,row=2,0
		b = Ribbon.LabelButton(group.frame,
				text="Region",
				image=icons.get("REGION"),
				command=self.addRegion)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Add REGION card [R]")

		# ----
		col,row=2,1
		b = Ribbon.LabelButton(group.frame, image=icons.get("ROT-DEFI"),
				text="Rotation",
				command=self.addRotdefi)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add transformation card(s) (Fluka:ROT-DEFI)")

		# ----
		col,row=2,2
		b = _ObjectMenuButton(group.frame, self,
				text="Object \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("pin"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add a flair object [space]")

		# ----
		col,row=4,0
		b = Ribbon.LabelButton(group.frame,
				text="Delete",
				image=icons.get("x"),
				command=self.delete)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Delete cards and remove their references [Del] or [BackSpace]")

		# ----
		# FIXME combine with the ROT-DEFI maybe a drop down menu
		#col,row=4,1
		#b = Ribbon.LabelButton(group.frame,
		#		image=icons.get("ROT-DEFI"),
		#		text="MATRIX",
		#		command=self.addRotMat)
		#b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.NSEW)
		#tkExtra.Balloon.set(b, "Add MATRIX card(s)")

		# ----
		col,row=4,2
		b = Ribbon.LabelButton(group.frame,
				text="Clone",
				image=icons.get("clone"),
				command=self.clone)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Clone cards and all associated cards to them [Ctrl-D]")

		# ========== Tools ==========
		group = _ToolGroup(self.ribbon, self, "Tools")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ----
		col,row=0,0
		b = Ribbon.LabelButton(group.frame,
				text="Move",
				image=icons.get("move"),
				state=tk.NORMAL if Global.developer else tk.DISABLED,
				command=self.viewers.setActionMove)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Move selected objects [g]rab")

		# ----
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text="Rotate",
				image=icons.get("rotate"),
				state=tk.NORMAL if Global.developer else tk.DISABLED,
				command=self.viewers.setActionRotate)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Rotate selected objects [r]")

		# ----
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text="Repeat",
				image=icons.get("repeat"),
				state=tk.NORMAL if Global.developer else tk.DISABLED,
				command=self.viewers.setActionRepeat)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Repeat selected objects [?]")

		# ====
		col,row=1,0
		b = Ribbon.LabelButton(group.frame,
				text="Volume",
				image=icons.get("RPP"),
				command=self.viewers.setActionVolume)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Calculate volume of selected regions")

		# ----
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text="Export",
				image=icons.get("save"),
				command=self.viewers.setActionExport)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Export viewport to image or another format [Ctrl-P]")

		# ----
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text="Movie",
				image=icons.get("movie"),
				state=tk.NORMAL,
				command=self.viewers.setActionMovie)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Create a Movie")

		# FIXME add common tools
		# Move
		# Rotate
		# Clone/Repeat
		# Export
		# Movie
		# ???

		# ========== Selection ==========
		group = Ribbon.LabelGroup(self.ribbon, "Selection")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ----
		col,row = 0,0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("light32"),
				anchor=tk.CENTER,
				command=self.toggleVisibility)
		b.grid(row=row, rowspan=2, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Toggle Visibility [v]")

		col,row = 0,2
		menulist = [	("Show",      "light_on",   self.setVisibility),
				("Hide",      "light_off",  self.unsetVisibility),
				None,
				("Show All",  "light_on",   self.setAllVisibility),
				("Hide All",  "light_off",  self.unsetAllVisibility)]
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Visibility",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Set/Unset visibility")

		# ---
		col,row=1,0
		b = Ribbon.LabelCheckbutton(group.frame,
				text="Lock",
				image=icons.get("lock"),
				compound=tk.LEFT,
				variable=self.lockEditing,
				command=self.lockEditingCommand)
				#value=ACTION.SELECT,
		b.grid(row=row, column=col, columnspan=3, padx=0, pady=0, sticky=tk.NSEW)
		#b = Ribbon.LabelButton(group.frame,
		#		image=icons.get("lock"),
		#		command=self.toggleLock)
		#b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		#tkExtra.Balloon.set(b, "Toggle editing lock [l]")
		#menulist = [	("Lock Bodies",   "lock",     self.lock),
		#		("Unlock bodies", "unlock",   self.unlock),
		#		None,
		#		("Lock All bodies",  "lock",   self.lockAll),
		#		("Unlock All bodies","unlock", self.unlockAll)]
		#b = Ribbon.MenuButton(group.frame, menulist,
		#		image=icons.get("triangle_down"),
		#		text="Lock",
		#		compound=tk.RIGHT)
		#b.grid(row=row, column=col+1, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Lock/Unlock graphical editing")

		# ---
		col,row=1,1
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("freeze"),
				command=self.toggleFreeze)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Freeze selection [F]")

		menulist = [	("Freeze",     "freeze",     self.freeze),
				("Unfreeze",   "heat",       self.unfreeze),
				None,
				("Unfreeze All","heat",      self.unfreezeAll)]
		b = Ribbon.MenuButton(group.frame, menulist,
				image=icons.get("triangle_down"),
				text="Freeze",
				compound=tk.RIGHT,
				anchor=tk.W)
		b.grid(row=row, column=col+1, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Freeze/Unfreeze selection")

		# ---
		col,row=1,2
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("wireframe"),
				command=self.toggle3D)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Toggle wireframe [#]")

		menulist = [	("Show",     "wireframe",     self.set3D),
				("Hide",     "wireframe_off", self.unset3D),
				None,
				("Hide All", "wireframe_off", self.unset3DAll)]
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Wireframe",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.W)
		b.grid(row=row, column=col+1, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Set/Unset wireframe")

		# ========== View ==========
		group = _ViewGroup(self.ribbon, self, "View")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		group.grid3rows()
		group.frame.grid_columnconfigure(1, weight=1)
		group.frame.grid_columnconfigure(3, weight=1)

		# ===
		col,row=0,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("run"))
		b.grid(row=row, column=col, padx=3, pady=0, sticky=tk.E)
		tkExtra.Balloon.set(b, "Select active run settings")

		self.runCombo = Ribbon.LabelCombobox(group.frame,
					command=self.runChanged,
					width=24)
		self.runCombo.grid(row=row, column=col+1, columnspan=4,
					padx=0, pady=0, sticky=tk.EW)

		# ---
		col,row=0,1
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("layers"),
				command=self.viewers.resetLayers)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Reset layers to default on all viewports")

		b = _LayerMenuButton(group.frame, self,
				text="Layer",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.W)
		b.grid(row=row, column=col+1, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Select layer for all viewports")

		# ---
		col,row=2,1
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("layout"),
				command=self.viewLayout)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Select windows layout 2x2 [#]")

		menulist = []
		for name in GeometryViewer.VIEWORDER:
			icon = "layout_"+name.lower()
			menulist.append((name, icon,
				lambda s=self,n=name : s.viewLayout(n)))

		b = Ribbon.MenuButton(group.frame, menulist,
				text="Layout",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.W)
		b.grid(row=row, column=col+1, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Select windows layout [w,W]")

		# ---
		col,row=0,2
		b = Ribbon.LabelButton(group.frame,
				text="Refresh",
				image=icons.get("refresh"),
				command=self.reload)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Refresh all viewports")

		# ---
		col,row=2,2
		b = Ribbon.LabelCheckbutton(group.frame,
				text="Synchronize",
				image=icons.get("sync"),
				compound=tk.LEFT,
				variable=self.syncViews,
				command=self.synchronizeViewports)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Synchronize the display of the Red,Blue,Magenta viewports")

		#menulist = []
		#for i in range(len(GeometryViewer.VIEWER.NAME)):
		#for i,n in enumerate(VIEWER.NAME):
		#	cmd  = lambda p=self.page,v=i: p.export(v)
		#	menulist.append((n,"save",cmd))

		#col,row=2,2
		#b = Ribbon.MenuButton(group.frame, menulist,
		#		text="Viewport",
		#		image=icons.get("save"),
		#b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.EW)
		#tkExtra.Balloon.set(b, "Save view port into image")

	#----------------------------------------------------------------------
	# Ribbon method to create context groups
	#----------------------------------------------------------------------
	def createContextGroups(self):
		# ========== Body ==========
		group = Ribbon.LabelGroup(self.ribbon, "Body")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		self.addContextGroup("body", group)
		group.grid2rows()

		# ---
		col,row=0,0
		b = Ribbon.LabelButton(group.frame,
				text="Rotdefi",
				image=icons.get("ROT-DEFI"),
				command=self.gotoRotdefi)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Show body ROT-DEFI transformation card")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text="Transform",
				image=icons.get("axes"),
				command=self.transformGeometry)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Transform bodies (Ctrl-T)")

		# ---
		col,row=1,0
		b = Ribbon.LabelButton(group.frame,
				text="Expand",
				image=icons.get("expand_parenthesis"),
				command=self.expandBody)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Expand macro bodies to infinite ones")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text="Orthogonal",
				image=icons.get("BOX"),
				command=self.makeOrthogonal)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Correct for orthogonality BOX, REC,...")

		# ========== Region ==========
		group = Ribbon.LabelGroup(self.ribbon, "Region")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		self.addContextGroup("region", group)
		group.grid3rows()

		# ---
		col,row=0,0
		b = Ribbon.LabelButton(group.frame,
				text="Material",
				image=icons.get("MATERIAL"),
				command=self.gotoMaterial)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Edit properties of selected material")

		# ---
		row += 1
		self._gotoRotdefiBtn = Ribbon.LabelButton(group.frame,
				text="Rotdefi",
				image=icons.get("ROT-DEFI"),
				command=self.gotoRotdefi)
		self._gotoRotdefiBtn.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self._gotoRotdefiBtn, "Show rotdefi card attached to region")

		# ---
		row += 1
		self._gotoLatticeBtn = Ribbon.LabelButton(group.frame,
				text="Lattice",
				image=icons.get("LATTICE"),
				command=self.gotoLattice)
		self._gotoLatticeBtn.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self._gotoLatticeBtn,
				"Go to prototype of selected lattice on last clicked viewport")

		# ---
		col,row=1,0
		b = Ribbon.LabelButton(group.frame,
				text="Expand",
				image=icons.get("expand_parenthesis"),
				command=self.expandZones)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Expand parenthesis if any. WARNING can take long")

		# ---
		row += 1
		self._joinBtn = Ribbon.LabelButton(group.frame,
				text="Optimize",
				image=icons.get("optimize"),
				command=self.optimizeZones)
		self._joinBtn.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(self._joinBtn,
				"Optimize zones from useless bodies and invalid bounding box")

		# ---
		row += 1
		self._joinBtn = Ribbon.LabelButton(group.frame,
				text="Join",
				image=icons.get("union"),
				command=self.joinRegions)
		self._joinBtn.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(self._joinBtn,
				"Join regions into a single multi-union region")

		# ---
		col,row=2,1
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("light"),
				command=self.toggleAlpha)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Toggle region transparency [T]")

		menulist = []
		for a in range(0,101,10):
			menulist.append((f"{a:02d}%", None,
				lambda s=self,a=a : s.toggleAlpha(a)))

		b = Ribbon.MenuButton(group.frame, menulist,
				text="Transparency",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT)
				#width=8)
		b.grid(row=row, column=col+1, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Set region transparency [T]")

		# ========== Mesh ==========
		group = Ribbon.LabelGroup(self.ribbon, "Mesh")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		self.addContextGroup("mesh", group)
		group.grid3rows()

		# ---
		col,row=0,0
		b = Ribbon.LabelButton(group.frame,
				text="Bodies",
				image=icons.get("TRC"),
				command=self.mesh2Bodies)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Create bodies out of the mesh")

		# ========== Spline ==========
		group = Ribbon.LabelGroup(self.ribbon, "Spline")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		self.addContextGroup("spline", group)
		group.grid3rows()

		# ---
		col,row=0,0
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Add Node",
				image=icons.get("add"),
				variable=self._mouseAction,
				value=ACTION.ADD_SPLINE_NODE)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Add a new node on spline")

		# ---
		row += 1
		b = Ribbon.LabelRadiobutton(group.frame,
				text="Del Node",
				image=icons.get("del"),
				variable=self._mouseAction,
				value=ACTION.DEL_SPLINE_NODE)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Delete a node from spline")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text="Sort Time",
				image=icons.get("TIME"),
				command=self.splineSortTime)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Sort time of nodes in ascending order")

		# ========== Stop ==========
		group = Ribbon.LabelGroup(self.ribbon, "Stop")
		group.label["background"] = Ribbon._BACKGROUND_GROUP2
		self.addContextGroup("stop", group)

		# ---
		b = Ribbon.LabelButton(group.frame,
				text="Stop",
				image=icons.get("stop32"),
				compound=tk.TOP,
				command=self.stopAction)
		b.pack(fill=tk.BOTH, expand=tk.YES)
		tkExtra.Balloon.set(b, "Stop the current action")

	#----------------------------------------------------------------------
	# GeometryEditor toplevel class
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage(False)

		# check buttons
		self.showBodies   = tk.BooleanVar();	self.showBodies.set(True)
		self.showRegions  = tk.BooleanVar();	self.showRegions.set(True)
		self.showRotdefi  = tk.BooleanVar();	self.showRotdefi.set(True)
		self.showMaterial = tk.BooleanVar();	self.showMaterial.set(False)
		self.showObjects  = tk.BooleanVar();	self.showObjects.set(True)
		self.showVisible  = tk.BooleanVar()
		self.snap2Grid    = tk.BooleanVar();	self.snap2Grid.set(True)
		self.lockEditing  = tk.BooleanVar();	self.lockEditing.set(False)
		self.syncViews    = tk.BooleanVar();	self.syncViews.set(False)
		self.orientSystem = tk.StringVar()

		self.orientSystem.set(Project.GeometryInfo.ORIENTATION)

		self._lastBodyName   = "body"
		self._lastRegionName = "REG"
		self._lastMaterial   = "VACUUM"

		self._mouseAction = tk.IntVar()
		self._mouseAction.set(ACTION.SELECT)
		self._mouseAction.trace("w",self.mouseActionChanged)

		# ---------- create layout -------------
		self.splitter = tkExtra.HSplitter(self.frame,
					tkFlair.config.getInt(tkFlair._PAGE_SECTION,
							self.name+".split", 250),
					absolute=True,
					sash=True)
		self.splitter.pack(side=tk.BOTTOM, expand=tk.YES, fill=tk.BOTH)

		# =============== TabSet Geometry/Layers/Debug=================
		self.tabSet = ttk.Notebook(self.splitter.leftFrame())
		self.tabSet.pack(fill=tk.BOTH, expand=tk.YES)
		self.tabSet.bind("<<NotebookTabChanged>>", self.changeTab)

		# Geometry
		geomframe = tk.Frame(self.tabSet)
		self.tabSet.add(geomframe,
				text="Geometry",
				image=icons.get("REGION"),
				compound=tk.LEFT)

		# layers
		self._layers = GeometryLayers.GeometryLayersFrame(self.tabSet, self.flair)
		self._layers.bind("<<LayersUpdate>>", self.updateLayers)
		self.tabSet.add(self._layers,
				text="Layers",
				image=icons.get("layers"),
				compound=tk.LEFT)

		# Errors
		self._errors = GeometryErrors.ErrorsFrame(self.tabSet, self)
		self.tabSet.add(self._errors,
				text="Errors",
				image=icons.get("debug"),
				compound=tk.LEFT)

		# Fill Geometry
		# --------------------------------------------------------------
		frame = tk.Frame(geomframe)
		frame.pack(fill=tk.X)

		# 3state buttons!
		b = tk.Checkbutton(frame,image=icons.get("TRC"),
				command=self.treeFill,
				variable=self.showBodies,
				indicatoron=False,
				highlightthickness=0,
				selectcolor="White",
				padx=0, pady=0)
		b.bind("<Control-Button-1>",	self.onlyBodies)
		b.bind("<Button-2>",		self.onlyBodies)
		b.bind("<Button-3>",		self.onlyBodies)
		b.pack(side=tk.LEFT, padx=1)
		tkExtra.Balloon.set(b, "Filter Bodies. "
				"<Ctrl-Click> or <Right-Click> to select only Bodies")

		b = tk.Checkbutton(frame,
				image=icons.get("REGION"),
				command=self.treeFill,
				variable=self.showRegions,
				indicatoron=False,
				highlightthickness=0,
				selectcolor="White",
				padx=0, pady=0)
		b.bind("<Control-Button-1>",	self.onlyRegions)
		b.bind("<Button-2>",		self.onlyRegions)
		b.bind("<Button-3>",		self.onlyRegions)
		b.pack(side=tk.LEFT, padx=1)
		tkExtra.Balloon.set(b, "Filter Regions. "
				"<Ctrl-Click>  or <Right-Click>to select only Regions")

		b = tk.Checkbutton(frame,image=icons.get("ROT-DEFI"),
				command=self.treeFill,
				variable=self.showRotdefi ,
				indicatoron=False,
				highlightthickness=0,
				selectcolor="White",
				padx=2, pady=0)
		b.bind("<Control-Button-1>",	self.onlyRotdefi)
		b.bind("<Button-2>",		self.onlyRotdefi)
		b.bind("<Button-3>",		self.onlyRotdefi)
		b.pack(side=tk.LEFT, padx=1)
		tkExtra.Balloon.set(b, "Filter Transformations. "
				"<Ctrl-Click> or <Right-Click> to select only Transformations")

		b = tk.Checkbutton(frame,image=icons.get("MATERIAL"),
				command=self.treeFill,
				variable=self.showMaterial,
				indicatoron=False,
				highlightthickness=0,
				selectcolor="White",
				padx=0, pady=0)
		b.bind("<Control-Button-1>",	self.onlyMaterials)
		b.bind("<Button-2>",		self.onlyMaterials)
		b.bind("<Button-3>",		self.onlyMaterials)
		b.pack(side=tk.LEFT, padx=1)
		tkExtra.Balloon.set(b, "Filter Materials. "
				"<Ctrl-Click> or <Right-Click> to select only Materials")

		b = tk.Checkbutton(frame,image=icons.get("pin"),
				command=self.treeFill,
				variable=self.showObjects,
				indicatoron=False,
				highlightthickness=0,
				selectcolor="White",
				padx=0, pady=0)
		b.bind("<Control-Button-1>",	self.onlyObjects)
		b.bind("<Button-2>",		self.onlyObjects)
		b.bind("<Button-3>",		self.onlyObjects)
		b.pack(side=tk.LEFT, padx=1)
		tkExtra.Balloon.set(b, "Filter Objects. "
				"<Ctrl-Click> or <Right-Click> to select only Objects")

		b = tk.Checkbutton(frame,image=icons.get("select"),
				command=self.treeFill,
				variable=self.showVisible,
				indicatoron=False,
				highlightthickness=0,
				selectcolor="White",
				padx=2, pady=0)
		b.pack(side=tk.LEFT, padx=1)
		tkExtra.Balloon.set(b, "Show Selected & Visible items")

		# --------------------------------------------------------------
		frame = tk.Frame(geomframe)
		frame.pack(fill=tk.X)
		self.filter = tkExtra.LabelEntry(frame,
					label="Filter",
					labelcolor="DarkGray",
					background="White",
					width=7)
		self.filter.pack(side=tk.LEFT, expand=tk.YES, fill=tk.X)
		self.filter.bind("<Return>",   self.treeFill)
		self.filter.bind("<KP_Enter>", self.treeFill)
		self.filter.bind("<<LabelEntryClear>>", self.treeFill)
		tkExtra.Balloon.set(self.filter, "Filter items")

		# --------------------------------------------------------------
		vsplit = tkExtra.VSplitter(geomframe, 0.5)
		vsplit.pack(expand=tk.YES,fill=tk.BOTH)

		# -----------------Object Tree ---------------------------------
		self.tree = tkExtra.Treeview(vsplit.topFrame())
		self.tree["columns"] = ("name",)
		self.tree.column("#0",    width=50*tkFlair.dpiscale, anchor=tk.W, stretch=tk.YES)
		self.tree.column("name",  width=50*tkFlair.dpiscale, anchor=tk.W, stretch=tk.YES)
		self.tree.heading("#0",   text="Type", anchor=tk.W)
		self.tree.heading("name", text="Name", anchor=tk.W)
		self.tree.pack(expand=tk.YES, fill=tk.BOTH, side=tk.LEFT)
		sb = ttk.Scrollbar(vsplit.topFrame(), orient=tk.VERTICAL, command=self.tree.yview)
		sb.pack(fill=tk.Y, side=tk.RIGHT)
		self.tree.config(yscrollcommand=sb.set)
		#self.tree["yscrollcommand"] = sb.set
		self.tree2card = {}			# Dictionary of every tree item to card/object
		self.card2tree = {}			# ... inverse dictionary ...
		self.tree.bindSearch(0)
		self.tree.bind("<<TreeviewSelect>>",	self.treeSelect)
		#self.tree.bind("<<TreeviewOpen>>",	self.treeOpen)
		self.tree.bind("<Delete>",		self.treeDelete)
		self.tree.bind("<Button-1>",		self.treeClick)
		self.tree.bind("<ButtonRelease-1>",	self.treeRelease)
		self.tree.bind("<Double-1>",		self.treeEditName)
		self.tree.bind("<Return>",		self.treeEditName)
		self.tree.bind("<KP_Enter>",		self.treeEditName)
		self.tree.bind("<Button-3>",		self.treePopup)

		self.tree.bind("<Control-Key-space>",	self.treePopup)
		self.tree.bind("<<Copy>>",		self.treeCopy)
		#self.tree.bind("<<Clone>>",		self.treeClone)
		self.tree.bind("<<Rename>>",		self.treeEditName)
		self.tree.bind("<<Paste>>",		self.treePaste)
		self.tree.bind("<<PasteSpecial>>",	self.treePasteSpecial)
		self.tree.bind("<<Cut>>",		self.treeCut)

		# -------------------- Properties/Attributes -------------------
		self.propTab = ttk.Notebook(vsplit.bottomFrame())
		self.propTab.pack(fill=tk.BOTH, expand=tk.YES)

		self.propList = tkExtra.MultiListbox(self.propTab,
					(("Property", 9, None),
					 ("Value",   10, None)),
					 header=False,
					 stretch="last",
					 background="White")
		self.propList.sortAssist = None
		self.propTab.add(self.propList,
				text="Properties",
				image=icons.get("properties"),
				compound=tk.LEFT)

		self.propList.bind("<<ListboxSelect>>",		self.propListSelect)
		self.propList.bindList("<Delete>",		self.propListDelete)
		self.propList.bindList("<BackSpace>",		self.propListDelete)
		self.propList.bindList("<<Rename>>",		self.propListEditEvent)
		self.propList.bindList("<Return>",		self.propListEditEvent)
		self.propList.bindList("<KP_Enter>",		self.propListEditEvent)
		self.propList.bindList("<<Splitter>>",	        self.toggleSplitter)
		self.propList.bindList("<ButtonRelease-1>",	self.propListRelease)
		self.propList.bindList("<Button-3>",		self.propListPopup)
		self.propList.bindList("<Control-Key-space>",	self.propListPopup)
		self.propList.bindList("<<Copy>>",		self.propListCopy)
		self.propList.bindList("<<Cut>>",		self.propListCut)
		self.propList.bindList("<<Paste>>",		self.propListPaste)
		self.propList.bindList("<<MoveUp>>",		self.propListMoveUp)
		self.propList.bindList("<<MoveDown>>",		self.propListMoveDown)
		self.propList.bindList("<<Clone>>",		self.propListClone)
		# FIXME maybe copy also bodies Copy&Paste
		self.propList.bindList("<<CloneSpecial>>",	self.propListClone)

		self.attrList = tkExtra.MultiListbox(self.propTab,
					(("Attribute",  9, None),
					 ("Value",     15, None)),
					 header=False,
					 stretch="last",
					 background="White")
		self.attrList.sortAssist = None
		self.propTab.add(self.attrList,
				text="Attributes",
				image=icons.get("attributes"),
				compound=tk.LEFT)

		#self.attrList.bind("<<ListboxSelect>>",	self.attrListSelect)
		self.attrList.bindList("<Delete>",		self.attrListDelete)
		self.attrList.bindList("<BackSpace>",		self.attrListDelete)
		self.attrList.bindList("<<Rename>>",		self.attrListEdit)
		self.attrList.bindList("<Return>",		self.attrListEdit)
		self.attrList.bindList("<KP_Enter>",		self.attrListEdit)
		self.attrList.bindList("<<Splitter>>",	        self.toggleSplitter)
		self.attrList.bindList("<ButtonRelease-1>",	self.attrListRelease)
		self.attrList.bindList("<Button-3>",		self.attrListPopup)
		self.attrList.bindList("<Control-Key-space>",	self.attrListPopup)
		#self.attrList.bindList("<<Copy>>",		self.attrListCopy)
		#self.attrList.bindList("<<Cut>>",		self.attrListCut)
		#self.attrList.bindList("<<Paste>>",		self.attrListPaste)
		#self.attrList.bindList("<<MoveUp>>",		self.attrListMoveUp)
		#self.attrList.bindList("<<MoveDown>>",		self.attrListMoveDown)

		# --------------------------------------------------------------
		self.viewers = GeometryViewer.Geometry4Frame(
					self.splitter.rightFrame(), self)
		self.viewers.pack(side=tk.RIGHT, expand=tk.YES, fill=tk.BOTH)
		self.viewers.focus_set()
		self.viewers.bind("<<Errors>>", self.updateErrors)

	#----------------------------------------------------------------------
	def createStatus(self):
		super().createStatus()
		tk.Label(self.status, text="x:", foreground="DarkGreen", pady=1).pack(side=tk.LEFT)
		self.xlabel = tk.Label(self.status, width=14, anchor=tk.W, pady=1)
		self.xlabel.pack(side=tk.LEFT, fill=tk.BOTH, padx=1, pady=1)

		tk.Label(self.status, text=" y:", foreground="DarkGreen", pady=1).pack(side=tk.LEFT)
		self.ylabel = tk.Label(self.status, width=14, anchor=tk.W, pady=1)
		self.ylabel.pack(side=tk.LEFT, fill=tk.BOTH, padx=1, pady=1)

		tk.Label(self.status, text=" z:", foreground="DarkGreen", pady=1).pack(side=tk.LEFT)
		self.zlabel = tk.Label(self.status, width=14, anchor=tk.W, pady=1)
		self.zlabel.pack(side=tk.LEFT, fill=tk.BOTH, padx=1, pady=1)

		return self.status

	# ----------------------------------------------------------------------
	def vsplit(self):
		if not self.created(): return None
		return self.viewers.vsplit.split

	# ----------------------------------------------------------------------
	def tsplit(self):
		if not self.created(): return None
		return self.viewers.tsplit.split

	# ----------------------------------------------------------------------
	def bsplit(self):
		if not self.created(): return None
		return self.viewers.bsplit.split

	# ----------------------------------------------------------------------
	def toggleTimeframe(self):
		self.viewers.timeframeToggle()

	# ----------------------------------------------------------------------
	def geometry(self):
		return self.viewers.geometry()

	# ----------------------------------------------------------------------
	def resetEditor(self):
		self.viewers.resetViewports()
		self.viewers.layout("2x2")
		self.reload()
		self.viewers.resetLayers()

	# ----------------------------------------------------------------------
	def orientationSystemChange(self):
		Project.GeometryInfo.ORIENTATION = \
			self.project.geometry["Orientation"] = self.orientSystem.get()
		if self.viewers.viewportOrientation(self.orientSystem.get()):
			self.viewers.resetViewports()
			self.viewers.resetLayers()
			self.viewers.synchronizeViewports()
			self.viewers.redraw()

	# ----------------------------------------------------------------------
	def synchronizeViewports(self):
		sv = int(self.syncViews.get())
		self.project.geometry["Synchronize"] = sv
		self.viewers.synchronizeViewports(self.syncViews.get())
		if self.syncViews.get():
			self.setStatus("Synchronized viewports motion", "DarkBlue")
		else:
			self.setStatus("Free viewports motion", "DarkBlue")

	# ----------------------------------------------------------------------
	def toggleSynchronizeViewports(self):
		self.syncViews.set(not self.syncViews.get())
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def reset(self):
		self._loadStatePending = True	# Force refreshing after showing the window

	# ----------------------------------------------------------------------
	def beforeDock(self):
		self.reset()

	# ----------------------------------------------------------------------
	def reload(self, event=None):
		dump("*-* GeometryEditor.reload")
		self.saveState()
		self.flair.checkProjectFile(self.page)
		self.refresh()

	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		dump("*-* GeometryEditor.refresh",what)
		super().refresh(what)
		self.project.input.refresh()
		self.show()

	# ----------------------------------------------------------------------
	# Show project on viewports
	# ----------------------------------------------------------------------
	def show(self):
		dump("*-* GeometryEditor.show")
		if self.project is None: return
		if self._projectName != self.project.name():
			for card in self.project.input.allcards():
				if card.ignore(): continue
				card[Card.SELECT] = card.get(Card.SELECT,0) & BIT.UNSELECT
			self._projectName = self.project.name()
			new = True
		else:
			new = False

		self.loadProject()

		if new and self.project.plots.empty():
			self.viewers.resetViewports()
			self.viewers.resetLayers()
		if self.tabSet.index(tk.CURRENT) == 1:	# Layers
			self._layers.changeLayer()
		if self._loadStatePending:
			self._loadStatePending = False
			self.loadState()
		self.selectShow()
		self.viewers.check4Frozen()

	# ----------------------------------------------------------------------
	# Load project and refresh information
	# ----------------------------------------------------------------------
	def loadProject(self):
		dump("*-* GeometryEditor.loadProject",bold=True)
		if Global.profiler:
			log.say("\n*** GeometryEditor.loadProject started")
			timer = tool.Timer()
		self._loadingProject = True
		cursor = self.page["cursor"]
		self.page["cursor"] = "watch"
		try:
			self.orientSystem.set(self.project.geometry["Orientation"])
		except KeyError:
			self.orientSystem.set(Project.GeometryInfo.ORIENTATION)
		self.viewers.viewportOrientation(self.orientSystem.get())
		self.syncViews.set(self.project.geometry("Synchronize"))
		self.lockEditing.set(self.project.geometry("Lock"))
		self._fillRuns()
		self.page["cursor"] = cursor
		self._loadingProject = False
		if Global.profiler:
			timer.stop("*** GeometryEditor.loadProject ended")
		self.synchronizeViewports()

		# display any pending messages
		for m in self._late_messages:
			self.flair.log(m)
		del self._late_messages[:]

	# ----------------------------------------------------------------------
	# Fill the runCombo with active runs from the flair project
	# ----------------------------------------------------------------------
	def _fillRuns(self):
		#dump("*-* GeometryEditor.fillRuns")
		activeRun = self.runCombo.get()
		self.runCombo.clear()
		found = False
		for run in self.project.runs():
			if run("parent"): continue
			runname = run.inputName2()
			self.runCombo.insert(tk.END, runname)
			if activeRun == runname:
				found = True

		# Set the run in the combo to actually load the project
		if found:
			self.runCombo.set(activeRun)
		else:
			self.runCombo.set(self.runCombo.get(0))

	# ----------------------------------------------------------------------
	# Called by runCombo when run is changed
	# ----------------------------------------------------------------------
	def runChanged(self):
		#dump("*-* GeometryEditor.runChanged")
		run = self.project.runs.getRunByName(self.runCombo.get())
		tkExtra.Balloon.set(self.runCombo, f"Active Run: {self.runCombo.get()}")
		self.viewers.loadProject(self.project, run)
		self.updateErrors()
		self.treeFill()
		self.redraw()

	# ----------------------------------------------------------------------
	# Save&Load current state and refresh flair window if any
	# ----------------------------------------------------------------------
	def loadState(self):
		if not self.created(): return
		self.viewers.loadState()

	# ----------------------------------------------------------------------
	def saveState(self):
		if not self.created(): return
		self.viewers.saveState()

	# ----------------------------------------------------------------------
	def get(self):
		self.saveState()

	# ----------------------------------------------------------------------
	# Handle the <FocusIn> event. Reload project if necessary
	# ----------------------------------------------------------------------
	def _focusIn(self, event=None):
		if self._inFocus: return
		super()._focusIn(event)
		# FocusIn is generated for all sub-windows
		if (event and event.widget is not self) or self.project is None: return
		self._inFocus = True
		self.flair.checkProjectFile(self)
		self._inFocus = False

	# ----------------------------------------------------------------------
	#def _focusOut(self, event):
	#	"""Handle the <FocusOut> event"""
	#	if self is not event.widget: return
	#	self.saveState()

	# ----------------------------------------------------------------------
	def cut(self, event=None):
		if event is None:
			self.treeCut(event)
		elif event.widget is self.propList or \
		     self.editType() == Card.REGION and \
		     self.propList.curselection():
			self.propListCut(event)
		#elif event.widget is self.attrList:
		#	self.attrListCut(event)
		elif isinstance(event.widget, (tk.Entry, tk.Text)):
			super().cut(event)
		else:
			self.treeCut(event)

	# ----------------------------------------------------------------------
	def copy(self, event=None):
		#dump("*-* GeometryEditor.copy")
		#dump("Widget=",event.widget)
		#dump("Layers=",self._layers)
		if event is None:
			self.treeCopy(event)
		elif event.widget is self.propList:
			self.propListCopy(event)
		elif event.widget is self._layers:
			dump("LAYERS")	 # XXX XXX XXX not working
		#elif event.widget is self.attrList:
		#	self.attrListCopy(event)
		elif isinstance(event.widget, (tk.Entry, tk.Text)):
			super().copy(event)
		else:
			self.treeCopy(event)

	# ----------------------------------------------------------------------
	def paste(self, event=None):
		#dump("*-* GeometryEditor.paste")
		if event is None:
			self.treePaste(event)
		elif event.widget is self.propList:
			self.propListPaste(event)
		#elif event.widget is self.attrList:
		#	self.attrListPaste(event)
		elif isinstance(event.widget, (tk.Entry, tk.Text)):
			super().paste(event)
		else:
			self.treePaste(event)

	# ----------------------------------------------------------------------
	def pasteSpecial(self, event=None):
		return self.treePasteSpecial(event)

	# ----------------------------------------------------------------------
	def clone(self, event=None):
		self.copy(event)
		self.pasteSpecial(event)
		return "break"

	# ----------------------------------------------------------------------
	def inputModified(self, m=True):
		self.flair.inputModified(m)

	# ----------------------------------------------------------------------
	def onlyBodies(self, event=None):
		self.showBodies.set(True)
		self.showRegions.set(False)
		self.showRotdefi.set(False)
		self.showMaterial.set(False)
		self.showObjects.set(False)
		self.treeFill()
		return "break"

	# ----------------------------------------------------------------------
	def onlyRegions(self, event=None):
		self.showBodies.set(False)
		self.showRegions.set(True)
		self.showRotdefi.set(False)
		self.showMaterial.set(False)
		self.showObjects.set(False)
		self.treeFill()
		return "break"

	# ----------------------------------------------------------------------
	def onlyRotdefi(self, event=None):
		self.showBodies.set(False)
		self.showRegions.set(False)
		self.showRotdefi.set(True)
		self.showMaterial.set(False)
		self.showObjects.set(False)
		self.treeFill()
		return "break"

	# ----------------------------------------------------------------------
	def onlyMaterials(self, event=None):
		self.showBodies.set(False)
		self.showRegions.set(False)
		self.showRotdefi.set(False)
		self.showMaterial.set(True)
		self.showObjects.set(False)
		self.treeFill()
		return "break"

	# ----------------------------------------------------------------------
	def onlyObjects(self, event=None):
		self.showBodies.set(False)
		self.showRegions.set(False)
		self.showRotdefi.set(False)
		self.showMaterial.set(False)
		self.showObjects.set(True)
		self.treeFill()
		return "break"

	# ----------------------------------------------------------------------
	# Populate object tree with the requested items
	# FIXME items with error highlight RED
	# ----------------------------------------------------------------------
	#@tool.timeit
	def treeFill(self, event=None):
		Global.dump("*-* GeometryEditor.treeFill",level=2)
		search = self.filter.get().upper()

		open_ = {}		# dictionary of open items
		#---------------------------------------------------------------
		def _scanOpen(parent=""):
			nonlocal open_
			try:
				card = self.tree2card[parent]
				open_[card] = self.tree.item(parent,"open")
			except KeyError:
				pass
			for item in self.tree.get_children(parent):
				_scanOpen(item)
		_scanOpen()

		# empty the list
		self.tree.delete(*self.tree.get_children())
		self.tree2card.clear()
		self.card2tree.clear()

		inp = self.project.input
		visible = self.showVisible.get()

		#---------------------------------------------------------------
		# Search for needle inside card
		#---------------------------------------------------------------
		def searchCard(card, needle):
			if card.tag.find(needle)>=0:               return True
			if card.name().upper().find(needle)>=0:    return True
			if card.comment().upper().find(needle)>=0: return True
			if card.text().upper().find(needle)>=0:    return True
			if card.type()==Card.REGION:
				for zone in card.getProperty(Card.ZONE,[]):
					if zone.upper().find(needle)>=0: return True
				try:
					if card[Card.MATERIAL].name().upper().find(needle)>=0:
						return True
				except AttributeError:
					pass
			for w in card.whats():
				if str(w).upper().find(needle)>=0:
					return True
			return False

		#---------------------------------------------------------------
		def _add(parent, card, name=None):
			nonlocal visible
			parentCard = card[Card.PARENT]
			if parentCard is not None:
				try:
					parent = self.card2tree[parentCard]
				except KeyError:
					parent = _add(parent, parentCard, parentCard.name())
			if visible:
				if card.get(Card.SELECT,0)>0:
					item = self.tree.insert(parent, tk.END,
								text=card.tag,
								image=icons.get(card.tag),
								values=name)
					self.tree2card[item] = card
					self.card2tree[card] = item
				else:
					return None
			else:
				item = self.tree.insert(parent, tk.END,
							text=card.tag,
							image=icons.get(card.tag),
							values=name)
				self.tree2card[item] = card
				self.card2tree[card] = item
			return item

		#---------------------------------------------------------------
		# Bodies
		if self.showBodies.get():
			parent = self.tree.insert("", tk.END,
							text="Bodies",
							image=icons.get("TRC"),
							open=True)
			for card in inp.cardsCache(Card.BODY).values():
				if card is None: continue
				if not search or searchCard(card, search):
					if card.tag=="VOXELS":
						name = "VOXEL"
					elif card.tag=="CELL":
						name = card.what(1)
					else:
						name = card.name()
					_add(parent, card, name)

			for card in inp.cardsCache(Card.SOLID).values():
				if not search or searchCard(card, search):
					_add(parent, card, card.name())

		#---------------------------------------------------------------
		# Regions
		if self.showRegions.get():
			parent = self.tree.insert("", tk.END,
							text="Regions",
							image=icons.get("REGION"),
							open=True)
			for card in inp.cardsCache(Card.REGION).values():
				if card.ignore() or card.name()=="&": continue
				if not search or searchCard(card, search):
					_add(parent, card, card.name())

		#---------------------------------------------------------------
		# Transformations
		if self.showRotdefi.get():
			parent = self.tree.insert("", tk.END,
							text="Transformations",
							image=icons.get("ROT-DEFI"),
							open=True)
			rotdefi = set()
			for card in inp.cardsSorted(Card.ROTATION,-1).values():
				if card.ignore(): continue
				if not search or searchCard(card, search):
					# FIXME for old inputs without names!!!
					item = (card.tag, card.name())
					if item in rotdefi: continue
					rotdefi.add(item)
					_add(parent, card, card.name())

		#---------------------------------------------------------------
		# Materials
		if self.showMaterial.get():
			parent = self.tree.insert("", tk.END,
							text="Materials",
							image=icons.get("MATERIAL"),
							open=True)
			for card in inp.cardsCache(Card.MATERIALS).values():
				if card.ignore(): continue
				if not search or searchCard(card, search):
					_add(parent, card, card.name())

		# Objects
		if self.showObjects.get():
			parent = self.tree.insert("", tk.END,
							text="Objects",
							image=icons.get("object"),
							open=True)
			cards = []
			for tag in CardInfo.FLAIR_TAGS:
				cards.extend(inp[tag])
			cards.extend(inp["BEAM"])
			cards.extend(inp["SPOTBEAM"])
			cards.extend(inp["/primary/select"])
			Card.sort(cards)

			for card in cards:
				if card.ignore(): continue
				if not search or searchCard(card, search):
					_add(parent, card, card.name())

		# scan items to select
		# FIXME same as treeColor
		selected = []
		for item,card in self.tree2card.items():
			flag = card.get(Card.SELECT,0)

			if card.invalid:
				self.tree.tag_add(ERROR_COLOR, item)
			elif flag & BIT.LOCK:
				self.tree.tag_add(LOCK_COLOR, item)
			elif flag & BIT.FREEZE:
				self.tree.tag_add(FREEZE_COLOR, item)
			elif flag & BIT.VISIBLE:
				self.tree.tag_add(VISIBLE_COLOR, item)
			elif flag & BIT.WIREFRAME:
				self.tree.tag_add(WIREFRAME_COLOR, item)

			if flag & BIT.SELECT: selected.append(item)

		for color in (ERROR_COLOR, LOCK_COLOR, FREEZE_COLOR, VISIBLE_COLOR, WIREFRAME_COLOR):
			self.tree.tag_configure(color, foreground=color)

		if selected:
			self.tree.selection_set(selected)
			self.tree.focus(selected[-1])
			self.tree.see(selected[-1])
		else:
			self.selectShow(False)

	# ----------------------------------------------------------------------
	# Object list popup menu
	# Create pop-up menu with default actions
	# ----------------------------------------------------------------------
	def treePopup(self, event):
		self.page.focus_set()
		menu=tk.Menu(self.page, tearoff=0)
		# --- Std menu ---
		menu.add_command(label="Cut", underline=2,
				compound=tk.LEFT, image=icons.get("cut"),
				command=self.cut,
				accelerator="Ctrl-X")
		menu.add_command(label="Copy", underline=0,
				compound=tk.LEFT, image=icons.get("copy"),
				command=self.copy,
				accelerator="Ctrl-C")
		menu.add_command(label="Paste", underline=0,
				compound=tk.LEFT, image=icons.get("paste"),
				command=self.paste,
				accelerator="Ctrl-V")
		menu.add_command(label="Clone", underline=2,
				compound=tk.LEFT, image=icons.get("clone"),
				command=self.clone,
				accelerator="Ctrl-D")

		# --- Add menu ---
		menu.add_separator()
		self.viewers._addMenus(menu)
		menu.add_command(label="Delete", underline=1,
				command=self.delete,
				accelerator="Del",
				compound=tk.LEFT, image=icons.get("x"))

		# --- Visibility and Selection ---
		menu.add_separator()
		self.viewers._visibilityMenu(menu)
		self.viewers._wireframeMenu(menu)
		self.viewers._freezeMenu(menu)

		# -----------------------------------
		menu.add_separator()
		menu.add_command(label="All", underline=0,
				command=self.treeSelectAll,
				compound=tk.LEFT, image=icons.get("select_all"))
		menu.add_command(label="Clear", underline=0,
				command=self.treeSelectClear,
				compound=tk.LEFT, image=icons.get("select_none"))
		menu.add_command(label="Invert", underline=0,
				command=self.treeSelectInvert,
				compound=tk.LEFT, image=icons.get("select_invert"))

		menu.tk_popup(event.x_root, event.y_root)
		return "break"

	# ----------------------------------------------------------------------
	# Clear the selection except when the Control is pressed
	# ----------------------------------------------------------------------
	def treeClick(self, event=None):
		if not event.state & STATE.CONTROL:
			# unselect everything
			self.selectClear(False)

	# ----------------------------------------------------------------------
	# Button-1 released
	# ----------------------------------------------------------------------
	def treeRelease(self, event):
		# Find activated item
		active = self.tree.identify_row(event.y)
		if active:
			try:
				self.viewers.updateEdit(self.tree2card[active])
				self.viewers.expose()
				#if card.type() == Card.BODY:
				#	self.geometry().edit("body",card[Card.ID])
				#	self.viewers.expose()
			except KeyError:
				pass

	# ----------------------------------------------------------------------
	# Called only if the selection on object tree is changed
	# Handle the <<TreeviewSelect>> message on the object tree
	# ----------------------------------------------------------------------
	def treeSelect(self, event=None):
		Global.dump("*-* GeometryEditor.treeSelect",level=2)
		if self.tree.size()==0: return
		# (Un)Select objects
		for card in self.tree2card.values():
			card[Card.SELECT] = card.get(Card.SELECT,0) & BIT.UNSELECT
		for item in self.tree.selection():
			try:
				card = self.tree2card[item]
				card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.SELECT
			except KeyError:
				pass

		# Find active entry and show cursor on it
		active = self.tree.focus()
		card = self.tree2card.get(active)
		self.viewers.updateEdit(card)
		self.selectShow(True)

	# ----------------------------------------------------------------------
	# Set selected items in objects tree
	# ----------------------------------------------------------------------
	def treeSelectSet(self):
		Global.dump("*-* GeometryEditor.treeSelectSet",level=2)
		if self.showVisible.get():
			self.treeFill()
			return
		selected = [i for i,c in self.tree2card.items()
				if c.get(Card.SELECT, 0) & BIT.SELECT]
		if selected:
			self.tree.selection_set(selected)
			self.tree.focus(selected[-1])
			self.tree.see(selected[-1])
		else:
			self.tree.selection_clear()

	# ----------------------------------------------------------------------
	# Double click on object list to change name
	# ----------------------------------------------------------------------
	def treeEditName(self, event=None):
		item = self.tree.focus()
		try:
			card = self.tree2card[item]
			edit = tkExtra.InPlaceEdit(self.tree, item=item, column=0)
			if edit.value is not None:
				self.changeName(card.name(), edit.value, card)
		except KeyError:
			pass

	# ----------------------------------------------------------------------
	# Restore/Set objects tree colors
	# Should be called after sort and after any change to
	#	selection concerning Visibility and Lock
	# FIXME should highlight the errors!
	# ----------------------------------------------------------------------
	def treeColors(self):
		Global.dump("*-* GeometryEditor.treeColors",level=2)
		for item,card in self.tree2card.items():
			flag = card.get(Card.SELECT,0)
			if card.invalid:
				color = ERROR_COLOR
			elif flag & BIT.LOCK:
				color = LOCK_COLOR
			elif flag & BIT.FREEZE:
				color = FREEZE_COLOR
			elif flag & BIT.VISIBLE:
				color = VISIBLE_COLOR
			elif flag & BIT.WIREFRAME:
				color = WIREFRAME_COLOR
			else:
				for tag in self.tree.item(item, "tags"):
					self.tree.tag_remove(tag, item)
				continue
			self.tree.tag_add(color, item)

	# ----------------------------------------------------------------------
	# Delete action on object list
	# delete only highlighted items in object tres
	# ----------------------------------------------------------------------
	def treeDelete(self, event=None):
		undoinfo = None
		bodiesRemoved = []
		regionsRemoved = []
		for item in reversed(self.tree.selection()):
			try:
				card = self.tree2card[item]
				if undoinfo is None:
					undoinfo = self.flair.undolistRefresh()

				card[Card.SELECT] = card.get(Card.SELECT,0) & BIT.UNSELECT
				if card.type() == Card.BODY:
					bodiesRemoved.append(card.name())
				elif card.type() == Card.REGION:
					regionsRemoved.append(card.name())
				undoinfo.append(self.flair.delCardUndo(card.pos(),True))
			except KeyError:
				pass

		# Correct references to bodies
		for body in bodiesRemoved:
			undoinfo.extend(self.flair.removeBodyRefUndo(body))

		# Correct references to regions
		for region in regionsRemoved:
			undoinfo.extend(self.flair.removeRegionRefUndo(region))

		if undoinfo:
			self.flair.addUndolistRefresh(undoinfo, "Delete card(s)")
		self.refreshUndoButton()

	# ----------------------------------------------------------------------
	# Perform the selection of objects
	# action < 0	un-select
	# action = 0	select
	# action > 0	toggle
	# ----------------------------------------------------------------------
	def select(self, card, clear=False, action=0):
		dump("*-* GeometryEditor.select card=",card,"clear=",clear,"action=",action)
		if isinstance(card, tuple):
			obj, idx, node = card
			if obj == "Z":
				# Special treatment select zones
				if clear:
					self.propList.selection_clear(0,tk.END)
				zz = 0
				if not isinstance(idx,list): idx = [idx]
				for z in idx:
					zz = z+3	# XXX WARNING should be in sync
							# with RegionProperties
					if action > 0:	# toggle
						if self.propList.selection_includes(zz):
							self.propList.selection_clear(zz)
						else:
							self.propList.selection_set(zz)
					else:
						self.propList.selection_set(zz)
				if zz: self.propList.see(zz)
				self.propListSelect(clear=clear)
				return None

		if clear:
			for c in self.project.input.allcards():
				c[Card.SELECT] = c.get(Card.SELECT,0) & BIT.UNSELECT

			for c in self.project.input.defaultMaterials.values():
				if c.get(Card.SELECT,0) & BIT.SELFREEZE:
					c[Card.SELECT] = c.get(Card.SELECT,0) & BIT.UNSELECT

			if self.project.ismode(Fluka):
				for c in Fluka.icruMaterials.values():
					if c.get(Card.SELECT,0) & BIT.SELFREEZE:
						c[Card.SELECT] = c.get(Card.SELECT,0) & BIT.UNSELECT

		if isinstance(card, Card):
			if action > 0:		# toggle
				card[Card.SELECT] = card.get(Card.SELECT,0) ^ BIT.SELECT
			elif action == 0:	# select
				card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.SELECT
			else:			# un-select
				card[Card.SELECT] = card.get(Card.SELECT,0) & BIT.UNSELECT

		elif isinstance(card, list):
			for c in card:
				if action > 0:
					if c.get(Card.SELECT,0) & BIT.SELECT:
						action = -1	# un-select
					else:
						action =  0	# select

				if action == 0:		# select
					c[Card.SELECT] = c.get(Card.SELECT,0) | BIT.SELECT
				else:			# un-select
					c[Card.SELECT] = c.get(Card.SELECT,0) & BIT.UNSELECT
			card = c # remember last card

		else:	# must be tuple (obj, idx, node)
			card = None
			if isinstance(idx, list):
				for i in idx:
					card = self.findCard(obj,i)
					if card is None: continue
					if action > 0:	# toggle
						if card.get(Card.SELECT,0) & BIT.SELECT:
							action = -1	# un-select
						else:
							action =  0	# select

					if action == 0:	# select
						card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.SELECT
						card[Card.NODE]   = node
					else:		# un-select
						card[Card.SELECT] = card.get(Card.SELECT,0) & BIT.UNSELECT
						card[Card.NODE]   = 0
			else:
				card = self.findCard(obj,idx)
				if card is not None:
					if action > 0:		# toggle
						card[Card.SELECT] = card.get(Card.SELECT,0) ^ BIT.SELECT
						card[Card.NODE]   = node
					elif action == 0:	# select
						card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.SELECT
						card[Card.NODE]   = node
					else:
						card[Card.SELECT] = card.get(Card.SELECT,0) & BIT.UNSELECT
						card[Card.NODE]   = 0

		# Populate listbox and trigger selecting
		self.treeSelectSet()

		# Update last edit
		self.viewers.updateEdit(card)

	# ----------------------------------------------------------------------
	# Show on all viewports the selected and visible objects
	# @return number of selected objects
	# ----------------------------------------------------------------------
	def selectShow(self, redraw=True, updateprop=True):
		Global.dump("*-* GeometryEditor.selectShow",level=2)
		self.viewers.body(  "*", "show",   0)
		self.viewers.region("*", "show",   0)
		self.viewers.object("*", "show",   0)
		self.viewers.object("*", "select", 0)	# select 0 clears node also

		selcards = []

		# FIXME check if a selection flag has changed from bodies/regions/zones
		# then only redraw!
		#needredraw = False

		# Scan on materials cards if any
		#if self.showMaterial.get():
		for card in self.project.input.defaultMaterials.values():
			if card.get(Card.SELECT,0) & BIT.SELFREEZE:
				selcards.append(card)
		for card in Fluka.icruMaterials.values():
			if card.get(Card.SELECT,0) & BIT.SELFREEZE:
				selcards.append(card)

		# Scan over all cards (input+voxel)
		for card in self.project.input.allcards():
			if card.ignore(): continue
			flag = card.get(Card.SELECT,0)
			#needredraw = True
			if flag == 0: continue
			if card.type() == Card.BODY:
				try:
					self.viewers.body(card[Card.ID], "show", flag)
				except IndexError:
					continue
				if flag & BIT.SELECT:
					self.viewers.body(card[Card.ID], "color",
						GeometryViewer._SELECT_COLOR)
					#self.viewers.body(card[Card.ID], "linewidth", 2)
				elif flag & BIT.FREEZE:
					self.viewers.body(card[Card.ID], "color",
						GeometryViewer._FREEZE_COLOR)
					#self.viewers.body(card[Card.ID], "linewidth", 0)
				elif flag & BIT.LOCK:
					self.viewers.body(card[Card.ID], "color",
						GeometryViewer._LOCK_COLOR)
					#self.viewers.body(card[Card.ID], "linewidth", 0)
				elif flag & BIT.VISIBLE:
					self.viewers.body(card[Card.ID], "color",
						GeometryViewer._VISIBLE_COLOR)
					#self.viewers.body(card[Card.ID], "linewidth", 0)

			elif card.type()==Card.REGION:
				#needredraw = True
				if card.name() == "&": continue
				try:
					self.viewers.region(card[Card.ID], "show", flag)
				except (IndexError,TypeError):
					pass

			elif card.type()==Card.OBJECT or card.tag in _CARD_OBJECTS:
				try:
					self.viewers.object(card[Card.ID], "show", 1)
					if flag & BIT.UNVISIBLE:
						self.viewers.object(card[Card.ID], "select", True)
						self.viewers.object(card[Card.ID], "selectnode",
								card[Card.NODE])
				except (IndexError,TypeError):
					pass

			# Append to the list of selected cards
			if flag & BIT.SELFREEZE: selcards.append(card)
			#if flag & BIT.SELECT: selcards.append(card)

		# update viewports if required
		if redraw: self.viewers.drawOnIdle()

		# Do not update properties
		if not updateprop: return

		# Update property list only if edit button is not pressed
		if not selcards:
			self.showNothing()
		else:
			self.showProperties(selcards)
		self.viewers.updateSelection(selcards)

	# ----------------------------------------------------------------------
	# @return true if any item is selected
	# ----------------------------------------------------------------------
	def isAnythingSelected(self, cardtype=None):
		Global.dump("*-* GeometryEditor.isAnythingSelected",level=2)
		for card in self.project.input.allcards():
			if card.ignore(): continue
			if cardtype is not None and card.type() != cardtype:
				continue
			if card.get(Card.SELECT,0) & BIT.SELECT:
				return True
		return False

	# ----------------------------------------------------------------------
	# Selection clear (ignore editing object)
	# ----------------------------------------------------------------------
	def selectClear(self, show=True, cardtype=None):
		Global.dump("*-* GeometryEditor.selectClear",level=2)
		for card in self.project.input.allcards():
			if card.ignore(): continue
			if cardtype is not None and card.type() != cardtype: continue
			card[Card.SELECT] = card.get(Card.SELECT,0) & BIT.UNSELECT
		if show:
			self.treeSelectSet()

	# ----------------------------------------------------------------------
	# @return true if object is selected
	# ----------------------------------------------------------------------
	def isSelected(self, card):
		if isinstance(card,tuple):
			card = self.findCard(*card)
		return card.get(Card.SELECT,0) & BIT.SELECT if card else 0

	# ----------------------------------------------------------------------
	# Return the list of selected objects with mask
	# ----------------------------------------------------------------------
	def getSelection(self, mask=BIT.SELECT, objtype=None):
		Global.dump("*-* GeometryEditor.getSelection",level=2)
		objects = []
		for card in self.project.input.allcards():
			if card.ignore(): continue
			if card.get(Card.SELECT,0) & mask:
				if objtype is None or card.type() == objtype:
					objects.append(card)
		return objects

	# ----------------------------------------------------------------------
	def selectAll(self):
		Global.dump("*-* GeometryEditor.selectAll",level=2)
		# with or without control
		if self.editObject is not None and \
		   self.editObject.type() == Card.REGION:
			# Select all bodies from selected regions
			if not self.selectAllBodies():
				self.treeSelectAll()
		else:
			self.treeSelectAll()

	#-----------------------------------------------------------------------
	def treeSelectAll(self):
		Global.dump("*-* GeometryEditor.treeSelectAll",level=2)
		item = self.tree.focus()
		if item:
			parent = self.tree.parent(item)
			if parent == "":
				self.tree.selectAll()
			else:
				self.tree.selection_add(self.tree.get_children(parent))
		else:
			self.tree.selectAll()

	#-----------------------------------------------------------------------
	def treeSelectClear(self):
		Global.dump("*-* GeometryEditor.treeSelectClear",level=2)
		self.tree.selectClear()

	#-----------------------------------------------------------------------
	def treeSelectInvert(self):
		Global.dump("*-* GeometryEditor.treeSelectInvert",level=2)
		self.tree.selectInvert()

	# ----------------------------------------------------------------------
	# Select all bodies from all selected regions
	# ----------------------------------------------------------------------
	def selectAllBodies(self):
		Global.dump("*-* GeometryEditor.selectAllBodies",level=2)
		found = False
		for region in self.project.input.cardsCache(Card.REGION,-1).values():
			if region.get(Card.SELECT,0) & BIT.SELECT == 0: continue
			found = True
			for name in csg.tokenize(region.expr()):
				if name in "+-|:()": continue
				body = self.findCard("B", name)
				if body is not None:
					body[Card.SELECT] = body.get(Card.SELECT,0) | BIT.SELECT
		self.treeSelectSet()
		return found

	# ----------------------------------------------------------------------
	# Return the list of visible objects
	# ----------------------------------------------------------------------
	def getVisible(self):
		return self.getSelection(BIT.VISIBLE)

	# ----------------------------------------------------------------------
	# set bit to all selected objects
	# ----------------------------------------------------------------------
	def setBit(self, bit, types, msg, tags=()):
		for card in self.project.input.cardlist:
			if card.ignore(): continue
			if card.type() not in types and card.tag not in tags: continue
			value = card.get(Card.SELECT,0)
			if value & BIT.SELECT:
				card[Card.SELECT] = value | bit
		self.setStatus(msg)

	# ----------------------------------------------------------------------
	# unset the selection bit to all selected objects
	# ----------------------------------------------------------------------
	def unsetBit(self, bit, types, msg, tags=()):
		unbit = ~bit
		mask  = bit | BIT.SELECT
		for card in self.project.input.cardlist:
			if card.ignore(): continue
			if card.type() not in types and card.tag not in tags: continue
			value = card.get(Card.SELECT,0)
			if value & mask == mask:
				card[Card.SELECT] = value & unbit
		self.setStatus("Un-"+msg)

	# ----------------------------------------------------------------------
	# Change bit on selected items
	# ----------------------------------------------------------------------
	def toggleBit(self, bit, types, msg, tags=()):
		for card in self.project.input.allcards():
			if card.ignore(): continue
			if card.type() not in types and card.tag not in tags: continue
			value = card.get(Card.SELECT,0)
			if value & BIT.SELECT:
				if value & bit:
					self.unsetBit(bit, types, msg, tags)
				else:
					self.setBit(bit, types, msg, tags)
				return

	# ----------------------------------------------------------------------
	# unset the selection bit to all objects
	# ----------------------------------------------------------------------
	def setBitAll(self, bit, types, msg, tags=()):
		for card in self.project.input.cardlist:
			if card.ignore(): continue
			if card.type() not in types and card.tag not in tags: continue
			card[Card.SELECT] = card.get(Card.SELECT,0) | bit
		self.setStatus(msg)

	# ----------------------------------------------------------------------
	# unset the selection bit to all objects
	# ----------------------------------------------------------------------
	def unsetBitAll(self, bit, types, msg, tags=()):
		unbit = ~bit
		for card in self.project.input.cardlist:
			if card.ignore(): continue
			if card.type() not in types and card.tag not in tags: continue
			value = card.get(Card.SELECT,0)
			if value & bit == bit:
				card[Card.SELECT] = value & unbit
		self.setStatus(f"Un-{msg}")

	# ----------------------------------------------------------------------
	# Change bit on all items
	# ----------------------------------------------------------------------
	def toggleBitAll(self, bit, types, msg, tags=()):
		for card in self.project.input.allcards():
			if card.ignore(): continue
			if card.type() not in types and card.tag not in tags: continue
			value = card.get(Card.SELECT,0)
			if value & bit:
				self.unsetBitAll(bit, types, msg, tags)
			else:
				self.setBitAll(bit, types, msg, tags)
			return

	# ----------------------------------------------------------------------
	def lockEditingCommand(self):
		self.project.geometry["Lock"] = int(self.lockEditing.get())
		if self.lockEditing.get():
			self.setStatus("Graphical editing is LOCKED")
		else:
			self.setStatus("Graphical editing is UN-LOCKED")

	# ----------------------------------------------------------------------
	def toggleLock(self):
		self.lockEditing.set(not self.lockEditing.get())
		self.lockEditingCommand()

	# ----------------------------------------------------------------------
	def _freezeEnd(self):
		self.viewers.check4Frozen()
		self.treeColors()
		self.updateProperties()
		self.selectShow()

	def toggleFreeze(self):
		self.toggleBit(BIT.FREEZE, (Card.BODY, Card.REGION), "Freeze selected objects")
		self._freezeEnd()

	def freeze(self):
		self.setBit(BIT.FREEZE, (Card.BODY, Card.REGION), "Freeze selected objects")
		self._freezeEnd()

	def unfreeze(self):
		self.unsetBit(BIT.FREEZE, (Card.BODY, Card.REGION), "Freeze selected objects")
		self._freezeEnd()

	# ----------------------------------------------------------------------
	def unfreezeAll(self):
		self.unsetBitAll(BIT.FREEZE, (Card.BODY, Card.REGION), "Freeze all objects")
		self._freezeEnd()

	# ----------------------------------------------------------------------
	def _3dEnd(self):
		self.treeColors()
		self.selectShow(True, False)
		self.updateProperties()

	def toggle3D(self):
		self.toggleBit(BIT.WIREFRAME, (Card.BODY,), "Set 3D wireframe")
		self._3dEnd()

	def set3D(self):
		self.setBit(BIT.WIREFRAME, (Card.BODY,), "Set 3D wireframe")
		self._3dEnd()

	def unset3D(self):
		self.unsetBit(BIT.WIREFRAME, (Card.BODY,), "Set 3D wireframe")
		self._3dEnd()

	def unset3DAll(self):
		self.unsetBitAll(BIT.WIREFRAME, (Card.BODY,), "Set 3D wireframe")
		self._3dEnd()

	# ----------------------------------------------------------------------
	def _visibilityEnd(self):
		self.treeColors()
		self.updateProperties()
		self.selectShow()

	def toggleVisibility(self):
		self.toggleBit(BIT.VISIBLE, (Card.BODY,Card.OBJECT),
				"Set visibility", _CARD_OBJECTS)
		self._visibilityEnd()

	def setVisibility(self):
		self.setBit(BIT.VISIBLE, (Card.BODY,Card.OBJECT),
				"Set visibility", _CARD_OBJECTS)
		self._visibilityEnd()

	def unsetVisibility(self):
		self.unsetBit(BIT.VISIBLE, (Card.BODY,Card.OBJECT),
				"Set visibility", _CARD_OBJECTS)
		self._visibilityEnd()

	def toggleVisibilityAll(self):
		self.toggleBitAll(BIT.VISIBLE, (Card.BODY,Card.OBJECT),
				"Set visibility", _CARD_OBJECTS)
		self._visibilityEnd()

	def setAllVisibility(self):
		self.setBitAll(BIT.VISIBLE, (Card.BODY,Card.OBJECT),
				"Set visibility", _CARD_OBJECTS)
		self._visibilityEnd()

	def unsetAllVisibility(self):
		self.unsetBitAll(BIT.VISIBLE, (Card.BODY,Card.OBJECT),
				"Set visibility", _CARD_OBJECTS)
		self._visibilityEnd()

	# ----------------------------------------------------------------------
	# Toggle the alpha channel for all selected regions
	# ----------------------------------------------------------------------
	def toggleAlpha(self, a=None):
		undoinfo = []
		for card in self.project.input.cardsCache(Card.REGION,-1).values():
			value = card.get(Card.SELECT,0)
			if value & BIT.SELECT:
				if a is None:
					alpha = 100 - card.get(Card.ALPHA,0)
				else:
					alpha = a
				undoinfo.append(self.flair.setCardUndo(card, Card.ALPHA, alpha, 0))
		if undoinfo:
			self.flair.addUndo(undoinfo,undoinfo[-1][0])
			self.redraw()
			self.updateProperties()

	# ----------------------------------------------------------------------
	# Return selection list of moveable objects (bodies and objects)
	# Movable are: bodies and objects
	# ----------------------------------------------------------------------
	def getMoveableSelection(self, mask=BIT.SELECT):
		objects = []
		bodies  = {}
		bodiesSelected = False	# Avoid moving the regions if bodies&regions are selected

		for card in self.project.input.cardlist:
			if card.ignore(): continue
			if card.get(Card.SELECT,0) & mask == 0: continue

			if card.type()==Card.BODY:
				bodies[card.name()] = card
				bodiesSelected = True

			elif card.type()==Card.OBJECT or \
			     card.tag in ("ROT-DEFI", "MATRIX", "ROTM"):
				objects.append(card)

			elif card.type()==Card.REGION:
				# Return bodies of region if only
				# regions are selected! No bodies
				if bodiesSelected: continue
				for name in csg.tokenize(card.expr()):
					if name in "+-|:()": continue
					if name in bodies: continue
					body = self.findCard("B", name)
					bodies[name] = body

		return (objects, list(bodies.values()))

	# ----------------------------------------------------------------------
	# Find Input Card from obj, idx
	# ----------------------------------------------------------------------
	def findCard(self, obj, idx):
		if obj == "B":
			if isinstance(idx,int):
				# Search by index
				for card in self.project.input.cardsCache(Card.BODY).values():
					if idx == card[Card.ID]:
						return card
			else:
				# Search by name
				for card in self.project.input.cardsCache(Card.BODY).values():
					if card.tag=="VOXELS":
						if idx=="VOXEL": return card
					elif idx==card.name():
						return card

		elif obj == "R":
			if isinstance(idx,int):
				if idx >= 0:
					# Scan in normal input
					for card in self.project.input.cardsCache(Card.REGION).values():
						if idx == card[Card.ID]:
							return card

				elif self.project.ismode(Fluka):
					# Scan in voxels by name!!!
					idx = -idx
					name = FlukaVoxel.regionName(idx)
					for voxel in self.project.input["VOXELS"]:
						if voxel.ignore() or voxel[FlukaCard.VOXEL] is None:
							continue
						for card in voxel[FlukaCard.VOXEL]["REGION"]:
							if card.name() == name:
							#if idx == card[Card.ID]:
								return card
			else:
				return self.project.input.cardsCache(Card.REGION).get(idx)

		elif obj=="O":
			rots = CardInfo.rotations[self.project.input.MODE]
			if isinstance(rots,str): rots = [rots]
			tags = tuple(CardInfo.FLAIR_TAGS + ["BEAM", "SPOTBEAM"] + rots)
			if isinstance(idx,int):
				for card in self.project.input.cardsCache(tags).values():
					if card.ignore(): continue
					if idx == card[Card.ID]:
						return card
			else:
				for card in self.project.input.cardsCache(tags).values():
					if card.ignore(): continue
					if idx == card.name():
						return card

		elif obj=="T":
			if isinstance(idx,int):
				for card in self.project.input.cardsCache(Card.ROTATION).values():
					if idx == card[Card.ID]:
						return card
			else:
				for card in self.project.input.cardsCache(Card.ROTATION).values():
					if idx == card.name():
						return card
		return None

	# ----------------------------------------------------------------------
	# Zoom on selected objects
	# ----------------------------------------------------------------------
	def zoomOnSelected(self, event=None):
		objects = self.getSelection()
		if objects is None: return

		for frame in self.viewers.frames:
			frame.zoomOnSelected(objects)

	# ----------------------------------------------------------------------
	def showNothing(self):
		self.propList.delete(0,tk.END)
		self.attrList.delete(0,tk.END)
		self.showContextGroup()
		self.editObject = None
		self.propList.setPopupMenu()

	# ----------------------------------------------------------------------
	# Show properties of the selected object
	# ----------------------------------------------------------------------
	def showProperties(self, cards):
		#dump("*-* GeometryEditor.showProperties",cards)
		select = [0]*10		# counters of selected cards
		freeze = [0]*10
		tags  = set()
		for c in cards:
			flag = c.get(Card.SELECT,0)
			if flag & BIT.SELECT:
				select[c.type()] += 1
			elif flag & BIT.FREEZE:
				freeze[c.type()] += 1
			tags.add(c.tag)
		tags  = list(tags)

		nbodies  = select[Card.BODY]
		nregions = select[Card.REGION]
		nobjects = select[Card.OBJECT]
		ngeneric = select[Card.GENERIC]
		#dump(">>> B:%d  R:%d  O:%d  G:%d"%(nbodies,nregions,nobjects,ngeneric))

		# XXX Check before if editObject contain the same cards
		if nregions==1 or \
		   (nregions==0 and freeze[Card.REGION]==1 and nobjects==0 and ngeneric==0):
			self.showContextGroup("region")
			# Find first region
			if nregions==1: # Which is selected
				for c in cards:
					if c.type() == Card.REGION and \
					   c.get(Card.SELECT,0) & BIT.SELECT:
						region = c
						break
			else: # or locked
				for c in cards:
					if c.type() == Card.REGION and \
					   c.get(Card.SELECT,0) & BIT.FREEZE:
						region = c
						break
			if self.editObject is not None and \
			   len(self.editObject.cards) == 1 and \
			   self.editObject.cards[0] is region and \
			   not self.editObject.isModified() and \
			   not self.flair.inundo():
				# Do nothing, most probably zone selection
				return
			self.editObject = RegionProperties(self, [region])

			state = tk.NORMAL if self._geometry().region(region[Card.ID],"hasmatrix") \
						else tk.DISABLED
			self._gotoRotdefiBtn.config(state=state)
			self._gotoLatticeBtn.config(state=state)
			self._joinBtn.config(state= tk.DISABLED)

		elif nbodies!=0 and nobjects==0 and nregions==0 and ngeneric==0:
			selcards = [x for x in cards if x[Card.SELECT]&BIT.SELECT != 0]
			if selcards:
				self.editObject = BodyProperties(self, selcards)
				self.showContextGroup("body")
			else:
				self.showNothing()
				return

		elif nbodies==0 and nobjects==0 and nregions!=0 and ngeneric==0:
			self.editObject = RegionProperties(self, cards)
			self.showContextGroup("region")
			self._gotoRotdefiBtn.config(state=tk.DISABLED)
			self._gotoLatticeBtn.config(state=tk.DISABLED)
			self._joinBtn.config(state= tk.NORMAL)

		elif nbodies==0 and nobjects!=0 and nregions==0 and ngeneric==0:
			self.editObject = ObjectProperties(self, cards)
			if self.editObject.objtype == "mesh":
				self.showContextGroup("mesh")
			elif self.editObject.objtype == "spline":
				self.showContextGroup("spline")
			else:
				self.showContextGroup("object")

		#elif nbodies==0 and nobjects==0 and nregions==0 and ngeneric!=0 and len(tags)==1:
		elif ngeneric!=0:
			if tags[0] in ("MATERIAL","M"):
				self.editObject = MaterialProperties(self, cards)
				self.showContextGroup("material")
			elif tags[0] == "ROT-DEFI":
				self.editObject = RotdefiProperties(self, cards)
				self.showContextGroup("rotdefi")
			elif tags[0] in ("MATRIX", "ROTM"):
				self.editObject = RotdefiProperties(self, cards)
				self.showContextGroup("rotdefi")
			elif tags[0] in ("TR","TR*"):
				self.editObject = TRProperties(self, cards)
				self.showContextGroup("rotdefi")
			elif tags[0] in ("BEAM","SPOTBEAM","/primary/select"):
				self.editObject = BeamProperties(self, cards)
				self.showContextGroup("beam")
			elif tags[0] == "VOXELS":
				self.editObject = VoxelProperties(self, cards)
			else:
				self.showNothing()
				return

		else:
			self.showNothing()
			return

		self.updateProperties()
		self.propListSelect()

		if self.editObject is not None and \
		   self.editObject.type()==Card.REGION:
			self.propList.setPopupMenu(
				[('Cut',         2, self.propListCut,      icons.get("cut"),   "Ctrl-X"),
				 ('Copy',        0, self.propListCopy,     icons.get("copy"),  "Ctrl-C"),
				 ('Paste',       0, self.propListPaste,    icons.get("paste"), "Ctrl-V"),
				 ('Clone',       1, self.propListClone,    icons.get("clone"), "Ctrl-D"),
				 None,
				 ('Delete',      0, self.propListDelete,   icons.get("x"),     "Del"),
				 ('Move Up',     5, self.propListMoveUp,   icons.get("up"),    "Ctrl-Up"),
				 ('Move Down',   5, self.propListMoveDown, icons.get("down"),  "Ctrl-Down")
				 ])
		else:
			self.propList.setPopupMenu()

	# ----------------------------------------------------------------------
	# Update the properties and attribute lists of the current object
	# ----------------------------------------------------------------------
	def updateProperties(self):
		dump("*-* GeometryEditor.updateProperties")
		if self.editObject is None: return
		if not self._loadingProject:
			self.flair.activateIfNeeded(self.editObject.cards[0])
		if self.editObject: self.editObject.fill()

	# ----------------------------------------------------------------------
	# Return type of editing objects if any else None
	# ----------------------------------------------------------------------
	def editType(self):
		if self.editObject is None or self.editObject.many:
			return None
		return self.editObject.type()

	# ----------------------------------------------------------------------
	# Property List popup menu
	# Create popup menu with default actions
	# ----------------------------------------------------------------------
	def propListPopup(self, event):
		if self.editObject is None: return
		self.page.focus_set()

		menu=tk.Menu(self.page, tearoff=0)

		menu.add_command(label="Cut", underline=2,
				command=self.propListCut,
				compound=tk.LEFT, image=icons.get("cut"),
				accelerator="Ctrl-X")
		menu.add_command(label="Copy", underline=0,
				command=self.propListCopy,
				compound=tk.LEFT, image=icons.get("copy"),
				accelerator="Ctrl-C")
		menu.add_command(label="Paste", underline=0,
				command=self.propListPaste,
				compound=tk.LEFT, image=icons.get("paste"),
				accelerator="Ctrl-V")
		menu.add_command(label="Clone", underline=0,
				state=tk.DISABLED,
				compound=tk.LEFT, image=icons.get("clone"))
		if self.editObject.type() == Card.REGION:
			menu.add_separator()
			menu.add_command(label="Edit Material", underline=0,
					command=self.gotoMaterial,
					compound=tk.LEFT, image=icons.get("MATERIAL"),
					accelerator="Ctrl-M")
		menu.add_separator()
		menu.add_command(label="Delete", underline=0,
				command=self.propListDelete,
				compound=tk.LEFT, image=icons.get("x"))
		menu.add_separator()

		menu.add_command(label="All", underline=0,
				command=self.propList.selectAll,
				compound=tk.LEFT, image=icons.get("select_all"))
		menu.add_command(label="Clear", underline=0,
				command=self.propList.selectClear,
				compound=tk.LEFT, image=icons.get("select_none"))
		menu.add_command(label="Invert", underline=0,
				command=self.propList.selectInvert,
				compound=tk.LEFT, image=icons.get("select_invert"))

		menu.tk_popup(event.x_root, event.y_root)
		return "break"

	# ----------------------------------------------------------------------
	# Copy properties
	# ----------------------------------------------------------------------
	def propListCopy(self, event=None):
		if self.editObject is None: return
		lst = self.propList.listbox(1)
		if self.editObject.type() == Card.REGION:
			txt = "\n|".join([str(lst.get(i)) for i in self.propList.curselection()])
		else:
			txt = " ".join([str(lst.get(i)) for i in self.propList.curselection()])
		self.page.clipboard_clear()
		self.page.clipboard_append(txt, type='STRING')
		return "break"

	# ----------------------------------------------------------------------
	# Cut properties
	# ----------------------------------------------------------------------
	def propListCut(self, event=None):
		if self.editObject is None: return
		self.propListCopy()
		self.propListDelete()
		return "break"

	# ----------------------------------------------------------------------
	# Paste properties
	# ----------------------------------------------------------------------
	def propListPaste(self, event=None):
		if self.editObject is None: return
		if self.editObject.type() == Card.REGION:
			self.addZone(self.flair.clipboard())
		return "break"

	# ----------------------------------------------------------------------
	# Return zone index from listbox item item
	# ----------------------------------------------------------------------
	def zoneIndex(self, item):
		name = self.propList.listbox(0).get(int(item))
		if not name.startswith("zone"): return None
		return int(name[4:])-1

	# ----------------------------------------------------------------------
	# Add or update zone information from mouse selection ACTION.ZONE
	# ----------------------------------------------------------------------
	def addZone(self, zone):
		if self.editObject.type() != Card.REGION: return
		if self.editObject.many: return

		# Find if any zone is selected
		sel = self.propList.curselection()
		if len(sel)==1:
			z = self.zoneIndex(sel[0])
			if z is not None:
				zones = self.editObject.cards[0][Card.ZONE]
				zones[z] = zone
				self.flair.addUndo(
					self.flair.setZonesUndo(self.editObject.cards[0], zones))
			else:
				self.flair.addUndo(
					self.flair.addZoneUndo(self.editObject.cards[0], zone))
			self.flair.refresh("redraw", self)
		elif len(sel)>1:
			self.flair.notify("Too many selected zones",
				"Too many zones are selected. Please select "
				"only one zone to update, or none to add a new one",
				tkFlair.NOTIFY_WARNING)
			return
		else:
			self.flair.addUndo(
				self.flair.addZoneUndo(self.editObject.cards[0], zone))
			self.flair.refresh("redraw", self)

		# FIXME Show properties after re-parsing the zones...
		self.updateProperties()
		self.forceRefresh()
		#self.showProperties(self.editObject.cards)
		#self.propListSelect()

	# ----------------------------------------------------------------------
	# Return # of zones selected:
	# ----------------------------------------------------------------------
	def propListZoneSelected(self):
		# Find if any zone is selected
		count = 0
		for sel in self.propList.curselection():
			if self.propList.get(sel)[0].startswith("zone"):
				count += 1
		return count

	# ----------------------------------------------------------------------
	# Return selected zones
	# ----------------------------------------------------------------------
	def propListSelectedZones(self):
		if self.editObject.type() == Card.REGION:
			zones = []
			for sel in self.propList.curselection():
				z = self.zoneIndex(sel)
				if z is not None:
					zones.append(z)
			return zones
		return None

	# ----------------------------------------------------------------------
	def propListClear(self):
		self.propList.selection_clear(0,tk.END)
		self.propListSelect()

	# ----------------------------------------------------------------------
	# Selection changed in propList
	# For Regions:
	#	- Select all bodies appearing in selected zones
	# ----------------------------------------------------------------------
	def propListSelect(self, event=None, clear=True):
		if self.editObject is None: return

		if self.editObject.type() == Card.REGION:
			# Select all bodies from zone
			bodies = set()
			selzones = []
			for sel in self.propList.curselection():
				name,zone = self.propList.get(sel)
				if not name.startswith("zone"): continue
				# find bodies
				try:
					for body in zone.translate(OPER_TRANS).split():
						bodies.add(body)
				except TypeError:	# Maybe some Unicode will not work
					pass

				# Select zone in viewers
				selzones.append(zone)

			if bodies:
				# Clear any body selection
				if clear:
					for card in self.project.input.cardlist:
						if card.ignore(): continue
						if card.type() == Card.BODY:
							card[Card.SELECT] = card.get(Card.SELECT,0) & BIT.UNSELECT

				# Select bodies, but don't use the self.select() to avoid
				# the filling of the propList and de-selection of zone
				for body in bodies:
					card = self.findCard("B",body)
					if card is not None:
						card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.SELECT
				self.treeFill()
				if selzones:
					self.viewers.stopThreads()
					self.viewers.zone("clear")
					self.viewers.zone("select", selzones)
					self.viewers.zone("show",
						self.viewers.region(
							self.editObject.cards[0][Card.ID],"show"))
			else:
				self.viewers.zone("clear")

			self.viewers.drawOnIdle()
			self.viewers.updateEdit()
			del bodies
		else:
			# FIXME
			pass
		self.viewers.updateSelection()

	# ----------------------------------------------------------------------
	# Button Release on the properties listbox
	# ----------------------------------------------------------------------
	def propListRelease(self, event):
		if self.editObject is None: return
		act = event.widget.nearest(event.y)
		self.propList.activate(act)
		# Find list box
		if self.propList.listboxIndex(event.widget) == 0: return
		self.propListEdit()
		return "break"

	# ----------------------------------------------------------------------
	def editName(self):
		if self.editObject is None: return
		self.selectGeometryPropertiesTab()
		self.propList.selection_clear(0,tk.END)
		self.propList.selection_set(0)
		self.propList.activate(0)
		self.propListEdit(False)

	# ----------------------------------------------------------------------
	def editMaterial(self):
		if self.editObject is None or self.editObject.type() != Card.REGION: return
		self.selectGeometryPropertiesTab()
		self.propList.selection_clear(0,tk.END)
		self.propList.selection_set(2)
		self.propList.activate(2)
		self.propListEdit(False)

	# ----------------------------------------------------------------------
	def editProperty(self):
		if self.editObject is None: return
		self.selectGeometryTab()
		if self.propTab.index(tk.CURRENT) == 0:
			self.propList.focus_set()
			self.propList.selection_clear(0,tk.END)
			self.propList.selection_set(tk.ACTIVE)
			self.propListEdit(False)
		else:
			self.attrList.focus_set()
			self.attrList.selection_clear(0,tk.END)
			self.attrList.selection_set(tk.ACTIVE)
			self.attrListEdit(False)

	# ----------------------------------------------------------------------
	# Show/Goto to material card of region
	# ----------------------------------------------------------------------
	def gotoMaterial(self, event=None):
		if self.editObject and self.editObject.type() == Card.REGION:
			self.selectGeometryPropertiesTab()
			materials = [card[Card.MATERIAL] for card in self.editObject.cards]
			self.select(materials, True)

	# ----------------------------------------------------------------------
	# Show body/region transformation ROT-DEFI card
	# ----------------------------------------------------------------------
	def gotoRotdefi(self):
		if not self.editObject: return
		if self.editObject.type() == Card.BODY:
			key = "@transform"
		elif self.editObject.type() == Card.REGION:
			key = "@rotdefi"
		lst = [card[key] for card in self.editObject.cards]
		if lst[0] is None: return
		# Find cards to display
		rotdefi = []
		for r in lst:
			if r and r[0]=="-": r=r[1:]
			rot = self.findCard("T",r)
			if rot: rotdefi.append(rot)
		self.selectGeometryAttributesTab()
		self.select(rotdefi, True)

	# ----------------------------------------------------------------------
	def gotoLattice(self):
		self.viewers.activeViewer().go2Lattice()

	# -----------------------------------------------------------------------
	# Delete action on property list
	# Regions: Delete zones
	# ----------------------------------------------------------------------
	def propListDelete(self, event=None):
		if self.editObject is None: return
		if self.editObject.type() == Card.REGION:
			card = self.editObject.cards[0]
			zones = card[Card.ZONE]
			for sel in reversed(self.propList.curselection()):
				z = self.zoneIndex(sel)
				if z is not None:
					del zones[z]
			self.flair.addUndo(self.flair.setZonesUndo(card,zones))
			self.flair.refresh("redraw", self)
			self.inputModified()

			# FIXME temporarily reload all
			self.loadProject()
			# FIXME Show properties after re-parsing the zones...
			self.updateProperties()

		else:
			cursel = self.propList.curselection()
			for sel in cursel:
				name = self.propList.listbox(0).get(sel)
				self.editObject.value(name,"")

			# Special check for ROT-DEFI
			if isinstance(self.editObject, RotdefiProperties):
				# if the selected card has all elements removed
				# remove also the card
				# Find index of cards
				idx = []
				for sel in cursel:
					name = self.propList.listbox(0).get(sel)
					try: i = int(name.split(".")[-1])
					except: continue
					if i not in idx: idx.append(i)
				idx.reverse()
				# Check if card is empty
				undoinfo = None
				for i in idx:
					card = self.editObject._rotdefi[i-1]
					for j in range(1,7):
						if card.what(j)!="": break
					else:
						if not undoinfo:
							undoinfo = self.flair.undolistRefresh()
						undoinfo.append(self.flair.delCardUndo(card.pos(),False))
				if undoinfo:
					self.flair.addUndolistRefresh(undoinfo, "Card property change")

		self.refreshUndoButton()

	# ----------------------------------------------------------------------
	# Move Up action in properties
	# Region: Move up zone
	# ----------------------------------------------------------------------
	def propListMoveUp(self, event=None):
		if self.editObject is None: return
		if self.editObject.type() == Card.REGION and \
		   len(self.editObject.cards)==1:
			card   = self.editObject.cards[0]
			zones  = card[Card.ZONE]
			moved  = False
			newsel = []
			for i,sel in enumerate(self.propList.curselection()):
				z = self.zoneIndex(sel)
				if z is None: continue
				if z==i:
					newsel.append(int(sel))
					continue
				newsel.append(int(sel)-1)
				zones.insert(z-1, zones.pop(z))
				moved = True

			if moved:
				self.flair.addUndo(self.flair.setZonesUndo(card,zones))
				self.flair.refresh("redraw", self)
				self.inputModified()
				self.updateProperties()
				if newsel:
					for i in newsel:
						self.propList.selection_set(i)
					self.propList.activate(newsel[0])
			return "break"

	# ----------------------------------------------------------------------
	# Move Down action in properties
	# Region: Move down zone
	# ----------------------------------------------------------------------
	def propListMoveDown(self, event=None):
		if self.editObject is None: return
		if self.editObject.type() == Card.REGION and \
		   len(self.editObject.cards)==1:
			card   = self.editObject.cards[0]
			zones  = card[Card.ZONE]
			n      = len(zones)-1
			moved  = False
			newsel = []
			for i,sel in enumerate(reversed(self.propList.curselection())):
				z = self.zoneIndex(sel)
				if z is None: continue
				if z==n-i:
					newsel.append(int(sel))
					continue
				newsel.append(int(sel)+1)
				zones.insert(z+1, zones.pop(z))
				moved = True

			if moved:
				self.flair.addUndo(self.flair.setZonesUndo(card,zones))
				self.flair.refresh("redraw", self)
				self.inputModified()
				self.updateProperties()
				if newsel:
					for i in newsel:
						self.propList.selection_set(i)
					self.propList.activate(newsel[0])
			return "break"

	# ----------------------------------------------------------------------
	# Clone action on property list
	# Regions: Clone zone
	# ----------------------------------------------------------------------
	def propListClone(self, event=None):
		if self.editObject is None: return
		if self.editObject.type() == Card.REGION and \
		   len(self.editObject.cards)==1:
			card   = self.editObject.cards[0]
			zones  = card[Card.ZONE]
			n = len(zones)
			for sel in self.propList.curselection():
				z = self.zoneIndex(sel)
				if z is None: continue
				zones.append(zones[z])

			self.flair.addUndo(self.flair.setZonesUndo(card,zones))
			self.flair.refresh("redraw", self)
			self.inputModified()
			self.updateProperties()
			for i in range(n,len(zones)):
				self.propList.selection_set(i+3)
			self.propList.activate(n)
			return "break"

	# ----------------------------------------------------------------------
	def propListEditEvent(self, event=None):
		if not self.propList.curselection():
			active = self.propList.index(tk.ACTIVE)
			self.propList.selection_set(active)
		self.propListEdit()

	# ----------------------------------------------------------------------
	def propListEdit(self, next_=True):
		if len(self.propList.curselection())!=1: return

		active = int(self.propList.curselection()[0])
		self.propList.activate(active)

		lbl, value = self.propList.get(active)
		edit = self.editObject.edit(self.propList)

		if edit is not None:
			if edit.value is None:
				return
			elif edit.value != value:
				if self.editObject and self.editObject.split():
					# Split value and propagate change to next fields
					if edit.value=="" or (edit.value and edit.value[0]=="="):
						values = [edit.value]
					else:
						values = _SPLIT_PAT.sub(" ", edit.value).split()
					for i, value in enumerate(values):
						self.propList.set(active,(lbl, value))
						self.editObject.value(lbl, value)
						if i == len(values)-1: break
						active += 1
						if active >= self.propList.size(): break
						self.propList.activate(active)
						lbl = self.propList.get(active)[0]
				else:
					self.editObject.value(lbl, edit.value)
					if self.editObject.type() == Card.REGION:
						self.updateProperties()
			if edit.lastkey == "Up":
				active -= 2
			elif edit.lastkey not in ("Return", "KP_Enter", "Down"):
				return

		# edit next item
		if not next_: return
		active += 1
		if active < self.propList.size():
			self.propList.selection_clear(0,tk.END)
			self.propList.selection_set(active)
			self.propList.activate(active)
			self.propList.see(active)
			self.page.event_generate("<Return>")

	# ----------------------------------------------------------------------
	# Attribute List popup menu
	# Create popup menu with default actions
	# ----------------------------------------------------------------------
	def attrListPopup(self, event):
		if self.editObject is None: return
		self.page.focus_set()

		menu=tk.Menu(self.page, tearoff=0)

		#menu.add_command(label="Cut", underline=2,
		#		command=self.propListCut,
		#		compound=tk.LEFT, image=icons.get("cut"),
		#		accelerator="Ctrl-X")
		#menu.add_command(label="Copy", underline=0,
		#		command=self.propListCopy,
		#		compound=tk.LEFT, image=icons.get("copy"),
		#		accelerator="Ctrl-C")
		#menu.add_command(label="Paste", underline=0,
		#		command=self.propListPaste,
		#		compound=tk.LEFT, image=icons.get("paste"),
		#		accelerator="Ctrl-V")
		#menu.add_command(label="Clone", underline=0,
		#		state=tk.DISABLED,
		#		compound=tk.LEFT, image=icons.get("clone"))
		#menu.add_separator()
		menu.add_command(label="Delete", underline=0,
				command=self.attrListDelete,
				compound=tk.LEFT, image=icons.get("x"))
		menu.add_separator()

		menu.add_command(label="All", underline=0,
				command=self.attrList.selectAll,
				compound=tk.LEFT, image=icons.get("select_all"))
		menu.add_command(label="Clear", underline=0,
				command=self.attrList.selectClear,
				compound=tk.LEFT, image=icons.get("select_none"))
		menu.add_command(label="Invert", underline=0,
				command=self.attrList.selectInvert,
				compound=tk.LEFT, image=icons.get("select_invert"))

		menu.tk_popup(event.x_root, event.y_root)
		return "break"

	# ----------------------------------------------------------------------
	# Button Release on the attribute listbox
	# ----------------------------------------------------------------------
	def attrListRelease(self, event):
		if self.editObject is None: return
		act = event.widget.nearest(event.y)
		self.attrList.activate(act)
		# Find list box
		if self.attrList.listboxIndex(event.widget) == 0: return

		name, value = self.attrList.get(act)
		if value in (tkFlair._ON, tkFlair._OFF, tkFlair._MULTI):
			if value == tkFlair._OFF:
				value = tkFlair._ON
			else:
				value = tkFlair._OFF
			self.editObject.value(name, value==tkFlair._ON)
			self.attrList.listbox(1).set(act, value)
			self.attrList.listbox(1).selection_set(act)
			self.viewers.check4Frozen()
			self.selectShow(True, False)
			self.treeColors()
		else:
			self.attrListEdit()
		return "break"

	# ----------------------------------------------------------------------
	def attrListDelete(self, event=None):
		if self.editObject is None: return
		for sel in self.attrList.curselection():
			name, value = self.attrList.get(sel)
			if value in (tkFlair._ON, tkFlair._OFF, tkFlair._MULTI):
				value = False
			else:
				value = ""
			self.editObject.value(name,value)

		self.refreshUndoButton()

	# ----------------------------------------------------------------------
	def attrListEdit(self, next_=True):
		if len(self.attrList.curselection())!=1: return

		active = int(self.attrList.curselection()[0])
		self.attrList.activate(active)
		lbl, value = self.attrList.get(active)
		edit = self.editObject.edit(self.attrList)

		if edit is not None:
			if edit.value is None:
				return
			elif edit.value != value:
				if self.editObject.split2():
					# Split value and propagate change to next fields
					if edit.value=="" or (edit.value and edit.value[0]=="="):
						values = [edit.value]
					else:
						values = _SPLIT_PAT.sub(" ", edit.value).split()
					for i, value in enumerate(values):
						self.attrList.set(active,(lbl, value))
						self.editObject.value(lbl, value)
						if i == len(values)-1: break
						active += 1
						if active >= self.attrList.size(): break
						self.propList.activate(active)
						lbl = self.attrList.get(active)[0]
				else:
					self.editObject.value(lbl, edit.value)
					if self.editObject.type() == Card.REGION:
						self.updateProperties()
			if edit.lastkey == "Up":
				active -= 2
			elif edit.lastkey not in ("Return", "KP_Enter", "Down"):
				return

		# edit next item
		if not next_: return
		active += 1
		if active < self.attrList.size():
			self.attrList.selection_clear(0,tk.END)
			self.attrList.selection_set(active)
			self.attrList.activate(active)
			self.page.event_generate("<Return>")

	# ----------------------------------------------------------------------
	def findUniqueName(self, obj, name=None):
		last = False
		if obj == Card.BODY:
			func = self.viewers.body
			if name is None:
				last = True
				name = self._lastBodyName
		elif obj == Card.REGION:
			func = self.viewers.region
			if name is None:
				last = True
				name = self._lastRegionName
		elif obj == Card.OBJECT:
			func = self.viewers.object
		else:
			return None

		# Find unique name
		if last:
			guess = f"{name[:7]}1"
		else:
			guess = name[:8]
		try:
			func(guess,"id")
			n = 1
			# find number in the end of the name
			pat = _LAST_NUM.match(name)
			if pat:
				name = pat.group(1)
				n = int(pat.group(2))+1
			while True:
				# Constraint in 8 characters
				sn = str(n)
				guess = "%s%s"%(name[:8-len(sn)],sn)
				try:
					func(guess,"id")
					n += 1
				except KeyError:
					break
		except KeyError:
			pass
		return guess

	# ----------------------------------------------------------------------
	def changeName(self, oldname, newname, card=None):
		if newname == "": return
		if card is None:
			card = self.editObject.cards[0]

		if card.type() in (Card.BODY, Card.REGION):
			pat = _LAST_NUM.match(newname)
			if pat: prefix = pat.group(1)
			else: prefix = newname

			count = 0
			if card.type() == Card.REGION:
				ctype = "region"
				stype = "Region"
				# Store last name
				self._lastRegionName = prefix
				# Count duplicates
				for c in self.project.input.cards["REGION"]:
					if c.name() == oldname:
						count += 1
			else:
				ctype = "body"
				stype = "Body"
				# Store last name
				self._lastBodyName = prefix
				# Count duplicates
				for body in CardInfo.bodies[card.MODE]:
					for c in self.project.input[body]:
						if c.name() == oldname:
							count += 1

			if oldname!="" and count>1:
				change = tkFlair.askyesno("Name changed",
						f"'{stype}' name has changed\n"
						f"'{oldname}' \N{RIGHTWARDS ARROW} '{newname}'\n"
						"Change all references in input?",
						parent=self.page)
			else:
				change = (oldname!="")

			if change:
				undoinfo = [ (self.flair.refreshUndo()),
					   self.flair.changeNameUndo(ctype, oldname, newname),
					   self.flair.setWhatUndo(card,0,newname),
					  (self.flair.refreshUndo()) ]
				self.treeFill()
			else:
				undoinfo = self.flair.setWhatUndo(card,0,newname)

			self.flair.addUndo(undoinfo)

		else:
			self.flair.addUndo(self.flair.setWhatUndo(card,0,newname))
		self.project.input.clearCache()

	# ----------------------------------------------------------------------
	# Find insertion position for bodies and/or regions/cells
	# ----------------------------------------------------------------------
	def _findInsertPosition(self, type_):
		forRegion = int(type_==Card.REGION)
		if not self.project.input.cardlist:
			# create the bare minimum input
			self.project.input.minimumInput()
			self.flair.redraw()
		if self.project.ismode(Fluka):
			ends = self.project.input["END"]
			if len(ends) != 2:
				self.flair.notify("Incorrect GEOMETRY card structure",
					"The basic geometry cards are not found in your input\n"
					"GEOBEGIN, END, END, GEOEND",
					tkFlair.NOTIFY_ERROR)
			try:
				return ends[forRegion].pos()
			except:
				pass
			if forRegion:
				if "LATTICE" in self.project.input.cards:
					return self.project.input.bestPosition("LATTICE")-1
				else:
					return self.project.input.bestPosition("GEOEND")-1
			else:
				return self.project.input.bestPosition("REGION")-1

		#---------------------------------------------------------------
		elif self.project.ismode((Mcnp,Phits)):
			if forRegion:
				cards = list(self.project.input.cardsCache(Card.REGION,-1).values())
				if cards: return cards[-1].pos()+1
				cards = self.project.input["[Cell]"]
				if cards: return cards[0].pos()+1
			else:
				cards = list(self.project.input.cardsCache(Card.BODY,-1).values())
				if cards: return cards[-1].pos()+1
				cards = self.project.input["[Surface]"]
				if cards: return cards[0].pos()+1

		#---------------------------------------------------------------
		try:
			return self.project.input.cardlist[-1].pos()
		except IndexError:
			return 0

	# ----------------------------------------------------------------------
	# Add a body at initial location, with size and z-direction
	# ----------------------------------------------------------------------
	def addBody(self, card):
		dump("*-* GeometryEditor.addBody")
		self.propTab.select(0)
		pos = self._findInsertPosition(False)
		undoinfo = [(self.flair.refreshUndo,)]
		undoinfo.append(self.flair.addCardUndo(card,pos,True))
		undoinfo.append((self.flair.refreshUndo,))
		self.addUndo(undoinfo, f"Add body {card.tag}")
		#self.flair.refresh("card", self)
		self.flair.refresh("card")
		self.select(card, True)		# Select
		card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.VISIBLE	# and make visible
		self.treeFill()
		self.treeColors()
		return card

	# ----------------------------------------------------------------------
	# Add a new region
	# ----------------------------------------------------------------------
	def addRegion(self, edit=True, mat=None, expr=None):
		self.propTab.select(0)
		pos = self._findInsertPosition(Card.REGION)
		name = self.findUniqueName(Card.REGION)

		undoinfo = [(self.flair.refreshUndo,)]
		if self.project.ismode(Fluka):
			# Create REGION card
			card = FlukaCard("REGION", [name])
			if expr: card.setExpr(expr)
			undoinfo.append(self.flair.addCardUndo(card, pos, True))
			# Create ASSIGNMA card
			pos = self.project.input.bestPosition("ASSIGNMA")
			if mat is None: mat = self._lastMaterial
			undoinfo.append(
				self.flair.addCardUndo(
					FlukaCard("ASSIGNMA", ["",mat, name]), pos, True))

		elif self.project.ismode(Moira):
			card = MoiraCard("CELL", [name, "", self._lastMaterial])
			if expr: card.setExpr(expr)
			undoinfo.append(self.flair.addCardUndo(card, pos, True))

		elif self.project.ismode(Mcnp, Phits):
			card = self.project.input.newCard("CELL",
						[name, self._lastMaterial])
			if expr: card.setExpr(expr)
			undoinfo.append(self.flair.addCardUndo(card, pos, True))
		else:
			card = None

		undoinfo.append((self.flair.refreshUndo,))
		self.addUndo(undoinfo, "Add region")
		self.flair.refresh("card")	# refresh self as well
		self.select(card, edit)
		if edit: self.editName()
		return card

	# ----------------------------------------------------------------------
	# Add a ROT-DEFI card
	# ----------------------------------------------------------------------
	def addRotdefi(self):
		self.propTab.select(0)
		if self.project.ismode(Moira):
			# tag  = "MATRIX"
			tag = "ROTM"
			card = MoiraCard(tag)
		elif self.project.ismode(Fluka):
			tag  = "ROT-DEFI"
			card = FlukaCard(tag)
		elif self.project.ismode(Mcnp):
			tag  = "TR"
			card = McnpCard(tag)
		elif self.project.ismode(Phits):
			tag  = "TR"
			card = PhitsCard(tag)
		pos = self.project.input.bestPosition(card.tag)
		undoinfo = [(self.flair.refreshUndo,)]
		undoinfo.append(self.flair.addCardUndo(card, pos, True))
		undoinfo.append(undoinfo[0])
		self.addUndo(undoinfo, "Add ROT-DEFI")
		self.flair.refresh("card", self)
		self.select(card, True)					# Select
		self.treeFill()
		card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.VISIBLE	# and make visible
		self.treeColors()
		self.propListEdit()
		self.editName()
		return card

	# ----------------------------------------------------------------------
	# Add a MATRIX card
	# ----------------------------------------------------------------------
	def addRotMat(self):
		self.propTab.select(0)
		tag  = "MATRIX"
		card = MoiraCard(tag)
		pos = self.project.input.bestPosition(card.tag)
		undoinfo = [(self.flair.refreshUndo,)]
		undoinfo.append(self.flair.addCardUndo(card, pos, True))
		undoinfo.append(undoinfo[0])
		self.addUndo(undoinfo, "Add MATRIX")
		self.flair.refresh("card", self)
		self.select(card, True)					# Select
		self.treeFill()
		card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.VISIBLE	# and make visible
		self.treeColors()
		self.propListEdit()
		self.editName()
		return card

	# ----------------------------------------------------------------------
	# Add a new object
	# ----------------------------------------------------------------------
	def addObject(self, card):
		self.propTab.select(0)
		pos = self.project.input.bestPosition(card.tag)
		# Add only card do not refresh the display
		undoinfo = [(self.flair.refreshUndo,)]
		undoinfo.append(self.flair.addCardUndo(card,pos,True))
		undoinfo.append(undoinfo[0])
		self.addUndo(undoinfo, "Add new object "+card.tag)
		self.flair.refresh("card", self)
		self.select(card, True)					# Select
		self.treeFill()
		card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.VISIBLE	# and make visible
		self.treeColors()
		self.propListEdit()
		return card

	# ----------------------------------------------------------------------
	# create bodies from mesh
	# ----------------------------------------------------------------------
	def mesh2Bodies(self, event=None):
		for card in self.project.input.cardlist:
			if card.ignore(): continue
			if not card.get(Card.SELECT,0) & BIT.SELECT: continue
			if card.type() != Card.OBJECT: continue
			if card.tag != "mesh": continue

			oid = card[Card.ID]
			log.say("Object:", card.name(), oid)
			bodies = self.geometry().object(oid,"bodies")
			log.say("Bodies:", bodies)
			for i,what in enumerate(bodies,1):
				tag = what[0]
				what[0] = f"{card.name()}{i}"
				body = FlukaCard(tag, what)
				self.project.input.add(body)
				log.say(">>>",body)
				#self.addBody(body)
		self.loadProject()
		self.redraw()

	# ======================================================================
	# Undo
	# ======================================================================
	def addUndo(self, undoinfo, msg=None): self.flair.addUndo(undoinfo,msg)
	def canUndo(self):	return self.flair.canUndo()
	def canRedo(self):	return self.flair.canRedo()

	# ----------------------------------------------------------------------
	def undo(self, event=None):
		self.flair.undoInput()

	# ----------------------------------------------------------------------
	def redo(self, event=None):
		self.flair.redoInput()

	# ----------------------------------------------------------------------
	# Called from flair when undo is requested
	# ----------------------------------------------------------------------
	def refreshUndo(self):
		dump("*-* GeometryEditor.refreshUndo")
		if not self.created(): return
		self.refreshLate()

	# ----------------------------------------------------------------------
	# Force refresh if any pending
	# @see GeometryViewer: cloning of objects with mouse!!!
	# FIXME maybe remove
	# ----------------------------------------------------------------------
	def forceRefresh(self):
		dump("*-* GeometryEditor.forceRefresh")
		if self.__afterRefresh is not None:
			self.page.after_cancel(self.__afterRefresh)
			self.__afterRefresh = None
		self._refresh()

	# ----------------------------------------------------------------------
	# Reload/draw waiting for all undo commands to finalize resizes the action handle
	#----------------------------------------------------------------------
	def refreshLate(self):
		if self.__afterRefresh is not None:
			self.page.after_cancel(self.__afterRefresh)
		self.__afterRefresh = self.page.after_idle(self._refresh)

	# ----------------------------------------------------------------------
	def _refresh(self):
		dump("*-* GeometryEditor.refreshLate")
		self.__afterRefresh = None
		self.loadProject()	# Load project calls fillRun -> runChange -> redraw
		self.updateErrors()
		if self.editObject is not None:
			self.editObject.fill()
			self.viewers.updateEdit()
		self.redraw()

	# ----------------------------------------------------------------------
	def activateIfNeeded(self, card):
		#dump("*-* GeometryEditor.activateIfNeeded")
		if self.isactive() and not self.flair.inundo(): return

		# Active only cards handled by GeometryEditor
		if card.type() not in (Card.BODY, Card.REGION, Card.OBJECT):
			return

		if self.editObject and card in self.editObject.cards:
			return

		self.select(card, True)

	# ----------------------------------------------------------------------
	# Standard cut, copy and paste
	# ----------------------------------------------------------------------
	def treeCut(self, event=None):
		self.treeCopy()
		self.delete()
		return "break"

	# ----------------------------------------------------------------------
	# Copy selected items to clipboard
	# ----------------------------------------------------------------------
	def treeCopy(self, event=None):
		cards = []
		for card in self.project.input.cardlist:
			if card.ignore(): continue
			if card.get(Card.SELECT,0) & BIT.SELECT:
				cards.append(card)
		if cards:
			self.flair.cards2Clipboard(cards)
		return "break"

	#-----------------------------------------------------------------------
	# Deep copy of a region card
	#-----------------------------------------------------------------------
	def regionDeepCopy(self, region, origname, bodies, bodyPos, undoinfo):
		expr = region.expr()

		# 1st clone the bodies
		for item in csg.tokenize(expr):
			if item in "+-|:()": continue
			if item in bodies: continue

			# clone the body
			body = self.findCard("B", item)
			if body is None: continue
			# cloning body body.name()
			card = body.clone()
			name = self.findUniqueName(card.type(),card.name())
			card.setName(name)
			self.geometry().body(name, "add", card.tag)
			bodies[item] = card
			undoinfo.append(self.flair.addCardUndo(card, bodyPos, False))
			bodyPos += 1

		# 2nd correct expression
		for name,card in bodies.items():
			expr = re.sub(rf"\b{name}\b", card.name(), expr)
		region.setExpr(expr)

		# 3rd add any associated card, ASSIGNMA, BIASING?
		# FIXME will not work properly as we are not renumbering before!!!!
		oldRegion = self.findCard("R",origname)
		if oldRegion:
			pos = self.project.input.bestPosition("ASSIGNMA")
			mat = oldRegion[Card.MATERIAL].name()
			undoinfo.append(self.flair.addCardUndo(
				FlukaCard("ASSIGNMA", ["", mat, region.name()]),
					pos, False))
		return len(bodies)

	# ----------------------------------------------------------------------
	# Paste cards and return undo information (with no refresh)
	# ----------------------------------------------------------------------
	def _cardsPaste(self, cards, undoinfo, deep=False):
		# find positions
		bodyPos = self._findInsertPosition(Card.BODY)
		regPos  = self._findInsertPosition(Card.REGION)
		objPos  = self.project.input.bestPosition("object")+1

		pasted   = []
		bodies = {}	# Duplicated bodies
		renamed = ""
		undoinfo.append(self.flair.renumberUndo())
		for card in cards:
			if card.type() == Card.BODY:
				pos = bodyPos

			elif card.type() == Card.REGION:
				pos = regPos

			elif card.type() == Card.OBJECT:
				pos = objPos

			else:
				continue

			# Increase pointers
			if bodyPos >= pos: bodyPos += 1
			if regPos  >= pos: regPos  += 1
			if objPos  >= pos: objPos  += 1

			# Set visibility flag for bodies/objects when duplicate
			visible = False

			origname = card.name()
			name = self.findUniqueName(card.type(),card.name())
			if name != origname:
				renamed += f"   {card.tag}:\t{card.name()}\t-> {name}\n"
				card.setName(name)
				visible = card.type() in (Card.BODY, Card.OBJECT)

			# Update geometry to register the name, to avoid
			# duplicate names from findUniqueName
			if card.type() == Card.BODY:
				try:
					whats = [card.numWhat(i) for i in range(1,card.nwhats()) ]
				except ValueError:
					whats = []
				card[Card.ID] = self.viewers.addBody(name, card.tag, whats)

			elif card.type() == Card.REGION:
				if deep:
					# Create a deep clone of the card!
					bodiesAdded = self.regionDeepCopy(card, origname, bodies,
									bodyPos, undoinfo)
					bodyPos += bodiesAdded
					if regPos  >= bodyPos: regPos  += bodiesAdded
					if objPos  >= bodyPos: objPos  += bodiesAdded
					if pos     >= bodyPos: pos += bodiesAdded
				card[Card.ID] = self.geometry().region(name, "add")
			if visible: card[Card.SELECT] = card.get(Card.SELECT,0) | BIT.VISIBLE

			pasted.append(card)
			undoinfo.append(self.flair.addCardUndo(card, pos, False))
		undoinfo.append(self.flair.renumberUndo())

		self.viewers.derive()

		return pasted, renamed

	# ----------------------------------------------------------------------
	# Paste clipboard items to input
	# ----------------------------------------------------------------------
	def treePaste(self, event=None):
		cards = self.flair.clipboard2Cards()
		if not cards:
			if self.editType() == Card.REGION:
				return self.propListPaste(event)
			return "break"

		self.selectClear(False)
		self.viewers.stopThreads()

		undoinfo = [self.flair.refreshUndo()]
		pasted,renamed = self._cardsPaste(cards,undoinfo)
		undoinfo.append(self.flair.refreshUndo())
		self.addUndo(undoinfo, "Paste card(s)")

		if renamed:
			self.flair.notify("Cards renamed",
				f"The following cards have been renamed:\n{renamed}",
				tkFlair.NOTIFY_WARNING)

		self.project.input.checkNumbering()	# XXX
		self.project.input.clearCache()

		if pasted:
			self.flair.activateIfNeeded(pasted[-1])
		return "break"

	# ----------------------------------------------------------------------
	# Paste special makes a deep copy of the objects
	# ----------------------------------------------------------------------
	def treePasteSpecial(self, event=None):
		cards = self.flair.clipboard2Cards()
		if not cards: return

		# find position to insert the cards
		#unknownpos = True
		#active = self.tree.focus()
		#if active:
		#	try:
		#		activecard = self.tree2card[active]
		#		unknownpos = False
		#		pos = activecard.card.pos()+1
		#	if activecard:

		self.selectClear(False)
		self.viewers.stopThreads()

		undoinfo = [self.flair.refreshUndo()]
		pasted,renamed = self._cardsPaste(cards, undoinfo, True)
		undoinfo.append(self.flair.refreshUndo())
		self.addUndo(undoinfo, "Paste card(s)")

		# Testing possibility to reload the page immediately and select
		#if self.__afterRefresh is not None:
		#	self.page.after_cancel(self.__afterRefresh)
		if renamed:
			self.flair.notify("Cards renamed",
				f"The following cards have been renamed:\n{renamed}",
				tkFlair.NOTIFY_WARNING)

		self.project.input.checkNumbering()	# XXX
		self.project.input.clearCache()

		# Unfortunately loading of the geometry is done at later state
		# from the refreshUndo() refreshLate
		# no possibility to select now

		# if we force the loading we can select
		# but we end up with double loading...
		if pasted:
			self.flair.activateIfNeeded(pasted[-1])
		return "break"

	# ----------------------------------------------------------------------
	# Delete key was pressed
	# Location     Action
	# --------     --------------------------
	# tree         Delete all selected items (call treeDelete)
	# propList     Delete zones if any (call propListDelete)
	# Viewports    1. If zone selected call propListDelete
	#              2. If bodies are selected delete only the bodies
	#              3. Delete all selected cards
	# ----------------------------------------------------------------------
	def delete(self, event=None):
		focus = self.page.focus_get()

		if focus is self.tree:
			self.treeDelete()

		elif focus is self.propList:
			self.propListDelete()

		else:
			t = False	# type of objects to delete
			# 1st if a region is selected
			if self.editType() == Card.REGION and \
			   self.propList.curselection():
				self.propListDelete()
				return

			# 2nd if objects are selected
			elif self.isAnythingSelected(Card.OBJECT):
				t = Card.OBJECT

			# 3rd if bodies are selected
			elif self.isAnythingSelected(Card.BODY):
				t = Card.BODY

			# bodies and regions removed, to correct references!
			bodiesRemoved = []
			regionsRemoved = []

			undoinfo = []
			for card in reversed(self.project.input.cardlist):
				if card.ignore(): continue
				if card.get(Card.SELECT,0) & BIT.SELECT == 0: continue
				if t and card.type() != t: continue
				if not undoinfo:
					# if we don't renumber cards everything gets screwed up
					#undoinfo.append(self.flair.renumberUndo())
					undoinfo.append(self.flair.refreshUndo())

				card[Card.SELECT] = card.get(Card.SELECT,0) & BIT.UNSELECT
				if card.type() == Card.BODY:
					bodiesRemoved.append(card.name())
				elif card.type() == Card.REGION:
					regionsRemoved.append(card.name())

				undoinfo.append(self.flair.delCardUndo(card.pos(),True))

			# Correct references to bodies
			for body in bodiesRemoved:
				undoinfo.extend(self.flair.removeBodyRefUndo(body))

			# Correct references to regions
			for region in regionsRemoved:
				undoinfo.extend(self.flair.removeRegionRefUndo(region))

			if undoinfo:
				#undoinfo.append(self.flair.renumberUndo())
				undoinfo.append(self.flair.refreshUndo())
				self.addUndo(undoinfo, "Delete card(s)")
			self.refreshUndoButton()

	# ----------------------------------------------------------------------
	# Update region and body properties in geometry
	# ----------------------------------------------------------------------
	def updateGeometry(self):
		self.viewers.updateGeometry()

	# ----------------------------------------------------------------------
	def updateLayers(self, event=None):
		self.modified()
		self.viewers.updateLayers(self._layers.layer.name)
		self.viewers.updateSelection()

	# ----------------------------------------------------------------------
	def updateErrors(self, event=None):
		if self.tabSet.index(tk.CURRENT) == 2:
			self._errors.fill()

	# ----------------------------------------------------------------------
	def selectGeometryTab(self):
		self.tabSet.select(0)

	# ----------------------------------------------------------------------
	def selectGeometryPropertiesTab(self):
		self.tabSet.select(0)
		self.propTab.select(0)

	# ----------------------------------------------------------------------
	def selectGeometryAttributesTab(self):
		self.tabSet.select(0)
		self.propTab.select(1)

	# ----------------------------------------------------------------------
	def selectLayersTab(self):
		self.tabSet.select(1)

	# ----------------------------------------------------------------------
	def selectErrorsTab(self):
		self.tabSet.select(2)

	# ----------------------------------------------------------------------
	def changeTab(self, event=None):
		page = self.tabSet.index(tk.CURRENT)
		if page == 1:	# Layers
			self._layers.show()
		elif page == 2:	# Errors
			self._errors.refresh()

	# ----------------------------------------------------------------------
	def showLayer(self, layer):
		self._layers.show(layer)
		self.selectLayersTab()

	# ----------------------------------------------------------------------
	# Force redrawing of viewports
	# ----------------------------------------------------------------------
	def redraw(self, event=None):
		dump("*-* GeometryEditor.redraw", self.created())
		if not self.created(): return
		if self.__afterRefresh: return		# If a reload/redraw is pending quit
		if self.tabSet.index(tk.CURRENT) == 1:	# Layers
			self._layers.refresh()
		try:
			if Global.profiler: timer = tool.Timer()
			self.viewers.redraw()
			if Global.profiler: timer.stop("+++ GeometryEditor.redraw:")
			self.refreshUndoButton()
		except tk.TclError:
			# XXX if we close the geoviewer then the
			# flair project is not informed containing
			# a self.geoedit pointer to a non Tk object
			pass

	# ----------------------------------------------------------------------
	# Set mouseAction IntVar with the appropriate action
	# ----------------------------------------------------------------------
	def setMouseAction(self, action):
		if self.lockEditing.get() and action in (ACTION.ADD, ACTION.MOVE, ACTION.ROTATE):
			self.flair.notify("Lock enabled",
					"Cannot add or edit geometry while the Lock is enabled",
					tkFlair.NOTIFY_WARNING)
			return
		self._mouseAction.set(action)

	# ----------------------------------------------------------------------
	# Inform viewers that mouse Action has changed
	# ----------------------------------------------------------------------
	def mouseActionChanged(self, *abc):
		self.viewers.setMouseAction(self._mouseAction.get())

	# ----------------------------------------------------------------------
	def toggleSnap2Grid(self, event=None):
		self.snap2Grid.set(not self.snap2Grid.get())

	# ----------------------------------------------------------------------
	def viewLayout(self, name=None):
		if name is None: name = "2x2"
		self.viewers.layout(name)

	# ----------------------------------------------------------------------
	def transformGeometry(self, event=None):
		cardlist = []
		for card in self.project.input.cardlist:
			if card.ignore(): continue
			if card.type() == Card.BODY and card.get(Card.SELECT,0)&BIT.SELECT:
				cardlist.append(card)
		GeometryBody.TransformGeometryDialog(self.page, self.flair, cardlist)

	# ----------------------------------------------------------------------
	def loadMaterialColors(self):
		self.viewers.loadMaterialColors()
		self.viewers.redraw()

	# ----------------------------------------------------------------------
	# Convert BOXes to RPPs
	# ----------------------------------------------------------------------
	#def box2rpp(self):
	#	if self.editObject is None: return
	#	cards = self.getSelection(objtype=Card.BODY)
	#	for card in reversed(cards):
	#		if card.tag != "BOX": continue
	#		GeometryBody.box2rpp(card)

	# ----------------------------------------------------------------------
	# Expand macro body to infinite ones
	# FIXME should be moved to GeometryBody
	# ----------------------------------------------------------------------
	def expandBody(self):
		if self.editObject is None: return
		cards = self.getSelection(objtype=Card.BODY)
		if not cards: return
		self.busy()

		#---------------------------------------------------------------
		# add a unique body and flag it as visible
		#---------------------------------------------------------------
		def addUniqueBody(card, tag, whats, pos, comment=""):
			nonlocal undoinfo
			n = self.findUniqueName(Card.BODY, card.name())
			self.geometry().body(n, "add", tag)
			whats[0] = n
			body = FlukaCard(tag, whats, comment)
			body.copyProperties(card)
			body[Card.SELECT] = body.get(Card.SELECT,0) | BIT.VISIBLE
			undoinfo.append(self.flair.addCardUndo(body, pos, True))
			return body

		#---------------------------------------------------------------
		# populate the dictionary replace with the pattern of body+expr
		#---------------------------------------------------------------
		def add2Replace(body, expr):
			nonlocal replace
			replace[re.compile(rf"\b{body.name()}\b")] = expr
			self._late_messages.append(f"    expr= {expr}")

		undoinfo = self.flair.undolistRefresh()
		bodies_count = 0
		regions_count = 0
		replace = {}
		for body in cards:
			pos = body.pos()+1
			bodies_count += 1
			#if body.tag not in ("RPP","BOX","RCC","TRC","WED","REC","ARB"): continue
			if body.tag == "RPP":
				x0 = addUniqueBody(body, "YZP", [None, body.numWhat(1)],
								pos, body.comment())
				x1 = addUniqueBody(body, "YZP", [None, body.numWhat(2)], pos+1)
				y0 = addUniqueBody(body, "XZP", [None, body.numWhat(3)], pos+2)
				y1 = addUniqueBody(body, "XZP", [None, body.numWhat(4)], pos+3)
				z0 = addUniqueBody(body, "XYP", [None, body.numWhat(5)], pos+4)
				z1 = addUniqueBody(body, "XYP", [None, body.numWhat(6)], pos+5)
				self._late_messages.append(f"Body '{body.name()}' expanded to:")
				self._late_messages.append(f"    body '{x0.name()}' {x0.tag}")
				self._late_messages.append(f"    body '{x1.name()}' {x1.tag}")
				self._late_messages.append(f"    body '{y0.name()}' {y0.tag}")
				self._late_messages.append(f"    body '{y1.name()}' {y1.tag}")
				self._late_messages.append(f"    body '{z0.name()}' {z0.tag}")
				self._late_messages.append(f"    body '{z1.name()}' {z1.tag}")
				add2Replace(body, f"(+{x1.name()} -{x0.name()} "
						   f"+{y1.name()} -{y0.name()} "
						   f"+{z1.name()} -{z0.name()})")
			elif body.tag in ("BOX","WED"):
				P = body.bodyP()
				X = body.bodyX()
				Y = body.bodyY()
				Z = body.bodyZ()
				PO = P+X+Y+Z
				self._late_messages.append(f"Body '{body.name()}' expanded to:")
				x0 = addUniqueBody(body, "PLA", ([None]+X)+P,  pos, body.comment())
				self._late_messages.append(f"    body '{x0.name()}' {x0.tag}")
				if body.tag == "BOX":
					pos += 1
					x1 = addUniqueBody(body, "PLA", ([None]+X)+PO, pos)
					self._late_messages.append(f"    body '{x1.name()}' {x1.tag}")
				pos += 1
				y0 = addUniqueBody(body, "PLA", ([None]+Y)+P,  pos)
				self._late_messages.append(f"    body '{y0.name()}' {y0.tag}")
				if body.tag == "BOX":
					pos += 1
					y1 = addUniqueBody(body, "PLA", ([None]+Y)+PO, pos)
					self._late_messages.append(f"    body '{y1.name()}' {y1.tag}")
				pos += 1
				z0 = addUniqueBody(body, "PLA", ([None]+Z)+P,  pos)
				self._late_messages.append(f"    body '{z0.name()}' {z0.tag}")
				pos += 1
				z1 = addUniqueBody(body, "PLA", ([None]+Z)+PO, pos)
				self._late_messages.append(f"    body '{z1.name()}' {z1.tag}")
				if body.tag == "WED":
					pos += 1
					xy = addUniqueBody(body, "PLA", ([None]+(Z^(Y-X)))+(P+X), pos)
					self._late_messages.append(f"    body '{xy.name()}' {xy.tag}")
					add2Replace(body, f"(+{z1.name()} -{z0.name()} "
							   f"-{x0.name()} -{y0.name()} "
							   f"+{xy.name()})")
				else:
					add2Replace(body, f"(+{x1.name()} -{x0.name()} "
							   f"+{y1.name()} -{y0.name()} "
							   f"+{z1.name()} -{z0.name()})")
			elif body.tag in ("RCC", "REC", "TRC"):
				P = body.bodyP()
				Z = body.bodyZ()
				PZ = P + Z
				dire = Z.direction(Input.zero)
				if dire=="X":
					p0 = addUniqueBody(body, "YZP", [None, P.x], pos,
									body.comment())
					p1 = addUniqueBody(body, "YZP", [None, PZ.x], pos+1)
				elif dire=="Y":
					p0 = addUniqueBody(body, "XZP", [None, P.y], pos,
									body.comment())
					p1 = addUniqueBody(body, "XZP", [None, PZ.y], pos+1)
				elif dire=="Z":
					p0 = addUniqueBody(body, "XYP", [None, P.z], pos,
									body.comment())
					p1 = addUniqueBody(body, "XYP", [None, PZ.z], pos+1)
				else:
					p0 = addUniqueBody(body, "PLA", ([None]+Z)+P, pos,
									body.comment())
					p1 = addUniqueBody(body, "PLA", ([None]+Z)+PZ, pos+1)
				if body.tag == "RCC":
					radius = body.numWhat(7)
					if dire=="X":
						c = addUniqueBody(body, "XCC",
							[None, P.y,P.z,radius], pos+2)
					elif dire=="Y":
						c = addUniqueBody(body, "YCC",
							[None, P.z,P.x,radius], pos+2)
					elif dire=="Z":
						c = addUniqueBody(body, "ZCC",
							[None, P.x,P.y,radius], pos+2)
					else:
						# convert card to "QUA" and then transform it
						c = addUniqueBody(body, "QUA",
							[None, 1, 1, 0, 0, 0, 0,
							-2*P.x, -2*P.y, -2*P.z,
							-radius**2 + P.length2()], pos+2)
						# FIXME to be done!!!!
						#c._transformQUA(card, matrix)
				elif body.tag == "REC":
					if dire=="X":
						# Not correct
						c = addUniqueBody(body, "XEC",
							[None, P.y,P.z,body.bodyX().length(),
							body.bodyY().length()],
							pos+2)
					elif dire=="Y":
						# Not correct
						c = addUniqueBody(body, "YEC",
							[None, P.z,P.y,body.bodyX().length(),
							body.bodyY().length()],
							pos+2)
					elif dire=="Z":
						# Not correct
						c = addUniqueBody(body, "ZEC",
							[None, P.x,P.y,body.bodyX().length(),
							body.bodyY().length()],
							pos+2)
				#elif body.tag == "TRC": FIXME
				self._late_messages.append(f"Body '{body.name()}' expanded to:")
				self._late_messages.append(f"    body '{p0.name()}' {p0.tag}")
				self._late_messages.append(f"    body '{p1.name()}' {p1.tag}")
				self._late_messages.append(f"    body '{c.name()}' {c.tag}")
				add2Replace(body, f"(+{p1.name()} -{p0.name()} +{c.name()})")
			else:
				bodies_count -= 1

		for region in self.project.input.cardsCache(Card.REGION,-1).values():
			region_expr = region.expr()
			changed = False
			for body,expr in replace.items():
				if body.search(region.expr()):
					region_expr = body.sub(expr,region_expr)
					regions_count += 1
					changed = True
			if changed:
				self._late_messages.append(f"Region '{region.name()}' bodies replaced")
				undoinfo.append(self.flair.setWhatUndo(region, -1, region_expr))

		self.flair.addUndolistRefresh(undoinfo, "Expand bodies")
		self.notBusy()
		if bodies_count:
			self.flair.notify("Expanded Bodies",
				f"{bodies_count} bodies expanded\n"
				f"and {regions_count} regions replaced",
				tkFlair.NOTIFY_NORMAL,
				"Output", self.flair.showOutput)
		else:
			self.flair.notify("Expanded Bodies",
				"No body expanded",
				tkFlair.NOTIFY_WARNING)

	# ----------------------------------------------------------------------
	# Correct bodies to make them orthogonal
	# ----------------------------------------------------------------------
	def makeOrthogonal(self):
		if self.editObject is None: return
		cards = self.getSelection(objtype=Card.BODY)
		if not cards: return
		self.busy()

		undoinfo = [(self.flair.refreshUndo,)]
		count = 0
		for body in cards:
			if body.tag in ("BOX","WED", "REC"):
				try:
					Xn,Yn = body.bodyZ().gramSchmidt(body.bodyX(), body.bodyY())
				except ZeroDivisionError:
					log.error(f"Invalid body parameters of {body.tag} {body.name()!r}")
					continue
				Xn = list(map(bmath.format, Xn))
				Yn = list(map(bmath.format, Yn))
				xi = 7 if body.tag == "REC" else 4
				undoinfo.append(self.flair.setWhatUndo(body, xi, Xn[0])); xi += 1
				undoinfo.append(self.flair.setWhatUndo(body, xi, Xn[1])); xi += 1
				undoinfo.append(self.flair.setWhatUndo(body, xi, Xn[2])); xi += 1

				undoinfo.append(self.flair.setWhatUndo(body, xi, Yn[0])); xi += 1
				undoinfo.append(self.flair.setWhatUndo(body, xi, Yn[1])); xi += 1
				undoinfo.append(self.flair.setWhatUndo(body, xi, Yn[2]))
				count += 1

		if len(undoinfo)>1:
			undoinfo.append(undoinfo[0])
			self.flair.addUndo(undoinfo,"Correct bodies for orthogonality")

		self.refresh()
		self.notBusy()
		if count:
			self.setStatus(f"{count} bodies corrected for orthogonality")
		else:
			self.setStatus("No body corrected for orthogonality")

	# ----------------------------------------------------------------------
	# expand parenthesis from regions
	# ----------------------------------------------------------------------
	def expandZones(self):
		if self.editObject is None: return
		regions = self.getSelection(objtype=Card.REGION)
		if not regions:
			self.setStatus("Please select a region to expand")
			return
		self.setStatus("Expanding zones...", "DarkBlue")
		self.setStatusPercent(0, update=True)
		# stop all threads, during the optimization
		self.viewers.stopThreads()
		self.busy()
		undoinfo = self.flair.undolistRefreshOnly()
		nexp  = 0
		nfail = 0
		ndel  = 0
		# process in reversed order since some REGIONS might have continuation
		# and they will be deleted, so it will screw up the input
		timer = tool.Timer()
		self.resetInterrupt()
		startTime = lastUpdate = time.time()
		for i, region in enumerate(reversed(regions)):
			changed = False
			zones = region[Card.ZONE]
			rid = region[Card.ID]
			# Check for bounding boxes
			if self.editObject.many:
				zonelist = range(len(zones)-1,-1,-1)
			else:
				zonelist = self.propListSelectedZones()
				if not zonelist:
					zonelist = range(len(zones)-1,-1,-1)
				else:
					zonelist = reversed(zonelist)
			for iz in zonelist:
				rc = self.geometry().region(rid,"expand",iz,
						GeometryEditor.EXPANSION_LIMIT)
				if rc:
					z = self.geometry().region(rid,"zone",iz)
					if z:
						#print("ZONES=",*z)
						zones[iz] = csg.toString(
								self.project.ismode((Mcnp,Phits)),
								*z)
						#print(">>>=",zones[iz])
						nz = zones[iz].count("|")+zones[iz].count(":")+1
						self._late_messages.append(
							f"Zone '{region.name()}:{iz+1}' "
							f"expanded to {nz} new zones")
						nexp += 1
					else:
						self._late_messages.append(
							f"Zone '{region.name()}:{iz+1}' deleted")
						del zones[iz]		# delete zone
						ndel += 1
					changed = True
				else:
					fmt,zone = self.geometry().region(rid,"zone",iz)
					if fmt==csg.RPN:
						self._late_messages.append(
							f"{log.WARNING} Zone '{region.name()}:{iz+1}'"
							"failed to expand")
						nfail += 1

				ti = time.time()
				if ti - startTime > 3.0:
					if self.interruptNotify():
						break
				if ti - lastUpdate > 0.25:
					self.flair.update()
					lastUpdate = ti
			if changed:
				undoinfo.append(self.flair.setZonesUndo(region,zones))
				# FIXME if REGION is empty
				# if not zones:
				#	maybe disable them?
			try:
				self.setStatusPercent((i*100) // len(regions), True)
			except tk.TclError:
				pass
			if self._interrupt: break

		self.resetInterrupt()

		self._late_messages.reverse()	# to show in correct order
		timer.stop()
		self._late_messages.append(
				f"Zones: Expanded={nexp}, Failed={nfail}, "
				f"Deleted={ndel} Time: {timer()} s")

		self.setStatusPercent(100, True)
		self.flair.addUndolistRefreshOnly(undoinfo,"Expanded zones")
		self.flair.refresh("redraw", self)
		if nexp or nfail:
			self.flair.notify("Zones expand",
				f"{nexp} zones expanded\n"
				f"{nfail} zones failed to expand\n"
				f"{ndel} zones deleted due to optimization",
				tkFlair.NOTIFY_NORMAL if nexp>0 else tkFlair.NOTIFY_WARNING,
				"Output", self.flair.showOutput)
		self.setStatus(f"{nexp} zones expanded", "DarkGreen")
		self.notBusy()
		self.setStatusPercent(0, True)

	# ----------------------------------------------------------------------
	# Remove zones with invalid bounding boxes
	# ----------------------------------------------------------------------
	def optimizeZones(self):
		if self.editObject is None: return
		regions = self.getSelection(objtype=Card.REGION)
		if not regions:
			self.setStatus("Please select a region/zones to optimize")
			return
		self.setStatus("Optimize zones...", "DarkBlue")
		self.setStatusPercent(0, update=True)
		# stop all threads, during the optimization
		self.viewers.stopThreads()
		self.busy()
		undoinfo = self.flair.undolistRefreshOnly()
		ndel = 0
		nopt = 0
		# process in reversed order since some REGIONS might have continuation
		# and they will be deleted, so it will screw up the input
		timer = tool.Timer()
		changed = False
		for i, region in enumerate(reversed(regions)):
			rid = region[Card.ID]
			if self.geometry().region(rid,"optimize"):
				changed = True
				zones = []
				for z in self.geometry().region(rid,"expr"):
					if z:
						zones.append(csg.toString(
								self.project.ismode((Mcnp,Phits)),
								*z))
				nopt += len(zones)
				ndel += len(region[Card.ZONE]) - len(zones)
				undoinfo.append(self.flair.setZonesUndo(region,zones))
			self.setStatusPercent((i*100) // len(regions), True)
		self._late_messages.reverse()	# to show in correct order
		timer.stop()
		self._late_messages.append(f"Zones: Optimized={nopt}, "
				f"Deleted={ndel} Time: {timer()} s")

		self.setStatusPercent(100, True)
		self.flair.addUndolistRefreshOnly(undoinfo, "Optimize zones")
		if changed:
			self.flair.refresh("redraw")
			self.flair.notify("Zones optimization",
				f"{nopt} zones optimized\n"
				f"{ndel} invalid zones deleted")
		self.setStatus(f"{ndel} invalid zones deleted", "DarkGreen")
		self.notBusy()
		self.setStatusPercent(0, True)

	# ----------------------------------------------------------------------
	# Convert multiple regions into a single multi-union
	# ----------------------------------------------------------------------
	def joinRegions(self):
		if self.editObject is None: return
		regions = self.getSelection(objtype=Card.REGION)
		if not regions or len(regions)==1:
			self.setStatus("Please select more than one regions to join")
			return
		self.busy()
		comments = []
		zones = []
		undoinfo = self.flair.undolistRefresh()
		delete = []
		for region in regions:
			if region.comment():
				comments.append(region.comment())
			zones.extend(region[Card.ZONE])
			if region is not regions[0]:
				delete.append(region.pos())
				assign = region[Card.ASSIGNMAT]
				if assign and assign.what(3)=="":
					delete.append(assign.pos())
		if comments:
			undoinfo.append(self.flair.setCommentUndo(regions[0], "\n".join(comments)))
		undoinfo.append(self.flair.setZonesUndo(regions[0], zones))
		self.flair.refresh("redraw", self)

		for pos in reversed(sorted(delete)):
			undoinfo.append(self.flair.delCardUndo(pos, False))

		self.flair.addUndolistRefresh(undoinfo, "Join regions")
		self.notBusy()
		self.setStatus(f"{len(regions)} regions joined into a single multi union region")

	# ----------------------------------------------------------------------
	def splineSortTime(self):
		if self.editObject is None: return
		count = 0
		undoinfo = self.flair.undolistRefreshOnly()
		for spline in self.getSelection(objtype=Card.OBJECT):
			if spline.tag != "spline": return
			undoinfo.append(self.flair.saveCardUndoInfo(spline.pos()))
			# sort time of spline as tuple
			# 1. first numerical (if it fails it will return 0)
			# 2. second the position
			# 3. third the actual string value to be saved for replacing later on
			times = [(spline.numProperty(f"node:{x}.t"), x, spline[f"node:{x}.t"])
					for x in range(1,spline["node:0"]+1)]
			for i,(num,_,value) in enumerate(sorted(times)):
				spline[f"node:{i+1}.t"] = value
			count += 1
		self.flair.addUndolistRefreshOnly(undoinfo, "Spline sort time")
		self.setStatus(f"{count} splines sorted in time")

	# ----------------------------------------------------------------------
	def stopAction(self):
		self.viewers.setActionStop()
		self.showContextGroup()

	# ----------------------------------------------------------------------
	def toggleSplitter(self, event=None):
		self.splitter.toggle()

	# ----------------------------------------------------------------------
	def help(self):
		Manual.show("","Geometry Editor")

#===============================================================================
# Return true if geoviewer is install and loaded
#===============================================================================
def installed():
	try:
		GeometryViewer.geoviewer.Viewer
		return True
	except AttributeError:
		return False

#-------------------------------------------------------------------------------
# Return true if PIL is installed
#-------------------------------------------------------------------------------
def PILInstalled():
	return GeometryViewer.PIL is not None #have_PIL

#-------------------------------------------------------------------------------
# Return geometry viewer errors during import
#-------------------------------------------------------------------------------
def geoviewer_error():
	return GeometryViewer.geoviewer_error

# -----------------------------------------------------------------------------
# Return geoviewer.so version
# -----------------------------------------------------------------------------
def version():
	try:
		return GeometryViewer.geoviewer.__version__
	except AttributeError:
		return None

# -----------------------------------------------------------------------------
# Check version against flair
# -----------------------------------------------------------------------------
def checkVersion(flair):
	try:
		v = version()
	except AttributeError:
		return
	if v and v != tkFlair.__version__:
		flair.write(f"{log.WARNING}flair - geoviewer. Version mismatch.\n")
		flair.write(f"{log.WARNINGCONT}flair     Version: {tkFlair.__version__}\n")
		flair.write(f"{log.WARNINGCONT}geoviewer Version: {v}\n")
		flair.write(f"{log.WARNINGCONT}Please use the same version for both packages\n")
		flair.notify("Version mismatch.",
			f"flair Version: {tkFlair.__version__}\n"
			f"geoviewer Version: {version()}\n"
			"Please use the same version for both packages",
			tkFlair.NOTIFY_WARNING)

#-------------------------------------------------------------------------------
# Read configuration file
#-------------------------------------------------------------------------------
def loadConfig(config):
	global VISIBLE_COLOR, LOCK_COLOR, FREEZE_COLOR, ERROR_COLOR, WIREFRAME_COLOR
	if GeometryViewer.geoviewer_error is None: return

	# === Colors ===
	section = tkFlair._COLOR_SECTION
	GeometryViewer._BACKGROUND_COLOR= config.getIColor(section,"geometry.background",
						GeometryViewer._BACKGROUND_COLOR)
	GeometryViewer._SELECT_COLOR    = config.getIColor(section,"geometry.select",
						GeometryViewer._SELECT_COLOR)
	GeometryViewer._VISIBLE_COLOR   = config.getIColor(section,"geometry.visible",
						GeometryViewer._VISIBLE_COLOR)
	GeometryViewer._LOCK_COLOR      = config.getIColor(section,"geometry.lock",
						GeometryViewer._LOCK_COLOR)
	GeometryViewer._FREEZE_COLOR    = config.getIColor(section,"geometry.freeze",
						GeometryViewer._FREEZE_COLOR)
	GeometryViewer._WIREFRAME_COLOR = config.getIColor(section,"geometry.wireframe",
						GeometryViewer._WIREFRAME_COLOR)
	GeometryViewer._BODY_BBIN_COLOR = config.getIColor(section,"geometry.bodybboxincolor",
						GeometryViewer._BODY_BBIN_COLOR)
	GeometryViewer._BODY_BBOUT_COLOR= config.getIColor(section,"geometry.bodybboxoutcolor",
						GeometryViewer._BODY_BBOUT_COLOR)
	GeometryViewer._ZONE_BB_COLOR   = config.getIColor(section,"geometry.zonebboxcolor",
						GeometryViewer._ZONE_BB_COLOR)
	GeometryViewer._REGION_BB_COLOR = config.getIColor(section,"geometry.regionbboxcolor",
						GeometryViewer._REGION_BB_COLOR)
	GeometryViewer._REGION_COLOR    = config.getIColor(section,"geometry.region",
						GeometryViewer._REGION_COLOR)
	GeometryViewer._ZONE_COLOR      = config.getIColor(section,"geometry.zone",
						GeometryViewer._ZONE_COLOR)
	GeometryViewer._ERROR_COLOR     = config.getIColor(section,"geometry.error",
						GeometryViewer._ERROR_COLOR)
	GeometryViewer._VERTEX_COLOR    = config.getIColor(section,"geometry.vertex",
						GeometryViewer._VERTEX_COLOR)
	GeometryViewer._LATTICE_COLOR   = config.getIColor(section,"geometry.lattice",
						GeometryViewer._LATTICE_COLOR)
	#GeometryViewer._REGERROR_COLOR  = config.getIColor(section,"geometry.region.error",
	#					GeometryViewer._REGERROR_COLOR)
	GeometryViewer._TITLE_COLOR     = config.getIColor(section,"geometry.title",
						GeometryViewer._TITLE_COLOR)
	GeometryViewer._GRIDTEXT_COLOR  = config.getIColor(section,"geometry.gridtext",
						GeometryViewer._GRIDTEXT_COLOR)
	GeometryViewer._PALETTETEXT_COLOR= config.getIColor(section,"geometry.palettetext",
						GeometryViewer._PALETTETEXT_COLOR)

	VISIBLE_COLOR   = f"#{GeometryViewer._VISIBLE_COLOR:06X}"
	LOCK_COLOR      = f"#{GeometryViewer._LOCK_COLOR:06X}"
	FREEZE_COLOR    = f"#{GeometryViewer._FREEZE_COLOR:06X}"
	WIREFRAME_COLOR = f"#{GeometryViewer._WIREFRAME_COLOR:06X}"
	ERROR_COLOR     = f"#{GeometryViewer._ERROR_COLOR:06X}"

	# === Geometry ===
	section = tkFlair._GEOMETRY_SECTION
	GeometryViewer.LAPTOP       = config.getBool(section, "laptop",      GeometryViewer.LAPTOP)
	GeometryViewer.INVERTWHEEL  = config.getBool(section, "invertwheel", GeometryViewer.INVERTWHEEL)
	GeometryViewer.PINZOOM      = config.getBool(section, "pinzoom",     GeometryViewer.PINZOOM)
	GeometryViewer.CORES        = config.getInt( section, "cores",       GeometryViewer.CORES)
	GeometryViewer.SNAPDISTANCE = config.getInt( section, "snap",        GeometryViewer.SNAPDISTANCE)
	GeometryViewer.CURSORSIZE   = config.getInt( section, "cursorsize",  GeometryViewer.CURSORSIZE)
	GeometryEditor.EXPANSION_LIMIT = config.getInt(section, "explimit",  GeometryEditor.EXPANSION_LIMIT)
	GeometryViewer.HIFONT       = config.getStr( section, "hifont",      GeometryViewer.HIFONT)

	# === Orientation ===
	section = tkFlair._ORIENTATION_SECTION
	Project.GeometryInfo.ORIENTATION = config.getStr(section, "default",
					Project.GeometryInfo.ORIENTATION)

	#-----------------------------------------------------------------------
	# Read orientation system from config
	# @return False on error
	#-----------------------------------------------------------------------
	def getOrientationSystem(i):
		try:
			name = config.get(section, f"name.{i}")
		except:
			return False
		system = {}
		for axis in GeometryViewer.AXES_ORDER:
			item = f"face.{i}.{axis}"
			face = config.getStr(section, item)
			system[axis] = face
		for view in VIEWER.NAME:
			item = f"view.{i}.{view}"
			value = config.getStr(section, item)
			values = []
			for x in value.split(","):
				try:
					x = int(x)
				except ValueError:
					x = x.strip()
				values.append(x)
			system[view] = values
		GeometryViewer.ORIENTATIONS[name] = system
		return True

	# System based
	for i in range(100,200):
		if not getOrientationSystem(i): break
	# User based
	for i in range(100):
		if not getOrientationSystem(i): break

	for item,value in config.items(section):
		if item.startswith("alias."):
			name = item[6:]
			GeometryViewer.ORIENTATIONS[name] = GeometryViewer.ORIENTATIONS[value]

#-------------------------------------------------------------------------------
def saveConfig(config):
	section = tkFlair._ORIENTATION_SECTION
	config.setStr(section, "default", Project.GeometryInfo.ORIENTATION)
