#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   13-Jun-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import math
import time
import traceback
import subprocess
import tkinter as tk

try:
	import PIL.Image
	import PIL.ImageTk	# Check if ImageTk exists to be used by ZoomImage
except ImportError:
	PIL = None

import csg
import log
import tool
import bmath
import icons
import Global
import Manual
import Project
import tkFlair
import tkExtra
import tkPalette
import GeometryBody
import GeometryTool
import CalibrateImage
import GeometryLayers
import GeometryToolbar

from tkconst    import STATE
from Global     import dump, BIT, DRAW, ACTION, VIEWER, SELECT
from deck       import Input, Card, CardInfo
from fluka      import Fluka, FlukaCard, FlukaParticle, FlukaLayout
from mcnp       import Mcnp
from moira      import Moira
from phits      import Phits
from penelope   import Penelope
from palette    import paletteStore, rndColor
from postscript import Postscript
from timeframe  import TimeFrameCanvas

try:
	import geoviewer
	geoviewer_error = ""
except ImportError as ex:
	geoviewer_error = str(ex)

# General variables
DEFZOOM    = 4.0
MAXHISTORY = 20			# Save last views
MAXORBIT   = 100.0		# maximum orbit depth * window dimensions
MAXZOOM    = 50.0
MOVESTEP   = 10.0
MOVESTEP2  = 100.0
ROTSTEP    = math.pi / 36.0	#  5deg
ROTSTEP90  = math.pi /  2.0	# 90deg
SZOOM      = 1.02

# states and size of orientation cube
OCUBE_SMALL    = 16		# Defined in Project Options.ocube
OCUBE_NSTEPS   =  5
OCUBE_STEP     =  4
OCUBE_ZOOMIN   =  1
OCUBE_ZOOMOUT  = -1
OCUBE_ANIMATE  = True

OCUBE_NOWHERE  =  0		# mouse click events/actions
OCUBE_ONAXIS   =  1
OCUBE_ONCUBE   =  2
OCUBE_ONHOME   =  3
OCUBE_ONORIGIN =  4
OCUBE_ONCW     =  5
OCUBE_ONCCW    =  6
OCUBE_ONLEFT   =  7
OCUBE_ONRIGHT  =  8

# animate timer it is better if all are multiples
# of the smallest timer
DRAW_TIMER    = 100
OCUBE_TIMER   =  25
PAN_TIMER     =  25
PROJECT_TIMER = 250

# Friction to apply during panning slow down
FRICTION = math.pow(1.0-0.0035, float(PAN_TIMER))
ZOOM      = 1.20		# Zoom In/Out factor

# XXX it should be at sync with the viewer/viewer.h enum class Projection : int
PROJECTION_NOTHING    =  0	# No projection information
PROJECTION_SPAWN      =  1
PROJECTION_START      =  2	# start projection (after stop)
PROJECTION_CONICS     =  3	# Transform bodies and prepare conics
PROJECTION_LOCATION   =  4	# Calculate the bodies&region location
PROJECTION_INTERSECT  =  5	# Intersect all bodies
PROJECTION_SCAN       =  6	# Scan segments
PROJECTION_DRAW       =  7	# Late drawing
PROJECTION_FINISHED   = 10	# Projection ended

# Viewport Matrix, names points the direction we are looking
UNIT_MATRIX   = bmath.Matrix(4,type=1)
BACK_MATRIX   = UNIT_MATRIX
FRONT_MATRIX  = bmath.Matrix.rotY(math.pi)
DOWN_MATRIX   = bmath.Matrix([	[0.0, 1.0, 0.0, 0.0],
				[0.0, 0.0, 1.0, 0.0],
				[1.0, 0.0, 0.0, 0.0],
				[0.0, 0.0, 0.0, 1.0]])
TOP_MATRIX90  = bmath.Matrix([	[1.0, 0.0, 0.0, 0.0],
				[0.0, 0.0,-1.0, 0.0],
				[0.0, 1.0, 0.0, 0.0],
				[0.0, 0.0, 0.0, 1.0]])
# Synchronize with the GeometryToolbar orientISO
UP_MATRIX     = DOWN_MATRIX * bmath.Matrix.rotY(math.pi)
RIGHT_MATRIX  = bmath.Matrix.rotY(math.pi/2.0)
LEFT_MATRIX   = bmath.Matrix.rotY(-math.pi/2.0)

# Orientation systems
AXES_ORDER   = [ "Z-", "Z+", "Y-", "Y+", "X-", "X+" ]	# Should be the same with ocube.cc
AXES_VECTORS = {
		"X-" : bmath.Vector(-1,0,0),
		"X+" : bmath.Vector( 1,0,0),
		"Y-" : bmath.Vector(0,-1,0),
		"Y+" : bmath.Vector(0, 1,0),
		"Z-" : bmath.Vector(0,0,-1),
		"Z+" : bmath.Vector(0,0, 1),
	}
ORIENTATIONS = {
	"flair" : {
			"X-" : "Right",
			"X+" : "Left",
			"Y-" : "Down",
			"Y+" : "Up",
			"Z-" : "Back",
			"Z+" : "Front",
			"Red"     : ("Media", True,  "Z+", "X+"),	# looking down
			"Green"   : ("3D",    False, 30,30,15),		# Rx:30, Ry:30, Rz:15
			"Blue"    : ("Media", True,  "Z+", "Y+"),	# looking left
			"Magenta" : ("Media", True,  "X+", "Y+"),	# looking back
		}
	}

DEF_CURSOR = "left_ptr"
MOUSE_CURSOR = {
	ACTION.SELECT          : DEF_CURSOR,
	ACTION.SELECT_REGION   : DEF_CURSOR,
	ACTION.SELECT_ZONE     : DEF_CURSOR,
	ACTION.SELECT_BODY     : DEF_CURSOR,
	ACTION.SELECT_OBJECT   : DEF_CURSOR,
	ACTION.SELECT_NODE     : DEF_CURSOR,
	ACTION.SELECT_SPLINE   : DEF_CURSOR,

	ACTION.AREA_SELECT     : "right_ptr",
	ACTION.ZONE            : "center_ptr",
	ACTION.ADD_ZONEPAINT   : "spraycan",		# "pencil"
	ACTION.PEN             : "pencil",
	#ACTION.PAINT           : "spraycan",
	ACTION.INFO            : "question_arrow",	 #"tcross",
	ACTION.CENTER          : "target",

	ACTION.PAN             : "fleur",
	ACTION.ORBIT           : "exchange",
	ACTION.PIVOT           : "target",
	ACTION.ZOOM_IN         : "sizing",
	ACTION.ZOOM_OUT        : "sizing",
	ACTION.ZOOM_ON         : "sizing",

	ACTION.VIEW_ORIGIN     : "cross",
	ACTION.VIEW_MOVE       : "fleur",
	ACTION.VIEW_ROTATE     : "exchange",

	ACTION.ADD             : "tcross",
	ACTION.ADD_NEXT        : "tcross",

	ACTION.ADD_SPLINE_NODE : "tcross",
	ACTION.DEL_SPLINE_NODE : "X_cursor",

	ACTION.MOVE            : "hand1",
	ACTION.MOVE_KEYBOARD   : "hand1",
	ACTION.ROTATE          : "exchange",
	ACTION.REPEAT          : "trek",

	ACTION.EDIT            : "pencil",
}

ACTION_MESSAGE = {
	ACTION.PAN             : "Pan viewport",
	ACTION.ORBIT           : "Orbit viewport with a trackball",
	ACTION.PIVOT           : "Set orbit pivot point and pan/center viewport to this location",
	ACTION.ZOOM_IN         : "Zoom in",
	ACTION.ZOOM_OUT        : "Zoom out",
	ACTION.ZOOM_ON         : "Zoom on object",
	ACTION.CENTER          : "Center viewport to location",
	ACTION.ADD_NEXT        : "Add next node",
	ACTION.ADD_ZONE        : "Add/Edit zone",
	ACTION.ADD_ZONEPAINT   : "Add multiple zones by painting",
	ACTION.ADD_SPLINE_NODE : "Add spline nodes",
	ACTION.DEL_SPLINE_NODE : "Delete spline nodes",
	ACTION.EDIT            : "Edit/Correct overlaps",
	ACTION.MOVE            : "Move selected objects",
	ACTION.MOVE_KEYBOARD   : "Move selected objects",
	ACTION.ROTATE          : "Rotate selected objects",
	ACTION.REPEAT          : "Repeat and distribute selected objects",
	ACTION.ZONE            : "Select zone master action",
	ACTION.PEN             : "Draw using pen",
	#ACTION.PAINT           : "Paint region properties",
	ACTION.INFO            : "Display info at cursor",
	ACTION.VOLUME          : "Display volume calculator tool",
	ACTION.EXPORT          : "Export selected viewport",
	ACTION.MOVIE           : "Create movie from selected viewport",
	ACTION.PLANNER         : "Dose planner reporting tool",
	ACTION.LINK            : "Link node with spline",
}

# Viewers
VIEWFORMAT = {	VIEWER.NAME[0] : (1.0, 1.0, 0.5),
		VIEWER.NAME[1] : (1.0, 0.0, 0.5),
		VIEWER.NAME[2] : (0.0, 0.5, 1.0),
		VIEWER.NAME[3] : (0.0, 0.5, 0.0),
		"2x2"          : (0.5, 0.5, 0.5),
		"2x1"          : (1.0, 0.5, 0.5),
		"1x2"          : (0.5, 1.0, 1.0),
		"1+2"          : (0.5, 1.0, 0.5),
		"2+1"          : (0.5, 0.5, 1.0) }
VIEWORDER = VIEWER.NAME[:]
VIEWORDER.extend(["2x2", "2x1", "1x2", "1+2", "2+1", "custom"])

# Color definitions
_BACKGROUND_COLOR = 0x707070	# geometry.background
_BODY_BBIN_COLOR  = 0xE6E600	# geometry.bodyBBoxInColor
_BODY_BBOUT_COLOR = 0x00FF00	# geometry.bodyBBoxOutColor
_ERROR_COLOR      = 0xFF0000	# geometry.error
_FREEZE_COLOR     = 0x00FFFF	# geometry.freeze
_LABEL_COLOR      = 0xA05010	# geometry.label
_LATTICE_COLOR    = 0xFFFFFF	# geometry.lattice
_LOCK_COLOR       = 0xFF8033	# geometry.lock
_REGION_BB_COLOR  = 0xFF8000	# geometry.regionBBoxColor
_REGION_COLOR     = 0x000000	# geometry.region
_SELECT_COLOR     = 0xFF00FF	# geometry.select
_VERTEX_COLOR     = 0xFFFF00	# geometry.vertex
_VISIBLE_COLOR    = 0x7F007F	# geometry.visible
_WIREFRAME_COLOR  = 0x007070	# geometry.wireframe
_ZONE_BB_COLOR    = 0x00FFFF	# geometry.zoneBBoxColor
_ZONE_COLOR       = 0x606060	# geometry.zone

#_REGERROR_COLOR  = 0xFFB0B0	# geometry.region.error

_TITLE_COLOR      = 0xD0D000	# geometry.title
_GRIDTEXT_COLOR   = 0xFFE090	# geometry.gridlabel
_PALETTETEXT_COLOR= 0x000000	# palette text color
#_COORD_COLOR      = # geometry.coordinate

# Configuration variables
TRACKBALLSIZE  = 0.8
CORES          = 0		# CPU/cores to use per viewport
LAPTOP         = False		# Laptop mode swap button-2 and button-3
INVERTWHEEL    = False		# Invert zooming with wheel
PINZOOM        = False		# Zoom in as google, center on mouse location
SNAPDISTANCE   = 4		# Snapping distance in pixels
CURSORSIZE     = 40		# Grab/Rotation cursor size
CROSSHAIR      = 10		# Cross hair size
CLOSE_DISTANCE = 4.0		# Close distance to search
HIFONT         = ""		# orientation system font to use, especially for HiDPI

GOBJECT_NODE_NONE = -1000000	# XXX must be the same as geo.h::NodeId_NONE
GOBJECT_NODE_X    = -1000001	# X-axis -//- NodeId_axisX
GOBJECT_NODE_Y    = -1000002	# Y-axis -//- NodeId_axisY
GOBJECT_NODE_Z    = -1000003	# Z-axis -//- NodeId_axisZ
GOBJECT_NODE_R    = -1000004	# R-axis -//- NodeId_axisR

# ------------------------------------------------------------------------------
# Variables to retrieve when manipulating an object
_OBJECT_MOVE_VARS = {
	"arrow" : (("x",   1), ("y",   2), ("z",   3),
		   ("dx",  7), ("dy",  8), ("dz",  9)),
	"camera": (("x",   1), ("y",   2), ("z",   3),
		   ("atx", 7), ("aty", 8), ("atz", 9),
		   ("upx",10), ("upy",11), ("upz",12),
		   ("focal",13)),
	"light" : (("x",   1), ("y",   2), ("z",   3),
		   ("dx",  7), ("dy",  8), ("dz",  9)),
	"mesh"  : (("x",   1), ("y",   2), ("z",   3)),
	"ruler" : (("x",   1), ("y",   2), ("z",   3),
		   ("xe",  7), ("ye",  8), ("ze",  9),
		   ("xa", 10), ("ya", 11), ("za", 12)),
	"point" : (("x",   1), ("y",   2), ("z",   3)),
	"spline": (("x",   1), ("y",   2), ("z",   3)),
}

# Additional sequence of items to manipulate on ACTION.ADD_NEXT
_ADD_SEQUENCE = {
	"arrow"  : ( 2,),
	"camera" : ( 2, 3),
	"ruler"  : ( 2,),
	"light"  : ( 2,),
	#"ARB"   : (),
	"BOX"    : (10, 4),
	"C/X"    : ( 1,),
	"C/Y"    : ( 1,),
	"C/Z"    : ( 1,),
	"CX"     : ( 1,),
	"CY"     : ( 1,),
	"CZ"     : ( 1,),
	"ELL"    : (10, 2),
	#"GQ"    : (),			# only position
	#"HEX"   : (),
	"K/X"    : ( 1,),
	"K/Y"    : ( 1,),
	"K/Z"    : ( 1,),
	"KX"     : ( 1,),
	"KY"     : ( 1,),
	"KZ"     : ( 1,),
	"P3"     : (12,),
	"P"      : (GOBJECT_NODE_R,),
	"PLA"    : (GOBJECT_NODE_R,),
	#"PX"    : (),			# only position
	#"PY"    : (),			# only position
	#"PZ"    : (),			# only position
	#"QUA"   : (),			# only position
	"RAW"    : (11,15),
	"RCC"    : (10, 1),
	"REC"    : (10, 1),
	#"RHP"   : (),
	"RPP"    : (18, 6),
	"S"      : ( 1,),
	"SO"     : ( 1,),
	"SPH"    : ( 1,),
	#"SQ"     : (),
	"SX"     : ( 1,),
	"SY"     : ( 1,),
	"SZ"     : ( 1,),
	"TRC"    : (10, 4, 1),
	#"TRX"   : (),
	#"TRY"   : (),
	#"TRZ"   : (),
	#"TX"    : (),
	#"TY"    : (),
	#"TZ"    : (),
	"WED"    : (11,15),
	#"X"     : (),
	"XCC"    : ( 1,),
	"XEC"    : ( 1, 2),
	#"XYP"   : (),			# only position
	#"XZP"   : (),			# only position
	#"Y"     : (),
	"YCC"    : ( 1,),
	"YEC"    : ( 1, 2),
	#"YZP"   : (),			# only position
	#"Z"     : (),
	"ZCC"    : ( 1,),
	"ZEC"    : ( 1, 2),
}

# ------------------------------------------------------------------------------
def mouseCursor(action):
	return MOUSE_CURSOR.get(action, DEF_CURSOR)

# ------------------------------------------------------------------------------
# Convert EMFCUT what1 to kinetic energy
# ------------------------------------------------------------------------------
def _emfwhat1(x):
	if x<0.0: return abs(x)
	return max(0.0,x-0.510998928E-03)

# ------------------------------------------------------------------------------
# return a signed rotdefi index
# ------------------------------------------------------------------------------
def signedRotdefi(rotdefi, idx):
	if idx and idx[0]=='-':
		return -rotdefi.get(idx[1:],0)
	else:
		return rotdefi.get(idx,0)

# ------------------------------------------------------------------------------
# Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet
# if we are away from the center of the sphere.
# ------------------------------------------------------------------------------
def projection2Sphere(r, x, y):
	d = math.sqrt(x**2 + y**2)
	if d < r:				# Inside sphere
		return math.sqrt(r**2 - d**2)
	else:					# On hyperbola
		#t = r / 1.41421356237309504880
		#return t*t / d
		return 0.0

# ------------------------------------------------------------------------------
# Ok, simulate a track-ball.  Project the points onto the virtual
# trackball, then figure out the axis of rotation, which is the cross
# product of P1 P2 and O P1 (O is the center of the ball, 0,0,0)
# Note:  This is a deformed trackball-- is a trackball in the center,
# but is deformed into a hyperbolic sheet of rotation away from the
# center.  This particular function was chosen after trying out
# several variations.
#
# It is assumed that the arguments to this routine are in the range [-1..1]
# ------------------------------------------------------------------------------
def trackball(p1x, p1y, p2x, p2y, axis):
	if abs(p1x-p2x)<=1.0e-5 and abs(p1y-p2y)<=1.0e-5:
		# Zero rotation
		return None

	# First, figure out z-coordinates for projection of P1 and P2 to
	# deformed sphere
	p1 = bmath.Vector(p1x, p1y, projection2Sphere(TRACKBALLSIZE, p1x, p1y))
	p2 = bmath.Vector(p2x, p2y, projection2Sphere(TRACKBALLSIZE, p2x, p2y))

	if axis is not None:
		# find projection of p1,p2 on the perpendicular plane to axis (passing from 0)
		pp1 = p1 - (axis*p1)*axis
		pp2 = p2 - (axis*p2)*axis

		# Then the cross product should be parallel to axis
		# and the length to the sin()
		pp1.norm()
		pp2.norm()
		d = pp1 - pp2

		# Figure out how much to rotate around that axis.
		axis = pp2 ^ pp1
		l = axis.norm()
		if abs(l)<0.0001: axis = bmath.Vector.Z
		t = d.length() / 2.0

	else:
		axis = p2 ^ p1		# rotation axis
		l = axis.norm()
		if abs(l)<0.0001: axis = bmath.Vector.Z

		# Figure out how much to rotate around that axis.
		d = p1-p2
		t = d.length() / (2.0*TRACKBALLSIZE)

	# Avoid problems with out-of-control values...
	if   t >  1.0:	t = 1.0
	elif t < -1.0:	t = -1.0
	phi = 2.0 * math.asin(t)		# how much we rotate on this axis
	return bmath.Quaternion(axis, phi)	# using quaternions

#===============================================================================
# Geometry viewer frame
#===============================================================================
class GeometryViewer(tk.Frame):
	def __init__(self, master, parent, **kw):
		"""
		Initialize GeometryViewer
		master is the GeometryFrame
		parent is the Geometry4Frame (grandparent)
		"""
		super().__init__(master, **kw)
		self._viewer   = geoviewer.Viewer(parent._geometry, self._w, self.tk.interpaddr())
		self._viewer.set("cores", CORES)
		self._viewer.set("bordercolor", 0xFFFF00)
		self._viewer.set("dpiscale", tkFlair.dpiscale)
		if tkFlair.dpiscale>1:
			self._viewer.set("xfont", HIFONT)
		self.parent    = parent
		self.frame     = master
		self.editor    = parent.editor
		self._cursor   = DEF_CURSOR
		self._geometry = parent._geometry
		self.config(background="", cursor=self._cursor)

		self.bindings()
		self.viewports  = []		# other viewports
		self.camera     = None		# position linked to camera

		self.action     = ACTION.SELECT	# current action
		self.selected   = None		# item selected
		self.curHistory = -1		# current history pointer
		self.history    = []		# location history
		self._save      = None		# General purpose variable
		self.defSystem  = bmath.Matrix(type=1)	# default orientation system
		self._sync      = False		# Synchronize viewport if true
		self.defSync    = False		# Default value of synchronize for def system

		self.__node       = 0		# During adding option to modify
		self.x            = 0		# starting position of the last action or segment
		self.y            = 0		#             -//-
		self._x           = 0		# Mouse position when clicked
		self._y           = 0		#             -//-
		self.__x          = 0		# starting position during adding
		self.__y          = 0		#             -//-
		self._orbitDepth  = 0.		# distance of the last centered <B2> object for orbit
		self._msg         = ""
		self._clone       = False	# Clone or Move (Control was pressed?)
		self._ocubeAction = OCUBE_ZOOMOUT
		self.ocubeSet(OCUBE_SMALL)

		self.resetAnchor()
		if self._viewer.get("depth")<24:
			self.flair().notify("Insufficient Display color depth",
				"Display color depth is %d bits while 24 bits required"
				% (self._viewer.get("depth")),
				tkFlair.NOTIFY_ERROR)

		# Animations variables
		self.__animateId  = None	# after id
		self._animateTime = 0		# animate timer in ms
		self._animate     = {}		# animate in background item True/False
		self.__drawafter  = None	# draw on idle
		self.__drawEndCmd = None	# tuple (command,args...) to execute at end of draw

	# ----------------------------------------------------------------------
	def bindings(self):
		self.bind("<Enter>",		self.enter)
		self.bind("<Leave>",		self.leave)
		self.bind("<Expose>",		self._viewer.expose)
		self.bind("<Configure>",	self.configureEvent)
		#self.bind("<Visibility>",	self.visibility)
		#self.bind("<Property>",	self.property)
		self.bind("<<Project>>",	self._project)
		#self.bind("<<EndProjection>>",	self.endProjection)
		#self.bind("<<EndDraw>>",	self.endDraw)

		self.bind("<Motion>",		self.motion)

		self.bind("<Button-1>",		self.button1)
		self.bind("<Button-2>",		self.button2)
		self.bind("<Button-3>",		self.button3)

		self.bind("<Double-1>",		self.double1)

		self.bind("<B1-Motion>",	self.buttonMotion)
		self.bind("<B2-Motion>",	self.buttonMotion)
		self.bind("<B3-Motion>",	self.buttonMotion)

		self.bind("<ButtonRelease-1>",	self.buttonRelease)
		self.bind("<ButtonRelease-2>",	self.buttonRelease)
		self.bind("<ButtonRelease-3>",	self.buttonRelease)

		if INVERTWHEEL:
			self.bind("<Button-4>",	self.cursorOrZoomOut)
			self.bind("<Button-5>",	self.cursorOrZoomIn)
		else:
			self.bind("<Button-4>",	self.cursorOrZoomIn)
			self.bind("<Button-5>",	self.cursorOrZoomOut)
		self.bind("<MouseWheel>",	self.wheel)

		self.bind("<Control-Button-4>",	self.panBack)
		self.bind("<Control-Button-5>",	self.panFront)

		self.bind("<KeyPress-Control_R>",   self.controlPress )
		self.bind("<KeyPress-Control_L>",   self.controlPress )
		self.bind("<KeyRelease-Control_R>", self.controlRelease )
		self.bind("<KeyRelease-Control_L>", self.controlRelease )

		self.bind("<Key>",		self.handleKey)

		# Some global bindings
		self.bind("<Control-Key-l>",	self.frame.layersDialog)
		self.bind("<Control-Key-m>",	self.editor.gotoMaterial)
		self.bind("<Control-Key-t>",	self.editor.transformGeometry)
		self.bind("<<PasteSpecial>>",	self.editor.pasteSpecial)
		self.bind("<<Splitter>>",	self.editor.toggleSplitter)
		self.bind("<<Hardcopy>>",	self.export)	# Ctrl-P
		self.bind("<Control-Key-n>",	self.toNotes)
		if Global.developer:
			self.bind("<Control-Key-b>",	self.parent.autoBodyDialog)
			self.bind("<Control-Key-M>",	self.editor.mesh2Bodies)

	# ----------------------------------------------------------------------
	def flair(self):	return self.editor.flair
	def project(self):	return self.parent.project
	def input(self):	return self.parent.project.input
	def name(self):		return self.frame.name
	def index(self):	return self.parent.frames.index(self.frame)
	def log(self, msg):	return self.parent.log(msg)

	# ----------------------------------------------------------------------
	def addViewport(self, viewer, color):
		idx = len(self.viewports)
		self._viewer.viewport(idx, "viewer", viewer._viewer)
		self._viewer.viewport(idx, "color",  color)
		self.viewports.append(viewer)

	# ----------------------------------------------------------------------
	def set(self, var, value=None):
		if value is None:
			return self._viewer.set(var)
		else:
			self._viewer.set(var,value)

	# ----------------------------------------------------------------------
	# Mouse entering viewport
	# Used for setting the focus and check for VIEWPORT selection
	# ----------------------------------------------------------------------
	def enter(self, event):
		if not self.editor.isactive(): return
		# Steal keyboard focus when mouse enters only if window has the
		try: focus = self.focus_get()
		except: return
		if focus is None: return

		# Avoid stealing the focus from the Projection Dialog
		if isinstance(focus, tk.Entry): return

		# Steal the focus for all other widgets withing the same toplevel
		if focus.winfo_toplevel() == self.winfo_toplevel():	# for V2
			self.focus_set()

		if self.parent.mouseAction == ACTION.SELECT_VIEWPORT:
			self._viewer.set("borderwidth",10)
			self._viewer.expose()

	# ----------------------------------------------------------------------
	# Leave event from window
	#   WARNING <Enter>,<Leave> events are generated also on mouse clicks
	#           but with a non zero state
	# ----------------------------------------------------------------------
	def leave(self, event):
		#FIXME find the correct state to eliminate ButtonPress
		#dump("*-* GeometryViewer.leave", bin(event.state), self._ocubeAction, level=3)
		if event.state<STATE.BUTTON1 and \
		   OCUBE_ANIMATE and \
		   self._ocubeAction == OCUBE_ZOOMIN:
			self.ocubeZoomOut()

		if self.parent.mouseAction == ACTION.SELECT_VIEWPORT:
			self._viewer.set("borderwidth",0)
			self._viewer.expose()

	# ----------------------------------------------------------------------
	def expose(self, event=None):
		#if self.name()=="Red":
		#	print("expose", self.name())
		#	import traceback; traceback.print_stack()
		self._viewer.expose()

	# ----------------------------------------------------------------------
	def configureEvent(self, event):
		#if self.name()=="Red": print("<Configure>", event)
		#w,h = self._viewer.size()
		#expose = (event.width<=w) and (event.height<=h)
		self._viewer.configure(event.width, event.height)
		self.drawFast3D()
		self.draw(late=True)

	# ----------------------------------------------------------------------
	#def visibility(self, event):
		#dump("*-* visibility", self.name())

	# ----------------------------------------------------------------------
	#def property(self, event):
		#dump("*-* property", self.title())

	# ----------------------------------------------------------------------
	def _project(self, event=None):
		self.draw(True)

	# ----------------------------------------------------------------------
	# Detect what type of drawing/projection is needed
	# @param invalid	force projection
	# @param late		spawn a drawing instead of synchronous
	# @param history	append to history if position changed
	# ----------------------------------------------------------------------
	def draw(self, invalid=False, late=False, history=True):
		#dump("*-* draw", self.name(), "invalid=",invalid,
		#	"late=",late, "viewer.invalid=",self._viewer.invalid(), level=3)
		# Kill any background animation request to draw
		self._animate["draw"] = 0
		if invalid or self._viewer.invalid() or self._viewer.zoom()>MAXZOOM:
			self._viewer.calcWindow(DEFZOOM)
			self._viewer.viewOto0()		# reset offset
			if history and not self._animate.get("pan"):
				self.historyAdd()

			# Viewport is invalid start projection
			self._animate["project"] = 0	# FIXME do I need that?
			self._viewer.project(True)
			self.animate("project", PROJECT_TIMER)

			# Set as grid label the closest axis match
			# Base the name on the direction of the w
			axes = self._viewer.grid("axes")
			self._viewer.grid("label", axes)

		self._viewer.draw(False, -1, True)
		if self._viewer.projectionChanged():
			for v in self.viewports:
				v.expose()
			self._viewer.projectionChanged(False)
		if not self._animate.get("project",0):	# do not expose during projection yet
			#if self.name()=="Red":
			#	import traceback; traceback.print_stack()
			#	print("expose draw1", self.name())
			self._viewer.expose()
			if late: # otherwise it will be called from EndProjection
				self.drawOnIdle()
		self.frame.fillOrigin()

	# ----------------------------------------------------------------------
	# Handler of the <<EndProjection>> message send from the Viewer()
	# ----------------------------------------------------------------------
	#def endProjection(self, event=None):
	#	dump("*-* endProjection", self.name(), level=2)
	#	if self.__projId is not None:
	#		self.after_cancel(self.__projId)
	#		self.__projId = None
	#	self.projectAfter()

	# ----------------------------------------------------------------------
	# Force redraw when idle
	# ----------------------------------------------------------------------
	def drawOnIdle(self):
		dump("*-* drawOnIdle", self.name(), level=3)
		if self.__drawafter is not None:
			self.after_cancel(self.__drawafter)
		self.__drawafter = self.after_idle(self.startDraw)

	# ----------------------------------------------------------------------
	# Span a drawing thread and start animate to visualize the drawing
	# at periodic intervals
	# ----------------------------------------------------------------------
	def startDraw(self):
		dump("*-* startDraw", self.name(), self._viewer.state(), level=3)
		#if self.name()=="Green": traceback.print_stack()
		if self.__drawafter is not None:
			self.after_cancel(self.__drawafter)
			self.__drawafter = None
		self._animate["draw"] = 0		# to avoid an additional expose() from animate
		self._viewer.draw(True)			# Spawn draw
		self.animate("draw", DRAW_TIMER)	# animate, but do not expose

	# ----------------------------------------------------------------------
	# Fast drawing only if 3D is active
	# ----------------------------------------------------------------------
	def drawFast3D(self):
		if not self._viewer.get("3D"): return False
		dump("*-* drawFast3D.draw", self.name(), "_viewer.draw(False,DRAW.FAST,True)", level=3)
		self._viewer.draw(False, DRAW.FAST, True)
		self._viewer.expose()
		return True

	# ----------------------------------------------------------------------
	# <<EndDraw>> End of drawing message has been received from GeometryViewer
	# ----------------------------------------------------------------------
	def endDraw(self, event=None):
		dump("*-* endDraw", self.name(), self.__drawEndCmd, level=3)
		#if self.name()=="Green": traceback.print_stack()
		self._animate["draw"] = 0
		self.check4Errors()
		if self.__drawEndCmd is not None:
			cmd,*args = self.__drawEndCmd
			self.__drawEndCmd = None
			cmd(*args)

	# ----------------------------------------------------------------------
	# Stop all pending threads
	# ----------------------------------------------------------------------
	def stop(self):
		dump("*-* stop", self.name(), level=3)
		if self.__drawEndCmd is not None:
			self.flair().notify("Stopped",
				"Background jobs are stopped",
				tkFlair.NOTIFY_WARNING)
			self.__drawEndCmd = None
		self._animate["project"] = 0
		self._animate["draw"]    = 0
		self._viewer.stopThread()

	# ----------------------------------------------------------------------
	# Main animation loop
	# ----------------------------------------------------------------------
	def animate(self, name=None, dt=None):
		#dump("*-* animate", self.name(), self._animateTime, name, dt, level=3)
		nextTime = 0		# next event time

		#---------------------------------------------------------------
		# Calculate time for next event
		#---------------------------------------------------------------
		def nextEvent(name, dt):
			nonlocal nextTime
			self._animate[name] = t = self._animateTime + dt
			if nextTime==0 or t < nextTime:
				nextTime = t

		#---------------------------------------------------------------
		# Check if we have to perform this event
		# and update the nextTime if needed
		#---------------------------------------------------------------
		def doEvent(name):
			nonlocal nextTime
			t = self._animate.get(name,0)
			if t==0: return False
			if nextTime==0 or t < nextTime:
				nextTime = t
			return t == self._animateTime

		#---------------------------------------------------------------
		# name defined: Start event, calculate time of next event
		#---------------------------------------------------------------
		if name is not None:
			nextEvent(name, dt)
			if self.__animateId is not None: return
			#if name == "project": self.__start = time.time()

		#---------------------------------------------------------------
		# Animate projection progress
		# perform fast synchronous draw during the projection phase
		#---------------------------------------------------------------
		if doEvent("project"):
			#print("*-* animate project", self.name(), self._animate["project"], self._animateTime)
			if self._viewer.state() == PROJECTION_FINISHED:
				self._animate["project"] = 0
				if self._viewer.error():
					# FIXME we should clear all previous errors
					self.parent.event_generate("<<Errors>>")
				self.startDraw()
				#print(self.name(), "Render", time.time()-self.__start)
			else:
				self._viewer.draw(False, -1, True)
				#self.name()=="Red" and print("expose project", self.name())
				self._viewer.expose()
				nextEvent("project", PROJECT_TIMER)

		#---------------------------------------------------------------
		# Animate late drawing typically after projection or in 3D
		#---------------------------------------------------------------
		if doEvent("draw"):
			if self._viewer.state() == PROJECTION_DRAW:
				nextEvent("draw", DRAW_TIMER)
				# wait for some time to avoid a flash
				#if self._viewer.state("%")>0:
			else:
				self.endDraw()
			#self.name()=="Red" and print("*-* animate draw", self.name(),
			#		self._animate["draw"], self._animateTime)
			self._viewer.expose()

		#---------------------------------------------------------------
		# Animate panning of display, during mouse motion
		#---------------------------------------------------------------
		if doEvent("pan"):
			#print("*-* animate pan", self.name(), self._animate["pan"],
			#		self._animateTime)
			if self.panSlowdown():
				nextEvent("pan", PAN_TIMER)

		#---------------------------------------------------------------
		# Animate zoom in/out of orientation cube
		#---------------------------------------------------------------
		if doEvent("orientcube"):
			oc = self._viewer.ocube("length")
			if self._ocubeAction == OCUBE_ZOOMIN:
				if oc<self._ocubeBig:
					oc += OCUBE_STEP*tkFlair.dpiscale
					nextEvent("orientcube", OCUBE_TIMER)
				else:
					self._viewer.ocube("detail", True)
					self._animate["orientcube"] = 0
			elif self._ocubeAction == OCUBE_ZOOMOUT:
				self._viewer.ocube("detail", False)
				if oc>self._ocubeSmall:
					oc -= OCUBE_STEP*tkFlair.dpiscale
					nextEvent("orientcube", OCUBE_TIMER)
				else:
					self._animate["orientcube"] = 0
			self._viewer.ocube("length",oc)
			self._viewer.expose()

		#---------------------------------------------------------------
		# if nexttime is calculated then set the new timer
		#---------------------------------------------------------------
		if nextTime:
			# Call our self if animation tasks are still pending
			#print(">>> animate next dt=", nextTime-self._animateTime)
			self.__animateId = self.after(nextTime-self._animateTime, self.animate)
			self._animateTime = nextTime
		else:
			# Kill the timer if there are no more animation tasks pending
			self.__animateId  = None
			self._animateTime = 0

	# ----------------------------------------------------------------------
	# Check for errors and notify
	# ----------------------------------------------------------------------
	def check4Errors(self):
		if self._viewer.error("kernel"):
			self.flair().notify("Kernel Error",
				self._viewer.error("kernel"),
				tkFlair.NOTIFY_ERROR)

	# ----------------------------------------------------------------------
	# Return  snap to grid status or closest handle
	# ----------------------------------------------------------------------
	def snap(self, event):
		snap = self.editor.snap2Grid.get()
		if event.state & STATE.SHIFT:
			if event.state & STATE.CONTROL:
				self._geometry.set("snapdistance",  5*SNAPDISTANCE)
			else:
				self._geometry.set("snapdistance",  SNAPDISTANCE)
				snap = not snap
		return snap

	# ---------------------------------------------------------------------
	def controlPress(self, event):
		if self.action in (ACTION.VIEW_ORIGIN,
				ACTION.MOVE, ACTION.ADD, ACTION.ADD_NEXT):
			# When Control is pressed check for the appropriate action
			event.state |= STATE.CONTROL
			self.buttonMotion(event)

	# ---------------------------------------------------------------------
	def controlRelease(self, event):
		if self.action in (ACTION.VIEW_ORIGIN,
				ACTION.MOVE, ACTION.ADD, ACTION.ADD_NEXT):
			# When Control is pressed check for the appropriate action
			event.state &= ~STATE.CONTROL
			self.buttonMotion(event)

	# ----------------------------------------------------------------------
	# Common action before a new mouse click
	# ----------------------------------------------------------------------
	def _buttonPrepare(self, event):
		if self.focus_get() != self:
			# Don't do anything on the first click if we don't have the focus
			self.focus_set()
			return False
		self.focus_set()
		self.parent.setActive(self._viewer)
		if self.action == ACTION.PAN_SLOWDOWN:
			self.action = ACTION.SELECT
			self._animate["pan"] = 0
			self.historyAdd()
			self.draw()
			self.synchronizeViewports()
		self.x = self._x = event.x
		self.y = self._y = event.y
		return True

	# ----------------------------------------------------------------------
	# Mouse button1 (Left) pressed
	# ----------------------------------------------------------------------
	def button1(self, event):
		if not self._buttonPrepare(event): return

		if self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD, ACTION.ROTATE):
			self.actionMoveStop(event)
			return
		elif self._viewer.ocube("detail"):
			self.ocubeClicked(event)
			return

		self._cursor = self.cget("cursor")

		if self.parent.mouseAction in (ACTION.ZOOM_IN, ACTION.ZOOM_OUT):
			self.action = self.parent.mouseAction
			self.config(cursor=mouseCursor(self.action))

		elif self.parent.mouseAction == ACTION.PAN:
			self.actionPan(event)

		elif self.parent.mouseAction == ACTION.ORBIT:
			self.actionOrbit(event, True)

		elif self.parent.mouseAction == ACTION.ADD_ZONE:
			self.actionZone(event.x, event.y)
			self.editor.setMouseAction(ACTION.SELECT)

		elif self.parent.mouseAction == ACTION.ADD_ZONEPAINT:
			self.action = self.parent.mouseAction
			self._geometry.zone("clear")
			self.actionZoneShow(event.x, event.y)
			self.actionZone(event.x, event.y)

		elif self.parent.mouseAction == ACTION.ADD:
			self.actionAdd(event)

		elif self.parent.mouseAction == ACTION.ADD_NEXT:
			pass	# do nothing all action will be done in button release
		#	if self.__node == 0:
		#		self._viewer.edit("start",
		#			(self._viewer.pixel2uv(event.x, event.y), self.snap(event)))
		#		self.resetAnchor()

		#elif event.state & STATE.SHIFT:
		#	if Global.developer:
		#		self.action = ACTION.PEN
		#		self._viewer.pen("clear")
		#		self._viewer.pen("add",(event.x,event.y))
		#		self._viewer.expose()
		#	else:
		#		self.action = ACTION.AREA_SELECT
		#	self.config(cursor=mouseCursor(self.action))

		elif self.parent.mouseAction == ACTION.ADD_SPLINE_NODE:
			self.actionAddSplineNode(event)

		elif self.parent.mouseAction == ACTION.DEL_SPLINE_NODE:
			self.actionDelSplineNode(event)

		elif self.parent.mouseAction == ACTION.CENTER:
			self.centerViewports(event)

		elif self.parent.mouseAction == ACTION.INFO:
			self.actionInfo(event)

		elif self.parent.mouseAction == ACTION.PIVOT:
			self.actionSetPivot(event)

		else:
			self.actionSelect(event)

	# ----------------------------------------------------------------------
	# Mouse button2 (Middle or Wheel) pressed
	# ----------------------------------------------------------------------
	def button2(self, event):
		if self.action == ACTION.ADD_NEXT or \
		   not self._buttonPrepare(event):
			return
		elif self._viewer.ocube("detail"):
			self.parent.toggleAction(ACTION.ORBIT)
		elif LAPTOP:
			self.buttonMenu(event)
		else:
			self.buttonPan(event)

	# ----------------------------------------------------------------------
	# Mouse button3 (Right) pressed
	# ----------------------------------------------------------------------
	def button3(self, event):
		if self.action == ACTION.ADD_NEXT:
			self.actionAddEnd()
			return
		elif self.parent.mouseAction < 0 or self.parent.mouseAction > ACTION.ADD:
			self.escape()
			return
		elif not self._buttonPrepare(event):
			return

		if event.state & STATE.BUTTON2:
			self.actionOrbit(event, False)
		elif self._viewer.ocube("detail"):
			self.parent.toggleAction(ACTION.ORBIT)
		elif not LAPTOP:
			self.buttonMenu(event)
		elif event.state & STATE.SHIFT:
			self.actionPan(event)
		else:
			self.buttonPan(event)

	# ----------------------------------------------------------------------
	def double1(self, event):
		#dump("*-* double1", level=2)
		if self.action == ACTION.ADD_NEXT:
			self.actionAddEnd()
		elif self.action < 0:
			self.escape()

	# ----------------------------------------------------------------------
	# Event: move mouse without any button pressed
	# Display the position of the mouse, and/or perform object
	# transformations like move,rotate,scale
	# ----------------------------------------------------------------------
	def motion(self, event):
		self.parent.setStatusPos(self._viewer.pixel2xyz(event.x, event.y))
		if tkExtra.Balloon.isshown() and \
		   self.parent.mouseAction not in (ACTION.ADD_ZONE, ACTION.INFO):
			tkExtra.Balloon.hide()

		if self.action in (ACTION.MOVE, ACTION.ADD_NEXT):
			self.actionMove(event)

		elif self.action == ACTION.ROTATE:
			self.actionRotate(event)

		elif self.parent.mouseAction in (ACTION.ADD_ZONE, ACTION.ADD_ZONEPAINT):
			self.actionZoneShow(event.x, event.y)

		else:
			self.ocubeMotion(event)

	# ----------------------------------------------------------------------
	# Event: move mouse while pressing any button
	# ----------------------------------------------------------------------
	def buttonMotion(self, event):
		if self.snap(event):
			uv  = self._viewer.pixel2uv(event.x, event.y)
			xyz = self._viewer.snap(uv[0], uv[1])
		else:
			xyz = self._viewer.pixel2xyz(event.x, event.y)
		self.parent.setStatusPos(xyz)

		if self.action == ACTION.SEL_OR_MOVE:
			# Check for the type of action
			if abs(event.x-self._x)>1 or abs(event.y-self._y)>1 and \
			   self.parent.closest[0] in ('B','O','C'):
				self._actionMoveOrRotate(event)

		elif self.action == ACTION.SEL_OR_AREA:
			if abs(event.x-self._x)>1 or abs(event.y-self._y)>1:
				if self.parent.mouseAction == ACTION.EDIT:
					self.action = ACTION.PEN
					self._viewer.pen("add", (event.x, event.y))
					self._viewer.expose()
				else:
					self.action = ACTION.AREA_SELECT
					self.config(cursor=mouseCursor(self.action))
					self._viewer.rectangle(self.x, self.y, event.x, event.y)
					self._viewer.expose()

		elif self.action == ACTION.PEN:
			self._viewer.pen("add", (event.x, event.y))
			self._viewer.expose()

		elif self.action in (ACTION.MOVE, ACTION.ADD, ACTION.ADD_NEXT):
			self.actionMove(event)

		elif self.action == ACTION.ROTATE:
			self.actionRotate(event)

		elif self.action==ACTION.VIEW_MOVE: # or \
		#   (self.action==ACTION.VIEW_ORIGIN and event.state&STATE.CONTROL):
			matrix  = self._viewer.matrix()
			matrixv = bmath.Matrix(self.viewports[self.selected]._viewer.matrix())
			w  = bmath.Vector(matrix[0][2], matrix[1][2], matrix[2][2])
			wv = bmath.Vector(matrixv[0][2],matrixv[1][2],matrixv[2][2])

			# Move along _axis (xyz-save) * axis
			d = (xyz[0]-self._save[0])*self._axis[0] + \
			    (xyz[1]-self._save[1])*self._axis[1] +\
			    (xyz[2]-self._save[2])*self._axis[2]

			# The center calculated as
			#     center = self._save + d*self._axis
			#        C   = S + d*A
			# is along the projection of the real center
			# Therefore we need to find the intersection of the two
			# lines:
			#    1) perpendicular to the screen passing from 'center'
			#        S' = C + w*t = S + d*A + w*t
			#    2) along the viewport center towards the wv vector
			#        S' = S + wv*tv
			# From 1,2
			#       wv*tv = d*A + w*t
			# Taking the cross product with w
			#       w^wv*tv = d*w^A + t*w^w = d*w^A
			# Therefore we can solve for the tv and projected it
			# along the line 2
			A = w ^ wv
			B = d * (w^self._axis)
			try: tv =B[0]/A[0]
			except: pass
			try: tv = B[1]/A[1]
			except: pass
			try: tv = B[2]/A[2]
			except: pass

			center = self._save + wv*tv
			self.viewports[self.selected]._viewer.origin(center)
			self.viewports[self.selected].drawFast3D()
			self.parent.expose()

		elif self.action == ACTION.VIEW_ORIGIN:
			# Move center of viewport
			if not self.moveAbsolute(event):
				xyz = self._viewer.moveUV(self._save, xyz)
			self.viewports[self.selected]._viewer.origin(xyz)
			self.viewports[self.selected].drawFast3D()
			self.parent.expose()

		elif self.action == ACTION.VIEW_ROTATE:
			# Rotate viewport
			x,y = self._viewer.viewport(self.selected, "centerview")
			phi = math.atan2((event.y-y)*self._viewer.aspect(), event.x-x)
			#if event.state & STATE.SHIFT:
			#	phi = round(phi/ROTSTEP)*ROTSTEP
			phi -= self._phi
			m = bmath.Matrix(4)
			m.rotate(-phi, self._axis)
			m = m*self._save
			if event.state & STATE.SHIFT:
				rx, ry, rz = m.getEulerRotation()
				rx = round(rx/ROTSTEP)*ROTSTEP
				ry = round(ry/ROTSTEP)*ROTSTEP
				rz = round(rz/ROTSTEP)*ROTSTEP
				m = bmath.Matrix.eulerRotation(rx,ry,rz)
			xyz = self.viewports[self.selected]._viewer.origin()
			self.viewports[self.selected]._viewer.matrix(m)
			self.viewports[self.selected]._viewer.origin(xyz)
			self.viewports[self.selected]._sync = False
			self.viewports[self.selected].drawFast3D()
			self.parent.expose()

		elif self.action in (ACTION.AREA_SELECT, ACTION.ZOOM_IN, ACTION.ZOOM_OUT):
			self._viewer.rectangle(self.x, self.y, event.x, event.y)
			self._viewer.expose()

		elif self.action == ACTION.PAN:
			dx = event.x - self.x
			dy = event.y - self.y
			dt = float(event.time - self.t)
			if dt>0.0:
				self.vx = float(dx) / dt
				self.vy = float(dy) / dt

			ox, oy = self._viewer.offset()
			# Perspective or Combo
			if self._viewer.get("3D") and self._viewer.get("projection"):
				z = self._viewer.zoom()
			else:
				z = 1.0
			self._viewer.offset(ox-dx/self._viewer.Sx()*z, oy+dy/self._viewer.Sy()*z)
			self.x = event.x
			self.y = event.y
			self.t = event.time
			#dump("+++ buttonMotion.draw2", self.name(), level=3)
			if not self.drawFast3D():
				self.draw()
				self.synchronizeViewports()

		elif self.action == ACTION.ORBIT:
			width  = float(self.winfo_width())
			height = float(self.winfo_height())
			m = min(width, height)

			if self._anchorAxis == 0:
				axis = bmath.Vector(	self._save[0][0],
							self._save[0][1],
							self._save[0][2])
			elif self._anchorAxis == 1:
				axis = bmath.Vector(	self._save[1][0],
							self._save[1][1],
							self._save[1][2])
			elif self._anchorAxis == 2:
				axis = bmath.Vector(	self._save[2][0],
							self._save[2][1],
							self._save[2][2])
			else:
				axis = None

			#axis,phi = \
			quat = trackball((2*self.x   - width)  / m,
					  (height - 2*self.y)  / m,
					  (2*event.x - width)  / m,
					  (height - 2*event.y) / m, axis)
			if quat is None: return
			#print(f"Q={self._quat} q={quat} Q*q={self._quat*quat}")
			self._quat @= quat

			if event.state & STATE.SHIFT:
				rx,ry,rz = self._quat.matrix().getEulerRotation()
				# round with ROTSTEP
				rrx = round(rx/ROTSTEP)*ROTSTEP
				rry = round(ry/ROTSTEP)*ROTSTEP
				rrz = round(rz/ROTSTEP)*ROTSTEP
				if self._anchorAxis == 0:
					rry = ry
					rrz = rz
				elif self._anchorAxis == 1:
					rrx = rx
					rrz = rz
				elif self._anchorAxis == 2:
					rrx = rx
					rry = ry
				m = bmath.Matrix.eulerRotation(rrx,rry,rrz)
				self._viewer.matrix(m)
			else:
				self._viewer.matrix(self._quat.matrix())

			# Restore origin
			if self._viewer.get("3D"):
				# Pan Forward to restore orbit depth
				w  = self._viewer.basis('w')
				x  = self._origin[0] + self._orbitDepth*w[0]
				y  = self._origin[1] + self._orbitDepth*w[1]
				z  = self._origin[2] + self._orbitDepth*w[2]
				self._viewer.origin((x,y,z))
				#dump("+++ buttonMotion.draw2", self.name(),
				#	"_viewer.draw(False, DRAW.FAST, True)", level=3)
				self._viewer.draw(False, DRAW.FAST, True)
			else:
				self._viewer.origin(self._origin)

			self._viewer.expose()
			self._sync = False
			self.x = event.x
			self.y = event.y

		elif self.parent.mouseAction == ACTION.INFO:
			self.actionInfo(event)

		elif self.action == ACTION.ADD_ZONEPAINT:
			self.actionZoneShow(event.x, event.y)
			self.actionZone(event.x, event.y)

	# ----------------------------------------------------------------------
	def buttonRelease(self, event):
		self.config(cursor=self._cursor)

		if self.action in (ACTION.SEL_OR_MOVE, ACTION.SEL_OR_AREA):
			if self.parent.closest is not None:
				if event.state & STATE.CONTROL or \
				   self.parent.closest[0] not in ('B','O','C') or \
				   self.parent.closest[2]>GOBJECT_NODE_NONE:
					self.actionDoSelection(event)
				else:
					self._actionMoveOrRotate(event)
					return		# Return not to change action

		elif self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD, ACTION.ROTATE):
			self.actionMoveStop(event)

		elif self.action in (ACTION.VIEW_ORIGIN, ACTION.VIEW_MOVE, ACTION.VIEW_ROTATE):
			viewer = self.viewports[self.selected]
			viewer.draw(True)
			viewer.expose()
			for v in self.viewports: v.expose()
			self.parent.setStatus()

		elif self.action == ACTION.AREA_SELECT:
			clear = (self.parent.mouseAction >= 0)
			action = SELECT.TOGGLE
			if event.state & STATE.SHIFT:
				clear = False
				if event.state & STATE.CONTROL:
					action = SELECT.UNSELECT
				else:
					action = SELECT.SELECT
			elif event.state & STATE.CONTROL:
				clear = False

			if self.editor.editType() == Card.REGION and \
			   self._geometry.selection("b",BIT.SELECT)==0:
				clear = bool(event.state & STATE.SHIFT)

			# Bodies selected?
			objs = self._viewer.enclosed(self.x, self.y,
					event.x, event.y, "B")
			if objs:
				self.editor.select(("B",objs,0), clear, action)
				clear = False

			# Objects selected?
			if self.parent.mouseAction != ACTION.SELECT_BODY:
				objs = self._viewer.enclosed(self.x, self.y,
						event.x, event.y, "O")
				if objs:
					self.editor.select(("O",objs,0), clear, action)
			self._viewer.rectangle()
			self.draw()

		elif self.action == ACTION.PAN:
			# Only click
			if abs(event.x-self._x)<2 and abs(event.y-self._y)<2:
				self.actionSetPivot(event)

			# if we are still moving
			elif event.time - self.t < 100:
				self.action = ACTION.PAN_SLOWDOWN
				self.animate("pan", PAN_TIMER)
				return
			else:
				# Set action to SELECT before draw
				# to ask for a spawned draw
				self.draw(late=True)

		elif self.action in (ACTION.ZOOM_IN, ACTION.ZOOM_OUT):
			x1,y1 = self._viewer.pixel2uv(self.x, self.y)
			x2,y2 = self._viewer.pixel2uv(event.x, event.y)
			self._viewer.offset(0.5*(x1+x2), 0.5*(y1+y2))

			zoom = self.action==ACTION.ZOOM_IN
			if event.state & STATE.CONTROL: zoom = not zoom
			if abs(event.x-self._x)>1 or (event.y-self._y)>1:
				if zoom:
					w = abs(x2-x1)
					h = abs(y2-y1)
					if w>0 and h>0:
						x1,y1,x2,y2 = self._viewer.extends()
						zx = (x2-x1) / w
						zy = (y2-y1) / h
						self._viewer.zoom(min(zx,zy))
				else:
					z = self._viewer.zoom()
					zx = float(abs(event.x-self._x)) / \
					     float(self.winfo_width())  * z
					zy = float(abs(event.y-self._y)) / \
					     float(self.winfo_height()) * z
					self._viewer.zoom(max(zx,zy))
			else:
				if zoom:
					self.zoomIn(event)
				else:
					self.zoomOut(event)
			self._viewer.rectangle()
			self.draw(False)

		elif self.action == ACTION.ORBIT:
			if self.parent.mouseAction != ACTION.ORBIT:
				self._viewer.set("trackball", 0)
			self.draw(True)

		elif self.action in (ACTION.ADD, ACTION.ADD_NEXT):
			self.actionAddNext(event)
			return

		self.action = ACTION.SELECT

	# ----------------------------------------------------------------------
	def buttonMenu(self, event):
		if self.action not in (ACTION.SELECT, ACTION.PAN, ACTION.ORBIT): return
		if event.state & STATE.CONTROL:
			self.parent.insertMenu()
		else:
			self.popupMenu(event)

	# ----------------------------------------------------------------------
	def buttonPan(self, event):
		if event.state & STATE.CONTROL:
			self.actionOrbit(event, False)
		elif event.state & STATE.SHIFT:
			self.actionZoomRect(event)
		else:
			self.actionPan(event)

	# ----------------------------------------------------------------------
	def panSlowdown(self):
		self.vx *= FRICTION
		self.vy *= FRICTION
		dx = self.vx * PAN_TIMER
		dy = self.vy * PAN_TIMER
		if self.action== ACTION.PAN_SLOWDOWN and (abs(dx)>=2.0 or abs(dy)>=2.0):
			ox, oy = self._viewer.offset()
			self._viewer.offset(ox-dx/self._viewer.Sx(), oy+dy/self._viewer.Sy())
			if self._viewer.get("3D"):
				#dump("+++ panSlowdown.draw1", self.name(),
				#	"_viewer.draw(False, DRAW.FAST, True)", level=3)
				self._viewer.draw(False, DRAW.FAST, True)
				self._viewer.expose()
			else:
				#dump("+++ panSlowdown.draw2", self.name(), level=3)
				self.draw()
			self.synchronizeViewports()
			return True
		else:
			self.action = ACTION.SELECT
			self._animate["pan"] = 0
			self.historyAdd()
			#dump("+++ panSlowdown.draw3", self.name(), level=3)
			self.draw(late=True)
			return False

	# ----------------------------------------------------------------------
	# Stop action restore everything
	# ----------------------------------------------------------------------
	def actionStop(self, event=None):
		if self.action == ACTION.ORBIT:
			self._viewer.matrix(self._save)
			self._viewer.set("trackball", 0)
		elif self.action in (ACTION.VIEW_ORIGIN, ACTION.VIEW_MOVE):
			self.viewports[self.selected]._viewer.origin(self._save)
		elif self.action == ACTION.VIEW_ROTATE:
			self.viewports[self.selected]._viewer.matrix(self._save)
		elif self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD, ACTION.ROTATE):
			self.actionMoveCancel()
		elif self.action == ACTION.ADD_NEXT:
			self.actionMoveCancel()
			self.parent.loadProject()
			self.parent.redraw()
		else:
			self._viewer.rectangle()

		if self.parent.mouseAction == ACTION.SELECT_VIEWPORT:
			self._viewer.set("borderwidth",0)
		self.editor.setMouseAction(ACTION.SELECT)
		self.action = ACTION.SELECT
		self._viewer.expose()

	# ----------------------------------------------------------------------
	# Default mouse action depending on the proximity of the mouse to
	# objects as well the key-combinations Ctrl/Shift
	# ----------------------------------------------------------------------
	def actionSelect(self, event):
		if self.parent.mouseAction < 0:
			kind = name = None
			if self.parent.mouseAction == ACTION.SELECT_BODY:
				kind = "B"
			elif self.parent.mouseAction == ACTION.SELECT_REGION:
				kind = "R"
			elif self.parent.mouseAction == ACTION.SELECT_ZONE:
				if self.editor.editType() == Card.REGION:
					kind = "Z"
					name = self.editor.editObject.name()
				else:
					kind = "RZ"
			elif self.parent.mouseAction in (ACTION.SELECT_OBJECT, ACTION.SELECT_NODE):
				kind = "O"
			elif self.parent.mouseAction == ACTION.SELECT_SPLINE:
				kind = "sO"	# search object with spline modifier
			elif self.parent.mouseAction == ACTION.SELECT_VIEWPORT:
				self._viewer.set("borderwidth",0)
				self._viewer.expose()
				self.parent.selectViewer(self)
				return
			else:
				assert False
			self.parent.closest = self._viewer.closest(event.x, event.y,
							CLOSE_DISTANCE, kind, name)

		elif self.editor.editType() == Card.REGION:
			self.parent.closest = self._viewer.closest(event.x, event.y,
						CLOSE_DISTANCE,
						None,
						self.editor.editObject.name())

		else:
			self.parent.closest = self._viewer.closest(event.x, event.y,
						CLOSE_DISTANCE)

		dump("GeometryViewer.actionSelect: self.parent.closest=", self.parent.closest)

		if self.parent.closest is None:
			self.action = ACTION.SEL_OR_AREA
			self.parent.setStatus()
			return

		obj,idx,node = self.parent.closest

		if self.parent.mouseAction == ACTION.SELECT_NODE:
			self.parent.updateActionToolSelection()

		elif obj=="V" and idx>=0:
			self.viewports[idx]._viewer.viewOto0()
			self.selected = idx

			x,y = self._viewer.viewport(idx, "centerpixel")

			width  = self.winfo_width()
			height = self.winfo_height()
			ww = width  / 20
			hh = height / 20

			# default action
			self.action = ACTION.VIEW_MOVE

			# check for rotation, if we click far from center
			# and close to borders
			if (event.x-x)**2 + (event.y-y)**2 > ww*hh:
				if event.x<=ww<x or (event.x>=width-ww  and x<width-ww) or \
				   event.y<=hh<y or (event.y>=height-hh and y<height-hh):
					self.action = ACTION.VIEW_ROTATE

			elif abs(event.x-x) < 5 and abs(event.y-y) < 5:
				self.action = ACTION.VIEW_ORIGIN

			# remember the position of the viewport
			self._save = bmath.Vector(self.viewports[idx]._viewer.origin())
			matrix  = self._viewer.matrix()
			matrixv = bmath.Matrix(self.viewports[idx]._viewer.matrix())

			w  = bmath.Vector(matrix[0][2], matrix[1][2], matrix[2][2])
			wv = bmath.Vector(matrixv[0][2],matrixv[1][2],matrixv[2][2])

			if self.action == ACTION.VIEW_ORIGIN:
				self._axis = wv ^ w

			elif self.action == ACTION.VIEW_MOVE:
				# Tricky we have to move along the wv
				# however the viewport line is NOT the intersection
				# of the two planes rather a vertical line at the CENTER position!
				ww = w ^ wv
				ww.norm()
				self._axis = ww ^ w

			else:
				# remember the matrix of the viewport
				self._save = matrixv
				self._phi  = math.atan2((event.y-y)*self._viewer.aspect(), event.x-x)
				self._axis = bmath.Vector(matrix[0][2],matrix[1][2],matrix[2][2])

			self.config(cursor=mouseCursor(self.action))

		# Initial hit on Zone or Region?
		elif obj in ("R","Z"):
			if self.parent.mouseAction == ACTION.EDIT:
				self._viewer.pen("clear")
				self._viewer.pen("add",(event.x,event.y))
				self._viewer.expose()
			# Postpone the selection after checking if
			# the cursor moved in the mean time
			self.action = ACTION.SEL_OR_AREA

		# Cursor
		elif obj == "C":
			self.action = ACTION.SEL_OR_MOVE

		# Check if we hit on a selected item
		elif self.editor.isSelected((obj,idx)):
			if obj=="B" and self._geometry.cursor("body") != idx:
				self._geometry.cursor("body", idx)
				self.parent.closest = self._viewer.closest(
					event.x, event.y, CLOSE_DISTANCE, "B")
			elif obj=="O":
				self._geometry.object(idx, "selectnode", node)
				#self.parent.selectObjectNode(idx, node)
			# Postpone the selection after checking if
			# the cursor moved in the mean time
			self.action = ACTION.SEL_OR_MOVE

		else:
			self.actionDoSelection(event)

	# ----------------------------------------------------------------------
	# Perform the actual selection
	# ----------------------------------------------------------------------
	def actionDoSelection(self, event):
		#dump("*-* actionDoSelection", level=2)
		obj, idx, node = self.parent.closest
		if obj=="R":
			msg = "Region: "
			for i in idx:
				name = self._geometry.region(i,"name")
				if name == "VOXEL":
					if self._viewer.get("3D"):
						uv = self._viewer.pixel2uv(event.x, event.y)
						xyz = self._viewer.hit(*uv)
					else:
						xyz = self._viewer.pixel2xyz(event.x, event.y)
					voxelReg = self._geometry.voxel("get",xyz)+1
					idx  = -voxelReg
					msg += f"VOXEL{voxelReg:03}"
				else:
					msg += f"{name} "
			self.parent.setStatus(msg)

		elif obj=="Z":
			self.parent.setStatus("Zones: %s"%
				(" ".join(["zone%02d"%(i+1) for i in idx])))

		elif obj=="B":
			self.parent.setStatus("Body: %s [%s]" %
				(self._geometry.body(idx, "name"),
				 self._geometry.body(idx, "type")))

		elif obj=="O":
			wasselected = self._geometry.object(idx, "select")
			self.parent.setStatus("Object: %s" %
				(self._geometry.object(idx, "name")))

		# Do the selection
		action = SELECT.TOGGLE
		clear  = (self.parent.mouseAction >= 0)
		if event.state & STATE.SHIFT:
			clear = False
			if event.state & STATE.CONTROL:
				action = SELECT.UNSELECT
			else:
				action = SELECT.SELECT
		elif event.state & STATE.CONTROL:
			clear = False

		self.editor.select((obj,idx,node), clear, action)

		# after the selection check if we have to select a node as well
		if obj=="O" and wasselected:
			self._geometry.object(idx, "selectnode", node)
			#self.parent.selectObjectNode(idx, node)

		if self.parent.mouseAction == ACTION.EDIT:
			self.actionOverlaps()

	# ----------------------------------------------------------------------
	# Correct overlaps in 2D
	# ----------------------------------------------------------------------
	def actionOverlaps(self):
		if self.editor.editType() != Card.REGION: return
		region = self.editor.editObject.cards[0]
		zones  = self.editor.propListSelectedZones()
		if not zones: return
		if self._viewer.region(region[Card.ID], "correct", zones):
			zones = []
			for z in self._geometry.region(region[Card.ID],"expr"):
				if z: zones.append(csg.toString(isinstance(self.input(),(Mcnp,Phits)),*z))
			self.flair().addUndo(self.flair().setZonesUndo(region,zones))
			self.flair().refresh("card")

	# ----------------------------------------------------------------------
	# Display object information at location
	# ----------------------------------------------------------------------
	def actionInfo(self, event):
		closest = self._viewer.closest(event.x, event.y, CLOSE_DISTANCE, "BR")
		if self.frame.getInt("3D"):
			uv = self._viewer.pixel2uv(event.x, event.y)
			xyz = self._viewer.hit(*uv)
		else:
			xyz = self._viewer.pixel2xyz(event.x, event.y)
		message = "Location:\n  x: %s\n  y: %s\n  z: %s" % \
			(tuple(bmath.format(x,10) for x in xyz))

		# Specific information
		if closest is None:
			pass

		# Body information
		elif closest[0] == "B":
			idx = closest[1]
			message += "\nBody: %s [%s]"%(
				self._geometry.body(idx,"name"),
				self._geometry.body(idx,"type"))
			what = self._geometry.body(idx,"what")
			message += "\nWhat: %s"%(" ".join([bmath.format(x,10) for x in what]))
			card = self.editor.findCard("B",idx)
			if card:
				dx  = card["@dx"]
				dy  = card["@dy"]
				dz  = card["@dz"]
				exp = card["@expansion"]
				tra = card["@transform"]
				if dx or dy or dz:
					message += f"\nTranslat: {dx}, {dy}, {dz}"
				if exp:
					message += f"\nExpansion: {exp}"
				if tra:
					message += f"\nTransform: {tra}"

		# Region information
		elif closest[0] == "R":
			indices = []
			names   = []
			idx     = closest[1][0]
			indices.append(idx)

			name = self._geometry.region(idx,"name")
			names.append(name)

			if name == "VOXEL" and self.frame.getInt("Show.voxel"):
				voxelReg = self._geometry.voxel("get",xyz)+1
				idx -= voxelReg
				names.append(f"VOXEL{voxelReg:03}")
				indices.append(idx)

			elif self._geometry.region(idx,"hasmatrix") and \
			     self.frame.getInt("Show.lattice"):
				idx = self._viewer.where(*xyz)
				if idx is not None:
					names.append(self._geometry.region(idx,"name"))
				else:
					names.append("-error-")
				indices.append(idx)

			message += "\nPath: " if len(names)>1 else "\nRegion:"
			message += ' \N{BLACK RIGHT-POINTING TRIANGLE} '.join(names)
			prefix = "   "
			for i,n in zip(indices,names):
				if len(indices)>1:
					message += f"\nRegion: {n}"
				card = self.editor.findCard("R",i)
				if card is None: continue
				mat = card[Card.MATERIAL]
				if mat: message += f"\n{prefix}Material: {mat.name()}"
				mat2 = card["@matDecay"]
				if mat2 and mat2 is not mat:
					message += f"\n{prefix}matDecay: {mat2.name()}"

				#if name2:
				#	message += "\nValue: %g"%(self._viewer.voxel(card[Card.ID],"value"))
				#else:
				#	message += "\nValue: %g"%(self._viewer.region(card[Card.ID],"value"))

				rot = card["@rotdefi"]
				if rot: message += f"\n{prefix}Transform: {rot}"

				for prop,value in card.property.items():
					if prop in (	# skip the following
						Card.ASSIGNMAT,
						#Card.CACHE,
						Card.COLOR,
						Card.CONT,
						Card.ID,
						Card.LATTICE,
						Card.MATERIAL,
						"@matDecay",
						"@n",
						"@rotdefi",
						Card.SELECT,
						Card.TYPE,
						Card.ZONE): continue
					if prop[0] == "@": prop = prop[1:]
					if isinstance(value,Card):
						if prop in ("biasingA","biasingH","biasingE","biasingN"):
							message += f"\n{prefix}{prop}:"
							message += f"\n{prefix}   Imp: {value.numWhat(3)}"
							message += f"\n{prefix}   Split: {value.numWhat(2)}"
						else:
							message += f"\n{prefix}{prop}: {value.name()}"
					elif value:
						message += f"\n{prefix}{prop}: {value}"

				# Check for magnetic field
				if Card.ASSIGNMAT in card.property:
					try:
						field = card[Card.ASSIGNMAT].intWhat(5)
						if field > 0:
							message += f"\n{prefix}field: "\
								f"{FlukaLayout.ASSIGNMA_field[field][0]}"
					except AttributeError:
						pass

			# Show information on voxel
			if name == "VOXEL" and self.frame.getInt("Show.voxel"):
				for card in self.input()["VOXELS"]:
					if card.ignore() or card[FlukaCard.VOXEL] is None: continue
					voxel = card[FlukaCard.VOXEL]
					message += "\nVoxel:"
					message += "\n   Data: %s = %d -> %d"%(
								str(self._geometry.voxel("index",xyz)),
								self._geometry.voxel("data",xyz),
								self._geometry.voxel("get",xyz))
					message += f"\n   File:  {voxel.filename}"
					message += f"\n   Title: {voxel.title}"
					message += f"\n   Nxyz:  {voxel.nx} x {voxel.ny} x {voxel.nz}"
					message += f"\n   dxyz:  {voxel.dx} x {voxel.dy} x {voxel.dz}"
					message += f"\n   nregions: {voxel.nregions+1}"
					message += f"\n   nvoxels:  {voxel.nvoxels}"
					message += f"\n   Rtstruct: {voxel.roiN}"

			# Any usrbin visible?
			for i,tag in enumerate(Project.GeometryLayerInfo.USRBINNAMES):
				if self.frame.getInt(tag):
					if not self._viewer.usrbin(i,"isregion"):
						value = self._viewer.usrbin(i,"value",xyz)
					else:
						value = self._viewer.usrbin(i,"value",idx)
					if value is not None:
						message += "\n%s: %s"%(tag, bmath.format(value,10))

						if not self._viewer.usrbin(i,"isregion"):
							value = self._viewer.usrbin(i,"error",xyz)
						else:
							value = self._viewer.usrbin(i,"error",idx)
						if value is not None:
							message += " \N{PLUS-MINUS SIGN} %s %%"%(bmath.format(value,5))

		self.clipboard_clear()
		self.clipboard_append(message)
		self.parent.showBalloon(message, 60000)

	# ----------------------------------------------------------------------
	# Orbit viewport with mouse
	# ----------------------------------------------------------------------
	def actionOrbit(self, event, lockAxis):
		if self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD, ACTION.ROTATE, ACTION.ADD_NEXT):
			return

		self.action = ACTION.ORBIT
		self.config(cursor=mouseCursor(self.action))
		self._viewer.viewOto0()

		# Save matrix before any transformation
		self._save   = self._viewer.matrix()
		self._quat   = bmath.Quaternion(bmath.Matrix(self._viewer.matrix()))
		self._origin = bmath.Vector(self._viewer.origin())

		width  = float(self.winfo_width())
		height = float(self.winfo_height())
		m = min(width,height)
		x = (2.0*event.x - width)  / m
		y = (height - 2.0*event.y) / m
		v = bmath.Vector(x, y, projection2Sphere(TRACKBALLSIZE, x, y))

		if lockAxis:
			THRESHOLD = 0.05
			self._anchorValue = ""
			vv = tuple(map(abs, bmath.Matrix(self._viewer.matrix()).multNoTranslation(v)))
			if vv[0] <= vv[1] and vv[0] <= vv[2] and abs(vv[0])<THRESHOLD:
				self._anchorAxis = 0
			elif vv[1] <= vv[0] and vv[1] <= vv[2] and abs(vv[1])<THRESHOLD:
				self._anchorAxis = 1
			elif vv[2] <= vv[0] and vv[2] <= vv[1] and abs(vv[2])<THRESHOLD:
				self._anchorAxis = 2
			else:
				self._anchorAxis = -1
		else:
			self.resetAnchor()

		if self.frame.getInt("3D"):
			if self._viewer.get("projection") == 1:
				minw = -self._viewer.camera("focal")	# Perspective
			else:
				minw = 0.0				# Orthographic and Combo
			# Limit orbit distance to minw
			if self._orbitDepth < minw:
				self._orbitDepth = minw/2.0
			# Pan backward
			self._origin -= self._orbitDepth * bmath.Vector(self._viewer.basis('w'))

		self._viewer.set("trackball", 1)

		self._viewer.expose()
		self._sync = False

	# ----------------------------------------------------------------------
	# Zoom on rectangle drawn
	# ----------------------------------------------------------------------
	def actionZoomRect(self, event):
		self.action = ACTION.ZOOM_IN
		self.config(cursor=mouseCursor(self.action))

	# ----------------------------------------------------------------------
	# Pan viewport with mouse
	# ----------------------------------------------------------------------
	def actionPan(self, event):
		self.action = ACTION.PAN
		self.config(cursor=mouseCursor(self.action))
		self.vx = 0.0
		self.vy = 0.0
		self.t  = event.time

	#-----------------------------------------------------------------------
	# Center viewport and set pivot point
	#-----------------------------------------------------------------------
	def actionSetPivot(self, event):
		self._viewer.viewOto0()

		# Center current position
		u,v = self._viewer.pixel2uv(event.x, event.y)
		xyz = bmath.Vector(self._viewer.uv2xyz(u,v))
		origin = xyz

		# For 3D displays find hit position to set as viewers pivot point and center
		if self._viewer.get("3D"): # and self._viewer.get("projection"):
			# For Perspective or Combo
			# Calculate distance for orbit operations
			hit,regtype = self._viewer.hit(u,v,"T")
			if regtype != "BLCKHOLE":
				hit = bmath.Vector(hit)
				# Is not correct, it should be calculated
				# after the move but should be a reasonable
				# approximation
				xyz0 = bmath.Vector(self._viewer.uv2xyz(0.,0.))
				cpos = bmath.Vector(self._viewer.camera("position"))
				dhit = (hit-cpos).length()
				#width = float(self.winfo_width())/self._viewer.Sx()
				if self._viewer.get("projection")==1:	# Perspective
					# Find the location of the hit object above/below
					# viewing plane
					#if (hit-xyz) * bmath.Vector(self._viewer.basis('w')) > 0.0:
					#	self._orbitDepth = -self._orbitDepth
					self._orbitDepth = 0.0
					origin = hit
				else:
					self._orbitDepth = (hit-xyz).length()
					if self._viewer.get("projection")==2:	# Combo
						dxyz  = (xyz-cpos).length()
						origin = dhit/dxyz*xyz + (1.0-dhit/dxyz)*xyz0

		# Set new origin
		self._viewer.origin(*origin)
		self.draw(True)

	# ----------------------------------------------------------------------
	# Add a new object/body
	# Sequence of Actions:
	#   - actionAdd                 from Action.ADD
	#   - actionMove
	#   - actionAddNext             from Multiple Action.ADD_NEXT
	#      - actionMoveStop         when no more next item exists
	#         - actionUpdate
	#         - actionMoveFinalize
	#   - actionAddEnd
	#   - actionMove
	#      - _actionDoMove
	# ----------------------------------------------------------------------
	def actionAdd(self, event):
		dump("*-* GeometryViewer.actionAdd", level=2)
		if self.editor.lockEditing.get(): return
		if self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD,
				ACTION.ROTATE, ACTION.ADD_NEXT): return

		u,v = self._viewer.pixel2uv(event.x, event.y)
		if self.snap(event):
			P = bmath.Vector(self._viewer.snap(u, v))
		else:
			P = bmath.Vector(self._viewer.pixel2xyz(event.x, event.y))

		self._viewer.edit("start", ((u,v), self.snap(event)))
		self.parent.stopThreads()

		grid_size = self._viewer.grid("size")[0]
		matrix    = self._viewer.matrix()
		self.__node = 0
		self.__x   = self.x
		self.__y   = self.y

		# Adding flair object
		if self.parent.actionOption in CardInfo.FLAIR_TAGS:
			tag  = self.parent.actionOption
			what = [self.editor.findUniqueName(Card.OBJECT,tag.lower())]
			what.extend(P)			# 1,2,3 position
			what.extend([0, 0, 0])		# 4,5,6 Options,Anchor,Color
			if   tag == "arrow":
				what.extend([0.,0.,0.])	# direction
			elif tag == "camera":
				what.extend(P)		# at
				what.extend([0.,1.,0.])	# up
				what.append(30)		# focal length
				what.append(60)		# field of view
				what.append(4/3)	# aspect
			elif tag == "light":
				what.extend([0.,0.,0.])	# direction
				what.append(1.0)	# power
				what.append(0.0)	# falloff
				what.append(0.0)	# specular
				what.append(0.0)	# radius
				what.append(1)		# samples
			elif tag == "mesh":
				pass
			elif tag == "ruler":
				what.extend(P)		# other end
				what.extend(P)		# angle
			elif tag == "point":
				pass

			card = Card(tag, what)
			# Late settings for properties
			if tag == "spline":
				card["node:0"] = 1
				card["node:1.t"] = 0
				card["node:1.x"] = 0
				card["node:1.y"] = 0
				card["node:1.z"] = 0
				self.__node = 1

			card[Card.ID] = self.parent.addObject(card)
			self.parent.selectObjectsBodies([card],None)
			self._geometry.object(self.parent.objectsid, "save", True)
			self._geometry.object(self.parent.objectsid, "show", True)
			self._geometry.object(self.parent.objectsid, "selectcolor", _SELECT_COLOR)
			self._geometry.object(self.parent.objectsid, "select", True)
			self._msg = f"Add object {tag}"

		else:	# Adding body
			card = GeometryBody.newBody(
					self.parent.actionOption,
					P,
					grid_size,
					matrix)
			card.setName(self.editor.findUniqueName(Card.BODY))
			whats = [card.numWhat(i) for i in range(1,card.nwhats()) ]
			card[Card.ID] = self.parent.addBody(card.name(), card.tag, whats)

			# FIXME a bit overshoot!
			self.parent.derive()

			self.parent.viewportProject(False, False)
			self.parent.draw()
			self.parent.expose()	# all viewports

			self.parent.selectObjectsBodies(None,[card])
			self._geometry.cursor("body", self.parent.bodiesid[0])
			self._geometry.body(self.parent.bodiesid, "save", True)
			self._geometry.body(self.parent.bodiesid, "show",
						BIT.SELECT|BIT.WIREFRAME|BIT.MOVE)
			self._msg = f"Add body {card.tag}"

			# Special treatment for the add sequence
			matrix = self._viewer.matrix()
			w = bmath.Vector(matrix[0][2], matrix[1][2], matrix[2][2])

			# Tweak sequences for specific bodies depending the viewport
			if card.tag == "RPP":
				if   abs(abs(w[0])-1.0) < 1.0E-5:
					_ADD_SEQUENCE[card.tag] = [16]
				elif abs(abs(w[1])-1.0) < 1.0E-5:
					_ADD_SEQUENCE[card.tag] = [17]
				elif abs(abs(w[2])-1.0) < 1.0E-5:
					_ADD_SEQUENCE[card.tag] = [18]
				else:
					_ADD_SEQUENCE[card.tag] = [18, 6]
			elif card.tag == "XEC":
				if   abs(abs(w[1])-1.0) < 1.0E-5:
					_ADD_SEQUENCE[card.tag] = [2]
				elif abs(abs(w[2])-1.0) < 1.0E-5:
					_ADD_SEQUENCE[card.tag] = [1]
				else:
					_ADD_SEQUENCE[card.tag] = [1, 2]
			elif card.tag == "YEC":
				if   abs(abs(w[0])-1.0) < 1.0E-5:
					_ADD_SEQUENCE[card.tag] = [1]
				elif abs(abs(w[2])-1.0) < 1.0E-5:
					_ADD_SEQUENCE[card.tag] = [2]
				else:
					_ADD_SEQUENCE[card.tag] = [1, 2]
			elif card.tag == "ZEC":
				if   abs(abs(w[0])-1.0) < 1.0E-5:
					_ADD_SEQUENCE[card.tag] = [2]
				elif abs(abs(w[1])-1.0) < 1.0E-5:
					_ADD_SEQUENCE[card.tag] = [1]
				else:
					_ADD_SEQUENCE[card.tag] = [1, 2]
		self.grab_set()
		self.actionAddNext(event)

	# ----------------------------------------------------------------------
	# Move to the next add step
	# ----------------------------------------------------------------------
	def actionAddNext(self, event=None):
		dump("*-* GeometryViewer.actionAddNext", level=2)
		self.editor.setMouseAction(ACTION.ADD_NEXT)
		self.action = ACTION.ADD_NEXT

		#self._viewer.edit("start",
		#		(self._viewer.pixel2uv(event.x, event.y),
		#			self.snap(event)))
		self.resetAnchor()

		if self.parent.actionOption in CardInfo.FLAIR_TAGS:
			tag = self.parent.objects[0].tag
			self._geometry.object(self.parent.objectsid, "save", True)
			#self._viewer.edit("object", self.parent.objectsid[0])

		else:
			tag = self.parent.bodies[0].tag
			self._geometry.body(self.parent.bodiesid, "save", True)
			self._viewer.edit("body", self.parent.bodiesid[0])

		seq = _ADD_SEQUENCE.get(tag)
		if seq is not None:
			try:	# Find current item
				idx = seq.index(self.__node)
			except ValueError:	# On error start with first one
				self.__node = seq[0]
				return
			else:
				# If button1 released at the same position
				# set parameter without click-dragging, but simply dragging
				if idx==0 and abs(event.x-self.__x)<=1 and abs(event.y-self.__y)<=1:
					return
				else:
					try:	# Get the next item
						self.__node = seq[idx+1]
						return
					except:	# On end stop
						pass

		elif tag == "spline":	# special case
			n = self._geometry.object(self.parent.objectsid[0], "nodes")
			t,x,y,z = self._geometry.object(self.parent.objectsid[0], "node", n)
			if n>=2: # Check with previous point
				_,xp,yp,zp = self._geometry.object(self.parent.objectsid[0],
								"node", n-1)
				# ignore and continue editing the same
				if x==xp and y==yp and z==zp: return
				if n>2:	# Check for a closed spline to end editing
					_,xp,yp,zp = self._geometry.object(self.parent.objectsid[0],
								"node", 1)
					if x==xp and y==yp and z==zp:
						self.actionAddEnd()
						return

			self._geometry.object(self.parent.objectsid[0], "node", (t+1,x,y,z))
			self.__node = self._geometry.object(self.parent.objectsid[0], "nodes")

			# start editing the last node
			self._viewer.edit("object", self.parent.objectsid[0], self.__node)
			self._geometry.object(self.parent.objectsid, "save", True)
			return

		self.actionAddEnd()

	#-----------------------------------------------------------------------
	# Cleanup and end the Adding action
	#-----------------------------------------------------------------------
	def actionAddEnd(self):
		if self.parent.objects:
			if self.parent.objects[0].tag == "spline":	# special case
				# get last node
				n = self._geometry.object(self.parent.objectsid[0], "nodes")
				_,x,y,z = self._geometry.object(self.parent.objectsid[0], "node", n)
				# delete last node anyhow
				self._geometry.object(self.parent.objectsid[0], "del", n)
				if n>2:	# check for a closed spline
					_,xp,yp,zp = self._geometry.object(self.parent.objectsid[0],
									"node", 1)
					if x==xp and y==yp and z==zp:
						# Convert spline to CLOSED
						option = self._geometry.object(self.parent.objectsid[0],
									"option") | 1
						self._geometry.object(self.parent.objectsid[0],
									"option", option)

		# Stop all movement and update and register body/object
		self.actionMoveStop()

		if self.parent.bodies:
			self.editor.addBody(self.parent.bodies[0])
			try:
				self._geometry.cursor("body", self.parent.bodiesid[0])
			except IndexError:
				# Hide error when body was not added for xy reason
				pass
			self.parent.viewportProject(False, False)
			self.parent.draw()
			self.parent.expose()	# all viewports
		else:
			self.editor.addObject(self.parent.objects[0])

		self.editor.setMouseAction(ACTION.SELECT)

	# ----------------------------------------------------------------------
	# Decide whether to rotate or move next
	# ----------------------------------------------------------------------
	def _actionMoveOrRotate(self, event):
		dump("*-* GeometryViewer._actionMoveOrRotate", level=2)
		# Clone if Control is pressed
		self._clone = bool(event.state & STATE.SHIFT)
		if self._clone:
			self.editor.clone()
			self.editor.forceRefresh()

		self.resetAnchor()
		if GOBJECT_NODE_Z <= self.parent.closest[2] <= GOBJECT_NODE_X:
			self.resetAnchor(-self.parent.closest[2]+GOBJECT_NODE_X)
			self.actionMoveStart(event, 0)
		elif self.parent.closest[2] == GOBJECT_NODE_R:
			self.actionRotateStart(event)
		else:
			self.actionMoveStart(event, self.parent.closest[2])

		# Correct the cursor after cloning
		if self._clone:
			if self.parent.bodiesid:
				self._geometry.cursor("body", self.parent.bodiesid[0])
			self._geometry.cursor("cursor", False)

	# ----------------------------------------------------------------------
	# Start moving action on objects or bodies
	# Actions:
	#  - actionMoveStart
	#  - actionMove ...
	#     - _actionDoMove ...
	#  - actionMoveStop
	#     - actionUpdate
	#     - actionMoveFinalize
	# ----------------------------------------------------------------------
	def actionMoveStart(self, event, node=0):
		dump("*-* GeometryViewer.actionMoveStart", self.parent.closest, node, level=2)
		if self.action in (ACTION.ADD_NEXT, ACTION.MOVE,
				ACTION.MOVE_KEYBOARD, ACTION.ROTATE):
			return False
		if not self.parent.selectObjectsBodies(*self.editor.getMoveableSelection()):
			return False

		self.parent.stopThreads()

		self.action = ACTION.MOVE
		self.parent.setStatus("Move mouse or Type coordinates "
					"e.g. 5 10 20 or x50 /5 or x10 *5",
					"Blue","White")
		self._cursor = self.cget("cursor")
		if node==GOBJECT_NODE_R:
			self.config(cursor=mouseCursor(ACTION.ROTATE))
		else:
			self.config(cursor=mouseCursor(self.action))

		self.x = event.x
		self.y = event.y

		#---------------------------------------------------------------
		# Start editing relative to mouse
		#---------------------------------------------------------------
		def moveRelative2Mouse():
			self._viewer.edit("start",
				(self._viewer.pixel2uv(event.x, event.y),
				 self.snap(event)))

		if self.parent.closest is None:
			moveRelative2Mouse()
		elif self.parent.closest[0] == 'B':
			self._viewer.edit("body", self.parent.closest[1])
		elif self.parent.closest[0] == 'O':
			if GOBJECT_NODE_NONE < node <= 0:
				moveRelative2Mouse()
				for card in self.parent.objects:
					card[Card.NODE] = 0
			else:
				self._viewer.edit("object", self.parent.closest[1], node)
				for card in self.parent.objects:
					card[Card.NODE] = node
		elif self.parent.closest[0] == 'C':
			self._viewer.edit("cursor", None)
		else:
			moveRelative2Mouse()

		if self.parent.objectsid:
			self._geometry.object(self.parent.objectsid, "save", True)
		if self.parent.bodiesid:
			self._geometry.body(self.parent.bodiesid, "save", True)
			self._geometry.body(self.parent.bodiesid, "show",
					BIT.SELECT|BIT.WIREFRAME|BIT.MOVE)
		# XXX It should be saved for each object!!!
		self.__node = node
		if node<=0:
			self._msg = "Move selected objects"
		else:
			self._msg = "Resize selected objects"

		self._geometry.cursor("cursor", False)		# hide the cursor
		self.grab_set()
		return True

	# ----------------------------------------------------------------------
	# FIXME merge _anchorAxis with _axis
	# ----------------------------------------------------------------------
	def actionMove(self, event):
		#dump("\n*-* GeometryViewer.actionMove", level=2)
		if self.editor.lockEditing.get(): return
		if not self._anchorValue:
			if self.parent.closest and self.parent.closest[0] == 'B': # Bodies
				# Position absolute everything else relative
				if self.__node>0:
					relative = False
				else:
					relative = not bool(event.state & STATE.CONTROL)
			else:	# Objects and with keyboard
				# Everything absolute unless CONTROL is pressed
				relative = not bool(event.state & STATE.CONTROL)
			# Calculate move distance
			self._viewer.edit("move",
					self._viewer.pixel2uv(event.x, event.y),	# u,v [0]
					(self._anchorAxis,				# axis [1]
					 self.snap(event),				# snap [2]
					 relative))					# relative [3]
			move = self._viewer.edit("move")
			self.parent.setStatus("Move: dx=%g  dy=%g dz=%g"%tuple(move), timeout=0)
		else:
			# Parse anchorValue
			val = []
			for x in self._anchorValue.split():
				try: val.append(float(x))
				except ValueError: pass
			if val: self._viewer.edit("value", val, self._anchorAxis)

		# Move objects
		self._actionDoMove()

	# ----------------------------------------------------------------------
	# Move selected objects in the dx.dy direction
	# called from the keyboard event handler
	# ----------------------------------------------------------------------
	def actionMoveBy(self, event, dx, dy):
		dump("*-* GeometryViewer.actionMoveBy",dx,dy, level=2)
		# Find cursor position
		x,y,z = self._viewer.uv2xyz(*self._viewer.edit("start"))
		u,v,w = self._viewer.edit("move")
		u,v   = self._viewer.xyz2uv(x+u, y+v, z+w)
		# keyboard event, doesn't have x,y
		event.x, event.y = self._viewer.uv2pixel(u,v)
		self.actionMoveStart(event)
		gx, gy = self._viewer.grid("size")
		u += gx*dx/20.
		v += gy*dy/20.
		self._viewer.edit("move", (u,v), (-1, False, True))
		# Move objects
		self._actionDoMove()
		self.action = ACTION.MOVE_KEYBOARD

	# ----------------------------------------------------------------------
	# Perform the move and redraw
	# ----------------------------------------------------------------------
	def _actionDoMove(self):
		#dump("*-* GeometryViewer._actionDoMove", level=2)
		self.parent.stopThreads()

		# Move objects
		if self.parent.objectsid:
			try:
				self._viewer.object(self.parent.objectsid, "move", self.__node)
			except IndexError:
				return

		# Move bodies
		if self.parent.bodiesid:
			try:
				self._viewer.body(self.parent.bodiesid, "move", self.__node)
			except IndexError:
				return

		# Redraw if needed
		if self.parent.objectLight: self._geometry.setLights()
		if self.parent.bodiesid:
			self.parent.viewportProject(False,False)
			self.parent.draw(False)
		else:
			# fast redraw all viewports if lights
			# or camera are selected
			redraw = False
			for obj in self.parent.objects:
				if obj.tag in ("light", "camera"):
					redraw = True
					break
			if redraw:
				self.drawFast3D()
				for view in self.viewports:
					if not view._viewer.get("3D"): continue
					view.loadCamera(fromobject=True)
					view._viewer.draw(False, DRAW.FAST, True)
		self.parent.expose()	# all viewports

	# ----------------------------------------------------------------------
	# Start rotating action on objects or bodies
	# ----------------------------------------------------------------------
	def actionRotateStart(self, event):
		Global.dump("*-* GeometryViewer.actionRotateStart", level=2)
		if self.action in (ACTION.ADD_NEXT, ACTION.MOVE,
				ACTION.MOVE_KEYBOARD, ACTION.ROTATE): return
		if self.editor.lockEditing.get(): return
		if not self.parent.selectObjectsBodies(*self.editor.getMoveableSelection()): return

		cursor = self._geometry.cursor("cursor")
		if cursor is None:
			if not self.parent.objectsid: return
			self._geometry.cursor("object", self.parent.objectsid[0])
			cursor = self._geometry.cursor("cursor")

		self.parent.stopThreads()

		self.action  = ACTION.ROTATE
		self.parent.setStatus("Move mouse or Type coordinates e.g. 180", "Blue","White")
		self._cursor = self.cget("cursor")
		self.config(cursor=mouseCursor(self.action))

		self._viewer.edit("start",
				(self._viewer.pixel2uv(event.x, event.y), self.snap(event)))

		if self.parent.objectsid:
			self._geometry.object(self.parent.objectsid, "save", True)
		if self.parent.bodiesid:
			self._geometry.body(self.parent.bodiesid, "save", True)
			self._geometry.body(self.parent.bodiesid, "show",
					BIT.SELECT|BIT.WIREFRAME|BIT.MOVE)

			#dump("+++ actionRotateStart.draw", self.name(), "_viewer.draw(False)", level=3)
			self._viewer.draw(False, -1, True)
			self.parent.expose()

		# Set pivot point
		self._viewer.edit("pivot", self._viewer.xyz2uv(*cursor))

		matrix = self._viewer.matrix()
		self._axis = bmath.Vector(matrix[0][2], matrix[1][2], matrix[2][2])

		self._msg = "Rotate selected items"
		self.grab_set()

	# ----------------------------------------------------------------------
	def actionRotate(self, event):
		#dump("*-* GeometryViewer.actionRotate", level=2)
		if self.editor.lockEditing.get(): return
		if not self._anchorValue:
			self._viewer.edit("rotate", (
					self._viewer.pixel2uv(event.x, event.y),
					self._axis,
					self.snap(event)))
			self.parent.setStatus("Angle: %g"%(self._viewer.edit("rotate")),timeout=0)
		else:
			try: val = float(self._anchorValue)
			except: return
			self._viewer.edit("value", val, self._axis)

		if self.parent.objectsid:
			self._viewer.object(self.parent.objectsid, "rotate", self.__node)
		if self.parent.bodiesid:
			self._viewer.body(self.parent.bodiesid, "rotate")

		if self.parent.bodiesid:
			#dump("*-* GeometryViewer.actionRotate, project", self.name(), level=2)
			self._viewer.project(False, False)
		self.parent.draw(False)	# all viewports
		self.parent.expose()

	# ----------------------------------------------------------------------
	def actionMoveStop(self, event=None):
		dump("*-* GeometryViewer.actionMoveStop", level=2)
		if self.editor.lockEditing.get(): return
		self.grab_release()
		undoinfo = []
		refresh  = False
		if self.parent.bodies:
			refresh = True
		else:
			# If a 3D viewport is enabled!
			for view in self.viewports:
				if view._viewer.get("3D"):
					# and there is a light selected force a refresh
					if self.parent.objectLight:
						refresh = True
					break

		if refresh:
			undoinfo.append((self.flair().refreshUndo,))

		if self._clone:
			for x in self._anchorValue.split():
				if "*" in x:
					try:
						x = int(x.replace("*",""))
						if self.action == ACTION.MOVE:
							self._actionMoveClone(x, undoinfo)
						elif self.action == ACTION.ROTATE:
							self._actionRotateClone(x, undoinfo)
					except ValueError:
						pass
					break
				if "/" in x:
					try:
						x = -int(x.replace("/",""))
						if self.action == ACTION.MOVE:
							self._actionMoveClone(x, undoinfo)
						elif self.action == ACTION.ROTATE:
							self._actionRotateClone(x, undoinfo)
					except ValueError:
						pass
					break
			else:
				undoinfo.extend(self.actionUpdate())
		else:
			undoinfo.extend(self.actionUpdate())

		if undoinfo:
			if refresh:
				undoinfo.append((self.flair().refreshUndo,))
			else:
				self.editor.updateProperties()
			self.flair().addUndo(undoinfo, self._msg)
			self._msg = ""
		self.actionMoveFinalize()

	# ----------------------------------------------------------------------
	# Create clones of the moved objects
	# Multiply length if times > 0
	# Divide   length if times < 0
	# ----------------------------------------------------------------------
	def _actionMoveClone(self, times, undoinfo):
		dump("*-* GeometryViewer._actionMoveClone", level=2)
		self.parent.stopThreads()

		mx,my,mz = self._viewer.edit("move")

		if self.parent.closest[2] > 0:
			s = self._viewer.uv2xyz(*self._viewer.edit("start"))
			#log.say("s=",s)
			#val = []
			#for x in self._anchorValue.split():
			#	try: val.append(float(x))
			#	except ValueError: pass
			#if not self._anchorValue:
			uv  = self._viewer.pixel2uv(self.x, self.y)
			xyz = self._viewer.snap(uv[0], uv[1])
			#log.say("xyz=",xyz)
			sx  = xyz[0] - s[0]
			sy  = xyz[1] - s[1]
			sz  = xyz[2] - s[2]
			mx -= sx
			my -= sy
			mz -= sz
			#log.say("sxyz=",sx,sy,sz)
			#log.say("mxyz=",mx,my,mz)
		else:
			sx = sy = sz = 0.0

		if times > 0:
			# Multiply
			start = 0	# Skip first copy (is already there)
			stop  = times	# Do the rest
					# step is equal to move
		else:
			times = -times
			start = 0	# start from the first copy
			stop  = times	# until one before the last
			mx   /= float(times)	# divide step
			my   /= float(times)
			mz   /= float(times)

		for i in range(start, stop):
			dx = sx + mx * (i+1)
			dy = sy + my * (i+1)
			dz = sz + mz * (i+1)
			#log.say("dxyz=",dx,dy,dz)

			if i>0:
				# paste the original bodies still in buffer
				self.editor.selectClear(False)
				cards = self.flair().clipboard2Cards()
				self.editor._cardsPaste(cards,undoinfo)
				#pasted, renamed = self.editor._cardsPaste(cards,undoinfo)
				self.parent.selectObjectsBodies(*self.editor.getMoveableSelection())
				if self.parent.objectsid:
					self._geometry.object(self.parent.objectsid, "save", True)
				if self.parent.bodiesid:
					self._geometry.body(self.parent.bodiesid, "save", True)
			self._viewer.edit("value",[dx,dy,dz])
			self._actionDoMove()
			undoinfo.extend(self.actionUpdate())

	# ----------------------------------------------------------------------
	# Update the variables back to the object
	# ----------------------------------------------------------------------
	def actionUpdate(self):
		dump("*-* GeometryViewer.actionUpdate", self.action, level=2)
		cursorType = self._geometry.cursor("type")
		cursorId   = self._geometry.cursor("id")

		undoinfo = []
		for card in self.parent.objects:
			if card.tag in ("ROT-DEFI","ROTM"):
				x,y,z = self._geometry.object(card[Card.ID], "pos")
				card["x"] = x
				card["y"] = y
				card["z"] = z
			else:
				if self.action != ACTION.ADD_NEXT:
					undoinfo.append(self.flair().saveCardUndoInfo(card.pos()))
				for var,w in _OBJECT_MOVE_VARS[card.tag]:
					card.setWhatIfValue(w,
						bmath.format(self._geometry.object(card[Card.ID], var)))

				# additional treatment of spline
				if card.tag == "spline":
					option = self._geometry.object(card[Card.ID], "option")
					n = self._geometry.object(card[Card.ID], "nodes")
					card["node:0"] = n + (option & 1)	# +1 if closed
					for i in range(1,n+1):
						t,x,y,z = self._geometry.object(
								card[Card.ID], "node", i)
						card.setPropertyIfValue(f"node:{i}.t", t)
						card.setPropertyIfValue(f"node:{i}.x", x)
						card.setPropertyIfValue(f"node:{i}.y", y)
						card.setPropertyIfValue(f"node:{i}.z", z)
					if option&1 and n>2:	# if closed
						try:
							# <-- int
							t = self._geometry.object(card[
								Card.ID], "node", n)[0]+1
						except IndexError:
							t = 0.0
						card.setPropertyIfValue(f"node:{n+1}.t", t)
					card.setWhat(4, self._geometry.object(card[Card.ID], "option"))

		for card in self.parent.bodies:
			try:
				tag = self._geometry.body(card[Card.ID], "type")
			except TypeError:
				continue
			if self.action != ACTION.ADD_NEXT:
				undoinfo.append(self.flair().saveCardUndoInfo(card.pos()))
				if card.tag not in (tag, "VOXELS"):
					if Input._useBOX:
						self.flair().notify("Changed body type",
							f"Body '{card.name()}' changed type\n"
							f"{card.tag} \N{RIGHTWARDS ARROW} {tag}",
							tkFlair.NOTIFY_WARNING)
						card.input.changeTag(card,tag)
					elif tag not in ("BOX","WED","RAW","ARB"):
						card.input.changeTag(card,tag)
						self.flair().notify("Changed body type",
							f"Body '{card.name()}' changed type\n"
							f"{card.tag} \N{RIGHTWARDS ARROW} {tag}",
							tkFlair.NOTIFY_WARNING)
			what = [bmath.format(x, 15, includeSign=False)
					for x in self._geometry.body(card[Card.ID], "what")]
			if card.tag == "VOXELS":
				# Get it from the RPP
				if tag == "RPP":
					card.setWhatIfValue(1, what[0])
					card.setWhatIfValue(2, what[2])
					card.setWhatIfValue(3, what[4])
				else:	# BOX we have to assign a rotation
					card.setWhatIfValue(1, what[0])
					card.setWhatIfValue(2, what[1])
					card.setWhatIfValue(3, what[2])
			else:
				what.insert(0,card.name())
				if not Input._useBOX:
					if tag not in ("BOX","WED","RAW","ARB"):
						card.setWhatsIfValue(what)
				else:
					card.setWhatsIfValue(what)

		self.flair().inputModified()

		if self.parent.objectLight: self._geometry.setLights()

		# For move,rotate refresh everything
		if self.action != ACTION.ADD_NEXT:
			# FIXME could be more selective, e.g. restoring back
			# properties during editing of bodies etc.
			# and refresh anything but geometry editor
			self.flair().refresh("card") #, self.editor)	# refresh everything

		# Restore edit body
		try: self._geometry.cursor(cursorType, cursorId)
		except: pass
		return undoinfo

	# ----------------------------------------------------------------------
	# Create clones of the rotated objects
	# Multiply angle if times > 0
	# Divide   angle if times < 0
	# ----------------------------------------------------------------------
	def _actionRotateClone(self, times, undoinfo):
		dump("*-* GeometryViewer.actionRotateClone", level=2)

	# ----------------------------------------------------------------------
	def actionMoveFinalize(self):
		dump("*-* GeometryViewer.actionMoveFinalize", level=2)
		self.resetAnchor()
		self.config(cursor=self._cursor)
		self.action = ACTION.SELECT
		self._clone = False

	# ----------------------------------------------------------------------
	def actionMoveCancel(self, event=None):
		dump("*-* GeometryViewer.actionMoveCancel", level=2)
		self.grab_release()
		if self._clone:
			# Call undo for the last cloning
			log.say("FIXME actionMoveCancel: mustUndo")
		self._clone = False
		self._geometry.cursor("cursor", True)
		if self.parent.objectsid:
			self._geometry.object(self.parent.objectsid, "restore")
		if self.parent.bodiesid:
			self._geometry.body(self.parent.bodiesid, "restore")
		self.actionMoveFinalize()
		self._viewer.project(False, False)
		#dump("+++ actionMoveCancel.draw", self.name(), "_viewer.draw(False)", level=3)
		self._viewer.draw(False, -1, True)
		self.parent.expose()	# all viewports

	# ----------------------------------------------------------------------
	# Reset orientation of viewport
	# ----------------------------------------------------------------------
	def resetViewport(self):
		self._viewer.matrix(self.defSystem)
		self._sync = self.defSync

	# ----------------------------------------------------------------------
	# Reset anchor axis and value
	# ----------------------------------------------------------------------
	def resetAnchor(self, axis=-1):
		self._anchorAxis  = axis
		self._anchorValue = ""

	# ----------------------------------------------------------------------
	# Set and toggle anchor axis
	# ----------------------------------------------------------------------
	def setAnchorAxis(self, axis):
		a = ord(axis) - ord('x')
		if self._anchorAxis == a:
			self._anchorAxis  = -1
			self.parent.setStatus("Clear axis")
		else:
			self.parent.setStatus(f"Anchor on '{axis}' axis")
			self._anchorAxis = a
		self._anchorValue = ""

	# ----------------------------------------------------------------------
	# Return True is move has to be absolute on 2D plane or relative on 3D
	# ----------------------------------------------------------------------
	def moveAbsolute(self, event):
		if event.state & STATE.CONTROL:
			moveAbs = False
		else:
			moveAbs = True
		if self.parent.move3D:
			moveAbs = 1 - moveAbs
		return moveAbs

	# ----------------------------------------------------------------------
	# Show zone as the mouse moves
	# ----------------------------------------------------------------------
	def actionZoneShow(self, x, y, clear=True):
		u,v = self._viewer.pixel2uv(x, y)
		zone = self._viewer.zone("find",(u,v,BIT.SELFREEZE))
		if zone and not self._geometry.zone("has",zone):
			if clear: self._geometry.zone("clear")
			self._geometry.zone("select", zone)
			self.parent.draw()
			self.parent.setStatus(
				csg.toString(isinstance(self.input(),(Mcnp,Phits)),*zone),
				timeout=10000)

	# ----------------------------------------------------------------------
	# create a new zone definition on the location we clicked
	# ----------------------------------------------------------------------
	def actionZone(self, x, y):
		dump("*-* actionZone", x, y)
		u,v = self._viewer.pixel2uv(x, y)
		self.parent.stopThreads()
		zone = self._geometry.zone("select")
		self.parent.draw()
		if zone is None or not zone:
			self.parent.setStatus("-empty-")
			return

		# Check if zone already exists in REGION
		if self.action == ACTION.ADD_ZONEPAINT and \
		   self.editor.editType() == Card.REGION: # and not self.editObject.many:
			rid = self.editor.editObject.cards[0][Card.ID]
			if self._geometry.zone("contained",rid):
				return

		szone = csg.toString(isinstance(self.input(), (Mcnp,Phits)), *zone[0])

		if Global.verbose:
			xx,yy,zz = self._viewer.pixel2xyz(x, y)
			log.say(f"\nMatrix   =\n{str(bmath.Matrix(self._viewer.matrix()))}")
			log.say("Origin   =",self._viewer.origin())
			log.say("OriginUV =",self._viewer.offset())
			log.say("Zoom     =",self._viewer.zoom())
			log.say("Extends  =",self._viewer.extends())
			log.say("\nxi=",x)
			log.say("yi=",y)
			log.say("u=",repr(u))
			log.say("v=",repr(v))
			log.say("x=",repr(xx))
			log.say("y=",repr(yy))
			log.say("z=",repr(zz))
			log.say("Zone=", szone)

		self.parent.setStatus(szone)

		#if self.action == ACTION.ADD_ZONEPAINT:
		#	return
		if self.editor.editType() == Card.REGION:
			if self.action == ACTION.ADD_ZONE:
				self.editor.selectClear(False, Card.BODY)
			self.editor.addZone(szone)
		else:
			self.editor.addRegion(False, expr=szone)

	# ----------------------------------------------------------------------
	# Add a new node to the clicked location and start dragging
	# ----------------------------------------------------------------------
	def actionAddSplineNode(self, event):
		if self.editor.editType() != Card.OBJECT: return
		if self.editor.editObject.tag() != "spline": return
		closest = self._viewer.closest(event.x, event.y,
						CLOSE_DISTANCE,
						"O",
						self.editor.editObject.name())
		card = self.editor.editObject.cards[0]
		self.parent.selectObjectsBodies([card],None)
		oid = card[Card.ID]
		nodes = self._geometry.object(oid,"nodes")
		if closest is None:
			node = nodes	# append a new node at mouse location
		else:
			node = abs(closest[2])
		if node > nodes:
			# insert after
			node -= nodes
		elif node == 1:
			node = 0	# special case when clicked on first node
					# to insert before that

		P = bmath.Vector(self._geometry.object(oid, "pos"))
		xyz = bmath.Vector(self._viewer.pixel2xyz(event.x, event.y))
		if node == nodes:
			# Append node at the end
			self._geometry.object(oid,"node",xyz-P)
			# continue the add action until escape is pressed
			undoinfo = self.actionUpdate()
			self.flair().addUndo(undoinfo, f"Node added to {card.name()}")
		else:
			# and enter the move phase
			node += 1
			self._geometry.object(oid,"insert",(node,xyz-P))
			self.parent.closest = ("O", oid, node)
			self.actionMoveStart(event, node)
			#self.action =
		self.parent.expose()

	# ----------------------------------------------------------------------
	def actionDelSplineNode(self, event):
		if self.editor.editType() != Card.OBJECT: return
		if self.editor.editObject.tag() != "spline": return
		closest = self._viewer.closest(event.x, event.y,
						CLOSE_DISTANCE,
						"O",
						self.editor.editObject.name())
		card = self.editor.editObject.cards[0]
		self.parent.selectObjectsBodies([card],None)
		oid = card[Card.ID]
		nodes = self._geometry.object(oid,"nodes")
		if closest is None:
			node = nodes	# append a new node at mouse location
		else:
			node = abs(closest[2])
		if node > nodes:
			# insert after
			node -= nodes
		elif node == 1:
			node = 0	# special case when clicked on first node
					# to insert before that

		self._geometry.object(oid, "del", node)
		# continue the add action until escape is pressed
		undoinfo = self.actionUpdate()
		self.flair().addUndo(undoinfo, f"Node deleted from {card.name()}")
		self.parent.expose()

	# ----------------------------------------------------------------------
	def panLeft(self, event=None):
		x,y = self._viewer.offset()
		if event is not None and event.state & STATE.SHIFT:
			step = MOVESTEP2
		else:
			step = MOVESTEP
		x -= float(self.winfo_width())/step / self._viewer.Sx()
		self._viewer.offset(x,y)
		self.draw(late=True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def panRight(self, event=None):
		x,y = self._viewer.offset()
		if event is not None and event.state & STATE.SHIFT:
			step = MOVESTEP2
		else:
			step = MOVESTEP
		x += float(self.winfo_width())/step / self._viewer.Sx()
		self._viewer.offset(x,y)
		self.draw(late=True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def panTop(self, event=None):
		x,y = self._viewer.offset()
		if event is not None and event.state & STATE.SHIFT:
			step = MOVESTEP2
		else:
			step = MOVESTEP
		y += float(self.winfo_height())/step / self._viewer.Sy()
		self._viewer.offset((x,y))
		self.draw(late=True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def panBottom(self, event=None):
		x,y = self._viewer.offset()
		if event is not None and event.state & STATE.SHIFT:
			step = MOVESTEP2
		else:
			step = MOVESTEP
		y -= float(self.winfo_height())/step / self._viewer.Sy()
		self._viewer.offset((x,y))
		self.draw(late=True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def panFront(self, event=None):
		self._viewer.viewOto0()
		matrix = bmath.Matrix(self._viewer.matrix())
		width = float(self.winfo_width())/self._viewer.Sx()
		if event is not None and event.state & STATE.SHIFT:
			step = MOVESTEP2
		else:
			step = MOVESTEP
		# Perspective or Combo
		if self._viewer.get("3D") and self._viewer.get("projection"):
			width = max(abs(self._orbitDepth)/10.0,width)
		self._viewer.matrix(matrix*bmath.Matrix.translate(0.0, 0.0, width/2.0/step))
		self.drawFast3D()
		self.draw(True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def panBack(self, event=None):
		self._viewer.viewOto0()
		matrix = bmath.Matrix(self._viewer.matrix())
		width = float(self.winfo_width())/self._viewer.Sx()
		if event is not None and event.state & STATE.SHIFT:
			step = MOVESTEP2
		else:
			step = MOVESTEP
		# Perspective or Combo
		if self._viewer.get("3D") and self._viewer.get("projection"):
			width = max(abs(self._orbitDepth)/10.0,width)
		self._viewer.matrix(matrix*bmath.Matrix.translate(0.0, 0.0, -width/2.0/step))
			#-float(self.winfo_width())/step/2.0/self._viewer.Sx()))
		self.drawFast3D()
		self.draw(True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def rotateX(self, ang):
		self._viewer.viewOto0()
		self._viewer.matrix(bmath.Matrix(self._viewer.matrix())*bmath.Matrix.rotX(ang))
		self.drawFast3D()
		self.draw(True)

	# ----------------------------------------------------------------------
	def rotateY(self, ang):
		self._viewer.viewOto0()
		self._viewer.matrix(bmath.Matrix(self._viewer.matrix())*bmath.Matrix.rotY(ang))
		self.drawFast3D()
		self.draw(True)

	# ----------------------------------------------------------------------
	def rotateZ(self, ang):
		self._viewer.viewOto0()
		self._viewer.matrix(bmath.Matrix(self._viewer.matrix())*bmath.Matrix.rotZ(ang))
		self.drawFast3D()
		self.draw(True)

	# ----------------------------------------------------------------------
	# Increase rotation cursor size
	# ----------------------------------------------------------------------
	def cursorZoomIn(self, event):
		if self._geometry.cursor("cursor"):
			global CURSORSIZE
			CURSORSIZE = self._geometry.set("cursor")-5
			self._geometry.set("cursor", CURSORSIZE)
			self._viewer.expose()

	# ----------------------------------------------------------------------
	# Decrease rotation cursor size
	# ----------------------------------------------------------------------
	def cursorZoomOut(self, event):
		if self._geometry.cursor("cursor"):
			global CURSORSIZE
			CURSORSIZE = self._geometry.set("cursor")+5
			self._geometry.set("cursor", CURSORSIZE)
			self._viewer.expose()

	# ----------------------------------------------------------------------
	# ZoomIn or increase cursor size
	# ----------------------------------------------------------------------
	def cursorOrZoomIn(self, event):
		self.parent.setActive(self._viewer)
		if event.state & (STATE.SHIFT|STATE.CONTROL) == STATE.SHIFT|STATE.CONTROL and \
		   self._geometry.cursor("cursor"):
			self.cursorZoomIn(event)
		else:
			self.zoomIn(event)

	# ----------------------------------------------------------------------
	# ZoomOut or decrease cursor size
	# ----------------------------------------------------------------------
	def cursorOrZoomOut(self, event):
		self.parent.setActive(self._viewer)
		if event.state & (STATE.SHIFT|STATE.CONTROL) == STATE.SHIFT|STATE.CONTROL and \
		   self._geometry.cursor("cursor"):
			self.cursorZoomOut(event)
		else:
			self.zoomOut(event)

	# ----------------------------------------------------------------------
	def zoomIn(self, event):
		# Hold Shift to zoom on center!
		self.zoomPin(ZOOM, event.x, event.y, not bool(event.state & STATE.SHIFT))
		self.drawFast3D()
		self.draw(late=True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def zoomOut(self, event):
		# Hold Shift to zoom on center!
		self.zoomPin(1.0/ZOOM, event.x, event.y, not bool(event.state & STATE.SHIFT))
		self.drawFast3D()
		self.draw(late=True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def wheel(self, event):
		zoom = math.pow(ZOOM*(event.delta//120))
		self.zoomPin(zoom, event.x, event.y, not bool(event.state & STATE.SHIFT))
		self.drawFast3D()
		self.draw(late=True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	# Zoom around a pin point
	# ----------------------------------------------------------------------
	def zoomPin(self, zoomFactor, x, y, pin=True):
		if (PINZOOM and pin) or (not PINZOOM and not pin):
			pu, pv = self._viewer.pixel2uv(x, y)
			ou, ov = self._viewer.offset()
			du = (ou - pu) / zoomFactor
			dv = (ov - pv) / zoomFactor
			self._viewer.offset(pu + du, pv + dv)
		self._viewer.zoom(self._viewer.zoom() * zoomFactor)

	# ----------------------------------------------------------------------
	# Zoom on selected object
	# ----------------------------------------------------------------------
	def zoomOn(self, xmin, ymin, xmax, ymax, uv=None):
		# Origin
		if uv is None:
			self._viewer.offset(0.5*(xmin+xmax), 0.5*(ymin+ymax))
		else:
			self._viewer.offset(*uv)

		# Zoom and Window
		x1,y1,x2,y2 = self._viewer.extends()
		if abs(xmax-xmin) < 1.0e-8:
			zx = MAXZOOM
		else:
			zx = 0.5*(x2-x1)/(xmax-xmin)

		if abs(ymax-ymin) < 1.0e-8:
			zy = MAXZOOM
		else:
			zy = 0.5*(y2-y1)/(ymax-ymin)
		self._viewer.zoom(min(zx,zy))
		if self.drawFast3D():
			self.draw(True)
		else:
			self.draw(late=True)
			self.synchronizeViewports()

	# ----------------------------------------------------------------------
	def zoomOnError(self, err):
		if not 1<=err<=self._viewer.error(): return
		error = self._viewer.error("get",err)
		if err is None: return
		x,y,z,xmin,xmax,ymin,ymax,body,regIn,regOut = error
		self.zoomOn(xmin,ymin, xmax,ymax, self._viewer.xyz2uv(x,y,z))

	# ----------------------------------------------------------------------
	# Zoom on selected objects
	# ----------------------------------------------------------------------
	def zoomOnSelected(self, objects=None):
		if objects is None:
			objects = self.editor.getSelection()
			if objects is None: return

		cardTypes = {
			Card.BODY:   "B",
			Card.REGION: "R",
			Card.OBJECT: "O" }

		# Search in 2 and 3D
		bbox  = None
		bbox2 = None
		bbox3 = None

		for card in objects:
			t = cardTypes.get(card.type())
			if t is None: continue
			if card[Card.ID] is None: continue

			# First in 2D
			bb = self._viewer.bbox2D(t,card[Card.ID])
			if bbox2 is None:
				bbox2 = bb
			elif bb is not None:
				# make the union with bbox2
				bbox2[0] = min(bbox2[0], bb[0])
				bbox2[1] = min(bbox2[1], bb[1])
				bbox2[2] = max(bbox2[2], bb[2])
				bbox2[3] = max(bbox2[3], bb[3])

			# Geometry 3D bounding box
			bb = self._geometry.bbox(t,card[Card.ID])
			if bbox is None:
				bbox = bb
			elif bb is not None:
				# make the union with bbox
				bbox[0] = min(bbox[0], bb[0])
				bbox[1] = min(bbox[1], bb[1])
				bbox[2] = min(bbox[2], bb[2])
				bbox[3] = max(bbox[3], bb[3])
				bbox[4] = max(bbox[4], bb[4])
				bbox[5] = max(bbox[5], bb[5])

			# Viewer3D transformed for calculating viewing extends
			bb = self._viewer.bbox(t,card[Card.ID])
			if bbox3 is None:
				bbox3 = bb
			elif bb is not None:
				# make the union with bbox
				bbox3[0] = min(bbox3[0], bb[0])
				bbox3[1] = min(bbox3[1], bb[1])
				bbox3[2] = min(bbox3[2], bb[2])
				bbox3[3] = max(bbox3[3], bb[3])
				bbox3[4] = max(bbox3[4], bb[4])
				bbox3[5] = max(bbox3[5], bb[5])

		if bbox2 is not None:
			self.zoomOn(*bbox2)

		elif bbox is not None:
			x = (bbox[0] + bbox[3])/2.0
			y = (bbox[1] + bbox[4])/2.0
			z = (bbox[2] + bbox[5])/2.0
			self._viewer.offset(0.,0.)
			self._viewer.origin(x,y,z)

			# get limits
			du = (bbox3[3] - bbox3[0])/2
			dv = (bbox3[4] - bbox3[1])/2
			if du > 100000000000000 or dv > 100000000000000:
				# on infinite just show a message
				self.flair().notify("Infinite bounding box",
					"Cannot calculate bounding box",
					tkFlair.NOTIFY_WARNING)
				return
			self._viewer.zoom(1.0)
			self._viewer.extends(-du,-dv,du,dv)
			self.draw(True)
			self._viewer.expose()
			self.synchronizeViewports()

	# ----------------------------------------------------------------------
	# Center all viewports on mouse location
	# ----------------------------------------------------------------------
	def centerViewports(self, event, only3d=False):
		if self._viewer.get("3D"):
			u,v = self._viewer.pixel2uv(event.x, event.y)
			xyz = self._viewer.hit(u,v)
		elif self.snap(event):
			uv  = self._viewer.pixel2uv(event.x, event.y)
			xyz = self._viewer.snap(uv[0], uv[1])
		else:
			xyz = self._viewer.pixel2xyz(event.x, event.y)

		if self._sync:	# Like B2 click
			# center viewport to location
			self._viewer.viewOto0()
			u,v = self._viewer.pixel2uv(event.x, event.y)
			self._viewer.origin(*bmath.Vector(self._viewer.uv2xyz(u,v)))
			self.draw(True)

		for view in self.viewports:
			if not only3d and not view._viewer.get("3D") or \
			       only3d and     view._viewer.get("3D"):
				view._viewer.origin(xyz)
				view._viewer.expose()
				view.draw(True)
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	# motion event on orientation cube
	# ----------------------------------------------------------------------
	def ocubeMotion(self, event):
		oc = self._viewer.ocube("length")
		if OCUBE_ANIMATE:
			if oc >= self._ocubeBig:
				self._viewer.ocube("mouse",(event.x, event.y))
				self._viewer.expose()
			oc *= 5
			# Top-Right corner
			#if event.y < oc and event.x > self.winfo_width()-oc:
			# Bottom-Left corner
			if event.x < oc and event.y > self.winfo_height()-oc:
				# Inside corner
				if self._ocubeAction == OCUBE_ZOOMOUT:
					self.ocubeZoomIn()
					self.config(cursor=DEF_CURSOR)
			elif self._ocubeAction == OCUBE_ZOOMIN:
					self.ocubeZoomOut()
					self.config(cursor=self._cursor)
		else:
			oc *= 5
			# Inside corner of ocube
			# Top-Right corner
			#if event.y < oc and event.x > self.winfo_width()-oc:
			# Bottom-Left corner
			if event.x < oc and event.y > self.winfo_height()-oc:
				self._viewer.ocube("detail", True)
				self._viewer.ocube("mouse",(event.x, event.y))
				self._viewer.expose()
			else:
				self._viewer.ocube("detail", False)
				self._viewer.expose()

	# ----------------------------------------------------------------------
	# click event on ocube
	# ----------------------------------------------------------------------
	def ocubeClicked(self, event):
		im,ix,iy,iz = self._viewer.ocube("selection")
		#self._viewer.ocube("detail", False)
		#self.ocubeZoomOut()
		#dump("ocube selection=",im,ix,iy,iz, level=3)

		if im==OCUBE_NOWHERE:		# Mouse Nowhere
			return
		elif im==OCUBE_ONCUBE and ix>=0:	# Mouse On Cube
			self._ocubeClickedOnCube(ix,iy,iz)

		elif im==OCUBE_ONHOME:		# Mouse On Home
			# reset viewport on the default orientation
			matrix  = self._viewer.matrix()
			# Copy transformations only (no translation)
			for i in range(3):
				for j in range(3):
					matrix[i][j] = self.defSystem[i][j]
			self._viewer.matrix(matrix)

		elif im==OCUBE_ONORIGIN:		# Mouse on origin
			self._viewer.origin(0.0, 0.0, 0.0)

		elif im==OCUBE_ONCW:		# Mouse on clock wise
			self.rotateZ(ROTSTEP90)
			return

		elif im==OCUBE_ONCCW:		# Mouse on counter clock wise
			self.rotateZ(-ROTSTEP90)
			return

		elif im==OCUBE_ONAXIS:		# Mouse on counter clock wise
			self._ocubeClickedOnAxis(ix)

		elif im==OCUBE_ONLEFT:
			self.historyBackward()
			return

		elif im==OCUBE_ONRIGHT:
			self.historyForward()
			return

		else:
			raise Exception("Invalid return from ocube.selection")

		self.draw(True)

	# ----------------------------------------------------------------------
	# Set the length of the default ocube
	# ----------------------------------------------------------------------
	def ocubeSet(self, length):
		self._ocubeSmall = length
		self._ocubeBig   = self._ocubeSmall + OCUBE_NSTEPS * OCUBE_STEP * tkFlair.dpiscale
		self._ocubeBig   = min(40*tkFlair.dpiscale,max(24*tkFlair.dpiscale, self._ocubeBig))
		self._viewer.ocube("length", self._ocubeSmall)

	#-----------------------------------------------------------------------
	# Try to align depending on the type of item hit
	#   Corners    Centers    Edges (all other)
	#    0 0 0      1 1 0
	#    0 0 2      1 1 2
	#    0 2 0
	#    0 2 2      1 0 1
	#               1 2 1
	#    2 0 0
	#    2 0 2      0 1 1
	#    2 2 0      2 1 1
	#    2 2 2
	#
	# CENTER:
	#    try to keep as much as possible
	#    the orientation of the present system
	# EDGE:
	#-----------------------------------------------------------------------
	def _ocubeClickedOnCube(self, ix, iy, iz):
		if ix<0: return

		# We clicked on the cube rotate system
		matrix = bmath.Matrix(self._viewer.matrix())
		origin = bmath.Vector(self._viewer.origin())
		if self.frame.getInt("3D"):
			if self._viewer.get("projection") == 1:
				minw = -self._viewer.camera("focal")	# Perspective
			else:
				minw = 0.0				# Orthographic and Combo
			# Limit orbit distance to minw
			if self._orbitDepth < minw:
				self._orbitDepth = minw/2.0
			# Pan backward
			origin -= self._orbitDepth * bmath.Vector(self._viewer.basis('w'))

		# Current w
		w = bmath.Vector(matrix[0][2], matrix[1][2], matrix[2][2])

		# Values are from 0..2, subtract 1 to make them from -1, 0, 1
		# W vector is given by ix-1,iz-1,iw-1
		wn = bmath.Vector(ix-1, iy-1, iz-1)
		wn.normalize()

		# If are the same then do nothing
		#if abs(w*wn) > 0.999:
		#	dump("Do nothing")
		#	return

		# Find angle to rotate as well axis
		axis  = w ^ wn
		alen  = axis.normalize()
		if alen<0.001: return
		if alen<-1.:  alen = -1.
		elif alen>1.: alen =  1.
		angle = math.asin(alen)
		rot = bmath.Matrix()
		rot.rotate(angle, axis)
		matrix = rot * matrix

		# Find for each axis the azimuthal angle on the screen
		# and search for the minimal angle vs any of the axis vertical/horizontal up/down
		# FIXME maybe more wait should be given to the vertical than the horizontal axis
		# e.g. search of 60 - 30 degrees
		# Potentially I could click to the axis to align!!!!
		angle = 1000
		for i in range(3):
			if abs(matrix[i][2]) > 0.9999: continue	# ignore w vector
			phi = math.atan2(matrix[i][1], matrix[i][0])
			if phi < -math.pi * 0.75:
				phi = -math.pi - phi		# along -u
			elif phi < -math.pi * 0.25:
				phi = -math.pi/2 - phi	# along -v
			elif phi < math.pi * 0.25:
				phi = - phi			# along +u
			elif phi < math.pi * 0.75:
				phi = math.pi/2 - phi		# along +v
			else:
				phi = math.pi - phi		# along -u

			if abs(phi) < abs(angle):
				angle = phi

		rot.rotate(-angle, wn)
		matrix = rot * matrix

		# Fix elements to zero in case 1
		for i in range(3):
			for j in range(3):
				if abs(matrix[i][j])>0.99999999:
					for k in range(3):
						if k!=j: matrix[i][k] = 0
						if k!=i: matrix[k][j] = 0

		self._viewer.matrix(matrix)

		if self._viewer.get("3D"):
			# Pan Forward to orbit depth, restoring the pivot point
			w  = self._viewer.basis('w')
			x  = origin[0] + self._orbitDepth*w[0]
			y  = origin[1] + self._orbitDepth*w[1]
			z  = origin[2] + self._orbitDepth*w[2]
			self._viewer.origin((x,y,z))
		else:
			self._viewer.origin(origin)

	#-----------------------------------------------------------------------
	def _ocubeClickedOnAxis(self, ia):
		xyz    = self._viewer.origin()
		matrix = bmath.Matrix(self._viewer.matrix())
		w = bmath.Vector(matrix[0][2], matrix[1][2], matrix[2][2])

		if abs(matrix[ia][2]) > 0.9999: return	# ignore w vector
		phi = math.atan2(matrix[ia][1], matrix[ia][0])
		if phi < -math.pi * 0.75:
			phi = -math.pi - phi		# along -u
		elif phi < -math.pi * 0.25:
			phi = -math.pi/2 - phi	# along -v
		elif phi < math.pi * 0.25:
			phi = - phi			# along +u
		elif phi < math.pi * 0.75:
			phi = math.pi/2 - phi		# along +v
		else:
			phi = math.pi - phi		# along -u

		# Rotate matrix to align with the axis
		rot = bmath.Matrix()
		rot.rotate(-phi, w)
		matrix = rot * matrix

		# Fix elements to zero in case 1
		for i in range(3):
			for j in range(3):
				if abs(matrix[i][j])>0.99999999:
					for k in range(3):
						if k!=j: matrix[i][k] = 0
						if k!=i: matrix[k][j] = 0
		self._viewer.matrix(matrix)
		self._viewer.origin(xyz)

	# ----------------------------------------------------------------------
	def ocubeZoomIn(self):
		#dump("*-* GeometryViewer.ocubeZoomIn", level=3)
		self._ocubeAction = OCUBE_ZOOMIN
		self.animate("orientcube", OCUBE_TIMER)

	# ----------------------------------------------------------------------
	def ocubeZoomOut(self):
		#dump("*-* GeometryViewer.ocubeZoomOut", level=3)
		self._ocubeAction = OCUBE_ZOOMOUT
		self.animate("orientcube", OCUBE_TIMER)

	# ----------------------------------------------------------------------
	# Synchronize all viewports to the same origin and zoom
	# ----------------------------------------------------------------------
	def synchronizeViewports(self):
		if not self.parent._syncViewports or not self._sync: return
		xyz  = self._viewer.origin()
		Sx   = self._viewer.Sx()	# pixel scaling
		#Sy   = self._viewer.Sy()	# assume aspect 1	FIXME
		# FIXME check the viewports that are NOT rotated
		# otherwise it doesn't work
		for v in self.viewports:
			if not v._sync: continue
			# Check if viewport is on plane
			uvw = v._viewer.xyz2uvw(*xyz)
			# By setting the origin we are forced to re-draw.
			# it would be better to set the offset relative
			# to the motion
			# Don't set the "center" but move the offset
			# if possible check if we are on the same plane
			if abs(uvw[2])>1e-10:
				#print("uvw=",uvw)
				# We are not on the same plane move the origin
				v._viewer.origin(xyz)
				invalid = True
			else:
				v._viewer.offset(uvw[0], uvw[1])
				invalid = False

			# Set the zooming?
			if abs(Sx - v._viewer.Sx())>1e-10:
				z = Sx / v._viewer.Sx() * v._viewer.zoom()
				v._viewer.zoom(z)
				# Redraw as invalid (FIXME it should be more smart)
				invalid = True
			# FIXME Drawing starts generates many expose events
			# on other viewports and it shows some incomplete plots
			v.draw(invalid) #, expose=False)
		#for v in self.viewports:
		#	if not v._sync: continue
		#	v.expose()

	# ----------------------------------------------------------------------
	# Transform viewport to go to the lattice/prototype position
	# ----------------------------------------------------------------------
	def go2Lattice(self):
		# It has to be a REGION with a ROT-DEFI associated
		if self.editor.editType() != Card.REGION: return
		region = self.editor.editObject.cards[0]
		trans = self._geometry.region(region[Card.ID],"matrix")
		if not trans: return
		self.historyAdd()
		self._viewer.viewOto0()
		matrix = bmath.Matrix(self._viewer.matrix())
		self._viewer.matrix(bmath.Matrix(trans) * matrix)
		self.draw(True)

	# ----------------------------------------------------------------------
	# Display from the point of view of the camera
	# @param camera    card to set as viewing camera
	# @param fromobj   during the actionMove, use the object id
	#                  information instead of the card whats
	# ----------------------------------------------------------------------
	def loadCamera(self, camera=None, fromobject=False):
		#dump("*-* loadCamera", self.name(), camera, self.camera, level=3)
		if not fromobject:
			if camera is not None: self.camera = camera
			if self.camera is None: return

			P = bmath.Vector(self.camera.numWhat( 1),
					 self.camera.numWhat( 2),
					 self.camera.numWhat( 3))
			# w=Z on viewer is towards the viewer
			Z = P - bmath.Vector(self.camera.numWhat( 7),
					     self.camera.numWhat( 8),
					     self.camera.numWhat( 9))
			Y =  bmath.Vector(self.camera.numWhat(10),
					  self.camera.numWhat(11),
					  self.camera.numWhat(12))
			focal  = self.camera.numWhat(13)
		elif self.camera is None:
			return
		else:
			oid = self.camera[Card.ID]
			P = bmath.Vector(self._geometry.object(oid, "pos"))
			Z = P - bmath.Vector(	self._geometry.object(oid, "atx"),
						self._geometry.object(oid, "aty"),
						self._geometry.object(oid, "atz"))
			Y = bmath.Vector(	self._geometry.object(oid, "upx"),
						self._geometry.object(oid, "upy"),
						self._geometry.object(oid, "upz"))
			focal = self._geometry.object(oid, "focal")
		fov = math.radians(self.camera.numWhat(14))
		if fov <= 0.0: fov = math.radians(60)
		aspect = self.camera.numWhat(15)
		if aspect <= 0.0: aspect = 4./3

		if abs(Z.normalize()) < 1e-10: Z = bmath.Vector.Z
		if abs(Y.normalize()) < 1e-10: Y = bmath.Vector.Y.clone()
		if abs(abs(Z*Y)-1.0) > 1e-5:
			Y = Z.gramSchmidt(Y)
			Y.normalize()
			X = Y ^ Z
		else:
			# In case Z // (Y=up) then use previous information if any
			Y = Z.gramSchmidt(Y.orthogonal())
			Y.normalize()
			X = Y ^ Z

		M = bmath.Matrix(3)
		M.make(X,Y,Z)
		self._viewer.matrix(M.transpose())
		w2 = focal*math.tan(fov/2)
		h2 = w2/aspect

		# origin in viewer is at the focal plane
		self._viewer.origin(P - Z*focal)
		self._viewer.extends(-w2, -h2, w2, h2)
		self._viewer.camera("fov", fov)
		self._viewer.zoom(1.0)

	# ----------------------------------------------------------------------
	def resetCamera(self):
		self.camera = None

	# ----------------------------------------------------------------------
	# Views History
	# ----------------------------------------------------------------------
	def historyAdd(self):
		t = time.time()
		if self.curHistory>0 and t-self.history[-1][0]<2.0:
			del self.history[-1]
			self.curHistory -= 1

		self.curHistory += 1

		if self.curHistory == MAXHISTORY:
			del self.history[0]
			self.curHistory -= 1

		if 0 <= self.curHistory < len(self.history):
			del self.history[self.curHistory:]

		m = self._viewer.matrix()
		e = self._viewer.extends()
		o = self._viewer.offset()
		z = self._viewer.zoom()

		self.history.append((t,m,e,o,z))
		self.frame.historyButton(self.curHistory>0, False)

	# ----------------------------------------------------------------------
	def historyShow(self):
		t,m,e,o,z = self.history[self.curHistory]
		self._viewer.matrix(m)
		self._viewer.extends(e)
		self._viewer.offset(o)
		self._viewer.zoom(z)
		self.draw(True,True,False)
		self.synchronizeViewports()
		self.frame.historyButton(self.curHistory>0,
			self.curHistory<len(self.history)-1)

	# ----------------------------------------------------------------------
	def historyBackward(self):
		if self.curHistory > 0:
			self.curHistory -= 1
			self.historyShow()

	# ----------------------------------------------------------------------
	def historyForward(self):
		if self.curHistory < len(self.history)-1:
			self.curHistory += 1
			self.historyShow()

	# ----------------------------------------------------------------------
	# Export viewport
	# ----------------------------------------------------------------------
	def export(self, event=None):
		self.parent.setActionExport()
		self.parent.selectViewer(self)

	# ----------------------------------------------------------------------
	def _checkPIL(self):
		if PIL is None:
			self.flair().notify("Missing PIL module",
				"PIL libraries are missing.\nCannot export the image.",
				tkFlair.NOTIFY_ERROR)
			return False
		return True

	# ----------------------------------------------------------------------
	# Restore the viewer size
	# ----------------------------------------------------------------------
	def restoreSize(self):
		if self._viewer.size("isfixed"):
			self._viewer.size("restore")	# restore viewer size
			self.draw(True,True)
		else:
			self._viewer.image("restore")	# restore the image data

	# ----------------------------------------------------------------------
	# export viewport as image
	# ----------------------------------------------------------------------
	def exportImage(self, filename, width=None, height=None, option=None):
		if not self._checkPIL(): return
		self.flair().notify("Export started",
			f"Export to {filename} started\n"
			"Please wait the rendering to finish",
			tkFlair.NOTIFY_WARNING)
		self.parent.setStatus(f"Exporting to {filename}")
		self.__drawEndCmd = (self._exportImageEnd, filename, option)
		if width and height:
			self._viewer.size("set", width, height)
		self.draw(late=True)

	# ----------------------------------------------------------------------
	# Save view port into notes
	# ----------------------------------------------------------------------
	def toNotes(self, event=None, width=None, height=None):
		if not self._checkPIL(): return
		filename = f"{self.project().name()}-{self.name()}-{self.frame.layerCombo.get()}.png"
		self.flair().notify("Export started",
			"Export to notes started\n"
			"Please wait the rendering to finish",
			tkFlair.NOTIFY_WARNING)
		self.parent.setStatus("Exporting to notes")
		self.__drawEndCmd = (self._exportImageEnd, filename, None, True)
		if width and height:
			self._viewer.size("set", width, height)
		self.draw(late=True)

	#-----------------------------------------------------------------------
	# export image when drawing is finished
	#-----------------------------------------------------------------------
	def _exportImageEnd(self, filename, option, tonotes=False):
		data = self._viewer.image("get", None, option)
		image = PIL.Image.frombuffer("RGBA",
			self._viewer.size(),
			data, "raw", "RGBA", 0, 1)
		try:
			image.save(filename)
			rc = True
		except Exception as ex:
			self.flair().notify("Python Imaging Error",
				f"Unknown type: {ex}",
				tkFlair.NOTIFY_ERROR)
			rc = False
		del image
		del data
		self.restoreSize()
		if tonotes:
			link = f"{{img:{filename}}}"
			if link not in self.project().notes:
				self.project().notes += f"\n\n{link}"
			self.flair().modified()
			self.flair().refresh("image")
			self.flair().notify("Viewport to Notes",
				f"Viewport {self.name()} added to notes\n"
				f"with the name: {filename}")
		else:
			self.flair().notify("Export Image",
				f"Viewport exported to '{filename}'")
		return rc

	# ----------------------------------------------------------------------
	# export to postscript format
	# ----------------------------------------------------------------------
	def exportPostscript(self, filename, paper, landscape,
				halign=tk.CENTER, valign=tk.CENTER,
				width=None, height=None):
		w = self.winfo_width()
		h = self.winfo_height()
		if width is None and height is None:
			sx = sy = 1
			width  = sx
			height = sy
		elif width is not None and height is not None:
			sx = w/width
			sy = h/height
		elif width is not None:
			sx = sy = w/width
			height = h/sy
		else:
			sx = sy = h/height
			width = w/sx

		ps = Postscript(filename, paper, landscape)
		if paper == "EPS":
			ps.setPaperSize(width, height)
			ps.setMargin()
			ps.header = None
		else:
			ps.header = f"{self.project().name()}: {self.name()}"
		ps.footer = None
		ps.comment()
		ps.prolog()
		data = self._viewer.image("get")

		if halign == tk.CENTER:
			dx = (ps.margin_right - ps.margin_left - width)/2 + ps.margin_left
		elif halign == tk.LEFT:
			dx = ps.margin_left
		elif halign == tk.RIGHT:
			dx = ps.margin_right - width

		if valign == tk.CENTER:
			dy = (ps.margin_top - ps.margin_bottom - height)/2 + ps.margin_bottom
		elif valign == tk.BOTTOM:
			dy = ps.margin_bottom
		elif valign == tk.TOP:
			dy = ps.margin_top - height

		ps.translate(dx,dy)
		ps(w, h, 8)			# width, height, bit per pixel
		ps("[",sx, 0, 0, -sy, 0, h,"]")	# transformation matrix
		ps("{<")
		for i in range(0, len(data), 4):
			if i and i%32==0: ps.write("\n")
			ps.write(data[i:i+3].hex())	# skip A channel
		if len(data)%32: ps.write("\n")
		ps(">} false 3 colorimage")
		ps.close()

		self.flair().modified()
		self.flair().refresh("image")

	# ----------------------------------------------------------------------
	# Export movie by changing the 'frame' variable and export
	# one image at a time
	# ----------------------------------------------------------------------
	def exportMovie(self, filename, **args):
		if not self._checkPIL(): return
		args["frametime"] = time.time()
		args["starttime"] = time.time()

		self.log("+++ Creating frames:")
		frame = self.__skipFrames(filename, args["start"], args)
		if frame >= args["stop"]:
			self._createMovie(filename, args)
		else:
			self.parent.timeframeSet(frame)
			self.parent.timeframe.setCurrent(frame)
			self.parent.loadProject()
			self.loadCamera()
			self._viewer.size("set", args["width"], args["height"])
			self.__drawEndCmd = (self._exportMovieEnd, filename, args)
			self.draw(late=True)
		self.editor.showContextGroup("stop")

	#-----------------------------------------------------------------------
	# skip existing frames
	#-----------------------------------------------------------------------
	def __skipFrames(self, filename, frame, args):
		fn = os.path.splitext(filename)[0]
		while frame <= args["stop"]:
			framename = f"{fn}{frame:04}.png"
			if not os.path.isfile(framename):
				break
			self.log(f"Skip existing {framename}")
			frame += 1
		return frame

	#-----------------------------------------------------------------------
	# export image when drawing is finished
	#-----------------------------------------------------------------------
	def _exportMovieEnd(self, filename, args):
		self._viewer.expose()
		fn = os.path.splitext(filename)[0]
		frame = self.input().localDict["frame"]

		data = self._viewer.image("get", None)
		image = PIL.Image.frombuffer("RGBA",
			self._viewer.size(),
			data, "raw", "RGBA", 0, 1)

		framename = f"{fn}{frame:04}.png"
		try:
			image.save(framename)
			dt = time.time() - args["frametime"]
			tt = time.time() - args["starttime"]
			perc = 100*(frame-args["start"]) / (args["stop"]-args["start"]+1)
			remain = (args["stop"]-frame)*tt/(frame-args["start"]+1)

			self.log(f">>> {perc:.1f}% Image '{framename}' rendered in {dt:.2f}s")
			msg = f"{perc:.1f}% Movie '{filename}' " \
				f"completed in {tool.friendlyTime(tt)} " \
				f"Frame: {frame} " \
				f"ETA: {tool.friendlyTime(remain)}"
			self.parent.setStatusPercent(perc)
		except Exception as ex:
			self.flair().notify("Python Imaging Error",
				f"Unknown type: {ex}",
				tkFlair.NOTIFY_ERROR)
			frame = None
		del image
		del data

		if frame is not None:
			frame = self.__skipFrames(filename, frame+1, args)
			if frame >= args["stop"]:
				self._createMovie(filename, args)
				frame = None
			else:
				self.parent.timeframeSet(frame)
				self.parent.timeframe.setCurrent(frame)
				self.parent.loadProject()
				self.loadCamera()
				self.parent.setStatus(msg, "DarkRed", "LightYellow", 0)
				args["frametime"] = time.time()
				self.__drawEndCmd = (self._exportMovieEnd, filename, args)
				self.draw(late=True)

		if frame is None:
			self.parent.timeframeSet(args["start"])
			self.parent.timeframe.setCurrent(args["start"])
			self.parent.loadProject()
			self.loadCamera()
			self.restoreSize()
			self._viewer.expose()
			self.editor.showContextGroup()

	#-----------------------------------------------------------------------
	# Create movie using ffmpeg
	# Options:
	#	-y		yes to replace possible existing file to avoid stop
	#	-i filepattern	pattern of input files like file%04d.png
	#	-start_number #	starting frame number
	#	-vframes #	number of total frames stop-start+1
	#	-r fps		frame rate typically 24
	#	-crf #		quality for x264 1=best 10=good, 25=decent => higher worse
	#	-vcodec libx264	conversion codec
	#	-pix_fmt yuv420p pixel format
	#	output.mp4	default in mp4 with libx264
	#-----------------------------------------------------------------------
	def _createMovie(self, filename, args):
		self.log("<<< rendering ended")

		cmd = ["ffmpeg"]
		cmd.append("-y")
		cmd.append("-i")

		fn = os.path.splitext(filename)[0]
		cmd.append(f"{fn}%04d.png")
		cmd.append("-start_number")
		cmd.append(str(self.parent.timeframe.start))
		cmd.append("-vframes")
		cmd.append(str(self.parent.timeframe.nframes()))
		cmd.append("-r")
		cmd.append(str(args["fps"]))
		if args["ffopts"]:
			cmd.append(args["ffopts"])
		cmd.append(filename)
		cmd = " ".join(cmd)
		self.log("Converting to movie")
		self.log(f"Cmd: {cmd}")
		try:
			output = subprocess.check_output(cmd,
					stderr=subprocess.STDOUT,
					shell=True)
		except subprocess.CalledProcessError as ex:
			self.flair().notify("Movie Tool",
				f"Error {ex}",
				tkFlair.NOTIFY_ERROR)
			return

		for line in output.decode(errors="replace").splitlines():
			self.log(line)

		tt = time.time() - args["starttime"]
		self.flair().notify("Movie Tool",
			f"Movie {filename} completed in {tt:.1f}s",)
		self.parent.resetPercent()
		self.parent.setStatus(f"Movie: '{filename}' "
					f"Completed in {tt:.1f}s", "DarkGreen")

	# ---------------------------------------------------------------------
	# Save input&view to a special file to be used by the test program
	# ---------------------------------------------------------------------
	def exportNox(self, filename):
		fout = open(filename,"w")
		inp = self.input()
		inp.preprocess()
		rotdefi = inp.cardProperties()
		fout.write(f"mode {inp.MODE}\n")

		# Write matrix sub-function
		def writeMatrix(matrix):
			for row in range(4):
				for col in range(4):
					fout.write(" %.15g"%(matrix[row][col]))
			fout.write("\n")

		# Write Transformations
		fout.write("# Transformations\n")
		for key,matrix in rotdefi.items():
			if isinstance(matrix,int): continue
			fout.write(f"rotdefi {key}")
			writeMatrix(matrix)
		fout.write("\n")

		# --------------------
		#       Bodies
		# --------------------
		fout.write("# Bodies\n")
		hasVoxel = False
		matrix = [bmath.Matrix(4,type=1)]	# list of transformation matrices
		for card in inp.cardsCache(Card.BODY).values():
			if card is None: continue
			if card.tag == "VOXELS":
				if card.what(4):
					mat = inp.getTransformation(card.evalWhat(4),rotdefi)
					fout.write("matrix ")
					writeMatrix(mat)
				fout.write(f"body {card.tag} {card.name().strip()}.vxl "
					f"{card.numWhat(1)} {card.numWhat(2)} {card.numWhat(3)}\n")
				hasVoxel = True
			elif card.tag=="END":
				continue
			else:
				# Add the body
				matrix = card[Card.MATRIX]
				if matrix:
				#if len(matrix)>1:
					try:
						fout.write("matrix ")
						writeMatrix(matrix)
					except:
						pass
						#breakpoint()
				fout.write("body %s %s %s\n"%(card.tag, card.name(),
					" ".join([str(card.numWhat(x))
							for x in range(1, card.nwhats())])))

			#elif card.tag[0]=="$":
			#	if card.tag=="$start_translat":
			#		matrix.append(matrix[-1]*bmath.Matrix.translate(
			#			card.numWhat(0), card.numWhat(1), card.numWhat(2)))
			#	elif card.tag=="$start_transform":
			#		# To be compatible with FLUKA
			#		matrix.append(matrix[-1] *
			#			inp.getTransformation(card.evalWhat(0),rotdefi))
			#	elif card.tag=="$start_expansion":
			#		matrix.append(matrix[-1]*bmath.Matrix.scale(card.numWhat(0)))
			#	elif card.tag[:5]=="$end_":
			#		if len(matrix)>1:
			#			del matrix[-1]
			#		else:
			#			self.parent.addError("Too many $end_xxx cards")
		fout.write("\n")

		# --------------------
		#       Regions
		# --------------------
		def writeRegion(region,expr):
			rot = signedRotdefi(rotdefi, region["@rotdefi"])
			color = self._viewer.region(region.name(),"color")
			alpha = self._viewer.region(region.name(),"alpha")
			sex = '\n\t'.join(expr.splitlines())
			if rot:
				fout.write("matrix ")
				writeMatrix(inp.getTransformation(rot, rotdefi))
			fout.write(f"region {region.name()} {region[Card.TYPE]} {rot} "
				f"{color} {alpha} NORM {sex}\n")

		fout.write("# Regions\n")
		expr   = None
		color  = None
		region = None
		for card in inp.cardsSorted(Card.REGION,-1).values():
			if card.name() == "&": # append to previous expression
				expr += card.expr()
				continue
			if expr is not None:
				writeRegion(region,expr)
			# move old card to region and process new card
			region = card
			expr   = card.expr()

		# Add last region
		writeRegion(region,expr)
		fout.write("\n")

		# Write voxel information
		if hasVoxel:
			fout.write("# Voxel information\n")
			for i in range(1,self._geometry.voxel("nregions")+1):
				color = self._viewer.voxel("color", i)
				fout.write(f"voxel color {i} {color}\n")
			fout.write("\n")

		# Write view
		fout.write("# View\n")
		fout.write("viewmatrix ")
		writeMatrix(self._viewer.matrix())
		fout.write("extends %g %g %g %g\n"%self._viewer.extends())
		fout.write("zoom %g\n"%self._viewer.zoom())
		fout.write("origin %g %g %g\n"%self._viewer.origin())

		# Write view
		fout.close()
		return True

	# ----------------------------------------------------------------------
	# Export to postscript and send to printer
	# ----------------------------------------------------------------------
	def hardcopy(self, event=None, width=None, height=None):
		printer = tkFlair.printer(self.parent)
		if not printer: return
		printer.open()
		self.exportPostscript(printer, printer.paper, printer.landscape,
				width=width,
				height=height)
		printer.close()

	# ---------------------------------------------------------------------
	# Scan other viewports for errors along one axis, stop on first error
	#
	# FIXME should be done in parallel and stop on errors
	# ---------------------------------------------------------------------
	def scan4Errors(self):
		# FIXME somehow the user should enter the requested axis!
		dump("*-* scanning geometry for errors")
		axis = 'x'
		idx = ord(axis)-ord('x')
		origins = [list(x._viewer.origin()) for x in self.viewports]
		vertices = self._viewer.vertex(axis)
		if len(vertices)<2: return
		a = vertices[0]
		for b in vertices[1:]:
			m = 0.5*(a+b)
			for o,v in zip(origins, self.viewports):
				o[idx] = m
				v._viewer.origin(o)
				v.draw(True)
			a = b
		for o,v in zip(origins, self.viewports):
			v._viewer.origin(o)
			v.draw(True)

	# ---------------------------------------------------------------------
#	def _scan4ErrorsOnAxis(self, axis, vertices):
#		origin = list(self._viewer.origin())
#		a = ord(axis)-ord('x')
#		for v in vertices:
#			origin[a] = v
#			self._viewer.origin(origin)
#			self.draw(True)
#			# wait to finish and check for errors

	# ---------------------------------------------------------------------
	# Escape, cancel last action progressively
	# ---------------------------------------------------------------------
	def escape(self, event=None):
		if self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD, ACTION.ROTATE) and \
		   self._anchorValue != "":
			self._anchorValue = ""
			# Call action Move or Rotate to update
			if self.action == ACTION.ROTATE:
				self.actionRotate(event)
			else:
				self.actionMove(event)
			return
		self.grab_release()

		# 1st stop current Action
		if self.action != ACTION.SELECT or self.parent.mouseAction < 0:
			if self.parent.mouseAction in (ACTION.ADD_ZONE, ACTION.ADD_ZONEPAINT):
				self._geometry.zone("clear")
			self.actionStop()
			self.parent.setStatus("Cancel action")

		# 2nd stop any pending action
		elif self.parent.mouseAction in (ACTION.ADD, ACTION.ADD_NEXT, ACTION.EDIT, ACTION.ADD_ZONEPAINT):
			if self.parent.mouseAction in (ACTION.ADD_ZONE, ACTION.ADD_ZONEPAINT):
				self._geometry.zone("clear")
			self.editor.setMouseAction(ACTION.SELECT)
			self.parent.setStatus("Cancel action")

		# 3rd if a region is selected
		elif self.editor.editType() == Card.REGION:
			# 3.1 clear property list selection if any
			if self.editor.propList.curselection():
				self.editor.propList.selection_clear(0,tk.END)
				self.editor.propListSelect()
				self._geometry.zone("clear")
				self.parent.setStatus("Properties cleared")

			# 3.2 clear objects
			elif self.editor.isAnythingSelected(Card.OBJECT):
				self.editor.selectClear(True,Card.OBJECT)
				self.parent.setStatus("Objects unselected")

			# 3.3 clear bodies
			elif self.editor.isAnythingSelected(Card.BODY):
				self.editor.selectClear(True,Card.BODY)
				self.parent.setStatus("Bodies unselected")

			# 3.4 clear everything else
			elif self.editor.isAnythingSelected():
				self.editor.selectClear()
				self.parent.setStatus("Unselected all")

		# 4rd unselect anything selected
		elif self.editor.isAnythingSelected():
			self.editor.selectClear()
			self.parent.setStatus("Unselected all")

		# 5th do nothing
		else:
			return

		self.editor.selectShow(False)
		self.parent.updateEdit()
		self.parent.draw()
		self.parent.expose()

	# ---------------------------------------------------------------------
	# Handle keyboard commands
	# ---------------------------------------------------------------------
	def handleKey(self, event):
		self.parent.setActive(self._viewer)

		if event.keysym in ("Shift_L", "Shift_R", "Alt_L", "Alt_R"):
			return

		# Cancel action
		elif event.keysym == "Escape":
			self.escape(event)

		# Move
		elif event.keysym == "Left":
			if self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD) and \
			   event.state & STATE.CTRLSHIFT:
				dd = 0.1 if event.state & STATE.SHIFT else 1
				self.actionMoveBy(event, -dd, 0)
			elif event.state & STATE.CONTROL:
				if event.state & STATE.SHIFT:
					ang = ROTSTEP90
				else:
					ang = ROTSTEP
				self.rotateY(ang)
			else:
				self.panLeft(event)

		elif event.keysym == "Right":
			if self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD) and event.state & STATE.CTRLSHIFT:
				dd = 0.1 if event.state & STATE.SHIFT else 1
				self.actionMoveBy(event, dd, 0)
			elif event.state & STATE.CONTROL:
				if event.state & STATE.SHIFT:
					ang = ROTSTEP90
				else:
					ang = ROTSTEP
				self.rotateY(-ang)
			else:
				self.panRight(event)

		elif event.keysym == "Up":
			if self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD) and event.state & STATE.CTRLSHIFT:
				dd = 0.1 if event.state & STATE.SHIFT else 1
				self.actionMoveBy(event, 0, dd)
			elif event.state & STATE.CONTROL:
				if event.state & STATE.SHIFT:
					ang = ROTSTEP90
				else:
					ang = ROTSTEP
				self.rotateX(ang)
			else:
				self.panTop(event)

		elif event.keysym == "Down":
			if self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD) and event.state & STATE.CTRLSHIFT:
				dd = 0.1 if event.state & STATE.SHIFT else 1
				self.actionMoveBy(event, 0, -dd)
			elif event.state & STATE.CONTROL:
				if event.state & STATE.SHIFT:
					ang = ROTSTEP90
				else:
					ang = ROTSTEP
				self.rotateX(-ang)
			else:
				self.panBottom(event)

		elif event.keysym == "Prior":
			if event.state & STATE.CONTROL:
				if event.state & STATE.SHIFT:
					ang = ROTSTEP90
				else:
					ang = ROTSTEP
				self.rotateZ(-ang)	# counter clock wise
			else:
				self.panFront(event)

		elif event.keysym == "Next":
			if event.state & STATE.CONTROL:
				if event.state & STATE.SHIFT:
					ang = ROTSTEP90
				else:
					ang = ROTSTEP
				self.rotateZ(ang)	# clock wise
			else:
				self.panBack(event)

		elif self.action in (ACTION.MOVE, ACTION.MOVE_KEYBOARD, ACTION.ROTATE,
					ACTION.ADD_NEXT, ACTION.ORBIT):
			self._handleKeyMove(event)

		else:
			self._handleKeyNormal(event)

	# ---------------------------------------------------------------------
	# Handle key method for all other actions
	# ---------------------------------------------------------------------
	def _handleKeyNormal(self, event):
		invalid = False

		if event.keysym == "space":
			if event.state & STATE.CONTROL:
				self.popupMenu(event)
			else:
				self.parent.insertMenu()
			return

		elif event.keysym == "Menu":
			self.popupMenu(event)
			return

		elif event.keysym == 'a':
			# with or without control
			self.editor.selectAll()

		elif event.keysym == 'A':
			# with or without control
			self.editor.selectClear()

		elif event.char == 'b':
			self.parent.insertBodyMenu()

		elif event.char == 'B':
			if self.parent.actionOption:
				self.parent.newBody()
			else:
				self.parent.insertBodyMenu()

		elif event.char == 'd':
			if self.parent.actionTool is self.parent.tool[ACTION.ZONE]:
				self.parent.actionTool.setAction(ACTION.ADD_ZONE)
			else:
				self.editor.setMouseAction(ACTION.ZONE)
			self.actionZoneShow(event.x, event.y)
			return

		elif event.char == 'D':
			if self.parent.actionTool is self.parent.tool[ACTION.ZONE]:
				self.parent.actionTool.setAction(ACTION.ADD_ZONEPAINT)
			else:
				self.editor.setMouseAction(ACTION.ZONE)
				self.parent.actionTool.setAction(ACTION.ADD_ZONEPAINT)
			self.actionZoneShow(event.x, event.y)
			return

		elif event.char == 'G':
			self.editor.toggleSnap2Grid()

		elif event.char == 'c':
			self.centerViewports(event, False)
			return

		elif event.char == 'C':
			self.centerViewports(event, True)
			return

		elif event.keysym == 'e':
			if Global.developer and event.state & STATE.CONTROL:
				self.scan4Errors()
			else:
				self.editor.editProperty()
			return

		elif event.char == 'E':
			self.parent.toggleAction(ACTION.EDIT)
			return

		elif event.char == 'f':	# find
			self.zoomOnSelected()
			return

		elif event.char == 'F':
			self.editor.toggleFreeze()

		elif event.char == 'g':
			self.parent.closest = None
			self.resetAnchor()
			if self.actionMoveStart(event, 0):
				self.actionMove(event)
			else:
				# FIXME move viewport
				pass
			return

		elif event.char == 'i':
			self.parent.toggleAction(ACTION.INFO)
			return

		elif event.char == 'k':
			self.parent.timeframeToggle()

		elif event.char == 'K':
			if Global.developer:
				log.say("\nMemory Information:",self.name())
				self._geometry.memory("dump")
				self._viewer.memory("dump")
				# Self destroy for MEMORY leak testing
				# needs to be compiled with MEM=yes
				try:
					for view in self.parent._viewers:
						view._destroy()
					self._geometry._destroy()
				except AttributeError:
					pass
				return

		elif event.char == 'l':
			self.go2Lattice()

		elif event.char == 'L':
			self.editor.toggleLock()

		elif event.char == 'm':
			self.editor.editMaterial()

		elif event.char == 'n' or event.keysym == "F2":
			self.editor.editName()
			return

		elif event.char == 'o':
			self.frame.toggleOrigin()

		elif event.char == 'O':
			self.frame.nextOrigin()

		elif event.char == 'p':
			self.parent.newPoint()
			return

		elif event.char == 'P':
			self.parent.setMouseAction(ACTION.PIVOT)
			return

		elif event.char == 'r':		# rotate object
			self.resetAnchor()
			self.actionRotateStart(event)
			return

		elif event.char == 'R':
			self.editor.addRegion()
			return

		elif event.char == 's':
			self.parent.toggleAction(ACTION.SELECT)
			return

		elif event.char == 'S':
			self.editor.toggleSynchronizeViewports()
			return

		elif event.char == 't':
			self.parent.toggleAction(ACTION.ORBIT)
			return

		elif event.char == 'T':
			self.editor.toggleAlpha()

		elif event.char == 'v':
			self.editor.toggleVisibility()

		elif event.char == 'w':
			self.parent.rotateLayout()

		elif event.char == 'W':
			self.parent.customLayout()

		elif event.char == 'x':
			self.parent.toggleAction(ACTION.PAN)
			return

		elif event.char == 'X':
			self.frame.rotateLayer()

		elif event.char == 'z':
			self.parent.toggleAction(ACTION.ZOOM_IN)
			return

		elif event.char == 'Z':
			self.parent.toggleAction(ACTION.ZOOM_OUT)
			return

		elif event.keysym in ("Insert", "Return", "KP_Enter"):
			self.parent.insertMenu()
			return

		elif event.keysym in ("Delete", "BackSpace", "KP_Delete"):
			self.editor.delete()

		elif event.char == '.':
			self.parent.repeatLastAction()
			return

		elif event.char == "#":
			self.parent.layout("2x2")

		elif event.char == "%":
			self.editor.toggle3D()
			return

		# Zoom
		elif event.char in ('=','+'):
			self.zoomIn(event)
			return

		elif event.char in ('-','_'):
			self.zoomOut(event)
			return

		elif event.char == '<':
			self.historyBackward()
			return

		elif event.char == '>':
			self.historyForward()
			return

		# Views
		elif event.keysym == '1' and event.state & STATE.CONTROL:
			xyz = self._viewer.origin()
			self._viewer.matrix(FRONT_MATRIX)
			self._viewer.origin(xyz)
			self._sync = False
			invalid = True

		elif event.keysym == '2' and event.state & STATE.CONTROL:
			xyz = self._viewer.origin()
			self._viewer.matrix(BACK_MATRIX)
			self._viewer.origin(xyz)
			self._sync = False
			invalid = True

		elif event.keysym == '3' and event.state & STATE.CONTROL:
			xyz = self._viewer.origin()
			self._viewer.matrix(DOWN_MATRIX)
			self._viewer.origin(xyz)
			self._sync = False
			invalid = True

		elif event.keysym == '4' and event.state & STATE.CONTROL:
			xyz = self._viewer.origin()
			self._viewer.matrix(UP_MATRIX )
			self._viewer.origin(xyz)
			self._sync = False
			invalid = True

		elif event.keysym == '5' and event.state & STATE.CONTROL:
			xyz = self._viewer.origin()
			self._viewer.matrix(LEFT_MATRIX)
			self._viewer.origin(xyz)
			self._sync = False
			invalid = True

		elif event.keysym == '6' and event.state & STATE.CONTROL:
			xyz = self._viewer.origin()
			self._viewer.matrix(RIGHT_MATRIX)
			self._viewer.origin(xyz)
			self._sync = False
			invalid = True

		elif (event.keysym == '0' and event.state & STATE.CONTROL) or \
		     event.keysym == "Home":		# center screen
			self._viewer.origin(0.0, 0.0, 0.0)
			invalid = True

		else:
			if Global.developer:
				if event.keysym == "f" and event.state & STATE.CONTROL:
					dump("+++ FPE ON")
					self._geometry.fpe(True)
				elif event.keysym == "F" and event.state & STATE.CONTROL:
					dump("+++ FPE off")
					self._geometry.fpe(False)
				Global.dump("??? GeometryViewer::handleKey: char=%s keysym=%s (%X) state=%X" %
					(event.char, event.keysym, event.keycode, event.state))
			return

		# FIXME late= not invalid
		# should check every action if it needs a late or note drawing
		self.draw(invalid, not invalid)
		self.synchronizeViewports()

	# ---------------------------------------------------------------------
	# Handle key method for move actions
	# ---------------------------------------------------------------------
	def _handleKeyMove(self, event):
		if event.char == 'G':
			self.editor.toggleSnap2Grid()

		#elif event.char == 'u':
			#self.setAnchorAxis(event.char)
		#elif event.char == 'v':
			#self.setAnchorAxis(event.char)
		#elif event.char == 'w':
			#self.setAnchorAxis(event.char)
		elif event.char == 'x':
			self.setAnchorAxis(event.char)
		elif event.char == 'y':
			self.setAnchorAxis(event.char)
		elif event.char == 'z':
			self.setAnchorAxis(event.char)

		# Zoom
		elif event.char and event.char in "0123456789.+-eE ":
			self._anchorValue += event.char
			self.parent.setStatus(f"Value: {self._anchorValue}","Blue","White")

		elif event.char in ("*","/"):
			if self._anchorValue and self._anchorValue[-1] != ' ':
				self._anchorValue += ' '	# add a space
			self._anchorValue += event.char
			self.parent.setStatus(f"Value: {self._anchorValue}","Blue","White")

		elif event.char == ",":
			self._anchorValue += ' '	# replace to space
			self.parent.setStatus(f"Value: {self._anchorValue}","Blue","White")

		elif event.keysym  == "BackSpace":
			if self._anchorValue:
				self._anchorValue = self._anchorValue[:-1]
				self.parent.setStatus(f"Value: {self._anchorValue}","Blue","White")

		elif event.keysym in ("Return", "KP_Enter"):
			# Equivalent to release button
			self.buttonRelease(event)
			return

		elif Global.verbose:
			log.say("GeometryViewer::handleKeyMove: char=%s keysym=%s state=%X" %
				(event.char, event.keysym, event.state))

		# Call action Move or Rotate to update
		if self.action == ACTION.ROTATE:
			self.actionRotate(event)
		else:
			self.actionMove(event)

	# ----------------------------------------------------------------------
	# Menus
	# ----------------------------------------------------------------------
	def popupMenu(self, event):
		menu = tk.Menu(self, tearoff=0)

		menu.add_command(label="Repeat Last Action", underline=0,
				command=self.parent.repeatLastAction,
				accelerator=".",
				compound=tk.LEFT, image=icons.get("refresh"))

		# --- Std cut & copy & paste menus ---
		menu.add_command(label="Cut", underline=2,
				command=self.editor.cut,
				accelerator="Ctrl-X",
				compound=tk.LEFT, image=icons.get("cut"))
		menu.add_command(label="Copy", underline=0,
				command=self.editor.copy,
				accelerator="Ctrl-C",
				compound=tk.LEFT, image=icons.get("copy"))
		menu.add_command(label="Paste", underline=0,
				command=self.editor.paste,
				accelerator="Ctrl-V",
				compound=tk.LEFT, image=icons.get("paste"))
		menu.add_command(label="Clone", underline=1,
				command=self.editor.clone,
				accelerator="Ctrl-D",
				compound=tk.LEFT, image=icons.get("clone"))
		menu.add_separator()

		# --- Add Menus ---
		submenu = tk.Menu(menu)
		menu.add_cascade(label="Add", menu=submenu, underline=0,
				compound=tk.LEFT, image=icons.get("add"),
				accelerator="Space")
		self.parent._addMenus(submenu)

		menu.add_command(label="Delete", underline=1,
				command=self.editor.delete,
				accelerator="Del",
				compound=tk.LEFT, image=icons.get("x"))

		# --- General ---
		menu.add_separator()
		if self.editor.editType() == Card.REGION and \
		   self._geometry.region(self.editor.editObject.cards[0][Card.ID],"hasmatrix"):
			menu.add_command(label="Goto Lattice", underline=0,
					command=self.go2Lattice,
					accelerator="l",
					compound=tk.LEFT, image=icons.get("LATTICE"))

		menu.add_command(label="Zoom On", underline=0,
				command=self.zoomOnSelected,
				accelerator="f",
				compound=tk.LEFT, image=icons.get("zoom_on"))

		menu.add_separator()
		self.parent._visibilityMenu(menu)
		self.parent._wireframeMenu(menu)
		self.parent._freezeMenu(menu)

		menu.tk_popup(event.x_root, event.y_root)

#===============================================================================
# Geometry frame
#===============================================================================
class GeometryFrame(tk.Frame):
	# ----------------------------------------------------------------------
	def __init__(self, master, parent, name, **kw):
		super().__init__(master, **kw)
		self.parent   = parent
		self.name     = name
		self.defLayer = "Media"

		self.viewer = GeometryViewer(self, parent, **kw)
		self.viewer.pack(side=tk.BOTTOM, expand=tk.YES, fill=tk.BOTH)
		self.createToolbar()

		# Remember loaded image
		self._clearImage()

	# ----------------------------------------------------------------------
	# FIXME convert to properties
	# ----------------------------------------------------------------------
	#@property
	def flair(self):	return self.parent.editor.flair
	#@property
	def project(self):	return self.parent.project
	#@property
	def input(self):	return self.parent.project.input

	# ----------------------------------------------------------------------
	def createToolbar(self):
		toolbar = tk.Frame(self, relief=tk.RAISED, takefocus=0, borderwidth=2)
		toolbar.pack(side=tk.TOP, fill=tk.X)
		self.maxWindow = tk.IntVar()

		self.comboButton = tk.Button(toolbar,image=icons.get("axes"),
				text=" \N{BLACK DOWN-POINTING TRIANGLE}",
				compound=tk.LEFT,
				command=self.changeOrigin,
				padx=0, pady=0)
		self.comboButton.grid(row=0, column=0, sticky=tk.N+tk.EW)
		self._lastOrigin = 1
		tkExtra.Balloon.set(self.comboButton, "Origin and projection [o]")

		self._bar = []
		self._bar.append(GeometryToolbar._NavigationFrame(toolbar,self))
		self._bar.append(GeometryToolbar._OriginFrame(toolbar,    self))
		self._bar.append(GeometryToolbar._BasisFrame(toolbar,     self))
		self._bar.append(GeometryToolbar._ExtendsFrame(toolbar,   self))
		self._bar.append(GeometryToolbar._MoveFrame(toolbar,      self))
		self._bar.append(GeometryToolbar._EulerFrame(toolbar,     self))
		self._bar.append(GeometryToolbar._TransformFrame(toolbar, self))
		self.lastFrame = None
		self.showOrigin(self._bar[0])

		# FIXME dirty hack
		self.layerCombo    = self._bar[0].layerCombo
		self.bookmarkCombo = self._bar[0].bookmarkCombo

		# max button
		self.maxButton = tk.Checkbutton(toolbar, text="\N{BLACK UP-POINTING TRIANGLE}",
						command=self.maximizeWindow,
						variable=self.maxWindow,
						indicatoron=tk.FALSE,
						highlightthickness=0,
						padx=2, pady=0)
		self.maxButton.grid(column=2, row=0, pady=2, sticky=tk.NE)
		tkExtra.Balloon.set(self.maxButton, "Maximize viewport [W]")

		toolbar.grid_columnconfigure(1, weight=1)

	# ----------------------------------------------------------------------
	def changeOrigin(self):
		menu = tk.Menu(self, tearoff=0)
		for frame in self._bar:
			if frame is self.lastFrame:
				state = tk.DISABLED
			else:
				state = tk.NORMAL
			menu.add_command(label=frame.name,
					image=icons.get(frame.icon),
					state=state,
					compound=tk.LEFT,
					command=lambda s=self,f=frame:s.showOrigin(f))
		menu.tk_popup(self.comboButton.winfo_rootx(),
			      self.comboButton.winfo_rooty() + self.comboButton.winfo_height())

	# ----------------------------------------------------------------------
	def toggleOrigin(self):
		idx = self._bar.index(self.lastFrame)
		if idx == 0:
			idx = self._lastOrigin
		else:
			idx = 0
		self.showOrigin(self._bar[idx])

	# ----------------------------------------------------------------------
	def nextOrigin(self):
		idx = self._bar.index(self.lastFrame) + 1
		if idx >= len(self._bar): idx = 0
		self.showOrigin(self._bar[idx])

	# ----------------------------------------------------------------------
	def showOrigin(self, frame):
		if self.lastFrame: self.lastFrame.grid_forget()
		self.comboButton["image"] = icons.get(frame.icon)
		frame.grid(row=0, column=1, sticky=tk.EW)
		frame.fill()
		self.lastFrame = frame
		idx = self._bar.index(self.lastFrame)
		if idx>0: self._lastOrigin = idx

	# ----------------------------------------------------------------------
	def loadOrigin(self):
		for frame in self._bar:
			frame.load()

	# ----------------------------------------------------------------------
	def fillOrigin(self):
		self.lastFrame.fill()

	# ----------------------------------------------------------------------
	def setColor(self, color):
		hexcolor = f"#{color:06X}"
		self.maxButton.config(background=hexcolor,
				selectcolor=hexcolor,
				activebackground=hexcolor)

	# ----------------------------------------------------------------------
	# Force regeneration of projections & plots
	# ----------------------------------------------------------------------
	def redraw(self):
		self.layer()
		self.viewer.loadCamera()
		self.viewer.draw(True)

	# ----------------------------------------------------------------------
	def changeLayer(self):
		self.viewer.stop()
		self.layer()
		try:
			if self.viewer.winfo_ismapped():
				self.viewer.draw(late=True)
		except tk.TclError:
			pass
		self.parent.updateSelection()

	# ----------------------------------------------------------------------
	def resetLayer(self):
		self.layerCombo.set(self.defLayer)

	# ----------------------------------------------------------------------
	def layersDialog(self, event=None):
		self.parent.showLayer(self.layerCombo.get())

	# ----------------------------------------------------------------------
	# update layer on viewer
	# ----------------------------------------------------------------------
	def layer(self):
		if self.project() is None: return
		# FIXME
		# Maybe we should avoid reloading if the last layer was
		# the same. Probably with a hash on the layer
		# Problem. On large voxel files it takes too long
		# to update the regions etc..

		self.activeLayer = self.layerCombo.get()
		if self.activeLayer not in self.project().geometry.layers:
			self.activeLayer = "Media"
			if self.layerCombo.get() != self.activeLayer:
				self.layerCombo.set(self.activeLayer)

		viewer = self.viewer._viewer

		viewer.set("2D",      0)
		viewer.set("vertex",  0)
		viewer.set("fill",    0)
		viewer.set("lattice", 0)
		viewer.set("voxel",   0)
		viewer.set("palette", 0)
		viewer.set("userdump",0)
		viewer.set("usrbin",  0)
		viewer.set("image",   0)
		viewer.set("3D",      0)
		viewer.set("projection",0)
		viewer.set("labels", self.getVar("Show.label")!="None")

		viewer.set("axes",		self.getInt("Options.axes"))
		viewer.set("viewport",		self.getInt("Options.viewport"))
		viewer.set("grid",		self.getInt("Options.grid"))
		viewer.set("errors",		self.getInt("Options.errors"))
		viewer.set("gridlevel",		self.getInt("Options.gridlevel"))
		viewer.set("latticelevel",	self.getInt("Options.latticelevel"))
		viewer.set("voxellevel",	self.getInt("Options.latticelevel"))	# <-- same as lattice
		viewer.set("crosshair",		self.getInt("Options.crosshair"))
		viewer.set("textbackground",	self.getInt("Options.textbackground"))
		self.viewer.ocubeSet(self.getInt("Options.ocube")*tkFlair.dpiscale)

		# Check fonts
		self.setFont("general",	self.getVar("Options.font"))
		self.setFont("grid",	self.getVar("Options.gridfont"))
		self.setFont("palette",	self.getVar("Options.palettefont"))

		# First set palette colors
		for tag in Project.GeometryLayerInfo.PALETTENAMES:
			if self.getInt(tag+"._visible"):
				self._paletteLayer()
				break
		else:
			# reset all palettes
			for i in range(len(Project.GeometryLayerInfo.PALETTENAMES)):
				viewer.palette(i,"reset")

		self._showLayer()	# setup information to display

		if self.getInt("Show"):
			if self.getVar("Show.color") != "None":
				viewer.set("fill", 1)
			viewer.set("2D", 1)
			viewer.set("vertex",  self.getInt("Show.vertex"))

		if self.getInt("Image"):
			self._imageLayer()
		elif self._imageData:
			# Free memory
			self._clearImage()

		if self.getInt("3D"):		self._3dLayer()	# check before the usrbin
		if self.getInt("Userdump"):	self._userdumpLayer()

		viewer.voxel("roiclear")
		if self.getInt("Voxel"):	self._voxelLayer()

		for tag in Project.GeometryLayerInfo.USRBINNAMES:
			if self.getInt(tag):
				self._usrbinLayer()
				break

		# Common
		try:
			viewer.set("lattice", self.getInt("Show.lattice"))
		except TypeError:
			pass

		try:
			if self.getInt("Show.voxel") and self.parent.voxels is not None:
				viewer.set("voxel", self.getInt("Show.voxel"))
		except TypeError:
			pass

	# ----------------------------------------------------------------------
	# Load fonts
	# ----------------------------------------------------------------------
	def setFont(self, name, font):
		vfont = self.viewer._viewer.font(name)
		if vfont is None or font != os.path.splitext(os.path.basename(vfont))[0]:
			# Find font and load it
			fn = os.path.join(tkFlair.iniDir,f"fonts/{font}.tga")
			try: os.stat(fn)
			except: fn = os.path.join(tkFlair.prgDir,"fonts",f"{font}.tga")

			try: os.stat(fn)
			except: return

			self.viewer._viewer.font(name,fn)

	# ----------------------------------------------------------------------
	def _regionsLayer(self):
		_viewer = self.viewer._viewer
		showlabel = self.getVar("Show.label")
		for card in self.input().cardsCache(Card.REGION).values():
			if card.name()=="&": continue
			try:
				color = self.parent.regionColor(card)
				rid = _viewer.region(card.name(), "id")
				_viewer.region(rid, "color", color)
				if showlabel == "Material":
					try: _viewer.region(rid, "label", card[Card.MATERIAL].name())
					except KeyError: pass
				else:
					_viewer.region(rid, "label", card.name())
			except KeyError:
				pass

	# ----------------------------------------------------------------------
	def _materialsLayer(self, key=Card.MATERIAL):
		_viewer   = self.viewer._viewer
		showlabel = self.getVar("Show.label")

		# Scan regions
		for card in self.input().cardsCache(Card.REGION).values():
			if card.name()=="&": continue
			try:
				mat   = card[key]
				color = self.parent.materialColor(mat)
				rid   = card[Card.ID]
				_viewer.region(rid, "color", color)

				if showlabel in ("Material","Value"):
					if mat: _viewer.region(rid, "label", mat.name())
				else:
					_viewer.region(rid, "label", card.name())

				alpha = int(card.get(Card.ALPHA,0))
				if alpha:
					alpha = int(round((alpha*255)/100))
					_viewer.region(rid, "alpha", alpha)
				elif card[Card.TYPE]==Card.REGION_NORMAL:
					alpha = self.parent.materialAlpha(mat)
					alpha = int(round((alpha*255)/100))
					_viewer.region(rid, "alpha", alpha)
				else:
					_viewer.region(rid, "alpha", 0)
			except (KeyError, TypeError):
				pass

		# Scan for voxel assignments if needed
		_3d = self.getInt("3D")
		if self.parent.voxels and (self.getInt("Show.voxel") or _3d):
			matDict = self.input().materialDict()
			_viewer.region("VOXEL", "alpha", 0)

			# set the voxel and user assignmats
			for card in self.input().cardsSorted("ASSIGNMA",-1).values():
				if card.what(3):
					toVoxel = Fluka._VOXELPAT.match(card.what(3))
					if toVoxel is None: continue
				else:
					toVoxel = None

				fromVoxel = Fluka._VOXELPAT.match(card.what(2))
				if fromVoxel is None: continue

				fromVoxel = int(fromVoxel.group(1))
				if toVoxel:
					toVoxel = int(toVoxel.group(1))
				else:
					toVoxel = fromVoxel

				step = card.intWhat(4)
				if step<=0: step=1

				for voxel in range(fromVoxel, toVoxel+1, step):
					try:
						material = matDict.get(card.what(1))
						alpha = self.parent.materialAlpha(material)
						if material:
							if alpha == 100 and _3d: # transparent
								color = 0xFFFFFFFF
							else:
								color = self.parent.materialColor(material)
						else:
							color = _LATTICE_COLOR
							if alpha>0:
								color |= int((255*alpha)/100)<<24
						_viewer.voxel("color", voxel-1, color)
					except KeyError:
						pass

	# ----------------------------------------------------------------------
	def _showLayer(self):
		_viewer   = self.viewer._viewer
		colorValue = self.getVar("Show.color")
		# None will get the default value of 1.0
		if colorValue == "Zone":
			self._regionsLayer()
			return
		elif colorValue == "Region":
			self._regionsLayer()
			return
		elif colorValue == "Material":
			self._materialsLayer()
			return
		elif colorValue == "Material-Decay":
			self._materialsLayer("@matDecay")
			return

		# Setup viewer
		palette = self.getVar("Show.palette")
		try:
			_viewer.palette(Project.GeometryLayerInfo.PALETTENAMES.index(palette), "default", 0)
		except ValueError:
			pass
		cblog = self.getInt(palette+".log")

		prop = f"@{colorValue.lower()}"
		if colorValue=="Corrfactor":
			self.input().addCardProperty("REGION",
						"CORRFACT",
						prop,
						2,4,5,6)
		elif colorValue=="Corrfactor-dE/dx":
			self.input().addCardProperty("REGION",
						"CORRFACT",
						prop,
						1,4,5,6)
		elif colorValue=="Corrfactor-rho":
			self.input().addCardProperty("REGION",
						"CORRFACT",
						"@corrfactor",
						2,4,5,6)
		elif colorValue=="e-Production":
			self.input().addCardProperty("MATERIAL",
						"EMFCUT",
						prop,
						1,4,5,6,
						"PROD-CUT")
		elif colorValue=="e-Transport":
			self.input().addCardProperty("REGION",
						"EMFCUT",
						prop,
						1,4,5,6,"",
						_emfwhat1)
		elif colorValue=="g-Production":
			self.input().addCardProperty("MATERIAL",
						"EMFCUT",
						prop,
						1,4,5,6,"PROD-CUT",
						_emfwhat1)
		elif colorValue=="g-Transport":
			self.input().addCardProperty("REGION",
						"EMFCUT",
						prop,
						2,4,5,6,"")
		elif colorValue=="Deltaray":
			self.input().addCardProperty("MATERIAL",
						"DELTARAY",
						prop,
						1,4,5,6)
		elif colorValue=="Field":
			self.input().addCardProperty("REGION",
						"ASSIGNMA",
						prop,
						5,2,3,4)

		# ---- @return color value ----
		def findColorValue(card):
			try:
				if colorValue=="Density":
					value = card[Card.MATERIAL].numWhat(3)
				elif colorValue.startswith("Importance"):	# Suffix -H, -E, -N
					part = colorValue[11:]	# particle suffix
					biasing = card["@biasing"+part]
					if biasing is None: biasing = card["@biasingA"]
					if biasing is not None:
						value = biasing.numWhat(3)
					else:
						value = 1.0
				elif colorValue.startswith("Splitting"):
					part = colorValue[10:]	# particle suffix
					biasing = card["@biasingH"]
					if biasing is None: biasing = card["@biasingA"]
					if biasing is not None:
						value = biasing.numWhat(2)
					else:
						value = 1.0
				elif colorValue=="Corrfactor-rho":
					try:
						value = float(card["@corrfactor"])
						if value == 0.0: value = 1.0
					except KeyError: value = 1.0
					value *= card[Card.MATERIAL].numWhat(3)
					#log.say(card.name(), card["@corrfactor"], card[Card.MATERIAL].numWhat(3), value)
				elif colorValue=="e-Production":
					value = float(card[Card.MATERIAL][prop])
				elif colorValue=="g-Production":
					value = float(card[Card.MATERIAL][prop])
				elif colorValue=="Deltaray":
					value = float(card[Card.MATERIAL][prop])
				else:
					value = float(card[prop])

			except (TypeError, AttributeError):
				value = 1.0

			label = str(value)

			if cblog:
				try: value = math.log10(value)
				except: value = 0.0
			else:
				try: value = float(value)
				except: value = 0.0

			return label, value

		# Loop over regions
		showlabel = self.getVar("Show.label")
		for card in self.input().cardsCache(Card.REGION).values():
			if card.name()=="&": continue
			label, value = findColorValue(card)
			try:
				rid = _viewer.region(card.name(), "id")
				_viewer.region(rid, "value", value)
				if showlabel == "Material":
					try: _viewer.region(rid, "label", card[Card.MATERIAL].name())
					except KeyError: pass
				elif showlabel == "Value":
					_viewer.region(rid, "label", label)
				else:
					_viewer.region(rid, "label", card.name())

				alpha = int(card.get(Card.ALPHA,0))
				if alpha:
					alpha = int(round((alpha*255)/100))
					_viewer.region(rid, "alpha", alpha)
				elif card[Card.TYPE]==Card.REGION_NORMAL:
					alpha = self.parent.materialAlpha(card[Card.MATERIAL])
					alpha = int(round((alpha*255)/100))
					_viewer.region(rid, "alpha", alpha)
				else:
					_viewer.region(rid, "alpha", 0)
			except KeyError:
				pass

		# Loop voxel regions
		if self.parent.voxels and (self.getInt("Show.voxel") or self.getInt("3D")):
			for voxel in self.input()["VOXELS"]:
				if voxel.ignore() or voxel[FlukaCard.VOXEL] is None: continue
				for card in voxel[FlukaCard.VOXEL]["REGION"]:
					label, value = findColorValue(card)
					pat = Fluka._VOXELPAT.match(card.name())
					if pat is None: continue
					vid = int(pat.group(1))
					_viewer.voxel("value", vid-1, value)

	# ----------------------------------------------------------------------
	def _clearImage(self):
		self._imageName  = None
		self._imageTime  = 0
		self._imageBlack = 0
		self._imageWhite = 0
		self._imageData  = None

	# ----------------------------------------------------------------------
	# Overlay an image on viewport
	# ----------------------------------------------------------------------
	def _imageLayer(self):
		_viewer = self.viewer._viewer
		_viewer.set("image", 1+self.getInt("Image.prompt"))	# 1=Late, 2=Prompt

		# Check for background image
		markers = GeometryLayers.getMarkers(self.project().geometry.layers[self.activeLayer])

		R,M = CalibrateImage.compute(markers)
		if M is None:
			self.flair().notify("Image Calibration",
				"Error in calibrating image",
				tkFlair.NOTIFY_ERROR)
			return

		iw = bmath.Vector(M[2][0], M[2][1], M[2][2])
		invmatrix = _viewer.matrix(-1)
		u  = bmath.Vector(invmatrix[0][0], invmatrix[0][1], invmatrix[0][2])
		v  = bmath.Vector(invmatrix[1][0], invmatrix[1][1], invmatrix[1][2])
		w  = bmath.Vector(invmatrix[2][0], invmatrix[2][1], invmatrix[2][2])
		dot = w * iw

		if abs(dot)<=0.9999:
			_viewer.viewOto0()

			# Not visible rotate the viewport
			if abs(dot)<0.01:
				u = iw.orthogonal()
			else:
				if dot<0.0: iw = -iw
				u = v^iw
			u.normalize()
			v = iw^u
			v.normalize()
			matrix = bmath.Matrix(3)
			matrix.make(u,v,iw)
			_viewer.matrix(matrix.transpose())
			self.viewer.draw(True)

		alpha    = self.getInt("Image.alpha")
		black    = self.getInt("Image.black")
		white    = self.getInt("Image.white")
		filename = self.getVar("Image.file")

		load = False	# Flag whether to load image

		# Check if color range has changed
		if black!=self._imageBlack or \
		   white!=self._imageWhite:
			load = True
			self._imageBlack = black
			self._imageWhite = white

		# Check if file is different
		if not load:
			load = (filename != self._imageName)
			self._imageName = filename

		# Check if image has been modified
		if not load:
			try:	# Check time
				mtime = os.path.getmtime(filename)
				load = (mtime > self._imageTime)
				self._imageTime = mtime
			except OSError:
				return

		if load:
			try:
				image = PIL.Image.open(filename)
				self._imageName = filename
			except:
				self.flair().notify("Error opening background image",
					"Cannot open file %r"%(filename),
					tkFlair.NOTIFY_ERROR)
				return

			image = image.convert("RGBA")
			image.load()
			try:
				self._imageData = image.tobytes()
			except AttributeError:
				self._imageData = image.tostring()
			size = image.size
			del image
			try:
				_viewer.image("data", self._imageData)
			except MemoryError as ex:
				del image
				self.flair().notify("Not enough memory",
					str(ex),
					tkFlair.NOTIFY_ERROR)
				return
			_viewer.image("size", size)
			if black!=0 or white!=0xffffff:
				_viewer.image("level", (black,white))

		_viewer.image("alpha", alpha)
		_viewer.image("matrix", (R,M))

	# ----------------------------------------------------------------------
	def _usrbinLayer(self):
		_viewer = self.viewer._viewer
		_viewer.set("usrbin", 1+self.getInt("Usrbin.prompt"))

		for i,tag in enumerate(Project.GeometryLayerInfo.USRBINNAMES):
			if not self.getInt(tag):
				_viewer.usrbin(i, "cleanup")
				continue

			# FIXME remember and do not reload if image is not modified
			try:    det = self.getInt(tag+".det")
			except: det = 1

			if self.getInt(tag+".input"):
				# Find usrbin from input
				try:
					usrbins = list(self.input().cardsSorted('USRBIN',-1).values())
					card    = usrbins[det-1]
					case    = card.case()
					default = card.info.default[case]
					_viewer.usrbin(i, "checker",
						(card.intWhat(1),
						 card.numWhat(7), card.numWhat(4),
						 card.intWhat(10, int(default[10])),
						 card.numWhat(8), card.numWhat(5),
						 card.intWhat(11, int(default[11])),
						 card.numWhat(9), card.numWhat(6),
						 card.intWhat(12, int(default[12]))))
					_viewer.usrbin(i, "offset", (0.0, 0.0, 0.0))
				except (KeyError, IndexError):
					continue
			else:
				# Load usrbin from file
				filename = self.getVar(tag+".file")
				if not filename:
					_viewer.usrbin(i, "cleanup")
					continue
				try:
					_viewer.usrbin(i, "load", filename, det)
				except IOError as ex:
					self.flair().notify("Error loading USRBIN",
						str(ex),
						tkFlair.NOTIFY_ERROR)
					continue

				palette = self.getVar(tag+".palette")
				try:
					_viewer.usrbin(i, "palette",
							Project.GeometryLayerInfo.PALETTENAMES.index(palette))
				except ValueError:
					palette = "Palette"
					_viewer.usrbin(i, "palette", 0)

				norm = self.project().input.eval(self.getVar(tag+".norm"))
				_viewer.usrbin(i, "norm", norm)
				_viewer.usrbin(i, "showerror", self.getInt(tag+".showerror"))
				_viewer.usrbin(i, "log",       self.getInt(palette+".log"))
				#if _viewer.usrbin(i, "type") in (2,8,12,18):
				if _viewer.usrbin(i, "isregion"):
					_viewer.set("fill",1)
					_viewer.usrbin(i, "regioncolor")

				try: xofs = float(self.getVar(tag+".xofs"))
				except: xofs = 0.0
				try: yofs = float(self.getVar(tag+".yofs"))
				except: yofs = 0.0
				try: zofs = float(self.getVar(tag+".zofs"))
				except: zofs = 0.0
				_viewer.usrbin(i, "offset", (xofs, yofs, zofs))

			_viewer.usrbin(i, "matrix", None)

			rot = self.getVar(tag+".rotdefi")
			if rot is not None and rot!="":
				_viewer.usrbin(i, "matrix", self.input().getTransformation(rot))

			rot = self.getVar(tag+".rotdefi2")
			if rot is not None and rot!="":
				_viewer.usrbin(i, "matrix", self.input().getTransformation(rot))

			_viewer.usrbin(i, "alpha", self.getInt(tag+".alpha"))

	# ----------------------------------------------------------------------
	def _3dLayer(self):
		self.viewer._geometry.setLights()
		self.viewer._viewer.set("3D", 1)
		self.viewer._viewer.set("drawtime",       15)
		self.viewer._viewer.set("ambient",         self.getInt("3D.ambient"))
		self.viewer._viewer.set("antialias",       self.getInt("3D.antialias"))
		self.viewer._viewer.set("deflights",       self.getInt("3D.deflights"))
		self.viewer._viewer.set("edgedetect",      self.getInt("3D.edgedetect"))
		self.viewer._viewer.set("projection",      self.getInt("3D.projection"))
		self.viewer._viewer.set("reflections",     self.getInt("3D.reflections"))
		self.viewer._viewer.set("shadows",         self.getInt("3D.shadows"))
		self.viewer._viewer.set("aoc",             self.getInt("3D.aoc"))
		self.viewer._viewer.set("skipblack",       self.getInt("3D.skipblack"))
		self.viewer._viewer.set("usrbinastexture", self.getInt("3D.usrbinastexture"))
		self.viewer._viewer.set("xray",            self.getInt("3D.xray"))
		self.viewer._viewer.camera("fov",          self.getFloat("3D.fov"))
		self.viewer._viewer.error("clear")

		self.viewer._viewer.set("clipbody", -1)		# Clear clipping bodies
		for i in range(3):
			try:
				clipbody = self.viewer._geometry.body(self.getVar(f"3D.clip.{i}"),"id")
				self.viewer._viewer.set("clipbody", clipbody)
				if self.getInt(f"3D.clipnegative.{i}"):
					self.viewer._viewer.set("clipnegative", i)
			except (KeyError, TypeError):
				pass

		self.viewer._viewer.set("projectbody",-1)	# Clear projection bodies
		for i in range(100):
			try:
				projbody = self.viewer._geometry.body(self.getVar(f"3D.project.{i}"),"id")
				self.viewer._viewer.set("projectbody", projbody)
			except (KeyError, TypeError):
				if i>3: break

	# ----------------------------------------------------------------------
	def _userdumpLayer(self):
		self.viewer._viewer.set("userdump", 1)
		self.viewer._viewer.userdump("reset")
		self.viewer._viewer.userdump("file",	self.getVar("Userdump.file"))
		self.viewer._viewer.userdump("start",	self.getInt("Userdump.start")-1)
		self.viewer._viewer.userdump("n",	self.getInt("Userdump.n"))
		self.viewer._viewer.userdump("alpha",	self.getInt("Userdump.alpha"))

		for particle in FlukaParticle.list:
			if not particle or particle[0]=='@': continue
			try:
				idx = FlukaParticle.get(particle).id
			except:
				idx = 0
			try:
				show = self.getInt(f"Userdump.{particle}")
			except TypeError:
				show = False
			if show:
				self.viewer._viewer.userdump("show", idx)
				try:
					emin = self.getFloat(f"Userdump.emin.{particle}")
					self.viewer._viewer.userdump("emin", idx, emin)
				except ValueError:
					pass
				try:
					emax = self.getFloat(f"Userdump.emax.{particle}")
					self.viewer._viewer.userdump("emax", idx, emax)
				except ValueError:
					pass

				color = self.getStr(f"Userdump.color.{particle}")
				if color:
					self.viewer._viewer.userdump("color", idx,
								tkPalette.tk2int(color))
			else:
				self.viewer._viewer.userdump("hide", idx)

	# ----------------------------------------------------------------------
	def _voxelLayer(self):
		for card in self.input()["VOXELS"]:
			if card.ignore() or card[FlukaCard.VOXEL] is None: continue
			break
		else:
			return
		for roi in map(int,self.getStr('Voxel.roi').split()):
			self.viewer._viewer.voxel("roi", roi, 0)
		self.viewer._viewer.voxel("roialpha", self.getInt('Voxel.roialpha'))

	# ----------------------------------------------------------------------
	def _paletteLayer(self):
		_viewer = self.viewer._viewer
		_viewer.set("palette", 1)
		for idx,tag in enumerate(Project.GeometryLayerInfo.PALETTENAMES):
			_viewer.palette(idx, "reset")
			if not self.getInt(f"{tag}._visible"):
				continue
			cbmin = self.getFloat(f"{tag}.min")
			cbmax = self.getFloat(f"{tag}.max")
			cbn   = self.getInt(f"{tag}.n")
			cblog = self.getInt(f"{tag}.log")
			if cbn<=0: cbn = 32
			elif cbn>256: cbn=256
			try:
				palette = paletteStore[self.getVar(f"{tag}.palette")].asRGB(cbn)
			except KeyError:
				palette = paletteStore[""].asRGB(cbn)
			if cblog:
				try: cbmin = math.log10(cbmin)
				except: cbmin = -4.0
				try: cbmax = math.log10(cbmax)
				except: cbmax = 4.0

			_viewer.palette(idx, "show", self.getInt(tag))
			_viewer.palette(idx, "min", cbmin)
			_viewer.palette(idx, "max", cbmax)
			_viewer.palette(idx, "log", cblog)
			_viewer.palette(idx, "smooth",  self.getInt(tag+".smooth"))
			_viewer.palette(idx, "invert",  self.getInt(tag+".inv"))
			_viewer.palette(idx, "alphamin",self.getInt(tag+".alphamin"))
			_viewer.palette(idx, "alphamax",self.getInt(tag+".alphamax"))
			_viewer.palette(idx, "label",   self.getStr(tag+".label"))
			_viewer.palette(idx, "palette", palette)

	# ----------------------------------------------------------------------
	def rotateLayer(self):
		self.layerCombo.select(self.nextLayer())

	# ----------------------------------------------------------------------
	def nextLayer(self):
		return (self.layerCombo.select()+1)%self.layerCombo.size()

	# ----------------------------------------------------------------------
	def getVar(self, name, default=None):
		try:
			layer = self.project().geometry.layers[self.activeLayer]
		except KeyError:
			return Project.GeometryLayerInfo.DEFAULT[name]

		if default is not None:
			return layer.setdefault(name, default)
		else:
			return layer(name)

	# ----------------------------------------------------------------------
	def getStr(self, name, default=None):
		return str(self.getVar(name,default))

	# ----------------------------------------------------------------------
	def getInt(self, name, default=None):
		return int(self.getVar(name,default))

	# ----------------------------------------------------------------------
	def getFloat(self, name, default=None):
		return float(self.getVar(name,default))

	# ----------------------------------------------------------------------
	def maximizeWindow(self, callLayout=True):
		if self.maxWindow.get():
			self.maxButton["text"] = "\N{BLACK DOWN-POINTING TRIANGLE}"
			if callLayout:
				self.parent.layout(self.name)
		else:
			self.maxButton["text"] = "\N{BLACK UP-POINTING TRIANGLE}"
			if callLayout:
				self.parent.layout("custom")

	# ----------------------------------------------------------------------
	def matrix(self, matrix=None):
		if matrix is None:
			return self.viewer._viewer.matrix()
		else:
			self.viewer._viewer.matrix(matrix)

	# ----------------------------------------------------------------------
	def historyButton(self, left, right):
		self._bar[0].historyButton(left, right)

	# ----------------------------------------------------------------------
	def zoomOnSelected(self, objects=None):
		self.viewer.zoomOnSelected(objects)

	# ----------------------------------------------------------------------
	def loadCamera(self, name):
		for camera in self.input()["camera"]:
			if camera.ignore(): continue
			if camera.name() == name:
				self.viewer.loadCamera(camera)
				return True
		return False

	# ----------------------------------------------------------------------
	# get projection information for flair PlotInfo
	# ----------------------------------------------------------------------
	def loadPlot(self, name=None):
		if self.project() is None: return
		if name is None:		# keep the initial argument
			plotname = self.name
		else:
			plotname = name
		self.viewer.resetCamera()

		# Check for camera with the special prefix
		if plotname.startswith(GeometryToolbar.CAMERA_PREFIX):
			self.loadCamera(plotname[len(GeometryToolbar.CAMERA_PREFIX):].strip())
			self.viewer.draw()
			return

		# Check for plot
		for plot in self.project().plots():
			if plot.name==plotname and plot.type=="Geometry":
				break
		else:
			return

		# Maybe it is still a camera, only for the default loading
		if name is None and plot("kind").startswith(Project.PlotInfo.CAMERA_KIND):
			plotname = plot("kind")[len(Project.PlotInfo.CAMERA_KIND):].strip()
			if self.loadCamera(plotname):
				self.bookmarkCombo.set(GeometryToolbar.CAMERA_PREFIX + plotname)
			return

		u = bmath.Vector(map(float,plot("basisu").split()))
		v = bmath.Vector(map(float,plot("basisv").split()))
		w = u ^ v
		M = bmath.Matrix(3)
		M.make(u,v,w)
		self.viewer._viewer.matrix(M.transpose())

		origin = list(map(float,plot("origin").split()))
		self.viewer._viewer.origin(origin[0], origin[1], origin[2])

		extU,extV = list(map(float, plot("extends").split()))
		self.viewer._viewer.extends(-2.0*extU, -2.0*extV, 2.0*extU, 2.0*extV)
		self.viewer._viewer.zoom(2.0)
		self.viewer._viewer.aspect(float(plot("aspect")))

		self.viewer.draw(True,True)	# force complete redraw
		self.layerCombo.set(plot("layer"))

	# ----------------------------------------------------------------------
	# set projection information to flair PlotInfo
	# ----------------------------------------------------------------------
	def savePlot(self, name=None):
		if name is None:
			name = self.name
		else:
			self.parent.setStatus(f"Plot saved as: {name}")
		for plot in self.project().plots():
			if plot.name == name and plot.type == "Geometry": break
		else:
			plot = Project.PlotInfo(self.project, name)
			plot.type = "Geometry"
			plot.dirname("Geometry")
			self.project().plots.add(plot)
			self.flair().refresh("plot")

		plot["origin"] = "%g %g %g" % self.viewer._viewer.origin()
		plot["basisu"] = "%g %g %g" % self.viewer._viewer.basis('u')
		plot["basisv"] = "%g %g %g" % self.viewer._viewer.basis('v')
		x1,y1,x2,y2 = self.viewer._viewer.extends()
		z = self.viewer._viewer.zoom()
		plot["extends"] = f"{(x2-x1)/2.0/z} {(y2-y1)/2.0/z}"
		plot["aspect"]  = str(self.viewer._viewer.aspect())

		u,v = self.viewer._viewer.grid("axes")	# will return {-}[UXYZ]:{-}[VXYZ]
		u = u[-1]	# get rid of sign if any
		v = v[-1]
		if u=="U" or v=="V":
			plot["coord"] = "U-V"
		else:
			plot["coord"] = f"{u}-{v}"
		plot["boundaries"] = 1
		try:	# when a second main window is opened and we close the
			# the window through the "protocol" I get a tk.TclError
			dt = self.layerCombo.get()
			plot["layer"] = dt
			if self.viewer.camera is None:
				plot["kind"] = "Geometry" if dt=="Borders" else "Material"
			else:
				plot["kind"] = Project.PlotInfo.CAMERA_KIND+self.viewer.camera.name()
		except tk.TclError:
			pass
		plot["labels"]  = 1

#===============================================================================
# Geometry4Frame parent frame of the 4 viewports
#
#     +-----------+-------------+
#     |-----------|-------------|
#     |           |             |
#     |    Red    |    Green    |
#     |           |             |
#     +-----------+-------------+
#     |-----------|-------------|
#     |           |             |
#     |   Blue    |   Magenta   |
#     |           |             |
#     +-----------+-------------+
#     |       Timeframe         |
#     +-------------------------+
#
#===============================================================================
class Geometry4Frame(tk.Frame):
	# ----------------------------------------------------------------------
	def __init__(self, master, editor, **kw):
		super().__init__(master, **kw)
		self._geometry = geoviewer.Geometry()
		#if Global.developer:
		#	self._geometry.fpe()		# enable floating point exceptions
		self._geometry.set("trackballsize", int(TRACKBALLSIZE*100))
		self._geometry.set("snapdistance",  SNAPDISTANCE)
		self._geometry.set("cursor",        CURSORSIZE)
		self._geometry.set("developer",     Global.developer)

		self.project        = None		# project currently loaded
		self.run            = None		# run currently loaded
		self.editor         = editor
		self.oldMouseAction = ACTION.SELECT
		self.mouseAction    = ACTION.SELECT	# Current mouse action when mouse is clicked
		self.active         = 0			# Active viewport
		self.selectedViewer = None		# selected viewport
		self.actionTool     = None		# Tool dialog of action
		self.actionOption   = None		# option of mouse action e.g. what to add
		self.errors         = []
		self.warnings       = []
		self.voxels         = None		# voxels filename
		self.colorValue     = 0			# which color value to plot
		self.move3D         = False		# moving 2D or 3D
		self.orientation    = None		# current orientation system

		self.closest        = None		# tuple with the closest mouse hit on viewports
		self.objects        = []		# Selected moveable objects
		self.objectsid      = []		# @see _select()
		self.objectLight    = False		# bool if object has a light selected @see _select
		self.bodies         = []		# Selected moveable bodies
		self.bodiesid       = []

		self._regionColor   = {}
		self._autoBody      = None		# auto body dialog
		self._syncViewports = False		# synchronize views
		self.__reload       = None		# reload project on idle
		self.__afterDraw    = None		# draw on idle
		self.__selcards     = []		# keep track of selected cards

		# insert time frame for movie editing at the bottom
		self.timeframe = TimeFrameCanvas(self, takefocus=True)
		self.timeframe.draw()
		self.timeframe.bind("<F1>",    self.timeframeHelp)
		self.timeframe.bind("<Key-k>", self.timeframeToggle)
		self.timeframe.bind("<ButtonRelease-1>",   self.timeframeRelease)
		self.timeframe.bind("<<TimeFrameActive>>", self.timeframeActiveDragged)
		tkExtra.bindData(self.timeframe, "<<TimeFrameChange>>", self.timeframeChange)

		# Create frames
		frame = tk.Frame(self)
		frame.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
		self.vsplit = tkExtra.VSplitter(frame,
					tkFlair.config.getFloat(tkFlair._PAGE_SECTION,
						"Geometry.vsplit", 0.5))
		self.vsplit.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

		self.tsplit = tkExtra.HSplitter(self.vsplit.topFrame(),
					tkFlair.config.getFloat(tkFlair._PAGE_SECTION,
						"Geometry.tsplit", 0.5))
		self.tsplit.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
		self.bsplit = tkExtra.HSplitter(self.vsplit.bottomFrame(),
					tkFlair.config.getFloat(tkFlair._PAGE_SECTION,
						"Geometry.bsplit", 0.5))
		self.bsplit.pack(side=tk.BOTTOM, expand=tk.YES, fill=tk.BOTH)

		# The 4 geometry viewer frames
		view1 = GeometryFrame(self.tsplit.leftFrame(), self,
				VIEWER.NAME[0], takefocus=False)
		view1.pack(expand=tk.YES, fill=tk.BOTH)
		view2 = GeometryFrame(self.tsplit.rightFrame(), self,
				VIEWER.NAME[1], takefocus=False)
		view2.pack(expand=tk.YES, fill=tk.BOTH)
		view3 = GeometryFrame(self.bsplit.leftFrame(), self,
				VIEWER.NAME[2], takefocus=False)
		view3.pack(expand=tk.YES, fill=tk.BOTH)
		view4 = GeometryFrame(self.bsplit.rightFrame(), self,
				VIEWER.NAME[3], takefocus=False)
		view4.pack(expand=tk.YES, fill=tk.BOTH)

		# Save frames
		self.frames   = (view1, view2, view3, view4)
		self._viewers = [x.viewer._viewer for x in self.frames]
		self.viewportOrientation(Project.GeometryInfo.DEFAULT["Orientation"])
		self.resetViewports()

		for i,f in enumerate(self.frames):
			f.setColor(VIEWER.COLOR[i])
			for j,f2 in enumerate(self.frames):
				if f is f2: continue
				f.viewer.addViewport(f2.viewer, VIEWER.COLOR[j])

		self.custom = (0.5, 0.5, 0.5)		# custom layer
		self._layout = "custom"
		self.resetLayers()
		self.layoutUpdateButtons()

		# Create tools
		self.tool = {}
		self.tool[ACTION.ZONE]    = GeometryTool.ZoneTool(self)
		self.tool[ACTION.MOVE]    = GeometryTool.MoveTool(self)
		self.tool[ACTION.ROTATE]  = GeometryTool.RotateTool(self)
		self.tool[ACTION.REPEAT]  = GeometryTool.RepeatTool(self)
		self.tool[ACTION.VOLUME]  = GeometryTool.VolumeTool(self)
		self.tool[ACTION.EXPORT]  = GeometryTool.ExportTool(self)
		self.tool[ACTION.MOVIE]   = GeometryTool.MovieTool(self)
		self.tool[ACTION.PLANNER] = GeometryTool.PlannerTool(self)
		self.tool[ACTION.LINK]    = GeometryTool.LinkTool(self)

		self.bind("<Configure>", self.configureEvent)

	# ----------------------------------------------------------------------
	# Access
	def geometry(self):		return self._geometry
	def flair(self):		return self.editor.flair
	def __getitem__(self, item):	return self._viewers[item]

	# ----------------------------------------------------------------------
	def setActive(self, viewport):
		self.active = self._viewers.index(viewport)

	# ----------------------------------------------------------------------
	def activeViewer(self):
		return self.frames[self.active].viewer

	# ----------------------------------------------------------------------
	def configureEvent(self, event):
		if self.actionTool:
			self.actionTool.event_generate("<Configure>")

	# ----------------------------------------------------------------------
	# @return currently selected cards
	# ----------------------------------------------------------------------
	def selectedCards(self):
		return self.__selcards

	# ----------------------------------------------------------------------
	# Update the information for the selected/drawn elements on
	# the action tool and timeframe
	# Note: it is dangerous to have [] as default value on the
	#       argument. However here the self.__selcards is not
	#       supposed to be changed
	# FIXME: maybe convert to "None" and set to selcards as []
	# ----------------------------------------------------------------------
	def updateSelection(self, selcards=[]):
		self.__selcards = selcards
		self.updateActionToolSelection()

	# ----------------------------------------------------------------------
	# Prepare selection lists for objects and bodies
	# ----------------------------------------------------------------------
	def selectObjectsBodies(self, objects, bodies):
		if objects:
			self.objects = objects
			self.objectsid = [x[Card.ID] for x in self.objects]
			for card in objects:
				if card.tag == "light":
					self.objectLight = True
					break
			else:
				self.objectLight = False
		else:
			del self.objects[:]
			del self.objectsid[:]

		if bodies:
			self.bodies = bodies
			self.bodiesid = [x[Card.ID] for x in self.bodies if x]
		else:
			del self.bodies[:]
			del self.bodiesid[:]

		return self.objects or self.bodies

	# ----------------------------------------------------------------------
	# Set viewport orientation system and sync options
	# @return True if the system has changed
	# ----------------------------------------------------------------------
	def viewportOrientation(self, systemname):
		if self.orientation == systemname:
			return False
		self.orientation = systemname

		try:
			system = ORIENTATIONS[self.orientation]
		except KeyError:
			self.orientation = "flair"
			system = ORIENTATIONS[self.orientation]

		for i,name in enumerate(VIEWER.NAME):
			layer,sync,*axes = system[name]
			if len(axes)==3:
				Rx,Ry,Rz = map(math.radians, axes)
				matrix = bmath.Matrix.eulerRotation(Rx,Ry,Rz)
			else:
				U = AXES_VECTORS[axes[0]]
				V = AXES_VECTORS[axes[1]]
				W = U ^ V
				matrix = bmath.Matrix()
				matrix.make(U,V,W)
				matrix.transpose()

			self.frames[i].defLayer  = layer
			self.frames[i].viewer.defSync   = sync
			self.frames[i].viewer.defSystem = matrix

			for face,aname in enumerate(AXES_ORDER):
				self._viewers[i].ocube("facename",
					face, system[aname])
		return True

	# ----------------------------------------------------------------------
	def resetViewports(self):
		for frame in self.frames:
			frame.viewer.resetViewport()

	# ----------------------------------------------------------------------
	def synchronizeViewports(self, s=None):
		if s is not None:
			if s and not self._syncViewports:
				# set the orientation as in the resetViewports only
				# when it is changed state, to avoid resetting
				# on undo/redo and project reloads
				#for i in (0,2,3):
				#for i in range(4):
				for frame in self.frames:
					if not frame.viewer._sync: continue
					_viewer = frame.viewer._viewer
					xyz = _viewer.origin()
					frame.viewer.synchronizeViewports()
					_viewer.origin(xyz)
					frame.viewer.draw(True)
			self._syncViewports = s
		self.frames[self.active].viewer.synchronizeViewports()

	# ----------------------------------------------------------------------
	def loadOrigin(self):
		for frame in self.frames:
			frame.loadOrigin()

	# ----------------------------------------------------------------------
	def loadMaterialColors(self):
		for card in self.project.input.cardsCache(Card.MATERIALS).values():
			# FIXME do I really need that?
			card[Card.COLOR] = None
			self.materialColor(card)
			self.updateMaterial(card)

	# ----------------------------------------------------------------------
	def materialColor(self, card):
		if card is None: return _BACKGROUND_COLOR
		color = card[Card.COLOR]
		if color is not None: return color
		mat = self.project.getMaterial(card.name())
		card[Card.COLOR] = color = mat.color32()
		return color

	# ----------------------------------------------------------------------
	def materialAlpha(self, card):
		if card is None: return 0
		return int(self.project.getMaterial(card.name())("alpha"))

	# ----------------------------------------------------------------------
	def regionColor(self, card):
		if card is None: return _BACKGROUND_COLOR
		color = card[Card.COLOR]
		if color is not None: return color
		try:
			card[Card.COLOR] = self._regionColor[card.name()]
		except KeyError:
			card[Card.COLOR] \
				= self._regionColor[card.name()] \
				= rndColor(hash(card.name()))
		return card[Card.COLOR]

	# ----------------------------------------------------------------------
	# Set Status mouse coordinates
	# ----------------------------------------------------------------------
	def setStatusPos(self, xyz):
		self.editor.xlabel.config(text=f"{xyz[0]:.10g}")
		self.editor.ylabel.config(text=f"{xyz[1]:.10g}")
		self.editor.zlabel.config(text=f"{xyz[2]:.10g}")

	# ----------------------------------------------------------------------
	# Set status message
	# In order not to show the balloon set timeout=0
	# ----------------------------------------------------------------------
	def setStatus(self, msg="", fg="DarkBlue", bg=None, timeout=4000):
		self.editor.setStatus(msg, fg, bg)
		if timeout and msg:
			self.showBalloon(msg, timeout)

	# ----------------------------------------------------------------------
	def setStatusPercent(self, *args, **kwargs):
		self.editor.setStatusPercent(*args, **kwargs)

	# ----------------------------------------------------------------------
	def resetPercent(self, *args, **kwargs):
		self.editor.resetPercent(*args, **kwargs)

	# ---------------------------------------------------------------------
	def showBalloon(self, msg, timeout=4000):
		self._help = msg
		tkExtra.Balloon.setWidget(self,
					self.winfo_pointerx(),
					self.winfo_pointery())
		tkExtra.Balloon.show(timeout)

	# ----------------------------------------------------------------------
	# Select viewer and update action tool if necessary
	# ----------------------------------------------------------------------
	def selectViewer(self, viewer=None):
		self.selectedViewer = viewer
		if self.selectedViewer:
			self.updateSelection()

	# ----------------------------------------------------------------------
	# Select maximized viewport
	# ----------------------------------------------------------------------
	def selectMaxViewer(self):
		for frame in self.frames:
			if frame.maxWindow.get():
				self.selectViewer(frame.viewer)
				return
		self.selectViewer()

	# ----------------------------------------------------------------------
	# Update the selection of the current action tool if any
	# ----------------------------------------------------------------------
	def updateActionToolSelection(self):
		if self.actionTool:
			self.actionTool.updateSelection()
		self.timeframeUpdate()

	# ----------------------------------------------------------------------
	# Perform the mouse action in all frames
	# ----------------------------------------------------------------------
	def setMouseAction(self, action=ACTION.SELECT):
		if self.mouseAction == action: return
		if self.mouseAction not in (ACTION.ADD_ZONE, ACTION.ADD_ZONEPAINT):
			self.oldMouseAction = self.mouseAction
		self.mouseAction = action

		# Hide and Show action tool dialog
		if action>=0:
			self.timeframeHide()
			if self.actionTool: self.actionTool.hide()
			try:
				self.actionTool = self.tool[action]
				if action in (ACTION.MOVIE, ACTION.PLANNER):
					self.timeframeShow()
			except KeyError:
				self.actionTool = None
			if self.actionTool: self.actionTool.show()

		elif action == ACTION.SELECT_VIEWPORT and \
		     self.actionTool and self.actionTool.hasViewportSelection():
			self.selectMaxViewer()

		cursor = mouseCursor(action)
		#dump("+++ Action=",action, cursor)
		for f in self.frames:
			f.viewer._cursor = cursor
			f.viewer.config(cursor=cursor)
		self._geometry.zone("clear")

		# Color level
		ll = self._geometry.set("lighterlevel")
		if action in (ACTION.ADD_ZONE, ACTION.ADD_ZONEPAINT):
			newll = 200
		else:
			newll = 0
		if newll != ll:
			self._geometry.set("lighterlevel",newll)
			self.draw()

		# Update viewers if needed
		for v in self._viewers:
			v.set("trackball", action==ACTION.ORBIT)
			v.expose()

		if self.actionTool and self.mouseAction>0:
			self.actionTool.nextAction()

		try:
			self.setStatus(ACTION_MESSAGE[action])
		except KeyError:
			pass

	# ----------------------------------------------------------------------
	# Toggle action to new or previous
	# ----------------------------------------------------------------------
	def toggleAction(self, action):
		if self.mouseAction != action:
			self.editor.setMouseAction(action)
		else:
			self.editor.setMouseAction(self.oldMouseAction)

	# ----------------------------------------------------------------------
	# Repeat last action
	# ----------------------------------------------------------------------
	def repeatLastAction(self):
		self.editor.setMouseAction(self.oldMouseAction)

	# ----------------------------------------------------------------------
	def setLayers(self, dt=None):
		if dt is None:
			focus = self.focus_get()
			if focus.master in self.frames:
				dt = focus.master.nextLayer()
			else:
				dt = self.frames[0].nextLayer()
		for frame in self.frames:
			frame.layerCombo.select(dt)

	# ----------------------------------------------------------------------
	# Set default layout with:
	#	Red     = Media
	#	Green   = 3D
	#	Blue    = Media
	#	Magenta = Media
	# ----------------------------------------------------------------------
	def resetLayers(self):
		for frame in self.frames:
			frame.resetLayer()

	# ----------------------------------------------------------------------
	def showLayer(self, layer):
		self.editor.showLayer(layer)

	# ----------------------------------------------------------------------
	def rotateLayout(self):
		n = VIEWORDER.index(self._layout) + 1
		if n>=len(VIEWORDER): n = 0
		self.layout(VIEWORDER[n])

	# ----------------------------------------------------------------------
	def customLayout(self):
		focus = self.focus_get()
		if focus.master in self.frames:
			fmt = focus.master.name
		else:
			fmt = ""
		if self._layout==fmt:
			self.layout("custom")
		else:
			self.layout(fmt)

	# ----------------------------------------------------------------------
	def compareLayout(self, v, t, b):
		if abs(self.vsplit.split-v) > 0.01: return False
		if v <= 0.01:
			# Compare only the bottom
			return abs(self.bsplit.split-b) <= 0.01
		elif v >= 0.99:
			# Compare only the top
			return abs(self.tsplit.split-t) <= 0.01
		else:
			# Compare both top and bottom
			return  abs(self.bsplit.split-b) <= 0.01 and \
				abs(self.tsplit.split-t) <= 0.01

	# ----------------------------------------------------------------------
	def layout(self, fmt):
		v,t,b = VIEWFORMAT.get(fmt, self.custom)
		if self.compareLayout(v,t,b): return	# same

		self._layout = fmt
		self.custom = ( self.vsplit.split,
				self.tsplit.split,
				self.bsplit.split)

		self.vsplit.setSplit(v)
		self.tsplit.setSplit(t)
		self.bsplit.setSplit(b)
		self.layoutUpdateButtons()

	#-----------------------------------------------------------------------
	def layoutUpdateButtons(self):
		for f in self.frames:
			f.maxWindow.set(False)
			f.maxButton["text"] = "\N{BLACK UP-POINTING TRIANGLE}"
		if self.vsplit.split>=0.99:
			if self.tsplit.split>=0.99:	# top left is maximized
				self.frames[0].maxWindow.set(True)
				self.frames[0].maxButton["text"] = "\N{BLACK DOWN-POINTING TRIANGLE}"
			elif self.tsplit.split<=0.01:	# top right is maximized
				self.frames[1].maxWindow.set(True)
				self.frames[1].maxButton["text"] = "\N{BLACK DOWN-POINTING TRIANGLE}"
		elif self.vsplit.split<=0.01:
			if self.bsplit.split>=0.99:	# bottom left is maximized
				self.frames[2].maxWindow.set(True)
				self.frames[2].maxButton["text"] = "\N{BLACK DOWN-POINTING TRIANGLE}"
			elif self.bsplit.split<=0.01:	# bottom right is maximized
				self.frames[3].maxWindow.set(True)
				self.frames[3].maxButton["text"] = "\N{BLACK DOWN-POINTING TRIANGLE}"

	# ----------------------------------------------------------------------
	# Save views as flair plots
	# ----------------------------------------------------------------------
	def saveState(self):
		if self.project is None: return
		for frame in self.frames:
			frame.savePlot()
		self.setOption("Frame.vsplit", self.vsplit.split)
		self.setOption("Frame.tsplit", self.tsplit.split)
		self.setOption("Frame.bsplit", self.bsplit.split)
		self.project.geometry["StartFrame"] = self.timeframe.start
		self.project.geometry["StopFrame"]  = self.timeframe.stop

	# ----------------------------------------------------------------------
	# Load views from flair plots
	# ----------------------------------------------------------------------
	def loadState(self):
		if self.project is None: return
		for frame in self.frames:
			frame.loadPlot()
		start = self.project.geometry("StartFrame")
		stop  = self.project.geometry("StopFrame")
		self.timeframe.setLimits(start, stop)

	# ----------------------------------------------------------------------
	def log(self, msg):
		if not self.flair() or not self.flair().log:
			log.say(msg)
		else:
			self.flair().log(msg)

	# ----------------------------------------------------------------------
	def addError(self, msg):
		if msg.startswith(log.ERROR_PREFIX):
			self.log(f"{log.ERROR}{msg[4:]}")
			self.errors.append(msg)
		if msg.startswith(log.WARNING_PREFIX):
			self.log(f"{log.WARNING}{msg[4:]}")
			self.warnings.append(msg)
		else:
			self.log(f"{log.ERROR}{msg}")
			self.errors.append(msg)

	# ----------------------------------------------------------------------
	def addWarning(self, msg):
		self.log(f"{log.WARNING}{msg}")
		self.warnings.append(msg)

	# ----------------------------------------------------------------------
	# set a variable in all views
	# ----------------------------------------------------------------------
	def set(self, var, value=None):
		if value is None:
			return self._geometry.set(var)
		else:
			self._geometry.set(var, value)

	# ----------------------------------------------------------------------
	# Add material in geometry
	# ----------------------------------------------------------------------
	def addMaterial(self, card):
		try:
			return self._geometry.material(card.name(),"add")
		except Exception as ex:
			self.addError(f"Material '{card.name()}' {ex}")
			return None

	# ----------------------------------------------------------------------
	def updateMaterial(self, card):
		mid = card[Card.ID]
		if mid is None: return
		if isinstance(card, FlukaCard):
			self._geometry.material(mid, "Z",       card.intWhat(1))
			self._geometry.material(mid, "weight",  card.numWhat(2))
			self._geometry.material(mid, "density", card.numWhat(3))
			self._geometry.material(mid, "A",       card.intWhat(6))

		mat = self.project.getMaterial(card.name())
		self._geometry.material(mid, "bxdf",     mat.get("bxdf"))
		self._geometry.material(mid, "specular", float(mat.get("specular")))
		self._geometry.material(mid, "shine",    float(mat.get("shine")))
		self._geometry.material(mid, "fuzz",     float(mat.get("fuzz")))
		self._geometry.material(mid, "roughness",float(mat.get("roughness")))
		self._geometry.material(mid, "ior",      float(mat.get("ior")))
		#self._geometry.material(mid, "alpha",    int(mat.get("alpha")))

	# ----------------------------------------------------------------------
	# add a body in all views
	# ----------------------------------------------------------------------
	def addBody(self, name, tag, what, matrix=None):
		#Global.dump(f"*-* addBody: {name} {tag} {what} {matrix}")
		try:
			bid = self._geometry.body(name, "add", tag)
		except Exception as ex:
			self.addError(f"Body '{name}' {ex}")
			return None
		try:
			self._geometry.body(bid, "what", what)
		except Warning as ex:
			self.addWarning(f"Body '{name}' {ex}")
		except Exception as ex:
			self.addError(f"Body '{name}' {ex}")
		if matrix is not None:
			self._geometry.body(bid, "matrix", matrix)
		return bid

	# ----------------------------------------------------------------------
	# modify body in geometry
	# ----------------------------------------------------------------------
	def body(self, idx, var, value=None):
		if idx is None: return
		try:
			if value is None:
				return self._geometry.body(idx, var)
			else:
				self._geometry.body(idx, var, value)
		except IndexError as e:
			self.addError(f"Body #{idx} {e}")

	#-----------------------------------------------------------------------
	# Scan for unused bodies and show warnings
	#-----------------------------------------------------------------------
	def unuusedBodies(self):
		for i,body in enumerate(self._geometry.error("unusedbodies")):
			self.addWarning(f"Body '{body}' unused")
			if i>50:
				self.addWarning("Too many unused bodies")
				break

	# ----------------------------------------------------------------------
	# modify region in geometry
	# ----------------------------------------------------------------------
	def region(self, idx, var, value=None):
		if idx is None: return
		if value is None:
			return self._geometry.region(idx, var)
		else:
			self._geometry.region(idx, var, value)

	# ----------------------------------------------------------------------
	# Manipulate zone selection
	# ----------------------------------------------------------------------
	def zone(self, cmd, value=None):
		if value is None:
			self._geometry.zone(cmd)
		else:
			# Ignore invalid bodies from zones
			try:
				self._geometry.zone(cmd, value)
			except ValueError:
				pass
			except Warning as e:
				self.flair().notify("Error selecting zone",
					str(e),
					tkFlair.NOTIFY_ERROR)

	# ----------------------------------------------------------------------
	# Add object
	# ----------------------------------------------------------------------
	def addObject(self, card):
		oid = card[Card.ID] = self._geometry.object(card.name(), "add", card.tag)
		self._geometry.object(oid, "x", card.numWhat(1))
		self._geometry.object(oid, "y", card.numWhat(2))
		self._geometry.object(oid, "z", card.numWhat(3))
		option = card.intWhat(4)
		self._geometry.object(oid, "type", option)
		s,a = divmod(card.intWhat(5),100)
		self._geometry.object(oid, "size",   s)
		self._geometry.object(oid, "anchor", a)
		self._geometry.object(oid, "color", tkPalette.tk2int(card.colorWhat(6)))

		# Additional variables
		if   card.tag == "arrow":
			self._geometry.object(oid, "dx", card.numWhat(7))
			self._geometry.object(oid, "dy", card.numWhat(8))
			self._geometry.object(oid, "dz", card.numWhat(9))
		elif card.tag == "camera":
			self._geometry.object(oid, "atx",    card.numWhat(7))
			self._geometry.object(oid, "aty",    card.numWhat(8))
			self._geometry.object(oid, "atz",    card.numWhat(9))
			self._geometry.object(oid, "upx",    card.numWhat(10))
			self._geometry.object(oid, "upy",    card.numWhat(11))
			self._geometry.object(oid, "upz",    card.numWhat(12))
			self._geometry.object(oid, "focal",  card.numWhat(13))
			self._geometry.object(oid, "fov",    card.numWhat(14))
			self._geometry.object(oid, "aspect", card.numWhat(15))
		elif card.tag == "light":
			self._geometry.object(oid, "dx",      card.numWhat(7))
			self._geometry.object(oid, "dy",      card.numWhat(8))
			self._geometry.object(oid, "dz",      card.numWhat(9))
			self._geometry.object(oid, "power",   card.numWhat(10))
			self._geometry.object(oid, "falloff", card.intWhat(11))
			self._geometry.object(oid, "specular",card.numWhat(12))
			self._geometry.object(oid, "radius",  card.numWhat(13))
			self._geometry.object(oid, "nsamples",card.intWhat(14))
		elif card.tag == "ruler":
			self._geometry.object(oid, "xe", card.numWhat(7))
			self._geometry.object(oid, "ye", card.numWhat(8))
			self._geometry.object(oid, "ze", card.numWhat(9))
			self._geometry.object(oid, "xa", card.numWhat(10))
			self._geometry.object(oid, "ya", card.numWhat(11))
			self._geometry.object(oid, "za", card.numWhat(12))
		elif card.tag == "mesh":
			self._geometry.object(oid, "file", card.text())
		elif card.tag == "spline":
			self._geometry.object(oid, "tension",  card.numWhat(7,0.5))
			self._geometry.object(oid, "segments", card.intWhat(8,5))
			n = card.intProperty("node:0",0)
			if option&1: n -= 1	# last point is only used for time
			for i in range(1,n+1):
				x = card.numProperty(f"node:{i}.x")
				y = card.numProperty(f"node:{i}.y")
				z = card.numProperty(f"node:{i}.z")
				t = card.numProperty(f"node:{i}.t")
				self._geometry.object(oid, "node", (t,x,y,z))
			if option&1:
				n += 1
				t = card.numProperty(f"node:{n}.t")
				self._geometry.object(oid, "node", (t,))	# <<< add last time
		return oid

	# ----------------------------------------------------------------------
	def object(self, idx, var, value=None):
		if value is None:
			return self._geometry.object(idx, var)
		else:
			self._geometry.object(idx, var, value)

	# ----------------------------------------------------------------------
	# Add a rotdefi as object to geometry
	# ----------------------------------------------------------------------
	def addRotdefi(self, card, name, matrix):
		oid = card[Card.ID] = self._geometry.object(name, "add", "rotdefi")
		self._geometry.object(oid, "axissize", 40)
		self._geometry.object(oid, "linewidth", 1)
		self._geometry.object(oid, "axiswidth", 3)
		self._geometry.object(oid, "size", 7)
		x = card["x"]
		if x: self._geometry.object(oid, "x", float(x))
		y = card["y"]
		if y: self._geometry.object(oid, "y", float(y))
		z = card["z"]
		if z: self._geometry.object(oid, "z", float(z))
		self._geometry.object(oid, "matrix", matrix)
		return oid

	# ----------------------------------------------------------------------
	# Add a BEAM card as object to geometry
	# ----------------------------------------------------------------------
	def addBeam(self, card, name):
		oid = card[Card.ID] = self._geometry.object(name, "add", "beam")
		if card.tag == "BEAM":
			ew = 1
			dw = 3
		else:
			ew = 2
			dw = 5
		self._geometry.object(oid, "energy", abs(card.numWhat(ew)))
		try:
			scale = abs(float(card.get("scale",1.0)))
			self._geometry.object(oid, "scale", scale)
		except ValueError:
			pass
		self._geometry.object(oid, "divergence", card.numWhat(dw)/1000.0)
		self._geometry.object(oid, "size",      5)
		self._geometry.object(oid, "linewidth", 2)
		self._geometry.object(oid, "color",     tkPalette.tk2int(card.get("color","")))
		return oid

	# ----------------------------------------------------------------------
	# Update the edit handle of the geometry
	# ----------------------------------------------------------------------
	def updateEdit(self, card=None):
		# Don't show handler within an action
		# FIXME needs better thinking
		if self.mouseAction<0:
			self._geometry.cursor("clear")
			return

		# Check last card
		if card is not None and \
		   card.get(Card.SELECT,0) & BIT.SELECT:
			if card.type() == Card.BODY:
				try:
					self._geometry.cursor("body", card[Card.ID])
				except IndexError:
					pass	# Body might not exist e.g. editing inactive branch
		elif self.project is not None:
			#lastBody = lastRegion = lastObject = None
			lastBody = lastRegion = None
			for card in self.project.input.allcards():
				if card.ignore(): continue
				if card.get(Card.SELECT,0) & BIT.SELECT == 0: continue
				t = card.type()
				if t == Card.BODY:
					lastBody = card
				elif t == Card.REGION:
					lastRegion = card
				#elif t == Card.OBJECT:
				#	lastObject = card

			# Highest priority bodies
			if lastBody is not None:
				try:
					self._geometry.cursor("body", lastBody[Card.ID])
				except IndexError:
					pass

			elif lastRegion is not None:
				try:
					self._geometry.cursor("region", lastRegion[Card.ID])
				except IndexError:
					# Voxels will fail
					pass

			#elif lastObject is not None:
			#	self._geometry.cursor("object", lastObject[Card.ID])

			else:
				self._geometry.cursor("clear")

	# ----------------------------------------------------------------------
	# Update region and body properties in geometry
	# ----------------------------------------------------------------------
	def updateGeometry(self):
		rotdefi = self.project.input.getTransformation()
		for region in self.project.input.cardsCache(Card.REGION).values():
			if region.name()=="&": continue
			try:
				rid = region[Card.ID]
				self._geometry.region(rid, "type", region[Card.TYPE])
				if region["@rotdefi"]:
					rot = signedRotdefi(rotdefi, region["@rotdefi"])
					self._geometry.region(rid, "matrix",
						self.project.input.getTransformation(rot,rotdefi))
			except (KeyError, TypeError):
				pass

	# ----------------------------------------------------------------------
	# Reload project to be called from a timer
	# ----------------------------------------------------------------------
	def _reloadProject(self):
		self.loadProject()
		#self.editor.selectShow(False)
		for frame in self.frames:
			frame.viewer.loadCamera()
			if frame.viewer._viewer.get("3D"):
				frame.viewer._viewer.draw(False, DRAW.FAST, True)
			frame.viewer._viewer.expose()
			frame.viewer.draw(True)
		self.__reload = None

	# ----------------------------------------------------------------------
	def reloadProjectOnIdle(self):
		if self.__reload:
			self.after_cancel(self.__reload)
		self.__reload = self.after_idle(self._reloadProject)

	# ----------------------------------------------------------------------
	# Load project
	# ----------------------------------------------------------------------
	def loadProject(self, project=None, run=None):
		#dump("*-* GeometryViewer.loadProject")
		self.flair().logReset()
		if project: self.project = project
		self.fillLayers()

		# Remember cursor
		cursor     = self._geometry.cursor("cursor")
		cursorType = self._geometry.cursor("type")
		cursorId   = self._geometry.cursor("id")

		# Stop threads
		self.stopThreads()	# ignore thread stopping from profiling

		if Global.profiler:
			log.say("\n*** GeometryViewer.loadProject started")
			timer = tool.Timer()

		# Cleanup everything
		self._geometry.cleanup()
		if Global.profiler: timer.lap("\tGeometryViewer.loadProject cleanup")

		# Load the new project
		self._geometry.error("clear")
		self._geometry.lock()
		try:
			self._loadInput(run)
		except:
			traceback.print_exc()
		self._geometry.unlock()

		# Derive...
		self.derive()
		if Global.profiler: timer.lap("\tGeometryViewer.loadProject derive")

		# Restore selection cursor if any
		if cursor:
			try: self._geometry.cursor(cursorType, cursorId)
			except: pass
		if Global.profiler: timer.lap("\tGeometryViewer.loadProject input")

		# Load information
		self.loadOrigin()
		if Global.profiler: timer.lap("\tGeometryViewer.loadProject origin")

		# Check for possible errors
		if self.errors or self.warnings:
			self.setStatus("Errors found while loading geometry.", "Red")

		if Global.profiler: timer.stop("GeometryViewer.loadProject ended")

		for i in range(self._geometry.error()):
			self.addError(self._geometry.error("get",i))

		warn = bool(self.warnings) and not bool(self.errors)
		self.flair().errorsDuringLoading(warn)

	# ----------------------------------------------------------------------
	# common initialization for loading the input
	# ----------------------------------------------------------------------
	def _loadInput(self, run):
		del self.errors[:]
		del self.warnings[:]
		self.voxels = None

		# Scan input for body and region properties
		inp = self.project.input
		if run is not None: self.run = run
		if inp.preprocess(self.run.defines()):
			self.addError("Pre-processing errors found")

		# Setup properties
		rotdefi = inp.cardProperties()
		#inp.validate()

		# --------------------
		#      Objects
		# --------------------
		for card in inp.cardsSorted(CardInfo.FLAIR_TAGS,-1).values():
			self.addObject(card)
		self._geometry.setLights()

		# --------------------
		# Load specific input
		# --------------------
		self._geometry.mode(self.project.input.MODE)
		if isinstance(self.project.input, Fluka):
			self._loadFluka(rotdefi)
		elif isinstance(self.project.input, Moira):
			self._loadMoira(rotdefi)
		elif isinstance(self.project.input, Mcnp):
			self._loadMcnp(rotdefi)
		elif isinstance(self.project.input, Phits):
			self._loadPhits(rotdefi)
		elif isinstance(self.project.input, Penelope):
			self._loadPenelope(rotdefi)

	# ======================================================================
	# FLUKA
	# ======================================================================
	def _loadFluka(self, rotdefi):
		if Global.profiler: timer = tool.Timer()
		inp = self.project.input

		# --------------------
		#       BEAM
		# --------------------
		for card in inp["BEAM"]:
			if card.ignore(): continue
			oid = self.addBeam(card, "beam")
			break
		else:
			oid = None

		dx = dy = 0.0	# default beam direction
		dz = 1.0
		if oid is not None:
			beamaxes = None
			for card in inp["BEAMAXES"]:
				if card.ignore(): continue
				bX = bmath.Vector(card.numWhat(1),
						card.numWhat(2),
						card.numWhat(3))
				bZ = bmath.Vector(card.numWhat(4),
						card.numWhat(5),
						card.numWhat(6))
				bX.normalize()
				bZ.normalize()
				bY = bZ ^ bX
				beamaxes = bmath.Matrix(4)
				beamaxes.make(bX,bY,bZ)
				beamaxes.transpose()
				break
			for card in inp["BEAMPOS"]:
				if card.ignore(): continue
				sdum = card.sdum().strip()
				if sdum not in ("","NEGATIVE"): continue
				card[Card.ID] = oid
				self._geometry.object(oid, "x", card.numWhat(1))
				self._geometry.object(oid, "y", card.numWhat(2))
				self._geometry.object(oid, "z", card.numWhat(3))
				dx = card.numWhat(4)
				dy = card.numWhat(5)
				try:
					dz = math.sqrt(1.0 - dx*dx - dy*dy)
				except:
					dz = 0.0
				if sdum=="NEGATIVE": dz = -dz
				if beamaxes:
					dx,dy,dz = beamaxes * bmath.Vector(dx,dy,dz)
				self._geometry.object(oid, "dx", dx)
				self._geometry.object(oid, "dy", dy)
				self._geometry.object(oid, "dz", dz)
			if Global.profiler: timer.lap("\t\tGeometryViewer._loadFluka:beam:")

		# --------------------
		#  All SPOTxxx cards
		# --------------------
		if "SPOTBEAM" in inp.cards:
			nspot = 0
			spots = {}
			for card in inp.allcards():
				if card.ignore(): continue
				if card.tag == "SPOTBEAM":
					if card.intWhat(0)==0:
						nspot += 1
					else:
						nspot = card.intWhat(0)
					oid = self.addBeam(card, f"spot{nspot}")
					spots[nspot] = oid
				elif card.tag == "SPOTPOS":
					if card.intWhat(0)!=0:
						try:
							oid = spots[card.intWhat(0)]
						except:
							log.error("mix-up in spots id")
					card[Card.ID] = oid
					self._geometry.object(oid, "x", card.numWhat(1))
					self._geometry.object(oid, "y", card.numWhat(2))
					self._geometry.object(oid, "z", card.numWhat(3))
					self._geometry.object(oid, "dx", dx)
					self._geometry.object(oid, "dy", dy)
					self._geometry.object(oid, "dz", dz)
				elif card.tag == "SPOTDIR":
					if card.intWhat(0)!=0:
						try:
							oid = spots[card.intWhat(0)]
						except:
							log.error("mix-up in spots id")
					card[Card.ID] = oid
					self._geometry.object(oid, "dx", card.numWhat(1))
					self._geometry.object(oid, "dy", card.numWhat(2))
					self._geometry.object(oid, "dz", card.numWhat(3))
				elif card.tag == "SPOTTRAN":
					inv    = card.sdum()=="INVERSE"
					matrix = inp.getTransformation(card.what(1), rotdefi, inv)
					from_  = card.intWhat(4)
					to_    = card.intWhat(5)
					step   = card.intWhat(6)
					if step == 0: step = 1
					for i in range(from_, to_+1, step):
						try:
							self._geometry.object(f"spot{i}",
								"matrix", matrix)
						except KeyError:
							pass

			if Global.profiler: timer.lap("\t\tGeometryViewer._loadFluka:spotbeam:")

		# --------------------
		#       ROT-DEFI
		# --------------------
		rot = set()
		for card in inp.cardsSorted(Card.ROTATION,-1).values():
			name = card.name().strip()
			#if not name: ... find name from what(1)
			if name in rot: continue
			rot.add(name)
			matrix = inp.getTransformation(name, rotdefi)
			if matrix: self.addRotdefi(card, name, matrix)

		if Global.profiler: timer.lap("\t\tGeometryViewer._loadFluka:transformation:")

		# --------------------
		#      MATERIAL
		# --------------------
		for card in inp.cardsCache(Card.MATERIALS).values():
			card[Card.ID] = self.addMaterial(card)
			self.updateMaterial(card)

		if Global.profiler: timer.lap("\t\tGeometryViewer._loadFluka:materials:")

		# --------------------
		#  Bodies & Regions
		# --------------------
		self._loadBodies(rotdefi)
		self._loadRegions(rotdefi)

		# --------------------
		#       Usrbins
		# --------------------
		#idx = 0
		#for card in inp["USRBIN"]:
		#	if card.ignore(): continue
		#	idx += 1
		#	name = "USRBIN.%d"%(idx)

		#	if card.intWhat(1) in (0,10):
		#		tag   = "RPP"
		#		whats = [card.numWhat(7), card.numWhat(4),
		#		         card.numWhat(8), card.numWhat(5),
		#		         card.numWhat(9), card.numWhat(6)]
		#		self.addBody(name, tag, whats)

		#	elif card.intWhat(1) in (1,11):
		#		tag   = "RCC"
		#		whats = [card.numWhat(7), card.numWhat(4),
		#		         card.numWhat(8), card.numWhat(5),
		#		         card.numWhat(9), card.numWhat(6)]
		#		self.addBody(name, tag, whats)

		#	else:
		#		continue

	# ======================================================================
	# MOIRA
	# ======================================================================
	def _loadMoira(self, rotdefi):
		if Global.profiler: timer = tool.Timer()
		inp = self.project.input

		# --------------------
		#       BEAM
		# --------------------
		X = Y = None
		beamdir = bmath.Vector(0, 0, 1)
		for card in inp["/primary/select"]:
			if card.ignore(): continue
			oid = card[Card.ID] = self._geometry.object(card.name(),
							"add", "beam")
			break
		else:
			oid = None
		if oid is not None:
			for card in inp.cardlist:
				if card.tag == "/gps/pos/centre":
					card[Card.ID] = oid
					self._geometry.object(oid, "x", card.numWhat(0))
					self._geometry.object(oid, "y", card.numWhat(1))
					self._geometry.object(oid, "z", card.numWhat(2))
				elif card.tag == "/gps/ene/mono":
					card[Card.ID] = oid
					self._geometry.object(oid, "energy", card.name())
				elif card.tag == "/gps/energy":
					card[Card.ID] = oid
					self._geometry.object(oid, "energy", card.name())
				if card.tag == "/gps/ang/maxtheta":
					card[Card.ID] = oid
					self._geometry.object(oid, "scale", 1.0) # XXX FIXME
					self._geometry.object(oid, "divergence", 2*card.numWhat(0)/1000.0 )
					self._geometry.object(oid, "size",      5)
					self._geometry.object(oid, "linewidth", 2)
					self._geometry.object(oid, "color",
								tkPalette.tk2int(card.get("color","")))
				if card.tag == "/gps/ang/rot2":
					card[Card.ID] = oid
					X  = bmath.Vector(card.numWhat(0), card.numWhat(1), card.numWhat(2))
				if card.tag == "/gps/ang/rot1":
					card[Card.ID] = oid
					Y  = bmath.Vector(card.numWhat(0), card.numWhat(1), card.numWhat(2))
			if X is not None and Y is not None :
				beamdir = X ^ Y
			if oid is not None:
				self._geometry.object(oid, "dx", beamdir[0])
				self._geometry.object(oid, "dy", beamdir[1])
				self._geometry.object(oid, "dz", beamdir[2])

		if Global.profiler: timer.lap("\tGeometryViewer._loadMoira:beam:")

		# ------------------------
		# Load materials
		# ------------------------
		for card in inp.materialDict().values():
			card[Card.ID] = self.addMaterial(card)
			self.updateMaterial(card)

		# ------------------------
		# Dic. MATRIX.name[Bodies]
		# ------------------------
		rotMat = {}
		curmat = None
		for card in inp.allcards():
			if card.tag == "MATRIX":
				curmat = card.what(0)
			elif card.type() == Card.BODY and curmat is not None:
				rotMat[card.what(0)] = curmat
			elif card.tag not in ("MATRIX","BODY") and curmat is not None:
				break
			else:
				continue

		# ------------------------
		# Finding transformation
		# ------------------------
		for card in inp.cardsSorted(Card.ROTATION,-1).values():
			name = card.name()
			matrix = inp.getTransformation(name, rotdefi)
			if matrix: self.addRotdefi(card, name, matrix)
		if Global.profiler: timer.lap("\tGeometryViewer._loadMoira:transformation:")

		self._loadBodies(rotdefi)
		self._loadCells(rotdefi)

	# ======================================================================
	# MCNP
	# ======================================================================
	def _loadMcnp(self, rotdefi):
		self._loadBodies(rotdefi)
		self._loadCells(rotdefi)

	# ======================================================================
	# PHITS
	# ======================================================================
	def _loadPhits(self, rotdefi):
		self._loadBodies(rotdefi)
		self._loadCells(rotdefi)

	# ----------------------------------------------------------------------
	# Bodies
	# ----------------------------------------------------------------------
	def _loadBodies(self, rotdefi):
		if Global.profiler: timer = tool.Timer()
		inp = self.project.input

		# --------------------
		#       Bodies
		# --------------------
		for card in inp.cardsSorted(Card.BODY,-1).values():
			if card is None: continue
			if card.tag == "VOXELS":
				if self.voxels is not None:
					self.addError("More than one voxel cards found")
					continue

				fn = card.name().strip()
				if not fn: continue
				self.voxels = f"{fn}.vxl"
				try:
					self._geometry.voxel("load", self.voxels)
				except Exception as ex:
					self.addError(f"Cannot load voxel file '{self.voxels}'\n{ex}")
					self.voxels = None
					continue

				try:
					x = card.numWhat(1)
					y = card.numWhat(2)
					z = card.numWhat(3)
					rot = card.what(4)

					self._geometry.voxel("xlow", x)
					self._geometry.voxel("ylow", y)
					self._geometry.voxel("zlow", z)

					card[Card.ID] = self.addBody("VOXEL", "RPP",
						[ x, self._geometry.voxel("xhigh"),
						  y, self._geometry.voxel("yhigh"),
						  z, self._geometry.voxel("zhigh") ])

					if rot:
						irot = inp.getTransformation(rot,rotdefi,True)
						# Unfortunately bodies and regions/voxels
						# use opposite transformation
						self._geometry.voxel("matrix",irot)
						if irot is not None:
							self._geometry.body(card[Card.ID], "matrix",
								inp.getTransformation(rot,rotdefi))
					else:
						self._geometry.voxel("matrix",None)
						self._geometry.body(card[Card.ID], "matrix", None)
				except Exception as ex:
					self.addError(f"loading voxel {ex}")

			elif card.tag=="END":
				continue

			else:
				if card.invalid:
					self.addError(f"Body '{card.name()}' {card.errorMessage()}")

				try:	# We have to check if body already exists
					self._geometry.body(card.name(),"id")
					self.addError(f"Body '{card.name()}' duplicate body")
					#card.addInvalid("Duplicate body")
				except KeyError:
					pass
				except Exception as ex:
					log.error(ex)
					log.errorCont(card)
					log.errorCont(card.name(),type(card.name()))

				# FIXME it will not work if the expression is changed
				# i don't have a link of references from each expression
				# Also if the expression is based on another body!!
				#whats = card[Card.CACHE]
				#if whats is None:
				try:
					whats = [card.numWhat(i) for i in range(1,card.nwhats()) ]
				except ValueError:
					self.addError(f"Body '{card.name()}' invalid whats")
					continue
				#card[Card.CACHE] = whats

				# Add the body
				card[Card.ID] = self.addBody(card.name(),
								card.tag,
								whats,
								card[Card.MATRIX])
				bbox = card["bbox"]
				if bbox:
					try:
						self._geometry.body(card[Card.ID], "bbox",
								list(map(float,bbox.split())))
					except ValueError:
						log.error("Invalid value in body's bbox")
						log.errorCont(card)
						log.errorCont(card.name(),type(card.name()))

		# Cleanup maybe old voxel
		if self.voxels is None:
			self._geometry.voxel("free")

		if Global.profiler: timer.stop("\t\tGeometryViewer._loadBodies:")

	# ----------------------------------------------------------------------
	# Regions or Cells
	# ----------------------------------------------------------------------
	def _loadRegions(self, rotdefi):
		if Global.profiler: timer = tool.Timer()
		inp = self.project.input
		if isinstance(inp, Fluka):
			tagname = "Region"
		elif isinstance(inp, (Moira, Mcnp, Phits)):
			tagname = "Cell"
		else:
			assert False

		#---------------------------------------------------------------
		# addRegion
		#---------------------------------------------------------------
		def _addRegion(region, expr):
			name = region.name()
			if not region.NAMEPAT.match(name):
				self.addError(f"{tagname} '{name}' invalid name")

			# add last region
			if region.invalid:
				self.addError(f"{tagname} '{name}' {region.errorMessage()}")
			try:	# We have to check if region already exists
				self._geometry.region(name,"id")
				self.addError(f"{tagname} '{name}' invalid")
			except KeyError:
				pass
			except Exception as ex:
				log.error(ex)
				log.errorCont(region)

			if region[Card.MATERIAL] is None:
				self.addWarning(f"{tagname} '{name}' "
						"is not assigned any material")

			rid = None
			try:
				region[Card.ID] = rid = self._geometry.region(name, "add")
				self._geometry.region(rid, "type", region[Card.TYPE])
				region[Card.ZONE] = self._geometry.region(rid, "expr", expr, True)
			except Warning as ex:
				self.addWarning(f"{tagname} '{name}' {ex}")
			except Exception as ex:
				self.addError(f"{tagname} '{name}' {ex}")

			if rid is not None:
				if region["@rotdefi"] or region["@rotdefi2"]:
					rot  = signedRotdefi(rotdefi, region["@rotdefi"])
					rot2 = signedRotdefi(rotdefi, region["@rotdefi2"])
					if not rot and not rot2:
						self.addWarning(f"{tagname} '{name}' "
							"cannot display LATTICE no ROTDEFI is assigned")
					if rot is not None:
						if isinstance(inp, Moira):
							rot = rot.clone()
							rot[0][3] = region.get("@dx",0)
							rot[1][3] = region.get("@dy",0)
							rot[2][3] = region.get("@dz",0)
							rot.inv()
							self._geometry.region(rid, "matrix", rot)
						elif isinstance(inp, Fluka):
							self._geometry.region(rid, "matrix",
									inp.getTransformation(rot,
										rotdefi))
					if rot2 is not None:
						self._geometry.region(rid, "matrix",
								inp.getTransformation(rot2, rotdefi))
				mat = region[Card.MATERIAL]
				if mat and mat[Card.ID] is not None:
					self._geometry.region(rid, "material", mat[Card.ID])

				# Check for bounding boxes
				nonlocal invalidBBoxCount, invalidBBoxShow
				for i in range(self._geometry.region(rid, "zone")):
					try:
						bbox = self._geometry.bbox("Z",rid,i)
					except IndexError:
						bbox = None
					if bbox is None or bbox[0]>bbox[3]:
						invalidBBoxCount += 1
						if invalidBBoxCount <= invalidBBoxShow:
							self.addWarning(f"{tagname} '{name}:{i+1}'"
								"invalid bounding box")

		#---------------------------------------------------------------
		# regions
		#---------------------------------------------------------------
		expr   = None
		region = None
		invalidBBoxCount = 0		# counter of invalid bounding box
		invalidBBoxShow  = 50		# messages to show
		for card in inp.cardsSorted(Card.REGION,-1).values():
			# append to previous expression if needed
			if card.name() == "&" and expr is not None:
				expr += card.expr()
				# FIXME should moved to regionProperties
				region[Card.CONT] = True
				continue
			if expr is not None: _addRegion(region, expr)
			# move old card to region and process new card
			region = card
			# create a region color if needed
			self.regionColor(region)
			# use by default the materialColor, we change it latter
			self.materialColor(region[Card.MATERIAL])
			expr = card.expr()

		if expr is not None: _addRegion(region, expr)
			#if self.voxels:
			#	rid = self.addRegion("VOXEL", [("STD", ["VOXEL","$"])])
			#	if rid is not None:
			#		self._geometry.region(rid, "type", Card.REGION_VOXEL)

		if invalidBBoxCount > invalidBBoxShow:
			self.addWarning(f"{invalidBBoxCount-invalidBBoxShow+1} "
					"invalid bounding box messages suppressed")

		if Global.profiler: timer.stop("\t\tGeometryViewer._loadRegions:")

	_loadCells = _loadRegions

	# ======================================================================
	# Penelope
	# ======================================================================
	def _loadPenelope(self, rotdefi):
		# ------------------------
		# Load materials
		# ------------------------
		for card in self.project.input.cardsSorted(Card.MATERIALS, -1).values():
			card[Card.ID] = self.addMaterial(card)
			self.updateMaterial(card)
		self._loadSurfaces(rotdefi)
		self._loadModules(rotdefi)

	# ----------------------------------------------------------------------
	# Load Penelope surfaces
	# FIXME move most of the functionality in the Penelope.bodyProperties
	# ----------------------------------------------------------------------
	def _loadSurfaces(self, rotdefi):
		for card in self.project.input.cardsSorted(Card.SURFACE).values():
			if card._children:
				matrix,scale = card.matrix()
			else:
				matrix = None
				scale  = [1,1,1]

			sform = card.surfaceForm()
			stype = None
			if sform == "plane":
				stype = "xyp"
				what  = [-card.numWhat(5)*scale[2]]

			elif sform == "cylinder":
				if scale[0] == scale[1]:
					stype = "zcc"
					what  = [0.0, 0.0, scale[0]]
				else:
					stype = "zec"
					what  = [0.0, 0.0, scale[0], scale[1]]

			elif sform == "sphere":
				if scale[0]==scale[1] and scale[1]==scale[2]:
					stype = "sph"
					what  = [0.0, 0.0, 0.0, scale[0]]

			#elif sform == "cone":
			#	stype = "k/z"

			elif sform == "quadric0":
				stype = "qua"
				what = [0.0] * 10
				for child in card._children:
					if child.ignore(): continue
					try:
						idx = ("AXX", "AYY", "AZZ",
						       "AXY", "AXZ", "AYZ",
						       "AX",  "AY",  "AZ", "A0").index(child.tag[9:])
						what[idx] = child.numWhat(0)
					except ValueError:
						pass

			# if nothing was result fall to the default one
			if stype is None:
				stype = "qua"
				what = [card.numWhat(1), card.numWhat(2), card.numWhat(3),
					0.0, 0.0, 0.0,
					0.0, 0.0, card.numWhat(4),
					card.numWhat(5)]
				what[0] /= scale[0]**2	# Cxx
				what[1] /= scale[1]**2	# Cyy
				what[2] /= scale[2]**2	# Czz
				what[8] /= scale[2]	# Cz

			#print("-"*80)
			#print(f"SURFACE={card.name()} {stype=} {what=}\n{matrix=}")
			card[Card.ID] = self.addBody(card.name(), stype, what, matrix)

	# ----------------------------------------------------------------------
	# Load Penelope modules and bodies
	# FIXME move most of the functionality in the Penelope.regionProperties
	# ----------------------------------------------------------------------
	def _loadModules(self, rotdefi):
		surfaces = self.project.input.cardsSorted(Card.SURFACE)
		regions  = self.project.input.cardsSorted(Card.MODULE)

		#---------------------------------------------------------------
		# Transform recursively region surfaces and of all its children
		# including possible CLONEs
		#---------------------------------------------------------------
		def transformRegion(reg):
			#print(f"+++ transformRegion {reg.name()!r}")
			nonlocal matrix, invMatrix, transformed

			if reg.tag == "CLONE":
				# Apply an additional transformation
				self._geometry.region(reg[Card.ID], "matrix", invMatrix)

			for item in reg.expr().split():
				if item[0]=='#':
					if item not in transformed:
						transformRegion(regions[item[1:]])
						transformed.add(item)
				elif item[1:] not in transformed:
					surf = surfaces[item[1:]]
					if not isinstance(surf,int) and surf.tag == "SURFACE*":
						continue
					transformed.add(item[1:])
					sid = surf if isinstance(surf,int) else surf[Card.ID]
					self._geometry.body(sid, "matrix", matrix)
		#---------------------------------------------------------------
		# Load all modules/bodies [=regions]
		# and transform surfaces when needed
		for region in regions.values():
			expr = []
			#if region.name()=="6": breakpoint()
			matrix,_ = region.matrix()	# possible transformation matrix
			if matrix:
				invMatrix = matrix.inverse()
			else:
				invMatrix = None
			#print("-"*80)
			#print(f"{region.tag}={region.name()}")
			#if matrix:
			#	print(f"\tmatrix:\n{matrix}")
			#	print(f"\tmatrix-1:\n{invMatrix}")

			#-------------------------------------------------------
			# clone duplicate expression and associated bodies
			#-------------------------------------------------------
			if region.tag == "CLONE":
				if matrix is None:
					self.addError(f"Clone '{region.name()}' "
						"doesn't have any transformation associated")
					continue
				try:
					srcreg = regions[region.what(1)]
					#print(f"\tCLONE={region.what(1)}: {srcreg.expr()}")
				except KeyError:
					self.addError(f"Clone '{region.name()}' "
						f"refers to non-existing region '{region.what(1)}'")
					continue

				surf_matrix,_ = srcreg.matrix()
				#print(f"\tsurf_matrix:\n{surf_matrix}")

				region_expr = []
				for i,item in enumerate(srcreg.expr().split()):
					# skip bodies/modules
					if item[0]=='#': continue
					try:
						surf = surfaces[item[1:]]
					except KeyError:
						self.addError(f"Module '{srcreg.name()}' "
							f"refers to non-existing surface '{item[1:]}'")
						continue

					if surf.tag == "SURFACE*":
						region_expr.append(item)
						continue
					else:
						name = f"{surf.name()}_{region.name()}"
						# get the transformed what's
						btype = self._geometry.body(surf[Card.ID], "type")
						what  = self._geometry.body(surf[Card.ID], "what")
						m = matrix * bmath.Matrix(self._geometry.body(surf[Card.ID], "matrix"))
						#print(f"surface: {name} {what=}")
						surfaces[name] = self.addBody(name, btype, what, m)
						#print("    >>>:", self._geometry.body(surfaces[name],"what"))
						region_expr.append(f"{item[0]}{name}")

				#print(f"\tCLONE={region_expr}")
				region.setExpr(" ".join(region_expr))

				region[Card.ID] = rid = self._geometry.region(region.name(), "add")
				region[Card.ZONE] = self._geometry.region(rid, "expr", region.expr(), True)
				#mat = srcreg[Card.MATERIAL]
				#if mat is not None:
				#	self._geometry.region(rid, "material", mat[Card.ID])
				# define it as lattice
				self._geometry.region(rid, "type", Card.REGION_LATTICE)
				self._geometry.region(rid, "matrix", invMatrix)
				continue

			#-------------------------------------------------------
			# Normal expression
			#-------------------------------------------------------
			#print(f"\tEXPR={region.expr()}")
			region_expr = region.expr().split()

			transformed = set()
			# Expand expression for possible sub-modules
			for item in region_expr:
				if item[0]=='#':
					try:
						reg = regions[item[1:]]
						if matrix: transformRegion(reg)
						# FIXME expand the region
						expr.append("#(")
						for i in reg.expr().split():
							if i[0]=='#': continue
							expr.append(i)
						expr.append(")")
					except KeyError:
						self.addError(f"Module '{region.name()}' "
							f"refers to non-existing region '{item[1:]}'")
						continue
				else:
					if matrix and item[1:] not in transformed:
						try:
							surf = surfaces[item[1:]]
							if isinstance(surf,int) or surf.tag != "SURFACE*":
								transformed.add(item[1:])
								sid = surf if isinstance(surf,int) else surf[Card.ID]
								#print(f"Transform {item[1:]}")
								self._geometry.body(sid, "matrix", matrix)
						except KeyError:
							self.addError(f"Module '{region.name()}' "
								f"refers to non-existing surface '{item[1:]}'")
							continue
					expr.append(item)
			#print(f"\tEXPA= {' '.join(expr)}")

			region[Card.ID] = rid = self._geometry.region(region.name(), "add")
			region[Card.ZONE] = self._geometry.region(rid, "expr", ' '.join(expr), True)
			mat = region[Card.MATERIAL]
			if mat is not None:
				self._geometry.region(rid, "material", mat[Card.ID])

	# ----------------------------------------------------------------------
	# Display message if objects are locked
	# ----------------------------------------------------------------------
	def check4Frozen(self):
		frozen = False
		for card in self.project.input.cardlist:
			if card.ignore(): continue
			if card.get(Card.SELECT,0) & BIT.FREEZE:
				frozen = True
				break
		if frozen:
			self._geometry.message("Objects frozen", _FREEZE_COLOR)
		else:
			self._geometry.message("")
		self.draw()

	# ----------------------------------------------------------------------
	# project all viewports
	# ----------------------------------------------------------------------
	def viewportProject(self, asthread=False, all_=False):
		for view in self._viewers:
			view.project(asthread, all_)

	# ----------------------------------------------------------------------
	# Derive generation in all viewers
	# ----------------------------------------------------------------------
	def derive(self):
		self._geometry.derive()
		for view in self._viewers:
			view.derive()

	# ----------------------------------------------------------------------
	# Expose all viewports
	# ----------------------------------------------------------------------
	def expose(self, event=None):
		for view in self._viewers:
			view.expose()

	# ----------------------------------------------------------------------
	# Stop all threads
	# ----------------------------------------------------------------------
	def stopThreads(self):
		self._geometry.volume("stop")
		for frame in self.frames:
			frame.viewer.stop()

	# ----------------------------------------------------------------------
	# Redraw all viewers
	# ----------------------------------------------------------------------
	def draw(self, late=True):
		self.__afterDraw = None
		for frame in self.frames:
			frame.viewer.draw(late=late)

	# ----------------------------------------------------------------------
	# Fast Redraw only 3D viewers
	# ----------------------------------------------------------------------
	def drawFast3D(self):
		for _view in self._viewers:
			if _view.get("3D"):
				_view.draw(False, DRAW.FAST, True)

	# ----------------------------------------------------------------------
	# Draw viewers on idle, accumulate events
	# ----------------------------------------------------------------------
	def drawOnIdle(self):
		if self.__afterDraw is not None:
			self.after_cancel(self.__afterDraw)
		self.__afterDraw = self.after_idle(self.draw)

	# ----------------------------------------------------------------------
	# Force regeneration of projections & plots
	# ----------------------------------------------------------------------
	def redraw(self):
		#dump("*-* Geometry4Frame.redraw")
		# Update variables (just in case)
		self._geometry.color("background",	_BACKGROUND_COLOR)
		self._geometry.color("select",		_SELECT_COLOR)
		self._geometry.color("visible",		_VISIBLE_COLOR)
		self._geometry.color("region",		_REGION_COLOR)
		self._geometry.color("zone",		_ZONE_COLOR)
		self._geometry.color("error",		_ERROR_COLOR)
		self._geometry.color("wireframe",	_WIREFRAME_COLOR)
		self._geometry.color("bodybboxin",	_BODY_BBIN_COLOR)
		self._geometry.color("bodybboxout",	_BODY_BBOUT_COLOR)
		self._geometry.color("zonebbox",	_ZONE_BB_COLOR)
		self._geometry.color("regionbbox",	_REGION_BB_COLOR)
		self._geometry.color("title",		_TITLE_COLOR)
		self._geometry.color("gridtext",	_GRIDTEXT_COLOR)
		self._geometry.color("palettetext",	_PALETTETEXT_COLOR)
		self._geometry.color("label",		_LABEL_COLOR)
		self._geometry.color("lattice",		_LATTICE_COLOR)

		# refresh viewports
		for frame in self.frames:
			frame.viewer.set("backgroundcolor", _BACKGROUND_COLOR)
			frame.redraw()
		self.synchronizeViewports()

	# ----------------------------------------------------------------------
	# Fill the combo with the list of layers
	# ----------------------------------------------------------------------
	def fillLayers(self):
		for frame in self.frames:
			frame.layerCombo.fill(self.project.geometry.layersList())

	# ----------------------------------------------------------------------
	# Update all viewports with the new layer configuration
	# ----------------------------------------------------------------------
	def updateLayers(self, layer):
		self.fillLayers()
		for frame in self.frames:
			if layer == frame.layerCombo.get():
				frame.viewer.stop()
				frame.layer()
				frame.viewer.drawFast3D()
				frame.viewer.draw(late=True)

	# ----------------------------------------------------------------------
	def autoBodyDialog(self, event=None):
		if self._autoBody is None:
			self._autoBody = GeometryBody.GeometryAutoBody(self,
						self.editor, self.project)
		else:
			try:
				self._autoBody.deiconify()
				self._autoBody.lift()
			except:
				self._autoBody = GeometryBody.GeometryAutoBody(self,
					self.editor, self.project)

	# ----------------------------------------------------------------------
	def setOption(self, name, value):
		self.project.geometry[name] = value

	# ----------------------------------------------------------------------
	def getVar(self, name):
		return self.project.geometry.layers[0](name)

	# ----------------------------------------------------------------------
	def setVar(self, name, value):
		self.project.geometry.layers[0][name] = value

	# ----------------------------------------------------------------------
	# Add menus
	# ----------------------------------------------------------------------
	def _addObjectMenu(self, menu):
		menu.add_command(label="Arrow", underline=0, command=self.newArrow,
				compound=tk.LEFT, image=icons.get("arrow"))
		menu.add_command(label="Camera", underline=0, command=self.newCamera,
				compound=tk.LEFT, image=icons.get("movie"))
		menu.add_command(label="Light", underline=0, command=self.newLight,
				compound=tk.LEFT, image=icons.get("light"))
		menu.add_command(label="Mesh", underline=0, command=self.newMesh,
				compound=tk.LEFT, image=icons.get("mesh"))
		menu.add_command(label="Point", underline=0, command=self.newPoint,
				compound=tk.LEFT, image=icons.get("point"), accelerator="p")
		menu.add_command(label="Ruler", underline=0, command=self.newRuler,
				compound=tk.LEFT, image=icons.get("ruler"))
		menu.add_command(label="Spline", underline=0, command=self.newSpline,
				compound=tk.LEFT, image=icons.get("spline"))
		#menu.add_command(label="Rotdefi", underline=0, command=self.newRotdefi,
		#		compound=tk.LEFT, image=icons.get("axes"))

	# ----------------------------------------------------------------------
	def _visibilityMenu(self, menu):
		submenu = tk.Menu(menu)
		menu.add_cascade(label="Visibility", menu=submenu, underline=0,
				compound=tk.LEFT, image=icons.get("view"))
		submenu.add_command(label="Set",     underline=0,
				compound=tk.LEFT, image=icons.get("enable"),
				command=self.editor.setVisibility)
		submenu.add_command(label="Clear",   underline=0,
				compound=tk.LEFT, image=icons.get("disable"),
				command=self.editor.unsetVisibility)
		submenu.add_command(label="Toggle",  underline=0,
				compound=tk.LEFT, image=icons.get("toggle"),
				accelerator="v",
				command=self.editor.toggleVisibility)

	# ----------------------------------------------------------------------
	def _wireframeMenu(self, menu):
		submenu = tk.Menu(menu)
		menu.add_cascade(label="Wireframe", menu=submenu, underline=0,
				compound=tk.LEFT, image=icons.get("wireframe"))
		submenu.add_command(label="Show",     underline=0,
				compound=tk.LEFT, image=icons.get("wireframe"),
				command=self.editor.set3D)
		submenu.add_command(label="Hide",   underline=0,
				compound=tk.LEFT, image=icons.get("wireframe_off"),
				command=self.editor.unset3D)
		submenu.add_command(label="Toggle",  underline=0,
				compound=tk.LEFT, image=icons.get("toggle"),
				accelerator="#",
				command=self.editor.toggle3D)

	# ----------------------------------------------------------------------
	def _freezeMenu(self, menu):
		submenu = tk.Menu(menu)
		menu.add_cascade(label="Selection", menu=submenu, underline=0,
				compound=tk.LEFT, image=icons.get("freeze"))
		submenu.add_command(label="Freeze",     underline=0,
				compound=tk.LEFT, image=icons.get("freeze"),
				command=self.editor.freeze)
		submenu.add_command(label="UnFreeze",   underline=0,
				compound=tk.LEFT, image=icons.get("heat"),
				command=self.editor.unfreeze)
		submenu.add_command(label="Toggle",  underline=0,
				compound=tk.LEFT, image=icons.get("toggle"),
				accelerator="F",
				command=self.editor.toggleFreeze)

	# ----------------------------------------------------------------------
	def _addMenus(self, menu):
		bodymenu = tk.Menu(menu, tearoff=1)
		menu.add_cascade(label="Bodies", menu=bodymenu, underline=0,
				compound=tk.LEFT, image=icons.get("TRC"),
				accelerator="b")
		tkFlair.bodiesMenu(bodymenu, self.project.mode(), self.newBody)

		menu.add_command(label="Zone", underline=0,
				compound=tk.LEFT, image=icons.get("zone"),
				command=self.setActionZone,
				accelerator="d")

		menu.add_command(label="REGION", underline=0,
				compound=tk.LEFT, image=icons.get("REGION"),
				command=self.editor.addRegion,
				accelerator="R")

		menu.add_command(label="Rotation", underline=2,
				compound=tk.LEFT, image=icons.get("ROT-DEFI"),
				command=self.editor.addRotdefi)

		objmenu = tk.Menu(menu)
		menu.add_cascade(label="Object", underline=0, menu=objmenu,
				compound=tk.LEFT, image=icons.get("pin"))
		self._addObjectMenu(objmenu)

	# ----------------------------------------------------------------------
	def insertMenu(self, event=None):
		menu = tk.Menu(self, tearoff=0)
		self._addMenus(menu)
		menu.tk_popup(self.winfo_pointerx(), self.winfo_pointery())

	# ----------------------------------------------------------------------
	def insertBodyMenu(self):
		menu = tk.Menu(self, tearoff=1)
		tkFlair.bodiesMenu(menu, self.project.mode(), self.newBody)
		menu.tk_popup(self.winfo_pointerx(), self.winfo_pointery())

	# ----------------------------------------------------------------------
	def insertObjectMenu(self):
		menu = tk.Menu(self, tearoff=0)
		self._addObjectMenu(menu)
		menu.tk_popup(self.winfo_pointerx(), self.winfo_pointery())

	# ----------------------------------------------------------------------
	# Add objects
	# ----------------------------------------------------------------------
	def newObject(self, tag):
		self.setStatus(f"Adding {tag.capitalize()} object")
		self.actionOption = tag
		self.editor.setMouseAction(ACTION.ADD)

	#-----------------------------------------------------------------------
	def newArrow(self):   self.newObject("arrow")
	def newCamera(self):  self.newObject("camera")
	def newLight(self):   self.newObject("light")
	def newMesh(self):    self.newObject("mesh")
	def newPoint(self):   self.newObject("point")
	def newRotdefi(self): self.newObject("rotdefi")
	def newRuler(self):   self.newObject("ruler")
	def newSpline(self):  self.newObject("spline")

	# ----------------------------------------------------------------------
	def newBody(self, card=None):
		if card is not None: self.actionOption = card
		self.setStatus(f"Adding body \'{self.actionOption}\'")
		self.editor.setMouseAction(ACTION.ADD)

	# ----------------------------------------------------------------------
	# Action zone called from popup menu
	# ----------------------------------------------------------------------
	#def setActionNewZone(self):
	#	self.editor.propList.selectClear()
	#	self.editor.setMouseAction(ACTION.ZONE)

	# ----------------------------------------------------------------------
	def setActionZone(self):	self.editor.setMouseAction(ACTION.ZONE)
	def setActionMove(self):	self.editor.setMouseAction(ACTION.MOVE)
	def setActionRotate(self):	self.editor.setMouseAction(ACTION.ROTATE)
	def setActionRepeat(self):	self.editor.setMouseAction(ACTION.REPEAT)

	# ----------------------------------------------------------------------
	def setActionVolume(self):
		self.stopThreads()
		self.editor.setMouseAction(ACTION.VOLUME)

	# ----------------------------------------------------------------------
	def setActionExport(self):	self.editor.setMouseAction(ACTION.EXPORT)
	def setActionMovie(self):	self.editor.setMouseAction(ACTION.MOVIE)
	def setActionPlanner(self):	self.editor.setMouseAction(ACTION.PLANNER)
	def setActionLink(self):	self.editor.setMouseAction(ACTION.LINK)

	# ----------------------------------------------------------------------
	def setActionStop(self):
		self.stopThreads()
		self.loadProject()
		for frame in self.frames:
			frame.viewer.restoreSize()
			frame.viewer.loadCamera()
		self.setMouseAction(ACTION.SELECT)
		self.redraw()

	# ----------------------------------------------------------------------
	def timeframeShow(self):
		self.timeframe.pack(side=tk.BOTTOM, fill=tk.X)
		self.timeframe.show()
		#self.timeframe.update_idletasks()
		self.after_idle(self.timeframeUpdate)

	# ----------------------------------------------------------------------
	def timeframeHide(self):
		self.timeframe.pack_forget()

	# ----------------------------------------------------------------------
	def timeframeToggle(self, event=None):
		if self.timeframe.winfo_ismapped():
			self.timeframeHide()
		else:
			self.timeframeShow()

	# ----------------------------------------------------------------------
	def timeframeSet(self, frame):
		self.project.input.setLocalDict(frame=int(frame))

	# ----------------------------------------------------------------------
	# return active spline card and node
	# ----------------------------------------------------------------------
	def timeframeActiveSpline(self):
		if not self.timeframe.active: return None,None
		name, node = self.timeframe.active
		node += 1
		if ":" in name: name = name.split(":")[1]
		splineDict = self.project.input.cardsSorted("spline")
		try:
			return splineDict[name], node
		except KeyError:
			return None,None

	# ----------------------------------------------------------------------
	# When time frame current position changed inform to redraw screen
	# ----------------------------------------------------------------------
	def timeframeChange(self, event=None):
		self.timeframeSet(event.data)
		spline,node = self.timeframeActiveSpline()
		if spline is not None:
			spline[Card.NODE] = node
		self.reloadProjectOnIdle()

	# ----------------------------------------------------------------------
	# Active node is being dragged
	# ----------------------------------------------------------------------
	def timeframeActiveDragged(self, event=None):
		spline,node = self.timeframeActiveSpline()
		if spline is None: return
		# FIXME no undo.... somehow I need to collapse
		# the movement undo and into one
		spline[f"node:{node}.t"] = self.timeframe.current
		spline[Card.NODE] = node

	# ----------------------------------------------------------------------
	# Button 1 is being released
	# ----------------------------------------------------------------------
	def timeframeRelease(self, event):
		if not self.timeframe.active: return
		# FIXME print check nodes and re-order them if needed

	# ----------------------------------------------------------------------
	# timeframe update the selection if mapped
	# ----------------------------------------------------------------------
	def timeframeUpdate(self):
		if not self.timeframe.winfo_ismapped(): return
		if not self.__selcards: return
		keys = []
		splineDict = self.project.input.cardsSorted("spline")
		selected = set()	# selected keys
		for card in self.__selcards:
			if card.type() not in (Card.OBJECT, Card.BODY):
				continue

			if card.tag == "spline":
				n = card.intProperty("node:0")
				nodes = [card.intProperty(f"node:{x}.t")
								for x in range(1,n+1)]
				keys.append((f"{card.name()}",nodes))
				if card[Card.NODE]:
					selected.add((card.name(), card[Card.NODE]-1))
				continue

			# for all other objects
			splines = set()
			for w in card.whats():
				splinename = Card.isSplineStr(w)
				if splinename:
					splines.add(splinename)
			# check if there is a spline associated with the card
			if splines:
				for sname in sorted(list(splines)):
					try:
						spline = splineDict[sname]
						n = spline.intProperty("node:0")
						nodes = [spline.intProperty(f"node:{x}.t")
								for x in range(1,n+1)]
						if spline[Card.NODE]:
							selected.add((f"{card.name()}:{sname}",
									spline[Card.NODE]-1))
					except KeyError:
						nodes = []
					keys.append((f"{card.name()}:{sname}",nodes))
			else:
				keys.append((card.name(),[]))
		self.timeframe.setKeys(keys, selected)

	# ----------------------------------------------------------------------
	def timeframeHelp(self, event=None):
		Manual.show("","Movie_Tool")
		return "break"
