#
# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   16-Jun-2006

# FIXME
#  - displayCards()  don't use a cardlist but an attribute on the cards
#                    like @show = True/False
#                    This way when the display is stopped due to timeout
#                    and we click to expand, show only the filtered cards
#
#  - showIcons       Should be smart. Display list without icons which is
#                    way faster and then insert them later on with
#                    an after_idle() thread

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import sys
import time
import ctypes
import traceback
import tkinter as tk
from tkinter  import dnd, ttk
from operator import attrgetter
from collections import defaultdict

import log
#import tool
import icons
import tkFlair
import tkExtra

import awari
import inputs
import Global
import Manual
import Ribbon
import Project
import Validate
import InputExtra
import FlairRibbon
import MaterialsPage

from tkconst  import STATE
from deck     import Input, Card, CardInfo, Layout
from fluka    import Fluka, FlukaCard, FlukaLayout
from mcnp     import Mcnp, McnpLayout
from phits    import Phits, PhitsLayout
from moira    import Moira, MoiraLayout
from penelope import Penelope, PenelopeLayout
from pstext   import PSTextExporter

# FIXME enable for show debugging information on screen
DEBUG = False

# Global variables
_ALL            = "*all*"
_INDENT         = 48
_TABPOS         = [90+i*150 for i in range(6)]
_ACTIVE_BORDER  = 2
_TRIANGLE_DOWN  = " \N{BLACK DOWN-POINTING SMALL TRIANGLE}"

# Show flags
_SHOW_SCALE        = True
_SHOW_PREPROCESSOR = True
_SHOW_ICONS        = True
_SHOW_LINES        = False

_INSERT_COMMENT    = False
_BODY_INSERT       = True
_REGEXP            = False	# regular expression search

_KEY_TIMEOUT    = 10000		# after how much time to commit the change (undo)
_CARDS_TIMEOUT  = 3.0		# maximum time to show cards
_CARDS_UPDATE   = 0.5		# interval for updating display during long insertions

# Edit type enum
_EDIT_NORMAL    = 0
#_EDIT_NAME     = 0
_EDIT_LIST      = 1
_EDIT_MULTILINE = 2
_EDIT_FILE      = 3
_CARD_HEIGHT    = 4

# Colors
_ACTIVE_COLOR  = "LightYellow"
_COMMENT_COLOR = "Blue"
_DEFINE_COLOR  = "Orange"
_DIFF_COLOR    = "Yellow"
_DISABLE_COLOR = "DarkGray"
_ERROR_COLOR   = "Red"
_HIDDEN_BGCOLOR= "LightGray"
_HIDDEN_COLOR  = "DarkBlue"
_INVALID_COLOR = "MistyRose"
_LABEL_COLOR   = "#40B040"
_MODE_COLOR    = "DarkMagenta"
_NAME_COLOR    = "Magenta"
_NORMAL_COLOR  = "White"
_TAG_COLOR     = "DarkRed"
_CHILD_COLOR   = "Maroon"
_VALUE_COLOR   = "Black"

# Fonts
#_COMMENT_FONT = ("Sans", "12")
_COMMENT_FONT = ("Mono", "11")
_FIXED_FONT   = ("Mono", "11")
_HIDDEN_FONT  = ("Sans", "10", "italic")
_LABEL_FONT   = ("Sans", "10")
_TAG_FONT     = ("Sans", "12", "bold")
_CHILD_FONT   = ("Sans", "11")
_VALUE_FONT   = ("Sans", "11")

_ACTIVE_TAG   = tk.ACTIVE
_COMMENT_TAG  = "comment"
_DEFINE_TAG   = "define"
_DISABLE_TAG  = "disable"
_ERROR_TAG    = "error"
_HIDDEN_TAG   = "hidden"
_HR_TAG       = "hr"
_ID_TAG       = "id"
_INDENT_TAG   = "indent%d"
_LABEL_TAG    = "label"
_MODE_TAG     = "mode"		# different mode/type tag
_NAME_TAG     = "name"
_SELECT_TAG   = "select"
_TAG_TAG      = "tag"
_CHILD_TAG    = "child"
_TEXT_TAG     = "text"
_VALUE_TAG    = "value"


#===============================================================================
# MyDndHandler to avoid the problem of selecting the text during dragging
#===============================================================================
class MyDndHandler(dnd.DndHandler):
	#-----------------------------------------------------------------------
	def on_motion(self, event):
		super().on_motion(event)
		return "break"

	#-----------------------------------------------------------------------
	@staticmethod
	def start(source, event):
		h = MyDndHandler(source, event)
		return h if h.root else None

#===============================================================================
# Every CardWidget has the syntax
#	 {ID}{COMMENT | HIDDEN text}
#	 {TAG\t} [\t | {LABEL\t} {VALUE\t} ]...
#		{\t | {LABEL\t} {VALUE\t} ]...
#		...
# y marks the first line of the Widget, yend the next line
# The {ID} = id(self)
#===============================================================================
class CardWidgetBase:
	TAG = _HIDDEN_TAG

	# ----------------------------------------------------------------------
	def __init__(self, editor, card):
		self.editor = editor
		self.card   = None
		self.y      = 0
		self.yend   = 0

	# ----------------------------------------------------------------------
	def page(self):		return self.editor.page
	def input(self):	return self.editor.input
	def flair(self):	return self.editor.flair
	def dy(self):		return self.yend - self.y

	# ----------------------------------------------------------------------
	# Last draw time - set last time modified
	# ----------------------------------------------------------------------
	def modified(self):
		self._modifiedTime = time.time()

	# ----------------------------------------------------------------------
	# Return true if card is newer than the widget
	# ----------------------------------------------------------------------
	def isModified(self):
		return True

	# ----------------------------------------------------------------------
	# Update color state
	# ----------------------------------------------------------------------
	def updateState(self, validate=False):
		pass

	# ----------------------------------------------------------------------
	# edit Item, it should never be called here!
	# ----------------------------------------------------------------------
	#def editItem(self, n=None, select=True):

	# ----------------------------------------------------------------------
	def indent(self):
		pass

	# ----------------------------------------------------------------------
	# update label
	# ----------------------------------------------------------------------
	def updateLabel(self):
		self.page().updateLabel(self.card)

	# ----------------------------------------------------------------------
	def index(self):    return f"{self.y}.0"
	def indexEnd(self): return f"{self.yend}.0"

	# ----------------------------------------------------------------------
	# update end position, typically after text editing
	# ----------------------------------------------------------------------
	def updateEnd(self):
		#Global.dump("*-* updateEnd yend0=",self.yend)
		try:
			nextid = self.editor.tag_nextrange(_ID_TAG, f"{self.y}.end", tk.END)[0]
			self.yend = int(nextid.split(".")[0])
			#Global.dump("yend1=",self.yend)
		except IndexError:
			nextid = self.editor.index(tk.END)
			self.yend = int(nextid.split(".")[0])-1
					# XXX FIXME XXX redraw of the last
					# card changes the yend
			#Global.dump("yend2=",self.yend)

	# ----------------------------------------------------------------------
	# Check if y..yend correspond to the actual position of CardWidget
	# ----------------------------------------------------------------------
	def checkPosition(self):
		index = f"{self.y}.end"
		thisid = self.editor.tag_prevrange(_ID_TAG, index, '0.0')
		try:
			idx = int(self.editor.get(*thisid))
		except ValueError:
			if Global.developer:
				traceback.print_stack()
				Global.dump(f"{log.ERROR} checkPosition idx expected at {thisid!r} found {self.editor.get(*thisid)!r}")
				return
		if id(self) != idx:
			Global.dump(f"{log.ERROR} checkPosition: Wrong id self={id(self)}!={idx}")
			Global.dump(self.card)
			return False
		y = int(thisid[0].split(".")[0])
		if self.y != y:
			Global.dump(f"{log.ERROR} checkPosition: Wrong start position self={self.y}!={y}")
			Global.dump(self.card)
			return False
		try:
			nextid = self.editor.tag_nextrange(_ID_TAG, index, tk.END)[0]
			yend = int(nextid.split(".")[0])
		except IndexError:
			nextid = self.editor.index(tk.END)
			yend = int(nextid.split(".")[0])-1
			return True	# XXX FIXME XXX redraw of the last
					# card changes the yend
		if self.yend != yend:
			Global.dump(f"{log.ERROR} checkPosition: Wrong end position self={self.yend}!={yend}")
			Global.dump(self.card)
		return self.yend == yend

	# ----------------------------------------------------------------------
	# make visible
	# ----------------------------------------------------------------------
	def see(self, item=None, end=False):
		if item is None:
			if not end:
				self.editor.see(self.index()+"+1l")
			else:
				self.editor.see(self.indexEnd()+"-1l")
		else:
			# FIXME check the postList for a similar functionality
			self.editor.see(item)

	# ----------------------------------------------------------------------
	def item(self, index):
		return None

	# ----------------------------------------------------------------------
	# Position of starting card
	# ----------------------------------------------------------------------
	def fromPos(self):
		return None

	# ----------------------------------------------------------------------
	# Position of ending card
	# ----------------------------------------------------------------------
	toPos = fromPos

	# ----------------------------------------------------------------------
	# Commit changes
	# WARNING: Return True if LAYOUT has changed, else False
	# ----------------------------------------------------------------------
	def commit(self, timeout=False):
		return False

	# ----------------------------------------------------------------------
	# editNext
	# ----------------------------------------------------------------------
	def editNext(self):
		pass

	# ----------------------------------------------------------------------
	# editPrev
	# ----------------------------------------------------------------------
	def editPrev(self):
		pass

#===============================================================================
# Hidden card widget
#===============================================================================
class HiddenCardWidget(CardWidgetBase):
	# ----------------------------------------------------------------------
	def __init__(self, editor, fromCard, toCard, ypos=0):
		self.editor   = editor
		self.card     = None
		self.fromCard = fromCard
		self.toCard   = toCard
		self.create(ypos)

	# ----------------------------------------------------------------------
	# Return a list of cards
	# ----------------------------------------------------------------------
	def cards(self):
		return self.editor.input.cardlist[self.fromCard.pos() : self.toCard.pos()+1]

	# ----------------------------------------------------------------------
	# Position of starting card
	# ----------------------------------------------------------------------
	def fromPos(self):
		return self.fromCard.pos()

	# ----------------------------------------------------------------------
	# Position of ending card
	# ----------------------------------------------------------------------
	def toPos(self):
		return self.toCard.pos()

	# ----------------------------------------------------------------------
	def text(self):
		if self.fromCard is self.toCard:
			return f"----------- {self.fromCard.tag} : 1 card hidden-----------"
		else:
			return "----------- %s ... %s : %d cards hidden -----------" % \
				(self.fromCard.tag, self.toCard.tag,
				 self.toCard.pos() - self.fromCard.pos() + 1 )

	# ----------------------------------------------------------------------
	# create a HiddenCardWidget
	# Format: {ID}{text}
	# ----------------------------------------------------------------------
	def create(self, y):
		self.y    = y
		self.yend = y+1
		# First make space if we are inside an existing line
		self.editor.insert(f"{y}.0", "\n")
		index = f"{y}.end"
		self.editor.insert(index, str(id(self)), _ID_TAG)
		self.editor.insert(index, self.text(), self.TAG)
		self.modified()

	# ----------------------------------------------------------------------
	# redraw hidden card
	# @return if height has changed
	# ----------------------------------------------------------------------
	def redraw(self, y=None):
		assert self.editor["state"] == tk.NORMAL
		insert = self.editor.index(tk.INSERT)	# remember insert
		if y is not None:
			self.y    = y
			self.yend = y+1
		hidden_text = self.editor.tag_nextrange(self.TAG,
					self.index(), self.indexEnd())
		if hidden_text: self.editor.delete(*hidden_text)
		self.editor.insert(f"{self.y}.end", self.text(), self.TAG)
		self.modified()
		self.editor.mark_set(tk.INSERT, insert)	# set insert
		return False

	# ----------------------------------------------------------------------
	# edit Item
	# ----------------------------------------------------------------------
	def editItem(self, *args, **kwargs):
		self.editor.disable()

	# ----------------------------------------------------------------------
	# Change card state
	# ----------------------------------------------------------------------
	def setEnable(self, enable):
		undoinfo = []
		for card in self.cards():
			undoinfo.append(self.flair().setEnableUndo(card, enable))
		return undoinfo

#===============================================================================
# Basic Card Widget
#===============================================================================
class CardWidget(CardWidgetBase):
	TAG  = _TAG_TAG
	ICON = None		# ICON to be filled later when icons are loaded

	# ----------------------------------------------------------------------
	def __init__(self, editor, card, ypos=0):
		self.editor = editor
		self.card   = card
		self.layout = editor.layout[card.MODE].get(card)
		self._save  = (None, None)
		self.create(ypos)

	# ----------------------------------------------------------------------
	# Return true if card is newer than the widget
	# ----------------------------------------------------------------------
	def isModified(self):
		return self.card.modifiedTime() > self._modifiedTime

	# ----------------------------------------------------------------------
	# Return a list of cards
	# ----------------------------------------------------------------------
	def cards(self):
		return [self.card]

	# ----------------------------------------------------------------------
	# Position from starting card
	# ----------------------------------------------------------------------
	def fromPos(self):
		return self.card.pos()

	# ----------------------------------------------------------------------
	# Position from ending card
	# ----------------------------------------------------------------------
	toPos = fromPos

	# ----------------------------------------------------------------------
	# create editor items
	# don't use marks very slow
	# {ID}
	# {COMMENT,indent}
	# {TAG,indent} [{LABEL} {VALUE}]...
	#      [{LABEL,ident} {VALUE}]...
	# ----------------------------------------------------------------------
	def create(self, y):
		self.y = y
		card   = self.card
		editor = self.editor
		self._indent = card.indent()
		card.clearInvalid()
		indent = _INDENT_TAG%(self._indent)
		# Make space if line is not empty!!!
		editor.insert(f"{y}.0", "\n")
		editor.insert(f"{y}.0", f"{id(self)}\n", _ID_TAG)
		y += 1
		if _SHOW_LINES and y>1:
			editor.insert(f"{y}.0", "\n", _HR_TAG)
			y += 1
		# comment in blue
		s = card.comment()
		if s:
			if s[-1] != "\n": s += "\n"
			editor.insert(f"{y}.end", s, (_COMMENT_TAG,indent))
			y += s.count("\n")
		# Create the items
		for i,(lbl, dat, *_) in enumerate(self.layout):
			index = f"{y}.end"
			if self.layout[i] is Layout.newline:
				editor.insert(index, "\n\t", indent)
				y += 1
				index = f"{y}.end"
				continue

			if i==0:
				#Global.dump("tag=",card.tag,"alias=",card.info.alias)
				card_tag = card.info.alias if card.info is not CardInfo.NONE else card.tag
				if card_tag and card_tag[0] in ("#", "$"):
					tag = (indent, self.TAG, _DEFINE_TAG)
				else:
					#Global.dump(">>> card mode", card_tag,
					#		card.MODE, self.input().MODE)
					if card.MODE != self.input().mode:
						card_tag = f"{card.MODE}:{card_tag}"
						tag = (indent, self.TAG, _MODE_TAG)
					else:
						tag = (indent, self.TAG)
				# Add the image
				if editor.showIcons:
					# FIXME maybe get it from info
					try:    icon = icons.get(card.tag)
					except: icon = self.ICON
					editor.image_create(index, image=icon)
					editor.tag_add(indent,f"{y}.0",index)
				editor.insert(index, f" {card_tag}\t\t", tag)
			elif lbl is None and dat is None:
				editor.insert(index,"\t\t", _LABEL_TAG)
				continue
			elif lbl is not None:
				editor.insert(index, f"{lbl}:\t", _LABEL_TAG)
			else:
				editor.insert(index, "\t", _LABEL_TAG)

			# Create data entry
			tag = _VALUE_TAG
			dataerror = False
			if isinstance(dat, tuple):
				datatxt = str(self.listValue(dat)) + _TRIANGLE_DOWN
				if datatxt[0] == '?':
					dataerror = True
					datatxt = datatxt[1:]
			elif callable(dat):
				datatxt = dat(editor.layout[card.MODE], card)
				if datatxt is None:
					datatxt = ""
				else:
					datatxt = str(datatxt)
			elif isinstance(dat, int):
				datatxt = str(card.what(dat))
				if dat==0 and lbl is None:
					tag = (_VALUE_TAG, _NAME_TAG)
				elif dat==-1:
					y += datatxt.count("\n")
					tag = (_VALUE_TAG, _TEXT_TAG)
			elif isinstance(dat, str):
				datatxt = str(card.get(dat,""))
			else:
				datatxt = ""
				tag = _LABEL_TAG

			# Highlight error
			if dataerror and card.enable:
				tag = (tag, _ERROR_TAG)

			editor.insert(index, datatxt+"\t", tag)
		self.yend = y+1
		if not card.enable or card.invalid:
			self.updateState()
		self.modified()		# record when it was drawn

	# ----------------------------------------------------------------------
	# redraw card
	# force re-generation of the card
	# @return if height has changed
	# ----------------------------------------------------------------------
	def redraw(self, y=None):
		#Global.dump("*-* CardWidget.redraw")
		assert self.editor["state"] == tk.NORMAL
		try:
			indexEnd = self.editor.tag_nextrange(
						_ID_TAG,
						f"{self.y}.end",
						tk.END)[0]
		except:
			indexEnd = self.editor.index(tk.END + "-1c")

		self.editor.delete(self.index(), indexEnd)
		self.create(self.y if y is None else y)

		if self.editor.active is self:
			self.editor.tag_add(_ACTIVE_TAG, self.index(), self.indexEnd())
			self.editor.tag_add(_SELECT_TAG, self.index(), self.indexEnd())
		return self.yend != self.y	# needs repositioning

	# ----------------------------------------------------------------------
	# correct indent level of the card if needed
	# ----------------------------------------------------------------------
	def indent(self):
		if self._indent == self.card.indent(): return
		oldIndent = _INDENT_TAG%(self._indent)
		newIndent = _INDENT_TAG%(self.card.indent())
		self._indent = self.card.indent()
		f   = self.index()
		end = self.indexEnd()
		# find all indent tags and replace them with the new one
		while True:
			try:
				f,t = self.editor.tag_nextrange(oldIndent, f, end)
			except ValueError:
				return
			self.editor.tag_remove(oldIndent, f,t)
			self.editor.tag_add(newIndent, f,t)

	# ----------------------------------------------------------------------
	# find list value
	# ----------------------------------------------------------------------
	def listValue(self, dat):
		(w, lst) = dat
		if callable(w):
			# Call function to return result
			v = vn = w(self.editor.layout[self.card.MODE], self.card)
		elif isinstance(w,int):
			# Get it from card directly
			vn = self.card.what(w)			# Actual what
			v  = self.card.evalWhat(w,True)		# Evaluated
		else:
			vn = self.card.get(w,"")
			v = vn

		# Dynamic lists
		if callable(lst):
			lst = lst(self.editor.layout[self.card.MODE])

		if not isinstance(lst, list): return vn

		# Convert to integer if necessary
		try:
			vf = float(v)
			vi = int(vf)
			if vf == float(vi): v = vi
		except (ValueError, TypeError):
			pass

		# Check also length 8
		try:
			v8 = v[:8]
		except TypeError:
			v8 = None

		# Make first a fast check
		if v in lst or vn in lst or (v8 is not None and v8 in lst) or str(v) in lst:
			return vn
		else:
			# check for tuples in list
			for item in lst:
				if isinstance(item,tuple) and \
				   (v in item or (v8 is not None and v8 in item)):
					return item[0]

		# Not found in list, issue an error
		if v or not lst:
			# FIXME Accept patterns with * and ranges with ".."
			#if self.card.enable and ("*" not in v or ".." not in v):
			if self.card.notIgnore() and isinstance(v,str) and "*" not in v:
				msg = f"item '{v}' do not exist"
				try:
					int(w)
					self.card.addInvalid(w)
				except:
					self.card.addInvalid(msg)
				self.flair().write(f"{log.WARNING}Card #{self.card.pos1()} {msg}\n")
				cardstr = str(self.card).replace("\n",f"\n{log.WARNINGCONT}")
				self.flair().write(f"{log.WARNINGCONT}{cardstr}\n")
			return f"?{vn}"

		item = lst[0]
		if isinstance(item, tuple):
			return item[0]
		else:
			return item

	# ----------------------------------------------------------------------
	# change the layout
	# ----------------------------------------------------------------------
	def updateLayout(self):
		layout = self.editor.layout[self.card.MODE].get(self.card)
		if self.layout is layout: return False
		# Compare layouts to see if they are identical
		if len(self.layout) == len(layout):
			equal = True
			for i in range(len(self.layout)):
				if self.layout[i][0] != layout[i][0]:
					equal = False
					break
			if equal: return False
		self.layout = layout
		return True

	# ----------------------------------------------------------------------
	# highlight label
	# ----------------------------------------------------------------------
	def highlightLabel(self, what=None):
		self.page().highlightLabel(what)

	# ----------------------------------------------------------------------
	# Insert comment
	# ----------------------------------------------------------------------
	def insComment(self):
		if self.editor.tag_nextrange(_COMMENT_TAG, self.index(), self.indexEnd()):
			return False
		idtag = self.editor.tag_nextrange(_ID_TAG, self.index(), self.indexEnd())
		if not idtag: return False
		self.editor.insert(idtag[1], "\n", (_COMMENT_TAG, _INDENT_TAG%(self._indent)))
		self.yend += 1
		self.modified()
		return True

	# ----------------------------------------------------------------------
	# Change card state
	# ----------------------------------------------------------------------
	def setEnable(self, enable):
		undoinfo = self.flair().setEnableUndo(self.card, enable)
		self.updateState()
		return undoinfo

	# ----------------------------------------------------------------------
	# change colors to display card state
	# ----------------------------------------------------------------------
	def updateState(self, validate=False):
		indexEnd = self.indexEnd()
		self.editor.tag_remove(_ERROR_TAG, self.index(), indexEnd)

		if not self.card.enable:	# Disabled
			self.editor.tag_add(_DISABLE_TAG, self.index(), indexEnd)
			return

		# Enabled
		self.editor.tag_remove(_DISABLE_TAG, self.index(), self.indexEnd())
		# Check for errors in case they exist
		err = None
		if self.card.invalid or validate:
			err  = self.card.validate(self.card.case())
		if not err: return

		# Highlight errors
		tagrange = self.editor.tag_nextrange(	self.TAG,
							self.index(),
							indexEnd)
		for lbl,dat,*_ in self.layout:
			if dat is None or not tagrange: continue
			tagrange = self.editor.tag_nextrange(
							_VALUE_TAG,
							tagrange[1],
							indexEnd)
			if isinstance(dat,tuple):
				w = dat[0]
			elif isinstance(dat,int):
				w = dat
			else:
				w = None

			if err[0] is None or w in err:
				self.editor.tag_add(_ERROR_TAG, *tagrange)

	# ----------------------------------------------------------------------
	# find item from index in the editor
	# ----------------------------------------------------------------------
	def item(self, index):
		tags = self.editor.tag_names(index)
		if _COMMENT_TAG in tags: return -1
		# Check for labels and values
		if _VALUE_TAG not in tags: return None
		# Scan value/name tags, starting from TAG until index
		item = -1

		#---------------------------------------------------------------
		def _scan(k, v, i):
			nonlocal item
			if k=="tagon" and v == _VALUE_TAG:
				item += 1
				try:
					while self.layout[item][1] is None:
						item += 1	# Skip empty data fields
				except IndexError:
					# FIXME It should never happen
					log.error(f"Invalid leyout item {item}")
					log.errorCont(f"Card: {self.card}")
					log.errorCont(f"Layout: {self.layout}")
		#---------------------------------------------------------------
		tagindex = self.editor.tag_nextrange(
						self.TAG,
						self.index(),
						self.indexEnd())[1]
		self.editor.dump(tagindex, index+"+1c", command=_scan)
		if item<0: return None
		return item

	# ----------------------------------------------------------------------
	# Find starting index of item
	# ----------------------------------------------------------------------
	def itemRange(self, item):
		if item == -1:	# Comment
			try:
				return self.editor.tag_nextrange(
							_COMMENT_TAG,
							self.index(),
							self.indexEnd())
			except IndexError:
				return ()

		index    = self.index()
		indexEnd = self.indexEnd()
		index    = self.editor.tag_nextrange(
						self.TAG,
						index,
						indexEnd)[1]
		for i,(_,dat,*_) in enumerate(self.layout):
			if dat is None: continue
			tagrange = self.editor.tag_nextrange(
							_VALUE_TAG,
							index,
							indexEnd)
			if i==item: return tagrange
			index = tagrange[1]
		return ()

	# ----------------------------------------------------------------------
	# @return text from item
	# ----------------------------------------------------------------------
	def itemText(self, item):
		tagrange = self.itemRange(item)
		if tagrange:
			# XXX when it is a list it will contain the triangle
			return self.editor.get(*tagrange)[:-1]
		else:
			return ""

	# ----------------------------------------------------------------------
	# edit Item
	# @param n	item to edit cursor should be in the location
	# @param select	select item
	# FIXME Requires that INSERT is in items location!!!!!
	# ----------------------------------------------------------------------
	def editItem(self, n=None, select=True, postlist=True):
		Global.dump(f"*-* InputPage.editItem n={n} select={select} postlist={postlist} "
			f"edit={self.editor.edit}")
		if self.editor.edit == n and n is not None:
			self.editor.enable()
			return
		self.editor.editLast = self.editor.edit
		self.editor.edit = n

		tags = self.editor.tag_names(tk.INSERT)
		if n is None or _LABEL_TAG in tags:	# FIXME for the moment
			self.editor.disable()
			self.editor.unpostList()
			return

		# FIXME Verify if tk.INSERT gives back the same item()!!!
		try:
			assert self.item(tk.INSERT) == n
		except:
			Global.dump(">e> ASSERT failed editItem:")
			traceback.print_stack()
			Global.dump("\ttk.INSERT=",  self.editor.index(tk.INSERT))
			Global.dump("\ttk.CURRENT=", self.editor.index(tk.CURRENT))
			Global.dump("\titem=",       self.item(tk.INSERT))
			Global.dump("\ttags=",       self.editor.tag_names(tk.INSERT))

		self.editor.enable()
		self.editor.tag_remove(tk.SEL, "1.0", tk.END)
		self.editor.see(tk.INSERT)

		lst = None
		if n<0:
			# Comment
			self.editType = _EDIT_MULTILINE
		else:
			dat  = self.layout[n][1]
			what = None
			if isinstance(dat, int):
				what = dat
			elif isinstance(dat, tuple):
				# create an option menu
				(var, lst) = dat
				if isinstance(var, int): what = var
				if callable(lst):
					lst = lst(self.editor.layout[self.card.MODE])
				w = dat[0]
				if callable(w):
					sel = w(self.editor.layout[self.card.MODE], self.card)
				elif isinstance(w,int):
					sel = self.card.what(w)
				else:
					sel = self.card.get(w,"")
			self.highlightLabel(what)

		# Find limits of datafield
		for tag in (_COMMENT_TAG, _VALUE_TAG):
			if tag in tags:
				self.editor.editTag = tag
				break
		else:
			self.editor.disable()
			self.editor.edit = None
			return

		# Check both previous and next range to see in which one the index is inside
		tagrange = self.editor.tag_range(self.editor.editTag, tk.INSERT)
		if not tagrange:
			self.editor.disable()
			self.editor.edit = None
			return

		if lst:
			# save value (without the down arrow and tab)
			self._save = (self.editor.edit, self.editor.get(*tagrange)[:-3])
			if isinstance(lst, list):
				self.editor.editType = _EDIT_LIST
				if postlist:
					self.editor.postList(lst, sel)
				else:
					self.editor.popupList.fill(lst, sel)
				select = True
			elif isinstance(lst, tuple):
				self.editor.editType = _EDIT_FILE
				# FIXME create a new type Layout.File
				#       that holds the information
				# Opening bFiledialog FocusOut will send a commit
				# and reset the text.edit to None
				# WARNING opening the dialog
				# filename is saved in InputExtra.filename
				InputExtra.openFile(self.editor, self._save[1], lst)
				# commit the change to force setting the card
				self.commit()
				self.editItem()
				select = False
			else:
				raise Exception(f"Invalid layout defined card: {self.card.tag}")
		else:
			# save value
			self.editor.editType = _EDIT_LIST
			self._save = (self.editor.edit, self.editor.get(*tagrange)[:-1])
			if n<0 or (isinstance(dat,int) and dat<0):
				self.editor.editType = _EDIT_MULTILINE
			else:
				self.editor.editType = _EDIT_NORMAL
		if select:
			self.editor.selectAll()

	# ----------------------------------------------------------------------
	# edit next item
	# ----------------------------------------------------------------------
	def editNext(self):
		item = self.editor.edit
		if item is None:
			# try last item edited
			item = self.editor.editLast
			if item is not None:
				try:	# does it exist?
					self.layout[item][1]
					tagrange = self.itemRange(item)
					self.editor.mark_set(tk.INSERT, tagrange[0])
					self.editItem(item)
					return
				except IndexError:
					pass
			item = None
		tagrange = ()
		if item is None:
			# try the comment
			item = -1
			tagrange = self.itemRange(item)
		if not tagrange:
			start_item = item
			while True:
				item += 1
				if item >= len(self.layout): item = 0
				if self.layout[item][1] is not None: break
				if item == start_item: return	# none found
				start_item = max(start_item, 0)

		tagrange = self.itemRange(item)
		self.editor.mark_set(tk.INSERT, tagrange[0])
		self.editItem(item)

	# ----------------------------------------------------------------------
	# edit previous item
	# ----------------------------------------------------------------------
	def editPrev(self):
		item = self.editor.edit
		if item is None: item = 0
		start_item = item
		while True:
			item -= 1
			if item < 0: item = len(self.layout)-1
			if self.layout[item][1] is not None: break
			if item == start_item: return # none found

		tagrange = self.itemRange(item)
		self.editor.mark_set(tk.INSERT, tagrange[0])
		if isinstance(self.layout[item][1], tuple):
			self.editor.mark_set(tk.INSERT, tagrange[1]+"-3c")
		else:
			self.editor.mark_set(tk.INSERT, tagrange[1]+"-1c")
		self.editItem(item)

	# ----------------------------------------------------------------------
	# Commit changes to active card
	# @return True if LAYOUT has changed, else False
	# timeout = True is the committing is done due to timeout (see handleKey)
	#           False(default) when committing is due to a user action
	# ----------------------------------------------------------------------
	def commit(self, timeout=False):
		Global.dump("*-* CardWidget.commit")
		# active can be different than self
		active  = self.editor.active
		edit    = self.editor.edit
		changed = False

		# Comment?
		if edit is None or active is None:
			return False
		elif edit < 0:
			val = self.itemText(edit)
			if val != active._save[1]:
				self.editor.page.inputModified()
				sval = str(val)
				#if len(val)==0:
				#	active.redraw()
				#	#reposition()??
				self.editor.addUndo(
					self.editor.flair.setCommentUndo(active.card, sval))
			return changed

		try:
			lbl, dat, *desc = active.layout[edit]
		except IndexError:
			Global.dump(f"+++ card {active.card}")
			Global.dump(f"+++ edit={edit}")
			return False
		Global.dump(f"+++ edit={edit} lbl={lbl} dat={dat} *desc={desc}")

		# ---------------------------
		# --- Listbox or filename ---
		# ---------------------------
		if isinstance(dat, tuple):
			self.card.clearInvalid()
			if isinstance(dat[1], tuple):	# Filename
				txt = val = self.editor.project.relpath(InputExtra.filename)
				# if user hits cancel then do nothing
				if txt == "": return False
				if active.card.tag == "VOXELS": # XXX strip extension
					# Special case
					fn,ext = os.path.splitext(val)
					if ext == ".vxl":
						txt = val = fn
			else:	# selection list
				txt,val = self.editor.popupList.selection()
				if txt is None:
					# No match found
					# Accept edited text instead of list selection
					# only if it starts with = or $ and is not
					# contains in the selection list
					ptxt = self.itemText(edit)
					if ptxt and (ptxt[0] in "$=" or "*" in ptxt):
						txt = val = ptxt[:-2]
				else:
					txt = self.itemText(edit)

			if txt is not None: # Text, Value
				txt = str(txt) + _TRIANGLE_DOWN
				w = dat[0]
				if callable(w):
					if w(self.editor.layout[self.card.MODE], active.card) != val:
						w(self.editor.layout[self.card.MODE], active.card, val)
						changed = True
				elif active.card.get(w,"") != str(val):
					self.editor.addUndo(
						self.editor.flair.setCardUndo(
							active.card,
							w,
							str(val),
							""))
					changed = True
				if changed:
					self.editor.enable()
					active.redraw()
					self.flair().refresh("card", self.editor.page)

			# Check for a possible change in the layout
			if active.card.tag == "#include":
				self.editor.refresh(force=True)
			elif active.card.tag == "VOXELS":
				if dat[0]==0:	# Special case!!!
					# Load new voxel file
					active.card.loadVoxel()
					self.editor.refresh(force=True)
			else:
				self.editor.page.inputModified()
				# If the layout was changed then then redraw card
				if active.updateLayout() or txt is None:
					self.editor.enable()
					if active.redraw():
						self.editor.reposition(
							self.editor.activePos())
					self.editor.disable()

		# ------------------------
		# --- Normal edit text ---
		# ------------------------
		else:
			val = self.itemText(edit)
			if self.editor.editType == _EDIT_LIST:
				val = val[:-2]
			if val != active._save[1]:
				changed = True
				if callable(dat):
					if not timeout:
						dat(self.editor.layout[self.card.MODE], active.card, val)
						active.redraw()
				else:
					Global.dump(f">>> what={dat} {active.card.get(dat,'')!r} val={val!r}")
					# Check if value didn't change
					if not timeout and dat==0 and val is not None:
						Global.dump(">>> setCardUndo(), changeName()")
						undoinfo = [self.editor.flair.setCardUndo(
									active.card,
									dat,
									str(val),
									"")]
						undoinfo.extend(active.changeName(
									active._save[1],
									val))
						self.editor.addUndo(undoinfo, undoinfo[0][0])
					elif active.card.get(dat,"") != str(val):
						Global.dump(">>> setCardUndo()")
						self.editor.addUndo(self.editor.flair.setCardUndo(
								active.card,
								dat,
								str(val),
								""))
				if not timeout:
					active._save = (active._save[0], val)
				self.editor.page.inputModified()
				self.flair().refresh("card", self.editor.page)

				# if a body name has changed
				if dat==0 and active.card.tag in \
						CardInfo.bodies[self.editor.project.mode()]:
					self.input().clearCache(Card.BODY)
				self.input().clearCache(active.card.tag)

		active.updateState(True)
		active.updateLabel()
		return changed

	# ----------------------------------------------------------------------
	# check for a possible name change in the card
	# FIXME should be project mode dependent
	# FIXME get the information from the range definition!!!!
	# add SOLID.xxx VOLU.xxx
	# ----------------------------------------------------------------------
	def changeName(self, old, new):
		if old=="" or old is None: return []
		tag = self.card.tag

		# Check card type
		kind = self.card.info.define
		if not kind:
			return []

		lst = self.input().cardsDefineType(kind)

		# Check for duplicate (apart from ROT-DEFI where duplicated are permitted!)
		# FIXME
		# Find dynamically list of a type
		if tag=="ROT-DEFI" or len([x for x in lst if x.name()==old])>=1:
			change = tkFlair.askyesno("Multiple defined name changed",
					f"{kind} name has changed,\n"
					f"{old} -> {new}\n"
					 "but it is defined more than once\n"
					 "Change all references in input?",
					parent=self.editor)
		else:
			change = True

		if change and new and old:
			self.input().clearCache()
			self.editor.layoutCache(False)
			undoinfo = [self.flair().changeNameUndo(kind, old, new)]
			self.flair().notify("Name change",
				f"Changed all references of {kind}\n{old} -> {new}")
			self.editor.refresh()
			self.editor.page._refreshOtherFrames()
			return undoinfo
		else:
			return []

	# ----------------------------------------------------------------------
	# Find selected item from list and inserted as text to the active item
	# at the cursor position. Mark it also as selected. Move the cursor
	# either to the right or the left(unchanged) of the insertion point
	# ----------------------------------------------------------------------
	def insertText(self, text):
		try: body = text.split('[')[0]
		except: return
		self.editor.tag_remove(tk.SEL, "1.0", tk.END)
		endindex = self.editor.index(self.editor._insertText+" wordend")
		text = self.editor.get(self.editor._insertText, endindex)
		if text.strip():	# If we typed nothing the wordend will contain the ending Tab
			self.editor.delete(self.editor._insertText, endindex)
		self.editor.pasteText(body)
		if body == Layout.BODY_PARENTHESIS:
			self.editor.mark_set(tk.INSERT, tk.INSERT+"-1c")

#===============================================================================
# Continuation Card Widget
#===============================================================================
class ChildCardWidget(CardWidget):
	TAG  = _CHILD_TAG
	ICON = None		# ICON to be filled later when icons are loaded

	#-----------------------------------------------------------------------
	def __init__(self, editor, card, ypos=0):
		super().__init__(editor, card,  ypos)

#===============================================================================
# Input Canvas filled with CardWidget
#===============================================================================
class InputEditor(tk.Text):
	def __init__(self, master, page, *arg, **kw_args):
		super().__init__(master, *arg, **kw_args)
		self["wrap"] = tk.NONE
		self["font"] = _VALUE_FONT	# default font

		self._tabs = _TABPOS[:]

		# Master link
		self.page  = page
		self.flair = page.flair
		self.input = None

		# Create input editor layouts
		self.layout = {}
		self.layout[Input.MODE] = Layout(self.flair)
		self.layout[Fluka.MODE] = FlukaLayout(self.flair)
		if Global.developer:
			self.layout[Moira.MODE]    = MoiraLayout(self.flair)
			self.layout[Mcnp.MODE]     = McnpLayout(self.flair)
			self.layout[Phits.MODE]    = PhitsLayout(self.flair)
			self.layout[Penelope.MODE] = PenelopeLayout(self.flair)

		# Interrupt long process
		self._interrupt = False

		# Undo/Redo list
		self.lastKey  = (-1,0)

		# Card list
		self.cardlist       = []
		self._resizeWidth   = 0
		self._highlightItem = None

		# Active field
		self.active      = None		# Active card widget
		self.cardAnchor  = None		# Card selection range
		self._anchor     = None		# button click position for selection
						# FIXME replace with self.anchorName()
		self._selword    = False	# False = character, True = Word
		self.edit        = None		# editing field in active card
		self.editTag     = None		# currently editing tag
		self.editType    = _EDIT_NORMAL	# Editing item type
		self.editLast    = None		# Last edited item
		self.selection   = set()	# selection set
		self._autoId     = ""		# dragging the mouse outside window
		self._insertText = None		# insert body starting position
		self.dnd_time    = None		# Potential dnd start time
		self.dnd_item    = None		# Drag item
		self.dnd_active  = None		# Old active
		self.dnd_label   = None		# Label to show
		self.scandrag    = None		# Scan drag
		self.showIcons   = _SHOW_ICONS	# Display icons on cards
		self.delay       = tkExtra.Balloon.delay	# Delay from Balloon
		self._motionX    = 0
		self._motionY    = 0
		self._key        = ""		# key buffer for multi key commands

		# Popup list
		self.popupList  = InputExtra.SelectionListFrame(self,
						cursor="arrow",
						relief=tk.RAISED,
						borderwidth=1)

		# Editor bindings
		self.bind('<Button-1>',		self.button1)
		self.bind('<ButtonRelease-1>',	self.release1)
		self.bind('<Double-1>',		self.double1)
		self.bind('<Triple-1>',		self.triple1)
		self.bind('<B1-Leave>',		self.autoRepeat)
		self.bind('<B1-Enter>',		self.cancelRepeat)
		self.bind('<B2-Motion>',	self.scanDrag)
		self.bind('<ButtonRelease-2>',	self.scanRelease)
		self.bind('<Button-3>',		self.popupMenu)
		self.bind('<Control-Button-3>',	self.popupInsertMenu)
		self.bind('<Button-4>',		lambda e,s=self: s.yview(tk.SCROLL, -1, tk.UNITS))
		self.bind('<Button-5>',		lambda e,s=self: s.yview(tk.SCROLL,  1, tk.UNITS))
		self.bind('<MouseWheel>',	lambda e,s=self: s.yview(tk.SCROLL,
							-e.delta//120, tk.UNITS))

		self.bind('<Key>',		self.handleKey)
		#self.bind('<KeyPress>',	self.keyPress)
		#self.bind('<KeyRelease>',	self.keyRelease)
		self.bind('<Tab>',		self.tabKey)
		self.bind('<Shift-Tab>',	self.tabShiftKey)
		self.bind('<Control-Tab>',	self.flair.tabs.activatePrevious)
		if sys.platform in ("linux","linux2"):
			self.bind('<ISO_Left_Tab>', self.tabShiftKey)

		self.bind('<Control-Key-slash>',self.selectAll)

		# Ignore events are handled by the class_bind in flair.py
		self.bind('<Control-Key-Return>',self.popupInsertMenu)
		self.bind('<<MoveUp>>',		self.moveUp)
		self.bind('<Control-Key-k>',	self.moveUp)
		self.bind('<<MoveDown>>',	self.moveDown)
		self.bind('<Control-Key-j>',	self.moveDown)
		self.bind('<Control-Key-space>',self.popupMenu)
		self.bind("<Insert>",		self.page.add)
		self.bind('<Control-Key-a>',	self.selectAll)
		self.bind('<Control-Key-A>',	self.selectClear)
		self.bind('<<Clone>>',		self.clone)
		self.bind('<<CloneSpecial>>',	self.cloneSpecial)
		self.bind("<Control-Key-f>",	self.page.focusSearch)
		self.bind("<Control-Key-F>",	self.page.focusSearch)
		self.bind("<Control-Key-e>",	self.page.editCard)
		self.bind("<Control-Key-g>",	self.page.find)
		self.bind("<Control-Key-G>",	self.page.find)
		self.bind("<Control-Key-h>",	self.replace)
		self.bind("<Control-Key-H>",	self.replace)

		self.bind('<Control-Key-i>',	self.toggleShow)
		self.bind('<Control-Key-h>',	self.collapseCards)
		self.bind('<Control-Key-Left>',	self.collapseCards)
		self.bind('<Control-Key-Right>',self.expandCards)
		# Ignore and let Flair Ribbon Tab to process them
		self.bind('<Alt-Control-Key-Left>',	self.doNothing)
		self.bind('<Alt-Control-Key-Right>',	self.doNothing)
		self.bind('<Alt-Control-Key-h>',	self.doNothing)
		self.bind('<Alt-Control-Key-l>',	self.doNothing)

		self.bind('<Control-Key-l>',	self.toggleState)
		self.bind('<Control-Key-L>',	self.toggleState)
		self.bind('<Control-Key-m>',	self.toggleComment)
		self.bind("<<Hardcopy>>",	self.page.hardcopy)
		#self.bind("<Control-Key-t>",	self.page.transformGeometry)
		#self.bind("<Control-Key-T>",	self.page.transformGeometry)
		self.bind("<<Splitter>>",	self.page.toggleSplitter)
		self.bind('<Leave>',		self.hideBalloon)
		self.bind('<FocusOut>',		self._focusOut)
		#self.bind('<Configure>',	self.resize)

		# The flair bindings are on top level which is the lowest priority
		# widget -> class -> toplevel
		# Warnings that class has global binding in tkExtra to override
		# the paste behaviour
		self.bind("<<Paste>>",		self.paste)
		self.bind("<<PasteSpecial>>",	self.paste)
		#self.bind("<<PasteSelection>>",	self.paste2)
		#self.bind("<<Undo>>",		self.undo)
		#self.bind("<<Redo>>",		self.redo)
		self.bind('<<Reposition>>',	self.repositionEvent)
		#self.bind('<<Selection>>',	self.selectionChanged)
		self.popupList.bind("<<Selected>>", self.popupListSelected)
		self.motionBind()
		self.disable()
		self.setTabs()
		self.tagConfigure()
		#self["exportselection"] = 1

	# ----------------------------------------------------------------------
	def doNothing(self, event=None):
		pass

	# ----------------------------------------------------------------------
	# GENERIC Text additions
	# ----------------------------------------------------------------------
	def anchorName(self):
		return self.tk.call("::tk::TextAnchor", self._w)

	# ----------------------------------------------------------------------
	# reset selection text anchor (call taken from text.tcl)
	# it should be called on before any selection
	# ----------------------------------------------------------------------
	def setAnchor(self, index=None):
		if index is None:
			self.mark_unset(self.anchorName())
		else:
			self.mark_set(self.anchorName(), index)

	# ----------------------------------------------------------------------
	# @return tag range under index
	# ----------------------------------------------------------------------
	def tag_range(self, tag, index):
		# Check both previous to see in which one the index is inside
		tagrange = self.tag_prevrange(tag, index, "1.0")
		if tagrange and \
		   self.compare(tagrange[0], "<=", index) and \
		   self.compare(index,       "<=", tagrange[1]):
			return tagrange
		# ... and next
		tagrange = self.tag_nextrange(tag, index, tk.END)
		if tagrange and \
		   self.compare(tagrange[0], "<=", tk.INSERT) and \
		   self.compare(tk.INSERT,   "<=", tagrange[1]):
			return tagrange
		return ()

	# ----------------------------------------------------------------------
	# @return text at index position marked by tag
	# ----------------------------------------------------------------------
	def tag_get(self, tag, index):
		tagrange = self.tag_range(tag, index)
		if tagrange:
			return self.get(*tagrange)
		else:
			return ""

	# ----------------------------------------------------------------------
	# @return tagrange of currently editing item, trimming last tab and arrow
	# ----------------------------------------------------------------------
	def edit_range(self, index):
		try:
			f,t = self.tag_range(self.editTag, index)
		except ValueError:
			return ()
		# Adjust text range
		if self.editType == _EDIT_LIST:
			t += "-3c"	# remove last space-downarrow-tab
		else:
			t += "-1c"	# remove last tab
		return f,t

	# ----------------------------------------------------------------------
	# @return text under tag at index
	# ----------------------------------------------------------------------
	def edit_get(self, index):
		tagrange = self.edit_range(index)
		if tagrange:
			return self.get(*tagrange)
		else:
			return ""

	# ----------------------------------------------------------------------
	# set/update input
	# ----------------------------------------------------------------------
	def setInput(self, inp=None):
		self.active = None
		self.edit   = None
		if inp is None:
			self.input = self.project.input
		else:
			self.input = inp
		for layout in self.layout.values():
			layout.setInput(self.input)

	# ----------------------------------------------------------------------
	# Manipulate cache of layouts
	# ----------------------------------------------------------------------
	def layoutCache(self, flag):
		for layout in self.layout.values():
			layout.cache(flag)

	# ----------------------------------------------------------------------
	# Configuration
	# ----------------------------------------------------------------------
	# There is a bug/conflict with dnd motion bind!
	# ----------------------------------------------------------------------
	def motionBind(self):
		self._motionId   = None
		self._motionTime = 0.0
		self.bind('<Motion>',	 self.motion)
		self.bind('<B1-Motion>', self.b1motion)

	# ----------------------------------------------------------------------
	def tagConfigure(self):
		self.tag_config(_COMMENT_TAG,	foreground=_COMMENT_COLOR, font=_COMMENT_FONT,
						borderwidth=2)
		self.tag_config(_HIDDEN_TAG,	foreground=_HIDDEN_COLOR,  font=_HIDDEN_FONT,
						background=_HIDDEN_BGCOLOR, justify='center')
		self.tag_config(_LABEL_TAG,	foreground=_LABEL_COLOR,   font=_LABEL_FONT)
		self.tag_config(_VALUE_TAG,	foreground=_VALUE_COLOR,   font=_VALUE_FONT)
		self.tag_config(_NAME_TAG,	foreground=_NAME_COLOR,    font=_VALUE_FONT)
		if DEBUG:
			self.tag_config(_VALUE_TAG, background="#F0FFF0")
			self.tag_config(_NAME_TAG,  background="#F0FFF0")
		self.tag_config(_HR_TAG,	background="#D0D0D0",      font=("Symbol",-1))
		self.tag_config(_TAG_TAG,	foreground=_TAG_COLOR,     font=_TAG_FONT)
		self.tag_config(_CHILD_TAG,	foreground=_CHILD_COLOR,   font=_CHILD_FONT)
		self.tag_config(_MODE_TAG,	foreground=_MODE_COLOR,    font=_TAG_FONT)
		self.tag_config(_DEFINE_TAG,	foreground=_DEFINE_COLOR,  font=_TAG_FONT)
		if DEBUG:
			self.tag_config(_ID_TAG, foreground="Cyan", background="Gray")
		else:
			self.tag_config(_ID_TAG, elide=True)
		self.tag_config(_ERROR_TAG,	foreground=_ERROR_COLOR)
		self.tag_config(_DISABLE_TAG,	foreground=_DISABLE_COLOR)
		self.tag_config(_SELECT_TAG,	background=_ACTIVE_COLOR)
		self.tag_config(_ACTIVE_TAG,	borderwidth=_ACTIVE_BORDER,
						relief=tk.GROOVE)
		for i in range(10):
			self.tag_config(_INDENT_TAG%(i), lmargin1=i*_INDENT)
		self.tag_raise(tk.SEL)

	# ----------------------------------------------------------------------
	# Get/Set
	# ----------------------------------------------------------------------
	@property
	def project(self):
		return self.flair.project

	# ----------------------------------------------------------------------
	# Generic operations
	# ----------------------------------------------------------------------
	def enable(self):
		Global.dump("*-* InputEditor.enable", level=2)
		#traceback.print_stack()
		self.config(state=tk.NORMAL)

	# ----------------------------------------------------------------------
	def disable(self):
		Global.dump("*-* InputEditor.disable", level=2)
		#traceback.print_stack()
		self.config(state=tk.DISABLED)

	# ----------------------------------------------------------------------
	def ignore(self, event=None):
		pass

	# ----------------------------------------------------------------------
	def addUndo(self, undoinfo, msg=None):
		self.flair.addUndo(undoinfo, msg)

	# ----------------------------------------------------------------------
	# unpost list if posted during horizontal moves
	# ----------------------------------------------------------------------
	def xview(self, *args):
		if self.popupList.isPosted():
			self.unpostList()
		super().xview(*args)

	# ----------------------------------------------------------------------
	def setTabs(self):
		tabs = ()
		for t in self._tabs:
			x  = str(t)
			x1 = str(t+1)
			tabs += (x,tk.RIGHT, x1,tk.LEFT)
		dt = self._tabs[-1] - self._tabs[-2]

		while len(tabs)<20:
			t += dt
			x  = str(t)
			x1 = str(t+1)
			tabs += (x,tk.RIGHT, x1,tk.LEFT)
		self["tabs"] = tabs
		self.tag_config(_TEXT_TAG, lmargin1=self._tabs[0]+4)
		#self["spacing1"] = 1		# FIXME
		#self["spacing2"] = 1
		#self["spacing3"] = 1
		if self.popupList.isPosted():
			tagrange = self.tag_range(self.editTag, tk.INSERT)
			if not tagrange: return
			bb0 = self.bbox(tagrange[0])
			bb1 = self.bbox(tagrange[1])
			if bb1:
				w = max(bb1[0] - bb0[0], 100)
			else:
				w = self.winfo_width()-bb0[0]
			self.popupList.place(width=w)

	# ----------------------------------------------------------------------
	# @return index of active in cardlist
	# ----------------------------------------------------------------------
	def activePos(self):
		try:
			return self.cardlist.index(self.active)
		except ValueError:
			return -1

	# ----------------------------------------------------------------------
	# FIXME experimental
	# ----------------------------------------------------------------------
	def _printPosition(self, from_=0, to_=-1):
		if to_<0: to_ = len(self.cardlist)
		for i,cw in enumerate(self.cardlist[from_:to_]):
			if cw.card:
				tag = cw.card.tag
			else:
				tag = "Hidden"
			Global.dump("%3d %-10s   y= %3d .. %3d  (dy=%d)" %
				(i+from_, tag, cw.y, cw.yend, (cw.yend-cw.y)))

	# ----------------------------------------------------------------------
	# Check position of all CardWidget in cardlist
	# ----------------------------------------------------------------------
	def checkPosition(self, _raise_exception_=False):
		for i,cw in enumerate(self.cardlist):
			if not cw.checkPosition():
				Global.dump(f"{log.ERROR}Out of position {i}")
				Global.dump(cw.card)
				Global.dump("*** Repositioning ***")
				traceback.print_stack()
				if _raise_exception_:
					raise Exception("checkPosition-reposition infinite loop")
				self.reposition(i)
				return False
		return True

	# ----------------------------------------------------------------------
	# Reposition all CardWidgets starting from start
	# ----------------------------------------------------------------------
	def reposition(self, start=0, stop=-1):
		#Global.dump("*-* InputEditor.reposition", start, stop)
		if start<=0:
			start = 0
			y = 1
		else:
			y = self.cardlist[start-1].yend

		if stop<0:
			stop = len(self.cardlist)

		for cw in self.cardlist[start:stop]:
			dy   = cw.dy()
			cw.y = y
			y   += dy
			cw.yend = y

	# ----------------------------------------------------------------------
	# Reposition all cardwidgets
	# ----------------------------------------------------------------------
	def repositionEvent(self, event):
		# FIXME send also data with event range!
		# look the tkExtra.bindData
		self.reposition(self.activePos())

	# ----------------------------------------------------------------------
	# Find CardWidget from index
	# ----------------------------------------------------------------------
	def findCardWidget(self, index):
		if _ID_TAG in self.tag_names(index):
			# extreme case with index == cw.index()
			thisid = self.tag_nextrange(_ID_TAG, index, tk.END)
		else:
			thisid = self.tag_prevrange(_ID_TAG, index, "1.0")
		if not thisid: return None
		try:
			idx = int(self.get(*thisid))
		except:
			if Global.developer: raise
			return None
		cw = ctypes.cast(idx, ctypes.py_object).value
		return cw

	# ----------------------------------------------------------------------
	# VERTICAL MOTION
	# ----------------------------------------------------------------------
	def yview(self, *args):
		self.unpostList()
		return super().yview(*args)

	# ----------------------------------------------------------------------
	# middle mouse button released
	# ----------------------------------------------------------------------
	def scanRelease(self, event):
		if not self.scandrag or \
		   (abs(self.scandrag[0]-event.x)<=2 and abs(self.scandrag[1]-event.y)<=2):
			if self.edit is not None:
				try:
					self.pasteText(self.selection_get(
							selection='PRIMARY'))
				except tk.TclError:
					pass
		self.scandrag = None
		self.config(cursor="")
		return "break"

	# ----------------------------------------------------------------------
	def scanDrag(self, event):
		if self.scandrag:
			self.scan_dragto(event.x, event.y)
		else:
			self.config(cursor="hand2")
			self.scan_mark(event.x, event.y)
			self.scandrag = (event.x, event.y)
		return "break"

	# ----------------------------------------------------------------------
	# Focus
	# ----------------------------------------------------------------------
	def focusIn(self):
		self.page._focusIn()
		self.focus_set()

	# ----------------------------------------------------------------------
	def _focusOut(self, event=None):
		try:
			if self.focus_get() is None:
				return
		except: return
		self.commit()

	# ----------------------------------------------------------------------
	# EVENT handlers
	# ----------------------------------------------------------------------
	# Keyboard event handler
	# ----------------------------------------------------------------------
	def handleKey(self, event):
		Global.dump("*-* InputPage.handleKey", event.keysym, event.char,
				"insert=",self.index(tk.INSERT))
		if event.state & STATE.ALT: return	# Ignore ALT key
		self.hideBalloon()

		self.delay = tkExtra.Balloon.delay	# reset delay when key pressed
		shift = event.state & STATE.SHIFT

		# keep track of key buffer
		prevkey = self._key
		self._key = ""

		# --------------------------------------------------------------
		# Action keys while not editing any field
		# --------------------------------------------------------------
		if self.edit is None:
			cardAnchor = self.cardAnchor	# remember before operation
			activePos  = self.activePos()	# remember past active position

			# operate on editor
			if event.keysym in ("Prior", "u", "U"):
				self.yview(tk.SCROLL, -1, tk.PAGES)
				top = self.index("@0,0")
				cw = self.findCardWidget(top)
				if cw is None and self.cardlist:
					cw = self.cardlist[0]
				if self.active is cw: return
				self.activate(cw)
			elif event.keysym in ("Next", "d", "D"):
				self.yview(tk.SCROLL,  1, tk.PAGES)
				bottom = self.index(f"@0,{self.winfo_height()}")
				cw = self.findCardWidget(bottom)
				if cw is None and self.cardlist: cw = self.cardlist[-1]
				if self.active is cw: return
				self.activate(cw)
			elif event.keysym == "Home":
				if self.cardlist:
					self.activate(0)
					self.active.see()
			elif event.keysym == "g":
				if prevkey == "g":
					if self.cardlist:
						self.activate(0)
						self.active.see()
				else:
					self._key = event.keysym
			elif event.keysym in ("End", "G"):
				if event.keysym == "G": shift = False
				if self.cardlist:
					self.activate(len(self.cardlist)-1)
					self.active.see(end=True)
			elif event.keysym in ("Up", "k", "K"):
				if not self.activatePrevious(): return "break"
			elif event.keysym in ("Down", "j", "J"):
				if not self.activateNext(): return "break"
			elif event.keysym in ("Left", "h"):
				self.xview(tk.SCROLL, -1, tk.UNITS)
				return "break"
			elif event.keysym == "H":
				# move cursor to top of screen (vi)
				if not self.cardlist: return
				shift = False
				cw = self.findCardWidget("@2,2")
				if self.active is cw: return
				self.activate(cw)
				self.active.see(end=True)
			elif event.keysym == "L":
				# move cursor to bottom of screen (vi)
				if not self.cardlist: return
				shift = False
				cw = self.findCardWidget(f"@2,{self.winfo_height()-2}")
				if self.active is cw: return
				self.activate(cw)
				self.active.see(end=True)
			elif event.keysym == "M":
				# move cursor to middle of screen (vi)
				if not self.cardlist: return
				shift = False
				cw = self.findCardWidget(f"@2,{self.winfo_height()//2}")
				if self.active is cw: return
				self.activate(cw)
				self.active.see(end=True)
			elif event.keysym in ("Right", "l"):
				self.xview(tk.SCROLL, 1, tk.UNITS)
				return "break"
			elif event.keysym in ("Return", "KP_Enter", "F2", "i"):
				if self.active:
					self.active.editNext()
				return "break"
			elif event.keysym == "Escape":
				if self.dnd_item is not None:
					self.dnd_end(None,None)
				else:
					self.cardSelectClear()
					if self.active is not None:
						self.cardSelect(self.active)
				return "break"
			elif event.keysym in ("BackSpace", "Delete", "KP_Delete") or event.char=="x":
				self.delSelectedCards()
				return "break"
			elif event.keysym == "F12" and Global.developer:
				if event.state & STATE.CONTROL:
					self.checkEditor()
				else:
					self.dumpActive()
			#elif event.keysym == "z":
			#	if prevkey == "z":
			#		# center active line
			#		print("CENTER")
			#	else:
			#		self._key = "z"

			elif event.char == "a":
				self.page.focusAdd()
				return "break"
			elif event.char == "c":
				self.page.toggleComment()
				return "break"
			elif event.char == "t":
				self.page.focusTree()
				return "break"
			elif event.char in ("/","?"):
				self.page.focusSearch()
				return "break"
			elif event.char == "#":
				self.page.focusGoto()
				return "break"
			else:
				# All other keys pass to the system
				return

			# Common action on every move
			if shift:
				self.cardUnSelect(cardAnchor, activePos)
				self.cardSelect(cardAnchor,   self.cardAnchor)
				self.cardAnchor = cardAnchor	# restore anchor
			else:
				# self.cardAnchor should be the activePos now
				if self.cardAnchor not in self.selection:
					self.cardSelectClear()
				self.cardSelect(self.cardAnchor)
			return "break"

		# --------------------------------------------------------------
		# Action keys while editing a field
		# --------------------------------------------------------------
		# Get tag range we will needed it
		try:
			startIndex, endIndex = self.edit_range(tk.INSERT)
			Global.dump(f"+++ insert={self.index(tk.INSERT)} "
				f"tag={self.editTag} range={startIndex}..{endIndex}")
		except ValueError as ex:
			Global.dump(f"+++ insert={self.index(tk.INSERT)} tag={self.editTag}")
			Global.dump(f">>> Exception: {ex}")
			self.unpostList()
			self.enable()
			self.commit()
			self.disable()
			self.active.editItem()
			return "break"

		# Save undo info in case of too long wait
		if self.lastKey[0]==self.edit and self.active and \
		   not self.popupList.isPosted() and \
		   event.time-self.lastKey[1] >= _KEY_TIMEOUT:
			self.active.commit(True)

		# editing
		if event.state & STATE.CONTROL:
			if len(event.keysym)>1:
				# Ignore all special keys with control
				# especially BackSpace and Delete
				return "break"
			return	# send all control-key to system
		elif event.keysym == "Escape":
			if self.popupList.isPosted():
				self.unpostList()
				self.popupList.unselect()
				#if self.editType == _EDIT_MULTILINE:
				return "break"

			if self.editType == _EDIT_FILE:
				return "break"

			try:	# Try if there is a selection
				self.index(tk.SEL_FIRST)
				self.tag_remove(tk.SEL, "1.0", tk.END)
			except tk.TclError:
				# No selection
				self.commit()
				self.active.editItem()

		# navigation
		elif event.keysym == "Home":
			if self.editType == _EDIT_MULTILINE:
				pos = self.search("\n", tk.INSERT, startIndex, backwards=True)
				if not pos:
					pos = startIndex
				else:
					pos += "+1c"
			else:
				pos = startIndex

			if shift:
				self.setAnchor(tk.INSERT)
				self.tag_add(tk.SEL, pos, tk.INSERT)
			self.mark_set(tk.INSERT, pos)

		elif event.keysym == "End":
			if self.editType == _EDIT_MULTILINE:
				pos = self.search("\n", tk.INSERT, endIndex)
				if not pos: pos = endIndex
			else:
				pos = endIndex
			if shift:
				self.setAnchor(tk.INSERT)
				self.tag_add(tk.SEL, tk.INSERT, pos)
			self.mark_set(tk.INSERT, pos)

		elif event.keysym == "Left":
			if self.compare(tk.INSERT, "<=", startIndex):
				# Try if there is a selection
				#if not event.state & STATE.SHIFT:
				#	self.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)
				return "break" # Unselect before moving to next
				#if not shift and self.edit != -1:
				#	self.unpostList()
				#	self.commit()
				#	self.active.editPrev()
			else:
				return	# < Use the default handler

		elif event.keysym == "Right":
			if self.compare(tk.INSERT, ">=", endIndex):	# ignore last tab
				# Try if there is a selection
				#if not event.state & STATE.SHIFT:
				#	self.tag_remove(tk.SEL, tk.SEL_FIRST, tk.SEL_LAST)
				return "break" # Unselect before moving to next
				#if not shift and self.edit != -1:
				#	self.unpostList()
				#	self.commit()
				#	self.active.editNext()
			else:
				return	# < Use the default handler

		elif event.keysym == "Up":
			if self.popupList.isPosted():
				self.popupList.up()
			else:
				index = self.index(tk.INSERT + "- 1 displaylines")
				if shift and self.compare(index, "<", startIndex):
					# constraint in the edit range when shift is pressed
					index = startIndex
				self.activateIndex(index, shift, True)
				if not event.state & STATE.SHIFT:
					self.tag_remove(tk.SEL, "1.0", tk.END)
			return "break"

		elif event.keysym == "Down":
			if self.popupList.isPosted():
				self.popupList.down()
			else:
				index = self.index(tk.INSERT + "+ 1 displaylines")
				if shift and self.compare(index, ">", endIndex):
					# constraint in the edit range when shift is pressed
					index = endIndex
				self.activateIndex(index, shift, True)
				if not event.state & STATE.SHIFT:
					self.tag_remove(tk.SEL, "1.0", tk.END)
			return "break"

		if event.keysym == "Prior":
			if self.popupList.isPosted():
				self.popupList.prior()
			else:
				# similar code to text.tcl, ::tk::TextScrollPages
				bbox = self.bbox(tk.INSERT)
				self.yview(tk.SCROLL, -1, tk.PAGES)
				if bbox:
					index = f"@{bbox[0]},{bbox[1]}"
				else:
					index = f"@0,{self.winfo_height()/2}"
				self.activateIndex(index, shift, False)

		elif event.keysym == "Next":
			if self.popupList.isPosted():
				self.popupList.next()
			else:
				# similar code to text.tcl, ::tk::TextScrollPages
				bbox = self.bbox(tk.INSERT)
				self.yview(tk.SCROLL,  1, tk.PAGES)
				if bbox:
					index = f"@{bbox[0]},{bbox[1]}"
				else:
					index = f"@0,{self.winfo_height()/2}"
				self.activateIndex(index, shift, False)

		elif event.keysym in ("Return", "KP_Enter"):
			if self.popupList.isPosted():
				self.unpostList()
				if self._insertText is not None:
					self.active.insertText(self.popupList.selection()[0])
					self.commit()
					return "break"
			elif self.editType == _EDIT_LIST:
				self.postList()
				return "break"

			if self.editType != _EDIT_MULTILINE:
				self.commit()
				self.active.editNext()
			else:
				self.insert(tk.INSERT, "\n",
						self.tag_names(tk.INSERT))
				self.see(tk.INSERT)
				self.active.updateEnd()
				self.event_generate("<<Reposition>>")
				self.lastKey = (self.edit, event.time)
				self.page.inputModified()

		elif event.keysym in ("BackSpace", "Terminate_Server"):
			if self.isTextSelected():
				self.delete(tk.SEL_FIRST, tk.SEL_LAST)
			elif self.compare(tk.INSERT, "<=", startIndex):
				return "break"
			else:
				self.delete(tk.INSERT+"-1c")
			self.see(tk.INSERT)
			self.active.updateEnd()
			self.event_generate("<<Reposition>>")
			self.lastKey = (self.edit, event.time)
			self.page.inputModified()
			if self.popupList.isPosted():
				self.filterList("BackSpace")

		elif event.keysym == "Delete":
			if self.isTextSelected():
				self.delete(tk.SEL_FIRST, tk.SEL_LAST)
			elif self.compare(tk.INSERT, ">=", endIndex):	# ignore last tab
				return "break"
			else:
				self.delete(tk.INSERT)
			self.see(tk.INSERT)
			self.active.updateEnd()
			self.event_generate("<<Reposition>>")
			self.lastKey = (self.edit, event.time)
			self.page.inputModified()
			if self.popupList.isPosted():
				self.filterList("Delete")

		# anything else editing
		elif event.char >= " ":
			key = event.char
			try:
				# Protect against accidental selection
				if self.compare(tk.SEL_FIRST, "<", startIndex):
					selFirst = startIndex
				else:
					selFirst = tk.SEL_FIRST
				if self.compare(tk.SEL_LAST, ">", endIndex):
					selLast = endIndex
				else:
					selLast = tk.SEL_LAST
				self.delete(selFirst, selLast)
				key = None
			except tk.TclError:
				pass
			self.insert(tk.INSERT, event.char,
					self.tag_names(tk.INSERT))
			self.see(tk.INSERT)

			# only for bodies
			if self.popupList.isPosted():
				if self._insertText is not None and \
				   event.char in (" ","+","-","|","(",")","#"):
					self.unpostList()
				else:
					self.filterList(key)
			elif self.editType == _EDIT_LIST:
				self.postList()

			self.lastKey = (self.edit, event.time)
			self.page.inputModified()
			if _BODY_INSERT and \
			   self.active.card.type()==Card.REGION and \
			   self.editType == _EDIT_MULTILINE and \
			   self.edit >= 0:
				if event.char in ("+","-"):
					self.insertBody()
				elif event.char == "#":
					self.insertRegion()

		elif event.keysym == "F1":
			self.help()

		# Default behaviour stop here!!!
		return "break"

	# ----------------------------------------------------------------------
#	def keyPress(self, event):
#		Global.dump("*-* InputPage.keyPress keysym=",event.keysym)
#		return "break"
#
#	# ----------------------------------------------------------------------
#	def keyRelease(self, event):
#		Global.dump("*-* InputPage.keyRelease keysym=",event.keysym)
#		return "break"

	# ----------------------------------------------------------------------
	# Popup balloon help, bound to motion without any button pressed
	# ----------------------------------------------------------------------
	def motion(self, event):
		tkExtra.Balloon.hide()
		if self._motionId is not None:
			if event.time - self._motionTime < self.delay:
				self.after_cancel(self._motionId)
				self._motionId = None
		if self._motionId is None:
			self._motionX = event.x
			self._motionY = event.y
			self._motionId = self.after(self.delay, self.showBalloon)
		self._motionTime = event.time

		if DEBUG:
			# FIXME only for debugging, to be removed
			self.page.setStatus(
				f"Index={self.index(tk.CURRENT)}  "
				f"Tags={self.tag_names(tk.CURRENT)}")
		return "break"

	# ----------------------------------------------------------------------
	# Mouse button1
	# Event order
	#   button1
	#   release1
	#   double1     (in case of double click the button1 is not called)
	#   release1
	#   triple1
	#   release1
	# ----------------------------------------------------------------------
	def button1(self, event):
		Global.dump("*-* InputPage.button1 tk.CURRENT=",self.index(tk.CURRENT))
		current = self.index(tk.CURRENT)	# remember index
		self.focusIn()
		self.tag_remove(tk.SEL, "1.0", tk.END)

		self.dnd_time = None	# Mark as dnd for the moment

		# Find CardWidget
		cw = self.findCardWidget(current)
		self._anchor  = None
		self._selword = False
		if cw is None:
			self.commit()
			self.unpostList()
			self.cardSelectClear()
			self.activate()
			return "break"

		# Item editing
		cid = self.cardlist.index(cw)

		# Commit changes if needed
		cardAnchor = self.cardAnchor # remember cardAnchor
		activePos  = self.activePos()
		try:
			# Check if we clicked on an image to be used as a card toggle
			self.image_cget(current,"name")
			onimage = True
		except tk.TclError:
			onimage = False
		if self.active is not cw:
			self.unpostList()
			self.commit()		# commit might change the tk.CURRENT
			cw.editItem()
			self.activate(cid)
			# Control hit or first column
			if event.state & STATE.CONTROL or onimage:
				if event.state & STATE.SHIFT:
					self.cardSelect(cardAnchor, activePos, toggle=True)
					self.cardSelect(cardAnchor, self.cardAnchor, toggle=True)
					self.cardAnchor = cardAnchor	# restore anchor
				else:
					self.cardSelectToggle(cid)
					return
			elif event.state & STATE.SHIFT:
				self.cardUnSelect(cardAnchor, activePos)
				self.cardSelect(cardAnchor, self.cardAnchor)
				self.cardAnchor = cardAnchor	# restore anchor
			elif cid not in self.selection:
				self.cardSelectClear()
				self.cardSelect(cid)
		elif event.state & STATE.CONTROL or onimage:
			self.unpostList()
			self.commit()		# commit might change the tk.CURRENT
			cw.editItem()
			self.cardSelectToggle(cid)
			return

		item = cw.item(current)
		if item is None:		# No item is selected
			self.commit()
			cw.editItem()
			# Potential dnd
			self.dnd_time = event.time

		elif item != self.edit:		# Edit new item
			self.unpostList()
			self.commit()		# commit might change the tk.CURRENT
			#self.cardSelectClear()
			#self.cardSelect(cid)
			self.mark_set(tk.INSERT, current)
			cw.editItem(item, select=False)

			# Constraint click into the tag-range
			ft = self.edit_range(tk.INSERT)
			if ft:
				if self.compare(tk.INSERT, ">", ft[1]):
					self.mark_set(tk.INSERT, ft[1])
				self._anchor = self.index(tk.INSERT)

		else:	# Click into currently editing item
			# Constraint end of range without tab or arrow down
			ft = self.edit_range(current)
			if ft:
				if self.compare(current, ">", ft[1]):
					self.mark_set(tk.INSERT, ft[1])
					if self.popupList.isPosted():
						self.unpostList()
						self.popupList.unselect()
					elif self.editType == _EDIT_LIST:
						self.postList()
				else:
					if self.editType == _EDIT_LIST:
						self.postList()
					self.mark_set(tk.INSERT, current)
				self._anchor = self.index(tk.INSERT)
		return "break"			# <<< break same as above

	# ----------------------------------------------------------------------
	# Button1 release
	# ----------------------------------------------------------------------
	def release1(self, event):
		Global.dump("*-* release1", level=2)
		self.cancelRepeat()
		return "break"

	# ----------------------------------------------------------------------
	# Double click
	# ----------------------------------------------------------------------
	def double1(self, event):
		Global.dump("*-* double1", level=2)
		if self.edit is not None:
			self._selword = True
			index = self.index(f"@{event.x},{event.y}")
			self._anchor = self.index(f"{index} wordstart")

			# Constraint end of range without tab or arrow down
			try:
				f,t = self.edit_range(index)
			except ValueError:
				return "break"
			index += " wordend"
			if self.compare(index, ">", t):
				self.mark_set(tk.INSERT, t)
			else:
				self.mark_set(tk.INSERT, index)
			self.setAnchor()
			self.tag_remove(tk.SEL, "1.0", tk.END)
			self.tag_add(tk.SEL, self._anchor, tk.INSERT)
			return "break"

		# Out in the blue...
		elif self.active is None:
			self.cardSelectClear()

		# On hidden cards
		elif self.active.card is None:
			self.expandCards()

		# On preprocessor block #if...#endif

		# XXX if double click on preprocessor #if..#else..#endif
		# select all block
		elif self.active.card.tag in Input._PREPRO_BLOCK:
			cid = self.activePos()
			indent = self.active._indent

			# Scan for the starting card with the same ident level
			fromid = cid
			while fromid>0:
				cw = self.cardlist[fromid]
				if cw.card is not None and \
				   cw.card.tag in ("#if", "#ifdef", "#ifndef") and \
				   cw._indent == indent: break
				fromid -= 1

			# Scan for the #endif card with the same indent level
			toid = cid
			while toid<len(self.cardlist):
				cw = self.cardlist[toid]
				if cw.card is not None and \
				   cw.card.tag == "#endif" and \
				   cw._indent == indent: break
				toid += 1

			self.cardSelect(fromid, toid, cid in self.selection)

		# if on #include..#endinclude
		elif self.active.card.tag in ("#include", "#endinclude"):
			self.cardSelectTagRange("#include", "#endinclude",
						self.activePos() in self.selection)

		# start..end transform
		elif self.active.card.tag in ("$start_transform", "$end_transform"):
			self.cardSelectTagRange("$start_transform", "$end_transform",
						self.activePos() in self.selection)

		# start..end translat
		elif self.active.card.tag in ("$start_translat", "$end_translat"):
			self.cardSelectTagRange("$start_translat", "$end_translat",
						self.activePos() in self.selection)

		# start..end expansion
		elif self.active.card.tag in ("$start_expansion", "$end_expansion"):
			self.cardSelectTagRange("$start_expansion", "$end_expansion",
						self.activePos() in self.selection)

		# GEOBEGIN..GEOEND
		elif self.active.card.tag in ("GEOBEGIN", "GEOEND"):
			self.cardSelectTagRange("GEOBEGIN", "GEOEND",
						self.activePos() in self.selection)

		# Select range of similar cards
		else:
			if self.active is None: return "break"
			if isinstance(self.active, HiddenCardWidget):
				pass
			else:
				cid = self.activePos()
				tag = self.cardlist[cid].card.tag
				begin = cid
				while begin>=0:
					card = self.cardlist[begin].card
					if card and card.tag == tag:
						begin -= 1
					else:
						break
				end = cid
				while end < len(self.cardlist):
					card = self.cardlist[end].card
					if card and card.tag == tag:
						end += 1
					else:
						break
				# Propagate the selection from button1
				self.cardSelect(begin+1, end-1, cid in self.selection)
		return "break"

	# ----------------------------------------------------------------------
	# Triple click
	# ----------------------------------------------------------------------
	def triple1(self, event):
		Global.dump("*-* triple1", level=2)
		self.selectAll()
		return "break"

	# ----------------------------------------------------------------------
	# Select using the mouse
	# ----------------------------------------------------------------------
	def b1motion(self, event=None):
		cw = self.findCardWidget("@%d,%d"%(event.x,event.y))
		if self is not self.focus_get() or cw is None:
			# if we are not the current window don't do anything
			# e.g. click and drag an item that opens a new dialog
			# (file dialog)
			pass
		elif self.edit is None:
			if cw is None or cw is self.active: return "break"
			# Moved out of the widget
			# Should it move out of the tag or the line?
			if self.active and \
			   self.dnd_time is not None and \
			   event.time - self.dnd_time>Global.DND_TIME:
				self.dnd_time = None
				event.num = 1	# Fake B1 press
				if MyDndHandler.start(self, event):
					self.unpostList()
					# Remove conflict with Drag Motion bind
					# we will fix it at the end of the dnd event
					self.unbind("<B1-Motion>")
					self.dnd_label = None

					# find cards to drag
					ranges = self.selection2Ranges()
					S = 0
					for f,t in ranges:
						S += self.cardlist[t].toPos() \
						   - self.cardlist[f].fromPos() + 1
					if S==1: # only one card
						self.dnd_label = self.active.card.tag
					else:
						cw = self.cardlist[ranges[0][0]]
						fromTag = cw.card.tag if cw.card is not None \
									else cw.fromCard.tag
						cw = self.cardlist[ranges[-1][1]]
						toTag   = cw.card.tag if cw.card is not None \
									else cw.toCard.tag
						self.dnd_label = f"{fromTag} .. {toTag} [{S} cards]"
					self.cut()
					self.selectClear()
				#self.tk.call("tk::CancelRepeat")
			else:
				self.dnd_time = None
				cid  = self.cardlist.index(cw)
				cardAnchor = self.cardAnchor	# remember before operation
				activePos  = self.activePos()	# remember past active position
				self.activate(cid)
				self.cardUnSelect(cardAnchor, activePos)
				self.cardSelect(cardAnchor, cid)
				self.cardAnchor = cardAnchor

		# We are editing an item, ==> Select text
		elif cw is self.active:
			index = self.index(f"@{event.x},{event.y}")
			item = cw.item(index)
			if item == self.edit and self._anchor != index:
				try:
					if self.compare(self._anchor,"<",index):
						f = self._anchor
						t = index
						fi = False
					else:
						f = index
						t = self._anchor
						fi = True

					if self._selword:
						f += " wordstart"
						t += " wordend"

					tagrange = self.edit_range(self._anchor)
					try:
						if self.compare(t, ">", tagrange[1]):
							t = tagrange[1]
					except IndexError:
						pass
					self.setAnchor()
					self.tag_remove(tk.SEL,"1.0",tk.END)
					self.tag_add(tk.SEL, f, t)
					self.mark_set(tk.INSERT, f if fi else t)
				except tk.TclError:
					pass
		else:
			# Select up to the maximum of the tagrange
			index    = self.index(f"@{event.x},{event.y}")
			tagrange = self.edit_range(self._anchor)
			if tagrange:
				if self.compare(index,"<", tagrange[0]):
					self.setAnchor()
					self.tag_remove(tk.SEL,"1.0",tk.END)
					self.tag_add(tk.SEL, tagrange[0], self._anchor)
					self.mark_set(tk.INSERT, tagrange[0])
				elif self.compare(index, ">", tagrange[1]):
					self.setAnchor()
					self.tag_remove(tk.SEL,"1.0",tk.END)
					self.tag_add(tk.SEL, self._anchor, tagrange[1])
					self.mark_set(tk.INSERT, tagrange[1])
		return "break"

	# ----------------------------------------------------------------------
	def cancelRepeat(self, event=None):
		if self._autoId:
			self.after_cancel(self._autoId)
		self._autoId = None
		return "break"

	# ----------------------------------------------------------------------
	# Repeat Button-1 motion when the cursor is out of the screen and the
	# button is pressed
	# Adapted from text.tcl TextAutoScan
	# ----------------------------------------------------------------------
	def autoRepeat(self, event=None):
		self._autoId = None
		if not self.winfo_exists(): return "break"
		if event is None:
			x = self.winfo_pointerx() - self.winfo_rootx()
			y = self.winfo_pointery() - self.winfo_rooty()
		else:
			x = event.x
			y = event.y
		if y > self.winfo_height():
			self.yview(tk.SCROLL, 1+y-self.winfo_height(), "pixels")
		elif y < 0:
			self.yview(tk.SCROLL, -1+y, "pixels")
		elif x > self.winfo_width():
			self.xview(tk.SCROLL,  2, tk.UNITS)
		elif x < 0:
			self.xview(tk.SCROLL, -2, tk.UNITS)
		else:
			return "break"
		# TextSelectTo $w $Priv(x) $Priv(y)
		self._autoId = self.after(50, self.autoRepeat)
		return "break"

	# ----------------------------------------------------------------------
	# DRAG & DROP
	# ----------------------------------------------------------------------
	# Accept dnd messages, i.e. we're a legit drop target, and we do
	# implement d&d functions.
	# ----------------------------------------------------------------------
	def dnd_accept(self, source, event):
		return self

	# ----------------------------------------------------------------------
	# Create drag object
	# ----------------------------------------------------------------------
	def dnd_enter(self, source, event):
		#self.focusIn()
		self.focus_set()
		self.dnd_item = tk.Label(self,	text=source.dnd_label,
						font=_TAG_FONT,
						foreground=_TAG_COLOR,
						background=_ACTIVE_COLOR,
						border=_ACTIVE_BORDER,
						relief=tk.RAISED)
		self.dnd_motion(source, event)

	# ----------------------------------------------------------------------
	def dnd_leave(self, source, event):
		self.dnd_item.place_forget()
		self.dnd_item = None

	# ----------------------------------------------------------------------
	# Move drag icon
	# ----------------------------------------------------------------------
	def dnd_motion(self, source, event):
		x = event.x_root - self.winfo_rootx()
		y = event.y_root - self.winfo_rooty()
		self.dnd_item.place(x=x, y=y)

	# ----------------------------------------------------------------------
	# Notification that dnd processing has been ended. It DOES NOT imply
	# that we've been dropped somewhere useful, we could have just been
	# dropped into deep space and nothing happened to any data structures,
	# or it could have been just a plain mouse-click w/o any dragging.
	# ----------------------------------------------------------------------
	def dnd_end(self, target, event):
		if self.dnd_item:
			self.dnd_item.place_forget()
			self.dnd_item = None

		if target is None and self.dnd_label is not None:
			self.page.undo()	# Restore cards
			#self.activate(self.dnd_active.card)
		self.motionBind()	# Restore motion binding
		self.dnd_label = None

	# ----------------------------------------------------------------------
	# Object has been dropped here
	# ----------------------------------------------------------------------
	def dnd_commit(self, source, event):
		if self.dnd_item:
			self.dnd_item.place_forget()
			self.dnd_item = None
		if source.dnd_label is None: return

		index = self.index("@%d,%d"%(event.x,event.y))
		line  = int(index.split(".")[0])
		cw = self.findCardWidget(index)
		if cw is None and self.cardlist:
			cw = self.cardlist[-1]

		up = False		# insert above or below
		if cw is not None:
			if line == cw.y:
				x,y,w,h,b = self.dlineinfo(index)
				up = event.y < y+h/2
		self.activate(cw)
		#if source is not self and source.dnd_active is not None:
		#	source.activate(source.dnd_active)
		self.paste()
		self.clipboard_clear()
		if up: self.moveUp()

	# ----------------------------------------------------------------------
	# ADD/DEL CARDs
	# ----------------------------------------------------------------------
	# Add a card (Normal or Hidden) or range of cards [card..to_card]
	# to InputEditor
	# @param card     card to be added
	# @param to_card  range as hidden cards
	# @param pos      position where to add the card
	# @param first    when multiple cards are inserted (in the middle)
	#                 mark what is is to be moved when first is True
	# @param last     when multiple cards are inserted (in the middle)
	#                 perform the move when last is True
	# @return card widget
	# ----------------------------------------------------------------------
	def addCard(self, pos, card, to_card=None):
		if not self.cardlist:
			pos  = 0
			ypos = 1		# Text is line1 based
		elif pos is None or pos>=len(self.cardlist):
			# Add after the last one
			pos  = len(self.cardlist)
			ypos = self.cardlist[-1].yend
		elif pos<=0:
			pos  = 0
			ypos = 1		# Text is line1 based
		else:
			ypos = self.cardlist[pos-1].yend

		# Create new card widget
		if isinstance(card, int):
			card = self.input.cardlist[card]

		if to_card is not None:
			if isinstance(to_card, int):
				to_card = self.input.cardlist[to_card]
			cw = HiddenCardWidget(self, card, to_card, ypos)
		else:
			cw = CardWidget(self, card, ypos)

		# Append card to list
		if pos>=len(self.cardlist):
			self.cardlist.append(cw)
		else:
			self.cardlist.insert(pos, cw)

		# Insert children
		if to_card is None:
			ypos = cw.yend
			for child in card._children:
				if child.parent is not None:
					cwc = ChildCardWidget(self, child, ypos)
				else:
					cwc = CardWidget(self, child, ypos)
				ypos = cwc.yend
				# Append card to list
				pos += 1
				if pos>=len(self.cardlist):
					self.cardlist.append(cwc)
				else:
					self.cardlist.insert(pos, cwc)
		return cw

	# ----------------------------------------------------------------------
	# Add card in correct position, expand HiddenCardWidget if needed
	# ----------------------------------------------------------------------
	def addCardSorted(self, card):
		cardpos = card.pos()
		for i, cw in enumerate(self.cardlist):
			if cw.toPos() < cardpos: continue
			if cw.card is not None: return None	# It should never come here
			# From here only HiddenCardWidget should be found
			if cardpos == cw.fromPos():
				# Same starting position
				if cw.fromPos() == cw.toPos():	# Only ONE card
					# Is hidden with only one card
					# replace the hidden with a normal cardwidget
					self.enable()
					self.delCards(i)
					cw = self.addCard(i, card)
					self.reposition(i+1)
					self.disable()
					assert self.checkPosition()
					return cw
				else:
					# Add a card at the start
					self.enable()
					newcw = self.addCard(i, card)
					self.reposition(i+1)
					# and update the widget
					cw.fromCard = self.input.cardlist[cardpos+1]
					cw.redraw()
					self.disable()
					assert self.checkPosition()
					return newcw
			elif cardpos == cw.toPos():
				# insert after the end of HiddenCardWidget
				cw.toCard = self.input.cardlist[cardpos-1]
				self.enable()
				cw.redraw()
				cw = self.addCard(i+1, card)
				self.disable()
				self.reposition(i+2)
				assert self.checkPosition()
				return cw
			else:	# In the middle of a HiddenCardWidget
				# by splitting into two
				tocard = cw.toCard
				cw.toCard = self.input.cardlist[cardpos-1]
				self.enable()
				cw.redraw()
				newcw = self.addCard(i+1, card)
				self.addCard(i+2, self.input.cardlist[cardpos+1], tocard)
				self.reposition(i+3)
				self.disable()
				assert self.checkPosition()
				return newcw

		# At the end of the file (it should never arrive here)
		self.enable()
		cw = self.addCard(None, card)
		self.disable()
		assert self.checkPosition()
		return cw

	# ----------------------------------------------------------------------
	# Delete a card(s) widgets from the editor box
	# Warning. it doesn't:
	#	- set the active
	#	- clean up the self.selection
	#	- reposition card widgets
	# ----------------------------------------------------------------------
	def delCardRange(self, from_, to_):
		cf = self.cardlist[from_]
		ct = self.cardlist[to_]
		self.delete(cf.index(), ct.indexEnd())
		del self.cardlist[from_:to_+1]

	# ----------------------------------------------------------------------
	# Delete card(s) widget range from editor
	# ----------------------------------------------------------------------
	def delCards(self, cards):
		if not self.cardlist: return
		if isinstance(cards,int):
			# delete one card
			cw = self.cardlist[cards]
			self.delete(cw.index(), cw.indexEnd())
			del self.cardlist[cards]
		else:
			for f,t in reversed(cards):
				self.delCardRange(f,t)

	# ----------------------------------------------------------------------
	# Delete all cards widgets
	# ----------------------------------------------------------------------
	def deleteAll(self):
		self.delete("1.0",tk.END)
		del self.cardlist[:]
		self.active = None
		self.selectionClear()

	# ----------------------------------------------------------------------
	# Delete selected cards
	# ----------------------------------------------------------------------
	#@tool.timeit
	def delSelectedCards(self):
		if self.active is None: return
		if not self.selection: return

		# remember active position
		actid = self.activePos()

		selection = self.selection2Ranges()

		# Prepare undo info
		undoinfo = self.flair.undolistRenumberOnly()
		for f,t in reversed(selection):
			if f<=actid<=t:
				actid = f
			if actid > t:
				actid -= t-f+1
			for c in range(self.cardlist[t].toPos(),
					self.cardlist[f].fromPos()-1, -1):
				try:
					undoinfo.append(self.flair.delCardUndo(c, False))
				except IndexError:
					# if the card range is modified by geoviewer but the
					# inputeditor is not updated
					pass
		self.flair.addUndolistRenumberOnly(undoinfo, "Delete card(s)")

		# Delete cardwidgets
		self.enable()
		self.delCards(selection)
		self.disable()

		self.selectionClear()
		self.reposition(selection[0][0])
		assert self.checkPosition()
		self.activate(actid)
		self.cardSelect(self.activePos())

		self.page.inputModified()
		self.page.refreshStatus()
		self.page.preprocess()	# FIXME only if #define/#undef is present
		self.page._refreshOtherFrames()
		self.input.clearCache(Card.BODY)
		self.input.checkNumbering()	# XXX

	# ----------------------------------------------------------------------
#	def _displayCardsAfter(self):
#		fromCard = self._fromCard
#		startTime = time.time()
#		for i,card in enumerate(self._cardlist):
#			pos = card.pos()
#			if fromCard and fromCard.pos() != pos:
#				toCard = self.input.cardlist[pos-1]
#				self.addCard(None, fromCard, toCard)
#			self.addCard(None, card)
#			if pos+1 < len(self.input.cardlist):
#				fromCard = self.input.cardlist[pos+1]
#			else:
#				fromCard = None
#			if time.time()-startTime > 0.5:
#				Global.dump("+++ idle",i)
#				self._fromCard = fromCard
#				del self._cardlist[:i]
#				self.after_idle(self._displayCardsAfter)
#				return
#		# Try to find which card to activate
#		for cw in self.cardlist:
#			if cw.fromPos() <= self._actpos <= cw.toPos():
#				self.activate(cw)
#				break
#		if self.active is None and self.cardlist:
#			self.activate(0)
#		self.page.refreshStatus()
#		if self.active is not None:
#			self.active.see()
#		self.tag_config(_COMMENT_TAG, foreground=_COMMENT_COLOR, font=_COMMENT_FONT)
#		self.tag_config(_TAG_TAG,     foreground=_TAG_COLOR,     font=_TAG_FONT)
#		self.tag_config(_LABEL_TAG,   foreground=_LABEL_COLOR,   font=_LABEL_FONT)
#		self.tag_config(_VALUE_TAG,   foreground=_VALUE_COLOR,   font=_VALUE_FONT)

	# ----------------------------------------------------------------------
	# CARD DISPLAY
	# ----------------------------------------------------------------------
	# Create the CardWidgets for the cardlist provided
	# ----------------------------------------------------------------------
	def displayCards(self, cardlist):
		#Global.dump("*-* InputEditor.displayCards:",[x.tag for x in cardlist])
		self.page.busy()
		# Remember active
		if self.active is not None:
			active = self.active.fromPos()
		else:
			active = 0

		# Remember vertical position and restore it!
		yview = self.yview()

		# Remember selections
		#selection = self.selection2Cards()

		# insert to editor
		startTime = self.resetInterrupt()
		self.enable()
		self.deleteAll()
		#self.page.setStatus("Loading....", update=True)
		self.layoutCache(True)		# Enable caching of lists
		if self.input.cardlist:
			fromCard = self.input.cardlist[0]
			# FIXME display icons on long displays
			if len(cardlist)>2000:
				self.showIcons = False
			else:
				self.showIcons = _SHOW_ICONS
			for i,card in enumerate(cardlist):
				pos = card.pos()
				if fromCard and fromCard.pos() != pos:
					toCard = self.input.cardlist[pos-1]
					self.addCard(None, fromCard, toCard)
				self.addCard(None, card)
				if pos+1 < len(self.input.cardlist):
					fromCard = self.input.cardlist[pos+1]
				else:
					fromCard = None
				if time.time()-startTime > _CARDS_TIMEOUT:
					break
			if fromCard:
				self.addCard(None, fromCard, self.input.cardlist[-1])
			del cardlist
		#self.page.resetPercent()

		# restore yview
		self.yview(tk.MOVETO, yview[0])

		# Try to find which card to activate
		for i,cw in enumerate(self.cardlist):
			if cw.fromPos() <= active <= cw.toPos():
				self.activate(i)
				self.cardSelect(i)
			# FIXME
			#if cw.card is not None:
			#	if cw.card.pos() in selection:
			#		self.cardSelect(i)

		if self.active is None and self.cardlist:
			self.activate(0)
			self.cardSelect(0)

		self.page.refreshStatus()

		if self.active is not None:
			self.active.see()
			self.cardSelect(self.active)
		if Global.profiler:
			Global.dump("+++ InputPage.displayCards: t=",time.time()-startTime,
				"rate=",len(self.cardlist)/(time.time()-startTime),"cards/sec")
		self.disable()
		self.page.notBusy()

	# ----------------------------------------------------------------------
	# Check card widget list and input card list for non-existing cards
	# ----------------------------------------------------------------------
	def insertAdditionalCards(self):
		self.page.busy()
		lastpos    = len(self.input.cardlist)
		startTime  = self.resetInterrupt()
		lastUpdate = startTime
		longNotify = False
		self.layoutCache(True)		# Enable caching of lists
		for i in range(len(self.cardlist)-1,0,-1):	# In reverse order
			cw = self.cardlist[i]
			if cw.card is not None:
				if cw.card.input is None:	# Card deleted
					self.delCards(i)
				elif cw.card.pos() < 0:
					# delete card
					self.delCards(i)
				elif cw.card.pos() < lastpos-1:
					# insert additional cards
					for card in self.input.cardlist[cw.card.pos()+1 : lastpos]:
						i += 1
						# timeout add them as HiddenCardWidget
						ti = time.time()
						if ti-startTime > _CARDS_TIMEOUT:
							if not self.flair.notifyIsVisible():
								self.flair.notify("Too long",
									"Process takes too long",
									tkFlair.NOTIFY_NORMAL,
									"Interrupt?", self.interrupt)
								longNotify = True
							if self._interrupt:
								self.addCard(i, card,
									self.input.cardlist[lastpos])
								startTime = self.resetInterrupt()
								break
						if ti-lastUpdate > _CARDS_UPDATE:
							lastUpdate = ti
							self.flair.update()
						self.addCard(i, card)
					lastpos = cw.card.pos()
				else:
					lastpos = cw.card.pos()
			elif cw.fromPos() < 0 or cw.toPos() < 0:
				self.delCards(i)
			elif cw.toPos() < lastpos-1:
				# insert additional cards
				for card in self.input.cardlist[cw.toPos()+1 : lastpos]:
					i += 1
					# timeout add them as HiddenCardWidget
					ti = time.time()
					if ti-startTime > _CARDS_TIMEOUT:
						if not self.flair.notifyIsVisible():
							self.flair.notify("Too long",
								"Process takes too long",
								tkFlair.NOTIFY_NORMAL,
								"Interrupt?", self.interrupt)
							longNotify = True
						if self._interrupt:
							self.addCard(i, card,
								self.input.cardlist[lastpos])
							startTime = self.resetInterrupt()
							break
					if ti-lastUpdate > _CARDS_UPDATE:
						lastUpdate = ti
						self.flair.update()
					self.addCard(i, card)
				lastpos = cw.fromPos()
			else:
				lastpos = cw.fromPos()
		self.reposition()
		assert self.checkPosition()
		self.layoutCache(False)		# Disable for normal editing
		if longNotify: self.flair.notifyHide()
		self.page.notBusy()

	# ----------------------------------------------------------------------
	# Refresh editor, refresh only cardwidgets with new values
	# ----------------------------------------------------------------------
	def refresh(self, start=0, force=False):
		Global.dump("*-* InputEditor.refresh start=",start)
		self.layoutCache(True)		# Enable caching of lists
		startTime = time.time()
		if not force: assert self.checkPosition()
		self.enable()
		y = 1 if start==0 else self.cardlist[start].y
		for cw in self.cardlist[start:]:
			if y!=cw.y: cw.reposition(y)
			if force or cw.isModified():
				#dump(f"+++ Redraw: {cw.card.tag} {y=} {cw.y=} {cw.yend=}")
				#breakpoint()
				#dump(self.dump(cw.index(), "end"))
				cw.redraw()
				#dump(f">>> {cw.card.tag} {cw.y=} {cw.yend=}")
				#dump("+++ check=", cw.checkPosition())
			y = cw.yend
		#self.delete(f"{y}.0", tk.END)
		self.disable()
		self.cardShowSelection()
		self.layoutCache(False)
		self.page.refreshStatus()
		if Global.profiler:
			Global.dump("+++ InputPage.refresh: t=",time.time()-startTime)
		assert self.checkPosition()
		assert self.input.checkNumbering()	# XXX

	# ----------------------------------------------------------------------
	# Card ACTIVATION
	# ----------------------------------------------------------------------
	# Activate the CardWidget containing the card
	# ----------------------------------------------------------------------
	def activate(self, card=None, select=False):
		Global.dump("*-* activate",card)
		#traceback.print_stack()
		# If normal Input card then search for card widget
		if isinstance(card, Card):
			for cw in self.cardlist:
				if cw.card is card:
					card = cw
					break
			else:
				card = self.addCardSorted(card)
				if card is None: return
				#raise Exception("Card to be activated not found")
			cid  = self.cardlist.index(card)
		elif isinstance(card, int):
			cid  = card
			if cid>=len(self.cardlist):
				cid = len(self.cardlist)-1
			try:
				card = self.cardlist[cid]
			except IndexError:
				cid   = 0
				card = self.cardlist[cid] if self.cardlist else None
		elif card is None:
			pass
		else:
			cid  = self.cardlist.index(card)

		# Card Widget
		if self.active is not None:
			if self.active is card:
				# redraw frame
				self.tag_add(_ACTIVE_TAG, card.index(),
							card.indexEnd())
				if cid in self.selection:
					self.tag_add(_SELECT_TAG,
							card.index(),
							card.indexEnd())
				if self.compare(tk.INSERT, "<", card.index()) or \
				   self.compare(tk.INSERT, ">", card.indexEnd()):
					self.mark_set(tk.INSERT, card.index())
				return

			# Delete previous tag
			self.tag_remove(_ACTIVE_TAG, "1.0", tk.END)
			self.tag_remove(tk.SEL,      "1.0", tk.END)

		# If card is not found do nothing
		if card is None:
			self.active     = None
			self.cardAnchor = None
			self.page.updateLabel()
			return

		# Activate
		card.updateState()
		self.active = card
		self.cardAnchor = cid
		self.tag_add(_ACTIVE_TAG, card.index(), card.indexEnd())
		if cid in self.selection:
			self.tag_add(_SELECT_TAG, card.index(), card.indexEnd())
		if select:
			self.cardSelect(cid)
		if self.compare(tk.INSERT, "<", card.index()) or \
		   self.compare(tk.INSERT, ">", card.indexEnd()):
			self.mark_set(tk.INSERT, card.index())

		card.updateLabel()

		# Avoid unnecessary scrolling for undo/redo on the second editor
		# If editor doesn't have the focus do not scroll to see it
		#try:
		#	if self is self.focus_get():
		#		card.see()
		#except KeyError:
		#	# In case of "tearoff" for a submenu!
		#	pass
		self.page.refreshStatus()
		#return undoinfo

	# ----------------------------------------------------------------------
	# Activate card if needed. This is typically called from the undo
	# and activates the card where the action is taken
	# ----------------------------------------------------------------------
	def activateIfNeeded(self, card):
		Global.dump("*-* InputEditor.activateIfNeeded",self.page.isactive(), self.flair.inundo())
		if self.page.isactive() and not self.flair.inundo(): return
		if self.active is None or self.active.card is not card:
			if not self.flair.inundo(): self.commit()
			self.cardSelectClear()
			self.activate(card)
			self.cardSelect(self.cardAnchor)
			self.active.see()

		if self.active is not None and self.active.card is not None:
			self.page.updateLabel(self.active.card)
			self.active.updateLayout()
			oldState = self["state"]
			self.enable()
			if self.active.redraw():
				self.cardSelectClear()
				self.cardSelect(self.cardAnchor)
				self.reposition(self.activePos())
			self.config(state=oldState)

	# ----------------------------------------------------------------------
	# Activate previous card
	# ----------------------------------------------------------------------
	def activatePrevious(self):
		if not self.cardlist: return False
		if not self.active:
			self.activate(0)
			self.active.see(end=True)
			return False
		y = self.active.y
		if isinstance(self.active, CardWidget): y += 1
		if self.bbox(f"{y}.0") is None:
			# Too long text scroll window
			self.yview(tk.SCROLL, -1, tk.UNITS)
			return False
		a = self.activePos()-1
		if a < 0: return False
		self.activate(a)
		self.active.see(end=True)
		return True

	# ----------------------------------------------------------------------
	# Activate next card
	# ----------------------------------------------------------------------
	def activateNext(self):
		if not self.cardlist: return False
		if not self.active:
			self.activate(len(self.cardlist)-1)
			self.active.see()
			return False
		if self.bbox(f"{self.active.yend-1}.0") is None:
			# Too long text scroll window
			self.yview(tk.SCROLL, 1, tk.UNITS)
			return False
		a = self.activePos() + 1
		if a >= len(self.cardlist): return False
		self.activate(a)
		self.active.see()
		return True

	# ----------------------------------------------------------------------
	# Activate and edit at index, commit if needed
	# ----------------------------------------------------------------------
	def activateIndex(self, index, shift, edit):
		cw = self.findCardWidget(index)
		if cw is None:
			return
		elif cw is self.active:
			item = cw.item(index)
			if item == self.edit:
				if shift:
					self.setAnchor(tk.INSERT)
					if self.compare(tk.INSERT, "<=", index):
						self.tag_add(tk.SEL, tk.INSERT, index)
					else:
						self.tag_add(tk.SEL, index, tk.INSERT)
				self.mark_set(tk.INSERT, index)
			elif not shift: # only without Shift pressed
				self.unpostList()
				self.commit()
				self.mark_set(tk.INSERT, index)
				self.active.editItem(item, postlist=False)
			self.see(tk.INSERT)
		elif not shift:	# only without Shift pressed
			self.unpostList()
			self.commit()
			self.active.editItem()
			self.cardSelectClear()
			self.activate(cw)
			self.cardSelect(cw)
			if edit:
				item = cw.item(index)
				self.mark_set(tk.INSERT, index)
				cw.editItem(item,postlist=False)

		# Correct INSERT position if needed
		if self.edit is not None:
			tagrange = self.edit_range(tk.INSERT)
			if not tagrange:
				self.edit = None
			elif self.compare(tk.INSERT, ">", tagrange[1]):
				self.mark_set(tk.INSERT, tagrange[1])
				self._anchor = self.index(tk.INSERT)

	# ----------------------------------------------------------------------
	# Commit changes to active card and to selected range with similar cards
	# ----------------------------------------------------------------------
	def commit(self):
		if self.edit   is None: return
		if self.active is None: return

		Global.dump("*-* InputEditor.commit")
		Global.dump("\tCard:", self.active.card)
		Global.dump("\tEdit:", self.edit)

		# Remember active (to be restored in the end)
		active = self.active

		yview = self.yview()
		try:
			layoutchanged = active.commit()
			self.yview(tk.MOVETO, yview[0])
			if not layoutchanged: return
		except:
			# Recover from possibly fatal errors
			self.edit = None
			active = None
			raise

		# Commit on multiple cards
		if len(self.selection)<=1: return

		if self.edit < 0:	# Comment changed
			val = active.card.comment()
		else:
			dat = active.layout[self.edit][1]
			# Get the value!
			if isinstance(dat, tuple): #List
				dat = dat[0]
			if callable(dat):
				val = dat(self.layout[active.card.MODE], active.card)
			else:
				val = active.card[dat]

		undoinfo = []
		self.enable()
		for f,t in reversed(self.selection2Ranges()):
			# Skip expression change in regions
			if active.card.tag == "REGION" and dat==-1:
				#self.page._refreshOtherFrames()
				return

			# Check for the same card type
			for cw in reversed(self.cardlist[f:t+1]):
				if cw is active: continue	# ignore active card
				if cw.card is None: continue	# Ignore HiddenCardWidget

				self.active = cw	# temporary fool the system

				if self.edit < 0:
					undoinfo.append(self.flair.setCommentUndo(cw.card, val))
					cw.redraw()
					continue

				# Ignore other type of cards
				if cw.card.tag != active.card.tag: continue

				# Modify the value
				if callable(dat):
					dat(self.layout[cw.card.MODE], cw.card, val)
				else:
					undoinfo.append(self.flair.setCardUndo(cw.card, dat, val))
				cw.redraw()
				self.cardSelect(cw)
		self.disable()

		if undoinfo:
			self.addUndo(undoinfo, undoinfo[0][0])
			self.reposition(f)

		self.active = active

		# multiple cards will set the ACTIVE_TAG around them
		self.tag_remove(_ACTIVE_TAG, "1.0", tk.END)
		self.tag_add(_ACTIVE_TAG, self.active.index(), self.active.indexEnd())

		self.page._refreshOtherFrames()

	# ----------------------------------------------------------------------
	# CARD SELECTION
	# ----------------------------------------------------------------------
	# Clear Card selection
	# ----------------------------------------------------------------------
	def cardSelectClear(self):
		#Global.dump("*-* InputPage.cardSelectClear")
		if not self.selection: return
		self.tag_remove(_SELECT_TAG, "1.0", tk.END)
		self.selectionClear()

	# ----------------------------------------------------------------------
	def selectionClear(self):
		# FIXME do we need to flag also the selected bit for GeoViewer?
		self.selection.clear()
#		for card in self.input.cardlist:
#			card[Card.SELECT]  = card.get(Card.SELECT,0) & BIT.UNSELECT

	# ----------------------------------------------------------------------
	# Handler of the <<Selection>> event
	# ----------------------------------------------------------------------
#	def selectionChanged(self, event):
#		Global.dump("*-* selectionChanged")
#		# remove selection during drag and drop
#		self.tag_remove(tk.SEL, "1.0", tk.END)

	# ----------------------------------------------------------------------
	# Highlight the current selection set
	# ----------------------------------------------------------------------
	def cardShowSelection(self):
		Global.dump("*-* InputPage.cardShowSelection")
		self.tag_remove(_SELECT_TAG, "1.0", tk.END)
		for f,t in self.selection2Ranges():
			self.tag_add(_SELECT_TAG,
					self.cardlist[f].index(),
					self.cardlist[t].indexEnd())

	# ----------------------------------------------------------------------
	# Select card range [from..to] (including end)
	# ----------------------------------------------------------------------
	def cardSelect(self, from_card, to_card=None, select=True, toggle=False):
		#Global.dump("*-* InputPage.cardSelect")
		if from_card is None:
			return
		elif isinstance(from_card, (list,set)):
			self.selection.update(from_card)
			self.cardShowSelection()
			return
		elif isinstance(from_card, int):
			f = from_card
		else:
			f = self.cardlist.index(from_card)

		if to_card is None:
			t = f
		elif isinstance(to_card, int):
			t = min(to_card, len(self.cardlist)-1)
		elif to_card==tk.END:
			t = len(self.cardlist)-1
		else:
			t = self.cardlist.index(to_card)

		# swap values
		if f>t: f,t = t,f

		# check for toggle and override select if any
		if toggle:
			select = f not in self.selection

		# add selection tag
		try:
			if select:
				self.tag_add(_SELECT_TAG,
						self.cardlist[f].index(),
						self.cardlist[t].indexEnd())
				for i in range(f,t+1):
					# FIXME do we need to flag also the selected bit for GeoViewer?
					self.selection.add(i)
			else:
				self.tag_remove(_SELECT_TAG,
						self.cardlist[f].index(),
						self.cardlist[t].indexEnd())
				for i in range(f,t+1):
					try:
						# FIXME do we need to flag also the selected bit for GeoViewer?
						self.selection.remove(i)
					except KeyError:
						pass
		except IndexError:
			pass

	# ----------------------------------------------------------------------
	# Unselect cards
	# ----------------------------------------------------------------------
	def cardUnSelect(self, from_card, to_card=None):
		Global.dump("*-* InputPage.cardUnSelect")
		self.cardSelect(from_card, to_card, False)

	# ----------------------------------------------------------------------
	# Create selection from ranges
	# ----------------------------------------------------------------------
	def cardSelectFromRanges(self, ranges):
		Global.dump("*-* InputPage.cardSelectFromRanges")
		for f,t in ranges:
			self.cardSelect(f,t)

	# ----------------------------------------------------------------------
	# card selection toggle
	# ----------------------------------------------------------------------
	def cardSelectToggle(self, card):
		Global.dump("*-* InputPage.cardSelectToggle")
		idx = card if isinstance(card, int) else self.cardlist.index(card)

		# FIXME do we need to flag also the selected bit for GeoViewer?
		if idx in self.selection:
			if len(self.selection)>1:
				self.tag_remove(_SELECT_TAG,
							self.cardlist[idx].index(),
							self.cardlist[idx].indexEnd())
				self.selection.remove(idx)
				if self.cardlist[idx] is self.active:
					# activate another card from the remaining selection
					self.activate(list(self.selection)[-1])
		else:
			self.tag_add(_SELECT_TAG,
						self.cardlist[idx].index(),
						self.cardlist[idx].indexEnd())
			self.selection.add(idx)

	# ----------------------------------------------------------------------
	# Select a range of cards enclosed with a start_tag .. stop_tag
	# ----------------------------------------------------------------------
	def cardSelectTagRange(self, start_tag, stop_tag, select=True):
		Global.dump("*-* InputPage.cardSelectTagRange")
		cid   = self.activePos()
		level = 1
		if self.active.card.tag == start_tag:
			fromid = cid
			toid   = fromid+1
			while toid<len(self.cardlist):
				cw = self.cardlist[toid]
				if cw.card is None:
					pass
				elif cw.card.tag == start_tag:
					level += 1
				elif cw.card.tag == stop_tag:
					level -= 1
					if level==0: break
				toid += 1
			self.cardSelect(fromid, toid, select)
		else:
			toid   = cid
			level  = 1
			fromid = toid-1
			while fromid>0:
				cw = self.cardlist[fromid]
				if cw.card is None:
					pass
				elif cw.card.tag == start_tag:
					level -= 1
					if level==0: break
				elif cw.card.tag == stop_tag:
					level += 1
				fromid -= 1
			self.cardSelect(fromid, toid, select)

	# ----------------------------------------------------------------------
	# Select All, text or cards
	# ----------------------------------------------------------------------
	def selectAll(self, event=None):
		Global.dump("*-* InputPage.selectAll")
		if self.edit is None:
			self.cardSelect(0, tk.END)
		else:
			tagrange = self.edit_range(tk.INSERT)
			if tagrange:
				self.setAnchor()
				self.tag_remove(tk.SEL, "1.0", tk.END)
				self.tag_add(tk.SEL, *tagrange)
		return "break"

	# ----------------------------------------------------------------------
	def selectClear(self, event=None):
		Global.dump("*-* InputPage.selectClear")
		if self.edit is None:
			self.cardSelectClear()
			if self.active is not None:
				self.cardSelect(self.active)
		else:
			self.tag_remove(tk.SEL, "1.0", tk.END)
		return "break"

	# ----------------------------------------------------------------------
	def isTextSelected(self):
		try:
			self.index(tk.SEL_FIRST)
			return True
		except tk.TclError:
			return False

	# ----------------------------------------------------------------------
	# @return a sorted list of selected ranges (from, to)
	# ----------------------------------------------------------------------
	def selection2Ranges(self):
		ranges = []
		if not self.selection: return ranges
		selection = sorted(list(self.selection))
		start = selection[0]
		stop  = selection[0]
		for cid in selection[1:]:
			if cid-stop != 1:
				ranges.append((start,stop))
				start = cid
			stop = cid
		ranges.append((start,stop))
		return ranges

	# ----------------------------------------------------------------------
	# @return a sorted list of selected ranges (from, to)
	# ----------------------------------------------------------------------
	def _selection2Ranges(self):
		ranges = []
		if not self.selection: return ranges
		selection = sorted(list(self.selection))
		start = selection[0]
		stop  = selection[0]
		for cid in selection[1:]:
			if cid-stop != 1:
				ranges.append((start,stop+1))
				start = cid
			stop = cid
		ranges.append((start,stop+1))
		return ranges

	# ----------------------------------------------------------------------
	# @return a set with all selected cards (not CardWidget)
	# ----------------------------------------------------------------------
	def selection2Cards(self):
		selection = set()
		for i in sorted(self.selection):
			cw = self.cardlist[i]
			for j in range(cw.fromPos(), cw.toPos()+1):
				selection.add(j)
		return selection

	# ----------------------------------------------------------------------
	# CLIPBOARD
	# ----------------------------------------------------------------------
	# Copy selected content to clipboard
	# ----------------------------------------------------------------------
	def copy(self, event=None):
		if self.edit is not None:
			try:
				self.index(tk.SEL_FIRST)
				self.clipboard_clear()
				self.clipboard_append(self.get(tk.SEL_FIRST, tk.SEL_LAST))
			except tk.TclError:
				pass
		elif self.active:
			self._copy2Clipboard()
		return "break"

	# ----------------------------------------------------------------------
	# Cut selected content to clipboard
	# ----------------------------------------------------------------------
	def cut(self, event=None):
		if self.edit is not None:
			try:
				self.index(tk.SEL_FIRST)
				self.copy()
				self.delete(tk.SEL_FIRST, tk.SEL_LAST)
			except tk.TclError:
				pass
		elif self.active:
			self._copy2Clipboard()
			self.delSelectedCards()
		return "break"

	# ----------------------------------------------------------------------
	# Paste clipboard content
	# ----------------------------------------------------------------------
	def paste(self, event=None):
		if self.edit is not None:
			self.pasteText(self.flair.clipboard())
		else:
			cards = self.flair.clipboard2Cards()
			if not cards: return "break"
			self.cardSelectClear()
			card_from, card_to = self._pasteBuffer(cards)
			self.activate(card_from)
			self.active.see()
			if card_to >= card_from:
				self.cardSelect(card_from, card_to)
			self.focusIn()
			self.page.refreshStatus()
		return "break"

	# ----------------------------------------------------------------------
	# Clone selected cards
	# ----------------------------------------------------------------------
	def clone(self, event=None):
		# During editing it pastes the old text....
		if self.active is None or not self.selection: return
		self.commit()
		self.active.editItem()
		# Find last item to active
		last = max(self.selection)
		self.copy()
		self.activate(last)
		self.active.see()
		self.paste()
		return "break"

	# ----------------------------------------------------------------
	# Clone selected cards, only their tags not values
	# ----------------------------------------------------------------
	def cloneSpecial(self, event=None):
		return self.clone(event)	# FIXME not working for the moment
		## During editing it pastes the old text....
		#if self.active is None or not self.selection: return
		#self.commit()
		#self.active.editItem()
		## Find last item to active
		#last = max(self.selection)
		#cards = []
		#for f,t in self.selection2Ranges():
		#	for cw in self.cardlist[f : t+1]:
		#		if cw.card is None:
		#			pass
		#		else:
		#			newcard = cw.card.__class__(cw.card.tag)
		#			cards.append(newcard)

		#pos = last+1
		#for card in cards:
		#	self.addCard
		#	pos += 1
		#self.activate(last)
		#self.active.see()
		#return "break"

	# ----------------------------------------------------------------
	# Paste text
	# ----------------------------------------------------------------
	def pasteText(self, txt):
		try:
			self.delete(tk.SEL_FIRST, tk.SEL_LAST)
		except tk.TclError:
			pass
		ft = self.edit_range(tk.INSERT)
		if not ft:
			Global.dump(">>> pateText: Cannot insert text...")
			return
		if self.editType != _EDIT_MULTILINE:
			if txt.find("\n")>=0 or len(txt)>80:
				self.flair.notify("Cannot paste",
					"String too long to paste",
					tkFlair.NOTIFY_ERROR)
				return
		self.insert(tk.INSERT, txt, self.tag_names(tk.INSERT))
		self.see(tk.INSERT)
		self.active.updateEnd()
		self.reposition(self.activePos())
		return "break"

	# ----------------------------------------------------------------------
	# Copy selected range to clipboard
	# ----------------------------------------------------------------------
	def _copy2Clipboard(self):
		hidden = 0		# Hidden group
		if not self.selection: return
		cards = []
		for f,t in self.selection2Ranges():
			for cw in self.cardlist[f : t+1]:
				if cw.card is None:
					hidden += 1
					for c in cw.cards():
						c[Card.HIDDEN] = hidden
						cards.append(c)
				else:
					cw.card[Card.HIDDEN] = 0
					cards.append(cw.card)
		self.flair.cards2Clipboard(cards)

	# ----------------------------------------------------------------------
	# Insert clipboard at active position
	# ----------------------------------------------------------------------
	def _pasteBuffer(self, cards):
		pos   = 0
		cwpos = 0
		if self.active is not None:
			if isinstance(self.active, HiddenCardWidget):
				pos = self.active.toPos()+1
			else:
				pos = self.active.fromPos()+1
			cwpos = self.activePos()+1

		fromcwpos = cwpos
		hidden    = 0
		fromcard  = None
		newcard   = None
		undoinfo = self.flair.undolistRenumberOnly()
		self.layoutCache(False)	# Clean caching of lists

		renamed = ""
		self.enable()
		for card in cards:
			h = card.get(Card.HIDDEN,0)
			tag = card.tag
			if card.type() == Card.BODY:
				lst = self.layout[card.MODE].bodyList(False)
			elif card.type() == Card.CELL:
				lst = self.layout[card.MODE].cellList()
			elif card.type() == Card.SOLID:
				lst = self.layout[card.MODE].typedefList("solid")
			elif card.type() == Card.MATERIAL:
				lst = self.layout[card.MODE].materialList()
			elif tag == "#define":
				lst = self.layout[card.MODE].defineList()
			elif tag in ("EVENTBIN", "RESNUCLE", "USRBDX", "USRBIN",
				     "USRCOLL", "USRTRACK", "USRYIELD"):
				lst = self.layout[card.MODE].binningList()
			#elif tag == "ROT-DEFI":
			#	lst = self.layout[card.MODE].rotationList()
			else:
				lst = None

			lastcard = newcard
			newcard  = card.clone()

			# Check for duplicate
			if card.name() and lst is not None and card.name() in lst:
				# Find unique name
				# find number in the end of the name
				pat = Project._LAST_NUM.match(card.name())
				if pat:
					name = pat.group(1)
					n = int(pat.group(2))+1
				else:
					name = card.name()
					n = 1
				while True:
					sn = str(n)
					guess = name[:8-len(sn)]+sn
					if guess not in lst:
						renamed += f"   {tag}:\t{name}\t-> {guess}\n"
						newcard.setName(guess)
						break
					n += 1

			# Add to undo
			undoinfo.append(self.flair.addCardUndo(newcard, pos, False))
			pos += 1
			if h != 0:
				if h != hidden:
					if fromcard:
						# Create widget
						self.addCard(cwpos, fromcard, lastcard)
						cwpos += 1
					hidden = h
					fromcard = newcard
			else:
				if fromcard:
					# Create widget
					self.addCard(cwpos, fromcard, lastcard)
					cwpos += 1
					fromcard = None
				self.addCard(cwpos, newcard)
				cwpos += 1
		if fromcard:
			self.addCard(cwpos, fromcard, newcard)
		self.disable()

		self.flair.addUndolistRenumberOnly(undoinfo, "Paste card(s)")

		if renamed:
			self.flair.notify("Cards renamed",
				f"The following cards have been renamed:\n{renamed}",
				tkFlair.NOTIFY_WARNING)

		self.layoutCache(False)		# Disable caching of lists
		self.reposition(fromcwpos)
		assert self.checkPosition()
		assert self.input.checkNumbering()	# XXX
		self.input.clearCache()
		self.page.preprocess()	# FIXME only if #define/#undef is present
		self.page._refreshOtherFrames()
		return fromcwpos, cwpos-1

	# ----------------------------------------------------------------
	# POPUP LIST
	# ----------------------------------------------------------------
	# listbox handling methods
	# ----------------------------------------------------------------------
	def postList(self, lst=None, sel=None):
		self.hideBalloon()

		tagrange = self.tag_range(self.editTag, tk.INSERT)
		if not tagrange: return

		bb0 = self.bbox(tagrange[0])
		bb1 = self.bbox(tagrange[1])
		if not bb0 or not bb1:
			# Try to scroll item in view
			self.see(tagrange[0])
			bb1 = self.bbox(tagrange[1])
			if not bb1:
				self.xview(tk.SCROLL, 20, tk.UNITS)
				bb1 = self.bbox(tagrange[1])
			bb0 = self.bbox(tagrange[0])
			if not bb0: return

		margin = 2
		x = bb0[0]
		if bb1:
			w = max(bb1[0] - x, 100)
		else:
			w = self.winfo_width()-x

		x -= margin
		y = bb0[1]+bb0[3]
		h = self.popupList.height()
		if y+h > self.winfo_height():
			y = bb0[1]-h - margin

		self._insertText = None
		if lst is not None:
			self.popupList.fill(lst, sel)

		self.popupList.post(x,y,w,h)

	# ----------------------------------------------------------------------
	# Display a popup list with current bodies
	# ----------------------------------------------------------------------
	def _insertList(self, lst):
		bb = self.bbox(tk.INSERT)
		x = bb[0]
		w = self.popupList.width()
		y = bb[1]+bb[3]

		h = self.popupList.height()
		if y+h > self.winfo_height():
			y = bb[1]-h

		self._insertText = self.index(tk.INSERT)	# remember starting position
		self.popupList.fill(lst, "")
		self.popupList.post(x,y,w,h)

	#-----------------------------------------------------------------------
	def insertBody(self):
		lst = self.layout[self.active.card.MODE].bodyList()
		lst.insert(1, Layout.BODY_PARENTHESIS)
		self._insertList(lst)

	# ----------------------------------------------------------------------
	def insertRegion(self):
		lst = self.layout[self.active.card.MODE].regionList()
		lst.insert(1, Layout.BODY_PARENTHESIS)
		self._insertList(lst)

	# ----------------------------------------------------------------------
	def unpostList(self):
		self.popupList.unpost()

	# ----------------------------------------------------------------------
	def popupListSelected(self, event=None):
		Global.dump("*-* InputPage.popupListSelected")
		self.focus_set()
		if self._insertText is not None:
			self.active.insertText(self.popupList.selection()[0])
			self.commit()
		else:
			item = self.edit
			self.commit()
			tagrange = self.active.itemRange(item)
			if tagrange:
				self.mark_set(tk.INSERT, tagrange[1]+"-3c")
				self.active.editItem(item)

	# ----------------------------------------------------------------------
	# Populate filter search listbox
	# ----------------------------------------------------------------------
	def filterList(self, key=None):
		if self._insertText is None:
			txt = self.edit_get(tk.INSERT).strip()
		else:
			txt = self.get(	self._insertText,
					self._insertText+" wordend").strip()

		if not txt:
			self.popupList.resetSearch()
			self.popupList.filter(refill=True)
			return

		refill = False
		if key=="Delete":
			self.popupList.resetSearch()
			refill = True

		if self.active._save[1] == txt:
			self.popupList.resetSearch()
			self.popupList.filter(refill=True)
			self.popupList.select(txt)
		else:
			if key=="BackSpace": refill = True
			self.popupList.filter(txt, refill)

	# ----------------------------------------------------------------
	# Tab key
	# ----------------------------------------------------------------------
	def tabKey(self, event):
		if self.edit is not None:
			self.commit()
			self.active.editNext()
		return "break"

	# ----------------------------------------------------------------------
	def tabShiftKey(self, event):
		if self.edit is not None:
			self.commit()
			self.active.editPrev()
		return "break"

	# ----------------------------------------------------------------------
	# BALLOON
	# ----------------------------------------------------------------------
	def showBalloon(self):
		self._motionId = None
		self.delay     = 500	# reduce delay to 500 ms for next popup

		cw = self.findCardWidget(tk.CURRENT)
		if cw is None: return

		# Prepare the text
		self._help = "Sorry: no info available..."	# default message

		if cw.card is None:	# Hidden Cards
			self._help = f"{cw.toCard.pos()-cw.fromCard.pos()+1} hidden cards"
		else:
			item = cw.item(tk.CURRENT)
			if item is None:
				self._help = f"Card: {cw.card.tag}\n{cw.card.info.meaning}"
			elif item == -1:
				self._help = "Comment text"
			else:
				try:
					lbl,dat,*desc = cw.layout[item]
				except IndexError:
					return

				if lbl is None: lbl=""
				if isinstance(dat, tuple): dat = dat[0]
				if isinstance(dat, int):
					if dat==0:
						what = "#0"
					else:
						what = f"#{dat}"
				else:
					what = None

				# if description is present
				if desc:
					if what is None:
						self._help = desc[0]
					else:
						self._help = f"{what}: {desc[0]}"
				elif what is not None:
					case = cw.card.case()
					if case is not None:
						try:    guide = cw.card.info.guide[case][dat]
						except: guide = lbl
						try:    default = cw.card.info.default[case][dat]
						except: default = "-"
						try:	type_ = cw.card.info.range[case][dat]
						except: type_ = ""
						self._help = f"{cw.card.tag} {lbl}: {what} [{type_}]\n"\
							f"{guide}\nDefault: {default}"
					else:
						self._help = what

				# REGION expression, check for bodies
				if cw.card.tag in ("TITLE","REGION",":CELL") and dat==-1:
					# get word under cursor
					name = self.get(tk.CURRENT+" wordstart",
							tk.CURRENT+" wordend")
					body = None
					if name=="VOXEL" and "VOXELS" in self.input.cards:
						body = self.input.cards["VOXELS"][0]
					else:
						for tag in CardInfo.bodies[self.project.mode()]:
							try:
								for card in self.input.cards[tag]:
									if card.name() == name:
										body = card
										break
							except KeyError:
								pass
							if body: break
					if body:
						self._help = str(body)
		# Show the balloon
		tkExtra.Balloon.setWidget(self,
			self._motionX + self.winfo_rootx(),
			self._motionY + self.winfo_rooty())
		tkExtra.Balloon.show()

	# ----------------------------------------------------------------------
	def hideBalloon(self, event=None):
		if self._motionId is not None:
			self.after_cancel(self._motionId)
			self._motionId = None
		else:
			tkExtra.Balloon.hide()

	# ----------------------------------------------------------------------
	# POPUP MENU
	# ----------------------------------------------------------------------
	def _addCardLambda(self, tag):
		if isinstance(tag, str):
			ci = CardInfo.get(self.project.mode(), tag)
		else:
			ci = tag
		return lambda t=ci.tag,s=self:s.addNewCardUndo(t)

	# ----------------------------------------------------------------------
	# Create add a new card menu
	# ----------------------------------------------------------------------
	def _addCardMenu(self, menu, tags, popup=True, prepro=True):
		groups = set()
		if tags:
			# Specific cards
			tags = tags[:]
			tags.sort()

			count = defaultdict(int)	# count # of cards per group

			# Find all groups from cards
			for t in tags:
				ci = CardInfo.get(self.project.mode(), t)
				for group in ci.group:
					groups.add(group)
					count[group] += 1

			#-------------------------------------------------------
			def tagsRemoveGroup(group):
				nonlocal tags
				i = 0
				while i<len(tags):
					tag = tags[i]
					ci = CardInfo.get(self.project.mode(), tag)
					if group in ci.group:
						del tags[i]
					else:
						i += 1
			#-------------------------------------------------------

			if CardInfo.GEOMETRY_GROUP in groups:
				for group in (  CardInfo.GEOMETRY_BODY_GROUP,
						#CardInfo.GEOMETRY_SURFACE_GROUP,
						CardInfo.GEOMETRY_SOLID_GROUP,
						CardInfo.GEOMETRY_VOLUME_GROUP,
						CardInfo.GEOMETRY_TRANSFORM_GROUP):
					try:
						groups.remove(group)
					except KeyError:
						pass

			# Check if solids/volumes are present and remove them
			ngroups = len(groups)
			for group in (  CardInfo.GEOMETRY_BODY_GROUP,
					CardInfo.GEOMETRY_SOLID_GROUP,
					CardInfo.GEOMETRY_VOLUME_GROUP,
					CardInfo.GEOMETRY_TRANSFORM_GROUP):
				if group not in groups: continue
				groups.remove(group)
				tagsRemoveGroup(group)
				if ngroups>1 or (popup and prepro):
					submenu = tk.Menu(menu,
							activebackground=Ribbon._ACTIVE_COLOR)
				else:
					submenu = menu
				if group == CardInfo.GEOMETRY_BODY_GROUP:
					tkFlair.bodiesMenu(submenu,
							self.project.mode(),
							self.addNewCardUndo)
				else:
					tkFlair.cardsGroupMenu(submenu,
							self.project.mode(),
							group,
							self.page.editor().addNewCardUndo,
							activebackground=Ribbon._ACTIVE_COLOR)
				if submenu is not menu:
					menu.add_cascade(label=group.split("/")[-1],
							menu=submenu,
							image=icons.get(CardInfo.groups_icons[group]),
							compound=tk.LEFT,
							underline=0)

			#-------------------------------------------------------
			def addSubmenu(menu, group):
				submenu = tk.Menu(menu, activebackground=Ribbon._ACTIVE_COLOR)
				if tkFlair.cardsGroupMenu(submenu,
							self.project.mode(),
							group,
							self.page.editor().addNewCardUndo,
							activebackground=Ribbon._ACTIVE_COLOR):
					try:
						icon = icons.get(CardInfo.groups_icons[group])
					except:
						icon = icons.get("add")
					menu.add_cascade(label=group,
							menu=submenu,
							image=icon,
							compound=tk.LEFT,
							underline=0)
				return submenu
			#-------------------------------------------------------

			# Show groups
			if not groups:
				pass
			elif len(groups)>1:
				first = True
				# First pass all cards in group
				for group in groups:
					if count[group] == len([x for x in CardInfo.cards(self.project.mode())
							if group in x.group]):
						if not first: menu.add_separator()
						first = False
						tkFlair.cardsGroupMenu(menu,
								self.project.mode(),
								group,
								self.page.editor().addNewCardUndo,
								activebackground=Ribbon._ACTIVE_COLOR)
				# Second pass all other groups
				for group in groups:
					if count[group] != len([x for x in CardInfo.cards(self.project.mode())
							if group in x.group]):
						if not first: menu.add_separator()
						# include it as a submenu
						submenu = addSubmenu(menu, group)
			else:
				group = list(groups)[0]
				# inline the only one group here
				if len(tags)>10 and popup:
					submenu = addSubmenu(menu, group)
				else:
					tkFlair.cardsGroupMenu(menu,
							self.project.mode(),
							group,
							self.page.editor().addNewCardUndo,
							activebackground=Ribbon._ACTIVE_COLOR)

			#underline = tkFlair.initUnderline(tags, self.project.mode())
			#if not popup and len(tags) <= 16:
			#	for t in tags:
			#		tkFlair.addCardCommand(menu,
			#				self.project.mode(),
			#				t,
			#				self._addCardLambda(t),
			#				underline)

		else:	# Groups
			if popup:
				submenu = tk.Menu(menu, activebackground=Ribbon._ACTIVE_COLOR)
				menu.add_cascade(label="Add",
						menu=submenu,
						image=icons.get("add"),
						compound=tk.LEFT,
						underline=0)
			else:
				submenu = menu

			tkFlair.cardsAllMenu(submenu,
					self.page.flair.project.mode(),
					self.page.editor().addNewCardUndo,
					tearoff=1,
					activebackground=Ribbon._ACTIVE_COLOR)

		if prepro and CardInfo.PREPROCESSOR_GROUP not in groups:
			submenu = tk.Menu(menu, activebackground=Ribbon._ACTIVE_COLOR)
			tkFlair.cardsGroupMenu(submenu,
						self.project.mode(),
						CardInfo.PREPROCESSOR_GROUP,
						self.page.editor().addNewCardUndo,
						activebackground=Ribbon._ACTIVE_COLOR)
			menu.add_cascade(label=CardInfo.PREPROCESSOR_GROUP,
						menu=submenu,
						image=icons.get(
							CardInfo.groups_icons[
								CardInfo.PREPROCESSOR_GROUP]),
						compound=tk.LEFT,
						underline=0)

	# ----------------------------------------------------------------------
	# Filter menu
	# ----------------------------------------------------------------------
	def _addFilterMenu(self, menu):
		filtermenu = tk.Menu(menu, activebackground=Ribbon._ACTIVE_COLOR)
		if InputExtra.filterMenu(self.page,
					filtermenu,
					self.active.card):
			menu.add_cascade(label="Filter on", menu=filtermenu,
					compound=tk.LEFT, image=icons.get("filter"),
					underline=0)

	# ----------------------------------------------------------------------
	# Create the popup to add a new card
	# ----------------------------------------------------------------------
	def popupInsertMenu(self, event=None):
		if self.active:
			self.commit()
			self.active.editItem()
		tags = self.page._tags
		if tags is not None and len(tags)==1:
			self.addNewCardUndo(tags[0])
		else:
			menu = tk.Menu(self, tearoff=0, activebackground=Ribbon._ACTIVE_COLOR)
			self._addCardMenu(menu, tags, False, True)
			try:
				menu.tk_popup(self.winfo_pointerx(), self.winfo_pointery())
			except tk.TclError:
				pass
			return "break"	# is important otherwise the handlekey
					# sends the focus to the editor

	# ----------------------------------------------------------------------
	# Create the popup menu
	# ----------------------------------------------------------------------
	def popupMenu(self, event):
		self.unpostList()
		if int(event.type) == 4:	# Mouse event
			cw = self.findCardWidget(tk.CURRENT)
			if cw is not None:
				# Find if is already selected
				if _SELECT_TAG in self.tag_names(tk.CURRENT):
					if self.active is not None:
						self.commit()
					self.button1(event)
					cw.editItem()

		menu = tk.Menu(self, tearoff=0, activebackground=Ribbon._ACTIVE_COLOR)
		active = self.active

		menu.add_command(label="Cut", underline=2, command=self.cut,
				compound=tk.LEFT, image=icons.get("cut"),
				accelerator="Ctrl-X")
		menu.add_command(label="Copy", underline=0, command=self.copy,
				compound=tk.LEFT, image=icons.get("copy"),
				accelerator="Ctrl-C")
		menu.add_command(label="Paste", underline=0, command=self.paste,
				compound=tk.LEFT, image=icons.get("paste"),
				accelerator="Ctrl-V")
		menu.add_command(label="Clone", underline=0, command=self.clone,
				compound=tk.LEFT, image=icons.get("clone"),
				accelerator="Ctrl-D")

		menu.add_separator()
		self._addCardMenu(menu, self.page._tags)

		if active is not None:
			menu.add_command(label="Delete", underline=0,
					command=self.page.delete,
					compound=tk.LEFT, image=icons.get("x"),
					accelerator="Del")

			self._addFilterMenu(menu)
			menu.add_separator()

			menu.add_command(label='Comment', underline=0,
					compound=tk.LEFT, image=icons.get("comment"),
					command=self.toggleComment,
					accelerator='Ctrl-M')

			menu.add_command(label='Enable/Disable', underline=0,
					compound=tk.LEFT, image=icons.get("enable"),
					command=self.toggleState,
					accelerator='Ctrl-L')

			menu.add_command(label='Show/Hide', underline=0,
					compound=tk.LEFT, image=icons.get("expand"),
					command=self.toggleShow,
					accelerator='Ctrl-I')

			menu.add_command(label="Edit Card", underline=0,
					compound=tk.LEFT, image=icons.get("edit"),
					command=self.editCard,
					accelerator='Ctrl-E')

		menu.add_separator()
		menu.add_command(label='Help',
				compound=tk.LEFT, image=icons.get("info"),
				command=self.help,
				accelerator="F1")

		try:
			menu.tk_popup(event.x_root, event.y_root)
		except tk.TclError:
			pass
		return "break"	# is important otherwise the handlekey
				# sends the focus to the editor

	# ----------------------------------------------------------------------
	# COMMENTS
	# ----------------------------------------------------------------------
	# Insert comment
	# ----------------------------------------------------------------------
	def insComment(self):
		if self.active is None or \
		   isinstance(self.active, HiddenCardWidget): return
		self.commit()
		self.active.editItem()
		self.enable()
		rc = self.active.insComment()
		self.disable()
		if rc:
			pos = self.activePos()
			self.reposition(pos+1)
			self.activate(self.active)
			self.active.see()
		self.mark_set(tk.INSERT, self.active.itemRange(-1)[0])
		self.active.editItem(-1)
		self.page.inputModified()
		assert self.checkPosition()

	# ----------------------------------------------------------------------
	# Delete comment
	# ----------------------------------------------------------------------
	def delComment(self):
		if self.active is None or not self.selection: return
		self.commit()
		self.active.editItem()
		undoinfo = []
		self.enable()
		for f,t in reversed(self.selection2Ranges()):
			for cw in reversed(self.cardlist[f:t+1]):
				for card in cw.cards():
					undoinfo.append(self.flair.setCommentUndo(card,""))
				cw.redraw()
		self.reposition(f)
		self.disable()
		self.addUndo(undoinfo, "Delete Comment")

	# ----------------------------------------------------------------------
	def toggleComment(self, event=None):
		if self.active is None or \
		   isinstance(self.active, HiddenCardWidget): return
		if self.active.card.comment():
			self.delComment()
		else:
			self.insComment()

	# ----------------------------------------------------------------------
	# START..STOP cards
	# ----------------------------------------------------------------------
	# Add start..stop cards around the selected ones
	# ----------------------------------------------------------------------
	def addStartStop(self, startTag, stopTag):
		if not self.selection: return
		undoinfo = self.flair.undolistRenumberOnly()
		for f,t in reversed(self.selection2Ranges()):
			self.activate(t)
			undoinfo.append(self.addNewCard(stopTag, False, False))
			self.activate(max(0,f-1))
			undoinfo.append(self.addNewCard(startTag, False, False))
			# re-indent cards
			for cw in self.cardlist[f+1:t+3]:
				cw.indent()
			self.active.see()
		self.flair.addUndolistRenumberOnly(undoinfo, "Add %s..%s"%(startTag,stopTag))

	# ----------------------------------------------------------------------
	# INCLUDE
	# ----------------------------------------------------------------------
	# Add #include..#endinclude only if a range of cards is selected
	# otherwise add an empty #include..#endinclude
	# ----------------------------------------------------------------------
	def addIncludeCard(self):
		selranges = self.selection2Ranges()
		if not selranges: return
		if len(selranges)>1:
			self.flair.notify("Cannot add #include",
				"Too many ranges of card(s) selected.\n"
				"Please select only one card or one range",
				tkFlair.NOTIFY_WARNING)
			return
		card_from, card_to = selranges[0]
		undoinfo = [self.flair.refreshUndo()]
		# fake/special tag to avoid triggering an infinite loop
		# addNewCard will insert the correct one
		self.activate(card_to)
		undoinfo.append(self.addNewCard("#endinclude", edit=False, refresh=False))
		self.activate(card_from if card_from==card_to else max(0, card_from-1))
		undoinfo.append(self.addNewCard("# INCLUDE", edit=False, refresh=False))
		undoinfo.append(self.flair.refreshUndo())
		self.addUndo(undoinfo,"Add #include")
		self.active.see()
		#self.refresh()

	# ----------------------------------------------------------------------
	# add cards around the selected ones
	# #if ... #endif
	# ----------------------------------------------------------------------
	def ifendif(self):		self.addStartStop("#if", "#endif")
	def startend_expansion(self):	self.addStartStop("$start_expansion","$end_expansion")
	def startend_translat(self):	self.addStartStop("$start_translat", "$end_translat")
	def startend_transform(self):	self.addStartStop("$start_transform","$end_transform")

	# ----------------------------------------------------------------------
	# Add a new card and return undo information
	# Warning tag will be like mode:tag    fluka:RPP
	# @param tag      new card tag to add as mode:tag
	# @param edit     if true start editing first item after addition
	# @param refresh  if true refresh display and see card
	# ----------------------------------------------------------------------
	def addNewCard(self, tag, edit=True, refresh=True):
		Global.dump("*-* InputExtra.addNewCard",tag)
		if self.active is not None:
			self.commit()
			self.active.editItem()
		elif self.cardlist:
			self.activate(0)

		if isinstance(tag,str):
			# Special treatment
			t = tag.split(":")[-1]	# skip mode
			if t == CardInfo.IFENDIF:
				self.ifendif()
				return
			elif t == CardInfo.EXPANSION:
				self.startend_expansion()
				return
			elif t == CardInfo.TRANSLAT:
				self.startend_translat()
				return
			elif t == CardInfo.TRANSFORM:
				self.startend_transform()
				return
			elif tag == "# INCLUDE":
				tag = "#include"

		self.cardSelectClear()
		project = self.project

		# Create new card
		if isinstance(tag, Card):
			card = tag.clone()
		else:
			card = Card.new(tag, self.input.mode)
			if _INSERT_COMMENT:
				card.setComment("Comment")
			# Special case(s)
			if card.tag == "TITLE":
				card.setText(project.title)

		self.enable()
		# Find location to add card
		if self.active is None:
			# append at the end of file
			cwpos = len(self.cardlist)
			inpos = len(self.input.cardlist)
		elif self.active.card is not None:
			# after active card
			cwpos = self.activePos()+1
			inpos = self.active.card.pos()+1
		else:
			# inside a HiddenCardWidget
			best = project.input.bestPosition(card.tag)
			if best is None or best > self.active.toPos():
				# insert after active card
				Global.dump("*** Below")
				cwpos = self.activePos()+1
				inpos = self.active.toPos()+1
			elif best <= self.active.fromPos():
				# insert before active card
				Global.dump("*** Above")
				cwpos = self.activePos()
				inpos = self.active.fromPos()
			else:
				# inside the HiddenCardWidget
				# split it up to the necessary position
				Global.dump("*** Inside")
				cwpos = self.activePos()+1	 # append
				inpos = best
				to_card = self.active.toCard
				self.active.toCard = self.input.cardlist[inpos-1]
				self.active.redraw()
				self.addCard(cwpos, self.input.cardlist[inpos], to_card)

		#Global.dump("+++ cwpos=",cwpos)
		#Global.dump("+++ inpos=",inpos)
		#Global.dump("+++ card=",card)
		undoinfo = self.flair.addCardUndo(card, inpos, True)
		self.activate()

		cw = self.addCard(cwpos, card)
		self.reposition(cwpos)
		self.disable()
		self.cardSelectClear()
		self.activate(cwpos, True)
		self.active.see()

		if refresh:
			self.update_idletasks()
			cw.see()
			self.focusIn()
			self.page.refreshStatus()
			if tag in ("#define","#undef"):
				self.page.preprocess()
			self.page._refreshOtherFrames()
		if edit:
			self.active.editNext()

		assert self.checkPosition()
		assert self.input.checkNumbering()
		return undoinfo

	# ----------------------------------------------------------------------
	# Add a new card and add also the undo + refresh information
	# ----------------------------------------------------------------------
	def addNewCardUndo(self, tag, edit=True, refresh=True):
		if tag == "#include":
			self.addIncludeCard()
		else:
			undoinfo = [	(self.flair.refreshUndo,),
					self.addNewCard(tag, edit, refresh),
					(self.flair.refreshUndo,)]
			if undoinfo[1]:	# if addNewCard returned undo info
					# if..endif, start..stop do not return anything
				self.addUndo(undoinfo, f"New card {tag}")

	# ----------------------------------------------------------------------
	# Change Card type
	# ----------------------------------------------------------------------
	def changeTag(self, tag):
		if self.active is None or not self.selection: return
		self.commit()
		self.active.editItem()
		undoinfo = [self.flair.refreshUndo()]
		ranges = self.selection2Ranges()

		# FIXME do delete/edit of new card instead of refreshing the display
		for f,t in ranges:
			for cw in self.cardlist[f : t+1]:
				if isinstance(cw,HiddenCardWidget): continue
				for card in cw.cards():
					if f"{card.MODE}:{card.tag}"==tag: continue
					undoinfo.append(self.flair.changeTagUndo(card, tag))
		undoinfo.append(self.flair.refreshUndo())
		#self.cardSelectClear()
		#self.cardSelectFromRanges(ranges)
		self.active.updateLabel()
		self.addUndo(undoinfo, "Change card type")
		self.page.inputModified()

	# ----------------------------------------------------------------------
	# Change card state
	# ----------------------------------------------------------------------
	def changeState(self, enable):
		if self.active is None or not self.selection: return
		undoinfo = []
		for cid in reversed(sorted(self.selection)):
			cw = self.cardlist[cid]
			undoinfo.append(cw.setEnable(enable))
		self.addUndo(undoinfo, "Change state")
		self.active.updateLabel()
		self.page.preprocess()	# FIXME only if #define/#undef is present
		self.page._refreshOtherFrames()

	# ----------------------------------------------------------------------
	def toggleState(self, event=None):
		if self.active is None: return
		card = self.input.cardlist[self.active.fromPos()]
		self.changeState(not card.enable)

	# ----------------------------------------------------------------------
	# Reset interrupt
	# ----------------------------------------------------------------------
	def resetInterrupt(self):
		self._interrupt = False
		return time.time()

	# ----------------------------------------------------------------------
	# Interrupt a long process from notify
	# ----------------------------------------------------------------------
	def interrupt(self):
		self.flair.notifyHide()
		self._interrupt = True

	# ----------------------------------------------------------------------
	# Show/Expand hidden cards
	# ----------------------------------------------------------------------
	#@tool.timeit
	def expandCards(self, event=None):
		if self.active is None or not self.selection: return "break"
		if self.edit is not None:
			self.commit()
			self.active.editItem()
		self.page.busy()
		startTime  = self.resetInterrupt()
		lastUpdate = startTime
		activePos  = self.activePos()
		selection  = []			# new selection
		correction = []			# shift to be applied while collapsing cards
		self.activate()
		self.enable()
		self.layoutCache(True)		# Enable caching of lists
		for i in reversed(sorted(self.selection)):
			cw = self.cardlist[i]
			if isinstance(cw, HiddenCardWidget):
				cards = cw.cards()
				if self.showIcons and len(cards)>2000:
					self.showIcons = False
				selection.append(i)
				correction.append(len(cards)-1)
				self.delCards(i)
				pos = i
				for card in cards:
					ti = time.time()
					if ti-startTime > _CARDS_TIMEOUT:
						if not self.flair.notifyIsVisible():
							self.flair.notify("Too long",
								"Process takes too long",
								tkFlair.NOTIFY_NORMAL,
								"Interrupt?", self.interrupt)
						if self._interrupt:
							cw = self.addCard(pos, card, cards[-1])
							correction[-1] = pos-i
							break
						if ti-lastUpdate > _CARDS_UPDATE:
							lastUpdate = ti
							self.flair.update()
					cw = self.addCard(pos, card)
					pos += 1
			if self._interrupt: break
		self.disable()
		self.layoutCache(False)	# Disable for normal editing

		if selection:
			self.flair.notifyHide()
			self.reposition(selection[-1])
			assert self.checkPosition()

			# Correct selection for shift on subsequent cards
			shift = 0
			for i in range(len(selection)-1, -1, -1):
				if activePos == selection[i]:
					activePos += shift
				selection[i] += shift
				shift += correction[i]

			# Set the selection
			self.selectionClear()
			for f,d in zip(selection, correction):
				self.cardSelect(f,f+d)

		# Set the active card
		self.activate(activePos)
		if self.active: self.active.see()
		self.page.notBusy()
		return "break"

	# ----------------------------------------------------------------------
	# Hide selected cards
	# ----------------------------------------------------------------------
	def collapseCards(self, event=None):
		# remove selected cards from card list and replace them
		# with a HiddenCardWidget
		if self.active is None or not self.selection: return "break"
		if self.edit is not None:
			self.commit()
			self.active.editItem()
		activePos  = self.activePos()
		selection  = []			# new selection
		correction = []			# shift to be applied while collapsing cards

		self.enable()
		for f,t in reversed(self.selection2Ranges()):
			if f==t: continue	# ignore single cards
			if f<=activePos<=t:
				activePos = f
			elif f<activePos:
				activePos -= t-f
			if isinstance(self.cardlist[f], HiddenCardWidget):
				fromCard = self.cardlist[f].fromCard
			else:
				fromCard = self.cardlist[f].card

			if isinstance(self.cardlist[t], HiddenCardWidget):
				toCard = self.cardlist[t].toCard
			else:
				toCard = self.cardlist[t].card

			self.delCardRange(f, t)
			self.addCard(f, fromCard, toCard)

			selection.append(f)
			correction.append(t-f)
		self.disable()

		if selection:
			self.reposition(selection[-1])
			assert self.checkPosition()

			# Correct selection for shift on subsequent cards
			shift = 0
			for i in range(len(selection)-1, -1, -1):
				selection[i] -= shift
				shift += correction[i]

			# Set the selection
			self.selectionClear()
			self.cardSelect(selection)

		# Set the active card
		self.activate(activePos)
		if self.active: self.active.see()
		return "break"

	# ----------------------------------------------------------------------
	# Show/Hide selected cards
	# ----------------------------------------------------------------------
	def toggleShow(self, event=None):
		if not self.selection: return
		first = sorted(self.selection)[0]
		if isinstance(self.cardlist[first], HiddenCardWidget):
			self.expandCards()
		else:
			self.collapseCards()

	# ----------------------------------------------------------------------
	# Edit card in dialog
	# ----------------------------------------------------------------------
	def editCard(self):
		if self.active is None or self.active.card is None: return
		self.commit()
		dialog = InputExtra.EditCardDialog(self, self.flair, self.active.card)
		if dialog.show():
			self.active.updateLayout()
			self.enable()
			self.active.redraw()
			self.disable()
			self.activate(self.active)
			self.reposition(self.activePos())
			self.active.updateLabel()
			self.page._refreshOtherFrames()

	# ----------------------------------------------------------------------
	# Move selected cards up
	# ----------------------------------------------------------------------
	def moveUp(self, event=None):
		if self.active is None: return
		if self.edit is not None:
			self.commit()
			self.active.editItem()
		self.addUndo(self.moveUpCards(self.selection2Ranges()))
		self.cardAnchor = self.activePos()
		self.page.inputModified()
		return "break"

	# ----------------------------------------------------------------------
	# Move selected cards down
	# ----------------------------------------------------------------------
	def moveDown(self, event=None):
		if self.active is None: return
		if self.edit is not None:
			self.commit()
			self.active.editItem()
		self.addUndo(self.moveDownCards(self.selection2Ranges()))
		self.cardAnchor = self.activePos()
		self.page.inputModified()
		return "break"

	# ----------------------------------------------------------------------
	# Move range of cards(fromid,toid) up one card and return undo info
	# ----------------------------------------------------------------------
	def moveUpCards(self, cards):
		if self.active is None: return None
		if cards[0][0]==0: return None		# first element in the list

		low    = 99999999
		high   = 0
		ranges = []

		self.enable()
		for f,t in reversed(cards):
			Global.dump("*-* moveUpCards: range=",f,t)
			# Move downwards the previous cards [fromPos-1] to position toPos+1
			ranges.append((f-1,t-1))
			hidden  = isinstance(self.cardlist[f-1], HiddenCardWidget)
			fromPos = self.cardlist[f-1].fromPos()
			cwcards = self.cardlist[f-1].cards()
			toPos   = self.cardlist[t].toPos()+1
			low     = min(low, fromPos)
			high    = max(high,toPos)
			for _ in cwcards:
				self.input.moveCard(fromPos, toPos, False)
			self.input.renumber(fromPos,toPos)

			# Delete from display the previous card
			self.delCards(f-1)
			self.reposition(f-1, t)	# warning it leaves an incomplete positioning
			if hidden:
				self.addCard(t, cwcards[0], cwcards[-1])
			else:
				self.addCard(t, cwcards[0])

			# re-indent cards
			for i in range(f-1,t):
				self.cardlist[i].indent()

		self.input.renumber(low-1, high)
		assert self.input.checkNumbering()
		assert self.checkPosition()

		self.cardSelectClear()
		self.cardSelectFromRanges(ranges)
		self.disable()
		if self.active is not None: self.active.see()

		self.page._refreshOtherFrames()
		return ("Move up cards", self.moveDownCards, ranges)

	# ----------------------------------------------------------------------
	# Move range of cards(fromid,toid) down one card and return undo info
	# ----------------------------------------------------------------------
	def moveDownCards(self, cards):
		if self.active is None: return None
		if cards[-1][1] == len(self.cardlist)-1: return None

		low    = 99999999
		high   = 0
		ranges = []

		self.enable()
		for f,t in reversed(cards):
			Global.dump("*-* moveDownCards: range=",f,t)
			# Move upwards the next cards [toPos+1] to position fromPos
			ranges.append((f+1,t+1))
			fromPos = self.cardlist[f].fromPos()
			toPos   = self.cardlist[t+1].toPos()
			cwcards = self.cardlist[t+1].cards()
			hidden  = isinstance(self.cardlist[t+1], HiddenCardWidget)
			low     = min(low, fromPos)
			high    = max(high,toPos)

			for _ in cwcards:
				self.input.moveCard(toPos, fromPos, False)
			self.input.renumber(fromPos-1,toPos+1)

			# Delete from display the previous card
			self.delCards(t+1)
			if hidden:
				self.addCard(f, cwcards[0], cwcards[-1])
			else:
				self.addCard(f, cwcards[0])
			self.reposition(f+1, t+2)

			# re-indent cards
			for i in range(f+1,t+2):
				self.cardlist[i].indent()

		self.input.renumber(low, high+1)
		assert self.checkPosition()
		assert self.input.checkNumbering()

		self.cardSelectClear()
		self.cardSelectFromRanges(ranges)
		self.disable()
		if self.active is not None: self.active.see()

		self.page._refreshOtherFrames()
		return ("Move down cards", self.moveUpCards, ranges)

	# ----------------------------------------------------------------------
	# FIND/REPLACE
	# ----------------------------------------------------------------------
	# Search for item
	# @param findStr	string pattern to match
	# @param matchCase	False: case insensitive
	#			True:  exact match
	# @param loop		False: search from index to end only
	#			True:  search all file index->end and start->index
	# @param editOnly	False: search all items editable and/or not
	#			True:  search only editable items
	# @return True is a match is found
	#	Inspect the self.edit if the match is on an editable item
	# ----------------------------------------------------------------------
	def find(self, findStr, matchCase=False, loop=True, editOnly=True):
		if not matchCase: findStr = findStr.upper()
		if self.active and self.edit is not None:
			self.commit()
			self.active.editItem()
		start = start0 = self.index(tk.INSERT)

		# loop to ignore false matches on non-editable items if requested
		while True:
			# Search from index to end
			position = self.search(findStr, start, tk.END,
						regexp=_REGEXP,
						nocase=not matchCase)
			#Global.dump("1. start=",start,"pos=",position, "start0=",start0)
			if not position and loop:
				# Search from beginning to index
				position = self.search(findStr, "1.0", start,
						regexp=_REGEXP,
						nocase=not matchCase)
				#Global.dump("2. start=",start,"pos=",position, "start0=",start0)
				if not position or self.compare(position, ">=", start0):
					return False

			if position:
				cw = self.findCardWidget(position)
				item = cw.item(position)
				self.mark_set(tk.INSERT, f"{position}+{len(findStr)}c")
				if item is None and editOnly:
					start = self.index(tk.INSERT)
					continue
				self.cardSelectClear()
				self.activate(cw)
				self.active.see()
				self.cardSelect(cw)
				if item is not None:
					self.active.editItem(item, select=False)
					self.setAnchor()
					self.tag_add(tk.SEL, position,
							f"{position}+{len(findStr)}c")
				return True
			else:
				return False

	# ----------------------------------------------------------------------
	def replace(self, findStr, replaceStr, matchCase=True):
		Global.dump("*-* InputPage.replace", findStr, replaceStr, matchCase)
		if self.edit is not None:
			# first verify if the sel is on the right location
			try:
				selFirst = self.index(tk.SEL_FIRST)
				selLast  = self.index(tk.SEL_LAST)
				seltxt = self.get(selFirst, selLast)
				if (matchCase and findStr==seltxt) or \
				   (not matchCase and findStr.upper() == seltxt.upper()):
					self.enable()
					selFirst = self.index(tk.SEL_FIRST)
					selLast  = self.index(tk.SEL_LAST)
					# WARNING, first insert and then delete to
					# avoid deleting also the tags
					# FIXME Maybe I should keep all tags and restore them
					self.insert(selLast, replaceStr)
					self.delete(selFirst, selLast)
					self.disable()
					self.commit()
			except tk.TclError:
				pass
		# Find next
		self.find(findStr, matchCase, False, True)
		return self.edit is not None

	# ----------------------------------------------------------------------
	def replaceAll(self, findStr, replaceStr, matchCase=True):
		count = 0
		while self.replace(findStr, replaceStr, matchCase):
			count += 1
		return count

	# ----------------------------------------------------------------------
	# Show card at position pos in input file
	# ----------------------------------------------------------------------
	def goto(self, pos):
		try:
			card = self.input.cardlist[pos]
		except IndexError:
			return
		self.cardSelectClear()
		self.activate(card, True)
		self.active.see()

	# ----------------------------------------------------------------------
	# HELP
	# ----------------------------------------------------------------------
	def help(self):
		if self.active is not None and self.active.card is not None:
			tag = self.active.card.tag
			#if tag == "REGION":
			#	tag = ("8.2.7", "8.6")
			#elif tag == "LATTICE":
			#	tag = ("8.2.10", "8.7")
			#elif tag == "GEOEND":
			#	tag = ("8.2.11","8.8")
			#elif tag == "VOXELS":
			#	tag = ("8.9", "8.3")
			#elif tag in ("$start_expansion","$end_expansion"):
			#	tag = ("8.2.5.1", "8.4")
			#elif tag in ("$start_translat","$end_translat"):
			#	tag = ("8.2.5.2", "8.4")
			#elif tag in ("$start_transform","$end_transform"):
			#	tag = ("8.2.5.3", "8.4")
			#elif tag[0] == "$":
			#	tag = ("8.2.5", "8.4")
			#elif tag == "END":
			#	tag = "8.2.6"
			#elif tag[0] == "!":
			#	tag = "F3"
			Manual.show(self.active.card.MODE, tag)
		else:
			Manual.show("","Input")
		return "break"

	# ----------------------------------------------------------------------
	# For debugging dump cardlist
	# ----------------------------------------------------------------------
	#def printCardList(self):
	#	for i, cw in enumerate(self.cardlist):
	#		if isinstance(cw,HiddenCardWidget):
	#			log.say(i,cw.fromCard.tag,":",cw.fromPos(),"..",cw.toCard.tag,":",cw.toPos())
	#		else:
	#			log.say(i,cw.card.tag,":",cw.fromPos())

	# ----------------------------------------------------------------------
	def dumpActive(self, active=True):
		Global.dump("*-* Dump active card")
		Global.dump(self.active.card)
		#fonts = {}
		tagstack = [{}]
		prefix = ""
		if active:
			start = self.active.index()
			stop  = self.active.indexEnd()+" + 1c"
		else:
			start = "1.0"
			stop  = tk.END
		for tag,value,pos in self.dump(start, stop):
			if tag=="tagoff":
				prefix = prefix[:-1]
			Global.dump(prefix,repr((tag, value, pos)))
			tagstack.append(tagstack[-1].copy())
			if tag=="tagon":
				for n,a,b,c,d in self.tag_config(value).values():
					if a=='' and b=='' and c=='' and d=='': continue
					tagstack[-1][n] = d
				prefix += "\t"
				Global.dump(prefix,tagstack[-1])
		self.checkPosition()

	# ----------------------------------------------------------------------
	# For debugging, check if all cards are in the proper sequence
	# ----------------------------------------------------------------------
	def checkEditor(self):
		Global.dump("*-* checkEditor")
		import re
		self.checkPosition()
		tagactive = defaultdict(bool)
		error = 0

		#---------------------------------------------------------------
		def showerror(msg):
			nonlocal error
			Global.dump(msg)
			error += 1

		for tag,value,pos in self.dump("1.0", tk.END):
			#print(f"tag={tag!r} value={value!r} pos={pos!r}")
			if tag=="tagon":
				tagactive[value] = True
				if tagactive[_ID_TAG] and value != _ID_TAG:
					showerror(f"@{pos} cannot have another tag inside 'id'")
			elif tag=="tagoff":
				tagactive[value] = False
			elif tag=="text":
				active = [x for x,v in tagactive.items() if v]
				if tagactive[_ID_TAG]:
					if re.match(r"^[0-9]+\n?$", value) is None:
						showerror(f"@{pos} Invalid {_ID_TAG} text {value!r}")
					#else:
					#	# FIXME
					#	idx = int(value.strip())
					#	cw = ctypes.cast(idx, ctypes.py_object).value
				elif tagactive[_TAG_TAG]:
					if re.match(r"^ #?[A-Za-z0-9/-]+\t\t$", value) is None:
						showerror(f"@{pos} invalid {_TAG_TAG} text {value!r}")
				elif tagactive[_TEXT_TAG] or \
				     tagactive[_VALUE_TAG] or \
				     tagactive[_HIDDEN_TAG] or \
				     tagactive[_LABEL_TAG]:
					if "\n" in value:
						showerror(f"@{pos} Invalid \\n in value")
				elif tagactive[_COMMENT_TAG]:
					pass
				elif value in ("\t","\n"):
					try: active.remove("active")
					except ValueError: pass
					try: active.remove("select")
					except ValueError: pass
					active = [x for x in active if not x.startswith('indent')]
					if active:
						showerror(f"@{pos} text {value!r} out of context {active}")
				else:
					showerror(f"@{pos} text {value!r} out of context {active}")
			elif tag=="image":
				pass
			elif tag=="mark":
				pass
			else:
				Global.dump(f"Error: invalid tag={tag}")

#===============================================================================
# Input Editor with the scrollbars
#===============================================================================
class InputEditorFrame(tk.Frame):
	def __init__(self, master, page, *args, **kwargs):
		super().__init__(master, *args, **kwargs)
		self.editor = InputEditor(self, page, cursor="xterm")
		self.editor.grid(row=0, column=0, sticky=tk.NSEW)

		sbv = ttk.Scrollbar(self, takefocus=0)
		sbv.grid(row=0, column=1, sticky=tk.NS)

		sbh = tkExtra.AutoScrollbar(self,
					orient=tk.HORIZONTAL,
					takefocus=0)
		sbh.grid(row=1, column=0, sticky=tk.EW)

		self.editor.config( yscrollcommand=sbv.set,
				    xscrollcommand=sbh.set )
		sbv.config(command=self.editor.yview)
		sbh.config(command=self.editor.xview)

		self.grid_rowconfigure(0, weight=1)
		self.grid_columnconfigure(0, weight=1)

		# Place resize tab buttons
		self.tabs = []
		for idx in range(len(_TABPOS)):
			self.tabs.append(
				tk.Label(self.editor,
					text="\N{UP TACK}",
					font=_COMMENT_FONT,
					foreground=_COMMENT_COLOR,
					background="White",
					cursor="arrow",
					anchor=tk.E,
					padx=0,
					pady=1,
					takefocus=False))
			self.tabs[idx].bind("<B1-Motion>",
					lambda e,idx=idx:self.tabsMotion(e,idx))
			tkExtra.Balloon.set(self.tabs[idx],
				"Tab for reposition column. "
				"Hold Shift or Control to make all following tabs equal width")
		self.placeTabs()

	# ----------------------------------------------------------------------
	def placeTabs(self):
		for tab,pos in zip(self.tabs, self.editor._tabs):
			tab.place(x=pos, y=0, anchor=tk.E)

	# ----------------------------------------------------------------------
	def tabsMotion(self, event, idx):
		x = event.x_root - self.editor.winfo_rootx()	# XXX bbox size
		#x1,y1,x2,y2 = self.editor.master.bbox('all')
		if idx>0:
			prev = self.editor._tabs[idx-1]
		else:
			prev = 0
		if x < prev + 32:
			x = prev + 32
		dx = x - self.editor._tabs[idx]
		self.editor._tabs[idx] = x
		if event.state & (STATE.CONTROL | STATE.SHIFT):
			dx = self.editor._tabs[idx]
			if idx>0: dx -= self.editor._tabs[idx-1]
			for i in range(idx+1,len(self.tabs)):
				x += dx
				self.editor._tabs[i] = x
		else:
			for i in range(idx+1,len(self.tabs)):
				self.editor._tabs[i] += dx
		self.placeTabs()
		self.editor.setTabs()

#===============================================================================
# Add Card Group Menu
#===============================================================================
class _AddCardGroupMenu(FlairRibbon.FlairMenu):
	def createMenu(self):
		# FIXME for a single item it should added immediately, no sub-menus selection etc...
		menu = tk.Menu(self.page.ribbon, tearoff=0, activebackground=Ribbon._ACTIVE_COLOR)
		self.page.editor()._addCardMenu(menu, self.page._tags, False, False)
		return menu

#===============================================================================
class _AddCardAllMenu(FlairRibbon.FlairMenu):
	def createMenu(self):
		menu = tk.Menu(self.page.ribbon, tearoff=0, activebackground=Ribbon._ACTIVE_COLOR)
		tkFlair.cardsAllMenu(menu,
				self.page.flair.project.mode(),
				self.page.editor().addNewCardUndo,
				tearoff=1,
				activebackground=Ribbon._ACTIVE_COLOR)

		# All cards in alphabetical order
		menu.add_separator()
		tkFlair.cardsABCMenu(menu,
				self.page.flair.project.mode(),
				self.page.editor().addNewCardUndo,
				activebackground=Ribbon._ACTIVE_COLOR)

		# populate all other modes
		menu.add_separator()
		for mode in sorted(inputs.MODES):
			if mode in ("", self.page.flair.project.mode()): continue
			submenu = tk.Menu(menu,
					tearoff=1,
					activebackground=Ribbon._ACTIVE_COLOR)
			menu.add_cascade(label=mode.capitalize(),
					menu=submenu,
					compound=tk.LEFT,
					image=icons.get(mode),
					underline=0)
			tkFlair.cardsABCMenu(submenu,
					mode,
					self.page.editor().addNewCardUndo,
					activebackground=Ribbon._ACTIVE_COLOR)
		return menu

#===============================================================================
class _AddCardPreprocessorMenu(FlairRibbon.FlairMenu):
	def createMenu(self):
		menu = tk.Menu(self.page.ribbon,
				tearoff=0,
				activebackground=Ribbon._ACTIVE_COLOR)
		tkFlair.cardsGroupMenu(menu,
					self.page.flair.project.mode(),
					CardInfo.PREPROCESSOR_GROUP,
					self.page.editor().addNewCardUndo,
					activebackground=Ribbon._ACTIVE_COLOR)
		return menu

#===============================================================================
class _ChangeCardMenu(FlairRibbon.FlairMenu):
	def createMenu(self):
		menu = tk.Menu(self.page.ribbon,
				tearoff=0,
				activebackground=Ribbon._ACTIVE_COLOR)
		tkFlair.cardsAllMenu(menu,
				self.page.flair.project.mode(),
				self.page.editor().changeTag,
				tearoff=0,
				activebackground=Ribbon._ACTIVE_COLOR)
		return menu

#===============================================================================
# Filter group menu
#===============================================================================
class _FilterGroup(FlairRibbon.FlairMenuGroup):
	#----------------------------------------------------------------------
	def createMenu(self):
		if self.page.editor().active is None: return None
		menu = tk.Menu(self, tearoff=0, activebackground=Ribbon._ACTIVE_COLOR)
		InputExtra.filterMenu(  self.page,
					menu,
					self.page.editor().active.card,
					self.page.findString.get(),
					self.page.findCase.get())
		return menu

#===============================================================================
# Input Tab Page
#===============================================================================
class InputPage(FlairRibbon.FlairPage):
	"""FLUKA input editor."""
	_name_ = "Input"
	_icon_ = "input"

	#-------------------------------------------------------------------------------
	def init(self):
		self.notebook   = None
		self.editors    = []
		self.hash       = 0
		self.tags       = None
		self.filters    = None
		# Remember for diff
		self.labelCard  = None
		self.labelWhat  = None
		self.labelText  = None
		self.labelError = None
		self._mode      = None		# For for which info was generated
		self.refreshOn  = ("card", "run", "project", "mode", "redraw")
		if CardWidget.ICON is None:
			CardWidget.ICON      = icons.get("tag")
			ChildCardWidget.ICON = icons.get("child")

	#----------------------------------------------------------------------
	def createRibbon(self):
		super().createRibbon()

		# ========== Input ==============
		group = Ribbon.LabelGroup(self.ribbon, "Input")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col,row=0,0
		b = Ribbon.LabelButton(group.frame,
				command=self.newInput,
				image=icons.get("new"))
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "New Input from the basic template")

		col,row=1,0
		b = FlairRibbon._TemplateMenuButton(group.frame, self,
				text="New",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "New input from template")

		# ---
		col,row=0,1
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("OPEN"),
				anchor=tk.W,
				command=self.flair.loadInput)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Load existing input")

		# ---
		col,row=1,1
		b = Ribbon.LabelButton(group.frame,
				text="Load",
				anchor=tk.W,
				command=self.flair.loadInput)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Load existing input")

		# ---
		col,row=0,2
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("save"),
				anchor=tk.W,
				command=self.flair.saveProject)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Save input and project")

		col,row=1,2
		b = Ribbon.LabelButton(group.frame,
				text="Save",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.W,
				command=self.flair.saveInputAs)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Save input as. WARNING this saves only the "
					"input and not the project")

		# ---
		col,row=2,0
		menulist = []
		for mode in sorted(inputs.MODES):
			if mode=="": continue
			menulist.append((f"\N{RIGHTWARDS ARROW} {mode.capitalize()}", mode,
					lambda m=mode:self.changeMode(m)))
		self.modeBtn = Ribbon.MenuButton(group.frame, menulist,
					text="Mode \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
					image=icons.get("change"),
					compound=tk.LEFT,
					foreground="DarkBlue",
					anchor=tk.W)
		self.modeBtn.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(self.modeBtn, "Change simulation input file mode")

		# ---
		row += 1
		menulist = [	("Fluka", "fluka",   self.flair.importFluka),
				("Gdml",  "geant",   self.flair.importGdml),
				("Mcnp",  "mcnp",    self.flair.importMcnp),
				("Moira", "moira",   self.flair.importMoira)]
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Import \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("import"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Import input from other formats")

		# ---
		row += 1
		menulist = [	("Fluka",    "fluka",   self.flair.exportFluka),
				("Mcnp",     "mcnp",    self.flair.exportMcnp),
				("Moira",    "moira",   self.flair.exportMoira),
				("OpenSCAD", "openscad",self.flair.exportOpenscad),
				("Povray",   "povray",  self.flair.exportPovray)]
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Export \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("export"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Export input to other formats")

		# ========== Cards ==============
		group = Ribbon.LabelGroup(self.ribbon, "Card")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col,row=0,0
		b = _AddCardGroupMenu(group.frame, self,
					image=icons.get("add32"))
		b.grid(row=row, column=col, rowspan=2, padx=3, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add Card from currently selected group [Ins] [Ctrl-Enter]")

		col,row=0,2
		b = _AddCardAllMenu(group.frame, self,
				text="Add",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add FLUKA Card")

		# ---
		col,row=1,0
		b = _AddCardPreprocessorMenu(group.frame, self,
				text="Preprocessor \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("define"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Add a preprocessor card")

		# ---
		row += 1
		self._materialBtn = Ribbon.LabelButton(group.frame,
				image=icons.get("MATERIAL"),
				text="Material \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				anchor=tk.W,
				command=self.addMaterial)
		self._materialBtn.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self._materialBtn, "Add Material from database")

		# ---
		row += 1
		self.addSearchEntry = tkExtra.LabelEntry(group.frame,
					"Add",
					"DarkGray",
					background="White", #Ribbon._BACKGROUND,
					font=Ribbon._FONT,
					width=12)
		self.addSearchEntry.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self.addSearchEntry , "Search for a new card to be added [a]")
		self.addSearchEntry.bind("<KeyRelease-Return>",   self.addSearchCmd)
		self.addSearchEntry.bind("<KeyRelease-KP_Enter>", self.addSearchCmd)
		self.addSearchEntry.bind("<Button-3>",            self.addSearchCmdClear)
		self.addSearchEntry.bind("<Escape>",              self.focusEditor)

		# ---
		col,row=2,0
		b = Ribbon.LabelButton(group.frame,
				text="Delete",
				image=icons.get("x"),
				command=self.delete,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Delete cards [Del]")

		# ---
		row += 1
		b = _ChangeCardMenu(group.frame, self,
				text="Change \N{BLACK DOWN-POINTING SMALL TRIANGLE}",
				image=icons.get("change"),
				compound=tk.LEFT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Change card to another type")

		# ---
		row += 1
		b = Ribbon.LabelButton(group.frame,
				text="Clone",
				image=icons.get("clone"),
				anchor=tk.W,
				command=self.clone)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Clone cards [Ctrl-D]")

		# ========== Edit ==============
		group = Ribbon.LabelGroup(self.ribbon, "Edit")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()

		# ---
		col,row=0,0
		b = Ribbon.LabelButton(group.frame,
				image=icons.get("toggle32"),
				anchor=tk.CENTER,
				command=self.toggle)
		b.grid(row=row, column=col, rowspan=2, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Change card state (Enable/Disable) [Ctrl-L]")

		col,row=0,2
		menulist = [	("Enable",  "enable",  self.enableCard),
				("Disable", "disable", self.disableCard)]
		b = Ribbon.MenuButton(group.frame, menulist,
				text="State",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Change card state (Enable/Disable) [Ctrl-L]")

		# ---
		col,row=1,0
		b = Ribbon.LabelButton(group.frame, image=icons.get("expand"),
				command=self.toggleShow)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Expand or collapse cards [Ctrl-Double-click] [Ctrl-I]")

		col,row=2,0
		menulist = [	("Show",  "expand", self.expandCards),
				("Hide",  "shrink", self.collapseCards)]
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Show",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT,
				anchor=tk.W)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Expand or collapse cards [Ctrl-Double-click] [Ctrl-I]")

		# ---
		col,row=1,1
		b = Ribbon.LabelButton(group.frame, image=icons.get("comment"),
				command=self.toggleComment)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Toggle cards comment [c or Ctrl-M]")

		col,row=2,1
		menulist = [	("Add Comment",    "comment_add",    self.insComment),
				("Remove Comment", "comment_delete", self.delComment)]
		b = Ribbon.MenuButton(group.frame, menulist,
				text="Comment",
				image=icons.get("triangle_down"),
				compound=tk.RIGHT)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.NSEW)
		tkExtra.Balloon.set(b, "Add/Remove cards comment")

		# ---
		col,row=1,2
		b = Ribbon.LabelButton(group.frame,
				text="Edit Card",
				image=icons.get("edit"),
				command=self.editCard)
		b.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Edit contents of current card [Ctrl-E]")

		# ---
		col,row=3,0
		#f = tk.Frame(group.frame)
		#f.grid(row=row, column=col, rowspan=3, padx=0, pady=0, sticky=tk.NSEW)

		b = Ribbon.LabelButton(group.frame,
				text="Move Up",
				image=icons.get("up"),
				command=self.moveUp)
		#b.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Move up card [Ctrl-Up]")

		# -----
		col,row=3,1
		self.gotoEntry = tkExtra.LabelEntry(group.frame,
					"Goto",
					"DarkGray",
					font=Ribbon._FONT,
					background="White", #Ribbon._BACKGROUND,
					width=8)
		self.gotoEntry.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self.gotoEntry , "Goto specific card index")
		self.gotoEntry.bind("<Return>",   self.goto)
		self.gotoEntry.bind("<KP_Enter>", self.goto)
		self.gotoEntry.bind("<Button-3>", self.gotoEntry.clear)
		self.gotoEntry.bind("<Escape>",   self.focusEditor)

		# ---
		col,row=3,2
		b = Ribbon.LabelButton(group.frame,
				text="Move Down",
				image=icons.get("down"),
				command=self.moveDown)
		#b.pack(side=tk.BOTTOM, expand=tk.YES, fill=tk.BOTH)
		b.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Move down card [Ctrl-Down]")

		# ========== Filter ==============
		group = _FilterGroup(self.ribbon, self, "Filter")
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)
		group.grid3rows()
		group.frame.grid_columnconfigure(0, weight=1)

		# ===
		col,row = 0,0
		self.runCombo = Ribbon.LabelCombobox(group.frame,
					width=24,
					command=self.runChanged)
		self.runCombo.grid(row=row, column=col, columnspan=2, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self.runCombo, "Show only active cards for the selected run")

		b = Ribbon.LabelButton(group.frame, image=icons.get("run"),
				command=self.showActiveCards)
		b.grid(row=row, column=col+2, padx=3, pady=0, sticky=tk.E)
		tkExtra.Balloon.set(b, "Show only active cards for the selected run")

		# ===
		col,row = 0,1
		self.findString = tkExtra.LabelEntry(group.frame,
					"Search",
					tkFlair._LABEL_ENTRY_COLOR,
					font=Ribbon._FONT,
					background="White", #Ribbon._BACKGROUND,
					width=20)
		self.findString.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self.findString , "Find cards with string")
		self.findString.bind("<Return>",   self.find)
		self.findString.bind("<KP_Enter>", self.find)
		self.findString.bind("<Button-3>", self.findString.clear)
		self.findString.bind("<Escape>",   self.focusEditor)

		# ===
		row += 1	# <-- put in focus order the search first
		self.replaceString = tkExtra.LabelEntry(group.frame,
					"Replace",
					tkFlair._LABEL_ENTRY_COLOR,
					font=Ribbon._FONT,
					width=20)
		self.replaceString.grid(row=row, column=col, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(self.replaceString, "Replace with string")
		self.replaceString.bind("<Return>",   self.replace)
		self.replaceString.bind("<KP_Enter>", self.replace)
		self.replaceString.bind("<Button-3>", self.replaceString.clear)
		self.replaceString.bind("<Escape>",   self.focusEditor)

		# ===
		col,row = 0,1	# <--- back to the find
		self.findCase = tk.BooleanVar()
		self.findCase.set(False)
		b = tk.Checkbutton(group.frame,
				variable=self.findCase,
				padx=0,
				pady=0,
				borderwidth=0,
				highlightthickness=0,
				font=Ribbon._FONT,
				background=Ribbon._BACKGROUND,
				activebackground="LightYellow")
		b.grid(row=row, column=col+1, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Tick for case sensitive")

		b = Ribbon.LabelButton(group.frame,
					image=icons.get("find"),
					command=self.find)
		b.grid(row=row, column=col+2, padx=3, pady=0, sticky=tk.E)
		tkExtra.Balloon.set(b, "Find cards with string")

		# ===
		col,row = 0,2
		self.replaceAll = tk.BooleanVar()
		self.replaceAll.set(False)
		b = tk.Checkbutton(group.frame,
				variable=self.replaceAll,
				padx=0,
				pady=0,
				borderwidth=0,
				highlightthickness=0,
				font=Ribbon._FONT,
				background=Ribbon._BACKGROUND,
				activebackground="LightYellow")
		b.grid(row=row, column=col+1, padx=0, pady=0, sticky=tk.EW)
		tkExtra.Balloon.set(b, "Replace all matches")

		b = Ribbon.LabelButton(group.frame, image=icons.get("replace"),
				command=self.replace)
		b.grid(row=row, column=col+2, padx=3, pady=0, sticky=tk.E)
		tkExtra.Balloon.set(b, "Replace with string")

		# ========== Validation ===========
		if Global.experimental:
			group = Ribbon.LabelGroup(self.ribbon, "Validation")
			group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

			# ---
			b = Ribbon.LabelButton(group.frame, text="Validate",
					image=icons.get("viewer32"),
					compound=tk.TOP,
					anchor=tk.W,
					command=self.validate)
			b.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
			tkExtra.Balloon.set(b, "Validate input")

		# ========== View ===========
		group = FlairRibbon.FlairMenuGroup(self.ribbon, self, "View",
			[("Application", "project", self.openInApplication),
			 ("Print",       "print",   self.hardcopy)])
		group.pack(side=tk.LEFT, fill=tk.Y, padx=0, pady=0)

		# ---
		b = Ribbon.LabelButton(group.frame,
				text="Refresh",
				image=icons.get("refresh"),
				command=self.redraw)
		b.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
		tkExtra.Balloon.set(b, "Refresh Cards [Ctrl-R]")

		# ---
		b = Ribbon.LabelButton(group.frame,
				text="Viewer",
				image=icons.get("view"),
				anchor=tk.W,
				command=lambda s=self: s.openInEditor(False))
		b.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
		tkExtra.Balloon.set(b, "Open in internal viewer")

		# ---
		b = Ribbon.LabelButton(group.frame,
				text="Editor",
				image=icons.get("edit"),
				anchor=tk.W,
				command=self.openInEditor)
		b.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
		tkExtra.Balloon.set(b, "Open in external editor")

	#----------------------------------------------------------------------
	# Create Input Editor page
	#----------------------------------------------------------------------
	def createPage(self):
		super().createPage(False)

		# reinitialize some variables
		self.notebook = None
		self.editors  = []
		frame = self.frame

		self.splitter = tkExtra.HSplitter(frame,
					tkFlair.config.getInt(tkFlair._PAGE_SECTION,
						f"{self.name}.split", 130*tkFlair.dpiscale),
						absolute=True,
						sash=True)
		self.splitter.pack(expand=tk.YES, fill=tk.BOTH)
		treeFrame = self.splitter.leftFrame()
		cardFrame = self.splitter.rightFrame()

		# Tree
		self.createInputTreeList()
		self.tree = tkExtra.Treeview(treeFrame, show="tree", style=tkFlair._TREEVIEW_STYLE)
		self.tree.bind("<Escape>", self.focusEditor)
		self.tree.bind("<<TreeviewSelect>>", self.treeSelect)
		#self.tree.bind("<<TreeviewOpen>>", self.treeOpen)
		self.tree.bind("<Escape>", self.focusEditor)
		self.tree.grid(row=0, column=0, sticky=tk.NSEW)

		# Scroll bars
		sb=tkExtra.AutoScrollbar(treeFrame, takefocus=0)
		sb.grid(row=0, column=1, sticky=tk.NS)
		self.tree.configure(yscrollcommand=sb.set)
		sb.configure(command=self.tree.yview)

		sb=tkExtra.AutoScrollbar(treeFrame,
				orient=tk.HORIZONTAL, takefocus=0)
		sb.grid(row=1, column=0, sticky=tk.EW)
		self.tree.configure(xscrollcommand=sb.set)
		sb.configure(command=self.tree.xview)

		treeFrame.grid_rowconfigure(0, weight=1)
		treeFrame.grid_columnconfigure(0, weight=1)

		# Create a fake dynamic tab
		self.notebook = ttk.Notebook(cardFrame, style=tkFlair._NOTEBOOK_STYLE)
		self.notebook.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
		self.notebook.enable_traversal()
		self._tablayout = tkFlair.style.layout(tkFlair._NOTEBOOKTAB_STYLE)

		# Add default editor
		self.editorAdd()

		# Card Interpretation
		self.labelFrame = tk.Frame(cardFrame)
		self.label = tk.Text(self.labelFrame, font=_FIXED_FONT,
				relief=tk.SUNKEN,
				takefocus=0,
				#state=tk.DISABLED,
				height=_CARD_HEIGHT,
				wrap=tk.NONE)
		self.label.bind("<ButtonRelease-1>",self.labelFocus)
		#self.label.bind("<Double-1>",      self.editCard)
		self.label.bind("<Double-3>",       self.awari)
		sbv2 = ttk.Scrollbar(self.labelFrame, takefocus=0)
		sbh2 = tkExtra.AutoScrollbar(self.labelFrame,
					orient=tk.HORIZONTAL, takefocus=0)
		self.labelFrame.grid_columnconfigure(0, weight=1)

		# Configure label
		self.label.tag_configure("card",     foreground="Black")
		self.label.tag_configure("scale",    foreground=_COMMENT_COLOR)
		self.label.tag_configure(_TAG_TAG,   foreground=_TAG_COLOR)
		self.label.tag_configure("diff",     background=_DIFF_COLOR)
		self.label.tag_configure("what",     background=_ACTIVE_COLOR)
		self.label.tag_configure(_ERROR_TAG, foreground=_ERROR_COLOR)

		# To initialize the grid
		self.label.grid(row=0,column=0, sticky=tk.NSEW)
		sbv2.grid(row=0, column=1, sticky=tk.NS)
		sbh2.grid(row=1, column=0, sticky=tk.EW)

		self.label.config( yscrollcommand=sbv2.set,
				   xscrollcommand=sbh2.set )
		sbv2.config(command=self.label.yview)
		sbh2.config(command=self.label.xview)

		if _CARD_HEIGHT>0:
			txt = "\N{BLACK UP-POINTING TRIANGLE}"
		else:
			txt = "\N{BLACK DOWN-POINTING TRIANGLE}"
		self.labelButton = tk.Button(cardFrame, text=txt,
				command=self.labelButtonToggle,
				font=_COMMENT_FONT,
				border=0,
				highlightthickness=0,
				width=2,
				activeforeground=tkFlair._TITLE_FOREGROUND_COLOR,
				takefocus=False,
				padx=1, pady=1)
		self.labelButtonToggle()
		self.labelButton.bind("<Button-1>",  self.labelButtonClicked)
		self.labelButton.bind("<B1-Motion>", self.labelButtonResize)
		tkExtra.Balloon.set(self.labelButton, "Hide/Show/Drag to Resize")

		b = tk.Button(self.label, image=icons.get("edit"),
				command=self.editCard,
				cursor="arrow",
				takefocus=False)
		b.place(relx=1.0, x=-20, rely=0.0)
		tkExtra.Balloon.set(b, "Edit contents of current card [Ctrl-E]")

	# ----------------------------------------------------------------------
	def addTab(self, event=None):
		Global.dump("*-* addTab")
		self.editorAdd()

	# ----------------------------------------------------------------------
	def delTab(self, event=None):
		Global.dump("*-* delTab")
		#self.editorDel()

	# ----------------------------------------------------------------------
	def editorAdd(self):
		#Global.dump("*-* editorAdd", len(self.editors))
		editorFrame = InputEditorFrame(self.notebook, self)
		self.notebook.add(editorFrame,
				text="Input FileName",
				image=icons.get("input"),
				compound=tk.LEFT)
		self.editors.append(editorFrame)
		self.editorAutoHide()
		return editorFrame.editor

	# ----------------------------------------------------------------------
	def editorAutoHide(self):
		# Hack to update the tabs (M.Lange)
		def refreshTabHack():
			info = self.notebook.pack_info()
			px = int(info['padx'])
			del info['padx']
			self.notebook.pack(padx=str(px+1), **info)
			self.notebook.update_idletasks()
			self.notebook.pack(padx=str(px), **info)

		if len(self.editors)==1:
			# Update the tablayout just in case
			self._tablayout = tkFlair.style.layout(tkFlair._NOTEBOOKTAB_STYLE)
			tkFlair.style.layout(tkFlair._NOTEBOOKTAB_STYLE, [])
			refreshTabHack()
		else:
			tkFlair.style.layout(tkFlair._NOTEBOOKTAB_STYLE, self._tablayout)
			refreshTabHack()

	# ----------------------------------------------------------------------
	# Return current active editor
	# ----------------------------------------------------------------------
	def editor(self):
		if self.notebook is None or self.page is None: return None
		tab = self.notebook.index(tk.CURRENT)
		return self.editors[tab].editor

	# ----------------------------------------------------------------------
	# Update current editor name
	# ----------------------------------------------------------------------
	def setInput(self, tab=tk.CURRENT):
		tab = self.notebook.index(tk.CURRENT)
		editor = self.editors[tab].editor
		editor.setInput()
		self.notebook.tab(tab, text=editor.project.name())

	# ----------------------------------------------------------------------
	# Input I/O
	# ----------------------------------------------------------------------
	def newInput(self):
		self.flair.openTemplate(
				tkFlair.inputTemplate(self.flair.project.mode()),
				onlyInput=True)

	# ----------------------------------------------------------------------
	def openTemplate(self, template):
		self.flair.openTemplate(template, onlyInput=True)

	# ----------------------------------------------------------------------
	# Focus
	# ----------------------------------------------------------------------
	def focusEditor(self, event=None):	self.editor().focus_set()
	def focusTree(self, event=None):	self.tree.focus_set()
	def focusAdd(self, event=None):		self.addSearchEntry.focus_set()
	def focusSearch(self, event=None):	self.findString.focus_set()
	def focusGoto(self, event=None):	self.gotoEntry.focus_set()

	#-------------------------------------------------------------------------------
	# generate Input Tree List from CardInfo
	#-------------------------------------------------------------------------------
	def createInputTreeList(self):
		#-----------------------------------------------------------------------
		# special treatment
		#-----------------------------------------------------------------------
		_inputSpecial = {
			"USRBIN"   : self.treeUSRXXX ,
			"USRBDX"   : self.treeUSRXXX ,
			"USRCOLL"  : self.treeUSRXXX ,
			"EVENTBIN" : self.treeUSRXXX ,
			"RESNUCLE" : self.treeUSRXXX ,
			"USRTRACK" : self.treeUSRXXX ,
			"USRYIELD" : self.treeUSRXXX }

		# ----------------------------------------
		# return a subbranch recursively for group
		# ----------------------------------------
		def branch(group):
			lst  = []
			tags = []
			for g in sorted(CardInfo.groups_set):
				if g.startswith(group+"/"): # add a sublist
					subgroup = g[len(group)+1:]	# FIXME until next "/"
					fullgroup,t,l = branch(group+"/"+subgroup)
					if t: lst.append((subgroup, t, l))
					tags.extend(t)

			# scan all cards belonging to a group
			cards = [x for x in CardInfo.cards(self.project.mode())
						if group in x.group ]
			cards.sort(key=attrgetter("order"))
			tags.extend([x.tag for x in cards])
			for c in cards:
				p = _inputSpecial.get(c.tag, None)
				# FIXME do I need that?
				if c.tag=="GEOBEGIN":
					# Special treatment, insert it to top
					lst.insert(0,(c.tag, c.tag, p))
				else:
					lst.append((c.tag, c.tag, p))
			return group, tags, lst

		#---------------------------------------------------------------
		# create a tree structure
		# inputTreeList is a list of list with
		#      [name, card-tags, sublist]
		#---------------------------------------------------------------
		self.inputTreeList = []
		for group in CardInfo.groups:
			if group == "Developers" and not Global.developer: continue
			g,t,l = branch(group)
			if t: self.inputTreeList.append((g, t, l))

	#===============================================================================
	# Tree build up
	#-------------------------------------------------------------------------------
	# default USRxxx cards
	#-------------------------------------------------------------------------------
	def treeUSRXXX(self, item):
		tag = item.split("\t")[1]
		try:
			cardlist = self.project.input.cards[tag]
		except KeyError:
			return

		# scan for the available units and particles
		if tag == "RESNUCLE":
			pwhat = None
			uwhat = 2
		else:
			pwhat = 2
			uwhat = 3

		units = []
		parts = []
		for card in cardlist:
			# Check if it exists
			u = abs(card.intWhat(uwhat))
			if u not in units: units.append(u)

			if pwhat is not None:
				p = card.what(pwhat)
				if p not in parts: parts.append(p)

		# sort units
		icon = icons.get("new")
		units.sort()
		for i in units:
			idname = f"{id(item)}\t{tag}\tia\t{uwhat}\t{i}"
			self.tree.insert(item, tk.END,
					id=idname,
					text=str(i),
					image=icon)

		# particles
		if pwhat is not None:
			icon = icons.get("particle")
			parts.sort()
			for i in parts:
				idname = f"{id(item)}\t{tag}\ts\t{pwhat}\t{i}"
				self.tree.insert(item, tk.END,
						id=idname,
						text=str(i),
						image=icon)

	#-----------------------------------------------------------------------
	# Fill tree with the input list
	#-----------------------------------------------------------------------
	def treeFill(self):
		#---------------------------------------------------------------
		# scan for open items
		#---------------------------------------------------------------
		def _scanOpen(parent=""):
			nonlocal open_
			if self.tree.item(parent,"open"):
				open_.add(parent)
			for item in self.tree.get_children(parent):
				_scanOpen(item)

		#---------------------------------------------------------------
		def _fillBranch(parent, treelist):
			nonlocal open_
			if callable(treelist):
				treelist(parent)
			else:
				for name,tag,sublist in treelist:
					if tag:
						if isinstance(tag,list):
							if all([x not in self.project.input.cards for x in tag]):
								continue
						elif tag not in self.project.input.cards:
							continue

					if isinstance(tag,list):
						icon = icons.get(tag[0], name, "tag")
						tag  = " ".join(tag)
					else:
						icon = icons.get(tag, name, "tag")

					idx = f"{id(treelist)}\t{tag}"
					item = self.tree.insert(parent, tk.END,
									id=idx,
									text=name,
									image=icon,
									open=idx in open_)

					if sublist:
						_fillBranch(item, sublist)

		#---------------------------------------------------------------
		open_ = set()
		_scanOpen()
		selection = self.tree.selection()
		self.tree.delete(*self.tree.get_children())
		parent = self.tree.insert("", tk.END,
						id="/",
						text="Input",
						image=icons.get("input"),
						open=True)
		_fillBranch(parent, self.inputTreeList)
		# HOW TO DISABLE THE SELECTION?
		self._treeSelection = False
		self.tree.selection_set(selection)
		self.tree.update()			# <<<<< FIXME is it dangerous?
		self._treeSelection = True

	#-----------------------------------------------------------------------
	def treeSelect(self, event):
		if not self._treeSelection: return
		tags    = set()
		filters = []
		for item in self.tree.selection():
			#print(f"\t{item=} {self.tree.item(item)}")
			if item=="/":
				self.filterCards(None)
				return
			else:
				_,tag,*flt = item.split("\t")
			for t in tag.split(): tags.add(t)
			if flt:
				f, w, v = flt
				# just to speed up
				if f[0]=='i':
					v = int(v)
				elif f[0]=='f':
					v = float(v)
				try: w = int(w)
				except: pass
				filters.append((f, w, v, True))
		# FIXME do I need to convert tags to string?
		self.filterCards(" ".join(list(tags)), filters)

	#-----------------------------------------------------------------------
	#def treeOpen(self, event):
	#	...
	#	#print("treeOpen")

	# ----------------------------------------------------------------------
	# label
	# ----------------------------------------------------------------------
	def labelButtonClicked(self, event):
		# remember height to check if we changed it
		self._labelHeight = self.label["height"]

	# ----------------------------------------------------------------------
	def labelButtonToggle(self):
		global _CARD_HEIGHT
		# We changed the height in between don't do anything
		try:
			if self._labelHeight != self.label["height"]: return
		except AttributeError:
			pass
		if self.labelButton["text"] == "\N{BLACK DOWN-POINTING TRIANGLE}":
			# Don't be based on that if we have more InputPages open
			_CARD_HEIGHT = -abs(_CARD_HEIGHT)
			self.labelButton["text"] = "\N{BLACK UP-POINTING TRIANGLE}"
			self.labelButton.place(in_=self.splitter.rightFrame(),
					relx=0.5, rely=1.0,
					anchor=tk.S)
			self.labelFrame.pack_forget()
		else:
			_CARD_HEIGHT = abs(_CARD_HEIGHT)
			self.labelFrame.pack(side=tk.BOTTOM, fill=tk.BOTH)
			self.labelButton["text"] = "\N{BLACK DOWN-POINTING TRIANGLE}"
			self.labelButton.place(in_=self.labelFrame,
					relx=0.5, rely=0.0,
					anchor=tk.N)

	# ----------------------------------------------------------------------
	def labelButtonResize(self, event):
		global _CARD_HEIGHT
		if event.y < -3:
			h = self.label["height"] + 1
			if h<15: self.label["height"] = h
		elif event.y > self.labelButton.winfo_height() + 3:
			h = self.label["height"] - 1
			self.label["height"] = max(h,1)
		_CARD_HEIGHT = self.label["height"]

	#-------------------------------------------------------------------------------
	def updateLabel(self, card=None):
		if self.project is None: return
		if card is not None:
			# FIXME if card is the same then do nothing
			cardText = card.toStr()
			if isinstance(card,FlukaCard):
				prop = card.evalWhatStr(FlukaCard.EXPR_OLD)
				if prop:
					prop = prop.replace("!@","        ")
					cardText += f"\n{prop}"
			cardError = None
			if card.invalid:
				try:
					cardError = card.errorMessage(
						self.editor().layout[card.MODE].what2labels(
							self.editor().active.layout))
				except TypeError:
					cardError = "Internal error, cannot find layout " \
							f"for card: {card}"
					log.error(f"cannot find layout for card: {card}")
			if card is self.labelCard and \
			   self.labelText == cardText and \
			   self.labelError == cardError:
				return

		self.label.config(state=tk.NORMAL)
		self.label.delete("1.0", tk.END)

		if card is None:
			self.labelText  = None
			self.labelError = None
			self.labelCard  = card
			self.label.config(state=tk.DISABLED)
			return

		if card.info.meaning:
			self.label.insert(tk.END, f"{card.info.meaning}\n", _COMMENT_TAG)

		# Show errors if needed
		if card.invalid:
			# what to labels conversion
			self.label.insert(tk.END, f"{cardError}\n", _ERROR_TAG)

		# Show alignment scale
		if _SHOW_SCALE and self.label["height"]>1:
			self.label.insert(tk.END, FlukaCard.SCALE, "scale")

		# Insert text and highlight
		self.labelIndex = str(self.label.index(tk.INSERT))
		for i,line in enumerate(cardText.splitlines()):
			self.label.insert(tk.END, f"{line:80}\n", "card")
			if line and (line[0] != ' ' or card.info.compact_mode):
				nlen = len(line.split(maxsplit=1)[0])
				if card.info.compact_mode:
					leading_white_space = len(line) - len(line.lstrip())
					nlen += leading_white_space
				self.label.tag_add(_TAG_TAG,
					f"{self.labelIndex} + {i} lines",
					f"{self.labelIndex} + {i} lines + {nlen} chars")

		# find differences and mark them
		if card is self.labelCard:
			start = None
			stop  = None
			self.labelText = self.labelText.ljust(len(cardText))
			for i in range(min(len(cardText), len(self.labelText))):
				new = cardText[i]
				old = self.labelText[i]
				if new!=old:
					if start is None: start = i
					stop  = i
				else:
					if start is not None:
						istart = f"{self.labelIndex} + {start} chars"
						istop  = f"{self.labelIndex} + {stop+1} chars"
						self.label.tag_add("diff",istart,istop)
						start = None
				if start is not None:
					istart = f"{self.labelIndex} + {start} chars"
					istop  = f"{self.labelIndex} + {stop+1} chars"
					self.label.tag_add("diff",istart,istop)
		else:
			self.labelWhat = None

		self.label.config(state=tk.DISABLED)

		# Save card
		self.labelText  = cardText
		self.labelError = cardError
		self.labelCard  = card

	# ----------------------------------------------------------------------
	# highlight label
	# ----------------------------------------------------------------------
	def highlightLabel(self, what):
		if self.labelCard is None: return
		if what == self.labelWhat: return
		if self.labelCard.isGeo() and \
			self.labelCard.tag not in ("LATTICE","VOXELS"): return

		if isinstance(self.labelCard, FlukaCard):
			self.label.tag_remove("what","1.0",tk.END)
			if what is not None:
				if what == 0:
					self.label.tag_add("what",
						self.labelIndex+" + 70 chars",
						self.labelIndex+" + 1 lines")
				else:
					lines = (what-1) // 6
					if lines>0:
						linesstr = f"+ {lines} lines "
						what = (what-1) % 6 + 1
					else:
						linesstr = ""

					self.label.tag_add("what",
						"%s %s+ %d chars" % (self.labelIndex,linesstr,10*what),
						"%s %s+ %d chars" % (self.labelIndex,linesstr,10*(what+1)))
		self.labelWhat = what

	# ----------------------------------------------------------------------
	# editor hardcopy
	# ----------------------------------------------------------------------
	def hardcopy(self, event=None):
		self.get()
		printer = tkFlair.printer(self.page)
		if not printer: return
		printer.open()
		ps = PSTextExporter(self.editor(), printer, printer.paper, printer.landscape)
		ps.header(f"{self.project.name()}: {self.project.title}")
		ps.comment()
		ps.prolog()
		ps.export()
		ps.close()
		printer.close()
		self.setStatus("Input printed",  "DarkBlue")

	# ----------------------------------------------------------------------
	# delete all
	# ----------------------------------------------------------------------
	def deleteAll(self):
		if self.page is None: return
		self.editor().deleteAll()
		self.updateLabel()

	# ----------------------------------------------------------------------
	# Activate and set focus
	# ----------------------------------------------------------------------
	def activate(self):
		super().activate()
		self.editor().focus_set()

	# ----------------------------------------------------------------------
	# redraw all widgets
	# ----------------------------------------------------------------------
	def redraw(self):
		self.hash = 0
		self.showCards(self.tags, self.filters, False)

	# ----------------------------------------------------------------------
	# Refresh editor
	# ----------------------------------------------------------------------
	def refresh(self, what=None):
		Global.dump("*-* InputPage.refresh", what)
		super().refresh(what)
		if what=="redraw":
			self.redraw()
			return
		if self._mode != self.project.mode():
			self._mode = self.project.mode()
			self.modeBtn["text"]  = self._mode.capitalize() +  _TRIANGLE_DOWN
			self.modeBtn["image"] = icons.get(self._mode)
			self.createInputTreeList()
		self.busy()
		self._fillRuns()
		self.treeRefresh()
		self.editor().setTabs()
		if self.editor().active is not None and self.editor().edit is not None:
			self.editor().commit()
		self.editor().refresh()
		self.notBusy()

	#-----------------------------------------------------------------------
	def treeRefresh(self):
		self.treeFill()

	#-----------------------------------------------------------------------
	# Change mode; update default mode and convert the input file if possible
	#-----------------------------------------------------------------------
	def changeMode(self, mode):
		if self.flair.convert(mode):
			Project.default_mode = mode

	# ----------------------------------------------------------------------
	# Fill the runCombo with active runs from the flair project
	# ----------------------------------------------------------------------
	def _fillRuns(self):
		activeRun = self.runCombo.get()
		self.runCombo.clear()
		self.runCombo.insert(tk.END, _ALL)
		found = activeRun == _ALL
		for run in self.project.runs():
			if run("parent"): continue
			runname = run.inputName2()
			self.runCombo.insert(tk.END, runname)
			if activeRun == runname:
				found = True
		# Set the run in the combo to actually load the project
		if found:
			self.runCombo.set(activeRun)
		else:
			self.runCombo.set(self.runCombo.get(0))

	# ----------------------------------------------------------------------
	# refresh status
	# ----------------------------------------------------------------------
	def refreshStatus(self):
		super().refreshStatus()
		active = self.editor().active
		if active:
			if isinstance(active,HiddenCardWidget):
				num = "%d..%d"%(active.fromCard.pos()+1,
					active.toCard.pos()+1)
			else:
				num = str(active.card.pos()+1)
		else:
			num = "-"
		# Count cards
		ncards = 0
		for cw in self.editor().cardlist:
			if isinstance(cw,CardWidget): ncards += 1

		if not self.editor().selection or len(self.editor().selection)==1:
			selected = ""
		else:
			selected = f"Selected:{len(self.editor().selection)}"

		if ncards != len(self.project.input.cardlist):
			self.setStatus(f"Current:{num} {selected} Displayed:{ncards}"
				f" Total:{len(self.project.input.cardlist)}")
		else:
			self.setStatus(f"Current:{num} {selected}"
				f" Total:{len(self.project.input.cardlist)}")

	# ----------------------------------------------------------------------
	def addMaterial(self):
		if not InputExtra.MaterialList.shown:
			dlg = InputExtra.MaterialList(self.ribbon)
			dlg.fill()
			x = self._materialBtn.winfo_rootx()
			y = self._materialBtn.winfo_rooty() + self._materialBtn.winfo_height()
			mat = dlg.show(x,y)
			if mat:
				MaterialsPage.insert2Input(self.flair, mat)
		else:
			InputExtra.MaterialList.shown.destroy()

	# ----------------------------------------------------------------------
	def addSearchCmd(self, event=None):
		pattern = self.addSearchEntry.get()
		if not pattern and InputExtra.AddSearchList.shown:
			InputExtra.AddSearchList.shown.destroy()
			return
		if not InputExtra.AddSearchList.shown:
			dlg = InputExtra.AddSearchList(self.ribbon)
			dlg.fill(self._mode, pattern)
			x = self.addSearchEntry.winfo_rootx()
			y = self.addSearchEntry.winfo_rooty() + self._materialBtn.winfo_height()
			tag = dlg.show(x,y)
			if tag:
				self.editor().addNewCardUndo(tag)
		else:
			InputExtra.AddSearchList.shown.destroy()

	# ----------------------------------------------------------------------
	def addSearchCmdClear(self, event=None):
		self.addSearchEntry.delete(0,tk.END)
		if InputExtra.AddSearchList.shown:
			InputExtra.AddSearchList.shown.destroy()

	# ----------------------------------------------------------------------
	# Refresh other frames of flair
	# ----------------------------------------------------------------------
	def _refreshOtherFrames(self):
		self.flair.refreshGeometry()

	# ----------------------------------------------------------------------
	# filter card and add card to editor
	# [list-of-OR] = [(f,w,v)...]
	# ----------------------------------------------------------------------
	def filterCard(self, cardlist, card, filters):
		if filters:
			for flt in filters: # list of OR's
				if card.test(*flt):
					cardlist.append(card)
					return
		else:
			cardlist.append(card)

	# ----------------------------------------------------------------------
	# Filter and show cards
	# ----------------------------------------------------------------------
	def filterCards(self, tags=None, filters=None):
		self.editor().setTabs()
		self.showCards(tags, filters)

	# ----------------------------------------------------------------------
	def filterOn(self, f):
		case = self.findCase.get()
		self.showCards(None,[('as',f[0],f[1],case)])

	# ----------------------------------------------------------------------
	# Show project cards
	# tags - tags as words to show
	# filters - commands see main program
	# ----------------------------------------------------------------------
	def showCards(self, tags=None, filters=None, addUndo=True):
		Global.dump("*-* InputPage.showCards: tags=",tags)

		if self.project.input is not self.editor().input:
			# Set the input file to use
			self.setInput()

		# Measure time to show
		startTime = time.time()

		# Validate last item
		if self.editor().active is not None:
			if self.editor().edit is not None:
				self.editor().commit()
				self.editor().active.editItem()
			#self.editor().cardSelectClear()

		inp = self.project.input
		if inp is None:
			self.deleteAll()
			self.label.config(state=tk.NORMAL)
			self.label.delete("1.0", tk.END)
			self.label.config(state=tk.DISABLED)
			return

		# Prepare hash from parameters
		h = hash(self.project.filename) \
		  ^ hash(self.project.input) \
		  ^ hash(tags) \
		  ^ len(self.project.input.cardlist)
		if filters:
			for lst in filters:
				if isinstance(lst,list):
					for (f,w,v) in lst:
						h ^= hash(f) \
						  ^  hash(w) \
						  ^  hash(v)
				else:
					h ^= hash(lst[0]) \
					  ^  hash(lst[1]) \
					  ^  hash(lst[2])

		# Is the same page?
		if self.hash == h: return

		self.busy()
		self.hash    = h
		self.tags    = tags
		self.filters = filters

		# Exclude cards that are not active
		defines = None		# defines to use for preprocessing
		ignore  = False		# ignore non active ones
		run = self.runCombo.get()
		if run != _ALL:
			run = self.project.runs.getRunByName(run)
			if run is not None:
				defines = run.defines()
				ignore = True

		# Pre process file
		errors = self.project.input.preprocess(defines)
		if errors:
			for card,err in errors:
				self.flair.write(f"{log.ERROR}Card #{card.pos1()} {err}\n")
				self.flair.write(f"{log.ERRORCONT}{card}\n")
			self.flair.notify("Pre-process errors",
				"There are several preprocessor errors",
				tkFlair.NOTIFY_WARNING,
				"Output",self.flair.showOutput)

		# Fill with visible cards
		self.editor().layoutCache(True)	# Enable caching of lists
		cardlist = []
		if tags is None:
			self._tags = None
			for card in inp.cardlist:
				if ignore and not card.active: continue
				self.filterCard(cardlist, card, filters)
		else:
			self._tags = tags.split()
			for t in self._tags:
				try:
					for card in inp.cards[t]:
						if ignore and not card.active: continue
						self.filterCard(cardlist, card, filters)
				except KeyError:
					pass

			# Add preprocessor cards unfiltered if needed
			if _SHOW_PREPROCESSOR:
				for t in [x.tag for x in CardInfo.cards(self.project.mode())
						if CardInfo.PREPROCESSOR_GROUP in x.group ]:
					# Is already added?
					if t in tags: continue
					try:
						for card in inp.cards[t]:
							cardlist.append(card)
					except KeyError:
						pass
		# Sort list
		Card.sort(cardlist)

		# Remove preprocessor cards and sub content when everything is hidden
		rmUnwanted = False
		if _SHOW_PREPROCESSOR:
			if tags and tags[0][0]=='#' and "#include" not in self.project.input.cards:
				rmUnwanted = False
			else:
				rmUnwanted = True
		preerror = False
		if rmUnwanted:
			start   = [-1]	# starting position of block
			count   = [0]	# line count of disabled cards
			include = 0	# expanded or hidden include lines
			incinc  = not filters and not tags
			i = 0
			while i < len(cardlist):
				card = cardlist[i]
				tag  = card.tag
				if not card.enable:
					# Increase count
					count[-1] += 1
				elif tag in ("#if", "#ifdef", "#ifndef"):
					# Start new branch
					start.append(i)
					count.append(0)
				elif tag == "#include" and incinc:
					start.append(i)
					count.append(0)
					include += 1
				elif tag in ("#elif", "#else"):
					pass
				elif tag=="#endif" or (tag=="#endinclude" and incinc):
					if len(start)==1:
						# Error in input TOO many endif
						if not preerror:
							self.flair.notify("Preprocessing error",
								f"Card #{i+1} Too many closing #endif",
								tkFlair.NOTIFY_ERROR,
								"Output",self.flair.showOutput)
							preerror = True
					else:
						# check last count
						s = start.pop()
						c = count.pop()
						if tag=="#endinclude":
							include -= 1
							# FIXME this should be treated only for
							# the EXPANDED cards
							# Maybe I have to put a special icon
							if include==0:
								# remove cards
								del cardlist[s+1:i]
								i = s + 2
								# left #include..#endinclude cards
								count[-1] += 2
								continue
						elif c == 0:
							# remove cards
							del cardlist[s:i+1]
							i = s
							continue
						else:
							# keep the card and move count
							count[-1] += c
				else:
					# Increase count
					count[-1] += 1
				i += 1

			if len(start)>1:
				if not preerror:
					err = "Incorrect number of closing #endif or #endinclude"
					self.flair.write(f"{log.ERROR}Card #{card.pos1()} {err}\n")
					self.flair.notify("Preprocessing error",
							err,
							tkFlair.NOTIFY_ERROR,
							"Output",
							self.flair.showOutput)
		self.notBusy()

		# Display cards
		self.editor().displayCards(cardlist)

		# If only one card is selected then activate it and show it
		if self._tags is not None and len(self._tags)==1 and \
		   (self.editor().active is None or
		    self.editor().active.card is None or
		    self.editor().active.card.tag != self._tags[0]):
			for card in cardlist:
				if card.tag == self._tags[0]:
					self.editor().cardSelectClear()
					self.editor().activate(card, True)
					self.editor().active.see()
					break

		del cardlist
		self.editor().layoutCache(False)	# Disable caching of lists

		# Measure time to show
		if Global.profiler:
			dt = time.time() - startTime
			if dt > 1.0:
				self.flair.write(f"InputPage: draw time= {dt:5.3g}s\n")

	# ----------------------------------------------------------------------
	# See active card
	# ----------------------------------------------------------------------
	def seeActive(self):
		if self.editor().active:
			self.editor().active.see()

	# ----------------------------------------------------------------------
	def showInvalid(self):
		cardlist = [ x.card for x in self.editor().cardlist
				if x.card is not None ]
		if not cardlist: self.showCards()
		self.hash = 0	# Force to re-filter the cards
		# Remove invalid cards
		for i in range(len(cardlist)-1,-1,-1):
			card = cardlist[i]
			if not card.invalid:
				if cardlist[i] == self.editor().active:
					self.editor().active = None
				del cardlist[i]
		self.editor().displayCards(cardlist)

	# ----------------------------------------------------------------------
	def afterDock(self):
		self.hash = 0	# Force to re-filter the cards
		if self.page is not None: self.refresh()

	# ----------------------------------------------------------------------
	def togglePreprocessor(self):
		global _SHOW_PREPROCESSOR
		_SHOW_PREPROCESSOR = not _SHOW_PREPROCESSOR
		tkFlair.config.setBool(tkFlair._INPUT_SECTION,
				"showpreprocessor", _SHOW_PREPROCESSOR)
		self.hash = 0	# Force to re-filter the cards
		self.showCards(self.tags, self.filters)

	# ----------------------------------------------------------------------
	# Run has changed show active cards
	# ----------------------------------------------------------------------
	def runChanged(self):
		self.showCards(self.tags, self.filters)
		if self.runCombo.get() != _ALL: self.showActiveCards()

	# ----------------------------------------------------------------------
	# Filter only active cards
	# ----------------------------------------------------------------------
	def showActiveCards(self):
		cardlist = [ x.card for x in self.editor().cardlist
				if x.card is not None ]

		if not cardlist: self.showCards()
		self.hash = 0	# Force to re-filter the cards
		# Exclude cards that are not active
		run = self.runCombo.get()
		if run != _ALL:
			run = self.project.runs.getRunByName(run)
			errors = self.project.input.preprocess(run.defines() if run else None)
			if errors:
				for card,err in errors:
					self.flair.write(f"{log.ERROR}Card #{card.pos1()} {err}\n")
					self.flair.write(f"{log.ERRORCONT}{card}\n")

				self.flair.notify("Pre-process errors",
					"There are several preprocessor errors",
					tkFlair.NOTIFY_WARNING,
					"Output",self.flair.showOutput)

			for i in range(len(cardlist)-1,-1,-1):
				card = cardlist[i]
				#if card.tag[0]!='#' and card.ignore():
				if card.tag[0]!='#' and not card.active:
					if cardlist[i] == self.editor().active:
						self.editor().active = None
					del cardlist[i]
			self.editor().displayCards(cardlist)
		else:
			self.showCards(self.tags, self.filters)

	# ----------------------------------------------------------------------
	# Run input preprocessor based on run in case something has changed
	# on a #define/#undef
	# - added or deleted
	# - changed state enable/disable
	# ----------------------------------------------------------------------
	def preprocess(self):
		run = self.runCombo.get()
		if run != _ALL:
			run = self.project.runs.getRunByName(run)
			self.project.input.preprocess(run.defines() if run else None)
		else:
			self.project.input.preprocess()

	# ======================================================================
	# Undo
	# ======================================================================
	def canUndo(self): return self.flair.canUndo()
	def canRedo(self): return self.flair.canRedo()

	def resetUndo(self):
		self.deleteAll()
		self.flair.undoredo.reset()

	# ----------------------------------------------------------------------
	def undo(self, event=None): self.flair.undoInput()
	def redo(self, event=None): self.flair.redoInput()

	# ----------------------------------------------------------------------
	# Commit changes in editor
	# ----------------------------------------------------------------------
	def commit(self):
		if self.page is not None and self.editor().active is not None:
			self.editor().commit()
			self.editor().active.editItem()

	# ----------------------------------------------------------------------
	# Editor commands
	# ----------------------------------------------------------------------
	# Add a new card or body if we are in the region expression
	# ----------------------------------------------------------------------
	def add(self, event=None):
		active = self.editor().active
		if active and \
		   active.card and \
		   active.card.type()==Card.REGION and \
		   self.editor().edit > 0:
			self.editor().insertBody()
		else:
			self.editor().popupInsertMenu()

	# ----------------------------------------------------------------------
	def delete(self, event=None):
		if self.editor().active is None or not self.editor().selection: return
		self.editor().active.commit()
		self.editor().active.editItem()
		self.editor().delSelectedCards()

	# ----------------------------------------------------------------------
	def toggle(self):		self.editor().toggleState()
	def toggleShow(self,event=None):self.editor().toggleShow()
	def toggleComment(self):	self.editor().toggleComment()
	def ifendif(self):		self.editor().ifendif()
	def selectAll(self):		self.editor().selectAll()
	def selectClear(self):		self.editor().selectClear()
	def insComment(self):		self.editor().insComment()
	def delComment(self):		self.editor().delComment()
	def enableCard(self):		self.editor().changeState(True)
	def disableCard(self):		self.editor().changeState(False)
	def expandCards(self):		self.editor().expandCards()
	def collapseCards(self):	self.editor().collapseCards()
	def moveUp(self):		self.editor().moveUp()
	def moveDown(self):		self.editor().moveDown()

	# ----------------------------------------------------------------------
	def activateIfNeeded(self,card):
		if self.project.input is self.editor().input:
			self.editor().activateIfNeeded(card)

	# ----------------------------------------------------------------------
	def cut(self, event=None):	self.editor().cut(event)

	# ----------------------------------------------------------------------
	def copy(self, event=None):
		focus = self.page.focus_get()
		if focus is self.editor():
			self.editor().copy(event)
		elif focus is self.label:
			self.labelCopy()

	def paste(self, event=None):	self.editor().paste(event)
	def clone(self, event=None):	self.editor().clone(event)
	def editCard(self,event=None):	self.editor().editCard()
	def help(self):			self.editor().help()

	# ----------------------------------------------------------------------
	def labelFocus(self, event=None):
		self.label.focus_set()
		return "break"

	# ----------------------------------------------------------------------
	def labelCopy(self, event=None):
		self.page.clipboard_clear()
		try:
			self.page.clipboard_append(self.label.get("sel.first", "sel.last"))
		except tk.TclError:
			pass
		return "break"

	# ----------------------------------------------------------------------
	# Show card from number in gotoEntry
	# ----------------------------------------------------------------------
	def goto(self, event=None):
		try:
			index = int(self.gotoEntry.get().replace("#",""))
			self.editor().goto(index-1)
		except ValueError:
			self.editor().active.see()

	# ----------------------------------------------------------------------
	def find(self, event=None):
		findStr = self.findString.get()
		if not findStr:
			self.flair.notify("Find empty",
				"Search item is not specified",
				tkFlair.NOTIFY_WARNING)
			return
		case = self.findCase.get()
		if not self.editor().find(findStr, case, editOnly=self.replaceString.get()!=""):
			self.flair.notify("Not Found",
				f"Target \'{findStr}\' not found",
				tkFlair.NOTIFY_WARNING)

	# ----------------------------------------------------------------------
	def replace(self, event=None):
		findStr = self.findString.get()
		if not findStr:
			self.flair.notify("Find empty",
				"Search item is not specified",
				tkFlair.NOTIFY_WARNING)
			return
		replaceStr = self.replaceString.get()
		if not replaceStr:
			self.flair.notify("Replace empty",
				"Search item is not specified",
				tkFlair.NOTIFY_WARNING)
			return

		case = self.findCase.get()

		if self.replaceAll.get():
			self.editor().replaceAll(findStr, replaceStr, case)
			count = self.replaceAll.set(False)
			if count == 0:
				self.flair.notify("Replace All",
					f"Target \'{findStr}\' not found",
					tkFlair.NOTIFY_WARNING)
			else:
				self.flair.notify("Replace All",
					f"'{findStr}' --> '{replaceStr}'\n"
					f"{count} instances replaced",
					tkFlair.NOTIFY_NORMAL)
		else:
			if not self.editor().replace(findStr, replaceStr, case):
				self.flair.notify("Not Found",
					f"Target \'{findStr}\' not found",
					tkFlair.NOTIFY_WARNING)

	# ----------------------------------------------------------------------
	def _saveIfNeeded(self):
		if self.project.filename=="":
			if self.flair.saveProject(): return False
		elif self.project.isModified():
			if tkFlair.askyesno("Save Project",
					"Current project is modified\n"
					"Save before opening?", parent=self.page):
				self.flair.saveProject()
		return True

	# ----------------------------------------------------------------------
	def openInEditor(self, editor=True):
		if not self._saveIfNeeded(): return
		files = [ self.project.filename,
			  self.project.inputName() ]
		if self.project.ismode(Moira):
			files.append(self.project.name() + Moira.EXTMAC)

		if editor:
			tkFlair.editor(*files)
		else:
			self.flair.viewer(*files)

	# ----------------------------------------------------------------------
	def openInApplication(self):
		if not self._saveIfNeeded(): return
		if self.project.ismode(Fluka):
			self.flair.notify("No associated application",
				"FLUKA has no associated visualization application",
				tkFlair.NOTIFY_WARNING)
		elif self.project.ismode(Moira):
			self.flair.application(self.project.inputName())
		elif self.project.ismode(Mcnp):
			self.flair.application("ip",f"i={self.project.inputName()}")
		#elif self.project.ismode(Phits):
		#	self.flair.application("ip",f"i={self.project.inputName()}")

	# ----------------------------------------------------------------------
	def get(self):
		if self.page is None: return
		if self.editor().active is not None:
			self.editor().commit()

	# ----------------------------------------------------------------------
	def inputModified(self):
		self.flair.inputModified()

	# ----------------------------------------------------------------------
#	def transformGeometry(self, event=None):
#		if self.editor().active is None: return
#		self.editor().commit()
#		self.editor().active.editItem()
#		cardlist = []
#		card_from, card_to = self.editor().getSelectRange()
#		if card_from is None: return
#		for c in self.editor().cardlist[card_from : card_to]:
#			if c.card is not None:
#				cardlist.append(c.card)
#			else:
#				cardlist.extend(c.cards())
#		GeometryBody.TransformGeometryDialog(self.page, self.flair, cardlist)

	# ----------------------------------------------------------------------
	# Validate input
	# ----------------------------------------------------------------------
	def validate(self):
		validate = Validate.Validate(self.project, self.flair.log)
		run = self.runCombo.get()
		if run != _ALL:
			run = self.project.runs.getRunByName(run)
		else:
			run = None
		validate.check(run)

	# ----------------------------------------------------------------------
	def toggleSplitter(self, event=None):
		self.splitter.toggle()

	# ----------------------------------------------------------------------
	def awari(self, event):
		iniFil = os.path.join(tkFlair.iniDir,"awari.var")
		imgDir = os.path.join(tkFlair.prgDir,"images")
		awari.startGame(self.master, iniFil, imgDir)

	# ----------------------------------------------------------------------
	def saveConfig(self, config):
		section = tkFlair._INPUT_SECTION
		if self.editor() is not None:
			config.set(section, "tabs",
				" ".join(map(str, self.editor()._tabs)))
		config.setInt(section, "cardheight",  _CARD_HEIGHT)

#-------------------------------------------------------------------------------
def loadConfig(config):
	global _ACTIVE_COLOR, _NORMAL_COLOR, _TAG_COLOR, _DEFINE_COLOR, \
	       _LABEL_COLOR, _VALUE_COLOR, _COMMENT_COLOR, _DISABLE_COLOR, \
	       _DIFF_COLOR, _ERROR_COLOR, _HIDDEN_COLOR, _HIDDEN_BGCOLOR, \
	       _MODE_COLOR, _NAME_COLOR, \
	       _LABEL_FONT, _COMMENT_FONT, _VALUE_FONT, _TAG_FONT, \
	       _FIXED_FONT, _HIDDEN_FONT, _INVALID_COLOR, \
	       _INSERT_COMMENT, _SHOW_SCALE, _SHOW_PREPROCESSOR, _SHOW_LINES, \
	       _BODY_INSERT, _CARDS_TIMEOUT, _REGEXP, _TABPOS, _CARD_HEIGHT

	section= tkFlair._COLOR_SECTION
	_ACTIVE_COLOR  = config.getColor(section,"input.active",   _ACTIVE_COLOR)
	_COMMENT_COLOR = config.getColor(section,"input.comment",  _COMMENT_COLOR)
	_DEFINE_COLOR  = config.getColor(section,"input.define",   _DEFINE_COLOR)
	_DIFF_COLOR    = config.getColor(section,"input.diff",     _DIFF_COLOR)
	_DISABLE_COLOR = config.getColor(section,"input.disable",  _DISABLE_COLOR)
	_ERROR_COLOR   = config.getColor(section,"input.error",    _ERROR_COLOR)
	_HIDDEN_BGCOLOR= config.getColor(section,"input.hiddenbg", _HIDDEN_BGCOLOR)
	_HIDDEN_COLOR  = config.getColor(section,"input.hidden",   _HIDDEN_COLOR)
	_INVALID_COLOR = config.getColor(section,"input.invalid",  _INVALID_COLOR)
	_LABEL_COLOR   = config.getColor(section,"input.label",    _LABEL_COLOR)
	_MODE_COLOR    = config.getColor(section,"input.mode",     _MODE_COLOR)
	_NAME_COLOR    = config.getColor(section,"input.name",     _NAME_COLOR)
	_NORMAL_COLOR  = config.getColor(section,"input.normal",   _NORMAL_COLOR)
	_TAG_COLOR     = config.getColor(section,"input.tag",      _TAG_COLOR)
	_VALUE_COLOR   = config.getColor(section,"input.value",    _VALUE_COLOR)

	section= tkFlair._FONT_SECTION
	_COMMENT_FONT  = config.getFont(section,"input.comment",   _COMMENT_FONT)
	_FIXED_FONT    = config.getFont(section,"input.fixed",     _FIXED_FONT)
	_HIDDEN_FONT   = config.getFont(section,"input.hidden",    _HIDDEN_FONT)
	_LABEL_FONT    = config.getFont(section,"input.label",     _LABEL_FONT)
	_TAG_FONT      = config.getFont(section,"input.tag",       _TAG_FONT)
	_VALUE_FONT    = config.getFont(section,"input.value",     _VALUE_FONT)

	section = tkFlair._INPUT_SECTION
	# Only configurable by hand in the .ini file
	_SHOW_PREPROCESSOR= config.getBool(section, "showpreprocessor",_SHOW_PREPROCESSOR)
	_INSERT_COMMENT   = config.getBool(section, "insertcomment",   _INSERT_COMMENT)
	_SHOW_SCALE       = config.getBool(section, "showscale",       _SHOW_SCALE)
	_SHOW_LINES       = config.getBool(section, "hrlines",         _SHOW_LINES)
	_BODY_INSERT      = config.getBool(section, "bodyinsert",      _BODY_INSERT)
	_REGEXP           = config.getBool(section, "regexp",          _REGEXP)
	_CARD_HEIGHT      = config.getInt( section, "cardheight",      _CARD_HEIGHT)
	tabs              = config.getStr(section,  "tabs",            " ".join(map(str,_TABPOS)))
	_TABPOS = list(map(int, tabs.split()))
	_CARDS_TIMEOUT    = float(config.getInt(section,"cardstimeout",int(_CARDS_TIMEOUT)))
